/*


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end


*/
// LiveCacheaw.cpp : implementation file
//

#include "LcWizard/StdAfx.h"
#include "LcWizard/LiveCache.h"
#include "LcWizard/LiveCacheAw.h" /* PTS 1108379 LiveCacheaw -> LiveCacheAw */
#include "LcWizard/chooser.h"     /* PTS 1108379 Chooser -> chooser */

#ifdef _PSEUDO_DEBUG
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/* Prototypes for local Procedures */
void procConfigureMFC(IConfiguration *pIConfiguration);
void procConfigureMidl(IConfiguration *pIConfiguration, CString	 &strRoot);
void procConfigureCompiler(IConfiguration *pIConfiguration, CString	 &strRoot);
void procConfigureCustomBuild(IConfiguration *pIConfiguration, CString	 &strRoot);
void procConfigureLink(IConfiguration *pIConfiguration, CString	 &strRoot);

// This is called immediately after the custom AppWizard is loaded.  Initialize
//  the state of the custom AppWizard here.
void CLiveCacheAppWiz::InitCustomAppWiz()
{
	// Create a new dialog chooser; CDialogChooser's constructor initializes
	//  its internal array with pointers to the steps.
	m_pChooser = new CDialogChooser;

	// Set the maximum number of steps.
	SetNumberOfSteps(LAST_DLG);

	// TODO: Add any other custom AppWizard-wide initialization here.
	// Inform AppWizard that we're making a DLL.
	m_pApplication = 0;
	m_Dictionary[_T("PROJTYPE_DLL")] = _T("1");

	// create random GUID
	OLECHAR*		oleGUID;
	GUID			guid = GUID_NULL;
	
	// create GUID for TYPELIB
	::CoCreateGuid(&guid);
	StringFromCLSID( guid, &oleGUID ); 
    wcstombs(  m_strLIBCLSID.GetBuffer(wcslen(oleGUID)-2), &oleGUID[1], wcslen(oleGUID)-2 ); 
	m_Dictionary[_T("LIB_CLSID_ODL")] = _T(m_strLIBCLSID.GetBuffer(wcslen(oleGUID)-2));

	// create GUID for ComponentClass
	::CoCreateGuid(&guid);
	StringFromCLSID( guid, &oleGUID ); 
    wcstombs(  m_strAPP_CLSID.GetBuffer(wcslen(oleGUID)-2), &oleGUID[1], wcslen(oleGUID)-2 ); 
	m_Dictionary[_T("APP_CLSID")] = _T(m_strAPP_CLSID.GetBuffer(wcslen(oleGUID)-2));

	// create GUID for INTERFACE
	::CoCreateGuid(&guid);
	StringFromCLSID( guid, &oleGUID ); 
    wcstombs(  m_strDISPIID_CLSID_ODL.GetBuffer(wcslen(oleGUID)-2), &oleGUID[1], wcslen(oleGUID)-2 ); 
	m_Dictionary[_T("DISPIID_CLSID_ODL")] = _T(m_strDISPIID_CLSID_ODL.GetBuffer(wcslen(oleGUID)-2));

	// create UUID for UDT
	::CoCreateGuid(&guid);
	StringFromCLSID( guid, &oleGUID ); 
    wcstombs(  m_strUDT_UUID.GetBuffer(wcslen(oleGUID)-2), &oleGUID[1], wcslen(oleGUID)-2 ); 
	m_Dictionary[_T("UDT_UUID")] = _T(m_strUDT_UUID.GetBuffer(wcslen(oleGUID)-2));
}

// This is called just before the custom AppWizard is unloaded.
void CLiveCacheAppWiz::ExitCustomAppWiz()
{
	// Deallocate memory used for the dialog chooser
	ASSERT(m_pChooser != NULL);
	delete m_pChooser;
	m_pChooser = NULL;

	// TODO: Add code here to deallocate resources used by the custom AppWizard 
}

// This is called when the user clicks "Create..." on the New Project dialog
//  or "Next" on one of the custom AppWizard's steps.
CAppWizStepDlg* CLiveCacheAppWiz::Next(CAppWizStepDlg* pDlg)
{
    // Delegate to the dialog chooser
	return m_pChooser->Next(pDlg);
}

// This is called when the user clicks "Back" on one of the custom
//  AppWizard's steps.
CAppWizStepDlg* CLiveCacheAppWiz::Back(CAppWizStepDlg* pDlg)
{
	// Delegate to the dialog chooser
	return m_pChooser->Back(pDlg);
}

void CLiveCacheAppWiz::CustomizeProject(IBuildProject* pProject)
{
	// TODO: Add code here to customize the project.  If you don't wish
	//  to customize project, you may remove this virtual override.
	
	// This is called immediately after the default Debug and Release
	//  configurations have been created for each platform.  You may customize
	//  existing configurations on this project by using the methods
	//  of IBuildProject and IConfiguration such as AddToolSettings,
	//  RemoveToolSettings, and AddCustomBuildStep. These are documented in
	//  the Developer Studio object model documentation.

	// WARNING!!  IBuildProject and all interfaces you can get from it are OLE
	//  COM interfaces.  You must be careful to release all new interfaces
	//  you acquire.  In accordance with the standard rules of COM, you must
	//  NOT release pProject, unless you explicitly AddRef it, since pProject
	//  is passed as an "in" parameter to this function.  See the documentation
	//  on CCustomAppWiz::CustomizeProject for more information.

	m_Dictionary.Lookup(_T("FULL_DIR_PATH"), m_strDirPath);
	m_Dictionary.Lookup(_T("root"), m_strRoot);
	
    HRESULT			hr;
	long			cnt_configurations;
	IConfigurations *pIConfigurations;
	IConfiguration  *pIConfiguration;
	VARIANT			Idx;
	::VariantInit(&Idx);
	Idx.vt		= VT_I2;
	Idx.iVal	= 0;
	
	hr = pProject->get_Configurations(&pIConfigurations);
	if ( SUCCEEDED(hr) && (pIConfigurations != NULL) ) {
		hr = pIConfigurations->get_Count(&cnt_configurations);
		if ( SUCCEEDED(hr) ) {
			for ( Idx.iVal=1; Idx.iVal <= cnt_configurations; Idx.iVal++ ) {
				hr = pIConfigurations->Item(Idx,&pIConfiguration); 
				if ( SUCCEEDED(hr) ) {
					procConfigureMFC(pIConfiguration);
					procConfigureMidl(pIConfiguration, m_strRoot);
					procConfigureCompiler(pIConfiguration, m_strRoot);
					procConfigureCustomBuild(pIConfiguration, m_strRoot);
					procConfigureLink(pIConfiguration, m_strRoot);
					pIConfiguration->Release();
				}
			}
		}
		pIConfigurations->Release();
	}
}


// Here we define one instance of the CLiveCacheAppWiz class.  You can access
//  m_Dictionary and any other public members of this class through the
//  global LiveCacheaw.
CLiveCacheAppWiz LiveCacheaw;

/* local procedure implementations */
void procConfigureMFC(IConfiguration *pIConfiguration)
{
	HRESULT		hr;
	VARIANT     varReserved;

	::VariantInit(&varReserved);

	/* BASE MFC Configurations */
	hr = pIConfiguration->RemoveToolSettings(L"mfc",L"2",varReserved);
	hr = pIConfiguration->AddToolSettings(L"mfc",L"0",varReserved);
}

void procConfigureMidl(IConfiguration *pIConfiguration, CString	 &strRoot)
{
	HRESULT		hr;
	VARIANT     varReserved;
	wchar_t		wszFN[32];
	wchar_t		wszST[256];
	wchar_t		wszST2[256];
	CString	    cStr;

	::VariantInit(&varReserved);

	cStr = strRoot + ".idl";
	for ( int i=0; i < cStr.GetLength(); i++) wszFN[i] = cStr[i];
	wszFN[i] = 0;

	/* BASE MIDL Configurations */
	cStr = "midl /Oicf /h \"" + strRoot + ".h\" /iid \"" + strRoot + "_i.c\" \"" + strRoot + ".idl\" \n"
	     + "GenInspecCode \"" + strRoot + ".tlb\" \"" + strRoot + "Ispc\" DBPROC";
	for ( i=0; i < cStr.GetLength(); i++) wszST[i] = cStr[i];
	wszST[i] = 0;
    cStr = ".\\" + strRoot + ".tlb \n" +
           ".\\" + strRoot + ".h \n" +
           ".\\" + strRoot + "_i.c \n" +
           ".\\" + strRoot + "Ispc.h \n" +
           ".\\" + strRoot + "Ispc.cpp ";
	for ( i=0; i < cStr.GetLength(); i++) wszST2[i] = cStr[i];
	wszST2[i] = 0;
	hr = pIConfiguration->AddCustomBuildStepToFile(wszFN,wszST,wszST2, 
		                                           L"Performing MIDL Step", varReserved );

};

void procConfigureCompiler(IConfiguration *pIConfiguration, CString	 &strRoot)
{
	HRESULT		hr;
	VARIANT     varReserved;
	BSTR		bstrName;

  ::VariantInit(&varReserved);
  
  hr = pIConfiguration->get_Name(&bstrName);

  if ( wcsstr(bstrName, L"Release") ) {
		//hr = pIConfiguration->RemoveToolSettings(L"cl.exe",L"/MD",varReserved);
    //hr = pIConfiguration->AddToolSettings(L"cl.exe",L"/MT",varReserved);
      hr = pIConfiguration->AddToolSettings(L"cl.exe",L"/MD",varReserved);
    }
    else {
    //hr = pIConfiguration->RemoveToolSettings(L"cl.exe",L"/MDd",varReserved);
    //hr = pIConfiguration->AddToolSettings(L"cl.exe",L"/MTd",varReserved);
      hr = pIConfiguration->AddToolSettings(L"cl.exe",L"/MDd",varReserved);
    }
  
  SysFreeString (bstrName);
}

void procConfigureCustomBuild(IConfiguration *pIConfiguration, CString	 &strRoot)
{
	HRESULT		hr;
	VARIANT   varReserved;

	::VariantInit(&varReserved);

	hr = pIConfiguration->AddCustomBuildStep(
		L"regsvr32 /s /c \"$(TargetPath)\"	echo regsvr32 exec. time > \"$(OutDir)\\regsvr32.trg\" ",
		L"$(OutDir)\\regsvr32.trg",
		L"Performing registration",
		varReserved );
}

void procConfigureLink(IConfiguration *pIConfiguration, CString	 &strRoot)
{
	HRESULT		hr;
	VARIANT   varReserved;
	BSTR		  bstrName;
	wchar_t		wszOUT[64];
	CString	  cStr;
  int i;

  ::VariantInit(&varReserved);
  
  hr = pIConfiguration->get_Name(&bstrName);
  if ( wcsstr(bstrName, L"Release") ) {
    cStr = "/out:.\\Release\\" + strRoot + ".dll";
    for ( i=0; i < cStr.GetLength(); i++) wszOUT[i] = cStr[i];
    wszOUT[i] = 0;
  }
  else {
    cStr = "/out:.\\Debug\\" + strRoot + "d.dll";
    for ( i=0; i < cStr.GetLength(); i++) wszOUT[i] = cStr[i];
    wszOUT[i] = 0;
  }
  hr = pIConfiguration->AddToolSettings(L"link.exe",wszOUT,varReserved);
  hr = pIConfiguration->AddToolSettings(L"link.exe",L"liboms.lib",varReserved);
  
  SysFreeString (bstrName);
}