/*!***************************************************************************

  module      : RTESys_Spinlock.c

  -------------------------------------------------------------------------

  responsible : JoergM

  special area: RunTime
  description : Spinlock C Routines

  last changed: 2001-05-25  14:03
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin LGPL
  Copyright (C) 2001 SAP AG

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  ========== licence end

*****************************************************************************/



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

typedef unsigned char bool;
#include "RunTime/System/RTESys_Spinlock.h"

#if defined(WIN32)
#  include <windows.h>
#endif /* WIN32 */

#if defined(OSF1)
#  include <alpha/builtins.h>
#endif /*OSF1*/

#if defined(HPUX)
#  include "RunTime/RTE_SystemPA.h" /* nocheck */
#endif

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/



/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/



/*===========================================================================*
 *  STRUCTURES, TYPES, UNIONS ...                                            *
 *===========================================================================*/



/*===========================================================================*
 *  STATIC/INLINE FUNCTION PROTOTYPES                                        *
 *===========================================================================*/

#if defined(SUN)
/*SUN*/extern void RTESys_AtomicSwapSun(int *val1, int *val2);
#endif /*SUN*/

#if defined(LINUX) && defined(I386)
  /*
   * function: LINUX_AtomicSwap (LINUX only)
   * description: Does an atomic swap of two integer values
   * 
   * After this call first value and second values are exchanged
   *
   * arguments: val1 [inout] First value
   *            val2 [inout] Second value
   */
/*LINUX && I386*/static void LINUX_AtomicSwap(int *val1, int *val2)
/*LINUX && I386*/{
/*LINUX && I386*/   asm("lock; xchg %0, (%1)"
/*LINUX && I386*/        : "=r"( *val2 ), "=r" ( val1 )
/*LINUX && I386*/        : "0" ( *val2 ), "1"  ( val1 ) );
/*LINUX && I386*/}
#endif /*LINUX*/



/*===========================================================================*
 *  METHODS                                                                  *
 *===========================================================================*/

#if defined(AIX)
/*AIX*/  void RTESys_InitLock( RTE_Lock * lock )
/*AIX*/  {
/*AIX*/    _clear_lock(&(*lock)[0], 0);
/*AIX*/  }
/*AIX*/
/*AIX*/  bool RTESys_TestAndLock( RTE_Lock * lock )
/*AIX*/  {
/*AIX*/    return _check_lock(&(*lock)[0], 0, 1);
/*AIX*/  }
/*AIX*/
/*AIX*/  void RTESys_Unlock( RTE_Lock * lock )
/*AIX*/  {
/*AIX*/    _clear_lock(&(*lock)[0], 0);
/*AIX*/  }
#endif /* AIX */

#if defined(HPUX)
/*HPUX*/ /* PTS 1108388 Somehow this calling sequence does not work while inside liboms */
/*HPUX*/  void RTESys_InitLock( RTE_Lock * lock )
/*HPUX*/  {
/*HPUX*/    volatile SAPDB_Int4 * AlignedPointer = (volatile SAPDB_Int4 *)( ((volatile char *)0) + 
/*HPUX*/                                               ( ( (((char *)lock)-(char *)0) + (sizeof(SAPDB_Int4)*CACHELINE_SIZE - 1) ) 
/*HPUX*/          & (~(sizeof(SAPDB_Int4)*CACHELINE_SIZE - 1)) ) );
/*HPUX*/    *AlignedPointer = 1;
/*HPUX*/  }
/*HPUX*/
/*HPUX*/  bool RTESys_TestAndLock( RTE_Lock * lock )
/*HPUX*/  {
/*HPUX*/    volatile SAPDB_Int4 * AlignedPointer = (volatile SAPDB_Int4 *)( ((volatile char *)0) + 
/*HPUX*/                                               ( ( (((char *)lock)-(char *)0) + (sizeof(SAPDB_Int4)*CACHELINE_SIZE - 1) ) 
/*HPUX*/          & (~(sizeof(SAPDB_Int4)*CACHELINE_SIZE - 1)) ) );
/*HPUX*/    /* Make the cache line "dirty" to improve performance */
/*HPUX*/    /* This statement writes a zero to the high byte      */
/*HPUX*/    *(volatile char *)AlignedPointer = 0;
/*HPUX*/    return (RTE_PATestAndSet((int *)AlignedPointer) == 0);
/*HPUX*/  }
/*HPUX*/    
/*HPUX*/  void RTESys_Unlock( RTE_Lock * lock )
/*HPUX*/  {
/*HPUX*/    volatile SAPDB_Int4 * AlignedPointer = (volatile SAPDB_Int4 *)( ((volatile char *)0) + 
/*HPUX*/                                               ( ( (((char *)lock)-(char *)0) + (sizeof(SAPDB_Int4)*CACHELINE_SIZE - 1) ) 
/*HPUX*/          & (~(sizeof(SAPDB_Int4)*CACHELINE_SIZE - 1)) ) );
/*HPUX*/    *AlignedPointer = 1;
/*HPUX*/  }
#endif /* HPUX */

#if defined(LINUX) && defined(I386)
/*LINUX && I386*/  void RTESys_InitLock( RTE_Lock * lock )
/*LINUX && I386*/  {
/*LINUX && I386*/    int NewValue = 0;
/*LINUX && I386*/    LINUX_AtomicSwap(&(*lock)[0], &NewValue);
/*LINUX && I386*/  }
/*LINUX && I386*/
/*LINUX && I386*/  bool RTESys_TestAndLock( RTE_Lock * lock )
/*LINUX && I386*/  {
/*LINUX && I386*/    int NewValue = 1;
/*LINUX && I386*/    LINUX_AtomicSwap(&(*lock)[0], &NewValue);
/*LINUX && I386*/    return ( NewValue != 0 );
/*LINUX && I386*/  }
/*LINUX && I386*/
/*LINUX && I386*/  void RTESys_Unlock( RTE_Lock * lock )
/*LINUX && I386*/  {
/*LINUX && I386*/    int NewValue = 0;
/*LINUX && I386*/    LINUX_AtomicSwap(&(*lock)[0], &NewValue);
/*LINUX && I386*/  }
#endif /* LINUX && I386 */

#if defined(LINUX) && defined(IA64)
/*LINUX && IA64*/  void RTESys_InitLock( RTE_Lock * lock )
/*LINUX && IA64*/  {
/*LINUX && IA64*/  	int *ptr;
/*LINUX && IA64*/
/*LINUX && IA64*/	ptr = &(*lock)[0];
/*LINUX && IA64*/	*ptr = 0;
/*LINUX && IA64*/  }
/*LINUX && IA64*/
/*LINUX && IA64*/  #if defined __INTEL_COMPILER
/*LINUX && IA64*/  #include <ia64intrin.h>
/*LINUX && IA64*/  #endif
/*LINUX && IA64*/
/*LINUX && IA64*/  bool RTESys_TestAndLock( RTE_Lock * lock )
/*LINUX && IA64*/  {
/*LINUX && IA64*/	int old;
/*LINUX && IA64*/
/*LINUX && IA64*/  #if defined __GNUC__
/*LINUX && IA64*/	__asm__ __volatile__ ("mov ar.ccv = 0 ;;");
/*LINUX && IA64*/	__asm__ __volatile__ ("cmpxchg4.acq %0=%2,%3,ar.ccv" \
/*LINUX && IA64*/			: "=r"(old),"=m"(*(int **)lock) \
/*LINUX && IA64*/			: "m"(*(int **)lock), "r"(1));
/*LINUX && IA64*/  #endif
/*LINUX && IA64*/  #if defined __INTEL_COMPILER
/*LINUX && IA64*/	old = _InterlockedCompareExchange_acq (lock, 1, 0);
/*LINUX && IA64*/  #endif
/*LINUX && IA64*/	return (old != 0);
/*LINUX && IA64*/  }
/*LINUX && IA64*/
/*LINUX && IA64*/  void RTESys_Unlock( RTE_Lock * lock )
/*LINUX && IA64*/  {
/*LINUX && IA64*/  #if defined __GNUC__
/*LINUX && IA64*/	int old;
/*LINUX && IA64*/
/*LINUX && IA64*/	__asm__ __volatile__ ("mov ar.ccv = 1 ;;");
/*LINUX && IA64*/	__asm__ __volatile__ ("cmpxchg4.acq %0=%2,%3,ar.ccv" \
/*LINUX && IA64*/			: "=r"(old),"=m"(*(int **)lock) \
/*LINUX && IA64*/			: "m"(*(int **)lock), "r"(0));
/*LINUX && IA64*/  #endif
/*LINUX && IA64*/  #if defined __INTEL_COMPILER
/*LINUX && IA64*/	(void)_InterlockedCompareExchange_rel (lock, 0, 1);
/*LINUX && IA64*/  #endif
/*LINUX && IA64*/  }
#endif /*LINUX && IA64*/

#if defined(NMP)
/*NMP*/  void RTESys_InitLock( RTE_Lock * lock )
/*NMP*/  {
/*NMP*/    initspin(lock, -1);
/*NMP*/  }
/*NMP*/
/*NMP*/  bool RTESys_TestAndLock( RTE_Lock * lock )
/*NMP*/  {
/*NMP*/    return ( cspinlock(lock) != 0 );
/*NMP*/  }
/*NMP*/
/*NMP*/  void RTESys_Unlock( RTE_Lock * lock )
/*NMP*/  {
/*NMP*/    spinunlock(lock);
/*NMP*/  }
#endif /* NMP */

#if defined(OSF1)
/*OSF1*/  void RTESys_InitLock( RTE_Lock * lock )
/*OSF1*/  {
/*OSF1*/    (*lock)[0] = 0;
/*OSF1*/  }
/*OSF1*/
/*OSF1*/  bool RTESys_TestAndLock( RTE_Lock * lock )
/*OSF1*/  {
/*OSF1*/    return ( __LOCK_LONG_RETRY(&(*lock)[0], 1) == 0 ); /* returns non-zero on lock success */
/*OSF1*/  }
/*OSF1*/
/*OSF1*/  void RTESys_Unlock( RTE_Lock * lock )
/*OSF1*/  {
#pragma message disable unavolacc
/*OSF1*/    __UNLOCK_LONG(&(*lock)[0]);
#pragma message enable unavolacc
/*OSF1*/  }
#endif /* OSF1 */

#if defined(SUN)
/*SUN*/  void RTESys_InitLock( RTE_Lock * lock )
/*SUN*/  {
/*SUN*/    int NewValue = 0;
/*SUN*/    RTESys_AtomicSwapSun(&(*lock)[0], &NewValue);
/*SUN*/  }
/*SUN*/
/*SUN*/  bool RTESys_TestAndLock( RTE_Lock * lock )
/*SUN*/  {
/*SUN*/    int NewValue = 1;
/*SUN*/    RTESys_AtomicSwapSun(&(*lock)[0], &NewValue);
/*SUN*/    return ( NewValue != 0 );
/*SUN*/  }
/*SUN*/
/*SUN*/  void RTESys_Unlock( RTE_Lock * lock )
/*SUN*/  {
/*SUN*/    int NewValue = 0;
/*SUN*/    RTESys_AtomicSwapSun(&(*lock)[0], &NewValue);
/*SUN*/  }
#endif /* SUN */

#if defined(WIN32)
/*WIN32*/ void RTESys_InitLock( RTE_Lock * lock )
/*WIN32*/ {
/*WIN32*/   InterlockedExchange(lock, 0); 
/*WIN32*/ }
/*WIN32*/ 
/*WIN32*/ /*---------------------------------------------------------------------------*/
/*WIN32*/ 
/*WIN32*/ bool RTESys_TestAndLock( RTE_Lock * lock )
/*WIN32*/ {
/*WIN32*/   return ( InterlockedExchange(lock, 1) != 0 ); 
/*WIN32*/ }
/*WIN32*/ 
/*WIN32*/ /*---------------------------------------------------------------------------*/
/*WIN32*/ 
/*WIN32*/ void RTESys_Unlock( RTE_Lock * lock )
/*WIN32*/ {
/*WIN32*/   InterlockedExchange(lock, 0); 
/*WIN32*/ }
#endif /* WIN32 */
/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/
