/*!***************************************************************************

  module      : RTESys_Spinlock.h

  -------------------------------------------------------------------------

  responsible : JoergM

  special area: RunTime
  description : Spinlock C_Functions

  last changed: 2001-05-25  13:57
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin LGPL
  Copyright (C) 2001 SAP AG

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  ========== licence end

*****************************************************************************/


#ifndef RTESYS_SPINLOCK_H
#define RTESYS_SPINLOCK_H



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "RunTime/RTE_Types.h"
#include <sys/types.h>

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!
   Function: RTESys_InitLock
   Description: Initialize a given Spinlock

   Needed before a spinlock can be used,, but not necessarily inside the program.
   If the spinlock i.e. is found in a shared memory segment, it could have been
   initialized outside already.

   Arguments: lock [inout] the lock to initialize
 */
void RTESys_InitLock( RTE_Lock * lock );

/*!
   Function: RTESys_TestAndLock
   Description: Test and lock a given Spinlock

   If locking fails, it is the caller responsibility to react (i.e. by giving up timeslice)

   Arguments: lock [inout] the lock to initialize
   Return value: true if lock was already in use, false if lock was aquiered successfully
 */
bool RTESys_TestAndLock( RTE_Lock * lock );

/*!
   Function: RTESys_Unlock
   Description: Unlock a given Spinlock

   The spinlock is reset into unlocked state if locked or not locked...

   Arguments: lock [inout] the lock to initialize
 */
void RTESys_Unlock( RTE_Lock * lock );

#endif  /* RTESYS_SPINLOCK_H */
