/*


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end


*/

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "geo200.h"

#ifndef _SAPDBErr_MessageList_HPP_
#define _SAPDBErr_MessageList_HPP_

class SAPDBErr_MessageList 
{

  public:

    typedef enum { Error = 1,
                   Info,
                   Warning } MessageType;

    typedef SAPDB_UInt4  MessageID;
    
    SAPDBErr_MessageList( )
    {
      ClearMessageList();
    }
    
    SAPDBErr_MessageList( SAPDB_Char  const * const  Component,
                          SAPDB_Char  const * const  FileName,
                          SAPDB_UInt4 const          LineNumber,
                          MessageType const          Type,
                          MessageID   const          ID,
                          SAPDB_Char  const * const  Message,
                          SAPDB_UInt4 const          NumOfArgs)
    {
      ClearMessageList();

      teo200_EventList newList("", ID, (teo200_EventList::teo200_EventType) Type, Component, Message);
      aEventList = newList;
    }

    SAPDBErr_MessageList (SAPDBErr_MessageList const& MessageList)
    {
      aEventList.eo200_ClearEventList();
      aEventList = MessageList.aEventList;
    }

    void ClearMessageList () 
    {
      aEventList.eo200_ClearEventList();
    }

    bool const     IsEmpty() const 
    { 
      return aEventList.IsEmpty();
    }

     const SAPDB_UInt4    ID() const
     {
      return aEventList.eo200_EventID();
     }

     const MessageType    Type() const
     {
       return (MessageType) aEventList.eo200_EventType();
     }

     SAPDB_Char const * const  Component() const
     {
       return aEventList.eo200_EventLabel();
     }

     SAPDB_Char const * const  Message() const
     {
       return aEventList.eo200_EventMsg();
     }

     SAPDB_UInt4  NumOfMessages() const
     {
        SAPDB_UInt4  nCount = 0;

        const teo200_EventList * pEvent = &aEventList;

        while (pEvent != 0) {
          nCount = pEvent->IsEmpty() ? nCount : nCount + 1;
          pEvent = pEvent->eo200_NextEvent();
        }

        return nCount;
     }

     SAPDBErr_MessageList * const   NextMessage() const
     {
       static SAPDBErr_MessageList aMsgList(aEventList.eo200_NextEvent());
       if (aEventList.eo200_NextEvent() == NULL) {
         return NULL;
       } else {
         return &aMsgList;
       } 
     }

    void   AppendNewMessage ( SAPDBErr_MessageList const &MessageList )
      {
        if (aEventList.IsEmpty()) {
          teo200_EventList newList("", 
                                   MessageList.ID(), 
                                   (teo200_EventList::teo200_EventType) MessageList.Type(), 
                                   MessageList.Component(), 
                                   MessageList.Message());

          aEventList.eo200_ClearEventList();
          aEventList = newList;
        } else {
          teo200_EventList newList(aEventList, "", 
                                   MessageList.ID(), 
                                   (teo200_EventList::teo200_EventType) MessageList.Type(), 
                                   MessageList.Component(), 
                                   MessageList.Message());

          aEventList.eo200_ClearEventList();
          aEventList = newList;
        }

      } 

  private:

    SAPDBErr_MessageList (teo200_EventList * pEventList) 
    {
      aEventList.eo200_ClearEventList();
      if (pEventList != NULL) {
        aEventList = *pEventList;
      }
    } 

    teo200_EventList aEventList;

};

#endif
