/*! 
  -----------------------------------------------------------------------------
 
  module: SAPDBWeb_List.hpp
 
  -----------------------------------------------------------------------------
 
  responsible:  BerndV
 
  special area: universal stuff for web access
   
  description:  a list class like STL list

  version:      7.3 and higher 
  -----------------------------------------------------------------------------
 
                          Copyright by SAP AG, 2000
 
  -----------------------------------------------------------------------------


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/

#ifndef _SAPDBWeb_List_HPP_
#define _SAPDBWeb_List_HPP_

/*
  -----------------------------------------------------------------------------
  macros
  -----------------------------------------------------------------------------
 */
#define SAPDBWEB_LIST_SIZE_STEP 10
#define SAPDBWEB_EMPTY          -1

/*
  -----------------------------------------------------------------------------
  Class:        SAPDBWeb_List
  -----------------------------------------------------------------------------

  description:  List

  -----------------------------------------------------------------------------
 */
template <class E>
class SAPDBWeb_List
{
public:
  SAPDBWeb_List            ( );
  ~SAPDBWeb_List           ( );

  void      push_back      ( const E & aElement );
  bool      empty          ( ) {return (m_nCurrent == SAPDBWEB_EMPTY);}
  typedef E * iterator;
  iterator  begin          ( ) {return &( m_theList[0]);}
  iterator  end            ( ) {return &(m_theList[m_nCurrent + 1]); }

private:

  E                 *  m_theList;
  int                  m_nCurrent;
  int                  m_nSize;

};

/* EndClass: SAPDBWeb_List */

/*
  =============================================================================
  implementation 
  =============================================================================
 */


/*
  -----------------------------------------------------------------------------
  public constructor SAPDBWeb_List :: SAPDBWeb_List
  -----------------------------------------------------------------------------
 */
template <class E>
inline SAPDBWeb_List<E> :: SAPDBWeb_List ( )
{
  m_nSize    = SAPDBWEB_LIST_SIZE_STEP;
  m_nCurrent = SAPDBWEB_EMPTY;

  m_theList = new E[m_nSize];

  for (int nIndex = 0;nIndex < m_nSize;++nIndex) {
    m_theList[nIndex] = 0;
  } // end for

} // end SAPDBWeb_List :: SAPDBWeb_List

/*
  -----------------------------------------------------------------------------
  public destructor SAPDBWeb_List :: ~SAPDBWeb_List
  -----------------------------------------------------------------------------
 */
template <class E>
inline SAPDBWeb_List<E> :: ~SAPDBWeb_List ( )
{
  if (m_theList != 0) {
    delete [] m_theList;
    m_theList = 0;
  } // end if
} // end SAPDBWeb_List :: ~SAPDBWeb_List

/*
  -----------------------------------------------------------------------------
  public function SAPDBWeb_List :: push_back
  -----------------------------------------------------------------------------
 */
template <class E>
inline void SAPDBWeb_List<E> :: push_back ( const E & aElement )
{
  int nIndex = 0;

  if ((m_nSize - 1) == m_nCurrent) {
    m_nSize =  m_nSize + SAPDBWEB_LIST_SIZE_STEP + 1;

    E *  m_newList = new E[m_nSize];

    for (nIndex = 0;nIndex < (m_nSize - SAPDBWEB_LIST_SIZE_STEP);++nIndex) {
      m_newList[nIndex] = m_theList[nIndex];
    } // end for
    for (nIndex = (m_nSize - SAPDBWEB_LIST_SIZE_STEP);nIndex < m_nSize;++nIndex) {
      m_newList[nIndex] = 0;
    } // end for

    delete [] m_theList;
    m_theList = m_newList;
  } // end if

  ++m_nCurrent;

  m_theList[m_nCurrent] = aElement;

} // end SAPDBWeb_List :: push_back

#endif // _SAPDBWeb_List_HPP_
