/*! 
  -----------------------------------------------------------------------------
  module: SAPDBWeb_Template.cpp
  -----------------------------------------------------------------------------
  responsible:  BerndV
  special area: universal stuff for web access
  description:  a class for creating simple Tables in SAP MiniApps style
  version:      7.3 and higher 
  -----------------------------------------------------------------------------
                          Copyright by SAP AG, 2000
  -----------------------------------------------------------------------------


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/

/*
  -----------------------------------------------------------------------------
  includes
  -----------------------------------------------------------------------------
 */
//#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "heo06.h"
#include "SAPDBCommon/Web/SAPDBWeb_Template.hpp"


/*
  -----------------------------------------------------------------------------
  private macros
  -----------------------------------------------------------------------------
 */
#define COMMENT_BEGIN           "<!--"
#define COMMENT_END             " -->"
#define TEMPLATE_BEGIN          "TemplateKey"
#define TEMPLATE_END            "EndTemplateKey"
#define VALUE_MARKER            "%"

#define VALLIST_CHAR            '*'

#define VALLIST_VALMARKER_BEGIN "%*"
#define VALLIST_VALMARKER_END   "*%"

/*    
  =============================================================================
  class:        SAPDBWeb_TemplateValue
  =============================================================================
 */
SAPDBWeb_TemplateValue :: SAPDBWeb_TemplateValue ( const SAPDB_Char * szName,
                                                   const SAPDB_Char * pBegin,
                                                   const SAPDB_Char * pEnd )
{
  m_strName.assign(szName);
  m_pBegin = pBegin;
  m_pEnd = pEnd;  
}

SAPDBWeb_TemplateValue :: ~SAPDBWeb_TemplateValue ( )
{
}

/*    
  =============================================================================
  class:        SAPDBWeb_Template
  =============================================================================
 */

/*
  -----------------------------------------------------------------------------
  public constructor SAPDBWeb_Template :: SAPDBWeb_Template
  -----------------------------------------------------------------------------
 */
SAPDBWeb_Template  :: SAPDBWeb_Template (       sapdbwa_WebAgent & wa,
                                          const SAPDB_Char       * szFile ) 
{
  m_pFileBuffer       = NULL;
  m_pMaster           = this;
  m_bError            = false;

  SAPDB_Char   szPath[256] = "";

  wa.GetDocumentRoot(szPath, 200); 

  strcat(szPath,"/");
  strcat(szPath, SAPDBWA_DOCUMENT_ROOT_HTML);
  strcat(szPath,"/");
  strcat(szPath, szFile);

  readFile(szPath);

} // end SAPDBWeb_Template :: SAPDBWeb_Template
/*
  -----------------------------------------------------------------------------
 */
SAPDBWeb_Template :: SAPDBWeb_Template (       SAPDBWeb_Template * pMaster,
                                         const SAPDB_Char        * szName,
                                         const SAPDB_Char        * pBegin, 
                                         const SAPDB_Char        * pEnd,
                                         const SAPDB_Bool          bValueList)
{
  m_pFileBuffer       = NULL;
  m_pMaster           = pMaster;
  m_bError            = false;

  analyzeTemplate(szName, pBegin, pEnd, bValueList);

} // end SAPDBWeb_Template :: SAPDBWeb_Template

/*
  -----------------------------------------------------------------------------
  public destructor SAPDBWeb_Template  :: ~SAPDBWeb_Template
  -----------------------------------------------------------------------------
 */
SAPDBWeb_Template  :: ~SAPDBWeb_Template ()
{
  // delete objects in the value list
  HTMLTemplateValueList::iterator  aValueListIterator = m_TemplateValueList.begin();
  while (aValueListIterator != m_TemplateValueList.end()) {
    delete (*aValueListIterator);
    (*aValueListIterator) = NULL;
    ++aValueListIterator;
  } // end while

  // delete objects in the template list
  HTMLTemplateList::iterator  aTemplateListIterator = m_TemplateList.begin();
  while (aTemplateListIterator != m_TemplateList.end()) {
    delete (*aTemplateListIterator);
    (*aTemplateListIterator) = NULL;
    ++aTemplateListIterator;
  } // end while

  // delete the filebuffer
  if (m_pFileBuffer != NULL) {
    delete [] m_pFileBuffer;
    m_pFileBuffer = NULL;
  } // end if
} // end SAPDBWeb_Template  :: ~SAPDBWeb_Template

/*
  -----------------------------------------------------------------------------
  public function SAPDBWeb_Template :: writePage
  -----------------------------------------------------------------------------
 */
void SAPDBWeb_Template :: writePage ( const SAPDBWeb_TemplateWriter & aWriter,
                                      bool                      bWithHeader)
{
  if (m_bError) {
    if (bWithHeader) {
      aWriter.sendHeader();
    } // end if

    aWriter.writeBuffer("<html><head><title>SAPDBWeb_Template - Error</title></head><body>");
    aWriter.writeBuffer(m_strError.c_str());
    aWriter.writeBuffer("</body></html>");
  } else {
  
    SAPDB_Int2 nWriteCount = 0;
    SAPDB_Int2 nIndex      = 0;
  
    if ( initTemplate( ) ) {

      if (bWithHeader) {
        aWriter.sendHeader();
      } // end if

      if (m_strName.empty()) {
        nWriteCount = 1;
      } else {
        nWriteCount = m_pMaster->askForWriteCount(m_strName.c_str());
      } // end if

      if (nWriteCount >= 0) {
        for (nIndex = 0; nIndex < nWriteCount; ++nIndex) {
          if (m_TemplateList.empty()) {
            if (!m_bValueList) {
              writeValue(aWriter);
            } else {
              writeValueList(aWriter);
            }
          } else {
            writeTemplates(aWriter);
          } // end if
        } // end for
      } else {
        nIndex = 0;
        while (m_pMaster->askForContinue(m_strName.c_str())) {
          if (m_TemplateList.empty()) {
            writeValue(aWriter);
          } else {
            writeTemplates(aWriter);
          } // end if
          nIndex++;
        } // end while
      } // end if

    } // end if

  } // end if

} // end SAPDBWeb_Template :: writePage

/*
  -----------------------------------------------------------------------------
  public function SAPDBWeb_Template :: writeTemplates
  -----------------------------------------------------------------------------
 */
void SAPDBWeb_Template :: writeTemplates ( const SAPDBWeb_TemplateWriter & aWriter )
{
  const SAPDB_Char * pCurrent = m_pBegin;

  // iterate the trough the objects list
  HTMLTemplateList::iterator  aListIterator = m_TemplateList.begin();

  while (aListIterator != m_TemplateList.end()) {

    // write before the next template / respectively after the last template
    if (((*aListIterator)->objectBegin() - pCurrent) > 0) {
      aWriter.writeBuffer(pCurrent, (*aListIterator)->objectBegin() - pCurrent);
    } // end if

    // write the template
    (*aListIterator)->writePage(aWriter, false);
    
    pCurrent = (*aListIterator)->objectEnd();

    ++aListIterator;
  } // end while

  // write after the last template
  if ((m_pEnd - pCurrent) > 0) {
    aWriter.writeBuffer(pCurrent, m_pEnd - pCurrent);
  } // end if

} // end SAPDBWeb_Template :: writeTemplates

/*
  -----------------------------------------------------------------------------
  public function SAPDBWeb_Template :: writeValue
  -----------------------------------------------------------------------------
 */
void SAPDBWeb_Template :: writeValue ( const SAPDBWeb_TemplateWriter & aWriter )
{
  const SAPDB_Char * pCurrent = m_pBegin;

  if (m_pBeginOfValue != NULL) {
    // write before the value
    if ((m_pBeginOfValue - m_pBegin) > 0) {
      aWriter.writeBuffer(m_pBegin, m_pBeginOfValue - m_pBegin);
    } // end if

    // write the value
    aWriter.writeBuffer(m_pMaster->askForValue(m_strName.c_str()).c_str());

    // write after the value
    if ((m_pEnd - m_pEndOfValue) > 0) {
      aWriter.writeBuffer(m_pEndOfValue, m_pEnd - m_pEndOfValue);
    } // end if
  } else {
    if ((m_pEnd - m_pBegin) > 0) {
      aWriter.writeBuffer(m_pBegin, m_pEnd - m_pBegin);
    } // end if
  } // end if

} // end SAPDBWeb_Template :: writeValue

/*
  -----------------------------------------------------------------------------
  public function SAPDBWeb_Template :: writeValueList
  -----------------------------------------------------------------------------
 */
void SAPDBWeb_Template :: writeValueList ( const SAPDBWeb_TemplateWriter & aWriter )
{
  const SAPDB_Char * pCurrent = m_pBegin;

  // iterate the trough the value list
  HTMLTemplateValueList::iterator  aListIterator = m_TemplateValueList.begin();

  while (aListIterator != m_TemplateValueList.end()) {

    // write before the next template / respectively after the last template
    if (((*aListIterator)->objectBegin() - pCurrent) > 0) {
      aWriter.writeBuffer(pCurrent, (*aListIterator)->objectBegin() - pCurrent);
    } // end if

    // write the value
    aWriter.writeBuffer(m_pMaster->askForValue((*aListIterator)->objectName()).c_str());

    pCurrent = (*aListIterator)->objectEnd();

    ++aListIterator;
  } // end while

  // write after the last value
  if ((m_pEnd - pCurrent) > 0) {
    aWriter.writeBuffer(pCurrent, m_pEnd - pCurrent);
  } // end if

} // end SAPDBWeb_Template :: writeValueList

/*
  -----------------------------------------------------------------------------
  private function SAPDBWeb_Template :: readFile
  -----------------------------------------------------------------------------
 */
void SAPDBWeb_Template :: readFile ( const SAPDB_Char * szFile )
{
  bool               bState  = false;
  SAPDB_Int4         hFile;
  tsp5_rte_fileerror fileError;
  tsp5_rte_fileerror closeError;
  tsp5_rte_fileinfo  fileInfo;
  tsp_longint        outLen;

  if (m_pFileBuffer != NULL) {
    delete [] m_pFileBuffer;
    m_pFileBuffer = NULL;
  } // end if

  // determine file size
  sqlfinfoc (szFile, &fileInfo, &fileError);
  
  if (fileError.sp5fe_result == vf_ok ) {
    // allocate buffer
    m_pFileBuffer = new SAPDB_Char[fileInfo.sp5fi_size + 1];

    if (m_pFileBuffer != NULL) {
     m_pFileBuffer[0] = '\0';

     // open template file
     sqlfopenc (szFile, SP5VF_BINARY, SP5VF_READ, SP5BK_BUFFERED, &hFile, &fileError);

      if (fileError.sp5fe_result == vf_ok ) {
        // read file
        sqlfreadc (hFile, m_pFileBuffer, fileInfo.sp5fi_size, &outLen, &fileError);
        if (fileError.sp5fe_result == vf_ok ) {
          bState = true;
        } // end if

        // close file
        sqlfclosec (hFile, SP5VF_CLOSE_NORMAL, &closeError);
      } // end if

    } // end if

  } // end if

  if (bState) {
    m_pFileBuffer[fileInfo.sp5fi_size] = '\0';
    analyzeTemplate("", m_pFileBuffer, &m_pFileBuffer[fileInfo.sp5fi_size], SAPDB_FALSE);
  } else {
    if (m_pFileBuffer != NULL) {
      delete [] m_pFileBuffer;
      m_pFileBuffer = NULL;
    } // end if
    m_bError = true;

    m_strError.assign("Error reading file <b>\"")
              .append(szFile)
              .append("\"</b>! (")
              .append(fileError.sp5fe_text)
              .append(")");
  } // end if

} // end SAPDBWeb_Template :: readFile


/*
  -----------------------------------------------------------------------------
  private function SAPDBWeb_Template :: analyzeTemplate
  -----------------------------------------------------------------------------
 */
void SAPDBWeb_Template :: analyzeTemplate ( const SAPDB_Char * szName,
                                            const SAPDB_Char * pBegin, 
                                            const SAPDB_Char * pEnd,
                                            const SAPDB_Bool   bValueList )
{

  m_strName.assign(szName);

  m_pBegin         = pBegin;
  m_pEnd           = pEnd;  
  m_pBeginOfValue  = NULL;
  m_pEndOfValue    = NULL;

  const SAPDB_Char * pBeginOfTemplate = NULL;
  const SAPDB_Char * pEndOfTemplate   = NULL;
  int                nCounter         = 0;
  SAPDBWeb_String    strTemplate;
  SAPDB_Bool         bTemplValList = SAPDB_FALSE;

  pBeginOfTemplate = findNextTemplate(m_pBegin, m_pEnd, strTemplate, &bTemplValList);

  // read nested objects
  while (pBeginOfTemplate != NULL && pBeginOfTemplate <= m_pEnd) {
    pEndOfTemplate = findEndOfTemplate(pBeginOfTemplate, m_pEnd, strTemplate.c_str());
    if (pEndOfTemplate != NULL && pEndOfTemplate  <= m_pEnd) {
      SAPDBWeb_Template * aNewTemplate = new SAPDBWeb_Template(m_pMaster, strTemplate.c_str(), pBeginOfTemplate, pEndOfTemplate, bTemplValList);

      if (aNewTemplate != NULL) {
        m_TemplateList.push_back(aNewTemplate);
        nCounter++;
      } // end if
    } else {
      m_bError   = true;
      m_strError.assign("Error end-of-template <b>\"")
                .append(strTemplate)
                .append("\"</b> not found!");
    } // end if

    pBeginOfTemplate = findNextTemplate(pEndOfTemplate, m_pEnd, strTemplate, &bTemplValList);
  } // end while

  if (m_TemplateList.empty()) {

    m_bValueList = bValueList;

    if (!bValueList) {
      // look for value
      SAPDBWeb_String strSearch;

      strSearch.assign(VALUE_MARKER)
               .append(szName)
               .append(VALUE_MARKER);

      m_pBeginOfValue = strstr(pBegin, strSearch.c_str());
      if (m_pBeginOfValue != NULL) {
        m_pEndOfValue = m_pBeginOfValue + strSearch.length();
      } // end if
    } else {
      // look for values
      const SAPDB_Char * pCurrent = NULL;
      const SAPDB_Char * pBeginOfValue = NULL;
      const SAPDB_Char * pEndOfValue   = NULL;
      SAPDBWeb_String    strValue;

      pCurrent = m_pBegin;
      while ( pCurrent != NULL ) {

        pCurrent = findNextValue(pCurrent, m_pEnd, &pBeginOfValue, &pEndOfValue, strValue);

        if ( pBeginOfValue != NULL && pEndOfValue != NULL ) {
          SAPDBWeb_TemplateValue * aNewTemplateValue = new SAPDBWeb_TemplateValue(strValue.c_str(), pBeginOfValue, pEndOfValue);

          if (aNewTemplateValue != NULL) {
            m_TemplateValueList.push_back(aNewTemplateValue);
          } // end if
        } // end if
      
      } // end while

    } // end if
    
  } // end if

} // end SAPDBWeb_Template :: analyzeTemplate

/*
  -----------------------------------------------------------------------------
  private function SAPDBWeb_Template :: findNextTemplate
  -----------------------------------------------------------------------------
 */
const SAPDB_Char * SAPDBWeb_Template :: findNextTemplate (const SAPDB_Char * pBuffer,
                                                          const SAPDB_Char * pEnd,
                                                          SAPDBWeb_String  & strTemplate,
                                                          SAPDB_Bool       * bTemplValList)
{
   
  const SAPDB_Char  *  pReturn       = NULL;
  const SAPDB_Char  *  pCurrent      = NULL;
  const SAPDB_Char  *  pBeginOfToken = NULL;
  const SAPDB_Char  *  pEndOfToken   = NULL;
  SAPDBWeb_String      strToken;

  if (pBuffer != NULL) {

    // look for comment
    pCurrent = strstr(pBuffer, COMMENT_BEGIN);

    while (pCurrent != NULL && pReturn == NULL && pCurrent <= pEnd) {
      // copy pointer
      pBeginOfToken = pCurrent;
      // skip comment
      pBeginOfToken = pBeginOfToken + strlen(COMMENT_BEGIN);
      // skip whitescpace
      while (isspace(*pBeginOfToken) && *pBeginOfToken != '\0') ++pBeginOfToken;
      // search end of token
      pEndOfToken = pBeginOfToken;
      while (!isspace(*pEndOfToken) && *pEndOfToken != '\0') ++pEndOfToken;
      // extract token
      strToken.assign(pBeginOfToken, pEndOfToken - pBeginOfToken);
      // look for keyword
      if (strToken.compare(TEMPLATE_BEGIN) == 0) {
        pBeginOfToken = pEndOfToken;
        // skip whitescpace
        while (isspace(*pBeginOfToken) && *pBeginOfToken != '\0') ++pBeginOfToken;
        // search end of token
        pEndOfToken = pBeginOfToken;
        while (!isspace(*pEndOfToken) && *pEndOfToken != '\0') ++pEndOfToken;
        // Extract Name of Template
        strTemplate.assign(pBeginOfToken, pEndOfToken - pBeginOfToken);
        // value list ?
        if (strTemplate[strTemplate.length() - 1] == VALLIST_CHAR) {
          *bTemplValList = SAPDB_TRUE;
        } else {
          *bTemplValList = SAPDB_FALSE;
        }
        // find end of comment
        pReturn = strstr(pEndOfToken, COMMENT_END);
        if (pReturn != NULL) {
          // skip end of comment
          pReturn = pReturn + strlen(COMMENT_END);
        } // end if
      } //  end if

      pCurrent = strstr(pEndOfToken, COMMENT_BEGIN);
    } // end while

  } // end if

  return pReturn;
} // end SAPDBWeb_Template :: findNextTemplate

/*
  -----------------------------------------------------------------------------
  private function SAPDBWeb_Template :: findEndOfTemplate
  -----------------------------------------------------------------------------
 */
const SAPDB_Char * SAPDBWeb_Template :: findEndOfTemplate ( const SAPDB_Char * pBuffer,
                                                            const SAPDB_Char * pEnd,
                                                            const SAPDB_Char * szTemplate)
{
   
  const SAPDB_Char  *  pReturn       = NULL;
  const SAPDB_Char  *  pCurrent      = NULL;
  const SAPDB_Char  *  pBeginOfToken = NULL;
  const SAPDB_Char  *  pEndOfToken   = NULL;
  SAPDBWeb_String      strToken;

  if (pBuffer != NULL) {

    // look for comment
    pCurrent = strstr(pBuffer, COMMENT_BEGIN);

    while (pCurrent != NULL && pReturn == NULL && pCurrent <= pEnd) {
      // copy pointer
      pBeginOfToken = pCurrent;
      // skip comment
      pBeginOfToken = pBeginOfToken + strlen(COMMENT_BEGIN);
      // skip whitescpace
      while (isspace(*pBeginOfToken) && *pBeginOfToken != '\0') ++pBeginOfToken;
      // search end of token
      pEndOfToken = pBeginOfToken;
      while (!isspace(*pEndOfToken) && *pEndOfToken != '\0') ++pEndOfToken;
      // extract token
      strToken.assign(pBeginOfToken, pEndOfToken - pBeginOfToken);
      // look for keyword
      if (strToken.compare(TEMPLATE_END) == 0) {
        pBeginOfToken = pEndOfToken;
        // skip whitescpace
        while (isspace(*pBeginOfToken) && *pBeginOfToken != '\0') ++pBeginOfToken;
        // search end of token
        pEndOfToken = pBeginOfToken;
        while (!isspace(*pEndOfToken) && *pEndOfToken != '\0') ++pEndOfToken;
        // Extract Name of Template
        strToken.assign(pBeginOfToken, pEndOfToken - pBeginOfToken);
        if (strcmp(strToken.c_str(), szTemplate) == 0) {
          // copy pointer
          pReturn = pCurrent;
        } // end if
      } //  end if

      pCurrent = strstr(pEndOfToken, COMMENT_BEGIN);
    } // end if

  } // end if

  return pReturn;
} // end SAPDBWeb_Template :: findEndOfTemplate

/*
  -----------------------------------------------------------------------------
  private function SAPDBWeb_Template :: findNextValue
  -----------------------------------------------------------------------------
 */
const SAPDB_Char * SAPDBWeb_Template :: findNextValue (const SAPDB_Char * pBuffer,
                                                       const SAPDB_Char * pEnd,
                                                       const SAPDB_Char ** pBeginOfValue,
                                                       const SAPDB_Char ** pEndOfValue,
                                                       SAPDBWeb_String  & strValue)
{
   
  const SAPDB_Char  *  pReturn       = NULL;
  const SAPDB_Char  *  pCurrent      = NULL;
  const SAPDB_Char  *  pBeginOfToken = NULL;
  const SAPDB_Char  *  pEndOfToken   = NULL;

  *pBeginOfValue = NULL;
  *pEndOfValue = NULL;

  if (pBuffer != NULL) {

    // look for value begin
    pCurrent = strstr(pBuffer, VALLIST_VALMARKER_BEGIN);

    if (pCurrent != NULL && pCurrent <= pEnd) {
      // copy pointer
      *pBeginOfValue = pCurrent;

      // skip value marker
      pCurrent = pCurrent + strlen(VALLIST_VALMARKER_BEGIN);

      // copy pointer
      pBeginOfToken = pCurrent;

      if (pCurrent <= pEnd) {

        // look for value end
        pCurrent = strstr(pCurrent, VALLIST_VALMARKER_END);

        if (pCurrent != NULL && pCurrent <= pEnd) {
          // copy pointer
          pEndOfToken = pCurrent;

          // extract token
          strValue.assign(pBeginOfToken, pEndOfToken - pBeginOfToken);
          
          // skip value marker
          pCurrent = pCurrent + strlen(VALLIST_VALMARKER_END);

          // copy pointer
          *pEndOfValue = pCurrent;

          pReturn = pCurrent;
        } // end if
      } // end if
    } // end if
  } // end if

  return pReturn;
} // end SAPDBWeb_Template :: findNextValue
