/*


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end


*/
#if defined(R73) && defined(_WIN32)
  #line 3 "SAPDBWeb_TemplateMsgBox.cpp"
#endif
/*!
  -----------------------------------------------------------------------------

  module: SAPDBWeb_TemplateMsgBox.cpp

  -----------------------------------------------------------------------------

  responsible:  BerndV

  special area: universal stuff for web access

  description:  a class for creating simple "Messageboxes" in SAP MiniApps style

  version:      7.3 and higher
  -----------------------------------------------------------------------------

                          Copyright by SAP AG, 2000

  -----------------------------------------------------------------------------
*/

/*
  -----------------------------------------------------------------------------
  includes
  -----------------------------------------------------------------------------
 */
#include <string.h>
#include "SAPDBCommon/Web/SAPDBWeb_TemplateMsgBox.hpp"

/*
  -----------------------------------------------------------------------------
  private macros
  -----------------------------------------------------------------------------
 */
#define OBJ_LOGO          "Logo"
#define OBJ_TITLE         "Title"
#define OBJ_MESSAGE       "Message"
#define OBJ_BUTTON        "Button"
#define OBJ_MSGLIST       "MessageList"
#define OBJ_MSGNO         "MessageNumber"
#define OBJ_MSGTXT        "MessageText"

/*
  =============================================================================
  class:        SAPDBWeb_TemplateMsgBox
  =============================================================================
*/

/*
  -----------------------------------------------------------------------------
  private function SAPDBWeb_TemplateMsgBox :: askForWriteCount
  -----------------------------------------------------------------------------
 */
SAPDB_Int2 SAPDBWeb_TemplateMsgBox :: askForWriteCount  ( const SAPDB_Char * szName )
{
  SAPDB_Int2 nReturn = 0;

  if (strcmp(szName, OBJ_LOGO) == 0) {
    nReturn = (m_bLogo) ? 1 : 0;
  } else if (strcmp(szName, OBJ_TITLE) == 0) {
    nReturn = 1;
  } else if (strcmp(szName, OBJ_MESSAGE) == 0) {
    nReturn = m_strMessage.Empty() ? 0 : 1;
  } else if (strcmp(szName, OBJ_MSGLIST) == 0) { 
    nReturn = m_oMsgList.IsEmpty() ? 0 : m_oMsgList.NumOfMessages();
    m_pCurrentMsg = &m_oMsgList;
  } else if (strcmp(szName, OBJ_MSGNO) == 0) {   
    nReturn = 1;
  } else if (strcmp(szName, OBJ_MSGTXT) == 0) {  
    nReturn = 1;
  } else if (strcmp(szName, OBJ_BUTTON) == 0) {
    nReturn = (m_strBtnText.empty()) ? 0 : 1;
  } // end if

  return nReturn;
} // end SAPDBWeb_TemplateMsgBox :: askForWriteCount

/*
  -----------------------------------------------------------------------------
  private function SAPDBWeb_TemplateMsgBox :: askForValue
  -----------------------------------------------------------------------------
 */
const SAPDBWeb_String SAPDBWeb_TemplateMsgBox :: askForValue  ( const SAPDB_Char * szName )
{
  const SAPDB_Char * pReturn = NULL;
  SAPDBWeb_String    strReturn;

  if (strcmp(szName, OBJ_TITLE) == 0) {
    pReturn = m_strTitle.c_str();
  } else if (strcmp(szName, OBJ_MESSAGE) == 0) {
    return m_strMessage;
  } else if (strcmp(szName, OBJ_MSGNO) == 0) {   
    while (m_pCurrentMsg != NULL && m_pCurrentMsg->IsEmpty()) {
      m_pCurrentMsg = m_pCurrentMsg->NextMessage();
    } // end while
    if (m_pCurrentMsg != NULL) {
      strReturn.assign((int) m_pCurrentMsg->ID());
      pReturn = strReturn.c_str();
    } // end if
  } else if (strcmp(szName, OBJ_MSGTXT) == 0) {  
    if (m_pCurrentMsg != NULL) {
      pReturn = m_pCurrentMsg->Message();
    } // end if
      m_pCurrentMsg = m_pCurrentMsg->NextMessage();
  } else if (strcmp(szName, OBJ_BUTTON) == 0) {
    m_strButton.assign("'")
               .append(m_strBtnText)
               .append("','")
               .append(m_strBtnAction)
			   .append("','")
			   .append(m_strBtnTarget)
               .append("'");
	
    pReturn = m_strButton.c_str();
  } // end if

  if (pReturn == NULL) {
    pReturn = "";
  } // end if

  return pReturn;
} // end SAPDBWeb_TemplateMsgBox :: askForValue
