/*


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end


*/
#if defined(R73) && defined(_WIN32)
  #line 3 "SAPDBWeb_TemplateMsgBox.hpp"
#endif
/*! 
  -----------------------------------------------------------------------------
 
  module: SAPDBWeb_TemplateMsgBox.hpp
 
  -----------------------------------------------------------------------------
 
  responsible:  BerndV
 
  special area: universal stuff for web access
   
  description:  a class for creating simple "Messageboxes" in SAP MiniApps style

  version:      7.3 and higher 
  -----------------------------------------------------------------------------
 
                          Copyright by SAP AG, 2000
 
  -----------------------------------------------------------------------------
*/

#ifndef _SAPDBWeb_TemplateMsgBox_HPP_
#define _SAPDBWeb_TemplateMsgBox_HPP_

#include "SAPDBCommon/Web/SAPDBWeb_Template.hpp"
#include "SAPDBCommon/Web/SAPDBWeb_String.hpp"
#include "SAPDBCommon/ErrorsAndMessages/SAPDBErr_MessageList.hpp"

#include "SAPDBCommon/SAPDB_DynamicUTF8String.hpp"

/*
  -----------------------------------------------------------------------------
  includes
  -----------------------------------------------------------------------------
 */

/*!
  -----------------------------------------------------------------------------
  Class:        SAPDBWeb_TemplateMsgBox
  -----------------------------------------------------------------------------

  description:  SAPDBWeb 

                An instance of this class represents a HTML-Page with a Table
  -----------------------------------------------------------------------------
 */
class SAPDBWeb_TemplateMsgBox : public SAPDBWeb_Template 
{
  public:

    /*!
      -------------------------------------------------------------------------
      function:     SAPDBWeb_TemplateMsgBox
      -------------------------------------------------------------------------
      description:  Constructor

      arguments:    wa [IN]  - webagent instance
      -------------------------------------------------------------------------
     */
     SAPDBWeb_TemplateMsgBox  ( sapdbwa_WebAgent & wa ) : SAPDBWeb_Template ( wa, "SAPDBMsgBox.htm" ) 
        {
          m_bLogo         = false;
          m_strTitle      = "Information";
          m_strMessage    = "";
          m_strBtnText    = "Back";
          m_strBtnAction  = "javascript:history.back()";
          m_strButton     = "";
          m_oMsgList.ClearMessageList();
          m_pCurrentMsg   = NULL;
		  m_strBtnTarget  =	"_self";
        } 
 
    /*!
      -------------------------------------------------------------------------
      function:     ~SAPDBWeb_TemplateMsgBox
      -------------------------------------------------------------------------
      description:  Destructor
      -------------------------------------------------------------------------
     */
     ~SAPDBWeb_TemplateMsgBox ( ) {}

    /*!
      -------------------------------------------------------------------------
      function:     enableLogo
      -------------------------------------------------------------------------
      description:  enable the Logo

      arguments:    bEnable   [IN] - enable or disable the logo (default: disabled)

      return value: a reference to this object
      -------------------------------------------------------------------------
     */
     SAPDBWeb_TemplateMsgBox & enableLogo ( bool bEnable = true )
        {
          m_bLogo = bEnable;
          return *this;
        } 

    /*!
      -------------------------------------------------------------------------
      function:     setTitle
      -------------------------------------------------------------------------
      description:  sets the title message box

      arguments:    szText   [IN] - the title (default: "Information")

      return value: a reference to this object
      -------------------------------------------------------------------------
     */
     SAPDBWeb_TemplateMsgBox & setTitle ( const SAPDB_Char * szText )
        {
          m_strTitle = szText;
          return *this;
        } 

    /*!
      -------------------------------------------------------------------------
      function:     setMessage
      -------------------------------------------------------------------------
      description:  sets the text of the message

      arguments:    szText  [IN] - the message text

      return value: a reference to this object
      -------------------------------------------------------------------------
     */
     SAPDBWeb_TemplateMsgBox & setMessage ( const SAPDB_Char * szText )
        {
          m_strMessage = szText;
          return *this;
        } 

    /*!
      -------------------------------------------------------------------------
      function:     setMessage
      -------------------------------------------------------------------------
      description:  sets the text of the message

      arguments:    szText  [IN] - the message text

      return value: a reference to this object
      -------------------------------------------------------------------------
     */
     SAPDBWeb_TemplateMsgBox & setMessage ( const SAPDB_DynamicUTF8String&  szText )
        {
          m_strMessage = szText;
          return *this;
        } 

    /*!

      -------------------------------------------------------------------------
      function:     setMessage
      -------------------------------------------------------------------------
      description:  sets the text of the message

      arguments:    szText  [IN] - the message text

      return value: a reference to this object
      -------------------------------------------------------------------------
     */
     SAPDBWeb_TemplateMsgBox & setMessage ( SAPDBErr_MessageList & oMsgList )
        {
          m_oMsgList = oMsgList;
          return *this;
        } 
                     
    /*!
      -------------------------------------------------------------------------
      function:     setButtonText
      -------------------------------------------------------------------------
      description:  sets the text of the Button

      arguments:    szText  [IN] - text for button (Default: "back")

      return value: a reference to this object
      -------------------------------------------------------------------------
     */
     SAPDBWeb_TemplateMsgBox & setButtonText ( const SAPDB_Char * szText )
        {
          m_strBtnText = szText;
          return *this;
        } 

    /*!
      -------------------------------------------------------------------------
      function:     setButtonAction
      -------------------------------------------------------------------------
      description:  sets the action of the Button

      arguments:    szText  [IN] - action for button 
                                     (Default: "javascript:history.back()")

      return value: a reference to this object
      -------------------------------------------------------------------------
     */
     SAPDBWeb_TemplateMsgBox & setButtonAction ( const SAPDB_Char * szText )
        {
          m_strBtnAction = szText;
          return *this;
        } 

	     /*!
      -------------------------------------------------------------------------
      function:     setButtonTarget
      -------------------------------------------------------------------------
      description:  sets the target of the Button

      arguments:    szText  [IN] - target for button 
                                     (Default: "javascript:history.back()")

      return value: a reference to this object
      -------------------------------------------------------------------------
     */
     SAPDBWeb_TemplateMsgBox & setButtonTarget ( const SAPDB_Char * szText )
        {
          m_strBtnTarget = szText;
          return *this;
        } 

  private:
     virtual SAPDB_Int2 askForWriteCount  ( const SAPDB_Char * szName );

     virtual const SAPDBWeb_String askForValue ( const SAPDB_Char * szName );

     bool                   m_bLogo;
     SAPDBWeb_String        m_strTitle;
     SAPDB_DynamicUTF8String        m_strMessage;
     SAPDBWeb_String        m_strBtnText;
     SAPDBWeb_String        m_strBtnAction;
     SAPDBWeb_String        m_strButton;
	 SAPDBWeb_String        m_strBtnTarget;
     SAPDBErr_MessageList   m_oMsgList;
     SAPDBErr_MessageList * m_pCurrentMsg;
}; 

/*! EndClass: SAPDBWeb_TemplateMsgBox */

#endif // _SAPDBWeb_TemplateMsgBox_HPP_

