/*! 
  -----------------------------------------------------------------------------
 
  module: SAPDBWeb_TemplateSimpleTable.hpp
 
  -----------------------------------------------------------------------------
 
  responsible:  BerndV
 
  special area: universal stuff for web access
   
  description:  a class for creating simple Tables in SAP MiniApps style

  version:      7.3 and higher 
  -----------------------------------------------------------------------------
 
                          Copyright by SAP AG, 2000
 
  -----------------------------------------------------------------------------


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/

#ifndef _SAPDBWeb_TemplateSimpleTable_HPP_
#define _SAPDBWeb_TemplateSimpleTable_HPP_

#include "SAPDBCommon/Web/SAPDBWeb_Template.hpp"
#include "SAPDBCommon/Web/SAPDBWeb_String.hpp"

/*
  -----------------------------------------------------------------------------
  includes
  -----------------------------------------------------------------------------
 */

/*!
  -----------------------------------------------------------------------------
  Class:        SAPDBWeb_TemplateSimpleTable
  -----------------------------------------------------------------------------

  description:  SAPDBWeb 

                An instance of this class represents a HTML-Page with a Table
  -----------------------------------------------------------------------------
 */
class SAPDBWeb_TemplateSimpleTable : public SAPDBWeb_Template 
{
  public:

    /*!
      -------------------------------------------------------------------------
      function:     SAPDBWeb_TemplateSimpleTable
      -------------------------------------------------------------------------
      description:  Constructor

      arguments:    wa [IN]  - webagent instance

      -------------------------------------------------------------------------
     */
     SAPDBWeb_TemplateSimpleTable  ( sapdbwa_WebAgent       &   wa );

 
    /*!
      -------------------------------------------------------------------------
      function:     ~SAPDBWeb_TemplateSimpleTable
      -------------------------------------------------------------------------
      description:  Destructor
      -------------------------------------------------------------------------
     */
     ~SAPDBWeb_TemplateSimpleTable ( );

     /*!
      -------------------------------------------------------------------------
      function:     enableLogo
      -------------------------------------------------------------------------
      description:  enable the Logo

      arguments:    bEnable   [IN] - enable or disable the logo (default: disabled)

      return value: a reference to this object
      -------------------------------------------------------------------------
     */
     SAPDBWeb_TemplateSimpleTable & enableLogo ( bool bEnable = true )
        {
          m_bLogo = bEnable;
          return *this;
        } 

     /*!
      -------------------------------------------------------------------------
      function:     enableGroupbox
      -------------------------------------------------------------------------
      description:  enable the Groupbox style

      arguments:    bEnable   [IN] - enable or disable the style (default: disabled)

      return value: a reference to this object
      -------------------------------------------------------------------------
     */
     SAPDBWeb_TemplateSimpleTable & enableGroupbox ( bool bEnable = true )
        {
          m_bGroup = bEnable;
          return *this;
        } 

    /*!
      -------------------------------------------------------------------------
      function:     setHeadLine
      -------------------------------------------------------------------------
      description:  sets the headline of the table

      arguments:    szHeadLine   [IN] - the headline

      return value: a reference to the object
      -------------------------------------------------------------------------
     */
     SAPDBWeb_TemplateSimpleTable & setHeadLine      ( const SAPDB_Char * szHeadLine   );

    /*!
      -------------------------------------------------------------------------
      function:     setTableTitle
      -------------------------------------------------------------------------
      description:  sets the title of the table

      arguments:    szTableTitle   [IN] - the title

      return value: a reference to the object
      -------------------------------------------------------------------------
     */
     SAPDBWeb_TemplateSimpleTable &  setTableTitle    ( const SAPDB_Char * szTableTitle );

    /*!
      -------------------------------------------------------------------------
      function:     setColumnTitle
      -------------------------------------------------------------------------
      description:  sets the title of a column

      arguments:    szColumnTitle   [IN] - the title
                    nColumn         [IN] - no of column

      return value: a reference to the object
      -------------------------------------------------------------------------
     */
     SAPDBWeb_TemplateSimpleTable &  setColumnTitle   ( const SAPDB_Char * szColumnTitle,    
                                        const  int         nColumn     );

    /*!
      -------------------------------------------------------------------------
      function:     setCell
      -------------------------------------------------------------------------
      description:  sets a cell of the table

      arguments:    szCell     [IN] - the content for the cell
                    nRow       [IN] - no of row
                    nColumn    [IN] - no of column

      return value: a reference to the object
      -------------------------------------------------------------------------
     */
     SAPDBWeb_TemplateSimpleTable &  setCell          ( const SAPDB_Char * szCell, 
                                        const int          nRow, 
                                        const int          nColumn      );

    /*!
      -------------------------------------------------------------------------
      function:     setColumnSummary
      -------------------------------------------------------------------------
      description:  sets the summary of a column

      arguments:    szColumnSummary [IN] - the summary
                    nColumn         [IN] - no of column

      return value: a reference to the object
      -------------------------------------------------------------------------
     */
     SAPDBWeb_TemplateSimpleTable &  setColumnSummary ( const SAPDB_Char * szColumnSummary,  
                                        const int          nColumn      );

     /*!
      -------------------------------------------------------------------------
      function:     addButton
      -------------------------------------------------------------------------
      description:  adds a Button

      arguments:    szName    [IN] - button text
                    szLink    [IN] - button link
					szTarget  [IN] - button target
                    bDisabled [IN] - button state

      return value: a reference to the object
      -------------------------------------------------------------------------
     */
     SAPDBWeb_TemplateSimpleTable &  addButton   ( const SAPDB_Char * szName,  
                                   const SAPDB_Char * szLink,
								   const SAPDB_Char * szTarget,
                                   const bool         bDisabled = false);

  private:
     virtual SAPDB_Int2 askForWriteCount  ( const SAPDB_Char * szName );

     virtual const SAPDBWeb_String askForValue ( const SAPDB_Char * szName );


     SAPDB_Int2        m_nMaxRow;
     SAPDB_Int2        m_nMaxCol;
     bool              m_bTitleRow;
     bool              m_bSummaryRow;
     SAPDB_Int2        m_nCurrentRow;
     SAPDB_Int2        m_nCurrentCol;
     SAPDB_Int2        m_nMaxButton;
     SAPDB_Int2        m_nCurrentButton;
     bool              m_bLogo;
     bool              m_bGroup;

     SAPDBWeb_String   m_strHeader;
     SAPDBWeb_String   m_strTableTitle;

     typedef SAPDBWeb_Array<SAPDBWeb_String> StringMap;
     typedef SAPDBWeb_Array<StringMap>       RowMap;

     StringMap  m_ColumnTitleMap;
     StringMap  m_ColumnSummaryMap;
     RowMap     m_RowMap;
     StringMap  m_ButtonMap;
}; 

/*! EndClass: SAPDBWeb_TemplateSimpleTable */

#endif // _SAPDBWeb_TemplateSimpleTable_HPP_

