/*


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end


*/

#ifndef StudioOAL_COLUMN_HPP
#define StudioOAL_COLUMN_HPP

/*!**********************************************************************

  module: StudioOAL_COLUMN.hpp

  -----------------------------------------------------------------------

  responsible:  WolfgangA

  special area: OAL

  description:  Column Class for OAL

  see also:

  -----------------------------------------------------------------------

  copyright:                Copyright by SAP AG, 2000

************************************************************************/

/*=======================================================================
 *  INCLUDES
 *=====================================================================*/

#include "SAPDBCommon/SAPDB_DynamicUTF8String.hpp"
//includes different enum definitons
#include "SAPDB/SQLStudio/OAL/StudioOAL_Common.h"	   




class StudioOAL_Column  
{
public:
	StudioOAL_Column();
	virtual ~StudioOAL_Column();

private:		

	//name of the column
	SAPDB_DynamicUTF8String		m_strColName;        

	SQLSMALLINT			m_enumODBCColType;
	SQLUINTEGER			m_lColLength;       
	SQLSMALLINT			m_nColDecimal;

	//private properties
	bool		m_bColNullable;
	bool		m_bColIsNull;
	bool		m_bColIsKey;
	bool		m_bColIsUpdatable;

public:
	
	/*!-------------------------------------------------------------------------
    function:     setColName
    description:  set column name
	arguments:	  [in] strColName, new column name
				  
    return value: /
    --------------------------------------------------------------------------*/
	void setColName(const SAPDB_DynamicUTF8String& strColName){m_strColName = strColName;}

	/*get column name*/
	void getColName(SAPDB_DynamicUTF8String& ColName)  {ColName = m_strColName;}

	/*set a value that indicates whether the column allows NULL values*/
	void setColNullable(bool ColNullable) {m_bColNullable = ColNullable;}

	/*get a value that indicates whether the column allows NULL values*/
	bool getColNullable() { return m_bColNullable;}

	/*sets the number of decimal digits of the column on the data source*/
	void setColDecimal(SQLSMALLINT nColDecimal) {m_nColDecimal = nColDecimal;}

	/*set the size of the column on the data source*/
	void setColLength(SQLUINTEGER lColLength){	m_lColLength = lColLength;}

	/*get the size of the column on the data source*/
	SQLUINTEGER  getColLength(){return m_lColLength;}

	/*set if data is NULL on the data source*/
	void setColIsNull(bool ColIsNull){m_bColIsNull = ColIsNull;}

	/*set if data is NULL on the data source*/
	bool getColIsNull() { return m_bColIsNull;}

	/*set the sql (odbc) data type of the column*/
	void setODBCColType (SQLSMALLINT nODBCColType)	{m_enumODBCColType = nODBCColType;}

	/*returns the sql (odbc) data type of the column*/
	 SQLSMALLINT getODBCColType ()	{return m_enumODBCColType;}


};

#endif