/*


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end


*/
#ifndef StudioOAL_ERROR_HPP
#define StudioOAL_ERROR_HPP

/*!**********************************************************************

  module: StudioOAL_Error.h

  -----------------------------------------------------------------------

  responsible:  WolfgangA

  special area: OAL

  description:  Error Class for OAL

  see also:

  -----------------------------------------------------------------------

  copyright:                Copyright by SAP AG, 2000

************************************************************************/

/*=======================================================================
 *  INCLUDES
 *=====================================================================*/


#include <WINDOWS.H>
#include <sql.h>		// Main include for ODBC core functions
#include <sqlext.h>     // Used for applications using Microsoft SQL Extensions
#include <sqlucode.h>	// Unicode Types for ODBC


#include "SAPDBCommon/SAPDB_DynamicUTF8String.hpp"
#include "SAPDB/SQLStudio/OAL/StudioOAL_Common.h"
#include "SAPDBCommon/SAPDB_Types.hpp"

class StudioOAL_Error  
{
public:
	StudioOAL_Error(SQLHANDLE hDhc,SQLHENV hEnv);
	virtual ~StudioOAL_Error();

//Variables
private:
	//EnvironmentHandle
	SQLHENV		m_hEnv ;

	//ConnectionHandle
	SQLHDBC		m_hDbc;
	

	//error object contains an error
	bool		m_bIsError;  

	//contains a warning
	bool		m_bIsWarning;

	//sql return code
	SQLRETURN	m_lRC;

	//position of the error in sql statement
	SAPDB_Int4		m_lErrPos;

	//native error number
	SQLINTEGER		m_nNativeError;

	SAPDB_DynamicUTF8String		m_strNativeError;
	SAPDB_DynamicUTF8String		m_strErrorText;
	SAPDB_DynamicUTF8String		m_strSQLState;
	SAPDB_DynamicUTF8String		m_strSQLStateErrorNumber;
	//////////////////////////////////////////////////////
	
	//if you want a SQLState to be ignored set member
	SAPDB_DynamicUTF8String		m_strNoSQLState;

	//if you want a NativeError to be ignored set member
	SAPDB_DynamicUTF8String		m_strNoNativeError;   

//Functions
private:
	//set ConnectionHandle
	void setDbc(SQLHANDLE hDhc);

	//set EnvironmentHandle
	void setEnv(SQLHENV hEnv);

		
	/*!-------------------------------------------------------------------------
    function:     scanErrorText
    description:  scans test which is returned by SQLError
	arguments:	  [in] strErrorText,input error text returned by SQLError
				  [in] strSQLState,input sql state text returned by SQLError
				  [in] nNativeError, native error number
				  [out] strOutErrorText, 
				  [out] strOutSQLState
				  [out] strOutNativeError, native error number
				  [out] *lOutErrPos,position of error in sql statement

				  
    return value: /
    --------------------------------------------------------------------------*/
	void scanErrorText(const SAPDB_DynamicUTF8String &strErrorText,
  					   const SAPDB_DynamicUTF8String &strSQLState,
					   SQLINTEGER  nNativeError,
					   SAPDB_DynamicUTF8String &strOutErrorText,
					   SAPDB_DynamicUTF8String &strOutSQLState,
					   SAPDB_DynamicUTF8String &strOutNativeError,
					   SAPDB_Int4 *lOutErrPos);
public:
		
	/*!-------------------------------------------------------------------------
    function:     checkSQLReturnCode
    description:  check if SQLReturnCode of a staement is an error
	arguments:	  [in] ReturnCode, return code of a SQL.... function
				  [in] hStmt, statement handle
				  

				  
    return value: true, if success
    --------------------------------------------------------------------------*/
	bool checkSQLReturnCode(SQLRETURN ReturnCode, SQLHSTMT hStmt);

	void clear();

	bool isError(){	return m_bIsError;}
	bool isWarning(){	return m_bIsWarning;}


	//returns error pos in error text
	SAPDB_Int4 getErrorPos(){	return m_lErrPos;}
	SQLRETURN  getReturnCode(){	return m_lRC;}


	//return ConnectionHandle
	SQLHANDLE getDbc() {return m_hDbc;}

	//return EnvironmentHandle
	SQLHENV getEnv(){return m_hEnv;}
	
	//return SQL State
	const SAPDB_DynamicUTF8String& getSQLState()		      {	return m_strSQLState;}
	const SAPDB_DynamicUTF8String& getSQLStateErrorNumber()   {	return m_strSQLStateErrorNumber;}
	const SAPDB_DynamicUTF8String& getNativeErrorAsString()	  {	return m_strNativeError;}
	const SAPDB_DynamicUTF8String& getErrorText()			  {	return m_strErrorText;}
	const SAPDB_DynamicUTF8String& getNoSQLState()			  { return m_strNoSQLState;}

	SQLINTEGER getNativeError()								   {	return m_nNativeError;}
	

	void  setNoSQLState(const SAPDB_DynamicUTF8String& NoSqlState){  m_strNoSQLState = NoSqlState;}
	
};


#endif