/*
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/


/*!**********************************************************************

  module: StudioOAL_Parameters.hpp

  -----------------------------------------------------------------------

  mail to:  sapdb-general@sap.com

  special area: OAL

  description:  class for parameters

  version    :  7.3

  see also:

  -----------------------------------------------------------------------

  copyright:                Copyright by SAP AG, 2000

************************************************************************/



#define	   	  PARAM_UNKNOWN		"Unknown"        
#define       PARAM_IN			"In"
#define       PARAM_INOUT		"InOut"
#define		  PARAM_RESULTCOL	"ResultCol"
#define 	  PARAM_OUT			"Out"
#define		  PARAM_RETURNVALUE "ReturnValue"

const int	  MX_BYTEVALUE = 32768;	

#include	  "SAPDB/SQLStudio/OAL/StudioOAL_Common.h"

#include	  "SAPDB/SAPDBCommon/SAPDB_Types.hpp"

#include	  "SAPDBCommon/SAPDB_DynamicUTF8String.hpp"

/*

SQLRETURN SQLBindParameter(
     SQLHSTMT         StatementHandle,
     SQLUSMALLINT     ParameterNumber,
     SQLSMALLINT      InputOutputType,
     SQLSMALLINT      ValueType,
     SQLSMALLINT      ParameterType,
     SQLUINTEGER      ColumnSize,
     SQLSMALLINT      DecimalDigits,
     SQLPOINTER       ParameterValuePtr,
     SQLINTEGER       BufferLength,
     SQLINTEGER *     StrLen_or_IndPtr);

*/
class StudioOAL_Parameters
{


public:
		/*!-------------------------------------------------------------------------
		  function:     StudioOAL_Parameters
		  description:  Constructor
		  arguments:	[in] hStmt, handle of param statement
						[in] nParameterNo, number of param
		  										
      
		  return value: / 
		--------------------------------------------------------------------------*/
		StudioOAL_Parameters(SQLHANDLE	hStmt,SAPDB_Int2  nParameterNo);
		
		 /*!-------------------------------------------------------------------------
		  function:     ~StudioOAL_Parameters
		  description:  Destructor
		  arguments:    /

		  return value: / 
		--------------------------------------------------------------------------*/
		~StudioOAL_Parameters();


		enum  emParameterType{

		  ptUnknown = 0,
		  ptIn = 1,
		  ptInOut = 2,
		  ptResultCol = 3,
		  ptOut = 4,
		  ptReturnValue = 5
		};

		/*!-------------------------------------------------------------------------
		  function:     setParameterType
		  description:  set the paramtype of parameter
		  arguments:    [in] emParameterType, paramtype of parameter

		  return value: / 
		--------------------------------------------------------------------------*/
		void setParameterType(emParameterType parameterType){m_emParameterType = parameterType;}

		/*!-------------------------------------------------------------------------
		  function:     getParameterType
		  description:  get the type of the parameter
		  arguments:    /

		  return value: m_emParameterType,type of parameter
		--------------------------------------------------------------------------*/
		SAPDB_Int2 getParameterType(){return m_emParameterType;}

		/*!-------------------------------------------------------------------------
		  function:     setColumnType
		  description:  set the column type of parameter
		  arguments:    [in] nColumnType, columntype of parameter

		  return value: / 
		--------------------------------------------------------------------------*/
		void setColumnType(SAPDB_Int2 nColumnType){m_nColumnType = nColumnType;}

		/*!-------------------------------------------------------------------------
		  function:     getColumnType
		  description:  get the column type of parameter
		  arguments:    /

		  return value: m_nColumnType, columntype of parameter 
		--------------------------------------------------------------------------*/
		SAPDB_Int2 getColumnType(){return m_nColumnType;}
		
		/*!-------------------------------------------------------------------------
		  function:     getParameterNo
		  description:  returns parameter number
		  arguments:    /

		  return value: m_nParameterNo
		--------------------------------------------------------------------------*/
		SAPDB_Int2 getParameterNo(){return m_nParameterNo;}

	   /*!-------------------------------------------------------------------------
		  function:     setValue
		  description:  set paramvalue
		  arguments:    sValue, reference to string class

		  return value: /
		-------------------------------------------------------------------------*/
		void setValue ( SAPDB_DynamicUTF8String sValue) ;

		/*!-------------------------------------------------------------------------
		  function:     getValue
		  description:  get paramvalue
		  arguments:    sValue, reference to string class

		  return value: /
		-------------------------------------------------------------------------*/
		void getValue (SAPDB_DynamicUTF8String &sValue);

		/*!-------------------------------------------------------------------------
		  function:     setValueLength
		  description:  set paramvaluelength
		  arguments:    lValueLength

		  return value: /
		-------------------------------------------------------------------------*/
		inline void setValueLength (SAPDB_Long lValueLength) { m_lValueLength = lValueLength;}

		/*!-------------------------------------------------------------------------
		  function:     getValueLength
		  description:  get paramvaluelength
		  arguments:    sValue, reference to string class

		  return value: /
		-------------------------------------------------------------------------*/
		inline SAPDB_Long getValueLength ( ) {  return m_lValueLength;}



		/*!-------------------------------------------------------------------------
		  function:     bindParameter
		  description:  bind parameters
		  arguments:    /

		  return value: true if success
		--------------------------------------------------------------------------*/
		SAPDB_Bool bindParameter();
		SAPDB_Bool bindSQLParameter();
		

private:

	SQLHANDLE  m_hStmt;
	SAPDB_Int2 m_nParameterNo;
	SAPDB_Int2 m_nColumnType;

	SAPDB_Long m_lValueLength;

	emParameterType m_emParameterType;	

	SAPDB_Byte* m_pValue;	

};
