/*
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/


/*!**********************************************************************

  module: Studio_DBFSObject.cpp

  -----------------------------------------------------------------------

  mail to:  sapdb-general@sap.com

  special area: WebQuery

  description:  class for reading and writing to database file system

  see also:		/

  -----------------------------------------------------------------------

  copyright:                Copyright by SAP AG, 2000

************************************************************************/

#include "SAPDB/SQLStudio/Studio_DBFSObject.hpp"
#include "SAPDB/SQLStudio/Studio_Errors.h"

Studio_DBFSObject::Studio_DBFSObject(Studio_Connection* poConnection,
									 SAPDB_UInt2	   nFSPrefix)
									 
:m_nFSPrefix(nFSPrefix),
m_poConnection(poConnection)
{
	m_pDBFS = NULL; 	 
}

Studio_DBFSObject::~Studio_DBFSObject()
{
	sapdbwa_ErrP pError = NULL; 

	//destroy dbfs
	if (m_pDBFS != NULL ) {

		if (wd101DestroyDBFS(m_pDBFS) == 0 ) {
     
			pError = wd101GetErr(m_pDBFS);
			if (sapdbwa_GetErrId(pError) != sapdbwa_Error_NoError) {
				//wa.WriteLogMsg(sapdbwa_GetErrMsg(pError));
			}
		}		 
		m_pDBFS = NULL;
	}
}



SAPDB_Int4 Studio_DBFSObject::readFile(const SAPDB_DynamicUTF8String& sFilePath, 
									   SAPDB_UInt2			  nMaxFileSize,
									   SAPDB_DynamicUTF8String& sContent)
				
{
	//init output
	sContent.Erase();

	//if there is no database filesystem, get one
	if (m_pDBFS == NULL ){

		if (getValidDBFS(sContent) == SAPDB_FALSE) return -1;		
	}

	twd101DBFile lFile;
	SAPDB_Int4  lByteRead = 0;

	sapdbwa_ErrP pError = NULL;
	SAPDB_DynamicUTF8String sError;		

	SAPDB_Char	sAsciiFilePath[512];
	memset(sAsciiFilePath,0,512);	
	SAPDB_Char* szpConvert = NULL;
	SAPDB_DynamicUTF8String strCopyOfFilePath = sFilePath;
	assert(strCopyOfFilePath.ConvertToASCII_Latin1((char*)sAsciiFilePath,sAsciiFilePath+sizeof (sAsciiFilePath),szpConvert) == SAPDB_UTF8Basis::Success);	
	
	//NOT UNICODE BECAUSE OF WA
	lFile = wd101Open(m_pDBFS,sAsciiFilePath,access_read );
	if (lFile == -1){
		
		if (getDBFSError(sError) == SAPDB_FALSE){
			 
			sContent = sError;
			return -1;
		 }
		 else{
			//open file again
			lFile = wd101Open(m_pDBFS,sAsciiFilePath,access_read );
			if (lFile == -1 ) {

				getDBFSError(sError);
				sContent = sError;
				return -1;
			}
		}	
	}	

	char *psBuffer = NULL;
	psBuffer = new char[nMaxFileSize];
	if(psBuffer == NULL) return -1;

	lByteRead = wd101Read(m_pDBFS,lFile,psBuffer,nMaxFileSize);
	if (lByteRead == -1) {
		
		getDBFSError(sError);
		sContent = sError;		
		lByteRead = -1;
	}

	if (wd101Close(m_pDBFS,lFile) == SAPDB_FALSE){

		getDBFSError(sError);
		sContent = sError;
		delete[] psBuffer;
		return -1;
	}

	if(lByteRead > 0)
		sContent = SAPDB_DynamicUTF8String((SAPDB_DynamicUTF8String::ConstPointer) psBuffer,lByteRead);
	

	delete[] psBuffer;
	
	return lByteRead;
}

SAPDB_Bool Studio_DBFSObject::getDBFSError(SAPDB_DynamicUTF8String &sOut)
{		    

	if(m_poConnection == NULL){

		sOut = (SAPDB_UTF8*)ST_ERR016;        
		return SAPDB_FALSE;
	}

	//check some pointers
	if (m_poConnection->getDBCHandle() == NULL) {
		

		sOut = ST_ERR001;
		return SAPDB_FALSE;
	}

	if (m_pDBFS == NULL) {

		sOut = (SAPDB_UTF8*)ST_ERR002;
		return SAPDB_FALSE;
	}

	char sBuffer[512];
	memset(sBuffer,0,512);

	//init SAPDB_Bool value
	sapdbwa_Bool connectionDown = SAPDB_TRUE;	
	//get error object
	sapdbwa_ErrP pError = wd101GetErr(m_pDBFS);
	if (pError == NULL) return SAPDB_FALSE;	
	
	if (sapdbwa_CheckDBC(m_poConnection->getDBCHandle(),
					sapdbwa_GetNativeOdbcErr(pError),
					sapdbwa_GetSqlState(pError),
					&connectionDown)) {

		if (connectionDown == SAPDB_FALSE) {			 

			 if (sapdbwa_GetErrId(pError) != sapdbwa_Error_NoError) {

	 			sapdbwa_CopyErrMsg(pError, sBuffer, sizeof (sBuffer));
			 }
			//NOT UNICODE BECAUSE OF WA
			sOut.ConvertFromASCII_Latin1(
							(SAPDB_DynamicUTF8String::ASCIIConstPointer)sBuffer,
							(SAPDB_DynamicUTF8String::ASCIIConstPointer)sBuffer+strlen(sBuffer));

			return SAPDB_FALSE;
		}
		else
			sOut = (SAPDB_UTF8*)ST_ERR014;		
	}
	else
		sOut = (SAPDB_UTF8*)ST_ERR015;					
	
	
	return SAPDB_TRUE;
}

SAPDB_Bool  Studio_DBFSObject::getValidDBFS(SAPDB_DynamicUTF8String &sError)
{
	
	if(m_poConnection == NULL){

		sError = (SAPDB_UTF8*)ST_ERR016;        
		return SAPDB_FALSE;
	}
	if (m_poConnection->getDBCPool() == NULL) {

		sError = (SAPDB_UTF8*)ST_ERR013;        
		return SAPDB_FALSE;
	}
#if defined(WIN32)
	if (m_poConnection->getODBCDriver().Length() < 1) {

		sError = (SAPDB_UTF8*)ST_ERR003;        
		return SAPDB_FALSE;
	}
#endif
	
	if (m_pDBFS != NULL) {

        if (wd101DestroyDBFS(m_pDBFS) == 0 ) {    
         
			getDBFSError(sError);			
			return SAPDB_FALSE;
		}
		m_pDBFS = NULL;		
     }

	SAPDB_Char sAsciiPassWord[128];
	memset(sAsciiPassWord,0,128);
	SAPDB_Char	sAsciiDBUser[128];
	memset(sAsciiDBUser,0,128);
	SAPDB_Char	sAsciiDBServer[128];
	memset(sAsciiDBServer,0,128);
	SAPDB_Char	sAsciiDBName[128];
	memset(sAsciiDBName,0,128);
	SAPDB_Char	sAsciiPrefix[128];
	memset(sAsciiPrefix,0,128);

	SAPDB_Char	sAsciiODBCDriver[128];
	memset(sAsciiODBCDriver,0,128);

	SAPDB_Char* szpConvert = NULL;


	//NOT UNICODE BECAUSE OF WA
	SAPDB_DynamicUTF8String		sCopyOfDBServer = m_poConnection->getDBServer();
	SAPDB_DynamicUTF8String		sCopyOfDBName = m_poConnection->getDBName();
	SAPDB_DynamicUTF8String		sCopyOfDBUser = m_poConnection->getDBUser();
	SAPDB_DynamicUTF8String		sCopyOfPassWord = m_poConnection->getPassWord();
	SAPDB_DynamicUTF8String		sCopyOfODBCDriver = m_poConnection->getODBCDriver();
	
	if(sCopyOfDBServer.Size() > 0)
		assert(sCopyOfDBServer.ConvertToASCII_Latin1((char*)sAsciiDBServer,sAsciiDBServer+sizeof (sAsciiDBServer),szpConvert) == SAPDB_UTF8Basis::Success);

	if(sCopyOfDBName.Size() > 0)
		assert(sCopyOfDBName.ConvertToASCII_Latin1((char*)sAsciiDBName,sAsciiDBName+sizeof (sAsciiDBName),szpConvert) == SAPDB_UTF8Basis::Success);

	if(sCopyOfDBUser.Size() > 0)
		assert(sCopyOfDBUser.ConvertToASCII_Latin1((char*)sAsciiDBUser,sAsciiDBUser+sizeof (sAsciiDBUser),szpConvert) == SAPDB_UTF8Basis::Success);

	if(sCopyOfPassWord.Size() > 0)
		assert(sCopyOfPassWord.ConvertToASCII_Latin1(sAsciiPassWord,sAsciiPassWord+sizeof (sAsciiPassWord),szpConvert) == SAPDB_UTF8Basis::Success);

	if(sCopyOfODBCDriver.Size() > 0)
		assert(sCopyOfODBCDriver.ConvertToASCII_Latin1(sAsciiODBCDriver,sAsciiODBCDriver+sizeof (sAsciiODBCDriver),szpConvert) == SAPDB_UTF8Basis::Success);
	
	
	m_pDBFS = wd101CreateDbcDBFS(m_poConnection->getDBCPool(),
								  "",
								  sAsciiODBCDriver,
								  sAsciiDBServer,
								  sAsciiDBName,
								  sAsciiDBUser,
								  sAsciiPassWord,
								  "",
								  m_nFSPrefix);
	if (m_pDBFS == NULL) {

		sError = (SAPDB_UTF8*)ST_ERR004;        
		return SAPDB_FALSE;
	}		

	return SAPDB_TRUE;

}


SAPDB_Int2  Studio_DBFSObject::openDir(const SAPDB_DynamicUTF8String &sPath,SAPDB_DynamicUTF8String &sError)
{

	//if there is no database filesystem, get one
	if (m_pDBFS == NULL ){

		if (getValidDBFS(sError) == SAPDB_FALSE) return -1;		
	}

	SAPDB_Char	sAsciiPath[256];
	memset(sAsciiPath,0,256);
	SAPDB_Char* szpConvert = NULL;

	SAPDB_DynamicUTF8String		sCopyOfPath = sPath;
	assert(sCopyOfPath.ConvertToASCII_Latin1((char*)sAsciiPath,sAsciiPath+sizeof (sAsciiPath),szpConvert) == SAPDB_UTF8Basis::Success);	

	SAPDB_Int2 lfno = wd101OpenDir(m_pDBFS,sAsciiPath,type_asc_name_asc);  
	if ( lfno == -1 ) {    
         
		 if (getDBFSError(sError) == SAPDB_FALSE)			 
			return -1;		 
		 else{
			 //open directory again, for reconnect
			 lfno = wd101OpenDir(m_pDBFS,sAsciiPath, type_asc_name_asc);  
			 if ( lfno == -1 ) {

				getDBFSError(sError);				
				return -1;
			}
		}
    }

	return lfno;
}

SAPDB_Bool  Studio_DBFSObject::status(SAPDB_DynamicUTF8String &sPath,st_db_stat &Info,SAPDB_DynamicUTF8String &sError)
{

	SAPDB_Char	sAsciiPath[1024];
	memset(sAsciiPath,0,1024);	
	SAPDB_Char* szpConvert = NULL;

	SAPDB_DynamicUTF8String strCopyOfPath = sPath;
	assert(strCopyOfPath.ConvertToASCII_Latin1((char*)sAsciiPath,sAsciiPath+sizeof (sAsciiPath),szpConvert) == SAPDB_UTF8Basis::Success);	

	sError.Erase();

	if ( wd101Stat(m_pDBFS,(const char*)(char*) sAsciiPath, &Info) == 0 ) {
	    
        getDBFSError(sError);
		return SAPDB_FALSE;			
    }

	return SAPDB_TRUE;
}

SAPDB_Bool  Studio_DBFSObject::closeDir(SAPDB_Int2 lfno,SAPDB_DynamicUTF8String &sError)
{

	sError.Erase();

	if (wd101CloseDir(m_pDBFS, lfno) == 0 ) {    
         
		 getDBFSError(sError);		
		 return SAPDB_FALSE;
    }        	
	return SAPDB_TRUE;
}

SAPDB_Bool  Studio_DBFSObject::getDirEntry(SAPDB_Int2 lfno,SAPDB_DynamicUTF8String &sOut)
{

	SAPDB_Char sBuffer[1024];
	memset(sBuffer,0,1024);
	SAPDB_Bool bRet = SAPDB_TRUE;

	sOut.Erase();

    //get directory entries
	SAPDB_Int4 lRet = wd101GetDirEntry(m_pDBFS, lfno, sBuffer, sizeof (sBuffer), 0);
    if (lRet == -1) {    
         
		 sapdbwa_ErrP pError = wd101GetErr(m_pDBFS);
		 if (sapdbwa_GetErrId(pError) != sapdbwa_Error_NoError &&
			(sapdbwa_GetErrId(pError) != sapdbwa_Error_NoMoreEntries)){

			sapdbwa_CopyErrMsg(pError, sBuffer, sizeof (sBuffer));			
		 }
		 bRet =  SAPDB_FALSE;
	}

	if(strlen(sBuffer) > 0){

		 //NOT UNICODE BECAUSE OF WA
		sOut.ConvertFromASCII_Latin1(
								(SAPDB_DynamicUTF8String::ASCIIConstPointer)sBuffer,
								(SAPDB_DynamicUTF8String::ASCIIConstPointer)sBuffer+strlen(sBuffer));
	}
	
	 return bRet;
}

SAPDB_Bool Studio_DBFSObject::createFolder(const SAPDB_DynamicUTF8String &sCompletePath,
										   SAPDB_DynamicUTF8String &sError)
{

	SAPDB_Char	sAsciiPath[1024];
	memset(sAsciiPath,0,1024);	
	SAPDB_Char* szpConvert = NULL;

	SAPDB_DynamicUTF8String strCopyOfPath = sCompletePath;
	assert(strCopyOfPath.ConvertToASCII_Latin1((char*)sAsciiPath,sAsciiPath+sizeof (sAsciiPath),szpConvert) == SAPDB_UTF8Basis::Success);	
	
	if (wd101MkDir(m_pDBFS,sAsciiPath) == sapdbwa_False){
    
        getDBFSError(sError);		
		return SAPDB_FALSE;
    }

	return SAPDB_TRUE;
}


SAPDB_Bool Studio_DBFSObject::deleteFolder(const SAPDB_DynamicUTF8String &sCompletePath,
										   SAPDB_DynamicUTF8String &sError)
{
	sError.Erase();

	SAPDB_Char	sAsciiPath[1024];
	memset(sAsciiPath,0,1024);	
	SAPDB_Char* szpConvert = NULL;

	SAPDB_DynamicUTF8String strCopyOfPath = sCompletePath;
	assert(strCopyOfPath.ConvertToASCII_Latin1((char*)sAsciiPath,sAsciiPath+sizeof (sAsciiPath),szpConvert) == SAPDB_UTF8Basis::Success);		
	
	if (wd101RmDir(m_pDBFS,sAsciiPath,sapdbwa_True) == sapdbwa_False){
    
        getDBFSError(sError);		
		return SAPDB_FALSE;
    }

	return SAPDB_TRUE;
}

SAPDB_Bool Studio_DBFSObject::deleteFile(const SAPDB_DynamicUTF8String &sCompletePath,
										 SAPDB_DynamicUTF8String &sError)
{

	sError.Erase();

	SAPDB_Char	sAsciiPath[1024];
	memset(sAsciiPath,0,1024);	
	SAPDB_Char* szpConvert = NULL;

	SAPDB_DynamicUTF8String strCopyOfPath = sCompletePath;
	assert(strCopyOfPath.ConvertToASCII_Latin1((char*)sAsciiPath,sAsciiPath+sizeof (sAsciiPath),szpConvert) == SAPDB_UTF8Basis::Success);		
	
	if (wd101Delete(m_pDBFS,sAsciiPath) == sapdbwa_False){
    
        getDBFSError(sError);		
		return SAPDB_FALSE;
    }

	return SAPDB_TRUE;
}

SAPDB_Bool Studio_DBFSObject::renameItem(const SAPDB_DynamicUTF8String &sCompleteOldPath,
										 const SAPDB_DynamicUTF8String &sCompleteNewPath,
										 SAPDB_DynamicUTF8String &sError)
{
	sError.Erase();

	SAPDB_Char	sAsciiNewPath[1024];
	memset(sAsciiNewPath,0,1024);	
	SAPDB_Char* szpConvert = NULL;

	SAPDB_DynamicUTF8String strCopyOfNewPath = sCompleteNewPath;
	assert(strCopyOfNewPath.ConvertToASCII_Latin1((char*)sAsciiNewPath,sAsciiNewPath+sizeof (sAsciiNewPath),szpConvert) == SAPDB_UTF8Basis::Success);		

	SAPDB_Char	sAsciiOldPath[1024];
	memset(sAsciiOldPath,0,1024);	

	SAPDB_DynamicUTF8String strCopyOfOldPath = sCompleteOldPath;
	assert(strCopyOfOldPath.ConvertToASCII_Latin1((char*)sAsciiOldPath,sAsciiOldPath+sizeof (sAsciiOldPath),szpConvert) == SAPDB_UTF8Basis::Success);		
	
	if (wd101Rename(m_pDBFS,sAsciiOldPath,sAsciiNewPath) == sapdbwa_False){
    
        getDBFSError(sError);		
		return SAPDB_FALSE;
    }

	return SAPDB_TRUE;
}


SAPDB_Bool Studio_DBFSObject::moveItem(const SAPDB_DynamicUTF8String &sCompleteOldPath,
									   const SAPDB_DynamicUTF8String &sCompleteNewPath,
									   SAPDB_DynamicUTF8String &sError)
{

	sError.Erase();

	SAPDB_Char	sAsciiNewPath[1024];
	memset(sAsciiNewPath,0,1024);	
	SAPDB_Char* szpConvert = NULL;

	SAPDB_DynamicUTF8String strCopyOfNewPath = sCompleteNewPath;
	assert(strCopyOfNewPath.ConvertToASCII_Latin1((char*)sAsciiNewPath,sAsciiNewPath+sizeof (sAsciiNewPath),szpConvert) == SAPDB_UTF8Basis::Success);		

	SAPDB_Char	sAsciiOldPath[1024];
	memset(sAsciiOldPath,0,1024);	

	SAPDB_DynamicUTF8String strCopyOfOldPath = sCompleteOldPath;
	assert(strCopyOfOldPath.ConvertToASCII_Latin1((char*)sAsciiOldPath,sAsciiOldPath+sizeof (sAsciiOldPath),szpConvert) == SAPDB_UTF8Basis::Success);		

	//first check if folder exists, because move create a item if it does noit exits
	// here a error code is wanted

	SAPDB_Int2 lfno = openDir(sCompleteNewPath,sError);  
    if ( lfno == -1 )  return SAPDB_FALSE;

	if(closeDir(lfno,sError) == SAPDB_FALSE) return SAPDB_FALSE;
	
	if (wd101Mv(m_pDBFS,sAsciiOldPath,sAsciiNewPath) == sapdbwa_False){
    
        getDBFSError(sError);		
		return SAPDB_FALSE;
    }

	return SAPDB_TRUE;
}


SAPDB_Bool Studio_DBFSObject::copyItem(const SAPDB_DynamicUTF8String &sCompleteOldPath,
										 const SAPDB_DynamicUTF8String &sCompleteNewPath,
										 SAPDB_DynamicUTF8String &sError)
{

	SAPDB_Char	sAsciiNewPath[1024];
	memset(sAsciiNewPath,0,1024);	
	SAPDB_Char* szpConvert = NULL;

	SAPDB_DynamicUTF8String strCopyOfNewPath = sCompleteNewPath;
	assert(strCopyOfNewPath.ConvertToASCII_Latin1((char*)sAsciiNewPath,sAsciiNewPath+sizeof (sAsciiNewPath),szpConvert) == SAPDB_UTF8Basis::Success);		

	SAPDB_Char	sAsciiOldPath[1024];
	memset(sAsciiOldPath,0,1024);	

	SAPDB_DynamicUTF8String strCopyOfOldPath = sCompleteOldPath;
	assert(strCopyOfOldPath.ConvertToASCII_Latin1((char*)sAsciiOldPath,sAsciiOldPath+sizeof (sAsciiOldPath),szpConvert) == SAPDB_UTF8Basis::Success);		
	
	if (wd101Copy(m_pDBFS,sAsciiOldPath,sAsciiNewPath) == sapdbwa_False){
    
        getDBFSError(sError);		
		return SAPDB_FALSE;
    }

	return SAPDB_TRUE;
}

SAPDB_Int2 Studio_DBFSObject::createFile(const SAPDB_DynamicUTF8String &sCompletePath,
										 SAPDB_DynamicUTF8String &sError)
{
	sError.Erase();

	SAPDB_Char	sAsciiPath[1024];
	memset(sAsciiPath,0,1024);	
	SAPDB_Char* szpConvert = NULL;

	SAPDB_DynamicUTF8String strCopyOfPath = sCompletePath;
	assert(strCopyOfPath.ConvertToASCII_Latin1((char*)sAsciiPath,sAsciiPath+sizeof (sAsciiPath),szpConvert) == SAPDB_UTF8Basis::Success);		
	
	SAPDB_Int2 lfno = wd101Create(m_pDBFS,sAsciiPath,access_write);
	     
	if(lfno == -1) getDBFSError(sError);							

	return lfno;
}


SAPDB_Int2 Studio_DBFSObject::openFile(const SAPDB_DynamicUTF8String &sCompletePath,
									   SAPDB_DynamicUTF8String &sError)
{
	sError.Erase();

	SAPDB_Char	sAsciiPath[1024];
	memset(sAsciiPath,0,1024);	
	SAPDB_Char* szpConvert = NULL;

	SAPDB_DynamicUTF8String strCopyOfPath = sCompletePath;
	assert(strCopyOfPath.ConvertToASCII_Latin1((char*)sAsciiPath,sAsciiPath+sizeof (sAsciiPath),szpConvert) == SAPDB_UTF8Basis::Success);		
	
	SAPDB_Int2 lfno = wd101Open(m_pDBFS,sAsciiPath,access_write);
	     
	if(lfno == -1) getDBFSError(sError);							

	return lfno;
}


SAPDB_Int4 Studio_DBFSObject::writeFile(SAPDB_Int2   lfno,
										void         *buffer,
										SAPDB_Int4   nBufferSize,
										SAPDB_DynamicUTF8String &sError)
{

	sError.Erase();

	SAPDB_Int4 bytewritten = wd101Write(m_pDBFS,lfno,buffer,nBufferSize);	     
	if(bytewritten == -1) getDBFSError(sError);							

	return bytewritten;
}

SAPDB_Bool Studio_DBFSObject::closeFile(SAPDB_Int2 lfno,SAPDB_DynamicUTF8String &sError)
{	
	
	sError.Erase();

	if (wd101Close(m_pDBFS,lfno) == sapdbwa_False){
    
        getDBFSError(sError);		
		return SAPDB_FALSE;
    }

	return SAPDB_TRUE;
}