/*
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/

/*!**********************************************************************

  module: StudioWeb_WebRequest.cpp

  -----------------------------------------------------------------------

  mail to:  sapdb-general@sap.com

  special area: WebSQL

  description:  functions for web query service

  see also:		for sapdbwa_WebAgent in hwd02waapi.h

  -----------------------------------------------------------------------

  copyright:                Copyright by SAP AG, 2000

************************************************************************/


/*=======================================================================
 *  INCLUDES
 *=====================================================================*/

//html code fragments
#include "SAPDB/SQLStudio/WebQuery/StudioWeb_HtmlFragments.h"

//error messages
#include "SAPDB/SQLStudio/WebQuery/StudioWeb_Error.h"

#include "SAPDB/SQLStudio/WebQuery/StudioWeb_WebRequest.hpp"

#include "SAPDB/SQLStudio/WebQuery/StudioWeb_WebQuery.hpp"

#include "SAPDBCommon/Web/SAPDBWeb_TemplateMsgBox.hpp"

#include "SAPDBCommon/SAPDB_DynamicUTF8String.hpp"



//global variables
static sapdbwa_DBCPoolP m_pDBCPool = NULL;

SAPDB_DynamicUTF8String m_sODBCDriver; 


#define WEBODBC		"SAP DB Web 7.3"

//called on closeWebSession
sapdbwa_Bool destroySessionData(sapdbwa_WebAgent &wa,void *pData)
{	

	StudioWeb_WebQuery *poWebQuery = (StudioWeb_WebQuery*) pData;
	if (poWebQuery != NULL){

		delete poWebQuery;  
		poWebQuery = NULL;
	}

	return sapdbwa_True;
}



sapdbwa_Int2 WebSQLInit(sapdbwa_WebAgent &wa)
{

	 wa.WriteLogMsg("StudioWeb-WebSQLInit\n");

	 char sValue[32];
	 memset(sValue,0,32);	 

	 int nTimeOut = sapdbwa_DBC_NO_TIMEOUT;
	 //get user defined value for db timeout
	 if (wa.GetParam(WQ_DBTIMEOUT_TAG,sValue,32) == 1) nTimeOut = atoi(sValue);	 

#if defined(WIN32)
	 /*char sODBCDriver[256];
	 memset(sODBCDriver,0,256);	 

	 //get user defined value for db timeout
	 if (wa.GetParam(WQ_ODBCDRIVER_TAG,sODBCDriver,256) == SAPDB_FALSE) wa.WriteLogMsg(WQ_ERR023);	 

	 //NOT UNICODE BECAUSE OF WA
	 m_sODBCDriver = (SAPDB_UTF8*)sODBCDriver;*/

	 m_sODBCDriver = WEBODBC;
#endif	 

	 m_pDBCPool = sapdbwa_CreateDBCPool(nTimeOut);    
     if(m_pDBCPool == NULL)  wa.WriteLogMsg(WQ_ERR018);
     
	 wa.SetUseWebSessions( sapdbwa_True );
	 	
    return 1;
}; 

sapdbwa_Int2 WebSQLExit(sapdbwa_WebAgent &wa)
{
	wa.WriteLogMsg(_T("StudioWeb-WebSQLExit\n"));

	if(m_pDBCPool != NULL){

		sapdbwa_DestroyDBCPool( m_pDBCPool );	
		m_pDBCPool = NULL;
	}		
    return 1;
}; 




/*!-------------------------------------------------------------------------
    function:     WebSQLService
    description:  called by webagent 
    arguments:	  [in] wa, webagent object	
				  [in] req, webagent http request object
				  [in] rep, webagent http reply object
				  
    return value: SAPDB_TRUE if success
    --------------------------------------------------------------------------*/
sapdbwa_Int2 WebSQLService(sapdbwa_WebAgent &wa,sapdbwa_HttpRequest &req,sapdbwa_HttpReply &rep)
{


   sapdbwa_WebSession    webSession;
   StudioWeb_WebQuery    *poWebQuery  = NULL;

   SAPDB_Bool            bValidSession = SAPDB_FALSE;
   SAPDB_Bool			 bCloseSession = SAPDB_FALSE;

#ifdef QUICK

   const char		*queryString = req.GetQueryString();
   if (queryString == NULL){
	   wa.WriteLogMsg("queryString is null");	 
	   wa.WriteLogMsg("\n");
   }
   else{
	  wa.WriteLogMsg(queryString);	 
	  wa.WriteLogMsg("\n");
   }
#endif

   
	// look for existing Session
	if(wa.GetWebSession( webSession ) == sapdbwa_True) {

      poWebQuery = (StudioWeb_WebQuery*) webSession.GetSessionData();
      if (poWebQuery == NULL)  
		  wa.CloseWebSession (webSession);	  	
	  else 
		  bValidSession = SAPDB_TRUE;

    } 

	if (wa.GetErr().GetErrId() == sapdbwa_Error_WebSessionTimeout) {
  
		SAPDBWeb_TemplateMsgBox aMsgBox(wa);

		aMsgBox.setMessage(WQ_ERR028);
		aMsgBox.setButtonText("OK");	
		
		aMsgBox.setButtonTarget(WQ_TARGETTOP);
		aMsgBox.setButtonAction(WQ_SERVICEDLL);
		aMsgBox.writePage(SAPDBWeb_TemplateWriterWA(rep));
	} 
	else{
		if(!bValidSession){

			// create session if not exist
			poWebQuery = new StudioWeb_WebQuery();			
			if (poWebQuery != NULL) {

				poWebQuery->setODBCDriver(m_sODBCDriver);
				//poWebQuery->setDBCPool((const sapdbwa_DBCPoolP)m_pDBCPool);
				poWebQuery->setDBCPool(m_pDBCPool);
				if (wa.StartWebSession(webSession) == sapdbwa_True) {
					if (webSession.SetSessionData(poWebQuery, destroySessionData) == sapdbwa_True) 

						bValidSession = SAPDB_TRUE;					
					else
						delete poWebQuery;
					
				}			
				else 
					delete poWebQuery;				
			}// end if
		}// end if
	}// end if
	  

	if (bValidSession){

		poWebQuery->doService(wa, req, rep,bCloseSession);          
		if (bCloseSession)  wa.CloseWebSession (webSession);    
	}
	else 
		sendErrorMessageBox(wa,rep,WQ_ERR016);


	return sapdbwa_RTC_Accepted;

};











