/*!**************************************************************************

  module      : Trace_Writer.cpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area:
  description : 

  last changed: 2000-11-27  18:15
  see also    :

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "ggg00.h"      // PASCAL: GG_kernel_constants_and_types
#include "hbd12_2.h"    // PASCAL: b12dump
#include "heo51.h"	    // RTE   : vfinish
#include "heo57.h"      // RTE   : Vmalloc/Vmfree
#include "heo58.h"      // RTE   : vIsDumpWanted/vIsTraceWanted

#include "SAPDB/Trace/Trace_Writer.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define ENTER_REGIONS	1
#define TRUE            1
#define FALSE           0


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/



/*===========================================================================*
 *  GLOBAL VARIABLES                                                         *
 *===========================================================================*/

Trace_Writer*	Trace_Writer::m_Instance = NULL;


/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/



/*===========================================================================*
 *  STATIC/INLINE FUNCTIONS (PROTOTYPES)                                     *
 *===========================================================================*/



/*===========================================================================*
 *  METHODS                                                                  *
 *===========================================================================*/

/*---------------------------------------------------------------------------*/
/* Singleton interface                                                       */
/*---------------------------------------------------------------------------*/

bool
Trace_Writer::CreateInstance( tsp00_TaskId TaskId )
{
    if ( NULL == m_Instance )
    {
        tsp00_Bool	bOkay	= FALSE;
        void		*pObj	= NULL;

        vmalloc( sizeof( Trace_Writer ), &pObj, &bOkay );

        if (TRUE != bOkay)
            return( false );

        m_Instance = REINTERPRET_CAST( Trace_WriterPtr, new( pObj ) Trace_Writer( TaskId ));

        return( m_Instance->Initialize());
    }
    return( true );
};

/*---------------------------------------------------------------------------*/

Trace_Writer::Trace_Writer( tsp00_TaskId &TaskId )
        :
        m_TaskId( TaskId ),
        m_TraceManager( Trace_Manager::GetInstance() ),
        m_pPage( NULL) // PTS 1120896 TS 2003-03-03
{
};

/*---------------------------------------------------------------------------*/

bool
Trace_Writer::Initialize() // PTS 1120896 TS 2003-03-03
{
    if( m_pPage != NULL )
        return( true );

    tsp00_Bool	bOkay	  = FALSE;
    tsp00_Int4  available = 0;

    vnewbuf( 1, &available, &m_pPage, &bOkay );

    if(( TRUE != bOkay ) || (NULL == m_pPage ))
        return( false );
    else
        return( true );
}

/*---------------------------------------------------------------------------*/

void
Trace_Writer::Dump( tsp00_TaskId &TaskId )
{
    tsp00_VfReturn		HostError;
    tsp00_ErrText		ErrorText;
    tgg00_DumpStateSet  DumpStateSet;
    tsp00_VFilename     DumpFile = m_TraceManager.GetDumpFile();

    DumpStateSet.clear();
    DumpStateSet.addElement( dumpAll_egg00 );

    memset( m_pPage, 0xFE, sizeof( tsp00_Page ));

    b12dump ( TaskId, ! ENTER_REGIONS, DumpStateSet,
              DumpFile, *m_pPage, HostError, ErrorText );
};

/*---------------------------------------------------------------------------*/

tsp00_TaskId
Trace_Writer::GetTaskId()
{
    return( m_TaskId );
};

/*---------------------------------------------------------------------------*/

void
Trace_Writer::Run()
{
    m_TraceManager.TraceWriterRegistration( m_TaskId );

    while( true )
    {
        m_TraceManager.WaitForWork( m_TaskId );

        // In case of vabort is called the trace writer starts
        // to flush the complete trace and the dump
        if ( TRUE == vIsTraceWanted() )
        {
            m_TraceManager.Flush( m_TaskId );
            if( TRUE == vIsDumpWanted() )
                Dump( m_TaskId );
            break;
        }
        else
            m_TraceManager.Flush( m_TaskId );
    }
    vfinish();
};

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/
