#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "SAPDB/WebAgent/HTTP/WAHTTP_Log.h"
#include "SAPDB/WebAgent/HTTP/WAHTTP_Misc.h"
#include "SAPDB/WebAgent/HTTP/WAHTTP_Connect.h"
#include "SAPDB/WebAgent/HTTP/WAHTTP_HTTP.h"
#include "SAPDB/WebAgent/HTTP/WAHTTP_Agent.h"
#include "SAPDB/WebAgent/HTTP/WAHTTP_Server.h"
#include "SAPDB/WebAgent/HTTP/WAHTTP_IniPar.h"
#include "SAPDB/WebAgent/HTTP/WAHTTP_FileAgent.h"
#include "SAPDB/WebAgent/HTTP/WAHTTP_WebAgent.h"
#include "SAPDB/WebAgent/HTTP/WAHTTP_Memory.h"

static struct st_agent
{
 int			Initialized;
 struct st_agent_client *AgentClientHandle;
 int			(*AgentInit)(struct st_agent_server *, struct st_agent_client **);
 enum en_af_returncode	(*AgentFunc)(struct st_agent_server *, struct st_agent_client *, struct st_agent_req *, struct st_agent_rep *);
 int			(*AgentExit)(struct st_agent_server *, struct st_agent_client *);
} Agents[] = 
{
 { FALSE, NULL,	InitWebAgent,	TryWebAgent,	ExitWebAgent },
 { FALSE, NULL,	InitFileAgent,	TryFileAgent,	ExitFileAgent },
};

#define NOOF_AGENTS	sizeof(Agents)/sizeof(struct st_agent)

static struct st_agent_server AgentServer;

static struct st_charp_int
{
 char	*V;
} Hmap[] =
{
 HTTP_SET_COOKIE,
 HTTP_VERSION,
 HTTP_EXPIRES,
 HTTP_CONTENT_TYPE,
 HTTP_CONTENT_LENGTH,
 HTTP_LAST_MODIFIED,
 HTTP_LOCATION,
 HTTP_DATE,
 HTTP_IF_MODIFIED_SINCE,
 HTTP_SERVER,
 HTTP_CONNECTION,
 HTTP_HOST,
 HTTP_PORT,
};
#define MAX_Hmap_LG 100

static struct st_charp
{
 char	*V;
} Smap[] =
{
 HTTP_STATUS_OK,
 HTTP_STATUS_NOT_MODIFIED,
 HTTP_STATUS_BAD_REQUEST,
 HTTP_STATUS_NOT_FOUND,
 HTTP_STATUS_LENGTH_REQUIRED,
 HTTP_STATUS_CREATED,
 HTTP_STATUS_NO_CONTENT,
 HTTP_STATUS_MOVE_PERM,
 HTTP_STATUS_REDIRECT,
 HTTP_STATUS_UNAUTHORIZED,
 HTTP_STATUS_FORBIDDEN,
 HTTP_STATUS_METHOD_NOT_ALLOWED,
 HTTP_STATUS_CONFLICT,
 HTTP_STATUS_SERVER_ERROR,
 HTTP_STATUS_NOT_IMPLEMENTED,
};
#define MAX_Smap_LG 100


static int AgentAddHeader_char
(
 struct st_agent_rep	*rep,
 int			Header,
 char			*Value
)
{
 return(AddHeader_char(rep->ServerRep, Hmap[Header].V, Value));
}

static int AgentAddHeaderByName_char
(
 struct st_agent_rep	*rep,
 char			*Header,
 char			*Value
)
{
 return(AddHeader_char(rep->ServerRep, Header, Value));
}

static int AgentAddHeader_date
(
 struct st_agent_rep	*rep,
 int			Header,
 time_t			Value
)
{
 return(AddHeader_date(rep->ServerRep, Hmap[Header].V, Value));
}

static int AgentAddHeader_size_t
(
 struct st_agent_rep	*rep,
 int			Header,
 size_t			Value
)
{
 return(AddHeader_size_t(rep->ServerRep, Hmap[Header].V, Value));
}

static char *AgentGetHeaderValue
(
 struct st_agent_req	*req,
 int			Header
)
{
 switch(Header)
 {
	case AGENT_HOST	: return(req->ServerReq->Host);
	case AGENT_PORT	: return(req->ServerReq->Port);
	default		: return(FindHeader(req->ServerReq, Hmap[Header].V));
 }
}

static char *AgentGetHeaderValueByName
(
 struct st_agent_req	*req,
 char			*Header
)
{
 return(FindHeader(req->ServerReq, Header));
}

static int AgentGetHeaderNo
(
 struct st_agent_req	*req,
 size_t			HeaderNo,
 char			**HeaderName,
 char			**HeaderValue
)
{
 return(GetHeaderNo(req->ServerReq, HeaderNo, HeaderName, HeaderValue));
}

static size_t AgentReceive
(
 struct st_agent_req	*req,
 char			*b,
 size_t			n
)
{
 return(req->ServerReq->Body.read(&req->ServerReq->Body, b, n));
}

static size_t AgentSend
(
 struct st_agent_rep	*rep,
 char			*b,
 size_t			n
)
{
 size_t	slg;

 if (SendHttpBody(rep->ServerRep, b, n, &slg))
	return(slg);
 else
	return(0);
}

static int AgentSendHeader
(
 struct st_agent_rep	*rep,
 int			Status
)
{
 char	tmph[sizeof(HTTP_VERSION " ") + MAX_Smap_LG + 1];

 strcpy(tmph, HTTP_VERSION " ");
 strcat(tmph, Smap[Status].V);
 return(SendHttpHeader(tmph, rep->ServerRep));
}

int InitAgents
(
 void
)
{
 struct st_agent	*a;

 memset(&AgentServer, 0, sizeof(struct st_agent_server));
 AgentServer.AddHeader_char		= AgentAddHeader_char;
 AgentServer.AddHeaderByName_char	= AgentAddHeaderByName_char;
 AgentServer.AddHeader_date		= AgentAddHeader_date;
 AgentServer.AddHeader_size_t		= AgentAddHeader_size_t;
 AgentServer.GetHeaderValue		= AgentGetHeaderValue;
 AgentServer.GetHeaderValueByName	= AgentGetHeaderValueByName;
 AgentServer.GetHeaderNo		= AgentGetHeaderNo;
 AgentServer.Receive			= AgentReceive;
 AgentServer.Send			= AgentSend;
 AgentServer.SendHeader			= AgentSendHeader;

 for (a=Agents; a<Agents+NOOF_AGENTS; a++)
	if (a->AgentInit)
		a->Initialized = a->AgentInit(&AgentServer, &a->AgentClientHandle);

 AgentServer.InitOk = TRUE;
 return(TRUE);
}

int TryAgents
(
 struct st_http_client	*client,
 struct st_http_req	*HttpReq,
 int			*Persistent
)
{
 struct st_agent	*a;
 struct	st_agent_req	AgentRequest;
 struct	st_agent_rep	AgentReply;

 *Persistent = HttpIsPersistent(HttpReq);
 AgentRequest.ServerReq = HttpReq;
 AgentRequest.method = HttpReq->method;
 AgentRequest.uri = HttpReq->uri;
 AgentRequest.version = HttpReq->hversion;
 AgentRequest.params = HttpReq->params;
 AgentRequest.querystring = HttpReq->querystring;
 AgentReply.ServerRep = client;

 for (a=Agents; a<Agents+NOOF_AGENTS; a++)
	if (a->Initialized)
		switch (a->AgentFunc(&AgentServer, a->AgentClientHandle, &AgentRequest, &AgentReply))
		{
		case af_accepted:
			return(TRUE);
		case af_ignored:
			continue;
		case af_error:
			return(FALSE);
		}

 return(FALSE);
}

int ExitAgents
(
 void
)
{
 struct st_agent	*a;

 for (a=Agents; a<Agents+NOOF_AGENTS; a++)
	if (a->Initialized)
		if (a->AgentExit)
			a->AgentExit(&AgentServer, a->AgentClientHandle);
 return(TRUE);
}

