#include <string.h>
#include <stdio.h>
#include <time.h>

#ifdef WIN32
#include <winsock2.h>
#else
#include <sys/socket.h>
#endif

#include "SAPDB/WebAgent/HTTP/WAHTTP_Misc.h"
#include "SAPDB/WebAgent/HTTP/WAHTTP_Connect.h"
#include "SAPDB/WebAgent/HTTP/WAHTTP_HTTP.h"
#include "SAPDB/WebAgent/HTTP/WAHTTP_Log.h"
#include "SAPDB/WebAgent/HTTP/WAHTTP_Memory.h"

#define TCP_SEND_BLOCK_SZ	1400

#define FMT_HTTP_HEADER_CHAR	"%s: %s%s"
#define FMT_HTTP_HEADER_SIZE_T	"%s: %ld%s"

int AddHeader_char
(
 struct st_http_client	*client,
 char			*Header,
 char			*Value
)
{
 size_t	l = strlen(Header) + 1 + (Value ? strlen(Value) : 0) + HTTP_HEADER_SEP_LG + 1;

 if (!DSextent(&client->HeadersLine, l))
	return(FALSE);

 /* MaO: check if header includes a colon or not	*/
 client->HeadersLine->lg += sprintf(&client->HeadersLine->str[client->HeadersLine->lg], FMT_HTTP_HEADER_CHAR, Header, Value ? Value : "", HTTP_HEADER_SEP);
 return(TRUE);
}

int AddHeader_size_t
(
 struct st_http_client	*client,
 char			*Header,
 size_t			Value
)
{
 size_t	l = strlen(Header) + 1 + 15 + HTTP_HEADER_SEP_LG + 1;

 if (!DSextent(&client->HeadersLine, l))
	return(FALSE);

 client->HeadersLine->lg += sprintf(&client->HeadersLine->str[client->HeadersLine->lg], FMT_HTTP_HEADER_SIZE_T, Header, Value, HTTP_HEADER_SEP);
 return(TRUE);
}

int AddHeader_date
(
 struct st_http_client	*client,
 char			*Header,
 time_t			Value
)
{
 size_t	tmplg;
 char	tmp[50];

/*#pragma ctime-is-not-thread-safe*/
 tmplg = sprintf(tmp, ctime(&Value));

/* remove the trailing eoln which is added by ctime on WINDOWS */
 if (tmp[tmplg - 1] == '\n')
 {
	tmp[tmplg - 1] = '\0';
	tmplg--;
 }

 strcat(tmp, " GMT");

 AddHeader_char(client, Header, tmp);

 return(TRUE);
}

static int SendHeader
(
 SOCKET		cs,
 char		*HeaderFieldName,
 size_t		HeaderFieldNameLg,
 char		*HeaderFieldValue,
 size_t		HeaderFieldValueLg
)
{
 size_t	slg;

 if (HeaderFieldNameLg > 0)
 {
	slg = send (cs,  HeaderFieldName, HeaderFieldNameLg, 0);
	if (SOCKET_ERROR == slg)
	{
		SyscallError("send to client failed", slg);
		return(FALSE);
	}
	else if (slg != HeaderFieldNameLg)
	{
		SyscallError("send to client failed (slg != rlg)", 0);
		return(FALSE);
	}
 }

 if (HeaderFieldValueLg > 0)
 {
	slg = send (cs,  HeaderFieldValue, HeaderFieldValueLg, 0);
	if (SOCKET_ERROR == slg)
	{
		SyscallError("send to client failed", slg);
		return(FALSE);
	}
	else if (slg != HeaderFieldValueLg)
	{
		SyscallError("send to client failed (slg != rlg)", 0);
		return(FALSE);
	}
 }

 slg = send (cs,  HTTP_HEADER_SEP, HTTP_HEADER_SEP_LG, 0);
 if (SOCKET_ERROR == slg)
 {
	SyscallError("send to client failed", slg);
	return(FALSE);
 }
 else if (slg != HTTP_HEADER_SEP_LG)
 {
	SyscallError("send to client failed (slg != rlg)", 0);
	return(FALSE);
 }
 return(TRUE);
}

int SendHttpHeader
(
 char			*Status,
 struct st_http_client	*client
)
{
 size_t		slg;

/* send the headers */

/*Log(Status, strlen(Status));*/

 if (!SendHeader(client->tcp_con->s, Status, strlen(Status), NULL, 0))
	return(FALSE);

 if (client)
 {
/*Log(client->HeadersLine->str, client->HeadersLine->lg);*/
	slg = send (client->tcp_con->s,  client->HeadersLine->str, client->HeadersLine->lg, 0);
	if (SOCKET_ERROR == slg)
	{
		SyscallError("send to client failed", slg);
		return(FALSE);
	}
	else if (slg != client->HeadersLine->lg)
	{
		SyscallError("send to client failed (slg != rlg)", 0);
		return(FALSE);
	}

	if (!SendHeader(client->tcp_con->s, NULL, 0, NULL, 0))
		return(FALSE);
 }
 else
 {
	if (!SendHeader(client->tcp_con->s, NULL, 0, NULL, 0))
		return(FALSE);
 }
 client->HeadersSent = TRUE;
 return(TRUE);
}

int SendHttpBody
(
 struct st_http_client	*client,
 void			*b,
 size_t			n,
 size_t			*rn
)
{
 if (!client->HeadersSent)
	return(FALSE);

 *rn = send (client->tcp_con->s,  b, n, 0);
 if (SOCKET_ERROR == *rn)
 {
	SyscallError("send to client failed", *rn);
	return(FALSE);
 }
 else if (*rn != n)
 {
	SyscallError("send to client failed (*rn != n)", 0);
	return(FALSE);
 }
 else
	return(TRUE);
}

#define HTTP_PREFIX	"HTTP_"
#define HTTP_PREFIX_LG	sizeof(HTTP_PREFIX)-1

static int httpheadercmp
(
 char	*HeaderField,
 char	*Header
)
/* compare the forms:
 * HttpHeader:	If-Modified-Since
 * with	Header:	HTTP_IF_MODIFIED_SINCE
 * or   Header: If-Modified-Since
 */
{
 int	HttpPrefix = FALSE;
 size_t	n =0;
 int	rtc;

 if (!strncmp(HTTP_PREFIX, Header, HTTP_PREFIX_LG))
 {
	Header += HTTP_PREFIX_LG;
	HttpPrefix = TRUE;
 }

 while (HeaderField[n] && Header[n])
 {
	if (HttpPrefix)
	{
		if (HeaderField[n] == '-' && Header[n] != '_')
			return(HeaderField[n] - Header[n] != '_');
		if (0 != (rtc = toupper(HeaderField[n]) - Header[n]))
			return(rtc);
	}
	else
	{
		if (0 != (rtc = toupper(HeaderField[n]) - toupper(Header[n])))
			return(rtc);
	}
	n++;
 }
 return(HeaderField[n] - Header[n]);
}

/*===========================================================================*/

char *FindHeader( struct st_http_req *HttpReq,
                  char               *Header )
{

    char    *p;
    size_t   i;

    /* insert webdav */
    for (i=1; i < HttpReq->NoOfHeaderLines; i++) {
        p=HttpReq->HeaderNamesArray[i];

        if (!httpheadercmp(p, Header))
            return(HttpReq->HeaderValuesArray[i]);
    }

    return(NULL);
}

/*===========================================================================*/

void LogHeaders
(
 struct st_http_req	*HttpReq
)
{
 char tmp[100], *p;
 size_t		i;
 
 for (i=1, p=HttpReq->HeaderNamesArray[i];
      i < HttpReq->NoOfHeaderLines;
      p=HttpReq->HeaderNamesArray[++i])
 {
	sprintf(tmp, "Header %s", p);
	Log(p, -1);
 }
}

int GetHeaderNo
(
 struct st_http_req	*HttpReq,
 size_t			HeaderNo,
 char			**HeaderName,
 char			**HeaderValue
)
{
 if (0 == HeaderNo || HeaderNo >= HttpReq->NoOfHeaderLines)
	return(FALSE);

 *HeaderName = HttpReq->HeaderNamesArray[HeaderNo];
 *HeaderValue = HttpReq->HeaderValuesArray[HeaderNo];

 while (*HeaderValue && **HeaderValue == ' ') (*HeaderValue)++;
 return(TRUE);
}

int GetContentLength
(
 struct st_http_req	*HttpReq,
 size_t			*Cl
)
{
 char	*cl = FindHeader(HttpReq, HTTP_CONTENT_LENGTH);

 if (!cl)
	return(FALSE);
 else
 {
	*Cl = atol(cl);
	return(TRUE);
 }
}

int HttpIsReqConsistent
(
 struct st_http_req	*HttpReq,
 char			**Status
)
/* check consistency */
{
 if (HttpReq->HMajorVersion >= 1)
 {
	if (HttpReq->HMinorVersion < 1)
		HttpReq->PersistentRequested = FALSE;
	else
	{
		if (!HttpReq->ContentLengthValid)
			HttpReq->PersistentRequested = FALSE;
	}

	if (HttpReq->HMinorVersion >= 0)
	{
		if (HttpReq->Method == rq_post &&
		     !HttpReq->ContentLengthValid)
		{
			Error("HTTP/1.0 with POST method requires Content-Length", 0);
			*Status = HTTP_STATUS_LENGTH_REQUIRED;
			return(FALSE);
		}
	}

	if (HttpReq->HMinorVersion >= 1)
	{
		if (!HttpReq->Host)
		{
			Error("HTTP/1.0 requires Host header field", 0);
			*Status = HTTP_STATUS_BAD_REQUEST;
			return(FALSE);
		}
	}
 }
 else
 {
	/* keep-alive is not supported for 1.0 or lower */
	HttpReq->PersistentRequested = FALSE;

 }
 return(TRUE);
}

int HttpGetVersion
(
 char	*Hversion,
 int	*Major,
 int	*Minor
)
{
 if (!strncmp(Hversion, "HTTP/", 5))
 {
	return(2 == sscanf(Hversion+5, "%d.%d", Major, Minor));
 }
 else
 {
	Error("Invalid HTTP Version format", 0);
	return(FALSE);
 }
}

int HttpGetHost
(
 struct st_http_req	*HttpReq,
 char			**Host
)
{
 char	*h = FindHeader(HttpReq, HTTP_HOST);

 if (!h)
	return(FALSE);

 *Host = h;
 return(TRUE);
}

int HttpIsPersistent
(
 struct st_http_req	*HttpReq
)
{
 char	*h = FindHeader(HttpReq, HTTP_CONNECTION);

 return(h && !strcmp(h, HTTP_KEEP_ALIVE));
}

int GetHostPort
(
 struct st_http_req	*HttpReq,
 char			**Host,
 char			**Port
)
{
 *Host = FindHeader(HttpReq, HTTP_HOST);
 if (!*Host)
 {
	*Host = NULL;
	*Port = NULL;
	return(FALSE);
 }
 else
 {
	*Port = strchr(*Host, ':');
	if (*Port)
	{
		**Port = '\0';
		(*Port)++;
	}
	return(TRUE);
 }
}
