#ifdef xysss
#include <windows.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <ctype.h>

#include "SAPDB/WebAgent/HTTP/WAHTTP_Misc.h"
#include "SAPDB/WebAgent/HTTP/WAHTTP_Log.h"
#include "SAPDB/WebAgent/HTTP/WAHTTP_Connect.h"
#include "SAPDB/WebAgent/HTTP/WAHTTP_IniPar.h"
#include "SAPDB/WebAgent/HTTP/WAHTTP_Memory.h"

#include "hsp100.h"

#define INI_LOGFILEPATH		"LogFilePath"
#define DEF_LOGFILE		"httpreq.log"
#define DEF_ERRFILE		"httperr.log"

#define LOG_FORMAT_LEAD		"%s: %s"
#define LOG_FORMAT_TRAIL	""
#define TRC_FORMAT		"%s: "
#define MALLOC_ERROR		"Malloc error"

typedef tsp_int4	My_FILE;

static My_FILE	*LogFd = NULL;
static My_FILE	*ErrFd = NULL;
static Excl	ErrExcl;
static Excl	LogExcl;
static Excl	LogReqExcl;
static sapdbwa_Bool LogReq = sapdbwa_True;
static int LogRes = FALSE;
static char LogFilePath[MAX_FILENAME_LEN+1] = "";
static char DefLogFilePath[1000] = "";
static int TraceLevel = 0;
static char StartTimeStr[11];

static DynStr	ReqLogStart = NULL,
		ResLogStart = NULL;
static DynStr	ReqLogEnd = NULL,
		ResLogEnd = NULL;

static DynStr	ReqLeader = NULL,
		ReqTrailer = NULL;
static DynStr	ResLeader = NULL,
		ResTrailer = NULL;

void SetDefLogFilePath
(
 char	*path
)
{
 strcpymax(sizeof(DefLogFilePath), DefLogFilePath, path);
}

int InitLog
(
)
{
 char		pLogReq[100];
 char		Ipar[100];
 time_t		t;
 char		tmps[100];
 tsp5_rte_fileerror	err;
 tsp00_Bool		bOk;

 time(&t);
 strftime(StartTimeStr, sizeof(StartTimeStr), "%m%d%y%H%M", localtime(&t));

 initexcl(&LogExcl, 0, NULL);
 initexcl(&ErrExcl, 0, NULL);
 initexcl(&LogReqExcl, 0, NULL);

 strncpy( LogFilePath, GetLogDirectory(), MAX_FILENAME_LEN );
 LogFilePath[MAX_FILENAME_LEN] = '\0';

 if (strlen(LogFilePath) == 0)
	strcpymax(sizeof(LogFilePath), LogFilePath, DefLogFilePath);

 sp77sprintf( Ipar, sizeof(Ipar), "%s%c%s", LogFilePath, DIRPATH_SEP, DEF_LOGFILE);

 sqlfopenc (Ipar, SP5VF_BINARY, SP5VF_APPEND, SP5BK_UNBUFFERED, &LogFd, &err);
 if (err.sp5fe_result != vf_ok)
	LogFd = 0;

/* sprintf(tmps, "Starting Log (build %s)", s100buildnumber());
 Log(tmps, -1);*/

 if (strlen(LogFilePath) == 0)
	strcpymax(sizeof(LogFilePath), LogFilePath, DefLogFilePath);

 sp77sprintf( Ipar, sizeof(Ipar), "%s%c%s", LogFilePath, DIRPATH_SEP, DEF_ERRFILE);

 sqlfopenc (Ipar, SP5VF_BINARY, SP5VF_APPEND, SP5BK_UNBUFFERED, &ErrFd, &err);
 if (err.sp5fe_result != vf_ok)
	ErrFd = 0;

 sprintf(tmps, "Starting ErrLog (build %s)", s100buildnumber());
 Error(tmps, -1);


 printf("Using (logfile:%s, tracelevel:%d)\n", Ipar, TraceLevel);

 LogReq = GetLogRequests();

 LogRes = FALSE;

 return(TRUE);
}

void Log
(
 char	text[],
 int	lg
)
{
 char			tdbuffer[100];
 char			wbuffer[100];
 time_t			t;
 size_t			tmplg = 0;
 size_t			ulg = 0;
 tsp00_Bool		bOk;
 tsp5_rte_fileerror	ferror;

 if (!LogFd)
	return;

 if (lg < 0)
	ulg = strlen(text);
 else
	ulg = lg;

 begexcl(LogExcl);

 time(&t);
 strftime(tdbuffer, sizeof(tdbuffer), "[%d/%b/%Y:%H:%M:%S]",
		 localtime(&t));
 if (LogFd)
 {
	sp77sprintf( wbuffer, sizeof(wbuffer), LOG_FORMAT_LEAD, tdbuffer, "");
	sqlfwritec (LogFd, wbuffer, strlen(wbuffer), &ferror);
 }

 if (LogFd)
	sqlfwritec (LogFd, text, ulg, &ferror);
 
 if (LogFd)
 {
	sqlfwritec (LogFd, LOG_FORMAT_TRAIL, sizeof(LOG_FORMAT_TRAIL)-1, &ferror);
#ifdef WIN32
	sqlfwritec (LogFd, "\r\n", sizeof("\r\n")-1, &ferror);
#else
	sqlfwritec (LogFd, "\n", sizeof("\n")-1, &ferror);
#endif
 }

 endexcl(LogExcl);

}

void Error
(
 char	text[],
 int	lg
)
{
 char			tdbuffer[100];
 char			wbuffer[100];
 time_t			t;
 size_t			tmplg = 0;
 size_t			ulg = 0;
 tsp00_Bool		bOk;
 tsp5_rte_fileerror	ferror;

 if (!ErrFd)
	return;

 if (lg < 0)
	ulg = strlen(text);
 else
	ulg = lg;

 begexcl(ErrExcl);

 time(&t);
 strftime(tdbuffer, sizeof(tdbuffer), "[%d/%b/%Y:%H:%M:%S]",
		 localtime(&t));
 if (ErrFd)
 {
	sp77sprintf( wbuffer, sizeof(wbuffer), LOG_FORMAT_LEAD, tdbuffer, "");
	sqlfwritec (ErrFd, wbuffer, strlen(wbuffer), &ferror);
 }

 if (ErrFd)
	sqlfwritec (ErrFd, text, ulg, &ferror);

 if (ErrFd)
 {
	sqlfwritec (ErrFd, LOG_FORMAT_TRAIL, sizeof(LOG_FORMAT_TRAIL)-1, &ferror);
#ifdef WIN32
	sqlfwritec (ErrFd, "\r\n", sizeof("\r\n")-1, &ferror);
#else
	sqlfwritec (ErrFd, "\n", sizeof("\n")-1, &ferror);
#endif
 }

 endexcl(ErrExcl);
}

void SetTraceLevel
(
 int	Level
)
{
 TraceLevel = Level;
}

void Trace
(
 int	Level,
 char	*Txt,
 size_t	lg
)
{
 if (Level > TraceLevel)
	return;

 Error(Txt, lg);
}

void ExitLog
(
 void
)
{
 tsp5_rte_fileerror	ferror;

 Log("Terminating Logging", -1);
 begexcl(LogExcl);

 exitexcl(LogReqExcl);
 if (LogFd)
	sqlfclosec (LogFd, SP5VF_CLOSE_NORMAL, &ferror);
 LogFd = NULL;
 if (ErrFd)
	sqlfclosec (ErrFd, SP5VF_CLOSE_NORMAL, &ferror);
 ErrFd = NULL;
 endexcl(LogExcl);
 endexcl(ErrExcl);

 exitexcl(LogExcl);
}

