#ifdef WIN32
#include <windows.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <sapdbwa_define.h>

#include "SAPDB/WebAgent/HTTP/WAHTTP_Misc.h"
#include "SAPDB/WebAgent/HTTP/WAHTTP_Control.h"

void PrintHelp();


/****************************************************************************
 ***
 *** main
 ***
 ****************************************************************************/

main(int argc, char *argv[])
{

	char  clo = '\0';
	char  execDir[MAX_FILENAME_LEN+1] = "";
	char *optarg = NULL;

	/* Get current directory	*/
	if ( !GetExecDirectory( argv[0], execDir, MAX_FILENAME_LEN ) ) {
		exit( EXIT_FAILURE );
	}
	execDir[MAX_FILENAME_LEN] = '\0';

	/* Check command line arguments	*/
    while ( (clo = GetOpt( argc, argv, "f:iudh", &optarg )) != -1) {
		switch ( clo ) {

#ifdef WIN32
			/* Install as service	*/
			case 'i':
				if ( !InstallAsService( execDir ) ) {
					return EXIT_FAILURE;
				} else {
					return EXIT_SUCCESS;
				}

				break;

			/* Remove as service	*/
			case 'u':
				if ( !RemoveAsService() ) {
					return EXIT_FAILURE;
				} else {
					return EXIT_SUCCESS;
				}

				break;

			/* Start to debug	*/
			case 'd':
				Start( execDir, sapdbwa_True, NULL );
				return EXIT_SUCCESS;

				break;
#else
			case 'i':
			case 'u':
				printf( "Not supported on this system!");
				break;
#endif

			/* Specify an alternate registry path*/
			case 'f':
				if ( optarg ) {
					Start( execDir, sapdbwa_True, optarg );
				} else {
					PrintHelp();
				}

				return EXIT_SUCCESS;

				break;

			/* Help	*/
			case 'h':
			default:
				PrintHelp();

				return EXIT_SUCCESS;

				break;
		}

		optarg = NULL;
	}

#ifdef WIN32
	SetDispatchTable();
#else
	Start( execDir, sapdbwa_False, NULL );
#endif

	return EXIT_SUCCESS;

}

void PrintHelp()
{

#ifdef WIN32
	printf("Usage: wahttp [-i] [-u] [-d] [-f file] [-h]\n\n");

	printf("Options:\n");
	printf("  -i      : install an SAP DB WWW service\n");
	printf("  -u      : uninstall an SAP DB WWW service\n");
	printf("  -d      : to start SAP DB WWW from command line\n");
	printf("  -f file : specify an alternate registry path\n");
	printf("  -h      : this help\n\n");

	printf("Specify registry path in wahttp.conf\n");
	printf("or use option -f\n\n");

	printf("To start service : net start sapdbwww\n");
	printf("To stop service  : net stop sapdbwww\n");
#else
	printf("Usage: wahttp [-f file] [-h]\n\n");

	printf("Options:\n");
	printf("  -f file : specify an alternate ini file\n");
	printf("  -h      : this help\n\n");

	printf("To start service specify ini file in wahttp.conf\n");
	printf("or use option -f\n");
#endif

}
