/*!**********************************************************************

  module: WAHTTP_Server.c

  -----------------------------------------------------------------------

  responsible:  Markus Oezgen

  special area: SAP DB WWW

  description:  Sockets, Request delivering

  see also:

  -----------------------------------------------------------------------

                         Copyright by SAP AG, 2001


    ========== licence begin LGPL
    Copyright (C) 2001 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end


************************************************************************/

#define MODULE_NAME "WAHTTP_Server.c"

/****************************************************************************
 ***
 *** Include
 ***
 ****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>

#ifdef WIN32
#include <winsock2.h>
#else
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <signal.h>
#endif

#ifdef WIN32
#define ThreadSleep	Sleep
#define closesocket(a)	closesocket(a)
#else
/*#define ThreadSleep(msecs)	sleep(msecs/1000)*/
#define ThreadSleep(msecs)	{ signal(SIGALRM, OurSleeper); alarm(msecs/1000); }
#define closesocket(a)		close(a)
#endif


static void OurSleeper
(
 int	sig
)
{
}

#include <sapdbwa_define.h>

#include "SAPDB/WebAgent/HTTP/WAHTTP_Misc.h"
#include "SAPDB/WebAgent/HTTP/WAHTTP_Log.h"
#include "SAPDB/WebAgent/HTTP/WAHTTP_IniPar.h"
#include "SAPDB/WebAgent/HTTP/WAHTTP_Connect.h"
#include "SAPDB/WebAgent/HTTP/WAHTTP_HTTP.h"
#include "SAPDB/WebAgent/HTTP/WAHTTP_Agent.h"
#include "SAPDB/WebAgent/HTTP/WAHTTP_Server.h"
#include "SAPDB/WebAgent/HTTP/WAHTTP_Workers.h"
#include "SAPDB/WebAgent/HTTP/WAHTTP_Memory.h"

/*#define PERSISTENT_CONNECTIONS*/

#define NO_OF_WORKERS	10


#define MAXHOSTNAMESZ	40
#define MAXREQS			1
#define MAXHEADERSIZE	2000
#define HTTP_PROT		"HTTP/1.0"
#define	DEF_HTTP_PORT	80

/*#define chunk 1000*/
/*#define BADPROXYREQ "Content-Type: text/html\r\n<body><h1>Bad Proxy Request</h1></body>"*/
/*#define PROXYREQERR "Content-Type: text/html\r\n<body><h1>Error during Proxy Request</h1></body>"*/
/*#define S501 "HTTP/1.0 501 Not Implemented\r\n\r\n"*/

#define DEFAULT_CLIENT_KEEPALIVE FALSE
#define DEFAULT_SERVER_KEEPALIVE FALSE

#define INI_MULTI_THREADED	"MultiThreading"
#define INI_TRCLEVEL		"TraceLevel"
#define INI_LISTEN_IP		"ListenIp"
#define INI_LISTEN_PORT		"ListenPort"
#define INI_ALIAS		"Alias"

#define DEF_TRCLEVEL		0
#define DEF_MULTI_THREADED	FALSE
#define DEF_LISTEN_PORT		80

#define TRC_HTTP_HEAD	8
#define TRC_HTTP_BODY	9

#define INI_FNAME		"SapDbPort80.ini"

static teo07_Thread thisThread;
static sapdbwa_Bool	threading = sapdbwa_True;
static int JUSTLOG = FALSE;

static int Delegate2Proxy = FALSE;
static char Delegate2ProxyHost[100] = "";
static unsigned short Delegate2ProxyPort = 0;
static DynStr	ReqCookieFileStr = NULL,
		ResCookieFileStr = NULL;


static SOCKET		ServerSocket;
static unsigned short	LocalPort = DEF_LISTEN_PORT;
typedef char	*charp;
#ifdef WITH_LISTEN_IP
static charp		*LocalIps = NULL;
static size_t		NoOfLocalIps = 0;
static char		LocalIp[100];
#endif

typedef char t_header[MAXHEADERSIZE];

static sapdbwa_Bool		stopWebServer = sapdbwa_False;
static sapdbwa_UInt4	workerThreadCount = 0;

/****************************************************************************
 ***
 *** 
 ***
 ****************************************************************************/

int recv_line
(
 SOCKET	cs,
 DynStr	*RecLine
)
/* Receive a HTTP header line including the "\r\n".
 * Return it terminated by '\0'
 * Allocate the buffer dynamically.
 * Add all data to the buffer.
 */
{
 char		c1 = '\0', c2 = '\0';
 int		lg;
 int		selectRtc;
 fd_set		readfds;
 struct timeval timeout;
 int		fromlg = 0;

 timeout.tv_sec = CONNECTIOM_TIMEOUT;
 timeout.tv_usec = 0;

 FD_ZERO(&readfds);
 FD_SET(cs, &readfds);
 /* recv one byte at a time */
 do
 {
	if (!DSextent(RecLine, 2))
	{
		FD_CLR(cs, &readfds);
		return(FALSE);
	}

	(*RecLine)->str[(*RecLine)->lg] = '\0';

	/* Wait for date or time-out */
/*
	selectRtc = select(1, &readfds, 0, 0, &timeout);

	switch (selectRtc)
	{
		case 0:	* Time-out *
			FD_CLR(cs, &readfds);
			SyscallError("Receive data from remote client timed out", -1);
			return(FALSE);
		case SOCKET_ERROR:
			FD_CLR(cs, &readfds);
			SyscallError("Waiting for data from remote client failed", -1);
			return(FALSE);
	}
*/
	c1 = c2;
	lg = recv( cs, &c2, 1, 0 );
	if (lg == SOCKET_ERROR) 
	{
		FD_CLR(cs, &readfds);
		SyscallError("Receive data from remote client failed", -1);
		return(FALSE);
	}
	else if (lg == 0)
	{
		FD_CLR(cs, &readfds);
		return(FALSE);
	}

	(*RecLine)->str[(*RecLine)->lg] = c2;
	(*RecLine)->lg++;

	if ((c1 == '\r' && c2 == '\n') /*||
	    (c1 == '\n' && c2 == '\n')*/)
	{
		FD_CLR(cs, &readfds);
		(*RecLine)->str[(*RecLine)->lg] = '\0';
		return(TRUE);
	}
 } while (1);
 
    /* PTS (MaO): 1109893 */

}

char *IllHeader[] =
{ "Proxy-Connection:",
  NULL
};

int ForbidHeader
(
 char	*Header,
 size_t	HeaderLg
)
{
 int	i = 0;

 while (IllHeader[i])
	if (HeaderLg == strlen(IllHeader[i]) &&
	    !strncmp(Header, IllHeader[i], HeaderLg))
		return(TRUE);
	else
		i++;
 return(FALSE);
}

static SOCKET AcceptingSocket = 0;

int AcceptConnection
(
 SOCKET			ps,
 int			KeepAlive,
 struct st_tcp_connection	*C
)
{
#if  WIN32 || OSF1 || HPUX || (SUN && !BIT64)
int            clen;
#else
socklen_t      clen;
#endif

int            rtc = TRUE;
FILE          *f = stdout;
struct linger  Linger;
int            just_say_no = 1;
int            keepalive_value = 1;  
int            one = 1;
   

 clen=sizeof(C->sa);
 AcceptingSocket = ps;
#ifdef WIN32
 if ((C->s = accept(ps, (struct sockaddr FAR  *)&C->sa, &clen)) == INVALID_SOCKET) 
#else
 if ((C->s = accept(ps, (struct sockaddr *)&C->sa, &clen)) == INVALID_SOCKET) 
#endif
 {
	AcceptingSocket = 0;
	Error("Init: accept failed", -1);
	return(FALSE);
 }
 AcceptingSocket = 0;

 clen = sizeof(C->sa);
#ifdef WIN32
 if(getsockname(C->s, (struct sockaddr FAR  *)&C->sa, &clen) < 0)
#else
 if(getsockname(C->s, (struct sockaddr *)&C->sa, &clen) < 0)
#endif
 {
	SyscallError("could not get  client socket name", -1);
	Error("Init: getsockname failed", -1);
	shutdown(C->s, SD_BOTH);
	closesocket(C->s);
	return(FALSE);
 }

 if((setsockopt(C->s, SOL_SOCKET, SO_REUSEADDR,(char *)&one, sizeof(one)))
       == -1) {
	SyscallError("could not set client socket option (SO_REUSEADDR)", -1);
	shutdown(C->s, SD_BOTH);
	closesocket(C->s);
	return(FALSE);
 }

 if((setsockopt(C->s,SOL_SOCKET, SO_KEEPALIVE, (void *)&keepalive_value,
        sizeof(keepalive_value))) == -1) {
	SyscallError("could not set client socket option (SO_KEEPALIVE)", -1); 
	shutdown(C->s, SD_BOTH);
	closesocket(C->s);
	return(FALSE);
 }

 if((setsockopt(C->s,IPPROTO_TCP, TCP_NODELAY,(char *)&just_say_no,
        sizeof(just_say_no))) == -1) {
	SyscallError("could not set client socket option (TCP_NODELAY)", -1); 
	shutdown(C->s, SD_BOTH);
	closesocket(C->s);
	return(FALSE);
 }

 Linger.l_onoff = !0; 
 Linger.l_linger = 30;
#ifdef WIN32
    if (setsockopt (C->s, SOL_SOCKET, SO_LINGER, (const char FAR *)&Linger, sizeof(Linger)))
#else
    #if (SUN && !BIT64)
        if (setsockopt (C->s, SOL_SOCKET, SO_LINGER, (const char *)&Linger, sizeof(Linger)))
    #else
        if (setsockopt (C->s, SOL_SOCKET, SO_LINGER, &Linger, sizeof(Linger)))
    #endif
#endif
 {
	SyscallError("could not set client socket option (SO_LINGER)", -1); 
	shutdown(C->s, SD_BOTH);
	closesocket(C->s);
	return(FALSE);
 }

 C->Connected = TRUE;
 C->KeepAlive = KeepAlive;
 return(TRUE);
}

void CloseConnection
(
 struct st_tcp_connection	*C
)
{
 fd_set readfds;
 struct timeval timeout;
 char	b[512];
 int	err;

 if (!C->Connected)
	  return;

 /* Shutdown the connection for sending data */
 shutdown(C->s, SD_SEND);

 /* Wait for data or an error on socket */
 /* discard rest in socket, then close */
 FD_ZERO(&readfds);
 FD_SET(C->s, &readfds);
 timeout.tv_sec = 2;	/* not too long in case remote browser forgets to close and we get stuck */
 timeout.tv_sec = 0;

 while (0 < select (C->s + 1, &readfds, NULL, NULL, &timeout))
 {
	err = recv( C->s, b, sizeof(b), 0 );
	if (err == 0 || err == SOCKET_ERROR)
		break;
 }

 /* Now really close */
 closesocket(C->s);
 C->Connected = FALSE;
 C->KeepAlive = FALSE;
}

static size_t hdllen
(
 char	*line
)
{
 char	*tmp = strstr(line, "\r\n");

 if (tmp)
	return(tmp - line);
 else
	return(0);
}

static int ReceiveHttpHeaders
(
 SOCKET	cs,
 DynStr	*HeadersLine,
 size_t	*NoOfHeaderLines
)
{
 size_t	LastHeaderLineAt = 0;
 size_t	LastHeaderLineLg = 0;

 while (recv_line(cs, HeadersLine))
 {
	LastHeaderLineLg = hdllen((*HeadersLine)->str+LastHeaderLineAt);
	if (0 == LastHeaderLineLg)
		return(TRUE);

	(*HeadersLine)->str[LastHeaderLineAt + LastHeaderLineLg] = '\0';
	(*NoOfHeaderLines)++;
	LastHeaderLineAt += LastHeaderLineLg + 2;
 }
 return(FALSE);
}

static int PrepareHeaders
(
 struct st_http_req	*HttpReq
)
{
 char			*p = NULL;
 char			*hl = HttpReq->HeadersLine->str;
 size_t			hll = HttpReq->HeadersLine->lg;
 size_t			HeaderNo = 0, pl;
 char			*tmp;
 tsp00_Bool		bOk;

 wahttp_SQLALLOCAT( HttpReq->NoOfHeaderLines * sizeof(char*),
		  &HttpReq->HeaderNamesArray,
		  &bOk);
 if (!bOk)
	HttpReq->HeaderNamesArray = NULL;

 wahttp_SQLALLOCAT( HttpReq->NoOfHeaderLines * sizeof(char*),
		  &HttpReq->HeaderValuesArray,
		  &bOk);
 if (!bOk)
	HttpReq->HeaderValuesArray = NULL;

 if (!HttpReq->HeaderNamesArray || !HttpReq->HeaderValuesArray)
 {
	if (HttpReq->HeaderNamesArray) wahttp_SQLFREE(HttpReq->HeaderNamesArray);
	HttpReq->HeaderNamesArray = NULL;
	if (HttpReq->HeaderValuesArray) wahttp_SQLFREE(HttpReq->HeaderValuesArray);
	HttpReq->HeaderValuesArray = NULL;
	return(FALSE);
 }

 for (p=hl;
      p<hl+hll && HeaderNo < HttpReq->NoOfHeaderLines;
      p+=pl + HTTP_HEADER_SEP_LG, HeaderNo++)
 {
	pl = strlen(p);
	HttpReq->HeaderNamesArray[HeaderNo] = p;
	if (HeaderNo == 0)
		HttpReq->HeaderValuesArray[HeaderNo] = NULL;
	else
	{
		tmp = strchr(p, ' ');
		if (!tmp)
		{
			if (HttpReq->HeaderNamesArray) wahttp_SQLFREE(HttpReq->HeaderNamesArray);
			HttpReq->HeaderNamesArray = NULL;
			if (HttpReq->HeaderValuesArray) wahttp_SQLFREE(HttpReq->HeaderValuesArray);
			HttpReq->HeaderValuesArray = NULL;
			return(FALSE);
		}
		*tmp = '\0';
		tmp++;
		while (*tmp == ' ') tmp++;
		HttpReq->HeaderValuesArray[HeaderNo] = tmp;

		/* remove the trailing colon */
		tmp = strchr(p, ':');
		if (*tmp == ':')
			*tmp = '\0';
	}
 }
 return(TRUE);
}

struct st_our_socket
{
 struct st_tcp_connection	*tcp_con;
 struct st_http_req		*req;
};

static size_t our_recv
(
 struct st_data *This,
 void		*Buf,
 size_t		Lg
)
{
 struct st_our_socket	*os = This->H;
 size_t			rlg;

 if (!os->req->ContentLength)
	return(0);

 rlg = WAHTTP_MIN(os->req->ContentLength, Lg);
 rlg = recv(os->tcp_con->s, Buf, rlg, 0 );
 if (rlg == SOCKET_ERROR) 
 {
	SyscallError("recv from server failed", rlg);
	return(0);
 }
 return(rlg);
}

static int BadRequest
(
 struct st_http_client	*client
)
{
 time_t			now;

 AddHeader_char(client, HTTP_SERVER, THIS_SERVER);
 time( &now );
#ifdef WIN32
 AddHeader_date(client, HTTP_DATE, now + _timezone);
#else
 AddHeader_date(client, HTTP_DATE, now + timezone);
#endif
/* AddHeader_char(&rep, CONTENT_TYPE, DEFAULTMIMETYPE);*/
 SendHttpHeader(HTTP_VERSION " " HTTP_STATUS_BAD_REQUEST, client);

 return(TRUE);
}

static int ProcessRequest
(
 struct st_http_client	*client,
 struct st_http_req	*HttpReq
)
{
 struct st_our_socket	os;
 char			*Status;
 int			Persistent = FALSE;

 Log(HttpReq->HeadersLine->str, -1);

 HttpReq->method = HttpReq->HeadersLine->str;
 if (!HttpReq->method)
	return(FALSE);

 HttpReq->uri = strchr(HttpReq->method, ' ');
 if (!HttpReq->uri)
	return(FALSE);
 else
 {
	*HttpReq->uri = '\0';		/* this is the end of method */
	HttpReq->uri++;
	while (*HttpReq->uri &&
	       *HttpReq->uri == ' ')	/* skip any more SP */
		HttpReq->uri++;
	if (!HttpReq->uri)
		return(FALSE);
	HttpReq->Type = hmt_simple_request;
 }

 HttpReq->hversion = strchr(HttpReq->uri, ' ');
 if (HttpReq->hversion)
 {
	*HttpReq->hversion = '\0';
	HttpReq->hversion++;
	while (*HttpReq->hversion &&
	       *HttpReq->hversion == ' ')	/* skip any more SP */
		HttpReq->hversion++;
	if (!HttpReq->hversion)
		return(FALSE);
	HttpReq->Type = hmt_full_request;
 }

 if (!strcmp(HttpReq->method, "GET"))
	HttpReq->Method = rq_get;
 else if (!strcmp(HttpReq->method, "HEAD"))
	HttpReq->Method = rq_head;
 else if (!strcmp(HttpReq->method, "POST"))
	HttpReq->Method = rq_post;
 else
	return(FALSE);
	
 if (HttpReq->Method == rq_post)
	HttpReq->ContentLengthValid = GetContentLength(HttpReq, &HttpReq->ContentLength);

 if (HttpReq->Type == hmt_simple_request)
 {
	HttpReq->HMajorVersion = 0;
	HttpReq->HMinorVersion = 0;
 }
 else
	if (!HttpGetVersion(HttpReq->hversion, &HttpReq->HMajorVersion, &HttpReq->HMinorVersion))
		return(FALSE);

 HttpReq->params = strchr(HttpReq->uri, ';');
 if (HttpReq->params)
 {
	*HttpReq->params = '\0';
	HttpReq->params++;
 }

 if (HttpReq->params)
	HttpReq->querystring = strchr(HttpReq->params, '?');
 else
	HttpReq->querystring = strchr(HttpReq->uri, '?');
 if (HttpReq->querystring)
 {
	*HttpReq->querystring = '\0';
	HttpReq->querystring++;
 }

 GetHostPort(HttpReq, &HttpReq->Host, &HttpReq->Port);
#ifdef PERSISTENT_CONNECTIONS
 HttpReq->PersistentRequested = HttpIsPersistent(HttpReq);
#else
 HttpReq->PersistentRequested = FALSE;
#endif

/* check consistency */
 if (!HttpIsReqConsistent(HttpReq, &Status))
 {
	BadRequest(client);
	return(TRUE);
 }

 if (HttpReq->PersistentRequested)
 {
	client->tcp_con->KeepAlive = TRUE;
	AddHeader_char(client, HTTP_CONNECTION, HTTP_KEEP_ALIVE);
 }

 os.tcp_con = client->tcp_con;
 os.req = HttpReq;
 HttpReq->Body.H = &os;
 HttpReq->Body.read = our_recv;
 HttpReq->Body.destroy = NULL;
 HttpReq->Body.rewind = NULL;

 if (!TryAgents(client, HttpReq, &Persistent))
 {
	Error("TryAgents failed", 0);
	return(FALSE);
 }
 return(TRUE);
}


static void DoHttp
(
 struct st_tcp_connection	*c,
 int				*KeepAlive
)
{
 char			*reqbuf = NULL;
 size_t			reqbufsize = 0, reqbuflg = 0;
 DynStr			ReqLine = NULL;
 int			recv_ok = TRUE;
 int			hno = 0;
 int			SimpleRequest = TRUE;
 size_t			cnt = 0;
 long			ContentLength = -1;
 int			ContentLengthDone = FALSE;
 char			Uri[MAXURISIZE] = "";
 DynStr			Body = NULL;
 size_t			rBodyLg = 0;
 size_t			ContentLg = 0;
 char			*cl = NULL;
 struct st_http_req	HttpReq;
 struct st_http_client	client;
 int			stop = FALSE;

	memset(&HttpReq, 0, sizeof(HttpReq));
	HttpReq.HeadersLine = DSinit();
	HttpReq.NoOfHeaderLines = 0;
	memset(&client, 0, sizeof(client));
	client.HeadersLine = DSinit();
	client.tcp_con = c;

	if (!ReceiveHttpHeaders(c->s, &HttpReq.HeadersLine, &HttpReq.NoOfHeaderLines))
	{
		Error("ReceiveHttpHeader (Request) failed", 0);
		stop = TRUE;
	}
	else if (!PrepareHeaders(&HttpReq))
	{
		Error("PrepareHeaders (Request) failed", 0);
		stop = TRUE;
	}
	else if (!ProcessRequest(&client, &HttpReq))
	{
		Error("ProcessRequest (Request) failed", 0);
		BadRequest(&client);
		stop = TRUE;
	}

	if (HttpReq.HeaderNamesArray) wahttp_SQLFREE(HttpReq.HeaderNamesArray);
	HttpReq.HeaderNamesArray = NULL;
	if (HttpReq.HeaderValuesArray) wahttp_SQLFREE(HttpReq.HeaderValuesArray);
	HttpReq.HeaderValuesArray = NULL;
	DSdelete(client.HeadersLine);
	DSdelete(HttpReq.HeadersLine);

 if (stop)
	*KeepAlive = FALSE;
 else
	*KeepAlive = c->KeepAlive && c->Connected;

 return ;
}

static  struct st_workq	*HttpReqQ;

static void *HttpWorker
(
  void	*something
)
{
 int			stop;
 struct st_work		*work;
 int			WorkNo = 0;
 int			KeepAlive = FALSE;

 do
 {
	work = GetWork(HttpReqQ);
	if (!work)
	{
		Error("Got NULL work", -1);
		break;
	}
	else
		if (work->stop)
		{
			Error("This HttpWorker got stop request", -1);
			DestroyWork(work);
			break;
		}

	WorkNo++;

	DoHttp(work->ActualWork, &KeepAlive);
	stop = work->stop;

	if (KeepAlive)
		PutWork(work, HttpReqQ);
	else
		DestroyWork(work);

 } while (!stop);

 Error("This HttpWorker is terminating", -1);
 sqlendthread(0);
 return (void*)0;
}

void *MakeTcpConnection
(
 void
)
{
 struct st_tcp_connection	*c;
 tsp00_Bool			bOk;

 wahttp_SQLALLOCAT( sizeof(struct st_tcp_connection),
		    &c,
		    &bOk);
 if (!bOk)
 {
	Error("MakeTcpConnection: could not wahttp_SQLALLOCAT", -1);
	return(NULL);
 }
 memset(c, 0, sizeof(struct st_tcp_connection));
 return(c);
}

void DestroyTcpConnection
(
 void *c
)
{
 CloseConnection(c);
 wahttp_SQLFREE(c);
}

static int Service_http
(
 SOCKET		ss
)
{
 int				rtc = TRUE;
 FILE				*f = stdout;
 struct	st_work			*work;
 struct st_tcp_connection	*c;

 work = MakeWork(MakeTcpConnection, DestroyTcpConnection);
 if (!work)
 {
	Error("Service_http: could not MkWork", -1);
	return(FALSE);
 }

 c = work->ActualWork;
 c->KeepAlive = DEFAULT_CLIENT_KEEPALIVE;
 c->Connected = FALSE;

 if (!AcceptConnection(ss, DEFAULT_CLIENT_KEEPALIVE, c))
 {
	DestroyWork(work);
	return(FALSE);
 }

 PutWork(work, HttpReqQ);
 if (threading)
 {
 }
 else
 {
	HttpWorker(0);
 }

 return(TRUE);
}


static ServiceLoop = TRUE;

static void *Service_https
(
 void *ss
)
{
 if (!StartWorkers())
	return (void *)0;

 Error("starting service loop", -1);
 ServiceLoop = TRUE;
 while (ServiceLoop)
	 Service_http(*(SOCKET*)ss);
 Error("terminating service loop", -1);
 return (void *)0;
}

/* Initialize for request processing:
 * - FileSystemRetrieval
 * - WebAgent
 */
static int InitRequestProcessing()
{

    tsp00_ErrTextc		errtext;
    teo07_ThreadErr	err;
    teo07_Thread		this_thread;

    if (!InitAgents()) {
        Error("Initializing Processing Agents failed", -1);
        return(FALSE);
    }

    return (TRUE);

}

static int ExitRequestProcessing
(
)
{
 if (!ExitAgents())
 {
	Error("Exitalizing Processing Agents failed", -1);
	return(FALSE);
 }

 return(TRUE);
}

enum en_cmd
{
 cmd_terminate,
 cmd_continue,
 cmd_err,
 cmd_huh
};

enum en_watchwhat
{
 ww_nothing,
 ww_console,
 ww_flag
} WatchWhat;

static int FlagStop = TRUE;

static enum en_cmd GetConsoleCommand
(
 void
)
{
 char	line[1000];

 switch (WatchWhat)
 {
  case ww_console:
	do
	{
		printf("Enter command (q for quit, no other commands available) :");
		if (!fgets(line, sizeof(line), stdin))
			return(cmd_err);
		if (!strcasecmp(line, "q\n"))
			return(cmd_terminate);
		else
		if (!strcasecmp(line, "c\n"))
			return(cmd_continue);
		else
			printf("huh ?\n");
	}
	while (1);
	break;
  case ww_flag:
	if (FlagStop)
		return(cmd_terminate);
	else
		return(cmd_continue);
	break;
  case ww_nothing:
  default:
		return(cmd_terminate);
 }
}

int InitWebServer( char	        IniFilePath[],
                   char        *alternateIniFile,
                   tsp00_Bool   debug)
{
#ifdef WIN32
 WSADATA	WsaData;
#endif
 int		err; 
 char		errtxt[MAX_ERRORTEXT_LG];
 char		errtxt2[MAX_ERRORTEXT_LG];
 char		Ipar[100];
 int		trcl = DEF_TRCLEVEL;
/*
 char		inifile[1000];
*/
 char		*sepat = 0;
 unsigned long	tid = 0;
 int		InitIniParFailed = FALSE;
 char		*tmp1;
 charp		*tmp2;
 char		errorText[MAX_ERRORTEXT_LG+1];
 tsp00_Bool bOk;
 tsp00_ErrTextc		threadErrText;
 teo07_ThreadErr	threadErr;

    /* Init threading environment           */
    /* Also important for usage of mutex    */
    /* PTS: 1110225 (MaO)                   */
    sqlinit_multithreading(&thisThread, threadErrText, &threadErr);
    if (threadErr != THR_OK_EO07) {
#ifdef WIN32
        if (debug) {
            printf(threadErrText);
            printf("\n");
        }
#else
        printf(threadErrText);
        printf("\n");
#endif

        return(FALSE);
    }

 wahttp_SQLMALLOCINIT(&bOk);

#ifndef WIN32
 signal(SIGPIPE, SIG_IGN);
#endif

 SetDefLogFilePath(IniFilePath);

/*
 if (strlen(IniFilePath) > 0)
	sprintf(inifile, "%s%s%s", IniFilePath, DIRPATH_SEP, INI_FNAME);
 else
	sprintf(inifile, "%s", INI_FNAME);
*/
 if ( !InitIniPar( IniFilePath, alternateIniFile, errorText, MAX_ERRORTEXT_LG ) )
	InitIniParFailed = TRUE;

 InitLog();

 if (InitIniParFailed)
 {
	sp77sprintf( errtxt2, MAX_ERRORTEXT_LG, "Init web server failed:%s", errorText );
	Error( errtxt2, -1);

	sp77sprintf( errtxt2, MAX_ERRORTEXT_LG, "Using default settings to init web server." );
	Error( errtxt2, -1);
 }

	trcl = DEF_TRCLEVEL;
/*
 if (!IniReadPar(INI_TRCLEVEL, sizeof(Ipar), Ipar, 0, NULL))
	trcl = DEF_TRCLEVEL;
 else
 {
	trcl = atoi(Ipar);
	if (trcl == 0)
		trcl = DEF_TRCLEVEL;
 }
*/

 SetTraceLevel(trcl);

	/* Get mime types	*/
	if ( !ReadMimeTypes( errorText, MAX_ERRORTEXT_LG ) ) {
		sp77sprintf( errtxt2, MAX_ERRORTEXT_LG, "Read mime types failed:%s", errorText );
		Error( errtxt2, -1);
	}

 threading = GetThreading();
/*
 if (!IniReadPar(INI_MULTI_THREADED, sizeof(Ipar), Ipar, 0, NULL))
	MULTI_THREADED = DEF_MULTI_THREADED;
 else
	MULTI_THREADED = !strcasecmp(Ipar, "yes");
*/

 LocalPort = GetPort();
/*
 if (!IniReadPar(INI_LISTEN_PORT, sizeof(Ipar), Ipar, 0, NULL))
	LocalPort = DEF_LISTEN_PORT;
 else
	LocalPort = atoi(Ipar);
*/

#ifdef WITH_LISTEN_IP
 *LocalIp = '\0';
/*
 if (!IniReadPar(INI_LISTEN_IP, sizeof(LocalIp), LocalIp, 0, NULL))
	*LocalIp = '\0';
*/

 NoOfLocalIps = 0;
 LocalIps = NULL;

 tmp1 = LocalIp;
 while (tmp1 && *tmp1)
 {
	tmp2 = LocalIps;
	wahttp_SQLREALLOCAT( ( NoOfLocalIps + 1 ) * sizeof(LocalIps),
		    &tmp2,
		    &bOk);
	if (!bOk)
	{
		if (LocalIps) wahttp_SQLFREE(LocalIps);
		Error("InitWebServer: could not wahttp_SQLREALLOCAT LocalIps", -1);
		return(FALSE);
	}
	LocalIps = tmp2;
	LocalIps[NoOfLocalIps] = tmp1;
	if (tmp1 = strchr(tmp1, ';'))
	{
		*tmp1 = '\0';
		tmp1++;
	}
	NoOfLocalIps++;
 }
 sprintf(errtxt, "Listening on IP %s on port %d", LocalIp, LocalPort);
#else
 sprintf(errtxt, "Listening on port %d", LocalPort);
#endif
 
 Error(errtxt, -1);

#ifdef WIN32
 err = WSAStartup (0x0101, &WsaData);
 if (err == SOCKET_ERROR)
 {
	SyscallError("Init: WSAStartup Failed", -1);
	ExitLog();
	return FALSE;
 }
#endif

 if (!InitWorkers(threading))
 {
	Error("Init: Failed", -1);
	ExitLog();
	return FALSE;
 }

 InitRequestProcessing();

 Error("Init: Initialized", -1);

 return(TRUE);
}

static teo07_Thread listenerThread = NULL;

int StartWebServer
(
)
{
 SOCKADDR_IN		sa_server;
 IN_ADDR		bind_address;
 int			b1, b2, b3, b4;
 size_t			i;
 tsp00_ErrTextc		errtext;
 teo07_ThreadErr	err;
#ifndef WIN32
 struct protoent	*pe;
#endif
 int				just_say_no = 1;
 int				keepalive_value = 1;  
 int				one = 1;
	sapdbwa_UInt4	maxThreads = 0;
	
	maxThreads = GetMaxThreads();
	if ( maxThreads < 1 ) {
		maxThreads = NO_OF_WORKERS;
	}

 if (listenerThread)
 {
	Error("Already running, startup request ignored", -1);
	return(EXIT_SUCCESS);
 }

 if (!MakeWorkers( maxThreads, HttpWorker ))
	return(EXIT_FAILURE);


#ifdef WIN32
 if ((ServerSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) {
#else
 if ((ServerSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) {
#endif
        SyscallError("Init: could not get listen socket", -1);
	ExitLog();
	return EXIT_FAILURE;
 }

 if((setsockopt(ServerSocket,SOL_SOCKET,SO_REUSEADDR,(char *)&one,sizeof(one)))
       == -1) {
	SyscallError("Init: could not set listen socket option (SO_REUSEADDR)", -1);
	ExitLog();
	return EXIT_FAILURE;
 }

 if((setsockopt(ServerSocket,SOL_SOCKET,SO_KEEPALIVE,(void *)&keepalive_value,
        sizeof(keepalive_value))) == -1) {
        SyscallError("Init: could not set socket option (SO_KEEPALIVE)", -1); 
	ExitLog();
	return EXIT_FAILURE;
 }

 if((setsockopt(ServerSocket,IPPROTO_TCP, TCP_NODELAY,(char *)&just_say_no,
        sizeof(just_say_no))) == -1) {
        SyscallError("Init: could not set socket option (TCP_NODELAY)", -1); 
	ExitLog();
	return EXIT_FAILURE;
 }

 memset((char *) &sa_server, 0, sizeof(sa_server));
 sa_server.sin_family=AF_INET;

#ifdef WITH_LISTEN_IP
 for (i=0; i<=NoOfLocalIps; i++)
 {
#ifdef WIN32
	if (i > 0 && LocalIps[i] &&
	    4 == sscanf(LocalIps[i], "%3d.%3d.%3d.%3d", &b1, &b2, &b3, &b4))
	{
		bind_address.S_un.S_un_b.s_b1 = b1;
		bind_address.S_un.S_un_b.s_b2 = b2;
		bind_address.S_un.S_un_b.s_b3 = b3;
		bind_address.S_un.S_un_b.s_b4 = b4;
	}
	else
		bind_address.s_addr = htonl(INADDR_ANY);
#else
/*
	if (i > 0 && LocalIps[i] &&
            !inet_aton(LocalIps[i], &bind_address.s_addr)) {
		Error("Invalid IP address format", -1);
		ExitLog();
		return EXIT_FAILURE;
	}
*/
	if (i > 0 && LocalIps[i] ) {
		if ( ( bind_address.s_addr = inet_addr( LocalIps[i] )) == -1 ) {
			SyscallError("Invalid IP address format", -1);
			ExitLog();
			return EXIT_FAILURE;
		}
	}
	else
		bind_address.s_addr = htonl(INADDR_ANY);
#endif
 }
#else
 bind_address.s_addr = htonl(INADDR_ANY);
#endif

 sa_server.sin_addr= bind_address;
 sa_server.sin_port = htons(LocalPort);

 if(bind(ServerSocket,(struct sockaddr *) &sa_server, sizeof(sa_server)) == -1)
 {
	if (bind_address.s_addr != htonl(INADDR_ANY))
		SyscallError("Init: could not bind to address port, listen socket", -1);
	else
		SyscallError("Init: could not bind to port %d, listen socket", -1);
	ExitLog();
	return EXIT_FAILURE;
 }

 if (listen (ServerSocket, 256))
/* if (listen (ServerSocket, 512))	for the time being */
 {
	Error("Init: listen failed", -1);
	ExitLog();
	return EXIT_FAILURE;
 }

/* listenerThread = _beginthread( &Service_https, 0, &ServerSocket);*/

  sqlbeginthread(WORKER_STACK_WAHTTP,
                    Service_https, 
                    &ServerSocket, 
                    THR_CREATE_SUSPENDED_EO07, 
                    &listenerThread,
                    errtext, 
                    &err);

 if (err == THR_OK_EO07)
 {
	sqlresumethread(listenerThread, errtext, &err);
	if (err == THR_OK_EO07)
	{
		Error("Init: starting Web Server", -1);
		return(EXIT_SUCCESS);
	}
	else
	{
		Error(errtext, -1);
		return(EXIT_FAILURE);
	}
 }
 else
 {
	Error(errtext, -1);
	return(EXIT_FAILURE);
 }
}

/****************************************************************************
 ***
 *** WaitWebServer
 ***
 ****************************************************************************/

int WaitWebServer( sapdbwa_Bool debug )
{

	tsp00_Int4		 status;
	tsp00_ErrTextc	 errtext;
	teo07_ThreadErr	 err;

	if ( !listenerThread ) {
		Error( "Not running, termination request ignored", -1 );
		return EXIT_FAILURE;
	}

#ifdef WIN32
	if ( debug == sapdbwa_False ) {
		/* Report the status SERVICE_RUNNING	*/
		ServiceIsRunning();
	}
#endif

/*
	ServiceLoop = FALSE;
	do {
		ThreadSleep(1000);

		if (!sqlthreadalive( listenerThread, errtext, &err)) {
			break;
		}

	} while ( err == THR_OK_EO07 && !(stopWebServer == sapdbwa_True) );
*/
/*
#pragma "Tony is anoyed by the cpu eating jointhread ... "
Sleep(3600000);
*/
	sqljointhread( listenerThread, &status, errtext, &err );

	if (err != THR_OK_EO07) {
		Error(errtext, -1);
	}

	listenerThread = NULL;
	Error("terminated", -1);

	return EXIT_SUCCESS;

}

/****************************************************************************
 ***
 *** StopWebServer
 ***
 ****************************************************************************/

int StopWebServer()
{
	tsp00_Int4		status;
	tsp00_ErrTextc	errtext;
	teo07_ThreadErr	err;

	if ( !listenerThread ) {
		Error( "Not running, termination request ignored", -1 );
		return EXIT_FAILURE;
	}

	shutdown(AcceptingSocket, SD_BOTH);
	closesocket( AcceptingSocket );
	ServiceLoop = FALSE;
	stopWebServer = sapdbwa_True;
/*
	do {
		ThreadSleep( 1000 );
		if (!sqlthreadalive( listenerThread, errtext, &err)) {
			break;
		}
	} while ( err == THR_OK_EO07 );
*/
	sqljointhread( listenerThread, &status, errtext, &err );

	if (err != THR_OK_EO07) {
		Error(errtext, -1);
	}

	listenerThread = NULL;
	Error( "terminated", -1 );

	return EXIT_SUCCESS;
}

/****************************************************************************
 ***
 *** ExitWebServer 
 ***
 ****************************************************************************/

int ExitWebServer()
{

	ExitRequestProcessing();

#ifdef WITH_LISTEN_IP
	if (LocalIps) free(LocalIps);
#endif

	ExitIniPar();
	ExitLog();

	return(EXIT_SUCCESS);

}

#ifdef kannweg
/****************************************************************************
 ***
 *** TcpGetIp
 ***
 ****************************************************************************/

unsigned long TcpGetIp( struct st_tcp_connection	*tcp_con )
{
	return( ntohl(tcp_con->sa.sin_addr.S_un.S_addr) );
}
#endif