#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "SAPDB/WebAgent/HTTP/WAHTTP_Agent.h"
#include "SAPDB/WebAgent/HTTP/WAHTTP_WebAgent.h"
#include "SAPDB/WebAgent/HTTP/WAHTTP_Memory.h"

#include "gwd00.h"
#include "hwd01wadef.h"
#include "hwd03waapi.h"
#include "hwd20control.h"
#include "hwd21req.h"
#include "hwd22rep.h"
#include "hwd28dict.h"
#include "hwd29date.h"

#define MODULE_NAME "WAHTTP_WebAgent.c"
#ifndef __FILE__
#define __FILE__ MODULE_NAME
#endif

sapdbwa_TraceDefModule( MODULE_NAME )

struct st_agent_client
{
 int something;
};

int InitWebAgent
(
 struct st_agent_server *as,
 struct st_agent_client	**wah
)
{
 sapdbwa_TraceDefFunction( "InitWebAgent" );

 if (!wd20ApiInit(NULL, API_SAPDBWWW_WD20 ))
	return false;

 sapdbwa_TraceAdd( "WebAgent API Init successfull." );

 return(true);
}

int ExitWebAgent
(
 struct st_agent_server *as,
 struct st_agent_client	*wah
)
{
 return(FALSE);
}

typedef struct st_wa_sapw3_server_handle
{
 struct st_agent_server *as;
 struct st_agent_req	*req;
 struct st_agent_rep	*rep;
 twd28DictionaryP	headerFields;	/* allocated header fields of the
					 * request allocated by
					 * wd52GetHeaderField are stored here */
} *twd52SapW3ServerHandle;

twd52SapW3ServerHandle wd52_CreateServerHandle
(
 struct st_agent_server *as,
 struct st_agent_req	*req,
 struct st_agent_rep	*rep
)
{
 twd52SapW3ServerHandle	handle;
 sapdbwa_Bool		allocatOK;

 wahttp_SQLALLOCAT( sizeof( struct st_wa_sapw3_server_handle), &handle, &allocatOK );
 if (!allocatOK)
 {
	/* memory allocation error */
	wd20WriteLogMsg( ERR_MSG_MEMORY_WD26 );
	return NULL;
 }; /* if */

 handle->headerFields = wd28CreateDictionary();
 if (handle->headerFields == NULL)
 {
	wahttp_SQLFREE( handle );
	wd20WriteLogMsg( ERR_MSG_MEMORY_WD26 );
	return NULL;
 }; /* if */
    
 handle->as = as;
 handle->req = req;
 handle->rep = rep;

 return handle;
} /* wd52_CreateServerHandle */

void wd52_DestroyServerHandle
(
 twd52SapW3ServerHandle handle
)
{
    wd28DestroyDictionary( handle->headerFields );        
    wahttp_SQLFREE( handle );
} /* wd52_DestroyServerHandle */

sapdbwa_UInt4 wd52SapW3ReadBody
(
 twd00ServerHandle	handle,
 char			*b,
 sapdbwa_UInt4		len
)
{
 twd52SapW3ServerHandle	sapw3Handle = (twd52SapW3ServerHandle) handle;

 return(sapw3Handle->as->Receive(sapw3Handle->req, b, len));
} /* wd52SapW3ReadBody */

const char* wd52SapW3GetHeader
(
 twd00ServerHandle	handle,
 const char		*fieldName
)
{
 twd52SapW3ServerHandle	sapw3Handle = (twd52SapW3ServerHandle) handle;
 const char		*result;
 char			*value;
    
  /* did we already retrieve this field ? */
 result = wd28GetValueByKey( sapw3Handle->headerFields, fieldName );
 if (result == NULL)
 {
	/* try to get value from web server */

	/* get the value */
	value = sapw3Handle->as->GetHeaderValueByName( sapw3Handle->req, (char*) fieldName);
	if (value)
	{
		if (!wd28AddCopyEntry( sapw3Handle->headerFields, fieldName, value ))
		{
			result = NULL;
	        }
		else
		{
			result = wd28GetValueByKey( sapw3Handle->headerFields, fieldName );
		}; /* else */
	}
	else
	{
		result = NULL;
	}; /* else */
 }; /* if */
    
 return result;
} /* wd52SapW3GetHeader */

sapdbwa_Bool wd52SapW3IsSecure
(
 twd00ServerHandle	handle
)
{
 const char *isSecure = NULL;
    
 isSecure = wd52SapW3GetHeader( handle, "SERVER_PORT_SECURE" );

 return (isSecure != NULL && (strcmp( isSecure, "1" ) == 0));
} /* wd52SapW3IsSecure */

sapdbwa_Bool wd52SapW3SendHeader
(
 twd00ServerHandle	handle,
 sapdbwa_HttpReplyP	rep
)
{
 twd52SapW3ServerHandle	sapw3Handle;
 char			*tmp;
 struct st_agent_server *sapw3_as;
 struct st_agent_rep	*sapw3_rep;
 char			Date[MAX_DATE_LEN_WD29] = "";
 sapdbwa_UInt4		iHeader;
 twd28DictionaryP	headers = wd22GetHeaders( rep );
 char			*name, *value;
 int			status;

 sapw3Handle = (twd52SapW3ServerHandle) handle;
 sapw3_as = sapw3Handle->as;
 sapw3_rep = sapw3Handle->rep;

 wd29HttpDate(Date);
 sapw3_as->AddHeader_char(sapw3_rep, AGENT_DATE, Date);

 tmp = wd22GetLocation(rep);
 if( tmp != NULL)
 {
	sapw3_as->AddHeader_char(sapw3_rep, AGENT_LOCATION, tmp);
 };

 tmp = wd22GetLastModified(rep);
 if( tmp != NULL)
 {
	sapw3_as->AddHeader_char(sapw3_rep, AGENT_LAST_MODIFIED, tmp);
 };

 tmp = wd22GetContentLength(rep);
 if( tmp != NULL)
 {
	sapw3_as->AddHeader_char(sapw3_rep, AGENT_CONTENT_LENGTH, tmp);
 };

 tmp = wd22GetContentType(rep);
 if( tmp != NULL)
 {
	sapw3_as->AddHeader_char(sapw3_rep, AGENT_CONTENT_TYPE, tmp);
 };

 tmp = wd22GetExpires(rep);
 if( tmp != NULL && wd29WADateToHttpDate( tmp, Date))
 {
	sapw3_as->AddHeader_char(sapw3_rep, AGENT_EXPIRES, Date);
 };

 for (iHeader = 0; iHeader < wd28NumEntries( headers ); iHeader++)
 {
	if (wd28GetEntryByIndex( headers, iHeader, (const char**)&name, &value ))
	{
		/* there must not be a space between the ':' and the value;
		 * else IE 5 is confused */	/* Check this with sapw3 to make sure (TG) */
		sapw3_as->AddHeaderByName_char(sapw3_rep, name, value);
	}; /* if */
 }; /* for */
    
 switch (wd22GetStatus(rep))
 {
	case sapdbwa_Status_Ok:
		status = AGENT_STATUS_OK;
		break;
	case sapdbwa_Status_Created:
		status = AGENT_STATUS_CREATED;
		break;
	case sapdbwa_Status_Response:
		status = AGENT_STATUS_NO_CONTENT;
		break;
	case sapdbwa_Status_Moveto:
		status = AGENT_STATUS_MOVE_PERM;
		break;
	case sapdbwa_Status_Redirect:
		status = AGENT_STATUS_REDIRECT;
		break;
	case sapdbwa_Status_Modified:
		status = AGENT_STATUS_NOT_MODIFIED;
		break;
	case sapdbwa_Status_BadRequest:
		status = AGENT_STATUS_BAD_REQUEST;
		break;
	case sapdbwa_Status_Unauthorized:
		status = AGENT_STATUS_UNAUTHORIZED;
		break;
	case sapdbwa_Status_Forbidden:
		status = AGENT_STATUS_FORBIDDEN;
		break;
	case sapdbwa_Status_NotFound:
		status = AGENT_STATUS_NOT_FOUND;
		break;
	case sapdbwa_Status_MethodNotAllowed:
		status = AGENT_STATUS_METHOD_NOT_ALLOWED;
		break;
	case sapdbwa_Status_Conflict:
		status = AGENT_STATUS_CONFLICT;
		break;
	case sapdbwa_Status_ServerError:
		status = AGENT_STATUS_SERVER_ERROR;
		break;
	case sapdbwa_Status_NotImplemented:
		status = AGENT_STATUS_NOT_IMPLEMENTED;
		break;
	default:
		status = AGENT_STATUS_SERVER_ERROR;
 }

 if (!sapw3_as->SendHeader(sapw3_rep, status))
 {
	wd20WriteLogMsg( __FILE__ ": sapw3_as->SendHeader failed.\n" );
	return(false);
 };
 return(true);
} /* wd52SapW3SendHeader */

sapdbwa_UInt4 wd52SapW3SendBody
(
 twd00ServerHandle	handle,
 const void		*Buf,
 sapdbwa_UInt4		len
)
{
 twd52SapW3ServerHandle	sapw3Handle = (twd52SapW3ServerHandle) handle;

 return(sapw3Handle->as->Send(sapw3Handle->rep, (void *)Buf, len));
} /* wd52SapW3SendBody */


enum en_af_returncode TryWebAgent
(
 struct st_agent_server *sapw3_as,
 struct st_agent_client	*sapw3_wah,
 struct st_agent_req	*sapw3_req,
 struct st_agent_rep	*sapw3_rep
)
{
 DWORD			rtc = af_ignored;
 twd21HttpRequestP	req = NULL;
 sapdbwa_HttpReplyP	rep = NULL;
 twd52SapW3ServerHandle handle = NULL;

 handle = wd52_CreateServerHandle( sapw3_as , sapw3_req , sapw3_rep );
 if (handle == NULL)
 {
	/* memory allocation error */
	wd20WriteLogMsg( ERR_MSG_MEMORY_WD26 );
	return rtc;
 } /* if */
        
 req = wd21CreateRequest( handle, wd52SapW3ReadBody, wd52SapW3GetHeader, wd52SapW3IsSecure );
 if (req == NULL)
 {
	/* memory allocation error */
	wd20WriteLogMsg( ERR_MSG_MEMORY_WD26 );
	wd52_DestroyServerHandle( handle );
	return rtc;
 }; /* if */

 rep = wd22CreateReply( handle, wd52SapW3SendHeader, wd52SapW3SendBody );
 if (rep == NULL)
 {
	/* memory allocation error */
	wd20WriteLogMsg( ERR_MSG_MEMORY_WD26 );
	wd21DestroyRequest( req );        
	wd52_DestroyServerHandle( handle );
	return rtc;
 }; /* if */
    
 wd21SetUri( req, sapw3_req->uri );

 /*wd52_BillsIfModifiedSince( req );*/
 wd21SetQueryString( req, sapw3_req->querystring );

#ifdef notyet
 wd21SetPathInfo( req, sapw3_req->path_info );
#else
/*#pragma path_info_not_implemented_yet*/
 /*wd20WriteLogMsg( "path_info not implemented yet...\n" );*/
#endif

 wd21SetRequestMethod( req, sapw3_req->method );
 wd21SetContentType( req, sapw3_as->GetHeaderValue(sapw3_req, AGENT_CONTENT_TYPE ));
 wd21SetContentLength( req, sapw3_as->GetHeaderValue(sapw3_req, AGENT_CONTENT_LENGTH ));

#ifdef notyet
 wd21SetPhysicalPath( req, sapw3_req->physical_path );
#else
/*#pragma physical_path_not_implemented_yet*/
 /*wd20WriteLogMsg( "physical_path not implemented yet..." );*/
#endif

#ifdef notyet
 wd21SetServerName( req, sapw3_as->server_name );
#else
/*#pragma server_name_not_implemented_yet*/
 /*wd20WriteLogMsg( "server_name not implemented yet..." );*/
#endif

 switch ( wd20CallServiceFunc( req, rep ))
 {
	case sapdbwa_RTC_Ignored:
		rtc = af_ignored;
		break;
	case sapdbwa_RTC_Accepted:
		rtc = af_accepted;
            break;
	case sapdbwa_RTC_Error:
		rtc = af_error;
		break;
        default:
		rtc = af_error;
		break;
 }; /* switch */

 wd21DestroyRequest( req );
 wd22DestroyReply( rep );
 wd52_DestroyServerHandle( handle );
    
 return(rtc);
} /* RequestMap */
