/* @lastChanged: "1998-06-02  10:18"

  * @filename:   vak263.cpp
  * @purpose:    "Stream Handle Functions"
  * @release:    7.1.0.0
  * @see:        ""
  *
  * @copyright:  "Copyright by   SAP AG, 1998"


    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*/

/* ========================================================================== */
/* ==========           REPLACEMENTS BY THE PREPROCESSOR            ========= */
/* ========================================================================== */

/* ========================================================================== */
/* ==========              DECLARATION OF TYPES                     ========= */
/* ========================================================================== */

/* ========================================================================== */
/* ==========              DECLARATION OF FUNCTIONS                 ========= */
/* ========================================================================== */

#include "vak001.h"
#include "hak263.h"
#include "hgg01.h"
#include "hak07.h"
#include "hgg17.h"
#include "hak12.h"
#include "hak34.h"

inline void swapInt2 (tsp00_Uint2& i2) {
  unsigned char* p  = REINTERPRET_CAST(unsigned char*,&i2);
  unsigned char aux = *p;
  *p     = *(p+1);
  *(p+1) = aux;
}

/*----------------------------------------------------------------------------*/

void buildNullTerminatedIdentifier(const tsp00_KnlIdentifier& id, char* dest) {
  int len = sizeof(id);
  while ((len > 0) && (' ' == id[len-1])) {
    --len;
  }
  memcpy(dest, &id[0], len);
  dest[len] = 0;
}

/*----------------------------------------------------------------------------*/

void evalDBProcAndParameterName(tak_all_command_glob& acv,
                                tsp00_KnlIdentifier&  dbproc,
                                int                   paramNo,
                                char*                 pParameterName)
{
  tak_param_info_ptr pInfo;
  a12FindParameter (acv, acv.a_curr_user_id, dbproc, paramNo, pInfo);
  if (NULL != pInfo) {
    int len = pInfo->param_name_len;
    if (len > sizeof(tsp00_KnlIdentifier)) {
      len = sizeof(tsp00_KnlIdentifier);
    }
    memcpy(pParameterName, &pInfo->param_name[0], len);
    pParameterName[len] = 0;
  }
  else {
    sprintf(pParameterName, "Parameter(%d)", paramNo);
  }
}
 
/*----------------------------------------------------------------------------*/
 
void incompatibleAbapDatatypes(ABTypes abapType, tsp00_DataType_Enum paramDataType, char* errorMsg) {
  strcpy(errorMsg, "incompatible datatypes (");
  switch (abapType) {
  case ABTYPC :
    strcat(errorMsg, "ABTYPC");
    break;
  case ABTYPDATE :
    strcat(errorMsg, "ABTYPDATE");
    break;
  case ABTYPP :
    strcat(errorMsg, "ABTYPP");
    break;
  case ABTYPTIME :
    strcat(errorMsg, "ABTYPTIME");
    break;
  case ABTYPX :
    strcat(errorMsg, "ABTYPX");
    break;
  case ABTYPFLOAT :
    strcat(errorMsg, "ABTYPFLOAT");
    break;
  case ABTYPINT :
    strcat(errorMsg, "ABTYPINT");
    break;
  case ABTYPINT2 :
    strcat(errorMsg, "ABTYPINT2");
    break;
  default :
    sprintf(errorMsg, "incompatible datatypes (%d", abapType);
  }
  strcat(errorMsg, ",");
  integer len = (integer)strlen(errorMsg);
  int     pos = (int)strlen(errorMsg);
  g17paramspec_to_line (paramDataType, 0, 0, len, errorMsg);
  // remove length and fraction info
  while ((errorMsg[pos]) && (errorMsg[pos] != '(')) {
    ++pos;
  }
  if ('(' == errorMsg[pos]) {
    errorMsg[pos] = 0;
  }
  strcat(errorMsg, ")");
}

/*----------------------------------------------------------------------------*/

inline void checkABAPDataTypes (ABTypes abapType,                  tsp00_Int2 abapLength, 
                                tsp00_DataType_Enum paramDataType, tsp00_Int2 paramLength,
                                bool& swapColumn, tgg00_BasisError& e, char* errorMsg) {
  switch (abapType) {
  case ABTYPC :
    {
      if ((paramDataType != dchb) && (paramDataType != dcha)) {
        e = e_incompatible_datatypes;
      }
      else {
        if (abapLength != paramLength) {
          sprintf(errorMsg, "different length : ABAP(%d), C++(%d)", abapLength, paramLength);
          e = e_incompatible_datatypes;
          return;
        }
      }
      break;
    }
  case ABTYPDATE :
    {
      if ((paramDataType != dchb) && (paramDataType != dcha)) {
        e = e_incompatible_datatypes;
      }
      break;
    }
  case ABTYPP :
    {
      if (paramLength != 8) {
        sprintf(errorMsg, "wrong length for packed_8_3: C++(%d)", paramLength);
        e = e_incompatible_datatypes;
        return;
      }
      swapColumn = true;
      break;
    }
  case ABTYPTIME :
    {
      break;
    }
  case ABTYPX :
    {
      break;
    }
  case ABTYPTABH : 
    {
      break;
    }
  case ABTYPNUM : 
    {
      break;
    }
  case ABTYPFLOAT :
    {
      if (paramDataType != dfloat) {
        e = e_incompatible_datatypes;
      }
      swapColumn = true;
      break;
    }
  case ABTYPINT :
    {
      if (paramDataType != dinteger) {
        e = e_incompatible_datatypes;
      }
      swapColumn = true;
      break;
    }
  case ABTYPINT2 :
    {
      if (paramDataType != dsmallint) {
        e = e_incompatible_datatypes;
      }
      swapColumn = true;
      break;
    }
  case ABTYPINT1 :
    {
      break;
    }
  case ABTYPW :
    {
      break;
    }
  case ABTYP1 :
    {
      swapColumn = true;
      break;
    }
  case ABTYP2 :
    {
      swapColumn = true;
      break;
    }
  }
  if (e != e_ok) {
    incompatibleAbapDatatypes(abapType, paramDataType, errorMsg);
  }
}

/*----------------------------------------------------------------------------*/

inline void checkStreamDataTypes (OmsStreamTypes streamType, tsp00_DataType_Enum paramDataType,
                                bool& swapColumn, tgg00_BasisError& e) {
  switch (streamType) {
  case STYPE_CHAR :
    {
      if ((paramDataType != dchb) && (paramDataType != dcha)) {
        e = e_incompatible_datatypes;
        return;
      }
      break;
    }
  case 	STYPE_I4  :
  case	STYPE_INT :
    {
      if (paramDataType != dinteger) {
        e = e_incompatible_datatypes;
        return;
      }
      swapColumn = true;
      break;
    }
  case 	STYPE_UI2 :
  case  STYPE_UI4 :
  case  STYPE_UINT:
    {
      if (paramDataType != dfixed) {
        e = e_incompatible_datatypes;
        return;
      }
      swapColumn = true;
      break;
    }
  case  STYPE_I2  :
    {
      if (paramDataType != dsmallint) {
        e = e_incompatible_datatypes;
        return;
      }
      swapColumn = true;
      break;
    }
  case STYPE_I1 :
  case STYPE_UI1:
    {
      break;
    }
  case STYPE_R4 :
  case STYPE_R8 :
    {
      if (paramDataType != dfloat) {
        e = e_incompatible_datatypes;
        return;
      }
      swapColumn = true;
      break;
    }

  default :
    e = e_not_implemented;
  }
}

/*----------------------------------------------------------------------------*/
/* PTS 1107790 */
bool buildStreamDescriptorElement(tsp00_DataType      dataType,
                                       int                 offset,
                                       int                 length,
                                       OmsTypeABAPColDesc& colDesc,
                                       tgg00_BasisError&   e)
{
  e             = e_ok;
  colDesc.inout = 0;
  colDesc.dec   = 0;
  switch(dataType) {
  case dfixed :
    {
      switch (length) {
      case 2 : /* 2 digits => 1 Byte Integer */
        return false;
      case 5 : /* 5 digits => 2 Byte Integer */
        colDesc.abap_type = STYPE_I2;
        colDesc.length    = 2;
        break;
      case 10 : /* 10 digits => 4 Byte Integer */
        colDesc.abap_type = STYPE_I4;
        colDesc.length    = 4;
        break;
      case 19 : /* 19 digits => 8 Byte Integer */
        colDesc.abap_type = STYPE_I8;
        colDesc.length    = 8;
        break;
      default :
        {
          e = e_invalid;
        }
      }
      colDesc.offset = offset;
      return true;
    }
  case dfloat :
    if (6 == length) {
      colDesc.abap_type = STYPE_R4;
      colDesc.length    = 4;
    }
    else {
      if (15 == length) {
        colDesc.abap_type = STYPE_R8;
        colDesc.length    = 8;
      }
      else {
        e = e_invalid;
      }
    }
    colDesc.offset = offset;
    return true;
  case dinteger :
    if (10 == length) {
      colDesc.abap_type = STYPE_I4;
      colDesc.length    = 4;
    }
    else {
      if (19 == length) {
        colDesc.abap_type = STYPE_I8;
        colDesc.length    = 8;
      }
      else {
        e = e_invalid;
      }
    }
    colDesc.offset    = offset;
    return true;
  case dsmallint :
    colDesc.abap_type = STYPE_I2;
    colDesc.length    = sizeof(tsp00_Int2);
    colDesc.offset    = offset;
    return true;
  }
  return false;
}

/* ========================================================================== */
/* ==========              DEFINITION OF GLOBAL FUNCTIONS           ========= */
/* ========================================================================== */

externC
void ak263CheckHandle (tak_all_command_glob& acv,
            tsp00_Int4&           st_ix,
            ak263_StreamHandle*   phandle,
            int                   handleLen,
            bool                  isABAP,
            tgg00_BasisError&     e,
            tsp00_Int4            paramNo,
            tsp00_KnlIdentifier&  methodName)

{ 
      const pasbool isError = true;
 
      tsp00_DataType          paramDataType;
      int                     ix;
      long                    col_count    = 0;
      long                    offset       = 0;
      unsigned short          swapColCount = 0;
      bool                    swap;
      bool                    swapColumn;
      tsp00_Int4              errIndex = 0;
      OmsTypeABAPColDesc      *colDesc;
      char                    errorMsg[256];
      char                    cMethodName[sizeof(tsp00_KnlIdentifier)+1];
      char                    cParamName [sizeof(tsp00_KnlIdentifier)+1];
 
      /* DEC OSF1 cxx had problem with tsp00_swap_kind and tsp00_SwapKind mixup */
      swap = ( ((int)acv.a_out_packet->sp1_header.sp1h_mess_swap) 
            != ((int)g01code.kernel_swap));
      if (swap) {
        // swap handle
        swapInt2 (phandle->tabHandle.rowSize);
        swapInt2 (phandle->tabHandle.colCount);
        if ((int)(sizeof(*phandle) - sizeof(phandle->tabHandle.colDesc)  +
          phandle->tabHandle.colCount * sizeof(phandle->tabHandle.colDesc[0])) > handleLen) {
          phandle->tabHandle.colCount = (unsigned short)(handleLen - 
            (sizeof(*phandle) - sizeof(phandle->tabHandle.colDesc))) / sizeof(phandle->tabHandle.colDesc[0]);
        }
        for (ix = 0; ix < phandle->tabHandle.colCount; ix++) {
          colDesc = &phandle->tabHandle.colDesc[ix];
          swapInt2 (colDesc->dec);
          swapInt2 (colDesc->length);
          swapInt2 (colDesc->offset);
        }
        phandle->tabHandle.rowCount = (isABAP) ? 1 : 0;
      }
      bool isApplicationHandle = (0 != phandle->tabHandle.colCount); 
      OmsTypeABAPColDesc* abapColDesc = &phandle->tabHandle.colDesc[0];
      tgg00_StackEntry*   st_curr     = &((*acv.a_mblock.mb_st())[st_ix]);
      tgg00_StackEntry*   st_upb      = &((*acv.a_mblock.mb_st())
             [acv.a_mblock.mb_qual()->mqual_pos() - 1 + acv.a_mblock.mb_qual()->mqual_cnt() - 1]);
      e = e_ok;
      do {
        if  (st_curr > st_upb) {
          break;
        }
        else
          if (st_curr->etype() != st_op) {
            break;
          }
          else {
            ++errIndex;
            ++st_ix;
            paramDataType.becomes((tsp00_DataType_Enum)  st_curr->epos());
            ++col_count;
            if (isABAP) {
              if (col_count > phandle->tabHandle.colCount) {
                sprintf(errorMsg, "too many members in C++ definition");
                e = e_too_many_columns;
              }
            }
            swapColumn = false;
            if (isABAP) {
              if (abapColDesc->offset != st_curr->ecol_pos()) {
                sprintf(errorMsg, "different offsets : ABAP(%d), C++(%d)", abapColDesc->offset, st_curr->ecol_pos());
                e = e_incompatible_datatypes;
              }
              else {
                checkABAPDataTypes ((ABTypes) abapColDesc->abap_type, abapColDesc->length, 
                  paramDataType, st_curr->elen_var(), swapColumn, e, errorMsg);
              }
            }
            else {
              if (!isApplicationHandle) {
                /* PTS 1107790 */
                swapColumn = buildStreamDescriptorElement(paramDataType, 
                  st_curr->ecol_pos(), st_curr->elen_var(), *abapColDesc, e);
              }
              else {
                checkStreamDataTypes ((OmsStreamTypes) abapColDesc->abap_type, paramDataType, swapColumn, e);
              }
            }
            if (e_ok != e) {
              char OpMsg[400];
              buildNullTerminatedIdentifier(methodName, cMethodName);
              evalDBProcAndParameterName(acv, methodName, paramNo, cParamName);   
              tsp00_KnlIdentifier errorIdentifier;
              memset(&errorIdentifier[0], ' ', sizeof(errorIdentifier));
              sprintf((char*) &errorIdentifier[0], "(Param %d,Member %d)", paramNo, errIndex);
              a07_nb_put_error(acv, e, 1, errorIdentifier);
              sprintf(OpMsg, "%d %s::%s(member%d) : %s", acv.a_return_segm->sp1r_returncode(), 
                cMethodName, cParamName, errIndex, errorMsg); 
              a34OpMsg (isError, *REINTERPRET_CAST(tsp00_MoveObj*, OpMsg), (int)strlen(OpMsg));
              return;
            }
            if ((swap) && (swapColumn)) {
               phandle->tabHandle.colDesc[swapColCount] = *abapColDesc;
               ++swapColCount;
            }
            ++abapColDesc;
            ++st_curr;
          }
      }
      while (true);
      if (isABAP) {
        if (col_count != phandle->tabHandle.colCount) {
          sprintf(errorMsg, "too few members in C++ definition, ABAP(%d),C++(%d)", 
            phandle->tabHandle.colCount, col_count);
          a34OpMsg (isError, *REINTERPRET_CAST(tsp00_MoveObj*, errorMsg), (int)strlen(errorMsg));
          e = e_too_few_columns;
          a07_b_put_error(acv, e, 1);
        }
      }
      phandle->tabHandle.colCount = swapColCount;
}

/*----------------------------------------------------------------------------*/

tsp00_Int4 ak263EvalHandleLength (tsp00_Int4 ColCount) {
  ak263_StreamHandle h;
  h.tabHandle.colCount = 0; /* just to suppress compiler warning */
  return (sizeof (h) - sizeof(h.tabHandle.colDesc) + 
    ColCount * sizeof(h.tabHandle.colDesc[0])); 
}
