/* @lastChanged: "1999-04-21  15:46"

  * @filename:   vak341.cpp
  * @purpose:    "Kernel Sink"
  * @release:    7.2.0.0
  * @see:        ""
  *
  * @copyright:  "Copyright by   SAP AG, 1998"


    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*/


/* ========================================================================== */
/* ==========           REPLACEMENTS BY THE PREPROCESSOR            ========= */
/* ========================================================================== */

#include "hgg10.h"
#include "hkb04.h"
#include "hbd01.h"
#include "hbd02.h"
#include "hbd04.h"
#include "hak34.h"
#include "hak341.h"
#include "vak001.h"
#include "hak07.h"
#include "hgg01.h"
#include "gco00.h"
#include "hkb50.h"
#include "hak05.h"
#include "heo01.h"
#include "heo58.h"
#include "heo78.h"
#include "heo02x.h"
#include "heo38.h"
#include "gco02.h"
#include "hgg01.h"
#include "hgg04.h"

/* ========================================================================== */
/* ==========              DECLARATION OF TYPES                     ========= */
/* ========================================================================== */


typedef bool (*QUALTEST) (unsigned char*, void*);

/* ========================================================================== */
/* ==========              DECLARATION OF FUNCTIONS                 ========= */
/* ========================================================================== */

bool ak341QualificationTest (void* QualFunc,
                   void*          QualValues,
                   unsigned char* Obj);

/* ========================================================================== */
/* ==========                     Globals                           ========= */
/* ========================================================================== */

static LibOmsInterface_cco02* ak341LibOmsInterfacePtr;

/* ========================================================================== */
/* ==========              DEFINITION OF GLOBAL FUNCTIONS           ========= */
/* ========================================================================== */

externC void
ak341CreateObjFile (tgg00_TransContext &Trans,
                    tgg00_ContainerId  &ObjFile,
                    tsp00_Int4	       ObjLen,
                    tsp00_Int4	       KeyLen)
                    
{
  Trans.trError_gg00 = e_ok;
  kb04CreateObjFile (Trans, ObjFile, ObjLen, KeyLen);
}

/*-----------------------------------------------------------------------------------*/

externCpp void
ak341CreateVersion (tak_all_command_glob&     acv,
                     tak_oms_version_id&      versionId,
                     tgg00_OmsVersionContext& versionContext,
                     tgg00_BasisError&        e)
{
  struct tSurrogateRecord {
    tsp00_C2   filler;
    tsp00_C2   hi;
    tsp00_Int4 lo;
  };

  pasbool         IsEotExcl;
  pasbool         AuxPermExists;
  pasbool         ObjLogExists;
  tsp00_TaskId    TransTaskId;
  union {
     tgg00_Surrogate surrogate;
     tSurrogateRecord surrogateRec;
  } versionSurrogate;

  e = e_ok;
  k50my_state (acv.a_transinf.tri_trans.trIndex_gg00,
      IsEotExcl, AuxPermExists,
      ObjLogExists, TransTaskId);
  if (ObjLogExists) {
    e = e_transaction_end_required;
    return;
  }
  a05surrogate_get (acv, versionSurrogate.surrogate);
  g04build_temp_tree_id (versionContext.ovc_tree, acv.a_transinf.tri_trans);
  versionContext.ovc_tree.fileSession_gg00().gg90SetNil();    
  versionContext.ovc_tree.fileType_gg00().addElement(ftsShared_egg00);
  versionContext.ovc_tree.fileType_gg00().addElement(ftsConcurrent_egg00);
  versionContext.ovc_tree.fileTfnTemp_gg00().becomes(ttfnOmsVersion_egg00);
  versionContext.ovc_tree.fileLevel_gg00()   = versionSurrogate.surrogateRec.hi;
  versionContext.ovc_tree.fileTempCnt_gg00() = versionSurrogate.surrogateRec.lo;
  versionContext.ovc_index_tree              = versionContext.ovc_tree;
  versionContext.ovc_key                     = 0;
  b01tcreate_file (acv.a_transinf.tri_trans, versionContext.ovc_tree);
  e = acv.a_transinf.tri_trans.trError_gg00;
  if (e_ok == e) {
     versionContext.ovc_index_tree.fileTfnTemp_gg00().becomes(ttfnOmsVersionIndex_egg00);
     b01tcreate_file (acv.a_transinf.tri_trans, versionContext.ovc_index_tree);
     e = acv.a_transinf.tri_trans.trError_gg00;
  }
}

/*-----------------------------------------------------------------------------------*/

externC
void ak341DropContainer (tak_all_command_glob& acv,
                         tgg00_ContainerId&    ObjFile,
                         tgg00_BasisError&     e) 
{
  acv.a_transinf.tri_trans.trError_gg00 = e_ok;
  bd04DropObjFile (acv.a_transinf.tri_trans, ObjFile);
  e = acv.a_transinf.tri_trans.trError_gg00;
}

/*-----------------------------------------------------------------------------------*/

void
ak341DropVersion (tak_all_command_glob&    acv,
                  tgg00_OmsVersionContext& VersionContext,
                  tgg00_BasisError&        e)
{
  
  ak341EndConsistentView (&acv, &VersionContext.ovc_trans_version, &e);
  if (e_ok == e) {
    b01destroy_file (acv.a_transinf.tri_trans, VersionContext.ovc_tree);
    e = acv.a_transinf.tri_trans.trError_gg00;
  }
  if (e_ok == e) {
    b01destroy_file (acv.a_transinf.tri_trans, VersionContext.ovc_index_tree);
    e = acv.a_transinf.tri_trans.trError_gg00;
  }
}

/*-----------------------------------------------------------------------------------*/

externC void
ak341EndConsistentTrans (tgg00_TransContext &Trans)
{
  Trans.trError_gg00 = e_ok;
  if (!Trans.trConsistView_gg00.gg90IsNil()) {
    kb04EndConsistentTrans (Trans);
  }
}

/*-----------------------------------------------------------------------------------*/

externCpp void 
ak341InquireKey (tgg00_TransContext &Trans,
                 tgg00_ContainerId  &ObjFile,
                 tgg00_ObjColDesc   &ColDesc)
{
  Trans.trError_gg00 = e_ok;
  a34InquireKey (Trans, ObjFile, ColDesc);
}

/*-----------------------------------------------------------------------------------*/

externC tsp00_Bool
ak341GetMonitorInfo (tak_all_command_glob& acv,
             tsp00_Addr&                   handle,
             tsp00_C16&                    iid,
             tsp00_Int4&                   dispid,
             tgg00_COMMonitorInfo&         monInfo)
{
  return ak341LibOmsInterfacePtr->GetMonitorInfo (
    REINTERPRET_CAST(IliveCacheSink*, &acv.a_kernel_sink), handle, iid, dispid, monInfo);
}

/*-----------------------------------------------------------------------------------*/

externC
void ak341ResetMonitor(tak_all_command_glob& acv)
{
  ak341LibOmsInterfacePtr->ResetMonitor(REINTERPRET_CAST(IliveCacheSink*, &acv.a_kernel_sink));
}

/*-----------------------------------------------------------------------------------*/

externC void
ak341GetObjFileInfo (tgg00_TransContext  &Trans,
					 tgg00_ContainerId   &ObjFileId,
					 tsp00_Int4          &NumAllPages,
					 tsp00_Int4          &NumFreePages,
					 tsp00_Int4          &NumEmptyPages,
           tsp00_Int4          &NumKeyLeaves,
					 tsp00_Int4          &Occupancy)
{				
	Trans.trError_gg00 = e_ok;

	/* change request 1000139, T.A. 14.10.1998 */
	/* PTS            1107153, A.K. 17.07.2000 */
	bd04GetObjFileStatistics (Trans, ObjFileId, NumAllPages, 
		NumFreePages, NumEmptyPages, NumKeyLeaves, Occupancy);    
}

/*-----------------------------------------------------------------------------------*/

externCpp
IliveCacheSink* ak341GetSink() {
  tak_all_command_glob* pAcv = REINTERPRET_CAST(tak_all_command_glob*, vGetAcvPtrFromCurrentTask());
  return REINTERPRET_CAST(IliveCacheSink*, &pAcv->a_kernel_sink);
}

/*-----------------------------------------------------------------------------------*/

externC 
void ak341LoadLibOms(tak_all_command_glob& acv) {
//#if defined(NMP)    PTS 1107243
//  return; // not yet implemented
//#else
  tsp00_Bool        ok;
  tsp00_Pathc       szInstPath;
  tsp01_RteError    rteError;
  teo78_ModuleName  OmsLibPath;
  tsp_errtext       errtext;
  HANDLE            hd = 0;
#if defined(WIN32)
  ok = sqlGetDbrootPgmPath (szInstPath, TERM_WITH_DELIMITER_EO01, &rteError );
#else
  ok = sqlGetDbrootLibPath (szInstPath, TERM_WITH_DELIMITER_EO01, &rteError );
#endif
  if (ok) 
  {
    strcpy(OmsLibPath, szInstPath);
    strcat(OmsLibPath, "liboms" );
  	eo38LoadLibraryEx(OmsLibPath, true, &hd, errtext, sizeof(tsp_errtext)-1);
  }
  if (0 == hd) {
    a07_b_put_error (acv, e_no_coclass_obj_found, 1);
  }
  else {
    ProcGetInterfacePtr_co02 I = (ProcGetInterfacePtr_co02) sqlGetProcAddress(hd, (char*) "co10_GetInterface",
      REINTERPRET_CAST(char*, &errtext[0]), sizeof(errtext));
    if (NULL == I) {
      a07_b_put_error (acv, e_invalid, 1);
      return;
    }
    ak341LibOmsInterfacePtr = (*I)();
    ak341LibOmsInterfacePtr->AdviseGetSinkFunc(ak341GetSink);
    ak341LibOmsInterfacePtr->AdviseHeapLimit(g01oms_heap_limit());
    ak341LibOmsInterfacePtr->AdviseOmsRegionCnt(g01region_cnt(rgnOMS_egg00));
    tak_kernel_sink* pSink = &acv.a_kernel_sink;
    ak341LibOmsInterfacePtr->InitLibOms(REINTERPRET_CAST(IliveCacheSink*, pSink));
  }
//#endif
}

/*-----------------------------------------------------------------------------------*/

externC
void ak341OmsHeapInfo(tak_all_command_glob& acv, tsp00_8ByteCounter& totalHeap, tsp00_8ByteCounter& freeSpaceInFreeChunks) {
  tak_kernel_sink* pSink = &acv.a_kernel_sink;
  ak341LibOmsInterfacePtr->GetMemoryUsageInfo(REINTERPRET_CAST(IliveCacheSink*, pSink), totalHeap, freeSpaceInFreeChunks);
}

/*-----------------------------------------------------------------------------------*/

externC
tsp00_Bool ak341NextVersion (tak_all_command_glob& acv,
                       tsp00_Bool          isFirstCall,
                       tsp00_C24&          versionId,
                       tsp00_Date&         createDate,
                       tsp00_Time&         createTime,
                       tsp00_Date&         openDate,
                       tsp00_Time&         openTime,
                       tgg91_TransNo&      consistentView,
                       tsp00_Bool&         isMarked,
                       tsp00_Bool&         isOpen,
                       tsp00_Bool&         isUnloaded,
                       tsp00_8ByteCounter& heapUsage,
                       tgg00_FileId&       versionTree,
	                     tgg00_FileId&       versionInvTree) {
  tak_kernel_sink* pSink = &acv.a_kernel_sink;
  if (isFirstCall) {
    return ak341LibOmsInterfacePtr->GetFirstVersion(REINTERPRET_CAST(IliveCacheSink*, pSink), versionId, createDate, createTime,
                                            openDate, openTime, consistentView, isMarked, isOpen, isUnloaded,
                                            heapUsage, versionTree, versionInvTree);
  }
  else {
    return ak341LibOmsInterfacePtr->GetNextVersion(REINTERPRET_CAST(IliveCacheSink*, pSink), versionId, createDate, createTime,
                                            openDate, openTime, consistentView, isMarked, isOpen, isUnloaded,
                                            heapUsage, versionTree, versionInvTree);
  }
}

/*-----------------------------------------------------------------------------------*/

externCpp void
ak341NextObj (tak_all_command_glob  *acv,
            tgg91_TransNo           *ConsistentView,
            tgg00_ContainerId       *ObjFile,
            tgg00_OmsVersionContext *VersionContext,
            tgg92_KernelOid         *CurrOid,
            tsp00_Int4              ObjBodySize,
            void*                   pObjBody,
            tgg91_PageRef           *ResultObjVers,
            short                   *e)
           
{
  bool       bWithObject;
  tsp00_Int4 ObjBodyLen;
  bWithObject = (NULL != pObjBody);
  ak341SetConsistentView (acv, ConsistentView);
  
  if ((!CurrOid->gg92IsNil()) && (VERSION_OBJ_PAGE_NO_GG92 == CurrOid->gg92GetPno())) {
    ak341VersionNextObj (*acv, VersionContext,
      *CurrOid, 
      ObjFile->fileObjFileNo_gg00(), ObjBodySize, 
      REINTERPRET_CAST(tsp00_MoveObj*, pObjBody), 
      *ResultObjVers, *e); 
  }
  else {
    if (acv->a_transinf.tri_trans.trConsistView_gg00.gg90IsNil()) {
      kb04StartConsistentTrans (acv->a_transinf.tri_trans);
      *e = acv->a_transinf.tri_trans.trError_gg00;
      if (e_ok != *e) {
        return;
      }
      *ConsistentView = acv->a_transinf.tri_trans.trConsistView_gg00; 
    }
    do {
      kb04NextObjConsistent (acv->a_transinf.tri_trans,
        *ObjFile, *CurrOid, bWithObject, ObjBodySize, pObjBody, *ResultObjVers);  
      *e = acv->a_transinf.tri_trans.trError_gg00;
      if (e_ok == *e) {
        if (VersionContext) {
          if (ak341UpdatedInVersion (*acv, VersionContext->ovc_tree, *CurrOid, ObjBodySize, ObjBodyLen, 
            REINTERPRET_CAST(tsp00_MoveObj*, pObjBody), *e)) {
            if (e_object_not_found == *e) {
              *e = e_ok;
              continue; /* ADIS 1001698, T.A. 10.11.1998 */
            }
          }
        }
        return;
      }
    }
    while (e_ok == *e);
    
    if ((e_no_next_object == *e) && 
      (VersionContext) && (VERSION_OBJ_PAGE_NO_GG92 != CurrOid->gg92GetPno() ))
    {
      CurrOid->gg92SetPno(VERSION_OBJ_PAGE_NO_GG92);
      CurrOid->gg92SetPos(0);
      ak341VersionNextObj (*acv, VersionContext,
        *CurrOid,
        ObjFile->fileObjFileNo_gg00(), ObjBodySize, 
        REINTERPRET_CAST(tsp00_MoveObj*, pObjBody), 
        *ResultObjVers, *e);      
    }
  }
}

/*-----------------------------------------------------------------------------------*/


externC void
ak341NilConsistentTrans (tgg00_TransContext&Trans)
{
  Trans.trConsistView_gg00.gg90SetNil();
}

/*-----------------------------------------------------------------------------------*/

externCpp void
ak341OpenVersion (tak_all_command_glob& acv,
                  tgg91_TransNo&        ConsistentView,
                  tgg00_BasisError&     e) {
  pasbool         IsEotExcl;
  pasbool         AuxPermExists;
  pasbool         ObjLogExists;
  tsp00_TaskId    TransTaskId;
  
  e = e_ok;
  k50my_state (acv.a_transinf.tri_trans.trIndex_gg00,
    IsEotExcl, AuxPermExists,
    ObjLogExists, TransTaskId);
  if (ObjLogExists) {
    e = e_transaction_end_required;
  }
}

/*-----------------------------------------------------------------------------------*/

externC pasbool
ak341SetOmsTraceLevel(tsp00_KnlIdentifier& lvl, pasbool enable) 
{
  return ak341LibOmsInterfacePtr->SetTraceLevel(lvl, enable != 0);
}

/*-----------------------------------------------------------------------------------*/

externC void
ak341StartConsistentTrans (tgg00_TransContext &Trans)
{
  Trans.trError_gg00 = e_ok;
  kb04StartConsistentTrans (Trans);
}

/*-----------------------------------------------------------------------------------*/

// externC void
// ak341UnlockObj (tgg00_TransContext	&Trans,
//                 tgg00_FileId		    &ObjFile,	
//                 tgg92_KernelOid       Oid)
     
// {
//  Trans.trError_gg00 = e_ok;
//  kb04UnlockObj (Trans, ObjFile, Oid);
// }

/*-----------------------------------------------------------------------------------*/

externC void 
ak341VarObjGuid (tsp00_C16& guid) 
{
    const GUID g = { 0x8fb0963d, 0xe5aa, 0x11d1, { 0xa9, 0x5c, 0x0, 0xa0, 0xc9, 0x43, 0x7, 0x31 } };

    memcpy (&guid, &g, sizeof(GUID));
}

/*-----------------------------------------------------------------------------------*/

bool ak341QualificationTest (void* QualFunc,
                   void*          QualValues,
                   unsigned char* Obj)
{
    class MyObj {
        bool f (void*) { return true; }
    };
    union {
        bool (MyObj::* FuncPtr) (void*);
        void* voidptr;
    } aux;
    aux.voidptr = QualFunc;
    Obj -= sizeof(void*);
    return (((MyObj*) Obj->*aux.FuncPtr)(QualValues));
}

/*-----------------------------------------------------------------------------------*/

externC
void ak341LockVersionDirectory(tak_all_command_glob& acv) {
  tak_kernel_sink* pSink = &acv.a_kernel_sink;
  ak341LibOmsInterfacePtr->LockVersionDirectory(REINTERPRET_CAST(IliveCacheSink*, pSink));
}

/*-----------------------------------------------------------------------------------*/

externC
void ak341UnLockVersionDirectory(tak_all_command_glob& acv) {
  tak_kernel_sink* pSink = &acv.a_kernel_sink;
  ak341LibOmsInterfacePtr->UnLockVersionDirectory(REINTERPRET_CAST(IliveCacheSink*, pSink));
}

/*-----------------------------------------------------------------------------------*/

void ak341MultiPurpose (tak_all_command_glob* acv,
            tgg00_MessType  MessType,
            tgg00_MessType2 MessType2,
            tsp00_Int4*     BufLen,
            void*           Buf,
            tsp00_Int2*     e)
{
  *e = e_ok;
  switch (MessType) {
  case m_change :
    *e      = a34ReturnCode (*REINTERPRET_CAST(tsp00_Int2*, Buf));
    *BufLen = 0;
    break;
  case m_diagnose :
/* +++ PTS 1106505 FF +++++++++++++++  */
    if ( mm_dump == MessType2 ) {
      eo38CTraceBack(*REINTERPRET_CAST(tsp00_Addr*, Buf) );
    }
    else {
/* --- PTS 1106505 FF ---------------  */
      g01vtrace.vtrBdObject_gg00 = true;
      g01vtrace.vtrOmsNew_gg00   = true;
      g01vtrace.vtrOmsGet_gg00   = true;
      g01vtrace.vtrOmsUpd_gg00   = true;
      g01vtrace.vtrOmsFree_gg00  = true;
      g01vtrace.vtrAny_gg00      = true;
    }  /* PTS 1106505 FF */
    *e      = 0;
    *BufLen = 0;
    break;
  case m_get : {
      switch (MessType2) {
    case mm_length :
      *REINTERPRET_CAST(tsp00_Int4*, Buf) =
        g01packet_size() - sizeof (tsp1_packet_header) -
        sizeof (tsp1_segment_header) - sizeof (tsp1_part_header);
      *BufLen = sizeof(tsp00_Int4);
      break;
    default :
      *e      = e_invalid;
      *BufLen = 0;
    }
    break;
               }
  case m_set : 
    {
      switch (MessType2) {
      case mm_lock_timeout :
        acv->a_transinf.tri_trans.trObjReqTimeout_gg00 = 
          *REINTERPRET_CAST (tsp00_Int2*, Buf);
        *BufLen = 0;
        break;
      default :
        *e      = e_invalid;
        *BufLen = 0;
      }
      break;
    }
  case m_lock :
    a34Lock (*acv, REINTERPRET_CAST(tsp00_MoveObj*, Buf), *BufLen, mm_lock == MessType2, *e);
    *BufLen = 0;
    break;
  case m_show :
    a34OpMsg (mm_log != MessType2, *REINTERPRET_CAST(tsp00_MoveObj*, Buf), *BufLen);
    *BufLen = 0;
    break;
  case m_unlock :
    a34UnLock (*acv, REINTERPRET_CAST(tsp00_MoveObj*, Buf), *BufLen, mm_lock == MessType2, *e);
    *BufLen = 0;
    break;
  default :
    *e = e_invalid;
    *BufLen = 0;
  }
}

/*-----------------------------------------------------------------------------------*/

externC
void ak341Shutdown(tak_all_command_glob& acv)
{
#if defined(NMP)
  return; // not yet implemented
#else
  tak_kernel_sink* pSink = &acv.a_kernel_sink;
  tgg00_BasisError e = ak341LibOmsInterfacePtr->Shutdown(REINTERPRET_CAST(IliveCacheSink*, pSink));
  if (e_ok != e) {
    a07_b_put_error (acv, e, 1);
  }
#endif
}

/*-----------------------------------------------------------------------------------*/

class tak341_VersionRecBuf {
  struct tak341_VersionRecInfo {
    unsigned char   m_header[cgg_rec_key_offset];
    tgg92_KernelOid m_oid;
    tgg00_ObjFileNo m_containerId;
    bool            m_deleted;
  };
public :
  tak341_VersionRecBuf() {}
  tak341_VersionRecBuf(tgg00_ObjFileNo ContainerId, tgg92_KernelOid& Oid, tsp00_Int4 BodySize, tsp00_MoveObj* Body);
  
  tgg00_ObjFileNo GetContainerId() {
    return  REINTERPRET_CAST(tak341_VersionRecInfo*, &m_rec)->m_containerId;
  }
  bool IsDeleted() {
    return REINTERPRET_CAST(tak341_VersionRecInfo*, &m_rec)->m_deleted;
  }
  tgg00_BasisError IsOk() {
    return (m_rec.recLen_gg00() < 0) ? m_rec.recLen_gg00() : e_ok;
  }
  void MarkDeleted() {
    REINTERPRET_CAST(tak341_VersionRecInfo*, &m_rec)->m_deleted = true;
  }
  tsp00_Int4 ObjBodyLen() {
    return  m_rec.recLen_gg00() - OBJ_OFFSET;
  }
  void BuildKeyRec(tgg00_ObjFileNo ContainerId, tsp00_MoveObj* pKey, int KeyLen, tgg92_KernelOid& Oid, tgg00_BasisError& e);

  tgg00_BasisError CopyObj(tsp00_Int4 Size, tsp00_MoveObj* copyBuf) {
    tgg00_BasisError e = e_ok;
    g10mv ("VAK341", 1,
      sizeof(m_rec), Size,
      &m_rec, OBJ_OFFSET + 1, copyBuf, 1, ObjBodyLen(), e);
    return e;
  }
  tsp00_MoveObj* GetInfoPtr() {
    return REINTERPRET_CAST(tsp00_MoveObj*, &m_rec.recBody_gg00()[m_rec.recKeyLen_gg00()]);
  }

  tgg92_KernelOid& GetOid() {
    return *REINTERPRET_CAST(tgg92_KernelOid*, &m_rec.recBody_gg00());
  }

  operator tgg00_Rec&() { 
    return m_rec;
  }
private :

  tgg00_Rec m_rec;
 
  enum {OBJ_OFFSET = cgg_rec_key_offset + sizeof(tgg92_KernelOid) + sizeof(tgg00_ObjFileNo) + sizeof(bool) };
};

/*-----------------------------------------------------------------------------------*/

tak341_VersionRecBuf::tak341_VersionRecBuf(tgg00_ObjFileNo ContainerId, tgg92_KernelOid& Oid, tsp00_Int4 BodySize, tsp00_MoveObj* Body) {
  tgg00_BasisError e = e_ok;
  m_rec.recLen_gg00()          = OBJ_OFFSET + BodySize; 
  m_rec.recKeyLen_gg00()       = sizeof(Oid);
  m_rec.recVarcolCnt_gg00()    = 0;
  m_rec.recVarcolOffset_gg00() = m_rec.recLen_gg00() -
              cgg_rec_key_offset - m_rec.recKeyLen_gg00();
  tak341_VersionRecInfo* pInfo = REINTERPRET_CAST(tak341_VersionRecInfo*, &m_rec);
  pInfo->m_oid         = Oid;
  pInfo->m_containerId = ContainerId;
  pInfo->m_deleted     = false;
  g10mv ("VAK341", 2,
    BodySize, sizeof(m_rec),
    Body, 1, &m_rec, OBJ_OFFSET + 1, BodySize,  m_rec.recLen_gg00());

}

/*-----------------------------------------------------------------------------------*/

void tak341_VersionRecBuf::BuildKeyRec(tgg00_ObjFileNo ContainerId, tsp00_MoveObj* pKey, int KeyLen, tgg92_KernelOid& Oid, tgg00_BasisError& e) {
  e                            = e_ok;
  m_rec.recKeyLen_gg00()       = (tsp00_Int4)(sizeof(ContainerId) + KeyLen);
  m_rec.recLen_gg00()          = cgg_rec_key_offset + m_rec.recKeyLen_gg00() + sizeof(Oid);
  m_rec.recVarcolCnt_gg00()    = 0;
  m_rec.recVarcolOffset_gg00() = m_rec.recLen_gg00() -
              cgg_rec_key_offset - m_rec.recKeyLen_gg00();
  *REINTERPRET_CAST(tgg00_ObjFileNo*, &m_rec.recBody_gg00()) = ContainerId;
  g10mv ("VAK341", 3,
    KeyLen, sizeof(m_rec.recBody_gg00()),
    pKey, 1, &m_rec.recBody_gg00(), sizeof(ContainerId) + 1, KeyLen, e);
  g10mv ("VAK341", 4,
    sizeof(Oid), sizeof(m_rec.recBody_gg00()),
    &Oid, 1, &m_rec.recBody_gg00(), sizeof(ContainerId) + KeyLen + 1, sizeof(Oid), e);
}
            
/*-----------------------------------------------------------------------------------*/

externCpp
bool ak341UpdatedInVersion (tak_all_command_glob& acv,
                        tgg00_FileId&         VersionFile,
                        tgg92_KernelOid&      Oid,
                        tsp00_Int4            ObjBodySize,
                        tsp00_Int4&           ObjBodyLen,
                        tsp00_MoveObj*        Body,
                        tgg00_BasisError&     e)
{

  tgg_lkey             rk;
  tak341_VersionRecBuf Rec;
  rk.len() = sizeof(Oid);
  memcpy(&(rk.k()[0]), &Oid, sizeof(Oid));
  b02get_record (acv.a_transinf.tri_trans, VersionFile, rk, Rec);
  e = acv.a_transinf.tri_trans.trError_gg00;
  if (e_ok == e) {
    if (Rec.IsDeleted()) {
      e = e_object_not_found;
    }
    else {
      ObjBodyLen = Rec.ObjBodyLen();
      e = Rec.CopyObj(ObjBodySize, Body);
    }
  }
  if (e_key_not_found == e) {
    e = e_ok;
    return false;
  }
  else {
    return true;
  }
}

/*-----------------------------------------------------------------------------------*/

externCpp
void ak341VersionDelObj(tak_all_command_glob& acv,
                        tgg00_FileId&         VersionFile,
                        tgg00_ObjFileNo       ContainerId,
                        tgg92_KernelOid&      Oid,
                        bool                  DoInsert,
                        tgg00_BasisError&     e)
{
  tak341_VersionRecBuf Rec(ContainerId, Oid, 0, NULL);
  e = Rec.IsOk();
  if (e_ok == e) {
    Rec.MarkDeleted();
    do {
      if (DoInsert) {
        b02add_record (acv.a_transinf.tri_trans, VersionFile, Rec);
      }
      else {
        b02repl_record (acv.a_transinf.tri_trans, VersionFile, Rec);
      }
      e = acv.a_transinf.tri_trans.trError_gg00;
      if ((e == e_duplicate_key) && (DoInsert)) {
        DoInsert = false;
      }
      else {
        break;
      }
    }
    while (true);
  }
}
                        
/*-----------------------------------------------------------------------------------*/                        

externCpp
void ak341VersionGetObjViaKey (tak_all_command_glob& acv,
                              tgg00_OmsVersionContext* VersionContext,
                              tsp00_Int4               KeyLen,
                              tsp00_MoveObj*           pKey,
                              tgg00_ObjFileNo          ClsId,
                              tsp00_Int4               ObjBodySize,
                              tsp00_MoveObj*           Body,
                              tgg92_KernelOid&         Oid,
                              tgg00_BasisError&        e)
{
  tgg_lkey lKey;
  tak341_VersionRecBuf Rec;
  
  e = e_ok;
  lKey.len() = (tsp00_Int4)sizeof(tgg00_ObjFileNo) + KeyLen;
  *REINTERPRET_CAST(tgg00_ObjFileNo*, &lKey.k()) = ClsId;
  g10mv ("VAK341", 5, 
    KeyLen, sizeof(lKey.k()), pKey, 1, &lKey.k(), sizeof(tgg00_ObjFileNo) + 1, KeyLen, e);
  if (e_ok == e) {
    b02get_record (acv.a_transinf.tri_trans, VersionContext->ovc_index_tree, lKey, Rec);
    e = acv.a_transinf.tri_trans.trError_gg00;
  }
  if (e_ok == e) {
    memcpy (&Oid, Rec.GetInfoPtr(), sizeof(Oid));
    if (ak341UpdatedInVersion (acv, VersionContext->ovc_tree, Oid, ObjBodySize, ObjBodySize, Body, e)) {
      if (e_object_not_found == e) {
        e = e_hash_key_not_found;
      }
    }
    else {
      e = e_invalid;
    }
  }
  else {
    if (e_key_not_found == e) {
       e = e_hash_key_not_found;
    }
  }
}

/*-----------------------------------------------------------------------------------*/     

externC
void ak341VersionNextObj (tak_all_command_glob& acv,
                          tgg00_OmsVersionContext* VersionContext,
                          tgg92_KernelOid&         Oid,
                          tgg00_ObjFileNo          wantedContainerId,
                          tsp00_Int4               ObjBodySize,
                          tsp00_MoveObj*           Body,
                          tgg91_PageRef&           ResultObjVers,
                          tgg00_BasisError&        e)
{
  
  const pasbool c_inclusive = true;
  tgg00_Lkey           lKey;
  tak341_VersionRecBuf Rec;

  lKey.len() = sizeof(Oid);
  acv.a_transinf.tri_trans.trError_gg00 = e_ok;
  do {
    *REINTERPRET_CAST(tgg92_KernelOid*, &lKey.k()) = Oid;
    b02next_record (acv.a_transinf.tri_trans, VersionContext->ovc_tree,
      lKey, !c_inclusive, Rec);
    e = acv.a_transinf.tri_trans.trError_gg00;
    if (e_key_not_found == e) {
      e = e_ok;
    }
    if (e_ok == e) {
      Oid = Rec.GetOid();
      if ((wantedContainerId == Rec.GetContainerId()) && (!Rec.IsDeleted())) {
        Rec.CopyObj(ObjBodySize, Body);
        ResultObjVers.gg91SetDummyRef();
        return;
      }
    }
    else {
      if (e_no_next_record == e) {
        e = e_no_next_object;
      }
      return;
    }
  }
  while (true);
}

/*-----------------------------------------------------------------------------------*/     

externCpp
void ak341VersionUpdObj(tak_all_command_glob& acv,
                        tgg00_OmsVersionContext* VersionContext,
                        tgg00_ObjFileNo          ContainerId,
                        tgg92_KernelOid&         Oid,
                        bool                     DoInsert,
                        tsp00_Int4               BodySize,
                        tsp00_MoveObj*           Body,
                        tgg00_ObjColDesc&        ColDesc,
                        tgg00_BasisError&        e)
{
  tak341_VersionRecBuf Rec(ContainerId, Oid, BodySize, Body);
  e = Rec.IsOk();
  if (e_ok == e) {
    do {
      if (DoInsert) {
        b02add_record (acv.a_transinf.tri_trans, VersionContext->ovc_tree, Rec);
        if ((VERSION_OBJ_PAGE_NO_GG92 == Oid.gg92GetPno()) && (e_ok == acv.a_transinf.tri_trans.trError_gg00)) {
          if (ColDesc.colHashCount_gg00 > 0) {
            tgg00_StackEntry& st = *REINTERPRET_CAST(tgg00_StackEntry*, ColDesc.colHashList_gg00);
            Rec.BuildKeyRec(ContainerId, &Body[st.epos()-1], st.elen_var(), Oid, e);
            if (e_ok == e) {
              b02add_record (acv.a_transinf.tri_trans, VersionContext->ovc_index_tree, Rec);
            }
          }
        }
      }
      else {
        b02repl_record (acv.a_transinf.tri_trans, VersionContext->ovc_tree, Rec);
      }
      e = acv.a_transinf.tri_trans.trError_gg00;
      if ((e == e_duplicate_key) && (DoInsert)) {
        DoInsert = false;
      }
      else {
        break;
      }
    }
    while (true);
  }
}
	  
 
