/****************************************************************************

  module      : gbd11.h

  -------------------------------------------------------------------------

  responsible : FrankH

  special area: Free Block Management
  description : defines / macros / typedefs / prototypes 

  last changed: 1999-10-26  16:06
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1998



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


#ifndef GBD11_H
#define GBD11_H

#ifndef TRUE
#define FALSE 0
#define TRUE  1
#endif
 
/*----------------------------------------------------------------------------*
 *  INCLUDES                                                                  *
 *----------------------------------------------------------------------------*/

#include "gsp00.h"
#include "ggg00.h"
#include "gbd00.h"
#include "gsp03.h"
#include "veo00c.h"
#include "geo60.h"

/*----------------------------------------------------------------------------*
 *  DEFINES                                                                   *
 *----------------------------------------------------------------------------*/


#ifdef TRACE

#define TRACE_BD_FBM bd_psm

extern void                                 t01int4() ;
extern void                                 t01p2int4() ;
extern void                                 cccprint() ;

#else

#define                                     t01int4(a,b,c) {}
#define                                     t01p2int4(a,b,c,d,e) {}
#define                                     cccprint(x) {}

#endif


#define ERR_SHM_TO_SMALL_FOR      55000,sp3m_error, csp3_n_fbm, "not enough memory (shm) for %s"
#define ERR_FST_FREE_PNO_INVALID  55001,sp3m_error, csp3_n_fbm, "invalid first pno(%d) < fst_pno(%d)"
#define ERR_LST_USED_PNO_INVALID  55002,sp3m_error, csp3_n_fbm, "last used pno(%d) not in [%d..%d]"
#define ERR_LST_USED_OFF_INVALID  55003,sp3m_error, csp3_n_fbm, "last used offs(%d) > dev_size(%d) "
#define ERR_MAXFBMPNO_TO_SMALL    55005,sp3m_error, csp3_n_fbm, "max_fbm_pno is:%d, needed:%d"
#define ERR_INVALID_DEVNO         55006,sp3m_error, csp3_n_fbm, "invalid devno %d !, devnos allowed: %d - %d"
#define ERR_BLOCK_NOT_USED        55007,sp3m_error, csp3_n_fbm, "tried to mark free Block (%d/%d) for svp/b_up"

#define ERR_PNO_ALREADY_MARKED_BUP   55008,sp3m_error, csp3_n_fbm, "pno %d already marked for bup (%d/%d)"
#define ERR_BLOCK_ALREADY_MARKED_BUP 55009,sp3m_error, csp3_n_fbm, "block already marked for b_up (%d/%d)"

#define ERR_BLOCK_ALREADY_IN_USE  55010,sp3m_error, csp3_n_fbm, "block already in use (%d/%d)"
#define ERR_INVALID_MARKER_COMBI  55011,sp3m_error, csp3_n_fbm, "invalid marker combination for block (%d/%d)"
#define ERR_MARKER_COMBI_IS       55011,sp3m_error, csp3_n_fbm, "act marker is %d, new marker is %d"
#define ERR_INVALID_DEVNO_OR_OFFS 55012,sp3m_error, csp3_n_fbm, "invalid devno(%d) or offset(%d)"
#define ERR_DEVSP_IS_FULL         55013,sp3m_error, csp3_n_fbm, "devno %d full!"

#define ERR_NUM_UNSAVED_PAGES     55014,sp3m_error, csp3_n_fbm, "%d pages not saved!"
#define ERR_UNSAVED_PNO           55014,sp3m_error, csp3_n_fbm, "pno %ld not saved"
#define ERR_PNO_NOT_MARKED_BUP    55015,sp3m_error, csp3_n_fbm, "pno %d, block (%d/%d) not marked for b_up"
#define ERR_PNO_NOT_BUP_PNO_LIST  55016,sp3m_error, csp3_n_fbm, "pno %d not in backup_pno_list"

/* PTS 1107140 */
#define ERR_FBM_TO_SMALL_IN_PARAM 55017,sp3m_error, csp3_n_fbm, "Number of FBM pages to small in paramfile"
#define ERR_FBM_WANTED_GIVEN      55018,sp3m_error, csp3_n_fbm, "wanted:%d, given:%d"

#define FBM_GLOB  GLOB_STRUCT->


/*----------------------------------------------------------------------------*
 *  MACROS                                                                    *
 *----------------------------------------------------------------------------*/

#define FBM_PAGE_ID(_fbm_page)   (_fbm_page->variant.C_TRUE.fbmp_id_F)
#define FBM_PAGE_PT(_fbm_page)   (_fbm_page->variant.C_TRUE.fbmp_pt_F)
#define FBM_PAGE_F1(_fbm_page)   (_fbm_page->variant.C_TRUE.fbmp_fill_1_F)
#define FBM_PAGE_C1(_fbm_page)   (_fbm_page->variant.C_TRUE.fbmp_check_1_F)
#define FBM_PAGE_F2(_fbm_page)   (_fbm_page->variant.C_TRUE.fbmp_fill_2_F)
#define FBM_PAGE_C2(_fbm_page)   (_fbm_page->variant.C_TRUE.fbmp_check_2_F)
#define FBM_PAGE_MX(_fbm_page)   (_fbm_page->variant.C_TRUE.fbmp_max_F)
#define FBM_PAGE_DATA(_fbm_page) (_fbm_page->variant.C_TRUE.fbmp_list_F)


/*-----------------------------------------------------------*/
/*           MACROS  for  "backup_pno_list"  access          */
/*-----------------------------------------------------------*/
 
#define CLEAR_FLAG_IN_BACKUP_LIST( _pno_list, _pno )            \
            { ( _pno_list.pnos [ _pno >> 3 ] ^= ( 1 << (_pno & 7) ));\
                --_pno_list.num_marked                          ;  \
            }
 
#define BACKUP_LIST_DATA_PNOS(_BitmapPos, _BitmapSize, _FirstDataPno, _LastDataPno ) \
            { _FirstDataPno = _BitmapPos << 3 ;                                    \
              _LastDataPno  = _FirstDataPno + (_BitmapSize << 3) -1;               \
            }

#define SET_FLAG_IN_BACKUP_LIST( _pno_list, _pno )              \
            { ( _pno_list.pnos [ _pno >> 3 ] |= ( 1 << (_pno & 7) ));\
                _pno_list.num_marked++ ;  \
            }
 
#define IS_BACKUP_PNO_FLAG_SET( _pno_list, _pno )                   \
            ( _pno_list [ _pno >> 3 ] &  ( 1 << (_pno & 7) ))
 
/*-----------------------------------------------------------*/
/*   MACROS  to save first_free / last_used fbm-values       */
/*-----------------------------------------------------------*/

/*********************************************************************/
#ifdef DEBUG
/*********************************************************************/
 
#define SAVE_DEV_FIRST_FREE( _devsp, _offs, _fbm_pno, _fbm_offs )  \
      if ( _fbm_pno >= _devsp->fst_fbm_page )                  \
        {  _devsp->fst_free_offs     = _offs     ;             \
           _devsp->fst_free.fbm_page = _fbm_pno  ;             \
           _devsp->fst_free.fbm_offs = _fbm_offs ;             \
        }  \
      else \
        {  \
          MSGD((ERR_FST_FREE_PNO_INVALID, _fbm_pno, _devsp->fst_fbm_page )) ; \
          vabort(TRUE);              \
        }

#define SAVE_DEV_LAST_USED( _devsp, _offs, _fbm_pno, _fbm_offs )  \
      if (_fbm_pno <= _devsp->lst_fbm_page && _fbm_pno >= _devsp->fst_fbm_page)\
        {  _devsp->lst_used_offs     = _offs     ;              \
           _devsp->lst_used.fbm_page = _fbm_pno  ;              \
           _devsp->lst_used.fbm_offs = _fbm_offs ;              \
           if ( _offs      >= _devsp->dev_size ||               \
                _fbm_offs  >= _devsp->dev_size )                \
             { \
               MSGD(( ERR_LST_USED_OFF_INVALID, _offs, _devsp->dev_size )); \
               vabort(TRUE);              \
             }                            \
        }  \
      else \
        {  \
          MSGD(( ERR_LST_USED_PNO_INVALID, \
                  _fbm_pno, _devsp->fst_fbm_page,_devsp->lst_fbm_page )) ; \
          vabort(TRUE);              \
        }
 
/*********************************************************************/
#else
/*********************************************************************/
 
#define SAVE_DEV_FIRST_FREE( _devsp, _offs, _fbm_pno, _fbm_offs )  \
        {  _devsp->fst_free_offs     = _offs     ;             \
           _devsp->fst_free.fbm_page = _fbm_pno  ;             \
           _devsp->fst_free.fbm_offs = _fbm_offs ;             \
        } 
 
#define SAVE_DEV_LAST_USED( _devsp, _offs, _fbm_pno, _fbm_offs )  \
        {  _devsp->lst_used_offs     = _offs     ;              \
           _devsp->lst_used.fbm_page = _fbm_pno  ;              \
           _devsp->lst_used.fbm_offs = _fbm_offs ;              \
           if ( _offs      >= _devsp->dev_size ||               \
                _fbm_offs  >= _devsp->dev_size )                \
             { \
               MSGD(( ERR_LST_USED_OFF_INVALID, _offs, _devsp->dev_size )); \
               vabort(TRUE);              \
             }                            \
        } 

/*********************************************************************/
#endif
/*********************************************************************/

/*-----------------------------------------------------------*/
/*           MACROS  for  "fbm_dirty_list"  access           */
/*-----------------------------------------------------------*/
 
#define CLEAR_FLAG_IN_DIRTY_LIST( fbm_pno, marker )            \
              ( FBM_GLOB fbm_dirty_list [ fbm_pno ] ^= ( 1 << (marker) ))
 
#define SET_FLAG_IN_DIRTY_LIST( fbm_pno, marker )              \
              ( FBM_GLOB fbm_dirty_list [ fbm_pno ] |= ( 1 << (marker) ))
 
#define IS_FBM_PAGE_DIRTY( fbm_pno, marker )                   \
              ( FBM_GLOB fbm_dirty_list [ fbm_pno ] &  ( 1 << (marker) ))
 
/*-----------------------------------------------------------*/
/*           MACROS  for  "devno check"                      */
/*-----------------------------------------------------------*/
 
#define FIRST_DEVNO 1
#define MAX_DATA_DEVSPACES  ( FIRST_DEVNO + FBM_GLOB max_data_devs )

#define VALID_DEVNO(devno) \
     ( devno >= FIRST_DEVNO && devno <= FBM_GLOB perm_devsp_used )

#define VALID_OFFSET(_offset, _dev_size) \
     ( _offset >= 0 && _offset <= _dev_size )

/*-----------------------------------------------------------*/
/*           MACROS  for  fbm_page access                    */
/*-----------------------------------------------------------*/
 
#ifdef DEVSPACE_FOR_FBM_IN_USE
static tbd_fbmpage                         *bd11_get_page ();
static tbd_fbm                             *bd11_get_list ();
static int                                  bd11_srch_oldest_cache_entry() ;
#define BD11_GET_LIST( t, fbm_pno , tr_err) \
        bd11_get_list( t, fbm_pno , tr_err)
#else

#define LAST_VALID_OFFSET(_fbm_pno)         \
    FBM_PAGE_MX( (&FBM_GLOB fbm_cache.cache_page[_fbm_pno]) ) 

#define BD11_GET_LIST( t, fbm_pno , tr_err) \
    FBM_PAGE_DATA((&FBM_GLOB fbm_cache.cache_page[fbm_pno]))

#define bd11_get_page(_t, _fbm_pno, _t_err) \
   &(FBM_GLOB fbm_cache.cache_page[_fbm_pno]) 

#endif

/*-----------------------------------------------------------*/
/*           MACROS to interpret block value                 */
/*-----------------------------------------------------------*/

#define HALF_BYTE_PER_BLOCK
 
#define cgg_max_temp_devs          4

#ifdef TEST
#define MAX_FBM_LIST_BD00            80
#endif


#ifdef ONE_BYTE_PER_BLOCK

#define  cbd_maxfbmlist_entries    (MAX_FBM_LIST_BD00)

#define MARK_BLOCK( block_list, offset, marker) ( block_list[offset]=marker )
#define ACT_BLOCK_VAL( block_list, offset) ( block_list[offset] )

#else
#ifdef HALF_BYTE_PER_BLOCK

#define  cbd_maxfbmlist_entries    (MAX_FBM_LIST_BD00<<1)

#define MARK_BLOCK( block_list, offset, marker)                       \
                      { if ( (offset & 1) == 1 )                      \
                          { block_list [ offset >> 1 ] &= 0xF3 ;      \
                            block_list [ offset >> 1 ] += marker<<2;  \
                          }                                           \
                        else                                          \
                          { block_list [ offset >> 1 ] &= 0x3F ;      \
                            block_list [ offset >> 1 ] += marker<<6 ; \
                          }                                           \
                      }

#define ACT_BLOCK_VAL( block_list, offset)                          \
          ( tbd_fbm )( ( offset & 1 ) == 1                          \
                       ? (block_list [ offset >> 1 ] & 0x0C )>>2    \
                       : (block_list [ offset >> 1 ] >> 6   ) ) 

#define MARK_AFTER_BACKUP( block_list, offset, marker)                       \
                      { if ( (offset & 1) == 1 )                      \
                          { block_list [ offset >> 1 ] &= 0xFC ;      \
                            block_list [ offset >> 1 ] += marker   ;  \
                          }                                           \
                        else                                          \
                          { block_list [ offset >> 1 ] &= 0xCF ;      \
                            block_list [ offset >> 1 ] += marker<<4 ; \
                          }                                           \
                      }

#define VAL_AFTER_BACKUP( block_list, offset)                       \
          ( tbd_fbm )( ( offset & 1 ) == 1                          \
                       ? (block_list [ offset >> 1 ] & 0x03 )       \
                       : (block_list [ offset >> 1 ] & 0x30 )>>4 ) 


          
#else
#define  cbd_maxfbmlist_entries    (MAX_FBM_LIST_BD00<<2)

#define MARK_BLOCK( block_list, offset, marker)                     \
           { switch ( offset & 3 )                                  \
              { case 0 :  block_list [ offset >> 2 ] &= 0x3F ;      \
                          block_list [ offset >> 2 ] += marker<<6 ; \
                          break ;                                   \
                case 1 :  block_list [ offset >> 2 ] &= 0xCF ;      \
                          block_list [ offset >> 2 ] += marker<<4 ; \
                          break ;                                   \
                case 2 :  block_list [ offset >> 2 ] &= 0xF3 ;      \
                          block_list [ offset >> 2 ] += marker<<2 ; \
                          break ;                                   \
                case 3 :  block_list [ offset >> 2 ] &= 0xFC ;      \
                          block_list [ offset >> 2 ] += marker    ; \
                          break ;                                   \
               }                                                    \
             }

#define ACT_BLOCK_VAL( block_list, offset)                          \
          ( tbd_fbm )( ( offset & 3 ) == 3                          \
                       ?  block_list [ offset >> 2 ] & 0x03         \
                     : ( offset & 2 ) == 2                          \
                       ? (block_list [ offset >> 2 ] & 0x0C )>>2    \
                     : ( offset & 1 ) == 1                          \
                       ? (block_list [ offset >> 2 ] & 0x30 )>>4    \
                       : (block_list [ offset >> 2 ] & 0xC0 )>>6 ) 

#endif
#endif

/*-----------------------------------------------------------*/
/*           MACROS  for  "chache_stamp"  access             */
/*-----------------------------------------------------------*/

#ifdef ONE_LONG_STAMP

#define INCR_STAMP_VALUE                  ( ++FBM_GLOB stamp_value )
#define STAMP1_GT_STAMP2( stamp1, stamp2) ( stamp1 > stamp2 )

#else

#define INCR_STAMP_VALUE             { if ( ++FBM_GLOB stamp_value.low == 0 ) \
                                         ++FBM_GLOB stamp_value.high ;        \
                                     }
 
#define STAMP1_GT_STAMP2( stamp1, stamp2) ( stamp1.high > stamp2.high || \
              ( stamp1.high == stamp2.high && stamp1.low > stamp2.low ) )

#endif
#define COPY_STAMP( dest, src )           ( dest   = src    )
#define STAMP_CACHE_ENTRY( stamp )                          \
                { INCR_STAMP_VALUE;                         \
                  COPY_STAMP (stamp,FBM_GLOB stamp_value) ; \
                }


/*----------------------------------------------------------------------------*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                    *
 *----------------------------------------------------------------------------*/
#ifdef ONE_LONG_STAMP
 
typedef unsigned long               stamp_type ;
 
#else
 
struct  stamp_struct
        { tsp00_Int4 high;
          tsp00_Int4 low ;
        } ;
typedef struct stamp_struct         stamp_type ;
 
#endif
 
struct  fbm_position
        { tsp00_PageNo fbm_page;
          tsp00_Int4    fbm_offs;
        } ;
typedef struct fbm_position         fbm_pos ;
 
#define  MAX_MULTIPLE_BLOCKS 16
#define  NOT_AVAIL           -1

struct   fbm_dev_info
         { tsp00_Int4      dev_size             ;
           tsp00_Int4      used_blocks          ;
           tsp00_Int4      tot_blocks_per_usage ;
           tsp00_Int4      act_blocks_per_usage ;
           tsp00_PageNo   fst_fbm_page         ; /* fst fbmpage for this devsp */
           tsp00_PageNo   lst_fbm_page         ; /* lst fbmpage for this devsp */
           tsp00_Int4      fst_free_offs        ; /* fst free offset (absolut)  */
           tsp00_Int4      lst_used_offs        ; /* lst used offset (absolut)  */
           fbm_pos       fst_free             ; /* fbm_pos for fst free block */
           fbm_pos       lst_used             ; /* fbm_pos for lst used block */
           fbm_pos       fst_frag             ; /* fst pos used for fragment  */
           fbm_pos       lst_frag             ; /* last pos used for fragment */
           fbm_pos       lst_mult_used[MAX_MULTIPLE_BLOCKS]    ;
                                                /* lst pos used for mult I/O  */
           fbm_pos       act_bup_pos          ; /* fbm_pos for lst bup block  */
           tsp00_Int4      num_marked_b_up      ; /* b_up pages for this devsp  */
           tsp00_Int4      num_marked_svp_free  ; /* svp  pages for this devsp  */
         };
 
typedef struct   fbm_dev_info   FBM_DEV_INFO ;
 
#define PID(t) ( t->trTaskId_gg00 )
 
struct   fbm_cache_info
         {
           tsp00_Int4                 cache_size ;
           tsp00_Int4                 cache_used ;
           tbd_fbmpage             *cache_page ;
           stamp_type              *stamp ;
           tsp00_Int4                 all_in_cache ;
           tsp00_Int4                 filler_alignment ;
         } ;
 
typedef struct   fbm_cache_info   FBM_CACHE_INFO ;
 
struct   fbm_backup_list
         {
           tsp00_Int4                 list_size  ;
           tsp00_Int4                 num_marked;
           tsp00_Int4                 act_pos   ;
           tsp00_PageNo              act_bitmap_pno;
           unsigned char            *pnos ;
         } ;
 
typedef struct   fbm_backup_list  FBM_BACKUP_LIST;
 
/*-----------------------------------------------------------*/
/*   GLOBAL variables used for FBM  ( from this mudul only ) */
/*-----------------------------------------------------------*/
 
typedef struct fbm_glob_var_struc
        { tsp00_Int4       perm_devsp_used ;
          tsp00_Int4       next_devsp_to_use_for_new_bl ;
          tsp00_Int4       max_data_devs ;
          tsp00_Int4       last_devsp_used_for_frag ;
          stamp_type       stamp_value  ;           /* 2 x tsp00_Int4 */
          tsp00_PageNo     max_fbm_pno ;
          tsp00_Int4       filler_1 ;
          tsp00_Int4       tot_blocks ;
          tsp00_Int4       tot_free_blocks ;
          tsp00_Int4       tot_svp_blocks ;
          tsp00_Int4       svp_triggered ;
          FBM_DEV_INFO   *devsp_info ;
          FBM_CACHE_INFO fbm_cache ;              /* 4 x tsp00_Int4 + 2 x ptr */
          tsp00_PageNo    *is_fbm_in_cache ;
          unsigned char  *fbm_dirty_list ;
          FBM_BACKUP_LIST bup_pno_list ;          /* 4 x tsp00_Int4 + 1 x ptr */
          FBM_BACKUP_LIST save_pg_pno_list ;      /* 4 x tsp00_Int4 + 1 x ptr */
        } t_fbm_glob ;

/*----------------------------------------------------------------------------*
 *  FUNCTION PROTOTYPES                                                       *
 *----------------------------------------------------------------------------*/

 
externC void bd11_init_bup_struc     ( FBM_BACKUP_LIST *bup_struc ) ;
 
externC void bd11_change_devsp_usage ( tgg00_TransContext   *t, 
                                       tgg00_BasisError     *tr_err ) ;

#endif  /* GBD11_H */
