/****************************************************************************

  module      : gbd900.h

  -------------------------------------------------------------------------

  responsible : Alexander Kley (d025251 - Alexander.Kley@SAP-AG.de)

  special area: "BD_Object_Constants_And_Types"
  description : for the implementation of the non inline methods see vbd900.cpp


  version     : 7.2.0.0
  last changed: 1999-08-02  11:01
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1998

*****************************************************************************/

#ifndef GBD900_H
#define GBD900_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/
 
#include "ggg92.h"
#include "gbd00.h"
#include "gbd920.h"

#include "hbd01.h"
#include "hbd92.h"
#include "hgg01.h"

#if COMPILEMODE_MEO00 >= SLOW_MEO00 
#include "hta99.h" /* no check */
#include "hta01.h" /* no check */
#endif

#if COMPILEMODE_MEO00 >= QUICK_MEO00 
#include "hgg01_1.h" 
#include "gsp03.h" 
#include "gsp03_3.h" 
#endif

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define MAX_EMPTY_PAGES_BD900                   1
#define MAX_PORTION_OF_EMPTY_PAGES_BD900        20

		/* The garbage collector (=GC) is started to remove empty pages */ 
        /* from the file if there are more than MAX_EMPTY_PAGES_BD900   */
		/* empty pages and if the portion of empty pages (in %) with    */
        /* respect to all pages of an object file is larger than        */
        /* MAX_PORTION_OF_EMPTY_PAGES_BD900 the GC                      */

#define MAX_NUM_CHAINS_PER_CONTAINER_BD900      1

        /* gives the maximum number of object page chains in one       */
        /* object container                                            */ 

#define BALANCE_LIMIT_BD900                     50
   
        /* if an object is shortened by an update so that it occupies less  */
        /* than the object frame length of the currrent file times          */
		/* BALANCE_LIMIT_GGbd900 (in %) divided by                          */
        /* CONT_FRAME_ENLARGE_FACTOR_BD900 an other file with smaler frames */
        /* is choosen to accomodate the the object                          */

#define CONT_FRAME_ENLARGE_FACTOR_BD900         200

        /* factor (in %) by which the obj frames in the continuation       */
        /* files increase from a file to the file with next larger frames  */

#define MAX_NUM_CONT_OBJ_FILES_BD900            12

        /* maximum number of continuation files which belong to a */
		/* file containing variable long objects                  */

#define PRIM_CONT_OBJ_FILE_NO_BD900             0

         /* this numbers is used in the fileid if the file          */
         /* contains no continuation object. i.e the object frames  */
         /* contain the object header with lock an log infos        */

#define  HASH_WIDTH_BD900                       1024
     
         /* number of entry slots of the hash container which  */ 
		 /* accomodated the object file infos                  */

#define  NUM_OBJ_FILE_INFOS_TO_ALLOCATE_BD900   2000
 
         /* number of ObjFileInfo frames the hash allocates if a new       */ 
         /* ObjFileInfo should be inserted but no empty frame is available */

#define  NUM_CONT_OBJ_FILE_INFOS_TO_ALLOCATE_BD900 500

         /* number of ContObjFileInfo frames the hash allocates if a new       */ 
         /* ContObjFileInfo should be inserted but no empty frame is available */

#define STORE_OBJECTS_ON_SUBROOT_BD900 YES

         /* if this variable is defined objects are stored on the subroots too */

#define NIL_OBJ_FILLCHAR_BD900                  0

#define NIL_GC_INDEX_BD900                     -1

#define DUMPMARKSIZE_BD900                      8

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  FORWARD DECLARATIONS  OF CLASSES, STRUCTURES, TYPES, UNIONS              *
 *===========================================================================*/

class cbd900_ObjFileInfo;

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

struct tbd900_ObjFrame
{
    tgg92_ObjHeader     objHeader_bd900;
    tgg91_ObjAnyBody    objBody_bd900;
};

/*---------------------------------------------------------------------------*/

struct tbd900_ContObjHeader
{
    tsp00_PageNo        cvoRootOIdPno_bd900;
    tgg00_ObjPagePos    cvoRootOIdPos_bd900;
    tgg00_ObjPagePos    cvoNextFreeObj_bd900;
};

/*---------------------------------------------------------------------------*/

struct tbd900_ContObjFrame
{
    tbd900_ContObjHeader cofHeader_bd900;
    tgg91_ObjAnyBody     cofBody_bd900;
};

/*---------------------------------------------------------------------------*/

class tbd900_VarObjBodyHeader
{
public:
    tsp00_Int4          vhdObjSize_bd900;
    tsp00_Int4          vhdContObjFileNo_bd900;
    tsp00_PageNo        vhdContObjPno_bd900;
    tgg00_ObjPagePos    vhdContObjPos_bd900;
    tsp00_Int2          vhdFiller_bd900;

	void bd900SetToEmptyHeader()
	{
		vhdContObjPno_bd900    = NIL_PAGE_NO_GG00;
		vhdContObjPos_bd900    = NIL_OBJ_PAGE_POS_GG92;
		vhdContObjFileNo_bd900 = PRIM_CONT_OBJ_FILE_NO_BD900;
		vhdObjSize_bd900       = 0;
		vhdFiller_bd900        = 0;
	}
};

/*---------------------------------------------------------------------------*/

struct tbd900_VarObjBody
{
    tbd900_VarObjBodyHeader vbdHeader_bd900;
    tgg91_ObjAnyBody        vbdBody_bd900;
};

/*---------------------------------------------------------------------------*/

typedef tsp00_Int4       tbd900_PageCounter;

/*---------------------------------------------------------------------------*/

class  cbd900_ObjChainInfo
{
	friend class cbd900_ObjFileInfo;
	
public://members 
    tsp00_PageNo    ociSubRoot_bd900;
    tsp00_PageNo    ociStartPageRelPage_bd900;
    tsp00_Int4      ociAllObjCnt_bd900;        // # of all object frames which are not free           
	tsp00_Int4      ociAllPageCnt_bd900;       // # of all pages in chain starting at the subroot    
	tsp00_Int2      ociCurrActivGC_bd900;      // index of the GC currently active on this chain 
    tsp00_Bool      ociRelPageIsReq_bd900;
	tsp00_Bool      ociFiller1_bd900;
	
public://methods
    cbd900_ObjChainInfo();
	void  bd900SetChainInfo   (tbd_node_ptrs   &NptrsSubRoot); 
	void  bd900ResetChainInfo ();        

private://methods
	void  bd900DumpChainInfo (tsp00_Int4          BufferSize,
	                          tsp00_Page         &Buffer,
                              tsp00_Int4         &OutPos);

	/* note that the dump routines in vxt08 rely on the size of this structure */
    CHECK_MAPPABLE_CLASS_GIVEN_SIZE_SP00 (cbd900_ObjChainInfo, 20)
};

/*---------------------------------------------------------------------------*/ 

class cbd900_ObjFileInfo
{
public: // members
	tsp00_Int4         ofiCurrChain_bd900;
    tsp00_Int4         ofiMaxChain_bd900;
	tsp00_Int4         ofiMaxObjPerPageCnt_bd900;
	tsp00_Int4         ofiMaxObjBodySize_bd900;
	tsp00_PageNo       ofiRoot_bd900;
	tsp00_PageNo       ofiKeyRoot_bd900;
	tsp00_Bool         ofiFileIsForVarObj_bd900;
    tsp00_Bool         ofiRelFileIsReq_bd900;
    tsp00_Int2         ofiFiller_bd900;

    ARRAY_VAR_MSP00 (ofiChain_bd900, cbd900_ObjChainInfo, MAX_NUM_CHAINS_PER_CONTAINER_BD900);    

public: //methods
	cbd900_ObjFileInfo ();
	cbd900_ObjFileInfo (tbd_current_tree  &Current,
		                tbd_node_ptrs     &NptrsRoot,
						tsp00_Int4         MaxObjBodySize,
						tsp00_Int4         MaxObjPerPageCnt);
	tsp00_Int4   bd900CurrChain ();        // current chain in which the next new object will be inserted   
	void         bd900ResetFileInfo (); 
	bool         bd900FileInfoIsEmpty() const;    
    tsp00_Int4   bd900NumAllPages()     const; // Number of all pages in the file  
	tsp00_Int4   bd900NumOccPages()     const; // Number of pages required to accomodate all none free object frames 
    void         bd900GetNumOfActiveGCs (tsp00_Int4 & NumReqGC,
                                         tsp00_Int4 & NumActiveGC);
	void         bd900SetFileInfo (tbd_current_tree     &Current,
		                           tbd_node_ptrs        &NptrsRoot,
								   tsp00_Int4            MaxObjBodySize,
								   tsp00_Int4            MaxObjPerPageCnt);
    void         bd900DumpFileInfo (tgg00_VfFileref    &HostFile,
    								tsp00_Uint4         BufferSize,
									tsp00_Page         &Buffer,
									tsp00_Int4         &OutPno,
									tsp00_Int4         &OutPos,
									tsp00_VfReturn     &HostError,
									tsp00_ErrText      &ErrText,
									tsp00_Int2         &ElemsOnThisPage,
									tsp00_Int4         &NumElemPos,
									tsp00_Int4          FileNo,
									tsp00_Int4          ContFileNo,
									tsp00_Int2          DumpCode,
									char               *DumpMark);

    cbd900_ObjChainInfo & operator[] (tsp00_Int4 i) {return ofiChain_bd900[i];};

	void  bd900BuildFileId    (tsp00_Int4            FileNo,
							   tgg00_FileId         &ObjFileId);
};

/*---------------------------------------------------------------------------*/

class  cbd900_ObjFileDir : 
public cbd920_ChainedListHash <tgg00_ObjFileNo, cbd900_ObjFileInfo, HASH_WIDTH_BD900, NUM_OBJ_FILE_INFOS_TO_ALLOCATE_BD900>

{
public://methods 
	void  bd900DumpObjFileDir (tgg00_VfFileref      &HostFile,
		                       tsp00_Int4            BufferSize,
		                       tsp00_Page           &Buffer,
		                       tsp00_Int4           &OutPno,
		                       tsp00_Int4           &OutPos,
		                       tsp00_VfReturn       &HostError,
	                           tsp00_ErrText        &ErrText);
};

/*---------------------------------------------------------------------------*/

typedef cbd920_HashIter <tgg00_ObjFileNo, cbd900_ObjFileInfo, HASH_WIDTH_BD900, NUM_OBJ_FILE_INFOS_TO_ALLOCATE_BD900> 
cbd900_ObjFileDirIter;

/*---------------------------------------------------------------------------*/

class  cbd900_ContObjFileInfo
{
public://members 
	tsp00_Int4          cfiNumContFiles_bd900;
private://members
	ARRAY_VAR_MSP00    (cfiContFile_bd900, cbd900_ObjFileInfo, MAX_NUM_CONT_OBJ_FILES_BD900);
	
public://methods
	cbd900_ContObjFileInfo() : cfiNumContFiles_bd900(0) {};
	cbd900_ObjFileInfo & operator[] (tsp00_Int4 i) {return cfiContFile_bd900[i-1];};

	void bd900BuildFileId (tsp00_Int4       FileNo,
		                   tsp00_Int4       ContFileNo,
		                   tgg00_FileId    &ContObjFileId);
};

/*---------------------------------------------------------------------------*/

class  cbd900_ContObjFileDir : 
public cbd920_ChainedListHash <tgg00_ObjFileNo, cbd900_ContObjFileInfo, HASH_WIDTH_BD900, NUM_CONT_OBJ_FILE_INFOS_TO_ALLOCATE_BD900>
{
public://methods
   	void bd900DumpContObjFileDir (tgg00_VfFileref      &HostFile,
	                              tsp00_Int4            BufferSize,
		                          tsp00_Page           &Buffer,
		                          tsp00_Int4           &OutPno,
		                          tsp00_Int4           &OutPos,
		                          tsp00_VfReturn       &HostError,
		                          tsp00_ErrText        &ErrText);
};

/*---------------------------------------------------------------------------*/

typedef cbd920_HashIter <tgg00_ObjFileNo, cbd900_ContObjFileInfo, HASH_WIDTH_BD900, NUM_CONT_OBJ_FILE_INFOS_TO_ALLOCATE_BD900>
cbd900_ContObjFileDirIter;

/*---------------------------------------------------------------------------*/

class cbd900_ObjKeyFileId : public tgg00_FileId
{
public:
    cbd900_ObjKeyFileId (const cbd900_ObjFileInfo &ObjFileInfo, 
		                 const tsp00_Int4          FileNo);
};

/*---------------------------------------------------------------------------*/

class cbd900_RootPageData
{
	friend class cbd900_RootPageInfo;

	tsp00_Uint2   rpdNumChains_bd900;                                        // # of chains in this container
	tsp00_Bool    rbdFileIsDeleted_bd900;                                    // file is marked for delete
	tsp00_Bool    rpdFiller1_bd900;
	tsp00_Uint2   rpdNumContFile_bd900;                                      // # of continuation files
	tsp00_Uint2   rpdFiller2_bd900; 
	tsp00_PageNo  rpdSubRoot_bd900  [MAX_NUM_CHAINS_PER_CONTAINER_BD900];    // list of subroots of this root
	tsp00_PageNo  rpdContRoot_bd900 [MAX_NUM_CONT_OBJ_FILES_BD900];          // list of roots of continuation files
};

/*---------------------------------------------------------------------------*/

class cbd900_SubRootPageData
{
	friend class cbd900_SubRootPageInfo;

    tbd900_PageCounter   spdAllPageCnt_bd900;        // # all pages in chain starting at the subroot 
    tbd900_PageCounter   spdEmptyPageCnt_bd900;      // # empty pages 
    tbd900_PageCounter   spdFreePageCnt_bd900;       // # pages with free objetcs 
    tsp00_Int4           spdEmptyPageLimit_bd900;    // current limit of empty pages (in %) before GC is started
	tsp00_Int4           spdMaxPageSeqNo_bd900;      // max page sequence number
	tsp00_Int4           spdAllObjCnt_bd900;         // dirty counter of all objects which are not free  
};

/*---------------------------------------------------------------------------*/

class cbd900_RootAndSubRootPageData
{
	friend class cbd900_RootPageInfo;
	friend class cbd900_SubRootPageInfo;

	cbd900_SubRootPageData rsdSubRootPageData_bd900;
    cbd900_RootPageData    rsdRootPageData_bd900;

	CHECK_MAPPABLE_CLASS_UPPER_SIZE_SP00 (cbd900_RootAndSubRootPageData,  FULLCOVERING_BD00)
};

/*---------------------------------------------------------------------------*/

class cbd900_RootAndSubRootPageInfo
{
	friend class cbd900_RootPageInfo;
	friend class cbd900_SubRootPageInfo;

private:
    cbd900_RootAndSubRootPageData * rsiRootAndSubRootPageDataPtr_bd900;
	cbd900_RootAndSubRootPageInfo (tbd_node_ptrs  &Nptrs);
};

/*---------------------------------------------------------------------------*/

class cbd900_RootPageInfo : public cbd900_RootAndSubRootPageInfo
{
public:
	cbd900_RootPageInfo (tbd_node_ptrs   &Nptrs);
	tsp00_Uint2           &bd900NumChains();
	tsp00_Bool           &bd900FileIsDeleted();
    const tsp00_PageNo   &bd900SubRoot (const int ChainNo);
	tsp00_Uint2          &bd900NumContFiles();
    tsp00_PageNo         &bd900ContRoot (const int ContFileNo);
	void                  bd900InitRootPageInfo();
	void                  bd900AddSubRoot (tsp00_PageNo SubRootPageNo);

private:
    tsp00_PageNo         &bd900_SubRoot (const int ChainNo);
};

/*---------------------------------------------------------------------------*/

class cbd900_SubRootPageInfo : public cbd900_RootAndSubRootPageInfo
{
public:
	cbd900_SubRootPageInfo (tbd_node_ptrs  &Nptrs);
	tsp00_Int4        &bd900AllPageCnt ();
	tsp00_Int4        &bd900EmptyPageCnt ();
	tsp00_Int4        &bd900FreePageCnt ();
	tsp00_Int4        &bd900EmptyPageLimit ();
	tsp00_Int4        &bd900MaxPageSeqNo ();
	tsp00_Int4        &bd900AllObjCnt ();
	void               bd900InitSubRootPageInfo();
};

/*===========================================================================*
 *  DEFINITION OF METHODS                                                    *
 *===========================================================================*/

/*---------------------------------------------------------------------------*/
/* METHODS of cbd900_ObjChainInfo                                            */
/*---------------------------------------------------------------------------*/

inline  void
cbd900_ObjChainInfo::bd900DumpChainInfo (tsp00_Int4          BufferSize,
										 tsp00_Page         &Buffer,
										 tsp00_Int4         &OutPos)
{
    ROUTINE_DBG_MEO00 ("bd900DumpChainInfo");

	memcpy (Buffer + OutPos - 1, this,  sizeof(*this));
	OutPos += sizeof(*this);
}

/*---------------------------------------------------------------------------*/

inline void
cbd900_ObjChainInfo::bd900ResetChainInfo()
{
    ociRelPageIsReq_bd900     = false;
    ociSubRoot_bd900          = NIL_PAGE_NO_GG00;
    ociStartPageRelPage_bd900 = NIL_PAGE_NO_GG00;
    ociAllObjCnt_bd900        = 0;
    ociAllPageCnt_bd900       = 0;   
    ociCurrActivGC_bd900      = NIL_GC_INDEX_BD900;
}

/*---------------------------------------------------------------------------*/

inline void  
cbd900_ObjChainInfo::bd900SetChainInfo (tbd_node_ptrs   &NptrsSubRoot)
{    
	bd900ResetChainInfo();
	cbd900_SubRootPageInfo SubRootPageInfo (NptrsSubRoot);
    ociSubRoot_bd900          = NptrsSubRoot.np_ptr()->nd_id();
    ociAllObjCnt_bd900        = SubRootPageInfo.bd900AllObjCnt();
    ociAllPageCnt_bd900       = SubRootPageInfo.bd900AllPageCnt();  
}

/*---------------------------------------------------------------------------*/

inline
cbd900_ObjChainInfo::cbd900_ObjChainInfo()
:ociFiller1_bd900 (0)
{
	bd900ResetChainInfo();
}

/*---------------------------------------------------------------------------*/
/* METHODS of cbd900_ObjFileInfo                                             */
/*---------------------------------------------------------------------------*/

inline
cbd900_ObjFileInfo::cbd900_ObjFileInfo (tbd_current_tree     &Current,
										tbd_node_ptrs        &NptrsRoot,
									    tsp00_Int4            MaxObjBodySize,
									    tsp00_Int4            MaxObjPerPageCnt)
{ 
	bd900SetFileInfo (Current, NptrsRoot, MaxObjBodySize, MaxObjPerPageCnt);
}

/*---------------------------------------------------------------------------*/

inline
cbd900_ObjFileInfo::cbd900_ObjFileInfo()
{    
    bd900ResetFileInfo();
}

/*---------------------------------------------------------------------------*/

inline tsp00_Int4
cbd900_ObjFileInfo::bd900CurrChain ()
{ 
    tsp00_Int4 CurrChain (ofiCurrChain_bd900);
    CurrChain  = (++CurrChain % ofiMaxChain_bd900);
    ofiCurrChain_bd900 = CurrChain;
    return (CurrChain);
};

/*---------------------------------------------------------------------------*/

inline	bool         
cbd900_ObjFileInfo::bd900FileInfoIsEmpty() const
{
	return (ofiRoot_bd900 == NIL_PAGE_NO_GG00);
};

/*---------------------------------------------------------------------------*/
/* METHODS of cbd900_ObjFileInfo                                             */
/*---------------------------------------------------------------------------*/

inline externCpp void
cbd900_ObjFileInfo::bd900BuildFileId (tsp00_Int4       FileNo ,      //[inp]
									 tgg00_FileId    &ObjFileId)     //[out]
{
	ROUTINE_DBG_MSP00 ("bd900BuildFileId");
	
	ObjFileId = b01niltree_id;
	ObjFileId.fileRoot_gg00()          = ofiRoot_bd900;
    ObjFileId.fileObjFileNo_gg00()     = FileNo;
	ObjFileId.fileContObjFileNo_gg00() = PRIM_CONT_OBJ_FILE_NO_BD900;
	ObjFileId.fileRootCheck_gg00()     = ROOT_CHECK_GG00;
	ObjFileId.fileTfn_gg00().becomes(tfnObj_egg00); 
    ObjFileId.fileBdUse_gg00().addElement (bd_write_acc);
	ObjFileId.fileType_gg00().clear();
    ObjFileId.fileType_gg00().addElement (ftsObject_egg00);      
    ObjFileId.fileType_gg00().addElement (ftsConcurrent_egg00); 
    ObjFileId.fileType_gg00().addElement (ftsPerm_egg00);
	if (this->ofiFileIsForVarObj_bd900)
		ObjFileId.fileObjFileType_gg00().becomes (oftVarLenObjFile_egg00);
	else
		ObjFileId.fileObjFileType_gg00().becomes (oftFixLenObjFile_egg00);
}

/*---------------------------------------------------------------------------*/
/* METHODS of cbd900_ContObjFileInfo                                         */
/*---------------------------------------------------------------------------*/
inline void 
cbd900_ContObjFileInfo::bd900BuildFileId (tsp00_Int4      FileNo,
										  tsp00_Int4       ContFileNo,
										  tgg00_FileId    &ContObjFileId)
{
	ROUTINE_DBG_MSP00 ("bd900BuildFileId");
	
	ContObjFileId = b01niltree_id;
	ContObjFileId.fileRoot_gg00()          = operator[](ContFileNo).ofiRoot_bd900;
    ContObjFileId.fileObjFileNo_gg00()     = FileNo;
	ContObjFileId.fileContObjFileNo_gg00() = ContFileNo;
	ContObjFileId.fileRootCheck_gg00()     = ROOT_CHECK_GG00;
	ContObjFileId.fileTfn_gg00().becomes(tfnObj_egg00); 
    ContObjFileId.fileBdUse_gg00().addElement (bd_write_acc);
	ContObjFileId.fileType_gg00().delElement (ftsTemp_egg00);
    ContObjFileId.fileType_gg00().addElement (ftsObject_egg00);      
    ContObjFileId.fileType_gg00().addElement (ftsConcurrent_egg00); 
    ContObjFileId.fileType_gg00().addElement (ftsPerm_egg00);
	ContObjFileId.fileObjFileType_gg00().becomes (oftVarLenObjFile_egg00);
}

/*---------------------------------------------------------------------------*/
/* METHODS of cbd900_ObjKeyFileId                                            */
/*---------------------------------------------------------------------------*/

inline
cbd900_ObjKeyFileId:: cbd900_ObjKeyFileId (const cbd900_ObjFileInfo  &ObjFileInfo, 
										   const tsp00_Int4           FileNo)
//:tgg00_FileId (g01tabid.obj_hash) /* PTS 1105458 AK 26/01/2000 */ 
{
	memcpy (STATIC_CAST (tgg00_FileId *, this), &(g01tabid.obj_hash), sizeof (g01tabid.obj_hash));
    fileRoot_gg00()      = ObjFileInfo.ofiKeyRoot_bd900;
	fileObjFileNo_gg00() = FileNo;
}

/*---------------------------------------------------------------------------*/
/* METHODS of cbd900_RootAndSubRootPageInfo                                  */
/*---------------------------------------------------------------------------*/

inline
cbd900_RootAndSubRootPageInfo::cbd900_RootAndSubRootPageInfo (tbd_node_ptrs  &Nptrs)
{
  rsiRootAndSubRootPageDataPtr_bd900 = reinterpret_cast <cbd900_RootAndSubRootPageData *>  (&(Nptrs.np_ptr()->nd_body()));
}

/*---------------------------------------------------------------------------*/
/* METHODS of cbd900_RootPageInfo                                             */
/*---------------------------------------------------------------------------*/

inline
cbd900_RootPageInfo::cbd900_RootPageInfo(tbd_node_ptrs  &Nptrs)
:
cbd900_RootAndSubRootPageInfo (Nptrs)

{
#   if COMPILEMODE_MSP00 >= QUICK_MSP00 
	if (Nptrs.np_ptr()->nd_root() != Nptrs.np_ptr()->nd_id())
		g01abort (csp3_bd_msg, csp3_n_obj, "BD900RPI: no root       ", Nptrs.np_ptr()->nd_id());
#   endif
}

/*---------------------------------------------------------------------------*/

inline tsp00_PageNo &
cbd900_RootPageInfo::bd900ContRoot (const int ContFileNo)
{
	return rsiRootAndSubRootPageDataPtr_bd900->rsdRootPageData_bd900.rpdContRoot_bd900 [ContFileNo-1];
}

/*---------------------------------------------------------------------------*/

inline tsp00_Uint2 &        
cbd900_RootPageInfo::bd900NumChains()
{
   return rsiRootAndSubRootPageDataPtr_bd900->rsdRootPageData_bd900.rpdNumChains_bd900;
}
/*---------------------------------------------------------------------------*/

inline tsp00_Uint2 &        
cbd900_RootPageInfo::bd900NumContFiles()
{
   return rsiRootAndSubRootPageDataPtr_bd900->rsdRootPageData_bd900.rpdNumContFile_bd900;
}
/*---------------------------------------------------------------------------*/

inline tsp00_Bool &        
cbd900_RootPageInfo::bd900FileIsDeleted()
{
return rsiRootAndSubRootPageDataPtr_bd900->rsdRootPageData_bd900.rbdFileIsDeleted_bd900;
}

/*---------------------------------------------------------------------------*/


inline tsp00_PageNo &
cbd900_RootPageInfo::bd900_SubRoot (const int ChainNo)
{
	return rsiRootAndSubRootPageDataPtr_bd900->rsdRootPageData_bd900.rpdSubRoot_bd900 [ChainNo];
}

/*---------------------------------------------------------------------------*/


inline const tsp00_PageNo &
cbd900_RootPageInfo::bd900SubRoot (const int ChainNo)
{
	return bd900_SubRoot (ChainNo);
}

/*---------------------------------------------------------------------------*/

inline void
cbd900_RootPageInfo::bd900AddSubRoot (tsp00_PageNo SubRootPageNo)
{
	bd900_SubRoot (bd900NumChains()) = SubRootPageNo;
    ++bd900NumChains();
}

/*---------------------------------------------------------------------------*/

inline void 
cbd900_RootPageInfo::bd900InitRootPageInfo()
{
	bd900FileIsDeleted()                                                       = false;
	bd900NumChains()                                                           = 0;
	bd900NumContFiles()                                                        = 0;   
	rsiRootAndSubRootPageDataPtr_bd900->rsdRootPageData_bd900.rpdFiller1_bd900 = false;
	rsiRootAndSubRootPageDataPtr_bd900->rsdRootPageData_bd900.rpdFiller2_bd900 = 0;
	
	for (int iChain=0; iChain<MAX_NUM_CHAINS_PER_CONTAINER_BD900; ++iChain)           
		bd900_SubRoot(iChain) = NIL_PAGE_NO_GG00;
	
	for (int iContFile=0; iContFile<MAX_NUM_CONT_OBJ_FILES_BD900; ++iContFile)           
		bd900ContRoot(iContFile) = NIL_PAGE_NO_GG00;
}

/*---------------------------------------------------------------------------*/
/* METHODS of cbd900_SubRootPageInfo                                         */
/*---------------------------------------------------------------------------*/

inline
cbd900_SubRootPageInfo::cbd900_SubRootPageInfo(tbd_node_ptrs  &Nptrs)
:
cbd900_RootAndSubRootPageInfo (Nptrs)
{
#   if COMPILEMODE_MSP00 >= QUICK_MSP00 
	if (Nptrs.np_ptr()->nd_sub_root() != Nptrs.np_ptr()->nd_id())
		g01abort (csp3_bd_msg, csp3_n_obj, "BD900RPI: no sub root   ", Nptrs.np_ptr()->nd_id());
#   endif
}

/*---------------------------------------------------------------------------*/

inline tsp00_PageNo &
cbd900_SubRootPageInfo::bd900AllPageCnt ()
{
	return rsiRootAndSubRootPageDataPtr_bd900->rsdSubRootPageData_bd900.spdAllPageCnt_bd900;
}

/*---------------------------------------------------------------------------*/

inline tsp00_PageNo &
cbd900_SubRootPageInfo::bd900EmptyPageCnt ()
{
	return rsiRootAndSubRootPageDataPtr_bd900->rsdSubRootPageData_bd900.spdEmptyPageCnt_bd900;
}

/*---------------------------------------------------------------------------*/

inline tsp00_PageNo &
cbd900_SubRootPageInfo::bd900FreePageCnt ()
{
	return rsiRootAndSubRootPageDataPtr_bd900->rsdSubRootPageData_bd900.spdFreePageCnt_bd900;
}

/*---------------------------------------------------------------------------*/

inline tsp00_PageNo &
cbd900_SubRootPageInfo::bd900EmptyPageLimit ()
{
	return rsiRootAndSubRootPageDataPtr_bd900->rsdSubRootPageData_bd900.spdEmptyPageLimit_bd900;
}

/*---------------------------------------------------------------------------*/

inline tsp00_PageNo &
cbd900_SubRootPageInfo::bd900MaxPageSeqNo ()
{
	return rsiRootAndSubRootPageDataPtr_bd900->rsdSubRootPageData_bd900.spdMaxPageSeqNo_bd900;
}

/*---------------------------------------------------------------------------*/

inline tsp00_PageNo &
cbd900_SubRootPageInfo::bd900AllObjCnt ()
{
	return rsiRootAndSubRootPageDataPtr_bd900->rsdSubRootPageData_bd900.spdAllObjCnt_bd900;
}

/*---------------------------------------------------------------------------*/

inline void
cbd900_SubRootPageInfo::bd900InitSubRootPageInfo()
{
	bd900AllPageCnt()       = 1;
	bd900FreePageCnt()      = 0;
	bd900EmptyPageCnt()     = 0;
	bd900EmptyPageLimit()   = MAX_PORTION_OF_EMPTY_PAGES_BD900;
	bd900MaxPageSeqNo()     = 0;
	bd900AllObjCnt()        = 0; 
}

/*---------------------------------------------------------------------------*/


#endif  /* GBD900_H */
