/****************************************************************************

  module      : gbd910.h

  -------------------------------------------------------------------------

  responsible : AlexanderK

  special area: object node classes
  description : description ...


  version     : 7.2.0.0
  last changed: 1999-07-27  14:47
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1998



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/

#ifndef GBD910_H
#define GBD910_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/
 
#include "ggg92.h"
#include "gbd00.h"
#include "gbd900.h"

#include "hbd13.h"
#include "hbd20.h"
#include "hbd92.h"
#include "hgg01_1.h" 
#include "gsp03.h" 
#include "gsp03_3.h" 

#if COMPILEMODE_MEO00 >= SLOW_MEO00 
#include "hta99.h"   
#include "hta01.h" 
#endif

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#ifndef TRUE
 #define TRUE 1
#endif

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/

extern cbd900_ObjFileDir      bd90ObjFileDir;
extern cbd900_ContObjFileDir  bd93ContObjFileDir;

/*===========================================================================*
 *  FORWARD DECLARATUIN                                                      *
 *===========================================================================*/

class cbd910_ObjNode;
class cbd910_ObjDataNode;
class cbd910_RootNode; 
class cbd910_SubRootNode;
class cbd910_ContObjDataNode;
class cbd910_PrimObjDataNode;

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

class cbd910_ObjNode : public tbd_node_ptrs
{
public:
    cbd910_ObjNode (tbd_current_tree       &Current,
	                tbd_node_request_Enum   NodeRequest);
    cbd910_ObjNode (tbd_current_tree       &Current, 
		            tsp00_PageNo            PageNo,
	                tbd_node_request_Enum   NodeRequest,
		            tbd00_PageLockMode_Enum PageLockMode = plmLock_ebd00);
	~cbd910_ObjNode();
	void bd910ReleasePage();
	void bd910ReleasePageWithoutCheck();
	void bd910AssignToPage (tsp00_PageNo            PageNo, 
		                    tbd00_PageLockMode_Enum  PageLockMode = plmLock_ebd00);
	void bd910FreePage();
	void bd910MoveToNextPage();
	void bd910AssignToNewPage (tsp00_Int4          ObjBodyLen,
		                       tsp00_Int4          ObjKeyLen,
		                       tsp00_PageNo        SubRoot,
		                       tsp00_Int4          ChainNo,
	                           tsp00_Int4          PageSeqNo);
	bool bd910PageIsAvailable();
	bool bd910PageIsRequested();
	void bd910GetAnotherObjPage (tsp00_PageNo                  NewPageNo,
	                             const tbd00_LockRequestSeqNo  NewPageSeqNo,
		                         cbd910_ObjNode               &NptrsNewPage);
	
	void bd910SetTo (const cbd910_ObjNode & ObjNode);

private:   // methods
	virtual void bd910_CheckNodeType ();
	void         bd910_InitNode (tbd_current_tree      &Current,
		                         tbd_node_request_Enum  NodeRequest);

	void   bd910InitObjPage (tsp00_Int4      ObjBodyLen,
		                     tsp00_Int4      ObjKeyLen,
		                     tsp00_PageNo    SubRoot,
		                     tsp00_Int4      ChainNo,
	                         tsp00_Int4      PageSeqNo);

protected: // members
	tbd_current_tree *m_ptrCurrent;

private:   //members
	tbd_node_request  m_NodeRequest;
	bool              m_PrimaryReference;
};

/*---------------------------------------------------------------------------*/

class cbd910_RootNode : public cbd910_ObjNode
{
public:
    cbd910_RootNode (tbd_current_tree        &Current,
		             tbd_node_request_Enum    NodeRequest);
    cbd910_RootNode (tbd_current_tree        &Current, 
                     tsp00_PageNo             PageNo,
		             tbd_node_request_Enum    NodeRequest,
		             tbd00_PageLockMode_Enum  PageLockMode = plmLock_ebd00);
private:
	virtual void bd910_CheckNodeType ();
};

/*---------------------------------------------------------------------------*/

class cbd910_SubRootNode : public cbd910_ObjNode
{
public:
    cbd910_SubRootNode (tbd_current_tree        &Current,
		                tbd_node_request_Enum     NodeRequest);
    cbd910_SubRootNode (tbd_current_tree        &Current, 
                        tsp00_PageNo             PageNo,
		                tbd_node_request_Enum     NodeRequest,
		                tbd00_PageLockMode_Enum   PageLockMode = plmLock_ebd00);

	bool    bd910ChainContainsTooManyEmptyPages();

private:
	virtual void bd910_CheckNodeType ();
};

/*---------------------------------------------------------------------------*/

class cbd910_ObjDataNode : public cbd910_ObjNode
{
public:
    cbd910_ObjDataNode (tbd_current_tree        &Current,
		                tbd_node_request_Enum    NodeRequest);
    cbd910_ObjDataNode (tbd_current_tree        &Current, 
	                    tsp00_PageNo             PageNo,
		                tbd_node_request_Enum    NodeRequest,
		                tbd00_PageLockMode_Enum  PageLockMode = plmLock_ebd00);

	void                    bd910AssignToNewPageInDataChain (cbd910_SubRootNode      &NptrsSubRoot);
    bool                    bd910RemovePageFromDataChain    (cbd910_SubRootNode      &NptrsSubRoot,
							                                 tsp00_PageNo             PrevFreePage,
							                                 tbd00_LockRequestSeqNo   PrevFreePageSeqNo);
	void                    bd910InsertPageIntoFreeChain    (cbd910_SubRootNode &NptrsSubRoot);
	void                    bd910IncreaseEmptyPageCount     (cbd910_SubRootNode &NptrsSubRoot);
	bool                    bd910RemovePageFromFreeChain    (cbd910_ObjNode     &NptrsPrevPage,
						            					     cbd910_SubRootNode &NptrsSubRoot);
	bool                    bd910PageMustBeInsertedIntoFreeChain ();
	bool                    bd910EmptyPageCountMustBeIncreased   ();
	tbd900_ObjFrame        *bd910ObjFramePtr (tsp00_Int4  ObjFramePos);
	static tbd900_ObjFrame *bd910ObjFramePtr (tsp00_Byte *pPageBodyBeg, tsp00_Int4 ObjFramePos);

private:
	void   bd910_SetAllPageCntInChainInfo (const tgg00_FileId  &FileId,
		        					       tsp00_Int4           ChainNo,
								           tsp00_Int4           AllPageCnt);

	void   bd910_CheckFreePageCnt (cbd910_SubRootNode &NptrsSubRoot);
};

/*---------------------------------------------------------------------------*/

class cbd910_ContObjDataNode : public cbd910_ObjDataNode
{
public:
    cbd910_ContObjDataNode (tbd_current_tree        &Current,
		                    tbd_node_request_Enum     NodeRequest);
    cbd910_ContObjDataNode (tbd_current_tree        &Current, 
                            tsp00_PageNo             PageNo,
		                    tbd_node_request_Enum    NodeRequest,
		                    tbd00_PageLockMode_Enum  PageLockMode = plmLock_ebd00);

private:
	virtual void bd910_CheckNodeType ();
};

/*---------------------------------------------------------------------------*/

class cbd910_PrimObjDataNode : public cbd910_ObjDataNode
{
public:
    cbd910_PrimObjDataNode (tbd_current_tree        &Current,
    		                tbd_node_request_Enum    NodeRequest);
    cbd910_PrimObjDataNode (tbd_current_tree        &Current, 
                            tsp00_PageNo             PageNo,
		                    tbd_node_request_Enum    NodeRequest,
		                    tbd00_PageLockMode_Enum  PageLockMode = plmLock_ebd00);

	void         bd910InsertPageIntoFreeChainAndUpdateEmptyPageCount (cbd900_ObjFileInfo &ObjFileInfo);

private:
	virtual void bd910_CheckNodeType ();
};


/*===========================================================================*
 *  DEFINITION OF METHODS                                                    *
 *===========================================================================*/

/*---------------------------------------------------------------------------*/
/* METHODS of cbd910_ObjNode                                                 */
/*---------------------------------------------------------------------------*/

inline void
cbd910_ObjNode::bd910AssignToPage(tsp00_PageNo              PageNo, 
								  tbd00_PageLockMode_Enum   PageLockMode)
{
#   if COMPILEMODE_MSP00 >= QUICK_MSP00 
	if (np_ptr() != NULL)
		g01abort (csp3_bd_msg, csp3_n_obj, "BD910Assign: Nptr!=NULL ", np_ptr()->nd_id());
#   endif	
	
	bd13GetNode (*m_ptrCurrent, PageNo, PageLockMode, m_NodeRequest, *this); 
	
	m_PrimaryReference = true;

	if ((oftUnknownType_egg00 == m_ptrCurrent->curr_tree_id.fileObjFileType_gg00()) &&
		(e_ok                 == m_ptrCurrent->curr_trans->trError_gg00))
	{

#       if COMPILEMODE_MSP00 >= QUICK_MSP00 
		if (pt2ContObject_egg00 == this->np_ptr()->nd_pt2())
			g01abort (csp3_bd_msg, csp3_n_obj, "BD910Assign: Wrong Type ", np_ptr()->nd_pt2());
#       endif
		
		m_ptrCurrent->curr_tree_id.fileObjFileNo_gg00() = this->np_ptr()->nd_obj_file_no();
        m_ptrCurrent->curr_tree_id.fileRoot_gg00()      = this->np_ptr()->nd_root();
		if (pt2Object_egg00 == this->np_ptr()->nd_pt2())
			m_ptrCurrent->curr_tree_id.fileObjFileType_gg00().becomes(oftFixLenObjFile_egg00);
		else
            m_ptrCurrent->curr_tree_id.fileObjFileType_gg00().becomes(oftVarLenObjFile_egg00);
	}	

#   if COMPILEMODE_MSP00 >= QUICK_MSP00 
	if (e_ok == m_ptrCurrent->curr_trans->trError_gg00)
		bd910_CheckNodeType();
#   endif
}

/*---------------------------------------------------------------------------*/

inline void 
cbd910_ObjNode::bd910_InitNode (tbd_current_tree      &Current,
			         			tbd_node_request_Enum  NodeRequest)
{
    np_ptr()      = NULL;
    np_cbptr()    = NULL;
    m_ptrCurrent  = &Current;
    m_NodeRequest.becomes(NodeRequest);
	m_PrimaryReference = true;
}

/*---------------------------------------------------------------------------*/

inline 
cbd910_ObjNode::cbd910_ObjNode (tbd_current_tree    &Current,
			        			tbd_node_request_Enum  NodeRequest)
{
	bd910_InitNode (Current, NodeRequest);
}

/*---------------------------------------------------------------------------*/

inline 
cbd910_ObjNode::cbd910_ObjNode (tbd_current_tree        &Current, 
								tsp00_PageNo             PageNo,
								tbd_node_request_Enum    NodeRequest,
								tbd00_PageLockMode_Enum  PageLockMode)
{
    bd910_InitNode     (Current, NodeRequest);
	bd910AssignToPage  (PageNo,  PageLockMode);
}

/*---------------------------------------------------------------------------*/

inline void
cbd910_ObjNode::bd910ReleasePage()
{ 
#   if COMPILEMODE_MEO00 >= SLOW_MEO00 
	if ((NULL != np_ptr())  &&
		m_PrimaryReference  &&
		(e_ok == m_ptrCurrent->curr_trans->trError_gg00))
	{
		bd92VerifyObjPage (*m_ptrCurrent, *this);
	}
#   endif
				
	bd910ReleasePageWithoutCheck();
}

/*---------------------------------------------------------------------------*/

inline void
cbd910_ObjNode::bd910ReleasePageWithoutCheck()
{
	if (NULL != np_ptr())
    { 
		if (m_PrimaryReference)
		{		
			if ((nr_for_update == m_NodeRequest) && (e_ok == m_ptrCurrent->curr_trans->trError_gg00)) 
				b13w_release_node (*this, *m_ptrCurrent);
			else  // m_NodeRequest == nr_for_read 
				b13r_release_node (*this, *m_ptrCurrent, lru_normal);
		}
		else
		{
			np_ptr()      = NULL;
			np_cbptr()    = NULL;
		}
    }
}

/*---------------------------------------------------------------------------*/

inline
cbd910_ObjNode::~cbd910_ObjNode() 
{     
    bd910ReleasePage();
}

/*---------------------------------------------------------------------------*/

inline void 
cbd910_ObjNode::bd910FreePage()
{
#   if COMPILEMODE_MSP00 >= QUICK_MSP00 
	if (nr_for_update != m_NodeRequest)
		g01abort (csp3_bd_msg, csp3_n_obj, "BD910FreePage: wrong Req", np_ptr()->nd_id());
	if (! m_PrimaryReference)
		g01abort (csp3_bd_msg, csp3_n_obj, "BD910FreePage: NoPrimRef", np_ptr()->nd_id());
#   endif
	
	b13free_node (*this, *m_ptrCurrent);
}    

/*---------------------------------------------------------------------------*/

inline void
cbd910_ObjNode::bd910MoveToNextPage()
{                       
	/* get pointer to next page while keeping the current  */
	/* page to ensure that the nd_right pointer is valid   */
	if (NULL == np_ptr())
		return;
	else
	{
		const tsp00_PageNo  RightPageNo = np_ptr()->nd_right();
		cbd910_ObjNode      NptrsPrevPage(*this);     
		
		/* reset page pointer */
		np_ptr()   = NULL;
		np_cbptr() = NULL;
		
		if (NIL_PAGE_NO_GG00 != RightPageNo)
			bd910AssignToPage (RightPageNo);
		else
		{
			/* no more pages in this chain, goto next chain of this file */
            const tgg00_ObjFileNo &CurrFileNo  = m_ptrCurrent->curr_tree_id.fileObjFileNo_gg00();
			const tsp00_PageNo    &CurrChainNo = NptrsPrevPage.np_ptr()->nd_chain_no();
			cbd900_ObjFileInfo    &ObjFileInfo = bd90ObjFileDir[CurrFileNo];
			
			if (ObjFileInfo.ofiMaxChain_bd900 - 1 <= CurrChainNo)
				return;
			else
			{   /* goto sub root of next chain */
				const tsp00_PageNo  NextSubRoot = ObjFileInfo[CurrChainNo+1].ociSubRoot_bd900;
				bd910AssignToPage (NextSubRoot);
				NptrsPrevPage.bd910ReleasePage();
                // bd910MoveToNextPage();  PTS 1107145 AK 10/07/2000
			}
		}
		
		/* check that the page moved to has a lower LockRequestSeqNo */
#       if COMPILEMODE_MSP00 >= QUICK_MSP00 
		if ((np_ptr() != NULL) && (np_ptr()->nd_id() != np_ptr()->nd_sub_root())) /* PTS 1107145 AK 10/07/2000*/
		{
			if (this->np_ptr()->ndLockRequestSeqNo_bd00() >= NptrsPrevPage.np_ptr()->ndLockRequestSeqNo_bd00() )
			{
				g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, "BD910MoveTo: wrong SeqNo",  
					NptrsPrevPage.np_ptr()->nd_id());
				g01abort (csp3_bd_msg, csp3_n_obj, "BD910MoveTo: wrong SeqNo", np_ptr()->nd_id());
			}
		}
#       endif	
	}
}

/*---------------------------------------------------------------------------*/

inline bool 
cbd910_ObjNode::bd910PageIsAvailable()
{
	return (NULL != this->np_ptr());
};

/*---------------------------------------------------------------------------*/

inline bool
cbd910_ObjNode::bd910PageIsRequested()
{
	return (TRUE == bd20IsPageRequested(this->np_cbptr()));
}


/*---------------------------------------------------------------------------*/

inline void
cbd910_ObjNode::bd910_CheckNodeType ()
{
	if ((this->np_ptr()->nd_pt2() != pt2ContObject_egg00)         &&
		(this->np_ptr()->nd_pt2() != pt2Object_egg00)             &&
		(this->np_ptr()->nd_pt2() != pt2VarObject_egg00))
	{
    	g01abort (csp3_bd_msg, csp3_n_obj, "BD910_Check: No ObjNode ", np_ptr()->nd_id());
	}
}


/*---------------------------------------------------------------------------*/

inline	void
cbd910_ObjNode::bd910SetTo (const cbd910_ObjNode & ObjNode)
{
#   if COMPILEMODE_MSP00 >= QUICK_MSP00 
	if (np_ptr() != NULL)
		g01abort (csp3_bd_msg, csp3_n_obj, "BD910SetTo: Nptr!=NULL  ", np_ptr()->nd_id());
#   endif

	*this              = ObjNode;
	m_PrimaryReference = false;
}

/*---------------------------------------------------------------------------*/
/* METHODS of cbd910_ObjDataNode                                             */
/*---------------------------------------------------------------------------*/

inline
cbd910_ObjDataNode::cbd910_ObjDataNode (tbd_current_tree        &Current,
										tbd_node_request_Enum    NodeRequest)
:cbd910_ObjNode (Current, NodeRequest)
{}

/*---------------------------------------------------------------------------*/

inline
cbd910_ObjDataNode::cbd910_ObjDataNode (tbd_current_tree        &Current, 
										tsp00_PageNo             PageNo,
										tbd_node_request_Enum    NodeRequest,
										tbd00_PageLockMode_Enum  PageLockMode)
:cbd910_ObjNode (Current, PageNo, NodeRequest, PageLockMode)
{}

/*---------------------------------------------------------------------------*/

inline void
cbd910_ObjNode::bd910AssignToNewPage (tsp00_Int4      ObjBodyLen,
									  tsp00_Int4      ObjKeyLen,
									  tsp00_PageNo    SubRoot,
									  tsp00_Int4      ChainNo,
									  tsp00_Int4      PageSeqNo)
{
#   if COMPILEMODE_MSP00 >= QUICK_MSP00 
	if (np_ptr() != NULL)
		g01abort (csp3_bd_msg, csp3_n_obj, "BD910NewPage: Nptr!=NULL", np_ptr()->nd_id());
	if (nr_for_update != m_NodeRequest)
		g01abort (csp3_bd_msg, csp3_n_obj, "BD910NewPage: wrong Req ", np_ptr()->nd_id());
#   endif
	
	b13new_object_node (*this, *m_ptrCurrent);

	m_PrimaryReference = true;

	this->bd910InitObjPage (ObjBodyLen, ObjKeyLen, SubRoot, ChainNo, PageSeqNo);
	
#   if COMPILEMODE_MSP00 >= QUICK_MSP00 
	bd910_CheckNodeType();
#   endif
}

/*---------------------------------------------------------------------------*/

inline void  
cbd910_ObjDataNode::bd910_SetAllPageCntInChainInfo (const tgg00_FileId  &FileId,
													tsp00_Int4           ChainNo,
													tsp00_Int4           AllPageCnt)
{
	ROUTINE_DBG_MEO00 ("bd910_SetAllPageCntInChainInfo");
	
	const tgg00_ObjFileNo &FileNo     = FileId.fileObjFileNo_gg00();
    const tsp00_Int4       ContFileNo = FileId.fileContObjFileNo_gg00();
	
	if (PRIM_CONT_OBJ_FILE_NO_BD900 == ContFileNo)
		bd90ObjFileDir[FileNo][ChainNo].ociAllPageCnt_bd900 = AllPageCnt;
	else
		bd93ContObjFileDir[FileNo][ContFileNo][ChainNo].ociAllPageCnt_bd900 = AllPageCnt; 
}

/*---------------------------------------------------------------------------*/

inline void
cbd910_ObjDataNode::bd910_CheckFreePageCnt (cbd910_SubRootNode &NptrsSubRoot)
{
	cbd900_SubRootPageInfo  SubRootPageInfo (NptrsSubRoot);

	int MinFreePageCnt = 0;
	int MaxFreePageCnt = 0;
	
	if (NptrsSubRoot.np_ptr()->nd_in_free_chain())
		++MinFreePageCnt;
	
	if (NptrsSubRoot.np_ptr()->nd_next_free() == NIL_PAGE_NO_GG00)
		MaxFreePageCnt = MinFreePageCnt;
	else
	{
		++MinFreePageCnt;
		if (NptrsSubRoot.np_ptr()->nd_id() != np_ptr()->nd_id())
		{
			if (NptrsSubRoot.np_ptr()->nd_next_free() == np_ptr()->nd_id())
			{
				if (np_ptr()->nd_next_free() == NIL_PAGE_NO_GG00)
					MaxFreePageCnt = MinFreePageCnt;
				else
					++MinFreePageCnt;
			}
			else
			{ 
				if (np_ptr()->nd_in_free_chain())
					++MinFreePageCnt;
				
				if (np_ptr()->nd_next_free() != NIL_PAGE_NO_GG00)
					++MinFreePageCnt;
			}
		}
	}
	
	if ((MinFreePageCnt > SubRootPageInfo.bd900FreePageCnt()) ||
		((MaxFreePageCnt !=0) && (MaxFreePageCnt < SubRootPageInfo.bd900FreePageCnt())))
	{
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj,
			"BD910InsFree:Odd FreeCnt",  	np_ptr()->nd_id());
	}
}

/*---------------------------------------------------------------------------*/

inline void  
cbd910_ObjDataNode::bd910InsertPageIntoFreeChain  (cbd910_SubRootNode &NptrsSubRoot)
{
    ROUTINE_DBG_MEO00 ("bd910InsertPageIntoFreeChain");
	
	/* insert page into free page chain */
	if (NptrsSubRoot.np_ptr()->nd_id() != np_ptr()->nd_id())
	{
		np_ptr()->nd_next_free()              = NptrsSubRoot.np_ptr()->nd_next_free();
		NptrsSubRoot.np_ptr()->nd_next_free() = np_ptr()->nd_id();
	}
	np_ptr()->nd_in_free_chain()          = true;
	
	/* increase free page counter on subroot page */ 
	cbd900_SubRootPageInfo  SubRootPageInfo (NptrsSubRoot);
	++SubRootPageInfo.bd900FreePageCnt();
	
	/* perform some simple checks */
#   if COMPILEMODE_MEO00 >= QUICK_MEO00 	
    bd910_CheckFreePageCnt (NptrsSubRoot);
#   endif

	/* write a vtrace entry */		
#   if COMPILEMODE_MEO00 >= SLOW_MEO00 
    t01name (bi,"ins into free list");
	t01int4 (bi,"ins page no=",np_ptr()->nd_id()); 
#   endif  
};

/*---------------------------------------------------------------------------*/

inline bool 
cbd910_ObjDataNode::bd910RemovePageFromFreeChain (cbd910_ObjNode       &NptrsPrevPage,
												  cbd910_SubRootNode   &NptrsSubRoot)
{
    ROUTINE_DBG_MEO00 ("bd910RemovePageFromFreeChain");
	
	/* remove page from free page chain */
	if (NptrsSubRoot.np_ptr()->nd_id() == np_ptr()->nd_id())
	{
		np_ptr()->nd_in_free_chain()           = false;

		/* decrease free page counter on subroot page */ 
		cbd900_SubRootPageInfo  SubRootPageInfo (NptrsSubRoot);
		--SubRootPageInfo.bd900FreePageCnt();
	}
	else if (NptrsPrevPage.np_ptr()->nd_next_free() == np_ptr()->nd_id())
	{
		NptrsPrevPage.np_ptr()->nd_next_free() = np_ptr()->nd_next_free();
		np_ptr()->nd_next_free()               = NIL_PAGE_NO_GG00;
		np_ptr()->nd_in_free_chain()           = false;
		
	   /* decrease free page counter on subroot page */ 
		cbd900_SubRootPageInfo  SubRootPageInfo (NptrsSubRoot);
		--SubRootPageInfo.bd900FreePageCnt();
	}

	/* perform some simple checks */
#   if COMPILEMODE_MEO00 >= QUICK_MEO00 	
    bd910_CheckFreePageCnt (NptrsSubRoot);
#   endif

	return (!np_ptr()->nd_in_free_chain());
};

/*---------------------------------------------------------------------------*/

inline void  
cbd910_ObjDataNode::bd910IncreaseEmptyPageCount (cbd910_SubRootNode &NptrsSubRoot)
{
    ROUTINE_DBG_MEO00 ("bd910IncreaseEmptyPageCount");
	
	/* increase empty page counter on subroot page */ 
	cbd900_SubRootPageInfo SubRootPageInfo (NptrsSubRoot);
	++SubRootPageInfo.bd900EmptyPageCnt();
	this->np_ptr()->nd_empty_page() = true;  
};

/*---------------------------------------------------------------------------*/

inline tbd900_ObjFrame*
cbd910_ObjDataNode::bd910ObjFramePtr (tsp00_Byte *pPageBegin,
									  tsp00_Int4  ObjFramePos)
{
	return ((tbd900_ObjFrame *) (pPageBegin + ObjFramePos));   
}

/*---------------------------------------------------------------------------*/

inline tbd900_ObjFrame*
cbd910_ObjDataNode::bd910ObjFramePtr (tsp00_Int4 ObjFramePos)
{
	return bd910ObjFramePtr ((tsp00_Byte *) np_ptr(),ObjFramePos);
}


/*---------------------------------------------------------------------------*/

inline	bool                    
cbd910_ObjDataNode::bd910PageMustBeInsertedIntoFreeChain ()
{
  return ((this->np_ptr()->nd_free_obj_cnt() > 0) && 
	     !this->np_ptr()->nd_in_free_chain());
}

/*---------------------------------------------------------------------------*/
	
inline	bool                    
cbd910_ObjDataNode::bd910EmptyPageCountMustBeIncreased ()
{
  return ((this->np_ptr()->nd_free_obj_cnt() == this->np_ptr()->nd_max_obj_cnt()) &&
	       !this->np_ptr()->nd_empty_page()); 
}

/*---------------------------------------------------------------------------*/
/* METHODS of cbd910_ContObjDataNode                                         */
/*---------------------------------------------------------------------------*/

inline
cbd910_ContObjDataNode::cbd910_ContObjDataNode (tbd_current_tree        &Current,
												tbd_node_request_Enum    NodeRequest)
:cbd910_ObjDataNode (Current, NodeRequest)
{}

/*---------------------------------------------------------------------------*/

inline
cbd910_ContObjDataNode::cbd910_ContObjDataNode (tbd_current_tree        &Current, 
												tsp00_PageNo             PageNo,
												tbd_node_request_Enum    NodeRequest,
												tbd00_PageLockMode_Enum  PageLockMode)
:cbd910_ObjDataNode (Current, PageNo, NodeRequest, PageLockMode)
{}

/*---------------------------------------------------------------------------*/

inline void
cbd910_ContObjDataNode::bd910_CheckNodeType ()
{
	if (this->np_ptr()->nd_pt2() != pt2ContObject_egg00)   
	{
    	g01abort (csp3_bd_msg, csp3_n_obj, "BD910_Check: No ContNode", np_ptr()->nd_id());
	}
}

/*---------------------------------------------------------------------------*/
/* METHODS of cbd910_PrimObjDataNode                                             */
/*---------------------------------------------------------------------------*/

inline
cbd910_PrimObjDataNode::cbd910_PrimObjDataNode (tbd_current_tree        &Current,
										tbd_node_request_Enum    NodeRequest)
:cbd910_ObjDataNode (Current, NodeRequest)
{}

/*---------------------------------------------------------------------------*/

inline
cbd910_PrimObjDataNode::cbd910_PrimObjDataNode (tbd_current_tree        &Current, 
										tsp00_PageNo             PageNo,
										tbd_node_request_Enum    NodeRequest,
										tbd00_PageLockMode_Enum  PageLockMode)
:cbd910_ObjDataNode (Current, PageNo, NodeRequest, PageLockMode)
{}

/*---------------------------------------------------------------------------*/

inline void
cbd910_PrimObjDataNode::bd910_CheckNodeType ()
{
	if ((this->np_ptr()->nd_pt2() != pt2Object_egg00)  &&
		(this->np_ptr()->nd_pt2() != pt2VarObject_egg00))
	{
    	g01abort (csp3_bd_msg, csp3_n_obj, "BD910_Check: No PObjNode", np_ptr()->nd_id());
	}
}

/*---------------------------------------------------------------------------*/
/* METHODS of cbd910_SubRootNode                                         */
/*---------------------------------------------------------------------------*/

inline
cbd910_SubRootNode::cbd910_SubRootNode (tbd_current_tree        &Current,
										tbd_node_request_Enum    NodeRequest)
:cbd910_ObjNode (Current, NodeRequest)
{}

/*---------------------------------------------------------------------------*/

inline 
cbd910_SubRootNode::cbd910_SubRootNode (tbd_current_tree        &Current, 
										tsp00_PageNo             PageNo,
										tbd_node_request_Enum    NodeRequest,
										tbd00_PageLockMode_Enum  PageLockMode)
:cbd910_ObjNode (Current, PageNo, NodeRequest, PageLockMode)
{}

/*---------------------------------------------------------------------------*/

inline bool 
cbd910_SubRootNode::bd910ChainContainsTooManyEmptyPages()
{
    ROUTINE_DBG_MEO00 ("bd910ChainContainsTooManyEmptyPages");    
	

	cbd900_SubRootPageInfo  SubRootPageInfo (*this);

    tsp00_Int4   PortionOfEmptyPages = 
		100 * SubRootPageInfo.bd900EmptyPageCnt()/(SubRootPageInfo.bd900AllPageCnt() + 1);
    
    return ((PortionOfEmptyPages > SubRootPageInfo.bd900EmptyPageLimit()) &&
		(SubRootPageInfo.bd900EmptyPageCnt() > MAX_EMPTY_PAGES_BD900));
}
	
/*---------------------------------------------------------------------------*/

inline void
cbd910_SubRootNode::bd910_CheckNodeType ()
{
	if ((this->np_ptr()->nd_id()   != this->np_ptr()->nd_sub_root())  ||
        ((this->np_ptr()->nd_pt2() != pt2ContObject_egg00)         &&
		 (this->np_ptr()->nd_pt2() != pt2Object_egg00)             &&
		 (this->np_ptr()->nd_pt2() != pt2VarObject_egg00)))
	{
    	g01abort (csp3_bd_msg, csp3_n_obj, "BD910_Check: No SubRoot ", np_ptr()->nd_id());
	}
}

/*---------------------------------------------------------------------------*/
/* METHODS of cbd910_RootNode                                                */
/*---------------------------------------------------------------------------*/

inline
cbd910_RootNode::cbd910_RootNode (tbd_current_tree        &Current,
								  tbd_node_request_Enum    NodeRequest)
:cbd910_ObjNode (Current, NodeRequest)
{}

/*---------------------------------------------------------------------------*/

inline 
cbd910_RootNode::cbd910_RootNode (tbd_current_tree        &Current, 
								  tsp00_PageNo             PageNo,
								  tbd_node_request_Enum    NodeRequest,
								  tbd00_PageLockMode_Enum  PageLockMode)
:cbd910_ObjNode (Current, PageNo, NodeRequest, PageLockMode)
{}

/*---------------------------------------------------------------------------*/

inline void
cbd910_RootNode::bd910_CheckNodeType ()
{
	if ((this->np_ptr()->nd_id()   != this->np_ptr()->nd_root())  ||
        ((this->np_ptr()->nd_pt2() != pt2ContObject_egg00)         &&
		 (this->np_ptr()->nd_pt2() != pt2Object_egg00)             &&
		 (this->np_ptr()->nd_pt2() != pt2VarObject_egg00)))
	{
    	g01abort (csp3_bd_msg, csp3_n_obj, "BD910_Check: No    Root ", np_ptr()->nd_id());
	}
}

/*---------------------------------------------------------------------------*/


#endif  /* GBD910_H */
