/****************************************************************************

  module      : gbd900.h

  -------------------------------------------------------------------------

  responsible : AlexanderK

  special area: "BD_Object_Constants_And_Types"
  description : for the implementation of the non inline methods see vbd900.cpp


  version     : 7.2.0.2
  last changed: 2000-20-01  11:01
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1998



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/

#ifndef GBD930_H
#define GBD930_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include <new.h>

#include "gsp03.h"
#include "gsp03_3.h"
#include "ggg12.h"
#include "gbd94.h"   //bd94TruncateHistory
#include "hbd12.h"   //vtrace routine b12vtrace
#include "hgg01.h"   //get cserv parameter g01maxuser
#include "hgg01_1.h" //g01abort,g01opmsg 
#include "heo55k.h"  //vbegexcl
#include "heo56k.h"  //vsuspend,vresume
#include "heo57.h"   //vallocat
#include "hkb28.h"   //k28emergency_shutdown

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define NIL_HISTORY_FILE_INDEX_BD930  -1

/* local defines uses in this file only */
#define GC_IS_STARTED                 true
#define GC_IS_STOPPED                 false

#if COMPILEMODE_MEO00 >= QUICK_MEO00 
#  define BD930_CHECK_GC  bd930_CheckGC()
#else 
#  define BD930_CHECK_GC
#endif

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  FORWARD DECLARATIONS  OF CLASSES, STRUCTURES, TYPES, UNIONS              *
 *===========================================================================*/

class cbd930_GarbageCollector;
class cbd930_GCState;
class cbd930_GarbCollController;
class cbd930_HistoryFileInfo;
class cbd930_HistoryFileDir;
class cbd930_HistoryFileIter;

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *==========================================================f=================*/

class cbd930_HistoryFileInfo
{
public: //members
	tbd94_HistoryFileHandle hfiHistoryFileHandle_bd930;
	tsp00_Int2              hfiCurrActivGC_bd930;
	tsp00_Bool              hfiHistoryGetsTruncated_bd930;
	tsp00_Bool              hfiFiller1_bd930;
#ifdef BIT64
	tsp00_Int4              hfiFiller2_bd930;
#endif

public: //methods
	     cbd930_HistoryFileInfo   ();
	bool bd930ClearHistoryFile    (tgg00_TransContext &Trans);
	void bd930TruncateHistoryFile (tgg00_TransContext &Trans);

private:
    /* this check is necessary for vallocat */
#  ifdef BIT64
    CHECK_MAPPABLE_CLASS_GIVEN_SIZE_SP00 (cbd930_HistoryFileInfo, 16)
#  else
    CHECK_MAPPABLE_CLASS_GIVEN_SIZE_SP00 (cbd930_HistoryFileInfo,  8)
#  endif
};

/*---------------------------------------------------------------------------*/

class cbd930_HistoryFileDir
{
	friend class cbd930_HistoryFileIter;

public: // methods
         cbd930_HistoryFileDir   ();
	void bd930Init               ();
	void bd930AddHistoryFile     (tgg00_TransContext        &Trans, 
		                          const tbd94_HistoryFileHandle &HistoryFileHandle);
	void bd930DumpHistoryFileDir (tgg00_VfFileref     &HostFile,
								  tsp00_Int4            BufferSize,
							      tsp00_Page           &Buffer,
								  tsp00_Int4           &OutPno,
								  tsp00_Int4           &OutPos,
							      tsp00_VfReturn       &HostError,
								  tsp00_ErrText        &ErrText);
private: // members
    cbd930_HistoryFileInfo *hfdFirstHFInfoPtr_bd930;    // pointer to the first entry in the directory
	tsp00_Int4              hfdNumHFInfos_bd930;        // number of HistoryFileInfo in the directory
	tsp00_Int4              hfdMaxNumHFInfos_bd930;     // maximum number of HistoryFileInfo which can be stored  
};

/*---------------------------------------------------------------------------*/

class cbd930_HistoryFileIter
{
public: // methods
		cbd930_HistoryFileInfo* operator () ();
        cbd930_HistoryFileInfo* operator ++ ();
        cbd930_HistoryFileInfo* bd930First  ();
		cbd930_HistoryFileInfo* bd930SetToHistFileWithOldestTrans(tgg00_TransContext &Trans);
		cbd930_HistoryFileIter (cbd930_HistoryFileDir & HistoryFileDir);

private: // members
	cbd930_HistoryFileDir  &hitHFDir_bd930;
	cbd930_HistoryFileInfo *hitCurrHFInfoPtr_bd930;
	int                     hitCurrHFIndex_bd930;
};

/*-----------------------------------------------------------*/

class cbd930_GarbageCollector
{
public:
	cbd930_GarbageCollector (tgg00_TransContext          &Trans,
	                         cbd930_GarbCollController   &GarbCollController);

private: // members
	tgg00_TransContext         &gcoTrans_bd930;
	const int                   gcoIndex_bd930;
	cbd930_GarbCollController  &gcoController_bd930;

private: // methods
	void bd930_HistoryErrorHandler           ();
	void bd930_PageErrorHandler              (cbd900_ObjFileInfo  &ObjFileInfo);
	void bd930_WriteMessageIntoVtrace        (const bool bStartStopFlag);
	void bd930_WritePageStatusIntoVtrace     (tgg00_TransContext    &Trans,
                                              tbd_nodeptr            pPage,
                                              tgg00_VtraceType_Enum  TraceObject);
	void bd930_ReleaseUnusedPagesOfAllFiles  ();
    void bd930_ReleaseUnusedPagesOfOneFile   (tgg00_FileId        &ObjFileId,
											  cbd900_ObjFileInfo  &ObjFileInfo,
											  bool                &bSomethingDone,
											  bool                &bFileIsEmpty);
	void bd930_ReleaseUnusedPagesOfOneChain  (tgg00_FileId        &ObjFileId,
	                                          cbd900_ObjFileInfo  &ObjFileInfo,
	                                          tsp00_Int4           ChainNo);
	void bd930_ReleaseAllPagesOfOneChain     (tgg00_FileId        &ObjFileId,
	                                          cbd900_ObjFileInfo  &ObjFileInfo,
	                                          tsp00_Int4           ChainNo);
    void bd930_RemoveFile                    (tgg00_FileId        &ObjFileId); 	//PTS 1107306 AK 01/08/2000
	bool bd930_TryToReleaseUnusedObjDataPages();
	bool bd930_DataBaseIsTooFull             ();
};

/*---------------------------------------------------------------------------*/

class cbd930_GCState
{
public:
	tsp00_TaskId gcsTaskId_bd930;
	tsp00_Bool   gcsActive_bd930;
	cbd930_GCState();
};

/*---------------------------------------------------------------------------*/

class  cbd930_GarbCollController
{
	friend class cbd930_GarbageCollector;

private://members
    tsp00_Int4                   gccNumGC_bd930;                    // # of all garbage collectors (GC) 
    tsp00_Int4                   gccNumGCUsed_bd930;                // # of used GCs  
	tsp00_Int4                   gccNumGCUsedToReleasePages_bd930;  // # of GCs used to remove empty pages or deleted files
	tsp00_Int4                   gccNumGCReqToReleasePages_bd930;   // # of chains which request the GC to release empty pages
    tsp00_TaskId                 gccStopGCManagerTID_bd930;         // Id of the task coordinating the suspension of all GCs
	cbd900_ObjFileDir           *gccObjFileDirPtr_bd930;
	cbd900_ContObjFileDir       *gccContObjFileDirPtr_bd930;
	cbd930_HistoryFileDir        gccHistoryFileDir_bd930;
	tsp00_Bool                   gccStopGC_bd930;                   // flag that displays whether all GCs have to stoped
    tsp00_Bool                   gccSavepointIsActive_bd930;        // displays if an savepoint is currently active	
	tsp00_Bool                   gccPrepareTruncHistory_bd930;       
#  if COMPILEMODE_MEO00 >= QUICK_MEO00 
    tsp00_Bool                   gccRestartOfAllGCsInAction;
#  endif
	
    ARRAY_VAR_MSP00 (gccGCState_bd930, cbd930_GCState,   MAX_OBJ_GARB_COLL_BD00);

public://methods
	void bd930Init                      (cbd900_ObjFileDir     &ObjFileDir,
		                                 cbd900_ContObjFileDir &ContObjFileDir);
	void bd930AddHistoryFile            (tgg00_TransContext          &Trans, 
										 const tbd94_HistoryFileHandle   &HistoryFileHandle);
	void bd930StartAllGCs               (tgg00_TransContext &Trans,
										 bool                bStartOnlyIfDataCacheIsFull = true);
	void bd930StartAllGCsAfterSavepoint (tgg00_TransContext &Trans);
	void bd930StopAllGCsForSavepoint    (tgg00_TransContext &Trans);
	void bd930StartGCToRemoveEmptyPages (tgg00_TransContext   &Trans,
		                                 cbd900_ObjChainInfo  &ChainInfo);
	void bd930StartGCToRemoveFile       (tgg00_TransContext   &Trans, 
		                                 cbd900_ObjFileInfo   &ObjFileInfo,
										 bool                  bRestart = false);	
	void bd930DumpGCController          (tgg00_VfFileref     &HostFile,
								         tsp00_Int4            BufferSize,
							             tsp00_Page           &Buffer,
								         tsp00_Int4           &OutPno,
								         tsp00_Int4           &OutPos,
							             tsp00_VfReturn       &HostError,
								         tsp00_ErrText        &ErrText);
	bool bd930CheckIfTaskIsAGC          (const tsp00_TaskId        TaskId) const;

private : // methods used by the friend class cbd930_GarbageCollector
	tsp00_Int4 bd930_AddGC        (tgg00_TransContext &Trans);
	void       bd930_FindUnusedGC (tsp00_TaskId       &UnusedGCTaskId);
	void       bd930_CheckGC      ();   
};

/*---------------------------------------------------------------------------*/
/* METHOD IMPLEMENTATION FOR cbd930_GarbCollController                       */
/*---------------------------------------------------------------------------*/

inline void
cbd930_GarbCollController::bd930AddHistoryFile (tgg00_TransContext              &Trans, 
											    const tbd94_HistoryFileHandle   &HistoryFileHandle)
{
	gccHistoryFileDir_bd930.bd930AddHistoryFile (Trans, HistoryFileHandle);
}

/*---------------------------------------------------------------------------*/

inline tsp00_Int4
cbd930_GarbCollController::bd930_AddGC(tgg00_TransContext   &Trans)
{
	vbegexcl (Trans.trTaskId_gg00, g08garbage);	
  	
	/* update number of registerd GCs */
	const int IndexGC = gccNumGC_bd930++;
	
    /* check Number of  garbage collectores */
	if (gccNumGC_bd930 > MAX_OBJ_GARB_COLL_BD00)
		g01abort (csp3_bd_msg, csp3_n_obj, "BD930RegGC:No more slots", IndexGC);
	
	/* check that this slot is really free */
	if (cgg_nil_pid != gccGCState_bd930[IndexGC].gcsTaskId_bd930)
		g01abort (csp3_bd_msg, csp3_n_obj, "BD930RegGC:Slot not free", IndexGC);

	vendexcl (Trans.trTaskId_gg00, g08garbage); 
	
	/* assign TaskIds */ 
	gccGCState_bd930[IndexGC].gcsTaskId_bd930 = Trans.trTaskId_gg00;
	
    return IndexGC;   
}


/*-----------------------------------------------------------*/

inline void
cbd930_GarbCollController::bd930_FindUnusedGC (tsp00_TaskId  &UnusedGCTaskId)	   
{
	
    ROUTINE_DBG_MEO00 ("bd930_FindUnusedGC");
	
	/* This function looks for an unused GC  */
	/* If one found it is marked and         */
	/* bUnusedGCExists ist set to true       */

	/* check if one is exclusively in the g08garbage region */
#   if COMPILEMODE_MEO00 >= SLOW_MEO00 
    g08check_excl (g08garbage);
#   endif
	
	UnusedGCTaskId = cgg_nil_pid;
				
	/* decide if there is still an unused GC which can be started  */
	if (gccNumGCUsed_bd930 < gccNumGC_bd930 && ! gccSavepointIsActive_bd930) 
	{				
		/* find free GC */
		for  (int iGC=0; iGC< gccNumGC_bd930; ++iGC)
		{
			if(!gccGCState_bd930 [iGC].gcsActive_bd930)
			{
				gccGCState_bd930 [iGC].gcsActive_bd930 = true;
				UnusedGCTaskId = gccGCState_bd930 [iGC].gcsTaskId_bd930;
				++gccNumGCUsed_bd930;
				break;
			}
		}
	}
}

/*---------------------------------------------------------------------------*/

inline void
cbd930_GarbCollController::bd930StartAllGCsAfterSavepoint (tgg00_TransContext &Trans)
{

#  if COMPILEMODE_MEO00 >= QUICK_MEO00 
   gccRestartOfAllGCsInAction = true;
#  endif

   gccSavepointIsActive_bd930 = false;
   bd930StartAllGCs (Trans); 

#  if COMPILEMODE_MEO00 >= QUICK_MEO00 
   gccRestartOfAllGCsInAction = false;
#  endif
}

/*-----------------------------------------------------------*/

inline void
cbd930_GarbCollController::bd930StartGCToRemoveEmptyPages (tgg00_TransContext     &Trans,
														   cbd900_ObjChainInfo    &ChainInfo)
{
	ROUTINE_DBG_MSP00 ("bd930StartGCToRemoveEmptyPages");

	/* This function starts a garbage collector. The garbage    */
	/* collector removes empty pages from the given chain. If   */
	/* there is already a GC activ no additionael GC is started */
	
	tsp00_TaskId  GCTaskId = cgg_nil_pid; 
    
	/* check if the file is already marked for a GC Service                */
	/* first try a dirty read to avoid unnessecary acces to a region       */   
	if (!ChainInfo.ociRelPageIsReq_bd900)
	{                               
		vbegexcl (Trans.trTaskId_gg00, g08garbage);
		
		/* check if the file is already marked for a GC Service  */
		/* but now within the datacache region                   */
		if (!ChainInfo.ociRelPageIsReq_bd900)
		{
			/*  mark that the object file requests a GC service  */
			ChainInfo.ociRelPageIsReq_bd900 = true;		
			++gccNumGCReqToReleasePages_bd930;
			bd930_FindUnusedGC (GCTaskId);
		}
		BD930_CHECK_GC;
		
		vendexcl (Trans.trTaskId_gg00, g08garbage);   
		
		/* start unused GC           ===================================== */
		if (cgg_nil_pid != GCTaskId) vresume (GCTaskId, cgg_suspend_lvl_0);
		/*                           ===================================== */
	}
}  

/*-----------------------------------------------------------*/

inline void
cbd930_GarbCollController::bd930StartGCToRemoveFile (tgg00_TransContext   &Trans, 
													 cbd900_ObjFileInfo   &ObjFileInfo,
													 bool                  bRestart)
{
	
   ROUTINE_DBG_MEO00 ("bd930StartGCToRemoveFile");

	tsp00_TaskId           GCTaskId[MAX_NUM_CHAINS_PER_CONTAINER_BD900];
	int              	   NumGCToBeStarted = 0;
	
	/* check if the file is already marked as to be deleted*/ 
	if (( bRestart && !ObjFileInfo.ofiRelFileIsReq_bd900) ||
		(!bRestart &&  ObjFileInfo.ofiRelFileIsReq_bd900))
	{	
#       if COMPILEMODE_MEO00 >= QUICK_MEO00 
		g01abort (csp3_bd_msg, csp3_n_obj, 
			"BD91_StartFile:File del ", ObjFileInfo.ofiRoot_bd900);
#		else
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
			"BD91_StartFile:File del ", ObjFileInfo.ofiRoot_bd900);
#       endif
	}
	
	
	/* mark the file as deleted in the file directory */
	vbegexcl (Trans.trTaskId_gg00, g08garbage);
	
	/* mark that the object file request a GC */
	ObjFileInfo.ofiRelFileIsReq_bd900 = true;
	
	/* find an unused GC */
	for (int iChain=0; iChain < ObjFileInfo.ofiMaxChain_bd900; ++iChain)
	{
		if (!ObjFileInfo[iChain].ociRelPageIsReq_bd900)
		{
			++gccNumGCReqToReleasePages_bd930;
			bd930_FindUnusedGC (GCTaskId[NumGCToBeStarted]);
			if (cgg_nil_pid != GCTaskId[NumGCToBeStarted]) ++NumGCToBeStarted;
		}
	}
	BD930_CHECK_GC;
	
	vendexcl (Trans.trTaskId_gg00, g08garbage);   
	
	/* start unused GC  ===================================== */
	for (int iGC=0; iGC<NumGCToBeStarted; ++iGC)
		vresume (GCTaskId[iGC], cgg_suspend_lvl_0);
	/*======================================================= */
}

/*-----------------------------------------------------------*/

inline void
cbd930_GarbCollController::bd930StopAllGCsForSavepoint (tgg00_TransContext   &Trans) 
{
	ROUTINE_DBG_MEO00  ("bd930StopAllGCs");
	
	/* This routine stops all running garbage collectors */
		
	vbegexcl (Trans.trTaskId_gg00, g08garbage);	

	gccSavepointIsActive_bd930 = true;

	if (0<gccNumGCUsed_bd930)
	{
		gccStopGC_bd930            = true;
		gccStopGCManagerTID_bd930  = Trans.trTaskId_gg00;
	};
	vendexcl (Trans.trTaskId_gg00, g08garbage); 
	
	if (gccStopGC_bd930)
	{
        /*=====================================================*/
        vsuspend (Trans.trTaskId_gg00, cgg_suspend_lvl_0, 229);
		/*=====================================================*/

		/* reset stop order for garbage collectors */
		gccStopGCManagerTID_bd930 = cgg_nil_pid;
	}
	
	/* check if really all GCs are suspended */
	if ((0<gccNumGCUsed_bd930) || gccStopGC_bd930 )
	{
#       if COMPILEMODE_MEO00 >= QUICK_MEO00 
			g01abort (csp3_bd_msg, csp3_n_obj, "BD91_StopGC: GC ACTIVE  ", gccNumGCUsed_bd930);
#		else
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
				"BD91_StopGC: GC ACTIVE  ", gccNumGCUsed_bd930);
#       endif
	}
}

/*---------------------------------------------------------------------------*/

inline bool
cbd930_GarbCollController::bd930CheckIfTaskIsAGC (const tsp00_TaskId  TaskId) const
{
	/* check if the given task id belongs to a Garabage Collector */
	for  (int iGC=0; iGC< gccNumGC_bd930; ++iGC)
	{
		if(gccGCState_bd930 [iGC].gcsTaskId_bd930 == TaskId)
			return true;
	}
	return false;
}

/*---------------------------------------------------------------------------*/

inline void
cbd930_GarbCollController::bd930Init (cbd900_ObjFileDir     &ObjFileDir,
			                          cbd900_ContObjFileDir &ContObjFileDir)
{
	gccNumGC_bd930                   = 0;
	gccNumGC_bd930                   = 0;
	gccNumGCUsed_bd930               = 0;                    
	gccNumGCUsedToReleasePages_bd930 = 0;
	gccNumGCReqToReleasePages_bd930  = 0;
	gccStopGCManagerTID_bd930        = cgg_nil_pid;
    gccObjFileDirPtr_bd930           = &ObjFileDir;
    gccContObjFileDirPtr_bd930       = &ContObjFileDir;
	gccStopGC_bd930                  = false;
	gccSavepointIsActive_bd930       = false;
	gccPrepareTruncHistory_bd930     = false;

#   if COMPILEMODE_MEO00 >= QUICK_MEO00 
	gccRestartOfAllGCsInAction       = false;
#   endif
	gccHistoryFileDir_bd930.bd930Init();
}
	
/*---------------------------------------------------------------------------*/
/* METHOD IMPLEMENTATION FOR cbd930_GarbageCollector                         */
/*---------------------------------------------------------------------------*/

inline void
cbd930_GarbageCollector::bd930_WriteMessageIntoVtrace (const bool bStart)
{
	if (g01vtrace.vtrOmsFree_gg00 )
	{
	    tgg12_OmsGarbCollTrace TrLineGC;

		TrLineGC.ogrbTrType_gg12.becomes(otGarbColl_egg12);
		TrLineGC.ogrbTrType2_gg12.becomes(ot2Nil_egg12);
		TrLineGC.ogrbCollectorIndex_gg12 = gcoIndex_bd930;
		TrLineGC.ogrbIsActiv_gg12        = bStart;
		b12vtrace (gcoTrans_bd930, bd, b930Garbcol, sizeof (TrLineGC), &TrLineGC);
	}
}

/*---------------------------------------------------------------------------*/

inline void 
cbd930_GarbageCollector::bd930_PageErrorHandler (cbd900_ObjFileInfo  &ObjFileInfo)
{
    switch (gcoTrans_bd930.trError_gg00)
    {
    case e_ok            : /* nothing to do                                    */
        break;
    case e_file_not_found: /* the file has been removed and the ObjFileInfo    */
		                   /* has to be updated                                */
		ObjFileInfo.bd900ResetFileInfo();
        gcoTrans_bd930.trError_gg00 = e_ok;
        break;
    default :             /* unknown error, shutdown because of safety reasons */ 
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
			"BD930GC_Error: Index    ", gcoIndex_bd930);
        k28emergency_shutdown (gcoTrans_bd930, m_nil);
    }
}

/*---------------------------------------------------------------------------*/

inline void 
cbd930_GarbageCollector::bd930_HistoryErrorHandler()
{
	if (e_ok != gcoTrans_bd930.trError_gg00)
	{
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
			"BD930HistGC: Error      ", gcoTrans_bd930.trError_gg00);
		g01abort (csp3_bd_msg, csp3_n_obj,"BD930HistGC: Index      ", gcoIndex_bd930);
	}
}

/*---------------------------------------------------------------------------*/
/* METHOD IMPLEMENTATION FOR cbd930_GCState                                  */
/*---------------------------------------------------------------------------*/

inline 
cbd930_GCState::cbd930_GCState()
:
gcsTaskId_bd930 (cgg_nil_pid),
gcsActive_bd930 (false)
{};

/*---------------------------------------------------------------------------*/
/* METHOD IMPLEMENTATION FOR cbd930_HistoryFileInfo                          */
/*---------------------------------------------------------------------------*/

inline
cbd930_HistoryFileInfo::cbd930_HistoryFileInfo()
:
hfiCurrActivGC_bd930          (NIL_GC_INDEX_BD900),
hfiHistoryGetsTruncated_bd930 (false),
hfiFiller1_bd930              (false) 
#ifdef BIT64
,hfiFiller2_bd930             (0)
#endif
{}

/*---------------------------------------------------------------------------*/

inline void
cbd930_HistoryFileInfo::bd930TruncateHistoryFile (tgg00_TransContext & Trans)
{
    bd94History.bd94RemoveOldestTransId (Trans, hfiHistoryFileHandle_bd930);
}

/*---------------------------------------------------------------------------*/
/* METHOD IMPLEMENTATION FOR cbd930_HistoryFileDir                           */
/*---------------------------------------------------------------------------*/

inline void
cbd930_HistoryFileDir::bd930AddHistoryFile (tgg00_TransContext            &Trans, 
											const tbd94_HistoryFileHandle &HistoryFileHandle)
{
	vbegexcl (Trans.trTaskId_gg00, g08garbage);
	if (hfdNumHFInfos_bd930 < hfdMaxNumHFInfos_bd930)
	{
		(hfdFirstHFInfoPtr_bd930 + hfdNumHFInfos_bd930)->hfiHistoryFileHandle_bd930 = HistoryFileHandle;
		hfdNumHFInfos_bd930 ++;
	}
	else
	   g01abort (csp3_bd_msg, csp3_n_obj,"BD930:Too Many HistFiles", hfdNumHFInfos_bd930);
	
	vendexcl (Trans.trTaskId_gg00, g08garbage); 
}

/*---------------------------------------------------------------------------*/

inline void
cbd930_HistoryFileDir::bd930Init()
{
	if (g01is_livecache ()) /* PTS 1106791 Alexander Kley 01/06/2000 */
	{
		tsp00_Bool bOk;
		
		hfdNumHFInfos_bd930    = 0;
		hfdMaxNumHFInfos_bd930 = g01maxuser();
		
		/* allocate memory for the history file entries */
		vallocat (hfdMaxNumHFInfos_bd930 * sizeof (cbd930_HistoryFileInfo), (void **) &hfdFirstHFInfoPtr_bd930, &bOk);
		
		/* write note about allocation into knldiag */
		g01allocate_msg (csp3_n_dynpool, "HIST_DIR  item size    :", sizeof (cbd930_HistoryFileInfo));
		g01allocate_msg (csp3_n_dynpool, "HIST_DIR  num items    :", hfdMaxNumHFInfos_bd930);
		g01allocate_msg (csp3_n_dynpool, "DYNP_B930_HIST_DIR     :", hfdMaxNumHFInfos_bd930 * sizeof (cbd930_HistoryFileInfo));
		
		/* initialize allocated memory */
		if(bOk)
			new (hfdFirstHFInfoPtr_bd930) cbd930_HistoryFileInfo [hfdMaxNumHFInfos_bd930]; 
		else
			g01abort (csp3_sysbuf_storage_exceeded, csp3_n_config,"DYNP_B930_HIST_DIR      ", 0);
	}
}

/*---------------------------------------------------------------------------*/

inline
cbd930_HistoryFileDir::cbd930_HistoryFileDir()
:
hfdFirstHFInfoPtr_bd930 (NULL),
hfdNumHFInfos_bd930     (0),
hfdMaxNumHFInfos_bd930  (0)
{};

/*---------------------------------------------------------------------------*/
/* METHOD IMPLEMENTATION FOR cbd930_HistoryFileIter                          */
/*---------------------------------------------------------------------------*/

inline 
cbd930_HistoryFileIter::cbd930_HistoryFileIter (cbd930_HistoryFileDir & HistoryFileDir)
:
hitHFDir_bd930         (HistoryFileDir),
hitCurrHFInfoPtr_bd930 (NULL),
hitCurrHFIndex_bd930   (NIL_HISTORY_FILE_INDEX_BD930)
{};

/*---------------------------------------------------------------------------*/

inline cbd930_HistoryFileInfo*
cbd930_HistoryFileIter::operator () ()
{
	return hitCurrHFInfoPtr_bd930;
}

/*---------------------------------------------------------------------------*/

inline cbd930_HistoryFileInfo*
cbd930_HistoryFileIter::operator ++ ()
{
	hitCurrHFIndex_bd930 ++;

    if (hitCurrHFIndex_bd930 < hitHFDir_bd930.hfdNumHFInfos_bd930)
		hitCurrHFInfoPtr_bd930 = hitHFDir_bd930.hfdFirstHFInfoPtr_bd930 + hitCurrHFIndex_bd930;
	else
		hitCurrHFInfoPtr_bd930 = NULL;
	
	return hitCurrHFInfoPtr_bd930;
}

/*---------------------------------------------------------------------------*/

inline cbd930_HistoryFileInfo*
cbd930_HistoryFileIter::bd930First()
{
	if (hitHFDir_bd930.hfdNumHFInfos_bd930 > 0)
	{
		hitCurrHFIndex_bd930   = 0;
		hitCurrHFInfoPtr_bd930 = hitHFDir_bd930.hfdFirstHFInfoPtr_bd930;
	}
	else
	{
		hitCurrHFIndex_bd930   = NIL_HISTORY_FILE_INDEX_BD930;
		hitCurrHFInfoPtr_bd930 = NULL;
	}
	
	return hitCurrHFInfoPtr_bd930;
}

/*---------------------------------------------------------------------------*/

#endif  /* GBD930_H */
