/*!**************************************************************************

  module      : gbd94.h
  special area: History Images
  see also    : 
  responsible : UweH
  last changed: 2000-01-21  12:00
  copyright:    Copyright by SAP AG, 2000
  description : 



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


#ifndef GBD94_H
#define GBD94_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"
#include "ggg00.h"
#include "ggg92.h"

#include "gkb08.h"

#if COMPILEMODE_MEO00 >= SLOW_MEO00 
#include "hta99.h" /* no check */
#endif

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#undef FALSE
#define FALSE ((pasbool)0)
#undef TRUE
#define TRUE  ((pasbool)1)

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*! class:       cbd94_HistoryFile
    description: This handle is exchanged between GC and Histman
*/
class cbd94_HistoryFile
{
public:
    /*! function: bd94FileId */
    tgg00_FileId& bd94FileId()        { return FileId_bd94;        }
    /*! function: bd94RightmostPage */
    tsp00_PageNo& bd94RightmostPage() { return RightmostPage_bd94; }
    /*! function: bd94CheckTaskId */
    tsp00_TaskId& bd94CheckTaskId()  { return CheckTaskId_bd94;   }
protected:
    tgg00_FileId  FileId_bd94;
    tsp00_PageNo  RightmostPage_bd94;
    tsp00_TaskId  CheckTaskId_bd94;
private:
    /* this check is necessary for vallocat */
    CHECK_MAPPABLE_CLASS_GIVEN_SIZE_SP00 (cbd94_HistoryFile, 48)
};
/*! endclass: cbd94_HistoryFile */

/*---------------------------------------------------------------------------*/

/*!
definition:  tbd94_HistoryFileHandle
description: This handle is exchanged between GC and Histman
*/
typedef cbd94_HistoryFile* tbd94_HistoryFileHandle;

/*---------------------------------------------------------------------------*/
/*! class:       cbd94_HistoryManager
    description: This is used to insert BeforeImages.
*/
class cbd94_HistoryManager
{   
public:
    /*!
    function:    cbd94_HistoryManager
    description: assigns member
    */
    cbd94_HistoryManager()
        : m_pHistoryFiles(NULL),
          m_MaxHistoryFiles(0)
    {}
    /*!
    function:    bd94Initialize
    description: This constructor creates all necessary HistoryFiles.
    */
    void bd94Initialize (tgg00_TransContext &Trans);
    /*!
    function:    bd94InsertImage
    description: A given BeforeImage and its identification is inserted into a container
    */
    void bd94InsertImage (tgg00_TransContext &Trans,
                          tgg92_KernelOid     TestOid,
                          tgg00_ObjFileNo     ObjFileNo,
                          tgg00_MessType      MessType,
                          tgg91_TransNo       ConsistentView,
                          tgg91_TransNo       PrevUpdTrans,
                          tgg91_PageRef       BeforePageRef,
                          tgg91_PageRef      &NextPageRef,
                          tsp00_Int4          BodySize,
                          void               *pBody);
    /*!
    function:    bd94GetHistory
    arguments:   Given is the PageRef.
                 pObjBody is filled with before image.
                 BodyLen  is assigned the current Length. (variable objects)
    description: This is used by ReadConsistent.
    */
    void bd94GetHistory (tgg00_TransContext &Trans,            // in/out
                         tgg92_KernelOid     TestOid,          // in
                         bool                ChangeDeleteEntry,// in
                         tgg91_PageRef       WantedPageRef,    // in
                         tgg91_PageRef      &NextPageRef,      // out
                         tgg00_MessType     &Action,           // out
                         tgg91_TransNo      &ConsistentView,   // out
                         tgg91_TransNo      &PrevUpdTrans,     // out
                         tgg00_ObjFrameVers &FrameVers,        // out
                         tsp00_Int4         &BodyLen,          // out
                         tsp00_Int4          BodySize = 0,     // in
                         void               *pBody    = NULL); // in(out)
    /*!
    function:    bd94GetObjectBody
    arguments:   Given is the PageRef.
                 pObjBody is filled with before image.
                 BodyLen is assigned the current Length. (variable objects)
    description: This is used by rollback.
                 If RollbackAction is m_delete,
                 then the HistoryEntry is marked with m_drop.
    */
    void bd94GetObjectBody (tgg00_TransContext &Trans,
                            tgg92_KernelOid     TestOid,
                            tgg00_MessType      RollbackAction,
                            tgg91_PageRef       WantedPageRef,
                            tsp00_Int4         &BodyLen,
                            tsp00_Int4          BodySize,
                            void               *pBody);
    /*!
    function:    bd94BeforeImageExists
    arguments:   Given is the PageRef and the OID.
    returns:     true, if a valid BeforeImage exists, else false.
    description: If the PageRef exists and entry is valid and is not m_drop,
                 then true is returned.
    */
    bool bd94BeforeImageExists (tgg00_TransContext &Trans,
                                tgg92_KernelOid     TestOid,
                                tgg91_PageRef       WantedPageRef);
    /*!
    function:    bd94Dump
    description: Dumps the complete Object to the knldump.
    */
    void bd94Dump (tgg00_VfFileref &HostFile,
				   tsp00_Int4       BufferSize,
				   tsp00_Page      &Buffer,
				   tsp00_Int4      &OutPno,
				   tsp00_Int4      &OutPos,
				   tsp00_VfReturn  &HostError,
				   tsp00_ErrText   &ErrText);
    /*!
    function:    bd94GetOldestTransIdToBeRemoved
    arguments:   Trans, HistFileHandle, [in]
                 OldestTransId          [out]
    description: If OldestTransId is not Nil then it could be removed via bd94RemoveOldestTransId.
    */
    void bd94GetOldestTransIdToBeRemoved (tgg00_TransContext      &Trans,
                                          tbd94_HistoryFileHandle  HistFileHandle,
                                          tgg91_TransNo           &OldestTransId);
    /*!
    function:    bd94GetHistFileInfo
    */
    void bd94GetHistFileInfo (tgg00_TransContext &Trans,
                              tsp00_Int4         &HistFileNo,
                              tgg91_TransNo      &FirstTransId,
                              pasbool            &FirstTransIsUsed,
                              tgg91_TransNo      &LastTransId,
                              pasbool            &LastTransIsUsed,
                              tsp00_Int4         &PageCount,
                              pasbool            &bFileIsEmpty);
    /*!
    function:    bd94RemoveHistory
    */
    void bd94RemoveHistory (tgg00_TransContext       &Trans,
                            tbd94_HistoryFileHandle  &HistFileHandle,
                            bool                      bDoExactUsageCheck,
                            bool                     &bAnythingDeleted);
    /*!
    function:    bd94RemoveOldestTransId
    */
    void bd94RemoveOldestTransId (tgg00_TransContext      &Trans,
                                  tbd94_HistoryFileHandle  HistFileHandle);
    /*!
    function:    bd94SaveHistfileInfo
    */
    void bd94SaveHistfileInfo (tgg00_TransContext &Trans);
protected:
    cbd94_HistoryFile*  m_pHistoryFiles; /* PTS 1107331 UH 2000-10-05 */
    tsp00_Int4          m_MaxHistoryFiles;
    /*!
    function:    bd94_GetHistoryFileOfHistFileNo
    arguments:   HistFileNo
    returns:     A HistoryFile.
    */
    cbd94_HistoryFile& bd94_GetHistoryFileOfHistFileNo (tsp00_Int4 HistFileNo)
    {
        return m_pHistoryFiles[HistFileNo]; /* PTS 1107331 UH 2000-10-05 */
    }
    /*!
    function:    bd94_GetHistoryFile
    arguments:   TaskId
    returns:     A HistoryFile.
    */
    cbd94_HistoryFile& bd94_GetHistoryFileOfTaskId (tsp00_TaskId TaskId);
};
/*! endclass: cbd94_HistoryManager */

/*===========================================================================*
 *  GLOBAL OBJECTS                                                           *
 *===========================================================================*/

extern cbd94_HistoryManager bd94History;

/*===========================================================================*
 *  DEFINITION OF INLINE METHODS                                              *
 *===========================================================================*/

/*---------------------------------------------------------------------------*/

#endif  /* GBD94_H */
