/****************************************************************************

  module      : hbd04.h

  -------------------------------------------------------------------------

  responsible : AlexanderK

  special area: kb-bd interface for the bd-funtions of the liveCache
  description : BD FileSystemInterface4: Operations for processing object files ..


  version     : 7.2.0.0
  last changed: 1999-06-22  15:53
  see also    :

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1998



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/

#ifndef HBD04_H
#define HBD04_H


/*****************************************************
 *     MD5: '\321\206\245\336<\016\310\273\203\011\355\230\367xf\324'
 *****************************************************/


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"
#include "ggg00.h"
#include "ggg91.h"
#include "ggg92.h"

/*===========================================================================*
 *  FUNCTION PROTOTYPES                                                      *
 *===========================================================================*/

externCpp void
bd04DelAllObj (tgg00_TransContext  &Trans,
			   tgg00_FileId        &ObjFileId,
			   tsp00_Int4          &NumDelObj);

/*-----------------------------------------------------------------------------------*/

externCpp void
bd04DelObj (tgg00_TransContext  &Trans,
            tgg00_FileId		&ObjFile,
            tgg92_KernelOid      Oid,
            tgg91_PageRef        WantedObjVers);

/*-----------------------------------------------------------------------------------*/
                     
externC void
bd04DropObjFile (tgg00_TransContext    &Trans,
                 tgg00_FileId          &FileId);

/*-----------------------------------------------------------------------------------*/

externCpp void
bd04GetObjConsistent (tgg00_TransContext  &Trans,              //[inp/out]  
                      tgg00_FileId        &ObjFileId,          //[inp]
                      tgg92_KernelOid     &OId,                //[inp/out] 
                      tsp00_Int4           ExtObjBodySize,     //[inp=size of buffer supplied for the object]
                      void                *pExtObjBody,        //[inp=position to which the object hast to be copied]
                      tsp00_Int4           HashKeySize,        //[inp]
                      void                *pHashKey,           //[inp]
					  const bool           bDoLock,            //[inp=true if the object should be locked] 
					  tsp00_Int4          &ObjBodySize,        //[out=real size of object]
                      tsp00_Int4          &KeyLen,             //[out=key len of the object]
                      tsp00_Byte*         *pKey,               //[out] 
                      tgg91_PageRef       &ResultObjVers,      //[out]
    				  tsp00_Int4          &LogReadAccesses);   //[out] PTS 1107819 AK 12/09/2000

/*-----------------------------------------------------------------------------------*/
/* PTS 1107153 AK 17/07/2000 */ 

externCpp void
bd04GetObjFileStatistics (tgg00_TransContext  &Trans,
						  tgg00_FileId        &ObjFileId,
						  tsp00_Int4          &NumAllPages,
						  tsp00_Int4          &NumFreePages,
						  tsp00_Int4          &NumEmptyPages,
						  tsp00_Int4          &NumKeyLeaves,        /* PTS 1107153 AK 17/07/2000 */
						  tsp00_Int4          &Occupancy);          /* PTS 1107153 AK 17/07/2000 */ 

/*-----------------------------------------------------------------------------------*/

externCpp void
bd04GetObjWithKeyConsistent (tgg00_TransContext  &Trans,            //[inp/out]
                             tgg00_FileId        &ObjFileId,        //[inp]
                             tsp00_Int4           ExtObjBodySize,   //[inp]
                             void                *pExtObjBody,      //[inp] 
                             tsp00_Int4          &KeyLen,           //[inp]
                             void                *pKey,             //[inp]
							 const bool           bDoLock,          //[inp] 
                             tgg92_KernelOid     &ResultOid,        //[out]
                             tgg91_PageRef       &ResultObjVers,    //[out]
            				 tsp00_Int4          &LogReadAccesses); //[out] PTS 1107819 AK 12/09/2000

/*-----------------------------------------------------------------------------------*/
/* CRS 1106284 AK 11-03-2000 */
externCpp void
bd04IsObjHistoryUsed (tgg00_TransContext &Trans,
                      tgg00_FileId       &ObjFileId,
                      tgg92_KernelOid     OId,
                      bool               &bHistoryIsUsed);

/*---------------------------------------------------------------------------*/

/* PTS 1001745 UH 23-11-1998 new */
externCpp void
bd04IsObjSelfLocked (tgg00_TransContext &Trans,
                     tgg00_FileId       &ObjFileId,
                     tgg92_KernelOid     OId,
				     bool               &IsSelfLocked);

/*-----------------------------------------------------------------------------------*/

externCpp void
bd04LockObj (tgg00_TransContext &Trans,
             tgg00_FileId       &ObjFileId,
             tgg92_KernelOid     OId,
             tgg91_PageRef       WantedObjVers);

/*-----------------------------------------------------------------------------------*/

externCpp void
bd04NewObjKey (tgg00_TransContext   &Trans,
               tgg00_FileId         &ObjFileId,
               tsp00_Int4		     KeyLen,
               void    		        *pKey,
               tgg92_KernelOid       OId,
               tgg92_KernelOid      &ReUsedOId,
               tgg91_PageRef        &NewObjVers);

/*-----------------------------------------------------------------------------------*/

externCpp void
bd04NewObj (tgg00_TransContext     &Trans,
            tgg00_FileId           &ObjFileId,
            tsp00_Int4             &FrameLen,
            tsp00_Int4		        KeyLen,
            void    			   *pKey,
            tgg92_KernelOid        &ResultOId,
            tgg91_PageRef          &ResultObjVers);

/*-----------------------------------------------------------------------------------*/
 
externCpp void
bd04NextObjConsistent (tgg00_TransContext     &Trans,
					   tgg00_FileId           &ObjFileId,
					   tgg92_KernelOid        &CurrOId,          //[inp,out]
					   bool                    bWithObject,      
					   tsp00_Int4              ExtObjBodySize,   //[inp=size of buffer supplied for the object]
					   void                   *pExtObjBody,
					   tsp00_Int4             &ObjBodySize,      //[out=real size of object]
					   tgg91_PageRef          &ResultObjVers,    
					   tsp00_Int4             &LogReadAccesses); //[out] PTS 1107819 AK 12/09/2000

/*-----------------------------------------------------------------------------------*/

externC void
bd04OmsReadAheadParticipant (tgg00_TransContext  &Trans,
							 tgg00_FileId       &ObjFileId);

/*---------------------------------------------------------------------------*/

externCpp void
bd04ResetObjLock (tgg00_TransContext &Trans,
                  tgg91_TransNo       LockTransId,
                  tgg00_TransIndex    LockTransIndex,
                  tgg92_KernelOid     WantedOId,
				  tgg00_ObjFileNo     ObjFileNo );

/*-----------------------------------------------------------------------------------*/

externCpp void
bd04RollbackObj (tgg00_TransContext  &Trans,
                 tgg00_FileId        &ObjFileId,
                 tgg00_MessType_Enum  RollbackType,
                 tgg92_KernelOid      WantedOId,
                 tgg00_ObjState       NewState,
                 tgg91_PageRef        ObjVers,
                 tgg91_TransNo        ObjUpdTrans,
                 tsp00_Int4           ObjSize,
                 void                *pObject);

/*-----------------------------------------------------------------------------------*/

externCpp void
bd04UnlockObj (tgg00_TransContext   &Trans,
               tgg00_FileId         &ObjFileId,
               tgg92_KernelOid       OId);

/*-----------------------------------------------------------------------------------*/

externCpp void
bd04UpdateObj (tgg00_TransContext       &Trans,
               tgg00_FileId             &ObjFileId,
               tgg92_KernelOid           OId,
               tgg91_PageRef            &ObjVers,
               tsp00_Int4		         NewObjBodySize,
               void					    *pNewObjBody);

/*-----------------------------------------------------------------------------------*/

#endif /* HBD04_H */
