/* @lastChanged: "1999-08-02  14:04"
 * @filename:    vbd11.c
 * @purpose:     "Free Block Management / init, mark, dump"
 * @release:     7.2
 * @see:         ""
 *
 * @copyright:   "Copyright by SAP AG, 1998"


    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

 */

/*----------------------------------------------------------------------------*
 *  INCLUDES                                                                  *
 *----------------------------------------------------------------------------*/

#include "gsp03.h"
#include "gsp03_3.h"
#include "gbd11.h"
#include "hbd11.h"
#include "hgg01_1.h" /*g01opmsg */
#include "heo55k.h"
#include "heo60.h"

/*----------------------------------------------------------------------------*
 *  DEFINES                                                                   *
 *----------------------------------------------------------------------------*/

#undef MOD__ 
#define MOD__  "VBD11.C:  "

#ifdef NO_INLINES
# define _INLINE
#else
#if defined(WIN32)
# define _INLINE	__inline
#else
# define _INLINE
#endif
#endif

#define d_fbmvars    301
#define d_fbmdevinfo 302
#define FBM_GLOB_VARS_MARK "FBMVARS "
 
#define FBM_PERMDEVSP_MARK "PERMDEVS"
#define d_perm_dev_info 303
 
#define d_perm_devs 304
#define FBM_TEMPDEVSP_MARK "TEMPDEVS"
#define d_temp_devs 304
#define FBM_CACHE_INFO_MARK "FBMCACHE"
#define d_cache_info 305
#define FBM_CACHE_PAGE_MARK "CACHE   "
#define d_cache_page 306
 
#define DUMP_MARKER_LEN 8
#define DUMP_MARKER_ID_LEN (sizeof (tsp00_Int2))
#define FULL_DUMP_MARKER_LEN (DUMP_MARKER_LEN+DUMP_MARKER_ID_LEN)

/*----------------------------------------------------------------------------*
 *  MACROS                                                                    *
 *----------------------------------------------------------------------------*/

#define DUMP_PUT_PAGE_HEADER( pno, buf, pos )        \
        { memcpy ( buf, pno, sizeof ( tsp00_Int4 ) ) ; \
          buf [ 4 ] = ptDump_egg00 ; pos = 5 ;           \
        }

/*-----------------------------------------------------------*/
/*           MACROS  for  "fbm_cache_pos"  access            */
/*-----------------------------------------------------------*/
 
#define GET_FBM_CACHE_POS( fbm_pno, cache_index ) \
          ( cache_index = (int ) FBM_GLOB is_fbm_in_cache [ fbm_pno ] )
 
#define SET_FBM_CACHE_POS(fbm_pno, cache_index) \
          ( FBM_GLOB is_fbm_in_cache [ fbm_pno ] = cache_index )
 
#define UNSET_FBM_CACHE_POS(fbm_pno) \
          ( FBM_GLOB is_fbm_in_cache [ fbm_pno ] = -1 )
 
 
/*----------------------------------------------------------------------------*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                              *
 *----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*
 *  EXTERNAL VARIABLES                                                        *
 *----------------------------------------------------------------------------*/

/*********************************/
#ifdef PTOC_Y
#define PASCAL_GLOBAL(_t,_v) extern _t * _v
#define PASCAL_GLOBAL_VALUE(_v) * _v
#else
#define PASCAL_GLOBAL(_t,_v) extern _t _v
#define PASCAL_GLOBAL_VALUE(_v) _v
#endif
/*********************************/

PASCAL_GLOBAL(tsp00_Int4,g08fbm);
#define FBM_REGION       PASCAL_GLOBAL_VALUE(g08fbm)

PASCAL_GLOBAL(tsp00_Int4,g08log);
#define LOG_REGION       PASCAL_GLOBAL_VALUE(g08log)
 
PASCAL_GLOBAL(char, b73spage_ref_statistic);
#define IS_REF_STATISTIC PASCAL_GLOBAL_VALUE(b73spage_ref_statistic)

PASCAL_GLOBAL(t_fbm_glob *, b01_fbm_glob);
#define GLOB_STRUCT      PASCAL_GLOBAL_VALUE(b01_fbm_glob)

PASCAL_GLOBAL(char, g01restart_local);
#define IS_RESTART_LOCAL PASCAL_GLOBAL_VALUE(g01restart_local)


/*----------------------------------------------------------------------------*
 *  LOCAL FUNCTION PROTOTYPES                                                 *
 *----------------------------------------------------------------------------*/
_INLINE
static void bd11_init_hash_and_cache_stamp  ( tsp00_Int4  fbm_cache_size, 
                                              tsp00_Int4 *init_ok ) ;

_INLINE
static void bd11_reset_globvars_and_mem     () ;

_INLINE
static void bd11_expand_fbm_pages           ( tgg00_TransContext      *t, 
                                              FBM_DEV_INFO            *devsp, 
                                              tgg00_BasisError        *tr_err );

_INLINE
static tsp00_Int2 bd11srch_devno_for_free_bl  ( tsp00_Int2 free_bl_wanted) ;

_INLINE
static tsp00_Int4 bd11srch_offset_for_free_bl ( tgg00_TransContext    *t, 
                                              FBM_DEV_INFO            *devsp, 
                                              tgg00_BasisError        *tr_err ) ;
_INLINE
static tsp00_Int4 bd11srch_mfree_offsets   ( tgg00_TransContext     *t,
                                           FBM_DEV_INFO             *devsp, 
                                           tsp00_Int2                min_block_cnt,
                                           tsp00_Int2               *max_block_cnt, 
                                           tgg00_BasisError         *tr_err ) ;

_INLINE
static boolean bd11_get_start_pos        ( fbm_pos         *start_pos, 
                                           FBM_DEV_INFO    *devsp, 
                                           tsp00_Int2      *min_wanted) ;
_INLINE
static int bd11_find_first_free_pos      ( tgg00_TransContext      *t, 
                                           FBM_DEV_INFO     *devsp, 
                                           fbm_pos           act_fbm_pos, 
                                           fbm_pos          *found_fbm_pos, 
                                           tgg00_BasisError *tr_err ) ;
_INLINE
static int bd11_mfind_free_pos           ( tgg00_TransContext      *t, 
                                           FBM_DEV_INFO     *devsp, 
                                           fbm_pos          *act_fbm_pos , 
                                           tsp00_Int2         num_free_wanted, 
                                           tgg00_BasisError *tr_err ) ;
_INLINE
static void bd11_mmark_found_pos         ( tgg00_TransContext      *t, 
                                           fbm_pos          *act_fbm_pos, 
                                           tsp00_Int2        num_to_mark, 
                                           tgg00_BasisError *tr_err );
_INLINE
static void bd11_free_marked_bl_from_dev ( tgg00_TransContext      *t, 
                                           FBM_DEV_INFO     *devsp, 
                                           tbd_fbm           marker, 
                                           tgg00_BasisError *tr_err ) ;

_INLINE
static void bd11dump_put_marker          ( char            *marker, 
                                           tsp00_Int2       marker_id, 
                                           char            *buf, 
                                           tsp00_Int4      *pos) ;
_INLINE
static void bd11dump_next_page           ( tgg_vf_fileref *hostfile, 
                                           tsp00_Page      buf, 
                                           tsp00_PageNo   *dump_pno, 
                                           tsp00_Int4     *pos, 
                                           tsp_vf_return  *hosterr, 
                                           tsp_errtext    *errtext);
_INLINE
static void bd11dump_devinfo             ( tgg_vf_fileref *hostfile, 
                                           tsp00_Page      buf, 
                                           tsp00_PageNo   *dump_pno, 
                                           tsp00_Int4     *pos, 
                                           tsp_vf_return  *hosterr, 
                                           tsp_errtext    *errtext) ;
_INLINE
static void bd11dump_cache_info          ( tgg_vf_fileref *hostfile,
                                           tsp00_Page      buf, 
                                           tsp00_PageNo   *dump_pno, 
                                           tsp00_Int4     *pos, 
                                           tsp_vf_return  *hosterr, 
                                           tsp_errtext    *errtext) ;
/*
 * ========================== GLOBAL FUNCTIONS ================================
 */

/*-----------------------------------------------------------*/
externC void b11_init ( tsp00_Int2 max_data_devs, 
                        tsp00_Int4 max_data_pages,
                        tsp00_Int4 fbm_cache_pages )
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"b11_init"

  int init_ok, avail ;
  tsp00_Int4 total_vallocat_bytes ;
 
  cccprint("BD11  b11_init  " ) ;
  DBG3((MF__, "max_data_pg = %d, fbm_pg = %d", max_data_pages ,  fbm_cache_pages )) ;
  t01p2int4 (TRACE_BD_FBM, "max_data_pg ", max_data_pages,
                           "fbm_cache_pg", fbm_cache_pages ) ;
  t01int4 (TRACE_BD_FBM, "fbm_glob_siz", sizeof ( t_fbm_glob ) ) ;
  g01allocate_msg (csp3_n_dynpool, "entries per FBM page   :",
                   cbd_maxfbmlist_entries);
  g01allocate_msg (csp3_n_dynpool, "FBM_GLOBAL_STRUCT total:",
                   sizeof ( t_fbm_glob ));
  vallocat ( sizeof ( t_fbm_glob ) , & GLOB_STRUCT , &init_ok ) ;
  total_vallocat_bytes = sizeof ( t_fbm_glob ) ;
  if ( init_ok )
  {
    memset ( GLOB_STRUCT, 0, sizeof ( t_fbm_glob ) ) ;
    FBM_GLOB max_data_devs = max_data_devs;
    max_data_devs += FIRST_DEVNO ; /* use devno as index */
    g01allocate_msg (csp3_n_dynpool, "FBM_DEV_INFO total     :",
                     sizeof ( FBM_DEV_INFO ) * max_data_devs );
    g01allocate_msg (csp3_n_dynpool, "MAX_DATA_DEVSPACES     :",
                     max_data_devs ) ;
    g01allocate_msg (csp3_n_dynpool, "FBM_DEVINFO element    :",
                     sizeof (FBM_DEV_INFO));
    vallocat ( sizeof ( FBM_DEV_INFO ) * max_data_devs ,
                      &(FBM_GLOB devsp_info ) , &init_ok ) ;
    total_vallocat_bytes += sizeof ( FBM_DEV_INFO ) * max_data_devs ;
  }
  if ( init_ok )
  {
    memset ( FBM_GLOB devsp_info , 0, sizeof(FBM_DEV_INFO) * max_data_devs ) ;
    g01allocate_msg (csp3_n_dyndata, "DYND_B11_FBM_CACHE     :",
                     fbm_cache_pages );
    vnewbuf ( fbm_cache_pages, &avail, &(FBM_GLOB fbm_cache.cache_page) ,
              &init_ok);
  }
  if ( init_ok )
  {
    FBM_GLOB max_fbm_pno = max_data_pages / cbd_maxfbmlist_entries
                         + FBM_GLOB max_data_devs;
    /* PTS 1107140 */
    if ( fbm_cache_pages < FBM_GLOB max_fbm_pno )
	{
      MSGD (( ERR_FBM_TO_SMALL_IN_PARAM )) ;
      MSGD (( ERR_FBM_WANTED_GIVEN, FBM_GLOB max_fbm_pno, fbm_cache_pages )) ;
      vabort(FALSE);
	}
  }

  if ( init_ok )
  {
    DBG3((MF__,"max_fbm_pno = %d", FBM_GLOB max_fbm_pno ));
    t01int4 (TRACE_BD_FBM, "max_fbm_pno ", FBM_GLOB max_fbm_pno) ;
    g01allocate_msg (csp3_n_dynpool, "FBM_DIRTY_LIST total   :",
                FBM_GLOB max_fbm_pno * sizeof( tsp_int1 ));
    g01allocate_msg (csp3_n_dynpool, "MAX_FBM_PNO            :",
                FBM_GLOB max_fbm_pno );
    g01allocate_msg (csp3_n_dynpool, "FBM_DIRTY element      :",
                sizeof(tsp_int1));
    vallocat ( FBM_GLOB max_fbm_pno * sizeof( tsp_int1 ),
               &(FBM_GLOB fbm_dirty_list), &init_ok );
    total_vallocat_bytes += FBM_GLOB max_fbm_pno * sizeof( tsp_int1 ) ;
  }
  if ( init_ok )
  { 
    tsp00_Int4 pno_list_size = (max_data_pages >> 3) + 1 ;
    FBM_GLOB bup_pno_list.list_size = FBM_GLOB save_pg_pno_list.list_size
                                    = pno_list_size ;
    t01int4 (TRACE_BD_FBM, "tot_pno_list", pno_list_size << 1 ) ;
    g01allocate_msg (csp3_n_dynpool, "FBM_BUP_PNO_LIST total :",
                     (pno_list_size << 1 ) * sizeof( tsp_int1 ));
    vallocat ( FBM_GLOB bup_pno_list.list_size * sizeof( tsp_int1 ),
               &(FBM_GLOB bup_pno_list.pnos ), &init_ok );
    bd11_init_bup_struc( &(FBM_GLOB bup_pno_list) ) ;
    vallocat ( FBM_GLOB save_pg_pno_list.list_size * sizeof( tsp_int1 ),
               &(FBM_GLOB save_pg_pno_list .pnos ), &init_ok );
    memset ( FBM_GLOB save_pg_pno_list.pnos, 0, pno_list_size ) ;
    bd11_init_bup_struc( &(FBM_GLOB save_pg_pno_list) ) ;
    total_vallocat_bytes += ( pno_list_size << 1 ) * sizeof( tsp_int1 ) ;
    g01allocate_msg (csp3_n_dynpool, "DYNP_B11_FBM_STRUC     :", total_vallocat_bytes );
  }
  if ( init_ok )
  { 
    FBM_GLOB fbm_cache.cache_size = fbm_cache_pages ;
    if ( FBM_GLOB max_fbm_pno > FBM_GLOB fbm_cache.cache_size )
      bd11_init_hash_and_cache_stamp( FBM_GLOB fbm_cache.cache_size,
                                      &init_ok ) ;
    else
      DBG3((MF__,"dont't need memory for hash and stamp")) ;
    DBG3((MF__, "global memory successfully allocated")) ;
#ifdef TEST
{ int b11_test_fbm_handling () ;
  b11_test_fbm_handling () ;
}
    { int i, tr_err = e_ok;
      char buffer[4096], *buf = buffer ;
      for ( i=0; i<1000; i++ ) 
        if ( (i & 7) == 7 )   
         {
           b11_mark_save_pages ( NULL, i, 1, i, fbm_occupied, &tr_err );
         } 
       b11_get_remove_pages_bitmap ( NULL, buf, 4096, &i, &tr_err );
       b11_remove_pages( NULL, 1, buf, i,  &tr_err );
    }
#endif

  }
  else
  {
    MSGD (( ERR_SHM_TO_SMALL_FOR,"global variables" )) ;
    vabort(FALSE);
  }
  return ;
}

/*-----------------------------------------------------------*/
externC void b11_init_fbm_for_dev ( tgg00_TransContext      *t,
                                    tsp00_Int2        devno, 
                                    tsp00_Int4        dev_size, 
                                    tgg00_BasisError *tr_err  )
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"b11_init_fbm_for_dev"

  cccprint("BD11  b11_init_f" ) ;
  t01p2int4 (TRACE_BD_FBM, "devno       ", devno,
                                 "dev_size    ", dev_size ) ;
  DBG3((MF__,"devno / dev_size  (%d/%d)", devno, dev_size ));
  if ( devno == FIRST_DEVNO )
    bd11_reset_globvars_and_mem () ;
 
  if ( devno >= FIRST_DEVNO && devno < MAX_DATA_DEVSPACES )
  { 
    static tsp00_Int2 last_devno_created ;
    FBM_DEV_INFO   *devsp = & FBM_GLOB devsp_info [ devno ] ;
    devsp->fst_fbm_page = devno == FIRST_DEVNO ? 0
                        : FBM_GLOB devsp_info [last_devno_created].lst_fbm_page+1;
    devsp->lst_fbm_page        = devsp->fst_fbm_page
                               + dev_size / cbd_maxfbmlist_entries ;
    if ( devsp->lst_fbm_page < FBM_GLOB max_fbm_pno )
    { 
      devsp->dev_size           = dev_size ;
      devsp->used_blocks          = devsp->num_marked_b_up = 0 ;
      devsp->num_marked_svp_free  = 0 ;
      devsp->tot_blocks_per_usage = 1 ;
      devsp->act_blocks_per_usage = 0 ;
      memset( devsp->lst_mult_used, 0, sizeof (devsp->lst_mult_used)) ;
      SAVE_DEV_FIRST_FREE( devsp, 0, devsp->fst_fbm_page, 0 ) ;
      devsp->fst_frag             = devsp->fst_free ;
      devsp->lst_frag.fbm_page = 0 ;
      devsp->lst_frag.fbm_offs = -1 ;
      devsp->act_bup_pos.fbm_page = devsp->fst_fbm_page ;
      devsp->act_bup_pos.fbm_offs = 0 ;
      SAVE_DEV_LAST_USED( devsp, 0, devsp->fst_fbm_page, 0 ) ;
      FBM_GLOB perm_devsp_used++ ;
      FBM_GLOB tot_blocks      += dev_size ;
      FBM_GLOB tot_free_blocks += dev_size ;
      last_devno_created = devno ;
      bd11_expand_fbm_pages ( t, devsp, tr_err  ) ;
    }
    else
    {
      *tr_err = e_illegal_page_address ;
      MSGD ((ERR_MAXFBMPNO_TO_SMALL, FBM_GLOB max_fbm_pno, devsp->lst_fbm_page )) ;
    }
  }
  else
  {
    *tr_err = e_no_of_dev_invalid ;
    MSGD ((ERR_INVALID_DEVNO, devno, FIRST_DEVNO, MAX_DATA_DEVSPACES ));
  }
  return ;
}
 
/*-----------------------------------------------------------*/
externC void b11_mark_block ( tgg00_TransContext      *t, 
                              tsp00_PageNo      pno, 
                              tsp00_Int2        devno, 
                              tsp00_Int4        offset, 
                              tbd_fbm           marker, 
                              tgg00_BasisError *tr_err )
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"b11_mark_block"

  BOOLEAN mark_block = TRUE ;
  FBM_DEV_INFO   *devsp = & FBM_GLOB devsp_info [ devno ] ;
  cccprint("BD11  b11_mark_b" ) ;
  t01p2int4 (TRACE_BD_FBM, "devno       ", devno,
                           "offset      ", offset ) ;
  t01p2int4 (TRACE_BD_FBM, "pno         ", pno  ,
                           "marker      ", marker) ;
  DBG3((MF__,"pno %d (%d/%d), mark %d)", pno, devno, offset, marker )) ;

  *tr_err = e_ok ;
  vbegexcl ( PID(t) , FBM_REGION ) ;

  if ( VALID_DEVNO ( devno ) && VALID_OFFSET( offset, devsp->dev_size ) )
  { 
    int fbm_pno = offset < cbd_maxfbmlist_entries ? devsp->fst_fbm_page
                : devsp->fst_fbm_page + offset / cbd_maxfbmlist_entries ;
    tsp00_Int4 offs_in_fbm_list = (offset % cbd_maxfbmlist_entries);
    tbd_fbm *fbm_list = BD11_GET_LIST ( t, fbm_pno, tr_err ) ;
    if ( *tr_err == e_ok )
      switch ( ACT_BLOCK_VAL ( fbm_list, offs_in_fbm_list ) )
      { 
        case fbm_free     :
             if ( marker != fbm_free )
               if ( marker == fbm_occupied )
               { 
                 devsp->used_blocks++ ;
                 FBM_GLOB tot_free_blocks-- ;
                 if ( devsp->lst_used_offs < offset )
                   SAVE_DEV_LAST_USED  ( devsp, offset, fbm_pno,
                                         offs_in_fbm_list ) ;
                }
                else
                {
                  *tr_err = e_invalid_fbm_mark ;
                  MSGD ((ERR_BLOCK_NOT_USED, devno,  offset )) ;
                }
                break ;
         case fbm_occupied :
               switch ( marker )
               { 
                 case fbm_free : devsp->used_blocks-- ;
                                 FBM_GLOB tot_free_blocks++ ;
                                 if ( devsp->fst_free_offs > offset )
                                   SAVE_DEV_FIRST_FREE (devsp, offset, fbm_pno,
                                                        offs_in_fbm_list ) ;
                                 break ;
                 case fbm_backup  :
                                 if ( IS_BACKUP_PNO_FLAG_SET(
                                         FBM_GLOB bup_pno_list.pnos, pno ) )
                                 { 
                                   *tr_err = e_invalid_fbm_mark ;
                                   MSGD ((ERR_PNO_ALREADY_MARKED_BUP,
                                               pno, devno,  offset )) ;
                                 }
                                 else
                                 {
                                   devsp->num_marked_b_up++ ;
                                   MARK_AFTER_BACKUP( fbm_list,
                                                      offs_in_fbm_list,
                                                      fbm_occupied);
                                   SET_FLAG_IN_BACKUP_LIST( 
                                               FBM_GLOB bup_pno_list, pno ) ;
                                 }
                                 break ;
                 case fbm_occupied: *tr_err = e_invalid_fbm_mark ;
                                    MSGD(( ERR_BLOCK_ALREADY_IN_USE,
                                            devno,  offset )) ;
                                    break ;
                 case fbm_svp_free: devsp->num_marked_svp_free++ ;
                                    FBM_GLOB tot_svp_blocks++ ;
                                    break ;
               }
               break ;
         case fbm_backup :
              if ( marker != fbm_backup )
              {
                MARK_AFTER_BACKUP( fbm_list, offs_in_fbm_list, marker);
                mark_block = FALSE ;
              }
              else
              { 
                *tr_err = e_multiple_converter_entry ;
                MSGD ((ERR_BLOCK_ALREADY_MARKED_BUP, devno, offset )) ;
              }
              break ;
         case fbm_svp_free :
             *tr_err = e_invalid_fbm_mark ;
             MSGD (( ERR_INVALID_MARKER_COMBI, devno,  offset )) ;
             MSGD (( ERR_MARKER_COMBI_IS , 
                      ACT_BLOCK_VAL ( fbm_list, offs_in_fbm_list ) ,marker )) ;
     }
     if ( *tr_err != e_ok )
        vabort (TRUE) ;
     else
       if ( mark_block )
       { 
         MARK_BLOCK ( fbm_list ,offs_in_fbm_list, marker ) ;
         SET_FLAG_IN_DIRTY_LIST ( fbm_pno, marker ) ;
       }
  }
  else
  { 
    *tr_err = e_illegal_page_address ;
    MSGD(( ERR_INVALID_DEVNO_OR_OFFS, devno, offset )) ;
  }
  vendexcl ( PID(t) , FBM_REGION ) ;
  return ;
}
 
/*-----------------------------------------------------------*/
externC void b11_get_free_block ( tgg00_TransContext      *t, 
                                  tsp00_Int2       *devno,
                                  tsp00_Int4       *offset, 
                                  boolean           perm_page, 
                                  tgg00_BasisError *tr_err )
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"b11_get_free_block"

  int savepoint_wanted = FALSE ;
  cccprint("BD11  b11_get_fr" ) ;
#ifdef TEST_MGET
/*----------------------*/
{ int cnt = 1 ;
  b11_mget_free_blocks ( t, &cnt, devno, offset, tr_err ) ;
  return ;
}
/*----------------------*/
#endif

  vbegexcl ( PID(t) , FBM_REGION ) ;

  *devno = bd11srch_devno_for_free_bl ( 1 ) ;

  if ( *devno < FIRST_DEVNO )
    *tr_err = e_no_more_space ;
  else
  { 
    *offset = bd11srch_offset_for_free_bl ( t, 
                              &(FBM_GLOB devsp_info[*devno]), tr_err ) ;
    if ( *offset < 0 )  /* should never be true */
      *tr_err = e_no_more_space ;
    else
    { 
      t01p2int4 (TRACE_BD_FBM, "free devno  ", *devno,
                               "free offset ", *offset ) ;
      DBG3 ((MF__,"free_pos: (%d/%d)", *devno, *offset )) ;
    }
  }

  if ( *tr_err == e_ok && FBM_GLOB tot_svp_blocks > --FBM_GLOB tot_free_blocks
                       && ! FBM_GLOB svp_triggered )
  { 
    FBM_GLOB svp_triggered = savepoint_wanted = TRUE ;
    t01p2int4 (TRACE_BD_FBM, "tot_free_bl ", FBM_GLOB tot_free_blocks ,
                             "tot_svp_bl  ", FBM_GLOB tot_svp_blocks ) ;
    DBG3((MF__,"svp_wanted! tot_free / tot_svp (%d/%d)",
                FBM_GLOB tot_free_blocks, FBM_GLOB tot_svp_blocks )) ;
  }

  vendexcl ( PID(t) , FBM_REGION ) ;
 
  if ( savepoint_wanted && !IS_RESTART_LOCAL)
  { 
    BOOLEAN isexcl = visexcl ( PID(t) , LOG_REGION ) ;
    if ( !isexcl ) vbegexcl ( PID(t) , LOG_REGION ) ;
    k561set_savepoint_wanted() ;
    if ( !isexcl ) vendexcl ( PID(t) , LOG_REGION ) ;
  }
 
  if (*tr_err == e_no_more_space )
  {
    MSGD (( ERR_DEVSP_IS_FULL, *devno )) ;
  }
}
 
/*-----------------------------------------------------------*/
externC void b11_mget_free_blocks ( tgg00_TransContext       *t, 
                                    tsp00_Int2         min_block_cnt, 
                                    tsp00_Int2        *max_block_cnt,
                                    tsp00_Int2        *devno, 
                                    tsp00_Int4        *offset, 
                                    tgg00_BasisError  *tr_err)
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"b11_mget_free_blocks"

  int savepoint_wanted = FALSE ;
  cccprint("BD11  b11_mget_f" ) ;
  t01p2int4 (TRACE_BD_FBM, "free_min    ",  min_block_cnt ,
                           "free_max    ", *max_block_cnt ) ;
 
  DBG3((MF__,"free_wanted: %d-%d", min_block_cnt, *max_block_cnt )) ;
  if ( min_block_cnt > *max_block_cnt )
  {
    DBG3((MF__,"free_wanted: %d-%d", min_block_cnt, *max_block_cnt )) ;
    vabort(TRUE) ;
  }
  vbegexcl ( PID(t) , FBM_REGION ) ;

  *devno = bd11srch_devno_for_free_bl ( *max_block_cnt ) ;
 
  if ( *devno < FIRST_DEVNO )
    *tr_err = e_no_more_space ;
  else
  { 
    *offset = bd11srch_mfree_offsets ( t, &(FBM_GLOB devsp_info[*devno]),
                                       min_block_cnt, max_block_cnt, tr_err );
    if ( *offset < 0 )  /* should never be true */
      *tr_err = e_no_more_space ;
    else
    { 
      t01p2int4 (TRACE_BD_FBM, "free devno  ", *devno,
                               "free offset ", *offset ) ;
      t01int4   (TRACE_BD_FBM, "free_found  ", *max_block_cnt ) ;
      DBG3((MF__, "mget_free: free_pos: (%d/%d), num_free: %d",
                   *devno, *offset, *max_block_cnt )) ;
    }
  }
 
 if ( *tr_err == e_ok &&
      FBM_GLOB tot_svp_blocks > (FBM_GLOB tot_free_blocks -= *max_block_cnt ) &&
      ! FBM_GLOB svp_triggered )
  { 
    FBM_GLOB svp_triggered = savepoint_wanted = TRUE ;
    t01p2int4 (TRACE_BD_FBM, "tot_free_bl ", FBM_GLOB tot_free_blocks ,
                             "tot_svp_bl  ", FBM_GLOB tot_svp_blocks ) ;
    DBG3((MF__,"svp_wanted! tot_free / tot_svp (%d/%d)",
               FBM_GLOB tot_free_blocks, FBM_GLOB tot_svp_blocks )) ;
  }
  vendexcl ( PID(t) , FBM_REGION ) ;
 
  if ( savepoint_wanted && !IS_RESTART_LOCAL)
  { 
    vbegexcl ( PID(t) , LOG_REGION ) ;
    k561set_savepoint_wanted() ;
    vendexcl ( PID(t) , LOG_REGION ) ;
  }
 
  if (*tr_err == e_no_more_space )
  {
    DBG3((MF__,"b11devsp_mget_free_blocks: devno %d full!", *devno )) ;

    g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_fbm,
        "No Free Block available ", FBM_GLOB tot_free_blocks) ;

  }
}

/*-----------------------------------------------------------*/
externC void b11_free_marked_blocks ( tgg00_TransContext      *t, 
                                      tbd_fbm           marker , 
                                      tgg00_BasisError *tr_err)
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"b11_free_marked_blocks"

  int i, devno ;
  cccprint("BD11  b11_free_m" ) ;
  *tr_err = e_ok ;
  t01int4 (TRACE_BD_FBM, "free_marker ", marker) ;
  DBG3((MF__,"srch for mark %d", marker )) ;

  vbegexcl ( PID(t) , FBM_REGION ) ;
  for ( devno = FIRST_DEVNO;
        devno <= FBM_GLOB perm_devsp_used && *tr_err == e_ok ; devno++ )
  { 
    DBG3((MF__,"dev : %d", devno )) ;
    t01int4 (TRACE_BD_FBM, "srch_devno  ", devno) ;
    bd11_free_marked_bl_from_dev ( t, & FBM_GLOB devsp_info [ devno ] ,
                                   marker, tr_err ) ;
  }
  bd11_change_devsp_usage ( t, tr_err ) ;
 
  if ( marker == fbm_svp_free )
  { 
    FBM_GLOB tot_svp_blocks = 0 ;
    FBM_GLOB svp_triggered = FALSE ;
  }

  vendexcl ( PID(t) , FBM_REGION ) ;

#ifdef TEST_FRAG
{
  tsp00_Int2 devno;
  tsp00_Int4 offset ;
  tsp00_Int4 num_free ;
  BOOLEAN in_use ;
  do
    { b11_get_last_frag_block ( t, &devno, &offset, &num_free, tr_err ) ;
      if (  devno >= FIRST_DEVNO )
        {
           b11_is_block_in_use ( t, devno, offset, &in_use, tr_err ) ;
           if ( in_use )
             b11_bup_pno_ready ( t, NIL_PAGE_NO_GG00 , devno, offset, tr_err ) ;
        }
    }
  while ( devno >= FIRST_DEVNO ) ;
}
#endif

  return ;
}
 
/*-----------------------------------------------------------*/
externC tsp00_Int4 b11_blocks_used ( tsp00_Int2 devno )
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"b11_blocks_used"

  tsp00_Int4 blocks_used ;

  blocks_used = VALID_DEVNO ( devno ) ? FBM_GLOB devsp_info[devno].used_blocks : 0 ;

  DBG3((MF__,"devno %d , used %d", devno, blocks_used )) ;

  return ( blocks_used );
}
 
/*-----------------------------------------------------------*/
externC void b11_fbm_statistics ( tsp00_Int4 *available,
                                  tsp00_Int4 *used     , 
                                  tsp00_Int4 *svp_free )
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"b11_fbm_statistics"

  *svp_free   = FBM_GLOB tot_svp_blocks ;
  *available  = FBM_GLOB tot_free_blocks ;
  *used       = FBM_GLOB tot_blocks - *svp_free - *available ;

  DBG3((MF__,"used %d, svp_free %d, avail %d", 
              *used, *svp_free, *available )) ;

  return ;
}
 
/*-----------------------------------------------------------*/
externC void b11dump_fbm ( tgg_vf_fileref *hostfile, 
                           tsp00_Page     *b, 
                           tsp00_PageNo   *dump_pno, 
                           tsp00_Int4     *pos, 
                           tsp_vf_return  *hosterror, 
                           tsp_errtext    *errtext)
/*-----------------------------------------------------------*/
{
  tsp00_Int4  buf_pos ;
  char     *buf = (char *) *b ;
  char     h_buf[10] ;
  tsp00_Int4 i, j, len;
 
  cccprint("BD11  b11dump_fb" ) ;
 
  DUMP_PUT_PAGE_HEADER( dump_pno, buf, buf_pos ) ;
  bd11dump_put_marker( FBM_GLOB_VARS_MARK , d_fbmvars , buf, &buf_pos ) ;
 
  len = (int ) ( (char *)( &FBM_GLOB devsp_info ) - (char *)GLOB_STRUCT ) ;
  memcpy ( buf + buf_pos, GLOB_STRUCT, len ) ;
  buf_pos += len ;
 
  bd11dump_put_marker( FBM_CACHE_INFO_MARK , d_cache_info, buf, &buf_pos ) ;
  len = sizeof ( FBM_CACHE_INFO ) ;
  memcpy ( buf + buf_pos, & FBM_GLOB fbm_cache , len ) ;
  buf_pos += len ;
 
  bd11dump_devinfo( hostfile, (tsp00_Byte *)buf, dump_pno, &buf_pos, hosterror, errtext) ;
 
  if ( ! FBM_GLOB fbm_cache.all_in_cache )
    bd11dump_cache_info( hostfile, (tsp00_Byte *)buf, dump_pno, &buf_pos, 
                         hosterror, errtext ) ;

  bd11dump_next_page( hostfile, (tsp00_Byte *)buf, dump_pno, &buf_pos, hosterror, errtext ) ;

  for ( i = 0; i< FBM_GLOB fbm_cache.cache_used && *hosterror == vf_ok ; i++ )
    vfwrite ( hostfile->no, &(FBM_GLOB fbm_cache.cache_page[i]) ,
              hosterror, *errtext ) ;
    
  *pos = 1 ;
  return ;
}
 
/*-----------------------------------------------------------*/
externC void bd11_init_bup_struc( FBM_BACKUP_LIST *bup_struc )
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"bd11_init_bup_struc"

  bup_struc->num_marked     = 0;
  bup_struc->act_pos        = 0;
  bup_struc->act_bitmap_pno = NIL_PAGE_NO_GG00;
  memset( bup_struc->pnos, 0, bup_struc->list_size) ;
}
 
/*-----------------------------------------------------------*/
externC void bd11_change_devsp_usage ( tgg00_TransContext      *t, 
                                       tgg00_BasisError *tr_err )
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"bd11_change_devsp_usage"

  tsp00_Int2       first_dev = FIRST_DEVNO ;
  tsp00_Int2       last_dev  = FBM_GLOB perm_devsp_used ;
  FBM_DEV_INFO   *devsp ;
  tsp00_Int2       devno ;
  tsp00_Int4       most_used , few_used ;
  tsp00_Int4       blocks_free , blocks_to_this_dev ;
 
  cccprint("BD11  bd11_chang" ) ;
 
  /*************************************************************************/
  /* search for the devsp with the fewest used blocks --> take it         */
  /*************************************************************************/
  if ( first_dev != last_dev )
  {
    tsp00_Int2 few_dev   = first_dev ;
    few_used = most_used = FBM_GLOB devsp_info [ first_dev ] . used_blocks ;
    for ( devno = first_dev + 1, devsp = & FBM_GLOB devsp_info [ devno ];
         devno <= last_dev ; devno++, devsp++ )
    { 
      if ( most_used < devsp->used_blocks )
        most_used = devsp->used_blocks ;
      if ( few_used  > devsp->used_blocks )
      { 
        few_used = devsp->used_blocks ;
        few_dev = devno ;
      }
    }

    FBM_GLOB next_devsp_to_use_for_new_bl = few_dev ;
    t01p2int4 (TRACE_BD_FBM, "most_used_bl", most_used, "few_used_dev",
                              FBM_GLOB next_devsp_to_use_for_new_bl ) ;
    DBG3((MF__,"next_devsp_to_use_for_new_bl %d", few_dev ));
  }
  else
  {  
    devsp          = & FBM_GLOB devsp_info [ first_dev ];
    most_used      = devsp->used_blocks ;
  }
 
  /*************************************************************************/
  /* calculate the number of blocks used for the next time for each devsp  */
  /*************************************************************************/
  for ( devno = first_dev, devsp = & FBM_GLOB devsp_info [ devno ];
        devno <= last_dev ; devno++, devsp++ )
  { 
    blocks_free = devsp->dev_size - devsp->used_blocks ;
    blocks_to_this_dev = most_used - devsp->used_blocks ;
    if ( blocks_free > blocks_to_this_dev )
         devsp->act_blocks_per_usage = devsp->tot_blocks_per_usage -
                                       blocks_to_this_dev ;
    else
      devsp->act_blocks_per_usage = -blocks_free ;
 
    t01p2int4 (TRACE_BD_FBM, "devno       ", devno, "act_bl_per_u",
                              devsp->act_blocks_per_usage );
    if ( (devsp->lst_used_offs - devsp->used_blocks) >
         (devsp->dev_size >> 7) )
    { 
      int i ; 
     /*************************************************************************/
     /* calculate the pos of last multiple blocks used for multiple I/O       */
     /*************************************************************************/
      for ( i=0; i < MAX_MULTIPLE_BLOCKS; i++ )
        devsp->lst_mult_used[i] = devsp->fst_free ;
    }
  }
 
  FBM_GLOB last_devsp_used_for_frag = -1 ;
  return ;
}

/*
 * ========================== LOCAL FUNCTIONS =================================
 */
/*-----------------------------------------------------------*/
_INLINE
static void bd11_init_hash_and_cache_stamp( tsp00_Int4  fbm_cache_size, 
                                            tsp00_Int4 *init_ok )
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"bd11_init_hash_and_cache_stamp"

  cccprint("BD11  bd11_init_" ) ;
  g01allocate_msg (csp3_n_dynpool, "FBM_CACHE_STAMP total  :",
                   sizeof ( stamp_type ) * fbm_cache_size );
  g01allocate_msg (csp3_n_dynpool, "FBM_CACHE_PAGES        :",fbm_cache_size );
  g01allocate_msg (csp3_n_dynpool, "FBM_CACHE_STAMP elemnt :",
                   sizeof(stamp_type));
  vallocat ( sizeof ( stamp_type ) * fbm_cache_size ,
            &(FBM_GLOB fbm_cache.stamp) , init_ok ) ;
  if ( *init_ok )
  {
    g01allocate_msg (csp3_n_dynpool, "FBM_HASH total         :",
                     FBM_GLOB max_fbm_pno * sizeof( tsp00_PageNo ));
    g01allocate_msg (csp3_n_dynpool, "MAX_FBM_PNO            :",
                     FBM_GLOB max_fbm_pno );
    g01allocate_msg (csp3_n_dynpool, "FBM_HASH element       :",
                     sizeof(tsp00_PageNo));
    vallocat ( FBM_GLOB max_fbm_pno * sizeof( tsp00_PageNo ),
                    &(FBM_GLOB is_fbm_in_cache) , init_ok );
  }
  if ( ! *init_ok )
  { 
    MSGD (( ERR_SHM_TO_SMALL_FOR,"fbm hash and stamp" )) ;
    vabort(FALSE);
  }
  return ;
}
 
/*-----------------------------------------------------------*/
_INLINE
static void bd11_reset_globvars_and_mem ()
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"bd11_reset_globvars_and_mem"

  cccprint("BD11  bd11_reset" ) ;
  DBG3((MF__,"called")) ;
  FBM_GLOB fbm_cache.cache_used = 0 ;
  FBM_GLOB fbm_cache.all_in_cache = TRUE ;
  FBM_GLOB next_devsp_to_use_for_new_bl = FIRST_DEVNO ;
  FBM_GLOB perm_devsp_used              = 0 ;
  FBM_GLOB tot_free_blocks              = FBM_GLOB tot_svp_blocks = 0 ;
  FBM_GLOB tot_blocks                   = 0 ;
  FBM_GLOB svp_triggered                = FALSE ;
  memset ( FBM_GLOB fbm_dirty_list , 0x00,
           FBM_GLOB max_fbm_pno * sizeof ( tsp_int1 ) ) ;
  if ( FBM_GLOB max_fbm_pno > FBM_GLOB fbm_cache.cache_size )
  { 
    memset ( FBM_GLOB is_fbm_in_cache, 0xFF,
             FBM_GLOB max_fbm_pno * sizeof ( tsp00_PageNo ) ) ;
    memset ( FBM_GLOB fbm_cache.stamp, 0,
                      FBM_GLOB fbm_cache.cache_size * sizeof ( stamp_type ) );
  }
  return ;
}
 
/*-----------------------------------------------------------*/
_INLINE
static void bd11_expand_fbm_pages ( tgg00_TransContext      *t, 
                                    FBM_DEV_INFO     *devsp, 
                                    tgg00_BasisError *tr_err )
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"bd11_expand_fbm_pages"

  int i;
  tbd_fbmpage *fbm_page ;
  cccprint("BD11  bd11_expan" ) ;
  t01p2int4 (TRACE_BD_FBM, "first_page  ", devsp->fst_fbm_page,
                           "last_page   ", devsp->lst_fbm_page ) ;
  DBG3((MF__,"from %d to %d", devsp->fst_fbm_page , devsp->lst_fbm_page ));
  for ( i=devsp->fst_fbm_page, *tr_err = e_ok  ;
        i <= devsp->lst_fbm_page && *tr_err == e_ok ; i++ )
  {
    if ( i >= FBM_GLOB fbm_cache.cache_size )
      FBM_GLOB fbm_cache.all_in_cache = FALSE ;
    else
    { 
      if ( FBM_GLOB max_fbm_pno > FBM_GLOB fbm_cache.cache_size )
        SET_FBM_CACHE_POS( i, i ) ;
      FBM_GLOB fbm_cache.cache_used++ ;
    }
    fbm_page                        = bd11_get_page (t, i, tr_err ) ;
    if ( *tr_err == e_ok )
    {
      FBM_PAGE_ID(fbm_page) = i ;
      FBM_PAGE_PT(fbm_page) = ptFbm_egg00 ;
      FBM_PAGE_F1(fbm_page) = 0 ;
      FBM_PAGE_C1(fbm_page) = 0 ;
      FBM_PAGE_F2(fbm_page) = 0 ;
      FBM_PAGE_C2(fbm_page) = 0 ;
      FBM_PAGE_MX(fbm_page) = i != devsp->lst_fbm_page
                                ? cbd_maxfbmlist_entries
                                : devsp->dev_size % cbd_maxfbmlist_entries ;
      memset ( FBM_PAGE_DATA(fbm_page), 0, sizeof(FBM_PAGE_DATA(fbm_page) ) ) ;
      SET_FLAG_IN_DIRTY_LIST ( i, fbm_free ) ;
    }
  }
  return ;
}

/*-----------------------------------------------------------*/
_INLINE
static tsp00_Int2 bd11srch_devno_for_free_bl ( tsp00_Int2 free_bl_wanted)
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"bd11srch_devno_for_free_bl"

  int found ;
  FBM_DEV_INFO   *devsp ;
  tsp00_Int2 num_devs   = FBM_GLOB perm_devsp_used;
  tsp00_Int2 srch_beg   = FBM_GLOB next_devsp_to_use_for_new_bl ;
  tsp00_Int2 act_devno  = srch_beg, dev_skipped = 0 ;
 
  cccprint("BD11  bd11srch_d" ) ;
  t01p2int4 (TRACE_BD_FBM, "first_devno ", FIRST_DEVNO,
                           "num_devs    ", num_devs   ) ;
  t01p2int4 (TRACE_BD_FBM, "srch_beg_dev", act_devno,
                           "free_wanted ", free_bl_wanted ) ;
  DBG3((MF__,"devno %d - %d", FIRST_DEVNO, num_devs  )) ;
  DBG3((MF__,"start dev %d to srch %d free blocks", act_devno, free_bl_wanted )) ;

  do
  { 
    devsp = &FBM_GLOB devsp_info[act_devno] ;
    found = devsp->dev_size > ( devsp->used_blocks + free_bl_wanted ) ;
    if ( found )
    { 
      if ( devsp->act_blocks_per_usage >= devsp->tot_blocks_per_usage )
      { 
        found = FALSE ; dev_skipped = act_devno ;
        devsp->act_blocks_per_usage = 0 ;
      }
    }
    else
    { 
      DBG3((MF__,"devsp %d hasn't %d free blocks", act_devno , free_bl_wanted )) ;
      t01int4 (TRACE_BD_FBM, "dev_no_full ", act_devno) ;
    }
    if ( !found )
      if ( ++act_devno > num_devs ) act_devno = FIRST_DEVNO ;
  }
  while ( !found && act_devno != srch_beg ) ;
 
  if ( !found )
    act_devno = dev_skipped ;
  else
    if ( act_devno == num_devs ) 
      FBM_GLOB next_devsp_to_use_for_new_bl = FIRST_DEVNO ;
    else
      FBM_GLOB next_devsp_to_use_for_new_bl = act_devno+1 ;
 
  DBG3((MF__,"found_devno: %d", act_devno )) ;
  t01int4 (TRACE_BD_FBM, "dev_no_found", act_devno) ;
  return ( act_devno ) ;
}
 
/*-----------------------------------------------------------*/
_INLINE
static tsp00_Int4 bd11srch_offset_for_free_bl ( tgg00_TransContext    *t, 
                                              FBM_DEV_INFO     *devsp, 
                                              tgg00_BasisError *tr_err )
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"bd11srch_offset_for_free_bl"

  tsp00_Int4 new_offset ;
  tbd_fbm *act_fbm_list ;
  tsp00_PageNo beg_fbm_pno = devsp->fst_free.fbm_page ,
              act_fbm_pno = beg_fbm_pno ;
  tsp00_Int4 srch_beg_offs= devsp->fst_free.fbm_offs, found ;
 
  cccprint("BD11  bd11srch_o" ) ;
  t01p2int4 (TRACE_BD_FBM, "beg_fbm_pno ", beg_fbm_pno,
                           "beg_fbm_offs", srch_beg_offs ) ;
  DBG3((MF__,"srch_beg_pos (%d/%d)", beg_fbm_pno, srch_beg_offs )) ;
  do
  { 
    if ( IS_FBM_PAGE_DIRTY( act_fbm_pno, fbm_free ) )
    { 
      act_fbm_list = BD11_GET_LIST ( t, act_fbm_pno, tr_err);
      if ( *tr_err == e_ok )
      { 
        while ( ACT_BLOCK_VAL ( act_fbm_list, srch_beg_offs ) != fbm_free 
                && ++srch_beg_offs < cbd_maxfbmlist_entries ) ;
              found = srch_beg_offs < cbd_maxfbmlist_entries ;

        if ( !found ) 
        { 
          CLEAR_FLAG_IN_DIRTY_LIST( act_fbm_pno, fbm_free );
          DBG3((MF__,"fbm %d set full", act_fbm_pno )) ;
        }
      }
    }
    else
    { 
      found = FALSE ;
      DBG3 ((MF__,"dont look in full fbm %d", act_fbm_pno )) ;
    }
    if ( !found )
      srch_beg_offs = 0 ;
  }
  while ( !found && *tr_err == e_ok && ++act_fbm_pno <= devsp->lst_fbm_page ) ;
 
  if ( found )
  { 
    new_offset = (act_fbm_pno - devsp->fst_fbm_page)
                  * cbd_maxfbmlist_entries + srch_beg_offs ;
    MARK_BLOCK ( act_fbm_list , srch_beg_offs , fbm_occupied ) ;
    devsp->used_blocks++ ;
    devsp->act_blocks_per_usage++ ;
    if ( new_offset > devsp->lst_used_offs )
      SAVE_DEV_LAST_USED ( devsp, new_offset, act_fbm_pno, srch_beg_offs );
    SAVE_DEV_FIRST_FREE ( devsp, new_offset, act_fbm_pno, srch_beg_offs ) ;
  }
  else
    new_offset = -1 ;

  DBG3 ((MF__,"returning offset %d", new_offset )) ;

  return ( new_offset ) ;
}
/*-----------------------------------------------------------*/
_INLINE
static tsp00_Int4 bd11srch_mfree_offsets ( tgg00_TransContext    *t,
                                         FBM_DEV_INFO     *devsp, 
                                         tsp00_Int2        min_block_cnt,
                                         tsp00_Int2       *max_block_cnt, 
                                         tgg00_BasisError *tr_err )
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"bd11srch_mfree_offsets"

  tsp00_Int4 new_offset ;
  tbd_fbm *act_fbm_list ;
  tsp00_Int4 found  ;
  fbm_pos     srch_beg_pos, found_fbm_pos ; 
  BOOLEAN     first_free_success ;
  tsp00_Int2    min_avail = min_block_cnt, num_found = 0 ;
 
  first_free_success = bd11_get_start_pos (&srch_beg_pos, devsp, &min_avail);

  cccprint("BD11  bd11srch_m" ) ;
  t01p2int4 (TRACE_BD_FBM, "beg_fbm_pno ", srch_beg_pos.fbm_page,
                           "beg_fbm_offs", srch_beg_pos.fbm_offs ) ;
  t01p2int4 (TRACE_BD_FBM, "free_min    ",  min_block_cnt ,
                                 "free_max    ", *max_block_cnt ) ;
  DBG3((MF__,"srch_beg (%d/%d), free_wanted: %d-%d",
              srch_beg_pos.fbm_page, srch_beg_pos.fbm_offs,
              min_block_cnt, *max_block_cnt )) ;
 
  found = bd11_find_first_free_pos ( t, devsp, 
                                     srch_beg_pos , &found_fbm_pos, tr_err ) ;
 
  if ( !found && !first_free_success )
    first_free_success = found = bd11_find_first_free_pos ( t, devsp, 
                                 devsp->fst_free, &found_fbm_pos, tr_err ) ;
 
  if ( found )
  { 
    tsp00_Int4    save_new_dev_offset ;
 
    if ( *max_block_cnt > 1 )
    { 
      tsp00_Int4    tmp_found ;
      fbm_pos     new_fbm_pos, tmp_fbm_pos = found_fbm_pos ; 

      num_found = bd11_mfind_free_pos ( t, devsp, &tmp_fbm_pos,
                                        *max_block_cnt, tr_err ) ;
 
      if ( num_found < min_avail && *tr_err == e_ok )
      { 
        tmp_found = num_found ;
        do
        {
          DBG3((MF__,"%d(found) != %d(wanted)", tmp_found, *max_block_cnt )) ;
          DBG3((MF__,"new_srch_pos (%d/%d)", tmp_fbm_pos.fbm_page, 
                                             tmp_fbm_pos.fbm_offs )) ;
          found = bd11_find_first_free_pos ( t, devsp, tmp_fbm_pos ,
                                            &tmp_fbm_pos, tr_err ) ;
          if ( found )
          { 
            new_fbm_pos  = tmp_fbm_pos ;
            tmp_found = bd11_mfind_free_pos ( t, devsp, &tmp_fbm_pos,
                                             *max_block_cnt, tr_err );
            if ( tmp_found < min_avail && tmp_found > num_found )
            { 
              DBG3 ((MF__,"%d>%d", tmp_found, num_found )) ;
              if ( num_found < MAX_MULTIPLE_BLOCKS &&
                   ( devsp->lst_mult_used[num_found-1].fbm_page == NOT_AVAIL ||
                     (!devsp->lst_mult_used[num_found-1].fbm_page &&
                      !devsp->lst_mult_used[num_found-1].fbm_page)))

              { 
                devsp->lst_mult_used[num_found-1] = found_fbm_pos;
                DBG3 ((MF__,"by the way: %d:(%d/%d)",
                             tmp_found, found_fbm_pos.fbm_page,
                             found_fbm_pos.fbm_offs )) ;
              }
              found_fbm_pos  = new_fbm_pos ;
              num_found      = tmp_found ;
              first_free_success = FALSE ;
            }
          }
        }
        while ( *tr_err == e_ok && found && tmp_found < min_avail ) ;

        if ( tmp_found >= min_avail )
        { 
          found_fbm_pos  = new_fbm_pos ;
          num_found      = tmp_found ;
          first_free_success = FALSE ;
          DBG3((MF__,"%d bl found at pos (%d/%d)",
                     num_found, found_fbm_pos.fbm_page, found_fbm_pos.fbm_offs ));
        }
      }
    }
    else
      num_found = *max_block_cnt ;
 
    new_offset = (found_fbm_pos.fbm_page - devsp->fst_fbm_page)
                 * cbd_maxfbmlist_entries + found_fbm_pos.fbm_offs;
    bd11_mmark_found_pos ( t, &found_fbm_pos, num_found , tr_err ) ;
    devsp->used_blocks += num_found ;
    devsp->act_blocks_per_usage += num_found ;
    save_new_dev_offset = new_offset + num_found - 1 ;
    if ( save_new_dev_offset > devsp->lst_used_offs )
      SAVE_DEV_LAST_USED ( devsp, save_new_dev_offset ,
                             found_fbm_pos.fbm_page, found_fbm_pos.fbm_offs );
    if ( first_free_success )
      SAVE_DEV_FIRST_FREE ( devsp, save_new_dev_offset ,
                            found_fbm_pos.fbm_page, found_fbm_pos.fbm_offs );

    if ( num_found < MAX_MULTIPLE_BLOCKS )
      if ( num_found < min_avail )
      { 
        int i ;     /* don't look next time for this */
                    /* or higher blockcount          */
        for ( i = min_avail; i > num_found; i-- )
        { 
          devsp->lst_mult_used[i-1].fbm_page = NOT_AVAIL ;
          DBG3((MF__, "bl not avail %d -->%d",i , num_found )) ;
        }
      }
      else
        if ( num_found > 1 )
          devsp->lst_mult_used[num_found-1] = found_fbm_pos ;
  }
  else
    new_offset = -1 ;
 
  *max_block_cnt = num_found ;
  DBG3((MF__,"found offs : %d, free_found: %d", new_offset, *max_block_cnt )) ;
  return ( new_offset ) ;
}

/*-----------------------------------------------------------*/
_INLINE
static boolean bd11_get_start_pos ( fbm_pos         *start_pos, 
                                    FBM_DEV_INFO    *devsp, 
                                    tsp00_Int2      *min_wanted)
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"bd11_get_start_pos"

  BOOLEAN  start_at_first_free ;

  DBG3 ((MF__,"wanted: %d", *min_wanted )) ;
  if ( *min_wanted <= 1 )
  { /************************************/
    /* only one block per I/O wanted    */
    /************************************/
    *start_pos  = devsp->fst_free ;
    start_at_first_free = TRUE ;
  } 
  else
  { 
    tsp00_Int2 bl_cnt ;

    if ( *min_wanted > MAX_MULTIPLE_BLOCKS )
      *min_wanted = MAX_MULTIPLE_BLOCKS ;

    /********************************************************************/
    /* srch last used pos for wanted block_cnt. If it is not available, */
    /* use lower greatest available smaller bl_cnt                      */
    /********************************************************************/

    for ( bl_cnt = *min_wanted - 1 ; 
          bl_cnt > 0 && devsp->lst_mult_used[bl_cnt].fbm_page == NOT_AVAIL ;
          bl_cnt-- ) ;

    if ( bl_cnt == 0 ||
         devsp->lst_mult_used[bl_cnt].fbm_page < devsp->fst_free.fbm_page ||
        ( devsp->lst_mult_used[bl_cnt].fbm_page == devsp->fst_free.fbm_page &&
          devsp->lst_mult_used[bl_cnt].fbm_offs <  devsp->fst_free.fbm_offs ))
    { /************************************/
      /* only one block per I/O available */
      /************************************/
      *start_pos          = devsp->fst_free ;
      start_at_first_free = TRUE ;
    }
    else
    { 
      *start_pos           = devsp->lst_mult_used[bl_cnt] ;
                             start_at_first_free  = FALSE ;
    }

    DBG3 ((MF__,"want/avail: %d/%d (%d/%d)", *min_wanted, bl_cnt+1,
                   start_pos->fbm_page, start_pos->fbm_offs )) ;
    *min_wanted = bl_cnt + 1 ;
  }
  return ( start_at_first_free ) ;
}
 
/*-----------------------------------------------------------*/
_INLINE
static int bd11_find_first_free_pos ( tgg00_TransContext      *t, 
                                      FBM_DEV_INFO     *devsp, 
                                      fbm_pos           act_fbm_pos, 
                                      fbm_pos          *found_fbm_pos, 
                                      tgg00_BasisError *tr_err )
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"bd11_find_first_free_pos"

  tbd_fbm *act_fbm_list ;
  BOOLEAN found = FALSE, set_fbm_to_full ;
  tsp00_Int4 last_offs ;
  cccprint("BD11  bd11_find_" ) ;
  t01p2int4 (TRACE_BD_FBM, "find_beg_fbm", act_fbm_pos.fbm_page,
                           "find_beg_off", act_fbm_pos.fbm_offs ) ;
  DBG3((MF__,"srch_beg (%d/%d)", act_fbm_pos.fbm_page, act_fbm_pos.fbm_offs )) ;

  do
  { 
    set_fbm_to_full = act_fbm_pos.fbm_offs == 0 ;
    if ( IS_FBM_PAGE_DIRTY( act_fbm_pos.fbm_page, fbm_free ) )
    { 
      act_fbm_list = BD11_GET_LIST ( t, act_fbm_pos.fbm_page, tr_err);
      last_offs = LAST_VALID_OFFSET( act_fbm_pos.fbm_page ) ;
      if ( *tr_err == e_ok )
      { 
        while ( ( act_fbm_pos.fbm_offs < last_offs ) &&
                  (ACT_BLOCK_VAL ( act_fbm_list, act_fbm_pos.fbm_offs) 
                  != fbm_free ) )
                ++act_fbm_pos.fbm_offs ;

        found = act_fbm_pos.fbm_offs < last_offs ;
        if ( !found && set_fbm_to_full )
        { 
          CLEAR_FLAG_IN_DIRTY_LIST( act_fbm_pos.fbm_page, fbm_free );
          DBG3 ((MF__,"fbm %d set full", act_fbm_pos.fbm_page )) ;
        }
      } 
    }
    else
    { 
      found = FALSE ;
      DBG3 ((MF__,"dont look in full fbm %d ", act_fbm_pos.fbm_page)) ;
    }

    if ( ! found )
      act_fbm_pos.fbm_offs = 0 ;
    else
      *found_fbm_pos  = act_fbm_pos  ;
  }
  while ( !found && *tr_err == e_ok 
                 && ++act_fbm_pos.fbm_page <= devsp->lst_fbm_page ) ;
  if ( found )
  { 
    DBG3((MF__,"f_first_free: found_pos (%d/%d)",
                found_fbm_pos->fbm_page, found_fbm_pos->fbm_offs )) ;
    t01p2int4 (TRACE_BD_FBM, "found_fbm_pn", found_fbm_pos->fbm_page , 
                             "found_fbm_of", found_fbm_pos->fbm_offs);
  }
  else
    DBG3((MF__,"no free block found !"));
 
  return ( found ) ;
}
 
/*-----------------------------------------------------------*/
_INLINE
static int bd11_mfind_free_pos ( tgg00_TransContext      *t, 
                                 FBM_DEV_INFO     *devsp, 
                                 fbm_pos          *act_fbm_pos , 
                                 tsp00_Int2        num_free_wanted, 
                                 tgg00_BasisError *tr_err )
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"bd11_mfind_free_pos"

  tsp00_PageNo      act_fbm_pno = act_fbm_pos->fbm_page ;
  tsp00_Int4         act_fbm_offs= act_fbm_pos->fbm_offs ;
  int end = FALSE ;
  int num_free_found = 0 ;
  tbd_fbm *fbm_list  = BD11_GET_LIST ( t, act_fbm_pno, tr_err ) ;
  tsp00_Int4 last_offs = LAST_VALID_OFFSET( act_fbm_pno ) ;

  cccprint("BD11  bd11_mfind" ) ;
  DBG3((MF__,"srch_beg (%d/%d) for %d blocks",
                      act_fbm_pno, act_fbm_offs, num_free_wanted )) ;
  t01p2int4 (TRACE_BD_FBM, "find_beg_fbm", act_fbm_pno,
                           "find_beg_off", act_fbm_offs ) ;
  t01int4   (TRACE_BD_FBM, "num_to_find ", num_free_wanted ) ;
 
  while ( *tr_err == e_ok && !end && num_free_found < num_free_wanted &&
          ACT_BLOCK_VAL ( fbm_list, act_fbm_offs ) == fbm_free )
  { 
    num_free_found++ ;
    act_fbm_offs++ ;
    end = act_fbm_pno == devsp->lst_fbm_page && act_fbm_offs >= last_offs ;
    if ( ! end )
      if ( act_fbm_offs == last_offs )
      { 
        act_fbm_offs = 0 ;
        act_fbm_pno++ ;
        fbm_list  = BD11_GET_LIST ( t, act_fbm_pno, tr_err ) ;
        last_offs = LAST_VALID_OFFSET( act_fbm_pno ) ;
      }
  }

  DBG3((MF__,"end_pos (%d/%d) for %d blocks",
                       act_fbm_pno, act_fbm_offs, num_free_found )) ;
  t01p2int4 (TRACE_BD_FBM, "find_end_fbm", act_fbm_pno,
                           "find_end_off", act_fbm_offs ) ;
  t01int4   (TRACE_BD_FBM, "num_found   ", num_free_found  ) ;
  act_fbm_pos->fbm_page = act_fbm_pno  ;
  act_fbm_pos->fbm_offs = act_fbm_offs ;
  return ( num_free_found ) ;
}

/*-----------------------------------------------------------*/
_INLINE
static void bd11_mmark_found_pos ( tgg00_TransContext      *t, 
                                   fbm_pos          *act_fbm_pos, 
                                   tsp00_Int2        num_to_mark, 
                                   tgg00_BasisError *tr_err )
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"bd11_mmark_found_pos"

  int i ;
  tsp00_PageNo     act_fbm_pno = act_fbm_pos->fbm_page ;
  tsp00_Int4        act_fbm_offs= act_fbm_pos->fbm_offs ;
  tbd_fbm *fbm_list = BD11_GET_LIST ( t, act_fbm_pno, tr_err ) ;
  cccprint("BD11  bd11_mmark" ) ;
  DBG3((MF__,"mark_beg_pos (%d/%d) for %d blocks",
              act_fbm_pno, act_fbm_offs, num_to_mark )) ;
  t01p2int4 (TRACE_BD_FBM, "mark_beg_fbm", act_fbm_pno,
                           "mark_beg_off", act_fbm_offs ) ;
  t01int4   (TRACE_BD_FBM, "num_to_mark ", num_to_mark) ;
 
  for ( i=0; i<num_to_mark && *tr_err == e_ok ; )
  {
    MARK_BLOCK ( fbm_list , act_fbm_offs , fbm_occupied ) ;
    if ( ++i < num_to_mark && ++act_fbm_offs >= cbd_maxfbmlist_entries )
    { 
      act_fbm_offs = 0 ;
      act_fbm_pno++ ;
      DBG3((MF__,"use nxt fbm (%d/%d)", act_fbm_pno, act_fbm_offs )) ;
      fbm_list = BD11_GET_LIST ( t, act_fbm_pno, tr_err ) ;
    }
  }
  DBG3((MF__,"last_mark_pos (%d/%d)", act_fbm_pno, act_fbm_offs )) ;

  t01p2int4 (TRACE_BD_FBM, "new_free_fbm", act_fbm_pno,
                           "new_free_off", act_fbm_offs ) ;

  act_fbm_pos->fbm_page = act_fbm_pno  ;
  act_fbm_pos->fbm_offs = act_fbm_offs ;
  return ;
}
 
 
/*-----------------------------------------------------------*/
_INLINE
static void bd11_free_marked_bl_from_dev ( tgg00_TransContext      *t, 
                                           FBM_DEV_INFO     *devsp, 
                                           tbd_fbm           marker, 
                                           tgg00_BasisError *tr_err )
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"bd11_free_marked_bl_from_dev"

  tsp00_Int4 ind, last_offs, sv_first_free_offs ;
  tsp00_PageNo fbm, sv_first_fbm_page ;
  tbd_fbm *act_fbm_list ;
  tsp00_Int4 num_to_srch = devsp->num_marked_svp_free ;
  BOOLEAN last_used_found ;
 
  cccprint("BD11  bd11_free_" ) ;
 
  /*************************************************************************/
  /* change all "svp_free" - blocks to "free" = blocks                     */
  /*************************************************************************/
  for ( fbm       = devsp->lst_fbm_page,
        last_offs = devsp->lst_used.fbm_offs ;
        fbm >= devsp->fst_fbm_page && num_to_srch > 0  && *tr_err == e_ok ;             fbm--, last_offs = cbd_maxfbmlist_entries - 1 )
    if ( IS_FBM_PAGE_DIRTY ( fbm , marker ) )
    { 
      DBG3((MF__,"found fbm_pno : %d", fbm )) ;
      CLEAR_FLAG_IN_DIRTY_LIST ( fbm, marker ) ;
      SET_FLAG_IN_DIRTY_LIST ( fbm, fbm_free ) ;
      if ( (act_fbm_list = BD11_GET_LIST ( t, fbm, tr_err ) ) != NULL )
      { 
      for ( ind = last_offs;
            ind >= 0 && num_to_srch > 0 ; ind -- )
        if ( ACT_BLOCK_VAL ( act_fbm_list , ind ) == marker )
        {  
          MARK_BLOCK ( act_fbm_list , ind , fbm_free ) ;
          if ( --num_to_srch == 0 )
          { 
            tsp00_Int4 first_offs = ( fbm - devsp->fst_fbm_page )
                                  * cbd_maxfbmlist_entries + ind ;
            if ( first_offs < devsp->fst_free_offs )
              SAVE_DEV_FIRST_FREE( devsp, first_offs, fbm, ind);
          }
        }
      }
    }
  else
    DBG3((MF__,"skipping pno %d", fbm )) ;
 
  /*************************************************************************/
  /* search for the last used block-pos after svp_free ==> free            */
  /*************************************************************************/
  for ( fbm =  devsp->lst_fbm_page, last_used_found = FALSE ,
        last_offs = devsp->lst_used.fbm_offs ;
        fbm >= devsp->fst_fbm_page && !last_used_found && *tr_err == e_ok ; 
        last_offs = cbd_maxfbmlist_entries - 1 )
  { 
    if ( (act_fbm_list = BD11_GET_LIST ( t, fbm, tr_err ) ) != NULL )
    { 
      for ( ind = last_offs; ind >= 0 && !last_used_found; )
        if ( ACT_BLOCK_VAL (act_fbm_list, ind ) != fbm_free )
          last_used_found = TRUE ;
        else
          ind -- ;
    }
    if ( !last_used_found )
      fbm-- ;
  }
 
  if ( last_used_found )
    SAVE_DEV_LAST_USED  ( devsp,
                         ((fbm - devsp->fst_fbm_page) * cbd_maxfbmlist_entries
                         + ind) , fbm, ind )
  else
    SAVE_DEV_LAST_USED  ( devsp, 0, devsp->fst_fbm_page, 0 ) ;
 
  devsp->num_marked_svp_free -= num_to_srch ;
  DBG3((MF__,"%d bl. freed", devsp->num_marked_svp_free )) ;
  t01int4 (TRACE_BD_FBM, "freed_blocks", devsp->num_marked_svp_free ) ;
  devsp->used_blocks       -= devsp->num_marked_svp_free ;
  FBM_GLOB tot_free_blocks += devsp->num_marked_svp_free ;
  devsp->num_marked_svp_free = num_to_srch ;
  return ;
}

/*-----------------------------------------------------------*/
_INLINE
static void bd11dump_put_marker ( char       *marker, 
                                  tsp00_Int2  marker_id, 
                                  char       *buf, 
                                  tsp00_Int4 *pos)
/*-----------------------------------------------------------*/
{
  cccprint("BD11  bd11dump_p" ) ;
  memcpy ( buf + *pos, marker, DUMP_MARKER_LEN);
  *pos += DUMP_MARKER_LEN ;
  memcpy ( buf+ *pos, &marker_id, sizeof ( marker_id ) );
  *pos += sizeof ( marker_id )  ;
  return ;
}
 
/*-----------------------------------------------------------*/
_INLINE
static void bd11dump_next_page( tgg_vf_fileref *hostfile, 
                                tsp00_Page      buf, 
                                tsp00_PageNo   *dump_pno, 
                                tsp00_Int4     *pos, 
                                tsp_vf_return  *hosterr, 
                                tsp_errtext    *errtext)
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"bd11dump_next_page"

  int len = sizeof ( tsp00_Page ) - *pos ;
  if ( len > 0 )
    memset ( buf + *pos, 254, len ) ;
  vfwrite ( hostfile->no, buf, hosterr, *errtext ) ;
  if ( *hosterr == vf_ok )
  {
    DBG3((MF__,"%d bytes written to dump_pno %d",
                      sizeof ( tsp00_Page ), *dump_pno)) ;
    (*dump_pno)++ ;
    DUMP_PUT_PAGE_HEADER( dump_pno, buf, *pos );
  }
  else
    DBG3((MF__,"%d %s", *hosterr, *errtext)) ;
  return ;
}
 
 
/*-----------------------------------------------------------*/
_INLINE
static void bd11dump_devinfo( tgg_vf_fileref *hostfile, 
                              tsp00_Page      buf, 
                              tsp00_PageNo   *dump_pno, 
                              tsp00_Int4     *pos, 
                              tsp_vf_return  *hosterr, 
                              tsp_errtext    *errtext)
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"bd11dump_devinfo"

  int i, info_len = sizeof ( FBM_DEV_INFO ) ,
      entry_len = FULL_DUMP_MARKER_LEN + info_len ;
  tsp00_Int2  entries_per_page = ( sizeof (tsp00_Page) - *pos -
                                 FULL_DUMP_MARKER_LEN -  /* first marker */
                                 sizeof(entries_per_page) ) / entry_len ;
  char     h_buf[10] ;
 
  if ( FBM_GLOB perm_devsp_used < entries_per_page )
    entries_per_page = FBM_GLOB perm_devsp_used;
 
  bd11dump_put_marker( FBM_PERMDEVSP_MARK , d_perm_dev_info , (char *)buf, pos ) ;
  memcpy ( buf + *pos, &entries_per_page, sizeof( entries_per_page )  ) ;
  *pos += sizeof( entries_per_page ) ;
 
  for ( i=1; i<= FBM_GLOB perm_devsp_used && *hosterr == vf_ok ; i++ )
  {
    if ( *pos + entry_len > sizeof ( tsp00_Page ) )
    { 
      bd11dump_next_page( hostfile, buf, dump_pno, pos, hosterr, errtext ) ;
      bd11dump_put_marker( FBM_PERMDEVSP_MARK , d_perm_dev_info ,
                           (char *)buf, pos ) ;
      entries_per_page = ( sizeof ( tsp00_Page ) - *pos -
                           sizeof(entries_per_page) ) / entry_len ;
      if ( (FBM_GLOB perm_devsp_used-i) < entries_per_page )
        entries_per_page = FBM_GLOB perm_devsp_used - i + 1;
      memcpy ( buf + *pos, &entries_per_page, sizeof( entries_per_page )  ) ;
      *pos += 2 ;
    }
    sprintf ( h_buf, "P_DEV %2d", i ) ;
    bd11dump_put_marker( h_buf , d_perm_devs , (char *)buf, pos ) ;
    memcpy ( buf + *pos, & (FBM_GLOB devsp_info[i]) , info_len ) ;
    *pos += info_len ;
  }
}
 
/*-----------------------------------------------------------*/
_INLINE
static void bd11dump_cache_info( tgg_vf_fileref *hostfile,
                                 tsp00_Page      buf, 
                                 tsp00_PageNo   *dump_pno, 
                                 tsp00_Int4     *pos, 
                                 tsp_vf_return  *hosterr, 
                                 tsp_errtext    *errtext)
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"bd11dump_cache_info"

  int i, len,
      entry_len = sizeof ( i ) + sizeof ( tsp00_PageNo ) + sizeof ( stamp_type );
  tsp00_Int4  buf_pos = *pos ;
  tsp00_Int2  entries_per_page = ( sizeof ( tsp00_Page ) - buf_pos -
                                 FULL_DUMP_MARKER_LEN -  /* first marker */
                                 sizeof(entries_per_page) ) / entry_len ;
 
  if ( FBM_GLOB fbm_cache.cache_used < entries_per_page )
    entries_per_page = FBM_GLOB fbm_cache.cache_used ;
 
  bd11dump_put_marker( FBM_CACHE_PAGE_MARK , d_cache_page, (char *)buf, &buf_pos ) ;
  memcpy ( buf + buf_pos, &entries_per_page, sizeof( entries_per_page )  ) ;
  buf_pos += 2 ;
  for ( i = 0; i < FBM_GLOB fbm_cache.cache_used && *hosterr == vf_ok ;
        i++ )
  {
    if ( buf_pos + entry_len > sizeof ( tsp00_Page ) )
    {
      bd11dump_next_page( hostfile, buf, dump_pno, &buf_pos,
                                         hosterr, errtext ) ;
      bd11dump_put_marker( FBM_CACHE_PAGE_MARK ,
                           d_cache_page, (char *)buf, &buf_pos ) ;
      entries_per_page = ( sizeof ( tsp00_Page ) - buf_pos -
                           sizeof(entries_per_page) ) / entry_len ;
      if ( (FBM_GLOB fbm_cache.cache_used-i) < entries_per_page )
        entries_per_page = FBM_GLOB fbm_cache.cache_used  - i ;
      memcpy ( buf + buf_pos, &entries_per_page,
               sizeof( entries_per_page )  ) ;
      buf_pos += 2 ;
    }
    len = sizeof(i) ;
    memcpy ( buf + buf_pos, &i, len ) ;
    buf_pos += len ;
    len = sizeof ( tsp00_PageNo ) ;
    memcpy ( buf + buf_pos,
             & (FBM_PAGE_ID( (&FBM_GLOB fbm_cache.cache_page[i]) ) ), len ) ;
    buf_pos += len ;
    len = sizeof ( stamp_type ) ;
    if ( FBM_GLOB max_fbm_pno > FBM_GLOB fbm_cache.cache_size )
      memcpy ( buf + buf_pos, & (FBM_GLOB fbm_cache.stamp[i]), len ) ;
    else
    {
      stamp_type no_stamp ;
      memset ( &no_stamp, 0, len );
      memcpy ( buf + buf_pos, & no_stamp , len ) ;
    }
    buf_pos += len ;
  }
  return ;
}

/*
 * =============================== END ========================================
 */
