/* @lastChanged: "1998-12-30  16:51"
 * @filename:    vbd110.c
 * @purpose:     "Free Block Management / backup. fragmenation handling
 * @release:     7.2
 * @see:         ""
 *
 * @copyright:   "Copyright by SAP AG, 1998"


    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

 */

/*----------------------------------------------------------------------------*
 *  INCLUDES                                                                  *
 *----------------------------------------------------------------------------*/

#include "gbd11.h"
#include "hbd11.h"
#include "heo55k.h"
#include "heo60.h"

/*----------------------------------------------------------------------------*
 *  DEFINES                                                                   *
 *----------------------------------------------------------------------------*/
#undef MOD__ 
#define MOD__  "VBD110.C: "

#ifdef NO_INLINES
# define _INLINE
#else
# if defined(WIN32)
#  define _INLINE	__inline
# else
#  define _INLINE
# endif
#endif


/*----------------------------------------------------------------------------*
 *  MACROS                                                                    *
 *----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                              *
 *----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------*
 *  EXTERNAL VARIABLES                                                        *
 *----------------------------------------------------------------------------*/

/*********************************/
#ifdef PTOC_Y
#define PASCAL_GLOBAL(_t,_v) extern _t * _v
#define PASCAL_GLOBAL_VALUE(_v) * _v
#else
#define PASCAL_GLOBAL(_t,_v) extern _t _v
#define PASCAL_GLOBAL_VALUE(_v) _v
#endif
/*********************************/

PASCAL_GLOBAL(tsp00_Int4,g08fbm);
#define FBM_REGION       PASCAL_GLOBAL_VALUE(g08fbm)

PASCAL_GLOBAL(t_fbm_glob *, b01_fbm_glob);
#define GLOB_STRUCT      PASCAL_GLOBAL_VALUE(b01_fbm_glob)

/*----------------------------------------------------------------------------*
 *  GLOBAL VARIABLES                                                          *
 *----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------*
 *  LOCAL VARIABLES                                                           *
 *----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------*
 *  LOCAL FUNCTION PROTOTYPES                                                 *
 *----------------------------------------------------------------------------*/

_INLINE
static tsp00_Int4 bd11_get_avg_for_balance () ;

_INLINE
static void bd11_mark_frag_blocks        ( FBM_DEV_INFO     *devsp , 
                                           tsp00_Int4        srch_end_pos, 
                                           tsp00_Int4        max_to_mark, 
                                           tgg00_BasisError *tr_err ) ;

_INLINE
static int bd11_get_last_frag_from_dev   ( tgg00_TransContext      *t, 
                                           tsp00_Int2        devno, 
                                           tsp00_Int4       *offset, 
                                           tgg00_BasisError *tr_err );

_INLINE
static void bd11_restore_backup_marker   ( FBM_DEV_INFO    *devsp,
                                           tsp00_Int4       offset,
                                           tsp00_PageNo     fbm_pno,
                                           tbd_fbm         *fbm_list, 
                                           tsp00_Int4       offs_in_fbm_list );

_INLINE
static int bd11_nxt_b_up_pos             ( tgg00_TransContext      *t, 
                                           FBM_DEV_INFO     *devsp, 
                                           tsp00_Int4       *found_fbm_offs, 
                                           tgg00_BasisError *tr_err ) ;

/*
 * ========================== GLOBAL FUNCTIONS ================================
 */

/*-----------------------------------------------------------*/
externC void b11_mark_save_pages ( tgg00_TransContext     *t, 
                                  tsp00_PageNo      pno, 
                                  tsp00_Int2        devno, 
                                  tsp00_Int4        offset, 
                                  tbd_fbm           marker, 
                                  tgg00_BasisError *tr_err )
/*-----------------------------------------------------------*/
{

#undef MF__
#define MF__   MOD__"b11_mark_save_pages"

  cccprint("BD11  b11_mark_s" ) ;
  t01p2int4 (TRACE_BD_FBM, "devno       ", devno,
                           "offset      ", offset ) ;
  t01p2int4 (TRACE_BD_FBM, "pno         ", pno  ,
                           "marker      ", marker) ;

  DBG3((MOD__,"pno %d (%d/%d), mark %d)", pno, devno, offset, marker  )) ;

  if ( marker == fbm_backup )
    b11_mark_block ( t, pno, devno, offset, marker, tr_err );

  if ( *tr_err == e_ok )
  { 
    if ( ! IS_BACKUP_PNO_FLAG_SET( FBM_GLOB save_pg_pno_list.pnos, pno ) )
    {
      SET_FLAG_IN_BACKUP_LIST( FBM_GLOB save_pg_pno_list, pno ) ;
      if ( FBM_GLOB save_pg_pno_list.act_bitmap_pno == NIL_PAGE_NO_GG00 ) 
           FBM_GLOB save_pg_pno_list.act_bitmap_pno = 0;
    }
    else
    { 
      *tr_err = e_multiple_converter_entry ;
      MSGD (( ERR_PNO_ALREADY_MARKED_BUP, pno, devno, offset )) ;
    }
  }
}

 
/*----------------------------------------------------------------------------*/
externC void b11_get_remove_pages_bitmap ( tgg00_TransContext         *t, 
                                           tsp_moveobj_ptr      buf, 
                                           tsp00_Int4           bytes_wanted, 
                                           tsp00_Int4          *bytes_given,
                                           tsp00_PageNo        *bitmap_pno,
                                           tgg00_BasisError    *tr_err)
/*----------------------------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"b11_get_remove_pages_bitmap"

  cccprint("BD11  b11_get_re" ) ;
  DBG3((MF__,"wanted %d ", bytes_wanted )) ;
  DBG3((MF__,"act_pos %d, list_size %d ", 
                          FBM_GLOB save_pg_pno_list.act_pos ,
                          FBM_GLOB save_pg_pno_list.list_size )) ;

  vbegexcl ( PID(t) , FBM_REGION ) ;

  if ( FBM_GLOB save_pg_pno_list.act_bitmap_pno == 0 )
    {
      DBG3 ((MF__, "%d PNOS marked for b_up in sv_pag_list", FBM_GLOB save_pg_pno_list.num_marked ));
    }

  if ( FBM_GLOB save_pg_pno_list.act_bitmap_pno != NIL_PAGE_NO_GG00 )
    {
      if ( FBM_GLOB save_pg_pno_list.act_pos < FBM_GLOB save_pg_pno_list.list_size )
        { 
          *bitmap_pno = FBM_GLOB save_pg_pno_list.act_bitmap_pno ;
          FBM_GLOB save_pg_pno_list.act_bitmap_pno ++ ;

          if ( FBM_GLOB save_pg_pno_list.act_pos + bytes_wanted >
               FBM_GLOB save_pg_pno_list.list_size  )
            *bytes_given = FBM_GLOB save_pg_pno_list.list_size - 
                           FBM_GLOB save_pg_pno_list.act_pos ;
          else
            *bytes_given = bytes_wanted ;
  
          memcpy ( (char *)buf, 
                    FBM_GLOB save_pg_pno_list.pnos + FBM_GLOB save_pg_pno_list.act_pos,
                    *bytes_given ) ;

          FBM_GLOB save_pg_pno_list.act_pos += *bytes_given  ;
        }
      else
        { 
          *bytes_given = 0 ;
          *bitmap_pno  = NIL_PAGE_NO_GG00 ;
        }

      if ( *bytes_given < bytes_wanted )
        bd11_init_bup_struc( &(FBM_GLOB save_pg_pno_list) ) ;
    }
  else
    { 
      *bytes_given = 0 ;
      *bitmap_pno  = NIL_PAGE_NO_GG00 ;
    }

  vendexcl ( PID(t) , FBM_REGION ) ;
  DBG3((MF__,"given  %d ", *bytes_given )) ;
}


/*----------------------------------------------------------------------------*/
externC void b11_remove_pages( tgg00_TransContext         *t, 
                               tsp00_Int4           BitmapPos, 
                               tsp_moveobj_ptr      buf, 
                               tsp00_Int4           bytes_given, 
                               tgg00_BasisError    *tr_err)
/*----------------------------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"b11_remove_pages"

  tsp00_PageNo start_pno, end_pno, act_abs_pno, act_rel_pno ;
  void       *cbptr = NULL ;

  cccprint("BD11  b11_remove" ) ;

  BACKUP_LIST_DATA_PNOS(BitmapPos, bytes_given, start_pno, end_pno ) ;

  t01p2int4 (TRACE_BD_FBM, "bitmap_pno  ", BitmapPos,
                           "bytes_given ", bytes_given ) ;
  t01p2int4 (TRACE_BD_FBM, "start_pno   ", start_pno,
                           "end_pno     ", end_pno ) ;

  DBG3((MF__,"BitmapPos %d, given %d ", BitmapPos, bytes_given )) ;
  DBG3((MF__,"from %d - %d", start_pno, end_pno )) ;

  for ( act_abs_pno = start_pno, act_rel_pno = 0; 
        act_abs_pno <= end_pno && *tr_err == e_ok ; 
        act_abs_pno++, act_rel_pno++  )
    if ( ! IS_BACKUP_PNO_FLAG_SET( ((char *)buf) , act_rel_pno ) )
      {
        DBG3((MF__,"pno %d/%d", act_abs_pno, act_rel_pno )) ;
        b10bup_free_pno (t, act_abs_pno, &cbptr );
      }
    if ( *tr_err == e_illegal_page_address )
      *tr_err = e_ok ;
}

/*-----------------------------------------------------------*/
externC void b11devsp_get_free_buffer ( tgg00_TransContext      *t,
                                        tsp00_Int2        devno,
                                        tsp00_Int2        block_cnt,
                                        tsp_int1          mark_as_used,
                                        tsp00_Int4       *offset,
                                        tgg00_BasisError *tr_err )
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"b11devsp_get_free_buffer"

  FBM_DEV_INFO   *devsp = &(FBM_GLOB devsp_info[devno]) ;
 
  cccprint("BD11  b11devsp_g" ) ;
  t01p2int4 (TRACE_BD_FBM, "devno       ", devno,
                           "block_cnt   ", block_cnt ) ;
  DBG3((MF__,"devno %d , block_cnt %d", devno, block_cnt )) ;
 
  if ( devsp->lst_used_offs + 1 + block_cnt > devsp->dev_size )
  { 
      /* devsp->lst_used_offs = 0 ;*/
    *tr_err = e_no_more_space ;
    MSGD (( ERR_DEVSP_IS_FULL, devno )) ;
  }
  else
  { 
    *offset = devsp->lst_used_offs+1  ;
    devsp->lst_used_offs += block_cnt ;
    DBG3((MF__,"free pos: %d / %d", devno , *offset )) ;
    if ( mark_as_used )
    { 
      tsp00_Int4 i=0, bl_to_mark = *offset;
      for ( ; i < block_cnt && *tr_err == e_ok ; i++, bl_to_mark++ )
      { 
        int fbm_pno = bl_to_mark < cbd_maxfbmlist_entries
                      ? devsp->fst_fbm_page
                      : devsp->fst_fbm_page + bl_to_mark /
                                                  cbd_maxfbmlist_entries ;
        tsp00_Int4 offs_in_fbm_list = bl_to_mark % cbd_maxfbmlist_entries;
        tbd_fbm *fbm_list = BD11_GET_LIST ( t, fbm_pno, tr_err ) ;
        if ( *tr_err == e_ok )
        { 
          DBG3((MF__,"mark_block: dev / offs / mark = (%d/%d/%d)",
                      devno,  bl_to_mark, fbm_occupied )) ;
          MARK_BLOCK ( fbm_list ,bl_to_mark, fbm_occupied ) ;
        }
      }
    if ( *tr_err != e_ok )
      vabort (TRUE) ;
    }
  }
  return ;
}

/*-----------------------------------------------------------*/
externC void b11_check_fbm_balance ( tgg00_TransContext      *t, 
                                     tsp00_Int4        devsp_compression,
                                     tsp00_Int4        devsp_balance, 
                                     boolean          *is_fragmented, 
                                     tgg00_BasisError *tr_err)
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"b11_check_fbm_balance"

  float cmpr_factor ;
  tsp00_Int4 change_devsp_usage = FALSE ;
  tsp00_Int4 devsp_used, devno, check_allowed ;
 
  cccprint("BD11  b11_balanc" ) ;

  *is_fragmented = FALSE ;
  if ( devsp_compression <= 0 )
    return ;

  cmpr_factor = (100.0 / (float )devsp_compression) ;
 
  DBG3((MF__,"compr %d, bal %d", devsp_compression, devsp_balance )) ;
  t01p2int4 (TRACE_BD_FBM, "compression ", devsp_compression,
                           "balance     ", devsp_balance ) ;
 
  vbegexcl ( PID(t) , FBM_REGION ) ;

  for ( devno = FIRST_DEVNO, check_allowed = TRUE ;
        devno <= FBM_GLOB perm_devsp_used && check_allowed ; devno++ )
    check_allowed = FBM_GLOB devsp_info [ devno ] . num_marked_b_up == 0 ;

  if ( check_allowed )
  { 
    tsp00_Int4 avg_used_per_dev = bd11_get_avg_for_balance () ;
    tsp00_Int4 max_diff_per_dev = avg_used_per_dev * (devsp_balance>>1) / 100;
    tsp00_Int4 max_allowed_per_dev = avg_used_per_dev + max_diff_per_dev ;
    FBM_DEV_INFO   *devsp ;
    DBG3((MF__,"avg:%d, max: %d", avg_used_per_dev, max_allowed_per_dev));
 
    for ( devno = FIRST_DEVNO;
          devno <= FBM_GLOB perm_devsp_used && *tr_err == e_ok ; devno++ )
    {
      t01int4 (TRACE_BD_FBM, "srch_devno  ", devno) ;
      devsp = & FBM_GLOB devsp_info [ devno ] ;
      devsp_used = devsp->used_blocks - devsp->num_marked_svp_free;
      DBG3((MF__,"devno : %d, used: %d", devno, devsp_used )) ;
      if ( devsp_used > max_allowed_per_dev )
      { 
        tsp00_Int4 num_to_mark = devsp_used - avg_used_per_dev ;
        if ( num_to_mark > MAX_MULTIPLE_BLOCKS )
        { /* !!!!  take blocks from this dev !!!! */
          DBG3((MF__,"dev %d , to_mark %d", devno, num_to_mark )) ;
          bd11_mark_frag_blocks ( devsp, 0, num_to_mark, tr_err ) ;
          if ( ! *is_fragmented )
            *is_fragmented = devsp->num_marked_b_up > 0 ;
          change_devsp_usage = TRUE ;
        }
      }
      else
      { 
        if ( devsp->used_blocks > 0 &&(devsp->lst_used_offs >
                  (tsp00_Int4 ) ((float )devsp->used_blocks * cmpr_factor )))
        { 
          DBG3((MF__,"compression for dev %d", devno )) ;
          bd11_mark_frag_blocks ( devsp, devsp->used_blocks, devsp->dev_size,tr_err ) ;
          if ( ! *is_fragmented )
            *is_fragmented = devsp->num_marked_b_up > 0 ;
        }
      }
    }
    if ( change_devsp_usage )
      bd11_change_devsp_usage (t, tr_err);
  }
  vendexcl ( PID(t) , FBM_REGION ) ;
}

/*-----------------------------------------------------------*/
externC void b11_get_last_frag_block ( tgg00_TransContext      *t, 
                                       tsp00_Int2       *devno, 
                                       tsp00_Int4       *offset, 
                                       tsp00_Int4       *free_blocks, 
                                       tgg00_BasisError *tr_err )
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"b11_get_last_frag_block"

  tsp00_Int2 srch_devno, beg_devno ;
  tsp00_Int2 found = FALSE ;
  cccprint("BD11  b11_get_la" ) ;
  vbegexcl ( PID(t) , FBM_REGION ) ;
 
  if ( FBM_GLOB last_devsp_used_for_frag < FIRST_DEVNO ||
       FBM_GLOB last_devsp_used_for_frag >= FBM_GLOB perm_devsp_used )
    beg_devno = srch_devno = FIRST_DEVNO ;
  else
    beg_devno = srch_devno = FBM_GLOB last_devsp_used_for_frag + 1 ;
 
  do
  {
    DBG3((MF__,"srch_dev %d", srch_devno ));
    if ( ! (found = bd11_get_last_frag_from_dev ( t, srch_devno,
                                                  offset, tr_err )) )
    if ( ++srch_devno > FBM_GLOB perm_devsp_used )
      srch_devno = FIRST_DEVNO ;
  }
  while ( !found &&  beg_devno != srch_devno ) ;
 
  if ( found )
  { 
    FBM_GLOB last_devsp_used_for_frag = srch_devno ;
    *devno = srch_devno ;
    DBG3((MF__,"dev: %d, offs: %d", *devno, *offset));
  }
  else
  { 
    FBM_GLOB last_devsp_used_for_frag = -1 ;
    *devno = -1 ;
    DBG3((MF__,"no block found"));
  }

  *free_blocks = FBM_GLOB tot_free_blocks ;
  vendexcl ( PID(t) , FBM_REGION ) ;
}
 
 
/*-----------------------------------------------------------*/
externC void b11_is_block_in_use ( tgg00_TransContext      *t, 
                                   tsp00_Int2        devno, 
                                   tsp00_Int4        offset, 
                                   boolean          *is_in_use, 
                                   tgg00_BasisError *tr_err )
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"b11_is_block_in_use"

  FBM_DEV_INFO *devsp = & FBM_GLOB devsp_info [ devno ];
  cccprint("BD11  b11_is_blo" ) ;

  *tr_err = e_ok ;
  *is_in_use = FALSE ;

  vbegexcl ( PID(t) , FBM_REGION ) ;

  if ( VALID_DEVNO ( devno ) && VALID_OFFSET( offset, devsp->dev_size ) )
  { 
    int fbm_pno = offset < cbd_maxfbmlist_entries ? devsp->fst_fbm_page
                 : devsp->fst_fbm_page + offset / cbd_maxfbmlist_entries ;
    tsp00_Int4 offs_in_fbm_list = (offset % cbd_maxfbmlist_entries);
    tbd_fbm *fbm_list = BD11_GET_LIST ( t, fbm_pno, tr_err ) ;
    if ( *tr_err == e_ok )
      *is_in_use = ACT_BLOCK_VAL( fbm_list, offs_in_fbm_list ) == fbm_backup &&
                   VAL_AFTER_BACKUP( fbm_list, offs_in_fbm_list) == fbm_occupied;

    DBG3((MF__,"block %d/%d is %sin use", 
               devno, offset, *is_in_use ? "" :"NOT "));
  }
  else
  {
    *tr_err = e_illegal_page_address ;
    MSGD ((ERR_INVALID_DEVNO_OR_OFFS, devno,  offset )) ;
  }
 
  vendexcl ( PID(t) , FBM_REGION ) ;
  return ;
}
 
/*-----------------------------------------------------------*/
externC void b11check_b_up ( tsp00_Int2  devno, 
                             tsp00_Int4 *num_b_up_bl, 
                             tsp00_Int4 *num_b_up_pg )
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"b11check_b_up"

  int num_pg_marked = FBM_GLOB bup_pno_list.num_marked ;
  int max_err_pg_to_diag = num_pg_marked > 100 ? 100 : num_pg_marked ;
  cccprint("BD11  b11check_b" ) ;
  *num_b_up_bl = FBM_GLOB devsp_info [ devno ] . num_marked_b_up ;
  *num_b_up_pg = num_pg_marked ;
  t01p2int4 (TRACE_BD_FBM, "devno       ", devno,
                           "num_b_up_blo", *num_b_up_bl );
  if ( devno == 1 && num_pg_marked > 0 )
  {
    long i ;
    MSGD ((ERR_NUM_UNSAVED_PAGES , num_pg_marked)) ;

    for ( i=0;
          i < FBM_GLOB bup_pno_list.list_size && max_err_pg_to_diag > 0; i++ )
    if ( FBM_GLOB bup_pno_list.pnos [i] )
    {
      int bit;
      for ( bit=0; bit<8; bit++ )
        if ( FBM_GLOB bup_pno_list.pnos [i] & (1 << bit ) )
        { 
          max_err_pg_to_diag-- ;
          MSGD ((ERR_UNSAVED_PNO , (i << 3) + bit ) ) ;
        }
    }
  }

  FBM_GLOB devsp_info [ devno ] . act_bup_pos.fbm_page = 
                      FBM_GLOB devsp_info [ devno ] . fst_fbm_page ;
  FBM_GLOB devsp_info [ devno ] . act_bup_pos.fbm_offs = 0 ;

  DBG3((MF__,"devno(%d), marked bl %d / pg %d",
              devno, *num_b_up_bl, *num_b_up_pg)) ;
  return ;
}
 
 
/*-----------------------------------------------------------*/
externC void b11_clear_backup ( tgg00_TransContext      *t, 
                                tgg00_BasisError *tr_err )
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"b11_clear_backup"

  int devno ;
  cccprint("BD11  b11clear_b" ) ;
  *tr_err = e_ok ;

  vbegexcl ( PID(t) , FBM_REGION ) ;
 
  if ( FBM_GLOB bup_pno_list.num_marked > 0 )
  {
    DBG3((MF__,"pno_list: marked pg %d", FBM_GLOB bup_pno_list.num_marked));
    bd11_init_bup_struc( &(FBM_GLOB bup_pno_list) ) ;
  }

  if ( FBM_GLOB save_pg_pno_list.num_marked > 0 )
  {
    DBG3((MF__,"sv_pg_pno_list: marked pg %d", FBM_GLOB save_pg_pno_list.num_marked));
    bd11_init_bup_struc( &(FBM_GLOB save_pg_pno_list) ) ;
  }
 
  for ( devno = FIRST_DEVNO;
        devno <= FBM_GLOB perm_devsp_used && *tr_err == e_ok ; devno++ )
  {
    FBM_DEV_INFO *devsp = & FBM_GLOB devsp_info [ devno ] ;
    if ( FBM_GLOB devsp_info [ devno ] . num_marked_b_up > 0)
    { 
      tsp00_PageNo fbm;
      tbd_fbm *act_fbm_list ;
      tsp00_Int4 ind, last_offs ;
      DBG3((MF__,"clr_bup: devno : %d, num_bup %d", devno, devsp->num_marked_b_up )) ;
      t01p2int4 (TRACE_BD_FBM, "srch_devno  ", devno,
                               "num_to_clear", devsp->num_marked_b_up) ;
      for ( fbm =  devsp->fst_fbm_page;
            fbm <= devsp->lst_used.fbm_page && *tr_err == e_ok ; fbm++ )
        if ( IS_FBM_PAGE_DIRTY ( fbm , fbm_backup ) )
        { 
          tsp00_Int4 base_offs = ( fbm - devsp->fst_fbm_page ) * cbd_maxfbmlist_entries ;
          DBG3((MF__,"fbm_pno : %d", fbm )) ;
          CLEAR_FLAG_IN_DIRTY_LIST ( fbm, fbm_backup ) ;
          if ( (act_fbm_list = BD11_GET_LIST ( t, fbm, tr_err ) ) != NULL )
          { 
            last_offs = ( fbm < devsp->lst_used.fbm_page
                      ? cbd_maxfbmlist_entries - 1
                      : devsp->lst_used.fbm_offs );
            for ( ind = 0; ind <= last_offs && devsp->num_marked_b_up > 0 ;
                  ind ++ )
              if ( ACT_BLOCK_VAL ( act_fbm_list , ind ) == fbm_backup )
                bd11_restore_backup_marker ( devsp, base_offs+ind, fbm,
                                             act_fbm_list, ind ) ;
           }
        }

        DBG3((MF__,"devno : %d, num_bup %d", devno ,
                    devsp->num_marked_b_up )) ;
    }
    devsp->act_bup_pos.fbm_page = devsp->fst_fbm_page ;
    devsp->act_bup_pos.fbm_offs = 0 ;
  }
 
  vendexcl ( PID(t) , FBM_REGION ) ;
  return ;
}
 
 
/*-----------------------------------------------------------*/
externC void b11_bup_pno_ready ( tgg00_TransContext      *t, 
                                 tsp00_PageNo      pno, 
                                 tsp00_Int2        devno, 
                                 tsp00_Int4        offset, 
                                 tgg00_BasisError *tr_err )
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"b11_bup_pno_ready"

  FBM_DEV_INFO   *devsp = & FBM_GLOB devsp_info [ devno ] ;
  cccprint("BD11  b11_bup_pn" ) ;
  t01p2int4 (TRACE_BD_FBM, "devno       ", devno,
                           "offset      ", offset ) ;
  t01int4   (TRACE_BD_FBM, "pno         ", pno    ) ;
  *tr_err = e_ok ;
 
  DBG3((MF__,"pno %6d at %2d/%5d", pno, devno,  offset)) ;

  vbegexcl ( PID(t) , FBM_REGION ) ;

  if ( VALID_DEVNO ( devno ) && VALID_OFFSET( offset, devsp->dev_size ) )
  { 
    int fbm_pno = offset < cbd_maxfbmlist_entries ? devsp->fst_fbm_page
                 : devsp->fst_fbm_page + offset / cbd_maxfbmlist_entries ;
    tsp00_Int4 offs_in_fbm_list = (offset % cbd_maxfbmlist_entries);
    tbd_fbm *fbm_list = BD11_GET_LIST ( t, fbm_pno, tr_err ) ;
    if ( pno == NIL_PAGE_NO_GG00 || 
         IS_BACKUP_PNO_FLAG_SET(FBM_GLOB bup_pno_list.pnos, pno ) )
    { 
      if ( ACT_BLOCK_VAL ( fbm_list, offs_in_fbm_list ) == fbm_backup )
      {
        bd11_restore_backup_marker ( devsp, offset, fbm_pno, fbm_list,
                                     offs_in_fbm_list ) ;
        if ( pno != NIL_PAGE_NO_GG00 )
          CLEAR_FLAG_IN_BACKUP_LIST( FBM_GLOB bup_pno_list, pno ) ;
      }
      else
      { 
        *tr_err = e_invalid_fbm_mark ;
        MSGD ((ERR_PNO_NOT_MARKED_BUP , pno, devno,  offset )) ;
      }
    }
    else
    { 
      *tr_err = e_bad_datapage ;
      MSGD ((ERR_PNO_NOT_BUP_PNO_LIST , pno )) ;
    }
  }
  else
  {
    *tr_err = e_illegal_page_address ;
     MSGD ((ERR_INVALID_DEVNO_OR_OFFS,  devno,  offset )) ;
  }
  vendexcl ( PID(t) , FBM_REGION ) ;
  return ;
}

/*-----------------------------------------------------------*/
externC void b11backup_block ( tgg00_TransContext      *t, 
                               tsp00_Int2        devno, 
                               tsp00_Int4        max_num, 
                               tsp00_Int4       *num, 
                               tsp00_Int4       *nxt_offset, 
                               tgg00_BasisError *tr_err )
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"b11backup_block"

  cccprint("BD11  b11backup_" ) ;
  t01int4 (TRACE_BD_FBM, "devno       ", devno ) ;
  DBG3((MF__,"devno(%d)", devno )) ;
  *tr_err = e_ok ;
  *num    = 0 ;

  if ( VALID_DEVNO ( devno ) )
  { 
    FBM_DEV_INFO   *devsp = & FBM_GLOB devsp_info [ devno ] ;

    vbegexcl ( PID(t) , FBM_REGION ) ;

    if ( devsp->act_bup_pos.fbm_offs >= 0 &&
         bd11_nxt_b_up_pos ( t, devsp, nxt_offset, tr_err ) )
    { 
      tsp00_Int4 abs_offset = *nxt_offset ;
      int fbm_pno = devsp->act_bup_pos.fbm_page ;
      tsp00_Int4 offs_in_fbm_list = devsp->act_bup_pos.fbm_offs ;
      int end = FALSE ;
      tbd_fbm *fbm_list = BD11_GET_LIST ( t, fbm_pno, tr_err ) ;

      while ( *tr_err == e_ok && !end && *num < max_num &&
              ACT_BLOCK_VAL ( fbm_list, offs_in_fbm_list ) == fbm_backup )
      {
        (*num)++ ;
        if ( ! (end = ++abs_offset > devsp->lst_used_offs ) )
          if ( ++offs_in_fbm_list == cbd_maxfbmlist_entries )
          { 
            offs_in_fbm_list = 0 ;
            fbm_list = BD11_GET_LIST ( t, ++fbm_pno, tr_err ) ;
          }
      }
 
      if ( *num > 0 && *tr_err == e_ok && !end )
      {
        devsp->act_bup_pos.fbm_page = fbm_pno ;
        devsp->act_bup_pos.fbm_offs = offs_in_fbm_list ;
      }
      else
        devsp->act_bup_pos.fbm_offs = -1 ;
    }
    vendexcl ( PID(t) , FBM_REGION ) ;
  }
  else
  { 
    *tr_err = e_illegal_page_address ;
    MSGD (( ERR_INVALID_DEVNO , devno, 
                                 FIRST_DEVNO, FBM_GLOB perm_devsp_used )) ;
  }

  DBG3((MF__,"num_found(%d), nxt_offs(%d)", *num, *nxt_offset )) ;
  t01p2int4 (TRACE_BD_FBM, "num_found   ", *num,
                           "nxt_offset  ", *nxt_offset ) ;
  return ;
}

/*-----------------------------------------------------------*/
externC void b11_block_b_up_finish ( tgg00_TransContext      *t, 
                                     tsp00_Int2        devno, 
                                     tsp00_Int4        offset, 
                                     tbd_fbm           svp_mark, 
                                     tgg00_BasisError *tr_err )
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"b11_block_b_up_finish"

  FBM_DEV_INFO   *devsp = & FBM_GLOB devsp_info [ devno ] ;
  cccprint("BD11  b11_block_" ) ;
  t01p2int4 (TRACE_BD_FBM, "devno       ", devno,
                           "offset      ", offset ) ;
  t01int4 (TRACE_BD_FBM, "svp_mark    ", svp_mark) ;
  *tr_err = e_ok ;
  DBG3((MF__,"dev / offs / svp_mark = (%d/%d/%d)",
              devno,  offset,  svp_mark  )) ;

  vbegexcl ( PID(t) , FBM_REGION ) ;

  if ( VALID_DEVNO ( devno ) && VALID_OFFSET( offset, devsp->dev_size ) )
  {
    int fbm_pno = offset < cbd_maxfbmlist_entries ? devsp->fst_fbm_page
                : devsp->fst_fbm_page + offset / cbd_maxfbmlist_entries ;
    tsp00_Int4 offs_in_fbm_list = offset % cbd_maxfbmlist_entries;
    tbd_fbm *fbm_list = BD11_GET_LIST ( t, fbm_pno, tr_err ) ;
    if ( *tr_err == e_ok &&
         ACT_BLOCK_VAL ( fbm_list, offs_in_fbm_list ) == fbm_backup)
    {
      MARK_BLOCK ( fbm_list ,offs_in_fbm_list, svp_mark ) ;
      SET_FLAG_IN_DIRTY_LIST ( fbm_pno, svp_mark ) ;
    }
  }
  else
  {
    *tr_err = e_illegal_page_address ;
    MSGD (( ERR_INVALID_DEVNO_OR_OFFS , devno,  offset )) ;
  }
  vendexcl ( PID(t) , FBM_REGION ) ;
  return ;
}


/*
 * ========================== LOCAL FUNCTIONS =================================
 */
 
/*-----------------------------------------------------------*/
_INLINE
static tsp00_Int4 bd11_get_avg_for_balance ()
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"bd11_get_avg_for_balance"

  tsp00_Int4 total_used  = FBM_GLOB tot_blocks -
                         FBM_GLOB tot_free_blocks -
                         FBM_GLOB tot_svp_blocks ;
  tsp00_Int4 avg_used_per_dev;
  tsp00_Int2 devsp_for_balance = FBM_GLOB perm_devsp_used , devno,
           sv_dev_for_bal   ;
  FBM_DEV_INFO   *devsp ;
 
  cccprint("BD11  bd11_get_a" ) ;
 
  do
  {
    sv_dev_for_bal   = devsp_for_balance ;
    avg_used_per_dev = total_used / sv_dev_for_bal ;
    DBG3((MF__,"avg_for_bal : %d devs for avg : %d",
                sv_dev_for_bal, avg_used_per_dev )) ;
    t01p2int4 (TRACE_BD_FBM, "devs_to_chck", sv_dev_for_bal,
                             "avg_per_dev ", avg_used_per_dev ) ;
    for ( devno = FIRST_DEVNO;
          devno <= FBM_GLOB perm_devsp_used ; devno++ )
    {
      devsp = & FBM_GLOB devsp_info [ devno ] ;
      if ( devsp->act_blocks_per_usage != -devsp->dev_size )
        if ( devsp->dev_size < avg_used_per_dev )
        { 
          if ( devsp->used_blocks < devsp->dev_size )
          { 
            tsp00_Int4 new_blocks = devsp->dev_size - devsp->used_blocks ;
            total_used -= devsp->dev_size ;
            DBG3((MF__,"dev %d too small (%d)",devno, devsp->dev_size )) ;
          }
          else
            DBG3((MF__,"dev %d full!", devno )) ;
          /* !!!! don't look for this dev again !!!! */
          devsp->act_blocks_per_usage = -devsp->dev_size ; /* NOT THIS */
          devsp_for_balance-- ;
        }
    }
  }
  while ( devsp_for_balance != sv_dev_for_bal ) ;

  if ( devsp_for_balance != FBM_GLOB perm_devsp_used ) /* RESET !! */
    for ( devno = FIRST_DEVNO; devno <= FBM_GLOB perm_devsp_used ; devno++ )
    { 
      devsp = & FBM_GLOB devsp_info [ devno ] ;
      if ( devsp->act_blocks_per_usage == -devsp->dev_size )
        devsp->act_blocks_per_usage = devsp->tot_blocks_per_usage ;
    }

  DBG3((MF__,"avg_used_per_dev: %d ", avg_used_per_dev )) ;
  t01int4 (TRACE_BD_FBM, "avg_per_dev ", avg_used_per_dev ) ;
  return ( avg_used_per_dev ) ;
}
 
 
/*-----------------------------------------------------------*/
_INLINE
static void bd11_mark_frag_blocks ( FBM_DEV_INFO     *devsp , 
                                    tsp00_Int4        srch_end_pos, 
                                    tsp00_Int4        max_to_mark, 
                                    tgg00_BasisError *tr_err )
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"bd11_mark_frag_blocks"

  tsp00_Int4 srch_end_offs, offs, last_offs ;
  tsp00_PageNo srch_end_fbm, fbm;
  tbd_fbm *act_fbm_list ;
  cccprint("BD11  bd11_mark_f" ) ;
 
  if ( srch_end_pos > 0 )
  {
    srch_end_fbm  = devsp->fst_fbm_page + srch_end_pos / cbd_maxfbmlist_entries;
    srch_end_offs = srch_end_pos % cbd_maxfbmlist_entries ;
    if ( srch_end_offs > 0 )  srch_end_offs--;
    else                      srch_end_fbm-- ;
  }
  else
  { 
    srch_end_fbm = devsp->fst_fbm_page ;
    srch_end_offs= 0 ;
  }
 
  devsp->lst_frag.fbm_page = fbm = devsp->lst_used.fbm_page ;
  devsp->lst_frag.fbm_offs = offs= devsp->lst_used.fbm_offs ;
 
  DBG3((MF__,"mark_frag_bl: srch_beg (%d/%d)", fbm, offs )) ;
  DBG3((MF__,"mark_frag_bl: srch_end (%d/%d)", srch_end_fbm, srch_end_offs));
  do
  {
    last_offs = fbm == srch_end_fbm ? srch_end_offs : 0 ;
    if ( (act_fbm_list = BD11_GET_LIST ( t, fbm, tr_err ) ) != NULL )
      while ( offs >= last_offs && max_to_mark != 0 )
      {
        if ( ACT_BLOCK_VAL ( act_fbm_list, offs ) == fbm_occupied )
        { 
          devsp->num_marked_b_up++ ;
          max_to_mark-- ;
          MARK_AFTER_BACKUP( act_fbm_list, offs, fbm_occupied);
          MARK_BLOCK ( act_fbm_list ,offs, fbm_backup) ;
          SET_FLAG_IN_DIRTY_LIST ( fbm, fbm_backup ) ;
          DBG3((MF__,"mark (%d/%d)", fbm, offs));
          devsp->fst_frag.fbm_page = fbm ;
          devsp->fst_frag.fbm_offs = offs ;
        }
      offs-- ;
      }
    offs = cbd_maxfbmlist_entries - 1;
    fbm-- ;
  }
  while ( fbm > srch_end_fbm && max_to_mark != 0 && *tr_err == e_ok ) ;
}
 
/*-----------------------------------------------------------*/
_INLINE
static int bd11_get_last_frag_from_dev ( tgg00_TransContext      *t, 
                                         tsp00_Int2        devno, 
                                         tsp00_Int4       *offset, 
                                         tgg00_BasisError *tr_err )
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"bd11_get_last_frag_from_dev"

  FBM_DEV_INFO *devsp = & FBM_GLOB devsp_info [ devno ];
  tsp00_PageNo   fbm_pno  ;
  tsp00_Int4      offs     ;
  tsp00_Int2      found = FALSE ;
 
  if ( devsp->num_marked_b_up > 0 && devsp->lst_frag.fbm_offs >= 0 )
  { 
    tsp00_Int4  last_offs ;
    tbd_fbm  *act_fbm_list ;

    fbm_pno  = devsp->lst_frag.fbm_page ;
    offs     = devsp->lst_frag.fbm_offs ;
 
    while ( fbm_pno >= devsp->fst_frag.fbm_page && !found )
    {
      DBG3((MF__,"srch_beg %d/%d", fbm_pno, offs ));
      act_fbm_list = BD11_GET_LIST ( t, fbm_pno, tr_err);
      last_offs = fbm_pno <= devsp->fst_frag.fbm_page
                  ? devsp->fst_frag.fbm_offs : 0 ;
 
      while ( offs >= last_offs && !found )
        if (! (found = ACT_BLOCK_VAL ( act_fbm_list, offs ) == fbm_backup))
          offs-- ;
 
      if ( !found )
      { 
        --fbm_pno ;
        offs = cbd_maxfbmlist_entries - 1 ;
      }
    }
  }
 
  if ( found )
  { 
    *offset = (fbm_pno - devsp->fst_fbm_page ) * cbd_maxfbmlist_entries +
               offs ;
    if ( offs > 0 )
    { 
      devsp->lst_frag.fbm_offs = offs - 1 ;
      devsp->lst_frag.fbm_page = fbm_pno ;
    }
    else
     if ( fbm_pno > devsp->fst_frag.fbm_page )
     { 
       devsp->lst_frag.fbm_offs = cbd_maxfbmlist_entries - 1 ;
       devsp->lst_frag.fbm_page = fbm_pno - 1 ;
     }
     else
       devsp->lst_frag.fbm_offs = -1 ;
    DBG3((MF__,"found: %d/%d", devsp->lst_frag.fbm_page, 
                               devsp->lst_frag.fbm_offs ));
  }
  else
  {
    *offset = -1 ;
    devsp->lst_frag.fbm_offs = -1 ;
    DBG3((MF__,"nothing found"));
  }
  return ( found ) ;
}
 
/*-----------------------------------------------------------*/
_INLINE
static void bd11_restore_backup_marker ( FBM_DEV_INFO    *devsp,
                                         tsp00_Int4       offset,
                                         tsp00_PageNo     fbm_pno,
                                         tbd_fbm         *fbm_list, 
                                         tsp00_Int4       offs_in_fbm_list )
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"bd11_restore_backup_marker"

  tbd_fbm new_marker = VAL_AFTER_BACKUP( fbm_list, offs_in_fbm_list);
  cccprint("BD11  bd11_resto" ) ;

  switch ( new_marker )
  { 
    case fbm_free      : devsp->used_blocks-- ;
                         FBM_GLOB tot_free_blocks++ ;
                         if ( devsp->fst_free_offs > offset )
                            SAVE_DEV_FIRST_FREE (devsp, offset, fbm_pno,
                                                 offs_in_fbm_list ) ;
                           break ;
    case fbm_svp_free  : FBM_GLOB tot_svp_blocks++ ;
                         devsp->num_marked_svp_free++ ;
                         break ;
  }
 
  MARK_BLOCK ( fbm_list ,offs_in_fbm_list, new_marker ) ;
  DBG3 ((MF__," (%d) new marker: %d", offs_in_fbm_list, new_marker ));
  SET_FLAG_IN_DIRTY_LIST ( fbm_pno, new_marker ) ;
  devsp->num_marked_b_up-- ;
 
  return ;
}
 
/*-----------------------------------------------------------*/
_INLINE
static int bd11_nxt_b_up_pos ( tgg00_TransContext      *t, 
                               FBM_DEV_INFO     *devsp, 
                               tsp00_Int4       *found_fbm_offs, 
                               tgg00_BasisError *tr_err )
/*-----------------------------------------------------------*/
{
#undef MF__
#define MF__   MOD__"bd11_nxt_b_up_pos"

  tsp00_Int4    last_offs, found = FALSE;
  tsp00_PageNo fbm = devsp->act_bup_pos.fbm_page;
  tsp00_Int4    ind = devsp->act_bup_pos.fbm_offs;
  tbd_fbm    *act_fbm_list ;
  cccprint("BD11  bd11_nxt_b_" ) ;
  DBG3((MF__,"srch_beg: %d/%d", fbm, ind )) ;
    
  for ( ;!found && fbm <= devsp->lst_used.fbm_page && *tr_err == e_ok ; fbm++ )
    if ( IS_FBM_PAGE_DIRTY ( fbm , fbm_backup ) )
      if ( (act_fbm_list = BD11_GET_LIST ( t, fbm, tr_err ) ) != NULL )
      { 
        last_offs = fbm < devsp->lst_used.fbm_page
                  ? cbd_maxfbmlist_entries - 1
                  : devsp->lst_used.fbm_offs ;
        for ( ; ind <= last_offs && !found ; ind ++ )
          if ( ACT_BLOCK_VAL ( act_fbm_list, ind ) == fbm_backup)
          { 
            found = TRUE ;
            *found_fbm_offs = ind + cbd_maxfbmlist_entries *
                                    ( fbm - devsp->fst_fbm_page ) ;
            devsp->act_bup_pos.fbm_page = fbm ;
            devsp->act_bup_pos.fbm_offs = ind ;               
          }
        ind = 0 ;
        }
     
  DBG3 ((MF__,"found:%d (%d/%d)", found, fbm, ind )) ;
  return ( found ) ;
}


/*
 * =============================== END ========================================
 */
