/*!**************************************************************************

  module:       vbd22.cpp
  special area: OMS_KeyManagement
  responsible:  UweH
  last changed: 1999-11-05  17:00
  version:      7.2.0.0
  see also: 
  copyright:    Copyright by SAP AG, 1998
  description:  Map an object key to an OID, there is no consistent-check done.



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "hbd22.h"

#include "ggg12.h"
#include "gbd900.h"
#include "gsp03_3.h"

#include "hbd01.h"
#include "hbd03.h"
#include "hbd12.h"
#include "hbd17.h" 
#include "hbd92.h"
#include "hgg01.h"
#include "hgg01_1.h"
#include "hgg10.h"

#if COMPILEMODE_MEO00 >= SLOW_MEO00 
#include "hta99.h"
#endif

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#ifdef NO_INLINES
#  define _INLINE
#else
#  define _INLINE   inline
#endif

/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/

extern cbd900_ObjFileDir  bd90ObjFileDir;

/*===========================================================================*
 *  LOCAL VARIABLES                                                          *
 *===========================================================================*/

/*===========================================================================*
 *  LOCAL FUNCTIONS (DECLARATION)                                            *
 *===========================================================================*/

inline static void 
bd22_GetOId (tgg00_TransContext   &Trans,
			 cbd900_ObjKeyFileId   ObjKeyFileId,
			 tsp00_Int4            KeyLen,
			 void                 *pKey,
			 tgg92_KernelOid      &Oid);

/*===========================================================================*
 *  GLOBAL FUNCTIONS (CODE)                                                  *
 *===========================================================================*/

externCpp void 
bd22CreateObjKeyFile (tgg00_TransContext &Trans,
                      tgg00_ObjFileNo     ObjFileNo)
{
    ROUTINE_DBG_MEO00 ("bd22CreateObjKeyFile");

	cbd900_ObjFileInfo    &ObjFileInfo = bd90ObjFileDir [ObjFileNo];
    cbd900_ObjKeyFileId    ObjKeyFileId (ObjFileInfo, ObjFileNo);

    b01fcreate_file (Trans, ObjKeyFileId, 0, cgg_nil_varcol_cnt_off, false);

	/* store root pno in the object file info */
    if (e_ok == Trans.trError_gg00)
		ObjFileInfo.ofiKeyRoot_bd900 = ObjKeyFileId.fileRoot_gg00();
    else if (e_duplicate_filename == Trans.trError_gg00)
	{
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
			"BD22CreKeyFile:Dupl File", ObjFileNo);
		Trans.trError_gg00 = e_wrong_class_id;
	}
}

/*-----------------------------------------------------------------------------------*/
/* PTS 1107153 AK 17/07/2000 */

externCpp void
bd22GetLeaveCountOfObjKeyFile (tgg00_TransContext  &Trans,
							   tgg00_ObjFileNo      ObjFileNo,
							   cbd900_ObjFileInfo  &ObjFileInfo,
							   tsp00_Int4          &NumKeyLeaves)
{
	const tsp00_Bool       cCountLeavesOnly = true;
	tsp00_Int4             DummyAvgPrimKeyCountPerPage;
	cbd900_ObjKeyFileId    ObjKeyFileId (ObjFileInfo, ObjFileNo);
	
	bd03StatisticsUpdate (Trans, ObjKeyFileId, NumKeyLeaves,
		DummyAvgPrimKeyCountPerPage, cCountLeavesOnly);
}

/*-----------------------------------------------------------------------------------*/

externCpp void
bd22ReleaseObjKeyFile (tgg00_TransContext &Trans,
                       tgg00_ObjFileNo     ObjFileNo)
{
	ROUTINE_DBG_MEO00 ("bd22ReleaseObjKeyFile");
	
	cbd900_ObjFileInfo    &ObjFileInfo = bd90ObjFileDir [ObjFileNo];
    cbd900_ObjKeyFileId    ObjKeyFileId (ObjFileInfo, ObjFileNo);
		
    b01pdestroy_perm_file (Trans, ObjKeyFileId);
	
    if (e_ok == Trans.trError_gg00)
		ObjFileInfo.ofiKeyRoot_bd900 = NIL_PAGE_NO_GG00;
}
             
/*-----------------------------------------------------------------------------------*/

externCpp void 
bd22GetOId (tgg00_TransContext  &Trans,
			tgg00_ObjFileNo      ObjFileNo,
			tsp00_Int4           KeyLen,
			void                *pKey,
			tgg92_KernelOid     &Oid)
{
    ROUTINE_DBG_MEO00 ("bd22GetOid");
	
	cbd900_ObjKeyFileId  ObjKeyFileId (bd90ObjFileDir [ObjFileNo], ObjFileNo);
	bd22_GetOId (Trans, ObjKeyFileId , KeyLen, pKey, Oid);		
}

/*-----------------------------------------------------------------------------------*/

externCpp void 
bd22InsertOId (tgg00_TransContext &Trans,
			   tgg00_ObjFileNo     ObjFileNo,
			   tsp00_Int4          KeyLen,
			   void               *pKey,
			   tgg92_KernelOid     OId,
			   tgg92_KernelOid    &FoundOId)
{
    ROUTINE_DBG_MEO00 ("bd22InsertOId");

    tgg00_BasisError &TrError   = Trans.trError_gg00;

    if ( TrError != e_ok) return;

    cbd900_ObjKeyFileId  ObjKeyFileId (bd90ObjFileDir [ObjFileNo], ObjFileNo);

    tsp00_Int4 FoundOIdLen;
    bd03OmsAddInv (Trans, ObjKeyFileId,
        tsp00_BytePtr (pKey), KeyLen, tsp00_BytePtr(&OId), sizeof(OId),
        sizeof(FoundOId), tsp00_BytePtr(&FoundOId), FoundOIdLen);

	switch (TrError)
	{
	case e_duplicate_invkey:
	case e_duplicate_key   : TrError = e_duplicate_hash_key; break;
    case e_inv_not_found   : TrError = e_hash_not_found;     break;
	}
}

/*-----------------------------------------------------------------------------------*/

externCpp void 
bd22RemoveOid (tgg00_TransContext &Trans,
               tgg00_ObjFileNo     ObjFileNo,
               tsp00_Int4          KeyLen,
               void               *pKey,
               tgg92_KernelOid     Oid)
{
    ROUTINE_DBG_MEO00 ("bd22RemoveOid");

    tgg00_BasisError    &TrError = Trans.trError_gg00;
    tgg92_KernelOid      OidOfKey;
    cbd900_ObjKeyFileId  ObjKeyFileId (bd90ObjFileDir [ObjFileNo], ObjFileNo);

    bd22_GetOId (Trans, ObjKeyFileId, KeyLen, pKey, OidOfKey);
    
    if (TrError != e_ok) return;
    
    if (! Oid.gg92IsEqIgnoreFrameVers (OidOfKey)) 
    {
        if (g01vtrace.vtrOmsUpd_gg00)
        {
            tgg12_OmsOidTrace TrLineOId;
            
            TrLineOId.ooidTrType_gg12.becomes(otOid_egg12);
            TrLineOId.ooidTrType2_gg12.becomes(ot2Input_egg12);
            TrLineOId.ooidObjId_gg12   = Oid;
            b12vtrace (Trans, bd, b22rem_oid, sizeof (TrLineOId), &TrLineOId);
    
            TrLineOId.ooidObjId_gg12   = OidOfKey;
            TrLineOId.ooidTrType2_gg12.becomes(ot2Output_egg12);

            b12vtrace (Trans, bd, b22rem_oid, sizeof (TrLineOId), &TrLineOId);

            bd92VtraceObjKey (Trans, ObjFileNo, KeyLen, pKey, b22rem_oid);
        }
        
        return;
    }

    bd03OmsDelInv (Trans, ObjKeyFileId, tsp00_BytePtr (pKey), KeyLen);

    if (TrError == e_inv_not_found)
        TrError = e_hash_not_found;
}

/*===========================================================================*
 *  LOCAL FUNCTIONS (CODE)                                                   *
 *===========================================================================*/

inline static void 
bd22_GetOId (tgg00_TransContext   &Trans,
			 cbd900_ObjKeyFileId   ObjKeyFileId,
			 tsp00_Int4            KeyLen,
			 void                 *pKey,
			 tgg92_KernelOid      &Oid)
{
    ROUTINE_DBG_MEO00 ("bd22_GetOid");

    tgg00_BasisError &TrError = Trans.trError_gg00;

	if (TrError != e_ok) return;
        
    tsp00_Int4 DummyOidLen;
    bd03OmsGetInv (Trans, ObjKeyFileId, tsp00_BytePtr (pKey), KeyLen, sizeof(Oid), tsp00_BytePtr(&Oid), DummyOidLen);

	switch (TrError)
	{
	case e_inv_list_not_found:
	case e_key_not_found     : 	TrError = e_hash_key_not_found; break;
	case e_inv_not_found     :  TrError = e_hash_not_found;     break;
	}
}

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/ 
