/****************************************************************************

  module      : vbd90.cpp

  -------------------------------------------------------------------------

  responsible : AlexanderK

  special area: BD :       Operations for processing object files
  description : 


  version     : 7.2.0.0
  last changed: 2000-03-15
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1998



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/

/*  WAIT FOR LOCK RELEASE 
*
* kb50LockObj
*   \kb67LockObj
*     |\kb67NewObj
*     |  |\bd90ReuseObj
*     |  |   \bd04_ReuseObj
*     |  |      \bd04NewObjKey (wait)
*     |   \bd90_NewObjFromThisPage (nowait)
*     |\bd90LockObj
*     |   \bd04LockObj (wait)
*     |\bd90_DelAllObjOnThisPage (nowait)
*      \bd90_NextObjConsistentOnThisPage (nowait)
*         \bd90_NextObjConsistentOnThisPage
*/

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include <stdlib.h>

#include "hbd90.h"

#include "gsp03.h"
#include "gsp03_3.h"
#include "ggg12.h"
#include "gbd05.h"
#include "gbd94.h"
#include "gbd900.h"
#include "gbd910.h"
#include "gbd920.h"
#include "gbd930.h"

#include "hbd10.h"  
#include "hbd15.h"
#include "hbd20.h"
#include "hbd22.h"
#include "hbd91.h"
#include "hbd92.h"
#include "hbd93.h"
#include "hbd300.h" // SetRootCheck
#include "hgg01.h" 
#include "hgg01_1.h" 
#include "hgg10.h"
#include "hgg17.h"
#include "hkb50.h"
#include "hkb67.h"

#if COMPILEMODE_MEO00 >= SLOW_MEO00 
#include "hta99.h"
#include "hta01.h"
#endif

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#ifdef NO_INLINES
#  define _INLINE
#else
#  define _INLINE   inline
#endif

#define FIRST_CHAIN_NO 0

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

#if COMPILEMODE_MEO00 >= SLOW_MEO00 

#   define BD90_CHECK_KEY_SIZE(TR_ERROR, HASH_KEY_SIZE, OBJ_KEY_SIZE) \
			if (HASH_KEY_SIZE != OBJ_KEY_SIZE)                        \
			{                                                         \
				TR_ERROR = e_illegal_keylength;                       \ 
				return;                                               \ 
			}                                                        

#else

#   define BD90_CHECK_KEY_SIZE(TR_ERROR, HASH_KEY_SIZE, OBJ_KEY_SIZE) 

#endif

/*---------------------------------------------------------------------------*/

#define MIN(a, b)  (((a) < (b)) ? (a) : (b))
#define MAX(a, b)  (((a) > (b)) ? (a) : (b)) 

/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/

enum ebd90_FunctionName {Delete, Get, KeyGet, GetCons, KeyGetCons,
                         Lock, Reuse, Unlock, Update};

/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/

extern cbd900_ContObjFileDir      bd93ContObjFileDir;
extern cbd930_GarbCollController  bd91GarbCollController;

/*===========================================================================*
 *  GLOBAL VARIABLES                                                          *
 *===========================================================================*/
 
cbd900_ObjFileDir     bd90ObjFileDir;

/*===========================================================================*
 *  LOCAL FUNCTIONS (PROTOTYPES)                                             *
 *===========================================================================*/


static void 
bd90_CopyChainInfoToSubRoots (tbd_current_tree   &Current,
							  cbd900_ObjFileInfo &ObjFileInfo);

/*---------------------------------------------------------------------------*/

static _INLINE void
bd90_DelAllObjOnThisPage (tbd_current_tree       &Current, 
						  cbd910_PrimObjDataNode &NptrsPage, 
						  tsp00_Int4             &NumDelObj);

/*---------------------------------------------------------------------------*/

static _INLINE void
bd90_GetObjBody   (tgg00_TransContext &Trans,
				   tgg92_KernelOid     OId, 
				   tbd_nodeptr         pPage, 
				   tsp00_Int4          ExtObjBodySize, 
				   void               *pExtObjBody, 
				   tbd900_ObjFrame    *pObjFrame);

/*---------------------------------------------------------------------------*/

static _INLINE void
bd90_NextObjConsistentOnThisPage (tbd_current_tree        &Current,
                                  tgg92_KernelOid         &OId,
                                  tgg91_PageRef           &ResultObjVers,
                                  bool                     bWithObject,
                                  tsp00_Int4               ExtObjBodySize,
                                  void                    *pExtObjBody, 
                                  cbd910_PrimObjDataNode  &NptrsPage,    
								  tsp00_Int4              &ObjBodySize,              //[out=length of the object] 
                                  bool                    &bNextObjIsFound, 
								  tsp00_Int4              &LogReadAccessesPerPage);  //[out] PTS 1107819 AK 12/09/2000

/*---------------------------------------------------------------------------*/

static void 
bd90_NewObjFromThisPage (tbd_current_tree         &Current,
    					 cbd910_SubRootNode       &NptrsSubRoot,
						 cbd910_PrimObjDataNode   &NptrsPage, 
                         tsp00_Int4                KeyLen,
                         void                     *pKey,
                         tgg92_KernelOid          &OId,
                         tgg91_PageRef            &ResultObjVers);

/*---------------------------------------------------------------------------*/

static _INLINE void
bd90_NoteVersionErrorInKnldiag (ebd90_FunctionName  FunctionName,
                                tgg92_KernelOid     WantedOId,
                                tgg00_ObjFrameVers  ExpectedVersion);

/*---------------------------------------------------------------------------*/

static void
bd90_UpdateObjBody (tbd_current_tree  &Current,
					tgg92_KernelOid    OId,
					tgg91_PageRef	   ObjVers,
					tbd_nodeptr        pPage,
					tbd900_ObjFrame   *pObjFrame, 
					tsp00_Int4         NewObjBodySize,
					void              *pNewObjBody);

/*---------------------------------------------------------------------------*/

static _INLINE tsp00_Int4
bd90_VarObjBodySize (tbd900_ObjFrame *pObjFrame);

/*===========================================================================*
 *  GLOBAL FUNCTIONS (CODE)                                                  * 
 *===========================================================================*/

externCpp void
bd90CreateObjFile (tbd_current_tree &Current,
				   tsp00_Int4        ObjBodyLen,
				   tsp00_Int4        ObjKeyLen)
{	
    ROUTINE_DBG_MEO00 ("bd90CreateObjFile");

    tgg00_BasisError      &TrError    = Current.curr_trans->trError_gg00;
	int                    iChain;
	
	/* check if there is sufficient space to create a new object file */	
	b10check_space (MAX_NUM_CHAINS_PER_CONTAINER_BD900, Current);
   	if (e_ok != TrError) return;  // PTS 1107864 AK 15/09/2000

	/* create root page */
	cbd910_RootNode   NptrsRoot (Current, nr_for_update);
	NptrsRoot.bd910AssignToNewPage (ObjBodyLen, ObjKeyLen, NIL_PAGE_NO_GG00,
          FIRST_CHAIN_NO, MAX_LOCK_REQUEST_SEQ_NO_BD00);
	if (e_ok != TrError) return;
	
	/* update file id */
	Current.curr_tree_id.fileRoot_gg00() = NptrsRoot.np_ptr()->nd_id();
	bd300SetRootCheck (Current.curr_tree_id);
	
	/* create sub roots */
	cbd910_SubRootNode   NptrsSubRoot (Current, nr_for_update);
    cbd900_RootPageInfo  RootPageInfo (NptrsRoot);							 
		
    /* store the pageno again at the root page */
	RootPageInfo.bd900AddSubRoot (NptrsRoot.np_ptr()->nd_id());
	
	/* introduce additional chains to the file */
    for (iChain=1; (iChain < MAX_NUM_CHAINS_PER_CONTAINER_BD900) && (TrError == e_ok);  ++iChain)
    {
		/* create a new Subroot */
        NptrsSubRoot.bd910AssignToNewPage (ObjBodyLen, ObjKeyLen, NIL_PAGE_NO_GG00, iChain, MAX_LOCK_REQUEST_SEQ_NO_BD00); 
        if (e_ok != TrError) return;
		
		/*store the pageno of the subroot at the root page */
		RootPageInfo.bd900AddSubRoot (NptrsSubRoot.np_ptr()->nd_id());

		/* since the file is not yet added to the ObjFileDir checks  */
		/* would find differences between the content of the subroot */
		/* and the ObjFileDir. PTS 1107145 AK 10/07/2000             */                 
        NptrsSubRoot.bd910ReleasePageWithoutCheck(); 
    }
	
	/* insert the new file into the object file directory */
	if (PRIM_CONT_OBJ_FILE_NO_BD900 == Current.curr_tree_id.fileContObjFileNo_gg00())
		bd91AddFileToObjFileDir     (*Current.curr_trans, Current.curr_tree_id, NptrsRoot);
	else
		bd91AddFileToContObjFileDir (*Current.curr_trans, Current.curr_tree_id, NptrsRoot);
	
	/* remove pages if the file existed already in the file directory */
	if (e_duplicate_filename == TrError)
	{
		TrError = e_ok;
		for (iChain=1; iChain < RootPageInfo.bd900NumChains(); ++iChain)
		{
			NptrsSubRoot.bd910AssignToPage (RootPageInfo.bd900SubRoot(iChain));
			NptrsSubRoot.bd910FreePage();
		}
		NptrsRoot.bd910FreePage();
		TrError = e_duplicate_filename; 		
	}	
}

/*---------------------------------------------------------------------------*/

externCpp void 
bd90DelObj (tbd_current_tree       &Current,           //[inp\out]
            tgg92_KernelOid        &OId,               //[inp\out] 
            tgg91_PageRef           WantedObjVers)     //[inp]             
{
    ROUTINE_DBG_MEO00 ("bd90DelObj");
    
    /*  This routine marks an object as deleted, i.e. the state  */
    /*  it's state  switched to free afetr eot                   */ 
    
    tgg00_BasisError    &TrError = Current.curr_trans->trError_gg00;
        
	/* check if there is sufficient space to write rollback and history entries */
    b10check_space (10, Current);
   	if (e_ok != TrError) return;  // PTS 1107864 AK 15/09/2000

    /*  get page pointer  */
    cbd910_PrimObjDataNode    NptrsPage (Current, OId.gg92GetPno(), nr_for_update);
    if (TrError != e_ok) {return;};
    
    /* determine pointer to object frame */
    BD92CHECK_OID_POSITION (TrError, OId, NptrsPage);
    
    tbd900_ObjFrame  *pObjFrame = NptrsPage.bd910ObjFramePtr (OId.gg92GetPos());
    
    /*  check object state  */
    if ((pObjFrame->objHeader_bd900.ohdState_gg92 != obsReserved_egg00    ) &&
        (pObjFrame->objHeader_bd900.ohdState_gg92 != obsKeyReserved_egg00 ) &&
        (pObjFrame->objHeader_bd900.ohdState_gg92 != obsOccupied_egg00   ))
    {
        TrError = e_wrong_object_state;
        return;
    }
    
    /* check object version  */  
    if (pObjFrame->objHeader_bd900.ohdFrameVersion_gg92 != OId.gg92GetFrameVers()) 
    {
        TrError = e_wrong_object_version;
        bd90_NoteVersionErrorInKnldiag 
            (Delete,OId,pObjFrame->objHeader_bd900.ohdFrameVersion_gg92);
        return;
    }
    
    /* check if object is locked and create a log entry  */
	const bool  bPageIsForVarObj = (pt2VarObject_egg00 == NptrsPage.np_ptr()->nd_pt2());
	
    if (bPageIsForVarObj)
        bd93LogDelVarObj (Current, OId, WantedObjVers, NptrsPage.np_ptr(), pObjFrame);
    else
    {
        const int ObjBodySize = NptrsPage.np_ptr()->nd_obj_frame_len() - sizeof(tgg92_ObjHeader);
        
        kb67DelUpdObj (*Current.curr_trans, m_delete, Current.curr_tree_id,
            OId, NptrsPage.np_ptr()->nd_obj_key_len(), WantedObjVers,
            pObjFrame->objHeader_bd900, ObjBodySize, pObjFrame->objBody_bd900);
    }        
    if (TrError != e_ok) return;
    
    /* mark the object as deleted  */
    pObjFrame->objHeader_bd900.ohdState_gg92.becomes(obsFreeAfterEot_egg00);
    
    /* update page header */  
    ++NptrsPage.np_ptr()->nd_free_eot_obj_cnt();
    --NptrsPage.np_ptr()->nd_occ_obj_cnt();
}

/*---------------------------------------------------------------------------*/

externCpp void 
bd90DelAllObj (tbd_current_tree  &Current,           //[inp\out]
			   tsp00_Int4        &NumDelObj)
{
    ROUTINE_DBG_MEO00 ("bd90DelAllObj");
    
    /* This routine marks all objects of the current file as deleted */
    /* Completely changed by PTS 1107145 AK 10/07/2000               */
	
	tgg00_BasisError         &TrError      = Current.curr_trans->trError_gg00;
	
   	/* check if there is sufficient space to write rollback and history entries */
    b10check_space (20, Current);
   	if (e_ok != TrError) return;  // PTS 1107864 AK 15/09/2000
    
    const tgg00_ObjFileNo    &FileNo       = Current.curr_tree_id.fileObjFileNo_gg00();
    const tsp00_PageNo       &FirstSubRoot = bd90ObjFileDir[FileNo][FIRST_CHAIN_NO].ociSubRoot_bd900;   
	cbd910_PrimObjDataNode    NptrsPage (Current, FirstSubRoot, nr_for_update);
	
    /* initialize counters */
    NumDelObj = 0;
	
	/* go through all pages of the chain and delete all objects */
	do 
	{	
		/* delete objects on this page */
		bd90_DelAllObjOnThisPage (Current, NptrsPage, NumDelObj);
		if (TrError != e_ok) return;
		
		/* move to next page */ 		
		NptrsPage.bd910MoveToNextPage();
		
	} while ((e_ok == TrError) && NptrsPage.bd910PageIsAvailable());
}

/*---------------------------------------------------------------------------*/

externCpp void
bd90DropObjFile (tbd_current_tree   &Current)
{
    ROUTINE_DBG_MEO00 ("bd90DropObjFile");
	
	/* This function marks a file an all its continuation files as to be */
	/* deleted. Moreover GCs are started which actually remove the files */ 
	
    tgg00_BasisError      &TrError     = Current.curr_trans->trError_gg00;
	const tgg00_ObjFileNo  FileNo      = Current.curr_tree_id.fileObjFileNo_gg00();
	cbd900_ObjFileInfo    &ObjFileInfo = bd90ObjFileDir[FileNo]; 
	
	/* check if the root of the input fileid corresponds */
	/* to the root stored in the object file directory   */
	if ((ObjFileInfo.ofiRoot_bd900 != Current.curr_tree_id.fileRoot_gg00()) ||
	    ObjFileInfo.ofiRelFileIsReq_bd900)
	{
		TrError = e_wrong_obj_file_id;
		return;
	}

	/* get the root page */ 
	cbd910_RootNode   NptrsRoot (Current, Current.curr_tree_id.fileRoot_gg00(), nr_for_update);
	if (e_ok != TrError) return;
	
    /* check if the root page is already marked as to be deleted */
	cbd900_RootPageInfo  RootPageInfo (NptrsRoot);
	if (RootPageInfo.bd900FileIsDeleted()) 	return;
	
    /* mark the file as to be deleted */
	RootPageInfo.bd900FileIsDeleted() = true;

	/* release root page before a GC is started to avoid a dead lock */
	NptrsRoot.bd910ReleasePageWithoutCheck(); //CRS 1107329 AK 01/08/2000 
				
	/* start GCs and mark the file in the object file directory */
	bd91GarbCollController.bd930StartGCToRemoveFile(*Current.curr_trans, ObjFileInfo);
			
	/* mark continuation files (if they exist at all) as to be deleted */		
	if (ObjFileInfo.ofiFileIsForVarObj_bd900)
	{ 	
        tbd_current_tree         ContCurrent (Current);
		cbd900_ContObjFileInfo & ContObjFileInfo = bd93ContObjFileDir[FileNo];
		
		for (int iContFileNo = 1; iContFileNo <=ContObjFileInfo.cfiNumContFiles_bd900; ++iContFileNo)
		{
			tsp00_PageNo ContObjFileRoot = ContObjFileInfo[iContFileNo].ofiRoot_bd900;
			if (NIL_PAGE_NO_GG00 != ContObjFileRoot)
			{
				/* construct currrent for the continuation objects */
				ContObjFileInfo.bd900BuildFileId (FileNo, iContFileNo, ContCurrent.curr_tree_id);
				
				/* mark all continuation files as deleted */
				cbd910_RootNode   NptrsContRoot (ContCurrent, ContObjFileRoot, nr_for_update);
				if (e_ok != TrError) return;
				
				/* get handler for the data accomodated on the root page body */
				cbd900_RootPageInfo  ContRootPageInfo (NptrsContRoot);
				
				/* check if the root page is already marked as to be deleted */
				if (ContRootPageInfo.bd900FileIsDeleted()) 	
				{
#                   if COMPILEMODE_MEO00 >= QUICK_MEO00 
					g01abort (csp3_bd_msg, csp3_n_obj, 
						"BD90DropFile:File del   ", ContObjFileRoot);
#		            else
					g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
						"BD90DropFile:File del   ", ContObjFileRoot);
#                   endif
				}
				
				/* mark the file as to be deleted */
				ContRootPageInfo.bd900FileIsDeleted() = true;

#               if COMPILEMODE_MEO00 >= QUICK_MEO00 
				NptrsContRoot.bd910ReleasePageWithoutCheck(); //CRS 1107329 AK 01/08/2000 
#               endif
			}
			
			/* start GCs and mark the file in the object file directory */
			bd91GarbCollController.bd930StartGCToRemoveFile(*Current.curr_trans, ContObjFileInfo[iContFileNo]);
		}
	}
}

/*---------------------------------------------------------------------------*/ 
 
externC void
bd90DumpObjFileDir (tgg00_VfFileref      &HostFile,
					tsp00_Int4            BufferSize,
					tsp00_Page           &Buffer,
					tsp00_Int4           &OutPno,
					tsp00_Int4           &OutPos,
					tsp00_VfReturn       &HostError,
					tsp00_ErrText        &ErrText)
{
    ROUTINE_DBG_MEO00 ("bd90DumpObjFileDir");

	bd90ObjFileDir.bd900DumpObjFileDir         (HostFile, BufferSize, Buffer, OutPno, OutPos, HostError, ErrText);    
	bd93ContObjFileDir.bd900DumpContObjFileDir (HostFile, BufferSize, Buffer, OutPno, OutPos, HostError, ErrText);  
}

/*---------------------------------------------------------------------------*/

externCpp void
bd90GetObjConsistent (tbd_current_tree  &Current,            //[inp\out] 
                      tgg92_KernelOid   &OId,                //[inp\out] 
                      tsp00_Int4         ExtObjBodySize,     //[inp= buffersize for object]
                      void              *pExtObjBody,        //[inp]
                      tsp00_Int4         HashKeySize,        //[inp]
                      void              *pHashKey,           //[inp]  
					  tsp00_Int4        &ObjBodySize,        //[out=length of the object] 
                      tsp00_Int4        &KeyLen,             //[out]
                      tsp00_Byte*       *pKey,               //[out]
                      tgg91_PageRef     &ResultObjVers,       //[out]
                      tsp00_Int4        &LogReadAccesses)    //[out] PTS 1107819 AK 12/09/2000
{
    ROUTINE_DBG_MEO00 ("bd90GetObjConsistent");
	
    tgg00_BasisError    &TrError    = Current.curr_trans->trError_gg00;
    const bool          bWithObject = true;
	LogReadAccesses                 = 0;                    // PTS 1107819 AK 12/09/2000                
	
    /*  get page pointer  */   
    cbd910_PrimObjDataNode   NptrsPage (Current, OId.gg92GetPno(), nr_for_read);
    if( TrError != e_ok) return;
	
    /* determine pointer to object frame */
    BD92CHECK_OID_POSITION (TrError, OId, NptrsPage);	
    tbd900_ObjFrame  *pObjFrame = NptrsPage.bd910ObjFramePtr (OId.gg92GetPos());

    /* check the object state  */
    if (pObjFrame->objHeader_bd900.ohdState_gg92 == obsFree_egg00)
    {
        TrError = e_wrong_object_state;
        return;
    }
	
    /* set parameters needed for the search in the log */ 
    tgg00_ObjState     ObjState        = pObjFrame->objHeader_bd900.ohdState_gg92;
    tgg00_ObjTransInfo ObjTransInfo    = pObjFrame->objHeader_bd900.ohdTransInfo_gg92;
    tgg00_ObjFrameVers ResultFrameVers = pObjFrame->objHeader_bd900.ohdFrameVersion_gg92; //???NIL_OBJ_FRAME_VERS_GG92;
	
	/* get object body */
	const bool  bPageIsForVarObj = (pt2VarObject_egg00 == NptrsPage.np_ptr()->nd_pt2());
	
	if (bPageIsForVarObj)
	{
		ObjBodySize = bd90_VarObjBodySize (pObjFrame);
		bd93GetVarObjBody (Current, OId, ExtObjBodySize, pExtObjBody, pObjFrame);
	}
	else
	{
		/* If a key is defined the key length and the pointer at */
		/* key within the returned ObjBody are determined        */  
		ObjBodySize  = NptrsPage.np_ptr()->nd_obj_frame_len() - sizeof(tgg92_ObjHeader);
		KeyLen = NptrsPage.np_ptr()->nd_obj_key_len();
		if (KeyLen > 0) *pKey = (tsp00_Byte*)pExtObjBody + ObjBodySize - KeyLen;
		
		/* if the OId is found with a key from the hash (displayed by HashKeyLen>0)  */
		/* the key in the object and the key used for searching in the hash have     */
		/* to be equal                                                               */
		if (HashKeySize > 0) 
		{
			BD90_CHECK_KEY_SIZE (TrError, HashKeySize, KeyLen);
			
			const void *pKeyInObjFrame = pObjFrame->objBody_bd900 +	ObjBodySize - KeyLen;
			
			if (memcmp (pHashKey, pKeyInObjFrame, KeyLen))
			{
				TrError = e_wrong_object_version;
				return;
			}
		}
       	/*  move object and key into output array assigned by pExtObjBody */
		bd90_GetObjBody (*Current.curr_trans, OId, NptrsPage.np_ptr(),
			ExtObjBodySize, pExtObjBody, pObjFrame);
	}
    if (TrError != e_ok)  return;
    
    /*  the current object may be inconsistent with the wanted view        */
    /*  which consists of objects commited before the current              */
    /*  transaction had been started in that case the old object image has */
    /*  to be restored from the log. Now  the  current page is released    */
    /*  since the check if the object is consistent and the search for the */
    /*  old image of the object takes places at the log pages and not here */
    
    /*  release page pointer  */
    NptrsPage.bd910ReleasePage(); 
    
 /*  the log pages are searched for the right image    */
    kb67GetObjConsistent (*Current.curr_trans, OId, ObjState, ObjTransInfo, 
        bWithObject, KeyLen, ObjBodySize, ExtObjBodySize, pExtObjBody,
        ResultFrameVers, ResultObjVers, LogReadAccesses);                    // PTS 1107819 AK 12/09/2000

    if (TrError != e_ok)  return;                                            // PTS 1001331 AK 31/07/1998
	
    if (HashKeySize > 0) 
    {
        // set OId version, because it is wrong, if it is coming from the 
        // key-management	
        OId.gg92SetFrameVers (ResultFrameVers);
    }
    else
    {
        /*  check the object version */   
        if (OId.gg92GetFrameVers() != ResultFrameVers)
        {
            TrError = e_wrong_object_version;
            bd90_NoteVersionErrorInKnldiag 
                (GetCons, OId, pObjFrame->objHeader_bd900.ohdFrameVersion_gg92);
        }
    }
}

/*---------------------------------------------------------------------------*/
					  
externCpp void
bd90GetObjConsistentWithLock (tbd_current_tree  &Current,            //[inp\out] 
							  tgg92_KernelOid   &OId,                //[inp\out] 
							  tsp00_Int4         ExtObjBodySize,     //[inp= buffersize for object]
							  void              *pExtObjBody,        //[inp]
							  tsp00_Int4         HashKeySize,        //[inp]
							  void              *pHashKey,           //[inp]  
							  tsp00_Int4        &ObjBodySize,        //[out=length of the object] 
							  tsp00_Int4        &KeyLen,             //[out]
							  tsp00_Byte*       *pKey,               //[out]
							  tgg91_PageRef     &ResultObjVers,      //[out]
							  tsp00_Int4        &LogReadAccesses)    //[out] PTS 1107819 AK 12/09/2000
{
	ROUTINE_DBG_MEO00 ("bd90GetObjConsistentWithLock");
	
	tgg00_BasisError   &TrError         = Current.curr_trans->trError_gg00;
	const bool          bDontGetObject  = false;
    tsp00_Bool          bConsistent;
	/* PTS 1107819 AK 12/09/2000; note: only in case of the error=e_object_dirty        */
	/* a value for LogReadAccesses will be returned which is not zero will be returned  */
	LogReadAccesses                     = 0;  

	/* get page pointer */   
	cbd910_PrimObjDataNode   NptrsPage (Current, OId.gg92GetPno(), nr_for_update);
	if( TrError != e_ok) return;
	
	/* determine pointer to object frame */
	BD92CHECK_OID_POSITION (TrError, OId, NptrsPage);	
	tbd900_ObjFrame *const pObjFrame = NptrsPage.bd910ObjFramePtr (OId.gg92GetPos());
	
	/* check the object state  */
	switch (pObjFrame->objHeader_bd900.ohdState_gg92)
	{
	case obsFree_egg00 :
		TrError = e_wrong_object_state;
		return;
	case obsFreeAfterEot_egg00 :
		tgg00_ObjFrameVers  ResultFrameVers = NIL_OBJ_FRAME_VERS_GG92;
		kb67GetObjConsistent (*Current.curr_trans, OId, pObjFrame->objHeader_bd900.ohdState_gg92,
			pObjFrame->objHeader_bd900.ohdTransInfo_gg92, bDontGetObject, KeyLen, ObjBodySize, 
			ExtObjBodySize, pExtObjBody, ResultFrameVers, ResultObjVers, LogReadAccesses);  //PTS 1107819 AK 12/09/2000
		if (e_ok == TrError) TrError = e_object_dirty;
		return;
	}
	
	/* check object frame version */
	const tgg00_ObjFrameVers & CurrFrameVers = pObjFrame->objHeader_bd900.ohdFrameVersion_gg92;
	ObjBodySize = NptrsPage.np_ptr()->nd_obj_frame_len() - sizeof(tgg92_ObjHeader);
	KeyLen      = NptrsPage.np_ptr()->nd_obj_key_len();
	if (KeyLen > 0) *pKey = (tsp00_Byte*)pExtObjBody + ObjBodySize - KeyLen;
	
	if (HashKeySize > 0) 
	{
		/* if the OId is found with a key from the hash (displayed by HashKeyLen>0)  */
		/* the key in the object and the key used for searching in the hash have     */
		/* to be equal                                                          	 */
		
		BD90_CHECK_KEY_SIZE (TrError, HashKeySize, KeyLen);
		const void *pKeyInObjFrame = pObjFrame->objBody_bd900 +	ObjBodySize - KeyLen;
		
		if (memcmp (pHashKey, pKeyInObjFrame, KeyLen))
		{
			TrError = e_wrong_object_version;
			return;
		}
		OId.gg92SetFrameVers (CurrFrameVers);
	}
	else
	{   
		if (CurrFrameVers != OId.gg92GetFrameVers()) 
		{
			TrError = e_wrong_object_version;
			return;
		}
	}
	
	/* check if the current object version is consistent with this view*/
	tgg00_ObjTransInfo & ObjTransInfo = pObjFrame->objHeader_bd900.ohdTransInfo_gg92;
	kb50ObjConsistentCheck (*Current.curr_trans, ObjTransInfo.otrUpdTransId_gg00, bConsistent);
	//                      without otrBeforeRef:  PTS 1105492 JA 2000-03-15 
	if (e_ok != TrError) return;

	/* return if the consistent version does not correspond to the current version */ 
	if (!bConsistent)
	{
		TrError = e_object_dirty;
		return;
	}
	
	/* lock object */
	ResultObjVers = ObjTransInfo.otrBeforeRef_gg00;
	kb67LockObj (*Current.curr_trans, Current.curr_tree_id, OId, ResultObjVers,
		pObjFrame->objHeader_bd900.ohdState_gg92, pObjFrame->objHeader_bd900.ohdTransInfo_gg92); 
	if (e_ok != TrError) return;
	
	/* get object body */
	if (pt2VarObject_egg00 == NptrsPage.np_ptr()->nd_pt2())
	{
		ObjBodySize = bd90_VarObjBodySize (pObjFrame);
		bd93GetVarObjBody (Current, OId, ExtObjBodySize, pExtObjBody, pObjFrame);
	}
	else
	{
		ObjBodySize  = NptrsPage.np_ptr()->nd_obj_frame_len() - sizeof(tgg92_ObjHeader);
		bd90_GetObjBody (*Current.curr_trans, OId, NptrsPage.np_ptr(),ExtObjBodySize, pExtObjBody, pObjFrame);
	}
}

/* ----------------------------------------------------------------------------------*/

externCpp void
bd90GetObjFileStatistics (tbd_current_tree   &Current,
						  tsp00_Int4         &NumAllPages,
						  tsp00_Int4         &NumFreePages,
						  tsp00_Int4         &NumEmptyPages,
						  tsp00_Int4         &NumKeyLeaves,        /* PTS 1107153 AK 17/07/2000 */
						  tsp00_Int4         &Occupancy)           /* PTS 1107153 AK 17/07/2000 */         
{
	
	ROUTINE_DBG_MEO00 ("bd90GetObjFileStatistics");
	
    /* initialize counters */
    NumAllPages   = 0;
    NumFreePages  = 0;
    NumEmptyPages = 0;
	NumKeyLeaves  = 0;

	const tgg00_ObjFileNo  &ObjFileNo   = Current.curr_tree_id.fileObjFileNo_gg00();
	cbd900_ObjFileInfo     &ObjFileInfo = bd90ObjFileDir[ObjFileNo];

    /* check input */
	if ((ObjFileNo < MIN_OBJ_FILE_NO_GG92) ||
        (ObjFileNo > MAX_OBJ_FILE_NO_GG92) ||
        (ObjFileInfo.ofiRoot_bd900 == NIL_PAGE_NO_GG00))
    {
        Current.curr_trans->trError_gg00 = e_wrong_class_id;
    }

    /* go trough all subroots of the file and sum the page counters */      
    for (int iChain=0; iChain<ObjFileInfo.ofiMaxChain_bd900; ++iChain)
    {
        const tsp00_PageNo SubRoot = ObjFileInfo[iChain].ociSubRoot_bd900;       
        cbd910_SubRootNode NptrsSubRoot (Current, SubRoot, nr_for_read);
        if (Current.curr_trans->trError_gg00 != e_ok) return;
		
		/* get handler for the data stored on the subroot page body */
		cbd900_SubRootPageInfo  SubRootPageInfo (NptrsSubRoot);
		
        NumAllPages     += SubRootPageInfo.bd900AllPageCnt();
        NumFreePages    += SubRootPageInfo.bd900FreePageCnt();
        NumEmptyPages   += SubRootPageInfo.bd900EmptyPageCnt();
    }
	
	/* add continuation pages */
	if (ObjFileInfo.ofiFileIsForVarObj_bd900)
	{ 
		/* initialize counters */
		int NumAllContPages     = 0;
		int NumFreeContPages    = 0;
		int NumEmptyContPages   = 0;
		
        bd93GetContObjFileStatistics (Current, NumAllContPages, NumFreeContPages, NumEmptyContPages);
		
		NumAllPages   += NumAllContPages;
		NumEmptyPages += NumEmptyContPages;
		NumFreePages  += NumFreeContPages;
	}
	
	/* get number of leaves from the key tree  PTS 1107153 AK 17/07/2000 */
	if (ObjFileInfo.ofiKeyRoot_bd900 != NIL_PAGE_NO_GG00)
		bd22GetLeaveCountOfObjKeyFile (*Current.curr_trans, ObjFileNo, ObjFileInfo, NumKeyLeaves);

	/* determines the average occupancy rate (in %) of the pages of this file */
    tsp00_Int4  NumOccPages  = ObjFileInfo.bd900NumOccPages();  
    tsp00_Int4  NumAllPages2 = ObjFileInfo.bd900NumAllPages();  	
	
	if (ObjFileInfo.ofiFileIsForVarObj_bd900)
	{
		/* add contributions of continuation files */
		cbd900_ContObjFileInfo & ContObjFileInfo = bd93ContObjFileDir[ObjFileNo];
		for (int iContFileNo=1;iContFileNo <= ContObjFileInfo.cfiNumContFiles_bd900; ++iContFileNo)
		{
			NumOccPages  += ContObjFileInfo[iContFileNo].bd900NumOccPages();  
			NumAllPages2 += ContObjFileInfo[iContFileNo].bd900NumAllPages(); 
		}
	}
				
	if (NumAllPages > 0) 
		Occupancy = NumOccPages/NumAllPages;
	else
		Occupancy = 0;	
	
	/* check if the numbers of pages from the objfiledir and the subroots are equal */
#   if COMPILEMODE_MEO00 >= QUICK_MEO00 
	if (NumAllPages != NumAllPages2)
	{
		g01opmsg (sp3p_knldiag, sp3m_warning, csp3_bd_msg, csp3_n_obj,   
			"bd90Stat:odd AllPageCnt", ObjFileNo);
		g01opmsg (sp3p_knldiag, sp3m_warning, csp3_bd_msg, csp3_n_obj,   
			"bd90Stat: AllPageCnt1 =", NumAllPages);
		g01opmsg (sp3p_knldiag, sp3m_warning, csp3_bd_msg, csp3_n_obj,   
			"bd90Stat: AllPageCnt2 =", NumAllPages2);
    }
#   endif
}

/*----------------------------------------------------------------------------*/
/* CRS 1106284 AK 11-03-2000 */

externCpp void
bd90IsObjHistoryUsed (tbd_current_tree &Current,
                     tgg92_KernelOid   &OId,
                     bool              &bHistoryIsUsed)
{
    ROUTINE_DBG_MEO00 ("bd90IsObjSelfLocked");
    
    tgg00_BasisError  &TrError = Current.curr_trans->trError_gg00;

    /*  get page pointer  */
    cbd910_PrimObjDataNode  NptrsPage (Current, OId.gg92GetPno(),  nr_for_read);
    if (TrError != e_ok) return; 
    
    /* determine pointer to object frame */
    BD92CHECK_OID_POSITION  (TrError, OId, NptrsPage);    
    tbd900_ObjFrame  *pObjFrame = NptrsPage.bd910ObjFramePtr (OId.gg92GetPos());

    /*  check the object version */
    if (pObjFrame->objHeader_bd900.ohdFrameVersion_gg92 != OId.gg92GetFrameVers()) 
    {
        TrError = e_wrong_object_version;
        bd90_NoteVersionErrorInKnldiag 
            (Lock, OId, pObjFrame->objHeader_bd900.ohdFrameVersion_gg92);
        return;
    }
    
    /* check object state */
    if (pObjFrame->objHeader_bd900.ohdState_gg92 != obsOccupied_egg00)
    {
        TrError = e_wrong_object_state;
        return;
    }
    
	bHistoryIsUsed = 
		kb67IsObjHistoryUsed (*Current.curr_trans, OId, pObjFrame->objHeader_bd900.ohdTransInfo_gg92);
}

/*----------------------------------------------------------------------------*/
/* PTS 1001745 UH 23-11-1998 new */

externCpp void
bd90IsObjSelfLocked (tbd_current_tree   &Current,
                     tgg92_KernelOid    &OId,
                     bool               &IsSelfLocked)
{
    ROUTINE_DBG_MEO00 ("bd90IsObjSelfLocked");
    
    tgg00_BasisError  &TrError = Current.curr_trans->trError_gg00;

    IsSelfLocked = false;

    /*  get page pointer  */
    cbd910_PrimObjDataNode  NptrsPage (Current, OId.gg92GetPno(),  nr_for_read);
    if (TrError != e_ok) return; 
    
    /* determine pointer to object frame */
    BD92CHECK_OID_POSITION  (TrError, OId, NptrsPage);    
    tbd900_ObjFrame  *pObjFrame = NptrsPage.bd910ObjFramePtr (OId.gg92GetPos());

    /*  check the object version */
    if (pObjFrame->objHeader_bd900.ohdFrameVersion_gg92 != OId.gg92GetFrameVers()) 
    {
        TrError = e_wrong_object_version;
        bd90_NoteVersionErrorInKnldiag 
            (Lock, OId, pObjFrame->objHeader_bd900.ohdFrameVersion_gg92);
        return;
    }
    
    /* check object state */
    if (pObjFrame->objHeader_bd900.ohdState_gg92 == obsFree_egg00)
    {
        TrError = e_wrong_object_state;
        return;
    }
    
	kb50CheckObjSelfLocked (*Current.curr_trans, OId, pObjFrame->objHeader_bd900.ohdTransInfo_gg92);
        
    /* PTS 1001769 Torsten Strahl 1998-11-30 */
    if (e_ok == TrError) 
        IsSelfLocked = true;
    else
        if (e_object_not_locked == TrError)         
            TrError = e_ok;
}

/*---------------------------------------------------------------------------*/

externCpp void
bd90LockObj (
             tbd_current_tree   &Current,
             tgg92_KernelOid    &OId,
             tgg91_PageRef       WantedObjVers
             )
             
{
    // NOTE! if this function returns e_wait_for_lock_release then wait outside bd-locking 

    ROUTINE_DBG_MEO00 ("bd90LockObj");

    tgg00_BasisError    &TrError = Current.curr_trans->trError_gg00;
    
    /*  get page pointer  */
    cbd910_PrimObjDataNode    NptrsPage(Current, OId.gg92GetPno(),  nr_for_update);
    if (TrError != e_ok) return; 
    
    /* determine pointer to object frame */
    BD92CHECK_OID_POSITION (TrError, OId, NptrsPage);    
    tbd900_ObjFrame  *pObjFrame = NptrsPage.bd910ObjFramePtr (OId.gg92GetPos());

    /*  check object version  */
    if (pObjFrame->objHeader_bd900.ohdFrameVersion_gg92 != OId.gg92GetFrameVers()) 
    {
        TrError = e_wrong_object_version;
    }
    else
    {
        switch (pObjFrame->objHeader_bd900.ohdState_gg92)
        {
        case obsFree_egg00 :
            TrError = e_wrong_object_state;
            break;
        case obsFreeAfterEot_egg00 :
            TrError = e_object_dirty;
            break;
        default :
            kb67LockObj (*Current.curr_trans, Current.curr_tree_id, OId, WantedObjVers,
                pObjFrame->objHeader_bd900.ohdState_gg92,
                pObjFrame->objHeader_bd900.ohdTransInfo_gg92);      
        }
    }   
}

/*---------------------------------------------------------------------------*/

externCpp void 
bd90NewObj (tbd_current_tree        &Current,                  // [in\out]
            tsp00_Int4              &FrameLen,                 // [out]
            tsp00_Int4               KeyLen,                   // [in]
            void                    *pKey,                     // [in]
            tgg92_KernelOid         &ResultOId,                // [out]
            tgg91_PageRef           &ResultObjVers)            // [out]
{
    ROUTINE_DBG_MEO00 ("bd90NewObj");
	
    /*  This function yields a new OId. First it tries to get    */
    /*  the new OId from the free chain. If it does not find     */
    /*  a new OId in this chain a new page is included into the  */
    /*  object file and an OId from this page is served          */

    tgg00_BasisError        &TrError = Current.curr_trans->trError_gg00;

	/* check if there is sufficient space to get a new  object frame, */
	/* a rollback entry and to insert a new key into the key tree     */
    b10check_space (10, Current);
   	if (e_ok != TrError) return;  // PTS 1107864 AK 15/09/2000
    
    const tgg00_ObjFileNo   &FileNo  = Current.curr_tree_id.fileObjFileNo_gg00(); 
    cbd910_PrimObjDataNode   NptrsPage    (Current, nr_for_update);
	
    /* get current chain  and  pointer to the corresponding sub root page */
	cbd900_ObjFileInfo & ObjFileInfo = bd90ObjFileDir[FileNo];
	
	/* check if the file has been deleted already */
	if (ObjFileInfo.ofiRelFileIsReq_bd900 ||
		ObjFileInfo.ofiRoot_bd900 != Current.curr_tree_id.fileRoot_gg00() )
	{
		TrError = e_wrong_class_id;
		return;
	}
	
	/* find the chain (i.e. subroot) where the new object should be inserted */
    const tsp00_Int4      ChainNo   = ObjFileInfo.bd900CurrChain();
    cbd900_ObjChainInfo  &ChainInfo = ObjFileInfo[ChainNo];
    const tsp00_PageNo    SubRoot   = ChainInfo.ociSubRoot_bd900; 
    
	cbd910_SubRootNode  NptrsSubRoot (Current, SubRoot, nr_for_update);
    
	if (e_ok != TrError)
	{
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_page_chain, csp3_n_obj,   
			"bd90new: FileNo=        ", FileNo);
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_page_chain, csp3_n_obj,   
			"bd90new: ChainNo=       ", ChainNo);
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_page_chain, csp3_n_obj,   
			"bd90new: CurrSubRoot=   ", SubRoot);
		return;
	}
	
#   if COMPILEMODE_MEO00 >= QUICK_MEO00 
	/* check if this subroot page belongs really to this file */
	if ((NptrsSubRoot.np_ptr()->nd_root()        != ObjFileInfo.ofiRoot_bd900) || 
		(NptrsSubRoot.np_ptr()->nd_obj_file_no() != FileNo)                    ||
		(NptrsSubRoot.np_ptr()->nd_sub_root()    != SubRoot)                   ||
		(NptrsSubRoot.np_ptr()->nd_chain_no()    != ChainNo))
	{
		g01abort (csp3_bd_msg, csp3_n_obj, 
			"BD90_NewObj:wrong SRoot ", NptrsSubRoot.np_ptr()->nd_id());
    }
#   endif

	/* set return parameter */
	FrameLen  = NptrsSubRoot.np_ptr()->nd_obj_frame_len();
	
	/* get page handle for the first page with free objects */
	/* CRS 1106534 Alexander Kley 2000-05-11                */
#   ifdef STORE_OBJECTS_ON_SUBROOT_BD900
	if (NptrsSubRoot.np_ptr()->nd_free_obj_cnt() > 0)
		NptrsPage.bd910SetTo (NptrsSubRoot);
	else
#   endif
	{
		if (NIL_PAGE_NO_GG00 != NptrsSubRoot.np_ptr()->nd_next_free())
			NptrsPage.bd910AssignToPage(NptrsSubRoot.np_ptr()->nd_next_free());
		else
			NptrsPage.bd910AssignToNewPageInDataChain (NptrsSubRoot);
		if (e_ok != TrError) return;
	}
	
	/* get a new OId and initialize the corresponding object */
	bd90_NewObjFromThisPage (Current, NptrsSubRoot, NptrsPage, 
		KeyLen, pKey, ResultOId, ResultObjVers);     
}

/*---------------------------------------------------------------------------*/

externCpp void 
bd90NextObjConsistent (tbd_current_tree       &Current,
					   tgg92_KernelOid        &OId,
					   bool                    bWithObject,
					   tsp00_Int4              ExtObjBodySize,
					   void                   *pExtObjBody, 
					   tsp00_Int4             &ObjBodySize,         //[out=length of the object] 
					   tgg91_PageRef          &ObjVers,
					   tsp00_Int4             &AllLogReadAccesses)  //[out] PTS 1107819 AK 12/09/2000
{
    // NOTE! if this function returns e_wait_for_lock_release then wait outside bd-locking 
	
    ROUTINE_DBG_MEO00 ("bd90NextObjConsistent");
    
    tgg00_BasisError      &TrError         = Current.curr_trans->trError_gg00;
    bool                   bNextObjIsFound = false; 
    cbd910_PrimObjDataNode NptrsPage(Current, nr_for_read);
	tsp00_Int4             LogReadAccessesPerPage;   // PTS 1107819 AK 12/09/2000
    
    /*---------------------------------*/
    /*  find starting page for search  */
    /*---------------------------------*/
    
    if (!OId.gg92IsNil() )
    {
        /* if OID is not nil look for object following the given OID  */
		NptrsPage.bd910AssignToPage(OId.gg92GetPno());
        if (e_ok != TrError) return;   
		
		/* check fileno and state of input object */
		BD92CHECK_OID_POSITION (TrError, OId, NptrsPage);                                    
		tbd900_ObjFrame *pObjFrame = NptrsPage.bd910ObjFramePtr (OId.gg92GetPos());
		
		if (pObjFrame->objHeader_bd900.ohdState_gg92 == obsFree_egg00)
		{
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
				"BD90NextObj:WrongIter   ", OId.gg92GetPno());
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
				"BD90NextObj:WrongIter   ", OId.gg92GetPos());
			/* note: The previous check  is very simple in principle one             */
			/*  	 should even check if the input OId belongs to an object which   */
			/*       owns a consistent image with respect to the current transaction */
			
			TrError = e_wrong_object_state;
			return;
		}
    }
    else
    {   
        /*  look for very first object in file, which  */
        /*  should be on the  subroot                  */
        const tgg00_ObjFileNo &FileNo           = Current.curr_tree_id.fileObjFileNo_gg00();
        const tsp00_PageNo     SubRoot          = bd90ObjFileDir[FileNo][FIRST_CHAIN_NO].ociSubRoot_bd900;   
		
        NptrsPage.bd910AssignToPage(SubRoot);
        if (e_ok != TrError) return;   
    }
    
    /*-------------------------------------------------*/
    /*  start looping over chains and pages of chains  */
    /*-------------------------------------------------*/
    
    /* loop over  object page */
    do
    {   
		/*  look for the next object on current page  */            
		bd90_NextObjConsistentOnThisPage (Current, OId, ObjVers, bWithObject, ExtObjBodySize, 
			pExtObjBody, NptrsPage, ObjBodySize, bNextObjIsFound, LogReadAccessesPerPage);

		/* accumulalte the number of read accesses to object history pages; PTS 1107819 AK 12/09/2000 */
		AllLogReadAccesses += LogReadAccessesPerPage;
		
		/* return immediately if the next object is found */
		if (bNextObjIsFound || TrError != e_ok) return;
		
        /* check if the command has been canceled. PTS 1105431 Alexander Kley 25/01/2000 */
		if (Current.curr_trans->trRteCommPtr_gg00->to_cancel)
		{
			TrError = e_cancelled;
			break;
		}

        /* reset OId and  goto next page */
		OId.gg92SetNil();
		NptrsPage.bd910MoveToNextPage();
	}
	while ((e_ok == TrError) && NptrsPage.bd910PageIsAvailable());   
    /* end of loop over pages */  
	
    /* nothing was found */
    OId.gg92SetNil();
	if (e_ok == TrError) 
		TrError = e_no_next_object;
}

/*---------------------------------------------------------------------------*/

externCpp void 
bd90ReadAhead (tbd_current_tree  &Current)
{
    ROUTINE_DBG_MSP00 ("bd90ReadAhead");
    
	/* this function reads all pages of a file in order to load them into the data cache       */
	/* however reading is stopped if the data cache is full (i.e. occupied by more than 99.9%) */
	/* PTS 1107145 AK 11/07/00 function rewritten                                              */
	
    tgg00_BasisError        &TrError         = Current.curr_trans->trError_gg00;
    const tgg00_ObjFileNo    FileNo          = Current.curr_tree_id.fileObjFileNo_gg00();
    const tsp00_PageNo       RootPno         = bd90ObjFileDir[FileNo].ofiRoot_bd900;   
	cbd910_ObjNode           NptrsPage (Current, RootPno, nr_for_read);
	const tsp00_Int4         c100_Percent    = 1000;
	const tsp00_Int4         c99Dot9_Percent =  999;
    const tsp00_Int4         MaxNumPagesToReadTillNextCheck = 20;
	tsp00_Int4               NumPagesToReadTillNextCheck    = 0;  
	tsp00_Int4               DataCacheSize;
	tsp00_Int4               RollbackPageCnt;
	tsp00_Int4               OmsHistoryPageCnt;
	tsp00_Int4               OmsDataPageCnt;
	tsp00_Int4               SqlDataPageCnt;
	tsp00_Int4               DataCacheOccupancy;
	
	while ((e_ok == TrError) && NptrsPage.bd910PageIsAvailable())
	{
		/* Stop reading if the cache is completely filled; PTS 1107820 AK 21/09/2000 */
		if (NumPagesToReadTillNextCheck <= 0)
		{
			NumPagesToReadTillNextCheck = MaxNumPagesToReadTillNextCheck;
			
			/* get data cache occupation */
			bd20GetWorkloadCounter (DataCacheSize, RollbackPageCnt, OmsHistoryPageCnt, OmsDataPageCnt, SqlDataPageCnt);
		    DataCacheOccupancy = RollbackPageCnt + OmsHistoryPageCnt + OmsDataPageCnt + SqlDataPageCnt;
			
			/* if the datacache is too full not only a cheap but not accurate check       */
			/* must be peforemed to decide  if an object can be released but an exact one */ 
			if 	(DataCacheSize * c99Dot9_Percent < DataCacheOccupancy * c100_Percent)
				break;
		}
		/* move to next page, i.e. read next page into the data cache */
		NptrsPage.bd910MoveToNextPage();
		
		--NumPagesToReadTillNextCheck;
	}
}

/*---------------------------------------------------------------------------*/

externC void
bd90RefreshObjFileStatistics (tgg00_TransContext &Trans)
{
    ROUTINE_DBG_MEO00 ("bd90RefreshObjFileStatistics");
    
    /* This function copies the number of all objects stored in the    */
    /* global structure bd90ObjFileDir to the subroots in order to     */
    /* maintain this number after a restart                            */
	/* changed by PTS 1107306 AK 01/08/2000                            */     
    
	cbd05_Current          Current     (Trans, m_update);           
	cbd900_ObjFileDirIter  ObjFileIter (bd90ObjFileDir);
    int                    iContFileNo;
    
	for  (ObjFileIter.bd920First(); NULL != ObjFileIter(); ++ObjFileIter) 
    {  
		const tgg00_ObjFileNo ObjFileNo = ObjFileIter.bd920Key();
		ObjFileIter()->bd900BuildFileId (ObjFileNo, Current.curr_tree_id);
		bd90_CopyChainInfoToSubRoots (Current, *ObjFileIter());
		
		/* copy the root pno of the continuation file to the root page */
		const tsp00_PageNo RootPno = ObjFileIter()->ofiRoot_bd900;                  
		cbd910_RootNode  NptrsRoot (Current, RootPno, nr_for_update);
		
		if (pt2VarObject_egg00 == NptrsRoot.np_ptr()->nd_pt2())
		{
			cbd900_ContObjFileInfo & ContObjFileInfo = bd93ContObjFileDir[ObjFileNo];
			cbd900_RootPageInfo      RootPageInfo (NptrsRoot);
			
			RootPageInfo.bd900NumContFiles() = ContObjFileInfo.cfiNumContFiles_bd900;
			for (iContFileNo=1; iContFileNo<= ContObjFileInfo.cfiNumContFiles_bd900; ++iContFileNo)
			{
				RootPageInfo.bd900ContRoot (iContFileNo) = ContObjFileInfo[iContFileNo].ofiRoot_bd900;
				if (NIL_PAGE_NO_GG00 != ContObjFileInfo[iContFileNo].ofiRoot_bd900)
				{
					ContObjFileInfo.bd900BuildFileId (ObjFileNo, iContFileNo, Current.curr_tree_id);			
					bd90_CopyChainInfoToSubRoots (Current, ContObjFileInfo[iContFileNo]);
				}
			}
		}
	}
}   

/*---------------------------------------------------------------------------*/

externCpp void
bd90ResetObjLock (tbd_current_tree   &Current,
                  tgg92_KernelOid    &OId,
                  tgg91_TransNo       LockTransId,
                  tgg00_TransIndex    LockTransIndex)                  
{
    ROUTINE_DBG_MSP00 ("bd90ResetObjLock");
    
    tgg00_BasisError    &TrError   = Current.curr_trans->trError_gg00;
	tgg00_ObjFileNo     &ObjFileNo = Current.curr_tree_id.fileObjFileNo_gg00();

    /*  get page pointer  */
    cbd910_PrimObjDataNode   NptrsPage(Current, OId.gg92GetPno(),  nr_for_update);

	/* check if the page belongs to the right container  PTS 1106143 AK 29/03/2000 */
    if ((e_ok      != TrError) || 
		(ObjFileNo != NptrsPage.np_ptr()->nd_obj_file_no())) 
	{
		/* check if the error is due to a drop of the file */
		if ((e_page_in_wrong_tree == TrError) || (e_no_converter_entry == TrError))
		{
			if ((NIL_PAGE_NO_GG00 == bd90ObjFileDir[ObjFileNo].ofiRoot_bd900) ||
				(bd90ObjFileDir[ObjFileNo].ofiRelFileIsReq_bd900))
				TrError = e_object_not_found;
		}
		return;
	}

    /* determine pointer to object frame */
    BD92CHECK_OID_POSITION (TrError, OId, NptrsPage);    
    
    tbd900_ObjFrame  *pObjFrame = NptrsPage.bd910ObjFramePtr (OId.gg92GetPos());
    
    if (pObjFrame->objHeader_bd900.ohdState_gg92 == obsFree_egg00)
        TrError = e_wrong_object_state;
    else 
    {
        kb67ResetObjLock (*Current.curr_trans, LockTransId, LockTransIndex,
            OId, pObjFrame->objHeader_bd900.ohdTransInfo_gg92);
    }
}

/*---------------------------------------------------------------------------*/

externCpp void
bd90ReuseObj (tbd_current_tree   &Current,
              tgg92_KernelOid    &OId,
              tgg91_PageRef      &ResultObjVers,
              tsp00_Int4          KeySize,
              void               *pKey,
              bool               &bTryAgain)
{
    // NOTE! if this function returns e_wait_for_lock_release then wait outside bd-locking 

    ROUTINE_DBG_MEO00 ("bd90ReuseObj");
   
    tgg00_BasisError    &TrError = Current.curr_trans->trError_gg00;
    bTryAgain = false;

	/* check if there is sufficient space to write a rollback  entry */
    b10check_space (10, Current);
   	if (e_ok != TrError) return;  // PTS 1107864 AK 15/09/2000

	/* get pointer to page */
    cbd910_PrimObjDataNode    NptrsPage (Current, OId.gg92GetPno(),  nr_for_update);
    
    if (TrError != e_ok)
    {
        if ((TrError == e_page_in_wrong_tree) || (TrError == e_no_converter_entry))
        {
            // OID provided from key-management isn't valid yet 
            TrError = e_too_old_oid;
            bTryAgain = true;
        }
        return;
    }
    
	/* get pointer to object frame*/
    BD92CHECK_OID_POSITION (TrError, OId, NptrsPage);    
    tbd900_ObjFrame  *pObjFrame = NptrsPage.bd910ObjFramePtr (OId.gg92GetPos());
    
    if (pObjFrame->objHeader_bd900.ohdState_gg92 == obsFree_egg00)
    {
        TrError   = e_wrong_object_state;
        bTryAgain = true;
        return;
    }
    
    const int  ObjBodySize = NptrsPage.np_ptr()->nd_obj_frame_len() - sizeof(tgg92_ObjHeader);
    void   *pKeyInObjFrame = pObjFrame->objBody_bd900 + ObjBodySize - KeySize;
    
    if (memcmp (pKey, pKeyInObjFrame, KeySize))
    {
        // meanwhile the information from the hash that there exists an
		// object frame with the key pKey may be wrong since the garbage 
		// collector removed this object, therefore this frame cannot be 
		// reused and one has to try again to insert the key pKey into the hash

        TrError   = e_wrong_object_version;
        bTryAgain = true;
        return;
    }
    
    kb67NewObj (*Current.curr_trans, Current.curr_tree_id, OId, pObjFrame->objHeader_bd900,
        ObjBodySize, KeySize);
    
    if (e_wait_for_lock_release == TrError)
    {
        bTryAgain = true;
        return;
    }
    
    if (TrError == e_ok)
    {
        ResultObjVers = pObjFrame->objHeader_bd900.ohdTransInfo_gg92.otrBeforeRef_gg00;
        
        -- NptrsPage.np_ptr()->nd_free_eot_obj_cnt();
		++ NptrsPage.np_ptr()->nd_occ_obj_cnt();
		
		pObjFrame->objHeader_bd900.ohdState_gg92.becomes(obsKeyReserved_egg00);
		
		if(pObjFrame->objHeader_bd900.ohdFrameVersion_gg92 >= MAX_OBJ_FRAME_VERS_GG92)
			pObjFrame->objHeader_bd900.ohdFrameVersion_gg92 = MIN_OBJ_FRAME_VERS_GG92;
		else
			pObjFrame->objHeader_bd900.ohdFrameVersion_gg92 ++;
		
		OId.gg92SetFrameVers (pObjFrame->objHeader_bd900.ohdFrameVersion_gg92);
    }
}

/*---------------------------------------------------------------------------*/

externCpp void 
bd90RollbackObj (tbd_current_tree    &Current,            //[inp\out]
                 tgg00_MessType_Enum  RollbackType,       //[inp]
                 tgg92_KernelOid     &OId,                //[inp]
                 tgg00_ObjState       NewState,           //[inp]
                 tgg91_PageRef        BeforeObjRef,       //[inp]
                 tgg91_TransNo        BeforeObjUpdVers,   //[inp]
                 tsp00_Int4           BeforeObjSize,      //[inp]
                 void                *pBeforeObj)         //[inp]
{
    ROUTINE_DBG_MEO00 ("bd90RollbackObj");

    /* possible actions which can be rollbacked by this procedure */

	/* current state  |   action   |   new state    */
	/* ---------------------------------------------*/
	/*  reserved     --- m_new    ---> free         */
    /*  key_reserved --- m_new    ---> free         */
	/*  key_reserved --- m_new    ---> feot         */
	/*  occupied     --- m_insert ---> reserved     */
	/*  occupied     --- m_insert ---> key_reserved */
	/*  occupied     --- m_update ---> occupied     */
	/*  feot         --- m_delete ---> occupied     */
	/*  feot         --- m_delete ---> reserved     */
	/*  feot         --- m_delete ---> key_reserved */

    tgg00_BasisError     &TrError                      = Current.curr_trans->trError_gg00;
    bool                  bTooManyEmptyPagesInContFile = false;
    tsp00_Int4            ContChainNo;           
	tsp00_Int4            ContObjFileNo;

    /* get page pointer */
    cbd910_PrimObjDataNode  NptrsPage(Current, OId.gg92GetPno(), nr_for_update);
    if (TrError != e_ok)      	
	{
		if ((e_page_in_wrong_tree == TrError) || (e_no_converter_entry == TrError)) /* PTS 1105630 Alexander Kley 16/02/2000 */
		{
			/* check if this error occured since the container was dropped */
			cbd900_ObjFileInfo &ObjFileInfo = bd90ObjFileDir[Current.curr_tree_id.fileObjFileNo_gg00()];
			if ((ObjFileInfo.ofiRoot_bd900 == NIL_PAGE_NO_GG00) || ObjFileInfo.ofiRelFileIsReq_bd900)
			    TrError = e_ok;
		}
		return; 
	}

    tbd_nodeptr &pPage            = NptrsPage.np_ptr();
	const bool   bPageIsForVarObj = (pt2VarObject_egg00 == NptrsPage.np_ptr()->nd_pt2());

	/* check if the page still belongs to the wanted file PTS 1105630 Alexander Kley 16/02/2000 */
	if (pPage->nd_obj_file_no() != Current.curr_tree_id.fileObjFileNo_gg00())
	{
		/* check if this error occured since the container was dropped */
			cbd900_ObjFileInfo &ObjFileInfo = bd90ObjFileDir[Current.curr_tree_id.fileObjFileNo_gg00()];
			if ((ObjFileInfo.ofiRoot_bd900 == NIL_PAGE_NO_GG00) || ObjFileInfo.ofiRelFileIsReq_bd900)
			    TrError = e_ok;
			else
				TrError = e_wrong_class_id;
		return; 
	}

    /* determine pointer to object frame */
    BD92CHECK_OID_POSITION (TrError, OId, NptrsPage);
    tbd900_ObjFrame  *pObjFrame = NptrsPage.bd910ObjFramePtr (OId.gg92GetPos());
    
    /* check if the wanted rollback is possible */
    switch (pObjFrame->objHeader_bd900.ohdState_gg92)
    {
    case obsReserved_egg00:
		 if (m_new != RollbackType || obsFree_egg00 != NewState)
        {
            TrError = e_wrong_object_state;
            return;
        }                    
        break;

    case obsKeyReserved_egg00:
        if (m_new != RollbackType                 || 
		   (obsFree_egg00         != NewState  &&
		    obsFreeAfterEot_egg00 != NewState))
        {
            TrError = e_wrong_object_state;
            return;
        }                    
        break;
        
    case obsOccupied_egg00:
        if (!((m_insert == RollbackType                                            &&
			((obsReserved_egg00    == NewState && pPage->nd_obj_key_len() == 0)  || 
		  	 (obsKeyReserved_egg00 == NewState && pPage->nd_obj_key_len() > 0 )))    ||
            (m_update == RollbackType && obsOccupied_egg00 == NewState)))
        {
            TrError = e_wrong_object_state;
            return;
        }                    
        break;
        
    case obsFreeAfterEot_egg00:
        if (m_delete != RollbackType
           ||
           (obsOccupied_egg00    != NewState &&
            obsReserved_egg00    != NewState &&
            obsKeyReserved_egg00 != NewState))
        {
            TrError = e_wrong_object_state;
            return;
        }                    
        break;

    default:
        TrError = e_wrong_object_state;
        return;
    }
	
	/* rollback object body if necessary */
	switch (NewState)
	{
	case obsReserved_egg00:
		if (bPageIsForVarObj)
		{
			/* remove continuation object */
			tbd900_VarObjBodyHeader & VarObjBodyHeader = ((tbd900_VarObjBody*) (pObjFrame->objBody_bd900))->vbdHeader_bd900;
			
			if (PRIM_CONT_OBJ_FILE_NO_BD900 != VarObjBodyHeader.vhdContObjFileNo_bd900)
			{
				bd93RemoveContObj (Current, OId, VarObjBodyHeader.vhdContObjFileNo_bd900,
					VarObjBodyHeader.vhdContObjPno_bd900, VarObjBodyHeader.vhdContObjPos_bd900);

				/* reset pointer to continuation object */
				VarObjBodyHeader.vhdContObjFileNo_bd900 = PRIM_CONT_OBJ_FILE_NO_BD900;
				VarObjBodyHeader.vhdContObjPno_bd900    = NIL_PAGE_NO_GG00;
				VarObjBodyHeader.vhdContObjPos_bd900    = NIL_OBJ_PAGE_POS_GG92;
				VarObjBodyHeader.vhdObjSize_bd900       = 0; 
			}		
		}
		break;

	case obsOccupied_egg00:
		if (pObjFrame->objHeader_bd900.ohdState_gg92 != obsFreeAfterEot_egg00)
		{
			/* in case the object state is obsFreeAfterEot_egg00 there */
			/* is nothing to do since the object body is still as it   */
			/* was before the object was deleted otherwise the before  */
			/* image of the object hast to be inserted                 */ 
			
			if (bPageIsForVarObj)
			{
				const bool bRollBack = true;
				tgg91_PageRef    DummyObjVers;
				DummyObjVers.gg91SetDummyRef();
				bd93UpdateVarObjBody (Current, OId, DummyObjVers, NptrsPage,
					pObjFrame, BeforeObjSize, pBeforeObj, bRollBack, 
					bTooManyEmptyPagesInContFile, ContObjFileNo, ContChainNo);
			}
			else
			{	
				/* check before object size */
				if (0 != BeforeObjSize || (obsOccupied_egg00 == NewState))
				{
					if (BeforeObjSize != tsp00_Int4(pPage->nd_obj_frame_len() - sizeof(tgg92_ObjHeader) -
						pPage->nd_obj_key_len()))
					{
						TrError = e_illegal_object_length;
						return;
					}              
					
					g10mv ("VBD90 ", 1,
						BeforeObjSize, 
						PAGE_MXSP00 - sizeof (tgg00_PageCheckTrailer), 
						pBeforeObj, 1,
						pPage, 
						OId.gg92GetPos() + POS_OFF_DIFF_BD00 + sizeof (tgg92_ObjHeader),
						BeforeObjSize, 
						TrError);
				}
			}
			if (TrError != e_ok) return;
		}
		break;
	}	
	
	/* update page header */
    if (obsFreeAfterEot_egg00 == pObjFrame->objHeader_bd900.ohdState_gg92)
    {    
        ++pPage->nd_occ_obj_cnt();
        --pPage->nd_free_eot_obj_cnt();
        
        /*  even if there is no more feot object the page  is not removed from the     */ 
        /*  chain of free after eot pages  since  it would cost too much. Removing the */
        /*  page from the chain will be done later by the GC                           */
    }
  
    /* update object header */
    pObjFrame->objHeader_bd900.ohdState_gg92                        = NewState;
    pObjFrame->objHeader_bd900.ohdTransInfo_gg92.otrBeforeRef_gg00  = BeforeObjRef; 
    pObjFrame->objHeader_bd900.ohdTransInfo_gg92.otrUpdTransId_gg00 = BeforeObjUpdVers;  
    
    if (obsFreeAfterEot_egg00 == NewState)
	{	
		/* check if the before image still exists PTS 1106144 29/03/2000 */
        if (!bd94History.bd94BeforeImageExists (*Current.curr_trans, OId, BeforeObjRef))
		{
			NewState.becomes(obsFree_egg00);
			pObjFrame->objHeader_bd900.ohdState_gg92 = NewState;
		}
		else
		{
			/* OId has its origin in the log  */
			pObjFrame->objHeader_bd900.ohdFrameVersion_gg92 = OId.gg92GetFrameVers(); 
			
			/* update page header */  
			++pPage->nd_free_eot_obj_cnt();
			--pPage->nd_occ_obj_cnt();    
		}
	}

    if (obsFree_egg00 == NewState)
    {  
        /* OId has its origin in the log  */
        pObjFrame->objHeader_bd900.ohdFrameVersion_gg92 = OId.gg92GetFrameVers(); 
        	
		/* update page header */  
		++pPage->nd_free_obj_cnt();
		--pPage->nd_occ_obj_cnt();
		
		/*  insert object into free object chain */
		pObjFrame->objHeader_bd900.ohdNextFreeObj_gg92 = pPage->nd_first_free_obj();           
		pPage->nd_first_free_obj()                     = OId.gg92GetPos(); 
		
		/* remove oid from hash */
		if (pPage->nd_obj_key_len()>0)
		{
			void *const pKey = (tsp00_Byte*) pPage + OId.gg92GetPos() + 
				pPage->nd_obj_frame_len() - pPage->nd_obj_key_len();
			bd22RemoveOid (*Current.curr_trans, Current.curr_tree_id.fileObjFileNo_gg00(),
				pPage->nd_obj_key_len(), pKey, OId);
			/*  PTS 1106012 Alexander Kley 16/03/00 */ 
			if ((e_hash_key_not_found == TrError) && (obsFree_egg00 == NewState))
				TrError = e_ok;
		}
		
		/* update global counter of all objects */
		const tsp00_Int4  ObjFileNo = pPage->nd_obj_file_no();
		const tsp00_Int2 &ChainNo   = pPage->nd_chain_no();
		bd90ObjFileDir[ObjFileNo][ChainNo].ociAllObjCnt_bd900--;
		
		/* update free page and empty counter on root page */
		if (NptrsPage.bd910PageMustBeInsertedIntoFreeChain() ||
		    NptrsPage.bd910EmptyPageCountMustBeIncreased())  
		{
			cbd900_ObjFileInfo &ObjFileInfo = bd90ObjFileDir[ObjFileNo];
			NptrsPage.bd910InsertPageIntoFreeChainAndUpdateEmptyPageCount(ObjFileInfo);
			
			if (e_ok != TrError)
			{
				/* check if this error occured since the container was dropped */
				/* PTS 1106671  Alexander Kley 19/05/2000 */
 				cbd900_ObjFileInfo &ObjFileInfo = bd90ObjFileDir[ObjFileNo];
				if ((ObjFileInfo.ofiRoot_bd900 == NIL_PAGE_NO_GG00) || ObjFileInfo.ofiRelFileIsReq_bd900)
				{
					TrError = e_ok;
					g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
						"bd90Rollba: File Droped ",	ObjFileNo);
				}
				else
				{
					g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
						"bd90Rollba:InsPageFailed",	ObjFileNo);
				}
				return; 
			}
		}
				
		// check if the number of objects is within reasonable limits  
#       if COMPILEMODE_MEO00 >= SLOW_MEO00 
		const cbd900_ObjChainInfo & ChainInfo = bd90ObjFileDir[ObjFileNo][ChainNo];
		if((0>ChainInfo.ociAllObjCnt_bd900) ||
			(ChainInfo.ociAllObjCnt_bd900 >
			ChainInfo.ociAllPageCnt_bd900 * bd90ObjFileDir[ObjFileNo].ofiMaxObjPerPageCnt_bd900))
		{
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
				"bd90Rollba:odd AllObjCnt",	ChainInfo.ociAllObjCnt_bd900 );
		}
#       endif
	}
    
    /* start garbage collector to remove empty pages from a continuation file */
	if (bTooManyEmptyPagesInContFile)
	{
		/* before the GC can be started all resources have to be returned */
        NptrsPage.bd910ReleasePage();
		Current.curr_tree_id.fileBdUse_gg00().delElement(bd_write_acc);
		const tgg00_ObjFileNo & FileNo = Current.curr_tree_id.fileObjFileNo_gg00(); 
		cbd900_ObjChainInfo &ChainInfo = bd93ContObjFileDir[FileNo][ContObjFileNo][ContChainNo];
        bd91GarbCollController.bd930StartGCToRemoveEmptyPages (*Current.curr_trans, ChainInfo);
	}
}

/*---------------------------------------------------------------------------*/

externCpp void
bd90UnlockObj (
             tbd_current_tree &Current,
             tgg92_KernelOid  &OId
             )
             
{
    ROUTINE_DBG_MEO00 ("bd90UnlockObj");
    
    tgg00_BasisError    &TrError = Current.curr_trans->trError_gg00;
    
    /*  get page pointer  */
	cbd910_PrimObjDataNode    NptrsPage(Current, OId.gg92GetPno(), nr_for_update);
    if (TrError != e_ok) return; 
    
    /* determine pointer to object frame */
    BD92CHECK_OID_POSITION (TrError, OId, NptrsPage);    
    tbd900_ObjFrame  *pObjFrame = NptrsPage.bd910ObjFramePtr (OId.gg92GetPos());

    /*  check object version  */
    if (pObjFrame->objHeader_bd900.ohdFrameVersion_gg92 != OId.gg92GetFrameVers()) 
    {
        TrError = e_wrong_object_version;
            bd90_NoteVersionErrorInKnldiag 
                (Unlock, OId, pObjFrame->objHeader_bd900.ohdFrameVersion_gg92);
        return;
    }
    
    /* check object state */
    if ((pObjFrame->objHeader_bd900.ohdState_gg92 != obsReserved_egg00    ) &&
        (pObjFrame->objHeader_bd900.ohdState_gg92 != obsKeyReserved_egg00 ) &&
        (pObjFrame->objHeader_bd900.ohdState_gg92 != obsOccupied_egg00   ))
    {
        TrError = e_wrong_object_state;
        return;
    }
    
    kb50UnlockObj (Current.curr_trans->trTaskId_gg00,
        Current.curr_trans->trTransId_gg00, Current.curr_trans->trIndex_gg00,
        OId, pObjFrame->objHeader_bd900.ohdTransInfo_gg92);   
}

/*---------------------------------------------------------------------------*/

externCpp void
bd90UpdateObj (
               tbd_current_tree       &Current,
               tgg92_KernelOid        &OId,
               tgg91_PageRef          &ObjVers,
               tsp00_Int4             NewObjBodySize,
               void                   *pNewObjBody
               )
               
{
    ROUTINE_DBG_MEO00 ("bd90UpdateObj");
    
    tgg00_BasisError      &TrError                      = Current.curr_trans->trError_gg00;
	bool                   bTooManyEmptyPagesInContFile = false;
    tsp00_Int4             ContFileNo;
	tsp00_Int4             ContChainNo;  
	const tgg00_ObjFileNo &FileNo                       = Current.curr_tree_id.fileObjFileNo_gg00();

	/* check if there is sufficient space to write rollback, history entries */
	/* and a new continuation object in case of objects of variable length   */                      
    b10check_space (10, Current);
   	if (e_ok != TrError) return;  // PTS 1107864 AK 15/09/2000

    /*  get page pointer  */
	cbd910_PrimObjDataNode    NptrsPage (Current, OId.gg92GetPno(), nr_for_update);
    if (TrError != e_ok) return;
    
	/* determine pointer to object frame */
    BD92CHECK_OID_POSITION (TrError, OId, NptrsPage);
    tbd900_ObjFrame  *pObjFrame = NptrsPage.bd910ObjFramePtr (OId.gg92GetPos());

	/* check object version */
    if (pObjFrame->objHeader_bd900.ohdFrameVersion_gg92 != OId.gg92GetFrameVers()) 
    {           
        TrError = e_wrong_object_version;
		bd90_NoteVersionErrorInKnldiag 
			(Update, OId, pObjFrame->objHeader_bd900.ohdFrameVersion_gg92);
		return;
    }

	const bool bPageIsForVarObj = (pt2VarObject_egg00 == NptrsPage.np_ptr()->nd_pt2());
 
	/* update object body */
	if (bPageIsForVarObj)
	{
		const bool bNoRollBack = false;
		bd93UpdateVarObjBody (Current, OId, ObjVers, NptrsPage, pObjFrame, 
			NewObjBodySize, pNewObjBody, bNoRollBack, bTooManyEmptyPagesInContFile, ContFileNo, ContChainNo); 
	}
	else
	{
		bd90_UpdateObjBody (Current, OId, ObjVers, NptrsPage.np_ptr(),
			pObjFrame,	NewObjBodySize, pNewObjBody);
	}
	
	if (TrError != e_ok) return;
	
	/*  set return value of result object sequence  */
	ObjVers = pObjFrame->objHeader_bd900.ohdTransInfo_gg92.otrBeforeRef_gg00;
	
	/*  update object frame header */
	pObjFrame->objHeader_bd900.ohdState_gg92.becomes(obsOccupied_egg00);

   /* start garbage collector to remove empty pages from a continuation file */
	if (bTooManyEmptyPagesInContFile)
	{
		/* before the GC can be started all resources have to be returned */
		NptrsPage.bd910ReleasePage();
		Current.curr_tree_id.fileBdUse_gg00().delElement(bd_write_acc);
        cbd900_ObjChainInfo &ChainInfo      = bd93ContObjFileDir[FileNo][ContFileNo][ContChainNo];
        bd91GarbCollController.bd930StartGCToRemoveEmptyPages (*Current.curr_trans, ChainInfo);		
	}
}

/*===========================================================================*
 *  LOCAL FUNCTIONS (CODE)                                                   *
 *===========================================================================*/

static void 
bd90_CopyChainInfoToSubRoots (tbd_current_tree   &Current,
							  cbd900_ObjFileInfo &ObjFileInfo)
{
	for (int iChain = 0; iChain<ObjFileInfo.ofiMaxChain_bd900; ++iChain)
	{    
		cbd900_ObjChainInfo & ChainInfo = ObjFileInfo[iChain];
		
		const tsp00_PageNo &SubRoot = ChainInfo.ociSubRoot_bd900;
		if (NIL_PAGE_NO_GG00 != SubRoot)
		{
			cbd910_SubRootNode  NptrsSubRoot (Current, SubRoot, nr_for_update);
			if (Current.curr_trans->trError_gg00 != e_ok)
			{
				g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
					"bd90Refresh: Error      ",	Current.curr_trans->trError_gg00 );
				g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
					"bd90Refresh: Subroot    ",	SubRoot);
				g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
					"bd90Refresh: FileNo     ",	Current.curr_tree_id.fileContObjFileNo_gg00());
				g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
					"bd90Refresh: ChainNo    ",	iChain);
				return;
			}
			
			/* get handler for the data stored on the root page body */
			cbd900_SubRootPageInfo  SubRootPageInfo (NptrsSubRoot);
			
			/* note that ociAllObjCnt_bd900 is updated dirty and need not be correct. Here it is */
			/* tried to keep the trivial restriction that the number of all objects must be >=0  */
			SubRootPageInfo.bd900AllObjCnt() = MAX (ChainInfo.ociAllObjCnt_bd900, 0);
		}
	}
}

/*---------------------------------------------------------------------------*/

static _INLINE void
bd90_DelAllObjOnThisPage (tbd_current_tree          &Current, 
						  cbd910_PrimObjDataNode    &NptrsPage, 
						  tsp00_Int4                &NumDelObj)
{
	ROUTINE_DBG_MEO00 ("bd90_DelAllObjOnThisPage");

	tgg00_BasisError    &TrError            = Current.curr_trans->trError_gg00;
	tbd_nodeptr         &pPage              = NptrsPage.np_ptr();
    tsp00_Uint2         &ObjFrameLen        = pPage->nd_obj_frame_len();


	const int            AlignedObjFrameLen = BD92ALIGN_OBJ_FRAME (ObjFrameLen);
	tgg00_ObjPagePos     ObjPos;
    tbd900_ObjFrame     *pObjFrame;
    tgg92_KernelOid      OId;
	tsp00_Uint2          AuxReqTimeout;
	
    /* go trough all objects on this page */
	for (ObjPos = pPage->nd_first_obj();
	(ObjPos + ObjFrameLen) <= BODY_END_BD00;
	ObjPos += AlignedObjFrameLen)
    {
#       if COMPILEMODE_MEO00 >= QUICK_MEO00 
        BD92CHECK_OBJ_POSITION (TrError, ObjPos, NptrsPage);
#       endif		
        pObjFrame = NptrsPage.bd910ObjFramePtr (ObjPos);

		/* find objects which are not free */
        if (obsFree_egg00 != pObjFrame->objHeader_bd900.ohdState_gg92)
		{
            /* create OId for current object */
            OId.gg92Set (
                pPage->nd_id(),
                ObjPos,
                pObjFrame->objHeader_bd900.ohdFrameVersion_gg92);
            
			const tgg91_PageRef  & WantedObjVers = 
				pObjFrame->objHeader_bd900.ohdTransInfo_gg92.otrBeforeRef_gg00;
			
			/*  get kb lock (nowait) for object */    
			AuxReqTimeout = Current.curr_trans->trObjReqTimeout_gg00; //save
			Current.curr_trans->trObjReqTimeout_gg00 = 0; // set nowait; PTS 1002114 JA 1999-03-02
			
			kb67LockObj (*Current.curr_trans, Current.curr_tree_id, OId, WantedObjVers,
				pObjFrame->objHeader_bd900.ohdState_gg92,
				pObjFrame->objHeader_bd900.ohdTransInfo_gg92);
			
			Current.curr_trans->trObjReqTimeout_gg00 = AuxReqTimeout; // restore timeout
			
			if (e_ok != TrError) break;
			
			if (obsFreeAfterEot_egg00 != pObjFrame->objHeader_bd900.ohdState_gg92)
			{
				/* check if object is locked and create a log entry  */
				
				const bool bPageIsForVarObj = (pt2VarObject_egg00 == NptrsPage.np_ptr()->nd_pt2());

				/* update object body */
				if (bPageIsForVarObj)
					bd93LogDelVarObj (Current, OId, WantedObjVers, pPage, pObjFrame);
				else
				{
					const int ObjBodySize = NptrsPage.np_ptr()->nd_obj_frame_len() - sizeof(tgg92_ObjHeader);
					
					kb67DelUpdObj (*Current.curr_trans, m_delete, Current.curr_tree_id,
						OId, pPage->nd_obj_key_len(), WantedObjVers, 
						pObjFrame->objHeader_bd900, ObjBodySize, pObjFrame->objBody_bd900);
				}        
				if (TrError != e_ok) break;
				
				/* increase counter of deleted objects */
				++NumDelObj;
				
				/* mark the object as deleted  */
				pObjFrame->objHeader_bd900.ohdState_gg92.becomes(obsFreeAfterEot_egg00); 
				
				/* update page header */  
				++pPage->nd_free_eot_obj_cnt();
				--pPage->nd_occ_obj_cnt();
			}
		}
	}
}

/*---------------------------------------------------------------------------*/

static _INLINE void
bd90_GetObjBody   (
				   tgg00_TransContext &Trans,
				   tgg92_KernelOid     OId, 
				   tbd_nodeptr         pPage, 
				   tsp00_Int4          ExtObjBodySize, 
				   void               *pExtObjBody, 
				   tbd900_ObjFrame     *pObjFrame  
				   )
{
	ROUTINE_DBG_MEO00 ("bd90_GetObjBody");
	
	/* this function copiess the object body identified by      */ 
	/* the OId at the position given by the pointer pExtObjBody */
	
	tgg00_BasisError    &TrError = Trans.trError_gg00;
	const tsp00_Uint2   KeyLen   = pPage->nd_obj_key_len();
	const tsp00_Int4 ObjBodySize = pPage->nd_obj_frame_len() - sizeof(tgg92_ObjHeader);
	
	/* check if the given memory is large enough to accomodate the object */
	if (ExtObjBodySize < ObjBodySize)
	{
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
			"BD90GOB:ExtObjLen<ObjLen", ExtObjBodySize);
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
			"BD90GOB:ExtObjLen<ObjLen", ObjBodySize);
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj,                           //PTS 1107471 AK 09/08/2000
			"BD90GOB: Pno            ", pPage->nd_id());
		b15dump_bad_page (Trans.trTaskId_gg00,'d', ".obj", pPage->nd_id(), (void *) pPage, 1); //PTS 1107471 AK 09/08/2000
		TrError = e_buffer_limit;
		return;
	}
	
	if (pObjFrame->objHeader_bd900.ohdState_gg92 == obsOccupied_egg00)
	{
		/* copy object */
		g10mv ("VBD90 ", 2, 
			PAGE_MXSP00 - sizeof (tgg00_PageCheckTrailer),
			ExtObjBodySize,
			pPage, 
			OId.gg92GetPos()+ POS_OFF_DIFF_BD00 + sizeof (tgg92_ObjHeader),
			pExtObjBody,         1,
			ObjBodySize, 
			TrError);
	}
	else
	{
		if (KeyLen > 0)
		{
			/* copy key only */
			g10mv ("VBD90 ", 3, 
				PAGE_MXSP00 - sizeof (tgg00_PageCheckTrailer), 
				ExtObjBodySize,
				pPage, 
				OId.gg92GetPos() + POS_OFF_DIFF_BD00 + sizeof (tgg92_ObjHeader) + ObjBodySize - KeyLen,
				pExtObjBody, 1 + ObjBodySize - KeyLen,
				KeyLen, 
				TrError);
			
			if (TrError != e_ok) {return;} 
		}
		
		/* fill the remaining object body with NIL_OBJ_FILLCHAR_BD900 */
		g10fil ("VBD90 ", 4, 
			ExtObjBodySize, pExtObjBody, 1, ObjBodySize - KeyLen, NIL_OBJ_FILLCHAR_BD900, TrError);
	} 
    
#   if COMPILEMODE_MEO00 >= SLOW_MEO00
    if (e_ok == TrError) t01moveobj (bi, pExtObjBody, 1, ObjBodySize);
#   endif   
    
}

/*---------------------------------------------------------------------------*/

static void 
bd90_NewObjFromThisPage (tbd_current_tree         &Current,
    					 cbd910_SubRootNode       &NptrsSubRoot,
						 cbd910_PrimObjDataNode   &NptrsPage, 
                         tsp00_Int4                KeyLen,
                         void                     *pKey,
                         tgg92_KernelOid          &OId,
                         tgg91_PageRef            &ResultObjVers)
{
    ROUTINE_DBG_MEO00 ("bd90_NewObjFromThisPage");
    
    
    /*  This routine finds the OId on a page presuming that at least one free  */
    /*  object frame exists. Moreover all pointers are restored corresponding  */
    /*  to changing the state of the frame whichs OId is returned              */

    tgg00_BasisError    &TrError          = Current.curr_trans->trError_gg00;
	tbd_nodeptr         &pPage            = NptrsPage.np_ptr();
    tsp00_Int4           ObjBodySize;
	const bool           bPageIsForVarObj = (pt2VarObject_egg00 == pPage->nd_pt2());

    /* check if the position of the first free obj is valid */
#   if COMPILEMODE_MEO00 >= QUICK_MEO00 
	if ( !bd92ObjPosIsValid (pPage->nd_first_free_obj(), pPage->nd_first_obj(), pPage->nd_obj_frame_len()))
	{	
		g01abort (csp3_b90c1_wrong_slotno, csp3_n_obj, 
			"BD90NewObj:wrong FreePos", pPage->nd_first_free_obj());
    };
#   endif

	/* construct position of the object frame  header */
	tbd900_ObjFrame  *pObjFrame = NptrsPage.bd910ObjFramePtr (pPage->nd_first_free_obj());
	
	/* check if the object frame is really free */
	if (obsFree_egg00 != pObjFrame->objHeader_bd900.ohdState_gg92)
	{
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
			"BD90NewObj: wrong State", pPage->nd_first_free_obj());
		g01abort (csp3_bd_msg, csp3_n_obj, 
				"BD90NewObj: wrong State ", pObjFrame->objHeader_bd900.ohdState_gg92);
	}

    /* construct new OId   */
    OId.gg92Set (
        pPage->nd_id(),
        pPage->nd_first_free_obj(),
        MIN_OBJ_FRAME_VERS_GG92);
    
    if (pObjFrame->objHeader_bd900.ohdFrameVersion_gg92 < MAX_OBJ_FRAME_VERS_GG92)
        OId.gg92SetFrameVers (pObjFrame->objHeader_bd900.ohdFrameVersion_gg92 + 1);
	else
		OId.gg92SetFrameVers (MIN_OBJ_FRAME_VERS_GG92);
    
    /* determine objbodysize */
	if (bPageIsForVarObj)
		ObjBodySize = 0;  // since for variable objects the later length is unknown
	else
		ObjBodySize = pPage->nd_obj_frame_len() - sizeof(tgg92_ObjHeader);

	/* check if the KeyLen corresponds to the KeyLen stored in the page header */
#   if COMPILEMODE_MEO00 >= QUICK_MEO00 
	g01check (csp3_b90_wrong_obj_key_len , csp3_n_obj, "BD90_NewObj: bad KeyLen ",  
		KeyLen, pPage->nd_obj_key_len() == KeyLen);
#   endif

	/* insert key into object body */
    if  (KeyLen > 0) 
    {        
        g10mv ("VBD90 ", 5, 
            KeyLen, ObjBodySize,
            pKey, 1, &(pObjFrame->objBody_bd900), ObjBodySize - KeyLen + 1, 
            KeyLen, Current.curr_trans->trError_gg00);  
		if (TrError != e_ok) {return;}  
    }
    
    // new unused object ==> lock collision impossible ==> set nowait to avoid any kind of request 
    tsp00_Uint2 AuxReqTimeout = Current.curr_trans->trObjReqTimeout_gg00; // save timeout
    Current.curr_trans->trObjReqTimeout_gg00 = 0; // set nowait 
    
    /* put current lock id into object frame header and create a log entry   */
    kb67NewObj (*Current.curr_trans, Current.curr_tree_id, OId, pObjFrame->objHeader_bd900,
        ObjBodySize, KeyLen);
    
    Current.curr_trans->trObjReqTimeout_gg00 = AuxReqTimeout; // restore timeout
    
    if (TrError != e_ok) {return;}
    
	// necessary for NewHashKey
	ResultObjVers = pObjFrame->objHeader_bd900.ohdTransInfo_gg92.otrBeforeRef_gg00;
	
	/* update page header */
	-- pPage->nd_free_obj_cnt();
	++ pPage->nd_occ_obj_cnt();
	pPage->nd_first_free_obj() = pObjFrame->objHeader_bd900.ohdNextFreeObj_gg92;
	
	/* update subroot page header */
	if (pPage->nd_free_obj_cnt() <= 0)              
		NptrsPage.bd910RemovePageFromFreeChain (NptrsSubRoot, NptrsSubRoot);
	
	/*  update empty page counter  */
	if (pPage->nd_empty_page())
	{
		pPage->nd_empty_page() = false;
		
        cbd900_SubRootPageInfo  SubRootPageInfo (NptrsSubRoot);
		--SubRootPageInfo.bd900EmptyPageCnt();  
	}
	
	/* update object frame header */
	pObjFrame->objHeader_bd900.ohdNextFreeObj_gg92  = NIL_OBJ_PAGE_POS_GG92;
	pObjFrame->objHeader_bd900.ohdFrameVersion_gg92 = OId.gg92GetFrameVers();
	pObjFrame->objHeader_bd900.ohdState_gg92.becomes(
		(KeyLen == 0) ? obsReserved_egg00 : obsKeyReserved_egg00);
	
	/* initialize the object body, i.e. the first four bytes  */
	/* of the body are filled with                            */
	int  ObjSize       = ObjBodySize - KeyLen;
	char InitString[5] = "\x00\x00\x00\x00";
	
	if (bPageIsForVarObj)
	{
		tbd900_VarObjBody & VarObjBody = *((tbd900_VarObjBody*) (pObjFrame->objBody_bd900));
		VarObjBody.vbdHeader_bd900.bd900SetToEmptyHeader();
		memcpy (VarObjBody.vbdBody_bd900, InitString, MIN (4,ObjSize));
	}
	else
		memcpy (pObjFrame->objBody_bd900,InitString, MIN (4,ObjSize));  
	
	/* increase global counter of used object frames in the current chain */   
	++ bd90ObjFileDir[pPage->nd_obj_file_no()][pPage->nd_chain_no()].ociAllObjCnt_bd900;
	
#   if COMPILEMODE_MEO00 >= SLOW_MEO00 
    // check if the number of objects is within reasonable limits  
	const cbd900_ObjChainInfo & ChainInfo =	bd90ObjFileDir[pPage->nd_obj_file_no()][pPage->nd_chain_no()];
	if((0>ChainInfo.ociAllObjCnt_bd900) ||
		(ChainInfo.ociAllObjCnt_bd900 >
		ChainInfo.ociAllPageCnt_bd900 * bd90ObjFileDir[pPage->nd_obj_file_no()].ofiMaxObjPerPageCnt_bd900))
	{
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
			"bd90NewObj:odd AllObjCnt",	ChainInfo.ociAllObjCnt_bd900 );
	}
#   endif

};

/*---------------------------------------------------------------------------*/

static _INLINE void
bd90_NextObjConsistentOnThisPage (tbd_current_tree        &Current,
                                  tgg92_KernelOid         &OId,
                                  tgg91_PageRef           &ResultObjVers,
                                  bool                     bWithObject,
                                  tsp00_Int4               ExtObjBodySize,
                                  void                    *pExtObjBody, 
                                  cbd910_PrimObjDataNode  &NptrsPage,    
								  tsp00_Int4              &ObjBodySize,             //[out=length of the object] 
                                  bool                    &bNextObjIsFound,
								  tsp00_Int4              &LogReadAccessesPerPage)  //[out] PTS 1107819 AK 12/09/2000
{
    ROUTINE_DBG_MEO00 ("bd90_NextObjConsistentOnThisPage");
    
    bNextObjIsFound = false;

    tgg00_ObjPagePos          ObjPos; 
    tbd900_ObjFrame            *pObjFrame; 
    tsp00_Int4                ObjPosBeg;
    tgg00_ObjState            ObjState;
    tgg00_ObjFrameVers        FrameVers;
    tgg00_ObjTransInfo        ObjTransInfo;
    const tsp00_Int4          ObjFrameLen      = NptrsPage.np_ptr()->nd_obj_frame_len();        
    tsp00_Int4                ObjKeyLen        = NptrsPage.np_ptr()->nd_obj_key_len();
    tgg00_BasisError         &TrError          = Current.curr_trans->trError_gg00;
	const bool                bPageIsForVarObj = (pt2VarObject_egg00 == NptrsPage.np_ptr()->nd_pt2());
	tsp00_Int4                LogReadAccesses ;           // PTS 1107819 AK 12/09/2000
                              LogReadAccessesPerPage = 0; // PTS 1107819 AK 12/09/2000
							  
    /* find starting object position for search */    
    const int AlignedObjFrameLen = BD92ALIGN_OBJ_FRAME(ObjFrameLen);  
    
    if ( ! OId.gg92IsNil() )
        ObjPosBeg = OId.gg92GetPos() + AlignedObjFrameLen;
    else
	{
		/* return immediately if there are no objects on this page. PTS 1107750 AK 05/09/2000 */
		if (NptrsPage.np_ptr()->nd_max_obj_cnt() == NptrsPage.np_ptr()->nd_free_obj_cnt())
			return;

        ObjPosBeg = NptrsPage.np_ptr()->nd_first_obj();
	}
    
    /*  start search */
    for ( ObjPos = ObjPosBeg;
	    ((ObjPos + ObjFrameLen) <= BODY_END_BD00) && (!bNextObjIsFound);
	      ObjPos += AlignedObjFrameLen )
	{ 
	    /*  get pointer to the current object frame  */
#       if COMPILEMODE_MEO00 >= QUICK_MEO00 
		BD92CHECK_OBJ_POSITION (TrError, ObjPos, NptrsPage);     
#       endif 	  
		pObjFrame = NptrsPage.bd910ObjFramePtr (ObjPos);
			  
	    /* continue if the frame does not accomodate an object, PTS 1107750 AK 05/09/2000 */
		if (pObjFrame->objHeader_bd900.ohdState_gg92 == obsFree_egg00)
		continue;
			  
		/*  check current object frame and get a consistent image (if possible) */
		if ((pObjFrame->objHeader_bd900.ohdState_gg92 == obsOccupied_egg00    )    ||
		    (pObjFrame->objHeader_bd900.ohdState_gg92 == obsFreeAfterEot_egg00)    ||
		    ((pObjFrame->objHeader_bd900.ohdState_gg92 == obsKeyReserved_egg00) &&
		    !pObjFrame->objHeader_bd900.ohdTransInfo_gg92.otrBeforeRef_gg00.gg91IsNilRef())) /* PTS 1001768 Torsten Strahl 1998-11-30 */
		{           
            /* create OId for current object */
            OId.gg92Set (
                NptrsPage.np_ptr()->nd_id(),
                ObjPos,
                pObjFrame->objHeader_bd900.ohdFrameVersion_gg92);
            
            /* move object and key into given array assigned by pExtObjBody */
            ObjState     = pObjFrame->objHeader_bd900.ohdState_gg92;
            FrameVers    = pObjFrame->objHeader_bd900.ohdFrameVersion_gg92;
            ObjTransInfo = pObjFrame->objHeader_bd900.ohdTransInfo_gg92;
			
			/* get object body length */
			if (bPageIsForVarObj)
				ObjBodySize = bd90_VarObjBodySize (pObjFrame);
			else
				ObjBodySize = NptrsPage.np_ptr()->nd_obj_frame_len() - sizeof(tgg92_ObjHeader); 
            
            if (bWithObject)
            {     
                if (bPageIsForVarObj)
					bd93GetVarObjBody (Current, OId, ExtObjBodySize, pExtObjBody, pObjFrame);
				else
					bd90_GetObjBody (*Current.curr_trans, OId, NptrsPage.np_ptr(),
					    ExtObjBodySize, pExtObjBody, pObjFrame);
          
                if (TrError != e_ok) {return;}
            }
            
            /*  the current object may be inconsistent with the wanted view        */
            /*  which consists of objects commited before the current              */
            /*  transaction had been started in that case the old object image has */
            /*  to be restored from the log.                                       */
            
            kb67GetObjConsistent (*Current.curr_trans, OId, ObjState, ObjTransInfo,
                bWithObject, ObjKeyLen, ObjBodySize, ExtObjBodySize, pExtObjBody,
				FrameVers, ResultObjVers, LogReadAccesses);  // PTS 1107819 AK 12/09/2000
            
            /* increase counter of all object history pages accesses  PTS 1107819 AK 12/09/2000 */
			LogReadAccessesPerPage += LogReadAccesses;
             
            /* evaluate error message from kb67GetObjConsistent */            
            
            switch (TrError)
            {
            case e_ok:
                /*  a consistent image was found  */ 
                bNextObjIsFound = true;
                OId.gg92SetFrameVers (FrameVers);  // PTS 1001332 Alexander Kley 31.7.98 
                return;
                
            case e_object_not_found:
                /*  there is no consistent image of this object the */
                /*  search for the next object has to be continued  */
                /*  reset error to e_ok                             */
                TrError = e_ok;
                break;
                
            default :
                /* a real error occured */
                return;
            }
        }
        else  // T.A.
        {
			/* now the frame state must be either obsReserved_egg00 or obsKeyReserved_egg00 */
#           if COMPILEMODE_MEO00 >= QUICK_MEO00 
			if (pObjFrame->objHeader_bd900.ohdState_gg92 != obsReserved_egg00 &&
                pObjFrame->objHeader_bd900.ohdState_gg92 != obsKeyReserved_egg00)
            {
				/* additional safety output, PTS 1107750 AK 05/09/2000 */
				g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
					"BD90_NextObj: OId.Pno  ", NptrsPage.np_ptr()->nd_id());
				g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
					"BD90_NextObj: OId.Pos  ", ObjPos);
				g01abort (csp3_bd_msg, csp3_n_obj, 
					"BD90_NextObj:Wrong State", pObjFrame->objHeader_bd900.ohdState_gg92);
			}
#           endif

			/* check if new object has been created by current transaction */
			tsp00_Uint2 AuxReqTimeout = Current.curr_trans->trObjReqTimeout_gg00; //save
			Current.curr_trans->trObjReqTimeout_gg00 = 0; // set nowait;
			
			kb67LockObj(*Current.curr_trans, Current.curr_tree_id, OId,
				pObjFrame->objHeader_bd900.ohdTransInfo_gg92.otrBeforeRef_gg00,
				pObjFrame->objHeader_bd900.ohdState_gg92,
				pObjFrame->objHeader_bd900.ohdTransInfo_gg92);
			
			Current.curr_trans->trObjReqTimeout_gg00 = AuxReqTimeout; // restore timeout
			
			if (e_ok == TrError)
			{
				/*  object has been created by current transaction, return OId */
				/*  note : Object Body is of no account                        */
				
				OId.gg92Set (
					NptrsPage.np_ptr()->nd_id(),
					ObjPos,
					pObjFrame->objHeader_bd900.ohdFrameVersion_gg92);
				
				bNextObjIsFound = true;
				return;
			}
			else
			{
				if ((e_lock_collision == TrError) || (e_object_dirty == TrError)) 
					TrError = e_ok;
			}
        } // End T.A.
    }
}


/*---------------------------------------------------------------------------*/

static _INLINE void
bd90_NoteVersionErrorInKnldiag ( 
                                ebd90_FunctionName FunctionName,
                                tgg92_KernelOid    WantedOId,
                                tgg00_ObjFrameVers ExpectedVersion
                                )
{
#if COMPILEMODE_MEO00 >= SLOW_MEO00 

    ROUTINE_DBG_MEO00 ("bd90_NoteVersionErrorInKnldiag");

    /* first line */
    {   
        char Message[] = bsp_c40 ;
        int Len        = 0;
        
        switch (FunctionName)
        {
        case Delete     : g17nameto_line ("Del:              ",Len,Message);break;
        case Get        : g17nameto_line ("Get:              ",Len,Message);break; 
        case GetCons    : g17nameto_line ("GetCons:          ",Len,Message);break; 
        case KeyGet     : g17nameto_line ("KeyGet:           ",Len,Message);break;
        case KeyGetCons : g17nameto_line ("KeyGetCons:       ",Len,Message);break;
        case Lock       : g17nameto_line ("Lock:             ",Len,Message);break;
        case Reuse      : g17nameto_line ("Reuse:            ",Len,Message);break;
        case Unlock     : g17nameto_line ("Unlock:           ",Len,Message);break;
        case Update     : g17nameto_line ("Update:           ",Len,Message);break;
        default         : g17nameto_line ("OMS:              ",Len,Message);break;
        }
        
        ++Len;
        g17nameto_line ("Wrong Vers=         ", Len, Message);
        g17trimint4_to_line (WantedOId.gg92GetFrameVers(), Len, Message);
        ++Len;
        g17nameto_line ("Corr Vers=          ", Len, Message);
        g17trimint4_to_line (ExpectedVersion, Len, Message);
        g01optextmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, Message);
    }
    
    /* second line */
    {
        char Message[] = bsp_c40 ;
        int Len        = 0;
        
        g17nameto_line ("Wanted OId :        ", Len, Message);
        ++Len;
        WantedOId.gg92AppendToLine (Len, *(tsp00_Line*) &Message);

        g01optextmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, Message);  
    }
# endif
}

/*---------------------------------------------------------------------------*/

static void
bd90_UpdateObjBody (
					tbd_current_tree  &Current,
					tgg92_KernelOid    OId,
					tgg91_PageRef	   ObjVers,
					tbd_nodeptr        pPage,
					tbd900_ObjFrame    *pObjFrame, 
					tsp00_Int4         NewObjBodySize,
					void              *pNewObjBody
					)
{
	ROUTINE_DBG_MEO00 ("bd90_UpdateObjBody");

	/* This function moves the object given by the pointer  */
	/* pNewObjBody into the object frame given by the OId   */

	tgg00_BasisError    &TrError = Current.curr_trans->trError_gg00;

	/* check object state */
	if ((pObjFrame->objHeader_bd900.ohdState_gg92 != obsReserved_egg00   ) &&
        (pObjFrame->objHeader_bd900.ohdState_gg92 != obsKeyReserved_egg00) &&
        (pObjFrame->objHeader_bd900.ohdState_gg92 != obsOccupied_egg00   ))
    {   
        TrError = e_wrong_object_state;
        return;
    }
	
    /*  check if the NewObjectBody fits into the ObjectFrame */
	tsp00_Int4  ObjBodySize = pPage->nd_obj_frame_len() - sizeof(tgg92_ObjHeader);   
    if (NewObjBodySize != ObjBodySize)
    {
        TrError = e_illegal_object_length;
        return;
    }

    /*  lock object and  creates an entry in the logging */
    /*  list. if the object is locked already an error   */
    /*  code = e_lock_collision is returned.             */
    /*  if the input sequence number ObjVers  is too old */
    /*  an error  code is sent                           */
    
    kb67DelUpdObj (*Current.curr_trans, m_update, Current.curr_tree_id,
        OId, pPage->nd_obj_key_len(), ObjVers,
        pObjFrame->objHeader_bd900, ObjBodySize,
        pObjFrame->objBody_bd900);                              
    
    if (TrError != e_ok) {return;}
    
    /*  insert new object body into the object frame */
    g10mv ("VBD90 ", 6,
        ObjBodySize, 
        PAGE_MXSP00 - sizeof (tgg00_PageCheckTrailer), 
        pNewObjBody,         1,
        pPage, 
        OId.gg92GetPos() + POS_OFF_DIFF_BD00 + sizeof (tgg92_ObjHeader),
        NewObjBodySize - pPage->nd_obj_key_len(), // no key update
        TrError); 
}

/*---------------------------------------------------------------------------*/

static _INLINE tsp00_Int4
bd90_VarObjBodySize (
					 tbd900_ObjFrame *pObjFrame
					 )
{
   ROUTINE_DBG_MEO00 ("bd90_VarObjBodySize");

   	tbd900_VarObjBody       &VarObjBody = *((tbd900_VarObjBody*) (pObjFrame->objBody_bd900));
    return VarObjBody.vbdHeader_bd900.vhdObjSize_bd900;
};

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/ 
