/****************************************************************************

  module      : vbd900.cpp

  -------------------------------------------------------------------------

  responsible : AlexanderK

  special area: BD_Object_Constants_And_Types
  description : 


  version     : 7.2.0.0
  last changed: 1999-09-08  11:49
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1998



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gbd900.h"
#include "gbd910.h"
 
#include "hgg01.h"
#include "hgg10.h"

#if COMPILEMODE_MEO00 >= SLOW_MEO00 
#include "hta99.h"
#endif

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define HUNDERT_PERCENT   100

/*===========================================================================*
 *  LOCAL FUNCTIONS (CODE)                                                   *
 *===========================================================================*/

static void
bd900_NewDumpPage (tgg00_VfFileref      &HostFile,
				   tsp00_Int4            BufferSize,
				   tsp00_Page           &Buffer,
				   tsp00_Int4           &OutPno,
				   tsp00_Int4           &OutPos,
				   tsp00_VfReturn       &HostError,
				   tsp00_ErrText        &ErrText,
				   const tsp00_Int2     &DumpCode,  
				   char *                DumpMark)
{
    tgg00_BasisError  MoveError = e_ok;

	/* start at a new dump page */
	g01new_dump_page (HostFile, Buffer, OutPno, OutPos, HostError, ErrText);
    if (vf_ok != HostError) return;
	
	g10mv ("VBD900", 1,    
		DUMPMARKSIZE_BD900, BufferSize, DumpMark, 1,
		Buffer, OutPos, DUMPMARKSIZE_BD900 , MoveError);
	OutPos += DUMPMARKSIZE_BD900;
	
	memcpy (Buffer + OutPos - 1, &DumpCode, sizeof(DumpCode));
	OutPos         += sizeof(DumpCode);
}

/*===========================================================================*
 *  IMPLEMENTATION OF METHODS DECLARED IN gbd900.h                           * 
 *===========================================================================*/

/*---------------------------------------------------------------------------*/
/* METHODS of cbd900_ObjFileInfo                                             */
/*---------------------------------------------------------------------------*/

externCpp void 
cbd900_ObjFileInfo::bd900DumpFileInfo (               
									   tgg00_VfFileref    &HostFile,
									   tsp00_Uint4         BufferSize,
									   tsp00_Page         &Buffer,
									   tsp00_Int4         &OutPno,
									   tsp00_Int4         &OutPos,
									   tsp00_VfReturn     &HostError,
									   tsp00_ErrText      &ErrText,
									   tsp00_Int2         &ElemsOnThisPage,
									   tsp00_Int4         &NumElemPos,
									   tsp00_Int4          FileNo,
									   tsp00_Int4          ContFileNo,
									   tsp00_Int2          DumpCode,
									   char               *DumpMark)
{
    ROUTINE_DBG_MEO00 ("bd900DumpFileInfo");
	
	tsp00_Int4 iChain;

	/* check if the complete dump still fits into the    */ 
	/* current dump page. if not a new dump page is used */
    if ((OutPos - 1 
		+ sizeof (FileNo)
		+ sizeof (ContFileNo)
        + sizeof (ofiCurrChain_bd900)
        + sizeof (ofiMaxChain_bd900) 
        + sizeof (ofiMaxObjPerPageCnt_bd900)
        + sizeof (ofiMaxObjBodySize_bd900)
        + sizeof (ofiRoot_bd900)
		+ sizeof (ofiKeyRoot_bd900) 
		+ sizeof (ofiFileIsForVarObj_bd900)
		+ sizeof (ofiRelFileIsReq_bd900)
		+ ofiMaxChain_bd900 * 
		  (sizeof (cbd900_ObjChainInfo) + sizeof (iChain))) > BufferSize)
	{
        memcpy (Buffer + NumElemPos - 1, &ElemsOnThisPage, sizeof(ElemsOnThisPage));
        
		bd900_NewDumpPage (HostFile, BufferSize, Buffer, OutPno, OutPos, HostError,
			ErrText, DumpCode, DumpMark);
			
		NumElemPos      = OutPos;
	    ElemsOnThisPage = 0;
	    OutPos          += sizeof (ElemsOnThisPage);
    }
	
	++ElemsOnThisPage;
	memcpy (Buffer + NumElemPos - 1, &ElemsOnThisPage, sizeof(ElemsOnThisPage));
	
	memcpy (Buffer + OutPos - 1, &FileNo, sizeof(FileNo));
	OutPos         += sizeof(FileNo);

	memcpy (Buffer + OutPos - 1, &ContFileNo, sizeof(ContFileNo));
	OutPos         += sizeof(ContFileNo);
	
	memcpy (Buffer + OutPos - 1, &ofiMaxChain_bd900, sizeof(ofiMaxChain_bd900));
	OutPos         += sizeof(ofiMaxChain_bd900);
	
	memcpy (Buffer + OutPos - 1, &ofiCurrChain_bd900, sizeof(ofiCurrChain_bd900));
	OutPos         += sizeof(ofiCurrChain_bd900);
	
	memcpy (Buffer + OutPos - 1, &ofiMaxObjPerPageCnt_bd900, sizeof(ofiMaxObjPerPageCnt_bd900));
	OutPos         += sizeof(ofiMaxObjPerPageCnt_bd900);
	
	memcpy (Buffer + OutPos - 1, &ofiMaxObjBodySize_bd900, sizeof(ofiMaxObjBodySize_bd900));
	OutPos         += sizeof(ofiMaxObjBodySize_bd900);
	
	memcpy (Buffer + OutPos - 1, &ofiRoot_bd900, sizeof(ofiRoot_bd900));
	OutPos         += sizeof(ofiRoot_bd900);
	
	memcpy (Buffer + OutPos - 1, &ofiKeyRoot_bd900, sizeof(ofiKeyRoot_bd900));
	OutPos         += sizeof(ofiKeyRoot_bd900);

	memcpy (Buffer + OutPos - 1, &ofiFileIsForVarObj_bd900, sizeof(ofiFileIsForVarObj_bd900));
	OutPos         += sizeof(ofiFileIsForVarObj_bd900);

    memcpy (Buffer + OutPos - 1, &ofiRelFileIsReq_bd900, sizeof(ofiRelFileIsReq_bd900));
	OutPos         += sizeof(ofiRelFileIsReq_bd900);

	// --- CHAIN STATE ---
	
	for (iChain = 0; iChain < ofiMaxChain_bd900; ++iChain)
	{
		memcpy (Buffer + OutPos - 1, &iChain, sizeof(iChain));
		OutPos         += sizeof(iChain);
		
		operator[](iChain).bd900DumpChainInfo (BufferSize, Buffer, OutPos);
	}   
}

/*---------------------------------------------------------------------------*/

externCpp void 
cbd900_ObjFileInfo::bd900ResetFileInfo ()
{
    ROUTINE_DBG_MEO00 ("bd900ResetFileInfo");
    
	ofiRoot_bd900             = NIL_PAGE_NO_GG00;
	ofiKeyRoot_bd900          = NIL_PAGE_NO_GG00;
    ofiCurrChain_bd900        = 0;
    ofiMaxChain_bd900         = 0;
	ofiMaxObjPerPageCnt_bd900 = 0;
    ofiMaxObjBodySize_bd900   = 0;
	ofiRelFileIsReq_bd900     = false;
	ofiFileIsForVarObj_bd900  = false;
	ofiFiller_bd900           = 0;
	
    for (int iChain=0; iChain < MAX_NUM_CHAINS_PER_CONTAINER_BD900; ++iChain)
    {
        ofiChain_bd900[iChain].bd900ResetChainInfo();
    }
}

/*---------------------------------------------------------------------------*/

externCpp tsp00_Int4 
cbd900_ObjFileInfo::bd900NumAllPages() const
{
    ROUTINE_DBG_MEO00 ("bd900NumAllPages");
	
	int        iChain; 
	tsp00_Int4 AllPageInFileCnt = 0;
	
	/* go trough all subroots of the file and sum up the number of  pages */
	for (iChain=0; iChain < ofiMaxChain_bd900; ++iChain)
		AllPageInFileCnt +=  ofiChain_bd900[iChain].ociAllPageCnt_bd900;

	return AllPageInFileCnt;
}

/*---------------------------------------------------------------------------*/

externCpp tsp00_Int4 
cbd900_ObjFileInfo::bd900NumOccPages() const
{
    ROUTINE_DBG_MEO00 ("bd900NumOccPages");
	
	int        iChain;
    tsp00_Int4 OccPageInFileCnt = 0;
	
	/* go trough all subroots of the file and sum up the number of  pages */
	for (iChain=0; iChain<= ofiMaxChain_bd900-1; ++iChain)
	{
		if ((ofiMaxObjPerPageCnt_bd900                 > 0)  && 
			(ofiChain_bd900[iChain].ociAllObjCnt_bd900 > 0))  //PTS 1103239 AK 08-09-1999
		{
			OccPageInFileCnt += (ofiChain_bd900[iChain].ociAllObjCnt_bd900*HUNDERT_PERCENT)
				/ofiMaxObjPerPageCnt_bd900;
		}
	}
		
	return OccPageInFileCnt;
}

/*---------------------------------------------------------------------------*/

externCpp void
cbd900_ObjFileInfo::bd900GetNumOfActiveGCs (tsp00_Int4 & NumReqGC,
                                            tsp00_Int4 & NumActiveGC) 
{	
	int  iChain;

	/* calculate number of chains which request a GC service */
	if (ofiRelFileIsReq_bd900)
	{
		NumReqGC = ofiMaxChain_bd900;
		
		for (iChain=0; iChain<ofiMaxChain_bd900; ++iChain) 
		{
			if (NIL_PAGE_NO_GG00 == operator[](iChain).ociSubRoot_bd900 )
				--NumReqGC;
		}
	}
	else
	{
		NumReqGC = 0;

		for (iChain=0; iChain <ofiMaxChain_bd900; ++iChain) 
		{
			if (operator[](iChain).ociRelPageIsReq_bd900)
				++NumReqGC;
		};
	}
	
	/* calculate number of chains where the GC is activ */
	NumActiveGC = 0;
	for (iChain=0; iChain<ofiMaxChain_bd900; ++iChain) 
	{		
		if (NIL_GC_INDEX_BD900 != operator[](iChain).ociCurrActivGC_bd900)
			++NumActiveGC;
	}
}

/*---------------------------------------------------------------------------*/

void 
cbd900_ObjFileInfo::bd900SetFileInfo (tbd_current_tree     &Current,
									  tbd_node_ptrs        &NptrsRoot,
									  tsp00_Int4            MaxObjBodySize,
									  tsp00_Int4            MaxObjPerPageCnt)
{      
	int                 iChain;
	cbd900_RootPageInfo RootPageInfo (NptrsRoot);

	bd900ResetFileInfo();
    ofiRoot_bd900             = NptrsRoot.np_ptr()->nd_id();
    ofiMaxChain_bd900         = RootPageInfo.bd900NumChains();
    ofiCurrChain_bd900        = 0;
    ofiMaxObjPerPageCnt_bd900 = MaxObjPerPageCnt;
    ofiMaxObjBodySize_bd900   = MaxObjBodySize;
	ofiFileIsForVarObj_bd900  = (pt2VarObject_egg00 == NptrsRoot.np_ptr()->nd_pt2());
    ofiRelFileIsReq_bd900     = RootPageInfo.bd900FileIsDeleted(); /* PTS 1106900 Alexander Kley 09/06/2000 */	
	
	for (iChain=0; iChain < RootPageInfo.bd900NumChains(); ++iChain)
	{
		cbd910_SubRootNode NptrsSubRoot (Current, RootPageInfo.bd900SubRoot(iChain), nr_for_read);
		ofiChain_bd900[iChain].bd900SetChainInfo (NptrsSubRoot);

		/* note: that since the current objectfileinfo is not yet inserted */
		/* into the directory some of the checks would report errors       */
		NptrsSubRoot.bd910ReleasePageWithoutCheck();
	}
	
	for (iChain=RootPageInfo.bd900NumChains(); iChain < MAX_NUM_CHAINS_PER_CONTAINER_BD900; ++iChain)
		ofiChain_bd900[iChain].bd900ResetChainInfo();
}

/*---------------------------------------------------------------------------*/
/* METHODS of cbd900_ObjFileDir                                              */
/*---------------------------------------------------------------------------*/
 
void
cbd900_ObjFileDir::bd900DumpObjFileDir (tgg00_VfFileref      &HostFile,
                                        tsp00_Int4            BufferSize,
                                        tsp00_Page           &Buffer,
                                        tsp00_Int4           &OutPno,
                                        tsp00_Int4           &OutPos,
                                        tsp00_VfReturn       &HostError,
                                        tsp00_ErrText        &ErrText)
{
    ROUTINE_DBG_MEO00 ("bd900DumpObjFileDir");
     
    tsp00_Int2        DumpCodeFileDir    = 250;
	tsp00_Int2        DumpCodeFileInfo   = 251;
    char              DumpMarkFileDir[]  = "B90OFDIR"; 
	char              DumpMarkFileInfo[] = "B90OFINF";
    tgg00_BasisError  MoveError        = e_ok;

    // OutPos is handled as a Pascal (!) Position
    // only for C-Routines 1 must be subtracted
			
    HostError.becomes(vf_ok);
	bd900_NewDumpPage (HostFile, BufferSize, Buffer, OutPno, OutPos, HostError,
		ErrText, DumpCodeFileDir, DumpMarkFileDir);
	
	/* dump maximum number of files */
	const tsp00_Int4 MaxNumFiles = MAX_OBJ_FILE_NO_GG92;
	memcpy (Buffer + OutPos - 1, &MaxNumFiles, sizeof(MaxNumFiles));
	OutPos         += sizeof(MaxNumFiles);
	
	/* dump number of files in the object file directory */
	int NumFiles = bd920NumItems();
	memcpy (Buffer + OutPos - 1, &NumFiles, sizeof(NumFiles));
	OutPos         += sizeof(NumFiles);
	
    /* dump entries in the Object file directory */
    g10mv ("VBD900", 2, DUMPMARKSIZE_BD900, BufferSize, DumpMarkFileInfo,
		1, Buffer, OutPos, DUMPMARKSIZE_BD900 , MoveError);
	OutPos += DUMPMARKSIZE_BD900;
	
	memcpy (Buffer + OutPos - 1, &DumpCodeFileInfo, sizeof(DumpCodeFileInfo));
	OutPos         += sizeof(DumpCodeFileInfo);

	tsp00_Int4	NumElemPos      = OutPos;
	tsp00_Int2  ElemsOnThisPage = 0;
	OutPos      += sizeof (ElemsOnThisPage);
	
	cbd900_ObjFileDirIter Iter (*this);
	
	for (Iter.bd920First(); (NULL != Iter()) && (HostError == vf_ok); ++Iter)
	{
		Iter()->bd900DumpFileInfo ( HostFile, BufferSize,
			Buffer, OutPno, OutPos,	HostError, ErrText, ElemsOnThisPage, 
			NumElemPos,	Iter.bd920Key(), /* = FileNo */
			PRIM_CONT_OBJ_FILE_NO_BD900, DumpCodeFileInfo, DumpMarkFileInfo);
	}
	
	//flush number of entries an this page 
	memcpy (Buffer + NumElemPos - 1, &ElemsOnThisPage, sizeof(ElemsOnThisPage));
}

/*---------------------------------------------------------------------------*/
/* METHODS of cbd900_ObjFileDir                                              */
/*---------------------------------------------------------------------------*/

void 
cbd900_ContObjFileDir::bd900DumpContObjFileDir (tgg00_VfFileref      &HostFile,
												tsp00_Int4            BufferSize,
												tsp00_Page           &Buffer,
												tsp00_Int4           &OutPno,
												tsp00_Int4           &OutPos,
												tsp00_VfReturn       &HostError,
												tsp00_ErrText        &ErrText)
{
	ROUTINE_DBG_MEO00 ("bd900DumpContObjFileDir");	

    tsp00_Int2                 DumpCode   = 254;
    char                       DumpMark[] = "B93CODIR";
	cbd900_ContObjFileDirIter  ContObjFileIter (*this);
	
    // OutPos is handled as a Pascal (!) Position
    // only for C-Routines 1 must be subtracted
	
    HostError.becomes(vf_ok);
	
	bd900_NewDumpPage (HostFile, BufferSize, Buffer, OutPno, OutPos, HostError,
		ErrText, DumpCode, DumpMark);
	
	tsp00_Int4 NumElemPos       = OutPos;
	tsp00_Int2 ElemsOnThisPage  = 0;
	OutPos                     += sizeof (ElemsOnThisPage);
	
	for (ContObjFileIter.bd920First(); 
	(NULL != ContObjFileIter()) && (HostError == vf_ok);
	++ContObjFileIter)
	{	
		for (tsp00_Int4 iContFileNo=1; 
		(iContFileNo <= ContObjFileIter()->cfiNumContFiles_bd900) && (HostError == vf_ok);
		++iContFileNo)
		{
			if ((*ContObjFileIter())[iContFileNo].ofiRoot_bd900 != NIL_PAGE_NO_GG00)
			{
				(*ContObjFileIter())[iContFileNo].bd900DumpFileInfo (
					HostFile, BufferSize, Buffer, OutPno, OutPos,
					HostError, ErrText, ElemsOnThisPage, NumElemPos, 
					ContObjFileIter.bd920Key(), /* = FileNo*/
			        iContFileNo, DumpCode, DumpMark);
			}
		}
	}
	//flush number of entries an this page 
	memcpy (Buffer + NumElemPos - 1, &ElemsOnThisPage, sizeof(ElemsOnThisPage));
}

/*---------------------------------------------------------------------------*/

// end of vbd900.cpp
