/****************************************************************************

  module      : vbd91.cpp

  -------------------------------------------------------------------------

  responsible : AlexanderK

  special area: interface for the pascal routines to the OMSGarbageCollection
  description : description ...


  version     : 7.2.0.0
  last changed: 1999-08-02  11:32
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1998



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/
#include "hbd91.h"

#include "gsp03.h" 
#include "gsp03_3.h"
#include "gbd05.h"
#include "gbd900.h"
#include "gbd910.h" 
#include "gbd920.h"
#include "gbd930.h"

#include "hbd01.h"
#include "hbd17.h"
#include "hbd22.h"
#include "hbd92.h"
#include "hbd93.h"
#include "hbd30.h"
#include "hbd30_1.h"
#include "hgg01.h"
#include "hgg01_1.h"

#include "heo55k.h" //vbegexcl/vendexcl

#if COMPILEMODE_MEO00 >= SLOW_MEO00
#include "hta99.h"
#include "hta01.h"
#endif

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define RESTART true

#ifndef TRUE
 #define TRUE 1
#endif

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/

class cbd91_ErrorTraceObject
{
public:
	cbd91_ErrorTraceObject (tgg00_TransContext &Trans,
	    				  tgg00_FileId          &ObjFileId,
						  const tgg00_ObjFileNo &ObjFileNo,
						  tgg92_KernelOid       &OId);
    ~cbd91_ErrorTraceObject();

	void bd91WriteVersionErrorMessage (const tgg00_ObjFileNo   ObjFileNoOnPage,
		                               const tgg91_PageRef     WantedObjVers,
	 	                               tbd900_ObjFrame        *pObjFrame);

	void bd91SetStatus (tgg12_RelObjStatus_Enum Status){m_RelObjStatus.becomes(Status);};
private:
	tgg00_TransContext     &m_Trans;
	tgg00_FileId           &m_ObjFileId;
    const tgg00_ObjFileNo  &m_ObjFileNo;
    tgg92_KernelOid        &m_OId;
	tgg12_RelObjStatus       m_RelObjStatus;

    void bd91_WriteStartErrorMessage ();
	void bd91_WriteFinalErrorMessage ();
};

/*---------------------------------------------------------------------------*/

class cbd91_HistoryNode  // PTS 1106180 AK 29/03/2000
{
private:
	tsp00_PageNo            m_HistoryPageNo;
    tbd_node_ptrs          &m_HistoryNodePtrs;
	tbd_current_tree       &m_HistoryCurrent;

public:

	cbd91_HistoryNode (tbd_node_ptrs &HistoryNodePtrs, tbd_current_tree &HistoryCurrent)
		:m_HistoryNodePtrs (HistoryNodePtrs), m_HistoryCurrent (HistoryCurrent) {};

	~cbd91_HistoryNode()
	{
		if (NULL == m_HistoryNodePtrs.np_ptr() && (e_ok == m_HistoryCurrent.curr_trans->trError_gg00))
			bd13GetNode (m_HistoryCurrent, m_HistoryPageNo, plmLock_ebd00, nr_for_update, m_HistoryNodePtrs);
	}

	void bd91Release()
	{
		if (NULL != m_HistoryNodePtrs.np_ptr())
		{
			m_HistoryPageNo = m_HistoryNodePtrs.np_ptr()->nd_id();
            b13w_release_node (m_HistoryNodePtrs, m_HistoryCurrent);
		}
	}
};

/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/

extern cbd900_ObjFileDir      bd90ObjFileDir;
extern cbd900_ContObjFileDir  bd93ContObjFileDir;

/*===========================================================================*
 *  LOCAL VARIABLES                                                          *
 *===========================================================================*/

cbd930_GarbCollController bd91GarbCollController;

/*===========================================================================*
 *  LOCAL FUNCTIONS (CODE)                                                   *
 *===========================================================================*/

inline void
cbd91_ErrorTraceObject::bd91WriteVersionErrorMessage (const tgg00_ObjFileNo   ObjFileNoOnPage,
												      const tgg91_PageRef     WantedObjVers,
													  tbd900_ObjFrame        *pObjFrame)
{
	/* write message into vtrace */
	tgg12_OmsVersionErrorTrace TrLineVersError;
	tgg12_OmsErrorTrace        TrLineError;
	
	TrLineVersError.oveTrType_gg12.becomes  (otVersionError_egg12);
	TrLineVersError.oveTrType2_gg12.becomes (ot2Nil_egg12);
	TrLineVersError.oveObjId_gg12           = m_OId;
	TrLineVersError.oveWantedVers_gg12      = WantedObjVers;
	TrLineVersError.oveObjFileNoOnPage_gg12 = ObjFileNoOnPage;
	
	if (NULL != pObjFrame)
	{
		TrLineVersError.oveObjVers_gg12         = pObjFrame->objHeader_bd900.ohdTransInfo_gg92.otrBeforeRef_gg00;
		TrLineVersError.oveFrameVers_gg12       = pObjFrame->objHeader_bd900.ohdFrameVersion_gg92;
	}
	
	b12vtrace (m_Trans, bd, b91garbcol, sizeof (TrLineVersError), &TrLineVersError);
	
	TrLineError.oerrTrType_gg12.becomes (otError_egg12);
	TrLineError.oerrTrType2_gg12.becomes (ot2Nil_egg12);
	TrLineError.oerrRoot_gg12      = m_ObjFileId.fileRoot_gg00();
	TrLineError.oerrError_gg12     = m_Trans.trError_gg00;
	TrLineError.oerrObjFileNo_gg12 = m_ObjFileNo;
	
	b12vtrace (m_Trans, bd, b91garbcol, sizeof (TrLineError), &TrLineError);
	
	/* write note into knldiag */
	g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj,
				"BD91RelObj:WrongObjState", m_ObjFileNo);
}

/*---------------------------------------------------------------------------*/

inline void
cbd91_ErrorTraceObject::bd91_WriteStartErrorMessage()
{
	tgg12_OmsOidTrace TrLineOId;

	TrLineOId.ooidTrType_gg12.becomes(otOid_egg12);
	TrLineOId.ooidTrType2_gg12.becomes(ot2Input_egg12);
	TrLineOId.ooidObjId_gg12   = m_OId;
	b12vtrace (m_Trans, bd, b91garbcol, sizeof (TrLineOId), &TrLineOId);
}

/*---------------------------------------------------------------------------*/

inline void
cbd91_ErrorTraceObject::bd91_WriteFinalErrorMessage ()
{	
	tgg12_OmsRelOidTrace TrLineRelOId;
	
	TrLineRelOId.roidTrType_gg12.becomes(otRelOid_egg12);
	TrLineRelOId.roidTrType2_gg12.becomes(ot2Output_egg12);
	TrLineRelOId.roidObjId_gg12        = m_OId;
	TrLineRelOId.roidRelObjStatus_gg12 = m_RelObjStatus;

	b12vtrace (m_Trans, bd, b91garbcol, sizeof (TrLineRelOId), &TrLineRelOId);
		
	if (e_ok != m_Trans.trError_gg00)
	{
		tgg12_OmsErrorTrace TrLineOut;
		
		TrLineOut.oerrTrType_gg12.becomes  (otError_egg12);
		TrLineOut.oerrTrType2_gg12.becomes (ot2Nil_egg12);
		TrLineOut.oerrRoot_gg12            = m_ObjFileId.fileRoot_gg00();
		TrLineOut.oerrObjFileNo_gg12       = m_ObjFileNo;
		TrLineOut.oerrError_gg12           = m_Trans.trError_gg00;
		
		b12vtrace (m_Trans, bd, b91garbcol, sizeof (TrLineOut), &TrLineOut);
	}
}

/*---------------------------------------------------------------------------*/

inline
cbd91_ErrorTraceObject::cbd91_ErrorTraceObject (tgg00_TransContext &Trans,
											tgg00_FileId          &ObjFileId,
											const tgg00_ObjFileNo &ObjFileNo,
											tgg92_KernelOid       &OId)
:
m_Trans       (Trans),
m_ObjFileId   (ObjFileId),
m_ObjFileNo   (ObjFileNo),
m_OId         (OId)
{
	m_RelObjStatus.becomes(rosObjSucessfullyReleased_gg12);

	if (g01vtrace.vtrOmsFree_gg00)
		bd91_WriteStartErrorMessage();
};

/*---------------------------------------------------------------------------*/

inline
cbd91_ErrorTraceObject::~cbd91_ErrorTraceObject ()
{
	if (g01vtrace.vtrOmsFree_gg00 || (m_Trans.trError_gg00 != e_ok))
		bd91_WriteFinalErrorMessage();
}

/*===========================================================================*
 *  GLOBAL FUNCTIONS (CODE)                                                  *
 *===========================================================================*/

externCpp cbd900_ObjFileInfo*
bd91AddFileToObjFileDir (tgg00_TransContext   &Trans,
						 tgg00_FileId         &FileId,
						 cbd910_RootNode      &NptrsRoot) 
{
	ROUTINE_DBG_MEO00 ("bd91AddFileToObjFileDir");

    /* PTS 1106716 Alexander Kley 01/06/2000                 */
	
	tbd_nodeptr            &pRoot            = NptrsRoot.np_ptr();
	const tgg00_ObjFileNo  &FileNo           = FileId.fileObjFileNo_gg00();
	const tsp00_Int4        ContFileNo       = FileId.fileContObjFileNo_gg00();
	const bool              bFileIsForVarObj = (pt2VarObject_egg00 == pRoot->nd_pt2());
	const int               ObjFrameLen      = pRoot->nd_obj_frame_len();
	int                     MaxObjBodySize;
    int                     MaxObjPerPageCnt;
	tgg00_BasisError       &TrError          = Trans.trError_gg00;
	
	/* determine maximum number of objects which can be stored on this page */
	const int AlignedObjFrameLen = BD92ALIGN_OBJ_FRAME(ObjFrameLen);
	if  (FULLCOVERING_BD00 >= ObjFrameLen)
		MaxObjPerPageCnt = 1 + ((FULLCOVERING_BD00 - ObjFrameLen)/AlignedObjFrameLen);
	else
	{
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj,
			"BD91AddFile: TooLongObj ", ObjFrameLen);
		Trans.trError_gg00 = e_too_long_objects;
		return NULL;
	}
	
	/* register object file at the global structure bd90ObjFileDir */
	if (bFileIsForVarObj)
		MaxObjBodySize = ObjFrameLen - sizeof (tgg92_ObjHeader) - sizeof (tbd900_VarObjBodyHeader);
	else
		MaxObjBodySize = ObjFrameLen - sizeof (tgg92_ObjHeader);
	
	/* create entry for the object file directory */
	cbd05_Current Current (Trans, m_select, FileId);
	cbd900_ObjFileInfo ObjFileInfo (Current, NptrsRoot, MaxObjBodySize, MaxObjPerPageCnt);
	
	/* insert entry into object file directory */
	cbd900_ObjFileInfo*  pObjFileInfo =
		bd90ObjFileDir.bd920InsItem (Trans.trTaskId_gg00, FileNo, ObjFileInfo, TrError);
	
	/* check if the item was successfully inserted. If the key existed */
	/* already or there are no more entrie possible NULL is returned   */
	if (NULL == pObjFileInfo)
	{
		if (e_duplicate_hash_key == TrError)
		{
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj,
				"BD91AddFile: Dupl FileNo", FileNo);
			Trans.trError_gg00 = e_duplicate_filename;
		}
		else if (e_hash_memory_exceeded == TrError)
		{
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj,
				"BD91AddFile:TooManyItems", FileNo);
		}
		else
		{
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj,
				"BD91AddFile: Error      ", TrError);
		}
	}
	
	/* create an empty entry in the continuation file directory for the continuation files */
	if  (bFileIsForVarObj)
	{
		/* create file directory entry for continuation files */
		const tgg00_ObjFileNo   ObjFileNo = FileId.fileObjFileNo_gg00();
		cbd900_ContObjFileInfo  ContObjFileInfo;
		tsp00_Int4              NextContObjBodyLen = MaxObjBodySize;
		
		while (bd93NextContObjBodyLen (NextContObjBodyLen)  && 
			(ContObjFileInfo.cfiNumContFiles_bd900 <= MAX_NUM_CONT_OBJ_FILES_BD900))
		{
            ++ContObjFileInfo.cfiNumContFiles_bd900;
            ContObjFileInfo[ContObjFileInfo.cfiNumContFiles_bd900].ofiMaxObjBodySize_bd900 = NextContObjBodyLen;
		}				
		
		bd93ContObjFileDir.bd920InsItem (Trans.trTaskId_gg00, ObjFileNo, ContObjFileInfo, TrError);
		if (e_ok != TrError) 
		{
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
				"BD91AddFile:InsertFailed", TrError);
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
				"BD91AddFile:InsertFailed", ObjFileNo);
		}
		
		/* check if there really exists an entry in the continuation file directory */
#       if COMPILEMODE_MEO00 >= QUICK_MEO00 
		if (bd93ContObjFileDir[ObjFileNo].cfiNumContFiles_bd900 <1)
			g01abort (csp3_bd_msg, csp3_n_obj, "BD91AddFile: No Entry   ", ObjFileNo);
#       endif
	}
	return pObjFileInfo;	
};

/*---------------------------------------------------------------------------*/

externCpp cbd900_ObjFileInfo*
bd91AddFileToContObjFileDir (tgg00_TransContext   &Trans,
							 tgg00_FileId         &FileId,
    						 cbd910_RootNode      &NptrsRoot)
{

    ROUTINE_DBG_MEO00 ("bd91AddFileToContObjFileDir");

	/* register object file at the global structure bd900ContObjFileDir */
	
	const tgg00_ObjFileNo   ObjFileNo     = FileId.fileObjFileNo_gg00();
	const tsp00_Int4        ContObjFileNo = FileId.fileContObjFileNo_gg00();
	cbd900_ObjFileInfo     &ObjFileInfo   = bd93ContObjFileDir[ObjFileNo][ContObjFileNo];
	
	vbegexcl (Trans.trTaskId_gg00, g08garbage);	
	if (!ObjFileInfo.bd900FileInfoIsEmpty())
		Trans.trError_gg00 = e_duplicate_filename;
	else
	{
		const int ObjFrameLen        = NptrsRoot.np_ptr()->nd_obj_frame_len();
		const int MaxObjBodySize     = ObjFrameLen - sizeof (tbd900_ContObjHeader);
		if (MaxObjBodySize != ObjFileInfo.ofiMaxObjBodySize_bd900)			   
			g01abort (csp3_bd_msg, csp3_n_obj, "BD91AddCFi:WrongBodySize", NptrsRoot.np_ptr()->nd_id());
		const int AlignedObjFrameLen = BD92ALIGN_OBJ_FRAME(ObjFrameLen);
		const int MaxObjPerPageCnt   = 1 + ((FULLCOVERING_BD00 - ObjFrameLen)/AlignedObjFrameLen);
		cbd05_Current Current (Trans, m_select, FileId);
		ObjFileInfo.bd900SetFileInfo (Current, NptrsRoot, MaxObjBodySize, MaxObjPerPageCnt);
	}
	vendexcl (Trans.trTaskId_gg00, g08garbage);	

	return (&ObjFileInfo); 
}

/*---------------------------------------------------------------------------*/

externC void
bd91DumpGarbColl (tgg00_VfFileref     &HostFile,
				  tsp00_Int4            BufferSize,
				  tsp00_Page           &Buffer,
				  tsp00_Int4           &OutPno,
				  tsp00_Int4           &OutPos,
				  tsp00_VfReturn       &HostError,
				  tsp00_ErrText        &ErrText)
{
    ROUTINE_DBG_MEO00 ("bd91DumpGarbColl");

    bd91GarbCollController.bd930DumpGCController (HostFile, BufferSize,  Buffer, OutPno,  OutPos, HostError, ErrText);
}

/*---------------------------------------------------------------------------*/

externC void
bd91InitObjFileDir (tgg00_TransContext &Trans)
{
	ROUTINE_DBG_MEO00 ("bd91InitObjFileDir");

   /* initialization of global structures bd90ObjFileDir */
	const bool           bNotShared          = false;
	const bool           bDontIncludeStrcols = false;
	tgg00_BasisError    &TrError             = Trans.trError_gg00;
	cbd05_Current        Current (Trans, m_select);
	tbd_fileinfo         FileInfo;
    tsp00_Int4           iContFileNo;
	cbd900_ObjFileInfo  *pObjFileInfo;

    /* initialize all entries in bd90ObjFileDir and bd93ContObjFileDir to "zero" */
	bd90ObjFileDir.bd920DelAllItems(Trans.trTaskId_gg00);
	bd93ContObjFileDir.bd920DelAllItems(Trans.trTaskId_gg00);

	/* set start FileId */
	tgg00_FileId   ObjFileId  = b01niltree_id;
	ObjFileId.fileTfn_gg00().becomes(tfnObj_egg00);
	
    /* insert all obj files into bd90ObjFileDir */
	while (TrError == e_ok)
	{
		/* get Fileinfo from BD-file directory */
		b17succ_fdir (ObjFileId.fileName_gg00(), ObjFileId.fileName_gg00(), FileInfo,
			bNotShared, Trans, bDontIncludeStrcols);
		
		if ((TrError == e_ok)                      &&
			(FileInfo.fi_root != NIL_PAGE_NO_GG00))
		{
			if (tfnObj_egg00 != ObjFileId.fileTfn_gg00()) 
				break;

			b30fdir_to_treeid (FileInfo, ObjFileId);
			Current.curr_tree_id = ObjFileId;
			
			cbd910_RootNode  NptrsRoot (Current, ObjFileId.fileRoot_gg00(), nr_for_read);
            if (TrError != e_ok) return;
			
			pObjFileInfo = bd91AddFileToObjFileDir (Trans, ObjFileId, NptrsRoot);
			if (e_ok != TrError) break;
			
			/* add key file root to object file directory  PTS 1106760 Alexander Kley 25/06/2000 */
			if (NptrsRoot.np_ptr()->nd_obj_key_len() > 0)  
			{
				cbd900_ObjKeyFileId  ObjKeyFileId (*pObjFileInfo, ObjFileId.fileObjFileNo_gg00());
				
				b17state_fdir (ObjKeyFileId.fileName_gg00(), FileInfo, bNotShared, Trans);
				if (e_ok != TrError)
				{
					g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
						"BD91InitDir:No Key Entry", NptrsRoot.np_ptr()->nd_id());
					return;
				}
				
				pObjFileInfo->ofiKeyRoot_bd900 = FileInfo.fi_root;
			}

			/* add continuation files to the cont obj file directory */
			/* PTS 1106716 Alexander Kley 01/06/2000                 */
			if (pt2VarObject_egg00 == NptrsRoot.np_ptr()->nd_pt2())
			{
               	const tgg00_ObjFileNo   &ObjFileNo       = ObjFileId.fileObjFileNo_gg00();
				cbd900_ContObjFileInfo  &ContObjFileInfo = bd93ContObjFileDir[ObjFileNo];
				cbd900_RootPageInfo RootPageInfo (NptrsRoot);
				for (iContFileNo=1; iContFileNo <= RootPageInfo.bd900NumContFiles(); ++iContFileNo)
				{
					const tsp00_PageNo ContFileRoot = RootPageInfo.bd900ContRoot (iContFileNo);
					if (NIL_PAGE_NO_GG00 != ContFileRoot)
					{
						ContObjFileInfo.bd900BuildFileId(ObjFileNo, iContFileNo, Current.curr_tree_id);
						Current.curr_tree_id.fileRoot_gg00()  = ContFileRoot;
						cbd910_RootNode  NptrsContRoot (Current, ContFileRoot, nr_for_read);
						if (e_ok != TrError)
						{
							g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
								"BD91InitDir:No Cont Root", NptrsRoot.np_ptr()->nd_id());
							g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
								"BD91InitDir:No Cont Root", ContFileRoot);
							return;
						}
						
						/* PTS 1107306  AK 28/07/2000 */
						bd91AddFileToContObjFileDir (Trans, Current.curr_tree_id, NptrsContRoot);
					}
				}
			}

			/* start GCs for deleted files */ 
           if (pObjFileInfo->ofiRelFileIsReq_bd900)
		   {
			   /* release root page before a GC is started to avoid a dead lock */
			   NptrsRoot.bd910ReleasePage(); 
			   
			   bd91GarbCollController.bd930StartGCToRemoveFile(Trans, *pObjFileInfo, RESTART);
			   
			   if (pObjFileInfo->ofiFileIsForVarObj_bd900)
			   {
				   cbd900_ContObjFileInfo & ContObjFileInfo = bd93ContObjFileDir[ObjFileId.fileObjFileNo_gg00()];				   
				   for (int iContFileNo = 1; iContFileNo <=ContObjFileInfo.cfiNumContFiles_bd900; ++iContFileNo)
				   {
					   tsp00_PageNo ContObjFileRoot = ContObjFileInfo[iContFileNo].ofiRoot_bd900;
					   if (NIL_PAGE_NO_GG00 != ContObjFileRoot)
						   bd91GarbCollController.bd930StartGCToRemoveFile(Trans, ContObjFileInfo[iContFileNo], RESTART);
				   }
			   }
		   }		   
	   }
	}
		
	/* The error code which is send when the end of     */
	/* the file directory is reached is reset to e_ok   */
	if (TrError == e_file_not_found) TrError = e_ok;
}

/*---------------------------------------------------------------------------*/

externC void
bd91InitOMSGarbageCollection ()
{
	ROUTINE_DBG_MEO00 ("bd91InitOMSGarbageCollection");

	bd91GarbCollController.bd930Init (bd90ObjFileDir ,bd93ContObjFileDir);
}

/*---------------------------------------------------------------------------*/

externCpp void
bd91RegisterHistoryFile   (tgg00_TransContext            &Trans,
						   const tbd94_HistoryFileHandle &HistoryFileHandle)
{
	ROUTINE_DBG_MEO00  ("bd91RegisterHistoryFile");

	bd91GarbCollController.bd930AddHistoryFile (Trans, HistoryFileHandle);
}

/*---------------------------------------------------------------------------*/

externC void
bd91StartOMSGarbageCollector (tgg00_TransContext &Trans)
{
    ROUTINE_DBG_MEO00 ("bd91StartOMSGarbageCollector");

	cbd930_GarbageCollector  GarbageCollector(Trans, bd91GarbCollController);
}

/*---------------------------------------------------------------------------*/

externC void
bd91StartOMSGarbageCollection (tgg00_TransContext   &Trans,
							   tsp00_Bool            bStartOnlyIfDataCacheIsFull)
{
	ROUTINE_DBG_MEO00("bd91StartOMSGarbageCollection");

	bd91GarbCollController.bd930StartAllGCs (Trans, (TRUE == bStartOnlyIfDataCacheIsFull));
}

/*---------------------------------------------------------------------------*/

externC void
bd91StartOMSGarbageCollectionAfterSavepoint (tgg00_TransContext   &Trans)
{
	ROUTINE_DBG_MEO00("bd91StartOMSGarbageCollection");

	bd91GarbCollController.bd930StartAllGCsAfterSavepoint (Trans);
}

/*---------------------------------------------------------------------------*/

externC void
bd91StopOMSGarbageCollectionForSavepoint (tgg00_TransContext   &Trans)

{
	ROUTINE_DBG_MEO00("bd91StopOMSGarbageCollectionForSavepoint");

	/* this function stops all garbage collectors, which remain suspended  */
	/* until the method bd930StartAllGCsAfterSavepoint is called           */

	bd91GarbCollController.bd930StopAllGCsForSavepoint (Trans);
}

/*---------------------------------------------------------------------------*/

externC void
bd91ReleaseUnusedObject (tbd_current_tree       &HistoryCurrent,
                         tgg92_KernelOid        &OId,
						 tgg00_ObjFileNo         ObjFileNo,
                         tgg91_PageRef          &WantedObjVers,
						 tbd_node_ptrs          &HistoryNodePtrs)
{
    /* this routine removes a single object and updates the object */
	/* counter and the free page chain, if this is neccessary      */

	tgg00_TransContext &Trans      = *HistoryCurrent.curr_trans;
    tgg00_BasisError   &TrError    = Trans.trError_gg00;
	tgg00_FileId        ObjFileId;

	/* this object writes a message into the vtrace at beginning and end of this routine */
	cbd91_ErrorTraceObject ErrorTraceObject (Trans, ObjFileId, ObjFileNo, OId);

	/* this object restores the HistoryNodePtrs if */
	/* it had to be released to avoid a dead lock  */
	cbd91_HistoryNode  HistoryNode (HistoryNodePtrs, HistoryCurrent);

	/* create FileId for the file which contains the object */
	cbd900_ObjFileInfo  &ObjFileInfo = bd90ObjFileDir[ObjFileNo];
	ObjFileInfo.bd900BuildFileId (ObjFileNo, ObjFileId);

	/* check if the file was already dropped */
	if ((NIL_PAGE_NO_GG00 == ObjFileInfo.ofiRoot_bd900) ||
		(ObjFileInfo.ofiRelFileIsReq_bd900))
	{
		/* since the file is dropped already one can return immediately */
		ErrorTraceObject.bd91SetStatus (rosFileAlreadyDeleted_gg12);
		return;
	}

    /* create current to access page */
	cbd05_Current Current (Trans, m_update, ObjFileId);

    /*  get page pointer  */
    cbd910_PrimObjDataNode    NptrsPage (Current, nr_for_update);
	NptrsPage.bd910AssignToPage(OId.gg92GetPno(), plmLockNoWait_ebd00);

	/* release history page to avoid a deadlock; PTS 1106180 AK 29/03/2000 */
    if (e_bd_page_collision == TrError)
	{
		TrError       = e_ok;
		HistoryNode.bd91Release();
		NptrsPage.bd910AssignToPage(OId.gg92GetPno());
	}

    if (e_ok != TrError)
	{
		/* check if the error is due to a drop of the file */
		if ((e_page_in_wrong_tree == TrError) || (e_no_converter_entry == TrError))
		{
			if ((NIL_PAGE_NO_GG00 == bd90ObjFileDir[ObjFileNo].ofiRoot_bd900) ||
				(bd90ObjFileDir[ObjFileNo].ofiRelFileIsReq_bd900))
			{
				ErrorTraceObject.bd91SetStatus (rosFileAlreadyDeleted_gg12);
				TrError = e_ok;
			}
			else if (NIL_PAGE_NO_GG00 != bd90ObjFileDir[ObjFileNo].ofiKeyRoot_bd900)
			{
				ErrorTraceObject.bd91SetStatus (rosPageNotFound_gg12);
				TrError = e_ok;
			}
			else
				ErrorTraceObject.bd91SetStatus (rosUnknownError_gg12);
		}
		else
			ErrorTraceObject.bd91SetStatus (rosUnknownError_gg12);
		
		return;
	}

    /*------------------------------------------------------------------*/
	/* check if the found object frame contains realy the wanted object */
    /*------------------------------------------------------------------*/
	
	/* determine pointer to the page accomodating the object frame */
	tbd_nodeptr         &pPage     = NptrsPage.np_ptr();
	
	const tsp00_Int4     ObjKeyLen = pPage->nd_obj_key_len();

	/*  check consistency of key infos stored on the page and stored in the obj file directory */
#   if COMPILEMODE_MEO00 >= QUICK_MEO00
	if ((ObjKeyLen > 0) != (NIL_PAGE_NO_GG00 != bd90ObjFileDir[ObjFileNo].ofiKeyRoot_bd900))
	{
		ErrorTraceObject.bd91SetStatus (rosWrongKeyInfo_ggg12);
			TrError = e_hash_not_found;
		return;
	}
#   endif

	/* check the obj file no on page */
	if (ObjFileNo != pPage->nd_obj_file_no())
	{
		/* check if the error is due to a drop of the file */
		if ((NIL_PAGE_NO_GG00 == bd90ObjFileDir[ObjFileNo].ofiRoot_bd900) ||
			(bd90ObjFileDir[ObjFileNo].ofiRelFileIsReq_bd900))
		{
			ErrorTraceObject.bd91SetStatus (rosFileAlreadyDeleted_gg12);
			TrError = e_ok;
		}
        else if (ObjKeyLen > 0)
		{
			ErrorTraceObject.bd91SetStatus (rosWrongFileNo_gg12);
			TrError = e_ok;
		}
		else
		{
			TrError = e_wrong_class_id;
			ErrorTraceObject.bd91WriteVersionErrorMessage (pPage->nd_obj_file_no(), WantedObjVers, NULL);
		}
		return;
	}

	/* determine pointer to object frame */
	BD92CHECK_OID_POSITION (TrError, OId, NptrsPage);
    tbd900_ObjFrame  *pObjFrame = NptrsPage.bd910ObjFramePtr (OId.gg92GetPos());

    /*--------------------------------------------------------------------------------------------------*/
    /* It can happen that the GC cannot find an object it was requested to delete. See for              */
	/* instance the history sequence : => new - upd - del - new - upd - del. Releasing an object frame  */
	/* due to the first delete must result in the error 'wrong object version', which has to be ignored */
	/* However, this can only happen in case that the object has a key and its frame was reused         */
    /*--------------------------------------------------------------------------------------------------*/

    /* check object version */
    if (WantedObjVers != pObjFrame->objHeader_bd900.ohdTransInfo_gg92.otrBeforeRef_gg00)
    {
		/* check if this could be regular error which can be ignored */
		if (ObjKeyLen > 0)
		{
	   		ErrorTraceObject.bd91SetStatus (rosWrongObjVersion_gg12);
			TrError = e_ok;
		}
		else
		{
			TrError = e_wrong_object_version;
			ErrorTraceObject.bd91WriteVersionErrorMessage (pPage->nd_obj_file_no(), WantedObjVers, pObjFrame);
		}
        return;
    }

#   if COMPILEMODE_MEO00 >= QUICK_MEO00
    /*  check object state  */
    if (obsFreeAfterEot_egg00  != pObjFrame->objHeader_bd900.ohdState_gg92)
		// note: the Frameversion is not decreased in case of a rollback
		//       therefore the next equation need not be valid
		// (OId.gg92GetFrameVers() != pObjFrame->objHeader_bd900.ohdFrameVersion_gg92)
    {
        TrError = e_wrong_object_version;
		ErrorTraceObject.bd91WriteVersionErrorMessage (pPage->nd_obj_file_no(), WantedObjVers, pObjFrame);
        return;
    }
#   endif

    /* =========================================================== */
	/* ----- after all checks the object can be released now ----- */
    /* =========================================================== */

    /*-----------------------*/
	/* remove primary object */
    /*-----------------------*/

	/*  update object counter in page header  */
	++pPage->nd_free_obj_cnt();
	--pPage->nd_free_eot_obj_cnt();

    /*  update object state  */
    pObjFrame->objHeader_bd900.ohdState_gg92.becomes(obsFree_egg00);

    /*  insert object into free object chain */
    pObjFrame->objHeader_bd900.ohdNextFreeObj_gg92 = pPage->nd_first_free_obj();
    pPage->nd_first_free_obj()        = OId.gg92GetPos();

    /*-----------------------*/
	/* remove OMS key        */
    /*-----------------------*/
	if (ObjKeyLen > 0)
	{
		/* construct pointer to the key */
		void *const pKey = (tsp00_Byte*) pPage + OId.gg92GetPos() +	pPage->nd_obj_frame_len() - ObjKeyLen;

		/* remove OId from hash */
		bd22RemoveOid (*Current.curr_trans, Current.curr_tree_id.fileObjFileNo_gg00(),ObjKeyLen, pKey, OId);
		if (e_hash_key_not_found == TrError)  TrError = e_ok;
		if (e_ok != TrError) 
		{
	   		ErrorTraceObject.bd91SetStatus (rosDelKeyFailed_gg12);   /* PTS 1106760  Alexander Kley 25/06/2000 */
			return;
		}
    }

	/*--------------------------------*/
    /* remove the continuation object */
    /*--------------------------------*/
	if (pt2VarObject_egg00 == pPage->nd_pt2())
	{
		tbd900_VarObjBody &VarObjBody  = *((tbd900_VarObjBody*) (pObjFrame->objBody_bd900));
		tgg00_ObjFileNo  ContObjFileNo = VarObjBody.vbdHeader_bd900.vhdContObjFileNo_bd900;
		
		/* check if there is really a continuation object */
		if (PRIM_CONT_OBJ_FILE_NO_BD900 != ContObjFileNo)
		{
			tsp00_PageNo     ContObjPno = VarObjBody.vbdHeader_bd900.vhdContObjPno_bd900;
			tgg00_ObjPagePos ContObjPos = VarObjBody.vbdHeader_bd900.vhdContObjPos_bd900;
			
			/* remove continuation object */
			bd93RemoveContObj (Current, OId, ContObjFileNo, ContObjPno, ContObjPos);
			if (e_ok != TrError) 
			{
				/* check if this error occured since the container was dropped */
				/* PTS 1107433 Alexander Kley 07/08/2000                       */
				cbd900_ObjFileInfo &ObjFileInfo = bd90ObjFileDir[ObjFileNo];
				if ((ObjFileInfo.ofiRoot_bd900 == NIL_PAGE_NO_GG00) || ObjFileInfo.ofiRelFileIsReq_bd900)
				{
					ErrorTraceObject.bd91SetStatus (rosFileJustDeleted_gg12);
					TrError = e_ok;
				}
				else
				{
					ErrorTraceObject.bd91SetStatus (rosDelContObjFailed_gg12);   /* PTS 1106760  Alexander Kley 25/06/2000 */
					return;
				}
			}
		}
		
		/* reset var object header */
		VarObjBody.vbdHeader_bd900.bd900SetToEmptyHeader();
	}

    /*---------------------------------------------*/
	/* update counter of all objects in chaininfo  */
	/*---------------------------------------------*/

	-- ObjFileInfo[NptrsPage.np_ptr()->nd_chain_no()].ociAllObjCnt_bd900;

	/*-------------------------------------------*/
	/* update free page chain and page counters  */
	/*-------------------------------------------*/

	if (NptrsPage.bd910PageMustBeInsertedIntoFreeChain() ||
		NptrsPage.bd910EmptyPageCountMustBeIncreased())
	{
		/* release history page to avoid a deadlock; PTS 1106180 AK 29/03/2000 */
		HistoryNode.bd91Release();
		
		NptrsPage.bd910InsertPageIntoFreeChainAndUpdateEmptyPageCount (ObjFileInfo);
		
		if (e_ok != TrError)
		{
			/* check if this error occured since the container was dropped */
            /* PTS 1106724 Alexander Kley 23/05/2000                       */
			cbd900_ObjFileInfo &ObjFileInfo = bd90ObjFileDir[ObjFileNo];
			if ((ObjFileInfo.ofiRoot_bd900 == NIL_PAGE_NO_GG00) || ObjFileInfo.ofiRelFileIsReq_bd900)
			{
				ErrorTraceObject.bd91SetStatus (rosFileJustDeleted_gg12);
				TrError = e_ok;
			}
			else
			{
				ErrorTraceObject.bd91SetStatus (rosUnknownError_gg12);
				g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
					"bd90RelObj:InsPageFailed",	ObjFileNo);
			}
			return; 
		}
	}
}

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/
