/****************************************************************************

  module      : vbd92.cpp

  -------------------------------------------------------------------------

  responsible : AlexanderK

  special area: BD_AuxilleryFunctionsForObjectHandling
  description : description ...


  version     : 7.2.0.0
  last changed: 2000-08-11  8:31
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1998



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/
 
#include "hbd92.h"

#include "gsp03.h"
#include "gsp03_3.h"
#include "ggg12.h"
#include "gbd05.h"
#include "gbd94.h"
#include "gbd900.h"
#include "gbd910.h"
#include "gbd930.h"

#include "hbd10.h"
#include "hbd12.h"
#include "hbd15.h"
#include "hbd22.h"
#include "hbd90.h"
#include "hgg01.h"
#include "hgg01_1.h"
#include "hkb50.h"

#if COMPILEMODE_MEO00 >= SLOW_MEO00 
#include "hta99.h"
#include "hta01.h"
#endif

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#ifdef NO_INLINES
#  define _INLINE
#else
#  define _INLINE   inline
#endif

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/

struct  tbd92_ObjPageChainStat
{
	int    FreePageCnt;           
    int    PageWithFreeLabelCnt;          
    int    FreePageInFreeChainCnt;           
    int    PageInFreeChainCnt;                    
    int    PageWithFreeLabelInFreeChainCnt;       
    int    FreePageWithFreeLabelInFreeChainCnt;    
    
    int    FeotPageCnt;
    
    int    EmptyPageCnt;
    int    PageWithEmptyLabelCnt;
    int    PageWithEmptyLabelInFreeChainCnt;
    int    EmptyPageInFreeChainCnt;
    int    EmptyPageWithEmptyLabelInFreeChainCnt;
    
    int    AllPageCnt; 

	int    FreeObjCnt;
	int    FeotObjCnt;
	int    OccObjCnt;
	int    ResObjCnt;

    inline void ResetAllCountersToZero ()		
	{
		FreePageCnt                          = 0;
		PageWithFreeLabelCnt                 = 0;
		FreePageInFreeChainCnt               = 0;
		PageInFreeChainCnt                   = 0;
		PageWithFreeLabelInFreeChainCnt      = 0;
		FreePageWithFreeLabelInFreeChainCnt  = 0;      
		FeotPageCnt                          = 0;
		EmptyPageCnt                         = 0;
		PageWithEmptyLabelCnt                = 0;
		PageWithEmptyLabelInFreeChainCnt     = 0;
		EmptyPageInFreeChainCnt              = 0;
		EmptyPageWithEmptyLabelInFreeChainCnt= 0;
		AllPageCnt                           = 0; 
		FreeObjCnt                           = 0;
		FeotObjCnt                           = 0;
		OccObjCnt                            = 0;
		ResObjCnt                            = 0;
	};
};



/*---------------------------------------------------------------------------*/

struct tbd92_ObjPageContainerStat
{
	int FreeObjCnt;
	int FeotObjCnt;
	int OccObjCnt;	
	int ResObjCnt;
    int AllPageCnt;
    int FreePageCnt;
	int FeotPageCnt;
	int EmptyPageCnt;
	
    tbd92_ObjPageContainerStat ()		
		:FreeObjCnt(0),
		FeotObjCnt(0),
		OccObjCnt(0),
		ResObjCnt (0),
		AllPageCnt(0),
		FreePageCnt(0),
		FeotPageCnt(0),
		EmptyPageCnt(0)
	{};
};

/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/

extern cbd900_ObjFileDir          bd90ObjFileDir;
extern cbd900_ContObjFileDir      bd93ContObjFileDir;
extern cbd930_GarbCollController  bd91GarbCollController;

/*===========================================================================*
 *  GLOBAL VARIABLES                                                         *
 *===========================================================================*/

/*===========================================================================*
 *  LOCAL VARIABLES                                                          *
 *===========================================================================*/

/*===========================================================================*
 *  LOCAL FUNCTIONS (PROTOTYPES)                                             *
 *===========================================================================*/


static _INLINE void
bd92_CheckBeforeImage (tgg00_TransContext   &Trans,      
                       tgg92_KernelOid       WantedOid,
					   tgg00_ObjTransInfo   &ObjTransInfo);  

/*---------------------------------------------------------------------------*/

static _INLINE void
bd92_CheckConsistencyOfChain (const  tbd92_ObjPageChainStat  &ChainStat,
							  cbd900_SubRootPageInfo         &SubRootPageInfo);

/*---------------------------------------------------------------------------*/

static _INLINE void
bd92_CheckHashEntry (tbd_current_tree       &Current, 
					 tbd900_ObjFrame         *pObjFrame, 
					 tgg92_KernelOid        &CurrentOId,
					 int                    ObjFrameLen,
                     int                    ObjKeyLen); 

/*---------------------------------------------------------------------------*/

static _INLINE  void
bd92_CheckAndUpdObjChainStatistics (tgg00_ObjFileNo               ObjFileNo,
									int                           ChainNo,
									const tbd92_ObjPageChainStat &ChainStat,
									cbd900_ObjFileInfo           *pObjFileInfo);

/*---------------------------------------------------------------------------*/

static _INLINE  void
bd92_VerifyObjPageChain (tbd_current_tree        &Current,
						boolean                  bConverterUpdate,
						tsp00_PageNo             SubRoot,
						tbd92_ObjPageChainStat  &ChainStat,
						const tsp00_Bool         bChainIsDeleted);

/*---------------------------------------------------------------------------*/

static _INLINE void
bd92_VerifyObjRootPage (tbd_node_ptrs  &NptrsRoot);

/*---------------------------------------------------------------------------*/

static _INLINE void
bd92_VerifyObjSubRootPage (tbd_node_ptrs  &NptrsSubRoot);

/*---------------------------------------------------------------------------*/

static _INLINE void
bd92_VerifyContObjDataPage (tbd_current_tree   &Current,
							tbd_nodeptr         pPage);

/*---------------------------------------------------------------------------*/

static _INLINE void
bd92_VerifyObjDataPage (tbd_current_tree  &Current,
						tbd_nodeptr        pPage,
						tsp00_Int4        &ResObjCnt);

/*---------------------------------------------------------------------------*/

static _INLINE void
bd92_VerifyObjPage (tbd_current_tree  &Current,
   				    tbd_node_ptrs     &NptrsPage,
				    tsp00_Int4        &ResObjCnt);

/*---------------------------------------------------------------------------*/

static _INLINE void
bd92_VerifyObjPageHeader (tbd_current_tree  &Current,
						  tbd_nodeptr        pPage);

/*---------------------------------------------------------------------------*/

static _INLINE void
bd92_WriteContainerStateToVtrace (tbd_current_tree                     &Current, 
								  const  tbd92_ObjPageContainerStat    &ContainerStat,
								  const  cbd900_ObjFileInfo            *pObjFileInfo);

/*---------------------------------------------------------------------------*/

static _INLINE void
bd92_WriteChainStateToVtrace (tbd_current_tree                &Current, 
							  const  tbd92_ObjPageChainStat   &ChainStat,
							  cbd910_SubRootNode              &NptrsSubRoot,
							  const tsp00_Bool                 bChainIsDeleted);

/*---------------------------------------------------------------------------*/

static _INLINE void
bd92_WritePageStateToVtrace (tgg00_TransContext    &Trans,
							 tbd_nodeptr            pPage,
							 tsp00_Int4             ResObjCnt);

/*===========================================================================*
 *  GLOBAL FUNCTIONS (CODE)                                                  *
 *===========================================================================*/


externCpp bool 
bd92ObjPosIsValid (tgg00_ObjPagePos  ObjPos,
                   tgg00_ObjPagePos  FirstObjPos,
                   tsp00_Uint2       ObjFrameLen)
{
    ROUTINE_DBG_MEO00 ("bd92ObjPosIsValid");

    /* This function returns true if the ObjPos points    */
    /* onto a valid position of an object otherwise false */
    /* is returned.                                       */
    
    bool      bValidPos;
    const int AlignedObjFrameLen = BD92ALIGN_OBJ_FRAME (ObjFrameLen);

    bValidPos = (
		(FirstObjPos >= NODE_HEADER_MXBD00                     ) &&  /* PTS 1107145 AK 10/07/2000 */
        (ObjPos >= FirstObjPos                                 ) &&
        ((ObjPos + ObjFrameLen) <= BODY_END_BD00               ) &&
        (((ObjPos - FirstObjPos) % AlignedObjFrameLen)     == 0) &&
        ((FirstObjPos - BD92ALIGN_OBJ_FRAME (FirstObjPos)) == 0) 
        );
	
	if (! bValidPos)
		bd92WriteWrongOIdToVtrace (ObjPos, FirstObjPos,  ObjFrameLen);
    
    return (bValidPos);
}

/*---------------------------------------------------------------------------*/

externC void
bd92VerifyObjFile (tbd_current_tree       &Current,
				   boolean                 bConverterUpdate)
{
    ROUTINE_DBG_MEO00 ("bd92VerifyObjFile");
	
	/* This function checks all object page chains which */
	/* belong to the current object file.      if any    */
	/* error is detected an op message is sent           */
	/* Moreover, if bConverterUpdate = true all pages    */
	/* of the container are flaged in the FBM as used    */
	
	tbd92_ObjPageContainerStat ContainerStat;
	tbd92_ObjPageChainStat     ChainStat;
	tsp00_PageNo               SubRoot;
	int                        iChain;
	const tgg00_ObjFileNo     &FileNo       = Current.curr_tree_id.fileObjFileNo_gg00();
	const int                  ContFileNo   = Current.curr_tree_id.fileContObjFileNo_gg00();
    cbd900_ObjFileInfo        *pObjFileInfo = NULL;

    b15file_opmsg (csp3_check_file, Current);

    /* check if the file to be checked is really an object file */
	if (Current.curr_tree_id.fileTfn_gg00() != tfnObj_egg00)
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
		"BD92_VOF: No Obj File   ", Current.curr_tree_id.fileTfn_gg00());

    /* get shortcut to the object file info */
	if (PRIM_CONT_OBJ_FILE_NO_BD900 == ContFileNo)
		pObjFileInfo = &(bd90ObjFileDir[FileNo]);
	else
		pObjFileInfo = &(bd93ContObjFileDir[FileNo][ContFileNo]);
	
	/* check if the  root is correctly stored in the ObjFileInfo */
    if (pObjFileInfo->ofiRoot_bd900 != Current.curr_tree_id.fileRoot_gg00())
	{
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
			"BD92_VOF: Wrong FileId R1", Current.curr_tree_id.fileRoot_gg00());
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
			"BD92_VOF: Wrong FileId R2", pObjFileInfo->ofiRoot_bd900);
	}

	/* flag FBM */
    if (bConverterUpdate)
        b10use_pno (*Current.curr_trans, Current.curr_tree_id.fileRoot_gg00());
	
	/* check each container chain of the file */
    const int       &MaxChain       = pObjFileInfo->ofiMaxChain_bd900; 
    const tsp00_Bool bFileIsDeleted = pObjFileInfo->ofiRelFileIsReq_bd900;

	for (iChain=0; (iChain<MaxChain) && (e_ok == Current.curr_trans->trError_gg00); ++iChain)
	{	
		SubRoot = pObjFileInfo->ofiChain_bd900[iChain].ociSubRoot_bd900;
		
		if (NIL_PAGE_NO_GG00 == SubRoot)
		{
			if (!bFileIsDeleted)
			{
				g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
					"BD92VOF:Missing Subroot ", FileNo);
				g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
					"BD92VOF:Missing Subroot ", ContFileNo);
			}	
		}
		else 
		{	
	    	bd92_VerifyObjPageChain (Current, bConverterUpdate, SubRoot, ChainStat, bFileIsDeleted);
			
			/* count objects */
			ContainerStat.FreeObjCnt    += ChainStat.FreeObjCnt;
			ContainerStat.FeotObjCnt    += ChainStat.FeotObjCnt;
			ContainerStat.OccObjCnt     += ChainStat.OccObjCnt;
			ContainerStat.ResObjCnt     += ChainStat.ResObjCnt;
			
			/* count pages */
			ContainerStat.AllPageCnt    += ChainStat.AllPageCnt;
			ContainerStat.FreePageCnt   += ChainStat.FreePageCnt;
			ContainerStat.FeotPageCnt   += ChainStat.FeotPageCnt;
			ContainerStat.EmptyPageCnt  += ChainStat.EmptyPageCnt;
			
			/* check the number of pages and objects stored in ObjFileInfo */
			bd92_CheckAndUpdObjChainStatistics (FileNo, iChain, ChainStat, pObjFileInfo);
		}
	}
	
	/* write a vtrace entry */
	bd92_WriteContainerStateToVtrace (Current, ContainerStat, pObjFileInfo);
	
	/* verify continuation files PTS 1106716 Alexander Kley 01/06/2000 */
	if ((oftVarLenObjFile_egg00 == Current.curr_tree_id.fileObjFileType_gg00()) &&
		(PRIM_CONT_OBJ_FILE_NO_BD900 == ContFileNo))                              // PTS 1107352 AK 01/08/2000
	{
		cbd900_ContObjFileInfo &ContObjFileInfo = bd93ContObjFileDir[FileNo];
        cbd05_Current ContCurrent (*Current.curr_trans, m_select);
		for (int iContFileNo = 1; iContFileNo <= ContObjFileInfo.cfiNumContFiles_bd900; ++iContFileNo)
		{
			if (NIL_PAGE_NO_GG00 != ContObjFileInfo[iContFileNo].ofiRoot_bd900)  // PTS 1107306 AK 01/08/2000
			{
				ContObjFileInfo.bd900BuildFileId (FileNo,iContFileNo,ContCurrent.curr_tree_id);
				bd92VerifyObjFile (ContCurrent,bConverterUpdate);
			}
		}
	}
}

/*---------------------------------------------------------------------------*/

externCpp void
bd92VerifyObjPage (tbd_current_tree  &Current,
				   tbd_node_ptrs     &NptrsPage)
{
    ROUTINE_DBG_MEO00 ("bd92VerifyObjPage");
	
	/* This routine checks the consistency */
	/* of the contents of an object page   */

	/* declare dummy return */

	tsp00_Int4 DummyResObjCnt = 0;

	bd92_VerifyObjPage (Current, NptrsPage, DummyResObjCnt);
}

/*---------------------------------------------------------------------------*/

externCpp void
bd92VtraceObjKey (
				  tgg00_TransContext    &Trans, 
				  tgg00_ObjFileNo        ObjFileNo,
				  tsp00_Int4             KeyLen,
				  void                  *pKey,
				  tgg00_VtraceType_Enum  TraceObject
				  )
{
    /*  This function writes the fileno and the key of a   */
    /*  BD-function at the KB-BD interface into the trace  */
    
    const int         MaxVtraceKey = sizeof (tsp00_Line) - sizeof (tgg12_OmsTraceKeyHeader);   
    tgg12_OmsKeyTrace TrLineOut;
    int               OutKeyLen;
    
    TrLineOut.okyHead_gg12.okhTrType_gg12.becomes  ( otKey_egg12);
    TrLineOut.okyHead_gg12.okhTrType2_gg12.becomes ( ot2Nil_egg12);
    TrLineOut.okyHead_gg12.okhObjFileNo_gg12       = ObjFileNo;
    TrLineOut.okyHead_gg12.okhConsView_gg12        = Trans.trConsistView_gg00;
    
    TrLineOut.okyHead_gg12.okhKeyLen_gg12 = 
        ((KeyLen >= 0) && (KeyLen <= MAX_UINT1_SP00)) ? KeyLen : MAX_UINT1_SP00;
    
    OutKeyLen = (TrLineOut.okyHead_gg12.okhKeyLen_gg12 > MaxVtraceKey)
        ? MaxVtraceKey : TrLineOut.okyHead_gg12.okhKeyLen_gg12;
    
    if (OutKeyLen > 0)
    {
        memcpy (&TrLineOut.okyKey_gg12, pKey, OutKeyLen);
    }
    b12vtrace (Trans, bd, TraceObject, (tsp00_Int2)(sizeof (TrLineOut.okyHead_gg12) + OutKeyLen), &TrLineOut);
}

/*---------------------------------------------------------------------------*/

externCpp void
bd92WriteWrongOIdToVtrace (tgg00_ObjPagePos  ObjPos,
                           tgg00_ObjPagePos  FirstObjPos,
                           tsp00_Uint2       ObjFrameLen)
{
	g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
		"BD92IsPosValid: Obj Pos ", ObjPos);
	
	g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
		"BD92IsPosValid:First Pos", FirstObjPos);
	
	g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
		"BD92IsPosValid: FrameLen", ObjFrameLen);
	
	g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
		"BD92IsPosValid:AFrameLen", BD92ALIGN_OBJ_FRAME (ObjFrameLen));
}

/*===========================================================================*
 *  LOCAL FUNCTIONS (CODE)                                                   *
 *===========================================================================*/

static _INLINE void
bd92_CheckBeforeImage (tgg00_TransContext   &Trans,      
                       tgg92_KernelOid       WantedOid,
					   tgg00_ObjTransInfo   &ObjTransInfo)   					   
{
	/* check if there exists an before image. for */
	/* a feot object there must always exist one  */
	tsp00_Bool bIsActive;
	
	kb50CheckActiveTrans (Trans.trTaskId_gg00, 
		ObjTransInfo.otrLockTransId_gg00,
		ObjTransInfo.otrLockTransIndex_gg00,
		bIsActive);

	if (bIsActive != TRUE)
	{
		if (!bd94History.bd94BeforeImageExists (Trans, WantedOid, ObjTransInfo.otrBeforeRef_gg00))
		{
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
				"BD92_CBI:odd BI OId.Pno=", WantedOid.gg92GetPno());
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
				"BD92_CBI:odd BI OId.Pos ", WantedOid.gg92GetPos());
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
				"BD92_CBI:odd BI OId.Ver ", WantedOid.gg92GetFrameVers());
#          if COMPILEMODE_MEO00 >= QUICK_MEO00
			g01abort (csp3_b90_damaged_obj_chain, csp3_n_obj, 
				"BD92_CBI:odd BeforeImage", Trans.trError_gg00);
#          else			
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
				"BD92_CBI:odd BeforeImage", Trans.trError_gg00);
#          endif
		}
	}
}

/*---------------------------------------------------------------------------*/

static _INLINE void
bd92_CheckConsistencyOfChain (const  tbd92_ObjPageChainStat  &ChainStat,
							  cbd900_SubRootPageInfo         &SubRootPageInfo)
{
	ROUTINE_DBG_MEO00 ("bd92_CheckConsistencyOfChain");
	
    /* this routine checks if the data collected in the ChainStat and   */
	/* on the subroot page are consistent, if not an op-message is sent */
	
    if (ChainStat.AllPageCnt != SubRootPageInfo.bd900AllPageCnt())
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
		"BD92_CC:oddAllPageChain ", ChainStat.AllPageCnt);
	
	if ((ChainStat.FreePageWithFreeLabelInFreeChainCnt >  ChainStat.FreePageInFreeChainCnt)          ||
		(ChainStat.FreePageInFreeChainCnt              >  ChainStat.FreePageCnt)                     ||
		(ChainStat.FreePageCnt                         >  ChainStat.PageWithFreeLabelInFreeChainCnt) ||
        (ChainStat.PageWithFreeLabelInFreeChainCnt     != SubRootPageInfo.bd900FreePageCnt())        ||
		(SubRootPageInfo.bd900FreePageCnt()            != ChainStat.PageWithFreeLabelCnt)            ||
		(ChainStat.PageWithFreeLabelCnt                != ChainStat.PageInFreeChainCnt ))
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
		"BD92_CC:oddFreePageChain", ChainStat.FreePageCnt);
	
	if ((ChainStat.EmptyPageInFreeChainCnt            >  SubRootPageInfo.bd900EmptyPageCnt()) ||
		(SubRootPageInfo.bd900EmptyPageCnt()          != ChainStat.PageWithEmptyLabelCnt )    ||
		(ChainStat.PageWithEmptyLabelCnt              >  ChainStat.EmptyPageCnt ))
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
		"BD92_CC:oddEmptPageChain", ChainStat.EmptyPageCnt);    
}

/*---------------------------------------------------------------------------*/


static _INLINE void
bd92_CheckHashEntry (
					 tbd_current_tree       &Current, 
					 tbd900_ObjFrame         *pObjFrame, 
					 tgg92_KernelOid        &CurrentOId,
					 int                    ObjFrameLen,
					 int                    ObjKeyLen
					 ) 
					 
{
	ROUTINE_DBG_MEO00 ("bd92_CheckHashEntry");
	
	/* This function checks if the OId found from the hash        */
	/* by using the key of the current object points at           */ 
	/* the current object                                         */
	/* Note, for an object of the feot state one cannot check     */
	/* the hash since there can exist an object with the same key */
	/* which was reused. In this case the current object was      */
    /* deleted and is of the state free after eot now.            */
	/* The Version ist not checked because it is not any more     */
	/* correct after an object is reused                          */
	
	tgg92_KernelOid HashOId;
	
	void * pObjKey = (void *) ((tsp00_Byte *) pObjFrame  + ObjFrameLen - ObjKeyLen);
	
	bd22GetOId (*Current.curr_trans, Current.curr_tree_id.fileObjFileNo_gg00(), ObjKeyLen, pObjKey, HashOId);
	
	if (Current.curr_trans->trError_gg00	!= e_ok) return;
	
    if (! HashOId.gg92IsEqIgnoreFrameVers(CurrentOId) &&
       (pObjFrame->objHeader_bd900.ohdState_gg92 == obsOccupied_egg00)) 
    {
        /*-----------------------------------*/
		/* an error is detected in the hash  */
		/*-----------------------------------*/
		
		/* write a message into knldiag */
		
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_page, csp3_n_obj,   
			"BD92_ChkHash:wrong Hash ", CurrentOId.gg92GetPno());
		
		/* output of input OId and input key into vtrace */

		tgg12_OmsOidTrace TrLineOId;
		
		TrLineOId.ooidTrType_gg12.becomes(otOid_egg12);
		TrLineOId.ooidTrType2_gg12.becomes(ot2Input_egg12);
		TrLineOId.ooidObjId_gg12   = CurrentOId;
		b12vtrace (*Current.curr_trans, bd, b92check_hash, sizeof (TrLineOId), &TrLineOId);
		
		bd92VtraceObjKey (*Current.curr_trans,
			Current.curr_tree_id.fileObjFileNo_gg00(), ObjKeyLen, pObjKey, b92check_hash);

        /* output of output OId or error reason into vtrace */

		if (Current.curr_trans->trError_gg00 == e_ok) 
		{
			TrLineOId.ooidObjId_gg12   = HashOId;
			TrLineOId.ooidTrType2_gg12.becomes(ot2Output_egg12);
			b12vtrace (*Current.curr_trans, bd, b92check_hash, sizeof (TrLineOId), &TrLineOId);
		}
		else
		{
			tgg12_OmsErrorTrace TrLineOut;
			
			TrLineOut.oerrTrType_gg12.becomes  ( otError_egg12);
			TrLineOut.oerrTrType2_gg12.becomes ( ot2Nil_egg12);
			TrLineOut.oerrRoot_gg12            = Current.curr_tree_id.fileRoot_gg00();
			TrLineOut.oerrObjFileNo_gg12       = Current.curr_tree_id.fileObjFileNo_gg00();
			TrLineOut.oerrError_gg12           = Current.curr_trans->trError_gg00;
			
			b12vtrace (*Current.curr_trans, bd, b92check_hash, sizeof (TrLineOut), &TrLineOut);
		}
	}
}



/*---------------------------------------------------------------------------*/

static _INLINE  void
bd92_CheckAndUpdObjChainStatistics (
									tgg00_ObjFileNo               ObjFileNo,
									int                           ChainNo,
									const tbd92_ObjPageChainStat &ChainStat,
									cbd900_ObjFileInfo           *pObjFileInfo
									)
{
	ROUTINE_DBG_MEO00 ("bd92_CheckAndUpdObjChainStatistics");
			
	/* Since the ociAllObjCnt_bd900 is updated dirty it may differ from the  */ 
	/* counted number of all occupied object frames. Here ociAllObjCnt_bd900 */
	/* is given again a reasonable value. PTS 1103239 AK 08-09-1999          */
	const tsp00_Int4 AllObjCnt = ChainStat.FeotObjCnt + ChainStat.OccObjCnt + ChainStat.ResObjCnt;
	pObjFileInfo->ofiChain_bd900[ChainNo].ociAllObjCnt_bd900 = AllObjCnt;
	
	/* The number of pages in ObjFileInfo is checked  */
	if (ChainStat.AllPageCnt != pObjFileInfo->ofiChain_bd900[ChainNo].ociAllPageCnt_bd900)
	{
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
			"BD92_COFI: WPC FileNo = ", ObjFileNo);
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
			"BD92_COFI: WPC ChainNo= ", ChainNo);
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
			"BD92_COFI:wrong Page Cnt", ChainStat.AllPageCnt);
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
			"BD92_COFI:wrong Page Cnt", pObjFileInfo->ofiChain_bd900[ChainNo].ociAllPageCnt_bd900);
	};
}

/*---------------------------------------------------------------------------*/

static _INLINE  void
bd92_VerifyObjPageChain (tbd_current_tree        &Current,
						 boolean                  bConverterUpdate,
						 tsp00_PageNo             SubRoot,
						 tbd92_ObjPageChainStat  &ChainStat,
						 const tsp00_Bool         bChainIsDeleted)
{
    ROUTINE_DBG_MEO00 ("bd92_VerifyObjPageChain");
    
    /* This function checks the counter of  empty, free,          */
    /* free_after_eot and all pages contained in the info         */
    /* block on the root page. The counters of occupied, free and */ 
    /* free_after_eot objects are verified for each page          */
    /* in the chain, too. In case the chain does not belong to a  */
	/* file which is marked as to be deleted it is also checked   */
	/* if the chains of free and free_after eot pages and objects */
    /* are consistent                                             */
    
    tgg00_BasisError        &TrError      = Current.curr_trans->trError_gg00;
	tsp00_Int4              PrevPageSeqNo = MAX_INT4_SP00;
    tsp00_Int4              ResObjCnt;
	tsp00_PageNo            NextPageNo;
	
	ChainStat.ResetAllCountersToZero();
    
    /*------------------------------------------------------------*/
    /* go through chains of all pages following the right pointer */
    /* and count the number of free, empty and free_after_eot     */
    /* pages. For each page the object counters are verified      */
    /* as well as the free object chain on each page              */
    /*------------------------------------------------------------*/
   
	/* keep subroot to avoid any chanhe of the chain structure */
	cbd910_SubRootNode   NptrsSubRoot (Current, SubRoot, nr_for_read);
    if (TrError != e_ok) return;
    
	cbd900_SubRootPageInfo SubRootPageInfo (NptrsSubRoot);
	const tsp00_Int4  ChainNo  = NptrsSubRoot.np_ptr()->nd_chain_no();  
    
	/* create page handler which runs over all pages of  */
	/* this chain and set it to the subroot (start) page */
	cbd910_ObjDataNode    NptrsPage (Current, nr_for_read);
    NptrsPage.bd910SetTo (NptrsSubRoot);

    /* traverse chain of all pages  */
	do
	{
		if (bConverterUpdate)
			b10use_pno (*Current.curr_trans,  NptrsPage.np_ptr()->nd_id());
		
		/* check that the page sequence numbers decreases */
		if ((NptrsPage.np_ptr()->ndLockRequestSeqNo_bd00() >= PrevPageSeqNo) &&
			(NptrsPage.np_ptr()->nd_id() != NptrsPage.np_ptr()->nd_sub_root()))		
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
			"BD92_VOC:wrong PageSeqNo",NptrsPage.np_ptr()->nd_id());
		
		PrevPageSeqNo = NptrsPage.np_ptr()->ndLockRequestSeqNo_bd00();
		
		/* check that the page has the same chain_no as the subroot */
		if (NptrsPage.np_ptr()->nd_chain_no() != ChainNo)
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
			"BD92_VOC:wrong ChainNo  ",NptrsPage.np_ptr()->nd_id());
		
		/* check object counters and object chains on current page */
		bd92_VerifyObjPage (Current, NptrsPage, ResObjCnt);     
		
		/* update page and object counter */
		++ChainStat.AllPageCnt;        
		
		if (NptrsPage.np_ptr()->nd_free_obj_cnt()     > 0)      ++ChainStat.FreePageCnt;
		if (NptrsPage.np_ptr()->nd_free_eot_obj_cnt() > 0)      ++ChainStat.FeotPageCnt;
		if (NptrsPage.np_ptr()->nd_free_obj_cnt() == NptrsPage.np_ptr()->nd_max_obj_cnt())
			++ChainStat.EmptyPageCnt;
		
		if (NptrsPage.np_ptr()->nd_in_free_chain())     ++ChainStat.PageWithFreeLabelCnt;
		if (NptrsPage.np_ptr()->nd_empty_page() && (NptrsPage.np_ptr()->nd_max_obj_cnt() > 0))
			++ChainStat.PageWithEmptyLabelCnt;         
		
		ChainStat.FreeObjCnt += NptrsPage.np_ptr()->nd_free_obj_cnt();
		ChainStat.FeotObjCnt += NptrsPage.np_ptr()->nd_free_eot_obj_cnt();
		ChainStat.OccObjCnt  += NptrsPage.np_ptr()->nd_occ_obj_cnt();
		ChainStat.ResObjCnt  += ResObjCnt;
		
		/* goto next page in chain */
		NextPageNo = NptrsPage.np_ptr()->nd_right();
		NptrsPage.bd910ReleasePage();
		
		if (NIL_PAGE_NO_GG00 == NextPageNo)
			break;
		else
		{
			NptrsPage.bd910AssignToPage (NextPageNo);
			if (e_ok != TrError) 
				return;
		}
	}
	while (ChainStat.AllPageCnt <= SubRootPageInfo.bd900AllPageCnt());
    
	/* traverse chain of free pages (= pages with free object frames ) */
	if (! bChainIsDeleted)
	{		
		/* set page handler to fisrt page of free page chain */
		NptrsPage.bd910SetTo (NptrsSubRoot);
		do
		{	
            if ((NptrsPage.np_ptr()->nd_id() != NptrsPage.np_ptr()->nd_sub_root()) ||
                (NptrsPage.np_ptr()->nd_free_obj_cnt() > 0)) 
				++ChainStat.PageInFreeChainCnt;
			
			if (NptrsPage.np_ptr()->nd_free_obj_cnt() > 0)
				++ChainStat.FreePageInFreeChainCnt;
			
			if ((NptrsPage.np_ptr()->nd_free_obj_cnt() == NptrsPage.np_ptr()->nd_max_obj_cnt()) &&
				(NptrsPage.np_ptr()->nd_max_obj_cnt() > 0)) 
				ChainStat.EmptyPageInFreeChainCnt++;
			
			if (NptrsPage.np_ptr()->nd_in_free_chain()) 
			{
				++ChainStat.PageWithFreeLabelInFreeChainCnt;
				if (NptrsPage.np_ptr()->nd_free_obj_cnt() > 0) 
					++ChainStat.FreePageWithFreeLabelInFreeChainCnt;
			}
			
			if (NptrsPage.np_ptr()->nd_empty_page() && (NptrsPage.np_ptr()->nd_max_obj_cnt ()>0))
			{
				++ChainStat.PageWithEmptyLabelInFreeChainCnt; 
				if (NptrsPage.np_ptr()->nd_free_obj_cnt() == NptrsPage.np_ptr()->nd_max_obj_cnt())
					++ChainStat.EmptyPageWithEmptyLabelInFreeChainCnt;
			}
			
			/* goto next page in free chain */
			NextPageNo = NptrsPage.np_ptr()->nd_next_free();
			NptrsPage.bd910ReleasePage();
			
			if (NIL_PAGE_NO_GG00 == NextPageNo)
				break;
			else
			{
				NptrsPage.bd910AssignToPage (NextPageNo);
				if (e_ok != TrError) return;
			}
		}
		while (ChainStat.PageInFreeChainCnt <= ChainStat.AllPageCnt);
	}
	
    /* check consistency of the infos in page chain info */ 
	if (!bChainIsDeleted)
		bd92_CheckConsistencyOfChain (ChainStat, SubRootPageInfo);
	else
	{
		/* note: if a file is droped only the number of all pages is updated */
		/* during the drop procedures all other counters are dirty           */
		if (ChainStat.AllPageCnt != SubRootPageInfo.bd900AllPageCnt())
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
			"BD92_VOP:oddAllPageChain", ChainStat.AllPageCnt);
	}
	
	/* write vtrace entry */
	bd92_WriteChainStateToVtrace (Current, ChainStat, NptrsSubRoot, bChainIsDeleted);
}

/*---------------------------------------------------------------------------*/

static _INLINE void
bd92_VerifyContObjDataPage (
							tbd_current_tree   &Current,
							tbd_nodeptr         pPage
							)
{
    ROUTINE_DBG_MEO00 ("bd92_VerifyContObjDataPage");

	/* This function checks the counter of free, free_after_eot */
	/* and occupied objects for continuation object data pages. */
	/* Moreover, the free object chain is verified              */
	
	const int             MaxObjCnt          = pPage->nd_max_obj_cnt();
	const int             ObjFrameLen        = pPage->nd_obj_frame_len();
	const int             AlignedObjFrameLen = BD92ALIGN_OBJ_FRAME(ObjFrameLen);
	const int             NoReservedObjects  = 0;
	int                   ObjPos             = pPage->nd_first_obj();
	int                   NextFreeObjPos;
	int                   OccObjCnt          = 0;
	int                   FreeObjCnt         = 0;
	int                   FreeChainObjCnt    = 0;
	tbd900_ContObjFrame   *pContObjFrame;  
	
    /* recalculate object counters */	
	for (int ObjCnt=1; ObjCnt <= MaxObjCnt; ++ObjCnt)
	{
		/*  construct position of the object frame  header             */		
		pContObjFrame = (tbd900_ContObjFrame *) ((tsp00_Byte*) pPage + ObjPos);
		
		/* count occurence of the various object  states */
		if (pContObjFrame->cofHeader_bd900.cvoRootOIdPno_bd900  != NIL_PAGE_NO_GG00      &&
			pContObjFrame->cofHeader_bd900.cvoRootOIdPos_bd900  != NIL_OBJ_PAGE_POS_GG92 &&
			pContObjFrame->cofHeader_bd900.cvoNextFreeObj_bd900 == NIL_OBJ_PAGE_POS_GG92)
		{
			++OccObjCnt;
		};	
		ObjPos += AlignedObjFrameLen;
	}
	FreeObjCnt = MaxObjCnt - OccObjCnt;
	

	/* traverse free obj chain */	
	NextFreeObjPos = pPage->nd_first_free_obj();
	
	while (NextFreeObjPos != NIL_OBJ_PAGE_POS_GG92)
	{
		++FreeChainObjCnt;
		
		if (!bd92ObjPosIsValid (NextFreeObjPos, pPage->nd_first_obj(), ObjFrameLen))
		{
			bd92_WritePageStateToVtrace (*Current.curr_trans, pPage, NoReservedObjects);  
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj,
				"BD92_VER:wrong obj pos  ", NextFreeObjPos);
		}
		
		pContObjFrame = (tbd900_ContObjFrame *) ((tsp00_Byte*) pPage + NextFreeObjPos);
		
		if (pContObjFrame->cofHeader_bd900.cvoRootOIdPno_bd900  != NIL_PAGE_NO_GG00      ||
			pContObjFrame->cofHeader_bd900.cvoRootOIdPos_bd900  != NIL_OBJ_PAGE_POS_GG92 )
		{
			bd92_WritePageStateToVtrace (*Current.curr_trans, pPage, NoReservedObjects); 
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
				"BD92_VER:wrong cont obj ", NextFreeObjPos);
		}
		
		if (FreeChainObjCnt > FreeObjCnt)
		{
			bd92_WritePageStateToVtrace (*Current.curr_trans, pPage, NoReservedObjects); 
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
				"BD92_VER:too long chain ", FreeObjCnt);
			break;
		}

		NextFreeObjPos = pContObjFrame->cofHeader_bd900.cvoNextFreeObj_bd900;
	}
	

	/* compare calculated counters with counters on page */
	if (FreeObjCnt        != pPage->nd_free_obj_cnt()     ||
        FreeChainObjCnt   != pPage->nd_free_obj_cnt()     ||
		0                 != pPage->nd_free_eot_obj_cnt() ||
		OccObjCnt         != pPage->nd_occ_obj_cnt())
	{
		bd92_WritePageStateToVtrace (*Current.curr_trans, pPage, NoReservedObjects);  
		
		if (FreeObjCnt != pPage->nd_free_obj_cnt())
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
			"BD92_VER:wrong FreeCnt  ", FreeObjCnt);
		
		if (0  != pPage->nd_free_eot_obj_cnt())
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
			"BD92_VER:wrong FeotCnt  ", pPage->nd_free_eot_obj_cnt() );
		
		if (OccObjCnt  != pPage->nd_occ_obj_cnt())
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
			"BD92_VER:wrong OccCnt   ",OccObjCnt );

		if (FreeChainObjCnt != pPage->nd_free_obj_cnt())
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
			"BD92_VER:wrong chain len", FreeChainObjCnt);
	}
};

/*---------------------------------------------------------------------------*/


static _INLINE void
bd92_VerifyObjDataPage (tbd_current_tree   &Current,
						tbd_nodeptr         pPage,
						tsp00_Int4         &ResObjCnt)
{
    ROUTINE_DBG_MEO00 ("bd92_VerifyObjDataPage");

	/* This function checks the counter of free, free_after_eot */
	/* and occupied objects for primary object data pages.      */
	/* Moreover, the free object chain is verified              */
	
	const int         MaxObjCnt          = pPage->nd_max_obj_cnt();
	const int         ObjFrameLen        = pPage->nd_obj_frame_len();
	const int         ObjKeyLen          = pPage->nd_obj_key_len();
	const int         AlignedObjFrameLen = BD92ALIGN_OBJ_FRAME(ObjFrameLen);
	int               ObjPos             = pPage->nd_first_obj();
	int               NextFreeObjPos;
	int               OccObjCnt          = 0;
	int               FreeObjCnt         = 0;
	int               FeotObjCnt         = 0;
	int               FreeChainObjCnt    = 0;
	                  ResObjCnt          = 0;
    bool              bThisTaskIsAGarbColl = bd91GarbCollController.bd930CheckIfTaskIsAGC (Current.curr_trans->trTaskId_gg00);
	tbd900_ObjFrame  *pObjFrame;
	tgg92_KernelOid   OId;
		
	/* preset common parameters of OId */
	OId.gg92SetPno    (pPage->nd_id());

	/* recalculate  object counters */

	if (ObjKeyLen <= 0)
	{
		/* check data page without keys */		
		for (int ObjCnt=1; ObjCnt <= MaxObjCnt; ++ObjCnt)
		{
			/*  construct position of the object frame  header */		
			pObjFrame = cbd910_ObjDataNode::bd910ObjFramePtr ((tsp00_Byte *) pPage, ObjPos);  

			/* create OId for this object frame */
			OId.gg92SetPos       (ObjPos);
			OId.gg92SetFrameVers (pObjFrame->objHeader_bd900.ohdFrameVersion_gg92);
			
			/* count occurence of the various object  states */
			switch (pObjFrame->objHeader_bd900.ohdState_gg92)
			{
			case obsOccupied_egg00     : ++OccObjCnt  ;	break; 
			case obsFree_egg00         : ++FreeObjCnt ; break;
			case obsFreeAfterEot_egg00 : 
				++FeotObjCnt ;
				if (!bThisTaskIsAGarbColl && // a GC could create a dead lock
					!g01restart_local   )   // lock list not yet initialized PTS 1107074 AK 30/06/2000
				{
						bd92_CheckBeforeImage (*Current.curr_trans, OId, 
							pObjFrame->objHeader_bd900.ohdTransInfo_gg92);
				}
				break;
			case obsReserved_egg00     : ++OccObjCnt  ; ++ResObjCnt; break; 	
			case obsKeyReserved_egg00  :
				g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_page, csp3_n_obj, 
					"BD92_VER:wrong State kre",ObjPos ); 
				break;
			default	                    :
				g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_page, csp3_n_obj, 
					"BD92_VER:unknown State  ",ObjPos ); 
				break; 
			};
			ObjPos += AlignedObjFrameLen;
		}
	}
	else
	{
		/* check data page with keys */		
		for (int ObjCnt=1; ObjCnt <= MaxObjCnt; ++ObjCnt)
		{
			/*  construct position of the object frame  header */
			pObjFrame = cbd910_ObjDataNode::bd910ObjFramePtr ((tsp00_Byte *) pPage, ObjPos);  

			/* create OId for this object frame */
			OId.gg92SetPos       (ObjPos);
			OId.gg92SetFrameVers (pObjFrame->objHeader_bd900.ohdFrameVersion_gg92);
			
			/* count occurence of the various object  states and check the hash entries */
			switch (pObjFrame->objHeader_bd900.ohdState_gg92)
			{
			case obsOccupied_egg00     : 
				bd92_CheckHashEntry (Current, pObjFrame, OId, ObjFrameLen, ObjKeyLen);
				++OccObjCnt;
				break; 
			case obsFreeAfterEot_egg00 :
				/* the hash is not tested because of the objects     */
				/* created temporaly  when an older object is reused */
				++FeotObjCnt;
				if (!bThisTaskIsAGarbColl && // a GC could create a dead lock
					!g01restart_local   )   // lock list not yet initialized PTS 1107074 AK 30/06/2000
				{
					bd92_CheckBeforeImage (*Current.curr_trans, OId, 
						pObjFrame->objHeader_bd900.ohdTransInfo_gg92);
				}
				break;
			case obsFree_egg00         : 
				++FreeObjCnt;
				break;
			case obsKeyReserved_egg00  :
				/* The hash is not tested since an object could be Keyreserved  */ 
				/* but not yet inserted into the hash (see NewObj in vkb04.cpp) */
				++OccObjCnt;
				++ResObjCnt;
				break;  
			case obsReserved_egg00     : 
				g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_page, csp3_n_obj,   
					"BD92_VER:wrong State res",ObjPos ); 
				break; 
            default	                   :
				g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_page, csp3_n_obj, 
					"BD92_VER:unknown State  ",ObjPos ); 
				break; 
			}			
			ObjPos += AlignedObjFrameLen;
		}
	}
	
	/* compare object counters */
	if (FreeObjCnt != pPage->nd_free_obj_cnt()     ||
		FeotObjCnt != pPage->nd_free_eot_obj_cnt() ||
		OccObjCnt  != pPage->nd_occ_obj_cnt()     )
	{
		bd92_WritePageStateToVtrace (*Current.curr_trans, pPage, ResObjCnt);  
		
		if (FreeObjCnt != pPage->nd_free_obj_cnt())
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
			"BD92_VER:wrong FreeCnt  ", FreeObjCnt);
		
		if (FeotObjCnt != pPage->nd_free_eot_obj_cnt())
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
			"BD92_VER:wrong FeotCnt  ", FeotObjCnt );
		
		if (OccObjCnt  != pPage->nd_occ_obj_cnt())
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
			"BD92_VER:wrong OccCnt   ",OccObjCnt );
	}
	
	/* check if all pages with free  objects are             */
	/* contained in the corresponding object page chains     */
	/* note that it may happen that the GC has changed the   */
	/* page state to free but the page is still not icluded  */
	/* in the free list since the GC was interupted by an    */
	/* user. This user finds the page free but not included  */
	/* in the free chain.                                    */

#   if COMPILEMODE_MEO00 >= SLOW_MEO00 
//	if (FreeObjCnt>0 && ! pPage->nd_in_free_chain())
//		g01opmsg (sp3p_knldiag, sp3m_info, csp3_b90_damaged_obj_chain, csp3_n_obj, 
//		"BD92_VER:ERR FREE List  ", pPage->nd_id());
//	
#   endif
	
	/* traverse free obj chain */	
	NextFreeObjPos = pPage->nd_first_free_obj();
	
	while (NextFreeObjPos != NIL_OBJ_PAGE_POS_GG92)
	{
		++FreeChainObjCnt;
		
		if (!bd92ObjPosIsValid (NextFreeObjPos, pPage->nd_first_obj(), ObjFrameLen))
		{
			bd92_WritePageStateToVtrace (*Current.curr_trans, pPage, ResObjCnt);  
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj,
				"BD92_VER:wrong obj pos  ", NextFreeObjPos);
		}
		
		pObjFrame = cbd910_ObjDataNode::bd910ObjFramePtr ((tsp00_Byte *) pPage, NextFreeObjPos);  
		
		if (pObjFrame->objHeader_bd900.ohdState_gg92 != obsFree_egg00)
		{
			bd92_WritePageStateToVtrace (*Current.curr_trans, pPage, ResObjCnt); 
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
				"BD92_VER:wrong obj state", pObjFrame->objHeader_bd900.ohdState_gg92);
		}
		
		if (FreeChainObjCnt > FreeObjCnt)
		{
			bd92_WritePageStateToVtrace (*Current.curr_trans, pPage, ResObjCnt); 
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
				"BD92_VER:too long chain ", FreeObjCnt);
			break;
		}
		
		NextFreeObjPos = pObjFrame->objHeader_bd900.ohdNextFreeObj_gg92;
	}
	
	/* compare calculated counters with counters on page */
	if (FreeChainObjCnt != pPage->nd_free_obj_cnt())
	{
		bd92_WritePageStateToVtrace (*Current.curr_trans, pPage, ResObjCnt); 
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
			"BD92_VER:wrong chain len", FreeChainObjCnt);
	}	
}

/*---------------------------------------------------------------------------*/

static _INLINE void
bd92_VerifyObjPage (tbd_current_tree  &Current,
   				    tbd_node_ptrs     &NptrsPage,
				    tsp00_Int4        &ResObjCnt)
{
    ROUTINE_DBG_MEO00 ("bd92_VerifyObjPage");
	
	/* This routine checks the consistency */
	/* of the contents of an object page   */

	/* initialize return parameter */
	ResObjCnt = 0;
	
    tbd_nodeptr   & pPage = NptrsPage.np_ptr();

	/* check page type */
    if (pPage->nd_pt()   !=  ptData_egg00           ||
		(pPage->nd_pt2()  !=  pt2Object_egg00     &&
	     pPage->nd_pt2()  !=  pt2VarObject_egg00  && 
		 pPage->nd_pt2()  !=  pt2ContObject_egg00)) 
	{
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
			"BD92_VOP: bad page type ", pPage->nd_id());
	}
	else
	{
        /* perform some general checks */

		/* check that the page has no next free page */
        /* if it is not a member of the free chain   */
		if (!pPage->nd_in_free_chain()                      && 
			(pPage->nd_id()        != pPage->nd_sub_root()) &&
			(pPage->nd_next_free() != NIL_PAGE_NO_GG00))
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
			"BD92_VOC:wrong NextFreeP", pPage->nd_id());
		
		/* check that if the nd_empty flag is set the page */
		/* is really  empty and member of the free chain   */ 
		if (pPage->nd_empty_page() &&
			((!pPage->nd_in_free_chain() && (pPage->nd_num_chain_updater() == 0)) || 
			(pPage->nd_free_obj_cnt() != pPage->nd_max_obj_cnt())))
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_obj_chain, csp3_n_obj, 
			"BD92_VOC: Page Not Empty", pPage->nd_id());

		/* check page header */
        bd92_VerifyObjPageHeader (Current, pPage);

		/* check page contents */
		if (pPage->nd_id() == pPage->nd_root()) 
			bd92_VerifyObjRootPage (NptrsPage);	
		
		if (pPage->nd_id() == pPage->nd_sub_root())
			bd92_VerifyObjSubRootPage (NptrsPage);
		
		if (pPage->nd_pt2()  ==  pt2ContObject_egg00)
			bd92_VerifyContObjDataPage (Current, pPage);
		else
			bd92_VerifyObjDataPage (Current, pPage, ResObjCnt);
	}
}

/*---------------------------------------------------------------------------*/

static _INLINE void
bd92_VerifyObjPageHeader (tbd_current_tree  &Current,
						  tbd_nodeptr        pPage)
{						  
	/* check if the data in the page header are consistent */
	/* with those in the object file directory             */

	const tsp00_Int4      FileNo        = pPage->nd_obj_file_no();
	cbd900_ObjFileInfo   *pObjFileInfo  = NULL;
	bool                  bHeaderOk     = true;
	
	/* get shortcut to the appropriate ObjFileInfo */
	if (pPage->nd_pt2() == pt2ContObject_egg00)
	{
		const int ContFileNo = Current.curr_tree_id.fileContObjFileNo_gg00();
		pObjFileInfo         = &(bd93ContObjFileDir[FileNo][ContFileNo]);
	}
	else
		pObjFileInfo = &(bd90ObjFileDir[FileNo]);
	
	/* check root */
	if (pPage->nd_root() != pObjFileInfo->ofiRoot_bd900)
	{
		bHeaderOk = false;
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
			"BD92_VOH:wrong root     ", pObjFileInfo->ofiRoot_bd900);
	}
	
	/* check maximum number of objects and position of first object */ 
	if (pPage->nd_root() == pPage->nd_id() || pPage->nd_sub_root() == pPage->nd_id() )
	{
		if (pPage->nd_max_obj_cnt() > pObjFileInfo->ofiMaxObjPerPageCnt_bd900)
		{
			bHeaderOk = false;
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
				"BD92_VOH:wrong MaxObjCnt", pObjFileInfo->ofiMaxObjPerPageCnt_bd900);
		}

		if (pPage->nd_first_obj() != BD92ALIGN_OBJ_FRAME (NODE_HEADER_MXBD00 + sizeof (cbd900_RootAndSubRootPageData)))
		{
			bHeaderOk = false;
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
				"BD92_VOH:wrong FirstOPos", pPage->nd_first_obj());
		}
	}
	else
	{
		if (pPage->nd_max_obj_cnt() != pObjFileInfo->ofiMaxObjPerPageCnt_bd900)
		{
			bHeaderOk = false;
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
				"BD92_VOH:wrong MaxObjCnt", pObjFileInfo->ofiMaxObjPerPageCnt_bd900);
		}

		if (pPage->nd_first_obj() != NODE_HEADER_MXBD00)
		{
			bHeaderOk = false;
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
				"BD92_VOH:wrong FirstOPos", pPage->nd_first_obj());
		}
	}

	if ((pPage->nd_first_free_obj() != NIL_OBJ_PAGE_POS_GG92) && 
		!bd92ObjPosIsValid (pPage->nd_first_free_obj(),  pPage->nd_first_obj(), pPage->nd_obj_frame_len()))
			bHeaderOk = false;
	
	/* check page type */
	if (((pPage->nd_pt2() == pt2VarObject_egg00)        &&
		(!pObjFileInfo->ofiFileIsForVarObj_bd900))       ||
		(((pPage->nd_pt2() == pt2Object_egg00) ||
		(pPage->nd_pt2() == pt2ContObject_egg00))    &&
		(pObjFileInfo->ofiFileIsForVarObj_bd900)))
	{
		bHeaderOk = false;
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, "BD92_VOH:wrong PType    ", pPage->nd_id());
	}

	if (!bHeaderOk)
	{
#          if COMPILEMODE_MSP00 >= QUICK_MSP00
			g01abort (csp3_bd_msg, csp3_n_obj, 
				"BD92_VOH:WrongPageHeader", pPage->nd_id());
#          else
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
				"BD92_VOH:WrongPageHeader", pPage->nd_id());
#          endif
	}
}

/*---------------------------------------------------------------------------*/

static _INLINE void
bd92_VerifyObjRootPage (tbd_node_ptrs  &NptrsRoot)
{
    ROUTINE_DBG_MEO00 ("bd92VerifyObjRootPage");
	
	/* check root page and compare data with ObjFileInfo */
	
	int iChain;
	cbd900_RootPageInfo   RootPageInfo (NptrsRoot); 
    tbd_nodeptr          &pRoot         = NptrsRoot.np_ptr();
	const tsp00_Int4      MaxChain      = RootPageInfo.bd900NumChains();
	const tsp00_Int4      FileNo        = pRoot->nd_obj_file_no();
	cbd900_ObjFileInfo   *pObjFileInfo  = NULL;
	
	/* get shortcut to the appropriate ObjFileInfo */
    if (pRoot->nd_pt2() == pt2ContObject_egg00)
	{
		cbd900_ContObjFileInfo & ContObjFileInfo = bd93ContObjFileDir[FileNo];

		for (int iContFileNo=1; iContFileNo<=ContObjFileInfo.cfiNumContFiles_bd900; ++iContFileNo)           
		{
			if (pRoot->nd_root() == ContObjFileInfo[iContFileNo].ofiRoot_bd900)
			{
				pObjFileInfo = &(ContObjFileInfo [iContFileNo]);
				break;
			}
		}
		if (pObjFileInfo == NULL)
			g01abort (csp3_bd_msg, csp3_n_obj,"BD92_VER:ContRoot NoReg ",pRoot->nd_id());
	}
	else
		pObjFileInfo = &(bd90ObjFileDir[FileNo]);
	
    /* check number of subroots */
    if (MaxChain != pObjFileInfo->ofiMaxChain_bd900)
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj,   
		"BD92_VER:WrongMaxChainNo",pRoot->nd_id());

	/* check that the drop info on the root page ist consistent */
	/* with those in the objfiledir , PTS 1106900 14/06/2000 AK */
	if (RootPageInfo.bd900FileIsDeleted() != pObjFileInfo->ofiRelFileIsReq_bd900)
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj,   
		"BD92_VER: WrongDropInfo ",pRoot->nd_id());
		
    /* check each entry in subroot list on root page */
	for (iChain=0; iChain < MaxChain; ++iChain)
	{
		if (pObjFileInfo->ofiRelFileIsReq_bd900)
		{
			/* note: if a chain is deleted the subroot entry on the root  */
			/* page is not updated automatically but the object file info */
			if ((NIL_PAGE_NO_GG00                  != pObjFileInfo->ofiChain_bd900[iChain].ociSubRoot_bd900) &&
                (RootPageInfo.bd900SubRoot(iChain) != pObjFileInfo->ofiChain_bd900[iChain].ociSubRoot_bd900))
			{
				g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj,   
					"BD92_VER:wrong RootPage1",pRoot->nd_id());
				g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj,   
					"SubrootList!=ObjFileInfo", iChain);
				g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj,   
					"ObjFileInfo.SubRoot     ", pObjFileInfo->ofiChain_bd900[iChain].ociSubRoot_bd900);
			}
		}
		else
		{
			if (RootPageInfo.bd900SubRoot(iChain) == NIL_PAGE_NO_GG00)
			{
				g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj,   
					"BD92_VER:wrong Root page",pRoot->nd_id());
				g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj,   
					"incomplete  subroot list", iChain);
			}
			else
			{   /* compare subroot list on root and in object file info */ 
				if (RootPageInfo.bd900SubRoot(iChain) != pObjFileInfo->ofiChain_bd900[iChain].ociSubRoot_bd900)
				{
					g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj,   
						"BD92_VER:wrong RootPage2",pRoot->nd_id());
					g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj,   
						"SubrootList!=ObjFileInfo", iChain);
					g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj,   
					"ObjFileInfo.SubRoot     ", pObjFileInfo->ofiChain_bd900[iChain].ociSubRoot_bd900);
				}
			}
		}
	}

	for (iChain=MaxChain; iChain<MAX_NUM_CHAINS_PER_CONTAINER_BD900; ++iChain)
	{
		if (RootPageInfo.bd900SubRoot(iChain) != NIL_PAGE_NO_GG00)
		{
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_page_chain, csp3_n_obj,   
				"BD92_VER:wrong Root page",pRoot->nd_id());
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_page_chain, csp3_n_obj,   
				"too many subroots       ", iChain);
		}
	}
}

/*---------------------------------------------------------------------------*/

static _INLINE void
bd92_VerifyObjSubRootPage (tbd_node_ptrs  &NptrsSubRoot)
{
    ROUTINE_DBG_MEO00 ("bd92VerifyObjSubRootPage");

	cbd900_SubRootPageInfo  SubRootPageInfo (NptrsSubRoot);
    cbd900_ObjChainInfo    *pObjChainInfo = NULL;
	tbd_nodeptr            &pSubRoot      = NptrsSubRoot.np_ptr();
	const tsp00_Int4        FileNo        = pSubRoot->nd_obj_file_no();
    const tsp00_Int4        ChainNo       = pSubRoot->nd_chain_no();

	/* get shortcut to the appropriate ObjChainInfo */
    if (pSubRoot->nd_pt2() == pt2ContObject_egg00)
	{
		cbd900_ContObjFileInfo & ContObjFileInfo = bd93ContObjFileDir[FileNo];

		for (int iContFileNo=1; iContFileNo<=ContObjFileInfo.cfiNumContFiles_bd900; ++iContFileNo)    
		{
			if (pSubRoot->nd_root() == ContObjFileInfo[iContFileNo].ofiRoot_bd900)
			{
				pObjChainInfo = &(ContObjFileInfo[iContFileNo][ChainNo]);
				break;
			}
		}
       if (pObjChainInfo == NULL)
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj,   
				"BD92_VER:ContRoot NoReg ",pSubRoot->nd_id());
	}
	else
		pObjChainInfo = &(bd90ObjFileDir[FileNo][ChainNo]);

	/* compare number of all pages on subroot and in ObjChainInfo */
	if (pObjChainInfo->ociAllPageCnt_bd900 != SubRootPageInfo.bd900AllPageCnt())
	{	
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj,   
			"BD92_VER:wrong SubRoot  ",pSubRoot->nd_id());
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj,   
			"AllPgCnt:  SubRoot     =", SubRootPageInfo.bd900AllPageCnt());
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj,   
			"AllPgCnt:  ObjFileInfo =", pObjChainInfo->ociAllPageCnt_bd900);
	}

	/* check ranges of page counters */	
	if (SubRootPageInfo.bd900AllPageCnt() < 0)
	{
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_page_chain, csp3_n_obj,   
			"BD92_VER:wrong SubRoot  ",pSubRoot->nd_id());
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_page_chain, csp3_n_obj,   
			"         wrong AllPgCnt ", SubRootPageInfo.bd900AllPageCnt());
	}
	
	if ((SubRootPageInfo.bd900FreePageCnt() < 0) ||
		(SubRootPageInfo.bd900FreePageCnt() > SubRootPageInfo.bd900AllPageCnt()))
	{
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_page_chain, csp3_n_obj,   
			"BD92_VER: odd  SubRoot  ",pSubRoot->nd_id());
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_page_chain, csp3_n_obj,  
			"         wrong FreePgCnt", SubRootPageInfo.bd900FreePageCnt());
	}
	
	if ((SubRootPageInfo.bd900EmptyPageCnt() < 0) ||
		(SubRootPageInfo.bd900EmptyPageCnt() > SubRootPageInfo.bd900FreePageCnt()))
	{
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_page_chain, csp3_n_obj,   
			"BD92_VER: odd  SubRoot  ",pSubRoot->nd_id());
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_b90_damaged_page_chain, csp3_n_obj,  
			"        wrong EmptyPgCnt", SubRootPageInfo.bd900EmptyPageCnt());
	}
}

/*---------------------------------------------------------------------------*/

static _INLINE void
bd92_WriteChainStateToVtrace (tbd_current_tree                &Current, 
							  const  tbd92_ObjPageChainStat   &ChainStat,
							  cbd910_SubRootNode              &NptrsSubRoot,
							  const tsp00_Bool                 bChainIsDeleted)
{
	ROUTINE_DBG_MEO00 ("bd92_WriteChainStateToVtrace");
    
	tgg12_OmsFilenameTrace TrLineIn;

	cbd900_SubRootPageInfo SubRootPageInfo (NptrsSubRoot);

	TrLineIn.ofnTrType_gg12.becomes(otFilename_egg12);
	TrLineIn.ofnTrType2_gg12.becomes(ot2Nil_egg12);
	TrLineIn.ofnFileName_gg12 = Current.curr_tree_id.fileName_gg00();
	b12vtrace (*Current.curr_trans, bd, b92verify, sizeof (TrLineIn), &TrLineIn);
	
	tgg12_OmsPageChainStatTrace TrLineChain;
	
	TrLineChain.opctSubRoot_gg12             =  NptrsSubRoot.np_ptr()->nd_id();
	TrLineChain.opctTrType_gg12.becomes(otPageChainStat_egg12);
	TrLineChain.opctAllPageCnt_gg12          =  ChainStat.AllPageCnt;
	TrLineChain.opctEmptyPageCnt_gg12        =  ChainStat.EmptyPageCnt;
	TrLineChain.opctFreePageCnt_gg12         =  ChainStat.FreePageCnt;
	TrLineChain.opctFeotPageCnt_gg12         =  ChainStat.FeotPageCnt;
	
	TrLineChain.opctAllPageRootCnt_gg12      =  SubRootPageInfo.bd900AllPageCnt();
	TrLineChain.opctEmptyPageRootCnt_gg12    =  SubRootPageInfo.bd900EmptyPageCnt();
	TrLineChain.opctFreePageRootCnt_gg12     =  SubRootPageInfo.bd900FreePageCnt();
	
	TrLineChain.opctPageInFreeChainCnt_gg12  =  ChainStat.PageInFreeChainCnt;
	
	TrLineChain.opctFreePageInFreeChainCnt_gg12  = ChainStat.FreePageInFreeChainCnt; 
	TrLineChain.opctEmptyPageInFreeChainCnt_gg12 = ChainStat.EmptyPageInFreeChainCnt;
	
	TrLineChain.opctPageWithEmptyLabelCnt_gg12 = ChainStat.PageWithEmptyLabelCnt;
	TrLineChain.opctPageWithFreeLabelCnt_gg12  = ChainStat.PageWithFreeLabelCnt;
	
	TrLineChain.opctPageWithFreeLabelInFreeChainCnt_gg12 = ChainStat.PageWithFreeLabelInFreeChainCnt;
	
	TrLineChain.opctFreePageWithFreeLabelInFreeChainCnt_gg12 = ChainStat.FreePageWithFreeLabelInFreeChainCnt;
	
	TrLineChain.opctFreeObjCnt_gg12 = ChainStat.FreeObjCnt;
	TrLineChain.opctFeotObjCnt_gg12 = ChainStat.FeotObjCnt;
	TrLineChain.opctOccObjCnt_gg12  = ChainStat.OccObjCnt;
	TrLineChain.opctResObjCnt_gg12  = ChainStat.ResObjCnt;
	
	TrLineChain.opctIsChainDeleted_gg12 = bChainIsDeleted;
	
	b12vtrace (*Current.curr_trans, bd, b92verify, sizeof (TrLineChain), &TrLineChain);
}


/*---------------------------------------------------------------------------*/

static _INLINE void
bd92_WriteContainerStateToVtrace (
							  tbd_current_tree                     &Current, 
							  const  tbd92_ObjPageContainerStat    &ContainerStat,
							  const  cbd900_ObjFileInfo            *pObjFileInfo
							  )
{ 
	ROUTINE_DBG_MEO00 ("bd92_WriteContainerStateToVtrace");
	
	tgg12_OmsFilenameTrace TrLineIn;
	
	TrLineIn.ofnTrType_gg12.becomes(otFilename_egg12);
	TrLineIn.ofnTrType2_gg12.becomes(ot2Nil_egg12);
	TrLineIn.ofnFileName_gg12 = Current.curr_tree_id.fileName_gg00();
	b12vtrace (*Current.curr_trans, bd, b92verify, sizeof (TrLineIn), &TrLineIn);
	
	tgg12_OmsPageContainerStatTrace TrLineCon;
	
	TrLineCon.opcoTrType_gg12.becomes(otPageContainerStat_egg12);
	TrLineCon.opcoFreeObjCnt_gg12    =  ContainerStat.FreeObjCnt;       
	TrLineCon.opcoFeotObjCnt_gg12    =  ContainerStat.FeotObjCnt; 
	TrLineCon.opcoOccObjCnt_gg12     =  ContainerStat.OccObjCnt;
	TrLineCon.opcoResObjCnt_gg12     =  ContainerStat.ResObjCnt;
	TrLineCon.opcoAllPageCnt_gg12    =  ContainerStat.AllPageCnt; 
	TrLineCon.opcoEmptyPageCnt_gg12  =  ContainerStat.EmptyPageCnt;
	TrLineCon.opcoFreePageCnt_gg12   =  ContainerStat.FreePageCnt;
	TrLineCon.opcoFeotPageCnt_gg12   =  ContainerStat.FeotPageCnt;
	TrLineCon.opcoNumSubRoots_gg12   =  pObjFileInfo->ofiMaxChain_bd900;
	TrLineCon.opcoIsFileDeleted_gg12 =  pObjFileInfo->ofiRelFileIsReq_bd900;
	
	b12vtrace (*Current.curr_trans, bd, b92verify, sizeof (TrLineCon), &TrLineCon);
}

/*---------------------------------------------------------------------------*/

static _INLINE void
bd92_WritePageStateToVtrace (  
							 tgg00_TransContext     &Trans,
							 tbd_nodeptr             pPage,
							 tsp00_Int4              ResObjCnt
							 )
{                             
	ROUTINE_DBG_MEO00 ("bd92_WritePageStateToVtrace");
    
    tgg12_OmsPageStatTrace     TrLinePage;
    
    TrLinePage.opsTrType_gg12.becomes(otPageStat_egg12);
    TrLinePage.opsPno_gg12           = pPage->nd_id();
    TrLinePage.opsOccObjCnt_gg12     = pPage->nd_occ_obj_cnt();
    TrLinePage.opsFreeObjCnt_gg12    = pPage->nd_free_obj_cnt();
    TrLinePage.opsFreeEotObjCnt_gg12 = pPage->nd_free_eot_obj_cnt();
    TrLinePage.opsResObjCnt_gg12     = ResObjCnt;                
    
    b12vtrace (Trans, bd, b92verify, sizeof (TrLinePage), &TrLinePage);
    
}

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/