/*
  -----------------------------------------------------------------------------

  module: gcn003.h

  -----------------------------------------------------------------------------

  responsible:  BerndV, MartinR, TiloH

  special area: DBMServer

  description:  general header for backup/recovery area

  -----------------------------------------------------------------------------


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/
#ifndef _GCN003_H_
#define _GCN003_H_

/* syntax key words */
#define ME_KEY_LOCATION_CN003       "LOCATION"
#define ME_KEY_TYPE_CN003           "TYPE"
#define ME_KEY_SAVETYPE_CN003       "SAVETYPE"
#define ME_KEY_SIZE_CN003           "SIZE"
#define ME_KEY_OVERWRITE_CN003      "OVERWRITE"
#define ME_KEY_BLOCKSIZE_CN003      "BLOCKSIZE"
#define ME_KEY_AUTOLOADER_CN003     "AUTOLOADER"
#define ME_KEY_OSCOMMAND_CN003      "OSCOMMAND"
#define ME_KEY_CREATED_CN003        "CREATED"
#define ME_KEY_MODIFIED_CN003       "MODIFIED"
#define ME_KEY_EXTERN_CN003         "EXTERN"

#define ME_KEY_MEDIAID_CN003        "MEDIA_ID"    /* for old media file */
#define ME_KEY_BLANK_CN003          ' '           /* for old media file */
#define ME_KEY_PROPSEP_CN003        '|'           /* for old media file */
#define ME_KEY_PARALLEL_CN003       "PARALLEL"    /* for old media file */

/* special characters and strings */
#define ME_COMMENT_CN003			    '\''
#define ME_OLDCOMMENT_CN003			  '#'
#define ME_NAME_BEGIN_CN003			  '['
#define ME_GROUP_NAME_SEP1_CN003  '\\'
#define ME_GROUP_NAME_SEP2_CN003  '/'
#define ME_NAME_END_CN003		      ']'
#define ME_PROP_EQUAL_CN003		    '='

/* property value keywords */
#define ME_PROP_YES_CN003           "YES"
#define ME_PROP_NO_CN003            "NO"
#define ME_PROP_VERSION_CN003       "VERSION"

/* old property value keywords */
#define ME_OLD_PROP_YES_CN003           "Y"
#define ME_OLD_PROP_NO_CN003            "N"
#define ME_OLD_PROP_VERSION_CN003       "V"

/* type property value keywords */
#define ME_TYPE_TAPE_CN003          "TAPE"
#define ME_TYPE_FILE_CN003          "FILE"
#define ME_TYPE_NORE_CN003          "NOREWIND"
#define ME_TYPE_PIPE_CN003          "PIPE"
#define ME_TYPE_UNKN_CN003          "UNKNOWN"

/* external property value keywords */
#define ME_EXT_NONE_CN003           "NONE"
#define ME_EXT_NSR_CN003            "NSR"
#define ME_EXT_ADSM_CN003           "BACK"
#define ME_EXT_BACK_CN003           "BACK"
#define ME_EXT_OMNI_CN003           "OMNI"

/* savetype property value keywords */
#define ME_SAVETYPE_DAT_CN003       "DATA"
#define ME_SAVETYPE_PAG_CN003       "PAGES"
#define ME_SAVETYPE_LOG_CN003       "LOG"
#define ME_SAVETYPE_AUT_CN003       "AUTO"
#define ME_SAVETYPE_UNK_CN003       "UNKNOWN"

/* type property value keywords for old media file*/
#define ME_OLD_TYPE_TAPE_CN003      "T"
#define ME_OLD_TYPE_FILE_CN003      "F"
#define ME_OLD_TYPE_NORE_CN003      "N"
#define ME_OLD_TYPE_PIPE_CN003      "P"
#define ME_OLD_TYPE_AUTO_CN003      "A"
#define ME_OLD_TYPE_AUTL_CN003      "L"
#define ME_OLD_TYPE_UNKN_CN003      "U"

/* default values for medium */
#define ME_DEFAULT_NAME_CN003           ""
#define ME_DEFAULT_LOCATION_CN003       ""
#define ME_DEFAULT_TYPE_CN003           ME_TYPE_UNKN_CN003
#define ME_DEFAULT_SAVETYPE_CN003       ME_SAVETYPE_UNK_CN003
#define ME_DEFAULT_SIZE_CN003           "0"
#define ME_DEFAULT_OVERWRITE_CN003      ME_PROP_NO_CN003
#define ME_DEFAULT_BLCCKSIZE_CN003      "8"
#define ME_DEFAULT_AUTOLOADER_CN003     ME_PROP_NO_CN003
#define ME_DEFAULT_OSCMD_CN003          ""
#define ME_DEFAULT_DATE_CN003           "19980101000000"
#define ME_DEFAULT_EXTERN_CN003         ME_EXT_NONE_CN003

#define ME_PROPERTY_NAME_LENGTH_MXCN003 64
#define ME_PROPERTY_VAL_LENGTH_MXCN003  256

#define ME_LINE_LENGTH_MXCN003          4096

#define ME_NAME_LENGTH_MXCN003          64
#define ME_LOCATION_LENGTH_MXCN003      256
#define ME_TYPE_LENGTH_MXCN003          32
#define ME_SIZE_LENGTH_MXCN003          16
#define ME_OVERWRITE_LENGTH_MXCN003     16
#define ME_BLOCKSIZE_LENGTH_MXCN003     16
#define ME_AUTOLOADER_LENGTH_MXCN003    16
#define ME_OSCMD_LENGTH_MXCN003         256
#define ME_SAVETYPE_LENGTH_MXCN003      16
#define ME_DATE_LENGTH_MXCN003          16
#define ME_EXTERNAL_LENGTH_MXCN003      16

/* medium structure */
typedef struct tcn003_Medium{
  char                  szName[ME_NAME_LENGTH_MXCN003 + 1];             /* name of medium                */
  char                  szLocation[ME_LOCATION_LENGTH_MXCN003 + 1];     /* location for medium           */
  char                  szType[ME_TYPE_LENGTH_MXCN003 + 1];             /* type of medium                */
  char                  szSaveType[ME_SAVETYPE_LENGTH_MXCN003 + 1];     /* use medium for DATA/PAGES/LOG */
  char                  szSize[ME_SIZE_LENGTH_MXCN003 + 1];             /* max size in pages of medium   */
  char                  szOverwrite[ME_OVERWRITE_LENGTH_MXCN003 + 1];   /* medium can overwrite          */
  char                  szBlocksize[ME_BLOCKSIZE_LENGTH_MXCN003 + 1];   /* blocksize for read/write      */
  char                  szAutoLoader[ME_AUTOLOADER_LENGTH_MXCN003 + 1]; /* is it a autoloder             */
  char                  szOSCommand[ME_OSCMD_LENGTH_MXCN003 + 1];       /* execute os command            */
  char                  szCreated[ME_DATE_LENGTH_MXCN003 + 1];          /* create date                   */
  char                  szModified[ME_DATE_LENGTH_MXCN003 + 1];         /* modify date                   */
  char                  szExtern[ME_EXTERNAL_LENGTH_MXCN003 + 1];       /* external type                 */

  bool                  bOld;                                       /* medium from old definition   */

  struct tcn003_Medium  * pNext;                                    /* pointer to next medium       */
  struct tcn003_Medium  * pPrev;                                    /* pointer to previous medium   */
} tcn003_Medium;

/* mediagroup structure */
typedef struct tcn003_MediaGroup{
  char                      szName[ME_NAME_LENGTH_MXCN003];         /* name of mediagroup           */
  tcn003_Medium             * pFirst;                               /* pointer to first medium      */
  tcn003_Medium             * pLast;                                /* pointer to last medium       */
  struct tcn003_MediaGroup  * pNext;                                /* pointer to next mediagroup   */
  struct tcn003_MediaGroup  * pPrev;                                /* pointer to previous mediagr. */
} tcn003_MediaGroup;

/* media structure */
typedef struct tcn003_Media{
  tcn003_MediaGroup     * pFirstGroup;                              /* pointer to first mediagroup  */
  tcn003_MediaGroup     * pLastGroup;                               /* pointer to last mediagroup   */
} tcn003_Media;

/* key words for save/restore statements   */
#define BA_KEY_SAVE_CN003           "SAVE"
#define BA_KEY_RESTORE_CN003        "RESTORE"
#define BA_KEY_STATE_SAVE_CN003     "STATE SAVE"
#define BA_KEY_STATE_RESTORE_CN003  "STATE RESTORE"

#define BA_KEY_RESTART_INFO_CN003   "GET RESTART VERSION"
#define BA_KEY_READLABEL_CN003      "READ LABEL"

#define BA_KEY_DATA_CN003           "DATA"
#define BA_KEY_PAGES_CN003          "PAGES"
#define BA_KEY_LOG_CN003            "LOG"
#define BA_KEY_CONFIG_CN003         "CONFIG"
#define BA_KEY_DEVSPACE_CN003       "DEVSPACE"

#define BA_KEY_QUICK_CN003          "QUICK"
#define BA_KEY_TO_CN003             "TO"
#define BA_KEY_FROM_CN003           "FROM"

#define BA_KEY_UNTIL_CN003          "UNTIL"
#define BA_KEY_EBIDLONG_CN003       "ExternalBackupID"
#define BA_KEY_EBID_CN003           "EBID"
#define BA_KEY_IBID_CN003           "IBID"
#define BA_KEY_KEY_CN003            "KEY"
#define BA_KEY_LABEL_CN003          "LABEL"
#define BA_KEY_STAGE_CN003          "STAGE"
#define BA_KEY_FILENUMBERLIST_CN003 "FILENUMBERLIST"
#define BA_KEY_FNL_CN003            "FNL"
#define BA_KEY_VERIFY_CN003         "VERIFY"
#define BA_KEY_NOVERIFY_CN003       "NOVERIFY"
#define BA_KEY_REMOVE_CN003         "REMOVE"
#define BA_KEY_KEEP_CN003           "KEEP"

#define BA_KEY_REPLACE_CN003        "REPLACE"
#define BA_KEY_AUTOIGNORE_CN003     "AUTOIGNORE"
#define BA_KEY_IGNORE_CN003         "IGNORE"
#define BA_KEY_CANCEL_CN003         "CANCEL"

#define BE_KEY_COUNT_CN003          "COUNT"

#define BA_KEY_BLOCKSIZE_CN003      "BLOCKSIZE"
#define BA_KEY_CASCADE_CN003        "CASCADE"
#define BA_KEY_NOCHECKPOINT_CN003   "NO CHECKPOINT"
#define BA_KEY_FVERSION_CN003       "FVERSION"
#define BA_KEY_MEDIANAME_CN003      "MEDIANAME"

#define BA_KEY_RECOVERY_CN003       "RECOVERY"
#define BA_KEY_MIGRATION_CN003      "MIGRATION"

#define BA_KEY_DEFAULTDEV_CN003     "'DEV'"

#define BA_KEY_BLANK_CN003          " "
#define BA_KEY_QUOTE_CN003          "\'"
#define BA_KEY_COMMA_CN003          ","
#define BA_KEY_DBLQUOTE_CN003       "\""

#define BA_KEY_BAKSLASH_CN003       "\\"
#define BA_KEY_FORSLASH_CN003       "/"

#define BA_KEY_AUTOSAVE_CN003       "AUTOSAVE"
#define BA_KEY_ON_CN003             "ON"
#define BA_KEY_END_CN003            "END"
#define BA_KEY_CANCEL_CN003         "CANCEL"
#define BA_KEY_SHOW_CN003           "SHOW"

#define BA_KEY_TIMEOUTOFF_CN003     "SET TIMEOUT OFF"

#define BA_TYPE_LENGTH_MXCN003      64
#define BA_FILENUM_LENGTH_MXCN003   32
#define BA_UNTIL_LENGTH_MXCN003     32

#define BA_FILESPEC_LENGTH_MXCN003  4096
#define BA_COMMAND_LENGTH_MXCN003   5120

#define BA_NEXTMEDIUM_RC_CN003      (-8020)

#define BA_AUTOSAVE_RC_ON_CN003       0
#define BA_AUTOSAVE_RC_OFF_CN003      100
#define BA_AUTOSAVE_RC_NEXT_CN003     BA_NEXTMEDIUM_RC_CN003


#define BA_AUTOSAVE_RESULT_ON_CN003   "AUTOSAVE IS ON"
#define BA_AUTOSAVE_RESULT_OFF_CN003  "AUTOSAVE IS OFF"
#define BA_AUTOSAVE_RESULT_NEXT_CN003 "NEXT VOLUME REQUIRED"

/* backup result structur */
typedef struct tcn003_BackupResult{
  int   nBackupRC;
  char  szDate[18+1];
  char  szTime[18+1];
  char  szDatabase[18+1];
  char  szServer[64+1];
  char  szKernelVersion[40+1];
  char  szPagesTransfered[10+1];
  char  szPagesLeft[10+1];
  char  szVolumes[10+1];
  char  szMediaName[40+1];
  char  szLocation[64+1];
  char  szErrortext[40+1];
  char  szLabel[10+1];
  char  szIsConsistent[5+1];
  char  szFirstLogPageNo[10+1];
  char  szLastLogPageNo[10+1];
  char  szDBStamp1Date[8+1];
  char  szDBStamp1Time[8+1];
  char  szDBStamp2Date[8+1];
  char  szDBStamp2Time[8+1];
  char  szBDPageCount[10+1];
  char  szDevicesUsed[10+1];
  char  szDatabaseID[132+1];
  char  szMaxUsedDataPage[10+1];
  char  szRedoTransRead[10+1];
  char  szRedoTransDone[10+1];
} tcn003_BackupResult;

#define BA_RESULT_PARAM_COUNT_CN003   25

#define BA_RESULT_RC_CN003         "Returncode"

#define BA_RESULT_DATE_CN003       "Date"
#define BA_RESULT_TIME_CN003       "Time"
#define BA_RESULT_DATABASE_CN003   "Database"
#define BA_RESULT_SERVER_CN003     "Server"
#define BA_RESULT_KNLVER_CN003     "Kernel Version"
#define BA_RESULT_PGSTRAN_CN003    "Pages Transferred"
#define BA_RESULT_PGSLEFT_CN003    "Pages Left"
#define BA_RESULT_VOLUMES_CN003    "Volumes"
#define BA_RESULT_MEDNAME_CN003    "Medianame"
#define BA_RESULT_LOCATION_CN003   "Location"
#define BA_RESULT_ERRTEXT_CN003    "Errortext"
#define BA_RESULT_LABEL_CN003      "Label"
#define BA_RESULT_ISCONS_CN003     "Is Consistent"
#define BA_RESULT_FLOGPG_CN003     "First LOG Page"
#define BA_RESULT_LLOGPG_CN003     "Last LOG Page"
#define BA_RESULT_DBSTMD1_CN003    "DB Stamp 1 Date"
#define BA_RESULT_DBSTMT1_CN003    "DB Stamp 1 Time"
#define BA_RESULT_DBSTMD2_CN003    "DB Stamp 2 Date"
#define BA_RESULT_DBSTMT2_CN003    "DB Stamp 2 Time"
#define BA_RESULT_PGCNT_CN003      "Page Count"
#define BA_RESULT_DEVUSE_CN003     "Devices Used"
#define BA_RESULT_DBID_CN003       "Database ID"
#define BA_RESULT_MUSEDATAPG_CN003 "Max Used Data Page"
#define BA_RESULT_REDOTRANSR_CN003 "Redo Transactions Read"
#define BA_RESULT_REDOTRANSD_CN003 "Redo Transactions Done"

#define BA_RESULT_STAGEFILES_PROCESSED_CN003   "Processed Stage Files"
#define BA_RESULT_STAGEFILES_UNPROCESSED_CN003 "Remaining Stage Files"
#define BA_RESULT_STAGEFILES_DELETED_CN003     "Deleted Stage Files"
#define BA_RESULT_STAGEFILES_BYTES_CN003       "Bytes processed"

/* backup restart info */
typedef struct tcn003_BackupRestartInfo{
  char  szUsedLogPage[10+1];
  char  szFirstLogPage[10+1];
  char  szRestartable[10+1];
  char  szDBIdentRestart[132+1];
  char  szDBIdentLog[132+1];
  char  szConsistent[10+1];
} tcn003_BackupRestartInfo;

#define BA_RESTARTINFO_PARAM_COUNT_CN003   6

#define BA_RESTARTINFO_USEDLOGPAGE_CN003        "Used LOG Page"
#define BA_RESTARTINFO_FIRSTLOGPAGE_CN003       "First LOG Page"
#define BA_RESTARTINFO_RESTARTABLE_CN003        "Restartable"
#define BA_RESTARTINFO_IDRESTART_CN003          "Id Restart Record"
#define BA_RESTARTINFO_IDLOG_CN003              "Id LOG Info"
#define BA_RESTARTINFO_CONSISTENT_CN003         "Consistent"

class tcn35b_DBMExternalBackupController;   //a forward declaration to avoid include of hcn35b.h
class tcn35c_DBMStageBackupController;      //a forward declaration to avoid include of hcn35c.h
class tin101_XDBBackupOperation;  //a forward declaration to avoid include of hin101.h

/* backup session structur */
typedef struct tcn003_BackupSesion{
  int                     nSessionType;
  char                    szMediaName[ME_NAME_LENGTH_MXCN003];
  char                    szBackupType[BA_TYPE_LENGTH_MXCN003];

  char                    szFileNumber[BA_FILENUM_LENGTH_MXCN003];

  char                    szUntilDate[BA_UNTIL_LENGTH_MXCN003];
  char                    szUntilTime[BA_UNTIL_LENGTH_MXCN003];

  tcn003_Media            tMedia;

  char                    szCommand[BA_COMMAND_LENGTH_MXCN003];

  bool                    bCheckOnly; // only check with service kernel
  bool                    bCancelActive;
  bool                    bAutoignore;

  tcn003_BackupResult     tBackupResult;

  tcn35b_DBMExternalBackupController * pBackupTool;
  tcn35c_DBMStageBackupController * StageController;
  tin101_XDBBackupOperation * pXDB;
} tcn003_BackupSession;

#define BA_SESSION_TYPE_SAVE_CN003      1
#define BA_SESSION_TYPE_RESTORE_CN003   2
#define BA_SESSION_TYPE_COMMAND_CN003   3

/* backup history */
#define BA_HISTORY_LISTPARAM_COLS_CN003C     "-C"
#define BA_HISTORY_LISTPARAM_KEY_CN003C      "-K"
#define BA_HISTORY_LISTPARAM_LABEL_CN003C    "-L"
#define BA_HISTORY_LISTPARAM_ACTION_CN003C   "-A"
#define BA_HISTORY_LISTPARAM_RESTORE_CN003C  "-R"
#define BA_HISTORY_LISTPARAM_MEDIA_CN003C    "-M"
#define BA_HISTORY_LISTPARAM_EBIDS_CN003C    "-E"
#define BA_HISTORY_LISTPARAM_UNTIL_CN003C    "-U"

#define BA_HISTORY_LINE_LENGTH_MXCN003      512
#define BA_HISTORY_EBIDLINE_LENGTH_MXCN003 1024

#define BA_HISTORY_CONT_FLAG_END_CN003C       "END     \n"
#define BA_HISTORY_CONT_FLAG_CONTINUE_CN003C  "CONTINUE\n"
#define BA_HISTORY_CONT_LEN_CN003C            (strlen(BA_HISTORY_CONT_FLAG_END_CN003C))

#define BA_HISTORY_COL_ALL_CN003C       "*"

#define BA_HISTORY_COL_KEY_CN003C       "KEY"
#define BA_HISTORY_LEN_KEY_CN003        12
#define BA_HISTORY_COL_LBL_CN003C       "LABEL"
#define BA_HISTORY_LEN_LBL_CN003        9
#define BA_HISTORY_COL_ACT_CN003C       "ACTION"
#define BA_HISTORY_LEN_ACT_CN003        9
#define BA_HISTORY_COL_ST1_CN003C       "STAMP1"
#define BA_HISTORY_LEN_ST1_CN003        19
#define BA_HISTORY_COL_ST2_CN003C       "STAMP2"
#define BA_HISTORY_LEN_ST2_CN003        19
#define BA_HISTORY_COL_STA_CN003C       "START"
#define BA_HISTORY_LEN_STA_CN003        19
#define BA_HISTORY_COL_STO_CN003C       "STOP"
#define BA_HISTORY_LEN_STO_CN003        19
#define BA_HISTORY_COL_FLO_CN003C       "FIRSTLOG"
#define BA_HISTORY_LEN_FLO_CN003        10
#define BA_HISTORY_COL_LLO_CN003C       "LASTLOG"
#define BA_HISTORY_LEN_LLO_CN003        10
#define BA_HISTORY_COL_LOG_CN003C       "LOG"
#define BA_HISTORY_LEN_LOG_CN003        3
#define BA_HISTORY_COL_MED_CN003C       "MEDIA"
#ifdef CN71
  #define BA_HISTORY_LEN_MED_CN003      8
#else  /* #ifdef CN71 */
  #define BA_HISTORY_LEN_MED_CN003      64
#endif /* #ifdef CN71 */
#define BA_HISTORY_COL_PAG_CN003C       "PAGES"
#define BA_HISTORY_LEN_PAG_CN003        10
#define BA_HISTORY_COL_VOL_CN003C       "VOLUMES"
#define BA_HISTORY_LEN_VOL_CN003        10
#define BA_HISTORY_COL_RCO_CN003C       "RC"
#define BA_HISTORY_LEN_RCO_CN003        10
#define BA_HISTORY_COL_RTX_CN003C       "ERROR"
#define BA_HISTORY_LEN_RTX_CN003        40

#define BA_HISTORY_COLSEP_CN003C        "|"

#define BA_HISTORY_KEY_ALL_CN003C        "*"
#define BA_HISTORY_LBL_ALL_CN003C        "*"
#define BA_HISTORY_ACT_ALL_CN003C        "*"

#define BA_HISTORY_LBLDAT_CN003C        "DAT"
#define BA_HISTORY_LBLLOG_CN003C        "LOG"
#define BA_HISTORY_LBLPAG_CN003C        "PAG"

#define BA_HISTORY_ACTSAV_CN003C        "SAVE"
#define BA_HISTORY_ACTRES_CN003C        "RESTORE"
#define BA_HISTORY_ACTHIS_CN003C        "HISTLOST"

#define BA_HISTORY_RES_ALL_CN003C        "*"
#define BA_HISTORY_RES_LST_CN003C        "LAST"
#define BA_HISTORY_RES_CNT_CN003C        "CONTINUE"

#define BA_HISTORY_UNT_ALL_CN003C        "*"

#define BA_HISTORY_NOLOG_CN003C          (0x7FFFFFFF)

// strings for output of the type of an EBID
#define BA_EBIDS_TYPE_DAT_REC_CN003C    "DATA  RECOVERY "
#define BA_EBIDS_TYPE_DAT_MIG_CN003C    "DATA  MIGRATION"
#define BA_EBIDS_TYPE_PAG_REC_CN003C    "PAGES RECOVERY "
#define BA_EBIDS_TYPE_PAG_MIG_CN003C    "PAGES MIGRATION"
#define BA_EBIDS_TYPE_LOG_REC_CN003C    "LOG   RECOVERY "


typedef struct tcn003_BackupHistoryLine{
  char szKEY             [BA_HISTORY_LEN_KEY_CN003 + 1];
  char szLABEL           [BA_HISTORY_LEN_LBL_CN003 + 1];
  char szACTION          [BA_HISTORY_LEN_ACT_CN003 + 1];
  char szDBSTAMP1        [BA_HISTORY_LEN_ST1_CN003 + 1];
  char szDBSTAMP2        [BA_HISTORY_LEN_ST2_CN003 + 1];
  char szSTARTDATE       [BA_HISTORY_LEN_STA_CN003 + 1];
  char szSTOPDATE        [BA_HISTORY_LEN_STO_CN003 + 1];
  char szSTARTLOGPAGE    [BA_HISTORY_LEN_FLO_CN003 + 1];
  char szSTOPLOGPAGE     [BA_HISTORY_LEN_LLO_CN003 + 1];
  char szLOGREQUIRED     [BA_HISTORY_LEN_LOG_CN003 + 1];
  char szMEDIANAME       [BA_HISTORY_LEN_MED_CN003 + 1];
  char szPAGESTRANSFERED [BA_HISTORY_LEN_PAG_CN003 + 1];
  char szVOLUMES         [BA_HISTORY_LEN_VOL_CN003 + 1];
  char szRETURNCODE      [BA_HISTORY_LEN_RCO_CN003 + 1];
  char szRETURNTEXT      [BA_HISTORY_LEN_RTX_CN003 + 1];
} tcn003_BackupHistoryLine;

#define BA_HISTMED_LEN_KEY_CN003        BA_HISTORY_LEN_KEY_CN003
#define BA_HISTMED_LEN_LBL_CN003        BA_HISTORY_LEN_LBL_CN003
#define BA_HISTMED_LEN_MED_CN003        BA_HISTORY_LEN_MED_CN003
#define BA_HISTMED_LEN_TYP_CN003        10
#define BA_HISTMED_LEN_OVR_CN003        3
#define BA_HISTMED_LEN_SIZ_CN003        10
#define BA_HISTMED_LEN_BSZ_CN003        10
#define BA_HISTMED_LEN_KND_CN003        1
#define BA_HISTMED_LEN_LOC_CN003        64

typedef struct tcn003_BackupHistoryMediumLine{
  char szKEY[BA_HISTMED_LEN_KEY_CN003 + 1];
  char szLABEL[BA_HISTMED_LEN_LBL_CN003 + 1];
  char szMEDIANAME[BA_HISTMED_LEN_MED_CN003 + 1];
  char szTYPE[BA_HISTMED_LEN_TYP_CN003 + 1];
  char szOVERWRITE[BA_HISTMED_LEN_OVR_CN003 + 1];
  char szSIZE[BA_HISTMED_LEN_SIZ_CN003 + 1];
  char szBLOCKSIZE[BA_HISTMED_LEN_BSZ_CN003 + 1];
  char szKIND[BA_HISTMED_LEN_KND_CN003 + 1];
  char szLOCATION[BA_HISTMED_LEN_LOC_CN003 + 1];
} tcn003_BackupHistoryMediumLine;

struct tcn003_BackupHistoryEBIDLine
{
    char   EmptyString;     // place szKey, szLabel, ... can point to
    char * CompleteMemory;  // pointer for handling of allocated memory

    char * szKey;           // pointers to parts of (*CompleteMemory) or
    char * szLabel;         // EmptyString, if 0==CompleteMemory
    char * szEBID;
                            // extra information needed only for restore without history
    char * szType;          // type of the save like ACTION in dbm.knl
    char * szDateTime;      // date and time in a format like STAMP1 or STAMP2 in dbm.knl

    char * BackupToolString;// used backup tool as contained in the new dbm.ebf format
    char * DBMReturnCode;   // result of a backup as contained in the new dbm.ebf format

    char * MediumName;      //medium information used for backups not mentioned in the Backup Medium History
    char * MediumType;
    char * MediumOverwriteFlag;
    char * MediumSize;
    char * MediumBlockSize;
    char * MediumKind;
    char * MediumLocation;

    int    PartNumber;
};

enum tcn003_BackupHistoryEBIDStatus //describing possible state of an EBID
{
    EBIDStatusUnknown,
    EBIDStatusNone,
    EBIDStatusAvailable,
    EBIDStatusUnavailable
};

//strings for output of a tcn003_BackupHistoryEBIDStatus
#define BA_HISTORY_EBIDSTATUS_UNKNOWN_CN003C     "UNKNOWN"
#define BA_HISTORY_EBIDSTATUS_NONE_CN003C        "NONE"
#define BA_HISTORY_EBIDSTATUS_AVAILABLE_CN003C   "AVAILABLE"
#define BA_HISTORY_EBIDSTATUS_UNAVAILABLE_CN003C "UNAVAILABLE"

struct tcn003_BackupHistoryEBID
{
    tcn003_BackupHistoryEBIDStatus   Status;
    tcn003_BackupHistoryEBIDLine     Line;
    tcn003_BackupHistoryEBID       * pNext;
    tcn003_BackupHistoryEBID       * pPrev;
};

struct tcn003_BackupHistoryEBIDList
{
    tcn003_BackupHistoryEBID * pFirst;
    tcn003_BackupHistoryEBID * pLast;

    tcn003_BackupHistoryEBID * pCurrent;
};

typedef struct tcn003_BackupHistoryMedium{
  tcn003_BackupHistoryMediumLine       tLine;

  struct tcn003_BackupHistoryMedium  * pNext;       /* pointer to next history medium      */
  struct tcn003_BackupHistoryMedium  * pPrev;       /* pointer to previous history medium  */
} tcn003_BackupHistoryMedium;

typedef struct tcn003_BackupHistoryMedia{
  tcn003_BackupHistoryMedium         * pFirst;      /* pointer to first history medium     */
  tcn003_BackupHistoryMedium         * pLast;       /* pointer to last history medium      */
} tcn003_BackupHistoryMedia;

typedef struct tcn003_BackupHistoryEntry{
  tcn003_BackupHistoryLine             tLine;

  char                                 szStamp[20];
  struct tcn003_BackupHistoryEntry   * pNext;       /* pointer to next history entry      */
  struct tcn003_BackupHistoryEntry   * pPrev;       /* pointer to previous history entry  */
} tcn003_BackupHistoryEntry;

typedef struct tcn003_BackupHistory{
  tcn003_BackupHistoryEntry          * pFirst;      /* pointer to first history entry      */
  tcn003_BackupHistoryEntry          * pLast;       /* pointer to last history entry       */

  tcn003_BackupHistoryEntry          * pCurrent;    /* pointer to current history entry    */

  char                                 szCols      [255];
  char                                 szKey       [255];
  char                                 szLabel     [255];
  char                                 szAction    [255];
  char                                 szRestore   [255];
  char                                 szUntil     [255];

  bool                                 bForRestore;
  bool                                 bRcOKOnly;
  bool                                 bWithMedia;
  bool                                 bWithEBIDs;

  int                                  nUsedLogPage;
  int                                  nFirstLogPage;
  bool                                 bRestartable;
} tcn003_BackupHistory;

/* backup history session structur */
typedef struct tcn003_BackupHistorySession{
  tcn003_BackupHistory          tHistory;
  tcn003_BackupHistoryMedia     tMedia;
  tcn003_BackupHistoryEBIDList  tEBIDs;
} tcn003_BackupHistorySesion;

#endif /* _GCN003_H_ */
