/* @lastChanged: "1998-06-08  9:33"
 * @filename:    vcn20.cpp
 * @purpose:     "param management"
 * @release:     7.2. / 6.2.
 * @see:         "-.-"
 *
 * @copyright:   "Copyright by SAP AG, 1998"
 *
 * ==================================================================
 *
 * responsible:   Bernd Vorsprach (D025588 - Bernd.Vorsprach@SAP-AG.de)
 *
 * created:       1998-01-20 by Bernd Vorsprach
 *
 * purpose:       basic module for parameter management ("XParam")
 *
 * ==================================================================


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 */

/* ------------------------------------------------------------------
 * IMPORT                                                                    
 * ------------------------------------------------------------------
 */
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>

#include "heo02.h"
#include "heo06.h"
#include "heo11.h"
#include "hsp100.h"

#include "hcn20.h"
#include "hcn23.h"
#include "hcn29.h"
#include "hcn50.h"
#include "hcn51.h"
#include "hcn52.h"
#include "hcn80.h"
#include "hcn90.h"

/* ------------------------------------------------------------------
 * specification private macros
 * ------------------------------------------------------------------
 */
#ifndef XP_PARAM_CHANGED
  #define XP_PARAM_CHANGED "__PARAM_CHANGED___"
#endif

#define PARAM_OPT_NOCLOSE      "NOCLOSE"

/* ------------------------------------------------------------------
 * specification private function
 * ------------------------------------------------------------------
 */
static tcn00_Error cn20_ParamCommitSession
      ( tsp00_DbNamec       szDbName,
        tcn002_XParamData * pXParamData,
        char              * replyData,
        int               * replyLen,
        int                 replyLenMax);

/* ------------------------------------------------------------------
 * PUBLIC FUNCTION cn20FreeXParamSection
 * ------------------------------------------------------------------
 * purpose: frees the XParamSection from the VControlDataT structure
 *
 */
void cn20FreeXParamSection 
      ( VControlDataT * vcontrol )
{

  if (vcontrol != NULL) {
    tcn002_XParamData * pXParamData = STATIC_CAST(tcn002_XParamData *, vcontrol->pXParamSection);

    cn29FreeXParamSection(pXParamData);

    vcontrol->pXParamSection = NULL;
  } /* end if */

} /* end cn20FreeXParamSection */

/* INTERN FUNCTION cn20XParamGetValue
 * ------------------------------------------------------------------
 *
 */
tcn00_Error cn20XParamGetValue
      ( const tsp00_DbNamec      & DBNameCIn,
        const tsp00_XpKeyTypec   & XpKeyCIn,
        tsp00_C64c               & XpValueC )
{
  tcn00_Error        nFuncReturn = OK_CN00;
  tsp01_RteError     RteError ;
  tsp11_XpReturnCode xp_ret;
  tsp00_XpValueRec   xp_value;
  tsp00_Int4         XparamId ;
  tsp00_DbNamec      DBNameC = DBNameCIn;
  tsp00_XpKeyTypec   XpKeyC  = XpKeyCIn;


  XpValueC[0] = 0 ;

  xp_ret = sqlXparamOpen (DBNameC, &XparamId , XP_OPEN_READONLY_SP11, &RteError) ;

  if ( xp_ret == xp_ok_sp11 ) {
    xp_ret = sqlXparamGet (XparamId, XpKeyC, &xp_value , &RteError );


    if ( xp_ret == xp_ok_sp11 ) {
      cn29ParamValueToString(XpValueC, &xp_value);
    } // end if

    sqlXparamClose ( XparamId, XP_CLOSE_NO_FLUSH_SP11, &RteError );
  } // end if

  if (xp_ret == xp_key_not_found_sp11) {
    nFuncReturn = ERR_XPNOTFOUND_CN00;
  } else if (xp_ret != xp_ok_sp11) {
    nFuncReturn = ERR_RTEEXT_CN00;
  } // end if

  return nFuncReturn;

} /* end c20XParamGetValue */

/* INTERN FUNCTION cn20XParamGetValue
 * ------------------------------------------------------------------
 *
 */
tcn00_Error cn20XParamGetValue
      ( const tsp00_DbNamec      & DBNameC,
        const tsp00_XpKeyTypec   & XpKeyC,
        tsp00_Int4               & XpNum )
{
  tcn00_Error        nFuncReturn = OK_CN00;
  tsp00_C64c         XpValueC;

  XpNum = 0;

  nFuncReturn = cn20XParamGetValue ( DBNameC, XpKeyC, XpValueC);
  if (nFuncReturn == OK_CN00) {
    XpNum = atol(XpValueC);
  } // end if

  return nFuncReturn;

} /* end c20XParamGetValue */

/* INTERN FUNCTION cn20XParamPutValue
 * ------------------------------------------------------------------
 *
 */
tcn00_Error cn20XParamPutValue
      ( const tsp00_DbNamec      & DBNameCIn,
        const tsp00_XpKeyTypec   & XpKeyCIn,
        const tsp00_C64c         & XpValueC )
{
  tcn00_Error        nFuncReturn = OK_CN00;
  tsp01_RteError     RteError ;
  tsp11_XpReturnCode xp_ret;
  tsp00_XpValueRec   xp_value;
  tsp00_Int4         XparamId ;
  tsp00_DbNamec      DBNameC = DBNameCIn;
  tsp00_XpKeyTypec   XpKeyC  = XpKeyCIn;

  xp_ret = sqlXparamOpen (DBNameC, &XparamId , XP_OPEN_READ_WRITE_SP11, &RteError) ;

  if ( xp_ret == xp_ok_sp11 ) {
    xp_ret = sqlXparamGet (XparamId, XpKeyC, &xp_value , &RteError );

    if ( xp_ret == xp_ok_sp11 ) {
      cn29StringToParamValue(&xp_value, XpValueC);
      xp_ret = sqlXparamPut (XparamId, XpKeyC, &xp_value , &RteError );
    } // end if

    sqlXparamClose ( XparamId, XP_CLOSE_FLUSH_SP11, &RteError );
  } // end if

  if (xp_ret == xp_key_not_found_sp11) {
    nFuncReturn = ERR_XPNOTFOUND_CN00;
  } else if (xp_ret != xp_ok_sp11) {
    nFuncReturn = ERR_RTEEXT_CN00;
  } // end if

  return nFuncReturn;

} /* end cn20XParamPutValue */

/* INTERN FUNCTION cn20XParamPutValue
 * ------------------------------------------------------------------
 *
 */
tcn00_Error cn20XParamPutValue
      ( const tsp00_DbNamec      & DBNameC,
        const tsp00_XpKeyTypec   & XpKeyC,
        const tsp00_Int4         & XpNum )
{
  tsp00_C64c         XpValueC;

  _stprintf(XpValueC, "%d", XpNum);

  return cn20XParamGetValue ( DBNameC, XpKeyC, XpValueC);

} /* end cn20XParamPutValue */

/* INTERN FUNCTION cn20XParamPutValue
 * ------------------------------------------------------------------
 *
 */
tcn00_Error cn20XParamPutValue
      ( const tsp00_DbNamec      & DBNameCIn,
        const tsp00_XpKeyTypec   & XpKeyCIn,
        tsp00_XpValueRec         * xp_value)
{
  tcn00_Error        nFuncReturn = OK_CN00;
  tsp01_RteError     RteError ;
  tsp11_XpReturnCode xp_ret;
  tsp00_Int4         XparamId ;
  tsp00_DbNamec      DBNameC = DBNameCIn;
  tsp00_XpKeyTypec   XpKeyC  = XpKeyCIn;

  xp_ret = sqlXparamOpen (DBNameC, &XparamId , XP_OPEN_READ_WRITE_SP11, &RteError) ;

  if ( xp_ret == xp_ok_sp11 ) {
    xp_ret = sqlXparamPut (XparamId, XpKeyC, xp_value , &RteError );

    sqlXparamClose ( XparamId, XP_CLOSE_FLUSH_SP11, &RteError );
  } // end if

  if (xp_ret == xp_key_not_found_sp11) {
    nFuncReturn = ERR_XPNOTFOUND_CN00;
  } else if (xp_ret != xp_ok_sp11) {
    nFuncReturn = ERR_RTEEXT_CN00;
  } // end if

  return nFuncReturn;

} /* end cn20XParamPutValue */

/* INTERN FUNCTION cn20XParamGetValue
 * ------------------------------------------------------------------
 *
 */
tcn00_Error cn20XParamGetValue
      ( VControlDataT * vcontrol,
        const char    * szParameter,
        char          * szValue,
        char          * szError)
{
  tcn00_Error   nFuncReturn = OK_CN00;
  char          szReply[256];
  char        * pReply = &szReply[0];
  char        * pValue = &szValue[0];
  int           nLen   = 0;

  nFuncReturn = cn29ParamGetIntern( vcontrol->dbname, STATIC_CAST(tcn002_XParamData *, vcontrol->pXParamSection), szParameter, szReply, &nLen, XPRequestValueIntern);

  if (nFuncReturn == OK_CN00) {
    /* skip the OK line */
    while (*pReply != CHAR_STRINGTERM_CN90 && *pReply != '\n') {
      pReply++;
    } /* end while */
    if (*pReply != CHAR_STRINGTERM_CN90) {
      pReply++;
    } /* end if */
    while (*pReply != CHAR_STRINGTERM_CN90 && *pReply != '\n') {
      *pValue = *pReply;
      pValue++;
      pReply++;
    } /* end while */
    *pValue = CHAR_STRINGTERM_CN90;
  } else {
    if (szError != 0) {
      strcpy(szError, pReply);
    } /* end if */
  } /* end if */

  return nFuncReturn;

} /* end cn20XParamGetValue */

/* INTERN FUNCTION cn20XParamGetServiceValues
 * ------------------------------------------------------------------
 *
 */
tcn00_Error cn20XParamGetServiceValues
      ( tsp00_DbNamec     & szDBName,
        const tsp00_Pathc & szDbRoot)
{
  tsp00_UserNamec  szUserName;
  tsp00_Namec      szPassword;
  
   return cn20XParamGetServiceValues ( szDBName,
                                       szDbRoot,
                                       szUserName,
                                       szPassword );

} // end cn20XParamGetServiceValues

/* INTERN FUNCTION cn20XParamGetServiceValues
 * ------------------------------------------------------------------
 *
 */
tcn00_Error cn20XParamGetServiceValues
      ( tsp00_DbNamec     & szDBName,
        const tsp00_Pathc & szDbRoot,
        tsp00_UserNamec   & szUserName,
        tsp00_Namec       & szPassword )
{
  tcn00_Error         nFuncReturn = OK_CN00;

  tsp01_RteError      RteError;
  tsp11_XpReturnCode  XpReturn;
  tsp00_Int4          hParam = 0;
  tsp00_XpValueRec    XParamValue;
  tsp00_C64c          szTemp;

//  if (!sqlXParamGetServiceParamFilename ( ServiceParamFilename, &RteError ) ) {
//    nFuncReturn = ERR_XPNOTFOUND_CN00;
//  } // end if

  cn90ServiceDB(szDBName, szDbRoot);
  if (szDBName.length() == 0) {
    nFuncReturn = ERR_XPNOTFOUND_CN00;
  } // end if
  
  if (nFuncReturn == OK_CN00) {
//    _tcscpy(szDBName, ServiceParamFilename);
    XpReturn = sqlXparamOpen   ( szDBName, &hParam, XP_OPEN_READONLY_SP11, &RteError);
    if (XpReturn != xp_ok_sp11) {
      nFuncReturn = ERR_RTEEXT_CN00;
    } // end if
  } // end if

  if (nFuncReturn == OK_CN00) {
    tsp00_XpKeyTypec XpKey;

    XpKey.rawAssign(PAN_CONTROLID);

    XpReturn = sqlXparamGet   ( hParam, XpKey, &XParamValue, &RteError);
    if (XpReturn != xp_ok_sp11) {
      nFuncReturn = ERR_XPNOTFOUND_CN00;
    } else {
      cn29ParamValueToString(szTemp, &XParamValue);
      _tcscpy(szUserName, szTemp);
    } // end if
    XpKey.rawAssign(PAN_CONTROLPW);
    XpReturn = sqlXparamGet   ( hParam, XpKey, &XParamValue, &RteError);
    if (XpReturn != xp_ok_sp11) {
      nFuncReturn = ERR_XPNOTFOUND_CN00;
    } else {
      cn90CalculateUncrypt(XParamValue.xp_crypt(), szTemp, true);
      _tcscpy(szPassword, szTemp);
    } // end if

    sqlXparamClose  ( hParam, false, &RteError);
  } // end if

  return nFuncReturn;

} /* end cn20XParamGetServiceValues */


/* PUBLIC FUNCTION cn20ParamDirectGet
 * ------------------------------------------------------------------
 * purpose: get named parameter with value from param file
 *
 */
tcn00_Error cn20ParamDirectGet
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax )
{
  char szParameter[20];

  cn90GetToken(command->args, szParameter, 1);
  cn90StrUpperCopy(szParameter, szParameter, false);
  return cn29ParamGetIntern( vcontrol->dbname,STATIC_CAST(tcn002_XParamData *, vcontrol->pXParamSection), szParameter, replyData, replyLen, XPRequestDirectGet);
} /* end cn20ParamDirectGet */

/* PUBLIC FUNCTION cn20ParamDirectGetAll
 * ------------------------------------------------------------------
 * purpose: get all parameters with values from param file
 *
 */
tcn00_Error cn20ParamDirectGetAll
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax )
{
  return cn29ParamGetAllIntern( vcontrol->dbname, STATIC_CAST(tcn002_XParamData *, vcontrol->pXParamSection), command->args, replyData, replyLen, replyLenMax, false, false, vcontrol->szLastParameter);
} /* end cn20ParamDirectGetAll */

/* PUBLIC FUNCTION cn20ParamDirectGetAllNext
 * ------------------------------------------------------------------
 * purpose: get all parameters with values from param file
 *
 */
tcn00_Error cn20ParamDirectGetAllNext
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax )
{
  return cn29ParamGetAllIntern( vcontrol->dbname, STATIC_CAST(tcn002_XParamData *, vcontrol->pXParamSection), command->args, replyData, replyLen, replyLenMax, false, true, vcontrol->szLastParameter);
} /* end cn20ParamDirectGetAllNext */

/* PUBLIC FUNCTION cn20ParamDirectPut
 * ------------------------------------------------------------------
 * purpose: write one named paramter with value to param file
 *
 */
tcn00_Error cn20ParamDirectPut
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  _TCHAR szParameter [XP_STRING_LENGTH_MXCN002 +1 ];
  _TCHAR szValue     [XP_STRING_LENGTH_MXCN002 +1 ];

  cn90GetToken(command->args, szParameter, 1);
  cn90StrUpperCopy(szParameter, szParameter, false);
  cn90GetToken(command->args, szValue,     2);

  return cn20ParamDirectPut( vcontrol,
                             szParameter,
                             szValue,
                             replyData,
                             replyLen,
                             replyLenMax);

} /* end cn20ParamDirectPut */

/* PUBLIC FUNCTION cn20ParamDirectPut
 * ------------------------------------------------------------------
 * purpose: write one named paramter with value to param file
 *
 */
tcn00_Error cn20ParamDirectPut
      ( VControlDataT * vcontrol,
        _TCHAR        * szParameter,
        _TCHAR        * szValue,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  tcn002_XParamData * pXParamData = (tcn002_XParamData *) vcontrol->pXParamSection;
  tcn002_XParameter * pParameter  = NULL;
  tcn002_XParameter * pDynHelp    = NULL;
  tcn002_XParameter * pTemp       = NULL;
  bool                bInit  = false;
  tsp00_XpKeyType     xpKey;
  tsp_errtext         szError;
  tsp00_XpReturnCode  nXPReturn;
  tsp00_Int4          hParamFile;
  bool                bMallocOK = false;
  bool                bParamOK  = false;
  tsp00_DbName        dbNameP;

  /* init param management for commandline use or tcontrol or rcontrol */
  if (pXParamData == NULL) {
    nFuncReturn = cn29ParamStartSession( vcontrol->dbname, pXParamData, replyData, replyLen);
    bInit = true;
  } else {
    nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_XPSESSION_CN00);
  } /* end if */

  if (nFuncReturn == OK_CN00) {

    /* search parameter in List */
    pParameter = cn29SearchParameter(pXParamData->pFirst, szParameter);

    if (pParameter == NULL) {
      pDynHelp = cn29SearchDynamic(pXParamData->pFirst, szParameter);

      /* is there an dynamic match */
      if (pDynHelp != NULL) {
        cn00MyMalloc (sizeof (tcn002_XParameter), &pParameter, &bMallocOK);
        if ((pParameter == NULL) || (!bMallocOK)) {
          nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_MEM_CN00);
        } else {
          /* add the parameter in the list of parameters in the XParam section */
          if (pXParamData->pFirst == NULL) {
            pXParamData->pFirst = pParameter;
          } /* end if */
          if (pXParamData->pLast != NULL) {
            pXParamData->pLast->pNext = pParameter;
          } /* end if */

          /* copy data from dynamic record */
          nFuncReturn = cn29CopyParameter(pParameter, pDynHelp);
          cn90AnswerIError(replyData, replyLen, nFuncReturn);
          /* clear the dynamic/intern property */
          pParameter->xpProperties[XPPropClassDynamic] = XPPropValueNo;
          pParameter->xpProperties[XPPropClassIntern] = XPPropValueNo;
          /* set the real name */
          cn90SaveStrcpy(pParameter->szParameterName, szParameter, sizeof(tsp00_XpKeyTypec));

          /* list management */
          pParameter->pNext = NULL;
          pParameter->pPrev = pXParamData->pLast;
          pXParamData->pLast = pParameter;
        } /* end if */
      } /* end if */
    } /* end if */

    /* write the parameter */
    if (pParameter != NULL) {

      nFuncReturn = cn29CheckRange(pParameter, szValue);
      cn90AnswerIError(replyData, replyLen, nFuncReturn);

      if (nFuncReturn == OK_CN00) {

        /* open param file */
        sqlxpopenxparam ( dbNameP.c2p(vcontrol->dbname), 0, &hParamFile , szError , &bParamOK );

        if (bParamOK) {
          /* write parameter */
          cn90StrUpperCopy (pParameter->szInput, szValue,
                            (bool   ) (pParameter->xpProperties[XPPropClassCasesensitive] == XPPropValueYes));
          strcpy(pParameter->szCalc, pParameter->szInput);
          pParameter->nCalcState = XPCalcStateUnknown; 
          strcpy(pParameter->szConfirmed, "");

          cn90StringCtoP(xpKey, szParameter, sizeof(tsp00_XpKeyType));
          sqlputxparam ( hParamFile,
                         xpKey,
                         cn29StringToParamValue(&(pParameter->xpValueRecord), pParameter->szInput),
                         szError , &nXPReturn );
          cn29WriteParamHistory(vcontrol->dbname, pParameter, PAHI_CHANGE_CN002);

          if (nXPReturn != xp_ok) {
            /* don't flush */
            cn90AnswerRTEError (replyData, replyLen, ERR_RTE_CN00, szError, szError.length(), nXPReturn );
            sqlxpclosexparam ( hParamFile , false, szError ,  &bParamOK );
          } else {
            /* flush */
            sqlxpclosexparam ( hParamFile , true, szError ,  &bParamOK );
            if (!bParamOK) {
              cn90AnswerRTEError (replyData, replyLen, ERR_RTE_CN00, szError, szError.length(), nXPReturn );
            } else {
              if (!pParameter->bParamExist) {
                /* it's now a param paramemter */
                pParameter->bParamExist = true;
                pTemp = pXParamData->pParam;
                if (pTemp != NULL) {
                  while (pTemp->pParam != NULL) {
                    pTemp = pTemp->pParam;
                  } /* end while */
                  pTemp->pParam = pParameter;
                } else {
                  pXParamData->pParam = pParameter;
                } /* end if */
              } /* end if */
            } /* end if */
          } /* end if */
        } else {
          nFuncReturn = cn90AnswerRTEError (replyData, replyLen, ERR_RTE_CN00, szError, szError.length(), 0 );
        } /* end if */
      } /* end if */

    } else {
      nFuncReturn = cn29AnswerParamNotFound(replyData, replyLen, szParameter, pXParamData->pFirst);
    } /* end if */

    if (nFuncReturn == OK_CN00) {
      cn90AnswerOK(replyData, replyLen, NULL);
    } /* end if */
  } /* end if */

  /* if init then stop param management */
  if (bInit) {
    cn29FreeXParamSection(pXParamData);
  } /* end if */

  *replyLen = (int)strlen(replyData);

  return nFuncReturn;
} /* end cn20ParamDirectPut */

/* PUBLIC FUNCTION cn20ParamDirectDel
 * ------------------------------------------------------------------
 * purpose: write one named paramter with value to param file
 *
 */
tcn00_Error cn20ParamDirectDel
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  _TCHAR szParameter [XP_STRING_LENGTH_MXCN002 +1 ];

  cn90GetToken(command->args, szParameter, 1);
  cn90StrUpperCopy(szParameter, szParameter, false);

  return cn20ParamDirectDel( vcontrol,
                             szParameter,
                             replyData,
                             replyLen,
                             replyLenMax);

} /* end cn20ParamDirectDel */

/* PUBLIC FUNCTION cn20ParamDirectDel
 * ------------------------------------------------------------------
 * purpose: delete one named paramter from the param file
 */
tcn00_Error cn20ParamDirectDel
      ( VControlDataT * vcontrol,
        _TCHAR        * szParameter,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  tcn002_XParamData * pXParamData = (tcn002_XParamData *) vcontrol->pXParamSection;
  tcn002_XParameter * pParameter  = NULL;
  tcn002_XParameter * pDynHelp    = NULL;
  tcn002_XParameter * pTemp       = NULL;
  tcn002_XParameter * pPrev       = NULL;
  bool                bInit  = false;
  tsp00_XpKeyType     xpKey;
  tsp_errtext         szError;
  tsp00_XpReturnCode  nXPReturn;
  tsp00_Int4          hParamFile;
  bool                bMallocOK = false;
  bool                bParamOK  = false;
  tsp00_DbName        dbNameP;
  
  /* init param management for commandline use or tcontrol or rcontrol */
  if (pXParamData == NULL) {
    nFuncReturn = cn29ParamStartSession( vcontrol->dbname, pXParamData, replyData, replyLen);
    bInit = true;
  } else {
    nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_XPSESSION_CN00);
  } /* end if */

  if (nFuncReturn == OK_CN00) {

    /* search parameter in List */
    pParameter = cn29SearchParameter(pXParamData->pParam, szParameter);

    /* delete the parameter */
    if (pParameter != NULL) {

      /* open param file */
      sqlxpopenxparam ( dbNameP.c2p(vcontrol->dbname), 0, &hParamFile , szError ,  &bParamOK );

      if (bParamOK) {
        /* delete parameter */
        cn90StringCtoP(xpKey, szParameter, sizeof(tsp00_XpKeyType));
        sqldelxparam ( hParamFile,
                       xpKey,
                       szError , &nXPReturn );
        cn29WriteParamHistory(vcontrol->dbname, pParameter, PAHI_DELETE_CN002);

        if (nXPReturn != xp_ok) {
          /* don't flush */
          cn90AnswerRTEError (replyData, replyLen, ERR_RTE_CN00, szError, szError.length(), nXPReturn );
          sqlxpclosexparam ( hParamFile , false, szError ,  &bParamOK );
        } else {
          /* flush */
          sqlxpclosexparam ( hParamFile , true, szError ,  &bParamOK );
          if (!bParamOK) {
            cn90AnswerRTEError (replyData, replyLen, ERR_RTE_CN00, szError, szError.length(), nXPReturn );
          } else {

            /* Maintain Param List */
            if (pParameter->bParamExist) {
              /* it was a param paramemter */
              pParameter->bParamExist = false;
              pTemp = pXParamData->pParam;
              pPrev = NULL;
              while (pTemp != pParameter && pTemp != NULL) {
                pPrev = pTemp;
                pTemp = pTemp->pParam;
              } /* end while */

              if (pTemp != NULL) {
                if (pPrev != NULL) {
                  pPrev->pParam = pTemp->pParam;
                } else if (pPrev == NULL && pTemp != NULL) {
                  pXParamData->pParam = pTemp->pParam;                                      
                } /* end if */
                pTemp->pParam = NULL;
              } /* end if */

            } /* end if */

            /* Maintain help List */
            if (!pParameter->bHelpExist) {
              if (pParameter->pPrev != NULL) {
                pParameter->pPrev->pNext = pParameter->pNext;
              } else {
                pXParamData->pFirst = pParameter->pNext;
              } /* end if */
              if (pParameter->pNext != NULL) {
                pParameter->pNext->pPrev = pParameter->pPrev;
              } else {
                pXParamData->pLast = pParameter->pPrev;
              } /* end if */
              cn29DeleteParameter(pParameter);
            } else {
              pParameter->szCalc[0] = CHAR_STRINGTERM_CN90;
              pParameter->szConfirmed[0] = CHAR_STRINGTERM_CN90;
              cn90SaveStrcpy(pParameter->szInput, pParameter->szDefault, XP_STRING_LENGTH_MXCN002 + 1);
            } /* end if */

          } /* end if */
        } /* end if */
      } else {
        nFuncReturn = cn90AnswerRTEError (replyData, replyLen, ERR_RTE_CN00, szError, szError.length(), 0 );
      } /* end if */
    } else {
      nFuncReturn = cn29AnswerParamNotFound(replyData, replyLen, szParameter, pXParamData->pParam, true);
    } /* end if */

    if (nFuncReturn == OK_CN00) {
      cn90AnswerOK(replyData, replyLen, NULL);
    } /* end if */
  } /* end if */

  /* if init then stop param management */
  if (bInit) {
    cn29FreeXParamSection(pXParamData);
  } /* end if */

  *replyLen = (int)strlen(replyData);

  return nFuncReturn;
} /* end  cn20ParamDirectDel */

/* PUBLIC FUNCTION cn20ParamStartSession
 * ------------------------------------------------------------------
 * purpose: initialize a XParam-Session
 *
 */
tcn00_Error cn20ParamStartSession
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn002_XParamData * pXParam     = STATIC_CAST(tcn002_XParamData *, vcontrol->pXParamSection);
  tcn00_Error         nFuncReturn = cn29ParamStartSession(vcontrol->dbname, pXParam, replyData, replyLen);

  vcontrol->pXParamSection = STATIC_CAST(void *, pXParam);

  return nFuncReturn;
} /* end cn20ParamStartSession */

/* PUBLIC FUNCTION cn20ParamExtGetAll
 * ------------------------------------------------------------------
 * purpose: get a all parameters with extended information
 *
 */
tcn00_Error cn20ParamExtGetAll
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  return cn29ParamGetAllIntern( vcontrol->dbname, STATIC_CAST(tcn002_XParamData *, vcontrol->pXParamSection), command->args, replyData, replyLen, replyLenMax, true, false, vcontrol->szLastParameter);
} /* end cn20ParamExtGetAll */

/* PUBLIC FUNCTION cn20ParamExtGetAllNext
 * ------------------------------------------------------------------
 * purpose: get a all parameters with extended information
 *
 */
tcn00_Error cn20ParamExtGetAllNext
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  return cn29ParamGetAllIntern( vcontrol->dbname, STATIC_CAST(tcn002_XParamData *, vcontrol->pXParamSection), command->args, replyData, replyLen, replyLenMax, true, true, vcontrol->szLastParameter);
} /* end cn20ParamExtGetAllNext */

/* PUBLIC FUNCTION cn20ParamExtGet
 * ------------------------------------------------------------------
 * purpose: get a named parameter with extended information
 *
 */
tcn00_Error cn20ParamExtGet
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  char szParameter[20];

  cn90GetToken(command->args, szParameter, 1);
  cn90StrUpperCopy(szParameter, szParameter, false);
  return cn29ParamGetIntern( vcontrol->dbname, STATIC_CAST(tcn002_XParamData *, vcontrol->pXParamSection), szParameter, replyData, replyLen, XPRequestExtGet);
} /* end cn20ParamExtGet */

/* PUBLIC FUNCTION cn20ParamGetType
 * ------------------------------------------------------------------
 * purpose: get type of parameter
 */
tcn00_Error cn20ParamGetType
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax )
{
  char szParameter[20];

  cn90GetToken(command->args, szParameter, 1);
  cn90StrUpperCopy(szParameter, szParameter, false);
  return cn29ParamGetIntern( vcontrol->dbname, STATIC_CAST(tcn002_XParamData *, vcontrol->pXParamSection), szParameter, replyData, replyLen, XPRequestType);
} /* end cn20ParamGetType */

/* PUBLIC FUNCTION cn20ParamGetDefault
 * ------------------------------------------------------------------
 * purpose: get default value of parameter
 */
tcn00_Error cn20ParamGetDefault
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax )
{
  char szParameter[20];

  cn90GetToken(command->args, szParameter, 1);
  cn90StrUpperCopy(szParameter, szParameter, false);
  return cn29ParamGetIntern( vcontrol->dbname, STATIC_CAST(tcn002_XParamData *, vcontrol->pXParamSection), szParameter, replyData, replyLen, XPRequestDefault);
} /* end cn20ParamGetDefault */

/* PUBLIC FUNCTION cn20ParamGetValue
 * ------------------------------------------------------------------
 * purpose: get value of parameter
 */
tcn00_Error cn20ParamGetValue
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax )
{
  char szParameter[20];

  cn90GetToken(command->args, szParameter, 1);
  cn90StrUpperCopy(szParameter, szParameter, false);
  return cn29ParamGetIntern( vcontrol->dbname, STATIC_CAST(tcn002_XParamData *, vcontrol->pXParamSection), szParameter, replyData, replyLen, XPRequestValue);
} /* end cn20ParamGetValue */

/* PUBLIC FUNCTION cn20ParamGetProperties
 * ------------------------------------------------------------------
 * purpose: get properties of parameter
 */
tcn00_Error cn20ParamGetProperties
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax )
{
  char szParameter[20];

  cn90GetToken(command->args, szParameter, 1);
  cn90StrUpperCopy(szParameter, szParameter, false);
  return cn29ParamGetIntern( vcontrol->dbname, STATIC_CAST(tcn002_XParamData *, vcontrol->pXParamSection), szParameter, replyData, replyLen, XPRequestProperties);
} /* end cn20ParamGetProperties */

/* PUBLIC FUNCTION cn20ParamGetHelp
 * ------------------------------------------------------------------
 * purpose: get explain section of parameter
 */
tcn00_Error cn20ParamGetHelp
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax )
{
  char szParameter[20];

  cn90GetToken(command->args, szParameter, 1);
  cn90StrUpperCopy(szParameter, szParameter, false);
  return cn29ParamGetIntern( vcontrol->dbname,STATIC_CAST(tcn002_XParamData *, vcontrol->pXParamSection), szParameter, replyData, replyLen, XPRequestHelp);
} /* end cn20ParamGetHelp */

/* PUBLIC FUNCTION cn20ParamGetExplain
 * ------------------------------------------------------------------
 * purpose: get explain section of parameter
 */
tcn00_Error cn20ParamGetExplain
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax )
{
  char szParameter[20];

  cn90GetToken(command->args, szParameter, 1);
  cn90StrUpperCopy(szParameter, szParameter, false);
  return cn29ParamGetIntern( vcontrol->dbname, STATIC_CAST(tcn002_XParamData *, vcontrol->pXParamSection), szParameter, replyData, replyLen, XPRequestExplain);
} /* end cn20ParamGetExplain */

/* PUBLIC FUNCTION cn20ParamGetFull
 * ------------------------------------------------------------------
 * purpose: get all data of parameter
 */
tcn00_Error cn20ParamGetFull
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax )
{
  char szParameter[20];

  cn90GetToken(command->args, szParameter, 1);
  cn90StrUpperCopy(szParameter, szParameter, false);
  return cn29ParamGetIntern( vcontrol->dbname, STATIC_CAST(tcn002_XParamData *, vcontrol->pXParamSection), szParameter, replyData, replyLen, XPRequestFull);
} /* end cn20ParamGetFull */

/* PUBLIC FUNCTION cn20ParamPut
 * ------------------------------------------------------------------
 * purpose: save the new value internal param session structure
 *          - no persitent write to param file
 */
tcn00_Error cn20ParamPut
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  tcn002_XParamData * pXParamData = (tcn002_XParamData *) vcontrol->pXParamSection;
  tsp00_XpKeyTypec    szParameter;
  char                szValue[XP_STRING_LENGTH_MXCN002 +1 ];
  char                szOverride[PARAMETER_MAXLEN_CN90];
  bool                bOverride = false;

  /* is there an active param session */
  if (nFuncReturn == OK_CN00) {
    if (pXParamData == NULL) {
      nFuncReturn = ERR_NOXPSESSION_CN00;
    } /* end if */
  } /* end if */

  /* setting the new value */
  if (nFuncReturn == OK_CN00) {
    cn90GetToken(command->args, szParameter, 1);
    cn90StrUpperCopy(szParameter, szParameter, false);
    cn90GetToken(command->args, szValue,     2);
    if (cn90GetToken(command->args, szOverride,  3)) {
      if (_tcsicmp(szOverride, XP_PROP_OVERRIDE_CN002) == 0) {
        bOverride = true;
      } else {
        nFuncReturn = ERR_PARAM_CN00;
      } // end if
    } // end if
  } // end if

  if (nFuncReturn == OK_CN00) {
    nFuncReturn = cn29SoftPutIntern(pXParamData, szParameter, szValue, false, bOverride);
  } /* end if */

  if (nFuncReturn == OK_CN00) {
    cn90AnswerOK(replyData, replyLen, NULL);
  } else if (nFuncReturn == ERR_XPNOTFOUND_CN00) {
    nFuncReturn = cn29AnswerParamNotFound(replyData, replyLen, szParameter, pXParamData->pFirst);
  } else {
    cn90AnswerIError(replyData, replyLen, nFuncReturn);
  } /* end if */

  return nFuncReturn;
} /* end cn20ParamPut */

/* PUBLIC FUNCTION cn20_ParamCheckAll
 * ------------------------------------------------------------------
 * purpose: check of all parameters
 */
static tcn00_Error cn20_ParamCheckAll
      ( tsp00_DbNamec       szDbName,
        tcn002_XParamData * pXParamData,
        const char        * szGroup,
        char              * replyData,
        int               * replyLen,
        int                 replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  tcn002_XParameter * pParameter  = NULL;
  tcn002_XPPropValue  nGroup = XPPropValueUnknown;
  tsp00_Versionc      VersionString;
  bool                bInit = false;

  /* init param management for commandline use or tcontrol or rcontrol */
  if (pXParamData == NULL) {
    nFuncReturn = cn29ParamStartSession( szDbName, pXParamData, replyData, replyLen);
    bInit = true;
  } /* end if */

  /* look for request mode */
  if (nFuncReturn == OK_CN00) {
    if (_tcslen(szGroup) > 0) {
      if (strcmp(szGroup, XP_PROP_SUPPORT_CN002) == 0) {
        nGroup    = XPPropValueSupport;
      } else if (strcmp(szGroup, XP_PROP_EXTENDED_CN002) == 0) {
        nGroup    = XPPropValueExtended;
      } else if (strcmp(szGroup, XP_PROP_GENERAL_CN002) == 0) {
        nGroup    = XPPropValueGeneral;
      } else {
        nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
      } /* end if */
    } /* end if */
  } /* end if */

  /* calculate all parameters */
  if (nFuncReturn == OK_CN00) {

    /* KERNELVERSION   */
    pParameter = cn29SearchParameter(pXParamData->pFirst, PAN_PARAM_KERNELVERSION);
    if (_tcslen(pParameter->szDefault) > 0) {
      VersionString.rawAssign(pParameter->szDefault);
    } else {
      sp100_GetVersionString(COMP_NAME_KERNEL_SP100, s100buildnumber, VersionString);
    } // end if
    cn29SoftPutIntern( pXParamData, PAN_PARAM_KERNELVERSION, VersionString, true);

    /* init answer */
    cn90AnswerOK(replyData, replyLen, NULL);

    /* calculate all parameters */
    pParameter = pXParamData->pFirst;
    /* checking while no error */
    while ((pParameter != NULL) && (nFuncReturn == OK_CN00)) {
      nFuncReturn = cn23CalculateParameter(pXParamData, pParameter, replyData, replyLen);
      if (nFuncReturn == OK_CN00) {
        nFuncReturn = cn23GetCalcState(pParameter, replyData, nGroup);
      } /* end if */
      if (nFuncReturn == OK_CN00) {
        pParameter = pParameter->pNext;
      } /* end if */
    } /* end while */

    /* check constraints for all parameters */
    pParameter = pXParamData->pFirst;
    /* checking while no error */
    while ((pParameter != NULL) && (nFuncReturn == OK_CN00)) {
      nFuncReturn = cn23CheckConstraint(pXParamData, pParameter, replyData, replyLen);
      if (nFuncReturn == OK_CN00) {
        nFuncReturn = cn23GetCalcState(pParameter, replyData, nGroup);
      } /* end if */
      if (nFuncReturn == OK_CN00) {
        pParameter = pParameter->pNext;
      } /* end if */
    } /* end while */

  } /* end if */

  if (bInit) {
    if (nFuncReturn == OK_CN00) {
      nFuncReturn = cn20_ParamCommitSession(szDbName, pXParamData, replyData, replyLen, replyLenMax); 
      cn29FreeXParamSection(pXParamData);
    } // end if
  } // end if

  *replyLen = (int)strlen(replyData);

  return nFuncReturn;
} /* end cn20_ParamCheckAll */

/* PUBLIC FUNCTION cn20ParamCheckAll
 * ------------------------------------------------------------------
 * purpose: check of all parameters
 */
tcn00_Error cn20ParamCheckAll
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  char               szGroup[XP_STRING_LENGTH_MXCN002 +1 ];

  cn90GetToken(command->args, szGroup, 1);

  return cn20_ParamCheckAll(vcontrol->dbname,
                            STATIC_CAST(tcn002_XParamData *, vcontrol->pXParamSection),
                            szGroup,
                            replyData,
                            replyLen,
                            replyLenMax);
} // end cn20ParamCheckAll

/* PUBLIC FUNCTION cn20ParamPutConfirm
 * ------------------------------------------------------------------
 * purpose: confirm an override request
 */
tcn00_Error cn20ParamPutConfirm
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error        nFuncReturn = OK_CN00;
  tcn002_XParamData * pXParamData = (tcn002_XParamData *) vcontrol->pXParamSection;
  tcn002_XParameter * pParameter  = NULL;
  tsp00_XpKeyTypec       szParameter;
  char               szValue[XP_STRING_LENGTH_MXCN002 +1 ];

  /* is there an active param session */
  if (nFuncReturn == OK_CN00) {
    if (pXParamData == NULL) {
      nFuncReturn = ERR_NOXPSESSION_CN00;
      cn90AnswerIError(replyData, replyLen, nFuncReturn);
    } /* end if */
  } /* end if */

  /* setting the new value */
  if (nFuncReturn == OK_CN00) {

    cn90GetToken(command->args, szParameter, 1);
    cn90GetToken(command->args, szValue, 2);

    if (strlen(szParameter) > 0) {

      /* search parameter in List */
      pParameter = cn29SearchParameter(pXParamData->pFirst, szParameter);
      /* set value to parameter */
      if (pParameter != NULL) {
        switch (pParameter->nCalcState) {
          case XPCalcStateRequest:
            if (cn90Stricmp(szValue, pParameter->szInput) == 0) {
              /* user select own value */
              strcpy(pParameter->szCalc, szValue);
              strcpy(pParameter->szConfirmed, szValue);
            } else if (cn90Stricmp(szValue, pParameter->szCalc)  == 0) {
              /* user select computed value */
              strcpy(pParameter->szInput, szValue);
              strcpy(pParameter->szConfirmed, szValue);
            } else {
              nFuncReturn = ERR_XPINVCONF_CN00;
              cn90AnswerIError(replyData, replyLen, nFuncReturn);
            } /* end if */
            break;
          default:
            nFuncReturn = ERR_XPINVCONF_CN00;
            cn90AnswerIError(replyData, replyLen, nFuncReturn);
            break;
        } /* end switch */

      } else {
        nFuncReturn = ERR_XPNOTFOUND_CN00;
        cn90AnswerIError(replyData, replyLen, nFuncReturn);
      } /* end if */

    } else {
      nFuncReturn = ERR_PARAM_CN00;
      cn90AnswerIError(replyData, replyLen, nFuncReturn);
    } /* end if */

  } /* end if */

  if (nFuncReturn == OK_CN00) {
    cn90AnswerOK(replyData, replyLen, NULL);
  } /* end if */

  *replyLen = (int)strlen(replyData);

  return nFuncReturn;

} /* end cn20ParamPutConfirm */

/* PUBLIC FUNCTION cn20_ParamCommitSession
 * ------------------------------------------------------------------
 * purpose: terminate param session _WITH_ writing changes to the param file
 *
 */
tcn00_Error cn20_ParamCommitSession
      ( tsp00_DbNamec       szDbName,
        tcn002_XParamData * pXParamData,
        char              * replyData,
        int               * replyLen,
        int                 replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  tcn002_XParameter * pParameter  = NULL;
  tcn002_XParameter * pDynHelp    = NULL;
  tcn002_XParameter * pTemp       = NULL;
  tsp00_XpKeyType     xpKey;
  tsp_errtext         szError;
  tsp00_XpReturnCode  nXPReturn;
  tsp00_Int4          hParamFile;
  bool                bMallocOK = false;
  bool                bParamOK  = false;
  char                szTemp[XP_STRING_LENGTH_MXCN002 + 1];
  tsp00_DbName        dbNameP;

  /* is there an active param session */
  if (nFuncReturn == OK_CN00) {
    if (pXParamData == NULL) {
      nFuncReturn = ERR_NOXPSESSION_CN00;
      cn90AnswerIError(replyData, replyLen, nFuncReturn);
    } /* end if */
  } /* end if */

  /* open param file */
  if (nFuncReturn == OK_CN00) {
    /* open param file */
    sqlxpopenxparam ( dbNameP.c2p(szDbName), 0, &hParamFile , szError ,  &bParamOK );
    if (!bParamOK) {
      nFuncReturn = ERR_RTE_CN00;
      cn90AnswerRTEError (replyData, replyLen, nFuncReturn, szError, szError.length(), 0 );
    } /* end if */
  } /* end if */

  /* looping around the parameters */
  if (nFuncReturn == OK_CN00) {
    pParameter = pXParamData->pFirst;

    while ((pParameter != NULL) && (nFuncReturn == OK_CN00)) {
      /* could we write the parameter */
      nFuncReturn = cn23GetCalcState(pParameter, replyData, XPPropValueNo);
 
      /* ok write it, if there is a value and it is not an internal parameter */
      if ( ( nFuncReturn                                 == OK_CN00       ) &&
           ( strlen(pParameter->szInput)                  > 0             ) &&
           ( pParameter->xpProperties[XPPropClassIntern] == XPPropValueNo )    ) {

        cn90StrUpperCopy (szTemp, pParameter->szInput, 
                          (bool   ) (pParameter->xpProperties[XPPropClassCasesensitive] == XPPropValueYes));
        cn90StringCtoP(xpKey, pParameter->szParameterName, sizeof(tsp00_XpKeyType));
        sqlputxparam ( hParamFile,
                       xpKey,
                       cn29StringToParamValue(&(pParameter->xpValueRecord), szTemp),
                       szError , &nXPReturn );
        // write history
        cn29WriteParamHistory(szDbName, pParameter, PAHI_CHANGE_CN002);
        if (nXPReturn != xp_ok) {
          nFuncReturn = ERR_RTE_CN00;
          cn90AnswerRTEError (replyData, replyLen, nFuncReturn, szError, szError.length(), nXPReturn );
        } /* end if */
      } /* end if */
 
      pParameter = pParameter->pNext;
    } /* end while */
  } /* end if */

  /* close param file (always close if open successful) */
  if (nFuncReturn == OK_CN00) {
    sqlxpclosexparam ( hParamFile , true, szError ,  &bParamOK );
    if (!bParamOK) {
      nFuncReturn = ERR_RTE_CN00;
      cn90AnswerRTEError (replyData, replyLen, nFuncReturn, szError, szError.length(), 0 );
      sqlxpclosexparam ( hParamFile , false, szError ,  &bParamOK );
    } /* end if */
  } else {
    if (bParamOK) {
      sqlxpclosexparam ( hParamFile , false, szError ,  &bParamOK );
    } /* end if */
  } /* end if */

  if (nFuncReturn == OK_CN00) {
    cn90AnswerOK(replyData, replyLen, NULL);
  } /* end if */

  *replyLen = (int)strlen(replyData);

  return nFuncReturn;
} /* end cn20_ParamCommitSession */

/* PUBLIC FUNCTION cn20ParamCommitSession
 * ------------------------------------------------------------------
 * purpose: terminate param session _WITH_ writing changes to the param file
 *
 */
tcn00_Error cn20ParamCommitSession
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  char                szSwitch[XP_STRING_LENGTH_MXCN002 +1 ];
  bool                bClose = true;

  // look for noclose option
  if (cn90GetToken(command->args, szSwitch, 1)) {
    if (_tcsicmp(PARAM_OPT_NOCLOSE, szSwitch) == 0) {
      bClose = false;
    } else {
      nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
    } // end if
  } // end if

  if (nFuncReturn == OK_CN00) {
    nFuncReturn = cn20_ParamCommitSession(vcontrol->dbname,
                                         STATIC_CAST(tcn002_XParamData *, vcontrol->pXParamSection),
                                         replyData,
                                         replyLen,
                                         replyLenMax);
  } // end if

  if (nFuncReturn == OK_CN00 && bClose) {
    /* free XParam section */
    cn20FreeXParamSection (vcontrol);
  } // end if

  return nFuncReturn;
} // end cn20ParamCommitSession

/* PUBLIC FUNCTION cn20ParamAbortSession
 * ------------------------------------------------------------------
 * purpose: terminate param session _WITHOUT_ writing changes to the param file
 *
 */
tcn00_Error cn20ParamAbortSession
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  /* free the XParam section */
  cn20FreeXParamSection (vcontrol);

  /* Answer always ok */
  cn90AnswerOK(replyData, replyLen, NULL);

  return OK_CN00;
} /* end cn20ParamAbortSession */

/* PUBLIC FUNCTION cn20ParamInit
 * ------------------------------------------------------------------
 * purpose: initializes the param file for a new database
 *
 */
tcn00_Error cn20ParamInit
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  tcn002_XParamData * pXParamData = (tcn002_XParamData *) vcontrol->pXParamSection;
  tcn002_XParameter * pParameter  = NULL;
  _TCHAR              szValue[XP_STRING_LENGTH_MXCN002 + 1];
  tsp00_Versionc      VersionString;
  tsp00_C256          szWrk;
  tsp_cryptpw         cryptPwd;
  tsp_namec           szPwd;
  _TCHAR              szInstance[PARAMETER_MAXLEN_CN90];
  cn51DBMConfigValue  cfgColdUsr(vcontrol->dbname, CFG_COLDUSR_CN51);

  cn90GetToken(command->args, szInstance, 1);

  /* is there an active param session */
  if (nFuncReturn == OK_CN00) {
    if (pXParamData == NULL) {
      nFuncReturn = ERR_NOXPSESSION_CN00;
    } /* end if */
  } /* end if */

  /* is there an existing param file */
  if (nFuncReturn == OK_CN00) {
    if (pXParamData->pParam != NULL) {
      nFuncReturn = ERR_PARAMEXIST_CN00;
    } /* end if */
  } /* end if */

  // remove parameters from other instance types
  if ((nFuncReturn == OK_CN00) && (_tcslen(szInstance) > 0)) {
      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_INSTANCE_TYPE, szInstance, true);
    if (nFuncReturn == OK_CN00) {
      nFuncReturn = cn29CheckInstance(szInstance, pXParamData);
    } // end if
  } // end if

  /* assign defaults */
  if (nFuncReturn == OK_CN00) {
//    /* SERVERDB        */
//    if (nFuncReturn == OK_CN00) {
//      nFuncReturn = cn29SoftPutIntern( pXParamData, XP_ID_SERVERDB_CN00, vcontrol->dbname, true);
//    } /* end if */
    /* RUNDIRECTORY    */
    if (nFuncReturn == OK_CN00) {
      cn90GetWrkPath(szWrk);
      sprintf(szValue, "%s%s", (char *) szWrk, (char *) vcontrol->dbname);
      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_RUNDIR, szValue, true);
    } /* end if */
    /* KERNELVERSION   */
    if (nFuncReturn == OK_CN00) {
      pParameter = cn29SearchParameter(pXParamData->pFirst, PAN_PARAM_KERNELVERSION);
      if (_tcslen(pParameter->szDefault) > 0) {
        VersionString.rawAssign(pParameter->szDefault);
      } else {
        sp100_GetVersionString(COMP_NAME_KERNEL_SP100, s100buildnumber, VersionString);
      } // end if
      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_PARAM_KERNELVERSION, VersionString, true);
    } /* end if */
    /* CONTROLUSERID */
    if (nFuncReturn == OK_CN00) {
      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_CONTROLID, vcontrol->pCurrentUser->getUserName(), true);
    } /* end if */
    /* CONTROLUSERPASSWORD */
    if (nFuncReturn == OK_CN00) {
      cryptPwd = vcontrol->pCurrentUser->getMasterPwd();
      cn90CalculateUncrypt(cryptPwd, szPwd, false);
      nFuncReturn = cn29SoftPutIntern( pXParamData,  PAN_CONTROLPW, szPwd, true);
    } /* end if */
  } /* end if */

  /* generate answer */
  if (nFuncReturn == OK_CN00) {
    vcontrol->pCurrentUser->setColdUser();
    vcontrol->pCurrentUser->save();
    cfgColdUsr = vcontrol->pCurrentUser->getUserName();
    cn90AnswerOK(replyData, replyLen, NULL);
  } else {
    cn90AnswerIError(replyData, replyLen, nFuncReturn);
  } /* end if */

  return nFuncReturn;
} /* end cn20ParamInit */

/* PUBLIC FUNCTION cn20ParamCopy
 * ------------------------------------------------------------------
 * purpose: copy parameters from a source-db
 *
 */
tcn00_Error cn20ParamCopy
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  tcn002_XParamData * pXParamData = (tcn002_XParamData *) vcontrol->pXParamSection;
  tsp_dbname          szDatabase;
  tsp_errtext         szErrtext;
  bool                bStatus;
  bool                bParamFileOpen = false;
  tsp00_Int4          hParamFile;
  tsp00_XpKeyType     xpKey;
  tsp00_XpKeyType     xpNextKey;
  tsp00_XpKeyTypec    szParamC;
  tsp00_XpValueRec    xpValue;
  tsp00_XpReturnCode  nXPReturn;
  tcn002_XParameter * pParameter  = NULL;
  tcn002_XParameter * pDynHelp    = NULL;
  bool                bMallocOK   = true;
  char                szValue[XP_STRING_LENGTH_MXCN002 + 1];
  tsp00_Versionc      VersionString;
  tsp00_C256          szWrk;
  tsp_cryptpw         cryptPwd;
  tsp_namec           szPwd;
  cn51DBMConfigValue  cfgColdUsr(vcontrol->dbname, CFG_COLDUSR_CN51);

  /* is there an active param session */
  if (nFuncReturn == OK_CN00) {
    if (pXParamData == NULL) {
      nFuncReturn = ERR_NOXPSESSION_CN00;
      cn90AnswerIError(replyData, replyLen, nFuncReturn);
    } /* end if */
  } /* end if */

  /* is there an existing param file */
  if (nFuncReturn == OK_CN00) {
    if (pXParamData->pParam != NULL) {
      nFuncReturn = ERR_PARAMEXIST_CN00;
      cn90AnswerIError(replyData, replyLen, nFuncReturn);
    } /* end if */
  } /* end if */

  /* get source name */
  if (nFuncReturn == OK_CN00) {
    if (!cn90GetToken(command->args, szDatabase, 1)) {
      nFuncReturn = ERR_PARAM_CN00;
      cn90AnswerIError(replyData, replyLen, nFuncReturn);
    } /* end if */
  } /* end if */

  /* open Param file of db */
  if ((nFuncReturn == OK_CN00) && (vcontrol->dbname[0] != CHAR_STRINGTERM_CN90)) {
    sqlxpopenxparam ( szDatabase, 0, &hParamFile , szErrtext ,  &bStatus );
    if  ( !bStatus ) {
      nFuncReturn = ERR_RTE_CN00;
      cn90AnswerRTEError (replyData, replyLen, nFuncReturn, szErrtext, szErrtext.length(), xp_not_ok );
    } else {
      bParamFileOpen = true;
    } /* end if */
  } /* end if */

  /* File abarbeiten */
  if (nFuncReturn == OK_CN00) {

    memset(xpKey, ' ', sizeof (tsp00_XpKeyType));
    do {
      sqlgnextxparam(hParamFile, xpKey, xpNextKey, &xpValue, szErrtext , &nXPReturn );

      if ( nXPReturn == xp_ok ) {
        cn90StringPtoC ( szParamC, xpNextKey, sizeof(xpNextKey));

        /* search parameter in List */
        pParameter = cn29SearchParameter(pXParamData->pFirst, szParamC);

        /* No match -> look for dynamic record */
        if (pParameter == NULL) {
          /* is there a dynamic match */
          pDynHelp = cn29SearchDynamic(pXParamData->pFirst, szParamC);
          if (pDynHelp != NULL) {
            if (pDynHelp->xpProperties[XPPropClassClear] == XPPropValueNo) {

              /* only none clear-on-copy parameters */
              cn00MyMalloc (sizeof (tcn002_XParameter), &pParameter, &bMallocOK);
              if ((pParameter == NULL) || (!bMallocOK)) {
                nFuncReturn = ERR_MEM_CN00;
                cn90AnswerIError(replyData, replyLen, nFuncReturn);
              } else {
                /* add the parameter in the list of parameters in the XParam section */
                if (pXParamData->pFirst == NULL) {
                  pXParamData->pFirst = pParameter;
                } /* end if */
                if (pXParamData->pLast != NULL) {
                  pXParamData->pLast->pNext = pParameter;
                } /* end if */

                /* copy data from dynamic record */
                nFuncReturn = cn29CopyParameter(pParameter, pDynHelp);
                /* clear the dynamic/intern property */
                pParameter->xpProperties[XPPropClassDynamic] = XPPropValueNo;
                pParameter->xpProperties[XPPropClassIntern] = XPPropValueNo;

                /* list management */
                pParameter->pNext = NULL;
                pParameter->pPrev = pXParamData->pLast;
                pXParamData->pLast = pParameter;

              } /* end if */
            } /* end if */
          } /* end if */
        } else {
          /* exclude clear-on-copy parameters */
          pParameter = (pParameter->xpProperties[XPPropClassClear] == XPPropValueYes) ? NULL : pParameter;
        } /* end if */

        /* Copy Data */
        if (pParameter != NULL) {
          strcpy(pParameter->szParameterName, szParamC);
          memcpy(&pParameter->xpValueRecord, &xpValue, sizeof(tsp00_XpValueRec));
          cn29ParamValueToString(pParameter->szInput, &pParameter->xpValueRecord);
        } /* end if */

        /* and the next */
        strncpy ( xpKey , xpNextKey , sizeof(xpKey) );
      } /* end if */

    }  while ( (nXPReturn == xp_ok) && (nFuncReturn == OK_CN00));
  } /* end if */

  /* close Param file of db (close always if open, independent from nFuncReturn) */
  if (bParamFileOpen) {
    sqlxpclosexparam ( hParamFile , false, szErrtext , &bStatus );
    /* set errortext only, if there no preceding error */
    if  ( (!bStatus) && (nFuncReturn == OK_CN00) ) {
      nFuncReturn = ERR_RTE_CN00;
      cn90AnswerRTEError (replyData, replyLen, nFuncReturn, szErrtext, szErrtext.length(), xp_not_ok );
    } /* end if */
  } /* end if */

  /* assign defaults */
  if (nFuncReturn == OK_CN00) {
//    /* SERVERDB        */
//    if (nFuncReturn == OK_CN00) {
//      nFuncReturn = cn29SoftPutIntern( pXParamData, XP_ID_SERVERDB_CN00, vcontrol->dbname, true);
//    } /* end if */
    /* RUNDIRECTORY    */
    if (nFuncReturn == OK_CN00) {
      cn90GetWrkPath(szWrk);
      sprintf(szValue, "%s%s", (char *) szWrk, (char *) vcontrol->dbname);
      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_RUNDIR, szValue, true);
    } /* end if */
    /* KERNELVERSION   */
    if (nFuncReturn == OK_CN00) {
      sp100_GetVersionString(COMP_NAME_KERNEL_SP100, s100buildnumber, VersionString);
      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_PARAM_KERNELVERSION, VersionString, true);
    } /* end if */
    /* CONTROLUSERID */
    if (nFuncReturn == OK_CN00) {
      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_CONTROLID, vcontrol->pCurrentUser->getUserName(),true);
    } /* end if */
    /* CONTROLUSERPASSWORD */
    if (nFuncReturn == OK_CN00) {
      cryptPwd = vcontrol->pCurrentUser->getMasterPwd();
      cn90CalculateUncrypt(cryptPwd, szPwd, false);
      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_CONTROLPW, szPwd, true);
    } /* end if */
    if (nFuncReturn == OK_CN00) {
      cn90AnswerIError(replyData, replyLen, nFuncReturn);
    } /* end if */
  } /* end if */

  /* generate answer */
  if (nFuncReturn == OK_CN00) {
    vcontrol->pCurrentUser->setColdUser();
    vcontrol->pCurrentUser->save();
    cfgColdUsr = vcontrol->pCurrentUser->getUserName();
    cn90AnswerOK(replyData, replyLen, NULL);
  } /* end if */

  return nFuncReturn;
} /* end cn20ParamCopy */

/* PUBLIC FUNCTION cn20ParamAddDevSpace
 * ------------------------------------------------------------------
 * purpose: add devspace params
 *
 */
tcn00_Error cn20ParamAddDevSpace
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  long nReturnSize = 0;
  return cn20ParamAddDevSpace(vcontrol, command, replyData, replyLen, replyLenMax, nReturnSize);
} // end cn20ParamAddDevSpace

/* PUBLIC FUNCTION cn20ParamAddDevSpace
 * ------------------------------------------------------------------
 * purpose: add devspace params
 *
 */
tcn00_Error cn20ParamAddDevSpace
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax,
        long          & nReturnSize)
{
  tcn00_Error            nFuncReturn    = OK_CN00;
  char                   szNumber   [XP_STRING_LENGTH_MXCN002 + 1];
  char                   szMode     [XP_STRING_LENGTH_MXCN002 + 1];
  tsp00_C64c             szType;
  tsp00_C64c             szName;
  tsp00_C64c             szSize;
  tsp00_XpKeyTypec       szParameter;
  char                   szRundir   [PARAMETER_MAXLEN_CN90];
  tcn00_DevspaceKeywords aKeyTable  [] = XP_DEVSPACE_KEYWORDS_CN00;
  tcn00_DevspaceParams   aParamTable[] = XP_DEVSPACE_PARAMS_CN00;
  long                   nKey          = 0;   
  long                   nParam        = 0;   
  int                    nNumber       = 0;
  tsp_int4               nCompSize     = 0;
  long                   nSize         = 0;
  tcn00_DBState          aState;
  tsp00_C64c             szChanged;
  tsp00_XpKeyTypec       szChangeName;
  bool                   bDirty        = false;
  tsp00_Bool             bOk;
  tsp00_ErrTextc         szError;
  tcn002_XParamData    * pXParamData = NULL;
  tcn002_XParameter    * pParameter  = NULL;

  nReturnSize = 0;

  /* is there an active param session */
  if (nFuncReturn == OK_CN00) {
    if (vcontrol->pXParamSection != NULL) {
      nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_XPSESSION_CN00);
    } /* end if */
  } /* end if */

  // start a temporary param session
  if (nFuncReturn == OK_CN00) {
    nFuncReturn = cn29ParamStartSession( vcontrol->dbname, pXParamData, replyData, replyLen);
  } // end if

  // check dirty flag
  if (nFuncReturn == OK_CN00) {
    aState = cn90DBState(vcontrol->dbname);
    if (aState == STATE_ERROR_CN00) {
      nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_STATE_CN00);
    } else if (aState != STATE_OFFLINE_CN00 && aState != STATE_RUIN_CN00) {
      szChangeName.rawAssign(XP_PARAM_CHANGED);
      if (cn20XParamGetValue( vcontrol->dbname, szChangeName, szChanged) == OK_CN00) {
        if (atol(szChanged) != 0) {
          bDirty = true;
        } // end if
      } // end if
    } // end if
  } // end if

  /* get parameters and check mode */
  if (nFuncReturn == OK_CN00) {
    cn90GetToken(command->args, szNumber, 1);
    cn90GetToken(command->args, szMode, 2);
    cn90GetToken(command->args, szName, 3, szName.size());
    cn90GetToken(command->args, szType, 4, szType.size());
    cn90GetToken(command->args, szSize, 5, szSize.size());

    /* search in keywordtable */
    for (nKey = 0; (aKeyTable[nKey].nDevSpace != XPDevSpaceUnknown_ecn00) &&
                   (_tcsicmp(aKeyTable[nKey].szText, szMode) != 0  );    nKey++);

    /* search in paramtable */
    for (nParam = 0; aKeyTable[nKey].nDevSpace != aParamTable[nParam].nDevSpace; nParam++);

    if (aKeyTable[nKey].nDevSpace == XPDevSpaceUnknown_ecn00) {
      nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
    } else if (bDirty && !aKeyTable[nKey].bMirror) {
      nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_NOADDD_CN00);
    } /* end if */
  } /* end if */

  /* check arguments */
  if (nFuncReturn == OK_CN00) {
    if ((strlen(szNumber) == 0) ||
        (strlen(szType)   != 1) ||
        (strlen(szName)   == 0) ||
        ((strlen(szSize)  == 0) && (strlen(aParamTable[nParam].szSize) > 0)) ) {
      nFuncReturn = ERR_PARAM_CN00;
      cn90AnswerIError(replyData, replyLen, nFuncReturn);
    } /* end if */
  } /* end if */

  if (nFuncReturn == OK_CN00) {
    /* set number of devspace */
    nNumber = atoi(szNumber);

    /* get the rundir */
    nFuncReturn = cn20XParamGetValue(vcontrol, PAN_RUNDIR, szRundir, replyData);
  } /* end if */

  /* raw check the devspace */
  if (nFuncReturn == OK_CN00) {
    if ((szType[0] == _T('R') || szType[0] == _T('r')) && (strlen(aParamTable[nParam].szSize) > 0)) {
      SqlDevSize( szName, &nCompSize, szError, &bOk );
      if (!bOk) {
        nFuncReturn = cn90AnswerRTEError (replyData, replyLen, ERR_DEVCHECK_CN00, szError, szError.length(), 1 );
      } else if (nCompSize > 0) {
        nSize = atol(szSize);
        nSize = (nSize == 0) ? nCompSize : nSize;
        sprintf(szSize, "%ld", nSize);
      } /* end if */
    } /* end if */
  } /* end if */

  /* set name of devspace */
  if (nFuncReturn == OK_CN00) {
    sprintf(szParameter, aParamTable[nParam].szName, nNumber);
    nFuncReturn = cn29SoftPutIntern( pXParamData, szParameter, szName, true);
    if (nFuncReturn != OK_CN00) {
      cn90AnswerIError(replyData, replyLen, nFuncReturn);
    } else {
      pParameter = cn29SearchParameter(pXParamData->pFirst, szParameter);

      nFuncReturn = cn23CalculateParameter(pXParamData, pParameter, replyData, replyLen);
      if (nFuncReturn == OK_CN00) {
        nFuncReturn = cn23GetCalcState(pParameter, replyData, XPPropValueGeneral);
      } /* end if */
      if (nFuncReturn == OK_CN00) {
        nFuncReturn = cn23CheckConstraint(pXParamData, pParameter, replyData, replyLen);
      } /* end if */
      if (nFuncReturn == OK_CN00) {
        nFuncReturn = cn23GetCalcState(pParameter, replyData, XPPropValueGeneral);
      } /* end if */

      if (nFuncReturn == OK_CN00) {
         cn29StringToParamValue(&(pParameter->xpValueRecord), szName);
        nFuncReturn = cn20XParamPutValue(vcontrol->dbname, szParameter, &(pParameter->xpValueRecord));
        if (nFuncReturn != OK_CN00) {
          cn90AnswerIError(replyData, replyLen, nFuncReturn);
        } // end if
      } // end if
    } // end if
  } /* end if */
  /* set type of devspace */
  if (nFuncReturn == OK_CN00) {
    sprintf(szParameter, aParamTable[nParam].szType, nNumber);
    nFuncReturn = cn29SoftPutIntern( pXParamData, szParameter, szType, true);
    if (nFuncReturn != OK_CN00) {
      cn90AnswerIError(replyData, replyLen, nFuncReturn);
    } else {
      pParameter = cn29SearchParameter(pXParamData->pFirst, szParameter);

      nFuncReturn = cn23CalculateParameter(pXParamData, pParameter, replyData, replyLen);
      if (nFuncReturn == OK_CN00) {
        nFuncReturn = cn23GetCalcState(pParameter, replyData, XPPropValueGeneral);
      } /* end if */
      if (nFuncReturn == OK_CN00) {
        nFuncReturn = cn23CheckConstraint(pXParamData, pParameter, replyData, replyLen);
      } /* end if */
      if (nFuncReturn == OK_CN00) {
        nFuncReturn = cn23GetCalcState(pParameter, replyData, XPPropValueGeneral);
      } /* end if */

      if (nFuncReturn == OK_CN00) {
         cn29StringToParamValue(&(pParameter->xpValueRecord), szType);
        nFuncReturn = cn20XParamPutValue(vcontrol->dbname, szParameter, &(pParameter->xpValueRecord));
        if (nFuncReturn != OK_CN00) {
          cn90AnswerIError(replyData, replyLen, nFuncReturn);
        } // end if
      } // end if
    } // end if
  } /* end if */
  /* set size of devspace */
  if ((nFuncReturn == OK_CN00) && (strlen(aParamTable[nParam].szSize) > 0)) {
    sprintf(szParameter, aParamTable[nParam].szSize, nNumber);
    nFuncReturn = cn29SoftPutIntern( pXParamData, szParameter, szSize, true);
    if (nFuncReturn != OK_CN00) {
      cn90AnswerIError(replyData, replyLen, nFuncReturn);
    } else {
      pParameter = cn29SearchParameter(pXParamData->pFirst, szParameter);

      nFuncReturn = cn23CalculateParameter(pXParamData, pParameter, replyData, replyLen);
      if (nFuncReturn == OK_CN00) {
        nFuncReturn = cn23GetCalcState(pParameter, replyData, XPPropValueGeneral);
      } /* end if */
      if (nFuncReturn == OK_CN00) {
        nFuncReturn = cn23CheckConstraint(pXParamData, pParameter, replyData, replyLen);
      } /* end if */
      if (nFuncReturn == OK_CN00) {
        nFuncReturn = cn23GetCalcState(pParameter, replyData, XPPropValueGeneral);
      } /* end if */

      if (nFuncReturn == OK_CN00) {
         cn29StringToParamValue(&(pParameter->xpValueRecord), szSize);
        nFuncReturn = cn20XParamPutValue(vcontrol->dbname, szParameter, &(pParameter->xpValueRecord));
        if (nFuncReturn != OK_CN00) {
          cn90AnswerIError(replyData, replyLen, nFuncReturn);
        } // end if
      } // end if
    } // end if
  } /* end if */

  // free session
  if (pXParamData != NULL) {
    cn29FreeXParamSection(pXParamData);
  } // end if

  if (nFuncReturn == OK_CN00) {
    if (aState == STATE_OFFLINE_CN00 || aState == STATE_RUIN_CN00) {
      command->args = "";
      nFuncReturn = cn20CalcDevices(vcontrol, command, replyData, replyLen, replyLenMax);
    } // end if
  } // end if

  /* generate answer */
  if (nFuncReturn == OK_CN00) {
    nReturnSize = nSize;
    cn90AnswerOK(replyData, replyLen, NULL);
  } /* end if */

  *replyLen = (int)strlen(replyData);

  return nFuncReturn;
} /* end cn20ParamAddDevSpace */

/* PUBLIC FUNCTION cn20ParamDelDevSpace
 * ------------------------------------------------------------------
 * purpose: del devspace params
 *
 */
tcn00_Error cn20ParamDelDevSpace
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error            nFuncReturn    = OK_CN00;
  char                   szNumber   [XP_STRING_LENGTH_MXCN002 + 1];
  char                   szMode     [XP_STRING_LENGTH_MXCN002 + 1];
  char                   szParameter[XP_STRING_LENGTH_MXCN002 + 1];
  tcn00_DevspaceKeywords aKeyTable  [] = XP_DEVSPACE_KEYWORDS_CN00;
  tcn00_DevspaceParams   aParamTable[] = XP_DEVSPACE_PARAMS_CN00;
  long                   nKey          = 0;   
  long                   nParam        = 0;   
  int                    nNumber       = 0;
  tcn00_DBState          aState;
  tsp00_C64c             szChanged;
  tsp00_XpKeyTypec       szChangeName;
  bool                   bDirty        = false;

  /* is there an active param session */
  if (nFuncReturn == OK_CN00) {
    if (vcontrol->pXParamSection != NULL) {
      nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_XPSESSION_CN00);
    } /* end if */
  } /* end if */

  // check dirty flag
  if (nFuncReturn == OK_CN00) {
    aState = cn90DBState(vcontrol->dbname);
    if (aState == STATE_ERROR_CN00) {
      nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_STATE_CN00);
    } else if (aState != STATE_OFFLINE_CN00 && aState != STATE_RUIN_CN00) {
      szChangeName.rawAssign(XP_PARAM_CHANGED);
      if (cn20XParamGetValue( vcontrol->dbname, szChangeName, szChanged) == OK_CN00) {
        if (atol(szChanged) != 0) {
          bDirty = true;
        } // end if
      } // end if
    } // end if
  } // end if

  /* get parameters and check mode */
  if (nFuncReturn == OK_CN00) {
    cn90GetToken(command->args, szNumber, 1);
    cn90GetToken(command->args, szMode, 2);

    /* search in keywordtable */
    for (nKey = 0; (aKeyTable[nKey].nDevSpace != XPDevSpaceUnknown_ecn00) &&
                   (_tcsicmp(aKeyTable[nKey].szText, szMode) != 0  );    nKey++);

    /* search in paramtable */
    for (nParam = 0; aKeyTable[nKey].nDevSpace != aParamTable[nParam].nDevSpace; nParam++);

    if (aKeyTable[nKey].nDevSpace == XPDevSpaceUnknown_ecn00) {
      nFuncReturn = ERR_PARAM_CN00;
      cn90AnswerIError(replyData, replyLen, nFuncReturn);
    } else if (bDirty && !aKeyTable[nKey].bMirror) {
      nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_XPDIRTY_CN00);
    } else {
      /* set number of devspace */
      nNumber = atoi(szNumber);
    } /* end if */
  } /* end if */

  /* set name of devspace */
  if (nFuncReturn == OK_CN00) {
    sprintf(szParameter, aParamTable[nParam].szName, nNumber);
    nFuncReturn = cn20ParamDirectDel( vcontrol, szParameter, replyData, replyLen, replyLenMax);
  } /* end if */
  /* set type of devspace */
  if (nFuncReturn == OK_CN00) {
    sprintf(szParameter, aParamTable[nParam].szType, nNumber);
    nFuncReturn = cn20ParamDirectDel( vcontrol, szParameter, replyData, replyLen, replyLenMax);
  } /* end if */
  /* set size of devspace */
  if ((nFuncReturn == OK_CN00) && (strlen(aParamTable[nParam].szSize) > 0)) {
    sprintf(szParameter, aParamTable[nParam].szSize, nNumber);
    nFuncReturn = cn20ParamDirectDel( vcontrol, szParameter, replyData, replyLen, replyLenMax);
  } /* end if */

  /* generate answer */
  if (nFuncReturn == OK_CN00) {
    cn90AnswerOK(replyData, replyLen, NULL);
  } /* end if */

  return nFuncReturn;
} // end cn20ParamDelDevSpace

/* PUBLIC FUNCTION cn20ParamGetDevSpace
 * ------------------------------------------------------------------
 * purpose: get devspace params
 *
 */
tcn00_Error cn20ParamGetDevSpace
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error            nFuncReturn   = OK_CN00;
  char                   szNumber   [XP_STRING_LENGTH_MXCN002 + 1];
  char                   szMode     [XP_STRING_LENGTH_MXCN002 + 1];
  tcn00_DevspaceKeywords aKeyTable  [] = XP_DEVSPACE_KEYWORDS_CN00;
  tcn00_DevspaceParams   aParamTable[] = XP_DEVSPACE_PARAMS_CN00;
  long                   nKey          = 0;   
  long                   nParam        = 0;   
  int                    nNumber       = 0;
  tsp00_C64c             szName;
  tsp00_C64c             szType;
  tsp00_C64c             szSize;
  tsp00_XpKeyTypec       szParameter;

  /* get parameters and check mode */
  if (nFuncReturn == OK_CN00) {
    cn90GetToken(command->args, szNumber, 1);
    cn90GetToken(command->args, szMode, 2);

    /* search in keywordtable */
    for (nKey = 0; (aKeyTable[nKey].nDevSpace != XPDevSpaceUnknown_ecn00) &&
                   (_tcsicmp(aKeyTable[nKey].szText, szMode) != 0  );    nKey++);

    /* search in paramtable */
    for (nParam = 0; aKeyTable[nKey].nDevSpace != aParamTable[nParam].nDevSpace; nParam++);

    if (aKeyTable[nKey].nDevSpace == XPDevSpaceUnknown_ecn00) {
      nFuncReturn = ERR_PARAM_CN00;
      cn90AnswerIError(replyData, replyLen, nFuncReturn);
    } else {
      /* set number of devspace */
      nNumber = atoi(szNumber);
    } /* end if */
  } /* end if */

  szName.Init();
  szType.Init();
  szSize.Init();
  /* get name of devspace */
  if (nFuncReturn == OK_CN00) {
    sprintf(szParameter, aParamTable[nParam].szName, nNumber);
    nFuncReturn = cn20XParamGetValue(vcontrol, szParameter, szName, replyData);
  } /* end if */
  /* get type of devspace */
  if (nFuncReturn == OK_CN00) {
    sprintf(szParameter, aParamTable[nParam].szType, nNumber);
    nFuncReturn = cn20XParamGetValue(vcontrol, szParameter, szType, replyData);
  } /* end if */
  /* get size of devspace */
  if ((nFuncReturn == OK_CN00) && (strlen(aParamTable[nParam].szSize) > 0)) {
    sprintf(szParameter, aParamTable[nParam].szSize, nNumber);
    nFuncReturn = cn20XParamGetValue(vcontrol, szParameter, szSize, replyData);
  } /* end if */

  /* generate answer */
  if (nFuncReturn == OK_CN00) {
    sprintf(replyData,"%s%s%s%s%s%s%s%s",
                      ANSWER_OK_CN00, 
                      LINE_SEPSTRING_CN00,
                      szName.asCharp(),
                      LINE_SEPSTRING_CN00,
                      szType.asCharp(),
                      LINE_SEPSTRING_CN00,
                      szSize.asCharp(),
                      LINE_SEPSTRING_CN00);

  } /* end if */

  *replyLen = (int)strlen(replyData);

  return nFuncReturn;
} // end cn20ParamGetDevSpace

/* PUBLIC FUNCTION cn20ParamGetDevSpaceAll
 * ------------------------------------------------------------------
 * purpose: get devspace params
 *
 */
tcn00_Error cn20ParamGetDevSpaceAll
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error            nFuncReturn   = OK_CN00;
  bool                   bInit         = false;
  tcn002_XParamData    * pXParamData   = STATIC_CAST( tcn002_XParamData *, vcontrol->pXParamSection);
  _TCHAR                 szMode     [XP_STRING_LENGTH_MXCN002 + 1] = "";
  tcn002_XParameter    * pParameter    = NULL;
  tcn002_XParameter    * pSizeParam    = NULL;
  tcn002_XParameter    * pTypeParam    = NULL;
  _TCHAR               * pCurrent      = replyData;
  tcn00_DevspaceParams   aParamTable[] = XP_DEVSPACE_PARAMS_CN00;
  int                    nIndex        = 0;
  int                    nMIndex       = 0;
  int                    nDevspace     = 0;
  tsp00_XpKeyTypec       szParameter;
  tsp00_XpKeyTypec       szDevspace;

  if (cn90GetToken(command->args, szMode, 1)) {
    if (_tcsicmp(szMode, XP_DEV_DAT_CN00) != 0 && _tcsicmp(szMode, XP_DEV_LOG_CN00) != 0) {
      nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
    } // end if
  } // end if

  // init param management for commandline use or tcontrol or rcontrol
  if (nFuncReturn == OK_CN00) {
    if (pXParamData == NULL) {
      nFuncReturn = cn29ParamStartSession( vcontrol->dbname, pXParamData, replyData, replyLen);
      bInit = true;
    } // end if
  } // end if

  if (nFuncReturn == OK_CN00) {
    // init answer
    _stprintf(pCurrent, "%s%s", ANSWER_OK_CN00, LINE_SEPSTRING_CN00);
    pCurrent = pCurrent + _tcslen (pCurrent);

    // look for log params
    if (_tcslen(szMode) == 0 || _tcsicmp(szMode, XP_DEV_LOG_CN00) == 0) {
      pParameter = cn29SearchParameter(pXParamData->pFirst, PAN_LOG_MODE);
      _stprintf(pCurrent, "%-*s%s%s", szDevspace.size() + 1, PAN_LOG_MODE, (pParameter != NULL) ? pParameter->szInput : _T(""), LINE_SEPSTRING_CN00);
      pCurrent = pCurrent + _tcslen(pCurrent);

      pParameter = cn29SearchParameter(pXParamData->pFirst, PAN_MAXARCHIVELOGS);
      _stprintf(pCurrent, "%-*s%s%s", szDevspace.size() + 1, PAN_MAXARCHIVELOGS, (pParameter != NULL) ? pParameter->szInput : _T(""), LINE_SEPSTRING_CN00);
      pCurrent = pCurrent + _tcslen(pCurrent);
    } // end if

    // look for DATA params
    if (_tcslen(szMode) == 0 || _tcsicmp(szMode, XP_DEV_DAT_CN00) == 0) {
      pParameter = cn29SearchParameter(pXParamData->pFirst, PAN_MAXDATADEVSPACES);
      _stprintf(pCurrent, "%-*s%s%s", szDevspace.size() + 1, PAN_MAXDATADEVSPACES, (pParameter != NULL) ? pParameter->szInput : _T(""), LINE_SEPSTRING_CN00);
      pCurrent = pCurrent + _tcslen(pCurrent);

      pParameter = cn29SearchParameter(pXParamData->pFirst, PAN_MAXDATAPAGES);
      _stprintf(pCurrent, "%-*s%s%s", szDevspace.size() + 1, PAN_MAXDATAPAGES, (pParameter != NULL) ? pParameter->szInput : _T(""), LINE_SEPSTRING_CN00);
      pCurrent = pCurrent + _tcslen(pCurrent);
    } // end if

    // retrieve log devices
    if (_tcslen(szMode) == 0 || _tcsicmp(szMode, XP_DEV_LOG_CN00) == 0) {
      nIndex = 0;
      for (nIndex = 0; aParamTable[nIndex].nDevSpace != XPDevSpaceLog_ecn00; ++nIndex);
      nMIndex = 0;
      for (nMIndex = 0; aParamTable[nMIndex].nDevSpace != XPDevSpaceMLog_ecn00; ++nMIndex);

      nDevspace = 1;
      _stprintf(szDevspace, aParamTable[nIndex].szName, nDevspace);
      pParameter = cn29SearchParameter(pXParamData->pFirst, szDevspace);

      while (pParameter != NULL) {
        // name
        _stprintf(pCurrent, "%-*s", szDevspace.size() + 1, szDevspace.asCharp());
        pCurrent = pCurrent + _tcslen(pCurrent);

        // size
        _stprintf(szParameter, aParamTable[nIndex].szSize, nDevspace);
        pSizeParam = cn29SearchParameter(pXParamData->pFirst, szParameter);
        _stprintf(pCurrent, "%-11s", (pSizeParam != NULL) ? pSizeParam->szInput : _T(""));
        pCurrent = pCurrent + _tcslen(pCurrent);

        // type
        _stprintf(szParameter, aParamTable[nIndex].szType, nDevspace);
        pTypeParam = cn29SearchParameter(pXParamData->pFirst, szParameter);
        _stprintf(pCurrent, "%-2s", (pTypeParam != NULL) ? pTypeParam->szInput : _T(""));
        pCurrent = pCurrent + _tcslen(pCurrent);

        // path
        _stprintf(pCurrent, " %s%s", pParameter->szInput, LINE_SEPSTRING_CN00);
        pCurrent = pCurrent + _tcslen(pCurrent);

        // mirror
        _stprintf(szDevspace, aParamTable[nMIndex].szName, nDevspace);
        pParameter = cn29SearchParameter(pXParamData->pFirst, szDevspace);
        if (pParameter != NULL) {
          // m-name
          _stprintf(pCurrent, "%-*s", szDevspace.size() + 1, szDevspace.asCharp());
          pCurrent = pCurrent + _tcslen(pCurrent);

          // m-size
          _stprintf(pCurrent, "%-11s", _T(""));
          pCurrent = pCurrent + _tcslen(pCurrent);

          // m-type
          _stprintf(szParameter, aParamTable[nMIndex].szType, nDevspace);
          pTypeParam = cn29SearchParameter(pXParamData->pFirst, szParameter);
          _stprintf(pCurrent, "%-2s", (pTypeParam != NULL) ? pTypeParam->szInput : _T(""));
          pCurrent = pCurrent + _tcslen(pCurrent);

          // m-path
          _stprintf(pCurrent, " %s%s", pParameter->szInput, LINE_SEPSTRING_CN00);
          pCurrent = pCurrent + _tcslen(pCurrent);
        } // end if

        ++nDevspace;
        _stprintf(szDevspace, aParamTable[nIndex].szName, nDevspace);
        pParameter = cn29SearchParameter(pXParamData->pFirst, szDevspace);
      } // end while

    } // end if

    // retrieve data devices
    if (_tcslen(szMode) == 0 || _tcsicmp(szMode, XP_DEV_DAT_CN00) == 0) {
      nIndex = 0;
      for (nIndex = 0; aParamTable[nIndex].nDevSpace != XPDevSpaceDat_ecn00; ++nIndex);
      nMIndex = 0;
      for (nMIndex = 0; aParamTable[nMIndex].nDevSpace != XPDevSpaceMDat_ecn00; ++nMIndex);

      nDevspace = 1;
      _stprintf(szDevspace, aParamTable[nIndex].szName, nDevspace);
      pParameter = cn29SearchParameter(pXParamData->pFirst, szDevspace);

      while (pParameter != NULL) {
        // name
        _stprintf(pCurrent, "%-*s", szDevspace.size() + 1, szDevspace.asCharp());
        pCurrent = pCurrent + _tcslen(pCurrent);

        // size
        _stprintf(szParameter, aParamTable[nIndex].szSize, nDevspace);
        pSizeParam = cn29SearchParameter(pXParamData->pFirst, szParameter);
        _stprintf(pCurrent, "%-11s", (pSizeParam != NULL) ? pSizeParam->szInput : _T(""));
        pCurrent = pCurrent + _tcslen(pCurrent);

        // type
        _stprintf(szParameter, aParamTable[nIndex].szType, nDevspace);
        pTypeParam = cn29SearchParameter(pXParamData->pFirst, szParameter);
        _stprintf(pCurrent, "%-2s", (pTypeParam != NULL) ? pTypeParam->szInput : _T(""));
        pCurrent = pCurrent + _tcslen(pCurrent);

        // path
        _stprintf(pCurrent, " %s%s", pParameter->szInput, LINE_SEPSTRING_CN00);
        pCurrent = pCurrent + _tcslen(pCurrent);

        // mirror
        _stprintf(szDevspace, aParamTable[nMIndex].szName, nDevspace);
        pParameter = cn29SearchParameter(pXParamData->pFirst, szDevspace);
        if (pParameter != NULL ) {
          // m-name
          _stprintf(pCurrent, "%-*s", szDevspace.size() + 1, szDevspace.asCharp());
          pCurrent = pCurrent + _tcslen(pCurrent);

          // m-size
          _stprintf(pCurrent, "%-11s", _T(""));
          pCurrent = pCurrent + _tcslen(pCurrent);

          // m-type
          _stprintf(szParameter, aParamTable[nMIndex].szType, nDevspace);
          pTypeParam = cn29SearchParameter(pXParamData->pFirst, szParameter);
          _stprintf(pCurrent, "%-2s", (pTypeParam != NULL) ? pTypeParam->szInput : _T(""));
          pCurrent = pCurrent + _tcslen(pCurrent);

          // m-path
          _stprintf(pCurrent, " %s%s", pParameter->szInput, LINE_SEPSTRING_CN00);
          pCurrent = pCurrent + _tcslen(pCurrent);
        } // end if

        ++nDevspace;
        _stprintf(szDevspace, aParamTable[nIndex].szName, nDevspace);
        pParameter = cn29SearchParameter(pXParamData->pFirst, szDevspace);
      } // end while

    } // end if

  } // end if

  *replyLen = (int)strlen(replyData);

  // if init then stop param management 
  if (bInit) {
    cn29FreeXParamSection(pXParamData);
  } // end if

  return nFuncReturn;
} // end cn20ParamGetDevSpaceAll

/* PUBLIC FUNCTION cn20ParamCalcDevSpace
 * ------------------------------------------------------------------
 * purpose: recalculate parameters dependent on dev spaces
 *
 */
tcn00_Error cn20ParamCalcDevSpace
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error        nFuncReturn = OK_CN00;
  tcn002_XParamData * pXParamData = (tcn002_XParamData *) vcontrol->pXParamSection;
  tcn002_XParameter * pParameter  = NULL;
  long               nSize    = 0;
  long               nMaxSize = 0;
  long               nSumSize = 0;
  char               szSize[XP_STRING_LENGTH_MXCN002 + 1];
  char               szNumber[XP_STRING_LENGTH_MXCN002 + 1];
  char               szParameter[XP_STRING_LENGTH_MXCN002 + 1];
  tcn00_DevspaceParams aParamTable[] = XP_DEVSPACE_PARAMS_CN00;
  long               nParam  = 0;   
  int                nNumber = 0;

  /* is there an active param session */
  if (nFuncReturn == OK_CN00) {
    if (pXParamData == NULL) {
      nFuncReturn = ERR_NOXPSESSION_CN00;
      cn90AnswerIError(replyData, replyLen, nFuncReturn);
    } /* end if */
  } /* end if */

  /* get parameters and check mode */
  if (nFuncReturn == OK_CN00) {
    while ((aParamTable[nParam].nDevSpace != XPDevSpaceUnknown_ecn00) && (nFuncReturn == OK_CN00)) {

      /* check and if necessary write max size (MAXDATAPAGES) */
      if ((nFuncReturn == OK_CN00) && (strlen(aParamTable[nParam].szMaxSize) > 0)) {
        nNumber = 0;
        /* calculate sizes */
        do {
          nNumber++;
          sprintf(szParameter, aParamTable[nParam].szSize, nNumber);
          pParameter = cn29SearchParameter(pXParamData->pFirst, szParameter);
          if (pParameter != NULL) {
            nSize = atol(pParameter->szInput);
            nMaxSize = (nMaxSize < nSize) ? nSize : nMaxSize;
            nSumSize = nSumSize + nSize;
          } /* end if */
        } while (pParameter != NULL);

        /* write size */
        sprintf(szSize, "%ld", (nSumSize + nMaxSize));
        pParameter = cn29SearchParameter(pXParamData->pFirst, aParamTable[nParam].szMaxSize);
        if (atol(pParameter->szInput) <  (nSumSize + nMaxSize)) {
          nFuncReturn = cn29SoftPutIntern( pXParamData, aParamTable[nParam].szMaxSize, szSize);
        } /* end if */
      } /* end if */

      /* check and if necessary write maxcount (MAXARCHIVELOGS or MAXDATADEVSPACES) */
      if ((nFuncReturn == OK_CN00) && (strlen(aParamTable[nParam].szMaxCount) > 0)) {
        nNumber = 0;
        /* calculate count */
        do {
          nNumber++;
          sprintf(szParameter, aParamTable[nParam].szName, nNumber);
          pParameter = cn29SearchParameter(pXParamData->pFirst, szParameter);
          if (pParameter != NULL) {
            pParameter = (strlen(pParameter->szInput) != 0) ? pParameter : NULL;
          } /* end if */
        } while (pParameter != NULL);

        /* write count */
        pParameter = cn29SearchParameter(pXParamData->pFirst, aParamTable[nParam].szMaxCount);
        if (atoi(pParameter->szInput) < nNumber) {
          do {
            sprintf(szNumber, "%d", nNumber);
            nFuncReturn = cn29SoftPutIntern( pXParamData, aParamTable[nParam].szMaxCount, szNumber);
            --nNumber;
          } while (nFuncReturn == ERR_XPCHECK_CN00);
        } /* end if */
      } /* end if */

      nParam++;
    } /* end while */

  } /* end if */

  /* generate answer */
  if (nFuncReturn == OK_CN00) {
    cn90AnswerOK(replyData, replyLen, NULL);
  } /* end if */

  return nFuncReturn;
} /* end cn20ParamCalcDevSpace */

/* ------------------------------------------------------------------
 * IMPLEMENTATION PUBLIC FUNCTION cn20CalcDevices
 * ------------------------------------------------------------------
 */
#undef  CN_FUNCTION
#define CN_FUNCTION "cn20CalcDevices"
tcn00_Error cn20CalcDevices
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  CONSOLE_TRACE_ENTER_CN00;

  tcn00_Error   nTmpState = OK_CN00;
  const char  * pArgs;

  pArgs = command->args;
  command->args = "";

  // calc devspaces
  if (cn20ParamStartSession( vcontrol, 
                             command, 
                             replyData, 
                             replyLen, 
                             replyLenMax) == OK_CN00) {
    nTmpState = cn20ParamCalcDevSpace( vcontrol, 
                                       command, 
                                       replyData, 
                                       replyLen, 
                                       replyLenMax);

   
    if (nTmpState == OK_CN00) {
      nTmpState = cn20ParamCheckAll( vcontrol, 
                                     command, 
                                     replyData, 
                                     replyLen, 
                                     replyLenMax);
    } // end if
    if (nTmpState == OK_CN00) {
      nTmpState = cn20ParamCommitSession( vcontrol, 
                                           command, 
                                           replyData, 
                                           replyLen, 
                                           replyLenMax);
    } else {
      nTmpState = cn20ParamAbortSession( vcontrol, 
                                         command, 
                                         replyData, 
                                         replyLen, 
                                         replyLenMax);
    } // end if
  } // end if

  command->args = pArgs;

  return nTmpState;
} // cn20CalcDevices

/* ------------------------------------------------------------------
 * IMPLEMENTATION PUBLIC FUNCTION cn20CreateSrvParam
 * ------------------------------------------------------------------
 */
tcn00_Error cn20CreateSrvParam
      ( tsp00_DbNamec   szDbName,
        const bool      bCheckOnly)
{
  FUNCTION_DBG_MCN00_1(_T("cn20CreateSrvParam"));

  tcn00_Error         nFuncReturn   = OK_CN00;
  tcn002_XParamData * pXParamData   = NULL;
  tsp00_Pathc         szRundir;
  tsp00_C256          szWrk;
  tsp00_Versionc      VersionString;
  _TCHAR              replyData[1000];
  int                 replyLen;
  int                 replyLenMax = 1000;
  tsp01_RteError      rteError;

  // start session
  if (!bCheckOnly) {
    sqlXParamRemoveAll(szDbName, &rteError);
  } // end if
  nFuncReturn = cn29ParamStartSession( szDbName, pXParamData, replyData, &replyLen);
  if (nFuncReturn == OK_CN00) {

    // collect some values
    cn90GetWrkPath(szWrk);
    if (strlen(szWrk) > 0) {
      sprintf(szRundir, "%s%s", szWrk.asCharp(), szDbName.asCharp());
      nFuncReturn = cn90DirCreate(szRundir, &rteError);
      if (nFuncReturn == OK_CN00) {
        sp100_GetVersionString(COMP_NAME_KERNEL_SP100, s100buildnumber, VersionString);
      } else {
        teo200_EventList aEvent1(FUNCTION_NAME_MCN00_1, rteError.OsErrCode,  TERR_CN00_1, _T("DBM"), rteError.OsErrText);
        teo200_EventList aEvent2(aEvent1, FUNCTION_NAME_MCN00_1, rteError.RteErrCode,  TERR_CN00_1, _T("DBM"), rteError.RteErrText);
        teo200_EventList aEvent3(aEvent2, FUNCTION_NAME_MCN00_1, ERR_RTEEXT_CN00_1);
        tin100_GlobalLog::writeEntry (aEvent3);
      } // end if
    } else {
      nFuncReturn = ERR_FILE_CN00;
      teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, nFuncReturn,  TERR_CN00_1, _T("DBM"), _T("Empty Wrk Path"));
      tin100_GlobalLog::writeEntry (aEvent);
    } // end if

    // set some parameters
//    if (nFuncReturn == OK_CN00) {
//      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_SERVERDB             , szDbName      , true);
//    } // end if
    if (nFuncReturn == OK_CN00) {
      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_RUNDIR               , szRundir      , true);
      if (nFuncReturn != OK_CN00) {
        teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, nFuncReturn,  TERR_CN00_1, _T("DBM"), _T("Error putting %s"), PAN_RUNDIR);
        tin100_GlobalLog::writeEntry (aEvent);
      } // end if
    } // end if
    if (nFuncReturn == OK_CN00) {
      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_PARAM_KERNELVERSION  , VersionString , true);
      if (nFuncReturn != OK_CN00) {
        teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, nFuncReturn,  TERR_CN00_1, _T("DBM"), _T("Error putting %s"), PAN_PARAM_KERNELVERSION);
        tin100_GlobalLog::writeEntry (aEvent);
      } // end if
    } // end if
    if (nFuncReturn == OK_CN00) {
      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_CONTROLID            , "SERVICE"     , true);
      if (nFuncReturn != OK_CN00) {
        teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, nFuncReturn,  TERR_CN00_1, _T("DBM"), _T("Error putting %s"), PAN_CONTROLID);
        tin100_GlobalLog::writeEntry (aEvent);
      } // end if
    } // end if
    if (nFuncReturn == OK_CN00) {
      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_CONTROLPW            , "SERVICE"     , true);
      if (nFuncReturn != OK_CN00) {
        teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, nFuncReturn,  TERR_CN00_1, _T("DBM"), _T("Error putting %s"), PAN_CONTROLPW);
        tin100_GlobalLog::writeEntry (aEvent);
      } // end if
    } // end if
//    if (nFuncReturn == OK_CN00) {
//      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_MAXBACKUPDEVS        , "32"          , true);
//    } // end if
//    if (nFuncReturn == OK_CN00) {
//      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_BACKUP_BLOCK_CNT     , "128"         , true);
//    } // end if
    if (nFuncReturn == OK_CN00) {
      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_LOG_SEGMENT_SIZE     , "0"           , true);
      if (nFuncReturn != OK_CN00) {
        teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, nFuncReturn,  TERR_CN00_1, _T("DBM"), _T("Error putting %s"), PAN_LOG_SEGMENT_SIZE);
        tin100_GlobalLog::writeEntry (aEvent);
      } // end if
    } // end if
//    if (nFuncReturn == OK_CN00) {
//      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_OBJ_LOG_SIZE         , "0"           , true);
//    } // end if
    if (nFuncReturn == OK_CN00) {
      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_LOG_MODE             , "DEMO"        , true);
      if (nFuncReturn != OK_CN00) {
        teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, nFuncReturn,  TERR_CN00_1, _T("DBM"), _T("Error putting %s"), PAN_LOG_MODE);
        tin100_GlobalLog::writeEntry (aEvent);
      } // end if
    } // end if
//    if (nFuncReturn == OK_CN00) {
//      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_OBJ_ROLLBACK_CACHE   , "8"           , true);
//    } // end if
//    if (nFuncReturn == OK_CN00) {
//      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_OBJ_LOG_IO_QUEUE     , "8"           , true);
//    } // end if
    if (nFuncReturn == OK_CN00) {
      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_MAXGARBAGE_COLLECTOR , "0"           , true);
      if (nFuncReturn != OK_CN00) {
        teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, nFuncReturn,  TERR_CN00_1, _T("DBM"), _T("Error putting %s"), PAN_MAXGARBAGE_COLLECTOR);
        tin100_GlobalLog::writeEntry (aEvent);
      } // end if
    } // end if
    if (nFuncReturn == OK_CN00) {
      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_MAXACTIVE            , "1"           , true);
      if (nFuncReturn != OK_CN00) {
        teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, nFuncReturn,  TERR_CN00_1, _T("DBM"), _T("Error putting %s"), PAN_MAXACTIVE);
        tin100_GlobalLog::writeEntry (aEvent);
      } // end if
    } // end if
    if (nFuncReturn == OK_CN00) {
      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_DATA_CACHE_PAGS      , "10"          , true);
      if (nFuncReturn != OK_CN00) {
        teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, nFuncReturn,  TERR_CN00_1, _T("DBM"), _T("Error putting %s"), PAN_DATA_CACHE_PAGS);
        tin100_GlobalLog::writeEntry (aEvent);
      } // end if
    } // end if
    if (nFuncReturn == OK_CN00) {
      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_INDEXFILE_LIST_SIZE  , "0"           , true);
      if (nFuncReturn != OK_CN00) {
        teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, nFuncReturn,  TERR_CN00_1, _T("DBM"), _T("Error putting %s"), PAN_INDEXFILE_LIST_SIZE);
        tin100_GlobalLog::writeEntry (aEvent);
      } // end if
    } // end if
    if (nFuncReturn == OK_CN00) {
      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_TREE_REGIONS         , "1"           , true);
      if (nFuncReturn != OK_CN00) {
        teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, nFuncReturn,  TERR_CN00_1, _T("DBM"), _T("Error putting %s"), PAN_TREE_REGIONS);
        tin100_GlobalLog::writeEntry (aEvent);
      } // end if
    } // end if
    if (nFuncReturn == OK_CN00) {
      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_DATACACHE_REGIONS    , "1"           , true);
      if (nFuncReturn != OK_CN00) {
        teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, nFuncReturn,  TERR_CN00_1, _T("DBM"), _T("Error putting %s"), PAN_DATACACHE_REGIONS);
        tin100_GlobalLog::writeEntry (aEvent);
      } // end if
    } // end if
    if (nFuncReturn == OK_CN00) {
      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_TRANS_REGIONS        , "1"           , true);
      if (nFuncReturn != OK_CN00) {
        teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, nFuncReturn,  TERR_CN00_1, _T("DBM"), _T("Error putting %s"), PAN_TRANS_REGIONS);
        tin100_GlobalLog::writeEntry (aEvent);
      } // end if
    } // end if
    if (nFuncReturn == OK_CN00) {
      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_TAB_REGIONS          , "1"           , true);
      if (nFuncReturn != OK_CN00) {
        teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, nFuncReturn,  TERR_CN00_1, _T("DBM"), _T("Error putting %s"), PAN_TAB_REGIONS);
        tin100_GlobalLog::writeEntry (aEvent);
      } // end if
    } // end if
    if (nFuncReturn == OK_CN00) {
      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_ROW_REGIONS          , "1"           , true);
      if (nFuncReturn != OK_CN00) {
        teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, nFuncReturn,  TERR_CN00_1, _T("DBM"), _T("Error putting %s"), PAN_ROW_REGIONS);
        tin100_GlobalLog::writeEntry (aEvent);
      } // end if
    } // end if
    if (nFuncReturn == OK_CN00) {
      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_MAXDATAPAGES         , "3000"        , true);
      if (nFuncReturn != OK_CN00) {
        teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, nFuncReturn,  TERR_CN00_1, _T("DBM"), _T("Error putting %s"), PAN_MAXDATAPAGES);
        tin100_GlobalLog::writeEntry (aEvent);
      } // end if
    } // end if
//    if (nFuncReturn == OK_CN00) {
//      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_MAXLOCKS             , "50"          , true);
//    } // end if
    if (nFuncReturn == OK_CN00) {
#if defined(page32KB)
      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_MAXUSERSTACK         , "600"         , true);
#else
      nFuncReturn = cn29SoftPutIntern( pXParamData, PAN_MAXUSERSTACK         , "400"         , true);
#endif
      if (nFuncReturn != OK_CN00) {
        teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, nFuncReturn,  TERR_CN00_1, _T("DBM"), _T("Error putting %s"), PAN_MAXUSERSTACK);
        tin100_GlobalLog::writeEntry (aEvent);
      } // end if
    } // end if
    if ( (nFuncReturn == OK_CN00) && !bCheckOnly) {
#if defined(SAPDBR73)
      if (nFuncReturn == OK_CN00) {
        nFuncReturn = cn29SoftPutIntern( pXParamData, "SYSDEV_001"             , "CHECKSAVE"   , true);
        if (nFuncReturn != OK_CN00) {
          teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, nFuncReturn,  TERR_CN00_1, _T("DBM"), _T("Error putting %s"), "SYSDEV_001");
          tin100_GlobalLog::writeEntry (aEvent);
        } // end if
      } // end if
      if (nFuncReturn == OK_CN00) {
        nFuncReturn = cn29SoftPutIntern( pXParamData, "SYS_TYPE_001"           , "F"           , true);
        if (nFuncReturn != OK_CN00) {
          teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, nFuncReturn,  TERR_CN00_1, _T("DBM"), _T("Error putting %s"), "SYS_TYPE_001");
          tin100_GlobalLog::writeEntry (aEvent);
        } // end if
      } // end if
#endif // SAPDBR73
      if (nFuncReturn == OK_CN00) {
        nFuncReturn = cn29SoftPutIntern( pXParamData, "DATADEV_0001"           , "CHECKDEV"    , true);
        if (nFuncReturn != OK_CN00) {
          teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, nFuncReturn,  TERR_CN00_1, _T("DBM"), _T("Error putting %s"), "DATADEV_0001");
          tin100_GlobalLog::writeEntry (aEvent);
        } // end if
      } // end if
      if (nFuncReturn == OK_CN00) {
        nFuncReturn = cn29SoftPutIntern( pXParamData, "DATA_TYPE_0001"         , "F"           , true);
        if (nFuncReturn != OK_CN00) {
          teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, nFuncReturn,  TERR_CN00_1, _T("DBM"), _T("Error putting %s"), "DATA_TYPE_0001");
          tin100_GlobalLog::writeEntry (aEvent);
        } // end if
      } // end if
      if (nFuncReturn == OK_CN00) {
        nFuncReturn = cn29SoftPutIntern( pXParamData, "DATA_SIZE_0001"         , "3000"        , true);
        if (nFuncReturn != OK_CN00) {
          teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, nFuncReturn,  TERR_CN00_1, _T("DBM"), _T("Error putting %s"), "DATA_SIZE_0001");
          tin100_GlobalLog::writeEntry (aEvent);
        } // end if
      } // end if
    } // end if

    // check and save
    if (nFuncReturn == OK_CN00) {
      nFuncReturn = cn20_ParamCheckAll( szDbName, pXParamData, "", replyData, &replyLen, replyLenMax);
      if (nFuncReturn == OK_CN00) {
        nFuncReturn = cn20_ParamCommitSession(szDbName, pXParamData, replyData, &replyLen, replyLenMax);
        if (nFuncReturn != OK_CN00) {
          teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, nFuncReturn,  TERR_CN00_1, _T("DBM"), _T("Error commit param session\n%s"), replyData);
          tin100_GlobalLog::writeEntry (aEvent);
        } // end if
      } else {
        teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, nFuncReturn,  TERR_CN00_1, _T("DBM"), _T("Error check params\n%s"), replyData);
        tin100_GlobalLog::writeEntry (aEvent);
      } // end if
    } // endif

    cn29FreeXParamSection(pXParamData);

  } else {
    teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, nFuncReturn,  TERR_CN00_1, _T("DBM"), _T("Error start param session\n%s"), replyData);
    tin100_GlobalLog::writeEntry (aEvent);
  } // end if

  return nFuncReturn;
} // cn20CreateSrvParam

/* PUBLIC FUNCTION cn20ParamRmFile
 * ------------------------------------------------------------------
 *
 */
tcn00_Error cn20ParamRmFile
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error        nFuncReturn = OK_CN00;
  tsp01_RteError     rteError;

  /* read the parameters */
  if (nFuncReturn == OK_CN00) {
    if (cn90GetToken(command->args, NULL, 1)         ||
       (vcontrol->dbname[0] == CHAR_STRINGTERM_CN90)    ) {
      nFuncReturn = ERR_PARAM_CN00;
      cn90AnswerIError(replyData, replyLen, nFuncReturn);
    } /* end if */
  } /* end if */

  /* check existence of session */
  if (nFuncReturn == OK_CN00) {
    if (vcontrol->pXParamSection != NULL) {
      nFuncReturn = ERR_XPSESSION_CN00;
      cn90AnswerIError(replyData, replyLen, nFuncReturn);
    } /* end if */
  } /* end if */

  /* delete the paramfile */
  if (nFuncReturn == OK_CN00) {
    if (!sqlXParamRemoveOneParamfile(vcontrol->dbname, 0, &rteError)) {
//    if (!sqlXParamRemoveAll(vcontrol->dbname, &rteError)) {
      nFuncReturn = ERR_RTEEXT_CN00;
      cn90AnswerNewRTEError(replyData, replyLen, nFuncReturn, &rteError);
    } /* end if */
  } /* end if */

  if (nFuncReturn == OK_CN00) {
    cn90AnswerOK(replyData, replyLen, NULL);
  } /* end if */

  return nFuncReturn;
} /* end cn20ParamRmFile */

/* PUBLIC FUNCTION cn20ParamRestore
 * ------------------------------------------------------------------
 * purpose: Restore old param file 
 */
tcn00_Error cn20ParamRestore
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error        nFuncReturn = OK_CN00;
  char               szVersion[PARAMETER_MAXLEN_CN90];
  char               szSource[PARAMETER_MAXLEN_CN90];
  char               szTarget[PARAMETER_MAXLEN_CN90];
  tsp00_C256         szConfig;

  tsp01_RteError     aRteError;

  /* read the parameters */
  if (nFuncReturn == OK_CN00) {
    if (!cn90GetToken(command->args, szVersion, 1)    ||
        cn90GetToken(command->args, NULL, 2)          ||
       (vcontrol->dbname[0] == CHAR_STRINGTERM_CN90)    ) {
      nFuncReturn = ERR_PARAM_CN00;
      cn90AnswerIError(replyData, replyLen, nFuncReturn);
    } /* end if */
  } /* end if */

  /* check existence of session */
  if (nFuncReturn == OK_CN00) {
    if (vcontrol->pXParamSection != NULL) {
      nFuncReturn = ERR_XPSESSION_CN00;
      cn90AnswerIError(replyData, replyLen, nFuncReturn);
    } /* end if */
  } /* end if */

  // write the history
  if (nFuncReturn == OK_CN00) {
    tsp11_XpReturnCode xpReturn;
    tsp00_Int4         xpVersId;
    tsp00_Int4         xpOrgId;
    tsp01_RteError     RteError;
    tsp00_XpKeyTypec   acXParamKey;
    tsp00_XpKeyTypec   acXParamNextKey;
    tcn002_XParameter  aParameter;

    // open Version file
    xpReturn = sqlXparamOpenVersion( vcontrol->dbname,
                                     atoi(szVersion),
		  	 		                         &xpVersId,
		  	                             XP_OPEN_READONLY_SP11 | XP_OPEN_EXISTING_SP11,
                                     &RteError);
    if (xpReturn == xp_ok_sp11) {

      // open current file
      xpReturn = sqlXparamOpen ( vcontrol->dbname,
		  	 		                     &xpOrgId,
		  	                         XP_OPEN_READONLY_SP11 | XP_OPEN_EXISTING_SP11,
                                 &RteError);
      if (xpReturn == xp_ok_sp11) {
        
        // look for parameters changed or new after the restore
        acXParamKey.Init();
        acXParamNextKey.Init();
        while (xpReturn == xp_ok_sp11) {
          // read from version file
          acXParamKey = acXParamNextKey; 
          xpReturn = sqlXparamGetNext( xpVersId,
                                       acXParamKey,
                                       acXParamNextKey,
                                       &aParameter.xpValueRecord,
                                       &RteError );
          if (xpReturn == xp_ok_sp11) {
            strcpy(aParameter.szParameterName, acXParamNextKey);
            cn29ParamValueToString(aParameter.szInput, &aParameter.xpValueRecord);

            xpReturn = sqlXparamGet   ( xpOrgId,
                                        acXParamNextKey,
                                        &aParameter.xpValueRecord,
                                        &RteError );
             if (xpReturn == xp_ok_sp11 || xpReturn == xp_key_not_found_sp11) {

               if (xpReturn == xp_ok_sp11) {
                 cn29ParamValueToString(aParameter.szOriginal, &aParameter.xpValueRecord);
               } else {
                 strcpy(aParameter.szOriginal, "");
               } // end if

               cn29WriteParamHistory(vcontrol->dbname, &aParameter, PAHI_CHANGE_CN002);

               xpReturn = xp_ok_sp11;
             } // end if

          } // end if

        } // end while

        xpReturn = (xpReturn == xp_eof_sp11) ? xp_ok_sp11 : xpReturn;

        // look for parameters deleted after the restore
        acXParamKey.Init();
        acXParamNextKey.Init();
        while (xpReturn == xp_ok_sp11) {
          // read from version file
          acXParamKey = acXParamNextKey; 
          xpReturn = sqlXparamGetNext( xpOrgId,
                                       acXParamKey,
                                       acXParamNextKey,
                                       &aParameter.xpValueRecord,
                                       &RteError );
          if (xpReturn == xp_ok_sp11) {
            strcpy(aParameter.szParameterName, acXParamNextKey);
            cn29ParamValueToString(aParameter.szOriginal, &aParameter.xpValueRecord);

            xpReturn = sqlXparamGet   ( xpVersId,
                                        acXParamNextKey,
                                        &aParameter.xpValueRecord,
                                        &RteError );
             if (xpReturn == xp_key_not_found_sp11) {

               strcpy(aParameter.szInput, "");
               cn29WriteParamHistory(vcontrol->dbname, &aParameter, PAHI_DELETE_CN002);

               xpReturn = xp_ok_sp11;

             } // end if

          } // end if

        } // end while

        // close current file
        sqlXparamClose(xpOrgId, false, &RteError);
      } // end if

      // close version file
      sqlXparamClose(xpVersId, false, &RteError);
    } // end if
  } // end if

  /* copy the version file */
  if (nFuncReturn == OK_CN00) {

    cn90GetCfgPath(szConfig);

    sprintf(szTarget, "%s%s",  (char *) szConfig, (char *) vcontrol->dbname);
    sprintf(szSource, "%s.%02d", (char *) szTarget, atoi(szVersion));

    if (!sqlfilecopyc( szSource, szTarget, &aRteError )) {
      nFuncReturn = ERR_RTEEXT_CN00;
      cn90AnswerNewRTEError (replyData, replyLen, nFuncReturn, &aRteError );
    } /* end if */

  } /* end if */

  if (nFuncReturn == OK_CN00) {
    cn90AnswerOK(replyData, replyLen, NULL);
  } /* end if */

  return nFuncReturn;
} /* end  cn20ParamRestore */

/* PUBLIC FUNCTION cn20ParamVersions
 * ------------------------------------------------------------------
 * purpose: delete one named paramter from the param file
 */
tcn00_Error cn20ParamVersions
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error        nFuncReturn = OK_CN00;
  char             * pCurrPos;
  void             * hDirectory  = NULL;  
  bool               bOpen       = false;
  tsp_c256           szFile;
  tsp_c256           szConfig;
  tsp5_rte_fileerror rteFileErr;
  
  /* read the parameters */
  if (nFuncReturn == OK_CN00) {
    if (cn90GetToken(command->args, NULL, 1)         ||
       (vcontrol->dbname[0] == CHAR_STRINGTERM_CN90)    ) {
      nFuncReturn = ERR_PARAM_CN00;
      cn90AnswerIError(replyData, replyLen, nFuncReturn);
    } /* end if */
  } /* end if */

  /* open config dir */
  if (nFuncReturn == OK_CN00) {
    cn90GetCfgPath(szConfig);

    szConfig[strlen(szConfig) - 1] = CHAR_STRINGTERM_CN90;
    sqlfopendirc ( &szConfig, &hDirectory,  &szFile, &rteFileErr);
    if  ( rteFileErr.sp5fe_result != vf_ok) {
      nFuncReturn = ERR_RTE_CN00;
    } else {
      bOpen = true;
    } /* end if */
  } /* end if */

  /* analyze files */
  if (nFuncReturn == OK_CN00) {

    /* Init Output */
    sprintf(replyData, "%s%s", ANSWER_OK_CN00, LINE_SEPSTRING_CN00);
    pCurrPos = replyData + strlen (replyData);

    while ( rteFileErr.sp5fe_result == vf_ok) {

      if ( ( strncmp(szFile, vcontrol->dbname, strlen(vcontrol->dbname)) == 0 ) &&
           ( ( strlen(vcontrol->dbname) == strlen(szFile)           ) ||
             ( (strlen(vcontrol->dbname) + 3) == strlen(szFile) &&
               szFile[strlen(vcontrol->dbname)] == '.'          &&
               isdigit(szFile[strlen(vcontrol->dbname) + 1])    &&
               isdigit(szFile[strlen(vcontrol->dbname) + 2])        )         )    ) {

       sprintf(pCurrPos, "%s%c%s%s", (char *) szConfig, PATH_DELIMITER_CN90, (char *) szFile, LINE_SEPSTRING_CN00);
       pCurrPos = replyData + strlen (replyData);
      } /* end if */

      /* get the next name */
      sqlfreaddirc(hDirectory,  &szFile, &rteFileErr);
    } /* end while */

  } /* end if */

  /* close */
  if (bOpen) {
    sqlfclosedirc (hDirectory, &rteFileErr);
    if (nFuncReturn == OK_CN00 && rteFileErr.sp5fe_result != vf_ok) {
      nFuncReturn = ERR_RTE_CN00;
    } /* end if */
  } /* end if */

  /* check result */
  if (nFuncReturn == OK_CN00) {
    *replyLen = (int)strlen(replyData);
  } else if (nFuncReturn == ERR_RTE_CN00) {
    cn90AnswerRTEError (replyData, replyLen, nFuncReturn, rteFileErr.sp5fe_text, rteFileErr.sp5fe_text.length(), rteFileErr.sp5fe_result );
  } else {
    cn90AnswerIError(replyData, replyLen, nFuncReturn);
  } /* end if */

  return nFuncReturn;
} /* end  cn20ParamVersions */

/* PUBLIC FUNCTION cn20ParamHistory
 * ------------------------------------------------------------------
 */
tcn00_Error cn20ParamHistory
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error              nFuncReturn = OK_CN00;
  tsp00_ErrTextc           szError;
  _TCHAR                 * pCurrent    = replyData;
  bool                     bOk         = true;
  bool                     bBufferLen  = true;
  cn52ParamHistoryRecord * pRecord;

  cn20FreeParamHistory(vcontrol);
  pRecord = new cn52ParamHistoryRecord(vcontrol->dbname);
  vcontrol->pParamHistory = (void *) pRecord;

  tsp00_C256c szSelection;
  bOk = pRecord->setSelection(szSelection.rawAssign(command->args));

  if (bOk) {

    _stprintf(pCurrent, "%s%s", ANSWER_OK_CN00, LINE_SEPSTRING_CN00);
    pCurrent = pCurrent + _tcslen (pCurrent);
    _stprintf(pCurrent, "%s%s", pRecord->getHeader().asCharp(), LINE_SEPSTRING_CN00);
    pCurrent = pCurrent + _tcslen (pCurrent);

    bOk = pRecord->readFirst(szError);

    while (bOk && bBufferLen) {

      _stprintf(pCurrent, "%s%s", pRecord->getRecord().asCharp(), LINE_SEPSTRING_CN00);
      pCurrent = pCurrent + _tcslen (pCurrent);

      bOk = pRecord->readNext(szError);
      if (bOk && ((pCurrent - replyData + pRecord->getRecord().length() + 5) > replyLenMax)) {
        bBufferLen = false;
      } // end if

    } // end if
  
    if (!bOk) {
      if (szError.length() > 0) {
        nFuncReturn = cn90AnswerRTEError (replyData, replyLen, ERR_FILE_CN00, szError, szError.length(), vf_notok );
      } // end if
      cn20FreeParamHistory(vcontrol);
    } // end if
  
  } else {
    nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
    cn20FreeParamHistory(vcontrol);
  } // end if

  *replyLen = (int)_tcslen(replyData);

  return nFuncReturn;
} // end cn20ParamHistory

/* PUBLIC FUNCTION cn20ParamHistoryNext
 * ------------------------------------------------------------------
 */
tcn00_Error cn20ParamHistoryNext
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error              nFuncReturn = OK_CN00;
  tsp00_ErrTextc           szError;
  _TCHAR                 * pCurrent = replyData;
  bool                     bOk         = true;
  bool                     bBufferLen  = true;
  cn52ParamHistoryRecord * pRecord = (cn52ParamHistoryRecord *) vcontrol->pParamHistory;
  
  _stprintf(pCurrent, "%s%s", ANSWER_OK_CN00, LINE_SEPSTRING_CN00);
  pCurrent = pCurrent + _tcslen (pCurrent);

  if (pRecord != NULL) {
    _stprintf(pCurrent, "%s%s", pRecord->getHeader().asCharp(), LINE_SEPSTRING_CN00);
    pCurrent = pCurrent + _tcslen (pCurrent);

    while (bOk && bBufferLen) {

      _stprintf(pCurrent, "%s%s", pRecord->getRecord().asCharp(), LINE_SEPSTRING_CN00);
      pCurrent = pCurrent + _tcslen (pCurrent);

      bOk = pRecord->readNext(szError);
      if (bOk && ((pCurrent - replyData + pRecord->getRecord().length() + 5) > replyLenMax)) {
        bBufferLen = false;
      } // end if

    } // end if
  
    if (!bOk) {
      if (szError.length() > 0) {
        nFuncReturn = cn90AnswerRTEError (replyData, replyLen, ERR_FILE_CN00, szError, szError.length(), vf_notok );
      } // end if
      cn20FreeParamHistory(vcontrol);
    } // end if
  } // end if

  *replyLen = (int)_tcslen(replyData);

  return nFuncReturn;
} // end cn20ParamHistoryNext

/* PUBLIC FUNCTION cn20FreeParamHistory
 * ------------------------------------------------------------------
 */
void cn20FreeParamHistory(VControlDataT * vcontrol)
{

  if (vcontrol != NULL) {
    cn52ParamHistoryRecord * pRecord = (cn52ParamHistoryRecord *) vcontrol->pParamHistory;

    if (pRecord != NULL) {
      delete pRecord;
      vcontrol->pParamHistory = NULL;
    } // end if
  } // end if

} // end cn20FreeParamHistory

/* PUBLIC FUNCTION cn20SaveParamsInfoToDB
 * ------------------------------------------------------------------
 * purpose: save param properties to database table
 */

#define DELETE_PARAMSINFO _T("DELETE DBPARAMSINFO")
#define INSERT_PARAMSINFO _T("INSERT INTO DBPARAMSINFO VALUES ('%.18s','%.5s','%.8s','%.128s','%.64s') UPDATE DUPLICATES")
//                                                              PARAMID
//                                                                   TYPE
//                                                                        GROUPSPEC
//                                                                             HELP
//                                                                                  DEFVALUE

tcn00_Error cn20SaveParamsInfoToDB
      ( const tsp00_DbNamec & szDbName )
{
  tcn00_Error         nFuncReturn   = OK_CN00;
  tcn002_XParamData * pXParamData   = NULL;
  _TCHAR              replyData[1000];
  int                 replyLen;
  int                 replyLenMax = 1000;
  // start session
  nFuncReturn = cn29ParamStartSession( szDbName, pXParamData, replyData, &replyLen);

  if (nFuncReturn == OK_CN00) {
    cn51DBMConfigValue    cfgDBAUsr ( szDbName, CFG_COLDUSR_CN51);
    cn50DBMUser           usrDBA    ( szDbName, cfgDBAUsr);
    tin01_sql_session     aSQLSession;

    nFuncReturn = cn80ConnectSQL (usrDBA, &aSQLSession);

    if (nFuncReturn == OK_CN00) {
      _TCHAR                szSqlAnswer[PARAMETER_MAXLEN_CN90];
      int                   nAnswerLength;
      tcn002_XParameter   * pParameter    = NULL;
      _TCHAR                szStatement[PARAMETER_MAXLEN_CN90];
      const _TCHAR        * pHelp         = NULL;
      _TCHAR                szHelp[129];
      tcn00_ValueToText     aTypeTable[]  = XP_TYPES_CN002;
      int                   nType         = 0;
      struct {
        tcn002_XPPropValue   nValue;
        const char         * szText;
      }                      aValueTable[] = XP_PROP_VALUES_CN002;
      int                   nValue         = 0;
      int                   i,j;

      cn80ExecuteSQL( &aSQLSession,
                      DELETE_PARAMSINFO,
                      szSqlAnswer,
                      PARAMETER_MAXLEN_CN90,
                      nAnswerLength);

      pParameter = pXParamData->pParam;
      while (pParameter != NULL && nFuncReturn == OK_CN00) {

        // help
        pHelp = (pParameter->pHelp != NULL) ? pParameter->pHelp->szLine : "";
        for (i = 0,j = 0; i < 128 && pHelp[i] != 0; ++i) {
          szHelp[j] = pHelp[i];
          if (szHelp[j] == _T('\'') ) {
            ++j;
            szHelp[j] = pHelp[i];
          } // end if
          ++j;
        } // end if
        szHelp[j] = 0;

        // type
        nType = 0;
        while ( (_tcslen(aTypeTable[nType].szText) > 0) && (pParameter->xpValueRecord.xp_value_t != aTypeTable[nType].nValue) ) {
          ++nType;
        } // while
        // group
        nValue = 0;
        while ( (aValueTable[nValue].nValue != XPPropValueUnknown                         ) && 
                (aValueTable[nValue].nValue != pParameter->xpProperties[XPPropClassGroup] )    ) {
          nValue++;
        } // end while 

        _stprintf(szStatement, INSERT_PARAMSINFO, 
                               pParameter->szParameterName,   // PARAMID
                               aTypeTable[nType].szText,      // TYPE
                               (aValueTable[nValue].szText == NULL) ? "" : aValueTable[nValue].szText,    // GROUPSPEC
                               szHelp,                        // HELP
                               pParameter->szDefault);        // DEFVALUE

        nFuncReturn = cn80ExecuteSQL( &aSQLSession,
                                      szStatement,
                                      szSqlAnswer,
                                      PARAMETER_MAXLEN_CN90,
                                      nAnswerLength);
        pParameter = pParameter->pParam;
      } // end while


      cn80ReleaseSQL(&aSQLSession);
    } // end if


    cn29FreeXParamSection(pXParamData);
  } // end if

  return nFuncReturn;
} // end cn20SaveParamsInfoToDB
