/* @lastChanged: "1998-06-05  15:29"
 * @filename:    vcn29.cpp
 * @filename:    vcn29.cpp
 * @purpose:     "param access service functions"
 * @release:     7.1. / 6.2
 * @see:         "-.-"
 *
 * @copyright:   "Copyright by SAP AG, 1998"
 *
 * ==================================================================
 *
 * responsible:   Bernd Vorsprach (D025588 - Bernd.Vorsprach@SAP-AG.de)
 *
 * created:       1998-03-04 by Bernd Vorsprach
 *
 * purpose:       service functions for all vcn2*c.cpp modules
 *
 * ==================================================================


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 */

/* ------------------------------------------------------------------
 * IMPORT                                                                    
 * ------------------------------------------------------------------
 */
#include <stdio.h>
#include <stdlib.h>


#include "heo02.h"
#include "hsp02.h"
#if defined(SAPDBR73)
  #include "geo002xp.h"
#elif defined(SAPDBR74)
  #define XPARAM_NAMES_ONLY
  #include "geo002.h"
#endif
#include "heo06.h"
#include "heo11.h"
#include "hsp100.h"

#include "hcn21.h"
#include "hcn23.h"
#include "hcn52.h"
#include "hcn80.h"
#include "hcn90.h"

#include "hcn29.h"

/* SPECIFICATION PRIVATE FUNCTION cn29_CopyLines
 * ------------------------------------------------------------------
 * purpose: copy a line section from source to dest
 */
static tcn00_Error cn29_CopyLines 
      ( tcn002_XPLine ** pDest,
        tcn002_XPLine *  pSource );

/* SPECIFICATION PRIVATE FUNCTION cn29_RequestedProperties
 * ------------------------------------------------------------------
 */
typedef tcn002_XPPropValue tReqPropArray[XPPropClassMAX][XP_MAX_VALUES_PER_CLASS_CN002];

static tcn00_Error cn29_RequestedProperties
      ( const char         * szLine,
        tReqPropArray        xpProperties);

/* ------------------------------------------------------------------
 * START OF IMPLEMENTATION
 * ------------------------------------------------------------------
 */

/* ------------------------------------------------------------------
 * IMPLEMENTATION INTERN FUNCTION cn29SearchDynamic
 * ------------------------------------------------------------------
 */
tcn002_XParameter * cn29SearchDynamic 
      ( tcn002_XParameter * pDynHelp,
        const char        * szName )
{
  tsp00_XpKeyTypec     szDynamic;
  char               * pEndDyn     = NULL;
  const char         * pEndParam   = NULL;
  bool                 bFound = false;

  /* search for dynamic match */
  bFound = false;
  while ((pDynHelp != NULL) && (bFound == false)) {
    if (pDynHelp->xpProperties[XPPropClassDynamic] == XPPropValueYes) {
      /* special check for dynamics */
      strcpy(szDynamic, pDynHelp->szParameterName);
      pEndDyn = strstr(szDynamic, XP_KEY_DYNAMIC_CN002);
      pEndDyn[0] = CHAR_STRINGTERM_CN90;
      pEndDyn++;
      pEndParam = &szName[strlen(szName) - strlen(pEndDyn)];
      if ( (strncmp(szName, szDynamic, strlen(szDynamic)) == 0) &&
           (strncmp(pEndDyn, pEndParam , strlen(pEndDyn))  == 0)    ) {
        bFound = true;
      } /* end if */
    } /* end if */
    pDynHelp = (bFound) ? pDynHelp : pDynHelp->pNext;
  } /* end while */

  return pDynHelp;
} /* end cn29SearchDynamic */

/* IMPLEMENTATION INTERN FUNCTION cn29SearchParameter
 * ------------------------------------------------------------------
 */
tcn002_XParameter * cn29SearchParameter 
      ( tcn002_XParameter * pParameter,
        const char        * szName )
{
  bool                 bFound = false;

  /* search for dynamic match */
  bFound = false;
  while ((pParameter != NULL) && (bFound == false)) {
    if (strcmp(pParameter->szParameterName, szName) == 0) {
      bFound = true;
    } else {
      pParameter = pParameter->pNext;
    } /* end if */
  } /* end while */

  return pParameter;
} /* end cn29SearchParameter */

/* IMPLEMENTATION INTERN FUNCTION cn29CopyParameter
 * ------------------------------------------------------------------
 */
tcn00_Error cn29CopyParameter 
      ( tcn002_XParameter * pDest,
        tcn002_XParameter * pSource )
{
  tcn00_Error          nFuncStatus  = OK_CN00;
  long                 i            = 0;

  strcpy(pDest-> szParameterName, pSource->szParameterName);
  strcpy(pDest->szDefault,        pSource->szDefault);
  memcpy(&(pDest->xpValueRecord), &(pSource->xpValueRecord), sizeof(tsp00_XpValueRec));
  pDest->bHelpExist     = pSource->bHelpExist;
  pDest->bParamExist    = pSource->bParamExist;

  pDest->nProperties    = pSource->nProperties;

  for (i = 0; i < XPPropClassMAX; i++) {
    pDest->xpProperties[i] = pSource->xpProperties[i];
  } /* end for */

  pDest->pNext          = NULL;
  pDest->pPrev          = NULL;
  pDest->pParam         = NULL;

  /* lines */
  nFuncStatus = cn29_CopyLines(&(pDest->pCode), pSource->pCode);
  if (nFuncStatus == OK_CN00) {
    nFuncStatus = cn29_CopyLines(&(pDest->pHelp), pSource->pHelp);
  } /* end if */
  if (nFuncStatus == OK_CN00) {
    nFuncStatus = cn29_CopyLines(&(pDest->pConstraint), pSource->pConstraint);
  } /* end if */
  if (nFuncStatus == OK_CN00) {
    nFuncStatus = cn29_CopyLines(&(pDest->pExplain), pSource->pExplain);
  } /* end if */

  return nFuncStatus;
} /* end cn29CopyParameter */

/* IMPLEMENTATION INTERN FUNCTION cn29_RequestedProperties
 * ------------------------------------------------------------------
 */
static tcn00_Error cn29_RequestedProperties
      ( const char         * szLine,
        tReqPropArray        xpProperties)
{
  tcn00_Error           nFuncStatus   = OK_CN00;
  int                   nClassIndex   = 0;
  int                   nValueIndex   = 0;
  int                   nOrIndex      = 0;
  char                * pClass        = NULL;
  char                * pValue        = NULL;
  char                * pOneValue     = NULL;
  long                  nToken        = 1;
  char                  szProperty[XP_PARAMLINE_LENGTH_MXCN002];
  struct {
    tcn002_XPPropValue   nValue;
    const char        * szText;
  }                     aValueTable[] = XP_PROP_VALUES_CN002;
  struct tcn002_PropClassTable {
    tcn002_XPPropClass   nClass;
    tcn002_XPPropValue   nDefault;
    const char        * szText;
  }                     aClassTable[] = XP_PROP_CLASSES_CN002;

  /* init array */
  for (nClassIndex = 0; nClassIndex < XPPropClassMAX; nClassIndex ++) {
    for (nOrIndex = 0; nOrIndex < XP_MAX_VALUES_PER_CLASS_CN002; nOrIndex ++) {
      xpProperties[nClassIndex][nOrIndex] = XPPropValueUnknown;
    } /* end for */
  } /* end for */

  while (cn90GetToken(szLine, szProperty, nToken++) && (nFuncStatus == OK_CN00)) {
    pClass = strtok(szProperty, "=");
    pValue = strtok(NULL, "=");
    pOneValue = strtok(pValue, ",");
    
    if ((pClass != NULL) && (pOneValue != NULL) ) {

      nOrIndex = 0;
      while (pOneValue != NULL && nFuncStatus == OK_CN00) {
        nClassIndex=0;
        nValueIndex=0;
        /* search class */
        while ( (aClassTable[nClassIndex].nClass != XPPropClassUnknown) && (strcmp(aClassTable[nClassIndex].szText, pClass) != 0) ) {
          nClassIndex++;
        } /* end while */
        /* search Value */
        while ( (aValueTable[nValueIndex].nValue != XPPropValueUnknown) && (strcmp(aValueTable[nValueIndex].szText, pOneValue) != 0) ) {
          nValueIndex++;
        } /* end while */

        if ((aValueTable[nValueIndex].nValue != XPPropValueUnknown) && (aClassTable[nClassIndex].nClass != XPPropClassUnknown) &&  nOrIndex < XP_MAX_VALUES_PER_CLASS_CN002) {
          xpProperties[aClassTable[nClassIndex].nClass][nOrIndex] = aValueTable[nValueIndex].nValue;
        } else {
          nFuncStatus =  ERR_PARAM_CN00;
        } /* end if */

        nOrIndex++;
        pOneValue = strtok(NULL, ",");
      } // end while

    } else {
      nFuncStatus =   ERR_PARAM_CN00;
    } /* end if */

  } /* end while */

  return nFuncStatus;
} /* end cn29_RequestedProperties */

/* IMPLEMENTATION INTERN FUNCTION cn29PrintProperties
 * ------------------------------------------------------------------
 */
void cn29PrintProperties 
      ( char               * pBuffer,
        tcn002_XPPropClass   nClass, 
        tcn002_XPPropValue   nValue,
        tsp00_C70c           szValue)
{
  struct {
    tcn002_XPPropValue   nValue;
    const char         * szText;
  }                     aValueTable[] = XP_PROP_VALUES_CN002;
  struct tcn002_PropClassTable {
    tcn002_XPPropClass   nClass;
    tcn002_XPPropValue   nDefault;
    const char         * szText;
  }                     aClassTable[] = XP_PROP_CLASSES_CN002;
  int                   j                 = 0;

  if (nClass != XPPropClassUnknown) {
    /* search value */
    while ( (aValueTable[j].nValue != XPPropValueUnknown) && (aValueTable[j].nValue != nValue) ) {
      j++;
    } /* end while */
    if (aValueTable[j].nValue == XPPropValueUnknown) {
      sprintf(pBuffer, "%s%s%s%s", aClassTable[nClass].szText, VALUE_SEPSTRING_CN00, (char *) szValue, LINE_SEPSTRING_CN00);
    } else {
      sprintf(pBuffer, "%s%s%s%s", aClassTable[nClass].szText, VALUE_SEPSTRING_CN00, aValueTable[j].szText, LINE_SEPSTRING_CN00);
    } // end if
  } /* end if */

} /* end  cn29PrintProperties */

/* IMPLEMENTATION INTERN FUNCTION cn29PrintExtendedGet
 * ------------------------------------------------------------------
 */
void cn29PrintExtendedGet
      ( char              * pBuffer,
        tcn002_XParameter * pParameter )
{
  tcn00_ValueToText  aTable[]    = XP_TYPES_CN002;
  long               i           = 0;
  tsp00_XpKeyTypec   aDummy;

  /* search typ */
  while ( (strlen(aTable[i].szText) > 0) && (pParameter->xpValueRecord.xp_value_t != aTable[i].nValue) ) {
    i++;
  } /* end while */

  // print param
  sprintf(pBuffer,
          "%-*s%s%s%s%s%s",
          aDummy.size() + 1,
          pParameter->szParameterName,
          VALUE_SEPSTRING_CN00,
          aTable[i].szText,
          VALUE_SEPSTRING_CN00,
          (pParameter->xpValueRecord.xp_value_t != xp_crypt_type) ?
           pParameter->szInput : "",
          LINE_SEPSTRING_CN00);

} /* end cn29PrintExtendedGet */

/* IMPLEMENTATION PRIVATE FUNCTION cn29_CopyLines
 * ------------------------------------------------------------------
 */
static tcn00_Error cn29_CopyLines 
      ( tcn002_XPLine ** pDest,
        tcn002_XPLine *  pSource )
{
  tcn00_Error           nFuncStatus  = OK_CN00;
  tcn002_XPLine      *   pNewLine    = NULL;
  tcn002_XPLine      *   pLastLine   = NULL;
  char              *   pLineText   = NULL;
  bool                  bMallocOK   = false;

  /* code lines */
  while (pSource && (nFuncStatus == OK_CN00)) {

    cn00MyMalloc (sizeof (tcn002_XPLine), &pNewLine, &bMallocOK);
    if ((pNewLine == NULL) || (!bMallocOK)) {
      nFuncStatus = ERR_MEM_CN00;
    } else {
      cn00MyMalloc (strlen(pSource->szLine) + 1, &pLineText, &bMallocOK);
      if ((pLineText == NULL) || (!bMallocOK)) {

        nFuncStatus = ERR_MEM_CN00;
        cn00MyFree(pNewLine);

      } else {

        pNewLine->szLine  = pLineText;
        pNewLine->xpLine  = pSource->xpLine;
        pNewLine->nNumber = pSource->nNumber;
        strcpy(pNewLine->szLine, pSource->szLine);

        if (*pDest == NULL) {
          *pDest = pNewLine;
        } else {
          pLastLine->pNext = pNewLine;
        } /* end if */
        pLastLine = pNewLine;

        pSource = pSource->pNext; 

      } /* end if */
    } /* end if */
  } /* end while */

  return nFuncStatus;
} /* end cn29_CopyLines */

/* IMPLEMENATION INTERN FUNCTION cn29SyntaxError
 * ------------------------------------------------------------------
 */
void cn29SyntaxError(
        char          * pData,
        int           * pnDataLen,
        tcn002_XPError   nSyntaxError,
        char          * szBuffer,
        const char    * szFile,
        long            nLine)
{
  int i = 0;
  tcn00_ValueToText errorArray[] = XPERROR_TO_TEXT_CN002;

  while ((errorArray[i].nValue !=  XPERR_UNKNOWN_CN002) && (errorArray[i].nValue !=  nSyntaxError)) {
    i++;
  } /* end while */

  sprintf (&pData[strlen(pData)], "%d,%s\n%s(%ld) : %s\n", nSyntaxError,
                                                           errorArray[i].szText,
                                                           szFile,
                                                           nLine,
                                                           szBuffer);
  *pnDataLen = (int)strlen(pData);

} /* end cn29SyntaxError */

/* IMPLEMENTATION INTERN FUNCTION cn29StringToParamValue
 * ------------------------------------------------------------------
 * purpose: copy a zero terminated string in the param value record
 *          (for strings use pascal representation)
 */
tsp00_XpValueRec * cn29StringToParamValue
    ( tsp00_XpValueRec  * pValue,
      const char    * szValue )
{
  char          szTemp[PARAMETER_MAXLEN_CN90];
  tsp00_Name    szName;       /* char [18] */
  tsp00_CryptPw cryptPwTemp;  /* char [24] */

  /* initialize the whole memory of the union with spaces */
  memset(&(pValue->variant), CHAR_PASCALFILL_CN90, sizeof(pValue->variant));
 
  switch ( pValue->xp_value_t ) {
    case xp_int2_type   :
      pValue->xp_int2() = atoi ( szValue );
      break ;
    case xp_int4_type   :
      pValue->xp_int4() = atol ( szValue );
      break ;
    case xp_crypt_type  :
      strcpy(szTemp, szValue);
      if (strlen(szTemp) == sizeof(tsp_cryptname) * 2) {
        cn90Uncrypt(szTemp, false);
      } /* end if */
      cn90StringCtoP(szName, szTemp, sizeof(tsp_name));
      s02encrypt(szName, cryptPwTemp);
      memcpy(&pValue->xp_crypt()[0], &cryptPwTemp[0], sizeof(cryptPwTemp));
      break;
    case xp_c8_type     :
      cn90StringCtoP(pValue->xp_c8(), szValue, sizeof(tsp_c8));
      break ;
    case xp_c18_type    :
      cn90StringCtoP(pValue->xp_c18(), szValue, sizeof(tsp_c18));
      break ;
    case xp_c24_type    :
      cn90StringCtoP(pValue->xp_c24(), szValue, sizeof(tsp_c24));
      break ;
    case xp_c40_type    :
      cn90StringCtoP(pValue->xp_c40(), szValue, sizeof(tsp_c40));
      break ;
    case xp_c64_type    :
      cn90StringCtoP(pValue->xp_c64(), szValue, sizeof(tsp_c64));
      break ;
  } /* end switch */

  return pValue;
} /* end cn90StringToParamValue */

/* IMPLEMENTATION INTERN FUNCTION cn29ParamValueToString
 * ------------------------------------------------------------------
 */
char * cn29ParamValueToString
    ( char               * szValue,
      const tsp00_XpValueRec * pValue )
{
  tsp00_Namec   szClearPwd;
  tsp00_C64c    szCryptPwd;

  switch ( pValue->xp_value_t ) {
    case xp_int2_type   :
      sprintf ( szValue , "%hd" , pValue->xp_int2() );
      break ;
    case xp_int4_type   :
      sprintf ( szValue , "%ld" , pValue->xp_int4() );
      break ;
    case xp_crypt_type  :
      cn90CalculateUncrypt(pValue->xp_crypt(), szClearPwd, true);
      cn90ClearStringToCryptString(szClearPwd, szCryptPwd);
      strcpy(szValue, szCryptPwd);
      break;
    case xp_c8_type     :
      cn90StringPtoC(szValue, (char*) &pValue->xp_c8(), sizeof(tsp_c8));
      break;
    case xp_c18_type    :
      cn90StringPtoC(szValue, (char*) &pValue->xp_c18(), sizeof(tsp_c18));
      break ;
    case xp_c24_type    :
      cn90StringPtoC(szValue, (char*) &pValue->xp_c24(), sizeof(tsp_c24));
      break ;
    case xp_c40_type    :
      cn90StringPtoC(szValue, (char*) &pValue->xp_c40(), sizeof(tsp_c40));
      break ;
    case xp_c64_type    :
      cn90StringPtoC(szValue, (char*) &pValue->xp_c64(), sizeof(tsp_c64));
      break ;
  } /* end switch */

  return szValue;
} /* end cn29ParamValueToString */

/* IMPLEMENTATION INTERN FUNCTION cn29ParamGetAllIntern
 * ------------------------------------------------------------------
 * purpose: get all parameters with values from param file
 *
 */
tcn00_Error cn29ParamGetAllIntern
      ( tsp00_DbNamec       szDbName,
        tcn002_XParamData * pXParamData,
        const _TCHAR      * szCommand,
        char              * replyData,
        int               * replyLen,
        const int           replyLenMax,
        const bool          bExtended,
        const bool          bNext,
        tsp00_XpKeyTypec  & szLastParameter)

{
  tcn00_Error          nFuncReturn = OK_CN00;
  tcn002_XParameter  * pParameter  = NULL;
  bool                 bInit = false;
  char               * pCurrPos;
  char                 szValue[XP_PARAMLINE_LENGTH_MXCN002];
  tReqPropArray        xpProperties;
  bool                 bMatch      = true;
  bool                 bOrMatch    = false;
  long                 nClassIndex = 0;
  long                 nOrIndex    = 0;
  tsp00_XpKeyTypec     aDummy;
    
  /* init param management for commandline use or tcontrol or rcontrol */
  if (pXParamData == NULL) {
    nFuncReturn = cn29ParamStartSession( szDbName, pXParamData, replyData, replyLen);
    bInit = true;
  } /* end if */

  if (nFuncReturn == OK_CN00) {

    if (bExtended) {
      nFuncReturn = cn90AnswerIError(replyData, replyLen, cn29_RequestedProperties(szCommand, xpProperties));
    } /* end if */

    if (nFuncReturn == OK_CN00) {
      /* Output */
      sprintf(replyData, "%s%s", ANSWER_OK_CN00, LINE_SEPSTRING_CN00);
      pCurrPos = replyData + strlen (replyData);

      /* in extended mode order by help file, otherwise by param file */
      pParameter = (bExtended) ? pXParamData->pFirst : pXParamData->pParam;

      if (bNext) {
        if (szLastParameter.length() > 0) {
          while (pParameter != NULL) {
            if (strcmp(szLastParameter, pParameter->szParameterName) == 0) {
              break;
            } // end if
            pParameter = (bExtended) ? pParameter->pNext : pParameter->pParam;
          } // end while
        } else {
          pParameter = NULL;
        } // end if
      } // end if
    } /* end if */

    while ((pParameter != NULL) && (nFuncReturn == OK_CN00)) {

      if (bExtended) {
        /* check propeties */
        bMatch = true;
        for (nClassIndex = 0; nClassIndex < XPPropClassMAX; nClassIndex ++) {
          if (xpProperties[nClassIndex][0] != XPPropValueUnknown) {
            bOrMatch = false;
            for (nOrIndex = 0; nOrIndex < XP_MAX_VALUES_PER_CLASS_CN002; nOrIndex++) {
              if (xpProperties[nClassIndex][nOrIndex] != XPPropValueUnknown) {
                bOrMatch = bOrMatch || xpProperties[nClassIndex][nOrIndex] == pParameter->xpProperties[nClassIndex];
              } // end if
            } // end for
            bMatch = bMatch && bOrMatch;
          } // end if
        } // end for

        /* print only requested parameters */
        if ( bMatch) {
          if ((int)(strlen(replyData) + (XP_PARAMLINE_LENGTH_MXCN002 * 2)) > replyLenMax) {
            /* so sorry, buffer too small */
//            nFuncReturn = ERR_TOOSMALL_CN00;
//            cn90AnswerIError(replyData, replyLen, nFuncReturn);
            szLastParameter.rawAssign(pParameter->szParameterName);
            break;
          } else {
            cn29PrintExtendedGet(pCurrPos, pParameter);
            pCurrPos = replyData + strlen (replyData);
          } /* end if */
        } /* end if */

      } else {
        /* in normal mode print only param parameter */
        if (pParameter->bParamExist) {
          if ((int)(strlen(replyData) + XP_PARAMLINE_LENGTH_MXCN002) > replyLenMax) {
//            /* so sorry, buffer too small */
//            nFuncReturn = ERR_TOOSMALL_CN00;
//            cn90AnswerIError(replyData, replyLen, nFuncReturn);
            szLastParameter.rawAssign(pParameter->szParameterName);
            break;
          } else {
            sprintf(pCurrPos,
                    "%-*s%s%s%s",
                    aDummy.size() + 1,
                    pParameter->szParameterName,
                    VALUE_SEPSTRING_CN00,
                    (pParameter->xpValueRecord.xp_value_t != xp_crypt_type) ?
                      cn29ParamValueToString(szValue, &(pParameter->xpValueRecord)) :
                      "",
                    LINE_SEPSTRING_CN00);
            pCurrPos = replyData + strlen (replyData);
          } /* end if */
        } /* end if */
      } /* end if */

      /* in extended mode order by help file, otherwise by param file */
      pParameter = (bExtended) ? pParameter->pNext : pParameter->pParam;

    } /* end while */
    if (pParameter == NULL) {
      szLastParameter.rawAssign("");
    } // end if

    *replyLen = (int)strlen(replyData);
  } /* end if */

  /* if init then stop param management */
  if (bInit) {
    cn29FreeXParamSection(pXParamData);
  } /* end if */

  return nFuncReturn;
} /* end cn29ParamGetAllIntern */

/* IMPLEMENTATION INTERN FUNCTION cn29ParamGetIntern
 * ------------------------------------------------------------------
 * purpose: get the specified data about the parameter
 */
tcn00_Error cn29ParamGetIntern
      ( tsp00_DbNamec       szDbName,
        tcn002_XParamData * pXParamData,
        const char        * szParameter,
        char              * replyData,
        int               * replyLen,
        tcn002_XPRequest nRequest )
{
  tcn00_Error          nFuncReturn = OK_CN00;
  tcn002_XParameter   * pParameter  = NULL;
  tcn002_XParameter   * pDynamic    = NULL;
  tcn002_XParameter     aTmpParam;
  bool                 bInit  = false;
  char                 szValue[XP_PARAMLINE_LENGTH_MXCN002];
  char               * pCurrPos;
  tcn00_ValueToText    aTable[]    = XP_TYPES_CN002;
  long                 i           = 0;
  tcn002_XPLine       * pLine       = NULL;

  /* init param management for commandline use or tcontrol or rcontrol */
  if (pXParamData == NULL) {
    nFuncReturn = cn29ParamStartSession( szDbName, pXParamData, replyData, replyLen);
    bInit = true;
  } /* end if */

  if (nFuncReturn == OK_CN00) {

    if (strlen(szParameter) > 0) {

      /* search parameter in List */
      pParameter = cn29SearchParameter(pXParamData->pFirst, szParameter);

      if (pParameter == NULL) {
        pDynamic = cn29SearchDynamic(pXParamData->pFirst, szParameter);

        /* is there an dynamic then generate tmp param */
        if (pDynamic != NULL) {
          pParameter = &aTmpParam;
          memset(pParameter, 0, sizeof(tcn002_XParameter));

          /* copy data from dynamic record */
          nFuncReturn = cn29CopyParameter(pParameter, pDynamic);
          cn90AnswerIError(replyData, replyLen, nFuncReturn);

          /* clear the dynamic/intern property */
          pParameter->xpProperties[XPPropClassDynamic] = XPPropValueNo;
          pParameter->xpProperties[XPPropClassIntern] = XPPropValueNo;
          /* set the real name */
          cn90SaveStrcpy(pParameter->szParameterName, szParameter, sizeof(tsp00_XpKeyTypec));
        } /* end if */
      } /* end if */

      /* copy match to reply buffer */
      if (pParameter != NULL) {
        /* Output OK */
        sprintf(replyData, "%s%s", ANSWER_OK_CN00, LINE_SEPSTRING_CN00);
        pCurrPos = replyData + strlen (replyData);
 
        /* Output Requested Data */
        switch (nRequest) {
          case XPRequestDirectGet:
            /* in normal mode print only param-parameter */
            if (pParameter->bParamExist) {
              sprintf(pCurrPos, 
                      "%s%s%s%s", 
                      pParameter->szParameterName, 
                      VALUE_SEPSTRING_CN00,  
                      (pParameter->xpValueRecord.xp_value_t != xp_crypt_type) ?
                        cn29ParamValueToString(szValue, &(pParameter->xpValueRecord)) :
                        "",
                      LINE_SEPSTRING_CN00);
            } else {
              nFuncReturn = cn29AnswerParamNotFound(replyData, replyLen, szParameter, pXParamData->pParam, true);
            } /* end if */
            break;

          case XPRequestExtGet:
            /* in extended */
            cn29PrintExtendedGet(pCurrPos, pParameter);
            break;

          case XPRequestType:
            /* type */
            while ( (strlen(aTable[i].szText) > 0) && (pParameter->xpValueRecord.xp_value_t != aTable[i].nValue) ) {
              i++;
            } /* end while */
            sprintf(pCurrPos, "%s%s", aTable[i].szText, LINE_SEPSTRING_CN00);
            break;

          case XPRequestDefault:
            /* Default */
            sprintf(pCurrPos, "%s%s", pParameter->szDefault, LINE_SEPSTRING_CN00);
            break;

          case XPRequestValue:
            /* value */
            if (pParameter->xpValueRecord.xp_value_t != xp_crypt_type) {
              sprintf(pCurrPos, "%s%s", pParameter->szInput, LINE_SEPSTRING_CN00);
            } else {
              sprintf(pCurrPos, "%s", LINE_SEPSTRING_CN00);
            } /* end if */
            break;

          case XPRequestValueIntern:
            /* value */
            sprintf(pCurrPos, "%s", pParameter->szInput, LINE_SEPSTRING_CN00);
            if (pParameter->xpValueRecord.xp_value_t == xp_crypt_type) {
              cn90Uncrypt(pCurrPos, false);
            } /* end if */
            sprintf(pCurrPos, "%s%s", pCurrPos, LINE_SEPSTRING_CN00);
            break;

          case XPRequestProperties:
            /* properties */
            for (i = 0; i < XPPropClassMAX; i++) {
              cn29PrintProperties(pCurrPos, (tcn002_XPPropClass) i, pParameter->xpProperties[i], pParameter->szProperties[i]);
              pCurrPos = replyData + strlen (replyData);
            } /* end if */
            break;

          case  XPRequestGood:
            /* good value */
            sprintf(pCurrPos, "%s%s%s%s", XP_KEY_GOOD_CN002, VALUE_SEPSTRING_CN00, (char *) pParameter->szGood, LINE_SEPSTRING_CN00);
            pCurrPos = replyData + strlen (replyData);
            break;

          case XPRequestHelp:
            /* help */
            pLine = pParameter->pHelp;
            while (pLine != NULL) {
              sprintf(pCurrPos, "%s%s", pLine->szLine, LINE_SEPSTRING_CN00);
              pCurrPos = replyData + strlen (replyData);
              pLine = pLine->pNext;
            } /* end while */
            break;
 
          case XPRequestExplain:
            /* explain */
            pLine = pParameter->pExplain;
            while (pLine != NULL) {
              sprintf(pCurrPos, "%s%s", pLine->szLine, LINE_SEPSTRING_CN00);
              pCurrPos = replyData + strlen (replyData);
              pLine = pLine->pNext;
            } /* end while */
            break;

          case XPRequestFull:
            /* type */
            while ( (strlen(aTable[i].szText) > 0) && (pParameter->xpValueRecord.xp_value_t != aTable[i].nValue) ) {
              i++;
            } /* end while */
            sprintf(pCurrPos, "%s%s", aTable[i].szText, LINE_SEPSTRING_CN00);
            pCurrPos = replyData + strlen (replyData);

            /* Default */
            sprintf(pCurrPos, "%s%s", pParameter->szDefault, LINE_SEPSTRING_CN00);
            pCurrPos = replyData + strlen (replyData);

            /* value */
            if (pParameter->xpValueRecord.xp_value_t != xp_crypt_type) {
              sprintf(pCurrPos, "%s%s", pParameter->szInput, LINE_SEPSTRING_CN00);
            } else {
              sprintf(pCurrPos, "%s", LINE_SEPSTRING_CN00);
            } /* end if */
            pCurrPos = replyData + strlen (replyData);

            /* properties */
            for (i = 0; i < XPPropClassMAX; i++) {
              cn29PrintProperties(pCurrPos, (tcn002_XPPropClass) i, pParameter->xpProperties[i], pParameter->szProperties[i]);
              pCurrPos = replyData + strlen (replyData);
            } /* end if */
            pCurrPos = replyData + strlen (replyData);

            /* good value */
            sprintf(pCurrPos, "%s%s%s%s", XP_KEY_GOOD_CN002, VALUE_SEPSTRING_CN00, (char *) pParameter->szGood, LINE_SEPSTRING_CN00);
            pCurrPos = replyData + strlen (replyData);

            /* help */
            sprintf(pCurrPos, "%s%s", XP_KEY_HELP_CN002, LINE_SEPSTRING_CN00);
            pCurrPos = replyData + strlen (replyData);

            pLine = pParameter->pHelp;
            while (pLine != NULL) {
              sprintf(pCurrPos, "%s%s", pLine->szLine, LINE_SEPSTRING_CN00);
              pCurrPos = replyData + strlen (replyData);
              pLine = pLine->pNext;
            } /* end while */
            pCurrPos = replyData + strlen (replyData);
 
            /* explain */
            sprintf(pCurrPos, "%s%s", XP_KEY_EXPLAIN_CN002, LINE_SEPSTRING_CN00);
            pCurrPos = replyData + strlen (replyData);

            pLine = pParameter->pExplain;
            while (pLine != NULL) {
              sprintf(pCurrPos, "%s%s", pLine->szLine, LINE_SEPSTRING_CN00);
              pCurrPos = replyData + strlen (replyData);
              pLine = pLine->pNext;
            } /* end while */
            pCurrPos = replyData + strlen (replyData);
            break;

         } /* end switch */
      } else {
        nFuncReturn = cn29AnswerParamNotFound(replyData, replyLen, szParameter, pXParamData->pFirst, (nRequest == XPRequestDirectGet));
      } /* end if */

    } else {
      nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
    } /* end if */

  } /* end if */

  /* if init then stop param management */
  if (bInit) {
    cn29FreeXParamSection(pXParamData);
  } /* end if */

  *replyLen = (int)strlen(replyData);

  return nFuncReturn;
} /* end cn29ParamGetIntern */

/* IMPLEMENTATION INTERN FUNCTION cn29CheckRange
 * ------------------------------------------------------------------
 */
tcn00_Error cn29CheckRange
      ( tcn002_XParameter * pParameter,
        const char        * szValue )
{
  tcn00_Error        nFuncReturn   = OK_CN00;
  char               szBuffer[XP_STRING_LENGTH_MXCN002];
  char               szToken[XP_STRING_LENGTH_MXCN002];
  int                nToken;
  int                nIndex;

  // check Valueset
  if (nFuncReturn == OK_CN00) {
    if (strlen(pParameter->szProperties[XPPropClassValueset]) > 0) {

      nFuncReturn = ERR_XPVALUESET_CN00;
      nToken = 1;
      while (cn90GetToken(pParameter->szProperties[XPPropClassValueset], szToken, nToken)) {
        if (pParameter->xpValueRecord.xp_value_t == xp_int2_type ||
            pParameter->xpValueRecord.xp_value_t == xp_int4_type    ) {
          strcpy(szBuffer, szToken);
        } else {
          cn90AnalyzeString(szToken, szBuffer, XP_STRING_LENGTH_MXCN002);
        } // end if
        if (pParameter->xpProperties[XPPropClassCasesensitive] == XPPropValueYes) {
          if (_tcscmp(szBuffer, szValue) == 0) {
            nFuncReturn = OK_CN00;
          } // end if
        } else {
          if (_tcsicmp(szBuffer, szValue) == 0) {
            nFuncReturn = OK_CN00;
          } // end if
        } // end if

        nToken++;
      } // end while
    } // end if
  } // end if

  // check maximum
  if (nFuncReturn == OK_CN00) {
    if (strlen(pParameter->szProperties[XPPropClassMax]) > 0) {
      switch ( pParameter->xpValueRecord.xp_value_t ) {
        case xp_int2_type   :
        case xp_int4_type   :
          nFuncReturn = atol(szValue) > atol(pParameter->szProperties[XPPropClassMax]) ? ERR_XPMAX_CN00 : OK_CN00;
          break;
        case xp_c8_type     :
        case xp_c18_type    :
        case xp_c24_type    :
        case xp_c40_type    :
        case xp_c64_type    :
          cn90AnalyzeString(pParameter->szProperties[XPPropClassMax], szBuffer, XP_STRING_LENGTH_MXCN002);
          if (pParameter->xpProperties[XPPropClassCasesensitive] == XPPropValueYes) {
            nFuncReturn = _tcscmp(szValue, szBuffer) > 0 ? ERR_XPMAX_CN00 : OK_CN00;
          } else {
            nFuncReturn = _tcsicmp(szValue, szBuffer) > 0 ? ERR_XPMAX_CN00 : OK_CN00;
          } // end if
          break;
      } /* end switch */
    } // end if
  } // end if

  // check minimum
  if (nFuncReturn == OK_CN00) {
    if (strlen(pParameter->szProperties[XPPropClassMin]) > 0) {
      switch ( pParameter->xpValueRecord.xp_value_t ) {
        case xp_int2_type   :
        case xp_int4_type   :
          nFuncReturn = atol(szValue) < atol(pParameter->szProperties[XPPropClassMin]) ? ERR_XPMIN_CN00 : OK_CN00;
          break;
        case xp_c8_type     :
        case xp_c18_type    :
        case xp_c24_type    :
        case xp_c40_type    :
        case xp_c64_type    :
          cn90AnalyzeString(pParameter->szProperties[XPPropClassMin], szBuffer, XP_STRING_LENGTH_MXCN002);
          if (pParameter->xpProperties[XPPropClassCasesensitive] == XPPropValueYes) {
            nFuncReturn = _tcscmp(szValue, szBuffer) < 0 ? ERR_XPMIN_CN00 : OK_CN00;
          } else {
            nFuncReturn = _tcsicmp(szValue, szBuffer) < 0 ? ERR_XPMIN_CN00 : OK_CN00;
          } // end if
          break;
      } /* end switch */
    } // end if
  } // end if

  // check range of datatype
  if (nFuncReturn == OK_CN00) {
    switch ( pParameter->xpValueRecord.xp_value_t ) {
      case xp_int2_type   :
        sprintf ( szBuffer , "%hd" , atoi(szValue) );
        nFuncReturn = (strcmp(szBuffer, szValue) == 0) ? nFuncReturn : ERR_XPOVERFLOW_CN00;
        break ;
      case xp_int4_type   :
        sprintf ( szBuffer , "%ld" , atol(szValue) );
        nFuncReturn = (strcmp(szBuffer, szValue) == 0) ? nFuncReturn : ERR_XPOVERFLOW_CN00;
        break ;

      case xp_crypt_type  :
        break;

      case xp_c8_type     :
        nFuncReturn = (sizeof(tsp_c8) < strlen(szValue)) ? ERR_XPOVERFLOW_CN00 : nFuncReturn;
        break ;

      case xp_c18_type    :
        nFuncReturn = (sizeof(tsp_c18) < strlen(szValue)) ? ERR_XPOVERFLOW_CN00 : nFuncReturn;
        break ;

      case xp_c24_type    :
        nFuncReturn = (sizeof(tsp_c24) < strlen(szValue)) ? ERR_XPOVERFLOW_CN00 : nFuncReturn;
        break ;

      case xp_c40_type    :
        nFuncReturn = (sizeof(tsp_c40) < strlen(szValue)) ? ERR_XPOVERFLOW_CN00 : nFuncReturn;
        break ;

      case xp_c64_type    :
        nFuncReturn = (sizeof(tsp_c64) < strlen(szValue)) ? ERR_XPOVERFLOW_CN00 : nFuncReturn;
        break ;
 
    } /* end switch */
  } // end if

  // check 7-Bit ASCII
  if (nFuncReturn == OK_CN00) {
    switch ( pParameter->xpValueRecord.xp_value_t ) {
      case xp_crypt_type  :
      case xp_c8_type     :
      case xp_c18_type    :
      case xp_c24_type    :
      case xp_c40_type    :
      case xp_c64_type    :
        for (nIndex = 0; szValue[nIndex] != 0; ++nIndex) {
          nFuncReturn = (((unsigned char) szValue[nIndex]) < 128) ?  nFuncReturn : ERR_ASCII_CN00;
        } // end if

        break;
    } /* end switch */
  } // end if

  return nFuncReturn;
} /* cn29CheckRange */

/* IMPLEMENTATION INTERN FUNCTION cn29SoftPutIntern
 * ------------------------------------------------------------------
 * purpose: save the new value internal param session structure
 *          - no check 
 *          - no persitent write to param file
 */
tcn00_Error cn29SoftPutIntern
      ( tcn002_XParamData  * pXParamData,
        const char         * szParameter,
        const char         * szValueIn,
        bool                 bIgnoreReadOnly,
        bool                 bOverride)
{
  tcn00_Error        nFuncReturn   = OK_CN00;
  tcn002_XParameter * pParameter    = NULL;
  tcn002_XParameter * pParamChanged = NULL;
  tcn002_XParameter * pParamVerified= NULL;
  tcn002_XParameter * pDynHelp      = NULL;
  bool               bMallocOK     = false;
  char               szValue[XP_STRING_LENGTH_MXCN002 +1 ];

  if (strlen(szParameter) > 0) {

    /* search parameter in List */
    pParameter = cn29SearchParameter(pXParamData->pFirst, szParameter);

    if (pParameter == NULL) {
      pDynHelp = cn29SearchDynamic(pXParamData->pFirst, szParameter);

      /* is there an dynamic match then create new instance */
      if (pDynHelp != NULL) {
        cn00MyMalloc (sizeof (tcn002_XParameter), &pParameter, &bMallocOK);
        if ((pParameter == NULL) || (!bMallocOK)) {
          nFuncReturn = ERR_MEM_CN00;
        } else {
          /* add the parameter in the list of parameters in the XParam section */
          if (pXParamData->pFirst == NULL) {
            pXParamData->pFirst = pParameter;
          } /* end if */
          if (pXParamData->pLast != NULL) {
            pXParamData->pLast->pNext = pParameter;
          } /* end if */

          /* copy data from dynamic record */
          nFuncReturn = cn29CopyParameter(pParameter, pDynHelp);
          /* clear the dynamic/intern property */
          pParameter->xpProperties[XPPropClassDynamic] = XPPropValueNo;
          pParameter->xpProperties[XPPropClassIntern] = XPPropValueNo;
          /* set the real name */
          cn90SaveStrcpy(pParameter->szParameterName, szParameter, sizeof(tsp00_XpKeyTypec));

          /* list management */
          pParameter->pNext = NULL;
          pParameter->pPrev = pXParamData->pLast;
          pXParamData->pLast = pParameter;
        } /* end if */
      } /* end if */
    } /* end if */

    /* parameter found? */
    if (pParameter != NULL) {

      if (pParameter->xpValueRecord.xp_value_t == xp_int2_type ||
          pParameter->xpValueRecord.xp_value_t == xp_int4_type    ) {
        strcpy(szValue, szValueIn);
      } else {
        if (!cn90AnalyzeString(szValueIn, szValue, XP_STRING_LENGTH_MXCN002)) {
          strcpy(szValue, szValueIn);
        } // end if
      } // end if

      /* check if write allowed */
      if (pParameter->xpProperties[XPPropClassChange] == XPPropValueNo && !bIgnoreReadOnly) {
        /* never writes read-only */
        nFuncReturn = ERR_XPREADONLY_CN00;
      } // end if
      if (nFuncReturn == OK_CN00) {
        pParamVerified = cn29SearchParameter(pXParamData->pFirst, XP_PARAM_VERSION_TIME);
        if (pParamVerified != NULL) {
          if ((pParamVerified->bParamExist)                                &&
            (pParameter->xpProperties[XPPropClassModify] == XPPropValueNo) && 
            !bIgnoreReadOnly) {
            nFuncReturn = ERR_XPREADONLY_CN00;
          } // end if
        } // end if
      } // end if

      if (nFuncReturn == OK_CN00) {
        pParamChanged = cn29SearchParameter(pXParamData->pFirst, XP_PARAM_CHANGED);
        if (pParamChanged != NULL) {
          if (pParamChanged->bParamExist) {
            if (pParameter->xpProperties[XPPropClassDevspace] == XPPropValueYes) {
//              /* don't change devspace parameters if kernel parameters dirty */
//              if ( (atol(pParamChanged->szInput) != 0) || pXParamData->bKernelDirty) {
//                nFuncReturn = ERR_XPDIRTY_CN00;
//              } /* end if */
            } else if (pParameter->xpProperties[XPPropClassDevspace] == XPPropValueNo) {
              /* don't change kernel parameters if devspace parameters dirty */
              if (pXParamData->bDevspaceDirty) {
                nFuncReturn = ERR_XPDIRTY_CN00;
              } /* end if */
            } /* end if */
          } /* end if */
        } /* end if */
      } /* end if */

      /* check the value */
      if (nFuncReturn == OK_CN00) {
        nFuncReturn = cn29CheckRange(pParameter, szValue);
      } /* end if */

      /* "write" the parameter */
      if (nFuncReturn == OK_CN00) {
        cn90StrUpperCopy (pParameter->szInput, szValue, 
                          (bool   ) (pParameter->xpProperties[XPPropClassCasesensitive] == XPPropValueYes));
        strcpy(pParameter->szCalc, pParameter->szInput);
        if (bOverride) {
        strcpy(pParameter->szConfirmed, pParameter->szInput);
        } // end if
        if (pParameter->xpProperties[XPPropClassDevspace] == XPPropValueYes) {
          pXParamData->bDevspaceDirty = true;
        } else {
          pXParamData->bKernelDirty = true;
        } /* end if */
      } /* end if */
    } else {
      nFuncReturn = ERR_XPNOTFOUND_CN00;
    } /* end if */

  } else {
    nFuncReturn = ERR_PARAM_CN00;
  } /* end if */

  return nFuncReturn;
} /* end cn29SoftPutIntern */


/* IMPLEMENTATION INTERN FUNCTION cn29ReadParamFile
 * ------------------------------------------------------------------
 * purpose: - open the param help of the specified db
 *          - call processing of file contents
 *          - close the file
 *
 */
tcn00_Error cn29ReadParamFile (
     tsp00_DbNamec       szDbName,
     tcn002_XParamData * pXParamData,
     char              * replyData,
     int               * replyLen)
{
  tcn00_Error         nFuncReturn    = OK_CN00;
  tsp_errtext         szErrtext;
  bool                bStatus;
  bool                bParamFileOpen = false;
  tsp_int4            hParamFile;
  tsp_int4            hParamFileV;
  tsp00_DbName        dbNameP;
  tsp00_DbNamec       dbNameTemp;

  /* open Param file of db */
  sqlxpopenxparam ( dbNameP.c2p(szDbName), 0, &hParamFile , szErrtext , &bStatus );
  if  ( !bStatus ) {
    nFuncReturn = ERR_RTE_CN00;
    cn90AnswerRTEError (replyData, replyLen, nFuncReturn, szErrtext, szErrtext.length(), xp_not_ok );
  } else {
    strcpy(dbNameTemp, szDbName);
    strcat(dbNameTemp, ".01");
    sqlxpopenxparam ( dbNameP.c2p(dbNameTemp), 0, &hParamFileV , szErrtext , &bStatus );
    if ( !bStatus ) {
      hParamFileV = -1;
    } // end if
    bParamFileOpen = true;
  } /* end if */

  /* Files abarbeiten */
  if (nFuncReturn == OK_CN00) {
    nFuncReturn = cn29ReadParamFileContents(szDbName, pXParamData, replyData, replyLen, hParamFile, hParamFileV);
  } /* end if */

  /* close Param file of db (close always if open, independent from nFuncReturn) */
  if (bParamFileOpen) {
    if (hParamFileV != -1) {
      sqlxpclosexparam ( hParamFileV , false, szErrtext , &bStatus );
    } // end if
    sqlxpclosexparam ( hParamFile , true, szErrtext , &bStatus );
    /* set errortext only, if there no preceding error */
    if  ( (!bStatus) && (nFuncReturn == OK_CN00) ) {
      nFuncReturn = ERR_RTE_CN00;
      cn90AnswerRTEError (replyData, replyLen, nFuncReturn, szErrtext, szErrtext.length(), xp_not_ok );
    } /* end if */
  } /* end if */

  return nFuncReturn;

} /* cn29ReadParamFile */

/* IMPLEMENTATION INTERN FUNCTION cn29ReadParamFileContents
 * ------------------------------------------------------------------
 * purpose: read and analyze the contents of XParam help file
 *
 */
tcn00_Error cn29ReadParamFileContents (
      tsp00_DbNamec       szDbName,
      tcn002_XParamData * pXParamData,
      char             * replyData,
      int              * replyLen,
      const tsp_int4     hParamFile,
      const tsp_int4     hParamFileV )
{
  tcn00_Error           nFuncReturn = OK_CN00;
  tsp00_XpKeyType       xpKey;
  tsp00_XpKeyType       xpNextKey;
  tsp00_XpKeyTypec      szParamC;
  tsp00_XpValueRec      xpValue;
  tsp00_XpReturnCode    nXPReturn;
  tsp00_XpReturnCode    nXPReturnV;
  tsp00_ErrText         szErrtext;
  tcn002_XParameter     aParameter;
  tcn002_XParameter   * pParameter  = NULL;
  tcn002_XParameter   * pDynHelp    = NULL;
  tcn002_XParameter   * pLast       = NULL;
  bool                 bMallocOK   = true;

  memset(xpKey, ' ', sizeof (tsp00_XpKeyType));
  do {
    sqlgnextxparam(hParamFile, xpKey, xpNextKey, &xpValue, szErrtext , &nXPReturn );

    if ( nXPReturn == xp_ok ) {
      cn90StringPtoC ( szParamC, xpNextKey, sizeof(xpNextKey));

      /* search parameter in List */
      pParameter = cn29SearchParameter(pXParamData->pFirst, szParamC);

      /* No match -> check for dynamic */
      if (pParameter == NULL) {

        pDynHelp = cn29SearchDynamic(pXParamData->pFirst, szParamC);
        if (pDynHelp != NULL) {

          cn00MyMalloc (sizeof (tcn002_XParameter), &pParameter, &bMallocOK);
          if ((pParameter == NULL) || (!bMallocOK)) {
            nFuncReturn = ERR_MEM_CN00;
            cn90AnswerIError(replyData, replyLen, nFuncReturn);
          } else {
            /* add the parameter in the list of parameters in the XParam section */
            if (pXParamData->pFirst == NULL) {
              pXParamData->pFirst = pParameter;
            } /* end if */
            if (pXParamData->pLast != NULL) {
              pXParamData->pLast->pNext = pParameter;
            } /* end if */

            /* copy data from dynamic record */
            nFuncReturn = cn29CopyParameter(pParameter, pDynHelp);
            /* clear the dynamic/intern property */
            pParameter->xpProperties[XPPropClassDynamic] = XPPropValueNo;
            pParameter->xpProperties[XPPropClassIntern] = XPPropValueNo;

            /* list management */
            pParameter->pNext = NULL;
            pParameter->pPrev = pXParamData->pLast;
            pXParamData->pLast = pParameter;
          } // end if
      
        } // end if

      } // end if

      if (pParameter != NULL) {
        /* Copy Data */
        strcpy(pParameter->szParameterName, szParamC);
        memcpy(&pParameter->xpValueRecord, &xpValue, sizeof(tsp00_XpValueRec));
        cn29ParamValueToString(pParameter->szInput, &pParameter->xpValueRecord);
        strcpy(pParameter->szOriginal, pParameter->szInput);

        /* set flags */
        pParameter->bParamExist = true;

        /* maintain second list (order by param file) */
        if (pXParamData->pParam == NULL) {
          pXParamData->pParam = pParameter;
          pLast               = pParameter;
        } else {
          pLast->pParam = pParameter;
        } /* end if */
        pLast = pParameter;

        /* an version value */
        if (hParamFileV != -1) {
          sqlgetxparam (hParamFileV, xpNextKey, &xpValue, szErrtext , &nXPReturnV);
          if (nXPReturnV == xp_ok) {
            cn29ParamValueToString(pParameter->szGood, &xpValue);
          } // end if
        } // end if

        /* and the next */
        strncpy ( xpKey , xpNextKey , sizeof(xpKey) );
      } else {
        strcpy(aParameter.szParameterName, szParamC);
        memcpy(&aParameter.xpValueRecord, &xpValue, sizeof(tsp00_XpValueRec));
        cn29ParamValueToString(aParameter.szInput, &aParameter.xpValueRecord);
        strcpy(aParameter.szOriginal, aParameter.szInput);

        // parameter not known in cserv.pcf -> delete it
        sqldelxparam (hParamFile, xpNextKey, szErrtext, &nXPReturn);
        cn29WriteParamHistory(szDbName, &aParameter, PAHI_DELETE_CN002);
      } /* end if */
    } /* end if */

  }  while ( (nXPReturn == xp_ok) && (nFuncReturn == OK_CN00));

  return nFuncReturn;
} /* end cn29ReadParamFileContents */

/* IMPLEMENTATION INTERN FUNCTION cn29DeleteParameter
 * ------------------------------------------------------------------
 * purpose: read and analyze the contents of XParam help file
 *
 */
void cn29DeleteParameter (
      tcn002_XParameter * pParameter )
{
  tcn002_XPLine     * pLine       = NULL;
  tcn002_XPLine     * pNextLine   = NULL;

  /* free the list of code lines */
  pLine = pParameter->pCode;
  while (pLine != NULL) {
    pNextLine = pLine->pNext;
    cn00MyFree(pLine->szLine);
    cn00MyFree(pLine);
    pLine = pNextLine;
  } /* end while */

  /* free the list of constarint lines */
  pLine = pParameter->pConstraint;
  while (pLine != NULL) {
    pNextLine = pLine->pNext;
    cn00MyFree(pLine->szLine);
    cn00MyFree(pLine);
    pLine = pNextLine;
  } /* end while */

  /* free the list of explain lines */
  pLine = pParameter->pExplain;
  while (pLine != NULL) {
    pNextLine = pLine->pNext;
    cn00MyFree(pLine->szLine);
    cn00MyFree(pLine);
    pLine = pNextLine;
  } /* end while */

  /* free the list of help lines */
  pLine = pParameter->pHelp;
  while (pLine != NULL) {
    pNextLine = pLine->pNext;
    cn00MyFree(pLine->szLine);
    cn00MyFree(pLine);
    pLine = pNextLine;
  } /* end while */

  /* free the parameter */
  cn00MyFree(pParameter);

} /* end cn29DeleteParameter */

/* IMPLEMENTATION INTERN FUNCTION cn29ParamStartSession
 *---------------------------------------------------------------------------*
 * purpose: initialize a XParam-Session
 *
 */
tcn00_Error cn29ParamStartSession
      ( tsp00_DbNamec         szDbName,
        tcn002_XParamData * & pXParamData,
        char              *   replyData,
        int               *   replyLen)
{
  FUNCTION_DBG_MCN00_1(_T("vcn29::cn29ParamStartSession"));

  tcn00_Error         nFuncReturn = OK_CN00;
  tsp00_Pathc         szNewFile;
  tsp01_RteError      RteError;

  if (!sqlGetDbrootEnvPath ( szNewFile, TERM_WITH_DELIMITER_EO01, &RteError)) {
    nFuncReturn = cn90AnswerNewRTEError(replyData, replyLen, ERR_RTEEXT_CN00, &RteError);
  } else {
    strcat(szNewFile, XP_NEWFILE_CN002);
  } // end if

  if (nFuncReturn == OK_CN00) {
    /* free existing XParam section */
    cn29FreeXParamSection(pXParamData);

    /* allocate new XParam section */
    pXParamData = new tcn002_XParamData;
    if (pXParamData == NULL) {
      nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_MEM_CN00);
    } else {
      memset(pXParamData, 0,sizeof(tcn002_XParamData));
    } /* end if */

    /* read the help file */
    if (nFuncReturn == OK_CN00) {
      nFuncReturn = cn21ReadHelpFile(pXParamData, replyData, replyLen, szNewFile, true);
    } /* end if */

    /* reading the param file  !!! after the help file */
    if (nFuncReturn == OK_CN00) {
      nFuncReturn = cn29ReadParamFile(szDbName, pXParamData, replyData, replyLen);
    } /* end if */

    // check the paramfile version
    if (nFuncReturn == OK_CN00) {
      tcn002_XParameter * pParameter = NULL;
      tsp00_Versionc      VersionString;

      pParameter = cn29SearchParameter(pXParamData->pFirst, PAN_PARAM_KERNELVERSION);
      if (pParameter != NULL) {
        if (_tcslen(pParameter->szDefault) > 0) {
          sp100_GetVersionString(COMP_NAME_DBMSERVER_SP100, s100buildnumber, VersionString);
          if (_tcsncmp(&VersionString[BEG_OF_1ST_VERS_SUBSTR_SP100], 
                       &pParameter->szDefault[BEG_OF_1ST_VERS_SUBSTR_SP100], 
                       END_OF_1ST_VERS_SUBSTR_SP100 - BEG_OF_1ST_VERS_SUBSTR_SP100) != 0) {
            teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, ERR_PARAMVER_CN00_1, VersionString.asCharp(), pParameter->szDefault);
            nFuncReturn = cn90AnswerEvent(replyData, replyLen, aEvent);
          } // end if
        } // end if
      } // end if
    } // end if

    if (nFuncReturn == OK_CN00) {
      tcn002_XParameter * pParameter = NULL;

      /* initialize the checking  */
      cn23InitializeChecking(pXParamData->pFirst);

      pParameter = cn29SearchParameter(pXParamData->pFirst, PAN_INSTANCE_TYPE);
      if ( pParameter != NULL ) {
        if ( _tcslen(pParameter->szInput) > 0 ) {
          nFuncReturn = cn29CheckInstance(pParameter->szInput, pXParamData);
        } // end if
      } // end if
    } // end if

    if (nFuncReturn == OK_CN00) {
#ifdef XPCHECKDEP
      /* print dependencies to stdout */
      cn23PrintDependencies (pXParamData->pFirst);
#endif
      /* generate Answer */
      cn90AnswerOK(replyData, replyLen, NULL);

      pXParamData->bKernelDirty = false;
      pXParamData->bDevspaceDirty = false;

    } else {
      /* free the XParam section */
      cn29FreeXParamSection(pXParamData);
      pXParamData = NULL;
    } /* end if */
  } /* end if */

  return nFuncReturn;

} /* end cn29ParamStartSession */

/* ------------------------------------------------------------------
 * PUBLIC FUNCTION cn29FreeXParamSection
 * ------------------------------------------------------------------
 * purpose: frees the XParamSection
 *
 */
void cn29FreeXParamSection 
      ( tcn002_XParamData * pXParamData )
{
  tcn002_XParameter * pParameter  = NULL;
  tcn002_XParameter * pNext       = NULL;

  if (pXParamData != NULL) {
    /* free the list of parameters */
    pParameter = pXParamData->pFirst;
    while (pParameter != NULL) {
      pNext = pParameter->pNext;
      cn29DeleteParameter(pParameter);
      pParameter = pNext;
    } /* end while */

    /* free the main structure */
    delete pXParamData;

  } /* end if */

} /* end cn29FreeXParamSection */

/* IMPLEMENTATION PRIVATE FUNCTION cn29_InvalidForInstance
 *---------------------------------------------------------------------------*
 */
bool cn29_InvalidForInstance
      ( const _TCHAR  *szInstance, 
        const _TCHAR  *szPossible )
{
  bool    bInvalid = false;
  _TCHAR  szToken[PARAMETER_MAXLEN_CN90];
  int     nToken = 1;

  if ((_tcslen(szInstance) > 0) && (_tcslen(szPossible) > 0)) {
    bInvalid = true;
    while (bInvalid && cn90GetToken(szPossible, szToken, nToken) ) {
      cn90StripQuotes(szToken);
      if (_tcsicmp(szToken, szInstance) == 0) {
        bInvalid = false;
      } // end if
      ++nToken;
    } // end while
  } // end if

  return bInvalid;
} // end cn29_InvalidForInstance

/* IMPLEMENTATION INTERN FUNCTION cn29CheckInstance
 *---------------------------------------------------------------------------*
 */
tcn00_Error cn29CheckInstance 
      ( const _TCHAR      * szInstance,
        tcn002_XParamData * pXParamData)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  tcn002_XParameter * pParameter  = NULL;
  tcn002_XParameter * pPrev       = NULL;
  tcn002_XParameter * pNext       = NULL;

  if (pXParamData != NULL) {

    // maintain the list
    pParameter = pXParamData->pFirst;
    while (pParameter != NULL) {

      pNext = pParameter->pNext;

      if (cn29_InvalidForInstance(szInstance, pParameter->szProperties[XPPropClassInstance])) {
        pParameter->xpProperties[XPPropClassGroup] = XPPropValueNo;
      } // end if

      // set defaults
      if (_tcslen(pParameter->szDefaultFor[0][XP_DEFAULTFORINSTANCE_CN002]) > 0 ) {
        int nIndex = 0;

        for (nIndex = 0; nIndex < XP_DEFAULTFOR_MXCN002; ++nIndex) {
          if (_tcsicmp(szInstance, pParameter->szDefaultFor[nIndex][XP_DEFAULTFORINSTANCE_CN002]) == 0) {
            if (!pParameter->bParamExist) {
              _tcscpy(pParameter->szInput,   pParameter->szDefaultFor[nIndex][XP_DEFAULTFORVALUE_CN002]);
            } // end if
            _tcscpy(pParameter->szDefault, pParameter->szDefaultFor[nIndex][XP_DEFAULTFORVALUE_CN002]);
            break;
          } // end if
        } // end for
      } // end if

      pParameter = pNext;
    } // end while

  } // end if

  return nFuncReturn;
} // end cn29CheckInstance

/* IMPLEMENTATION INTERN FUNCTION cn29AnswerParamNotFound
 *---------------------------------------------------------------------------*
 */
tcn00_Error cn29AnswerParamNotFound 
      ( _TCHAR            * replyData,
        int               * replyLen,
        const _TCHAR      * szParameter,
        tcn002_XParameter * pParameter,
        bool                bDirect)
{
  tcn00_Error         nFuncReturn = ERR_XPNOTFOUND_CN00;
  _TCHAR            * pData;
  int                 nHits = 0;

  cn90AnswerIError(replyData, replyLen, nFuncReturn);

  if (_tcslen(szParameter) > 2) {

    pData = replyData + _tcslen(replyData);

    while (pParameter != NULL) {

      if (strstr(pParameter->szParameterName, szParameter) != NULL) {
        if (nHits == 0) {
          _stprintf(pData, "\n");
          pData = pData + _tcslen(pData);
        } // end if

        ++nHits;

        _stprintf(pData, "%-20s", pParameter->szParameterName);
        pData = pData + _tcslen(pData);

        if ((nHits % 3) == 0) {
          _stprintf(pData, "\n");
          pData = pData + _tcslen(pData);
        } // end if

      } // end if

      pParameter = bDirect ? pParameter->pParam : pParameter->pNext;
    } // end while

    if ((nHits % 3) != 0) {
      _stprintf(pData, "\n");
      pData = pData + _tcslen(pData);
    } // end if

  } // end if

  *replyLen = (int)_tcslen(replyData);

  return nFuncReturn;
} // end cn29AnswerParamNotFound

/* IMPLEMENTATION INTERN FUNCTION cn29WriteParamHistory
 *---------------------------------------------------------------------------*
 */
void cn29WriteParamHistory 
      ( tsp00_DbNamec       szDBName,
        tcn002_XParameter * pParameter, 
        _TCHAR              cOperation)
{
  if ( ( (_tcscmp(pParameter->szInput, pParameter->szOriginal) != 0 ) ||
         (cOperation == PAHI_DELETE_CN002                           )     ) &&
       (_tcscmp(pParameter->szParameterName, XP_PARAM_CHANGED     )  != 0 ) &&
       (_tcscmp(pParameter->szParameterName, XP_PARAM_VERSION_TIME)  != 0 ) &&
       (pParameter->xpValueRecord.xp_value_t  != xp_crypt_type            )    ) {
  
    tsp00_XpKeyTypec szName;
    tsp00_C64c       szNewValue;
    tsp00_C64c       szOldValue;
    tsp00_ErrTextc   szError;
    tsp00_C8c        szGroup;


    szName.rawAssign(pParameter->szParameterName);
    szNewValue.rawAssign(pParameter->szInput);
    szOldValue.rawAssign(pParameter->szOriginal);

    struct {
      tcn002_XPPropValue   nValue;
      const char         * szText;
    }                     aValueTable[] = XP_PROP_VALUES_CN002;
    int                   j                 = 0;
    /* search value */
    while ( (aValueTable[j].nValue != XPPropValueUnknown) && (aValueTable[j].nValue != pParameter->xpProperties[XPPropClassGroup]) ) {
      j++;
    } /* end while */
    if (aValueTable[j].szText != NULL) {
      szGroup.rawAssign(aValueTable[j].szText);
    } else {
      szGroup.Init();
    } // end if

    cn52ParamHistoryRecord aRecord(szDBName,
                                   szName,
                                   szNewValue,
                                   szOldValue,
                                   szGroup,
                                   (cOperation == PAHI_DELETE_CN002)? cn52ParamHistoryRecord::StDeleted : cn52ParamHistoryRecord::StChanged);

    aRecord.appendToFile (szError);

  } // end if

} // end cn29WriteParamHistory
