/*******************************************************************************
  module:       vcn37.cpp
  ------------------------------------------------------------------------------
  responsible:  TiloH
  special area: dbm server
  description:  implementation of pipe2file
  last change:  2000-11-29 15:31
  version:      7.2.5
  see also:     
  ------------------------------------------------------------------------------
                         Copyright by SAP AG, 2000


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*******************************************************************************/

#include<stdio.h>
#include<time.h>

#include "hni32.h"
#include "hni34.h"
#include "hni35.h"
#include "hcn36.h"

#if defined(WIN32)
#elif defined(UNIX)
    #include<unistd.h>
#else
    #error vcn37.cpp only coded for UNIX and WIN32 (define one of them)
#endif


ni32_RestrictedDoubleParameter Direction_cn37("-d", ni32_RestrictedDoubleParameter::TempValueList<<"file2pipe"<<"pipe2file"<<"test2pipe"<<"compare");
ni32_DoubleParameter FileName_cn37("-f");
ni32_DoubleParameter FileSize_cn37("-s");

ni32_SingleParameter HiddenParameterReadOnly_cn37("-readonly");
ni32_SingleParameter HiddenParameterRemoveFiles_cn37("-removefiles");
ni32_SingleParameter HiddenParameterShowSpeed_cn37("-showspeed");
ni32_SingleParameter HiddenParameterNoWait_cn37("-nowait");

ni32_TrippleParameter HiddenParameterWait_cn37("-wait");


class cn37_Pipe2FileProgramParameters:public ni32_ProgramParameters
{
  public:
    cn37_Pipe2FileProgramParameters(int                            argc,
                                    const char * const           * argv,
                                    ni31_List<ni32_AParameter *> & ParamList)
        :ni32_ProgramParameters(argc, argv, ParamList),
         CurrentProgramName(argv[0]),
         UsageString(0),
         UsageStringConstructed(0)
    {}

    ~cn37_Pipe2FileProgramParameters()
    {
        if(UsageStringConstructed)
            delete [] UsageString;
    }

    virtual const char * UsageMessage()
    {
        if(0==UsageString)
        {
            static char ConstUsageString[]="usage: pipe2file <options> [-d file2pipe|pipe2file] [-s <max_file_size>]\n"
                                           "    <options>:\n"
                                           "        -f <file_name>      (name of the file)\n"
                                           "        -p <pipe_name>      (name of the pipe)\n"
                                           "    <max_file_size>:\n"
                                           "        An integer specifying the maximal size of the files processed. Only if\n"
                                           "        a maximal file size is specified, consecutive numbers are added to the\n"
                                           "        <file_name>.\n";

            const char * RelativCurrentProgramName=FileNameInPath(CurrentProgramName);

            if(0==RelativCurrentProgramName || !strcmp(RelativCurrentProgramName, "pipe2file"))
                UsageString=ConstUsageString;
            else
            {
                char * ProgramName=ConstUsageString+strlen("usage: ");
                char * Options    =ProgramName+strlen("pipe2file");

                *ProgramName='\0';

                size_t Length=strlen(ConstUsageString)+strlen(RelativCurrentProgramName)+1+strlen(Options);

                UsageString=new char [Length+1];

                if(0==UsageString)
                {
                    *ProgramName='p';
                    UsageString=ConstUsageString;
                }
                else
                {
                    UsageStringConstructed=1;
                    sprintf(UsageString, "%s%s%s", ConstUsageString, RelativCurrentProgramName, Options);
                }
            }
        }

        return UsageString;
    }

    virtual int          UsageMessageAvailable(){return 1;}

    int ExtendedParameterCheck()
    {
        int rc=01;

        if(0==strcmp(Direction_cn37.GiveValue(), "pipe2file") ||
           0==strcmp(Direction_cn37.GiveValue(), "file2pipe"))
        {
            if(!FileName_cn37.IsSet())
                rc=0;
        }
        else
            if(0==strcmp(Direction_cn37.GiveValue(), "test2pipe"))
            {
                if(!FileSize_cn37.IsSet())
                    rc=0;
                else
                    HiddenParameterNoWait_cn37.TakeValue("-nowait");    // the "-nowait" is not necessary
            }
            else
                if(0==strcmp(Direction_cn37.GiveValue(), "compare"))
                {
                    if(!FileName_cn37.IsSet())
                        rc=0;
                }

        if(!rc)
            fprintf(stderr, "%s\n", (const char *)UsageMessage());

        return rc;
    }

  private:
    const char * CurrentProgramName;
    char       * UsageString;
    int          UsageStringConstructed;

    const char * FileNameInPath(const char *Name)
    {
        const char *h;
        #if defined(WIN32)
        h=strrchr(Name, '\\');
        #elif defined (UNIX)
        h=strrchr(Name, '/');
        #else
            #error vcn37.cpp was only coded for WIN32 and UNIX (define one of them).
        #endif

        if(h==0)
            h=Name;
        else
            h++;

        return h;
    }
};

static void cn37_ShowSpeedIfRequested(teo00_Int8 TotalBytesWritten, time_t StartTime)
{
    static time_t LastTime=0;

    if(HiddenParameterShowSpeed_cn37.IsSet())
    {    
        time_t EndTime=time(0);

        if(0==LastTime || 0<EndTime-LastTime)   // for the first time or if last show was a second ago
        {
            if(EndTime>StartTime)
            {
                if(MAX_UINT8_CN36==TotalBytesWritten)
                    printf(">= ");

                printf("%d MB/s\n", (double(TotalBytesWritten)/double(1024*1024))/double(EndTime-StartTime));
            }
            else
                printf(">%d MB/s\n", (double(TotalBytesWritten)/double(1024*1024)));

            LastTime=EndTime;   // remember, that we showed the speed in this second
        }
    }
}

static void IncreaseTotalBytesCounter(teo00_Uint8 & TotalBytes, long Increment)
{
    if(MAX_UINT8_CN36-Increment>=TotalBytes)
        TotalBytes+=Increment;
    else
        TotalBytes=MAX_UINT8_CN36;
}

static int cn37_ReportFileAsDoneAndWaitForAnswer(const char *CompleteFileName, int * WasLastFile=0)
{
    int rc=1;

    if(!HiddenParameterNoWait_cn37.IsSet())
    {
        printf("d");
        fflush(stdout);

        if(HiddenParameterRemoveFiles_cn37.IsSet() && 0!=CompleteFileName)
        {
            tni34_AVirtualFile TheFile(CompleteFileName, tni34_AVirtualFile::ToDiskAppend, tni34_AVirtualFile::DoNotOpenNow); // do not open

            if(!TheFile.Remove())
                rc=0;

            if(0!=WasLastFile)
                *WasLastFile=0;
        }
        else
        {
            char Line[1024+1];
            int  i=0, LineBreak=0, EndOfFileReached=0;

            while(!LineBreak && !EndOfFileReached && i<1024)
            {
                if(EOF==scanf("%c", Line+i))
                    EndOfFileReached=1;
                else
                    if('\n'==Line[i])
                        LineBreak=1;
                    else
                        i++;
            }

            Line[i]='\0';

            if(EndOfFileReached || (Line[0]!='n' && Line[0]!='l'))
                rc=0;
            else
                if(0!=WasLastFile)
                    *WasLastFile=(Line[0]=='l');
        }
    }
    else
    {
        if(0!=WasLastFile)
            *WasLastFile=0;
    }

    return rc;
}

static int cn37_ReadPipeWriteFiles(const char * PipeName,
                                   const char * FileName,
                                         char * Buffer,
                                         int    BufferSize,
                                   const int    UseFileSizeBound,
                                   teo00_Int8   FileSizeBound,
                                   int          DelayTime,
                                   teo00_Uint8  DelayPoint,
                                   int          NumberOfWaits,
                                   teo00_Uint8  WaitAfterEveryBytes)
{
    int rc=1;

    tni34_AVirtualPipe Pipe(PipeName, tni34_AVirtualPipe::In);

    if(Pipe.IsGood())
    {
        tni34_AVirtualFile * File=0;

        int  PostFix=0;
        char CompleteFileName[256];
        strncpy(CompleteFileName, FileName, 256-12-1); // at least 12 char's for postfix
        CompleteFileName[256-12-1]='\0';

        char * StartOfPostFix=CompleteFileName+strlen(CompleteFileName);

        long BytesRead=0;
        long BufferBytesWritten;
        teo00_Int8 SpaceInFile;
        
        teo00_Uint8 TotalBytesRead=0;
        teo00_Uint8 TotalBytesWritten=0;    // is used for undocumented show speed feature only
        
        teo00_Uint8 NextWaitPoint=WaitAfterEveryBytes;

        int Delayed=0;

        time_t StartTime=time(0);

        do
        {
            int DoneDelayOrExtraWait=0;

            do
            {
                int Delay=0;
                int ExtraWait=0;

                DoneDelayOrExtraWait=0;

                if(0!=DelayTime && TotalBytesRead+BufferSize>=DelayPoint && !Delayed)
                    Delay=1;

                if(0<NumberOfWaits && TotalBytesRead+BufferSize>=NextWaitPoint)
                    ExtraWait=1;

                if(Delay || ExtraWait)
                {
                    long BytesToReadBeforeDelay;
                
                    if(Delay)
                        BytesToReadBeforeDelay=long(DelayPoint-TotalBytesRead); //long() is ok, because of if(...TotalBytesRead+BufferSize>=DelayPoint...)
                    else
                        //if(ExtraWait)
                        BytesToReadBeforeDelay=long(NextWaitPoint-TotalBytesRead); //long() is ok, because of if(...TotalBytesRead+BufferSize>=NextWaitPoint...)

                    if(0<BytesToReadBeforeDelay)
                    {
                        if(!Pipe.Read(Buffer, BytesToReadBeforeDelay, BytesRead))  
                            rc=0;
                    }
                    
                    if(rc && BytesToReadBeforeDelay==BytesRead) //just stop, if all was ok and we have all the bytes we asked for
                    {
                        if(Delay)
                            cn36_Sleep(DelayTime);
                        else
                            cn37_ReportFileAsDoneAndWaitForAnswer(CompleteFileName);
                        
                        DoneDelayOrExtraWait=1;

                        if(Delay)
                            Delayed=1;
                        else
                        {
                            NumberOfWaits--;
                            NextWaitPoint+=WaitAfterEveryBytes;
                        }
                    }
                }
                else
                {
                    if(!Pipe.Read(Buffer, BufferSize, BytesRead))
                        rc=0;
                }
            }
            while(rc && DoneDelayOrExtraWait && 0==BytesRead);  // repeat with delays or extra waits, until we run into an error or we realy read some bytes from the pipe

            if(rc)
            {
                BufferBytesWritten=0;
                TotalBytesRead+=BytesRead;

                if(HiddenParameterReadOnly_cn37.IsSet())
                {
                    IncreaseTotalBytesCounter(TotalBytesWritten, BytesRead);

                    BufferBytesWritten=BytesRead;
                    
                    cn37_ShowSpeedIfRequested(TotalBytesWritten, StartTime);
                }

                while(BufferBytesWritten!=BytesRead && rc)
                {
                    if(0==File)
                    {
                        if(UseFileSizeBound)
                        //if(FileSizeBound>0)
                        {
                            sprintf(StartOfPostFix, ".%d", PostFix++);
                            File=new tni34_AVirtualFile(CompleteFileName, tni34_AVirtualFile::ToDiskTruncate);
                            SpaceInFile=FileSizeBound;
                        }
                        else
                        {
                            if(PostFix==0)
                            {
                               PostFix++;
                               File=new tni34_AVirtualFile(CompleteFileName, tni34_AVirtualFile::ToDiskTruncate);
                               SpaceInFile=MAX_INT8_CN36;
                            }
                            else
                                rc=0;   // if we do not use a file size bound we can not open more then one file
                        }
                    }

                    if(rc)
                    {
                        if(0==File || !File->IsGood())
                        {
                            rc=0;
                            ErrorHandler_ni35->Report(tni35_AStandardError::OutOfMemory.Set());
                        }
                        else    // we have a file for writing
                        {
                            long BytesWritten;
                            long h=BytesRead-BufferBytesWritten;

                            if(h>SpaceInFile)
                                h=long(SpaceInFile);  // conversion correct, because h>SpaceInFile>0

                            if(File->Write(Buffer+BufferBytesWritten, h, BytesWritten))
                            {
                                IncreaseTotalBytesCounter(TotalBytesWritten, BytesWritten);

                                BufferBytesWritten+=BytesWritten;

                                if(UseFileSizeBound)            //if we do not use a file size bound, we have unlimited space in the file
                                    SpaceInFile-=BytesWritten;
                            }
                            else
                                rc=0;
                        }
                    }

                    if(0>=SpaceInFile)  // we have no more space in file -> close it and make sure that we open the next one in the next loop
                    {
                        if(!File->Close())
                            rc=0;
                        else
                            if(cn37_ReportFileAsDoneAndWaitForAnswer(CompleteFileName))
                                cn37_ShowSpeedIfRequested(TotalBytesWritten, StartTime);
                            else
                                rc=0;

                        delete File;
                        File=0;
                    }
                }
            }
        }
        while(rc && 0!=BytesRead);     //no error and not end of pipe reached

        if(0!=File)
        {
            if(!File->Close())
                rc=0;
            else           
                if(!cn37_ReportFileAsDoneAndWaitForAnswer(CompleteFileName))
                    rc=0;

            if(rc && HiddenParameterShowSpeed_cn37.IsSet() && !UseFileSizeBound)
            {
                teo00_Uint8 FileSize;

                if(!File->GetSize(FileSize))
                    fprintf(stderr, "Could not determine the size of '%s'.\n", (const char *)File->GiveName());
                else
                {
                    char Buf[50];

                    if(cn36_UInt8ToStrN(Buf, 50, FileSize))
                        printf("\nCreated a file of size %s.\n", Buf);
                    else
                        fprintf(stderr, "\nInternal error.\n");
                }
            }

            delete File;
            File=0;
        }
        
        cn37_ShowSpeedIfRequested(TotalBytesWritten, StartTime);
    }
    else
        rc=0;

    return rc;
}

static int cn37_ReadFilesWritePipe(const char * FileName,
                                   const char * PipeName,
                                         char * Buffer,
                                         int    BufferSize,
                                   const int    UseFileSizeBound)
{
    int rc=1;

    tni34_AVirtualPipe Pipe(PipeName, tni34_AVirtualPipe::Out);

    if(Pipe.IsGood())
    {
        tni34_AVirtualFile * File=0;

        int  PostFix=0;
        char CompleteFileName[256];
        strncpy(CompleteFileName, FileName, 256-12-1); // at least 12 char's for postfix
        CompleteFileName[256-12-1]='\0';

        char * StartOfPostFix=CompleteFileName+strlen(CompleteFileName);

        int LastFileDone=0;
        long BytesRead;

        teo00_Uint8 TotalBytesWritten=0;

        do
        {
            if(UseFileSizeBound)
            {
                sprintf(StartOfPostFix, ".%d", PostFix++);
                File=new tni34_AVirtualFile(CompleteFileName, tni34_AVirtualFile::FromDisk);
            }
            else
            {
                if(PostFix==0)
                {
                    PostFix++;
                    File=new tni34_AVirtualFile(CompleteFileName, tni34_AVirtualFile::FromDisk);
                }
                else
                    rc=0;   // if we do not use a file size bound we can not open more then one file 
            }

            if(rc)
            {
                if(0==File || !File->IsGood()) //something wrong?
                {
                    rc=0;
                    ErrorHandler_ni35->Report(tni35_AStandardError::OutOfMemory.Set());
                }
                else
                {
                    do
                    {
                        if(File->Read(Buffer, BufferSize, BytesRead))
                        {
                            if(0!=BytesRead)
                            {
                                long BytesWritten=-1;
                        
                                if(!Pipe.Write(Buffer, BytesRead, BytesWritten) || BytesWritten!=BytesRead)
                                    rc=0;
                                else
                                    IncreaseTotalBytesCounter(TotalBytesWritten, BytesWritten);
                            }
                        }
                        else
                            rc=0;
                    }
                    while(rc && 0!=BytesRead);

                    if(!File->Close())
                        rc=0;
                
                    if(rc)
                    {
                        if(!cn37_ReportFileAsDoneAndWaitForAnswer(CompleteFileName, &LastFileDone))
                            rc=0;
                        else
                            if(!UseFileSizeBound)
                                LastFileDone=1; //without a file size bound we process only one file and this is the last
                    }

                    delete File;
                    File=0;
                }
            }
        }
        while(rc && !LastFileDone);     //no error and not end of last file reached

        if(!Pipe.Close())
            rc=0;

        if(0!=File)
            delete File;
    }
    else
        rc=0;

    return rc;
}

static int cn37_ConstructTestDataWritePipe(const char * PipeName,
                                                 char * Buffer,
                                                 int    BufferSize,
                                           teo00_Uint8  TestSize,
                                           int          DelayTime,
                                           teo00_Uint8  DelayPoint)
{
    int rc=1;
    int Delayed=0;

    tni34_AVirtualPipe Pipe(PipeName, tni34_AVirtualPipe::Out);

    if(Pipe.IsGood())
    {
        long BytesRead;

        teo00_Uint8 TotalBytesWritten=0;
        teo00_Uint8 TotalBytesConstructed=0;

        int i=0;

        for(i=0; i<BufferSize; i++)
            Buffer[i]=i;

        do
        {
            if(TotalBytesConstructed+BufferSize<TestSize)
                BytesRead=BufferSize;
            else
                BytesRead=long(TestSize-TotalBytesConstructed); // cast is ok, because of if()...else

            TotalBytesConstructed+=BytesRead;

            if(0!=DelayTime && TotalBytesConstructed>=DelayPoint && !Delayed)
            {
                long BytesReadBeforeDelay=long(DelayPoint-(TotalBytesConstructed-BytesRead));
                long BytesReadAfterDelay=long(TotalBytesConstructed-DelayPoint);
                
                long BytesWritten=-1;

                // write before the delay
                if(!Pipe.Write(Buffer, BytesReadBeforeDelay, BytesWritten) || BytesWritten!=BytesReadBeforeDelay)
                    rc=0;
                else
                    IncreaseTotalBytesCounter(TotalBytesWritten, BytesWritten);

                if(rc)
                {
                    cn36_Sleep(DelayTime);  // delay

                    //write after the delay
                    if(!Pipe.Write(Buffer+BytesReadBeforeDelay, BytesReadAfterDelay, BytesWritten) || BytesWritten!=BytesReadAfterDelay)
                        rc=0;
                    else
                        IncreaseTotalBytesCounter(TotalBytesWritten, BytesWritten);
                }

                Delayed=1;
            }
            else
                if(0!=BytesRead)
                {
                    long BytesWritten=-1;
        
                    if(!Pipe.Write(Buffer, BytesRead, BytesWritten) || BytesWritten!=BytesRead)
                        rc=0;
                    else
                        IncreaseTotalBytesCounter(TotalBytesWritten, BytesWritten);
                }
            
        }
        while(rc && 0!=BytesRead);
    
        if(rc)
            if(!cn37_ReportFileAsDoneAndWaitForAnswer(0))
                rc=0;

        if(!Pipe.Close())
            rc=0;
    }
    else
        rc=0;

    return rc;
}

static int cn37_CompareFileAndPipe(const char * FileName,
                                   const char * PipeName,
                                         char * Buffer,
                                         char * Buffer2,
                                         int    BufferSize,
                                   const int    UseFileSizeBound)
{
    int rc=1;

    tni34_AVirtualPipe Pipe(PipeName, tni34_AVirtualPipe::In);

    if(Pipe.IsGood())
    {
        tni34_AVirtualFile * File=0;
        
        int  PostFix=0;
        char CompleteFileName[256];
        strncpy(CompleteFileName, FileName, 256-12-1); // at least 12 char's for postfix
        CompleteFileName[256-12-1]='\0';

        char * StartOfPostFix=CompleteFileName+strlen(CompleteFileName);

        int  LastFileDone=0;
        long BytesRead;
        long BytesRead2;

        do
        {
            if(UseFileSizeBound)
            {
                sprintf(StartOfPostFix, ".%d", PostFix++);
                File=new tni34_AVirtualFile(CompleteFileName, tni34_AVirtualFile::FromDisk);
            }
            else
            {
                if(PostFix==0)
                {
                    PostFix++;
                    File=new tni34_AVirtualFile(CompleteFileName, tni34_AVirtualFile::FromDisk);
                }
                else
                    rc=0;   // if we do not use a file size bound we can not open more then one file 
            }

            if(rc)
            {
                if(0==File || !File->IsGood()) //something wrong?
                {
                    rc=0;
                    ErrorHandler_ni35->Report(tni35_AStandardError::OutOfMemory.Set());
                }
                else
                {
                    do
                    {
                        if(File->Read(Buffer, BufferSize, BytesRead))
                        {
                            if(0!=BytesRead)
                            {
                                if(!Pipe.Read(Buffer2, BytesRead, BytesRead2)) //BufferSize>=BytesRead
                                    rc=0;
                                else
                                {
                                    if(BytesRead!=BytesRead2 || 0!=memcmp(Buffer, Buffer2, BytesRead))
                                    {
                                        rc=2;
                                        fprintf(stderr, "Sources differ.\n");
                                    }
                                }
                            }
                        }
                        else
                            rc=0;
                    }
                    while(1==rc && 0!=BytesRead);   //as long as we did not find a difference and did not reach the end of the file

                    if(!File->Close())
                        rc=0;
                
                    if(rc)
                    {
                        if(!cn37_ReportFileAsDoneAndWaitForAnswer(CompleteFileName, &LastFileDone))
                            rc=0;
                        else
                            if(!UseFileSizeBound)
                                LastFileDone=1; //without a file size bound we process only one file and this is the last
                    }

                    delete File;
                    File=0;
                }
            }
        }
        while(1==rc && !LastFileDone);     //no error, no difference of byte streams (rc==2) and not end of last file reached

        if(1==rc && LastFileDone)
        {
            if(!Pipe.Read(Buffer2, BufferSize, BytesRead2)) //BufferSize>=BytesRead
                rc=0;
            else
                if(0<BytesRead2)
                {
                    rc=2; //all files ended but still data in the pipe
                    fprintf(stderr, "Sources differ.\n");
                }
        }

        if(!Pipe.Close())
            rc=0;

        if(0!=File)
            delete File;
    }
    else
        rc=0;

    return rc;
}


int main(int argc, char ** argv)
{
    int rc=1;

    ni32_DoubleParameter PipeName("-p", ni32_AParameter::Required);
    ni32_DoubleParameter BlockSize("-b");
    ni32_TrippleParameter DelayAfter("-delay");

    cn37_Pipe2FileProgramParameters Parameters(argc,
                                               argv,
                                               ni32_ProgramParameters::TempParamList<<&Direction_cn37
                                                                                    <<&FileName_cn37
                                                                                    <<&PipeName
                                                                                    <<&FileSize_cn37
                                                                                    <<&BlockSize
                                                                                    <<&DelayAfter
                                                                                    <<&HiddenParameterReadOnly_cn37
                                                                                    <<&HiddenParameterRemoveFiles_cn37
                                                                                    <<&HiddenParameterShowSpeed_cn37
                                                                                    <<&HiddenParameterNoWait_cn37
                                                                                    <<&HiddenParameterWait_cn37);

    if(!Parameters.Check())
        rc=0;
    else
    {
        teo00_Uint8 FileSizeBound=teo00_Uint8(0);
        int         UseFileSizeBound=0;
        int         BufferSize=2*64*1024;   // we use a 128KB Buffer by default

        int         DelayTime=0;
        teo00_Uint8 DelayPoint=0;

        int         NumberOfWaits=0;
        teo00_Uint8 WaitAfterEveryBytes=0;

        if(FileSize_cn37.IsSet())
        {
            if(cn36_StrToUInt8(FileSizeBound, FileSize_cn37.GiveValue()))
            {
                if(0<FileSizeBound)
                    UseFileSizeBound=1;
                else
                    rc=0;
            }
            else
                rc=0;
        }

        if(rc && BlockSize.IsSet())
            if(!cn36_StrToInt(BufferSize, BlockSize.GiveValue()))
                rc=0;

        if(rc && DelayAfter.IsSet())
        {
            const char * DelayTimeAsString;
            const char * DelayPointAsString;

            DelayAfter.GiveValues(DelayTimeAsString, DelayPointAsString);

            if(!cn36_StrToInt(DelayTime, DelayTimeAsString) ||
               !cn36_StrToUInt8(DelayPoint, DelayPointAsString))
            {
                rc=0;
            }
        }

        if(rc && HiddenParameterWait_cn37.IsSet())
        {
            const char * NumberOfWaitsAsString;
            const char * WaitAfterEveryBytesAsString;

            HiddenParameterWait_cn37.GiveValues(NumberOfWaitsAsString, WaitAfterEveryBytesAsString);

            if(!cn36_StrToInt(NumberOfWaits, NumberOfWaitsAsString) ||
               !cn36_StrToUInt8(WaitAfterEveryBytes, WaitAfterEveryBytesAsString))
            {
                rc=0;
            }
        }

        if(rc)
        {
            char *Buffer;
                
            Buffer=new char [BufferSize];

            if(0==Buffer)
            {
                rc=0;
                ErrorHandler_ni35->Report(tni35_AStandardError::OutOfMemory.Set());
            }
            else
            {
                if(!strcmp(Direction_cn37.GiveValue(), "pipe2file"))
                {
                    if(!cn37_ReadPipeWriteFiles(PipeName.GiveValue(),
                                                FileName_cn37.GiveValue(),
                                                Buffer,
                                                BufferSize,
                                                UseFileSizeBound,
                                                FileSizeBound,
                                                DelayTime,
                                                DelayPoint,
                                                NumberOfWaits,
                                                WaitAfterEveryBytes))
                    {
                        rc=0;
                    }
                }
                else
                    if(!strcmp(Direction_cn37.GiveValue(), "file2pipe"))
                    {
                        if(!cn37_ReadFilesWritePipe(FileName_cn37.GiveValue(),
                                                    PipeName.GiveValue(),
                                                    Buffer,
                                                    BufferSize,
                                                    UseFileSizeBound))
                        {
                            rc=0;
                        }
                    }
                    else
                        if(!strcmp(Direction_cn37.GiveValue(), "test2pipe"))
                        {
                            if(!cn37_ConstructTestDataWritePipe(PipeName.GiveValue(),
                                                                Buffer,
                                                                BufferSize,
                                                                FileSizeBound,
                                                                DelayTime,
                                                                DelayPoint))
                            {
                                rc=0;
                            }
                        }
                        else
                            if(!strcmp(Direction_cn37.GiveValue(), "compare"))
                            {
                                char * Buffer2=new char [BufferSize];

                                if(0==Buffer)
                                {
                                    rc=0;
                                    ErrorHandler_ni35->Report(tni35_AStandardError::OutOfMemory.Set());
                                }
                                else
                                {
                                    rc=cn37_CompareFileAndPipe(FileName_cn37.GiveValue(),
                                                               PipeName.GiveValue(),
                                                               Buffer,
                                                               Buffer2,
                                                               BufferSize,
                                                               UseFileSizeBound);

                                    delete [] Buffer2;
                                }
                            }

                delete [] Buffer;
            }
        }
    }

    if(1==rc)      // change return code from C-meaning to process return code meaning and don't change rc==2
        rc=0;
    else
        if(0==rc)
            rc=1;

    return rc;
}
