/* 
  -----------------------------------------------------------------------------
 
  module: vcn41.cpp
 
  -----------------------------------------------------------------------------
 
  responsible:  BerndV
 
  special area: DBMServer Configuration
   
  description:  DBMServer Configuration - Implementation

  -----------------------------------------------------------------------------


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/

/*
  -----------------------------------------------------------------------------
  includes
  -----------------------------------------------------------------------------
 */
#include <stdio.h>

#include "vsp001.h"
#include "vsp009c.h"
#include "geo28.h"

#include "hin101.h"

#include "hcn20.h"
#include "hcn50.h"
#include "hcn42.h"
#include "hcn51.h"
#include "hcn53.h"
#include "hcn90.h"
#include "hcn41.h"

/*
  -----------------------------------------------------------------------------
  private macros
  -----------------------------------------------------------------------------
 */
#define CFG_RAWDATA "-raw"
/*
  -----------------------------------------------------------------------------
  function:     cn41DBMSetValue
  -----------------------------------------------------------------------------
 */
tcn00_Error cn41DBMSetValue
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  _TCHAR  szValue [PARAMETER_MAXLEN_CN90];
  _TCHAR  szName  [PARAMETER_MAXLEN_CN90];
  boolean bRaw = false;

  cn90GetToken(command->args, szName,     1);
  cn90GetToken(command->args, szValue,    2);

  if (_tcsicmp(szName, CFG_RAWDATA) == 0) {
    bRaw = true;
    cn90GetToken(command->args, szName, 2);
    if (!cn90GetToken(command->args, szValue, 3)) {
      szValue[0]='\0';  //force the deletion if no value was supplied (for type raw)
    } // end if
  } // end if

  if (nFuncReturn == OK_CN00) {
    cn51DBMConfigValue cfgValue(vcontrol->dbname, *(tsp00_Namec *) szName);

    if (bRaw) {
      tsp00_C512c aValue;
      aValue.rawAssign(szValue);
      cfgValue = aValue;
    } else {
      tsp00_Namec aValue;
      aValue.rawAssign(szValue);
      cfgValue = aValue;
    } // end if
    if (cfgValue.LastEvent()) {
      nFuncReturn = cn90AnswerEvent(replyData, replyLen, cfgValue.LastEvent());
    } else {
      cn90AnswerOK(replyData, replyLen, NULL);
    } // end if
  } // end if

  return nFuncReturn;
} // end cn41DBMSetValue

/*
  -----------------------------------------------------------------------------
  function:     cn41DBMGetValue
  -----------------------------------------------------------------------------
 */
tcn00_Error cn41DBMGetValue
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  _TCHAR  szName  [PARAMETER_MAXLEN_CN90];
  boolean bRaw = false;
  
  cn90GetToken(command->args, szName, 1);
  if (_tcsicmp(szName, CFG_RAWDATA) == 0) {
    bRaw = true;
    if (!cn90GetToken(command->args, szName, 2)) {
      nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
    } // end if
  } // end if

  if (nFuncReturn == OK_CN00) {
    if (bRaw) {
      cn51DBMConfigValue cfgValue(vcontrol->dbname,  *(tsp00_Namec *) szName);
      sprintf(replyData, "%s%s%s%s", ANSWER_OK_CN00, 
                                     LINE_SEPSTRING_CN00,
                                     (const char *) (tsp00_C512c) cfgValue,
                                     LINE_SEPSTRING_CN00);
    } else {
      cn51DBMConfigValue cfgValue(vcontrol->dbname,  *(tsp00_Namec *) szName);
      sprintf(replyData, "%s%s%s%s", ANSWER_OK_CN00, 
                                     LINE_SEPSTRING_CN00,
                                     (const char *) (tsp00_Namec) cfgValue,
                                     LINE_SEPSTRING_CN00);
    } // end if
  } // end if

  *replyLen = (int)strlen(replyData);

  return nFuncReturn;
} // end cn41DBMGetValue

/*
  -----------------------------------------------------------------------------
  function:     cn41UserLogon
  -----------------------------------------------------------------------------
 */
tcn00_Error cn41UserLogon
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  _TCHAR              szToken    [PARAMETER_MAXLEN_CN90];
  _TCHAR            * pPassword   = NULL; 
  tsp00_UserNamec     szUser;
  tsp00_Namec         szPassword;
  tsp_cryptpw         cryptPwd;
  tsp_namec           szPwd;
  tsp9_rte_xerror     xError;
  tsp00_C64c          szColdUser;
  tsp00_C64c          szColdPwd;
  bool                bErrSysLogon = false;

  cn90GetToken(command->args, szToken, 1);
  pPassword = _tcschr(szToken, PWD_SEPARATOR_CN00);
  if (pPassword != NULL) {
    *pPassword = CHAR_STRINGTERM_CN90;
    pPassword++;
    cn90Uncrypt(pPassword, false);
    szUser.rawAssign(szToken);
    szPassword.rawAssign(pPassword);

    vcontrol->pCurrentUser->setDBName   ( vcontrol->dbname )
                           .setUserName ( szUser );

    nFuncReturn = vcontrol->pCurrentUser->reload();
    if (nFuncReturn == ERR_USRREAD_CN00) {
      // check if upc is empty 
      tsp00_Int4              nLength = 0;
      tsp00_Uint4             nToken = FIRST_TOKEN_EO28;
      tsp00_UserNamec         szUserName;
      teo200_EventList        EventList;
      cn51DBMConfigValue      cfgColdUsr(vcontrol->dbname, CFG_COLDUSR_CN51);

      {
        teo28_ProfileContainer  Profile(vcontrol->dbname, &EventList);
        Profile.eo28_ScanProfiles(nToken, nLength, NULL, szUserName, &EventList);
      }
    
      if(!EventList && nToken == NO_TOKEN_EO28) {
        // Yes, upc is empty -> create user
        nFuncReturn = OK_CN00;

        // check User against param file 
        if (cn20XParamGetValue(vcontrol->dbname, XP_ID_CONTROLUSERID_CN00, szColdUser)  == OK_CN00 &&
            cn20XParamGetValue(vcontrol->dbname, XP_ID_CONTROLPASSWORD_CN00, szColdPwd) == OK_CN00     ) {
          cn90Uncrypt( szColdPwd, false);
          if (_tcsicmp(szColdUser, szUser)     != 0 || 
              _tcsicmp(szColdPwd,  szPassword) != 0    ) {
            nFuncReturn = ERR_USRFAIL_CN00;
          } // end if
        } // end if

        if (nFuncReturn == OK_CN00) {
#ifndef _WIN32
          // restore upc at UNIX only if a system logon is given // PTS 1108551
          if (vcontrol->bLocalSysLogon || vcontrol->bSysLogon) {
#endif
            vcontrol->pCurrentUser->setMasterPwd    ( szPassword  )
                                   .setDBMUser      ( true )
                                   .setColdUser     ( )
                                   .setServerRights ( DBMMaskAll_CN50, DBMUserDefault_CN50 )
                                   .setGUIRights    ( DBMMaskAll_CN50, DBMUserDefaultGUI_CN50 );

            nFuncReturn = vcontrol->pCurrentUser->save(true);
            if (nFuncReturn == OK_CN00) {
              cfgColdUsr = vcontrol->pCurrentUser->getUserName();
              // make a copy from the upc
              cn42BackupFile(vcontrol->dbname, FGET_USER_CN42);
            } else {
              nFuncReturn = (nFuncReturn != ERR_USRFRM_CN00) ? ERR_USRFAIL_CN00 : nFuncReturn;
            } // end if
#ifndef _WIN32
          } else {
            nFuncReturn = ERR_USRREAD_CN00;
          } // end if 
#endif
        } // end if 
      } else {
        // No, upc is not empty -> wrong username!
        nFuncReturn = ERR_USRFAIL_CN00;
      } // end if 
    } // end if 

    if (nFuncReturn == OK_CN00) {
      if (!vcontrol->pCurrentUser->checkMasterPwd( szPassword) &&
          !vcontrol->pCurrentUser->checkSecondPwd( szPassword)    ) {
        nFuncReturn = ERR_USRFAIL_CN00;
      } else {
        // system logon
        if (vcontrol->pCurrentUser->getSystemName().length() > 0 && !vcontrol->bLocalSysLogon) {
          cryptPwd = vcontrol->pCurrentUser->getSystemPwd();
          cn90CalculateUncrypt(cryptPwd, szPwd, false);
          sqlxlogon (vcontrol->pCurrentUser->getSystemName(), szPwd, &xError);
          if (xError.xe_result != 0) {
            bErrSysLogon = true;
#ifndef _WIN32
            // try only impersonate with os user
            sqlxlogon (vcontrol->pCurrentUser->getSystemName(), NULL, &xError);
            if (xError.xe_result != 0) {
              // try impersonate with UPC owner
              if (vcontrol->pCurrentUser->getUPCOwner().length() > 0) {
                sqlxlogon (vcontrol->pCurrentUser->getUPCOwner(), NULL, &xError);
                if (xError.xe_result != 0) {
                  nFuncReturn = ERR_IMPERSONATE_CN00;
                } // end if
              } else {
                nFuncReturn = ERR_IMPERSONATE_CN00;
              } // end if 
            } // end if
#endif // _WIN32
          } else {
            vcontrol->bSysLogon = true;
          } // end if
#ifndef _WIN32
        } else { 
          // impersonate
          if (!vcontrol->bLocalSysLogon && !vcontrol->bSysLogon) {
            if (vcontrol->pCurrentUser->getUPCOwner().length() > 0) {
              sqlxlogon (vcontrol->pCurrentUser->getUPCOwner(), NULL, &xError);
              if (xError.xe_result != 0) {
                nFuncReturn = ERR_IMPERSONATE_CN00;
              } // end if
            } else {
              nFuncReturn = ERR_IMPERSONATE_CN00;
            } // end if 
          } // end if 
#endif // _WIN32
        } // end if
      } // end if
    } // end if
  } else {
    nFuncReturn = ERR_PARAM_CN00;
  } // end if

  if (nFuncReturn == OK_CN00 || bErrSysLogon) {

    // enable protocol
    tin100_GlobalLog::deleteObject();
    tin100_GlobalLog * aLog = tin100_GlobalLog::createObject();
    if (aLog != NULL) {
      tsp00_Pathc szLogFile;

      cn42GetFileName(vcontrol->dbname, FGET_DBMPRT_CN42, szLogFile);

      tin105_SequentialLogFile * aLogFile = new tin105_SequentialLogFile(szLogFile, LOG_ALL_MIN105);
      if (aLogFile != NULL) {
        if (aLogFile->checkFile()) {
          aLog->addFile(aLogFile, true);
        } else {
          delete aLogFile;
          cn42GetFileName(vcontrol->dbname, FGET_DBMSRV_CN42, szLogFile);
          aLogFile = new tin105_SequentialLogFile(szLogFile, LOG_ALL_MIN105);
          if (aLogFile != NULL) {
            aLog->addFile(aLogFile, true);
          } // end if
        } // end if
      } // end if
    } // end if

    // check XDB
    if (vcontrol->dbname.length() > 0) {
      cn51DBMConfigValue  cfgXDBUsr (vcontrol->dbname, CFG_XDBSTATE_CN51);
      if (cfgXDBUsr != 0) {
        tin101_XDBOperation::enable ();
      } // end if
      cn51DBMConfigValue  cfgDbaDtl (vcontrol->dbname, CFG_DBADTL_CN51);
      if (cfgDbaDtl != 0) {
        tcn53_DBAAction::enable ();
      } // end if
    } // end if

    // check Unicode
    if (vcontrol->dbname.length() > 0) {
      tsp00_C64c        szUnicode;
      tsp00_XpKeyTypec  szParameter;
      szParameter.rawAssign(PAN_UNICODE_CATALOG);
      if (cn20XParamGetValue(vcontrol->dbname, szParameter, szUnicode) == OK_CN00) {
        vcontrol->bUnicode = (stricmp(szUnicode, XP_VAL_YES_CN00) == 0);
      } // end if

    } // end if

  } // end if

  if (nFuncReturn == OK_CN00) {
    cn90AnswerOK(replyData, replyLen, NULL);
  } else {
    if (bErrSysLogon) {
      _stprintf(replyData, "%s%s%s%s%s%s",
                           ANSWER_OK_CN00,
                           LINE_SEPSTRING_CN00,
                           _T("INFO: Logon to operating system failed!"),
                           LINE_SEPSTRING_CN00,
                           xError.xe_text,
                           LINE_SEPSTRING_CN00);
      *replyLen = (int)_tcslen(replyData);
    } else {
      vcontrol->pCurrentUser->clear();
      cn90AnswerIError(replyData, replyLen, nFuncReturn);
    } // end if
  } // end if

  return nFuncReturn;
} // cn41UserLogon

/*
  -----------------------------------------------------------------------------
  function:     cn41UserSystem
  -----------------------------------------------------------------------------
 */
tcn00_Error cn41UserSystem
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;

  _TCHAR              szToken    [PARAMETER_MAXLEN_CN90];
  _TCHAR            * pPassword   = NULL; 
  tsp00_UserNamec     szUser;
  tsp00_Namec         szPassword;
  tsp9_rte_xerror     xError;

  cn90GetToken(command->args, szToken, 1);
  pPassword = _tcschr(szToken, PWD_SEPARATOR_CN00);
  if (pPassword != NULL) {
    *pPassword = CHAR_STRINGTERM_CN90;
    pPassword++;
    cn90Uncrypt(pPassword, false);
    szUser.rawAssign(szToken);
    szPassword.rawAssign(pPassword);

    sqlxlogon (szUser.asCharp(), szPassword.asCharp(), &xError);
    if (xError.xe_result != 0) {
      nFuncReturn = cn90AnswerRTEError (replyData, replyLen, ERR_SYSLOGON_CN00, xError.xe_text, sizeof(xError.xe_text), xError.xe_result);
    } else {
      vcontrol->bSysLogon = true;
    } // end if
  
  } else {
    nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
  } // end if

  if (nFuncReturn == OK_CN00) {
    cn90AnswerOK(replyData, replyLen, NULL);
  } // end if

  return nFuncReturn;
} // cn41UserSystem

/*
  -----------------------------------------------------------------------------
  function:     cn41UserCreate
  -----------------------------------------------------------------------------
 */
tcn00_Error cn41UserCreate
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  _TCHAR              szToken    [PARAMETER_MAXLEN_CN90];
  _TCHAR            * pPassword   = NULL; 
  tsp00_UserNamec     szUser;
  tsp00_Namec         szPassword;
  _TCHAR              szTemplate  [PARAMETER_MAXLEN_CN90];
  cn50DBMUser         dbUser;
  bool                bTemplate   = false;
    
  // read parameters
  cn90GetToken(command->args, szToken, 1);
  if ( cn90GetToken(command->args, szTemplate, 2)) {
    bTemplate = true;
  } // end if

  // extract password
  pPassword = _tcschr(szToken, PWD_SEPARATOR_CN00);
  if (pPassword != NULL) {
    *pPassword = CHAR_STRINGTERM_CN90;
    pPassword++;
    szUser.rawAssign(szToken);
    szPassword.rawAssign(pPassword);
  } else {
    nFuncReturn = ERR_PARAM_CN00;
  } // end if

  // check existence of user
  if (nFuncReturn == OK_CN00) {
    dbUser.setDBName   ( vcontrol->dbname )
          .setUserName ( szUser );

    if (dbUser.reload() == OK_CN00) {
      nFuncReturn = ERR_USREXISTS_CN00;
    } // end if
  } // end if

  // read Template
  if (nFuncReturn == OK_CN00 && bTemplate) {
    dbUser.setDBName   ( vcontrol->dbname )
          .setUserName ( *(tsp_usernamec *) szTemplate );

    nFuncReturn = dbUser.reload();
  } // end if

  if (nFuncReturn == OK_CN00) {
    cn90Uncrypt(pPassword, false);
    dbUser.setDBName    ( vcontrol->dbname )
          .setUserName  ( szUser )
          .setMasterPwd ( szPassword )
          .setColdUser  ( false )
          .setSDBAUser  ( false );
    
    nFuncReturn = dbUser.save();
  } // end if
  
  if (nFuncReturn == OK_CN00) {
    // make a copy from the upc
    cn42BackupFile(vcontrol->dbname, FGET_USER_CN42);
    cn90AnswerOK(replyData, replyLen, NULL);
  } else {
    cn90AnswerIError(replyData, replyLen, nFuncReturn);
  } // end if

  return nFuncReturn;
} // cn41UserCreate

/*
  -----------------------------------------------------------------------------
  function:     cn41UserGet
  -----------------------------------------------------------------------------
 */
tcn00_Error cn41UserGet
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error    nFuncReturn = OK_CN00;
  _TCHAR       * pData;
  _TCHAR         szName      [PARAMETER_MAXLEN_CN90];

  cn90GetToken(command->args, szName,     1);

  cn50DBMUser  dbUser( vcontrol->dbname, 
                       *(tsp_usernamec *) szName );
  
  nFuncReturn = dbUser.reload();
  if (nFuncReturn == OK_CN00) {
    cn90AnswerOK(replyData, replyLen, NULL);
    pData = replyData + _tcslen(replyData);
    dbUser.getData(pData);
    *replyLen = (int)_tcslen(replyData);
  } else {
    cn90AnswerIError(replyData, replyLen, nFuncReturn);
  } // end if

  return nFuncReturn;
} // cn41UserGet

/*
  -----------------------------------------------------------------------------
  function:     cn41UserGetRights
  -----------------------------------------------------------------------------
 */
tcn00_Error cn41UserGetRights
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error    nFuncReturn = OK_CN00;
  _TCHAR       * pData;
  _TCHAR         szName      [PARAMETER_MAXLEN_CN90];
  _TCHAR         szRights    [PARAMETER_MAXLEN_CN90];

  cn90GetToken(command->args, szName, 1);
  cn90GetToken(command->args, szRights, 2);

  cn50DBMUser  dbUser( vcontrol->dbname, 
                       *(tsp_usernamec *) szName );
    
  nFuncReturn = dbUser.reload();
  if (nFuncReturn == OK_CN00) {
    cn90AnswerOK(replyData, replyLen, NULL);
    pData = replyData + _tcslen(replyData);
    nFuncReturn = dbUser.listRights(pData, szRights);
  } // end if
  
  if (nFuncReturn == OK_CN00) {
    *replyLen = (int)_tcslen(replyData);
  } else {
    cn90AnswerIError(replyData, replyLen, nFuncReturn);
  } // end if

  return nFuncReturn;
} // cn41UserGetRights

/*
  -----------------------------------------------------------------------------
  function:     cn41UserGetAll
  -----------------------------------------------------------------------------
 */
tcn00_Error cn41UserGetAll
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error             nFuncReturn = OK_CN00;
  teo200_EventList        EventList;
  teo28_ProfileContainer  Profile(vcontrol->dbname, &EventList);
  tsp00_Uint4             nToken = FIRST_TOKEN_EO28;
  tsp00_UserNamec         szUserName;
  _TCHAR                * pCurrent = replyData;
  tsp00_Int4              nLength = 0;

  if (!EventList) {
    Profile.eo28_ScanProfiles(nToken, nLength, NULL, szUserName, &EventList);
    if(!EventList) {
      if (nToken == NO_TOKEN_EO28) {
        nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_NOUSERS_CN00);
      } else {
        // Init OK Message
        _stprintf(pCurrent, "%s%s%s%s", ANSWER_OK_CN00,        LINE_SEPSTRING_CN00,
                                        (_TCHAR *) szUserName, LINE_SEPSTRING_CN00);
        pCurrent = pCurrent + _tcslen(pCurrent);
        while (nToken != LAST_TOKEN_EO28 && !EventList) {
          Profile.eo28_ScanProfiles(nToken, nLength, NULL, szUserName, &EventList);
          if(!EventList && _tcsicmp(USERNAME_SUPERUSER_CN50, szUserName) != 0) {
            _stprintf(pCurrent, "%s%s", (_TCHAR *) szUserName, LINE_SEPSTRING_CN00);
            pCurrent = pCurrent + _tcslen(pCurrent);
          } // end if
        } // end while
      } // end if
    } // end if
  } // end if

  if (EventList) {
    nFuncReturn = cn90AnswerRTEError(replyData, replyLen, ERR_RTE_CN00, 
                                     EventList.eo200_EventMsg(),
                                     (int)_tcslen(EventList.eo200_EventMsg()),
                                     EventList.eo200_EventID());
  } // end if

  *replyLen = (int)_tcslen(replyData);

  return nFuncReturn;
} // cn41UserGetAll

/*
  -----------------------------------------------------------------------------
  function:     cn41UserPut
  -----------------------------------------------------------------------------
 */
tcn00_Error cn41UserPut
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error    nFuncReturn = OK_CN00;
  _TCHAR         szName      [PARAMETER_MAXLEN_CN90];
  const _TCHAR * pData       = NULL;
  cn50DBMUser    dbUser;
  
  cn90GetToken(command->args, szName, 1);

  dbUser.setDBName  ( vcontrol->dbname )
        .setUserName( *(tsp_usernamec *) szName           );

  nFuncReturn = dbUser.reload();

  if (nFuncReturn == OK_CN00) {
    pData = _tcschr(command->args, _T(' '));
    
    nFuncReturn = dbUser.setData(pData);

    if (nFuncReturn == OK_CN00) {
      nFuncReturn = dbUser.save();
      if (nFuncReturn == OK_CN00) {
        if (vcontrol->pCurrentUser->isThis(*(tsp_usernamec *) szName)) {
          nFuncReturn = vcontrol->pCurrentUser->reload();
        } // end if
      } // end if
    } // end if

  } // end if

  if (nFuncReturn == OK_CN00) {
    // make a copy from the upc
    cn42BackupFile(vcontrol->dbname, FGET_USER_CN42);
    cn90AnswerOK(replyData, replyLen, NULL);
  } else {
    cn90AnswerIError(replyData, replyLen, nFuncReturn);
  } // end if

  return nFuncReturn;
} // cn41UserPut

/*
  -----------------------------------------------------------------------------
  function:     cn41UserDelete
  -----------------------------------------------------------------------------
 */
tcn00_Error cn41UserDelete
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error    nFuncReturn = OK_CN00;
  _TCHAR         szName      [PARAMETER_MAXLEN_CN90];
  cn50DBMUser    dbUser;
  
  cn90GetToken(command->args, szName, 1);
  dbUser.setDBName  ( vcontrol->dbname )
        .setUserName( *(tsp_usernamec *) szName           );

  nFuncReturn = dbUser.reload();

  if (nFuncReturn == OK_CN00) {
    if (vcontrol->pCurrentUser->isThis(dbUser.getUserName())) {
      nFuncReturn = ERR_USRDELSELF_CN00;
    } else if (dbUser.isColdUser()) {
      nFuncReturn = ERR_USRDELCOLD_CN00;
    } else if (dbUser.isSDBAUser()) {
      nFuncReturn = ERR_USRDELSDBA_CN00;
    } else {
      nFuncReturn = dbUser.deleteUsr();
    } // end if

  } // end if

  if (nFuncReturn == OK_CN00) {
    cn90AnswerOK(replyData, replyLen, NULL);
  } else {
    cn90AnswerIError(replyData, replyLen, nFuncReturn);
  } // end if

  return nFuncReturn;
} // end cn41UserDelete

#ifdef CN71

#define USRCLASS_DBM        "DBM"
#define USRCLASS_SYS        "SYS"
#define USRCLASS_DBA        "DBA"
#define USRCLASS_SQL        "SQL"

/*
  -----------------------------------------------------------------------------
  function:     cn41_71_SetUser
  -----------------------------------------------------------------------------
 */
tcn00_Error cn41_71_SetUser
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error    nFuncReturn = OK_CN00;
  char           szUserClass [PARAMETER_MAXLEN_CN90];
  char           szUser      [PARAMETER_MAXLEN_CN90];
  char           szPassword  [PARAMETER_MAXLEN_CN90];
  
  // read arguments
  cn90GetToken(command->args, szUserClass, 1);
  cn90GetToken(command->args, szUser,      2);
  cn90GetToken(command->args, szPassword,  3);
  cn90Uncrypt(szPassword, false);

  // generate OK-answer
  cn90AnswerOK(replyData, replyLen, NULL);

  if (strcmp(szUserClass, USRCLASS_DBM) == 0) {
    // check/create user
    _stprintf(command->args, "%s,%s", (char *) szUser, (char *) szPassword);
    nFuncReturn = cn41UserLogon( vcontrol,
                                 command,
                                 replyData,
                                 replyLen,
                                 replyLenMax );

  } else if (strcmp(szUserClass, USRCLASS_SYS) == 0) {
    // attach os logon to current user
    _stprintf(command->args, "%s SYSTEMNAME=%s SYSTEMPASSWORD=%s", 
                             (char *) vcontrol->pCurrentUser->getUserName(),
                             (char *) szUser, 
                             (char *) szPassword);
    nFuncReturn = cn41UserPut( vcontrol,
                               command,
                               replyData,
                               replyLen,
                               replyLenMax );
    if (nFuncReturn == OK_CN00) {
      // reconnect to activate system logon
      _stprintf(command->args, "%s,%s", 
                               (char *) vcontrol->pCurrentUser->getUserName(),
                               (char *) vcontrol->pCurrentUser->getMasterPwd(true));
      nFuncReturn = cn41UserLogon( vcontrol,
                                   command,
                                   replyData,
                                   replyLen,
                                   replyLenMax );
    } // end if
  } else if (strcmp(szUserClass, USRCLASS_SQL) == 0) {
    // do nothing
  } else if (strcmp(szUserClass, USRCLASS_DBA) == 0) {
    // check/create dba
  } else {
    nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
  } // end if

  return nFuncReturn;

} // end cn41_71_SetUser

/*
  -----------------------------------------------------------------------------
  function:     cn41_71_ParamLogon
  -----------------------------------------------------------------------------
 */
tcn00_Error cn41_71_ParamLogon
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error    nFuncReturn = OK_CN00;
  char           szUser      [PARAMETER_MAXLEN_CN90];
  char           szPassword  [PARAMETER_MAXLEN_CN90];
  
  // read arguments
  cn90GetToken(command->args, szUser,      1);
  cn90GetToken(command->args, szPassword,  2);
  cn90Uncrypt(szPassword, false);

  // check/create user
  _stprintf(command->args, "%s,%s", (char *) szUser, (char *) szPassword);
  nFuncReturn = cn41UserLogon( vcontrol,
                               command,
                               replyData,
                               replyLen,
                               replyLenMax );

  return nFuncReturn;

} // end cn41_71_ParamLogon

/*
  -----------------------------------------------------------------------------
  function:     cn41_71_Logon
  -----------------------------------------------------------------------------
 */
tcn00_Error cn41_71_Logon
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error    nFuncReturn = OK_CN00;
  char           szUser      [PARAMETER_MAXLEN_CN90];
  char           szPassword  [PARAMETER_MAXLEN_CN90];
  
  // read arguments
  cn90GetToken(command->args, szUser,      1);
  cn90GetToken(command->args, szPassword,  2);
  cn90Uncrypt(szPassword, false);

  // attach os logon to current user
  _stprintf(command->args, "%s SYSTEMNAME=%s SYSTEMPASSWORD=%s", 
                           (char *) vcontrol->pCurrentUser->getUserName(),
                           (char *) szUser, 
                           (char *) szPassword);
  nFuncReturn = cn41UserPut( vcontrol,
                             command,
                             replyData,
                             replyLen,
                             replyLenMax );
  if (nFuncReturn == OK_CN00) {
    // reconnect to activate system logon
    _stprintf(command->args, "%s,%s", 
                             (char *) vcontrol->pCurrentUser->getUserName(),
                             (char *) vcontrol->pCurrentUser->getMasterPwd(true));
    nFuncReturn = cn41UserLogon( vcontrol,
                                 command,
                                 replyData,
                                 replyLen,
                                 replyLenMax );
  } // end if

  return nFuncReturn;

} // end cn41_71_Logon

#endif /* CN71 */
