/* 
  -----------------------------------------------------------------------------
 
  module: vcn50.cpp
 
  -----------------------------------------------------------------------------
 
  responsible:  BerndV
 
  special area: DBMServer User Management
   
  description:  DBMServer User Management - Implementation

  version:      7.2.* 

  -----------------------------------------------------------------------------


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/

/*
  -----------------------------------------------------------------------------
  includes
  -----------------------------------------------------------------------------
 */
#include "heo02.h"
#include "hsp02.h"
#include "geo47.h"
#include "heo11.h"
#include "geo28.h"
#include "hcn29.h"
#include "hcn51.h"
#include "hcn80.h"
#include "hcn90.h"
#include "hcn50.h"

/*
  -----------------------------------------------------------------------------
  private macros
  -----------------------------------------------------------------------------
 */
// server right to text structure
#define ServerRightToText                                    \
{                                                            \
  {DBMRightDBInfoRead_CN50      , _T("DBInfoRead"        ) , _T("Request status data"               )}, \
  {DBMRightExecLoad_CN50        , _T("ExecLoad"          ) , _T("Execute the LOAD program"          )}, \
  {DBMRightSystemCmd_CN50       , _T("SystemCmd"         ) , _T("Execute operating system commands" )}, \
  {DBMRightUserMgm_CN50         , _T("UserMgm"           ) , _T("User management"                   )}, \
  {DBMRightFileFull_CN50        , _T("FileFull"          ) , _T("File access (read and write)"      )}, \
  {DBMRightFileRead_CN50        , _T("FileRead"          ) , _T("File access (read only)"           )}, \
  {DBMRightDBFileRead_CN50      , _T("DBFileRead"        ) , _T("Database file access (read only)"  )}, \
  {DBMRightBackupSave_CN50      , _T("Backup"            ) , _T("Saving backups"                    )}, \
  {DBMRightInstallMgm_CN50      , _T("InstallMgm"        ) , _T("Installation management"           )}, \
  {DBMRightLoadSysTab_CN50      , _T("LoadSysTab"        ) , _T("Load the system tables"            )}, \
  {DBMRightParamCheckWrite_CN50 , _T("ParamCheckWrite"   ) , _T("Parameter access (checked write)"  )}, \
  {DBMRightParamFull_CN50       , _T("ParamFull"         ) , _T("Parameter access (read and write)" )}, \
  {DBMRightParamRead_CN50       , _T("ParamRead"         ) , _T("Parameter access (read only)"      )}, \
  {DBMRightDBStart_CN50         , _T("DBStart"           ) , _T("Start database instance"           )}, \
  {DBMRightDBStop_CN50          , _T("DBStop"            ) , _T("Stop database instance"            )}, \
  {DBMRightBackupRestore_CN50   , _T("Recovery"          ) , _T("Restoring backups"                 )}, \
  {DBMRightAccessSQL_CN50       , _T("AccessSQL"         ) , _T("Access to SQL session"             )}, \
  {DBMRightAccessUtility_CN50   , _T("AccessUtility"     ) , _T("Access to utility session"         )}, \
  {DBMRightUnused11_CN50        , _T(""                  ) , _T(""                                  )}, \
  {DBMRightUnused12_CN50        , _T(""                  ) , _T(""                                  )}, \
  {DBMRightUnused20_CN50        , _T(""                  ) , _T(""                                  )}, \
  {DBMRightUnused21_CN50        , _T(""                  ) , _T(""                                  )}, \
  {DBMRightUnused22_CN50        , _T(""                  ) , _T(""                                  )}, \
  {DBMRightUnused23_CN50        , _T(""                  ) , _T(""                                  )}, \
  {DBMRightUnused24_CN50        , _T(""                  ) , _T(""                                  )}, \
  {DBMRightUnused25_CN50        , _T(""                  ) , _T(""                                  )}, \
  {DBMRightUnused26_CN50        , _T(""                  ) , _T(""                                  )}, \
  {DBMRightUnused27_CN50        , _T(""                  ) , _T(""                                  )}, \
  {DBMRightUnused28_CN50        , _T(""                  ) , _T(""                                  )}, \
  {DBMRightUnused29_CN50        , _T(""                  ) , _T(""                                  )}, \
  {DBMRightUnused30_CN50        , _T(""                  ) , _T(""                                  )}, \
  {DBMRightUnused31_CN50        , _T(""                  ) , _T(""                                  )}, \
  {DBMMaskNothing_CN50          , _T(""                  ) , _T(""                                  )}  \
}

#define GUIRightToText                                       \
{                                                            \
  {DBMGUIRight01_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight02_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight03_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight04_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight05_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight06_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight07_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight08_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight09_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight10_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight11_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight12_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight13_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight14_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight15_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight16_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight17_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight18_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight19_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight20_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight21_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight22_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight23_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight24_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight25_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight26_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight27_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight28_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight29_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight30_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMGUIRight31_CN50           , _T(""                  ) , _T(""                                  )}, \
  {DBMMaskNothing_CN50          , _T(""                  ) , _T(""                                  )}  \
}

// Keywords
#define KEY_SVRRIGHTS            _T("SERVERRIGHTS"   )
#define KEY_GUIRIGHTS            _T("GUIRIGHTS"      )
#define KEY_PWD                  _T("PASSWORD"       )
#define KEY_2PWD                 _T("SECONDPASSWORD" )
#define KEY_SYSNAME              _T("SYSTEMNAME"     )
#define KEY_SYSPWD               _T("SYSTEMPASSWORD" )
#define KEY_DBMUSR               _T("DBMUSER"        )
#define KEY_SQLUSR               _T("SQLUSER"        )
#define KEY_SQLUSERMODE          _T("SQLUSERMODE"    )
#define KEY_DISABLED             _T("DISABLED"       )
#define KEY_COMMENT              _T("COMMENT"        )
#define KEY_YES                  _T("YES"            )
#define KEY_NO                   _T("NO"             )
#define KEY_DBA                  _T("DBA"            )
#define KEY_RESOURCE             _T("RESOURCE"       )
#define KEY_STANDARD             _T("STANDARD"       )

// Special Characters
#define CHAR_ASSIGN              _T('=')
#define CHAR_RIGHTDELIMETER      _T(',')
#define CHAR_RIGHTGRANT          _T('+')
#define CHAR_RIGHTREMOVE         _T('-')
   
    
/*    
  =============================================================================
  class:        cn50DBMUser
  =============================================================================
 */

/*
  -----------------------------------------------------------------------------
  public constructor cn50DBMUser :: cn50DBMUser
  -----------------------------------------------------------------------------
 */
cn50DBMUser :: cn50DBMUser ( tsp00_DbNamec const szDBName, tsp00_UserNamec const szUserName)
{
  clear();
  setDBName(szDBName);
  setUserName(szUserName);
  reload();
} // end cn50DBMUser :: cn50DBMUser

/*
  -----------------------------------------------------------------------------
  public constructor cn50DBMUser :: cn50DBMUser
  -----------------------------------------------------------------------------
 */
cn50DBMUser :: cn50DBMUser ( )
{
  clear();
} // end cn50DBMUser :: cn50DBMUser

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: setDBName
  -----------------------------------------------------------------------------
 */
cn50DBMUser & cn50DBMUser :: setDBName ( tsp00_DbNamec const szValue )
{
 
  Data.szDBName = szValue;

  bDirty = true;
  return *this;
} // end cn50DBMUser :: setDBName

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: setUserName
  -----------------------------------------------------------------------------
 */
cn50DBMUser & cn50DBMUser :: setUserName ( tsp00_UserNamec const szValue )
{
 
  Data.szUserName = szValue;

  bDirty = true;
  return *this;
} // end cn50DBMUser :: setUserName

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: setMasterPwd
  -----------------------------------------------------------------------------
 */
cn50DBMUser & cn50DBMUser :: setMasterPwd ( tsp00_Namec const szValue )
{
  tsp00_Namec tmpName;
  tsp00_Name  tmpValue;

  cn90StrUpperCopy(tmpName, szValue, false);

  tmpValue.c2p(tmpName);

  s02applencrypt( tmpValue, Data.cryptMasterPwd);

  bDirty = true;
  bPwdDirty = true;

  return *this;
} // end cn50DBMUser :: setMasterPwd

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: setSecondPwd
  -----------------------------------------------------------------------------
 */
cn50DBMUser & cn50DBMUser :: setSecondPwd ( tsp00_Namec const szValue )
{

  if (szValue.length() == 0) {
    Data.bSecondPwd = false;
  } else {
    tsp00_Namec tmpName;
    tsp00_Name  tmpValue;

    cn90StrUpperCopy(tmpName, szValue, false);

    tmpValue.c2p(tmpName);
 
    s02applencrypt( tmpValue, Data.cryptSecondPwd);
    Data.bSecondPwd = true;

  } // end if

  bDirty = true;
  bSecondPwdDirty = true;

  return *this;
} // end cn50DBMUser :: setSecondPwd

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: checkMasterPwd
  -----------------------------------------------------------------------------
 */
bool cn50DBMUser :: checkMasterPwd ( tsp00_Namec const szValue )
{
  tsp00_Namec     tmpName;
  tsp00_CryptPw   tmpMasterPwd;
  tsp00_Name      tmpValue;

  cn90StrUpperCopy(tmpName, szValue, false);

  tmpValue.c2p(tmpName);

  s02applencrypt( tmpValue, tmpMasterPwd);

  bLoggedOn =     (memcmp(&tmpMasterPwd, &Data.cryptMasterPwd, sizeof(tsp00_CryptPw)) == 0) 
               && !isDisabled() 
               && isDBMUser();

  return  bLoggedOn;
} // end cn50DBMUser :: checkMasterPwd

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: checkSecondPwd
  -----------------------------------------------------------------------------
 */
bool cn50DBMUser :: checkSecondPwd ( tsp00_Namec const szValue ) 
{
  tsp00_Namec     tmpName;
  tsp00_CryptPw   tmpSecondPwd;
  tsp00_Name      tmpValue;

  cn90StrUpperCopy(tmpName, szValue, false);

  tmpValue.c2p(tmpName);

  s02applencrypt( tmpValue, tmpSecondPwd);

  bLoggedOn =    (memcmp(&tmpSecondPwd, &Data.cryptSecondPwd, sizeof(tsp00_CryptPw)) == 0) 
              && Data.bSecondPwd  
              && !isDisabled() 
              && isDBMUser();

  return bLoggedOn;
} // end cn50DBMUser :: checkSecondPwd

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: setSystemName
  -----------------------------------------------------------------------------
 */
cn50DBMUser & cn50DBMUser :: setSystemName ( tsp00_UserNamec const szValue )
{
 
  Data.szSystemName = szValue;

  bDirty = true;
  return *this;
} // end cn50DBMUser :: setSystemName

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: setSystemPwd
  -----------------------------------------------------------------------------
 */
cn50DBMUser & cn50DBMUser :: setSystemPwd ( tsp00_Namec const szValue )
{
  tsp00_Name tmpValue;

  tmpValue.c2p(szValue);
 
  s02applencrypt( tmpValue, Data.cryptSystemPwd);

  bDirty = true;
  return *this;
} // end cn50DBMUser :: setSystemPwd

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: setDBMUser
  -----------------------------------------------------------------------------
 */
cn50DBMUser & cn50DBMUser :: setDBMUser ( bool const bValue )
{
  Data.bDBMUser = bValue;

  bDirty = true;
  return *this;
} // end cn50DBMUser :: setDBMUser

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: setSQLUser
  -----------------------------------------------------------------------------
 */
cn50DBMUser & cn50DBMUser :: setSQLUser ( bool const bValue )
{
  Data.bSQLUser = bValue;

  bDirty = true;
  return *this;
} // end cn50DBMUser :: setSQLUser

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: setColdUser
  -----------------------------------------------------------------------------
 */
cn50DBMUser & cn50DBMUser :: setColdUser ( bool const bValue )
{
  Data.bColdUser = bValue;

  bDirty = true;
  return *this;
} // end cn50DBMUser :: setColdUser

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: setSDBAUser
  -----------------------------------------------------------------------------
 */
cn50DBMUser & cn50DBMUser :: setSDBAUser ( bool const bValue )
{
  Data.bSDBAUser   = bValue;

  bDirty = true;
  return *this;
} // end cn50DBMUser :: setSDBAUser

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: setSQLUsermode
  -----------------------------------------------------------------------------
 */
cn50DBMUser & cn50DBMUser :: setSQLUsermode  ( cn50DBMUser::SQLUserMode const eValue )
{
  Data.eUserMode = eValue;

  bDirty    = true;
  bSQLDirty = true;
  return *this;
} // end cn50DBMUser :: setSQLUsermode

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: setDisabled
  -----------------------------------------------------------------------------
 */
cn50DBMUser & cn50DBMUser :: setDisabled ( bool const bValue )
{
  Data.bDisabled = bValue;

  bDirty = true;
  return *this;
} // end cn50DBMUser :: setDisabled

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: setServerRights
  -----------------------------------------------------------------------------
 */
cn50DBMUser & cn50DBMUser :: setServerRights ( tsp00_Uint4 const nMask, tsp00_Uint4 const nValues )
{
  Data.dbmServerRights = (Data.dbmServerRights & ~nMask) | nValues;

  bDirty = true;
  return *this;
} // end cn50DBMUser :: setServerRights

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: setGUIRights
  -----------------------------------------------------------------------------
 */
cn50DBMUser & cn50DBMUser :: setGUIRights ( tsp00_Uint4 const nMask, tsp00_Uint4 const nValues )
{
  Data.dbmGUIRights = (Data.dbmGUIRights & ~nMask) | nValues;

  bDirty = true;
  return *this;
} // end cn50DBMUser :: setGUIRights

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: setComment
  -----------------------------------------------------------------------------
 */
cn50DBMUser & cn50DBMUser :: setComment ( tsp00_C64c const szValue )
{
 
  Data.szComment = szValue;

  bDirty = true;
  return *this;
} // end cn50DBMUser :: setComment

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: getDBName
  -----------------------------------------------------------------------------
 */
tsp00_DbNamec cn50DBMUser :: getDBName ( ) const
{
  return Data.szDBName;
} // end cn50DBMUser :: getDBName

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: getUserName
  -----------------------------------------------------------------------------
 */
tsp00_UserNamec cn50DBMUser :: getUserName ( ) const
{
  return Data.szUserName;
} // end cn50DBMUser :: getUserName

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: getMasterPwd
  -----------------------------------------------------------------------------
 */
tsp00_CryptPw cn50DBMUser :: getMasterPwd ( ) const
{
  return Data.cryptMasterPwd;
} // end cn50DBMUser :: getMasterPwd

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: getMasterPwd
  -----------------------------------------------------------------------------
 */
tsp00_C64c cn50DBMUser :: getMasterPwd ( bool bUncrypt ) const
{
  tsp00_C64c      szPwd;

  if (bUncrypt) {
    cn90CalculateUncrypt(Data.cryptMasterPwd, szPwd, false);
  } else {
    cn90CryptToCryptString(Data.cryptMasterPwd, szPwd);
  } // end if

  return szPwd;
} // end cn50DBMUser :: getMasterPwd

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: getClearMasterPwd
  -----------------------------------------------------------------------------
 */
tsp00_Namec cn50DBMUser :: getClearMasterPwd ( ) const
{
  tsp00_C64c      szPwd;
  tsp00_Namec     szReturn;

  szPwd = getMasterPwd(true);

  szReturn.rawAssign(szPwd.asCharp());

  return szReturn;

} // end cn50DBMUser :: getClearMasterPwd

  /*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: getSecondPwd
  -----------------------------------------------------------------------------
 */
tsp00_CryptPw cn50DBMUser :: getSecondPwd ( ) const
{
  return Data.cryptSecondPwd;
} // end cn50DBMUser :: getSecondPwd

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: getSecondPwd
  -----------------------------------------------------------------------------
 */
tsp00_C64c cn50DBMUser :: getSecondPwd ( bool bUncrypt ) const
{
  tsp00_C64c      szPwd;

  if (bUncrypt) {
    cn90CalculateUncrypt(Data.cryptSecondPwd, szPwd, false);
  } else {
    cn90CryptToCryptString(Data.cryptSecondPwd, szPwd);
  } // end if

  return szPwd;
} // end cn50DBMUser :: getSecondPwd

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: isSecondPwd
  -----------------------------------------------------------------------------
 */
bool cn50DBMUser :: isSecondPwd ( ) const
{
  return Data.bSecondPwd;
} // end cn50DBMUser :: isSecondPwd

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: getSystemName
  -----------------------------------------------------------------------------
 */
tsp00_UserNamec cn50DBMUser :: getSystemName ( ) const
{
  return Data.szSystemName;
} // end cn50DBMUser :: getSystemName

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: getSystemPwd
  -----------------------------------------------------------------------------
 */
tsp00_CryptPw cn50DBMUser :: getSystemPwd ( ) const
{
  return Data.cryptSystemPwd;
} // end cn50DBMUser :: getSystemPwd

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: getSystemPwd
  -----------------------------------------------------------------------------
 */
tsp00_C64c cn50DBMUser :: getSystemPwd ( bool bUncrypt ) const
{
  tsp00_C64c      szPwd;

  if (bUncrypt) {
    cn90CalculateUncrypt(Data.cryptSystemPwd, szPwd, false);
  } else {
    cn90CryptToCryptString(Data.cryptSystemPwd, szPwd);
  } // end if

  return szPwd;
} // end cn50DBMUser :: getSystemPwd

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: isDBMUser
  -----------------------------------------------------------------------------
 */
bool cn50DBMUser :: isDBMUser ( ) const
{
  return Data.bDBMUser;
} // end cn50DBMUser :: isDBMUser

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: isSQLUser
  -----------------------------------------------------------------------------
 */
bool cn50DBMUser :: isSQLUser ( ) const
{
  return Data.bSQLUser;
} // end cn50DBMUser :: isSQLUser

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: isColdUser
  -----------------------------------------------------------------------------
 */
bool cn50DBMUser :: isColdUser ( ) const
{
  return Data.bColdUser;
} // end cn50DBMUser :: isColdUser

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: isSDBAUser
  -----------------------------------------------------------------------------
 */
bool cn50DBMUser :: isSDBAUser ( ) const
{
  return Data.bSDBAUser;
} // end cn50DBMUser :: isSDBAUser

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: getSQLUsermode
  -----------------------------------------------------------------------------
 */
cn50DBMUser::SQLUserMode cn50DBMUser :: getSQLUsermode  ( ) const
{
  return Data.eUserMode;
} // end cn50DBMUser :: getSQLUsermode

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: isDisabled
  -----------------------------------------------------------------------------
 */
bool cn50DBMUser :: isDisabled ( ) const
{
  return Data.bDisabled;
} // end cn50DBMUser :: isDisabled

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: getServerRights
  -----------------------------------------------------------------------------
 */
tsp00_Uint4 cn50DBMUser :: getServerRights ( ) const
{
  return Data.dbmServerRights;
} // end cn50DBMUser :: getServerRights

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: getGUIRights
  -----------------------------------------------------------------------------
 */
tsp00_Uint4 cn50DBMUser :: getGUIRights ( ) const
{
  return Data.dbmGUIRights;
} // end cn50DBMUser :: getGUIRights

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: isServerRight
  -----------------------------------------------------------------------------
 */
bool cn50DBMUser :: isServerRight ( tsp00_Uint4 const nRights, bool const bAnd ) const
{
  bool bReturn = false;

  if (bAnd) {
    bReturn = ((Data.dbmServerRights & nRights) == nRights);
  } else {
    bReturn = ((Data.dbmServerRights & nRights) != DBMMaskNothing_CN50);
  } // end if

  return bReturn || (nRights == DBMMaskNothing_CN50);
} // end cn50DBMUser :: isServerRight

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: isGUIRight
  -----------------------------------------------------------------------------
 */
bool cn50DBMUser :: isGUIRight ( tsp00_Uint4 const nRights, bool const bAnd ) const
{
  bool bReturn = false;

  if (bAnd) {
    bReturn = ((Data.dbmGUIRights & nRights) == nRights);
  } else {
    bReturn = ((Data.dbmGUIRights & nRights) != DBMMaskNothing_CN50);
  } // end if

  return bReturn;
} // end cn50DBMUser :: isGUIRight

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: getComment
  -----------------------------------------------------------------------------
 */
tsp00_C64c cn50DBMUser :: getComment ( ) const
{
  return Data.szComment;
} // end cn50DBMUser :: getComment

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: getData
  -----------------------------------------------------------------------------
 */
void cn50DBMUser :: getData ( _TCHAR * pBuffer ) const
{
  _TCHAR       * pCurrent       = pBuffer;
  int            nIndex         = 0;
  RightToTextT   SvrRightList[] = ServerRightToText;
  RightToTextT   GUIRightList[] = GUIRightToText;
  const _TCHAR * pValue         = NULL;

  if (pCurrent != NULL) {

    // server rights
    _stprintf(pCurrent, _T("%s%c"), KEY_SVRRIGHTS, CHAR_ASSIGN);
    pCurrent = pCurrent + _tcslen(pCurrent);
    for (nIndex = 0; SvrRightList[nIndex].nRight != DBMMaskNothing_CN50; nIndex++) {
      if (isServerRight(SvrRightList[nIndex].nRight) && _tcslen(SvrRightList[nIndex].szName) > 0) {
        _stprintf(pCurrent, _T("%s%c"), SvrRightList[nIndex].szName, 
                                        CHAR_RIGHTDELIMETER);
        pCurrent = pCurrent + _tcslen(pCurrent);
      } // end if
    } // end for
    if (*(pCurrent - 1) == CHAR_RIGHTDELIMETER) {
      --pCurrent;
    } // end if
    *pCurrent = _T('\n');
    ++pCurrent;

    // GUI rights
    _stprintf(pCurrent, _T("%s%c"), KEY_GUIRIGHTS, CHAR_ASSIGN);
    pCurrent = pCurrent + _tcslen(pCurrent);
    for (nIndex = 0; GUIRightList[nIndex].nRight != DBMMaskNothing_CN50; nIndex++) {
      if (isGUIRight(GUIRightList[nIndex].nRight) && _tcslen(GUIRightList[nIndex].szName) > 0) {
        _stprintf(pCurrent, _T("%s%c"), GUIRightList[nIndex].szName, 
                                        CHAR_RIGHTDELIMETER);
        pCurrent = pCurrent + _tcslen(pCurrent);
      } // end if
    } // end for
    if (*(pCurrent - 1) == CHAR_RIGHTDELIMETER) {
      --pCurrent;
    } // end if
    *pCurrent = _T('\n');
    ++pCurrent;

    // exists second pwd
    _stprintf(pCurrent, _T("%s%c%s\n"), KEY_2PWD, 
                                        CHAR_ASSIGN,
                                        isSecondPwd() ? KEY_YES : KEY_NO);
    pCurrent = pCurrent + _tcslen(pCurrent);

    // DBM user
    _stprintf(pCurrent, _T("%s%c%s\n"), KEY_DBMUSR, 
                                        CHAR_ASSIGN,
                                        isDBMUser() ? KEY_YES : KEY_NO);
    pCurrent = pCurrent + _tcslen(pCurrent);

    // SQL user
    _stprintf(pCurrent, _T("%s%c%s\n"), KEY_SQLUSR, 
                                        CHAR_ASSIGN,
                                        isSQLUser() ? KEY_YES : KEY_NO);
    pCurrent = pCurrent + _tcslen(pCurrent);

    // SQL User Mode
    switch (getSQLUsermode()) {
      case SQLUsrModeDBA:
        pValue = KEY_DBA;
        break;        
      case SQLUsrModeRESOURCE:
        pValue = KEY_RESOURCE;
        break;        
      case SQLUsrModeSTANDARD:
        pValue = KEY_STANDARD;
        break;        
      default:
        pValue = _T("");
        break;        
    } // end switch
    _stprintf(pCurrent, _T("%s%c%s\n"), KEY_SQLUSERMODE, 
                                        CHAR_ASSIGN,
                                        pValue);
    pCurrent = pCurrent + _tcslen(pCurrent);

    // Disabled
    _stprintf(pCurrent, _T("%s%c%s\n"), KEY_DISABLED, 
                                        CHAR_ASSIGN,
                                        isDisabled() ? KEY_YES : KEY_NO);
    pCurrent = pCurrent + _tcslen(pCurrent);

    // System Name
    _stprintf(pCurrent, _T("%s%c%s\n"), KEY_SYSNAME, 
                                        CHAR_ASSIGN,
                                        getSystemName().asCharp());
    pCurrent = pCurrent + _tcslen(pCurrent);

    // Comment
    _stprintf(pCurrent, _T("%s%c%s\n"), KEY_COMMENT, 
                                        CHAR_ASSIGN,
                                        getComment().asCharp());
    pCurrent = pCurrent + _tcslen(pCurrent);

  } // end if

} // end cn50DBMUser :: getData

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: getUPCOwner
  -----------------------------------------------------------------------------
 */
tsp00_UserNamec cn50DBMUser :: getUPCOwner (  ) const
{
  return szUPCOwner;
} // end cn50DBMUser :: getUPCOwner

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: listRights
  -----------------------------------------------------------------------------
 */
tcn00_Error cn50DBMUser :: listRights ( _TCHAR * pBuffer, const _TCHAR * szMode ) const
{
  tcn00_Error    nFuncReturn    = OK_CN00;
  _TCHAR       * pCurrent       = pBuffer;
  int            nIndex         = 0;
  RightToTextT   SvrRightList[] = ServerRightToText;
  RightToTextT   GUIRightList[] = GUIRightToText;
  RightToTextT * RightList      = NULL;
  bool           bServer        = true;

  if (pCurrent != NULL) {

    if ( _tcsicmp(szMode, KEY_SVRRIGHTS ) == 0 ) {
      RightList = &SvrRightList[0];
    } else if ( _tcsicmp(szMode, KEY_GUIRIGHTS ) == 0 ) {
      RightList = &GUIRightList[0];
      bServer = false;
    } // end if

    if (RightList != NULL) {
      for (nIndex = 0; RightList[nIndex].nRight != DBMMaskNothing_CN50; nIndex++) {
        if (_tcslen(RightList[nIndex].szName) > 0) {
          _stprintf(pCurrent, _T("%-20s%s%c%s%s%s"),
                                          RightList[nIndex].szName, 
                                          VALUE_SEPSTRING_CN00,
                                          bServer ?  isServerRight(RightList[nIndex].nRight) ? CHAR_RIGHTGRANT : CHAR_RIGHTREMOVE
                                                  :  isGUIRight(RightList[nIndex].nRight)    ? CHAR_RIGHTGRANT : CHAR_RIGHTREMOVE,
                                          VALUE_SEPSTRING_CN00,
                                          RightList[nIndex].szDescription,
                                          LINE_SEPSTRING_CN00);
          pCurrent = pCurrent + _tcslen(pCurrent);
        } // end if
      } // end for
    } else {
      nFuncReturn = ERR_PARAM_CN00;
    } // end if
  } // end if

  return nFuncReturn;
} // end cn50DBMUser :: listRights

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: existsOnDisk
  -----------------------------------------------------------------------------
 */
bool cn50DBMUser :: existsOnDisk ( ) const
{
  return bExistsOnDisk;
} // end cn50DBMUser :: existsOnDisk

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: isDirty
  -----------------------------------------------------------------------------
 */
bool cn50DBMUser :: isDirty ( ) const
{
  return bDirty;
} // end cn50DBMUser :: isDirty

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: isLoggedOn
  -----------------------------------------------------------------------------
 */
bool cn50DBMUser :: isLoggedOn ( ) const
{
  return bLoggedOn;
} // end cn50DBMUser :: isLoggedOn

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: isThis
  -----------------------------------------------------------------------------
 */
bool cn50DBMUser :: isThis ( tsp00_UserNamec const szUserName) const
{
  return (_tcsicmp(szUserName, Data.szUserName) == 0);
} // end cn50DBMUser :: isThis

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: save
  -----------------------------------------------------------------------------
 */
tcn00_Error cn50DBMUser :: save ( bool bRecover )
{
  tcn00_Error   nFuncReturn = OK_CN00;
  
  if ((_tcslen(Data.szDBName) == 0) || (_tcslen(Data.szUserName) == 0)) {
    nFuncReturn = ERR_PARAM_CN00;
  } else {
    if (!bRecover) {
      if (Data.bColdUser) {
        nFuncReturn = saveColdUsr();
      } else if (Data.bSQLUser) {
        nFuncReturn = saveSQLUsr();
      } else if (Data.bExistsInDB) {
        nFuncReturn = deleteSQLUsr();
      } // end if
    } // end if

    if (nFuncReturn == OK_CN00) {
      teo200_EventList        EventList;
      teo28_ProfileContainer  Profile(Data.szDBName, &EventList);

      if (EventList) {
        nFuncReturn = ERR_USRSAVE_CN00;
      } else {
        setSizes(Data);
        Data.recHeader.nRecordId   = UserRecordFormat_Curr_CN50;
        Data.recHeader.nRecordSize = sizeof(Data);
        if (!Profile.eo28_SetProfile(Data.szUserName, sizeof(DBMUserData_1), &Data, &EventList)) {
          nFuncReturn = ERR_USRSAVE_CN00;
        } else {
          bExistsOnDisk   = true;
          bDirty          = false;
          bPwdDirty       = false;
          bSecondPwdDirty = false;
          bSQLDirty       = false;
        } // end if
      } // end if
    } // end if

  } // end if

  return nFuncReturn;

} // end cn50DBMUser :: save

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: reload
  -----------------------------------------------------------------------------
 */
tcn00_Error cn50DBMUser :: reload ( )
{

  tcn00_Error     nFuncReturn = OK_CN00;
  tsp00_Int4      nLength     = sizeof(DBMUserData_1);
  tsp00_Int4      nHeadLength = sizeof(DBMUserData_1);
  DBMUserData_1 * pHeader     = NULL;

  if ((_tcslen(Data.szDBName) == 0) || (_tcslen(Data.szUserName) == 0)) {
    nFuncReturn = ERR_PARAM_CN00;
  } else {
    teo200_EventList        EventList;
    teo28_ProfileContainer  Profile(Data.szDBName, &EventList);

    if (EventList) {
      nFuncReturn = ERR_USRREAD_CN00;
    } else {

      // read profile into header structure with dynamic buffer
      pHeader = REINTERPRET_CAST(DBMUserData_1 *, new char[nHeadLength]); 
      if (pHeader != NULL) {
        if (!Profile.eo28_GetProfile(Data.szUserName, nHeadLength, pHeader, &EventList)) {
          if (nHeadLength > sizeof(DBMUserData_1)) {
            delete pHeader;
            pHeader = REINTERPRET_CAST(DBMUserData_1 *, new char[nHeadLength]); 
            if (pHeader != NULL) {
              if (!Profile.eo28_GetProfile(Data.szUserName, nHeadLength, pHeader, &EventList)) {
                nFuncReturn = ERR_USRREAD_CN00;
              } // end if
            } else {
              nFuncReturn = ERR_MEM_CN00;
            } // end if
          } else {
            nFuncReturn = ERR_USRREAD_CN00;
          } // end if
        } // end if

        // check header data
        if (nFuncReturn == OK_CN00) {
          if (pHeader->recHeader.nRecordId   != UserRecordFormat_Curr_CN50 ||
              pHeader->recHeader.nRecordSize != sizeof(Data)                  ) {
            // shit happens! check format!
            nFuncReturn = checkRecord(Data.szUserName, &Profile, pHeader);
          } // end if
        } // end if
        delete pHeader;

      } else {
        nFuncReturn = ERR_MEM_CN00;
      } // end if

      // read really
      if (nFuncReturn == OK_CN00) {
        if (!Profile.eo28_GetProfile(Data.szUserName, nLength, &Data, &EventList)) {
          nFuncReturn = ERR_USRREAD_CN00;
        } else {
          if (!Profile.eo28_GetOwnerName(szUPCOwner, &EventList)) {
            szUPCOwner.Init();
          } // end if
          bExistsOnDisk   = true;
          bDirty          = false;
          bPwdDirty       = false;
          bSecondPwdDirty = false;
          bSQLDirty       = false;        
        } // end if
      } // end if

    } // end if
  } // end if

  return nFuncReturn;

} // end cn50DBMUser :: reload

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: deleteUsr
  -----------------------------------------------------------------------------
 */
tcn00_Error cn50DBMUser :: deleteUsr ( )
{
  tcn00_Error   nFuncReturn = OK_CN00;
  
  if (_tcslen(Data.szDBName) == 0 || _tcslen(Data.szUserName) == 0) {
    nFuncReturn = ERR_PARAM_CN00;
  } else {

    if (Data.bExistsInDB) {
      nFuncReturn = deleteSQLUsr();
    } // end if

    if (nFuncReturn == OK_CN00) {
      teo200_EventList        EventList;
      teo28_ProfileContainer  Profile(Data.szDBName, &EventList);

      if (EventList) {
        nFuncReturn = ERR_USRSAVE_CN00;
      } else {
        if (!Profile.eo28_DelProfile(Data.szUserName, &EventList)) {
          nFuncReturn = ERR_USRSAVE_CN00;
        } // end if
      } // end if


    } // end if

  } // end if

  return nFuncReturn;

} // end cn50DBMUser :: deleteUsr

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: clear
  -----------------------------------------------------------------------------
 */
void cn50DBMUser :: clear ( )
{
  setDBName       ( (*(tsp00_DbNamec    *) _T("")) );
  setUserName     ( (*(tsp00_UserNamec  *) _T("")) );
  setMasterPwd    ( (*(tsp00_Namec      *) _T("")) );
  setSecondPwd    ( (*(tsp00_Namec      *) _T("")) );
  setSystemName   ( (*(tsp00_UserNamec  *) _T("")) );
  setSystemPwd    ( (*(tsp00_Namec      *) _T("")) );
  setComment      ( (*(tsp00_C64c       *) _T("")) );

  setDBMUser      ( false );
  setSQLUser      ( false );
  setColdUser     ( false );
  setSDBAUser     ( false );

  setDisabled     ( false );

  setServerRights ( DBMMaskAll_CN50, DBMMaskNothing_CN50 );
  setGUIRights    ( DBMMaskAll_CN50, DBMMaskNothing_CN50 );

  Data.bExistsInDB = false;

  szUPCOwner.Init();

  bExistsOnDisk   = false;
  bDirty          = false;
  bPwdDirty       = false;
  bSecondPwdDirty = false;
  bSQLDirty       = false;
  bLoggedOn       = false;

} // end cn50DBMUser :: clear

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: setData
  -----------------------------------------------------------------------------
 */
tcn00_Error  cn50DBMUser :: setData ( const _TCHAR * szData )
{
  _TCHAR        szToken[1024];
  _TCHAR      * pValue   = NULL;
  long          nToken   = 1;
  tcn00_Error   nReturn  = OK_CN00;
  
  if (szData != NULL) {

    // analyze tokens
    while (cn90GetToken(szData, szToken, nToken)) {
      pValue = _tcschr(szToken, CHAR_ASSIGN);

      if (pValue != NULL) {
        *pValue = CHAR_STRINGTERM_CN90;
        pValue++;

        if        ( _tcsicmp(szToken, KEY_SVRRIGHTS ) == 0 ) {
          RightToTextT  RightList[] = ServerRightToText;
          setRights(pValue, Data.dbmServerRights, RightList);
        } else if ( _tcsicmp(szToken, KEY_GUIRIGHTS ) == 0 ) {
          RightToTextT  RightList[] = GUIRightToText;
          setRights(pValue, Data.dbmGUIRights, RightList);
        } else if ( _tcsicmp(szToken, KEY_PWD       ) == 0 ) {
          cn90Uncrypt(pValue, false);
          setMasterPwd(*(tsp00_Namec *) pValue);
        } else if ( _tcsicmp(szToken, KEY_2PWD      ) == 0 ) {
          cn90Uncrypt(pValue, false);
          setSecondPwd(*(tsp00_Namec *) pValue);
        } else if ( _tcsicmp(szToken, KEY_SYSNAME   ) == 0 ) { 
          setSystemName(*(tsp00_UserNamec *) pValue);
        } else if ( _tcsicmp(szToken, KEY_SYSPWD    ) == 0 ) { 
          cn90Uncrypt(pValue, false);
          setSystemPwd(*(tsp00_Namec *) pValue);
        } else if ( _tcsicmp(szToken, KEY_DBMUSR    ) == 0 ) { 
          if        ( _tcsicmp(pValue, KEY_YES ) == 0 ) {
            setDBMUser();
          } else if ( _tcsicmp(pValue, KEY_NO  ) == 0 ) {
            setDBMUser(false);
          } // end if    
        } else if ( _tcsicmp(szToken, KEY_SQLUSR    ) == 0 ) { 
          if        ( _tcsicmp(pValue, KEY_YES ) == 0 ) {
            setSQLUser();
          } else if ( _tcsicmp(pValue, KEY_NO  ) == 0 ) {
            setSQLUser(false);
          } // end if    
        } else if ( _tcsicmp(szToken, KEY_SQLUSERMODE  ) == 0 ) { 
          if        ( _tcsicmp(pValue, KEY_DBA ) == 0 ) {
            setSQLUsermode(SQLUsrModeDBA);
          } else if ( _tcsicmp(pValue, KEY_RESOURCE  ) == 0 ) {
            setSQLUsermode(SQLUsrModeRESOURCE);
          } else if ( _tcsicmp(pValue, KEY_STANDARD  ) == 0 ) {
            setSQLUsermode(SQLUsrModeSTANDARD);
          } // end if    
        } else if ( _tcsicmp(szToken, KEY_DISABLED  ) == 0 ) { 
          if        ( _tcsicmp(pValue, KEY_YES ) == 0 ) {
            setDisabled();
          } else if ( _tcsicmp(pValue, KEY_NO  ) == 0 ) {
            setDisabled(false);
          } // end if    
        } else if ( _tcsicmp(szToken, KEY_COMMENT   ) == 0 ) { 
          cn90StripQuotes(pValue);
          setComment(*(tsp00_C64c *) pValue);
        } else {
          nReturn = ERR_PARAM_CN00;
        } // end if    

      } // end if

      nToken++;
    } // end while

  } // end if

  return nReturn;
} // end cn50DBMUser :: setData

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: setRights
  -----------------------------------------------------------------------------
 */
cn50DBMUser &  cn50DBMUser :: setRights(_TCHAR * szValue, tsp00_Uint4 & nValue, RightToTextT RightList[])
{
  _TCHAR   * pRight     = szValue;
  _TCHAR   * pGrant     = NULL;
  _TCHAR   * pNextRight = NULL;
  long       nIndex     = 0;

  while (pRight) {

    // Terminate right entry
    pNextRight = _tcschr(pRight, CHAR_RIGHTDELIMETER);
    if (pNextRight != NULL) {
      *pNextRight = CHAR_STRINGTERM_CN90;
      pNextRight++;
    } // end if

    // look for grant sign
    pGrant = pRight;
    pRight++;
    
    // look for right value
    for (nIndex = 0; RightList[nIndex].nRight != DBMMaskNothing_CN50; nIndex++) {
      if (_tcsicmp(pRight, RightList[nIndex].szName) == 0) {
        break;
      } // end if
    } // end for

    // set/unset right
    switch (*pGrant) {
      case CHAR_RIGHTGRANT:
        nValue = nValue | RightList[nIndex].nRight;        
        break;
      case CHAR_RIGHTREMOVE:
        nValue = nValue & ~RightList[nIndex].nRight;        
        break;
      default:
        break;
    } // end switch

    // an the next
    pRight = pNextRight;
  } // end while

  return *this;
} // end cn50DBMUser :: setRights

/*
  -----------------------------------------------------------------------------
  protected function cn50DBMUser :: saveColdUsr
  -----------------------------------------------------------------------------
 */
tcn00_Error cn50DBMUser :: saveColdUsr()
{
  tcn00_Error         nFuncReturn = OK_CN00;
  tsp00_DbNamec       dbName      = getDBName();
  tsp00_XpValueRec    XpValue;

  #if defined (CN71)
    tsp00_XpReturnCode  nXPReturn;
    tsp_errtext         szError;
    tsp00_Int4          hParamFile;
    bool                bParamOK  = false;
    tsp00_DbName        dbNameP;
    tsp00_XpKeyType     xpKey;

    /* open/read/close param file */
    sqlxpopenxparam ( dbNameP.c2p(dbName), 0, &hParamFile , szError , UCHAR_P_CAST_CN00 &bParamOK );
    if (bParamOK) {
      XpValue.xp_value_t = xp_crypt_type;
      cn29StringToParamValue(&XpValue, getMasterPwd(true));

      sqlputxparam ( hParamFile, xpKey.c2p(XP_ID_CONTROLPASSWORD_CN00), &XpValue, szError, &nXPReturn );
      if (nXPReturn == xp_ok) {
        sqlxpclosexparam ( hParamFile , true, szError , UCHAR_P_CAST_CN00 &bParamOK );
      } else {
        sqlxpclosexparam ( hParamFile , false, szError , UCHAR_P_CAST_CN00 &bParamOK );
        nFuncReturn = ERR_RTE_CN00;
      } // end if
    } // end if
  #else
    tsp11_XpReturnCode  XpReturn;
    tsp01_RteError      RteError;

    if (bPwdDirty) {
      if (cn90DBState(dbName) != STATE_OFFLINE_CN00) {
        nFuncReturn = ERR_DBRUN_CN00;
      } else {

        XpValue.xp_value_t.becomes(xp_crypt_type);
        cn29StringToParamValue(&XpValue, getMasterPwd(true));

        dbName = getDBName();
        XpReturn = eo11PutParam ( dbName,
                                XP_ID_CONTROLPASSWORD_CN00,
                                &XpValue,
                                &RteError );

        if (XpReturn != xp_ok_sp11) {
          nFuncReturn = ERR_RTE_CN00;
        } // end if
      } // end if

    } // end if
  #endif

  return nFuncReturn; 
} // end cn50DBMUser :: saveColdUsr

/*
  -----------------------------------------------------------------------------
  protected function cn50DBMUser :: saveSQLUsr
  -----------------------------------------------------------------------------
 */
tcn00_Error cn50DBMUser :: saveSQLUsr()
{
  tcn00_Error         nFuncReturn = OK_CN00;
  _TCHAR              szCommand[200];
  cn51DBMConfigValue  cfgDBAUsr ( Data.szDBName, CFG_DBAUSR_CN51);
  cn50DBMUser         usrDBA    ( Data.szDBName, cfgDBAUsr);

  if (!existsInDB()) {
    if (!isSDBAUser()) {
      // Create user
      if (cn90DBState(Data.szDBName) != STATE_WARM_CN00) {
        nFuncReturn = ERR_DBNORUN_CN00;
      } else {
        _stprintf(szCommand, _T("CREATE USER %s PASSWORD %s"), (_TCHAR *) getUserName(), (_TCHAR *) getMasterPwd(true) );
        nFuncReturn = cn80ExecuteSQL(usrDBA, szCommand);
        if (nFuncReturn == OK_CN00) {
          Data.bExistsInDB = true;
        } else {
          // possible user exist?
          _stprintf(szCommand, _T("ALTER PASSWORD %s %s"), (_TCHAR *) getUserName(), (_TCHAR *) getMasterPwd(true) );
          nFuncReturn = cn80ExecuteSQL(usrDBA, szCommand);
          if (nFuncReturn == OK_CN00) {
            Data.bExistsInDB = true;
          } // end if
        } // end if
      } // end if
    } else {
      setExistsInDB();
    } // end if
  } else if (bPwdDirty) {
    // only change password
    if (cn90DBState(Data.szDBName) != STATE_WARM_CN00) {
      nFuncReturn = ERR_DBNORUN_CN00;
    } else {
        _stprintf(szCommand, _T("ALTER PASSWORD %s %s"), (_TCHAR *) getUserName(), (_TCHAR *) getMasterPwd(true) );
        nFuncReturn = cn80ExecuteSQL(usrDBA, szCommand);
    } // end if
  } // end if

  if (bSQLDirty && nFuncReturn == OK_CN00) {
    const _TCHAR            * pUserMode = NULL;

    if (cn90DBState(Data.szDBName) != STATE_WARM_CN00) {
      nFuncReturn = ERR_DBNORUN_CN00;
    } else {
      switch (getSQLUsermode()) {
        case SQLUsrModeDBA:
          pUserMode = KEY_DBA;
          break;        
        case SQLUsrModeRESOURCE:
          pUserMode = KEY_RESOURCE;
          break;        
        case SQLUsrModeSTANDARD:
          pUserMode = KEY_STANDARD;
          break;        
        default:
          pUserMode = _T("");
          break;        
      } // end switch
      _stprintf(szCommand, _T("ALTER USER %s %s"), (_TCHAR *) getUserName(), pUserMode );
      nFuncReturn = cn80ExecuteSQL(usrDBA, szCommand);
    } // end if
  } // end if

  if (bSecondPwdDirty && nFuncReturn == OK_CN00) {
    // change second password
    if (cn90DBState(Data.szDBName) != STATE_WARM_CN00) {
      nFuncReturn = ERR_DBNORUN_CN00;
    } else {
      _stprintf(szCommand, _T("ALTER USER %s %s PASSWORD %s"), 
                          (_TCHAR *) getUserName(), 
                          isSecondPwd() ?  _T("ADD") : _T("DROP"),
                          isSecondPwd() ? (_TCHAR *) getSecondPwd(true) : _T(""));

      nFuncReturn = cn80ExecuteSQL(usrDBA, szCommand);
    } // end if
  } // end if

  return nFuncReturn; 
} // end cn50DBMUser :: saveSQLUsr

/*
  -----------------------------------------------------------------------------
  protected function cn50DBMUser :: deleteSQLUsr
  -----------------------------------------------------------------------------
 */
tcn00_Error cn50DBMUser :: deleteSQLUsr()
{
  tcn00_Error         nFuncReturn = OK_CN00;
  _TCHAR              szCommand[200];
  cn51DBMConfigValue  cfgDBAUsr ( Data.szDBName, CFG_DBAUSR_CN51);
  cn50DBMUser         usrDBA    ( Data.szDBName, cfgDBAUsr);

  if (!isColdUser() && !isSDBAUser()) {

    if (cn90DBState(Data.szDBName) != STATE_WARM_CN00) {
      nFuncReturn = ERR_DBNORUN_CN00;
    } else {
      _stprintf(szCommand, _T("DROP USER %s"), (_TCHAR *) getUserName(), (_TCHAR *) getMasterPwd(true) );
      nFuncReturn = cn80ExecuteSQL(usrDBA, szCommand);
      if (nFuncReturn == OK_CN00) {
        setExistsInDB(false);
      } // end if
    } // end if
         
  } // end if

  return nFuncReturn; 
} // end cn50DBMUser :: deleteSQLUsr

/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: setExistsInDB
  -----------------------------------------------------------------------------
 */
cn50DBMUser & cn50DBMUser :: setExistsInDB ( bool bValue )
{
  Data.bExistsInDB  = bValue;

  return *this;
} // end cn50DBMUser :: setExistsInDB
/*
  -----------------------------------------------------------------------------
  public function cn50DBMUser :: existsInDB
  -----------------------------------------------------------------------------
 */
bool cn50DBMUser :: existsInDB ( ) const
{
  return Data.bExistsInDB;
} // end cn50DBMUser :: existsInDB

/*
  -----------------------------------------------------------------------------
  pivate function cn50DBMUser :: setSizes
  -----------------------------------------------------------------------------
 */
void cn50DBMUser :: setSizes ( DBMUserData_1 & locData ) const
{
  locData.s_szDBName        = sizeof(locData.szDBName);
  locData.s_szUserName      = sizeof(locData.szUserName);
  locData.s_cryptMasterPwd  = sizeof(locData.cryptMasterPwd);
  locData.s_cryptSecondPwd  = sizeof(locData.cryptSecondPwd);
  locData.s_szSystemName    = sizeof(locData.szSystemName);
  locData.s_cryptSystemPwd  = sizeof(locData.cryptSystemPwd);
  locData.s_bSecondPwd      = sizeof(locData.bSecondPwd);
  locData.s_bDBMUser        = sizeof(locData.bDBMUser);
  locData.s_bSQLUser        = sizeof(locData.bSQLUser);
  locData.s_bExistsInDB     = sizeof(locData.bExistsInDB);
  locData.s_bColdUser       = sizeof(locData.bColdUser);
  locData.s_bSDBAUser       = sizeof(locData.bSDBAUser);
  locData.s_eUserMode       = sizeof(locData.eUserMode);
  locData.s_bDisabled       = sizeof(locData.bDisabled);
  locData.s_dbmServerRights = sizeof(locData.dbmServerRights);
  locData.s_dbmGUIRights    = sizeof(locData.dbmGUIRights);
  locData.s_szComment       = sizeof(locData.szComment);
} // end cn50DBMUser :: setSizes

/*
  -----------------------------------------------------------------------------
  pivate function cn50DBMUser :: checkRecord
  -----------------------------------------------------------------------------
 */
tcn00_Error cn50DBMUser :: checkRecord ( const tsp00_UserNamec    szUserName,
                                         teo28_ProfileContainer * pProfile,
                                         const DBMUserData_1    * pHeader ) const
{
  tcn00_Error         nFuncReturn   = OK_CN00;
  tsp00_Uint2         nSourceRecord = pHeader->recHeader.nRecordId;
  tsp00_Uint2         nTargetRecord = 0;

  // look for version 0 records
  nSourceRecord = (nSourceRecord  > UserRecordFormat_Curr_CN50) ? 0 : nSourceRecord;

  if (nSourceRecord < UserRecordFormat_Curr_CN50) {
    nTargetRecord = nSourceRecord + 1;
    do {
      switch (nTargetRecord) {
        case UserRecordFormat_1_CN50:
          nFuncReturn = convFrom0To1(szUserName, pProfile);
          break;
        default:
          // do nothing, version doesn't  exist
          break;
      } // end switch
    } while (nTargetRecord <  UserRecordFormat_Curr_CN50 && 
           nFuncReturn   == OK_CN00);

  } else if (pHeader->recHeader.nRecordSize != sizeof(DBMUserData_1)) {
    // only different sizes
    nFuncReturn = convFromXToX(szUserName, pProfile);
  } // end if

  return nFuncReturn;
} // end cn50DBMUser :: checkRecord

/*
  -----------------------------------------------------------------------------
  pivate function cn50DBMUser :: convFrom0To1
  -----------------------------------------------------------------------------
 */
tcn00_Error cn50DBMUser :: convFrom0To1 ( const tsp00_UserNamec    szUserName,
                                          teo28_ProfileContainer * Profile ) const
{
  tcn00_Error         nFuncReturn   = OK_CN00;
  teo200_EventList    EventList;
  DBMUserData_0       Data_0;
  tsp00_Int4          nLength_0 = sizeof(DBMUserData_0);
  DBMUserData_1       Data_1;
  
  // read old version
  if (!Profile->eo28_GetProfile(szUserName, nLength_0, &Data_0, &EventList)) {
    nFuncReturn = ERR_USRFRM_CN00;
  } else {
    if (nLength_0 != sizeof(DBMUserData_0)) { 
      nFuncReturn = ERR_USRFRM_CN00;
    } else {

      Data_1.szDBName        = Data_0. szDBName;
      Data_1.szUserName      = Data_0.szUserName;
      Data_1.cryptMasterPwd  = Data_0.cryptMasterPwd;
      Data_1.cryptSecondPwd  = Data_0.cryptSecondPwd;
      Data_1.szSystemName    = Data_0.szSystemName;
      Data_1.cryptSystemPwd  = Data_0.cryptSystemPwd;
      Data_1.bSecondPwd      = Data_0.bSecondPwd;
      Data_1.bDBMUser        = Data_0.bDBMUser;
      Data_1.bSQLUser        = Data_0.bSQLUser;
      Data_1.bExistsInDB     = Data_0.bExistsInDB;
      Data_1.bColdUser       = Data_0.bColdUser;
      Data_1.bSDBAUser       = Data_0.bSDBAUser;
      Data_1.eUserMode       = Data_0.eUserMode;
      Data_1.bDisabled       = Data_0.bDisabled;
      Data_1.dbmServerRights = Data_0.dbmServerRights;
      Data_1.dbmGUIRights    = Data_0.dbmGUIRights;
      Data_1.szComment       = Data_0.szComment;

      setSizes(Data_1);
      Data_1.recHeader.nRecordId   = UserRecordFormat_Curr_CN50;
      Data_1.recHeader.nRecordSize = sizeof(DBMUserData_1);

      if (!Profile->eo28_DelProfile(Data_1.szUserName, &EventList)) {
        nFuncReturn = ERR_USRFRM_CN00;
      } else {
        if (!Profile->eo28_SetProfile(Data_1.szUserName, sizeof(DBMUserData_1), &Data_1, &EventList)) {
          nFuncReturn = ERR_USRFRM_CN00;
        } // end if
      } // end if
    } // end if
  } // end if
   
  return nFuncReturn;
} // end cn50DBMUser :: convFrom0To1

/*
  -----------------------------------------------------------------------------
  pivate function cn50DBMUser :: convFromXToX
  -----------------------------------------------------------------------------
 */
tcn00_Error cn50DBMUser :: convFromXToX ( const tsp00_UserNamec    szUserName,
                                          teo28_ProfileContainer * Profile ) const
{
  return ERR_USRFRM_CN00;
} // end cn50DBMUser :: convFromXToX
