/*


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end


*/
#if defined(CN72) && defined(_WIN32)
  #line 3 "d:\\v72\\sys\\src\\cn\\vcn84.cpp"
#endif 

/*! 
  -----------------------------------------------------------------------------
 
  module: vcn83.cpp
 
  -----------------------------------------------------------------------------
 
  responsible:  MartinR
 
  special area: DBMServer
 
  description:  implementation module for kernel event access

  version:      7.2.

  -----------------------------------------------------------------------------
 
                          Copyright by SAP AG, 2000
 
  -----------------------------------------------------------------------------
*/

/* 
  -----------------------------------------------------------------------------
  includes
  -----------------------------------------------------------------------------
*/

#include <stdio.h>

#include "hcn40.h"
#include "hcn90.h"

#include "hcn84.h"

/* 
  -----------------------------------------------------------------------------
  specification private types and macros
  -----------------------------------------------------------------------------
*/
#define TRACE_PROTOPT_TITLE _T("name,option\n")

#define TRACE_PROTOPT \
  {{_T("a"), _T("Order Interface (AK)")},  \
   {_T("b"), _T("Record Interface (BD)")},  \
   {_T("k"), _T("Show Message Block (KB)")},  \
   {_T("m"), _T("Message Block")},  \
   {_T("e"), _T("No Entrypos Output")},  \
   {_T("s"), _T("Strategy")},  \
   {_T("t"), _T("Time")},  \
   {_T("x"), _T("Switch Output (Slow Kernel)")},  \
   {NULL,    NULL}}

typedef struct tcn84TraceProtOpt {
    const _TCHAR * szOption;
    const _TCHAR * szName;
} tcn84TraceProtOpt;

/* 
  -----------------------------------------------------------------------------
  specification private functions
  -----------------------------------------------------------------------------
*/

/* 
  -----------------------------------------------------------------------------
  function:     cn84TraceShow
  -----------------------------------------------------------------------------
 */
tcn00_Error cn84TraceShow
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error       nFuncReturn  = OK_CN00;

  cn90AnswerOK(replyData, replyLen, NULL);

  return nFuncReturn;
} // end cn84TraceShow

/* 
  -----------------------------------------------------------------------------
  function:     cn84TraceOn
  -----------------------------------------------------------------------------
 */
tcn00_Error cn84TraceOn
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error       nFuncReturn  = OK_CN00;

  cn90AnswerOK(replyData, replyLen, NULL);

  return nFuncReturn;
} // end cn84TraceOn

/* 
  -----------------------------------------------------------------------------
  function:     cn84TraceOff
  -----------------------------------------------------------------------------
 */
tcn00_Error cn84TraceOff
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error       nFuncReturn  = OK_CN00;

  cn90AnswerOK(replyData, replyLen, NULL);

  return nFuncReturn;
} // end cn84TraceOff

/* 
  -----------------------------------------------------------------------------
  function:     cn84TraceFlush
  -----------------------------------------------------------------------------
 */
tcn00_Error cn84TraceFlush
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error       nFuncReturn  = OK_CN00;

  cn90AnswerOK(replyData, replyLen, NULL);

  return nFuncReturn;
} // end cn84TraceFlush

/* 
  -----------------------------------------------------------------------------
  function:     cn84TraceClear
  -----------------------------------------------------------------------------
 */
tcn00_Error cn84TraceClear
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error       nFuncReturn  = OK_CN00;

  cn90AnswerOK(replyData, replyLen, NULL);

  return nFuncReturn;
} // end cn84TraceClear

/* 
  -----------------------------------------------------------------------------
  function:     cn84TraceProtOpt
  -----------------------------------------------------------------------------
 */
tcn00_Error cn84TraceProtOpt
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  long                nIndex          = 0;
  tcn84TraceProtOpt   sTraceProtOpt[] = TRACE_PROTOPT;
  _TCHAR            * pCurrent        = replyData;

  // init OK Answer
  _stprintf (pCurrent, "%s%s%s", ANSWER_OK_CN00, LINE_SEPSTRING_CN00, (_TCHAR *) TRACE_PROTOPT_TITLE);
  pCurrent = pCurrent + _tcslen(pCurrent);

  while (sTraceProtOpt[nIndex].szOption != NULL) {
    _stprintf (pCurrent, "%s%s%s%s", 
                         (_TCHAR *) sTraceProtOpt[nIndex].szOption,
                         VALUE_SEPSTRING_CN00,
                         (_TCHAR *) sTraceProtOpt[nIndex].szName,
                         LINE_SEPSTRING_CN00);
    pCurrent = pCurrent + _tcslen(pCurrent);

    nIndex++;
  }

  *replyLen = (int)_tcslen(replyData);

  return OK_CN00;
} // end cn84TraceProtOpt


/* 
  -----------------------------------------------------------------------------
  function:     cn84TraceProt
  -----------------------------------------------------------------------------
 */
tcn00_Error cn84TraceProt
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  return cn40ExecXKernprot(vcontrol, command, replyData, replyLen, replyLenMax);
} // end cn84TraceProt
