/*! ******************************************************************************

  module: vcn90.cpp

  ------------------------------------------------------------------------------

  responsible:  BerndV

  special area: Control-Server

  description:  implementation of useful utility functions
                for the cn (Control) layer

  version:      7.2. / 6.2

  ------------------------------------------------------------------------------

                          Copyright by SAP AG, 1998



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

  *******************************************************************************/

/* ------------------------------------------------------------------
 * IMPORT
 * ------------------------------------------------------------------
 */
#include <stdlib.h>
#include <ctype.h>
#include <stdio.h>
#include <time.h>

//extern "C" int chdir(const char * szDir);

#include "vsp001.h"
#include "vsp009c.h"
#include "heo06.h"
#include "heo02.h"
#include "hsp02.h"
#include "heo01.h"
#include "hcn90.h"

#if !defined(_WIN32)
  #include "gen00.h"/* no check */
#endif /* !WIN32 */

/* ------------------------------------------------------------------
 * SPECIFICATION PRIVATE MACROS
 * ------------------------------------------------------------------
 */

/* directories */
#define DB_UX_CFG_DIR   _T("/usr/spool/sql/config")
#define DB_CFG_DIR      _T("config")
#define DB_WRK_DIR      _T("wrk")
#define DB_ENV_DIR      _T("env")
#define DB_BIN_DIR      _T("bin")
#define ENV_DBROOT      _T("DBROOT")

/* so sorry, errno values aren't present on every system in the same matter, so
   that we include the usage of errno values in some ifdef's                   */
#ifdef E2BIG
  #define E2BIG_CN90     { E2BIG     , _T("E2BIG : argument list to long")             },
#else
  #define E2BIG_CN90
#endif

#ifdef EACCES
  #define EACCES_CN90    { EACCES    , _T("EACCES : permission denied")                },
#else
  #define EACCES_CN90
#endif

#ifdef EAGAIN
  #define EAGAIN_CN90    { EAGAIN    , _T("EAGAIN : No more processes or not enough memory or maximum nesting level reached") },
#else
  #define EAGAIN_CN90
#endif

#ifdef EBADF
  #define EBADF_CN90     { EBADF     , _T("EBADF : bad file number")                   },
#else
  #define EBADF_CN90
#endif

#ifdef ECHILD
  #define ECHILD_CN90    { ECHILD    , _T("ECHILD : no spawnde processes")             },
#else
  #define ECHILD_CN90
#endif

#ifdef EDEADLOCK
  #define EDEADLOCK_CN90 { EDEADLOCK , _T("EDEADLOCK : resource deadlock would error") },
#else
  #define EDEADLOCK_CN90
#endif

#ifdef EDOM
  #define EDOM_CN90      { EDOM      , _T("EDOM : math argument")                      },
#else
  #define EDOM_CN90
#endif

#ifdef EEXIST
  #define EEXIST_CN90    { EEXIST    , _T("EEXIST : file exist")                       },
#else
  #define EEXIST_CN90
#endif

#ifdef EINVAL
  #define EINVAL_CN90    { EINVAL    , _T("EIVAL : invalid argument")                  },
#else
  #define EINVAL_CN90
#endif

#ifdef EMFILE
  #define EMFILE_CN90    { EMFILE    , _T("EMFILE : too many open files")              },
#else
  #define EMFILE_CN90
#endif

#ifdef ENOENT
  #define ENOENT_CN90    { ENOENT    , _T("ENOENT : no such file or directory")        },
#else
  #define ENOENT_CN90
#endif

#ifdef ENOEXEC
  #define ENOEXEC_CN90   { ENOEXEC   , _T("ENOEXEC : exec format error")               },
#else
  #define ENOEXEC_CN90
#endif

#ifdef ENOMEM
  #define ENOMEM_CN90    { ENOMEM    , _T("ENOMEM : not enough memory")                },
#else
  #define ENOMEM_CN90
#endif

#ifdef ENOSPC
  #define ENOSPC_CN90    { ENOSPC    , _T("ENOSPC : no space left on device")          },
#else
  #define ENOSPC_CN90
#endif

#ifdef ERANGE
  #define ERANGE_CN90    { ERANGE    , _T("ERANGE : result too large")                 },
#else
  #define ERANGE_CN90
#endif

#ifdef EXDEV
  #define EXDEV_CN90     { EXDEV     , _T("EXDEV : cross-device link")                 },
#else
  #define EXDEV_CN90
#endif

#define ERRNO_TO_TEXT { \
   E2BIG_CN90           \
   EACCES_CN90          \
   EAGAIN_CN90          \
   EBADF_CN90           \
   ECHILD_CN90          \
   EDEADLOCK_CN90       \
   EDOM_CN90            \
   EEXIST_CN90          \
   EINVAL_CN90          \
   EMFILE_CN90          \
   ENOENT_CN90          \
   ENOEXEC_CN90         \
   ENOMEM_CN90          \
   ENOSPC_CN90          \
   ERANGE_CN90          \
   EXDEV_CN90           \
  { 0 , NULL       },   \
  { 0 , _T("unknown errno") }}

/* create dir */
#define CHAR_SLASH        _T('/')
#define CHAR_BACKSLASH    _T('\\')

/*
 * ==================================================================
 * START OF IMPLEMENTATION PUBLIC
 * ==================================================================
 */

/* PUBLIC FUNCTION cn90StrUpperCopy
 * ------------------------------------------------------------------
 */
_TCHAR * cn90StrUpperCopy
    ( _TCHAR             * szDest,
      const _TCHAR       * szSource,
      const bool           bKeepCase )
{
  long i = 0;

  for (i = 0; szSource[i] != CHAR_STRINGTERM_CN90; i++) {
    szDest[i] = bKeepCase ? szSource[i] : _totupper(szSource[i]);
  } /* end for */

  szDest[i] = CHAR_STRINGTERM_CN90;

  return szDest;
} /* end cn90StrUpperCopy */

/* PUBLIC FUNCTION cn90StringPtoC
 * ------------------------------------------------------------------
 */
_TCHAR * cn90StringPtoC
    ( _TCHAR             * szDestC,
      const _TCHAR       * pSourceP,
      const int            nLength )
{
  int     i;
  bool bFound = false;

  // at first looking for zero termination in source
  for (i = 0; (i < (nLength - 1)) && (pSourceP[i] != CHAR_STRINGTERM_CN90); i++);

  if (pSourceP[i] == CHAR_STRINGTERM_CN90) {
    _tcscpy(szDestC, pSourceP);
  } else {
    szDestC[nLength] = CHAR_STRINGTERM_CN90;
    for (i = nLength - 1; i >= 0; i--) {
      if ((pSourceP[i] == CHAR_PASCALFILL_CN90) && !bFound) {
        szDestC[i] = CHAR_STRINGTERM_CN90;
      } else {
        bFound = true;
        szDestC[i] = pSourceP[i];
      } /* end if */
    } /* end while */
  } /* end while */

  return szDestC;
} /* end cn90StringPtoC */

/* PUBLIC FUNCTION cn90StringCtoP
 * ------------------------------------------------------------------
 */
_TCHAR * cn90StringCtoP
    ( _TCHAR             * pDestP,
      const _TCHAR       * szSourceC,
      const int            nLength )
{
  int i;

  memset(pDestP, CHAR_PASCALFILL_CN90, nLength);
  for (i = 0; ((szSourceC[i] != CHAR_STRINGTERM_CN90) && (i < nLength)); i++) {
    pDestP[i] = szSourceC[i];
  } /* end while */

  return pDestP;
} /* end cn90StringCtoP */

/* PUBLIC FUNCTION cn90StripLeadingChar
 * ------------------------------------------------------------------
 */
_TCHAR * cn90StripLeadingChar
    ( _TCHAR     * szString,
      const _TCHAR cStrip )
{
  char * pSource      = szString;
  char * pDestination = szString;

  while ((*pSource == cStrip) && (*pSource != CHAR_STRINGTERM_CN90)) {
    pSource++;
  } /* end while */
  while (*pSource != CHAR_STRINGTERM_CN90) {
    *pDestination = *pSource;
    pDestination++;
    pSource++;
  } /* end while */

  *pDestination = CHAR_STRINGTERM_CN90;

  return szString;
} /* end cn90StripLeadingChar */

/* PUBLIC FUNCTION cn90StripTrailingChar
 * ------------------------------------------------------------------
 */
_TCHAR * cn90StripTrailingChar
    ( _TCHAR       * szString,
      const _TCHAR   cStrip )
{
  char * pSource      = &szString[strlen(szString) - 1];
  char * pDestination = &szString[strlen(szString) - 1];

  while ((*pSource == cStrip) && (pSource != szString)) {
    *pDestination = CHAR_STRINGTERM_CN90;
    pSource--;
    pDestination--;
  } /* end while */

  return szString;
} /* end cn90StripTrailingChar */

/* PUBLIC FUNCTION cn90Strip
 * ------------------------------------------------------------------
 */
_TCHAR * cn90Strip
    ( _TCHAR * szString )
{
  return cn90StripLeadingChar(
             cn90StripTrailingChar(
                 cn90StripLeadingChar(
                     cn90StripTrailingChar(
                         szString,
                         _T('\t')),
                     _T('\t')),
                 _T(' ')),
             _T(' '));
} /* end cn90Strip */

/* PUBLIC FUNCTION cn90StripQuotes
 * ------------------------------------------------------------------
 */
_TCHAR * cn90StripQuotes
    ( _TCHAR * szString )
{
  return cn90StripLeadingChar(
             cn90StripTrailingChar(
                 cn90StripLeadingChar(
                     cn90StripTrailingChar(
                         szString,
                         _T('\'')),
                     _T('\'')),
                 _T('"')),
             _T('"'));
} /* end cn90StripQuotes */

/* PUBLIC FUNCTION cn90Stricmp
 * ------------------------------------------------------------------
 */
int  cn90Stricmp
    ( const _TCHAR * szString1,
      const _TCHAR * szString2 )
{
  int nReturn = 0;
  int nIndex  = -1;

  do {
    nIndex++;
    nReturn = _totupper(szString1[nIndex]) - _totupper(szString2[nIndex]);
  } while (szString1[nIndex] != CHAR_STRINGTERM_CN90 &&
           szString2[nIndex] != CHAR_STRINGTERM_CN90 &&
           nReturn           == 0);

  return nReturn;
} /* end cn90Stricmp */

/* PUBLIC FUNCTION cn90StrNicmp
 * ------------------------------------------------------------------
 */
int cn90StrNicmp(const _TCHAR * szString1,
                 const _TCHAR * szString2,
                 size_t         MaxLength)
{
    int    nReturn=0;
    size_t nIndex =0;

    while(0==nReturn &&     //not found any difference
          nIndex<MaxLength) //not reached maximal length
    {
        nReturn=_totupper(szString1[nIndex]) - _totupper(szString2[nIndex]);

        if(CHAR_STRINGTERM_CN90==szString1[nIndex] ||   //reached end of string1 
           CHAR_STRINGTERM_CN90==szString2[nIndex]   )  //reached end of string2
        {
            nIndex=MaxLength;   //end the loop (same as a break)
        }
        else
            nIndex++;
    }
    
    return nReturn;
} /* end cn90StrNicmp */

/* PUBLIC FUNCTION cn90CheckKeyword
 * ------------------------------------------------------------------
 */
bool cn90CheckKeyword
    ( const _TCHAR *     szLine,
      const _TCHAR *     szKeyword,
      const bool         bCaseSens)
{
  int nLength = 0;

  nLength = (int)_tcslen(szKeyword);

  return ( ( (bCaseSens ? _tcsncmp(szLine,szKeyword, nLength)
                        : _tcsnicmp(szLine,szKeyword, nLength)) == 0      ) &&
           ( ( isspace(szLine[nLength]) != 0                         ) ||
             ( szLine[nLength]          == CHAR_STRINGTERM_CN90)     )     );
} /* end cn90CheckKeyword */

/* PUBLIC FUNCTION cn90SaveStrcpy
 * ------------------------------------------------------------------
 */
_TCHAR * cn90SaveStrcpy
    ( _TCHAR       * szDest,
      const _TCHAR * szSource,
      int            nLength )
{
  _tcsncpy(szDest, szSource, nLength);
  szDest[nLength - 1] = CHAR_STRINGTERM_CN90;
  return szDest;
} /* end  cn90SaveStrcpy */

/* PRIVATE FUNCTION cn90GetToken
 * ------------------------------------------------------------------
 */
bool cn90GetToken
    ( const _TCHAR       * szBuffer,
      _TCHAR             * szToken,
      const long           nRequest,
      const long           nLength )
{
  long               nToken      = 0;
  long               i           = 0;
  long               j           = 0;
  bool            bOutside    = true;
  bool            bString     = false;

  /* is buffer valid */
  if (szBuffer != 0) {
    /* loop until we've reached the end of buffer or the requested token is complete */
    for (i = 0; (szBuffer[i] != CHAR_STRINGTERM_CN90) && (nToken <= nRequest) && ((j < nLength) || (nLength == -1)); i++) {
      if (bOutside) {
        /* if we're outside of a token look out for a non whitesapaces */
        if (!isspace((unsigned char) szBuffer[i])) {
          bOutside = false;
          nToken++;
          /* copy the first char of the token */
          if ((szToken != NULL) && (nToken == nRequest)) szToken[j++] = szBuffer[i];
          /* set a flag for string handling */
          bString = szBuffer[i] == CHAR_SYNTAXSTRING_CN90;
        } /* end if */
      } else {
        if (bString) {
          /* copy the char */
          if ((szToken != NULL) && (nToken == nRequest)) szToken[j++] = szBuffer[i];
          /* look out for string terminator */
          if (szBuffer[i] == CHAR_SYNTAXSTRING_CN90 && isspace(szBuffer[i+1])) {
            bOutside = true;
            /* attach the null char */
            if ((szToken != NULL) && (nToken == nRequest)) szToken[j++] = CHAR_STRINGTERM_CN90;
          } /* end if */
        } else {
          /* look out for whitespaces */
          if (isspace((unsigned char) szBuffer[i])) {
            bOutside = true;
            /* attach the null char */
            if ((szToken != NULL) && (nToken == nRequest)) szToken[j++] = CHAR_STRINGTERM_CN90;
          } else {
            /* copy the current char */
            if ((szToken != NULL) && (nToken == nRequest)) szToken[j++] = szBuffer[i];
            bString = szBuffer[i] == CHAR_SYNTAXSTRING_CN90;
          } /* end if */
        } /* end if */
      } /* end if */
    } /* end for */
  } /* end for */

  /* if the token was at the end of the buffer attach a null char */
  if (szToken != NULL && szBuffer[i] == CHAR_STRINGTERM_CN90) szToken[j] = CHAR_STRINGTERM_CN90;

  return ((nToken >= nRequest) && ((j <= nLength) ||(nLength == -1)));
} /* end cn90GetToken */

/* PRIVATE FUNCTION cn90GetStartAndEndOfToken
 * ------------------------------------------------------------------
 */
bool cn90GetStartAndEndOfToken(const _TCHAR *   szBuffer,
                               const _TCHAR * & TokenStart,
                               const _TCHAR * & TokenEnd,
                               const long       nRequest)
{
    long nToken  =0;
    long i       =0;
    bool bOutside=true;
    bool bString =false;

    TokenEnd=TokenStart=0;

    if(0!=szBuffer) // is buffer valid
    {
        while(CHAR_STRINGTERM_CN90!=szBuffer[i] && 0==TokenEnd)  //loop until we've reached the end of buffer or the requested token is complete
        {
            if(bOutside)    // if we're outside of a token ...
            {
                if(!isspace(szBuffer[i]))   //... look out for a non whitesapaces
                {
                    bOutside=false;
                    nToken++;

                    if(nToken==nRequest)
                        TokenStart=szBuffer+i;

                    bString=(CHAR_SYNTAXSTRING_CN90==szBuffer[i]);  // set a flag for string handling
                }
            }
            else            // if we're inside of a token ...
            {
                if(bString)
                {
                    if(CHAR_SYNTAXSTRING_CN90==szBuffer[i] && isspace(szBuffer[i+1]))   // look out for string terminator
                    {
                        bOutside=true;
                        i++;    //position on the space
                    }
                }
                else
                {
                    if(isspace(szBuffer[i]))
                        bOutside=true;
                    else
                        bString=(CHAR_SYNTAXSTRING_CN90==szBuffer[i]);  // set a flag for string handling
                }

                if(nToken==nRequest && bOutside)    //right token and szBuffer[i] is the first space after the token ...
                    TokenEnd=szBuffer+i;            //thats the end we looked for
            }

            i++;
        }
    }

    if(CHAR_STRINGTERM_CN90==szBuffer[i] && 0!=TokenStart && 0==TokenEnd)
        TokenEnd=szBuffer+i;

    return (0!=TokenStart && 0!=TokenEnd);
} // end cn90GetStartAndEndOfToken

/* PUBLIC FUNCTION cn90AnalyzeNumber
 * ------------------------------------------------------------------
 */
bool cn90AnalyzeNumber
    ( _TCHAR           * szBuffer,
      long             * pValue )
{
  long               i           = 0;
  bool            bValid      = true;

  cn90Strip(szBuffer);
  for (i = 0; szBuffer[i] != CHAR_STRINGTERM_CN90; i++) {
    bValid = isdigit(szBuffer[i]) && bValid;
  } /* end if */

  *pValue = (bValid) ? atol(szBuffer) : 0;

  return bValid;
} /* end cn90AnalyzeNumber */



/* PUBLIC FUNCTION cn90AnalyzeString
 * ------------------------------------------------------------------
 */
bool cn90AnalyzeString
    ( const _TCHAR     * szBuffer,
      _TCHAR           * szValue,
      const long         nMaxLen )
{
  long               i           = 0;
  long               nLength     = 0;
  bool               bValid      = true;
  char *             pTemp       = 0;

  nLength = (long)_tcslen(szBuffer);
  pTemp = new char[nLength + 1];

  if (pTemp) {

    _tcscpy(pTemp, szBuffer);
    cn90Strip(pTemp);
    nLength = (long)_tcslen(pTemp);

    if (nLength > 1) {
      if ((pTemp[0]           == CHAR_SYNTAXSTRING_CN90) &&
          (pTemp[nLength - 1] == CHAR_SYNTAXSTRING_CN90)    ) {
        for (i = 1; (i < (nLength - 1)) && bValid; i++) {
          if ( (pTemp[i] == CHAR_SYNTAXSTRING_CN90   ) ||
//               (pTemp[i] == CHAR_SYNTAXFOLLOW_CN90   ) ||
               (pTemp[i] <  CHAR_FIRSTPRINT_CN90     ) ||
               (pTemp[i] >  CHAR_LASTPRINT_CN90      )    ) {
            bValid = false;
          } else {
            szValue[i - 1] = pTemp[i];
          } /* end if */
        } /* end for */

        if (i > (nMaxLen + 1)) {
          bValid = false;
        } /* end if */

      } else {
        bValid = false;
      } /* end if */
    } else {
      bValid = false;
    } /* end if */

    szValue[bValid ? i - 1 : 0] = CHAR_STRINGTERM_CN90;

    delete pTemp;
  } else {
    bValid = false;
  } // end if

  return bValid;
} /* end cn90AnalyzeString */

/* IMPLEMENTATION PUBLIC FUNCTION cn90AnswerIError
 * ------------------------------------------------------------------
 */
tcn00_Error cn90AnswerIError
      ( _TCHAR            * szBuffer,
        int               * pnDataLen,
        const tcn00_Error   nError )
{
  int i = 0;
  struct {
    tcn00_Error   nValue;
    const _TCHAR* szText;
  }               errorArray[] = ERROR_TO_TEXT_CN00;


  while ((errorArray[i].nValue !=  ERR_UNKNOWN_CN00) && (errorArray[i].nValue !=  nError)) {
    i++;
  } /* end while */

  _stprintf (szBuffer, _T("%s%s%d%s%s%s"), ((nError == OK_CN00) ? ANSWER_OK_CN00 : ANSWER_ERR_CN00),
                                           LINE_SEPSTRING_CN00,
                                           nError,
                                           TOKEN_SEPSTRING_CN00,
                                           errorArray[i].szText,
                                           LINE_SEPSTRING_CN00);
  *pnDataLen = (int)_tcslen(szBuffer);

  return nError;
} /* end cn90AnswerIError */

/* IMPLEMENTATION PUBLIC FUNCTION cn90AnswerEvent
 * ------------------------------------------------------------------
 */
teo200_EventList::teo200_EventID cn90AnswerEvent
      ( _TCHAR                 * szBuffer,
        int                    * pnDataLen,
        const teo200_EventList & EvtLst )
{

  _TCHAR           * pCurrent = szBuffer;
  teo200_EventList * pNextEvt = EvtLst.eo200_NextEvent();
  
  _stprintf(pCurrent, "%s%s%d,%s%s",
            (EvtLst.eo200_EventID() == cn00_1_ErrId(OK_CN00_1) ) ? ANSWER_OK_CN00 : ANSWER_ERR_CN00,
            LINE_SEPSTRING_CN00,
            (int) EvtLst.eo200_EventID(),
            EvtLst.eo200_EventMsg(),
            LINE_SEPSTRING_CN00);

  while (pNextEvt != NULL) {
    pCurrent = pCurrent + _tcslen(pCurrent);

    _stprintf(pCurrent, "%d,%s%s",
              (int) pNextEvt->eo200_EventID(),
              pNextEvt->eo200_EventMsg(),
              LINE_SEPSTRING_CN00);

    pNextEvt = pNextEvt->eo200_NextEvent();
  } // end while

  *pnDataLen = (int)_tcslen(szBuffer);

  return EvtLst.eo200_EventID();
} /* end cn90AnswerEvent */

/* IMPLEMENTATION PUBLIC FUNCTION cn90AnswerRTEError
 * ------------------------------------------------------------------
 */
tcn00_Error cn90AnswerRTEError
      ( _TCHAR            * szBuffer,
        int               * pnDataLen,
        const tcn00_Error   nError,
        const _TCHAR      * szRteMsg,
        const int           nMsgLen,
        const int           nRteError )
{
  int i = 0;
  struct {
    tcn00_Error   nValue;
    const _TCHAR* szText;
  }               errorArray[] = ERROR_TO_TEXT_CN00;

  while ((errorArray[i].nValue !=  ERR_UNKNOWN_CN00) && (errorArray[i].nValue !=  nError)) {
    i++;
  } /* end while */

  _TCHAR szText[256];

  cn90StringPtoC(szText, szRteMsg, (nMsgLen > 255) ? 255 : nMsgLen);
  
  _stprintf (szBuffer, _T("%s%s%d%s%s%s%d%s%s%s"), ((nError == OK_CN00) ? ANSWER_OK_CN00 : ANSWER_ERR_CN00),
                                                   LINE_SEPSTRING_CN00,
                                                   nError,
                                                   TOKEN_SEPSTRING_CN00,
                                                   errorArray[i].szText,
                                                   LINE_SEPSTRING_CN00,
                                                   nRteError,
                                                   TOKEN_SEPSTRING_CN00,
                                                   szText,
                                                   LINE_SEPSTRING_CN00);

  *pnDataLen = (int)_tcslen(szBuffer);

  return nError;
} /* end cn90AnswerRTEError */

/* IMPLEMENTATION PUBLIC FUNCTION cn90AnswerNewRTEError
 * ------------------------------------------------------------------
 */
tcn00_Error cn90AnswerNewRTEError
      ( _TCHAR            *   szBuffer,
        int               *   pnDataLen,
        const tcn00_Error     nError,
        const tsp01_RteError * pRteError)
{
  int i = 0;
  struct {
    tcn00_Error   nValue;
    const _TCHAR* szText;
  }               errorArray[] = ERROR_TO_TEXT_CN00;

  while ((errorArray[i].nValue !=  ERR_UNKNOWN_CN00) && (errorArray[i].nValue !=  nError)) {
    i++;
  } /* end while */

  _stprintf (szBuffer, _T("%s%s%d%s%s%s%d%s%s%s%d%s%s%s"),
                       ((nError == OK_CN00) ? ANSWER_OK_CN00 : ANSWER_ERR_CN00),
                       LINE_SEPSTRING_CN00,
                       nError,
                       TOKEN_SEPSTRING_CN00,
                       errorArray[i].szText,
                       LINE_SEPSTRING_CN00,
                       pRteError->RteErrCode,
                       TOKEN_SEPSTRING_CN00,
                       (_TCHAR *) pRteError->RteErrText,
                       LINE_SEPSTRING_CN00,
                       pRteError->OsErrCode,
                       TOKEN_SEPSTRING_CN00,
                       (_TCHAR *) pRteError->OsErrText,
                       LINE_SEPSTRING_CN00);

  *pnDataLen = (int)_tcslen(szBuffer);

  return nError;
} /* end cn90AnswerNewRTEError */

/* IMPLEMENTATION PUBLIC FUNCTION cn90AnswerSQLError
 * ------------------------------------------------------------------
 */
tcn00_Error cn90AnswerSQLError
      ( _TCHAR            * szBuffer,
        int               * pnDataLen,
        const tcn00_Error   nError,
        const _TCHAR      * szSQLMsg,
        const int           nSQLError )
{
  int i = 0;
  struct {
    tcn00_Error   nValue;
    const _TCHAR* szText;
  }               errorArray[] = ERROR_TO_TEXT_CN00;

  while ((errorArray[i].nValue !=  ERR_UNKNOWN_CN00) && (errorArray[i].nValue !=  nError)) {
    i++;
  } /* end while */

  _stprintf (szBuffer, _T("%s%s%d%s%s%s%d%s%s%s"), ((nError == OK_CN00) ? ANSWER_OK_CN00 : ANSWER_ERR_CN00),
                                                   LINE_SEPSTRING_CN00,
                                                   nError,
                                                   TOKEN_SEPSTRING_CN00,
                                                   errorArray[i].szText,
                                                   LINE_SEPSTRING_CN00,
                                                   nSQLError,
                                                   TOKEN_SEPSTRING_CN00,
                                                   szSQLMsg,
                                                   LINE_SEPSTRING_CN00);

  *pnDataLen = (int)_tcslen(szBuffer);

  return nError;
} /* cn90AnswerSQLError */

/* IMPLEMENTATION PUBLIC FUNCTION cn90AnswerErrnoError
 * ------------------------------------------------------------------
 */
tcn00_Error cn90AnswerErrnoError
      ( _TCHAR            * szBuffer,
        int               * pnDataLen,
        const tcn00_Error   nError,
        const int           nErrno,
        _TCHAR            * szInfo)
{
  int i = 0;
  int j = 0;
  struct {
    tcn00_Error   nValue;
    const _TCHAR* szText;
  }               errorArray[] = ERROR_TO_TEXT_CN00;
  struct {
    int           nValue;
    const _TCHAR* szText;
  }               errnoArray[] = ERRNO_TO_TEXT;

  while ((errorArray[i].nValue !=  ERR_UNKNOWN_CN00) && (errorArray[i].nValue !=  nError)) {
    i++;
  } /* end while */
  while ((errnoArray[j].szText !=  NULL) && (errnoArray[j].nValue != nErrno)) {
    j++;
  } /* end while */
  if (errnoArray[j].szText ==  NULL) {
    j++;
  } /* end if */

  _stprintf (szBuffer, _T("%s%s%d%s%s%s%d%s%s%s%s%s"), ((nError == OK_CN00) ? ANSWER_OK_CN00 : ANSWER_ERR_CN00),
                                                       LINE_SEPSTRING_CN00,
                                                       nError,
                                                       TOKEN_SEPSTRING_CN00,
                                                       errorArray[i].szText,
                                                       LINE_SEPSTRING_CN00,
                                                       nErrno,
                                                       TOKEN_SEPSTRING_CN00,
                                                       errnoArray[j].szText,
                                                       LINE_SEPSTRING_CN00,
                                                       (szInfo == NULL ? "" : szInfo),
                                                       (szInfo == NULL ? "" : LINE_SEPSTRING_CN00));

  *pnDataLen = (int)_tcslen(szBuffer);

  return nError;
} /* end cn90AnswerErrnoError */

/* IMPLEMENTATION PUBLIC FUNCTION cn90AnswerExecute
 * ------------------------------------------------------------------
 */
tcn00_Error cn90AnswerExecute
      ( _TCHAR            * szBuffer,
        int               * pnDataLen,
        const int           nMaxLen,
        const tcn00_Error   nError,
        const int           nSysReturn,
        const _TCHAR      * szCommand,
        const _TCHAR      * szTmpFile )
{
  int    i     = 0;
  FILE * hFile = NULL;
  _TCHAR szLine[PARAMETER_MAXLEN_CN90 + 1];

  struct {
    tcn00_Error    nValue;
    const _TCHAR * szText;
  }               errorArray[] = ERROR_TO_TEXT_CN00;

  while ((errorArray[i].nValue !=  ERR_UNKNOWN_CN00) && (errorArray[i].nValue !=  nError)) {
    i++;
  } /* end while */

  _stprintf (szBuffer, _T("%s%s%d%s%s%s%d%s%s%s"), ((nError == OK_CN00) ? ANSWER_OK_CN00 : ANSWER_ERR_CN00),
                                                   LINE_SEPSTRING_CN00,
                                                   nError,
                                                   TOKEN_SEPSTRING_CN00,
                                                   errorArray[i].szText,
                                                   LINE_SEPSTRING_CN00,
                                                   nSysReturn,
                                                   TOKEN_SEPSTRING_CN00,
                                                   szCommand,
                                                   LINE_SEPSTRING_CN00);

  if (szTmpFile != NULL) {
    hFile = fopen(szTmpFile, _T("r"));
    if (hFile) {
      while (fgets(szLine, PARAMETER_MAXLEN_CN90, hFile) != NULL && ((int) (strlen(szBuffer) + strlen(szLine) + 5) < nMaxLen)) {
        strcat(szBuffer, cn90StripTrailingChar(szLine, _T('\n')));
        strcat(szBuffer, LINE_SEPSTRING_CN00);
      } /* end while */
      fclose(hFile);
    } /* end if */
  } /* end if */

  *pnDataLen = (int)_tcslen(szBuffer);

  return nError;
} /* end cn90AnswerExecute */

/* IMPLEMENTATION PUBLIC FUNCTION cn90AnswerOK
 * ------------------------------------------------------------------
 */
tcn00_Error cn90AnswerOK
      ( _TCHAR       * szBuffer,
        int          * pnDataLen,
        const _TCHAR * szMsg )
{
  if (szMsg == NULL) {
    _stprintf (szBuffer, _T("%s%s"), ANSWER_OK_CN00,
                                     LINE_SEPSTRING_CN00);
  } else {
    _stprintf (szBuffer, _T("%s%s%s%s"), ANSWER_OK_CN00,
                                         TOKEN_SEPSTRING_CN00,
                                         szMsg,
                                         LINE_SEPSTRING_CN00);
  } /* end if */

  *pnDataLen = (int)_tcslen (szBuffer);

  return OK_CN00;
} /* end cn90AnswerOK */

/* PUBLIC FUNCTION cn90Trace
 * ------------------------------------------------------------------
 */
#ifdef CNTRACE

#define TRACE_FILE_CN90     "dbmsrv.prt"
#define TRACE_MAX_CN90      800000

void cn90Trace
      ( const char   * szString1,
        const char   * szString2 )
{
  FILE          * hTrace;
  FILE          * hFile;
  tsp00_C256      szDir;
  char            szFilename[100];
  time_t          aTime;
  size_t          nSize = 0;
  void          * pBuffer = NULL;

  cn90GetWrkPath(szDir);
  strcpy(szFilename, szDir);
  strcat(szFilename, TRACE_FILE_CN90);

  time(&aTime);

  hTrace = fopen(szFilename  , "a+");
  if (hTrace) {
    fseek(hTrace, 0, SEEK_END);
    nSize = ftell(hTrace);
    if (nSize > TRACE_MAX_CN90) {
      fclose(hTrace);
      hTrace = NULL;
      hFile = fopen(szFilename , "rb");
      if (hFile) {
        fseek(hFile, TRACE_MAX_CN90 / 2, SEEK_SET);
        pBuffer = malloc(nSize - TRACE_MAX_CN90 / 2);
        if (pBuffer) {
          fread(pBuffer, nSize - TRACE_MAX_CN90 / 2, 1, hFile);
          fclose(hFile);
          hFile = fopen(szFilename , "wb");
          if (hFile) {
            fwrite(pBuffer, nSize - TRACE_MAX_CN90 / 2, 1, hFile);
            fclose(hFile);
            hTrace = fopen(szFilename  , "a");
          } /* end if */
        } else {
          fclose(hFile);
        } /* end if */
      } /* end if */
    } /* end if */
  } /* end if */

  if (hTrace) {
    fprintf (hTrace, "%s %s : %s\n", ctime(&aTime), szString1, szString2);
    fclose(hTrace);
  } /* end if */

} /* end cn90Trace */
#endif /* CNTRACE */

/* IMPLEMENTATION PUBLIC FUNCTION cn90Uncrypt
 * ------------------------------------------------------------------
 */
#ifndef MAX_CRYPT
  #define MAX_CRYPT csp_cryptname
#endif
#define CRYPT_BASE      16

char * cn90Uncrypt
      ( char      * szPassword,
        bool        bForDatabase)
{
  tsp_cryptname pwCrypted;
  int           i;
  int           nLenNumber = sizeof(tsp_cryptname) / MAX_CRYPT * 2;
  char          szNumber[2];
  char        * pEnd;

  memset(&pwCrypted, 0, sizeof(pwCrypted));

  if ((int) strlen(szPassword) == nLenNumber * MAX_CRYPT) {
    szNumber[1] = CHAR_STRINGTERM_CN90;
    for (i = 0; szPassword[i] != CHAR_STRINGTERM_CN90; i++) {
      szNumber[0] = szPassword[i];
      pwCrypted[i / nLenNumber] = (pwCrypted[i / nLenNumber] * CRYPT_BASE) + strtoul(szNumber, &pEnd, CRYPT_BASE);
    } /* end for */

    cn90CalculateUncrypt(pwCrypted, szPassword, bForDatabase);
  } /* end if */

  return szPassword;
} /* end cn90Uncrypt */

/* IMPLEMENTATION PUBLIC FUNCTION cn90CalculateUncrypt
 * ------------------------------------------------------------------
 */
char * cn90CalculateUncrypt
      ( const tsp_cryptpw     pwCrypted,
        char                * passwd,
        bool                  bForDatabase)
{
  tsp_name pwClear;
  tsp_namec pwClearc;

  if (bForDatabase) {
    s02decrypt(pwClear, pwCrypted);
  } else {
    s02appldecrypt(pwClear, pwCrypted);
  } // end if

  pwClearc.p2c(pwClear);
  strcpy(passwd, pwClearc);

  return passwd;
} // end cn90CalculateUncrypt

/* IMPLEMENTATION PUBLIC FUNCTION cn90CalculateUncrypt
 * ------------------------------------------------------------------
 */

char * cn90CalculateUncrypt
      ( const tsp_cryptname   pwCrypted,
        char                * passwd,
        bool               bForDatabase)
{
  tsp_cryptpw pwTemp;

  memcpy(&pwTemp[0], &pwCrypted[0], sizeof(tsp_cryptpw));

  return cn90CalculateUncrypt(pwTemp, passwd, bForDatabase);
} // end cn90CalculateUncrypt

/* IMPLEMENTATION PUBLIC FUNCTION cn90DirCreate
 * ------------------------------------------------------------------
 */
tcn00_Error cn90DirCreate
      ( char           * szDirectory,
        tsp01_RteError * pRteError )
{
  tcn00_Error      nFuncReturn = OK_CN00;
  unsigned int     nIndex      = 0;
  tsp01_RteError   aRteError;

  /* change slashes */
  cn90ChangePathDelimiter(szDirectory);

  /* try to create directory */
  if (! sqlcreate_dirc (szDirectory, (pRteError == NULL) ? &aRteError : pRteError )) {
    nFuncReturn = ERR_RTEEXT_CN00;
  } /* end if */

  return nFuncReturn;
} /* end cn90DirCreate */

/* IMPLEMENTATION PUBLIC FUNCTION cn90ChangePathDelimiter
 * ------------------------------------------------------------------
 */
char * cn90ChangePathDelimiter
      ( char          * szPath )
{
  unsigned long  nIndex      = 0;

  /* change backslash and slash to PATH_DELIMITTER */
  for (nIndex = 0; nIndex < strlen(szPath); nIndex++) {
    szPath[nIndex] = (szPath[nIndex] == CHAR_BACKSLASH) ? CHAR_SLASH          : szPath[nIndex];
    szPath[nIndex] = (szPath[nIndex] == CHAR_SLASH)     ? PATH_DELIMITER_CN90 : szPath[nIndex];
  } /* end for */

  return szPath;
} /* end cn90ChangePathDelimiter */

/* IMPLEMENTATION PUBLIC FUNCTION cn90GetCfgPath
 * ------------------------------------------------------------------
 */
char * cn90GetCfgPath
      ( tsp00_C256 & szDir)
{
  tsp01_RteError   aRteError;
  tsp00_Pathc      szLokDir;

  sqlGetIndependentConfigPath(szLokDir, TERM_WITH_DELIMITER_EO01, &aRteError);
  _tcscpy(szDir, szLokDir);

  return (char *) szDir;
} /* end cn90GetCfgPath */

/* IMPLEMENTATION PUBLIC FUNCTION cn90GetWrkPath
 * ------------------------------------------------------------------
 */
char * cn90GetWrkPath
      ( tsp00_C256 & szDir)
{
  tsp01_RteError  aRteError;
  tsp00_Pathc     szLokDir;

  sqlGetIndependentWrkPath(szLokDir, TERM_WITH_DELIMITER_EO01, &aRteError);
  _tcscpy(szDir, szLokDir);

  return (char *) szDir;
} /* end cn90GetWrkPath */

/* IMPLEMENTATION PUBLIC FUNCTION cn90GetVersionEnvPath
 * ------------------------------------------------------------------
 */
char * cn90GetVersionEnvPath
      ( tsp00_C256 & szDir)
{

  tsp01_RteError  aRteError;
  tsp00_Pathc      szLokDir;

  sqlGetDbrootEnvPath(szLokDir, TERM_WITH_DELIMITER_EO01, &aRteError);
  _tcscpy(szDir, szLokDir);

  return (char *) szDir;
} /* end cn90GetVersionEnvPath */

/* IMPLEMENTATION PUBLIC FUNCTION cn90GetVersionBinPath
 * ------------------------------------------------------------------
 */
char * cn90GetVersionBinPath
      ( tsp00_C256 & szDir)
{
  tsp01_RteError  aRteError;
  tsp00_Pathc     szLokDir;

  sqlGetDbrootBinPath(szLokDir, TERM_WITH_DELIMITER_EO01, &aRteError);
  _tcscpy(szDir, szLokDir);

  return (char *) szDir;
} /* end cn90GetVersionBinPath */

/*
  -----------------------------------------------------------------------------
  function:     cn90DBState
  -----------------------------------------------------------------------------
*/
  tcn00_DBState cn90DBState
        ( const tsp00_DbNamec szDbName )
{
  tcn00_DBState     DBState = STATE_UNKNOWN_CN00;
  tsp9_rte_xerror   xError;
#if defined(_WIN32)
  void            * hMonitor      = NULL;
  const char      * szDescription = NULL;
  char              szBuffer[PARAMETER_MAXLEN_CN90];
  int               nLine         = 0;
#else
  tsp00_Uint4       kernelDbState;
#endif /* _WIN32 */

  /* set global variables for sqlxknlopenmonitor */
  tsp00_Pathc     szDBRoot;
  sqlxinit (szDbName, cn90DBRoot(szDbName, szDBRoot));

#if defined(_WIN32)
  sqlxknlopenmonitor ("state", 0, &hMonitor, &szDescription, &xError);

  if (xError.xe_result != csp9_xrte_ok) {
    DBState = STATE_OFFLINE_CN00;
    sqlxcanstart( &xError );
    if ( xError.xe_result != csp9_xrte_ok ) {
      DBState = STATE_RUIN_CN00;
    } else {
      DBState = STATE_OFFLINE_CN00;
    } // end if
  } else {
    while (xError.xe_result == csp9_xrte_ok && nLine < 2) {
      sqlxknlmonitor (hMonitor,
                      csp9_read_next,
                      szBuffer,
                      PARAMETER_MAXLEN_CN90,
                      &xError);
      nLine++;
    } /* end while */

    if (xError.xe_result == csp9_xrte_atend || xError.xe_result == csp9_xrte_ok) {

      cn90Strip(szBuffer);

      if (_tcsicmp(szBuffer, TXT_STATE_OFFLINE_CN00) == 0) {
        DBState = STATE_OFFLINE_CN00;     
      } else if (_tcsicmp(szBuffer, TXT_STATE_STARTING_CN00) == 0) {
        DBState = STATE_STARTING_CN00;    
      } else if (_tcsicmp(szBuffer, TXT_STATE_COLD_CN00) == 0) {
        DBState = STATE_COLD_CN00;        
      } else if (_tcsicmp(szBuffer, TXT_STATE_WARM_CN00) == 0) {
        DBState = STATE_WARM_CN00;        
      } else if (_tcsicmp(szBuffer, TXT_STATE_SHTDWN_CN00) == 0) {
        DBState = STATE_SHTDWN_CN00;      
      } else if (_tcsicmp(szBuffer, TXT_STATE_SHTDWNREI_CN00) == 0) {
        DBState = STATE_SHTDWNREI_CN00;   
      } else if (_tcsicmp(szBuffer, TXT_STATE_SHTDWNKILL_CN00) == 0) {
        DBState = STATE_SHTDWNKILL_CN00;  
      } else if (_tcsicmp(szBuffer, TXT_STATE_STOP_CN00) == 0) {
        DBState = STATE_STOP_CN00;        
      } else if (_tcsicmp(szBuffer, TXT_STATE_KILL_CN00) == 0) {
        DBState = STATE_KILL_CN00;        
      } else if (_tcsicmp(szBuffer, TXT_STATE_ABORT_CN00) == 0) {
        DBState = STATE_ABORT_CN00;       
      } else if (_tcsicmp(szBuffer, TXT_STATE_RUIN_CN00) == 0) {
        DBState = STATE_RUIN_CN00;        
      } else if (_tcsicmp(szBuffer, TXT_STATE_RUNNING_CN00) == 0) {
        DBState = STATE_RUNNING_CN00;     
      } else if (_tcsicmp(szBuffer, TXT_STATE_UNKNOWN_CN00) == 0) {
        DBState = STATE_UNKNOWN_CN00;     
      } else {
        DBState = STATE_UNKNOWN_CN00;
      } /* end if */

    } else {
      DBState = STATE_ERROR_CN00;
    } /* end if */

    sqlxknlclosemonitor (hMonitor);
  } /* end if */
#else
  kernelDbState = sqlxdbstate( &xError );
  if ( xError.xe_result != csp9_xrte_ok )
  {
      DBState = STATE_ERROR_CN00;
  }
  else
  {
      switch( kernelDbState )
      {
        case SERVER_STARTING:
            DBState = STATE_UNKNOWN_CN00;
            break;
        case SERVER_COLD:
            DBState = STATE_COLD_CN00;
            break;
        case SERVER_WARM:
            DBState = STATE_WARM_CN00;
            break;
        case SERVER_KILL:
            DBState = STATE_KILL_CN00;
            break;
        case SERVER_STOP:
            DBState = STATE_STOP_CN00;
            break;
        case SERVER_SHUTDOWN:
            DBState = STATE_SHTDWN_CN00;
            break;
        case SERVER_SHUTDOWNREINIT:
            DBState = STATE_SHTDWNREI_CN00;
            break;
        case SERVER_SHUTDOWNKILL:
            DBState = STATE_SHTDWNKILL_CN00;
            break;
        default:
            sqlxcanstart( &xError );
            if ( xError.xe_result != csp9_xrte_ok )
            {
                DBState = STATE_RUIN_CN00;
            } else {
                DBState = STATE_OFFLINE_CN00;
            } // end if
            break;
      }
  } // end if
#endif /* _WIN32 */

  sqlxfinish();

  return DBState;
} /* end cn90DBState */

/*
  -----------------------------------------------------------------------------
  function cn90DBSpeed
  -----------------------------------------------------------------------------
*/
tsp9_pgm_kind cn90DBSpeed
      ( const tsp00_DbNamec szDbName )
{
  tsp9_pgm_kind nDBSpeed = csp9_any_pgm;
  tsp9_rte_xerror     xError;
  void              * hEnum;
  tsp9_rte_dbinfo     dbInfo;

  sqlxopen_db_enum (&hEnum, true, &xError);

  if (xError.xe_result == csp9_xrte_ok) {
    while (xError.xe_result == csp9_xrte_ok) {
      sqlxnext_db (hEnum, &dbInfo, &xError);
      if (xError.xe_result == csp9_xrte_ok) {
        if ((_tcsicmp(dbInfo.dbname, szDbName) == 0) && dbInfo.state == csp9_state_running ) {
          nDBSpeed = dbInfo.pgmKind;
        } // end if
      } // end if
    } // end while

    sqlxclose_db_enum (hEnum);
  } // end if

  return nDBSpeed;
} // cn90DBSpeed

/*
  -----------------------------------------------------------------------------
  function cn90DBExists
  -----------------------------------------------------------------------------
*/
bool cn90DBExists
      (const tsp_dbnamec szDbName)
{
  tsp_int4 nCounter = 0;

  return cn90DBExists(szDbName, nCounter);
} // cn90DBExists
/*
  -----------------------------------------------------------------------------
*/
bool cn90DBExists
      (const tsp_dbnamec szDbName, tsp_int4 & nCounter)
{
  bool bDBExists = false;
  tsp9_rte_xerror     xError;
  void              * hEnum;
  tsp9_rte_dbinfo     dbInfo;

  sqlxopen_db_enum (&hEnum, false, &xError);

  nCounter = 0;

  if (xError.xe_result == csp9_xrte_ok) {
    while (xError.xe_result == csp9_xrte_ok) {
      sqlxnext_db (hEnum, &dbInfo, &xError);
      if (xError.xe_result == csp9_xrte_ok) {
         if (_tcsicmp (dbInfo.dbname, szDbName) == 0) {
           bDBExists = true;
           nCounter++;
         } // end if
      } // end if
    } // end while

    sqlxclose_db_enum (hEnum);
  } // end if

  return bDBExists;

} // cn90DBExists
/*
  -----------------------------------------------------------------------------
*/
bool cn90DBExists
      (const tsp_dbnamec szDbName, tsp9_pgm_kind pgmKind)
{
  bool bDBExists = false;
  tsp9_rte_xerror     xError;
  void              * hEnum;
  tsp9_rte_dbinfo     dbInfo;

  sqlxopen_db_enum (&hEnum, false, &xError);

  if (xError.xe_result == csp9_xrte_ok) {
    while (xError.xe_result == csp9_xrte_ok) {
      sqlxnext_db (hEnum, &dbInfo, &xError);
      if (xError.xe_result == csp9_xrte_ok) {
         if (_tcsicmp (dbInfo.dbname, szDbName) == 0 && dbInfo.pgmKind == pgmKind) {
           bDBExists = true;
         } // end if
      } // end if
    } // end while

    sqlxclose_db_enum (hEnum);
  } // end if

  return bDBExists;

} // cn90DBExists

/*
  -----------------------------------------------------------------------------
  function cn90ClearStringToCryptString
  -----------------------------------------------------------------------------
*/
void cn90ClearStringToCryptString
      ( tsp00_Namec    szClearString,
        tsp00_C64c   & szCryptString)
{
  tsp00_CryptPw   stCrypt;
  tsp00_Name      pasString;

  s02applencrypt(pasString.c2p(szClearString), stCrypt);
  cn90CryptToCryptString(stCrypt, szCryptString);
} // end cn90CryptToString

/*
  -----------------------------------------------------------------------------
  function cn90CryptStringToClearString
  -----------------------------------------------------------------------------
*/
void cn90CryptStringToClearString
      ( tsp00_C64c     szCryptString,
        tsp00_Namec  & szClearString,
        bool           bForDatabase)
{
  tsp00_C64c szTemp;

  szTemp = szCryptString;

  szClearString.rawAssign(cn90Uncrypt(szTemp, bForDatabase));

} // end cn90CryptStringToClearString

/*
  -----------------------------------------------------------------------------
  function cn90CryptToCryptString
  -----------------------------------------------------------------------------
*/
void cn90CryptToCryptString
      ( tsp00_CryptPw     stCrypt,
        tsp00_C64c      & szCryptString)
{
  tsp00_CryptName stTemp;

  memcpy (&stTemp[0], &stCrypt[0], sizeof(stTemp));

  cn90CryptToCryptString(stTemp, szCryptString);
} // end cn90CryptToString
/*
  -----------------------------------------------------------------------------
*/
void cn90CryptToCryptString
      ( tsp00_CryptName   stCrypt,
        tsp00_C64c      & szCryptString)
{
  _stprintf ( szCryptString , _T("%08x%08x%08x%08x%08x%08x"),
                              (int) stCrypt[0],
                              (int) stCrypt[1],
                              (int) stCrypt[2],
                              (int) stCrypt[3],
                              (int) stCrypt[4],
                              (int) stCrypt[5]);
} // end cn90CryptToString

/*
  -----------------------------------------------------------------------------
  function cn90DBRoot
  -----------------------------------------------------------------------------
*/
_TCHAR * cn90DBRoot
      (const tsp_dbnamec & szDbName, tsp00_Pathc & szDBRoot)
{
  // PTS 1109637, 1109638, 1109639
#if 1
  static tsp_dbnamec oldDbName;
  static tsp00_Pathc oldDBRoot;

  if ( strcmp(oldDbName.asCharp(), szDbName.asCharp()) != 0 )
  {
    oldDbName = szDbName;
    if (!sqlGetDBRootOfDB( oldDbName.asCharp(),
                           oldDBRoot.asCharp(),
                           oldDBRoot.size())) {
      oldDBRoot.Init();
    } // end if
  }

  szDBRoot = oldDBRoot;
#else
  if (!sqlGetDBRootOfDB( szDbName.asCharp(), 
                         szDBRoot.asCharp(),
                         szDBRoot.size())) {
    szDBRoot.Init();
  } // end if
#endif
  return szDBRoot.asCharp();

/*
  tsp9_rte_xerror     xError;
  void              * hEnum;
  tsp9_rte_dbinfo     dbInfo;

  szDBRoot.Init();

  sqlxopen_db_enum (&hEnum, false, &xError);

  if (xError.xe_result == csp9_xrte_ok) {
    while (xError.xe_result == csp9_xrte_ok && szDBRoot.length() == 0) {
      sqlxnext_db (hEnum, &dbInfo, &xError);
      if (xError.xe_result == csp9_xrte_ok) {
         if (_tcsicmp(dbInfo.dbname, szDbName.asCharp()) == 0) {
           szDBRoot.rawAssign(dbInfo.dbroot);
         } // end if
      } // end if
    } // end while

    sqlxclose_db_enum (hEnum);
  } // end if

  return szDBRoot.asCharp();
*/
} // cn90DBRoot

/*
  -----------------------------------------------------------------------------
  function cn90ServiceDB
  -----------------------------------------------------------------------------
*/
#ifndef SERVICE_PARAMFILE_ID_EO01
  #define SERVICE_PARAMFILE_ID_EO01       "_"
#endif

_TCHAR * cn90ServiceDB
      (tsp_dbnamec & szDbName, const tsp00_Pathc & szDBRoot)
{
  tsp9_rte_xerror     xError;
  void              * hEnum;
  tsp9_rte_dbinfo     dbInfo;

  szDbName.Init();

  sqlxopen_db_enum (&hEnum, false, &xError);

  if (xError.xe_result == csp9_xrte_ok) {
    while (xError.xe_result == csp9_xrte_ok && szDbName.length() == 0) {
      sqlxnext_db (hEnum, &dbInfo, &xError);
      if (xError.xe_result == csp9_xrte_ok) {
        if ( ( _tcsicmp(dbInfo.dbroot, szDBRoot                                                     ) == 0 ) &&
             ( _tcsncmp(dbInfo.dbname, SERVICE_PARAMFILE_ID_EO01, _tcslen(SERVICE_PARAMFILE_ID_EO01)) == 0 )    ) {
           szDbName.rawAssign(dbInfo.dbname);
         } // end if
      } // end if
    } // end while

    sqlxclose_db_enum (hEnum);
  } // end if

  return szDbName.asCharp();
} // cn90ServiceDB

/* 
  -----------------------------------------------------------------------------
  function cn90Timestamp
  -----------------------------------------------------------------------------
*/
const tsp00_Timestampc & cn90Timestamp
    ( tsp00_Timestampc & szTimestamp )
{
  tsp00_Date  DateP;
  tsp00_Time  TimeP;
  tsp00_Datec DateC;
  tsp00_Timec TimeC;
    
  sqldattime(DateP, TimeP);
  DateC.p2c(DateP);
  TimeC.p2c(TimeP);

  _stprintf(szTimestamp, "%.8s%.6s", DateC.asCharp(), (TimeC.asCharp() + 2));

  return szTimestamp;

} // end cn90Timestamp
