/*******************************************************************************
*
* module:       gco001.h
*
* ------------------------------------------------------------------------------
*
* responsible:  T.Anhaus
*
* lastChanged:  1999-07-20
*
* purpose:      Base Types of OMS
*
* generatedBy: 
*
* ------------------------------------------------------------------------------
*
* Copyright by SAP AG, 1999
*


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

********************************************************************************/

#ifndef GCO001_H
#define GCO001_H

#include "ggg92.h" 
#include "ilivecachesink.h" /* no check */
#include "gco00.h"
#include "hco30-k.h"

#if defined(BIT64)
#define USE_CONTAINER_HANDLE_CO001
#endif

enum {
  RGN_STACK_HEAP     = 7,
  RGN_SHARED_HEAP    = 6,
  RGN_VERSION_DIR    = 5,
  RGN_CLASS_DIR      = 4,
  RGN_UNLOAD_VERSION = 3,
  RGN_MONITOR        = 2,
  RGN_HEAP_COUNTER   = 1
};

#define OMS_DEFAULT_SCHEMA_HANDLE        1
#define INTERNAL_CRITICAL_SECTIONS_CO001 8
#define MAX_CRITICAL_SECTIONS_CO001      16

//typedef unsigned int OmsContainerNo;

struct OmsVarOid : public OmsObjectId
{
public :
  bool operator==( const OmsObjectId&r ) const
  { 
    return ((OmsObjectId) (*this) == (OmsObjectId) r);
  }
  bool operator!=( const OmsObjectId&r ) const 
	{ 
    return ((OmsObjectId) (*this) != (OmsObjectId) r);
  }
  bool operator<( const OmsObjectId&r ) const 
  { return (memcmp (this, &r, sizeof(r)) <  0);}
  bool operator<=( const OmsObjectId&r ) const 
  { return (memcmp (this, &r, sizeof(r)) <= 0);}
};

#define VAR_OBJ_CHUNK_SIZE (MAX_RECLEN_GG00 - cgg_rec_key_offset - sizeof(OmsObjectId) - sizeof(tgg00_ObjFileNo) - 4 - sizeof (OmsVarOid) - sizeof (tsp00_Uint4))

struct OmsVarObjContainer_co001 {
  unsigned char vobjInfo[VAR_OBJ_CHUNK_SIZE];
  tsp00_Uint4 m_vobjSize;
  OmsVarOid   m_vobjNext;
};

struct OmsVarObjChunk_co001 {
  tgg91_PageRef            m_seq;
  OmsObjectId              m_oid;
  tsp00_Uint4              m_containerHandle;
  OmsVarObjContainer_co001 m_container;
  void Clear() {
    m_oid.setNil();
  }
};

class CHResult_co001 {
public :
  CHResult_co001() : m_hresult(0) {}

  CHResult_co001(HRESULT hr) : m_hresult(hr) {
    if (FAILED(hr)) {
      throw DbpError(DbpError::HRESULT_ERROR, hr);
    }
  }

  void operator= (HRESULT hr) {
    if (FAILED(hr)) {
      throw DbpError(DbpError::HRESULT_ERROR, hr);
    }
    m_hresult = hr;
  }
private :
  HRESULT m_hresult;
};

class CriticalSection_co001 {
private :
  IliveCacheSink*   m_lcSink;
  int               m_regionId;
  bool              m_inSection;
public :
  CriticalSection_co001(IliveCacheSink* lcSink,int regionId) : m_lcSink(lcSink), m_regionId(regionId), m_inSection(false) {} 

  void Enter() {
    if (m_inSection) {
      throw DbpError(DbpError::RTE_ERROR, -1); 
    }
    else {
      m_lcSink->EnterCriticalSection(m_regionId);
      m_inSection = true;
    }
  }

  void Leave() {
    if (m_inSection) {
      m_lcSink->LeaveCriticalSection(m_regionId);
      m_inSection = false;
    }
    else {
      throw DbpError(DbpError::RTE_ERROR, -1); 
    }
  }
  ~CriticalSection_co001() {
    if (m_inSection) {
      Leave();
    }
  }
};

class IGuidHash_co001 {
public :
  static tsp00_Int4 HashValue(const GUID& guid, tsp00_Int4 maxSlot) {
    return guid.Data1 % maxSlot;
  }
  static tsp00_Int4 HashValue(const GUID& guid, OmsSchemaHandle s, OmsContainerNo c, tsp00_Int4 maxSlot) {
    return (guid.Data1 ^ (s * 48623) ^ (c * 7)) % maxSlot;
  }
};

inline tsp00_Int4 co001_Min(tsp00_Int4 t1, tsp00_Int4 t2) {
  if (t1 < t2) {
    return t1;
  }
  else {
    return t2;
  }
}

#endif
