/* @lastChanged: "1998-10-22  8:37"
 * @filename:    gco01.h
 * @purpose:     "LiveCache Interface used by DB-Procedures"
 * @release:     7.2.0.0
 * @see:         ""
 *
 * @copyright:   "Copyright by SAP AG, 1999" 


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 */

#if defined(GCO01_ONLY_MEMBER_FUNCS)
#undef  GCO01_H
#undef  VIRTUAL
#undef  PURE_VIRTUAL
#endif
#ifndef GCO01_H
#define GCO01_H

#include "geo00.h"
#include "ggg00.h"
//#include "ggg91.h"

#if defined(GCO01_ONLY_MEMBER_FUNCS)
#define VIRTUAL 
#define PURE_VIRTUAL
#else
#define VIRTUAL       virtual
#define PURE_VIRTUAL  =0

externC const IID IID_IliveCacheSink;

externC {

  struct IliveCacheSink : public IUnknown
    {
    public:
#endif
        VIRTUAL HRESULT STDMETHODCALLTYPE ABAPRead( 
            /* [out][in] */ OmsTypeInt4  *TabHandle,
            /* [out][in] */ tsp00_Int4   *count,
            /* [in] */ tsp00_Int4  *lpbuflen,
            /* [size_is][out] */ unsigned char  *lpv,
            /* [out] */ short  *lpb_error) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE ABAPWrite( 
            /* [in] */ OmsTypeInt4 *TabHandle,
            /* [in] */ tsp00_Int4 row_size,
            /* [in] */ tsp00_Int4 row_count,
            /* [in] */ unsigned char  *lpv,
            /* [out] */ short  *lpb_error) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE AllocPacket( 
            /* [out] */ unsigned char  * *p,
            /* [out] */ tsp00_Int4  *sz) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE BasisError( 
            /* [in] */ short b_error) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE CloseVersion( 
            /* [in] */ OmsVersionId  *lpVersionId,
            /* [in] */ tsp00_Int4 lastKey,
            /* [out] */ short  *lpb_error) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE ConsoleMsg( 
            /* [in] */ tsp00_Int4 buflen,
            /* [size_is][in] */ unsigned char  *lpv) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE CreateOpenVersion( 
            /* [in] */ unsigned char  *lpConsistentView,
            /* [in] */ OmsVersionId  *lpVersionId,
            /* [in] */ boolean createVersion,
            /* [out] */ unsigned char  *lpVersionContext,
            /* [out] */ short  *lpb_error) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE DeleteAllObj( 
            /* [in] */ unsigned char  *lpContainerId,
            /* [in] */ unsigned char  *lpVersionFile,
            /* [out] */ int  *lpDeleted,
            /* [out] */ short  *lpb_error) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE DeleteObj( 
            /* [in] */ unsigned char  *lpConsistentView,
            /* [in] */ unsigned char  *lpContainerId,
            /* [in] */ unsigned char  *lpVersionContext,
            /* [size_is][in] */ unsigned char  *lpoid,
            /* [size_is][in] */ unsigned char  *lpobjseq,
            /* [in] */ unsigned char  *lpcoldesc,
            /* [out] */ short  *lpb_error) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE EndConsistentView( 
            /* [out][in] */ unsigned char  *lpConsistentView,
            /* [out] */ short  *lpb_error) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE EnterCriticalSection(/*in */ short regionId) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE GetErrorMsg( 
            /* [in] */ short errorno,
            /* [out][in] */ tsp00_Int4  *lpbuflen,
            /* [size_is][out][in] */ unsigned char  *lpv) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE GetObj( 
            /* [out][in] */ tgg91_TransNo&                 consistentView,
            /* [out][in] */ tgg00_ContainerId&             containerId, 
            /*      [in] */ const tgg00_OmsVersionContext* pVersionContext,
            /*      [in] */ const OmsTypeOid&              oid,
            /*      [in] */ bool                           doLock,
            /* [out]     */ tgg91_PageRef&                 objSeq,
            /*      [in] */ tsp00_Int4                     bodySize,
            /* [out]     */ unsigned char*                 pBody,
            /* [out]     */ tsp00_Int4&                    bodyLen,
            /* [out]     */ tsp00_Int4&                    histLogCount,
            /* [out]     */ tgg00_BasisError&              e) PURE_VIRTUAL;
      
        VIRTUAL HRESULT STDMETHODCALLTYPE GetObjWithKey( 
            /* [out][in] */ unsigned char  *lpConsistentView,
            /* [in] */ unsigned char  *lptree,
            /* [in] */ unsigned char  *lpVersionContext,
            /* [in] */ tsp00_Int4 keylen,
            /* [in] */ unsigned char  *lpkey,
            /* [in] */ boolean doLock,
            /* [in] */ tsp00_Int4 bodysize,
            /* [size_is][out] */ unsigned char  *lpbody,
            /* [out] */ OmsTypeOid  *lpoid,
            /* [size_is][out] */ unsigned char  *lpobjseq,
            /* [out] */ int  *hist_log_count,
            /* [out] */ short  *lpb_error) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE HResultError( 
            /* [in] */ HRESULT h_error) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE IsDebugEnvironment( void) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE IsLocked( 
            /* [in] */ unsigned char  *lpContainerId,
            /* [in] */ OmsTypeOid  *lpoid,
            /* [out] */ short  *lpb_error) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE LeaveCriticalSection(/*in */ short regionId) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE LockObj( 
            /* [in] */ unsigned char  *lpConsistentView,
            /* [in] */ unsigned char  *lpContainerId,
            /* [in] */ OmsTypeOid  *lpoid,
            /* [size_is][in] */ unsigned char  *lpobjseq,
            /* [out] */ short  *lpb_error) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE MultiPurpose( 
            /* [in] */ unsigned char MessType,
            /* [in] */ unsigned char MessType2,
            /* [out][in] */ tsp00_Int4  *lpbuflen,
            /* [out][in] */ unsigned char  *lpv,
            /* [out] */ short  *lpb_error) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE NewConsistentView( 
            /* [out][in] */ unsigned char  *lpConsistentView,
            /* [out] */ short  *lpb_error) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE NewObj( 
            /* [in] */ unsigned char  *lpConsistentView,
            /* [in] */ unsigned char  *lpContainerId,
            /* [in] */ unsigned char  *lpVersionContext,
            /* [in] */ tsp00_Int4 KeyPos,
            /* [in] */ tsp00_Int4 KeyColSize,
            /* [in] */ unsigned char  *KeyColumn,
            /* [out] */ OmsTypeOid  *lpoid,
            /* [size_is][out] */ unsigned char  *lpobjseq,
            /* [out] */ short  *lpb_error) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE NextObj( 
            /* [out][in] */ unsigned char  *lpConsistentView,
            /* [in] */ unsigned char  *lpContainerId,
            /* [in] */ unsigned char  *lpVersionContext,
            /* [out][in] */ OmsTypeOid  *lpoid,
            /* [size_is][out] */ unsigned char  *lpobjseq,
            /* [in] */ tsp00_Int4 bodysize,
            /* [size_is][out] */ unsigned char  *lpbody,
            /* [out] */ short  *lpb_error) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE RegisterClass(
            /* [in] */ tsp00_Int4          SchemaId,
            /* [in] */ GUID                *lpGUID,
            /* [in] */ tsp00_KnlIdentifier *lpClassName,
            /* [in] */ tsp00_Int4          size,
            /* [in] */ OmsContainerNo      ContainerNo,
            /* [out][in] */ unsigned char  *lpcoldesc,
            /* [out] */ unsigned char      *lpContainerId,
            /* [out] */ short  *lpb_error) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE Rollback( 
            /* [in] */ tsp00_Int4 buflen,
            /* [size_is][in] */ unsigned char  *lpv,
            /* [out] */ short  *lpb_error) PURE_VIRTUAL;
             
        VIRTUAL HRESULT STDMETHODCALLTYPE SetError( 
            /* [in] */ short errorno,
            /* [in] */ tsp00_Int4 buflen,
            /* [in] */ unsigned char *lpv,
            /* [in] */ tsp00_CodeType codeType ) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE SQLReceive( 
            /* [out] */ unsigned char  * *lpv) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE SQLRequest( 
            /* [in] */ tsp00_Int4 buflen,
            /* [size_is][in] */ unsigned char  *lpv,
            /* [out] */ short  *lp_error) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE SubtransCommit( 
            /* [out] */ short  *lpb_error) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE SubtransRollback( 
            /* [out] */ short  *lpb_error) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE SubtransStart( 
            /* [out] */ short  *lpb_error) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE UnlockObj( 
            /* [in] */ unsigned char  *lpContainerId,
            /* [in] */ OmsTypeOid  *lpoid,
            /* [out] */ short  *lpb_error) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE UpdateObj( 
            /* [in] */ unsigned char  *lpConsistentView,
            /* [in] */ unsigned char  *lpContainerId,
            /* [in] */ unsigned char  *lpVersionContext,
            /* [size_is][in] */ unsigned char  *lpoid,
            /* [size_is][out][in] */ unsigned char  *lpobjseq,
            /* [in] */ tsp00_Int4 bodysize,
            /* [size_is][in] */ unsigned char  *lpbody,
            /* [in] */ unsigned char  *lpcoldesc,
            /* [out] */ short  *lpb_error) PURE_VIRTUAL;
        
        VIRTUAL HRESULT STDMETHODCALLTYPE Vtrace( 
            /* [in] */ tsp00_Int4 buflen,
            /* [size_is][in] */ unsigned char  *lpv) PURE_VIRTUAL;
        VIRTUAL HRESULT STDMETHODCALLTYPE GetDefaultContext(/* [out] */ void** p) PURE_VIRTUAL;
        VIRTUAL HRESULT STDMETHODCALLTYPE SetDefaultContext(/* [in] */ void* p) PURE_VIRTUAL;
        VIRTUAL HRESULT STDMETHODCALLTYPE OpenFile(/* [in] */ char  *lpFileName,
          /* [in]  */ tsp00_Uint4 FileSize,
          /* [in]  */ boolean     doAppend, 
          /* [out] */ tsp00_Int4  *FileHandle,
          /* [out] */ short  *lpb_error) PURE_VIRTUAL; 
        VIRTUAL HRESULT STDMETHODCALLTYPE WriteFile(/* [in] */ tsp00_Int4 FileHandle,
          /* [in]  */ char   *msg,
          /* [out] */ short  *lpb_error) PURE_VIRTUAL; 
        VIRTUAL HRESULT STDMETHODCALLTYPE CloseFile(/* [in] */ tsp00_Int4 FileHandle,
          /* [out] */ short  *lpb_error) PURE_VIRTUAL;
        VIRTUAL HRESULT STDMETHODCALLTYPE SyncFile(/* [in] */ char  *lpFileName,
          /* [in] */ tsp00_Int4 FileHandle,
          /* [out] */ short  *lpb_error) PURE_VIRTUAL;
        VIRTUAL HRESULT STDMETHODCALLTYPE GetDateTime(/* [out] */ tsp00_Date* Date,
          /* [out] */ tsp00_Time* Time) PURE_VIRTUAL;
        VIRTUAL HRESULT STDMETHODCALLTYPE GetClock(/* [out] */ tsp00_Int4* Sec,
          /* [out] */ tsp00_Int4* MSec) PURE_VIRTUAL;
        VIRTUAL HRESULT STDMETHODCALLTYPE DropVersion( /* in */ unsigned char  *lpVersionContext,
            /* [out] */ short  *lpb_error) PURE_VIRTUAL;
        VIRTUAL HRESULT STDMETHODCALLTYPE CreateSchema(/* [in] */ tsp00_KnlIdentifier *lpSchemaName,
            /* [out] */ tsp00_Int4 *lpSchemaId,
            /* [out] */ short      *lpb_error) PURE_VIRTUAL;
        VIRTUAL HRESULT STDMETHODCALLTYPE ExistsSchema(/* [in] */ tsp00_KnlIdentifier *lpSchemaName,
            /* [out] */ tsp00_Int4 *lpSchemaId,
            /* [out] */ short      *lpb_error) PURE_VIRTUAL;
        VIRTUAL HRESULT STDMETHODCALLTYPE DropSchema(/* [in] */ tsp00_Int4 SchemaId,
            /* [out] */ short      *lpb_error) PURE_VIRTUAL;
        VIRTUAL HRESULT STDMETHODCALLTYPE GetSchemaName(/* [in] */ tsp00_Int4 SchemaId,
            /* [out] */ tsp00_KnlIdentifier *lpSchemaName,
            /* [out] */ short      *lpb_error) PURE_VIRTUAL;
        VIRTUAL HRESULT STDMETHODCALLTYPE ExistsFile(/* [in] */ char  *lpFileName,
          /* [out] */ bool    &existsFile,
          /* [out] */ bool    &isDirectory,
          /* [out] */ short   &error) PURE_VIRTUAL; 
        VIRTUAL HRESULT STDMETHODCALLTYPE DropFile(/* [in] */ char  *lpFileName,
          /* [out] */ short   &error) PURE_VIRTUAL; 
        VIRTUAL HRESULT STDMETHODCALLTYPE GetWorkingDirectory(/* [out] */ tsp00_Pathc& pathName,
          /* [out] */ short  &error) PURE_VIRTUAL; 
        VIRTUAL HRESULT STDMETHODCALLTYPE GetDbrootSapPath(/* [out] */ tsp00_Pathc& pathName,
          /* [out] */ short  &error) PURE_VIRTUAL; 
        VIRTUAL HRESULT STDMETHODCALLTYPE LoadLibraryEx( /* in */ const char* path, 
          /* in  */ bool        loadGlobal,
          /* out */ HANDLE      *hd,
          /* out */ tsp_errtext &errtext,
          /* in  */ int         errtext_size) PURE_VIRTUAL;
        VIRTUAL HRESULT STDMETHODCALLTYPE FreeLibrary( /* in */ HANDLE hd,
          /* out */ bool        &ok,
          /* out */ tsp_errtext &errtext,
          /* in  */ int         errtext_size) PURE_VIRTUAL;
        VIRTUAL HRESULT STDMETHODCALLTYPE DropContainer(/* [in] */ OmsSchemaHandle schema,
            /* [in] */ const GUID&        guid,
            /* [in] */ OmsContainerNo     cno,
            /* [out] */ tgg00_BasisError& error) PURE_VIRTUAL;
        VIRTUAL HRESULT STDMETHODCALLTYPE GetContainerInfo(/* [in] */ tsp00_Uint4 containerHandle,
            /* [out] */ GUID& guid,
            /* [out] */ OmsSchemaHandle& schema,
            /* [out] */ OmsContainerNo&  cno,
            /* [out] */ short& error) PURE_VIRTUAL;
        VIRTUAL HRESULT STDMETHODCALLTYPE IsObjHistoryUsed( 
            /* [out][in] */ tgg00_ContainerId& containerId, 
            /*      [in] */ const OmsTypeOid&  oid,
            /* [out]     */ bool&              isUsed,
            /* [out]     */ tgg00_BasisError&  e) PURE_VIRTUAL;

#if !defined(GCO01_ONLY_MEMBER_FUNCS)
    };
}
#endif

#endif  /* GCO01_H */
