/*!================================================================
 module:        gco02.h
 
 responsible:   ThomasA
 
 special area:  OMS
 
 description:   LibOms Interface
 
 see:
 
 copyright:     SAP AG 1998 


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 ===================================================================*/



#if !defined (__GCO02_H)
#define __GCO02_H

#include "gsp00.h"
#include "ggg92.h"

struct IliveCacheSink;

typedef IliveCacheSink* (*GetSinkPtr_co02)(void);

class LibOmsInterface_cco02 {
public :
  virtual void GetMemoryUsageInfo(
    IliveCacheSink*     lcSink,
    tsp00_8ByteCounter& totalHeapUsed,
    tsp00_8ByteCounter& freeChunkSpace) = 0;

  virtual tsp00_Bool GetFirstVersion(
    IliveCacheSink*     lcSink,
    tsp00_C24&          versionId,
    tsp00_Date&         createDate,
    tsp00_Time&         createTime,
    tsp00_Date&         openDate,
    tsp00_Time&         openTime,
    tgg91_TransNo&      consistentView,
    tsp00_Bool&         isMarked,
    tsp00_Bool&         isOpen,
    tsp00_Bool&         isUnloaded,
    tsp00_8ByteCounter& heapUsage,
    tgg00_FileId&       versionTree,
	  tgg00_FileId&       versionInvTree) = 0;

  virtual tsp00_Bool GetMonitorInfo(
    IliveCacheSink*       lcSink,
    tsp00_Addr&           handle,
    tsp00_C16&            iid,
    tsp00_Int4&           dispid,
    tgg00_COMMonitorInfo& monInfo) = 0;

  virtual tsp00_Bool GetNextVersion(
    IliveCacheSink*     lcSink,
    tsp00_C24&          versionId,
    tsp00_Date&         createDate,
    tsp00_Time&         createTime,
    tsp00_Date&         openDate,
    tsp00_Time&         openTime,
    tgg91_TransNo&      consistentView,
    tsp00_Bool&         isMarked,
    tsp00_Bool&         isOpen,
    tsp00_Bool&         isUnloaded,
    tsp00_8ByteCounter& heapUsage,
    tgg00_FileId&       versionTree,
	  tgg00_FileId&       versionInvTree) = 0;

  virtual void AdviseGetSinkFunc(GetSinkPtr_co02 func) = 0;

  virtual void AdviseHeapLimit(tsp00_Uint4 limit) = 0;

  virtual void AdviseOmsRegionCnt(tsp00_Int4) = 0;

  virtual void InitLibOms(IliveCacheSink* lcSink) = 0;

  virtual tgg00_BasisError Shutdown(IliveCacheSink* lcSink) = 0;

  virtual void LockVersionDirectory(IliveCacheSink* lcSink) = 0;

  virtual void ResetMonitor(IliveCacheSink* lcSink) = 0;
  
  virtual void UnLockVersionDirectory(IliveCacheSink* lcSink) = 0;

  virtual bool SetTraceLevel(tsp00_KnlIdentifier& lvl, bool enable) = 0;
};

typedef LibOmsInterface_cco02* (*ProcGetInterfacePtr_co02)(void);

typedef IliveCacheSink* (*GetSinkPtr_co02)(void);

# endif