/*!================================================================
 module:        gco10.h
 
 responsible:   ThomasA
 
 special area:  OMS
 
 description:   C++ Persistent Object Interface
 
 see:
 
 copyright:     SAP AG 1998 


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 ===================================================================*/



#if !defined (__GCO10_H)
#define __GCO10_H

// Assertions : compile with option /-D_ASSERT_OMS

#ifdef _ASSERT_OMS
#define _OMSASSERT(session,expr) if (!(expr)) { session->ThrowDBError(-28999, #expr);}
#else
#define _OMSASSERT(h,expr) 
#endif

#ifndef _DEBUG

/****************************************************************************
*
* Debug OFF
* Debug OFF
* Debug OFF
*
***************************************************************************/

#define _TRACE_METHOD_ENTRY(h,methodname)
#define _TRACE_OID(h,oid)
#else

/****************************************************************************
*
* Debug ON
* Debug ON
* Debug ON
*
***************************************************************************/

#define _TRACE_METHOD_ENTRY(h,methodname) (h)->omsTracef(methodname);

#define _TRACE_OID(h,oid) (h)->omsTraceOid(oid);

#endif

#include "gco001.h"
#include "gco02.h"
#include "hco13.h"
#include "gco14.h"
#include "hco15.h"
#include "hco16.h"
#include "hco18.h"
#include "gco90.h"
#include "ggg92.h"
#include "ggg00.h"
#include "hco10.h"
#include "hco30-k.h"


#define FOR_UPD_CO10       true

const GUID VAR_OBJ_GUID_GCO10 = { 0x8fb0963d, 0xe5aa, 0x11d1, { 0xa9, 0x5c, 0x0, 0xa0, 0xc9, 0x43, 0x7, 0x31 } };

externCpp
void* omsMalloc(size_t size);

/*----------------------------------------------------------------------*/

externCpp
void omsFree(void* p);

/*----------------------------------------------------------------------------*/

class AFX_EXT_CLASS OmsDuplicateKey : public DbpError
{
public:
  OmsDuplicateKey(long e, const OmsObjectId& oid); 
};

class AFX_EXT_CLASS OmsLockTimeout : public DbpError
{
public:
  OmsLockTimeout(const OmsObjectId& oid);
};

class AFX_EXT_CLASS OmsOutOfDate : public DbpError
{
public:
  OmsOutOfDate();
  OmsOutOfDate(const OmsObjectId& oid);
};

class OmsContext_co12;
class OmsSession_co12;

class OmsCallbackInterface {
  friend class OmsSession_co12;
  friend class OmsHandle;
public :
  enum Event {
      BeforeCommit,
      AfterCommit,
      AfterRollback,
      ByApplication
  };
private :
  virtual ~OmsCallbackInterface() {};
  virtual void omsValidate (OmsHandle&, Event, bool) = 0;
  virtual void omsInvalidate (OmsHandle&, Event, bool) = 0;
  virtual void omsDestroySelf() = 0;
};

class AFX_EXT_CLASS OmsHandle : public DbpBase {
  
  friend class  OmsAbstractObject;
  friend class  OmsCriticalSection;
  friend class  OmsLockObject;
  friend class  OmsStream;
  friend class  OmsObjByClsIterBase;
  friend class  cco13_BeforeImages;
  friend void   co10_SetCurrMonitorMethod(OmsHandle& h, tgg00_COMMonitorInfo* p);
public :
    
  OmsHandle  (OmsTypeInt4 sz = 1007);
  ~OmsHandle ();
  HRESULT     omsAttachDefaultContext();
  OmsObjectId omsCastOid(const GUID guid, const OmsObjectId& oid);
  bool        omsCheckOid    (const OmsObjectId&, const GUID& guid);
  bool        omsCheckOid    (const OmsVarOid& );
  void        omsCleanUpAfterCrash();
  OmsSchemaHandle omsCreateSchema(const char* SchemaName);
  const void* omsDeRef(const OmsVarOid&);
  void        omsDropSchema(OmsSchemaHandle);
  void        omsDropContainer(const GUID&, OmsSchemaHandle, OmsContainerNo);
  void        omsOidInfo (const OmsObjectId&, GUID&, OmsSchemaHandle&, char*, int, char*, int, OmsContainerNo&); 
  int         omsCurrentSubtransLevel() const;
  void        omsLoadVarObject (const OmsVarOid& oid, unsigned long size, void* buf);
  void        omsLock( const OmsObjectId& );
  void        omsLock( const OmsVarOid& oid);
  void        omsLocks( const OmsObjectId*, int);
  int         omsGetTimeOut() const;
  OmsContainerNo  omsGetContainerNo (const OmsObjectId&);
  void        omsGetObj(const OmsObjectId& oid, unsigned char* cnsView, int& ObjLen, unsigned char* obj); 
  OmsSchemaHandle omsGetSchemaHandle(const char* SchemaName) const;
  OmsSchemaHandle omsGetSchemaHandle(const OmsObjectId&);
  void        omsGetSchemaName(OmsSchemaHandle,char*,int) const;
  OmsCallbackInterface* omsGetTransactionInterface (const GUID&) const;
  unsigned long omsGetVarObjSize (const OmsVarOid& oid);
  void        omsInitSession (IUnknown** s);
  void        omsNewConsistentView();
  void        omsNewConsistentView (OmsObjectId* pOids, int cntOids, short timeout);
  void        omsNewConsistentView (OmsObjectId* pOids, int cntOids, short timeout, OmsObjectId*, int&);
  void        omsReleaseConsistentView();
  void        omsPrintOid (char* buf, size_t bufSize, const OmsObjectId oid);
  const       OmsVarOid& omsNewVarObject(OmsSchemaHandle, OmsContainerNo);
  bool        omsIsLocked ( const OmsObjectId& );
  void		    omsRaiseError (short msgno, const char* errmsg);
  void        omsReadOnly (bool = true);
  void        omsRegClass (OmsSchemaHandle, const GUID& guid, const char* ClassName, size_t PersistentSize, size_t ObjectSize, const GUID* pBaseClass, OmsAbstractObject*, OmsContainerNo);
  void        omsRegClassAndKey (OmsSchemaHandle,const GUID& guid, const char* ClassName, size_t PersistentSize, size_t KeySize, size_t ObjectSize, OmsAbstractObject*, OmsContainerNo);
  void        omsRegTransactionInterface (const GUID&, OmsCallbackInterface*);
  void        omsRelease ( const OmsObjectId& );
  void        omsRelease ( const OmsVarOid& oid );
  void        omsReleaseAll();
  void        omsReleaseUnconditional( const OmsObjectId& ); // not exported in livecache.h
  void        omsReleaseVarObject(const OmsVarOid& oid);
  HRESULT     omsReleaseDefaultContext();
  void        omsSetTimeOut( short );
  void        omsStoreVarObject ( const OmsVarOid& oid, const void*, unsigned long objLength);
  short       omsCommit   (bool);
  short       omsCommitHandling(int&);
  void        omsRollback (bool);
  void 		    omsDelete ( const OmsObjectId& );
  void        omsDelete ( const OmsVarOid& );
  void        omsDelete ( OmsAbstractObject* );
  void		    omsDeleteAll(const GUID& guid, OmsSchemaHandle, OmsContainerNo);
  int         omsStartSubtrans();
  void        omsCommitSubtrans(int = -1);
  void        omsRollbackSubtrans( int = - 1);
  unsigned long  omsGetVarObjectSize (const OmsVarOid& );
  void           omsTerminate();
  void           omsTrace ( const char* tracemsg, long TraceMsgLen);
  void           omsTracef( const char*fmt, ... );
  void           omsTraceOid (const OmsObjectId& oid);
  void           omsTransRollback();
  void           omsUnlock( const OmsObjectId& ); 
  void           omsExceptionHandler(DbpError&);					
  void	         omsDump();  
  const OmsVersionId* omsCurrentVersionId () const;
  void           omsCreateVersion    ( const OmsVersionId& );
  void           omsMarkVersion      ( const OmsVersionId& );
  void           omsOpenVersion      ( const OmsVersionId& );
  void           omsCloseVersion     ( const OmsVersionId& );
  void           omsDropVersion      ( const OmsVersionId& );
  void           omsForceDropVersion ( const OmsVersionId& );
  bool           omsExistsVersion    ( const OmsVersionId& );
  void           omsRenameVersion    ( const OmsVersionId&, const OmsVersionId&);
  void           omsValidate();
  void           omsGetMonitorInfo (tgg00_COMMonitorTransferBuf* lpMonInfo);
  void           omsRegMonitor     (void* pMon); 
  void           omsSetCurrMonitorMethod(tgg00_COMMonitorInfo* p);
 
  void           TraceCommitPhase(int);
  protected :
    OmsAbstractObject* omsDeRef      (const OmsObjectId& oid, const GUID& guid, bool forUpd, bool do_lock);
    OmsAbstractObject* omsDeRefKey (unsigned char* key, const GUID& guid, bool forUpd, bool do_lock, OmsSchemaHandle, OmsContainerNo);
  private :
    void SessionEnd();
    IliveCacheSink**      m_session;
    co90_Monitor*         m_monitor;
  public :
    OmsSession_co12*      m_pSession;
    OmsObjByClsIterBase omsAllOids( const GUID&, OmsSchemaHandle, OmsContainerNo );
    OmsAbstractObject* omsNewObject ( const GUID&, OmsSchemaHandle, OmsContainerNo );
    OmsAbstractObject* omsNewRegistryObject (size_t sz, const GUID guid);
    OmsAbstractObject* omsNewKeyedObject( const GUID& guid, const unsigned char* key, OmsSchemaHandle, OmsContainerNo );
    tco13_ClsIdEntry*  omsAddContainerInfo (ContainerInfo_co18* ContainerInfo);
};

class AFX_EXT_CLASS OmsCriticalSection {
public :
  OmsCriticalSection(OmsHandle*,int);
  void omsEnter();
  void omsLeave();
  ~OmsCriticalSection();
private :
  OmsSession_co12* m_session;
  int              m_regionId;
  bool             m_inSection;
};

class AFX_EXT_CLASS OmsLockObject {
  friend class OmsHandle;
  friend class OmsSession_co12;
public :
   OmsLockObject(OmsHandle&);
   ~OmsLockObject();                       
   void omsEnable ();
   void omsEnable (const OmsLockHandle&);

   const OmsLockHandle& omsGetLockHandle() const;
   void  omsLockShare();
   void  omsLockExclusive();
   void  omsUnlock();
   bool  omsIsLocked() const;
   bool  omsShareLocked() const;
   bool  omsExclusiveLocked() const;
   void  InvalidateHandle();
   OmsLockObject* Next() const;
private :
  enum OmsLockMode{ NO_LOCK, LOCK_SHARE, LOCK_EXCLUSIVE };
  OmsHandle*     m_omsHandle;
  OmsLockObject* m_next;
  OmsLockHandle  m_lockHandle;
  OmsLockMode    m_mode;
};

class AFX_EXT_CLASS OmsObjByClsIterBase {
  friend class OmsHandle;
  friend class OmsAbstractObject;
public :
  ~OmsObjByClsIterBase();
protected:
  operator bool () const;
  const OmsObjectId& operator() () const;
  OmsAbstractObject* omsDeRef(bool,bool) const;
  void omsNextNewObjectInVersion(OmsObjectContainerPtr);
  void operator++ ();
#if !defined(WIN32)
public :
#endif
  OmsObjByClsIterBase(const OmsObjByClsIterBase&);
  OmsObjByClsIterBase(OmsHandle&);
private:
  OmsSession_co12*       m_session;
  tco13_ClsIdEntry*      m_containerInfo;
  OmsObjectContainerPtr  m_curr;
  OmsObjectContainerPtr  m_next;
  bool                   m_iterNewObjectsInVersion;
  tsp00_Int4             m_headIndex;
};

class AFX_EXT_CLASS OmsAbstractObject {
  
  friend class OmsObjectContainer;
  friend class OmsHandle;
  friend class cco13_OidHash;
  friend class OmsObjByClsIterBase;
protected :
  
public :
  virtual void       omsCleanUp (OmsHandle&);
  virtual void       omsDump    (OmsHandle& h, const char* flags) const;
  void               omsDelete          (OmsHandle& h);
  void               omsFlush           (OmsHandle& h);
  OmsSchemaHandle    omsGetSchemaHandle(OmsHandle&) const;
  bool               omsIsLocked        (OmsHandle& h) const;
  void               omsLock            (OmsHandle& h);
  void               omsRelease         (OmsHandle& h);
  void               omsStore           (OmsHandle& h);
  virtual void       omsVerify  (OmsHandle& h) const;
  const OmsObjectId& omsGetOid() const;
protected :
  OmsAbstractObject ();
  const OmsObjectId& omsOid() const;
  OmsAbstractObject* omsForUpdPtr(OmsHandle& h, bool doLock) const;
  bool  omsHistoryInUse(OmsHandle&) const;
  const OmsAbstractObject* omsDeRefNext(OmsHandle& Handle,bool do_lock) const;
  void omsKey(OmsHandle& Handle, unsigned char* key) const;
  static void omsDeleteAll(OmsHandle& h, const GUID& guid, OmsSchemaHandle, OmsContainerNo);
  static OmsAbstractObject* omsNewObject (OmsHandle& Handle, const GUID& guid, OmsSchemaHandle, OmsContainerNo );
  static OmsAbstractObject* omsNewRegistryObject(size_t sz, OmsHandle& Handle, const GUID& guid);
  static OmsAbstractObject* omsNewKeyedObject (OmsHandle& Handle, const GUID& guid, const unsigned char* key, OmsSchemaHandle, OmsContainerNo ); 
  static OmsObjByClsIterBase omsAllOids( OmsHandle& h, const GUID& guid, OmsSchemaHandle, OmsContainerNo);
  static void omsRegClass (OmsHandle& Handle, const GUID& Guid, const char* ClassName, size_t PersistentSize, size_t ObjectSize, const GUID* pBaseClass, OmsAbstractObject* Obj, OmsSchemaHandle, OmsContainerNo);
  static void omsRegClassAndKey (OmsHandle&, const GUID&, const char* ClassName, size_t PersistentSize, size_t KeySize, size_t ObjectSize, OmsAbstractObject* Obj, OmsSchemaHandle, OmsContainerNo);
  static const OmsAbstractObject* omsDeRef (const OmsObjectId& oid, OmsHandle& Handle, const GUID& guid);
  static OmsAbstractObject* omsDeRefForUpd (const OmsObjectId& oid, OmsHandle& Handle, const GUID& guid, bool do_lock);
  static const OmsAbstractObject* omsDeRefKey (unsigned char* key, OmsHandle& Handle, const GUID& guid, OmsSchemaHandle, OmsContainerNo);
  static OmsAbstractObject* omsDeRefKeyForUpd (unsigned char* key, OmsHandle& Handle, const GUID& guid, bool do_lock, OmsSchemaHandle, OmsContainerNo);
private :
  OmsAbstractObject(const OmsAbstractObject&);              // no copy
  OmsAbstractObject& operator= (const OmsAbstractObject&r); // no assignment
  
};

class OmsObjectContainer{ 
  friend class OmsHandle;
  friend class cco13_OidHash;
public :
  enum ct_state {
    CN_STORE = 1, CN_LOCKED = 2, CN_DELETE = 4, CN_NEW = 8, CN_VAROBJ = 16 };
    OmsObjectContainerPtr m_hashnext;
    OmsObjectId           m_oid;
    tgg91_PageRef         m_objseq;
private :
    unsigned char         m_state;
public :
    unsigned char         m_filler;
    tsp00_Uint4           m_beforeImages;
#if defined(USE_CONTAINER_HANDLE_CO001)
    tsp00_Uint4           m_containerHandle;
#else
    tco13_ClsIdEntryPtr   m_containerInfo;
#endif
    OmsAbstractObject     m_pobj;
        
    static tsp00_Int4 headerSize() {
      return (sizeof (OmsObjectContainer) - sizeof (OmsAbstractObject));
    }

    inline void ChainFree(OmsContext_co12* context);

    bool DeletedFlag() {
      return ((m_state & OmsObjectContainer::CN_DELETE) > 0);
    }

    bool existBeforeImage (int subtransLvl, bool inVersion) {
      if (0 == subtransLvl) {
        return true;
      }
      if ((!inVersion) && (1 == subtransLvl)) {
        return true;
      }
      return ((m_beforeImages & (1 << (subtransLvl - 1))) > 0);
    }

    inline tco13_ClsIdEntry* GetContainerInfo(OmsContext_co12* context);
    
    inline tco13_ClsIdEntry* GetContainerInfoNoCheck(OmsContext_co12* context);

    tsp00_Uint4 GetContainerHandle() const {
#if defined(USE_CONTAINER_HANDLE_CO001)
      return m_containerHandle;
#else
      return m_containerInfo->GetContainerHandle();
#endif
    }

    void InitObjContainer(tco13_GuidEntry& classInfo) {
      void** pvtptr  = REINTERPRET_CAST(void**, (&m_pobj));
      *pvtptr        = classInfo.GetVirtualTablePtr();
      this->InitState(&classInfo);
      m_beforeImages = 0;
      m_hashnext     = NULL;
#if defined(USE_CONTAINER_HANDLE_CO001)
      m_containerHandle = 0;
#else
      m_containerInfo   = NULL;
#endif
    }

    void InitObjContainer(tco13_ClsIdEntry& clsinfo) { 
      void** pvtptr  = REINTERPRET_CAST(void**, (&m_pobj));
      *pvtptr        = clsinfo.GetVirtualTablePtr();
      this->InitState(&clsinfo);
      m_beforeImages = 0;
      m_hashnext     = NULL;
#if defined(USE_CONTAINER_HANDLE_CO001)
      m_containerHandle = 0; 
#else
      m_containerInfo   = NULL;
#endif
    }

    void InitState(tco13_GuidEntry* classInfo) {
      m_state = 0;
       if ((NULL != classInfo) && (classInfo->IsVarObject())) {
        this->MarkVarObj();
      }
    }

    void InitState(tco13_ClsIdEntry* clsinfo) { // TODO ?
      m_state = 0;
       if ((NULL != clsinfo) && (clsinfo->IsVarObject())) {
        this->MarkVarObj();
      }
    }

    bool LockedFlag() {
      return ((m_state & OmsObjectContainer::CN_LOCKED) > 0);
    }

    void MarkDeleted() {
      m_state &= (OmsObjectContainer::CN_STORE ^ 0xFF);
      m_state |= (OmsObjectContainer::CN_DELETE);
    }
   
    void MarkLocked() {
      m_state |= OmsObjectContainer::CN_LOCKED;
    }

    void MarkNew() {
      m_state |= OmsObjectContainer::CN_NEW;
    }
    
    void MarkStored() {
      m_state |= OmsObjectContainer::CN_STORE;
    }

    void MarkVarObj() {
      m_state |= OmsObjectContainer::CN_VAROBJ;
    }

    bool NewFlag() {
      return ((m_state & OmsObjectContainer::CN_NEW) > 0);
    }

    void RemoveStoredFlag() {
      m_state &= (OmsObjectContainer::CN_STORE ^ 0xFF);
    }

    void setBeforeImage (int subtransLvl) {
      m_beforeImages |= (1 << (subtransLvl-1)); 
    }

    static void store (OmsHandle& h, OmsAbstractObject* o) {
      OmsObjectContainerPtr p = containerPtr(o);
      p->m_state |= OmsObjectContainer::CN_STORE;
    }

#if defined(USE_CONTAINER_HANDLE_CO001)
    void SetContainerHandle(tsp00_Uint4 ContainerHandle) {
      m_containerHandle = ContainerHandle;

    }
#else
    void SetContainerInfo(tco13_ClsIdEntryPtr ContainerInfo) {
      m_containerInfo = ContainerInfo;
    }
#endif

    bool StoredFlag() {
      return ((m_state & OmsObjectContainer::CN_STORE) > 0);
    }

    bool VarObjFlag() {
      return ((m_state & OmsObjectContainer::CN_VAROBJ) > 0);
    }

    static OmsObjectContainerPtr containerPtr (OmsAbstractObject* pobj) {
      return (OmsObjectContainerPtr) (((unsigned char*) pobj) - ((sizeof(OmsObjectContainer)) - sizeof (OmsAbstractObject))); 
    }
   
    void Copy (OmsContext_co12* c, OmsObjectContainerPtr beforeImage, int ObjectSize);

public :
};

#include "hco12.h"

inline void OmsObjectContainer::ChainFree(OmsContext_co12* context) {
#if defined(USE_CONTAINER_HANDLE_CO001)
  tco13_ClsIdEntry* containerInfo = context->GetContainerInfo(m_containerHandle);
#else
  tco13_ClsIdEntry* containerInfo = m_containerInfo;
#endif
  OmsObjectContainerPtr p = REINTERPRET_CAST(OmsObjectContainerPtr, this);
  containerInfo->chainFree(*context, p);
}

inline tco13_ClsIdEntry* OmsObjectContainer::GetContainerInfo(OmsContext_co12* context) {
#if defined(USE_CONTAINER_HANDLE_CO001)
  return context->GetContainerInfo(m_containerHandle);
#else
  context->CheckIfContainerNotDropped(m_containerInfo);
  return m_containerInfo;
#endif
}

inline tco13_ClsIdEntry* OmsObjectContainer::GetContainerInfoNoCheck(OmsContext_co12* context) {
#if defined(USE_CONTAINER_HANDLE_CO001)
  return context->GetContainerInfoNoCheck(m_containerHandle);
#else
  return m_containerInfo;
#endif
}

class AFX_EXT_CLASS OmsStream {
protected :
  void Init(OmsHandle* h, OmsTypeABAPTabHandle& sh, long rsz, long sz, bool istream);

  OmsStream(OmsHandle* h, OmsTypeABAPTabHandle& sh, long rsz, long sz, bool istream = true);
  OmsStream(OmsHandle* h, OmsTypeStreamHandle& sh, long rsz, long sz, bool istream = true);
  ~OmsStream();
  void           omsClose();
  void           omsFlush();
  unsigned char* omsNxt();
  void           omsSwapABAPRow (unsigned char* p);
  void           omsSwapStream  (unsigned char*); 
  void           omsWrite       (unsigned char*);
private :
  OmsHandle*             m_handle;
  OmsTypeABAPTabHandle*  m_stream_handle;
  tsp00_Int4             m_row_curr;
  tsp00_Int4             m_row_count;
  tsp00_Int4             m_row_size;
  tsp00_Int4             m_buf_size;
  bool                   m_eot;
  bool                   m_is_istream;
  unsigned char*         m_row_buf;
};

class AFX_EXT_CLASS OmsTimeStamp
{
public:
  OmsTimeStamp(const OmsTypeABAPTimeStamp& timeStamp);
  
  typedef struct _TimeStamp
  {
    unsigned char nYear1   : 4;
    unsigned char nPad     : 4;
    
    unsigned char nYear3   : 4;
    unsigned char nYear2   : 4;
    
    unsigned char nMonth1  : 4;
    unsigned char nYear4   : 4;
    
    unsigned char nDay1    : 4;
    unsigned char nMonth2  : 4;
    
    unsigned char nHour1   : 4;
    unsigned char nDay2    : 4;
    
    unsigned char nMinute1 : 4;
    unsigned char nHour2   : 4;
    
    unsigned char nSecond1 : 4;
    unsigned char nMinute2 : 4;

    unsigned char nSign    : 4;
    unsigned char nSecond2 : 4;
  } TimeStamp;

  // Auslesen der Daten:
  const OmsTypeABAPTimeStamp& omsGetTimeStamp() const;
  tsp00_Uint4 omsSecsSince01011970() { return m_DateTime.getDTValue(); }  
protected:
  // Konvertierung der Daten:
  void ToTimeStamp();
  void ToDateTime();

  // Schnelle Multiplikations- und Divisionsfunktionen
  inline int mult10(int n)
  {
    return (n << 3) + (n << 1);
  }

  inline int mult100(int n)
  {
    return (n << 6) + (n << 5) + (n << 2);
  }

  inline int mult1000(int n)
  {
    register int a;
    a = mult100(n);
    return mult10(a);
  }

  inline void div10(int n, int& n10, int& n1)
  {
    n1  = n;
    n10 = 0;

    while (n1 >= 10)
    {
      n1 -= 10;
      n10++;
    }
  }

  inline void div1000(int n, int& n1000, int& n100, int& n10, int& n1)
  {
    n100  = n;
    n1000 = 0;

    while (n100 >= 1000)
    {
      n100 -= 1000;
      n1000++;
    }

    n10  = n100;
    n100 = 0;
    
    while (n10 >= 100)
    {
      n10 -= 100;
      n100++;
    }

    n1  = n10;
    n10 = 0;
    
    while (n1 >= 10)
    {
      n1 -= 10;
      n10++;
    }
  }
protected:
  cco15_DateTime m_DateTime;
  union {
    TimeStamp            m_TimeStamp;
    OmsTypeABAPTimeStamp m_ABAPTimeStamp;
  };
};

/*---------------------------------------------------------------------------*/

class LibOmsInterface_cco10 : public LibOmsInterface_cco02 {
  virtual void GetMemoryUsageInfo(IliveCacheSink* lcSink,
    tsp00_8ByteCounter& totalHeapUsed,
    tsp00_8ByteCounter& freeChunkSpace);
  virtual tsp00_Bool GetFirstVersion(
    IliveCacheSink*     lcSink,
    tsp00_C24&          versionId,
    tsp00_Date&         createDate,
    tsp00_Time&         createTime,
    tsp00_Date&         openDate,
    tsp00_Time&         openTime,
    tgg91_TransNo&      consistentView,
    tsp00_Bool&         isMarked,
    tsp00_Bool&         isOpen,
    tsp00_Bool&         isUnloaded,
    tsp00_8ByteCounter& heapUsage,
    tgg00_FileId&       versionTree,
	  tgg00_FileId&       versionInvTree);
  virtual tsp00_Bool GetMonitorInfo(
    IliveCacheSink*       lcSink,
    tsp00_Addr&           handle,
    tsp00_C16&            iid,
    tsp00_Int4&           dispid,
    tgg00_COMMonitorInfo& monInfo);
  virtual tsp00_Bool GetNextVersion(
    IliveCacheSink*     lcSink,
    tsp00_C24&          versionId,
    tsp00_Date&         createDate,
    tsp00_Time&         createTime,
    tsp00_Date&         openDate,
    tsp00_Time&         openTime,
    tgg91_TransNo&      consistentView,
    tsp00_Bool&         isMarked,
    tsp00_Bool&         isOpen,
    tsp00_Bool&         isUnloaded,
    tsp00_8ByteCounter& heapUsage,
    tgg00_FileId&       versionTree,
	  tgg00_FileId&       versionInvTree);
  virtual void AdviseGetSinkFunc(GetSinkPtr_co02 func);
  virtual void AdviseHeapLimit(tsp00_Uint4 limit);
  virtual void AdviseOmsRegionCnt(tsp00_Int4 cntRgn);
  virtual void InitLibOms(IliveCacheSink* lcSink);
  virtual tgg00_BasisError Shutdown(IliveCacheSink* lcSink);
  virtual void LockVersionDirectory(IliveCacheSink* lcSink);
  virtual void ResetMonitor(IliveCacheSink* lcSink);
  virtual void UnLockVersionDirectory(IliveCacheSink* lcSink);
  virtual bool SetTraceLevel(tsp00_KnlIdentifier& lvl, bool enable);
};

/*---------------------------------------------------------------------------*/

class DummySink_cco10 : public IliveCacheSink {
  virtual HRESULT STDMETHODCALLTYPE QueryInterface( 
    /* [in] */ REFIID riid,
    /* [iid_is][out] */ void  **ppvObject) { return S_OK; }
    virtual ULONG STDMETHODCALLTYPE AddRef( void) { return S_OK; }
  virtual ULONG STDMETHODCALLTYPE Release( void) { return S_OK; }
  virtual HRESULT STDMETHODCALLTYPE ABAPRead( 
    /* [out][in] */ OmsTypeInt4  *TabHandle,
    /* [out][in] */ tsp00_Int4   *count,
    /* [in] */ tsp00_Int4  *lpbuflen,
    /* [size_is][out] */ unsigned char  *lpv,
    /* [out] */ short  *lpb_error) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE ABAPWrite( 
    /* [in] */ OmsTypeInt4 *TabHandle,
    /* [in] */ tsp00_Int4 row_size,
    /* [in] */ tsp00_Int4 row_count,
    /* [in] */ unsigned char  *lpv,
    /* [out] */ short  *lpb_error) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE AllocPacket( 
    /* [out] */ unsigned char  * *p,
    /* [out] */ tsp00_Int4  *sz) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE BasisError( 
    /* [in] */ short b_error) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE CloseVersion( 
    /* [in] */ OmsVersionId  *lpVersionId,
    /* [in] */ tsp00_Int4 lastKey,
    /* [out] */ short  *lpb_error) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE ConsoleMsg( 
    /* [in] */ tsp00_Int4 buflen,
    /* [size_is][in] */ unsigned char  *lpv) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE CreateOpenVersion( 
    /* [in] */ unsigned char  *lpConsistentView,
    /* [in] */ OmsVersionId  *lpVersionId,
    /* [in] */ boolean createVersion,
    /* [out] */ unsigned char  *lpVersionContext,
    /* [out] */ short  *lpb_error) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE DeleteAllObj( 
    /* [in] */ unsigned char  *lpContainerId,
    /* [in] */ unsigned char  *lpVersionFile,
    /* [out] */ int  *lpDeleted,
    /* [out] */ short  *lpb_error) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE DeleteObj( 
    /* [in] */ unsigned char  *lpConsistentView,
    /* [in] */ unsigned char  *lpContainerId,
    /* [in] */ unsigned char  *lpVersionContext,
    /* [size_is][in] */ unsigned char  *lpoid,
    /* [size_is][in] */ unsigned char  *lpobjseq,
    /* [in] */ unsigned char  *lpcoldesc,
    /* [out] */ short  *lpb_error) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE EndConsistentView( 
    /* [out][in] */ unsigned char  *lpConsistentView,
    /* [out] */ short  *lpb_error) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE EnterCriticalSection(/*in */ short regionId) { return S_OK; };
  
  virtual HRESULT STDMETHODCALLTYPE GetErrorMsg( 
    /* [in] */ short errorno,
    /* [out][in] */ tsp00_Int4  *lpbuflen,
    /* [size_is][out][in] */ unsigned char  *lpv) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE GetObj(
    /* [out][in] */ tgg91_TransNo&                 consistentView,
    /* [out][in] */ tgg00_ContainerId&             containerId, 
    /*      [in] */ const tgg00_OmsVersionContext* pVersionContext,
    /*      [in] */ const OmsTypeOid&              oid,
    /*      [in] */ bool                           doLock,
    /* [out]     */ tgg91_PageRef&                 objSeq,
    /*      [in] */ tsp00_Int4                     bodySize,
    /* [out]     */ unsigned char*                 pBody,
    /* [out]     */ tsp00_Int4&                    bodyLen,
    /* [out]     */ tsp00_Int4&                    histLogCount,
    /* [out]     */ tgg00_BasisError&              e) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE GetObjWithKey( 
    /* [out][in] */ unsigned char  *lpConsistentView,
    /* [in] */ unsigned char  *lptree,
    /* [in] */ unsigned char  *lpVersionContext,
    /* [in] */ tsp00_Int4 keylen,
    /* [in] */ unsigned char  *lpkey,
    /* [in] */ boolean doLock,
    /* [in] */ tsp00_Int4 bodysize,
    /* [size_is][out] */ unsigned char  *lpbody,
    /* [out] */ OmsTypeOid  *lpoid,
    /* [size_is][out] */ unsigned char  *lpobjseq,
    /* [out] */ int  *hist_log_count,
    /* [out] */ short  *lpb_error) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE HResultError( 
    /* [in] */ HRESULT h_error) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE IsDebugEnvironment( void) { return S_OK; };
  
  virtual HRESULT STDMETHODCALLTYPE IsLocked( 
    /* [in] */ unsigned char  *lpContainerId,
    /* [in] */ OmsTypeOid  *lpoid,
    /* [out] */ short  *lpb_error) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE LeaveCriticalSection(/*in */ short regionId) { return S_OK; };
  
  virtual HRESULT STDMETHODCALLTYPE LockObj( 
    /* [in] */ unsigned char  *lpConsistentView,
    /* [in] */ unsigned char  *lpContainerId,
    /* [in] */ OmsTypeOid  *lpoid,
    /* [size_is][in] */ unsigned char  *lpobjseq,
    /* [out] */ short  *lpb_error) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE MultiPurpose( 
    /* [in] */ unsigned char MessType,
    /* [in] */ unsigned char MessType2,
    /* [out][in] */ tsp00_Int4  *lpbuflen,
    /* [out][in] */ unsigned char  *lpv,
    /* [out] */ short  *lpb_error) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE NewConsistentView( 
    /* [out][in] */ unsigned char  *lpConsistentView,
    /* [out] */ short  *lpb_error) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE NewObj( 
    /* [in] */ unsigned char  *lpConsistentView,
    /* [in] */ unsigned char  *lpContainerId,
    /* [in] */ unsigned char  *lpVersionContext,
    /* [in] */ tsp00_Int4 KeyPos,
    /* [in] */ tsp00_Int4 KeyColSize,
    /* [in] */ unsigned char  *KeyColumn,
    /* [out] */ OmsTypeOid  *lpoid,
    /* [size_is][out] */ unsigned char  *lpobjseq,
    /* [out] */ short  *lpb_error) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE NextObj( 
    /* [out][in] */ unsigned char  *lpConsistentView,
    /* [in] */ unsigned char  *lpContainerId,
    /* [in] */ unsigned char  *lpVersionContext,
    /* [out][in] */ OmsTypeOid  *lpoid,
    /* [size_is][out] */ unsigned char  *lpobjseq,
    /* [in] */ tsp00_Int4 bodysize,
    /* [size_is][out] */ unsigned char  *lpbody,
    /* [out] */ short  *lpb_error) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE RegisterClass( 
    /* [in] */ tsp00_Int4          SchemaId,
    /* [in] */ GUID                *lpGUID,
    /* [in] */ tsp00_KnlIdentifier *lpClassName,
    /* [in] */ tsp00_Int4          size,
    /* [in] */ OmsContainerNo      ContainerNo,
    /* [out][in] */ unsigned char  *lpcoldesc,
    /* [out] */ unsigned char      *lpContainerId,
    /* [out] */ short  *lpb_error) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE Rollback( 
    /* [in] */ tsp00_Int4 buflen,
    /* [size_is][in] */ unsigned char  *lpv,
    /* [out] */ short  *lpb_error) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE SetError( 
    /* [in] */ short errorno,
    /* [in] */ tsp00_Int4 buflen,
    /* [in] */ unsigned char  *lpv,
    /* [in] */ unsigned char codeType ) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE SQLReceive( 
    /* [out] */ unsigned char  * *lpv) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE SQLRequest( 
    /* [in] */ tsp00_Int4 buflen,
    /* [size_is][in] */ unsigned char  *lpv,
    /* [out] */ short  *lp_error) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE SubtransCommit( 
    /* [out] */ short  *lpb_error) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE SubtransRollback( 
    /* [out] */ short  *lpb_error) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE SubtransStart( 
    /* [out] */ short  *lpb_error) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE UnlockObj( 
    /* [in] */ unsigned char  *lpContainerId,
    /* [in] */ OmsTypeOid  *lpoid,
    /* [out] */ short  *lpb_error) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE UpdateObj( 
    /* [in] */ unsigned char  *lpConsistentView,
    /* [in] */ unsigned char  *lpContainerId,
    /* [in] */ unsigned char  *lpVersionContext,
    /* [size_is][in] */ unsigned char  *lpoid,
    /* [size_is][out][in] */ unsigned char  *lpobjseq,
    /* [in] */ tsp00_Int4 bodysize,
    /* [size_is][in] */ unsigned char  *lpbody,
    /* [in] */ unsigned char  *lpcoldesc,
    /* [out] */ short  *lpb_error) { return S_OK; };
    
    virtual HRESULT STDMETHODCALLTYPE Vtrace( 
    /* [in] */ tsp00_Int4 buflen,
    /* [size_is][in] */ unsigned char  *lpv) { return S_OK; };
    virtual HRESULT STDMETHODCALLTYPE GetDefaultContext(/* [out] */ void** p) { return S_OK; };
    virtual HRESULT STDMETHODCALLTYPE SetDefaultContext(/* [in] */ void* p) { return S_OK; };
    virtual HRESULT STDMETHODCALLTYPE OpenFile(/* [in] */ char  *lpFileName,
    /* [in]  */ tsp00_Uint4 FileSize,
    /* [in]  */ boolean     doAppend,
    /* [out] */ tsp00_Int4  *FileHandle,
    /* [out] */ short  *lpb_error) { return S_OK; }
    virtual HRESULT STDMETHODCALLTYPE WriteFile(/* [in] */ tsp00_Int4 FileHandle,
    /* [in]  */ char   *msg,
    /* [out] */ short  *lpb_error) { return S_OK; } 
    virtual HRESULT STDMETHODCALLTYPE CloseFile(/* [in] */ tsp00_Int4 FileHandle,
    /* [out] */ short  *lpb_error) { return S_OK; }
    virtual HRESULT STDMETHODCALLTYPE SyncFile(/* [in] */ char  *lpFileName,
    /* [in] */ tsp00_Int4 FileHandle,
    /* [out] */ short  *lpb_error) { return S_OK; } 
    virtual HRESULT STDMETHODCALLTYPE GetDateTime(/* [out] */ tsp00_Date *Date,
    /* [out] */ tsp00_Time *Time) { return S_OK; } 
    virtual HRESULT STDMETHODCALLTYPE GetClock(/* [out] */ tsp00_Int4 *Sec,
    /* [out] */ tsp00_Int4 *MSec) { return S_OK; } 
    virtual HRESULT STDMETHODCALLTYPE DropVersion( /* in */ unsigned char  *lpVersionContext,
    /* [out] */ short  *lpb_error) { return S_OK; }
    virtual HRESULT STDMETHODCALLTYPE CreateSchema(/* [in] */tsp00_KnlIdentifier *SchemaName,
    /* [out] */ tsp00_Int4           *SchemaId,
    /* [out] */ short                *lpb_error) { return S_OK; }
    virtual HRESULT STDMETHODCALLTYPE ExistsSchema(/* [in] */tsp00_KnlIdentifier *SchemaName,
    /* [out] */ tsp00_Int4           *SchemaId,
    /* [out] */ short                *lpb_error) { return S_OK; }
    virtual HRESULT STDMETHODCALLTYPE DropSchema(/* [in] */ tsp00_Int4 SchemaId,
            /* [out] */ short      *lpb_error) { return S_OK; };
    virtual HRESULT STDMETHODCALLTYPE GetSchemaName(/* [in] */ tsp00_Int4 SchemaId,
            /* [out] */ tsp00_KnlIdentifier *lpSchemaName,
            /* [out] */ short      *lpb_error) { return S_OK; };
    virtual HRESULT STDMETHODCALLTYPE ExistsFile(/* [in] */ char  *lpFileName,
          /* [out] */ bool     &existsFile,
          /* [out] */ bool     &isDirectory,
          /* [out] */ short    &error)  { return S_OK; };
    virtual HRESULT STDMETHODCALLTYPE DropFile(/* [in] */ char  *lpFileName,
          /* [out] */ short  &error)  { return S_OK; };
    virtual HRESULT STDMETHODCALLTYPE GetWorkingDirectory(/* [out] */ tsp00_Pathc& pathName,
          /* [out] */ short  &error) { return S_OK; };
    virtual HRESULT STDMETHODCALLTYPE GetDbrootSapPath(/* [out] */ tsp00_Pathc& pathName,
          /* [out] */ short  &error) { return S_OK; };
    virtual HRESULT STDMETHODCALLTYPE LoadLibraryEx( /* in */ const char* path, 
          /* in  */ bool        loadGlobal,
          /* out */ HANDLE      *hd,
          /* out */ tsp_errtext &errtext,
          /* in  */ int         errtext_size) { return S_OK; };
    virtual HRESULT STDMETHODCALLTYPE FreeLibrary( /* in */ HANDLE hd,
          /* out */ bool        &ok,
          /* out */ tsp_errtext &errtext,
          /* in  */ int         errtext_size) { return S_OK; };
    virtual HRESULT STDMETHODCALLTYPE DropContainer(/* [in] */ OmsSchemaHandle schema,
          /* [in] */ const GUID&        guid,
          /* [in] */ OmsContainerNo     cno,
          /* [out] */ tgg00_BasisError& error) { return S_OK; };
    virtual HRESULT STDMETHODCALLTYPE GetContainerInfo(/* [in] */ tsp00_Uint4 containerHandle,
          /* [out] */ GUID& guid,
          /* [out] */ OmsSchemaHandle& schema,
          /* [out] */ OmsContainerNo&  cno,
          /* [out] */ short& error) { return S_OK; };
    virtual HRESULT STDMETHODCALLTYPE IsObjHistoryUsed( 
            /* [out][in] */ tgg00_ContainerId& containerId, 
            /*      [in] */ const OmsTypeOid&  oid,
            /* [out]     */ bool&              isUsed,
            /* [out]     */ tgg00_BasisError&  e) { return S_OK; };
};

class cco10_OmsGlob {
public :
  OmsObjectId    m_nilOid;
  tgg91_PageRef  m_nilSeq;
  cco10_OmsGlob() : m_nilOid() {
   m_nilSeq.gg91SetNilRef();
  }
};

#if defined(WIN32)
/*int co10_EvalException (OmsHandle* pHandle, LPEXCEPTION_POINTERS lpexceptInfo); PTS 1106505 FF */
int co10_EvalException (IliveCacheSink* m_sysLcSink, LPEXCEPTION_POINTERS lpexceptInfo );
#endif

/*---------------------------------------------------------------------------*/

inline void omsIfNilOidThrowError (OmsHandle* h, const OmsObjectId& oid, const char* method) {
  if (!oid) {
    h->m_pSession->ThrowDBError (e_nil_pointer, method); 
  }
};

/*----------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* inline functions, to ensure inlining on all platforms                     */
/* Exported class methods cannot be defined as inline on most platforms      */
/*---------------------------------------------------------------------------*/

#define __HCO13_IMPL
#include "hco13.h" 

static cco10_OmsGlob OmsGlob_co10;

/*----------------------------------------------------------------------*/

inline void co10_SetCurrMonitorMethod(OmsHandle& h, tgg00_COMMonitorInfo* p) {
  h.m_pSession->SetCurrentMonitorMethod(p); // PTS 1107744
}

/*----------------------------------------------------------------------*/

inline OmsObjectContainerPtr omsGetContainer(const OmsAbstractObject* pObj) {
  if (NULL == pObj) throw DbpError (DbpError::RTE_ERROR, e_nil_pointer);
  return (OmsObjectContainerPtr) ((unsigned char*) pObj - OmsObjectContainer::headerSize());
};

/*----------------------------------------------------------------------*/
/*           OmsHandle  ---  inline methods                             */
/*----------------------------------------------------------------------*/

inline void OmsObjectContainer::Copy (OmsContext_co12* c, OmsObjectContainerPtr beforeImage, int ObjectSize) {
  tsp00_Int2 refCnt;
  if (this->VarObjFlag()) {
    refCnt = REINTERPRET_CAST(OmsVarObjInfo_co12*, &m_pobj)->m_refCnt;
    REINTERPRET_CAST(OmsVarObjInfo_co12*, &m_pobj)->freeVarObj(c);
  }
  memcpy (&m_oid, &beforeImage->m_oid, ObjectSize - sizeof(m_hashnext));
  if (this->VarObjFlag()) {
    REINTERPRET_CAST(OmsVarObjInfo_co12*, &m_pobj)->m_refCnt = refCnt;
    REINTERPRET_CAST(OmsVarObjInfo_co12*, &beforeImage->m_pobj)->m_pvobj = NULL;
  }
}

#endif // __GCO10_H

