/* @lastChanged: "1998-09-21  9:43"

  * @filename:   hco00.h
  * @purpose:    "User Interface for C++ Persistent Objects"
  * @release:    7.1.0.0
  * @see:        "http://www.bea.sap-ag.de/doku/rel-7.1/hco00.html"
  *
  * @copyright:  "Copyright by   SAP AG, 1998"


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/

#if !defined (__liveCache_INCLUDED)
#define __liveCache_INCLUDED

#if _MSC_VER > 1100
#define PLACEMENT_DELETE 1
#endif

#undef  AFX_EXT_CLASS
#define AFX_EXT_CLASS DECLSPEC_IMPORT

#include "ilivecachesink.h" /* no check */
#include "livecachebase.h"  /* no check */

DEFINE_GUID(IID_IliveCacheSink, 
0x8580E621,0x243F,0x11d1,0xB5,0xA1,0xAC,0x7F,0x01,0x06,0x8C,0x36);

#define OMS_DEFAULT_SCHEMA_HANDLE 1

class OmsAbstractObject; // forward declarations
class OmsHandle;
class OmsExternalOid ;

class AFX_EXT_CLASS OmsCriticalSection {
public :
  OmsCriticalSection(OmsHandle*, int id = 0);
  void omsEnter();
  void omsLeave();
  ~OmsCriticalSection();
private :
  void* m_handle;
  int   m_regionId;
  bool  m_inSection;
};

struct OmsObjectId : public OmsTypeOid
{
public :
  inline bool operator< ( const OmsObjectId&r ) const;
  inline bool operator<=( const OmsObjectId&r ) const;
  inline bool operator==( const OmsObjectId&r ) const;
  inline bool operator!=( const OmsObjectId&r ) const;
  inline unsigned int   getPno() const;
  inline int            getPagePos() const;
  inline int            getGeneration() const;
  inline long           omsHashValue(long upb) const; /* PTS 1103189 */
  inline bool operator! () const; 
  inline operator bool  () const; 
protected:
  inline OmsObjectId();
  inline void operator= (const OmsExternalOid&r);
  inline OmsObjectId(OmsTypeUInt4 p, unsigned short pp, unsigned short g);
};

struct OmsVarOid : public OmsObjectId
{
public :
  inline OmsVarOid();
  inline OmsVarOid(OmsHandle&h, const OmsExternalOid&r);
  inline bool operator< ( const OmsVarOid&r ) const;
  inline bool operator<=( const OmsVarOid&r ) const;
  inline bool operator==( const OmsVarOid&r ) const;
  inline bool operator!=( const OmsVarOid&r ) const;
  inline bool operator! () const; 
  inline operator bool  () const; 
  inline bool omsCheckOid ( OmsHandle& ) const;
  inline const void* operator()(OmsHandle&) const;
  inline const void* omsDeRef(OmsHandle&) const;
};


// External Oid

class OmsExternalOid : public OmsObjectId
{
public:
  OmsExternalOid() {}
  OmsExternalOid( const OmsObjectId&r ) : OmsObjectId(r) {}
  OmsExternalOid( const OmsTypeOid&r  ) : OmsObjectId(r.pno, r.pagePos, r.generation) {}
  OmsExternalOid( OmsTypeByte* r) { memcpy (this, r, sizeof(*this)); }
  void omsCopyTo( OmsTypeByte* r) { memcpy (r, this, sizeof(*this)); }
private:
};

// type save Oid

template<class T>
class OmsOid : public OmsObjectId
{
public:
  typedef bool (T::* OmsQualFunc) ();
  inline OmsOid();
  inline OmsOid( OmsHandle&h, const OmsExternalOid&r );
  inline OmsOid( OmsHandle&h, const OmsObjectId&from ); 
  inline bool omsCheckOid ( OmsHandle&, const OmsExternalOid& r ) const;
  inline bool omsCheckOid ( OmsHandle& ) const;
  inline const T* operator()(OmsHandle&) const;
  inline T* operator()(OmsHandle&,bool do_lock) const;
  inline const T* omsDeRef(OmsHandle&) const;
  inline T* omsDeRefForUpd(OmsHandle&,bool do_lock=true) const;
  inline bool operator==( const OmsOid<T>&r ) const;
  inline bool operator!=( const OmsOid<T>&r ) const;
  inline bool operator! () const;
  inline operator bool () const; 
};

class AFX_EXT_CLASS OmsDuplicateKey : public DbpError
{
public:
  OmsDuplicateKey(long e);
};

class AFX_EXT_CLASS OmsLockTimeout : public DbpError
{
public:
  OmsLockTimeout(const OmsObjectId&);
};

class AFX_EXT_CLASS OmsOutOfDate : public DbpError
{
public:
  OmsOutOfDate();
  OmsOutOfDate(const OmsObjectId&);
};

// iterator for all objects of a class

class AFX_EXT_CLASS OmsObjByClsIterBase {
friend class OmsHandle;
public :
  ~OmsObjByClsIterBase();
protected:
	operator bool () const;
	const OmsObjectId& operator() () const;
  OmsAbstractObject* omsDeRef(bool,bool) const;
 	void operator++ ();
#if !defined(WIN32)
public :
#endif
  OmsObjByClsIterBase(const OmsObjByClsIterBase&);
  OmsObjByClsIterBase (OmsHandle&);
private:
	OmsHandle*   m_phandle;
	void*        m_clsinfo;
	void*        m_curr;
  void*        m_next;
  bool         m_iterNewObjectsInVersion;
  int          m_headIndex;
};

template< class T3 >
class OmsObjByClsIter : public OmsObjByClsIterBase
{
public:
  OmsObjByClsIter( OmsHandle& h ) : OmsObjByClsIterBase (h) {}
  inline operator bool () const;
  inline const OmsOid<T3> operator() () const;                
  inline const T3* operator() ( OmsHandle&h ) const;         // old style
  inline T3* operator() ( OmsHandle&h, bool do_lock ) const; // old style
  inline const T3* omsDeRef () const;
  inline T3* omsDeRefForUpd (bool do_lock) const;
  inline T3* omsDeRefForUpd (OmsHandle&h, bool do_lock) const; // old style
  inline void operator++ ();
private :
  void operator= (const OmsObjByClsIter<T3>&); // no copy
};

// Call Back Interface

class OmsCallbackInterface {
  friend class OmsHandle;
public :
  enum Event {
      BeforeCommit,
      AfterCommit,
      AfterRollback,
      ByApplication
  };
protected :
  virtual ~OmsCallbackInterface() {};
  virtual void omsValidate (OmsHandle&, Event, bool inVersion) = 0;
  virtual void omsInvalidate (OmsHandle&, Event, bool inVersion) = 0;
  virtual void omsDestroySelf() = 0;
};

// Handle

class AFX_EXT_CLASS OmsHandle : public DbpBase {
  
  friend class OmsAbstractObject;
  friend class DbpError;
  
public :
  enum OmsMode { OPTIMISTIC, PESSIMISTIC, IMPLICIT };
  
  OmsHandle (OmsTypeInt4 sz = 1007);
  ~OmsHandle ();
  OmsObjectId         omsCastOid(const GUID guid, const OmsObjectId& oid);
  void                omsCommitSubtrans( int requiredLevel= -1 );
  bool                omsCheckOid ( const OmsObjectId&, const GUID& guid );
  bool                omsCheckOid ( const OmsVarOid&);
  void                omsCloseVersion ( const OmsVersionId& );
  OmsSchemaHandle     omsCreateSchema(const char* SchemaName);
  int                 omsCurrentSubtransLevel() const;
  void                omsCreateVersion ( const OmsVersionId& );
  const OmsVersionId* omsCurrentVersionId () const;
  void                omsDelete ( const OmsObjectId& );
  void                omsDelete ( const OmsVarOid& );
  void                omsDelete ( OmsAbstractObject* );
  const void*         omsDeRef  ( const OmsVarOid& );
  void                omsDropSchema(OmsSchemaHandle);
  void                omsDropContainer(const GUID&, OmsSchemaHandle, OmsContainerNo);
  void                omsDropVersion( const OmsVersionId& );
  void                omsDump( void );
  void                omsExceptionHandler(DbpError&);
  bool                omsExistsVersion( const OmsVersionId& );
  void                omsForceDropVersion ( const OmsVersionId& );
  OmsContainerNo      omsGetContainerNo (const OmsObjectId&);
  OmsSchemaHandle     omsGetSchemaHandle(const OmsObjectId&);
  OmsSchemaHandle     omsGetSchemaHandle(const char* SchemaName) const;
  void                omsGetSchemaName(OmsSchemaHandle,char*,int) const;
  OmsCallbackInterface* omsGetTransactionInterface (const GUID&) const;
  int                 omsGetTimeOut() const;
  unsigned long       omsGetVarObjectSize (const OmsVarOid& oid);
  bool                omsIsLocked ( const OmsObjectId& );
  void                omsInitSession (IUnknown** s);
  void                omsLoadVarObject (const OmsVarOid& oid, unsigned long size, void*);
  void                omsLock( const OmsObjectId&);
  void                omsLock( const OmsVarOid& );
  void                omsLocks( const OmsObjectId*, int);
  void                omsMarkVersion ( const OmsVersionId& );
  void                omsNewConsistentView();
  void                omsNewConsistentView(OmsObjectId* pOids, int cntOids, short timeout);
  void                omsNewConsistentView(OmsObjectId* pOids, int cntOids, short timeout, OmsObjectId* pErrOids, int& cntErrOids); 
  const OmsVarOid&    omsNewVarObject(OmsSchemaHandle,OmsContainerNo);
  void                omsOidInfo (const OmsObjectId&, GUID&, OmsSchemaHandle&, 
                                  char* SchemaName, int SchemaNameBufSize,
                                  char* ClassName,  int ClassNameBufSize, OmsContainerNo&); 
  void                omsOpenVersion( const OmsVersionId& );
  void                omsPrintOid (char* buf, size_t bufSize, const OmsObjectId oid);
  void			          omsRaiseError (short msgno, const char* errmsg);
  void                omsReadOnly (bool = true);
  long                omsRefresh();
  void                omsRegTransactionInterface (const GUID&, OmsCallbackInterface*);
  void                omsRelease  (const OmsObjectId&);
  void                omsRelease  (const OmsVarOid&);
  void                omsReleaseAll ();
  void                omsReleaseConsistentView();
  void                omsReleaseVarObject(const OmsVarOid&);
  void                omsRollbackSubtrans( int requiredLevel = -1 );
  void                omsRenameVersion( const OmsVersionId& oldId, const OmsVersionId& newId); 
  void                omsSetTimeOut( short );
  int                 omsStartSubtrans();
  void                omsStoreVarObject ( const OmsVarOid& oid, const void*, unsigned long objLength);
  void                omsTerminate();
  void                omsTrace     ( const char* tracemsg, long TraceMsgLen);
  void                omsTracef    ( const char*fmt, ... );
  void                omsTraceOid  (const OmsObjectId& oid);
  void                omsTransRollback();
private :
  OmsHandle (const OmsHandle&);             // no copy
  OmsHandle& operator= (const OmsHandle&r); // no assignment
  void*  dummy1;
  double dummy2;
  GUID   dummy3;
  void*  dummy4;
  void*  dummy5;
  void*  dummy6;
  void*  dummy7;
};

//
// Definition of abstract persistent Object
//
class AFX_EXT_CLASS OmsAbstractObject {
  
  friend class OmsHandle;
public :
  enum OmsRegEnum { USE_IN_REGISTRY_ONLY };
  
  virtual void      omsCleanUp  ( OmsHandle& );
  virtual void      omsDump     ( OmsHandle&, const char* flags) const;
  void              omsDelete   ( OmsHandle& );
  void              omsFlush    ( OmsHandle& );
  const OmsObjectId &omsGetOid  () const;
  bool              omsIsLocked ( OmsHandle& ) const;
  void              omsLock     ( OmsHandle& );
  void              omsRelease  ( OmsHandle& );
  void              omsStore    ( OmsHandle& );
  virtual void      omsVerify   ( OmsHandle& ) const;
protected :
  OmsAbstractObject ();
  const OmsObjectId &omsOid() const;
  OmsAbstractObject* omsForUpdPtr(OmsHandle& h, bool doLock) const;
  OmsAbstractObject* omsDeRefNext(OmsHandle&,bool do_lock) const;
  bool   omsHistoryInUse(OmsHandle&) const;
  void   omsKey(OmsHandle& Handle, unsigned char* key) const;
  static void omsDeleteAll(OmsHandle& h, const GUID& guid, OmsSchemaHandle, OmsContainerNo ContainerNo);
  static OmsAbstractObject* omsNewObject (OmsHandle&, const GUID&, OmsSchemaHandle, OmsContainerNo );
  static OmsAbstractObject* omsNewRegistryObject (size_t, OmsHandle&, const GUID&);
  static OmsAbstractObject* omsNewKeyedObject( OmsHandle& Handle, const GUID&, const unsigned char* key, OmsSchemaHandle, OmsContainerNo );
  static OmsObjByClsIterBase omsAllOids( OmsHandle& h, const GUID& guid, OmsSchemaHandle, OmsContainerNo);
  static void omsRegClass (
    OmsHandle&, 
    const GUID&,
    const char* ClassName,
    size_t      PersistentSize,
    size_t      ObjectSize, 
    const GUID* pBaseClass,
    OmsAbstractObject* Obj,
    OmsSchemaHandle,
    OmsContainerNo);
  static void omsRegClassAndKey (OmsHandle& Handle,
    const GUID& Guid, 
    const char* ClassName,
    size_t PersistentSize, 
    size_t KeySize, 
    size_t ObjectSize,
    OmsAbstractObject* Obj,
    OmsSchemaHandle,
    OmsContainerNo);
  static const OmsAbstractObject* omsDeRef (const OmsObjectId& oid, OmsHandle& Handle, const GUID& guid);
  static OmsAbstractObject* omsDeRefForUpd (const OmsObjectId& oid, OmsHandle& Handle, const GUID& guid, bool);
  static const OmsAbstractObject* omsDeRefKey (unsigned char* key, OmsHandle&, const GUID&, OmsSchemaHandle, OmsContainerNo);
  static OmsAbstractObject* omsDeRefKeyForUpd (unsigned char* key, OmsHandle&, const GUID&, bool, OmsSchemaHandle, OmsContainerNo);
private :
  OmsAbstractObject(const OmsAbstractObject&);              // no copy
  OmsAbstractObject& operator= (const OmsAbstractObject&r); // no assignment
  
};

// template for persistent objects

template<class T2>
class OmsObject : public OmsAbstractObject
{
  friend class OmsHandle;
  friend class OmsOid<T2>;
public:
  static const GUID guid;
  inline const OmsOid<T2> omsGetOid() const;
  inline T2* omsForUpdPtr(OmsHandle& h, bool doLock=true) const; 
  inline const T2* omsDeRefNext (OmsHandle&ah, bool do_lock=false);
  inline void* operator new ( size_t sz, OmsHandle&ah, OmsSchemaHandle, OmsContainerNo);
  inline void* operator new ( size_t sz, OmsHandle& ah, OmsRegEnum e);
  inline bool  omsHistoryInUse(OmsHandle&) const;
  static void  omsDeleteAll ( OmsHandle&h, OmsSchemaHandle sh, OmsContainerNo ContainerNo )
  {
    OmsAbstractObject::omsDeleteAll( h, guid, sh, ContainerNo );
  }
  static OmsObjByClsIter<T2> omsAllOids( OmsHandle&h, OmsSchemaHandle sh, OmsContainerNo ContainerNo )
  { 
#if defined(WIN32)
    return *REINTERPRET_CAST (OmsObjByClsIter<T2>*, &OmsAbstractObject::omsAllOids( h, guid, sh, ContainerNo ));
#else
    OmsObjByClsIterBase iter = OmsAbstractObject::omsAllOids( h, guid, sh, ContainerNo );
    return *REINTERPRET_CAST (OmsObjByClsIter<T2>*, &iter);
#endif
  }
  static bool omsOidIsInstanceOf(OmsHandle&h, const OmsExternalOid&oid) {
    return h.omsCheckOid( (OmsObjectId) oid, T2::guid );
  }
  static void omsRegClass( OmsHandle&ah, const char* className, 
    size_t sz, OmsObject<T2>* pobj, OmsSchemaHandle sh, OmsContainerNo ContainerNo)
  { 
    OmsAbstractObject::omsRegClass( ah, guid, className, sz, sizeof(T2), NULL, pobj, sh, ContainerNo ); 
  }

  static void omsRegClass( OmsHandle&ah, const char* className, 
    size_t sz, const GUID& BaseClassGuid, OmsObject<T2>* pobj, OmsSchemaHandle sh, OmsContainerNo ContainerNo)
  { 
    OmsAbstractObject::omsRegClass( ah, guid, className, sz, sizeof(T2), &BaseClassGuid, pobj, sh, ContainerNo ); 
  }
protected:
  OmsObject() {}
#if defined(WIN32)
private:
#endif
  ~OmsObject() {} // derived class must not be destructed
#if !defined(WIN32)
private :
#endif

  static const T2* omsDeRef( const OmsOid<T2>&aoid, OmsHandle&ah )   
  { 
    return (T2*)OmsAbstractObject::omsDeRef( aoid, ah, guid );
  }
  static T2* omsDeRefForUpd( const OmsOid<T2>&aoid, OmsHandle&ah, bool doLock )
  { 
    return (T2*)OmsAbstractObject::omsDeRefForUpd( aoid, ah, guid, doLock );
  }
  OmsObject( const OmsObject<T2>& );       // no copy
  void operator= ( const OmsObject<T2>& ); // no assignment
};

// template for persistent objects with key

template<class Obj, class Key>
class OmsKeyedObject : public OmsObject<Obj>
{
public:
  inline void  omsGetKey( OmsHandle&ah, Key& k ) const;
  void* operator new ( size_t sz, OmsHandle&ah, const Key&akey ,OmsSchemaHandle sh, OmsContainerNo ContainerNo)
  {
    return (void*) OmsAbstractObject::omsNewKeyedObject( ah, OmsObject<Obj>::guid, (unsigned char*) &akey, sh, ContainerNo );
  }
  void* operator new ( size_t sz, OmsHandle& ah, OmsAbstractObject::OmsRegEnum e)
  {
    return (void*) OmsAbstractObject::omsNewRegistryObject( sz + sizeof(Key), ah, OmsObject<Obj>::guid); 
  }
  static void  omsRegClass( OmsHandle&ah, const char* className, size_t sz, OmsObject<Obj>* pobj, OmsSchemaHandle sh, OmsContainerNo cno )
  { 
    OmsAbstractObject::omsRegClassAndKey( ah, OmsObject<Obj>::guid, className, sz, sizeof(Key), sizeof(Obj), pobj, sh, cno ); 
  }
  static const Obj* omsKeyAccess( OmsHandle&h, const Key&key, OmsSchemaHandle shdl, OmsContainerNo cno )  
  {
    return (Obj*)OmsAbstractObject::omsDeRefKey( (unsigned char*)(&key), h, OmsObject<Obj>::guid, shdl, cno );
  }
  static Obj* omsKeyAccessForUpd( OmsHandle&h, const Key&key, bool doLock, OmsSchemaHandle sh, OmsContainerNo cno )  
  {
    return (Obj*)OmsAbstractObject::omsDeRefKeyForUpd((unsigned char*)(&key), h, OmsObject<Obj>::guid, doLock, sh, cno );
  }
protected:
};

// lock object

class AFX_EXT_CLASS OmsLockObject {
public :
   OmsLockObject(OmsHandle&);
   ~OmsLockObject();
                           
   void omsEnable ();
   void omsEnable (const OmsLockHandle&);

   const OmsLockHandle& omsGetLockHandle() const;
   void  omsLockShare();
   void  omsLockExclusive();
   void  omsUnlock();
   bool  omsIsLocked() const;
   bool  omsShareLocked() const;
   bool  omsExclusiveLocked() const;

private :
  enum OmsLockMode{ NO_LOCK, LOCK_SHARE, LOCK_EXCLUSIVE };
  OmsHandle*     m_omsHandle;
  void*          m_void;
  OmsLockHandle  m_lockHandle;
  OmsLockMode    m_mode;
};

// abap table stream

class AFX_EXT_CLASS OmsStream {
protected :
  OmsStream(OmsHandle* h, OmsTypeABAPTabHandle& sh, long rsz, long sz, bool istream = true);    
  OmsStream(OmsHandle* h, OmsTypeStreamHandle& sh, long rsz, long sz, bool istream = true);
  ~OmsStream();
  void           omsClose();
  void           omsFlush();
  unsigned char* omsNxt();
  void           omsWrite(unsigned char*);
private :
  OmsHandle*             m_handle;
  OmsTypeABAPTabHandle*  m_stream_handle;
  long                   m_row_curr;
  long                   m_row_count;
  long                   m_row_size;
  long                   m_buf_size;
  bool                   m_eot;
  bool                   m_is_istream;
  unsigned char*         m_row_buf;
};

// abap table input stream

template <class HandleType, class RowType> class OmsIStream : public OmsStream {
public :
  OmsIStream (OmsHandle* h, HandleType& th, long cached_rows = -1) :
       OmsStream (h, th.Handle, sizeof(RowType), cached_rows) {
         RowType* p = th.RowDef;
       };
       RowType* omsRead() { return (RowType*) omsNxt(); };
};

// abap table output stream

template <class HandleType, class RowType> class OmsOStream : public OmsStream {
public :
  OmsOStream (OmsHandle* h, HandleType& th, long cached_rows = -1) :
       OmsStream (h, th.Handle, sizeof(RowType), cached_rows, false) {
         RowType* p = th.RowDef;
       };
  void omsClose () { OmsStream::omsClose (); };
  void omsWrite (RowType& row) { OmsStream::omsWrite ((unsigned char*) &row); };
       
};

// abap timestamp

class AFX_EXT_CLASS OmsTimeStamp
{
public:
  OmsTimeStamp(const OmsTypeABAPTimeStamp& timeStamp);
  const    OmsTypeABAPTimeStamp& omsGetTimeStamp() const;
  unsigned long omsSecsSince01011970();
private :
  unsigned long        m_DateTime;
  OmsTypeABAPTimeStamp m_ABAPTimeStamp;
};

#define OMS_REG_CLASS(a,s,cno,h) {a::omsRegClass(h, #a, sizeof(a), new(h, OmsObject<a>::USE_IN_REGISTRY_ONLY) a, s,cno);}
#define OMS_REG_CLASS_PREFIX(a,prefixSize,h) {a::omsRegClass(h, #a, prefixSize, new(h, OmsObject<a>::USE_IN_REGISTRY_ONLY) a);}
#define OMS_REG_CLASS_INH(a,s,cno,b,h) {a::omsRegClass(h, #a, sizeof(a), b::guid,new(h, OmsObject<a>::USE_IN_REGISTRY_ONLY) a, s,cno);}

extern
void* omsMalloc(size_t size);

extern 
void omsFree(void* p);

////////////////////////////////////////////////////////////////////////////////////////////////////////
//                           Start of Implementation part                                             //
////////////////////////////////////////////////////////////////////////////////////////////////////////

#define NIL_OBJ_PAGE_NO  2147483647
#define NIL_OBJ_PAGE_POS 0
#define NIL_OBJ_VERSION  0

inline OmsObjectId::OmsObjectId() {
  pno        = NIL_OBJ_PAGE_NO;
  pagePos    = NIL_OBJ_PAGE_POS;
  generation = NIL_OBJ_VERSION;
}

inline bool OmsObjectId::operator< ( const OmsObjectId&r ) const
{ 
    return (memcmp (this, &r, sizeof(r)) <  0);
}

inline bool OmsObjectId::operator<=( const OmsObjectId&r ) const
{
    return (memcmp (this, &r, sizeof(r)) <= 0);
}
inline bool OmsObjectId::operator==( const OmsObjectId&r ) const 
{ 
    return (memcmp (this, &r, sizeof(r)) == 0); 
}
inline bool OmsObjectId::operator!=( const OmsObjectId&r ) const
{
    return (memcmp (this, &r, sizeof(r)) != 0);
}

inline OmsObjectId::OmsObjectId(OmsTypeUInt4 p, unsigned short pp, unsigned short g)
{ 
  pno        = p;
  pagePos    = pp;
  generation = g;
}

inline void OmsObjectId::operator= (const OmsExternalOid&r) 
{
  memcpy (this, &r, sizeof(r));
}

inline bool OmsObjectId::operator! () const 
{
  return ((NIL_OBJ_PAGE_NO==pno) && (NIL_OBJ_PAGE_POS==pagePos) && (NIL_OBJ_VERSION==generation));
}

inline OmsObjectId::operator bool () const 
{ 
  return ((NIL_OBJ_PAGE_NO!=pno) || (NIL_OBJ_PAGE_POS!=pagePos) || (NIL_OBJ_VERSION!=generation));
}

inline unsigned int OmsObjectId::getPno() const {
  return pno;
}

inline int OmsObjectId::getPagePos() const {
  return pagePos;
}

inline int OmsObjectId::getGeneration() const {
  return generation;
}

/* PTS 1103189 */
inline long OmsObjectId::omsHashValue(long upb) const { 
  return (((getPno() << 13) + pagePos) % upb);
}

// OmsVarOid

inline OmsVarOid::OmsVarOid() : OmsObjectId() {};

inline OmsVarOid::OmsVarOid( OmsHandle&h, const OmsExternalOid&r )
{
  OmsObjectId::operator=( r );
  if (!omsCheckOid( h )) {
    throw DbpError(E_INCOMPATIBLE_OID);
  }
}
  

inline bool OmsVarOid::operator< ( const OmsVarOid&r ) const
{ 
    return (memcmp (this, &r, sizeof(r)) <  0);
}

inline bool OmsVarOid::operator<=( const OmsVarOid&r ) const
{
    return (memcmp (this, &r, sizeof(r)) <= 0);
}

inline bool OmsVarOid::operator==( const OmsVarOid&r ) const 
{ 
    return (memcmp (this, &r, sizeof(r)) == 0); 
}

inline bool OmsVarOid::operator!=( const OmsVarOid&r ) const
{
    return (memcmp (this, &r, sizeof(r)) != 0);
}

inline bool OmsVarOid::operator! () const 
{ 
  return (OmsObjectId::operator!()) ;
}  

inline OmsVarOid::operator bool () const 
{ 
  return (OmsObjectId::operator bool()) ;  
}

inline bool OmsVarOid::omsCheckOid ( OmsHandle&h ) const
{
  return h.omsCheckOid(*this);
}

inline const void* OmsVarOid::operator()(OmsHandle& h) const 
{
  return h.omsDeRef(*this);
}

inline const void* OmsVarOid::omsDeRef(OmsHandle& h) const
{
  return h.omsDeRef(*this);
}

// OmsOid

template<class T>
inline
OmsOid<T>::OmsOid() : OmsObjectId(NIL_OBJ_PAGE_NO, NIL_OBJ_PAGE_POS, NIL_OBJ_VERSION) {}

template<class T>
inline 
OmsOid<T>::OmsOid( OmsHandle&h, const OmsExternalOid&r )
{
    if (omsCheckOid( h, r ))
        OmsObjectId::operator=( r );
    else
        throw DbpError(E_INCOMPATIBLE_OID);
}

template<class T>
inline OmsOid<T>::OmsOid( OmsHandle&h, const OmsObjectId&from )
{
  OmsObjectId::operator=(h.omsCastOid(T::guid, from));
}

template<class T>
inline
bool OmsOid<T>::omsCheckOid( OmsHandle&h, const OmsExternalOid& r ) const
{
    return h.omsCheckOid( (OmsObjectId) r, T::guid );
}

template<class T>
inline
bool OmsOid<T>::omsCheckOid( OmsHandle&h ) const
{
    return h.omsCheckOid( (OmsObjectId) *this, T::guid );
}

template<class T> 
inline 
const T* OmsOid<T>::operator() ( OmsHandle& ah ) const 
{
    return T::omsDeRef( *this, ah ); 
}

template<class T> 
inline 
T* OmsOid<T>::operator() ( OmsHandle& ah, bool doLock) const 
{
    return T::omsDeRefForUpd( *this, ah, doLock ); 
}

template<class T> 
inline 
const T* OmsOid<T>::omsDeRef ( OmsHandle& ah ) const 
{
    return T::omsDeRef( *this, ah ); 
}

template<class T> 
inline 
T* OmsOid<T>::omsDeRefForUpd ( OmsHandle& ah, bool doLock) const 
{
    return T::omsDeRefForUpd( *this, ah, doLock ); 
}

template<class T> 
inline 
bool OmsOid<T>::operator==( const OmsOid<T>&r ) const
{ 
    return (memcmp (this, &r, sizeof(r)) == 0);
}
template<class T> 
inline 
bool OmsOid<T>::operator!=( const OmsOid<T>&r ) const
{ 
    return (memcmp (this, &r, sizeof(r)) != 0);
}
template<class T> 
inline 
bool OmsOid<T>::operator! () const 
{
  return (OmsObjectId::operator! ());
}
template<class T> 
inline 
OmsOid<T>::operator bool () const
{ 
  return (OmsObjectId::operator bool ()); 
}

template< class T3 >
inline
OmsObjByClsIter<T3>::operator bool () const
{ 
    return OmsObjByClsIterBase::operator bool ();
}

template< class T3 >
inline
const T3* OmsObjByClsIter<T3>::operator() ( OmsHandle&h ) const 
{ 
    return REINTERPRET_CAST(T3*, OmsObjByClsIterBase::omsDeRef(false, false));
}

template< class T3 >
inline
T3* OmsObjByClsIter<T3>::operator() ( OmsHandle&h, bool do_lock ) const 
{ 
    return REINTERPRET_CAST(T3*, OmsObjByClsIterBase::omsDeRef(true, do_lock));
}

template< class T3 >
inline
const T3* OmsObjByClsIter<T3>::omsDeRef () const 
{ 
    return REINTERPRET_CAST(T3*, OmsObjByClsIterBase::omsDeRef(false,false));
}

template< class T3 >
inline
T3* OmsObjByClsIter<T3>::omsDeRefForUpd ( bool do_lock ) const 
{ 
    return REINTERPRET_CAST(T3*, OmsObjByClsIterBase::omsDeRef(true,do_lock));
}

template< class T3 >
inline
void OmsObjByClsIter<T3>::operator++ () 
{
    OmsObjByClsIterBase::operator++();
}

template< class T3 >
inline
const OmsOid<T3> OmsObjByClsIter<T3>::operator() () const 
{
    return *REINTERPRET_CAST(const OmsOid<T3>*, &OmsObjByClsIterBase::operator()());
}

template<class T2>
inline 
const T2* OmsObject<T2>::omsDeRefNext (OmsHandle&ah, bool do_lock)
{
  return (T2*)OmsAbstractObject::omsDeRefNext(ah, do_lock );
}

template<class T2>
inline 
const OmsOid<T2> OmsObject<T2>::omsGetOid() const 
{ 
    return *(OmsOid<T2>*)(&omsOid());
}

template<class T2>
inline
bool OmsObject<T2>::omsHistoryInUse(OmsHandle& h) const
{
  return OmsAbstractObject::omsHistoryInUse(h);
}

template<class T2>
inline 
T2* OmsObject<T2>::omsForUpdPtr(OmsHandle& h, bool doLock) const
{
  return (T2*) OmsAbstractObject::omsForUpdPtr(h, doLock); 
}

template<class T2>
inline 
void* OmsObject<T2>::operator new ( size_t sz, OmsHandle&ah, OmsSchemaHandle sh, OmsContainerNo cno )
{
    return (void*) OmsAbstractObject::omsNewObject( ah, guid, sh, cno );
}

template<class T2>
inline 
void* OmsObject<T2>::operator new ( size_t sz, OmsHandle& ah, OmsRegEnum e ) 
{
    return (void*) OmsAbstractObject::omsNewRegistryObject( sz, ah, guid ); 
}

template<class Obj, class Key>
inline
void OmsKeyedObject<Obj,Key>::omsGetKey( OmsHandle&ah, Key& k ) const
{ 
    omsKey ( ah, REINTERPRET_CAST(unsigned char*, &k) );
}

inline bool operator== (OmsTypePacked_8_3& l, OmsTypePacked_8_3& r) {
  return (0 == memcmp(&l, &r, sizeof(r)));
}

inline bool operator!= (OmsTypePacked_8_3& l, OmsTypePacked_8_3& r) {
  return (0 != memcmp(&l, &r, sizeof(r)));
}

# endif

