/*!================================================================
 module:        hco12.h
 
 responsible:   ThomasA
 
 special area:  OMS
 
 description:   Implementation of the OMS Context   
 
 see:
 
 copyright:     SAP AG 1998 


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 ===================================================================*/

#ifndef HCO12_H
#define HCO12_H


/*****************************************************
 *     RefType: C++
 *****************************************************/


/*****************************************************
 *     I N C L U D E S                               *
 *****************************************************/

#include "hco13.h"
#include "hco16.h"
#include "ggg92.h"
#include "ggg250.h"
#include "ggg251.h"
#include "gco001.h" 
#include "heo381.h"

extern   co16_Allocator          Allocator_co16;
extern   co16_SharedMemAllocator SharedMemAllocator_co16;

/*****************************************************
 *     P R O T O T Y P E S                           *
 *****************************************************/

enum VarObjLockType_co12 {
  VarObjNoLock,
  VarObjShareLock,
  VarObjExclusiveLock
};

class CUnknownContainerId_co12 : public tgg00_ContainerId {
public :
  CUnknownContainerId_co12() {
    fileObjFileType_gg00().becomes(oftUnknownType_egg00);
  }
  tsp00_Uint4 GetContainerHandle() {
    return fileObjFileNo_gg00();
  }
  bool IsVarObjFile() {
    return (oftVarLenObjFile_egg00 == fileObjFileType_gg00());
  }
};

class OmsSession_co12;

/*!
  class : OmsContext_co12
  description : Implements an OMS Context. A Context is a consistent view togethher 
                with cached Objects belonging to this view.
*/
class OmsContext_co12 {
  class ContextHeap : public ceo381_Heap {
  public :
    ContextHeap(IliveCacheSink** pSink);
    virtual teo381_ChunkPtr GetMem(size_t sz);
    virtual void FreeMem(teo381_ChunkPtr p);
    virtual ~ContextHeap(); // required by sun compiler
  private :
    IliveCacheSink** m_sysLcSink;
  };
   ContextHeap              m_heap;
public :
   cco13_OidHash            m_oidDir;
   cco13_ClsIdHash          m_containerDir;
   tgg91_TransNo            m_consistentView;
   OmsSession_co12*         m_session;
   OmsContext_co12(OmsSession_co12* session, const OmsVersionId* vid = NULL, const tgg00_OmsVersionContext* vctxt = NULL);
   /*!
     function : CloseVersion()

     description : closes the current open version 
   */
   ~OmsContext_co12();
   inline void                  AssignLcSink(IliveCacheSink* pSink);
   inline bool                  IsBoundToTrans() { return m_boundToTrans; }
   inline void                  CheckDBError(tsp00_Int2 e, const char* msg, const OmsObjectId& oid);
   inline void                  CheckCompatibility(const GUID&, OmsObjectContainerPtr);
   inline void                  CheckIfContainerNotDropped(tco13_ClsIdEntry*);
   inline bool                  CheckOid (const OmsObjectId& oid, const GUID& guid);
   inline bool                  CheckOid (const OmsVarOid& oid);
          void                  CleanContainerDir();
   inline void                  ClearObjCache();
   inline void                  CloseVersion();
   inline tsp00_Int4            CurrentVersionKey() const;
          void                  DeleteSelf();
   inline void                  DeleteObjInContext (OmsObjectContainerPtr p, tco13_ClsIdEntry*);
          void                  DumpObjCache(OmsHandle&);
          void                  EmptyObjCache(tsp00_Uint4 containerHandle = 0xFFFFFFFF);
   inline bool                  ExistsConsistentView();
   inline OmsObjectContainerPtr FindObjInContext (const OmsObjectId* pOid);
   inline OmsObjectContainerPtr FindVarObjInContext (const OmsObjectId& oid, VarObjLockType_co12, unsigned long size, void* buf);
          void                  FlushObj(OmsObjectContainerPtr p);
          void                  FlushDelete (OmsObjectContainerPtr pObj, bool deleteInHash); 
          void                  FlushObjCache(bool emptyObjCache = false);
          void                  FlushVarObject (OmsObjectContainerPtr p); 
   inline void                  deallocate(void* p);
   inline void                  FreeVarObj(unsigned char* p, size_t sz);
   inline tco13_GuidEntry*      GetClassInfo(const GUID& guid);
   inline tco13_ClsIdEntry*     GetContainerInfo(tsp00_Uint4 containerHandle);
   inline tco13_ClsIdEntry*     GetContainerInfoNoCheck(tsp00_Uint4 containerHandle);
   inline OmsSession_co12*      GetCurrentSession(); 
   inline OmsObjectContainerPtr GetObj(const OmsObjectId& oid, bool doLock);
   inline OmsObjectContainerPtr GetObj(const GUID& guid, const OmsObjectId& oid, bool doLock);
   inline OmsObjectContainerPtr GetObjFromLiveCacheBase(const OmsObjectId& oid, bool);
   inline OmsObjectContainerPtr GetObjFromLiveCacheBase(const GUID&, const OmsObjectId& oid, bool);
   inline OmsObjectContainerPtr GetObjViaKey(tco13_ClsIdEntry& clsinfo, unsigned char* key, bool doLock);
   const tsp00_Date&            GetCreateDate() const { return m_date; }
   const tsp00_Time&            GetCreateTime() const { return m_time; }
   const tsp00_Date&            GetOpenDate() const { return m_lastOpenDate; }
   const tsp00_Time&            GetOpenTime() const { return m_lastOpenTime; }
   const OmsVersionId&          GetVersionId() { return m_version; }
   inline int                   GetVarObjFromLiveCacheBase(const OmsObjectId& oid, bool);
   inline void                  IncVersionKey ();
   bool                         IsDropped() const { return m_isDropped; }
   bool                         IsMarked() { return m_marked; } 
   inline bool                  IsVersionOpen();
   inline bool                  IsOpenVersion (const OmsVersionId& versionId);
   inline bool                  IsUnloadable();
   inline bool                  IsUnloaded() const { return (m_pVersionContext != NULL); }
   inline bool                  IsVersion();
   IliveCacheSink*              LcSink() const { return m_currLcSink; }
   inline OmsObjectContainerPtr LoadVarObject (const OmsVarOid& oid, VarObjLockType_co12, tsp00_Int4 objSize, OmsVarObjChunk_co001& chunk, void* buf);
   inline void                  LockObj(const OmsObjectId&oid);
   inline void                  LockResult (tsp00_Int2 DBError, OmsObjectContainerPtr p, const OmsObjectId& oid, const char* msg); 
   inline void*                 allocate(size_t sz);
   void                         MarkBoundToTrans() { m_boundToTrans = true; }
   void                         MarkDropped() { m_isDropped = true; }
   void                         MarkNotBoundToTrans();
   void                         MarkVersion() { m_marked = true; }
   void                         NewConsistentView();
   void                         OpenVersion(OmsSession_co12* session);
   inline void                  PutObjectIntoContext (OmsObjectContainerPtr p, tsp00_Uint4 ContainerHandle);
   inline void                  RenameVersion(const OmsVersionId& OldId, const OmsVersionId& NewId);
   inline void                  ResetConsistentView ();
          void                  ResetLcSink(); 
   inline void                  RemoveUnlockedObjFromCache();
   inline void                  SetConsistentView(const tgg91_TransNo&);
          void                  UnLoad();
          void                  VersionClearObjCache();
   inline unsigned char*        VersionContext();
   inline OmsObjectId           VersionNewOid();
   inline void*                 operator new(size_t sz, IliveCacheSink* lcSink);
#if defined(PLACEMENT_DELETE_MEO00)
   inline void                  operator delete (void* p, IliveCacheSink* lcSink);
#endif
   inline tsp00_8ByteCounter    HeapUsed();
          tsp00_8ByteCounter    VarObjHeapUsed() { return m_varObjHeapUsed; }
          tgg00_FileId&         VersionFile()    { return m_versionContext.ovc_tree; }
          tgg00_FileId&         VersionInvFile() { return m_versionContext.ovc_index_tree; }
private :
  OmsVersionId             m_version;
  tgg00_OmsVersionContext  m_versionContext;
  tgg00_OmsVersionContext* m_pVersionContext;
  public :
    tsp00_8ByteCounter       m_varObjHeapUsed;
  private :
    tsp00_Uint4              m_lastDropId;
    bool                     m_isOpen;
    bool                     m_isDropped;
    bool                     m_isVersion;
    bool                     m_boundToTrans;
    bool                     m_marked;
    tsp00_Date               m_date;
    tsp00_Time               m_time;
    tsp00_Date               m_lastOpenDate;
    tsp00_Time               m_lastOpenTime;
    IliveCacheSink*          m_currLcSink;
    IliveCacheSink*          m_sysLcSink; 
};
/*! endclass: OmsContext_co12 */

/*!
  class : OmsSession_co12
  description : Implements an OMS Session representing a kernel session in the OMS.
*/

typedef OmsContext_co12* OmsContextPtr_co12;

class OmsSession_co12 {
public :
                               OmsSession_co12(IliveCacheSink* lcSink, int cntRegions);
                               ~OmsSession_co12();
  inline void* operator new(size_t sz, IliveCacheSink* lcSink);
#if defined(PLACEMENT_DELETE_MEO00)
  inline void  operator delete(void* p, IliveCacheSink* lcSink);
#endif
  inline void                  ActivateDefaultContext();
  inline void                  AddContainerInfo(tco13_ClsIdEntry* classInfo);
  const  OmsObjectId&          CastOid(const GUID& castToGuid, const OmsObjectId& oid);
         void                  ChangedConsistentView();
         void                  CleanAfterDropSchema();
         void                  ClearDefaultContext();
  inline void                  ClearFreeLists();
  inline void                  CloseVersion();
  inline void                  CommitSubtrans(int requiredLevel);
  inline void                  ClearTransVersions();
  inline void                  CreateDefaultContext();
         void                  CreateVersion(const OmsVersionId& versionId);
  inline int                   CurrentSubtransLevel() const;
  inline OmsContext_co12*      CurrentContext() const;
  inline void                  DecSubtransLevel();
         void                  DeleteAll(const GUID& guid, OmsSchemaHandle Schema, OmsContainerNo ContainerNo);
  inline void                  DeleteSession();
         void                  DeleteSelf(IliveCacheSink* lcSink);
         void                  DeleteVarObject(const OmsVarOid& oid);
  inline OmsAbstractObject*    DeRef (const OmsObjectId& oid, const GUID& guid, bool forUpd, bool doLock);
  inline OmsAbstractObject*    DeRefViaKey (unsigned char* key, const GUID&, bool forUpd, bool do_lock, OmsSchemaHandle, OmsContainerNo);
  const  void*                 DeRefVarObject(const OmsVarOid& oid);
         void                  DropVersion(OmsContext_co12* context);
  inline void                  ExecuteSubtransCommit(int requiredLevel);
  inline void                  deallocate(void* p);
  inline void                  FreeLockObjAndBeforeImages();
  inline void                  FreeStackHeap();
  inline OmsAbstractObject*    ForUpdPtr(OmsObjectContainerPtr pObj);
  inline OmsCallbackInterface* GetCallbackInterface() const;
  inline OmsCallbackInterface* GetCallbackInterface(const GUID& guid) const;
  inline tco13_ClsIdEntry*     GetClsInfo(const GUID& guid, OmsSchemaHandle Schema, OmsContainerNo ContainerNo);
  inline tco13_GuidEntry*      GetClassInfo(const GUID& guid);
  inline tco13_ClsIdEntry*     GetContainerInfo(tsp00_Uint4 containerHandle);
  inline OmsObjectContainerPtr GetMemory (tco13_ClsIdEntry&);
  inline OmsObjectContainerPtr GetMemory (tco13_GuidEntry& classInfo);
  inline unsigned char*        GetVarObjMemory(size_t sz);
  inline tsp00_Int4            GetVarObjSize(const OmsVarOid& oid);
  inline tsp00_8ByteCounter    HeapUsed(); 
         bool                  HistoryInUse(const OmsObjectId&);
  inline bool                  InVersion();
  inline void                  IncRefCnt();
  inline void                  IncSubtransLevel ();
  inline void                  IncDeref();
  inline void                  IncDereflC();
  inline void                  IncDerefKey();
  inline void                  IncStore(); 
  inline void                  IncStore_lC();
  inline void                  IncDelete (); 
  inline void                  IncDelete_lC(); 
  inline void                  IncDelete_lC(tsp00_Int4 cntDeleted); 
  inline void                  IncLock();
  inline void                  IncLogHop(int cnt); 
  inline void                  IncExceptions();
  inline void                  IncOutOfDate(); 
  inline void                  IncTimeout(); 
  inline void                  IncSubtransCommit(); 
  inline void                  IncSubtransRollback(); 
  inline void                  IncLoadVarObj(); 
  inline void                  IncStoreVarObj();
  inline void                  InsertTransVersion(OmsContext_co12* context);
         bool                  IsLocked (const OmsObjectId& oid);
  inline OmsObjectContainerPtr LoadVarObject (const OmsVarOid& oid, VarObjLockType_co12 lockReq, tsp00_Int4 size, void* buf);
  inline void                  MonitorSubtransLevel(); 
  inline void                  SetCacheSize(tsp00_Uint4 sz); 
  inline void                  IncCreateVersion (); 
  inline void                  IncOpenVersion (); 
  inline void                  IncCloseVersion (); 
  inline void                  IncDropVersion (); 
  inline void                  IncDeleteVarObject (); 
  inline bool                  InsertBeforeImage (OmsObjectContainerPtr p);
  inline void                  InsertNewBeforeImage (OmsObjectContainerPtr p, tco13_ClsIdEntry* clsinfo);
  inline bool                  IsDefaultContext(OmsContext_co12* context);
  inline bool                  IsSubtransOpen();
  inline void                  LockObj(OmsObjectContainerPtr p);
  inline void*                 allocate(size_t sz);
  inline void                  MonitorVarObjSize(tsp00_Int4);
  inline void                  MonitorWaitNewConsistentView(tsp00_Int4);
         void                  NewConsistentView (OmsObjectId* pOids, int cntOids, short timeout, OmsObjectId* pErrOids, int& cntErrOids); 
         OmsAbstractObject*    NewKeyedObject(const GUID& guid, const unsigned char* key, OmsSchemaHandle Schema, OmsContainerNo ContainerNo);
         OmsAbstractObject*    NewObject(const GUID& guid, OmsSchemaHandle Schema, OmsContainerNo ContainerNo);
         const OmsVarOid&      NewVarObject(OmsSchemaHandle Schema, OmsContainerNo ContainerNo);
  inline void                  OpenVersion(OmsContext_co12* context);
         void                  RegContainer(OmsSchemaHandle Schema,
                                 const GUID&        guid,
                                 const char*        ClassName, 
                                 size_t             PersistentSize,
                                 size_t             ObjectSize,
                                 const GUID*        pBaseClass,
                                 OmsAbstractObject* Obj, 
                                 OmsContainerNo     ContainerNo);
         void                  RegVarObjectContainer(OmsSchemaHandle Schema, OmsContainerNo cno);
         void                  RegisterCallbackInterface(const GUID& guid, OmsCallbackInterface* pInterface);
  inline void                  ResetSubtransLevel ();
         void                  ReleaseCallbackInterface();
  inline void                  ReleaseObj(const OmsObjectId& oid);
  inline void                  ReleaseObj(OmsObjectContainerPtr p);
  inline void                  ReleaseObjUnconditional(const OmsObjectId& oid);
         void                  ReleaseVarObject(const OmsVarOid& oid);
         void                  RemoveFromTransVersion(OmsContext_co12* context);
  inline void                  RollbackSubtrans(int requiredLevel);
  inline void                  SetContext(OmsContext_co12* context) {m_context = context;}
         void                  SetCurrentMonitorMethod(tgg00_COMMonitorInfo* p); // PTS 1107744 
         void                  SetTimeout(short timeout);
  inline void*                 StackHeapMalloc(size_t size);
  inline int                   StartSubtrans();
  inline void                  StoreVarObject (const OmsVarOid& oid, const void* pVarObj, unsigned long objLength);
         void                  ThrowDBError(tsp00_Int2 e, const char* msg);
         void                  ThrowDBError(tsp00_Int2 e, const char* msg, const OmsObjectId& oid);
         void                  ThrowReturnCode(tsp00_Int2 e, const char* msg);
         void                  Trace( const char*fmt, ... );
         void                  TransEnd();
         void                  UnlockObj (const OmsObjectId& oid);
         bool                  VersionBoundByMe(OmsContext_co12* context);

  GUID                     m_callbackInterfaceGuid;
  OmsCallbackInterface*    m_callbackInterface;
  IliveCacheSink*          m_lcSink;
  OmsTypeInt64             m_sizeVarobj;
  OmsTypeInt64             m_cntVarobj;
  tsp00_Int4               m_timeout;
  int                      m_refCnt;
  int                      m_subtrans_lvl;
  bool                     m_stream_io;
  bool                     m_read_only;
  bool*                    m_critical_section_in_use;
  OmsLockObject*           m_lockObjects;
  OmsContext_co12*         m_context;
  OmsContext_co12*         m_defaultContext;
  cco13_BeforeImages       m_beforeImages;
 
  OmsVarObjChunk_co001     m_currVarObjChunk;
  cgg251DCUList<OmsContext_co12*,OmsSession_co12> m_versionsBoundToTrans;
  private :
    void operator delete (void*) {
      throw e_invalid;
    }
    tgg00_COMMonitorInfo*  m_monitor_curr;
    co16_OmsStackHeap      m_stackHeap;
    tsp00_8ByteCounter     m_heapUsed; 
    tsp00_8ByteCounter     m_heapAtMethodBegin;
};
/*! endclass: OmsSession_co12 */


class OmsVarObjInfoNode_co12 {
public :
  OmsVarObjInfoNode_co12() : m_vobjOid(), m_vobjSeq(), m_vobjNext(NULL) {
  }

  OmsVarObjInfoNode_co12(const OmsObjectId& oid, const tgg91_PageRef& objSeq) : m_vobjOid(oid), m_vobjSeq(objSeq), m_vobjNext(NULL) {
  }

  static void* operator new(size_t size, OmsContext_co12& c) {
    if (c.IsVersion()) {
      return c.allocate(sizeof(OmsVarObjInfoNode_co12));
    }
    else {
      return c.m_session->StackHeapMalloc(sizeof(OmsVarObjInfoNode_co12));
    }
  }
#if defined(PLACEMENT_DELETE_MEO00)
  void operator delete(void* p, OmsContext_co12& c) {
    if (c.IsVersion()) {
      c.deallocate(p);
    }
  }
#endif

//  static void deleteVarObjInfoNode (OmsContext_co12& c, OmsVarObjInfoNode_co12& p) {
//  }
  OmsObjectId             m_vobjOid;
  tgg91_PageRef           m_vobjSeq;
  OmsVarObjInfoNode_co12* m_vobjNext;
private :
};

class OmsVarObjInfo_co12 : public OmsAbstractObject {  
public :
  OmsVarObjInfo_co12() : m_vobjSize(0), m_refCnt(0), m_pvobj(NULL), m_vobjInfo() {}
 
  OmsVarObjInfo_co12(const OmsObjectId& oid, tsp00_Uint4 size, const tgg91_PageRef& objSeq) : 
    m_vobjSize(size), m_refCnt(0), m_pvobj(NULL), m_vobjInfo(oid, objSeq) {}

  void freeVarObj (OmsContext_co12* context) {
    context->FreeVarObj(m_pvobj, m_vobjSize);
    m_pvobj = NULL;
  }

  void lock(VarObjLockType_co12 lockReq) {
    switch (lockReq) {
    case VarObjNoLock :
      return;
    case VarObjShareLock :
      if (m_refCnt >= 0) {
        ++m_refCnt;
        return;
      }
      break;
    case VarObjExclusiveLock :
      if (0 == m_refCnt) {
        m_refCnt = -1;
        return;
      }
    }
    throw DbpError (DbpError::DB_ERROR, e_duplicate_object);
  }

  void unlock() {
    if (m_refCnt > 0) {
      /* share lock */
      --m_refCnt;
    }
    else {
      if (m_refCnt < 0) {
        /* exclusive lock */
        m_refCnt = 0;
      }
    }
  }

  void* operator new(size_t sz, OmsAbstractObject* pObj) {
    return pObj;
  }
#if defined(PLACEMENT_DELETE_MEO00)
  void operator delete(void* p, OmsAbstractObject* pObj) {}
#endif

  tsp00_Uint4            m_vobjSize;
  tsp00_Int2             m_refCnt;
public :
  unsigned char*         m_pvobj;
  OmsVarObjInfoNode_co12 m_vobjInfo;

};


class CriticalSection_co12 : public CriticalSection_co001 {
public :
  CriticalSection_co12(OmsSession_co12* session,int regionId) : CriticalSection_co001(session->m_lcSink, regionId) {} 
};

/*****************************************************
*     I M P L E M E N T A T I O N                   *
*****************************************************/

/*----------------------------------------------------------------------*/

inline void OmsContext_co12::AssignLcSink(IliveCacheSink* pSink) {
  if (NULL != m_currLcSink) {
    if (!m_session->VersionBoundByMe(this)) {
      throw DbpError (e_invalid, "AssignLcSink");
    }
  }
  m_currLcSink = pSink; 
  m_sysLcSink  = co10_GetLcSink();
}

/*----------------------------------------------------------------------*/

inline void OmsContext_co12::CheckDBError(tsp00_Int2 e, const char* msg, const OmsObjectId& oid) {
  if (0 != e) {
    m_session->ThrowDBError (e, msg, oid);
  }
}

/*----------------------------------------------------------------------*/

inline void OmsContext_co12::CheckCompatibility(const GUID& guid, OmsObjectContainerPtr pObj) {
  if (NULL == pObj) {
    return;
  }
  if (pObj->GetContainerInfo(this)->GetGuid() != guid) {
    tco13_ClsIdEntry* pContainerInfo = pObj->GetContainerInfo(this);
    if (NULL != pContainerInfo) {
      if (pContainerInfo->IsDerivedClassOf(guid)) {
        return;
      }
    }
    tco13_GuidEntry* pClassInfo = GetClassInfo(guid);
    if (pClassInfo->IsBaseClass()) {
    }
    m_session->ThrowDBError (e_invalid, "OmsContext_co12::CheckCompatibility");
  }
}

/*----------------------------------------------------------------------*/

inline void OmsContext_co12::CheckIfContainerNotDropped(tco13_ClsIdEntry* pContainerInfo) {
  m_containerDir.CheckIfNotDropped(pContainerInfo);
}

/*----------------------------------------------------------------------*/

inline bool OmsContext_co12::CheckOid (const OmsObjectId& oid, const GUID& guid) {
  
  if (oid) {
    try {
      tco13_ClsIdEntry* pContainerInfo;
      OmsObjectContainerPtr pObj = this->FindObjInContext (&oid);
      if (NULL != pObj) {
        if (pObj->DeletedFlag()) {
          return false;
        } 
        pContainerInfo = pObj->GetContainerInfo(this);
        if (pContainerInfo->GetGuid() != guid ) {
          return false;
        }
      }
      else {
        pObj = GetObjFromLiveCacheBase(oid, false);
        if (NULL == pObj) {
          return false;
        }
        pContainerInfo = pObj->GetContainerInfo(this);
        if (pContainerInfo->GetGuid() != guid ) {
          return false;
        }
      }
    }
    catch (DbpError&) {
      return false;
    }
  }
  return true;
}

/*----------------------------------------------------------------------*/

inline bool OmsContext_co12::CheckOid (const OmsVarOid& oid)
{
  if (oid) {
    try {
      tco13_ClsIdEntry* pContainerInfo;
      OmsObjectContainerPtr pObj = this->FindObjInContext (&oid);
      if (NULL != pObj) {
        if (pObj->DeletedFlag()) {
          return false;
        } 
        pContainerInfo = pObj->GetContainerInfo(this);
        if (!pContainerInfo->IsVarObject()) {
          return false;
        }
      }
      else {
        pObj = GetObjFromLiveCacheBase(oid, false);
        if (NULL == pObj) {
          return false;
        }
        if (!pObj->VarObjFlag()) {
          return false;
        }
      }
    }
    catch (DbpError&) {
      return false;
    }
  }
  return true;  
}

/*----------------------------------------------------------------------*/

inline void  OmsContext_co12::ClearObjCache() {
   m_oidDir.Clear();
}

/*----------------------------------------------------------------------*/

inline void OmsContext_co12::CloseVersion() {
  m_containerDir.ClearFreeLists();
  m_isOpen = false;
}

/*----------------------------------------------------------------------*/

inline tsp00_Int4 OmsContext_co12::CurrentVersionKey() const {
  return m_versionContext.ovc_key;
}

/*----------------------------------------------------------------------*/

inline void OmsContext_co12::DeleteObjInContext (OmsObjectContainerPtr p, tco13_ClsIdEntry* clsinfo) {
  m_oidDir.HashDelete(p->m_oid);
  if (NULL != clsinfo) {
    clsinfo->chainFree(*this, p);
  }
}

/*----------------------------------------------------------------------*/

inline bool OmsContext_co12::ExistsConsistentView() {
  return (!(m_consistentView.gg90IsNil()));
}

/*----------------------------------------------------------------------*/

inline OmsObjectContainerPtr OmsContext_co12::FindObjInContext (const OmsObjectId* pOid) {
  return m_oidDir.HashFind(pOid);
}

/*----------------------------------------------------------------------*/

inline OmsObjectContainerPtr OmsContext_co12::FindVarObjInContext (const OmsObjectId& oid, VarObjLockType_co12 lockReq, unsigned long size, void* buf) {
  const char* msg = "OmsContext_co12::FindVarObjInContext";
  OmsObjectContainerPtr po = this->FindObjInContext (&oid);
  if (NULL != po) {
    if (po->DeletedFlag()) {
      m_session->ThrowDBError (e_object_not_found, msg, oid);
    }
    OmsVarObjInfo_co12* pObjInfo = REINTERPRET_CAST (OmsVarObjInfo_co12*, &po->m_pobj);
    pObjInfo->lock(lockReq);
    if (pObjInfo->m_vobjSize > size) {
      m_session->ThrowDBError (e_too_small_packet_size, msg, oid);
    }
    if (NULL != buf) {
      memcpy (buf, pObjInfo->m_pvobj, pObjInfo->m_vobjSize);
    }
    return po;
  }
  return NULL;
}

/*----------------------------------------------------------------------*/

inline void OmsContext_co12::deallocate(void* p) {
  m_heap.deallocate(p);
}

/*----------------------------------------------------------------------*/

inline void OmsContext_co12::FreeVarObj(unsigned char* p, size_t sz) {
  if (NULL != p) {
    m_varObjHeapUsed -= sz;
  }
  this->deallocate (p);
}

/*----------------------------------------------------------------------*/

inline tco13_GuidEntry* OmsContext_co12::GetClassInfo(const GUID& guid) {
  return m_containerDir.GetClassInfo (guid);
}

/*----------------------------------------------------------------------*/

inline tco13_ClsIdEntry* OmsContext_co12::GetContainerInfo(tsp00_Uint4 containerHandle) {
  return m_containerDir.HashFindViaContainerHandle(containerHandle);
}

/*----------------------------------------------------------------------*/

inline tco13_ClsIdEntry* OmsContext_co12::GetContainerInfoNoCheck(tsp00_Uint4 containerHandle) {
  return m_containerDir.HashFindViaContainerHandle(containerHandle, false);
}

/*----------------------------------------------------------------------*/

inline OmsSession_co12* OmsContext_co12::GetCurrentSession() {
  return m_session;
}

/*----------------------------------------------------------------------*/

inline OmsObjectContainerPtr OmsContext_co12::GetObj(const OmsObjectId& oid, bool doLock) {
  m_session->IncDeref();
  OmsObjectContainerPtr pObj = m_oidDir.HashFind(&oid);//this->FindObjInContext(&oid);
  if (NULL != pObj) {
    if (pObj->DeletedFlag()) {
      return NULL; // PTS 1107036
    }
    if ((doLock) && (!pObj->LockedFlag()) && (!this->IsVersion())) {
      this->LockObj(pObj->m_oid);
    }
  }
  else {
    pObj = this->GetObjFromLiveCacheBase(oid, doLock);
  }
  return pObj;
}


/*----------------------------------------------------------------------*/

inline OmsObjectContainerPtr OmsContext_co12::GetObj(const GUID& guid, const OmsObjectId& oid, bool doLock) {
  m_session->IncDeref();
  OmsObjectContainerPtr pObj = this->FindObjInContext(&oid);
  if (NULL != pObj) {
    if (pObj->DeletedFlag()) {
      return NULL; // PTS 1107036
    }
    if ((doLock) && (!pObj->LockedFlag()) && (!this->IsVersion())) {
      this->LockObj(pObj->m_oid);
    }
  }
  else {
    pObj = this->GetObjFromLiveCacheBase(guid, oid, doLock);
  }
  this->CheckCompatibility(guid, pObj);
  
  return pObj;
}

/*----------------------------------------------------------------------*/

inline OmsObjectContainerPtr OmsContext_co12::GetObjFromLiveCacheBase(const OmsObjectId& oid,
                                                                      bool               doLockObj)
                                                                      
{
  const char* msg = "OmsContext_co12::GetObjFromLiveCacheBase_1";
  int              objLen;
  int              objHistReadCount;
  tgg00_BasisError DBError;
  if (this->IsVersion()) {
    doLockObj = false;
  }
  CUnknownContainerId_co12 FileId;
  m_session->m_currVarObjChunk.Clear();
  CHResult_co001 hr = m_currLcSink->GetObj (
    m_consistentView,
    FileId,
    m_pVersionContext,
    *REINTERPRET_CAST(OmsTypeOid*, CONST_CAST(OmsObjectId*, &oid)),
    doLockObj,  
    m_session->m_currVarObjChunk.m_seq,
    sizeof(m_session->m_currVarObjChunk.m_container), 
    ((unsigned char*) &m_session->m_currVarObjChunk.m_container), 
    objLen,
    objHistReadCount,
    DBError);
  if (0 == DBError) {
    OmsObjectContainerPtr pObj;
    if (FileId.IsVarObjFile()) {
      m_session->m_currVarObjChunk.m_containerHandle = FileId.GetContainerHandle();
      if (objLen <= VAR_OBJ_CHUNK_SIZE) {
        m_session->m_currVarObjChunk.m_container.m_vobjSize = objLen;
      }
      pObj = this->LoadVarObject (*REINTERPRET_CAST(const OmsVarOid*, &oid), VarObjNoLock, 
        m_session->m_currVarObjChunk.m_container.m_vobjSize, m_session->m_currVarObjChunk, NULL);
    }
    else {
      tco13_ClsIdEntry* containerInfo = this->GetContainerInfo(FileId.GetContainerHandle());
      pObj = m_session->GetMemory (*containerInfo);
      memcpy (REINTERPRET_CAST(unsigned char*, &pObj->m_pobj) + sizeof(void*), &m_session->m_currVarObjChunk.m_container,
        containerInfo->GetPersistentSize());
      pObj->m_oid    = oid;
      pObj->m_objseq = m_session->m_currVarObjChunk.m_seq;
      PutObjectIntoContext(pObj, FileId.GetContainerHandle());
    }
    if (doLockObj) {
      pObj->MarkLocked();
    }
    return pObj;
  }
  else {
    if (e_object_not_found != DBError) {
      if (doLockObj) {
        LockResult (DBError, NULL, oid, msg);
      }
      this->CheckDBError(DBError, msg, oid);
    }
    return NULL;
  }
}

/*----------------------------------------------------------------------*/

inline OmsObjectContainerPtr OmsContext_co12::GetObjFromLiveCacheBase(const GUID&        guid, 
                                                                      const OmsObjectId& oid,
                                                                      bool               doLockObj)
                                                                      //bool               addIntoObjCache) 
{
  int              dummyLen;
  int              objHistReadCount;
  tgg00_BasisError DBError;
  const char* msg = "OmsContext_co12::GetObjFromLiveCacheBase_2";
  
  m_session->IncDereflC();
  tco13_GuidEntry* pClassInfo = this->GetClassInfo(guid);
  if (pClassInfo->IsBaseClass()) {
    // oid may identifiy an object of a derived class, i.e. size is unknown
    return this->GetObjFromLiveCacheBase(oid, doLockObj);
  }
  else {
    OmsObjectContainerPtr pObj = m_session->GetMemory (*pClassInfo);
    if (this->IsVersion()) {
      doLockObj = false;
    }
    CUnknownContainerId_co12 FileId;
    CHResult_co001 hr = m_currLcSink->GetObj (
      m_consistentView,
      FileId,
      m_pVersionContext,
      *REINTERPRET_CAST(OmsTypeOid*, CONST_CAST(OmsObjectId*, &oid)),
      doLockObj,  
      pObj->m_objseq,
      pClassInfo->GetObjectSize(), 
      ((unsigned char*) &pObj->m_pobj) + sizeof(void*), 
      dummyLen,
      objHistReadCount,
      DBError);
    if (0 != DBError) {
      pClassInfo->ChainFree(*this, pObj); 
      if (e_object_not_found != DBError) {
        if (doLockObj) {
          LockResult (DBError, pObj, oid, msg);
        }
        this->CheckDBError(DBError, msg, oid);
      }
      return NULL;
    }
    pObj->m_oid = oid;
    if (doLockObj) {
      pObj->MarkLocked();
    }
    //if (addIntoObjCache) {
    PutObjectIntoContext(pObj, FileId.GetContainerHandle());
    //};
    m_session->IncLogHop(objHistReadCount);
    return pObj;
  }
}


/*----------------------------------------------------------------------*/

inline OmsObjectContainerPtr OmsContext_co12::GetObjViaKey(tco13_ClsIdEntry& ContainerInfo, unsigned char* key, bool doLock) {
  
  const char* msg = "OmsContext_co12::GetObjViaKey";

  m_session->IncDerefKey();
  int                 ObjHistReadCount;
  tgg00_BasisError    DBError;
  OmsObjectContainer* obj = m_session->GetMemory(ContainerInfo);
  if (this->IsVersion()) {
    doLock = false;
  }
  CHResult_co001 hr = m_currLcSink->GetObjWithKey (
    (unsigned char*) &m_consistentView,
    (unsigned char*) &ContainerInfo.GetFileId(),
    VersionContext(),
    (*ContainerInfo.GetInvColDesc().colHashList_gg00)[0].elen_var(),
    key, 
    doLock, 
    ContainerInfo.GetPersistentSize(),
    ((unsigned char*) &obj->m_pobj) + sizeof(void*),
    REINTERPRET_CAST(OmsTypeOid*, &obj->m_oid),
    (unsigned char*) &obj->m_objseq,
    &ObjHistReadCount,
    &DBError);
  m_session->IncLogHop(ObjHistReadCount);  
  OmsObjectContainerPtr found = NULL;
  if ( DBError != 0 ) {
    ContainerInfo.chainFree(*this, obj);
    if (e_hash_key_not_found != DBError) {
      this->LockResult (DBError, NULL, OmsObjectId(), msg);
    }
    else {
      if (this->IsVersion()) {
        found = ContainerInfo.VersionFindKey(key);
        if (NULL != found) {
          DBError = 0;
        }
      }
    }
  }
  if (0 == DBError) {
    if (NULL != found) {
      if (found->DeletedFlag()) {
        return NULL;
      }
      else {
        obj = found;
      }
    }
    else {
      if (doLock) {
        this->LockResult (DBError, obj, OmsObjectId(), msg);
      }
      found = this->FindObjInContext (&obj->m_oid);
      if (NULL == found) {
        this->PutObjectIntoContext (obj, ContainerInfo.GetContainerHandle());
      }
      else {
        found->m_objseq = obj->m_objseq;
        if (doLock) {
          found->MarkLocked();
        }
        ContainerInfo.chainFree(*this, obj);
        if (found->DeletedFlag()) {
          return NULL;
        }
        obj = found;
      }
    }
    return obj;
  }
  return NULL;
}

/*----------------------------------------------------------------------*/

inline int OmsContext_co12::GetVarObjFromLiveCacheBase(const OmsObjectId& oid, bool doLock) {
  
  int                 objLen;
  int                 objHistReadCount;
  tgg00_BasisError    DBError;
  CUnknownContainerId_co12 FileId;
  const char* msg = "OmsContext_co12::GetVarObjFromLiveCacheBase";
  CHResult_co001 hr = m_currLcSink->GetObj (
    m_consistentView,
    FileId,
    m_pVersionContext,
    *REINTERPRET_CAST(OmsTypeOid*, CONST_CAST(OmsObjectId*, &oid)),
    doLock, 
    m_session->m_currVarObjChunk.m_seq, 
    sizeof(m_session->m_currVarObjChunk.m_container),
    (unsigned char*) &m_session->m_currVarObjChunk.m_container,
    objLen, objHistReadCount, DBError);
  if ( DBError != 0 ) { 
    if (e_object_not_found != DBError) {
      if (doLock) {
        LockResult (DBError, NULL, oid, msg);
      }
      this->CheckDBError(DBError, msg, oid);
    }
    return -1;
  }
  if (!FileId.IsVarObjFile()) {
    m_session->ThrowDBError (e_object_not_found, msg, oid);
  }
  m_session->m_currVarObjChunk.m_containerHandle = FileId.GetContainerHandle();
  m_session->IncLogHop(objHistReadCount);
  m_session->m_currVarObjChunk.m_oid = oid;
  if (objLen <= VAR_OBJ_CHUNK_SIZE) {
    m_session->m_currVarObjChunk.m_container.m_vobjSize = objLen;
  }
  return m_session->m_currVarObjChunk.m_container.m_vobjSize;
}

/*----------------------------------------------------------------------*/

inline tsp00_8ByteCounter OmsContext_co12::HeapUsed() {
  ceo381_Heap::THeapInfo heapInfo;
  m_heap.getHeapInfo(heapInfo);
  return heapInfo.heapUsed;
}

/*----------------------------------------------------------------------*/

inline void OmsContext_co12::IncVersionKey () {
  ++m_versionContext.ovc_key;
}

/*----------------------------------------------------------------------*/

inline bool OmsContext_co12::IsOpenVersion (const OmsVersionId& versionId) {
  return (0 == memcmp (versionId, m_version, sizeof(versionId)));
}

/*----------------------------------------------------------------------*/

inline bool OmsContext_co12::IsUnloadable() {
  return ((!m_isOpen) && (!m_boundToTrans));
}

/*----------------------------------------------------------------------*/

inline bool OmsContext_co12::IsVersion() {
  return m_isVersion;
}

/*----------------------------------------------------------------------*/

inline bool OmsContext_co12::IsVersionOpen() {
	return m_isOpen;
}

/*----------------------------------------------------------------------*/

inline OmsObjectContainerPtr OmsContext_co12::LoadVarObject (const OmsVarOid& oid, VarObjLockType_co12 lockReq, 
                                                             tsp00_Int4 objSize, OmsVarObjChunk_co001& chunk, void* buf) {
  OmsVarObjInfo_co12*     pObjInfo;
  unsigned char*          p;
  tsp00_Int2              DBError;
  OmsVarObjInfoNode_co12* pInfoNode;
  OmsObjectId             nextOid;
  int                     objLen;
  int                     ObjHistReadCount;

  /* first chunk of var object ist expected in parameter chunk ! */

  const char* msg = "OmsContext_co12::LoadVarObject";
  
  CUnknownContainerId_co12   FileId;
  tco13_GuidEntry* pClassInfo   = this->GetClassInfo(VAR_OBJ_GUID_GCO10);
  OmsObjectContainerPtr po      = m_session->GetMemory(*pClassInfo);
  tsp00_Uint4           moveLen = co001_Min (objSize, VAR_OBJ_CHUNK_SIZE);
  po->m_oid     = oid;
  po->m_objseq  = chunk.m_seq;
  po->MarkVarObj();
  pObjInfo = new(&po->m_pobj) OmsVarObjInfo_co12(oid, objSize, chunk.m_seq); 
  pObjInfo->lock(lockReq);
  pObjInfo->m_pvobj               = m_session->GetVarObjMemory(objSize);
  pObjInfo->m_vobjInfo.m_vobjOid  = oid;
  pObjInfo->m_vobjInfo.m_vobjSeq  = chunk.m_seq;
  pObjInfo->m_vobjInfo.m_vobjNext = NULL;
  pInfoNode = &pObjInfo->m_vobjInfo;
  objSize  -= moveLen;
  try {
    this->PutObjectIntoContext (po, chunk.m_containerHandle);
  }
  catch (DbpError& e) {
    if (e_unknown_guid == e.dbpError()) {
      // auto register varobj container
      GUID             guid;
      OmsSchemaHandle  schema;
      OmsContainerNo   cno;
      tgg00_BasisError DBError;
      m_currLcSink->GetContainerInfo (chunk.m_containerHandle, guid, schema, cno, DBError);
      if (0 == DBError) {
        m_session->RegVarObjectContainer(schema, cno);
        this->PutObjectIntoContext (po, chunk.m_containerHandle);
      }
      else {
        m_session->ThrowDBError(DBError, msg);
      }
    }
    else {
      throw e;
    }
  }
  memcpy (pObjInfo->m_pvobj, &chunk.m_container, moveLen);
  p  = pObjInfo->m_pvobj + moveLen;
  while (objSize > 0) {
    nextOid = chunk.m_container.m_vobjNext;
    moveLen = co001_Min (objSize, VAR_OBJ_CHUNK_SIZE);
    CHResult_co001 hr = m_currLcSink->GetObj (
      m_consistentView,
      FileId,
      m_pVersionContext,
      *REINTERPRET_CAST(OmsTypeOid*, &nextOid),
      false,  
      chunk.m_seq,
      sizeof (chunk.m_container),
      (unsigned char*) &chunk.m_container, 
      objLen,
      ObjHistReadCount,
      DBError);
    this->CheckDBError(DBError, msg, oid);
    memcpy (p, &chunk.m_container, moveLen);
    p       += moveLen;
    objSize -= moveLen;
    pInfoNode->m_vobjNext = new(*this) OmsVarObjInfoNode_co12 (nextOid, chunk.m_seq);
    pInfoNode = pInfoNode->m_vobjNext;
  }
  if (buf != NULL) {
    memcpy (buf, pObjInfo->m_pvobj, pObjInfo->m_vobjSize);
  }
  return po;
}

/*----------------------------------------------------------------------*/


inline void OmsContext_co12::LockObj( const OmsObjectId&oid ) {
  
  const char* msg = "OmsContext_co12::LockObj";
  
  if (m_session->m_read_only) {
    m_session->ThrowDBError (e_file_read_only, msg);
  }
  if (this->IsVersion()) {
    return;
  }
  OmsObjectContainerPtr pObj = this->FindObjInContext(&oid);
  if (NULL != pObj) {
    if (pObj->DeletedFlag()) {
      m_session->ThrowDBError(e_object_not_found, msg, oid);
    }
    if (pObj->LockedFlag()) {
      // Object already locked
      return;
    }
    m_session->LockObj(pObj);
  }
  else {
    pObj = this->GetObjFromLiveCacheBase(oid, true); 
    if (NULL == pObj) {
      m_session->ThrowDBError (e_object_not_found, msg, oid);
    }
  }
}

/*----------------------------------------------------------------------*/

inline void OmsContext_co12::LockResult (tsp00_Int2 DBError, OmsObjectContainerPtr p, const OmsObjectId& oid, const char* msg) {
  if (DBError != 0) {
    if (e_object_dirty == DBError) {
      m_session->IncOutOfDate();
      throw OmsOutOfDate(oid);
    }
    else {
      m_session->ThrowDBError (DBError, msg, oid);
    }
    return;
  }
  else {
    if (NULL != p) {
      p->MarkLocked();
    }
    return;
  }
}

/*----------------------------------------------------------------------*/

inline void* OmsContext_co12::allocate(size_t sz) {
  return m_heap.allocate(sz);
}

/*----------------------------------------------------------------------*/

inline void OmsContext_co12::PutObjectIntoContext (OmsObjectContainerPtr p, tsp00_Uint4 ContainerHandle) {
  const char* msg = "PutObjectIntoContext";
  
  tco13_ClsIdEntry* pContainerInfo = this->GetContainerInfo(ContainerHandle);
#if defined(USE_CONTAINER_HANDLE_CO001)
  p->SetContainerHandle(ContainerHandle);
#else
  p->SetContainerInfo(pContainerInfo);
#endif
  if (NULL != pContainerInfo) {
    m_oidDir.HashInsert(p);
  }
  else {
    m_session->ThrowDBError(e_unknown_guid, msg);
  }
}

/*----------------------------------------------------------------------*/

inline void OmsContext_co12::RenameVersion(const OmsVersionId& OldId, const OmsVersionId& NewId) {
  memcpy (m_version, NewId, sizeof(m_version));
}

/*----------------------------------------------------------------------*/

inline void OmsContext_co12::ResetConsistentView () {
  m_consistentView.gg90SetNil();
}

/*----------------------------------------------------------------------*/

inline void OmsContext_co12::SetConsistentView(const tgg91_TransNo& v) {
   m_consistentView = v;
}

/*----------------------------------------------------------------------*/

void OmsContext_co12::RemoveUnlockedObjFromCache() {
  m_oidDir.RemoveUnlockedObjFromCache();
}

/*----------------------------------------------------------------------*/

inline unsigned char*  OmsContext_co12::VersionContext() {
  return REINTERPRET_CAST(unsigned char*, m_pVersionContext);
}

/*----------------------------------------------------------------------*/

inline OmsObjectId OmsContext_co12::VersionNewOid() {
  OmsObjectId oid;
  oid.setPno(VERSION_OBJ_PAGE_NO_GG92);
  IncVersionKey();
  tgg92_OmsTypeOid* pOid = REINTERPRET_CAST(tgg92_OmsTypeOid*, &oid);
  pOid->generation = CurrentVersionKey() / (256 * 256);  
  pOid->pagePos    = CurrentVersionKey() % (256 * 256);
  return oid;
}

/*----------------------------------------------------------------------*/

inline void* OmsContext_co12::operator new(size_t sz, IliveCacheSink* lcSink) {
  return SharedMemAllocator_co16.allocate(lcSink, sz);
}

/*----------------------------------------------------------------------*/
#if defined(PLACEMENT_DELETE_MEO00)
inline void OmsContext_co12::operator delete (void* p, IliveCacheSink* lcSink) {
  REINTERPRET_CAST(OmsContext_co12*, p)->DeleteSelf();
}
#endif

/*----------------------------------------------------------------------*/
/*               Implementation of OmsSession_co12                      */
/*----------------------------------------------------------------------*/

inline void* OmsSession_co12::operator new(size_t sz, IliveCacheSink* lcSink) {
  return SharedMemAllocator_co16.allocate(lcSink, sz);
}

/*----------------------------------------------------------------------*/
#if defined(PLACEMENT_DELETE_MEO00)
inline void OmsSession_co12::operator delete(void* p, IliveCacheSink* lcSink) {
  REINTERPRET_CAST(OmsSession_co12*, p)->DeleteSession();
}
#endif

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::ActivateDefaultContext() {
  m_context = m_defaultContext;
  this->ChangedConsistentView();
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::AddContainerInfo(tco13_ClsIdEntry* containerInfo) {
  m_context->m_containerDir.HashInsert (containerInfo);
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::CloseVersion() {
  this->IncCloseVersion();
  m_context->CloseVersion();
  this->ActivateDefaultContext();
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::ClearFreeLists() {
  m_defaultContext->m_containerDir.ClearFreeLists();
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::ClearTransVersions() {
  if (!m_versionsBoundToTrans.empty()) {
    CriticalSection_co12 region(this, RGN_VERSION_DIR);
    region.Enter();
    cgg251dclIterator<OmsContext_co12*,OmsSession_co12> iter = m_versionsBoundToTrans.begin();
    while (iter) {
      cgg251dclIterator<OmsContext_co12*,OmsSession_co12> delIter = iter;
      ++iter;
      (*delIter())->CleanContainerDir();
      if (!(*delIter())->IsVersionOpen()) {
        (*delIter())->MarkNotBoundToTrans();
        m_versionsBoundToTrans.remove(delIter);
      }
    }
  }
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::CreateDefaultContext() {
  m_defaultContext = new(m_lcSink) OmsContext_co12(this);
  m_context        = m_defaultContext;
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::CommitSubtrans(int requiredLevel) {
 
  if (!this->IsSubtransOpen()) {
	  return;
  }
  if (-1 == requiredLevel) {
    requiredLevel = this->CurrentSubtransLevel();
  }
  this->IncSubtransCommit();
  this->ExecuteSubtransCommit(requiredLevel);
}

/*----------------------------------------------------------------------*/

inline OmsContext_co12* OmsSession_co12::CurrentContext() const {
  return m_context;
}

/*----------------------------------------------------------------------*/

inline int OmsSession_co12::CurrentSubtransLevel() const {
  return m_subtrans_lvl;
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::DecSubtransLevel() {
  --m_subtrans_lvl;
}

/*----------------------------------------------------------------------*/

inline OmsAbstractObject* OmsSession_co12::DeRef (const OmsObjectId& oid, const GUID& guid, bool forUpd, bool doLock) {
 
  const char* msg = "OmsSession_co12::DeRef";
   
  OmsObjectContainerPtr found = m_context->GetObj (guid, oid, doLock);
  if (found) {
    if ((doLock) && (!found->LockedFlag()) && (!m_context->IsVersion())) {
      m_context->LockObj(found->m_oid);
    }
    if (forUpd) {
      return this->ForUpdPtr(found);
    }
    return &found->m_pobj;
  }
  else {
    this->ThrowDBError (e_object_not_found, msg, oid);
    return NULL;
  }
}

/*----------------------------------------------------------------------*/

inline OmsAbstractObject* OmsSession_co12::DeRefViaKey (unsigned char*  key, 
                                                        const GUID&     guid, 
                                                        bool            forUpd, 
                                                        bool            doLock, 
                                                        OmsSchemaHandle Schema,
                                                        OmsContainerNo  ContainerNo)
{
  tco13_ClsIdEntry*     clsinfo = this->GetClsInfo (guid, Schema, ContainerNo);
  OmsObjectContainerPtr found   = m_context->GetObjViaKey(*clsinfo, key, doLock);
  if (NULL != found) {
    if (forUpd) {
      return this->ForUpdPtr(found);
    }
    return &found->m_pobj;
  }
  return NULL;
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::DeleteSession() {
  --m_refCnt;
  if (0 == m_refCnt) {
    this->OmsSession_co12::~OmsSession_co12();
    SharedMemAllocator_co16.deallocate(co10_GetLcSink(), this);
  }
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::ExecuteSubtransCommit(int requiredLevel) {
  tsp00_Int2     DBError;
  CHResult_co001 hr;
  m_beforeImages.subtransCommit (requiredLevel); 
  while (this->CurrentSubtransLevel() >= requiredLevel) {
    if (this->CurrentSubtransLevel() > 1) {
      hr = m_lcSink->SubtransCommit( &DBError);
      if ( 0 != DBError ) { 
        this->ThrowReturnCode (DBError, "omsCommitSubtrans");
      }
    }
    this->DecSubtransLevel();
  }
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::deallocate(void* p) {
  m_heapUsed -= co10_chunkSize(p);
  SharedMemAllocator_co16.deallocate(m_lcSink, p);
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::FreeStackHeap() {
  m_stackHeap.Free();
}

/*----------------------------------------------------------------------*/

inline OmsAbstractObject* OmsSession_co12::ForUpdPtr(OmsObjectContainerPtr pObj) {
  const char* msg = "OmsSession_co12::ForUpdPtr";
  if (m_read_only) {
    this->ThrowDBError (e_file_read_only, msg, pObj->m_oid);
  }
  this->InsertBeforeImage (pObj);
  return &pObj->m_pobj;
}

/*----------------------------------------------------------------------*/

inline tco13_GuidEntry* OmsSession_co12::GetClassInfo(const GUID& guid) {
  return m_context->GetClassInfo (guid);
}

/*----------------------------------------------------------------------*/

inline OmsCallbackInterface* OmsSession_co12::GetCallbackInterface() const {
  return m_callbackInterface;
}

/*----------------------------------------------------------------------*/

inline OmsCallbackInterface* OmsSession_co12::GetCallbackInterface(const GUID& guid) const {
  if (m_callbackInterfaceGuid == guid) {
    return m_callbackInterface;
  }
  else {
    return NULL;
  }
}

/*----------------------------------------------------------------------*/

inline tco13_ClsIdEntry* OmsSession_co12::GetClsInfo(const GUID& guid, OmsSchemaHandle Schema, OmsContainerNo ContainerNo) {
  return m_context->m_containerDir.HashFind (guid, Schema, ContainerNo);
}

/*----------------------------------------------------------------------*/

inline tco13_ClsIdEntry* OmsSession_co12::GetContainerInfo(tsp00_Uint4 containerHandle) {
  return m_context->GetContainerInfo(containerHandle);
}

/*----------------------------------------------------------------------*/

inline OmsObjectContainerPtr OmsSession_co12::GetMemory (tco13_ClsIdEntry& clsid) { 
  OmsObjectContainerPtr p;

  p = clsid.m_freeHead->free;
  if (NULL != p) { 
    clsid.m_freeHead->free = p->m_hashnext;
  }
  else {
    if (this->InVersion()) {
      p = (OmsObjectContainerPtr) m_context->allocate(clsid.GetObjectSize());
    }
    else {
      p = (OmsObjectContainerPtr) StackHeapMalloc(clsid.GetObjectSize());
    }
  }
  if (NULL != p) {
    p->InitObjContainer(clsid);
  }
  return p;
}

/*----------------------------------------------------------------------*/

inline OmsObjectContainerPtr OmsSession_co12::GetMemory (tco13_GuidEntry& classInfo) {

  OmsObjectContainerPtr p;

  p = classInfo.m_freeHead->free;
  if (NULL != p) { 
    classInfo.m_freeHead->free = p->m_hashnext;
  }
  else {
    if (this->InVersion()) {
      p = (OmsObjectContainerPtr) m_context->allocate(classInfo.GetObjectSize());
    }
    else {
      p = (OmsObjectContainerPtr) StackHeapMalloc(classInfo.GetObjectSize());
    }
  }
  if (NULL != p) {
    p->InitObjContainer(classInfo);
  }
  return p;
}

/*----------------------------------------------------------------------*/

inline unsigned char* OmsSession_co12::GetVarObjMemory(size_t sz) {
  m_context->m_varObjHeapUsed += sz;
  this->MonitorVarObjSize(sz);
  return REINTERPRET_CAST(unsigned char*, m_context->allocate(sz));
}

/*----------------------------------------------------------------------*/

inline tsp00_Int4 OmsSession_co12::GetVarObjSize(const OmsVarOid& oid) {

  const char* msg = "OmsContext_co12::GetVarObjSize";
  if (oid == m_currVarObjChunk.m_oid) {
    return m_currVarObjChunk.m_container.m_vobjSize;
  }
  OmsObjectContainerPtr pObj = m_context->FindObjInContext (&oid);
  if (NULL != pObj) {
    if (pObj->DeletedFlag()) {
      ThrowDBError (e_object_not_found, msg, oid);
    }
    OmsVarObjInfo_co12* pObjInfo;
    pObjInfo = REINTERPRET_CAST (OmsVarObjInfo_co12*, &pObj->m_pobj);
    return pObjInfo->m_vobjSize;
  }
  tsp00_Int4 size = m_context->GetVarObjFromLiveCacheBase(oid, false);
  if (size < 0) {
    ThrowDBError (e_object_not_found, msg, oid);
    return 0;
  }
  else {
    return size;
  }
}

/*----------------------------------------------------------------------*/

inline tsp00_8ByteCounter OmsSession_co12::HeapUsed() {
  return m_heapUsed;
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::IncRefCnt() {
  ++m_refCnt;
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::IncDeref() {
  if (NULL != m_monitor_curr) {
    ++m_monitor_curr->cmiCounters_gg00[cmiOmsRead];
  }
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::IncDereflC() {
  if (NULL != m_monitor_curr) {
    if (InVersion()) {
       ++m_monitor_curr->cmiCounters_gg00[cmiLCReadInVersion];
    } 
    else {
      ++m_monitor_curr->cmiCounters_gg00[cmiLCRead];
    }
  }
}
/*----------------------------------------------------------------------*/

inline void OmsSession_co12::IncDerefKey() {
  if (NULL != m_monitor_curr) {
    if (InVersion()) {
      ++m_monitor_curr->cmiCounters_gg00[ cmiLCKeyReadInVersion];
    }
    else {
      ++m_monitor_curr->cmiCounters_gg00[cmiLCKeyRead];
    }
  }
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::IncStore() {
  if (NULL != m_monitor_curr) {
    ++m_monitor_curr->cmiCounters_gg00[cmiOmsStore];
  }
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::IncStore_lC() {
  if (NULL != m_monitor_curr) {
    if (InVersion()) {
      ++m_monitor_curr->cmiCounters_gg00[cmiStoreInVersion];
    }
    else {
      ++m_monitor_curr->cmiCounters_gg00[cmiLCStore];
    }
  }
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::IncDelete () {
  if (m_monitor_curr) {
    ++m_monitor_curr->cmiCounters_gg00[cmiOmsDelete];
  }
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::IncDelete_lC(tsp00_Int4 cntDeleted) {
  if (NULL != m_monitor_curr) {
    if (InVersion()) {
      m_monitor_curr->cmiCounters_gg00[cmiDeleteInVersion] += cntDeleted;
    }
    else {
      m_monitor_curr->cmiCounters_gg00[cmiLCDelete] += cntDeleted;
    }
  }
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::IncDelete_lC() {
  IncDelete_lC(1);
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::IncLock() {
  if (NULL != m_monitor_curr) {
    ++m_monitor_curr->cmiCounters_gg00[cmiLCLock];
  }
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::IncLogHop(int cnt) {
  if (NULL != m_monitor_curr) {
    m_monitor_curr->cmiCounters_gg00[cmiLogHop] += cnt;
  }
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::IncExceptions() {
  if (NULL != m_monitor_curr) {
    ++m_monitor_curr->cmiCounters_gg00[cmiExceptions];
  }
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::IncOutOfDate() {
  if (NULL != m_monitor_curr) {
    ++m_monitor_curr->cmiCounters_gg00[cmiOutOfDate];
  }
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::IncTimeout() {
  if (NULL != m_monitor_curr) {
    ++m_monitor_curr->cmiCounters_gg00[cmiTimeout];
  }
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::IncSubtransCommit() {
  if (NULL != m_monitor_curr) {
    ++m_monitor_curr->cmiCounters_gg00[cmiSubtransCommit];
  }
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::IncSubtransRollback() {
  if (NULL != m_monitor_curr) {
    ++m_monitor_curr->cmiCounters_gg00[cmiSubtransRollback];
  }
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::IncLoadVarObj() {
  if (NULL != m_monitor_curr) {
    ++m_monitor_curr->cmiCounters_gg00[cmiLoadVarObj];
  }
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::IncStoreVarObj() {
  if (NULL != m_monitor_curr) {
    ++m_monitor_curr->cmiCounters_gg00[cmiStoreVarObj];
  }
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::MonitorSubtransLevel() {
  if (NULL != m_monitor_curr) {
    if (m_subtrans_lvl > m_monitor_curr->cmiCounters_gg00[cmiMaxSubtransLevel]) { 
      m_monitor_curr->cmiCounters_gg00[cmiMaxSubtransLevel] = m_subtrans_lvl;
    }
  }
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::SetCacheSize(tsp00_Uint4 sz) {
  if (NULL != m_monitor_curr) {
    m_monitor_curr->cmiCounters_gg00[cmiCacheSize] = sz;
  }
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::IncCreateVersion () {
 if (NULL != m_monitor_curr) {
    ++m_monitor_curr->cmiCounters_gg00[cmiCreateVersion];
  }
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::IncOpenVersion () {
 if (NULL != m_monitor_curr) {
    ++m_monitor_curr->cmiCounters_gg00[cmiOpenVersion];
  }
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::IncCloseVersion () {
 if (NULL != m_monitor_curr) {
    ++m_monitor_curr->cmiCounters_gg00[cmiCloseVersion];
  }
}   

/*----------------------------------------------------------------------*/
        
inline void OmsSession_co12::IncDropVersion () {
 if (NULL != m_monitor_curr) {
    ++m_monitor_curr->cmiCounters_gg00[cmiDropVersion];
  }
}             

/*----------------------------------------------------------------------*/          

inline void OmsSession_co12::IncDeleteVarObject () {
 if (NULL != m_monitor_curr) {
    ++m_monitor_curr->cmiCounters_gg00[cmiDeleteVarObject];
  }
}              


/*----------------------------------------------------------------------*/

inline void OmsSession_co12::IncSubtransLevel () {
  ++m_subtrans_lvl;
  this->MonitorSubtransLevel();
}

/*----------------------------------------------------------------------*/

inline bool OmsSession_co12::InsertBeforeImage (OmsObjectContainerPtr p) {
  int subtransLvl = this->CurrentSubtransLevel();
  if (!(p->existBeforeImage (subtransLvl, m_context->IsVersion()))) {
    if ((subtransLvl > 1) || (this->InVersion())) {
      m_beforeImages.insertBeforeImage (p, subtransLvl);
      return true;
    }
  }
  return false;
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::InsertNewBeforeImage (OmsObjectContainerPtr p, tco13_ClsIdEntry* clsinfo) {
  int subtransLvl = this->CurrentSubtransLevel();
  if ((subtransLvl > 1) || (this->InVersion())) {
    m_beforeImages.insertNewBeforeImage (p, clsinfo, subtransLvl);
  }
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::InsertTransVersion(OmsContext_co12* context) {
  if (!this->VersionBoundByMe(context)) {
    m_versionsBoundToTrans.insert(context);
  }
  context->MarkBoundToTrans();
}

/*----------------------------------------------------------------------*/

inline bool OmsSession_co12::InVersion() {
  return (m_context != m_defaultContext);
}

/*----------------------------------------------------------------------*/

inline bool OmsSession_co12::IsDefaultContext( OmsContext_co12* context) {
  return (context == m_defaultContext);
}

/*----------------------------------------------------------------------*/

inline bool OmsSession_co12::IsSubtransOpen() {
  return (m_subtrans_lvl > 1);
}

/*----------------------------------------------------------------------*/

inline OmsObjectContainerPtr OmsSession_co12::LoadVarObject (const OmsVarOid& oid, VarObjLockType_co12 lockReq, tsp00_Int4 size, void* buf) {
  
  const char* msg = "OmsSession_co12::LoadVarObject";
  
  this->IncLoadVarObj();
  OmsObjectContainerPtr p = m_context->FindVarObjInContext(oid, lockReq, size, buf);
  if (NULL != p) {
    return p;
  }
  tsp00_Int4 objSize = m_context->GetVarObjFromLiveCacheBase(oid, false);
  if ((size < objSize) && (buf != NULL)) {
    this->ThrowDBError (e_too_small_packet_size, msg, oid);
  }
  p = m_context->LoadVarObject(oid, lockReq, objSize, m_currVarObjChunk, buf);
  m_currVarObjChunk.Clear();
  return p;
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::LockObj(OmsObjectContainerPtr p) {
  tgg00_BasisError    DBError;
  CUnknownContainerId_co12 FileId;
  CHResult_co001 hr = m_lcSink->LockObj (
    (unsigned char*) &m_context->m_consistentView,
    (unsigned char*) &FileId, //TODO so lassen ??
    REINTERPRET_CAST(OmsTypeOid*, &p->m_oid),
    (unsigned char*) &p->m_objseq,
    &DBError);
  m_context->LockResult (DBError, p, p->m_oid, "OmsSession_co12::LockObj");
  this->IncLock();
}

/*----------------------------------------------------------------------*/

inline void* OmsSession_co12::allocate(size_t sz) {
  void* p = SharedMemAllocator_co16.allocate(m_lcSink, sz);
  m_heapUsed += co10_chunkSize(p);
  return p;
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::MonitorVarObjSize(tsp00_Int4 size) {
  if (NULL != m_monitor_curr) {
    m_monitor_curr->cmiCounters_gg00[cmiSumVarObjSize] += size;
    ++m_monitor_curr->cmiCounters_gg00[cmiCntVarObj]; 
  }
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::MonitorWaitNewConsistentView(tsp00_Int4 runTime) {
  if ((NULL != m_monitor_curr) && (runTime > 0)) {
    m_monitor_curr->cmiCounters_gg00[cmiSumWaitNewConsView] += runTime;
    ++m_monitor_curr->cmiCounters_gg00[cmiCntWaitNewConsView]; 
    if (runTime > m_monitor_curr->cmiCounters_gg00[cmiMaxWaitNewConsView]) {
      m_monitor_curr->cmiCounters_gg00[cmiMaxWaitNewConsView] = runTime;
    }
    if (runTime < m_monitor_curr->cmiCounters_gg00[cmiMinWaitNewConsView]) {
      m_monitor_curr->cmiCounters_gg00[cmiMinWaitNewConsView] = runTime;
    }
  }
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::OpenVersion(OmsContext_co12* context) {
  short   DBError;
  HRESULT hr;
  m_defaultContext->VersionClearObjCache();
  if (IsDefaultContext(m_context)) {
    /* current context is default context, release current consistent view */
    if (m_context->ExistsConsistentView()) {
      hr = m_lcSink->EndConsistentView(
        (unsigned char*) &m_context->m_consistentView, &DBError);
      if FAILED (hr) {
         throw DbpError (DbpError::HRESULT_ERROR, hr);
      }
      if ( 0 != DBError ) { 
        throw DbpError (DbpError::DB_ERROR, DBError, "OmsSession::OpenVersion");
      }
    }
  }
  m_context = context;
  m_context->OpenVersion(this);
  this->ChangedConsistentView();
  this->InsertTransVersion(context);
}

/*----------------------------------------------------------------------*/


inline void OmsSession_co12::ReleaseObj(const OmsObjectId& oid) {
  OmsObjectContainerPtr found = m_context->FindObjInContext(&oid);
  if (found) {
    this->ReleaseObj(found);
  }
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::ReleaseObj(OmsObjectContainerPtr p) {
  const bool releaseUpdatedObj = ((!m_context->IsVersion()) && (!this->IsSubtransOpen()));
  if (p->DeletedFlag()) {
    if (releaseUpdatedObj) {
      m_context->FlushDelete (p, true);
    }
    return;
  }
  if (p->StoredFlag()) {
    if (releaseUpdatedObj) {
      m_context->FlushObj(p);
    }
    else {
      return;
    }
  }
  else {
    if (p->LockedFlag()) {
      if ((releaseUpdatedObj) && (!p->VarObjFlag())) {
        m_context->FlushObj(p);
      }
      else {
        return;
      }
    }
  }
  m_context->DeleteObjInContext(p, p->GetContainerInfo (m_context));
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::ReleaseObjUnconditional(const OmsObjectId& oid) {
  OmsObjectContainerPtr found = m_context->FindObjInContext(&oid);
  if (found) {
    m_context->DeleteObjInContext(found, found->GetContainerInfo (m_context));
  }
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::RollbackSubtrans(int requiredLvl) {
  tgg00_BasisError DBError;
  CHResult_co001    hr;
  
  if (-1 == requiredLvl) {
    requiredLvl = this->CurrentSubtransLevel();
  }
  m_beforeImages.subtransRollback (requiredLvl); 
  while (this->CurrentSubtransLevel() >= requiredLvl) {
    if (this->CurrentSubtransLevel() > 1) {
      hr = m_lcSink->SubtransRollback( &DBError);
      if ( e_ok != DBError ) { 
        this->ThrowReturnCode (DBError, "RollbackSubtrans");
      }
    }
    this->DecSubtransLevel();
  }
  m_currVarObjChunk.Clear();
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12::ResetSubtransLevel () {
  m_subtrans_lvl = 1;
}

/*----------------------------------------------------------------------*/

inline void* OmsSession_co12::StackHeapMalloc(size_t size) {
  return m_stackHeap.Malloc(size);
}

/*----------------------------------------------------------------------*/

inline int OmsSession_co12::StartSubtrans() {
  
  const char* msg = "OmsSession_co12::StartSubtrans";
  tsp00_Int2  DBError;
  if (this->CurrentSubtransLevel() >= MAX_SUBTRANS_LEVEL) {
    this->ThrowDBError (e_too_many_subtrans, msg);      
  }
  CHResult_co001 hr = m_lcSink->SubtransStart(&DBError);
  if ( 0 != DBError ) { 
    this->ThrowDBError (DBError, msg);
  }
  this->IncSubtransLevel();
  return this->CurrentSubtransLevel();
}

/*----------------------------------------------------------------------*/

inline void OmsSession_co12:: StoreVarObject (const OmsVarOid& oid, const void* pVarObj, unsigned long objLength) {
 
  bool                  freeBeforeImage = true;
  OmsObjectContainerPtr found           = this->LoadVarObject (oid, VarObjNoLock, -1, NULL);
  if (NULL == found) {
      this->ThrowDBError (e_object_not_found, "StoreVarObject", oid);
  }
  if (!found->LockedFlag()) {
    if (!IsLocked(oid)) {
      this->ThrowDBError (e_object_not_locked, "StoreVarObject", oid);
    }
  }
  if (m_currVarObjChunk.m_oid == oid) {
    m_currVarObjChunk.Clear();
  }
  OmsVarObjInfo_co12* objInfo = REINTERPRET_CAST (OmsVarObjInfo_co12*, &found->m_pobj);
  if (!(found->existBeforeImage (this->CurrentSubtransLevel(), this->InVersion()))) { 
    this->ForUpdPtr(found); 
    freeBeforeImage = false;
  }
  if ((objInfo->m_vobjSize != objLength) || (!freeBeforeImage)) {
    if (freeBeforeImage) {
      m_context->FreeVarObj(objInfo->m_pvobj, objInfo->m_vobjSize);
    }
    objInfo->m_pvobj = this->GetVarObjMemory(objLength);
  }
  objInfo->m_vobjSize = objLength;
  memcpy (objInfo->m_pvobj, pVarObj, objLength);
  found->MarkStored();
}

/*----------------------------------------------------------------------*/

inline void  OmsSession_co12::FreeLockObjAndBeforeImages() {
  /* unlock OmsLockObjects */
  OmsLockObject* curr = m_lockObjects;
  while (NULL != curr) {
    curr->m_mode = OmsLockObject::NO_LOCK;
    curr         = curr->m_next;
  }
  /* free before images */
  m_beforeImages.freeBeforeImages();
  m_subtrans_lvl = 1;
}
   
#endif  // __HCO12_H
