/* @lastChanged: "1999-04-21  8:45"

  * @filename:   hco13.h
  * @purpose:    "C++ Persistent Object Hashing"
  * @release:    7.2.0.0
  * @see:        "http://www.bea.sap-ag.de/doku/rel-7.1/hco13.html"
  *
  * @copyright:  "Copyright by   SAP AG, 1998"


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/

#if !defined (__HCO13_H)
#define __HCO13_H

#include "ilivecachesink.h" // no check
#include "ggg00.h"
#include "ggg92.h"
#include "gco001.h"
#include "gco102.h"
#include "hco18.h"
#include "ggg250.h"

#define MAX_SUBTRANS_LEVEL 32

class OmsAbstractObject;
class OmsHandle;
class OmsObjectContainer;
class OmsObjByClsIterBase;
class OmsStream;
class LogEntry_co10;
class OmsSession_co12;
class OmsContext_co12;

typedef short tco13_ClsId;
typedef OmsObjectContainer* OmsObjectContainerPtr; 

struct tco13_FreeListHeader {
  tco13_FreeListHeader* next;
  OmsObjectContainerPtr free;
  size_t                size;
  inline tco13_FreeListHeader(tsp00_Int4 sz);
  inline void* operator new(size_t sz, OmsContext_co12* context);
#if defined(PLACEMENT_DELETE_MEO00)
  inline void operator delete(void* p, OmsContext_co12* context);
#endif
  inline void DeleteSelf(OmsContext_co12* context);
  void operator delete(void*p); // not implemented, attempt to call will force linker error
};

struct tco13_GuidEntry {
  tco13_GuidEntry*      m_hashNext;
  ClassInfo_co18*       m_classInfo;
  tco13_FreeListHeader* m_freeHead;
  tco13_GuidEntry(tco13_GuidEntry* next, ClassInfo_co18* ci, tco13_FreeListHeader* fl) :
  m_hashNext(next),  m_classInfo(ci), m_freeHead(fl) {} 
  inline void* operator new (size_t sz, OmsContext_co12* context);
  void          ChainFree (OmsContext_co12&c, OmsAbstractObject* p);
  void          ChainFree (OmsContext_co12&c, OmsObjectContainerPtr& p);
  inline void   DeleteSelf(OmsContext_co12* context);
  inline const GUID& GetBaseGuid() const;
  inline int    GetKeyLen();
  inline size_t GetObjectSize();
  inline void*  GetVirtualTablePtr();
  inline bool   IsBaseClass() const;
  inline bool   IsDerivedClass() const;
  inline bool   IsDerivedClassOf(const GUID& guid) const;
  inline bool   IsVarObject();
#if defined(PLACEMENT_DELETE_MEO00)
  inline void operator delete(void* p, OmsContext_co12* context);
#endif
  void operator delete(void*p); // not implemented, attempt to call will force linker error
};

typedef tco13_GuidEntry* tco13_GuidEntryPtr;

class tco13_ClassHash : public IGuidHash_co001 {
public :
#define GUID_HASH_ENTRIES_CO13 49
  
  tco13_ClassHash();
  ~tco13_ClassHash();
  void Create(OmsContext_co12* context);
  void HashInsert(ClassInfo_co18* pClassInfo, tco13_FreeListHeader* freeHead);
  tco13_GuidEntryPtr HashFind(const GUID& guid);
private :
  OmsContext_co12*   m_context;
  tco13_GuidEntryPtr m_head[GUID_HASH_ENTRIES_CO13];
};

struct tco13_ClsIdEntry {
  tco13_ClsIdEntry*                                               m_clsidHashNext;
  tco13_ClsIdEntry*                                               m_guidHashNext;
  ContainerInfo_co18*                                             m_containerInfo;
  cgg250AvlTree<unsigned char*, tco13_ClsIdEntry,OmsContext_co12> m_index;
  tco13_FreeListHeader*                            m_freeHead;
  tco13_ClsIdEntry (ContainerInfo_co18* classInfo,OmsContext_co12* c);
  inline void* operator new    (size_t sz, OmsContext_co12* session);
#if defined(PLACEMENT_DELETE_MEO00)
  inline void  operator delete (void* p, OmsContext_co12* session);
#endif
  inline int                     Compare(const unsigned char*, const unsigned char*);
  inline void                    DeleteSelf(OmsContext_co12* session);
  inline const ClassInfo_co18*   GetClassInfoPtr() const;
  inline ContainerInfo_co18*     GetContainerInfoPtr() const;
  inline tsp00_Uint4             GetContainerHandle() const;
  inline OmsContainerNo          GetContainerNo() const;
  inline OmsSchemaHandle         GetSchema() const;
  inline tgg00_ContainerId&      GetFileId() const;
  inline const GUID&             GetGuid() const;
  inline const tgg00_ObjColDesc& GetInvColDesc() const;
  inline const tgg00_StackEntry& GetKeyDesc() const;
  inline int                     GetKeyLen() const;
  inline unsigned char*          GetKeyPtr(OmsObjectContainerPtr p);
  inline tsp00_Uint4             GetObjectSize() const;
  inline tsp00_Int4              GetPersistentSize() const;
  inline void*                   GetVirtualTablePtr() const;
  inline bool                    IsDerivedClassOf(const GUID&) const;
  inline bool                    IsVarObject();
  void                           chainFree (OmsContext_co12&c, OmsAbstractObject* p);
  void                           chainFree (OmsContext_co12&c, OmsObjectContainerPtr& p);
  void                           VersionAddKey(OmsObjectContainerPtr);
  void                           VersionDelKey(OmsObjectContainerPtr);
  void                           VersionDelIndex();
  OmsObjectContainer*            VersionFindKey(const unsigned char*);
private :
  void operator delete(void*);
};

typedef tco13_ClsIdEntry* tco13_ClsIdEntryPtr;

class cco13_ClsIdHash : public IGuidHash_co001 {
	tsp00_Int4            headentries;
	tco13_ClsIdEntryPtr*  m_clsidHead;
  tco13_ClsIdEntryPtr*  m_guidHead;
  tco13_FreeListHeader* freeListHeader;
  OmsContext_co12*      m_context;
  tco13_ClassHash       m_classDir;
public :
  class Iter {
    friend class cco13_ClsIdHash;
    cco13_ClsIdHash*    hash;
    long                headIndex;
    tco13_ClsIdEntryPtr curr;
  public :
    inline Iter(cco13_ClsIdHash*);
    inline Iter(const Iter&);
    inline void operator++();
    inline operator bool() const;
    inline tco13_ClsIdEntryPtr operator()() const;
  };
	cco13_ClsIdHash  ();
	~cco13_ClsIdHash ();
  tco13_ClsIdEntry*        AutoRegister (const GUID& guid, OmsSchemaHandle sh, OmsContainerNo ContainerNo);
  inline void              CheckIfNotDropped(tco13_ClsIdEntry* curr);
  void                     Clean();                    
  void                     Create(OmsContext_co12* context);
  inline Iter              First ();
  inline tco13_GuidEntry*  GetClassInfo(const GUID guid);
	inline void              HashInsert (tco13_ClsIdEntry* h);
	inline tco13_ClsIdEntry* HashFindViaContainerHandle (tsp00_Uint4 containerHandle, bool checkExistence = true);
  inline tco13_ClsIdEntry* HashFind   (const GUID& k, OmsSchemaHandle, OmsContainerNo);
         void              HashDelete (tco13_ClsIdEntry* containerInfo);
         bool              IsBaseClassOf(const GUID&, const GUID&);
         bool              IsDerivedClassOf(const GUID& , const GUID&);
  inline void              ClearFreeLists();
  tco13_FreeListHeader*    GetFreeListHeader(size_t size);
  void                     ThrowUnknownContainer(tco13_ClsIdEntryPtr);
  void                     ThrowUnknownContainer(const GUID& guid, OmsSchemaHandle Schema, OmsContainerNo ContainerNo);
  friend class cco13_ClsIdHash::Iter;
};

class cco13_OidHash {
friend class OmsHandle;
friend class OmsObjByClsIterBase;
	tsp00_Int4             m_headentries;
  OmsContext_co12*       m_context;
 	OmsObjectContainerPtr* m_head;
	OmsObjectContainerPtr* m_headcurr;
public :
	cco13_OidHash ();
	~cco13_OidHash ();
  void                         Clean();
  void                         Clear();
  tgg00_BasisError             ClearForOpenVersion();
  void                         Create (OmsContext_co12* s, const tsp00_Int4 sz = 10003);
  inline void                  HashFree       ();
  inline void                  HashInit       (tsp00_Int4 sz, bool throwError = true);
	inline void                  HashInsert     (OmsObjectContainer* h);
  inline void                  HashSlotInsert (OmsObjectContainer* h);
	inline OmsAbstractObject*    HashFind       (const OmsObjectId& k);
  inline OmsObjectContainerPtr HashFind       (const OmsObjectId* k);
	void                         HashDelete     (const OmsObjectId& key);
  void                         HashResize     (tsp00_Int4 entryCount);
  void                         HashCheck      ();
  void                         CheckChain     (OmsObjectContainerPtr curr);
  void                         RemoveUnlockedObjFromCache();
  void                         SetEmpty       ();
  class OidIter {
    friend class cco13_OidHash;
    cco13_OidHash*         m_hash;
    long                   m_headIndex;
    OmsObjectContainerPtr  m_curr;
  public :
    inline OidIter(cco13_OidHash*);
    inline OidIter(const OidIter&);
    inline void operator++();
    inline void operator++(int);
    inline operator bool() const;
    inline OmsObjectContainerPtr operator()() const;
  };
  friend class cco13_OidHash::OidIter;
  inline OidIter               First ();
private :
  OmsObjectContainerPtr* HeadPtr (const OmsObjectId& key) {
    tsp00_Uint4 idx = ((key.getPno() << 13) + key.getPos()) % m_headentries;
    m_headcurr = &m_head[idx];
    return m_headcurr;
  }
};


class cco13_BeforeImages {
public :
  cco13_BeforeImages ();
  ~cco13_BeforeImages();
  void freeBeforeImages     ();
  void init(OmsSession_co12* c);
  void insertBeforeImage    (OmsObjectContainerPtr, int subtransLvl);
  void insertNewBeforeImage (OmsObjectContainerPtr, tco13_ClsIdEntry* clsinfo, int subtransLvl);
  void removeContext        (OmsContext_co12* pContext);
  void subtransRollback     (int subtransLvl);
  void subtransCommit       (int subtransLvl);
private :
  OmsObjectContainerPtr m_beforeImages[MAX_SUBTRANS_LEVEL]; 
  OmsSession_co12*      m_session;
};

#endif // __HCO013_H

#if defined (__HCO13_IMPL)
#if !defined (__HCO13_H_I)
#define __HCO13_H_I

inline tco13_FreeListHeader::tco13_FreeListHeader(tsp00_Int4 sz) : next(NULL), size(sz), free(NULL) {
}

/*----------------------------------------------------------------------*/

inline void* tco13_FreeListHeader::operator new(size_t sz, OmsContext_co12* context) {
  return context->allocate(sz);
}

/*----------------------------------------------------------------------*/

#if defined(PLACEMENT_DELETE_MEO00)
inline void tco13_FreeListHeader::operator delete (void* p, OmsContext_co12* context) {
  context->deallocate(p);
}
#endif

/*----------------------------------------------------------------------*/

inline void tco13_FreeListHeader::DeleteSelf (OmsContext_co12* context) {
  context->deallocate(this);
}

/*----------------------------------------------------------------------*/

inline void* tco13_GuidEntry::operator new (size_t sz, OmsContext_co12* context) {
  return context->allocate(sz);
}

/*----------------------------------------------------------------------*/

inline void tco13_GuidEntry::DeleteSelf(OmsContext_co12* context) {
  context->deallocate(this);
}

/*----------------------------------------------------------------------*/

inline const GUID& tco13_GuidEntry::GetBaseGuid() const {
  return m_classInfo->GetBaseGuid();
}

/*----------------------------------------------------------------------*/

inline int tco13_GuidEntry::GetKeyLen() {
  return m_classInfo->GetKeyDesc().elen_var();
}

/*----------------------------------------------------------------------*/

inline size_t tco13_GuidEntry::GetObjectSize() {
  return m_classInfo->GetObjectSize();
}

/*----------------------------------------------------------------------*/

inline void* tco13_GuidEntry::GetVirtualTablePtr() {
  return m_classInfo->GetVTPtr();
}

/*----------------------------------------------------------------------*/

inline bool tco13_GuidEntry::IsBaseClass() const {
  return m_classInfo->IsBaseClass();
}

/*----------------------------------------------------------------------*/

inline bool tco13_GuidEntry::IsDerivedClass() const {
  return m_classInfo->IsDerivedClass();
}

/*----------------------------------------------------------------------*/

inline bool tco13_GuidEntry::IsDerivedClassOf(const GUID& guid) const {
  return m_classInfo->IsDerivedClassOf(guid);
}

/*----------------------------------------------------------------------*/

inline bool tco13_GuidEntry::IsVarObject() {
  return m_classInfo->IsVarObject();
}

/*----------------------------------------------------------------------*/

#if defined(PLACEMENT_DELETE_MEO00)
inline void tco13_GuidEntry::operator delete(void* p, OmsContext_co12* context) {
  context->deallocate(p);
}
#endif

/*----------------------------------------------------------------------*/

inline void* tco13_ClsIdEntry::operator new (size_t sz, OmsContext_co12* context) {
  return context->allocate(sz);
}

/*----------------------------------------------------------------------*/

#if defined(PLACEMENT_DELETE_MEO00)
inline void tco13_ClsIdEntry::operator delete (void* p, OmsContext_co12* context) {
  context->deallocate(p);
}
#endif

/*----------------------------------------------------------------------*/

inline int tco13_ClsIdEntry::Compare(const unsigned char* l, const unsigned char* r) {
  int cmp = memcmp (l, r, GetKeyLen());
  if (0 != cmp) {
    if (cmp > 0) {
      return 1;
    }
    else {
      return -1;
    }
  }
  else {
    return 0;
  }
}

/*----------------------------------------------------------------------*/

inline void tco13_ClsIdEntry::DeleteSelf(OmsContext_co12* context) {
  // release reference to global container dictionary 
  co10_DetachContainerInfo(context->LcSink(), m_containerInfo);
  // destroy self
  context->deallocate(this);
}

/*----------------------------------------------------------------------*/

inline const ClassInfo_co18* tco13_ClsIdEntry::GetClassInfoPtr() const {
  return m_containerInfo->GetClassInfoPtr();
}

/*----------------------------------------------------------------------*/

inline ContainerInfo_co18* tco13_ClsIdEntry::GetContainerInfoPtr() const {
  return m_containerInfo;
}

/*----------------------------------------------------------------------*/

inline tsp00_Uint4 tco13_ClsIdEntry::GetContainerHandle() const {
  return m_containerInfo->GetContainerHandle();
}

/*----------------------------------------------------------------------*/

inline OmsContainerNo tco13_ClsIdEntry::GetContainerNo() const {
  return m_containerInfo->m_ContainerNo;
}

/*----------------------------------------------------------------------*/

inline tgg00_FileId& tco13_ClsIdEntry::GetFileId() const { 
  return m_containerInfo->m_ContainerId;
}

/*----------------------------------------------------------------------*/

inline const GUID& tco13_ClsIdEntry::GetGuid() const {
  return m_containerInfo->GetGuid();
}

/*----------------------------------------------------------------------*/

inline const tgg00_ObjColDesc& tco13_ClsIdEntry::GetInvColDesc() const {
  return m_containerInfo->GetInvColDesc();
}

/*----------------------------------------------------------------------*/

inline const tgg00_StackEntry& tco13_ClsIdEntry::GetKeyDesc() const {
  return m_containerInfo->GetKeyDesc();
}

/*----------------------------------------------------------------------*/

inline int tco13_ClsIdEntry::GetKeyLen() const {
  return m_containerInfo->GetKeyDesc().elen_var();
}

/*----------------------------------------------------------------------*/

inline unsigned char* tco13_ClsIdEntry::GetKeyPtr(OmsObjectContainerPtr p) {
  return (((unsigned char*) &p->m_pobj) + /* vtblptr */ sizeof (void*) + GetKeyDesc().epos() - 1);
}

/*----------------------------------------------------------------------*/

inline tsp00_Uint4 tco13_ClsIdEntry::GetObjectSize() const {
  return m_containerInfo->GetObjectSize();
}

/*----------------------------------------------------------------------*/

inline tsp00_Int4 tco13_ClsIdEntry::GetPersistentSize() const {
  return m_containerInfo->m_PersistentSize;
}

/*----------------------------------------------------------------------*/

inline OmsContainerNo tco13_ClsIdEntry::GetSchema() const {
  return m_containerInfo->m_schema;
}

/*----------------------------------------------------------------------*/

inline void* tco13_ClsIdEntry::GetVirtualTablePtr() const {
   return m_containerInfo->GetVirtualTablePtr();
}

/*----------------------------------------------------------------------*/

inline bool tco13_ClsIdEntry::IsDerivedClassOf(const GUID& guid) const {
  return m_containerInfo->m_clsInfo->IsDerivedClassOf(guid);
}

/*----------------------------------------------------------------------*/

inline bool tco13_ClsIdEntry::IsVarObject() {
  return m_containerInfo->m_clsInfo->IsVarObject();
}

/*----------------------------------------------------------------------*/
/* Implementation of inline methods of class cco13_OidHash              */
/*----------------------------------------------------------------------*/

inline OmsAbstractObject* cco13_OidHash::HashFind (const OmsObjectId& k) {

  OmsObjectContainer* curr = *HeadPtr(k);
  while (curr) {
    if (curr->m_oid == k)
			return &curr->m_pobj;
		else
			curr = curr->m_hashnext;
  }
  return NULL;
};

/*----------------------------------------------------------------------*/

inline OmsObjectContainerPtr cco13_OidHash::HashFind (const OmsObjectId* k) {
	
	OmsObjectContainer* curr = *HeadPtr(*k);

	while (curr)
		if (curr->m_oid == *k)
			return curr;
    else {
			curr = curr->m_hashnext;
    }
	return NULL;
};

/*----------------------------------------------------------------------*/

inline void cco13_OidHash::HashInsert(OmsObjectContainer* h) {

  h->m_hashnext = *HeadPtr(h->m_oid);
	*m_headcurr     = h;
#ifdef _ASSERT_OMS
  HashCheck();
#endif
};

/*----------------------------------------------------------------------*/

inline void cco13_OidHash::HashSlotInsert (OmsObjectContainer* h) {
	h->m_hashnext = *m_headcurr;
	*m_headcurr     = h;
#ifdef _ASSERT_OMS
  HashCheck();
#endif
};

/*----------------------------------------------------------------------*/

inline cco13_OidHash::OidIter cco13_OidHash::First ()
{
  OidIter iter(this);
  for (iter.m_headIndex = 0; iter.m_headIndex < m_headentries; ++iter.m_headIndex) {
    iter.m_curr = this->m_head[iter.m_headIndex];
    if (NULL != iter.m_curr) {
      break;
    }
  }
  return iter;
};

/*----------------------------------------------------------------------*/

inline cco13_OidHash::OidIter::OidIter(cco13_OidHash* h) : m_hash(h), m_curr(NULL), m_headIndex(-1) {
}

/*----------------------------------------------------------------------*/

inline cco13_OidHash::OidIter::OidIter(const OidIter& iter) {
  *this = iter;
}

/*----------------------------------------------------------------------*/

inline void cco13_OidHash::OidIter::operator++() {
  if (NULL != m_curr) {
    m_curr = m_curr->m_hashnext;
  }
  if (NULL == m_curr) {
    ++m_headIndex; 
    while (m_headIndex < m_hash->m_headentries) {
      m_curr = m_hash->m_head[m_headIndex];
      if (NULL != m_curr) {
        break;
      }
      ++m_headIndex;
    }
  }
}

/*----------------------------------------------------------------------*/

inline void cco13_OidHash::OidIter::operator++(int dummy) {
  operator++();
}

/*----------------------------------------------------------------------*/

inline cco13_OidHash::OidIter::operator bool() const {
  return (m_curr != NULL);
}

/*----------------------------------------------------------------------*/

inline OmsObjectContainerPtr cco13_OidHash::OidIter::operator()() const {
  return m_curr;
}

/*----------------------------------------------------------------------*/
/* Implementation of class cco13_ClsIdHash inline methods               */
/*----------------------------------------------------------------------*/

inline void cco13_ClsIdHash::CheckIfNotDropped(tco13_ClsIdEntry* curr) {
  if (curr->m_containerInfo->IsDropped()) {
    this->ThrowUnknownContainer(curr);
  }
}

/*----------------------------------------------------------------------*/

inline void cco13_ClsIdHash::ClearFreeLists() {
  tco13_FreeListHeader* p = freeListHeader;
  while (p != NULL) {
    if (m_context->IsVersion()) {
      OmsObjectContainerPtr curr = p->free;
      OmsObjectContainerPtr toFree;
      while (NULL != curr) {
        toFree = curr;
        curr   = curr->m_hashnext;
        m_context->deallocate(toFree);
      }
    }
    p->free = NULL;
    p = p->next;
  }
}

/*----------------------------------------------------------------------*/

inline void cco13_ClsIdHash::HashInsert (tco13_ClsIdEntry* h)
{
  int hashSlot;
  hashSlot              = (h->GetContainerHandle()) % headentries;
  h->m_clsidHashNext    = m_clsidHead[hashSlot];
  m_clsidHead[hashSlot] = h;

  hashSlot = HashValue(h->GetGuid(), h->GetSchema(), h->GetContainerNo(), headentries);
  h->m_guidHashNext    = m_guidHead[hashSlot];
  m_guidHead[hashSlot] = h;

  h->m_freeHead = this->GetFreeListHeader(h->GetObjectSize()); 
  m_classDir.HashInsert(h->m_containerInfo->m_clsInfo, h->m_freeHead);
};

/*----------------------------------------------------------------------*/

inline tco13_ClsIdEntry* cco13_ClsIdHash::HashFindViaContainerHandle (tsp00_Uint4 containerHandle, bool checkExistence) 
{
  int hashSlot           = containerHandle % headentries;
  tco13_ClsIdEntry* curr = m_clsidHead[hashSlot];
  while (curr != NULL) {
    if (curr->GetContainerHandle() == containerHandle) {
      if (checkExistence) {
        CheckIfNotDropped(curr);
      }
      return curr;
    }
    curr = curr->m_clsidHashNext;
  }
  ContainerInfo_co18* pContainerInfo = co10_FindContainerInfo(m_context->LcSink(), containerHandle);
  if (NULL != pContainerInfo) {
    curr = new(m_context) tco13_ClsIdEntry (pContainerInfo, m_context);
    if (NULL != curr) {
      this->HashInsert (curr);
    }
    return curr;
  }
  OMS_TRACE(omsTrAlways, m_context->LcSink(), "cco13_ClsIdHash::HashFindViaContainerHandle : " << containerHandle << " not found");
  throw DbpError (DbpError::DB_ERROR, e_unknown_guid);
  return NULL;
};

/*----------------------------------------------------------------------*/

inline tco13_ClsIdEntry* cco13_ClsIdHash::HashFind (const GUID& guid, OmsSchemaHandle sh, OmsContainerNo ContainerNo) {
  int hashSlot = HashValue(guid, sh, ContainerNo, headentries);
  tco13_ClsIdEntry* curr = m_guidHead[hashSlot];
  while (curr != NULL) {
    if ((curr->GetGuid() == guid) && (curr->GetSchema() == sh) && (curr->GetContainerNo() == ContainerNo)) {
      CheckIfNotDropped(curr);
      return curr;
    }
    curr = curr->m_guidHashNext;
  }
  return this->AutoRegister(guid, sh, ContainerNo);
}

/*----------------------------------------------------------------------*/

inline cco13_ClsIdHash::Iter cco13_ClsIdHash::First ()
{
  Iter iter(this);
  for (iter.headIndex = 0; iter.headIndex < headentries; ++iter.headIndex) {
    iter.curr = m_clsidHead[iter.headIndex];
    if (NULL != iter.curr) {
      break;
    }
  }
  return iter;
};

/*----------------------------------------------------------------------*/

inline tco13_GuidEntry* cco13_ClsIdHash::GetClassInfo(const GUID guid) {
  tco13_GuidEntry* found = m_classDir.HashFind(guid);
  if (NULL == found) {
    ClassInfo_co18* pClassInfo = co10_FindClassInfo(m_context->LcSink(), guid);
    if (NULL != pClassInfo) {
      tco13_FreeListHeader* freeHead = this->GetFreeListHeader(pClassInfo->GetObjectSize()); 
      m_classDir.HashInsert(pClassInfo, freeHead);
      found = m_classDir.HashFind(guid); 
    }
    else {
      OMS_TRACE(omsTrAlways, m_context->LcSink(), "cco13_ClsIdHash::GetClassInfo : " << guid << " not found");
      throw DbpError (DbpError::DB_ERROR, e_unknown_guid);
    }
  }
  return found;
}

/*----------------------------------------------------------------------*/

inline cco13_ClsIdHash::Iter::Iter(cco13_ClsIdHash* h) : hash(h), curr(NULL), headIndex(0) {
}

/*----------------------------------------------------------------------*/

inline cco13_ClsIdHash::Iter::Iter(const Iter& iter) {
  *this = iter;
}

/*----------------------------------------------------------------------*/

inline void cco13_ClsIdHash::Iter::operator++() {
  if (NULL != curr) {
    curr = curr->m_clsidHashNext;
  }
  if (NULL == curr) {
    ++headIndex; 
    while (headIndex < hash->headentries) {
      curr = hash->m_clsidHead[headIndex];
      if (NULL != curr) {
        break;
      }
      ++headIndex;
    }
  }
}

/*----------------------------------------------------------------------*/

inline cco13_ClsIdHash::Iter::operator bool() const {
  return (curr != NULL);
}

/*----------------------------------------------------------------------*/

inline tco13_ClsIdEntryPtr cco13_ClsIdHash::Iter::operator()() const {
  return curr;
}
#endif // __HCO13_H_I
#endif // __HCO13_IMPL



