/*!================================================================
 module:        hco17.h
 
 responsible:   ThomasA
 
 special area:  OMS
 
 description:   OmsVersion Dictionary   
 
 see:
 
 copyright:     SAP AG 1998 


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 ===================================================================*/

#ifndef HCO17_H
#define HCO17_H


#include "hco12.h"
#include "hco16.h"
#include "ggg250.h"
#include "ggg251.h"

class CVersionId_co17 {
public :
  CVersionId_co17(const OmsVersionId& versionId) {
    memcpy (m_versionId, versionId, sizeof(versionId));
  }
#if defined(HPUX) || defined(SUN)
  CVersionId_co17(const OmsTypeChar* pVersionId) {
    memcpy (m_versionId, pVersionId, sizeof(m_versionId));
  }
#endif
  CVersionId_co17(const CVersionId_co17& versionId) {
    memcpy (this, &versionId, sizeof(versionId));
  }
  operator OmsVersionId&() {
    return m_versionId;
  }
  int Compare(const CVersionId_co17& versionId) const {
    return memcmp(m_versionId, versionId.m_versionId, sizeof(m_versionId));
  }
  ~CVersionId_co17() {}
private :
  OmsVersionId m_versionId;
};

class CCompareVersionId_co17 {
public :
  int Compare(const CVersionId_co17& v1, const CVersionId_co17& v2) {
    int cmp = v1.Compare(v2);
    if (0 != cmp) {
      if (cmp > 0) {
        return 1;
      }
      else {
        return -1;
      }
    }
    else {
      return 0;
    }
  }
};

class CVersionDictionary_co17 {
public :
  CVersionDictionary_co17() : m_VersionDictionary(), m_UnloadableVersions(&SharedMemAllocator_co16) {
    m_VersionDictionary.AdviseAllocator(&SharedMemAllocator_co16);
  }
  OmsContext_co12* FindVersion  (const CVersionId_co17&);
	tgg00_BasisError InsertVersion(const CVersionId_co17&, OmsContext_co12*); 
  tgg00_BasisError DropVersion  (const CVersionId_co17&);
  void             MarkUnloadable(OmsContext_co12* pContext);
  void             MarkNotUnloadable(IliveCacheSink* lcSink, OmsContext_co12* pContext);
  void             Shutdown();
  bool             UnloadOldestVersion(IliveCacheSink* lcSink);
  cgg250AvlInfoTree<CVersionId_co17,OmsContext_co12*,CCompareVersionId_co17,co16_SharedMemAllocator>::Iterator First();
private :
	cgg250AvlInfoTree<CVersionId_co17,OmsContext_co12*,CCompareVersionId_co17,co16_SharedMemAllocator> m_VersionDictionary;
  cgg251DCList<OmsContext_co12*, co16_SharedMemAllocator>                                            m_UnloadableVersions;
};

/*--------------------------------------------------------------------------*/
/*       Implementation of  CVersionDirctionary_co17                        */
/*--------------------------------------------------------------------------*/

tgg00_BasisError CVersionDictionary_co17::DropVersion  (const CVersionId_co17& versionId) {
  return m_VersionDictionary.Delete (versionId);
}

OmsContext_co12* CVersionDictionary_co17::FindVersion (const CVersionId_co17& versionId) {
  OmsContext_co12** found = m_VersionDictionary.Find(versionId);
  if (NULL != found) {
    return *found;
  }
  else {
    return NULL;
  }
}

cgg250AvlInfoTree<CVersionId_co17,OmsContext_co12*,CCompareVersionId_co17,co16_SharedMemAllocator>::Iterator CVersionDictionary_co17::First() {
  return m_VersionDictionary.First();
}

tgg00_BasisError CVersionDictionary_co17::InsertVersion(const CVersionId_co17& versionId,OmsContext_co12* pContext) {
  tgg00_BasisError rc;
  OmsContext_co12** p;
  p = m_VersionDictionary.Insert(versionId, rc);
  if (NULL != p) {
    *p = pContext;
  }
  return rc;
}

void CVersionDictionary_co17::Shutdown() {
  m_VersionDictionary.DeleteAll();
  m_UnloadableVersions.remove_all();
}

bool CVersionDictionary_co17::UnloadOldestVersion(IliveCacheSink* lcSink) {
  if (!m_UnloadableVersions.empty()) {
    OmsContext_co12* pContext = m_UnloadableVersions.back();
    m_UnloadableVersions.pop_back();
    pContext->AssignLcSink(lcSink);
    pContext->UnLoad();
    pContext->ResetLcSink();
    return true;
  }
  else {
    return false;
  }
}

void CVersionDictionary_co17::MarkNotUnloadable(IliveCacheSink* lcSink, OmsContext_co12* pContext) {
  lcSink->EnterCriticalSection(RGN_UNLOAD_VERSION);
  for (cgg251dclIterator<OmsContext_co12*, co16_SharedMemAllocator>iter = m_UnloadableVersions.begin(); iter; ++iter) {
    if (*iter() == pContext) {
      m_UnloadableVersions.remove(iter);
      break;
    }
  }
  lcSink->LeaveCriticalSection(RGN_UNLOAD_VERSION);
}

void CVersionDictionary_co17::MarkUnloadable(OmsContext_co12* pContext) {
  CriticalSection_co12 region(pContext->m_session, RGN_UNLOAD_VERSION);
  region.Enter();
  m_UnloadableVersions.push_front(pContext);
}

#endif