/*!================================================================
 module:        hco18.h
 
 responsible:   ThomasA
 
 special area:  OMS
 
 description:   Oms Class and Container Dictionary   
 
 see:
 
 copyright:     SAP AG 1998 


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 ===================================================================*/

#ifndef HCO18_H
#define HCO18_H

#include "ilivecachesink.h" // no check
#include "ggg00.h"
#include "gco01.h"
#include "gco001.h"
#include "gco102.h"

struct ClassInfo_co18 {
public :
  ClassInfo_co18(const char* ClassName, 
                 const GUID& Guid,
                 const GUID* pBaseClass,
                 const tgg00_ObjColDesc& InvColDesc,
                 size_t ObjSize, 
                 void* vtptr,
                 ClassInfo_co18* HashNext);
  void* operator new(size_t sz, IliveCacheSink* lcSink);
  void           DeleteSelf(IliveCacheSink* lcSink);
#if defined(PLACEMENT_DELETE_MEO00)
  void  operator delete (void* p, IliveCacheSink* lcSink);
#endif
  bool                    IsBaseClass()   const { return m_isBaseClass; }
  bool                    IsDerivedClass() const { return m_isDerivedClass; }
  bool                    IsDerivedClassOf(const GUID& guid) const { return (m_isDerivedClass && (m_baseGuid == guid)); }
  bool                    IsVarObject()   const { return m_isVarObject; }
  const GUID&             GetBaseGuid()   const { return m_baseGuid; }
  const char*             GetClassName()  const { return REINTERPRET_CAST(const char*, &m_ClassName[0]); }
  const GUID&             GetGuid()       const { return m_guid; }
  const tgg00_ObjColDesc& GetInvColDesc() const { return m_InvColDesc; }
  const tgg00_StackEntry& GetKeyDesc()    const { return m_KeyDesc; }
  size_t                  GetObjectSize() const { return m_ObjectSize; }
  void*                   GetVTPtr()      const { return m_VTPtr; }
  void                    SetBaseClass()        { m_isBaseClass = true; }
  ClassInfo_co18*         m_HashNext;
private :
  GUID                 m_guid;
  GUID                 m_baseGuid;
  tsp00_C64            m_ClassName;
  tgg00_ObjColDesc     m_InvColDesc;
  tgg00_StackEntry     m_KeyDesc; 
  size_t               m_ObjectSize;
  void*                m_VTPtr;
  bool                 m_isBaseClass;
  bool                 m_isDerivedClass;
  bool                 m_isVarObject;
};

struct ContainerInfo_co18 {
public :
  ContainerInfo_co18*  m_clsidHashNext;
  ContainerInfo_co18*  m_guidHashNext;
  ClassInfo_co18*      m_clsInfo;
  OmsContainerNo       m_ContainerNo;
  OmsSchemaHandle      m_schema;
  tgg00_ContainerId    m_ContainerId;
  size_t               m_PersistentSize;
  bool                 m_dropped;
  short                m_refCnt;
  ContainerInfo_co18 ( ClassInfo_co18* pClassInfo, const tgg00_ContainerId&, tsp00_Uint4 containerHandle, 
    size_t psz, OmsSchemaHandle sh, OmsContainerNo cno); 
  void* operator new(size_t sz, IliveCacheSink* lcSink);
  void           DeleteSelf(IliveCacheSink* lcSink);
#if defined(PLACEMENT_DELETE_MEO00)
  void  operator delete (void* p, IliveCacheSink* lcSink);
#endif
  inline short            DecRefCnt();
  const ClassInfo_co18*   GetClassInfoPtr()    const { return m_clsInfo; }
  tsp00_Uint4             GetContainerHandle() const { return m_ContainerId.fileObjFileNo_gg00(); }
  const GUID&             GetGuid()            const { return m_clsInfo->GetGuid(); }
  const tgg00_ObjColDesc& GetInvColDesc()      const { return m_clsInfo->GetInvColDesc(); }
  const tgg00_StackEntry& GetKeyDesc()         const { return m_clsInfo->GetKeyDesc(); }  
  size_t                  GetObjectSize()      const { return m_clsInfo->GetObjectSize(); }
  OmsSchemaHandle         GetSchema()          const { return m_schema; }
  void*                   GetVirtualTablePtr() const { return m_clsInfo->GetVTPtr(); }
  void                    IncRefCnt()          { ++m_refCnt; }
  inline void             IncRefCnt(IliveCacheSink* lcSink);
  bool                    IsDropped()          const { return m_dropped; } 
  void                    MarkDropped()        { m_dropped = true; }
};

typedef ContainerInfo_co18* pContainerInfo_co18;
typedef ClassInfo_co18*     pClassInfo_co18;

class CContainerDictionary_co18 : public IGuidHash_co001 {
  enum { MaxHeadEntries = 2013 };
  tsp00_Uint4         m_dropId;
  pClassInfo_co18     m_classInfoHead[MaxHeadEntries];
  pContainerInfo_co18 m_clsidHead[MaxHeadEntries];
  pContainerInfo_co18 m_guidHead [MaxHeadEntries];
  ClassInfo_co18*     FindGuid (const GUID guid); 
  ContainerInfo_co18* FindGuid (const GUID guid, OmsSchemaHandle sh, OmsContainerNo cno);
  void                Insert   (ContainerInfo_co18* pClassInfo);
  void                Init();
public :
  class ContainerIterator {
   CContainerDictionary_co18* m_clsDir;
    int                     m_index;
    pContainerInfo_co18     m_curr;
  public :
    ContainerIterator(CContainerDictionary_co18*);
    void operator++();
    inline operator bool() const;
    inline pContainerInfo_co18 operator()() const;
  };

  CContainerDictionary_co18();
  void                Delete (IliveCacheSink* lcSink, ContainerInfo_co18* pContainerInfo);
  tgg00_BasisError    DropContainer          (IliveCacheSink* lcSink, const GUID& guid, OmsSchemaHandle Schema, OmsContainerNo cno);
  tgg00_BasisError    DropSchema             (IliveCacheSink* lcSink, OmsSchemaHandle Schema);
  ClassInfo_co18*     FindGuid               (IliveCacheSink* lcSink, const GUID guid); 
  ContainerInfo_co18* FindViaContainerHandle (IliveCacheSink* lcSink, tsp00_Uint4 containerHandle);
  tsp00_Uint4         GetLastDropId() const  { return m_dropId; }
  ContainerInfo_co18* FindViaGuid (IliveCacheSink* lcSink, const GUID guid, OmsSchemaHandle sh, OmsContainerNo cno);
  ClassInfo_co18* RegisterClass (IliveCacheSink* lcSink, 
                                 const char* ClassName,
                                 const GUID& guid,
                                 const GUID* pBaseClass,
                                 const tgg00_ObjColDesc &ColDesc,
                                 size_t ObjectSize, 
                                 void* vtptr);
  ContainerInfo_co18* RegisterContainer (IliveCacheSink* lcSink, 
                                         const GUID& guid, 
                                         const char* ClassName, 
                                         size_t PersistentSize, 
                                         size_t ObjectSize, 
                                         const GUID* pBaseClass,
                                         const tgg00_ObjColDesc &ColDesc,
                                         OmsSchemaHandle sh,
                                         OmsContainerNo cno,
                                         void* vtptr);

  void ReleaseContainerInfo(IliveCacheSink* lcSink, pContainerInfo_co18);
  void Shutdown();
  friend class CContainerDictionary_co18::ContainerIterator;
};

/*----------------------------------------------------------------------------------*/

inline CContainerDictionary_co18::ContainerIterator::operator bool() const {
  return (NULL != m_curr);
}

/*----------------------------------------------------------------------------------*/

inline pContainerInfo_co18 CContainerDictionary_co18::ContainerIterator::operator()() const {
  return m_curr;
}

/*----------------------------------------------------------------------------------*/

inline short ContainerInfo_co18::DecRefCnt() {
  --m_refCnt;
  return m_refCnt; 
}

/*----------------------------------------------------------------------------------*/

inline void ContainerInfo_co18::IncRefCnt(IliveCacheSink* lcSink) {
  CriticalSection_co001 cs(lcSink, RGN_CLASS_DIR);
  ++m_refCnt;
  OMS_TRACE(omsTrContainerDir, lcSink, "ContainerInfo::IncRefCnt " << m_refCnt);
  OMS_TRACE(omsTrContainerDir, lcSink, "Name : " << GetClassInfoPtr()->GetClassName() << "Schema : " << GetSchema() << "CNo : "  << m_ContainerNo);
}

#endif