/*


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end


*/
&ifdef KERNEL
/*******************************************************************************
*
* module:       hco31.h
*
* ------------------------------------------------------------------------------
*
* responsible:  Thomas Anhaus
*
* lastChanged:  2000-01-19
*
* purpose:      definition of liveCache error class 
*
* generatedBy:   
*
* ------------------------------------------------------------------------------
&else
/********************************************************************************
&endif
*
* Copyright by SAP AG, 1999
*
********************************************************************************/
&ifdef KERNEL
#ifndef HCO31_H
#define HCO31_H


#include "ggg92.h"
#include "gco00.h"
#include "ilivecachesink.h" /* no check */
#include "gco01.h"

struct AFX_EXT_CLASS OmsObjectId : public tgg92_OmsObjectId
{
public :
    
  OmsObjectId() 
  {
    setNil();
  };

  OmsTypeOid* pOid() {
    return REINTERPRET_CAST (OmsTypeOid*, &m_oid);
  }
protected :
  OmsObjectId( unsigned short pos, unsigned int l ) 
  { m_oid.pagePos = pos; setPno(l); m_oid.generation = NIL_OBJ_VERSION; }
};
&else
#ifndef _LIVECACHE_DBPERROR
#define _LIVECACHE_DBPERROR

enum {
    E_INCOMPATIBLE_OID   = -28007,
    E_NUMBER_OVERFLOW    = -28008
};

struct OmsObjectId;
&endif

/*!
  class : DbpErrorBase

  description : Base Class for all errors thrown in OMS

  note : all classes thrown in a db-procedure should by derived from DbpErrorBase
 */
class AFX_EXT_CLASS DbpErrorBase 
{
public :
  DbpErrorBase() {}
};
/*! endclass: DbpErrorBase */


/*!
  class :  DbpError

  description: Error Class of SQL-Class and OMS

 */
class AFX_EXT_CLASS DbpError : public DbpErrorBase 
{
public :
  enum DbpErrorKinds { DB_ERROR, HRESULT_ERROR, RTE_ERROR, USER_DEFINED,
  DB_ERROR_UNICODE, USER_DEFINED_UNICODE };

&ifdef KERNEL
  DbpError (DbpErrorKinds ErrorKind, long ErrorNo);
  DbpError (DbpErrorKinds ErrorKind, long ErrorNo, const char* errtext);
  DbpError (DbpErrorKinds ErrorKind, long ErrorNo, const DbpTypeUnicode* errtext);
  DbpError (DbpErrorKinds ErrorKind, long ErrorNo, const char* errtext, const OmsObjectId& oid);
  DbpError (DbpErrorKinds ErrorKind, long ErrorNo, const DbpTypeUnicode* errtext, const OmsObjectId& oid);
&endif
  DbpError (long ErrorNo);
  DbpError (long ErrorNo, const char* errtext);
  DbpError (long ErrorNo, const DbpTypeUnicode* errtext);
  long  dbpError () const;
  short dbpDBError () const;
  short dbpRteError () const;
  long  dbpHresultError() const;
  const char* dbpErrorText() const;
  const DbpTypeUnicode* dbpErrorTextUnicode() const;
  const OmsObjectId& dbpErrorOid() const;

  DbpErrorKinds m_errorKind;
  long          m_errorNo;
&ifdef KERNEL
  OmsObjectId   m_oid;
&else
  OmsTypeOid    m_oid;
&endif
  char          m_errorText[256]; 
};
/*! endclass: DbpError */
#endif 