/*


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end


*/
// liveCacheDbg.h : Declaration of the CliveCacheDbg

#ifndef __LIVECACHEDBG_H_
#define __LIVECACHEDBG_H_

#include "hco60res.h"       // main symbols
#include "gsp00.h"       
#include "hos38.h"
#include "geo39.h"

DWORD __stdcall DBProcThread(void* pv);

///////////////////////////////////////////////////////////
//
// Thread procedure
//
/////////////////////////////////////////////////////////////////////////////
// CliveCacheDbg

class ATL_NO_VTABLE CliveCacheDbg : 
public CComObjectRootEx<CComMultiThreadModel>,
public CComCoClass<CliveCacheDbg, &CLSID_liveCacheDbg>,
public IliveCacheDbg
{
  
public:
  
  DWORD                       m_ThreadId;	           // ID of DBPROC-thread
  HANDLE                      m_hThread;	           // Handle to DBPROC-thread
  HANDLE                      m_hCoRequestEvent;     // CoRequest scheduled
  HANDLE                      m_hCoReplyEvent;       // CoReply   scheduled
  HANDLE                      m_hStopThreadEvent;    // Stop DB-Proc Thread
  HANDLE                      m_hDBProcTermEvent;    //
  
  unsigned char*              m_buffer_ptr;
  unsigned char*              m_buffer_ptr2;
  long                        m_DbgBuf2Length;
  HRESULT                     m_hr;
  unsigned char*				m_SQLOutbufPtr;
  tsp_dcom_init_sink_parms    m_sink_parms;          // 
  unsigned char               m_sink_container[256]; //
  teo39_DebugMsg*             m_DbprocParms;
  unsigned char*              m_DbprocInOnlyParms;
  unsigned char               m_DbprocParmsStack[2048];
  unsigned char*              m_CallBackBuf;
  
  CliveCacheDbg()
  {
    DWORD rc;
    
    m_buffer_ptr    = NULL;
    m_buffer_ptr2   = NULL;
    m_DbgBuf2Length = 0;
    m_hr            = NULL;
    m_sink_parms.acv_addr    = (char*)this;
    m_sink_parms.sink_addr   = (char*)(&m_sink_container[0]);
    m_sink_parms.sink_size   = sizeof(m_sink_container);
    m_DbprocParms            = (teo39_DebugMsg*)malloc( MARSHAL_BUF_MXEO39 );
    m_DbprocInOnlyParms      = (unsigned char*)malloc( MARSHAL_BUF_MXEO39 );
    m_CallBackBuf            = (unsigned char*)malloc( MARSHAL_BUF_MXEO39 );
    m_hThread   = ::CreateThread(NULL,             // Default security
      0,                // Default stack size
      DBProcThread, 
      (void*)this,
      CREATE_SUSPENDED, // Create the thread suspended.
      &m_ThreadId) ;    // Get the Thread ID.
    m_hCoRequestEvent         = ::CreateEvent(NULL, FALSE, FALSE, NULL) ;
    m_hCoReplyEvent           = ::CreateEvent(NULL, FALSE, FALSE, NULL) ;
    m_hDBProcTermEvent        = ::CreateEvent(NULL, FALSE, FALSE, NULL) ;
    m_hStopThreadEvent        = ::CreateEvent(NULL, FALSE, FALSE, NULL) ;
    
    // Thread was created suspended; start the thread.
    rc = ::ResumeThread(m_hThread) ;
    _ASSERT(rc != 0xffffffff) ;
    
  }
  
  ~CliveCacheDbg()
  {
    // Signal DB-Proc thread to terminate;
    SetEvent(m_hStopThreadEvent) ;
    DWORD dwReturn = ::WaitForMultipleObjects(1,
      &m_hDBProcTermEvent,
      FALSE,
      3000 );      /* PTS 1106838 13-Jun-2000 FF  max wait time == 3 seconds*/
      /* INFINITE);   PTS 1106838 13-Jun-2000 FF 
    while ( dwReturn != WAIT_OBJECT_0 )
		    dwReturn = ::WaitForMultipleObjects(1,
        &m_hDBProcTermEvent,
        FALSE,
        INFINITE); */
    CloseHandle(m_hCoRequestEvent);
    CloseHandle(m_hCoReplyEvent);
    CloseHandle(m_hDBProcTermEvent);
    CloseHandle(m_hStopThreadEvent);
    CloseHandle(m_hThread); 
  }
  
  DECLARE_REGISTRY_RESOURCEID(IDR_LIVECACHEDBG)
    DECLARE_NOT_AGGREGATABLE(CliveCacheDbg)
    
    BEGIN_COM_MAP(CliveCacheDbg)
    COM_INTERFACE_ENTRY(IliveCacheDbg)
    END_COM_MAP()
    
    // IliveCacheDbg
public:
  STDMETHOD(Debug)(/*[in]*/ long DbgBufSize,
    /*[in,out]*/ long *pDbgBufLength, 
    /*[size_is(*pDbgBufLength), in, out]*/ unsigned char pDbgBuffer[],
    /*[in]*/ long DbgBufSize2,
    /*[in,out]*/ long *pDbgBufLength2, 
    /*[size_is(*pDbgBufLength), in, out]*/ unsigned char pDbgBuffer2[]);
};

#endif //__LIVECACHEDBG_H_
