/*!***************************************************************************

  module      : hco80CGenInspecCode.h

  -------------------------------------------------------------------------

  responsible : FerdiF

  special area: IDL to Code Generator   
  description : Generate Inspection Code out of generated XML file
				
  
  last changed: 2000-03-08  11:22
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/

#ifndef HCO81CGENINSPECCODE_H
#define HCO81CGENINSPECCODE_H

#include "gsp00.h"
#include <fstream.h>

#ifndef HCO82IDL2CODE_H
#include "hco82Idl2Code.h"
#endif

#include "gco00.h"
#include "ggg250.h"
#include "ggg251.h"

class co81CGenInspecCode;

/*----------------------------------------------------------------------*/

struct co81CExprElem {
  XmlElementType_Enum_co82  m_token_type;
  char *                    m_token_value;
  co81CExprElem();
  ~co81CExprElem();
  const co81CExprElem & operator= (const co81CExprElem& right);
};

/*----------------------------------------------------------------------*/


class co81CTypedef {
  cgg251DCList<co81CTypedef*,co81CGenInspecCode> m_member_list;
  cgg251DCList<co81CExprElem*,co81CGenInspecCode> m_expr_list;
  GUID                m_uuid;
  XmlElementType_Enum_co82   m_def_type;
  char*               m_name;
  char*               m_type;
  XmlAttributeType_Enum_co82 m_value_type;
  char*               m_value_string;
  unsigned char       m_ptr;
  short               m_size;
  short               m_dimension;
  short               m_alignment;
  short               m_member_num;
  short               m_vt_type;
public:
  co81CTypedef();
  ~co81CTypedef();
  inline int          Compare(const co81CTypedef*, const co81CTypedef*);
  void  appendMember(co81CTypedef* m);
  void  appendToken(co81CExprElem* s);
  short getMemberNum(co81CGenInspecCode &CGenInsepcCode);
  int   calExprValue(co81CGenInspecCode &CGenInsepcCode);
  void  setDimension(short d);
  void  setName(const XmlElementType_Enum_co82 dt, const char* n);
  void  setType(co81CGenInspecCode &CGenInsepcCode, const char* t, short size, short align);
  void  setValue(const XmlAttributeType_Enum_co82 vt, const char* vs);
  void  setUuid(const char* uuid);
  void  setStructCharacteristics(co81CGenInspecCode &CGenInsepcCode);
  void  printSubParameter(co81CGenInspecCode &CGenInsepcCode, ofstream &fout, short &pos);
  void  checkIfAbapOrStreamHandle(co81CGenInspecCode &CGenInsepcCode, short *prm_type, 
    short *prm_size, short *prm_elem_num, co81CTypedef * const ** row_def );
  inline short getAlignment() const;
  inline short getDimension() const;
  inline cgg251DCList<co81CTypedef*,co81CGenInspecCode> getMemberList() const;
  inline short getMemberNum() const;
  inline XmlElementType_Enum_co82 getDefType() const;
  inline XmlAttributeType_Enum_co82 getValType() const;
  inline short getSize() const;
  inline short getVtType() const;
  inline const char* getType() const;
  inline const GUID* const getUuid() const;
  inline const char* const getValue() const;
  inline const char* const getName() const;
  inline unsigned char getPtr() const;
  inline void*  allocate(size_t sz);
  inline void   deallocate(void* p);
};

/*----------------------------------------------------------------------*/

inline short co81CTypedef::getAlignment() const {
  return m_alignment;
}

/*----------------------------------------------------------------------*/

inline short co81CTypedef::getDimension() const {
  return m_dimension;
}

/*----------------------------------------------------------------------*/

inline XmlElementType_Enum_co82 co81CTypedef::getDefType() const {
  return m_def_type;
};

/*----------------------------------------------------------------------*/

inline XmlAttributeType_Enum_co82 co81CTypedef::getValType() const {
  return m_value_type;
};

/*----------------------------------------------------------------------*/

inline cgg251DCList<co81CTypedef*,co81CGenInspecCode> co81CTypedef::getMemberList() const {
  return m_member_list;
};

/*----------------------------------------------------------------------*/

inline short co81CTypedef::getMemberNum() const {
  return m_member_num;
};

/*----------------------------------------------------------------------*/

inline unsigned char co81CTypedef::getPtr() const {
  return m_ptr;
};

/*----------------------------------------------------------------------*/

inline short co81CTypedef::getSize() const {
  return m_size;
}

/*----------------------------------------------------------------------*/

inline const char * co81CTypedef::getType() const {
  return m_type;
}

/*----------------------------------------------------------------------*/

inline short co81CTypedef::getVtType() const {
  return m_vt_type;
}

/*----------------------------------------------------------------------*/

inline const GUID* const co81CTypedef::getUuid() const {
  return &m_uuid;
};

/*----------------------------------------------------------------------*/

inline const char* const co81CTypedef::getValue() const {
  return m_value_string;
};

/*----------------------------------------------------------------------*/

inline const char* const co81CTypedef::getName() const {
  return m_name;
};

/*----------------------------------------------------------------------*/

inline int co81CTypedef::Compare(const co81CTypedef* l, const co81CTypedef* r) {
  if ( strcmp(l->m_name, r->m_name) > 0 ) 
    return 1;
  else if ( strcmp(l->m_name, r->m_name) < 0 )
    return -1;
  return 0;
}

/*----------------------------------------------------------------------*/

class co81CGenInspecCode {
  tsp00_C256c         m_filename;
  tsp00_C64c          m_name;
  ofstream            m_cpp_out;
  ofstream            m_h_out;
  co81CTypedef*       m_act_pType;          /* type description in work */
  bool                m_expr_flag;          /* an <expr> is expected    */
  co81CExprElem       m_expr_token;         /* expr token in work       */
  cgg250AvlTree<co81CTypedef*, co81CTypedef,co81CGenInspecCode> m_index;
  short getTypeDimension(const char*);
  int getTypeCharacteristics( const char* type, const char* usedby, const char* username, 
    short& size, short &align );
  void printGuidV1( ofstream& out, co81CTypedef& type );
  bool checkFirstParam(const char *pname, const char *puuid, const char *ptype, const char *pdirection);
public:
  co81CGenInspecCode();
  ~co81CGenInspecCode();
  int Init(char*  cFilename);
  void startElement(int &error, DboType_Enum_co82 &dboType, const char *name, const char **atts);
  void endElement(int &error, const char *name);
  void charData(int &error, const char *s, int len);
  inline void*  allocate(size_t sz);
  inline void   deallocate(void* p);
  co81CTypedef* const * findNode( co81CTypedef &findNode );
  short checkBasisType(const char* type);
};

/*----------------------------------------------------------------------*/

inline void*  co81CGenInspecCode::allocate(size_t sz) {
  return new char [sz];
  
};

/*----------------------------------------------------------------------*/

inline void   co81CGenInspecCode::deallocate(void* p) {
  /* delete p;                            PTS 1109282   */
  delete REINTERPRET_CAST( char*, p ); /* PTS 1109282   */
};

#endif

