/* @lastChanged: "1999-04-21  16:14"

  * @filename:   vco10.cpp
  * @purpose:    "Implementation of C++ Persistent Object Interface"
  * @release:    7.2.0.0
  * @see:        "http://www.bea.sap-ag.de/doku/rel-7.1/vco10.html"
  *
  * @copyright:  "Copyright by   SAP AG, 1998"


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/


/* ================================================================================ */
/* ==========              REPLACEMENTS BY THE PREPROCESSOR               ========= */
/* ================================================================================ */



/* ================================================================================ */
/* ==========                 DECLARATION OF TYPES                        ========= */
/* ================================================================================ */




/* ================================================================================ */
/* ==========                 DECLARATION OF FUNCTIONS                    ========= */
/* ================================================================================ */

#define EXTERN_CO102
#include "gco102.h"
#include "geo00.h"
#include "gco00.h"
#include "ggg00.h"
#include "gco10.h"
#include "hsp20.h"
#include "ggg200.h"
#include "hco17.h"
#include "hco18.h"
#include "hco19.h"
#include "heo381.h"

/*---------------------------------------------------------------------------*/
/*                              Globals                                      */
/*---------------------------------------------------------------------------*/

const char* VAR_OBJ_CLASS_NAME_CO10 = "VarObjContainer";


#if defined(_IBMR2) && defined (BIT64)
static DummySink_cco10*        pDummySink_co10;
static IliveCacheSink*         ProxySinkAddr_co10() { return pDummySink_co10; }
#else
static DummySink_cco10         DummySink_co10;
static IliveCacheSink*         ProxySinkAddr_co10() { return &DummySink_co10; }
#endif

#if defined(BIT64)
static OmsTypeInt64            co10_maxHeap = MAX_INT4_SP00;
#else
static unsigned long           co10_maxHeap = MAX_INT4_SP00;
#endif
static GetSinkPtr_co02         GetSinkAddrFunc_co10 = ProxySinkAddr_co10;

class CResetSink {
public :
  CResetSink() {}
  ~CResetSink() {
    GetSinkAddrFunc_co10 = ProxySinkAddr_co10;
  }
};

class CGlob_co10 {
public :
  CGlob_co10() : RegionCnt(0), VersionDictionary(), ClassDictionary(),
#if defined(SUN)
    Monitor(co19_Init()),
#else
    Monitor(),
#endif
    VersionIter(), Heap (1024, 1024, co10_maxHeap), ResetSink() 
  {
  }
  ~CGlob_co10() {
  }
  //
  // caution : order of member variables is important !!
  //
  int                        RegionCnt;
  // liboms shared heap, must be destructed after heap objects have been destructed 
  ceo381_Heap                Heap;
  CVersionDictionary_co17    VersionDictionary;
  CContainerDictionary_co18  ClassDictionary;
#if defined(SUN)
  // due to linker duplicates of static template destructors Monitor must be 
  // placed in vco19.cpp. 
  Monitor_co19&              Monitor;
#else
  Monitor_co19               Monitor;
#endif
  // global version iterator, used for 'select .. from sysdd.oms_versions
  cgg250AvlInfoTree<CVersionId_co17,OmsContext_co12*,CCompareVersionId_co17,co16_SharedMemAllocator>::Iterator VersionIter;
  // ResetSink must be last member, because the destructor ensures, that the
  // destructors of the other members don't ask for the kernel environment !
  CResetSink              ResetSink;
};

#if defined(_IBMR2) && defined (BIT64)
static CGlob_co10*             pGlob_co10;
static LibOmsInterface_cco10*  pLibOmsInterface_co10;
#define Glob_co10 (*pGlob_co10)
#else
static CGlob_co10              Glob_co10;
static LibOmsInterface_cco10   LibOmsInterface_co10;
#endif

#if defined(_IBMR2) && defined (BIT64)
// AIX does not initialize global variables, i.e. contructors are not called
bool InitAIX64Globals() {
  co10_maxHeap = MAX_INT4_SP00;
  GetSinkAddrFunc_co10 = ProxySinkAddr_co10;
  pDummySink_co10 = new DummySink_cco10;
  if (NULL == pDummySink_co10) {
    return false;
  }
  pGlob_co10 = new CGlob_co10;
  if (NULL == pGlob_co10) {
    return false;
  }
  pLibOmsInterface_co10 = new LibOmsInterface_cco10;
  if (NULL == pLibOmsInterface_co10) {
    return false;
  }
  return true;
}
#endif


/* ================================================================================ */
/* ==========                 DEFINITION OF GLOBAL FUNCTIONS              ========= */
/* ================================================================================ */

tsp00_Uint4 co10_chunkSize(void* p) {
  if (NULL == p) {
    return 0;
  }
  else {
    return Glob_co10.Heap.chunkSize(p);
  }
}

/*----------------------------------------------------------------------*/

inline unsigned char* malloc_co10(IliveCacheSink* session, size_t size) {
  
  unsigned char* p;
  CriticalSection_co001 cs(session, RGN_HEAP_COUNTER);
  cs.Enter();
  p = (unsigned char*) Glob_co10.Heap.allocate(size);
  return p;
}

/*----------------------------------------------------------------------*/

inline unsigned char* malloc_co10(size_t size) {
  unsigned char* p; 
  IliveCacheSink* session;
  session = (*GetSinkAddrFunc_co10)();
  CriticalSection_co001 cs(session, RGN_HEAP_COUNTER);
  cs.Enter();
  p = REINTERPRET_CAST(unsigned char*, Glob_co10.Heap.allocate(size));
  return p;
}

/*----------------------------------------------------------------------*/

inline void* operatorNew_co10 (size_t size) {
  unsigned char* p = malloc_co10(size); 
  if (NULL == p) {
    IliveCacheSink* lcSink = (*GetSinkAddrFunc_co10)();
    OMS_TRACE(omsTrMemory, lcSink, "malloc failed : required : " << size);
    CriticalSection_co001 cs(lcSink, RGN_UNLOAD_VERSION);
    cs.Enter();
    p = malloc_co10(size);
    if (NULL == p) {
      bool versionUnloaded;
      do {
        versionUnloaded = Glob_co10.VersionDictionary.UnloadOldestVersion(lcSink);
        if (versionUnloaded) {
          p = malloc_co10(size);
        }
      }
      while ((NULL == p) && (versionUnloaded));
    }
  }
  if (NULL == p) {
    throw DbpError (DbpError::RTE_ERROR, e_new_failed);
  }
  return p;
}

/*----------------------------------------------------------------------*/

void* co10_malloc (IliveCacheSink* lcSink, size_t size) {
  unsigned char* p = malloc_co10(lcSink, size); 
  if (NULL == p) {
    CriticalSection_co001 cs(lcSink, RGN_UNLOAD_VERSION);
    cs.Enter();
    p = malloc_co10(lcSink, size);
    if (NULL == p) {
      bool versionUnloaded;
      do {
        versionUnloaded = Glob_co10.VersionDictionary.UnloadOldestVersion(lcSink);
        if (versionUnloaded) {
          p = malloc_co10(lcSink, size);
        }
      }
      while ((NULL == p) && (versionUnloaded));
    }
  }
  if (NULL == p) {
    throw DbpError (DbpError::RTE_ERROR, e_new_failed);
  }
  return p;
}

/*----------------------------------------------------------------------*/

void co10_DetachContainerInfo(IliveCacheSink* lcSink, pContainerInfo_co18 p) {
  Glob_co10.ClassDictionary.ReleaseContainerInfo (lcSink, p);
}

/*----------------------------------------------------------------------*/

void co10_free(IliveCacheSink* session, void* p) {
  if (NULL == p) { 
    return;
  }
  CriticalSection_co001 cs(session, RGN_HEAP_COUNTER);
  cs.Enter();
  Glob_co10.Heap.deallocate(p);
 }

/*----------------------------------------------------------------------*/

IliveCacheSink* co10_GetLcSink() {
  return (*GetSinkAddrFunc_co10)();
}

/*----------------------------------------------------------------------*/

externCpp
void* omsMalloc(size_t size) {
  return operatorNew_co10(size);
}

/*----------------------------------------------------------------------*/

externCpp
void omsFree(void* p) {
  IliveCacheSink* session = (*GetSinkAddrFunc_co10)();
  CriticalSection_co001 cs(session, RGN_HEAP_COUNTER);
  cs.Enter();
  Glob_co10.Heap.deallocate(p);
}

/*----------------------------------------------------------------------*/

externC void* co10_GetInterface() {
#if defined(_IBMR2) && defined (BIT64)
  if (InitAIX64Globals()) {
    return REINTERPRET_CAST(void*, pLibOmsInterface_co10);
  }
  return NULL;
#else
  return REINTERPRET_CAST(void*, &LibOmsInterface_co10);
#endif
}

/*----------------------------------------------------------------------*/

void co10_MarkUnloadable(OmsContext_co12* pContext) {
  Glob_co10.VersionDictionary.MarkUnloadable(pContext);
}

/*----------------------------------------------------------------------*/

ContainerInfo_co18* co10_RegisterContainer (IliveCacheSink*        lcSink, 
                                            const GUID&            guid, 
                                            const char*            ClassName, 
                                            size_t                 PersistentSize, 
                                            size_t                 ObjectSize, 
                                            const GUID*            pBaseClass,
                                            const tgg00_ObjColDesc &ColDesc,
                                            OmsSchemaHandle        sh,
                                            OmsContainerNo         cno,
                                            void*                  vtptr)
{
  return Glob_co10.ClassDictionary.RegisterContainer (lcSink, guid, ClassName, PersistentSize,
    ObjectSize, pBaseClass, ColDesc, sh, cno, vtptr); 
}

/*----------------------------------------------------------------------*/

void co10_ToKnlIdentifier (const char* Source, tsp00_KnlIdentifier& Dest) {
  int copyLen;
  copyLen = strlen(Source);
  if (copyLen > sizeof(Dest)) {
    throw DbpError (DbpError::DB_ERROR, -2014, "too long identifier");
  }
  memcpy(&Dest[0], Source, copyLen);
  memset(&Dest[copyLen], 0, sizeof(Dest) - copyLen);
}

/*----------------------------------------------------------------------*/

void LibOmsInterface_cco10::AdviseGetSinkFunc(GetSinkPtr_co02 func) {
  GetSinkAddrFunc_co10 = func;
}

/*----------------------------------------------------------------------*/

void LibOmsInterface_cco10::AdviseOmsRegionCnt(tsp00_Int4 cntRgn) {
  Glob_co10.RegionCnt = cntRgn;
}

/*----------------------------------------------------------------------*/

void LibOmsInterface_cco10::InitLibOms(IliveCacheSink* lcSink) {
  struct VirtualObject {
    void* vtptr;
  }; 
  tgg00_ObjColDesc ColDesc;
  OmsVarObjInfo_co12 info;
  ColDesc.colHashCount_gg00 = 0;
  ColDesc.colInvCount_gg00  = 0;
  ColDesc.colHashList_gg00  = NULL;
  ColDesc.colInvList_gg00   = NULL;
  Glob_co10.ClassDictionary.RegisterClass (lcSink, VAR_OBJ_CLASS_NAME_CO10, VAR_OBJ_GUID_GCO10, NULL,
    ColDesc, sizeof(info) + OmsObjectContainer::headerSize(), (REINTERPRET_CAST(VirtualObject*, &info))->vtptr); 
}

/*----------------------------------------------------------------------*/

void LibOmsInterface_cco10::AdviseHeapLimit(tsp00_Uint4 limit) {
  Glob_co10.Heap.setHeapLimit((0 == limit) ? MAX_INT4_SP00 : limit);
}

/*----------------------------------------------------------------------*/

void LibOmsInterface_cco10::GetMemoryUsageInfo(IliveCacheSink*     lcSink, 
                                               tsp00_8ByteCounter& totalHeapUsed,
                                               tsp00_8ByteCounter& freeChunkSpace) 
{
  ceo381_Heap::THeapInfo heapInfo;
  Glob_co10.Heap.getHeapInfo(heapInfo);
  totalHeapUsed  = heapInfo.heapUsed;
  freeChunkSpace = Allocator_co16.EvalFreeSpace(lcSink);
};

/*----------------------------------------------------------------------*/

tsp00_Bool LibOmsInterface_cco10::GetFirstVersion(IliveCacheSink*     lcSink,
                                                  tsp00_C24&          versionId,
                                                  tsp00_Date&         createDate,
                                                  tsp00_Time&         createTime,
                                                  tsp00_Date&         openDate,
                                                  tsp00_Time&         openTime,
                                                  tgg91_TransNo&      consistentView,
                                                  tsp00_Bool&         isMarked,
                                                  tsp00_Bool&         isOpen,
                                                  tsp00_Bool&         isUnloaded,
                                                  tsp00_8ByteCounter& heapUsage,
                                                  tgg00_FileId&       versionTree,
                                                  tgg00_FileId&       versionInvTree) {
  lcSink->EnterCriticalSection(RGN_VERSION_DIR);
  Glob_co10.VersionIter = Glob_co10.VersionDictionary.First();
  return GetNextVersion(lcSink, versionId, createDate, createTime, 
    openDate, openTime, consistentView, isMarked, isOpen, isUnloaded,
    heapUsage, versionTree, versionInvTree);
}

/*----------------------------------------------------------------------*/

tsp00_Bool LibOmsInterface_cco10::GetMonitorInfo(
                                                 IliveCacheSink*       lcSink,
                                                 tsp00_Addr&           handle,
                                                 tsp00_C16&            iid,
                                                 tsp00_Int4&           dispid,
                                                 tgg00_COMMonitorInfo& monInfo)
{
  static Monitor_co19::Iterator Iter;
  if (NULL == handle) {
    lcSink->EnterCriticalSection(RGN_MONITOR);
    Iter = Glob_co10.Monitor.begin();
    handle = REINTERPRET_CAST(tsp00_Addr, &Iter);
  }
  if (Iter) {
    Iter.Get(*REINTERPRET_CAST(CLSID*, &iid), dispid, monInfo);
    ++Iter;
    return true;
  }
  else {
    lcSink->LeaveCriticalSection(RGN_MONITOR);
  }
  return false;
}

/*----------------------------------------------------------------------*/

tsp00_Bool LibOmsInterface_cco10::GetNextVersion(IliveCacheSink*     lcSink,
                                                 tsp00_C24&          versionId,
                                                 tsp00_Date&         createDate,
                                                 tsp00_Time&         createTime,
                                                 tsp00_Date&         openDate,
                                                 tsp00_Time&         openTime,
                                                 tgg91_TransNo&      consistentView,
                                                 tsp00_Bool&         isMarked,
                                                 tsp00_Bool&         isOpen,
                                                 tsp00_Bool&         isUnloaded,
                                                 tsp00_8ByteCounter& heapUsage,
                                                 tgg00_FileId&       versionTree,
                                                 tgg00_FileId&       versionInvTree) {
  
  while (true) {
    if (Glob_co10.VersionIter) {
      OmsContext_co12* pContext = *Glob_co10.VersionIter.GetInfo();
      if (pContext->IsDropped()) {
        // version is marked as dropped, try to drop it now
        ++Glob_co10.VersionIter;
        if (!pContext->IsBoundToTrans()) {
          // version no longer bound to transaction, i.e. can be dropped now
          pContext->AssignLcSink(lcSink);  /* PTS 1107209 */
          Glob_co10.VersionDictionary.MarkNotUnloadable(lcSink, pContext);
          Glob_co10.VersionDictionary.DropVersion(CVersionId_co17(pContext->GetVersionId())); 
          pContext->DeleteSelf();
          if (Glob_co10.VersionIter) {
            //  reset iterator
            pContext = *Glob_co10.VersionIter.GetInfo();
            Glob_co10.VersionIter = Glob_co10.VersionDictionary.First();
            while (*Glob_co10.VersionIter.GetInfo() != pContext) {
              ++Glob_co10.VersionIter;
            }
          }
        }
      }
      else {
        memset (versionId, ' ', sizeof(versionId));
        memcpy(&versionId[0], &pContext->GetVersionId()[0], sizeof(OmsVersionId)); 
        createDate     = pContext->GetCreateDate();
        createTime     = pContext->GetCreateTime();
        openDate       = pContext->GetOpenDate();
        openTime       = pContext->GetOpenTime();
        consistentView = pContext->m_consistentView;
        isMarked       = pContext->IsMarked();
        isOpen         = pContext->IsVersionOpen();
        isUnloaded     = pContext->IsUnloaded();
        heapUsage      = pContext->HeapUsed() + pContext->VarObjHeapUsed(); 
        versionTree    = pContext->VersionFile();
        versionInvTree = pContext->VersionInvFile();
        ++Glob_co10.VersionIter;
        return true;
      }
    }
    else {
      lcSink->LeaveCriticalSection(RGN_VERSION_DIR);
      return false;
    }
  }
}

/*----------------------------------------------------------------------*/

void LibOmsInterface_cco10::ResetMonitor(IliveCacheSink* lcSink) {
  CriticalSection_co001 cs(lcSink, RGN_MONITOR);
  cs.Enter();
  Glob_co10.Monitor.Reset();
}

/*----------------------------------------------------------------------*/

tgg00_BasisError LibOmsInterface_cco10::Shutdown(IliveCacheSink* lcSink) {
  Glob_co10.VersionDictionary.Shutdown();
  Glob_co10.ClassDictionary.Shutdown();
  Allocator_co16.Release(lcSink);
  Glob_co10.Monitor.Reset();
  /*Glob_co10.Heap.freeHeap();                  PTS 1107627 FF 24-Aug-2000 */
  GetSinkAddrFunc_co10 = ProxySinkAddr_co10; /* PTS 1107627 FF 24-Aug-2000 */
  return e_ok;
}

/*----------------------------------------------------------------------*/

void LibOmsInterface_cco10::LockVersionDirectory(IliveCacheSink* lcSink) {
  lcSink->EnterCriticalSection(RGN_VERSION_DIR);
}

/*----------------------------------------------------------------------*/

bool LibOmsInterface_cco10::SetTraceLevel(tsp00_KnlIdentifier& lvl, bool enable) {
  return TraceLevel_co102.Set(lvl, enable);
}

/*----------------------------------------------------------------------*/

void LibOmsInterface_cco10::UnLockVersionDirectory(IliveCacheSink* lcSink) {
  lcSink->LeaveCriticalSection(RGN_VERSION_DIR);
}

/*----------------------------------------------------------------------*/

static void co10_TraceGuid (OmsHandle *h, const GUID& guid) {
  h->omsTracef ("guid : %8X-%4X-%4X-%1X%1X%1X%1X%1X%1X%1X%1X", guid.Data1, guid.Data2, guid.Data3,
  guid.Data4[0], guid.Data4[1], guid.Data4[2], guid.Data4[3], guid.Data4[4], guid.Data4[5],
  guid.Data4[6], guid.Data4[7]);
}

/*----------------------------------------------------------------------*/

externCpp
ClassInfo_co18* co10_FindClassInfo(IliveCacheSink* lcSink, const GUID guid) {
  return Glob_co10.ClassDictionary.FindGuid(lcSink, guid);
}

/*----------------------------------------------------------------------*/

externCpp
ContainerInfo_co18* co10_FindContainerInfo(IliveCacheSink* lcSink, tsp00_Uint4 containerHandle) {
  return Glob_co10.ClassDictionary.FindViaContainerHandle(lcSink, containerHandle);
}

/*----------------------------------------------------------------------*/

externCpp
ContainerInfo_co18* co10_FindContainerInfo(IliveCacheSink* lcSink, const GUID& guid, OmsSchemaHandle Schema, OmsContainerNo ContainerNo) {
  return  Glob_co10.ClassDictionary.FindViaGuid(lcSink, guid, Schema, ContainerNo);
}

/*----------------------------------------------------------------------*/

externCpp
tsp00_Uint4 co10_GetLastDropId() {
  return Glob_co10.ClassDictionary.GetLastDropId();
}

/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/* Implementation of class OmsCriticalSection                           */
/*----------------------------------------------------------------------*/

OmsCriticalSection::OmsCriticalSection(OmsHandle* h, int regionId) : m_session(h->m_pSession), 
  m_inSection(false), m_regionId(regionId + INTERNAL_CRITICAL_SECTIONS_CO001)
{
}

/*----------------------------------------------------------------------*/

void OmsCriticalSection::omsEnter() {
  if ((m_inSection) || (m_session->m_critical_section_in_use[m_regionId-INTERNAL_CRITICAL_SECTIONS_CO001])) {
    m_session->ThrowDBError (e_already_in_critical_section, "omsEnter");
  }
  m_session->m_critical_section_in_use[m_regionId-INTERNAL_CRITICAL_SECTIONS_CO001] = true;
  m_session->m_lcSink->EnterCriticalSection(m_regionId);
  m_inSection = true;
}

/*----------------------------------------------------------------------*/

void OmsCriticalSection::omsLeave() {
  if (m_inSection) {
    m_session->m_lcSink->LeaveCriticalSection(m_regionId);
    m_session->m_critical_section_in_use[m_regionId-INTERNAL_CRITICAL_SECTIONS_CO001] = false;
    m_inSection = false;
  }
  else {
    m_session->ThrowDBError ( e_not_in_critical_section, "omsLeave");
  }
}

/*----------------------------------------------------------------------*/

OmsCriticalSection::~OmsCriticalSection() {
  if (m_inSection) {
    omsLeave();
  }
}

/*----------------------------------------------------------------------*/
/* Implementation of class OmsDuplicateKey                              */
/*----------------------------------------------------------------------*/

OmsDuplicateKey::OmsDuplicateKey(long e, const OmsObjectId& oid) : 
  DbpError(DbpError::DB_ERROR, e, "Duplicate Key", oid) {
};

/*----------------------------------------------------------------------*/ 
/* Implementation of class OmsHandle                                    */
/*----------------------------------------------------------------------*/ 

OmsHandle::OmsHandle (OmsTypeInt4 sz) : m_session(NULL),   
m_pSession(NULL), m_monitor(NULL)
{
};

/*----------------------------------------------------------------------*/

HRESULT OmsHandle::omsAttachDefaultContext() {
  HRESULT hr = (*m_session)->GetDefaultContext(REINTERPRET_CAST(void**, &m_pSession));
  if ((!FAILED(hr)) && (NULL == m_pSession)) {
    m_pSession = new(*m_session) OmsSession_co12(*m_session, Glob_co10.RegionCnt);
    if (NULL == m_pSession) {
      return E_OUTOFMEMORY ;
    }
    hr = (*m_session)->SetDefaultContext((void*) m_pSession);
    m_pSession->CreateDefaultContext();
    OMS_TRACE(omsTrSession, (*m_session), "New Session : " << m_pSession); 
  }
  dbpAdviseLcSink(*m_session);
  m_pSession->IncRefCnt();
  OMS_TRACE(omsTrSession, (*m_session), "Session " << m_pSession << " attached to Hndl " << this << " RefCnt " << m_pSession->m_refCnt);
  return hr;
}

/*----------------------------------------------------------------------*/

tco13_ClsIdEntry* OmsHandle::omsAddContainerInfo (ContainerInfo_co18* ContainerInfo) {
  tco13_ClsIdEntry* curr = new(m_pSession->m_context) tco13_ClsIdEntry (ContainerInfo, m_pSession->m_context);
  m_pSession->AddContainerInfo (curr);
  return curr;
}

/*----------------------------------------------------------------------*/

OmsObjByClsIterBase OmsHandle::omsAllOids( const GUID& guid, OmsSchemaHandle Schema, OmsContainerNo ContainerNo ) 
{
  _TRACE_METHOD_ENTRY(this,"omsAllOids");
  
  OmsObjByClsIterBase iterator(*this);
  tsp00_Int2            DBError;
  CHResult_co001        hr;
  OmsObjectContainerPtr obj = NULL;
  iterator.m_containerInfo = m_pSession->GetClsInfo (guid, Schema, ContainerNo);
  tsp00_Int4 size          = iterator.m_containerInfo->GetPersistentSize();
  iterator.m_next          = REINTERPRET_CAST(OmsObjectContainerPtr, 
    m_pSession->allocate (iterator.m_containerInfo->GetObjectSize()));
  iterator.m_next->InitObjContainer(*iterator.m_containerInfo);
  iterator.m_next->m_oid = OmsObjectId();
  do {
    hr = (*m_session)->NextObj (
      (unsigned char*) &m_pSession->m_context->m_consistentView,
      (unsigned char*) &iterator.m_containerInfo->GetFileId(),
      m_pSession->m_context->VersionContext(),
      iterator.m_next->m_oid.pOid(), 
      (unsigned char*) &iterator.m_next->m_objseq, size,
      ((unsigned char*) &iterator.m_next->m_pobj) + sizeof(void*), &DBError);
    if ( DBError != 0 ) {
      if (e_no_next_object == DBError) {
        if (!m_pSession->InVersion()) {
          return iterator; /* PTS 1107792 */
        }
        break;
      }
      else {
        m_pSession->ThrowDBError (DBError, "omsSelectOids");
      }
    }
    obj = m_pSession->m_context->m_oidDir.HashFind (&iterator.m_next->m_oid);
    if (NULL != obj) {
      if (!obj->DeletedFlag()) {
        break;
      }
    }
    else {
      break;
    }
  }
  while (true);
  /* PTS 1107792 */
  // We are in a version, look for new version objects
  if (!(iterator.m_next->m_oid)) {
    iterator.m_iterNewObjectsInVersion = true;
    iterator.omsNextNewObjectInVersion(NULL);
  }
  iterator.m_curr = REINTERPRET_CAST(OmsObjectContainerPtr, 
    m_pSession->allocate (iterator.m_containerInfo->GetObjectSize()));
  iterator.m_curr->InitObjContainer(*iterator.m_containerInfo);
  iterator.operator++();
  return iterator;
}

/*----------------------------------------------------------------------*/

OmsObjectId OmsHandle::omsCastOid(const GUID guid, const OmsObjectId& oid) {
  return m_pSession->CastOid(guid, oid);
}

/*----------------------------------------------------------------------*/

bool OmsHandle::omsCheckOid ( const OmsObjectId& oid, const GUID& guid) {
  OMS_TRACE(omsTrInterface, m_pSession->m_lcSink, "omsCheckOid : " << guid << "," << oid);
  return m_pSession->CurrentContext()->CheckOid(oid, guid);
}

/*----------------------------------------------------------------------*/

bool OmsHandle::omsCheckOid (const OmsVarOid& oid)
{
  OMS_TRACE(omsTrInterface, m_pSession->m_lcSink, "omsCheckOid : " << oid);
  return m_pSession->CurrentContext()->CheckOid(oid);
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsCloseVersion ( const OmsVersionId& versionId )
{
  tsp00_Int2 DBError;
  if (!m_pSession->InVersion()) {
    return;
  }
  OmsContext_co12* pVersion = m_pSession->m_context;
  if (pVersion->IsOpenVersion(versionId)) {
    OMS_TRACE(omsTrInterface, m_pSession->m_lcSink, "omsCloseVersion : " << CFixedLenChar_co102(versionId, sizeof(versionId)));
    CriticalSection_co12 region(this->m_pSession, RGN_VERSION_DIR);
    if (1 < m_pSession->CurrentSubtransLevel()) {
      m_pSession->ThrowDBError (e_not_implemented, "Close Version : Open Subtrans");
    }
// PTS 1106619
#if ( defined (AIX) && (__IBMCPP__ >= 500) )
    OmsVersionId LocalVersionID ;
    memcpy(LocalVersionID, versionId, sizeof(OmsVersionId) );
    CHResult_co001 hr = (*m_session)->CloseVersion (&LocalVersionID,
      m_pSession->m_context->CurrentVersionKey(), &DBError);  
#else
    CHResult_co001 hr = (*m_session)->CloseVersion (CONST_CAST(OmsVersionId*, &versionId),
      m_pSession->m_context->CurrentVersionKey(), &DBError);  
#endif

    if ( DBError != 0 ) {
      m_pSession->ThrowDBError (DBError, "omsCloseVersion");
    }
    region.Enter();
    m_pSession->CloseVersion();
    if (pVersion->IsDropped()) {
      region.Leave();
      // version has been marked as to be deleted, do it now
      try {
        omsForceDropVersion(versionId);
      }
      catch (DbpError&) {
      }
    }
  }
}

/*----------------------------------------------------------------------*/

short OmsHandle::omsCommitHandling (int& phase) {
  tsp00_Uint4            listCount  = 0;
  tsp00_Uint4            entryCount = 0;
  bool                   inVersion  = m_pSession->InVersion();
  try {
    OmsCallbackInterface* callbackInterface = m_pSession->GetCallbackInterface();
    if (NULL != callbackInterface) {
      callbackInterface->omsValidate (*this, OmsCallbackInterface::BeforeCommit,
        inVersion);
    }
    phase = 1;
    m_pSession->ExecuteSubtransCommit (1);
    phase = 2;
    m_pSession->m_defaultContext->FlushObjCache();
    phase = 3;
    m_pSession->FreeLockObjAndBeforeImages();
    phase = 4;
    m_pSession->ClearDefaultContext();
    phase = 5;
    if (NULL != callbackInterface) {
      bool read_only          = m_pSession->m_read_only;
      m_pSession->m_read_only = true;
      callbackInterface->omsInvalidate (*this, OmsCallbackInterface::AfterCommit,
        m_pSession->InVersion());
      m_pSession->m_read_only = read_only;
      m_pSession->ClearDefaultContext();
    }
    phase = 6;
    //if ((entryCount > 2 * listCount) && (!inVersion))   
    //{
    //  m_pSession->m_context->m_oidDir.HashResize (entryCount);
    //}
    //if (inVersion) {
    //  m_pSession->CurrentContext()->UnLoad();
    //}
    m_pSession->TransEnd();
    phase = 7;
  }
  catch (DbpError& e) {
    dbpOpError ("omsCommit : DbpError %d", e.dbpError());
    TraceCommitPhase(phase);
    omsRollback(false);
    return e_work_rolled_back;
  }
  return 0;
}

/*----------------------------------------------------------------------*/

void  OmsHandle::TraceCommitPhase(int phase) {
  char msg[30];
  switch (phase) {
  case 0 :
    strcpy (msg, "omsValidate");
    break;
  case 1 :
    strcpy (msg, "SubtransCommit");
    break;
  case 2 :
    strcpy (msg, "FlushObjCache");
    break;
  case 3 :
    strcpy (msg, "FreeBeforeImages");
    break;
  case 4 :
    strcpy (msg, "ClearDefaultContext");
    break;
  case 5 :
    strcpy (msg, "OmsInvalidate");
    break;
  case 6 :
    strcpy (msg, "TransEnd");
    break;
  default :
    strcpy(msg, "After TransEnd");
  }
  dbpTrace("Commit Phase : %s", msg);
}

/*----------------------------------------------------------------------*/

short OmsHandle::omsCommit (bool doRelease) {
  
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsCommit");
  short  e;
  int    phase = 0;
#if defined(WIN32)
  __try {
#else
    try {
#endif
      e = omsCommitHandling(phase);
      if (0 != e) {
        return e;
      }
    }
#if defined(WIN32)
/*    __except(co10_EvalException (this, GetExceptionInformation ())) { PTS 1106505 FF */
    __except(co10_EvalException (m_pSession->m_lcSink, GetExceptionInformation ())) {
      dbpOpError ("omsCommit failed");
      TraceCommitPhase(phase);
      omsRollback(false);
      return e_work_rolled_back;
    }
#else
    catch (...) {
      dbpOpError ("omsCommit : unknown exception");
      TraceCommitPhase(phase);
      omsRollback(false);
      return e_work_rolled_back;
    }
#endif
    if (doRelease) {
      this->SessionEnd(); /* PTS 1107209 */
    }
    return 0;
  }

/*----------------------------------------------------------------------*/

  void OmsHandle::SessionEnd() {  /* PTS 1107209 */
    if (m_pSession->InVersion()) {
      OmsContext_co12* pContext = m_pSession->CurrentContext();
      pContext->CloseVersion();
      if (pContext->IsDropped()) {
        omsForceDropVersion(pContext->GetVersionId());
      }
    }
    m_pSession->ClearDefaultContext();
    m_pSession->ReleaseCallbackInterface();
  }

/*----------------------------------------------------------------------*/

void OmsHandle::omsCommitSubtrans( int required_lvl ) {
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsCommitSubtrans : reqlvl = " << required_lvl);
  m_pSession->CommitSubtrans(required_lvl);
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsCleanUpAfterCrash() {
  OMS_TRACE (omsTrAlways, m_pSession->m_lcSink, "omsCleanUpAfterCrash");
  CriticalSection_co12 region(this->m_pSession, RGN_VERSION_DIR);
  region.Enter();
  cgg250AvlInfoTree<CVersionId_co17,OmsContext_co12*,CCompareVersionId_co17,co16_SharedMemAllocator>::Iterator iter = Glob_co10.VersionDictionary.First();
  while (iter) {
    OmsContext_co12* pContext = *iter.GetInfo();
    ++iter;
    if ((pContext->IsBoundToTrans()) && (m_pSession->VersionBoundByMe(pContext))) {
      OMS_TRACE (omsTrAlways, m_pSession->m_lcSink, "Implicit Drop Version : " << 
        CFixedLenChar_co102(pContext->GetVersionId(), sizeof(pContext->GetVersionId())));
      Glob_co10.VersionDictionary.MarkNotUnloadable(m_pSession->m_lcSink, pContext);
      m_pSession->DropVersion(pContext);
      Glob_co10.VersionDictionary.DropVersion(CVersionId_co17(pContext->GetVersionId())); 
    }
  }
  region.Leave();
  omsRollback(true);
}

/*----------------------------------------------------------------------*/

OmsSchemaHandle OmsHandle::omsCreateSchema(const char* SchemaName) {
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsCreateSchema : " << SchemaName);
  tsp00_KnlIdentifier SchemaIdentifier;
  tsp00_Int4          SchemaId = 0;
  tgg00_BasisError    DBError;
  co10_ToKnlIdentifier(SchemaName, SchemaIdentifier);
  CHResult_co001 hr = (*m_session)->CreateSchema (&SchemaIdentifier, &SchemaId, &DBError);
  if ( DBError != 0 ) {
    m_pSession->ThrowDBError (DBError, "omsCreateSchema");
  }
  return SchemaId;
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsCreateVersion( const OmsVersionId& versionId )
{
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsCreateVersion : " << CFixedLenChar_co102(versionId, sizeof(versionId)));
  tsp00_Int2              DBError;
  OmsContext_co12*        initContext = m_pSession->CurrentContext();
  CriticalSection_co12    region(this->m_pSession, RGN_VERSION_DIR);
  if (m_pSession->IsSubtransOpen()) {
    m_pSession->ThrowDBError (e_too_many_subtrans, "Create Version : Open Subtrans");
  }
  if (m_pSession->InVersion()) {
    m_pSession->ThrowDBError (e_already_in_version, "omsCreateVersion"); 
  }
  region.Enter();
  if (Glob_co10.VersionDictionary.FindVersion(CVersionId_co17(versionId))) {
    m_pSession->ThrowDBError (e_duplicate_name, "omsCreateVersion");
  }
  m_pSession->CreateVersion(versionId);
  DBError = Glob_co10.VersionDictionary.InsertVersion(CVersionId_co17(versionId), m_pSession->CurrentContext());
  if (DBError != 0) {
    m_pSession->m_context->DeleteSelf();
    m_pSession->SetContext(initContext);
    m_pSession->ThrowDBError (DBError, "omsCreateVersion");
  }
}

/*----------------------------------------------------------------------*/

int OmsHandle::omsCurrentSubtransLevel() const {
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsCurrentSubtransLevel : " << m_pSession->CurrentSubtransLevel() - 1);
  return m_pSession->CurrentSubtransLevel() - 1;
}

/*----------------------------------------------------------------------*/

const OmsVersionId*  OmsHandle::omsCurrentVersionId () const {
  if (m_pSession->InVersion()) {
    return REINTERPRET_CAST(const OmsVersionId*, &m_pSession->CurrentContext()->GetVersionId()[0]);
  }
  return NULL;
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsDelete( const OmsObjectId& oid ) {
  
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsDelete : " << oid);
  m_pSession->IncDelete ();
  OmsObjectContainerPtr found = m_pSession->CurrentContext()->GetObj(oid, false);
  if (!found) {
    m_pSession->ThrowDBError (e_object_not_found, "omsDelete", oid);
  }
  OmsAbstractObject* pObj = &found->m_pobj; /* PTS 1107093 */
  pObj->omsCleanUp(*this);
  m_pSession->InsertBeforeImage (found);
  found->MarkDeleted();
  
};

/*----------------------------------------------------------------------*/

void OmsHandle::omsDelete( OmsAbstractObject* pobj ) {
  omsDelete (omsGetContainer(pobj)->m_oid);
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsDelete( const OmsVarOid& oid ) {
  OMS_TRACE (omsTrInterface | omsTrVarObject, m_pSession->m_lcSink, "omsDelete(Var) : " << oid);
  m_pSession->DeleteVarObject(oid);
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsDeleteAll (const GUID& guid, OmsSchemaHandle Schema, OmsContainerNo ContainerNo) {
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsDeleteAll : " << guid << ", " 
    << Schema << ", " << ContainerNo);
  m_pSession->DeleteAll (guid, Schema, ContainerNo);
}

/*----------------------------------------------------------------------*/

const void* OmsHandle::omsDeRef(const OmsVarOid& oid) {
  omsIfNilOidThrowError (this, oid, "omsDeRef : VarObject");
  OMS_TRACE (omsTrInterface | omsTrVarObject, m_pSession->m_lcSink, "omsDeRef(Var) : " << oid);
  return m_pSession->DeRefVarObject(oid);
}

/*----------------------------------------------------------------------*/

OmsAbstractObject* OmsHandle::omsDeRef (const OmsObjectId& oid, const GUID& guid, bool forUpd, bool do_lock )
{
  if (!oid) {
    return NULL;
  }
  else {
    return m_pSession->DeRef (oid, guid, forUpd, do_lock);
  }
};

/*----------------------------------------------------------------------*/

OmsAbstractObject* OmsHandle::omsDeRefKey (unsigned char*  key, 
                                           const GUID&     guid, 
                                           bool            forUpd, 
                                           bool            do_lock,
                                           OmsSchemaHandle Schema,
                                           OmsContainerNo  ContainerNo) 
{
  return m_pSession->DeRefViaKey (key, guid, forUpd, do_lock, Schema, ContainerNo);  
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsDropSchema(OmsSchemaHandle Schema) {
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsDropSchema : " << Schema);
  if (m_pSession->InVersion()) {
    m_pSession->ThrowDBError (e_missing_privilege, "omsDropSchema : not allowed in version");
  }
  if (OMS_DEFAULT_SCHEMA_HANDLE == Schema) {
    m_pSession->ThrowDBError (e_missing_privilege, "omsDropSchema : OMS Schema not allowed");
  }
  tgg00_BasisError DBError = Glob_co10.ClassDictionary.DropSchema(m_pSession->m_lcSink, Schema);
  if ( 0 != DBError ) {
    m_pSession->ThrowDBError (DBError, "omsDropSchema");
  }
  m_pSession->CleanAfterDropSchema();
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsDropContainer(const GUID& guid, OmsSchemaHandle Schema, OmsContainerNo ContainerNo) {
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsDropContainer : " << guid << ", " << Schema << ", " << ContainerNo );
   if (m_pSession->InVersion()) {
    m_pSession->ThrowDBError (e_missing_privilege, "omsDropContainer : not allowed in version");
  }
  tgg00_BasisError DBError = Glob_co10.ClassDictionary.DropContainer(m_pSession->m_lcSink, guid, Schema, ContainerNo);
  if ( 0 != DBError ) {
    m_pSession->ThrowDBError (DBError, "omsDropSchema");
  }
  m_pSession->CleanAfterDropSchema();
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsDropVersion( const OmsVersionId& versionId )
{
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsDropVersion : " << CFixedLenChar_co102(versionId, sizeof(versionId)));
  OmsContext_co12*   context;
  CriticalSection_co12 region(this->m_pSession, RGN_VERSION_DIR);
  region.Enter();
  context = Glob_co10.VersionDictionary.FindVersion(CVersionId_co17(versionId));
  if (NULL == context) {
    m_pSession->ThrowDBError (e_unknown_version, "omsDropVersion");
  }
  Glob_co10.VersionDictionary.MarkNotUnloadable(this->m_pSession->m_lcSink, context);
  m_pSession->DropVersion(context);
  Glob_co10.VersionDictionary.DropVersion(CVersionId_co17(versionId)); 
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsForceDropVersion( const OmsVersionId& versionId )
{
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsDropVersion : " << CFixedLenChar_co102(versionId, sizeof(versionId)));
  OmsContext_co12*   context;
  CriticalSection_co12 region(this->m_pSession, RGN_VERSION_DIR);
  region.Enter();
  context = Glob_co10.VersionDictionary.FindVersion(CVersionId_co17(versionId));
  if (NULL == context) {
    m_pSession->ThrowDBError (e_unknown_version, "omsDropVersion");
  }
  Glob_co10.VersionDictionary.MarkNotUnloadable(this->m_pSession->m_lcSink, context);
  try {
    m_pSession->DropVersion(context);
    Glob_co10.VersionDictionary.DropVersion(CVersionId_co17(versionId));
  }
  catch (DbpError& e) {
    if (-5001 == e.dbpError()) {
      context->MarkDropped();
    }
    else {
      throw e;
    }
  }
}
/*----------------------------------------------------------------------*/

void OmsHandle::omsDump () {
  m_pSession->CurrentContext()->DumpObjCache(*this);
};

/*----------------------------------------------------------------------*/

bool OmsHandle::omsExistsVersion( const OmsVersionId& versionId )
{
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsExistsVersion : " << CFixedLenChar_co102(versionId, sizeof(versionId)));
  CriticalSection_co12 region(this->m_pSession, RGN_VERSION_DIR);
  region.Enter();
  OmsContext_co12* context = Glob_co10.VersionDictionary.FindVersion(CVersionId_co17(versionId));
  if ((NULL != context) && (context->IsDropped())) {  /* PTS 1107209 */ /* PTS 1107277 */
    region.Leave();
    try {
      omsForceDropVersion(versionId);
      context = NULL;
    }
    catch(DbpError&) {
    }
  }
  return (context != NULL);
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsExceptionHandler(DbpError& e) {
  switch (e.m_errorKind) {
  case DbpError::DB_ERROR :
    {
      omsRaiseError ((short) e.m_errorNo, &e.m_errorText[0]);
      break;
    }
  case DbpError::HRESULT_ERROR :
    {
      (*m_session)->HResultError (e.m_errorNo);
      break;
    }
  case DbpError::RTE_ERROR :
    {
      omsRaiseError ((short) e.m_errorNo, &e.m_errorText[0]);
      break;
    } 
  case DbpError::USER_DEFINED :
    {
      omsRaiseError ((short) e.m_errorNo, &e.m_errorText[0]);
      break;
    } 
  default :
    dbpOpError ("Invalid DbpError, kind = %d, errno = %d", e.m_errorKind, e.m_errorNo);
    throw;
  }
} 

/*----------------------------------------------------------------------*/

OmsContainerNo OmsHandle::omsGetContainerNo(const OmsObjectId& oid) {
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsGetContainerNo : " << oid);
  omsIfNilOidThrowError (this, oid, "omsGetContainerNo");
  OmsObjectContainerPtr pObj = m_pSession->CurrentContext()->GetObj(oid, false);
  if (NULL != pObj) {
    tco13_ClsIdEntry* pContainerInfo = pObj->GetContainerInfo(m_pSession->CurrentContext());
    return pContainerInfo->GetContainerNo();
  }
  else {
    m_pSession->ThrowDBError (e_object_not_found, "omsGetContainerNo", oid);
    return 0;
  }
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsGetMonitorInfo (tgg00_COMMonitorTransferBuf* lpMonInfo) {

  if (lpMonInfo->ctbDispId_gg00 < 0) {
    // reset statistics
    m_monitor->Reset();
    return;
  }
  while ((lpMonInfo->ctbDispId_gg00 < m_monitor->size()) &&  
    (0 == (*m_monitor)[lpMonInfo->ctbDispId_gg00].m_monInfo.cmiCallCnt_gg00)) {
    ++lpMonInfo->ctbDispId_gg00;
  }
  if (lpMonInfo->ctbDispId_gg00 >= m_monitor->size()) {
    lpMonInfo->ctbDispId_gg00 = -1;
    return;
  }
  memcpy (&lpMonInfo->ctbMonInfo_gg00, 
    &((*m_monitor)[lpMonInfo->ctbDispId_gg00].m_monInfo),
    sizeof (lpMonInfo->ctbMonInfo_gg00));
  if (csp_maxint4 == lpMonInfo->ctbMonInfo_gg00.cmiCounters_gg00[cmiMinWaitNewConsView]) {
    lpMonInfo->ctbMonInfo_gg00.cmiCounters_gg00[cmiMinWaitNewConsView] = 0;
  }
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsGetObj(const OmsObjectId& oid, unsigned char* cnsView, int& ObjLen, unsigned char* obj)
// exported for diagnostics only, not included into interface
{
  tgg91_TransNo cView;
  tgg91_TransNo currCView = m_pSession->m_defaultContext->m_consistentView;
  memcpy (&cView, cnsView, sizeof(cView));
  m_pSession->m_defaultContext->SetConsistentView(cView);
  m_pSession->m_defaultContext->SetConsistentView(cView);
  OmsObjectContainerPtr pObj = m_pSession->m_defaultContext->GetObjFromLiveCacheBase(oid, false);
  if (NULL != pObj) {
    tco13_ClsIdEntry* pContainerInfo = pObj->GetContainerInfo(m_pSession->m_defaultContext);
    ObjLen = pContainerInfo->GetClassInfoPtr()->GetObjectSize() - OmsObjectContainer::headerSize() - sizeof(void*);
    memcpy (obj, &pObj->m_pobj, ObjLen); 
  }
  else {
    ObjLen = 0;
    obj    = NULL;
  }
  m_pSession->m_defaultContext->SetConsistentView(currCView);
}

/*----------------------------------------------------------------------*/

OmsSchemaHandle OmsHandle::omsGetSchemaHandle(const char* SchemaName) const {
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsGetSchemaHandle : " << SchemaName);
  tsp00_KnlIdentifier SchemaIdentifier;
  tsp00_Int4          SchemaId = 0;
  tgg00_BasisError    DBError;
  co10_ToKnlIdentifier(SchemaName, SchemaIdentifier);
  CHResult_co001 hr = (*m_session)->ExistsSchema (&SchemaIdentifier, &SchemaId, &DBError);
  if ( 0 != DBError ) {
    m_pSession->ThrowDBError (DBError, "omsGetSchemaHandle");
  }
  return SchemaId;
}

/*----------------------------------------------------------------------*/

OmsSchemaHandle OmsHandle::omsGetSchemaHandle(const OmsObjectId& oid) {
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsGetSchemaHandle : " << oid);
  omsIfNilOidThrowError (this, oid, "omsGetSchemaHandle");
  OmsObjectContainerPtr pObj = m_pSession->CurrentContext()->GetObj(oid, false);
  if (NULL != pObj) {
    tco13_ClsIdEntry* pContainerInfo = pObj->GetContainerInfo(m_pSession->CurrentContext());
    return pContainerInfo->GetSchema();
  }
  else {
    m_pSession->ThrowDBError (e_object_not_found, "omsGetSchemaHandle", oid);
    return 0;
  }
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsGetSchemaName(OmsSchemaHandle Schema, char* SchemaName, int SchemaNameBufSize) const {
  
  if (SchemaNameBufSize > 0) {
    tsp00_KnlIdentifier Identifier;
    tgg00_BasisError    DBError;
    CHResult_co001 hr = m_pSession->m_lcSink->GetSchemaName(Schema, &Identifier, &DBError);
    int len = sizeof(Identifier);
    if (Identifier[len-1] == 0) {
      len = strlen(Identifier);
    }
    if (len > SchemaNameBufSize - 1) {
      m_pSession->ThrowDBError (e_buffer_too_small, "omsGetSchemaName");
    }
    memcpy (SchemaName, &Identifier[0], len);
    SchemaName[len] = 0;
  }
}

/*----------------------------------------------------------------------*/

int OmsHandle::omsGetTimeOut() const {
  return m_pSession->m_timeout;
}

/*----------------------------------------------------------------------*/

OmsCallbackInterface* OmsHandle::omsGetTransactionInterface (const GUID& guid) const {
  return m_pSession->GetCallbackInterface(guid);
}

/*----------------------------------------------------------------------*/

unsigned long OmsHandle::omsGetVarObjectSize (const OmsVarOid& oid) 
{
  OMS_TRACE (omsTrInterface | omsTrVarObject, m_pSession->m_lcSink, "omsGetVarObjSize : " << oid);
  return m_pSession->GetVarObjSize(oid);
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsInitSession (IUnknown** s) {
  m_session = (IliveCacheSink**) s;
}

/*----------------------------------------------------------------------*/

bool OmsHandle::omsIsLocked (const OmsObjectId& oid) {
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsIsLocked : " << oid);
  return m_pSession->IsLocked(oid);
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsLoadVarObject (const OmsVarOid& oid, unsigned long size, void* buf) {
  OMS_TRACE (omsTrInterface | omsTrVarObject, m_pSession->m_lcSink, "omsLoadVarObject : " << oid << ", size " << size);
  omsIfNilOidThrowError (this, oid, "omsLoadVarObject");
  m_pSession->LoadVarObject(oid, VarObjExclusiveLock, size, buf); 
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsLock( const OmsObjectId&oid )
{
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsLock : " << oid);
  m_pSession->CurrentContext()->LockObj(oid);
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsLock( const OmsVarOid& oid) { 
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsLock(Var) : " << oid);
  m_pSession->CurrentContext()->LockObj((OmsObjectId) oid);
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsLocks( const OmsObjectId*p, int n )
{ 
  _TRACE_METHOD_ENTRY(this,"omsLocks")
    
  tsp00_Int4 ix;
  for (ix = 0; ix < n; ++ix) {
    omsLock(*p);
    ++p;
  }
  return;
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsMarkVersion ( const OmsVersionId& versionId ) {
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsMarkVersion : " << CFixedLenChar_co102(versionId, sizeof(versionId)));
  OmsContext_co12*     context;
  CriticalSection_co12 region(m_pSession, RGN_VERSION_DIR);
  region.Enter();
  context = Glob_co10.VersionDictionary.FindVersion(CVersionId_co17(versionId));
  if (NULL == context) {
    m_pSession->ThrowDBError (e_unknown_version, "omsOpenVersion");
  }
  context->MarkVersion();
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsNewConsistentView () {
  
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsNewConsistentView : ");
  if (m_pSession->CurrentContext()->IsVersion()) {
    return;
  }
  m_pSession->CurrentContext()->RemoveUnlockedObjFromCache();
  m_pSession->CurrentContext()->NewConsistentView();
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsNewConsistentView (OmsObjectId* pOids, int cntOids, short timeout)
{
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsNewConsistentView : " << "cnt " << cntOids << "timeout " << timeout);
   int dummy = 0;
   omsNewConsistentView (pOids, cntOids, timeout, NULL, dummy);
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsNewConsistentView (OmsObjectId* pOids, int cntOids, short timeout, OmsObjectId* pErrOids, int& cntErrOids) {
  
  gg200QuickSort(&pOids[0], cntOids);
  
  m_pSession->NewConsistentView(pOids, cntOids, timeout, pErrOids, cntErrOids);
}

/*----------------------------------------------------------------------*/

OmsAbstractObject* OmsHandle::omsNewKeyedObject(const GUID&          guid, 
                                                const unsigned char* key,
                                                OmsSchemaHandle      Schema,
                                                OmsContainerNo       ContainerNo )
{
  
  _TRACE_METHOD_ENTRY(this,"omsNewKeyObject");

  return m_pSession->NewKeyedObject (guid, key, Schema, ContainerNo);
}

/*----------------------------------------------------------------------*/

OmsAbstractObject* OmsHandle::omsNewObject( const GUID& guid, OmsSchemaHandle Schema, OmsContainerNo ContainerNo )
{
  
  _TRACE_METHOD_ENTRY(this,"omsNewObject")
  return m_pSession->NewObject(guid, Schema, ContainerNo);    
}

/*----------------------------------------------------------------------*/

OmsAbstractObject* OmsHandle::omsNewRegistryObject (size_t sz, const GUID guid) {
  
  OmsObjectContainerPtr p;
  sz += OmsObjectContainer::headerSize();
  if (m_pSession->InVersion()) {
    p = (OmsObjectContainerPtr) m_pSession->m_context->allocate(sz);
  }
  else {
    p = (OmsObjectContainerPtr) m_pSession->StackHeapMalloc(sz);
  }
  if (p) {
    p->InitState(REINTERPRET_CAST(tco13_GuidEntry*, NULL));
    return &p->m_pobj;
  }
  return NULL;  
}

/*----------------------------------------------------------------------*/

const OmsVarOid& OmsHandle::omsNewVarObject(OmsSchemaHandle Schema, OmsContainerNo ContainerNo)
{
  OMS_TRACE (omsTrInterface | omsTrVarObject, m_pSession->m_lcSink, "omsNewVarObj : " << Schema << ", " << ContainerNo);
  m_pSession->RegVarObjectContainer (Schema, ContainerNo);
  return m_pSession->NewVarObject(Schema, ContainerNo);
}

/*----------------------------------------------------------------------*/

void  OmsHandle::omsPrintOid (char* buf, size_t bufSize, const OmsObjectId oid)
{

   char auxbuf[256];
   sprintf ( auxbuf, "OID : %d.%d (vers %d)", oid.getPno(), oid.getPos(), oid.getVersion());
   if (strlen(auxbuf) < bufSize) {
     strcat (buf, auxbuf);
   }
}

/*----------------------------------------------------------------------*/



void OmsHandle::omsOidInfo (const OmsObjectId& oid, GUID& Guid, 
                            OmsSchemaHandle& Schema, char* SchemaName, int SchemaNameBufSize,
                            char* ClassName, int ClassNameBufSize, OmsContainerNo& ContainerNo)
{
  
  omsIfNilOidThrowError (this, oid, "omsLoadVarObject");
  OmsObjectContainerPtr pObj = m_pSession->CurrentContext()->GetObj(oid, false);
  if (NULL != pObj) {
    tco13_ClsIdEntry* pContainerInfo = pObj->GetContainerInfo(m_pSession->CurrentContext());
    Guid        = pContainerInfo->GetGuid();
    Schema      = pContainerInfo->GetSchema();
    ContainerNo = pContainerInfo->GetContainerNo();
    omsGetSchemaName(Schema, SchemaName, SchemaNameBufSize);
    if (ClassNameBufSize > 0) {
      int ClassNameLength = strlen(pContainerInfo->GetClassInfoPtr()->GetClassName());
      if (ClassNameLength >= ClassNameBufSize) {
        ClassNameLength = ClassNameBufSize - 1;
      }
      memcpy (ClassName, pContainerInfo->GetClassInfoPtr()->GetClassName(), ClassNameLength); 
      ClassName[ClassNameLength] = 0;
    }
  }
  else {
    m_pSession->ThrowDBError (e_object_not_found, "omsOidInfo", oid);
  }
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsOpenVersion( const OmsVersionId& versionId )
{
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsOpenVersion : " << CFixedLenChar_co102(versionId, sizeof(versionId)));
  tsp00_Int2              DBError;
  tgg00_OmsVersionContext versionContext;
  OmsContext_co12*        context;
  const char*             msg = "omsOpenVersion";
  CriticalSection_co12    region(m_pSession, RGN_VERSION_DIR);
  region.Enter();
  context = Glob_co10.VersionDictionary.FindVersion(CVersionId_co17(versionId));
  if (NULL == context) {
    m_pSession->ThrowDBError (e_unknown_version, msg);
  }
  if (context->IsDropped()) {
    region.Leave();
    try {
      omsForceDropVersion(versionId);
    }
    catch(DbpError&) {
    }
    m_pSession->ThrowDBError (e_unknown_version, msg);
  }
  if (m_pSession->IsSubtransOpen()) {
    m_pSession->ThrowDBError (e_too_many_subtrans, "Open Version : Open Subtrans"); 
  }
  if ((context->IsBoundToTrans()) && (!m_pSession->VersionBoundByMe(context))) {
      m_pSession->ThrowDBError (e_version_already_open, msg);
  }
  if (m_pSession->InVersion()) {
    m_pSession->ThrowDBError (e_already_in_version, msg); 
  }
  CHResult_co001 hr = (*m_session)->CreateOpenVersion (
    (unsigned char*) &context->m_consistentView,
    (OmsVersionId*)  &versionId,
    false,
    (unsigned char*) &versionContext,
    &DBError);
  if ( DBError != 0 ) {
    m_pSession->ThrowDBError (DBError, msg);
  }
  m_pSession->ClearDefaultContext();
  m_pSession->OpenVersion(context);
  m_pSession->IncOpenVersion();
  Glob_co10.VersionDictionary.MarkNotUnloadable(m_pSession->m_lcSink, context);
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsRaiseError (short msgno, const char* errmsg) {
  _TRACE_METHOD_ENTRY(this,"omsRaiseError");
  m_pSession->m_lcSink->SetError (msgno, strlen (errmsg), 
    (unsigned char*) errmsg, csp_ascii );
};

/*----------------------------------------------------------------------*/

void OmsHandle::omsReadOnly (bool read_only) {
  m_pSession->m_read_only = read_only;
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsRegClass (OmsSchemaHandle    Schema,
                             const GUID&        guid,
                             const char*        ClassName, 
                             size_t             PersistentSize,
                             size_t             ObjectSize,
                             const GUID*        pBaseClass,
                             OmsAbstractObject* Obj, 
                             OmsContainerNo     ContainerNo)
{
  m_pSession->RegContainer(Schema, guid, ClassName, PersistentSize, ObjectSize, pBaseClass, Obj,  ContainerNo);
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsRegClassAndKey (OmsSchemaHandle    Schema, 
                                   const GUID&        guid,
                                   const char*        ClassName,
                                   size_t             PersistentSize, 
                                   size_t             KeySize, 
                                   size_t             ObjectSize, 
                                   OmsAbstractObject* Obj, 
                                   OmsContainerNo     ContainerNo) {
  
  _TRACE_METHOD_ENTRY(this,"OmsHandle::omsRegClassAndKey");
  struct VirtualObject {
    void* vtptr;
  };    
  tco13_ClsIdEntry* pContainerInfo;
  bool              isRegistered = false;
  
  try {
    pContainerInfo = m_pSession->GetClsInfo (guid, Schema, ContainerNo);
    isRegistered = true;
  }
  catch (DbpError& e) {
    if (e_unknown_guid != e.dbpError()) {
      throw e;
    }
  }
  if (!isRegistered) {
    tgg00_ObjColDesc    InvColDesc;
    tgg00_StackEntry    KeyDesc;
    ObjectSize     += (OmsObjectContainer::headerSize() + KeySize);
    PersistentSize += (KeySize - /* virtual table ptr */ sizeof (void*)); 
    InvColDesc.colHashCount_gg00 = 1;
    InvColDesc.colInvCount_gg00  = 0;
    InvColDesc.colHashList_gg00  = REINTERPRET_CAST(tgg00_StackListPtr, &KeyDesc);
    InvColDesc.colInvList_gg00   = NULL;
    KeyDesc.etype().becomes(st_object_col);
    KeyDesc.eop().becomes(op_none);
    KeyDesc.epos()     = PersistentSize - KeySize + 1;
    KeyDesc.elen_var() = KeySize;
    KeyDesc.ecol_pos() = -1;
    ContainerInfo_co18* p = Glob_co10.ClassDictionary.RegisterContainer (*m_session, guid, ClassName, PersistentSize,
      ObjectSize, NULL, InvColDesc, Schema, ContainerNo, (REINTERPRET_CAST(VirtualObject*, Obj))->vtptr);
    pContainerInfo = omsAddContainerInfo (p);
  }
  pContainerInfo->chainFree(*m_pSession->CurrentContext(), Obj);
};

/*----------------------------------------------------------------------*/

void OmsHandle::omsRegMonitor (void* pMon) { 
  m_monitor = REINTERPRET_CAST(co90_Monitor*, (pMon)); 
} 

/*----------------------------------------------------------------------*/

void OmsHandle::omsRegTransactionInterface (const GUID& guid, OmsCallbackInterface* pInterface) {
  m_pSession->RegisterCallbackInterface(guid, pInterface);
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsRelease( const OmsObjectId& oid ) {
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsRelease : " << oid);
  m_pSession->ReleaseObj(oid);
};

/*----------------------------------------------------------------------*/

void OmsHandle::omsRelease ( const OmsVarOid& oid ) {
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsRelease : " << oid);
  m_pSession->ReleaseObj ((OmsObjectId) oid); 
};

/*----------------------------------------------------------------------*/

void OmsHandle::omsReleaseAll() {
  
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsReleaseAll");
  
  const bool emptyObjCache = true;
  if (!m_pSession->IsSubtransOpen()) {
    if(m_pSession->IsDefaultContext(m_pSession->CurrentContext())) {
      m_pSession->CurrentContext()->FlushObjCache(emptyObjCache);
    }
    m_pSession->FreeLockObjAndBeforeImages(); 
  }
}

/*----------------------------------------------------------------------*/

HRESULT OmsHandle::omsReleaseDefaultContext() {
  if (NULL != m_monitor) {
    CriticalSection_co12 region(this->m_pSession, RGN_MONITOR);
    region.Enter();
    Glob_co10.Monitor.Insert(m_monitor); /* PTS 1107731 */
    region.Leave();
  }
  OMS_TRACE(omsTrSession, (*m_session), " Release Session " << m_pSession << " of Hndl " << this << " RefCnt " << m_pSession->m_refCnt);
  m_pSession->DeleteSession();
  return S_OK;
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsReleaseUnconditional( const OmsObjectId& oid) {
  m_pSession->ReleaseObjUnconditional (oid); 
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsReleaseVarObject(const OmsVarOid& oid) {
  OMS_TRACE (omsTrInterface | omsTrVarObject, m_pSession->m_lcSink, "omsReleaseVarObject : " << oid);
  m_pSession->ReleaseVarObject(oid);
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsReleaseConsistentView() {
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsReleaseConsistentView");
  tsp00_Int2  DummyError;
  tsp00_Int4  BufLen = 0;
  tsp00_Int4  Buf;
  CHResult_co001 hr = (*m_session)->MultiPurpose (m_end, mm_nil, &BufLen, 
    (unsigned char*) &Buf, &DummyError);
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsRollback (bool doRelease) {

  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsRollback");
  m_pSession->RollbackSubtrans(1);
  m_pSession->FreeLockObjAndBeforeImages();
  m_pSession->ClearDefaultContext();
  OmsCallbackInterface* callbackInterface = m_pSession->GetCallbackInterface();
  if (NULL != callbackInterface) {
    bool read_only = m_pSession->m_read_only;
    m_pSession->m_read_only = true;
    try {
      callbackInterface->omsInvalidate (*this, OmsCallbackInterface::AfterRollback,
        m_pSession->InVersion());
    }
    catch (DbpError e) {
      dbpOpError("omsInvalidate failed with DbpError %d", e.dbpError());
      m_pSession->m_currVarObjChunk.m_oid.setNil();
      m_pSession->ResetSubtransLevel();
      m_pSession->m_read_only = read_only;
      m_pSession->ClearDefaultContext();
      throw e;
    }
    catch (...) {
      dbpOpError("omsInvalidate failed with unknown exception");
      m_pSession->m_currVarObjChunk.m_oid.setNil();
      m_pSession->ResetSubtransLevel();
      m_pSession->m_read_only = read_only;
      m_pSession->ClearDefaultContext();
      throw;
    }
    m_pSession->m_currVarObjChunk.m_oid.setNil();
    m_pSession->m_read_only = read_only;
    if (!m_pSession->InVersion()) {
      m_pSession->ResetSubtransLevel();
      m_pSession->ClearDefaultContext();
    }
  }
  if (doRelease) {
    this->SessionEnd();  /* PTS 1107209 */
  }
  m_pSession->TransEnd();
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsRollbackSubtrans( int required_lvl ) {
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsRollbackSubtrans : " << required_lvl);
  if (!m_pSession->IsSubtransOpen()) {
    return;
  }
  m_pSession->IncSubtransRollback();
  m_pSession->RollbackSubtrans(required_lvl);
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsRenameVersion ( const OmsVersionId& oldId, const OmsVersionId& newId) {
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsRenameVersion : " << CFixedLenChar_co102(oldId, sizeof(oldId)) 
    << " to " << CFixedLenChar_co102(newId, sizeof(newId)));
 
  const char* msg = "omsRenameVersion";

  CriticalSection_co12 region(this->m_pSession, RGN_VERSION_DIR);
  region.Enter();
  OmsContext_co12* context = Glob_co10.VersionDictionary.FindVersion(CVersionId_co17(oldId));
  if (NULL == context) {
    m_pSession->ThrowDBError(e_unknown_version, msg);
  }
  tgg00_BasisError e = Glob_co10.VersionDictionary.InsertVersion(CVersionId_co17(newId), context);
  if (e_ok != e) {
    if (e_duplicate_key == e) {
      e = e_duplicate_name;
    }
    m_pSession->ThrowDBError (e_duplicate_name, msg);
  }
  Glob_co10.VersionDictionary.DropVersion(CVersionId_co17(oldId)); 
  context->RenameVersion (oldId, newId);
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsSetTimeOut( short timeout ) {
  tsp00_Int2  DummyError;
  tsp00_Int4  BufLen = sizeof(timeout);
  m_pSession->m_timeout = timeout;
  CHResult_co001 hr = (*m_session)->MultiPurpose (m_set, mm_lock_timeout, &BufLen, 
    (unsigned char*) &timeout, &DummyError);
}

/*----------------------------------------------------------------------*/

int OmsHandle::omsStartSubtrans() {
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsStartSubtrans");
  return m_pSession->StartSubtrans();
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsStoreVarObject ( const OmsVarOid& oid, const void* pVarObj, unsigned long objLength)
{
  OMS_TRACE (omsTrInterface | omsTrVarObject, m_pSession->m_lcSink, "omsStoreVarObject : " << oid << ", length is " << objLength);
  m_pSession->StoreVarObject(oid, pVarObj, objLength);
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsTerminate() {
  dbpOpMsg("omsTerminate called");
  tgg00_BasisError e = e_terminate; /* PTS 1107666 */
  tsp00_Int4  BufLen = sizeof(e);
  tsp00_Int2  SqlError;
  // transform basis error into sql return code
  m_pSession->m_lcSink->MultiPurpose (m_change, mm_nil, &BufLen, 
    (unsigned char*) &e, &SqlError);
  omsRaiseError (SqlError, "");
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsTrace (const char* tracemsg, long TraceMsgLen) {
  if (NULL != *m_session)
    (*m_session)->Vtrace (TraceMsgLen, (unsigned char*) tracemsg);
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsTracef( const char*fmt, ... ) {
#if defined(WIN32)
  char buffer[256];
#else
  char buffer[1024];
#endif

  va_list args;
  va_start (args, fmt);
#if defined(WIN32)
  _vsnprintf ( buffer, sizeof(buffer), fmt, args );
#else
  vsprintf ( buffer, fmt, args );
  buffer[256] = '\0';
#endif
  omsTrace (buffer, strlen(buffer));
  va_end (args);
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsTraceOid (const OmsObjectId& oid) {
  omsTracef ("OID : %d.%d (vers %d)", oid.getPno(), oid.getPos(), oid.getVersion());
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsTransRollback() {
  tsp00_Int2 DBError;
  CHResult_co001 hr = (*m_session)->Rollback (0, NULL, &DBError);
  if ( DBError != 0 ) {
    m_pSession->ThrowDBError (DBError, "omsTransRollback");
  }
  omsRollback(false);
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsUnlock (const OmsObjectId& oid)
{
  OMS_TRACE (omsTrInterface, m_pSession->m_lcSink, "omsUnlock : " << oid);
  m_pSession->UnlockObj(oid);
}

/*----------------------------------------------------------------------*/

void OmsHandle::omsValidate() {
};

/*----------------------------------------------------------------------*/

OmsHandle::~OmsHandle () {
  _TRACE_METHOD_ENTRY(this,"~OmsHandle");
}

/*----------------------------------------------------------------------*/
/* Implementation of OmsLockTimeout                                     */
/*----------------------------------------------------------------------*/

OmsLockTimeout::OmsLockTimeout(const OmsObjectId& oid) :
  DbpError(DbpError::DB_ERROR, -51, "Lock Request Timeout", oid) {
}

/*----------------------------------------------------------------------*/
/* Implementation of OmsOutOfDate                                       */
/*----------------------------------------------------------------------*/

OmsOutOfDate::OmsOutOfDate(const OmsObjectId& oid) :
  DbpError(DbpError::DB_ERROR, e_object_dirty, "Object out of date", oid) {
}

/*----------------------------------------------------------------------*/

OmsOutOfDate::OmsOutOfDate() :
  DbpError(e_object_dirty, "Object out of date") {
}

/*----------------------------------------------------------------------*/

//
// Implementation of OmsAbstractObject
//

OmsAbstractObject::OmsAbstractObject() {};

/*----------------------------------------------------------------------*/

OmsObjByClsIterBase OmsAbstractObject::omsAllOids( OmsHandle& h, const GUID& guid, OmsSchemaHandle Schema, OmsContainerNo ContainerNo) {
  OMS_TRACE(omsTrInterface, h.m_pSession->m_lcSink, "OmsAbstractObject::omsAllOids : " << guid << "," << Schema << "," << ContainerNo); 
  return h.omsAllOids (guid, Schema, ContainerNo);
};

/*----------------------------------------------------------------------*/

void OmsAbstractObject::omsCleanUp (OmsHandle&) {};

/*----------------------------------------------------------------------*/

void OmsAbstractObject::omsDelete (OmsHandle& Handle) {
 
  Handle.m_pSession->IncDelete ();
  OmsObjectContainerPtr pObjContainer = omsGetContainer(this);
  
  OMS_TRACE(omsTrInterface, Handle.m_pSession->m_lcSink, "OmsAbstractObject::omsDelete : " << pObjContainer->m_oid);
  
  if (pObjContainer->DeletedFlag()) {
    Handle.m_pSession->ThrowDBError (e_object_not_found, "OmsAbstractObj::omsDelete", pObjContainer->m_oid);
  }
  omsCleanUp (Handle);
  pObjContainer->MarkDeleted();
}

/*----------------------------------------------------------------------*/

void OmsAbstractObject::omsDeleteAll(OmsHandle& h, const GUID& guid, OmsSchemaHandle Schema, OmsContainerNo ContainerNo) {
  OMS_TRACE(omsTrInterface, h.m_pSession->m_lcSink, "OmsAbstractObject::omsDeleteAll : "
    << guid << "," << Schema << "," << ContainerNo);
  h.omsDeleteAll (guid, Schema, ContainerNo);
}

/*----------------------------------------------------------------------*/

const OmsAbstractObject* OmsAbstractObject::omsDeRef (const OmsObjectId& oid, OmsHandle& Handle, const GUID& guid) {
  OMS_TRACE(omsTrInterface, Handle.m_pSession->m_lcSink, "OmsAbstractObject::omsDeRef : " << oid << ", " << guid); 
#if defined(TEST_INLINE)
  OmsObjectContainerPtr p = Handle.m_pSession->m_context->FindObjInContext(&oid);
  if (NULL != p) {
    if (p->DeletedFlag()) {
      p = NULL;
    }
  }
  else {
    if (oid) {
      p = Handle.m_pSession->m_context->GetObjFromLiveCacheBase(oid, false);
    }
    else {
      return NULL;
    }
  }
  if (NULL != p) {
    Handle.m_pSession->m_context->CheckCompatibility(guid, p);
    return &p->m_pobj;
  }
  else {
    Handle.m_pSession->ThrowDBError (e_object_not_found, "OmsAbstractObject::omsDeRef", oid);
  }
  return NULL;
#else
  if (!oid) {
    return NULL;
  }
  else {
    return REINTERPRET_CAST(OmsAbstractObject*, Handle.m_pSession->DeRef (oid, guid, !FOR_UPD_CO10, false));
  }
#endif
};

/*----------------------------------------------------------------------*/

OmsAbstractObject* OmsAbstractObject::omsDeRefForUpd (const OmsObjectId& oid, OmsHandle& Handle, const GUID& guid, bool do_lock) {
  OMS_TRACE(omsTrInterface, Handle.m_pSession->m_lcSink, "OmsAbstractObject::omsDeRefForUpd : " << oid 
    << ", " << guid << ", do_lock : " << do_lock); 
  if (oid) {
    return REINTERPRET_CAST(OmsAbstractObject*, Handle.m_pSession->DeRef(oid, guid, FOR_UPD_CO10, do_lock));
  }
  else {
    return NULL;
  }
};

/*----------------------------------------------------------------------*/

const OmsAbstractObject* OmsAbstractObject::omsDeRefKey (unsigned char*  key,
                                                         OmsHandle&      Handle, 
                                                         const GUID&     guid, 
                                                         OmsSchemaHandle h,
                                                         OmsContainerNo  ContainerNo) 
{
  OMS_TRACE(omsTrInterface, Handle.m_pSession->m_lcSink, "OmsAbstractObject::omsDeRefKey : " << guid
    << "CNo : " << ContainerNo << " Key : " << CUnsignedChar_co102(key, Handle.m_pSession->GetClassInfo(guid)->GetKeyLen()));
  return REINTERPRET_CAST(OmsAbstractObject*, Handle.m_pSession->DeRefViaKey(key, guid, !FOR_UPD_CO10, false, h, ContainerNo));
};

/*----------------------------------------------------------------------*/

OmsAbstractObject* OmsAbstractObject::omsDeRefKeyForUpd (unsigned char*  key,
                                                         OmsHandle&      Handle, 
                                                         const GUID&     guid, 
                                                         bool            do_lock,
                                                         OmsSchemaHandle h,
                                                         OmsContainerNo  ContainerNo)
{
  OMS_TRACE(omsTrInterface, Handle.m_pSession->m_lcSink, "OmsAbstractObject::omsDeRefKeyForUpd : " << guid
    << "CNo : " << ContainerNo << " Key : " << CUnsignedChar_co102(key, Handle.m_pSession->GetClassInfo(guid)->GetKeyLen())
    << " do_lock : " << do_lock);
  return REINTERPRET_CAST(OmsAbstractObject*, Handle.m_pSession->DeRefViaKey(key, guid, FOR_UPD_CO10, do_lock, h, ContainerNo));
};

/*----------------------------------------------------------------------*/

void OmsAbstractObject::omsDump (OmsHandle& Handle, const char* flags) const {
  OmsObjectId oid = omsOid();
  Handle.omsTracef ("OID : %d.%d (vers %d) %s", 
    oid.getPno(), oid.getPos(), oid.getVersion(), flags); 
}

/*----------------------------------------------------------------------*/

void OmsAbstractObject::omsFlush (OmsHandle& Handle) {
  OmsObjectContainerPtr pObjContainer = omsGetContainer(this);
  Handle.m_pSession->m_context->FlushObj(pObjContainer);
};

/*----------------------------------------------------------------------*/

const OmsObjectId& OmsAbstractObject::omsGetOid() const {
  return omsOid();
}

/*----------------------------------------------------------------------*/

OmsAbstractObject* OmsAbstractObject::omsForUpdPtr(OmsHandle& h, bool doLock) const {
  OmsObjectContainerPtr pObjContainer = omsGetContainer(this);
  if (h.m_pSession->m_read_only) {
    h.m_pSession->ThrowDBError (e_file_read_only, "omsLock");
  }
  if (doLock) {
    if (!(pObjContainer->LockedFlag()) && (!h.m_pSession->InVersion())) {
      ((OmsAbstractObject*) this)->omsLock(h);
    }
  }
  h.m_pSession->InsertBeforeImage (pObjContainer);
  return (OmsAbstractObject*) this;
}

/*----------------------------------------------------------------------*/

OmsSchemaHandle OmsAbstractObject::omsGetSchemaHandle(OmsHandle& Handle) const {
  OmsObjectContainerPtr pObjContainer = omsGetContainer(this);
  tco13_ClsIdEntry* pContainerInfo = pObjContainer->GetContainerInfo(Handle.m_pSession->CurrentContext());
  return pContainerInfo->GetSchema();
}

/*----------------------------------------------------------------------*/

bool OmsAbstractObject::omsHistoryInUse(OmsHandle& h) const {
  OmsObjectContainerPtr pObjContainer = omsGetContainer(this);
  OMS_TRACE(omsTrInterface, h.m_pSession->m_lcSink, "OmsAbstractObject::omsHistoryInUse : " << pObjContainer->m_oid);
  return h.m_pSession->HistoryInUse(pObjContainer->m_oid);
}

/*----------------------------------------------------------------------*/

bool OmsAbstractObject::omsIsLocked (OmsHandle& Handle) const {
  
  _TRACE_METHOD_ENTRY(&Handle,"OmsAbstractObject::omsIsLocked");

  return Handle.m_pSession->IsLocked(omsGetContainer(this)->m_oid);
}

/*----------------------------------------------------------------------*/

void OmsAbstractObject::omsKey(OmsHandle& Handle, unsigned char* key) const {

  OmsObjectContainerPtr pObjContainer = omsGetContainer(this);
  tco13_ClsIdEntry* pContainerInfo = pObjContainer->GetContainerInfo (Handle.m_pSession->CurrentContext());
  unsigned char *p = REINTERPRET_CAST(unsigned char*, CONST_CAST(OmsAbstractObject*, this));
  memcpy (key, pContainerInfo->GetKeyPtr(pObjContainer), pContainerInfo->GetKeyLen()); 
}

/*----------------------------------------------------------------------*/

void OmsAbstractObject::omsLock (OmsHandle& h) {
  
  OmsObjectContainerPtr pObjContainer = omsGetContainer(this);

  OMS_TRACE(omsTrInterface, h.m_pSession->m_lcSink, "OmsAbstractObject::omsLock : " << pObjContainer->m_oid);
  
  if (h.m_pSession->m_read_only) {
    h.m_pSession->ThrowDBError (e_file_read_only, "omsLock");
  }
  if ((pObjContainer->LockedFlag()) || (h.m_pSession->InVersion())) {
    // instance already locked
    return;
  }
  h.m_pSession->LockObj(pObjContainer);
}

/*----------------------------------------------------------------------*/

OmsAbstractObject* OmsAbstractObject::omsNewObject (OmsHandle&      Handle, 
                                                    const GUID&     guid, 
                                                    OmsSchemaHandle h, 
                                                    OmsContainerNo  ContainerNo ) {
  OMS_TRACE(omsTrInterface, Handle.m_pSession->m_lcSink, "OmsAbstractObject::omsNewObject : " << guid 
    << "CNo : " << ContainerNo);
  return Handle.omsNewObject ( guid, h, ContainerNo ); 
};

/*----------------------------------------------------------------------*/

OmsAbstractObject* OmsAbstractObject::omsNewRegistryObject(size_t sz, OmsHandle& Handle, const GUID& guid) {
  return Handle.omsNewRegistryObject (sz, guid);
};

/*----------------------------------------------------------------------*/

OmsAbstractObject* OmsAbstractObject::omsNewKeyedObject (OmsHandle&           Handle, 
                                                         const GUID&          guid, 
                                                         const unsigned char* key,
                                                         OmsSchemaHandle      h,
                                                         OmsContainerNo       ContainerNo ) {
  OMS_TRACE(omsTrInterface, Handle.m_pSession->m_lcSink, "OmsAbstractObject::omsNewKeyedObject : " << guid
    << " CNo : " << ContainerNo << " Key : " << CUnsignedChar_co102(key, Handle.m_pSession->GetClassInfo(guid)->GetKeyLen()));
  return Handle.omsNewKeyedObject ( guid, key, h, ContainerNo ); 
};

/*----------------------------------------------------------------------*/

const OmsObjectId& OmsAbstractObject::omsOid() const {
  return omsGetContainer(this)->m_oid;
};

/*----------------------------------------------------------------------*/

void OmsAbstractObject::omsRegClass (OmsHandle&         Handle,
                                     const GUID&        Guid, 
                                     const char*        ClassName, 
                                     size_t             PersistentSize, 
                                     size_t             ObjectSize,
                                     const GUID*        pBaseClass,
                                     OmsAbstractObject* Obj,
                                     OmsSchemaHandle    h,
                                     OmsContainerNo     ContainerNo) 
{
  Handle.omsRegClass (h, Guid, ClassName, PersistentSize, ObjectSize, pBaseClass, Obj, ContainerNo);
};

/*----------------------------------------------------------------------*/

void OmsAbstractObject::omsRegClassAndKey (OmsHandle&         Handle,
                                           const GUID&        Guid,
                                           const char*        ClassName, 
                                           size_t             PersistentSize, 
                                           size_t             KeySize, 
                                           size_t             ObjectSize,
                                           OmsAbstractObject* Obj,
                                           OmsSchemaHandle    h,
                                           OmsContainerNo     ContainerNo) 
{
  Handle.omsRegClassAndKey (h, Guid, ClassName, PersistentSize, KeySize, ObjectSize, Obj, ContainerNo);
};

/*----------------------------------------------------------------------*/

void OmsAbstractObject::omsRelease(OmsHandle& Handle) {
 
 OMS_TRACE(omsTrInterface, Handle.m_pSession->m_lcSink, "OmsAbstractObject::omsRelease : " << omsGetContainer(this)->m_oid);
 Handle.m_pSession->ReleaseObj(omsGetContainer(this));
}

/*----------------------------------------------------------------------*/

void OmsAbstractObject::omsStore (OmsHandle& Handle) {
 
  OMS_TRACE(omsTrInterface, Handle.m_pSession->m_lcSink, "OmsAbstractObject::omsStore : " << omsGetContainer(this)->m_oid);
  Handle.m_pSession->IncStore();
  omsVerify(Handle);
  
  OmsObjectContainerPtr pObjContainer = omsGetContainer(this);
  _TRACE_OID(&Handle, pObjContainer->m_oid);
  if (!(pObjContainer->existBeforeImage (Handle.m_pSession->CurrentSubtransLevel(), Handle.m_pSession->InVersion()))) {
    Handle.m_pSession->ThrowDBError (e_missing_before_image, "missing before image");
  }
  else {
    pObjContainer->MarkStored();
  }
}

/*----------------------------------------------------------------------*/

void OmsAbstractObject::omsVerify (OmsHandle& h) const {}

/*----------------------------------------------------------------------*/

OmsLockObject::OmsLockObject(OmsHandle& h) {
  m_omsHandle = &h;
  m_mode      = NO_LOCK;
  m_next      = h.m_pSession->m_lockObjects; 
  h.m_pSession->m_lockObjects = this;
  memset (&m_lockHandle, 0, sizeof(m_lockHandle));
}

/*----------------------------------------------------------------------*/

OmsLockObject::~OmsLockObject() {
  if (NULL != m_omsHandle) {
    OmsLockObject*  curr = m_omsHandle->m_pSession->m_lockObjects;
    OmsLockObject** prev = &m_omsHandle->m_pSession->m_lockObjects; 
    while (curr) {
      if (curr == this) {
        *prev = m_next;
        return;
      }
      prev = &m_next;
      curr = curr->m_next;
    }
  }
}

/*----------------------------------------------------------------------*/

void OmsLockObject::omsEnable () {
  if (NULL != m_omsHandle) {
    tsp00_Int2  sqlCode;
    tsp00_Int4  BufLen = sizeof(m_lockHandle);
    CHResult_co001 hr = (*m_omsHandle->m_session)->MultiPurpose (m_get, mm_key, &BufLen, 
      (unsigned char*) &m_lockHandle, &sqlCode);
    if (0 != sqlCode) {
      m_omsHandle->m_pSession->ThrowReturnCode (sqlCode, "omsEnable");
    }
  }
}

/*----------------------------------------------------------------------*/

void OmsLockObject::omsEnable (const OmsLockHandle& lh) {
  m_lockHandle = lh;
}

/*----------------------------------------------------------------------*/

const OmsLockHandle& OmsLockObject::omsGetLockHandle() const {
  return m_lockHandle;
}

/*----------------------------------------------------------------------*/

void OmsLockObject::omsLockShare() {
  if (NULL != m_omsHandle) {
    tsp00_Int2  sqlCode;
    tsp00_Int4  BufLen = sizeof(m_lockHandle);
    if (NO_LOCK == m_mode) {
      CHResult_co001 hr = (*m_omsHandle->m_session)->MultiPurpose (m_lock, mm_nil, &BufLen, 
        (unsigned char*) &m_lockHandle, &sqlCode);
      if (0 != sqlCode) {
        m_omsHandle->m_pSession->ThrowDBError (sqlCode, "omsLockShare");
      }
      m_mode = LOCK_SHARE;
    }
  }
}

/*----------------------------------------------------------------------*/

void OmsLockObject::omsLockExclusive() {
  if (NULL != m_omsHandle) {
    tsp00_Int2  sqlCode;
    tsp00_Int4  BufLen = sizeof(m_lockHandle);
    if (LOCK_EXCLUSIVE != m_mode) {
      CHResult_co001 hr = (*m_omsHandle->m_session)->MultiPurpose (m_lock, mm_lock, &BufLen, 
        (unsigned char*) &m_lockHandle, &sqlCode);
      if (0 != sqlCode) {
        m_omsHandle->m_pSession->ThrowDBError (sqlCode, "omsLockShare");
      }
      m_mode = LOCK_EXCLUSIVE;
    }
  }
}

/*----------------------------------------------------------------------*/

void OmsLockObject::omsUnlock() {
  if (NULL != m_omsHandle) {
    tsp00_Int2  sqlCode;
    tsp00_Int4  BufLen = sizeof(m_lockHandle);
    if (NO_LOCK != m_mode) {
      tgg00_MessType2 MessType2;
      if (LOCK_SHARE == m_mode) {
        MessType2.becomes(mm_nil);
      }
      else {
        MessType2.becomes(mm_lock);
      }
      CHResult_co001 hr = (*m_omsHandle->m_session)->MultiPurpose (m_unlock, MessType2, &BufLen, 
        (unsigned char*) &m_lockHandle, &sqlCode);
      if (0 != sqlCode) {
        m_omsHandle->m_pSession->ThrowDBError (sqlCode, "omsLockShare");
      }
      m_mode = NO_LOCK;
    }
  }
}

/*----------------------------------------------------------------------*/

bool  OmsLockObject::omsIsLocked() const {
  return (m_mode != NO_LOCK);
}

/*----------------------------------------------------------------------*/

bool  OmsLockObject::omsShareLocked() const {
  return (LOCK_SHARE == m_mode);
}

/*----------------------------------------------------------------------*/

bool  OmsLockObject::omsExclusiveLocked() const {
  return (LOCK_EXCLUSIVE == m_mode);
}

/*----------------------------------------------------------------------*/

void OmsLockObject::InvalidateHandle() {
  m_omsHandle = NULL;
}

/*----------------------------------------------------------------------*/

OmsLockObject* OmsLockObject::Next() const {
  return m_next;
}

/*----------------------------------------------------------------------*/
//
// Implementation of OmsStream
//

OmsStream::OmsStream(OmsHandle* h, OmsTypeABAPTabHandle& sh, long rsz, long sz, bool istream) {
  Init(h, sh, rsz, sz, istream);
};

/*----------------------------------------------------------------------*/

OmsStream::OmsStream(OmsHandle* h, OmsTypeStreamHandle& sh, long rsz, long sz, bool istream) {
  Init(h, *REINTERPRET_CAST(OmsTypeABAPTabHandle*,(&sh)), rsz, sz, istream);
};

/*----------------------------------------------------------------------*/

void OmsStream::Init(OmsHandle* h, OmsTypeABAPTabHandle& sh, long rsz, long sz, bool istream) {
  m_handle        = h;
  m_stream_handle = &sh;
  if (sz < 0) {
    tsp00_Int4  PacketSize;
    tsp00_Int2  DummyError;
    tsp00_Int4  BufLen = 0;
    CHResult_co001 hr = (*h->m_session)->MultiPurpose (m_get, mm_length, &BufLen, 
      (unsigned char*) &PacketSize, &DummyError);
    sz = PacketSize / rsz;
    if (sz > csp_maxint2) {
      sz = csp_maxint2;
    }
  }
  m_buf_size    = rsz * sz;
 
  m_row_size    = rsz;
  m_row_curr    = -1;
  m_row_count   = 0;
  if ((istream) && (-1 == m_stream_handle->rowCount)) {
    // empty input stream
    m_row_buf = NULL;
    m_eot     = true;
  }
  else {
    m_row_buf = REINTERPRET_CAST(unsigned char*, co10_malloc (*h->m_session, m_buf_size)); 
    m_eot     = false;
  }
  m_is_istream  = istream;
};

/*----------------------------------------------------------------------*/

OmsStream::~OmsStream() {
  co10_free(*m_handle->m_session, m_row_buf);
};

/*----------------------------------------------------------------------*/

void OmsStream::omsClose() {
  if (m_row_count > 0)
    omsFlush();
}

/*----------------------------------------------------------------------*/

void OmsStream::omsFlush() {
  
  _TRACE_METHOD_ENTRY(m_handle,"OmsStream::omsFlush")
    if (m_handle->m_pSession->m_stream_io) {
      tsp00_Int2 DBError;
      CHResult_co001 hr = (*m_handle->m_session)->ABAPWrite (&m_stream_handle->ABAPTabId,
        m_row_size, m_row_count, &m_row_buf[0], &DBError);
      m_row_count = 0;
      if ( DBError != 0 ) {
        m_handle->m_pSession->m_stream_io = false; 
        m_handle->m_pSession->ThrowDBError (DBError, "omsFlush (stream)");
      }
    }
}

/*----------------------------------------------------------------------*/

unsigned char* OmsStream::omsNxt() {
  
  _TRACE_METHOD_ENTRY(m_handle,"OmsStream::omsNxt");
  if (NULL == m_row_buf) {
    return NULL;
  }
  if (m_handle->m_pSession->m_stream_io) {
    if (++m_row_curr == m_row_count) {
      if (m_eot) {
        return NULL;
      }
      tsp00_Int2 DBError;
      tsp00_Int4 required_rows;
      required_rows   = m_buf_size / m_row_size;
      m_row_count     = required_rows;
      CHResult_co001 hr = (*m_handle->m_session)->ABAPRead (&m_stream_handle->ABAPTabId, 
        &m_row_count, &m_buf_size, &m_row_buf[0], &DBError);
      if ( DBError != 0 ) 
        if ( e_no_next_record == DBError )
          m_eot = true;
        else {
          m_handle->m_pSession->m_stream_io = false;
          m_handle->m_pSession->ThrowDBError (DBError, "omsNxt");
        }
        m_row_curr = 0;
        if (m_row_count == 0) {
          co10_free(*m_handle->m_session, m_row_buf);
          m_row_buf = NULL;
          return NULL;
        }
    }
  }
  if (m_stream_handle->colCount) {
    if (m_stream_handle->rowCount) {
      omsSwapABAPRow (&m_row_buf[m_row_curr * m_row_size]);
    }
    else {
      omsSwapStream (&m_row_buf[m_row_curr * m_row_size]);
    }
  }
  return &m_row_buf[m_row_curr * m_row_size];
}

/*----------------------------------------------------------------------*/

void OmsStream::omsSwapStream (unsigned char* p) {
  OmsTypeStreamMemberDesc *memDesc;
  unsigned char           *curr;
  int                     jx;
  int                     swapLength;
  unsigned char           aux_double[sizeof(double)];

  tsp00_SwapKind source_swap;
  tsp00_SwapKind destin_swap;
  source_swap.becomes(sw_normal);
  destin_swap.becomes(sw_full_swapped);
  for (int ix = 0; ix < m_stream_handle->colCount; ix++) {
    memDesc = REINTERPRET_CAST(OmsTypeStreamMemberDesc*, (&m_stream_handle->colDesc[ix]));
    swapLength = 0;
    switch (memDesc->memberType) {
    case  STYPE_I2  :
    case  STYPE_UI2 :
      swapLength = 2;
      break;
    case STYPE_I4  :
    case STYPE_UI4 :
    case STYPE_R4  : /* ADIS 1001759, T.A. 25.11.1998 */
      swapLength = 4;
      break;
    case STYPE_I8  : /* ADIS 1107790, T.A. 07.11.2000 */
    case STYPE_UI8 :
    case STYPE_R8  : /* ADIS 1001759, T.A. 25.11.1998 */
      swapLength = 8;
      break;
    }
    if (swapLength > 0) {
      curr = p + memDesc->offset;
      memcpy (&aux_double[0], curr, swapLength);
      for (jx = swapLength - 1; jx >= 0; jx--) {
        *curr = aux_double[jx];
        curr++;
      }
    }
  }
}

/*----------------------------------------------------------------------*/

void OmsStream::omsSwapABAPRow (unsigned char* p) {
  OmsTypeABAPColDesc      *abapDesc;
  unsigned char           *curr;
  tsp00_Int4              *pI4;
  unsigned char           aux;
  int                     jx;
  unsigned char           aux_double[sizeof(double)];

  tsp00_SwapKind source_swap;
  tsp00_SwapKind destin_swap;
  source_swap.becomes(sw_normal);
  destin_swap.becomes(sw_full_swapped);
  for (int ix = 0; ix < m_stream_handle->colCount; ix++) {
    abapDesc = &m_stream_handle->colDesc[ix];
    switch (abapDesc->abap_type) {
    case ABTYPINT2  :
      curr      = p + abapDesc->offset;
      aux       = *curr;
      *curr     = *(curr+1);
      *(curr+1) = aux;
      break;
    case ABTYPINT : 
      pI4 = REINTERPRET_CAST(tsp00_Int4*, (p + abapDesc->offset));
      s20sw_int4 (*pI4, source_swap, destin_swap); 
      break;
    case ABTYPFLOAT : /* ADIS 1001759, T.A. 25.11.1998 */
      curr = p + abapDesc->offset;
      memcpy (&aux_double[0], curr, sizeof(aux_double));
      for (jx = sizeof(aux_double) - 1; jx >= 0; jx--) {
        *curr = aux_double[jx];
        curr++;
      }
      break;
    }
  }
}

/*----------------------------------------------------------------------*/

void OmsStream::omsWrite(unsigned char* obj) {
  
  _TRACE_METHOD_ENTRY(m_handle,"OmsStream::omsWrite")
    if (m_handle->m_pSession->m_stream_io) {
      if (m_row_count * m_row_size == m_buf_size)
        omsFlush();
      memcpy (&m_row_buf[m_row_count * m_row_size], obj, m_row_size);
      if (m_stream_handle->colCount) {
        if (m_stream_handle->rowCount) {
          omsSwapABAPRow (&m_row_buf[m_row_count * m_row_size]);
        }
        else {
          omsSwapStream (&m_row_buf[m_row_count * m_row_size]);
        }
      }
      m_row_count++;
    }
}

/*----------------------------------------------------------------------*/

//
// Implementation of OmsObjByClsIterBase
//
// PTS 1106619
OmsObjByClsIterBase::OmsObjByClsIterBase(const OmsObjByClsIterBase& iter) {
  OmsObjectContainerPtr   *p;
  memcpy (this, &iter, sizeof(iter));
  p = CONST_CAST(OmsObjectContainerPtr*, &iter.m_curr);
  *p = NULL;
   p = CONST_CAST(OmsObjectContainerPtr*, &iter.m_next);
  *p = NULL;
}

/*----------------------------------------------------------------------*/

OmsObjByClsIterBase::OmsObjByClsIterBase( OmsHandle& h ) : 
m_session(h.m_pSession), m_containerInfo(), m_iterNewObjectsInVersion(false), m_headIndex(-1), m_curr(NULL), m_next(NULL) {
};

/*----------------------------------------------------------------------*/

OmsObjByClsIterBase::operator bool () const {
  return ((m_curr) && (m_curr->m_oid)) ? true : false;
}

/*----------------------------------------------------------------------*/

const OmsObjectId& OmsObjByClsIterBase::operator() () const {
  return (m_curr) ? m_curr->m_oid : OmsGlob_co10.m_nilOid;
}

/*----------------------------------------------------------------------*/

OmsAbstractObject* OmsObjByClsIterBase::omsDeRef (bool forUpd, bool doLock) const{
  if ((m_curr) && (m_curr->m_oid)) {
    OmsObjectContainerPtr found = m_session->m_context->FindObjInContext (&m_curr->m_oid);
    if (found) {
      if (found->DeletedFlag()) {
        m_session->ThrowDBError (e_object_not_found, "OmsObjByClsIterBase()", m_curr->m_oid);
        return NULL;
      }
      else {
        if ((doLock) && (!found->LockedFlag()) && (!m_session->m_context->IsVersion())) {
          m_session->m_context->LockObj(found->m_oid);
        }
        if (forUpd) {
          return m_session->ForUpdPtr(found);
        }
        return &found->m_pobj;
      }
    }
    else {
      OmsObjectContainerPtr obj = m_session->GetMemory (*m_containerInfo); 
      memcpy (obj, m_curr,  m_containerInfo->GetObjectSize());
      m_session->m_context->PutObjectIntoContext (obj, m_containerInfo->GetContainerHandle());
      if ((doLock) && (!m_session->m_context->IsVersion())) {
        m_session->m_context->LockObj(obj->m_oid);
      }
      if (forUpd) {
        return m_session->ForUpdPtr(obj);
      }
      return &obj->m_pobj;
    }
  }
  return NULL;
}

/*----------------------------------------------------------------------*/

void OmsObjByClsIterBase::omsNextNewObjectInVersion(OmsObjectContainerPtr obj) {
  do {
    while (NULL != obj) {
      if ((VERSION_OBJ_PAGE_NO_GG92 == obj->m_oid.getPno()) && (obj->GetContainerHandle() == m_containerInfo->GetContainerHandle())) { 
        if ((!(obj->DeletedFlag())) && (!obj->m_objseq.gg91IsDummyRef())) {
          memcpy (m_next, obj, m_containerInfo->GetObjectSize());
          return;
        }
      }
      obj = obj->m_hashnext;
    }
    ++m_headIndex;
    if (m_headIndex < m_session->m_context->m_oidDir.m_headentries) {
      obj = m_session->m_context->m_oidDir.m_head[m_headIndex];
    }
    else {
      break;
    }
  }
  while (true);
  m_next->m_oid.setNil();
}

/*----------------------------------------------------------------------*/

void OmsObjByClsIterBase::operator++ ()
{
  OmsObjectContainerPtr curr = m_curr;
  m_curr = m_next;
  m_next = curr;
  if ((m_curr) && (m_curr->m_oid)) {
    OmsObjectContainerPtr obj = NULL;
    tsp00_Int2            DBError;
    tsp00_Int4            size = m_containerInfo->GetPersistentSize();
    
    if (m_iterNewObjectsInVersion) {
      omsNextNewObjectInVersion(m_curr->m_hashnext);
    }
    else {
      m_next->m_oid = m_curr->m_oid;
      do {
        CHResult_co001 hr = m_session->m_lcSink->NextObj (
          (unsigned char*) &m_session->m_context->m_consistentView,
          (unsigned char*) &m_containerInfo->GetFileId(),
          m_session->m_context->VersionContext(),
          m_next->m_oid.pOid(), 
          (unsigned char*) &m_next->m_objseq, size, 
          ((unsigned char*) &m_next->m_pobj) + sizeof(void*), &DBError);
         if ( DBError != 0 ) { 
          if ( DBError == e_no_next_object) {
            if (m_session->InVersion()) {
              m_iterNewObjectsInVersion = true;
              omsNextNewObjectInVersion(NULL);
            }
            else {
              m_next->m_oid.setNil();
            }
            return;
          }
          else {
            m_session->ThrowDBError (DBError, "Iter op++");
          }
        }
        else {
          obj = m_session->m_context->m_oidDir.HashFind (&m_next->m_oid);
          if (obj) {
            if (!obj->DeletedFlag()) {
              break;
            }
          }
          else {
            break;
          }
        }
      }
      while (true);
    }
  }
}

/*----------------------------------------------------------------------*/

OmsObjByClsIterBase::~OmsObjByClsIterBase() {
  if (m_curr) {
    m_session->deallocate (m_curr);
  }
  if (m_next) {
    m_session->deallocate (m_next);
  }
}

/*----------------------------------------------------------------------*/

//
// Implementation of OmsTimeStamp
//

OmsTimeStamp::OmsTimeStamp(const OmsTypeABAPTimeStamp& timeStamp)
{
  memcpy (&m_TimeStamp, &timeStamp, sizeof(timeStamp));
  ToDateTime();
}

/*----------------------------------------------------------------------*/

const OmsTypeABAPTimeStamp& OmsTimeStamp::omsGetTimeStamp() const
{
  return m_ABAPTimeStamp;
}

/*----------------------------------------------------------------------*/

void OmsTimeStamp::ToDateTime()
{
  int nSec   = m_TimeStamp.nSecond1 * 10 + m_TimeStamp.nSecond2;
  int nMin   = m_TimeStamp.nMinute1 * 10 + m_TimeStamp.nMinute2;
  int nHour  = m_TimeStamp.nHour1   * 10 + m_TimeStamp.nHour2;
  int nDay   = m_TimeStamp.nDay1    * 10 + m_TimeStamp.nDay2;
  int nMonth = m_TimeStamp.nMonth1  * 10 + m_TimeStamp.nMonth2;
  int nYear  = m_TimeStamp.nYear1 * 1000 + m_TimeStamp.nYear2 * 100 +
    m_TimeStamp.nYear3 * 10 + m_TimeStamp.nYear4;
  
  if (nYear < 1970)
  {
    nYear = 1970;
  }
  else if (nYear > 2099)
  {
    nYear = 2099;
  }
  m_DateTime = cco15_DateTime(nYear, nMonth, nDay, nHour, nMin, nSec);
}
/*----------------------------------------------------------------------*/

void OmsTimeStamp::ToTimeStamp()
{
  int n1, n2, n3, n4;

  div10(m_DateTime.seconds(),  n1, n2);
  m_TimeStamp.nSecond1 = n1;
  m_TimeStamp.nSecond2 = n2;
  
  div10(m_DateTime.minutes(),  n1, n2);
  m_TimeStamp.nMinute1 = n1;
  m_TimeStamp.nMinute2 = n2;

  div10(m_DateTime.hours(), n1, n2);
  m_TimeStamp.nHour1 = n1;
  m_TimeStamp.nHour2 = n2;

  div10(m_DateTime.day(), n1, n2);
  m_TimeStamp.nDay1 = n1;
  m_TimeStamp.nDay2 = n2;

  div10(m_DateTime.month(),  n1, n2);
  m_TimeStamp.nMonth1 = n1;
  m_TimeStamp.nMonth2 = n2;

  div1000(m_DateTime.year(), n1, n2, n3, n4);
  m_TimeStamp.nYear1 = n1;
  m_TimeStamp.nYear2 = n2;
  m_TimeStamp.nYear3 = n3;
  m_TimeStamp.nYear4 = n4;
}

/*----------------------------------------------------------------------*/

#if defined(WIN32)
/*int co10_EvalException (OmsHandle* pHandle, LPEXCEPTION_POINTERS lpexceptInfo )  PTS 1106505 FF */
int co10_EvalException (IliveCacheSink* m_sysLcSink, LPEXCEPTION_POINTERS lpexceptInfo )
{

#define STATUS_UNHANDLED_CPP_EXCEPTION (unsigned long)0xE06D7363L   
  int  ret = EXCEPTION_EXECUTE_HANDLER; 
  char cExceptCode[32];
  switch (lpexceptInfo->ExceptionRecord->ExceptionCode)
  {
  case STATUS_WAIT_0:                       // 0x00000000L    
    strcpy(cExceptCode,"WAIT_0"); break;
  case STATUS_ABANDONED_WAIT_0:             // 0x00000080L    
    strcpy(cExceptCode,"ABANDONED_WAIT_0"); break;
  case STATUS_USER_APC:                     // 0x000000C0L    
    strcpy(cExceptCode,"USER_APC"); break;
  case STATUS_TIMEOUT:                      // 0x00000102L    
    strcpy(cExceptCode,"TIMEOUT"); break;
  case STATUS_PENDING:                      // 0x00000103L    
    strcpy(cExceptCode,"PENDING"); break;
  case STATUS_SEGMENT_NOTIFICATION:         // 0x40000005L    
    strcpy(cExceptCode,"SEGMENT_NOTIFICATION"); break;
  case STATUS_GUARD_PAGE_VIOLATION:         // 0x80000001L    
    strcpy(cExceptCode,"GUARD_PAGE_VIOLATION"); break;
  case STATUS_NONCONTINUABLE_EXCEPTION:     // 0xC0000025L    
    strcpy(cExceptCode,"GUARD_PAGE_VIOLATION"); break;
  case STATUS_BREAKPOINT:                   // 0x80000003L
    return EXCEPTION_CONTINUE_SEARCH;
  case STATUS_SINGLE_STEP:                  // 0x80000004L    
    strcpy(cExceptCode,"SINGLE_STEP"); break;
  case STATUS_IN_PAGE_ERROR:                // 0xC0000006L    
    strcpy(cExceptCode,"IN_PAGE_ERROR"); break;
  case STATUS_ILLEGAL_INSTRUCTION:          // 0xC000001DL    
    strcpy(cExceptCode,"ILLEGAL_INSTRUCTION"); break;
  case STATUS_PRIVILEGED_INSTRUCTION:       // 0xC0000096L    
    strcpy(cExceptCode,"PRIVILEGED_INSTRUCTION"); break;
  case STATUS_STACK_OVERFLOW:               // 0xC00000FDL    
    strcpy(cExceptCode,"STACK_OVERFLOW"); break;
  case STATUS_CONTROL_C_EXIT:               // 0xC000013AL    
    strcpy(cExceptCode,"CONTROL_C_EXIT"); break;
  case STATUS_DATATYPE_MISALIGNMENT:        // 0x80000002L   
    strcpy(cExceptCode,"DATATYPE_MISALIGNMENT"); break;
  case STATUS_ACCESS_VIOLATION:             // 0xC0000005L    
    strcpy(cExceptCode,"ACCESS_VIOLATION"); break;
  case STATUS_INVALID_HANDLE:               // 0xC0000008L    
    strcpy(cExceptCode,"INVALID_HANDLE"); break;
  case STATUS_NO_MEMORY:                    // 0xC0000017L    
    strcpy(cExceptCode,"NO_MEMORY"); break;
  case STATUS_INVALID_DISPOSITION:          // 0xC0000026L    
    strcpy(cExceptCode,"INVALID_DISPOSITION"); break;
  case STATUS_ARRAY_BOUNDS_EXCEEDED:        // 0xC000008CL    
    strcpy(cExceptCode,"ARRAY_BOUNDS_EXCEEDED"); break;
  case STATUS_FLOAT_DENORMAL_OPERAND:       // 0xC000008DL    
    strcpy(cExceptCode,"FLOAT_DENORMAL_OPERAND"); break;
  case STATUS_FLOAT_DIVIDE_BY_ZERO:         // 0xC000008EL    
    strcpy(cExceptCode,"FLOAT_DIVIDE_BY_ZERO"); break;
  case STATUS_FLOAT_INEXACT_RESULT:         // 0xC000008FL    
    strcpy(cExceptCode,"FLOAT_INEXACT_RESULT"); break;
  case STATUS_FLOAT_INVALID_OPERATION:      // 0xC0000090L    
    strcpy(cExceptCode,"FLOAT_INVALID_OPERATION"); break;
  case STATUS_FLOAT_OVERFLOW:               // 0xC0000091L    
    strcpy(cExceptCode,"FLOAT_OVERFLOW"); break;
  case STATUS_FLOAT_STACK_CHECK:            // 0xC0000092L    
    strcpy(cExceptCode,"FLOAT_STACK_CHECK"); break;
  case STATUS_FLOAT_UNDERFLOW:              // 0xC0000093L    
    strcpy(cExceptCode,"FLOAT_UNDERFLOW"); break;
  case STATUS_INTEGER_DIVIDE_BY_ZERO:       // 0xC0000094L    
    strcpy(cExceptCode,"INTEGER_DIVIDE_BY_ZERO"); break;
  case STATUS_INTEGER_OVERFLOW:             // 0xC0000095L    
    strcpy(cExceptCode,"INTEGER_OVERFLOW"); break;
  case STATUS_UNHANDLED_CPP_EXCEPTION:      // 0xE06D7363L   
    strcpy(cExceptCode,"UNHANDLED_C++_EXCEPTION"); break;  
  default:
    sprintf(cExceptCode,"ExceptionCode: 0x%8.8X ", 
      lpexceptInfo->ExceptionRecord->ExceptionCode);
    ret = EXCEPTION_CONTINUE_SEARCH;
  }
//  pHandle->dbpOpError("%s at address 0x%8.8X", cExceptCode,   PTS 1106505 FF
//    lpexceptInfo->ExceptionRecord-> ExceptionAddress);        PTS 1106505 FF
/* PTS 1106505 FF +++++++++++++++++++++  */
  tsp00_Int2  DummyError;
  tsp00_Int4  BufLen = sizeof(tsp00_Addr);
  tsp00_Addr  Buf = REINTERPRET_CAST(tsp00_Addr,lpexceptInfo);
  CHResult_co001 hr = m_sysLcSink->MultiPurpose (m_diagnose, mm_dump, &BufLen, 
    (unsigned char*) &Buf, &DummyError);
/* PTS 1106505 FF ---------------------  */
  return ret;
}

#endif


