
/* @lastChanged: "1998-09-22  8:46"

  * @filename:   vco13.cpp
  * @purpose:    "Implementation Runtime Functions of Persistent Object Interface"
  * @release:    7.1.0.0
  * @see:        "http://www.bea.sap-ag.de/doku/rel-7.1/vco13.html"
  *
  * @copyright:  "Copyright by   SAP AG, 1998"


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/


/* ================================================================================ */
/* ==========              REPLACEMENTS BY THE PREPROCESSOR               ========= */
/* ================================================================================ */



/* ================================================================================ */
/* ==========                 DECLARATION OF TYPES                        ========= */
/* ================================================================================ */




/* ================================================================================ */
/* ==========                 DECLARATION OF FUNCTIONS                    ========= */
/* ================================================================================ */

#include "geo00.h"
#include "gco00.h"
#include "gco10.h"
#define __HCO13_IMPL
#include "hco13.h"

/* ================================================================================ */
/* ==========                 DEFINITION OF GLOBAL FUNCTIONS              ========= */
/* ================================================================================ */

void tco13_GuidEntry::ChainFree (OmsContext_co12&c, OmsAbstractObject* po) {
  if (NULL != po) {
    OmsObjectContainerPtr p = OmsObjectContainer::containerPtr(po);
#ifdef _ASSERT_OMS
    OmsObjectContainerPtr curr = m_freeHead->free;
    while (curr) {
      if (curr == p) {
        throw DbpError (-10, "Duplicate in freelist"); 
      }
      curr = curr->m_hashnext;
    }
#endif
    if (p->VarObjFlag()) {
      REINTERPRET_CAST(OmsVarObjInfo_co12*, po)->freeVarObj(&c);
    }
    p->m_hashnext    = m_freeHead->free;
    m_freeHead->free = p;
  }
}

/*----------------------------------------------------------------------*/

void tco13_GuidEntry::ChainFree (OmsContext_co12&c, OmsObjectContainerPtr& p) {
  if (NULL != p) {
#ifdef _ASSERT_OMS
    OmsObjectContainerPtr curr = m_freeHead->free;
    while (curr) {
      if (curr == p) {
        throw DbpError (-10, "Duplicate in freelist"); 
      }
      curr = curr->m_hashnext;
    }
#endif
    if (p->VarObjFlag()) {
      REINTERPRET_CAST(OmsVarObjInfo_co12*, &p->m_pobj)->freeVarObj(&c);
    }
    p->m_hashnext    = m_freeHead->free;
    m_freeHead->free = p;
    p = NULL;
  }
}

/*----------------------------------------------------------------------*/

tco13_ClassHash::tco13_ClassHash() : IGuidHash_co001(), m_context(NULL) {
  for (int ix = 0; ix < GUID_HASH_ENTRIES_CO13; ++ix) {
    m_head[ix] = NULL;
  }
}

/*----------------------------------------------------------------------*/

tco13_ClassHash::~tco13_ClassHash() {
  for (int ix = 0; ix < GUID_HASH_ENTRIES_CO13; ++ix) {
    tco13_GuidEntryPtr pDel;
    tco13_GuidEntryPtr curr = m_head[ix];
    while (NULL != curr) {
      pDel = curr;
      curr = curr->m_hashNext;
      pDel->DeleteSelf(m_context);
    }
  }
}

/*----------------------------------------------------------------------*/

void tco13_ClassHash::Create(OmsContext_co12* context) {
  m_context = context;
}

/*----------------------------------------------------------------------*/

void tco13_ClassHash::HashInsert(ClassInfo_co18* pClassInfo, tco13_FreeListHeader* freeHead) {
  if (!HashFind(pClassInfo->GetGuid())) {
    int slot = HashValue(pClassInfo->GetGuid(), GUID_HASH_ENTRIES_CO13);
    m_head[slot] = new(m_context) tco13_GuidEntry(m_head[slot], pClassInfo, freeHead); 
  }
}

/*----------------------------------------------------------------------*/
 
tco13_GuidEntryPtr tco13_ClassHash::HashFind(const GUID& guid) {
  int slot = HashValue(guid, GUID_HASH_ENTRIES_CO13);
  tco13_GuidEntryPtr p = m_head[slot];
  while (NULL != p) {
    if (p->m_classInfo->GetGuid() == guid) {
      return p;
    }
    p = p->m_hashNext;
  }
  return NULL;
}

/*----------------------------------------------------------------------*/

tco13_ClsIdEntry::tco13_ClsIdEntry (ContainerInfo_co18* pContainerInfo,OmsContext_co12* c) : m_clsidHashNext(NULL), 
    m_guidHashNext(NULL), m_containerInfo(pContainerInfo), m_index(), m_freeHead(NULL) {
  m_index.AdviseCompare(this);
  m_index.AdviseAllocator(c);
  pContainerInfo->IncRefCnt(c->GetCurrentSession()->m_lcSink);
}

/*----------------------------------------------------------------------*/

void tco13_ClsIdEntry::chainFree (OmsContext_co12&c, OmsAbstractObject* po) {
  if (po) {
    OmsObjectContainerPtr p = OmsObjectContainer::containerPtr(po);
#ifdef _ASSERT_OMS
    OmsObjectContainerPtr curr = m_freeHead->free;
    while (curr) {
      if (curr == p) {
        throw DbpError (-10, "Duplicate in freelist"); 
      }
      curr = curr->m_hashnext;
    }
#endif
    if (p->VarObjFlag()) {
      REINTERPRET_CAST(OmsVarObjInfo_co12*, po)->freeVarObj(&c);
    }
    p->m_hashnext    = m_freeHead->free;
    m_freeHead->free = p;
  }
}

/*----------------------------------------------------------------------*/

void tco13_ClsIdEntry::chainFree (OmsContext_co12&c, OmsObjectContainerPtr& p) {
  if (NULL != p) {
#ifdef _ASSERT_OMS
    OmsObjectContainerPtr curr = m_freeHead->free;
    while (curr) {
      if (curr == p) {
        throw DbpError (-10, "Duplicate in freelist"); 
      }
      curr = curr->m_hashnext;
    }
#endif
    if (p->VarObjFlag()) {
      REINTERPRET_CAST(OmsVarObjInfo_co12*, &p->m_pobj)->freeVarObj(&c);
    }
    p->m_hashnext    = m_freeHead->free;
    m_freeHead->free = p;
    p = NULL;
  }
}

/*----------------------------------------------------------------------*/

void tco13_ClsIdEntry::VersionAddKey(OmsObjectContainerPtr p) {
  tgg00_BasisError rc;
  rc = m_index.Insert(GetKeyPtr(p));
  if (e_ok != rc) {
    throw DbpError (DbpError::DB_ERROR, rc); 
  }
}

/*----------------------------------------------------------------------*/

void tco13_ClsIdEntry::VersionDelIndex() {
  m_index.DeleteAll();
}

/*----------------------------------------------------------------------*/

void tco13_ClsIdEntry::VersionDelKey(OmsObjectContainerPtr p) {
   m_index.Delete(GetKeyPtr(p));
}

/*----------------------------------------------------------------------*/

OmsObjectContainer* tco13_ClsIdEntry::VersionFindKey(const unsigned char* k) {
  unsigned char* const * p = m_index.Find(CONST_CAST(unsigned char*, k));
  if (NULL != p) {
    return REINTERPRET_CAST(OmsObjectContainer*, CONST_CAST(unsigned char*, (*p) + GetKeyLen() - GetObjectSize()));
  }
  else {
    return NULL;
  }
}

/*----------------------------------------------------------------------*/

cco13_OidHash::cco13_OidHash () : m_headentries(0), m_head(NULL),	m_headcurr(NULL), m_context(NULL) {
};

/*----------------------------------------------------------------------*/

// end constructor

cco13_OidHash::~cco13_OidHash () {
  tsp00_Int4 ix;
  if (m_context->IsVersion()) {
    for (ix = 0; ix < m_headentries; ++ix) {
      OmsObjectContainerPtr toDel;
      OmsObjectContainerPtr curr = m_head[ix];
      while (NULL != curr) {
        toDel = curr;
        curr  = curr->m_hashnext;
        m_context->deallocate(toDel);
      }
    }
  }
  HashFree();	
};

// end destructor

/*----------------------------------------------------------------------*/

void cco13_OidHash::Clean() {
  // removes all objects of dropped containers from OID Hash
  OmsObjectContainerPtr* prev;
  OmsObjectContainerPtr  curr;
  OmsObjectContainerPtr  toFree;
  tsp00_Int4             ix;
  for (ix = 0; ix < m_headentries; ix++) {
    prev = &m_head[ix];
    curr = *prev;
    while (NULL != curr) {
      tco13_ClsIdEntry* containerInfo = curr->GetContainerInfoNoCheck(m_context);
      if (containerInfo->GetContainerInfoPtr()->IsDropped()) {
        toFree = curr;
        *prev = curr->m_hashnext;
        curr  = curr->m_hashnext;
        containerInfo->chainFree(*m_context, toFree);
      }
      else {
        prev = &curr->m_hashnext;
        curr  = curr->m_hashnext;
      }
    }
  }
}

/*----------------------------------------------------------------------*/

void cco13_OidHash::Clear() {
  tsp00_Int4  ix;
  tsp00_Int4 listCount  = 0;
  tsp00_Int4 entryCount = 0;
  OmsObjectContainerPtr curr;
  for (ix = 0; ix < m_headentries; ++ix) {
    curr = m_head[ix];
    if (NULL != curr) {
      ++listCount;
      do {
        ++entryCount;
        if (curr->VarObjFlag()) {
          REINTERPRET_CAST(OmsVarObjInfo_co12*, &curr->m_pobj)->freeVarObj(m_context);
        }
        curr  = curr->m_hashnext;
      }
      while (NULL != curr);
    }
    m_head[ix] = NULL;
  }
  if (entryCount > 2 * listCount) {
    HashResize (entryCount);
  }
}

/*----------------------------------------------------------------------*/

void cco13_OidHash::Create (OmsContext_co12* c, const tsp00_Int4 sz) {
  m_context = c;
  HashInit(sz, true);
}

/*----------------------------------------------------------------------*/

void cco13_OidHash::CheckChain(OmsObjectContainerPtr curr) {
  OmsObjectContainerPtr p = curr->m_hashnext;
  while (p != NULL) {
    if (p == curr) {
      throw DbpError (-9, "Loop in Oms Cache"); 
    }
    p = p->m_hashnext;
  }
}

/*----------------------------------------------------------------------*/

void cco13_OidHash::HashCheck() {
  
  for (tsp00_Int4 ix = 0; ix < m_headentries; ix++) {
    if (m_head[ix] != NULL) {
      CheckChain (m_head[ix]);
    }
  }
}

/*----------------------------------------------------------------------*/

inline void cco13_OidHash::HashFree () 
{
  m_context->deallocate(m_head);
}

/*----------------------------------------------------------------------*/

inline void cco13_OidHash::HashInit (tsp00_Int4 sz, bool throwError)
{
  m_headentries= sz;
	m_head = REINTERPRET_CAST(OmsObjectContainerPtr*, m_context->allocate(sizeof(OmsObjectContainerPtr) * sz));
  for (tsp00_Int4 ix = 0; ix < sz; ix++) {
		m_head[ix] = NULL;
	};
	m_headcurr = NULL;
}

/*----------------------------------------------------------------------*/

void cco13_OidHash::HashDelete (const OmsObjectId& key) {

	OmsObjectContainerPtr* prev = HeadPtr(key);
	OmsObjectContainerPtr  curr = *prev;
  while (curr) {
		if (curr->m_oid == key) {
      if (curr->VarObjFlag()) {
        OmsVarObjInfo_co12* objInfo;
        objInfo = REINTERPRET_CAST(OmsVarObjInfo_co12*, (&curr->m_pobj));
        objInfo->freeVarObj(m_context);
      }
			*prev = curr->m_hashnext;
			return;
		}
		else {
			prev = &curr->m_hashnext;
			curr = curr->m_hashnext;
		}
  }
#ifdef _ASSERT_OMS
  HashCheck();
#endif
	return;
};

/*----------------------------------------------------------------------*/

void cco13_OidHash::HashResize(tsp00_Int4 entryCount)
{
  if (entryCount > m_headentries) {
    tsp00_Int4 currHeadEntries   = m_headentries;
    bool isPrime                 = false;
    OmsObjectContainerPtr* oldHead;
    oldHead = m_head;
    int  ix;
    currHeadEntries = entryCount;
    while (!isPrime) {
      currHeadEntries++;
      isPrime = (1 == (currHeadEntries % 2)) ? true : false; 
      ix = 3;
      while (((ix * ix) <= currHeadEntries) && (isPrime)) {
        isPrime = (0 == (currHeadEntries % ix)) ? false : true;
        ix += 2;
      }
    }
    HashInit(currHeadEntries, false);
    if (NULL != m_head) {
      m_context->deallocate (oldHead);
    }
    else {
      m_head = oldHead;
    }
  }
}
// end HashDelete

/*----------------------------------------------------------------------*/

void cco13_OidHash::RemoveUnlockedObjFromCache() {
  tsp00_Int4             ix;
  OmsObjectContainerPtr* prev;
  OmsObjectContainerPtr  curr;
  OmsObjectContainerPtr  toFree;
  tco13_ClsIdEntry*      pContainerInfo;
 
  for (ix = 0; ix < m_headentries; ix++) {
    prev       = &m_head[ix];
    curr       = m_head[ix];
    m_head[ix] = NULL;
    while (NULL != curr) {
      toFree = curr;
      curr   = curr->m_hashnext;
      if ((toFree->LockedFlag()) || (toFree->m_beforeImages != 0)) 
      {
        *prev              = toFree;
        toFree->m_hashnext = NULL;
        prev               = &toFree->m_hashnext;
      }
      else {
        pContainerInfo = toFree->GetContainerInfo(m_context);
        pContainerInfo->chainFree(*m_context, toFree);
      }
    }
  }
}

/*----------------------------------------------------------------------*/

void cco13_OidHash::SetEmpty() {
  tsp00_Int4 ix;
  for (ix = 0; ix < m_headentries; ++ix) {
    m_head[ix] = NULL;
  }
}

/*----------------------------------------------------------------------*/
/* Implementation of ClsId HashList                                     */
/*----------------------------------------------------------------------*/

cco13_ClsIdHash::cco13_ClsIdHash () : headentries(0), m_clsidHead(NULL), m_guidHead(NULL), freeListHeader(NULL), m_context(NULL) 
{
};

/*----------------------------------------------------------------------*/

tco13_ClsIdEntry* cco13_ClsIdHash::AutoRegister (const GUID& guid, OmsSchemaHandle Schema, OmsContainerNo ContainerNo) {
  tco13_ClsIdEntry* curr = NULL;
  ContainerInfo_co18* pContainerInfo = co10_FindContainerInfo(m_context->GetCurrentSession()->m_lcSink, guid, Schema, ContainerNo);
  if (NULL != pContainerInfo) {
    curr = new(m_context) tco13_ClsIdEntry (pContainerInfo, m_context);
    if (NULL != curr) {
      HashInsert (curr);
    }
  }
  else {
    this->ThrowUnknownContainer(guid, Schema, ContainerNo);
  }
  return curr;
}

/*----------------------------------------------------------------------*/

void cco13_ClsIdHash::Clean() {
  // remove all infos of dropped containers
  cco13_ClsIdHash::Iter iter = this->First();
  while (iter) {
    tco13_ClsIdEntry* curr = iter();
    ++iter;
    if (curr->m_containerInfo->IsDropped()) {
      this->HashDelete(curr);
    }
  }
}

/*----------------------------------------------------------------------*/

void cco13_ClsIdHash::Create (OmsContext_co12* context) {
  m_context   = context;
  headentries = 257;
  m_clsidHead = REINTERPRET_CAST(tco13_ClsIdEntryPtr*, 
    m_context->allocate(sizeof(tco13_ClsIdEntryPtr) * headentries));
  m_guidHead = REINTERPRET_CAST(tco13_ClsIdEntryPtr*, 
    m_context->allocate(sizeof(tco13_ClsIdEntryPtr) * headentries));
  for (tsp00_Int4 ix = 0; ix < headentries; ++ix) {
    m_clsidHead[ix]= NULL;
    m_guidHead[ix] = NULL;
  };
  m_classDir.Create(context);
  freeListHeader = new(m_context) tco13_FreeListHeader(-1);
};

// end constructor

/*----------------------------------------------------------------------*/

tco13_FreeListHeader* cco13_ClsIdHash::GetFreeListHeader(size_t size) 
{
  tco13_FreeListHeader* p = freeListHeader;
  while (p != NULL) {
    if (p->size == size) {
      return p;
    }
    else {
      p = p->next;
    }
  }
  p       = new(m_context) tco13_FreeListHeader(size);
  p->next = freeListHeader->next;
  freeListHeader->next = p;
  return p;
}

/*----------------------------------------------------------------------*/

void cco13_ClsIdHash::HashDelete (tco13_ClsIdEntry* containerInfo) {
  // remove from ContainerHandle Hash 
  int hashSlot              = containerInfo->GetContainerHandle() % headentries;
  tco13_ClsIdEntryPtr* prev = &m_clsidHead[hashSlot];
  tco13_ClsIdEntryPtr  curr = *prev;
  while (curr != NULL) {
    if (curr == containerInfo) {
      *prev = curr->m_clsidHashNext;
      break;
    }
    else {
      prev = &curr->m_clsidHashNext;
      curr = curr->m_clsidHashNext;
    }
  }
  // remove from Guid,Schema,ContainerNo Hash
  hashSlot = HashValue(containerInfo->GetGuid(), containerInfo->GetSchema(), 
    containerInfo->GetContainerNo(), headentries);
  prev     = &m_guidHead[hashSlot];
  curr     = *prev;
  while (curr != NULL) {
    if (curr == containerInfo) {
      *prev = curr->m_guidHashNext;
      break;
    }
    else {
      prev = &curr->m_guidHashNext;
      curr = curr->m_guidHashNext;
    }
  }
  // delete local ContainerInfo
  containerInfo->DeleteSelf(m_context);
}

/*----------------------------------------------------------------------*/

bool cco13_ClsIdHash::IsBaseClassOf(const GUID& baseGuid, const GUID& derivedGuid) {
  return  IsDerivedClassOf(baseGuid, derivedGuid);
}

/*----------------------------------------------------------------------*/

bool cco13_ClsIdHash::IsDerivedClassOf(const GUID& baseGuid, const GUID& derivedGuid) {
   tco13_GuidEntry* pDerivedClassInfo = GetClassInfo(derivedGuid);
   do {
     if (pDerivedClassInfo->IsDerivedClassOf(baseGuid)) {
       return true;
     }
     if (pDerivedClassInfo->IsDerivedClass()) {
       pDerivedClassInfo = GetClassInfo(pDerivedClassInfo->GetBaseGuid());
     }
     else {
       return false;
     }
   }
   while(1);
}

/*----------------------------------------------------------------------*/

void cco13_ClsIdHash::ThrowUnknownContainer(tco13_ClsIdEntryPtr containerInfo) {
  this->ThrowUnknownContainer(containerInfo->GetGuid(), containerInfo->GetSchema(), containerInfo->GetContainerNo());
}

/*----------------------------------------------------------------------*/


void cco13_ClsIdHash::ThrowUnknownContainer(const GUID& guid, OmsSchemaHandle Schema, OmsContainerNo ContainerNo) {
  char buf[256];
  sprintf (buf, "guid : %8X-%4X-%4X-%1X%1X%1X%1X%1X%1X%1X%1X, Schema(%8X) CNo(%8X)", guid.Data1, guid.Data2, guid.Data3,
    guid.Data4[0], guid.Data4[1], guid.Data4[2], guid.Data4[3], guid.Data4[4], guid.Data4[5],
    guid.Data4[6], guid.Data4[7], Schema, ContainerNo);
  throw DbpError (DbpError::DB_ERROR, e_unknown_guid, buf); 
}

/*----------------------------------------------------------------------*/

cco13_ClsIdHash::~cco13_ClsIdHash () {
  
  /* delete hash */
  tco13_ClsIdEntry* curr;
  tco13_ClsIdEntry* delClsInfo;
  for (tsp00_Int4 ix = 0; ix < headentries; ix++) {
    curr = m_clsidHead[ix];
    while (curr != NULL) {
      delClsInfo = curr;
      curr       = curr->m_clsidHashNext;
      delClsInfo->DeleteSelf(m_context);
    }
  }
  m_context->deallocate(m_clsidHead);
  m_context->deallocate(m_guidHead);
  /* delete free list headers */
  tco13_FreeListHeader* toDel;
  tco13_FreeListHeader* p = freeListHeader;
  while (p != NULL) {
    toDel = p;
    p     = p->next;
    toDel->DeleteSelf(m_context);
  }
};

// end destructor

// end Guid HashFind

/*----------------------------------------------------------------------*/

cco13_BeforeImages::cco13_BeforeImages () : m_session(NULL) {
  int ix;
  for (ix = 0; ix < MAX_SUBTRANS_LEVEL; ix++) {
    m_beforeImages[ix] = NULL;
  }
}

/*----------------------------------------------------------------------*/

cco13_BeforeImages::~cco13_BeforeImages () {
  
}

/*----------------------------------------------------------------------*/

void cco13_BeforeImages::init (OmsSession_co12* s) {
   m_session = s;
}

/*----------------------------------------------------------------------*/

void cco13_BeforeImages::freeBeforeImages () {

  OmsObjectContainerPtr pBeforeImage;
  OmsObjectContainerPtr pToFree;
  tco13_ClsIdEntry*     pContainerInfo;
  OmsContext_co12*      context;
  int                   subtransLvl = m_session->CurrentSubtransLevel();
  while (subtransLvl > 0) {
    pBeforeImage                  = m_beforeImages[subtransLvl-1];
    m_beforeImages[subtransLvl-1] = NULL;
    while (pBeforeImage) {
      context        = *REINTERPRET_CAST(OmsContext_co12**, &pBeforeImage->m_pobj);
      pContainerInfo = pBeforeImage->GetContainerInfo(context);
      pToFree        = pBeforeImage;
      pBeforeImage   = pBeforeImage->m_hashnext;
      pContainerInfo->chainFree(*context, pToFree);
    }
    subtransLvl--; 
  }
}

/*----------------------------------------------------------------------*/

void cco13_BeforeImages::removeContext (OmsContext_co12* pContext) {
  OmsObjectContainerPtr  pBeforeImage;
  OmsObjectContainerPtr* pPrev;
  OmsObjectContainerPtr  pToFree;
  tco13_ClsIdEntry*      pContainerInfo;
  OmsContext_co12*       currContext;
  int                    subtransLvl = m_session->CurrentSubtransLevel();
  while (subtransLvl > 0) {
    pBeforeImage = m_beforeImages[subtransLvl-1];
    pPrev        = &m_beforeImages[subtransLvl-1];
    while (NULL != pBeforeImage) {
      currContext = *REINTERPRET_CAST(OmsContext_co12**, &pBeforeImage->m_pobj);
      if (currContext == pContext) {
        pContainerInfo = pBeforeImage->GetContainerInfo(currContext);
        pToFree        = pBeforeImage;
        pBeforeImage   = pBeforeImage->m_hashnext;
        *pPrev         = pBeforeImage;
        pContainerInfo->chainFree(*currContext, pToFree);
      }
      else {
        pPrev        = &pBeforeImage->m_hashnext;
        pBeforeImage = pBeforeImage->m_hashnext;
      }
    }
    subtransLvl--; 
  } 
}

/*----------------------------------------------------------------------*/

void cco13_BeforeImages::subtransCommit (int wantedSubtransLvl) {
  OmsObjectContainerPtr pBeforeImage, p;
  OmsObjectContainerPtr found;
  OmsContext_co12*      context;
  int         subtransLvl = m_session->CurrentSubtransLevel();
  tsp00_Uint4 mask = (1 == wantedSubtransLvl) ? 0 : (0xFFFFFFFF >> (MAX_SUBTRANS_LEVEL - wantedSubtransLvl + 1));
  while (subtransLvl >= wantedSubtransLvl) {
    subtransLvl--;
    pBeforeImage                = m_beforeImages[subtransLvl];
    m_beforeImages[subtransLvl] = NULL;
    while (pBeforeImage) {
      p                  = pBeforeImage;
      pBeforeImage       = pBeforeImage->m_hashnext;
      context            = *REINTERPRET_CAST(OmsContext_co12**, &p->m_pobj);
      found = context->FindObjInContext(&p->m_oid);
      if (found) {
        found->m_beforeImages &= mask;
        if (!found->existBeforeImage(wantedSubtransLvl-1, m_session->InVersion())) {
          p->m_hashnext = m_beforeImages[wantedSubtransLvl-2];
          m_beforeImages[wantedSubtransLvl-2] = p;
          found->setBeforeImage (wantedSubtransLvl-1); 
        }
        else {
          p->ChainFree(context);
        }
      }
      else {
        // object must have been release by application, i.e. omsRelease of an
        // object, that has been read for update without omsStore, omsDelete
      }
    }
  }
}

/*----------------------------------------------------------------------*/

void cco13_BeforeImages::subtransRollback (int wantedSubtransLvl) {
  OmsObjectContainerPtr pBeforeImage, p;
  OmsObjectContainerPtr found;
  tco13_ClsIdEntry*     pContainerInfo;
  OmsContext_co12*      context;
  int         subtransLvl = m_session->CurrentSubtransLevel();
  tsp00_Uint4 mask = (1 == wantedSubtransLvl) ? 0 : (0xFFFFFFFF >> (MAX_SUBTRANS_LEVEL - wantedSubtransLvl + 1));
  while (subtransLvl >= wantedSubtransLvl) {
    subtransLvl--;
    pBeforeImage                = m_beforeImages[subtransLvl];
    m_beforeImages[subtransLvl] = NULL;
    while (pBeforeImage) {
      p                  = pBeforeImage;
      p->m_beforeImages &= mask; 
      pBeforeImage       = pBeforeImage->m_hashnext;
      context            = *REINTERPRET_CAST(OmsContext_co12**, &p->m_pobj);
      if (p->NewFlag()) {
        context->DeleteObjInContext(p, NULL);
        pContainerInfo = p->GetContainerInfo(context);
        if (context->IsVersion()) {
          pContainerInfo->VersionDelKey(p);
        }
        pContainerInfo->chainFree(*context, p);
      }
      else {
        pContainerInfo = p->GetContainerInfo(context);
        found = context->FindObjInContext(&p->m_oid);
        if (NULL != found) {
          found->Copy (context, p, pContainerInfo->GetObjectSize());
          /* restore vtable ptr */
          *REINTERPRET_CAST(void**, &found->m_pobj) = pContainerInfo->GetVirtualTablePtr();
          pContainerInfo->chainFree(*context, p);
        }
        else {
          // object must have been release by application, i.e. omsRelease of an
          // object, that has been read for update without omsStore, omsDelete
        }
      }
    }
  }
}

/*----------------------------------------------------------------------*/

void cco13_BeforeImages::insertBeforeImage (OmsObjectContainerPtr pObj, int subtransLvl) {
  if (subtransLvl > 0) {
#ifdef _ASSERT_OMS
    OmsObjectContainerPtr p = m_beforeImages[subtransLvl-1];
    while (p) {
      _OMSASSERT (m_session, pObj->m_oid != p->m_oid);
      p = p->m_hashnext;
    }
#endif
    tco13_ClsIdEntry* pContainerInfo = pObj->GetContainerInfo(m_session->m_context);
    OmsObjectContainerPtr pBeforeImage = m_session->GetMemory (*pContainerInfo);
    memcpy (pBeforeImage, pObj, pContainerInfo->GetObjectSize());
    pBeforeImage->m_hashnext      = m_beforeImages[subtransLvl-1];
    m_beforeImages[subtransLvl-1] = pBeforeImage;
    pObj->setBeforeImage (subtransLvl);
    /* use the vtable ptr of the before image to store the context the object is from */
    *REINTERPRET_CAST(OmsContext_co12**, (&pBeforeImage->m_pobj)) = m_session->m_context;
  }
}

/*----------------------------------------------------------------------*/

void cco13_BeforeImages::insertNewBeforeImage (OmsObjectContainerPtr pObj, tco13_ClsIdEntry* pContainerInfo, int subtransLvl) {

  if (subtransLvl > 0) {
#ifdef _ASSERT_OMS
    OmsObjectContainerPtr p = m_beforeImages[subtransLvl-1];
    while (p) {
      _OMSASSERT (m_session, pObj->m_oid != p->m_oid);
      p = p->m_hashnext;
    }
#endif
    OmsObjectContainerPtr pBeforeImage = m_session->GetMemory (*pContainerInfo);
    memcpy (pBeforeImage, pObj, pContainerInfo->GetObjectSize());
    pBeforeImage->m_hashnext      = m_beforeImages[subtransLvl-1];
    pBeforeImage->MarkNew();
    m_beforeImages[subtransLvl-1] = pBeforeImage;
    pObj->setBeforeImage (subtransLvl);
    /* use the vtable ptr of the before image to store the context the object is from */
    *REINTERPRET_CAST(OmsContext_co12**, (&pBeforeImage->m_pobj)) = m_session->m_context;
  }
}

