/* @lastChanged: "1999-04-21  16:14"

  * @filename:   vco18.cpp
  * @purpose:    "Implementation of OMS Class Dictionary"
  * @release:    7.2.0.0
  * @see:        "http://www.bea.sap-ag.de/doku/rel-7.1/vco10.html"
  *
  * @copyright:  "Copyright by   SAP AG, 1998"


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/


/* ================================================================================ */
/* ==========              REPLACEMENTS BY THE PREPROCESSOR               ========= */
/* ================================================================================ */



/* ================================================================================ */
/* ==========                 DECLARATION OF TYPES                        ========= */
/* ================================================================================ */




/* ================================================================================ */
/* ==========                 DECLARATION OF FUNCTIONS                    ========= */
/* ================================================================================ */

#include "geo00.h"
#include "gco00.h"
#include "gco10.h"
#include "hco16.h"
#include "hco18.h"

extern co16_SharedMemAllocator SharedMemAllocator_co16;

tsp00_Uint4 ContainerHandle_co18 = 0;

static void co18_ToKnlIdentifier (const char* Source, tsp00_KnlIdentifier& Dest) {
  int copyLen;
  copyLen = strlen(Source);
  if (copyLen > sizeof(Dest)) {
    copyLen = sizeof(Dest);
  }
  memcpy(&Dest[0], Source, copyLen);
  memset(&Dest[copyLen], ' ', sizeof(Dest) - copyLen);
}

/*----------------------------------------------------------------------------------*/

ClassInfo_co18::ClassInfo_co18(const char* ClassName, 
                               const GUID& Guid,
                               const GUID* pBaseClass,
                               const tgg00_ObjColDesc& ColDesc,
                               size_t ObjSize, 
                               void* vtptr, 
                               ClassInfo_co18* HashNext) :  
                               m_guid(Guid), m_ObjectSize(ObjSize), m_VTPtr(vtptr),
                               m_HashNext(HashNext), m_isBaseClass(false), m_isDerivedClass(pBaseClass!=NULL),
                               m_isVarObject(false)
{
  if (pBaseClass != NULL) {
    memcpy (&m_baseGuid, pBaseClass, sizeof(m_baseGuid));
  }
  m_InvColDesc = ColDesc;
  int ClassNameLength = strlen(ClassName);
  if (ClassNameLength >= sizeof(m_ClassName)) {
    ClassNameLength = sizeof(m_ClassName) - 1;
  }
  memcpy (&m_ClassName[0], ClassName, ClassNameLength);
  m_ClassName[ClassNameLength] = 0;
  if (0 < m_InvColDesc.colHashCount_gg00) {
    m_KeyDesc    = (*ColDesc.colHashList_gg00)[0];
    m_InvColDesc.colHashList_gg00  = REINTERPRET_CAST(tgg00_StackListPtr, (&m_KeyDesc));
  }
  m_isVarObject = (0 == memcmp(&Guid, &VAR_OBJ_GUID_GCO10, sizeof(Guid)));
}

/*----------------------------------------------------------------------------------*/

void* ClassInfo_co18::operator new(size_t sz, IliveCacheSink* lcSink) {
  return SharedMemAllocator_co16.allocate(lcSink, sz);
}

/*----------------------------------------------------------------------------------*/

void ClassInfo_co18::DeleteSelf(IliveCacheSink* lcSink) {
#if defined(_ASSERT_OMS)
  memset (this, 0, sizeof(*this));
#endif
  SharedMemAllocator_co16.deallocate(lcSink, this);
}

/*----------------------------------------------------------------------------------*/

#if defined(PLACEMENT_DELETE_MEO00)

void  ClassInfo_co18::operator delete (void* p, IliveCacheSink* lcSink) {
  SharedMemAllocator_co16.deallocate(lcSink, p);
}

#endif

/*----------------------------------------------------------------------------------*/

ContainerInfo_co18::ContainerInfo_co18 (ClassInfo_co18* pClassInfo, const tgg00_ContainerId& t, 
                                tsp00_Uint4 containerHandle, size_t psz, OmsSchemaHandle sh, OmsContainerNo cno) :
m_clsidHashNext(NULL), m_guidHashNext(NULL), m_clsInfo(pClassInfo),
m_PersistentSize(psz), m_ContainerNo(cno), m_schema(sh), m_dropped(false), m_refCnt(0) 
{
  memcpy (&m_ContainerId, &t, sizeof (m_ContainerId));
}

/*----------------------------------------------------------------------------------*/

void* ContainerInfo_co18::operator new(size_t sz, IliveCacheSink* lcSink) {
  return SharedMemAllocator_co16.allocate(lcSink, sz);
}

/*----------------------------------------------------------------------------------*/

void ContainerInfo_co18::DeleteSelf(IliveCacheSink* lcSink) {
#if defined(_ASSERT_OMS)
  memset (this, 0, sizeof(*this));
#endif
  SharedMemAllocator_co16.deallocate(lcSink, this);
}

/*----------------------------------------------------------------------------------*/
#if defined(PLACEMENT_DELETE_MEO00)

void  ContainerInfo_co18::operator delete (void* p, IliveCacheSink* lcSink) {
  SharedMemAllocator_co16.deallocate(lcSink, p);
}

#endif

/*----------------------------------------------------------------------------------*/

CContainerDictionary_co18::CContainerDictionary_co18() : IGuidHash_co001(), m_dropId(0) {
  this->Init();
}

/*----------------------------------------------------------------------------------*/

tgg00_BasisError CContainerDictionary_co18::DropContainer (IliveCacheSink* lcSink, 
                                                           const GUID& guid, 
                                                           OmsSchemaHandle schema,
                                                           OmsContainerNo cno) {
  tgg00_BasisError DBError = e_ok;
  CriticalSection_co001 cs(lcSink, RGN_CLASS_DIR);
  cs.Enter();
  CHResult_co001 hr = lcSink->DropContainer (schema, guid, cno, DBError);
  if (e_ok == DBError) {
    ContainerInfo_co18* p = this->FindGuid(guid, schema, cno);
    if (NULL != p) {
      p->MarkDropped();
    }
  }
  return DBError;
}

/*----------------------------------------------------------------------------------*/

tgg00_BasisError CContainerDictionary_co18::DropSchema (IliveCacheSink* lcSink, OmsSchemaHandle Schema) {
  tgg00_BasisError DBError;
  CriticalSection_co001 cs(lcSink, RGN_CLASS_DIR);
  cs.Enter();
  CHResult_co001 hr = lcSink->DropSchema (Schema, &DBError);
  if (0 == DBError) {
    for (ContainerIterator iter(this); iter; ++iter) {
      if (iter()->GetSchema() == Schema) {
        iter()->MarkDropped();
      }
    }
    ++m_dropId;
  }
  return DBError;
}

/*----------------------------------------------------------------------------------*/

ClassInfo_co18* CContainerDictionary_co18::FindGuid (const GUID guid) {
  int HashSlot  = HashValue(guid, CContainerDictionary_co18::MaxHeadEntries);
  ClassInfo_co18* p = m_classInfoHead[HashSlot];
  while (NULL != p) {
    if (p->GetGuid() == guid) {
      return p;
    }
    else {
      p = p->m_HashNext;
    }
  }
  return NULL;
}

/*----------------------------------------------------------------------------------*/

ClassInfo_co18* CContainerDictionary_co18::FindGuid (IliveCacheSink* lcSink, const GUID guid) {
  CriticalSection_co001 cs(lcSink, RGN_CLASS_DIR);
  cs.Enter();
  ClassInfo_co18* p = this->FindGuid(guid);
  return p;
}

/*----------------------------------------------------------------------------------*/

ContainerInfo_co18* CContainerDictionary_co18::FindViaContainerHandle  (IliveCacheSink* lcSink, tsp00_Uint4 containerHandle) {
  CriticalSection_co001 cs(lcSink, RGN_CLASS_DIR);
  cs.Enter();
  int clsidHashSlot = containerHandle % CContainerDictionary_co18::MaxHeadEntries;
  ContainerInfo_co18* p = m_clsidHead[clsidHashSlot];
  while (NULL != p) {
    if (p->GetContainerHandle() == containerHandle) {
      if (p->IsDropped()) {
        p = NULL;
      }
      break;
    }
    else {
      p = p->m_clsidHashNext;
    }
  }
  return p;
}

/*----------------------------------------------------------------------------------*/

ContainerInfo_co18* CContainerDictionary_co18::FindViaGuid (IliveCacheSink* lcSink, const GUID guid, OmsSchemaHandle sh, OmsContainerNo cno) {
  CriticalSection_co001 cs(lcSink, RGN_CLASS_DIR);
  cs.Enter();
  ContainerInfo_co18* p = this->FindGuid(guid, sh, cno);
  if ((p != NULL) && (p->IsDropped())) {
    p = NULL;
  }
  return p;
}

/*----------------------------------------------------------------------------------*/

ContainerInfo_co18* CContainerDictionary_co18::FindGuid(const GUID guid, OmsSchemaHandle sh, OmsContainerNo cno) {
  int guidHashSlot  = HashValue(guid, sh, cno, CContainerDictionary_co18::MaxHeadEntries);
  ContainerInfo_co18* p = m_guidHead[guidHashSlot];
  while (NULL != p) {
    if ((p->m_clsInfo->GetGuid() == guid) && (p->m_schema == sh) && (p->m_ContainerNo == cno)) {
      break;
    }
    else {
      p = p->m_guidHashNext;
    }
  }
  return p;
}

/*----------------------------------------------------------------------------------*/

void CContainerDictionary_co18::Init() {
  int ix;
  for (ix = 0; ix < MaxHeadEntries; ++ ix) {
    m_classInfoHead[ix] = NULL;
    m_clsidHead[ix]     = NULL;
    m_guidHead [ix]     = NULL;
  }
}

/*----------------------------------------------------------------------------------*/

ClassInfo_co18* CContainerDictionary_co18::RegisterClass (IliveCacheSink*        lcSink, 
                                                       const char*            ClassName,
                                                       const GUID&            guid,
                                                       const GUID*            pBaseClassGuid,
                                                       const tgg00_ObjColDesc &ColDesc,
                                                       size_t                 ObjectSize, 
                                                       void*                  vtptr) {
  ClassInfo_co18* p = FindGuid(guid);
  if (NULL == p) {
    int HashSlot  = HashValue(guid, CContainerDictionary_co18::MaxHeadEntries);
    if (NULL != pBaseClassGuid) {
      p = FindGuid(*pBaseClassGuid);
      if (NULL == p) {
        throw DbpError (DbpError::DB_ERROR, e_unknown_guid, "RegisterClass");
      }
      else {
        p->SetBaseClass();
      }
    }
    p = new(lcSink) ClassInfo_co18 (ClassName, guid, pBaseClassGuid, ColDesc, ObjectSize, vtptr, m_classInfoHead[HashSlot]);
    m_classInfoHead[HashSlot] = p;
  }
  return p;
}

/*----------------------------------------------------------------------------------*/

ContainerInfo_co18* CContainerDictionary_co18::RegisterContainer (IliveCacheSink* lcSink,
                                                       const GUID&            guid, 
                                                       const char*            ClassName, 
                                                       size_t                 PersistentSize, 
                                                       size_t                 ObjectSize,
                                                       const GUID*            pBaseClass,
                                                       const tgg00_ObjColDesc &ColDesc,
                                                       OmsSchemaHandle        sh, 
                                                       OmsContainerNo         ContainerNo,
                                                       void*                  vtptr) {
  HRESULT      hr = S_OK;
  tsp00_Int2   DBError = 0;
  CriticalSection_co001 cs(lcSink, RGN_CLASS_DIR);
  cs.Enter();
  ContainerInfo_co18* p = FindGuid (guid, sh, ContainerNo);
  if (NULL == p) {
    tgg00_ContainerId   containerId;
    tsp00_KnlIdentifier knlIdentifier;
    co18_ToKnlIdentifier(ClassName, knlIdentifier);
    hr = lcSink->RegisterClass (sh, CONST_CAST(GUID*, &guid), &knlIdentifier, PersistentSize, ContainerNo, 
      (unsigned char*) &ColDesc, (unsigned char*) &containerId, &DBError);
    if ((!FAILED(hr)) && (0 == DBError)) {
      ++ContainerHandle_co18;
      ClassInfo_co18* pClass = RegisterClass (lcSink, ClassName, guid, pBaseClass, ColDesc, ObjectSize, vtptr);
      p = new(lcSink) ContainerInfo_co18 (pClass, containerId, ContainerHandle_co18, PersistentSize, sh, ContainerNo);
      this->Insert(p);
    }
  }
  if (FAILED(hr)) {
    throw DbpError (DbpError::HRESULT_ERROR, hr);
  }
  if (0 != DBError) {
    throw DbpError (DbpError::DB_ERROR, DBError, "RegisterClass");
  }
  return p;
}

/*----------------------------------------------------------------------------------*/

void CContainerDictionary_co18::Delete (IliveCacheSink* lcSink, ContainerInfo_co18* pContainerInfo) {
  OMS_TRACE(omsTrContainerDir, lcSink, "CContainerDictionary_co18::Delete : " << pContainerInfo->GetClassInfoPtr()->GetClassName());
  int clsidHashSlot = pContainerInfo->GetContainerHandle() % CContainerDictionary_co18::MaxHeadEntries;
  pContainerInfo_co18* prev = &m_clsidHead[clsidHashSlot]; 
  pContainerInfo_co18  curr = *prev;
  bool found = false;
  while (NULL != curr) {
    if (curr == pContainerInfo) {
      *prev = pContainerInfo->m_clsidHashNext;
      found = true;
      break;
    }
    else {
      prev = &curr->m_clsidHashNext;
      curr = curr->m_clsidHashNext;
    }
  }
  if (found) {
    found = false;
    int guidHashSlot  = HashValue(pContainerInfo->m_clsInfo->GetGuid(), pContainerInfo->m_schema,
      pContainerInfo->m_ContainerNo, CContainerDictionary_co18::MaxHeadEntries);
    prev = &m_guidHead[guidHashSlot];
    curr = *prev;
    while (NULL != curr) {
      if (curr == pContainerInfo) {
        *prev = pContainerInfo->m_guidHashNext;
        found = true;
        break;
      }
      else {
        prev = &curr->m_guidHashNext;
        curr = curr->m_guidHashNext;
      }
    }
  }
  if (found) {
    pContainerInfo->DeleteSelf(lcSink);
  }
  else {
    throw DbpError (DbpError::DB_ERROR, -28999, "CContainerDictionary_co18::Delete");
  }
}

/*----------------------------------------------------------------------------------*/

void CContainerDictionary_co18::Insert (ContainerInfo_co18* pContainerInfo) { 
  int clsidHashSlot = pContainerInfo->GetContainerHandle() % CContainerDictionary_co18::MaxHeadEntries;
  int guidHashSlot  = HashValue(pContainerInfo->m_clsInfo->GetGuid(), pContainerInfo->m_schema,
    pContainerInfo->m_ContainerNo, CContainerDictionary_co18::MaxHeadEntries);
  pContainerInfo->m_clsidHashNext = m_clsidHead[clsidHashSlot]; 
  m_clsidHead[clsidHashSlot]      = pContainerInfo;
  pContainerInfo->m_guidHashNext  = m_guidHead[guidHashSlot];
  m_guidHead[guidHashSlot]        = pContainerInfo;
}

/*----------------------------------------------------------------------------------*/

void CContainerDictionary_co18::ReleaseContainerInfo(IliveCacheSink* lcSink, pContainerInfo_co18 p) {
  CriticalSection_co001 cs(lcSink, RGN_CLASS_DIR);
  cs.Enter();
  OMS_TRACE(omsTrContainerDir, lcSink, "ReleaseContainerInfo, RefCnt : " << p->m_refCnt);
  OMS_TRACE(omsTrContainerDir, lcSink, "Name : " << p->GetClassInfoPtr()->GetClassName() << "Schema : " 
    << p->GetSchema() << ", CNo : " << p->m_ContainerNo);
  if ((0 == p->DecRefCnt()) && (p->IsDropped())) {
    this->Delete(lcSink, p);
  }
}

/*----------------------------------------------------------------------------------*/

void CContainerDictionary_co18::Shutdown() {
  this->Init();
}

/*----------------------------------------------------------------------------------*/

CContainerDictionary_co18::ContainerIterator::ContainerIterator(CContainerDictionary_co18* clsDir) : 
m_clsDir(clsDir), m_index(0), m_curr(clsDir->m_clsidHead[0]) {
  while ((NULL == m_curr) && (++m_index < CContainerDictionary_co18::MaxHeadEntries)) {
    m_curr = m_clsDir->m_clsidHead[m_index];
  }
}

/*----------------------------------------------------------------------------------*/

void CContainerDictionary_co18::ContainerIterator::operator++() {
  if (NULL != m_curr) { 
    m_curr = m_curr->m_clsidHashNext;
    while (NULL == m_curr) { // PTS 1106820
      ++m_index;
      if (m_index < CContainerDictionary_co18::MaxHeadEntries) {
        m_curr = m_clsDir->m_clsidHead[m_index];
      }
      else {
        return;
      }
    }
  }
}

/*----------------------------------------------------------------------------------*/

