/* @lastChanged: "1998-07-10  16:03"
 * @filename:    vco20.cpp
 * @purpose:     "System Stored Procedures"
 * @release:     7.2.0.0  
 * @see:         ""
 *
 * @copyright:   "Copyright by SAP AG, 1998"


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 */

/****************************************************************************/
/* DbpInstallObj.cpp : Implementation of CDbpInstall                            */
/****************************************************************************/

#if defined(WIN32)
#include <string>
#else
#include <stdlib.h>
#endif
#include "heo02x.h"
#include "lcbasetypes.h"      /* no check */
#include "livecachetypes.h"   /* no check */
#include "hco20idl.h"         /* no check */
#include "hco20.h"

/*===========================================================================*
 *  LOCAL FUNCTIONS (PROTOTYPES)                                             *
 *===========================================================================*/

#define MX_KNL_IDENTIFIER   64
#define STR_IN(P) SqlCol(P, SqlCol::VSTRING, strlen(P))
#define STR_OUT(P) SqlCol(&P[0], SqlCol::VSTRING, sizeof(P))

void truncateBlanks (CPath& Path) {
  int len = sizeof(Path) - 1;
  while (' ' == Path[len]) {
    len--;
  }
  len++;
  if (len < sizeof(Path)) {
    Path[len] = 0;
  }
}

/****************************************************************************/
/* Method Register is the very first Method called after the COM-Objekt     */
/* has been created.                                                        */
/* The CALLBACK Interface (IKernelSink) has been activated and may be used  */
/* (within the "constructor" of the COM-object IKernelSink hasn't been      */
/*  activated yet !)                                                        */

STDMETHODIMP CDbpInstall::Register()
{
  HRESULT hr	= S_OK;
  try {
    
    co90CheckVersion ((IUnknown**)&this->m_pIUnknown, 
                      ILIVECACHESINK_VERSION, 
                      LIVECACHETYPES_VERSION, LIVECACHEOMS_VERSION, 
                      LIVECACHESQL_VERSION );
/* *** TODO ***                                                            */
/* Register all persistent objects used by DB-Procedures here              */
/* This Routine is called, before the first DB-Procedure gets control.     */
  }
  catch (DbpError e) {
    omsExceptionHandler (e);
  }
	return hr;
}

/****************************************************************************/
/* IDbpInstall method implementation  (to be inserted here)                   */
/*-------------------------------------------------------------------------------*/

STDMETHODIMP CDbpInstall::COPY_AND_REG_DLL(CPath * DllName, DllHandle * Handle, int regDll)
{
	enum { bufSize = 4096 };
    
 
  FILE  *dll = NULL;
  try {
    CPath         Path;
    int           bufCnt = 0;
    unsigned char buf[bufSize];
    unsigned char *p;
    char          *InstRoot;

    truncateBlanks(*DllName);
    InstRoot = getenv("INSTROOT");
    if (!InstRoot) {
      throw DbpError (-2, "INSTROOT not set");
    }
    if (strlen(InstRoot) + 5 + strlen ((char*) DllName) >= sizeof(Path)) {
      throw DbpError (-3, "Path too long");
    }
    strcpy ((char*) &Path[0], InstRoot);
    strcat ((char*) &Path[0], "\\sap\\");
    strcat ((char*) &Path[0], (char*) DllName);

    OmsIStream<DllHandle, unsigned char> inStream(this, *Handle);
    if( NULL == (dll  = fopen( REINTERPRET_CAST(char*,Path), "wb" )) ) {
      throw DbpError (-1, "could not open output file\n" );
    }
    while (p = inStream.omsRead()) {
      if (bufCnt < bufSize) {
        buf[bufCnt] = *p;
        bufCnt++;
      }
      else {
        fwrite( buf, 1, bufCnt, dll);
        buf[0] = *p;
        bufCnt = 1;
      }
    }
    if (bufCnt > 0) {
      fwrite( buf, 1, bufCnt, dll);
    }
    fclose (dll);
    
    if (regDll) {
      REG_DLL (&Path);
    }
  }
  catch (DbpError e)
  {
    if (dll) {
      fclose (dll);
    }
    omsExceptionHandler(e);
  }
  return S_OK;
}

/*-------------------------------------------------------------------------------*/

STDMETHODIMP CDbpInstall::ACTIVATE_DLL(CProgId * pProgId)
{
  try {
    char stmt[256];
    truncateBlanks(*pProgId);
    sqlSetMode(SqlOpt::SQLMODE_ADABAS);         
    SQL sel = sql ("Select OWNER, DBPROCNAME from domain.dbprocedures where Prog_id = ?");
    (sel << STR_IN((char*)pProgId)).sqlExecute();
    if (100 != sqlCode()) {
      int cnt = 0;
      int resCnt = sqlResultCount();
      SQL fetch  = sql ("Fetch Into ?, ?");
      char owner[MX_KNL_IDENTIFIER + 1];
      char dbproc[MX_KNL_IDENTIFIER + 1];
      fetch << STR_OUT(owner) << STR_OUT(dbproc);
      do {
        fetch.sqlExecute();
        if (0 == sqlCode()) {
          cnt++;
          strcpy(&stmt[0],"CREATE OR REPLACE DBPROC ");
          strcat(&stmt[0], dbproc);
          strcat(&stmt[0], " AS '");
          strcat(&stmt[0], (char*)pProgId);
          strcat(&stmt[0], "' EXECUTE INPROC");
          if ( cnt == resCnt ) {
            strcat(&stmt[0], " COMMIT");
          }
          sql (&stmt[0])();
        }
        else {
          break;
        }
      }
      while (0 == sqlCode());
    }
  }
  catch (DbpError e)
  {
    omsExceptionHandler(e);
  }
  
  return S_OK;
}

/*-------------------------------------------------------------------------------*/

STDMETHODIMP CDbpInstall::REG_DLL(CPath * Path)
{
  tsp_errtext       ErrText;
  int               ErrTextLen = sizeof(ErrText);

  typedef HRESULT (WINAPI *REGISTERSERVER) (void);
  
  truncateBlanks(*Path);

  try {
    HINSTANCE hInst = (HINSTANCE)sqlLoadLibrary(REINTERPRET_CAST(char*,Path), (char*) &ErrText[0], ErrTextLen);
    if ( NULL == hInst ) 
    {
      throw DbpError (-2, "could not load dll");
    }
    REGISTERSERVER r = (REGISTERSERVER)sqlGetProcAddress ( hInst ,"DllRegisterServer", (char*) &ErrText[0], ErrTextLen) ;
    
    if ( NULL == r )
    {  
      throw DbpError (-3, "could not get proc address"); 
    }
    
    HRESULT hr = r ( );
    if (0 != hr) {
      throw DbpError (-4, "could not register"); 
    }
    
    sqlFreeLibrary(hInst, (char*) &ErrText[0], ErrTextLen);
  }
  catch (DbpError e)
  {
    omsExceptionHandler(e);
  }
  return S_OK;
}

/*-------------------------------------------------------------------------------*/
/* PTS 1103927 */

STDMETHODIMP CDbpInstall::FORCE_AND_WAIT_FOR_CHECKPOINT() {
  char emsg[256];                                         /* PTS 1110813 FF */
  SqlHandle::sqlHandler orgHandler = sqlSetHandler(NULL); /* PTS 1110813 FF don't use exception handling */
  try {
    OmsTypeInt4 ckptId;
    this->FORCE_CHECKPOINT   (&ckptId);
    if (ckptId > 0) {
      this->WAIT_FOR_CHECKPOINT(ckptId);
      if (  (0 != sqlCode()) && ( 100 != sqlCode()) ) goto DBErrorExit; /* PTS 1110813 FF */
    }
    else {
      sprintf (emsg, "Checkpoint already working");               /* PTS 1110813 FF */
      STATIC_CAST(SqlHandle*,this)->sqlRaiseError(500,&emsg[0]);  /* PTS 1110813 FF */
      sqlSetHandler(orgHandler);                                  /* PTS 1110813 FF */
      return S_OK;                                                /* PTS 1110813 FF */
      /* throw DbpError(500, "Checkpoint already working");          PTS 1110813 FF */
    }
  }
  catch (DbpError e)
  {
    sqlExceptionHandler(e);
  }
  sqlSetHandler(orgHandler);    /* PTS 1110813 FF */
  return S_OK;

  /* DBErrorExist has been coded for PTS 1110813 FF */
DBErrorExit:
  sqlSetHandler(orgHandler);
  STATIC_CAST(SqlHandle*,this)->sqlRaiseError(sqlCode(),sqlErrorMsg());
  return S_OK;
}

/*-------------------------------------------------------------------------------*/

/* PTS 1102637, T.A. 1999-05-06 */

STDMETHODIMP CDbpInstall::FORCE_CHECKPOINT(OmsTypeInt4 *chptId) {
  SqlHandle::sqlHandler orgHandler = sqlSetHandler(NULL); /* PTS 1110813 FF don't use exception handling */
  /* try {  PTS 1110813 FF */
    SQL forceCheckpoint = sql("force checkpoint ?");
    forceCheckpoint << SqlCol(*chptId);
    forceCheckpoint.sqlExecute();
  /* }               PTS 1110813 FF
  catch (DbpError e) PTS 1110813 FF*/
  if (  (0 != sqlCode()) && ( 100 != sqlCode()) ) /* PTS 1110813 FF */
  {
    // PTS 1105865
    if ((500 == sqlCode()/* PTS 1110813 e.dbpError()*/) || 
      (-51 == sqlCode()/* PTS 1110813 e.dbpError()*/)) {
      *chptId = -1;
    }
    else {
      /* sqlExceptionHandler(e);  PTS 1110813 FF  */
      STATIC_CAST(SqlHandle*,this)->sqlRaiseError(sqlCode(),sqlErrorMsg()); /* PTS 1110813 FF */
    }
  }
  return S_OK;
}


/*-------------------------------------------------------------------------------*/

STDMETHODIMP CDbpInstall::WAIT_FOR_CHECKPOINT(OmsTypeInt4 chptId) {
  SqlHandle::sqlHandler orgHandler = sqlSetHandler(NULL); /* PTS 1110813 FF don't use exception handling */
  /* try {  PTS 1110813 FF */
    SQL waitForCheckpoint = sql("wait for checkpoint ?");
    waitForCheckpoint << SqlCol(chptId);
    waitForCheckpoint.sqlExecute();
  /* }               PTS 1110813 FF
  catch (DbpError e) PTS 1110813 FF*/
  if (  (0 != sqlCode()) && ( 100 != sqlCode()) ) /* PTS 1110813 FF */
  {
    /* sqlExceptionHandler(e);  PTS 1110813 FF  */
    STATIC_CAST(SqlHandle*,this)->sqlRaiseError(sqlCode(),sqlErrorMsg()); /* PTS 1110813 FF */
  }
  sqlSetHandler(orgHandler);  /* PTS 1110813 FF */
  return S_OK;
}
/* END PTS */
/*-------------------------------------------------------------------------------*/

STDMETHODIMP CDbpInstall::SYSDDLTRIGGER()
{
  
  /*returncode of a sub function call, like DropTable
	this value is necessary because global function sqlCode()
	keeps last error code e.g. 100 row not found*/
  long retSubFunction = 0;

  char action[MX_KNL_IDENTIFIER + 1];
  char objtype[MX_KNL_IDENTIFIER + 1];
  char param1[MX_KNL_IDENTIFIER + 1];
  char param2[MX_KNL_IDENTIFIER + 1];
  char param3[MX_KNL_IDENTIFIER + 1];
  char param4[MX_KNL_IDENTIFIER + 1];


  SqlHandle::sqlHandler orgHandler = sqlSetHandler(NULL); /* PTS 1110813 FF don't use exception handling */
  try {
    sqlSetMode(SqlOpt::SQLMODE_ADABAS);         
    SQL stmt  = sql ("SELECT SYSRESULT (ACTION, OBJECTTYPE, PARAM1, PARAM2, PARAM3, PARAM4) FROM SYSDD.PARAMETERS");
    SQL fetch = sql ("FETCH  SYSRESULT INTO ?,?,?,?,?,?");
    SQL close = sql ("CLOSE  SYSRESULT");
    stmt.sqlExecute();
    if (  (0 != sqlCode()) && ( 100 != sqlCode()) ) goto DBErrorExit;      /* PTS 1110813 FF */
    fetch << STR_OUT(action) << STR_OUT(objtype) << STR_OUT(param1) << STR_OUT(param2) << STR_OUT(param3)  << STR_OUT(param4);
    
    /*continue while sqlCode = 0 or retSubFunction is row not found*/
    while (0 == sqlCode() || retSubFunction == 100) {
      
      /*fetch next action*/
      fetch.sqlExecute();
      if (  (0 != sqlCode()) && ( 100 != sqlCode()) ) goto DBErrorExit;      /* PTS 1110813 FF */
      /*reinit retSubFunction*/
      retSubFunction = 0;
      
      /*if there is one more action handle it*/
      if (0 == sqlCode()) {
        if (!strcmp(&action[0], "DROP")) {
          if ( !strcmp(&objtype[0], "USER")) {
            
            retSubFunction = DropUser (param2);		
            if (  (0 != retSubFunction) && ( 100 != retSubFunction) ) goto DBErrorExit;  /* PTS 1110813 FF */
            continue;  
          }
          //SYSTEM Tabellen sind mit objtype[0], "TABLE" gekennzeichnet	
          if ( (!strcmp(&objtype[0], "TABLE")) || (!strcmp(&objtype[0], "SYNONYM")) || (!strcmp(&objtype[0], "VIEW")) ) {
            
            retSubFunction = DropTable (param1,param2);
            if (  (0 != retSubFunction) && ( 100 != retSubFunction) ) goto DBErrorExit;      /* PTS 1107376 FF */
            continue;  				
          }		  
          if (!strcmp(&objtype[0], "COLUMN")) {
            
            retSubFunction = DropColumn (param1,param2,param3);
            if (  (0 != retSubFunction) && ( 100 != retSubFunction) ) goto DBErrorExit;      /* PTS 1107376 FF */
            continue;  				
          }		  
        }
        if (! strcmp(&action[0], "RENAME")) {
          if ( ! strcmp(&objtype[0], "USER")) {
            
            retSubFunction = RenameUser (param2, param3);
            if (  (0 != retSubFunction) && ( 100 != retSubFunction) ) goto DBErrorExit;  /* PTS 1110813 FF */
            continue;  
          }		
          
          //SYSTEM Tabellen sind mit objtype[0], "TABLE" gekennzeichnet	
          if ( (!strcmp(&objtype[0], "TABLE"))  || (!strcmp(&objtype[0], "SYNONYM")) || (!strcmp(&objtype[0], "VIEW")) ) {		  
            
            retSubFunction = RenameTable (param1,param2, param3);
            if (  (0 != retSubFunction) && ( 100 != retSubFunction) ) goto DBErrorExit;      /* PTS 1107376 FF */
            continue;  
          }          
          
          if ( ! strcmp(&objtype[0], "COLUMN")) {
            
            retSubFunction = RenameColumn (param1, param2, param3, param4);
            if (  (0 != retSubFunction) && ( 100 != retSubFunction) ) goto DBErrorExit;      /* PTS 1107376 FF */
            continue;  
          }          
        }
      }
    }
    close.sqlExecute();
    if (  (0 != sqlCode()) && ( 100 != sqlCode()) ) goto DBErrorExit;      /* PTS 1110813 FF */
  }
  catch (DbpError e) {
    
  /*
		if error is -4004 then the system table which has to be updated doesn't exist,
    so the trigger doesn't throw a error, beacuase there is nothing to do.
    */
    long lRet = 0;
    if ((lRet = sqlCode()) != -4004)	{
      
      char msg[256];
      
      sprintf (msg, "DDL Trigger failed : %d", e.dbpError());
      DbpError err(-9406, msg);
      sqlExceptionHandler(err);
    }
  }
  sqlSetHandler(orgHandler); /* PTS 1110813 FF reset exception handling */
  return S_OK;

  /* DBErrorExist has been coded for PTS 1110813 FF */
DBErrorExit:
  char emsg[256];
  
  sqlSetHandler(orgHandler); 
  if ( sqlCode() != -4004)	{
    sprintf (emsg, "DDL Trigger failed : %d", sqlCode());
    STATIC_CAST(SqlHandle*,this)->sqlRaiseError(-9406,&emsg[0]);
    return -1;                 
  }
  return S_OK;
}

/*-------------------------------------------------------------------------------*/

STDMETHODIMP CDbpInstall::DROP_CURSOR()
{
  
  char cursor[MX_KNL_IDENTIFIER + 1];

  try {
    sqlSetMode(SqlOpt::SQLMODE_ADABAS);         
    SQL stmt  = sql ("SELECT SYSRESULT (TABLENAME) FROM SUT.RESULTTABLES");
    SQL fetch = sql ("FETCH  SYSRESULT INTO ?");
	SQL close = sql ("CLOSE  SYSRESULT");
    stmt.sqlExecute();
    fetch << STR_OUT(cursor);
    while (0 == sqlCode()) {
      fetch.sqlExecute();
      if (0 == sqlCode()) {

            CloseCursor (cursor);		
			continue;  
      }
    }
    close.sqlExecute();
  }
  catch (DbpError e) {

		char msg[256];

		sprintf (msg, "DDL Trigger failed : %d", e.dbpError());
		DbpError err(-9406, msg);
		sqlExceptionHandler(err);
	}

  return S_OK;
}


/*===========================================================================*
 *  LOCAL FUNCTIONS (CODE)                                                   *
 *===========================================================================*/

/*!-------------------------------------------------------------------------
function:     DropUser
description:  if a user ... is dropped, the item in the systemview SYSQP_MY_FAVORITES is cleared
arguments:    user [in]  name of user
              
              
return value: long, sql error code
--------------------------------------------------------------------------*/
long CDbpInstall::DropUser (char* user)
{
  SQL delAllObjects = sql ("DELETE SYSDBA.SYSQP_ALL_OBJECTS WHERE SYSQP_ALL_OBJECTS.OWNER = ?");
  delAllObjects << STR_IN(user);
  delAllObjects.sqlExecute();
  if (  (0 != sqlCode()) && ( 100 != sqlCode()) ) return sqlCode();      /* PTS 1110813 FF */

  SQL delAllUserProfiles = sql ("DELETE SYSDBA.SYSQP_ALL_USER_PROFILES WHERE \"UID\" = ?");
  delAllUserProfiles << STR_IN(user);
  delAllUserProfiles.sqlExecute();
  if (  (0 != sqlCode()) && ( 100 != sqlCode()) ) return sqlCode();      /* PTS 1110813 FF */

  SQL delAllSharedObjects = sql ("DELETE SYSDBA.SYSQP_ALL_SHARED_OBJECTS WHERE \"GRANTEE\" = ? OR \"OWNER\" = ?");
  delAllSharedObjects << STR_IN(user) << STR_IN(user);
  delAllSharedObjects.sqlExecute();
  if (  (0 != sqlCode()) && ( 100 != sqlCode()) ) return sqlCode();      /* PTS 1110813 FF */

  /*Die Spalte  FITEMTYPE = 7 entspricht einem FavoriteItem vom Typ User*/
  SQL delAllFavorites = sql ("DELETE SYSDBA.SYSQP_MY_FAVORITES WHERE SYSQP_MY_FAVORITES.FITEMNAME = ? AND SYSQP_MY_FAVORITES.FITEMTYPE = 7" );
  delAllFavorites << STR_IN(user);
  delAllFavorites.sqlExecute();
  if (  (0 != sqlCode()) && ( 100 != sqlCode()) ) return sqlCode();      /* PTS 1110813 FF */

  SQL delWAInode = sql ("DELETE  SYSDBA.WA_INODE WHERE \"VIEWOWNER\" = ?");
  delWAInode << STR_IN(user);
  delWAInode.sqlExecute();
  if (  (0 != sqlCode()) && ( 100 != sqlCode()) ) return sqlCode();      /* PTS 1110813 FF */

  SQL delWAContainer = sql ("DELETE  SYSDBA.WA_CONTAINER WHERE \"VIEWOWNER\" = ?");
  delWAContainer << STR_IN(user);
  delWAContainer.sqlExecute();
  if (  (0 != sqlCode()) && ( 100 != sqlCode()) ) return sqlCode();      /* PTS 1107376 FF */

  SQL delSysUpdStat = sql ("DELETE  SYSDBA.SYSUPDSTATWANTED WHERE OWNER = ?");
  delSysUpdStat << STR_IN(user);
  delSysUpdStat.sqlExecute();

  return sqlCode();
}

/*!-------------------------------------------------------------------------
function:     DropTable 
description:  if a table,view ... is dropped, the item in the system view SYSQP_MY_FAVORITES is cleared
arguments:    owner [in]  owner of the table, view ...
              table [in]  name of the table, view ...
              
return value: long, sql error code
--------------------------------------------------------------------------*/
long CDbpInstall::DropTable (char* owner,char* table)
{   	
  /*Die Spalte FITEMTYPE = 1 - 6 entspricht einem FavoriteItem vom Typ Table,View,Result,System,Snapshot,Synonym*/
  SQL delAllFavorites = sql ("DELETE SYSDBA.SYSQP_MY_FAVORITES WHERE SYSQP_MY_FAVORITES.FITEMOWNER = ? AND SYSQP_MY_FAVORITES.FITEMNAME = ? AND SYSQP_MY_FAVORITES.FITEMTYPE  BETWEEN  1 AND 6" );
  delAllFavorites << STR_IN(owner) << STR_IN(table);
  delAllFavorites.sqlExecute();
  if (  (0 != sqlCode()) && ( 100 != sqlCode()) ) return sqlCode();  

  SQL delSysUpdStat = sql ("DELETE  SYSDBA.SYSUPDSTATWANTED WHERE OWNER = ? AND TABLENAME = ?");
  delSysUpdStat << STR_IN(owner) << STR_IN(table);
  delSysUpdStat.sqlExecute();

  return sqlCode();
}

/*!-------------------------------------------------------------------------
function:     DropColumn
description:  if a column is dropped, the item in the system table SYSUPDSTATWANTED is cleared
arguments:    owner  [in]  owner of the table, view ...
              table  [in]  name of the table, view ...
              column [in]  name of the column ...
              
return value: long, sql error code
--------------------------------------------------------------------------*/
long CDbpInstall::DropColumn (char* owner,char* table,char* column)
{   	
  
  SQL delSysUpdStat = sql ("DELETE  SYSDBA.SYSUPDSTATWANTED WHERE OWNER = ? AND TABLENAME = ? AND COLUMNNAME = ?");
  delSysUpdStat << STR_IN(owner) << STR_IN(table) << STR_IN(column);
  delSysUpdStat.sqlExecute();

  return sqlCode();
}

/*-------------------------------------------------------------------------------*/

void CDbpInstall::CloseCursor (char* cursor)
{
  char stmt[256];
  strcpy(&stmt[0],"CLOSE ");
  strcat(&stmt[0], cursor);
  SQL closeCursor = sql (stmt);
  closeCursor.sqlExecute();

}

/*-------------------------------------------------------------------------------*/

/*!-------------------------------------------------------------------------
function:     RenameUser
description:  if a user ... is renamed, the item in the systemview SYSQP_MY_FAVORITES is updated
arguments:    oldUserName [in]  name of user
			  newUserName [in]  name of user
              
              
return value: long, sql error code
--------------------------------------------------------------------------*/
long CDbpInstall::RenameUser (char* oldUserName, char* newUserName)
{
  SQL updAllObjects = sql ("UPDATE SYSDBA.SYSQP_ALL_OBJECTS SET \"OWNER\" = ? WHERE \"OWNER\" = ?");
  updAllObjects << STR_IN(newUserName) << STR_IN(oldUserName);
  updAllObjects.sqlExecute();
  if (  (0 != sqlCode()) && ( 100 != sqlCode()) ) return sqlCode();      /* PTS 1110813 FF */

  SQL updAllUserProfiles = sql ("UPDATE SYSDBA.SYSQP_ALL_USER_PROFILES SET \"UID\" = ? WHERE \"UID\" = ?");
  updAllUserProfiles << STR_IN(newUserName) << STR_IN(oldUserName);
  updAllUserProfiles.sqlExecute();
  if (  (0 != sqlCode()) && ( 100 != sqlCode()) ) return sqlCode();      /* PTS 1110813 FF */

  SQL updAllSharedObjects_1 = sql("UPDATE  SYSDBA.SYSQP_ALL_SHARED_OBJECTS SET \"OWNER\" = ? WHERE  \"OWNER\" = ?");
  updAllSharedObjects_1 << STR_IN(newUserName) << STR_IN(oldUserName);
  updAllSharedObjects_1.sqlExecute(); 
  if (  (0 != sqlCode()) && ( 100 != sqlCode()) ) return sqlCode();      /* PTS 1110813 FF */

  SQL updAllSharedObjects_2 = sql("UPDATE  SYSDBA.SYSQP_ALL_SHARED_OBJECTS SET \"GRANTEE\" = ? WHERE \"GRANTEE\" = ?");
  updAllSharedObjects_2 << STR_IN(newUserName) << STR_IN(oldUserName);
  updAllSharedObjects_2.sqlExecute(); 
  if (  (0 != sqlCode()) && ( 100 != sqlCode()) ) return sqlCode();      /* PTS 1110813 FF */

  /*Die Spalte FITEMTYPE = 7 entspricht einem FavoriteItem vom Typ User */
  SQL upAllFavorites = sql ("UPDATE SYSDBA.SYSQP_MY_FAVORITES SET \"FITEMNAME\"  = ? WHERE SYSQP_MY_FAVORITES.FITEMNAME = ? AND SYSQP_MY_FAVORITES.FITEMTYPE = 7" );
  upAllFavorites << STR_IN(newUserName) << STR_IN(oldUserName);
  upAllFavorites.sqlExecute();
  if (  (0 != sqlCode()) && ( 100 != sqlCode()) ) return sqlCode();      /* PTS 1110813 FF */

  SQL updWAInode = sql ("UPDATE SYSDBA.WA_INODE SET \"VIEWOWNER\" = ? WHERE \"VIEWOWNER\" = ?");
  updWAInode << STR_IN(newUserName) << STR_IN(oldUserName);
  updWAInode.sqlExecute();
  if (  (0 != sqlCode()) && ( 100 != sqlCode()) ) return sqlCode();      /* PTS 1110813 FF */

  SQL updWAContainer = sql ("UPDATE SYSDBA.WA_CONTAINER SET \"VIEWOWNER\" = ? WHERE \"VIEWOWNER\" = ?");
  updWAContainer << STR_IN(newUserName) << STR_IN(oldUserName);
  updWAContainer.sqlExecute();
  if (  (0 != sqlCode()) && ( 100 != sqlCode()) ) return sqlCode();  

  /* auf SYSUPDSTATWANTED kann kein Key Update durchgefhrt werden, daher ...*/
  sqlSetMode(SqlOpt::SQLMODE_ADABAS);         
  SQL sel = sql ("SELECT SYSUPDSTAT_CURSOR (OWNER, TABLENAME, COLUMNNAME) FROM SYSDBA.SYSUPDSTATWANTED WHERE OWNER = ? FOR REUSE");
  sel << STR_IN(oldUserName);
  SQL fetch = sql ("FETCH  SYSUPDSTAT_CURSOR INTO ?, ?");
  char tname[MX_KNL_IDENTIFIER + 1];
  char cname[MX_KNL_IDENTIFIER + 1];
  fetch << STR_OUT(tname) << STR_OUT(cname);
  SQL close = sql ("CLOSE  SYSUPDSTAT_CURSOR");
  SQL del = sql ("DELETE  SYSDBA.SYSUPDSTATWANTED WHERE OWNER = ?");
  del << STR_IN(oldUserName);
  SQL ins = sql ("INSERT SYSDBA.SYSUPDSTATWANTED COLUMNS (OWNER, TABLENAME, COLUMNNAME) VALUES (?, ?, ?)");
  sel.sqlExecute();
  del.sqlExecute();
  while (0 == sqlCode()) {
    fetch.sqlExecute();
    if (0 == sqlCode()) {
      ins << STR_IN(newUserName) << STR_IN(tname) << STR_IN(cname);
      ins.sqlExecute();
      continue;  
    }
  }
  close.sqlExecute();

  return sqlCode();
}

/*!-------------------------------------------------------------------------
function:     RenameTable 
description:  if a table,view ... is renamed, the item in the system view of the SYSQP_MY_FAVORITES is updated
arguments:    owner [in]  owner of the table, view ...
              oldTableName  [in], old name of the table, view ...
			  newTableName [in], new name of the table, view ...
              
return value: long, sql error code
--------------------------------------------------------------------------*/
long CDbpInstall::RenameTable (char* owner,char* oldTableName, char* newTableName)
{

  /* Die Spalte FITEMTYPE = 1 - 6 entspricht einem FavoriteItem vom Typ Table,View,Result,System,Snapshot,Synonym */
  SQL upAllFavorites = sql ("UPDATE SYSDBA.SYSQP_MY_FAVORITES SET \"FITEMNAME\"  = ? WHERE SYSQP_MY_FAVORITES.FITEMOWNER = ? AND SYSQP_MY_FAVORITES.FITEMNAME = ? AND SYSQP_MY_FAVORITES.FITEMTYPE BETWEEN  1 AND 6");						 
  upAllFavorites << STR_IN(newTableName) << STR_IN(owner)<< STR_IN(oldTableName);
  upAllFavorites.sqlExecute();
  if (  (0 != sqlCode()) && ( 100 != sqlCode()) ) return sqlCode();  


  /* auf SYSUPDSTATWANTED kann kein Key Update durchgefhrt werden, daher ...*/
  sqlSetMode(SqlOpt::SQLMODE_ADABAS);         
  SQL sel = sql ("SELECT SYSUPDSTAT_CURSOR (COLUMNNAME) FROM SYSDBA.SYSUPDSTATWANTED WHERE OWNER = ? AND TABLENAME = ? FOR REUSE");
  sel << STR_IN(owner) << STR_IN(oldTableName);
  SQL fetch = sql ("FETCH  SYSUPDSTAT_CURSOR INTO ?");
  char cname[MX_KNL_IDENTIFIER + 1];
  fetch << STR_OUT(cname);
  SQL close = sql ("CLOSE  SYSUPDSTAT_CURSOR");
  SQL del = sql ("DELETE  SYSDBA.SYSUPDSTATWANTED WHERE OWNER = ? AND TABLENAME = ?");
  del << STR_IN(owner) << STR_IN(oldTableName);
  SQL ins = sql ("INSERT SYSDBA.SYSUPDSTATWANTED COLUMNS (OWNER, TABLENAME, COLUMNNAME) VALUES (?, ?, ?)");
  sel.sqlExecute();
  del.sqlExecute();
  while (0 == sqlCode()) {
    fetch.sqlExecute();
    if (0 == sqlCode()) {
      ins << STR_IN(owner) << STR_IN(newTableName) << STR_IN(cname);
      ins.sqlExecute();
      continue;  
    }
  }
  close.sqlExecute();
  return sqlCode();
  
}

/*!-------------------------------------------------------------------------
function:     RenameColumn 
description:  if a column within a table,view ... is renamed, the item in the system table SYSUPDSTATWANTED is updated
arguments:    owner [in]  owner of the table, view ...
              TableName  [in], name of the table, view ...
			  oldColumnName [in], old name of the column
			  newColumnName [in], new name of the column
              
return value: long, sql error code
--------------------------------------------------------------------------*/
long CDbpInstall::RenameColumn (char* owner, char* table, char* oldColumnName, char* newColumnName)
{

  /* auf SYSUPDSTATWANTED kann kein Key Update durchgefhrt werden, daher ...*/
  SQL del = sql ("DELETE  SYSDBA.SYSUPDSTATWANTED WHERE OWNER = ? AND TABLENAME = ? AND COLUMNNAME = ?");
  del << STR_IN(owner) << STR_IN(table) << STR_IN(oldColumnName);
  del.sqlExecute();

  SQL ins = sql ("INSERT SYSDBA.SYSUPDSTATWANTED COLUMNS (OWNER, TABLENAME, COLUMNNAME) VALUES (?, ?, ?)");
  ins << STR_IN(owner) << STR_IN(table) << STR_IN(newColumnName);
  ins.sqlExecute();

  return sqlCode();
  
}

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/
