/****************************************************************************

  module      : vco31.cpp
  
  -------------------------------------------------------------------------
    
  responsible : ThomasA
      
  special area: liveCache Error Class
  description : Defines Functions of an error thrown by OMS and SQL class
        
  last changed: 2000-01-10  9:34
           
  -------------------------------------------------------------------------
            
  copyright:    Copyright by SAP AG, 1999
              


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


/*===========================================================================*
*  INCLUDES                                                                 *
*===========================================================================*/


#include "hco31-k.h"
#include "hsp81.h"

/*===========================================================================*
*  DEFINES                                                                  *
*===========================================================================*/

DbpError::DbpError (DbpErrorKinds ErrorKind, long ErrorNo) {
  m_errorKind    = ErrorKind;
  m_errorNo      = ErrorNo;
  m_errorText[0] = 0;
};

/*----------------------------------------------------------------------*/

DbpError::DbpError (DbpErrorKinds ErrorKind, long ErrorNo, const char* errtext) {
  m_errorKind    = ErrorKind;
  m_errorNo      = ErrorNo;
  strcpy (m_errorText, errtext);
};

/*----------------------------------------------------------------------*/

DbpError::DbpError (DbpErrorKinds ErrorKind, long ErrorNo, const DbpTypeUnicode* errtext) {
  m_errorKind    = ErrorKind;
  m_errorNo      = ErrorNo;
  sp81UCS2strcpy ((tsp81_UCS2Char*)(m_errorText), (tsp81_UCS2Char*)errtext);
};

/*----------------------------------------------------------------------*/

DbpError::DbpError (DbpErrorKinds ErrorKind, long ErrorNo, const char* errtext, const OmsObjectId& oid) {
  m_errorKind    = ErrorKind;
  m_errorNo      = ErrorNo;
  m_oid          = oid;
  strcpy (m_errorText, errtext);
};

/*----------------------------------------------------------------------*/

DbpError::DbpError (DbpErrorKinds ErrorKind, long ErrorNo, const DbpTypeUnicode* errtext, const OmsObjectId& oid) {
  m_errorKind    = ErrorKind;
  m_errorNo      = ErrorNo;
  m_oid          = oid;
  sp81UCS2strcpy ((tsp81_UCS2Char*)(m_errorText), (tsp81_UCS2Char*)errtext);
};

/*----------------------------------------------------------------------*/

DbpError::DbpError (long ErrorNo) {
  m_errorKind    = USER_DEFINED;
  m_errorNo      = ErrorNo;
  m_errorText[0] = 0;
};

/*----------------------------------------------------------------------*/

DbpError::DbpError (long ErrorNo, const char* errtext) {
  m_errorKind    = USER_DEFINED;
  m_errorNo      = ErrorNo;
  strcpy (m_errorText, errtext);
};

/*----------------------------------------------------------------------*/

DbpError::DbpError (long ErrorNo, const DbpTypeUnicode* errtext) {
  m_errorKind    = USER_DEFINED_UNICODE;
  m_errorNo      = ErrorNo;
  sp81UCS2strcpy ((tsp81_UCS2Char*)(m_errorText), (tsp81_UCS2Char*)errtext);
};

/*----------------------------------------------------------------------*/

long DbpError::dbpError () const { 
  return m_errorNo; 
};

/*----------------------------------------------------------------------*/

short DbpError::dbpDBError () const { 
  return (m_errorKind == DB_ERROR) ? m_errorNo : 0; 
};

/*----------------------------------------------------------------------*/

short DbpError::dbpRteError () const { 
  return (m_errorKind == RTE_ERROR) ? m_errorNo : 0; 
};

/*----------------------------------------------------------------------*/

long  DbpError::dbpHresultError() const { 
  return (m_errorKind == HRESULT_ERROR) ? m_errorNo : 0; 
};

/*----------------------------------------------------------------------*/

const char* DbpError::dbpErrorText() const { 
  return &m_errorText[0]; 
};

/*----------------------------------------------------------------------*/

const DbpTypeUnicode* DbpError::dbpErrorTextUnicode() const {
  return (DbpTypeUnicode*)&m_errorText[0]; 
};

/*----------------------------------------------------------------------*/

const OmsObjectId& DbpError::dbpErrorOid() const { 
  return m_oid; 
};