/*!***************************************************************************

  module      : vco81CGenInspecCode.cpp

  -------------------------------------------------------------------------

  responsible : FerdiF

  special area: IDL to Code Generator   
  description : Generate Inspection Code out of generated XML file
				
  
  last changed: 2000-03-08  11:22
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/

#include <fstream.h>
#include <iostream.h>
#include "hco83ErrorHandling.h"
#define  INITGUID    1     /* Define needed CoClassIDs + IIDs in geo36.h */
#include "hco81CGenInspecCode.h"
#include "vsp001.h"
#include "geo36.h"

static const GUID g_null_guid = 
{ 0x0, 0x0, 0x0, { 0x0, 0x0, 0x0, 0xa0, 0x0, 0x0, 0x0, 0x0 } };

static GUID g_udt_guid = 
{ 0x0, 0x0, 0x0, { 0x0, 0x0, 0x0, 0xa0, 0x0, 0x0, 0x0, 0x0 } };


static cgg251DCList<co81CTypedef*,co81CGenInspecCode> g_struct_level;
static co81CTypedef *g_interface = NULL;
static co81CTypedef *g_coclass = NULL;
static short        g_func_no  = 0;
static short        g_param_no = 0;
static char         g_act_method[65];

static unsigned int hex2int( char x )
{
  unsigned int i = 0;
    if (('0' <= x) && ( x <= '9'))
      i = x - '0';
    if (('A' <= x) && ( x <= 'F'))
      i = (x - 'A') + 10;
    if (('a' <= x) && ( x <= 'f'))
      i = (x - 'a') + 10;
  return i;
};

static bool co81ExecuteOperator(co81CGenInspecCode &CGenInsepcCode,
                                cgg251dclIterator<co81CExprElem*,co81CGenInspecCode> &iter, 
                                co81CExprElem &ResultValue);
static bool co81GetParm(co81CGenInspecCode &CGenInsepcCode,
                        cgg251dclIterator<co81CExprElem*,co81CGenInspecCode> &iter, 
                        co81CExprElem &ResultValue);
static bool co81isOperator(co81CGenInspecCode &CGenInsepcCode,
                           XmlElementType_Enum_co82 type);
static bool co81IsInteger(co81CGenInspecCode &CGenInsepcCode,
                          co81CExprElem &token );
static bool co81Add(co81CGenInspecCode &CGenInsepcCode,
                    cgg251dclIterator<co81CExprElem*,co81CGenInspecCode> &iter, 
                    co81CExprElem &ResultValue);
static bool co81Mul(co81CGenInspecCode &CGenInsepcCode,
                    cgg251dclIterator<co81CExprElem*,co81CGenInspecCode> &iter, 
                    co81CExprElem &ResultValue);
static bool co81Div(co81CGenInspecCode &CGenInsepcCode,
                    cgg251dclIterator<co81CExprElem*,co81CGenInspecCode> &iter, 
                    co81CExprElem &ResultValue);
static bool co81Mod(co81CGenInspecCode &CGenInsepcCode,
                    cgg251dclIterator<co81CExprElem*,co81CGenInspecCode> &iter, 
                    co81CExprElem &ResultValue);
static bool co81Subtract(co81CGenInspecCode &CGenInsepcCode,
                         cgg251dclIterator<co81CExprElem*,co81CGenInspecCode> &iter, 
                         co81CExprElem &ResultValue);
static double co81GetFloat(co81CExprElem &t);
static int co81GetInteger(co81CExprElem &t );
static int co81GetInteger(char* s, int &i );
static void co81PutFloat(double i, co81CExprElem &r);
static void co81PutInteger(long i, co81CExprElem &r);

/*----------------------------------------------------------------------*/
/* struct co81CExprElem                                                 */
/*----------------------------------------------------------------------*/

co81CExprElem::co81CExprElem() {
  this->m_token_type  = elemUnknown_co82;
  this->m_token_value = NULL;
};

co81CExprElem::~co81CExprElem() {
  delete this->m_token_value;
};

const co81CExprElem & co81CExprElem::operator= (const co81CExprElem& right) {
  this->m_token_type  = right.m_token_type;
  if ( right.m_token_value ) {
    this->m_token_value = new char[ strlen(right.m_token_value) + 1 ];
    strcpy(this->m_token_value, right.m_token_value);
  }
  else {
    this->m_token_value  = NULL;
  }
  return *this;
};


/*----------------------------------------------------------------------*/
/* class co81CTypedef                                                   */
/*----------------------------------------------------------------------*/

co81CTypedef::co81CTypedef() {
  m_uuid.Data1    = 0;
  m_uuid.Data2    = 0;
  m_uuid.Data3    = 0;
  memset(m_uuid.Data4, '\0', sizeof(m_uuid.Data4));
  m_def_type      = elemUnknown_co82;
  m_name          = NULL;
  m_type          = NULL;
  m_value_type    = attUnknown_co82;
  m_value_string  = NULL;
  m_ptr           = 0;
  m_size          = -1;
  m_dimension     = -1;
  m_alignment     = -1;
  m_member_num    = 0;
  m_vt_type       = csp1_vt_empty;
};

/*----------------------------------------------------------------------*/

co81CTypedef::~co81CTypedef() {
  delete m_name;
  delete m_type;
  delete m_value_string;
};

/*----------------------------------------------------------------------*/

void  co81CTypedef::appendMember(co81CTypedef* m) {
  co81CTypedef findNode;
  m_member_list.push_back(m);
  m_member_num++;
};

/*----------------------------------------------------------------------*/

void  co81CTypedef::appendToken(co81CExprElem* t) {
  m_expr_list.push_back(t);
};

/*----------------------------------------------------------------------*/

void  co81CTypedef::setDimension(short d) {
  m_dimension = d;
};

/*----------------------------------------------------------------------*/

void co81CTypedef::setName(const XmlElementType_Enum_co82 vt, const char* n) {
  m_def_type  = vt;
  m_name      = new char[ strlen(n) + 1 ];
  strcpy(m_name,n);
  if ( vt == elemStruct_co82 ) 
    m_vt_type = csp1_vt_userdefined;
};

/*----------------------------------------------------------------------*/

void   co81CTypedef::setStructCharacteristics(co81CGenInspecCode &CGenInsepcCode) {
  co81CTypedef *curr;
  m_size = 0;
  m_alignment = 0;
  co81CTypedef findNode;
  
  for ( cgg251dclIterator<co81CTypedef*,co81CGenInspecCode> iter = m_member_list.begin(); iter; ++iter) {
    curr = *iter();
    
    short l_dim = 1;
    if ( curr->m_dimension > 0 )
      l_dim = curr->m_dimension;
    if ( curr->m_type ) {
      findNode.setName( elemUnknown_co82, curr->m_type );
      co81CTypedef* const *foundptr = CGenInsepcCode.findNode(findNode);
      if ( foundptr && (*foundptr)->m_dimension > 0 )
        l_dim = (*foundptr)->m_dimension * l_dim;
    }
    
    if ( curr->m_alignment > m_alignment ) m_alignment = curr->m_alignment;
    
    m_size = m_size + (curr->m_alignment - 1);
    m_size = m_size - ( m_size % curr->m_alignment );
    
    m_size = m_size + (l_dim * curr->m_size);
  }
  
};

/*----------------------------------------------------------------------*/

void co81CTypedef::setType(co81CGenInspecCode &CGenInsepcCode, 
                           const char* t, short size, short align) {
  if ( strchr( t, '*' )) {
    m_type = new char[ strlen(t) ];
    strncpy(m_type,t,strlen(t) - 1);
    m_type[strlen(t) - 1] = '\0';
    m_ptr = 1;
  }
  else {
    m_type = new char[ strlen(t) + 1 ];
    strcpy(m_type,t);
    m_ptr = 0;
  };

  m_size  = size;
  m_alignment = align;
  if ( !strcmp("long",t) || !strcmp("unsigned long",t) ) {
    if ( !strcmp("OmsTypeInt64", m_name) ) {
      m_size      = 8;
      m_alignment = 8;
      m_vt_type   = csp1_vt_i8;
    }
    if ( !strcmp("OmsTypeUInt64", m_name) ) {
      m_size      = 8;
      m_alignment = 8;
      m_vt_type   = csp1_vt_ui8;
    }
  }

  /* check OMS Datatypes                                */

  
  if (0 == memcmp (&GUID_LC_CHAR, &m_uuid, sizeof(m_uuid))) {                                                         
    m_vt_type = csp1_vt_lc_char;
  }
  if (0 == memcmp (&GUID_LC_WCHAR, &m_uuid, sizeof(m_uuid))) {                                                         
    m_vt_type = csp1_vt_lc_wchar;
  }
  if (0 == memcmp (&GUID_LC_BYTE, &m_uuid, sizeof(m_uuid))) {                                                         
    m_vt_type = csp1_vt_lc_byte;
  }
  if (0 == memcmp (&GUID_LC_BOOL, &m_uuid, sizeof(m_uuid))) {                                                         
    m_vt_type = csp1_vt_lc_bool;
  }
  if (0 == memcmp (&GUID_LC_INT1, &m_uuid, sizeof(m_uuid))) {                                                         
    m_vt_type = csp1_vt_lc_int1;
  }
  if (0 == memcmp (&GUID_LC_UINT1, &m_uuid, sizeof(m_uuid))) {                                                         
    m_vt_type = csp1_vt_lc_uint1;
  }
  if (0 == memcmp (&GUID_LC_ABAP_TAB_HANDLE, &m_uuid, sizeof(m_uuid))) {                                                         
    m_vt_type = csp1_vt_lc_abap_tab_handle;
  }
  if (0 == memcmp (&GUID_OMSTYPESTREAMHANDLE, &m_uuid, sizeof(m_uuid))) {                                                         
    m_vt_type = csp1_vt_lc_stream_handle;
  }
  if (0 == memcmp (&GUID_LC_PARAM_DESC, &m_uuid, sizeof(m_uuid))) {                                                         
    m_vt_type = csp1_vt_lc_kb71_param_desc;
  }

  /* check basis C++ Types */
  if ( csp1_vt_empty == m_vt_type ) {
    m_vt_type = CGenInsepcCode.checkBasisType( m_type );
    
    /* check if user defined Type */
    if ( csp1_vt_empty == m_vt_type ) {
      co81CTypedef findNode;
      findNode.setName( elemUnknown_co82, m_type );
      co81CTypedef* const *foundptr = CGenInsepcCode.findNode(findNode);
      if ( foundptr )
      {
        m_vt_type = (*foundptr)->m_vt_type;
        while ( foundptr && (csp1_vt_userdefined == m_vt_type ) ) /* PTS 1108040 UH 2000-10-13 */
        {
          if ( (*foundptr)->m_type )
          {
            findNode.setName( elemUnknown_co82, (*foundptr)->m_type );
            foundptr = CGenInsepcCode.findNode(findNode);
            m_vt_type = (*foundptr)->m_vt_type;
          }
          else
            foundptr = NULL;
        };
        if ( m_vt_type == csp1_vt_empty )
          m_vt_type = csp1_vt_userdefined;
        foundptr  = NULL;
      };
    };
  }
};

/*----------------------------------------------------------------------*/

void co81CTypedef::setValue(const XmlAttributeType_Enum_co82 vt, const char* vs) {
  m_value_type = vt;
  m_value_string = new char[ strlen(vs) + 1 ];
  strcpy(m_value_string,vs);
};

/*----------------------------------------------------------------------*/

void co81CTypedef::setUuid(const char* uuid_p) {
  int i;
  int j;
  char uuid[64];
  const char *pc = uuid_p;

  i = 0;
  for ( ; *pc != '\0'; pc++ ) {
    if ( *pc != ' ' ) {
      uuid[i] = *pc;
      i++;
    }
  }
  uuid[i] = 0;
  

  m_uuid.Data1 = 0;
  m_uuid.Data2 = 0;
  m_uuid.Data3 = 0;

  if ( 0 == strlen(uuid)) {
    m_uuid = g_null_guid;
  }
  else {
    for ( i=0; i < 8; i = i + 2 )
      m_uuid.Data1 = m_uuid.Data1 * 256 + hex2int( uuid[i] ) * 16 + hex2int( uuid[i+1] );
    for ( i=9; i < 13; i = i + 2 ) 
      m_uuid.Data2 = m_uuid.Data2 * 256 + hex2int( uuid[i] ) * 16 + hex2int( uuid[i+1] );
    for ( i=14; i < 18; i = i + 2 ) 
      m_uuid.Data3 = m_uuid.Data3 * 256 + hex2int( uuid[i] ) * 16 + hex2int( uuid[i+1] );
    for ( i=19, j=0; i < 23; i = i + 2, j++ ) 
      m_uuid.Data4[j] = hex2int( uuid[i] ) * 16 + hex2int( uuid[i+1] );
    for ( i=24, j=2; i < 36; i = i + 2, j++ )
      m_uuid.Data4[j] = hex2int( uuid[i] ) * 16 + hex2int( uuid[i+1] );
  };
};

/*----------------------------------------------------------------------*/

void  co81CTypedef::checkIfAbapOrStreamHandle(co81CGenInspecCode &CGenInsepcCode, 
                                              short *prm_type,
                                              short *prm_size,
                                              short *prm_elem_num,
                                              co81CTypedef * const ** row_def ) {
  if ( 2 == m_member_num ) { 
    cgg251dclIterator<co81CTypedef*,co81CGenInspecCode> iter = m_member_list.begin();    
    co81CTypedef *first  = *iter();
    if ( first->m_ptr ) {
      ++iter;
      co81CTypedef *second = *iter();
      if ( csp1_vt_userdefined == second->m_vt_type ) {
        co81CTypedef findNode;
        findNode.setName( elemUnknown_co82, second->m_type );
        co81CTypedef* const *foundptr = CGenInsepcCode.findNode(findNode); 
        findNode.setName( elemUnknown_co82, first->m_type );
        if ((0 == memcmp (&GUID_LC_ABAP_TAB_HANDLE, &(*foundptr)->m_uuid, sizeof(m_uuid))) ||    
          (0 == memcmp (&GUID_OMSTYPESTREAMHANDLE, &(*foundptr)->m_uuid, sizeof(m_uuid))) ||                                                        
          (0 == memcmp (&GUID_LC_PARAM_DESC, &(*foundptr)->m_uuid, sizeof(m_uuid)))) {              

          if (0 == memcmp (&GUID_LC_ABAP_TAB_HANDLE, &(*foundptr)->m_uuid, sizeof(m_uuid)))
            *prm_type = csp1_vt_lc_abap_tab_handle;
          else if (0 == memcmp (&GUID_OMSTYPESTREAMHANDLE, &(*foundptr)->m_uuid, sizeof(m_uuid)))
            *prm_type = csp1_vt_lc_stream_handle;
          else 
            *prm_type = csp1_vt_lc_kb71_param_desc;

          co81CTypedef* const *foundptr = CGenInsepcCode.findNode(findNode); 
          *row_def = foundptr; 
          *prm_size = (*foundptr)->m_size;
          if ( (*foundptr)->m_def_type == elemTypedef_co82 )
            *prm_elem_num = 1;
          else
            *prm_elem_num = (*foundptr)->getMemberNum(CGenInsepcCode);
        }
      };
    }
  }
};

/*----------------------------------------------------------------------*/

bool co81isOperator(co81CGenInspecCode &CGenInsepcCode,
                    XmlElementType_Enum_co82 type) {
  if ( elemFloat_co82 == type ) return false;
  if ( elemFixed_co82 == type ) return false;
  if ( elemIdent_co82 == type ) return false;
  return true;
}

/*----------------------------------------------------------------------*/

bool co81GetParm(co81CGenInspecCode &CGenInsepcCode,
                 cgg251dclIterator<co81CExprElem*,co81CGenInspecCode> &iter, 
                 co81CExprElem &ResultValue) {
  iter++;
  if ( iter ) {
    co81CExprElem* const token = *iter();
    if ( co81isOperator( CGenInsepcCode, token->m_token_type ) ) {
      if ( !co81ExecuteOperator( CGenInsepcCode, iter, ResultValue )) {
        return false;
      }
    }
    else {
      if ( token->m_token_type == elemIdent_co82 ) {
        co81CTypedef findNode;
        findNode.setName( elemUnknown_co82, token->m_token_value );
        co81CTypedef* const *foundptr = CGenInsepcCode.findNode(findNode); 
        if ( (*foundptr)->getValType() == attNumber_co82 ) {
          if ( strchr((*foundptr)->getValue(),'.' )) 
            ResultValue.m_token_type = elemFloat_co82;
          else
            ResultValue.m_token_type = elemFixed_co82;
          ResultValue.m_token_value = new char[ strlen((*foundptr)->getValue()) + 1 ];
          strcpy(ResultValue.m_token_value, (*foundptr)->getValue());
        }
        else
          return false;
      }
      else
        ResultValue = *token;
    }
    return true;
  }
  else 
    return false;
};

/*----------------------------------------------------------------------*/

bool co81IsInteger(co81CGenInspecCode &CGenInsepcCode,
                   co81CExprElem &token ) {
  if ( token.m_token_type == elemFixed_co82 ) 
    return true;
  if ( token.m_token_type == elemFloat_co82 ) 
    if ( strchr(token.m_token_value,'.' )) 
      return true;
  return false;
};

/*----------------------------------------------------------------------*/

static int co81GetInteger(char* s, int &i) {
  char *c = strchr(s,'+');
  if ( c ) 
    *c = ' ';
  c = strchr(s,'-');
  while ( c && *c+1 && (' ' == *c+1) ) {
    *c = ' ';
    c++;
    *c = '-';
  }
  return sscanf(s,"%d",&i);
};

/*----------------------------------------------------------------------*/

int co81GetInteger(co81CExprElem &t ) {
  int l;
  if ( co81GetInteger(t.m_token_value, l) )
    return l;
  else 
    return 0;
};

/*----------------------------------------------------------------------*/

void co81PutInteger(long l, co81CExprElem &r) {
  char buf[32];
  r.m_token_type = elemFixed_co82;
  sprintf(buf,"%d",l);
  r.m_token_value = new char[ strlen(buf) + 1 ];
  strcpy(r.m_token_value, buf);
};

/*----------------------------------------------------------------------*/

double co81GetFloat(co81CExprElem &t) {
  return 0;
};

/*----------------------------------------------------------------------*/

void co81PutFloat(double i, co81CExprElem &r) {
};

/*----------------------------------------------------------------------*/

bool co81Add(co81CGenInspecCode &CGenInsepcCode,
             cgg251dclIterator<co81CExprElem*,co81CGenInspecCode> &iter, 
             co81CExprElem &ResultValue) {
  co81CExprElem prm1;
  co81CExprElem prm2;
  
  if ( !co81GetParm( CGenInsepcCode, iter, prm1 ) ) return false;
  if ( !co81GetParm( CGenInsepcCode, iter, prm2 ) ) return false;
  
  if ( co81IsInteger( CGenInsepcCode, prm1 ) && co81IsInteger( CGenInsepcCode, prm2 ) )
    co81PutInteger( co81GetInteger(prm1) + co81GetInteger(prm2), ResultValue );
  else {
    co81PutFloat( co81GetFloat(prm1) + co81GetFloat(prm2), ResultValue );
  }
  return true;
}

/*----------------------------------------------------------------------*/

bool co81Mod(co81CGenInspecCode &CGenInsepcCode,
             cgg251dclIterator<co81CExprElem*,co81CGenInspecCode> &iter, 
             co81CExprElem &ResultValue) {
  co81CExprElem prm1;
  co81CExprElem prm2;
  
  if ( !co81GetParm( CGenInsepcCode, iter, prm1 ) ) return false;
  if ( !co81GetParm( CGenInsepcCode, iter, prm2 ) ) return false;
  
  if ( co81IsInteger( CGenInsepcCode, prm1 ) && co81IsInteger( CGenInsepcCode, prm2 ) )
    co81PutInteger( co81GetInteger(prm1) % co81GetInteger(prm2), ResultValue );
  else {
    return false;
  }
  return true;
}

/*----------------------------------------------------------------------*/

bool co81Div(co81CGenInspecCode &CGenInsepcCode,
             cgg251dclIterator<co81CExprElem*,co81CGenInspecCode> &iter, 
             co81CExprElem &ResultValue) {
  co81CExprElem prm1;
  co81CExprElem prm2;
  
  if ( !co81GetParm( CGenInsepcCode, iter, prm1 ) ) return false;
  if ( !co81GetParm( CGenInsepcCode, iter, prm2 ) ) return false;
  
  if ( co81IsInteger( CGenInsepcCode, prm1 ) && co81IsInteger( CGenInsepcCode, prm2 ) )
    co81PutInteger( co81GetInteger(prm1) / co81GetInteger(prm2), ResultValue );
  else {
    co81PutFloat( co81GetFloat(prm1) / co81GetFloat(prm2), ResultValue );
  }
  return true;
}

/*----------------------------------------------------------------------*/

bool co81Mul(co81CGenInspecCode &CGenInsepcCode,
             cgg251dclIterator<co81CExprElem*,co81CGenInspecCode> &iter, 
             co81CExprElem &ResultValue) {
  co81CExprElem prm1;
  co81CExprElem prm2;
  
  if ( !co81GetParm( CGenInsepcCode, iter, prm1 ) ) return false;
  if ( !co81GetParm( CGenInsepcCode, iter, prm2 ) ) return false;
  
  if ( co81IsInteger( CGenInsepcCode, prm1 ) && co81IsInteger( CGenInsepcCode, prm2 ) )
    co81PutInteger( co81GetInteger(prm1) * co81GetInteger(prm2), ResultValue );
  else {
    co81PutFloat( co81GetFloat(prm1) * co81GetFloat(prm2), ResultValue );
  }
  return true;
}

/*----------------------------------------------------------------------*/

bool co81Subtract(co81CGenInspecCode &CGenInsepcCode,
                  cgg251dclIterator<co81CExprElem*,co81CGenInspecCode> &iter, 
                  co81CExprElem &ResultValue) {
  co81CExprElem prm1;
  co81CExprElem prm2;
  
  if ( !co81GetParm( CGenInsepcCode, iter, prm1 ) ) return false;
  if ( !co81GetParm( CGenInsepcCode, iter, prm2 ) ) return false;
  
  if ( co81IsInteger( CGenInsepcCode, prm1 ) && co81IsInteger( CGenInsepcCode, prm2 ) )
    co81PutInteger( co81GetInteger(prm1) - co81GetInteger(prm2), ResultValue );
  else {
    co81PutFloat( co81GetFloat(prm1) - co81GetFloat(prm2), ResultValue );
  }
  return true;
}

/*----------------------------------------------------------------------*/

bool co81ExecuteOperator(co81CGenInspecCode &CGenInsepcCode,
                         cgg251dclIterator<co81CExprElem*,co81CGenInspecCode> &iter, 
                         co81CExprElem &ResultValue) {
  co81CExprElem* const token = *iter();
  switch( token->m_token_type ) {
  case elemSub_co82:
    return co81Subtract(CGenInsepcCode, iter, ResultValue);
  case elemAdd_co82:
    return co81Add(CGenInsepcCode, iter, ResultValue);
  case elemMul_co82:
    return co81Mul(CGenInsepcCode, iter, ResultValue);
  case elemDiv_co82:
    return co81Div(CGenInsepcCode, iter, ResultValue);
  case elemMod_co82:
    return co81Mod(CGenInsepcCode, iter, ResultValue);
    break;
  default:
    break;
  }
  return false;
}

/*----------------------------------------------------------------------*/

int co81CTypedef::calExprValue(co81CGenInspecCode &CGenInsepcCode) {
  int rc = 0;
  m_value_type = attNumber_co82;
  cgg251dclIterator<co81CExprElem*,co81CGenInspecCode> iter = m_expr_list.begin(); 
  if ( iter ) {
    co81CExprElem* const token = *iter();
    if ( co81isOperator( CGenInsepcCode, token->m_token_type ) ) {
      co81CExprElem l_ResultValue;
      if ( !co81ExecuteOperator( CGenInsepcCode, iter, l_ResultValue )) {
        switch ( this->getDefType() ){
        case elemConst_co82:
          co83ErrorMsg( C_CONST_EXPR_NOT_VALID_CO83, m_value_string, m_name, m_type );
          rc = C_CONST_EXPR_NOT_VALID_CO83;
          break;
        default:
          break;
        }
      }
      else {
        if ( m_value_string ) delete m_value_string;
        m_value_string = new char[ strlen(l_ResultValue.m_token_value) + 1 ];
        strcpy(m_value_string, l_ResultValue.m_token_value);
      }
    }
    else {
      iter++;
      if ( iter ) {
        switch ( this->getDefType() ){
        case elemConst_co82:
          co83ErrorMsg( C_CONST_EXPR_NOT_VALID_CO83, m_value_string, m_name, m_type  );
          rc = C_CONST_EXPR_NOT_VALID_CO83;
          break;
        default:
          break;
        }
      }
      else {
        if ( m_value_string ) delete m_value_string;
        m_value_string = new char[ strlen(token->m_token_value ) + 1 ];
        strcpy(m_value_string,token->m_token_value);
      }
    }
  }
  else {
    switch ( this->getDefType() ){
    case elemConst_co82:
      co83ErrorMsg( C_CONST_EXPR_NOT_VALID_CO83, m_value_string, m_name, m_type  );
      rc = C_CONST_EXPR_NOT_VALID_CO83;
      break;
    default:
      break;
    }
  }
  return rc;
}

/*----------------------------------------------------------------------*/

short co81CTypedef::getMemberNum(co81CGenInspecCode &CGenInsepcCode) {
  short l_member_num = m_member_num;
  if ( m_def_type == elemTypedef_co82 )
    l_member_num = 1;
  else {
    for ( cgg251dclIterator<co81CTypedef*,co81CGenInspecCode> iter = m_member_list.begin(); iter; iter++ ){    
      co81CTypedef const *curr  = *iter();
      co81CTypedef findNode;
      findNode.setName( elemUnknown_co82, curr->m_type );
      co81CTypedef* const *foundptr = CGenInsepcCode.findNode(findNode);
      while ( foundptr && (0 == (*foundptr)->m_member_num) ) {
        findNode.setName( elemUnknown_co82, (*foundptr)->m_type );
        foundptr = CGenInsepcCode.findNode(findNode);
      }   
      if ( foundptr && ( 1 < (*foundptr)->m_member_num ) ) {
        l_member_num = l_member_num + (*foundptr)->getMemberNum(CGenInsepcCode) - 1;
      }
    }
  }
  return l_member_num;
}

/*----------------------------------------------------------------------*/

void co81CTypedef::printSubParameter(co81CGenInspecCode &CGenInsepcCode, ofstream &fout, short &pos) {
  co81CTypedef *curr = NULL;
  co81CTypedef *base = NULL;
  short l_typ;
  short l_subtyp;
  short l_dim;
  short l_size;
  unsigned int   i;

  cgg251dclIterator<co81CTypedef*,co81CGenInspecCode> iter = m_member_list.begin();
  if ( this->m_def_type == elemTypedef_co82 ) {
    curr = this;
  }
  else {
    if ( iter ) curr = *iter();
  } 
  
  for ( ; curr; ) {
    g_param_no++;
    const GUID * const guid = curr->getUuid();
    const char * t = curr->getType();

    if (( 0 == strcmp(t,"hyper") ) ||
      ( 0 == strcmp(t,"unsigned hyper") ) ||
      ( 0 == strcmp(t,"signed hyper") ) ||
      ( 0 == strcmp(t,"char") ) ||
      ( 0 == strcmp(t,"boolean") ) ||
      ( 0 == strcmp(t,"bool") ) ||
      ( 0 == strcmp(t,"signed long") ) ||
      ( 0 == strcmp(t,"signed long int") ) ||
      ( 0 == strcmp(t,"unsigned long int") ) ||
      ( 0 == strcmp(t,"unsigned") ) ||
      ( 0 == strcmp(t,"long") ) ||
      ( 0 == strcmp(t,"unsigned long") ) ||
      ( 0 == strcmp(t,"signed long") ) ||
      ( 0 == strcmp(t,"long int") ) ) {
      char l_buf[128];
      strcpy(l_buf, "member:");
      strcat(l_buf,curr->getName());
      l_size = -1;
      co83WarnMsg( C_TYPE_NOT_PORTABLE_CO83, l_buf, t );
    }
    
    if ( curr->m_dimension > 0 )
      l_dim = curr->m_dimension;
    else
      l_dim = 0;

    co81CTypedef findNode;
    findNode.setName( elemUnknown_co82, curr->m_type );
    co81CTypedef* const *foundptr = CGenInsepcCode.findNode(findNode);
    while ( foundptr && (0 == (*foundptr)->m_member_num) &&
      ( csp1_vt_userdefined /*empty **FF** */ == (*foundptr)->m_vt_type) ) {
      if ( (*foundptr)->m_dimension > 0 ) {
        if ( l_dim )
          l_dim = l_dim * (*foundptr)->m_dimension;
        else
          l_dim = (*foundptr)->m_dimension;
      }
      findNode.setName( elemUnknown_co82, (*foundptr)->m_type );
      foundptr = CGenInsepcCode.findNode(findNode);
    }
    
    if ( foundptr && ( 1 < (*foundptr)->m_member_num ) ) {
      g_param_no--;
      
      short struct_align = (*foundptr)->getAlignment();

      /* align position                                                  */
      if (pos % struct_align != 0) {
        pos += (struct_align - pos % struct_align);
        co83WarnMsg( C_STRUCT_NOT_ALIGNED_CO83, g_act_method, this->getName(), (*foundptr)->getName() );    
      }

      (*foundptr)->printSubParameter( CGenInsepcCode, fout, pos );

      if (pos % struct_align != 0) {
        pos += (struct_align - pos % struct_align);
        co83WarnMsg( C_STRUCT_NOT_ALIGNED_CO83, g_act_method, this->getName(), (*foundptr)->getName() );    
      }
    }
    else {
      if ( foundptr && ( 1 == (*foundptr)->m_member_num ) ) {
        base = *((*foundptr)->m_member_list.begin())();
        l_size = curr->getSize();
        l_typ = base->m_vt_type;
        l_subtyp = 0;
        if ( 0 < base->m_dimension )
          l_dim = base->m_dimension;
        if ( base->m_type ) {
          findNode.setName( elemUnknown_co82, base->m_type );
          foundptr = CGenInsepcCode.findNode(findNode);
        }
        while ( foundptr ) {
          l_typ = (*foundptr)->m_vt_type;
          if ( (*foundptr)->m_dimension > 0 ) {
            if ( l_dim )
              l_dim = l_dim * (*foundptr)->m_dimension;
            else
              l_dim = (*foundptr)->m_dimension;
          }
          if ( (*foundptr)->m_type ) {
            findNode.setName( elemUnknown_co82, (*foundptr)->m_type );
            foundptr = CGenInsepcCode.findNode(findNode);
          }
          else {
            foundptr = NULL;
          }
        }
      }
      else {
        l_size = curr->getSize();
        if ( m_ptr ) {
          l_typ = csp1_vt_ptr;
          l_subtyp = curr->getVtType();
        }
        else {
          l_typ = curr->getVtType();
          l_subtyp = 0;
          if ( curr->getVtType() == csp1_vt_empty ) {
            if ( foundptr ) {
              l_typ = (*foundptr)->m_vt_type;
            }
          }
          if ( foundptr && ( 0 < (*foundptr)->m_dimension )) {
            if ( l_dim )
              l_dim = (*foundptr)->m_dimension * l_dim;
            else
              l_dim = (*foundptr)->m_dimension;
          }
          if ( l_dim > 1 ) 
            l_size = l_size * l_dim;
        }
      }
      if (( l_dim > 1 ) && ( l_typ != csp1_vt_ptr )) {
          l_subtyp = l_typ;
          l_typ = csp1_vt_carray;
      }
      
      /* align position                                                  */
      if (pos % curr->getAlignment() != 0) {
        pos += (curr->getAlignment() - pos % curr->getAlignment());
        co83WarnMsg( C_STRUCT_NOT_ALIGNED_CO83, g_act_method, this->getName(), curr->getName() );    
      }
      
      if ( !m_ptr ) {
        
        fout << "," << endl << "    (char*)\"MAN\"";

        /* within struct-parameters don't signal csp1_vt_lc_uint1 */
        if ( l_typ == csp1_vt_lc_uint1 ) l_typ = csp1_vt_lc_byte;
        
        //       if ( l_dim > l_size) DebugBreak();
        char cBuf[6];
        strcpy(cBuf,"I/O");
        for (i = 0; i < strlen(cBuf); i++) cBuf[i] = toupper( cBuf[i] ); 
        fout << ", (char*)\"" << cBuf << "\""
          << ", " << l_typ 
          << ", " << l_subtyp
          << ", " << l_size
          << ", " << l_dim 
          << ", " << pos 
          << ", 0,";
        
         // fout.setf(ios::hex); doesn't work on sun
        char cData[9];
        sprintf(cData,"%x",guid->Data1);
        fout << " { 0x" << cData;
        sprintf(cData,"%x",guid->Data2);
        fout << ", 0x" << cData;
        sprintf(cData,"%x",guid->Data3);
        fout << ", 0x" << cData << ", {";
        //fout << " { 0x" << guid->Data1 << ", 0x" << guid->Data2 << ", 0x" 
        //  << guid->Data3 << ", {";
        
        for (i=0; i < 7; i++ ) {
          sprintf(cData,"%x",(unsigned short)guid->Data4[i]);
          fout << " 0x" << cData << ",";
          //fout << " 0x" << (unsigned short)guid->Data4[i] << ",";
        };
        //      fout << " 0x" << (unsigned short)guid->Data4[7] << " }}, (char*)\"" << curr->getName() << "\"";
        sprintf(cData,"%x",(unsigned short)guid->Data4[7]);
        fout << " 0x" << cData << " }}, (char*)\"\"";
        //fout << " 0x" << (unsigned short)guid->Data4[7] << " }}, (char*)\"\"";
        // fout.unsetf(ios::hex);
        
      }
      else {
        co83ErrorMsg( C_POINTER_NOT_ALLOWED_CO83, g_act_method, this->getName(), curr->getName() );
        return;
      }
      pos = pos + l_size;
    }
    curr = NULL;
    if ( this->m_def_type != elemTypedef_co82 ) {
      iter++;
      if ( iter ) curr = *iter();
    } 
  };
};

/*----------------------------------------------------------------------*/
/* class co81CGenInspecCode                                             */
/*----------------------------------------------------------------------*/

co81CGenInspecCode::co81CGenInspecCode() {
  m_act_pType = NULL;
  m_expr_flag = false;
};

/*----------------------------------------------------------------------*/

co81CGenInspecCode::~co81CGenInspecCode() {
  delete g_interface;
  delete g_coclass;

  m_cpp_out.close();
  m_h_out.close();
};

/*----------------------------------------------------------------------*/

int co81CGenInspecCode::Init(char*  cFilename) {
  strcpy(m_filename, cFilename);
  char* pc = strrchr(m_filename,'.');
  if ( pc ) 
    *pc = '\0';
#if defined(_WIN32)
  pc = strrchr(m_filename,'\\'); 
#else
  pc = strrchr(m_filename,'/'); 
#endif
  if ( pc )
    strcpy(m_name, pc + 1);
  else
    strcpy(m_name,&m_filename[0]);


  m_cpp_out.open(strcat(m_filename,"Ispc.cpp"));
  if ( m_cpp_out.fail() )
    return 0;

  pc = strrchr(m_filename,'.'); 
  if ( pc ) 
    *(pc) = '\0';
  m_h_out.open(strcat(m_filename,".h"));
  if ( m_h_out.fail() )
    return 0;

  pc = strrchr(m_filename,'\\'); 
  if ( pc ) 
    *(pc + 1) = '\0';
  m_cpp_out                                                  << endl;
  m_cpp_out << "#include \"" << &m_name[0] << "Ispc.h\""     << endl;
  m_cpp_out                                                  << endl;
  m_cpp_out << "const t_FuncBufDesc CO_FunctionDesc[] ="     << endl;
  m_cpp_out << "{"                                           << endl;

  m_h_out                                                    << endl;
  m_h_out << "#include \"lcbasetypes.h\""                    << endl;
  m_h_out                                                    << endl;
  m_h_out << endl;
  m_h_out << "typedef void ( __stdcall *CO_INTROSPECTFUNCTION) ("<< endl;
  m_h_out << "  unsigned short       ParmIdx," << endl;
  m_h_out << "  char                 **lplpMode," << endl;
  m_h_out << "  char                 **lplpIOtype," << endl;
  m_h_out << "  short int            *lpDatatype," << endl;
  m_h_out << "  short int            *lpSubdatatype," << endl;
  m_h_out << "  short int            *lpLength,"<< endl;
  m_h_out << "  short int            *lpDim,"<< endl;
  m_h_out << "  long                 *lpOffset,"<< endl;
  m_h_out << "  unsigned char        *lpFrac,"<< endl;
  m_h_out << "  LPGUID               *lplpGUID,"<< endl;
  m_h_out << "  char                 **lplpParmName);"<< endl;
  m_h_out << ""<< endl;
  m_h_out << "typedef struct"<< endl;
  m_h_out << "{"<< endl;
  m_h_out << "  unsigned short        FuncIdx;"<< endl;
  m_h_out << "  char                  *lpMethod;"<< endl;
  m_h_out << "  unsigned short        ParamCount;"<< endl;
  m_h_out << "  CO_INTROSPECTFUNCTION lpFunction;"<< endl;
  m_h_out << "} t_FuncBufDesc;"<< endl;
  m_h_out << ""<< endl;
  m_h_out << "typedef struct"<< endl;
  m_h_out << "{"<< endl;
  m_h_out << "  char                 *Mode;"<< endl;
  m_h_out << "  char                 *IOtype;"<< endl;
  m_h_out << "  short int            Datatype;"<< endl;
  m_h_out << "  short int            Subdatatype;"<< endl;
  m_h_out << "  short int            Length;"<< endl;
  m_h_out << "  short int            Dim;"<< endl;
  m_h_out << "  long  int            Offset;"<< endl;
  m_h_out << "  unsigned char        Frac;"<< endl;
  m_h_out << "  GUID                 guid;"<< endl;
  m_h_out << "  char                 *lpParmName;"<< endl;
  m_h_out << "} t_ParmsDesc;"<< endl;

  return 1;
};

/*----------------------------------------------------------------------*/

short co81CGenInspecCode::checkBasisType(const char* type) {
  
  if ( !strcmp(type, "hyper")) return csp1_vt_i8;     
  if ( !strcmp(type, "signed hyper")) return csp1_vt_i8;     
  if ( !strcmp(type, "unsigned hyper")) return csp1_vt_ui8;     
  if ( !strcmp(type, "long long")) return csp1_vt_i8; 
  
  if ( !strcmp(type, "short")) return csp1_vt_i2;
  if ( !strcmp(type, "short int")) return csp1_vt_i2;
  if ( !strcmp(type, "signed short")) return csp1_vt_i2;
  if ( !strcmp(type, "signed short int")) return csp1_vt_i2;
  if ( !strcmp(type, "unsigned short")) return csp1_vt_ui2;
  if ( !strcmp(type, "unsigned short int")) return csp1_vt_ui2;
  
  if ( !strcmp(type, "int")) return csp1_vt_int;
  if ( !strcmp(type, "long int")) return csp1_vt_int;
  if ( !strcmp(type, "signed int")) return csp1_vt_int;
  if ( !strcmp(type, "signed long int")) return csp1_vt_int;
  if ( !strcmp(type, "unsigned int")) return csp1_vt_uint;
  if ( !strcmp(type, "unsigned long int")) return csp1_vt_uint;
  if ( !strcmp(type, "unsigned")) return csp1_vt_uint;
  if ( !strcmp(type, "signed")) return csp1_vt_int;
  
  if ( !strcmp(type, "char")) return csp1_vt_i1;
  if ( !strcmp(type, "byte")) return csp1_vt_ui1;
  if ( !strcmp(type, "boolean")) return csp1_vt_i1;
  if ( !strcmp(type, "bool")) return csp1_vt_i4;
  if ( !strcmp(type, "signed char")) return csp1_vt_i1;
  if ( !strcmp(type, "unsigned char")) return csp1_vt_ui1;
  if ( !strcmp(type, "wchar_t")) return csp1_vt_i2;
  
  if ( !strcmp(type, "long")) return csp1_vt_i4;
  if ( !strcmp(type, "signed long")) return csp1_vt_i4;
  if ( !strcmp(type, "unsigned long")) return csp1_vt_ui4;

  if ( !strcmp(type, "float")) return csp1_vt_r4;
  if ( !strcmp(type, "double")) return csp1_vt_r8;

  return csp1_vt_empty;
};

/*----------------------------------------------------------------------*/

bool co81CGenInspecCode::checkFirstParam(const char *pname, const char *puuid, 
                                         const char *ptype, const char *pdirection) {
  if ( pname && puuid && ptype && pdirection ) {

    co81CTypedef findNode;
    findNode.setName( elemUnknown_co82, ptype );
    co81CTypedef* const *foundptr = (m_index.FindNode(&findNode))->GetKey();
    
    /* if first method */
    if ( g_udt_guid == g_null_guid ) {
      if ( !( foundptr && (csp1_vt_userdefined == (*foundptr)->getVtType()))) {
        co83ErrorMsg( C_P1_NO_USERDEFINED_TYPE_CO83, g_act_method, (*foundptr)->getType() );
        return true;  /* error */
      }
      if ( !strcmp("in", pdirection) ){
        co83ErrorMsg( C_P1_IO_TYPE_INPUT_CO83, g_act_method );
        return true;  /* error */
      }
      g_udt_guid = *(*foundptr)->getUuid();
    }
    else {
      if ( !strcmp("out", pdirection) ){
        co83ErrorMsg( C_P1_IO_TYPE_OUTPUT_CO83, g_act_method );
        return true;  /* error */
      }
      if ( g_udt_guid != *(*foundptr)->getUuid() ) {
        co83ErrorMsg( C_P1_WRONG_GUID_CO83, g_act_method );
      }     
    };
  }
  else {
    if ( !puuid)
      co83ErrorMsg( C_P1_NO_GUID_CO83, g_act_method );
    else
      co83ErrorMsg( C_P1_TYPE_SPEC_INCOMPLETE_CO83, g_act_method );
    return true;  /* error */
  };

  return false; /* no error */
};

/*----------------------------------------------------------------------*/

short co81CGenInspecCode::getTypeDimension(const char* d){
  short dim = 0;
  if ( d && strlen(d) ) {
    char  buf[64];
    strcpy(buf,d);
    bool found = false;
    co81CTypedef findNode;
    
    do {
      findNode.setName( elemUnknown_co82, buf );
      int l;
      if ( co81GetInteger(buf, l) ) {
        dim   = (short)l;
        found = true;
      }
      else {
        co81CTypedef* const *foundptr = (m_index.FindNode(&findNode))->GetKey();
        if ( foundptr && (*foundptr)->getValue() )
          strcpy(buf,(*foundptr)->getValue());
        else
          return (-1);
      };
    } while ( !found );
  }
  return dim;
};

/*----------------------------------------------------------------------*/

int  co81CGenInspecCode::getTypeCharacteristics(const char* type, 
                                                const char* usedby, 
                                                const char* username,
                                                short& size, 
                                                short &align ) {
  char   t[64];

  strcpy (t,type);

  /* trim trailing '*' char */
  char* pc = strrchr(t,'*');
  if ( pc ) *pc = '\0';

  if (( 0 == strcmp(t,"unsigned char") ) ||
    ( 0 == strcmp(t,"signed char") ) ||
    ( 0 == strcmp(t,"byte") ) ||
    ( 0 == strcmp(t,"boolean") ) ||
    ( 0 == strcmp(t,"bool") ) ||
    ( 0 == strcmp(t,"char") )) {
    size  = 1;
    align = 1;
    return 1;
  };
  
  if (( 0 == strcmp(t,"unsigned short") ) ||
    ( 0 == strcmp(t,"signed short") ) ||
    ( 0 == strcmp(t,"signed short int") ) ||
    ( 0 == strcmp(t,"unsigned short int") ) ||
    ( 0 == strcmp(t,"short int") ) ||
    ( 0 == strcmp(t,"short int") ) ||
    ( 0 == strcmp(t,"short") ) ||
    ( 0 == strcmp(t,"wchar_t") ) ) {
    size  = 2;
    align = 2;
    return 1;
  };
  
  if (( 0 == strcmp(t,"unsigned int") ) ||
    ( 0 == strcmp(t,"signed int") ) ||
    ( 0 == strcmp(t,"signed") ) ||
    ( 0 == strcmp(t,"unsigned") ) ||
    ( 0 == strcmp(t,"float") ) ||
    ( 0 == strcmp(t,"signed long") ) ||
    ( 0 == strcmp(t,"signed long int") ) ||
    ( 0 == strcmp(t,"unsigned long int") ) ||
    ( 0 == strcmp(t,"long") ) ||
    ( 0 == strcmp(t,"unsigned long") ) ||
    ( 0 == strcmp(t,"signed long") ) ||
    ( 0 == strcmp(t,"long int") ) ||
    ( 0 == strcmp(t,"int") ) ) {
    size  = sizeof(int);
    align = sizeof(int);
    return 1;
  };
  
  if (( 0 == strcmp(t,"hyper") ) ||
    ( 0 == strcmp(t,"signed hyper") ) ||
    ( 0 == strcmp(t,"unsigned hyper") ) ||
    ( 0 == strcmp(t,"double") ) ||
    ( 0 == strcmp(t,"signed long long") ) ||
    ( 0 == strcmp(t,"unsigned long long") ) ||
    ( 0 == strcmp(t,"long long") ) ) {
    size  = 8;
    align = 8;
    return 1;
  };

  co81CTypedef findNode;
  findNode.setName( elemUnknown_co82, t );

  co81CTypedef* const *foundptr = (m_index.FindNode(&findNode))->GetKey();
  if ( foundptr ) {
    size  = (*foundptr)->getSize();
    align = (*foundptr)->getAlignment();
    return 1;
  }
  else {
    char buf[128];
    strcpy(buf, usedby);
    strcat(buf,":");
    strcat(buf,username);
    size = -1;
    align = -1;
    co83ErrorMsg( C_TYPE_UNKNOWN_CO83, buf, t );
  };
  return 0;
};

/*----------------------------------------------------------------------*/

void co81CGenInspecCode::printGuidV1( ofstream& out, co81CTypedef& type ) {
  int i;
  
  const GUID * const guid = type.getUuid();

  //out.setf(ios::hex); doesn't work on sun
  char cData[9];
  sprintf(cData,"%x",guid->Data1);
  out << " { 0x" << cData;
  sprintf(cData,"%x",guid->Data2);
  out << ", 0x" << cData;
  sprintf(cData,"%x",guid->Data3);
  out << ", 0x" << cData << "," << endl << " {";

  //out << "{ 0x" << guid->Data1 << ", 0x" << guid->Data2 << ", 0x" 
  //  << guid->Data3 << "," << endl << "{";

  for ( i=0; i < 7; i++ ) {
    sprintf(cData,"%x",(unsigned short)guid->Data4[i]);
    out << " 0x" << cData << ",";
    //out << " 0x" << (unsigned short)guid->Data4[i] << ",";
  }
  sprintf(cData,"%x",(unsigned short)guid->Data4[7]);
  out << " 0x" << cData << " }};" << endl;
  //out << " 0x" << (unsigned short)guid->Data4[8] << " }};" << endl;
  //out.unsetf(ios::hex);
};

/*----------------------------------------------------------------------*/

void co81CGenInspecCode::startElement(int &error, DboType_Enum_co82 &dboType, 
                                      const char *name, const char **atts) {
  co81CTypedef* pType;
  XmlElementType_Enum_co82 elemtype = co82getElementType(name);
  unsigned int  i;
  char buf[64];
  const char *name_idx = NULL;
  const char *type_idx = NULL;
  tgg00_BasisError basis_err = e_ok;

  switch (elemtype) {
  case elemConst_co82: 
    m_act_pType = new co81CTypedef;
    for ( i = 0; atts[i]; i=i+2) {
      XmlAttributeType_Enum_co82 attrtype = co82getAttributeType(atts[i]);
      switch ( attrtype ) {
      case attName_co82:
        name_idx = atts[i+1];
        m_act_pType->setName( elemtype, atts[i+1] ); 
        break;
      case attType_co82:
        short  size;
        short  align;
        getTypeCharacteristics( atts[i+1], "definition of CONST", name_idx, size, align );
        m_act_pType->setType( *this, atts[i+1], size, align ); 
        break;
      case attString_co82:
      case attNumber_co82:
      case attChar_co82:
      case attExpression_co82:
        m_act_pType->setValue( attrtype, atts[i+1] ); 
        break;
      case attPack_co82:
      case attUuid_co82:
      case attDimension_co82:
      case attDirection_co82:
      case attUnknown_co82:
      default:
        break;
      }
    }
    break;
  case elemTypedef_co82:
    pType = new co81CTypedef;
    for ( i = 0; atts[i]; i=i+2) {
      XmlAttributeType_Enum_co82 attrtype = co82getAttributeType(atts[i]);
      switch ( attrtype ) {
      case attName_co82: 
        name_idx = atts[i+1];
        pType->setName( elemtype, atts[i+1] ); break;
      case attUuid_co82:
        pType->setUuid( atts[i+1] ); break;
      case attType_co82:
        short  size;
        short  align;
        getTypeCharacteristics( atts[i+1], "TYPEDEF", name_idx, size, align );
        pType->setType( *this, atts[i+1], size, align ); break;
      case attDimension_co82:
        pType->setDimension( getTypeDimension( atts[i+1] ) ); break;
      case attString_co82:
      case attNumber_co82:
      case attChar_co82:
      case attPack_co82:
      case attDirection_co82:
      case attUnknown_co82:
      default:
        break;
      }
    }
    basis_err = m_index.Insert(pType);
    if ( basis_err ) {
      co83ErrorMsg( C_DUP_DEFINITION_CO83, pType->getName() );
      if ( !error )
        error = - C_DUP_DEFINITION_CO83;
    }
    break;
  case elemStruct_co82:
    pType = new co81CTypedef;
    g_struct_level.push_front(pType);    
    for ( i = 0; atts[i]; i=i+2) {

      XmlAttributeType_Enum_co82 attrtype = co82getAttributeType(atts[i]);
      switch ( attrtype ) {
      case attName_co82: 
        pType->setName( elemtype, atts[i+1] ); break;
      case attUuid_co82:
        pType->setUuid( atts[i+1] ); break;
      case attType_co82:
      case attDimension_co82:
      case attString_co82:
      case attNumber_co82:
      case attChar_co82:
      case attPack_co82:
      case attDirection_co82:
      case attUnknown_co82:
      default:
        break;
      }
    }
    basis_err = m_index.Insert(pType);
    if ( basis_err ) {
      co83ErrorMsg( C_DUP_DEFINITION_CO83, pType->getName() );
      if ( !error )
        error = - C_DUP_DEFINITION_CO83;
    }
    break;
  case elemMember_co82:
    pType = new co81CTypedef;
    for ( i = 0; atts[i]; i=i+2) {
      short dim;
      XmlAttributeType_Enum_co82 attrtype = co82getAttributeType(atts[i]);
      switch ( attrtype ) {
      case attName_co82: 
        name_idx = atts[i+1];
        pType->setName( elemtype, atts[i+1] ); break;
      case attUuid_co82:
        pType->setUuid( atts[i+1] ); break;
      case attType_co82:
        type_idx = atts[i+1];
        short  size;
        short  align;
        getTypeCharacteristics( atts[i+1], "STRUCT member", name_idx, size, align );
        pType->setType( *this, atts[i+1], size, align ); break;
      case attDimension_co82:
        dim = getTypeDimension( atts[i+1] );
        if ( dim < 0 ) {
          co83ErrorMsg( C_DIMENSION_TYPE_UNDEF_CO83, atts[i+1], name_idx, type_idx );
          if ( !error )
            error = - C_PTR_OF_PTR_UNSUPPORTED_CO83;
        }
        else
          pType->setDimension( dim ); 
        break;
      case attString_co82:
      case attNumber_co82:
      case attChar_co82:
      case attPack_co82:
      case attDirection_co82:
      case attUnknown_co82:
      default:
        break;
      }
    }
    if ( !g_struct_level.empty() )
      (g_struct_level.front())->appendMember(pType);
    break;
  case elemInterface_co82:
    g_interface = new co81CTypedef;
    for ( i = 0; atts[i]; i=i+2) {
      XmlAttributeType_Enum_co82 attrtype = co82getAttributeType(atts[i]);
      switch ( attrtype ) {
      case attName_co82: 
        g_interface->setName( elemtype, atts[i+1] ); break;
      case attUuid_co82:
        g_interface->setUuid( atts[i+1] ); break;
      case attType_co82:
      case attDimension_co82:
      case attString_co82:
      case attNumber_co82:
      case attChar_co82:
      case attPack_co82:
      case attDirection_co82:
      case attUnknown_co82:
      default:
        break;
      }
    }

    m_h_out                                                    << endl;
    m_h_out << "static const   IID   IID_Intro ="              << endl;
    printGuidV1( m_h_out, *g_interface );
    break;
  case elemCoClass_co82:
    g_coclass = new co81CTypedef;
    for ( i = 0; atts[i]; i=i+2) {
      XmlAttributeType_Enum_co82 attrtype = co82getAttributeType(atts[i]);
      switch ( attrtype ) {
      case attName_co82: 
        g_coclass->setName( elemtype, atts[i+1] ); break;
      case attUuid_co82:
        g_coclass->setUuid( atts[i+1] ); 

        break;
      case attType_co82:
      case attDimension_co82:
      case attString_co82:
      case attNumber_co82:
      case attChar_co82:
      case attPack_co82:
      case attDirection_co82:
      case attUnknown_co82:
      default:
        break;
      }
    }
    m_h_out                                                    << endl;
    m_h_out << "static const   CLSID CLSID_Intro ="            << endl;
    printGuidV1( m_h_out, *g_coclass );

    m_h_out                                                    << endl;
    m_h_out << "#define PROCEDURE_NUM " << g_func_no           << endl;
    m_h_out << "#define VERSION_NUMBER 1"                      << endl;

    m_cpp_out << endl <<"};"                                   << endl << endl;                                                   
    m_cpp_out << "extern \"C\" void __stdcall"                                               << endl;
    m_cpp_out << "  Co_IntrospectObject (LPCLSID        *lpClsId,"                           << endl;
    m_cpp_out << "                       LPIID          *lpIId,"                             << endl;
    m_cpp_out << "                       unsigned short *lpNum,"                             << endl;
    m_cpp_out << "                       long           *lpVersion ) {"                      << endl;
    m_cpp_out << "  *lpClsId        = (LPCLSID)&CLSID_Intro;"                                << endl;
    m_cpp_out << "  *lpIId          = (LPIID)&IID_Intro;"                                    << endl;
    m_cpp_out << "  *lpNum          = PROCEDURE_NUM;"                                        << endl;
    m_cpp_out << "  *lpVersion      = VERSION_NUMBER;"                                       << endl;
    m_cpp_out << "};"                                                                << endl << endl;

    m_cpp_out << "enum Co_AccessType {"                                                      << endl;
    m_cpp_out << "  get_by_name,"                                                            << endl;
    m_cpp_out << "  get_by_index"                                                            << endl;
    m_cpp_out << "};"                                                                << endl << endl;

    m_cpp_out << "extern \"C\" void __stdcall"                                               << endl;
    m_cpp_out << "  Co_IntrospectMethod (Co_AccessType          accessType,"                 << endl;
    m_cpp_out << "                       char                 **lplpMethod,"                 << endl;
    m_cpp_out << "                       unsigned short         Num,"                        << endl;
    m_cpp_out << "                       unsigned short        *lpIdx,"                      << endl;
    m_cpp_out << "                       unsigned short        *lpParamCount,"               << endl;
    m_cpp_out << "                       CO_INTROSPECTFUNCTION *lplpIntrospecFunc ) {"       << endl;
    m_cpp_out << "  *lplpIntrospecFunc = NULL;"                                              << endl;
    m_cpp_out << "  if (accessType == get_by_name) {"                                        << endl;
    m_cpp_out << "    if ( **lplpMethod == '\\0' ) {"                                        << endl;
    m_cpp_out << "      *lpIdx             = CO_FunctionDesc[0].FuncIdx;"                    << endl;
    m_cpp_out << "      *lpParamCount      = CO_FunctionDesc[0].ParamCount;"                 << endl;
    m_cpp_out << "      *lplpIntrospecFunc = CO_FunctionDesc[0].lpFunction;"                 << endl;
    m_cpp_out << "      *lplpMethod        = CO_FunctionDesc[0].lpMethod;"                   << endl;
    m_cpp_out << "    }"                                                                     << endl;
    m_cpp_out << "    else"                                                                  << endl;
    m_cpp_out << "      for ( int i=0; i < PROCEDURE_NUM; i++ )"                             << endl;
    m_cpp_out << "        if ( 0 == strcmp(*lplpMethod, CO_FunctionDesc[i].lpMethod )) {"    << endl;
    m_cpp_out << "          *lpIdx             = CO_FunctionDesc[i].FuncIdx;"                << endl;
    m_cpp_out << "          *lpParamCount      = CO_FunctionDesc[i].ParamCount;"             << endl;
    m_cpp_out << "          *lplpIntrospecFunc = CO_FunctionDesc[i].lpFunction;"             << endl;
    m_cpp_out << "        }"                                                                 << endl;
    m_cpp_out << "  }"                                                                       << endl;
    m_cpp_out << "  else {"                                                                  << endl;
    m_cpp_out << "    if (( Num <= PROCEDURE_NUM ) && ( Num >= 0 )) {"                       << endl;
    m_cpp_out << "      *lpIdx             = CO_FunctionDesc[Num].FuncIdx;"                  << endl;
    m_cpp_out << "      *lpParamCount      = CO_FunctionDesc[Num].ParamCount;"               << endl;
    m_cpp_out << "      *lplpIntrospecFunc = CO_FunctionDesc[Num].lpFunction;"               << endl;
    m_cpp_out << "      *lplpMethod        = CO_FunctionDesc[Num].lpMethod;"                 << endl;
    m_cpp_out << "    }"                                                                     << endl;
    m_cpp_out << "  }"                                                                       << endl;
    m_cpp_out << "};"                                                                        << endl;
    
    break;
  case elemIdl_co82:
    g_func_no++;
    g_param_no = 0;
    
    m_h_out                                                    << endl;
    m_h_out <<"static void __stdcall co_IntrospectFunction" << g_func_no << " (" << endl;
    m_h_out <<"  unsigned short       ParmIdx,"                << endl;
    m_h_out <<"  char                 **lplpMode,"             << endl;
    m_h_out <<"  char                 **lplpIOtype,"           << endl;
    m_h_out <<"  short int            *lpDatatype,"            << endl;
    m_h_out <<"  short int            *lpSubdatatype,"         << endl;
    m_h_out <<"  short int            *lpLength,"              << endl;
    m_h_out <<"  short int            *lpDim,"                 << endl;
    m_h_out <<"  long                 *lpOffset,"              << endl;
    m_h_out <<"  unsigned char        *lpFrac,"                << endl;
    m_h_out <<"  LPGUID               *lplpGUID,"              << endl;
    m_h_out <<"  char                 **lplpParmName)"         << endl;
    m_h_out <<"{"                                              << endl;
    
    if ( g_func_no > 1 )
      m_cpp_out << "," << endl << "  ";
    else 
      m_cpp_out << endl << "  ";
    m_cpp_out << g_func_no + 2 << ", (char*)\"" << atts[1] << "\", ";
    strcpy( g_act_method, atts[1] );
    break;
  case elemParam_co82:
    {
      const char *name_idx      = NULL;
      const char *direction_idx = NULL;
      const char *uuid_idx      = NULL;
      const char *type_idx      = NULL;
      const char *dimension_idx = NULL;
      short is_ptr;
      char  generatedName[11];
      
      g_param_no++;
      sprintf(generatedName,"_idl_%5.5d",g_param_no );
      
      if ( 1 == g_param_no ) {
        m_h_out <<"  static const t_ParmsDesc CO_ParmDesc[] ="     << endl; 
        m_h_out <<"  {"                                            << endl;
      };

      for ( i = 0; atts[i]; i=i+2) {
        XmlAttributeType_Enum_co82 attrtype = co82getAttributeType(atts[i]);
        switch ( attrtype ) {
        case attName_co82:
          if ( 0 == strlen(atts[i+1] )) 
            name_idx = &generatedName[0];
          else
            name_idx = atts[i+1]; 
          break;
        case attUuid_co82:
          uuid_idx = atts[i+1]; break;
        case attType_co82:
          type_idx = atts[i+1]; break;
        case attDimension_co82:
          dimension_idx = atts[i+1]; break;
        case attString_co82:
        case attNumber_co82:
        case attChar_co82:
        case attPack_co82:
        case attDirection_co82:
          direction_idx = atts[i+1]; break;
        case attUnknown_co82:
        default:
          break;
        }
      }      

      if ( strrchr( type_idx, '*' )) {
        strncpy(buf, type_idx, strlen(type_idx) - 1);
        buf[strlen(type_idx) - 1] = '\0';
        is_ptr = 1;
        if ( strrchr( buf, '*' )) {
          co83ErrorMsg( C_PTR_OF_PTR_UNSUPPORTED_CO83, g_act_method, name_idx, type_idx );
          if ( !error )
            error = - C_PTR_OF_PTR_UNSUPPORTED_CO83;
          break;
        }
      }
      else {
        strcpy(buf, type_idx);
        is_ptr = 0;
      };
      
      if (dboType != dbo_method_co82 || checkFirstParam(name_idx, uuid_idx, buf, direction_idx)) {
        co81CTypedef findNode;
        findNode.setName( elemUnknown_co82, buf );
        co81CTypedef* const *foundptr = (m_index.FindNode(&findNode))->GetKey();
        while ( foundptr && 
          elemTypedef_co82  == (*foundptr)->getDefType() &&
          csp1_vt_userdefined == (*foundptr)->getVtType() &&
          (*foundptr)->getType() ) {
            findNode.setName( elemUnknown_co82, (*foundptr)->getType() );
            foundptr = (m_index.FindNode(&findNode))->GetKey();
        }

        if ( 1 == g_param_no )
          m_h_out << "    (char*)\"MAN\"";
        else
          m_h_out << "," << endl << "    (char*)\"MAN\"";

        if ( direction_idx && strlen(direction_idx)) {
          char cBuf[6];
          strcpy(cBuf, direction_idx);
          for (i = 0; i < strlen(cBuf); i++) {
            cBuf[i] = toupper( cBuf[i] ); 
            if ( ' ' == cBuf[i] ) 
              cBuf[i] = '\0';
          }

          m_h_out << ", (char*)\"" << cBuf << "\"";
        }
        else {
          co83WarnMsg( C_NO_DIR_SPECIFIC_CO83, g_act_method, name_idx );
          m_h_out << ", (char*)\"I/O\"";
        };
        
        short l_typ; 
        short l_subtyp;
        short l_size;
        short l_dim       = getTypeDimension( dimension_idx );
        
        /* parameter is a user defined type ?  */
        if ( foundptr ) {
          l_typ       = csp1_vt_ptr;
          l_subtyp    = csp1_vt_userdefined;
          l_size      = (*foundptr)->getSize();
/*        if ( (*foundptr)->getDimension() > 0 ) {
            if ( l_dim > 0 ) {
              co83ErrorMsg( C_UNSUPPORTED_VT_CARRAY1_CO83, g_act_method, type_idx);
            }
          } 
*/
          short l_elem_num  = (*foundptr)->getMemberNum(*this);
          
          const GUID * const  guid = (*foundptr)->getUuid();
          /* parameter is a struct ? */
          if (elemStruct_co82  == (*foundptr)->getDefType() ) {
            if ( is_ptr || (*foundptr)->getPtr() ) {
              (*foundptr)->checkIfAbapOrStreamHandle( *this, &l_typ, &l_size, &l_elem_num, &foundptr );
              short l_align = (*foundptr)->getAlignment();
              if ( 0 != l_size % l_align )
                l_size += (l_align - l_size % l_align);  /* xxx */
              m_h_out << ", " << l_typ 
                << ", " << l_subtyp
                << ", " << l_size
                << ", " << l_elem_num
                << ", 0, 0,";
              
              char cData[9];
              sprintf(cData,"%x",guid->Data1);
              m_h_out << " { 0x" << cData;
              sprintf(cData,"%x",guid->Data2);
              m_h_out << ", 0x" << cData;
              sprintf(cData,"%x",guid->Data3);
              m_h_out << ", 0x" << cData << ", {";

              //m_h_out.setf(ios::hex);
              //m_h_out << " { 0x" << guid->Data1 << ", 0x" << guid->Data2 << ", 0x" 
              //  << guid->Data3 << ", {";
              
              for ( i=0; i < 7; i++ ) {
                sprintf(cData,"%x",(unsigned short)guid->Data4[i]);
                m_h_out << " 0x" << cData << ",";
                //m_h_out << " 0x" << (unsigned short)guid->Data4[i] << ",";
              };
              sprintf(cData,"%x",(unsigned short)guid->Data4[7]);
              m_h_out << " 0x" << cData << " }}, (char*)\"" << name_idx <<"\"";
              //m_h_out << " 0x" << (unsigned short)guid->Data4[7] << " }}, (char*)\"" << name_idx <<"\"";
              //m_h_out.unsetf(ios::hex);
              
              short pos = 0;
              (*foundptr)->printSubParameter( *this, m_h_out, pos);
            }
            else {
              /* Struct must be passed via pointer */
              co83ErrorMsg( C_PASS_STRUCT_BY_PTR_CO83, g_act_method, name_idx );
              if ( !error )
                error = - C_PASS_STRUCT_BY_PTR_CO83;
              break;
            }
          }
          else { 
            /* parameter is no struct */
            co81CTypedef* const * searchptr = foundptr;
            findNode.setName( elemUnknown_co82, (*foundptr)->getType() );
            while ( searchptr && ((*searchptr)->getDefType() == elemTypedef_co82 ) 
              && (0 == (*searchptr)->getMemberNum()) ) {
              if ( (*searchptr)->getPtr() )
              {
                if ( is_ptr ) {
                  co83ErrorMsg( C_PTR_OF_PTR_UNSUPPORTED_CO83, g_act_method, name_idx, 
                    (*foundptr)->getType());
                  if ( !error )
                    error = - C_PTR_OF_PTR_UNSUPPORTED_CO83;
                  break;
                }
                else {
                  is_ptr = 1;
                }
              }
              if ( (*searchptr)->getDimension() > 0 ) {
                if ( l_dim > 0 ) {
                  co83ErrorMsg( C_UNSUPPORTED_VT_CARRAY2_CO83, g_act_method, name_idx );
                  if ( !error )
                    error = - C_UNSUPPORTED_VT_CARRAY2_CO83;
                  break;
                }
                else
                {
                  l_dim = (*searchptr)->getDimension();
                }
              }
              findNode.setName( elemUnknown_co82, (*searchptr)->getType() );
              searchptr = this->findNode(findNode);
            }
            if ( error ) break;

            if ( is_ptr ) {
              co81CTypedef findNode;
              l_subtyp  = (*foundptr)->getVtType();
              l_size = (*foundptr)->getSize();
              if ( foundptr && ( 1 == (*foundptr)->getMemberNum() ) && (l_typ == csp1_vt_userdefined)) {
                co81CTypedef *base = *((*foundptr)->getMemberList().begin())();
                if ( l_dim > 0 ) {
                  l_typ = csp1_vt_carray;
                  if ( l_size != 1 || ( l_dim > 0) ) {
                    co83ErrorMsg( C_UNSUPPORTED_VT_CARRAY1_CO83, g_act_method,
                         (*foundptr)->getType());
                    if ( !error )
                      error = - C_UNSUPPORTED_VT_CARRAY1_CO83;
                    /* break;          PTS 1109282 */
                    l_typ = l_subtyp = l_size = l_dim = -1;
                    break;          /* PTS 1109282 */
                  }
                  else 
                    l_size = l_size * l_dim;
                }
                else {
                  l_size = (*foundptr)->getSize();
                  l_dim = 0;
                  l_subtyp = base->getVtType();
                }
              }
              else {
                if ( l_dim ) {
                  l_typ = csp1_vt_carray;
                  l_dim = (*foundptr)->getDimension();
                  if ( l_size != 1 ) {
                    co83ErrorMsg( C_UNSUPPORTED_VT_CARRAY1_CO83, g_act_method, (*foundptr)->getType());
                    if ( !error )
                      error = - C_UNSUPPORTED_VT_CARRAY1_CO83;
                    l_typ = l_subtyp = l_size = l_dim = -1;
                    break;
                  }
                  else 
                    l_size = l_size * l_dim;
                }
              }
              
              if ( (l_dim < 1) && direction_idx && !strcmp("in",direction_idx) 
                && searchptr && elemStruct_co82 != (*searchptr)->getDefType() ) {

                l_dim = l_size = l_subtyp = -1;
                co83ErrorMsg( C_IN_SCALAR_BYPTR_CO83, g_act_method, name_idx, buf);
                if ( !error )
                  error = - C_IN_SCALAR_BYPTR_CO83;
                break;
              };
              
              if ( l_subtyp == csp1_vt_i8 ) {
                co83ErrorMsg( C_UNSUPPORTED_PARAM_TYPE_CO83, g_act_method, name_idx, type_idx);
                if ( !error )
                  error = - C_UNSUPPORTED_PARAM_TYPE_CO83;
                break;
              }

              /* following patch has been done, because liveCache uses in this
                 cases only l_dim to determine the length of the parameter      */
              if ((( l_typ == csp1_vt_ptr ) || ( l_typ == csp1_vt_userdefined )) &&
                ( 0 == l_dim ) && ( 1 == l_size ) )
                l_dim = l_size;

              m_h_out << ", " << l_typ 
                << ", " << l_subtyp
                << ", " << l_size
                << ", " << l_dim
                << ", 0, 0,";

            }
            else {
              if ( (*foundptr)->getVtType() == csp1_vt_i8 ) {
                co83ErrorMsg( C_UNSUPPORTED_PARAM_TYPE_CO83, g_act_method, name_idx, type_idx);
                if ( !error )
                  error = - C_UNSUPPORTED_PARAM_TYPE_CO83;
                break;              
              }
              if ( l_dim > 0 ) {
                co83ErrorMsg( C_PASS_STRUCT_BY_PTR_CO83, g_act_method, name_idx );
                if ( !error )
                  error = - C_PASS_STRUCT_BY_PTR_CO83;
                break;
              }
              else {
                m_h_out << ", " << (*foundptr)->getVtType()
                  << ", " << 0
                  << ", " << l_size
                  << ", " << 0
                  << ", 0, 0,";
              }
            }
            
            char cData[9];
            sprintf(cData,"%x",guid->Data1);
            m_h_out << " { 0x" << cData;
            sprintf(cData,"%x",guid->Data2);
            m_h_out << ", 0x" << cData;
            sprintf(cData,"%x",guid->Data3);
            m_h_out << ", 0x" << cData << ", {";
            //m_h_out.setf(ios::hex);
            //m_h_out << " { 0x" << guid->Data1 << ", 0x" << guid->Data2 << ", 0x" 
            //  << guid->Data3 << ", {";
            
            for ( i=0; i < 7; i++ ) {
              sprintf(cData,"%x",(unsigned short)guid->Data4[i]);
              m_h_out << " 0x" << cData << ",";
              //m_h_out << " 0x" << (unsigned short)guid->Data4[i] << ",";
            };
            sprintf(cData,"%x",(unsigned short)guid->Data4[i]);
            m_h_out << " 0x" << cData << " }}, (char*)\"" << name_idx <<"\"";
            //m_h_out << " 0x" << (unsigned short)guid->Data4[7] << " }}, (char*)\"" << name_idx <<"\"";
            //m_h_out.unsetf(ios::hex);
          }


        }
        else {  
          /* parameter is not a user defined type */

          if ( strchr( type_idx, '*' )) {
            strncpy(buf, type_idx, strlen(type_idx) - 1);
            buf[strlen(type_idx) - 1] = '\0';
            l_typ = csp1_vt_ptr;
            l_subtyp = checkBasisType( buf );
            if ( l_subtyp == csp1_vt_i8 ) {
              co83ErrorMsg( C_UNSUPPORTED_PARAM_TYPE_CO83, g_act_method, name_idx, type_idx);
              if ( !error )
                error = - C_UNSUPPORTED_PARAM_TYPE_CO83;
              break;              
            }
          }
          else {
            strcpy(buf, type_idx);
            l_typ = checkBasisType( buf );
            if ( l_typ == csp1_vt_i8 ) {
              co83ErrorMsg( C_UNSUPPORTED_PARAM_TYPE_CO83, g_act_method, name_idx, type_idx);
              if ( !error )
                error = - C_UNSUPPORTED_PARAM_TYPE_CO83;
              break;              
            }
            l_subtyp = 0;
          };

          if (( 0 == strcmp(buf,"hyper") ) ||
            ( 0 == strcmp(buf,"char") ) ||
            ( 0 == strcmp(buf,"boolean") ) ||
            ( 0 == strcmp(buf,"bool") ) ||
            ( 0 == strcmp(buf,"signed long") ) ||
            ( 0 == strcmp(buf,"signed long int") ) ||
            ( 0 == strcmp(buf,"unsigned long int") ) ||
            ( 0 == strcmp(buf,"unsigned") ) ||
            ( 0 == strcmp(buf,"long") ) ||
            ( 0 == strcmp(buf,"unsigned long") ) ||
            ( 0 == strcmp(buf,"signed long") ) ||
            ( 0 == strcmp(buf,"long int") ) ) {
            char l_buf[128];
            strcpy(l_buf, "parameter");
            strcat(l_buf,":");
            strcat(l_buf,name_idx);
            l_size = -1;
            co83WarnMsg( C_TYPE_NOT_PORTABLE_CO83, l_buf, buf );
          }

          short l_tmp;

          if ( !getTypeCharacteristics( buf, "parameter", name_idx, l_size, l_tmp ) ) break;

          if ( is_ptr ) {
            if ( direction_idx && !strcmp("in",direction_idx)) {
              co83ErrorMsg( C_IN_SCALAR_BYPTR_CO83, g_act_method, name_idx, buf);
              if ( !error )
                error = - C_IN_SCALAR_BYPTR_CO83;
              break;
            }
          }
          else {
            if ( l_dim > 0 ) {
              co83ErrorMsg( C_PASS_STRUCT_BY_PTR_CO83, g_act_method, name_idx );
              if ( !error )
                error = - C_PASS_STRUCT_BY_PTR_CO83;
              break;
            }
          }

          m_h_out << ", " << l_typ 
            << ", " << l_subtyp
            << ", " << l_size
            << ", " << l_dim
            << ", 0, 0, { 0x0, 0x0, 0x0, { 0x0, 0x0, 0x0, 0xa0, 0x0, 0x0, 0x0, 0x0 }}, (char*)\""
            << name_idx << "\""; 
        };
      };
    };
    break;
  case elemExpr_co82:
    m_expr_flag = true;
    break;
  case elemSub_co82:
  case elemAdd_co82:
  case elemFloat_co82:
  case elemFixed_co82:
  case elemMul_co82:
  case elemDiv_co82:
  case elemMod_co82:
  case elemIdent_co82:
    if ( m_expr_flag && ( NULL != m_act_pType )) {
       m_expr_token.m_token_type  = elemtype;
       m_expr_token.m_token_value = NULL;
    };
    break;
  case elemLibrary_co82:
  case elemCpp_co82:
  case elemPragma_co82:
  case elemImport_co82:
  case elemEnum_co82:
  case elemInclude_co82:
  case elemUnknown_co82:
    break;
  default:
    break;
  };
};

/*----------------------------------------------------------------------*/

co81CTypedef* const * co81CGenInspecCode::findNode( co81CTypedef &findNode ) {
   return (m_index.FindNode(&findNode))->GetKey();
};


/*----------------------------------------------------------------------*/

void co81CGenInspecCode::endElement(int &error, const char *name) {
  co81CTypedef* pType;
  XmlElementType_Enum_co82 elemtype = co82getElementType(name);
  tgg00_BasisError basis_err = e_ok;

  switch (elemtype) {
  case elemConst_co82:
    basis_err = m_index.Insert(m_act_pType);
    if ( basis_err ) {
      co83ErrorMsg( C_DUP_DEFINITION_CO83, m_act_pType->getName() );
      if ( !error )
        error = - C_DUP_DEFINITION_CO83;
    }
    m_act_pType = NULL;
    break;
  case elemStruct_co82:
    pType = g_struct_level.front();
    g_struct_level.pop_front();
    pType->setStructCharacteristics(*this);
    break;
  case elemImport_co82:
  case elemLibrary_co82:
  case elemCpp_co82:
  case elemPragma_co82:
  case elemInclude_co82:
  case elemTypedef_co82:
  case elemMember_co82:
  case elemEnum_co82:
  case elemInterface_co82:
    break;
  case elemIdl_co82:
    if ( g_param_no ) {
      m_h_out << "  };"                                                                               << endl; 
      m_h_out                                                                                         << endl;
      m_h_out << "  if (( ParmIdx > (sizeof(CO_ParmDesc) / sizeof(t_ParmsDesc))) || ( ParmIdx < 1 ))" << endl;
      m_h_out << "  {"                                                                                << endl;
      m_h_out << "    *lplpMode = NULL;"                                                              << endl;
      m_h_out << "  }"                                                                                << endl;
      m_h_out << "  else"                                                                             << endl;
      m_h_out << "  {"                                                                                << endl;
      m_h_out << "    *lplpMode       = CO_ParmDesc[ParmIdx - 1].Mode;"                               << endl;
      m_h_out << "    *lplpIOtype     = CO_ParmDesc[ParmIdx - 1].IOtype;"                             << endl;
      m_h_out << "    *lpDatatype     = CO_ParmDesc[ParmIdx - 1].Datatype;"                           << endl;
      m_h_out << "    *lpSubdatatype  = CO_ParmDesc[ParmIdx - 1].Subdatatype;"                        << endl;
      m_h_out << "    *lpLength       = CO_ParmDesc[ParmIdx - 1].Length;"                             << endl;
      m_h_out << "    *lpDim          = CO_ParmDesc[ParmIdx - 1].Dim;"                                << endl;
      m_h_out << "    *lpOffset       = CO_ParmDesc[ParmIdx - 1].Offset;"                             << endl;
      m_h_out << "    *lpFrac         = CO_ParmDesc[ParmIdx - 1].Frac;"                               << endl;
      m_h_out << "    *lplpGUID       = (GUID*)&CO_ParmDesc[ParmIdx - 1].guid;"                       << endl;
      m_h_out << "    *lplpParmName   = CO_ParmDesc[ParmIdx - 1].lpParmName;"                         << endl;
      m_h_out << "  }"                                                                                << endl;
      m_h_out << "};"                                                                                 << endl;
    }
    else {
      m_h_out << "  *lplpMode = NULL;"                                                                << endl;
      m_h_out << "}"                                                                                  << endl;
    }
    m_cpp_out << g_param_no << ", &co_IntrospectFunction" << g_func_no;
    break;
  case elemExpr_co82:
    if ( m_expr_flag && ( NULL != m_act_pType )) {
      int rc = m_act_pType->calExprValue(*this);
      if ( rc )
        if ( !error )
          error = - rc;
    }
    m_expr_flag = false;
    break;
  case elemSub_co82:
  case elemAdd_co82:
  case elemFloat_co82:
  case elemFixed_co82:
  case elemMul_co82:
  case elemDiv_co82:
  case elemMod_co82:
  case elemIdent_co82:
    if ( m_expr_flag && ( NULL != m_act_pType )) {
       co81CExprElem* p = new co81CExprElem;
       *p = m_expr_token;
       m_act_pType->appendToken(p);
    }
    break;
  case elemCoClass_co82:
  case elemUnknown_co82:
  default:
    break;
  }
}

/*----------------------------------------------------------------------*/

void co81CGenInspecCode::charData(int &error, const char *s, int len) {
  if ( m_expr_flag && ( NULL != m_act_pType )) {
    m_expr_token.m_token_value = new char[ len + 1 ];
    memcpy( m_expr_token.m_token_value, s, len );
    m_expr_token.m_token_value[ len ] = '\0';
  }
};
