/*!***************************************************************************

  module      : vco81CGenInspecCode.cpp

  -------------------------------------------------------------------------

  responsible : FerdiF

  special area: IDL to Code Generator   
  description : Main Program
				
  
  last changed: 2000-03-08  11:22
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/

#include <fstream.h>
#include <stdio.h>
#include "idl/expat/xmlparse.h"

#include "hco83ErrorHandling.h"
#include "hco82Idl2Code.h"
#include "hco80CGenHFile.h"
#include "hco81CGenInspecCode.h"

const int cmx_filename_len      = 256;
const int cmx_input_buffer_size = 4096;

typedef char t_filename[cmx_filename_len];

typedef struct t_xmlUserArgs {
  co80CGenHFile       m_co80CGenHFile;
  co81CGenInspecCode  m_co81CGenInspecCode;
  int                 m_depth;
  DboType_Enum_co82   m_dboType; 
  int                 m_error;
  XML_Parser          m_parser;
} t_xmlUserArgs;

/*----------------------------------------------------------------------*/
/* upper case conversion of a null terminated s                         */

inline char* strupper (char *s) {
  char *c = s;
  while (*c) {
    if (islower (*c))
      *c = toupper (*c);
    c++;
  }
  return s;
}

/*----------------------------------------------------------------------*/

XmlAttributeType_Enum_co82 co82getAttributeType(const char* attrib) {

  const char  cName[]        = "NAME";
  const char  cString[]      = "STRING";
  const char  cNumber[]      = "NUMBER";
  const char  cChar[]        = "CHAR";
  const char  cPack[]        = "PACK";
  const char  cUuid[]        = "UUID";
  const char  cType[]        = "TYPE";
  const char  cDimension[]   = "DIMENSION";
  const char  cDirection[]   = "DIRECTION";
  const char  cExpression[]  = "EXPRESSION";

  if ( 0 == strcmp(attrib,cName) )        return attName_co82;
  if ( 0 == strcmp(attrib,cString) )      return attString_co82;
  if ( 0 == strcmp(attrib,cNumber) )      return attNumber_co82;
  if ( 0 == strcmp(attrib,cChar) )        return attChar_co82;
  if ( 0 == strcmp(attrib,cPack) )        return attPack_co82;
  if ( 0 == strcmp(attrib,cUuid) )        return attUuid_co82;
  if ( 0 == strcmp(attrib,cType) )        return attType_co82;
  if ( 0 == strcmp(attrib,cDimension) )   return attDimension_co82;
  if ( 0 == strcmp(attrib,cDirection) )   return attDirection_co82;
  if ( 0 == strcmp(attrib,cExpression) )  return attExpression_co82;
  return attUnknown_co82;

};

/*----------------------------------------------------------------------*/

XmlElementType_Enum_co82 co82getElementType(const char* element) {

  const char  cLibrary[]     = "LIBRARY";
  const char  cCpp[]         = "CPP";
  const char  cPragma[]      = "PRAGMA";
  const char  cImport[]      = "IMPORT";
  const char  cConst[]       = "CONST";
  const char  cTypedef[]     = "TYPEDEF";
  const char  cStruct[]      = "STRUCT";
  const char  cMember[]      = "MEMBER";
  const char  cEnum[]        = "ENUM";
  const char  cInclude[]     = "INCLUDE";
  const char  cInterface[]   = "INTERFACE";
  const char  cIdl[]         = "IDL";
  const char  cParam[]       = "PARAM";
  const char  cCoClass[]     = "COCLASS";
  const char  cIdl2Xml[]     = "IDL2XML";
  const char  cExpr[]        = "EXPR";
  const char  cSub[]         = "SUB";
  const char  cAdd[]         = "ADD";
  const char  cFloat[]       = "FLOAT";
  const char  cFixed[]       = "FIXED";
  const char  cMul[]         = "MUL";
  const char  cDiv[]         = "DIV";
  const char  cMod[]         = "MOD";
  const char  cIdent[]       = "IDENT";

  if ( 0 == strcmp(element,cLibrary) )  return elemLibrary_co82;
  if ( 0 == strcmp(element,cCpp) )      return elemCpp_co82;
  if ( 0 == strcmp(element,cPragma) )   return elemPragma_co82;
  if ( 0 == strcmp(element,cImport) )   return elemImport_co82;
  if ( 0 == strcmp(element,cConst) )    return elemConst_co82;
  if ( 0 == strcmp(element,cTypedef) )  return elemTypedef_co82;
  if ( 0 == strcmp(element,cStruct) )   return elemStruct_co82;
  if ( 0 == strcmp(element,cMember) )   return elemMember_co82;
  if ( 0 == strcmp(element,cEnum) )     return elemEnum_co82;
  if ( 0 == strcmp(element,cInclude) )  return elemInclude_co82;
  if ( 0 == strcmp(element,cInterface) )return elemInterface_co82;
  if ( 0 == strcmp(element,cIdl) )      return elemIdl_co82;
  if ( 0 == strcmp(element,cCoClass) )  return elemCoClass_co82;
  if ( 0 == strcmp(element,cParam) )    return elemParam_co82;
  if ( 0 == strcmp(element,cIdl2Xml) )  return elemIdl2Xml_co82;
  if ( 0 == strcmp(element,cExpr) )     return elemExpr_co82;
  if ( 0 == strcmp(element,cSub) )      return elemSub_co82;
  if ( 0 == strcmp(element,cAdd) )      return elemAdd_co82;
  if ( 0 == strcmp(element,cFloat) )    return elemFloat_co82;
  if ( 0 == strcmp(element,cFixed) )    return elemFixed_co82;
  if ( 0 == strcmp(element,cMul) )      return elemMul_co82;
  if ( 0 == strcmp(element,cDiv) )      return elemDiv_co82;
  if ( 0 == strcmp(element,cMod) )      return elemMod_co82;
  if ( 0 == strcmp(element,cIdent) )    return elemIdent_co82;
  return elemUnknown_co82;
};

/*----------------------------------------------------------------------*/

bool co82isExprParam(const char* elem) {

  if ( elemFloat_co82 == co82getElementType(elem) ) return true;
  if ( elemFixed_co82 == co82getElementType(elem) ) return true;
  if ( elemIdent_co82 == co82getElementType(elem) ) return true;
  return false;
};

/*----------------------------------------------------------------------*/

static int ProcessCommandLine (int argc , 
                               char * argv[],
                               t_filename& pcInputXML,
                               DboType_Enum_co82& dboType ){

  if ( argc < 2 ) {
    co83ErrorMsg( C_WRONG_PARAMETERS_CO83 );
    return 0;
  }
  else {
    char buf[64];
    strcpy(&pcInputXML[0],argv[1]);
    if ( argc > 2 ) {
      strcpy(buf,argv[2]);
      
      if ( !strcmp("UDT",strupper(buf)) ) 
        dboType = dbo_method_co82;
      else if ( !strcmp("DBPROC",strupper(buf)) ) 
        dboType = dbo_proc_co82;
      else { 
        co83ErrorMsg( C_WRONG_PARAMETERS_CO83 );
        return 0;
      }
    }
    else {
      dboType = dbo_proc_co82;
    }
  }
  return 1;
}

/*----------------------------------------------------------------------*/

externC void charData(void *userData, const char *s, int len)
{
  t_xmlUserArgs *puserArgs = (t_xmlUserArgs*)userData;
  puserArgs->m_co81CGenInspecCode.charData(puserArgs->m_error, s, len);
};

/*----------------------------------------------------------------------*/

externC void startElement(void *userData, const char *name, const char **atts)
{
  t_xmlUserArgs *puserArgs = (t_xmlUserArgs*)userData;
  if ( co82isExprParam(name) ) 
    XML_SetCharacterDataHandler(puserArgs->m_parser, &charData);
  puserArgs->m_co80CGenHFile.startElement(name, atts);
  puserArgs->m_co81CGenInspecCode.startElement(puserArgs->m_error, puserArgs->m_dboType, name, atts);
  puserArgs->m_depth++;
}

/*----------------------------------------------------------------------*/

externC void endElement(void *userData, const char *name)
{
  t_xmlUserArgs *puserArgs = (t_xmlUserArgs*)userData;
  puserArgs->m_co80CGenHFile.endElement(name);
  puserArgs->m_co81CGenInspecCode.endElement(puserArgs->m_error, name);
  puserArgs->m_depth--;
  if ( co82isExprParam(name) ) 
    XML_SetCharacterDataHandler(puserArgs->m_parser, NULL);
}

/*----------------------------------------------------------------------*/

int main (int argc , char * argv[]) {
  t_filename        cInputXML;
  tsp00_VFilenamec  cFilename;
  char              buf[cmx_input_buffer_size];
  t_xmlUserArgs     userArgs;

  //DebugBreak();
  strcpy(cFilename,"");
  if ( ProcessCommandLine(argc , argv, cInputXML, userArgs.m_dboType) ) {
    userArgs.m_depth = 0;
    userArgs.m_error = 0;
    FILE *strmInputXML;
    if ( userArgs.m_co80CGenHFile.Init(cInputXML) && ((strmInputXML = fopen( cInputXML, "r" )) != NULL) ) {
      userArgs.m_co81CGenInspecCode.Init(cInputXML);
      userArgs.m_parser = XML_ParserCreate(NULL);
      int done;
      XML_SetUserData(userArgs.m_parser, &userArgs);
      XML_SetElementHandler(userArgs.m_parser, &startElement, &endElement);
      do {
        size_t len = fread(buf, 1, sizeof(buf), strmInputXML);
        done = len < sizeof(buf);
        if (!XML_Parse(userArgs.m_parser, buf, len, done)) {
          co83ErrorMsg( C_XML_PARSE_ERROR_CO83, 
            XML_ErrorString(XML_GetErrorCode(userArgs.m_parser)), 
            XML_GetCurrentLineNumber(userArgs.m_parser) );
          return C_XML_PARSE_ERROR_CO83;
        }
      } while (!done);
      XML_ParserFree(userArgs.m_parser);
    }
  }
  return userArgs.m_error;
}


