/*!***************************************************************************

  module      : vco83ErrorHandling.cpp

  -------------------------------------------------------------------------

  responsible : FerdiF

  special area: IDL to Code Generator   
  description : Error Handling
				
  
  last changed: 2000-03-08  11:22
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/

#include <fstream.h>
#include <iostream.h>
#include <stdio.h>
#include "hco83ErrorHandling.h"

static const char *EMsgPrefix = "xml2ispc -E- code:";
static const char *WMsgPrefix = "xml2ispc -W- code:";

static const char* co83_error_msg[C_NUMBER_OF_ERROR_CO83] = {
 "invalid call parameters passed \n\n   USAGE:xml2ispc <xml-file output of idlchecker> DBPROC | UDT", /* C_WRONG_PARAMETERS_CO83 */
 "XML Parse Error: %s at line %d",                                              /* C_XML_PARSE_ERROR_CO83         */
 "method: %s, first parameter MUST be of type VT_USERDEFINED; type found %d",   /* C_P1_NO_USERDEFINED_TYPE_CO83  */
 "method: %s, first parameter MUST not be of I/O type INPUT",                   /* C_P1_IO_TYPE_INPUT_CO83        */
 "method: %s, first parameter MUST not be of I/O type OUTPUT",                  /* C_P1_IO_TYPE_OUTPUT_CO83       */
 "method: %s, first parameter has wrong uid",                                   /* C_P1_WRONG_GUID_CO83           */
 "method: %s, first parameter has no uid assigned",                             /* C_P1_NO_GUID_CO83              */
 "method: %s, type specification of first parameter incomplete",                /* C_P1_TYPE_SPEC_INCOMPLETE_CO83 */
 "%s, type '%s' is unknown or not supported",                                   /* C_TYPE_UNKNOWN_CO83            */
 "method: %s, unsupported parameter type, VT_CARRAY with elements of type '%s'",/* C_UNSUPPORTED_VT_CARRAY1_CO83  */
 "method: %s, directional spec. on parameter %s missing, [IN,OUT] assumed",     /* C_NO_DIR_SPECIFIC_CO83         */
 "method: %s, structure or array %s must be passed by pointer",                          /* C_PASS_STRUCT_BY_PTR_CO83      */
 "method: %s, parameter: %s, scalar datatype '%s' passed by ptr but parameter is input only", /* C_IN_SCALAR_BYPTR_CO83 */
 "method: %s, parameter: %s, datatype '%s' not allowed",                        /* C_UNSUPPORTED_PARAM_TYPE_CO83  */
 "method: %s, parameter: %s, pointer of pointer of datatype '%s' not allowed",  /* C_PTR_OF_PTR_UNSUPPORTED_CO83  */
 "method: %s, parameter: %s, unsupported type, VT_CARRAY with dimension > 1",   /* C_UNSUPPORTED_VT_CARRAY2_CO83  */
 "%s, type '%s' is NOT portable",                                               /* C_TYPE_NOT_PORTABLE_CO83       */
 "method: %s, parameter:'%s' member:'%s' not properly aligned",                 /* C_STRUCT_NOT_ALIGNED_CO83      */
 "method: %s, parameter:'%s' member:'%s' pointer within struct not allowed",    /* C_POINTER_NOT_ALLOWED_CO83     */
 "dimension: '%s' of struct_member: '%s', type '%s' is undefined",              /* C_DIMENSION_TYPE_UNDEF_CO83    */
 "invalid expression: '%s' defining CONST with name: '%s', type '%s' failed",   /* C_CONST_EXPR_NOT_VALID_CO83    */
 "duplicate definition of '%s'"                                                 /* C_DUP_DEFINITION_CO83          */
};

/*----------------------------------------------------------------------*/

void co83ErrorMsg( const int err_no ) {
  if ( (0 < err_no) &&(err_no <= C_NUMBER_OF_ERROR_CO83 )) {
    cout << EMsgPrefix << err_no << " " << co83_error_msg[err_no -1] << endl;  
  };
};

/*----------------------------------------------------------------------*/

void co83ErrorMsg( const int err_no, const char* p1 ) {
  if ( (0 < err_no) &&(err_no <= C_NUMBER_OF_ERROR_CO83 )) {
    char buffer[256];
    if ( p1 ) {
      sprintf( buffer, co83_error_msg[err_no -1], p1); 
      cout << EMsgPrefix << err_no << " " << buffer << endl;  
    };
  };
};

/*----------------------------------------------------------------------*/

void co83ErrorMsg( const int err_no, const char* p1, const int p2 ) {
  if ( (0 < err_no) &&(err_no <= C_NUMBER_OF_ERROR_CO83 )) {
    char buffer[256];
    if ( p1 && p2 ) {
      sprintf( buffer, co83_error_msg[err_no -1], p1, p2 ); 
      cout << EMsgPrefix << err_no << " " << buffer << endl;  
    };
  };
};

/*----------------------------------------------------------------------*/

void co83ErrorMsg(const int err_no, const char* p1, const char* p2 ) {
  if ( (0 < err_no) &&(err_no <= C_NUMBER_OF_ERROR_CO83 )) {
    char buffer[256];
    if ( p1 && p2 ) {
      sprintf( buffer, co83_error_msg[err_no -1], p1, p2 ); 
      cout << EMsgPrefix << err_no << " " << buffer << endl;  
    };
  };
};

/*----------------------------------------------------------------------*/

void co83ErrorMsg(const int err_no, const char* p1, const char* p2, 
                  const char* p3 ) {
  if ( (0 < err_no) &&(err_no <= C_NUMBER_OF_ERROR_CO83 )) {
    char buffer[256];
    if ( p1 && p2 && p3 ) {
      sprintf( buffer, co83_error_msg[err_no -1], p1, p2, p3 ); 
      cout << EMsgPrefix << err_no << " " << buffer << endl;  
    };
  };
};

/*----------------------------------------------------------------------*/

void co83WarnMsg( const int err_no, const char* p1, const char* p2 ) {
  if ( (0 < err_no) &&(err_no <= C_NUMBER_OF_ERROR_CO83 )) {
    char buffer[256];
    if ( p1 && p2 ) {
      sprintf( buffer, co83_error_msg[err_no -1], p1, p2 ); 
      cout << WMsgPrefix << err_no << " " << buffer << endl;  
    };
  };
};

/*----------------------------------------------------------------------*/

void co83WarnMsg( const int err_no, const char* p1, const char* p2, const char* p3 ) {
  if ( (0 < err_no) &&(err_no <= C_NUMBER_OF_ERROR_CO83 )) {
    char buffer[256];
    if ( p1 && p2 && p3 ) {
      sprintf( buffer, co83_error_msg[err_no -1], p1, p2, p3 ); 
      cout << WMsgPrefix << err_no << " " << buffer << endl;  
    };
  };
};
