/****************************************************************************

  module      : ven35.cpp

  -------------------------------------------------------------------------

  responsible : JoergM

  special area: Kernel LZU UNIX DCOM Thread
  description : DCOMx_Function Thread support

  last changed: 2000-04-27  15:15
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1998



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/
#if defined(OSF1)
#define _XOPEN_SOURCE_EXTENDED
#endif /* OSF1 */

#include "ggg00.h"
#include "gsp00.h"
#include "geo00k.h"
#include "geo001.h"

#include "gen35.h"
#include "gco00.h"
#include "heo35.h"
#include "heo36.h"
#include "heo38.h"
#include "heo46.h"
#include "heo60.h"
#include "geo50_0.h"
#include "geo007_1.h"
#include "geo007_2.h"
#include "hen40.h"
#include "gen41.h"
#include "geo57.h"
#include "gen71.h"
#include "gen72.h"
#include "gen73.h"
#include "gen76.h"
#include "hen50.h"
#include "gen81.h"
/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define TASK_REL_EN35	    0
#define UKT_REL_EN35	    1

#define NUM_OF_DCOM_MXEO35 2

#define MAX_DCOM_ERRCODE                        6
#define ERRCODE_DCOM_CREATE_DCOMI               0
#define ERRCODE_DCOM_RESUME_DCOMI               1
#define ERRCODE_DCOMI_CLOSE_EVENT_SEM           2
#define ERRCODE_DCOM_MAX_DCOMIS_REACHED         3
#define ERRCODE_DCOM_GET_DCOMI_CTRL             4
#define ERRCODE_DCOM_REM_UKT_DCOMi_NOT_ALLOWED  5

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/



/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/

extern key_t en81_common_rte_key;

/*===========================================================================*
 *  GLOBAL VARIABLES                                                         *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL VARIABLES                                                          *
 *===========================================================================*/

static const char * en35_ErrTab[ MAX_DCOM_ERRCODE ] =
{
    ERRMSG_DCOM_CREATE_DCOMI,             // - ERRCODE_DCOM_CREATE_DCOMI
    ERRMSG_DCOM_RESUME_DCOMI,             // - ERRCODE_DCOM_RESUME_DCOMI
    ERRMSG_DCOMI_CLOSE_EVENT_SEM,         // - ERRCODE_DCOMI_CLOSE_EVENT_SEM
    ERRMSG_DCOM_MAX_DCOMIS_REACHED,       // - ERRCODE_DCOM_MAX_DCOMIS_REACHED
    ERRMSG_DCOM_GET_DCOMI_CTRL,           // - ERRCODE_DCOM_GET_DCOMI_CTRL
    ERRMSG_DCOM_REM_UKT_DCOMi_NOT_ALLOWED // - ERRCODE_DCOM_REM_UKT_DCOMi_NOT_ALLOWED
};



/*===========================================================================*
 *  LOCAL FUNCTIONS (PROTOTYPES)                                             *
 *===========================================================================*/

static void en35_ReqFromDCOMxToUKT ( struct DOUBLY_LINKED      *pRequest );

static void en35_ReqToDCOM0 ( struct DOUBLY_LINKED      *pRequest );

static void en35_ReqToDCOMi ( ten50_DCOMiCtrl *pDCOMiCtrl,
                              struct DOUBLY_LINKED *pRequest );

static struct DOUBLY_LINKED * en35_DCOMReceive (ten50_DCOM0Ctrl *pDCOM );

static ten50_DCOMiCtrl* en35_DCOMiInit ( struct TASK_TYPE *pTaskCtrl,
                                         tsp00_Uint4            ReqType,
                                         tsp00_ErrText    &Errtext ) ;

static ten50_DCOMiCtrl* en35_GetDCOMiCtrl ( struct DOUBLY_LINKED  *pRequest );

static void en35_ReleaseDCOMiCtrl ( tsp00_Uint4                     Relationship,
                                    ten50_DCOMiCtrl          **ppDCOMiCtrl );

static bool en35_DCOMiCreate ( struct DOUBLY_LINKED   *pRequest );

static bool en35_DCOMiRemove ( struct DOUBLY_LINKED    *pRequest );

static void* en35_DCOMi ( void * );

static void en35_DCOMiTerminate ( ten50_DCOMiCtrl *pDCOMiCtrl,
                                  ten50_DlqDCOMCtrlReqArgs *DCOMCtrlRequest );

static tsp00_Int4 en35_AllocDCOMiCtrl ( ten50_DCOMiCtrl **ppDCOMiCtrl );

static void en35_FreeDCOMiCtrl ( ten50_DCOMiCtrl *pDCOMiCtrl );

static void en35_init_dcomi_ctrl ( ten50_DCOMiCtrl *pDCOMiCtrl );

/*===========================================================================*
 *  GLOBAL FUNCTIONS (CODE)                                                  *
 *===========================================================================*/

externC void en35_VDCOMTransfer ( void *pTransportBuffer, 
				  bool *pOk,
				  tsp00_ErrText *pErrtext )
{
  ROUTINE_DBG_MEO00 ("en35_VDCOMTransfer");

  *pOk = sql35_VDCOMiRequest ( REQ_DCOM_TRANSFER_EO00K, pTransportBuffer, *pErrtext );

  return;
}

/*---------------------------------------------------------------------------*/

externC bool en35_RemoveDCOMi ( void  )
{
  ROUTINE_DBG_MEO00 ("en35_RemoveDCOMi");

ten50_UKT_Control *this_ukt = THIS_UKT_CTRL;
  struct TASK_TYPE      *pTaskCtrl = this_ukt->curr_task;
  struct DOUBLY_LINKED  &Request   = pTaskCtrl->qa;

  if ( pTaskCtrl->pDCOMiCtrl == NULL )
    return false;  // --- DCOMi not running!

  Request.taskid                                 = pTaskCtrl;
  Request.req_type                               = REQ_DCOM_REMOVE_DCOMI_EO00K;
  Request.args.DCOMCtrlRequest.Relationship_en50 = TASK_REL_EN35;
  
  en35_ReqToDCOM0( &Request );
  goto_disp();
  
  // --- wrong reply?
  if ( this_ukt->curr_quu != &Request )
  {
    MSGCD (( ERR_UNKNOWN_REQUEST ));
    vabort(WRITE_CORE);
  }
  
  if ( Request.args.DCOMCtrlRequest.Ok_en50 == false )
  {
    DBG3 (( MF__, "Can't create DCOMi for UKT %d", this_ukt->index ));
    MSGD  (( ERR_ERROR_TEXT, CONST_CAST(char *,en35_ErrTab[Request.args.DCOMRequest.ErrCode_en50]) ));
  }

  return Request.args.DCOMCtrlRequest.Ok_en50;
}

/*---------------------------------------------------------------------------*/

externC void *en35_DCOM0 ( void *arg  )
{
  ROUTINE_DBG_MEO00 ("en35_DCOM0");

  teo07_ThreadErr               err;
  struct DOUBLY_LINKED          *pRequest ;
  ten50_DCOM0Ctrl               &DCOM0 = KGS-> dcom0;

  en81_Set_KT_TLS ( & KGS-> dcom0 );

  // --- set thread control info
  DCOM0.state   = KT_RUNNING;
  DCOM0.start_time = time(0);

  // --- Runs forever
  while( DCOM0.start_time > 0 )
  {
    DBG3 (( MF__, "DCOM0.pDCOMWorkQue           is 0x%x", 
			DCOM0.pWorkQue_en50 ));
    DBG3 (( MF__, "DCOM0.pDCOMReqQue            is 0x%x", 
			DCOM0.pReqQue_en50 ));
    DBG3 (( MF__, "DCOM0.pDCOMWorkQue->pRequest is 0x%x", 
			DCOM0.pWorkQue_en50->pRequest ));

    //  --- Wait for a request to DCOM0
    pRequest = en35_DCOMReceive ( &DCOM0 );

    switch ( pRequest->req_type )
    {
    case REQ_DCOM_CREATE_DCOMI_EO00K  :
      en35_DCOMiCreate ( pRequest );
      en35_ReqFromDCOMxToUKT(pRequest); // -- send error / ok back to UKT
      break;
      
    case REQ_DCOM_REMOVE_DCOMI_EO00K  :
      en35_DCOMiRemove ( pRequest );
      en35_ReqFromDCOMxToUKT(pRequest); // -- send error / ok back to UKT
      break;

	default:
      MSGCD (( ERR_ILLEGAL_REQUEST, "DCOM0", pRequest->req_type ));
	  break;
    }
    DBG3 (( MF__, "next in loop" ));
  }
  /*NOTREACHED*/
  return NULL;
}

/*===========================================================================*
 *  LOCAL FUNCTIONS (CODE)                                                   *
 *===========================================================================*/

externC bool sql35_VDCOMiRequest (  tsp00_Uint4           ReqType, 
                                  void            *pParamInfo,
                                  tsp00_ErrText   &Errtext )
{
  ROUTINE_DBG_MSP00 ("sql35_VDCOMiRequest");

  ten50_DCOMiCtrl                  *pDCOMiCtrl;
  ten50_UKT_Control                   *this_ukt = THIS_UKT_CTRL;
  struct TASK_TYPE                 *pTaskCtrl = this_ukt->curr_task;
  struct DOUBLY_LINKED             &Request   = pTaskCtrl->qa;
  bool                             Ok         = false;

  pTaskCtrl->state = TSK_DCOM_OBJ_CALLED;

  pDCOMiCtrl = en35_DCOMiInit ( pTaskCtrl, ReqType, Errtext );

  if ( pDCOMiCtrl != NULL )
  {
    Request.taskid                               = pTaskCtrl;
    Request.req_type                             = ReqType;
    Request.args.DCOMRequest.pParam_en50      = pParamInfo;
    Request.args.DCOMRequest.Ok_en50          = true;
    Request.args.DCOMRequest.APIRetCode_en50  = NO_ERROR;

    en35_ReqToDCOMi ( pDCOMiCtrl, &Request );

    // --- goto dispatcher
    goto_disp();

    if ( this_ukt->curr_quu != &Request ) // --- wrong reply?
    {
      MSGCD (( ERR_UNKNOWN_REQUEST ));
      DBG1  (( MF__, "[T:0x%03x] pRequest: 0x%x (SEND) != this_ukt->curr_quu: 0x%x (REPLY)",
               pTaskCtrl->ulTaskIndex, pRequest, this_ukt->curr_quu ));
      vabort(WRITE_CORE);
    }

    Ok = Request.args.DCOMRequest.Ok_en50;

    if ( Ok == false )
        {
             eo46BuildPascalErrorStringRC ( (tsp00_Byte *)(char *)Errtext,
                      CONST_CAST(char *,en35_ErrTab[Request.args.DCOMRequest.ErrCode_en50]),
                      Request.args.DCOMRequest.APIRetCode_en50 );
        }

		DBG3 (( MF__, "[T:0x%03x] Returned from request UKT%d",
            pTaskCtrl->index, this_ukt->tid ));
    }

 pTaskCtrl->state = TSK_RUNNING;

 return Ok;
}


/*---------------------------------------------------------------------------*/


static void en35_ReqFromDCOMxToUKT ( struct DOUBLY_LINKED *pRequest )
{
  ROUTINE_DBG_MSP00 ("en35_ReqFromDCOMxToUKT");

  e71_ioc ( pRequest->taskid, pRequest );
  e72_wake ( pRequest->taskid->ukt );

  return;
}

/*---------------------------------------------------------------------------*/

static void en35_ReqToDCOM0 ( struct DOUBLY_LINKED      *pRequest )
{
  ROUTINE_DBG_MSP00 ("en35_ReqToDCOM0");
  ten50_DCOM0Ctrl             &DCOM0 = KGS-> dcom0;

  WAIT_UNTIL_ALONE ( DCOM0.alone );

  DCOM0.pReqQue_en50->pRequest = pRequest;
  DCOM0.pReqQue_en50           = DCOM0.pReqQue_en50->pNextElem;
  DCOM0.QueLen_en50++;

  CLEARLOCK ( DCOM0.alone );

  e72_wake ( REINTERPRET_CAST(ten50_UKT_Control *,&DCOM0) );

  return;
}

/*---------------------------------------------------------------------------*/


static void en35_ReqToDCOMi ( struct ten50_DCOMiCtrl *pDCOMiCtrl,
                              struct DOUBLY_LINKED *pRequest )
{
  ROUTINE_DBG_MSP00 ("en35_ReqToDCOMi");

  WAIT_UNTIL_ALONE ( pDCOMiCtrl->alone );

  pDCOMiCtrl->pReqQue_en50->pRequest = pRequest;
  pDCOMiCtrl->pReqQue_en50           = pDCOMiCtrl->pReqQue_en50->pNextElem;
  pDCOMiCtrl->QueLen_en50++;

  CLEARLOCK ( pDCOMiCtrl->alone );

  if (pDCOMiCtrl->QueLen_en50 > pDCOMiCtrl->MaxQueLen_en50)
      pDCOMiCtrl->MaxQueLen_en50 = pDCOMiCtrl->QueLen_en50;

  e72_wake ( REINTERPRET_CAST(ten50_UKT_Control *,pDCOMiCtrl) );

  return;
}

/*---------------------------------------------------------------------------*/

static struct DOUBLY_LINKED *en35_SetDCOM0_Sleeping(
ten50_DCOM0Ctrl  *pDCOM,
ten50_DCOMQueue  *pDCOMWorkQue)
{
  ROUTINE_DBG_MSP00 ("en35_SetDCOM0_Sleeping");

  pDCOM->state = KT_SLEEPING ;
  return pDCOMWorkQue->pRequest;
}

/*
 *   Tries to get a request from the DCOMx 'DCOMWork' queue, if no request
 *   is available then this thread should go to sleep until it is waked
 *   by a task/thread making a DCOM-request to it.
 */
static struct DOUBLY_LINKED * en35_DCOMReceive (
ten50_DCOM0Ctrl        *pDCOM )
{
    ROUTINE_DBG_MSP00 ("en35_DCOMReceive");
  
    tsp00_Int4               rc = NO_ERROR;
    struct DOUBLY_LINKED *   pRequest = NULL;

    while ( !pRequest )
    {
        pRequest = pDCOM->pWorkQue_en50->pRequest;

        if ( !pRequest )
        {
            pRequest = en35_SetDCOM0_Sleeping(pDCOM, pDCOM->pWorkQue_en50);
        }

        if ( pRequest ) 
        {
            pDCOM->state = KT_RUNNING;
        }
        else
        {
            if ( !pDCOM->fSemAlreadyPosted )
            {
                rc = semop ( pDCOM->semid, & semrcv, 1 ); /* wait until wakened */
                if ( rc == -1 )
                {
                    while ( (KGS-> state == SERVER_STOP)
                         || (KGS-> state == SERVER_KILL) )
                    {
                        sqlyieldthread();
                    }

                    if ( errno != EINTR )
                    {
                        vabort(WRITE_CORE);
                    }
                }
            }
            pDCOM->fSemAlreadyPosted = 0;
        }
    }
    
    WAIT_UNTIL_ALONE ( pDCOM->alone );
    
    pDCOM->pWorkQue_en50->pRequest = 0;
    pDCOM->pWorkQue_en50 = pDCOM->pWorkQue_en50->pNextElem;
    pDCOM->QueLen_en50--;
    
    CLEARLOCK ( pDCOM->alone );
    
    return ( pRequest );
}

/*---------------------------------------------------------------------------*/

static ten50_DCOMiCtrl* en35_DCOMiInit ( struct TASK_TYPE *pTaskCtrl,
                                         tsp00_Uint4       ReqType,
                                         tsp00_ErrText    &Errtext ) 
{
  ROUTINE_DBG_MSP00 ("en35_DCOMiInit");
  ten50_UKT_Control *this_ukt = THIS_UKT_CTRL;
  struct DOUBLY_LINKED  &Request   = pTaskCtrl->qa;
  tsp00_Uint4                                 Relationship;

  switch ( ReqType )
  {
  case REQ_DCOM_TRANSFER_EO00K:
    if ( pTaskCtrl->pDCOMiCtrl != NULL )
      return pTaskCtrl->pDCOMiCtrl;  // --- DCOMi already running!

    Relationship = TASK_REL_EN35;
    break;

  default:
    if ( this_ukt->pDCOMiCtrl != NULL )
      return this_ukt->pDCOMiCtrl;  // --- DCOMi already running!

    if ( !eo38CoInitialize() )
    {
      DBG3 (( MF__, "eo38CoInitialize failed for UKT %d", this_ukt->index ));
      eo46BuildPascalErrorStringRC ( (tsp00_Byte *)(char *)Errtext, (char *)ERRMSG_DCOM_COINITIALIZE_FAILED, 0 );
      return NULL;
    }

    Relationship = UKT_REL_EN35;
    break;
  }

  //
  // --- startup DCOMi
  //
  Request.taskid                                    = pTaskCtrl;
  Request.req_type                                  = REQ_DCOM_CREATE_DCOMI_EO00K;
  Request.args.DCOMCtrlRequest.Relationship_en50 = Relationship;
  
  en35_ReqToDCOM0( &Request );
  goto_disp();
  
  // --- wrong reply?
  if ( this_ukt->curr_quu != &Request )
  {
    MSGCD (( ERR_UNKNOWN_REQUEST ));
    DBG1  (( MF__, "[T:0x%03x] pRequest: 0x%x (SEND) != this_ukt->curr_quu: 0x%x (REPLY)",
             pTaskCtrl->ulTaskIndex, pRequest, this_ukt->curr_quu ));
    vabort(WRITE_CORE);
  }
  
  if ( Request.args.DCOMCtrlRequest.Ok_en50 )
  { 
    if ( Relationship == TASK_REL_EN35 )
      return pTaskCtrl->pDCOMiCtrl;
    else
      return this_ukt->pDCOMiCtrl;
  }
  else
  {
    if ( Relationship == UKT_REL_EN35 )
      eo38UnInitialize();

    DBG3 (( MF__, "Can't create DCOMi for UKT %d", this_ukt->ulUKTIndex ));
    eo46BuildPascalErrorStringRC ( (tsp00_Byte *)(char *)Errtext,
                                   CONST_CAST(char *,en35_ErrTab[Request.args.DCOMRequest.ErrCode_en50]), 
                                   Request.args.DCOMRequest.APIRetCode_en50 );
  }
  
  return NULL;
}

/*---------------------------------------------------------------------------*/

static ten50_DCOMiCtrl* en35_GetDCOMiCtrl ( struct DOUBLY_LINKED  *pRequest )
{
  ROUTINE_DBG_MSP00 ("en35_GetDCOMiCtrl");

  tsp00_Int4                             rc = NO_ERROR;
  ten50_DlqDCOMCtrlReqArgs &DCOMCtrlRequest = pRequest->args.DCOMCtrlRequest;
  ten50_DCOMiCtrl                 *pDCOMiCtrl       = NULL;
  ten50_DCOMiCtrl                 *pTmpDCOMiCtrl;
  tsp00_Uint4                            DCOMiNo;
  
  switch ( DCOMCtrlRequest.Relationship_en50 )
  {
  case TASK_REL_EN35 :

    rc = en35_AllocDCOMiCtrl( &pDCOMiCtrl );

    if ( pDCOMiCtrl == NULL )
    {
      DCOMCtrlRequest.ErrCode_en50    = ERRCODE_DCOM_MAX_DCOMIS_REACHED;
      DCOMCtrlRequest.APIRetCode_en50 = rc;
      DCOMCtrlRequest.Ok_en50         = false;
    }
    else
      en35_init_dcomi_ctrl ( pDCOMiCtrl );

    break;

  case UKT_REL_EN35:

    for ( DCOMiNo  = 0; 
          DCOMiNo < KGS-> ulNumOfDCOMis;
          DCOMiNo++ )
    {
      pTmpDCOMiCtrl = KGS-> pFirstDCOMiCtrl + DCOMiNo;
      
      if ( pTmpDCOMiCtrl->state == KT_INITIALIZED ) 
      {
        pDCOMiCtrl = pTmpDCOMiCtrl;
        en35_init_dcomi_ctrl ( pDCOMiCtrl );
        break;
      }
    }

    if ( pDCOMiCtrl == NULL )
    {
      MSGCD(( ERR_MAX_DCOMIS_REACHED ));
      
      DCOMCtrlRequest.ErrCode_en50    = ERRCODE_DCOM_MAX_DCOMIS_REACHED;
      DCOMCtrlRequest.APIRetCode_en50 = NO_ERROR;
      DCOMCtrlRequest.Ok_en50         = false;
    }
    break;
  }
   
  return pDCOMiCtrl;
}

/*---------------------------------------------------------------------------*/

static void en35_ReleaseDCOMiCtrl ( tsp00_Uint4                     Relationship,
                                     ten50_DCOMiCtrl          **ppDCOMiCtrl )
{
  ROUTINE_DBG_MSP00 ("en35_ReleaseDCOMiCtrl");
  
  
  switch ( Relationship )
  {
  case TASK_REL_EN35 :
    en35_FreeDCOMiCtrl ( *ppDCOMiCtrl );
    break;

  case UKT_REL_EN35  :
    // --- init the dcomi control record
    en35_init_dcomi_ctrl ( *ppDCOMiCtrl );
    break;
  }

  *ppDCOMiCtrl = NULL; 
  
  return;
}

/*---------------------------------------------------------------------------*/
extern "C" {
typedef void *(*ThreadFunc)(void *);
};

static bool en35_DCOMiCreate ( struct DOUBLY_LINKED   *pRequest )
{
  ROUTINE_DBG_MSP00 ("en35_DCOMiCreate");
  tsp00_Int4                       exit_code;
  teo07_ThreadErr                  err;
  tsp00_ErrTextc                   threadingErrorText;
  ten50_DCOMiCtrl                 *pDCOMiCtrl;
  ten50_DlqDCOMCtrlReqArgs &DCOMCtrlRequest  = pRequest->args.DCOMCtrlRequest;
  struct TASK_TYPE                *pTaskCtrl = pRequest->taskid;
  ten50_UKT_Control               *this_ukt  = pTaskCtrl->ukt;

  DCOMCtrlRequest.Ok_en50 = true;

  pDCOMiCtrl = en35_GetDCOMiCtrl ( pRequest );
  if ( pDCOMiCtrl == NULL )
  {
    // Resource problem
    DCOMCtrlRequest.ErrCode_en50    = ERRCODE_DCOM_CREATE_DCOMI;
    DCOMCtrlRequest.Ok_en50         = false;
    DCOMCtrlRequest.APIRetCode_en50 = NO_ERROR;
    return ( false );
  }
 
  pDCOMiCtrl->Relationship_en50 = DCOMCtrlRequest.Relationship_en50;

  volatile ten50_DCOMiCtrl *pTestDCOMiCtrl = (volatile ten50_DCOMiCtrl *)pDCOMiCtrl;

  pTestDCOMiCtrl->tid = 0;

  // --- create DCOMi-Thread
#ifdef LINUX
  sqlbeginthread( 0, /* PTS 1106187 */
#else
  sqlbeginthread( KT_DCOMi_STACK_SIZE,
#endif
                  (ThreadFunc)en35_DCOMi,
		  pDCOMiCtrl,
		  THR_CREATE_SUSPENDED_EO07,
		  &pDCOMiCtrl->thread,
		  threadingErrorText,
		  &err );
  if ( err == THR_OK_EO07 )
  {
    sqlresumethread(pDCOMiCtrl->thread, threadingErrorText, &err);

    while ( pTestDCOMiCtrl->tid == 0 )
    {
      sqlyieldthread();
    }

    if ( pTestDCOMiCtrl->semid < 0 )
    {
      sqljointhread( pDCOMiCtrl->thread, &exit_code, threadingErrorText, &err );
      DCOMCtrlRequest.Ok_en50         = false;
    }
  }
  else
  {
      MSGALL((ERR_CREATING_THREAD, "DCOMi", err));
      DCOMCtrlRequest.Ok_en50         = false;
  }

  if ( DCOMCtrlRequest.Ok_en50 )
  {
    while ( pTestDCOMiCtrl->state == KT_INITIALIZED )
    {
      sqlyieldthread();
    }

    if ( pDCOMiCtrl->state != KT_TERMINATING )
    {
      if ( DCOMCtrlRequest.Relationship_en50 == UKT_REL_EN35 )
      {
        this_ukt->pDCOMiCtrl  = pDCOMiCtrl;
      }
      else
      {
        pTaskCtrl->pDCOMiCtrl = pDCOMiCtrl;
      }
    }
    else
    {
      sqljointhread( pDCOMiCtrl->thread, &exit_code, threadingErrorText, &err );
      DCOMCtrlRequest.Ok_en50         = false;
    }
  }

  if ( !DCOMCtrlRequest.Ok_en50 )
  {
    en35_ReleaseDCOMiCtrl(pDCOMiCtrl->Relationship_en50, &pDCOMiCtrl);
    DCOMCtrlRequest.ErrCode_en50    = ERRCODE_DCOM_CREATE_DCOMI;
    DCOMCtrlRequest.APIRetCode_en50 = NO_ERROR;
  }

  return DCOMCtrlRequest.Ok_en50;
}

/*---------------------------------------------------------------------------*/

static bool en35_DCOMiRemove ( struct DOUBLY_LINKED    *pRequest )
{
  ROUTINE_DBG_MSP00 ("en35_DCOMiRemove");
  tsp00_Int4      exit_code;
  teo07_ThreadErr err;
  tsp00_ErrTextc  threadingErrorText;

  ten50_DCOMiCtrl         **ppDCOMiCtrl     = NULL;
  ten50_DlqDCOMCtrlReqArgs &DCOMCtrlRequest = pRequest->args.DCOMCtrlRequest;

  if ( DCOMCtrlRequest.Relationship_en50 == UKT_REL_EN35 )
  {
    DCOMCtrlRequest.Ok_en50         = false;
    DCOMCtrlRequest.ErrCode_en50    = ERRCODE_DCOM_REM_UKT_DCOMi_NOT_ALLOWED;
    DCOMCtrlRequest.APIRetCode_en50 = NO_ERROR;
    return false;
  }
  else
  {
    ppDCOMiCtrl = &pRequest->taskid->pDCOMiCtrl;
  }

  //
  // --- wait until the DCOMi has stop the activities
  //
  while (( (*ppDCOMiCtrl)->state != KT_SLEEPING ) ||
         ( (*ppDCOMiCtrl)->pWorkQue_en50 != (*ppDCOMiCtrl)->pReqQue_en50 ))
  {
    DBG3 (( MF__, "state       = %d",   (*ppDCOMiCtrl)->state ));
    DBG3 (( MF__, "pWorkQue    = 0x%x", (*ppDCOMiCtrl)->pWorkQue_en50 ));
    DBG3 (( MF__, "pDCOMReqQue = 0x%x", (*ppDCOMiCtrl)->pReqQue_en50 ));
    SLEEP( 100 );
  }

  // --- send request to DCOMi
  en35_ReqToDCOMi( *ppDCOMiCtrl, pRequest );

  sqljointhread( (*ppDCOMiCtrl)->thread, &exit_code, threadingErrorText, &err );
  if ( err != THR_OK_EO07 )
  {
    DBG3 (( MF__, "join failed = %s",   threadingErrorText ));
  }

  DCOMCtrlRequest.Ok_en50         = true;
  DCOMCtrlRequest.APIRetCode_en50 = exit_code;

  return true;
}

/*---------------------------------------------------------------------------*/

static void en35_DCOMiTerminate ( ten50_DCOMiCtrl          *pDCOMiCtrl,
                                  ten50_DlqDCOMCtrlReqArgs *DCOMCtrlRequest )
{
  ROUTINE_DBG_MSP00 ("en35_DCOMiTerminate");

  tsp00_Int4      rc = NO_ERROR;

  pDCOMiCtrl->state = KT_TERMINATING;
  
  (void)sql41_remove_sem( &pDCOMiCtrl->semid, (char *)"DEVi", (char *)KGS-> serverdb );

  en35_ReleaseDCOMiCtrl ( pDCOMiCtrl->Relationship_en50, &pDCOMiCtrl );

  return;
}

/*---------------------------------------------------------------------------*/

static void *en35_DCOMi ( void *arg )
{
  ROUTINE_DBG_MSP00 ("en35_DCOMi");
  
  ten50_DCOMiCtrl               *pDCOMiCtrl = (ten50_DCOMiCtrl *)arg;
  tsp00_Int4                    rc         = NO_ERROR;
  bool                          Terminate = false;
  struct DOUBLY_LINKED          *pRequest;
  teo07_ThreadErr               err;
  
  en81_Set_KT_TLS ( arg );

  // --- Create event semaphore (SET) for DCOMi
  rc = sql41_create_sem ( en81_common_rte_key, sqlgetthreadid() ,
		          KGS-> ulNumOfDCOMis, 0660 , getuid() , (char *)"db" , 
		          (char *)KGS-> serverdb );
  pDCOMiCtrl->semid = rc;
  if ( rc < 0 )
  {
     MSGD (( ERR_DCOMI_NO_SEMAPHORE, (long)pDCOMiCtrl->tid ));
  }
  pDCOMiCtrl->tid   = sqlgetthreadid();

  if ( rc > 0 )
  {
    try {
	  if ( eo38CoInitialize() )
	  {
		MSGD (( INFO_KERNEL_THREAD_STARTED, "DCOMi" ));
		
		DBG3 (( MF__, "DCOMi TID:%d started", pDCOMiCtrl->tid ));
		
		// --- set thread control info
		pDCOMiCtrl->state      = KT_RUNNING;
		pDCOMiCtrl->start_time = time((time_t *)0);
		
		// --- Runs until closed
		while ( Terminate == false )
		{
		  pRequest = en35_DCOMReceive ( REINTERPRET_CAST(ten50_DCOM0Ctrl *,pDCOMiCtrl) );
		  
		  switch ( pRequest->req_type )
		  {
			
		  case REQ_DCOM_PARAMINFO_EO00K :
			eo36ParamInfo ( (tsp_dcom_paraminfo_parms*)
											  pRequest->args.DCOMRequest.pParam_en50 );
			en35_ReqFromDCOMxToUKT(pRequest); // -- send reply back to UKT
			break;
			
		  case REQ_DCOM_CO_CREATE_INSTANCE_EO00K :
			eo38CoCreateInstance ( (tsp_dcom_co_create_instance_parms*)
									 pRequest->args.DCOMRequest.pParam_en50 );
			en35_ReqFromDCOMxToUKT(pRequest); // -- send reply back to UKT
			break;
			
		  case REQ_DCOM_RELEASE_INSTANCE_EO00K:
			eo38ReleaseInstance ( (tsp_dcom_release_instance_parms*)
									pRequest->args.DCOMRequest.pParam_en50 );
			en35_ReqFromDCOMxToUKT(pRequest); // -- send reply back to UKT
			break;
			
		  case REQ_DCOM_RELASE_ALL_EO00K:
			eo38VDcomReleaseAll (NULL);
			en35_ReqFromDCOMxToUKT(pRequest); // -- send reply back to UKT
			break;
			
		  case REQ_DCOM_REMOVE_DCOMI_EO00K:
			en35_DCOMiTerminate ( pDCOMiCtrl, &(pRequest->args.DCOMCtrlRequest));
			Terminate = true;                // --- exit for loop
			break;
			
		  case REQ_DCOM_CO_GET_CLASSOBJECT_EO00K :
			eo38CoGetClassObject ( (tsp_dcom_co_create_instance_parms*) 
									 pRequest->args.DCOMRequest.pParam_en50 );
			en35_ReqFromDCOMxToUKT(pRequest); // -- send reply back to UKT
			break;
			
		  default          :
			MSGCD (( ERR_ILLEGAL_REQUEST, "DCOMi", pRequest->req_type ));
			en35_ReqFromDCOMxToUKT(pRequest); // -- send reply back to UKT
			break;
		  }
		  DBG3 (( MF__, "next in loop"));
		}
		
		eo38UnInitialize();
		
	  }
          else
          {
            pDCOMiCtrl->state = KT_TERMINATING;
          }
    }
    catch(...)
    {
    }
  }

  MSGD (( INFO_KERNEL_THREAD_STOPPED, "DCOMi" ));

  return NULL;
}

/*---------------------------------------------------------------------------*/

static tsp00_Int4 en35_AllocDCOMiCtrl ( ten50_DCOMiCtrl **ppDCOMiCtrl )
{
  ROUTINE_DBG_MSP00 ("en35_AllocDCOMiCtrl");

  tsp00_Int4              rc             = NO_ERROR;
  void             *pBuff          = NULL;
  teo00_Uint4       NumOfDCOMs     = NUM_OF_DCOM_MXEO35;
  teo00_Uint4       TotalSize;

  TotalSize = sizeof(ten50_DCOMiCtrl) + sizeof(ten50_DCOMQueue) * NumOfDCOMs;

  rc = ALLOC_MEM_EO57 ( &pBuff, TotalSize );

  if ( rc != NO_ERROR_EO001 )
  {
	  MSGD (( INFO_DCOMI_NO_SEMAPHORE ));
      return rc;
  }

  // --- init memory
  memset (pBuff, 0, TotalSize);
  *ppDCOMiCtrl = (ten50_DCOMiCtrl*)pBuff;

  (*ppDCOMiCtrl)->pDCOMs_en50     = REINTERPRET_CAST(ten50_DCOMQueue*,(*ppDCOMiCtrl) + 1);
  (*ppDCOMiCtrl)->NumOfDCOMs_en50 = NumOfDCOMs;

   return rc;
}

/*---------------------------------------------------------------------------*/

static void en35_FreeDCOMiCtrl ( ten50_DCOMiCtrl *pDCOMiCtrl )
{
  ROUTINE_DBG_MSP00 ("en35_FreeDCOMiCtrl");

  FREE_MEM_EO57 ( pDCOMiCtrl );

  return;
}

#undef  MF__
#define MF__ MOD__"en35_init_dcomi_ctrl"
static void  en35_init_dcomi_ctrl ( ten50_DCOMiCtrl *pDCOMiCtrl )
{
  ten50_DCOMQueue *pDCOMs     = pDCOMiCtrl->pDCOMs_en50;
  tsp00_Int4       NumOfDCOMs = pDCOMiCtrl->NumOfDCOMs_en50;

  DBGIN;

  // --- init memory

  memset ( pDCOMiCtrl , 0, sizeof( ten50_DCOMiCtrl )) ;
  pDCOMiCtrl->pDCOMs_en50 = pDCOMs;

  e73_dev_ioc_queue_init ( REINTERPRET_CAST(struct IOC_QUEUE *,pDCOMs), NumOfDCOMs );

  memset ( &(pDCOMiCtrl->DCOMStat_en50), 0, sizeof( teo52_ActionTime )) ;

  pDCOMiCtrl->pDCOMs_en50         = pDCOMs;
  pDCOMiCtrl->NumOfDCOMs_en50     = NumOfDCOMs;

  pDCOMiCtrl->pWorkQue_en50       = pDCOMs;
  pDCOMiCtrl->pReqQue_en50        = pDCOMs;
  INIT_LOCK( pDCOMiCtrl->alone );

  DBGOUT;
}

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/
