/****************************************************************************

  module      : ven80.cpp

  -------------------------------------------------------------------------

  responsible : JoergM

  special area: KERNEL C++ main
  description : C++ Kernel Main
  
  last changed: 1999-08-04  11:20

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1998



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gen81.h"
#include "geo50_0.h"
#include "geo007_1.h"
#include "gen38.h"
#include <stdlib.h>
/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/

/*===========================================================================*
 *  GLOBAL VARIABLES                                                         *
 *===========================================================================*/

/*===========================================================================*
 *  LOCAL VARIABLES                                                          *
 *===========================================================================*/

/*===========================================================================*
 *  LOCAL FUNCTIONS (PROTOTYPES)                                             *
 *===========================================================================*/

/*===========================================================================*
 *  GLOBAL FUNCTIONS (CODE)                                                  *
 *===========================================================================*/

int
main ( int argc , char *argv[] )
{
    exit ( en81_kernel_main ( argc , argv ) ) ;
}

/*----------*/

extern "C" int
en80RunKernel( struct ten50_KT_Control *pMain )
{
    int rc = 111;
#ifndef USE_CATCH_ALL_CODE
    rc = en81RunKernel( pMain );
#else
    try 
    {
        rc = en81RunKernel( pMain );
    }
    catch (...) 
    {
        void *TlsPtr = en81_Get_KT_TLS();
        ten50_KT_Control *pKt = (ten50_KT_Control *)TlsPtr;

        if ( (pKt == NULL) 
          || (KGS == NULL) )
        {
            MSGD(( ERR_UNHANDLED_CPP_EXCEPTION_NO_KT,
                e38_dbstate_name(KGS != NULL ? KGS->state : -1) ));
        }
        else
        {
            ten50_UKT_Control *pUkt = (ten50_UKT_Control *)TlsPtr;

            if ( (pUkt >= KGS->first_kp)
              && (pUkt <= KGS->last_kp ) )
            {
                if ( pUkt->curr_task != NULL )
                {
                    MSGD(( ERR_UNHANDLED_CPP_EXCEPTION_UKT, 
                        e38_dbstate_name(KGS->state),
                        pUkt->tid,
                        pUkt->index,
                        en38KtStateName(pUkt->state),
                        pUkt->curr_task->index,
                        e38_type_name(pUkt->curr_task->type),
                        e38_tskstate_name(pUkt->curr_task->state) ));
                }
                else
                {
                    MSGD(( ERR_UNHANDLED_CPP_EXCEPTION_UKT, 
                        e38_dbstate_name(KGS->state),
                        pUkt->tid,
                        pUkt->index,
                        en38KtStateName(pUkt->state),
                        0,
                        "-",
                        "-" ));
                }
            }
            else
            {
                MSGD(( ERR_UNHANDLED_CPP_EXCEPTION_KT, 
                        e38_dbstate_name(KGS->state),
                        pKt->tid, 
                        en38KtStateName(pKt->state) ));
            }
        }
        en81FinalAction();
    }
#endif
    return rc;
}

/*===========================================================================*
 *  LOCAL FUNCTIONS (CODE)                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/