/* @lastChanged: "1998-10-22  9:25"
 * @filename:   geo00d.h
 * @purpose:    "DCOM Layer global definition file"
 * @release:    7.2.0.0
 * @see:        "-.-"
 * @author:     Ferdinand Flohe (D024849 - Ferdinand.Flohe@SAP-AG.de)
 * @copyright:  "Copyright by SAP AG, 1998"


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/

#ifndef GEO00D
#define GEO00D

#include <stdio.h>
#include "ilivecachesink.h"  /* no check */ 
#include "gco01.h"

#define  GEO00D_DCOM_WRONG_FIRST_PARAM              0x80040200L
#define  GEO00D_DCOM_TOO_MANY_OUT_PARAM             0x80040201L
#define  GEO00D_DCOM_UNSUPPORTED_PARAM_TYPE         0x80040202L
#define  GEO00D_DCOM_MISSING_USERTYPE_GUID          0x80040203L
#define  GEO00D_DCOM_TOO_MANY_INTERFACES            0x80040204L
#define  GEO00D_DCOM_PROGID_NOT_FOUND               0x80040205L
#define  GEO00D_DCOM_INPROCSERVER_NOT_FOUND         0x80040206L
#define  GEO00D_DCOM_LOCALSERVER_NOT_FOUND          0x80040207L
#define  GEO00D_DCOM_METHOD_NOT_FOUND               0x80040208L
#define  GEO00D_DCOM_COGUID_NOT_FOUND_IN_TYPELIB    0x80040209L
#define  GEO00D_DCOM_INSPEC_ROUTINE_NOT_FOUND       0x8004020AL
#define  GEO00D_DCOM_COCLASS_IN_REGISTRY_NOT_FOUND  0x8004020BL
#define  GEO00D_DCOM_ONLY_INPROC_SUPPORTED          0x8004020CL
#define  GEO00D_DCOM_LIBRARY_NOT_LOADABLE           0x8004020DL
#define  GEO00D_DCOM_DLL_MAIN_MISSING               0x8004020EL
#define  GEO00D_DCOM_DLL_MAIN_ERROR                 0x8004020FL
#define  GEO00D_DCOM_NO_COCLASS_OBJECT_FOUND        0x80040210L
#define  GEO00D_DCOM_DBPROC_CRASHED                 0x80040211L
#define  GEO00D_DCOM_BUFFER_LIMIT                   0x80040212L
#define  GEO00D_DCOM_UNEXPECTED_TYPELIB_STRUCTURE   0x80040313L
#define  GEO00D_DCOM_TRANSEND_FAILED                0x80040314L
#define  GEO00D_DCOM_INSTALLATION_PATH_NOT_FOUND    0x80040315L
#define  GEO00D_DCOM_INSTALLATION_NOT_REGISTERED    0x80040316L
#define  GEO00D_DCOM_COULD_NOT_RETRIEVE_PATH_VAL    0x80040317L
#define  GEO00D_DCOM_EXCEPTION_RAISED               0xFFFFFFFFL
#define  GEO00D_DCOM_TERMINATED                     0xFFFFFFFEL


#define  GEO00D_DBO_METHOD  0
#define  GEO00D_DBO_DBPROC  1
#define  GEO00D_DBO_DBFUNC  2

#if defined(WIN32)
#define   STATUS_SETERROR_EXCEPTION          (tsp00_Uint4   )0xC0004711L
#define   STATUS_TERMINATE_EXCEPTION         (tsp00_Uint4   )0xC0004712L
#endif

class CHRESULT_EO00D
{
private:
  IliveCacheSink* m_psession;
  HRESULT         m_hr;
  char            m_ModuleName[64];
  
public:
  CHRESULT_EO00D(tsp00_C4 session,const char* pcModuleName)
  {
    memcpy(&m_psession,&session[0],sizeof(m_psession));
    strcpy(&m_ModuleName[0],pcModuleName);
  }

  bool    succeeded()
  {
    if (SUCCEEDED(m_hr)) return true;
    return false;
  }

  bool    succeeded( const char* s)
  {
    char cVtraceMsg[256];

    if (SUCCEEDED(m_hr)) return true;
    sprintf ( &cVtraceMsg[0], "HRESULT: 0x%X, %s:%.64s", m_hr, m_ModuleName, s );
    m_psession->ConsoleMsg ((tsp00_Int4)strlen(&cVtraceMsg[0]), (unsigned char*)&cVtraceMsg[0] );
    return false;
  }

  bool    failed()
  {
    if (FAILED(m_hr)) return true;
    return false;
  }

  bool    failed( const char* s)
  {
    char cVtraceMsg[256];

    if (FAILED(m_hr)) 
    {
      sprintf ( &cVtraceMsg[0], "HRESULT: 0x%X, %s:%.64s", m_hr, m_ModuleName, s );
      m_psession->ConsoleMsg ((tsp00_Int4)strlen(&cVtraceMsg[0]), (unsigned char*)&cVtraceMsg[0] );
      return true;
    }
    else
      return false;
  }

  bool    noError()
  {
    if (m_hr == NO_ERROR) return true;
    return false;
  }

  bool    noError( const char* s)
  {
    char cVtraceMsg[256];

    if (m_hr == NO_ERROR) return true;
    sprintf ( &cVtraceMsg[0], "HRESULT: 0x%X, %s:%.64s", m_hr, m_ModuleName, s );
    m_psession->ConsoleMsg ((tsp00_Int4)strlen(&cVtraceMsg[0]), (unsigned char*)&cVtraceMsg[0] );
    return false;
  }

  HRESULT hresult()
  {
    return m_hr;
  }

  void    operator= (HRESULT hr)
  {
    m_hr = hr;
  };

};

extern HRESULT
sql36LoadTypeLibFromProgId (BOOLEAN            InProcServ,
                            char*              cProgID,
                            CLSID&             ClsID,
                            HINSTANCE          *lpHInstance,
                            tsp00_C4           session,
                            tsp_c64&           errtext);

#if defined (_WIN64)
 EXTERN_C LONG WINAPI sql38aVdcomCall(
    IUnknown *lpIUnknown,
    PROC  pProc,
    void* pBufAddr,
    long  BufLen,
		void* pfBufAddr,
    long  fBufLen,
    long  ParmStackSize
    );
#else
 typedef int (WINAPI *PROC)();
 EXTERN_C LONG WINAPI sql38aVdcomCall(
    IUnknown *lpIUnknown,
    PROC  pProc,
    void* pBufAddr,
    long  BufLen,
    long  ParmStackSize
    );
#endif  // _WIN64

EXTERN_C PROC WINAPI sql36GetProcAddress(
    HMODULE hModule,
    LPCSTR lpProcName
    );

EXTERN_C BOOL WINAPI sql36FreeLibrary(
    HMODULE hLibModule
    );



#endif  /*GEO00D*/
