/****************************************************************************

  module      : geo50_0.h

  -------------------------------------------------------------------------

  responsible : JoergM

  special area: OS dependend include file for LZU kernel specific structures
  description : Exists only to prevent vmake from complaining too often


  last changed: 2000-12-02  12:41
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1998



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


#ifndef GEO50_0_H
#define GEO50_0_H



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#if defined(_WIN32)
#ifdef KERNEL_LZU
#ifndef KERNEL
#define KERNEL
#endif /* KERNEL */
#endif /* KERNEL_LZU */
#include "vos00kc.h" /* nocheck */
#else
#ifdef KERNEL_LZU
#ifndef USE_KGS
#define USE_KGS
#endif /* KGS */
#endif /* KERNEL_LZU */
#include "gen50.h" /* nocheck */
#endif


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#if defined(_WIN32)

/* Pointer to Kernel Global Section Instance */
#  define KGS         (&kgs)

/* Task Control Block Pointer */
#  define TASK_CTRL              PTASK_CTRL_REC
/* UKT Control Block Pointer */
#  define UKT_CTRL               PUKT_CTRL_REC
/* Queue Element Control Block */
#  define QUEUE_CTRL             PDLQ_REC

/* Task containing UKT ( needs TASK_CTRL ) */
#  define TASK_UKT(_pTaskCtrl)              ((_pTaskCtrl)->pUKT)
/* Task Type ( needs TASK_CTRL ) */
#  define TASK_TYPE(_pTaskCtrl)             ((_pTaskCtrl)->TaskType)
/* Task State ( needs TASK_CTRL ) */
#  define TASK_STATE(_pTaskCtrl)            ((_pTaskCtrl)->TaskState)
/* Task Index ( needs TASK_CTRL ) */
#  define TASK_INDEX(_pTaskCtrl)            ((_pTaskCtrl)->ulTaskIndex)
/* Task Spinlock ( needs TASK_CTRL ) */
#  define TASK_EXCLUSIVE(_pTaskCtrl)       &((_pTaskCtrl)->lExclusive)
/* Task State Reason ( needs TASK_CTRL ) */
#  define TASK_STATE_REASON(_pTaskCtrl)     ((_pTaskCtrl)->ulStateReason)
/* Task Wait Time ( needs TASK_CTRL ) */
#  define TASK_WAIT_TIME(_pTaskCtrl)        ((_pTaskCtrl)->ulWaitTime)
/* Task Resume Count ( needs TASK_CTRL ) */
#  define TASK_RESUME_COUNT(_pTaskCtrl)     ((_pTaskCtrl)->lResumeCount)
/* Task Vsignal State Variable ( needs TASK_CTRL ) */
#  define TASK_VSIGNAL_OK(_pTaskCtrl)       ((_pTaskCtrl)->lVSignalOk)
/* Task Special Queue Element 1 ( needs TASK_CTRL ) */
#  define TASK_SPECIAL_QUEUE_1(_pTaskCtrl) &((_pTaskCtrl)->SpecDLQElem1)

/* Queue Request Element Type ( needs QUEUE_CTRL ) */
#  define QUEUE_REQTYPE(_pQueue)           ((_pQueue)->ulReqType)
/* Queue Requesting Task ( needs QUEUE_CTRL ) */
#  define QUEUE_REQTASK(_pQueue)           ((_pQueue)->pTaskCtrl)
/* (NT only) Target TaskCtrl of current request ( needs UKT Ctrl ) */
#  define QUEUE_REQFROMTASK(_pQueue)          ((_pQueue)->pFromTaskCtrl)

/* UKT current task ( needs UKT_CTRL ) */
#  define UKT_CURRENT_TASK(_pUKT)          ((_pUKT)->pCTask)
/* Current outstanding Request of given UKT ( needs UKT_CTRL ) */
#  define UKT_CURRENT_REQUEST(_pUKT)       ((_pUKT)->pCurrReq)
/* (NT only) Target TaskCtrl of current request ( needs UKT Ctrl ) */
#  define UKT_CURRENT_REQ_FROMTASK(_pUKT)   QUEUE_REQFROMTASK(UKT_CURRENT_REQUEST(_pUKT))

/* Suspend Reasons */
#  define KGS_SUSPEND_REASONS              (KGS->ulSuspendReasons)
/* Current time */
#  define KGS_CURRENT_TIME                 (KGS->ulCurrTime)

/* Kernel Abort */
#  define DO_ABORT                         ABORT()
/* Lock Initialisation ( NT only */
#  define INIT_LOCK(_Exclusive)            CLEARLOCK(_Exclusive)

#else

/* Pointer to Kernel Global Section Instance */
#  define KGS                    kgs

/* Task Control Block Pointer */
#  define TASK_CTRL              struct TASK_TYPE *
/* UKT Control Block Pointer */
#  define UKT_CTRL               ten50_UKT_Control *
/* Queue Element Control Block */
#  define QUEUE_CTRL             struct DOUBLY_LINKED *

/* Task containing UKT ( needs TASK_CTRL ) */
#  define TASK_UKT(_pTaskCtrl)              ((_pTaskCtrl)->ukt)
/* Task Type ( needs TASK_CTRL ) */
#  define TASK_TYPE(_pTaskCtrl)             ((_pTaskCtrl)->type)
/* Task State ( needs TASK_CTRL ) */
#  define TASK_STATE(_pTaskCtrl)            ((_pTaskCtrl)->state)
/* Task Index ( needs TASK_CTRL ) */
#  define TASK_INDEX(_pTaskCtrl)            ((_pTaskCtrl)->index)
/* Task Spinlock ( needs TASK_CTRL ) */
#  define TASK_EXCLUSIVE(_pTaskCtrl)        ((_pTaskCtrl)->exclusive)
/* Task State Reason ( needs TASK_CTRL ) */
#  define TASK_STATE_REASON(_pTaskCtrl)     ((_pTaskCtrl)->state_reason)
/* Task Wait Time ( needs TASK_CTRL ) */
#  define TASK_WAIT_TIME(_pTaskCtrl)        ((_pTaskCtrl)->wait_time)
/* Task Resume Count ( needs TASK_CTRL ) */
#  define TASK_RESUME_COUNT(_pTaskCtrl)     ((_pTaskCtrl)->resume_count)
/* Task Vsignal State Variable ( needs TASK_CTRL ) */
#  define TASK_VSIGNAL_OK(_pTaskCtrl)       ((_pTaskCtrl)->vsignal_ok)
/* Task Special Queue Element 1 ( needs TASK_CTRL ) */
#  define TASK_SPECIAL_QUEUE_1(_pTaskCtrl)  (QUEUE_CTRL)&((_pTaskCtrl)->qa)

/* UKT current task ( needs UKT_CTRL ) */
#  define UKT_CURRENT_TASK(_pUKT)          ((_pUKT)->curr_task)
/* Current outstanding Request of given UKT ( needs UKT_CTRL ) */
#  define UKT_CURRENT_REQUEST(_pUKT)       ((_pUKT)->curr_quu)

/* Queue Request Element Type ( needs QUEUE_CTRL ) */
#  define QUEUE_REQTYPE(_pQueue)           ((_pQueue)->req_type)
/* Queue Requesting Task ( needs QUEUE_CTRL ) */
#  define QUEUE_REQTASK(_pQueue)           ((_pQueue)->taskid)

/* Suspend Reasons */
#  define KGS_SUSPEND_REASONS              (KGS->SuspendReason)
/* Current time */
#  define KGS_CURRENT_TIME                 time( (time_t*)0 )
/* Kernel Abort */
#  define DO_ABORT                         vabort(WRITE_CORE)

#endif

/* Current Task ( Accesses Thread Local Storage ) */
#  define CURR_TASK                   UKT_CURRENT_TASK(THIS_UKT_CTRL)
/* Current Task ( Accesses Thread Local Storage ) */
#  define CURR_TASK_STATE             TASK_STATE(CURR_TASK)
/* Type of current outstanding Request ( needs UKT Ctrl ) */
#  define UKT_CURRENT_REQTYPE(_pUKT)  QUEUE_REQTYPE(UKT_CURRENT_REQUEST(_pUKT))

/* PTS 1106176 */
#define NUMBER_OF_BLOCKIO_RETRIES 3

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/


/*===========================================================================*
 *  FUNCTION PROTOTYPES                                                      *
 *===========================================================================*/


/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/



#endif  /* GEO50_0_H */
