/*!***************************************************************************

  module      : heo51.h

  -------------------------------------------------------------------------

  responsible : JoergM

  special area: kernel_initializations
  description : Kernel Initialization Functions


  last changed: 2000-11-27  18:07
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1999



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


#ifndef HEO51_H
#define HEO51_H



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"
#include "vsp001.h"
#include "vsp002.h"

/*===========================================================================*
 *  FUNCTION PROTOTYPES                                                      *
 *===========================================================================*/

externC void    vinitok      ( /* void */ );

externC void    vinit        ( tsp00_RteCommAddr VAR_VALUE_REF rte_comm_ptr,
						       tsp00_Int2                      len );

externC void    vdebug_break ( tsp00_Int4 i4DebugBreakPosition );

/*!
  Function: vgetuktid
  Description: Retrieve identifier of current running UKT

  Identifier is in range 1...max_ukt, where max_ukt is member 'sp2xp_max_ukt' 
  of tsp200_xparam_rec filled by call to vcold.

  Arguments: uktId [out] place where to store UKT id
 */
externC void    vgetuktid    ( tsp00_Int4 VAR_VALUE_REF uktId );

/*!
  Function: vgetpid
  Description: Retrieve Task Id of current running task
  Arguments: pid [out] place where to store task id
 */
externC void    vgetpid      ( tsp00_TaskId VAR_VALUE_REF pid );

externC void    vgetfirstuserpid ( tsp00_TaskId VAR_VALUE_REF pid );

externC void    vptype       ( tsp00_TaskId                    pid ,
						       tsp2_process_type VAR_VALUE_REF proc_type );

externC void    vversion     ( tsp00_Version VAR_ARRAY_REF kernelversion, 
				               tsp00_Version VAR_ARRAY_REF rteversion );

externC tsp1_utility_startup_Param vutil_startup( /* void */ );

externC void    vrestart     ( tsp00_TaskId                pid , 
						       tsp00_DevName VAR_ARRAY_REF sysdevspace );

externC void    vshutdown    ( tsp00_TaskId             pid ,
						       tsp00_ShutdownMode_Param mode );

externC void    vfinish      ( /* void */ );

externC void    vos          ( tsp00_Os VAR_VALUE_REF os );

#endif  /* HEO51_H */
