/****************************************************************************

  module      : heo52.h

  -------------------------------------------------------------------------

  responsible : FrankH

  special area: vclock      
  description : description ...


  last changed: 1999-10-28  13:20
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1998



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


#ifndef HEO52_H
#define HEO52_H


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/
#include "gsp00.h"
#include "veo00c.h"
#include "geo52.h"

#define EO52_IS_TIME_MEASURE_ACTIVE (KGS->fTimeMeasure)

#define EO52_TAKE_ABS_TIME(_StateStat, _ToMeasure)                            \
          { if ( EO52_IS_TIME_MEASURE_ACTIVE )                                \
              { teo00_Int4  _BegSec, _BegMicroSec, _EndSec, _EndMicroSec;     \
                                                                              \
                vclock ( &_BegSec, &_BegMicroSec );                           \
                _ToMeasure;                                                   \
                vclock ( &_EndSec, &_EndMicroSec );                           \
                EO52_TIME_DIFF(_BegSec         , _BegMicroSec,                \
                               _EndSec         , _EndMicroSec,                \
                               _StateStat.ulSec, _StateStat.ulMicroSec) ;     \
              }                                                               \
            else                                                              \
             _ToMeasure;                                                      \
          }


#define EO52_TAKE_REL_ABS_TIME(_pTaskCtrl, _StateStat, _ToMeasure)            \
          { if ( EO52_IS_TIME_MEASURE_ACTIVE )                                \
              { teo00_Int4  _BegSec, _BegMicroSec, _EndSec, _EndMicroSec;     \
                                                                              \
                vclock ( &_BegSec, &_BegMicroSec );                           \
                _ToMeasure;                                                   \
                vclock ( &_EndSec, &_EndMicroSec );                           \
                EO52_TIME_DIFF(_BegSec         , _BegMicroSec,                \
                          _EndSec         , _EndMicroSec,                     \
                          _StateStat.ulSec, _StateStat.ulMicroSec) ;          \
                EO52_TIME_DIFF(_BegSec            , _BegMicroSec,             \
                          _pTaskCtrl->TaskStateStat.iRelStopSec,              \
                          _pTaskCtrl->TaskStateStat.iRelStopMicroSec,         \
                          _StateStat.ulRelSec, _StateStat.ulRelMicroSec) ;    \
              }                                                               \
            else                                                              \
             _ToMeasure;                                                      \
          }

#define EO52_TAKE_UKT_RUNTIME(_RunStartSec, _RunStartMicroSec, _RunStat, _IOWaitStat , _WaitingForIO, _Do) \
  { _RunStat.ulCount++;                                               \
    if ( EO52_IS_TIME_MEASURE_ACTIVE )                                \
      { int  _ActSec, _ActMicroSec;                                   \
                                                                      \
        vclock ( &_ActSec, &_ActMicroSec );                           \
        if ( _RunStartSec )      /* only if start set */              \
          EO52_TIME_DIFF(_RunStartSec, _RunStartMicroSec,             \
                    _ActSec     , _ActMicroSec,                       \
                    _RunStat.ulSec, _RunStat.ulMicroSec) ;            \
        _Do;                                                          \
        vclock ( &_RunStartSec, &_RunStartMicroSec );                 \
        if ( _WaitingForIO )                                          \
        { _IOWaitStat.ulCount++;                                      \
          EO52_TIME_DIFF(_ActSec          , _ActMicroSec,             \
                    _RunStartSec     , _RunStartMicroSec,             \
                    _IOWaitStat.ulSec, _IOWaitStat.ulMicroSec) ;      \
        }                                                             \
    }                                                                 \
    else                                                              \
      _Do;                                                            \
  }

/*===========================================================================*
 *  FUNCTION PROTOTYPES                                                      *
 *===========================================================================*/

externC void
vclock            ( teo00_Int4      *Sec,
                    teo00_Int4      *MicroSec);

externC void
vtimestamp        ( tsp00_Timestamp VAR_ARRAY_REF pStamp);

externC void
vUTCtimestamp     ( tsp00_Timestamp VAR_ARRAY_REF pStamp);

externC void
vdattime          ( tsp00_Date VAR_ARRAY_REF pDate,
                    tsp00_Time VAR_ARRAY_REF pTime );

/*
  Function: vTimeZoneDelta
  Description: Return difference between local time and GMT in hours

  Berlin winter time result 1.

  Berlin summer time result 2.

  New Zealand Chatham Islands result 12,75 !!!!

  NOTE: If weekday difference between localtime and systemtime is larger
  than a day, 0. is returned and a warning message is issued in knldiag.

  Return value: Difference in hours
 */
externC tsp00_Longreal vTimeZoneDelta();
#endif  /* HEO52_H */
