/*!**************************************************************************

  module       : heo57.h

  -------------------------------------------------------------------------

  responsible : JoergM

  special area: Kernel Memory
  description : Kernel Memory Allocation routines

  first created:2000-11-27  20:38
  last changed: 2001-01-02  12:39
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/
#ifndef HEO57_H
#define HEO57_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"

/*===========================================================================*
 *  FUNCTION PROTOTYPES                                                      *
 *===========================================================================*/

externC void                   vallocat         ( tsp00_Int4      length,
												  void          **objptr,
												  tsp00_Bool     *ok);

externC void                   valloc_convcache ( tsp00_Int4      count,
												  tsp00_Int4     *available,
												  tsp00_PageAddr *bufptr,
												  tsp00_Bool     *ok );

externC void                   valloc_datacache ( tsp00_Int4      count,
												  tsp00_Int4     *available,
												  tsp00_PageAddr *bufptr,
												  tsp00_Bool     *ok );


externC void                   vmalloc          ( tsp00_Int4      length,
												  void          **objptr,
												  tsp00_Bool     *ok );

externC void                   vmfree           ( void           *objptr );

externC void                   vmnewbuf         ( tsp00_Int4      count,
												  tsp00_BufAddr  *bufptr,
												  tsp00_Bool     *ok );

externC void                   vmfreebuf        ( tsp00_PageAddr  bufptr );

externC void                   vnewbuf          ( tsp00_Int4      count,
												  tsp00_Int4     *available,
												  tsp00_PageAddr *bufptr,
												  tsp00_Bool     *ok );

externC void                   vnewtracebuf     ( tsp00_Int4     *available,
												  tsp00_PageAddr *bufptr0,
												  tsp00_PageAddr *bufptr1,
												  tsp00_Bool     *ok );

/* PTS 1108768 Space for Tracebuffer used for InMemory tracing */
/*!
  Function: vGetUktTraceBuffer
  Description: Get UKT specific tracebuffer

  For each UKT a tracebuffer is allocated by the LZU. The pointer to the first page in this buffer
  and the total size in pages of this buffer is returned.

  The given uktId is either the current Ukt (retrieved by a call to vgetuktid) or any valid ukt id
  in the range form 1...max_ukt. max_ukt was returned by vcold in the tsp200_xparam_rec.

  The memory returned by this call is save to be used for devspace I/O.

  Arguments: uktId [in] The id of the ukt with which the trace buffer is
             firstPage [out] the pointer to the first page
             sizeInPages [out] the number of pages
             ok [out] the indicator, that the call succeeded
 */
externC void                   vGetUktTraceBuffer( tsp00_Int4                   uktId,
                                                   tsp00_PageAddr VAR_VALUE_REF firstPage,
                                                   tsp00_Int4     VAR_VALUE_REF sizeInPages,
                                                   tsp00_Bool     VAR_VALUE_REF ok );

/*!
  Function: vGetTraceInfoPage
  Description: Get common tracebuffer info page and total trace buffer size

  There is a trace buffer for each UKT. This page is used to save the individual size
  information for these buffers. The totalSizeInPages can be used to dump the complete
  trace buffer pages with a single I/O.

  Arguments: infoPage [out] the pointer to the info page
             totalSizeInPages [out] the total size (including info page) in pages
 */
externC void vGetTraceInfoPage( tsp00_PageAddr VAR_VALUE_REF infoPage, 
                                tsp00_Int4 VAR_VALUE_REF totalSizeInPages );

#endif
