/*!***************************************************************************

  module      : heo69.h

  -------------------------------------------------------------------------

  responsible : JoergM

  special area: Accounting
  description : Accounting functions (Stack,Timing,IOs etc.)
  
  last changed: 1999-10-23  2:27
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1999



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


#ifndef HEO69_H
#define HEO69_H



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"

/*===========================================================================*
 *  FUNCTION PROTOTYPES                                                      *
 *===========================================================================*/

/*!
  Function:     vcinit
  description:  Reset Millisecond counter

	This call gets the current number of ticks used in subsequent
	calls (vccheck) to get the number of milliseconds needed.
	User and System times are added to a total time.
  arguments: 
	taskid [in] TaskId (only used in kernel RTE)
  return value: none
 */
externC void vcinit ( tsp00_TaskId  taskid );

/*!
  Function:     vccheck
  description:  Get number of milliseconds since last call to vcinit()

	This call gets the current number of ticks and returns the number
	of milliseconds used since last call to vcinit().
	User and System times are added to a total time.
  arguments: 
	taskid [in] TaskId (only used in kernel RTE)
	time   [out] Number of milliseconds since last vcinit
  return value: none
 */
externC void vccheck ( tsp00_TaskId  taskid,
					   tsp00_Int4    VAR_VALUE_REF time );

/*!
  Function:     vioinit
  description:  Reset task specific IO counter

	This call resets the global IO counter for the specific task
  arguments: 
	taskid [in] TaskId (only used in kernel RTE)
  return value: none
 */
externC void vioinit ( tsp00_TaskId  taskid );

/*!
  Function:     viocheck
  description:  Get direct IO counter since last call to vioinit()

	This call gets the global direct IO count for the specific task

  arguments: 
	taskid [in] TaskId (only used in kernel RTE)
	no   [out] Number of direct IO calls
  return value: none
 */
externC void viocheck ( tsp00_TaskId  taskid ,
						tsp00_Int4    VAR_VALUE_REF no );

/*!
  Function:     vmonitor
  description:  Get task specific counter

	This call gets the counter for the specific task.

  arguments: 
	taskid [in] TaskId (only used in kernel RTE)
	phys_ios [out] Sum of SelfIORead/Write and DevIORead/Write
	suspends [out] Number of vsuspend() calls
	waits    [out] Number of vwait() calls
  return value: none
 */
externC void vmonitor ( tsp00_TaskId  taskid ,
						tsp00_Int4    VAR_VALUE_REF phys_ios ,
						tsp00_Int4    VAR_VALUE_REF suspends ,
						tsp00_Int4    VAR_VALUE_REF waits );

/*!
  Function:     vsinit

  description:  Initialization of stack checking code

  Initialization of stack checking code (only needed for Client RTE)
  This function is a dummy for kernel RTE.

  arguments: 
	none
  return value: none
 */
externC void vsinit ( /* void */ );

/*!
  Function:     vscheck

  description:  Returns the number of bytes left on the stack

  This call returns the current number of bytes left on the stack.

  In kernel RTE the current task stack is used.

  arguments: 
	LeftStackSize [out] Number of bytes left on stack

  return value: none
 */
externC void vscheck ( tsp00_Int4 VAR_VALUE_REF LeftStackSize );

/*!
  Function:     vscheck

  description:  Fill stack with pattern

  Fill the remaning stack with a pattern.
  Leave the two-word chgstack()-test-pattern intact.
  Current function's additional stack requirements
  estimated as eight words.

  In kernel RTE the current task stack is used.

  arguments: 
    none

  return value: none
 */
externC void vsfill ( /* void */ );

/*!
  Function:     vsscan

  description:  Return number of bytes left since last call to vsfill()

  This call uses the first difference from the pattern setup with last
  vsfill() call. This difference is used to return the maximum number
  of bytes used on the staqck.

  In kernel RTE the current task stack is used.

  arguments: 
	MaxStackSize [out] Number of bytes left on stack

  return value: none
 */
externC void vsscan ( tsp00_Int4 VAR_VALUE_REF MaxStackSize );

/*!
  Function:     vsoverflow

  description:  Calls vabort of stack overflow was detected
  
  This call uses the vsscan() function and compares the result with
  the RTE internal known task stack size. If overflow occured, a
  vabort() call is done ( and a Stack Overflow message written to knldiag).

  arguments: 
    none

  return value: 
    none
 */
#if COMPILEMODE_MSP00 >= SLOW_MSP00	
#define SECURITY_BUFFER_MXEO69 sizeof(tsp00_Page)
#else
#define SECURITY_BUFFER_MXEO69 0
#endif
externC void vsoverflow ( /* void */ );

#endif  /* HEO69_H */
