/****************************************************************************

  module      : veo11.c

  -------------------------------------------------------------------------

  responsible : FrankH

  special area: paramfile access
  description : general xparam access routines ...


  last changed: 2000-12-01  11:22
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1998



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>

#include "hsp100.h"
#include "heo11.h"
#include "heo11_1.h"
#include "heo46.h"
#include "heo47.h"
#include "geo60.h"

#if defined (_WIN32)
# include <windows.h>
# include "gos44.h" /* nocheck */
#endif


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/
#define MOD__ "veo11.c :"


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/***************************************************************/
#if defined(_WIN32) || defined(OS2)
/***************************************************************/
bool    os44Filecopy ( PSZ               pszSrcName,
                       PSZ               pszDestName,
                       tsp01_RteError    *RteError );


#define FILECOPY(_Src,_Dest,_RteOSErr)  (os44Filecopy(_Src,_Dest,_RteOSErr) )

/***************************************************************/
#else
/***************************************************************/
#define FILECOPY(_Src,_Dest,_RteOSErr)  (e541_filecopy(_Src,_Dest,_RteOSErr))

/***************************************************************/
#endif
/***************************************************************/


/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/



/*===========================================================================*
 *  GLOBAL VARIABLES                                                         *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL VARIABLES                                                          *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL FUNCTIONS (PROTOTYPES)                                             *
 *===========================================================================*/
static tsp00_Bool eo11BuildParamFileNameWithPath ( tsp00_DbNamec   DBNameC,
                                                   tsp00_Int4      ParamVersion,
                                                   tsp00_Pathc     ParamFileNameWithPath,
                                                   tsp01_RteError *RteError);

static tsp00_Bool eo11CheckFileExists ( tsp00_Pathc FileToCheck );

static tsp00_Bool eo11DeleteFile ( tsp00_Pathc FileToDelete,
                                   tsp01_RteError  *RteError) ;

static tsp00_Bool eo11SaveParamFileVersion ( tsp00_DbNamec  VAR_ARRAY_REF  DBNameC,
                                             tsp00_Int4                    VersionToSave ,
                                             tsp01_RteError               *RteError)  ;

static void eo11CreateTimeStamp ( tsp00_C64  VAR_ARRAY_REF  TimeStamp ) ;

/*===========================================================================*
 *  GLOBAL FUNCTIONS (CODE)                                                  *
 *===========================================================================*/
externC void  sqlxpopenxparam ( tsp00_DbName     acServerDB,
                                tsp00_Int4       MaxParams,
  		  	   		            tsp00_Int4      *pXparamId,
  						        tsp00_ErrText    acErrorText,
                                tsp00_BoolAddr   pbOk      )
{
    #undef  MF__
    #define MF__ MOD__"sqlxpopenxparam"

    tsp00_DbNamec      DBNameC;
    tsp11_XpReturnCode XpRet ;
    tsp01_RteError     RteError;

    DBGIN;

    /* convert pascal string to c string and truncate database name */
    eo46PtoC ( DBNameC, acServerDB , sizeof (tsp00_DbName) );

    DBG3 (( MF__, "DBNameC = %s", DBNameC ));

    XpRet = sqlXparamOpen ( DBNameC, pXparamId,
                            XP_OPEN_CREATE_SP11 | XP_OPEN_REC_SP11 | XP_OPEN_BUFFERED_SP11, 
                            &RteError ) ;

    *pbOk = XpRet == xp_ok_sp11 ;

    if ( *pbOk )
      memset (( char* ) acErrorText, ' ', sizeof ( tsp00_ErrText ));
    else
      eo46CtoP ( acErrorText, RteError.RteErrText , sizeof (tsp00_ErrText) );

    DBGOUT;
    return;
}

/*==========================================================================*/

externC void    sqlxpclosexparam    (   tsp00_Int4     XparamId,
                                        tsp00_Bool     fFlush, 
                                        tsp00_ErrText  acErrorText,
                                        tsp00_BoolAddr pbOk            )
{
    #undef  MF__
    #define MF__ MOD__"sqlxpclosexparam"

    DBGIN;

    *pbOk = true;

    sql11c_CloseXParam ( XparamId, fFlush, acErrorText, pbOk );

    DBGOUT;
}

/*==========================================================================*/

externC void    sqldelxparam    (   tsp00_Int4           XparamId,
                                    tsp11_XpKeyType      acXParamKey,
                                    tsp00_ErrText        acErrorText,
                                    tsp_xp_return_code  *pucXParamReturnCode )
{
    #undef  MF__
    #define MF__ MOD__"sqldelxparam"

    tsp11_XpKeyTypec    acXParamKeyC ;

    DBGIN;

    eo46PtoC ( acXParamKeyC, acXParamKey , sizeof (tsp11_XpKeyType) );

    sql11c_DeleteXParam ( XparamId, acXParamKeyC,
                          acErrorText, pucXParamReturnCode );

    DBGOUT;
}

/*==========================================================================*/

externC void sqlgnextxparam ( 	tsp00_Int4				XparamId,
	                            tsp11_XpKeyType 		acXParamKey,			
                             	tsp11_XpKeyType 		acXParamNextKey,		
                            	tsp00_XpValueRec*		prcXParamValue,
                            	tsp00_ErrText 			acErrorText,		
                            	tsp_xp_return_code*		pucXParamReturnCode)
{
    #undef  MF__
    #define MF__ MOD__"sqlgnextxparam"


    tsp11_XpKeyTypec    acXParamKeyC ;
    tsp11_XpKeyTypec    acXParamNextKeyC ;

    DBGIN;

    eo46PtoC ( acXParamKeyC, acXParamKey , sizeof (tsp11_XpKeyType) );

    sql11c_GetNextXParam ( XparamId, acXParamKeyC, acXParamNextKeyC,
                           (XP_PVALUE_REC )prcXParamValue, 
                           acErrorText, pucXParamReturnCode );

    eo46CtoP ( acXParamNextKey, acXParamNextKeyC, sizeof (tsp11_XpKeyType) );

    DBGOUT;
}

/*==========================================================================*/

externC void sqlgetxparam ( tsp00_Int4          XparamId,
                        	tsp11_XpKeyType     acXParamKey,	
	                        tsp00_XpValueRec   *prcXParamValue,
                        	tsp00_ErrText 		acErrorText,
                        	tsp_xp_return_code *pucXParamReturnCode)
{
    #undef  MF__
    #define MF__ MOD__"sqlgetxparam"

    tsp11_XpKeyTypec    acXParamKeyC ;

    DBGIN;

    eo46PtoC ( acXParamKeyC, acXParamKey , sizeof (tsp11_XpKeyType) );
    sql11c_GetXParam ( XparamId, acXParamKeyC,(XP_PVALUE_REC ) prcXParamValue, 
                       acErrorText, pucXParamReturnCode );

    DBGOUT;
}

/*==========================================================================*/

externC void sqlputxparam ( tsp00_Int4         XparamId,
                        	tsp11_XpKeyType     acXParamKey,	
	                        tsp00_XpValueRec   *prcXParamValue,
                        	tsp00_ErrText 		acErrorText,
                        	tsp_xp_return_code *pucXParamReturnCode)
{
    #undef  MF__
    #define MF__ MOD__"sqlputxparam"

    tsp11_XpKeyTypec    acXParamKeyC ;


    DBGIN;

    eo46PtoC ( acXParamKeyC, acXParamKey , sizeof (tsp11_XpKeyType) );
    sql11c_PutXParam ( XparamId, acXParamKeyC, (XP_PVALUE_REC )prcXParamValue, 
                       acErrorText, pucXParamReturnCode );

    DBGOUT;
}

/*==========================================================================*/
externC tsp11_XpReturnCode 
        sqlXparamOpen   ( tsp00_DbNamec       DBNameC,
    	 		          tsp00_Int4         *pXparamId,
		  	 		      tsp00_Int4          OpenMode,
                          tsp01_RteError     *RteError )
{
  return(sqlXparamOpenVersion(DBNameC, XP_ACT_VERSION_SP11, pXparamId, OpenMode, RteError )) ;
}

/*==========================================================================*/

externC tsp11_XpReturnCode 
        sqlXparamOpenVersion   ( tsp00_DbNamec       DBNameC,
                                 tsp00_Int4          VersionToOpen,
		  	 		             tsp00_Int4         *pXparamId,
		  	 		             tsp00_Int4          OpenMode,
                                 tsp01_RteError     *RteError )
{
  tsp00_Bool         bOk;
  tsp11_XpReturnCode XpRet ;
  tsp00_Int4         DefaultMode = 0 ;
  tsp00_ErrText      acErrorText ; 
  tsp00_Pathc        ParamFile ;

  eo46_rte_error_init ( RteError ) ;

  if ( !(OpenMode & XP_OPEN_BIN_SP11) &&
       !(OpenMode & XP_OPEN_REC_SP11) )
    DefaultMode |= XP_OPEN_REC_SP11 ;

  if ( !(OpenMode & XP_OPEN_BUFFERED_SP11) &&
       !(OpenMode & XP_OPEN_DIRECT_SP11) )
    if (OpenMode & XP_OPEN_READONLY_SP11)
      DefaultMode |= XP_OPEN_DIRECT_SP11 ;
    else
      DefaultMode |= XP_OPEN_BUFFERED_SP11 ;

  bOk = eo11BuildParamFileNameWithPath (DBNameC, VersionToOpen, ParamFile, RteError ) ;

  if ( bOk )
  {
    DBG3 (( MF__, "ParamFile = %s", ParamFile ));
    sql11c_OpenXParam ( ParamFile, pXparamId, acErrorText, (DefaultMode | OpenMode), &bOk );
	if ( !bOk )
	{
      RteError->RteErrCode = RTE_ERROR_SP01 ;
      eo46PtoC ( RteError->RteErrText, acErrorText , sizeof (tsp00_ErrText) );
	}
  }

  XpRet = bOk ? xp_ok_sp11 : xp_not_ok_sp11 ;

  return ( XpRet ) ;
}


/*==========================================================================*/

externC tsp11_XpReturnCode 
        sqlXparamClose  ( tsp00_Int4          XparamId,
		  			      tsp00_Bool          fFlush,
                          tsp01_RteError     *RteError )
{
  tsp00_Bool         bOk;
  tsp00_ErrText      acErrorText ; 
  tsp11_XpReturnCode XpRet ;

  eo46_rte_error_init ( RteError ) ;

  sql11c_CloseXParam ( XparamId, fFlush, acErrorText, &bOk );

  if ( bOk )
  {
    XpRet = xp_ok_sp11 ;
  }
  else
  {
    XpRet = xp_not_ok_sp11 ;
    RteError->RteErrCode = RTE_ERROR_SP01 ;
    eo46PtoC ( RteError->RteErrText, acErrorText , sizeof (tsp00_ErrText) );
  }

  return ( XpRet ) ;
}

/*==========================================================================*/

externC tsp11_XpReturnCode 
        sqlXparamPut    ( tsp00_Int4          XparamId,
					      tsp11_XpKeyTypec    acXParamKeyC,
						  tsp00_XpValueRec   *prcXParamValue,
                          tsp01_RteError     *RteError )
{
  tsp00_ErrText      acErrorText ; 
  tsp11_XpReturnCode XpRet ;
  tsp_xp_return_code OldXParamReturnCode ;

  eo46_rte_error_init ( RteError ) ;

  sql11c_PutXParam ( XparamId, acXParamKeyC, (XP_PVALUE_REC )prcXParamValue, 
                     acErrorText, &OldXParamReturnCode );

  XpRet = OldXParamReturnCode;
  if ( XpRet != xp_ok_sp11 )
  {
    RteError->RteErrCode = RTE_ERROR_SP01 ;
    eo46PtoC ( RteError->RteErrText, acErrorText , sizeof (tsp00_ErrText) );
  }

  return ( XpRet ) ;
}

/*==========================================================================*/

externC tsp11_XpReturnCode 
        sqlXparamGet    ( tsp00_Int4          XparamId,
				          tsp11_XpKeyTypec    acXParamKeyC,
					      tsp00_XpValueRec   *prcXParamValue,
                          tsp01_RteError     *RteError )
{
  tsp00_ErrText      acErrorText ; 
  tsp11_XpReturnCode XpRet ;
  tsp_xp_return_code OldXParamReturnCode ;

  eo46_rte_error_init ( RteError ) ;

  sql11c_GetXParam ( XparamId, acXParamKeyC,(XP_PVALUE_REC ) prcXParamValue, 
                     acErrorText, &OldXParamReturnCode );

  XpRet = OldXParamReturnCode;

  if ( XpRet != xp_ok_sp11 )
  {
    RteError->RteErrCode = RTE_ERROR_SP01 ;
    eo46PtoC ( RteError->RteErrText, acErrorText , sizeof (tsp00_ErrText) );
  }

  return ( XpRet ) ;
}

/*==========================================================================*/

externC tsp11_XpReturnCode 
        sqlXparamGetNext( tsp00_Int4          XparamId,
				 	      tsp11_XpKeyTypec    acXParamKeyC,
						  tsp11_XpKeyTypec    acXParamNextKeyC,
						  tsp00_XpValueRec   *prcXParamValue,
                          tsp01_RteError     *RteError )
{
  tsp00_ErrText      acErrorText ; 
  tsp11_XpReturnCode XpRet ;
  tsp_xp_return_code OldXParamReturnCode ;

  eo46_rte_error_init ( RteError ) ;

  sql11c_GetNextXParam ( XparamId, acXParamKeyC, acXParamNextKeyC,
                         (XP_PVALUE_REC )prcXParamValue, 
                         acErrorText, &OldXParamReturnCode );

  XpRet = OldXParamReturnCode;

  if ( XpRet != xp_ok_sp11 )
  {
    RteError->RteErrCode = RTE_ERROR_SP01 ;
    eo46PtoC ( RteError->RteErrText, acErrorText , sizeof (tsp00_ErrText) );
  }

  return ( XpRet ) ;
}

/*==========================================================================*/

externC tsp11_XpReturnCode 
        sqlXparamDel    ( tsp00_Int4          XparamId,
		  		          tsp11_XpKeyTypec    acXParamKeyC,
                          tsp01_RteError     *RteError )

{
  tsp00_ErrText      acErrorText ; 
  tsp11_XpReturnCode XpRet ;
  tsp_xp_return_code OldXParamReturnCode ;

  eo46_rte_error_init ( RteError ) ;

  sql11c_DeleteXParam ( XparamId, acXParamKeyC,
                        acErrorText, &OldXParamReturnCode );

  XpRet = OldXParamReturnCode ;

  if ( XpRet != xp_ok_sp11 )
  {
    RteError->RteErrCode = RTE_ERROR_SP01 ;
    eo46PtoC ( RteError->RteErrText, acErrorText , sizeof (tsp00_ErrText) );
  }

  return ( XpRet ) ;
}

/*==========================================================================*/

externC tsp11_XpReturnCode 
eo11ResetXparamChangedFlag ( tsp00_DbNamec                   DBNameC,
                             tsp01_RteError                 *RteError )

{
  #undef  MF__
  #define MF__ MOD__"eo11ResetXparamChangedFlag"
  tsp11_XpReturnCode xp_ret;
  tsp00_Int4         XparamId ;

  DBGIN;

  xp_ret = sqlXparamOpen (DBNameC, &XparamId , XP_OPEN_DIRECT_SP11, RteError) ;
  if ( xp_ret == xp_ok_sp11 )
    {
      tsp00_Bool         bOk;
      tsp00_ErrText      acErrorText ; 
      sql11c_ResetParamChangedFlag ( XparamId, acErrorText, &bOk ) ;
      if ( bOk )
      {
        /* PTS 1104471 */
        XP_VALUE_REC     XpValue ;

        XpValue.xp_value_t = xp_c64_type_sp11 ;
        eo11CreateTimeStamp(XpValue.xp_type.xp_c64);
        eo46CtoP(XpValue.xp_type.xp_c64, (char *)XpValue.xp_type.xp_c64, sizeof(XpValue.xp_type.xp_c64) );

        /* ignore error, if only the timestamp is missing */
        sqlXparamPut ( XparamId, XP_PARAM_VERSION_TIME, (tsp00_XpValueRec *)&XpValue, RteError ) ;

        sqlXparamClose ( XparamId, XP_CLOSE_FLUSH_SP11, RteError );
      }
      else
      {
        tsp01_RteError   IgnoreCloseError ;
        RteError->RteErrCode = RTE_ERROR_SP01 ;
        eo46PtoC ( RteError->RteErrText, acErrorText , sizeof (tsp00_ErrText) );
        sqlXparamClose ( XparamId, XP_CLOSE_NO_FLUSH_SP11, &IgnoreCloseError );
      }
    }

  DBGOUT;
  return ( xp_ret );
}

/*==========================================================================*/

externC tsp11_XpReturnCode 
        eo11GetParam ( tsp00_DbNamec                   DBNameC ,
                       tsp11_XpKeyTypec                XpKeyC,
                       tsp00_C64c                      XpValueC,
                       tsp01_RteError                 *RteError )
{
  #undef  MF__
  #define MF__ MOD__"eo11GetParam"

  tsp01_RteError     IgnoreCloseError ;
  tsp11_XpReturnCode xp_ret;
  tsp00_Int4         XparamId ;
  XP_VALUE_REC       xp_value;
  
  DBGIN;

  XpValueC[0] = 0 ;

  xp_ret = sqlXparamOpen (DBNameC, &XparamId , XP_OPEN_READONLY_SP11, RteError) ;

  if ( xp_ret == xp_ok_sp11 )
    xp_ret = sqlXparamGet (XparamId, XpKeyC, 
                           (tsp00_XpValueRec *)&xp_value , RteError );

  if ( xp_ret == xp_ok_sp11 )
  {
    switch ( xp_value.xp_value_t )
      {
      case xp_int2_type_sp11   : sql47_ltoa((int )xp_value.xp_type.xp_int2, XpValueC, sizeof(tsp00_C64c) );
                                 break;
      case xp_int4_type_sp11   : sql47_ltoa((int )xp_value.xp_type.xp_int4, XpValueC, sizeof(tsp00_C64c) );
                                 break;
      case xp_c8_type_sp11     : eo46PtoC ( XpValueC , xp_value.xp_type.xp_c8 , 8 );
                                 break;
      case xp_c18_type_sp11    : eo46PtoC ( XpValueC , xp_value.xp_type.xp_c18 ,18 );
                                 break;
      case xp_c24_type_sp11    : eo46PtoC ( XpValueC , xp_value.xp_type.xp_c24 , 24 );
                                 break;
      case xp_c40_type_sp11    : eo46PtoC ( XpValueC , xp_value.xp_type.xp_c40 , 40 );
                                 break;
      case xp_c64_type_sp11    : eo46PtoC ( XpValueC , xp_value.xp_type.xp_c64 , 64 );
                                 break;
      case xp_crypt_type_sp11  : break;
      }
  }

  sqlXparamClose ( XparamId, XP_CLOSE_NO_FLUSH_SP11, &IgnoreCloseError );

  DBGOUT ;

  return ( xp_ret );
}

/*==========================================================================*/

externC tsp11_XpReturnCode 
        eo11PutParam ( tsp00_DbNamec                   DBNameC ,
                       tsp11_XpKeyTypec                XpKeyC,
                       tsp00_XpValueRec               *XpValue,
                       tsp01_RteError                 *RteError )
{
  #undef  MF__
  #define MF__ MOD__"eo11PutParam"

  tsp01_RteError     IgnoreCloseError ;
  tsp11_XpReturnCode xp_ret;
  tsp00_Int4         XparamId ;
  
  DBGIN;

  xp_ret = sqlXparamOpen (DBNameC, &XparamId , XP_OPEN_CREATE_SP11 | XP_OPEN_READ_WRITE_SP11, RteError) ;

  if ( xp_ret == xp_ok_sp11 )
  {
    xp_ret = sqlXparamPut( XparamId, XpKeyC, XpValue, RteError ) ;
    if ( xp_ret == xp_ok_sp11 )
      sqlXparamClose ( XparamId, XP_CLOSE_FLUSH_SP11, RteError );
    else
      sqlXparamClose ( XparamId, XP_CLOSE_NO_FLUSH_SP11, &IgnoreCloseError );
  }

  DBGOUT ;

  return ( xp_ret );
}

/*------------------------------*/
/* PTS 1109311 */
externC tsp11_XpReturnCode 
        eo11PutParamWithoutChgFlag (tsp00_DbNamec                   DBNameC,
                                    tsp11_XpKeyTypec                XpKeyC,
                                    tsp00_XpValueRec               *XpValue,
                                    tsp01_RteError                 *RteError)
{
    tsp11_XpReturnCode xp_ret;
    tsp00_Int4         XparamId ;
    tsp01_RteError     IgnoreCloseError ;


    xp_ret = sqlXparamOpen (DBNameC, &XparamId , XP_OPEN_DIRECT_SP11, RteError) ;
    if ( xp_ret == xp_ok_sp11 )
    {
        xp_ret = sqlXparamPut (XparamId, XpKeyC, XpValue, RteError ) ;

        if ( xp_ret == xp_ok_sp11 )
            sqlXparamClose ( XparamId, XP_CLOSE_FLUSH_SP11, RteError );
        else
            sqlXparamClose ( XparamId, XP_CLOSE_NO_FLUSH_SP11, &IgnoreCloseError );
    }

    return ( xp_ret );
}

/*------------------------------*/

externC tsp00_Bool sqlXParamCheckExists ( tsp00_DbNamec    DBNameC) 
{
  tsp00_Pathc    Paramfile ;
  tsp01_RteError RteError ;
  tsp00_Bool     Ok ;

  Ok = sqlXParamGetParamfileDir ( DBNameC, Paramfile,
                                  TERM_WITH_DELIMITER_EO01, &RteError ) ;
  if ( Ok )
  {
    strcat (Paramfile, DBNameC ) ;
    Ok =  eo11CheckFileExists ( Paramfile ) ;
  }
  return ( Ok );
}

/*------------------------------*/

externC tsp00_Bool sqlXparamFsize ( tsp00_DbNamec    DBNameC,
                                    tsp00_Int4      *Fsize,
                                    tsp01_RteError  *RteError )
{
  #undef  MF__
  #define MF__ MOD__"sqlXparamFsize"

  tsp00_Pathc    Paramfile ;
  tsp00_Bool     Ok ;

  DBGPAS;

  Ok = sqlXParamGetParamfileDir ( DBNameC, Paramfile,
                                  TERM_WITH_DELIMITER_EO01, RteError ) ;
  if ( Ok )
    strcat (Paramfile, DBNameC ) ;

  if ( Ok )
  {
#if defined (_WIN32)
    tsp00_Int      Attl ;
    tsp00_Uint     FSizeLow, FSizeHigh ;
    tsp00_Int4 rc;
    rc = sql44c_get_file_info ( Paramfile, &Attl, &FSizeLow, &FSizeHigh ) ;
    Ok = rc == NO_ERROR ;
    if ( Ok )
      *Fsize = FSizeLow ;
    else
      eo46_set_rte_error( RteError, rc, "", Paramfile ) ;
#else
     struct stat buf;
     Ok = stat ( Paramfile, &buf )  == 0 ;
     if ( Ok ) 
       *Fsize = buf.st_size ;
     else
       eo46_set_rte_error( RteError, undef_ceo00, "", Paramfile ) ;
#endif
  }

  DBG3 (( MF__, "ParamFile:%s. Size = %d", Paramfile, *Fsize ));

  return ( Ok );
}

/*------------------------------*/

externC tsp00_Bool sqlXParamRemoveAll ( tsp00_DbNamec    DBNameC,
                                        tsp01_RteError  *RteError) 
{
  tsp00_Bool     Ok ;
   
  Ok = sqlXParamRemoveOneParamfile ( DBNameC, XP_ACT_VERSION_SP11, RteError ) &&
       sqlXParamRemoveAllVersions  ( DBNameC, RteError );

  return ( Ok );
}

/*------------------------------*/
externC tsp00_Bool 
        sqlXParamRemoveAllVersions ( tsp00_DbNamec  VAR_ARRAY_REF  DBNameC,
                                     tsp01_RteError               *RteError) 
{
  tsp00_Bool     Ok = true ;
  tsp00_Int4 VersionToDelete ;

  for ( VersionToDelete  = XP_FIRST_VERSION_SP11; 
        VersionToDelete <= XP_MAX_VERSION_SP11 && Ok ; VersionToDelete++ )
    Ok = sqlXParamRemoveOneParamfile ( DBNameC, VersionToDelete, RteError ) ;
  
  return ( Ok );
}

/*------------------------------*/

externC tsp00_Bool sqlXParamRemoveOneParamfile ( tsp00_DbNamec  VAR_ARRAY_REF  DBNameC,
                                                 tsp00_Int4                    VersionToRemove,
                                                 tsp01_RteError               *RteError) 
{
  tsp00_Bool     Ok ;
  tsp00_Pathc    Paramfile ;

  Ok = eo11BuildParamFileNameWithPath (DBNameC, VersionToRemove, Paramfile, RteError );

  if ( Ok )
    if ( eo11CheckFileExists (Paramfile) )
      Ok = eo11DeleteFile ( Paramfile, RteError ) ;

  return ( Ok );
}

/*------------------------------*/

externC tsp00_Bool  eo11SaveActParamFileAsVersion ( tsp00_DbNamec  VAR_ARRAY_REF  DBNameC,
                                                    tsp01_RteError               *RteError) 
{
  #undef  MF__
  #define MF__ MOD__"eo11SaveActParamFileAsVersion"

  tsp00_Int4  VersionToHandle = XP_FIRST_VERSION_SP11;
  tsp00_Pathc NewParamFileVersion ;
  tsp00_Bool  VersionExists ;
  DBGIN;

  /* search for the first free version number to use */
  do 
  {
    DBG3 ((MF__, "checking for paramfile version'%s'", NewParamFileVersion));
    if ( eo11BuildParamFileNameWithPath (DBNameC, VersionToHandle, NewParamFileVersion, RteError ) )
      VersionExists = eo11CheckFileExists( NewParamFileVersion ) ;
  }
  while ( RteError->RteErrCode == RTE_NO_ERROR_SP01  && VersionExists &&
          ++VersionToHandle <= XP_MAX_VERSION_SP11 ) ;

  /* if XP_MAX_VERSION_SP11 reached, remove latest version and rename all other */
  if ( VersionToHandle > XP_MAX_VERSION_SP11 )
  {
    VersionToHandle = XP_MAX_VERSION_SP11 ;
    eo11DeleteFile ( NewParamFileVersion, RteError ) ;
  }

  /* save previous versions */
  for ( --VersionToHandle; RteError->RteErrCode == RTE_NO_ERROR_SP01 && 
                           VersionToHandle      >= XP_FIRST_VERSION_SP11;
        --VersionToHandle )
  {
    eo11SaveParamFileVersion ( DBNameC, VersionToHandle, RteError ) ;
    DBG3 ((MF__, "saving paramfile version %d", VersionToHandle));
  }

  /* now we are ready to save the actual version */
  if ( RteError->RteErrCode == RTE_NO_ERROR_SP01 )
  {
    eo11SaveParamFileVersion ( DBNameC, XP_ACT_VERSION_SP11, RteError ) ;
    DBG3 ((MF__, "saving actual paramfile "));
  }

  return( RteError->RteErrCode == RTE_NO_ERROR_SP01 );

}

/*------------------------------*/

externC tsp00_Bool sqlXParamGetParamfileDir ( tsp00_DbNamec    DBNameC,
                                              tsp00_Pathc      ParamDir, 
                                              int              TerninateWithDelimiter,
                                              tsp01_RteError  *RteError  )
{
  #undef  MF__
  #define MF__ MOD__"sqlXParamGetParamfileDir"

  tsp00_Bool                Ok ;

/* PTS 1110180 This leads to a read only directory...
  if ( sqlIsServiceParamFile ( DBNameC ) )
    Ok = sqlGetServiceParamPath ( ParamDir, TerninateWithDelimiter, RteError) ;
  else
 */
    Ok = sqlGetIndependentConfigPath ( ParamDir, TerninateWithDelimiter, RteError) ;

  DBG3 ((MF__, "Ok = %d, ParamDir = %s", Ok, ParamDir));

  return (Ok) ;
}

/*===========================================================================*
 *  LOCAL FUNCTIONS (CODE)                                                   *
 *===========================================================================*/
static tsp00_Bool eo11BuildParamFileNameWithPath ( tsp00_DbNamec   DBNameC,
                                                   tsp00_Int4      ParamVersion,
                                                   tsp00_Pathc     ParamFileNameWithPath,
                                                   tsp01_RteError *RteError)
{
  tsp00_Bool     Ok ;

  Ok = sqlXParamGetParamfileDir ( DBNameC, ParamFileNameWithPath,
                                  TERM_WITH_DELIMITER_EO01, RteError ) ;
  if ( Ok )
  {
    strcat (ParamFileNameWithPath, DBNameC);

    if ( ParamVersion != XP_ACT_VERSION_SP11 )
    {
      int ActFileNameLen = (int)strlen(ParamFileNameWithPath) ;
      ParamFileNameWithPath[ActFileNameLen] = '.' ;
      ActFileNameLen++ ;
      if ( ParamVersion < 10 ) 
      {
        ParamFileNameWithPath[ActFileNameLen] = '0' ;
        ActFileNameLen++ ;
      }
      sql47_itoa( ParamVersion, ParamFileNameWithPath + ActFileNameLen, 
                  sizeof(tsp00_Pathc) - ActFileNameLen );
    }
  }
  return ( Ok );
}

/*------------------------------*/

static tsp00_Bool eo11CheckFileExists ( tsp00_Pathc FileToCheck )
{
  tsp00_Bool     Ok ;
  struct stat buf;
  Ok = stat ( FileToCheck, &buf )  == 0 ;
  return(Ok);
}

/*------------------------------*/

static tsp00_Bool eo11DeleteFile ( tsp00_Pathc FileToDelete,
                                   tsp01_RteError  *RteError) 
{
  tsp00_Bool     Ok ;
#if defined (_WIN32)
       Ok = DeleteFile ( FileToDelete ) != 0 ; 
#else
       Ok = unlink ( FileToDelete )  == 0 ;
#endif

  if ( !Ok ) 
    eo46_set_rte_error( RteError, undef_ceo00, "", FileToDelete ) ;

  return(Ok);
}

/*------------------------------*/

static tsp00_Bool  eo11SaveParamFileVersion ( tsp00_DbNamec  VAR_ARRAY_REF  DBNameC,
                                              tsp00_Int4                    VersionToSave ,
                                              tsp01_RteError               *RteError) 
{
  #undef  MF__
  #define MF__ MOD__"eo11SaveParamFileVersion"

  tsp00_Pathc    ParamFileToSave, NewParamFileVersion;
  tsp00_Int4     NewVersion = VersionToSave + 1 ;
  tsp00_Bool     Ok ;

  DBGIN;

  Ok = eo11BuildParamFileNameWithPath (DBNameC, VersionToSave, ParamFileToSave    , RteError ) &&
       eo11BuildParamFileNameWithPath (DBNameC, NewVersion   , NewParamFileVersion, RteError ) ;

  if ( Ok )
  {
    if ( VersionToSave == XP_ACT_VERSION_SP11 )
    {
      FILECOPY( ParamFileToSave, NewParamFileVersion, RteError );
      DBG3 ((MF__, "copy '%s' --> '%s', rte_rc = %d, os_rc = %d" ,
             ParamFileToSave, NewParamFileVersion, RteError->RteErrCode, RteError->OsErrCode));
    }
    else
    {
      DBG3 ((MF__, "rename '%s' --> '%s'" ,ParamFileToSave, NewParamFileVersion));
      Ok = rename ( ParamFileToSave, NewParamFileVersion ) == 0;
    }
  }

  DBGOUT;
  return(Ok) ;
}

/*------------------------------*/

/* PTS 1104471 */

static void eo11CreateTimeStamp ( tsp00_C64  VAR_ARRAY_REF  TimeStamp )
{
  #undef  MF__
  #define MF__ MOD__"eo11GetTimeStamp"

#ifdef WIN32
  {
    SYSTEMTIME   DateTime;
    GetLocalTime(&DateTime);
    wsprintf( TimeStamp,
              "%d-%02d-%02d %02d:%02d:%02d",
              DateTime.wYear,
              DateTime.wMonth,
              DateTime.wDay,
              DateTime.wHour,
              DateTime.wMinute,
              DateTime.wSecond ) ;
  }
#else
  {
#include <time.h>
#ifdef _REENTRANT
    struct tm                         tm_buffer ;
#endif
    struct tm                       * tmbuf ;
    time_t                            clk = time ( (time_t *) 0 );

#ifdef _REENTRANT
    tmbuf = localtime_r ( & clk, &tm_buffer );
#else
    tmbuf = localtime ( & clk );
#endif
    (void) sprintf ( (char *)TimeStamp , "%d-%02d-%02d %02d:%02d:%02d",
                     1900 + tmbuf->tm_year, tmbuf->tm_mon+1, tmbuf->tm_mday,
                     tmbuf->tm_hour, tmbuf->tm_min  , tmbuf->tm_sec );
  }
#endif
}

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/
