/****************************************************************************

  module      : veo111u.c

  -------------------------------------------------------------------------

  responsible : FrankH

  special area: tool
  description : create param from save


  last changed: 1999-03-24  11:40
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1998



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


/*----------------------------------------------------------------------------*
 *  INCLUDES                                                                  *
 *----------------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>

#include "gsp00.h"
#include "ggg00.h"
#include "heo06.h"
#include "heo46.h"

#ifdef WIN32
#include <windows.h>
#endif

#include "heo11.h"
#include "heo102.h"


/*----------------------------------------------------------------------------*
 *  DEFINES                                                                   *
 *----------------------------------------------------------------------------*/

#ifdef NO_INLINES
# define _INLINE
#else
# define _INLINE	__inline
#endif


/*----------------------------------------------------------------------------*
 *  MACROS                                                                    *
 *----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                              *
 *----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------*
 *  EXTERNAL VARIABLES                                                        *
 *----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------*
 *  GLOBAL VARIABLES                                                          *
 *----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------*
 *  LOCAL VARIABLES                                                           *
 *----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------*
 *  LOCAL FUNCTION PROTOTYPES                                                 *
 *----------------------------------------------------------------------------*/



/*
 * ========================== GLOBAL FUNCTIONS ================================
 */

int	main ( int argc , char **argv )

{
    tsp_vf_return               error ;
    tsp00_ErrText               errtext ;
    tsp_vfilename               SaveFile ;
    tsp_int2                    BufCnt ;
    tsp_dbname                  DBName ;
    tsp_int2                    Len ;
    tsp00_CompName              Component;

    if ( argc < 3 || argc > 4 )
    {
	(void) fputs ( "usage: cr_param savefile blck_cnt <DBNAME>\n", stderr );
	return ( 1 );
    }

    if ( argc == 3 )
      memset ( DBName ,  ' ', sizeof ( DBName ) );
    else
      memcpy ( DBName , argv[3], sizeof ( DBName )) ;

    memset ( errtext , ' ', sizeof ( errtext ) );
    memset ( SaveFile, ' ', sizeof ( SaveFile ) );

    Len = strlen ( argv[1] ) ;
    memcpy ( SaveFile, argv[1], 
             Len < sizeof ( SaveFile ) ? Len : sizeof ( SaveFile ) );

    BufCnt = atoi ( argv[2] ) ;

    eo46CtoP( Component, "cr_param", sizeof(tsp00_CompName));
    sqlinit ( Component, (tsp00_BoolAddr)0 );
    sqlextract_xparamfile ( SaveFile, BufCnt, DBName, &error, errtext ) ;
    if ( error != vf_ok )
      { errtext[sizeof(tsp00_ErrText)-1] = '\0' ;
        fprintf ( stderr, "cr_param: %s \n" , errtext ) ;
      }
    else
       fprintf ( stderr, "paramfile %.8s created\n", DBName );

    sqlfinish ( 0 );
    return ( error != vf_ok );
}


/*
 * ========================== LOCAL FUNCTIONS =================================
 */


/*
 * =============================== END ========================================
 */