/* @lastChanged: "1998-06-26  16:34"
 * @filename:    veo11u.c
 * @purpose:     "extract param from save"
 * @release:     7.1.1.1
 * @see:         ""
 *
 * @copyright:   "Copyright by   SAP AG, 1998"


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 */

/*----------------------------------------------------------------------------*
 *  INCLUDES                                                                  *
 *----------------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>

#include "gsp00.h"
#include "ggg00.h"
#include "heo06.h"
#include "heo46.h"
#include "gkb03.h"

#include "geo00_0.h"
#include "geo001.h"
#include "geo57.h"
#include "heo11.h"
#include "heo11_1.h"
#include "geo002xp.h"

/*----------------------------------------------------------------------------*
 *  DEFINES                                                                   *
 *----------------------------------------------------------------------------*/

#ifdef NO_INLINES
# define _INLINE
#else
# define _INLINE	__inline
#endif


/*----------------------------------------------------------------------------*
 *  MACROS                                                                    *
 *----------------------------------------------------------------------------*/
#define MOD__ "VEO11U.C: "

#define XP_PAGE_PNO(_Page)        (_Page->variant.C_1.xp_pno_F)
#define XP_PAGE_PT(_Page)         (_Page->variant.C_1.xp_pt_F)
#define XP_PAGE_PT2(_Page)        (_Page->variant.C_1.xp_pt2_F)
#define XP_PAGE_LENGTH(_Page)     (_Page->variant.C_2.xp_length_F)
#define XP_PAGE_TOT_LENGTH(_Page) (_Page->variant.C_2.xp_total_length_F)
#define XP_PAGE_DATA(_Page)       (_Page->variant.C_2.xp_data_F)
#define XP_PAGE_CHKSUM(_Page)     (_Page->variant.C_2.xp_checksum_F)


#define IS_PARAM_PAGE(_page)      ( XP_PAGE_PT2(_page) == pt2Xparam_egg00 && \
                                    XP_PAGE_PT (_page) == ptSave_egg00 )

#define IS_DATA_PAGE(_page)       ( XP_PAGE_PT(_page) == ptData_egg00 || \
                                    XP_PAGE_PT(_page) == ptLog_egg00 )


#define ERRTEXT(v,t)  { DBG3(((char*)t,v));\
		      eo46CtoP(v,t,sizeof(tsp_errtext)); }


/***************************************************************/
#if _WIN32 || OS2
/***************************************************************/

#define CLOSE_XPARAM_FILE( _XpFd, _Error, _Errtext ) \
  sql11c_CloseXParam ( _XpFd , !XP_FLUSH, _Errtext , _Error );

/***************************************************************/
#else
/***************************************************************/

#define CLOSE_XPARAM_FILE( _XpFd, _Error, _Errtext ) \
    sql11c_close_xparam_file ( _XpFd , FALSE ) ;

/***************************************************************/
#endif
/***************************************************************/

/*----------------------------------------------------------------------------*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                              *
 *----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------*
 *  EXTERNAL VARIABLES                                                        *
 *----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------*
 *  GLOBAL VARIABLES                                                          *
 *----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------*
 *  LOCAL VARIABLES                                                           *
 *----------------------------------------------------------------------------*/


static char BAD_BLOCKSIZE_GIVEN []            = "bad blocksize given";
static char CREATE_ERRTEXT      []            = "XPARAM Create ERROR";
static char BAD_BLOCKSIZE_READ  []            = "bad blocksize read";
static char NO_PARAMDATA_FOUND  []            = "no data for param found";
static char MALLOC_FAILED       []            = "not enough memory";
static char PARAMSIZE_MISMATCH  []            = "paramsize mismatch";
static char CHECKSUM_ERROR      []            = "checksum error";
static char CANT_GET_DBNAME     []            = "Serverdbname not found" ;

/*----------------------------------------------------------------------------*
 *  LOCAL FUNCTION PROTOTYPES                                                 *
 *----------------------------------------------------------------------------*/


static	void CreateXparamFile ( tsp_dbname     DBName,
                                char           *XparamBuf,
                                tsp_int4       XparamSize,
                                tsp_vf_return  *Error,
                                tsp_errtext      Errtext ) ;

static	void GetXparamDBName  ( tsp_dbname     DBName,
                                char           *XparamBuf,
                                tsp_int4       XparamSize,
                                tsp_vf_return  *Error,
                                tsp_errtext    Errtext );

static	void OpenXparamFile (   tsp_dbname     DBName,
                                tsp_int4       * XparamFileNo,
                                tsp_vf_return  *Error,
                                tsp_errtext     Errtext ) ;
/*
static int checksum( tkb3_xparam_page *XparamPage ) 
       { return ( XparamPage->checksum ) ; } 
*/


/*
 * ========================== GLOBAL FUNCTIONS ================================
 */

externC void	sqlextract_xparamfile ( tsp_vfilename                   SaveFile ,
                                        tsp_int2                        BufCnt ,
                                        tsp_dbname                      DBName ,
                                        tsp_vf_return                   * error ,
                                        tsp_errtext                     errtext )

{
  #undef  MF__
  #define MF__ MOD__"sqlcreate_param"

  tsp_int4             PoolName = 0;
  tsp_int4             SaveFileNo ;
  tsp_longint          BytesRead,
                       RecLen = BufCnt * sizeof(tsp_page) ;
  char                 *Buf = NULL ;
  char                 *ParamBuf = NULL ;
  tsp5_rte_fileerror   RteError ;
  ULONG                rc;

  DBGIN;

  DBG3 ((MF__, "SaveFile = %s", SaveFile )) ;
  DBG3 ((MF__, "sqlcreate_param: BufCnt   = %d", BufCnt   )) ;
  DBG3 ((MF__, "sqlcreate_param: DBName   = %.8s\n", DBName   )) ;

  if ( BufCnt < 1 )
    { *error = vf_notok  ;
       ERRTEXT( errtext, BAD_BLOCKSIZE_GIVEN );
    }
  else
      *error = vf_ok  ;

  if ( *error == vf_ok )
    { sqlfopenp ( SaveFile, SP5VF_BINARY, SP5VF_READ,
                SP5BK_UNBUFFERED, &SaveFileNo, &RteError ) ;
      if ( ( *error = RteError.sp5fe_result ) != vf_ok )
        ERRTEXT( errtext, (char *)RteError.sp5fe_text )
      else
        {
        rc = ALLOC_MEM_EO57( (void **)&Buf, RecLen );

        if ( rc != NO_ERROR_EO001 )
          { *error = vf_notok ;
            ERRTEXT( errtext, MALLOC_FAILED );
          }
        }
    }

  if ( *error == vf_ok )
    { BOOLEAN ParamDataFound = FALSE, EndOfParam = FALSE ;
      tsp_int4 TotParamSize, ActParamSize = 0,
               ActBlock, TotalReads = 0 ;
      do
        { sqlfreadc ( SaveFileNo, Buf, RecLen, &BytesRead , &RteError ) ;

          DBG3 ((MF__, " %8d.sqlfread : %d Bytes", ++TotalReads, BytesRead )) ;


          if ( ( *error = RteError.sp5fe_result ) == vf_ok && BytesRead > 0 )
            { tkb3_xparam_page *XparamPage = (tkb3_xparam_page *) Buf ;
              ActBlock = 0 ;

              if ( !ParamDataFound )
                { for ( ; ActBlock < BufCnt &&
                          !(EndOfParam     = IS_DATA_PAGE( XparamPage)) &&
                          !(ParamDataFound = IS_PARAM_PAGE(XparamPage)) ;
                          ActBlock++, XparamPage++ ) ;
                  if ( ParamDataFound )
                    { DBG3 ((MF__, "ParamDataFound at block %d / %d",
                                    TotalReads , ActBlock )) ;
                      TotParamSize = XP_PAGE_TOT_LENGTH(XparamPage) ;

                      rc = ALLOC_MEM_EO57( (void **)&ParamBuf, TotParamSize );

                      if ( rc != NO_ERROR_EO001 )
                        { *error = vf_notok ;
                          ERRTEXT( errtext, MALLOC_FAILED );
                        }
                    }
                }

              if ( ParamDataFound && *error == vf_ok )
                for ( ; ActBlock < BufCnt &&
                        !(EndOfParam = !IS_PARAM_PAGE( XparamPage )) ;
                        ActBlock++, XparamPage++ )
                  { DBG3 ((MF__,"param page %d found, data len = %d",
                           XP_PAGE_PNO(XparamPage), XP_PAGE_LENGTH(XparamPage) ));
                    if ( FALSE /* XparamPage->checksum != checksum(XparamPage) */ )
                      { *error = vf_notok ;
                        ERRTEXT( errtext, CHECKSUM_ERROR );
                      }
                    else
                    if ( ActParamSize + XP_PAGE_LENGTH(XparamPage) <= TotParamSize )
                      { memcpy ( ParamBuf + ActParamSize,
                                 XP_PAGE_DATA(XparamPage), XP_PAGE_LENGTH(XparamPage) ) ;
                        ActParamSize += XP_PAGE_LENGTH(XparamPage) ;
                      }
                    else
                      { *error = vf_notok ;
                         DBG3 ((MF__,"len mismatch: act=%d, add=%d, tot=%d",
                                ActParamSize, XP_PAGE_LENGTH(XparamPage), 
                                TotParamSize));
                         ERRTEXT( errtext, PARAMSIZE_MISMATCH );
                      }
                  }
            }
          else
            if ( *error != vf_ok )
              ERRTEXT( errtext, (char *)RteError.sp5fe_text )
            else
              { *error = vf_notok ;
                ERRTEXT( errtext, BAD_BLOCKSIZE_READ );
              }
        }
      while ( *error == vf_ok && !EndOfParam ) ;

      sqlfclosec ( SaveFileNo, SP5VF_CLOSE_NORMAL, &RteError);

      if ( *error == vf_ok )
        if ( !ParamDataFound )
          { *error = vf_notok ;
            ERRTEXT( errtext, NO_PARAMDATA_FOUND );
          }
        else
          CreateXparamFile ( DBName, ParamBuf, TotParamSize, error, errtext );
    }
  if ( ParamBuf != NULL ) FREE_MEM_EO57 ( ParamBuf ) ;
  if ( Buf      != NULL ) FREE_MEM_EO57 ( Buf ) ;
  DBGOUT ;
}

/* ========================== LOCAL  FUNCTIONS ============================ */

static	void CreateXparamFile ( tsp_dbname     DBName,
                                char           *XparamBuf,
                                tsp_int4       XparamSize,
                                tsp_vf_return  *Error,
                                tsp_errtext      Errtext )

{
  #undef  MF__
  #define MF__ MOD__"CreateXParamFile"

  tsp_int4         XparamFileNo ;
  tsp_int4         bytesWritten;
  boolean          DummyOk = TRUE;
  tsp_errtext      DummyErrtext ;

  DBGIN;

  GetXparamDBName ( DBName, XparamBuf, XparamSize, Error, Errtext );

  if ( *Error == vf_ok )
    { OpenXparamFile( DBName, &XparamFileNo, Error, Errtext );

      if ( XparamFileNo >= 0 )
        { XP_RETURNCODE xp_rc ;
          sql11c_SeqBinWriteXParam ( XparamFileNo, XparamBuf, XparamSize,
                                     &bytesWritten, Errtext, &xp_rc ) ;

          DBG3 ((MF__, "%d of %d Bytes written to paramfile",
                       XparamSize, bytesWritten));

          CLOSE_XPARAM_FILE ( XparamFileNo, &DummyOk, DummyErrtext );
          if ( !DummyOk || xp_rc != xp_ok ) 
            *Error = vf_notok ;
        }
    }
  else
    ERRTEXT( Errtext, CANT_GET_DBNAME );

  DBGOUT ;
}

/*==========================================================================*/

static	void GetXparamDBName ( tsp_dbname     DBName,
                               char           *XparamBuf,
                               tsp_int4       XparamSize,
                               tsp_vf_return  *Error,
                               tsp_errtext     Errtext )

{
  #undef  MF__
  #define MF__ MOD__"GetXparamDBName"

  XP_HFILE XpFd ;

  DBGIN;

  if ( (XpFd = sql11c_InitXparamStruc ( XparamBuf, XparamSize)) >= 0 )
    {
      XP_KEY_TYPE       xp_key ;
      XP_VALUE_REC      xp_value ;
      XP_KEY_TYPEC      ParamId ;
      XP_RETURNCODE     xp_ret = XP_OK ;

      if ( *DBName == ' ' || *DBName == '\0' )  /* get DBName form Save */
        { do
            {
              sql11c_SeqNextXParam ( XpFd, xp_key, &xp_value, Errtext, &xp_ret);
              if ( xp_ret == XP_OK )
  	        { eo46PtoC ( ParamId , xp_key , sizeof(xp_key) )  ;
                  DBG3 ((MF__, "xp_key = '%.18s'" , xp_key ));
                }
            }
          while ( xp_ret == XP_OK && strcmp( ParamId , PAN_SERVERDB ) ) ;
          if ( xp_ret == XP_OK )
            { memcpy ( DBName, (char *)&xp_value.xp_type,
                       sizeof ( tsp_dbname ) ) ;
              DBG3 ((MF__, "found DBName = '%.8s'" , DBName  ));
            }
        }
      else                                      /* take over given DBame */
        { eo46CtoP ( xp_key , PAN_SERVERDB,  sizeof(xp_key) )  ;
          memset ( &xp_value, 0, sizeof ( xp_value ) );
          xp_value.xp_value_t = XP_C18_TYPE ;
          eo46CtoP ( xp_value.xp_type.xp_c18, (char *)DBName, sizeof(tsp_dbname) ) ;
          sqlputxparam ( XpFd, xp_key , (tsp00_XpValueRec *)&xp_value , Errtext , &xp_ret ) ;
        }

      if ( xp_ret != XP_OK )
        *Error = vf_notok  ;

      sql11c_FreeXparamStruc ( XpFd ) ;
    }

  DBGOUT ;
}


/*==========================================================================*/

static	void OpenXparamFile ( tsp_dbname     DBName,
                              tsp_int4       * XparamFileNo,
                              tsp_vf_return  *Error,
                              tsp_errtext     Errtext )

{
  #undef  MF__
  #define MF__ MOD__"OpenXParamFile"

  tsp01_RteError     RteError ;
  tsp00_Int4         DefaultMode = XP_OPEN_CREATE_SP11 | XP_OPEN_BIN_SP11 | XP_OPEN_DIRECT_SP11;

  sqlXparamOpen ((char *)DBName, XparamFileNo, DefaultMode, &RteError) ;

  if ( RteError.RteErrCode != RTE_NO_ERROR_SP01 )
    { *Error = vf_notok  ;
      eo46CtoP ( Errtext, RteError.RteErrText, sizeof (tsp00_ErrText) );
    }
  else
     *Error = vf_ok ;

  DBGOUT ;
}


/*
 * =============================== END ========================================
 */