/*@lastChanged: "1999-04-21  15:03"
* @filename:    veo38.cpp
* @purpose:     "COM_Interface" 
* @release:     7.2.0.0
* @see:         ""
*
* @copyright:   "Copyright by SAP AG, 1999"


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/

/*===========================================================================*
*  INCLUDES                                                                 *
*===========================================================================*/
#if defined(NMP)
#define RWSTD_NO_NAMESPACE  1
#include <sys/user.h>
#endif

#include "gsp00.h"
#include "ggg00.h"
#include "ggg92.h"
#include "heo01.h"
#include "heo52.h"
#include "heo15.h"
#include "veo00c.h"
#if defined (_WIN32)

#define DCOM_EXCLUSIVE  &kgs.DCOM_MEM.lExclusive
#define DCOM_HEAD       kgs.DCOM_MEM.lphead
#define DCOM_FREE       kgs.DCOM_MEM.lpfree
#include "gos00.h" /* nocheck */
/* gos00.h defines try as '__try' and catch as '__except' */ 
#undef try
#undef catch 
#include "geo007_1.h" /* nocheck */
#include "vos00kc.h" /* nocheck */
#include "ilivecachesink.h"  /* no check */
#include "hos38.h" /* nocheck */
#include "hos39.h" /* nocheck */

#else

#define DCOM_EXCLUSIVE  kgs->DCOM_MEM.lExclusive
#define DCOM_HEAD       kgs->DCOM_MEM.lphead
#define DCOM_FREE       kgs->DCOM_MEM.lpfree
#include "gco00.h" /* nocheck */
#include "ilivecachesink.h"  /* no check */
#include "vsp002c.h" /* nocheck */
#include "heo46.h" /* nocheck */
#include "heo60.h" /* nocheck */
#include "gen50.h" /* nocheck */
#include "gen500.h" /* nocheck */
#include "hen38.h" /* nocheck */
#endif


#if defined(USE_KGS)
#include "heo56.h" /* nocheck */
#include "geo60.h" /* nocheck */
#endif

#include "geo00d.h"
#include "geo39.h"
#include "hak07.h"
#include "hak34.h"
#include "hak341.h"
#include "geo38.h"
#include "heo38.h"

#include "geo36.h"
#include "heo02x.h"
#include "hco31-k.h"
#include "ggg251.h"

#if defined(USE_KGS)
#include "hgg01.h"     /* PTS 1110813 */
#endif

externC void vabort(tsp00_Bool dumpCore);

/*===========================================================================*
*  DEFINES                                                                  *
*===========================================================================*/

#ifdef NO_INLINES
# define _INLINE
#else
# define _INLINE  inline
#endif

#if defined (BIT64)
#define   MIN_PARAMETER_SIZE  8
#else
#define   MIN_PARAMETER_SIZE  4
#endif

#if defined (_WIN32)
#define   TRY                               __try 
#if defined(USE_KGS)
#define   CATCH_ALL                         __except (EXCEPTION_EXECUTE_HANDLER, eo38_Dump_Exception(GetExceptionInformation())) 
#else
#define   CATCH_ALL                         __except (EXCEPTION_EXECUTE_HANDLER) 
#endif
#else    
#define   TRY                               try
#define   CATCH_ALL                         catch (DbpError)   
#endif

//#define   DEBUG_FACTORY                     -1
#define   DEBUG_INTERFACE_LOCAL             -2
#define   DEBUG_INTERFACE_REMOTE            -3
#define   DEBUG_SESSION_ID                   1

#define   RC_DISPADDR_FAILED                -1;
#define   RC_DISPADDR_INVALIDATE_FAILED     -2;
#define   RC_DISPADDR_OK                     0;
#define   RC_DISPADDR_FACTORY_FOUND          1;
#define   RC_DISPADDR_INVALID_FOUND          2;
#define   RC_DISPADDR_INUSE                  3;

#define   C_SYS_DISPID                       5;

/*===========================================================================*
*  MACROS                                                                   *
*===========================================================================*/

#ifdef USE_KGS
#else
#ifdef WAIT_UNTIL_ALONE
#undef  WAIT_UNTIL_ALONE
#define WAIT_UNTIL_ALONE(x)
#endif

#ifdef CLEARLOCK
#undef  CLEARLOCK
#define CLEARLOCK(x)
#endif

#ifdef  MSGCD
#undef  MSGCD
#define MSGCD(x)
#endif

#if defined (WIN32)
#define os39VDcomGetDispAddr (parms);
#define os39CoCreateInstance (parms);
#define os39CoGetClassobject (parms);
#define os39VDcomCreateInstance (parms);
#define os39VDcomDBProc (parms);
#endif
#endif


/*===========================================================================*
*  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
*===========================================================================*/

#if defined(_IBMR2)

#if defined(BIT64)
/* 
 *  with xlC Compiler on AIX an interface pointer points to the following      
 *  structure:
 *
 *
 *
 *  Interface - pointer  
 *   
 *  | 
 *  |
 *  v
 *   ----------   0  -----------      
 *  |virt.Meth.|--->! Method 1  !    Method 1 is a 'normal' Function- 
 *   ----------   8  -----------     pointer ( see AIX 32 BIT )
 *                  ! Method 2  !    
 *               16  -----------                       
 *                  !     .     !                    
 *                        .                             
 *                  !     .     !                        
 *              n*8  -----------
 *                  ! Method n  !    
 *                   -----------     
 *                  !  0xdead   !    
 *                   -----------     
 *                  !  0xdead   !     
 *                   -----------     
 *                                          
 *                                   
 *                                           
 */
typedef long *t_lpptr;
#else  /* AIX 32 BIT */
/* 
 *  with xlC Compiler on AIX an interface pointer points to the following      
 *  structure:
 *
 *
 *
 *  Interface - pointer  
 *   
 *  | 
 *  |
 *  v
 *   ----------   0  -----------     each Method-Pointer is made up of 
 *  |virt.Meth.|--->!     0     !    2 words:
 *   ----------      -----------        
 *                  !     0     !      1-word: address (Method-Desc)
 *                8  -----------       2-word: pInterface + 2-word = 
 *                  ! Method 1  !              this-pointer of C++ Class
 *                   -         -               of the member function
 *                  ! pointer   !              
 *               16  -----------
 *                  ! Method 2  !    Method-Desc-i -> 0  -----------
 *                   -         -                        ! method i  !
 *                  ! pointer   !                     4  -----------
 *               24  -----------                        !   TOC     ! r2
 *                  !     .     !                     8  -----------
 *                        .                             !environment! r11
 *                  !     .     !                        ----------
 *              n*8  -----------
 *                  ! Method n  !    the Method-Desc-i has the 'normal'
 *                   -         -     structure for a function pointer
 *                  ! pointer   !    parameter
 *                   -----------     1. word: address of the entry point of
 *                  !  0xdead   !             the procedure
 *                   -         -     2. word: address of TOC of module in which 
 *                  !  0xbeaf   !             the procedure is bound
 *                   -----------     3. word: environment-pointer for languages 
 *                                            as Pascal. If used by C++ ????
*/
typedef struct t_tabentry
{
  void*     addr;
  long      this_displ;
} t_tabentry;
typedef t_tabentry *t_lpptr;
#endif /* AIX 32 BIT */

#elif defined(NMP)  /* Reliant Unix  -  32 & 64 Bit */

/* 
 *  with CC Compiler on NMP an interface pointer points to the following      
 *  structure:
 *
 *
 *
 *  Interface - pointer  
 *   
 *  | 
 *  |
 *  v
 *   ----------   0  -----------     each Method-Entry made up of 
 *  |virt.Meth.|--->!   RTTI    !    2 short fields +
 *   ----------      --       --     1 pointer (4 Byte aligned 32 Bit/
 *                  !   Entry   !               8 Byte aligned 64 Bit )
 *                i  -----------       1.st short: is the delta value added 
 *                  ! Method 1  !                  to the this pointer  
 *                   -         -       2.nd short: isn't used (cfront compatability)
 *                  !  Entry    !      3 pointer:  point to the virtual function
 *               2*i -----------
 *                  ! Method 2  !                                            
 *                   -         -                                       
 *                  !  Entry    !                                    
 *               3*i -----------                                          
 *                  !     .     !    i = sizeof(short)*2 + alignment + sizeof(pointer);
 *                        .                                               
 *                  !     .     !        32 Bit ==> alignment = 0; 
 *              n*i  -----------         64 Bit ==> alignment = 4;
 *                  ! Method n  !                                      
 *                   -         -                                      
 *                  !  Entry    !               
 *                   -----------                                            
 *                                                         
 *                                                                              
*/
typedef struct t_tabentry
{
  short     this_displ;
  short     not_used;
#if defined(BIT64)
  int       filler;
#endif
  void*     addr;
} t_tabentry;

typedef t_tabentry *t_lpptr;

//#end Reliant Unix NMP */
#elif defined(SUN)
#if (__SUNPRO_CC < 0x0500)     /* PTS 1105381 */
/* 
 *
 *  the SUN C++ Compiler interface pointer points to the following      
 *  structure:
 *
 *  pInterface -> 0  ----------        0  -----------  each Method-Pointer is  
 *                  |virt.Meth.|-------->!     0     !   made up of 2 words:
 *                   ----------           -----------        
 *                  |          |         !     0     ! 1-word:  
 *                                      8  -----------  1. half-word: pInterface +       
 *                                       ! Method 1  !     half-word = this-pointer 
 *                                        -         -                  of Class
 *                                       ! pointer   !  2. half-word: not used      
 *                                    16  -----------       
 *                                       ! Method 2  ! 2-word: address of method
 *                                        -         -         
 *                                       ! pointer   !        
 *                                    24  -----------         
 *                                       !     .     !        
 *                        .              
 *                                       !     .     !        
 *                                   n*8  -----------
 *                                       ! Method n  !    
 *                                        -         -     
 *                                       ! pointer   !    
 *                                        -----------
*/
typedef struct t_tabentry
{
  short     this_displ;
  short     filler;
  void*     addr;
} t_tabentry;

typedef t_tabentry *t_lpptr;
#else
/* 
 *
 *  pInterface -> 0  -----------        0  ------------
 *                  ! virt.Meth ! ------->|      0     |
 *                   -----------           ------------
 *                  !     :     !         |      0     |
 *                                      8  ------------
 *                                        | Method 1   |
 *                                     16  ------------
 *                                        | Method 2   |
 *                                         ------------
 *                                        |      :     |
 *                                               :
 *                                        |      :     |
 *                                         ------------
 *                                (n+1)*4 | Method n   |
 *                                         ------------
 *                                        
*/
typedef long *t_lpptr;
#endif
#elif defined(PA11) || defined(PA20W)
/* 
 *  the HP 32Bit aCC Compiler interface pointer points to the following      
 *  structure:
 *
 *  pInterface -> 0  -----------        0  ------------
 *                  ! virt.Meth ! ------->|      0     |
 *                   -----------           ------------
 *                  !     :     !         |      0     |
 *                                         ------------
 *                                        |      x     |
 *                                         ------------
 *                                        |      0     |
 *                                      16 ------------
 *                                        | Method 1   |
 *                                         ------------
 *                                        | Method 2   |
 *                                         ------------
 *                                        |      :     |
 *                                               :
 *                                        |      :     |
 *                                         ------------
 *                                        | Method n   |
 *                                         ------------
 *                                        
*/
typedef long *t_lpptr;
#elif defined(ALPHA)
/* 
 *  the ALPHA 64Bit cxx Compiler interface pointer points to the following      
 *  structure:
 *
 *  pInterface -> 0  -----------        0  ------------
 *                  ! virt.Meth ! ------->| Method 1   |
 *                   -----------           ------------
 *                  !     :     !       8 | Method 2   |
 *                                         ------------
 *                                        |      :     |
 *                                               :
 *                                        |      :     |
 *                                         ------------
 *                               (n-1)*8  | Method n   |
 *                                         ------------
 *                                        
 * on the ALPHA machine the C++ type 'long' and pointers are 64-bit long.
*/
typedef long *t_lpptr;
#elif defined(LINUX)
/* 
 *  gcc 2.95/2.96 Compiler interface pointer points to the following      
 *  structure:
 *
 *  pInterface -> 0  -----------        0  ------------
 *                  ! virt.Meth ! ------->|      0     |
 *                   -----------           ------------
 *                  !     :     !       4 | Type-Info  |
 *                                         ------------
 *                                      8 | Method 1   |
 *                                         ------------
 *                                        |      :     |
 *                                               :
 *                                        |      :     |
 *                                         ------------
 *                           8 + (n-1)*4  | Method n   |
 *                                         ------------
 * 
 *  gcc 3.2 for i386 Compiler interface pointer points to the following      
 *  structure:
 *
 *  pInterface -> 0  -----------        0  ------------
 *                  ! virt.Meth ! ------->| Method 1   |
 *                   -----------           ------------
 *                  !     :     !       4 | Method 2   |
 *                                         ------------
 *                                      8 | Method 3   |
 *                                         ------------
 *                                        |      :     |
 *                                               :
 *                                        |      :     |
 *                                         ------------
 *                             (n-1) * 4  | Method n   |
 *                                         ------------
 */
typedef long *t_lpptr;
#elif defined(_WIN32)

#if defined(_WIN64)
/* 
 *  MS VC++ Compiler interface pointer points to the following      
 *  structure:
 *
 *  pInterface -> 0  -----------        0  ------------
 *                  ! virt.Meth ! ------->| Method 1   |
 *                   -----------           ------------
 *                  !     :     !       8 | Method 2   |
 *                                         ------------
 *                                        |      :     |
 *                                               :
 *                                        |      :     |
 *                                         ------------
 *                               (n-1)*8  | Method n   |
 *                                         ------------
*/
typedef tsp00_Longint *t_lpptr;
#else
/* 
 *  MS VC++ Compiler interface pointer points to the following      
 *  structure:
 *
 *  pInterface -> 0  -----------        0  ------------
 *                  ! virt.Meth ! ------->| Method 1   |
 *                   -----------           ------------
 *                  !     :     !       4 | Method 2   |
 *                                         ------------
 *                                        |      :     |
 *                                               :
 *                                        |      :     |
 *                                         ------------
 *                               (n-1)*4  | Method n   |
 *                                         ------------
*/
typedef long *t_lpptr;
#endif

#else
typedef long *t_lpptr;
#endif

typedef struct
{
  union
  {
    char i1;
    short i2;
    long i4;
    int intVal;
    FLOAT flt;
    FLOAT dbl;              
    PVOID ptr;
  } t;
} t_Parm;

typedef struct
{
  t_Parm p[256];
} t_Parms;

typedef HRESULT (__stdcall * t_lpIFunc) (LPVOID lpIID, const t_Parms p);

class CDefaultAllocator_eo38 {
public :
  void*  allocate(size_t sz) { 
    void* p = malloc(sz);
    if ( !p ) throw DbpError(DbpError::RTE_ERROR, e_new_failed);
    return p; 
  };
  
  void   deallocate(void* p) { 
    free(p); 
  };
};

class CLoadedLibDirEntry_eo38
{
private:
  HANDLE m_hd;
  short  m_usedCount;
  char*  m_namePtr;
  bool   m_global;
public:
  CLoadedLibDirEntry_eo38(HANDLE hd, const char* p, bool g) : m_hd(hd), m_usedCount(1), m_global(g) {
    m_namePtr = (char *)malloc(strlen(p)+1);
    if ( !m_namePtr ) throw DbpError(DbpError::RTE_ERROR, e_new_failed);
    strcpy(m_namePtr,p);
  };

  ~CLoadedLibDirEntry_eo38() {
    free(m_namePtr);
  }

  void  incrementUsedCount() { 
    m_usedCount++; 
  };
  short  decrementUsedCount() {
    if (m_usedCount > 0) {
      m_usedCount--;
    }
    return m_usedCount; 
  };

  const char * const getNamePtr() const {
    return m_namePtr;
  };

  HANDLE getHandle() const {
    return m_hd;
  };

  bool getGlobalFlg() const {
    return m_global;
  };

  void* operator new( size_t sz )                  /* PTS 1110402 */
  {                                                /* PTS 1110402 */
    return malloc( sz );                           /* PTS 1110402 */
  };                                               /* PTS 1110402 */
  
  void operator delete(void* p)                    /* PTS 1110402 */
  {                                                /* PTS 1110402 */
    free( p );                                     /* PTS 1110402 */
  };                                               /* PTS 1110402 */

};

/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
*===========================================================================*/

/*===========================================================================*
*  GLOBAL VARIABLES                                                         *
*===========================================================================*/

#ifndef USE_KGS
externC KERNEL_GLOBAL_SEC kgs;
#endif

static CDefaultAllocator_eo38                                        DefaultAllocator;
static cgg251DCList<CLoadedLibDirEntry_eo38*,CDefaultAllocator_eo38> LoadedLibDir(&DefaultAllocator);

/*===========================================================================*
*  LOCAL VARIABLES                                                          *
*===========================================================================*/



/*===========================================================================*
*  LOCAL FUNCTIONS (PROTOTYPES)                                             *
*===========================================================================*/

static HRESULT eo38_ComCache (GUID * lpGuid, 
                              void *lpInterface, 
                              tsp00_Uint4 session,
                              bool  remoteDbgFlag,
                              teo38_HashEntryType hType);
static BOOLEAN eo38_Initialize (void);
static void eo38_Invalidate(tsp_dcom_co_create_instance_parms *lpParms,
                            GUID *lpGuid, 
                            tsp00_Uint4 session);
static void eo38_ReleaseInstance ( CHashEntryPtr    currEntry, tsp00_Addr SessionContext );
static HRESULT eo38_UnadviseInstance ( CHashEntryPtr    currEntry );
static void eo38DisposeInstances( tsp_dcom_dispatch_parms * parms );

/*===========================================================================*
*  GLOBAL FUNCTIONS (CODE)                                                  *
*===========================================================================*/

/*===========================================================================*
*  CHashEntry                                                               *
*===========================================================================*/
CHashEntry::CHashEntry (){
  m_lpnext            = NULL;
  m_session           = 0;
  m_iid               = GUID_NULL;
  m_IUnknown          = NULL;
  m_type              = Free_eeo38;
  m_is_committed      = false;
  m_idebug_used_flag  = false;
  m_is_remote_debug   = false;
};

/*===========================================================================*
*  CSessionHashEntry                                                        *
*===========================================================================*/

void CHashEntry::init(tsp00_Uint4 session, 
                      LPGUID      lpGuid, 
                      LPUNKNOWN   lpInterface, 
                      bool        remoteDbgFlag,
                      teo38_HashEntryType hType )
{
  m_lpnext            = NULL;
  m_session           = session;
  m_iid               = *lpGuid;
  m_IUnknown          = lpInterface;
  m_type              = hType;
  m_is_committed      = false;
  m_idebug_used_flag  = false;
  m_is_remote_debug   = remoteDbgFlag;
};

/*---------------------------------------------------------------------------*/
void* CHashEntry::operator new( size_t sz )                    /* PTS 1110402 */
{                                                              /* PTS 1110402 */
  return malloc( sz );                                         /* PTS 1110402 */
};                                                             /* PTS 1110402 */

/*---------------------------------------------------------------------------*/
void CHashEntry::operator delete(void* p)                      /* PTS 1110402 */
{                                                              /* PTS 1110402 */
  free( p );                                                   /* PTS 1110402 */
};                                                             /* PTS 1110402 */
  
/*---------------------------------------------------------------------------*/
CCoClassHash::CCoClassHash ()
{
  m_HashHdrSize   = 0;
  m_HashHdrPtr    = NULL;
  m_HashFreePtr   = NULL;
  m_HashCurrPtr   = NULL;
  m_HashCurrIdx   = 0;
  m_session       = 0;
  m_searchType    = SearchEQ_eeo38;
};
  


/*---------------------------------------------------------------------------*/
tsp00_Uint4 CCoClassHash::HashInit (const tsp00_Uint4 lcSessionHashHdrSize )
{
  tsp00_Uint4    ix;
  
  /* m_HashHdrPtr    = new CHashEntryPtr [ lcSessionHashHdrSize ];        PTS 1110402 */ 
  m_HashHdrPtr = (CHashEntryPtr*) malloc( sizeof( CHashEntryPtr )      /* PTS 1110402 */
    * lcSessionHashHdrSize );                                          /* PTS 1110402 */
  if ( NULL == m_HashHdrPtr )
  {
    return ERROR_NOT_ENOUGH_MEMORY;
  }
#if defined(USE_KGS) && defined(DEBUG_MEM_ALLOC)
    MSGD(( 11999, INFO_TYPE, "new   ", "veo38.cpp:HashInit - 0x%lx, sizeInBytes %ld",
                  m_HashHdrPtr, sizeof(CHashEntryPtr) * lcSessionHashHdrSize ));
#endif /* DEBUG_MEM_ALLOC */
  
  m_HashHdrSize = lcSessionHashHdrSize;
  for ( ix = 0; ix < lcSessionHashHdrSize; ix++ )
    m_HashHdrPtr[ix] = NULL;
  
  return S_OK;
};
 
/*---------------------------------------------------------------------------*/
void CCoClassHash::chainEntry(CHashEntryPtr p){
  tsp00_Uint4    ix;
  ix = SessionHashIdx(p->getSession());
  p->setNextEntry( m_HashHdrPtr[ix] );
  m_HashHdrPtr[ix] = p;
}


/*---------------------------------------------------------------------------*/
void CCoClassHash::ResetEntrySearch( tsp00_Uint4 session, 
                                    teo38_HashSearchType searchType ){
  m_session              = session;
  m_HashCurrPtr   = NULL;
  m_HashCurrIdx   = 0;
  m_searchType           = searchType;
};

/*---------------------------------------------------------------------------*/
CHashEntryPtr CCoClassHash::getFreeEntry(){
  CHashEntryPtr p;
  p = m_HashFreePtr;
  if ( NULL == p ) {
    p = new (CHashEntry);
#if defined(USE_KGS) && defined(DEBUG_MEM_ALLOC)
    MSGD(( 11999, INFO_TYPE, "new   ", "veo38.cpp:getFreeEntry() - 0x%lx, sizeInBytes %ld",
                  p, sizeof(CHashEntry) ));
#endif /* DEBUG_MEM_ALLOC */
  }
  else {
    m_HashFreePtr = p->getNextEntry();
  }
  return p;
};

/*---------------------------------------------------------------------------*/
CHashEntryPtr CCoClassHash::NextEntry( void ){
  
  if ( SearchEQ_eeo38 == m_searchType ) {
    if ( NULL == m_HashCurrPtr ) 
      m_HashCurrPtr = m_HashHdrPtr[SessionHashIdx( m_session )];
    else
      m_HashCurrPtr = m_HashCurrPtr->getNextEntry();
    while ( NULL != m_HashCurrPtr ) {
      if ( m_HashCurrPtr->sessionEQ( m_session )) 
        return m_HashCurrPtr;
      m_HashCurrPtr = m_HashCurrPtr->getNextEntry();
    }
  }
  else {
    for ( ; m_HashCurrIdx < m_HashHdrSize; m_HashCurrIdx++ ){
      bool  found;
     
      if ( NULL == m_HashCurrPtr ) 
        m_HashCurrPtr = m_HashHdrPtr[m_HashCurrIdx];
      else
        m_HashCurrPtr = m_HashCurrPtr->getNextEntry();
      while ( NULL != m_HashCurrPtr ) {
        switch ( m_searchType )
        {
        case SearchALL_eeo38:
          found = true;
          break;
        case SearchNE_eeo38:
          found = m_HashCurrPtr->sessionNE( m_session );
          break;
        default:
          found = false;
        }
        if ( found )
        {
          return m_HashCurrPtr;
        }
        m_HashCurrPtr = m_HashCurrPtr->getNextEntry();
      }
    }
  }
  m_HashCurrPtr   = NULL;
  m_HashCurrIdx   = 0;
  return NULL;    
}

/*---------------------------------------------------------------------------*/
CHashEntryPtr CCoClassHash::GetRechainFirstEntry( 
                              teo38_HashEntryType *lpEntryType ){
  tsp00_Uint4           ix;
  bool                  found;
  CHashEntryPtr  prev = NULL;
  CHashEntryPtr  curr = NULL;
  
  for ( ix = 0; ix < m_HashHdrSize; ix++) {
    if ( NULL == curr ) 
      curr = m_HashHdrPtr[ix];
    else {
      prev = curr;
      curr = curr->getNextEntry();
    }
    
    while ( NULL != curr ) {
      switch ( m_searchType )
      {
      case SearchALL_eeo38:
        found = true;
        break;
      case SearchEQ_eeo38:
        found = curr->sessionEQ( m_session );
        break;
      case SearchNE_eeo38:
        found = curr->sessionNE( m_session );
        break;
      default:
        found = false;
      }
      if ( found && ( NULL != lpEntryType ))
      {
        if ( curr->getType() != *lpEntryType )
          found = false;
      }
      if ( found )
      {
        if ( NULL == prev )
          m_HashHdrPtr[ix] = curr->getNextEntry();
        else
          prev->setNextEntry(curr->getNextEntry());
        /* chain into FreeChain */
        curr->setNextEntry(m_HashFreePtr);
        m_HashFreePtr = curr;
        return curr;
      }
      prev = curr;
      curr = curr->getNextEntry();
    }
    prev = NULL;
  }
  return NULL;    
}
  
/*---------------------------------------------------------------------------*/
void CCoClassHash::chainFreeEntry(CHashEntryPtr p) {
  p->setNextEntry(m_HashFreePtr);
  m_HashFreePtr = p;
};
  
/*---------------------------------------------------------------------------*/
CHashEntryPtr CCoClassHash::FindHashEntry( GUID *lpGuid, tsp00_Uint4 session ) {
  CHashEntryPtr  curr;
  
  curr = m_HashHdrPtr[ SessionHashIdx(session) ];
  while ( NULL != curr ) {
    if ( curr->sessionEQ( session ) && curr->GuidEQ( lpGuid ) )
      return curr;
    curr = curr->getNextEntry();
  }
  return NULL;
}

/*---------------------------------------------------------------------------*/
void* CCoClassHash::operator new( size_t sz )                  /* PTS 1110402 */
{                                                              /* PTS 1110402 */
  return malloc( sz );                                         /* PTS 1110402 */
};                                                             /* PTS 1110402 */

/*---------------------------------------------------------------------------*/
void CCoClassHash::operator delete(void* p)                    /* PTS 1110402 */
{                                                              /* PTS 1110402 */
  free( p );                                                   /* PTS 1110402 */
};                                                             /* PTS 1110402 */

/*===========================================================================*
*                                                                            *
*===========================================================================*/

externC void eo38CoCreateInstance ( tsp_dcom_co_create_instance_parms * parms )
{ 
  /* This Entry is only used for UDT's      */
  DWORD dwClsContext;
  GUID  CoClsId;
  IID   IId;
  
  parms->ifactory = 0;
  
#if defined(USE_KGS) && defined(WIN32)
  if (!parms->debug_mode)
  {
#endif
    if (parms->in_proc_server)
      dwClsContext = CLSCTX_INPROC_SERVER;
    else
      dwClsContext = CLSCTX_LOCAL_SERVER;
    
    memcpy(&CoClsId, &parms->coclsid[0], sizeof(CoClsId));
    memcpy(&IId, &parms->iid[0], sizeof(IId));
    parms->ret = sql38CoCreateInstance ( &CoClsId, 
      parms->debug_location, 
      dwClsContext, 
      IId, 
      (void**)&(parms->idispatch));
    if (!parms->ret)
    {
      parms->ret = eo38_ComCache (&IId, parms->idispatch, 0, false, IID_eeo38);
    }
#ifdef USE_KGS
    else
    {
      char cClsContext[80];
      short loc_len = 0;
      
      memset(cClsContext,'\0',sizeof(cClsContext));
      switch ( dwClsContext )
      {
      case CLSCTX_INPROC_SERVER:
        strcpy(cClsContext, "INPROC");
        break;
      case CLSCTX_LOCAL_SERVER:
        strcpy(cClsContext, "LOCAL");
        break;
      case CLSCTX_REMOTE_SERVER:
        strcpy(cClsContext, "REMOTE AT ");
        memcpy(&cClsContext[strlen(cClsContext)],&parms->debug_location[0], loc_len );
        break;
      default:
        break;
      }
      MSGCD ((ERR_DCOM_COCREATE, "CoCreateInstance", parms->ret, cClsContext ));
      MSGCD ((ERR_DCOM_GUID, "COCLSID",
        CoClsId.Data1, CoClsId.Data2, CoClsId.Data3,
        CoClsId.Data4[0], CoClsId.Data4[1], CoClsId.Data4[2],
        CoClsId.Data4[3], CoClsId.Data4[4], CoClsId.Data4[5],
        CoClsId.Data4[6], CoClsId.Data4[7] ));
    }
#if defined (WIN32)
  }
  else
    os39CoCreateInstance (parms);
#endif
#endif
}

/*---------------------------------------------------------------------------*/

externC void eo38CoGetClassObject ( tsp_dcom_co_create_instance_parms * parms )
{
  
  /* Create a ClassFactory                              */
  /* ClassFactories are globally used within liveCache  */
  /* They are only released on *SHUTDOWN* within        */
  /* vdcom_releaseall ( eo38VDcomReleaseAll )           */
  /* exception: debugger related factories aren't       */
  /*            preserved.                              */
  
  DWORD          dwClsContext;
  LPCLASSFACTORY lpIClassFactory;
  tsp00_Uint4    session;
  GUID           CoClsId;
#if defined(USE_KGS)
  short          loc_len;
#else
  IliveCacheSink*   pIliveCacheSink;
  char              msgBuffer[132];
  memcpy( &pIliveCacheSink, &parms->ikernel_sink, sizeof(pIliveCacheSink));
#endif
  
  memcpy (&session, &parms->session[0], sizeof (session));
  parms->idispatch = 0;
#if defined(USE_KGS) && defined (WIN32)
  if (!parms->debug_mode)
  {
#endif
    if (parms->in_proc_server)
      dwClsContext = CLSCTX_INPROC_SERVER;
    else
    {
      if ( parms->debug_location == NULL )
        dwClsContext = CLSCTX_LOCAL_SERVER;
      else
        dwClsContext = CLSCTX_REMOTE_SERVER;
    }
    memcpy(&CoClsId, &parms->coclsid[0], sizeof(CoClsId));
    parms->ret = sql38CoGetClassObject (&CoClsId,
      dwClsContext,
      parms->debug_location,
      IID_IClassFactory,
      (void**)&parms->ifactory);
    lpIClassFactory = (IClassFactory *) parms->ifactory;
    if (!parms->ret)
    {
      /* Debugger */
      /* Don't lock Server or maintain Factory-Pointer within Hash    */
      /* Hash Key (COClsid + SessionID) wouldn't be unique            */
      /* A Debugger factory can only be used once ! Why maintain it ? */
      if ((memcmp (&parms->coclsid, &CLSID_liveCacheDbg, sizeof (GUID))) &&
        (memcmp (&parms->coclsid, &CLSID_LiveCacheRMT, sizeof (GUID))))
      {
        if (!parms->ret) parms->ret = lpIClassFactory->LockServer (TRUE);
        if (parms->ret) 
        {
#ifdef USE_KGS
          MSGCD ((ERR_DCOM_HRESULT, "LockServer", parms->ret ));
          MSGCD ((ERR_DCOM_GUID, "COCLSID",
            CoClsId.Data1, CoClsId.Data2, CoClsId.Data3,
            CoClsId.Data4[0], CoClsId.Data4[1], CoClsId.Data4[2],
            CoClsId.Data4[3], CoClsId.Data4[4], CoClsId.Data4[5],
            CoClsId.Data4[6], CoClsId.Data4[7] ));
#else
          sprintf(&msgBuffer[0],"CreateInstance:LockServer HRESULT: 0x%X", parms->ret );
          pIliveCacheSink->ConsoleMsg (strlen(&msgBuffer[0]), (unsigned char*)&msgBuffer[0] );
          sprintf(&msgBuffer[0],"COCLSID:{%8.8X-%4.4X-%4.4X-%2.2X%2.2X-%2.2X%2.2X%2.2X%2.2X%2.2X%2.2X}",
            CoClsId.Data1, CoClsId.Data2, CoClsId.Data3,
            CoClsId.Data4[0], CoClsId.Data4[1], CoClsId.Data4[2],
            CoClsId.Data4[3], CoClsId.Data4[4], CoClsId.Data4[5],
            CoClsId.Data4[6], CoClsId.Data4[7] );
          pIliveCacheSink->ConsoleMsg (strlen(&msgBuffer[0]), (unsigned char*)&msgBuffer[0] );
#endif
        }
        parms->ret = eo38_ComCache (&CoClsId,
          parms->ifactory,
          0,
          false,
          CLSID_eeo38);
      }
    }
#ifdef USE_KGS
    else
    {
      char cClsContext[80];
      
      memset(cClsContext,'\0',sizeof(cClsContext));
      switch ( dwClsContext )
      {
      case CLSCTX_INPROC_SERVER:
        strcpy(cClsContext, "INPROC");
        break;
      case CLSCTX_LOCAL_SERVER:
        strcpy(cClsContext, "LOCAL");
        break;
      case CLSCTX_REMOTE_SERVER:
        strcpy(cClsContext, "REMOTE AT ");
        if (parms->debug_location)
        {
          for ( loc_len= 0; parms->debug_location[loc_len] != ' '; loc_len++ );
          memcpy(&cClsContext[strlen(cClsContext)],&parms->debug_location[0], loc_len );
        }
        break;
      default:
        strcpy(cClsContext, "UNKNOWN");
        break;
      }
      MSGCD ((ERR_DCOM_COCREATE, "CoGetClassObject", parms->ret, cClsContext ));
      MSGCD ((ERR_DCOM_GUID, "COCLSID",
        CoClsId.Data1, CoClsId.Data2, CoClsId.Data3,
        CoClsId.Data4[0], CoClsId.Data4[1], CoClsId.Data4[2],
        CoClsId.Data4[3], CoClsId.Data4[4], CoClsId.Data4[5],
        CoClsId.Data4[6], CoClsId.Data4[7] ));
    }
#if defined (WIN32)
  }
  else
    os39CoGetClassobject (parms);
#endif
#endif
}

/*---------------------------------------------------------------------------*/

externC BOOLEAN eo38CoInitialize ( void )
{  
  HRESULT hr;
  hr = sql38CoInitializeEx ();
  if (FAILED (hr))
  {
#ifdef USE_KGS
    MSGALL ((ERR_COINITIALIZE, hr));
#endif
    return (0);
  }
  WAIT_UNTIL_ALONE (DCOM_EXCLUSIVE);
  if (!eo38_Initialize ())
  {
    CLEARLOCK (DCOM_EXCLUSIVE);
    return (0);
  }
  CLEARLOCK (DCOM_EXCLUSIVE);
  return (1);
}

/*---------------------------------------------------------------------------*/

externC void eo38IIdILiveCacheSpec ( tsp_c16 * iid )
{
  memcpy (iid, &IID_IliveCacheSpec, sizeof (IID_IliveCacheSpec));
}

/*---------------------------------------------------------------------------*/

externC void eo38InitializeDebugServer (tsp_dcom_co_create_instance_parms * parms)
{
  tsp00_Uint4 session;
  LPCLASSFACTORY lpIClassFactory;
  IID guid_iid;

  /* RC_DISPADDR_INUSE */
  if ( parms->ret == 3 ) return;
  
  /* this procedure is only used to ensured the initialization    */
  /* of the LCDebug Server                                        */
  
  parms->idispatch = 0;
  parms->ifactory = 0;
  parms->in_proc_server = 0;
  parms->debug_mode = 0;
  memcpy (&session, &parms->session[0], sizeof (session));
  memcpy (&guid_iid, &IID_IliveCacheDbg, sizeof (IID));
  
  eo38CoGetClassObject (parms);
  
  lpIClassFactory = (IClassFactory *) parms->ifactory;
  if (!parms->ret)
  {
    parms->ret =
      lpIClassFactory->CreateInstance (NULL,
      IID_IliveCacheDbg,
      (void**)&parms->idispatch);
    if (SUCCEEDED (parms->ret))
    {
      if ( parms->debug_location != 0 ) 
        parms->ret = eo38_ComCache (&guid_iid,
        (IUnknown*)parms->idispatch,
        session,
        true,
        DbgInterface_eeo38);
      else
        parms->ret = eo38_ComCache (&guid_iid,
        (IUnknown*)parms->idispatch,
        session,
        false,
        DbgInterface_eeo38);
    }
    lpIClassFactory->Release();
  }
  else {
    MSGCD ((ERR_DCOM_HRESULT, "eo38InitializeDebugServer: Start of liveCacheProxy/liveCacheRMT failed", 
      parms->ret ));
  }
};

/*---------------------------------------------------------------------------*/

#ifdef USE_KGS
/* method is used to cleanup after a DB-Procedure crashed */
static void eo38DisposeInstances( tsp_dcom_dispatch_parms * parms ) {
  int                       j;
  bool                      EntryFound = true;
  CCoClassHashPtr           HashPtr;
  CHashEntryPtr             currEntry;
  unsigned int              session;
  CIliveCacheSink*          pIliveCacheSink;
  tsp1_packet*              SQLBuffer_ptr;
  IUnknown                  *lpIUnknown;
  HRESULT                   hr;
  IliveCacheSpec            *pILC;
  
  MSGCD ((ERR_DCOM_DUMP_INFO, "Drop all component objects active on session"));

  ak341GetDebugContext (REINTERPRET_CAST(tak_all_command_glob*, parms->session_context),
    *REINTERPRET_CAST(tgg91_SessionNo*, &session),
    *REINTERPRET_CAST(tsp00_Addr*, &pIliveCacheSink),
    SQLBuffer_ptr );

  HashPtr = REINTERPRET_CAST(CCoClassHash*,DCOM_HEAD);
  if ( NULL == HashPtr ) {
    MSGCD ((ERR_DCOM_DUMP_INFO, "eo38DisposeInstances DCOM_HEAD == NULL"));
    vabort(false);   
  };
  
  j = 0;
  while (EntryFound)
  {
    EntryFound = false;
    WAIT_UNTIL_ALONE (DCOM_EXCLUSIVE);
    HashPtr->ResetEntrySearch( session, SearchEQ_eeo38 );
    currEntry = HashPtr->NextEntry();
    CLEARLOCK (DCOM_EXCLUSIVE);

    if (NULL != currEntry)
    {
      EntryFound = true;
      
      lpIUnknown = currEntry->getIUnknown();
      if (currEntry->getType() != DbgInterface_eeo38)
      {
        hr = lpIUnknown->QueryInterface ( IID_IliveCacheSpec, (void **) &pILC);
        if (SUCCEEDED (hr))
        {
          TRY {
            hr = pILC->sysCleanUp ();
          }
#if defined(WIN32)
          __except (eo38_Eval_Exception (parms, GetExceptionInformation ())) {
            hr = parms->hresult;
          };
#else
          catch (RTE_DBProcError e) {
             hr = GEO00D_DCOM_EXCEPTION_RAISED;
          }
          catch (DbpError e) { 
             en38_Eval_Exception( parms );
             hr = GEO00D_DCOM_DBPROC_CRASHED; 
          }
#endif

          if ( FAILED(hr)) {  
            IID loc_iid;
            char    msgBuffer[132];
       
            MSGCD ((ERR_DCOM_HRESULT, "eo38DisposeInstances:sysCleanUp", hr ));
            memcpy (&loc_iid.Data1, currEntry->getGuid(), sizeof (loc_iid));
            sprintf(&msgBuffer[0],"IID:{%8.8X-%4.4X-%4.4X-%2.2X%2.2X-%2.2X%2.2X%2.2X%2.2X%2.2X%2.2X} hr:%d",
              loc_iid.Data1, loc_iid.Data2, loc_iid.Data3,
              loc_iid.Data4[0], loc_iid.Data4[1], loc_iid.Data4[2],
              loc_iid.Data4[3], loc_iid.Data4[4], loc_iid.Data4[5],
              loc_iid.Data4[6], loc_iid.Data4[7], hr);
            MSGCD ((ERR_DCOM_DUMP_INFO, &msgBuffer[0]));
          }
          pILC->Release ();
        }
        else
          MSGCD ((ERR_DCOM_HRESULT, "eo38DisposeInstances:QILC", hr ));
      }

      eo38_ReleaseInstance( currEntry, parms->session_context );
      WAIT_UNTIL_ALONE (DCOM_EXCLUSIVE);
      HashPtr->ResetEntrySearch( session, SearchEQ_eeo38 );
      currEntry = HashPtr->GetRechainFirstEntry(NULL);
      CLEARLOCK (DCOM_EXCLUSIVE);
    }
  }

  MSGCD ((ERR_DCOM_DUMP_INFO, "Drop of component objects done"));
};
#endif

/*---------------------------------------------------------------------------*/

externC void eo38ReleaseInstance ( tsp_dcom_release_instance_parms * parms )
{
  int                       j;
  tsp00_Uint4               session;
  bool                      EntryFound = true;
  CCoClassHashPtr           HashPtr;
  CHashEntryPtr             currEntry;
  
  HashPtr = REINTERPRET_CAST(CCoClassHash*,DCOM_HEAD);
  if ( NULL == HashPtr ) return;
  
  j = 0;
  while (EntryFound)
  {
    EntryFound = false;
    WAIT_UNTIL_ALONE (DCOM_EXCLUSIVE);
    memcpy (&session, &parms->session[0], sizeof (session));
#ifdef USE_KGS
    HashPtr->ResetEntrySearch( session, SearchEQ_eeo38 );
#else
    HashPtr->ResetEntrySearch( 0, SearchNE_eeo38 );
#endif
    currEntry = HashPtr->NextEntry();
    CLEARLOCK (DCOM_EXCLUSIVE);

    if (NULL != currEntry)
    {
      EntryFound = true;
      eo38_ReleaseInstance( currEntry, parms->session_context );
      WAIT_UNTIL_ALONE (DCOM_EXCLUSIVE);
#ifdef USE_KGS
      HashPtr->ResetEntrySearch( session, SearchEQ_eeo38 );
#else
      HashPtr->ResetEntrySearch( 0, SearchNE_eeo38 );
#endif
      currEntry = HashPtr->GetRechainFirstEntry(NULL);
      CLEARLOCK (DCOM_EXCLUSIVE);
    }
  }
}

/*---------------------------------------------------------------------------*/

externC void eo38ReleaseProxies ()
{
  /* Routine only called during exception handling (vco50kc) */
  /* Routine should clear up all Debug Processes, to avoid   */
  /* Debugger Processes to stay alive after liveCache Crash  */
  
  IUnknown                *lpIUnknown;
  CCoClassHash            *HashPtr;
  CHashEntryPtr           currEntry;
  
  HashPtr = REINTERPRET_CAST(CCoClassHash*, (DCOM_HEAD));
  if ( NULL == HashPtr ) return;
  
  HRESULT hr = sql38CoInitializeEx ();
  if (FAILED(hr)) return;
  
  /* hopefully we didn't crash within critical region !!     */
  /* ELSE: liveCache busy-loop forever    *CHANGE*           */
  WAIT_UNTIL_ALONE (DCOM_EXCLUSIVE);
  
  /* Release session related debugger                        */
  HashPtr->ResetEntrySearch( 0, SearchALL_eeo38 );
  currEntry = HashPtr->NextEntry();  
  while ( NULL != currEntry)
  {
    if (currEntry->getType()  == DbgInterface_eeo38)
    {
      lpIUnknown = currEntry->getIUnknown();
      TRY { lpIUnknown->Release (); }
      CATCH_ALL { 
        MSGCD ((ERR_DCOM_DUMP_INFO,
          "eo38ReleaseProxies:Release of debugger failed (session)"));
      };
    }
    currEntry = HashPtr->NextEntry();  
  }
  
  CLEARLOCK (DCOM_EXCLUSIVE);
}

/*---------------------------------------------------------------------------*/

externC void eo38RemoveProxy ( tsp00_Uint4 session )
{
  /* Routine is called on DB-Session-Release     */
  /* If there is a Debugger related to this      */
  /* session -> release it.                      */
  /*                                             */
  
  IUnknown              *lpIUnknown;
  CCoClassHash          *HashPtr;
  CHashEntryPtr         currEntry;
  teo38_HashEntryType  entryType = DbgInterface_eeo38;
  HashPtr = REINTERPRET_CAST(CCoClassHash*,(DCOM_HEAD));
  if ( NULL == HashPtr ) return;
  
  WAIT_UNTIL_ALONE (DCOM_EXCLUSIVE);
  HashPtr->ResetEntrySearch( session, SearchALL_eeo38 );
  currEntry = HashPtr->GetRechainFirstEntry(&entryType);  
  while ( NULL != currEntry)
  {
    lpIUnknown = currEntry->getIUnknown();
    CLEARLOCK (DCOM_EXCLUSIVE);
    TRY { lpIUnknown->Release (); }
    CATCH_ALL { 
      MSGCD ((ERR_DCOM_DUMP_INFO,
        "eo38RemoveProxy:Release of debugger failed"));
    };
    WAIT_UNTIL_ALONE (DCOM_EXCLUSIVE);
    HashPtr->ResetEntrySearch( 0, SearchALL_eeo38 );
    currEntry = HashPtr->GetRechainFirstEntry(&entryType);  
  }
  CLEARLOCK (DCOM_EXCLUSIVE);
}

/*------------- +++++ PTS 1106505 FF +++++++++++++++ ------------------------*/

externC void eo38CTraceBack( tsp00_Addr hd )
{
#ifdef USE_KGS 
#if defined (WIN32)
   eo38_Dump_Exception ( REINTERPRET_CAST(LPEXCEPTION_POINTERS, hd) );
#endif
#endif
};

/* --- PTS 1106505 FF ---------------  */

/*---------------------------------------------------------------------------*/

externC void eo38UnInitialize ( void )
{
  sql38CoUninitialize ();
  return;
}

/*---------------------------------------------------------------------------*/

externC void eo38VDcomCreateInstance ( tsp_dcom_co_create_instance_parms * parms )
{
  LPCLASSFACTORY lpIClassFactory;
  IUnknown *lpIUnknown;
  IliveCacheSpec *pILC;
  tsp00_Uint4 session;
  IID   IId;
#ifndef USE_KGS
  IliveCacheSink*   pIliveCacheSink;
  char              msgBuffer[132];
  memcpy( &pIliveCacheSink, &parms->ikernel_sink, sizeof(pIliveCacheSink));
#endif
  
#ifdef USE_KGS
#if defined (WIN32)
  if (!parms->debug_mode)
  {
#endif
#endif
    memcpy(&IId, &parms->iid[0], sizeof(IId));
    lpIClassFactory = (IClassFactory *) parms->ifactory;
    parms->ret = lpIClassFactory->CreateInstance ( NULL,
      IId,
      (void **)&(parms->idispatch));
    if (SUCCEEDED (parms->ret))
    {
      lpIUnknown = (IUnknown *) parms->idispatch;
      parms->ret = lpIUnknown->QueryInterface ( IID_IliveCacheSpec,(void **) &pILC );
      if (SUCCEEDED (parms->ret))
      {
        tsp_dcom_dispatch_parms disp_parms;
        disp_parms.idispatch = parms->idispatch;
        disp_parms.ifactory = parms->ifactory;
        disp_parms.dispid = -2;
#ifdef USE_KGS
        disp_parms.session_context = (char*)((CIliveCacheSink*)parms->ikernel_sink)->lpSessionContext;
        ((CIliveCacheSink*)parms->ikernel_sink)->CallbackIsActive = false;
#else
        memcpy( &disp_parms.session_context, &parms->ikernel_sink, sizeof(disp_parms.session_context));
#endif
        TRY
        {
          parms->ret = pILC->Advise ( (IUnknown *) parms->ikernel_sink);
        }
#if defined(WIN32)
        __except (eo38_Eval_Exception (&disp_parms, GetExceptionInformation ())) 
        {
          parms->ret = disp_parms.hresult;
        }
#else
        catch (RTE_DBProcError) {
          parms->ret = -1;
        }
        catch (DbpError &e) { 
          parms->ret = e.dbpError();
        }
#endif
        if (!parms->ret)
        {
          if (memcmp (&parms->iid[0], &IID_IliveCacheSpec, sizeof (parms->iid)))
          TRY 
          { 
            parms->ret = pILC->sysMonitorInit (); /* +++ Change Request 1000139 +++ */
            parms->ret = pILC->Register (); 
          }
#if defined(WIN32)
          __except (eo38_Eval_Exception (&disp_parms, GetExceptionInformation ())) 
          {
            parms->ret = disp_parms.hresult;
#else
            catch (RTE_DBProcError) {
              parms->ret = -1;
            }
            CATCH_ALL { 
              parms->ret = ERR_DCOM_COMETHOD_CRASHED;
#endif
              if ( parms->ret != GEO00D_DCOM_TRANSEND_FAILED )
              { 
#ifdef USE_KGS
                MSGCD ((ERR_DCOM_HRESULT, "CreateInstance:Register", parms->ret ));
#else
                sprintf(&msgBuffer[0],"CreateInstance:Register HRESULT: 0x%X", parms->ret );
                pIliveCacheSink->ConsoleMsg (strlen(&msgBuffer[0]), (unsigned char*)&msgBuffer[0] );
#endif
              }
            }
            pILC->Release ();
            memcpy (&session, &parms->session[0], sizeof (session));
            if (!parms->ret)
              parms->ret = eo38_ComCache (REINTERPRET_CAST(IID *, & (parms->iid)),
              parms->idispatch,
              session,
              false,
              IID_eeo38);
          }
          else
          {
#ifdef USE_KGS
            MSGCD ((ERR_DCOM_HRESULT, "eo38VDcomCreateInstance:Advise", parms->ret ));
#else
            sprintf(&msgBuffer[0],"eo38VDcomCreateInstance:Advise HRESULT: 0x%X", parms->ret );
            pIliveCacheSink->ConsoleMsg (strlen(&msgBuffer[0]), (unsigned char*)&msgBuffer[0] );
#endif
          }
        }
        else
        {
#ifdef USE_KGS
          MSGCD ((ERR_DCOM_HRESULT, "eo38VDcomCreateInstance:QI(ILCSpec)", parms->ret ));
#else
          sprintf(&msgBuffer[0],"eo38VDcomCreateInstance:QI(ILCSpec) HRESULT: 0x%X", parms->ret );
          pIliveCacheSink->ConsoleMsg (strlen(&msgBuffer[0]), (unsigned char*)&msgBuffer[0] );
#endif
        }
      }
      else
      {
#ifdef USE_KGS
        MSGCD ((ERR_DCOM_HRESULT, "eo38VDcomCreateInstance:CI", parms->ret ));
#else
        sprintf(&msgBuffer[0],"eo38VDcomCreateInstance:CI HRESULT: 0x%X", parms->ret );
        pIliveCacheSink->ConsoleMsg (strlen(&msgBuffer[0]), (unsigned char*)&msgBuffer[0] );
#endif
      }
#ifdef USE_KGS
#if defined (WIN32)
    }
    else
      os39VDcomCreateInstance (parms);
#endif
#endif
  }
  
/*---------------------------------------------------------------------------*/
  
static void eo38VDcomDispatch ( tsp_dcom_dispatch_parms * parms )
{
#if defined(USE_KGS)
  IliveCacheSpec  *pILC;
  tsp00_Int4      start_time_sec; 
  tsp00_Int4      start_time_msec;
  tsp00_Int4      stop_time_sec; 
  tsp00_Int4      stop_time_msec;
  tsp00_Int4      diff_time_sec; 
  tsp00_Int4      diff_time_msec;
#endif
  
  tsp00_Uint4     param_stack_size;
  
#if defined(LINUX) 
  //    vdebug_break(10001);
#endif
  
	tsp00_Uint										param_num;
  tak_dbproc_stack_param_debug *pDebugStack;
  IUnknown										 *lpIUnknown;
  IUnknown										 *lpClassPtr;
  t_lpIFunc											lp_IFunc;
  t_lpptr												lpVTable;

#if defined(WIN32) && defined(USE_KGS)
  PTASK_CTRL_REC pTaskCtrl = THIS_UKT_CTRL->pCTask;
  pTaskCtrl->TaskState = TSK_DCOM_OBJ_CALLED;
#endif
  lpIUnknown = (IUnknown *) parms->idispatch;
  lpClassPtr = lpIUnknown;

  lpVTable = (t_lpptr) * (tsp00_Longint *) lpIUnknown;

  param_num = parms->param_stack_len / sizeof(tak_dbproc_stack_param_debug);
  pDebugStack = REINTERPRET_CAST(tak_dbproc_stack_param_debug*,parms->param_stack);

#if defined(_IBMR2) || defined(SUN) || defined(PA11) || defined(PA20W) || defined(ALPHA) || defined(NMP)
  /* calculate entry address of method to call      */
#if defined(PA11) || defined(PA20W)
  lp_IFunc    = (t_lpIFunc) lpVTable[parms->dispid+4];
#elif defined(ALPHA) || ( defined(_IBMR2) && defined(BIT64) )
  lp_IFunc    = (t_lpIFunc) lpVTable[parms->dispid];
#elif defined(SUN) && (__SUNPRO_CC >= 0x0500)   /* PTS 1105381 *//* 15-Nov-2000 FF chg == to >= */ 
  lp_IFunc    = (t_lpIFunc) lpVTable[parms->dispid+2];
#else
  lp_IFunc    = (t_lpIFunc) lpVTable[parms->dispid+1].addr;
  /* calculate 'this' pointer of COM-Object           */
  lpClassPtr = (IUnknown*)((long)lpIUnknown + lpVTable[parms->dispid+1].this_displ);
#endif
  param_stack_size = sizeof(void *);      /* size for this pointer */
  while ( param_num > 0 )
  {
    if ( '\00' == pDebugStack->dbg_is_ptr )
    {
      /* C++ pushes 1&2 Bytes Types as words   */
      if ( pDebugStack->dbg_length <= MIN_PARAMETER_SIZE ) {
        param_stack_size = param_stack_size + MIN_PARAMETER_SIZE;
      }
      else  
      {
#if defined(PA11) || defined(BIT64)
        param_stack_size = ( param_stack_size + 7 ) & 0xFFFFFFF8;   /* align to double word */
#endif
#if defined(NMP)
        /* double is passed as 2 Fullwords 8 Byte aligend on NMP */
        if ( pDebugStack->dbg_vt == csp1_vt_r8 ) {
          param_stack_size = (param_stack_size + 7) & 0xFFFFFFF8;
          param_stack_size = param_stack_size + 8;
        }
        else
#endif
          param_stack_size = param_stack_size + pDebugStack->dbg_length;
      }
    }
    else {
#if defined(BIT64)
      param_stack_size = (param_stack_size + 7) & 0xFFFFFFF8;
#endif
      param_stack_size = param_stack_size + sizeof(void *);
    }
    param_num--;
    pDebugStack++;
  }
#elif defined(LINUX) || defined (_WIN32)
#if defined(LINUX) && __GNUC__ < 3
  lp_IFunc = (t_lpIFunc) lpVTable[parms->dispid+2];
#elif defined(LINUX) && (defined I386 && __GNUC__ == 3 && __GNUC_MINOR__ >= 2)
  lp_IFunc = (t_lpIFunc) lpVTable[parms->dispid];
#elif defined(LINUX) && (defined IA64 && __GNUC__ == 3 && __GNUC_MINOR__ >= 1)
  lp_IFunc = (t_lpIFunc) &lpVTable[2 * parms->dispid];
#elif defined(_WIN32)
  lp_IFunc = (t_lpIFunc) lpVTable[parms->dispid];
#endif
  param_stack_size = parms->param_stack_len;
#if defined	(_WIN64)
	tak_dbproc_stack_param	param_stack[cak260max_params];	//Define verwenden
	tak_dbproc_stack_param	fparam_stack[cak260max_fparams];	//Define verwenden

	tsp00_Uint				len = 0;
	tsp00_Uint				flen = 0;


	while (param_num > 0)
	{
    /* The first 8 floating point parameters go into the floating point    */
		/* parameter stack.  This allows the assembly code to easily determine */
		/* which parameters are passed in floating point registers.            */
    if ( pDebugStack->dbg_vt == csp1_vt_r8 ) 
		{
			if (flen < cak260max_fparams)
			{
				// Double precision floats are copied onto the floating point parameter
				// stack in their memory format.
				fparam_stack[flen].variant.C_6.r8_F = pDebugStack->dbg_value.r8;
				flen++;
			}
			else
			{
				MSGCD ((ERR_DCOM_MAX_FPARAM, cak260max_fparams));
			}
				//
    } /* end if */
    else if ( pDebugStack->dbg_vt == csp1_vt_r4 ) 
		{
			if (flen < cak260max_fparams)
			{
				// Single precision floats are converted to double-precision floats 
				// before placing them on the floating point parameter stack because 
				// the assembly code can't tell the difference between single and 
				// double precision.  This does not affect the target function.  The 
				// target function receives the parameter either as a floating point 
				// register or on the general parameter stack, where it is left 
				// in single precision memory format.
				fparam_stack[flen].variant.C_6.r8_F = pDebugStack->dbg_value.r4;
				flen++;
			}
			else
			{
				MSGCD ((ERR_DCOM_MAX_FPARAM, cak260max_fparams));
			}
    } /* end else-if */

    /* All parameters (including floating point parameters) are placed */
    /* in order onto the general parameter stack.                      */
		if (len < cak260max_params)
		{

			param_stack[len].variant.C_7.ptr_F = pDebugStack->dbg_value.ptr;
			len++;
			param_num--;
			pDebugStack++;
		}
		else
		{
			MSGCD ((ERR_DCOM_MAX_PARAM, cak260max_params));
			ABORT ();
		}
  } /* end for */

  len <<= 3; /* len = len * 8; // Convert len to bytes. */
  flen <<= 3; /* flen = flen * 8; // Convert flen to bytes. */
#endif

#else
  lp_IFunc = (t_lpIFunc) lpVTable[parms->dispid];
#endif
  
  /*--* +++ FF Change Request 1000139  +++ *-----------------------------------*/
#if defined USE_KGS
  if ( false &&     // PTS 1122612 FF no monitoring within V73
      parms->monitoring )
  {
    TRY 
    {
      parms->hresult = lpIUnknown->QueryInterface (IID_IliveCacheSpec, (void **)&pILC);
      if (FAILED(parms->hresult))
      {
        MSGCD ((ERR_DCOM_HRESULT, "os38kVDcomDispatch:IID_IliveCacheSpec", parms->hresult));
        return;
      }
      pILC->sysMonitorProlog(parms->monitor_slot_id, 
        REINTERPRET_CAST(unsigned char*,&parms->method_name) );
      vclock (&start_time_sec, &start_time_msec);
    }
    CATCH_ALL 
    {
      MSGCD ((ERR_DCOM_DUMP_INFO,"os38kVDcomDispatch:sysMonitorProlog : unhandled exception")); /* PTS 1107168 */
    }
  }
#endif
  /*--* --- FF Change Request 1000139  --- *-----------------------------------*/
  TRY             
  {
    /* following has been coded in Assembler          */
    /* to push the exact length of needed Parms       */
#if defined (_WIN64)
	sql38aVdcomCall(lpClassPtr,		/* class this pointer					*/ 
      (PROC)lp_IFunc,           /* entry-point								*/
      param_stack,							/* call-stack									*/ 
      len,											/* call-stack length					*/
			fparam_stack,							/* call-stack (floats)				*/ 
      flen,											/* call-stack length (floats)	*/
      param_stack_size);        /* param-stack size: is not needed for the assembly in this case */
#else
#if defined(SUN) && defined(BIT64)
    /* PTS 1111288 */ /* round up to multiple of 16 bytes */
    param_stack_size = ( param_stack_size + 15 ) & (~15);
#endif
    sql38aVdcomCall(lpClassPtr,  /* class this pointer*/ 
      (PROC)lp_IFunc,            /* entry-point       */
      parms->param_stack,        /* call-stack        */ 
      parms->param_stack_len,    /* call-stack length */
      param_stack_size);         /* param-stack size  */
#endif
    parms->hresult = 0;
  }
#if defined(_WIN32)
  __except (eo38_Eval_Exception (parms, GetExceptionInformation ()))
  {
#ifdef USE_KGS
    if ( parms->hresult != GEO00D_DCOM_EXCEPTION_RAISED ) {
      MSGCD ((ERR_DCOM_HRESULT, "eo38VDcomDispatch", parms->hresult ));
    }
#endif
  }
#else
  catch (RTE_DBProcError e) {
    if (e_terminate == e.error_number) { /* PTS 1107676 */
      parms->hresult = GEO00D_DCOM_TERMINATED;
    }
    else {
      parms->hresult = GEO00D_DCOM_EXCEPTION_RAISED;
    }
  }
  catch (DbpError e) { 
#ifdef USE_KGS
    char    msgBuffer[132];
    sprintf(&msgBuffer[0],"eo38aVDcomCall: DbpError:%d", e.dbpError() );
    MSGCD ((ERR_DCOM_DUMP_INFO,msgBuffer));
#endif
    en38_Eval_Exception( parms );
    parms->hresult = GEO00D_DCOM_DBPROC_CRASHED; }
#endif
  
#ifdef USE_KGS
  /*--* +++ FF Change Request 1000139  +++ *-----------------------------------*/
  if ( false &&     // PTS 1122612 FF no monitoring within V73
     ( parms->monitoring && ( 0 == parms->hresult )))
  {
    TRY
    {
      vclock (&stop_time_sec, &stop_time_msec);
      diff_time_sec  = stop_time_sec - start_time_sec;
      diff_time_msec = stop_time_msec - start_time_msec;
      if (stop_time_msec < start_time_msec)
      {
        diff_time_sec  = diff_time_sec - 1;
        diff_time_msec = diff_time_msec + 1000000 ;
      }
      pILC->sysMonitorEpilog(parms->monitor_slot_id, diff_time_sec * 1000 + diff_time_msec / 1000);
      pILC->Release();
    }
    CATCH_ALL 
    {
      MSGCD ((ERR_DCOM_DUMP_INFO,"os38kVDcomDispatch:sysMonitorEpilog : unhandled exception")); /* PTS 1107168 */
    }
  }
  /*--* --- FF Change Request 1000139  --- *-----------------------------------*/
#ifdef WIN32
  pTaskCtrl->TaskState = TSK_RUNNING;
#endif
#endif
  return;
}

/*---------------------------------------------------------------------------*/

externC void eo38VDcomDBProc ( tsp_dcom_dispatch_parms * parms )
{
#ifdef USE_KGS
  tsp00_C4          session;
  CIliveCacheSink*  pIliveCacheSink;
  tsp1_packet*      SQLBuffer_ptr;
  
  ak341GetDebugContext (REINTERPRET_CAST(tak_all_command_glob*, parms->session_context),
    *REINTERPRET_CAST(tgg91_SessionNo*, &session),
    *REINTERPRET_CAST(tsp00_Addr*, &pIliveCacheSink),
    SQLBuffer_ptr );
  pIliveCacheSink->CallbackIsActive = false;
  pIliveCacheSink->sqlErrorFlag     = false;   /* PTS 1110813 FF  */
  
#if defined (WIN32)
  if (parms->debug_mode)
  {
    os39VDcomDBProc (parms);
  }
  else
#endif
#endif
    eo38VDcomDispatch (parms);
#ifdef USE_KGS
  if (( GEO00D_DCOM_DBPROC_CRASHED == parms->hresult ) || ( GEO00D_DCOM_TERMINATED == parms->hresult )) {
      eo38DisposeInstances( parms );
      pIliveCacheSink->lpDefaultContext = NULL;
  }
  else                                                          /* PTS 1110813 FF  */
  {                                                             /* PTS 1110813 FF  */
    if ( pIliveCacheSink->sqlErrorFlag ) parms->hresult = -1;   /* PTS 1110813 FF  */
  }                                                             /* PTS 1110813 FF  */

#endif
}

/*---------------------------------------------------------------------------*/

externC void eo38VDcomGetDispAddr ( tsp_dcom_co_create_instance_parms * parms )
{
  GUID                    guid_coclsid;
  IID                     guid_iid;
  tsp00_Uint4             session;
  CCoClassHashPtr         HashPtr;
  CHashEntryPtr           currPtr;
  
  memcpy (&guid_coclsid, &parms->coclsid[0], sizeof (parms->coclsid));
  memcpy (&guid_iid, &parms->iid[0], sizeof (parms->iid));
  memcpy (&session, &parms->session[0], sizeof (session));
  parms->ret = RC_DISPADDR_FAILED;
  parms->idispatch = 0;
  parms->ifactory = 0;
  
#if defined(WIN32) && defined(USE_KGS)
  if (parms->debug_mode)
  {
    /* initialize Debug Server */
    os39VDcomGetDispAddr (parms);
    return;
  };
#endif
  
  HashPtr = REINTERPRET_CAST(CCoClassHashPtr, (DCOM_HEAD));
  if (NULL == HashPtr) return;
  
  WAIT_UNTIL_ALONE (DCOM_EXCLUSIVE);
  /* check if IID in cache ( Interface-Pointer for UDF's )   */
  
#ifdef USE_KGS
  currPtr = HashPtr->FindHashEntry( &guid_iid, session );
#else
  currPtr = HashPtr->FindHashEntry( &guid_iid, DEBUG_SESSION_ID );
#endif
  
  if (NULL != currPtr)
  {
    /* the ILiveCacheDbg Interface is used within LOCAL and REMOTE Servers */
    if (currPtr->getType() == DbgInterface_eeo38) 
    {
      parms->idispatch = REINTERPRET_CAST(char *,(currPtr->getIUnknown())); 
      currPtr->setDebugFlag(true);
      parms->ret = RC_DISPADDR_OK;
      CLEARLOCK (DCOM_EXCLUSIVE);
      return;
      //        }
    }
    else
    {
      parms->idispatch = REINTERPRET_CAST(char *,(currPtr->getIUnknown())); 
      if ( currPtr->getType() == Invalid_IID_eeo38 )
      {
        eo38_Invalidate( parms, &guid_iid, session );
      }
      else
      {
        parms->ret = RC_DISPADDR_OK;
        CLEARLOCK (DCOM_EXCLUSIVE);
      }
      return;
    }
  }
  
  /* check if CLSID in cache ( DBPROC Factory-Pointer)   */
  currPtr = HashPtr->FindHashEntry( &guid_coclsid, 0 );
  if ( NULL != currPtr )
  {
    parms->ifactory = REINTERPRET_CAST(char *,(currPtr->getIUnknown()));
    if ( currPtr->getType() == Invalid_CLSID_eeo38 )
    {
      parms->ret = RC_DISPADDR_INVALID_FOUND;
    }
    else
    {
      parms->ret = RC_DISPADDR_FACTORY_FOUND;
    }
  }
  CLEARLOCK (DCOM_EXCLUSIVE);
  return;
}

/*---------------------------------------------------------------------------*/

externC void eo38VDcomInitSink ( tsp_dcom_init_sink_parms * parms )
{
  CIliveCacheSink *lpCliveCacheSink;
  
  if (sizeof (CIliveCacheSink) > parms->sink_size)
  {
#ifdef USE_KGS
    MSGCD ((ERR_DCOM_DUMP_INFO, "eo38VDcomInitSink: sizeof(CIliveCacheSink) > reserved space for Object"));
    vabort(false);   
#endif
  }
  else
  {
    lpCliveCacheSink  = new ( parms->sink_addr ) CIliveCacheSink;
    lpCliveCacheSink->lpSessionContext = 
      REINTERPRET_CAST(tak_all_command_glob*,(parms->acv_addr));
  }
  return;
}

/*--* +++ FF Change Request 1000139  +++ *-----------------------------------*/
#if defined(USE_KGS)

externC void os38kVDcomMonOpen ( tsp_dcom_monitor_parms * parms )
{
  CCoClassHashPtr         HashPtr;
  CHashEntryPtr           currSessEntry, newSessEntry;
  IUnknown*               lpIUnknown;
  HRESULT                 hr;
  
  parms->session_handle   = NULL;
  parms->retcode          = -1;
  parms->idispatch        = NULL;
  
  WAIT_UNTIL_ALONE (DCOM_EXCLUSIVE);
  HashPtr = REINTERPRET_CAST(CCoClassHashPtr, (DCOM_HEAD));
  if (NULL != HashPtr) 
  {
    HashPtr->ResetEntrySearch( 0, SearchALL_eeo38 );
    currSessEntry = HashPtr->NextEntry();
    while ( currSessEntry )
    {
      /* only Interfaces, no Factories, no IDebug, no IID_IliveCacheSpec Interface */
      if ((IID_eeo38 == currSessEntry->getType()) &&
        (IID_IliveCacheSpec != *(currSessEntry->getGuid())))
      {
        /* get IUnknown for IID_IliveCacheSpec Interface */
        TRY
        {
          hr = (currSessEntry->getIUnknown())->QueryInterface(IID_IliveCacheSpec, 
            REINTERPRET_CAST(void **,&lpIUnknown)); 
        }
        CATCH_ALL
        {
          MSGCD ((ERR_DCOM_DUMP_INFO,"os38kVDcomMonOpen:QueryInterface : unhandled exception")); /* PTS 1107168 */
          hr = -1;
        }
        if (SUCCEEDED(hr))
        {
          newSessEntry = HashPtr->getFreeEntry();
          newSessEntry->init( currSessEntry->getSession(), 
            currSessEntry->getGuid(), 
            lpIUnknown, 
            currSessEntry->getRemoteDbgFlag(), 
            currSessEntry->getType());
          newSessEntry->setNextEntry((CHashEntryPtr)parms->session_handle);
          parms->session_handle = (tsp00_Addr)newSessEntry;
        }
      }
      currSessEntry = HashPtr->NextEntry();
    }
  }
  CLEARLOCK (DCOM_EXCLUSIVE);
  if ( NULL == parms->session_handle )
  {
    parms->idispatch  = NULL;
    parms->retcode    = -1;
  }
  else
  {
    tsp00_Uint4  session = newSessEntry->getSession();
    newSessEntry = REINTERPRET_CAST(CHashEntryPtr,parms->session_handle);
    parms->idispatch = REINTERPRET_CAST(char *, newSessEntry->getIUnknown());
    memcpy (&parms->iid[0], newSessEntry->getGuid(), sizeof (parms->iid));
    memcpy(&parms->session[0], &session, sizeof(parms->session)); 
    parms->retcode = S_OK;
  }
}

/*---------------------------------------------------------------------------*/

externC void os38kVDcomMonNext ( tsp_dcom_monitor_parms * parms )
{
  CHashEntryPtr           currSessEntry, newSessEntry;
  CCoClassHashPtr         HashPtr;

  parms->retcode = S_OK;

  if ( NULL == parms->session_handle )
  {
    parms->idispatch  = NULL;
    parms->retcode    = -1;
    return;
  }

  currSessEntry = REINTERPRET_CAST(CHashEntryPtr,parms->session_handle);
  if (parms->idispatch == REINTERPRET_CAST(char *, currSessEntry->getIUnknown()))
  {
    TRY {
      /* decrement count for previous instance */
      (REINTERPRET_CAST(IUnknown *, parms->idispatch))->Release();
    }
    /* if error -> write message & ignore error */
    CATCH_ALL {
      MSGCD ((ERR_DCOM_DUMP_INFO,"os38kVDcomMonNext:Release : unhandled exception")); /* PTS 1107168 */
    }
    
    newSessEntry  = currSessEntry->getNextEntry();
    
    /* re-chain to free chain */
    WAIT_UNTIL_ALONE (DCOM_EXCLUSIVE);
    HashPtr = REINTERPRET_CAST(CCoClassHashPtr, (DCOM_HEAD));
    HashPtr->chainFreeEntry(currSessEntry);
    CLEARLOCK (DCOM_EXCLUSIVE);
    
    if ( newSessEntry != NULL )
    {
      parms->session_handle = (tsp00_Addr)newSessEntry;
      tsp00_Uint4  session = newSessEntry->getSession();
      newSessEntry = REINTERPRET_CAST(CHashEntryPtr,parms->session_handle);
      parms->idispatch = REINTERPRET_CAST(char *, newSessEntry->getIUnknown());
      memcpy (&parms->iid[0], newSessEntry->getGuid(), sizeof (parms->iid));
      memcpy(&parms->session[0], &session, sizeof(parms->session)); 
    }
    else
    {
      parms->session_handle = NULL;
      parms->idispatch  = NULL;
      parms->retcode    = -1;
      return;
    }
  }
  else
  {
    MSGCD ((ERR_DCOM_DUMP_INFO,"os38kVDcomMonNext: parameter incorrect (heap not cleared)"));
    parms->retcode = -1;
  }
}

/*---------------------------------------------------------------------------*/

externC void os38kVDcomMonClose ( tsp_dcom_monitor_parms * parms )
{
  CHashEntryPtr           currSessEntry, newSessEntry;
  CCoClassHashPtr         HashPtr;
  parms->retcode = S_OK;
  
  if (NULL == parms->session_handle ) return;
  
  currSessEntry = REINTERPRET_CAST(CHashEntryPtr,parms->session_handle);
  while (parms->idispatch == REINTERPRET_CAST(char *, currSessEntry->getIUnknown()))
  {
    TRY {
      /* decrement count for previous instance */
      (REINTERPRET_CAST(IUnknown *, parms->idispatch))->Release();
    }
    /* if error -> write message & ignore error */
    CATCH_ALL {
      MSGCD ((ERR_DCOM_DUMP_INFO,"os38kVDcomMonNext:Release : unhandled exception")); /* PTS 1107168 */
    }
    
    newSessEntry  = currSessEntry->getNextEntry();
    
    /* re-chain to free chain */
    WAIT_UNTIL_ALONE (DCOM_EXCLUSIVE);
    HashPtr = REINTERPRET_CAST(CCoClassHashPtr, (DCOM_HEAD));
    HashPtr->chainFreeEntry(currSessEntry);
    CLEARLOCK (DCOM_EXCLUSIVE);
    currSessEntry = newSessEntry;
  }
  parms->session_handle = NULL;
  
  parms->idispatch = 0; 
}

#endif

/*--* --- FF Change Request 1000139  --- *-----------------------------------*/

/*---------------------------------------------------------------------------*/

externC void eo38VDcomInvalidate ( tsp_dcom_co_create_instance_parms * parms )
{
  IID   actIId;
  CLSID actCoClsId;
  CHashEntryPtr  currEntry;
  CCoClassHash  *HashPtr = REINTERPRET_CAST(CCoClassHash*,(DCOM_HEAD));
  
  if ( NULL == HashPtr ) return;
  
  memcpy( &actIId, &parms->iid[0], sizeof(actIId)); 
  memcpy( &actCoClsId, &parms->coclsid[0], sizeof(actCoClsId)); 
  
  WAIT_UNTIL_ALONE (DCOM_EXCLUSIVE);
  HashPtr->ResetEntrySearch( 0, SearchALL_eeo38 );
  currEntry = HashPtr->NextEntry();  
  while ( NULL != currEntry)
  {
    if ((*(currEntry->getGuid())  == actIId) ||
      (*(currEntry->getGuid())  == actCoClsId))  
    {
      switch(currEntry->getType())
      {
      case CLSID_eeo38: 
        currEntry->setType(Invalid_CLSID_eeo38); 
        break;
      case IID_eeo38: 
        currEntry->setType(Invalid_IID_eeo38); 
        break;
      case DbgFactory_eeo38:
      case DbgInterface_eeo38:
      default:
        break;
      }
    }
    currEntry = HashPtr->NextEntry();  
  }
  CLEARLOCK (DCOM_EXCLUSIVE);
}

/*---------------------------------------------------------------------------*/
externC void eo38VDcomReleaseAll ( tsp00_Addr pI )
{
  LPCLASSFACTORY        lpIClassFactory;
  tsp00_Uint4           cnt;
  CHashEntryPtr         currEntry;
  CCoClassHash          *HashPtr;
#ifndef USE_KGS
  IUnknown              *lpIUnknown;
  HRESULT               hr;
  char                  msgBuffer[132];
  IliveCacheSpec        *pILC;
  IliveCacheSink*       pIliveCacheSink = REINTERPRET_CAST( IliveCacheSink*, pI);
#endif
  
  HashPtr = REINTERPRET_CAST(CCoClassHash*,(DCOM_HEAD));
  if ( NULL == HashPtr ) return;
  
#ifndef USE_KGS
  HashPtr->ResetEntrySearch( 0, SearchNE_eeo38 );
  currEntry = HashPtr->GetRechainFirstEntry(NULL);  

  while ( NULL != currEntry)
  {
    /* within liveCacheProxy all instances have to be closed */
    TRY
    {
      lpIUnknown = currEntry->getIUnknown();
      hr = lpIUnknown->QueryInterface ( IID_IliveCacheSpec, (void **) &pILC);
      if (SUCCEEDED (hr))
      {
        hr = pILC->Unadvise ();
        if ( FAILED(hr))  {
          sprintf(&msgBuffer[0],"eo38VDcomReleaseAll:Unadvise HRESULT: 0x%X", hr );
          pIliveCacheSink->ConsoleMsg (strlen(&msgBuffer[0]), (unsigned char*)&msgBuffer[0] );
        } 
        pILC->Release ();
      }
      else {
        sprintf(&msgBuffer[0],"eo38VDcomReleaseAll:QILC HRESULT: 0x%X", hr );
        pIliveCacheSink->ConsoleMsg (strlen(&msgBuffer[0]), (unsigned char*)&msgBuffer[0] );
      } 
      cnt = lpIUnknown->Release ();
    }
    CATCH_ALL
    {
      sprintf(&msgBuffer[0],"eo38VDcomReleaseAll: unhandled exception "); /* PTS 1107168 */
      pIliveCacheSink->ConsoleMsg (strlen(&msgBuffer[0]), (unsigned char*)&msgBuffer[0] );
    }
    currEntry = HashPtr->GetRechainFirstEntry(NULL);  
  }
#endif

  WAIT_UNTIL_ALONE (DCOM_EXCLUSIVE);
  HashPtr->ResetEntrySearch( 0, SearchALL_eeo38 );
  currEntry = HashPtr->GetRechainFirstEntry(NULL);  
  while ( NULL != currEntry)
  {
    if ( 0 == currEntry->getSession() )
    {
      lpIClassFactory = REINTERPRET_CAST(IClassFactory *,(currEntry->getIUnknown()));
      cnt = lpIClassFactory->Release ();
      cnt = lpIClassFactory->LockServer (FALSE);
    }
    currEntry = HashPtr->GetRechainFirstEntry(NULL);  
  }
  //  CoFreeUnusedLibraries ();
  CLEARLOCK (DCOM_EXCLUSIVE);
}

/*---------------------------------------------------------------------------*/

externC void eo38VDcomTransEnd ( tsp_dcom_dispatch_parms * parms )
{
  const short      vt_ptr             = 26;
  bool             DebugSessionLocal  = false;
  bool             DebugSessionRemote = false;
  short            rc                 = 0;
  int                     i;
  tsp00_Uint4              session;
  short                   arrCnt = 0;
  IUnknown*               arrInterfaces[100];  
  CCoClassHashPtr         HashPtr;
  CHashEntryPtr           currEntry;
  teo39_Transend          *param_stack   = REINTERPRET_CAST(teo39_Transend*,parms->param_stack);
  tsp00_Int4              l_buflen;
  struct _aligned_buf
  {
    union 
    {
      teo39_TransendBuf       l_buf;
      integer                 align;
    } v;
  } aligned_buf;
#if defined(_IBMR2) || defined(SUN) || defined(PA11) || defined(PA20W) || defined(ALPHA) || defined(NMP) || defined(_WIN64) || (defined LINUX && defined IA64)
  teo39_Transend_desc   l_param;
  l_param.lpbuflen.plVal       = &l_buflen;  
  l_param.lpbuflen.vt          = vt_ptr;  
  l_param.lpbuflen.io_flag     = 1;  
  l_param.lpbuflen.ptr_flag    = 1;  
  l_param.lpbuflen.len         = 4;  
  l_param.lpbuf.byref          = (void*)&aligned_buf.v.l_buf;
  l_param.lpbuf.vt             = vt_ptr;  
  l_param.lpbuf.io_flag        = 1;  
  l_param.lpbuf.ptr_flag       = 1;  
  l_param.lpbuf.len            = sizeof( teo39_TransendBufMsg );  
#else
  teo39_Transend        l_param;
  l_param.lpbuflen      = &l_buflen;  
  l_param.lpbuf         = &aligned_buf.v.l_buf;
#endif
  
  parms->hresult = 0;
  HashPtr = REINTERPRET_CAST(CCoClassHash*,(DCOM_HEAD));
  if ( NULL == HashPtr ) return;
  
  memcpy(&session, parms->session, sizeof(session));
  WAIT_UNTIL_ALONE (DCOM_EXCLUSIVE);
#ifdef USE_KGS
  HashPtr->ResetEntrySearch( session, SearchEQ_eeo38 );
#else
  HashPtr->ResetEntrySearch( 0, SearchNE_eeo38 );
#endif
  currEntry = HashPtr->NextEntry();
  
  while ( NULL != currEntry)
  {
    if ((currEntry->getType() == IID_eeo38) ||
      (currEntry->getType() == Invalid_IID_eeo38))  
    {
      /* Ignore special Interface, which is activated by DDL (CREATE DBPROC) */
      if ( IID_IliveCacheSpec != *(currEntry->getGuid()))
      {
        arrInterfaces[arrCnt] = currEntry->getIUnknown();
        arrCnt++;
      }
    }
    else
    {
      if (currEntry->getDebugFlag())
      {
        if (currEntry->getRemoteDbgFlag()) 
          DebugSessionRemote = true;
        else
          DebugSessionLocal  = true;
      }
    }
    currEntry = HashPtr->NextEntry();  
  }
  
  CLEARLOCK (DCOM_EXCLUSIVE);
  
  HRESULT hr = 0;
  // for (i = 0; i < arrCnt; i++)  PTS 1107032 FF 27-Jun-2000
  i = 0;                         //PTS 1107032 FF 27-Jun-2000
  if ( i < arrCnt )              //PTS 1107032 FF 27-Jun-2000
  {
    aligned_buf.v.l_buf.msg.mtype     = param_stack->lpbuf->msg.mtype;
    aligned_buf.v.l_buf.msg.mtype2    = param_stack->lpbuf->msg.mtype2;
    l_buflen            = sizeof( teo39_TransendBufMsg );
#if defined(_IBMR2) || defined(SUN) || defined(PA11) || defined(PA20W) || defined(ALPHA) || defined(NMP) || defined(_WIN64) || (defined LINUX && defined IA64)
    parms->param_stack_len = 2 * sizeof(teo39_Variant);
#endif
    parms->param_stack  = REINTERPRET_CAST(tsp00_Void*,(&l_param));
    parms->hresult = arrInterfaces[i]->QueryInterface( IID_IliveCacheSpec, 
    REINTERPRET_CAST(void **,(&parms->idispatch)));
    if ( SUCCEEDED(parms->hresult))
    {
      eo38VDcomDispatch (parms);
      rc = aligned_buf.v.l_buf.err; 
#ifdef USE_KGS
      if ((0 != rc) || FAILED(parms->hresult))
      {
        char dmp_buf[132];
        MSGCD ((ERR_DCOM_HRESULT, "eo38VDcomTransEnd: Error calling sysRequest", rc ));
        sprintf(dmp_buf,"mtype: %X  mtype2: %X ", (long)(param_stack->lpbuf->msg.mtype),
            (long)(param_stack->lpbuf->msg.mtype2)); 
        MSGCD ((ERR_DCOM_DUMP_INFO,dmp_buf));
        hr = GEO00D_DCOM_TRANSEND_FAILED;       /* pts 1105922 FF 10-Mar-2000 */
        /*if ( !hr ) hr = parms->hresult;          pts 1105922                */
      }
#endif
      arrInterfaces[i]->Release();
     }
#ifdef USE_KGS
     else
     {
       MSGCD ((ERR_DCOM_HRESULT, "eo38VDcomTransEnd:QI(ILCSpec)", parms->hresult ));
       hr = GEO00D_DCOM_TRANSEND_FAILED;       /* pts 1105922 FF 10-Mar-2000 */
     }
#endif
  }
  
  parms->param_stack  = REINTERPRET_CAST(tsp00_Void*,param_stack);
#if defined(_IBMR2) || defined(SUN) || defined(PA11) || defined(PA20W) || defined(ALPHA) || defined(NMP) || defined(_WIN64) || (defined LINUX && defined IA64)
  parms->param_stack_len = sizeof( teo39_TransendBufMsg );
#endif
#if defined (WIN32)
#ifdef USE_KGS
  if ( DebugSessionLocal ) 
  {
    parms->debug_location = 0;
    os39VDcomTransEnd ( parms );
  }

  if (( DebugSessionRemote ) && (SUCCEEDED(parms->hresult)))
  {
    parms->debug_location = (char*)-1;
    os39VDcomTransEnd ( parms );
  }

  if (SUCCEEDED(parms->hresult))
  {
    if (( DebugSessionLocal ) || ( DebugSessionRemote )) 
    {
      WAIT_UNTIL_ALONE (DCOM_EXCLUSIVE);
      HashPtr->ResetEntrySearch( session, SearchEQ_eeo38 );
      currEntry = HashPtr->NextEntry();  
      while ( NULL != currEntry)
      {
        if (currEntry->getType() == DbgInterface_eeo38) 
          currEntry->setDebugFlag(false);
        currEntry = HashPtr->NextEntry();  
      }
      CLEARLOCK (DCOM_EXCLUSIVE);
    }
  }
#endif
#endif

   /*  if (( 0 != rc ) || FAILED(hr))  pts 1105922 */
  if ( 0 != hr )     /* pts 1105922 FF 10-Mar-2000 */
  {
#ifdef USE_KGS
    eo38DisposeInstances( parms );
#endif
    parms->hresult = GEO00D_DCOM_TRANSEND_FAILED; 
  }
}

/*---------------------------------------------------------------------------*/


void eo38LoadLibraryEx(const char *path, 
                       bool      loadGlobal,
                       HANDLE    *hd,
                       char      *errtext,
                       int       errtext_size) 
{
  WAIT_UNTIL_ALONE (DCOM_EXCLUSIVE);

  bool found = false;
  CLoadedLibDirEntry_eo38 *curr;
  for ( cgg251dclIterator<CLoadedLibDirEntry_eo38*,CDefaultAllocator_eo38> iter = LoadedLibDir.begin(); iter; ++iter) {
    curr = *iter();
    if ( !strcmp(curr->getNamePtr(), path) ) {
      found = true;
      break;
    }
  }
  if ( found ) {
    curr->incrementUsedCount();
    *hd = curr->getHandle();
  }
  else {
    *hd = sqlLoadLibraryEx(CONST_CAST(char*,path), (loadGlobal) ? RTLD_GLOBAL_EO02 : RTLD_LOCAL_EO02, errtext, errtext_size);
    if ( *hd ) {
      try {
        curr = new CLoadedLibDirEntry_eo38(*hd, path, loadGlobal);
        LoadedLibDir.push_front(curr);
      }
      catch (DbpError&) {
        sqlFreeLibrary(*hd, errtext, errtext_size);
        *hd = NULL;
      }
    }
#ifdef USE_KGS
    else  {
      char buf[256];                                  
      sprintf(&buf[0],"eo38LoadLibraryEx - File: %.100s  err: %.64s", path, errtext );
      MSGCD ((ERR_DCOM_DUMP_INFO, &buf[0] ));
    }
#endif
  }

  CLEARLOCK (DCOM_EXCLUSIVE);
};

/*---------------------------------------------------------------------------*/

void eo38FreeLibrary(HANDLE      hd,
                     bool        *ok,
                     char        *errtext,
                     int         errtext_size)
{
  WAIT_UNTIL_ALONE (DCOM_EXCLUSIVE);

  bool found = false;
  *ok = false;
  CLoadedLibDirEntry_eo38 *curr;
  for ( cgg251dclIterator<CLoadedLibDirEntry_eo38*,CDefaultAllocator_eo38> iter = LoadedLibDir.begin(); iter; ++iter) {
    curr = *iter();
    if ( curr->getHandle() == hd ) {
//      short useCnt = curr->decrementUsedCount();
//      if ( false /* PTS 1107191, 0 == useCnt */) {
//        LoadedLibDir.remove(iter);
//        delete curr;
//        if ( sqlFreeLibrary(hd, errtext, errtext_size)) 
//          *ok = true;
//        else
//          *ok = false;
//      }
//      else
        *ok = true;
      break;
    }
  }

  CLEARLOCK (DCOM_EXCLUSIVE);
};

/*---------------------------------------------------------------------------*/

void *CIliveCacheSink::operator new( size_t stAllocateBlock, void *addr )
{
  return addr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::QueryInterface ( REFIID iid, PVOID *ppv)
{
  if (memcmp (&iid, &IID_IUnknown, sizeof (IID)) == 0)
    *ppv = (void *) this;
  else if (memcmp (&iid, &IID_IliveCacheSink, sizeof (IID)) == 0)
    *ppv = (void *) this;
  else
  {
    *ppv = 0;
    return E_NOINTERFACE;
  }
  this->AddRef();
  return S_OK;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP_(ULONG) CIliveCacheSink::AddRef(void)
{
  return S_OK;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP_(ULONG) CIliveCacheSink::Release(void)
{
  return S_OK;
}

/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::GetDefaultContext (void** p)
{
  *p = lpDefaultContext;
  return S_OK;
}
#endif
/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::SetDefaultContext (void* p)
{
  lpDefaultContext = p;
  return S_OK;
}
#endif
/*---------------------------------------------------------------------------*/

#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::GetObj (
                                      tgg91_TransNo&                 consistentView,
                                      tgg00_ContainerId&             containerId, 
                                      const tgg00_OmsVersionContext* pVersionContext,
                                      const OmsTypeOid&              oid,
                                      bool                           doLock,
                                      tgg91_PageRef&                 objSeq,
                                      tsp00_Int4                     bodySize,
                                      unsigned char*                 pBody,
                                      tsp00_Int4&                    bodyLen,
                                      tsp00_Int4&                    histLogCount,
                                      tgg00_BasisError&              e)
{
  this->CallbackIsActive = true;
  ak341GetObj (this->lpSessionContext,
    consistentView,
    containerId,
    pVersionContext,
    *REINTERPRET_CAST(const tgg92_KernelOid*, &oid),
    doLock,
    bodySize,
    REINTERPRET_CAST(tgg00_Rec*, pBody),
    bodyLen, 
    objSeq,
    histLogCount,
    e);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif
/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::UpdateObj(
                                        unsigned char  * lpConsistentView,
                                        unsigned char  * lpContainerId,
                                        unsigned char  * lpVersionContext,
                                        unsigned char  * lpoid,
                                        unsigned char  * lpobjseq,
                                        tsp00_Int4 bodysize,
                                        unsigned char  * lpbody,
                                        unsigned char  * lpcoldesc,
                                        short  * b_error)
{
  this->CallbackIsActive = true;
  ak341UpdateObj (this->lpSessionContext,
     (tgg91_TransNo *)          lpConsistentView,
    (tgg00_ContainerId *)       lpContainerId,
    (tgg00_OmsVersionContext*)  lpVersionContext,
    REINTERPRET_CAST(tgg92_KernelOid *,         lpoid),
    (tgg91_PageRef *)           lpobjseq,
    bodysize,
    (tgg00_Rec *) lpbody,
    (tgg00_ObjColDesc *) lpcoldesc,
    b_error);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif
/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::NewObj(
                                     unsigned char   * lpConsistentView,
                                     unsigned char   * lpContainerId,
                                     unsigned char   * lpVersionContext,
                                     tsp00_Int4                       KeyPos,
                                     tsp00_Int4                       KeyColSize,
                                     unsigned char   * KeyColumn, 
                                     OmsTypeOid      * lpoid,
                                     unsigned char   * lpobjseq,
                                     short           * b_error)
{
  this->CallbackIsActive = true;
  ak341NewObj (this->lpSessionContext,
    (tgg91_TransNo *)           lpConsistentView,
    (tgg00_ContainerId *)       lpContainerId,
    (tgg00_OmsVersionContext*)  lpVersionContext,
    KeyPos, 
    KeyColSize,
    (tsp00_MoveObj*)    KeyColumn,
    REINTERPRET_CAST(tgg92_KernelOid *, lpoid),
    (tgg91_PageRef *)   lpobjseq,
    b_error);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif        
/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::RegisterClass(
                                            tsp00_Int4          SchemaId,
                                            GUID                *lpGUID,
                                            tsp00_KnlIdentifier *lpClassName,
                                            tsp00_Int4           size,
                                            OmsContainerNo       ContainerNo,
                                            unsigned char       *lpcoldesc,
                                            unsigned char       *lpContainerId,
                                            short               *b_error)
{
  this->CallbackIsActive = true;
  a34RegisterClass (*this->lpSessionContext,
    SchemaId,
    *REINTERPRET_CAST(tsp00_C16*, lpGUID),
    *REINTERPRET_CAST(tsp00_KnlIdentifier*, lpClassName),
    size,
    ContainerNo,
    *REINTERPRET_CAST(tgg00_ObjColDesc*, lpcoldesc),
    *REINTERPRET_CAST(tgg00_ContainerId*, lpContainerId),
    *b_error);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif
/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::DeleteAllObj( 
                      unsigned char  * lpContainerId,
                      unsigned char  * lpVersionContext,
                      int            * lpDeleted,                    
                      short          * b_error)
{
  this->CallbackIsActive = true;
  ak341DeleteAllObj (this->lpSessionContext,
                (tgg00_ContainerId*) lpContainerId,
                (tgg00_OmsVersionContext*) lpVersionContext,
                lpDeleted,
                b_error);
  this->CallbackIsActive = false;
  return S_OK;
}
        
#endif
/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::DeleteObj(
                                        unsigned char  * lpConsistentView,
                                        unsigned char  * lpContainerId,
                                        unsigned char  * lpVersionContext,
                                        unsigned char  * lpoid,
                                        unsigned char  * lpobjseq,
                                        unsigned char  * lpcoldesc,
                                        short  * b_error)
{
  this->CallbackIsActive = true;
  ak341DeleteObj (this->lpSessionContext,
    (tgg91_TransNo *)            lpConsistentView,
    (tgg00_ContainerId *)        lpContainerId,
    (tgg00_OmsVersionContext*)   lpVersionContext,
    (tgg92_KernelOid *)          lpoid,
    (tgg91_PageRef *)            lpobjseq,
    (tgg00_ObjColDesc *)         lpcoldesc,
    b_error);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif
/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::CloseVersion(
                      OmsVersionId    *lpVersionId,
                      tsp00_Int4             lastKey,
                      short  *b_error)
{
  this->CallbackIsActive = true;
  ak341CloseVersion (this->lpSessionContext, 
    (tak_oms_version_id*) lpVersionId,
    lastKey,
    b_error);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif
/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::CreateOpenVersion(
                      unsigned char  *lpConsistentView,
                      OmsVersionId   *lpVersionId,
                      boolean         createVersion,
                      unsigned char  *lpVersionContext,
                      short          *b_error)
{
  this->CallbackIsActive = true;
  ak341CreateOpenVersion (this->lpSessionContext,
    (tgg91_TransNo *)           lpConsistentView,
    (tak_oms_version_id*)       lpVersionId,
                                createVersion,
    (tgg00_OmsVersionContext*)  lpVersionContext,
    b_error);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif
/*---------------------------------------------------------------------------*/

#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::NewConsistentView(
                      unsigned char  *lpConsistentView,
                      short          *b_error) 
{
  this->CallbackIsActive = true;
  ak341NewConsistentView (this->lpSessionContext, 
    (tgg91_TransNo *)           lpConsistentView,
    b_error);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif
/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::EndConsistentView(
                      unsigned char  *lpConsistentView,
                      short          *b_error) 
{
  this->CallbackIsActive = true;
  ak341EndConsistentView (this->lpSessionContext, 
    (tgg91_TransNo *)           lpConsistentView,
    b_error);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif
/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::EnterCriticalSection(short regionId) 
{
  this->CallbackIsActive = true;
  ak341EnterCriticalSection (this->lpSessionContext, regionId);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif
/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::LeaveCriticalSection(short regionId) 
{
  this->CallbackIsActive = true;
  ak341LeaveCriticalSection (this->lpSessionContext, regionId);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif
/*---------------------------------------------------------------------------*/

#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::Rollback( 
                                       tsp00_Int4 buflen,
                                       unsigned char  * lpv,
                                       short  * lpb_error)
{
  this->CallbackIsActive = true;
  a34Rollback (*this->lpSessionContext,
    *REINTERPRET_CAST(tsp00_MoveObj *, lpv),
    buflen,
    *lpb_error);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif
/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::GetObjWithKey (
                                             unsigned char  * lpConsistentView,
                                             unsigned char  * lptree,
                                             unsigned char  * lpVersionContext,
                                             tsp00_Int4                      keylen,
                                             unsigned char  * lpkey,
                                             boolean                   getConsistent,
                                             tsp00_Int4                      bodysize,
                                             unsigned char  * lpbody,
                                             OmsTypeOid     * lpoid,
                                             unsigned char  * lpobjseq,
                                             int            * hist_log_count, /* +++ Change Request 1000139 +++ */
                                             short          * lpb_error)
{
  this->CallbackIsActive = true;
  ak341GetObjWithKey (this->lpSessionContext,
    (tgg91_TransNo *)          lpConsistentView,
    (tgg00_ContainerId *)      lptree,
    (tgg00_OmsVersionContext*) lpVersionContext,
    getConsistent,
    bodysize,
    lpbody,
    keylen,
    lpkey,
    (tgg91_PageRef *)   lpobjseq,
    REINTERPRET_CAST(tgg92_KernelOid *, lpoid),
    hist_log_count,
    lpb_error);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif
/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::SetError(short error,
                                       tsp00_Int4 buflen,
                                       unsigned char  * lpv,
                                       tsp00_CodeType codeType)
{
  this->CallbackIsActive = true;
  sqlErrorFlag = true;              /* PTS 1110813 FF  */
  a34SetError (*this->lpSessionContext,
    error,
    *REINTERPRET_CAST(tsp00_MoveObj*, lpv),
    buflen,
    codeType);
  this->CallbackIsActive = false;
  
#ifdef WIN32
  tsp00_Uint4 status;
  if (a07_return_code (e_cancelled, sqlm_adabas) == error) {
    status = STATUS_TERMINATE_EXCEPTION;
  }
  else {
    status = STATUS_SETERROR_EXCEPTION;
  }
  RaiseException (status,
    0,            /* continuable exception */
    0,            /* number of parms       */
    NULL);        /* &parms                */
#else
  if ( g01is_livecache() )  /* 1110813 & 1120088 FF */
    throw RTE_DBProcError(error);
#endif        
  return S_OK;
}

#endif        
/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::SubtransStart( 
                                            short  * lpb_error)
{
  this->CallbackIsActive = true;
  
  a34SubtransStart (*this->lpSessionContext,
    *lpb_error);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif
/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::SubtransCommit( 
                                             short  * lpb_error)
{
  this->CallbackIsActive = true;
  
  a34SubtransCommit (*this->lpSessionContext,
    *lpb_error);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif        
/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::SubtransRollback( 
                                               short  * lpb_error)
{
  this->CallbackIsActive = true;
  
  a34SubtransRollback (*this->lpSessionContext,
    *lpb_error);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif
/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::Vtrace( 
                                     tsp00_Int4 buflen,
                                     unsigned char  * lpv)
{
  this->CallbackIsActive = true;
  a34VTrace (*this->lpSessionContext,
    *REINTERPRET_CAST(tsp00_MoveObj*, lpv),
    buflen);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif
/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::LockObj(
                                      unsigned char  * lpConsistentView,
                                      unsigned char  * lpContainerId,
                                      OmsTypeOid     * lpoid,
                                      unsigned char  * lpobjseq,
                                      short          * lpb_error)
{
  this->CallbackIsActive = true;
  ak341LockObj (this->lpSessionContext, 
    (tgg91_TransNo *)     lpConsistentView,
    (tgg00_ContainerId *) lpContainerId,
    REINTERPRET_CAST(tgg92_KernelOid *,   lpoid),
    (tgg91_PageRef *)     lpobjseq,
    lpb_error);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif
/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::BasisError( 
                                         short basis_error)
{
  this->CallbackIsActive = true;
  a34BasisError (*this->lpSessionContext,
    basis_error);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif
/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::HResultError( 
                                           HRESULT hresult_error)
{
  this->CallbackIsActive = true;
  a34HResultError (*this->lpSessionContext,
    hresult_error);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif
/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::AllocPacket( 
                                          unsigned char  ** lpv,
                                          tsp00_Int4 *sz)
{
  this->CallbackIsActive = true;
  a34AllocPacket (*this->lpSessionContext, 
    *REINTERPRET_CAST(tsp1_packet_ptr *, lpv), 
    *sz);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif

/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::SQLRequest( 
                                         tsp00_Int4 buflen,
                                         unsigned char  * lpv,
                                         short *lperror)
{
  this->CallbackIsActive = true;
  a34Request (*this->lpSessionContext,
    (tsp1_packet_ptr) lpv);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif
/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::SQLReceive( 
                                         unsigned char  ** lpv)
{
  this->CallbackIsActive = true;
  a34Receive (*this->lpSessionContext,
    *REINTERPRET_CAST(tsp1_packet_ptr*, lpv));
  this->CallbackIsActive = false;
  return S_OK;
}

#endif
/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::NextObj(
                                      unsigned char  * lpConsistentView,
                                      unsigned char  * lpContainerId,
                                      unsigned char  * lpVersionContext,
                                      OmsTypeOid     * lpoid,
                                      unsigned char  * lpobjseq,
                                      tsp00_Int4 bodysize,
                                      unsigned char  * lpbody,
                                      short  * b_error)
{
  this->CallbackIsActive = true;
  ak341NextObj (this->lpSessionContext,
    (tgg91_TransNo *)          lpConsistentView,
    (tgg00_ContainerId *)      lpContainerId,
    (tgg00_OmsVersionContext*) lpVersionContext,
    REINTERPRET_CAST(tgg92_KernelOid *,        lpoid),
    bodysize,
    lpbody,
    REINTERPRET_CAST(tgg91_PageRef *,(lpobjseq)),
    b_error);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif
/*---------------------------------------------------------------------------*/

#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::ABAPRead( 
                                       OmsTypeInt4    * TabHandle,
                                       tsp00_Int4     * count,
                                       tsp00_Int4     * lpbuflen,
                                       unsigned char  * lpv,
                                       short          * lpb_error)
{
  this->CallbackIsActive = true;
  a34AbapRead (*this->lpSessionContext,
    *REINTERPRET_CAST(tsp00_MoveObj*, TabHandle),
    *count,
    STATIC_CAST(int,(*lpbuflen)),
    *REINTERPRET_CAST(tsp00_MoveObj *, lpv),
    *lpb_error);
  
  this->CallbackIsActive = false;
  return S_OK;
}

#endif
/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::ABAPWrite( 
                                        OmsTypeInt4   * TabHandle,
                                        tsp00_Int4    row_size,
                                        tsp00_Int4    row_count,
                                        unsigned char * lpv,
                                        short         * lpb_error)
{
  this->CallbackIsActive = true;
  a34AbapWrite (*this->lpSessionContext,
    *REINTERPRET_CAST(tsp00_MoveObj*, TabHandle),
    row_size,
    row_count,
    *REINTERPRET_CAST(tsp00_MoveObj*, lpv),
    *lpb_error);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif
/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::GetErrorMsg( 
                                          short errorno,
                                          tsp00_Int4 *lpbuflen,
                                          unsigned char  * lpv)
{
  this->CallbackIsActive = true;
  tsp00_Int2 e = errorno;
  a34GetErrormsg (*this->lpSessionContext,
    e,
    *REINTERPRET_CAST(tsp00_MoveObj*, lpv),
    *REINTERPRET_CAST(int *,(lpbuflen)));
  this->CallbackIsActive = false;
  return S_OK;
}

#endif

/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::MultiPurpose (unsigned char MessType,
                                            unsigned char MessType2,
                                            tsp00_Int4           *lpbuflen,
                                            unsigned char  *lpv,
                                            short          *lpb_error)
{
  tgg00_MessType  l_MessType;                            /* PTS 1106619 FF */
  tgg00_MessType2 l_MessType2;                           /* PTS 1106619 FF */
  l_MessType.becomes( (tgg00_MessType_Enum)MessType );   /* PTS 1106619 FF */
  l_MessType2.becomes( (tgg00_MessType2_Enum)MessType2 );/* PTS 1106619 FF */
  
  this->CallbackIsActive = true;
  ak341MultiPurpose (this->lpSessionContext,
    /*  *REINTERPRET_CAST(tgg00_MessType*,  (&MessType)),  PTS 1106619 FF */
    /*  *REINTERPRET_CAST(tgg00_MessType2*, (&MessType2)), PTS 1106619 FF */
    l_MessType,           /* PTS 1106619 FF */ 
    l_MessType2,          /* PTS 1106619 FF */
    REINTERPRET_CAST(int*, (lpbuflen)),
    (tsp00_MoveObj *) lpv,
    lpb_error);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif
/*---------------------------------------------------------------------------*/

#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::UnlockObj( 
                                        unsigned char  * lpContainerId,
                                        OmsTypeOid     * lpoid,
                                        short          * lpb_error)
{
  this->CallbackIsActive = true;
  ak341UnlockObj (this->lpSessionContext, 
    (tgg00_ContainerId *) lpContainerId,
    REINTERPRET_CAST(tgg92_KernelOid *, lpoid),
    lpb_error);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif

/*---------------------------------------------------------------------------*/

#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::IsLocked( 
                                       unsigned char  * lpContainerId,
                                       OmsTypeOid     * lpoid,
                                       short          * lpb_error)
{
  this->CallbackIsActive = true;
  ak341IsLocked (this->lpSessionContext, 
    (tgg00_ContainerId *) lpContainerId,
    REINTERPRET_CAST(tgg92_KernelOid *, lpoid),
    lpb_error);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::IsDebugEnvironment( void)
{
#ifdef USE_KGS
  return 0;
#else
  return 1;
#endif
}

/*---------------------------------------------------------------------------*/

#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::ConsoleMsg( 
                                         tsp00_Int4 buflen,
                                         unsigned char  * lpv)
{
  unsigned char dmp_buf[132];
  if ( buflen >= sizeof(dmp_buf)) buflen = sizeof(dmp_buf) - 1;
  memcpy(dmp_buf,lpv,buflen);
  dmp_buf[buflen] = '\0';
  MSGCD ((ERR_DCOM_PROXY_MSG,dmp_buf));
  return S_OK;
}

#endif

/*---------------------------------------------------------------------------*/

#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::DropFile(char  *lpFileName,
                                       short &error)
{
  tsp00_VfReturn_Param Error;
  eo15RemoveMessageFile(lpFileName, Error);
  if (vf_ok == Error) {
    error = e_ok;
  }
  else {
    char buf[132];                                  
    sprintf(&buf[0],"DropFile: %.100s  err: %d ", lpFileName, Error );
    MSGCD ((ERR_DCOM_DUMP_INFO,buf));
    error = a07_return_code (e_hostfile_error, sqlm_adabas);
  }
  return S_OK;
}

#endif

/*---------------------------------------------------------------------------*/


#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::ExistsFile(char  *lpFileName,
                                       bool     &doesExist,
                                       bool     &isDirectory,
                                       short    &error)
{
  tsp00_VfReturn_Param Error;
  eo15CheckExist(lpFileName, doesExist, isDirectory, Error);
  if (vf_ok == Error) {
    error = e_ok;
  }
  else {
    char buf[132];                                  
    sprintf(&buf[0],"ExistsFile: %.100s  err: %d ", lpFileName, Error );
    MSGCD ((ERR_DCOM_DUMP_INFO,buf));
    error = a07_return_code (e_hostfile_error, sqlm_adabas);
  }
  return S_OK;
}

#endif

/*---------------------------------------------------------------------------*/

#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::GetWorkingDirectory(tsp00_Pathc& path,
                                       short  &error)
{
  tsp00_VfReturn_Param Error;
  eo15GetWorkingDirectory(path, Error);
  if (vf_ok == Error) {
    error = e_ok;
  }
  else {
    char buf[132];                                  
    sprintf(&buf[0],"GetWorkingDirectory  err: %d ", Error );
    MSGCD ((ERR_DCOM_DUMP_INFO,buf));
    error = a07_return_code (e_hostfile_error, sqlm_adabas);
  }
  return S_OK;
}

#endif

/*---------------------------------------------------------------------------*/

#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::GetDbrootSapPath(tsp00_Pathc& path,
                                       short  &error)
{
  tsp01_RteError rteError;
  if (sqlGetDbrootSapPath (path, TERM_WITH_DELIMITER_EO01, &rteError)) {
    error = e_ok;
  }
  else {
    char buf[256];                                  
    memcpy (&path[0], &rteError.RteErrText[0], sizeof(rteError.RteErrText));
    path[sizeof(rteError.RteErrText)] = 0;
    sprintf(&buf[0],"GetDbrootSapPath  rteError: %d %.80s", rteError.RteErrCode, rteError.RteErrText );
    MSGCD ((ERR_DCOM_DUMP_INFO,buf));
    sprintf(&buf[0],"GetDbrootSapPath  osError: %d %.80s", rteError.OsErrCode, rteError.OsErrText );
    MSGCD ((ERR_DCOM_DUMP_INFO,buf));
    error = a07_return_code (e_hostfile_error, sqlm_adabas);
  }
  return S_OK;
}

#endif

/*---------------------------------------------------------------------------*/

#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::OpenFile(char  *lpFileName,
                                       tsp00_Uint4 FileSize,
                                       boolean     doAppend,
                                       tsp00_Int4  *FileHandle,
                                       short       *lpb_error)
{
  tsp00_VfReturn_Param Error;
  if (doAppend) {
    *FileHandle = eo15OpenToAppendToMessageFile(lpFileName, Error);
  }
  else {
    *FileHandle = eo15OpenMessageFile(lpFileName, FileSize, Error);
  }
  if (vf_ok == Error) {
    *lpb_error = e_ok;
  }
  else {
    char buf[256];                                  
    if (doAppend) {
      sprintf(&buf[0],"OpenFile: %.200s filesize %d append err: %d ", lpFileName, FileSize, Error );
    }
    else {
      sprintf(&buf[0],"OpenFile: %.200s filesize %d no append err: %d ", lpFileName, FileSize, Error );
    }
    MSGCD ((ERR_DCOM_DUMP_INFO,buf));
    *lpb_error = a07_return_code (e_hostfile_error, sqlm_adabas);
  }
  return S_OK;
}

#endif

/*---------------------------------------------------------------------------*/

#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::WriteFile(tsp00_Int4 FileHandle,
                                        char   *msg,
                                        short  *lpb_error)
{
  tsp00_VfReturn_Param Error;
  eo15WriteMessageFile(FileHandle, msg, Error);
  if (vf_ok == Error) {
    *lpb_error = e_ok;
  }
  else {
    char buf[132];                                  
    sprintf(&buf[0],"WriteFile hd: %d  err: %d ", FileHandle, Error );
    MSGCD ((ERR_DCOM_DUMP_INFO,buf));
    *lpb_error = a07_return_code (e_hostfile_error, sqlm_adabas);
  }                                     
  return S_OK;
}

#endif

/*---------------------------------------------------------------------------*/

#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::CloseFile(tsp00_Int4 FileHandle,
                                        short  *lpb_error)
{
  tsp00_VfReturn_Param Error;
  eo15CloseMessageFile(FileHandle, Error);
  if (vf_ok == Error) {
    *lpb_error = e_ok;
  }
  else {
    char buf[132];                                  
    sprintf(&buf[0],"CloseFile: %d  err: %d ", FileHandle, Error );
    MSGCD ((ERR_DCOM_DUMP_INFO,buf));
    *lpb_error = a07_return_code (e_hostfile_error, sqlm_adabas);
  }                                        
  return S_OK;
}

#endif

/*---------------------------------------------------------------------------*/

#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::SyncFile(char       *lpFileName,
                                       tsp00_Int4 FileHandle,
                                       short      *lpb_error) 
{
  tsp00_VfReturn_Param Error;
  if (NULL != lpFileName) {
    eo15SyncNamedMessageFile(lpFileName, Error);
  }
  else {
    eo15SyncMessageFile(FileHandle, Error);
  }
  if (vf_ok == Error) {
    *lpb_error = e_ok;
  }
  else {
    char buf[132];                                  
    if (NULL != lpFileName) 
      sprintf(&buf[0],"SyncFile %.100s err: %d ", lpFileName, Error );
    else
      sprintf(&buf[0],"SyncFile hd: %d err: %d ", FileHandle, Error );
    MSGCD ((ERR_DCOM_DUMP_INFO,buf));
    *lpb_error = a07_return_code (e_hostfile_error, sqlm_adabas);
  }                                 
  return S_OK;
}

#endif

/*---------------------------------------------------------------------------*/

#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::GetDateTime(tsp00_Date* Date,
                                          tsp00_Time* Time) 
{
  vdattime(*Date, *Time);                       
  return S_OK;
}

#endif

/*---------------------------------------------------------------------------*/

#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::GetClock(tsp00_Int4* Sec,
                                       tsp00_Int4* MSec) 
{
  vclock(Sec, MSec);                       
  return S_OK;
}

#endif

/*---------------------------------------------------------------------------*/

#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::DropVersion(unsigned char *lpVersionContext,
                                          short         *lpb_error)
{
  this->CallbackIsActive = true;
  ak341DropVersion (*this->lpSessionContext,
    *REINTERPRET_CAST(tgg00_OmsVersionContext*, lpVersionContext),
    *lpb_error);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif

/*---------------------------------------------------------------------------*/

#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::CreateSchema(tsp00_KnlIdentifier *SchemaName,
                                          tsp00_Int4           *SchemaId,
                                          short                *lpb_error)
{
  this->CallbackIsActive = true;
  a34CreateSchema (*this->lpSessionContext, *SchemaName, *SchemaId, *lpb_error);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif

/*---------------------------------------------------------------------------*/

#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::ExistsSchema(tsp00_KnlIdentifier *SchemaName,
                                          tsp00_Int4           *SchemaId,
                                          short                *lpb_error)
{
  this->CallbackIsActive = true;
  a34ExistsSchema (*this->lpSessionContext, *SchemaName, *SchemaId, *lpb_error);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif

/*---------------------------------------------------------------------------*/

#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::DropContainer(OmsSchemaHandle schema,
                                         const GUID&        guid,
                                         OmsContainerNo     cno,
                                         tgg00_BasisError&  error)
{
  this->CallbackIsActive = true;
  a34DropContainer (*this->lpSessionContext, schema, *REINTERPRET_CAST(tsp_c16*, CONST_CAST(GUID*, &guid)), cno, error);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif

/*---------------------------------------------------------------------------*/

#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::DropSchema(tsp00_Int4 SchemaId,
                                         short      *lpb_error)
{
  this->CallbackIsActive = true;
  a34DropSchema (*this->lpSessionContext, SchemaId, *lpb_error);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif

/*---------------------------------------------------------------------------*/

#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::GetSchemaName(tsp00_Int4 SchemaId,
                                         tsp00_KnlIdentifier *SchemaName,
                                         short               *lpb_error)
{
  this->CallbackIsActive = true;
  a34GetSchemaName (*this->lpSessionContext, SchemaId, *SchemaName, *lpb_error);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif

/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::LoadLibraryEx(const char  *path, 
                                            bool        loadGlobal,
                                            HANDLE      *hd,
                                            tsp_errtext &errtext,
                                            int         errtext_size)
{
  this->CallbackIsActive = true;
  eo38LoadLibraryEx(path, loadGlobal, hd, errtext, errtext_size);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif

/*---------------------------------------------------------------------------*/

#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::FreeLibrary(HANDLE      hd,
                                          bool        &ok,
                                          tsp_errtext &errtext,
                                          int         errtext_size)
{
  this->CallbackIsActive = true;
  eo38FreeLibrary(hd, &ok, errtext, errtext_size);
  this->CallbackIsActive = false;
  return S_OK;
}

#endif

/*---------------------------------------------------------------------------*/
#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::GetContainerInfo(tsp00_Uint4 containerHandle,
            GUID&            guid,
            OmsSchemaHandle& schema,
            OmsContainerNo&  cno,
            short&           error)
{
  this->CallbackIsActive = true;
  a34GetContainerInfo (*this->lpSessionContext,
    containerHandle,
    *REINTERPRET_CAST(tsp00_C16*, &guid),
    (tsp00_Int4&) schema, 
    (tsp00_Int4&) cno,
    error);
  this->CallbackIsActive = true;
  return S_OK;
}

#endif

#ifdef USE_KGS

STDMETHODIMP CIliveCacheSink::IsObjHistoryUsed( 
            tgg00_ContainerId& containerId, 
            const OmsTypeOid&  oid,
            bool&              isUsed,
            tgg00_BasisError&  e) 
{
  this->CallbackIsActive = true;
  a34IsObjHistoryUsed (*this->lpSessionContext,
    containerId,
    *REINTERPRET_CAST(const tgg92_KernelOid*, &oid),
    isUsed, 
    e);
  this->CallbackIsActive = true;
  return S_OK;
}

#endif
/*---------------------------------------------------------------------------*/

/*===========================================================================*
*  LOCAL FUNCTIONS (CODE)                                                   *
*===========================================================================*/

static HRESULT eo38_ComCache ( GUID * lpGuid, 
                              void *lpInterface, 
                              tsp00_Uint4 session,
                              bool  remoteDbgFlag,
                              teo38_HashEntryType hType )
{
  CCoClassHashPtr         HashPtr;
  CHashEntryPtr           currSessPtr;
  
  WAIT_UNTIL_ALONE (DCOM_EXCLUSIVE);
  
  HashPtr = REINTERPRET_CAST(CCoClassHashPtr,(DCOM_HEAD));
  if ( NULL == HashPtr ) 
  {
    eo38_Initialize ();
    HashPtr = REINTERPRET_CAST(CCoClassHashPtr,(DCOM_HEAD));
    if ( NULL == HashPtr ) 
    {
      CLEARLOCK (DCOM_EXCLUSIVE);
      return -1;
    }
  }
  
  currSessPtr = HashPtr->getFreeEntry();
  if ( NULL == currSessPtr )
  {
    CLEARLOCK (DCOM_EXCLUSIVE);
    return -1;
  }
  
#ifdef USE_KGS
  currSessPtr->init( session, lpGuid, (LPUNKNOWN)lpInterface, remoteDbgFlag, hType );
#else
  if ( CLSID_eeo38 == hType )
    currSessPtr->init( 0, lpGuid, (LPUNKNOWN)lpInterface, remoteDbgFlag, hType );
  else
    currSessPtr->init( DEBUG_SESSION_ID, lpGuid, (LPUNKNOWN)lpInterface, remoteDbgFlag, hType );
#endif
  HashPtr->chainEntry(currSessPtr);
  CLEARLOCK (DCOM_EXCLUSIVE);
  return S_OK;
}

/*---------------------------------------------------------------------------*/

static BOOLEAN eo38_Initialize ( void )
{
  BOOLEAN rc = 1;
  tsp00_Uint4  lRC;
  CCoClassHashPtr pMem;
  
  if (DCOM_HEAD == 0)
  {
    pMem = new (CCoClassHash);
    if ( NULL == pMem )
    {
      MSGCD ((ERR_DCOM_NOMEM, "eo38_Initialize" ));
      rc = 0;
    }
    else
    {
#if defined(USE_KGS) && defined(DEBUG_MEM_ALLOC)
      MSGD(( 11999, INFO_TYPE, "new   ", "veo38.cpp:eo38_Initialize - 0x%lx, sizeInBytes %ld",
        pMem, sizeof(CCoClassHash) ));
#endif /* DEBUG_MEM_ALLOC */
      lRC = pMem->HashInit( C_HASH_SESS_HDR_SIZE );
      if ( 0 != lRC )
      {
        MSGCD ((ERR_DCOM_NOMEM, "eo38_Initialize:HashInit" ));
        rc = 0;
      }
      else
      {
        DCOM_HEAD = pMem;
      }
    }
  }
  return (rc);
}

/*---------------------------------------------------------------------------*/

static void eo38_Invalidate(tsp_dcom_co_create_instance_parms *lpParms,
                            GUID *lpGuid, 
                            tsp00_Uint4 session)
{
#ifdef HUGO
  HRESULT hr;
  tsp_dcom_dispatch_parms disp_parms;
  tsp00_Uint4    idx, ui;
  IUnknown *lpIUnknown;
  
  tsp00_Int4 bufLen;
  
  struct _buffer
  {
    union 
    { 
      tgg00_MessType  MessType;
      tsp00_Int2      Rc;
    } v;
  } buffer;
  
  struct  _stack 
  {
    tsp00_Int4     *plong;
    tgg00_MessType *pMessType;
  } stack;
  
  parms->ret = RC_DISPADDR_INVALID_FOUND;
  memcpy(&act_IID_Dsc,lpIID_Dsc,sizeof(act_IID_Dsc));
  
  lpIID_Dsc->idispatch = NULL;
  lpIID_Dsc->session = 0;
  lpIID_Dsc->AdviseCookie = 0;
  lpIID_Dsc->type = Free_eeo38;
  /* chain to free - chain */
  if (lpIID_Dsc_prev == NULL)
    (*((t_Head *) DCOM_HEAD))[idx] = STATIC_CAST(t_IIDDsc *,(lpIID_Dsc->lpnext));
  else
    lpIID_Dsc_prev->lpnext = STATIC_CAST(t_IIDDsc *,(lpIID_Dsc->lpnext));
  lpIID_Dsc->lpnext = DCOM_FREE;
  DCOM_FREE = lpIID_Dsc;
  CLEARLOCK (DCOM_EXCLUSIVE);
  
  /*** VALIDATE REQUEST for OMS Layer !!! ***/
  
  disp_parms.idispatch = REINTERPRET_CAST(char *,(act_IID_Dsc.idispatch));
  disp_parms.ifactory = REINTERPRET_CAST(char *,(act_IID_Dsc.idispatch));
  disp_parms.session_context  = parms->session_context;
  disp_parms.dispid           = C_SYS_DISPID;
  disp_parms.hresult          = S_OK;
  disp_parms.param_stack      = (tsp00_MoveObjPtr)&stack;
  disp_parms.param_stack_len  = sizeof(stack);
  disp_parms.debug_mode       = 0;
  disp_parms.debug_location   = NULL;
  bufLen            = sizeof(buffer);
  stack.plong       = &bufLen;
  stack.pMessType   = &buffer.v.MessType;
  buffer.v.MessType.becomes(m_shutdown);
  
  lpIUnknown = (IUnknown *) disp_parms.idispatch;
  hr = lpIUnknown->QueryInterface( IID_IliveCacheSpec, 
    REINTERPRET_CAST(void **,(&disp_parms.idispatch)));
  if ( SUCCEEDED(hr) )
  {
    eo38VDcomDispatch (&disp_parms);
    hr = disp_parms.hresult;
    if ( FAILED(hr) )
      MSGCD ((ERR_DCOM_HRESULT, "eo38VDcomDispAddr:Invalidate(ILCSpec)", hr ));
  }
  else
    MSGCD ((ERR_DCOM_HRESULT, "eo38VDcomDispAddr:QI(ILCSpec)", hr ));
  if ( FAILED(hr) )  parms->ret = RC_DISPADDR_INVALIDATE_FAILED;
  if ( buffer.v.Rc != S_OK ) 
  {
    parms->ret = buffer.v.Rc;
    MSGCD ((ERR_DCOM_HRESULT, "eo38VDcomDispAddr:Invalidate_RC:", buffer.v.Rc ));
    return;
  }
  
  disp_parms.dispid = 2;
  TRY 
  {
    lpIUnknown = (IUnknown *) parms->idispatch;
    hr = lpIUnknown->QueryInterface ( IID_IConnectionPointContainer,
      (void **) &pCPC);
    if (SUCCEEDED (hr))
      hr = pCPC->FindConnectionPoint (IID_IliveCacheSink, &pCP);
    else
      MSGCD ((ERR_DCOM_HRESULT, "eo38VDComDispAddr:QI(ICPC)", hr ));
    if (SUCCEEDED (hr))
    {
      pCPC->Release ();
      hr = pCP->Unadvise (act_IID_Dsc.AdviseCookie);
      if ( FAILED(hr))  
        MSGCD ((ERR_DCOM_HRESULT, "eo38VDComDispAddr:Unadvise", hr ));
      pCP->Release ();
    }
    else
      MSGCD ((ERR_DCOM_HRESULT, "eo38VDComDispAddr:FCP", hr ));
    lpIUnknown->Release ();
  }
  CATCH_ALL
  {
    MSGCD ((ERR_DCOM_HRESULT, "eo38VDComDispAddr:Release", disp_parms.hresult ));
  }       
      }
#endif
};

/*---------------------------------------------------------------------------*/

static void eo38_ReleaseInstance ( CHashEntryPtr  currEntry, tsp00_Addr SessionContext  )
{
  tsp_dcom_dispatch_parms   disp_parms;
  tsp00_Uint4               cnt;
  IUnknown                  *lpIUnknown;
  HRESULT                   hr;
  IliveCacheSpec            *pILC;
  
  disp_parms.idispatch  = REINTERPRET_CAST(char *,(currEntry->getIUnknown()));
  disp_parms.ifactory   = REINTERPRET_CAST(char *,(currEntry->getIUnknown()));
  disp_parms.dispid     = 2;
  disp_parms.session_context = SessionContext;
  TRY
  {
    lpIUnknown = currEntry->getIUnknown();
    if (currEntry->getType() != DbgInterface_eeo38)
    {
      hr = lpIUnknown->QueryInterface ( IID_IliveCacheSpec, (void **) &pILC);
      if (SUCCEEDED (hr))
      {
        hr = pILC->Unadvise ();
#if defined(USE_KGS) 
        if ( FAILED(hr))  
          MSGCD ((ERR_DCOM_HRESULT, "eo38_ReleaseInstance:Unadvise", hr ));
#endif
        pILC->Release ();
      }
      else
        MSGCD ((ERR_DCOM_HRESULT, "eo38_ReleaseInstance:QILC", hr ));
    }
#if defined(USE_KGS) && defined(WIN32)
    else
    {
      os39VDcomReleaseAll ( SessionContext, currEntry->getRemoteDbgFlag(), &hr );
      if ( FAILED(hr)) {
        MSGCD ((ERR_DCOM_HRESULT, "eo38_ReleaseInstance:os39VDcomReleaseAll", hr ));
      }
    }
#endif
    cnt = lpIUnknown->Release ();
  }
#if defined(WIN32)
  __except (eo38_Eval_Exception (&disp_parms, GetExceptionInformation ()))
  {
     MSGCD ((ERR_DCOM_HRESULT, "eo38_ReleaseInstance:Release", disp_parms.hresult ));
  }
#else
  catch (RTE_DBProcError e) {
     disp_parms.hresult = GEO00D_DCOM_EXCEPTION_RAISED;
     MSGCD ((ERR_DCOM_HRESULT, "eo38_ReleaseInstance:Release", disp_parms.hresult ));
  }
  catch (DbpError e) { 
    char    msgBuffer[132];
    sprintf(&msgBuffer[0],"eo38_ReleaseInstance:Release: DbpError:%d", e.dbpError() );
    MSGCD ((ERR_DCOM_DUMP_INFO,msgBuffer));
  }
#endif
}

/*---------------------------------------------------------------------------*/

#ifdef USE_KGS

static HRESULT eo38_UnadviseInstance ( CHashEntryPtr    currEntry )
{
  tsp_dcom_dispatch_parms   disp_parms;
  IUnknown                  *lpIUnknown;
  HRESULT                   hr;
  IliveCacheSpec            *pILC;
  
  disp_parms.idispatch  = REINTERPRET_CAST(char *,(currEntry->getIUnknown()));
  disp_parms.ifactory   = REINTERPRET_CAST(char *,(currEntry->getIUnknown()));
  disp_parms.dispid     = 2;
  disp_parms.session_context = 0;
  TRY
  {
    lpIUnknown = currEntry->getIUnknown();
    if (currEntry->getType() != DbgInterface_eeo38) 
    {
      hr = lpIUnknown->QueryInterface ( IID_IliveCacheSpec,(void **) &pILC);
      if (SUCCEEDED (hr))
      {
        hr = pILC->Unadvise ();
        if ( FAILED(hr))  
          MSGCD ((ERR_DCOM_HRESULT, "eo38_UnadviseInstance:Unadvise", hr ));
        pILC->Release ();
      }
      else
        MSGCD ((ERR_DCOM_HRESULT, "eo38_UnadviseInstance:QI(ICPC)", hr ));
    }
  }
  CATCH_ALL  
  {
    disp_parms.hresult = GEO00D_DCOM_DBPROC_CRASHED;
    MSGCD ((ERR_DCOM_HRESULT, "eo38_UnadviseInstance:Release", disp_parms.hresult ));
    if ( SUCCEEDED(hr) ) hr = disp_parms.hresult;
  }
  return hr;
}

#endif

/*===========================================================================*
*  END OF CODE                                                              *
*===========================================================================*/
