/****************************************************************************

  module      : veo40.c

  -------------------------------------------------------------------------

  responsible : FrankH

  special area: TCP/IP routines via SAP NI ( network interface )
  description : description ...


  last changed: 2000-04-18  15:05
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1998



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/

/*===========================================================================*/
#ifndef DUMMY_MODULE
/*===========================================================================*/
/*--------------------------------------------------------------------*/
/* SAP includes                                                       */
/*--------------------------------------------------------------------*/
#ifdef WIN32
  #define SAPonNT
#endif

#ifdef _IBMR2 
  #define SAPonRS6000
#endif

#if PA11 || PA20W
  #define SAPonHPPA
#endif

#if defined(SVR4) && defined(SUN)
  #define SAPonSUN
#endif

#if defined(OSF1)
  #define SAPonALPHA
#endif

#if defined(LINUX)
#define SAPonLIN
#endif

#if defined(SVR4) && defined(NMP)
  #define SAPonRM600
#endif

#include "sap/saptype.h"
#include "sap/saptypeb.h"
#include "sap/saptypec.h"
#include "sap/sapuc.h"
#include "sap/err.h"
#include "sap/nixx.h"
/*===========================================================================*/
#endif /* DUMMY */
/*===========================================================================*/

/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
/* externC is defined in SAP-Includes (sapuc.h) too   */
#undef externC
/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "geo007_2.h"
#include "heo420.h"
#include "geo40.h"
#include "heo46.h"
#include "heo47.h"
#include "geo60.h"

#define MOD__ "veo40.c:"

/*===========================================================================*/
#ifndef DUMMY_MODULE
/*===========================================================================*/
/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#ifdef NO_INLINES
# define _INLINE
#else
# define _INLINE	__inline
#endif

#ifdef DEBUG
#define DBGNI(_arg) {printf("%s", MOD__); printf _arg ; printf("\n");}
#else
#define DBGNI(_arg)
#endif

#define NI_DEFAULT_SERVICE 7269
/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/



/*===========================================================================*
 *  GLOBAL VARIABLES                                                         *
 *===========================================================================*/

extern FILE *tf ; /* SAP global trace/error file descriptor */


/*===========================================================================*
 *  LOCAL VARIABLES                                                          *
 *===========================================================================*/



/*===========================================================================*
 *  STATIC/INLINE FUNCTIONS (PROTOTYPES)                                     *
 *===========================================================================*/
static void 
eo40NiBuildErrorString  ( tsp00_ErrTextc VAR_ARRAY_REF  pErrText,
                          char                         *ErrorMsgHeader,
                          SAPRETURN                     SapRc );

static void * 
eo40NiDupHandle         ( void                         *NiHandle );

static void 
eo40GetServiceAsString  ( unsigned short               *Service,
                          char                         *szService,
                          int                           szServiceLen) ;

/*===========================================================================*
 *  PUBLIC FUNCTIONS (CODE)                                                  *
 *===========================================================================*/

externC int
eo40NiIsAvailable() 
{
  return(TRUE);
}

/*===========================================================================*/

#define CTRC_TRUNC      1

externC tsp00_Longuint 
eo40NiInit        ( tsp00_Pathc     VAR_ARRAY_REF  TraceFile,
                    int                            TraceLevel,
                    tsp00_ErrTextc  VAR_ARRAY_REF  ErrText )
{
  SAPRETURN               SapRc ;
  tsp00_Longuint          ulCommState;
  extern FILE *CTrcInit ( const char   *  file,
                          SAP_BYTE     mode,
                          int          tracelevel);

  tf = (FILE *)CTrcInit (TraceFile[0] != '\0' ? (char *)TraceFile : NULL , 
                         CTRC_TRUNC, TraceLevel);

  SapRc = NiInit();

  if ( SapRc == SAP_O_K)
    ulCommState = commErrOk_esp01 ;
  else
  {
    ulCommState = commErrNotOk_esp01 ;
    eo40NiBuildErrorString ( ErrText, ERRMSG_NICOM_INIT_ERROR, SapRc );
  }
  return ( ulCommState ) ;
}

/*===========================================================================*/
#define ERRMSG_NICOM_GET_SD_ERROR "Can't get sd from NiHdl:"

externC tsp00_Longuint 
eo40NiHandleToSocket        ( void                         *NiHandle ,
                              int                          *Sd,
                              tsp00_ErrTextc  VAR_ARRAY_REF  ErrText )
{
  tsp00_Longuint          ulCommState;
  SAPRETURN               SapRc ;

  SapRc = NiHdlToSock 	 (*(NI_HDL *)NiHandle, Sd ) ;

  if ( SapRc == SAP_O_K)
    ulCommState = commErrOk_esp01 ;
  else
  {
    ulCommState = commErrNotOk_esp01 ;  
    eo40NiBuildErrorString ( ErrText, ERRMSG_NICOM_GET_SD_ERROR, SapRc );
    ErrTrace(tf);
  }

  return ( ulCommState );

}


/*===========================================================================*/

externC tsp00_Longuint 
eo40NiDataAvailable   ( void                         *NiHandle ,
                        tsp00_ErrTextc VAR_ARRAY_REF  pErrText )
{
  #undef  MF__
  #define MF__ MOD__"eo40NiDataAvailable"
  tsp00_Longuint          ulCommState;
  SAPRETURN               SapRc ;

  SapRc = NiPeek ( *(NI_HDL *)NiHandle, 0 ) ;

  switch (SapRc)
  {
    case SAP_O_K    : ulCommState = commErrOk_esp01 ;        break ;
    case NIETIMEOUT : ulCommState = commErrWouldBlock_esp01; break ;
    default         : ulCommState = commErrNotOk_esp01;      
                      eo40NiBuildErrorString ( pErrText,
                                               ERRMSG_NICOM_PEEK_ERROR, SapRc );
                      ErrTrace(tf);
                      break ;
  }
  return ( ulCommState );
}

/*===========================================================================*/

int eo40NiIsValidHandle      ( void                     *NiHandle )
{
  return ( NiHandle != NULL && *(NI_HDL *)NiHandle != NI_INVALID_HDL ) ;
}


/*===========================================================================*/

externC tsp00_Longuint 
eo40NiClose     ( void                         **NiHandle)
{
  tsp00_Longuint          ulCommState = commErrOk_esp01;
  SAPRETURN               SapRc ;

  if ( eo40NiIsValidHandle ( *NiHandle) )
  { 
    SapRc = NiCloseHandle (**(NI_HDL **)NiHandle) ;
    free ( *NiHandle ) ;
    *NiHandle = NULL ;
  }
  return ( ulCommState );
}

/*===========================================================================*/

externC tsp00_Longuint 
eo40NiFinish(tsp00_ErrTextc VAR_ARRAY_REF   pErrText)
{
  SAPRETURN               SapRc = NiExit() ;
  tsp00_Longuint          ulCommState ;
  
  if ( SapRc  == SAP_O_K )
    ulCommState = commErrOk_esp01 ;
  else
  {
    ulCommState = commErrNotOk_esp01 ;
    eo40NiBuildErrorString ( pErrText, ERRMSG_NICOM_FINISH_ERROR, SapRc );
    ErrTrace(tf);
  }
  return ( commErrOk_esp01 );
}

/*===========================================================================*/

externC tsp00_Longuint 
eo40NiListen    ( void                        **NiHandle,
                  unsigned short               *Service,
                  tsp00_ErrTextc VAR_ARRAY_REF  pErrText )
{
  tsp00_Longuint          CommState = commErrOk_esp01;
  SAPRETURN               SapRc ;
  NI_HDL                  ListenHandle ;
  char                    szService[6] ;

  eo40GetServiceAsString ( Service, szService, sizeof(szService) );

  SapRc = NiListen (szService, &ListenHandle) ;

  if ( SapRc == SAP_O_K)
  {
    *NiHandle = eo40NiDupHandle( &ListenHandle ) ;
    CommState = *NiHandle ? commErrOk_esp01 : commErrNotOk_esp01 ;
  }
  else
  {
    eo40NiBuildErrorString ( pErrText, ERRMSG_NICOM_LISTEN_ERROR, SapRc );
    CommState = commErrNotOk_esp01 ;
    ErrTrace(tf);
  }
  return ( CommState );
}

/*===========================================================================*/

externC tsp00_Longuint 
eo40NiAccept    ( void                        *NiListenHandle,
                  void                       **NiCommHandle,
                  char                        *ServerNode, 
                  int                          MaxServerNodeLen,
                  tsp00_ErrTextc VAR_ARRAY_REF ErrText )
{
  tsp00_Longuint          CommState = commErrOk_esp01;
  SAPRETURN               SapRc ;
  NI_HOSTADR              hostaddr ;
  NI_HDL                  NiHandle ;

  SapRc = NiAccept (*(NI_HDL *)NiListenHandle, NI_BLOCK, &NiHandle, &hostaddr );

  if ( SapRc == SAP_O_K)
  {
    char *HostName = NiAddrToHost (&hostaddr) ;

    if ( HostName == NULL )
      HostName = NiAdrToStr (&hostaddr);

    if ( HostName != NULL )
    {
      strncpy ( ServerNode, HostName, MaxServerNodeLen ) ;
      ServerNode [ MaxServerNodeLen -1 ] = '\0' ;
    }
    *NiCommHandle = eo40NiDupHandle( &NiHandle ) ;
    CommState = *NiCommHandle ? commErrOk_esp01 : commErrNotOk_esp01 ;
  }
  else
  {
    CommState = commErrNotOk_esp01 ;
    eo40NiBuildErrorString ( ErrText, ERRMSG_NICOM_ACCEPT_ERROR, SapRc );
    ErrTrace(tf);
  }

  return ( CommState );
}



/*===========================================================================*/

tsp00_Longuint 
eo40NiSend ( void                         *SendHandle,
             char                         *pData,
             int                           Len,
             tsp00_ErrTextc VAR_ARRAY_REF  pErrText )
{
  tsp00_Longuint          CommState = commErrOk_esp01;
  SAPRETURN               SapRc ;
  NI_HDL                  NiHandle = *((NI_HDL *)SendHandle ) ;
  int                     BytesSend;

  for (BytesSend = 0; 
       Len > 0 && CommState == commErrOk_esp01 ; 
       pData += BytesSend, Len -= BytesSend )
  {
    SapRc = NiRawWrite (NiHandle, (SAP_BYTE *)pData, Len, NI_BLOCK, &BytesSend);
    if ( SapRc != SAP_O_K)
    {
      eo40NiBuildErrorString ( pErrText, ERRMSG_NICOM_SOCKET_SND_ERROR, SapRc );
      CommState = commErrNotOk_esp01 ;
      ErrTrace(tf);
    }
  }
  return ( CommState );
}


/*===========================================================================*/

externC int   
eo40NiIsSaprouterSring     (  char                   *StrToCheck )
{
  return( memcmp (StrToCheck, NI_HOSTSEP, NI_SEPSIZ) == 0 );
}

/*===========================================================================*/

externC int   
eo40NiExtractDBNodeFromSaprouterSring ( char                        *SaprouterSring,
                                        tsp00_NodeIdc VAR_ARRAY_REF  ServerNode  ) 
{
  int IsSaprouterSring = eo40NiIsSaprouterSring(SaprouterSring);

  if ( IsSaprouterSring )
  {
    int   NodeLen ;
    char *NodeBeg = SaprouterSring + NI_SEPSIZ ,
         *Ptr = NodeBeg ;

    /* search last host in routerstring ==> server node */
    while ( (Ptr = strstr ( Ptr, NI_HOSTSEP )) != NULL )
    {
      NodeBeg = Ptr + NI_SEPSIZ;
      Ptr     = NodeBeg ;
    }

    Ptr = strchr(NodeBeg, *NI_HOSTSEP ) ;/* end of host */

    NodeLen = Ptr ? Ptr - NodeBeg : strlen(NodeBeg);
    
    if ( NodeLen >= sizeof(tsp00_NodeIdc) )  
      NodeLen = sizeof(tsp00_NodeIdc) - 1 ;

    memcpy( ServerNode, NodeBeg, NodeLen );
    ServerNode [ NodeLen+1 ] = '\0' ;
    }

  return ( IsSaprouterSring ) ;
}

/*===========================================================================*/
tsp00_Longuint 
eo40NiReceive ( void                         *RecvHandle,
                char                         *pData,
                tsp00_Longuint                ulMaxDataLen,
                tsp00_Longuint               *BytesReceived,
                tsp00_ErrTextc VAR_ARRAY_REF  pErrText )
{
  #undef  MF__
  #define MF__ MOD__"eo40NiReceive"
 
  tsp00_Longuint          CommState = commErrOk_esp01;
  SAPRETURN               SapRc ;
  SAP_INT                 readLen;
  NI_HDL                  NiHandle = *((NI_HDL *)RecvHandle ) ;

  SapRc = NiRawRead (NiHandle, (SAP_BYTE *)pData, ulMaxDataLen, NI_BLOCK, &readLen);

  DBGNI(( "header: wanted %d bytes, read %d, SapRc = %d",
           BytesRemaining, *BytesReceived, SapRc ));

  if ( SapRc != SAP_O_K )
  {
      eo40NiBuildErrorString ( pErrText, ERRMSG_NICOM_REC_ERROR, SapRc );
      CommState = commErrNotOk_esp01 ;
      ErrTrace(tf);
  }
  else
  {
     *BytesReceived = readLen;
  }

  return ( CommState );
}

/*===========================================================================*/
tsp00_Longuint 
eo40NiConnectToServer( teo40_NiConnectInfoRecord    *pNIConnInfo,
                       tsp00_ErrTextc VAR_ARRAY_REF  ErrText )
{
  SAPRETURN      SapRc ;
  tsp00_Longuint CommState;
  NI_HDL         NiHandle ;
  char           szService[6] ;

  eo40GetServiceAsString ( &pNIConnInfo->usServicePort, szService, sizeof(szService) );

  SapRc = NiRawConnect (pNIConnInfo->szSapRouterString, szService, 
                        NI_BLOCK, &NiHandle);

  if ( SapRc == SAP_O_K )
  {
    pNIConnInfo->NiHandle = eo40NiDupHandle( &NiHandle ) ;
    CommState = pNIConnInfo->NiHandle ? commErrOk_esp01 : commErrNotOk_esp01 ;
  } 
  else
  {
    eo40NiBuildErrorString ( ErrText, ERRMSG_NICOM_CONNECT_ERROR, SapRc );
    CommState = commErrNotOk_esp01 ;
    ErrTrace(tf);
  }
  /* ?? socket options ?? */
  return ( CommState ) ;
}


/*===========================================================================*
 *  LOCAL/PRIVATE/PROTECTED FUNCTIONS (CODE)                                 *
 *===========================================================================*/

/*===========================================================================*/

#define ERR_NIEINTERN           "internal error"
#define ERR_NIEHOST_UNKNOWN     "unknown Host"
#define ERR_NIESERV_UNKNOWN     "unknown Service"
#define ERR_NIESERV_USED        "Service in use"
#define ERR_NIETIMEOUT          "timeout"
#define ERR_NIECONN_BROKEN      "connection broken"
#define ERR_NIECONN_REFUSED     "connection refused"

#define ERR_NIEINVAL            "invalid parameter"
#define ERR_NIEPING             "ping received"
#define ERR_NIECONN_PENDING     "NIECONN_PENDING"
#define ERR_NIEVERSION          "bad version"
 
#define ERR_NIEROUT_HOST_UNKNOWN "ROUT: unknown host"
#define ERR_NIEROUT_SERV_UNKNOWN "ROUT: unknown service"
#define ERR_NIEROUT_PERM_DENIED  "ROUT: perm. denied"
#define ERR_NIEROUT_CONN_REFUSED "ROUT: conn. refused"
#define ERR_NIEROUT_CONN_BROKEN  "ROUT: conn. broken"
#define ERR_NIEROUT_INTERN       "ROUT: intern. error"
#define ERR_NIEROUT_VERSION      "ROUT: bad version"
#define ERR_NIEROUT_CANCELED     "ROUT: comm. canceled"

#define ERR_NIEROUT_SHUTDOWN     "ROUT: shutdown"
#define ERR_NIEROUT_INFO_DENIED  "ROUT: info denied"
#define ERR_NIEROUT_OVERFLOW     "ROUT: client overflow"
#define ERR_NIEROUT_MODE_DENIED  "ROUT: talkmode forbidden"
#define ERR_NIEROUT_NOCLIENT     "ROUT: client missing"
#define ERR_NIEROUT_EXTERN       "ROUT: ext. lib error"
#define ERR_NIEROUT_SNC_FAILURE  "ROUT: SNC Error"

#define ERR_NI_UNKNOWN_RC        "NI: RC="
/*===========================================================================*/

static void 
eo40NiBuildErrorString ( tsp00_ErrTextc VAR_ARRAY_REF ErrTextC,
                         char                        *ErrorMsgHeader,
                         SAPRETURN                    SapRc )
{
  char *NiErrorString ;
  char UnknownRC[sizeof(ERR_NI_UNKNOWN_RC) + 8];
  int  ErrorHeaderLen  = strlen(ErrorMsgHeader) ;
  int  ErrorTrailerLen = sizeof(tsp00_ErrTextc) - ErrorHeaderLen - 1 ;

  switch ( SapRc ) 
  {
    case NIEHOST_UNKNOWN     : NiErrorString = ERR_NIEHOST_UNKNOWN     ; break ; 
    case NIESERV_UNKNOWN     : NiErrorString = ERR_NIESERV_UNKNOWN     ; break ; 
    case NIESERV_USED        : NiErrorString = ERR_NIESERV_USED        ; break ; 
    case NIETIMEOUT          : NiErrorString = ERR_NIETIMEOUT          ; break ; 
    case NIECONN_BROKEN      : NiErrorString = ERR_NIECONN_BROKEN      ; break ; 
    case NIEINVAL            : NiErrorString = ERR_NIEINVAL            ; break ; 
    case NIECONN_REFUSED     : NiErrorString = ERR_NIECONN_REFUSED     ; break ;
    case NIEPING             : NiErrorString = ERR_NIEPING             ; break ;
    case NIECONN_PENDING     : NiErrorString = ERR_NIECONN_PENDING     ; break ;
    case NIEVERSION          : NiErrorString = ERR_NIEVERSION          ; break ;
    case NIEINTERN           : NiErrorString = ERR_NIEINTERN           ; break ; 

    case NIEROUT_HOST_UNKNOWN: NiErrorString = ERR_NIEROUT_HOST_UNKNOWN; break ;
    case NIEROUT_SERV_UNKNOWN: NiErrorString = ERR_NIEROUT_SERV_UNKNOWN; break ;
    case NIEROUT_PERM_DENIED : NiErrorString = ERR_NIEROUT_PERM_DENIED ; break ;
    case NIEROUT_CONN_REFUSED: NiErrorString = ERR_NIEROUT_CONN_REFUSED; break ;
    case NIEROUT_CONN_BROKEN : NiErrorString = ERR_NIEROUT_CONN_BROKEN ; break ;
    case NIEROUT_INTERN      : NiErrorString = ERR_NIEROUT_INTERN      ; break ; 
    case NIEROUT_VERSION     : NiErrorString = ERR_NIEROUT_VERSION     ; break ; 
    case NIEROUT_CANCELED    : NiErrorString = ERR_NIEROUT_CANCELED    ; break ; 

    case NIEROUT_SHUTDOWN    : NiErrorString = ERR_NIEROUT_SHUTDOWN    ; break ; 
    case NIEROUT_INFO_DENIED : NiErrorString = ERR_NIEROUT_INFO_DENIED ; break ; 
    case NIEROUT_OVERFLOW    : NiErrorString = ERR_NIEROUT_OVERFLOW    ; break ; 
    case NIEROUT_MODE_DENIED : NiErrorString = ERR_NIEROUT_MODE_DENIED ; break ; 
    case NIEROUT_NOCLIENT    : NiErrorString = ERR_NIEROUT_NOCLIENT    ; break ; 
    case NIEROUT_EXTERN      : NiErrorString = ERR_NIEROUT_EXTERN      ; break ; 
    case NIEROUT_SNC_FAILURE : NiErrorString = ERR_NIEROUT_SNC_FAILURE ; break ; 

    default                  : strcpy(UnknownRC, ERR_NI_UNKNOWN_RC) ;
                               sql47_itoa(SapRc, UnknownRC  + sizeof(ERR_NI_UNKNOWN_RC),
                                          sizeof(UnknownRC) - sizeof(ERR_NI_UNKNOWN_RC) ) ;
                               NiErrorString = UnknownRC               ; break ;
  }

  strcpy( ErrTextC, ErrorMsgHeader ) ;

  if ( ErrorTrailerLen > 0 ) 
  {
    ErrTextC [ ErrorHeaderLen ] = ':';
    strncpy( &ErrTextC [ ErrorHeaderLen + 1] , NiErrorString, ErrorTrailerLen ) ;
  }

}

/*===========================================================================*/

static void * 
eo40NiDupHandle ( void *NiHandle )
{
  void *NiDupHandle = malloc ( sizeof(NI_HDL) ) ;

  if ( NiDupHandle )
    memcpy( NiDupHandle, NiHandle, sizeof(NI_HDL) ) ;

  return ( NiDupHandle ) ;
}

/*===========================================================================*/
static void 
eo40GetServiceAsString ( unsigned short *Service,
                         char           *szService,
                         int             szServiceLen)
{
  unsigned short NiServiceNo ;

  if ( !Service || *Service == 0 )
  {
    NiServiceNo = NI_DEFAULT_SERVICE ;
    if (Service)
      *Service = NiServiceNo; /* give used ServiceNo back */
  }
  else
    NiServiceNo = *Service ;

  sql47_itoa( ( int )NiServiceNo, szService, szServiceLen);
}

/*===========================================================================*/
externC int   
eo40NiExtractServiceFromSaprouterSring ( char                        *SaprouterSring,
                                         unsigned short              *Service )
{
  int IsSaprouterSring = eo40NiIsSaprouterSring(SaprouterSring);

  if ( IsSaprouterSring )
  {
    char *LastHost = SaprouterSring ;
    char *Ptr      = LastHost;

    /* search last host in routerstring ==> server node */

    while ( (Ptr = strstr ( LastHost+1, NI_HOSTSEP ) ) )
      LastHost = Ptr ;
 
    if ( (Ptr = strstr ( LastHost, NI_SERVSEP )) == NULL ) 
    {
      *Service = 0 ;
    }
    else
    {
      /* There is only one NI_HOSTSEP AND a NI_SERVSEP */
      int i;
      char szService[6] ;

      for ( i=0, Ptr += NI_SEPSIZ ;
            i < sizeof(szService)-1 && isdigit(*Ptr) ; 
            i++, Ptr++ )
        szService[i] = *Ptr ;

       szService[i] = '\0' ;

      *Service = (unsigned short) atoi(szService);
    }
  }
  return ( IsSaprouterSring ) ;
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
#else /* USE_SAP_NI */
/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
externC int
eo40NiIsAvailable() 
{
  return( 0 );
}

externC tsp00_Longuint 
eo40NiInit        ( tsp00_Pathc     VAR_ARRAY_REF  TraceFile,
                    int                            TraceLevel,
                    tsp00_ErrTextc  VAR_ARRAY_REF  ErrText )
{
  return ( commErrNotOk_esp01 );
}

externC tsp00_Longuint 
eo40NiHandleToSocket        ( void                         *NiHandle ,
                              int                          *Sd,
                              tsp00_ErrTextc  VAR_ARRAY_REF  ErrText )
{
  return ( commErrNotOk_esp01 );
}

externC tsp00_Longuint 
eo40NiDataAvailable   ( void                         *NiHandle ,
                        tsp00_ErrTextc VAR_ARRAY_REF  pErrText )
{
  return ( commErrNotOk_esp01 );
}

externC tsp00_Longuint 
eo40NiSqlCancelDump       ( teo40_NiConnectInfoRecord    *pNIConnInfo,
                            tsp00_Longuint                ulReqType,
                            tsp00_ErrTextc VAR_ARRAY_REF  ErrText )
{
  return ( commErrNotOk_esp01 );
}


externC int   
eo40NiExtractDBNodeFromSaprouterSring ( char                         *SaprouterSring,
                                        tsp00_NodeIdc  VAR_ARRAY_REF  ServerNode  )
{
  return ( commErrNotOk_esp01 );
}

externC int   
eo40NiExtractServiceFromSaprouterSring ( char                        *SaprouterSring,
                                         unsigned short              *Service )
{
  *Service = 0 ;
  return ( commErrNotOk_esp01 );
}

externC int   
eo40NiIsSaprouterSring   (  char                  *StrToCheck )
{
  return ( 0);
}

externC int   
eo40NiIsValidHandle      ( void                   *NiHandle )
{
  return ( 0 );
}

externC tsp00_Longuint 
eo40NiClose     ( void                         **NiHandle)
{
  return ( commErrNotOk_esp01 );
}

externC tsp00_Longuint 
eo40NiFinish(tsp00_ErrTextc VAR_ARRAY_REF   pErrText)
{
  return ( commErrNotOk_esp01 );
}

externC tsp00_Longuint 
eo40NiListen    ( void                         **NiHandle,
                  unsigned short                *Service,
                  tsp00_ErrTextc VAR_ARRAY_REF   pErrText )
{
  return ( commErrNotOk_esp01 );
}

externC tsp00_Longuint 
eo40NiAccept    ( void                          *NiListenHandle,
                  void                         **NiCommHandle,
                  char                          *ServerNode, 
                  int                            MaxServerNodeLen,
                  tsp00_ErrTextc VAR_ARRAY_REF   pErrText )
{
  return ( commErrNotOk_esp01 );
}

externC tsp00_Longuint 
eo40NiSend              ( void                         *SendHandle,
                          char                         *pData,
                          int                           Len,
                          tsp00_ErrTextc VAR_ARRAY_REF  pErrText )
{
  return ( commErrNotOk_esp01 );
}

tsp00_Longuint 
eo40NiReceive ( void                         *RecvHandle,
                char                         *pData,
                tsp00_Longuint                ulMaxDataLen,
                tsp00_Longuint               *BytesReceived,
                tsp00_ErrTextc VAR_ARRAY_REF  pErrText )
{
  return ( commErrNotOk_esp01 );
}

tsp00_Longuint 
eo40NiConnectToServer   ( teo40_NiConnectInfoRecord    *pNIConnInfo,
                          tsp00_ErrTextc VAR_ARRAY_REF  ErrText )
{
  return ( commErrNotOk_esp01 );
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
#endif /* USE_SAP_NI */
/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/