/****************************************************************************

  module      : veo58.c

  -------------------------------------------------------------------------

  responsible : FrankH

  special area: kernel / LZU
  description : xparamfile access (configuration) 


  last changed: 2001-01-03  10:51
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1998



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/

#define MOD__ "veo58.c: "

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "geo00_0.h"
#include "geo50_0.h"
#include "geo007_1.h"
#include "heo46.h"
#include "heo52.h"
#include "heo58.h"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/

/*===========================================================================*
 *  GLOBAL VARIABLES                                                         *
 *===========================================================================*/

/*===========================================================================*
 *  LOCAL VARIABLES                                                          *
 *===========================================================================*/

/*===========================================================================*
 *  LOCAL FUNCTIONS (PROTOTYPES)                                             *
 *===========================================================================*/

/*===========================================================================*
 *  GLOBAL FUNCTIONS (CODE)                                                  *
 *===========================================================================*/

/* --------------------------------------------------------------------
Create a unique id like : p26202.bea.sap-ag.de:MUT_629_D19980908T174400
                          ------- NODE -------:--DB---_DYYYYMMDDTHHMMSS
                                                       Date     Time
-------------------------------------------------------------------- */
void	vget_uniqe_id ( tsp_line            uniqe_id )
{
#undef  MF__
#define MF__ MOD__"vget_uniqe_id"

/*********************************************************************/
#if _WIN32 || OS2 
/*********************************************************************/
SQL_DBNAMEC                             *pszServerDB   = &(kgs.szServerDB  );
SQL_NODEIDC                             *pszServerNode = &(kgs.szServerNode);
/*********************************************************************/
#else
/*********************************************************************/
SQL_NODEIDC                             *pszServerNode = & KGS->servernode ;
SQL_DBNAMEC                             *pszServerDB   = & KGS->serverdb ;
/*********************************************************************/
#endif
/*********************************************************************/

  int DBNodeLen     = (int)strlen ( (const char *)*pszServerNode ) ;
  int DBNameLen     = (int)strlen ( (const char *)*pszServerDB ) ;
  tsp_time TimeStamp ;
  int TimeLen       = sizeof(tsp_time) ;
  int DateLen       = sizeof(tsp_date) ;
  tsp00_Byte *PtrUniqueId =  uniqe_id;
  
  if ( DBNodeLen + 1 + DBNameLen + 1 + TimeLen + DateLen 
                        > sizeof (tsp_line ) )
  {
    DBNodeLen = sizeof(tsp_line) - DateLen - TimeLen - 1 - DBNameLen - 1 ;
  }
  memset ( PtrUniqueId, ' ', sizeof(tsp_line) ) ;

  memcpy( PtrUniqueId, *pszServerNode, DBNodeLen ) ;
  PtrUniqueId += DBNodeLen ;
  *PtrUniqueId = ':' ;
  PtrUniqueId++ ;
  memcpy( PtrUniqueId, *pszServerDB, DBNameLen ) ;
  PtrUniqueId += DBNameLen ;
  *PtrUniqueId = '_' ;
  PtrUniqueId++ ;
  vdattime ( PtrUniqueId, TimeStamp ) ;
  PtrUniqueId += DateLen ;
  *PtrUniqueId = '_' ;
  PtrUniqueId++ ;
  memcpy( PtrUniqueId, &TimeStamp[2], TimeLen-2 ) ; /* hours only 2 Byte */
}

/*
 *============================================================================
 */
void    vcold ( struct tsp200_xparam_rec           * prms ,
                tsp200_vcold_result                * vcold_result )

{
/*********************************************************************/
#if _WIN32 || OS2 
/*********************************************************************/
#undef  MF__
#define MF__ MOD__"vcold"
#ifdef DEBUG
PUKT_CTRL_REC   pUKT = THIS_UKT_CTRL;
#endif

SQL_DBNAMEC                             *pszServerDB   = &(kgs.szServerDB  );
SQL_NODEIDC                             *pszServerNode = &(kgs.szServerNode);
RTE_XPARAM_REC                          *XParam        = kgs.XParam ;

DBGIN_T (pUKT->pCTask->ulTaskIndex);
/*********************************************************************/
#else
/*********************************************************************/
#ifdef DEBUG
ten50_UKT_Ctrl *this_ukt = THIS_UKTR_CTRL;
#endif
SQL_NODEIDC                             *pszServerNode = & KGS->servernode ;
SQL_DBNAMEC                             *pszServerDB   = & KGS->serverdb ;
RTE_XPARAM_REC                          *XParam = KGS->XParam ;

DBG1 (( "vcold: called,    T%ld \n", this_ukt->curr_task->index ));
/*********************************************************************/
#endif
/*********************************************************************/

/* PTS 1105389 */
eo46CtoP ( prms->sp2xp_instance_type , XParam->szInstanceType, sizeof(tsp_c8) );
eo46CtoP ( prms->sp2xp_serverdb      , *pszServerDB   , sizeof(tsp_dbname) );
eo46CtoP ( prms->sp2xp_servernode    , *pszServerNode , sizeof(tsp_nodeid) );
eo46CtoP ( prms->sp2xp_controluser   , XParam->szControlUser , 
                                      sizeof(tsp_knl_identifier ) );
memcpy     ( prms->sp2xp_controlpassword  , XParam->crypControlPassw, 
                                      sizeof(tsp_cryptpw) );
eo46CtoP ( prms->sp2xp_rundirectory     , XParam->szRunDirectory ,
                                      sizeof(tsp_vfilename));
/*********************************************************************/
#if _WIN32 || OS2
/*********************************************************************/
eo46CtoP ( prms->sp2xp_opmsg1   , ""               , sizeof(tsp_vfilename));
eo46CtoP ( prms->sp2xp_opmsg2   , ""               , sizeof(tsp_vfilename));
/*********************************************************************/
#else
/*********************************************************************/
eo46CtoP ( prms->sp2xp_opmsg1   , XParam->szOpMsg1 , sizeof(tsp_vfilename));
eo46CtoP ( prms->sp2xp_opmsg2   , XParam->szOpMsg2 , sizeof(tsp_vfilename));
/*********************************************************************/
#endif
/*********************************************************************/
eo46CtoP ( prms->sp2xp_mbset1           , XParam->szMultiByteSet_1 , 
                                      sizeof(tsp_knl_identifier) );
eo46CtoP ( prms->sp2xp_mbset2           , XParam->szMultiByteSet_2 , 
                                      sizeof(tsp_knl_identifier) );
eo46CtoP ( prms->sp2xp_mbset3           , XParam->szMultiByteSet_3 , 
                                      sizeof(tsp_knl_identifier) );
eo46CtoP ( prms->sp2xp_mbset4           , XParam->szMultiByteSet_4 , 
                                      sizeof(tsp_knl_identifier) );
eo46CtoP ( prms->sp2xp_mbset5           , XParam->szMultiByteSet_5 , 
                                      sizeof(tsp_knl_identifier) );
eo46CtoP ( prms->sp2xp_default_code    , XParam->szDefaultCode , 
                                      sizeof(tsp_c18) );
eo46CtoP ( prms->sp2xp_date_time_format, XParam->szDateTimeFormat , 
                                      sizeof(tsp_c18) );

prms->sp2xp_no_of_backup_devs   = (tsp_int4 ) XParam->lNoOfBackupDevs;
prms->sp2xp_log_backup_to_pipe  = (tsp00_Bool)XParam->fLogBackToPipe;        /* PTS 1108249 */
prms->sp2xp_maxautobackupdevs   = (tsp_int4 ) XParam->ulMaxAutoBackupDevs;
prms->sp2xp_maxtrans            = (tsp_int4 ) XParam->lMaxTrans;
prms->sp2xp_data_cache_pages    = (tsp_int4 ) XParam->lDataCachePages;
prms->sp2xp_sequence_cache_pages  =  (tsp_int4)XParam->lSequenceCache;
prms->sp2xp_ccp_catalog_cache_pages =(tsp_int4)XParam->lCatalogCachePages;
prms->sp2xp_min_server_desc     = (tsp_int4 ) XParam->lMinServerDesc;
prms->sp2xp_distrib_desc_cache  = (tsp_int4 ) XParam->lDistribDescCache;
prms->sp2xp_distrib_cmd_cache   = (tsp_int4 ) XParam->lServerCmdCache;
prms->sp2xp_maxuser             = (tsp_int4 ) XParam->ulMaxUser;
prms->sp2xp_maxserverdb         = (tsp_int4 ) XParam->lMaxServerDB;
prms->sp2xp_maxservertask       = (tsp_int4 ) XParam->ulMaxServer;
prms->sp2xp_maxdatapages        = (tsp_int4 ) XParam->lMaxDataPages;
prms->sp2xp_log_queue_pages     = (tsp_int4 ) XParam->lLogQueuePages;
prms->sp2xp_log_cache_pages     = (tsp_int4 ) XParam->lLogCachePages;
prms->sp2xp_maxlocks            = (tsp_int4 ) XParam->lMaxLocks;
prms->sp2xp_deadlock_detection  = (tsp_int4 ) XParam->ulDeadlockDetection;
prms->sp2xp_lock_supply_block   = (tsp_int4 ) XParam->ulLockSupplyBlock;
prms->sp2xp_conv_cache_pages    = (tsp_int4 ) XParam->lConvCachePages;
prms->sp2xp_fbm_cache_pages     = (tsp_int4 ) XParam->lFbmCachePages;
prms->sp2xp_fbm_devsp_compr     = (tsp_int4 ) XParam->lFbmCompression;
prms->sp2xp_fbm_devsp_balance   = (tsp_int4 ) XParam->lFbmBalance;
prms->sp2xp_fbm_low_io_rate     = (tsp_int4 ) XParam->lFbmLowIORate;
prms->sp2xp_usm_cache_pages     = (tsp_int4 ) XParam->lUsmCachePages;
prms->sp2xp_backup_block_count  = (tsp_int4 ) XParam->lBackUpBlockCount;
prms->sp2xp_default_lru         = (tsp_int4 ) XParam->fDefaultLRU;
prms->sp2xp_db_for_sap          = (tsp_int4 ) XParam->fServerDBForSAP;
prms->sp2xp_join_search_level   = (tsp_int4 ) XParam->lJoinSearchLevel;
prms->sp2xp_join_maxtab_level4  = (tsp_int2 ) XParam->lJoinMaxTabLevel4;
prms->sp2xp_join_maxtab_level9  = (tsp_int2 ) XParam->lJoinMaxTabLevel9;
prms->sp2xp_distinct_overhead   = (tsp_int2 ) XParam->lDistinctOverhead;
prms->sp2xp_orderby_factor      = (tsp_int4 ) XParam->lOrderbyFactor;
prms->sp2xp_build_res_percent   = (tsp_int4 ) XParam->lBuildResultPercent ;
prms->sp2xp_fetch_res_percent   = (tsp_int4 ) XParam->lFetchResultPercent ;
prms->sp2xp_join_prefetch_percent=(tsp_int4 ) XParam->lJoinPrefetchPercent;
prms->sp2xp_key_inv_rate_percent= (tsp_int4 ) XParam->lKeyInvRatePercent;
prms->sp2xp_max_merge           = (tsp_int4 ) XParam->lMaxMerge;
prms->sp2xp_maxdevspaces        = (tsp_int4 ) XParam->lMaxDevspaces;
prms->sp2xp_maxdatadevspaces    = (tsp_int4 ) XParam->lMaxDataDevspaces;
prms->sp2xp_maxsysdevspaces     = (tsp_int4 ) XParam->lMaxSysDevspaces;
prms->sp2xp_maxlogdevspaces     = (tsp_int4 ) XParam->lMaxArchiveLogs;
prms->sp2xp_delay_commit        = (tsp_int4 ) XParam->fDelayCommit;
prms->sp2xp_maxuserstack        = (tsp_int4 ) XParam->ulMaxUsUtStack;
prms->sp2xp_delay_lw            = (tsp_int4 ) XParam->lDelayLW != 0 ;
prms->sp2xp_data_regions        = (tsp_int4 ) XParam->ulNumDataCacheReg ;
prms->sp2xp_tree_regions        = (tsp_int4 ) XParam->ulNumTreeReg ;
prms->sp2xp_trans_regions       = (tsp_int4 ) XParam->ulNumTransReg ;
prms->sp2xp_tab_regions         = (tsp_int4 ) XParam->ulNumTabReg ;
prms->sp2xp_row_regions         = (tsp_int4 ) XParam->ulNumRowReg ;
prms->sp2xp_oms_regions         = (tsp_int4 ) XParam->ulNumOMSReg ;
prms->sp2xp_unicode             = (tsp00_Bool)XParam->fUnicode ;
prms->sp2xp_mirrored_data       = (tsp00_Bool)XParam->fMirrData ;
prms->sp2xp_svp_1_conv_flush    = (tsp00_Bool)XParam->fSvpFstConvFlush ;
prms->sp2xp_log_mode            = (tsp_int4 ) XParam->lLogMode ;
prms->sp2xp_log_segment_size    = (tsp_int4 ) XParam->lLogSegmentSize ;
prms->sp2xp_session_timeout     = (tsp_int4 ) XParam->lSessionTimeout ;
prms->sp2xp_request_timeout     = (tsp_int4 ) XParam->lRequestTimeout ;
prms->sp2xp_net_timeout         = (tsp_int4 ) XParam->lNetTimeout ;
prms->sp2xp_lock_timeout        = (tsp_int4 ) XParam->lLockTimeout ;
prms->sp2xp_restart_time        = (tsp_int4 ) XParam->lRestartTime ;
prms->sp2xp_multi_io_block_count= (tsp_int4 ) XParam->ulMultIOBlockCount ;
prms->sp2xp_dw_io_block_count   = (tsp_int4 ) XParam->ulDwIOBufSize ;
prms->sp2xp_maxdatawriter       = (tsp_int4 ) XParam->ulMaxDataWriter ;
prms->sp2xp_packet_size         = (tsp_int4 ) XParam->ulPacketSize  ;
prms->sp2xp_pno_pool_size       = (tsp_int4 ) XParam->ulPnoPoolSize ;
prms->sp2xp_mb_data_size        = (tsp_int4 ) XParam->ulMBDataSize;
prms->sp2xp_mb_qual_size        = (tsp_int4 ) XParam->ulMBQualSize;
prms->sp2xp_mb_stack_size       = (tsp_int4 ) XParam->ulMBStackSize;
prms->sp2xp_kb_work_st_size     = (tsp_int4 ) XParam->ulKBWorkStSize;
prms->sp2xp_kb_workbuf_size     = (tsp_int4 ) XParam->ulKBWorkBufSize;
prms->sp2xp_ccs_catalog_cache_size  =(tsp_int4)XParam->ulCatalogCacheSize;
prms->sp2xp_maxcpu              = (tsp_int4 ) XParam->ulMaxCPU;
prms->sp2xp_mbtotalsize         = (tsp_int4 ) XParam->ulMultiByteTotalSize;
prms->sp2xp_dw_copy_boundary    = (tsp_int4 ) XParam->lDwCopyBoundary ;
prms->sp2xp_dw_io_area_size     = (tsp_int4 ) XParam->lDwIOAreaSize  ;
prms->sp2xp_dw_io_area_flush    = (tsp_int4 ) XParam->lDwIOAreaFlush ;
prms->sp2xp_dw_lru_tail_flush   = (tsp_int4 ) XParam->lDwLRUTailFlush ;
prms->sp2xp_log_start_copy      = (tsp_int4 ) XParam->lLogStartCopy;
prms->sp2xp_log_copy_distance   = (tsp_int4 ) XParam->lLogCopyDistance;
prms->sp2xp_reserved_redo_size  = (tsp_int4 ) XParam->lReservedRedoSize;
prms->sp2xp_obj_log_segment_size= (tsp_int4 ) XParam->ulObjLogSegmentSize;
prms->sp2xp_obj_log_ioqueue_pages=(tsp_int4 ) XParam->ulObjLogIOQueuePages;
prms->sp2xp_obj_log_rollbackcache=(tsp_int4 ) XParam->ulObjLogRollbackCache; 
prms->sp2xp_page_size           = (tsp_int4 ) XParam->ulDBPageSize;
prms->sp2xp_obj_log_size        = (tsp_int4 ) XParam->ulObjLogSize;
prms->sp2xp_maxeventtask        = (tsp_int4 ) XParam->ulMaxEventTasks;
prms->sp2xp_maxbackuptask       = (tsp_int4 ) XParam->ulMaxBackupTasks;
prms->sp2xp_maxgarbage_collector= (tsp_int4 ) XParam->ulMaxGarbageCollector;
prms->sp2xp_readahead_blobs_gt  = (tsp_int4 ) XParam->ulReadAheadBlobsGT;
prms->sp2xp_indexfile_list_size = (tsp_int4 ) XParam->lIndexfileListSize;
prms->sp2xp_oms_heap_limit      = (tsp_int4 ) XParam->ulOMSHeapLimit;
/*
XParam->lDistinctOverhead = (ULONG ) & prms->sp2xp_distinct_overhead   ;
*/
prms->sp2xp_max_ukt             = KGS->ulNumOfUKTs; /* PTS 1108768 */
* vcold_result = sp200vcr_ok ;

/*********************************************************************/
#if _WIN32 || OS2
/*********************************************************************/
DBG2 ((MF__, "xparam_rec:", prms, sizeof(*prms) ));

DBGOUT_T (pUKT->pCTask->ulTaskIndex);
/*********************************************************************/
#else
/*********************************************************************/
DBG1 (( "vcold: returning, T%ld \n", this_ukt->curr_task->index ));
/*********************************************************************/
#endif
/*********************************************************************/
}

tsp00_Bool sql58k_check_dup_devspace_name( DEVSPACES_DESCR *DevspacesDesc,
                                           tsp2_devnamec   szDevspaceName )
{
#undef  MF__
#define MF__ MOD__"sql58k_check_dup_devspace_name"

  tsp00_Bool        Ok = 0;
  DEVSPACES_DESCR   *pDevspDesc;
  ULONG             ulDevNo;
  tsp2_dev_type     TmpDevType;
  char              *pDevspaceName;

  DBGPAS;

  for ( TmpDevType = 0; TmpDevType <= sp2dt_trace; TmpDevType++ )
  {
    pDevspDesc = &(DevspacesDesc[TmpDevType]);

    for ( ulDevNo = 0; ulDevNo < pDevspDesc->ulNumDevs; ulDevNo++ )
    {
      if (pDevspDesc->Devspaces[ulDevNo].lDevspaceType == DEVSPACE_UNUSED) 
        continue;

      pDevspaceName = pDevspDesc->Devspaces[ulDevNo].szDevspaceName;

#     if defined (_WIN32)
       Ok = (_stricmp( pDevspaceName, szDevspaceName ) != 0);
#     else
       Ok = (strcmp( pDevspaceName, szDevspaceName ) != 0);
#     endif

      if ( !Ok )
      {
        MSGCD (( ERR_DEVSPACE_ALREADY_EXISTS, szDevspaceName ));
        return Ok;
      }
    }
  }

  return Ok;
}

/*
 * ==================================================================
 */

DEV_DESCR * sql58k_get_devsp_descr ( tsp2_dev_type dev_type ,
                                     tsp_int4      type_spec_devno )
                                     /* first_spec_devno = 0 !!! */
{
#undef  MF__
#define MF__ MOD__"sql58k_get_devsp_descr "

  RTE_XPARAM_REC                          *XParam        = KGS->XParam ;

  DBGPAS;

  if ( dev_type <= sp2dt_trace &&
       XParam->DevspDescr[dev_type].ulNumDevs > (ULONG) type_spec_devno ) 
    return ( &(XParam->DevspDescr[dev_type].Devspaces[type_spec_devno]) ) ;
  else
    return ( NULL ) ;
}

/*
 * ==================================================================
 */

void	vget_devsize ( tsp2_dev_type   dev_type ,
                       tsp_int4        type_spec_devno ,
                       tsp_int4        *dev_size ,
                       tsp00_BoolAddr  ok )
{
#undef  MF__
#define MF__ MOD__"vget_devsize"

  DEV_DESCR *devsp ;
  DBGPAS;

  devsp = sql58k_get_devsp_descr ( dev_type, --type_spec_devno ) ;

  *ok = devsp != NULL && devsp->lDevspaceType == dev_type ;

  if ( *ok )
    { *dev_size = devsp->lDevspaceSize ;
      DBG3 (( MF__, "type: %d, spec_devno: %d, size: %d\n", 
                            dev_type, type_spec_devno, *dev_size )) ;
    }
  return ;
}

/*
 * ==================================================================
 */
void	vget_devname ( tsp2_dev_type        dev_type ,
                       tsp_int4             type_spec_devno ,
                       tsp2_devname         dev_name ,
                       tsp00_BoolAddr       ok )
{
#undef  MF__
#define MF__ MOD__"vget_devname"

  DEV_DESCR *devsp ;

  DBGPAS;
  devsp = sql58k_get_devsp_descr ( dev_type, --type_spec_devno ) ;

  *ok = devsp != NULL && devsp->lDevspaceType == dev_type ;

  if ( *ok )
    {
      eo46CtoP ( dev_name , devsp->szDevspaceName, sizeof(tsp2_devname) );
      DBG3 (( MF__, "type: %d, spec_devno: %d, name: %s\n", 
                    dev_type, type_spec_devno, devsp->szDevspaceName));
    }
  return ;
}

/*
 * ==================================================================
 */
void	vnew_devspace( tsp2_dev_type        dev_type ,
                       tsp_int4             type_spec_devno ,
                       tsp2_devname         dev_name ,
                       tsp_int4             dev_size ,
                       tsp00_BoolAddr       ok )
{
#undef  MF__
#define MF__ MOD__"vnew_devspace"

  DEV_DESCR *new_devsp ;
  tsp2_devnamec   szDevspaceName;
  RTE_XPARAM_REC  *XParam = KGS->XParam;

  DBGPAS;

  eo46PtoC ( szDevspaceName, dev_name, sizeof(tsp2_devname));

  *ok = sql58k_check_dup_devspace_name ( XParam->DevspDescr, szDevspaceName );

  if ( *ok ) 
  {
    new_devsp = sql58k_get_devsp_descr ( dev_type, --type_spec_devno );

    *ok = new_devsp != NULL && new_devsp->lDevspaceType == DEVSPACE_UNUSED ;

    if ( *ok )
    { 
      eo46PtoC ( new_devsp->szDevspaceName, dev_name, sizeof(tsp2_devname));
      new_devsp->lDevspaceType = dev_type ;
      new_devsp->lDevspaceSize = dev_size ;
      DBG3 (( MF__, "type: %d, spec_devno: %d, name: %s, size %d\n", dev_type, 
        type_spec_devno, new_devsp->szDevspaceName, dev_size));
    }
  }
  return ;
}


/*
 * ==================================================================
 */
void	vrevoke_devspace( tsp2_dev_type        dev_type ,
                          tsp_int4             type_spec_devno )
{
#undef  MF__
#define MF__ MOD__"vrevoke_devspace"

  int        ok;
  DEV_DESCR *revoke_devsp ;
  DBGPAS;

  revoke_devsp = sql58k_get_devsp_descr ( dev_type, --type_spec_devno );

  ok = revoke_devsp != NULL && revoke_devsp->lDevspaceType == dev_type ;

  if ( ok )
    { 
      DBG3 (( MF__, "type: %d, spec_devno: %d, name: %s, size %d\n", 
                     dev_type, type_spec_devno, revoke_devsp->szDevspaceName, 
                    revoke_devsp->lDevspaceSize));
      revoke_devsp->lDevspaceType = DEVSPACE_UNUSED ;
      revoke_devsp->lDevspaceSize = 0 ;
    }
  return ;
}


/* PTS CR 1103950 */
/*
 * ==================================================================
 */
void	vinit_akdump_ptr( void  *acv_pointer ,        /* CR 1000209 */
                          void (*ak_dump_proc)() )    /* CR 1000209 */

{
#undef  MF__
#define MF__ MOD__"vinit_akdump_ptr"

/*********************************************************************/
#if _WIN32
/*********************************************************************/
  PUKT_CTRL_REC                  pUKT      = THIS_UKT_CTRL;
  PTASK_CTRL_REC                 pCurrTask = pUKT->pCTask;

  DBGIN_T (pCurrTask->ulTaskIndex);

/*********************************************************************/
#endif
/*********************************************************************/
  CURR_TASK->AcvPointer = acv_pointer ;
  CURR_TASK->AkDumpProc = ak_dump_proc ;
}

/*
 * ==================================================================
 */

/* CR 1104345 */
void *vGetAcvPtrFromCurrentTask()
{
#undef  MF__
#define MF__ MOD__"vGetAcvPtrFromCurrentTask"

return(CURR_TASK->AcvPointer); 
}

/* PTS 1108768 */
static tsp00_Bool wantTrace = false;
static tsp00_Bool wantDump = false;
/*!
  Function:    vInitTraceAndDump
  Description: Initialization Function to retrieve Trace and Dump Information
  Arguments: vtraceName [out] Name of Tracefile (knltrace)
             vdumpName  [out] Name of Dumpfile (knldump)
 */
void        vInitTraceAndDump ( tsp00_VFilename VAR_ARRAY_REF vtraceName,
                                tsp00_VFilename VAR_ARRAY_REF vdumpName
            					        )
{
  tsp00_Int4 traceDevsize;
  vdiaginit(&wantTrace, &wantDump, vtraceName, &traceDevsize, vdumpName);
}

/*!
  Function: vIsDumpWanted
  Description: Request current status of dump wanted flag
  Return value: true if dump of kernel structures is wanted, false if not
 */
externC tsp00_Bool  vIsDumpWanted  ( )
{
  return wantDump;
}

/*!
  Function: vIsTraceWanted
  Description: Request current status of trace wanted flag
  Return value: true if writing kernel trace is wanted, false if not
 */
externC tsp00_Bool  vIsTraceWanted ( )
{
  return wantTrace;
}

