/*!**************************************************************************

  module       : veo76.cpp

  -------------------------------------------------------------------------

  responsible : JoergM

  Special area: Spinlocks
  Description : System Independend RTE_Spinlock and RTE_LockedScope Interface

  first created:2000-10-27  14:07
  last changed: 2000-11-20  16:49

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "heo76.h"

#if defined(WIN32)

#  include <windows.h>

#else

#  if defined(OSF1)
#    define _XOPEN_SOURCE_EXTENDED
#  endif /* OSF1 */

#  include <unistd.h>
#  include <stdlib.h>

#  if defined(_REENTRANT)

#    if defined(SUN)
#      include <thread.h>
#    endif /*SUN*/

#    if defined(AIX)
/*AIX*/   extern "C" void yield(void);
#    endif

#    if defined(AIX)||defined(OSF1)||defined(HPUX)||defined(NMP)||defined(LINUX)
#      include <pthread.h>
#    endif /*AIX||OSF1||HPUX||NMP||LINUX*/

#  endif /*_REENTRANT*/

#  if defined(OSF1)
#    include <alpha/builtins.h>
#  endif /*OSF1*/

#  if defined(AIX)
#    include <sys/atomic_op.h>
#  endif

#  if defined(HPUX)
/* PTS 1108389 */
#    include <sys/mman.h>
#  endif

#endif /*WIN32*/
/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/

/*===========================================================================*
 *  STATIC/INLINE FUNCTIONS (PROTOTYPES)                                     *
 *===========================================================================*/

/*LINUX*/ static void LINUX_AtomicSwap(tsp00_Int4 *val1, tsp00_Int4 *val2);

/*===========================================================================*
 *=====================  END OF DECLARATIONS ================================*
 *===========================================================================*/

/*===========================================================================*
 *  EXPORTED FUNCTIONS, EXPORTED CLASS METHODS (IMPLEMENTATION)              *
 *===========================================================================*/
/*---------------------------------------------------------------------------*/

#if defined(WIN32)
/*WIN32*/ void   eo76GiveUpTimeSlice()
/*WIN32*/ {
/*WIN32*/   SleepEx(0, false); // This works for multithread as well as for singlethreaded version
/*WIN32*/ }
/*WIN32*/
/*WIN32*/ void   eo76InitLock( teo76_Lock &lock )
/*WIN32*/ {
/*WIN32*/   InterlockedExchange(&lock, 0); 
/*WIN32*/ }
/*WIN32*/
/*WIN32*/ bool   eo76TestAndLock( teo76_Lock &lock )
/*WIN32*/ {
/*WIN32*/   return ( InterlockedExchange(&lock, 1) != 0 ); 
/*WIN32*/ }
/*WIN32*/
/*WIN32*/ void   eo76Unlock( teo76_Lock &lock )
/*WIN32*/ {
/*WIN32*/   InterlockedExchange(&lock, 0); 
/*WIN32*/ }
/*WIN32*/
#else

void eo76GiveUpTimeSlice()
{
#if defined(_REENTRANT)

  // multithreading variant
# if defined(SUN)
/*SUN*/   thr_yield();
# endif /*SUN*/

# if defined(AIX)
/*AIX*/   yield();
# endif /*AIX*/

# if defined(OSF1)||defined(HPUX)||defined(NMP)||defined(LINUX)
/*OSF1||HPUX||NMP||LINUX*/   sched_yield();
# endif /*OSF1||HPUX||NMP||LINUX*/

#else

  // singlethreading variant
  sleep(0);

#endif /*_REENTRANT*/
}

#endif /* WIN32 */

#if defined(AIX)
/*AIX*/  void eo76InitLock( teo76_Lock & lock )
/*AIX*/  {
/*AIX*/    _clear_lock(&lock[0], 0);
/*AIX*/  }
/*AIX*/
/*AIX*/  bool eo76TestAndLock( teo76_Lock & lock )
/*AIX*/  {
/*AIX*/    return _check_lock(&lock[0], 0, 1);
/*AIX*/  }
/*AIX*/
/*AIX*/  void eo76Unlock( teo76_Lock & lock )
/*AIX*/  {
/*AIX*/    _clear_lock(&lock[0], 0);
/*AIX*/  }
#endif /* AIX */

#if defined(HPUX)
# if 1
/*HPUX*/ /* PTS 1108389 Somehow this calling sequence does not work while inside liboms */
/*HPUX*/  void eo76InitLock( teo76_Lock & lock )
/*HPUX*/  {
/*HPUX*/    volatile tsp00_Int4 * AlignedPointer = reinterpret_cast<volatile tsp00_Int4 *> ( // align pointer to chacheline size
/*HPUX*/    ( reinterpret_cast<unsigned long>(&lock) + (sizeof(tsp00_Int4)*CACHELINE_SIZE - 1) ) 
/*HPUX*/          & (~(sizeof(tsp00_Int4)*CACHELINE_SIZE - 1)) );
/*HPUX*/    *AlignedPointer = 1;
/*HPUX*/  }
/*HPUX*/
/*HPUX*/  extern "C" int e76_hpinterlock(volatile tsp00_Int4 *pLock);
/*HPUX*/  bool eo76TestAndLock( teo76_Lock & lock )
/*HPUX*/  {
/*HPUX*/    volatile tsp00_Int4 * AlignedPointer = reinterpret_cast<volatile tsp00_Int4 *> ( // align pointer to chacheline size
/*HPUX*/    ( reinterpret_cast<unsigned long>(&lock) + (sizeof(tsp00_Int4)*CACHELINE_SIZE - 1) ) 
/*HPUX*/          & (~(sizeof(tsp00_Int4)*CACHELINE_SIZE - 1)) );
/*HPUX*/    /* Make the cache line "dirty" to improve performance */
/*HPUX*/    /* This statement writes a zero to the high byte      */
/*HPUX*/    *(reinterpret_cast<volatile char *>(AlignedPointer)) = 0;
/*HPUX*/    return (e76_hpinterlock(AlignedPointer) == 0);
/*HPUX*/  }
/*HPUX*/    
/*HPUX*/  void eo76Unlock( teo76_Lock & lock )
/*HPUX*/  {
/*HPUX*/    volatile tsp00_Int4 * AlignedPointer = reinterpret_cast<volatile tsp00_Int4 *> ( // align pointer to chacheline size
/*HPUX*/    ( reinterpret_cast<unsigned long>(&lock) + (sizeof(tsp00_Int4)*CACHELINE_SIZE - 1) ) 
/*HPUX*/          & (~(sizeof(tsp00_Int4)*CACHELINE_SIZE - 1)) );
/*HPUX*/    *AlignedPointer = 1;
/*HPUX*/  }
# else
/*HPUX*/  void eo76InitLock( teo76_Lock & lock )
/*HPUX*/  {
/*HPUX*/    msemaphore * AlignedSem = reinterpret_cast<msemaphore *> ( // align pointer to chacheline size
/*HPUX*/    (char *)0 + ( ((reinterpret_cast<char *>(&lock)) - (char *)0) + (sizeof(tsp00_Int4)*CACHELINE_SIZE - 1)
/*HPUX*/          & (~(sizeof(tsp00_Int4)*CACHELINE_SIZE - 1)) ) );
/*HPUX*/    msem_init(AlignedSem, MSEM_UNLOCKED);
/*HPUX*/  }
/*HPUX*/
/*HPUX*/  bool eo76TestAndLock( teo76_Lock & lock )
/*HPUX*/  {
/*HPUX*/    msemaphore * AlignedSem = reinterpret_cast<msemaphore *> ( // align pointer to chacheline size
/*HPUX*/    (char *)0 + ( ((reinterpret_cast<char *>(&lock)) - (char *)0) + (sizeof(tsp00_Int4)*CACHELINE_SIZE - 1)
/*HPUX*/          & (~(sizeof(tsp00_Int4)*CACHELINE_SIZE - 1)) ) );
/*HPUX*/    /* Make the cache line "dirty" to improve performance */
/*HPUX*/    /* This statement writes a zero to the high byte      */
/*HPUX*/    *(reinterpret_cast<volatile char *>(AlignedSem)) = 0;
/*HPUX*/    return ( msem_lock(AlignedSem, MSEM_IF_NOWAIT) != -1 );
/*HPUX*/  }
/*HPUX*/    
/*HPUX*/  void eo76Unlock( teo76_Lock & lock )
/*HPUX*/  {
/*HPUX*/    msemaphore * AlignedSem = reinterpret_cast<msemaphore *> ( // align pointer to chacheline size
/*HPUX*/    (char *)0 + ( ((reinterpret_cast<char *>(&lock)) - (char *)0) + (sizeof(tsp00_Int4)*CACHELINE_SIZE - 1)
/*HPUX*/          & (~(sizeof(tsp00_Int4)*CACHELINE_SIZE - 1)) ) );
/*HPUX*/    msem_unlock(AlignedSem, 0);
/*HPUX*/  }
# endif
#endif /* HPUX */

#if defined(LINUX)
/*LINUX*/  void eo76InitLock( teo76_Lock & lock )
/*LINUX*/  {
/*LINUX*/    lock[0] = 0;
/*LINUX*/  }
/*LINUX*/
/*LINUX*/  bool eo76TestAndLock( teo76_Lock & lock )
/*LINUX*/  {
/*LINUX*/    tsp00_Int4 NewValue = 1;
/*LINUX*/    LINUX_AtomicSwap(&lock[0], &NewValue);
/*LINUX*/    return ( NewValue != 0 );
/*LINUX*/  }
/*LINUX*/
/*LINUX*/  void eo76Unlock( teo76_Lock & lock )
/*LINUX*/  {
/*LINUX*/    lock[0] = 0;
/*LINUX*/  }
#endif /* LINUX */

#if defined(NMP)
/*NMP*/  void eo76InitLock( teo76_Lock & lock )
/*NMP*/  {
/*NMP*/    initspin(&lock, -1);
/*NMP*/  }
/*NMP*/
/*NMP*/  bool eo76TestAndLock( teo76_Lock & lock )
/*NMP*/  {
/*NMP*/    return ( cspinlock(&lock) != 0 );
/*NMP*/  }
/*NMP*/
/*NMP*/  void eo76Unlock( teo76_Lock & lock )
/*NMP*/  {
/*NMP*/    spinunlock(&lock);
/*NMP*/  }
#endif /* NMP */

#if defined(OSF1)
/*OSF1*/  void eo76InitLock( teo76_Lock & lock )
/*OSF1*/  {
/*OSF1*/    lock[0] = 0;
/*OSF1*/  }
/*OSF1*/
/*OSF1*/  bool eo76TestAndLock( teo76_Lock & lock )
/*OSF1*/  {
/*OSF1*/    return ( __LOCK_LONG_RETRY(&lock[0], 1) == 0 ); // returns non-zero on lock success
/*OSF1*/  }
/*OSF1*/
/*OSF1*/  void eo76Unlock( teo76_Lock & lock )
/*OSF1*/  {
#pragma message disable unavolacc
/*OSF1*/    __UNLOCK_LONG(&lock[0]);
#pragma message enable unavolacc
/*OSF1*/  }
#endif /* OSF1 */

#if defined(SUN)
/*SUN*/  void eo76InitLock( teo76_Lock & lock )
/*SUN*/  {
/*SUN*/    lock[0] = 0;
/*SUN*/  }
/*SUN*/
/*SUN*/  extern "C" e76_interlock(tsp00_Int4 *pLock);
/*SUN*/  bool eo76TestAndLock( teo76_Lock & lock )
/*SUN*/  {
/*SUN*/    return e76_interlock(&lock[0]);
/*SUN*/  }
/*SUN*/
/*SUN*/  void eo76Unlock( teo76_Lock & lock )
/*SUN*/  {
/*SUN*/    lock[0] = 0;
/*SUN*/  }
#endif /* SUN */

/*===========================================================================*
 *  LOCAL/INLINE FUNCTIONS, LOCAL CLASS METHODS (IMPLEMENTATION)             *
 *===========================================================================*/

#if defined(LINUX) && defined (I386)
  /*
   * function: LINUX_AtomicSwap (LINUX only)
   * description: Does an atomic swap of two integer values
   * 
   * After this call first value and second values are exchanged
   *
   * arguments: val1 [inout] First value
   *            val2 [inout] Second value
   */
/*LINUX*/static void LINUX_AtomicSwap(tsp00_Int4 *val1, tsp00_Int4 *val2)
/*LINUX*/{
/*LINUX*/   asm("lock; xchg %0, (%1)"
/*LINUX*/        : "=r"( *val2 ), "=r" ( val1 )
/*LINUX*/        : "0" ( *val2 ), "1"  ( val1 ) );
/*LINUX*/}
#endif /*LINUX && I386*/

#if defined LINUX && defined IA64
/*LINUX*/#if defined __INTEL_COMPILER
/*LINUX*/#include <ia64intrin.h>
/*LINUX*/#endif
/*LINUX*/inline void
/*LINUX*/LINUX_AtomicSwap (tsp00_Int4 *val1, tsp00_Int4 *val2)
/*LINUX*/{
/*LINUX*/#if defined __GNUC__
/*LINUX*/	__asm__ __volatile__ ("mov ar.ccv = 0 ;;");
/*LINUX*/	__asm__ __volatile__ ("cmpxchg4.acq %0=%2,%3,ar.ccv"  \
/*LINUX*/	: "=r"(*val2), "=m"(*val1) \
/*LINUX*/	: "m"(*val1), "r"(*val2));
/*LINUX*/#endif
/*LINUX*/#if defined __INTEL_COMPILER
/*LINUX*/	*val2 = _InterlockedCompareExchange (val1, *val2, 0);
/*LINUX*/#endif
/*LINUX*/}
#endif /* LINUX && IA64 */

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/