/****************************************************************************

  module      : veo79.c

  -------------------------------------------------------------------------

  responsible : FrankH

  special area: kernel / LZU
  description : xparamfile access (configuration) 


  last changed: 2001-01-02  13:46
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1998



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#define MOD__ "veo79.c: "

#include "vsp003c.h"

#include "geo00_0.h"
#include "geo50_0.h"
#include "geo001.h"
#include "geo002.h"
#include "geo002xp.h"
#include "geo007_1.h"
#include "geo007_2.h"
#include "veo00c.h"

#include "heo01.h"
#include "heo11_1.h"

#include "heo46.h"
#include "heo47.h"
#include "heo79.h"
#include "geo50_0.h"
#include "geo57.h"
#include "heo58.h"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/
#define MAX_TASKCLUSTER_PARTS 10

#define MAX_EXCLUSIVE     250
#define MAX_REGIONS       40
#define ACT_PARAM_VERSION 0

/***************************************************************/
#if defined(_WIN32)
/***************************************************************/
#include <io.h>

#define MIN_DISP_LOOPS     1

/***************************************************************/
#else
/***************************************************************/

#define MIN_DISP_LOOPS     2

extern  tsp00_DbNamec      KERNEL_GLOBAL_DBNAME ;

/***************************************************************/
#endif
/***************************************************************/

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/
#define XPR_INT(xp_val)   sql79k_xpr_int (xp_val)

#define XPR_SCOPY( para_len, para, xp_val, xp_name) \
                                   sql79k_xpr_scopy ( para, xp_val )

/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/
#define XP_V_UPDATE_OFFLINE 0
#define XP_V_UPDATE_ONLINE  1

struct xparam_strncmp_table
{
	tsp11_XpKeyTypec name;		/* Name found in XParam file */
	size_t           len;       /* Maximum length to compare */

#define XP_V_SET_TASKCLUSTER  0
#define XP_V_SET_DEVSP_NAME   1
#define XP_V_SET_DEVSP_SIZE   2
#define XP_V_SET_M_DEVSP_NAME 3
	int              type;      /* Classification of expected type */
    int              UpdateMode;/* update param allowed in off/online state of db */

	tsp2_dev_type    dt;        /* Devspace type or filler */
	tsp2_dev_type    mdt;       /* Mirror Devspace type or filler */
};

static struct xparam_strncmp_table RteXparamTable1[] =
{
 { PAN_TASKCLUSTER_N ,     PAN_TASKCLUSTER_N_LN,      
		XP_V_SET_TASKCLUSTER, XP_V_UPDATE_OFFLINE, sp2dt_filler, sp2dt_filler    },
 { PAN_SYS_N ,             PAN_SYS_N_LN,              
		XP_V_SET_DEVSP_NAME,  XP_V_UPDATE_OFFLINE, sp2dt_sys,    sp2dt_filler    },
 { PAN_ARCHIVE_LOG_N ,     PAN_ARCHIVE_LOG_N_LN,
		XP_V_SET_DEVSP_NAME,  XP_V_UPDATE_OFFLINE, sp2dt_log,    sp2dt_filler    },
 { PAN_DATADEV_N ,         PAN_DATADEV_N_LN,
		XP_V_SET_DEVSP_NAME,  XP_V_UPDATE_OFFLINE, sp2dt_data,   sp2dt_filler    },
 { PAN_SYS_N_SIZE ,        PAN_SYS_N_SIZE_LN,
		XP_V_SET_DEVSP_SIZE,  XP_V_UPDATE_OFFLINE, sp2dt_sys,    sp2dt_filler    },
 { PAN_ARCHIVE_LOG_N_SIZE, PAN_ARCHIVE_LOG_N_SIZE_LN,
		XP_V_SET_DEVSP_SIZE,  XP_V_UPDATE_OFFLINE, sp2dt_log,    sp2dt_filler    },
 { PAN_DATADEV_N_SIZE,     PAN_DATADEV_N_SIZE_LN,
		XP_V_SET_DEVSP_SIZE,  XP_V_UPDATE_OFFLINE, sp2dt_data,   sp2dt_filler    },
 { PAN_M_SYS_N ,           PAN_M_SYS_N_LN,
		XP_V_SET_M_DEVSP_NAME,XP_V_UPDATE_OFFLINE, sp2dt_sys,    sp2dt_mirr_sys  },
 { PAN_M_ARCHIVE_LOG_N ,   PAN_M_ARCHIVE_LOG_N_LN,
		XP_V_SET_M_DEVSP_NAME,XP_V_UPDATE_OFFLINE, sp2dt_log,    sp2dt_mirr_log  },
 { PAN_M_DATADEV_N ,       PAN_M_DATADEV_N_LN,
		XP_V_SET_M_DEVSP_NAME,XP_V_UPDATE_OFFLINE, sp2dt_data,   sp2dt_mirr_data },
};
#define RTE_XPARAM_TABLE1_MXEO79 \
	(sizeof(RteXparamTable1)/sizeof(struct xparam_strncmp_table))

struct xparam_strcmp_table
{
	tsp11_XpKeyTypec name;		/* Name found in XParam file */

#define XP_V_BOOL_Y   0  /* Bool type (TRUE if Y or y) (for YES   */
#define XP_V_BOOL_N   1  /* Bool type (TRUE if N or n) (for NO)   */
#define XP_V_BOOL_A   2  /* Bool type (TRUE if A or a) (for AUTO) */
#define XP_V_BOOL_S   3  /* Bool type (TRUE if S or s) (for STACK) */
#define XP_V_LONG     4  /* Signed Long type */
#define XP_V_ULONG    5  /* Unsigned Long type */
#define XP_V_REGIONS  6  /* Unsigned Long Region Count */
#define XP_V_STRING   7  /* Any String type */
#define XP_V_IGNORED  8  /* Accepted, but ignored */
#define XP_V_CRYPT    9  /* Crypted Password */
#define XP_V_SPECIAL_STRING  10 /* String Parameter with side effect */
#define XP_V_SPECIAL_LONG    11 /* Long Parameter with side effect */
#define XP_V_SPECIAL_ULONG   12 /* Unsigned Long Parameter with side effect */
#define XP_V_SPECIAL_SPECIAL 13 /* Other Parameter with side effect */
	int              type;       /* Claasification of expected type */
    int              UpdateMode;/* update param allowed in off/online state of db */

    size_t           size;       /* Size in RTE struct */
	unsigned long    offset;     /* Offset in RTE struct */
};

#define XP_SZ_OFF(_x)  sizeof(((RTE_XPARAM_REC *)0)->_x), \
							((size_t)&(((RTE_XPARAM_REC *)0)->_x))

static struct xparam_strcmp_table RteXparamTable2[] =
{
 { PAN_KERNELTRACEFILE,    XP_V_SPECIAL_STRING, XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szTraceDevspace)    },
#define XP_T_SPECIAL_KERNELTRACEFILE  0
 { PAN_KERNELTRACESIZE,    XP_V_SPECIAL_LONG,   XP_V_UPDATE_ONLINE, XP_SZ_OFF(lTraceDevspaceSize) },
#define XP_T_SPECIAL_KERNELTRACESIZE  1
 { PAN_MAXDEVSPACES,       XP_V_SPECIAL_LONG,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lMaxDevspaces)      },
#define XP_T_SPECIAL_MAXDEVSPACES     2
 { PAN_MAXSYSDEVSPACES,    XP_V_SPECIAL_LONG,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lMaxSysDevspaces)   },
#define XP_T_SPECIAL_MAXSYSDEVSPACES  3
 { PAN_MAXARCHIVELOGS,     XP_V_SPECIAL_LONG,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lMaxArchiveLogs)    },
#define XP_T_SPECIAL_MAXARCHIVELOGS   4
 { PAN_PACKET_SIZE,        XP_V_SPECIAL_ULONG,  XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulPacketSize)       },
#define XP_T_SPECIAL_PACKET_SIZE      5
 { PAN_MIN_REPLY_SIZE,     XP_V_SPECIAL_ULONG,  XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMinReplySize)     },
#define XP_T_SPECIAL_MIN_REPLY_SIZE   6
 { PAN_MAXUSERSTACK,       XP_V_SPECIAL_ULONG,  XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMaxUsUtStack)     },
#define XP_T_SPECIAL_MAXUSERSTACK     7
 { PAN_TREE_REGIONS,       XP_V_SPECIAL_ULONG,  XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulNumTreeReg)       },
#define XP_T_SPECIAL_TREE_REGIONS     8
 { PAN_MP_DISP_LOOPS,      XP_V_SPECIAL_ULONG,  XP_V_UPDATE_ONLINE,  XP_SZ_OFF(ulDispLoopsBefSemop) },
#define XP_T_MP_DISP_LOOPS            9
 { PAN_LOG_MODE,           XP_V_SPECIAL_SPECIAL,XP_V_UPDATE_ONLINE, 0, 0 },
#define XP_T_LOG_MODE                 10
 { PAN_MIRR_DATA,          XP_V_SPECIAL_SPECIAL,XP_V_UPDATE_OFFLINE, 0, 0 },
#define XP_T_MIRR_DATA                11
 { PAN_MAXDATADEVSPACES,   XP_V_SPECIAL_LONG,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lMaxDataDevspaces)   },
#define XP_T_SPECIAL_MAXDATADEVSPACES 12
 { PAN_PAGE_SIZE,          XP_V_SPECIAL_SPECIAL,XP_V_UPDATE_OFFLINE, 0, 0 },
#define XP_T_SPECIAL_PAGE_SIZE        13
#if defined(_WIN32) 
 { PAN_RTE_ALLOC_POOL,     XP_V_SPECIAL_ULONG,  XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulRTEAllocPoolSize)  },
#define XP_T_RTE_ALLOC_POOL           14 
 { PAN_KERNEL_ALLOC_POOL,  XP_V_SPECIAL_ULONG,  XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulKernelAllocPoolSize) },
#define XP_T_KERNEL_ALLOC_POOL        15
#else
 { PAN_GATEWAYTYPE,        XP_V_SPECIAL_SPECIAL,XP_V_UPDATE_OFFLINE, 0, 0 },
#define XP_T_GATEWAYTYPE              14
#endif
/*
 * If there are any other XParam values ignored here, but used in kernel
 * add them here...
 */
 { PAN_SERVERDB,           XP_V_IGNORED,  XP_V_UPDATE_OFFLINE,   0, 0                         },
 { PAN_CONTROLPW,          XP_V_CRYPT,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(crypControlPassw)    },
 { PAN_UNICODE_CATALOG,    XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fUnicode)            },
 { PAN_SUBTREE,            XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fSubtree)            },
 { PAN_USE_ASYNC_IO,       XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fUseAsyncIO)         },
 { PAN_EVENTPRIO,          XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fEventPrio)          },
 { PAN_MAXBACKUPDEVS,      XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lNoOfBackupDevs)     },
 { PAN_MAXTRANS,           XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lMaxTrans)           },
 { PAN_ALLOW_LBACK_TO_PIPE,XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fLogBackToPipe)      }, /* PTS 1108249 */
 { PAN_DATA_CACHE_PAGS,    XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lDataCachePages)     },
 { PAN_SEQUENCE_CACHE,     XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lSequenceCache)      },
 { PAN_CATALOG_CACHE,      XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lCatalogCachePages)  },
 { PAN_SERVER_DESC_CACHE,  XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lDistribDescCache)   },
 { PAN_SERVER_CMD_CACHE,   XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lServerCmdCache)     },
 { PAN_MIN_SERVER_DESC,    XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lMinServerDesc)      },
 { PAN_MAXSERVERDB,        XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lMaxServerDB)        },
 { PAN_LOG_IO_QUEUE,       XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lLogQueuePages)      },
 { PAN_ROLLBACK_CACHE,     XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lLogCachePages)      },
 { PAN_MAXLOCKS,           XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lMaxLocks)           },
 { PAN_CONV_CACHE_PAGES,   XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lConvCachePages)     },
 { PAN_FBM_CACHE_PAGES,    XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lFbmCachePages)      },
 { PAN_USM_CACHE_PAGES,    XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lUsmCachePages)      },
 { PAN_BACKUP_BLOCK_CNT,   XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lBackUpBlockCount)   },
 { PAN_DYNDATA,            XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lSharedDynData)      },
 { PAN_DYNPOOL,            XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lSharedDynPool)      },
 { PAN_MAXIOTHREADS,       XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lDeviPerDevspace)    },
 { PAN_IOTHREADS_FOR_PRIO, XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lPrioDeviPerDevspace) },
 { PAN_MAXDATAPAGES,       XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lMaxDataPages)       },
 { PAN_REALTIME_PRIO,      XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lRealtimePrio)       },
 { PAN_LOG_SEGMENT_SIZE,   XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lLogSegmentSize)     },
 { PAN_LOCK_TIMEOUT,       XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lLockTimeout)        },
 { PAN_REQUEST_TIMEOUT,    XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lRequestTimeout)     },
 { PAN_NET_TIMEOUT,        XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lNetTimeout)         },
 { PAN_LOG_START_COPY,     XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lLogStartCopy)       },
 { PAN_RESERVED_REDO_SIZE, XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lReservedRedoSize)   },
 { PAN_LOG_COPY_DISTANCE,  XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lLogCopyDistance)    },
 { PAN_INDEXFILE_LIST_SIZE,XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lIndexfileListSize)  },
 { PAN_MAXAUTOBACKUPDEVS,  XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMaxAutoBackupDevs) },
 { PAN_USE_THRD_FOR_TSK,   XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fUseThreadForTask)   },  /* PTS 1108768 */
 { PAN_USE_PROC_SCOPE,     XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fUseProcessScope)    },  /* PTS 1108913 */
 { PAN_MAXACTIVE,          XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMaxUser)           },
 { PAN_MAXSERVER,          XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMaxServer)         },
 { PAN_LOCK_SUPPLY_BLOCK,  XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulLockSupplyBlock)   },
 { PAN_DIAG_SIZE,          XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulDiagSize)          },
 { PAN_UTILITY_PROTSIZE,   XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulUtilityProtSize)   },
 { PAN_MAXCPU,             XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMaxCPU)            },
 { PAN_PNOPOOLSIZE,        XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulPnoPoolSize)       },
 { PAN_MAXDATAWRITER,      XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMaxDataWriter)     },
 { PAN_MULTIBYTE_CACHE,    XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMultiByteTotalSize) },
 { PAN_MB_DATA_SIZE,       XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMBDataSize)        },
 { PAN_MB_QUAL_SIZE,       XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMBQualSize)        },
 { PAN_MB_STACK_SIZE,      XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMBStackSize)       },
 { PAN_KB_WORK_ST_SIZE,    XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulKBWorkStSize)      },
 { PAN_KB_WORKBUF_SIZE,    XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulKBWorkBufSize)     },
 { PAN_CATALOG_CACHE_SIZE, XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulCatalogCacheSize)  },
 { PAN_XPARAM_CHANGED,     XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulXparamChanged)     },
 { PAN_EVENT_SIZE,         XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulEventSize)         },
 { PAN_MAXEVENTTASKS,      XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMaxEventTasks)     },
 { PAN_MAX_BACKUP_TASKS,   XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMaxBackupTasks)    },
 { PAN_MAXGARBAGE_COLLECTOR, XP_V_ULONG,  XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMaxGarbageCollector) },
 { PAN_MAXEVENTS,          XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMaxEvents)         },
 { PAN_OBJ_LOG_SIZE,       XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulObjLogSize)        },
 { PAN_OBJ_LOG_SEGMENT_SIZE, XP_V_ULONG,  XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulObjLogSegmentSize) },
 { PAN_OBJ_LOG_IO_QUEUE,   XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulObjLogIOQueuePages) },
 { PAN_OBJ_ROLLBACK_CACHE, XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulObjLogRollbackCache) },
 { PAN_READAHEAD_BLOBS,    XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulReadAheadBlobsGT)  },
 { PAN_DATACACHE_REGIONS,  XP_V_REGIONS,  XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulNumDataCacheReg)   },
 { PAN_TRANS_REGIONS,      XP_V_REGIONS,  XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulNumTransReg)       },
 { PAN_TAB_REGIONS,        XP_V_REGIONS,  XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulNumTabReg)         },
 { PAN_ROW_REGIONS,        XP_V_REGIONS,  XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulNumRowReg)         },
 { PAN_OMS_REGIONS,        XP_V_REGIONS,  XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulNumOMSReg)         },
 { PAN_EVENT_ALIVE_CYCLE,  XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulEventAliveCycle)   },
 { PAN_PARAM_KERNELVERSION,XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szKernelVersion)     },
 { PAN_CONTROLID,          XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szControlUser)       },
 { PAN_DIAGFILE,           XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szDiagFile)          },
 { PAN_BACKUP_HISTFILE,    XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szBackupHistFile)    },
 { PAN_BACKUP_MED_DEF,     XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szBackupMedDefFile)  },
 { PAN_UTILITY_PROTFILE,   XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szUtilityProtFile)   },
 { PAN_RUNDIR,             XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szRunDirectory)      },
 { PAN_KERNELDUMPFILE,     XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szDumpFilename)      },
 { PAN_MULTIBYTE_SET_1,    XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szMultiByteSet_1)    },
 { PAN_MULTIBYTE_SET_2,    XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szMultiByteSet_2)    },
 { PAN_MULTIBYTE_SET_3,    XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szMultiByteSet_3)    },
 { PAN_MULTIBYTE_SET_4,    XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szMultiByteSet_4)    },
 { PAN_MULTIBYTE_SET_5,    XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szMultiByteSet_5)    },
 { PAN_RTEDUMPFILE,        XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szRTEDumpFilename)   },
 { PAN_EVENTFILE,          XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szEventFile)         },
 { PAN_DEFAULT_CODE,       XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szDefaultCode)       },
 { PAN_MP_RGN_QUEUE,       XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fExclQueue)          },
 { PAN_MAX_MESSAGE_FILES,  XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMaxMessageFiles)   },
/* PTS 1105389 */
 { PAN_INSTANCE_TYPE,      XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szInstanceType)      },
 { PAN_CRASH_HISTORY_NUM,  XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(uCrashHistoryNum)    },
 { PAN_CRASH_HISTORY_PATH, XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szCrashHistoryPath)  },
 { PAN_CRASH_SEM,			     XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(uCrashSem)           },
 /* PTS 1108975 */
 { PAN_TRACE_PAGES_TI,     XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(tracePagesTI)        },
 { PAN_TRACE_PAGES_GC,     XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(tracePagesGC)        },
 { PAN_TRACE_PAGES_AL,     XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(tracePagesAL)        },
 { PAN_TRACE_PAGES_DW,     XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(tracePagesDW)        },
 { PAN_TRACE_PAGES_US,     XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(tracePagesUS)        },
 { PAN_TRACE_PAGES_UT,     XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(tracePagesUT)        },
 { PAN_TRACE_PAGES_SV,     XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(tracePagesSV)        },
 { PAN_TRACE_PAGES_EV,     XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(tracePagesEV)        },
 { PAN_TRACE_PAGES_TW,     XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(tracePagesTW)        },
 { PAN_TRACE_PAGES_CS,     XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(tracePagesCS)        },
 { PAN_TRACE_PAGES_BUP,    XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(tracePagesBUP)       },

/***************************************************************/
/* OS Depending Parameter                                      */
/***************************************************************/
#if defined(_WIN32)
 { PAN_ALLOW_COMPR_DEVSP,  XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fAllowCompressedDevsp) },
 { PAN_DYNAMIC_SERVER,     XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fDynamicServerTasks)   },
 { PAN_RESTART_SHUTDOWN,   XP_V_BOOL_A,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fAutoRestartShutdown)  },
 { PAN_USE_MEM_ENHANCEMENT,XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fUseMemEnhancement)    },
 { PAN_MEM_ENHANCEMENT_LIM,XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lMemEnhancementLimit)  },
 { "_PSE36_USAGE",         XP_V_IGNORED,  XP_V_UPDATE_OFFLINE, 0, 0                             },
 { PAN_USE_FIBERS,         XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fUseFibers)            },
#else
 { PAN_DYNAMIC_TASK_STACK, XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fDynTaskStack)         },
 { PAN_TASK_STACK_AREA,    XP_V_IGNORED,  XP_V_UPDATE_OFFLINE, 0, 0                             },
 { PAN_OPMSG1,             XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szOpMsg1)              },
 { PAN_OPMSG2,             XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szOpMsg2)              },
 { PAN_SHMCHUNK,           XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lShmChunk)             },
 { PAN_SHMKERNEL,          XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lShmKernel)            },
 /* PTS 1118800 */
 { PAN_USE_OPEN_DIRECT,    XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(useOpenDirect)         },
#endif
/***************************************************************/
/* End of OS Depending Parameter                               */
/***************************************************************/
/***************************************************************/
/* ONLINE changable Parameter     using "diagnose" commands    */
/***************************************************************/
 { PAN_DELAY_LW,           XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lDelayLW)            },
 { PAN_FORMAT_DEVSPACE,    XP_V_BOOL_Y,   XP_V_UPDATE_ONLINE,  XP_SZ_OFF(fFormatDevspace)     },
 { PAN_DELAY_COMMIT,       XP_V_BOOL_Y,   XP_V_UPDATE_ONLINE,  XP_SZ_OFF(fDelayCommit)        },

 { PAN_SERVERDB_FOR_SAP,   XP_V_BOOL_Y,   XP_V_UPDATE_ONLINE,  XP_SZ_OFF(fServerDBForSAP)     },
 { PAN_DATE_TIME_FORMAT,   XP_V_STRING,   XP_V_UPDATE_ONLINE,  XP_SZ_OFF(szDateTimeFormat)    },
 { PAN_MULT_IO_BLOCK_CNT,  XP_V_ULONG,    XP_V_UPDATE_ONLINE,  XP_SZ_OFF(ulMultIOBlockCount)  },
 { PAN_RESTART_TIME,       XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lRestartTime)        },
 { PAN_DEADLOCK_DETECTION, XP_V_ULONG,    XP_V_UPDATE_ONLINE,  XP_SZ_OFF(ulDeadlockDetection) },
 { PAN_SESSION_TIMEOUT,    XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lSessionTimeout)     },
 { PAN_USE_IOTHREAD_ONLY,  XP_V_BOOL_N,   XP_V_UPDATE_ONLINE,  XP_SZ_OFF(fAllowSingleIO)      },
 { PAN_DEV_QUEUE,          XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lDevQueue)           },
 { PAN_DEFAULT_LRU,        XP_V_BOOL_Y,   XP_V_UPDATE_ONLINE,  XP_SZ_OFF(fDefaultLRU)         },
 { PAN_MP_RGN_DIRTY_READ,  XP_V_BOOL_Y,   XP_V_UPDATE_ONLINE,  XP_SZ_OFF(fBegExclDirtyRead)   },
 { PAN_MP_RGN_BUSY_WAIT,   XP_V_BOOL_Y,   XP_V_UPDATE_ONLINE,  XP_SZ_OFF(fEndExclBusyWait)    },
 { PAN_MP_DISP_PRIO,       XP_V_BOOL_Y,   XP_V_UPDATE_ONLINE,  XP_SZ_OFF(fDynDispQueSrch)     },
 { PAN_MP_RGN_LOOP,        XP_V_ULONG,    XP_V_UPDATE_ONLINE,  XP_SZ_OFF(ulRegionCollisionLoop) },
 { PAN_MP_RGN_PRIO,        XP_V_BOOL_Y,   XP_V_UPDATE_ONLINE,  XP_SZ_OFF(fBegExclQueWhenPrio) },
 { PAN_MAXRGN_REQUEST,     XP_V_ULONG,    XP_V_UPDATE_ONLINE,  XP_SZ_OFF(ulRegLockSlice)      },
 { PAN_PRIO_BASE_U2U,      XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lPrioBaseU2U)        },
 { PAN_PRIO_BASE_IOC,      XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lPrioBaseIOC)        },
 { PAN_PRIO_BASE_RAV,      XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lPrioBaseRAV)        },
 { PAN_PRIO_BASE_COM,      XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lPrioBaseCOM)        },
 { PAN_PRIO_BASE_REX,      XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lPrioBaseREX)        },
 { PAN_PRIO_FACTOR,        XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lPrioFactor)         },
 { PAN_SVP_1_CONV_FLUSH,   XP_V_BOOL_Y,   XP_V_UPDATE_ONLINE,  XP_SZ_OFF(fSvpFstConvFlush)    },
 { PAN_DW_COPY_BOUNDARY,   XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lDwCopyBoundary)     },
 { PAN_DW_IO_AREA_SIZE,    XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lDwIOAreaSize)       },
 { PAN_DW_IO_AREA_FLUSH,   XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lDwIOAreaFlush)      },
 { PAN_FBM_DEVSP_COMPR,    XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lFbmCompression)     },
 { PAN_FBM_DEVSP_BALANCE,  XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lFbmBalance)         },
 { PAN_FBM_LOW_IO_RATE,    XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lFbmLowIORate)       },
 { PAN_COPY_IO_BLOCK_CNT,  XP_V_ULONG,    XP_V_UPDATE_ONLINE,  XP_SZ_OFF(ulDwIOBufSize)       },
 { PAN_DW_LRU_TAIL_FLUSH,  XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lDwLRUTailFlush)     },
 { PAN_JOIN_SEARCH_LEVEL,  XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lJoinSearchLevel)    },
 { PAN_JOIN_MAXTAB_LEVEL4, XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lJoinMaxTabLevel4)   },
 { PAN_JOIN_MAXTAB_LEVEL9, XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lJoinMaxTabLevel9)   },
 { PAN_OPTIM_OR_DISTINCT,  XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lDistinctOverhead)   },
 { PAN_OPTIM_ORDERBY_IDX,  XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lOrderbyFactor)      },
 { PAN_OPTIM_BUILD_RESLT,  XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lBuildResultPercent) },
 { PAN_OPTIM_FETCH_RESLT,  XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lFetchResultPercent) },
 { PAN_OPTIM_MAX_MERGE,    XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lMaxMerge)           },
 { PAN_OPTIM_JOIN_FETCH  , XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lJoinPrefetchPercent)},
 { PAN_OPTIM_KEY_INV_RATE, XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lKeyInvRatePercent)  },
 { PAN_AK_DUMP_ALLOWED,    XP_V_BOOL_Y,   XP_V_UPDATE_ONLINE,  XP_SZ_OFF(fAKDumpAllowed)     },
 { PAN_OMS_HEAP_LIMIT  ,   XP_V_ULONG,    XP_V_UPDATE_ONLINE,  XP_SZ_OFF(ulOMSHeapLimit) },
 /* PTS 1111787 */
 { PAN_COMMENT,            XP_V_IGNORED,  XP_V_UPDATE_OFFLINE,   0, 0                         },
 /* PTS 1113403 */
 { PAN_SUPPRESS_CORE,      XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(suppressCore) },
 /* PTS 1116153 */
 { PAN_OFFICIAL_NODE,      XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szOfficialNode) },
};
#define RTE_XPARAM_TABLE2_MXEO79 \
	(sizeof(RteXparamTable2)/sizeof(struct xparam_strcmp_table))

#define XP_PTR(_x,_t) (((char *)_x)+_t->offset)
/*===========================================================================*
 *  GLOBAL VARIABLES                                                         *
 *===========================================================================*/

extern BOOLEAN                                 fGotControlPw ;

/*===========================================================================*
 *  LOCAL VARIABLES                                                          *
 *===========================================================================*/

static char * szDevType[ NUM_DEVSP_TYPES ] = { "SYS",  "MSYS",
                                               "ALOG" , "MALOG",
                                               "TLOG",
                                               "DATA", "MDATA",
                                               "TRACE" } ;


/*===========================================================================*
 *  LOCAL FUNCTIONS (PROTOTYPES)                                             *
 *===========================================================================*/

static int sql79k_find_param ( tsp11_XpKeyTypec                 name ,
                               XP_VALUE_REC                    *xpr );

void vabort           (BOOLEAN write_core) ;

static LONG  sql79k_xpr_int          ( XP_VALUE_REC  *xpr ) ;

static void  sql79k_xpr_scopy        ( char         *s ,
                                       XP_VALUE_REC * xpr  ) ;

static void  sql79k_alloc_devsp_descr ( RTE_XPARAM_REC            *XParam,
                                       LONG                      lDevspType,
                                       ULONG                     ulNumDevsp);

static int    sql79k_set_devsp_name  ( RTE_XPARAM_REC            *XParam,
                                       tsp11_XpKeyTypec           DevspName ,
                                       XP_VALUE_REC              *xpr ,
                                       LONG                      lDevType ,
                                       ULONG                     ulDevspNo );

static int    sql79k_set_devsp_size  ( RTE_XPARAM_REC            *XParam,
                                       LONG                      lDevType ,
                                       ULONG                     ulDevspNo ,
                                       ULONG                     ulDevspSize );

static void   sql79k_set_taskcluster ( RTE_XPARAM_REC            *XParam,
                                       char                      *szTaskCluster,
                                       ULONG                     ulTCNo );


static int    sql79k_save_param_version ( char               *param_dir,
                                          int                VersionToSave ,
                                          tsp_errtext        errtext) ;

static void   sql79k_init_param         ( RTE_XPARAM_REC        *XParam  );

static void   sql79k_analyze_param      ( RTE_XPARAM_REC        *XParam,
                                          tsp11_XpKeyTypec       name,
                                          XP_VALUE_REC          *xpr ) ;

static void   eo79OpenXparamBinary      ( tsp00_DbNamec      ServerDB,
                                          BOOLEAN            ReadOnly,
                                          XP_PHFILE          xp_fd ,
                                          tsp00_ErrText      errtext ,
                                          tsp_xp_return_code *xp_ret ) ;

/*===========================================================================*
 *  GLOBAL FUNCTIONS (CODE)                                                  *
 *===========================================================================*/

void    vxparam_save_good_config ( BOOLEAN             is_restart,        
                                   tsp_errtext         errtext ,
                                   tsp_xp_return_code *xp_ret )
{
  #undef  MF__
  #define MF__ MOD__"vxparam_save_good_config"


  DBGIN;
#ifdef VERSION_TEST
{
      int test_copy_param();
/*      vxparam_save_good_config(1, errtext, xp_ret ) ; */
      test_copy_param();
}
  vxparam_rec_put( "SCHNUDEL", "DUDEL", 0 ,errtext, xp_ret, errtext );
  vxparam_rec_put( "_PRIO_BASE_COM", "33", 1 ,errtext, xp_ret, errtext );
  vxparam_rec_put( "_TREE_RGNS", "10", 1 ,errtext, xp_ret, errtext );
  vxparam_rec_put( "_PRIO_BASE_COM", "10", 0 ,errtext, xp_ret, errtext );
  vxparam_rec_put( "_PRIO_BASE_COM", "10", 1 ,errtext, xp_ret, errtext );
#endif

  *xp_ret = XP_OK  ;

  if ( !is_restart || XPARAM(ulXparamChanged) )
  {
    tsp01_RteError RteError ;
    eo46_rte_error_init ( &RteError ) ;

    if ( !is_restart  ) /* !!! config changed (add devspace) !!! */
      sqlXParamRemoveAllVersions( KERNEL_GLOBAL_DBNAME, &RteError ) ;

    if ( RteError.RteErrCode == RTE_NO_ERROR_SP01 )
    {
      /* PTS 1103607: 1. Reset Changedflag / 2. save param as version*/
      eo11ResetXparamChangedFlag ( KERNEL_GLOBAL_DBNAME, &RteError ) ;

      if ( RteError.RteErrCode == RTE_NO_ERROR_SP01 )
        eo11SaveActParamFileAsVersion ( KERNEL_GLOBAL_DBNAME, &RteError ) ;

      if ( (RteError.RteErrCode == RTE_NO_ERROR_SP01) ) 
        XPARAM(ulXparamChanged) = 0;
      else
      {
        MSGCD (( ERR_XPARAM_VERSIONING, RteError.OsErrText ));
        memcpy(errtext, RteError.OsErrText, sizeof( tsp_errtext) ) ;
        *xp_ret = !XP_OK ;
      }
    }
  }
  DBGOUT;
  return ;
}

/*
 * ===========================================================================
 */

void    vxparam_open_binary ( tsp00_DbName       ServerDB,
                              BOOLEAN            ReadOnly,
                              XP_PHFILE          xp_fd ,
                              tsp00_ErrText      errtext ,
                              tsp_xp_return_code *xp_ret )
{
  #undef  MF__
  #define MF__ MOD__"vxparam_open_binary"

  tsp00_DbNamec     ServerDBc ;

  DBGIN;

  eo46PtoC ( ServerDBc, ServerDB, sizeof (tsp00_DbName) );

  eo79OpenXparamBinary (ServerDBc, ReadOnly, xp_fd, errtext,xp_ret) ;

  DBGOUT;
  return ;
}

/*
 * ===========================================================================
 */

void    vxparam_open ( XP_PHFILE            xp_fd ,
                       tsp00_ErrText        errtext ,
                       tsp_xp_return_code  *xp_ret )
{
  #undef  MF__
  #define MF__ MOD__"vxparam_open"

  DBGIN;
  eo79OpenXparamBinary ( KERNEL_GLOBAL_DBNAME, TRUE /* ReadOnly */,
                         xp_fd, errtext, xp_ret ) ;

  DBGOUT;
  return ;
}

/*
 * ===========================================================================
 */

void    vxparam_close ( XP_HFILE            xp_fd ,
                        tsp_errtext         errtext ,
                        tsp_xp_return_code *xp_ret )
{
  #undef  MF__
  #define MF__ MOD__"vxparam_close"

  tsp01_RteError    RteError ;
  DBGIN;

  eo46_rte_error_init ( &RteError ) ;

  sqlXparamClose ( xp_fd, XP_CLOSE_NO_FLUSH_SP11, &RteError );
  if ( RteError.RteErrCode != RTE_NO_ERROR_SP01 )
    {
      *xp_ret = !XP_OK ;
      eo46CtoP ( errtext, RteError.RteErrText, sizeof (tsp00_ErrText) );
    }
  else
    *xp_ret = XP_OK ;

  DBGOUT;
  return ;
}

/*
 * ===========================================================================
 */

void    vxparam_read_binary ( XP_HFILE            xp_fd ,
                              char               *buf ,
                              tsp_int4            bytes_to_read ,
                              tsp_int4           *bytes_read ,
                              tsp_errtext         errtext ,
                              tsp_xp_return_code *xp_ret )
{
  #undef  MF__
  #define MF__ MOD__"vxparam_read_binary"

  DBGIN;

  sql11c_SeqBinReadXParam ( xp_fd , (char *)buf, bytes_to_read ,
		           bytes_read, errtext , xp_ret ) ;

  DBGOUT;
  return ;
}

/*
 * ===========================================================================
 */

void    vxparam_write_binary ( XP_HFILE            xp_fd ,
                               char               *buf ,
                               tsp_int4            bytes_to_write ,
                               tsp_int4           *bytes_written ,
                               tsp_errtext         errtext ,
                               tsp_xp_return_code *xp_ret )
{
  #undef  MF__
  #define MF__ MOD__"vxparam_write_binary"

  DBGIN;

  sql11c_SeqBinWriteXParam ( xp_fd , (char *)buf, bytes_to_write ,
		                     bytes_written, errtext , xp_ret ) ;

  DBGOUT;
  return ;
}

/*
 * ===========================================================================
 */
void    vxparam_set_new_serverdb( tsp00_DbName       ServerDB,
                                  tsp00_ErrText      errtext ,
                                  tsp_xp_return_code *xp_ret )
{
  #undef  MF__
  #define MF__ MOD__"vxparam_new_set_new_serverdb"

  XP_VALUE_REC     XpValue ;
  tsp01_RteError   RteError;
  tsp00_DbNamec    DBNameC ;

  DBGIN;

  eo46_rte_error_init ( &RteError ) ;
  eo46PtoC ( DBNameC, ServerDB, sizeof (tsp00_DbName) );
/*
  strupr(DBNameC) ;
  eo46CtoP ( ServerDB, DBNameC, sizeof (tsp00_DbName) );
*/
  XpValue.xp_value_t = xp_c18_type_sp11 ;
  memcpy( (char *)&XpValue.xp_type, ServerDB, sizeof(tsp00_DbName) ) ;
  eo11PutParam ( DBNameC, PAN_SERVERDB, (tsp00_XpValueRec *)&XpValue, &RteError ) ;
  *xp_ret = RteError.RteErrCode == RTE_ERROR_SP01 ? !XP_OK : XP_OK ;

  DBGOUT;
  return ;
}

/*
 * ===========================================================================
 */

void    vxparam_get_filesize ( tsp_int4           *xp_fsize ,
                               tsp_errtext         errtext ,
                               tsp_xp_return_code *xp_ret )
{
  #undef  MF__
  #define MF__ MOD__"vxparam_get_filesize"
  tsp00_Bool       fOk ;
  tsp01_RteError   RteError;

  DBGIN;

  eo46_rte_error_init ( &RteError ) ;
  fOk = sqlXparamFsize ( KERNEL_GLOBAL_DBNAME, xp_fsize, &RteError ) ;

  if ( !fOk )
  {
    *xp_ret = !XP_OK ;
    eo46CtoP ( errtext, RteError.OsErrText, sizeof (tsp00_ErrText) );
  }
  else
    *xp_ret = XP_OK ;

  DBGOUT;
  return ;
}

/*
 * ===========================================================================
 */

void    vxparam_get_next( XP_PHFILE           xp_fd ,
                          XP_KEY_TYPE         xp_key ,
                          tsp_c64             value ,
                          tsp_errtext         errtext ,
                          tsp_xp_return_code *xp_ret )
{
  #undef  MF__
  #define MF__ MOD__"vxparam_get_next"

  BOOLEAN         fBadParameter ;
  XP_VALUE_REC    xp_value ;

  DBGIN;

  *xp_ret = XP_OK ;

  if ( *xp_fd < 0 )
    { 
      tsp01_RteError  RteError ;
      eo46_rte_error_init ( &RteError ) ;

      sqlXparamOpen (KERNEL_GLOBAL_DBNAME, xp_fd, XP_OPEN_READONLY_SP11, &RteError) ;

      if ( RteError.RteErrCode != RTE_NO_ERROR_SP01 )
        { *xp_ret = !XP_OK ;
          *xp_fd = -1 ;
          eo46CtoP ( errtext, RteError.RteErrText, sizeof (tsp00_ErrText) );
          return ;
        }
      else     /* read header page */
        sql11c_SeqNextXParam ( *xp_fd , xp_key , &xp_value , errtext , xp_ret );
    }

  do
    { char szDummy[80] ;
      fBadParameter = FALSE ;
      sql11c_SeqNextXParam ( *xp_fd , xp_key , &xp_value , errtext , xp_ret ) ;
      if ( *xp_ret == XP_OK )
	  {
          tsp11_XpKeyTypec    acXParamKeyC ;
          eo46PtoC ( acXParamKeyC, xp_key , sizeof (tsp11_XpKeyType) );
	      if ( sql79k_find_param ( acXParamKeyC , &xp_value ) < 0 )
		  {
              fBadParameter = TRUE ;
		  }
          else
		  {
              if ( xp_value.xp_value_t >= XP_INT2_TYPE &&
                   xp_value.xp_value_t <= XP_INT4_TYPE )
			  {
                  sprintf ( szDummy, "%12d", sql79k_xpr_int (&xp_value)) ;
			  }
              else if ( xp_value.xp_value_t >= XP_C8_TYPE &&
                       xp_value.xp_value_t <= XP_C64_TYPE )
			  {
                  sql79k_xpr_scopy ( szDummy, &xp_value) ;
			  }
              else if ( xp_value.xp_value_t != XP_CRYPT_TYPE )
			  {
                  MSGCD (( ERR_ILLEGAL_XP_TYPE, xp_value.xp_value_t ));
                  fBadParameter = TRUE ;
			  }

              if ( !fBadParameter )
              {
				   eo46CtoP ( value, szDummy, sizeof( tsp_c64 ) );
                   DBG3 ((MF__, "xp_key = '%.18s'" , xp_key  ));
                   DBG3 ((MF__, "xp_val = '%s'" , szDummy ));
              }
          }
	  }
      else
      {
          tsp01_RteError   IgnoreCloseError ;
          if ( *xp_ret != XP_EOF && *xp_ret != XP_KEY_NOT_FOUND )
  	  		  MSGCD (( ERR_READING_PARAMFILE , errtext )) ;

          DBG3 ((MF__, " paramfile closed'" ));
          sqlXparamClose ( *xp_fd, XP_CLOSE_NO_FLUSH_SP11, &IgnoreCloseError );
          *xp_fd = -1 ;
      }
    }
  while ( *xp_ret == XP_OK && fBadParameter ) ;

  DBGOUT;
  return ;
}

/*
 * ===========================================================================
 */

/* PTS 1103828 */ 
void vxparam_rec_put ( XP_KEY_TYPE         xp_key ,
                       tsp00_C64           value ,
                       BOOLEAN             IsNumeric,
                       tsp00_C64           Oldvalue,
                       tsp00_ErrText       errtext,
                       tsp_xp_return_code *xp_ret)
{
  int which ;
  tsp11_XpKeyTypec name;
  tsp00_C64c                  valueC ;
  tsp00_C64c                  OldvalueC ;
  struct xparam_strcmp_table *pE;
  
  eo46PtoC ( name, xp_key , sizeof (tsp11_XpKeyType) );
  eo46PtoC ( valueC, value , sizeof (tsp_c64) );

  if ( (which = sql79k_find_param(name, NULL) ) < 0 )
  {
    *xp_ret = XP_KEY_NOT_FOUND ;
    eo46CtoP ( errtext, ERRMSG_KEY_NOT_FOUND, sizeof (tsp00_ErrText) );
    MSGCD (( ERR_XP_UPDATE_UNKNOWN_KEY, name ));
  }
  else
  {
	which -= RTE_XPARAM_TABLE1_MXEO79;
    if ( which >= 0 )
    {
      pE = &RteXparamTable2[which];
      *xp_ret = pE->UpdateMode == XP_V_UPDATE_ONLINE ? XP_OK : XP_NOT_OK ;
    }
    else
      *xp_ret = XP_NOT_OK ; /* RteXparamTable1 contains unchangeable parameters only */

    if ( *xp_ret != XP_OK )
    {
      eo46CtoP ( errtext, ERRMSG_PARAM_UPDATE_NOT_ALLOWED, sizeof (tsp00_ErrText) );
      MSGCD (( ERR_XP_UPDATE_ACCESS_DENIED, name ));
    }
  }

  if ( *xp_ret == XP_OK )
  {
    RTE_XPARAM_REC             *XParam = KGS->XParam ;

	/*
	 * Reduced to offset into RteXparamTable2
	 */

    memset ( OldvalueC, 0, sizeof(OldvalueC)) ;

    switch( pE->type ) 
	{
		case XP_V_BOOL_Y:
            OldvalueC[0] = *(tsp00_Bool *)XP_PTR(XParam, pE) ? 'Y' : 'N' ;
			*(tsp00_Bool *)XP_PTR(XParam, pE) = 
					(tsp00_Bool)( *valueC == 'y' || *valueC == 'Y' );
			break;

		case XP_V_BOOL_N:
            OldvalueC[0] = *(tsp00_Bool *)XP_PTR(XParam, pE) ? 'N' : 'Y' ;
			*(tsp00_Bool *)XP_PTR(XParam, pE) = 
					(tsp00_Bool)( *valueC == 'n' || *valueC == 'N' );
			break;

		case XP_V_BOOL_A:
            OldvalueC[0] = *(tsp00_Bool *)XP_PTR(XParam, pE) ? 'A' : 'M' ;
			*(tsp00_Bool *)XP_PTR(XParam, pE) = 
					(tsp00_Bool)( *valueC == 'a' || *valueC == 'A' );
			break;

		case XP_V_BOOL_S:
            OldvalueC[0] = *(tsp00_Bool *)XP_PTR(XParam, pE) ? 'S' : 'D' ;
			*(tsp00_Bool *)XP_PTR(XParam, pE) = 
					(tsp00_Bool)( *valueC == 's' || *valueC == 'S' );
			break;

		case XP_V_LONG:
            if ( IsNumeric )
            {
              sql47_itoa( *(tsp00_Int *)XP_PTR(XParam, pE), OldvalueC, sizeof(OldvalueC) );
              *(tsp00_Int *)XP_PTR(XParam, pE) = atoi( valueC );
            }
            else
              *xp_ret = XP_NOT_OK ; 
			break;

		case XP_V_ULONG:
            if ( IsNumeric )
            {
              sql47_itoa( *(tsp00_Uint *)XP_PTR(XParam, pE), OldvalueC, sizeof(OldvalueC) );
  	          *(tsp00_Uint *)XP_PTR(XParam, pE) = atoi( valueC );
            }
            else
              *xp_ret = XP_NOT_OK ; 
			break;

		case XP_V_STRING:
			memcpy ( OldvalueC, XP_PTR(XParam, pE), pE->size);
			memcpy ( XP_PTR(XParam, pE), valueC, pE->size);
			break;

		case XP_V_SPECIAL_LONG:
            if ( IsNumeric )
            {
              sql47_itoa( *(tsp00_Int *)XP_PTR(XParam, pE), OldvalueC, sizeof(OldvalueC) );
     	      *(tsp00_Int *)XP_PTR(XParam, pE) = atoi( valueC );
            }
            else
              *xp_ret = XP_NOT_OK ; 
			break;

		case XP_V_SPECIAL_ULONG:
            if ( IsNumeric )
            {
              sql47_itoa( *(tsp00_Int *)XP_PTR(XParam, pE), OldvalueC, sizeof(OldvalueC) );
    	      *(tsp00_Uint *)XP_PTR(XParam, pE) = atoi( valueC );
			  switch( which ) 
              { 
			    case XP_T_MP_DISP_LOOPS:
				  if ( XParam->ulDispLoopsBefSemop < 2 )
				    	XParam->ulDispLoopsBefSemop = 2;
 				   break;
      			default:
                   *xp_ret = XP_NOT_OK ; 
	    			DBG3 ((MF__, "RteXparamTable no special entry %d for %s\n",
		    			which, name ));
			    	break;
              }
			}
            else
              *xp_ret = XP_NOT_OK ; 
			break;
		default:
            *xp_ret = XP_NOT_OK ; 
			DBG3 ((MF__, "RteXparamTable entry %d for %s wrong type %d\n",
					which, name, pE->type ));
			break;
		}

    if ( *xp_ret == XP_OK )
    {
      eo46CtoP ( Oldvalue, OldvalueC , sizeof (tsp_c64) );
      MSGCD (( INFO_DYNAMIC_XP_KEY_CHANGED, name, OldvalueC, valueC ));
    }
    else
    {
      eo46CtoP ( errtext, ERRMSG_BAD_FORMAT, sizeof (tsp00_ErrText) );
      MSGCD (( ERR_XP_UPDATE_BAD_FORMAT, valueC, name ));
    }
  }

}

extern  void   *vxparam_GetStringFromXParam (tsp11_XpKeyTypec   xparam_id,
                                             RTE_XPARAM_REC    *p_xparam)
{
/*===========================================================================*
 *  Locals                                                                   *
 *===========================================================================*/
  tsp00_Int4                  ix;
  struct xparam_strcmp_table *pTable;


/*===========================================================================*
 *  Instructions                                                             *
 *===========================================================================*/
  ix = sql79k_find_param (xparam_id, NULL);
  if (ix < RTE_XPARAM_TABLE1_MXEO79)
  {
    MSGD ((ERR_UNKNOWN_XP_PARAMETER, xparam_id));
    return (NULL);
  }

  ix -= RTE_XPARAM_TABLE1_MXEO79;
    
  pTable = &RteXparamTable2[ix];

  if (pTable->type != XP_V_STRING && pTable->type != XP_V_SPECIAL_STRING)
  {
    MSGD ((ERR_ILLEGAL_XP_TYPE, pTable->type));
    return (NULL);
  }

  return (XP_PTR(p_xparam, pTable));
}


/*===========================================================================*
 *  LOCAL FUNCTIONS (CODE)                                                   *
 *===========================================================================*/
static void    
eo79OpenXparamBinary ( tsp00_DbNamec      ServerDB,
                       BOOLEAN            ReadOnly,
                       XP_PHFILE          xp_fd ,
                       tsp00_ErrText      errtext ,
                       tsp_xp_return_code *xp_ret )
{
  #undef  MF__
  #define MF__ MOD__"eo79OpenXparamBinary"

  tsp01_RteError    RteError ;
  tsp00_Int4        DefaultMode = XP_OPEN_BIN_SP11 ;

  DBGIN;

  eo46_rte_error_init ( &RteError ) ;

  if ( ReadOnly ) 
    DefaultMode |= XP_OPEN_READONLY_SP11 ;
  else
    DefaultMode |= XP_OPEN_CREATE_SP11 ;

  *xp_ret = XP_OK ;

  sqlXparamOpen (ServerDB, xp_fd, DefaultMode, &RteError) ;

  if ( RteError.RteErrCode != RTE_NO_ERROR_SP01 )
    {
      *xp_ret = !XP_OK ;
      *xp_fd = -1 ;
      eo46CtoP ( errtext, RteError.RteErrText, sizeof (tsp00_ErrText) );
    }

  DBGOUT;
  return ;
}

/*
 * ===========================================================================
 */

static void sql79k_init_param ( RTE_XPARAM_REC                  *XParam  )
{
  #undef  MF__
  #define MF__ MOD__"sql79k_init_param"
  DBGIN;

  XParam->ulRegLockSlice        = MAX_EXCLUSIVE;
  XParam->ulRegionCollisionLoop = 0 ;
  XParam->lDelayLW              = 0 ;
  XParam->fSvpFstConvFlush      = FALSE;
  XParam->fDelayCommit          = FALSE ;
  XParam->fAllowSingleIO        = FALSE ;
  XParam->ulDispLoopsBefSemop   = MIN_DISP_LOOPS;
  XParam->fEndExclBusyWait      = FALSE;
  XParam->fBegExclDirtyRead     = FALSE;
  XParam->fDynDispQueSrch       = FALSE;
  XParam->fBegExclQueWhenPrio   = FALSE;
  XParam->fExclQueue            = TRUE;
  XParam->lPrioDeviPerDevspace  = 0;
  XParam->ulNumDataCacheReg     = 1;
  XParam->ulNumTreeReg          = 1;
  XParam->ulNumTransReg         = 1;
  XParam->ulNumTabReg           = 1;
  XParam->ulNumRowReg           = 1;
  XParam->ulNumOMSReg           = 8;
  XParam->lSequenceCache        = 1;
  XParam->lPrioBaseU2U          = 100;
  XParam->lPrioBaseRAV          = 80;
  XParam->lPrioBaseIOC          = 80;
  XParam->lPrioBaseCOM          = 10;
  XParam->lPrioBaseREX          = 40;
  XParam->lPrioFactor           = 80;
  XParam->lFbmCachePages        = 0;
  XParam->lFbmCompression       = 50;
  XParam->lFbmBalance           = 10;
  XParam->lFbmLowIORate         = 10;
  XParam->fDefaultLRU           = TRUE;
  XParam->ulMaxCPU              = 1;
  XParam->lJoinSearchLevel      = 0;
  XParam->lJoinMaxTabLevel4     = 16;
  XParam->lJoinMaxTabLevel9     = 5;
  XParam->lDistinctOverhead     = 5;
  XParam->lOrderbyFactor        = 2;
  XParam->lBuildResultPercent   = 15;
  XParam->lFetchResultPercent   = 50;
  XParam->lJoinPrefetchPercent  = 0;
  XParam->lJoinPrefetchPercent  = 0;
  XParam->lKeyInvRatePercent    = 0;
  XParam->lLogStartCopy         = 0;
  XParam->lReservedRedoSize     = 0;
  XParam->lLogCopyDistance      = 0;

  XParam->lDwCopyBoundary       = 50;
  XParam->lDwIOAreaSize         = 20;
  XParam->lDwIOAreaFlush        = 50;
  XParam->lDwLRUTailFlush       = 5;
  XParam->ulMaxDataWriter       = 0;
  XParam->ulDwIOBufSize         = 1;
  XParam->ulPnoPoolSize         = 10000;
  XParam->ulMultIOBlockCount    = 8;
  XParam->lBackUpBlockCount     = 8;

  XParam->ulMBDataSize          = 4096 ;
  XParam->ulMBQualSize          = 4096 ;
  XParam->ulMBStackSize         = 4 * 4096 ;
  XParam->ulKBWorkStSize        = 8 * 1024 ;
  XParam->ulKBWorkBufSize       = 4096 ;
  XParam->ulCatalogCacheSize    = 64 * 1024 ;
  XParam->ulLockSupplyBlock     = 100 ;
  XParam->fFormatDevspace       = TRUE;
  XParam->ulXparamChanged       = 0;

  XParam->ulMaxEvents           = 100;
  XParam->ulMaxMessageFiles     = 5;
  XParam->ulOMSHeapLimit        = 0;

/***************************************************************/
#if defined(_WIN32)
/***************************************************************/
  XParam->fUseMemEnhancement    = FALSE;
  XParam->lMemEnhancementLimit  = 0;

  XParam->ulRTEAllocPoolSize    = RTE_UKT_ALLOC_POOL_SIZE;
  XParam->fAllowCompressedDevsp = FALSE;
  XParam->fAutoRestartShutdown  = FALSE;
#if defined ( DYNAMIC_SERVER )
   XParam->fDynamicServerTasks  = TRUE;
#else
   XParam->fDynamicServerTasks  = FALSE;
#endif
#if defined ( _WIN64 )
   XParam->fUseFibers           = TRUE;
#else
   XParam->fUseFibers           = FALSE;
#endif

  XParam->fUseAsyncIO          = 1;
  XParam->lMaxServerDB         = 1;

/***************************************************************/
#else
/***************************************************************/
#ifdef OSF1
  XParam->fUseAsyncIO           = 1 ;
#else
  XParam->fUseAsyncIO           = 0 ;
#endif
  XParam->ulMaxUsUtStack        = 0;
  XParam->fDynTaskStack         = FALSE;
/***************************************************************/
#endif
/***************************************************************/

/* PTS 1106437 */
#if defined(LINUX) || defined(NMP) || defined(_WIN32)
  XParam->fStackStack           = TRUE;
#else
  XParam->fStackStack           = FALSE;
#endif

  XParam->ulDBPageSize          = sizeof ( tsp00_Page ) ;
  XParam->ulNoOfRegions         = MAX_REGIONS ;

  XParam->ulObjLogSize          = 0;
  XParam->ulMaxBackupTasks      = 0;
  XParam->ulMaxEventTasks       = 0;
  XParam->ulMaxGarbageCollector = 0;
  XParam->ulMaxObjectLogWriter  = 0;
  XParam->ulReadAheadBlobsGT    = 25;

  XParam->ulEventAliveCycle     = 0;
  XParam->lIndexfileListSize    = 0;

  DBGOUT ;
  return;
}


/*
 * Search parameter in table and return index if found. -1 if not found
 * The returned index can be used as index into RteXparamTable1.
 */
  #undef  MF__
  #define MF__ MOD__"sql79k_find_param"
static int sql79k_find_param ( tsp11_XpKeyTypec                 name ,
                               XP_VALUE_REC                    *xpr )
{
int which;

	/*
     * First entry is special! All other are compared with strcmp()...
     */
	for ( which = 0; which < RTE_XPARAM_TABLE1_MXEO79; which ++ )
	{
		struct xparam_strncmp_table *pN;

		pN = &RteXparamTable1[which];
		if ( !strncmp(name, pN->name, pN->len) )
		{
          if ( xpr )
			switch( pN->type ) 
            {
			case XP_V_SET_TASKCLUSTER:
			case XP_V_SET_DEVSP_NAME:
			case XP_V_SET_M_DEVSP_NAME:
				switch ( xpr->xp_value_t )
				{
				case XP_C8_TYPE:
				case XP_C18_TYPE:
				case XP_C24_TYPE:
				case XP_C40_TYPE:
				case XP_C64_TYPE:
					break;
				default:
					MSGCD (( ERR_ILLEGAL_XP_TYPE, xpr->xp_value_t ));
					which = -1;
					break;
				}
				break;

			case XP_V_SET_DEVSP_SIZE:
  				switch ( xpr->xp_value_t )
				{
				case XP_INT2_TYPE:
    			case XP_INT4_TYPE :
					break;
				default:
      				MSGCD (( ERR_ILLEGAL_XP_TYPE, xpr->xp_value_t ));
					which = -1;
					break;
				}
				break;
			}
			return which;
		}
	}

    for ( which = 0; which < RTE_XPARAM_TABLE2_MXEO79; which++ )
	{
		struct xparam_strcmp_table *pE;
		pE = &RteXparamTable2[which];

        if ( !strncmp( name, pE->name, strlen(pE->name) ) )
		{
          if ( xpr )
			switch( pE->type ) 
			{
			case XP_V_LONG:
			case XP_V_ULONG:
			case XP_V_REGIONS:
			case XP_V_SPECIAL_LONG:
			case XP_V_SPECIAL_ULONG:
  				switch ( xpr->xp_value_t )
				{
				case XP_INT2_TYPE:
    			case XP_INT4_TYPE :
					break;
				default:
      				MSGCD (( ERR_ILLEGAL_XP_TYPE, xpr->xp_value_t ));
					which = -1;
					break;
				}
				break;

			case XP_V_BOOL_Y:
			case XP_V_BOOL_N:
			case XP_V_BOOL_A:
			case XP_V_BOOL_S:
			case XP_V_STRING:
			case XP_V_SPECIAL_STRING:
  				switch ( xpr->xp_value_t )
				{
  				case XP_C8_TYPE:
  				case XP_C18_TYPE:
  				case XP_C24_TYPE:
  				case XP_C40_TYPE:
  				case XP_C64_TYPE:
					break;
				default:
    				MSGCD (( ERR_ILLEGAL_XP_TYPE, xpr->xp_value_t ));
					which = -1;
					break;
				}
				break;

			case XP_V_IGNORED:
			case XP_V_SPECIAL_SPECIAL:
				break;

			case XP_V_CRYPT:
  				switch ( xpr->xp_value_t )
				{
  				case XP_CRYPT_TYPE:
					break;
				default:
    				MSGCD (( ERR_ILLEGAL_XP_TYPE, xpr->xp_value_t ));
					which = -1;
					break;
				}
				break;

			default:
				DBG3 ((MF__, "RteXparamTable entry %d for %s wrong type %d\n",
					which, pE->name, pE->type ));
				which = -1;
				break;
			}
			break;
		}
	}

	if ( which == RTE_XPARAM_TABLE2_MXEO79 )
	{
		which = -1;
	}

	/*
	 * Add number of elements in first table (subtracted in analyze_param...
	 */
	if ( which >= 0 )
	{
		which += RTE_XPARAM_TABLE1_MXEO79;
	}

	return which;
}


/*
 * ===========================================================================
 */
int      eo79ReadAndAnalyzeParam ( tsp00_DbNamec     ServerDB,
                                   RTE_XPARAM_REC   *XParam,
                                   tsp01_RteError   *RteError )
{
  #undef  MF__
  #define MF__ MOD__"eo79ReadAndAnalyzeParam"
  tsp00_Int4                      xp_fd;
  tsp11_XpKeyTypec                xp_key;
  tsp11_XpKeyTypec                xp_next_key;
  tsp00_XpValueRec                xp_value;
  tsp11_XpReturnCode              xp_ret;

  memset (XParam, 0, sizeof(RTE_XPARAM_REC) );
  eo46_rte_error_init ( RteError ) ;

  /* --- Set XPARAM default values */
  sql79k_init_param ( XParam );

  xp_ret = sqlXparamOpen ( ServerDB, &xp_fd, XP_OPEN_READONLY_SP11, RteError) ;

  if ( xp_ret == XP_OK )
  {
    *xp_key = '\0';
    xp_ret = sqlXparamGetNext( xp_fd, xp_key, xp_next_key, 
                               &xp_value, RteError);

    while ( xp_ret == XP_OK )
    {
      DBG3 ((MF__, "xp_next_key = %s\n" , xp_next_key ));

      sql79k_analyze_param ( XParam , xp_next_key , (XP_PVALUE_REC )&xp_value );
      strcpy (xp_key , xp_next_key );

      xp_ret = sqlXparamGetNext( xp_fd, xp_key, xp_next_key, 
                                &xp_value, RteError);
    }
 
    sqlXparamClose ( xp_fd, XP_CLOSE_NO_FLUSH_SP11, RteError );
  }
  return ( xp_ret == XP_EOF );
}

/*
 * ===========================================================================
 */
static void     
sql79k_analyze_param ( RTE_XPARAM_REC                  *XParam  ,
                       tsp11_XpKeyTypec                 name ,
                       XP_VALUE_REC                    *xpr )
{
  #undef  MF__
  #define MF__ MOD__"sql79k_analyze_param"

int which;
char szDummy[80] ;
struct xparam_strcmp_table *pE;
tsp00_Uint ulDummy;

  DBGPAS;

	which = sql79k_find_param(name, xpr);
	if ( which < 0 )
	{
/*IST on request of Juergen Primsch 1993-01-05: suppress this warning */
		DBG3 ((MF__, "Unknown parameter in xparam : '%s'", name ));
		return;
	}

	if ( which < RTE_XPARAM_TABLE1_MXEO79 )
	{
		struct xparam_strncmp_table *pN;

		pN = &RteXparamTable1[which];
		ulDummy = atoi( name + pN->len + 1 ) - 1;

		switch( pN->type )
		{
		case XP_V_SET_TASKCLUSTER:
			XPR_SCOPY ( sizeof(szDummy) , szDummy , xpr , name );
			sql79k_set_taskcluster( XParam, szDummy, ulDummy ) ;
			break;

		case XP_V_SET_DEVSP_NAME:
			sql79k_set_devsp_name ( XParam, name, xpr, pN->dt, ulDummy ) ;
			break;

		case XP_V_SET_M_DEVSP_NAME:
			if ( sql79k_set_devsp_name ( XParam, name, xpr, 
						pN->mdt, ulDummy ) )
			{
				XParam->DevspDescr[ pN->mdt ].
					Devspaces [ ulDummy ].lDevspaceSize =
						XParam->DevspDescr[ pN->dt  ].
							Devspaces [ ulDummy ].lDevspaceSize ;
			}
			break;

		case XP_V_SET_DEVSP_SIZE:
			sql79k_set_devsp_size ( XParam, pN->dt, 
									ulDummy, XPR_INT ( xpr ) ) ;
			break;

		default:
			DBG3 ((MF__, "RteXparamTable entry %d for %s wrong type %d\n",
					which, name, pN->type ));
			break;
		}
	}
	else
	{

		/*
		 * Reduced to offset into RteXparamTable2
		 */
		which -= RTE_XPARAM_TABLE1_MXEO79;
		pE = &RteXparamTable2[which];

		switch( pE->type ) 
		{
		case XP_V_BOOL_Y:
			XPR_SCOPY ( sizeof(szDummy), szDummy, xpr, name );
			*(tsp00_Bool *)XP_PTR(XParam, pE) = 
					(tsp00_Bool)( *szDummy == 'y' || *szDummy == 'Y' );
			break;

		case XP_V_BOOL_N:
			XPR_SCOPY ( sizeof(szDummy), szDummy, xpr, name );
			*(tsp00_Bool *)XP_PTR(XParam, pE) = 
					(tsp00_Bool)( *szDummy == 'n' || *szDummy == 'N' );
			break;

		case XP_V_BOOL_A:
			XPR_SCOPY ( sizeof(szDummy), szDummy, xpr, name );
			*(tsp00_Bool *)XP_PTR(XParam, pE) = 
					(tsp00_Bool)( *szDummy == 'a' || *szDummy == 'A' );
			break;

		case XP_V_BOOL_S:
			XPR_SCOPY ( sizeof(szDummy), szDummy, xpr, name );
			*(tsp00_Bool *)XP_PTR(XParam, pE) = 
					(tsp00_Bool)( *szDummy == 's' || *szDummy == 'S' );
			break;

		case XP_V_LONG:
			*(tsp00_Int *)XP_PTR(XParam, pE) = XPR_INT( xpr );
			break;

		case XP_V_ULONG:
			*(tsp00_Uint *)XP_PTR(XParam, pE) = XPR_INT( xpr );
			break;

		case XP_V_REGIONS:
			ulDummy = (tsp00_Uint)XPR_INT( xpr );
			*(tsp00_Uint *)XP_PTR(XParam, pE) = ulDummy;
			XParam->ulNoOfRegions += ulDummy;
			break;

		case XP_V_STRING:
			XPR_SCOPY ( pE->size, XP_PTR(XParam, pE), xpr, name );
			break;

		case XP_V_CRYPT:
			fGotControlPw = TRUE ;
			COPY ( XP_PTR(XParam,pE), xpr->xp_type.xp_crypt , pE->size ) ;
			break;

		case XP_V_IGNORED:
			break;

		case XP_V_SPECIAL_LONG:
			*(tsp00_Int *)XP_PTR(XParam, pE) = XPR_INT( xpr );
			switch( which ) 
			{
			case XP_T_SPECIAL_KERNELTRACESIZE:
				if ( XParam->DevspDescr [ sp2dt_trace ].Devspaces == NULL )
					  sql79k_alloc_devsp_descr ( XParam, sp2dt_trace, 1 ) ;
				XParam->DevspDescr [ sp2dt_trace ].Devspaces[0].lDevspaceSize =
											  XPR_INT ( xpr ) ;
				break;

			case XP_T_SPECIAL_MAXDEVSPACES:
				XParam->lMaxDevspaces++;	/* Add TRACE_DEVSPACE */
				break;

			case XP_T_SPECIAL_MAXSYSDEVSPACES:
				sql79k_alloc_devsp_descr ( XParam, sp2dt_sys, 
						XParam->lMaxSysDevspaces ) ;
				if ( XParam->fMirrData )
				{
				  sql79k_alloc_devsp_descr ( XParam, sp2dt_mirr_sys, 
						XParam->lMaxSysDevspaces);
				}
				break;

			case XP_T_SPECIAL_MAXARCHIVELOGS:
				if ( XParam->lMaxArchiveLogs > 0 )
				{
				  sql79k_alloc_devsp_descr ( XParam, sp2dt_log, 
						XParam->lMaxArchiveLogs);
				}
				break;

			case XP_T_SPECIAL_MAXDATADEVSPACES:
				sql79k_alloc_devsp_descr ( XParam, sp2dt_data, 
						XParam->lMaxDataDevspaces ) ;
				if ( XParam->fMirrData )
				{
				  sql79k_alloc_devsp_descr ( XParam, sp2dt_mirr_data,
									 XParam->lMaxDataDevspaces);
				}
				break;

			default:
				DBG3 ((MF__, "RteXparamTable no special entry %d for %s\n",
					which, name ));
				break;
			}
			break;

		case XP_V_SPECIAL_ULONG:
			*(tsp00_Uint *)XP_PTR(XParam, pE) = XPR_INT( xpr );
			switch( which ) 
			{
			case XP_T_SPECIAL_PACKET_SIZE:
			case XP_T_SPECIAL_MIN_REPLY_SIZE:
				*(tsp00_Uint *)XP_PTR(XParam, pE) &= ~7;
				break;

			case XP_T_SPECIAL_MAXUSERSTACK:
				XParam->ulMaxUsUtStack <<= 10;
				break;

			case XP_T_SPECIAL_TREE_REGIONS:
				XParam->ulNoOfRegions += XParam->ulNumTreeReg ;
								/* NumSplitt = NumTree */
				XParam->ulNoOfRegions += XParam->ulNumTreeReg ;
				break;

			case XP_T_MP_DISP_LOOPS:
				if ( XParam->ulDispLoopsBefSemop < 2 )
					XParam->ulDispLoopsBefSemop = 2;
				break;

#if defined(_WIN32)
			case XP_T_RTE_ALLOC_POOL:
				XParam->ulRTEAllocPoolSize *= ( 1 MB );
				break;

			case XP_T_KERNEL_ALLOC_POOL:
				XParam->ulKernelAllocPoolSize *= sizeof(tsp00_Page);
				break;
#endif /* WIN32 */

			default:
				DBG3 ((MF__, "RteXparamTable no special entry %d for %s\n",
					which, name ));
				break;
			}
			break;

		case XP_V_SPECIAL_STRING:
			XPR_SCOPY ( pE->size, XP_PTR(XParam, pE), xpr, name );
			switch( which ) 
			{
			case XP_T_SPECIAL_KERNELTRACEFILE:
				if ( XParam->DevspDescr [ sp2dt_trace ].Devspaces == NULL )
					sql79k_alloc_devsp_descr ( XParam, sp2dt_trace, 1 ) ;
				XPR_SCOPY ( sizeof(tsp2_devname) + 1,
					(char *)XParam->DevspDescr [ sp2dt_trace ].Devspaces[0].szDevspaceName,
					xpr , name );
				XParam->DevspDescr [ sp2dt_trace ].Devspaces[0].lDevspaceType =
					sp2dt_trace ;
				break;

			default:
				DBG3 ((MF__, "RteXparamTable no special entry %d for %s\n",
					which, name ));
				break;
			}
			break;

		case XP_V_SPECIAL_SPECIAL:
			switch( which ) 
			{
			case XP_T_LOG_MODE:
				XPR_SCOPY ( sizeof(szDummy) , szDummy , xpr , name );
				switch ( szDummy[1] )
				{
					case 'u' :
					case 'U' : XParam->lLogMode = csp2_dual_log  ;
							   sql79k_alloc_devsp_descr ( XParam, sp2dt_mirr_log,
														  XParam->lMaxArchiveLogs );
							   break ;
					case 'i' :
					case 'I' : XParam->lLogMode = csp2_single_log ; 
							   break ;

					case 'e' :
					case 'E' : XParam->lLogMode = csp2_demo_log  ; 
							   break ;

					default  : MSGCD (( ERR_XP_INVALID_LOG_MODE, szDummy ));
							   vabort( FALSE );
							   break;
				}
				break;

			case XP_T_MIRR_DATA:
				XPR_SCOPY ( sizeof(szDummy) , szDummy , xpr , name );
				if ( (XParam->fMirrData = *szDummy == 'y' || *szDummy == 'Y') )
				{
					if ( XParam->lMaxSysDevspaces > 0 )
					  sql79k_alloc_devsp_descr ( XParam, sp2dt_mirr_sys,
										 XParam->lMaxSysDevspaces);
		/* ADIS 1001565 */
					if ( XParam->lMaxDataDevspaces > 0 )
					  sql79k_alloc_devsp_descr ( XParam, sp2dt_mirr_data,
										 XParam->lMaxDataDevspaces);
				}
				break;
				
			case XP_T_SPECIAL_PAGE_SIZE:
				ulDummy = XPR_INT ( xpr ) ;

				if ( ulDummy != XParam->ulDBPageSize )
				{
				  DBG3 ((MF__, "Illegal Pagesize in xparam ignored : %ld != %ld\n",
					ulDummy, XParam->ulDBPageSize ));
				}
				break;

#if !defined(_WIN32)
			case XP_T_GATEWAYTYPE:
				{
				    MSGCD (( ERR_GATEWAY_PARM_FOUND ));
					vabort( FALSE );
				}
				break;
#endif /* !WIN32 */
			default:
				DBG3 ((MF__, "RteXparamTable no special entry %d for %s\n",
					which, name ));
				break;
			}
			break;

		default:
			DBG3 ((MF__, "RteXparamTable entry %d for %s wrong type %d\n",
					which, name, pE->type ));
			break;
		}
	}
}

/*
 * ===========================================================================
 */


static LONG  sql79k_xpr_int ( XP_VALUE_REC  *xpr )
  {
  #undef  MF__
  #define MF__ MOD__"sql79k_xpr_int"

  DBGPAS;

  switch ( xpr->xp_value_t )
    {
    case XP_INT2_TYPE :
      return ( (int) xpr->xp_type.xp_int2 );

    case XP_INT4_TYPE :
      return ( xpr->xp_type.xp_int4 );

    default :
      MSGCD (( ERR_ILLEGAL_XP_TYPE, xpr->xp_value_t ));
      return ( -1 );
    }
  }

/*
 * ===========================================================================
 */

static void  sql79k_xpr_scopy ( char         *s ,
                                XP_VALUE_REC * xpr  )

  {
  #undef  MF__
  #define MF__ MOD__"sql79k_xpr_copy"

  static int     len [] = { sizeof  xpr->xp_type.xp_int2 ,
			    sizeof  xpr->xp_type.xp_int4 ,
			    sizeof  xpr->xp_type.xp_c8 ,
			    sizeof  xpr->xp_type.xp_c18 ,
  			    sizeof  xpr->xp_type.xp_c24 ,
			    sizeof  xpr->xp_type.xp_c40 ,
			    sizeof  xpr->xp_type.xp_c64 ,
  			    sizeof  xpr->xp_type.xp_crypt } ;

  DBGIN;

  if ( xpr->xp_value_t >= XP_C8_TYPE && xpr->xp_value_t <= XP_C64_TYPE )
    {
    eo46PtoC ( s , (tsp00_Byte *)&xpr->xp_type , (short) len [xpr->xp_value_t]  );
    }
  else
    {
    MSGCD (( ERR_ILLEGAL_XP_TYPE, xpr->xp_value_t ));
    }

  DBG3 ((MF__, "Value = %s" , s));

  DBGOUT;
  }

/*
 * ===========================================================================
 */

static  void    sql79k_alloc_devsp_descr ( RTE_XPARAM_REC         *XParam,
                                           LONG                   lDevspType,
                                           ULONG                  ulNumDevsp)
{
    ULONG i ;
    if ( ALLOC_MEM_EO57 ( (void **)&(XParam->DevspDescr[lDevspType].Devspaces),
                     ulNumDevsp * sizeof ( DEV_DESCR ) ) != NO_ERROR_EO001 )
      {
        MSGCD (( ERR_MALLOC_FAILED_FOR ,"DEVSPACE_DESCR" ));
        DBGOUT ;
        vabort( FALSE );
      }

    for ( i=0; i < ulNumDevsp; i++ )
      XParam->DevspDescr[lDevspType].Devspaces[i].lDevspaceType =
                                                  DEVSPACE_UNUSED  ;
    XParam->DevspDescr[lDevspType].ulNumDevs = ulNumDevsp ;
}

/*
 * ===========================================================================
 */

static  int    sql79k_set_devsp_name ( RTE_XPARAM_REC             *XParam,
                                       tsp11_XpKeyTypec           DevspName ,
                                       XP_VALUE_REC               *xpr ,
                                       LONG                       lDevType ,
                                       ULONG                      ulDevspNo )
{
  int fOk ;
  tsp2_devnamec szDevName ;

  XPR_SCOPY ( sizeof(tsp2_devname) + 1, (char *)szDevName, xpr, DevspName );

  if ( szDevName[0] == '\0' )
    fOk = FALSE ;
  else
  {
    if ( FALSE && /* PTS 1119731 */
        ( ulDevspNo > 0 ) && 
        ( XParam->DevspDescr[lDevType].Devspaces[ulDevspNo - 1].lDevspaceType == DEVSPACE_UNUSED ))
    {
      MSGCD (( ERR_INVALID_DEVSPACE_NO, ulDevspNo ));
      fOk = FALSE;
    }
    else
    {
      fOk = sql58k_check_dup_devspace_name( XParam->DevspDescr, DevspName );

      if ( fOk )
      {
        if (( fOk = XParam->DevspDescr [ lDevType ]. ulNumDevs > ulDevspNo ))
        { 
          DEV_DESCR *ActDevsp = XParam->DevspDescr[lDevType].Devspaces + ulDevspNo;
          memcpy (ActDevsp->szDevspaceName, szDevName, sizeof(tsp2_devname) + 1);
          ActDevsp->lDevspaceType = lDevType ;
        }
        else
        { 
          MSGCD (( ERR_XP_TOO_MANY_DEVSPACES, szDevType[lDevType] ,
                   XParam->DevspDescr [ lDevType ]. ulNumDevs ));
          vabort( FALSE );
        }
      }
    }
  }
  return (fOk) ;
}

/*
 * ===========================================================================
 */

static  int    sql79k_set_devsp_size ( RTE_XPARAM_REC            *XParam,
                                       LONG                      lDevType ,
                                       ULONG                     ulDevspNo ,
                                       ULONG                     ulDevspSize )
{
  int fOk ;

  if ( ulDevspSize == 0 )
    return ( TRUE ) ;

  if ( ( fOk = XParam->DevspDescr [ lDevType ]. ulNumDevs > ulDevspNo ) )
    { DEV_DESCR *ActDevsp = XParam->DevspDescr[lDevType].Devspaces + ulDevspNo;
      ActDevsp->lDevspaceSize = ulDevspSize ;
    }
  else
    { MSGCD (( ERR_XP_TOO_MANY_DEVSPACES, szDevType[lDevType] ,
                            XParam->DevspDescr [ lDevType ]. ulNumDevs )) ;
      vabort( FALSE );
    }

  return (fOk) ;
}

/*
 * ===========================================================================
 */

static  void   sql79k_set_taskcluster( RTE_XPARAM_REC            *XParam,
                                       char                      *szTaskCluster,
                                       ULONG                     ulTCNo )
{
  #undef  MF__
  #define MF__ MOD__"sql79k_set_taskcluster"
  int TaskClusterLen = (int)strlen ( szTaskCluster ) + 1;

  DBGIN;

  if ( XParam->TaskClusterArray == NULL )
    { if ( ALLOC_MEM_EO57 ( (void **) &(XParam->TaskClusterArray),
                     MAX_TASKCLUSTER_PARTS * sizeof ( char * ) ) != NO_ERROR_EO001 )
        { MSGCD (( ERR_MALLOC_FAILED_FOR ,"set TASKCLUSTER" ));
          DBGOUT ;
          vabort( FALSE );
        } 
      else
        memset ( XParam->TaskClusterArray, 0, 
                 MAX_TASKCLUSTER_PARTS * sizeof ( char * ) ) ;
    } 

  if ( ulTCNo < 0 || ulTCNo >= MAX_TASKCLUSTER_PARTS-1 )
    { MSGCD (( ERR_XP_TOO_MANY_TASKCLUSTER, ulTCNo, MAX_TASKCLUSTER_PARTS ));
      DBGOUT ;
      vabort( FALSE );
    } 

  if ( ALLOC_MEM_EO57 ( (void **) &(XParam->TaskClusterArray[ulTCNo]),
                   TaskClusterLen * sizeof ( char) ) != NO_ERROR_EO001 )
    { 
      MSGCD (( ERR_MALLOC_FAILED_FOR ,"set TASKCLUSTER" ));
      DBGOUT ;
      vabort( FALSE );
    } 
  else
    strcpy ( XParam->TaskClusterArray[ulTCNo] , szTaskCluster ) ;

  DBG3 ((MF__, "Taskcluster[%d] set to '%s'", 
         ulTCNo, XParam->TaskClusterArray[ ulTCNo ] ));
}


#ifdef VERSION_TEST
#define ERR_FSIZE_PARAMFILE            50310,ERR_TYPE,  "vgetxpar","Error filesize param file: %s"
#define ERR_WRITING_PARAMFILE          50310,ERR_TYPE,  "vgetxpar","Error writing param file: %s"
#define ERR_OPEN_PARAMFILE             50310,ERR_TYPE,  "vgetxpar","Error opening param file for %s: %s"

/*
 * ===========================================================================
 */
static int test_copy_param()
{
XP_HFILE          xp_fd, xp_fdw ;
tsp_errtext       errtext ;
tsp_int1          xp_ret ;
char              buf [ 4080 ] ;
tsp_int4          bytes_to_read = 4080 ;
tsp_int4          bytes_read, bytes_written ;
char             *NewDbname = "xyz";

vxparam_get_filesize ( &bytes_read, errtext , &xp_ret ) ;

if ( xp_ret != XP_OK )
{  MSGCD (( ERR_FSIZE_PARAMFILE , errtext )) ; }

eo79OpenXparamBinary ( KERNEL_GLOBAL_DBNAME, TRUE, &xp_fd, errtext , &xp_ret ) ;
eo79OpenXparamBinary ( NewDbname, FALSE, &xp_fdw, errtext , &xp_ret ) ;

if ( xp_ret == XP_OK )
  {
    eo79OpenXparamBinary ( NewDbname, FALSE, &xp_fdw, errtext , &xp_ret ) ;
    if ( xp_ret == XP_OK )
      { 
        do
          {
            vxparam_read_binary (  xp_fd, buf, bytes_to_read ,
                                  &bytes_read, errtext , &xp_ret ) ;
            if ( bytes_read > 0 )
              {
                vxparam_write_binary ( xp_fdw, buf, bytes_read ,
                                      &bytes_written, errtext , &xp_ret ) ;
                if ( xp_ret != XP_OK )
    	  		  MSGCD (( ERR_WRITING_PARAMFILE , errtext )) ;
              }
            else
  	  		  MSGCD (( ERR_READING_PARAMFILE , errtext )) ;
        }
        while ( xp_ret == XP_OK ) ;
        vxparam_close ( xp_fdw, errtext , &xp_ret ) ;
      }
    else
     { MSGCD (( ERR_OPEN_PARAMFILE , "write", errtext )) ; }

    vxparam_close ( xp_fd, errtext , &xp_ret ) ;
  }
else
  {  MSGCD (( ERR_OPEN_PARAMFILE , "read", errtext )) ; }

if ( xp_ret == XP_OK )
  vxparam_set_new_serverdb( NewDbname, errtext , &xp_ret ) ;

}

#endif
