/*!***************************************************************************

  module      : veo86.c

  -------------------------------------------------------------------------

  responsible : StefanP

  special area: Diagnostic files
  description : Saving of diagnostic files at restart if the database
                was shutting down abnormally the last time.
                Functions for the server site.

  last changed: 2001-02-07  14:46
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/
#include    "gsp00.h"
#include    "geo00.h"
#include    "geo00_2.h"
#include    "geo002.h"
#include    "geo002xp.h"
#include    "geo007_1.h"
#include    "heo11.h"
#include    "heo79.h"
#include    "geo001.h"
#include    "heo15.h"
#include    "heo47.h"
#include    "heo52.h"
#include    "heo86.h"
#include    "heo861.h"
#include    "geo50_0.h"
#include    "hsp77.h"
#include    "geo57.h"

#if defined (_WIN32)    /*&if $OSSPEC = WIN32*/ 
#include  "gos00.h"     /* nocheck */
#else                   /*&endif*/
#include <sys/types.h>
#include <dirent.h>
#endif


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#if defined(_WIN32)
/*--- NT    */
#define     SEPERATOR                   "\\"
#else
/*--- UNIX  */
#define     SEPERATOR                   "/"
#endif


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/

/*--- Table of diagnostic files to be saved ---*/
typedef struct
{
  tsp11_XpKeyTypec    xparamFilename;
  tsp00_VFilenamec    diagFilename;
  tsp00_Bool          useRunDir;
  tsp00_Pathc         diagPath;
  tsp00_Bool          movefile;   /* true: move file    false: copy file */ /* 1110184 */
} DIAGNOSTIC_FILE;

static  DIAGNOSTIC_FILE   diag_file_table[] =
{                                             /* 1110184 */          
  {PAN_DIAGFILE,        "\0",     true, "\0", false},  /*--- KNLDIAG  */
  {PAN_KERNELTRACEFILE, "\0",     true, "\0", false},  /*--- KNLTRACE */
  {PAN_RTEDUMPFILE,     "\0",     true, "\0", true },  /*--- RTEDUMP  */
  {PAN_KERNELDUMPFILE,  "\0",     true, "\0", true },  /*--- KNLDUMP  */
  {"\0",                "AK.dmp", true, "\0", true },
  {"\0",                "AK.buf", true, "\0", true },
  {"\0",                "AK.stm", true, "\0", true },
  {"\0",                "core",   true, "\0", true }   /*--- UNIX DB-core */  /* PTS 1109823 */
};

#define NUM_DIAG_FILES 	(sizeof(diag_file_table)/sizeof(DIAGNOSTIC_FILE))



/*===========================================================================*
 *  STATIC/INLINE FUNCTIONS (PROTOTYPES)                                     *
 *===========================================================================*/

  
static  void                RTEDiag_SaveDiagFiles           (tsp00_Char        *destPath,
                                                             RTE_XPARAM_REC    *p_xparam);
  
static  tsp00_Bool          RTEDiag_GenerateNewHistDir      (tsp00_VFilenamec   dirname,
                                                             tsp00_DbNamec      dbname); 

static  void                RTEDiag_MakeBackup              (tsp00_Char        *source, 
                                                             tsp00_Char        *dest,
                                                             tsp00_Bool         movefile); /* PTS 1110184 */

static  teo001_SystemRc     RTEDiag_CreateDirectory         (tsp00_Pathc        directory);


/*===========================================================================*
 *  EXTERN FUNCTIONS                                                         *
 *===========================================================================*/

/* For the documentation see header file */
extern	void	eo86_SetCrashSem    (tsp00_Uint4       state, 
                                     tsp00_DbNamec     dbname)
{
/*===========================================================================*
 *  Locals                                                                   *
 *===========================================================================*/
  tsp11_XpReturnCode  rc;
  tsp01_RteError      RteError;
  XP_VALUE_REC        XpValue ;

/*===========================================================================*
 *  Instructions                                                             *
 *===========================================================================*/
  XpValue.xp_value_t = XP_INT4_TYPE;
  XpValue.xp_type.xp_int4 = state;

  rc = eo11PutParamWithoutChgFlag (dbname, PAN_CRASH_SEM, (tsp00_XpValueRec *)&XpValue, 
                     &RteError);
  if (rc != xp_ok_sp11)
  {
    if (state == CRASH_SEM_RESET)
    {
      MSGD ((ERR_DIAG_HIST_SEM, "Resetting", RteError.RteErrText, RteError.RteErrCode));
    }
    else
    {
      MSGD ((ERR_DIAG_HIST_SEM, "Posting", RteError.RteErrText, RteError.RteErrCode));
    }
  }
  return;
}

/* For the documentation see header file */
extern  void  eo86_MakeCrashHist    (RTE_XPARAM_REC   *p_xparam,
                                     tsp00_DbNamec     dbname)
{
/*===========================================================================*
 *  Locals                                                                   *
 *===========================================================================*/
    static  tsp00_Char          routine[] = "eo86_MakeCrashHist";
            tsp00_Pathc         newHistPath;
            tsp00_VFilenamec    newHistDirname;
            teo001_SystemRc     rc = NO_ERROR_EO001;

/*===========================================================================*
 *  Instructions                                                             *
 *===========================================================================*/
  if (p_xparam->uCrashSem == CRASH_SEM_RESET)
    return;

  MSGD ((INFO_DIAG_HIST_START));

  rc = RTEDiag_CreateDirectory (p_xparam->szCrashHistoryPath);
  if (rc != NO_ERROR_EO001 && rc != ERROR_ALREADY_EXISTS_EO001)
  {
      eo861_OsError (routine, "RTEDiag_CreateDirectory", p_xparam->szCrashHistoryPath, rc, ERR_TYPE);
      return;
  }

  /* Caution: uCrashHistoryNum is always > 0. This is checked by cserv.pcf */
  eo861_DeleteOldHistories (p_xparam->uCrashHistoryNum - 1, dbname, p_xparam->szCrashHistoryPath);        

  /*--- Generate new history path */
  if (!RTEDiag_GenerateNewHistDir (newHistDirname, dbname))
  {
    MSGD ((ERR_DIAG_HIST_DIRNAME));
    return;
  }
  
  memset (&newHistPath, 0, sizeof (newHistPath));
  strcpy ((void *)&newHistPath, p_xparam->szCrashHistoryPath);
  strcat ((void *)&newHistPath, SEPERATOR);
  strcat ((void *)&newHistPath, newHistDirname);


  rc = RTEDiag_CreateDirectory (newHistPath);
  if (rc != NO_ERROR_EO001 && rc != ERROR_ALREADY_EXISTS_EO001)
  {
      eo861_OsError (routine, "RTEDiag_CreateDirectory", newHistPath, rc, ERR_TYPE);
      return;
  }

  /*--- Save diagnostic files */
  RTEDiag_SaveDiagFiles (newHistPath, p_xparam);
  
  MSGD ((INFO_DIAG_HIST_STOP));
  return;
}


/*===========================================================================*
 *  STATIC FUNCTIONS                                                         *
 *===========================================================================*/
/*!
  Chapter: RTE Diagnostic (Server, NT and UNIX)
 */
/*!
  Function:     RTEDiag_SaveDiagFiles (static)
  description:  This function executes the backup of the diagnostic files
                to be saved.
                
  arguments:
    path       [in] Path to the directory where the diagnostic files have to be saved
    p_xparam   [in] Pointer to XPARAM-Record

  return value: none
*/

static  void  RTEDiag_SaveDiagFiles (tsp00_Char       *path,
                                     RTE_XPARAM_REC   *p_xparam)
{
/*===========================================================================*
 *  Locals                                                                   *
 *===========================================================================*/
  tsp00_VFilenamec  sourceFilename;
  tsp00_VFilenamec  destFilename;
  tsp00_Pathc       sourcePath;
  tsp00_Pathc       destPath;
  tsp00_Uint4       ix;
  tsp00_Char       *pfilename;

/*===========================================================================*
 *  Instructions                                                             *
 *===========================================================================*/
  for (ix = 0; ix < NUM_DIAG_FILES; ix++)
  {
    /*--- Create path and filename of the file to be saved */
    if (*diag_file_table[ix].xparamFilename != '\0')
    {/*--- Take filename from XParam */
      pfilename = vxparam_GetStringFromXParam (diag_file_table[ix].xparamFilename, p_xparam);
      if (pfilename != NULL)
      {
        strcpy (sourceFilename, pfilename);
        strcpy (destFilename, sourceFilename);
      }
      else
      {
        MSGD ((ERR_DIAG_HIST_GET_FILENAME, diag_file_table[ix].xparamFilename));
        continue;
      }

      /*--- Special treatment for the KNLDIAG-file */
      if (!strcmp (diag_file_table[ix].xparamFilename, PAN_DIAGFILE))
      {
        strcat (sourceFilename, KNLDIAG_EXT_OLD);      
      }
    }
    else /*--- Take filename from table */
    {
      strcpy (sourceFilename, diag_file_table[ix].diagFilename);
      strcpy (destFilename, sourceFilename);
    }
 
    if (diag_file_table[ix].useRunDir)
    {
      strcpy (sourcePath, p_xparam->szRunDirectory);    
    }
    else /*--- Take Rundirectory from table */
    {
      strcpy (sourcePath, diag_file_table[ix].diagPath);    
    }

    strcat (sourcePath, SEPERATOR);
    strcat (sourcePath, sourceFilename);

    /*--- Create path and filename of the file to be saved to */
    strcpy (destPath, path);
    strcat (destPath, SEPERATOR);
    strcat (destPath, destFilename);

    /*--- Make Backup of diagnostic file */
    RTEDiag_MakeBackup (sourcePath, destPath, diag_file_table[ix].movefile); /* 1110184 */
  }
  
  return;
}

/*!
  Function:     RTEDiag_GenerateNewHistDir (static)
  description:  This function executes the backup of the diagnostic files
                to be saved.
                
  arguments:
    dirname     [out]   Name of new history directory
    dbname      [in]    Name of the database

  return value: true    Generation of new history name succeeded
  return value: false   Generation of new history name failed    

*/

static  tsp00_Bool  RTEDiag_GenerateNewHistDir (tsp00_VFilenamec dirname, 
                                                tsp00_DbNamec    dbname)
{
/*===========================================================================*
 *  Locals                                                                   *
 *===========================================================================*/
    
    tsp00_VFilenamec    tmpDirname;
    tsp00_Date          date;
    tsp00_Time          time;

    #define     VDATTIME_OFF_YEAR       0
    #define     VDATTIME_OFF_MONTH      4
    #define     VDATTIME_OFF_DAY        6
    #define     VDATTIME_OFF_HOUR       2
    #define     VDATTIME_OFF_MINUTE     4
    #define     VDATTIME_OFF_SECOND     6

    #define     PUTDATETIME(dt_elem, len, dt, off)                                  \
                                        {                                                                   \
                                            size_t              offset;                                     \
                                            size_t              length;                                     \
                                                                                                            \
                                            offset = strcspn (HIST_NAME_SKEL, (dt_elem));                   \   
                                            length = strspn (HIST_NAME_SKEL + offset, (dt_elem));           \
                                            if  (length != len)                                             \
                                            {                                                               \
                                                return (false);                                             \
                                            }                                                               \
                                                                                                            \
                    if (offset <= strlen (HIST_NAME_SKEL) - (len))                  \
                                            {                                                               \
                                                memcpy (tmpDirname + offset, (dt) + (off), (len));          \
                                            }                                                               \
                                            else                                                            \
                                            {                                                               \
                                                return (false);                                             \
                                            }                                                               \
                                        }                                                                   \


/*===========================================================================*
 *  Instructions                                                             *
 *===========================================================================*/

    strcpy (tmpDirname, HIST_NAME_SKEL); 
            
    vdattime (date, time);  /* date: YYYYMMDD     time:   00HHTTSS */

    PUTDATETIME (HIST_YEAR, HIST_LEN_YEAR, date, VDATTIME_OFF_YEAR);
    PUTDATETIME (HIST_MONTH, HIST_LEN_MONTH, date, VDATTIME_OFF_MONTH);
    PUTDATETIME (HIST_DAY, HIST_LEN_DAY, date, VDATTIME_OFF_DAY);
    PUTDATETIME (HIST_HOUR, HIST_LEN_HOUR, time, VDATTIME_OFF_HOUR);
    PUTDATETIME (HIST_MINUTE, HIST_LEN_MINUTE, time, VDATTIME_OFF_MINUTE);
    PUTDATETIME (HIST_SECOND, HIST_LEN_SECOND, time, VDATTIME_OFF_SECOND);

    strcpy (dirname, dbname);
    strcat (dirname, tmpDirname);

    return (true);
}


/*!
  EndChapter: RTE Diagnostic (Server, NT and UNIX)
 */


/***************************************************************/
#if defined(_WIN32)
/***************************************************************/
/*!
  Chapter: NT only 
 */

/*!
  Function:     RTEDiag_CreateDirectory (static)
  description:  A directory is created
                
  arguments:
    directory   [in]    Name of the directory to be created (full pathname) 

  return value: Creating of the directory failed    :   rc of GetLastError 
                Creating of the directory succeeded :   NO_ERROR_EO001
*/
static  teo001_SystemRc   RTEDiag_CreateDirectory (tsp00_Pathc   directory)
{
/*===========================================================================*
 *  Locals                                                                   *
 *===========================================================================*/
    teo001_SystemRc    rc = NO_ERROR_EO001;
    SECURITY_ATTRIBUTES   SA;
    ACCESS_RIGHTS_REC     Access;

/*===========================================================================*
 *  Instructions                                                             *
 *===========================================================================*/
  SA.nLength                = sizeof(SA);
  SA.bInheritHandle         = FALSE;
  SA.lpSecurityDescriptor   = NULL;

  memset ( &Access, 0, sizeof (Access) );
  Access.Creator.ulAccess       = SPECIFIC_RIGHTS_ALL | STANDARD_RIGHTS_ALL;
  Access.Creator.ulAceFlags     = CONTAINER_INHERIT_ACE | OBJECT_INHERIT_ACE ;
  Access.Admin.ulAccess         = SPECIFIC_RIGHTS_ALL | STANDARD_RIGHTS_ALL;
  Access.Admin.ulAceFlags       = CONTAINER_INHERIT_ACE | OBJECT_INHERIT_ACE ;
  Access.System.ulAccess        = SPECIFIC_RIGHTS_ALL | STANDARD_RIGHTS_ALL;
  Access.System.ulAceFlags      = CONTAINER_INHERIT_ACE | OBJECT_INHERIT_ACE ;
  Access.SAPDBOp.ulAccess       = SPECIFIC_RIGHTS_ALL | STANDARD_RIGHTS_ALL;
  Access.SAPDBOp.ulAceFlags     = CONTAINER_INHERIT_ACE | OBJECT_INHERIT_ACE ;
  Access.DomSAPDBOp.ulAccess    = SPECIFIC_RIGHTS_ALL | STANDARD_RIGHTS_ALL;
  Access.DomSAPDBOp.ulAceFlags  = CONTAINER_INHERIT_ACE | OBJECT_INHERIT_ACE ;
  Access.BackupOp.ulAccess      = GENERIC_READ;
  Access.BackupOp.ulAceFlags    = CONTAINER_INHERIT_ACE | OBJECT_INHERIT_ACE ;
  
  rc = sql49c_alloc_and_init_SD( &Access, &SA.lpSecurityDescriptor );

  if ( rc == NO_ERROR_EO001 )
  {
    if (!CreateDirectory (directory, &SA))
    {
        rc = GetLastError ();
    }
          
    if ( SA.lpSecurityDescriptor )
    {
      FREE_MEM ( SA.lpSecurityDescriptor );
    }
  }

  return (rc);
}

/*!
  Function:     RTEDiag_MakeBackup (static)
  description:  This functions moves a file even if the new destination is on another
                filesystem. 
                The security attributes are beeing restored after moving. 
                .
                                
  arguments:
    source  [in]    Name of the source file (full pathname)
    dest    [in]    Name of the destination file (full pathname) 
    movefile[in]    true:   move file 
                    false:  copy file


  return value: none 
  
*/
static  void  RTEDiag_MakeBackup (tsp00_Char  *source, 
                                  tsp00_Char  *dest,
                                  tsp00_Bool   movefile) /* PTS 1110184 */
{
/*===========================================================================*
 *  Locals                                                                   *
 *===========================================================================*/
    static  tsp00_Char          routine[] = "RTEDiag_MakeBackup";
            teo001_SystemRc     rc = NO_ERROR_EO001;


/*===========================================================================*
 *  Instructions                                                             *
 *===========================================================================*/
    if (movefile)
    {
        if (!MoveFile (source, dest))
        {
            rc = GetLastError ();
            if (ERROR_FILE_NOT_FOUND != rc)
            {
                eo861_OsError (routine, "MoveFile", source, rc, ERR_TYPE);
            }
        }
    }
    else    /* PTS 1110184 */
    {
        if (!CopyFile (source, dest, true))
        {
            rc = GetLastError ();
            if (ERROR_FILE_NOT_FOUND != rc)
            {
                eo861_OsError (routine, "CopyFile", source, rc, ERR_TYPE);
            }
        }

    }

  return;
}

/*!
  EndChapter: NT only
 */

/***************************************************************/
#else /* UNIX, LINUX */
/***************************************************************/
/*!
  Chapter: UNIX only
 */

/*!
  Function:     RTEDiag_CreateDirectory (static)
  description:  A directory is created
                
  arguments:
    directory   [in]    Name of the directory to be created (full pathname) 

  return value: Creating of the directory failed    :   errno
                Creating of the directory succeeded :   NO_ERROR_EO001
*/
static  teo001_SystemRc  RTEDiag_CreateDirectory (tsp00_Pathc    directory)
{
/*===========================================================================*
 *  Locals                                                                   *
 *===========================================================================*/
    teo001_SystemRc    rc = NO_ERROR_EO001;
    int                umask_old;
/*===========================================================================*
 *  Instructions                                                             *
 *===========================================================================*/
    umask_old = umask(0);
    if (mkdir (directory, 0777) != 0)
    {
        rc = errno; 
    }

    umask(umask_old);
    return (rc);
}

/*!
  Function:     RTEDiag_MakeBackup (static)
  description:  Moves or copies files. If the new destination is on the same
                filesystem the file is moved otherwise it is copied.
                .
                                
  arguments:
    source  [in]    Name of the source file (full pathname)
    dest    [in]    Name of the destination file (full pathname) 
    movefile[in]    true:   move file 
                    false:  copy file



  return value: none 
  
*/
static  void  RTEDiag_MakeBackup (tsp00_Char  *source, 
                                  tsp00_Char  *dest,
                                  tsp00_Bool   movefile)    /* PTS 1110184 */
{
/*===========================================================================*
 *  Locals                                                                   *
 *===========================================================================*/
    static  tsp00_Char              routine[] = "RTEDiag_MakeBackup";
            tsp00_VfReturn_Param    rcStatus;
            teo15_Handle            hFile;


/*===========================================================================*
 *  Instructions                                                             *
 *===========================================================================*/
  /* the following 'open'is only performed to determine if the file exists */
    hFile = open (source, O_RDWR);
    if (hFile != INVALID_HANDLE_EO15)
    {
        if (movefile)
        {
            if (link (source, dest) != NO_ERROR_EO001)
            {
                if (EXDEV == errno)   /* dest is located on a different file system */
                {
                    eo15_CopyFile ( source, dest, &rcStatus);
                    if (vf_ok == rcStatus)
                    {
                        unlink (source);
                    }
                }
                else
                {
                     eo861_OsError (routine, "link", source, UNDEF_ERROR, ERR_TYPE);
                }
            }
            else
            {
                unlink (source);
            }
        }
        else /* PTS 1110184 */
        {
            eo15_CopyFile ( source, dest, &rcStatus);
        }

        if (close (hFile) != 0)             
        {   
            eo861_OsError (routine, "close", source, UNDEF_ERROR, ERR_TYPE);
        } 
    }

    return;
}

/*!
  EndChapter: UNIX only
 */ 

/***************************************************************/
#endif
/***************************************************************/

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/
