/****************************************************************************

  module      : veo92.c

  -------------------------------------------------------------------------

  responsible : FrankH

  special area: console(x_cons)
  description : common x_cons / sysmon routines


  last changed: 2000-02-22  16:20
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1999



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "geo00_2.h"
#include "veo00xc.h"
#include "geo007_2.h"
#include "geo002.h"
#include "geo002xp.h"
#include "geo003.h"
#include "heo92.h"
#include "heo11.h"
#include "heo47.h"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#if _WIN32 
#define NEXT_TASK(_pCurrTask) _pCurrTask->pNextTaskCtrl
#define global     

#else

#define NEXT_TASK(_pCurrTask) _pCurrTask->next_task
global void                            e92_print_param () ;

#endif

#define pri printf

global int                             e92_set_new_param () ;
global teo00_Bool                      e92_strcont () ;

#define TYPE_STR  0
#define TYPE_BOOL 1
#define TYPE_INT2 2
#define TYPE_INT4 4
#define TYPE_LONG 8

#define SEC_PER_MIN  60
#define SEC_PER_HOUR 3600
#define SEC_PER_DAY  (24*SEC_PER_HOUR)


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/



/*===========================================================================*
 *  GLOBAL VARIABLES                                                         *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL VARIABLES                                                          *
 *===========================================================================*/



/*===========================================================================*
 *  STATIC/INLINE FUNCTIONS (PROTOTYPES)                                     *
 *===========================================================================*/

static int set_int_param( int   int_type , 
                          void *act_param, 
                          char *new_param, 
                          long  lower_limit, 
                          long  upper_limit ) ;

static int set_bool_param( teo00_Bool *act_param,
                           char       *new_param ) ;

static void e92_print_one_param ( char *param_id, 
                                  char *param_wanted, 
                                  int   wanted_len,
                                  void *value, 
                                  char  type ) ;

/*===========================================================================*
 *  PUBLIC FUNCTIONS (CODE)                                                  *
 *===========================================================================*/

char *eo92GetTaskStateName ( teo00_Longint lState,
                             teo00_Longint lStateReason)
  {
  #undef  MF__
  #define MF__ MOD__"eo92GetTaskStateName"

  DBGPAS;

  switch ( lState )
    {
    case TSK_ASYNCLOSE_EO00:          return ( XCONS_TSK_STATE_ASYNCLOSE      );
    case TSK_ASYNIO_EO00:             return ( XCONS_TSK_STATE_ASYNIO         );
    case TSK_ASYNOPEN_EO00:           return ( XCONS_TSK_STATE_ASYNOPEN       );
    case TSK_ASYNWAIT_READ_EO00:      return ( XCONS_TSK_STATE_ASYNWAIT_READ  );
    case TSK_ASYNWAIT_WRITE_EO00:     return ( XCONS_TSK_STATE_ASYNWAIT_WRITE );
    case TSK_CONNECTWAIT_EO00:        return ( XCONS_TSK_STATE_CONNECTWAIT    );
    case TSK_INACTIVE_EO00:           return ( XCONS_TSK_STATE_INACTIVE       );
    case TSK_INITIALIZED_EO00:        return ( XCONS_TSK_STATE_INITIALIZE     );
    case TSK_IO2_READ_EO00:           return ( XCONS_TSK_STATE_IO2_READ       );
    case TSK_IO2_WRITE_EO00:          return ( XCONS_TSK_STATE_IO2_WRITE      );
    case TSK_IO_READ_EO00:            return ( XCONS_TSK_STATE_IO_READ        );
    case TSK_IO_WRITE_EO00:           return ( XCONS_TSK_STATE_IO_WRITE       );
    case TSK_LOCKED_EO00:             return ( XCONS_TSK_STATE_LOCKED         );
    case TSK_MARKED_FOR_STARTUP_EO00: return ( XCONS_TSK_STATE_MARKED_FOR_STRT);
    case TSK_RUNNABLE_EO00:           return ( XCONS_TSK_STATE_RUNNABLE       );
    case TSK_RUNNING_EO00:            return ( XCONS_TSK_STATE_RUNNING        );
    case TSK_TERMINATED_EO00:         return ( XCONS_TSK_STATE_TERMINATED     );
    case TSK_V2INFO_EO00:             return ( XCONS_TSK_STATE_V2INFO         );
    case TSK_V2BLOCKIO_EO00:          return ( XCONS_TSK_STATE_V2BLOCKIO      );
    case TSK_VATTACH_EO00:            return ( XCONS_TSK_STATE_VATTACH        );
    case TSK_VBEGEXCL_EO00:           return ( XCONS_TSK_STATE_VBEGEXCL       );
    case TSK_VCREATE_EO00:            return ( XCONS_TSK_STATE_VCREATE        );
    case TSK_VDETACH_EO00:            return ( XCONS_TSK_STATE_VDETACH        );
    case TSK_VDEVSIZE_EO00:           return ( XCONS_TSK_STATE_VDEVSIZE       );
    case TSK_VDIAGINIT_EO00:          return ( XCONS_TSK_STATE_VDIAGINIT      );
    case TSK_VENDEXCL_EO00:           return ( XCONS_TSK_STATE_VENDEXCL       );
    case TSK_VFCLOSE_EO00:            return ( XCONS_TSK_STATE_VFCLOSE        );
    case TSK_VFOPEN_EO00:             return ( XCONS_TSK_STATE_VFOPEN         );
    case TSK_VFREAD_EO00:             return ( XCONS_TSK_STATE_VFREAD         );
    case TSK_VFWRITE_EO00:            return ( XCONS_TSK_STATE_VFWRITE        );
    case TSK_VKILL_EO00:              return ( XCONS_TSK_STATE_VKILL          );
    case TSK_VNCLEAR_EO00:            return ( XCONS_TSK_STATE_VNCLEAR        );
    case TSK_VNCLOSE_EO00:            return ( XCONS_TSK_STATE_VNCLOSE        );
    case TSK_VNOPEN_EO00:             return ( XCONS_TSK_STATE_VNOPEN         );
    case TSK_VNRECEIVE_EO00:          return ( XCONS_TSK_STATE_VNRECEIVE      );
    case TSK_VNREQUEST_EO00:          return ( XCONS_TSK_STATE_VNREQUEST      );
    case TSK_VNRESTART_EO00:          return ( XCONS_TSK_STATE_VNRESTART      );
    case TSK_VNSHUTDOWN_EO00:         return ( XCONS_TSK_STATE_VNSHUTDOWN     );
    case TSK_VOPMSG_EO00:             return ( XCONS_TSK_STATE_VOPMSG         );
    case TSK_VRECEIVE_EO00:           return ( XCONS_TSK_STATE_VRECEIVE       );
    case TSK_VRELEASE_EO00:           return ( XCONS_TSK_STATE_VRELEASE       );
    case TSK_VREPLY_EO00:             return ( XCONS_TSK_STATE_VREPLY         );
    case TSK_VSHUTDOWN_EO00:          return ( XCONS_TSK_STATE_VSHUTDOWN      );
    case TSK_VSLEEP_EO00:             return ( XCONS_TSK_STATE_VSLEEP         );
    case TSK_VSUSPEND_EO00:
         if ( lStateReason > 0 )
           return (SuspendReasonText[lStateReason].szShortReason) ;
         else            
           return ( XCONS_TSK_STATE_VSUSPEND       );
         break ;
    case TSK_VWAIT_EO00:              return ( XCONS_TSK_STATE_VWAIT          );
    case TSK_DCOM_OBJ_CALLED_EO00:    return ( XCONS_TSK_STATE_DCOM_OBJ_CALLED);
    case TSK_RESCHEDULE_MSEC_EO00:    return ( XCONS_TSK_STATE_RESCHEDULE_MSEC);
    case TSK_PSE36_DATA_READ_EO00:    return ( XCONS_TSK_STATE_PSE_DATA_READ  );
    case TSK_PSE36_DATA_WRITE_EO00:   return ( XCONS_TSK_STATE_PSE_DATA_WRITE );
    case TSK_PSE36_OBJLOG_READ_EO00:  return ( XCONS_TSK_STATE_PSE_OLOG_READ  );
    case TSK_PSE36_OBJLOG_WRITE_EO00: return ( XCONS_TSK_STATE_PSE_OLOG_WRITE );
    case TSK_INSERT_EVENT_EO00:       return ( XCONS_TSK_STATE_INSERT_EVENT );
    case TSK_WAIT_FOR_EVENT_EO00:     return ( XCONS_TSK_STATE_WAIT_FOR_EVENT );
    case TSK_STOPPED_EO00:            return ( XCONS_TSK_STATE_STOPPED );
    default: break;
  }
  return ( XCONS_TSK_STATE_UNKNOWN );
  }

/*------------------------------*/

char *eo92GetTaskTypeName ( teo00_Uint1 TaskType )
  {
  #undef  MF__
  #define MF__ MOD__"eo92GetTaskTypeName"

  DBGPAS;

  switch ( TaskType )
    {
    case TT_TI_EO00:     return ( XCONS_TSK_TYPE_TI );
    case TT_AL_EO00:     return ( XCONS_TSK_TYPE_AL );
    case TT_OL_EO00:     return ( XCONS_TSK_TYPE_OL );
    case TT_TW_EO00:     return ( XCONS_TSK_TYPE_TW );
    case TT_DW_EO00:     return ( XCONS_TSK_TYPE_DW );
    case TT_SN_EO00:     return ( XCONS_TSK_TYPE_SN );
    case TT_RC_EO00:     return ( XCONS_TSK_TYPE_RC );
    case TT_CS_EO00:     return ( XCONS_TSK_TYPE_CS );
    case TT_UT_EO00:     return ( XCONS_TSK_TYPE_UT );
    case TT_SV_EO00:     return ( XCONS_TSK_TYPE_SV );
    case TT_US_EO00:     return ( XCONS_TSK_TYPE_US );
    case TT_EV_EO00:     return ( XCONS_TSK_TYPE_EV );
    case TT_GC_EO00:     return ( XCONS_TSK_TYPE_GC );
    case TT_BUP_EO00:    return ( XCONS_TSK_TYPE_BUP );
    }

  return ( XCONS_TSK_TYPE_UNKNOWN );
  }

/*------------------------------*/

void eo92_SecToDDHHMMSS( long  Sec, 
                         char *Buf ) /* ddd:hh:mm:ss: */
{
#undef  MF__
#define MF__ MOD__"eo92_SecToDDHHMMSS"

  int  dd = Sec / SEC_PER_DAY , 
       hh = ( Sec % SEC_PER_DAY  ) / SEC_PER_HOUR , 
       mm = ( Sec % SEC_PER_HOUR ) / SEC_PER_MIN, 
       ss = ( Sec % SEC_PER_MIN );
  
  if ( dd )                     sprintf(Buf, "%3d:",dd%1000 );
  else                          strcpy(Buf, "    ") ;
  Buf+=4 ;

  if ( dd || hh )               sprintf(Buf, "%02d:",hh%100 );
  else                          strcpy(Buf, "   ") ;
  Buf+=3 ;

  if ( dd || hh || mm )         sprintf(Buf, "%02d:",mm%100 );
  else                          strcpy(Buf, "   ") ;
  Buf+=3 ;

  if ( dd || hh || mm || ss)    sprintf(Buf, "%02d",ss%100 );
  else                          strcpy(Buf, "  ") ;

  Buf+=2 ;
  *Buf='\0' ;

  return ;
} 

/*================================================*/
teo00_Bool e92_strcont ( char *string1 , 
                         char *string2 , 
                         int   minlgt )
/*================================================*/
{
#undef  MF__
#define MF__ MOD__"e92_strcont"
        int                     i , lgt ;

        /*
         *  Does string1 start with string2 of at least minlgt characters ?
         */
        if ( (int) strlen(string1) < (lgt = (int)strlen(string2)) ) return ( FALSE );
        if ( lgt < minlgt ) return ( FALSE );
        if ( minlgt == 0  ) return ( TRUE );
        
        for ( i = 0 ; i < lgt ; i ++ )
        {
            if ( string1 [ i ] != string2 [ i ] ) break ;
        }
        if ( i < lgt ) return ( FALSE );
        return ( TRUE );
}


#define SHOW_UNKNOWN    0
/*================================================*/
int     e92_get_tt_to_show (char *active_param , 
                            int *s_task_type )
/*================================================*/
{
    int i, rc = NO_ERROR;

    for ( i = 0 ; active_param [ i ] != '\0' ; i ++ )
      if ( islower ( active_param [ i ] ) )
         active_param [ i ] = active_param [ i ] - 'a' + 'A' ;

    if ( e92_strcont ( "DW"   , active_param , 2 ) )
      *s_task_type  = TT_DW_EO00 ;
    else
    if ( e92_strcont ( "SV"   , active_param , 2 ) )
      *s_task_type  = TT_SV_EO00 ;
    else
    if ( e92_strcont ( "US"   , active_param , 2 ) )
      *s_task_type  = TT_US_EO00 ;
    else
    if ( e92_strcont ( "GC"   , active_param , 2 ) ) /* PTS 1107070 */
      *s_task_type  = TT_GC_EO00 ;
    else
      rc = !NO_ERROR ;

    return ( rc ) ;
}


/*------------------------------*/

/*===========================================================================*/
#ifdef CONSOLE
/*===========================================================================*/

int e92_set_new_param ( RTE_XPARAM_REC *XParam, 
                        char           *param_id, 
                        char           *param_val )
{
#undef  MF__
#define MF__ MOD__"e92_set_new_param"

int                             set_ok ;

if ( *param_id == 'h' && *param_val == 'h' )
  { set_ok = FALSE ;
    pri ( "\nTunable parameters:\n\n" ) ;
    e92_print_param ( XParam, PAN_DELAY_LW ) ;
    e92_print_param ( XParam, PAN_USE_IOTHREAD_ONLY ) ;
    e92_print_param ( XParam, PAN_MP_RGN_LOOP ) ;
    e92_print_param ( XParam, PAN_MP_RGN_PRIO ) ;
    e92_print_param ( XParam, PAN_MP_DISP_PRIO ) ;
    e92_print_param ( XParam, PAN_MP_RGN_BUSY_WAIT ) ;
    e92_print_param ( XParam, PAN_MP_DISP_LOOPS ) ;
    e92_print_param ( XParam, PAN_PRIO_BASE_U2U ) ;
    e92_print_param ( XParam, PAN_PRIO_BASE_IOC ) ;
    e92_print_param ( XParam, PAN_PRIO_BASE_RAV ) ;
    e92_print_param ( XParam, PAN_PRIO_BASE_REX ) ;
    e92_print_param ( XParam, PAN_PRIO_BASE_COM ) ;
    e92_print_param ( XParam, PAN_PRIO_FACTOR   ) ;
/*
    e92_print_param ( XParam, PAN_DISTINCT_OVERHEAD ) ;
    e92_print_param ( XParam, PAN_DELAY_COMMIT ) ;
    e92_print_param ( XParam, PAN_REG_DIRTY_READ ) ;
    e92_print_param ( XParam, PAN_REG_LOCK_SLICE ) ;
*/
  }
else if ( ! strcmp( param_id , PAN_PRIO_BASE_U2U  ) )
  set_ok = set_int_param ( TYPE_LONG, &XParam->lPrioBaseU2U, param_val, 0, 10000 ) ; 

else if ( ! strcmp( param_id , PAN_PRIO_BASE_RAV  ) )
  set_ok = set_int_param ( TYPE_LONG, &XParam->lPrioBaseRAV, param_val, 0, 10000 ) ; 

else if ( ! strcmp( param_id , PAN_PRIO_BASE_IOC  ) )

  set_ok = set_int_param ( TYPE_LONG, &XParam->lPrioBaseIOC, param_val, 0, 10000 ) ; 

else if ( ! strcmp( param_id , PAN_PRIO_BASE_COM  ) )

  set_ok = set_int_param ( TYPE_LONG, &XParam->lPrioBaseCOM, param_val, 0, 10000 ) ; 
else if ( ! strcmp( param_id , PAN_PRIO_BASE_REX  ) )

  set_ok = set_int_param ( TYPE_LONG, &XParam->lPrioBaseREX, param_val, 0, 10000 ) ; 
else if ( ! strcmp( param_id , PAN_PRIO_FACTOR  ) )
  set_ok = set_int_param ( TYPE_LONG, &XParam->lPrioFactor, param_val, 0, 10000 ) ; 


else if ( ! strcmp( param_id , PAN_DELAY_LW ) )
  set_ok = set_int_param ( TYPE_LONG, &XParam->lDelayLW , param_val, 0, 100 ) ; 

else if ( ! strcmp( param_id , PAN_DELAY_COMMIT ) )
  set_ok = set_bool_param ( &XParam->fDelayCommit , param_val ) ; 

else if ( ! strcmp( param_id , PAN_USE_IOTHREAD_ONLY ) )
  set_ok = set_bool_param ( &XParam->fAllowSingleIO , param_val ) ; 

else if ( ! strcmp( param_id , PAN_MP_RGN_PRIO ) )
  set_ok = set_bool_param ( &XParam->fBegExclQueWhenPrio , param_val ) ; 

else if ( ! strcmp( param_id , PAN_MP_DISP_PRIO ) )
  set_ok = set_bool_param ( &XParam->fDynDispQueSrch , param_val ) ; 

else if ( ! strcmp( param_id , PAN_MP_RGN_DIRTY_READ ) )
  set_ok = set_bool_param ( &XParam->fBegExclDirtyRead , param_val ) ; 

else if ( ! strcmp( param_id , PAN_MP_RGN_BUSY_WAIT ) )
  set_ok = set_bool_param ( &XParam->fEndExclBusyWait , param_val ) ; 

else if ( ! strcmp( param_id , PAN_MAXRGN_REQUEST ) )
  set_ok = set_int_param ( TYPE_LONG, &XParam->ulRegLockSlice , 
                                      param_val, 25, 100000 ) ; 

else if ( ! strcmp( param_id , PAN_MP_DISP_LOOPS ) )
  set_ok = set_int_param ( TYPE_LONG, &XParam->ulDispLoopsBefSemop , 
                                      param_val, 2, 1000000 ) ; 

else if ( ! strcmp( param_id , PAN_MP_RGN_LOOP ) )
  if ( XParam->ulMaxCPU > 1 )
    set_ok = set_int_param ( TYPE_LONG, &XParam->ulRegionCollisionLoop, 
                                        param_val, 1, 100000 ) ; 
else
  { set_ok = FALSE ;
    pri ( "putparam for parameter '%s' not allowed!\n", param_id ) ;
  }
  return( set_ok ) ;
}

/*------------------------------*/
void     e92_print_param ( RTE_XPARAM_REC *XParam, 
                           char           *name )
{
#undef  MF__
#define MF__ MOD__"e92_print_param"

int                             name_len = (int)strlen(name) ;
teo00_Bool                      fUserIoThreadsOnly;

DBGPAS;

#define PRINT_ONE_PARAM( param_id, param_val, param_type ) \
    e92_print_one_param ( param_id, name , name_len,  param_val , param_type ) ;

PRINT_ONE_PARAM ( PAN_PARAM_KERNELVERSION, XParam->szKernelVersion, TYPE_STR ) ;

PRINT_ONE_PARAM ( PAN_MAXDEVSPACES ,     &XParam->lMaxDevspaces , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_MAXDATADEVSPACES , &XParam->lMaxDataDevspaces, TYPE_LONG );

PRINT_ONE_PARAM ( PAN_MAXSYSDEVSPACES  , &XParam->lMaxSysDevspaces, TYPE_LONG );

PRINT_ONE_PARAM ( PAN_MAXARCHIVELOGS   , &XParam->lMaxArchiveLogs, TYPE_LONG );

PRINT_ONE_PARAM ( PAN_LOG_MODE         , &XParam->lLogMode ,TYPE_LONG);

PRINT_ONE_PARAM ( PAN_LOG_SEGMENT_SIZE , &XParam->lLogSegmentSize ,TYPE_LONG);

PRINT_ONE_PARAM ( PAN_MIRR_DATA        , &XParam->fMirrData, TYPE_BOOL);

PRINT_ONE_PARAM ( PAN_SESSION_TIMEOUT  , &XParam->lSessionTimeout ,TYPE_LONG);

PRINT_ONE_PARAM ( PAN_REQUEST_TIMEOUT  , &XParam->lRequestTimeout ,TYPE_LONG);

PRINT_ONE_PARAM ( PAN_LOCK_TIMEOUT     , &XParam->lLockTimeout ,TYPE_LONG);

PRINT_ONE_PARAM ( PAN_RESTART_TIME     , &XParam->lRestartTime ,TYPE_LONG);

PRINT_ONE_PARAM ( PAN_NET_TIMEOUT      , &XParam->lNetTimeout ,TYPE_LONG);

PRINT_ONE_PARAM ( PAN_DEFAULT_CODE     , XParam->szDefaultCode,TYPE_STR);

PRINT_ONE_PARAM ( PAN_DATE_TIME_FORMAT , XParam->szDateTimeFormat,TYPE_STR);

PRINT_ONE_PARAM ( PAN_MAXBACKUPDEVS    , &XParam->lNoOfBackupDevs , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_ALLOW_LBACK_TO_PIPE,&XParam->fLogBackToPipe , TYPE_BOOL );

PRINT_ONE_PARAM ( PAN_BACKUP_BLOCK_CNT,  &XParam->lBackUpBlockCount, TYPE_LONG);

PRINT_ONE_PARAM ( PAN_MULT_IO_BLOCK_CNT, &XParam->ulMultIOBlockCount,TYPE_LONG);

PRINT_ONE_PARAM ( PAN_MAXTRANS          ,&XParam->lMaxTrans , TYPE_LONG ) ;

PRINT_ONE_PARAM ( PAN_MAXDATAPAGES ,     &XParam->lMaxDataPages , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_DATACACHE_REGIONS ,&XParam->ulNumDataCacheReg ,TYPE_LONG);

PRINT_ONE_PARAM ( PAN_TREE_REGIONS ,     &XParam->ulNumTreeReg ,TYPE_LONG);

PRINT_ONE_PARAM ( PAN_OMS_REGIONS  ,     &XParam->ulNumOMSReg ,TYPE_LONG);

PRINT_ONE_PARAM ( PAN_DATA_CACHE_PAGS   ,&XParam->lDataCachePages , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_SEQUENCE_CACHE    ,&XParam->lSequenceCache , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_CONV_CACHE_PAGES , &XParam->lConvCachePages , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_FBM_CACHE_PAGES ,  &XParam->lFbmCachePages , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_FBM_DEVSP_COMPR ,  &XParam->lFbmCompression, TYPE_LONG );

PRINT_ONE_PARAM ( PAN_FBM_DEVSP_BALANCE ,&XParam->lFbmBalance, TYPE_LONG );

PRINT_ONE_PARAM ( PAN_FBM_LOW_IO_RATE ,  &XParam->lFbmLowIORate, TYPE_LONG );

PRINT_ONE_PARAM ( PAN_USM_CACHE_PAGES ,  &XParam->lUsmCachePages , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_LOG_IO_QUEUE ,     &XParam->lLogQueuePages, TYPE_LONG );

PRINT_ONE_PARAM ( PAN_ROLLBACK_CACHE ,   &XParam->lLogCachePages , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_CATALOG_CACHE   ,  &XParam->lCatalogCachePages, TYPE_LONG );

PRINT_ONE_PARAM ( PAN_SERVER_DESC_CACHE, &XParam->lDistribDescCache, TYPE_LONG );

PRINT_ONE_PARAM ( PAN_SERVER_CMD_CACHE , &XParam->lServerCmdCache, TYPE_LONG );

PRINT_ONE_PARAM ( PAN_MAXACTIVE ,        &XParam->ulMaxUser , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_MIN_SERVER_DESC ,  &XParam->lMinServerDesc , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_MAXSERVER ,        &XParam->ulMaxServer , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_MAXSERVERDB ,      &XParam->lMaxServerDB , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_MAXLOCKS ,         &XParam->lMaxLocks , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_DEFAULT_LRU ,      &XParam->fDefaultLRU , TYPE_BOOL);

PRINT_ONE_PARAM ( PAN_SERVERDB_FOR_SAP,  &XParam->fServerDBForSAP , TYPE_BOOL);

PRINT_ONE_PARAM ( PAN_DYNDATA ,          &XParam->lSharedDynData , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_DYNPOOL ,          &XParam->lSharedDynPool , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_DIAGFILE  ,         XParam->szDiagFile, TYPE_STR );

PRINT_ONE_PARAM ( PAN_DIAG_SIZE ,        &XParam->ulDiagSize , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_EVENTFILE  ,        XParam->szEventFile, TYPE_STR );

PRINT_ONE_PARAM ( PAN_EVENT_SIZE ,       &XParam->ulEventSize , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_MAXEVENTS ,        &XParam->ulMaxEvents , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_RUNDIR ,           XParam->szRunDirectory , TYPE_STR );

PRINT_ONE_PARAM ( PAN_KERNELDUMPFILE  ,  XParam->szDumpFilename , TYPE_STR );

PRINT_ONE_PARAM ( PAN_KERNELTRACEFILE ,  XParam->szTraceDevspace, TYPE_STR );

PRINT_ONE_PARAM ( PAN_KERNELTRACESIZE , &XParam->lTraceDevspaceSize, TYPE_LONG );

fUserIoThreadsOnly = !XParam->fAllowSingleIO ;
PRINT_ONE_PARAM ( PAN_USE_IOTHREAD_ONLY , &fUserIoThreadsOnly , TYPE_BOOL );

PRINT_ONE_PARAM ( PAN_MAXIOTHREADS ,    &XParam->lDeviPerDevspace , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_IOTHREADS_FOR_PRIO, &XParam->lPrioDeviPerDevspace, TYPE_LONG );

PRINT_ONE_PARAM ( PAN_DEV_QUEUE ,         &XParam->lDevQueue , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_MAXCPU ,            &XParam->ulMaxCPU , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_READAHEAD_BLOBS ,   &XParam->ulReadAheadBlobsGT , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_PRIO_BASE_U2U ,     &XParam->lPrioBaseU2U , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_PRIO_BASE_IOC ,     &XParam->lPrioBaseIOC , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_PRIO_BASE_RAV ,     &XParam->lPrioBaseRAV , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_PRIO_BASE_COM ,     &XParam->lPrioBaseCOM , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_PRIO_BASE_REX ,     &XParam->lPrioBaseREX , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_PRIO_FACTOR   ,     &XParam->lPrioFactor  , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_MP_RGN_PRIO,        &XParam->fBegExclQueWhenPrio, TYPE_BOOL);

PRINT_ONE_PARAM ( PAN_MP_RGN_QUEUE,       &XParam->fExclQueue, TYPE_BOOL);

PRINT_ONE_PARAM ( PAN_MP_DISP_PRIO ,      &XParam->fDynDispQueSrch ,TYPE_BOOL);

PRINT_ONE_PARAM ( PAN_MP_RGN_DIRTY_READ , &XParam->fBegExclDirtyRead ,TYPE_BOOL);

PRINT_ONE_PARAM ( PAN_MP_RGN_BUSY_WAIT ,  &XParam->fEndExclBusyWait ,TYPE_BOOL);

PRINT_ONE_PARAM ( PAN_MP_DISP_LOOPS ,     &XParam->ulDispLoopsBefSemop,TYPE_LONG);

PRINT_ONE_PARAM ( PAN_MP_RGN_LOOP,        &XParam->ulRegionCollisionLoop,TYPE_LONG);

PRINT_ONE_PARAM ( PAN_MAXRGN_REQUEST ,    &XParam->ulRegLockSlice ,TYPE_LONG);

PRINT_ONE_PARAM ( PAN_DELAY_LW ,          &XParam->lDelayLW , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_DELAY_COMMIT ,      &XParam->fDelayCommit , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_MAXUSERSTACK ,      &XParam->ulMaxUsUtStack ,TYPE_LONG);

PRINT_ONE_PARAM ( PAN_MAXDATAWRITER ,     &XParam->ulMaxDataWriter ,TYPE_LONG);

/* PRINT_ONE_PARAM ( PAN_CONTROLID         , XParam->szControlUser     ,TYPE_STR); */

PRINT_ONE_PARAM ( PAN_JOIN_SEARCH_LEVEL , &XParam->lJoinSearchLevel, TYPE_LONG);

PRINT_ONE_PARAM ( PAN_JOIN_MAXTAB_LEVEL4 , &XParam->lJoinMaxTabLevel4, TYPE_LONG);

PRINT_ONE_PARAM ( PAN_JOIN_MAXTAB_LEVEL9 , &XParam->lJoinMaxTabLevel9, TYPE_LONG);

PRINT_ONE_PARAM ( PAN_OPTIM_OR_DISTINCT , &XParam->lDistinctOverhead, TYPE_LONG);

PRINT_ONE_PARAM ( PAN_OPTIM_ORDERBY_IDX,   &XParam->lOrderbyFactor, TYPE_LONG);

PRINT_ONE_PARAM ( PAN_OPTIM_BUILD_RESLT, &XParam->lBuildResultPercent,TYPE_LONG);

PRINT_ONE_PARAM ( PAN_OPTIM_FETCH_RESLT, &XParam->lFetchResultPercent,TYPE_LONG);

PRINT_ONE_PARAM ( PAN_OPTIM_MAX_MERGE  , &XParam->lMaxMerge,TYPE_LONG);

PRINT_ONE_PARAM ( PAN_OPTIM_JOIN_FETCH , &XParam->lJoinPrefetchPercent,TYPE_LONG);

PRINT_ONE_PARAM ( PAN_OPTIM_KEY_INV_RATE, &XParam->lKeyInvRatePercent ,TYPE_LONG);

PRINT_ONE_PARAM ( PAN_AK_DUMP_ALLOWED ,  &XParam->fAKDumpAllowed , TYPE_BOOL);

PRINT_ONE_PARAM ( PAN_UNICODE_CATALOG ,  &XParam->fUnicode , TYPE_BOOL);


PRINT_ONE_PARAM ( PAN_MULTIBYTE_SET_1 ,  XParam->szMultiByteSet_1, TYPE_STR);

PRINT_ONE_PARAM ( PAN_MULTIBYTE_SET_2 ,  XParam->szMultiByteSet_2, TYPE_STR);

PRINT_ONE_PARAM ( PAN_MULTIBYTE_SET_3 ,  XParam->szMultiByteSet_3, TYPE_STR);

PRINT_ONE_PARAM ( PAN_MULTIBYTE_SET_4 ,  XParam->szMultiByteSet_4, TYPE_STR);

PRINT_ONE_PARAM ( PAN_MULTIBYTE_SET_5 ,  XParam->szMultiByteSet_5, TYPE_STR);

PRINT_ONE_PARAM ( PAN_MULTIBYTE_CACHE , &XParam->ulMultiByteTotalSize , TYPE_LONG);
PRINT_ONE_PARAM ( PAN_LOG_START_COPY  , &XParam->lLogStartCopy , TYPE_LONG);

PRINT_ONE_PARAM ( PAN_RESERVED_REDO_SIZE,&XParam->lReservedRedoSize, TYPE_LONG);

PRINT_ONE_PARAM ( PAN_LOG_COPY_DISTANCE, &XParam->lLogCopyDistance , TYPE_LONG);

PRINT_ONE_PARAM ( PAN_DW_COPY_BOUNDARY , &XParam->lDwCopyBoundary , TYPE_LONG);

PRINT_ONE_PARAM ( PAN_DW_IO_AREA_SIZE  , &XParam->lDwIOAreaSize  , TYPE_LONG);

PRINT_ONE_PARAM ( PAN_DW_IO_AREA_FLUSH , &XParam->lDwIOAreaFlush  , TYPE_LONG);

PRINT_ONE_PARAM ( PAN_DW_LRU_TAIL_FLUSH, &XParam->lDwLRUTailFlush  , TYPE_LONG);

PRINT_ONE_PARAM ( PAN_COPY_IO_BLOCK_CNT, &XParam->ulDwIOBufSize , TYPE_LONG);

PRINT_ONE_PARAM ( PAN_SVP_1_CONV_FLUSH , &XParam->fSvpFstConvFlush, TYPE_BOOL );

PRINT_ONE_PARAM ( PAN_XPARAM_CHANGED ,   &XParam->ulXparamChanged, TYPE_LONG );

PRINT_ONE_PARAM ( PAN_OBJ_LOG_SIZE ,     &XParam->ulObjLogSize, TYPE_LONG );

PRINT_ONE_PARAM ( PAN_OBJ_LOG_SEGMENT_SIZE , &XParam->ulObjLogSegmentSize, TYPE_LONG );

PRINT_ONE_PARAM ( PAN_OBJ_LOG_IO_QUEUE , &XParam->ulObjLogIOQueuePages, TYPE_LONG );

PRINT_ONE_PARAM ( PAN_OBJ_ROLLBACK_CACHE , &XParam->ulObjLogRollbackCache, TYPE_LONG );

PRINT_ONE_PARAM ( PAN_MAXEVENTTASKS , &XParam->ulMaxEventTasks , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_MAX_BACKUP_TASKS , &XParam->ulMaxBackupTasks , TYPE_LONG );

PRINT_ONE_PARAM ( PAN_MAXGARBAGE_COLLECTOR , &XParam->ulMaxGarbageCollector , TYPE_LONG );

PRINT_ONE_PARAM ( "MAX_OBJECT_LOGWRITER" , &XParam->ulMaxObjectLogWriter , TYPE_LONG );
/******************************/
#if _WIN32 || OS2 
/******************************/

PRINT_ONE_PARAM ( PAN_KERNEL_ALLOC_POOL   , &XParam->ulKernelAllocPoolSize, TYPE_LONG);

PRINT_ONE_PARAM ( PAN_RTE_ALLOC_POOL      , &XParam->ulRTEAllocPoolSize, TYPE_LONG);

PRINT_ONE_PARAM ( PAN_FORMAT_DEVSPACE     , &XParam->fFormatDevspace, TYPE_BOOL);

PRINT_ONE_PARAM ( PAN_DYNAMIC_SERVER      , &XParam->fDynamicServerTasks, TYPE_BOOL);

PRINT_ONE_PARAM ( PAN_USE_MEM_ENHANCEMENT,  &XParam->fUseMemEnhancement, TYPE_BOOL);
PRINT_ONE_PARAM ( PAN_MEM_ENHANCEMENT_LIM,  &XParam->lMemEnhancementLimit, TYPE_LONG);

/******************************/
#else
/******************************/
PRINT_ONE_PARAM ( PAN_OPMSG1 , XParam->szOpMsg1, TYPE_STR );

PRINT_ONE_PARAM ( PAN_OPMSG2 , XParam->szOpMsg2, TYPE_STR );

PRINT_ONE_PARAM ( PAN_TASK_STACK_AREA ,XParam->fStackStack ? 
                                      "STACK" : "DATA" ,TYPE_STR);

PRINT_ONE_PARAM ( PAN_DYNAMIC_TASK_STACK ,&XParam->fDynTaskStack ,TYPE_BOOL);
#endif
return;
}

/*------------------------------*/

void eo92_compress_path ( char           *pszCompPathName,
                          char           *pszPathName,
                          teo00_Longuint  ulCompPathLen )
  {
  #undef  MF__
  #define MF__ MOD__"eo92_compress_path"
  teo00_Longuint           lPathLen;
  teo00_Longuint           lSplitPos;
  teo00_Longuint           lRest;

  DBGPAS;

  if ( ulCompPathLen < 4 )
    {
    pszCompPathName[0] = '\0';
    return;
    }

  lPathLen = (teo00_Longuint)strlen ( pszPathName );

  if ( lPathLen > (int)ulCompPathLen )
    {
    lSplitPos = (ulCompPathLen / 2) - 2;
    lRest     = ulCompPathLen - lSplitPos - 4;

    strncpy ( pszCompPathName, pszPathName, lSplitPos );
    pszCompPathName [ lSplitPos ] = '\0';
    strcat ( pszCompPathName, "...." );
    strcat ( pszCompPathName, pszPathName + lPathLen - lRest );
    }
  else
    {
    strcpy ( pszCompPathName, pszPathName );
    }

  return;
  }

/*===========================================================================*
 *  LOCAL/PRIVATE/PROTECTED FUNCTIONS (CODE)                                 *
 *===========================================================================*/

static void e92_print_one_param ( char *param_id, 
                                  char *param_wanted, 
                                  int   wanted_len,
                                  void *value, 
                                  char  type )
{
  #undef  MF__
  #define MF__ MOD__"e92_print_one_param"

  DBGPAS;

  if ( e92_strcont ( param_id, param_wanted , wanted_len ) )
    switch ( type )
    { case TYPE_STR  : pri ( "%-20s: %s\n", param_id , 
                           value ? (char* )value : "" ) ;
                 break ;
      case TYPE_LONG : pri ( "%-20s: %ld\n", param_id, *(tsp_int4*)value ) ;
                 break ;
      case TYPE_BOOL : pri ( "%-20s: %s\n", param_id , *(teo00_Bool *)value ? "YES" : "NO" ) ;
                 break ;
    }
 return ;
}

/*------------------------------*/

static int set_bool_param( teo00_Bool *act_param,
                           char *new_param ) 
{
#undef  MF__
#define MF__ MOD__"set_bool_param"
  int set_ok  = *new_param == 'y' || *new_param == 'Y' 
             || *new_param == 'n' || *new_param == 'N' ;

  if ( set_ok )
    *act_param = *new_param == 'y' || *new_param == 'Y' ;
  else 
    pri ( "parameter out of range! ( use : y | Y | n | N ) \n" );

  return ( set_ok ) ;
}

/*------------------------------*/

static int set_int_param( int   int_type , 
                          void *act_param, 
                          char *new_param, 
                          long  lower_limit, 
                          long  upper_limit ) 
{
#undef  MF__
#define MF__ MOD__"set_int_param"

  long new_val = atol ( new_param ) ;
  int set_ok  = /* new_val != *act_param && */
                new_val >= lower_limit && new_val <= upper_limit ;

  if ( set_ok )
    switch ( int_type )
      { case TYPE_INT2 : *( (tsp_int2 *) act_param ) = (tsp_int2 ) new_val ;
                         break ;
        case TYPE_INT4 : *( (tsp_int4 *) act_param ) = (tsp_int4 ) new_val ;
                         break ;
        case TYPE_LONG : *( (tsp_int4 *) act_param ) = (tsp_int4 ) new_val ;
                         break ;
      }
  else 
    { pri ( "parameter out of range! ( use : %d ... %d ) \n", 
                       lower_limit, upper_limit );
    }
  return ( set_ok ) ;
}
/*===========================================================================*/
#endif /* KERNEL */
/*===========================================================================*/

/*------------------------------*/

static void eo92ExpandMultipleTasks (  tsp00_Int4       NumTasks,
                                        char            *szTaskToken,
                                        tsp00_Int4      *TotalMultipleTasks,
                                        char            *szCluster,
                                        tsp00_Int4      *MaxClusterLen )

{
  #undef  MF__
  #define MF__ MOD__"eo92ExpandMultipleTasks"

  char  szNumOfXXTasks[20];


  if (NumTasks)
  {
    sql47_ltoa( (long )NumTasks, szNumOfXXTasks, sizeof(szNumOfXXTasks) );
    strcat ( szCluster, szNumOfXXTasks );
    strcat ( szCluster, szTaskToken);
    *TotalMultipleTasks -= NumTasks ;
    if ( *TotalMultipleTasks )
      strcat ( szCluster, ",");
  }
  return;
}

/*------------------------------*/

externC tsp00_Int4 
        eo92CreateTaskcluster (  TASK_CTRL           pFirstUKTTaskCtrl,
                                 char               *szCluster,
                                 tsp00_Int4          MaxClusterLen )
{
  #undef  MF__
  #define MF__ MOD__"eo92CreateTaskcluster"

  tsp00_Int4 NumOfDWTasks     = 0;
  tsp00_Int4 NumOfEVTasks     = 0;
  tsp00_Int4 NumOfServerTasks = 0;
  tsp00_Int4 NumOfUserTasks   = 0;
  tsp00_Int4 NumOfGCTasks     = 0;
  tsp00_Int4 NumOfBackUpTasks = 0;
  tsp00_Int4 NumOfOLTasks     = 0;
  tsp00_Int4 TotalTasks       = 0, NumMultipleTasks;
  TASK_CTRL                   pCurrTask;

  szCluster[0]      = '\0';

  for ( pCurrTask =  pFirstUKTTaskCtrl, TotalTasks = 0;
        pCurrTask != NULL;
        pCurrTask =  NEXT_TASK(pCurrTask), TotalTasks++ )
    {
    switch (TASK_TYPE(pCurrTask))
      {
      case TT_TI_EO00 : strcat ( szCluster, "TI"); break;
      case TT_AL_EO00 : strcat ( szCluster, "AL"); break;
      case TT_TW_EO00 : strcat ( szCluster, "TW"); break;
      case TT_CS_EO00 : strcat ( szCluster, "CS"); break;
      case TT_UT_EO00 : strcat ( szCluster, "UT"); break;
      case TT_OL_EO00 : NumOfOLTasks++          ; break;
      case TT_EV_EO00 : NumOfEVTasks++          ; break;
      case TT_DW_EO00 : NumOfDWTasks++          ; break;
      case TT_SV_EO00 : NumOfServerTasks++      ; break;
      case TT_US_EO00 : NumOfUserTasks++        ; break;
      case TT_GC_EO00 : NumOfGCTasks++          ; break;
      case TT_BUP_EO00: NumOfBackUpTasks++      ; break;
      default    : strcat ( szCluster, "??"); break;
      }

    if (( NEXT_TASK(pCurrTask) )     &&
        ( TASK_TYPE(pCurrTask) != TT_DW_EO00 ) &&
        ( TASK_TYPE(pCurrTask) != TT_SV_EO00 ) &&
        ( TASK_TYPE(pCurrTask) != TT_EV_EO00 ) &&
        ( TASK_TYPE(pCurrTask) != TT_US_EO00 ) &&
        ( TASK_TYPE(pCurrTask) != TT_GC_EO00 ) &&
        ( TASK_TYPE(pCurrTask) != TT_OL_EO00 ) &&
        ( TASK_TYPE(pCurrTask) != TT_BUP_EO00))
      strcat ( szCluster, ",");
    }

  NumMultipleTasks = NumOfEVTasks   + NumOfDWTasks + NumOfServerTasks +
                     NumOfUserTasks + NumOfGCTasks + NumOfBackUpTasks + NumOfOLTasks;


  eo92ExpandMultipleTasks ( NumOfOLTasks,  "*OL", &NumMultipleTasks, 
                            szCluster, &MaxClusterLen ) ;

  eo92ExpandMultipleTasks ( NumOfDWTasks,    "*DW", &NumMultipleTasks, 
                            szCluster, &MaxClusterLen ) ;

  eo92ExpandMultipleTasks ( NumOfGCTasks,    "*GC", &NumMultipleTasks, 
                            szCluster, &MaxClusterLen ) ;

  eo92ExpandMultipleTasks ( NumOfEVTasks,    "*EV", &NumMultipleTasks, 
                            szCluster, &MaxClusterLen ) ;

  eo92ExpandMultipleTasks ( NumOfBackUpTasks,"*BUP", &NumMultipleTasks, 
                            szCluster, &MaxClusterLen ) ;

  eo92ExpandMultipleTasks ( NumOfServerTasks,"*SV", &NumMultipleTasks, 
                            szCluster, &MaxClusterLen ) ;

  eo92ExpandMultipleTasks ( NumOfUserTasks,  "*US", &NumMultipleTasks, 
                            szCluster, &MaxClusterLen ) ;


  return ( TotalTasks ) ;
}
/*------------------------------*/

/* PTS 1114129 */
externC char *eo92GetConnectStateName(int connState, char *buffer, int len)
{
	if ( len < 80 )
	{
		return (char *)"buffer too small";
	}
	switch(connState)
	{
		case CON_UNUSED_EO003:
			sprintf(buffer, "unused[%d]", connState);
			break;
		case CON_INITIALIZED_EO003:
			sprintf(buffer, "initialized[%d]", connState);
			break;
		case CON_CONNECTING_EO003:
			sprintf(buffer, "connecting[%d]", connState);
			break;
		case CON_CONNECTED_EO003:
			sprintf(buffer, "connected[%d]", connState);
			break;
		case CON_REQUESTED_EO003:
			sprintf(buffer, "requested[%d]", connState);
			break;
		case CON_RECEIVED_EO003:
			sprintf(buffer, "received[%d]", connState);
			break;
		case CON_REPLIED_EO003:
			sprintf(buffer, "replied[%d]", connState);
			break;
		case CON_CANCELLED_EO003:
			sprintf(buffer, "cancelled[%d]", connState);
			break;
		case CON_RELEASED_EO003:
			sprintf(buffer, "released[%d]", connState);
			break;
		case CON_KEEP_ALIVE_EO003:
			sprintf(buffer, "keep alive[%d]", connState);
			break;
		case CON_INCOMING_EO003:
			sprintf(buffer, "incoming[%d]", connState);
			break;
		case CON_TIMEDOUT_EO003:
			sprintf(buffer, "timed out[%d]", connState);
			break;
		default:
			sprintf(buffer, "UNKNOWN[%d]", connState);
			break;
	}
	return buffer;
}

/*------------------------------*/

externC char *eo92GetConnectServiceName(int connService, char *buffer, int len)
{
	if ( len < 80 )
	{
		return (char *)"buffer too small";
	}

	switch(connService)
	{
        case srvUser_esp01:
			sprintf(buffer, "user task[%d]", connService);
			break;
        case srvAsyncUser_esp01:
			sprintf(buffer, "asynchron user task[%d]", connService);
			break;
        case srvUtility_esp01:
			sprintf(buffer, "utility task[%d]", connService);
			break;
        case srvDistribution_esp01:
			sprintf(buffer, "distribution[%d]", connService);
			break;
        case srvControl_esp01:
			sprintf(buffer, "data base manager[%d]", connService);
			break;
        case srvEvent_esp01:
			sprintf(buffer, "event task[%d]", connService);
			break;
        case srvStream_esp01:
			sprintf(buffer, "stream[%d]", connService);
			break;
		default:
			sprintf(buffer, "UNKNOWN[%d]", connService);
			break;
	}
	return buffer;
}

/*------------------------------*/

externC char *eo92GetConnectProtocolName(int connProtocol, char *buffer, int len)
{
	if ( len < 80 )
	{
		return (char *)"buffer too small";
	}
	switch(connProtocol)
	{
		case PROT_UNDEF_EO003:
			sprintf(buffer, "undefined[%d]", connProtocol);
			break;
		case PROT_SHM_EO003:
			sprintf(buffer, "shared memory[%d]", connProtocol);
			break;
		case PROT_BIGSHM_EO003:
			sprintf(buffer, "big shared memory[%d]", connProtocol);
			break;
		case PROT_SOCKET_EO003:
			sprintf(buffer, "socket[%d]", connProtocol);
			break;
		case PROT_NI_EO003:
			sprintf(buffer, "SAP network interface (NI)[%d]", connProtocol);
			break;
		case PROT_LOCAL_CONTROL_EO003:
			sprintf(buffer, "local control[%d]", connProtocol);
			break;
		default:
			sprintf(buffer, "UNKNOWN[%d]", connProtocol);
			break;
	}
	return buffer;
}

/*------------------------------*/

externC char *eo92GetSwapTypeName(int swapType, char *buffer, int len)
{
    switch ( swapType )
    {
    case 1:
		sprintf(buffer, "not swapped[%d]", swapType);
		break;
	case 2:
		sprintf(buffer, "swapped[%d]", swapType);
		break;
	case 3:
		sprintf(buffer, "half swapped[%d]", swapType);
		break;
    default:
		sprintf(buffer, "Illegal swap type[%d]", swapType);
		break;
    }
	return buffer;
}

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/
