/****************************************************************************

  module      : veo951.c

  -------------------------------------------------------------------------

  responsible : FrankH

  special area: application startup program
  description : param


  last changed: 2000-07-19  14:00
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1998



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp01.h"
#include "geo002xp.h"
#include "heo95.h"
#include "heo06.h"
#include "heo11.h"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#ifdef NO_INLINES
# define _INLINE
#else
# define _INLINE	__inline
#endif

#define PARAM_OPT_L "l"
#define PARAM_OPT_LL "ll"
#define PARAM_OPT_V "V"
#define PARAM_OPT_CLEAR   "CLEAR"
#define PARAM_BCHECK      "BCHECK"
#define PARAM_BCHECKC     "BCHECKC"
#define PARAM_BXHELP      "BXHELP"
#define PARAM_BINIT       "BINIT"
#define PARAM_NOCHECK     "NOCHECK"
#define PARAM_SHOWALL     "SHOWALL"
#define PARAM_NCP         "-ncp"
#define PARAM_NODE_ID     "-n"
#define PARAM_UID         "-u"
#define PARAM_DBID        "-d"
#define PARAM_GATEWAY_ID  "-g"
#define PARAM_RUNID       "-r"
#define PARAM_OPT_v       "-v"
#define PARAM_RC_BCHECKC 11
#define PARAM_RC_BATCH 4
#define PARAM_RC_REL24 3
#define PARAM_RC_QUIT  2
#define PARAM_RC_SAVE  0


#ifdef WIN32

#define PARAM_CHK   "xpcheck.prt"
#define PARAM_DBNAME_F "xparam.dbn"
#define PARAM_TMP   ""
#define LIST_PARAM_SHORT  "cmd /c dir /B"
#define LIST_PARAM_LONG   "cmd /c dir"
#define PARAM_CLEAR "del"

#else

#define PARAM_CHK   "xpcheck.prot"
#define PARAM_DBNAME_F "/tmp/xparam.dbname"
#define PARAM_TMP   "/tmp"
#define LIST_PARAM_SHORT  "ls"
#define LIST_PARAM_LONG   "ls -l"
#define PARAM_CLEAR "rm"

#endif

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/

typedef struct          xparam_args
                        { char *uid , *uid_val ;
                          char *dbid, *dbname;
                          char *gateway_id, *gateway;
                          char *node_id, *node;
                          char *rid,  *rundir ;
                          char *bchk, *bcheck ;
                          char *bini, *binit ;
                          char *showall, *npc, *opt_v ;
                          tsp00_Pathc param_dir ;
                        } XPARAM_ARGS ;

/*===========================================================================*
 *  GLOBAL VARIABLES                                                         *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL VARIABLES                                                          *
 *===========================================================================*/



/*===========================================================================*
 *  STATIC/INLINE FUNCTIONS (PROTOTYPES)                                     *
 *===========================================================================*/

#ifdef WIN32
static int e951_call_srvconf    ( XPARAM_ARGS *param_rec, 
                                  char        *srv_opt );
#endif

static int handle_param_retcode ( int          param_retcode, 
                                  XPARAM_ARGS *param_rec ) ;

static int filecopy             ( char        *src, 
                                  char        *dest );

static int exec_param_prog    ( PROG_ID       *param_prog, 
                                XPARAM_ARGS   *param_rec, 
                                int           *param_ret );

static int save_paramfile     ( XPARAM_ARGS   *param_rec );

static int fill_param_rec     ( XPARAM_ARGS   *param_rec, 
                                int            num_args, 
                                char         **args );

static int check_first_option ( PROG_ID       *param_prog, 
                                char         **options, 
                                XPARAM_ARGS   *param_rec ) ;

static int clear_paramfile    ( XPARAM_ARGS   *param_rec ) ;

static int check_param_env    ( XPARAM_ARGS   *param_rec, 
                                char          *tmp_dir ) ;

static int e951_getchar       () ;

static char *expand_env_vars  ( char *path );
/*===========================================================================*
 *  PUBLIC FUNCTIONS (CODE)                                                  *
 *===========================================================================*/

/*-----------------------------------------------------------*/
int e951_check_rundir ( char  *dbname, 
                        char **param_rundir, 
                        int    ask_for_creation )
/*-----------------------------------------------------------*/
{
  int check_ok ;

  DBG0 (( "e951_check_rundir for db <%s>\n", dbname ));
  check_ok = e951_getparam ( dbname, PAN_RUNDIR, param_rundir ) != CNULL;
  if ( check_ok )
    if ( !(check_ok = e951_f_access ( *param_rundir, S_IFDIR, FALSE ))
          && ask_for_creation )
      { printf ( "\nRun-directory %s is missing", *param_rundir );
        printf ( "\nDo you want that directory to be created ? (y/n): ");
#ifndef WIN32
        fflush ( stdout ) ;
#endif
        if ( (check_ok = (char )toupper ( e951_getchar() ) == 'Y') )
          { tsp01_RteError RteError ;
            printf ( "\nCreating %s ...", *param_rundir );
            if ( !(check_ok = sqlcreate_dirc( *param_rundir, &RteError )) )
              printf ( "\n%s\n%s", RteError.RteErrText, RteError.OsErrText ) ;
          }
        else
           printf ( "\nYou must do that BEFORE starting the database %s\n",
                                               dbname);
      }
  return ( check_ok ) ;
}

/*-----------------------------------------------------------*/
char *e951_getparam ( char  *dbname, 
                      char  *param_id, 
                      char **param )
/*-----------------------------------------------------------*/
{
  int check_ok;
  tsp01_RteError          RteError ;
  tsp00_C64c              ParamVal;

  DBG0 (( "e951_getparam for db <%s> called\n", dbname ));
  eo11GetParam ( dbname , param_id, ParamVal, &RteError ) ;

  if ( (check_ok = RteError.RteErrCode == RTE_NO_ERROR_SP01) )
  {
    expand_env_vars ( ParamVal ) ;
    *param = (char *) strdup ( ParamVal ) ;
    DBG0 (( "e951_getparam <%s>:\n", param_id,ParamVal ));
  }
  else
  {
    /* PTS 1107204 */
    DBG0 (( "\ne951_getparam: %s\n%s", RteError.RteErrText, RteError.OsErrText )) ;
    *param = NULL ;
  }

  return ( *param ) ;
}

/*-----------------------------------------------------------*/
int e951_f_access ( char *path, 
                    int   mode, 
                    int   pri_err )
/*-----------------------------------------------------------*/
{
  struct stat buf;
  int access_ok ;

  access_ok = stat ( path, &buf )  == 0 && (buf.st_mode & mode) == mode ;
  DBG0 (( "trying to access path <%s> with mode %02x\n", path, mode ));
  if ( !access_ok && pri_err )
    {
      perror ( "\nstat error" ) ;
      printf ( "can't access path  >%s<\n", path ) ;
    }

  return ( access_ok ) ;
}

/*-----------------------------------------------------------*/
int e951_call_param ( param, num_args, args )
/*-----------------------------------------------------------*/
PROG_ID *param;
int num_args ;
char **args ;
{
#if 0
  int call_ret, check_param_env(), check_first_option();
  int fill_param_rec(), handle_param_retcode();
  tsp00_Pathc tmp_dir;
  XPARAM_ARGS args_rec ;
#endif
  printf("\n'%s' not longer supported, use 'dbmcli' NOW !!\n",param->progid);
  return(0);
#if 0
  DBG0 (( "call_param called with %d parameter(s) \n", num_args ));
  memset ( &args_rec, '\0', sizeof ( XPARAM_ARGS ) ) ;
  if ( check_param_env (&args_rec, tmp_dir ) )
    { if ( num_args == 0 ||
           !check_first_option ( param, args, &args_rec ) )
        { if ( !fill_param_rec ( &args_rec, num_args, args ) )
            e950_call_help (param->progid, param->prog_helpfiles ); 
          else
            { int param_ret ;
              if ( (call_ret = exec_param_prog (param, &args_rec ,&param_ret)) )
                { handle_param_retcode ( param_ret, &args_rec ) ;
#ifdef WIN32
                  call_ret = param_ret >  PARAM_RC_BATCH 
                             /* && param_ret != PARAM_RC_BCHECKC  */
                             ? param_ret : 0;
#else
                  call_ret = (param_ret >> 8) >  PARAM_RC_BATCH 
                             /* && (param_ret >> 8) != PARAM_RC_BCHECKC */
                             ? param_ret : 0;
#endif
                }
            }
       }
      else
       call_ret = 0 ;
    }
  else
    call_ret = NOTOK ;
  return ( call_ret ) ;
#endif
}

/*-----------------------------------------------------------*/
char *e951_toupper_str ( char *str )
/*-----------------------------------------------------------*/
{
  char *base = str ;
  for ( ; *str != '\0'; str++ )
    if ( islower ( *str ) )
      *str = (char )toupper ( *str ) ;
  return ( base );
}

/*-----------------------------------------------------------*/
int e951_create_path ( char          *path_to_create, 
                       unsigned short mode )
/*-----------------------------------------------------------*/
{
  int create_ok = TRUE ;
  char *act_pos = path_to_create , *end_pos;

#ifdef WIN32
     if ( act_pos[1] == ':' ) act_pos += 2;
#endif
  if ( *act_pos == PATH_DELIMITER_EO01 ) act_pos++ ;

  do
    { if ( (end_pos = (char *)strchr ( act_pos, PATH_DELIMITER_EO01 )) != CNULL )
        *end_pos = '\0' ;
      if ( !e951_f_access ( path_to_create , S_IFDIR, FALSE ) )
        { DBG0 (( "\ncreating directory <%s>", path_to_create ));
#ifdef WIN32
          if ( !(create_ok = CreateDirectory ( path_to_create, NULL ) != 0 ) )
#else
          if ( !(create_ok = mkdir ( path_to_create, mode ) == 0 ) )
#endif
            perror ( "can't create directory" ) ;
        }
      if ( end_pos != CNULL )
        { *end_pos = PATH_DELIMITER_EO01; act_pos = end_pos+1 ; }
    }
  while ( create_ok && end_pos != CNULL ) ;
  return ( create_ok ) ;
}


/*===========================================================================*
 *  LOCAL/PRIVATE/PROTECTED FUNCTIONS (CODE)                                 *
 *===========================================================================*/
/*-----------------------------------------------------------*/
static char *expand_env_vars ( char *path )
/*-----------------------------------------------------------*/
{
  int len ;
  tsp00_Pathc TmpTmpFile,exp_path ;
  char *beg_env_var, *end_env_var, *env_var_val,
       *act_pos, *exp_path_pos;

  DBG0 (( "expand_env_vars: before expand <%s>\n", path ));
  
  do                                             /* PTS 1000713 */
  { 
    beg_env_var  = end_env_var = CNULL ;
    act_pos      = path;
    exp_path_pos = exp_path ;

    strcpy( TmpTmpFile, path ) ;
    while ( ( beg_env_var = strchr ( act_pos      , BEG_ENV_VAR ) ) != CNULL &&
          ( end_env_var = strchr ( beg_env_var+1, END_ENV_VAR ) ) != CNULL )
    { *end_env_var = '\0' ;
      if ( ( env_var_val = (char *)getenv ( beg_env_var+1 ) ) != CNULL )
        { len = (int ) (beg_env_var - act_pos) ;
#ifndef WIN32
          if ( * ( beg_env_var - 1 ) == '$' ) len-- ;
#endif
          if ( len > 0 )
            strncpy ( exp_path_pos, act_pos, len ) ;
          strcpy ( exp_path_pos + len, env_var_val );
          exp_path_pos += len + strlen ( env_var_val ) ;
          act_pos = end_env_var + 1 ;
        }
    }

    if ( end_env_var != CNULL || beg_env_var != CNULL )
    {
      strcpy ( exp_path_pos, act_pos ) ;
      strcpy ( path, exp_path ) ;
    }
    
  }
  while ( strcmp( TmpTmpFile, path ) != 0 ) ;

  DBG0 (( "expand_env_vars: after expand <%s>\n", path ));
  return ( path ) ;
}


#ifdef WIN32
/*-----------------------------------------------------------*/
static int e951_call_srvconf ( XPARAM_ARGS *param_rec, 
                               char        *srv_opt )
/*-----------------------------------------------------------*/
{
  int rc, num_params = 0 ;
  char *params[5] ;
  if ( param_rec->node )
    { params[num_params] = PARAM_NODE_ID; num_params++ ;
      params[num_params] = param_rec->node     ; num_params++ ;
    }
  if ( param_rec->dbname )
    { params[num_params] = PARAM_DBID         ; num_params++ ;
      params[num_params] = param_rec->dbname  ; num_params++ ;
    }
  if ( *srv_opt )
    { params[num_params] = srv_opt            ; num_params++ ;
    }
  rc = e950_search_and_call_sqlprog ( "x_service", num_params, params ) ;
  return ( rc ) ;
}

#endif

/*-----------------------------------------------------------*/
static int e951_getchar()
/*-----------------------------------------------------------*/
{
#ifdef WIN32
  HANDLE std_in = GetStdHandle ( STD_INPUT_HANDLE ) ;
  char   byte ;
  long   bytes_read ;

  ReadFile ( std_in, &byte, 1, &bytes_read, NULL);

  return ( byte );
#else
  return ( getchar() );
#endif

}

/*-----------------------------------------------------------*/
static int check_param_env ( XPARAM_ARGS *param_rec, 
                             char        *tmp_dir )
/*-----------------------------------------------------------*/

{
  int check_ok ;
  tsp01_RteError RteError ;

  DBG0 (( "check_param_env called ...\n" ));

  check_ok = sqlGetIndependentConfigPath( param_rec->param_dir,
	                                      TERM_WITHOUT_DELIMITER_EO01,&RteError);
  if ( check_ok )
    {
      strcpy ( tmp_dir, PARAM_TMP ) ;
      check_ok = ( *tmp_dir == '\0'    ||
                   e951_f_access ( tmp_dir, F_EX, TRUE ) ) &&
                 ( *param_rec->param_dir == '\0'  ||
                   e951_f_access ( param_rec->param_dir, F_EX, TRUE ) );
    }
  return ( check_ok ) ;
}

/*-----------------------------------------------------------*/
static int exec_dir_cmd ( dir_cmd, param_dir )
/*-----------------------------------------------------------*/
char *dir_cmd ;
char *param_dir ;
{
  int rc ;
  char sysbuf[255] ;
  sprintf ( sysbuf, "%s %s", dir_cmd, param_dir ) ;
  DBG0 (( "exec_dir_cmd: %s\n", sysbuf ));
  rc = system ( sysbuf ) ;
  return ( rc ) ;
}

/*-----------------------------------------------------------*/
static int clear_paramfile ( XPARAM_ARGS *param_rec )
/*-----------------------------------------------------------*/
{
  int clear_ok ;
  tsp00_Pathc paramfilename  ;
  sprintf ( paramfilename, "%s%c%s", param_rec->param_dir, PATH_DELIMITER_EO01, 
                           param_rec->dbname );
  if ( e951_f_access ( paramfilename,  F_EX, FALSE ) )
    { printf ( "\nDo you really want to delete your" ) ;
#ifdef WIN32
      e951_toupper_str ( param_rec->dbname ) ;
#endif
      printf ( "\nXPARAM file named %s ? (y/n): ", param_rec->dbname);
#ifndef WIN32
      fflush ( stdout ) ;
#endif
      if ( (char )toupper ( e951_getchar() ) == 'Y' )
        { if ( (clear_ok = unlink ( paramfilename ) == 0) )
            printf ( "\n%s deleted\n", paramfilename ) ;
          else
            { char buf[80];
              sprintf ( buf, "Paramfile deletion for %s failed\n",
                       paramfilename);
              perror ( buf ) ;
            }
#ifdef WIN32
          if ( clear_ok )
            clear_ok = e951_call_srvconf ( param_rec, "-r" ) == 0 ;
#endif
        }
     }
   else
     printf ( "\nParamfile for serverdb <%s> not found\n", param_rec->dbname ) ;
  return ( clear_ok ) ;
}

/*-----------------------------------------------------------*/
static int check_first_option( PROG_ID     *param_prog, 
                               char       **options, 
                               XPARAM_ARGS *param_rec )
/*-----------------------------------------------------------*/
{
  int exec_dir_cmd(), clear_paramfile();
  int opt_found = IS_OPT_CHR ( *(options[0]) ) ;
  DBG0 (( "check_first_option for param ( %s / %s / % s)\n",
           PARAM_OPT_L, PARAM_OPT_V, PARAM_OPT_CLEAR ));

  if ( opt_found )
    if ( (opt_found = !strcmp ( &options[0][1], PARAM_OPT_L )) )
      exec_dir_cmd ( LIST_PARAM_SHORT, param_rec->param_dir ) ;
    else
      if ( (opt_found = !strcmp ( &options[0][1], PARAM_OPT_LL )) )
        exec_dir_cmd ( LIST_PARAM_LONG, param_rec->param_dir ) ;
      else
        if ( (opt_found = !strcmp ( &options[0][1], PARAM_OPT_V )) )
          e950_call_component ( param_prog, 1, options ) ;
        else
          if ( (opt_found = !strcmp ( &options[0][1], PARAM_OPT_CLEAR )) )
            if ( options[1] != CNULL )
              { param_rec->dbid = PARAM_DBID; param_rec->dbname=options[1] ;
                clear_paramfile ( param_rec ) ;
              }
            else
              USAGE_PARAM_CLEAR(param_prog->progid)

  DBG0 (( "check_first_option: opt_found = %s\n",
                                 opt_found ? "TRUE" : "FALSE" )) ;
  return ( opt_found ) ;
}

/*-----------------------------------------------------------*/
static int fill_param_rec ( XPARAM_ARGS *param_rec, 
                            int          num_args, 
                            char       **args )
/*-----------------------------------------------------------*/
{
  int fill_ok = TRUE , act_arg = 0 ;
  while ( fill_ok && num_args )
    { if ( IS_OPT_CHR ( *(args[act_arg]) ) )
        { if ( !strcmp ( &args[act_arg][1], &PARAM_UID[1] ) )
            { if ( (fill_ok = num_args > 1) )
                { param_rec->uid = PARAM_UID;
                  param_rec->uid_val = args[act_arg+1] ;
                  num_args -= 2, act_arg+=2 ;
                  DBG0(("uid: %s %s\n",param_rec->uid,param_rec->uid_val ));
                }
            }
          else
          if ( !strcmp ( &args[act_arg][1], &PARAM_DBID[1] ) )
            { if ( (fill_ok = num_args > 1) )
                { param_rec->dbid = PARAM_DBID;
                  param_rec->dbname = args[act_arg+1] ;
                  num_args -= 2, act_arg+=2 ;
                  DBG0(("db: %s %s\n", param_rec->dbid, param_rec->dbname ));
                }
            }
          else
          if ( !strcmp ( &args[act_arg][1], &PARAM_GATEWAY_ID[1] ) )
            { if ( (fill_ok = num_args > 1) )
                { param_rec->gateway_id = PARAM_GATEWAY_ID;
                  param_rec->gateway = args[act_arg+1] ;
                  num_args -= 2, act_arg+=2 ;
                  DBG0(("gateway: %s %s\n", 
                        param_rec->gateway_id, param_rec->gateway));
                }
            }
          else
#ifdef WIN32
          if ( !strcmp ( &args[act_arg][1], &PARAM_NODE_ID[1] ) )
            { if ( (fill_ok = num_args > 1) )
                { param_rec->node_id = PARAM_NODE_ID;
                  param_rec->node = args[act_arg+1] ;
                  num_args -= 2, act_arg+=2 ;
                  DBG0(("db: %s %s\n", param_rec->node_id, param_rec->node ));
                }
            }
          else
#endif
          if ( !strcmp ( &args[act_arg][1], &PARAM_RUNID[1] ) )
            { if ( (fill_ok = num_args > 1) )
                { param_rec->rid = PARAM_RUNID;
                  param_rec->rundir = args[act_arg+1] ;
                  num_args -= 2, act_arg+=2 ;
                  DBG0(("rundir: %s %s\n",param_rec->rid, param_rec->rundir ));
                }
            }
          else
          if ( !strcmp ( &args[act_arg][1], &PARAM_NCP[1] ) )
            { param_rec->npc = PARAM_NCP ;
              --num_args, ++act_arg ;
            }
          else
          if ( !strcmp ( args[act_arg], PARAM_OPT_v ) )
            { param_rec->opt_v = PARAM_OPT_v ;
              --num_args, ++act_arg ;
              DBG0(("option: %s set\n",param_rec->opt_v ));
            }
          else fill_ok = FALSE ;
        }
      else
      if ( !strcmp ( args[act_arg], PARAM_SHOWALL ) )
        { param_rec->showall = PARAM_SHOWALL ;
          --num_args, ++act_arg ;
        }
      else
      if ( !strcmp ( args[act_arg], PARAM_NOCHECK ) )
        { param_rec->bchk = param_rec->bcheck = PARAM_NOCHECK ;
          --num_args, ++act_arg ;
        }
      else
      if ( !strcmp ( args[act_arg], PARAM_BCHECK ) )
        { param_rec->bchk = param_rec->bcheck = PARAM_BCHECK ;
          --num_args, ++act_arg ;
        }
      else
      if ( !strcmp ( args[act_arg], PARAM_BCHECKC ) )
        { param_rec->bchk = param_rec->bcheck = PARAM_BCHECKC ;
          --num_args, ++act_arg ;
        }
      else
      if ( !strcmp ( args[act_arg], PARAM_BXHELP ) )
        { param_rec->bchk = param_rec->bcheck = PARAM_BXHELP ;
          --num_args, ++act_arg ;
        }
      else
      if ( !strcmp ( args[act_arg], PARAM_BINIT ) )
        { param_rec->bini = param_rec->binit = PARAM_BINIT ;
          --num_args, ++act_arg ;
        }
      else fill_ok = FALSE ;
    }
  if ( fill_ok )
    { if ( !param_rec->dbid ) param_rec->dbname = e950_get_dbname_from_env () ;
      if ( ! (fill_ok = !param_rec->dbname ||
               (unsigned int ) strlen (param_rec->dbname) <= sizeof(tsp00_DbName) ) )
        printf("\n'%s' too long! limit is %d characters!\n",
                   param_rec->dbname, sizeof(tsp00_DbName) ) ;
#ifdef WIN32
      if ( param_rec->dbname != CNULL ) 
        e951_toupper_str ( param_rec->dbname ) ;
#endif
    }
  return ( fill_ok ) ;
}

/*-----------------------------------------------------------*/
static int save_paramfile ( XPARAM_ARGS *param_rec )
/*-----------------------------------------------------------*/
{
  int save_ok = TRUE;
  DBG0 (( "save_paramfile for db <%s>called ...\n", param_rec->dbname ));
  if ( param_rec->dbname != CNULL && *param_rec->dbname != '\0' )
    { tsp00_Pathc full_paramfile;
      sprintf ( full_paramfile, "%s%c%s", param_rec->param_dir,
                                 PATH_DELIMITER_EO01, param_rec->dbname ) ;
      if ( e951_f_access ( full_paramfile, F_EX, FALSE ) )
        if ( !e951_f_access ( full_paramfile, F_RW , FALSE) )
          { save_ok = FALSE ;
            perror ( "\naccess failed ") ;
            printf ( "\ncan't access paramfile <%s>\n", full_paramfile) ;
          }
     }
  return ( save_ok ) ;
}

/*-----------------------------------------------------------*/
static int exec_param_prog ( PROG_ID     *param_prog, 
                             XPARAM_ARGS *param_rec, 
                             int         *param_ret )
/*-----------------------------------------------------------*/
{
  int exec_ret=TRUE, num_args=0 ;
  char *args[20] ;

  DBG0 (( "exec_param_prog called ...\n" ));

  if ( param_rec->dbid != CNULL && param_rec->gateway_id != CNULL ) 
    { exec_ret = FALSE ;
      printf("\neither %s or %s allowed!\n" ,
               param_rec->dbid , param_rec->gateway_id ) ;
      return ( exec_ret ) ;
    }
  if ( param_rec->uid != CNULL )
    { args[num_args] = param_rec->uid ;
      args[num_args+1] = param_rec->uid_val ;
      num_args+=2 ;
    }
  if ( param_rec->rid == CNULL )
    { tsp00_Pathc std_rundir ;
      if ( (exec_ret = sqlGetWrkPath ( std_rundir,
                                       TERM_WITHOUT_DELIMITER_EO01 ) != CNULL) )
        param_rec->rundir = (char *)strdup ( std_rundir ) ;
      else
        printf("\ncan't build std_rundir_path!\n" ) ;
    }
  if ( param_rec->dbid != CNULL )
    { args[num_args] = param_rec->dbid ; args[num_args+1] = param_rec->dbname ;
      num_args+=2 ;
    }
  if ( param_rec->gateway_id != CNULL )
    { args[num_args]   = param_rec->gateway_id ; 
      args[num_args+1] = param_rec->gateway ;
      num_args+=2 ;
    }
  if ( param_rec->opt_v != CNULL )
    { args[num_args] = param_rec->opt_v; ++num_args ; }
  args[num_args] = param_rec->rundir ;
  num_args++ ;

  if ( param_rec->bini != CNULL )
    { args[num_args] = param_rec->binit; ++num_args ; 
      param_prog->init_term = FALSE ;
    }
  if ( param_rec->bchk != CNULL )
    { args[num_args] = param_rec->bcheck; ++num_args ; 
      param_prog->init_term = FALSE ;
    }
  if ( param_rec->showall != CNULL )
    { args[num_args] = param_rec->showall; ++num_args ; 
    }
  if ( param_rec->npc != CNULL )
    { args[num_args] = param_rec->npc; ++num_args ; 
    }

  if ( debug ) e950_print_params ( "exec_param_prog", num_args, args ) ;
  if ( exec_ret )
     { int save_paramfile() ;
       unlink ( PARAM_DBNAME_F ) ;
       if ( (exec_ret = param_rec->dbid == CNULL ||
             save_paramfile ( param_rec ) ) )
         *param_ret = e950_call_component ( param_prog, num_args, args ) ;
     }
  return ( exec_ret ) ;
}

/*-----------------------------------------------------------*/
static int get_srcfile_time ( srcfile, act_time, mod_time )
/*-----------------------------------------------------------*/
char *srcfile ;
time_t *act_time ;
time_t *mod_time ;
{
  struct stat buf ;
  int get_ok = stat ( srcfile, &buf ) == 0 ;

  if ( get_ok )
    { *act_time = buf.st_atime ; *mod_time = buf.st_mtime ; }
  else
      *act_time = *mod_time = (time_t )NULL ;

  return ( get_ok ) ;
}

/*-----------------------------------------------------------*/
static int filecopy ( char *src, 
                      char *dest )
/*-----------------------------------------------------------*/
{
  boolean copy_ok ;
  tsp01_RteError RteError ;

  copy_ok = sqlfilecopyc (  src, dest, &RteError);
  DBG0 (( "filecopy: <%s> ==> <%s> called: %s\n", 
           src, dest, copy_ok ? "OK" : "FAILED" ));
  if ( !copy_ok  )
    printf ( "\n%s\n%s\n", RteError.RteErrText, RteError.OsErrText ) ;
  return ( copy_ok ) ;
}

/*-----------------------------------------------------------*/
static int handle_param_retcode ( int          param_retcode, 
                                  XPARAM_ARGS *param_rec )
/*-----------------------------------------------------------*/
{
  tsp00_Pathc ParamCHK ;
  int filecopy() ;
  FILE *stream = fopen ( PARAM_DBNAME_F, "r" ) ;

  if ( ( param_retcode & 0xFF ) == 0 ) param_retcode >>= 8 ;
  DBG0 (( "handle_param_retcode: %d\n", param_retcode ));
  if ( stream )
    { char line [ MAX_DBNAME_LEN + 1 ], *pos;
      fgets ( line, MAX_DBNAME_LEN+1, stream );
      if ( ( pos = (char *)strchr ( line, ' ' ) ) ) *pos = '\0';
      if ( line [ strlen(line) -1 ] == '\n' )
        line [ strlen(line) - 1 ] = '\0' ;
      param_rec->dbname = (char *) strdup ( line ) ;
      DBG0 (( "new DBNAME is <%s>\n", param_rec->dbname));
      fclose ( stream ) ;
      unlink ( PARAM_DBNAME_F ) ;
    }
  if ( param_rec->dbname != CNULL )
    { 
      if ( param_retcode == PARAM_RC_SAVE || param_retcode == PARAM_RC_BATCH ||
           param_retcode == PARAM_RC_BCHECKC )
        { int ok ;
          if ( param_retcode  == PARAM_RC_SAVE )
            { 
              ok = e951_check_rundir ( param_rec->dbname,
                                      &param_rec->rundir, TRUE ) ;
            }
          else
            ok = e951_check_rundir ( param_rec->dbname,
                                    &param_rec->rundir, FALSE ) &&
                 e951_f_access ( PARAM_CHK, F_EX, FALSE ) ;
          if ( ok )
#ifdef WIN32
            if ( stricmp ( param_rec->rundir , getcwd ( ParamCHK, sizeof(tsp00_Pathc) )) )
#else
            if ( strcmp ( param_rec->rundir , getcwd ( ParamCHK, sizeof(tsp00_Pathc) )) )
#endif
              { sprintf ( ParamCHK, "%s%c%s", param_rec->rundir,
                                         PATH_DELIMITER_EO01, PARAM_CHK ) ;
                DBG0 (( "moving <%s> to <%s>\n", PARAM_CHK, ParamCHK));
                if ( filecopy ( PARAM_CHK, ParamCHK ) )
                  unlink ( PARAM_CHK ) ;
              }
#ifdef WIN32
          ok = e951_call_srvconf ( param_rec, "-u" ) == 0 ;
#endif
        }
      else
        unlink ( PARAM_CHK ) ;
    }
  return ( TRUE ) ;
}

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/