/****************************************************************************

  module      : veo953.c

  -------------------------------------------------------------------------

  responsible : FrankH

  special area: application startup program
  description : other applicaione (xcontrol, xkernprot, ... )


  last changed: 1999-03-15  11:40
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1998



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "heo95.h"
#include "heo12_1.h"
#include "geo002xp.h"
#include "gsp09.h"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define CNTRL_BATCH_ID    "-b"
#define CNTRL_UID         "-u"
#define CNTRL_DBID        "-d"
#define CNTRL_GWID        "-g"
#define CNTRL_NODE_ID     "-n"
#define CNTRL_CLR_ID      "-c"
#define CNTRL_VERS_ID     "-V"
#define CNTRL_VERSION     "@Version@"
#define CNTRL_BATCH       "BATCH"
#define CNTRL_TRACE       "TRACE"

#define CNTRL_RUN "control.run"
#define CNTRL_LCK "control.lck"

#define KERNPROT_DBID        "-d"
#define KERNPROT_COPYTRACE   "-c"
#define KERNPROT_USE_FILE    "-f"
#define DIAGNOSE_KERNPROT    "KERNPROT"
#define DIAGNOSE_COPYTRACE   "COPYTRACE"

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/

typedef struct          xcontrol_args
                        { char *uname ;
                          char *u_pwd ;
                          char *db_gw_id  ;
                          char *dbname;
                          char *dbnode ;
                          char *trace ;
                          char **batch_args ;
                          int  num_batch_args ;
                          int   clr_lock ;
                        } XCONTROL_ARGS ;

typedef struct          xkernprot_args
                        {
                          char *dbname;
                          char *cp_tr_id  ;
                          char *cp_trace_pages ;
                          char *use_inp_file ;
                          tsp00_Pathc inp_file  ;
                          tsp00_Pathc outp_file ;
                          char *switches;
                        } XKERNRPOT_ARGS ;

/*===========================================================================*
 *  GLOBAL VARIABLES                                                         *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL VARIABLES                                                          *
 *===========================================================================*/



/*===========================================================================*
 *  STATIC/INLINE FUNCTIONS (PROTOTYPES)                                     *
 *===========================================================================*/

static int clr_cntrl_lock ( char *dbname ) ;

static int fill_cntrl_rec ( XCONTROL_ARGS *cntrl_rec,
   					        int            num_args,
							char         **args ) ;

static int set_inp_outp_filename ( XKERNRPOT_ARGS *kernprot_rec ) ;

static int fill_kernprot_rec ( XKERNRPOT_ARGS *kernprot_rec,
   					           int             num_args,
					           char          **args );

/*===========================================================================*
 *  PUBLIC FUNCTIONS (CODE)                                                  *
 *===========================================================================*/
/*-----------------------------------------------------------*/
int e953_call_domain ( PROG_ID *domain,
					   int      num_args,
					   char   **args )
/*-----------------------------------------------------------*/
{
  int i, num_params, call_ret ;
  char **params = (char **)malloc ( ( num_args+2 ) * sizeof (char *)) ;

  DBG0 (( "\ncall_domain called with %d parameters\n", num_args ));
  if ( params != (char **) NULL )
    {
      tsp00_Pathc     PgmPath ;

      params[0] = "-R"; params[1] = "DOMAIN.SYSDD.START" ;
      for ( i=0, num_params=2 ; i < num_args ; i++, num_params++ )
        params [ num_params ] = args [ i ] ;

      if ( en950_GetProgramExecPath ( domain->ProgramLocation, PgmPath ) )
        { int i, found ;
          tsp00_Pathc test_path ;
          for ( i=0, found=FALSE; !found && i<3 ; i++ )
            { sprintf ( test_path, "%s%c%s", PgmPath, PATH_DELIMITER_EO01, domain->progname ) ;
              if ( ! ( found = e951_f_access ( test_path, F_EX, FALSE ) ) )
                domain->progname += strlen ( domain->progname ) + 1 ;
            }
          call_ret = !found ? NOTOK :
                     e950_call_component ( domain, num_params, params ) ;
        }
      else
        call_ret = NOTOK;
      free ( (char **)params ) ;
    }
  else
    { perror ( "\nmalloc error" ) ; call_ret = NOTOK ; }

  return ( call_ret ) ;
}

/*-----------------------------------------------------------*/
int e953_call_control ( PROG_ID *control,
 					    int      num_args,
						char   **args )
/*-----------------------------------------------------------*/
{
  int call_ret, fill_cntrl_rec() ;
  XCONTROL_ARGS args_rec ;
  DBG0 (( "\ncall_control called with %d parameters\n", num_args ));

  if ( !fill_cntrl_rec ( &args_rec, num_args, args ) )
    { e950_call_help (control->progid, control->prog_helpfiles );
      call_ret = NOTOK ;
    }
  else
    { char **params = (char **)malloc ( ( 6 + 2 + args_rec.num_batch_args )
                      * sizeof (char *)) ;
      tsp00_Pathc     PgmPath ;

      if ( params != (char **) NULL )
        { int num_params = 2;
          if ( args_rec.dbname && *args_rec.dbname != '\0' )
            { params[num_params++] = args_rec.db_gw_id ;
              params[num_params++] = args_rec.dbname ;
            }
          else
            { params[num_params++] = "_"  ;
              params[num_params++] = "_"  ;
            }
          if ( args_rec.uname  ) params[num_params++] = args_rec.uname  ;
          else                   params[num_params++] = "_"  ;
          if ( args_rec.u_pwd  ) params[num_params++] = args_rec.u_pwd  ;
          else                   params[num_params++] = "_"  ;
          if ( args_rec.dbnode ) params[num_params++] = args_rec.dbnode ;
          else                   params[num_params++] = "_"  ;
          if ( args_rec.trace  ) params[num_params++] = args_rec.trace  ;
          if ( args_rec.batch_args )
            { int i ;
              params[num_params++] = CNTRL_BATCH  ;
              for ( i=0; i < args_rec.num_batch_args ; i++ )
                params[num_params++] = args_rec.batch_args[i] ;
            }

          if ( args_rec.clr_lock && args_rec.dbname )
              clr_cntrl_lock ( args_rec.dbname ) ;

          if ( en950_GetProgramExecPath ( control->ProgramLocation, PgmPath ) )
            { sprintf ( PgmPath, "%s%c%s", PgmPath, PATH_DELIMITER_EO01, "control") ;
              if ( args_rec.batch_args ||
                   ( args_rec.dbname &&
                     strcmp (args_rec.dbname, CNTRL_VERSION ) == 0 )
                 )
                params[0] = "-b" ;
              else
                params[0] = "-r" ;
              params[1] = PgmPath ;
              call_ret = e950_call_component ( control, num_params, params ) ;
            }
          else
            call_ret = NOTOK;
          free ( (char **)params ) ;
        }
      else
        { perror ( "\nmalloc error" ) ; call_ret = NOTOK ; }
    }
  return ( call_ret ) ;
}

/*-----------------------------------------------------------*/
int e953_call_python ( PROG_ID *python,
 					     int      num_args,
						 char   **args )
/*-----------------------------------------------------------*/
{
#if defined (_WIN32)
  #define PATHVAR_DELIMITER ";"
#else
  #define PATHVAR_DELIMITER ":"
#endif
  int             call_ret = 1;
  tsp01_RteError  RteError;
  boolean         GetOk;
  tsp00_Pathc     DBRootPgm;
  int             DBRootPgmLen;
  char          * endPath;
  static const char    * PythonPathEnvVar = "PYTHONPATH";
  static const char    * PythonHomeEnvVar = "PYTHONHOME";
  const char    * OldPythonPath = getenv (PythonPathEnvVar);
  const char    * libPath1 = "lib/python";
  const char    * libPath2 = "misc";
  int             newPathLen;
  char          * newPathBuf;
#if !defined (_WIN32)
  int             newHomeLen;
  char          * newHomeBuf;
#endif

  DBG0 (( "\ncall_python called with %d parameters\n", num_args ));


  GetOk = sqlGetDbrootPgmPath( DBRootPgm, TERM_WITHOUT_DELIMITER_EO01, &RteError ) ;
  if (!GetOk) {
      perror ( "\nsqlGetDbrootPgmPath error" );
      return call_ret;
  }
  endPath = DBRootPgm + strlen (DBRootPgm) - 1;
  while ((*endPath != PATH_DELIMITER_EO01)) {
      --endPath;
  }
  DBRootPgmLen = endPath - DBRootPgm + 1;
  DBRootPgm [DBRootPgmLen] = '\0';
  /*
   * set python path
   */
  newPathLen = (int)(strlen (PythonPathEnvVar) + 1 + (2 * DBRootPgmLen)
    + strlen (libPath1) + 1 + strlen (libPath2));
  if (OldPythonPath != NULL) {
      newPathLen += (int)(1 + strlen (OldPythonPath));
  }
  newPathBuf = malloc (newPathLen + 1);
  if (newPathBuf == NULL) {
      perror ( "\nmalloc error" );
      return call_ret;
  }
  sprintf (newPathBuf, "%s=%s%s" PATHVAR_DELIMITER "%s%s", PythonPathEnvVar,
      DBRootPgm, libPath1, DBRootPgm, libPath2);
  if (OldPythonPath != NULL) {
      strcat (newPathBuf, PATHVAR_DELIMITER);
      strcat (newPathBuf, OldPythonPath);
  }
  putenv (newPathBuf);
  /*
   * set PYTHONHOME
   */
#if !defined (_WIN32)
  newHomeLen = strlen (PythonHomeEnvVar) + 1 + strlen (DBRootPgm);
  newHomeBuf = malloc (newHomeLen + 1);
  sprintf (newHomeBuf, "%s=%s", PythonHomeEnvVar, DBRootPgm);
  putenv (newHomeBuf);
#endif
  /*
   * set ld_load_path
   *    not necessary, python is linked statically on UNIX
   */
  /*
   * call python executable
   */
  if ( call_ret != NOTOK) {
      call_ret = e950_call_component ( python, num_args, args ) ;
  }
  return ( call_ret ) ;
}
/* PTS 1108471 */
/*-----------------------------------------------------------*/
int e953_call_tracesort ( PROG_ID *tracesort,
 					      int      num_args,
						  char   **args )
/*-----------------------------------------------------------*/
{
  int call_ret =1, num_kernprot_args = 0 ;

  call_ret = e950_call_component ( tracesort, num_args, args ) ;
  return ( call_ret ) ;
}

/*-----------------------------------------------------------*/
int e953_call_kernprot ( PROG_ID *kernprot,
 					     int      num_args,
						 char   **args )
/*-----------------------------------------------------------*/
{
  int call_ret =1, num_kernprot_args = 0 ;
  XKERNRPOT_ARGS args_rec ;
  char *kernprot_args [ 10 ] ;

  DBG0 (( "\ncall_kernprot called with %d parameters\n", num_args ));

  if ( num_args == 0 )
  {
    num_kernprot_args = 2 ; /* -d KERNPROT  */
    kernprot_args[0]  = KERNPROT_DBID ;
    kernprot_args[1]  = DIAGNOSE_KERNPROT ;
  }
  else
    if ( !fill_kernprot_rec ( &args_rec, num_args, args ) )
    {
      e950_call_help (kernprot->progid, kernprot->prog_helpfiles );
      call_ret = NOTOK ;
    }
  else
    {
      if ( set_inp_outp_filename ( &args_rec ) )
      {
        if ( args_rec.inp_file[0] != '\0' )
        {
          if ( e951_f_access ( args_rec.inp_file, F_EX, TRUE ) )
          {
            /* PTS 1108471 / PTS 1113500
			if ( !args_rec.use_inp_file ) */
			{
     		  char *TraceSortArgs[2] ;
	    	  tsp00_Pathc SortTraceOutpFile ;

			  sprintf(SortTraceOutpFile, "%s.dat", args_rec.inp_file );
              TraceSortArgs[0] = args_rec.inp_file ;
              TraceSortArgs[1] = SortTraceOutpFile ;
              call_ret = e950_search_and_call_sqlprog ( "xtracesort",  2, TraceSortArgs ) ;
			  if ( call_ret != 0  ) 
     			return (NOTOK) ;
			  else
                 strcpy( args_rec.inp_file, SortTraceOutpFile);
			}

            num_kernprot_args += 4 ; /* -d KERNPROT outp inp */
            kernprot_args[0] = KERNPROT_DBID ;
            if ( args_rec.cp_trace_pages )
            {
              kernprot_args[1] = DIAGNOSE_COPYTRACE ;
              kernprot_args[2] = args_rec.inp_file ;
              kernprot_args[3] = args_rec.cp_trace_pages ;
              if ( args_rec.switches )
                printf ( "\nignoring switches: %s", args_rec.switches ) ;
            }
            else
            {
              kernprot_args[1] = DIAGNOSE_KERNPROT ;
              kernprot_args[2] = args_rec.outp_file ;
              kernprot_args[3] = args_rec.inp_file ;
              if ( args_rec.switches )
              {
                num_kernprot_args++ ;  /* akbxmn */
                kernprot_args[4] = args_rec.switches ;
                printf ( "\nreading : %s"  , args_rec.inp_file ) ;
                printf ( "\ncreating: %s\n", args_rec.outp_file ) ;
              }
            }
          }
          else
            call_ret = NOTOK ;
        }
      }
      else
        call_ret = NOTOK ;

    }

  if ( call_ret != NOTOK)
     call_ret = e950_call_component ( kernprot, num_kernprot_args,
                                      kernprot_args ) ;
  return ( call_ret ) ;
}

/* PTS 1108827 FH 29-Dec-2000                                                    */
/*-----------------------------------------------------------*/
int e953_call_PrecompiledTools  ( PROG_ID *PrecompiledTool,
        					      int      num_args,
		      	    			  char   **args )
/*-----------------------------------------------------------*/
{
  int call_ret = TRUE; /* PTS 1109785 */
  tsp00_DbNamec  DefaultDBName ;
  tsp00_ErrTextc ErrTextc ;

  DBG0 (( "\ncall_PrecompiledTools called with %d parameters\n", num_args ));

  if ( GetDefaultDBName ( DefaultDBName, ErrTextc ) && *DefaultDBName != '\0' )
  {
    tsp00_Pathc   DBRootOfDB;
    DBG0(("DefaultDBName = %s\n", DefaultDBName))

	if ( sqlGetDBRootOfDB(DefaultDBName, DBRootOfDB, sizeof(DBRootOfDB) ) )
    {
      DBG0(("DBROOT of %s is %s\n", DefaultDBName, DBRootOfDB));
      e950_PutEnvDBRoot ( DBRootOfDB ) ;
	}
    else
	{
      printf("Cant find installation path for <%s>! \n", DefaultDBName);
      call_ret = NOTOK ;
	}
  }
  else
  {
    printf("Error during searching of the name of the database: ");
    printf("%s\n", *DefaultDBName != '\0' ? ErrTextc : "no database name found");
    call_ret = NOTOK ;
  }

  if ( call_ret != NOTOK)
     call_ret = e950_call_component ( PrecompiledTool, num_args, args ) ;

  return ( call_ret ) ;
}

/*===========================================================================*
 *  LOCAL/PRIVATE/PROTECTED FUNCTIONS (CODE)                                 *
 *===========================================================================*/

/*-----------------------------------------------------------*/
static int clr_cntrl_lock ( char *dbname )
/*-----------------------------------------------------------*/
{
  tsp00_Pathc lockfile;
  char *rundir;
  int clr_ok = e951_getparam ( dbname, PAN_RUNDIR, &rundir) != CNULL ;
  DBG0 (( "\nclr_cntrl_lock called for db: %s\n", dbname ));
  if ( clr_ok )
    { sprintf ( lockfile, "%s%c%s", rundir, PATH_DELIMITER_EO01, CNTRL_RUN ) ;
      DBG0 (( "\nclr_cntrl_lock: removing file '%s' \n", lockfile ));
      if ( (clr_ok = unlink ( lockfile ) == 0 ) )
        { sprintf ( lockfile, "%s%c%s", rundir, PATH_DELIMITER_EO01, CNTRL_LCK ) ;
          DBG0 (( "\nclr_cntrl_lock: removing file '%s' \n", lockfile ));
          clr_ok = unlink ( lockfile ) == 0 ;
        }
    }
  return ( clr_ok ) ;
}

/*-----------------------------------------------------------*/
static int fill_cntrl_rec ( XCONTROL_ARGS *cntrl_rec,
   					        int            num_args,
					        char         **args )
/*-----------------------------------------------------------*/
{
  int fill_ok = TRUE , act_arg = 0 ;
  memset ( cntrl_rec, '\0', sizeof ( XCONTROL_ARGS ) ) ;
  while ( num_args && fill_ok )
    if ( IS_OPT_CHR ( *(args[act_arg]) ) )
      { if ( args[act_arg][1] == CNTRL_UID[1] )
          { if ( args[act_arg][2] != '\0' )  /* no delimitter after "-u" */
              { cntrl_rec->uname = (args[act_arg])+2 ;
                if ( (cntrl_rec->u_pwd = strchr ( cntrl_rec->uname, ',' )) )
                  { *cntrl_rec->u_pwd = '\0'; cntrl_rec->u_pwd++ ; }
                num_args --, act_arg++ ;
              }
            else
            if ( (fill_ok = num_args > 1) )
              { cntrl_rec->uname = args[act_arg+1] ;
                if ( (cntrl_rec->u_pwd = strchr ( cntrl_rec->uname, ',' )) )
                  { *cntrl_rec->u_pwd = '\0'; cntrl_rec->u_pwd++ ; }
                num_args -= 2, act_arg+=2 ;
              }
            DBG0(("uname: %s, u_pwd: %s\n", cntrl_rec->uname,
                   cntrl_rec->u_pwd ? cntrl_rec->u_pwd : "" ));
          }
        else
        if ( !strcmp ( &args[act_arg][1], &CNTRL_DBID[1] ) ||
             !strcmp ( &args[act_arg][1], &CNTRL_GWID[1] ) )
          { if ( (fill_ok = num_args > 1) )
              { cntrl_rec->db_gw_id = &args[act_arg][1] ;
                cntrl_rec->dbname = args[act_arg+1] ;
                if ( (cntrl_rec->dbnode = strchr ( cntrl_rec->dbname, ':' )) )
                  { cntrl_rec->dbname = cntrl_rec->dbnode + 1 ;
                    *cntrl_rec->dbnode = '\0';
                    cntrl_rec->dbnode = args[act_arg+1] ;
                  }
                num_args -= 2, act_arg+=2 ;
                DBG0(("id: %s dbname: %s, dbnode: %s \n",
                       cntrl_rec->db_gw_id, cntrl_rec->dbname,
                       cntrl_rec->dbnode ? cntrl_rec->dbnode : "" ));
              }
          }
        else
        if ( !strcmp ( &args[act_arg][1], &CNTRL_NODE_ID[1] ) )
          { if ( (fill_ok = num_args > 1) )
              { cntrl_rec->dbnode = args[act_arg+1] ;
                num_args -= 2, act_arg+=2 ;
                DBG0(("dbnode: %s\n", cntrl_rec->dbnode ));
              }
          }
        else
        if ( !strcmp ( &args[act_arg][1], &CNTRL_CLR_ID[1] ) )
          { cntrl_rec->clr_lock = TRUE ;
            num_args -= 1, act_arg+=1 ;
            DBG0(("clear_lock: %d\n", cntrl_rec->clr_lock ));
          }
        else
        if ( !strcmp ( &args[act_arg][1], &CNTRL_VERS_ID[1] ) )
          { if ( (fill_ok = act_arg == 0 && num_args == 1 ) )
              { num_args-- ;
                cntrl_rec->db_gw_id = "_" ;
                cntrl_rec->dbname = CNTRL_VERSION ;
                DBG0(("VERSION ( dbname set to %s )\n", cntrl_rec->dbname ));
              }
          }
        else
        if ( !strcmp ( &args[act_arg][1], &CNTRL_BATCH_ID[1] ) )
          { if ( (fill_ok = num_args > 1) )
              { cntrl_rec->batch_args = &(args[act_arg+1]) ;
                cntrl_rec->num_batch_args = num_args - 1;
                num_args = 0 ;
                DBG0(("%d arguments for batchprocessing found \n",
                       cntrl_rec->num_batch_args )) ;
              }
          }
        else
          { fill_ok = FALSE ;
            DBG0(("fill_cntrl_rec: unknown option: %s\n", args[act_arg] ));
          }
      }
    else
    if ( !strcmp ( args[act_arg], CNTRL_TRACE ) )
      { num_args-- ; act_arg++ ;
        cntrl_rec->trace = CNTRL_TRACE ;
        DBG0(("control trace set to %s )\n", cntrl_rec->trace ));
      }
    else
      { fill_ok = FALSE ;
        DBG0(("fill_cntrl_rec: unknown parameter: %s\n", args[act_arg] ));
      }
  if ( fill_ok && cntrl_rec->dbname == CNULL )
    if ( (cntrl_rec->dbname = e950_get_dbname_from_env () ) )
       cntrl_rec->db_gw_id = &CNTRL_DBID[1];
#ifdef WIN32
  if ( fill_ok )
    if ( cntrl_rec->dbname != CNULL &&
         strcmp(cntrl_rec->dbname , CNTRL_VERSION ) )
      e951_toupper_str ( cntrl_rec->dbname ) ;
#endif
  return ( fill_ok ) ;
}

/*-----------------------------------------------------------*/
static int fill_kernprot_rec ( XKERNRPOT_ARGS *kernprot_rec,
   					           int             num_args,
					           char          **args )
/*-----------------------------------------------------------*/
{
  int fill_ok = TRUE , act_arg = 0 ;
  memset ( kernprot_rec, '\0', sizeof ( XKERNRPOT_ARGS ) ) ;
  while ( num_args && fill_ok )
    if ( IS_OPT_CHR ( *(args[act_arg]) ) )
    {
      if ( !strcmp ( &args[act_arg][1], &KERNPROT_DBID[1] ) )
      {
        if ( (fill_ok = num_args > 1) )
        {
          kernprot_rec->dbname = args[act_arg+1] ;
          num_args -= 2, act_arg+=2 ;
          DBG0(("fill_kernprot_rec: dbname: %s\n", kernprot_rec->dbname )) ;
        }
      }
      else
        if ( !strcmp ( &args[act_arg][1], &KERNPROT_COPYTRACE[1] ) )
        {
          if ( (fill_ok = num_args > 1) )
          {
            int i=0;
            kernprot_rec->cp_trace_pages = args[act_arg+1] ;
            do
              if ( kernprot_rec->cp_trace_pages[i] != '\0' )
                if ( isdigit (kernprot_rec->cp_trace_pages[i]) )
                  i++ ;
                else
                {
                  fill_ok = FALSE ;
                  printf("invalid pages per tracefile: %s\n",
                          kernprot_rec->cp_trace_pages ) ;
                }
            while ( fill_ok && kernprot_rec->cp_trace_pages[i] != '\0' ) ;
            num_args -= 2, act_arg+=2 ;
            DBG0(("fill_kernprot_rec: %s %s\n", DIAGNOSE_COPYTRACE,
                   kernprot_rec->cp_trace_pages )) ;
          }
        }
        else
          if ( !strcmp ( &args[act_arg][1], &KERNPROT_USE_FILE[1] ) )
          {
            if ( (fill_ok = num_args > 1) )
            {
              kernprot_rec->use_inp_file = args[act_arg+1] ;
              num_args -= 2, act_arg+=2 ;
              DBG0(("fill_kernprot_rec: use file %s \n",
                     kernprot_rec->use_inp_file )) ;
            }
          }
          else
          {
            fill_ok = FALSE ;
            DBG0(("fill_kernprot_rec: unknown option: %s\n", args[act_arg] ));
          }
    }
    else
      if ( ! kernprot_rec->switches )
      {
        kernprot_rec->switches = args[act_arg] ;
        num_args-- ; act_arg++ ;
        DBG0(("kernprot_rec: switches: %s\n",kernprot_rec->switches ));
      }
      else
      {
        fill_ok = FALSE ;
        DBG0(("fill_kernrpot_rec: unknown option: %s\n", args[act_arg] ));
      }
  return ( fill_ok ) ;
}

/*-----------------------------------------------------------*/
static int set_inp_outp_filename ( XKERNRPOT_ARGS *kernprot_rec )
/*-----------------------------------------------------------*/
{
  int   set_ok = TRUE ;

  if ( kernprot_rec->dbname )
  { char *dummy_tr, *dummy_rundir ;
  if ( e951_getparam ( kernprot_rec->dbname ,
    PAN_KERNELTRACEFILE, &dummy_tr) == CNULL )
    set_ok = FALSE ;
  else
    if ( EO01_IS_ABSOLUT_PATH(dummy_tr) )
      strcpy (kernprot_rec->inp_file, dummy_tr ) ;
    else
    {
      if ( e951_getparam ( kernprot_rec->dbname, PAN_RUNDIR, &dummy_rundir) == CNULL )
        set_ok = FALSE ;
      else
      {
        set_ok = strlen(dummy_rundir) + strlen(dummy_tr) + 1 <
          sizeof (kernprot_rec->inp_file) ;
        if ( set_ok )
          sprintf(kernprot_rec->inp_file, "%s%c%s",
                  dummy_rundir, PATH_DELIMITER_EO01, dummy_tr ) ;
        else
          printf("set_inp_filename: path to long (%s%c%s)\n",
                  dummy_rundir, PATH_DELIMITER_EO01, dummy_tr ) ;
      }
    }
    if ( set_ok )
      sprintf(kernprot_rec->outp_file,"%s.prt", kernprot_rec->dbname) ;
  }
  else
    if ( kernprot_rec->use_inp_file )
    {
      strcpy( kernprot_rec->inp_file, kernprot_rec->use_inp_file );
      sprintf(kernprot_rec->outp_file,"%s.prt", kernprot_rec->use_inp_file);
    }

  return ( set_ok ) ;
}

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/
