/* @lastChanged: "1998-07-28  16:57"
 * @filename:    veo99.c
 * @purpose:     "GETPARAM"
 * @release:     M.m.c.p
 * @see:         ""
 *
 * @copyright:   "Copyright by SAP AG, 1998"


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 */

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include <stdlib.h>
#include <stdio.h>

#include "gsp00.h"
#include "heo01.h"
#include "heo46.h"
#include "heo11.h"
#include "geo60.h"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#ifdef NO_INLINES
# define _INLINE
#else
# define _INLINE	__inline
#endif

#define  GETPARAM_USAGE                      "usage: %s <ServerDB> <ParamName> \r\n"
#define  GETPARAM_ERR_TEXT                   "%s : %s\n"


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/



/*===========================================================================*
 *  GLOBAL VARIABLES                                                         *
 *===========================================================================*/

#ifdef WIN32
void sql60_disable_message_output();
void sql60_enable_message_output ();
#endif


/*===========================================================================*
 *  LOCAL VARIABLES                                                          *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL FUNCTIONS (PROTOTYPES)                                             *
 *===========================================================================*/



/*===========================================================================*
 *  GLOBAL FUNCTIONS (CODE)                                                  *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL FUNCTIONS (CODE)                                                   *
 *===========================================================================*/

static int 	PrintAllParams ( tsp00_DbNamec      DBNameC,
                             tsp01_RteError    *RteError )
{
  #undef  MF__
  #define MF__ MOD__"GetAllParams"

  tsp01_RteError     IgnoreError ;
  tsp11_XpReturnCode xp_ret;
  teo00_Bool         fOk = true ;
  XP_HFILE           XparamFileNo ;
  XP_VALUE_REC       xp_value;
  tsp00_C64c         XpValueC;
  tsp11_XpKeyTypec   XpKeyC, XpNextKeyC ;
  
  DBGIN;

  XpValueC[0] = 0 ;
  XpKeyC[0]   = 0 ;

  xp_ret = sqlXparamOpen (DBNameC, &XparamFileNo , XP_OPEN_READONLY_SP11, RteError) ;

  while ( xp_ret == xp_ok_sp11 )
  {
    xp_ret = sqlXparamGetNext (XparamFileNo, XpKeyC, XpNextKeyC,
                              (tsp00_XpValueRec *)&xp_value , &IgnoreError );
  
    if ( xp_ret == xp_ok_sp11 )
    {
      switch ( xp_value.xp_value_t )
      {
      case xp_int2_type_sp11   : sprintf(XpValueC,"%d", xp_value.xp_type.xp_int2 );
                                 break;
      case xp_int4_type_sp11   : sprintf(XpValueC,"%d", xp_value.xp_type.xp_int4 );
                                 break;
      case xp_c8_type_sp11     : eo46PtoC ( XpValueC , xp_value.xp_type.xp_c8 , 8 );
                                 break;
      case xp_c18_type_sp11    : eo46PtoC ( XpValueC , xp_value.xp_type.xp_c18 ,18 );
                                 break;
      case xp_c24_type_sp11    : eo46PtoC ( XpValueC , xp_value.xp_type.xp_c24 , 24 );
                                 break;
      case xp_c40_type_sp11    : eo46PtoC ( XpValueC , xp_value.xp_type.xp_c40 , 40 );
                                 break;
      case xp_c64_type_sp11    : eo46PtoC ( XpValueC , xp_value.xp_type.xp_c64 , 64 );
                                 break;
      case xp_crypt_type_sp11  : strcpy( XpValueC, "***" ) ;
                                 break;
      }
      strcpy(XpKeyC, XpNextKeyC ) ;
      printf("%-18.18s: %s\n", XpKeyC, XpValueC );
    }
  }

  sqlXparamClose ( XparamFileNo, XP_CLOSE_NO_FLUSH_SP11, &IgnoreError );

  DBGOUT ;

  return ( RteError->RteErrCode == RTE_NO_ERROR_SP01 );
}

/*===========================================================================*/

int  main ( int argc, char **argv )
  {
  int                     argn;
  char                    * val;
  char                    * db;

  long                    rc ;
  tsp00_C64c              outs;
  tsp01_RteError          RteError ;
  tsp00_Pathc             pszProgName ;

  argn = 1;

  strcpy(pszProgName, "getparam" ) ;

  if ( argc < 2 || argc > 3 ) 
    {
    fprintf ( stderr, GETPARAM_USAGE, pszProgName );
    return ( 1 );
    }

  db  = argv [ 1 ];

#ifdef WIN32
  sql60_disable_message_output ();
#endif

  if ( argc == 3 )
  {
    val = argv [ 2 ];

    eo11GetParam ( db , val, outs, &RteError ) ;
    if ( RteError.RteErrCode == RTE_NO_ERROR_SP01 ) 
    {
      printf("%s\n", outs );
      rc = 0 ;
    }   
  }
  else
  {
    PrintAllParams ( db, &RteError ) ;
  }

  if ( RteError.RteErrCode != RTE_NO_ERROR_SP01 ) 
  {
    fprintf ( stderr, GETPARAM_ERR_TEXT, pszProgName, RteError.RteErrText );
    rc = 1 ;
  }

#ifdef WIN32
  sql60_enable_message_output();
#endif

  return ( rc );
  }


/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/