/* @lastChanged: "2000-03-15"

  * @filename:   vgg06.cpp
  * @purpose:    "GG_cpp_auxiliary_functions"
  * @release:    7.2.0.0
  * @see:        ""
  *
  * @copyright:  "Copyright by SAP AG, 1999"


    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*/

/* ========================================================================== */
/* ==========              DECLARATION OF TYPES                     ========= */
/* ========================================================================== */

#include <string.h>
#include "hgg06.h"
#include "hgg10.h" // g10mv
#include "ggg92.h"
#include "SAPDBCommon/SAPDB_Types.h"

#if COMPILEMODE_MEO00 >= SLOW_MEO00 
#include "hta99.h"
#include "hta01.h"	
#endif

/* ========================================================================== */
/* ==========              DEFINITION OF GLOBAL FUNCTIONS           ========= */
/* ========================================================================== */


// PASCAL interface
externC void gg06BuildAuxSession (/*[OUT]*/ tgg91_SessionNo      &Session,
                                  /*[IN] */ tgg00_MessType_Param  MessType)
{
    Session.gg91BuildAuxSession (MessType);
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06BuildInternalSession (/*[OUT]*/ tgg91_SessionNo      &Session,
                                       /*[IN] */ tgg00_MessType_Param  MessType,
                                       /*[IN] */ tgg91_TransNo        &TransNo)
{
    Session.gg91BuildInternalSession (MessType, TransNo);
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06BuildRef (/*[OUT]*/ tgg91_PageRef &PageRef,
                           /*[IN] */ tsp00_Int4     Pno,
                           /*[IN] */ tsp00_Int4     Pos)
{
    PageRef.gg91BuildRef (Pno, Pos);
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06BuildTaskSession (/*[OUT]*/ tgg91_SessionNo &Session,
                                   /*[IN] */ tsp00_Uint1      TaskType)
{
    Session.gg91BuildTaskSession (TaskType);
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06Cint6Decr (/*[IN OUT]*/ tgg90_Cint6 &Cint6)
{
    --Cint6;
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06Cint6Incr (/*[IN OUT]*/ tgg90_Cint6 &Cint6)
{
    ++Cint6;
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06Cint6ToLine (/*[IN]    */ tgg90_Cint6 &Cint6,
                              /*[IN OUT]*/ int         &LineLen,
                              /*[IN OUT]*/ tsp00_Line  &Line)
{
    Cint6.gg90AppendToLine (LineLen, Line);
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06FileVersIncr (/*[IN OUT]*/ tgg91_FileVersion &FileVers)
{
    ++FileVers;
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06FileVersToLine (/*[IN]    */ tgg91_FileVersion &FileVers,
                                 /*[IN OUT]*/ int               &LineLen,
                                 /*[IN OUT]*/ tsp00_Line        &Line)
{
    FileVers.gg91AppendToLine (LineLen, Line);
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06GetObjHeaderParam (/*[IN] */ void               *ObjHeaderBuf,
                                    /*[OUT]*/ tgg00_ObjTransInfo &ObjTransInfo,
                                    /*[OUT]*/ tgg00_ObjPagePos   &NextFreeObj,
                                    /*[OUT]*/ tgg00_ObjState     &ObjState,
                                    /*[OUT]*/ tgg00_ObjFrameVers &ObjFrameVers)
{
    tgg92_ObjHeader* pObjHeader = (tgg92_ObjHeader*) ObjHeaderBuf;

    ObjTransInfo = pObjHeader->ohdTransInfo_gg92;
    NextFreeObj  = pObjHeader->ohdNextFreeObj_gg92;
    ObjState     = pObjHeader->ohdState_gg92;
    ObjFrameVers = pObjHeader->ohdFrameVersion_gg92;
}

/*----------------------------------------------------------------------------*/

externC tsp00_Int4 gg06GetOidFrameVers (/*[IN]*/ tgg92_KernelOid &oid)
{
    return ( oid.gg92GetFrameVers() );
}

/*----------------------------------------------------------------------------*/

externC tsp00_Int4 gg06GetOidPos (/*[IN]*/ tgg92_KernelOid &oid)
{
    return ( oid.gg92GetPos() );
}

/*----------------------------------------------------------------------------*/

externC void gg06InitObjTransInfo (/*[OUT]*/ tgg00_ObjTransInfo &ObjTransInfo)
{
    ObjTransInfo.otrBeforeRef_gg00.gg91SetNilRef();
    ObjTransInfo.otrLockTransIndex_gg00 = cgg_nil_transindex;
    ObjTransInfo.otrLockTransId_gg00.gg90SetNil();
    ObjTransInfo.otrUpdTransId_gg00.gg90SetNil();
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC tsp00_Int4 gg06Int4Get (const tsp00_BytePtr Buffer,
                                tsp00_Int4          Pos)
{
    return ( REINTERPRET_CAST(tgg90_Cint4*,(Buffer+Pos-1))->gg90GetInt4() );
}

/*----------------------------------------------------------------------------*/

// PASCAL interface to handle Uint2
externC tsp00_Int4 gg06Int4FromUint2 (/*[IN]*/ tsp00_Uint2 Uint2)
{ 
    return Uint2; 
}

/*----------------------------------------------------------------------------*/

// PASCAL interface to handle Uint2
externC void
gg06Int4ToUint2 (/*[IN] */ tsp00_Int4   Int4,
                 /*[OUT]*/ tsp00_Uint2 &Uint2)
{
    Uint2 = Int4; 
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC tsp00_Bool gg06IsDummyFileVers (/*[IN]*/ tgg91_FileVersion &FileVers)
{
    return ( FileVers.gg91IsDummyVers() );
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC tsp00_Bool gg06IsDummyRef (/*[IN]*/ tgg91_PageRef &PageRef)
{
    return ( PageRef.gg91IsDummyRef() );
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC tsp00_Bool gg06IsDummyTrans (/*[IN]*/ tgg91_TransNo &TransNo)
{
    return ( TransNo.gg90IsDummy() );
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC tsp00_Bool gg06IsEqOidPnoPos (/*[IN]*/ tgg92_KernelOid &oid1,
                                /*[IN]*/ tgg92_KernelOid &oid2)
{
    return ( oid1.gg92IsEqPnoPos (oid2) );
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC tsp00_Bool gg06IsNilObjFileNo (/*[IN]*/ tgg00_ObjFileNo ObjFileNo)
{
    return (NIL_OBJ_FILE_NO_GG92 == ObjFileNo);
}

/*----------------------------------------------------------------------------*/

// PASCAL interface PTS 1105492 JA 2000-03-15
externC tsp00_Bool gg06IsNilOid (/*[IN]*/ tgg92_KernelOid &oid)
{
    return ( oid.gg92IsNil() );
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC tsp00_Bool gg06IsNilRef (/*[IN]*/ tgg91_PageRef &PageRef)
{
    return ( PageRef.gg91IsNilRef() );
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC tsp00_Bool gg06IsNilSession (/*[IN]*/ tgg91_SessionNo &SessionNo)
{
    return ( SessionNo.gg90IsNil() );
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC tsp00_Bool gg06IsNilTrans (/*[IN]*/ tgg91_TransNo &TransNo)
{
    return ( TransNo.gg90IsNil() );
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC tsp00_Bool gg06IsSpecialFileVers (/*[IN]*/ tgg91_FileVersion &FileVers)
{
    return ( FileVers.gg91IsSpecialVers() );
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC tsp00_Bool gg06IsZeroSubtrans (/*[IN]*/ tgg91_SubtransNo &SubtransNo)
{
    return ( SubtransNo.gg90IsZero() );
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06OidToLine (/*[IN]    */ tgg92_KernelOid &oid,
                            /*[IN OUT]*/ int             &LineLen,
                            /*[IN OUT]*/ tsp00_Line      &Line)
{
    oid.gg92AppendToLine (LineLen, Line);
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC tsp00_Int4 gg06PnoGet (/*[IN]*/ tgg90_Cint4 &Cint4)
{
    return ( Cint4.gg90GetInt4() );
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06PnoPut (/*[OUT]*/ tgg90_Cint4 &Cint4,
                         /*[IN] */ tsp00_Int4   int4)
{
    Cint4.becomes (int4);
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC tsp00_Int4 gg06RefPno (/*[IN]*/ tgg91_PageRef &PageRef)
{
    return (PageRef.gg91RefPno() );
};

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC	tsp00_Int4 gg06RefPos (/*[IN]*/ tgg91_PageRef &PageRef)
{
    return (PageRef.gg91RefPos() );
};

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06SessionBecomes (/*[OUT]*/ tgg91_SessionNo &SessionNo,
                                 /*[IN] */ tsp00_Int4       IntValue)
{
    SessionNo.becomes(IntValue);
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06SessionDecr (/*[IN OUT]*/ tgg91_SessionNo &SessionNo)
{
    --SessionNo;
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC tsp00_Int4 gg06SessionGetInt4 (/*[IN]*/ tgg91_SessionNo &SessionNo)
{ 
    return SessionNo.gg90GetInt4(); 
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06SessionIncr (/*[IN OUT]*/ tgg91_SessionNo &SessionNo)
{
    ++SessionNo;
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06SessionToLine (/*[IN]    */ tgg91_SessionNo &SessionNo,
                                /*[IN OUT]*/ int             &LineLen,
                                /*[IN OUT]*/ tsp00_Line      &Line)
{
    SessionNo.gg91AppendToLine (LineLen, Line);
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06SetDummyFileVers (/*[OUT]*/ tgg91_FileVersion &FileVers)
{
    FileVers.gg91SetDummyVers();
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06SetDummyRef (/*[OUT]*/ tgg91_PageRef &PageRef)
{ 
    PageRef.gg91SetDummyRef ();
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06SetDummyTrans (/*[OUT]*/ tgg91_TransNo &TransNo)
{
    TransNo.gg90SetDummy();
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06SetNilOid (/*[OUT]*/ tgg92_KernelOid &oid)
{
    oid.gg92SetNil ();
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06SetNilRef (/*[OUT]*/ tgg91_PageRef &PageRef)
{ 
    PageRef.gg91SetNilRef ();
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06SetNilSession (/*[OUT]*/ tgg91_SessionNo &SessionNo)
{
    SessionNo.gg90SetNil();
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06SetNilTrans (/*[OUT]*/ tgg91_TransNo &TransNo)
{
    TransNo.gg90SetNil();
}

/*----------------------------------------------------------------------------*/

// PASCAL interface

externC void gg06SetOid (/*[OUT]*/ tgg92_KernelOid &oid,
                         /*[IN] */ tsp00_Int4       pno,
                         /*[IN] */ tsp00_Int4       pos,
                         /*[IN] */ tsp00_Int4       FrameVers)
{
    oid.gg92Set (pno, pos, FrameVers);
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06SetOidFrameVers (/*[OUT]*/ tgg92_KernelOid &oid,
                                  /*[IN] */ tsp00_Int4       FrameVers)
{
    oid.gg92SetFrameVers (FrameVers);
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06SetOidPos (/*[OUT]*/ tgg92_KernelOid &oid,
                            /*[IN] */ tsp00_Int4       pos)
{
    oid.gg92SetPos (pos);
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06SetSpecialFileVers (/*[OUT]*/ tgg91_FileVersion &FileVers)
{
    FileVers.gg91SetSpecialVers();
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06SetZeroCint6 (/*[OUT]*/ tgg90_Cint6 &Cint6)
{
    Cint6.gg90SetZero();
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06SetZeroSession (/*[OUT]*/ tgg91_SessionNo &SessionNo)
{
    SessionNo.gg90SetZero();
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06SetZeroSubtrans (/*[OUT]*/ tgg91_SubtransNo &SubtransNo)
{
    SubtransNo.gg90SetZero();
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06SetZeroTrans (/*[OUT]*/ tgg91_TransNo &TransNo)
{
    TransNo.gg90SetZero();
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC tsp00_Int4 gg06SizeofObjHeader ()
{ 
    return ( sizeof (tgg92_ObjHeader) ); 
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06SubtransBecomes (/*[OUT]*/ tgg91_SubtransNo &SubtransNo,
                                  /*[IN] */ tsp00_Int4        IntValue)
{
    SubtransNo.becomes(IntValue);
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06SubtransDecr (/*[IN OUT]*/ tgg91_SubtransNo &SubtransNo)
{
    --SubtransNo;
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC tsp00_Int4 gg06SubtransGetInt4 (/*[IN]*/ tgg91_SubtransNo &SubtransNo)
{ 
    return SubtransNo.gg90GetInt4(); 
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06SubtransIncr (/*[IN OUT]*/ tgg91_SubtransNo &SubtransNo)
{
    ++SubtransNo;
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06SubtransToLine (/*[IN]    */ tgg91_SubtransNo &SubtransNo,
                                 /*[IN OUT]*/ int              &LineLen,
                                 /*[IN OUT]*/ tsp00_Line       &Line)
{
    SubtransNo.gg90AppendToLine (LineLen, Line);
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06TransBecomes (/*[OUT]*/ tgg91_TransNo &TransNo,
                               /*[IN] */ tsp00_Int4     IntValue)
{
    TransNo.becomes(IntValue);
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06TransDecr (/*[IN OUT]*/ tgg91_TransNo &TransNo)
{
    --TransNo;
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC tsp00_Int4 gg06TransGetInt4 (/*[IN]*/ tgg91_TransNo &TransNo)
{ 
    return TransNo.gg90GetInt4(); 
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06TransIncr (/*[IN OUT]*/ tgg91_TransNo &TransNo)
{
    ++TransNo;
}

/*----------------------------------------------------------------------------*/

// PASCAL interface
externC void gg06TransToLine (/*[IN]    */ tgg91_TransNo &TransNo,
                              /*[IN OUT]*/ int           &LineLen,
                              /*[IN OUT]*/ tsp00_Line    &Line)
{
    TransNo.gg90AppendToLine (LineLen, Line);
}

/*----------------------------------------------------------------------------*/
static inline tsp00_Uint4 gg06_NextPrimNumber (tsp00_Uint4 number)
{
    if ( number <= 2 ) return 2;

    tsp00_Uint4 nextNumber = number;
    bool        isPrim     = false;
    tsp00_Uint4 i;

    while ( ! isPrim )
    {
        nextNumber++;
        i = 3;
        isPrim = (nextNumber % 2) == 1;
        while (i*i <= nextNumber && isPrim)
        {
            isPrim = (nextNumber % i) != 0;
            i += 2;
        }
    }
    return nextNumber;
}

/*----------------------------------------------------------------------------*/

// PTS 1120350 UH 2003-01-31 new
#define MAX_ROWID_HASH_PRIM_NUMBERS 1024
#define START_PRIM_NUMBER           255
#define PRIMNUMBER_DISTANCE         3

// PTS 1120350 UH 2003-01-31 new
static tsp00_Uint4 gg06RowIdHashPrimNumbers[MAX_ROWID_HASH_PRIM_NUMBERS];

/*----------------------------------------------------------------------------*/
externPascal void gg06buildCompressedKey (
    const tsp00_Byte*                    key,
    tsp00_Int4                           keySize,
    tsp00_Int4                           keyStartPosition,
    tsp00_Int4                           keyLength,
    tsp00_Int4                           compressedKeySize,
    tsp00_Byte*                          compressedKey,
    tsp00_Uint4           VAR_VALUE_REF  hashvalue1,
    tsp00_Uint4           VAR_VALUE_REF  hashvalue2,
    tgg00_BasisError      VAR_VALUE_REF  moveError)
{
    // PTS 1121179 UH 2003-04-07 new

    ROUTINE_DBG_MEO00 ("gg06buildCompressedKey");

    hashvalue1 = 0;
    hashvalue2 = 0;
    
    if ( moveError != e_ok )
        return;
        
    tsp00_Int4 hashKeyLength;
    tsp00_Int4 moveKeyLength;
    
    if ( keyLength > MAX_ROWID_HASH_PRIM_NUMBERS )
        hashKeyLength = MAX_ROWID_HASH_PRIM_NUMBERS;
    else
        hashKeyLength = keyLength;
    
    if ( keyLength < compressedKeySize )
        moveKeyLength = keyLength;
    else
        moveKeyLength = compressedKeySize;

    // move into first part
    g10mv1 ( "VGG06 ", 1,
             keySize,       compressedKeySize,
             key,           keyStartPosition + keyLength - moveKeyLength,
             compressedKey, 1,
             moveKeyLength,  moveError);
                        
    if ( moveError != e_ok )
        return;

    // determine hash values
    tsp00_Int4  curroffset;
    tsp00_Uint4 newSummmand;

    for ( curroffset = keyStartPosition-1;
          curroffset < keyStartPosition+hashKeyLength-1;
          curroffset++ )
    {
        newSummmand = ( (*(key+curroffset) + curroffset)
                        *
                        gg06RowIdHashPrimNumbers[curroffset] );
                        
        if ( hashvalue1 < SAPDB_MAX_UINT4 - newSummmand )
            hashvalue1 += newSummmand;
        else
            hashvalue2 += newSummmand;
    }

    #if COMPILEMODE_MEO00 >= SLOW_MEO00 
    t01sname   (kb_lock, "original key");
    t01moveobj (kb_lock, const_cast<tsp00_Byte*>(key), keyStartPosition, keyLength);
    t01sname   (kb_lock, "compressed  ");
    t01moveobj (kb_lock, compressedKey, 1, moveKeyLength);
    t01p2int4  (kb_lock, "hashvalue1  ", hashvalue1,
                         "hashvalue2  ", hashvalue2);
    #endif
}

/*----------------------------------------------------------------------------*/
externPascal void gg06CalculateRowIdHashPrimNumbers ()
{
    // PTS 1120350 UH 2003-01-31 new
    // start: 255, distance: 3, count: 1024, number: 28697, sum: 3582848940
    ROUTINE_DBG_MEO00 ("gg06CalculateRowIdHashPrimNumbers");
    
    tsp00_Uint4 primnumber = START_PRIM_NUMBER;
    
    for ( tsp00_Uint4 primnumberCount = 0;
          primnumberCount < MAX_ROWID_HASH_PRIM_NUMBERS;
          ++primnumberCount )
    {
        primnumber = gg06_NextPrimNumber(primnumber);
        if ( primnumberCount % 2 == 0 )
            gg06RowIdHashPrimNumbers[primnumberCount] = primnumber;
        else
            gg06RowIdHashPrimNumbers[MAX_ROWID_HASH_PRIM_NUMBERS-primnumberCount] = primnumber;
        
        for (tsp00_Uint4 skip = 0; skip<PRIMNUMBER_DISTANCE; skip++)
            primnumber = gg06_NextPrimNumber(primnumber);
    }
    #define UWE 1
    #ifdef UWE
    #define KEYLENGTH 1024
    tsp00_Byte key[KEYLENGTH];
    memset (key, 255, KEYLENGTH);
    #define COMPRESSEDKEYSIZE 56
    tsp00_Byte compressedkey[COMPRESSEDKEYSIZE];
    memset (compressedkey, 0, COMPRESSEDKEYSIZE);
    tsp00_Uint4 hashvalue1;
    tsp00_Uint4 hashvalue2;
    tgg00_BasisError berr = e_ok;
    gg06buildCompressedKey (key, KEYLENGTH, 1, KEYLENGTH,
                            COMPRESSEDKEYSIZE, compressedkey, hashvalue1, hashvalue2, berr);
    #endif
}

// end of program
