/*
===========================================
   responsible:    DanielD
   description:    Interface to SAP DB
   target:         Perl
   last changed:   27.11.2001
   copyright:      2001 SAP AG

    ========== licence begin LGPL
    Copyright (C) 2000-2001 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end
===========================================*/


#include <stddef.h>

#ifdef __cplusplus
#include <math.h>
#include <stdlib.h>
extern "C" {
#else
#define bool char
#define HAS_BOOL 1
#endif

#define PERL_GLUE 1

#include "patchlevel.h"/* no check */

#if defined (ACTIVE_PERL)
    #if PATCHLEVEL < 6
        #define PERL_OBJECT
    #endif
#endif

#include "EXTERN.h"/* no check */
#include "perl.h"/* no check */
#if defined (PERL_OBJECT)
#include "perlCAPI.h"/* no check */
#endif
#include "XSUB.h"/* no check */
#undef free
#undef malloc
#undef realloc
#include <string.h>
#if defined (bool)
#undef bool
#endif
#if defined (PERL_OBJECT)
    #if !defined (PL_sv_undef)
        #define PL_sv_undef sv_undef
    #endif
    #if !defined (PL_sv_yes)
        #define PL_sv_yes sv_yes
    #endif
    #if !defined (PL_errgv)
        #define PL_errgv errgv
    #endif
#endif
#ifdef __cplusplus
}
#endif

#if defined (_WIN32)
#define GLUEEXPORT __declspec(dllexport)
#else
#define GLUEEXPORT
#endif

#define LIT(c) CONST_CAST(char *, c)
#define CROAKBUF_SIZE 200

#define MF__
#if defined (__cplusplus)
#define externC extern "C"
#else
#define externC
#endif

#define SL_None &PL_sv_undef
#define SL_isNone(val) (!SvOK (val))
#define SL_isTrue(val) SvTRUE (val)

static const char * invalidArgCount_C = "Invalid number of arguments";

static void *
getGluePointer (SV * sv, const char * className, char * croakBuf)
{
    IV   tmp;

    if (SvGMAGICAL(sv)) {
        mg_get(sv);
    }

    /* Check to see if this is an object */
    if (sv_isobject(sv)) {
        SV *tsv = (SV*) SvRV(sv);
        if ((SvTYPE(tsv) == SVt_PVHV)) {
            MAGIC *mg;
            if (SvMAGICAL(tsv)) {
                mg = mg_find(tsv,'P');
                if (mg) {
                    SV *rsv = mg->mg_obj;
                    if (sv_isobject(rsv)) {
                        tmp = SvIV((SV*)SvRV(rsv));
                    }
                }
            }
            else {
                strcpy (croakBuf, "Not a valid pointer value");
                return NULL;
            }
        }
        else {
            tmp = SvIV((SV*)SvRV(sv));
        }
    }
    else if (SL_isNone (sv)) {            /* Check for undef */
        strcpy (croakBuf, "Undef object not allowed");
        return NULL;
    }
    else if (SvTYPE(sv) == SVt_RV) {       /* Check for NULL pointer */
        if (!SvROK(sv)) {
            strcpy (croakBuf, "NULL object not allowed");
        }
        else {
            strcpy (croakBuf, "NULL object not allowed");
        }
        return NULL;
    }
    else {                                 /* Dont know what it is */
        strcpy (croakBuf, "Not a valid pointer value");
        return NULL;
    }
    /* Now see if the types match */

    if (!sv_isa (sv, (char *) className)) {
        sprintf (croakBuf, "Object is of class %s (should be %s)",
                HvNAME(SvSTASH(SvRV(sv))), className);
        return NULL;
    }
    return (void*) tmp;
}


/*
 * Exceptions
 */

static const char * CommunicationError_sapdb = "CommunicationError";
static const char * SQLError_sapdb = "SQLError";

/* snippet Generic_precode */
#include <ctype.h>

#if defined (PERL_GLUE)
    /* conflict with gsp00 enum */
    #undef na
    /* conflict with malloc.h realloc included later */
    #undef realloc
#endif
#if defined (__cplusplus)
#include "heo02.h"
#else
#include "heo102.h"
#endif
#include "hsp02.h"
#include "hsp30.h"
#include "hsp40.h"
#include "hsp41.h"
#include "hsp42.h"
#include "hsp77.h"
#include "hsp78.h"
#include "hsp100.h"
#include "hin28.h"

#if defined (PERL_GLUE)
    #include "patchlevel.h"/* no check */
    #if PATCHLEVEL >= 5
        #define na PL_na
    #else
        #define na Perl_na
    #endif
    #undef read
#endif

#if defined (PYTHON_GLUE)
#include <compile.h>
#include <frameobject.h>
typedef PyObject SL_ValueT;
#elif defined (PERL_GLUE)
typedef SV SL_ValueT;
#else
    #error Invalid Glue Environment
#endif

/*----------------------------------------*/

static void c2p (
    void * instanceData,
    int    instanceSize,
    const char * str)
{
    int len = strlen (str);

    if (len > instanceSize)
        len = instanceSize;
    memcpy (instanceData, str, len);
    if (len < instanceSize)
        memset ((char *) instanceData + len, ' ', instanceSize - len);
}

static void * tracingMalloc (size_t len)
{
    void * result = malloc (len);
    printf("malloc: 0x%p\n", result);
    return result;
}

static void tracingFree (void * ptr)
{
    printf("free: 0x%p\n", ptr);
    free (ptr);
}

/*
#define malloc(len) tracingMalloc(len)
#define free(ptr)   tracingFree(ptr)
*/

#if defined (__cplusplus) && (COMPILEMODE_MSP00 >= SLOW_MSP00)
    /*!
      Class: ScriptingTracer
     */
    class ScriptingTracer
    {
    public:
        ScriptingTracer (const char * name);
        ~ScriptingTracer ();
        void trace (int layer, const char *format, ...);
        static void switchTrace (bool on);
        static void initTrace ();
    private:
        const char * procName;
        static bool traceIsOn;
    };

    ScriptingTracer::ScriptingTracer (const char * name)
    {
        if (traceIsOn) {
            printf("<%s>\n", name);
        }
        this->procName = name;
    }

    ScriptingTracer::~ScriptingTracer ()
    {
        if (traceIsOn) {
            printf("</%s>\n", this->procName);
        }
    }

    void ScriptingTracer::trace (int layer, const char *format, ...)
    {
        if (traceIsOn) {
            va_list         ap;

            va_start (ap, format);
            vprintf(format, ap);
            va_end (ap);
        }
    }

    bool ScriptingTracer::traceIsOn = false;

    void ScriptingTracer::switchTrace (bool on)
    {
        traceIsOn = on;
    }

    void ScriptingTracer::initTrace ()
    {
        const char * sapdbTrace = getenv ("SAPDBTRACE");
        if (sapdbTrace == NULL) {
            traceIsOn = false;
        }
        else {
            traceIsOn = true;
        }
    }
    /*! EndClass: ScriptingTracer */

    #undef TRACER_CLASS_MSP00
    #define TRACER_CLASS_MSP00 ScriptingTracer

#endif

#if defined (__cplusplus)
#define REFPARM(name)   name
#define INLINE          inline
#else
#define REFPARM(name)   &name
#define INLINE

#undef ROUTINE_DBG_MSP00
#define ROUTINE_DBG_MSP00(lit) static const char * funcname_glue_ = lit

#define REINTERPRET_CAST(type, expr) ((type) expr)
#define STATIC_CAST(type, expr) ((type) expr)
#define CONST_CAST(type, expr) ((type) expr)
#endif

char * buildInfo ()
{
    static tsp00_Versionc versionString;
    static bool isInitialized;
#if defined (PYTHON_GLUE)
    tsp100_CompName compName = "sapdbpy  ";
#elif defined (PERL_GLUE)
    tsp100_CompName compName = "sapdbperl";
#else
    #error must be either Python or Perl
#endif

    if (!isInitialized) {
        sp100_GetVersionString (compName, s100buildnumber, versionString);
    }
    return versionString;
}


static const int internalErrorC = -16506;

typedef unsigned char byte_t;

typedef struct ParseInfoT {
    tin01_parseid     pid;
    int               paramCount;
    int               outputParamCount;
    int               inLongCount;
    int               outLongCount;
    tsp1_param_info * paramInfo;
    char            * sqlString;
} ParseInfoT;

typedef struct LongReaderT LongReaderT;

static SL_ValueT *
SL_fromString (
    const char * value,
    int len);

static SL_ValueT *
SL_fromUCS2 (
    const void * value,
    int charLen);

static int
SL_readCall (
    SL_ValueT * value,
    byte_t * streamBuf,
    int requestedLen);

static SL_ValueT *
readLong (
    LongReaderT * nself,
    int requested);

static bool
fillParamInfos (
    tin01_sql_session * session,
    ParseInfoT * parseInfo,
    bool initial);

static void
increaseRef (
    void * obj);

static void
decreaseRef (
    void * obj);

static void
setAutocommit (
    tin01_sql_session * session,
    bool autocommit);

static bool
getAutocommit (
    tin01_sql_session * session);

static bool
lastWasEOT (
    tin01_sql_session * session);

static void
enableCursorNames (
    tin01_sql_session * session,
    bool enable);

static int
nextCursorCount (
    tin01_sql_session * session);

static void
addCursorName (
    tin01_sql_session * session);

static void
clearEOTFlag (
    tin01_sql_session * session);

static void
setEOTFlag (
    tin01_sql_session * session);



/*----------------------------------------*/

static void
clearParseInfo (
    tin01_sql_session * session,
    ParseInfoT * parseInfo)
{
    ROUTINE_DBG_MSP00 ("clearParseInfo");
    if ((session != NULL) && session->is_connected) {
        if (parseInfo->paramCount != -1) {
            i28droppid (session, parseInfo->pid);
        }
    }
    if (parseInfo->paramInfo != NULL) {
        free (parseInfo->paramInfo);
        parseInfo->paramInfo = NULL;
    }
    parseInfo->paramCount = -1;
    if (parseInfo->sqlString != NULL) {
        free (parseInfo->sqlString);
        parseInfo->sqlString = NULL;
    }
}

/*----------------------------------------*/

typedef struct PreparedT {
    tin01_sql_session * session;
    ParseInfoT        parseInfo;
} PreparedT;

typedef enum fetchKindsT {
    nextFetchE, relativeFetchE, absoluteFetchE,
    firstFetchE, lastFetchE,
    fetchKindsE
} fetchKindsT;

static const char * fetchCommandsC [] = {
    " next ", " relative ? ", " pos (?) ", " first ", " last "
};

typedef struct ResultSetT {
    tin01_sql_session * session;
    tsp00_KnlIdentifierc cursorName;
    ParseInfoT        fetchInfo [fetchKindsE];
    byte_t            * dataBuf;
    int               dataSize;
    int               dataLen;
    byte_t            * currentRow;
    byte_t            * eoRecords;
    int               rowSize;
    int               maxRecordsPerFetch;
    const char      * colNames;
    tsp1_param_info * colInfo;
    int               colCount;
    int               currentPos;
} ResultSetT;

/*----------------------------------------*/

static int dbgSizes [] = {36, 256, 1111};

static void
checkMem ()
{
    ROUTINE_DBG_MSP00 ("checkMem");
    int i;
    for (i = 0; i < 3; ++i) {
        void * p  = malloc (dbgSizes [i]);
        free (p);
    }
}

/*----------------------------------------*/

static const int atEndStreamC = -2;
static const int errorStreamC = -1;

struct InputStreamVMT;

typedef struct InputStreamVMT* InputStreamT;

typedef int readFunctionT (InputStreamT * instream, byte_t * buf, int requestedLen);

typedef struct InputStreamVMT {
    readFunctionT * read;
} InputStreamVMT;

/*----------------------------------------*/

typedef struct StringInputStreamT {
    const InputStreamVMT * vmt;
    const char * data;
    const char * eoData;
    const char * pos;
} StringInputStreamT;

/*----------------------------------------*/

static int
stringInputRead (
    StringInputStreamT * self,
    byte_t * buf,
    int requestedLen)
{
    ROUTINE_DBG_MSP00 ("stringInputRead");
    int copyLen = requestedLen;
    int remaining = self->eoData - self->pos;

    if (requestedLen <= 0) {
        return errorStreamC;
    }
    if (remaining <= 0) {
        return atEndStreamC;
    }
    if (copyLen > remaining) {
        copyLen = remaining;
    }
    memcpy (buf, self->pos, copyLen);
    self->pos += copyLen;
    return copyLen;
}

/*----------------------------------------*/

static const InputStreamVMT StringInputStreamVMT_Imp = {
    (readFunctionT*) stringInputRead,
};

/*----------------------------------------*/

static InputStreamT *
initStringInputStream (
    StringInputStreamT * self,
    const char * data,
    int len)
{
    ROUTINE_DBG_MSP00 ("initStringInputStream");
    self->vmt = &StringInputStreamVMT_Imp;
    self->data = data;
    self->pos = data;
    self->eoData = data + len;
    return(InputStreamT*) self;
}

typedef struct CallableInputStreamT {
    const InputStreamVMT * vmt;
    SL_ValueT * callable;
    bool atEnd;
} CallableInputStreamT;

/*----------------------------------------*/

static int
callableInputRead (
    CallableInputStreamT * self,
    byte_t * buf,
    int requestedLen)
{
    ROUTINE_DBG_MSP00 ("callableInputRead");
    if (self->atEnd) {
        return atEndStreamC;
    }
    return SL_readCall (self->callable, buf, requestedLen);
}

/*----------------------------------------*/

static const InputStreamVMT CallableInputStreamVMT_Imp = {
    (readFunctionT*) callableInputRead,
};

/*----------------------------------------*/

static InputStreamT *
initCallableInputStream (
    CallableInputStreamT * self,
    SL_ValueT * callable)
{
    ROUTINE_DBG_MSP00 ("initStringInputStream");
    self->vmt = &CallableInputStreamVMT_Imp;
    self->callable = callable;
    self->atEnd = false;
    return(InputStreamT*) self;
}

/*----------------------------------------*/

struct LongReaderT {
    tin01_sql_session  * session;
    tsp00_LongDescriptor descriptor;
    byte_t               * dataBuf;
    int                  dataSize;
    int                  dataPos;
    int                  dataLen;
    bool                 lastGetvalFound;
};

/*----------------------------------------*/

bool keywordMatch (
    const char * kwArg,
    int kwArgLen,
    const char * token,
    int tokenLen)
{
    ROUTINE_DBG_MSP00 ("keywordMatch");

    if (kwArgLen != tokenLen) {
        return false;
    }
    if (memcmp (kwArg, token, kwArgLen) == 0) {
        return true;
    }
    else {
        return false;
    }
}

/*----------------------------------------*/

void setStringConfig (
    char * configField,
    int fieldLen,
    const char * value,
    int valueLen)
{
    ROUTINE_DBG_MSP00 ("setStringConfig");

    if (fieldLen <= valueLen) {
        memcpy (configField, value, fieldLen);
    }
    else {
        memcpy (configField, value, valueLen);
        memset (configField + valueLen, ' ', fieldLen - valueLen);
    }
}

/*----------------------------------------*/

int parseInt (
    const char * token,
    int tokenLen)
{
    ROUTINE_DBG_MSP00 ("parseInt");
    int result = 0;
    int i;

    for (i = 0; i < tokenLen; ++i) {
        result *= 10;
        result += token [i] - '0';
    }
    return result;
}

/*----------------------------------------*/

bool parseBool (
    const char * token,
    int tokenLen,
    bool defaultValue)
{
    ROUTINE_DBG_MSP00 ("parseBool");
    bool result = defaultValue;

    if (memcmp (token, "true", tokenLen) == 0) {
        result = true;
    }
    else if (memcmp (token, "on", tokenLen) == 0) {
        result = true;
    }
    else if (memcmp (token, "1", tokenLen) == 0) {
        result = true;
    }
    else if (memcmp (token, "false", tokenLen) == 0) {
        result = false;
    }
    else if (memcmp (token, "off", tokenLen) == 0) {
        result = false;
    }
    else if (memcmp (token, "0", tokenLen) == 0) {
        result = false;
    }
    return result;
}

/*----------------------------------------*/

static void
handleConfigArgument (
    tin01_sql_session * session,
    const char * token,
    int tokenLen,
    const char * value,
    int valueLen)
{
    ROUTINE_DBG_MSP00 ("handleConfigArgument");

    if (keywordMatch ("version", 7, token, tokenLen)) {
        setStringConfig ((char *)session->senderid, sizeof (tsp00_C5),
            value, valueLen);
    }
    else if (keywordMatch ("component", 9, token, tokenLen)) {
        setStringConfig ((char *) session->senderid + sizeof (tsp00_C5), sizeof (tsp00_C3),
            value, valueLen);
    }
    else if (keywordMatch ("sqlmode", 7, token, tokenLen)) {
        setStringConfig ((char*)session->xuser.xu_sqlmode, sizeof (session->xuser.xu_sqlmode),
            value, valueLen);
    }
    else if (keywordMatch ("cachelimit", 10, token, tokenLen)) {
        session->xuser.xu_cachelimit = parseInt (value, valueLen);
    }
    else if (keywordMatch ("timeout", 7, token, tokenLen)) {
        session->xuser.xu_timeout = parseInt (value, valueLen);
    }
    else if (keywordMatch ("isolation", 9, token, tokenLen)) {
        session->xuser.xu_isolation = parseInt (value, valueLen);
    }
    else if (keywordMatch ("autocommit", 10, token, tokenLen)) {
        setAutocommit (session, parseBool (value, valueLen, false));
    }
    else if (keywordMatch ("cursornames", 11, token, tokenLen)) {
        enableCursorNames (session, parseBool (value, valueLen, true));
    }
}

/*----------------------------------------*/

static void
parseSessionConfig (
    tin01_sql_session * session,
    const char * configString)
{
    ROUTINE_DBG_MSP00 ("parseSessionConfig");
    const char * argStart;
    const char * nextConfig;
    bool atEnd = (configString [0] == '\0');
    int tokenLen;
    int argLen;

    while (!atEnd) {
        argStart = strchr (configString, '=');
        if (argStart == NULL) {
            return;
        }
        tokenLen = argStart - configString;
        ++argStart;
        nextConfig = strchr (argStart, '&');
        if (nextConfig == NULL) {
            atEnd = true;
            nextConfig = argStart + strlen (argStart);
        }
        argLen = nextConfig - argStart;
        handleConfigArgument (session, configString, tokenLen, argStart, argLen);
        configString = nextConfig + 1;
    }

}

/*----------------------------------------*/

/* endsnippet Generic_precode */
/* snippet Perl_precode */

static int
allocFailed (void * ptr, char * croakBuf)
{
    if (ptr == NULL) {
        strcpy (croakBuf, "Out of memory");
        return 1;
    }
    else {
        return 0;
    }
}

static bool
SL_isString (
    SV * value)
{
    ROUTINE_DBG_MSP00 ("SL_isString");
    bool result;

    if (SvPOK (value)) {
        result = true;
    }
    else {
        result = false;
    }
    return result;
}

/*----------------------------------------*/

static bool
SL_isCallable (
    SV* value)
{
    ROUTINE_DBG_MSP00 ("SL_isCallable");
    bool result = false;
    SV*  referenced;

    if (SvTYPE (value) != SVt_RV) {
        return false;
    }
    referenced = SvRV (value);
    if (SvTYPE (referenced) == SVt_PVCV) {
        result = true;
    }
    return result;
}

/*----------------------------------------*/

static void
SL_getRawString (
    SV * value,
    const char ** chars,
    int * len)
{
    size_t perlValue;
    *chars = SvPV (value, perlValue);
    *len = (int) perlValue;
}

/*----------------------------------------*/

static int
SL_getNumber (
    SV * value,
    tsp1_param_info * paramInfo,
    tsp00_Number * number)
{
    ROUTINE_DBG_MSP00 ("SL_getNumber");
    double floatValue;
    tsp00_NumError numrc;
    int frac;

    if ((paramInfo->sp1i_data_type == dfloat) || (paramInfo->sp1i_data_type == dfloat)) {
        frac = csp_float_frac;
    }
    else {
        frac = paramInfo->sp1i_frac;
    }
    floatValue = SvNV (value);
    s41plrel (number, 1, csp_fixed, frac, floatValue, REFPARM (numrc));
    return 1;
}

/*----------------------------------------*/

static SL_ValueT *
SL_fromNone ()
{
    ROUTINE_DBG_MSP00 ("SL_fromNone");

    return &PL_sv_undef;
}

/*----------------------------------------*/

static SL_ValueT *
SL_fromInt (
    long value)
{
    ROUTINE_DBG_MSP00 ("SL_fromInt");

    return newSViv (value);
}

/*----------------------------------------*/

static SL_ValueT *
SL_fromDouble (
    double value)
{
    ROUTINE_DBG_MSP00 ("SL_fromDouble");

    return newSVnv (value);
}

/*----------------------------------------*/

static SL_ValueT *
SL_fromString (
    const char * value,
    int len)
{
    ROUTINE_DBG_MSP00 ("SL_fromString");

    if (len == -1) {
        len = strlen (value);
    }
    if (len == 0) {
        value = "";   /* because newSVpv uses 0 as 'unknown' PTS 1104234 */
    }
    return newSVpv (CONST_CAST (char *, value), len);
}

/*----------------------------------------*/

static SL_ValueT *
SL_fromUCS2 (
    const void * value,
    int charLen)
{
    ROUTINE_DBG_MSP00 ("SL_fromUCS2");
    SL_ValueT * result;
    size_t perlValue;

    tsp78ConversionResult convResult;
    tsp00_Uint4 destBytesWritten;
    tsp00_Uint4 srcBytesParsed;

    /* alloc object */
    result = newSV (charLen);
    if (result == NULL) {
        return result;
    }
    /* try conversion */
    convResult = sp78convertString (
        sp77encodingAscii, SvPV (result, perlValue), charLen, &destBytesWritten, false,
        sp77encodingUCS2Native, value, charLen * 2, &srcBytesParsed);
    /* fail: delete object */
    if (convResult != sp78_Ok) {
        SvREFCNT_dec (result);
        result = SL_fromNone ();
    }
    return result;
}

/*----------------------------------------*/

static int
SL_readCall (
    SV * value,
    byte_t * streamBuf,
    int requestedLen)
{
    ROUTINE_DBG_MSP00 ("SL_readCall");
    int  resultCount;
    SV * callResult;
    const char * rawResult;
    size_t readResult = 0;
    dSP;

    /*
     * call the Callable object
     */
    /* ENTER; */
    /* SAVETMPS; */
    PUSHMARK (sp);
    XPUSHs (sv_2mortal (newSViv (requestedLen)));
    PUTBACK;
    resultCount = perl_call_sv (value, G_SCALAR | G_EVAL);
    SPAGAIN;
    if (SvTRUE (GvSV (PL_errgv))) {
        POPs;
    }
    else {
        callResult = POPs;
    }
    PUTBACK;
    /* FREETMPS; */
    /* LEAVE; */
    if (resultCount != 1) {
        return -1;
    }
    /*
     * get string data from result
     */
    rawResult = SvPV (callResult, readResult);
    memcpy (streamBuf, rawResult, readResult);
    return (int) readResult;
}

/*----------------------------------------*/

static void
SL_forget (
    SV * value)
{
    sv_2mortal (value);
}

/*----------------------------------------*/

typedef struct SequenceT {
    AV * perlSeq;
} SequenceT;

static SequenceT EmptySequenceC = {NULL};

static int
sequencePerl2C (
    SV * perlRef,
    SequenceT * cSeq,
    char * croakBuf)
{
    ROUTINE_DBG_MSP00 ("sequencePerl2C");

    if (SvTYPE (perlRef) == SVt_PVAV) {
        cSeq->perlSeq = REINTERPRET_CAST (AV*, perlRef);
    }
    else {
        if (!SvROK (perlRef)) {
            sp77sprintf (croakBuf, CROAKBUF_SIZE, "Arguments must be given as reference");
            return 0;
        }
        if (SvTYPE (SvRV (perlRef)) != SVt_PVAV) {
            sp77sprintf (croakBuf, CROAKBUF_SIZE, "Arguments must be given as reference");
            return 0;
        }
        cSeq->perlSeq = REINTERPRET_CAST (AV*, SvRV (perlRef));
    }
    return 1;
}

/*----------------------------------------*/

static int
sequenceLength (
    SequenceT * cSeq)
{
    ROUTINE_DBG_MSP00 ("sequenceLength");
    int result;

    if (cSeq->perlSeq == NULL) {
        result = 0;
    }
    else {
        result = av_len (cSeq->perlSeq) + 1;
    }
    return result;
}

/*----------------------------------------*/

static SV *
getSequenceElement (
    SequenceT * cSeq,
    int index)
{
    ROUTINE_DBG_MSP00 ("getSequenceElement");
    SV ** fetchResult;
    SV * result;

    fetchResult = av_fetch (cSeq->perlSeq, index, 0);
    result = *fetchResult;
    return result;
}

/*----------------------------------------*/

typedef struct SQLResultT {
    union {
        SV * resultObj;
        AV * resultList;
    } value;
    bool isTuple;
    bool isOK;
    int additionalValues;
} SQLResultT;

static int
sqlResultC2Py (
    void * dummy,
    SQLResultT result,
    SV ** perlResult)
{
    ROUTINE_DBG_MSP00 ("sqlResultC2Py");
    *perlResult = result.value.resultObj;
    return 1;
}

/*----------------------------------------*/

static SL_ValueT *
sqlResultC2SL_Value (
    SQLResultT * result)
{
    ROUTINE_DBG_MSP00 ("sqlResultC2SL_Value");
    return result->value.resultObj;
}

/*----------------------------------------*/

static void
initResult (
    SQLResultT * result,
    int additional)
{
    ROUTINE_DBG_MSP00 ("initResult");
    result->value.resultObj = NULL;
    result->isTuple = false;
    result->isOK = true;
    result->additionalValues = additional;
}

/*----------------------------------------*/

static bool
allocResult (
    SQLResultT * result,
    int elements)
{
    ROUTINE_DBG_MSP00 ("allocResult");
    bool allocOK = true;

    /* ASSERT_DBG_MSP00 (result->value.resultObj == NULL); */
    result->value.resultList = newAV ();
    result->isTuple = true;
    allocOK = (result->value.resultObj != 0);
    if (allocOK) {
        av_unshift (result->value.resultList, elements + result->additionalValues);
    }
    return allocOK;
}

/*----------------------------------------*/

static bool
setResult (
    SQLResultT * result,
    SV * newElement)
{
    ROUTINE_DBG_MSP00 ("setResult");

    /* ASSERT_DBG_MSP00 (result->value.resultObj == NULL); */
    result->isTuple = false;
    if (newElement == NULL) {
        newElement = &PL_sv_undef;
    }
    result->value.resultObj = newElement;
    return true;
}

/*----------------------------------------*/

static bool
setResultElement (
    SQLResultT * result,
    int index,
    SV * newElement)
{
    ROUTINE_DBG_MSP00 ("setElement");
    bool assignOK = true;

    if (result->isTuple) {
        if (index < 0) {
            index = av_len (result->value.resultList) - index;
        }
        av_store (result->value.resultList, index, newElement);
    }
    else {
        if (index == 0) {
            assignOK = setResult (result, newElement);
        }
        else {
            sv_2mortal (newElement);
            /* makeError tuple assignment index out of range" */
            assignOK = false;
        }
    }
    return assignOK;
}

/*----------------------------------------*/

static int
sqlResultOK (
    SQLResultT * result)
{
    ROUTINE_DBG_MSP00 ("sqlResultOK");
    return(result->isOK);
}

/*----------------------------------------*/

static boolean
sqlErrOccured (
    tin01_sql_session * nself,
    char * croakBuf)
{
    ROUTINE_DBG_MSP00 ("sqlErrOccurred");
    boolean errOccurred = false;

    if (nself->lasterr.sqlresult.returnCode != 0) {
        errOccurred = true;
        if (nself->lasterr.sqlresult.errorPos > 1) {
            sp77sprintf (croakBuf, CROAKBUF_SIZE, "%s: (%d) (at %d) %s", SQLError_sapdb,
                nself->lasterr.sqlresult.returnCode,
                nself->lasterr.sqlresult.errorPos,
                CONST_CAST (const char *, nself->lasterr.sqlresult.sqlmsg));
        }
        else {
            sp77sprintf (croakBuf, CROAKBUF_SIZE, "%s: (%d) %s", SQLError_sapdb,
                nself->lasterr.sqlresult.returnCode,
                CONST_CAST (const char *, nself->lasterr.sqlresult.sqlmsg));
        }
    }
    else if (nself->lasterr.sp_rc != 0) {
        errOccurred = true;
        sp77sprintf (croakBuf, CROAKBUF_SIZE, "%s: %s",
            CommunicationError_sapdb,
            STATIC_CAST (const char *, nself->lasterr.errtext));
    }
    return errOccurred;
}

/*----------------------------------------*/

static boolean
isSessionConnected (
    tin01_sql_session * session,
    char * croakBuf)
{
    ROUTINE_DBG_MSP00 ("isSessionConnected");

    if ((session != NULL) && session->is_connected) {
        return true;
    }
    sp77sprintf (croakBuf, CROAKBUF_SIZE, "%s: Invalid Session", CommunicationError_sapdb);
    return false;
}

/*----------------------------------------*/

static INLINE int
requestReceive (
    tin01_sql_session * session,
    bool withInfo,
    bool allowRetry)
{
    ROUTINE_DBG_MSP00 ("requestReceive");
    int rc;

    rc = i28sql (session, NULL);
    if (rc == cin01_db_not_accessible) {
        rc = i28connect (session);
        if (rc == cin01_db_ok) {
            tin01_sqlresult * sqlresult = &(session->lasterr.sqlresult);
            sqlresult->sqlmsg [0] = '\0';
            if (lastWasEOT (session) && allowRetry) {
                rc = cin01_db_parse_again;
            }
            else {
                rc = cin01_db_cmd_timeout;
                strcpy (sqlresult->sqlmsg,
                    "Session inactivity timeout (work rolled back)");
            }
            sqlresult->returnCode = rc;
        }
    }
    clearEOTFlag (session);
    return rc;
}

/*----------------------------------------*/

/* endsnippet Perl_precode */
static const char * SapDB_LongReaderClassID = "SapDB_LongReader";

typedef struct SapDB_LongReaderObjectT {
    const char * classID;
    LongReaderT nself;
/* no code for key SapDB_LongReader Perl_/Generic_ in ['Generic_precode', 'Python_precode']*/
} SapDB_LongReaderObjectT;

#define isSapDB_LongReaderObject(op) ((op)->classID == SapDB_LongReaderClassID)

static const char * notClassSapDB_LongReader_C = "Object not of class SapDB_LongReader";

static SapDB_LongReaderObjectT *
newSapDB_LongReader ()
{
    SapDB_LongReaderObjectT * result;

    Newz (0, result, 1, SapDB_LongReaderObjectT);
    return result;
}

static void
freeSapDB_LongReader (
    SapDB_LongReaderObjectT * value)
{
    Safefree (value);
}
/* snippet SapDB_LongReader Generic_precode */

#define DESCFIELD(descriptor, name) (descriptor.variant.C_false.name##_F)
#if defined (__cplusplus)
#define BECOMES(lvalue, expr)       lvalue.becomes (expr)
#else
#define BECOMES(lvalue, expr)       lvalue = expr
#endif
/*----------------------------------------*/

static void
setReaderState (
    LongReaderT * nself,
    char * longBuf)
{
    ROUTINE_DBG_MSP00 ("setReaderState");
    tsp00_ValMode_Enum valMode = DESCFIELD (nself->descriptor, ld_valmode);
    int valLen = DESCFIELD (nself->descriptor, ld_vallen);
    const char * longData;

    if ((valMode == vm_datapart)
            || (valMode == vm_alldata)
            || (valMode == vm_lastdata)) {
        if (valLen > nself->dataSize) {
            if (nself->dataBuf == NULL) {
                nself->dataBuf = REINTERPRET_CAST (byte_t*, malloc (valLen));
            }
            else {
                nself->dataBuf = REINTERPRET_CAST (byte_t*, realloc (nself->dataBuf, valLen));
            }
            memset (nself->dataBuf, '%', valLen);
            nself->dataSize = valLen;
        }
        longData = longBuf + DESCFIELD (nself->descriptor, ld_valpos) - 1;
        memcpy (nself->dataBuf, longData, valLen);
        nself->dataPos = 0;
        nself->dataLen = valLen;
    }
    if ((valMode == vm_alldata) || (valMode == vm_lastdata)) {
        nself->lastGetvalFound = true;
    }
}

/*----------------------------------------*/

static void
initLongReader (
    LongReaderT * nself,
    tin01_sql_session * session,
    tsp00_LongDescriptor * descriptor)
{
    ROUTINE_DBG_MSP00 ("initLongReader");

    nself->session = session;
    increaseRef (session->more_data);
    memcpy (&nself->descriptor, descriptor, sizeof (*descriptor));
    setReaderState (nself, i28argsbuf (session));
}

/*----------------------------------------*/

#if defined (PYTHON_GLUE)
static PyObject *
createLongReader (
    tin01_sql_session * session,
    tsp00_LongDescriptor * descriptor)
{
    ROUTINE_DBG_MSP00 ("createLongReader");
    SapDB_LongReaderObjectT * reader = newSapDB_LongReader ();

    if (reader == NULL) {
        return NULL;
    }
    initLongReader (&reader->nself, session, descriptor);
    return REINTERPRET_CAST (PyObject *, reader);
}

/*----------------------------------------*/

#elif defined (PERL_GLUE)
static SV *
createLongReader (
    tin01_sql_session * session,
    tsp00_LongDescriptor * descriptor)
{
    ROUTINE_DBG_MSP00 ("createLongReader");
    char croakBuf [200];
    int resultCount;
    SV * perlObject = NULL;
    SapDB_LongReaderObjectT * reader = NULL;
    dSP;

    /*
     * create resultset object
     */
    /* ENTER; */
    /* SAVETMPS; */
    PUSHMARK (sp);
    PUTBACK;
    resultCount = perl_call_pv (LIT("sapdb::createReaderZombie"),
                      G_SCALAR | G_EVAL);
    SPAGAIN;
    if (SvTRUE (GvSV (PL_errgv))) {
        POPs;
    }
    else {
        perlObject = POPs;
        SvREFCNT_inc (perlObject);  /* increase refcount to prevent delete */
    }
    PUTBACK;
    /* FREETMPS; */
    /* LEAVE; */
    /*
     * retrieve actual pointer from this mess
     */
    if (resultCount != 1) {
        return NULL;
    }
    reader = REINTERPRET_CAST (SapDB_LongReaderObjectT*,
         getGluePointer (perlObject, SapDB_LongReaderClassID, croakBuf));
    if ((reader == NULL) || !isSapDB_LongReaderObject (reader)) {
        sv_2mortal (perlObject);
        return NULL;
    }
    /*
     * fill reader object with soul
     */
    initLongReader (&reader->nself, session, descriptor);
    return perlObject;
}
#endif
/*----------------------------------------*/

static void
closeReader (
    LongReaderT * nself)
{
    ROUTINE_DBG_MSP00 ("closeReader");
    if (nself->dataBuf != NULL) {
        free (nself->dataBuf);
    }
    decreaseRef (nself->session->more_data);
}

/*----------------------------------------*/

#if defined (PYTHON_GLUE)
static void
destructor_SapDB_LongReader (
    PyObject * pyself)
{
    ROUTINE_DBG_MSP00 ("destructor_SapDB_LongReader");
    SapDB_LongReaderObjectT* self = REINTERPRET_CAST (SapDB_LongReaderObjectT*, pyself);
    LongReaderT * nself = &self->nself;
    closeReader (nself);
    free (pyself);
}
#endif

/*----------------------------------------*/

static int
doGetval (
    LongReaderT * nself)
{
    ROUTINE_DBG_MSP00 ("doGetval");
    tin01_sql_session * session = nself->session;
    int rc;
    char * resultBuf;

    i28initspecial (session, sp1m_getval);
    i28newpart (session, sp1pk_longdata);
    i28parg (session, REINTERPRET_CAST (const char *, &nself->descriptor),
        sizeof (tsp00_LongDescriptor), sizeof (tsp00_LongDescriptor),
        1, '\0');
    rc = requestReceive (session, false, false);
    if (rc != cin01_db_ok) {
        nself->lastGetvalFound = true;
        nself->dataLen = 0;
        nself->dataPos = 0;
        return rc;
    }
    i28findpart (session, sp1pk_longdata);
    resultBuf = i28argsbuf (session);
    memcpy (&nself->descriptor, resultBuf + 1, sizeof (tsp00_LongDescriptor));
    setReaderState (nself, resultBuf);
    return 0;
}

/*----------------------------------------*/

static SL_ValueT *
readLong (
    LongReaderT * nself,
    int requested)
{
    ROUTINE_DBG_MSP00 ("readLong");
    SL_ValueT * result = NULL;
    char * resultBuf;
    long   resultLen = 0;
    int remaining = nself->dataLen - nself->dataPos;

    resultBuf = REINTERPRET_CAST (char*, malloc (requested));
    while (requested > 0) {
        if (remaining >= requested) {
            /* copy requested */
            memcpy (resultBuf + resultLen, nself->dataBuf + nself->dataPos, requested);
            nself->dataPos += requested;
            resultLen += requested;
            requested = 0;
        }
        else {
            /* copy remaining */
            if (remaining > 0) {
                memcpy (resultBuf + resultLen, nself->dataBuf + nself->dataPos, remaining);
                requested -= remaining;
                resultLen += remaining;
                nself->dataPos += remaining;
            }
            if (nself->lastGetvalFound) {
                break;
            }
            doGetval (nself);
            remaining = nself->dataLen - nself->dataPos;
        }
    }
    result = SL_fromString (resultBuf, resultLen);
    free (resultBuf);
    return result;
}

/* endsnippet SapDB_LongReader Generic_precode */
/*----------------------------------------*/

externC XS (SapDB_LongReader_read)
{
    ROUTINE_DBG_MSP00 ("SapDB_LongReader_read");
/* no code for key Perl_/Generic_code */
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_LongReaderObjectT* self;
    LongReaderT* nself;
    int count;
    SL_ValueT * result;
    
    cv = cv;
    if (items != 2) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_LongReaderObjectT*) getGluePointer (ST(0), SapDB_LongReaderClassID, croakBuf);
    if ((self == NULL) || !isSapDB_LongReaderObject (self)) {
        croak (croakBuf);
    }
    nself = &self->nself;
    switch (items) {
        case 2: count = (int) SvIV (ST(1));
    }
    if (!ok) {
        ok = 0;
        goto cleanup_label;
    }
    if (!isSessionConnected (self->nself.session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Perl_/Generic_nativeCall */
    result = readLong (nself, count);
    /* outargs */
    ST (argvi) = sv_newmortal ();
    sv_setsv (ST (argvi++), result);
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        croak (croakBuf);
    }
    XSRETURN (argvi);
}


/*----------------------------------------*/

externC XS (SapDB_LongReader_destructor)
{
    ROUTINE_DBG_MSP00 ("SapDB_LongReader_destructor");
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_LongReaderObjectT* self;
    LongReaderT* nself;

    cv = cv;
    if (items != 1) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_LongReaderObjectT*) getGluePointer (ST(0), SapDB_LongReaderClassID, croakBuf);
    if ((self == NULL) || !isSapDB_LongReaderObject (self)) {
        croak (croakBuf);
    }
    nself = &self->nself;
    closeReader (nself);
    freeSapDB_LongReader (self);
    XSRETURN (0);
}

static const char * SapDB_ResultSetClassID = "SapDB_ResultSet";

typedef struct SapDB_ResultSetObjectT {
    const char * classID;
    ResultSetT nself;
/* no code for key SapDB_ResultSet Perl_/Generic_ in ['Generic_precode']*/
} SapDB_ResultSetObjectT;

#define isSapDB_ResultSetObject(op) ((op)->classID == SapDB_ResultSetClassID)

static const char * notClassSapDB_ResultSet_C = "Object not of class SapDB_ResultSet";

static SapDB_ResultSetObjectT *
newSapDB_ResultSet ()
{
    SapDB_ResultSetObjectT * result;

    Newz (0, result, 1, SapDB_ResultSetObjectT);
    return result;
}

static void
freeSapDB_ResultSet (
    SapDB_ResultSetObjectT * value)
{
    Safefree (value);
}
/* snippet SapDB_ResultSet Generic_precode */
static int
openCursor (
    tin01_sql_session * session,
    ResultSetT * resultSet);

/*----------------------------------------*/
#if defined (PYTHON_GLUE)
static PyObject *
createResultSetFromReply (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("createResultSetFromReply");
    SapDB_ResultSetObjectT * newObj = newSapDB_ResultSet ();
    int rc;

    if (newObj == NULL) {
        return NULL;
    }
    rc = openCursor (session, &newObj->nself);
    if (rc != 0) {
        sqlErrOccured (session, NULL);
        Py_DECREF (newObj);
        newObj = NULL;
    }
    return REINTERPRET_CAST(PyObject *, newObj);
}
#elif defined (PERL_GLUE)
static SV *
createResultSetFromReply (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("createResultSetFromReply");
    char croakBuf [200];
    int resultCount;
    SV * perlObject = NULL;
    SapDB_ResultSetObjectT * resultSet = NULL;
    int rc;
    dSP;

    /*
     * create resultset object
     */
    /* ENTER; */
    /* SAVETMPS; */
    PUSHMARK (sp);
    PUTBACK;
    resultCount = perl_call_pv (LIT("sapdb::createCursorZombie"),
                      G_SCALAR | G_EVAL);
    SPAGAIN;
    if (SvTRUE (GvSV (PL_errgv))) {
        POPs;
    }
    else {
        perlObject = POPs;
    }
    PUTBACK;
    /* FREETMPS; */
    /* LEAVE; */
    /*
     * retrieve actual pointer from this mess
     */
    if (resultCount != 1) {
        return NULL;
    }
    resultSet = REINTERPRET_CAST (SapDB_ResultSetObjectT*,
         getGluePointer (perlObject, SapDB_ResultSetClassID, croakBuf));
    if ((resultSet == NULL) || !isSapDB_ResultSetObject (resultSet)) {
        sv_2mortal (perlObject);
        return NULL;
    }
    /*
     * fill resultset object with soul
     */
    rc = openCursor (session, &resultSet->nself);
    if (rc != 0) {
        sv_2mortal (perlObject);
        return NULL;
    }
    return perlObject;
}
#endif
/*----------------------------------------*/

static void
getOutputParameter (
    tin01_sql_session * session,
    ParseInfoT * parseInfo,
    SQLResultT * result)
{
    ROUTINE_DBG_MSP00 ("getOutputParameter");
    int         outputIndex = -1;
    int         i;
    tsp1_param_info * paramInfo = parseInfo->paramInfo;
    SL_ValueT * value;
    void      * addr;
    int         intVal;
    double      floatVal;
    tsp00_NumError  numErr;
    int         valLen;

    allocResult (result, parseInfo->outputParamCount);
    for (i = 0; i < parseInfo->paramCount; ++i, ++paramInfo) {
        if (paramInfo->sp1i_io_type != sp1io_input) {
            ++outputIndex;
            addr = i28argaddr (session, paramInfo->sp1i_bufpos);
            if (addr == NULL) {
                value = SL_fromNone ();
            }
            else {
                switch (paramInfo->sp1i_data_type) {
                    case dfixed:
                        if (paramInfo->sp1i_frac == 0) {
                            s40glint (REINTERPRET_CAST (tsp00_Number*, addr),
                                1, paramInfo->sp1i_length,
                                REFPARM (intVal), REFPARM (numErr));
                            value = SL_fromInt (intVal);
                        }
                        else {
                            s40glrel (REINTERPRET_CAST (tsp00_Number*, addr),
                                1, paramInfo->sp1i_length,
                                REFPARM (floatVal), REFPARM (numErr));
                            value = SL_fromDouble (floatVal);
                        }
                        break;
                    case dfloat:
                    case dvfloat:
                        s40glrel (REINTERPRET_CAST (tsp00_Number*, addr),
                            1, paramInfo->sp1i_length,
                            REFPARM (floatVal), REFPARM (numErr));
                        value = SL_fromDouble (floatVal);
                        break;
                    case dsmallint:
                    case dinteger:
                        s40glint (REINTERPRET_CAST (tsp00_Number*, addr),
                            1, paramInfo->sp1i_length,
                            REFPARM (intVal), REFPARM (numErr));
                        value = SL_fromInt (intVal);
                        break;
                    case dcha:
                    case dche:
                    case dvarchara:
                    case dvarchare:
                    case ddate:
                    case dtime:
                    case dtimestamp:
                        valLen = s30klen (addr, ' ', paramInfo->sp1i_length);
                        value = SL_fromString (REINTERPRET_CAST(const char *, addr),
                            valLen);
                        break;
                    case dchb:
                    case dvarcharb:
                        valLen = s30klen (addr, '\0', paramInfo->sp1i_length);
                        value = SL_fromString (REINTERPRET_CAST (const char *, addr),
                            valLen);
                        break;
                    case dstra:
                    case dstre:
                    case dstrb:
                    case dlonga:
                    case dlonge:
                    case dlongb:
                        value = createLongReader (session,
                            REINTERPRET_CAST (tsp00_LongDescriptor*, addr));
                        break;
                    default:
                        value = SL_fromNone ();
                        break;
                }
            }
            setResultElement (result, outputIndex, value);
        }
    }
}

/*----------------------------------------*/

static void *
argAddr (
    byte_t * dataBuf,
    int bufPos)
{
    ROUTINE_DBG_MSP00 ("argAddr");
    byte_t    *bufaddr;
    byte_t        *result;

    bufaddr = dataBuf + bufPos;
    if (bufaddr [-1] == (unsigned char) csp_undef_byte)
        result = NULL;
    else
        result = bufaddr;
    return result;
}

/*----------------------------------------*/

static void
getOutputParameter2 (
    tin01_sql_session * session,
    byte_t * dataBuf,
    ParseInfoT * parseInfo,
    SQLResultT * result)
{
    ROUTINE_DBG_MSP00 ("getOutputParameter");
    int         outputIndex = -1;
    int         i;
    tsp1_param_info * paramInfo = parseInfo->paramInfo;
    SL_ValueT * value;
    void      * addr;
    int         intVal;
    double      floatVal;
    tsp00_NumError  numErr;
    int         valLen;

    allocResult (result, parseInfo->outputParamCount);
    for (i = 0; i < parseInfo->paramCount; ++i, ++paramInfo) {
        if (paramInfo->sp1i_io_type != sp1io_input) {
            ++outputIndex;
            addr = argAddr (dataBuf, paramInfo->sp1i_bufpos);
            if (addr == NULL) {
                value = SL_fromNone ();
            }
            else {
                switch (paramInfo->sp1i_data_type) {
                    case dfixed:
                        if (paramInfo->sp1i_frac == 0) {
                            s40glint (REINTERPRET_CAST (tsp00_Number*, addr),
                                1, paramInfo->sp1i_length,
                                REFPARM (intVal), REFPARM (numErr));
                            if (numErr == num_ok) {
                                value = SL_fromInt (intVal);
                            }
                        }
                        else {
                            s40glrel (REINTERPRET_CAST (tsp00_Number*, addr),
                                1, paramInfo->sp1i_length,
                                REFPARM (floatVal), REFPARM (numErr));
                            if (numErr == num_ok) {
                                value = SL_fromDouble (floatVal);
                            }
                        }
                        if (numErr != num_ok) {
                            char strbuf [100];
                            int outlen;
                            s42gchr (addr, 1, paramInfo->sp1i_length,
                                paramInfo->sp1i_frac, paramInfo->sp1i_length,
                                strbuf, 1, REFPARM (outlen), REFPARM (numErr));
                            value = SL_fromString (strbuf, outlen);
                        }
                        break;
                    case dfloat:
                    case dvfloat:
                        s40glrel (REINTERPRET_CAST (tsp00_Number*, addr),
                            1, paramInfo->sp1i_length,
                            REFPARM (floatVal), REFPARM (numErr));
                        value = SL_fromDouble (floatVal);
                        break;
                    case dsmallint:
                    case dinteger:
                        s40glint (REINTERPRET_CAST (tsp00_Number*, addr),
                            1, paramInfo->sp1i_length,
                            REFPARM (intVal), REFPARM (numErr));
                        value = SL_fromInt (intVal);
                        break;
                    case dcha:
                    case dche:
                    case dvarchara:
                    case dvarchare:
                    case ddate:
                    case dtime:
                    case dtimestamp:
                        valLen = s30klen (addr, ' ', paramInfo->sp1i_length);
                        value = SL_fromString (REINTERPRET_CAST (const char *, addr),
                            valLen);
                        break;
                    case dchb:
                    case dvarcharb:
                        valLen = s30klen (addr, '\0', paramInfo->sp1i_length);
                        value = SL_fromString (REINTERPRET_CAST (const char *, addr),
                            valLen);
                        break;
                    case dunicode:
                    case dvarcharuni:
                        valLen = paramInfo->sp1i_length
                            - sp77encodingUCS2Native->countPadChars (addr, paramInfo->sp1i_length * 2, ' ');
                        value = SL_fromUCS2 (addr, valLen);
                        break;
                    case dboolean:
                        if (((byte_t*)addr)[0] == '\0') {
                            value = SL_fromString ("", 0);
                        }
                        else {
                            value = SL_fromInt (1);
                        }
                        break;
                    case dstra:
                    case dstre:
                    case dstrb:
                    case dlonga:
                    case dlonge:
                    case dlongb:
                        value = createLongReader (session, REINTERPRET_CAST (tsp00_LongDescriptor*, addr));
                        break;
                    default:
                        value = SL_fromNone ();
                        break;
                }
            }
            setResultElement (result, outputIndex, value);
        }
    }
}

/*----------------------------------------*/

static int
doDescribe (
    tin01_sql_session * session,
    const char * cursorName)
{
    ROUTINE_DBG_MSP00 ("doDescribe");
    int rc;

    if (cursorName [0] != '\0') {
        rc = i28adbsf (session, "DESCRIBE \"%s\"", cursorName);
    }
    else {
        rc = i28adbs_noinfo  (session, "DESCRIBE", false);
    }
    return rc;
}

/*----------------------------------------*/

static int
openCursor (
    tin01_sql_session * session,
    ResultSetT * resultSet)
{
    ROUTINE_DBG_MSP00 ("openCursor");
    int nameLen = -1;
    int rc = 0;
    int infoSize;
    int i;

    resultSet->session = session;
    increaseRef (session->more_data);
    for (i = 0; i < fetchKindsE; ++i) {
        resultSet->fetchInfo [i].paramCount = -1;
    }
    memset (resultSet->cursorName, '\0', sizeof (resultSet->cursorName));
    i28gresulttablename (session, &nameLen, resultSet->cursorName);
    if (i28findpart (session, sp1pk_columnnames)) {
        nameLen = i28nameslen (session);
    }
    else {
        nameLen = -1;
    }
    if (nameLen == -1) {
        rc = doDescribe (session, resultSet->cursorName);
        if (rc == 0) {
            nameLen = i28nameslen (session);
        }
    }
    if (rc == 0) {
        resultSet->colNames = REINTERPRET_CAST (const char *, malloc (nameLen));
        memcpy ((void *) resultSet->colNames , i28colnames (session), nameLen);
        resultSet->colCount = i28paramcount (session);
        infoSize = sizeof (tsp1_param_info) * resultSet->colCount;
        resultSet->colInfo = REINTERPRET_CAST (tsp1_param_info *, malloc (infoSize));
        memcpy (resultSet->colInfo, i28gparaminfo (session, 0), infoSize);
        resultSet->currentPos = 0;
    }
    return rc;
}

/*----------------------------------------*/

static bool
row2result (
    tin01_sql_session * session,
    ResultSetT * resultSet,
    SQLResultT * result)
{
    ROUTINE_DBG_MSP00 ("row2result");

    if (i28findpart (session, sp1pk_data)) {
        getOutputParameter (session, &resultSet->fetchInfo [nextFetchE], result);
    }
    else {
        setResult (result, NULL);
    }
    return true;
}

/*----------------------------------------*/

static int
getFetchBufPos (
    ResultSetT * resultSet)
{
    ROUTINE_DBG_MSP00 ("getFetchBufPos");
    return(resultSet->currentRow - resultSet->dataBuf)
    / resultSet->rowSize;
}

/*----------------------------------------*/

static int
getFetchBufRecords (
    ResultSetT * resultSet)
{
    ROUTINE_DBG_MSP00 ("getFetchBufRecords");
    return(resultSet->currentRow - resultSet->eoRecords)
    / resultSet->rowSize;
}

/*----------------------------------------*/

static boolean
moveFetchBufPos (
    ResultSetT * resultSet,
    int relative)
{
    ROUTINE_DBG_MSP00 ("moveFetchBufPos");
    boolean result;
    byte_t * newPos = resultSet->currentRow + (resultSet->rowSize * relative);
    if ((resultSet->dataBuf <= newPos) && (newPos < resultSet->eoRecords)) {
        result = true;
        resultSet->currentRow = newPos;
    }
    else {
        result = false;
    }
    return result;
}

/*----------------------------------------*/

static int
calculateFetchRecord (
    ParseInfoT * parseInfo)
{
    int result = 0;
    int i;

    for (i = 0; i < parseInfo->paramCount; ++i) {
        tsp1_param_info * paramInfo = &parseInfo->paramInfo [i];
        if (paramInfo->sp1i_io_type == sp1io_output) {
            result += paramInfo->sp1i_in_out_len;
        }
    }
    return result;
}

/*----------------------------------------*/

static int
parseFetch (
    ResultSetT * resultSet,
    fetchKindsT fetchKind)
{
    ROUTINE_DBG_MSP00 ("parseFetch");
    tin01_sql_session * session = resultSet->session;
    ParseInfoT * parseInfo;
    int rc;
    const char * fmt;
    int i;

    i28initparse (session);
    if (strlen (resultSet->cursorName) > 0) {
        fmt = "Fetch %s \"%s\"";
    }
    else {
        fmt = "Fetch %s ";
    }
    i28pcmdf (session, fmt,
        fetchCommandsC [fetchKind], STATIC_CAST (char *, resultSet->cursorName));
    i28pcmd (session, "into ?", -1);
    for (i = 1; i < resultSet->colCount; ++i) {
        i28pcmd (session, ", ?", -1);
    }
#if defined (__cplusplus)
    session->segment->sp1c_sqlmode ()
        = tsp1_sqlmode::fromInt (sp1sm_adabas);
#else
    session->segment->variant.C_3.sp1c_sqlmode_F = sp1sm_adabas;
#endif
    session->segment->variant.C_3.sp1c_mass_cmd_F = true;
    rc = requestReceive (session, false, false);
    if (rc != cin01_db_ok) {
        return rc;
    }
    parseInfo = &resultSet->fetchInfo [fetchKind];
    if (!fillParamInfos (session, parseInfo, true)) {
        return internalErrorC;
    }
    if (resultSet->rowSize == 0) {
        resultSet->rowSize = calculateFetchRecord (parseInfo);
    }
    return cin01_db_ok;
}

/*----------------------------------------*/

static int
copyFetchBuffer (
    ResultSetT * resultSet)
{
    ROUTINE_DBG_MSP00 ("copyFetchBuffer");
    int partLen;
    byte_t * newBuf;
    int recordsInFetch;

    if (!i28findpart (resultSet->session, sp1pk_data)) {
        return cin01_db_row_not_found;
    }
    if (resultSet->session->part->variant.C_2.sp1p_arg_count_F == 0) {
        /* workaround: current bug in fetch relative */
        return cin01_db_row_not_found;
    }
    /*
     * make sure buffer is large enough
     */
    partLen = i28partlen (resultSet->session);
    if (resultSet->dataSize < partLen) {
        newBuf = REINTERPRET_CAST (byte_t*, realloc (resultSet->dataBuf, partLen));
        if (newBuf == NULL) {
            return internalErrorC;
        }
        else {
            resultSet->dataBuf = newBuf;
            resultSet->dataSize = partLen;
        }
    }
    /*
     * copy data
     */
    resultSet->dataLen = partLen;
    memcpy (resultSet->dataBuf, i28argsbuf (resultSet->session), partLen);
    /*
     * set currentRow and eoRecords
     */
    recordsInFetch = resultSet->session->part->variant.C_2.sp1p_arg_count_F;
    if (recordsInFetch > resultSet->maxRecordsPerFetch) {
        resultSet->maxRecordsPerFetch = recordsInFetch;
    }
    resultSet->currentRow = resultSet->dataBuf;
    resultSet->eoRecords = resultSet->dataBuf
                           + (recordsInFetch * resultSet->rowSize);
    return cin01_db_ok;
}

/*----------------------------------------*/

static int
doFetch (
    ResultSetT * resultSet,
    fetchKindsT fetchKind,
    int parameter)
{
    ROUTINE_DBG_MSP00 ("doFetch");
    tin01_sql_session * session = resultSet->session;
    ParseInfoT     * parseInfo;
    int          rc;
    tsp00_Number     num;
    tsp00_NumError   res;
    tsp1_param_info * paramInfo;
    int          fetchCount = 30000;

    parseInfo = &resultSet->fetchInfo [fetchKind];
    if (parseInfo->paramCount == -1) {
        rc = parseFetch (resultSet, fetchKind);
        if (rc != cin01_db_ok) {
            return rc;
        }
    }
    i28initexecute (session, parseInfo->pid);
    switch (fetchKind) {
        case relativeFetchE:
            if (parameter < 0) {
                if (resultSet->maxRecordsPerFetch > 0) {
                    fetchCount = resultSet->maxRecordsPerFetch;
                }
                else {
                    fetchCount = 1;
                }
            }
            /*FALLTHROUGH*/
        case absoluteFetchE:
            /*
             * add parameter
             */
            i28newpart (session, sp1pk_data);
            s41plint (&num, 1, csp_fixed, 0, parameter, REFPARM (res));
            paramInfo = &parseInfo->paramInfo [0];
            i28parg (session, (const tin01_byte *) num, sizeof (num),
                paramInfo->sp1i_in_out_len, paramInfo->sp1i_bufpos,
                csp_defined_byte);
            break;
    }
    /*
     * add result count
     */
    i28newpart (session, sp1pk_resultcount);
    s41plint (&num, 1, csp_fixed, 0, fetchCount, REFPARM (res));
    i28parg (session, (const tin01_byte *) num, sizeof (num),
        sizeof (num) + 1, 1, csp_defined_byte);
    /*
     * execute
     */
    rc = requestReceive (session, false, false);
    if (rc == cin01_db_ok) {
        rc = copyFetchBuffer (resultSet);
    }
    return rc;
}

/*----------------------------------------*/

static void
getFetchOutput (
    SQLResultT * result,
    ResultSetT * resultSet,
    int          fetchKind,
    int          rc)
{
    ROUTINE_DBG_MSP00 ("getFetchOutput");
    switch (rc) {
        case cin01_db_ok:
            getOutputParameter2 (resultSet->session,
                resultSet->currentRow,
                &resultSet->fetchInfo [fetchKind],
                result);
            break;
        case cin01_db_row_not_found:
            setResult (result, SL_fromNone ());
            break;
        default:
            result->isOK = false;
            break;
    }
}

/*----------------------------------------*/

static SQLResultT
relativeRow (
    ResultSetT * resultSet,
    int count)
{
    ROUTINE_DBG_MSP00 ("relativeRow");
    int rc = cin01_db_ok;
    SQLResultT result;

    initResult (&result, 0);
    /*
     * quick fix: absolute passes to relative, but
     *    FETCH RELATIVE
     */
    if (resultSet->fetchInfo [relativeFetchE].paramCount == -1) {
        rc = parseFetch (resultSet, relativeFetchE);
        if (rc != cin01_db_ok) {
            return result;
        }
    }
    /* end quick fix */
    if (resultSet->dataBuf == NULL) {
        /*
         * no result yet requires FETCH
         */
        rc = doFetch (resultSet, relativeFetchE, count);
        if (rc == cin01_db_ok) {
            resultSet->currentRow= resultSet->dataBuf;
            resultSet->currentPos += count;
        }
    }
    else if (moveFetchBufPos (resultSet, count)) {
        resultSet->currentPos += count;
    }
    else {
        /*
         * moving outside of buffer requires FETCH
         */
        rc = doFetch (resultSet, relativeFetchE, count);
        if (rc != cin01_db_ok) {
            if (count > 0) {
                resultSet->currentRow = resultSet->dataBuf;
            }
            else {
                resultSet->currentRow = resultSet->eoRecords
                                        - resultSet->rowSize;
            }
        }
        else {
            resultSet->currentPos += count;
        }
    }
    getFetchOutput (&result, resultSet, relativeFetchE, rc);
    return result;
}

/*----------------------------------------*/

static SQLResultT
absoluteRow (
    ResultSetT * resultSet,
    int count)
{
    ROUTINE_DBG_MSP00 ("absoluteRow");
    int rc = cin01_db_ok;
    SQLResultT result;

    if (count == (resultSet->currentPos + 1)) {
        result = relativeRow(resultSet, 1);
    }
    else {
        initResult (&result, 0);
        rc = doFetch (resultSet, absoluteFetchE, count);
        resultSet->currentRow = resultSet->dataBuf;
        getFetchOutput (&result, resultSet, absoluteFetchE, rc);
        resultSet->currentPos = count;
    }
    return result;
}

/*----------------------------------------*/

#if defined (PYTHON_GLUE)
static PyObject *
absoluteRowBased0 (
    PyObject * pyself,
    int count)
{
    ROUTINE_DBG_MSP00 ("absoluteRowBased0");
    SapDB_ResultSetObjectT* self = REINTERPRET_CAST (SapDB_ResultSetObjectT*, pyself);
    ResultSetT* nself = &self->nself;
    SQLResultT result;
    PyObject * resultObj;

    result = absoluteRow (nself, count + 1);
    if (!result.isOK) {
        return NULL;
    }
    if (!sqlResultC2Py (self, result, &resultObj)) {
        return NULL;
    }
    if (SL_isNone (resultObj)) {
        PyErr_SetNone (PyExc_IndexError);
        return NULL;
    }
    return resultObj;
}

static PyObject *
__getitem___SapDB_ResultSet (
    PyObject * pyself,
    PyObject* args,
    PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("absoluteRowBased0");
    int index;
    static const char * kwlist [] = {
        LIT("index"), NULL};

    if (!PyArg_ParseTupleAndKeywords (args, keywds, LIT("i:SapDB_ResultSet.__getitem__"), (char **) kwlist, &index)) {
        return NULL;
    }
    return absoluteRowBased0 (pyself, index);
}

#endif

/*----------------------------------------*/

static SQLResultT
firstRow (
    ResultSetT * resultSet)
{
    ROUTINE_DBG_MSP00 ("first");
    int rc = cin01_db_ok;
    SQLResultT result;

    initResult (&result, 0);
    rc = doFetch (resultSet, firstFetchE, -1);
    resultSet->currentRow = resultSet->dataBuf;
    getFetchOutput (&result, resultSet, firstFetchE, rc);
    resultSet->currentPos = 1;
    return result;
}

/*----------------------------------------*/

static SQLResultT
lastRow (
    ResultSetT * resultSet)
{
    ROUTINE_DBG_MSP00 ("lastRow");
    int rc;
    SQLResultT result;

    initResult (&result, 0);
    rc = doFetch (resultSet, lastFetchE, -1);
    resultSet->currentRow = resultSet->eoRecords
                            - resultSet->rowSize;
    getFetchOutput (&result, resultSet, lastFetchE, rc);
    resultSet->currentPos = -1;
    return result;
}

/*----------------------------------------*/

static SQLResultT
currentRow (
    ResultSetT * resultSet)
{
    ROUTINE_DBG_MSP00 ("currentRow");
    SQLResultT result;
    int i;

    initResult (&result, 0);
    for (i = 0; i < fetchKindsE; ++i) {
        if (resultSet->fetchInfo [i].paramCount != -1) {
            getOutputParameter2 (resultSet->session,
                resultSet->currentRow,
                &resultSet->fetchInfo [i],
                &result);
            return result;
        }
    }
    result.isOK = false;
    return result;
}

/*----------------------------------------*/

static SQLResultT
nextRow (
    ResultSetT * resultSet)
{
    ROUTINE_DBG_MSP00 ("nextRow");
    return relativeRow (resultSet, 1);
}

/*----------------------------------------*/

static SQLResultT
previousRow (
    ResultSetT * resultSet)
{
    ROUTINE_DBG_MSP00 ("previousRow");
    return relativeRow (resultSet, -1);
}

/*----------------------------------------*/

static int
columnCount (
    ResultSetT * resultSet)
{
    return resultSet->colCount;
}

/*----------------------------------------*/

static const char * sqlTypeNamesC [] = {
    "Fixed",
    "Float",
    "Char",
    "Char EBCDIC",
    "Char Binary",
    "Rowid",
    "Long",
    "Long EBCDIC",
    "Long Binary",
    "Long Doublebyte",
    "Date",
    "Time",
    "Float",
    "Timestamp",
    "Unknown",
    "Number",
    "Nonumber",
    "Duration",
    "Doublebyte EBCDIC",
    "Long",
    "Long EBCDIC",
    "Long Binary",
    "Long Doublebyte",
    "Boolean",
    "Unicode",
    "filler1",
    "filler2",
    "filler3",
    "filler4",
    "Smallint",
    "Integer",
    "Varchar",
    "Varchar EBCDIC",
    "Varchar Binary",
    "Long Unicode",
    "Long Unicode",
    "Varchar Unicode",
    "UDT",
    "ABAP Table",
};

static int sqlTypeCodesC [] = {
    3,
    6,
    1,
    1,
    -1,
    -2,
    -1,
    -1,
    -4,
    1111,
    91,
    92,
    6,
    93,
    1111,
    6,
    1111,
    1111,
    1,
    -1,
    -1,
    -4,
    -1,
    -7,
    1,
    1111,
    1111,
    1111,
    1111,
    5,
    4,
    12,
    12,
    -3,
    -12,
    -1,
    12,
    1111,
    1111,
};

/*----------------------------------------*/

static SQLResultT
columnInfo (
    ResultSetT * resultSet,
    int colIndex)
{
    ROUTINE_DBG_MSP00 ("columnInfo");
    SQLResultT result;
    const char * namePos;
    tsp1_param_info * paramInfo;
    int i;

    initResult (&result, 0);
    if ((colIndex < 0) || (resultSet->colCount <= colIndex)) {
        result.isOK = false;
        return result;
    }
    allocResult (&result, 4);
    /*
     * get name
     */
    namePos = resultSet->colNames;
    for (i = 0; i < resultSet->colCount; ++i) {
        if (i == colIndex) {
            setResultElement (&result, 0, SL_fromString (namePos + 1, namePos [0]));
            break;
        }
        namePos += namePos [0] + 1;
    }
    /*
     * get type, len and frac
     */
    paramInfo = &resultSet->colInfo [i];
    setResultElement (&result, 1, SL_fromString (sqlTypeNamesC [paramInfo->sp1i_data_type], -1));
    setResultElement (&result, 2, SL_fromInt (paramInfo->sp1i_length));
    setResultElement (&result, 3, SL_fromInt (paramInfo->sp1i_frac));
    return result;
}

/*----------------------------------------*/

#if defined (__cplusplus)
#define ISMEMBER_M(set, element)    set.includes (element)
#else
#define ISMEMBER_M(set, element)    (set [0] & (1 << element))
#endif

/*----------------------------------------*/

static SL_ValueT *
paramMode2maybeNull (
    tsp1_param_opt_set paramMode)
{
    SL_ValueT * result;

    if (ISMEMBER_M (paramMode, sp1ot_mandatory)) {
        /* false */
        result = SL_None;
    }
    else {
        /* true */
        result = SL_fromInt (1);
    }
    return result;
}


/*----------------------------------------*/

static SL_ValueT *
ioType2string (
    tsp1_param_io_type ioType)
{
    const char * str;
    int slen;

    switch (ioType) {
        case sp1io_input:
            str = "IN";
            slen = 2;
            break;
        case sp1io_output:
            str = "OUT";
            slen = 3;
            break;
        case sp1io_inout:
            str = "IN/OUT";
            slen = 6;
            break;
        default:
            str = "UNKNOWN";
            slen = 7;
            break;
    }
    return SL_fromString (str, slen);
}

/*----------------------------------------*/

static SQLResultT
getDescription (
    int entryCount,
    const char * nameBuf,
    tsp1_param_info * paramInfos)
{
    ROUTINE_DBG_MSP00 ("getDescription");
    SQLResultT result;
    SQLResultT tuple;
    const char * namePos;
    tsp1_param_info * paramInfo;
    int i;

    initResult (&result, 0);
    allocResult (&result, entryCount);
    namePos = nameBuf;
    for (i = 0; i < entryCount; ++i) {
        initResult (&tuple, 0);
        allocResult (&tuple, 7);
        /*
         * get name
         */
        if (namePos != NULL) {
            setResultElement (&tuple, 0, SL_fromString (namePos + 1, namePos [0]));
            namePos += namePos [0] + 1;
        }
        else {
            setResultElement (&tuple, 0, SL_fromString ("", 0));
        }
        /*
         * get type, len and frac
         */
        paramInfo = &paramInfos [i];
        setResultElement (&tuple, 1, SL_fromString (sqlTypeNamesC [paramInfo->sp1i_data_type], -1));
        setResultElement (&tuple, 2, SL_fromInt (sqlTypeCodesC [paramInfo->sp1i_data_type]));
        setResultElement (&tuple, 3, SL_fromInt (paramInfo->sp1i_length));
        setResultElement (&tuple, 4, SL_fromInt (paramInfo->sp1i_frac));
        setResultElement (&tuple, 5, paramMode2maybeNull (paramInfo->sp1i_mode));
        setResultElement (&tuple, 6, ioType2string (paramInfo->sp1i_io_type));
        /*
         * set result from tuple
         */
        setResultElement (&result, i, sqlResultC2SL_Value(&tuple));
    }
    return result;
}

/*----------------------------------------*/

static SQLResultT
getCursorDescription (
    ResultSetT * resultSet)
{
    ROUTINE_DBG_MSP00 ("getCursorDescription");
    return getDescription (resultSet->colCount,
        resultSet->colNames, resultSet->colInfo);
}

/*----------------------------------------*/

static char *
cursorName (
    ResultSetT * nself)
{
    ROUTINE_DBG_MSP00 ("cursorName");
    return nself->cursorName;
}

/*----------------------------------------*/

static void
closeResultSet (
    ResultSetT * nself)
{
    ROUTINE_DBG_MSP00 ("closeResultSet");
    int i;

    for (i = 0; i < fetchKindsE; ++i) {
        clearParseInfo (nself->session, &nself->fetchInfo [i]);
    }
    if (strlen (nself->cursorName) > 0) {
        if (nself->session->is_connected) {
            i28adbsf (nself->session, "CLOSE \"%s\"", STATIC_CAST (char*, nself->cursorName));
        }
    }
    decreaseRef (nself->session->more_data);
    if (nself->colNames != NULL) {
        const void * tmpVoid = nself->colNames;
        free (CONST_CAST (void *, tmpVoid));
        nself->colNames = NULL;
    }
    if (nself->colInfo != NULL) {
        free (nself->colInfo);
        nself->colInfo = NULL;
    }
    nself->colCount = 0;
}

/*----------------------------------------*/

#if defined (PYTHON_GLUE)
static void
destructor_SapDB_ResultSet (
    PyObject * pyself)
{
    ROUTINE_DBG_MSP00 ("destructor_SapDB_ResultSet");
    SapDB_ResultSetObjectT* self = REINTERPRET_CAST (SapDB_ResultSetObjectT*, pyself);

    closeResultSet (&self->nself);
    free (pyself);
}
#endif

/*----------------------------------------*/
/* endsnippet SapDB_ResultSet Generic_precode */
/*----------------------------------------*/

externC XS (SapDB_ResultSet_next)
{
    ROUTINE_DBG_MSP00 ("SapDB_ResultSet_next");
/* no code for key Perl_/Generic_code */
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_ResultSetObjectT* self;
    ResultSetT* nself;
    SQLResultT result;
    
    cv = cv;
    if (items != 1) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_ResultSetObjectT*) getGluePointer (ST(0), SapDB_ResultSetClassID, croakBuf);
    if ((self == NULL) || !isSapDB_ResultSetObject (self)) {
        croak (croakBuf);
    }
    nself = &self->nself;
    if (!isSessionConnected (self->nself.session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Perl_/Generic_nativeCall */
    result = nextRow (nself);
    if (!sqlResultOK (&result) && sqlErrOccured (nself->session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
    ST (argvi) = sv_newmortal ();
    if (result.isTuple) {
        sv_setsv (ST (argvi++), newRV (REINTERPRET_CAST (SV*, result.value.resultList)));
    }
    else {
        sv_setsv (ST (argvi++), result.value.resultObj);
    };
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        croak (croakBuf);
    }
    XSRETURN (argvi);
}

/*----------------------------------------*/

externC XS (SapDB_ResultSet_previous)
{
    ROUTINE_DBG_MSP00 ("SapDB_ResultSet_previous");
/* no code for key Perl_/Generic_code */
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_ResultSetObjectT* self;
    ResultSetT* nself;
    SQLResultT result;
    
    cv = cv;
    if (items != 1) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_ResultSetObjectT*) getGluePointer (ST(0), SapDB_ResultSetClassID, croakBuf);
    if ((self == NULL) || !isSapDB_ResultSetObject (self)) {
        croak (croakBuf);
    }
    nself = &self->nself;
    if (!isSessionConnected (self->nself.session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Perl_/Generic_nativeCall */
    result = previousRow (nself);
    if (!sqlResultOK (&result) && sqlErrOccured (nself->session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
    ST (argvi) = sv_newmortal ();
    if (result.isTuple) {
        sv_setsv (ST (argvi++), newRV (REINTERPRET_CAST (SV*, result.value.resultList)));
    }
    else {
        sv_setsv (ST (argvi++), result.value.resultObj);
    };
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        croak (croakBuf);
    }
    XSRETURN (argvi);
}

/*----------------------------------------*/

externC XS (SapDB_ResultSet_relative)
{
    ROUTINE_DBG_MSP00 ("SapDB_ResultSet_relative");
/* no code for key Perl_/Generic_code */
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_ResultSetObjectT* self;
    ResultSetT* nself;
    int count;
    SQLResultT result;
    
    cv = cv;
    if (items != 2) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_ResultSetObjectT*) getGluePointer (ST(0), SapDB_ResultSetClassID, croakBuf);
    if ((self == NULL) || !isSapDB_ResultSetObject (self)) {
        croak (croakBuf);
    }
    nself = &self->nself;
    switch (items) {
        case 2: count = (int) SvIV (ST(1));
    }
    if (!ok) {
        ok = 0;
        goto cleanup_label;
    }
    if (!isSessionConnected (self->nself.session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Perl_/Generic_nativeCall */
    result = relativeRow (nself, count);
    if (!sqlResultOK (&result) && sqlErrOccured (nself->session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
    ST (argvi) = sv_newmortal ();
    if (result.isTuple) {
        sv_setsv (ST (argvi++), newRV (REINTERPRET_CAST (SV*, result.value.resultList)));
    }
    else {
        sv_setsv (ST (argvi++), result.value.resultObj);
    };
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        croak (croakBuf);
    }
    XSRETURN (argvi);
}

/*----------------------------------------*/

externC XS (SapDB_ResultSet_absolute)
{
    ROUTINE_DBG_MSP00 ("SapDB_ResultSet_absolute");
/* no code for key Perl_/Generic_code */
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_ResultSetObjectT* self;
    ResultSetT* nself;
    int i;
    SQLResultT result;
    
    cv = cv;
    if (items != 2) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_ResultSetObjectT*) getGluePointer (ST(0), SapDB_ResultSetClassID, croakBuf);
    if ((self == NULL) || !isSapDB_ResultSetObject (self)) {
        croak (croakBuf);
    }
    nself = &self->nself;
    switch (items) {
        case 2: i = (int) SvIV (ST(1));
    }
    if (!ok) {
        ok = 0;
        goto cleanup_label;
    }
    if (!isSessionConnected (self->nself.session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Perl_/Generic_nativeCall */
    result = absoluteRow (nself, i);
    if (!sqlResultOK (&result) && sqlErrOccured (nself->session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
    ST (argvi) = sv_newmortal ();
    if (result.isTuple) {
        sv_setsv (ST (argvi++), newRV (REINTERPRET_CAST (SV*, result.value.resultList)));
    }
    else {
        sv_setsv (ST (argvi++), result.value.resultObj);
    };
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        croak (croakBuf);
    }
    XSRETURN (argvi);
}

/*----------------------------------------*/

externC XS (SapDB_ResultSet_first)
{
    ROUTINE_DBG_MSP00 ("SapDB_ResultSet_first");
/* no code for key Perl_/Generic_code */
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_ResultSetObjectT* self;
    ResultSetT* nself;
    SQLResultT result;
    
    cv = cv;
    if (items != 1) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_ResultSetObjectT*) getGluePointer (ST(0), SapDB_ResultSetClassID, croakBuf);
    if ((self == NULL) || !isSapDB_ResultSetObject (self)) {
        croak (croakBuf);
    }
    nself = &self->nself;
    if (!isSessionConnected (self->nself.session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Perl_/Generic_nativeCall */
    result = firstRow (nself);
    if (!sqlResultOK (&result) && sqlErrOccured (nself->session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
    ST (argvi) = sv_newmortal ();
    if (result.isTuple) {
        sv_setsv (ST (argvi++), newRV (REINTERPRET_CAST (SV*, result.value.resultList)));
    }
    else {
        sv_setsv (ST (argvi++), result.value.resultObj);
    };
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        croak (croakBuf);
    }
    XSRETURN (argvi);
}

/*----------------------------------------*/

externC XS (SapDB_ResultSet_last)
{
    ROUTINE_DBG_MSP00 ("SapDB_ResultSet_last");
/* no code for key Perl_/Generic_code */
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_ResultSetObjectT* self;
    ResultSetT* nself;
    SQLResultT result;
    
    cv = cv;
    if (items != 1) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_ResultSetObjectT*) getGluePointer (ST(0), SapDB_ResultSetClassID, croakBuf);
    if ((self == NULL) || !isSapDB_ResultSetObject (self)) {
        croak (croakBuf);
    }
    nself = &self->nself;
    if (!isSessionConnected (self->nself.session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Perl_/Generic_nativeCall */
    result = lastRow (nself);
    if (!sqlResultOK (&result) && sqlErrOccured (nself->session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
    ST (argvi) = sv_newmortal ();
    if (result.isTuple) {
        sv_setsv (ST (argvi++), newRV (REINTERPRET_CAST (SV*, result.value.resultList)));
    }
    else {
        sv_setsv (ST (argvi++), result.value.resultObj);
    };
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        croak (croakBuf);
    }
    XSRETURN (argvi);
}

/*----------------------------------------*/

externC XS (SapDB_ResultSet_current)
{
    ROUTINE_DBG_MSP00 ("SapDB_ResultSet_current");
/* no code for key Perl_/Generic_code */
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_ResultSetObjectT* self;
    ResultSetT* nself;
    SQLResultT result;
    
    cv = cv;
    if (items != 1) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_ResultSetObjectT*) getGluePointer (ST(0), SapDB_ResultSetClassID, croakBuf);
    if ((self == NULL) || !isSapDB_ResultSetObject (self)) {
        croak (croakBuf);
    }
    nself = &self->nself;
    if (!isSessionConnected (self->nself.session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Perl_/Generic_nativeCall */
    result = currentRow (nself);
    if (!sqlResultOK (&result) && sqlErrOccured (nself->session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
    ST (argvi) = sv_newmortal ();
    if (result.isTuple) {
        sv_setsv (ST (argvi++), newRV (REINTERPRET_CAST (SV*, result.value.resultList)));
    }
    else {
        sv_setsv (ST (argvi++), result.value.resultObj);
    };
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        croak (croakBuf);
    }
    XSRETURN (argvi);
}

/*----------------------------------------*/

externC XS (SapDB_ResultSet_cursorName)
{
    ROUTINE_DBG_MSP00 ("SapDB_ResultSet_cursorName");
/* no code for key Perl_/Generic_code */
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_ResultSetObjectT* self;
    ResultSetT* nself;
    char * result;
    
    cv = cv;
    if (items != 1) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_ResultSetObjectT*) getGluePointer (ST(0), SapDB_ResultSetClassID, croakBuf);
    if ((self == NULL) || !isSapDB_ResultSetObject (self)) {
        croak (croakBuf);
    }
    nself = &self->nself;
    if (!isSessionConnected (self->nself.session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Perl_/Generic_nativeCall */
    result = cursorName (nself);
    /* outargs */
    ST (argvi) = sv_newmortal ();
    sv_setpv (ST (argvi++), result);
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        croak (croakBuf);
    }
    XSRETURN (argvi);
}

/*----------------------------------------*/

externC XS (SapDB_ResultSet_columnCount)
{
    ROUTINE_DBG_MSP00 ("SapDB_ResultSet_columnCount");
/* no code for key Perl_/Generic_code */
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_ResultSetObjectT* self;
    ResultSetT* nself;
    int result;
    
    cv = cv;
    if (items != 1) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_ResultSetObjectT*) getGluePointer (ST(0), SapDB_ResultSetClassID, croakBuf);
    if ((self == NULL) || !isSapDB_ResultSetObject (self)) {
        croak (croakBuf);
    }
    nself = &self->nself;
    if (!isSessionConnected (self->nself.session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Perl_/Generic_nativeCall */
    result = columnCount (nself);
    /* outargs */
    ST (argvi) = sv_newmortal ();
    sv_setiv (ST (argvi++), (IV) result);
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        croak (croakBuf);
    }
    XSRETURN (argvi);
}

/*----------------------------------------*/

externC XS (SapDB_ResultSet_columnInfo)
{
    ROUTINE_DBG_MSP00 ("SapDB_ResultSet_columnInfo");
/* no code for key Perl_/Generic_code */
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_ResultSetObjectT* self;
    ResultSetT* nself;
    int i;
    SQLResultT result;
    
    cv = cv;
    if (items != 2) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_ResultSetObjectT*) getGluePointer (ST(0), SapDB_ResultSetClassID, croakBuf);
    if ((self == NULL) || !isSapDB_ResultSetObject (self)) {
        croak (croakBuf);
    }
    nself = &self->nself;
    switch (items) {
        case 2: i = (int) SvIV (ST(1));
    }
    if (!ok) {
        ok = 0;
        goto cleanup_label;
    }
    if (!isSessionConnected (self->nself.session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Perl_/Generic_nativeCall */
    result = columnInfo (nself, i);
    if (!sqlResultOK (&result) && sqlErrOccured (nself->session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
    ST (argvi) = sv_newmortal ();
    if (result.isTuple) {
        sv_setsv (ST (argvi++), newRV (REINTERPRET_CAST (SV*, result.value.resultList)));
    }
    else {
        sv_setsv (ST (argvi++), result.value.resultObj);
    };
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        croak (croakBuf);
    }
    XSRETURN (argvi);
}

/*----------------------------------------*/

externC XS (SapDB_ResultSet_getDescription)
{
    ROUTINE_DBG_MSP00 ("SapDB_ResultSet_getDescription");
/* no code for key Perl_/Generic_code */
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_ResultSetObjectT* self;
    ResultSetT* nself;
    SQLResultT result;
    
    cv = cv;
    if (items != 1) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_ResultSetObjectT*) getGluePointer (ST(0), SapDB_ResultSetClassID, croakBuf);
    if ((self == NULL) || !isSapDB_ResultSetObject (self)) {
        croak (croakBuf);
    }
    nself = &self->nself;
    if (!isSessionConnected (self->nself.session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Perl_/Generic_nativeCall */
    result = getCursorDescription (nself);
    if (!sqlResultOK (&result) && sqlErrOccured (nself->session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
    ST (argvi) = sv_newmortal ();
    if (result.isTuple) {
        sv_setsv (ST (argvi++), newRV (REINTERPRET_CAST (SV*, result.value.resultList)));
    }
    else {
        sv_setsv (ST (argvi++), result.value.resultObj);
    };
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        croak (croakBuf);
    }
    XSRETURN (argvi);
}


/*----------------------------------------*/

externC XS (SapDB_ResultSet_destructor)
{
    ROUTINE_DBG_MSP00 ("SapDB_ResultSet_destructor");
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_ResultSetObjectT* self;
    ResultSetT* nself;

    cv = cv;
    if (items != 1) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_ResultSetObjectT*) getGluePointer (ST(0), SapDB_ResultSetClassID, croakBuf);
    if ((self == NULL) || !isSapDB_ResultSetObject (self)) {
        croak (croakBuf);
    }
    nself = &self->nself;
    closeResultSet (nself);
    freeSapDB_ResultSet (self);
    XSRETURN (0);
}

static const char * SapDB_PreparedClassID = "SapDB_Prepared";

typedef struct SapDB_PreparedObjectT {
    const char * classID;
    PreparedT nself;
/* no code for key SapDB_Prepared Perl_/Generic_ in ['Generic_precode', '::select Generic_nativeCall vars']*/
} SapDB_PreparedObjectT;

#define isSapDB_PreparedObject(op) ((op)->classID == SapDB_PreparedClassID)

static const char * notClassSapDB_Prepared_C = "Object not of class SapDB_Prepared";

static SapDB_PreparedObjectT *
newSapDB_Prepared ()
{
    SapDB_PreparedObjectT * result;

    Newz (0, result, 1, SapDB_PreparedObjectT);
    return result;
}

static void
freeSapDB_Prepared (
    SapDB_PreparedObjectT * value)
{
    Safefree (value);
}
/* snippet SapDB_Prepared Generic_precode */

typedef struct LongInputT {
    bool readStarted;
    bool hasData;
    unsigned short colIndex;
    tsp00_LongDescriptor descriptor;
    InputStreamT * stream;
    union {
        StringInputStreamT stringStream;
        CallableInputStreamT callableStream;
    } obj;
} LongInputT;

/*----------------------------------------*/

static void
putLongvals (
    tin01_sql_session * session,
    ParseInfoT * parseInfo,
    LongInputT * longs,
    bool forPutval)
{
    int    descriptorLen;
    int    descriptorPos;
    byte_t * dataPtr;
#if defined (PUTVAL_REQUIRES_TWO_BYTE_COUNT)
    byte_t * startPtr;
#endif
    int    bufPos;
    int    freeLen;
    int    bytesRead;
    tsp00_Int2 transferCount = 0;
    LongInputT * currentLong;
    InputStreamT * instream;
    tsp1_param_info * paramInfo;
    int i;

    if (forPutval) {
        descriptorLen = sizeof (tsp00_LongDescriptor) + 1;
    }
    else {
        descriptorLen = 0;
    }
    i28getFreePartInfo (session, REINTERPRET_CAST (void**, &dataPtr), &bufPos, &freeLen);
#if defined (PUTVAL_REQUIRES_TWO_BYTE_COUNT)
    startPtr = dataPtr;
    if (forPutval) {
        dataPtr += 2;
        bufPos += 2;
    }
#endif
    for (i = 0; (freeLen > descriptorLen) && (i < parseInfo->inLongCount); ++i) {
        currentLong = &longs [i];
        instream = currentLong->stream;
        paramInfo = &parseInfo->paramInfo [currentLong->colIndex];
        if (currentLong->hasData) {
            bytesRead = (*instream)->read (instream,
                            dataPtr + descriptorLen, freeLen - descriptorLen);
            if (bytesRead > 0) {
                if (bytesRead < (freeLen - descriptorLen)) {
                    if (currentLong->readStarted) {
                        BECOMES (DESCFIELD (currentLong->descriptor, ld_valmode), vm_lastdata);
                    }
                    else {
                        BECOMES (DESCFIELD (currentLong->descriptor, ld_valmode), vm_alldata);
                    }
                    currentLong->hasData = false;
                }
                else {
                    BECOMES (DESCFIELD (currentLong->descriptor, ld_valmode), vm_datapart);
                    currentLong->readStarted = true;
                }
                DESCFIELD (currentLong->descriptor, ld_valpos) = bufPos + descriptorLen;
                DESCFIELD (currentLong->descriptor, ld_vallen) = bytesRead;
                if (forPutval) {
                    descriptorPos = bufPos;
                }
                else {
                    descriptorPos = paramInfo->sp1i_bufpos;
                }
                i28parg (session, REINTERPRET_CAST (const char*, &currentLong->descriptor),
                    sizeof (currentLong->descriptor),
                    sizeof (currentLong->descriptor) + 1,
                    descriptorPos, '\0');
                ++transferCount;
                i28extendpart (session, bytesRead);
                dataPtr += bytesRead + descriptorLen;
                bufPos += bytesRead + descriptorLen;
                freeLen -= bytesRead + descriptorLen;
            }
            else {
                currentLong->hasData = false;
            }
        }
    }
    if (forPutval && !(longs [parseInfo->inLongCount - 1].hasData)) {
        /* at end of all longs */
        currentLong = &longs [parseInfo->inLongCount - 1];
        BECOMES (DESCFIELD (currentLong->descriptor, ld_valmode), vm_last_putval);
        i28parg (session, REINTERPRET_CAST (const char*, &currentLong->descriptor),
            sizeof (currentLong->descriptor),
            sizeof (currentLong->descriptor) + 1,
            bufPos, '\0');
        /* ++transferCount; */
    }
#if defined (PUTVAL_REQUIRES_TWO_BYTE_COUNT)
    if (forPutval) {
        memcpy (startPtr, &transferCount, sizeof (transferCount));
    }
#endif
    session->part->variant.C_2.sp1p_arg_count_F += transferCount;
}

/*----------------------------------------*/

static int
doPutval (
    tin01_sql_session * session,
    ParseInfoT * parseInfo,
    LongInputT * longs)
{
    ROUTINE_DBG_MSP00 ("doPutval");
    tsp00_LongDescriptor descriptor;
    int descriptorsReturned;
    byte_t * bufPos;
    int rc;
    int i;

    /*
     *  get descriptors
     */
    i28findpart (session, sp1pk_longdata);
    descriptorsReturned = session->part->variant.C_2.sp1p_arg_count_F;
#if defined (__cplusplus)
    bufPos = (byte_t *) session->part->sp1p_buf ().asCharp ();
#else
    bufPos = session->part->variant.C_1.sp1p_buf_F;
#endif
    for (i = 0; i < descriptorsReturned; ++i) {
        memcpy (&descriptor, bufPos + 1, sizeof (descriptor));
        memcpy (&longs [DESCFIELD (descriptor, ld_valind)].descriptor,
            &descriptor, sizeof (descriptor));
    }
    /*
     * send putval
     */
    i28initspecial (session, sp1m_putval);
    i28newpart (session, sp1pk_longdata);
    putLongvals (session, parseInfo, longs, true);
    rc = requestReceive (session, false, false);
    return rc;
}

/*----------------------------------------*/

static void
    setInputParameter (
    tin01_sql_session * session,
    ParseInfoT * parseInfo,
    SequenceT    args,
    LongInputT * longs)
{
    ROUTINE_DBG_MSP00 ("setInputParameter");
    tsp1_param_info * paramInfo = parseInfo->paramInfo;
    SL_ValueT  * value;
    int argCount;
    int inputIndex = -1;
    char fillChar = ' ';
    const char * addr;
    int dataLen;
    int i;
    tsp00_Number number;
    char boolBuf [1];
    int longIndex = 0;
    LongInputT * currentLong;

    if (parseInfo->paramCount == 0) {
        return;
    }
    argCount = sequenceLength (&args);
    i28newpart (session, sp1pk_data);
    for (i = 0; i < parseInfo->paramCount; ++i, ++paramInfo) {
        if (paramInfo->sp1i_io_type != sp1io_output) {
            ++inputIndex;
            if (inputIndex < argCount) {
                value = getSequenceElement (&args, inputIndex);
            }
            else {
                value = SL_None;
            }
            if (value == NULL) {
                /* set exception and exit */
            }
            fillChar = ' ';
            switch (paramInfo->sp1i_data_type) {
                case dfixed:
                case dfloat:
                case dvfloat:
                case dsmallint:
                case dinteger:
                    if (SL_isNone (value)) {
                        i28pnull (session,
                            paramInfo->sp1i_in_out_len,
                            paramInfo->sp1i_bufpos);
                    }
                    else {
                        SL_getNumber (value, paramInfo, &number);
                        i28parg (session, (tin01_byte *) number,
                            sizeof (number),
                            paramInfo->sp1i_in_out_len,
                            paramInfo->sp1i_bufpos, '\0');
                    }
                    break;
                case dchb:
                case dvarcharb:
                    fillChar = '\0';
                    /*FALLTHROUGH*/
                case dcha:
                case dche:
                case dvarchara:
                case dvarchare:
                case ddate:
                case dtime:
                case dtimestamp:
                    if (SL_isNone (value)) {
                        i28pnull (session,
                            paramInfo->sp1i_in_out_len,
                            paramInfo->sp1i_bufpos);
                    }
                    else {
                        SL_getRawString (value, &addr, &dataLen);
                        i28parg (session, addr, dataLen,
                            paramInfo->sp1i_in_out_len,
                            paramInfo->sp1i_bufpos, fillChar);
                    }
                    break;
                case dboolean:
                    if (SL_isTrue (value)) {
                        boolBuf [0] = '\1';
                    }
                    else {
                        boolBuf [0] = '\0';
                    }
                    i28parg (session, boolBuf, sizeof (boolBuf),
                        paramInfo->sp1i_in_out_len,
                        paramInfo->sp1i_bufpos, '\0');
                    break;
                case dstra:
                case dstre:
                case dstrb:
                case dlonga:
                case dlonge:
                case dlongb:
                    /*
                     * put descriptor
                     */
                    currentLong = &longs [longIndex];
                    currentLong->colIndex = i;
                    DESCFIELD (currentLong->descriptor, ld_valind) = longIndex;
                    BECOMES (DESCFIELD (currentLong->descriptor, ld_valmode), vm_nodata);
                    i28parg (session, REINTERPRET_CAST (const char*, &currentLong->descriptor),
                        sizeof (currentLong->descriptor),
                        paramInfo->sp1i_in_out_len,
                        paramInfo->sp1i_bufpos, '\0');
                    /*
                     * convert value to stream
                     */
                    if (SL_isString (value)) {
                        const char * data;
                        int len;
                        SL_getRawString (value, &data, &len);
                        currentLong->stream = initStringInputStream (
                            &currentLong->obj.stringStream, data, len);
                        currentLong->hasData = true;
                    }
                    else if (SL_isCallable (value)) {
                        currentLong->stream = initCallableInputStream (
                            &currentLong->obj.callableStream, value);
                        currentLong->hasData = true;
                    }
                    else {
                        currentLong->hasData = false;
                    }
                    /*
                     *
                     */
                    ++longIndex;
                    break;
            }
        }
    }
    if (longs != NULL) {
        putLongvals (session, parseInfo, longs, false);
    }
}

/*----------------------------------------*/

static bool
fillParamInfos (
    tin01_sql_session * session,
    ParseInfoT * parseInfo,
    bool initial)
{
    ROUTINE_DBG_MSP00 ("fillParamInfos");
    bool ok = true;
    int infoSize;
    int i;

    i28gparseid (session, parseInfo->pid);
    parseInfo->paramCount = i28paramcount (session);
    if (initial) {
        parseInfo->paramInfo = NULL;
    }
    if (parseInfo->paramCount > 0) {
        infoSize = parseInfo->paramCount * sizeof (tsp1_param_info);
        if (initial) {
            parseInfo->paramInfo = REINTERPRET_CAST (tsp1_param_info*, malloc (infoSize));
        }
        if (parseInfo->paramInfo == NULL) {
            ok = false;
        }
        else {
            memcpy (parseInfo->paramInfo, i28gparaminfo (session, 0), infoSize);
        }
    }
    parseInfo->outputParamCount = 0;
    for (i = 0; i < parseInfo->paramCount; ++i) {
        tsp1_param_info * paramInfo = &parseInfo->paramInfo [i];
        if (paramInfo->sp1i_io_type != sp1io_input) {
            ++parseInfo->outputParamCount;
        }
        switch (paramInfo->sp1i_data_type) {
            case dstra:
            case dstre:
            case dstrb:
            case dlonga:
            case dlonge:
            case dlongb:
                switch (paramInfo->sp1i_io_type) {
                    case sp1io_input:
                        ++parseInfo->inLongCount;
                        break;
                    case sp1io_output:
                        ++parseInfo->outLongCount;
                        break;
                    case sp1io_inout:
                        ++parseInfo->inLongCount;
                        ++parseInfo->outLongCount;
                        break;
                }
                break;
        }
    }
    return ok;
}

/*----------------------------------------*/

static int
prepare (
    tin01_sql_session * session,
    PreparedT * prepared,
    const char * cmd,
    bool initial)
{
    ROUTINE_DBG_MSP00 ("prepare");
    int rc = 0;
    int cmdLen = strlen (cmd);
    int retryCount = 3;

    prepared->session = session;
    increaseRef (session->more_data);
    prepared->parseInfo.paramCount = -1;
    do {
        i28initparse (prepared->session);
        i28pcmd (prepared->session, cmd, cmdLen);
        if (getAutocommit (session)) {
            session->segment->variant.C_3.sp1c_commit_immediately_F = true;
        }
        rc = requestReceive (prepared->session, true, true);
        --retryCount;
    } while ((rc == cin01_db_parse_again) && (retryCount > 0));
    if (rc == cin01_db_ok) {
        fillParamInfos (session, &prepared->parseInfo, initial);
        i28gparseid (prepared->session, prepared->parseInfo.pid);
        if (initial) {
            prepared->parseInfo.sqlString = (char *) malloc (cmdLen + 1);
            if (prepared->parseInfo.sqlString != NULL) {
                memcpy (prepared->parseInfo.sqlString, cmd, cmdLen + 1);
            }
        }
    }
    return rc;
}

/*----------------------------------------*/

static int
prepareFetch (
    tin01_sql_session * session,
    PreparedT * prepared)
{
    ROUTINE_DBG_MSP00 ("prepareFetch");
    char fetchCmd [100];
    int nameLen;
    int pos;

    strcpy (fetchCmd, "FETCH  ");
    pos = strlen ( fetchCmd);
    nameLen = sizeof (fetchCmd) - pos;
    if (i28gresulttablename (session, &nameLen, fetchCmd + pos)) {
        fetchCmd [pos - 1] = '"';
        fetchCmd [pos + nameLen] = '"';
        fetchCmd [pos + nameLen + 1] = '\0';
    }
    return prepare (session, prepared, fetchCmd, true);
}

/*----------------------------------------*/

static bool
packet2result (
    tin01_sql_session * session,
    int rc,
    ParseInfoT * parseInfo,
    SQLResultT * result)
{
    ROUTINE_DBG_MSP00 ("packet2result");

    if ((rc != 0) && (rc != 100)) {
        result->isOK = false;
        return false;
    }
    if (rc == 100) {
        session->lasterr.sqlresult.returnCode = 0;
    }
    if (i28findpart (session, sp1pk_data)) {
        ParseInfoT localInfo;
        bool requireDeallocation = false;
        if (parseInfo == NULL) {
            parseInfo = &localInfo;
            if (fillParamInfos (session, parseInfo, true)) {
                requireDeallocation = true;
            }
            else {
                parseInfo->paramInfo = NULL;
            }

        }
        i28findpart (session, sp1pk_data);
        if (parseInfo->paramInfo != NULL) {
            getOutputParameter (session, parseInfo, result);
            if (requireDeallocation) {
                free (localInfo.paramInfo);
            }
        }
        else {
            int partLen = i28partlen (session);
            char * partPtr = i28getpart (session);
            SL_ValueT * resultValue = SL_fromString (partPtr, partLen);
            setResult (result, resultValue);
        }
    }
    else if (i28findpart (session, sp1pk_resulttablename)) {
        setResult (result, createResultSetFromReply (session));
    }
    else if (i28findpart (session, sp1pk_resultcount)) {
        setResult (result, SL_fromInt (i28resultcount (session)));
    }
    else {
        setResult (result, SL_fromInt (0));
    }
    return true;
}

/*----------------------------------------*/

static SQLResultT
executePrepared (
    PreparedT * prepared,
    SequenceT   args)
{
    ROUTINE_DBG_MSP00 ("executePrepared");
    SQLResultT   result;
    LongInputT * longs = NULL;
    int          rc = cin01_db_ok;
    ParseInfoT * parseInfo = &prepared->parseInfo;
    int          retryCount = 3;

    initResult (&result, 0);
    if (parseInfo->inLongCount > 0) {
        longs = REINTERPRET_CAST (LongInputT*, calloc (parseInfo->inLongCount,
                    sizeof (LongInputT)));
    }
    do {
        if (rc == cin01_db_parse_again) {
            --retryCount;
            rc = prepare (prepared->session, prepared,
                prepared->parseInfo.sqlString, false);
            if (rc != cin01_db_ok) {
                break;
            }
        }
        i28initexecute (prepared->session, parseInfo->pid);
        prepared->session->segment->variant.C_3.sp1c_with_info_F = true;
        prepared->session->segment->variant.C_3.sp1c_commit_immediately_F
            = getAutocommit (prepared->session);
        addCursorName (prepared->session);
        setInputParameter (prepared->session, parseInfo, args, longs);
        rc = requestReceive (prepared->session, true, true);
    } while ((rc == cin01_db_parse_again) && (retryCount > 0));
    packet2result (prepared->session, rc, parseInfo, &result);
    if (longs != NULL) {
        while (longs [parseInfo->inLongCount - 1].hasData
               && (rc == cin01_db_ok)) {
            rc = doPutval (prepared->session, parseInfo, longs);
        }
    }
    if (longs != NULL) {
        free (longs);
    }
    return result;
}

/*----------------------------------------*/

static SQLResultT
nextPrepared (
    PreparedT * prepared)
{
    ROUTINE_DBG_MSP00 ("nextPrepared");
    SequenceT  args;
    args = EmptySequenceC;
    return executePrepared (prepared, args);
}

/*----------------------------------------*/

static SQLResultT
getPreparedDescription (
    PreparedT * prepared)
{
    ROUTINE_DBG_MSP00 ("getPreparedDescription");
    return getDescription (prepared->parseInfo.paramCount, NULL,
        prepared->parseInfo.paramInfo);
}

/*----------------------------------------*/

static void
closePrepared (
    PreparedT * nself)
{
    ROUTINE_DBG_MSP00 ("closePrepared");
    clearParseInfo (nself->session, &nself->parseInfo);
    decreaseRef (nself->session->more_data);
}

/*----------------------------------------*/

#if defined (PYTHON_GLUE)
static void
destructor_SapDB_Prepared (
    PyObject * pyself)
{
    ROUTINE_DBG_MSP00 ("destructor_SapDB_Prepared");
    SapDB_PreparedObjectT* self = REINTERPRET_CAST (SapDB_PreparedObjectT*, pyself);

    closePrepared (&self->nself);
    free (pyself);
}

/*----------------------------------------*/
#endif

/* endsnippet SapDB_Prepared Generic_precode */
/*----------------------------------------*/

externC XS (SapDB_Prepared_execute)
{
    ROUTINE_DBG_MSP00 ("SapDB_Prepared_execute");
/* no code for key Perl_/Generic_code */
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_PreparedObjectT* self;
    PreparedT* nself;
    SequenceT sqlParms =  EmptySequenceC;
    SQLResultT result;
    
    cv = cv;
    if ((items < 1) || (items > 2)) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_PreparedObjectT*) getGluePointer (ST(0), SapDB_PreparedClassID, croakBuf);
    if ((self == NULL) || !isSapDB_PreparedObject (self)) {
        croak (croakBuf);
    }
    nself = &self->nself;
    switch (items) {
        case 2: ok = ok && sequencePerl2C (ST(1), &sqlParms, croakBuf);;
    }
    if (!ok) {
        ok = 0;
        goto cleanup_label;
    }
    if (!isSessionConnected (self->nself.session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Perl_/Generic_nativeCall */
    result = executePrepared (nself, sqlParms);
    if (!sqlResultOK (&result) && sqlErrOccured (nself->session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
    ST (argvi) = sv_newmortal ();
    if (result.isTuple) {
        sv_setsv (ST (argvi++), newRV (REINTERPRET_CAST (SV*, result.value.resultList)));
    }
    else {
        sv_setsv (ST (argvi++), result.value.resultObj);
    };
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        croak (croakBuf);
    }
    XSRETURN (argvi);
}

/*----------------------------------------*/

externC XS (SapDB_Prepared_getDescription)
{
    ROUTINE_DBG_MSP00 ("SapDB_Prepared_getDescription");
/* no code for key Perl_/Generic_code */
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_PreparedObjectT* self;
    PreparedT* nself;
    SQLResultT result;
    
    cv = cv;
    if (items != 1) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_PreparedObjectT*) getGluePointer (ST(0), SapDB_PreparedClassID, croakBuf);
    if ((self == NULL) || !isSapDB_PreparedObject (self)) {
        croak (croakBuf);
    }
    nself = &self->nself;
    if (!isSessionConnected (self->nself.session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Perl_/Generic_nativeCall */
    result = getPreparedDescription (nself);
    if (!sqlResultOK (&result) && sqlErrOccured (nself->session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
    ST (argvi) = sv_newmortal ();
    if (result.isTuple) {
        sv_setsv (ST (argvi++), newRV (REINTERPRET_CAST (SV*, result.value.resultList)));
    }
    else {
        sv_setsv (ST (argvi++), result.value.resultObj);
    };
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        croak (croakBuf);
    }
    XSRETURN (argvi);
}


/*----------------------------------------*/

externC XS (SapDB_Prepared_destructor)
{
    ROUTINE_DBG_MSP00 ("SapDB_Prepared_destructor");
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_PreparedObjectT* self;
    PreparedT* nself;

    cv = cv;
    if (items != 1) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_PreparedObjectT*) getGluePointer (ST(0), SapDB_PreparedClassID, croakBuf);
    if ((self == NULL) || !isSapDB_PreparedObject (self)) {
        croak (croakBuf);
    }
    nself = &self->nself;
    closePrepared (nself);
    freeSapDB_Prepared (self);
    XSRETURN (0);
}

static const char * SapDB_SessionClassID = "SapDB_Session";

typedef struct SapDB_SessionObjectT {
    const char * classID;
    tin01_sql_session *nself;
    /* snippet SapDB_Session Generic_instance vars */
    tin01_sql_session session;
    bool              autocommit;
    int               cursorCount;
    bool              implicitCursorNames;
    bool              lastWasEOT;
    /* endsnippet SapDB_Session Generic_instance vars */
    /* snippet SapDB_Session Perl_instance vars */
    int refCount;
    /* endsnippet SapDB_Session Perl_instance vars */
} SapDB_SessionObjectT;

#define isSapDB_SessionObject(op) ((op)->classID == SapDB_SessionClassID)

static const char * notClassSapDB_Session_C = "Object not of class SapDB_Session";

static SapDB_SessionObjectT *
newSapDB_Session ()
{
    SapDB_SessionObjectT * result;

    Newz (0, result, 1, SapDB_SessionObjectT);
    return result;
}

static void
freeSapDB_Session (
    SapDB_SessionObjectT * value)
{
    Safefree (value);
}
/* snippet SapDB_Session Generic_precode */

static void
setAutocommit (
    tin01_sql_session * session,
    bool autocommit)
{
    ((SapDB_SessionObjectT*)session->more_data)->autocommit = autocommit;
}

/*----------------------------------------*/

static bool
getAutocommit (
    tin01_sql_session * session)
{
    if (((SapDB_SessionObjectT*)session->more_data)->autocommit) {
        return true;
    }
    else {
        return false;
    }
}

/*----------------------------------------*/

static bool
lastWasEOT (
    tin01_sql_session * session)
{
    if (((SapDB_SessionObjectT*)session->more_data)->autocommit) {
        return true;
    }
    if (((SapDB_SessionObjectT*)session->more_data)->lastWasEOT) {
        return true;
    }
    return false;
}

/*----------------------------------------*/

static void
enableCursorNames (
    tin01_sql_session * session,
    bool enable)
{
    ((SapDB_SessionObjectT*)session->more_data)->implicitCursorNames = enable;
}

/*----------------------------------------*/

static int
nextCursorCount (
    tin01_sql_session * session)
{
    SapDB_SessionObjectT * self = (SapDB_SessionObjectT *) session->more_data;
    ++self->cursorCount;
    return self->cursorCount;
}

/*----------------------------------------*/

static void
addCursorName (
    tin01_sql_session * session)
{
    #define NAME_LEN_C 13
    char cursorName [NAME_LEN_C];
    int cursorCount = nextCursorCount (session);
    int index = NAME_LEN_C - 1;

    if (((SapDB_SessionObjectT *) session->more_data)->implicitCursorNames) {
        memcpy (cursorName, "PCURSOR_00000", NAME_LEN_C);
        while (cursorCount > 0) {
            cursorName [index] = (cursorCount % 10) + '0';
            cursorCount /= 10;
            --index;
        }
        i28newpart (session, sp1pk_resulttablename);
        i28addpartdata (session, cursorName, NAME_LEN_C);
    }
    #undef NAME_LEN_C
}

/*----------------------------------------*/

static void
clearEOTFlag (
    tin01_sql_session * session)
{
    ((SapDB_SessionObjectT*)session->more_data)->lastWasEOT = false;
}

/*----------------------------------------*/

static void
setEOTFlag (
    tin01_sql_session * session)
{
    ((SapDB_SessionObjectT*)session->more_data)->lastWasEOT = true;
}

/*----------------------------------------*/



static boolean
initializeRTE ()
{
    ROUTINE_DBG_MSP00 ("initializeRTE");
    tsp_compname compName;

    c2p (compName, sizeof (compName), "R SAP DB Scripting");
    sqlinit (compName, NULL);
    return true;
}

/*----------------------------------------*/

static void
fillXUser (
    tsp4_xuser_record * xuser,
    const char * user,
    const char * pwd,
    const char * dbname,
    const char * node)
{
    ROUTINE_DBG_MSP00 ("fillXUser");
    tsp_name clearPwd;
    int i = 0;

    c2p (xuser->xu_user, sizeof (xuser->xu_user), user);
    c2p (clearPwd, sizeof (clearPwd), pwd);
    for (i = 0; i < sizeof (clearPwd); ++i) {
        clearPwd [i] = toupper (clearPwd [i]);
    }
    s02applencrypt (clearPwd, xuser->xu_password);
    c2p (xuser->xu_serverdb, sizeof (xuser->xu_serverdb), dbname);
    c2p (xuser->xu_servernode, sizeof (xuser->xu_servernode), node);
    c2p (xuser->xu_sqlmode, sizeof (xuser->xu_sqlmode), "ADABAS");
    xuser->xu_cachelimit = -1;
    xuser->xu_timeout = -1;
    xuser->xu_isolation = -1;
    c2p (xuser->xu_dblang, sizeof (xuser->xu_dblang), "");;
}

/*----------------------------------------*/

static void
localRelease (
    tin01_sql_session* nself)
{
    ROUTINE_DBG_MSP00 ("localRelease");

    if ((nself != NULL) && (nself->is_connected)) {
        i28release (nself, 0);
    }
}

/*----------------------------------------*/


#if defined (PYTHON_GLUE)
static void
destructor_SapDB_Session (
    PyObject * pyself)
{
    ROUTINE_DBG_MSP00 ("destructor_SapDB_Session");
    SapDB_SessionObjectT* self = REINTERPRET_CAST (SapDB_SessionObjectT*, pyself);

    localRelease (self->nself);
    free (pyself);
}

/*----------------------------------------*/
#endif

static int
sql (
    tin01_sql_session * session,
    const char * cmd,
    SQLResultT * result)
{
    ROUTINE_DBG_MSP00 ("sql");
    int sqlrc;
    int retryCount = 3;

    initResult (result, 0);
    do {
        i28initadbs (session);
        i28pcmd (session, cmd, UNDEF_SP00);
        if (getAutocommit (session)) {
            session->segment->variant.C_3.sp1c_commit_immediately_F = true;
        }
        addCursorName (session);
        sqlrc = requestReceive (session, true, true);
        --retryCount;
    } while ((sqlrc == cin01_db_parse_again) && (retryCount > 0));
    packet2result (session, sqlrc, NULL, result);
    return sqlrc;
}

/*----------------------------------------*/

static int
sqlX (
    tin01_sql_session * session,
    const char * cmd,
    SequenceT parms,
    SQLResultT * result)
{
    ROUTINE_DBG_MSP00 ("sqlX");
    int sqlrc;
    PreparedT prepared;

    memset (&prepared, '\0', sizeof (prepared));
    sqlrc = prepare (session, &prepared, cmd, true);
    if (sqlrc == 0) {
        *result = executePrepared (&prepared, parms);
        closePrepared (&prepared);
    }
    else {
        result->isOK = false;
    }
    return sqlrc;
}

/*----------------------------------------*/

static int
utility (
    tin01_sql_session * session,
    const char * cmd,
    SQLResultT * result)
{
    ROUTINE_DBG_MSP00 ("utility");
    int sqlrc;

    initResult (result, 1);
    sqlrc = i28utility (session, cmd);
    if ((sqlrc == 0) || (i28findpart (session, sp1pk_data))) {
        packet2result (session, 0, NULL, result);
        if (result->isTuple) {
            setResultElement (result, -1, SL_fromInt (sqlrc));
        }
        sqlrc = 0;
    }
    else {
        result->isOK = false;
    }
    return sqlrc;
}

/*----------------------------------------*/

static int
commit (
    tin01_sql_session * session)
{
    int result = i28adbs_noinfo (session, "COMMIT WORK", false);
    setEOTFlag (session);
    return result;
}

/*----------------------------------------*/

static int
rollback (
    tin01_sql_session * session)
{
    int result = i28adbs_noinfo (session, "ROLLBACK WORK", false);
    setEOTFlag (session);
    return result;
}

/*----------------------------------------*/

#if defined (PERL_GLUE)
static void
increaseRef (
    void * obj)
{
    ROUTINE_DBG_MSP00 ("increaseRef");
    SapDB_SessionObjectT * perlSession =
        REINTERPRET_CAST (SapDB_SessionObjectT*, obj);

    ++perlSession->refCount;
}

/*----------------------------------------*/

static void
decreaseRef (
    void * obj)
{
    ROUTINE_DBG_MSP00 ("decreaseRef");
    SapDB_SessionObjectT * perlSession =
        REINTERPRET_CAST (SapDB_SessionObjectT*, obj);

    --perlSession->refCount;
    if (perlSession->refCount < 0) {
        localRelease (perlSession->nself);
        freeSapDB_Session (perlSession);
    }
}

#endif
/*----------------------------------------*/

/* endsnippet SapDB_Session Generic_precode */
/*----------------------------------------*/

externC XS (SapDB_Session_sql)
{
    ROUTINE_DBG_MSP00 ("SapDB_Session_sql");
/* no code for key Perl_/Generic_code */
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_SessionObjectT* self;
    tin01_sql_session* nself;
    char * cmd;
    SQLResultT sqlResult;
    int result;
    
    cv = cv;
    if (items != 2) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_SessionObjectT*) getGluePointer (ST(0), SapDB_SessionClassID, croakBuf);
    if ((self == NULL) || !isSapDB_SessionObject (self)) {
        croak (croakBuf);
    }
    nself = self->nself;
    switch (items) {
        case 2: cmd = (char *) SvPV (ST(1), na);
    }
    if (!ok) {
        ok = 0;
        goto cleanup_label;
    }
    if (!isSessionConnected (self->nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Perl_/Generic_nativeCall */
    result = sql (nself, cmd, &sqlResult);
    if ((result != 0) && sqlErrOccured (nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
    ST (argvi) = sv_newmortal ();
    if (sqlResult.isTuple) {
        sv_setsv (ST (argvi++), newRV (REINTERPRET_CAST (SV*, sqlResult.value.resultList)));
    }
    else {
        sv_setsv (ST (argvi++), sqlResult.value.resultObj);
    };
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        croak (croakBuf);
    }
    XSRETURN (argvi);
}

/*----------------------------------------*/

externC XS (SapDB_Session_sqlX)
{
    ROUTINE_DBG_MSP00 ("SapDB_Session_sqlX");
/* no code for key Perl_/Generic_code */
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_SessionObjectT* self;
    tin01_sql_session* nself;
    char * cmd;
    SequenceT parms =  EmptySequenceC;
    SQLResultT sqlResult;
    int result;
    
    cv = cv;
    if ((items < 2) || (items > 3)) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_SessionObjectT*) getGluePointer (ST(0), SapDB_SessionClassID, croakBuf);
    if ((self == NULL) || !isSapDB_SessionObject (self)) {
        croak (croakBuf);
    }
    nself = self->nself;
    switch (items) {
        case 3: ok = ok && sequencePerl2C (ST(2), &parms, croakBuf);;
        case 2: cmd = (char *) SvPV (ST(1), na);
    }
    if (!ok) {
        ok = 0;
        goto cleanup_label;
    }
    if (!isSessionConnected (self->nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Perl_/Generic_nativeCall */
    result = sqlX (nself, cmd, parms, &sqlResult);
    if ((result != 0) && sqlErrOccured (nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
    ST (argvi) = sv_newmortal ();
    if (sqlResult.isTuple) {
        sv_setsv (ST (argvi++), newRV (REINTERPRET_CAST (SV*, sqlResult.value.resultList)));
    }
    else {
        sv_setsv (ST (argvi++), sqlResult.value.resultObj);
    };
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        croak (croakBuf);
    }
    XSRETURN (argvi);
}

/*----------------------------------------*/

externC XS (SapDB_Session_prepare)
{
    ROUTINE_DBG_MSP00 ("SapDB_Session_prepare");
/* no code for key SapDB_Session::prepare Perl_/Generic_ in ['Generic_nativeCall vars']*/
    SapDB_PreparedObjectT * newObj = NULL;
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_SessionObjectT* self;
    tin01_sql_session* nself;
    char * cmd;
    int result;
    
    cv = cv;
    if (items != 2) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_SessionObjectT*) getGluePointer (ST(0), SapDB_SessionClassID, croakBuf);
    if ((self == NULL) || !isSapDB_SessionObject (self)) {
        croak (croakBuf);
    }
    nself = self->nself;
    switch (items) {
        case 2: cmd = (char *) SvPV (ST(1), na);
    }
    if (!ok) {
        ok = 0;
        goto cleanup_label;
    }
    if (!isSessionConnected (self->nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    newObj = newSapDB_Prepared ();
    if (allocFailed (newObj, croakBuf)) {
        ok = 0;
        newObj = NULL;
        goto cleanup_label;
    }
    newObj->classID = SapDB_PreparedClassID;
/* snippet SapDB_Session::prepare Generic_nativeCall */
result = prepare (nself, &newObj->nself, cmd, true);
if ((result != 0) && sqlErrOccured (nself, croakBuf)) {
    ok = 0;
    goto cleanup_label;
}
/* endsnippet SapDB_Session::prepare Generic_nativeCall */
    /* outargs */
    ST (argvi) = sv_newmortal ();
    sv_setref_pv (ST (argvi++), (char *)SapDB_PreparedClassID, (void *) newObj);
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        if (newObj != NULL) {
            free (newObj);
        }
        croak (croakBuf);
    }
    XSRETURN (argvi);
}

/*----------------------------------------*/

externC XS (SapDB_Session_release)
{
    ROUTINE_DBG_MSP00 ("SapDB_Session_release");
/* no code for key Perl_/Generic_code */
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_SessionObjectT* self;
    tin01_sql_session* nself;
    int withCommit =  0;
    
    cv = cv;
    if ((items < 1) || (items > 2)) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_SessionObjectT*) getGluePointer (ST(0), SapDB_SessionClassID, croakBuf);
    if ((self == NULL) || !isSapDB_SessionObject (self)) {
        croak (croakBuf);
    }
    nself = self->nself;
    switch (items) {
        case 2: withCommit = (int) SvIV (ST(1));
    }
    if (!ok) {
        ok = 0;
        goto cleanup_label;
    }
    if (!isSessionConnected (self->nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Perl_/Generic_nativeCall */
    i28release (nself, withCommit);
    /* outargs */
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        croak (croakBuf);
    }
    XSRETURN (argvi);
}

/*----------------------------------------*/

externC XS (SapDB_Session_commit)
{
    ROUTINE_DBG_MSP00 ("SapDB_Session_commit");
/* no code for key Perl_/Generic_code */
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_SessionObjectT* self;
    tin01_sql_session* nself;
    int result;
    
    cv = cv;
    if (items != 1) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_SessionObjectT*) getGluePointer (ST(0), SapDB_SessionClassID, croakBuf);
    if ((self == NULL) || !isSapDB_SessionObject (self)) {
        croak (croakBuf);
    }
    nself = self->nself;
    if (!isSessionConnected (self->nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Perl_/Generic_nativeCall */
    result = commit (nself);
    if ((result != 0) && sqlErrOccured (nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        croak (croakBuf);
    }
    XSRETURN (argvi);
}

/*----------------------------------------*/

externC XS (SapDB_Session_rollback)
{
    ROUTINE_DBG_MSP00 ("SapDB_Session_rollback");
/* no code for key Perl_/Generic_code */
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_SessionObjectT* self;
    tin01_sql_session* nself;
    int result;
    
    cv = cv;
    if (items != 1) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_SessionObjectT*) getGluePointer (ST(0), SapDB_SessionClassID, croakBuf);
    if ((self == NULL) || !isSapDB_SessionObject (self)) {
        croak (croakBuf);
    }
    nself = self->nself;
    if (!isSessionConnected (self->nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Perl_/Generic_nativeCall */
    result = rollback (nself);
    if ((result != 0) && sqlErrOccured (nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        croak (croakBuf);
    }
    XSRETURN (argvi);
}

/*----------------------------------------*/

externC XS (SapDB_Session_setAutocommit)
{
    ROUTINE_DBG_MSP00 ("SapDB_Session_setAutocommit");
/* no code for key Perl_/Generic_code */
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_SessionObjectT* self;
    tin01_sql_session* nself;
    int autocommit =  1;
    
    cv = cv;
    if ((items < 1) || (items > 2)) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_SessionObjectT*) getGluePointer (ST(0), SapDB_SessionClassID, croakBuf);
    if ((self == NULL) || !isSapDB_SessionObject (self)) {
        croak (croakBuf);
    }
    nself = self->nself;
    switch (items) {
        case 2: autocommit = (int) SvIV (ST(1));
    }
    if (!ok) {
        ok = 0;
        goto cleanup_label;
    }
    if (!isSessionConnected (self->nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Perl_/Generic_nativeCall */
    setAutocommit (nself, autocommit);
    /* outargs */
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        croak (croakBuf);
    }
    XSRETURN (argvi);
}

/*----------------------------------------*/

externC XS (SapDB_Session_getAutocommit)
{
    ROUTINE_DBG_MSP00 ("SapDB_Session_getAutocommit");
/* no code for key Perl_/Generic_code */
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_SessionObjectT* self;
    tin01_sql_session* nself;
    int result;
    
    cv = cv;
    if (items != 1) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_SessionObjectT*) getGluePointer (ST(0), SapDB_SessionClassID, croakBuf);
    if ((self == NULL) || !isSapDB_SessionObject (self)) {
        croak (croakBuf);
    }
    nself = self->nself;
    if (!isSessionConnected (self->nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Perl_/Generic_nativeCall */
    result = getAutocommit (nself);
    /* outargs */
    ST (argvi) = sv_newmortal ();
    sv_setiv (ST (argvi++), (IV) result);
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        croak (croakBuf);
    }
    XSRETURN (argvi);
}

/*----------------------------------------*/

externC XS (SapDB_Session_utility)
{
    ROUTINE_DBG_MSP00 ("SapDB_Session_utility");
/* no code for key Perl_/Generic_code */
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_SessionObjectT* self;
    tin01_sql_session* nself;
    char * cmd;
    SQLResultT sqlResult;
    int result;
    
    cv = cv;
    if (items != 2) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_SessionObjectT*) getGluePointer (ST(0), SapDB_SessionClassID, croakBuf);
    if ((self == NULL) || !isSapDB_SessionObject (self)) {
        croak (croakBuf);
    }
    nself = self->nself;
    switch (items) {
        case 2: cmd = (char *) SvPV (ST(1), na);
    }
    if (!ok) {
        ok = 0;
        goto cleanup_label;
    }
    if (!isSessionConnected (self->nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Perl_/Generic_nativeCall */
    result = utility (nself, cmd, &sqlResult);
    if ((result != 0) && sqlErrOccured (nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
    ST (argvi) = sv_newmortal ();
    if (sqlResult.isTuple) {
        sv_setsv (ST (argvi++), newRV (REINTERPRET_CAST (SV*, sqlResult.value.resultList)));
    }
    else {
        sv_setsv (ST (argvi++), sqlResult.value.resultObj);
    };
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        croak (croakBuf);
    }
    XSRETURN (argvi);
}

/*----------------------------------------*/

externC XS (SapDB_Session_switch)
{
    ROUTINE_DBG_MSP00 ("SapDB_Session_switch");
/* no code for key Perl_/Generic_code */
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_SessionObjectT* self;
    tin01_sql_session* nself;
    char * trace;
    char * test;
    int result;
    
    cv = cv;
    if (items != 3) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_SessionObjectT*) getGluePointer (ST(0), SapDB_SessionClassID, croakBuf);
    if ((self == NULL) || !isSapDB_SessionObject (self)) {
        croak (croakBuf);
    }
    nself = self->nself;
    switch (items) {
        case 3: test = (char *) SvPV (ST(2), na);
        case 2: trace = (char *) SvPV (ST(1), na);
    }
    if (!ok) {
        ok = 0;
        goto cleanup_label;
    }
    if (!isSessionConnected (self->nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Perl_/Generic_nativeCall */
    result = i28switch (nself, trace, test);
    if ((result != 0) && sqlErrOccured (nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        croak (croakBuf);
    }
    XSRETURN (argvi);
}

/*----------------------------------------*/

externC XS (SapDB_Session_switchlimit)
{
    ROUTINE_DBG_MSP00 ("SapDB_Session_switchlimit");
/* no code for key Perl_/Generic_code */
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_SessionObjectT* self;
    tin01_sql_session* nself;
    char * trace;
    char * test;
    char * startroutine;
    char * stoproutine;
    int buflimit;
    int result;
    
    cv = cv;
    if (items != 6) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_SessionObjectT*) getGluePointer (ST(0), SapDB_SessionClassID, croakBuf);
    if ((self == NULL) || !isSapDB_SessionObject (self)) {
        croak (croakBuf);
    }
    nself = self->nself;
    switch (items) {
        case 6: buflimit = (int) SvIV (ST(5));
        case 5: stoproutine = (char *) SvPV (ST(4), na);
        case 4: startroutine = (char *) SvPV (ST(3), na);
        case 3: test = (char *) SvPV (ST(2), na);
        case 2: trace = (char *) SvPV (ST(1), na);
    }
    if (!ok) {
        ok = 0;
        goto cleanup_label;
    }
    if (!isSessionConnected (self->nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Perl_/Generic_nativeCall */
    result = i28switchlimit (nself, trace, test, startroutine, stoproutine, buflimit);
    if ((result != 0) && sqlErrOccured (nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        croak (croakBuf);
    }
    XSRETURN (argvi);
}

/*----------------------------------------*/

externC XS (SapDB_Session_buflimit)
{
    ROUTINE_DBG_MSP00 ("SapDB_Session_buflimit");
/* no code for key Perl_/Generic_code */
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_SessionObjectT* self;
    tin01_sql_session* nself;
    int buflimit;
    int result;
    
    cv = cv;
    if (items != 2) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_SessionObjectT*) getGluePointer (ST(0), SapDB_SessionClassID, croakBuf);
    if ((self == NULL) || !isSapDB_SessionObject (self)) {
        croak (croakBuf);
    }
    nself = self->nself;
    switch (items) {
        case 2: buflimit = (int) SvIV (ST(1));
    }
    if (!ok) {
        ok = 0;
        goto cleanup_label;
    }
    if (!isSessionConnected (self->nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Perl_/Generic_nativeCall */
    result = i28buflimit (nself, buflimit);
    if ((result != 0) && sqlErrOccured (nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        croak (croakBuf);
    }
    XSRETURN (argvi);
}

/*----------------------------------------*/

externC XS (SapDB_Session_minbuf)
{
    ROUTINE_DBG_MSP00 ("SapDB_Session_minbuf");
/* no code for key Perl_/Generic_code */
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_SessionObjectT* self;
    tin01_sql_session* nself;
    int result;
    
    cv = cv;
    if (items != 1) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_SessionObjectT*) getGluePointer (ST(0), SapDB_SessionClassID, croakBuf);
    if ((self == NULL) || !isSapDB_SessionObject (self)) {
        croak (croakBuf);
    }
    nself = self->nself;
    if (!isSessionConnected (self->nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Perl_/Generic_nativeCall */
    result = i28minbuf (nself);
    if ((result != 0) && sqlErrOccured (nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        croak (croakBuf);
    }
    XSRETURN (argvi);
}

/*----------------------------------------*/

externC XS (SapDB_Session_maxbuf)
{
    ROUTINE_DBG_MSP00 ("SapDB_Session_maxbuf");
/* no code for key Perl_/Generic_code */
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_SessionObjectT* self;
    tin01_sql_session* nself;
    int result;
    
    cv = cv;
    if (items != 1) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_SessionObjectT*) getGluePointer (ST(0), SapDB_SessionClassID, croakBuf);
    if ((self == NULL) || !isSapDB_SessionObject (self)) {
        croak (croakBuf);
    }
    nself = self->nself;
    if (!isSessionConnected (self->nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Perl_/Generic_nativeCall */
    result = i28maxbuf (nself);
    if ((result != 0) && sqlErrOccured (nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        croak (croakBuf);
    }
    XSRETURN (argvi);
}


/*----------------------------------------*/

externC XS (SapDB_Session_destructor)
{
    ROUTINE_DBG_MSP00 ("SapDB_Session_destructor");
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    SapDB_SessionObjectT* self;
    tin01_sql_session* nself;

    cv = cv;
    if (items != 1) {
        croak (invalidArgCount_C);
    }
    self = (SapDB_SessionObjectT*) getGluePointer (ST(0), SapDB_SessionClassID, croakBuf);
    if ((self == NULL) || !isSapDB_SessionObject (self)) {
        croak (croakBuf);
    }
    nself = self->nself;
    decreaseRef (self);
    XSRETURN (0);
}

/*----------------------------------------*/

externC XS (sapdb_connect)
{
    ROUTINE_DBG_MSP00 ("sapdb_connect");
/* no code for key connect_sapdb Perl_/Generic_ in ['Generic_nativeCall']*/
    SapDB_SessionObjectT * newObj = NULL;
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    char * user;
    char * pwd;
    char * dbname;
    char * host = LIT( "");
    char * config = LIT( "");
    tsp_int4 result;
    
    cv = cv;
    if ((items < 3) || (items > 5)) {
        croak (invalidArgCount_C);
    }
    switch (items) {
        case 5: config = (char *) SvPV (ST(4), na);
        case 4: host = (char *) SvPV (ST(3), na);
        case 3: dbname = (char *) SvPV (ST(2), na);
        case 2: pwd = (char *) SvPV (ST(1), na);
        case 1: user = (char *) SvPV (ST(0), na);
    }
    if (!ok) {
        ok = 0;
        goto cleanup_label;
    }
    newObj = newSapDB_Session ();
    if (allocFailed (newObj, croakBuf)) {
        ok = 0;
        newObj = NULL;
        goto cleanup_label;
    }
    newObj->classID = SapDB_SessionClassID;
/* snippet connect_sapdb Generic_nativeCall */
newObj->nself = &newObj->session;
i28initsession (newObj->nself, NULL);
newObj->session.lasterr.lasterr_on = true;
newObj->autocommit = false;
newObj->cursorCount = 0;
newObj->implicitCursorNames = true;
newObj->lastWasEOT = true;
/* set special senderid */
setStringConfig ((char *) newObj->session.senderid + sizeof (tsp00_C5),
    sizeof (tsp00_C3), "DOM", 3);
fillXUser (&newObj->nself->xuser, user, pwd, dbname, host);
newObj->nself->more_data = newObj;
parseSessionConfig (newObj->nself, config);
result = i28sqlconnect (newObj->nself);
if ((result != 0) && sqlErrOccured (newObj->nself, croakBuf)) {
    ok = 0;
    goto cleanup_label;
}
/* endsnippet connect_sapdb Generic_nativeCall */
    /* outargs */
    ST (argvi) = sv_newmortal ();
    sv_setref_pv (ST (argvi++), (char *)SapDB_SessionClassID, (void *) newObj);
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        if (newObj != NULL) {
            free (newObj);
        }
        croak (croakBuf);
    }
    XSRETURN (argvi);
}

/*----------------------------------------*/

externC XS (sapdb_utilitySession)
{
    ROUTINE_DBG_MSP00 ("sapdb_utilitySession");
/* no code for key utilitySession_sapdb Perl_/Generic_ in ['Generic_nativeCall']*/
    SapDB_SessionObjectT * newObj = NULL;
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    char * user;
    char * pwd;
    char * dbname;
    char * host = LIT( "");
    tsp_int4 result;
    
    cv = cv;
    if ((items < 3) || (items > 4)) {
        croak (invalidArgCount_C);
    }
    switch (items) {
        case 4: host = (char *) SvPV (ST(3), na);
        case 3: dbname = (char *) SvPV (ST(2), na);
        case 2: pwd = (char *) SvPV (ST(1), na);
        case 1: user = (char *) SvPV (ST(0), na);
    }
    if (!ok) {
        ok = 0;
        goto cleanup_label;
    }
    newObj = newSapDB_Session ();
    if (allocFailed (newObj, croakBuf)) {
        ok = 0;
        newObj = NULL;
        goto cleanup_label;
    }
    newObj->classID = SapDB_SessionClassID;
/* snippet utilitySession_sapdb Generic_nativeCall */
newObj->nself = &newObj->session;
i28initsession (newObj->nself, NULL);
newObj->session.lasterr.lasterr_on = true;
fillXUser (&newObj->nself->xuser, user, pwd, dbname, host);
newObj->nself->as_utility = 1;
result = i28sqlconnect (newObj->nself);
if ((result != 0) && sqlErrOccured (newObj->nself, croakBuf)) {
    ok = 0;
    goto cleanup_label;
}
newObj->nself->more_data = newObj;
/* endsnippet utilitySession_sapdb Generic_nativeCall */
    /* outargs */
    ST (argvi) = sv_newmortal ();
    sv_setref_pv (ST (argvi++), (char *)SapDB_SessionClassID, (void *) newObj);
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        if (newObj != NULL) {
            free (newObj);
        }
        croak (croakBuf);
    }
    XSRETURN (argvi);
}

/*----------------------------------------*/

externC XS (sapdb__buildInfo)
{
    ROUTINE_DBG_MSP00 ("sapdb__buildInfo");
/* no code for key Perl_/Generic_code */
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    char * result;
    
    cv = cv;
    if (items != 0) {
        croak (invalidArgCount_C);
    }
/* no code for key Perl_/Generic_nativeCall */
    result = buildInfo ();
    /* outargs */
    ST (argvi) = sv_newmortal ();
    sv_setpv (ST (argvi++), result);
    XSRETURN (argvi);
}

/*----------------------------------------*/

externC XS (sapdb_createCursorZombie)
{
    ROUTINE_DBG_MSP00 ("sapdb_createCursorZombie");
/* no code for key createCursorZombie Perl_/Generic_ in ['Generic_nativeCall vars']*/
    SapDB_ResultSetObjectT * newObj = NULL;
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    
    cv = cv;
    if (items != 0) {
        croak (invalidArgCount_C);
    }
    newObj = newSapDB_ResultSet ();
    if (allocFailed (newObj, croakBuf)) {
        ok = 0;
        newObj = NULL;
        goto cleanup_label;
    }
    newObj->classID = SapDB_ResultSetClassID;
/* snippet createCursorZombie Generic_nativeCall */
 /* create simply an empty ResultSet without initialization */
/* endsnippet createCursorZombie Generic_nativeCall */
    /* outargs */
    ST (argvi) = sv_newmortal ();
    sv_setref_pv (ST (argvi++), (char *)SapDB_ResultSetClassID, (void *) newObj);
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        if (newObj != NULL) {
            free (newObj);
        }
        croak (croakBuf);
    }
    XSRETURN (argvi);
}

/*----------------------------------------*/

externC XS (sapdb_createReaderZombie)
{
    ROUTINE_DBG_MSP00 ("sapdb_createReaderZombie");
/* no code for key createReaderZombie Perl_/Generic_ in ['Generic_nativeCall vars']*/
    SapDB_LongReaderObjectT * newObj = NULL;
    int ok = 1;
    int argvi = 0;
    char croakBuf [CROAKBUF_SIZE];
    dXSARGS;
    
    cv = cv;
    if (items != 0) {
        croak (invalidArgCount_C);
    }
    newObj = newSapDB_LongReader ();
    if (allocFailed (newObj, croakBuf)) {
        ok = 0;
        newObj = NULL;
        goto cleanup_label;
    }
    newObj->classID = SapDB_LongReaderClassID;
/* snippet createReaderZombie Generic_nativeCall */
 /* create simply an empty LongReader without initialization */
/* endsnippet createReaderZombie Generic_nativeCall */
    /* outargs */
    ST (argvi) = sv_newmortal ();
    sv_setref_pv (ST (argvi++), (char *)SapDB_LongReaderClassID, (void *) newObj);
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        if (newObj != NULL) {
            free (newObj);
        }
        croak (croakBuf);
    }
    XSRETURN (argvi);
}


#if defined (ACTIVE_PERL)
#define boot_SAP__DBTech__sapdb boot_SAP__DBTech__sapdbaperl
#else
#define boot_SAP__DBTech__sapdb boot_SAP__DBTech__sapdbcperl
#endif

externC GLUEEXPORT XS(boot_SAP__DBTech__sapdb)
{
    dXSARGS;
    char * file = CONST_CAST (char *, __FILE__);
    cv = cv;
    items = items;
    newXS (LIT ("SAP::DBTech::sapdbc::sapdb_connect"), sapdb_connect, file);
    newXS (LIT ("SAP::DBTech::sapdbc::sapdb_utilitySession"), sapdb_utilitySession, file);
    newXS (LIT ("SAP::DBTech::sapdbc::sapdb__buildInfo"), sapdb__buildInfo, file);
    newXS (LIT ("SAP::DBTech::sapdbc::sapdb_createCursorZombie"), sapdb_createCursorZombie, file);
    newXS (LIT ("SAP::DBTech::sapdbc::sapdb_createReaderZombie"), sapdb_createReaderZombie, file);
    newXS (LIT ("SAP::DBTech::sapdbc::SapDB_LongReader_read"), SapDB_LongReader_read, file);
    newXS (LIT ("SAP::DBTech::sapdbc::SapDB_ResultSet_next"), SapDB_ResultSet_next, file);
    newXS (LIT ("SAP::DBTech::sapdbc::SapDB_ResultSet_previous"), SapDB_ResultSet_previous, file);
    newXS (LIT ("SAP::DBTech::sapdbc::SapDB_ResultSet_relative"), SapDB_ResultSet_relative, file);
    newXS (LIT ("SAP::DBTech::sapdbc::SapDB_ResultSet_absolute"), SapDB_ResultSet_absolute, file);
    newXS (LIT ("SAP::DBTech::sapdbc::SapDB_ResultSet_first"), SapDB_ResultSet_first, file);
    newXS (LIT ("SAP::DBTech::sapdbc::SapDB_ResultSet_last"), SapDB_ResultSet_last, file);
    newXS (LIT ("SAP::DBTech::sapdbc::SapDB_ResultSet_current"), SapDB_ResultSet_current, file);
    newXS (LIT ("SAP::DBTech::sapdbc::SapDB_ResultSet_cursorName"), SapDB_ResultSet_cursorName, file);
    newXS (LIT ("SAP::DBTech::sapdbc::SapDB_ResultSet_columnCount"), SapDB_ResultSet_columnCount, file);
    newXS (LIT ("SAP::DBTech::sapdbc::SapDB_ResultSet_columnInfo"), SapDB_ResultSet_columnInfo, file);
    newXS (LIT ("SAP::DBTech::sapdbc::SapDB_ResultSet_getDescription"), SapDB_ResultSet_getDescription, file);
    newXS (LIT ("SAP::DBTech::sapdbc::SapDB_Prepared_execute"), SapDB_Prepared_execute, file);
    newXS (LIT ("SAP::DBTech::sapdbc::SapDB_Prepared_getDescription"), SapDB_Prepared_getDescription, file);
    newXS (LIT ("SAP::DBTech::sapdbc::SapDB_Session_sql"), SapDB_Session_sql, file);
    newXS (LIT ("SAP::DBTech::sapdbc::SapDB_Session_sqlX"), SapDB_Session_sqlX, file);
    newXS (LIT ("SAP::DBTech::sapdbc::SapDB_Session_prepare"), SapDB_Session_prepare, file);
    newXS (LIT ("SAP::DBTech::sapdbc::SapDB_Session_release"), SapDB_Session_release, file);
    newXS (LIT ("SAP::DBTech::sapdbc::SapDB_Session_commit"), SapDB_Session_commit, file);
    newXS (LIT ("SAP::DBTech::sapdbc::SapDB_Session_rollback"), SapDB_Session_rollback, file);
    newXS (LIT ("SAP::DBTech::sapdbc::SapDB_Session_setAutocommit"), SapDB_Session_setAutocommit, file);
    newXS (LIT ("SAP::DBTech::sapdbc::SapDB_Session_getAutocommit"), SapDB_Session_getAutocommit, file);
    newXS (LIT ("SAP::DBTech::sapdbc::SapDB_Session_utility"), SapDB_Session_utility, file);
    newXS (LIT ("SAP::DBTech::sapdbc::SapDB_Session_switch"), SapDB_Session_switch, file);
    newXS (LIT ("SAP::DBTech::sapdbc::SapDB_Session_switchlimit"), SapDB_Session_switchlimit, file);
    newXS (LIT ("SAP::DBTech::sapdbc::SapDB_Session_buflimit"), SapDB_Session_buflimit, file);
    newXS (LIT ("SAP::DBTech::sapdbc::SapDB_Session_minbuf"), SapDB_Session_minbuf, file);
    newXS (LIT ("SAP::DBTech::sapdbc::SapDB_Session_maxbuf"), SapDB_Session_maxbuf, file);
    newXS (LIT("SAP::DBTech::sapdbc::SapDB_LongReader_destructor"), SapDB_LongReader_destructor, file);
    newXS (LIT("SAP::DBTech::sapdbc::SapDB_ResultSet_destructor"), SapDB_ResultSet_destructor, file);
    newXS (LIT("SAP::DBTech::sapdbc::SapDB_Prepared_destructor"), SapDB_Prepared_destructor, file);
    newXS (LIT("SAP::DBTech::sapdbc::SapDB_Session_destructor"), SapDB_Session_destructor, file);
/* snippet Generic_initroutine */
    IFSLOW_MEO00 (ScriptingTracer::initTrace ());
    initializeRTE ();
/* endsnippet Generic_initroutine */
    ST(0) = &PL_sv_yes;
    XSRETURN(1);
}

