/*++ BUILD Version: 0001    // Increment this if a change has global effects


               INTEL CORPORATION PROPRIETARY INFORMATION
  This software is supplied under the terms of a license agreement or
  nondisclosure agreement with Intel Corporation and may not be copied
  or disclosed except in accordance with the terms of that agreement.
        Copyright (c) 1998 Intel Corporation. All Rights Reserved.

Module Name:

    pse36api.h

Abstract:

    This is the include file that defines all common constants and types
    for accessing the PSE36 driver device

Revision History:



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

--*/


#ifndef _PSE36API_H_
#define _PSE36API_H_

//
// NtDeviceIoControlFile that defines the APIs for the PSE36 driver
//

//
// Define the device type value.
//

#define FILE_DEVICE_PSE36               0x8650

//
// IoControlCode values for PSE36 driver devices.
//

#define IOCTL_PSE36_BASE                FILE_DEVICE_PSE36


#define IOCTL_PSE36_GET_SIZE         CTL_CODE(IOCTL_PSE36_BASE, 0x0A00, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define IOCTL_PSE36_PERFORMANCE      CTL_CODE(IOCTL_PSE36_BASE, 0x0A01, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define IOCTL_PSE36_READ_SCATTER     CTL_CODE(IOCTL_PSE36_BASE, 0x0A80,  METHOD_NEITHER, FILE_READ_ACCESS)
#define IOCTL_PSE36_WRITE_GATHER     CTL_CODE(IOCTL_PSE36_BASE, 0x0A81, METHOD_NEITHER, FILE_WRITE_ACCESS)
#define IOCTL_PSE36_WRITE_FROM       CTL_CODE(IOCTL_PSE36_BASE, 0x0A90, METHOD_NEITHER, FILE_READ_ACCESS)
#define IOCTL_PSE36_READ_TO          CTL_CODE(IOCTL_PSE36_BASE, 0x0A91, METHOD_NEITHER, FILE_WRITE_ACCESS)

//
// IOCTL_PSE36_GET_SIZE returns the size of the PSE36 driver device in bytes.
//
// Define the structures for get size
//

typedef struct _PSE36_SIZE {
    LONGLONG Size;
    ULONG NumberOfPdes;
} PSE36_SIZE, *PPSE36_SIZE;

//
// IOCTL_PSE36_PERFORMANCE returns summary disk request statistics used
// in measuring performance.
//
// The following structure is exchanged on an IOCTL_PSE36_PERFORMANCE
// request. This ioctl collects summary disk request statistics used
// in measuring performance.
//

#define PSE36_MAX_PDES          9            // 8 procs + 1 extra

typedef struct _PSE36_PERFORMANCE {
        LARGE_INTEGER BytesRead;                // 0x0
        LARGE_INTEGER BytesWritten;             // 0x8
        ULONG ReadCount;                        // 0x10
        ULONG WriteCount;                       // 0x14
        LARGE_INTEGER FastBytesRead;            // 0x18
        LARGE_INTEGER FastBytesWritten;         // 0x20
        ULONG FastReadCount;                    // 0x28
        ULONG FastWriteCount;                   // 0x2C
        ULONG IoctlReadDescriptorCount;         // 0x30
        ULONG IoctlWriteDescriptorCount;        // 0x34
        ULONG IoctlReadCount;                   // 0x38
        ULONG IoctlWriteCount;                  // 0x3C
        LARGE_INTEGER DiskBytesRead;            // 0x40
        LARGE_INTEGER DiskBytesWritten;         // 0x48
        ULONG DiskReadDescriptorCount;          // 0x50
        ULONG DiskWriteDescriptorCount;         // 0x54
        ULONG DiskReadCount;                    // 0x58
        ULONG DiskWriteCount;                   // 0x5C
        ULONG NumberOfPdes;                     // 0x60
        ULONG NumberOfPages;                    // 0x64
        ULONG WaitForPDE;                       // 0x68
        ULONG ExtraPdeChecks;                   // 0x6C
        ULONG LockCount[PSE36_MAX_PDES + 1];    // 0x70
} PSE36_PERFORMANCE, *PPSE36_PERFORMANCE;

//
// IOCTL_PSE36_READ_SCATTER and IOCTL_PSE36_WRITE_SCATTER do
// scatter/gather I/O to the PSE36 driver device.  They take a varible
// length list of buffer address disk offset pairs.
//

typedef struct _PSE36_SCATTER_GATHER_DESCRIPTOR {
    LONGLONG Offset;    
    PVOID    Buffer;
    ULONG    BufferSize;
} PSE36_SCATTER_GATHER_DESCRIPTOR, *PPSE36_SCATTER_GATHER_DESCRIPTOR;

typedef struct _PSE36_SCATTER_GATHER_IO {
    LONG     DescriptorCount;
    LONG     Pad;
    PSE36_SCATTER_GATHER_DESCRIPTOR ScatterGatherList[0];
} PSE36_SCATTER_GATHER_IO, *PPSE36_SCATTER_GATHER_IO;

//
// IOCTL_PSE36_WRITE_FROM and IOCTL_PSE36_READ_TO transfer data between
// the PSE36 driver device and a real disk.  They take the arguments for
// the disk I/O as well as the offset into the PSE36 driver device.
//

typedef struct _PSE36_INDIRECT_DESCRIPTOR {
    LARGE_INTEGER  PSE36DeviceOffset;
    LARGE_INTEGER  FileOffset;
    ULONG	   TransferSize;
    LONG           Pad;
} PSE36_INDIRECT_DESCRIPTOR, *PPSE36_INDIRECT_DESCRIPTOR;

typedef struct _PSE36_INDIRECT_IO {
    HANDLE         hFile;
    LONG           DescriptorCount;
    PSE36_INDIRECT_DESCRIPTOR DescriptorList[0];
} PSE36_INDIRECT_IO, *PPSE36_INDIRECT_IO;

//
// PSE36 driver device name strings
//

#define PSE36_DEFAULT_DOS_NAME    "\\\\.\\PSE36"
#define PSE36_DEVICE_NAME         "PSE36"

//
// The PSE36 driver device can only be used by a single process.  In order
// for performance tools to work a second device is created.  The name of
// this device is created by appending PSE36_PERF_SUFFIX to the original
// device name.  For example the default name for this device would be
// \\.\PSE36Perf
//

#define PSE36_PERF_SUFFIX         "Perf"

#endif // _PSE36API_H_
