/* @lastChanged: "1999-05-31"

  * @filename:   vkb04.cpp
  * @purpose:    "KB_object_handling_interface"
  * @release:    7.2.0.0
  * @see:        ""
  *
  * @copyright:  "Copyright by SAP AG, 1999"


    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*/


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include <string.h>

#include "hkb04.h"

#include "hbd01.h"
#include "hbd04.h"
#include "hbd12.h"
//???? kann weg, wenn LogReadAccess in der Parameterleiste steht
#include "hbd21.h"
#include "hgg01.h"
#include "hkb05.h"
#include "hkb51.h"
#include "hkb53.h"
#include "hkb56.h"
#include "hkb64.h"

#if COMPILEMODE_MEO00 >= SLOW_MEO00	
#include "hta99.h"
#include "hta01.h"
#endif

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define NIL_WORK_STACK		0
#define NIL_WORK_STACK_MAX	0

struct t_TraceOutput
{
    tgg00_MessType     trcMessType;
    tgg00_MessType2    trcMessType2;
    tgg00_BasisError   trcError; 
    tgg91_TransNo      trcConsistView;
};

/*===========================================================================*
 *  GLOBAL FUNCTIONS                                                         *
 *===========================================================================*/

inline static void kb04_EnterKb (tgg00_TransContext	   &Trans,
                                 tgg00_MessType_Enum    MessType,
                                 tgg00_MessType2_Enum	MessType2,
                                 bool                   WriteVtrace)
{
    
    Trans.trState_gg00.delElement (tsFunctRollbSet_egg00);
    Trans.trRteCommPtr_gg00->virt_reads = -MAX_INT2_SP00;
    
    if  (WriteVtrace)
    {
        t_TraceOutput TraceOutput;
        
        TraceOutput.trcMessType.becomes(MessType);
        TraceOutput.trcMessType2.becomes(MessType2);
        TraceOutput.trcError       = e_ok;
        TraceOutput.trcConsistView = Trans.trConsistView_gg00; 
        
        b12vtrace (Trans, kb, kb04oms, sizeof (TraceOutput), &TraceOutput);
    }
    
    if (Trans.trRteCommPtr_gg00->to_cancel)
    {
        Trans.trError_gg00 = e_cancelled;
        return;
    }
    
    if (cgg_nil_transindex == Trans.trIndex_gg00)
    {
        Trans.trError_gg00 = e_nil_transindex; // trans entry of this session not found
    }
    
#   if COMPILEMODE_MEO00 >= SLOW_MEO00
    t01trans (kb, "Trans (KB04)", Trans);
#   endif   
}


/*-----------------------------------------------------------------------------------*/

inline static void kb04_ErrorHandling (tgg00_TransContext   &Trans,
                                       tgg00_MessType_Enum	 MessType,
                                       tgg00_MessType2_Enum	 MessType2,
                                       bool                  WriteVtrace)
{
    if  (e_ok != Trans.trError_gg00)
    {
        kb05ReturnErrorHandling (Trans, MessType, MessType2,
            NIL_WORK_STACK, NIL_WORK_STACK_MAX);
    }
    
    if  (WriteVtrace)
    {
        t_TraceOutput TraceOutput;
        
        TraceOutput.trcMessType.becomes(m_return_error);
        TraceOutput.trcMessType2.becomes((m_begin == MessType) ? mm_trans : mm_nil); // mark StartConsistTrans
        TraceOutput.trcError       = Trans.trError_gg00;
        TraceOutput.trcConsistView = Trans.trConsistView_gg00;
        
        b12vtrace (Trans, kb, kb04oms, sizeof (TraceOutput), &TraceOutput);
    }
}

/*-----------------------------------------------------------------------------------*/

externCpp void
kb04CreateObjFile (tgg00_TransContext	&Trans,
                   tgg00_FileId 	    &ObjFile,
                   tsp00_Int4       	 ObjBodyLen,
                   tsp00_Int4            ObjKeyLen)			   
{ 
    ROUTINE_DBG_MEO00 ("kb04CreateObjFile");
    
    const tgg00_MessType_Enum	ThisMessType  = m_create_file;
    const tgg00_MessType2_Enum	ThisMessType2 = mm_object;
    const bool                  WriteVtrace   = (g01vtrace.vtrAll_gg00 || g01vtrace.vtrOmsNew_gg00 );
    
    tgg00_LockReqMode	GrantedLock;
    tgg00_Lkey			DummyKey;
    
    if (e_ok != Trans.trError_gg00) return;
    kb04_EnterKb (Trans, ThisMessType, ThisMessType2, WriteVtrace);
    
    if (e_ok != Trans.trError_gg00) return; // timeout !
    
#   if COMPILEMODE_MEO00 >= SLOW_MEO00	
    t01treeid (kb, "create Obj  ", ObjFile);
    t01trans  (kb, "trans       ", Trans );
#   endif
    
    GrantedLock.becomes(lckFree_egg00);
    k53eot_excl_lock (Trans, ObjFile, DummyKey, lckSysExcl_egg00, GrantedLock);
    
    if  (e_ok == Trans.trError_gg00)
    {
        ObjFile.fileVersion_gg00().gg90SetZero();
        
        k64create_bd_file (Trans, ObjFile, ThisMessType, ThisMessType2,
            0, ObjBodyLen, ObjKeyLen);
    }
    kb04_ErrorHandling (Trans, ThisMessType, ThisMessType2, WriteVtrace);
}

/*----------------------------------------------------------------------------*/

externCpp void
kb04CreateObjIndex (tgg00_TransContext  &Trans,
                    tgg00_FileId		&ObjFileId,
                    tgg00_ObjColDesc	&InvColDesc)
                    
{ 
    ROUTINE_DBG_MEO00 ("kb04CreateObjIndex");
    
#   define NO_TREE_INDEX        true
#   define INDEX_LOCK_BYTE      1
    
    const tgg00_MessType_Enum	ThisMessType  = m_create_index;
    const tgg00_MessType2_Enum	ThisMessType2 = mm_object;
    const bool                  WriteVtrace   = (g01vtrace.vtrAll_gg00 || g01vtrace.vtrOmsNew_gg00 );
    
    tgg00_LockReqMode           GrantedLock;
    tgg00_FileId                ObjInvFileId;
    tgg00_Lkey                  InvLockKey;
    
    
    if (e_ok != Trans.trError_gg00) return;
    
    kb04_EnterKb (Trans, ThisMessType, ThisMessType2, WriteVtrace);
    
    if (e_ok != Trans.trError_gg00) return; // timeout !
    
    InvLockKey.keyVal_gg00() [1] = INDEX_LOCK_BYTE;
    InvLockKey.keyLen_gg00()     = 1;
    
    // Initialize ObjInvFile
    
    k53eot_excl_lock (Trans, ObjInvFileId, InvLockKey, lckRowExcl_egg00, GrantedLock);
    
    if (e_ok == Trans.trError_gg00)
    {
        b01fcreate_file (Trans, ObjFileId, 0, cgg_nil_varcol_cnt_off,
            !NO_TREE_INDEX);
        
        if (e_ok == Trans.trError_gg00)
        {
            //bd04CreateObjIndex (Trans, ObjFileId, ObjInvFileId, InvColDesc);
            Trans.trError_gg00 = e_not_implemented;
        }
    }
    
    kb04_ErrorHandling (Trans, ThisMessType, ThisMessType2, WriteVtrace); 
}

/*----------------------------------------------------------------------------*/

externCpp void
kb04DeleteAllObj (tgg00_TransContext  &Trans,
				  tgg00_FileId		  &ObjFile,
				  tsp00_Int4          &NumDelObj)				  
{ 
	ROUTINE_DBG_MSP00 ("kb04DeleteAllObj");
	
	const tgg00_MessType_Enum	ThisMessType  = m_delete;
	const tgg00_MessType2_Enum	ThisMessType2 = mm_object;
	const bool                  WriteVtrace   = (cgg_is_true == g01vtrace.vtrOmsUpd_gg00);
	
	if (Trans.trError_gg00 != e_ok) return;

    kb04_EnterKb (Trans, ThisMessType, ThisMessType2, WriteVtrace);
	
    if (Trans.trError_gg00 != e_ok) return; // timeout !

    bd04DelAllObj (Trans, ObjFile, NumDelObj);
	
	kb04_ErrorHandling (Trans, ThisMessType, ThisMessType2, WriteVtrace);      
}


/*----------------------------------------------------------------------------*/

externCpp void
kb04DeleteObj (tgg00_TransContext	&Trans,
               tgg00_FileId			&ObjFile,
               tgg92_KernelOid  	 Oid,
               tgg91_PageRef	     WantedObjVers,
               tgg00_ObjColDesc		&ColDesc)
               
{ 
    ROUTINE_DBG_MEO00 ("kb04DeleteObj");
    
    const tgg00_MessType_Enum	ThisMessType  = m_delete;
    const tgg00_MessType2_Enum	ThisMessType2 = mm_object;
    const bool                  WriteVtrace   = (cgg_is_true == g01vtrace.vtrOmsUpd_gg00);
    
    if (e_ok != Trans.trError_gg00) return;
    
    kb04_EnterKb (Trans, ThisMessType, ThisMessType2, WriteVtrace);
    
    if (e_ok != Trans.trError_gg00) return; // timeout !
    
    bd04DelObj (Trans, ObjFile, Oid, WantedObjVers);
    
    kb04_ErrorHandling (Trans, ThisMessType, ThisMessType2, WriteVtrace);      
}


/*----------------------------------------------------------------------------*/

externCpp void
kb04EndConsistentTrans (tgg00_TransContext &Trans)
{ 
    ROUTINE_DBG_MEO00 ("kb04EndConsistentTrans");
    
    const tgg00_MessType_Enum	ThisMessType  = m_end;
    const tgg00_MessType2_Enum	ThisMessType2 = mm_trans;
    const bool                  WriteVtrace   = (cgg_is_true == g01vtrace.vtrOmsGet_gg00);
    
    if (e_ok != Trans.trError_gg00) return;
    
	// PTS 1107806 AK 09/09/2000
    Trans.trRteCommPtr_gg00->to_cancel = false;

    kb04_EnterKb (Trans, ThisMessType, ThisMessType2, WriteVtrace);
    
    if (e_ok != Trans.trError_gg00) return; // timeout !
    
    kb51DelConsistTrans (Trans.trTaskId_gg00, Trans.trConsistView_gg00 );
    Trans.trConsistView_gg00.gg90SetNil();
    
    kb04_ErrorHandling (Trans, ThisMessType, ThisMessType2, WriteVtrace);   
}

/*----------------------------------------------------------------------------*/

externCpp void
kb04GetObjConsistent (tgg00_TransContext &Trans,
                      tgg00_FileId	     &ObjFile,	   
                      tgg92_KernelOid     Oid,
                      tsp00_Int4	       ObjBodySize,  //[inp=buffer size for object] 
                      void			         *pObjBody,     //[inp]
          					  bool                bDoLock,      //[inp] 
                      tsp00_Int4         &ObjBodyLen,   //[out=real object length]
                      tgg91_PageRef      &ResultObjVers)
{ 
    ROUTINE_DBG_MEO00 ("kb04GetObjConsistent");
    
    const tgg00_MessType_Enum	ThisMessType  = m_get_object;
    const tgg00_MessType2_Enum	ThisMessType2 = mm_commit;
    const bool                  WriteVtrace   = (cgg_is_true == g01vtrace.vtrOmsGet_gg00);
	
    tsp00_Byte *pKey   = NULL;
    tsp00_Int4  KeyLen = 0;
    
    if (e_ok != Trans.trError_gg00) return;
    
    kb04_EnterKb (Trans, ThisMessType, ThisMessType2, WriteVtrace);
    
    if (e_ok != Trans.trError_gg00) return; // timeout !
    
    if ((cgg_nil_transindex == Trans.trIndex_gg00)
        ||
        Trans.trConsistView_gg00.gg90IsNil() )
    {
        Trans.trError_gg00 = e_unknown_consist_vers;
    }
    else
    {
        const tsp00_Int4  ZeroHashKeyLen  = 0;
        void             *DummyHashKeyPtr = NULL;

		//???? kann weg, wenn LogReadAccess in der Parameterleiste steht
		tsp00_Int4  LogReadAccesses;
        
        bd04GetObjConsistent (Trans, ObjFile, Oid, ObjBodySize, pObjBody,
            ZeroHashKeyLen, DummyHashKeyPtr, bDoLock, ObjBodyLen,
			KeyLen, &pKey, ResultObjVers, LogReadAccesses); //PTS 1107819 AK 12/09/2000
        
        //???? kann weg, wenn LogReadAccess in der Parameterleiste steht
        b21mp_obj_hist_read_count_put (Trans.trBdTcachePtr_gg00, LogReadAccesses);
    }
    
#   if COMPILEMODE_MEO00 >= SLOW_MEO00
    if (e_ok == Trans.trError_gg00) t01moveobj (kb, pObjBody, 1, ObjBodyLen);
#   endif   
    
    kb04_ErrorHandling (Trans, ThisMessType, ThisMessType2, WriteVtrace);   
}

/*----------------------------------------------------------------------------*/

externCpp void
kb04GetObjWithKeyConsistent (tgg00_TransContext	   &Trans,
                             tgg00_FileId	       &ObjFile,
                             tsp00_Int4		        ObjBodySize,
                             void			       *pObjBody,
                             tsp00_Int4	            KeyLen,
                             void		           *pKey,
                    		 bool                  bDoLock,      //[inp] 
                             tgg92_KernelOid       &ResultOid,
                             tgg91_PageRef         &ResultObjVers)
{ 
    ROUTINE_DBG_MEO00 ("kb04GetObjWithKeyConsistent");
    
    const tgg00_MessType_Enum	ThisMessType  = m_get_object;
    const tgg00_MessType2_Enum	ThisMessType2 = mm_object_id;
    const bool                  WriteVtrace   = (cgg_is_true == g01vtrace.vtrOmsGet_gg00);

    if (e_ok != Trans.trError_gg00) return;
    
    kb04_EnterKb (Trans, ThisMessType, ThisMessType2, WriteVtrace);
    
    if (e_ok != Trans.trError_gg00) return; // timeout !
    
    if ((cgg_nil_transindex == Trans.trIndex_gg00)
        ||
        Trans.trConsistView_gg00.gg90IsNil() )
    {
        Trans.trError_gg00 = e_unknown_consist_vers;
    }
    else
    {
		//???? kann weg, wenn LogReadAccess in der Parameterleiste steht
		tsp00_Int4  LogReadAccesses;

        bd04GetObjWithKeyConsistent (Trans, ObjFile, ObjBodySize,
			pObjBody, KeyLen, pKey, bDoLock, ResultOid, ResultObjVers, LogReadAccesses); //PTS 1107819 AK 12/09/2000
        
        //???? kann weg, wenn LogReadAccess in der Parameterleiste steht
        b21mp_obj_hist_read_count_put (Trans.trBdTcachePtr_gg00, LogReadAccesses);
 
        if (e_object_not_found == Trans.trError_gg00)
            Trans.trError_gg00 = e_hash_key_not_found;
    }
    
#   if COMPILEMODE_MEO00 >= SLOW_MEO00
    if (e_ok == Trans.trError_gg00) t01moveobj (kb, pObjBody, 1, ObjBodySize);
#   endif   
    
    kb04_ErrorHandling (Trans, ThisMessType, ThisMessType2, WriteVtrace); 
}


/*---------------------------------------------------------------------------*/

externCpp void
kb04IsObjHistoryUsed (tgg00_TransContext &Trans,
                      tgg00_FileId       &ObjFile,
                      tgg92_KernelOid     Oid,
                      bool               &bHistoryIsUsed)
{
    /* PTS 1106284 UH 07-04-2000 new */
    ROUTINE_DBG_MSP00 ("kb04IsObjHistoryUsed");
    if (e_ok != Trans.trError_gg00) return;

    const tgg00_MessType_Enum	ThisMessType  = m_history;
    const tgg00_MessType2_Enum	ThisMessType2 = mm_test;
    const bool                  WriteVtrace   = (cgg_is_true == g01vtrace.vtrOmsUpd_gg00);
    kb04_EnterKb (Trans, ThisMessType, ThisMessType2, WriteVtrace);
    if (e_ok != Trans.trError_gg00) return; // timeout !

    bd04IsObjHistoryUsed (Trans, ObjFile, Oid, bHistoryIsUsed);
    
    kb04_ErrorHandling (Trans, ThisMessType, ThisMessType2, WriteVtrace); 
}


/*----------------------------------------------------------------------------*/

externCpp void
kb04IsObjSelfLocked (tgg00_TransContext  &Trans,
                     tgg00_FileId        &ObjFile,
                     tgg92_KernelOid      Oid,
                     bool                &IsSelfLocked)
                     
                     /* PTS 1001745 UH 23-11-1998 new */
                     
{
    ROUTINE_DBG_MSP00 ("kb04IsObjSelfLocked");
    
    
    const tgg00_MessType_Enum	ThisMessType  = m_lock;
    const tgg00_MessType2_Enum	ThisMessType2 = mm_test;
    const bool                  WriteVtrace   = (cgg_is_true == g01vtrace.vtrOmsUpd_gg00);
    
    if (e_ok != Trans.trError_gg00) return;
    
    kb04_EnterKb (Trans, ThisMessType, ThisMessType2, WriteVtrace);
    
    if (e_ok != Trans.trError_gg00) return; // timeout !
    
    bd04IsObjSelfLocked (Trans, ObjFile, Oid, IsSelfLocked);
    
    kb04_ErrorHandling (Trans, ThisMessType, ThisMessType2, WriteVtrace); 
}

/*----------------------------------------------------------------------------*/

externCpp void
kb04LockObj (tgg00_TransContext	&Trans,
             tgg00_FileId	    &ObjFile,	   
             tgg92_KernelOid     Oid,
             tgg91_PageRef   	 WantedObjVers)
{ 
    ROUTINE_DBG_MEO00 ("kb04LockObj");
    
    
    const tgg00_MessType_Enum	ThisMessType  = m_lock;
    const tgg00_MessType2_Enum	ThisMessType2 = mm_object;
    const bool                  WriteVtrace   = (cgg_is_true == g01vtrace.vtrOmsUpd_gg00);
    
    if (e_ok != Trans.trError_gg00) return;
    
    kb04_EnterKb (Trans, ThisMessType, ThisMessType2, WriteVtrace);
    
    if (e_ok != Trans.trError_gg00) return; // timeout !
    
    bd04LockObj (Trans, ObjFile, Oid, WantedObjVers);
    
    /* PTS: 1001340, Alexander Kley 3.8.1998 */
    if (e_wrong_object_version == Trans.trError_gg00) 
    {
        Trans.trError_gg00 = e_object_dirty;
    }
    
    kb04_ErrorHandling (Trans, ThisMessType, ThisMessType2, WriteVtrace); 
}


/*----------------------------------------------------------------------------*/

externCpp void
kb04NewObj (tgg00_TransContext  &Trans,
            tgg00_FileId	    &ObjFile,
            tsp00_Int4		     KeyLen,
            void			    *Key,
            tgg92_KernelOid     &ResultOid,
            tgg91_PageRef       &ResultObjVers)
{ 
    ROUTINE_DBG_MEO00 ("kb04NewObj");
    
    const tgg00_MessType_Enum	ThisMessType  = m_new;
    const tgg00_MessType2_Enum	ThisMessType2 = mm_object;
    const bool                  WriteVtrace   = (cgg_is_true == g01vtrace.vtrOmsNew_gg00);
    
    tsp00_Int4  FrameLen = 0;
    
    if (e_ok != Trans.trError_gg00) return;
    
    kb04_EnterKb (Trans, ThisMessType, ThisMessType2, WriteVtrace);
    
    if (e_ok != Trans.trError_gg00) return; // timeout !
    
    bd04NewObj (Trans, ObjFile, FrameLen, KeyLen, Key, ResultOid, ResultObjVers);
    
    if (e_ok == Trans.trError_gg00 && KeyLen > 0)
    {
        tgg91_PageRef   ReUsedObjVers;
        tgg92_KernelOid	ReUsedOId;
        
        bd04NewObjKey (Trans, ObjFile, KeyLen, Key, ResultOid, ReUsedOId, ReUsedObjVers);
        
        if ((e_ok == Trans.trError_gg00) && ! ReUsedOId.gg92IsNilPno() )
		{
			// An object could be reused for the given key,
			// so we have to delete the newly created object.
			// The Key is not deleted, because the key's oid doesn't match.
			
			bd04DelObj (Trans, ObjFile, ResultOid, ResultObjVers);
			
			ResultOid     = ReUsedOId;
			ResultObjVers = ReUsedObjVers;
		}
		else if (e_duplicate_hash_key == Trans.trError_gg00) 
		{
			/* return OId of object which already belongs to the   */
			/* current key, PTS 1105663 Alexander Kley 29/02/2000  */                             
			ResultOid     = ReUsedOId;
            ResultObjVers = ReUsedObjVers;
		}
    }
    kb04_ErrorHandling (Trans, ThisMessType, ThisMessType2, WriteVtrace);
}

/*----------------------------------------------------------------------------*/

externCpp void
kb04NextObjConsistent (tgg00_TransContext   &Trans,
                       tgg00_FileId		    &ObjFile,	   
                       tgg92_KernelOid      &CurrOid,
                       bool                  bWithObject,
                       tsp00_Int4		     ObjBodySize,   //[inp=buffer size for object]
                       void				    *pObjBody,
                       //tsp00_Int4         &ObjBodyLen,    //[out=real object length]
                       tgg91_PageRef        &ResultObjVers)
{ 
    ROUTINE_DBG_MEO00 ("kb04NextObjConsistent");
    
    const tgg00_MessType_Enum	ThisMessType  = m_get_object;
    const tgg00_MessType2_Enum	ThisMessType2 = mm_qual;
    const bool                  WriteVtrace   = (cgg_is_true == g01vtrace.vtrOmsGet_gg00);
    tsp00_Int4                  ObjBodyLen;
    
    if (e_ok != Trans.trError_gg00) return;
    
    kb04_EnterKb (Trans, ThisMessType, ThisMessType2, WriteVtrace);
    
    if (e_ok != Trans.trError_gg00) return; // timeout !
    
	//???? kann weg, wenn LogReadAccess in der Parameterleiste steht
		tsp00_Int4  LogReadAccesses;

    bd04NextObjConsistent (Trans, ObjFile, CurrOid, bWithObject,
        ObjBodySize, pObjBody, ObjBodyLen, ResultObjVers, LogReadAccesses); //PTS 1107819 AK 12/09/2000
    
 //???? kann weg, wenn LogReadAccess in der Parameterleiste steht
        b21mp_obj_hist_read_count_put (Trans.trBdTcachePtr_gg00, LogReadAccesses);

#   if COMPILEMODE_MEO00 >= SLOW_MEO00
    if (e_ok == Trans.trError_gg00) t01moveobj (kb, pObjBody, 1, ObjBodySize);
#   endif   
    
    kb04_ErrorHandling (Trans, ThisMessType, ThisMessType2, WriteVtrace); 
}

/*----------------------------------------------------------------------------*/

externCpp void
kb04StartConsistentTrans (tgg00_TransContext &Trans)
{ 
    ROUTINE_DBG_MEO00 ("kb04StartConsistentTrans");
    
    const tgg00_MessType_Enum	ThisMessType  = m_begin;
    const tgg00_MessType2_Enum	ThisMessType2 = mm_trans;
    const bool                  WriteVtrace   = (cgg_is_true == g01vtrace.vtrOmsGet_gg00);
    
    if (e_ok != Trans.trError_gg00) return;
    
    kb04_EnterKb (Trans, ThisMessType, ThisMessType2, WriteVtrace);
    
    if (e_ok != Trans.trError_gg00) return; // timeout !
    
    kb51StartConsistTrans (Trans);
    
    kb04_ErrorHandling (Trans, ThisMessType, ThisMessType2, WriteVtrace); 
}


/*----------------------------------------------------------------------------*/

externCpp void
kb04UnlockObj (tgg00_TransContext   &Trans,
               tgg00_FileId         &ObjFile,	   
               tgg92_KernelOid       Oid)
{ 
    ROUTINE_DBG_MEO00 ("kb04UnlockObj");
    
    const tgg00_MessType_Enum	ThisMessType  = m_unlock;
    const tgg00_MessType2_Enum	ThisMessType2 = mm_object;
    const bool                  WriteVtrace   = (cgg_is_true == g01vtrace.vtrOmsUpd_gg00);
    
    if (e_ok != Trans.trError_gg00) return;
    
    kb04_EnterKb (Trans, ThisMessType, ThisMessType2, WriteVtrace);
    
    if (e_ok != Trans.trError_gg00) return; // timeout !
    
    bd04UnlockObj (Trans, ObjFile, Oid);
    
    kb04_ErrorHandling (Trans, ThisMessType, ThisMessType2, WriteVtrace); 
}

/*----------------------------------------------------------------------------*/

externCpp void
kb04UpdateObj (tgg00_TransContext	&Trans,
               tgg00_FileId			&ObjFile,
               tgg92_KernelOid  	 Oid,
               tgg91_PageRef    	&ObjVers,
               tsp00_Int4			 NewBodySize,
               void			        *pNewBody,
               tgg00_ObjColDesc		&ColDesc)	   
{ 
    ROUTINE_DBG_MEO00 ("kb04UpdateObj");
    
    const tgg00_MessType_Enum	ThisMessType  = m_update;
    const tgg00_MessType2_Enum	ThisMessType2 = mm_object;
    const bool                  WriteVtrace   = (cgg_is_true == g01vtrace.vtrOmsUpd_gg00);
    
    if (e_ok != Trans.trError_gg00) return;
    
    kb04_EnterKb (Trans, ThisMessType, ThisMessType2, WriteVtrace);
    
    if (e_ok != Trans.trError_gg00) return; // timeout !
    
    bd04UpdateObj (Trans, ObjFile, Oid, ObjVers, NewBodySize, pNewBody);
    
    kb04_ErrorHandling (Trans, ThisMessType, ThisMessType2, WriteVtrace); 
}
