/*!***************************************************************************

  module      : vkb531.cpp
  responsible : JuergenA, UweH
  special area: KB_object_rollback
  last changed: 2000-03-15
  copyright   : Copyright by SAP AG, 2000
  see also    :
  description : 



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include <string.h>

#include "ggg92.h"
#include "gkb08.h"
#include "gsp03_2.h"

#include "hbd01.h"
#include "hbd04.h"
#include "gbd94.h"
#include "hgg01.h"
#include "hgg01_1.h"  // g01optextmsg
#include "hkb50_1.h"
#include "hkb56_1.h"
#include "hkb85.h"
#include "hsp77.h"    // sp77sprintf

#if COMPILEMODE_MEO00 >= SLOW_MEO00	
#include "hta99.h"
#include "hta01.h"
#endif

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define IGNORE_ERROR true
#define ZERO_LEN     0

/*===========================================================================*
 *   LOCAL FUNCTIONS (PROTOTYPES)                                            *
 *===========================================================================*/

static void
kb531_BuildTreeId (const tgg92_KernelOid &oid,
                   tgg00_FileId          &FileId);


static void
kb531_GetObjectBody (tgg00_TransContext  &Trans,
                    tkb08_LogentryParam  &EntryParam,
                    tkb08_ObjUpdLogEntry &LogEntry,
                    tgg00_MessType        LogType,
                    tsp00_Int4            LogHeadLen,
                    tsp00_Int4            BodySize,
                    void                *pBody,
                    tsp00_Int4           &BodyLen);

static void
kb531_GetRollbObjEntry (tgg00_TransContext  &Trans,
                        tkb08_LogentryParam &EntryParam,
                        tgg00_LogPos        &CurrEntryPos,
                        tkb00_LogEntrySeq   &SkipUntilSeq,
                        bool                &SkipPartRollb);

static void
kb531_ObjEntryRollb (tgg00_TransContext	 &Trans,
                     bool				  IgnoreErr,
                     tgg00_StackListPtr	  StackAddr,
                     tsp00_Int2			  StackMax,
                     tkb08_LogentryParam &EntryParam);

/*===========================================================================*
 *  GLOBAL FUNCTIONS                                                         *
 *===========================================================================*/

// PASCAL interface
externC void
kb531ResetObjLock (/*[IN OUT]*/ tgg00_TransContext &Trans,
                   /*[IN]    */ tgg91_TransNo      &LockTransId,
                   /*[IN]    */ tgg00_TransIndex    LockTransIndex,
                   /*[IN]    */ tgg00_ObjFileNo     ObjFileNo,      // PTS 1105492 JA 2000-03-15
                   /*[IN]    */ tgg92_KernelOid    &Oid)
{ 
    ROUTINE_DBG_MSP00 ("kb531ResetObjLock");
    
    if (e_ok != Trans.trError_gg00) return;
    
	// PTS 1106143 AK 29/03/2000
    bd04ResetObjLock (Trans, LockTransId, LockTransIndex, Oid, ObjFileNo);
}


/*-----------------------------------------------------------------------------------*/

externC void
kb531PrintLogEntryParam (tkb08_LogentryParam &EntryParam,
                         tgg00_BasisError     TrError)
{
    /* PTS 1107529 UH 2000-09-13 new from  kb531_PrintObjLogEntryParam */

    tgg00_BasisError AuxErr = e_ok;
    char             Msg [sizeof(tsp00_C40)+1];

    sp77sprintf (Msg, sizeof(tsp00_C40), "Pos: %4s:%10d.%-10d        ",
        (ptData_egg00==EntryParam.lgeEntryLogPos_kb08.logPageType_gg00())?"Data":" Log",
        EntryParam.lgeEntryLogPos_kb08.logLpno_gg00(),
        EntryParam.lgeEntryLogPos_kb08.logPpos_gg00(),
        AuxErr);
    if ( AuxErr != e_ok ) return;
    g01optextmsg (sp3p_syscon, sp3m_error, kbMsg_csp03, csp3_n_log, Msg);
    g01opmsg (sp3p_syscon, sp3m_error, kbMsg_csp03, csp3_n_log,
          "BasisError              ", (tsp00_Int4)TrError);

    if ( NULL != EntryParam.lgeEntry_kb08  )
    {
        sp77sprintf (Msg, sizeof(tsp00_C40), "LogType=%1u,%1u                             ",
            (tsp00_Uint1)(EntryParam.lgeEntry_kb08->leType_kb08()),
            (tsp00_Uint1)(EntryParam.lgeEntry_kb08->leType2_kb08()),
            AuxErr);
        if ( AuxErr != e_ok ) return;
        g01optextmsg (sp3p_syscon, sp3m_error, kbMsg_csp03, csp3_n_log, Msg);

        if ( mm_object == EntryParam.lgeEntry_kb08->leType2_kb08() )
        {
            tkb08_ObjUpdLogEntry *LogEntry =
                REINTERPRET_CAST(tkb08_ObjUpdLogEntry*,EntryParam.lgeEntry_kb08);

            sp77sprintf (Msg, sizeof(tsp00_C40), "OID=%10d.%-10d               ",
                LogEntry->leOid_kb08.gg92GetPno(),
                LogEntry->leOid_kb08.gg92GetPos(),
                AuxErr);
            if ( AuxErr != e_ok ) return;
            g01optextmsg (sp3p_syscon, sp3m_error, kbMsg_csp03, csp3_n_log, Msg);

            sp77sprintf (Msg, sizeof(tsp00_C40), "HistoryRef=%10d.%-10d        ",
                LogEntry->leObjHistoryRef_kb08.gg91RefPno(),
                LogEntry->leObjHistoryRef_kb08.gg91RefPos(),
                AuxErr);
            if ( AuxErr != e_ok ) return;
            g01optextmsg (sp3p_syscon, sp3m_error, kbMsg_csp03, csp3_n_log, Msg);
        }
    }
}


/*-----------------------------------------------------------------------------------*/

// PASCAL interface
externC void
kb531RollbObjEntries (/*[IN OUT]*/ tgg00_TransContext    &Trans,
                      /*[IN]    */ tgg00_MessType_Param   MessType,
                      /*[IN]    */ tgg00_MessType2_Param  MessType2,
                      /*[MEMORY]*/ tkb08_LogentryParam	 &EntryParam, 
                      /*[MEMORY]*/ tgg00_StackListPtr	  WorkStackAddr,
                      /*[IN]    */ tsp00_Int2			  WorkStackMax,
                      /*[IN OUT]*/ tsp00_Int4			 &RollbackCnt)
{ 
    ROUTINE_DBG_MEO00 ("kb531RollbObjEntries");
    
    bool                SkipPartRollb;
    tkb00_LogEntrySeq   StopEntrySeq;
    tkb00_LogEntrySeq   SkipUntilSeq;
    tgg00_LogPos	    CurrEntryPos;
    tkb00_LogFileRef    RbFile;
   
    if (e_ok != Trans.trError_gg00) return;

    StopEntrySeq.gg90SetNil ();

    tkb00_LogEntrySeq DummyPrevEntrySeq;
    kb50GetObjPrevLogPos (Trans.trIndex_gg00, StopEntrySeq, DummyPrevEntrySeq, RbFile);

    if (! ((m_partial_rollback == MessType) && (mm_nil == MessType2)) )
        StopEntrySeq.gg90SetNil ();
    
#   if COMPILEMODE_MEO00 >= SLOW_MEO00
    ta01LogSeq (kb_log, "StopEntrySeq", StopEntrySeq);
#   endif
    
    if (NIL_PAGE_NO_GG00 == RbFile.lfrFileLastPos_kb00.logLpno_gg00())
    {
        return; /* nothing to rollback */
    }

    /* begin with rollback */

    kb85ParamLogEntryInit (EntryParam);
    
    EntryParam.lgeFile_kb08 = RbFile.lfrFileId_kb00;

    CurrEntryPos = RbFile.lfrFileLastPos_kb00;

    /* check initialized access handle */

    if  ( (ptData_egg00 != CurrEntryPos.logPageType_gg00()       ) ||
          (tfnLog_egg00 != EntryParam.lgeFile_kb08.fileTfn_gg00())  )
    {
        Trans.trError_gg00 = e_not_implemented;
        return;
    }
    
    SkipPartRollb = false;
    
    SkipUntilSeq.gg90SetNil (); 
    
    do
    {
        kb531_GetRollbObjEntry (Trans, EntryParam, CurrEntryPos,
            SkipUntilSeq, SkipPartRollb);
        if (e_ok != Trans.trError_gg00) break;
        
        if (m_partial_rollback == MessType)
        {
            if (mm_subtrans == MessType2)
            {
                if (Trans.trSubtransId_gg00	>
                    EntryParam.lgeEntry_kb08->lgHead_kb08().leSubtransId_kb08)
                {			
                    StopEntrySeq = EntryParam.lgeEntry_kb08->leSequence_kb08();
                    break;
                }
            }
            else // command rollback
            {
#               if COMPILEMODE_MEO00 >= SLOW_MEO00
                ta01LogSeq (kb_log, "leSequence  ", EntryParam.lgeEntry_kb08->leSequence_kb08());
#               endif
                
                if  (! StopEntrySeq.gg90IsNil()
                    &&  
                    StopEntrySeq >= EntryParam.lgeEntry_kb08->leSequence_kb08())
                {			
                    StopEntrySeq = EntryParam.lgeEntry_kb08->leSequence_kb08();
                    break;
                }
            }
        }
        
        if ( ! SkipPartRollb )
        {
            tgg00_MessType LogEntryType = EntryParam.lgeEntry_kb08->leType_kb08();
            
            if (m_delete == LogEntryType ||
                m_insert == LogEntryType || 
                m_new    == LogEntryType || 
                m_update == LogEntryType) 
            {
                RollbackCnt += 1;
                kb531_ObjEntryRollb (Trans, ! IGNORE_ERROR, WorkStackAddr, WorkStackMax, EntryParam);
                if (e_ok != Trans.trError_gg00) break;
            }
        }
        
    } while (NIL_PAGE_NO_GG00 != CurrEntryPos.logLpno_gg00());
    
    kb85LogPageRelease (Trans, EntryParam);
    
    if (e_ok != Trans.trError_gg00) return;

    if ((0 < RollbackCnt) && (m_partial_rollback == MessType))
    {
        kb56CmdRollbackLog (Trans, ptObjLog_egg00, StopEntrySeq);
    }
}


/*===========================================================================*
 *   LOCAL FUNCTIONS                                                         *
 *===========================================================================*/

static void
kb531_BuildTreeId (const tgg92_KernelOid &oid,
                   tgg00_ObjFileNo        ObjFileNo,
                   tgg00_FileId          &FileId)
                   
{
    ROUTINE_DBG_MEO00 ("kb531_BuildTreeId");
    
    FileId = b01niltree_id;
    FileId.fileObjFileType_gg00().becomes(oftUnknownType_egg00);
    FileId.fileObjFileNo_gg00() = ObjFileNo;
    
#   if COMPILEMODE_MEO00 >= SLOW_MEO00
    t01treeid      (kb_log, "FileId      ", FileId);
#   endif
}

/*-----------------------------------------------------------------------------------*/

static void
kb531_GetObjectBody (tgg00_TransContext	 &Trans,
                    tkb08_LogentryParam  &EntryParam,
                    tkb08_ObjUpdLogEntry &LogEntry,
                    tgg00_MessType        LogType,
                    tsp00_Int4            LogHeadLen,
                    tsp00_Int4            BodySize,
                    void                 *pBody,
                    tsp00_Int4           &BodyLen)
{
    ROUTINE_DBG_MEO00 ("kb531_GetObjectBody");
    
    const bool ReadAfterImage  = (m_delete != LogType && g01log.plogObjRestart_gg00);
    const bool ReadBeforeImage = (m_update == LogType || m_delete == LogType);
    
    tsp00_Int4 PosInEntry = LogHeadLen + 1;

    BodyLen = 0;
    
    if (ReadBeforeImage)
    {
        bd94History.bd94GetObjectBody (Trans,
                                       LogEntry.leOid_kb08, LogType,
                                       LogEntry.leObjHistoryRef_kb08,
                                       BodyLen, BodySize, pBody);

        if (e_ok != Trans.trError_gg00) return;

        if ( BodyLen != LogEntry.leBodyLenObj_kb08 - LogEntry.leKeyLenObj_kb08 )
        {
            Trans.trError_gg00 = e_log_error;
        }
    }
    
    if (e_ok != Trans.trError_gg00) return;

    if (ReadAfterImage)
    {
        Trans.trError_gg00 = e_not_implemented;
    }
    if (e_ok != Trans.trError_gg00) return;
}


/*-----------------------------------------------------------------------------------*/

static void
kb531_GetRollbObjEntry (tgg00_TransContext	&Trans,
                        tkb08_LogentryParam	&EntryParam,
                        tgg00_LogPos		&CurrEntryPos,
                        tkb00_LogEntrySeq	&SkipUntilSeq,
                        bool				&SkipPartRollb)
{ 
    // NOTE! Initialization before calling this function the first time:
    //          SkipUntilSeq.gg90SetNil()
    //          SkipPartRollb = false
    
    ROUTINE_DBG_MEO00 ("kb531_GetRollbObjEntry");
    
    if (e_ok != Trans.trError_gg00) return;
    
#   if COMPILEMODE_MEO00 >= SLOW_MEO00
    t01logpos (kb_log, "CurrEntryPos", CurrEntryPos);
#   endif
    
    kb85ReadLogEntry (Trans, CurrEntryPos, EntryParam);
    
    if (e_ok != Trans.trError_gg00) return;
    
    {
        tkb08_UnivLogEntry &CurrEntry = *(EntryParam.lgeEntry_kb08);
        
        if (SkipPartRollb && CurrEntry.leSequence_kb08() == SkipUntilSeq)
        {
            SkipPartRollb = false;
        }
        
        if  ( ! SkipPartRollb && m_partial_rollback == CurrEntry.leType_kb08() )
        {
            // skip rollbacked logs
            SkipPartRollb = true;
            SkipUntilSeq  = CurrEntry.lgCmdRollb_kb08().leStopEntrySeq_kb08;
            
#           if COMPILEMODE_MEO00 >= SLOW_MEO00
            ta01LogSeq (kb_log, "SkipUntilSeq", SkipUntilSeq);
#           endif
        }
        
        CurrEntryPos.logLpno_gg00() = CurrEntry.lgHead_kb08().leLpnoPrev_kb08;
        CurrEntryPos.logPpos_gg00() = CurrEntry.lgHead_kb08().lePposPrev_kb08;
    }
    
#   if COMPILEMODE_MEO00 >= SLOW_MEO00
    t01logpos      (kb_log, "cur_entrypos", CurrEntryPos);
    t01basis_error (kb_log, "get_rollb_en", Trans.trError_gg00);
#   endif
}


/*-----------------------------------------------------------------------------------*/

static void
kb531_ObjEntryRollb (tgg00_TransContext	 &Trans,
                     bool				  IgnoreErr,
                     tgg00_StackListPtr	  StackAddr,
                     tsp00_Int2			  StackMax,
                     tkb08_LogentryParam &EntryParam)
{
    ROUTINE_DBG_MEO00 ("kb531_ObjEntryRollb");
    
    const tgg00_MessType LogType = EntryParam.lgeEntry_kb08->leType_kb08();
    
    int                  LogHeadLen = 0;
    tkb08_ObjUpdLogEntry LogEntry;
    
    if (e_ok != Trans.trError_gg00) return;
    
    Trans.trState_gg00.addElement (tsRollback_egg00);
    
    switch (LogType)
    {
    case m_new:
    case m_delete:
    case m_insert:
    case m_update: 
        {
            LogHeadLen = sizeof (tkb08_ObjUpdLogEntry);
            break;
        }
    default:
        Trans.trError_gg00 = e_not_implemented;
    }
    
    if (e_ok != Trans.trError_gg00) return;
     
    kb85EntryHeadFromLogPage (Trans, EntryParam, LogHeadLen, LogEntry.leHead_kb08);
  
    if (e_ok != Trans.trError_gg00) return;
    
    switch (LogType)
    {
    case m_new:
        {
            tgg00_FileId  FileId;
            tkb00_PagePtr pNullBeforeObjBody = NULL;
            
            kb531_BuildTreeId (LogEntry.leOid_kb08, LogEntry.leObjFileNo_kb08, FileId);
            
            bd04RollbackObj (Trans, FileId, LogType,
                LogEntry.leOid_kb08, LogEntry.leObjPrevState_kb08,
                LogEntry.leObjPrevVersion_kb08, LogEntry.leObjPrevUpdTrans_kb08,
                ZERO_LEN, pNullBeforeObjBody);

            if ( e_ok != Trans.trError_gg00 )
            {
                kb531PrintLogEntryParam  (EntryParam, Trans.trError_gg00);
            }

            break;
        }
    case m_delete:
    case m_insert:
    case m_update: 
        {
            tsp00_Int4    BeforeObjBodyLen = 0;
            tkb00_Page    BeforeObjBody;
            tkb00_PagePtr pBeforeObjBody   = &BeforeObjBody;
        
            kb531_GetObjectBody (Trans, EntryParam,
                LogEntry, LogType, LogHeadLen,
                sizeof (BeforeObjBody), pBeforeObjBody, BeforeObjBodyLen);
            
            if ( e_ok == Trans.trError_gg00 )
            {
                tgg00_FileId  FileId;

                kb531_BuildTreeId (LogEntry.leOid_kb08, LogEntry.leObjFileNo_kb08, FileId);
            
                bd04RollbackObj (Trans, FileId, LogType,
                    LogEntry.leOid_kb08, LogEntry.leObjPrevState_kb08,
                    LogEntry.leObjPrevVersion_kb08, LogEntry.leObjPrevUpdTrans_kb08,
                    BeforeObjBodyLen, pBeforeObjBody);

                if ( e_ok != Trans.trError_gg00 )
                {
                    kb531PrintLogEntryParam  (EntryParam, Trans.trError_gg00);
                }
            }
            break;
        }
    default:
        Trans.trError_gg00 = e_not_implemented;
    }
    
    Trans.trState_gg00.delElement (tsRollback_egg00);
}



// end of progam
