/*!***************************************************************************

  module      : vkb67.cpp
  responsible : JuergenA, UweH
  special area: KB_object_handling_interface
  last changed: 2000-03-15
  copyright   : Copyright by SAP AG, 2000
  see also    :
  description : 



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include <string.h>

#include "hkb67.h"

#include "gkb08.h"
#include "gsp03.h"
#include "gsp03_2.h"

#include "gbd94.h"
#include "heo55k.h"
#include "hgg01.h"
#include "hgg01_1.h"
#include "hgg08.h"
#include "hgg10.h"
#include "hkb50.h"
#include "hkb51.h"
#include "hkb56_2.h"
#include "hsp77.h"    // sp77sprintf
#include "hkb85.h"

#if COMPILEMODE_MEO00 >= SLOW_MEO00	
#include "hta99.h"	
#include "hta01.h"	
#include "hta01_2.h"
#endif

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

// #define __CHECK_TIME       true

#define ALIGN_LOGENTRY_END true
#define INS_OBJ_MISSING    "InsObjLog missing       "
#define NIL_OBJ_FILLCHAR   0
#define WRONG_LOG_OID      "GetConsist:OID <> LogOID"

#ifdef __CHECK_TIME
externC void vclock (tsp_int4 *Sec, tsp_int4 *MicroSec);
#endif

#define CHANGE_ENTRY true

/*===========================================================================*
 *   LOCAL FUNCTIONS (PROTOTYPES)                                            *
 *===========================================================================*/

/*---------------------------------------------------------------------------*/

void kb67_GetPrevLogObj (tgg00_TransContext &Trans,	   
                         tgg92_KernelOid     Oid,
                         tgg91_PageRef       FirstBeforeRef,
                         bool                bWithObject,
                         tsp00_Int4          KeyLen,
                         tsp00_Int4         &ObjBodyLen,
                         tsp00_Int4          ObjBodySize,
                         void               *pObjBody,
                         tgg00_ObjFrameVers &FrameVers,
                         tgg91_PageRef	    &ResultObjVers,
                         tsp00_Int4         &LogReadAccesses);   // PTS 1107819 AK 12/09/2000

/*---------------------------------------------------------------------------*/

void kb67_LogEntryInsDelUpd (tgg00_TransContext	 &Trans,
                             tgg00_MessType_Enum  MessType,
                             tgg00_FileId		 &ObjFile,
                             tgg92_KernelOid      Oid,
                             tsp00_Int4           ObjKeySize,
                             tgg92_ObjHeader     &ObjHeader,
                             tsp00_Int4           ObjBodySize,
                             void                *pObjBody);

/*---------------------------------------------------------------------------*/

inline void kb67_PrintError (const char *s /*tsp00_C40*/)
{
    if (g01vtrace.vtrKbLock_gg00)
        g01optextmsg (sp3p_syscon, sp3m_info, kbMsg_csp03, csp3_n_history, s);
}


/*===========================================================================*
 *  GLOBAL FUNCTIONS                                                         *
 *===========================================================================*/

externCpp void
kb67DelUpdObj (tgg00_TransContext   &Trans,
               tgg00_MessType_Enum   WantedMessType,
               tgg00_FileId         &ObjFile,
               tgg92_KernelOid       Oid,
               tsp00_Int4            ObjKeySize,
               tgg91_PageRef         WantedObjVers,
               tgg92_ObjHeader      &ObjHeader,     
               tsp00_Int4            ObjBodySize,
               void                 *pObjBody)
{ 
    ROUTINE_DBG_MSP00 ("kb67DelUpdObj");
    
    tgg00_MessType MessType;
    if (e_ok != Trans.trError_gg00) return;
    
    if (m_update == WantedMessType)
    {
        if (obsOccupied_egg00 == ObjHeader.ohdState_gg92)
            MessType.becomes (m_update);
        else
            MessType.becomes (m_insert);   
    }
    else
        MessType.becomes (WantedMessType);
    
    kb50CheckObjSelfLocked (Trans, Oid, ObjHeader.ohdTransInfo_gg92);
    
    if (e_ok != Trans.trError_gg00) return;
    
    if ((m_insert != MessType) && (WantedObjVers != ObjHeader.ohdTransInfo_gg92.otrBeforeRef_gg00))
    {
        Trans.trError_gg00 = e_object_dirty;

        if (g01vtrace.vtrKbLock_gg00)
        {
            kb50XxxObjTrace (Trans, "  kb67DelUpd ",
                Oid,
                ObjHeader.ohdTransInfo_gg92.otrUpdTransId_gg00,
                ObjHeader.ohdTransInfo_gg92.otrBeforeRef_gg00,
                ObjHeader.ohdState_gg92);
        }
        return;
    }
    
    kb67_LogEntryInsDelUpd (Trans, MessType, ObjFile, Oid, ObjKeySize,
        ObjHeader, ObjBodySize, pObjBody);
}


/*-----------------------------------------------------------------------------------*/

externCpp void
kb67GetObjConsistent(tgg00_TransContext &Trans,
                     tgg92_KernelOid     Oid,
                     tgg00_ObjState      ObjState,
                     tgg00_ObjTransInfo	&ObjTransInfo,
                     bool                bWithObj,
                     tsp00_Int4          KeyLen,
                     tsp00_Int4         &ObjBodyLen,
                     tsp00_Int4			 ObjBodySize,
                     void				*pObjBody,
                     tgg00_ObjFrameVers &FrameVers,
                     tgg91_PageRef      &ResultObjVers,
					 tsp00_Int4         &LogReadAccesses)  //[out] PTS 1107819 AK 12/09/2000)
{ 
    ROUTINE_DBG_MEO00 ("kb67GetObjConsistent");
    
    tsp00_Bool bConsistent;  
    
    if (e_ok != Trans.trError_gg00) return;  
    
    kb50ObjConsistentCheck (Trans, ObjTransInfo.otrUpdTransId_gg00, bConsistent);
	//                      without otrBeforeRef:  PTS 1105492 JA 2000-03-15 
    
    if (e_ok != Trans.trError_gg00) return;  
    
    if (bConsistent)
    {
		LogReadAccesses = 0;                                                                       // PTS 1107819 AK 12/09/2000

        if (ObjState == obsFreeAfterEot_egg00)
        {
            // object deleted
            Trans.trError_gg00 = e_object_not_found;
            
            if (g01vtrace.vtrKbLock_gg00)
            {
                kb50XxxObjTrace (Trans, "  kb67GetCns",
                    Oid, ObjTransInfo.otrUpdTransId_gg00, ObjTransInfo.otrBeforeRef_gg00, ObjState);
            }
            return;
        } 
        ResultObjVers = ObjTransInfo.otrBeforeRef_gg00;
    }
    else
    {   
        kb67_GetPrevLogObj (Trans, Oid, ObjTransInfo.otrBeforeRef_gg00, bWithObj,
			KeyLen, ObjBodyLen, ObjBodySize, pObjBody, FrameVers, ResultObjVers, LogReadAccesses); // PTS 1107819 AK 12/09/2000
    }
}



/*---------------------------------------------------------------------------*/

externCpp bool
kb67IsObjHistoryUsed (tgg00_TransContext &Trans,
                      tgg92_KernelOid    &Oid,
                      tgg00_ObjTransInfo &ObjTransInfo)
{
    /* PTS 1106284 UH 07-04-2000 new */
    ROUTINE_DBG_MSP00 ("kb67IsObjHistoryUsed");
    if (e_ok != Trans.trError_gg00) return false;
    return TRUE == kb51IsObjTransUsed (Trans.trTaskId_gg00, Oid, ObjTransInfo);
}

/*-----------------------------------------------------------------------------------*/

externCpp void
kb67LockObj (tgg00_TransContext  &Trans,
             tgg00_FileId        &ObjFile,
             tgg92_KernelOid      Oid,
             tgg91_PageRef	      WantedObjVers,
             tgg00_ObjState_Enum  ObjState,
             tgg00_ObjTransInfo  &ObjTransInfo)
{ 
    // NOTE! if this function returns e_wait_for_lock_release then wait outside bd-locking 
    
    ROUTINE_DBG_MSP00 ("kb67LockObj");
    
    if (e_ok != Trans.trError_gg00) return;
    
    kb50LockObj (Trans, ObjFile.fileObjFileNo_gg00(), Oid, ObjTransInfo); // PTS 1105492 JA 2000-03-15
    
    if (e_ok != Trans.trError_gg00) return;
    
    if ( ! WantedObjVers.gg91IsDummyRef() ) 
    {
        if (WantedObjVers != ObjTransInfo.otrBeforeRef_gg00 && obsFreeAfterEot_egg00 != ObjState)
        {
            kb50UnlockObj (Trans.trTaskId_gg00, Trans.trTransId_gg00, Trans.trIndex_gg00,
                Oid, ObjTransInfo);
            
            Trans.trError_gg00 = e_object_dirty;

            if (g01vtrace.vtrKbLock_gg00)
            {
                kb50XxxObjTrace (Trans, "  kb67LckObj",
                    Oid, ObjTransInfo.otrUpdTransId_gg00, ObjTransInfo.otrBeforeRef_gg00, ObjState);
            }
            return;
        }
    }
    
#   if COMPILEMODE_MEO00 >= SLOW_MEO00
    t01page_ref    (kb, "WntedObjVers", WantedObjVers);
    t01page_ref    (kb, "otrBeforeRef", ObjTransInfo.otrBeforeRef_gg00);
    t01basis_error (kb, "trError     ", Trans.trError_gg00);
#   endif 
}


/*-----------------------------------------------------------------------------------*/

externCpp void
kb67NewObj (tgg00_TransContext &Trans,
            tgg00_FileId	   &ObjFile,
            tgg92_KernelOid     Oid,
            tgg92_ObjHeader    &ObjHeader,     
            tsp00_Int4		    ObjBodySize,
            tsp00_Int4		    ObjKeySize)
{   
    // NOTE! if this function returns e_wait_for_lock_release then wait outside bd-locking 
    
    ROUTINE_DBG_MEO00 ("kb67NewObj");
    
    tgg00_ObjState_Enum  WantedObjState = (ObjKeySize == 0) ? obsReserved_egg00 : obsKeyReserved_egg00;
    tgg00_MessType2      EntryPart;
    
    tkb00_LogFileRef     LogFile;
    tkb00_LogFileRefPtr  LogFilePtr = &LogFile;
    
    EntryPart.becomes (mm_nil);
    
    if (e_ok != Trans.trError_gg00) return;
    
    
    if ((0 == ObjKeySize) || (obsFree_egg00 == ObjHeader.ohdState_gg92))
    {
        ObjHeader.ohdTransInfo_gg92.otrBeforeRef_gg00.gg91SetNilRef();
        ObjHeader.ohdTransInfo_gg92.otrLockTransIndex_gg00 = cgg_nil_transindex;
        ObjHeader.ohdTransInfo_gg92.otrLockTransId_gg00.gg90SetNil();
        ObjHeader.ohdTransInfo_gg92.otrUpdTransId_gg00.gg90SetNil();   
    }
    
    kb67LockObj (Trans, ObjFile, Oid, ObjHeader.ohdTransInfo_gg92.otrBeforeRef_gg00,
        WantedObjState, ObjHeader.ohdTransInfo_gg92);    
    
    if (e_ok != Trans.trError_gg00) return;
    
    if ((obsKeyReserved_egg00 == ObjHeader.ohdState_gg92) ||
        (obsOccupied_egg00    == ObjHeader.ohdState_gg92)  )
    {
        Trans.trError_gg00 = e_duplicate_hash_key;
        
        if (g01vtrace.vtrKbLock_gg00)
        {
            kb50XxxObjTrace (Trans, "  kb67NewObj",
                Oid,
                ObjHeader.ohdTransInfo_gg92.otrUpdTransId_gg00,
                ObjHeader.ohdTransInfo_gg92.otrBeforeRef_gg00,
                ObjHeader.ohdState_gg92);
        }
        return;
    }
    
    kb67_LogEntryInsDelUpd (Trans, m_new, ObjFile, Oid, ObjKeySize,
        ObjHeader, ObjBodySize, NULL);  
}


/*---------------------------------------------------------------------------*/

externCpp void
kb67ResetObjLock (tgg00_TransContext &Trans,
                  tgg91_TransNo       LockTransId,
                  tgg00_TransIndex    LockTransIndex,
                  tgg92_KernelOid     Oid,
                  tgg00_ObjTransInfo &ObjTransInfo)
{ 
    ROUTINE_DBG_MSP00 ("kb67ResetObjLock");
    
    if (e_ok != Trans.trError_gg00) return;    
    
    kb50TryCollisionEntryRemove (Trans.trTaskId_gg00, LockTransId, LockTransIndex, Oid, ObjTransInfo);
}


/*===========================================================================*
 *  LOCAL FUNCTIONS (CODE)                                                   *
 *===========================================================================*/

void kb67_GetPrevLogObj (tgg00_TransContext &Trans,	   
                         tgg92_KernelOid     Oid,
                         tgg91_PageRef       FirstBeforeRef,
                         bool                bWithObject,
                         tsp00_Int4          KeyLen,
                         tsp00_Int4         &ObjBodyLen,
                         tsp00_Int4          ObjBodySize,
                         void               *pObjBody,
                         tgg00_ObjFrameVers &FrameVers,
                         tgg91_PageRef	    &ResultObjVers,
                         tsp00_Int4         &LogReadAccesses)  // [out] PTS 1107819 AK 12/09/2000
{
    ROUTINE_DBG_MEO00 ("kb67_GetPrevLogObj");
    
#   ifdef __CHECK_TIME
    tsp00_Int4 StartSec;
    tsp00_Int4 StartUsec;
    
    vclock (&StartSec, &StartUsec);
#   endif
    
    bool           bObjFound     = false;
    tgg91_PageRef  CurrEntryRef;
    tgg91_PageRef  NextPageRef;
    tgg91_TransNo  ConsistentView;
    tgg91_TransNo  PrevUpdTrans;
    tgg00_MessType Action;

    LogReadAccesses = 0;                                       // PTS 1107819 AK 12/09/2000
    CurrEntryRef    = FirstBeforeRef;
    
    if (Trans.trError_gg00 != e_ok) return;
    
#   if COMPILEMODE_MEO00 >= SLOW_MEO00	
    ta01Oid     (kb, "Oid         ", Oid);
    ta01TransNo (kb, "trCnsistVers", Trans.trConsistView_gg00);
#   endif
    
    do
    {
        if ( NIL_PAGE_NO_GG00 == CurrEntryRef.gg91RefPno() )
        {
            Trans.trError_gg00 = e_object_not_found;
            
            if (g01vtrace.vtrKbLock_gg00)
            {
                kb50XxxObjTrace (Trans, "  kb67_Get1 ",
                    Oid,
                    Trans.trTransId_gg00, // dummy param: UpdTransId not available
                    CurrEntryRef,
                    obsFree_egg00);       // dummy param: ObjState not available

            kb67_PrintError ("ObjSeq is nil                           ");
            }
        }
        else
        {
            if ( bWithObject )
            {
                bd94History.bd94GetHistory (Trans, Oid, ! CHANGE_ENTRY, CurrEntryRef,
                                            NextPageRef, Action,
                                            ConsistentView, PrevUpdTrans, FrameVers,
                                            ObjBodyLen, ObjBodySize - KeyLen, pObjBody);
            }
            else
            {
                bd94History.bd94GetHistory (Trans, Oid, ! CHANGE_ENTRY, CurrEntryRef,
                                            NextPageRef, Action,
                                            ConsistentView, PrevUpdTrans, FrameVers,
                                            ObjBodyLen);
            }
            ++LogReadAccesses;
        }
        
        if ( (e_ok == Trans.trError_gg00) && (ConsistentView == Trans.trConsistView_gg00) )
        {
            /* UH 14-03-2000 */
            /* It is forbidden to update an object within an explicite view.     */
            /* In other words, if a transaction has updated an object by itself, */
            /* then always the objects itself is consistent it.                  */
            /* The current ConsistentView cannot be found in an history entry !! */

            g01abort (kbMsg_csp03, csp3_n_history, "Invalid Upd in ExpliView", CurrEntryRef.gg91RefPno());
        }
        
        if ((e_ok == Trans.trError_gg00) && ! bObjFound)     
        {
            CurrEntryRef = NextPageRef;
            
            // check before image
            
            pasbool bConsistent;  
            
            kb50ObjConsistentCheck (Trans, PrevUpdTrans, bConsistent);
			//                      without NextPageRef:  PTS 1105492 JA 2000-03-15 
            
            if (bConsistent && e_ok == Trans.trError_gg00)
            {
                // before image is consistent
                
                if (m_new == Action)
                {
                    // there is no before image
                    Trans.trError_gg00 = e_object_not_found;
                    
                    if (g01vtrace.vtrKbLock_gg00)
                    {
                        kb50XxxObjTrace (Trans, "  kb67_Get4 ",
                            Oid,
                            PrevUpdTrans,
                            CurrEntryRef,
                            obsReserved_egg00);   // Reserved instead of m_new
                    }
                }
                else
                {
                    ResultObjVers = NextPageRef;
                    bObjFound     = true;
                }
            } 
            else // NOT bConsistent
            {
                if  ( e_ok == Trans.trError_gg00 && g01vtrace.vtrKbLock_gg00 ) 
                {
                    kb50XxxObjTrace (Trans, "  kb67GetPrv",
                        Oid,
                        PrevUpdTrans,
                        NextPageRef,
                        obsUpdBeforeImage_egg00);
                }
            }
        }
    } while (! bObjFound && (e_ok == Trans.trError_gg00) );
    
    if ( e_ok == Trans.trError_gg00 )
    {
        if ( ! bObjFound )
            Trans.trError_gg00 = e_object_not_found;
    }
    else if ( e_object_not_found != Trans.trError_gg00 )
    {
        char Msg[sizeof(tsp00_C40)+1];
        sp77sprintf (Msg, sizeof(tsp00_C40), "NoHist: %-10d.%10d %10d",
                     CurrEntryRef.gg91RefPno(), CurrEntryRef.gg91RefPos(),
                     Trans.trTaskId_gg00);
        kb67_PrintError (Msg);
        sp77sprintf (Msg, sizeof(tsp00_C40), "   OID: %-10d.%10d %10d",
                     Oid.gg92GetPno(), Oid.gg92GetPos(),
                     Trans.trError_gg00);
        kb67_PrintError (Msg);
        sp77sprintf (Msg, sizeof(tsp00_C40), "CONSVIEW: %d", Trans.trConsistView_gg00.gg90GetInt4());
        kb67_PrintError (Msg);
        Trans.trError_gg00 = e_obj_history_not_found;
    }
    
#   ifdef __CHECK_TIME
    g01time_vtrace (Trans, "OBJLOG CONS READ  ", StartSec, StartUsec);
#   endif
}

/*---------------------------------------------------------------------------*/

void kb67_LogEntryInsDelUpd (tgg00_TransContext	 &Trans,
                             tgg00_MessType_Enum  MessType,
                             tgg00_FileId		 &ObjFile,
                             tgg92_KernelOid      Oid,
                             tsp00_Int4           ObjKeySize,
                             tgg92_ObjHeader     &ObjHeader,
                             tsp00_Int4           ObjBodySize,
                             void                *pObjBody)
{
    ROUTINE_DBG_MSP00 ("kb67_LogEntryInsDelUpd");
    
    if ( e_ok != Trans.trError_gg00 ) return;

#   ifdef __CHECK_TIME
    tsp00_Int4 StartSec;
    tsp00_Int4 StartUsec;
    
    vclock (&StartSec, &StartUsec);
#   endif
    
          tsp00_Int4     ObjSize           = ObjBodySize - ObjKeySize;
    const bool           bWriteBeforeImage = ((MessType == m_delete) || (MessType == m_update));
    tkb08_ObjUpdLogEntry LogEntry;
    
    tkb08_LogEntryHeader &LHead = LogEntry.leHead_kb08;
    
    LHead.leLen_kb08 = sizeof(LogEntry);
    
    LHead.leType_kb08.becomes  ( MessType);
    LHead.leType2_kb08.becomes ( mm_object);
    LHead.leSubtransId_kb08    = Trans.trSubtransId_gg00;
    LHead.leTransId_kb08       = Trans.trTransId_gg00;
    
    LogEntry.leOid_kb08             = Oid;   
    LogEntry.leObjConsistView_kb08  = Trans.trConsistView_gg00;
    LogEntry.leObjPrevUpdTrans_kb08 = ObjHeader.ohdTransInfo_gg92.otrUpdTransId_gg00;
    LogEntry.leObjPrevState_kb08    = ObjHeader.ohdState_gg92;
    /*LogEntry.leFillObjUpd_kb08      = false;*/
    LogEntry.leBodyLenObj_kb08      = ObjBodySize; // ObjFrameSize - sizeof(pObjFrame->objHeader_gg92);
    LogEntry.leKeyLenObj_kb08       = ObjKeySize;
    LogEntry.leObjFileNo_kb08       = ObjFile.fileObjFileNo_gg00();
    LogEntry.leObjPrevVersion_kb08  = ObjHeader.ohdTransInfo_gg92.otrBeforeRef_gg00;
    
    if ( ! bWriteBeforeImage ) ObjSize = 0;

    bd94History.bd94InsertImage (Trans,
                                 Oid, ObjFile.fileObjFileNo_gg00(),
                                 LogEntry.leHead_kb08.leType_kb08,
                                 Trans.trConsistView_gg00,
                                 ObjHeader.ohdTransInfo_gg92.otrUpdTransId_gg00,
                                 LogEntry.leObjPrevVersion_kb08,
                                 LogEntry.leObjHistoryRef_kb08,
                                 ObjSize,
                                 pObjBody);

#   if COMPILEMODE_MEO00 >= SLOW_MEO00	
    ta01Oid     (kb_log, "leOid       ", LogEntry.leOid_kb08);
    ta01TransNo (kb_log, "leObjConsist", LogEntry.leObjConsistView_kb08);
    ta01TransNo (kb_log, "leObjPrevUpd", LogEntry.leObjPrevUpdTrans_kb08);
    t01page_ref (kb_log, "leObjPrevVer", LogEntry.leObjPrevVersion_kb08);
    t01page_ref (kb_log, "leObjHistory", LogEntry.leObjHistoryRef_kb08);
#   endif 
    
    tkb00_LogFileRef RbFile;
    tgg00_LogPos     FirstLogPos;
    tgg00_MessType2  EntryPart;
				
    kb56LogObjEntryHead (Trans, LogEntry.leHead_kb08,
        sizeof (LogEntry), FirstLogPos, EntryPart, RbFile);
    
    if (e_ok == Trans.trError_gg00)
    {
        ObjHeader.ohdTransInfo_gg92.otrUpdTransId_gg00 = Trans.trTransId_gg00;
        ObjHeader.ohdTransInfo_gg92.otrBeforeRef_gg00  = LogEntry.leObjHistoryRef_kb08;
    }
    
#   ifdef __CHECK_TIME
    g01time_vtrace (Trans, "OBJLOG WRITE      ", StartSec, StartUsec);
#   endif
}

// end of program
                             
 
