rem @(#) $Id: //sapdb/V73/c_00/b_48/sys/src/lctools/lcinit.bat#1 $ SAP
@echo off
setlocal

rem **********************************************************
rem *         Database  Initialization                       *
rem **********************************************************
rem *                                                        *
rem *         lcinit <lc_name> [ <parameter> [debug]]        *
rem *                          [ -uDBM <user>,<pwd> ]        *
rem *                          [ -uDBA <user>,<pwd> ]        *
rem *                          [ -uSQL <user>,<pwd> ]        *
rem *                          [ -ud   <pwd> ]               *
rem *                                                        *
rem *                                                        *
rem *                                                        *
rem *   lc_name   : database name                            *
rem *                                                        *
rem *   parameter : additional parameter                     *
rem *                                                        *
rem *               init                                     *
rem *               register                                 *
rem *               restart                                  *
rem *               slow                                     *
rem *               stop                                     *
rem *               shutdown                                 *
rem *                                                        *
rem *   userflags :                                          *
rem *                                                        *
rem *               -uDBM : DBM User (default : control)     *
rem *               -uDBA : DBA User (default : superdba)    *
rem *               -uSQL : SQL User (default : sapr3)       *
rem *               -ud   : Domain pwd (default : domain)    *
rem *                                                        *
rem **********************************************************

rem --------------------- input parameters -------------------------------
set LC=%1
set mode=restart
set debug=opt

if "%LC%" == "" (
  echo.
  echo USAGE : LCINIT db_name [parameter] [userflags]                                 
  echo.
  echo         parameter :
  echo                      init     [debug]
  echo                      restart  [debug]
  echo                      register [debug]
  echo                      slow     [debug]
  echo                      shutdown
  echo                      stop
  echo         userflags :
  echo                      -uDBM user,pwd
  echo                      -uDBA user,pwd
  echo                      -uSQL user,pwd
  echo                      -ud   pwd
  echo.
  goto end_err
)

rem --------------------- internal user/pwd ------------------------------
set db_user=control
set db_pwd=control
set dba_user=superdba
set dba_pwd=admin
set sap_user=sapr3
set sap_pwd=sap
set domain_pwd=domain

rem --------------------- check input parameters -------------------------
:check_input
if "%2" == "" goto end_input
if "%2" == "-uDBM" (
  set db_user=%3
  set db_pwd=%4
  shift
  shift  
) else if "%2" == "-uDBA" (
  set dba_user=%3
  set dba_pwd=%4
  shift
  shift  
) else if "%2" == "-uSQL" (
  set sap_user=%3
  set sap_pwd=%4
  shift
  shift  
) else if "%2" == "-ud" (
  set domain_pwd=%3
  shift  
) else if "%2" == "debug" (
  set debug=%2
) else (
  set mode=%2
)
shift
goto check_input
:end_input

rem echo mode=%mode%
rem echo debug=%debug%
rem echo db_user=%db_user%
rem echo db_pwd=%db_pwd%
rem echo dba_user=%dba_user%
rem echo dba_pwd=%dba_pwd%
rem echo sap_user=%sap_user%
rem echo sap_pwd=%sap_pwd%
rem echo domain_pwd=%domain_pwd%

rem ------------------------ DBMCLI --------------------------------------
set DBMCLI=dbmcli -d %LC% -u %db_user%,%db_pwd%

rem --------------------- internal variables -----------------------------
set state=NONE
set vers=0
set dbtype=LVC
set errflag=0
set donetxt=restarted

rem --------------------- check environment ------------------------------

%DBMCLI% show state > adatmp
if errorlevel 1 (
  echo.
  type adatmp
  echo ERROR on %LC% 
  echo.
  goto end_err
)

rem --------------------- get installation directory ---------------------

%DBMCLI% dbm_version INSTROOT > adatmp

if errorlevel 1 (
  rem use DBROOT
  set INSTROOT=%DBROOT%
  goto check_env
) 

rem build a command script to set INSTROOT

echo @echo off > instroot.bat
echo (set INSTROOT=^^>> instroot.bat
findstr -v "OK" adatmp >> instroot.bat
echo ) >> instroot.bat
call instroot
rem echo INSTROOT=%INSTROOT%

:check_env

if not exist %INSTROOT%\bin (
  echo.
  echo ERROR : %INSTROOT% does not exist
  echo.
  goto end_err
)

rem ------------------------ REGISTER tools  --------------------------------------

set REGCOM=%INSTROOT%\bin\xregcomp -c
set REGFCT=%INSTROOT%\sap\regLCCom -d %LC% -u %sap_user%,%sap_pwd% -f
set CHKPT=%INSTROOT%\sap\regLCCom -d %LC% -u %dba_user%,%dba_pwd% -w

rem --------------------- check database instance ------------------------

set dbtype=DB
%DBMCLI% param_directget INSTANCE_TYPE > adatmp

if errorlevel 1 (
  rem INSTANCE_TYPE is unknown, check OBJ_LOG_SIZE

  %DBMCLI% param_directget OBJ_LOG_SIZE > adatmp

  if not errorlevel 1 (
    rem check if OBJ_LOG_SIZE = 0
    findstr "\<0\>" adatmp > adatmp2
    if errorlevel 1 (
      rem liveCache, OBJ_LOG_SIZE <> 0
      set dbtype=LVC
    )
  )
) else (
  findstr "LVC" adatmp > adatmp2
  if not errorlevel 1 (
    rem liveCache, INSTANCE_TYPE is LVC
    set dbtype=LVC
  )
)

rem echo %dbtype%

if "%dbtype%" == "LVC" (
  set dbtext1=LIVECACHE
  set dbtext2=liveCache
)
if "%dbtype%" == "DB" (
  set dbtext1=DATABASE
  set dbtext2=Database
)

set logfile=%INSTROOT%\lcinit.log
set hisfile=%INSTROOT%\lcinit.his
set inst_path=%INSTROOT%\sap

rem --------------------- print out the header  --------------------------

echo.
echo %dbtext1% %LC% (%mode%)
echo ************************ START ***************************** > %logfile%
echo %dbtext1% %LC% (%mode%) >> %logfile%
date /t  >> %logfile%
time /t  >> %logfile%
echo ------------------------------------------------------------ >> %logfile%
echo installation path = %INSTROOT% >> %logfile%
echo.

rem --------------------- check state of the database --------------------

%DBMCLI% show state > adatmp
%DBMCLI% show state >> %logfile%

findstr "WARM ONLINE" adatmp > adatmp2
if not errorlevel 1 (

  set state=ONLINE

) else (

  findstr "COLD ADMIN" adatmp > adatmp2
  if not errorlevel 1 (

    set state=COLD

  ) else (

    set state=OFFLINE
    echo The database state is OFFLINE

    if "%mode%" == "register" (
      goto reg_COM
    ) else if "%mode%" == "REGISTER" (
      goto reg_COM
    ) else if "%mode%" == "stop"     (
      goto nothing_todo
    ) else if "%mode%" == "STOP"     (
      goto nothing_todo
    ) else if "%mode%" == "shutdown" (
      goto nothing_todo
    ) else if "%mode%" == "SHUTDOWN" (
      goto nothing_todo

    ) else if "%mode%" == "slow" (

      rem --------------- start DB into COLD mode (SLOW) -----------------

      echo starting %LC% SLOW KERNEL into COLD
      echo starting %LC% SLOW KERNEL into COLD >> %logfile%
      %DBMCLI% db_cold -slow >> %logfile%
      if errorlevel 1 (
        goto err
      ) else (
        set state=COLD
      )
    ) else (

      rem --------------- start DB into COLD mode ------------------------

      echo starting %LC% into COLD
      echo starting %LC% into COLD >> %logfile%
      %DBMCLI% db_cold >> %logfile%
      if errorlevel 1 (
        goto err
      ) else (
        set state=COLD
      )
    )
  )
)

echo The database state is %state%

rem ------------------- check version of the database --------------------

%DBMCLI% show version > adatmp
%DBMCLI% show version >> %logfile%
type adatmp
echo.

findstr "7\.1" adatmp > adatmp2
if not errorlevel 1 (
  set vers=71x
) else (
  findstr "7\.2\.1" adatmp > adatmp2
  if not errorlevel 1 (
    set vers=721
  ) else (
    findstr "7\.2\.2" adatmp > adatmp2
    if not errorlevel 1 (
      set vers=722
    ) else (
      findstr "7\.2\.3" adatmp > adatmp2
      if not errorlevel 1 (
        set vers=723
      ) else (
        findstr "7\.4\." adatmp > adatmp2
        if not errorlevel 1 (
          set vers=74+
        ) else (
          set vers=724+
        )
      )
    )
  )
)

rem echo %vers%

rem -------------------- call old version of lcinit ----------------------

if not "%vers%" == "724+" (
  if not "%vers%" == "74+" (
    echo calling %INSTROOT%\sap\LC723 .... >> %logfile%
    call %INSTROOT%\sap\lc723 %LC% %mode% %debug%
    goto end
  )
)

rem -------------------- parameter switch --------------------------------

if "%mode%" == "init"     goto init_lc
if "%mode%" == "INIT"     goto init_lc
if "%mode%" == "restart"  goto restart_lc
if "%mode%" == "RESTART"  goto restart_lc
if "%mode%" == "stop"     goto stop_lc
if "%mode%" == "STOP"     goto stop_lc
if "%mode%" == "shutdown" goto shutdown_lc
if "%mode%" == "SHUTDOWN" goto shutdown_lc
if "%mode%" == "register" goto reg_COM
if "%mode%" == "REGISTER" goto reg_COM
if "%mode%" == "slow"     goto restart_lc
if "%mode%" == "SLOW"     goto restart_lc

echo.
echo ERROR : parameter '%mode%' is not allowed!
echo.
goto end_err

rem ------------------- restart of the database --------------------------
:restart_lc

set donetxt=restarted

if "%state%" == "ONLINE" (

  rem DB is already online
  goto nothing_todo

) else (

  echo starting %LC% into ONLINE
  echo starting %LC% into ONLINE >> %logfile%
  %DBMCLI% db_warm >> %logfile%

  if errorlevel 1 (
    echo.
    echo ERROR : restart not possible [please check knldiag!!]
    echo ERROR : restart not possible [please check knldiag] >> %logfile%
    echo.
    goto err
  )
)

goto loading_data


rem ---------------------- stop the database -----------------------------
:stop_lc

set donetxt=stopped

if "%state%" == "OFFLINE" (
  rem DB is already offline
  goto nothing_todo

) else (

  echo stopping %LC% 
  echo stopping %LC% >> %logfile%
  %DBMCLI% db_stop >> %logfile%

  if errorlevel 1 (
    goto err
  ) else (
    goto ok
  )
)

rem -------------------- shutdown the database ---------------------------
:shutdown_lc

set donetxt=stopped

if "%state%" == "OFFLINE" (
  rem DB is already offline
  goto nothing_todo

) else (

  echo stopping %LC% 
  echo stopping %LC% >> %logfile%
  %DBMCLI% db_offline >> %logfile%

  if errorlevel 1 (
    goto err
  ) else (
    goto ok
  )
)

rem -------------------- register COM routines ---------------------------
:reg_COM

set donetxt=registered

if not "%state%" == "ONLINE" (
  rem LC is not online so restart it
  call %INSTROOT%\sap\lcinit %LC% restart
  goto end
)

goto reg_COM0

rem --------------------   nothing TODO ----------------------------------
:nothing_todo

echo.
echo %dbtext2% %LC% is already %state%
echo %dbtext2% %LC% is already %state% >> %logfile%
goto history

rem ======================================================================
rem ======================== init of the database ========================
rem ======================================================================
:init_lc

set donetxt=initialized

rem ---------------------- DCOM registration -----------------------------

echo.
echo %dbtext2% DCOM registration
echo %dbtext2% DCOM registration >> %logfile%

%REGCOM% %INSTROOT%\pgm\dbpinstall

echo.

rem -------------------- start DB into COLD mode -------------------------

echo starting %LC% into COLD
echo starting %LC% into COLD >> %logfile%
%DBMCLI% db_stop >> %logfile%
if errorlevel 1 goto err
%DBMCLI% db_cold >> %logfile%
if errorlevel 1 goto err

rem ----------------------- INIT CONFIG ----------------------------------

echo starting INIT CONFIG
echo starting INIT CONFIG >> %logfile%
%DBMCLI% -uUTL -c util_execute init config >> %logfile%
if errorlevel 1 goto err

rem --------------------- ACTIVATE SERVERDB ------------------------------

echo starting ACTIVATE SERVERDB
echo starting ACTIVATE SERVERDB >> %logfile%
%DBMCLI% -uUTL -c util_activate %dba_user%,%dba_pwd% >> %logfile%
if errorlevel 1 goto err

rem --------------------- loading system tables --------------------------

echo loading SYSTEM TABLES
echo loading SYSTEM TABLES >> %logfile%
%DBMCLI% -c load_systab -ud %domain_pwd% >> %logfile%
if errorlevel 1 goto err

rem --------------------- creating SAP user ------------------------------

echo CREATING user %sap_user%
echo CREATING user %sap_user% >> %logfile%
%DBMCLI% -uSQL %dba_user%,%dba_pwd% sql_execute create user %sap_user% password %sap_pwd% dba not exclusive >> %logfile%
if errorlevel 1 goto err

rem --------------------- restarting the database ------------------------

echo restarting %LC%
echo restarting %LC% >> %logfile%
%DBMCLI% db_offline >> %logfile%
if errorlevel 1 goto err
%DBMCLI% db_warm >> %logfile%
if errorlevel 1 goto err


rem ------ enable write transaction (only liveCache >= 7.2.3 b04) --------

if "%dbtype%" == "LVC" (
  if "%vers%" == "724+" (
    if not "%vers%" == "74+" (
      echo.
      echo ENABLE WRITE TRANSACTION
      echo ENABLE WRITE TRANSACTION >> %logfile%
      %DBMCLI% -uSQL %dba_user%,%dba_pwd% sql_execute enable write transaction >> %logfile%
      if errorlevel 1 goto err
    )
  )
)

rem ======================================================================
rem =================== END : init of the database =======================
rem ======================================================================


:loading_data

rem ----------------------- set monitoring -------------------------------

echo.
echo SET MONITOR ON
echo SET MONITOR ON >> %logfile%
%DBMCLI% -uSQL %sap_user%,%sap_pwd% sql_execute monitor on >> %logfile%
if errorlevel 1 goto err

rem ---------------------- DCOM registration -----------------------------
:reg_COM0

echo.
echo %dbtext2% DCOM registration
echo %dbtext2% DCOM registration >> %logfile%

%REGCOM% %INSTROOT%\pgm\dbpinstall

rem ----------------- creating system COM routines -----------------------

echo.
echo creating system COM routines
echo creating system COM routines >> %logfile%
%REGFCT% %inst_path%\dbpinstall.lst >> %logfile%
if errorlevel 1 goto err

echo.

rem ---------------------------------------------------------------------
rem --------------- definition of the application COM routines ----------
rem ---------------------------------------------------------------------
rem       sapapo     : APO                 (Reusch)
rem       sapts      : APO                 (Kilgus)
rem       sapatp     : ATP                 (Leffers)
rem       sphsapdb   : BW                  (Pfeiffer)
rem       dbpSearch  : DrFuzzy/Requisite   (Elvers)
rem       saprpm     : Automotive          (Haertig)
rem       saprep     : Automotive          (Toma)
rem       sapsea     : Automotive          (Haertig)
rem ---------------------------------------------------------------------

set com_list= sapapo sapatp saprpm sapts sphsapdb dbpSearch saprep sapsea
set com_flag=

if "%debug%" == "debug" set com_flag=d

rem ------------------ register the application COM routines ------------

for %%f in ( %com_list% ) do ( 
  if exist %inst_path%\%%f%com_flag%.dll (
    echo registration of %%f%com_flag%.dll
    echo registration of %%f%com_flag%.dll >> %logfile% 
    %REGCOM% %inst_path%\%%f%com_flag%     >> %logfile% 
    if errorlevel 1 goto err
  )
)

echo.

rem --------------------- create the application COM routines -----------

for %%f in ( %com_list% ) do ( 
  if exist %inst_path%\%%f.lst (
    echo creating COM routines of %%f%com_flag%.dll
    echo creating COM routines of %%f%com_flag%.dll >> %logfile%
    %REGFCT% %inst_path%\%%f.lst >> %logfile%
    if errorlevel 1 goto err
  )
)

rem ------------------ force checkpoint (only livecache) ----------------

if "%dbtype%" == "LVC" (
  if not "%vers%" == "74+" (
    echo.
    echo writing CHECKPOINT
    echo writing CHECKPOINT >> %logfile%
    %CHKPT% >> %logfile%
    if errorlevel 1 goto err
  )
)

rem ---------------------------------------------------------------------

:ok 
echo.
echo %dbtext2% %LC% successfully %donetxt%
echo %dbtext2% %LC% successfully %donetxt% >> %logfile%
goto history

:err
echo.
set errflag=1
echo ERROR : %dbtext2% %LC% not %donetxt% (see %logfile%)
echo ERROR : %dbtext2% %LC% not %donetxt% >> %logfile%

:history
echo ------------------------------------------------------------ >> %logfile%
date /t  >> %logfile%
time /t  >> %logfile%
echo ************************* END ****************************** >> %logfile%
rem  update the history file
type %logfile% >> %hisfile%

if %errflag% == 0 goto end

:end_err
rem set ERRORLEVEL
@verify error >NUL 2>&1

rem ---------------------------------------------------------------------
:end
