/*! 
  -----------------------------------------------------------------------------
 
  module: gls00.h
 
  -----------------------------------------------------------------------------
 
  responsible:  SteffenS and HansGeorgB

  special area: Replication Manager
 
  description:  "Fundamental definition module of the LS layer"
                
  version:      7.3.

  -----------------------------------------------------------------------------
 
  copyright:    SAP AG, 1997-2003
 
  -----------------------------------------------------------------------------


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/
#ifndef GLS00
#define GLS00


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#if defined(WIN32)
#include <strstrea.h>
#endif
#include <fstream.h>

#include "heo02.h"
#include "heo03.h"
#include "heo06.h"
#include "vsp007.h"

#include "SAPDBCommon/SAPDB_DynamicUTF8String.hpp"  // unicode support
#include "hsp77.h"                                  // buffer overflow safe sprintf function

#include "vin00.h"          // Header fuer zeit- und datumsumrechnungen und fuer ParseID-groessen

#include "geo47.h"
#include "gsp100.h"

/* START OF GENERAL DEFINITIONS: *******************************************************/

#if defined(_WIN32)
#define NEW_LINE_LS00 "\r\n"
const short NEW_LINE_LEN_LS00 = 2;
#else
#define NEW_LINE_LS00 "\n"
const short NEW_LINE_LEN_LS00 = 1;
#endif


const unsigned char UNDEF_BYTE_LS00         = STATIC_CAST(unsigned char, csp_undef_byte);
const char          DEF_BYTE_LS00           = csp_defined_byte;
const char          DEF_BYTE_ASCII_LS00     = '\x20';
const char          DEF_BYTE_EBCDIC_LS00    = '\x40';
const char          DEF_BYTE_UNICODE_LS00   = csp_unicode_def_byte;
const char          BLANK_LS00              = ' ';
const int           SLASH_LS00              = '/';
const int           BACKSLASH_LS00          = '\\';
const char          EOC_LS00                = STATIC_CAST(char,0x3B);

const char          QUOTE_LS00              = '"';
const char          COMMA_LS00              = ',';
const char          POINT_LS00              = '.';
const char          NO_THSD_GROUPS_LS00     = 'N';      // zur einstellung der tausendergruppen bei dezimalzahlen

const char          ProtOkHeader_ls00[5]    = "OK \n";
const char          ProtErrHeader_ls00[5]   = "ERR\n";
const int           ProtHeaderLength_ls00   = 4;

#define             PROTOCOL_SIZE_LS00        4*sizeof(char)

const int           MAX_PARAM_LS00          = 10;
const int           EOF_LS00                = -1;

const int           MAX_NUMERICLEN_LS00     = csp_fixed;
const int           MAX_SHORTREAL_LS00      = 4;
const int           MAX_LONGREAL_LS00       = 8;

const short         MAX_BOOLVALUE_LENGTH_LS00    = 10;

const int           MAX_PARSEID_LENGTH_LS00 = mxin_parsid;


#define             UNDEFINED_LS00  -2
#define             NULL_VALUE   "?"

// fuer tableextract/tableload
const tsp00_Uint1   MY_PAGE_LS00            = 255;

#ifdef UNICODE
// using this size value in char string definitions we have to consider the trailing NULL
// and the fact that names could be decorated with double quotes which cause the name
// to be 2 byte longer; the kernel however does not pay any attention to the additional bytes
//
// When using this size for kernel identifiers add 5 to be on the safe site
const short         KNL_IDENTIFIER_SIZE = sizeof(tsp00_KnlIdentifier);
#else
const short         KNL_IDENTIFIER_SIZE = sizeof(tsp00_KnlIdentifier)/2;
#endif

const int           MAX_REPLY_LENGTH_LS00   = 1024;

const int           BDINFO_SIZE_LS00        = sizeof(tsp7_bdinfo);

const char          COMP_NAME_LS00[65]      = "REPLICATION MANAGER (SERVER COMPONENT)                          ";
const char          APPL_NAME_LS00[4]       = csp_comp_load;

const char          HEADER_IDENTIFIER_LS00[33] = "REPLICATION SERVER SAP DB       ";

#define INSERT_EMPTY_VALUE        -10
#define INSERT_DEFAULT            -11
#define INSERT_NULL               -12
#define INSERT_ASSIGNED           -13


#define STOP_NOMORE_DATA_LS00     -111
#define STOP_LAST_DATA_LS00       -222
#define STOP_NO_DATA_LS00         -333


#define CMD_SEPARATOR_TOKEN_LS00       "//"
#define CMD_SEPARATOR_TOKEN_LEN_LS00   2


// Maximale LONG Daten Lnge im Eingabedatenstrom: 'Filename [von-bis]'
#define DEF_MAX_LONG_VALUE_LS00    1024

//  header sizes of an sql packet of the order interface
static tsp00_Int4 BlockHSize_ls00   = sizeof(tsp1_packet_header) + sizeof(tsp1_segment_header) + sizeof(tsp1_part_header);
static tsp00_Int4 PartHSize_ls00    = sizeof(tsp1_part_header);
static tsp00_Int4 SegmentHSize_ls00 = sizeof(tsp1_segment_header);
static tsp00_Int4 PacketHSize_ls00  = sizeof(tsp1_packet_header);

static tsp00_Int4 LongDescSize_ls00 = sizeof(tsp00_LongDescriptor);

/* Possible db data types

 0         1          2          3           4         5           6            7           8             9
 dfixed,   dfloat,    dcha,      dche,       dchb,     drowid,     dstra,       dstre,      dstrb,        dstrdb,
 10
 ddate,    dtime,     dvfloat,   dtimestamp, dunknown, dnumber,    dnonumber,   dduration,  ddbyteebcdic, dlonga,
 20
 dlonge,   dlongb,    dlongdb,   dboolean,   dunicode, ddtfiller1, ddtfiller2,  ddtfiller3, ddtfiller4,   dsmallint,
 30
 dinteger, dvarchara, dvarchare, dvarcharb,  dstruni,  dlonguni,   dvarcharuni, dudt,       dabaptabhandle
*/

// char data types
static char
szMap_CharDataType_ls00[39] =
    {0,0,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,1,1,1,0,0,1,0,0};

// long data types
static char
szMap_LongDataType_ls00[39] =
    {0,0,0,0,0,0,1,1,1,1,0,0,0,0,0,0,0,0,0,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0};

// number data types
static char
szMap_NumberDataType_ls00[39] =
    {1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0};


// special constant data types
static char
szMap_SpecConstDataType_ls00[39] =
    {0,0,0,0,0,0,0,0,0,0,1,1,0,1,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};


// blob data types
static char
szMap_BlobDataType_ls00[39] =
    {0,0,0,0,1,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0};


// Defines the necessity to send data of the particular data type as UCS2 when order interface
// is unicode (replication server works with unicode order interface if database is unicode)
static bool
bUnicodeOverOI_ls00[39] = {false,false,true ,true ,false,false,true ,true ,false,false,
                           true ,true ,false,true ,false,false,false,false,false,false,
                           false,false,false,false,true ,false,false,false,false,false,
                           false,true ,true ,false,true ,true ,true ,false,false};

// The following char string definition uses precompiler sql mode definition:
/*typedef enum tsp1_sqlmode_Enum
{
    sp1sm_nil,
    sp1sm_session_sqlmode,
    sp1sm_adabas,
    sp1sm_ansi,
    sp1sm_db2,
    sp1sm_oracle
} tsp1_sqlmode_Enum;
*/
static char const *pszSQLModeName_ls00[6] = {"INTERNAL", "INTERNAL", "INTERNAL", "ANSI", "DB2", "ORACLE"};

// The following char string definition uses precompiler sql mode definition:
/*typedef enum tpr00_SQLModeEnum {
  cpr_kind_empty         =   0,
  cpr_kind_internal      =   1,
  cpr_kind_db2           =   2,
  cpr_kind_ansi          =   3,
  cpr_kind_oracle        =   4,
  cpr_kind_sapr3         =   5
} tpr00_SQLModeEnum;
*/
static char const *pszPRSQLModeName_ls00[6] = {"INTERNAL", "INTERNAL", "DB2", "ANSI", "ORACLE", "INTERNAL"};

//  leading headers of an sql packet of the order interface
typedef char tls00_PktBlock [sizeof(tsp1_packet_header) + sizeof(tsp1_segment_header) + sizeof(tsp1_part_header)];


// Laenge des headers einer datendatei bei TABLEEXTRACT
#define TE_FILE_HEADER_LENGTH_LS00   1024


// DB return codes
typedef enum{
    dbNotLoadedCharset_els00   = -4025,
    dbUnknownTableName_els00   = -4004,
    dbRowNotFound_els00        = 100    
} tls00_DBReturnCodes;



//* Resulttypes of syntax parser
//* ACHTUNG: tls00_ScanResult und ScanResultName_ls00 sind miteinander korreliert!!!
typedef enum {
    srUnknown_els00,
    srRelease_els00,
    srFctCatExtr_els00,
    srFctCatLoad_els00,
    srFctDatExtr_els00,
    srFctDatLoad_els00,
    srFctDatUpd_els00,
    srFctDBExtr_els00,
    srFctDBLoad_els00,
    srFctFastLoad_els00,
    srFctTabExtr_els00,
    srFctTabLoad_els00,
    srFctTabUnlo_els00,
    srFctTabUpd_els00,  
    srFctSqlStmt_els00,
    srCmdSession_els00,
    srFctRemark_els00  //!!LogMessage
} tls00_ScanResult;

//* Namen fuer die oben definierten resultattypen
static const char* ScanResultName_ls00[] = {
//    "UNKNOWN",
    "RELEASE",
    "CATALOGEXTRACT",
    "CATALOGLOAD",
    "DATAEXTRACT",
    "DATALOAD",
    "DATAUPDATE",
    "DBEXTRACT",
    "DBLOAD",
    "FASTLOAD",
    "TABLEEXTRACT",
    "TABLELOAD",
    "TABLEUNLOAD",
    "TABLEUPDATE",
    "SQL_STMT",
    "SESSION_STMT",
    "REMARK"  ,          //!!LogMessage
    "EXTERN_STMT"
};

/* Modi zur zusaetzlichen klassifikation des parse-resultats */
typedef enum
{
    sfUnknown_els00,
    sfAutoCommitON_els00,
    sfUseUser_els00,
    sfRelease_els00
} tls00_SubFuncClass;

/* Modi for copying string to packet (ls04StringToPart) */
typedef enum {
    smUnknown_els00,
    smVarLeng_els00,
    smDefLeng_els00,
    smFixLeng_els00
} tls00_StrMode;


/* Character codes */
/* The order of the values MUST remain; if new values are necessary add them at the end*/
typedef enum {
    ctUnknown_els00,
    ctAscii_els00,
    ctEbcdic_els00,
    ctUTF8_els00,
    ctIgnore_els00,         // corresponds to BYTE in db and binary in data file struct
    ctUCS2_els00,           // not swapped UCS2 (high byte, low byte)
    ctUCS2Swapped_els00,    // swapped UCS2 (low byte, high byte)
    ctUCS2Native_els00,     // swapped/not swapped UCS2 depending on machine server running on
    ctCodeSet_els00
} tls00_CodeType;


/* Representation of date & time */
typedef enum {
    dtUnknown_els00 = -1,
    dtFreeMask_els00,
    dtISO_els00,
    dtUSA_els00,
    dtEUR_els00,
    dtJIS_els00,
    dtInternal_els00
} tls00_DateTime;


/* Errorcodes */
typedef enum {
    errOK_els00,
    errNoMemory_els00,
    errInitPckt_els00,
    errInitSegm_els00,
    errStr2Part_els00,
    errChr2Part_els00,
    errBuf2Part_els00,
    errFromPart_els00,
    errConvNumber_els00,
    errMoveRange_els00,
    errYaccAlloc_els00,
    errParseSyntax_els00,
    errFastLoad_els00,
    errVFClose_els00
} tls00_ErrorCodes;


typedef enum
{
    tptUnknown_els00 = -1,
    tptHeader_els00,
    tptData_els00,
    tptTableDescription_els00,
    tptDefault_els00,
    tptConstraint_els00,
    tptIndexDescription_els00,
    tptIndex_els00,
    tptLongString_els00,
    tptShortString_els00,
    tptEOF_els00,
    tptEOT_els00
} tls00_TEPageType;


typedef enum
{
    dbuUnknown_els00 = 0,
    dbuSysdba_els00,
    dbuDba_els00,
    dbuResource_els00,
    dbuColdUser_els00
} tls00_DBUserMode;

typedef enum 
{
   emExtractAsDDL_els31,
   emExtractAsXML_els31,
   emLoadAsDDL_els31   ,
   emLoadAsXML_els31   
} tls31_ExtrMode;


/* External data types */
typedef enum {
    ioChar_els00,
    ioDecimal_els00,
    ioZoned_els00,
    ioInteger_els00,
    ioReal_els00,
    ioOracleNumber_els00,
    ioOracleDate_els00
} tls00_InOutType;


/* Some more often used structures */

typedef struct tls00_String
{
    tsp00_Addr  strAddr_ls00;       // char pointer
    tsp00_Int4  strLeng_ls00;
}tls00_String;


typedef struct tls00_Buffer
{
    // size values correspond to the values buf len and buf size in part header of the order interface
    void                    *pBufPtr_ls00;
    tsp00_Longint            lBufSize_ls00;     // specifies the still available count of bytes in buffer - on creation time == the whole buffer
    tsp00_Longint            lBufLen_ls00;      // specifies count of used bytes in buffer - on creation time = 0
} tls00_Buffer;


typedef struct tls00_Buffer2
{
    void                    *pBufPtr_ls00;
    tsp00_Longint            lBufLen_ls00;      // buffer length, fix - determined when creating pBufPtr
    tsp00_Longint            lBufPos_ls00;      // used to navigate through the buffer
} tls00_Buffer2;


/*
 Explanation for TSP1_PARAM_INFO from vsp001.h

typedef struct                     tsp1_param_info
{
    tsp1_param_opt_set  sp1i_mode;          // enum type: mandatory, optional, default, escape_char,
    tsp1_param_io_type  sp1i_io_type;       // enum type: input, output, inout,
    tsp_data_type       sp1i_data_type;     // for explanation see gsp00.h
    tsp_int1            sp1i_frac;
    tsp_int2            sp1i_length;
    tsp_int2            sp1i_in_out_len;
    tsp_int4            sp1i_bufpos;
} tsp1_param_info;

*/

/*  DIE BEIDEN STRUKTUREN WERDEN Z.Z. NICHT BENOETIGT (21061999)

typedef char tls00_InfoBuf [csp_max_fields * sizeof(tsp1_param_info)];


typedef struct tls00_Table  // structure used for table descripion before any action on this table
{
    tsp00_KnlIdentifier tabTabName_ls00;    // table name is a kernel identifier and 32/64 (ascii/unicode) byte long
    tsp00_Int4          tabTabNLen_ls00;
    tsp00_Addr          tabColBuff_ls00;
    tsp00_Int2          tabColBLen_ls00;
    tsp00_Int2          tabFieldCount_ls00;                 // number of columns of a table
    tsp00_KnlIdentifier tabColName_ls00[csp_max_fields];    // max fields = 1024
    tsp00_Int2          tabColNLen_ls00[csp_max_fields];
    union
    {
        tsp1_param_info tabFieldInfo_ls00[csp_max_fields];  // info for every column of a table
        tls00_InfoBuf   tabInfoBuf_ls00;
    };
} tls00_Table;
*/

typedef struct tls00_FieldBasics
{
    tsp00_DataType  fbDataType_ls00;        // sp7fi_data_type
    teo00_Int1      fbFraction_ls00;        // sp7fi_frac           DER TYP MUSS BLEIBEN, DA SIGNED NOTWENDIG
    tsp00_Int2      fbLength_ls00;          // sp7fi_len
    tsp00_Int2      fbInOutLength_ls00;     // sp7fi_in_out_l
} tls00_FieldBasics;

typedef struct tls00_FieldInfo              //tsp7_ffield_info
{
    tsp00_Int2          fiStmtColNo_ls00;       // sp7fi_stmt_col_no
    tls00_FieldBasics   fiBasics_ls00;
    tsp00_Int2          fiDefColNo_ls00;        // sp7fi_def_col_no
    tsp00_ColMode       fiColMode_ls00;         // sp7fi_col_mode
    tsp7_len_type       fiLengthType_ls00;      // sp7fi_len_type
    tsp7_default        fiDefault_ls00;         // sp7fi_default
    tsp_c3              fiFiller_ls00;          // sp7fi_filler
} tls00_FieldInfo;


// hervorgegangen aus tsp1_param_info !! Bei aenderungen an der kernstruktur muss diese struktur also
// angepasst werden !!
typedef struct tls00_ParamInfo
{
    tsp1_param_opt_set  piMode_ls00;          // bit combination: mandatory, optional, default, escape_char,
    tsp1_param_io_type  piIoType_ls00;       // enum type: input, output, inout,
    tls00_FieldBasics   piBasicInfo_ls00;
    tsp00_Int4          piBufpos_ls00;
} tls00_ParamInfo;


// Die folgende struktur ist eine nachbildung der struktur tsp7_table_description aus vsp007.
// Dabei wurde allerdings auf die union verzichtet und die struktur tsp7_table_description_false
// nicht genutzt, da damit sonst sofort 80k speicher belegt werden wuerden.
// Ausserdem ist zusaetzlich eine struktur aufgenommen worden, die die spaltennamen der
// tabelle enthalten kann. Diese muss aber von der jeweiligen funktionalitaet nach eigenen
// erfordernissen gefuellt werden.

typedef struct tls00_TableDescription
{
    tsp00_KnlIdentifierc    tdAuthid_ls00;
    tsp00_KnlIdentifierc    tdTablename_ls00;

    //tsp00_Addr              tdColName_ls00[csp_max_fields];    // array of pointers
    tsp00_KnlIdentifierc    *tdColName_ls00[csp_max_fields];    // array of pointers

    tsp00_Int2              tdFieldCount_ls00;          // sp7t_field_cnt_F
    tsp00_SwapKind          tdKernelSwap_ls00;          // sp7t_kernel_swap_F
    tsp_int1                tdFiller_ls00;              // sp7t_filler_F
    tsp00_Int2              tdKernelLoc_ls00;           // sp7t_kernel_loc_F
    tsp00_Int2              tdConstraintCount_ls00;     // sp7t_constraint_cnt_F

    union
    {
        tls00_FieldInfo     *tdFieldsArray_ls00[csp_max_fields];        // sp7t_fields_F    csp_max_fields = 1024
        tls00_ParamInfo     *tdParamArray_ls00[csp_max_fields];
    };
} tls00_TableDescription;


// Sollte z.Z. (August 2000) nur fuer table load genutzt werden.
//
// Die folgende struktur ist eine nachbildung der struktur tsp7_index_description aus vsp007.
// Dabei wurde allerdings auf die union verzichtet und die strukturen tsp7_index_description_1
// und tsp7_index_description_3 nicht genutzt, da erstere nicht belegt wird
// und mit der zweiten sonst sofort mehr als 35k speicher belegt werden wuerden.
// Zusaetzlich ist ein zeiger array aufgenommen worden, welches die indexnamen der
// tabelle enthalten kann. Dieses muss aber von der jeweiligen funktionalitaet nach eigenen
// erfordernissen gefuellt werden.
typedef struct tls00_IndexDescription
{
    tsp00_Int2              idSingleIndexCnt_ls00;      // aenderung zu tsp7_index_description, dort ist das ein char
    tsp00_C2                idFiller_ls00;              // aenderung zu tsp7_index_description, dort ist das ein tsp_c3
    tsp00_Int2              idFirstMultIndex_ls00;
    tsp00_Int2              idMultIndexCnt_ls00;

    tsp7_info_single_index *idSingleIndex_ls00;
    tsp7_info_mult_index   *idMultIndex_ls00;    // array of pointers
    tsp00_KnlIdentifierc   *idIndexName_ls00;    // array of pointers

} tls00_IndexDescription;


typedef struct tls00_Parts4InternalSql
{
    tsp00_Int2            pfiCount_ls00;
    tsp1_part_kind_Enum  *pfiPartKind_ls00;
    tls00_String         *pfiData_ls00;
} tls00_Parts4InternalSql;


/***************************************************************************************/


/* Structure for LOAD-SESSION context */
typedef struct tls00_SessionInfo
{
    tsp00_KnlIdentifierc    siUserName_ls00;        // 18 byte (terminated 0) character array
    tsp00_CryptPw           siUserCPwd_ls00;        // 24 byte character
    tsp00_Namec             siUserPwd_ls00;         
    tsp00_KnlIdentifier     siUserGroup_ls00;       // 18 byte character; group user above belongs to
    tsp00_Addr              siUserKey_ls00;
    tsp00_NodeId            siDBNode_ls00;          // 64 byte character
    tsp00_DbNamec           siDBName_ls00;          // 18 byte character null terminated

    bool                    siDBConnection_ls00;    // db connection established?
    bool                    siSqlStat_ls00;
    tsp1_sqlmode_Enum       siSQLMode_ls00;
    tsp00_Int2              siIsoLevel_ls00;
    bool                    siNolog_ls00;
    bool                    siAutoCommit_ls00;
    bool                    siMessageOn_ls00;
    bool                    siInstallOn_ls00;

    // All parameter needed for codeset/termcharset conversions
    tsp00_KnlIdentifier     siTCharSet_ls00;
    bool                    siIgnoreTCS_ls00;
    tsp00_Ctable            siConversionTables_ls00[6];
    tls00_CodeType          siCodeType_ls00;

    tls00_DateTime          siDateType_ls00; 
    tls00_DateTime          siTimeType_ls00; 
    tls00_DateTime          siStampType_ls00; 
    char                    siThsdChar_ls00;
    char                    siDecPoint_ls00;
    char                    siSeperat_ls00[2];      // 1 byte sign and the null terminator
    char                    siDelimit_ls00[2];      // 1 byte sign and the null terminator

    tsp00_C20c              siNullSpec_ls00;
    
    // default value for bool-values,  10 bytes long + terminating 0
    char                    siTrueVal_ls00[MAX_BOOLVALUE_LENGTH_LS00 + 1];
    char                    siFalseVal_ls00[MAX_BOOLVALUE_LENGTH_LS00 + 1];
    tsp00_C3                siLanguage_ls00;
    tsp00_C1                siFiller_ls00;

    tsp00_Int4              siTASize_ls00;          // transaction size may be given by the user
    tsp00_Int4              siMaxAdmissibleErrors;

    bool                   *sipCancelByte_ls00;
} tls00_SessionInfo;


/* Structure for DATABASE-SESSION context */
typedef struct tls00_DBInfo
{
    tls00_String        dbiCommand_ls00;        // command to be continued
    tsp00_C3            dbiFiller_ls00;
    tsp00_TaskId        dbiSqlPid_ls00;     // 4 byte integer
    tsp00_TermId        dbiSqlTerm_ls00;    // 18 byte character
    tsp01_Service       dbiSqlServ_ls00;    // enum type: sql_user, sql_async_user, sql_utility, sqlDistribution_esp00
    tsp01_CommErr       dbiSqlRc_ls00;      // signed char (vsp001c.h), Communication Error
    tsp00_SwapKind      dbiOwnSwap_ls00;    // swap behavior (little endian, big endian); enum type: sw_dummy, sw_normal, sw_full_swapped, sw_part_swapped,
    tsp00_Int2          dbiPktCount_ls00;   // Number of packets
    tsp00_Int4          dbiPktRc_ls00;      // Return code for sending packet
    tls00_CodeType      dbiEncoding_ls00;   // encoding of db server connected to (unicode, ascii)

    tsp100_VersionID0   dbiDBVersion_ls00;
    tsp00_Int4          dbiPageSize_ls00;
    tsp00_Int4          dbiPktSize_ls00;    // Size of message packet
    tsp1_packet_list    dbiPktSndList_ls00; // Pointer to tsp1_packet
    tsp1_packet_list    dbiPktRcvList_ls00; // Pointer to tsp1_packet

    tls00_SessionInfo   dbiSession_ls00;    // see structure above
    tsp00_ErrText       dbiSqlErr_ls00;     // 40 byte character error text
    tsp00_Int4          dbiSqlRef_ls00;
    tls00_PktBlock      dbiPktInitBlock_ls00;  // packet-, segment- and part header at once;  see above for declaration
} tls00_DBInfo;


/* Structure for CLIENT/SERVER communication */
typedef struct tls00_Command
{
    tsp00_Addr          cmdReqDat_ls00;
    tsp00_Int4          cmdReqPos_ls00;
    tsp00_Int4          cmdReqErr_ls00;

    tsp00_Addr          cmdScnDat_ls00;
    tsp00_Addr          cmdRpyDat_ls00;         // wird in ls01-modul dynamisch mit 1024 byte allokiert und an die
                                                // funktionen weitergereicht
    tls00_ScanResult    cmdScanRC_ls00;
    tls00_SubFuncClass  cmdSubFunc_ls00;
    tsp00_C2            cmdFiller_ls00;
} tls00_CtrlSvrCmd;


typedef struct tls00_LargeNumber
{
    tsp00_Longint lnu2GB_Blocks_ls00;
    tsp00_Longint lnuRemainder_ls00;
} tls00_LargeNumber;

/* Structure for FILE access */
typedef struct tls00_VFile
{
    tsp00_Addr              vfFileName_ls00;        // name of file handled, null terminated
    tsp00_Int4              vfFileHndl_ls00;
    tsp5_rte_datakind       vfDataType_ls00;        // possible values: SP5VF_TEXT, SP5VF_BINARY
    tsp5_rte_filemode       vfFileMode_ls00;        // possible values: SP5VF_READ, SP5VF_WRITE, SP5VF_APPEND
    tsp5_rte_buffering_kind vfFileBuf_ls00;         // possible values: SP5BK_BUFFERED, SP5BK_UNBUFFERED, SP5BK_DISKSYNCED
    tsp5_rte_closeoption    vfFileCOpt_ls00;        // possible values: SP5VF_CLOSE_NORMAL, SP5VF_CLOSE_REWIND, SP5VF_CLOSE_NEXT_TAPE
    tls00_LargeNumber       vfFilePos_ls00;         // physical pos in file
    tls00_LargeNumber       vfFileLen_ls00;         // overall length of file
    tsp00_Int8              vfFileLineNum_ls00;
} tls00_VFile;


typedef struct tls00_Restart
{
    tls00_VFile             *rstRestartFile_ls00;
    tsp00_Addr               rstRestartDataFileName_ls00;
    tsp00_Addr               rstRestartEntry_ls00;
    tls00_LargeNumber        rstRestartPos_ls00;
} tls00_Restart;


/* Special structure for predefined default values. Used in FastLoadRecord structure
   and probably in other load structure. These defaults are delivered by the
   tls00_SessionInfo Structure. */
typedef struct tls00_PredefDefault
{
    tsp00_Addr    pdUserDefault;
    tsp00_Addr    pdUserGroupDefault;
    tsp00_Number  pdOracleUid_ls00;
    tsp00_C9      pdStamp_ls00;      // stamp length is actually 8 bytes + the undef/def byte, contains last inserted stamp value
    tsp00_C6      pdFiller1_ls00;
    tsp00_C1      pdFiller2_ls00;
} tls00_PredefDefault;

//*
//*     Structure used ONLY for fastload. Contains information about
//*     the page layout necessary to know if we want to fill the page.
//*
typedef struct tls00_PageInfo
{
    tsp00_Int4  pgiDataStartPos;
    tsp00_Int4  pgiDataEndPos;
    tsp00_Int4  pgiDataSpace;
} tls00_PageInfo;


typedef struct tls00_ParamInfo_LONG       // LONG Paramterinfo (verwendet bei dataload und dataextract)
{
    bool                 loiIsLong_ls00;          // TRUE: IS LONG COLUMN
    bool                 loiWithLongInfo_ls00;    // TRUE: Der Spalte wurde ein LONGFILE Parameter zugeordnet
    bool                 loiPart_ls00;            // bool, flagging if long is part of file or the whole file
    bool                 loibFileNumbered_ls00;   // nur beim Dataextract genutzt: wird die longfiledatei durchnumeriert?
    tsp00_Int2           loiNumberSize_ls00;      // nur beim Dataextract genutzt: wieviel ziffern beim durchnumerieren
                                                  // bzw. fuer positionsangaben verfuegbar
    tsp00_Int4           loiGreatestNum_ls00;     // nur bei DE, enthaelt groesste moegliche zahl fuer durchnumerierte datei

    tsp00_Addr           loiLongFileName_ls00;    // Command-Filename 

    tsp00_Addr           loiFileName_ls00;        // Inputfilename 
    tsp00_Longint        loiLength_ls00;          // 0: Laenge unbekannt, d.h. gesamter Fileinhalt ist signifikant 
    tls00_LargeNumber    loiFileOffset_ls00;      // File offset of the long data in the used long data file
    tsp00_Longint        loiCountOfSendData_ls00; // Anzahl bereits gesendeter Daten an den SQL Server

    tls00_VFile         *loiFileStruct_ls00;      // IO Fileinfo's 
    tls00_CodeType       loiFileEncoding_ls00;
} tls00_ParamInfo_LONG;


typedef struct tls00_DataSource 
{
    tls00_String             dasInputBuffer;
    tls00_String            *dasConvertedData;
    tsp00_Uint4              lReadFieldsCnt_ls00;
    tsp00_Uint4             *lFillIndex_ls00;
    tsp00_Uint4             *plDataToColIndex_ls00;
    tsp00_Uint4             *plDataToParamIndex_ls00;
    tsp00_Uint4             *lReadPos_ls00;
} tls00_DataSource;


// pattern is build from separators and delimiters; single separators do
// not happen but single delimiters must be handled; this is why we have
// patDelimit here
typedef struct tls00_Pattern
{
    tsp00_Addr               ppatPattern_ls00;
    tsp00_Longint            lpatPatternLen_ls00;
    tsp00_Addr               ppatDelimit_ls00;
    tsp00_Longint            lpatDelimiterLen_ls00;
    tsp00_Int4              *plpatFailure_ls00;
} tls00_Pattern;


typedef struct tls00_ColumnData
{
    tsp00_Longint       cdPosInStmt_ls00;
    tsp00_Addr          cdPtrToData_ls00;
    tsp00_Addr          cdDataBuffer_ls00;
    tsp00_Longint       cdDataBufferLen_ls00;
    tsp00_Longint       cdDataLen_ls00;
    tsp00_Longint       cdLONGLen_ls00;             // length of LONG value if column is of type long
    bool                cdPtrUsed_ls00;
    tsp00_DataType      cdSQLDataType_ls00;
    tsp00_Longint       cdSQLLen_ls00;
    tls00_InOutType     cdExtDataType_ls00;
    tls00_CodeType      cdExtEncoding_ls00;
    tls00_CodeType      cdExtLONGEncoding_ls00;
    bool                cdExtHexRepresentation_ls00;
    bool                cdIsNILValue_ls00;
    tsp_c1              cdFiller_ls00;
} tls00_ColumnData;


typedef struct tls00_MetaData
{
    tls00_VFile             *pmdMetaFile_ls00;          // not used; data file itself contains meta data 
    tsp00_Uint4              mdFieldCount_ls00;         // Count of columns in data file
    tsp00_Uint4              mdLongFieldCount_ls00;     // Count of columns of type long in data file
    tsp00_Uint4             *pmdLongFieldPos_ls00;       // array of positions of columns of type long in a single record
    tsp00_Int2              *pmdLengthArray_ls00;       // used to read meta data from file (length info of the data in file); may be negative to indicate a NULL value
    tsp00_Uint4              mdLengthArrayLen_ls00;     // length of length array
    tsp00_Uint2              mdLengthArrayPos_ls00;     // not used
    tsp00_Uint4              mdRecordLength_ls00;
    tsp00_SwapKind_Enum      mdSwapKind_ls00;           // swap kind of data in data file
} tls00_MetaData;

//*
//*     Structure for FASTLOAD execution --> FastLoadRecord
//*
typedef struct tls00_FloadRec
{
    tls00_TableDescription   flrTableInfo_ls00;
    
    tls00_DataSource         flrDataSource_ls00;
    tls00_Pattern            flrPattern_ls00;
    
    tsp7_bdinfo              flrBdInfo_ls00;         /* 48 byte character */

    tsp00_Key                flrLstKeyAcp_ls00;      /* Last key accepted */
    tsp00_Int2               flrLenKeyAcp_ls00;
    
    //* Special treatment for known default values. Delivered by tls00_SessionInfo
    tls00_PredefDefault      flrDefStruct_ls00;
    tls00_String            *flrDefValue_ls00[csp_max_fields];  // array of tls00_String pointers

    tsp00_Int2               flrPgPerPckt_ls00;          /* number of possible pages per packet */
    tsp00_Int2               flrRecordLen_ls00;          /* length of one record in a table */
    tsp00_Int2               flrFixColLen_ls00;          /* length of all fix columns together */
    tsp00_Int2               flrVarColCnt_ls00;          /* number of columns with variable length */
    tls00_PageInfo           flrPageInfo_ls00;

    //*
    //* The following 2 values are used to summarize the number of pages and the number of records loaded
    //* during one fastload run into a dedicated table. This information is used in the ENDLOAD command
    //* after loading the pages has successfully finished
    //*
    tsp00_Int4               flrPageCount_ls00;      // count of built pages sent to kernel; limited by kernel to MAX_INT4_SP00
    tsp00_Int8               flrRecordCount_ls00;    // count of processed records
    
    tsp00_Int8               flrLastCommLine_ls00;   // last committed input line
    tsp00_Int4               flrLRejected_ls00;      // count of rejected records
    
    tsp00_SwapKind           flrSwapKind_ls00;
    bool                     flrIsSyskey_ls00;       // syskey column (if no user defined key exists)
    bool                     flrTabEmpty_ls00;
    bool                     flrStampDef_ls00;
} tls00_FloadRec;



typedef struct tls00_LongExecInfo
{
    tsp00_Int2              *dliNNLongs_ls00;           // array of long column indexes in single record which are not null
    tsp00_Int2               dliCntNNLongs_ls00;         // Count of not null long cols in single record
    tsp00_Int2               dliFinishedNNLongs_ls00;    // Count of not null long cols in single records already processed          
} tls00_LongExecInfo;


// This structure is used in DATALOAD and DATAUPDATE processing for handling of data part
// in the order interface packet buffer
typedef struct tls00_DataPartHandling
{
    tls00_Buffer             dphBuffer_ls00;
    tsp00_Int4               dphRecordLength_ls00;

    // free space management in order interface buffer
    tsp00_Int4              *dphRecordNumbers_ls00; // record numbers in mass insert/update OI packet buffer
    tsp00_Int4              *dphRecordSpace_ls00;   // position/space of records in mass insert/update OI packet buffer
} tls00_DataPartHandling;


//*
//*     Structure for DATALOAD execution --> DataLoadRecord
//*
typedef struct tls00_DloadRec
{
    tls00_VFile             *dlrInfile_ls00;
    
    tls00_TableDescription   dlrTableInfo_ls00;
    
    tls00_DataSource         dlrDataSource_ls00;
    tls00_Pattern            dlrPattern_ls00;
    
    tls00_MetaData           dlrMetaData_ls00;

    char                     dlrParseId_ls00[MAX_PARSEID_LENGTH_LS00 + 1];
    unsigned char            dlrResultCount_ls00[sizeof(csp_rescnt_unknown)];
    tsp1_part_attributes     dlrPartAttr_ls00;
    tsp00_Int4               dlrMaxRowsPPckt_ls00;   // max. Anzahl von zeilen in einem insert packet = transaction size

    tls00_DataPartHandling   dlrDataPartHandling_ls00;

    // Long value information
    tls00_LongExecInfo       dlrDInfo_ls00;
    tsp00_Int4               dlrLongColCnt_ls00;     // Anzahl der long spalten in einer zu bearbeitenden tabelle
    tsp00_LongDescriptor    *dlrpLongDescriptorVector_ls00;

    tsp00_Int4               dlrRecordCount_ls00;    // Wird genutzt, um die anzahl geladener datensaetze zu halten
                                                    // und gibt gleichzeitig die anzahl geladener zeilen an (fuers
                                                    // log-file)

    tsp00_Longint            dlrInsertedRows_ls00;   // Anzahl der eingefuegten Datensaetze
    tsp00_Longint            dlrLastCommLine_ls00;   // Letzte 'commit'ted input line
    tsp00_Longint            dlrLRejected_ls00;      // Anzahl abgelehnter zeilen beim laden einer tabelle

    tsp00_SwapKind           dlrSwapKind_ls00;
} tls00_DloadRec;


//*
//*     Structure for DataUpdate execution --> DataUpdateRecord
//*
typedef struct tls00_DatUpdRec
{
     // databaseInfo
    tsp00_SwapKind          duprSwapKind_ls00;
    
    // orderInterface
    // part kind data : header
    tls00_String            duprpPktDatPartHeader;
    // part kind data : variable

    tls00_DataPartHandling   duprDataPartHandling_ls00;

    tsp00_Int4              duprMaxRowsPPckt_ls00;                            // max number of rows in update packet

    // data part: columns
    tls00_TableDescription  duprTableDescription_ls00;                        // sp1pk_shortinfo
    char                    duprParseId_ls00[MAX_PARSEID_LENGTH_LS00 + 1];    // sp1pk_parsid
    unsigned char           duprResultCount_ls00[sizeof(csp_rescnt_unknown)]; // sp1pk_resultcount
    tsp1_part_attributes    duprPartAttr_ls00;                                // sp1pa_first_packet, sp1pa_next_packet, sp1pa_last_packet
    // long data part : columns
    tls00_LongExecInfo      duprLongExecInfo_ls00;
    tsp00_LongDescriptor    *duprpVecLongDescriptor_ls00;                     // tsp_long_descriptor
    tsp00_Int4              duprLongColCnt_ls00;                              // Anzahl der long spalten in einer zu bearbeitenden tabelle
     
    // dataSource
    tls00_DataSource        duprDataSource_ls00;
    tls00_Pattern           duprPattern_ls00;
    tls00_VFile             *duprpvFile;
    //tls00_String          *pRecFields;
  
    // logging and recovery
    tsp00_Int4              duprRecordCount_ls00;                             // number of rows
    tsp00_Int4              duprUpdatedRows_ls00;                             // number of updated rows
    tsp00_Int4              duprLastCommLine_ls00;                            // last commited input line
    tsp00_Int4              duprRejectedRows_ls00;                            // number of rejected rows for update command
    
} tls00_DatUpdRec;


//*
//*     Structure for DATAEXTRACT execution --> DataExtractRecord
//*
typedef struct tls00_DextractRec
{
    tls00_VFile             *derCmdFile_ls00;         // command file in case of DE for DL/FL
    tls00_VFile             *derDataFile_ls00;        // data file

    tls00_TableDescription   derTableInfo_ls00;

    tls00_DataSource         derDataSource_ls00;
    tls00_Pattern           *pderPattern_ls00;          // pattern structure used when extracting to external format COMPRESSED

    tls00_ColumnData        *pderColumnData_ls00;

    char                     derParseId_ls00[MAX_PARSEID_LENGTH_LS00 + 1];
    tsp00_Addr               derRTableName_ls00;      // result table name
    
    bool                     derTabEmpty_ls00;        // indicates emptyness of table
    
    char                     derFillValue_ls00;

    tsp00_Int4              *derMapIndex_ls00;       // Mapping von column-struktur auf tableinfo-struktur
    tsp00_Int4             (*derExtLenFrac_ls00)[2]; // 2-dim-array; 1.wert = externe laenge, 2. wert = fraction part

    tls00_LongExecInfo       derLongInfo_ls00;        // info on longs processing/processed in a single record
    tsp00_LongDescriptor    *derLongDescArray_ls00;  // one-dim. array of long descriptors
    bool                     derUseGetvalPckt_ls00;   // indicates necessity of using getval packets to extract longs

    tsp00_Addr               derRawResultCount_ls00;  // result count in internal number representation
    tsp00_Int4               derResultCount_ls00;     // result count as integer value

    tsp00_Int2               derRecordLen_ls00;       // length of one record in a table
    tsp00_Int4               derMaxRowsPPckt_ls00;    // max. count of records in a single fetch packet
    tsp00_Int4               derRecordCount_ls00;     // count of extracted records

    tsp00_Int4               derErrNumInRecord_ls00;
    tsp00_Int4               derLRejected_ls00;
    tsp00_Int4               derLongColCnt_ls00;      // count of long columns in table to process

    tsp00_SwapKind           derSwapKind_ls00;
} tls00_DextractRec;


//*
//*     Structure for TABLEEXTRACT execution --> TableExtractRecord
//*
typedef struct tls00_TextractRec
{
    tsp7_bdinfo             terBdInfo_ls00;         // 48 byte character
    tsp7_bdinfo             terLongBdInfo_ls00;

    tsp00_Int2              terPgPerPckt_ls00;      // number of possible pages per packet
    tsp00_Int4              terPageSize_ls00;
    tsp00_Int4              terPageCount_ls00;

    tsp00_Int4              terPageNo_ls00;

    tsp00_SwapKind          terSwapKind_ls00;
    char                   *terpPageBuffer_ls00;

    tls00_VFile            *terMappingFile_ls00;

    // output file characteristics
    tls00_VFile            *terOutFile_ls00;
    tsp00_Int4              terFileCounter_ls00;

    tsp00_Int4              terCount_ls00;               // gibt an, wieviele datenseiten pro speichermedium genutzt werden sollen
//TODOTODO: gehoert eigentlich in tabextractcmd-struktur und muss dann hier ausgewertet werden mit der folgenden variablen
    tls00_LargeNumber       terMediumSize_ls00;
  
    // parameter for multitable extraction where single table is a special case
//    tsp00_Addr              terpSaveCatResultTab_ls00;
//    tls00_TableDescription *terpSaveCatResultTabTD_ls00;

//    tsp00_Addr             *terpTabNameArray_ls00;
//    tsp00_Addr             *terpUserNameArray_ls00;
    
//    tsp00_Int4              lNameCount_ls00;        // used for allocation of char-resources
    
    // Features of a single table
    tsp00_Int2              terConstraintCnt_ls00;
    tsp00_Int2              terDefaultCnt_ls00;         // user defined defaults
    tsp00_Int2              terLongColCnt_ls00;
    tsp00_Int2             *terDefaultArray_ls00;

    tls00_TEPageType        terPageTypeExtracting_ls00;

    tsp00_Int4              terNumTabInFile_ls00;
    tsp00_Int4              terNumTabExtracted_ls00;       // count of extracted tables
    tsp00_Int4              terNumTabExtractFailed_ls00;   // count of tables not extracted because of errors

    bool                    terTabEmpty_ls00;
} tls00_TextractRec;



typedef struct tls00_TloadRec
{
    tls00_VFile             *tlrInFile_ls00;
    tsp00_Int4               tlrCount_ls00;              // gibt an, wieviele datenseiten pro speichermedium beim te genutzt wurden

    tls00_VFile             *tlrMappingFile_ls00;

    tls00_Restart            tlrRestart_ls00;

    tls00_String             tlrTableInfo_ls00;
    tls00_String             tlrIndexInfo_ls00;

    tls00_String             tlrExtrTableInfo_ls00;
    tls00_String             tlrExtrIndexInfo_ls00;

    tsp7_bdinfo              tlrBdInfo_ls00;         // 48 byte character
    tsp7_bdinfo              tlrLongBdInfo_ls00;     // 48 byte character

    tsp00_Int4               tlrSndPcktNo_ls00;      // sent packet number
    tsp00_Int4               tlrRcvPcktNo_ls00;      // received packet number

    tsp00_Int2               tlrPgPerPckt_ls00;      // number of possible pages per packet to send

    tsp00_Int4               tlrPageSize_ls00;
    tsp00_Int4               tlrPageCount_ls00;

    tsp00_Int4               tlrPageNo_ls00;

    tls00_CodeType           tlrSourceEncoding_ls00;            // encoding of the source db; read from data file header

    tsp00_SwapKind           tlrSwapFromDb_ls00;
    tsp00_SwapKind           tlrSwapFromTE_ls00;

    tsp00_SwapKind           tlrTargetSwap_ls00;           // swap of the target db kernel

    tls00_TEPageType         tlrPageTypeExtracting_ls00;
    bool                     tlrContinueTable_ls00;          // soll neue tabelle geladen werden oder fortgesetzt
    bool                     tlrMultTabInFile_ls00;

    tsp00_Int4               tlrNumTabLoaded_ls00;
    tsp00_Int4               tlrNumTabRejected_ls00;

} tls00_TloadRec;


/* START OF LOAD-FUNCTION DESCRIPTIONS: (filled by vls11.ycc) **************************/


// Special abstract types derived from the known datatypes
typedef enum
{
    atUnknown_els00,
    atNumber_els00,
    atSpecConstant_els00,
    atLong_els00
} tls00_AbstractType;

/* Numerical Operators */
typedef enum {
    opUnknown_els00,
    opEqual_els00,
    opSmaller_els00,
    opLarger_els00,
    opSmallerEqual_els00,
    opLargerEqual_els00,
    opNotEqual_els00
} tls00_OpType;


/* Boolean Operators */
typedef enum
{
    boUnknown_els00,
    boAND_els00,
    boOR_els00,
    boNOT_els00
} tls00_BoolOpType;


/* Column assign values */
typedef enum
{
    asUnknown_els00,
    asLiteral_els00,
    asSeqno_els00,
    asUser_els00,
    asUsergroup_els00,
    asStamp_els00,
    asDate_els00,
    asTime_els00,
    asTimeStamp_els00,
    asDB2Sqlid_els00,
    asDB2Date_els00,
    asDB2Time_els00,
    asDB2Timestamp_els00,
    asOraSysdate_els00,
    asOraUid_els00
} tls00_Assign;


/* Modi for dataextract */
typedef enum
{
    resUnknown_els00,
    resDataload_els00,
    resFastload_els00,
    resDataUpdt_els00,
    resDataloadNoDDL_els00,
    resFastloadNoDDL_els00
} tls00_Restore;


/* Modi for dataload */
typedef enum
{
    dupUnknown_els00,
    dupIgnore_els00,
    dupReject_els00,
    dupUpdate_els00
} tls00_Duplicate;


/* TABLE SPECIFICATION: ****************************************************************/


/* Simple column condition */
typedef struct tls00_Condition
{
    tsp00_Int2      conStart_ls00;
    tsp00_Int2      conEnd_ls00;
    tls00_InOutType conFormat_ls00;
    bool            conIsHex_ls00;
    tsp00_Int2      conFrac_ls00;
    void            *conBinCompVal_ls00;
    tsp00_Addr      conCompVal_ls00;
    tsp00_Int4      conValLen_ls00;
    tls00_OpType    conCompOp_ls00;
    tsp00_Uint1     conColNo_ls00;
} tls00_Condition;


/* Node of complex column condition */
typedef struct tls00_CondSpec
{
    tls00_BoolOpType    cosBoolOp_ls00;
    bool                cosOtherwise_ls00;
    bool                cosDefNull_ls00;
    bool                cosCondValue_ls00;      // enthaelt true oder false je nach der auswertung
    tls00_Condition*    cosCond_ls00;
    tls00_CondSpec*     cosLeft_ls00;
    tls00_CondSpec*     cosRight_ls00;
} tls00_CondSpec;


/* Structure of table specification */
typedef struct tls00_TableSpec
{
    tsp00_KnlIdentifierc  tsTabName_ls00;
    tsp00_KnlIdentifierc  tsTabOwnr_ls00;
    tsp00_Addr            tsFullTabName_ls00;
    tls00_CondSpec       *tsCondSpec_ls00;
} tls00_TableSpec;


/* Position of a column */
typedef struct tls00_FieldPos
{
    tsp00_Int2          fpoStart_ls00;
    tsp00_Int2          fpoEnd_ls00;
} tls00_ColField;


/* External format of a column */
typedef struct tls00_FieldFormat
{
    tls00_InOutType     ffoFormat_ls00;
    bool                ffoIsHex_ls00;
    tsp00_Int2          ffoFrac_ls00;
} tls00_FieldFormat;


/* Transforming options of a column */
typedef struct tls00_FieldOption
{
    bool       fopScale_ls00;
    bool       fopTrunc_ls00;
    bool       fopRound_ls00;
    bool       fopAnyOperation_ls00;     // To optimize some numeric operation during fastload
    tsp00_Int4 fopScVal_ls00;
    tsp00_Int4 fopRTVal_ls00;
} tls00_FieldOption;


/* Internal combination of 2 structures */
typedef struct tls00_FormatSpec
{
    tls00_FieldFormat   fspFormat_ls00;
    tls00_FieldOption   fspOption_ls00;
} tls00_FormatSpec;


/* Assign value definition of a column */
typedef struct tls00_FieldAsgn
{
    tsp00_Addr      fasAsgnVal_ls00;
    tsp00_Int2      fasAsgnLen_ls00;
    tls00_Assign    fasAsgnTyp_ls00;
    tsp00_C1        fasFiller_ls00;
    tsp00_Int4      fasSeqNo_ls00;
    tsp00_Int4      fasSeqInc_ls00;
} tls00_FieldAsgn;


/* Structure of column specification */
typedef struct tls00_Column
{
    tsp00_KnlIdentifierc colName_ls00;  /* 32/64 byte in ascii/unicode */
    tsp00_Int2           colNumber_ls00;
    bool                 colIsLit_ls00;
    bool                 colIsKey_ls00;
    bool                 colIsSet_ls00;
    bool                 colIsWhere_ls00;
//    tsp00_Int2           colSpecialIndex;
    tls00_FieldPos       colFPos_ls00;
    tls00_FieldFormat    colFormat_ls00;
    tls00_FieldOption    colOption_ls00;
    tls00_FieldAsgn      colAssign_ls00;
    tls00_CondSpec*      colNullCond_ls00;
    tls00_ParamInfo_LONG colLongInfo;
    tsp00_C2             colFiller_ls00;
} tls00_Column;


/* List of columns */
typedef struct tls00_MultColSpec
{
    tsp00_Int4          mcsCount_ls00;
    tls00_Column*       mcsColumn_ls00[csp_max_fields];
} tls00_MultColSpec;


/* FILE SPECIFICATION: *****************************************************************/


/* Some file option specifications ... */
typedef struct tls00_CodeSpec
{
    tsp00_Addr      cdsCodeSet_ls00;
    tsp00_Int2      cdsCSetLen_ls00;
    tls00_CodeType  cdsCodeType_ls00;
    tsp00_C1        cdsFiller_ls00;
} tls00_CodeSpec;

/* The following structure is only used in vls11.ycc for parsing */
typedef struct tls00_DateTimeSpec
{
    tsp00_Addr      dtsMask_ls00;
    tls00_DateTime  dtsType_ls00;
    tsp00_C3        dtsFiller_ls00;
    tin_date_mask   dtsSpecialDateMask_ls00;    // This mask is needed only to be able to use date conversion
                                                // functions in vin41 (vin32)
} tls00_DateTimeSpec;

typedef struct tls00_ComprSpec
{
    tsp00_C1        cpsSeperat_ls00;
    tsp00_C1        cpsDelimit_ls00;
    bool            cpsIsFormat_ls00;
    bool            cpsIsCompr_ls00;
} tls00_ComprSpec;


//*
//*     Only (ONLY) used for parsing the command. The values are delivered to 
//*     structure tls00_FileFormat for later use.
//*
typedef struct tls00_BoolSpec
{
    tsp00_Int2      bspTValLen_ls00;
    tsp00_Addr      bspTrueVal_ls00;

    tsp00_Int2      bspFValLen_ls00;
    tsp00_Addr      bspFalseVal_ls00;
} tls00_BoolSpec;


/* Structure of file format options */
typedef struct tls00_FileFormat
{
    tls00_CodeType      ffoCodeType_ls00;       // Code type of the data file

    //*** date values given as in/outfile specification
    tls00_DateTime      ffoDateType_ls00;
    _TCHAR              ffoDateMask_ls00[128];           // null terminated string
    tin_date_mask       ffoSpecialDateMask_ls00;    // This mask is needed only to be able to use date conversion
                                                    // functions in vin41 (vin32)
    //*** time values given as in/outfile specification
    tls00_DateTime      ffoTimeType_ls00; 
    _TCHAR              ffoTimeMask_ls00[128];           // null terminated string
    tin_date_mask       ffoSpecialTimeMask_ls00;    // This mask is needed only to be able to use date conversion
    
    //*** time stamp values given as in/outfile specification
    tls00_DateTime      ffoTimeStampType_ls00; 
    _TCHAR              ffoTimeStampMask_ls00[128];          // null terminated string
    tin_date_mask       ffoSpecialTimeStampMask_ls00;   // This mask is needed only to be able to use date conversion

    //*** Decimal point specifier and 'thousand-separator' as in 1.000.000,00
    char                ffoThsdChar_ls00;
    char                ffoDecPoint_ls00;

    char                ffoSeperat_ls00[2];     // 1 byte sign and the null terminator
    char                ffoDelimit_ls00[2];     // 1 byte sign and the null terminator

    _TCHAR              ffoCodeSet_ls00[64];
    
    tsp00_Int2          ffoTValLen_ls00;        // boolean value TRUE - so far max 10 chars long
    _TCHAR              ffoTrueVal_ls00[11];
    
    tsp00_Int2          ffoFValLen_ls00;        // boolean value FALSE - so far max 10 chars long
    _TCHAR              ffoFalseVal_ls00[11];
    
    _TCHAR              ffoNullSpec_ls00[21];
    tsp00_Int4          ffoCntSpec_ls00; 

    tsp00_Int2          ffoCSetLen_ls00;
    tsp00_Int2          ffoNSpecLen_ls00;
    tsp00_SwapKind      ffoIntSpec_ls00; 

    //*** Values that describe the file format
    //*     The value ffoCompress_ls00 is so far only used for navigating within the
    //*     parser. Whithout that one could not decide what to do if COMPRESS and FORMATTED
    //*     are specified. In the program control flow only the value of ffoFormatted is
    //*     is used - if true the data is formatted, if false the data is compressed.
    bool                ffoBinary_ls00;         // angeschaltet liegt datei binaer vor, ansonsten in textformat
    bool                ffoFormatted_ls00;
    bool                ffoCompress_ls00;
    bool                ffoCompact_ls00;  //!!COMPACT
    bool                ffoXml_ls00;
    tsp00_Int4          ffoBytesPerLine_ls00;

} tls00_FileFormat;

typedef struct tls00_FileExtract
{
    tsp00_Longint feStart_ls00;
    tsp00_Longint feEnd_ls00;
} tls00_FileExtract;

/* File specification */
typedef struct tls00_FileSpec
{
    tsp00_Addr          fsFileName_ls00;        // null terminierter string
    tls00_FileFormat    fsFormat_ls00;
    tls00_FileExtract   fsExtract_ls00;
    bool                fsAppend_ls00;
    bool                fsFromOra_ls00;
    tsp00_Int2          fsLongColID_ls00;
    tsp00_Addr          fsLongColName_ls00;     // null terminierter string
} tls00_FileSpec;

typedef struct tls00_FileInfo
{
    tls00_VFile     vFile;
    tls00_FileSpec  fileSpec;
} tls00_FileInfo ;



/* List of file specifications */
typedef struct tls00_MultFileSpec
{
    tsp00_Int4      mfsCount_ls00;
    tls00_FileSpec* mfsFiles_ls00[csp_max_fields];
} tls00_MultFileSpec;

typedef struct tls00_Usage
{
    tsp00_Int4  uUsageValue_ls00;
    bool        uUsingRows_ls00;
    char        uFiller[3];
} tls00_Usage;

/* LOAD FUNCTIONS: FASTLOAD, DATALOAD, -EXTRACT & -UPDATE ******************************/
/* Additionally one db sql stmt structure is defined  **********************************/
typedef struct tls00_RemarkCmd   //!!LogMessage
{
    tsp00_Addr  dscRemark_ls00;
    tsp00_Int4  dscRemarkLen_ls00;
} tls00_RemarkCmd;

typedef struct tls00_DBSqlCmd
{
    tsp00_Addr  dscSqlStmt_ls00;
    tsp00_Int4  dscStmtLen_ls00;
} tls00_DBSqlCmd;

typedef struct tls00_FloadCmd
{
    tls00_Usage         flcUsage_ls00;
    tls00_TableSpec     flcTable_ls00;
    tls00_MultColSpec   flcColSpec_ls00;
    tls00_FileSpec      flcInfile_ls00;
} tls00_FloadCmd;

typedef struct tls00_DatLoadCmd
{
    bool                dloOracle_ls00;
    tls00_Duplicate     dloDplMode_ls00;
    tls00_TableSpec     dloTable_ls00;
    tls00_MultColSpec   dloColSpec_ls00;
    tls00_FileSpec      dloInfile_ls00;
    tls00_MultFileSpec  dloLongSpec_ls00;
} tls00_DatLoadCmd;

typedef struct tls00_DatExtrCmd
{
    tsp00_Addr          dexFullUserInput_ls00;
    tsp00_Addr          dexSelOrdStr_ls00;      // null terminated string pointer
    tls00_Restore       dexRestore_ls00;
    tls00_TableSpec     dexTable_ls00;
//    tls00_MultColSpec   dexColSpec_ls00;
    tls00_MultColSpec   dexOutSpec_ls00;
    tls00_FileSpec      dexCmdFile_ls00;
    tls00_FileSpec      dexDatFile_ls00;
    tls00_MultFileSpec  *dexLongSpec_ls00;
} tls00_DatExtrCmd;

typedef struct tls00_DatUpdCmd
{
    tls00_TableSpec     dupTable_ls00;
    tls00_MultColSpec   dupColSpec_ls00;
    tsp00_Int4          dupSetColIndex_ls00;    // specifies first column in dupColSpec_ls00 that specifies a column to be set
    tls00_FileSpec      dupInfile_ls00;
    tls00_MultFileSpec  dupLongSpec_ls00;
} tls00_DatUpdCmd;

/* LOAD FUNCTIONS: CATALOG, TABLE & DB-STATEMENTS **************************************/

typedef struct tls00_PartSpec
{
    tsp00_KnlIdentifierc psTabName_ls00;
    tsp00_KnlIdentifierc psTabOwnr_ls00;
    tsp00_Addr           psFullTabName_ls00;
    bool                 psSelUser_ls00;
    bool                 psSelAll_ls00;
    tsp00_C2             psFiller_ls00;
} tls00_PartSpec;

typedef struct tls00_CatExtrCmd
{
    tls00_PartSpec  cexPartSpec_ls00;
    tls00_FileSpec  cexOutfile_ls00;
} tls00_CatExtrCmd;

typedef struct tls00_CatLoadCmd
{
    tls00_PartSpec  cloPartSpec_ls00;
    tls00_FileSpec  cloInfile_ls00;
} tls00_CatLoadCmd;

typedef struct tls00_TabExtrCmd
{
    tls00_PartSpec  texPartSpec_ls00;
    tls00_FileSpec  texOutfile_ls00;
} tls00_TabExtrCmd;

typedef struct tls00_TabLoadCmd
{
    tls00_PartSpec  tloPartSpec_ls00;
    tls00_FileSpec  tloInfile_ls00;
    tls00_FileSpec  tloReport_ls00;
    bool            tloRestart_ls00;
    tsp00_C3        tloFiller_ls00;
} tls00_TabLoadCmd;

typedef struct tls00_TabUnloCmd
{
    tls00_PartSpec  tunPartSpec_ls00;
    tls00_FileSpec  tunOutfile_ls00;
} tls00_TabUnloCmd;

typedef struct tls00_TabUpdCmd
{
    tls00_PartSpec  tupPartSpec_ls00;
    tls00_FileSpec  tupInfile_ls00;
    tls00_FileSpec  tupReport_ls00;
} tls00_TabUpdCmd;

typedef struct tls00_DBExtrCmd
{
    tls00_FileSpec  dbeDatfile_ls00;
    tls00_FileSpec  dbeCatfile_ls00;
} tls00_DBExtrCmd;

typedef struct tls00_DBLoadCmd
{
    tls00_FileSpec  dblDatfile_ls00;
    tls00_FileSpec  dblCatfile_ls00;
    tls00_FileSpec  dblReport_ls00;
} tls00_DBLoadCmd;


/*!
  -----------------------------------------------------------------------------
 
  function:     ls00IsDBUnicode

  -----------------------------------------------------------------------------

  description:  Tests if db is UNICODE. To do so at least a connection 
                to the db must be established.
                The default encoding type is ASCII.
                The right value is set after connecting a user to the particular
                db (only then the db releases the information about its
                encoding type).

  arguments:    DBInfo          [IN]    - structure with all info on the running
                                           session and the database connected to

  returnvalue:  true    - db is unicode
                false   - the opposite
  -----------------------------------------------------------------------------
*/
inline bool
ls00IsDBUnicode(tls00_DBInfo *pDBInfo)
{
    assert(true == pDBInfo->dbiSession_ls00.siDBConnection_ls00);

    return (ctUCS2_els00 == pDBInfo->dbiEncoding_ls00);
}

#endif /*GLS00*/
