/*! 
  -----------------------------------------------------------------------------
 
  module: hls05.h
 
  -----------------------------------------------------------------------------
 
  responsible:  SteffenS

  special area: Replication Manager
 
  description:  String funktionen
                
  version:      7.3.

  -----------------------------------------------------------------------------
 
  copyright:    SAP AG, 1997-2002
 
  -----------------------------------------------------------------------------


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/

#ifndef HLS05
#define HLS05

#include "gls00.h"

/*!
  -----------------------------------------------------------------------------

  function:     ls05StrToUpr
  -----------------------------------------------------------------------------

  description:  Converts ASCII 8 bit Latin-1 string in place to UPPER CASE

  arguments:    pszString      [IN]  - string to convert
                lStringLength  [IN]  - length of the string

  returnvalue:  None
  -----------------------------------------------------------------------------
*/
void
ls05StrToUpr(tsp00_Addr pszString, tsp00_Int4 lStringLength);


/*!
  -----------------------------------------------------------------------------

  function:     ls05StrTrimRight - overloaded
  -----------------------------------------------------------------------------

  description:  A not NULL terminated ASCII string is trimmed right.
                The string itself is NOT CHANGED. Only its length is adapted.

                Does not work for UNICODE or MBCS strings.

  arguments:    pszString   [IN]      - string to trim
                lStrLength  [IN/OUT]  - string length before and after processing

  returnvalue:  None
  -----------------------------------------------------------------------------
*/
void
ls05StrTrimRight(tsp00_Addr pszString, tsp00_Int4& lStrLength);


/*!
  -----------------------------------------------------------------------------

  function:     ls05StrTrimRight - overloaded
  -----------------------------------------------------------------------------

  description:  A not NULL terminated ASCII string is trimmed right by a
                defined character.
                The string itself is NOT CHANGED. Only its length is adapted.

                Does not work for UNICODE or MBCS strings.
                
  arguments:    pszString   [IN]      - string to trim
                lStrLength  [IN/OUT]  - string length before and after processing
                cTrimVal    [IN]      - character to remove

  returnvalue:  None
  -----------------------------------------------------------------------------
*/
inline void
ls05StrTrimRight(tsp00_Addr pszString, tsp00_Longint& lStrLength, char cTrimVal)
{
    ROUTINE_DBG_MEO00 ("ls05StrTrimRight II");
    
    while ( (cTrimVal == pszString[lStrLength-1]) &&
            (lStrLength > 0) )
    {
        --lStrLength;
    }
}


/*!
  -----------------------------------------------------------------------------

  function:     ls05StrTrimRight - overloaded
  -----------------------------------------------------------------------------

  description:  A not NULL terminated string (ASCII or UNICODE) is trimmed right
                by a defined character (US7ASCII).
                The string itself is NOT CHANGED. Only its length is adapted.
                Length is delivered in bytes not in characters.

  arguments:    pszString  [IN]      - string to trim
                lStrLength [IN/OUT]  - string length before and after processing
                cTrimVal   [IN]      - character to remove - only US7ASCII
                CodeType   [IN]      - defines string encoding

  returnvalue:  Length of trimmed string in lStrLength in Bytes.
                
                No return code
  -----------------------------------------------------------------------------
*/
void
ls05StrTrimRight(tsp00_Addr      pszString,
                 tsp00_Longint  &lStrLength,
                 char            cTrimVal,
                 tls00_CodeType  CodeType);


/*!
  -----------------------------------------------------------------------------

  function:     ls05StrTrimLeft - overloaded, inline
  -----------------------------------------------------------------------------

  description:  A not NULL terminated ASCII string is trimmed left.

                Does not work for UNICODE or MBCS strings.

  arguments:    pszString   [IN]      - string to trim
                lStrLength  [IN/OUT]  - string length before and after processing

  returnvalue:  None
  -----------------------------------------------------------------------------
*/
inline void
ls05StrTrimLeft(tsp00_Addr& pszString, tsp00_Int4& lStrLength)
{
    ROUTINE_DBG_MEO00 ("ls05StrTrimLeft");

    while (isspace(*pszString))
    {
        pszString++;
        lStrLength -= 1;
    }
}


/*!
  -----------------------------------------------------------------------------

  function:     ls05StrTrimLeft - overloaded
  -----------------------------------------------------------------------------

  description:  Deletes leading chars defined by the caller (in place).
                The encoding type may one of the following: ASCII, UTF8 or UCS2.
                The char to delete must be given as US7ASCII.
                The string itself is modified and the length is adapted.
                
  arguments:    pBuffer        [IN/OUT]  - string to trim
                lBufferLength  [IN/OUT]  - string length before and after processing
                cTrimChar      [IN]      - char to remove (only US7ASCII)
                CodeType       [IN]      - encoding type

  returnvalue:  Length of trimmed string in lBufferLength in Bytes.
                Modified string in pBuffer.
                
                No return code.
  -----------------------------------------------------------------------------
*/
void
ls05StrTrimLeft(tsp00_Addr       pBuffer,
                tsp00_Longint   &lBufferLength,
                char             cTrimChar,
                tls00_CodeType   CodeType);


/*!
  -----------------------------------------------------------------------------

  function:     ls05StrTrimRightEx
  -----------------------------------------------------------------------------

  description:  Truncate trailing blanks from NULL terminated ASCII string.

                Does not work for unicode or MBCS strings.

  arguments:    pszString  [IN]  - string

  returnvalue:  None
  -----------------------------------------------------------------------------
*/
void
ls05StrTrimRightEx(tsp00_Addr pszString);


/*!
  -----------------------------------------------------------------------------

  function:     ls05StrTrimLeftEx
  -----------------------------------------------------------------------------

  description:  Removes leading blanks from a NULL terminated ASCII string.
                The given string length is adapted.

                Does not work for unicode or MBCS strings.

  arguments:    pszString   [IN]      - string to trim
                lStrLength  [IN/OUT]  - string length before and after processing

  returnvalue:  None
  -----------------------------------------------------------------------------
*/
void
ls05StrTrimLeftEx(tsp00_Addr pszString, tsp00_Longint& lStrLength);


/*!
  -----------------------------------------------------------------------------

  function:     ls05StrCmpWFill
  -----------------------------------------------------------------------------

  description:  Compares 2 values whereas the smaller is filled with blanks 
                before comparison.

  arguments:    pszValue1   [IN]  - first value
                lVal1Len    [IN]  - length of the first value
                pszValue2   [IN]  - second value
                lVal2Len    [IN]  - length of the second value

  returnvalue:  true    - values are equal
                false   - values are not equal
  -----------------------------------------------------------------------------
*/
bool
ls05StrCmpWFill(const tsp00_Addr pszValue1,
                tsp00_Int4       lVal1Len,
                const tsp00_Addr pszValue2,
                tsp00_Int4       lVal2Len);


/*!
  -----------------------------------------------------------------------------

  function:     ls05InsertFillBytes
  -----------------------------------------------------------------------------

  description:  Fills a value left aligned with padding bytes. The kind of
                padding bytes are defined using the db data type.
                Takes UNICODE (UCS2) into accout (with the right swap).

  arguments:    pszData        [IN]  - value to pad
                lDataLen       [IN]  - length of value
                DBDataType     [IN]  - data type
                lNeededDataLen [IN]  - length to fill to
                bUnicode       [IN]  - value is UCS2 or not
                mySwap         [IN]  - the swap

  returnvalue:  no
  -----------------------------------------------------------------------------
*/
void
ls05InsertFillBytes(unsigned char       *pszData,
                    tsp00_Longint        lDataLen,
                    tsp00_DataType       DBDataType,
                    tsp00_Longint        lNeededDataLen,
                    bool                 bUnicode,
                    tsp00_SwapKind_Enum  mySwap);


/*!
  -----------------------------------------------------------------------------
  function:     ls05AreEncodingsCompatible
  -----------------------------------------------------------------------------

  description:  

  arguments:    

  returnvalue:  errOK_els00                    - External encoding allowed
                errIncompatibleEncodings_els98 - Incompatible external and internal
                                                  encodings; a conversion from
                                                  external to internal or vice versa
                                                  would lose information
  -----------------------------------------------------------------------------
*/
tsp00_Longint
ls05AreEncodingsCompatible(tls00_CodeType DataFile_Encoding,
                           tls00_CodeType DB_Encoding,
                           bool           bLoad = true);



/*!
  -----------------------------------------------------------------------------

  function:     ls05ConvertToUCS2
  -----------------------------------------------------------------------------

  description:  Converts a UTF8 or ASCII string to native UCS2 encoding (native
                means that byte order of UCS2 chars correspond to machine swap
                of machine running on).
                Takes db data type into account, i.e. data of data type byte
                is not converted.

  arguments:    pszDestination  [OUT]     - destination string (ucs2)
                lDestinationLen [IN/OUT]  - length of destination string in byte
                pszSource       [IN]      - source string (ascii or utf8)
                lSourceLen      [IN]      - length of source string
                SrcEncoding     [IN]      - encoding of source string
                DBDataType      [IN]      - db data type
                pszErrText      [OUT]     - error text

  returnvalue:  destination string length in lDestinationLen
                
                return code:
                errInternal_els98 - when partial character in source, but hit end or
                                    insufficient space in target for conversion
                errConvertingEncoding_els98 - source contains invalid chars or char sequences
  -----------------------------------------------------------------------------
*/
tsp00_Longint
ls05ConvertToUCS2(unsigned char  *pszDestination,
                  tsp00_Longuint &lDestinationLen,
                  unsigned char  *pszSource,
                  tsp00_Longint   lSourceLen,
                  tls00_CodeType  SrcEncoding,
                  tsp00_DataType  DBDataType,
                  tsp00_Addr      pszErrText);


/*!
  -----------------------------------------------------------------------------

  function:     ls05ConvertToUCS2Simple
  -----------------------------------------------------------------------------

  description:  Converts UTF8 or ASCII string to native UCS2 encoding (native
                means that byte order of UCS2 chars corresponds to machine swap
                of machine running on).

  arguments:    pszDestination  [OUT] - destination string (ucs2)
                lDestinationLen [IN]  - length of destination string in byte
                pszSource       [IN]  - source string (ascii or utf8)
                lSourceLen      [IN]  - length of source string
                SrcEncoding     [IN]  - encoding of source string
                pszErrText      [OUT] - error text

  returnvalue:  destination string length in lDestinationLen
                
                return code:
                errOK_els00                 - successful conversion
                errInternal_els98           - when partial character in source,
                                                but hit end or insufficient
                                                space in target for conversion
                errConvertingEncoding_els98 - source contains invalid chars or
                                               char sequences
  -----------------------------------------------------------------------------
*/
tsp00_Longint
ls05ConvertToUCS2Simple(unsigned char  *pszDestination,
                        tsp00_Longuint &lDestinationLen,
                        unsigned char  *pszSource,
                        tsp00_Longint   lSourceLen,
                        tls00_CodeType  SrcEncoding,
                        tsp00_Addr      pszErrText);


/*!
  -----------------------------------------------------------------------------

  function:     ls05ConvertToUCS2NotSwapped
  -----------------------------------------------------------------------------

  description:  Converts a UTF8 or ASCII string to UCS2 encoding not swapped
                (means that byte order of UCS2 chars corresponds big endian).

  arguments:    pszDestination  [OUT]     - destination string (ucs2)
                lDestinationLen [IN/OUT]  - length of destination string in byte
                pszSource       [IN]      - source string (ascii or utf8)
                lSourceLen      [IN]      - length of source string
                SrcEncoding     [IN]      - encoding of source string
                pszErrText      [OUT]     - error text

  returnvalue:  destination string length in lDestinationLen
                
                return code:
                errOK_els00                 - successful conversion
                errInternal_els98           - when partial character in source,
                                                but hit end or insufficient
                                                space in target for conversion
                errConvertingEncoding_els98 - source contains invalid chars or
                                               char sequences
  -----------------------------------------------------------------------------
*/
tsp00_Longint
ls05ConvertToUCS2NotSwapped(unsigned char  *pszDestination,
                            tsp00_Longuint &lDestinationLen,
                            unsigned char  *pszSource,
                            tsp00_Longuint  lSourceLen,
                            tls00_CodeType  SrcEncoding,
                            tsp00_Addr      pszErrText);

/*!
  -----------------------------------------------------------------------------

  function:     ls05ConvertFromUCS2
  -----------------------------------------------------------------------------

  description:  Converts an UCS2 (always native UCS2 which means that byte order
                of UCS2 chars correspond to machine swap of machine running on)
                to UTF8 or ASCII .
                Takes db data type into account, i.e. data of data type byte
                is not converted.

  arguments:    pszDestination  [OUT]    - destination string (ASCII or UTF8)
                lDestinationLen [IN/OUT] - length of destination string in byte
                                            before and after conversion
                DestEncoding    [IN]     - encoding of destination string
                pszSource       [IN]     - source string (UCS2 native)
                lSourceLen      [IN]     - length of source string
                DBDataType      [IN]     - db data type
                pszErrText      [OUT]    - error text

  returnvalue:  destination string length in lDestinationLen

                return code:
                errOK_els00                 - successful conversion
                errInternal_els98           - when partial character in source,
                                                but hit end or insufficient
                                                space in target for conversion
                errConvertingEncoding_els98 - source contains invalid chars or
                                               char sequences
  -----------------------------------------------------------------------------
*/
tsp00_Longint
ls05ConvertFromUCS2(unsigned char  *pszDestination,
                    tsp00_Longint  &lDestinationLen,
                    tls00_CodeType  DestEncoding,
                    unsigned char  *pszSource,
                    tsp00_Longint   lSourceLen,
                    tsp00_DataType  DBDataType,
                    tsp00_Addr      pszErrText);


/*!
  -----------------------------------------------------------------------------

  function:     ls05ConvertFromUCS2Simple - overloaded
  -----------------------------------------------------------------------------

  description:  Converts a UCS2 encoded string to UTF8 or ASCII string 

  arguments:    pszDestination  [OUT] - destination string (ascii or utf8)
                pszSource       [IN]  - source string (ucs2)
                lDestinationLen [OUT] - length of destination string in byte
                lSourceLen      [IN]  - length of source string
                DestEncoding    [IN]  - encoding of destination string
                bSourceSwapped  [IN]  - source string UCS2 swapped or not
                pszErrText      [OUT] - error text

  returnvalue:  destination string length in lDestinationLen

                return code:
                errOK_els00                 - successful conversion
                errInternal_els98           - when partial character in source,
                                                but hit end or insufficient
                                                space in target for conversion
                errConvertingEncoding_els98 - source contains invalid chars or
                                               char sequences
  -----------------------------------------------------------------------------
*/
tsp00_Longint
ls05ConvertFromUCS2Simple(unsigned char  *pszDestination,
                          unsigned char  *pszSource,
                          tsp00_Longint  &lDestinationLen,
                          tsp00_Longint   lSourceLen,
                          tls00_CodeType  DestEncoding,
                          bool            bSourceSwapped,
                          tsp00_Addr      pszErrText);

/*!
  -----------------------------------------------------------------------------

  function:     ls05ConvertFromUCS2Simple - overloaded
  -----------------------------------------------------------------------------

  description:  Converts a UCS2 native encoded string to UTF8 or ASCII string 
                (native means that byte order of UCS2 chars corresponds to
                 machine swap of machine running on).

  arguments:    pszDestination  [OUT] - destination string (ascii or utf8)
                pszSource       [IN]  - source string (ucs2)
                lDestinationLen [OUT] - length of destination string in byte
                lSourceLen      [IN]  - length of source string
                DestEncoding    [IN]  - encoding of destination string
                pszErrText      [OUT] - error text

  returnvalue:  destination string length in lDestinationLen

                return code:
                errOK_els00                 - successful conversion
                errInternal_els98           - when partial character in source,
                                                but hit end or insufficient
                                                space in target for conversion
                errConvertingEncoding_els98 - source contains invalid chars or
                                               char sequences
  -----------------------------------------------------------------------------
*/
tsp00_Longint
ls05ConvertFromUCS2Simple(unsigned char  *pszDestination,
                          unsigned char  *pszSource,
                          tsp00_Longint  &lDestinationLen,
                          tsp00_Longint   lSourceLen,
                          tls00_CodeType  DestEncoding,
                          tsp00_Addr      pszErrText);


/*!
  -----------------------------------------------------------------------------

  function:     ls05ConvertFromUTF8 - overloaded
  -----------------------------------------------------------------------------

  description:  Converts a string from UTF8 to UCS2 or ASCII.

  arguments:    pszDestination  [OUT] - destination string (UCS2 or ASCII)
                lDestinationLen [OUT] - length of destination string in byte
                Source          [IN]  - source string (UTF8)
                DestEncoding    [IN]  - encoding of destination string; in case
                                         of UCS2 it must define the swap, too
                pszErrText      [OUT] - error text

  returnvalue:  destination string length in lDestinationLen

                return code:
                errOK_els00                 - successful conversion
                errInternal_els98           - when partial character in source,
                                                but hit end or insufficient
                                                space in target for conversion
                errConvertingEncoding_els98 - source contains invalid chars or
                                               char sequences
  -----------------------------------------------------------------------------
*/
tsp00_Longint
ls05ConvertFromUTF8(unsigned char           *pszDestination,
                    tsp00_Longuint          &lDestinationLen,
                    SAPDB_DynamicUTF8String &Source,
                    tls00_CodeType           DestEncoding,
                    tsp00_Addr               pszErrText);


/*!
  -----------------------------------------------------------------------------

  function:     ls05ConvertFromUTF8 - overloaded
  -----------------------------------------------------------------------------

  description:  Converts a string from UTF8 to UCS2 or ASCII.

  arguments:    pszDestination  [OUT] - destination string (UCS2 or ASCII)
                lDestinationLen [OUT] - length of destination string in Bytes
                pszSource       [IN]  - source string (UTF8)
                lSourceLen      [IN]  - source string length in Bytes
                DestEncoding    [IN]  - encoding of destination string; in case
                                         of UCS2 it must define the swap, too
                pszErrText      [OUT] - error text

  returnvalue:  destination string length in lDestinationLen

                return code:
                errOK_els00                 - successful conversion
                errInternal_els98           - when partial character in source,
                                                but hit end or insufficient
                                                space in target for conversion
                errConvertingEncoding_els98 - source contains invalid chars or
                                               char sequences
  -----------------------------------------------------------------------------
*/
tsp00_Longint
ls05ConvertFromUTF8(unsigned char           *pszDestination,
                    tsp00_Longuint          &lDestinationLen,
                    unsigned char           *pszSource,
                    tsp00_Longuint           lSourceLen,
                    tls00_CodeType           DestEncoding,
                    tsp00_Addr               pszErrText);

/*!
  -----------------------------------------------------------------------------

  function:     ls05ConvertToUTF8
  -----------------------------------------------------------------------------

  description:  Converts a string from UCS2 or ASCII to UTF8.

  arguments:    pszDestination  [OUT] - destination string (UTF8)
                pszSource       [IN]  - source string (UCS2 or ASCII)
                lSourceLen      [IN]  - length of source string in Bytes
                SrcEncoding     [IN]  - encoding of source string; in case
                                         of UCS2 it must define the swap, too
                pszErrText      [OUT] - error text

  returnvalue:  destination string length in lDestinationLen

                return code:
                errOK_els00                 - successful conversion
                errInternal_els98           - when partial character in source,
                                                but hit end or insufficient
                                                space in target for conversion
                errConvertingEncoding_els98 - source contains invalid chars or
                                               char sequences
  -----------------------------------------------------------------------------
*/
tsp00_Longint
ls05ConvertToUTF8(SAPDB_DynamicUTF8String &Destination,
                  unsigned char           *pszSource,
                  tsp00_Longuint           lSourceLen,
                  tls00_CodeType           SrcEncoding,     // must give a hint on swapped or not - e.g. ctUCS2_els00 or ctUCS2Native_els00 or ctUCS2Swapped_els00
                  tsp00_Addr               pszErrText);

#endif /*HLS05*/