/*! 
  -----------------------------------------------------------------------------
 
  module: hls26.h
 
  -----------------------------------------------------------------------------
 
  responsible:  SteffenS

  special area: Replication Server
 
  description:  Definition of exported functions for TABLEEXTRACT
                
  version:      7.3.

  -----------------------------------------------------------------------------
 
  copyright:    SAP AG, 1999-2002
 
  -----------------------------------------------------------------------------


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/
#ifndef HLS26
#define HLS26

#include "gls00.h"

#define MAX_MEDIUM_SIZE_IN_BYTES_LS26    (1024*1024*1024)    // 1 GB in Byte so far !

/*!
  -----------------------------------------------------------------------------

  function: ls26SwitchUser
  -----------------------------------------------------------------------------

  description:  Compares user name of logged on user and requested user. If
                user names are not equal it switches the session to the new
                user without changing other session context and
                without supplying a password.

  arguments:    pDBInfo     [IN]  - structure with all information on the
                                     running session and the db connected to.
                CurrentUser [IN]  - name of user currently logged on, UTF8 encoded
                NewUser     [IN]  - name of user to log on, UTF8 encoded
                pszErrText  [OUT] - error text

  returnvalue:  errOK_els00
                errInternSQL_els98 - error executing statement that switches the user
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls26SwitchUser(tls00_DBInfo             *pDBInfo,
               SAPDB_DynamicUTF8String   CurrentUser,
               SAPDB_DynamicUTF8String   NewUser,
               tsp00_Addr                pszErrText);

/*!
  -----------------------------------------------------------------------------
 
  function:     ls26ExecTableExtract
  -----------------------------------------------------------------------------

  description:  main function for table extract execution
  
                This function does the first things that are necessary to extract
                a table. It then starts the data export.

  arguments:    *TECmd   [IN]  - table extract cmd structure;  contains all
                                  necessary information for execution.
                *DBInfo  [IN]  - structure with all information on the
                                  running session and the db connected to.
                *ErrText [OUT] - error text

  returnvalue:  errOK_els00    - everything went fine
                errors of the called functions, error lyrics can be found in log file.
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls26ExecTableExtract(tls00_TabExtrCmd *TECmd,
                     tls00_DBInfo     *DBInfo,
                     tsp00_Addr        ErrText);


#endif /*HLS26*/