/*! 
  -----------------------------------------------------------------------------
 
  module: hls26_TETable.h
 
  -----------------------------------------------------------------------------
 
  responsible:  SteffenS

  special area: Replication Server
 
  description:  Definition of exported functions for TABLEEXTRACT for a single table
                
  version:      7.3.

  -----------------------------------------------------------------------------
 
  copyright:    SAP AG, 2000-2002
 
  -----------------------------------------------------------------------------


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/
#ifndef HLS26_TETABLE
#define HLS26_TETABLE

#include "gls00.h"

#define STOP_ANALYSE_LS26  -1

#define GET_TABLE_SIZE_STMT_LS26   _T("SELECT NUMERIC_VALUE + SHORT_COLUMN_PAGES FROM SYSDBA.PAGES WHERE OWNER = '%s' AND TABLENAME = '%s'")


#define SAVE_LONG_STMT_LS26        _T("BEGIN SAVE LONG \"%s\"")
#define SAVE_TABLE_STMT_LS26       _T("BEGIN SAVE TABLE \"%s\"")
#define SAVE_COLUMN_STMT_LS26      _T("BEGIN SAVE COLUMN")
#define DESCRIBE_C_STMT_LS26       _T("DESCRIBE CONSTRAINT \"%s\"")

const tsp00_Int4 lStmtLength_ls26 = 4 * sizeof(tsp00_KnlIdentifier);

#define FILE_EXTENSION_SIZE_LS26     4
#define FILECOUNT_MXLS26             9999


typedef tsp00_Int2     tls26_Co_De_LenType;     // constraint/default length - type;
                                                // this is the int type that the constraint/default length is stored with


// AUFBAU DES HEADERS FUER EINE DATENDATEI (1 KBYTE)
// swap des servers bei extract     1 byte
// code type (ascii, ebcdic, ...)   1
// identifier                       64
// version                          40
// datum                            8
// time                             8

// page size der extrahierten db    4

// count                            4 gibt an, ob und wieviele datenseiten pro datei geschrieben wurden
// mehrere tabellen pro datei       1 wahr oder falsch


// DIESE ANGABEN SIND EIGENTLICH UNNOETIG, DA SIE IN DER MITEXTRAHIERTEN TABELLENBESCHREIBUNG ENTHALTEN SIND
// table name                       64
// user name                        18
// constraint count                 2
// index count                      2
// user defined defaults            2



typedef struct tls26_TEFileTrailer         // trailer of tableextract data files
{
    tsp00_Int4       tftPageNo_ls26;
    tsp00_Uint1      tftType_ls26;          // special type, der eine seite als eigene beschreibung kennzeichnet
    tsp00_C3         tftFiller1_ls26;
    tls00_TEPageType tftPageType_ls26;              // type of trailer page: tptEOF_ls00
    tsp00_Int4       tftNumOfTabInFile_ls26;        // number of tables in data file
    tsp00_Int4       tftNumOfPagesInFile_ls26;      // number of pages in data file
    bool             tftTableIncomplete_ls26;       // bool value: is the last table saved in data file only incompletely saved
    tls00_TEPageType tftDataTypeAtInterrupt_ls26;   // in case last table has been saved incompletely - the last save type of data
} tls26_TEFileTrailer;


typedef struct tls26_DescPageHeader         // header of description pages
{
    tsp00_Int4       dphPageNo_ls26;
    tsp00_Uint1      dphType_ls26;          // das ist ein spezielle typ, der eine seite als eigene beschreibung kennzeichnet
    tsp00_C3         dphFiller1_ls26;
    tls00_TEPageType dphPageType_ls26;
    tsp00_Int4       dphSpecialID_ls26;
    tsp00_Int4       dphMore_ls26;
    tsp00_Int4       dphFillerInt_ls26;
} tls26_DescPageHeader;


/*!
  -----------------------------------------------------------------------------
 
  function:     ls26ExtractSingleTable
  -----------------------------------------------------------------------------

  description:  Extracts a single table in database internal format

  arguments:    DBInfo     [IN]  - structure with all info on the running
                                    session and the database connected to.
                TableName  [IN]  - table name (UTF8)
                TERec      [IN]  - structure with all info needed
                                    to execute table extract
                pszErrText [OUT] - error text

  returnvalue:  errOK_els00      - No error
                some errors
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls26ExtractSingleTable(tls00_DBInfo            *DBInfo,
                       SAPDB_DynamicUTF8String &TableName,
                       tls00_TextractRec       *TERec,
                       tsp00_Addr               pszErrText);


/*!
  -----------------------------------------------------------------------------
 
  function:     ls26WriteHeaderPage
  -----------------------------------------------------------------------------

  description:  Writes the data file header.

  arguments:    DBInfo           [IN]  - structure with all info on the running
                                          session and the database connected to.
                TERec            [IN]  - structure with all info needed
                                          to execute table extract
                bMultipleTables  [IN]  - shows if the data file contains data of
                                          multiple tables
                pszErrText       [OUT] - error text

  returnvalue:  errOK_els00          - No error
                errFileWrite_els98   - error writing header to data file
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls26WriteHeaderPage(tls00_DBInfo      *DBInfo,
                    tls00_TextractRec *TERec,
                    bool               bMultipleTables,
                    tsp00_Addr         pszErrText); 


/*!
  -----------------------------------------------------------------------------
 
  function:     ls26WriteFileTrailer

  -----------------------------------------------------------------------------

  description:  Writes the data file header.

  arguments:    DBInfo           [IN]  - structure with all info on the running
                                          session and the database connected to.
                TERec            [IN]  - structure with all info needed
                                          to execute table extract
                bTableIncomplete [IN]  - shows if last table is only incompletely
                                          saved in data file
                pszFileName      [IN]  - name of next used data file (only for 
                                          TE USER or TE ALL)
                pszErrText       [OUT] - error text

  returnvalue:  errOK_els00          - No error
                errFileWrite_els98   - error writing trailer to data file
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls26WriteFileTrailer(tls00_DBInfo      *DBInfo,
                     tls00_TextractRec *TERec,
                     bool               bTableIncomplete,
                     tsp00_Addr         pszFileName,
                     tsp00_Addr         pszErrText);


#endif /*HLS26_TETABLE*/