/*! 
  -----------------------------------------------------------------------------
 
  module: hls27_DEExec.h
 
  -----------------------------------------------------------------------------
 
  responsible:  SteffenS

  special area: Replication Manager
 
  description:  Functions for dataextract execution
                
  version:      7.3.

  -----------------------------------------------------------------------------
 
  copyright:    SAP AG, 1999-2002
 
  -----------------------------------------------------------------------------


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/

#ifndef HLS27_DEEXEC
#define HLS27_DEEXEC

#include "gls00.h"

typedef struct tls27_DataStruct
{
    tsp00_Addr  dasSimplePtr_ls27;
    tsp00_Addr  dasAllocMemPtr_ls27;
    tsp00_Int4  dasValLength_ls27;
    bool        dasSimplePtrUsed_ls27;
    tsp_c3      dasFiller_ls27;
} tls27_DataStruct;


typedef struct tls27_DataSink
{
    tls27_DataStruct dskStruct_ls27;
    tls00_Buffer2    dskMetaLine_ls27;
    tls00_Buffer2    dskOutLine_ls27;
} tls27_DataSink;


/*!
  -----------------------------------------------------------------------------
 
  function:     ls27ExtractTableData

  -----------------------------------------------------------------------------

  description:  Extracts data from a single table.

  arguments:    DBInfo       [IN]   - pointer; db session structure
                DextRec      [IN]   - pointer; extract record structure
                DextCmd      [OUT]  - pointer; extract command structure
                pszErrText   [OUT]  - fehlertext

  returnvalue:  errOK_els00           everything went fine
                errors
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls27ExtractTableData(tls00_DBInfo      *DBInfo,
                     tls00_DextractRec *DextRec,
                     tls00_DatExtrCmd  *DextCmd,
                     tsp00_Addr        pszErrText);


/*!
  -----------------------------------------------------------------------------
 
  function:     ls27BuildLoadCtrlFile

  -----------------------------------------------------------------------------

  description:  Creates load command file.
                Only called if user specified
                'dataextract for dataload/fastload/dataupdate ...'.
                
                Writes 'create table' command if table belongs to current user
                and is a basic table.
                Writes a 'dataload/...' command if table is not empty. 
                Otherwise generates a message to log file that table is empty.
                Writes 'create index' command if table has indexes.

  arguments:    DBInfo       [IN]   - pointer; db session structure
                DextRec      [IN]   - pointer; extract record structure
                DextCmd      [OUT]  - pointer; extract command structure
                pszErrText   [OUT]  - fehlertext

  returnvalue:  errOK_els00           everything went fine
                errFileOpen_els98     anlegen der kommandodatei klappt nicht
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls27BuildLoadCtrlFile(tls00_DBInfo      * DBInfo,
                      tls00_DextractRec * DextRec,
                      tls00_DatExtrCmd  * DextCmd,
                      tsp00_Addr          pszErrText);


#endif //HLS27_DEEXEC