/*! 
  -----------------------------------------------------------------------------
 
  module: hls28_TLTable.h
 
  -----------------------------------------------------------------------------
 
  responsible:  SteffenS

  special area: Replication Server
 
  description:  Definition of exported functions for TABLELOAD for single table
                
  version:      7.3.

  -----------------------------------------------------------------------------
 
  copyright:    SAP AG, 2000-2003
 
  -----------------------------------------------------------------------------


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/
#ifndef HLS28_TLTABLE
#define HLS28_TLTABLE

#include "gls00.h"

#define EXISTS_TABLE_STMT_LS28       _T("EXISTS TABLE \"%s\".\"%s\"")
#define DROP_TABLE_STMT_LS28         _T("DROP TABLE ")
#define EMPTY_TABLE_STMT_LS28        _T("SELECT * FROM \"%s\".\"%s\" WHERE ROWNO <= 1")
#define DELETE_TABLE_STMT_LS28       _T("DELETE \"%s\".\"%s\"")
#define INTO_TABLE_STMT_LS28         _T("INTO TABLE \"%s\".\"%s\"")
#define INTO_CONSTRAINT_STMT_LS28    _T("INTO CONSTRAINT \"%s\"")
#define INTO_DEFAULT_STMT_LS28       _T("INTO DEFAULT \"%s\"")
#define END_LOAD_TABLE_STMT_LS28     _T("END LOAD \"%s\"")
#define RESTORE_STRING_STMT_LS28     _T("RESTORE LONG \"%s\".\"%s\"")
#define RESTORE_DATA_STMT_LS28       _T("RESTORE TABLE \"%s\".\"%s\"")
#define RESTORE_LONG_STMT_LS28       _T("RESTORE COLUMN")


#define END_RESTORE_STMT_LS28        _T("END RESTORE")


#define STOP_LS28                    -1
#define STOP_NODATA_LS28             -2
#define UNKNOWN_TE_HEADER_LS28       -3
#define WRONG_TE_HEADER_LS28         -4

typedef enum
{
    tsaIncopyNone_ls28,
    tsaIncopySome_ls28,
    tsaIncopyAll_ls28
    
} tls28_TableSchemaAdaptMode;


typedef struct tls28_PageReadInfo
{
    tsp00_Int4 priCount_ls28;
    tsp00_Int4 priMore_ls28;
    tsp00_Addr priBuffer_ls28;
} tls28_PageReadInfo;


typedef struct tls28_DefaultInfo
{
    tsp00_Int2 diColNo_ls28;
    tsp00_Int4 diLength_ls28;
    tsp00_Addr diValue_ls28;
}
tls28_DefaultInfo;



/*!
  -----------------------------------------------------------------------------
 
  function:     ls28LoadSingleTable
  -----------------------------------------------------------------------------

  description:  Restores a single table with the data found in a data file 
                generated by TABLEEXTRACT.

  arguments:    DBInfo     [IN]  - structure that contains all information on the
                                    running session and the database connected to.
                TLCmd      [IN]  - tableload command structure that contains all
                                    necessary information for execution.
                TLRec      [IN]  - structure, holding information for executing
                                    TABLELOAD for a single table
                OwnerName  [IN]  - table owner name UTF8 encoded
                TableName  [IN]  - table name UTF8 encoded
                pszErrText [OUT] - error text

  returnvalue:  errOK_els00              - no error
                lots of errors may be
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls28LoadSingleTable(tls00_DBInfo            *pDBInfo,
                    tls00_TabLoadCmd        *TLCmd,
                    tls00_TloadRec          *TLRec,
                    SAPDB_DynamicUTF8String &OwnerName,
                    SAPDB_DynamicUTF8String &TableName,
                    tsp00_Addr               pszErrText);


/*!
  -----------------------------------------------------------------------------
 
  function:     ls28ReadDataFileHeader
  -----------------------------------------------------------------------------

  description:  Reads the header of a data file generated using TABLEEXTRACT and
                checks for corruption.
                Sets some necessary values in pTLRec: tlrSwapFromTE_ls00, tlrCount_ls00

                Does not check code type, version, date, time and if more than one
                table are in file.

  arguments:    pTLRec      [IN]   - pointer; structure, holding information for
                                      executing TABLELOAD for a single table
                pszErrText  [OUT]  - pointer; error text

  returnvalue:  errOK_els00              - no error
                errWrongTEFileType_els98 - wrong identifier in header (file was
                                           probably not generated using TE or is
                                           corrupted
                errWrongTEPageSize_els98 - page sizes of source and target system
                                           do not match
                some error from reading the file
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls28ReadDataFileHeader(tls00_TloadRec *pTLRec,
                       tsp00_Addr      pszErrText);

#endif
