/*! 
  -----------------------------------------------------------------------------
 
  module: hls29_dataupdate.h
 
  -----------------------------------------------------------------------------
 
  responsible:  Hans GeorgB

  special area: Replication Manager
 
  description:  dataupdate
                
  version:      7.3.

  -----------------------------------------------------------------------------
 
  copyright:    SAP AG, 1997-2002
 
  -----------------------------------------------------------------------------

    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/

#ifndef HLS29
#define HLS29

#include "gls00.h"

tsp00_Int4  ls29_InitDBInfo      ( tls00_DBInfo *pDBInfo ,tsp00_Addr errText );
void ls29_FreeDBInfo             ( tls00_DBInfo *pDBInfo ,tls00_DatUpdRec * pDatUpdRec );

tsp00_Int4  ls29_InitDatUpdRec   ( tls00_DBInfo *pDBInfo ,tls00_DatUpdCmd * pDatUpdCmd ,tls00_DatUpdRec *&pDatUpdRec,tsp00_Addr errText);
void ls29_FreeDatUpdRec          ( tls00_DatUpdRec *&pDatUpdRec,tsp00_Addr errText);

//tsp00_Int4  ls29_InitDatUpdCmd   ( tls00_DBInfo *pDBInfo ,tls00_DatUpdCmd * pDatUpdCmd ,tls00_DatUpdRec *&pDatUpdRec,tsp00_Addr errText);
void ls29_FreeDatUpdCmd          ( tls00_DatUpdCmd * pDatUpdCmd);

/*
  -----------------------------------------------------------------------------

  function:     ls29_ExecDatUpdCmd

  -----------------------------------------------------------------------------

  description:  main loop of dataupdate command
                terminated by:
                datasource reached eof
                more than maxerrorcount occured during processing

  arguments:    pDBInfo     [IN]    - A pointer to a structure that contains all
                                      information on the running session and the
                                      database connected to.
                pDatUpdCmd  [IN]    - A pointer to a structure that contains all
                                      information on the dataupdate command given
                                      by the user
                pDatUpdRec  [IN]    - A pointer to a structure that contains some
                                      additional information for processing the
                                      dataupdate command
                errText     [OUT]   - errText
                
  returnvalue:  errOK_els00         - in case of NO errors
                                
  -----------------------------------------------------------------------------
*/
tsp00_Int4  ls29_ExecDatUpdCmd   ( tls00_DBInfo *pDBInfo ,tls00_DatUpdCmd * pDatUpdCmd ,tls00_DatUpdRec *&pDatUpdRec,tsp00_Addr errText);

// part : data

// part : longdata

/*
  -----------------------------------------------------------------------------

  function:     ls29_GetNewLongDatPartBuf

  -----------------------------------------------------------------------------

  description:  msgtype: putval

  arguments:    pDBInfo     [IN]    - A pointer to a structure that contains all
                                      information on the running session and the
                                      database connected to.
                pDatUpdCmd  [IN]    - A pointer to a structure that contains all
                                      information on the dataupdate command given
                                      by the user
                pDatUpdRec  [IN]    - A pointer to a structure that contains some
                                      additional information for processing the
                                      dataupdate command
                errText     [OUT]  -  A pointer to a buffer where a description is
                                      available in case of errors

  returnvalue:  errOK_els00         - in case of NO errors
                errPutVal_els98     - in case something went wrong inserting the long
                some other errors may happen

-----------------------------------------------------------------------------
*/
tsp00_Int4 ls29_GetNewLongDatPartBuf ( tls00_DBInfo *pDBInfo,tls00_DatUpdCmd *pDatUpdCmd, tls00_DatUpdRec *pDatUpdRec,tsp00_Addr errText );

/*
  -----------------------------------------------------------------------------

  function:     ls29_AddLongDatPartBuf

  -----------------------------------------------------------------------------

  description:  This function builds one database record and writes it to a data part buffer.
                that will be sent by a msgtype: execute

  arguments:    pDatUpdCmd  [IN]    - A pointer to a structure that holds all user information
                                      to process the dataupdate command.
                pDatUpdRec  [IN]    - A pointer to a structure that contains some
                                      additional information for processing the
                                      dataupdate command
                pDatPartBuf [OUT]   - Data Part buffer to store the records
                bPutVal     [OUT]   - Indicates that a long column of the last 
                                      record in pDatPartBuf has to be sent
                                      using the msgtype putval :
                                      Is used by ls29_GetNewLongDatPartBuf
                errText     [OUT]   - A pointer to a buffer where a description is
                                      available in case of errors

  returnvalue:  errOK_els00         - in case of NO errors
                some other errors   - in case of errors

  -----------------------------------------------------------------------------
*/
tsp00_Int4 ls29_AddLongDatPartBuf    ( tls00_DatUpdCmd *pDatUpdCmd ,tls00_DatUpdRec *pDatUpdRec,tsp00_Addr pDatPartBuf,bool & bPutVal,tsp00_Addr errText);

tsp00_Int4 ls29_ClosePutVal          ( tls00_DBInfo *pDBInfo,tls00_DatUpdRec *pDatUpdRec, tsp00_LongDescriptor *pLongDescriptor ,tsp00_Addr errText );


// Recovery
/*
  -----------------------------------------------------------------------------

  function:     ls29_LoggingDatPartBuf

  -----------------------------------------------------------------------------

  description:  Checks a receive  packet for error. In case of NO ERROR the 
                iRecCount, bPutVal and PartBufSize in order interface packet are reinitialized.
    
  arguments:    pSqlPacket  [IN]     - packet to check
                pDatUpdRec  [IN]     - structure with information on
                                       processing the dataload command
                iRecCount   [IN/OUT] - count of rows sent to kernel for which
                                       the error check has to be performed
                iErrorPos   [OUT]    - in case of error indicating the erroneous
                                       record in packet
                errText     [OUT]    - error text

  returnvalue:  errOK_els00         - in case of NO errors
                kernel error while processing data

                In case of no errors the iRecCount is set to 0. The values for
                LastCommittedLine, Count of rejected and count of inserted rows
                are adapted.
                In case of error the function writes an error message.

  -----------------------------------------------------------------------------
*/
tsp00_Int4 ls29_LoggingDatPartBuf  ( tsp1_packet *pSqlPacket, tls00_DatUpdRec *pDatUpdRec,tsp00_Int4  &iRecCount ,tsp00_Int4 &iErrorPos,tsp00_Addr errText);

// Declarations - Application specific structures

/*!
  -----------------------------------------------------------------------------

  function:     ls29ExecDataUpdate

  -----------------------------------------------------------------------------

  description:  DataUpdate main function

  arguments:    pDBInfo     [IN]    - A pointer to a structure that contains all 
                                      information on the running session and the
                                      database connected to.
                pDatUpdCmd  [IN]    - A pointer to a structure that contains all
                                      information on the dataload command given
                                      by the user
                errText     [OUT]   - A pointer to a buffer where a description is
                                      available in case of errors
  returnvalue:  errOK_els00         - in case of NO errors
                some other errors   - in case of errors

  -----------------------------------------------------------------------------
*/
tsp00_Int4 ls29ExecDataUpdate( tls00_DatUpdCmd *pDatUpdCmd,tls00_DBInfo * pDBInfo,tsp00_Addr ErrText);

#endif /*HLS29*/