
#ifndef HLS31_ODBC_RESULT_H
#define HLS31_ODBC_RESULT_H

/*!**********************************************************************

  module: hls31_odbc_Result.h

  -----------------------------------------------------------------------

  responsible:  Hans-GeorgB and SteffenS

  special area: odbc 
  description:  
  version:      7.3.

  see also:

  -----------------------------------------------------------------------

  ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end
*/

// defines
//#define MEMTEST    1 
//#define UNICODE

// includes
#include "gls00.h"
#include "hls98msg.h"
// ConnectionSet
#include "hwd07dbc.h" 
// StatementAttribute
#include "gpr00.h"
// ResultSet
#ifndef UNICODE
#include "SAPDB/SQLStudio/OAL/StudioOAL_AResult.hpp"  // Ascii   Version using A Version of ODBC calls
#endif
#ifdef UNICODE
#include "SAPDB/SQLStudio/OAL/StudioOAL_WResult.hpp"  // Unicode Version using W Version of ODBC calls
#endif
// Table
#include "SQLStudio/TableDefinition/StudioTD_TableDef.h"

// forward declarations

#ifdef UNICODE
class ODBC_WResult          ;
#endif
#ifndef UNICODE
class ODBC_AResult          ;
#endif

// classes derived from ODBC_AResult resp. ODBC_WResult
class CatalogExtractDefODBC   ;
class CatalogLoadDefODBC      ;
class CatalogObjectDefODBC    ;
class UserModeDefODBC         ;
class CURUserDefODBC          ;
class UserDefODBC             ;
class RoleDefODBC             ;
class TableDefODBC            ;
class ViewDefODBC             ;
class SynonymDefODBC          ;
class TriggerDefODBC          ;
class DBProcedureDefODBC      ;
class GrantDBProcedureDefODBC ;
class IndexDefODBC            ;
class IndexColumnDefODBC      ;
class ColumnTypeDefODBC       ;
class SequenceDefODBC         ;
class DataExtractDefODBC      ;

class IndexColumnDef          ;
class sapdb_Error             ;

/*===========================================================================*
 *    class CatalogDefODBC 
 *===========================================================================*/
class CatalogDefODBC 
{
public:
     CatalogDefODBC();
    ~CatalogDefODBC();

    // Catalog
    void              setSqlMode       ( tpr00_SQLModeEnum catSqlMode ){ m_catSqlMode  = catSqlMode;}
    tpr00_SQLModeEnum getSqlMode       (){ return m_catSqlMode;}

    void            setExtrMode        ( tls31_ExtrMode   catextrMode){ m_catExtrMode = catextrMode;}
    SAPDB_Int4      getExtrMode        (){ return m_catExtrMode;}

    void            setPartSpec        ( tls00_PartSpec & catpartSpec ){m_catPartSpec = catpartSpec;}
    tls00_PartSpec  getPartSpec        (){ return m_catPartSpec ;}

    void            setMaxErrorCount   (SAPDB_Int4 lMaxErrorCount){m_catMaxAdmissibleErrors = lMaxErrorCount;}
    SAPDB_Int4      getMaxErrorCount   (){return m_catMaxAdmissibleErrors;}

    // class Database
    SAPDB_Bool      Connect            ( tls00_DBInfo *pDBSession     );
    SAPDB_Bool      Disconnect         ();
    void            setODBCConnectionPool( sapdbwa_DBCPool & odbcConnectionPool ){ m_odbcConnectionPool = odbcConnectionPool; }
    sapdbwa_DBCPool getODBCConnectionPool(){ return m_odbcConnectionPool ;}
    SAPDB_Bool      setODBCConnection  ( sapdbwa_DBC & odbcConnection );
    sapdbwa_DBC     getODBCConnection  (){ return m_odbcConnection; }
    SAPDB_Bool      setODBCDriver      ( tls00_DBInfo *pDBSession     );
    StudioOAL_String getOdbcDriver     (){ return m_odbcDriver    ; }
    // DataBase Authorization
    void             setCurrentUser     ( StudioOAL_String user );
    StudioOAL_String getCurrentUser     ( );

    sapdb_Error *   getError           () { return m_pError; }

    // class File
    void            setFileConnection  ( tls00_FileInfo  &FileSession ){ m_FileSession = FileSession ;}
    tls00_FileInfo  getFileConnection  (){ return m_FileSession ;}
    SAPDB_Bool      readLine           ( StudioOAL_String &record  );
    SAPDB_Bool      writeLine          ( StudioOAL_String  record  );


    CatalogExtractDefODBC    * getCatalogExtractDefODBC    () { return m_pCatalogExtractDefODBC;}
    CatalogLoadDefODBC       * getCatalogLoadDefODBC       () { return m_pCatalogLoadDefODBC;}
    UserModeDefODBC          * getUserModeDefODBC          () { return m_pUserModeDefODBC;}
    CURUserDefODBC           * getCURUserDefODBC           () { return m_pCURUserDefODBC;}

    // Catalog Objects
    UserDefODBC              * getUserDefODBC              () { return m_pUserDefODBC;}
    RoleDefODBC              * getRoleDefODBC              () { return m_pRoleDefODBC;}
    TableDefODBC             * getTableDefODBC             () { return m_pTableDefODBC;}
    TriggerDefODBC           * getTriggerDefODBC           () { return m_pTriggerDefODBC;}
    ViewDefODBC              * getViewDefODBC              () { return m_pViewDefODBC;}
    SynonymDefODBC           * getSynonymDefODBC           () { return m_pSynonymDefODBC;}
    ColumnTypeDefODBC        * getColumnTypeDefODBC        () { return m_pColumnTypeDefODBC;}
    IndexDefODBC             * getIndexDefODBC             () { return m_pIndexDefODBC;}
    SequenceDefODBC          * getSequenceDefODBC          () { return m_pSequenceDefODBC;}
    DBProcedureDefODBC       * getDBProcedureDefODBC       () { return m_pDBProcedureDefODBC;}
    GrantDBProcedureDefODBC  * getGrantDBProcedureDefODBC  () { return m_pGrantDBProcedureDefODBC;}
    
    // CatalogExtract Commands

    // CatalogExtract all
    SAPDB_Int4 ExtrCatalogCmd           ( StudioOAL_String user );
    // CatalogExtract user
    SAPDB_Int4 ExtrSchemaCmd            ( StudioOAL_String user );
    // CatalogExtract table
    SAPDB_Int4 ExtrTableCmd             ( tls00_PartSpec  catpartSpec );

    // CatalogLoad   Commands

    // CatalogLoad all
    SAPDB_Int4 LoadCatalogCmd           ( StudioOAL_String user );
    // CatalogLoad user
    SAPDB_Int4 LoadSchemaCmd            ( StudioOAL_String user );
    // CatalogLoad table
    SAPDB_Int4 LoadTableCmd             ( tls00_PartSpec  catpartSpec );
    
    // CatalogObjectExtract Commands

    SAPDB_Bool writeCatalogHeader       ( );    

    // Extract Authorization Objects from System Views DOMAIN 
    SAPDB_Int4 schemaExtrCmd            ( StudioOAL_String user  );
    SAPDB_Int4 userExtrCmd              ( tls00_PartSpec catpartSpec );
    SAPDB_Int4 roleExtrCmd              ( tls00_PartSpec catpartSpec );
        
    // Extract Catalog Objects from System Views DOMAIN 
    SAPDB_Int4 connectionExtrCmd        ( );
    SAPDB_Int4 columntypesetExtrCmd     ( );
    
    // Meta Data
    SAPDB_Int4 tabletypeExtrCmd         ( tls00_PartSpec  catpartSpec );
    
    SAPDB_Int4 tableExtrCmdExecDirect   ( StudioOAL_String owner, StudioOAL_String tablename );
    SAPDB_Int4 tableExtrCmdExec         ( StudioOAL_String owner, StudioOAL_String tablename );
    SAPDB_Int4 viewExtrCmdExecDirect    ( StudioOAL_String owner, StudioOAL_String viewname  );
    SAPDB_Int4 viewExtrCmdExec          ( StudioOAL_String owner, StudioOAL_String viewname  );
    SAPDB_Int4 synonymExtrCmdExecDirect ( StudioOAL_String owner, StudioOAL_String tablename );
    SAPDB_Int4 synonymExtrCmdExec       ( StudioOAL_String owner, StudioOAL_String tablename );
    SAPDB_Int4 indexExtrCmdExecDirect   ( StudioOAL_String owner, StudioOAL_String tablename );
    SAPDB_Int4 indexExtrCmdExec         ( StudioOAL_String owner, StudioOAL_String tablename );
    SAPDB_Int4 triggerExtrCmdExecDirect ( StudioOAL_String owner, StudioOAL_String tablename );
    SAPDB_Int4 triggerExtrCmdExec       ( StudioOAL_String owner, StudioOAL_String tablename );

    SAPDB_Int4 sequenceExtrCmd          ( tls00_PartSpec  catpartSpec );
    SAPDB_Int4 dbprocedureExtrCmd       ( tls00_PartSpec  catpartSpec );

    // Data
    SAPDB_Int4 dataExtrCmdExec          ( StudioOAL_String owner, StudioOAL_String tablename );
    DataExtractDefODBC                   *m_pDataExtractDefODBC;
    
    SAPDB_Bool writeCatalogFooter       ( );
    
private:

    // Init CatalogObjects 
    SAPDB_Bool initCatalogObjects       ( sapdbwa_DBC & odbcConnection);

    CatalogExtractDefODBC    *m_pCatalogExtractDefODBC;
    CatalogLoadDefODBC       *m_pCatalogLoadDefODBC;
    UserModeDefODBC          *m_pUserModeDefODBC;
    CURUserDefODBC           *m_pCURUserDefODBC;

    UserDefODBC              *m_pUserDefODBC;
    RoleDefODBC              *m_pRoleDefODBC;
    TableDefODBC             *m_pTableDefODBC;
    TriggerDefODBC           *m_pTriggerDefODBC;
    ViewDefODBC              *m_pViewDefODBC;
    SynonymDefODBC           *m_pSynonymDefODBC;
    ColumnTypeDefODBC        *m_pColumnTypeDefODBC;
    IndexDefODBC             *m_pIndexDefODBC;
    SequenceDefODBC          *m_pSequenceDefODBC;
    DBProcedureDefODBC       *m_pDBProcedureDefODBC;
    GrantDBProcedureDefODBC  *m_pGrantDBProcedureDefODBC;
        
    SAPDB_Int4                m_nSchemaCount;
    SAPDB_Int4                m_nTableCount;
    SAPDB_Int4                m_nViewCount;
    SAPDB_Int4                m_nSynonymCount;
    
    sapdb_Error              *m_pError;

    // Database Connection

    // ODBC Connection
    StudioOAL_String   m_odbcDriver;
    
    sapdbwa_DBCPool   m_odbcConnectionPool;
    sapdbwa_DBC       m_odbcConnection;     

    // OI   Connection hwd04wases.h  
    // sapdbwa_SessionPool   m_oiConnectionPool;
    // sapdbwa_DbConnection  m_oiConnection;     

    // File Connection
    tls00_FileInfo    m_FileSession;

    tls31_ExtrMode    m_catExtrMode;
    tls00_PartSpec    m_catPartSpec;
    tpr00_SQLModeEnum m_catSqlMode;
    SAPDB_Int4        m_catMaxAdmissibleErrors;


    StudioOAL_String   m_currentUser;
};
#ifdef UNICODE
/*===========================================================================*
 *    class ODBC_WResult
 *===========================================================================*/
class ODBC_WResult :  public StudioOAL_WResult
{
public:

    ODBC_WResult ( sapdbwa_DBC & odbcConnection ); 
    virtual ~ODBC_WResult ();
    
};
#endif
#ifndef UNICODE
/*===========================================================================*
 *    class ODBC_AResult
 *===========================================================================*/
class ODBC_AResult :  public StudioOAL_AResult
{
public:

    ODBC_AResult ( sapdbwa_DBC & odbcConnection ); 
    virtual ~ODBC_AResult ();
    
};
#endif
/*===========================================================================*
 *    class CatalogObjectDefODBC 
 *===========================================================================*/
class CatalogObjectDefODBC : 
#ifdef  UNICODE
                       public ODBC_WResult
#endif
#ifndef UNICODE
                       public ODBC_AResult
#endif
{
public:
    CatalogObjectDefODBC( sapdbwa_DBC & odbcConnection, CatalogDefODBC * pCatalog );
    CatalogObjectDefODBC( sapdbwa_DBC & odbcConnection );
    
    virtual ~CatalogObjectDefODBC();

    void               setCatalog( CatalogDefODBC * pCatalog ){ m_pCatalogDefODBC = pCatalog;}
    CatalogDefODBC   * getCatalog(){ return m_pCatalogDefODBC  ;}
        
    virtual SAPDB_Bool initResultSet ( sapdbwa_DBC & odbcConnection ) = 0;
        
    SAPDB_Bool executeSQL            ( StudioOAL_String& statement);
    SAPDB_Bool readResult            ( StudioOAL_String *record  );
    SAPDB_Bool writeLine             ( StudioOAL_String  record  );

    StudioOAL_String  string2XML        ( StudioOAL_String record  );
    StudioOAL_String& EscapeSingleQuotes( StudioOAL_String& string );
        
    SAPDB_Bool filterSQLExceptions   ( StudioOAL_String nativeError, StudioOAL_String errorText);

private:

    // Catalog
    CatalogDefODBC    *m_pCatalogDefODBC;

};
/*===========================================================================*
 *    class sapdb_Error
 *===========================================================================*/
class sapdb_Error
{
public:
     sapdb_Error();
    ~sapdb_Error();

    inline void             setErrorText( StudioOAL_String errorText ) { m_ErrorText = errorText; }
    inline StudioOAL_String getErrorText()                             { return m_ErrorText ; }
    inline void             setReturnCode( SQLRETURN ReturnCode )      { m_ReturnCode = ReturnCode; }
    inline SQLRETURN        getReturnCode()                            { return m_ReturnCode; }
    inline void             isError( SAPDB_Bool isError )              { m_isError = isError; }
    inline SAPDB_Bool       isError()                                  { return m_isError; }

    SAPDB_Bool       m_isError;
    SAPDB_Long       m_NativeError;
    SAPDB_Int4       m_ErrorPos;
    SQLRETURN        m_ReturnCode;
    StudioOAL_String m_ErrorText;
    StudioOAL_String m_NativeErrorAsString;
};
/*===========================================================================*
 *    class CatalogExtractDefODBC 
 *===========================================================================*/
class CatalogExtractDefODBC : public CatalogObjectDefODBC
{
public:
   CatalogExtractDefODBC( sapdbwa_DBC & odbcConnection , CatalogDefODBC * pCatalog );
   CatalogExtractDefODBC( sapdbwa_DBC & odbcConnection  );
  ~CatalogExtractDefODBC();

   SAPDB_Bool initResultSet ( sapdbwa_DBC & odbcConnection );
   SAPDB_Bool readResult    ( StudioOAL_String *record );
   SAPDB_Bool writeResult   ( StudioOAL_String  record );

   SAPDB_Int4  getObjectCount() { return m_nObjectCount; }

private:

   SAPDB_Int4      m_nObjectCount;
   SAPDB_Int4      m_nTableCount;
   SAPDB_Int4      m_nViewCount;
   SAPDB_Int4      m_nSynonymCount;

   StudioOAL_String m_servernode;
   StudioOAL_String m_dbname;
   StudioOAL_String m_username;
   
};
/*===========================================================================*
 *    class CatalogLoadDefODBC 
 *===========================================================================*/
class CatalogLoadDefODBC    : public CatalogObjectDefODBC
{
public:
   CatalogLoadDefODBC( sapdbwa_DBC & odbcConnection, CatalogDefODBC * pCatalog );
   CatalogLoadDefODBC( sapdbwa_DBC & odbcConnection );
  ~CatalogLoadDefODBC();

   void             setSQLMode ( StudioOAL_String  record );
   StudioOAL_String getSQLMode ( );

   SAPDB_Bool initResultSet    ( sapdbwa_DBC & odbcConnection );
   SAPDB_Bool readResult       ( StudioOAL_String *record );
   SAPDB_Bool writeResult      ( StudioOAL_String  record );

private:

   StudioOAL_String m_servernode;
   StudioOAL_String m_dbname;
   StudioOAL_String m_username;
   StudioOAL_String m_sqlmode;

};

// CatalogObjects created by Autorization Definition Statements

/*===========================================================================*
 *    class UserDefODBC 
 *===========================================================================*/
class UserDefODBC :  public CatalogObjectDefODBC
{
public:
   UserDefODBC( sapdbwa_DBC & odbcConnection, CatalogDefODBC * pCatalog );
   UserDefODBC( sapdbwa_DBC & odbcConnection );
  ~UserDefODBC();

   SAPDB_Bool initResultSet ( sapdbwa_DBC & odbcConnection );
   SAPDB_Bool readResult    ( StudioOAL_String *record  );
   SAPDB_Bool writeResult   ( StudioOAL_String *record ,tls31_ExtrMode catextrMode );

private:

   StudioOAL_String m_owner;
   StudioOAL_String m_groupName;
   StudioOAL_String m_userName;
   StudioOAL_String m_userMode;
   StudioOAL_String m_maxTimeOut;
   StudioOAL_String m_costWarning;
   StudioOAL_String m_costLimit;
   StudioOAL_String m_connectMode;
   StudioOAL_String m_comment;

};
/*===========================================================================*
 *    class UserModeDefODBC 
 *===========================================================================*/
class UserModeDefODBC :  public CatalogObjectDefODBC
{
public:
   UserModeDefODBC( sapdbwa_DBC & odbcConnection, CatalogDefODBC * pCatalog );
   UserModeDefODBC( sapdbwa_DBC & odbcConnection );
  ~UserModeDefODBC();

   SAPDB_Bool initResultSet ( sapdbwa_DBC & odbcConnection );
   SAPDB_Bool readResult    ( StudioOAL_String *record      );

   StudioOAL_String  getUserNameDBA() { return m_usernameDBA; }
   StudioOAL_String  getUserNameDBM() { return m_usernameDBM; }

private:

   StudioOAL_String m_usernameDBA;
   StudioOAL_String m_usernameDBM;
 

};
/*===========================================================================*
 *    class CURUserDefODBC 
 *===========================================================================*/
class CURUserDefODBC :  public CatalogObjectDefODBC
{
public:
   CURUserDefODBC( sapdbwa_DBC & odbcConnection, CatalogDefODBC * pCatalog );
   CURUserDefODBC( sapdbwa_DBC & odbcConnection );
  ~CURUserDefODBC();

   SAPDB_Bool initResultSet ( sapdbwa_DBC & odbcConnection );
   SAPDB_Bool writeResult   ( StudioOAL_String  record      );

private:

   StudioOAL_String  m_username;

};
/*===========================================================================*
 *    class RoleDefODBC 
 *===========================================================================*/
class RoleDefODBC :  public CatalogObjectDefODBC 
{
public:
   RoleDefODBC( sapdbwa_DBC & odbcConnection, CatalogDefODBC * pCatalog );
   RoleDefODBC( sapdbwa_DBC & odbcConnection );
  ~RoleDefODBC();

   SAPDB_Bool initResultSet ( sapdbwa_DBC & odbcConnection );
   SAPDB_Bool readResult    ( StudioOAL_String *record  );
   SAPDB_Bool writeResult   ( StudioOAL_String *record  );

private:

    StudioOAL_String m_role;
    StudioOAL_String m_passwordrequired;

};

// CatalogObjects created by Data Definition Statements

/*===========================================================================*
 *    class TableDefODBC 
 *===========================================================================*/
class TableDefODBC :  public CatalogObjectDefODBC ,
                      public StudioTD_TableDef      
{
public:
   TableDefODBC( sapdbwa_DBC & odbcConnection, CatalogDefODBC * pCatalog );
   TableDefODBC( sapdbwa_DBC & odbcConnection );
  ~TableDefODBC();

   // 1.Direct SQL Statement

   SAPDB_Int4 readTableDefinitionDirect ( StudioOAL_String owner, StudioOAL_String tablename );

   SAPDB_Bool executeSQL                ( StudioOAL_String statement );
   SAPDB_Bool readResult                ( StudioOAL_String *record , SAPDB_Bool* bIsNull );
   
   // 2. Prepared SQL Statement

   SAPDB_Bool initResultSet             ( sapdbwa_DBC & odbcConnection );

   SAPDB_Bool readTableDefinition       ( StudioOAL_String owner, StudioOAL_String tablename );
   SAPDB_Bool writeTableDefinition( );
   SAPDB_Bool readColumnDefResult       ( StudioOAL_String *record ) ;
   SAPDB_Bool readConstraintDefResult   ( StudioOAL_String *record ) ;
   SAPDB_Bool readForeignKeyDefResult   ( StudioOAL_String *record ) ;
   SAPDB_Bool readTableAttrDefResult    ( StudioOAL_String *record ) ;

   StudioOAL_String   genCreateStatement();
   
private:

#ifdef UNICODE
   ODBC_WResult     *m_pColumnDefResult;     
   ODBC_WResult     *m_pConstraintDefResult; 
   ODBC_WResult     *m_pForeignKeyDefResult; 
   ODBC_WResult     *m_pTableAttrDefResult;
#endif
#ifndef UNICODE
   ODBC_AResult     *m_pColumnDefResult;     
   ODBC_AResult     *m_pConstraintDefResult; 
   ODBC_AResult     *m_pForeignKeyDefResult; 
   ODBC_AResult     *m_pTableAttrDefResult; 
#endif

   StudioOAL_String   m_tableName ;
   StudioOAL_String   m_owner;
   
};
/*===========================================================================*
 *    class IndexDefODBC 
 *===========================================================================*/
class IndexDefODBC      : public CatalogObjectDefODBC
{
public:
   IndexDefODBC( sapdbwa_DBC & odbcConnection, CatalogDefODBC * pCatalog );
   IndexDefODBC( sapdbwa_DBC & odbcConnection );
  ~IndexDefODBC();

   SAPDB_Bool setODBCConnection            ( sapdbwa_DBC & odbcConnection) ;

   // 1.Direct SQL Statement

   SAPDB_Bool readIndexDefinitionDirect    ( StudioOAL_String owner, StudioOAL_String tablename);
   SAPDB_Bool readIndexColDefinitionDirect ( StudioOAL_String owner, StudioOAL_String tablename, StudioOAL_String indexname );

   // 2. Prepared SQL Statement

   SAPDB_Bool initResultSet                ( sapdbwa_DBC & odbcConnection );
   SAPDB_Bool readIndexDefResult           ( StudioOAL_String * record );
   SAPDB_Bool readIndexColDefResult        ( StudioOAL_String * record );

   SAPDB_Bool setIndexParameters           ( StudioOAL_String owner, StudioOAL_String tablename );
   SAPDB_Bool readIndexDefinition          ( StudioOAL_String owner, StudioOAL_String tablename );
   SAPDB_Bool readIndexColDefinition       ( StudioOAL_String owner, StudioOAL_String tablename, StudioOAL_String indexname );
   
   SAPDB_Bool writeIndexDefinition         ( );

   StudioOAL_String      genCreateStatement( );
   StudioOAL_String      genCreateXML();
   
   IndexColumnDef       *addColumn ();
   void                  reset     ();


   IndexColumnDefODBC   *m_pIndexColDefODBC;
   StudioOAL_List        m_IndexColumns; 

private:

#ifdef UNICODE
   ODBC_WResult     *m_pIndexDefResult;  
   ODBC_WResult     *m_pIndexColDefResult;
#endif
   
#ifndef UNICODE
   ODBC_AResult     *m_pIndexDefResult;  
   ODBC_AResult     *m_pIndexColDefResult;
#endif

   StudioOAL_String  m_tableName ;
   StudioOAL_String  m_owner;
   StudioOAL_String  m_indexName;
   StudioOAL_String  m_indexType;
   StudioOAL_String  m_comment;
};
/*===========================================================================*
 *    class IndexColumnDefODBC 
 *===========================================================================*/
class IndexColumnDefODBC : public CatalogObjectDefODBC
{
public:
    IndexColumnDefODBC( sapdbwa_DBC & odbcConnection, CatalogDefODBC * pCatalog );
    IndexColumnDefODBC( sapdbwa_DBC & odbcConnection );
   ~IndexColumnDefODBC();

   SAPDB_Bool initResultSet( sapdbwa_DBC & odbcConnection );

   void              setIndexDefODBC( IndexDefODBC *pIndexDefODBC) { m_pIndexDefODBC = pIndexDefODBC;}  
   IndexDefODBC     *getIndexDefODBC() { return m_pIndexDefODBC;}

   IndexDefODBC      *m_pIndexDefODBC;
};
/*===========================================================================*
 *    class IndexColumnDef
 *===========================================================================*/
class IndexColumnDef 
{
public:
   IndexColumnDef() {;}
   ~IndexColumnDef(){;}

   StudioOAL_String   genCreateStatement();
   StudioOAL_String   genCreateXML();

   void              setOwner     ( StudioOAL_String owner     ){m_owner     = owner    ;}
   void              setTableName ( StudioOAL_String tableName ){m_tableName = tableName;}
   void              setIndexName ( StudioOAL_String indexName ){m_indexName = indexName;}
   void              setIndexType ( StudioOAL_String indextype ){m_indexType = indextype;}
   void              setSorting   ( StudioOAL_String sorting   ){m_sorting   = sorting  ;}
   StudioOAL_String   getSorting   ( ){ return m_sorting ;}
   void              setComment   ( StudioOAL_String comment );
   StudioOAL_String   getComment   ( ){ return m_comment ;}

   StudioTD_ColumnDef m_column;

   StudioOAL_String   m_owner;
   StudioOAL_String   m_tableName ;
   StudioOAL_String   m_indexName;
   StudioOAL_String   m_indexType;
   StudioOAL_String    m_sorting;
   StudioOAL_String    m_comment;
};
/*===========================================================================*
 *    class SynonymDefODBC 
 *===========================================================================*/
class SynonymDefODBC  :  public CatalogObjectDefODBC
{
public:
   SynonymDefODBC ( sapdbwa_DBC & odbcConnection, CatalogDefODBC * pCatalog );
   SynonymDefODBC ( sapdbwa_DBC & odbcConnection );
  ~SynonymDefODBC ();

   // 1.Direct SQL Statement

   SAPDB_Bool readResult    ( StudioOAL_String *record );
   SAPDB_Bool writeResult   ( StudioOAL_String *record );

   // 2. Prepared SQL Statement
   
   SAPDB_Bool initResultSet        ( sapdbwa_DBC & odbcConnection );

   SAPDB_Bool readSynonymDefinition( StudioOAL_String owner, StudioOAL_String tablename );
   SAPDB_Bool writeSynonymDefinition( );
   SAPDB_Bool readSynonymDefResult ( StudioOAL_String *record ) ;

   StudioOAL_String     genCreateXML ( );
   StudioOAL_String     genCreateDDL ( );

private:

#ifdef UNICODE
   ODBC_WResult*     m_pSynonymDefResult ;
#endif
#ifndef UNICODE
   ODBC_AResult*     m_pSynonymDefResult ;
#endif

   StudioOAL_String  m_synonymName ;
   StudioOAL_String  m_owner;
   StudioOAL_String  m_synonymDefinition;

   StudioOAL_String  m_tablename;
   StudioOAL_String  m_tableowner;
   StudioOAL_String  m_public;
   StudioOAL_String  m_comment;

};
/*===========================================================================*
 *    class ViewDefODBC 
 *===========================================================================*/
class ViewDefODBC :  public CatalogObjectDefODBC
{
public:
   ViewDefODBC( sapdbwa_DBC & odbcConnection, CatalogDefODBC * pCatalog );
   ViewDefODBC( sapdbwa_DBC & odbcConnection );
  ~ViewDefODBC();

   // 1.Direct SQL Statement

   SAPDB_Bool readResult    ( StudioOAL_String *record );
   SAPDB_Bool writeResult   ( StudioOAL_String *record );

   // 2. Prepared SQL Statement
  
   SAPDB_Bool initResultSet    ( sapdbwa_DBC & odbcConnection );

   SAPDB_Bool readViewDefinition ( StudioOAL_String owner, StudioOAL_String tablename );
   SAPDB_Bool writeViewDefinition( );
   SAPDB_Bool readViewDefResult  ( StudioOAL_String *record ) ;

   StudioOAL_String  genCreateXML ( );

private:

#ifdef UNICODE
   ODBC_WResult*     m_pViewDefResult ;
#endif
#ifndef UNICODE
   ODBC_AResult*     m_pViewDefResult ;
#endif

   StudioOAL_String  m_viewName ;
   StudioOAL_String  m_owner;
   StudioOAL_String  m_viewDefinition;

};
/*===========================================================================*
 *    class TriggerDefODBC 
 *===========================================================================*/
class TriggerDefODBC      : public CatalogObjectDefODBC
{
public:
   TriggerDefODBC( sapdbwa_DBC & odbcConnection, CatalogDefODBC * pCatalog );
   TriggerDefODBC( sapdbwa_DBC & odbcConnection );
  ~TriggerDefODBC();

   // 1.Direct SQL Statement

   SAPDB_Bool readResult    (  StudioOAL_String *record );
   SAPDB_Bool writeResult   (  StudioOAL_String *record );

   // 2. Prepared SQL Statement

   SAPDB_Bool initResultSet ( sapdbwa_DBC & odbcConnection );

   SAPDB_Bool readTriggerDefinition( StudioOAL_String owner, StudioOAL_String tablename );
   SAPDB_Bool writeTriggerDefinition();
   SAPDB_Bool readTriggerDefResult ( StudioOAL_String *record ) ;
      
   StudioOAL_String       genCreateXML();
   StudioOAL_String       genCreateStatement ();

   void       reset();
 
private:

#ifdef UNICODE
   ODBC_WResult*     m_pTriggerDefResult ;
#endif
#ifndef UNICODE
   ODBC_AResult*     m_pTriggerDefResult ;
#endif

   StudioOAL_String  m_triggerName;
   StudioOAL_String  m_owner;
   StudioOAL_String  m_triggerDefinition;

   StudioOAL_String  m_tableName ;
   StudioOAL_String  m_comment;
};
/*===========================================================================*
 *    class DBProcedureDefODBC 
 *===========================================================================*/
class DBProcedureDefODBC  : public CatalogObjectDefODBC
{
public:
   DBProcedureDefODBC( sapdbwa_DBC & odbcConnection, CatalogDefODBC * pCatalog );
   DBProcedureDefODBC( sapdbwa_DBC & odbcConnection );
  ~DBProcedureDefODBC();

   SAPDB_Bool setODBCConnection ( sapdbwa_DBC & odbcConnection );

   // 1.Direct SQL Statement

   SAPDB_Bool readResult        ( StudioOAL_String *record );
   SAPDB_Bool writeResult       ( StudioOAL_String *record );

   // 2. Prepared SQL Statement

   SAPDB_Bool initResultSet     ( sapdbwa_DBC & odbcConnection );
  
   StudioOAL_String        genCreateXML();

   SAPDB_Int4 grantDBProcedureExtrCmd  ( StudioOAL_String dbprocedureName );
   GrantDBProcedureDefODBC * m_pGrantDBProcedureDefODBC;
 
private:

   StudioOAL_String       m_dbprocedureName;
   StudioOAL_String       m_owner;
   StudioOAL_String        m_dbprocedureDefinition;

   StudioOAL_String       m_comment;
};
/*===========================================================================*
 *    class GrantDBProcedureDefODBC 
 *===========================================================================*/
class GrantDBProcedureDefODBC : public CatalogObjectDefODBC
{
public:
   GrantDBProcedureDefODBC( sapdbwa_DBC & odbcConnection, CatalogDefODBC * pCatalog );
   GrantDBProcedureDefODBC( sapdbwa_DBC & odbcConnection );
  ~GrantDBProcedureDefODBC();

   // 1.Direct SQL Statement

   SAPDB_Bool readResult    ( StudioOAL_String *record );
   SAPDB_Bool writeResult   ( StudioOAL_String  record );

   // 2. Prepared SQL Statement

   SAPDB_Bool initResultSet ( sapdbwa_DBC & odbcConnection );
   
private:

   StudioOAL_String        m_grantdbprocedureDefinition;

};
/*===========================================================================*
 *    class ColumnTypeDefODBC
 *===========================================================================*/
class ColumnTypeDefODBC : public CatalogObjectDefODBC
{
public:
    ColumnTypeDefODBC( sapdbwa_DBC & odbcConnection, CatalogDefODBC * pCatalog );
    ColumnTypeDefODBC( sapdbwa_DBC & odbcConnection );
   ~ColumnTypeDefODBC();

   SAPDB_Bool setODBCConnection  ( sapdbwa_DBC & odbcConnection ) ;

   SAPDB_Bool initResultSet      ( sapdbwa_DBC & odbcConnection );
   SAPDB_Bool readResult         ( StudioOAL_String *record );
   SAPDB_Bool writeResult        ( StudioOAL_String *record );

private:

   StudioOAL_String m_typeName;
   StudioOAL_String m_dataType;
   StudioOAL_String m_precision;
   StudioOAL_String m_literalPrefix;
   StudioOAL_String m_literalSuffix;
   StudioOAL_String m_createParams;
   StudioOAL_String m_isNullable;
   StudioOAL_String m_isCaseSensitive;
   StudioOAL_String m_isSearchable;
   StudioOAL_String m_unsignedAttribute;
   StudioOAL_String m_money;
   StudioOAL_String m_autoIncrement;
   StudioOAL_String m_localTypeName;
   StudioOAL_String m_minimumScale;
   StudioOAL_String m_maximumScale;

};
/*===========================================================================*
 *    class SequenceDefODBC 
 *===========================================================================*/
class SequenceDefODBC :  public CatalogObjectDefODBC
{
public:
   SequenceDefODBC( sapdbwa_DBC & odbcConnection, CatalogDefODBC * pCatalog );
   SequenceDefODBC( sapdbwa_DBC & odbcConnection );
  ~SequenceDefODBC();
 
   SAPDB_Bool initResultSet ( sapdbwa_DBC & odbcConnection );
   SAPDB_Bool readResult    ( StudioOAL_String *record  );
   SAPDB_Bool writeResult   ( StudioOAL_String *record  );

private:

   StudioOAL_String m_sequencename;
   StudioOAL_String m_owner;
   StudioOAL_String m_incrementby;
   StudioOAL_String m_lastnumber;
   StudioOAL_String m_maxvalue;
   StudioOAL_String m_minvalue;
   StudioOAL_String m_cycleflag;
   StudioOAL_String m_cachesize;
   StudioOAL_String m_orderflag;
   
};

/*===========================================================================*
 *    class DataExtractDefODBC 
 *===========================================================================*/
class DataExtractDefODBC      : public CatalogObjectDefODBC
{
public:
   DataExtractDefODBC( sapdbwa_DBC & odbcConnection, CatalogDefODBC * pCatalog );
   DataExtractDefODBC( sapdbwa_DBC & odbcConnection );
  ~DataExtractDefODBC();

   // 1.Direct SQL Statement

   SAPDB_Bool readResult    ( StudioOAL_String *record );
   SAPDB_Bool writeResult   ( StudioOAL_String *record );

   // 2. Prepared SQL Statement

   SAPDB_Bool initResultSet ( sapdbwa_DBC & odbcConnection );

#ifdef UNICODE
   ODBC_WResult*     m_pDataExtractDefResult ;
#endif
#ifndef UNICODE
   ODBC_AResult*     m_pDataExtractDefResult ;
#endif

   StudioOAL_String  m_owner;
   StudioOAL_String  m_tableName ;
};

#endif


/*===========================================================================*
 *    class TableExtractDefODBC 
 *===========================================================================*/
class TableExtractDefODBC : public CatalogObjectDefODBC
{
public:
//   TableExtractDefODBC( sapdbwa_DBC & odbcConnection , CatalogDefODBC * pCatalog );
    TableExtractDefODBC( sapdbwa_DBC & odbcConnection, tls00_PartSpec *pPartSpec );
    ~TableExtractDefODBC();

    SAPDB_Bool initResultSet ( sapdbwa_DBC & odbcConnection );
//   SAPDB_Bool readResult    ( StudioOAL_String *record );
//   SAPDB_Bool writeResult   ( StudioOAL_String  record );

    SAPDB_Int4  getObjectCount() { return m_nTableCount; }
    SAPDB_Bool  GetOwnerTablePair( sapdbwa_DBC & odbcConnection,
                                       StudioOAL_String *pUserName,
                                       StudioOAL_String *pTableName);

private:

//   SAPDB_Int4      m_nObjectCount;
   SAPDB_Int4      m_nTableCount;
//   SAPDB_Int4      m_nViewCount;
//   SAPDB_Int4      m_nSynonymCount;

//   StudioOAL_String m_servernode;
//   StudioOAL_String m_dbname;
   StudioOAL_String m_username;
   tls00_PartSpec   *m_pPartSpec;
   
};

/*===========================================================================*
 *     End
 *===========================================================================*/
