/*! 
  -----------------------------------------------------------------------------
 
  module: vls05.cpp
 
  -----------------------------------------------------------------------------
 
  responsible:  SteffenS

  special area: Replication Manager
 
  description:  String functions
                
  version:      7.3.

  -----------------------------------------------------------------------------
 
  copyright:    SAP AG, 1997-2002
 
  -----------------------------------------------------------------------------


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/

#include <ctype.h>

#include "gls00.h"
#include "hls05.h"
#include "hls07.h"
#include "hls98msg.h"
#include "hsp78.h"
#include "hsp77.h"

#define _INTERNAL_FILE_ID_    _T("05")


static int
lLatin1_ToUpperMap_ls05[256] =
{
 0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,
 0x08,0x09,0x0A,0x0B,0x0C,0x0D,0x0E,0x0F,
 0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,
 0x18,0x19,0x1A,0x1B,0x1C,0x1D,0x1E,0x1F,
 0x20,0x21,0x22,0x23,0x24,0x25,0x26,0x27,
 0x28,0x29,0x2A,0x2B,0x2C,0x2D,0x2E,0x2F,
 0x30,0x31,0x32,0x33,0x34,0x35,0x36,0x37,
 0x38,0x39,0x3A,0x3B,0x3C,0x3D,0x3E,0x3F,
 0x40,0x41,0x42,0x43,0x44,0x45,0x46,0x47,
 0x48,0x49,0x4A,0x4B,0x4C,0x4D,0x4E,0x4F,
 0x50,0x51,0x52,0x53,0x54,0x55,0x56,0x57,
 0x58,0x59,0x5A,0x5B,0x5C,0x5D,0x5E,0x5F,
 0x60,0x41,0x42,0x43,0x44,0x45,0x46,0x47,
 0x48,0x49,0x4A,0x4B,0x4C,0x4D,0x4E,0x4F,
 0x50,0x51,0x52,0x53,0x54,0x55,0x56,0x57,
 0x58,0x59,0x5A,0x7B,0x7C,0x7D,0x7E,0x7F,
 0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,
 0x88,0x89,0x8A,0x8B,0x8C,0x8D,0x8E,0x8F,
 0x90,0x91,0x92,0x93,0x94,0x95,0x96,0x97,
 0x98,0x99,0x9A,0x9B,0x9C,0x9D,0x9E,0x9F,
 0xA0,0xA1,0xA2,0xA3,0xA4,0xA5,0xA6,0xA7,
 0xA8,0xA9,0xAA,0xAB,0xAC,0xAD,0xAE,0xAF,
 0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,
 0xB8,0xB9,0xBA,0xBB,0xBC,0xBD,0xBE,0xBF,
 0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,
 0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,
 0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,
 0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF,
 0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,
 0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,
 0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xF7,
 0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xFF
};

/*
  -----------------------------------------------------------------------------
  Chapter: Private functions declaration 
  -----------------------------------------------------------------------------
*/

/*
  -----------------------------------------------------------------------------

  function:     ls05_ConvertUTF8ToUCS2
  -----------------------------------------------------------------------------

  description:  Converts a string from UTF8 to UCS2.

  arguments:    pszUCS2    [OUT]     - destination string
                lUCS2Len   [IN/OUT]  - length of destination string in Bytes
                pszUTF8    [IN]      - source string
                lUTF8Len   [IN]      - length of source string in Bytes
                UCS2Type   [IN]      - type of UCS2 encoding (swapped or not)

  returnvalue:  destination string length in lUCS2Len

                return code:
                errOK_els00                 - successful conversion
                errInternal_els98           - when partial character in source,
                                                but hit end or insufficient
                                                space in target for conversion
                errConvertingEncoding_els98 - source contains invalid chars or
                                               char sequences
  -----------------------------------------------------------------------------
*/
static tsp00_Longint
ls05_ConvertUTF8ToUCS2(unsigned char       *pszUCS2,
                       tsp00_Longuint      &lUCS2Len,
                       const unsigned char *pszUTF8,
                       tsp00_Longuint       lUTF8Len,
                       tls00_CodeType       UCS2Type);


/*
  -----------------------------------------------------------------------------

  function:     ls05_ConvertUTF8ToASCII - overloaded
  -----------------------------------------------------------------------------

  description:  Converts a string from UTF8 to ASCII .

  arguments:    pszASCII     [OUT]     - destination string
                lASCIILen    [IN/OUT]  - length of destination string in Bytes
                Destination  [IN]      - source string
                
  returnvalue:  destination string length in lASCIILen

                return code:
                errOK_els00                 - successful conversion
                errInternal_els98           - when partial character in source,
                                                but hit end or insufficient
                                                space in target for conversion
                errConvertingEncoding_els98 - source contains invalid chars or
                                               char sequences
  -----------------------------------------------------------------------------
*/
static tsp00_Longint
ls05_ConvertUTF8ToASCII(char                          *pszASCII,
                        tsp00_Longuint                &lASCIILen,
                        const SAPDB_DynamicUTF8String &Source);


/*
  -----------------------------------------------------------------------------

  function:     ls05_ConvertUTF8ToASCII - overloaded
  -----------------------------------------------------------------------------

  description:  Converts a string from UTF8 to ASCII .

  arguments:    pszASCII   [OUT]     - destination string
                lASCIILen  [IN/OUT]  - length of destination string in Bytes
                pszUTF8    [IN]      - source string
                lUTF8Len   [IN]      - source string length in Bytes
                
  returnvalue:  destination string length in lASCIILen

                return code:
                errOK_els00                 - successful conversion
                errInternal_els98           - when partial character in source,
                                                but hit end or insufficient
                                                space in target for conversion
                errConvertingEncoding_els98 - source contains invalid chars or
                                               char sequences
  -----------------------------------------------------------------------------
*/
static tsp00_Longint
ls05_ConvertUTF8ToASCII(char            *pszASCII,
                        tsp00_Longuint  &lASCIILen,
                        unsigned char   *pszUTF8,
                        tsp00_Longuint   lUTF8Len);


/*
  -----------------------------------------------------------------------------

  function:     ls05_ConvertASCIIToUTF8
  -----------------------------------------------------------------------------

  description:  Converts a string from ASCII to UTF8.

  arguments:    Destination  [OUT]  - destination string
                pszASCII     [IN]   - source string
                lASCIILen    [IN]   - length of source string in Bytes

  returnvalue:  return code:
                errOK_els00                 - successful conversion
                errInternal_els98           - when partial character in source,
                                                but hit end or insufficient
                                                space in target for conversion
                errConvertingEncoding_els98 - source contains invalid chars or
                                               char sequences
  -----------------------------------------------------------------------------
*/
static tsp00_Longint
ls05_ConvertASCIIToUTF8(SAPDB_DynamicUTF8String &Destination,
                        unsigned char           *pszASCII,
                        tsp00_Longuint           lASCIILen);


/*
  -----------------------------------------------------------------------------

  function:     ls05_ConvertASCIIToUCS2
  -----------------------------------------------------------------------------

  description:  Converts a string from ASCII to UCS2.

  arguments:    pszUCS2     [IN]     - destination string
                lUCS2Len    [IN/OUT] - length of destination string in Bytes
                pszASCII    [OUT]    - source string
                lASCIILen   [IN]     - length of source string in Bytes
                UCS2Type    [IN]     - type of UCS2 encoding (swapped or not)

  returnvalue:  destination string length in lUCS2Len

                return code:
                errOK_els00                 - successful conversion
                errInternal_els98           - when partial character in source,
                                                but hit end or insufficient
                                                space in target for conversion
                errConvertingEncoding_els98 - source contains invalid chars or
                                               char sequences
  -----------------------------------------------------------------------------
*/
static tsp00_Longint
ls05_ConvertASCIIToUCS2(unsigned char       *pszUCS2,
                        tsp00_Longuint      &lUCS2Len,      // reference; contains real length on return
                        const unsigned char *pszASCII,
                        tsp00_Longuint       lASCIILen,
                        const tsp77encoding *UCS2Type);


/*
  -----------------------------------------------------------------------------

  function:     ls05_ConvertUCS2ToUTF8
  -----------------------------------------------------------------------------

  description:  Converts a string from UCS2 to UTF8.

  arguments:    Destination  [OUT]  - destination string
                pszSource    [IN]   - source string
                lSourceLen   [IN]   - length of source string in Bytes
                UCS2Type     [IN]   - type of UCS2 encoding (swapped or not)

  returnvalue:  return code:
                errOK_els00                 - successful conversion
                errInternal_els98           - when partial character in source,
                                                but hit end or insufficient
                                                space in target for conversion
                errConvertingEncoding_els98 - source contains invalid chars or
                                               char sequences
  -----------------------------------------------------------------------------
*/
static tsp00_Longint
ls05_ConvertUCS2ToUTF8(SAPDB_DynamicUTF8String  &Destination,
                       unsigned char            *pszSource,
                       tsp00_Longint             lSourceLen,
                       tls00_CodeType            UCS2Type);     // gives a hint on swap type


/*
  -----------------------------------------------------------------------------

  function:     ls05_ConvertUCS2ToASCII
  -----------------------------------------------------------------------------

  description:  Converts a string from UCS2 to ASCII Latin1.

  arguments:    pszASCII    [OUT]    - destination string
                lASCIILen   [IN/OUT] - length of destination string in Bytes
                pszUCS2     [IN]     - source string
                lUCS2Len    [IN]     - length of source string in Bytes
                UCS2Type    [IN]     - type of UCS2 encoding (swapped or not)

  returnvalue:  destination string length in lASCIILen

                return code:
                errOK_els00                 - successful conversion
                errInternal_els98           - when partial character in source,
                                                but hit end or insufficient
                                                space in target for conversion
                errConvertingEncoding_els98 - source contains invalid chars or
                                               char sequences
  -----------------------------------------------------------------------------
*/
static tsp00_Longint
ls05_ConvertUCS2ToASCII(unsigned char       *pszASCII,
                        tsp00_Longint       &lASCIILen,
                        unsigned char       *pszUCS2,
                        tsp00_Longint        lUCS2Len,
                        const tsp77encoding *UCS2Type);


/*
  -----------------------------------------------------------------------------
  EndChapter: Private functions declaration 
  -----------------------------------------------------------------------------
*/


/*
  -----------------------------------------------------------------------------
  Chapter: Public functions implementation
  -----------------------------------------------------------------------------
*/


/*
  -----------------------------------------------------------------------------
  function:     ls05StrToUpr
  -----------------------------------------------------------------------------
*/
void ls05StrToUpr(tsp00_Addr pszString, tsp00_Int4 lStringLength)
{
    ROUTINE_DBG_MEO00 ("ls05StrToUpr");

    for (int i = 0; i < lStringLength; ++i)
    {
        pszString[i] = lLatin1_ToUpperMap_ls05[STATIC_CAST(unsigned char, pszString[i])];
    }
}
//  ls05StrToUpr()


/*
  -----------------------------------------------------------------------------
  function:     ls05StrTrimRight - overloaded
  -----------------------------------------------------------------------------
*/
void ls05StrTrimRight(tsp00_Addr pszString, tsp00_Int4& lStrLength)
{
    ROUTINE_DBG_MEO00 ("ls05StrTrimRight");

    tsp00_Addr  szLast      = NULL;
    tsp00_Addr  pszMyString = pszString;
    int         i           = 0;

    while (i <= lStrLength - 1)
    {
        if (isspace(*pszMyString))
        {
            if (szLast == NULL)
                szLast = pszMyString;
        }
        else
        {
            szLast = NULL;
        }

        pszMyString++;
        i++;
    }

    if (szLast != NULL)
    {
        // calculate string length newly
        lStrLength = STATIC_CAST(tsp00_Int4, (szLast - (pszString)) );
    }
}
// ls05StrTrimRight()


/*
  -----------------------------------------------------------------------------
  function:     ls05StrTrimRight - overloaded
  -----------------------------------------------------------------------------
*/
void
ls05StrTrimRight(tsp00_Addr      pszString,
                 tsp00_Longint  &lStrLength,
                 char            cTrimVal,
                 tls00_CodeType  CodeType)
{
    ROUTINE_DBG_MEO00 ("ls05StrTrimRight III");

    tsp00_Uint4 lCount = 0;

    switch (CodeType)
    {
        case ctUCS2Swapped_els00:
        {
            lCount = sp77encodingUCS2Swapped->countPadChars(STATIC_CAST(void*, pszString),
                                                            STATIC_CAST(tsp00_Uint4, lStrLength),
                                                            cTrimVal);
            break;
        }
        case ctUCS2_els00:
        {
            lCount = sp77encodingUCS2->countPadChars(STATIC_CAST(void*, pszString),
                                                     STATIC_CAST(tsp00_Uint4, lStrLength),
                                                     cTrimVal);
            break;
        }
        default: // covers ctAscii_els00 and ctUTF8_els00 (trim char may only be US7ASCII)
        {
            ls05StrTrimRight(pszString, lStrLength, cTrimVal);
        }
    }
    lStrLength -= (2 * lCount); // Attention: in case of ASCII/UTF8 we rely on the initialization of lCount
    return;
}
// ls05StrTrimRight()


/*
  -----------------------------------------------------------------------------
  function:     ls05StrTrimLeft - overloaded
  -----------------------------------------------------------------------------
*/
void
ls05StrTrimLeft(tsp00_Addr       pBuffer,
                tsp00_Longint   &lBufferLength,
                char             cTrimChar,
                tls00_CodeType   CodeType)
{
    ROUTINE_DBG_MEO00 ("ls05StrTrimLeft II");

    bool          bdone = false;
    tsp00_Longint i     = 0;

    switch (CodeType)
    {
        case ctUCS2Swapped_els00:
        {
            for (; ( (i < lBufferLength) && (pBuffer[i] == cTrimChar) && (pBuffer[i+1] == 0) ); i += 2)
                ;
            bdone = true;
        // ATTENTION: Fall through  
        }
        case ctUCS2_els00:
        {
            if (false == bdone)
            {
                for (; ( (i < lBufferLength) && (pBuffer[i] == 0) && (pBuffer[i+1] == cTrimChar) ); i += 2)
                    ;
            }
            
            if ( (0 != i) && (lBufferLength != i) )
            {
                lBufferLength -= i;
                memmove(pBuffer, &pBuffer[i], lBufferLength*sizeof(char));              
            }
            else if (lBufferLength == i)
            {
                lBufferLength = 0;
            }
            break;
        }
        default: // covers ctAscii_els00 and ctUTF8_els00 (trim char may only be US7ASCII)
        {
            ls05StrTrimLeftEx(pBuffer, lBufferLength);
        }
    }

    return;
}
// ls05StrTrimLeft()


/*
  -----------------------------------------------------------------------------
  function:     ls05StrTrimRightEx
  -----------------------------------------------------------------------------
*/
void ls05StrTrimRightEx(tsp00_Addr pszString)
{   
    ROUTINE_DBG_MEO00 ("ls05StrTrimRightEx");

    tsp00_Addr pszLast = NULL;
    tsp00_Addr pszS    = pszString;

    while (0 != *pszS)
    {
        if (isspace(*pszS))
        {
            if (NULL == pszLast)
            {
                pszLast = pszS;
            }
        }
        else
        {
            pszLast = NULL;
        }

        ++pszS;
    }
    if (NULL != pszLast)    // truncate at trailing space start
    {
        *pszLast = '\0';
    }
}
// ls05StrTrimRightEx()


/*
  -----------------------------------------------------------------------------
  function:     ls05StrTrimLeftEx
  -----------------------------------------------------------------------------
*/
void ls05StrTrimLeftEx(tsp00_Addr pszString, tsp00_Longint& lStrLength)
{
    ROUTINE_DBG_MEO00 ("ls05StrTrimLeftEx");

    tsp00_Longint len = 0;

    while ( (isspace(pszString[len])) && (len < lStrLength) )
        ++len;


    if ( (0 != len) && (len != lStrLength) )
    {
        lStrLength -= len;
        memmove(pszString, &pszString[len], lStrLength);
    }
    else if (len == lStrLength)
    {
        lStrLength = 0;
    }
}
// ls05StrTrimLeftEx()


/*
  -----------------------------------------------------------------------------
  function:     ls05StrCmpWFill
  -----------------------------------------------------------------------------
*/
bool ls05StrCmpWFill(const tsp00_Addr pszValue1,
                     tsp00_Int4 lVal1Len,
                     const tsp00_Addr pszValue2,
                     tsp00_Int4 lVal2Len)
{
    ROUTINE_DBG_MEO00 ("ls05StrCmpWFill");

    bool        bEqual = true;

    tsp00_Int4  lCompLen = lVal1Len;
    tsp00_Addr  pszCompVal = pszValue2;

    tsp00_Int4  lDif = (lVal2Len - lCompLen);

    if (lDif < 0)
    {
        lCompLen    = lVal2Len;
        pszCompVal  = pszValue1;
    }

    if (0 == strncmp(pszValue2, pszValue1, lCompLen) )
    {
        if (lDif != 0)
        {
            tsp00_Int2 i = 0;

            pszCompVal = pszCompVal + lCompLen;

            lDif = (lDif > 0) ? lDif : -lDif;
            for (i; i < lDif; ++i)
            {
                if (pszCompVal[i] != ' ')
                {
                    bEqual = false;
                    break;
                }
            }
        }
    }
    else
    {
        bEqual = false;
    }

    return bEqual;
}
// ls05StrCmpWFill()


/*
  -----------------------------------------------------------------------------
  function:     ls05InsertFillBytes
  -----------------------------------------------------------------------------
*/
void
ls05InsertFillBytes(unsigned char       *pszData,
                    tsp00_Longint        lDataLen,
                    tsp00_DataType       DBDataType,
                    tsp00_Longint        lNeededDataLen,
                    bool                 bUnicode,
                    tsp00_SwapKind_Enum  mySwap)
{
    ROUTINE_DBG_MEO00 ("ls05InsertFillBytes");
    
    // All character based data is filled with the ascii filler ' ' except char byte fields
    size_t  lDif = lNeededDataLen - lDataLen;
    if (0 < lDif)
    {
        char cFill = ls07GetFillByte(DBDataType);
        if (DEF_BYTE_LS00 == cFill)
        {
            memset(pszData + lDataLen, cFill, lDif);
        }
        else
        {
            if (true == bUnicode)
            {
                char c1 = (sw_normal == mySwap) ? DEF_BYTE_LS00 : cFill;
                char c2 = (sw_normal == mySwap) ? cFill         : DEF_BYTE_LS00;

                for (int i = lDataLen; i < lNeededDataLen; pszData[i++] = c1, pszData[i++] = c2)
                    ;
            }
            else
            {
                memset(pszData + lDataLen, cFill, lDif);
            }
        }
    }
    return;
}
// ls05InsertFillBytes


/*!
  -----------------------------------------------------------------------------
  function:     ls05AreEncodingsCompatible
  -----------------------------------------------------------------------------
*/
tsp00_Longint
ls05AreEncodingsCompatible(tls00_CodeType DataFile_Encoding,
                           tls00_CodeType DB_Encoding,
                           bool           bLoad)
{
    tsp00_Longint rc = errOK_els00;

    if (true == bLoad)
    {
        if ( (ctAscii_els00 != DataFile_Encoding) && (ctAscii_els00 == DB_Encoding) )
        {
            rc = errIncompatibleEncodings_els98;
        }
    }
    else
    {
        if ( (ctAscii_els00 == DataFile_Encoding) && (ctAscii_els00 != DB_Encoding) )
        {
            rc = errIncompatibleEncodings_els98;
        }
    }
    return rc;
}
// ls05CheckExternalEncoding()


/*
  -----------------------------------------------------------------------------
  function:     ls05ConvertToUCS2
  -----------------------------------------------------------------------------
*/
tsp00_Longint
ls05ConvertToUCS2(unsigned char  *pszDestination,
                  tsp00_Longuint &lDestinationLen,
                  unsigned char  *pszSource,
                  tsp00_Longint   lSourceLen,
                  tls00_CodeType  SrcEncoding,
                  tsp00_DataType  DBDataType,
                  tsp00_Addr      pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls05ConvertToUCS2");

    tsp00_Longint rc = errOK_els00;

    // if db data type is binary or if external code type is UCS2  no conversion applied
    // external code type UCS2 is checked by called func
    if ( (dstrb != DBDataType) && (dstrdb != DBDataType) )
    {
        rc = ls05ConvertToUCS2Simple(pszDestination, lDestinationLen, pszSource, lSourceLen, SrcEncoding, pszErrText);
    }

    return rc;
}
// ls05ConvertToUCS2()


/*
  -----------------------------------------------------------------------------
  function:     ls05ConvertToUCS2Simple
  -----------------------------------------------------------------------------
*/
tsp00_Longint
ls05ConvertToUCS2Simple(unsigned char  *pszDestination,
                        tsp00_Longuint &lDestinationLen,
                        unsigned char  *pszSource,
                        tsp00_Longint   lSourceLen,
                        tls00_CodeType  SrcEncoding,
                        tsp00_Addr      pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls05ConvertToUCS2Simple");

    tsp00_Longint rc = errOK_els00;

    if ( (ctUCS2_els00 != SrcEncoding) && (ctUCS2Swapped_els00 != SrcEncoding) )        // source encoding may not be already ucs2
    {
        if (ctUTF8_els00 == SrcEncoding)
        {
            rc = ls05_ConvertUTF8ToUCS2(pszDestination, lDestinationLen, pszSource, lSourceLen, ctUCS2Native_els00);
        }
        else
        {
            rc = ls05_ConvertASCIIToUCS2(pszDestination, lDestinationLen, pszSource, lSourceLen, sp77encodingUCS2Native);
        }
    }

    // Both conversion functions return identical values for
    //  errOK_els00                      : conversion successful; length of converted string is returned in lDestinationLen
    //  errEncodingSourceExhausted_els98 : partial character in source, but hit end
    //  errConvertingEncoding_els98      : source contains invalid chars or char sequences)
    //  errEncodingTargetExhausted_els98 : insuff. space in target for conversion)
    if (errOK_els00 != rc)
    {
        if ( errConvertingEncoding_els98 != rc )
        {
            RepServMessages::Instance()->ls98Msg(layCursor_els98, pszErrText, errInternal_els98,
                                                                              _INTERNAL_FILE_ID_, __LINE__, rc);
            rc = errInternal_els98;
        }
        else
        {
            if (ctAscii_els00== SrcEncoding)
            {
                RepServMessages::Instance()->ls98Msg(layCursor_els98, pszErrText,
                                                                      errConvertingEncoding_els98, "ASCII", "UCS2");
            }
            else
            {
                RepServMessages::Instance()->ls98Msg(layCursor_els98, pszErrText,
                                                                      errConvertingEncoding_els98, "UTF8", "UCS2");
            }

            rc = errConvertingEncoding_els98;
        }
    }
    return rc;
}
// ls05ConvertToUCS2Simple()


/*
  -----------------------------------------------------------------------------
  function:     ls05ConvertToUCS2NotSwapped
  -----------------------------------------------------------------------------
*/
tsp00_Longint
ls05ConvertToUCS2NotSwapped(unsigned char  *pszDestination,
                            tsp00_Longuint &lDestinationLen,
                            unsigned char  *pszSource,
                            tsp00_Longuint  lSourceLen,
                            tls00_CodeType  SrcEncoding,
                            tsp00_Addr      pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls05ConvertToUCS2NotSwapped");

    tsp00_Longint rc = errOK_els00;

    if ( (ctUCS2_els00 != SrcEncoding) && (ctUCS2Swapped_els00 != SrcEncoding) )
    {
        if (ctUTF8_els00 == SrcEncoding)
        {
            rc = ls05_ConvertUTF8ToUCS2(pszDestination, lDestinationLen, pszSource, lSourceLen, ctUCS2_els00);
        }
        else
        {
            rc = ls05_ConvertASCIIToUCS2(pszDestination, lDestinationLen, pszSource, lSourceLen, sp77encodingUCS2);
        }
    }

    // Both conversion functions return identical values for
    //  errOK_els00                      : conversion successful; length of converted string is returned in lDestinationLen
    //  errEncodingSourceExhausted_els98 : partial character in source, but hit end
    //  errConvertingEncoding_els98      : source contains invalid chars or char sequences)
    //  errEncodingTargetExhausted_els98 : insuff. space in target for conversion)
    if (errOK_els00 != rc)
    {
        if ( errConvertingEncoding_els98 != rc )
        {
            RepServMessages::Instance()->ls98Msg(layCursor_els98, pszErrText, errInternal_els98,
                                                                              _INTERNAL_FILE_ID_, __LINE__, rc);
            rc = errInternal_els98;
        }
        else
        {
            if (ctAscii_els00== SrcEncoding)
            {
                RepServMessages::Instance()->ls98Msg(layCursor_els98, pszErrText,
                                                                      errConvertingEncoding_els98, "ASCII", "UCS2");
            }
            else
            {
                RepServMessages::Instance()->ls98Msg(layCursor_els98, pszErrText,
                                                                      errConvertingEncoding_els98, "UTF8", "UCS2");
            }

            rc = errConvertingEncoding_els98;
        }
    }
    return rc;
}
// ls05ConvertToUCS2NotSwapped()


/*
  -----------------------------------------------------------------------------
  function:     ls05ConvertFromUCS2
  -----------------------------------------------------------------------------
*/
tsp00_Longint
ls05ConvertFromUCS2(unsigned char  *pszDestination,
                    tsp00_Longint  &lDestinationLen,
                    tls00_CodeType  DestEncoding,
                    unsigned char  *pszSource,
                    tsp00_Longint   lSourceLen,
                    tsp00_DataType  DBDataType,
                    tsp00_Addr      pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls05ConvertFromUCS2");

    tsp00_Longint rc = errOK_els00;

    // if db data type is binary or if external code type is UCS2  no conversion applied
    // external code type UCS2 is checked by called func
    if ( (dstrb != DBDataType) && (dstrdb != DBDataType) )
    {
        short swapTest = 1;
        
        rc = ls05ConvertFromUCS2Simple(pszDestination,
                                       pszSource,
                                       lDestinationLen,
                                       lSourceLen,
                                       DestEncoding,
                                       ((*(char *)&swapTest) ? true : false),
                                       pszErrText);
    }
    
    return rc;
}
// ls05ConvertFromUCS2()


/*
  -----------------------------------------------------------------------------
  function:     ls05ConvertFromUCS2Simple - overloaded
  -----------------------------------------------------------------------------
*/
tsp00_Longint
ls05ConvertFromUCS2Simple(unsigned char  *pszDestination,
                          unsigned char  *pszSource,
                          tsp00_Longint  &lDestinationLen,
                          tsp00_Longint   lSourceLen,
                          tls00_CodeType  DestEncoding,
                          bool            bSourceSwapped,
                          tsp00_Addr      pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls05ConvertFromUCS2Simple I");

    tsp00_Longint rc = errOK_els00;

    const tsp77encoding *pSrcEncoding_sp77 = (true == bSourceSwapped) ? sp77encodingUCS2Swapped : sp77encodingUCS2;
    tls00_CodeType       pSrcEncoding      = (true == bSourceSwapped) ? ctUCS2Swapped_els00 : ctUCS2_els00;

    if ( (ctUCS2_els00 != DestEncoding) && (ctUCS2Swapped_els00 != DestEncoding) )        // destination encoding may not be already ucs2
    {
        if (ctUTF8_els00 == DestEncoding)
        {
            SAPDB_DynamicUTF8String s;

            rc = ls05_ConvertUCS2ToUTF8(s, pszSource, lSourceLen, pSrcEncoding);
            if (errOK_els00 == rc)
            {
                lDestinationLen = s.BasisSize();
                memcpy(pszDestination, s.StrPtr(), lDestinationLen);
                pszDestination[lDestinationLen] = 0;
            }
        }
        else
        {
            rc = ls05_ConvertUCS2ToASCII(pszDestination, lDestinationLen, pszSource, lSourceLen, pSrcEncoding_sp77);
        }
    }

   
    // Both conversion functions return identical values for
    //  errOK_els00                      : conversion successful; length of converted string is returned in lDestinationLen
    //  errEncodingSourceExhausted_els98 : partial character in source, but hit end
    //  errConvertingEncoding_els98      : source contains invalid chars or char sequences)
    //  errEncodingTargetExhausted_els98 : insuff. space in target for conversion)
    if (errOK_els00 != rc)
    {
        if (errConvertingEncoding_els98 != rc)
        {
            RepServMessages::Instance()->ls98Msg(layCursor_els98, pszErrText, errInternal_els98,
                                                                              _INTERNAL_FILE_ID_, __LINE__, rc);
            rc = errInternal_els98;
        }
        else
        {
            if (ctAscii_els00 == DestEncoding)
            {
                RepServMessages::Instance()->ls98Msg(layCursor_els98, pszErrText,
                                                                      errConvertingEncoding_els98, "UCS2", "ASCII");
            }
            else
            {
                RepServMessages::Instance()->ls98Msg(layCursor_els98, pszErrText,
                                                                      errConvertingEncoding_els98, "UCS2", "UTF8");
            }

            rc = errConvertingEncoding_els98;
        }
    }

    return rc;
}
// ls05ConvertFromUCS2Simple()


/*
  -----------------------------------------------------------------------------
  function:     ls05ConvertFromUCS2Simple
  -----------------------------------------------------------------------------
*/
tsp00_Longint
ls05ConvertFromUCS2Simple(unsigned char  *pszDestination,
                          unsigned char  *pszSource,
                          tsp00_Longint  &lDestinationLen,
                          tsp00_Longint   lSourceLen,
                          tls00_CodeType  DestEncoding,
                          tsp00_Addr      pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls05ConvertFromUCS2Simple II");

    short swapTest = 1;
    
    tsp00_Longint rc = ls05ConvertFromUCS2Simple(pszDestination,
                                                 pszSource,
                                                 lDestinationLen,
                                                 lSourceLen,
                                                 DestEncoding,
                                                 ((*(char *)&swapTest) ? true : false),
                                                 pszErrText);
    return rc;
}
// ls05ConvertFromUCS2Simple()


/*
  -----------------------------------------------------------------------------
  function:     ls05ConvertToUTF8
  -----------------------------------------------------------------------------
*/
tsp00_Longint
ls05ConvertToUTF8(SAPDB_DynamicUTF8String &Destination,
                  unsigned char           *pszSource,
                  tsp00_Longuint           lSourceLen,
                  tls00_CodeType           SrcEncoding,     // must give a hint on swapped or not - e.g. ctUCS2_els00 or ctUCS2Native_els00 or ctUCS2Swapped_els00
                  tsp00_Addr               pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls05ConvertToUTF8");

    tsp00_Longint rc = errOK_els00;

    if (ctUTF8_els00 != SrcEncoding)    // destination encoding MUST NOT be already UTF8
    {
        if (ctAscii_els00 == SrcEncoding)
        {
            rc = ls05_ConvertASCIIToUTF8(Destination, pszSource, lSourceLen);
        }
        else
        {
            rc = ls05_ConvertUCS2ToUTF8(Destination, pszSource, lSourceLen, SrcEncoding);
        }
    }

    // Both conversion functions return identical values for
    //  errOK_els00                      : conversion successful
    //  errEncodingSourceExhausted_els98 : partial character in source, but hit end
    //  errConvertingEncoding_els98      : source contains invalid chars or char sequences)
    //  errEncodingTargetExhausted_els98 : insuff. space in target for conversion)
    if (errOK_els00 != rc)
    {
        if ( errConvertingEncoding_els98 != rc )
        {
            RepServMessages::Instance()->ls98Msg(layCursor_els98, pszErrText, errInternal_els98,
                                                                              _INTERNAL_FILE_ID_, __LINE__, rc);
            rc = errInternal_els98;
        }
        else
        {
            if (ctAscii_els00 == SrcEncoding)
            {
                RepServMessages::Instance()->ls98Msg(layCursor_els98, pszErrText,
                                                                      errConvertingEncoding_els98, "ASCII", "UTF8");
            }
            else
            {
                RepServMessages::Instance()->ls98Msg(layCursor_els98, pszErrText,
                                                                      errConvertingEncoding_els98, "UCS2", "UTF8");
            }
            rc = errConvertingEncoding_els98;
        }
    }
    return rc;
}
// ls05ConvertToUTF8()


/*
  -----------------------------------------------------------------------------
  function:     ls05ConvertFromUTF8 I
  -----------------------------------------------------------------------------
*/
tsp00_Longint
ls05ConvertFromUTF8(unsigned char           *pszDestination,
                    tsp00_Longuint          &lDestinationLen,   // reference - contains the real length on return
                    SAPDB_DynamicUTF8String &Source,
                    tls00_CodeType           DestEncoding,      // must give a hint on swapped or not - e.g. ctUCS2_els00 or ctUCS2Native_els00 or ctUCS2Swapped_els00
                    tsp00_Addr               pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls05ConvertFromUTF8 I");

    tsp00_Longint rc = errOK_els00;

    if (ctUTF8_els00 != DestEncoding)   // destination encoding MUST NOT be already UTF8
    {
        if (ctAscii_els00 == DestEncoding)
        {
            rc = ls05_ConvertUTF8ToASCII(REINTERPRET_CAST(char *, pszDestination), lDestinationLen, Source);
        }
        else
        {
            rc = ls05_ConvertUTF8ToUCS2(pszDestination, lDestinationLen, Source.StrPtr(), Source.BasisSize(), DestEncoding);
        }
    }
    // Both conversion functions return identical values for
    //  errOK_els00                      : conversion successful
    //  errEncodingSourceExhausted_els98 : partial character in source, but hit end
    //  errConvertingEncoding_els98      : source contains invalid chars or char sequences)
    //  errEncodingTargetExhausted_els98 : insuff. space in target for conversion)
    if (errOK_els00 != rc)
    {
        if (errConvertingEncoding_els98 != rc)
        {
            RepServMessages::Instance()->ls98Msg(layCursor_els98, pszErrText, errInternal_els98,
                                                                              _INTERNAL_FILE_ID_, __LINE__, rc);
            rc = errInternal_els98;
        }
        else
        {
            if (ctAscii_els00 == DestEncoding)
            {
                RepServMessages::Instance()->ls98Msg(layCursor_els98, pszErrText,
                                                                      errConvertingEncoding_els98, "UTF8", "ASCII");
            }
            else
            {
                RepServMessages::Instance()->ls98Msg(layCursor_els98, pszErrText,
                                                                      errConvertingEncoding_els98, "UTF8", "UCS2");
            }
            rc = errConvertingEncoding_els98;
        }
    }
    return rc;
}
// ls05ConvertFromUTF8()


/*
  -----------------------------------------------------------------------------
  function:     ls05ConvertFromUTF8 II
  -----------------------------------------------------------------------------
*/
tsp00_Longint
ls05ConvertFromUTF8(unsigned char           *pszDestination,
                    tsp00_Longuint          &lDestinationLen,
                    unsigned char           *pszSource,
                    tsp00_Longuint           lSourceLen,
                    tls00_CodeType           DestEncoding,
                    tsp00_Addr               pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls05ConvertFromUTF8 II");

    tsp00_Longint rc = errOK_els00;

    if (ctUTF8_els00 != DestEncoding)   // destination encoding MUST NOT be already UTF8
    {
        if (ctAscii_els00 == DestEncoding)
        {
            rc = ls05_ConvertUTF8ToASCII(REINTERPRET_CAST(char*, pszDestination), lDestinationLen,
                                         pszSource,                               lSourceLen);
        }
        else
        {
            rc = ls05_ConvertUTF8ToUCS2(pszDestination, lDestinationLen, pszSource, lSourceLen, DestEncoding);
        }
    }
    // Both conversion functions return identical values for
    //  errOK_els00                      : conversion successful
    //  errEncodingSourceExhausted_els98 : partial character in source, but hit end
    //  errConvertingEncoding_els98      : source contains invalid chars or char sequences)
    //  errEncodingTargetExhausted_els98 : insuff. space in target for conversion)
    if (errOK_els00 != rc)
    {
        if (errConvertingEncoding_els98 != rc)
        {
            RepServMessages::Instance()->ls98Msg(layCursor_els98, pszErrText, errInternal_els98,
                                                                              _INTERNAL_FILE_ID_, __LINE__, rc);
            rc = errInternal_els98;
        }
        else
        {
            if (ctAscii_els00 == DestEncoding)
            {
                RepServMessages::Instance()->ls98Msg(layCursor_els98, pszErrText,
                                                                      errConvertingEncoding_els98, "UTF8", "ASCII");
            }
            else
            {
                RepServMessages::Instance()->ls98Msg(layCursor_els98, pszErrText,
                                                                      errConvertingEncoding_els98, "UTF8", "UCS2");
            }
            rc = errConvertingEncoding_els98;
        }
    }
    return rc;
}
// ls05ConvertFromUTF8()

/*
  -----------------------------------------------------------------------------
  EndChapter: Public functions implementation
  -----------------------------------------------------------------------------
*/

//////////////////////////////////////////////////////////////////////////////////////////////
/*
  -----------------------------------------------------------------------------
  Chapter: Private functions implementation
  -----------------------------------------------------------------------------
*/
static tsp00_Longint
ls05_ConvertUTF8ToUCS2(unsigned char       *pszUCS2,
                       tsp00_Longuint      &lUCS2Len,
                       const unsigned char *pszUTF8,
                       tsp00_Longuint       lUTF8Len,
                       tls00_CodeType       UCS2Type)
{
    ROUTINE_DBG_MEO00 ("ls05_ConvertUTF8ToUCS2");

    tsp00_Longint                     rc           = errOK_els00;
    SAPDB_UTF8Basis::ConversionResult rcConversion = SAPDB_UTF8Basis::Success;

    // Incoming source string is already UTF8
    SAPDB_DynamicUTF8String s;
    s.AssignRaw(STATIC_CAST(const SAPDB_UTF8*, pszUTF8), lUTF8Len);

    tsp81_UCS2Char *pszUCS2At = NULL;

    if (ctUCS2Native_els00 == UCS2Type)
    {
        rcConversion = s.ConvertToUCS2_Unaligned_Swap(REINTERPRET_CAST(tsp81_UCS2Char*, pszUCS2),
                                                      REINTERPRET_CAST(tsp81_UCS2Char*, (pszUCS2 + lUCS2Len)),
                                                      pszUCS2At);
    }
    else
    {
        bool bDestSwap = (ctUCS2_els00 == UCS2Type) ? false : true;
        rcConversion = s.ConvertToUCS2_Unaligned_Swap(REINTERPRET_CAST(tsp81_UCS2Char*, pszUCS2),
                                                      REINTERPRET_CAST(tsp81_UCS2Char*, (pszUCS2 + lUCS2Len)),
                                                      pszUCS2At,
                                                      bDestSwap);
    }
    if (SAPDB_UTF8Basis::Success == rcConversion)
    {
        lUCS2Len = REINTERPRET_CAST(unsigned char*, pszUCS2At) - pszUCS2;
    }
    else if (SAPDB_UTF8Basis::SourceExhausted == rcConversion)
    {
        rc = errEncodingSourceExhausted_els98;
    }
    else if (SAPDB_UTF8Basis::SourceCorrupted == rcConversion)
    {
        rc = errConvertingEncoding_els98;
    }
    else
    {
        rc = errEncodingTargetExhausted_els98;
    }

    return rc;
}
// ls05_ConvertUTF8ToUCS2()


static tsp00_Longint
ls05_ConvertASCIIToUCS2(unsigned char       *pszUCS2,
                        tsp00_Longuint      &lUCS2Len,      // reference; contains real length on return
                        const unsigned char *pszASCII,
                        tsp00_Longuint       lASCIILen,
                        const tsp77encoding *UCS2Type)
{
    ROUTINE_DBG_MEO00 ("ls05_ConvertASCIIToUCS2");

    tsp00_Longint rc = errOK_els00;

    tsp78ConversionResult rcConversion = sp78_Ok;

    tsp00_Uint4 lUCS2BytesWritten = 0;
    tsp00_Uint4 lASCIIBytesParsed = 0;

    rcConversion = sp78convertBuffer(UCS2Type,          pszUCS2,  STATIC_CAST(tsp00_Uint4, lUCS2Len),  &lUCS2BytesWritten,
                                     sp77encodingAscii, pszASCII, STATIC_CAST(tsp00_Uint4, lASCIILen), &lASCIIBytesParsed);
    if (sp78_Ok == rcConversion)
    {
        lUCS2Len = lUCS2BytesWritten;
    }
    else if (sp78_SourceExhausted == rcConversion)
    {
        rc = errEncodingSourceExhausted_els98;
    }
    else if (sp78_SourceCorrupted == rcConversion)
    {
        rc = errConvertingEncoding_els98;
    }
    else
    {
        rc = errEncodingTargetExhausted_els98;
    }
    return rc;
}
// ls05_ConvertASCIIToUCS2()


/*
  -----------------------------------------------------------------------------
  function:     ls05_ConvertUCS2ToUTF8
  -----------------------------------------------------------------------------
*/
static tsp00_Longint
ls05_ConvertUCS2ToUTF8(SAPDB_DynamicUTF8String  &Destination,
                       unsigned char            *pszSource,
                       tsp00_Longint             lSourceLen,
                       tls00_CodeType            UCS2Type)      // gives a hint on swap type
{
    ROUTINE_DBG_MEO00 ("ls05_ConvertUCS2ToUTF8");

    tsp00_Longint                     rc           = errOK_els00;
    SAPDB_UTF8Basis::ConversionResult rcConversion = SAPDB_UTF8Basis::Success;

    if (ctUCS2Native_els00 == UCS2Type)
    {
        rcConversion = Destination.ConvertFromUCS2_Unaligned_Swap(REINTERPRET_CAST(tsp81_UCS2Char *, pszSource),
                                                                  REINTERPRET_CAST(tsp81_UCS2Char *, (pszSource + lSourceLen) ) );
    }
    else
    {
        bool bSrcSwapped = (ctUCS2_els00 == UCS2Type) ? false : true;
        rcConversion = Destination.ConvertFromUCS2_Unaligned_Swap(REINTERPRET_CAST(tsp81_UCS2Char*, pszSource),
                                                                  REINTERPRET_CAST(tsp81_UCS2Char*, (pszSource + lSourceLen)),
                                                                  bSrcSwapped);
    }

    if (SAPDB_UTF8Basis::SourceExhausted == rcConversion)
    {
        rc = errEncodingSourceExhausted_els98;
    }
    else if (SAPDB_UTF8Basis::TargetExhausted == rcConversion)
    {
        rc = errEncodingTargetExhausted_els98;
    }
    else if (SAPDB_UTF8Basis::SourceCorrupted == rcConversion)
    {
        rc = errConvertingEncoding_els98;
    }

    return rc;
}
// ls05_ConvertUCS2ToUTF8()


/*
  -----------------------------------------------------------------------------
  function:     ls05_ConvertUCS2ToASCII
  -----------------------------------------------------------------------------
*/
static tsp00_Longint
ls05_ConvertUCS2ToASCII(unsigned char       *pszASCII,
                        tsp00_Longint       &lASCIILen,
                        unsigned char       *pszUCS2,
                        tsp00_Longint        lUCS2Len,
                        const tsp77encoding *UCS2Type)
{
    ROUTINE_DBG_MEO00 ("ls05_ConvertUCS2ToASCII");

    tsp00_Longint         rc           = errOK_els00;
    tsp78ConversionResult rcConversion = sp78_Ok;

    tsp00_Uint4 lASCIIBytesWritten = 0;
    tsp00_Uint4 lUCS2BytesParsed   = 0;

    rcConversion = sp78convertBuffer(sp77encodingAscii, pszASCII, STATIC_CAST(tsp00_Uint4, lASCIILen), &lASCIIBytesWritten,
                                     UCS2Type,          pszUCS2,  STATIC_CAST(tsp00_Uint4, lUCS2Len),  &lUCS2BytesParsed);
    if (sp78_Ok == rcConversion)
    {
        lASCIILen = STATIC_CAST(tsp00_Longint, lASCIIBytesWritten);
    }
    else if (sp78_SourceExhausted == rcConversion)
    {
        rc = errEncodingSourceExhausted_els98;
    }
    else if (sp78_SourceCorrupted == rcConversion)
    {
        rc = errConvertingEncoding_els98;
    }
    else
    {
        rc = errEncodingTargetExhausted_els98;
    }

    return rc;
}
// ls05_ConvertUCS2ToASCII()


/*
  -----------------------------------------------------------------------------
  function:     ls05_ConvertUTF8ToASCII I
  -----------------------------------------------------------------------------
*/
static tsp00_Longint
ls05_ConvertUTF8ToASCII(char                          *pszASCII,
                        tsp00_Longuint                &lASCIILen,
                        const SAPDB_DynamicUTF8String &Source)
{
    ROUTINE_DBG_MEO00 ("ls05_ConvertUTF8ToASCII I");

    tsp00_Longint rc = errOK_els00;

    SAPDB_UTF8Basis::ConversionResult rcConversion = SAPDB_UTF8Basis::Success;

    char *pszASCIIEnd = NULL;

    rcConversion = Source.ConvertToASCII_Latin1(pszASCII, (pszASCII + lASCIILen), pszASCIIEnd);
    if (SAPDB_UTF8Basis::Success == rcConversion)
    {
        lASCIILen = pszASCIIEnd - pszASCII;
    }
    else if (SAPDB_UTF8Basis::SourceExhausted == rcConversion)
    {
        rc = errEncodingSourceExhausted_els98;
    }
    else if (SAPDB_UTF8Basis::SourceCorrupted == rcConversion)
    {
        rc = errConvertingEncoding_els98;
    }
    else
    {
        rc = errEncodingTargetExhausted_els98;
    }

    return rc;
}
// ls05_ConvertUTF8ToASCII()


/*
  -----------------------------------------------------------------------------
  function:     ls05_ConvertUTF8ToASCII II
  -----------------------------------------------------------------------------
*/
static tsp00_Longint
ls05_ConvertUTF8ToASCII(char            *pszASCII,
                        tsp00_Longuint  &lASCIILen,
                        unsigned char   *pszUTF8,
                        tsp00_Longuint   lUTF8Len)
{
    ROUTINE_DBG_MEO00 ("ls05_ConvertUTF8ToASCII II");

    tsp00_Longint rc = errOK_els00;

    SAPDB_UTF8Basis::ConversionResult rcConversion = SAPDB_UTF8Basis::Success;

    // Incoming source string is already UTF8
    SAPDB_DynamicUTF8String Source;
    Source.AssignRaw(STATIC_CAST(const SAPDB_UTF8*, pszUTF8), lUTF8Len);

    char *pszASCIIEnd = NULL;

    rcConversion = Source.ConvertToASCII_Latin1(pszASCII, (pszASCII + lASCIILen), pszASCIIEnd);
    if (SAPDB_UTF8Basis::Success == rcConversion)
    {
        lASCIILen = pszASCIIEnd - pszASCII;
    }
    else if (SAPDB_UTF8Basis::SourceExhausted == rcConversion)
    {
        rc = errEncodingSourceExhausted_els98;
    }
    else if (SAPDB_UTF8Basis::SourceCorrupted == rcConversion)
    {
        rc = errConvertingEncoding_els98;
    }
    else
    {
        rc = errEncodingTargetExhausted_els98;
    }

    return rc;
}
// ls05_ConvertUTF8ToASCII()


/*
  -----------------------------------------------------------------------------
  function:     ls05_ConvertASCIIToUTF8
  -----------------------------------------------------------------------------
*/
static tsp00_Longint
ls05_ConvertASCIIToUTF8(SAPDB_DynamicUTF8String &Destination,
                        unsigned char           *pszASCII,
                        tsp00_Longuint           lASCIILen)
{
    ROUTINE_DBG_MEO00 ("ls05_ConvertASCIIToUTF8");

    tsp00_Longint rc = errOK_els00;

    SAPDB_UTF8Basis::ConversionResult rcConversion = SAPDB_UTF8Basis::Success;

    rcConversion = Destination.ConvertFromASCII_Latin1(REINTERPRET_CAST(char*, pszASCII),
                                                       REINTERPRET_CAST(char*, (pszASCII + lASCIILen)) );

    if (SAPDB_UTF8Basis::SourceExhausted == rcConversion)
    {
        rc = errEncodingSourceExhausted_els98;
    }
    else if (SAPDB_UTF8Basis::SourceCorrupted == rcConversion)
    {
        rc = errConvertingEncoding_els98;
    }
    else if (SAPDB_UTF8Basis::TargetExhausted == rcConversion)
    {
        rc = errEncodingTargetExhausted_els98;
    }

    return rc;
}
// ls05_ConvertASCIIToUTF8()


/*
  -----------------------------------------------------------------------------
  EndChapter: Private functions implementation
  -----------------------------------------------------------------------------
*/
