/*
  -----------------------------------------------------------------------------

  module: vls10.cpp

  -----------------------------------------------------------------------------

  responsible:  SteffenS

  special area: Replication Manager

  description:  dispatcher
    
  version:      7.3.

  -----------------------------------------------------------------------------

  copyright:    SAP AG, 1997-2002

  -----------------------------------------------------------------------------


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/
#include "gls00.h"
#include "hls98msg.h"
#include "hls10.h"
#include "hls13.h"
#include "hls15.h"
#include "hls20.h"
#include "hls21.h"
#include "hls26.h"
#include "hls27_dataextract.h"
#include "hls28_TLoad.h"
#include "hls29_dataupdate.h"
#include "hls31_catalogextract.h"
#include "hls32_dbextract.h"
#include "hls99.h"


/*!
  -----------------------------------------------------------------------------

  function:     ls10_Destroy

  -----------------------------------------------------------------------------

  description:  Pseudo destructor.
                Destroys load command structure built using the info in the
                command.

  arguments:    Type    [IN]   - type of command struct to destroy
                Struct  [IN]   - struct to destroy

  returnvalue:  None
  -----------------------------------------------------------------------------
*/
static void
ls10_Destroy(tls00_ScanResult Type, char * &Struct);



/* Local Functions ******************************************************************/



/*
  -----------------------------------------------------------------------------
  function:     ls10_Destroy
  -----------------------------------------------------------------------------
*/
static void
ls10_Destroy(tls00_ScanResult Type, char * &Struct)
{
    ROUTINE_DBG_MEO00 ("ls10_Destroy");

    int i = 0;
    tls00_MultColSpec *MultCol = NULL;

    switch (Type)
    {
        case srFctDBExtr_els00:
        {
            tls00_DBExtrCmd* pDBExtrCmd  = ( tls00_DBExtrCmd *) Struct;

            if (NULL != pDBExtrCmd->dbeCatfile_ls00.fsFileName_ls00)
            {
                delete [] pDBExtrCmd->dbeCatfile_ls00.fsFileName_ls00;
                pDBExtrCmd->dbeCatfile_ls00.fsFileName_ls00 = NULL;
            }
            if (NULL != pDBExtrCmd->dbeDatfile_ls00.fsFileName_ls00)
            {
                delete [] pDBExtrCmd->dbeDatfile_ls00.fsFileName_ls00;
                pDBExtrCmd->dbeDatfile_ls00.fsFileName_ls00 = NULL;
            }

            break;
        }
        case srFctDBLoad_els00:
        {
            tls00_DBLoadCmd * pDBLoadCmd = ( tls00_DBLoadCmd *) Struct;

            if (NULL != pDBLoadCmd->dblCatfile_ls00.fsFileName_ls00)
            {
                delete [] pDBLoadCmd->dblCatfile_ls00.fsFileName_ls00;
                pDBLoadCmd->dblCatfile_ls00.fsFileName_ls00 = NULL;
            }
            if (NULL != pDBLoadCmd->dblDatfile_ls00.fsFileName_ls00)
            {
                delete [] pDBLoadCmd->dblDatfile_ls00.fsFileName_ls00;
                pDBLoadCmd->dblDatfile_ls00.fsFileName_ls00 = NULL;
            }
            if (NULL != pDBLoadCmd->dblReport_ls00.fsFileName_ls00)
            {
                delete [] pDBLoadCmd->dblReport_ls00.fsFileName_ls00;
                pDBLoadCmd->dblReport_ls00.fsFileName_ls00 = NULL;
            }


            break;
        }
        case srFctCatExtr_els00:
        {
            tls00_CatExtrCmd* pCatExtrCmd = ( tls00_CatExtrCmd *) Struct;

            // release full qualified table name
            if (NULL != pCatExtrCmd->cexPartSpec_ls00.psFullTabName_ls00)
            {
                delete [] pCatExtrCmd->cexPartSpec_ls00.psFullTabName_ls00;
                pCatExtrCmd->cexPartSpec_ls00.psFullTabName_ls00 = NULL;
            }

            // release outfile name
            if (NULL != pCatExtrCmd->cexOutfile_ls00.fsFileName_ls00)
            {
                delete [] pCatExtrCmd->cexOutfile_ls00.fsFileName_ls00;
                pCatExtrCmd->cexOutfile_ls00.fsFileName_ls00 = NULL;
            }


            break;
        }
        case srFctCatLoad_els00:
        {
            tls00_CatLoadCmd* pCatLoadCmd = ( tls00_CatLoadCmd *) Struct;

            // release full qualified table name
            if (NULL != pCatLoadCmd->cloPartSpec_ls00.psFullTabName_ls00)
            {
                delete [] pCatLoadCmd->cloPartSpec_ls00.psFullTabName_ls00;
                pCatLoadCmd->cloPartSpec_ls00.psFullTabName_ls00 = NULL;
            }

            // release infile name
            if (NULL != pCatLoadCmd->cloInfile_ls00.fsFileName_ls00)
            {
                delete [] pCatLoadCmd->cloInfile_ls00.fsFileName_ls00;
                pCatLoadCmd->cloInfile_ls00.fsFileName_ls00 = NULL;
            }

            break;
        }
        case srFctFastLoad_els00:
        {
            tls00_FloadCmd* FC = (tls00_FloadCmd*) Struct;

            if (NULL != FC->flcTable_ls00.tsFullTabName_ls00)
            {
                delete [] FC->flcTable_ls00.tsFullTabName_ls00;
                FC->flcTable_ls00.tsFullTabName_ls00 = NULL;
            }

            // better solution would be recursion
            if (NULL != FC->flcTable_ls00.tsCondSpec_ls00)
            {
                if (NULL != FC->flcTable_ls00.tsCondSpec_ls00->cosCond_ls00)
                {
                    delete FC->flcTable_ls00.tsCondSpec_ls00->cosCond_ls00;
                    FC->flcTable_ls00.tsCondSpec_ls00->cosCond_ls00 = NULL;
                }

                if (NULL != FC->flcTable_ls00.tsCondSpec_ls00->cosLeft_ls00)
                {
                    if (NULL != FC->flcTable_ls00.tsCondSpec_ls00->cosLeft_ls00->cosCond_ls00)
                    {
                        delete FC->flcTable_ls00.tsCondSpec_ls00->cosLeft_ls00->cosCond_ls00;
                        FC->flcTable_ls00.tsCondSpec_ls00->cosLeft_ls00->cosCond_ls00 = NULL;
                    }

                    delete FC->flcTable_ls00.tsCondSpec_ls00->cosLeft_ls00;
                    FC->flcTable_ls00.tsCondSpec_ls00->cosLeft_ls00 = NULL;
                }

                if (NULL != FC->flcTable_ls00.tsCondSpec_ls00->cosRight_ls00)
                {
                    if (NULL != FC->flcTable_ls00.tsCondSpec_ls00->cosRight_ls00->cosCond_ls00)
                    {
                        delete FC->flcTable_ls00.tsCondSpec_ls00->cosRight_ls00->cosCond_ls00;
                        FC->flcTable_ls00.tsCondSpec_ls00->cosRight_ls00->cosCond_ls00 = NULL;
                    }

                    delete FC->flcTable_ls00.tsCondSpec_ls00->cosRight_ls00;
                    FC->flcTable_ls00.tsCondSpec_ls00->cosRight_ls00 = NULL;
                }
            }

            MultCol = &FC->flcColSpec_ls00;

            // delete member of structure flcInfile_ls00 (type: tls00_FileSpec)
            tls00_FileSpec *myFS = &FC->flcInfile_ls00;
            if (NULL != myFS->fsFileName_ls00)
            {
                delete [] myFS->fsFileName_ls00;
                myFS->fsFileName_ls00 = NULL;
            }

            break;
        }
        case srFctDatLoad_els00:
        {
            tls00_DatLoadCmd *pDC = REINTERPRET_CAST(tls00_DatLoadCmd*, Struct);

            if (NULL != pDC->dloTable_ls00.tsFullTabName_ls00)
            {
                delete [] pDC->dloTable_ls00.tsFullTabName_ls00;
                pDC->dloTable_ls00.tsFullTabName_ls00 = NULL;
            }

            // better solution would be recursion
            if (NULL != pDC->dloTable_ls00.tsCondSpec_ls00)
            {
                if (NULL != pDC->dloTable_ls00.tsCondSpec_ls00->cosCond_ls00)
                {
                    delete pDC->dloTable_ls00.tsCondSpec_ls00->cosCond_ls00;
                    pDC->dloTable_ls00.tsCondSpec_ls00->cosCond_ls00 = NULL;
                }

                if (NULL != pDC->dloTable_ls00.tsCondSpec_ls00->cosLeft_ls00)
                {
                    if (NULL != pDC->dloTable_ls00.tsCondSpec_ls00->cosLeft_ls00->cosCond_ls00)
                    {
                        delete pDC->dloTable_ls00.tsCondSpec_ls00->cosLeft_ls00->cosCond_ls00;
                        pDC->dloTable_ls00.tsCondSpec_ls00->cosLeft_ls00->cosCond_ls00 = NULL;
                    }

                    delete pDC->dloTable_ls00.tsCondSpec_ls00->cosLeft_ls00;
                    pDC->dloTable_ls00.tsCondSpec_ls00->cosLeft_ls00 = NULL;
                }

                if (NULL != pDC->dloTable_ls00.tsCondSpec_ls00->cosRight_ls00)
                {
                    if (NULL != pDC->dloTable_ls00.tsCondSpec_ls00->cosRight_ls00->cosCond_ls00)
                    {
                        delete pDC->dloTable_ls00.tsCondSpec_ls00->cosRight_ls00->cosCond_ls00;
                        pDC->dloTable_ls00.tsCondSpec_ls00->cosRight_ls00->cosCond_ls00 = NULL;
                    }

                    delete pDC->dloTable_ls00.tsCondSpec_ls00->cosRight_ls00;
                    pDC->dloTable_ls00.tsCondSpec_ls00->cosRight_ls00 = NULL;
                }
            }

            MultCol = &pDC->dloColSpec_ls00;
        
            // free long value information
            for (i = 0; i < MultCol->mcsCount_ls00; ++i)
            {
                tls00_Column *pCol = MultCol->mcsColumn_ls00[i];

                if (true == pCol->colLongInfo.loiIsLong_ls00)
                {
                    if (NULL != pCol->colLongInfo.loiFileName_ls00)
                    {
                        delete [] pCol->colLongInfo.loiFileName_ls00;
                        pCol->colLongInfo.loiFileName_ls00 = NULL;
                    }

                    if (NULL != pCol->colLongInfo.loiFileStruct_ls00)
                    {
                        delete pCol->colLongInfo.loiFileStruct_ls00;
                        pCol->colLongInfo.loiFileStruct_ls00 = NULL;
                    }
                }
            }
    

            // delete member of structure dloInfile_ls00 (type: tls00_FileSpec)
            tls00_FileSpec *myFS = &pDC->dloInfile_ls00;
            if (NULL != myFS->fsFileName_ls00)
            {
                delete [] myFS->fsFileName_ls00;
                myFS->fsFileName_ls00 = NULL;
            }

            break;
        }
        case srFctDatUpd_els00:
        {
            tls00_DatUpdCmd   *pDC      = REINTERPRET_CAST(tls00_DatUpdCmd*, Struct);
            tls00_MultColSpec *pMultCol = NULL;

            if (NULL != pDC->dupTable_ls00.tsFullTabName_ls00)
            {
                delete [] pDC->dupTable_ls00.tsFullTabName_ls00;
                pDC->dupTable_ls00.tsFullTabName_ls00 = NULL;
            }

            // better solution would be recursion

            if (NULL != pDC->dupTable_ls00.tsCondSpec_ls00)
            {
                if (NULL != pDC->dupTable_ls00.tsCondSpec_ls00->cosCond_ls00)
                {
                    delete pDC->dupTable_ls00.tsCondSpec_ls00->cosCond_ls00;
                    pDC->dupTable_ls00.tsCondSpec_ls00->cosCond_ls00 = NULL;
                }

                if (NULL != pDC->dupTable_ls00.tsCondSpec_ls00->cosLeft_ls00)
                {
                    if (NULL != pDC->dupTable_ls00.tsCondSpec_ls00->cosLeft_ls00->cosCond_ls00)
                    {
                        delete pDC->dupTable_ls00.tsCondSpec_ls00->cosLeft_ls00->cosCond_ls00;
                        pDC->dupTable_ls00.tsCondSpec_ls00->cosLeft_ls00->cosCond_ls00 = NULL;
                    }

                    delete pDC->dupTable_ls00.tsCondSpec_ls00->cosLeft_ls00;
                    pDC->dupTable_ls00.tsCondSpec_ls00->cosLeft_ls00 = NULL;
                }

                if (NULL != pDC->dupTable_ls00.tsCondSpec_ls00->cosRight_ls00)
                {
                    if (NULL != pDC->dupTable_ls00.tsCondSpec_ls00->cosRight_ls00->cosCond_ls00)
                    {
                        delete pDC->dupTable_ls00.tsCondSpec_ls00->cosRight_ls00->cosCond_ls00;
                        pDC->dupTable_ls00.tsCondSpec_ls00->cosRight_ls00->cosCond_ls00 = NULL;
                    }

                    delete pDC->dupTable_ls00.tsCondSpec_ls00->cosRight_ls00;
                    pDC->dupTable_ls00.tsCondSpec_ls00->cosRight_ls00 = NULL;
                }
            }

            pMultCol = &pDC->dupColSpec_ls00;

            // free long value information
            for (i = 0; i < pMultCol->mcsCount_ls00; ++i)
            {
                tls00_Column *pCol = pMultCol->mcsColumn_ls00[i];

                if (true == pCol->colLongInfo.loiIsLong_ls00)
                {
                    if (NULL != pCol->colLongInfo.loiFileName_ls00)
                    {
                        delete [] pCol->colLongInfo.loiFileName_ls00;
                        pCol->colLongInfo.loiFileName_ls00 = NULL;
                    }

                    if (NULL != pCol->colLongInfo.loiFileStruct_ls00)
                    {
                        delete pCol->colLongInfo.loiFileStruct_ls00;
                        pCol->colLongInfo.loiFileStruct_ls00 = NULL;
                    }
                }

                if (NULL != pCol)
                {
                    if (NULL != pCol->colAssign_ls00.fasAsgnVal_ls00)
                    {
                        delete pCol->colAssign_ls00.fasAsgnVal_ls00;
                        pCol->colAssign_ls00.fasAsgnVal_ls00 = NULL;
                    }

                    if (NULL != pCol->colNullCond_ls00)        // RECURSION REQUIRED
                    {
                        delete pCol->colNullCond_ls00;
                        pCol->colNullCond_ls00 = NULL;
                    }

                    delete pCol;
                    pCol = NULL;
                }
            }

            // delete member of structure dloInfile_ls00 (type: tls00_FileSpec)
            tls00_FileSpec *myFS = &pDC->dupInfile_ls00;
            if (NULL != myFS->fsFileName_ls00)
            {
                delete [] myFS->fsFileName_ls00;
                myFS->fsFileName_ls00 = NULL;
            }

            break;
        }
        case srFctDatExtr_els00:
        {
            tls00_DatExtrCmd* pDEC = (tls00_DatExtrCmd*) Struct;
            MultCol = &pDEC->dexOutSpec_ls00;       // later on used to free column names

            if (NULL != pDEC->dexTable_ls00.tsFullTabName_ls00)
            {
                delete [] pDEC->dexTable_ls00.tsFullTabName_ls00;
                pDEC->dexTable_ls00.tsFullTabName_ls00 = NULL;
            }

            if (NULL != pDEC->dexFullUserInput_ls00)
            {
                delete [] pDEC->dexFullUserInput_ls00;
                pDEC->dexFullUserInput_ls00 = NULL;
            }

            if (NULL != pDEC->dexSelOrdStr_ls00)
            {
                delete [] pDEC->dexSelOrdStr_ls00;
                pDEC->dexSelOrdStr_ls00 = NULL;
            }

            if (NULL != pDEC->dexCmdFile_ls00.fsFileName_ls00)
            {
                delete [] pDEC->dexCmdFile_ls00.fsFileName_ls00;
                pDEC->dexCmdFile_ls00.fsFileName_ls00 = NULL;
            }

            if (NULL != pDEC->dexDatFile_ls00.fsFileName_ls00)
            {
                delete [] pDEC->dexDatFile_ls00.fsFileName_ls00;
                pDEC->dexDatFile_ls00.fsFileName_ls00 = NULL;
            }

            if (NULL != pDEC->dexLongSpec_ls00)
            {
                if (0 != pDEC->dexLongSpec_ls00->mfsCount_ls00)
                {
                    for (i=0; i < pDEC->dexLongSpec_ls00->mfsCount_ls00; ++i)
                    {
                        if (NULL != pDEC->dexLongSpec_ls00->mfsFiles_ls00[i]->fsFileName_ls00)
                        {
                            delete [] pDEC->dexLongSpec_ls00->mfsFiles_ls00[i]->fsFileName_ls00;
                            pDEC->dexLongSpec_ls00->mfsFiles_ls00[i]->fsFileName_ls00 = NULL;
                        }

                        if (NULL != pDEC->dexLongSpec_ls00->mfsFiles_ls00[i]->fsLongColName_ls00)
                        {
                            delete [] pDEC->dexLongSpec_ls00->mfsFiles_ls00[i]->fsLongColName_ls00;
                            pDEC->dexLongSpec_ls00->mfsFiles_ls00[i]->fsLongColName_ls00 = NULL;
                        }

                        delete pDEC->dexLongSpec_ls00->mfsFiles_ls00[i];
                    }
                }

                delete pDEC->dexLongSpec_ls00;
            }


            break;
        }
        case srFctTabExtr_els00:
        {
            MultCol = NULL;     // later on used to free column names
            
            tls00_TabExtrCmd* pTEC = (tls00_TabExtrCmd*) Struct;
            
            // release full qualified table name
            if (NULL != pTEC->texPartSpec_ls00.psFullTabName_ls00)
            {
                delete [] pTEC->texPartSpec_ls00.psFullTabName_ls00;
                pTEC->texPartSpec_ls00.psFullTabName_ls00 = NULL;
            }

            // release outfile name
            if (NULL != pTEC->texOutfile_ls00.fsFileName_ls00)
            {
                delete [] pTEC->texOutfile_ls00.fsFileName_ls00;
                pTEC->texOutfile_ls00.fsFileName_ls00 = NULL;
            }

            break;
        }
        case srFctTabLoad_els00:
        {
            MultCol = NULL;     // later on used to free column names
            
            tls00_TabLoadCmd* pTLC = (tls00_TabLoadCmd*) Struct;
            
            // release full qualified table name
            if (NULL != pTLC->tloPartSpec_ls00.psFullTabName_ls00)
            {
                delete [] pTLC->tloPartSpec_ls00.psFullTabName_ls00;
                pTLC->tloPartSpec_ls00.psFullTabName_ls00 = NULL;
            }

            // release infile name
            if (NULL != pTLC->tloInfile_ls00.fsFileName_ls00)
            {
                delete [] pTLC->tloInfile_ls00.fsFileName_ls00;
                pTLC->tloInfile_ls00.fsFileName_ls00 = NULL;
            }

            // release report file name
            if (NULL != pTLC->tloReport_ls00.fsFileName_ls00)
            {
                delete [] pTLC->tloReport_ls00.fsFileName_ls00;
                pTLC->tloReport_ls00.fsFileName_ls00 = NULL;
            }

            break;
        }
        case srFctRemark_els00:
        {
            tls00_RemarkCmd* pRemark = (tls00_RemarkCmd*) Struct;

            if (NULL != pRemark->dscRemark_ls00)
            {
                delete [] pRemark->dscRemark_ls00;
                pRemark->dscRemark_ls00 = NULL;
            }

            break;
        }

    }

    // The following MUST be done for all load structs
    if (NULL != MultCol)
    {
        for (i=0; i < MultCol->mcsCount_ls00; ++i)
        {
            tls00_Column *myCol = MultCol->mcsColumn_ls00[i];
            if (NULL != myCol)
            {
                if (NULL != myCol->colAssign_ls00.fasAsgnVal_ls00)
                {
                    delete myCol->colAssign_ls00.fasAsgnVal_ls00;
                    myCol->colAssign_ls00.fasAsgnVal_ls00 = NULL;
                }

                if (NULL != myCol->colNullCond_ls00)        // RECURSION REQUIRED
                {
                    delete myCol->colNullCond_ls00;
                    myCol->colNullCond_ls00 = NULL;
                }

                delete myCol;
                myCol = NULL;
            }
        }
    }


    // NOW delete the structure struct and return
    delete Struct;
    Struct = NULL;

    return;
}
//  ls10_Destroy()



/* Globale Functions *****************************************************************/


/*
  -----------------------------------------------------------------------------
  function:     ls10ExecChangeUser
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls10ExecChangeUser(tls00_DBInfo *DBInfo, tls00_SessionInfo *Session, tsp00_Addr ErrText)
{
    ROUTINE_DBG_MEO00 ("ls10ExecChangeUser");

    tsp00_Int4          rc                  = errOK_els00;
    tls00_SessionInfo*  pOldSession         = NULL;
    tls00_SessionInfo*  pWorkingSession     = &DBInfo->dbiSession_ls00;
    bool                ReconnectAfterError = false;        // restore old connection in case of error?

    RepServMessages *pMsgObj = RepServMessages::Instance(); // Error reporting objekt

    if (NULL != Session->siUserKey_ls00)                    //* User key? Get it.
    {
        char pszTmpText[sizeof(tsp00_ErrText) + 1];

        rc = ls15SQLGetUser(Session, pszTmpText);
        if (errOK_els00 != rc)
        {
            pMsgObj->ls98Msg(laySQLConnect_els98, ErrText, errSqlConnectUserkey_els98,
                                                           Session->siUserKey_ls00,
                                                           pszTmpText);
            return rc;
        }
        else    // delete user key info because not longer necessary
        {
            delete [] Session->siUserKey_ls00;
            Session->siUserKey_ls00 = NULL;
        }
    }

    //*
    //*     Save old connection if existing
    //*
    if (errOK_els00 != (rc = ls15CommitUser (DBInfo, ErrText)) )
    {
        // If no connection (errConnectNoServerDB_els98) nothing to save
        if (errConnectNoServerDB_els98 != rc)
        {
            return rc;      // error text is logged by ls15CommitUser
        }
    }
    else
    {
        pOldSession = new tls00_SessionInfo;            // User 'committed': save old session structure
        memcpy(pOldSession, pWorkingSession, sizeof (tls00_SessionInfo));

        // new user may not have db connection and may not be logged on
        Session->siDBConnection_ls00 = false;
        Session->siSqlStat_ls00      = false;

        ReconnectAfterError = true;                     // restore if connection fails
    }

    // Copy new session struct into 'old' DBInfo-struct and overwrite old session struct
    memcpy (pWorkingSession, Session, sizeof (tls00_SessionInfo));


    //*
    //*     New connection
    //* 1. db connection; 2. user logon
    //*
    rc = errOK_els00;

    rc = ls15Connect(DBInfo, ErrText);
    if (errOK_els00 != rc)       // no connection; error is already logged
    {
        if (true == ReconnectAfterError)
        {
            // restore old session structure
            memcpy(pWorkingSession, pOldSession, sizeof (tls00_SessionInfo));

            rc = errOK_els00;
            rc = ls15Connect(DBInfo, ErrText);     // try to connect 'old' user
            if (0 == rc)        // success restoring old connection
            {
                pMsgObj->ls98Msg( layUserConnect_els98, ErrText, errPreviousUserRestored_els98,
                                                                 pOldSession->siUserName_ls00.asCharp());
                rc = errPreviousUserRestored_els98;
            }
            // else --> caller logs error
        }
        // else --> caller logs error
    }
    else        // log connected user and server db
    {
        if (true == pWorkingSession->siNolog_ls00)
        {
            pMsgObj->ls98PrintMessage(layUserConnect_els98, msgUserConnectionLog_els98,
                                                            pWorkingSession->siUserName_ls00.asCharp(),
                                                            DBInfo->dbiSession_ls00.siDBName_ls00.asCharp(),
                                                            "NOLOG");
        }
        else
        {
            pMsgObj->ls98PrintMessage(layUserConnect_els98, msgUserConnectionLog_els98,
                                                            pWorkingSession->siUserName_ls00.asCharp(),
                                                            DBInfo->dbiSession_ls00.siDBName_ls00.asCharp(),
                                                            " ");
        }
    }

    if (NULL != pOldSession)
        delete pOldSession;

    return rc;
}
//  ls10ExecChangeUser()


/*
  -----------------------------------------------------------------------------
  function:     ls10ExecLoadFunction
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls10ExecLoadFunction(tls00_Command *SvrCmd, tls00_DBInfo *DBInfo)
{
    ROUTINE_DBG_MEO00 ("ls10ExecLoadFunction");

    RepServMessages*  MsgInstance = RepServMessages::Instance();    // Error reporting objekt

    //*
    //*     connected to db?
    //*
    if (true != DBInfo->dbiSession_ls00.siSqlStat_ls00)
    {
        MsgInstance->ls98Msg(layBasicConnect_els98, SvrCmd->cmdRpyDat_ls00,
                                                    errNoUserConnection_els98,
                                                    DBInfo->dbiSession_ls00.siDBName_ls00.asCharp());
        SvrCmd->cmdReqErr_ls00 = errNoUserConnection_els98;
    }
    else
    {
        switch (SvrCmd->cmdScanRC_ls00)
        {
            case srFctCatExtr_els00:    // Catalog extract
            {
                tls00_CatExtrCmd* pCatExtrCmd = REINTERPRET_CAST( tls00_CatExtrCmd*, SvrCmd->cmdScnDat_ls00 );
                SvrCmd->cmdReqErr_ls00 = ls31ExecCatalogExtract (pCatExtrCmd, DBInfo, SvrCmd->cmdRpyDat_ls00);
                ls10_Destroy(srFctCatExtr_els00, SvrCmd->cmdScnDat_ls00 );
                break;
            }
            case srFctCatLoad_els00:    // Catalog load
            {
                tls00_CatLoadCmd* pCatLoadCmd = REINTERPRET_CAST( tls00_CatLoadCmd*, SvrCmd->cmdScnDat_ls00 );
                SvrCmd->cmdReqErr_ls00 = ls31ExecCatalogLoad    (pCatLoadCmd, DBInfo, SvrCmd->cmdRpyDat_ls00);
                ls10_Destroy(srFctCatLoad_els00, SvrCmd->cmdScnDat_ls00 );
                break;
            }
            case srFctDatExtr_els00:    // Data extract
            {
                tls00_DatExtrCmd* DextCmd = (tls00_DatExtrCmd*) SvrCmd->cmdScnDat_ls00;
                SvrCmd->cmdReqErr_ls00 = ls27ExecDataExtract (DextCmd, DBInfo, SvrCmd->cmdRpyDat_ls00);
                ls10_Destroy(srFctDatExtr_els00, SvrCmd->cmdScnDat_ls00 );
                break;
            }
            case srFctDatLoad_els00:    // Data load
            {
                tls00_DatLoadCmd* DatLoadCmd = (tls00_DatLoadCmd*) SvrCmd->cmdScnDat_ls00;
                SvrCmd->cmdReqErr_ls00 = ls20ExecDataload (DatLoadCmd, DBInfo, SvrCmd->cmdRpyDat_ls00);
                break;
            }
            case srFctDatUpd_els00:     // Data update
            {
                tls00_DatUpdCmd* pDatUpdCmd    = REINTERPRET_CAST( tls00_DatUpdCmd* , SvrCmd->cmdScnDat_ls00 );
                SvrCmd->cmdReqErr_ls00 = ls29ExecDataUpdate ( pDatUpdCmd, DBInfo, SvrCmd->cmdRpyDat_ls00);
                break;
            }
            case srFctDBExtr_els00:     // Database extract
            {
                tls00_DBExtrCmd* pDBExtrCmd   = REINTERPRET_CAST( tls00_DBExtrCmd* , SvrCmd->cmdScnDat_ls00 );
                SvrCmd->cmdReqErr_ls00 = ls32ExecDBExtract ( pDBExtrCmd, DBInfo, SvrCmd->cmdRpyDat_ls00);
                ls10_Destroy(srFctDBExtr_els00, SvrCmd->cmdScnDat_ls00 );
                break;
            }
            case srFctDBLoad_els00:     // Database load
            {
                tls00_DBLoadCmd* pDBLoadCmd = REINTERPRET_CAST( tls00_DBLoadCmd*, SvrCmd->cmdScnDat_ls00 );
                SvrCmd->cmdReqErr_ls00  = ls32ExecDBLoad ( pDBLoadCmd, DBInfo, SvrCmd->cmdRpyDat_ls00);
                ls10_Destroy(srFctDBLoad_els00, SvrCmd->cmdScnDat_ls00 );
                break;
            }
            case srFctFastLoad_els00:   // Fastload
            {
                tls00_FloadCmd* FloadCmd = (tls00_FloadCmd*) SvrCmd->cmdScnDat_ls00;
                SvrCmd->cmdReqErr_ls00 = ls21ExecFastload (FloadCmd, DBInfo, SvrCmd->cmdRpyDat_ls00);
                ls10_Destroy(srFctFastLoad_els00, SvrCmd->cmdScnDat_ls00 );

                break;
            }
            case srFctTabExtr_els00:    // Table extract
            {
                tls00_TabExtrCmd* TabExtrCmd = (tls00_TabExtrCmd*) SvrCmd->cmdScnDat_ls00;
                SvrCmd->cmdReqErr_ls00 = ls26ExecTableExtract(TabExtrCmd, DBInfo, SvrCmd->cmdRpyDat_ls00);

                ls10_Destroy(srFctTabExtr_els00, SvrCmd->cmdScnDat_ls00 );
                break;
            }
            case srFctTabLoad_els00:    // Table load
            {
                tls00_TabLoadCmd* TLCmd = REINTERPRET_CAST(tls00_TabLoadCmd*, SvrCmd->cmdScnDat_ls00);
                SvrCmd->cmdReqErr_ls00  = ls28ExecTableLoad(TLCmd, DBInfo, SvrCmd->cmdRpyDat_ls00);

                ls10_Destroy(srFctTabLoad_els00, SvrCmd->cmdScnDat_ls00 );
                break;
            }
            case srFctTabUnlo_els00:    // Table unload
            {
                MsgInstance->ls98Msg(layBasicConnect_els98, SvrCmd->cmdRpyDat_ls00,
                                                            errNotImplemented_els98,
                                                            "Table unload");
                SvrCmd->cmdReqErr_ls00 = errNotImplemented_els98;
                break;
            }
            case srFctTabUpd_els00:     // Table update
            {
                MsgInstance->ls98Msg(layBasicConnect_els98, SvrCmd->cmdRpyDat_ls00,
                                                            errNotImplemented_els98,
                                                            "Table update");
                SvrCmd->cmdReqErr_ls00 = errNotImplemented_els98;
                break;
            }
            default:
            {
                MsgInstance->ls98Msg(layBasicConnect_els98, SvrCmd->cmdRpyDat_ls00,
                                                            errNotImplemented_els98,
                                                            "Unknown Function");
                SvrCmd->cmdReqErr_ls00 = errNotImplemented_els98;
                break;
            }
        }
    }

    return SvrCmd->cmdReqErr_ls00;
}
//  ls10ExecLoadFunction()


/*
  -----------------------------------------------------------------------------
  function:     ls10ExecSQLStmt
  -----------------------------------------------------------------------------
*/
int ls10ExecSQLStmt(tls00_Command *SvrCmd, tls00_DBInfo *DBInfo)
{
    ROUTINE_DBG_MEO00 ("ls10ExecSQLStmt");

    tsp00_Int4  rc = errOK_els00;

    if (true != DBInfo->dbiSession_ls00.siSqlStat_ls00)     // connected ?
    {
        RepServMessages::Instance()->ls98Msg(layBasicConnect_els98, SvrCmd->cmdRpyDat_ls00,
                                                                    errNoUserConnection_els98,
                                                                    DBInfo->dbiSession_ls00.siDBName_ls00.asCharp());
        rc = errNoUserConnection_els98;
    }
    else
    {
        rc = ls13ExecExtSQLStmt(DBInfo,
                                REINTERPRET_CAST(tls00_DBSqlCmd*, SvrCmd->cmdScnDat_ls00),
                                SvrCmd->cmdRpyDat_ls00);
    }
    return rc;
}
//  ls10ExecSQLStmt()

/*
  -----------------------------------------------------------------------------
  function:     ls10ExecRemark
  -----------------------------------------------------------------------------
*/
tsp00_Int4 ls10ExecRemark( tls00_Command * SvrCmd, tls00_DBInfo * DBInfo) //!!LogMessage
{
    ROUTINE_DBG_MEO00 ("ls10ExecRemark");

    tsp00_Int4  rc = errOK_els00;

    RepServMessages *pMsgInstance = RepServMessages::Instance(); 

    tls00_RemarkCmd* pRemarkCmd = (tls00_RemarkCmd*)SvrCmd->cmdScnDat_ls00;

    SAPDB_DynamicUTF8String  remark; 
    
    remark.ConvertFromASCII_Latin1(pRemarkCmd->dscRemark_ls00,
                                   pRemarkCmd->dscRemark_ls00 + pRemarkCmd->dscRemarkLen_ls00  );

    if ( remark.Length() <= 1023 )
         pMsgInstance->ls98PrintMessage (layCatalogExtract_els98, msgRemark_els98 , remark.StrPtr() );
    else
    {
         remark  = remark.SubStr(0,1022 );
         pMsgInstance->ls98PrintMessage (layCatalogExtract_els98, msgRemark_els98 , remark.StrPtr() );
    }

    ls10_Destroy(srFctRemark_els00, SvrCmd->cmdScnDat_ls00);

    return rc;
}
//  ls10ExecRemark

