/* 
  -----------------------------------------------------------------------------
 
  module: vls11.cpp
 
  -----------------------------------------------------------------------------
 
  responsible:  SteffenS

  special area: Replication Manager
 
  description:  Dispatcher module
                
  version:      7.3.

  -----------------------------------------------------------------------------
 
  copyright:    SAP AG, 1997-2001
 
  -----------------------------------------------------------------------------


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/

#include "gls00.h"
#include "hls98msg.h"
#include "hls05.h"
#include "hls11.h"
#include "hls10.h"

#include "hls14_scan.h"     // parser and scanner
#include "hls14_parse.h"

#include "hls15.h"
#include "hls99.h"


/*
  -----------------------------------------------------------------------------
  function:     ls11ParseLoadCommand
  -----------------------------------------------------------------------------
*/
int ls11ParseLoadCommand(tls00_Command *SvrCmd, tls00_DBInfo *DBInfo)
{
    ROUTINE_DBG_MEO00 ("ls11ParseLoadCommand");
    
    RepServMessages *pMsgObj = RepServMessages::Instance();             // error reporting obj
        
    tsp00_Int4  ErrCode = errOK_els00;

    ReplicationParser *Parser = new ReplicationParser(SvrCmd, DBInfo);  // Initialize parser
    if (Parser == NULL)
    {
        return errYaccAlloc_els00;
    }
    // parser generates scanner -> successful ?
    ErrCode = Parser->ls14GetErrorCode();
    if (errOK_els00 != ErrCode)
    {
        pMsgObj->ls98EFromRcToFile(layDispatch_els98, SvrCmd->cmdRpyDat_ls00, 
                                                      STATIC_CAST(ErrorNumbers, ErrCode) );
    }

    //*
    //*     parse load command
    //*
    //* StartParsing will always return OK! One must check the error value using
    //* the Parser member function ls14GetErrorCode.
    //*
    void *MyStruct = NULL;

    Parser->StartParsing(MyStruct);
    ErrCode = Parser->ls14GetErrorCode();

    delete Parser;


    //*
    //*     Log command (excluding user user , remark command) and starting time
    //*
    if (sfRelease_els00   != SvrCmd->cmdSubFunc_ls00 &&    //!!LogMessage
		srFctRemark_els00 != SvrCmd->cmdScanRC_ls00     )
    {
        tsp00_Date  Date;
        tsp00_Time  Time;
        sqldattime(Date, Time);
        tsp00_Datec szDate;
        tsp00_Timec szTime;
        szDate.p2c(Date);
        szTime.p2c(Time);
        pMsgObj->ls98PrintMessage(layDispatch_els98, msgStartCommand_els98,  szDate.asCharp(),
                                                                             szTime.asCharp());
        
        pMsgObj->ls98PrintMessage(layDispatch_els98, msgSimpleMsg_els98, SvrCmd->cmdReqDat_ls00);
    }


    if (errOK_els00 != ErrCode)         // internal error while parsing? --> check the error code
    {
        // error message is only logged to return string; we need to put it into log file
        pMsgObj->ls98EFromRcToFile(layDispatch_els98, SvrCmd->cmdRpyDat_ls00,
                                                      STATIC_CAST(ErrorNumbers, ErrCode) );
        SvrCmd->cmdReqErr_ls00 = ErrCode    ;// Adapt command return code

        // In case of syntax error we return OK anyway and a message in the command
        // structure (it's filled by the parser already).
        // In case there is some other error than syntax error we return an internal
        // error which will cause the server to be stopped.
//        if (errSQLSyntax_els98 == ErrCode)
//        {
//            ErrCode = errOK_els00;
//        }
    }
    else
    {
        // Check for errors while parsing the command

        if (errOK_els00 != SvrCmd->cmdReqErr_ls00)
        {
            // error message is only logged to return string; we need to put it into log file
            pMsgObj->ls98EFromRcToFile(layDispatch_els98, SvrCmd->cmdRpyDat_ls00,
                                                          STATIC_CAST(ErrorNumbers, SvrCmd->cmdReqErr_ls00) );
        }
        else
        {
            //*
            //*     DISPATCH
            //*
            //* No error detected. So we can work on the command sent.
            //*
            switch (SvrCmd->cmdScanRC_ls00)
            {
				case srFctRemark_els00: //!!LogMessage
                {
					 SvrCmd->cmdReqErr_ls00 = ls10ExecRemark (SvrCmd, DBInfo);
                     break;
                }
                case srFctSqlStmt_els00:
                {
                    SvrCmd->cmdReqErr_ls00  = ls10ExecSQLStmt(SvrCmd, DBInfo);
                    break;
                }
                case srCmdSession_els00:
                {
                    if (NULL != MyStruct)
                    {
                        SvrCmd->cmdReqErr_ls00 = ls10ExecChangeUser(DBInfo,
                                                                    STATIC_CAST(tls00_SessionInfo*, MyStruct),
                                                                    SvrCmd->cmdRpyDat_ls00);
                        delete STATIC_CAST(tls00_SessionInfo*, MyStruct);
                        MyStruct = NULL;
                    }

                    // Im fall, dass der kunde autcommit explizit auf ON setzt, wird ein
                    // 'COMMIT WORK' an die datenbank gesendet.
                    if (sfAutoCommitON_els00 == SvrCmd->cmdSubFunc_ls00)
                    {
                        ls15RSCommit(DBInfo, SvrCmd->cmdRpyDat_ls00);
                    }

                    break;
                }
                case srRelease_els00:
                {
                    ls15CommitUser(DBInfo, SvrCmd->cmdRpyDat_ls00);
                    break;
                }
                case srFctCatExtr_els00:
                case srFctCatLoad_els00:
                case srFctDatExtr_els00:
                case srFctDatLoad_els00:
                case srFctDatUpd_els00:
                case srFctDBExtr_els00:
                case srFctDBLoad_els00:
                case srFctFastLoad_els00:
                case srFctTabExtr_els00:
                case srFctTabLoad_els00:
                case srFctTabUnlo_els00:
                case srFctTabUpd_els00:
                {
                    SvrCmd->cmdReqErr_ls00 = ls10ExecLoadFunction(SvrCmd, DBInfo);
                    break;
                }
            }

            //*
            //*     Traceausgaben
            //*
            if (SvrCmd->cmdReqErr_ls00 != errOK_els00)
            {
                TRACE_PRNF_MLS99(("ls11ParseLoadCommand", "Error executing command '%s' (Error: %s).\n",
                                                          SvrCmd->cmdReqDat_ls00,
                                                          SvrCmd->cmdRpyDat_ls00));
            }
            else
            {
                TRACE_PRNF_MLS99(("ls11ParseLoadCommand", "Info: Successful execution of '%s'.\n",
                                                          SvrCmd->cmdReqDat_ls00));
            }
        }
    }

    return ErrCode;
}
//  ls11ParseLoadCommand()
