/*!**********************************************************************

  module: vls32_dbextract.cpp

  -----------------------------------------------------------------------

  responsible:  Hans-GeorgB
  special area: Replication Manager
  description:  Implementation of the DBEXTRACT/DBLOAD functionality
  version:      7.3.

  -----------------------------------------------------------------------

   ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end
*/

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "hls98msg.h"
#include "hls26.h"                 // ls26ExecTableExtract
#include "hls28_TLoad.h"           // ls28ExecTableLoad
#include "hls31_catalogextract.h"  // ls31ExecCatalogExtract
#include "hls32_dbextract.h"       // ls31ExecCatalogLoad


/* -----------------------------------------------------------------------------
   ls32ExecDBExtract
/-------------------------------------------------------------------------------*/
tsp00_Int4 ls32ExecDBExtract( tls00_DBExtrCmd * pDBExtrCmd, tls00_DBInfo *pDBInfo, tsp00_Addr ErrText)
{
	 ROUTINE_DBG_MEO00 ("ls32ExecDBExtract");

	 // Log   - Writer
	 RepServMessages *pMsgInstance  = RepServMessages::Instance();

	 tsp00_Int4  rc = errOK_els00;

	 // interpret <databaseextract statement>::=
	 // dbextract OUTFILE <external_file_spec> <external_file_spec>

	 tls00_CatExtrCmd *pCatExtrCmd = new tls00_CatExtrCmd;
	 memset ( pCatExtrCmd , 0, sizeof ( tls00_CatExtrCmd ) );
	 pCatExtrCmd->cexPartSpec_ls00.psSelAll_ls00 = 1;
	 pCatExtrCmd->cexOutfile_ls00 = pDBExtrCmd->dbeCatfile_ls00;

	 rc = ls31ExecCatalogExtract( pCatExtrCmd, pDBInfo, ErrText);
	 
	 if ( rc != errOK_els00 )
	 {
		 delete pCatExtrCmd;
		 return rc;
	 }

	 tls00_TabExtrCmd* pTabExtrCmd = new tls00_TabExtrCmd;
	 memset ( pTabExtrCmd , 0, sizeof ( tls00_TabExtrCmd ) );
	 pTabExtrCmd->texPartSpec_ls00.psSelAll_ls00 = 1;
	 pTabExtrCmd->texOutfile_ls00 = pDBExtrCmd->dbeDatfile_ls00;
	 
	 rc = ls26ExecTableExtract(pTabExtrCmd, pDBInfo, ErrText );
	 if  ( rc != errOK_els00 )
	 {
		   delete pTabExtrCmd;
		   return rc;
	 }

	 delete pCatExtrCmd;
	 delete pTabExtrCmd;

	 return rc;
}
/* -----------------------------------------------------------------------------
   ls32ExecDBLoad
/-------------------------------------------------------------------------------*/
tsp00_Int4 ls32ExecDBLoad( tls00_DBLoadCmd * pDBLoadCmd, tls00_DBInfo *pDBInfo, tsp00_Addr ErrText)
{

	 ROUTINE_DBG_MEO00 ("ls32ExecDBLoad");

	 // Log   - Writer
	 RepServMessages *pMsgInstance  = RepServMessages::Instance();

	 tsp00_Int4  rc = errOK_els00;

	 // interpret <databaseload statement>::=
	 // dbload  INFILE <external_infile_spec> <external_infile_spec> OUTFILE <external_file_spec>

	 tls00_CatLoadCmd *pCatLoadCmd = new tls00_CatLoadCmd;
	 memset ( pCatLoadCmd , 0, sizeof ( tls00_CatLoadCmd ) );
	 pCatLoadCmd->cloPartSpec_ls00.psSelAll_ls00 = 1;
	 pCatLoadCmd->cloInfile_ls00 = pDBLoadCmd->dblCatfile_ls00;

	 rc = ls31ExecCatalogLoad( pCatLoadCmd, pDBInfo, ErrText);

	 delete pCatLoadCmd;

	 tls00_TabLoadCmd* pTabLoadCmd =  new tls00_TabLoadCmd;
     memset ( pTabLoadCmd , 0, sizeof ( tls00_TabLoadCmd ) );
	 pTabLoadCmd->tloPartSpec_ls00.psSelAll_ls00 = 1;
	 pTabLoadCmd->tloInfile_ls00  = pDBLoadCmd->dblDatfile_ls00;
	 pTabLoadCmd->tloReport_ls00  = pDBLoadCmd->dblReport_ls00;

	 rc = ls28ExecTableLoad(pTabLoadCmd, pDBInfo, ErrText );

	 delete pTabLoadCmd;
	
	 return rc;
}

/*===========================================================================*
 *     END
 *===========================================================================*/
