/* 
  -----------------------------------------------------------------------------
 
  module: vls99.cpp
 
  -----------------------------------------------------------------------------
 
  responsible:  

  special area: Replication Manager
 
  description:  trace
                
  version:      7.3.

  -----------------------------------------------------------------------------
 
  copyright:    SAP AG, 1997-2000
 
  -----------------------------------------------------------------------------


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/


#include <stdio.h>
#include <stdarg.h>
#include <assert.h>

#include "gls00.h"
#include "hls99.h"
#include "heo01.h"

// Initialisiere die statischen klassenvariablen
bool        tls99_TraceClass::bTraceFileInit_ls99       = false;
FILE        *tls99_TraceClass::TraceFile_ls99           = NULL;     // Initialisiere zu NULL pointer
char        *tls99_TraceClass::pszTraceFileName_ls99    = NULL;
teo00_Int2  tls99_TraceClass::lSubCount = 0;


char const *tls99_TraceClass::pszPartKindName[25] = {"not used",
                                                     "appl_parameter_description",
                                                     "columnnames",
                                                     "command",
                                                     "conv_tables_returned",
                                                     "data",
                                                     "errortext",
                                                     "getinfo",
                                                     "modulname",
                                                     "page",
                                                     "parsid",
                                                     "parsid_of_select",
                                                     "resultcount",
                                                     "resulttablename",
                                                     "shortinfo",
                                                     "user_info_returned",
                                                     "surrogate",
                                                     "bdinfo",
                                                     "longdata",
                                                     "tablename",
                                                     "session_info_returned",
                                                     "output_cols_no_parameter",
                                                     "key",
                                                     "serial",
                                                     "relative_pos"};

char const *tls99_TraceClass::pszSQLModeName[6] = {"NOT USED", "SESSION", "INTERNAL",
                                                   "ANSI",     "DB2",     "ORACLE"};

char const *tls99_TraceClass::pszDataTypeName[39] = {"fixed   \t", "float   \t", "char asc\t", "char ebc\t",
                                                     "char byt\t", "rowid   \t", "long asc\t", "long ebc\t",
                                                     "long byt\t", "9\t\t",      "date    \t", "time    \t",
                                                     "12\t\t",     "timestmp\t", "14\t\t",     "15\t\t",
                                                     "16\t\t",     "17\t\t",     "18\t\t",     "19\t\t",
                                                     "20\t\t",     "21\t\t",     "22\t\t",     "boolean \t",
                                                     "char uni\t", "25\t\t",     "26\t\t",     "27\t\t",
                                                     "28\t\t",     "smallint\t", "integer \t", "vchr asc\t",
                                                     "vchr ebc\t", "vchr byt\t", "long uni\t", "35\t\t",
                                                     "vchr uni\t", "37\t\t",     "38\t\t"};

/* Lokale Funktionen: ******************************************************************/
bool tls99_TraceClass::ls99_TraceIsOn (const char Method)
{
    /* Hier soll mal die Entscheidung hin, ob die Ausgabe der aktuellen
    Traceinformation gewuenscht ist, momentan wenig sinnvoll */
    if (Method != 0x00)
    {
        return false;
    }
    else
    {
        return bTraceFileInit_ls99;
    }
}
/* ls99_TraceIsOn */


void tls99_TraceClass::ls99_TraceBuf(tsp00_Int4 TraceBufLen, tsp00_Addr TraceBuf)
{
    assert(NULL != TraceBuf);

    if (0 == TraceBufLen)
    {
        return;
    }

    tsp00_Int4 BufLine = 0;
    tsp00_Int4 j       = 0;
    tsp00_Int4 k       = 0;

    tsp00_Int4 strCount = 0;
    tsp00_Int4 tmpCount = 0;        // must be a signed because of return code of sp77sprintf

    //*
    //*     Allocate write buffer
    //*
    //* size: 1. every char can be represented as hex (i.e. using 2 bytes)
    //*       2. every row begins with 2 tabs (= 2 * count of rows)
    //*       3. every 8th char to write we insert a blank (= 7 * count of rows)
    //*       4. every column is finished with a \n (= 1* count of rows)
    //*       5. we have a trailing \0;
    //*
    size_t lBufLen = 2*TraceBufLen + (10 * ((TraceBufLen + 31)/32)) + 1;
    char *pszMyBuffer = new char[lBufLen];
    
    //*
    //*     Write trace hex
    //*
    k=0;
    while ( ((k*32) < TraceBufLen) && (tmpCount >= 0) )
    {
        BufLine = 32;           
        if ((TraceBufLen - (k*32)) < BufLine)
        {
            BufLine = TraceBufLen - (k*32);
        }

        pszMyBuffer[strCount] = pszMyBuffer[strCount + 1] = '\t';
        strCount += 2*sizeof(char);
        for (j=0; ( (j < BufLine) && (tmpCount >= 0) ); ++j)
        {
            if ((j != 0) && ((j % 4) == 0))
            {
                pszMyBuffer[strCount] = BLANK_LS00;
                strCount += sizeof(char);
            }
            strCount += (tmpCount = sp77sprintf(pszMyBuffer + strCount,
                                                lBufLen - strCount,
                                                "%.2X", STATIC_CAST(unsigned char, TraceBuf[j + (k*32)]) ) );
        }

        // In case of error (writing over string limits) sp77sprintf returns the negative count of
        // sucessfully converted parameters; so strCount is decreased by this value and we have some
        // space to insert the 0 and stop writing the buffer
        if (tmpCount < 0)
        {
            pszMyBuffer[strCount] = 0;
        }
        else
        {
            pszMyBuffer[strCount] = '\n';
            strCount += sizeof(char);
            ++k;
        }
    }
    pszMyBuffer[strCount] = 0;                  // trailing 0
    fprintf (TraceFile_ls99, pszMyBuffer);      // Ausgeben in die datei

    //*
    //*     Write trace as CHAR
    //*
    pszMyBuffer[0]  = DEF_BYTE_LS00;        // puffer reinitialisieren
    strCount        = 0;                    // Ausgabezaehler reinitialisieren
    k               = 0; 
    size_t i        = 0;
    while ((k*64) < TraceBufLen) 
    {
        BufLine = 64;           
        if ((TraceBufLen - (k*64)) < BufLine)
        {
            BufLine = TraceBufLen - (k*64);
        }

        pszMyBuffer[0] = pszMyBuffer[1] = '\t';
        for (j=0, i=(2+(k*64)); j < BufLine; ++j, ++i)
        {
            if (STATIC_CAST(unsigned char, TraceBuf[j + (k*64)]) < STATIC_CAST(char, 0x20) )
            {
                pszMyBuffer[i] = '.';
            }
            else
            {
                pszMyBuffer[i] = TraceBuf[j + (k*64)];
            }
        }
        pszMyBuffer[i] = '\n';
        ++k;
    }
    pszMyBuffer[++i] = 0;
    fprintf (TraceFile_ls99, "%s", pszMyBuffer);      // Ausgeben in die datei

    delete [] pszMyBuffer;
    pszMyBuffer = NULL;
}
/* ls99_TraceBuf */


/* Globale Funktionen: *****************************************************************/


tls99_TraceClass::tls99_TraceClass (const char *Param)
{
    RoutineName_ls99 = Param;

    if ((RoutineName_ls99 != NULL) && ls99_TraceIsOn())
    {
        fprintf(TraceFile_ls99, "%*s==> %s\n", lSubCount*2, " ",RoutineName_ls99);
    }

    lSubCount++;
}
/* Konstruktor tls99_TraceClass */


// Der destruktor laesst die tracedatei offen, da er staendig (beim verlassen einer jeden
// funktion, in dem die trace-klasse aufgerufen wird) aufgerufen wird. Die datei wird erst
// ganz zum schluss des servers geschlossen mit ls99Ende.
tls99_TraceClass::~tls99_TraceClass ()
{
    lSubCount--;

    if ( (RoutineName_ls99 != NULL) && ls99_TraceIsOn() )
    {
        fprintf(TraceFile_ls99, "%*s<== %s\n", lSubCount*2, " ",RoutineName_ls99);
    }
}
/* Destruktor ~tls99_TraceClass */


void tls99_TraceClass::ls99PrnF(const _TCHAR *Layer, const _TCHAR *format, ...)
{
    va_list arglist;

    if ((RoutineName_ls99 != NULL) && ls99_TraceIsOn())
    {
        fprintf(TraceFile_ls99, "%s: ", Layer);

        va_start(arglist, format);
        vfprintf(TraceFile_ls99, format, arglist);
        va_end(arglist);
    }
}
/* ls99PrnF */


void tls99_TraceClass::ls99Info(const _TCHAR *Layer, const _TCHAR *Text, tls00_DBInfo* DBInfo)
{
    if ((RoutineName_ls99 != NULL) && ls99_TraceIsOn())
    {
        tls00_SessionInfo* S = &DBInfo->dbiSession_ls00;

        int i = 0;
        char szBuffer[1024];        // Das sollte hier ausreichen.

        i = sprintf(szBuffer, "%s: %s\nDBInfo:\t", Layer, Text);
        i += sprintf (szBuffer + i, "'%s'\n", S->siDBName_ls00.asCharp() );
        
        if (true == DBInfo->dbiSession_ls00.siSqlStat_ls00)
        {
            i += sprintf (szBuffer + i, "\t=> Status: connected, ");
        }
        else
        {
            i += sprintf (szBuffer + i, "\t=> Status: not connected, ");
        }
        
        i += sprintf (szBuffer + i, "%i x %i Bytes per Packet, on Node '%s'\n", DBInfo->dbiPktCount_ls00,
                                                                                DBInfo->dbiPktSize_ls00,
                                                                                S->siDBNode_ls00.asCharp());

        i += sprintf (szBuffer + i, "\t=> Pid: %i, Ref: %i, Service: %i, Terminal: '%s'\n", DBInfo->dbiSqlPid_ls00,
                                                                                            DBInfo->dbiSqlRef_ls00,
                                                                                            STATIC_CAST(int, DBInfo->dbiSqlServ_ls00),
                                                                                            DBInfo->dbiSqlTerm_ls00.asCharp());

        fprintf(TraceFile_ls99, szBuffer);      // Jetzt gib's aus.
    }
}
/* ls99Info */


void tls99_TraceClass::ls99User(const _TCHAR *Layer, tls00_DBInfo *DBInfo)
{
    if ( (RoutineName_ls99 != NULL) && ls99_TraceIsOn () )
    {
        int i = 0;
        char szBuffer[1024];        // Das sollte hier ausreichen.

        tls00_SessionInfo* S = &DBInfo->dbiSession_ls00;

        i = sprintf(szBuffer, "%s\nUser Information: ", Layer);
        i += sprintf(szBuffer + i,  "'%s' at '%s'\n", S->siUserName_ls00.asCharp(), S->siDBName_ls00.asCharp() );

        i += sprintf(szBuffer + i,  "\t%i Bytes per Page, ", DBInfo->dbiPageSize_ls00);

        if (S->siAutoCommit_ls00)
        {
            i += sprintf(szBuffer + i, "AUTOCOMMIT ON, ");
        }
        else
        {
            i += sprintf(szBuffer + i, "AUTOCOMMIT OFF, ");
        }

        if (S->siMessageOn_ls00)
        {
            i += sprintf(szBuffer + i, "MESSAGE ON, ");
        }
        else
        {
            i += sprintf(szBuffer + i, "MESSAGE OFF, ");
        }
        
        if (S->siSQLMode_ls00 > sp1sm_oracle)
        {
            i += sprintf(szBuffer + i, "SQLMODE UNKNOWN (%i), ", STATIC_CAST(int, S->siSQLMode_ls00) );
        }
        else
        {
            i += sprintf(szBuffer + i, "SQLMODE %s, ", pszSQLModeName[S->siSQLMode_ls00]);
        }

        if (true == S->siNolog_ls00)
        {
            i += sprintf(szBuffer + i, "NOLOG, ");
        }


        if (S->siIgnoreTCS_ls00)
        {
            i += sprintf(szBuffer + i, "IGNORE TERMCHARSET\n");
        }
        else
        {
            i += sprintf(szBuffer + i, "USE TERMCHARSET: '%s'\n", S->siTCharSet_ls00.asCharp() );
        }

        fprintf(TraceFile_ls99, szBuffer);      // Jetzt gib's aus.
    }
}
/* ls99User */

// Die funktion gibt nur die ersten 64 bytes des packetes aus!
void tls99_TraceClass::ls99Pckt(const _TCHAR *Layer, tsp1_packet* SqlPacket)
{
    tsp1_segment *SqlSegm;
    tsp1_part    *SqlPart;
    tsp00_Int4    i, MaxBufOutput;
    char          szBuffer[512];
    szBuffer[0] = 0;
    
    
    if ((RoutineName_ls99 != NULL) && ls99_TraceIsOn() )
    {
        if (SqlPacket == NULL)
        {
            fprintf(TraceFile_ls99, "%s: \tSQLPacket: Pointer to packet is NULL\n", Layer);
            return;
        }
        else
        {
            tsp00_Int4 lCount = 0;

            lCount += sprintf(szBuffer + lCount, "%s: \tSQLPacket:\n", Layer);
            SqlSegm = &SqlPacket->sp1_segm();
            
            //* Packet-Header ausgeben
            if (1 == SqlSegm->sp1s_segm_kind())
            {   
                lCount += sprintf(szBuffer + lCount, "[REQUEST] ");
            }
            else
            {
                lCount += sprintf(szBuffer + lCount, "[RECEIVE] ");
            }
            
            switch (SqlPacket->sp1_header.sp1h_mess_code)
            {
                case csp_ascii:
                {
                    lCount += sprintf(szBuffer + lCount, "ascii, ");
                    break;
                }
                case csp_ebcdic:
                {
                    lCount += sprintf(szBuffer + lCount, "ebcdic, ");
                    break;
                }
                case csp_codeneutral:
                {
                    lCount += sprintf(szBuffer + lCount, "neutral, ");
                    break;
                }
                case csp_unicode_swap:
                {
                    lCount += sprintf(szBuffer + lCount, "unicode swapped, ");
                    break;
                }
                case csp_unicode:
                {
                    lCount += sprintf(szBuffer + lCount, "unicode, ");
                    break;
                }
                case csp_unicode_native:
                {
                    lCount += sprintf(szBuffer + lCount, "unicode native, ");
                    break;
                }
                case csp_utf8:
                {
                    lCount += sprintf(szBuffer + lCount, "utf8, ");
                    break;
                }
                default:
                {
                    lCount += sprintf(szBuffer + lCount, "unknown msg-code (%i), ",
                                                         STATIC_CAST(int, SqlPacket->sp1_header.sp1h_mess_code) );
                    break;
                }
            }
            
            switch (SqlPacket->sp1_header.sp1h_mess_swap)
            {
                case sw_full_swapped:
                {
                    lCount += sprintf(szBuffer + lCount, "full swap, ");
                    break;
                }
                case sw_part_swapped:
                {
                    lCount += sprintf(szBuffer + lCount, "part swap, ");
                    break;
                }
                case sw_normal:
                {
                    lCount += sprintf(szBuffer + lCount, "no swap, ");
                    break;
                }
                default:
                {
                    lCount += sprintf(szBuffer + lCount, "unknown swap (%i), ",
                                                         STATIC_CAST(int, SqlPacket->sp1_header.sp1h_mess_swap) );
                    break;
                }
            }

            //produces output: LOA70300,    (Segmemts: 1, Len: 104, Size: 32288)
            //                 Segment 1 (Offset 0, Len: 104, Parts: 1)
            lCount += sprintf(szBuffer + lCount, "%s%.*s,\t(Segmemts = %i, Len = %i, Size = %i)\n\t"
                                                 "Segment %i (Offset = %i, Len = %i, Parts = %i)\n\t\t",
                                                 APPL_NAME_LS00,
                                                 5, SqlPacket->sp1_header.sp1h_appl_version.asCharp(),
                                                 SqlPacket->sp1_header.sp1h_no_of_segm,
                                                 SqlPacket->sp1_header.sp1h_varpart_len,
                                                 SqlPacket->sp1_header.sp1h_varpart_size,
                                                 SqlSegm->sp1s_own_index(),
                                                 SqlSegm->sp1s_segm_offset(),
                                                 SqlSegm->sp1s_segm_len(),
                                                 SqlSegm->sp1s_no_of_parts());
            fprintf(TraceFile_ls99, szBuffer);

            szBuffer[0] = 0;
            lCount      = 0;

            //* Segment-Header ausgeben
            //produces output:
            //  sql-mode = <sql mode>, msg-type = <msg type>, cmd-type = <cmd type>, commit_immediately = <true|false> 
            if (1 == SqlSegm->sp1s_segm_kind())     // Request
            {
                lCount += sprintf(szBuffer + lCount, "sql-mode = ");
                
                if (SqlSegm->sp1c_sqlmode() > sp1sm_oracle)
                {
                    lCount += sprintf(szBuffer + lCount, "UNKNOWN (%i)", STATIC_CAST(int, SqlSegm->sp1c_sqlmode()) );
                }
                else
                {
                    lCount += sprintf(szBuffer + lCount, pszSQLModeName[SqlSegm->sp1c_sqlmode()]);
                }

                lCount += sprintf(szBuffer + lCount, ", msg-type = ");
                switch (SqlSegm->sp1c_mess_type())
                {
                    case sp1m_dbs:
                    {
                        lCount += sprintf(szBuffer + lCount, "dbs");
                        break;
                    }
                    case sp1m_parse:
                    {
                        lCount += sprintf(szBuffer + lCount, "parse");
                        break;
                    }
                    case sp1m_getparse:
                    {
                        lCount += sprintf(szBuffer + lCount, "getparse");
                        break;
                    }
                    case sp1m_syntax:
                    {
                        lCount += sprintf(szBuffer + lCount, "syntax");
                        break;
                    }
                    case sp1m_execute:
                    {
                        lCount += sprintf(szBuffer + lCount, "execute");
                        break;
                    }
                    case sp1m_getexecute:
                    {
                        lCount += sprintf(szBuffer + lCount, "getexecute");
                        break;
                    }
                    case sp1m_putval:
                    {
                        lCount += sprintf(szBuffer + lCount, "putval");
                        break;
                    }
                    case sp1m_getval:
                    {
                        lCount += sprintf(szBuffer + lCount, "getval");
                        break;
                    }
                    case sp1m_load:
                    {
                        lCount += sprintf(szBuffer + lCount, "load");
                        break;
                    }
                    case sp1m_unload:
                    {
                        lCount += sprintf(szBuffer + lCount, "unload");
                        break;
                    }
                    default:
                    {
                        lCount += sprintf(szBuffer + lCount, "unknown (%i)", (int) SqlSegm->sp1c_mess_type());
                    }
                }

                lCount += sprintf(szBuffer + lCount, ", cmd-type = ");
                switch (SqlSegm->sp1c_producer())
                {
                    case sp1pr_user_cmd:
                    {
                        lCount += sprintf(szBuffer + lCount, "user");
                        break;
                    }
                    case sp1pr_internal_cmd:
                    {
                        lCount += sprintf(szBuffer + lCount, "internal");
                        break;
                    }
                    case sp1pr_kernel:
                    {
                        lCount += sprintf(szBuffer + lCount, "kernel");
                        break;
                    }
                    case sp1pr_installation:
                    {
                        lCount += sprintf(szBuffer + lCount, "installation");
                        break;
                    }
                    default:
                    {
                        lCount += sprintf(szBuffer + lCount, "unknown (%i)", STATIC_CAST(int, SqlSegm->sp1c_producer()) );
                    }
                }
                

                lCount += sprintf(szBuffer + lCount, ", commit_immediately = ");                
                if (SqlSegm->sp1c_commit_immediately())
                {
                    lCount += sprintf(szBuffer + lCount, "true");
                }
                else
                {
                    lCount += sprintf(szBuffer + lCount, "false");
                }

                if (SqlSegm->sp1c_ignore_costwarning())
                {
                    lCount += sprintf(szBuffer + lCount, ", ignore_costwarning");
                }
                if (SqlSegm->sp1c_prepare())
                {
                    lCount += sprintf(szBuffer + lCount, ", prepare");
                }
                if (SqlSegm->sp1c_with_info())
                {
                    lCount += sprintf(szBuffer + lCount, ", with_info");
                }
                if (SqlSegm->sp1c_mass_cmd())
                {
                    lCount += sprintf(szBuffer + lCount, ", mass_cmd");
                }
                if (SqlSegm->sp1c_parsing_again())
                {
                    lCount += sprintf(szBuffer + lCount, ", parsing_again");
                }
            }
            else     // Receive
            {
                sprintf(szBuffer, "State: %.*s, Return: %i, ErrPos: %i", SqlSegm->sp1r_sqlstate().length(),
                                                                         SqlSegm->sp1r_sqlstate().asCharp(),
                                                                         SqlSegm->sp1r_returncode(),
                                                                         SqlSegm->sp1r_errorpos() );

            }       // if (1 == SqlSegm->sp1s_segm_kind())
            
            lCount += sprintf(szBuffer + lCount, "\n");
            fprintf (TraceFile_ls99, szBuffer);
            
            
            tsp00_Int4 PartOffset = sizeof(tsp1_segment_header);
    
            for (i=1; i <= SqlSegm->sp1s_no_of_parts(); ++i)
            {
                tsp00_Int4  PartLen, AssignedLen;
                tsp00_Int4  HeadSize = sizeof(tsp1_part_header);

                _TCHAR szPartKind[64];

                SqlPart = (tsp1_part*) &SqlPacket->sp1_varpart()[PartOffset];

                //*
                //*     Part-Header ausgeben
                //*
                if (SqlPart->sp1p_part_kind() > sp1pk_relative_pos)
                {
                    sprintf(szPartKind, "unknown part-kind (%i), ", STATIC_CAST(int, SqlPart->sp1p_part_kind()) );
                }
                else
                {
                    strcpy(szPartKind, pszPartKindName[SqlPart->sp1p_part_kind()]);
                }


                // The next statement produces the following output:
//      1.Part: PartKind = bdinfo, Args = 1, Offset = 0, Len = 36192
                fprintf(TraceFile_ls99, "\t%i.Part: PartKind = %s, Args = %i, Offset = %i, Len = %i\n",
                                        i,
                                        szPartKind,
                                        SqlPart->sp1p_arg_count(),
                                        PartOffset,
                                        SqlPart->sp1p_buf_len());
        
                //* Write part buffer; output limit is 64 chars
                MaxBufOutput = 64;          
                if (SqlPart->sp1p_buf_len() < MaxBufOutput)
                {
                    MaxBufOutput = SqlPart->sp1p_buf_len();
                }
                
                ls99_TraceBuf(MaxBufOutput, SqlPart->sp1p_buf().asCharp());
                
                PartLen     = SqlPart->sp1p_buf_len();
                AssignedLen = ((HeadSize + PartLen + 7) / 8 * 8 );
                PartOffset += AssignedLen;
            
            }       // for (i=1; i <= SqlSegm->sp1s_no_of_parts(); ++i)

            fprintf (TraceFile_ls99, "\n");
        }
    }
}
// ls99Pckt()

// text ist ein nullterminierter string, Len ist die laenge von Buf
void tls99_TraceClass::ls99Buff(const _TCHAR *Layer, const _TCHAR *Text, tsp00_Int4 Len, _TCHAR *Buf)
{
    if ((RoutineName_ls99 != NULL) && ls99_TraceIsOn () )
    {
        fprintf(TraceFile_ls99, "%s: %s\tbuf(%i)\n", Layer, Text, Len);

        ls99_TraceBuf(Len, Buf);
    }
}
/* ls99Buff */


void tls99_TraceClass::ls99Tab2(const _TCHAR *Layer, tls00_TableDescription *Table)
{
    char        szBuffer[1024] = {'\0'};;       // Das sollte fuer unsere zwecke der zwischenpufferung genuegen
    tsp00_Int2  lBufCnt = 0;

    if ((RoutineName_ls99 != NULL) && ls99_TraceIsOn() )
    {
        lBufCnt =  sprintf(szBuffer, "%s\tTableDescription:\n", Layer);
        lBufCnt += sprintf(szBuffer + lBufCnt, "Tableowner: '%s'\n", Table->tdAuthid_ls00.asCharp());
        lBufCnt += sprintf(szBuffer + lBufCnt, "Tablename:  '%s'\n", Table->tdTablename_ls00.asCharp());
        lBufCnt += sprintf(szBuffer + lBufCnt, "FieldCount:  %i\t", Table->tdFieldCount_ls00 );
        lBufCnt += sprintf(szBuffer + lBufCnt, "KernelSwap:  %i\t", STATIC_CAST(int, Table->tdKernelSwap_ls00) );
        lBufCnt += sprintf(szBuffer + lBufCnt, "KernelLock:  %i\t", STATIC_CAST(int, Table->tdKernelLoc_ls00) );
        lBufCnt += sprintf(szBuffer + lBufCnt, "ConstrCount: %i\n", STATIC_CAST(int, Table->tdConstraintCount_ls00) );
        lBufCnt += sprintf(szBuffer + lBufCnt, "No\tstmt_no\tdef_no\tfrac\tlen\tio_len\tlentype\tmode\ttype\t\tdefault\tname\n");
        fprintf(TraceFile_ls99, szBuffer);
        lBufCnt = 0;

        for (int i=0; i<Table->tdFieldCount_ls00; i++)
        {
            tsp00_DataType DBDataType = Table->tdFieldsArray_ls00[i]->fiBasics_ls00.fbDataType_ls00;

            lBufCnt = sprintf(szBuffer, "%i\t%i\t%i\t%i\t%i\t%i\t",
                                        i,
                                        STATIC_CAST(int, Table->tdFieldsArray_ls00[i]->fiStmtColNo_ls00),
                                        STATIC_CAST(int, Table->tdFieldsArray_ls00[i]->fiDefColNo_ls00),
                                        STATIC_CAST(int, Table->tdFieldsArray_ls00[i]->fiBasics_ls00.fbFraction_ls00),
                                        STATIC_CAST(int, Table->tdFieldsArray_ls00[i]->fiBasics_ls00.fbLength_ls00),
                                        STATIC_CAST(int, Table->tdFieldsArray_ls00[i]->fiBasics_ls00.fbInOutLength_ls00) ); 
            
            switch (Table->tdFieldsArray_ls00[i]->fiLengthType_ls00)
            {
                case sp7_fix_len:
                    lBufCnt += sprintf(szBuffer + lBufCnt, "fix   \t");
                    break;
                case sp7_var_len:
                    lBufCnt += sprintf(szBuffer + lBufCnt, "var   \t");
                    break;
                case sp7_var_len_long:
                    lBufCnt += sprintf(szBuffer + lBufCnt, "var(l)\t");
                    break;
                default:
                    lBufCnt += sprintf(szBuffer + lBufCnt, "%i\t\t", (int) Table->tdFieldsArray_ls00[i]->fiLengthType_ls00);
                    break;
            }
            switch (Table->tdFieldsArray_ls00[i]->fiColMode_ls00)
            {
                case mod_syskey:
                    lBufCnt += sprintf(szBuffer + lBufCnt, "skey  \t");
                    break;
                case mod_key:
                    lBufCnt += sprintf(szBuffer + lBufCnt, "key   \t");
                    break;
                case mod_mandatory:
                    lBufCnt += sprintf(szBuffer + lBufCnt, "man   \t");
                    break;
                case mod_optional:
                    lBufCnt += sprintf(szBuffer + lBufCnt, "opt   \t");
                    break;
                case mod_dropped:
                    lBufCnt += sprintf(szBuffer + lBufCnt, "drop  \t");
                    break;
                default:
                    lBufCnt += sprintf(szBuffer + lBufCnt, "%i\t\t", (int) Table->tdFieldsArray_ls00[i]->fiColMode_ls00);
                    break;
            }

            // Write data type
            if (DBDataType > dabaptabhandle)
            {
                lBufCnt += sprintf(szBuffer + lBufCnt, "%i\t\t", STATIC_CAST(int, DBDataType));
            }
            else
            {
                lBufCnt += sprintf(szBuffer + lBufCnt, pszDataTypeName[DBDataType]);
            }

            switch (Table->tdFieldsArray_ls00[i]->fiDefault_ls00)
            {
                case sp7_no_default:
                    lBufCnt += sprintf(szBuffer + lBufCnt, "no def\t");
                    break;
                case sp7_default_user:
                    lBufCnt += sprintf(szBuffer + lBufCnt, "user  \t");
                    break;
                case sp7_default_usergroup:
                    lBufCnt += sprintf(szBuffer + lBufCnt, "group \t");
                    break;
                case sp7_default_date:
                    lBufCnt += sprintf(szBuffer + lBufCnt, "date  \t");
                    break;
                case sp7_default_time:
                    lBufCnt += sprintf(szBuffer + lBufCnt, "time  \t");
                    break;
                case sp7_default_stamp:
                    lBufCnt += sprintf(szBuffer + lBufCnt, "stamp \t");
                    break;
                case sp7_default_timestamp:
                    lBufCnt += sprintf(szBuffer + lBufCnt, "timest\t");
                    break;
                case sp7_default_uid:
                    lBufCnt += sprintf(szBuffer + lBufCnt, "uid   \t");
                    break;
                case sp7_default_true:
                    lBufCnt += sprintf(szBuffer + lBufCnt, "true  \t");
                    break;
                case sp7_default_false:
                    lBufCnt += sprintf(szBuffer + lBufCnt, "false \t");
                    break;
                case sp7_other_default:
                    lBufCnt += sprintf(szBuffer + lBufCnt, "other \t");
                    break;
                case sp7_default_serial:
                    lBufCnt += sprintf(szBuffer + lBufCnt, "serial\t");
                    break;
                default:
                    lBufCnt += sprintf(szBuffer + lBufCnt, "%i\t\t", (int) Table->tdFieldsArray_ls00[i]->fiDefault_ls00);
            }
            lBufCnt += sprintf(szBuffer + lBufCnt, "%s\n", Table->tdColName_ls00[i]);

            fprintf(TraceFile_ls99, szBuffer);  // End of for-loop reached -> flush the stuff
            lBufCnt = 0;
        }
        fprintf (TraceFile_ls99, "\n");
    }
}
/* ls99Tab2 */




void tls99_TraceClass::ls99TabParam(const _TCHAR *Layer, tls00_TableDescription *Table)
{
    char        szBuffer[1024] = {'\0'};;       // Das sollte fuer unsere zwecke der zwischenpufferung genuegen
    tsp00_Int2  lBufCnt = 0;

    if ((RoutineName_ls99 != NULL) && ls99_TraceIsOn() )
    {
        lBufCnt =  sprintf(szBuffer, "%s\tTableDescription:\n", Layer);
        lBufCnt += sprintf(szBuffer + lBufCnt, "Tableowner: '%s'\n", Table->tdAuthid_ls00.asCharp());
        lBufCnt += sprintf(szBuffer + lBufCnt, "Tablename:  '%s'\n", Table->tdTablename_ls00.asCharp());
        lBufCnt += sprintf(szBuffer + lBufCnt, "FieldCount:  %i\t", Table->tdFieldCount_ls00 );
        lBufCnt += sprintf(szBuffer + lBufCnt, "KernelSwap:  %i\t", STATIC_CAST(int, Table->tdKernelSwap_ls00) );
        lBufCnt += sprintf(szBuffer + lBufCnt, "KernelLock:  %i\t", STATIC_CAST(int, Table->tdKernelLoc_ls00) );
        lBufCnt += sprintf(szBuffer + lBufCnt, "ConstrCount: %i\n", STATIC_CAST(int, Table->tdConstraintCount_ls00) );
        lBufCnt += sprintf(szBuffer + lBufCnt, "no\tpos\tmode\t\tiotype\tdatatype\tfrac\tlen\tio_len\tname\n");
        fprintf(TraceFile_ls99, szBuffer);
        lBufCnt = 0;

        tls00_ParamInfo *PI = NULL;

        for (int i=0; i<Table->tdFieldCount_ls00; i++)
        {
            PI = Table->tdParamArray_ls00[i];
            lBufCnt = sprintf(szBuffer, "%i.\t%i\t", i, PI->piBufpos_ls00);

            if (true == PI->piMode_ls00.includes(sp1ot_mandatory))
            {
                lBufCnt += sprintf(szBuffer + lBufCnt, "mandatory\t");
            }
            else if (true == PI->piMode_ls00.includes(sp1ot_optional))
            {
                lBufCnt += sprintf(szBuffer + lBufCnt, "optional \t");
            }
            else if (true ==  PI->piMode_ls00.includes(sp1ot_default))
            {
                lBufCnt += sprintf(szBuffer + lBufCnt, "default  \t");
            }
            else        // sp1ot_escape_char
            {
                lBufCnt += sprintf(szBuffer + lBufCnt, "escape   \t");
            }


            switch (PI->piIoType_ls00)
            {
                case sp1io_input:
                {
                    lBufCnt += sprintf(szBuffer + lBufCnt, "input\t");
                    break;
                }
                case sp1io_output:
                {
                    lBufCnt += sprintf(szBuffer + lBufCnt, "output\t");
                    break;
                }
                default:
                {
                    lBufCnt += sprintf(szBuffer + lBufCnt, "inout\t");
                    break;
                }
            }

            // Write data type
            if (PI->piBasicInfo_ls00.fbDataType_ls00 > dabaptabhandle)
            {
                lBufCnt += sprintf(szBuffer + lBufCnt, "%i\t\t", STATIC_CAST(int, PI->piBasicInfo_ls00.fbDataType_ls00));
            }
            else
            {
                lBufCnt += sprintf(szBuffer + lBufCnt, pszDataTypeName[PI->piBasicInfo_ls00.fbDataType_ls00]);
            }


            lBufCnt += sprintf(szBuffer + lBufCnt, "%i\t%i\t%i\t",PI->piBasicInfo_ls00.fbFraction_ls00,
                                                                  PI->piBasicInfo_ls00.fbLength_ls00,
                                                                  PI->piBasicInfo_ls00.fbInOutLength_ls00);

            if (NULL != Table->tdColName_ls00[i])
            {
                lBufCnt += sprintf(szBuffer + lBufCnt, "%s\n", Table->tdColName_ls00[i]);
            }
            else
            {
                lBufCnt += sprintf(szBuffer + lBufCnt, "\n");
            }

            fprintf(TraceFile_ls99, szBuffer);  // End of for-loop reached -> flush the stuff
            lBufCnt = 0;
        }
        fprintf (TraceFile_ls99, "\n");
    }
}
// ls99TabParam()


// Hier wird genau einmal das trace file geoeffnet. Der dateizeiger ist eine statische variable
// der klasse und kann/sollte von allen objekten genutzt werden.
void tls99_TraceClass::ls99Init (const _TCHAR *Layer, const _TCHAR *Text, const _TCHAR *FileName)
{
    bTraceFileInit_ls99 = false;        // Initialisiere die statische variable
    
    tsp00_Date Date;
    tsp00_Time Time;
    sqldattime(Date, Time);
    
    tsp00_Datec szDate;
    tsp00_Timec szTime;
    szDate.p2c(Date);
    szTime.p2c(Time);
    
    // Hole speicher fuer den dateinamen. Dieser muss in der ende-funktion wieder
    // freigegeben werden!!! Dann uebernimm den gelieferten dateinamen in die interne variable.
    tsp01_RteError  RteError;
    tsp00_Pathc     tracefilepath;
        
    sqlGetIndependentWrkPath(tracefilepath,
                             TERM_WITH_DELIMITER_EO01,
                             &RteError);
        
    size_t len = tracefilepath.length() + strlen(FileName);
        
    pszTraceFileName_ls99 = new char[len + 1];      // Initialisiere die member variablen
    pszTraceFileName_ls99[0] = '\0';

    sprintf(pszTraceFileName_ls99, "%s%s", tracefilepath.asCharp(), FileName);
    
    // Oeffne die trace-ausgabedatei und "mache sie unbuffered". Wenn das schiefgeht, rumsstss ordentlich.
    // Es wird allerdings abgefangen durch die TraceFileInit-variable
    if (NULL != (TraceFile_ls99 = fopen(pszTraceFileName_ls99, "w")) )
    {
        // Keine (!) fehlerbehandlung hier straeflicherweise. Dazu muss allerdings
        // auch erst die aufrufphilosophie der trace-klasse geaendert werden
        setvbuf( TraceFile_ls99, NULL, _IONBF, 0 );

        fprintf(TraceFile_ls99, "%s: %s	%s%s\n", Layer, Text, szDate.asCharp(), szTime.asCharp());

        // Initialisierung hat geklappt. Schalte traceausgabe frei.
        bTraceFileInit_ls99 = true;
    }
}
// ls99Init()




// Hier wird schlussendlich die datei geschlossen, falls das nicht durch einen
// systemabsturz (:-(( ) sowieso schon geschehen ist. Diese funktion sollte genau
// einmal im gesamten server aufgerufen werden, naemlich am ende.
// Der inhalt irgendwelcher dateipuffer muss NICHT geflusht werden, da wir hier ungepuffert
// arbeiten.
void tls99_TraceClass::ls99Ende(const _TCHAR *Layer, const _TCHAR *Text)
{
    if ((RoutineName_ls99 != NULL) && ls99_TraceIsOn () )
    {
        fclose (TraceFile_ls99);
    }

    // Freigeben der in INIT angeforderten resourcen.
    delete pszTraceFileName_ls99;
    pszTraceFileName_ls99 = NULL;

    bTraceFileInit_ls99 = false;
}
// ls99Ende()
