/****************************************************************************

  module      : gos542.h

  -------------------------------------------------------------------------

  responsible : FrankH

  special area: Intel PSE 36 support 
  description : special "vblockio"


  last changed: 1999-08-19  9:39
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1999



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


#ifndef GOS542_H
#define GOS542_H



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

//
// +---------------------------------------------------------------+
// |   PSE36 MEMORY CONTROL                                        |
// +---------------------------------------------------------------+
//

typedef struct PSE36_Statistics
{
  tsp_int4      TotalPages ;           // ==> Total size of PSE36 in Pages
  tsp_int4      MaxDataPages ;         // ==> Size of PSE36 in Data-Pages
  tsp_int4      NumChangedDataPages;   // ==> Pages to save to devspace

  /* PTS 1103699 */

  tsp00_Longreal NumDataReads ;
  tsp00_Longreal NumDataReadsFail;
  tsp00_Longreal NumDataWrites ;
  tsp00_Longreal NumDataWritesFail;
} tos542_PSE36_Statistics;

typedef struct PSE36_control_struct
  {
  PVOID                           PhysPagesFrameNumbers; // ==> AWE virtual adress ro store page frame numbers
  int                             PhysPagesPerDBPage;    // ==> Number of physical pages per DB page 
  PPVOID                          AWEMappingAdresses;    // ==> Mapping addresses ( one DB-Page for each Task )
  char                           *DataPageArray;         // ==> Array with status information
                                                         //     for each datapage in PSE36
  tsp_int4                        LowestChangedDataPage; // ==> lowest changed DB-Page in PSE36
#ifdef OBJLOG_CACHE_INFO_USED
  char                           *ObjLogPageArray;       // ==> Array with status information
                                                         //     for each logpage in PSE36
#endif
  tsp_int4                        InfoPagePos;           // ==> reserved for informations       
  tsp_int4                        WriteMode ;            // ==> write operations to PSE36 AND/OR DISK
  long                            Exclusive;             // ==> used to grant exclusive access
  tsp_int4                        DBPageSize ;
  tos542_PSE36_Statistics         PSE36Stat;             // ==> statistics
  } tos542_PSE36Ctrl;

/*===========================================================================*
 *  FUNCTION PROTOTYPES                                                      *
 *===========================================================================*/

externC DWORD os542PSE36Init ();

/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/

#endif  /* GOS542_H */
