/* @lastChanged: "1999-04-21  9:36"
 * @filename:    vos35.cpp
 * @purpose:     "DCOMx_Functions"
 * @release:     7.2.0.0
 * @see:         ""
 *
 * @copyright:   "Copyright by SAP AG, 1998"


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 */

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/


#include "gos00.h"
#include "gsp00.h"
#include "geo00k.h"
#include "vsp002c.h"
#include "veo00c.h"
#include "geo002.h"
#include "geo007_1.h"
#include "geo007_2.h"
#include "vos00kc.h"

#include "heo46.h"
#include "heo35.h"
#include "gos35k.h"
#include "heo36.h"
#include "heo38.h"
#include "hos39.h"


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#ifdef NO_INLINES
# define _INLINE
#else
# define _INLINE	inline
#endif

#define MAX_DCOMI_WAIT_TIME               10
#define MAX_DCOMI_REMOVE_WAIT_TIME        20

#define MAX_DCOM_ERRCODE                        6
#define ERRCODE_DCOM_CREATE_DCOMI               0
#define ERRCODE_DCOM_RESUME_DCOMI               1
#define ERRCODE_DCOMI_CLOSE_EVENT_SEM           2
#define ERRCODE_DCOM_MAX_DCOMIS_REACHED         3
#define ERRCODE_DCOM_GET_DCOMI_CTRL             4
#define ERRCODE_DCOM_REM_UKT_DCOMi_NOT_ALLOWED  5


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/



/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/



/*===========================================================================*
 *  GLOBAL VARIABLES                                                         *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL VARIABLES                                                          *
 *===========================================================================*/

static CHAR* os35k_ErrTab[ MAX_DCOM_ERRCODE ] =
                        {
                        ERRMSG_DCOM_CREATE_DCOMI,             // - ERRCODE_DCOM_CREATE_DCOMI
                        ERRMSG_DCOM_RESUME_DCOMI,             // - ERRCODE_DCOM_RESUME_DCOMI
                        ERRMSG_DCOMI_CLOSE_EVENT_SEM,         // - ERRCODE_DCOMI_CLOSE_EVENT_SEM
                        ERRMSG_DCOM_MAX_DCOMIS_REACHED,       // - ERRCODE_DCOM_MAX_DCOMIS_REACHED
                        ERRMSG_DCOM_GET_DCOMI_CTRL,           // - ERRCODE_DCOM_GET_DCOMI_CTRL
                        ERRMSG_DCOM_REM_UKT_DCOMi_NOT_ALLOWED // - ERRCODE_DCOM_REM_UKT_DCOMi_NOT_ALLOWED
                        };


/*===========================================================================*
 *  LOCAL FUNCTIONS (PROTOTYPES)                                             *
 *===========================================================================*/

static _INLINE void os35k_ReqFromDCOMxToUKT    ( tos00k_Dlq                *pRequest );
static _INLINE void os35k_ReqToDCOM0           ( tos00k_Dlq                *pRequest );
static _INLINE void os35k_ReqToDCOMi           ( tos35k_DCOMiCtrl          *pDCOMiCtrl,
                                                 tos00k_Dlq                *pRequest);
static _INLINE tos00k_Dlq* os35k_DCOMReceive   ( tos00k_DCOMQueue          **ppDCOMWorkQue,
                                                 ULONG                     *pDCOMQueLen, 
                                                 LONG                      *pExclusive,
                                                 tos00k_ThreadCtrlHeader   *pThrdCtrlHeader );
static tos35k_DCOMiCtrl* os35k_DCOMiInit       ( tos00k_TaskCtrl           *pTaskCtrl,
                                                 ULONG                     ReqType,
                                                 tsp00_ErrText             &Errtext );

static tos35k_DCOMiCtrl* os35k_GetDCOMiCtrl    ( tos00k_Dlq                *pRequest );
static void os35k_KillDCOMi                    ( tos35k_DCOMiCtrl          **ppDCOMiCtrl,
                                                 ULONG                     Relationship );
static void os35k_ReleaseDCOMiCtrl             ( ULONG                     Relationship,
                                                 tos35k_DCOMiCtrl          **ppDCOMiCtrl );

static bool os35k_DCOMiCreate                  ( tos00k_Dlq                *pRequest );
static bool os35k_DCOMiRemove                  ( tos00k_Dlq                *pRequest );
static void os35k_DCOMiTerminate               ( tos35k_DCOMiCtrl          *ppDCOMiCtrl,
                                                 tos00k_DlqDCOMCtrlReqArgs *DCOMCtrlRequest );
static void _System os35k_DCOMi                ( tos35k_DCOMiCtrl          *pDCOMiCtrl );

static LONG os35k_AllocDCOMiCtrl               ( tos35k_DCOMiCtrl          **ppDCOMiCtrl );
static _INLINE void os35k_FreeDCOMiCtrl        ( tos35k_DCOMiCtrl          *pDCOMiCtrl );


/*===========================================================================*
 *  GLOBAL FUNCTIONS (CODE)                                                  *
 *===========================================================================*/


externC void sql35_VDCOMTransfer ( void             *pTransferBuffer,
                                   bool             *pOk,
                                   tsp00_ErrText    *pErrtext )
{
  ROUTINE_DBG_MSP00 ("sql35_VDCOMTransfer");
  
  *pOk = sql35_VDCOMiRequest ( REQ_DCOM_TRANSFER_EO00K, pTransferBuffer, *pErrtext );

  return;
}

/*---------------------------------------------------------------------------*/

externC bool sql35_RemoveDCOMi ( void )
{
  ROUTINE_DBG_MSP00 ("sql35_RemoveDCOMi");

  tos00k_TaskCtrl                  *pTaskCtrl = THIS_UKT_CTRL->pCTask;
  tos00k_Dlq                       &Request   = pTaskCtrl->SpecDLQElem1;

  if ( pTaskCtrl->pDCOMiCtrl == NULL )
    return false;  // --- DCOMi not running!

  Request.pTaskCtrl                                  = pTaskCtrl;
  Request.ulReqType                                  = REQ_DCOM_REMOVE_DCOMI_EO00K;
  Request.ReqArgs.DCOMCtrlRequest.Relationship_os00k = TASK_REL_OS35K;
  
  os35k_ReqToDCOM0( &Request );
  GOTO_DISP(pTaskCtrl->pUKT);
  
  // --- wrong reply?
  if ( pTaskCtrl->pUKT->pCurrReq != &Request )
  {
    MSGCD (( ERR_UNKNOWN_REQUEST ));
    ABORT();
  }
  
  if ( Request.ReqArgs.DCOMCtrlRequest.Ok_os00k == false )
  {
    DBG3 (( MF__, "Can't create DCOMi for UKT %d", pTaskCtrl->pUKT->ulUKTIndex ));
    MSGD  (( ERR_ERROR_TEXT, os35k_ErrTab[Request.ReqArgs.DCOMRequest.ErrCode_os00k] ));
  }

  return Request.ReqArgs.DCOMCtrlRequest.Ok_os00k;
}

/*---------------------------------------------------------------------------*/

externC void sql35_DCOM0 ( void )
{
  ROUTINE_DBG_MSP00 ("sql35_DCOM0");

  LONG                          rc = NO_ERROR;
  tos00k_Dlq                    *pRequest;
  tos35k_DCOM0Ctrl              &DCOM0          = *kgs.pDCOM0;

  #pragma warning( disable : 4509 )
  __try 
   {
    // --- Create event semaphore (SET) for DCOM0
    rc = sql41c_create_event_sem ( &DCOM0.ThrdCtrlHdr_os35k.hevSem, NULL,
                                   NULL, 0, false, NULL );
    if (rc != NO_ERROR)
      ABORT();
    
    // --- set the thread priority
    rc = SET_THREAD_PRIO ( DCOM0.ThrdCtrlHdr_os35k.lThrdPrio,
                           DCOM0.ThrdCtrlHdr_os35k.Tid,
                           DCOM0.ThrdCtrlHdr_os35k.hThrd );
    
    if ( rc != NO_ERROR )
    {
      MSGCD(( ERR_SETTING_THRD_PRIO, DCOM0.ThrdCtrlHdr_os35k.Tid));
      DBG1 ((MF__, "Cannot set the thread priority, TID: %u",
             DCOM0.ThrdCtrlHdr_os35k.Tid));
      ABORT();
    }

    MSGD (( INFO_KERNEL_THREAD_STARTED, "DCOM0" ));

    // --- set thread control info
    DCOM0.ThrdCtrlHdr_os35k.ThrdState   = KT_RUNNING;
    DCOM0.ThrdCtrlHdr_os35k.ulStartTime = kgs.ulCurrTime;

    // --- Runs forever
    for (;;)
    {
      DBG3 (( MF__, "DCOM0.pDCOMWorkQue           is 0x%x", DCOM0.pDCOMWorkQue ));
      DBG3 (( MF__, "DCOM0.pDCOMReqQue            is 0x%x", DCOM0.pDCOMReqQue ));
      DBG3 (( MF__, "DCOM0.pDCOMWorkQue->pRequest is 0x%x", DCOM0.pDCOMWorkQue->pRequest ));

      //  --- Wait for a request to DCOM0
      pRequest = os35k_DCOMReceive  ( &DCOM0.pWorkQue_os35k,
                                      &DCOM0.QueLen_os35k,
                                      &DCOM0.Exclusive_os35k,
                                      &DCOM0.ThrdCtrlHdr_os35k );

      switch ( pRequest->ulReqType )
      {
      case REQ_DCOM_CREATE_DCOMI_EO00K  :
        os35k_DCOMiCreate ( pRequest );
        os35k_ReqFromDCOMxToUKT(pRequest); // -- send error / ok back to UKT
        break;
        
      case REQ_DCOM_REMOVE_DCOMI_EO00K  :
        os35k_DCOMiRemove ( pRequest );
        os35k_ReqFromDCOMxToUKT(pRequest); // -- send error / ok back to UKT
        break;
        MSGCD (( ERR_ILLEGAL_REQUEST, "DCOM0", pRequest->ulReqType ));
      }
      DBG3 (( MF__, "next in loop" ));
    }
  }
  __except( sql50k_excepthandler(GetExceptionInformation()) ) {;}

  MSGD (( INFO_KERNEL_THREAD_STOPPED, "DCOM0" ));

  EXITTHREAD ( 0, DCOM0.ThrdCtrlHdr_os35k.hThrd );

  return;
}

/*===========================================================================*
 *  LOCAL FUNCTIONS (CODE)                                                   *
 *===========================================================================*/

externC bool sql35_VDCOMiRequest ( tsp00_Uint4     ReqType, 
                                   void            *pParamInfo,
                                   tsp00_ErrText   &Errtext )
{
  ROUTINE_DBG_MSP00 ("sql35_VDCOMiRequest");

  tos35k_DCOMiCtrl                 *pDCOMiCtrl;
  tos00k_UKTCtrl                   *pUKT      = THIS_UKT_CTRL;
  tos00k_TaskCtrl                  *pTaskCtrl = pUKT->pCTask;
  tos00k_Dlq                       &Request   = pTaskCtrl->SpecDLQElem1;
  bool                             Ok         = false;

  pTaskCtrl->TaskState = TSK_DCOM_OBJ_CALLED;

  pDCOMiCtrl = os35k_DCOMiInit ( pTaskCtrl, ReqType, Errtext );

  if ( pDCOMiCtrl != NULL )
    {
    Request.pTaskCtrl                             = pTaskCtrl;
    Request.ulReqType                             = ReqType;
    Request.ReqArgs.DCOMRequest.pParam_os00k      = pParamInfo;
    Request.ReqArgs.DCOMRequest.Ok_os00k          = true;
    Request.ReqArgs.DCOMRequest.APIRetCode_os00k  = NO_ERROR;

    os35k_ReqToDCOMi ( pDCOMiCtrl, &Request );

    // --- goto dispatcher
    GOTO_DISP(pUKT);

    if ( pUKT->pCurrReq != &Request ) // --- wrong reply?
      {
      MSGCD (( ERR_UNKNOWN_REQUEST ));
      DBG1  (( MF__, "[T:0x%03u] pRequest: 0x%x (SEND) != pUKT->pCurrReq: 0x%x (REPLY)",
               pTaskCtrl->ulTaskIndex, pRequest, pUKT->pCurrReq ));
      ABORT();
      }

    Ok = Request.ReqArgs.DCOMRequest.Ok_os00k;

    if ( Ok == false )
      {
      eo46BuildPascalErrorStringRC ( (teo00_Byte *)(char *)Errtext,
        os35k_ErrTab[Request.ReqArgs.DCOMRequest.ErrCode_os00k],
        Request.ReqArgs.DCOMRequest.APIRetCode_os00k );
    }

    DBG3 (( MF__, "[T:0x%03u] Returned from request pUKT: %x",
            pTaskCtrl->ulTaskIndex, pUKT ));
    }

 pTaskCtrl->TaskState = TSK_RUNNING;

 return Ok;
}


/*---------------------------------------------------------------------------*/


static _INLINE void os35k_ReqFromDCOMxToUKT ( tos00k_Dlq      *pRequest )
{
  ROUTINE_DBG_MSP00 ("os35k_ReqFromDCOMxToUKT");

  DBG3 (( MF__, "Send pRequest: 0x%x to TID:%u, [T:0x%03u]", pRequest,
          pRequest->pTaskCtrl->pUKT->ThrdCtrlHeader.Tid,
          pRequest->pTaskCtrl->ulTaskIndex ));

  sql73k_UKT_enqu ( pRequest->pTaskCtrl, pRequest );
  sql72k_wake_ukt ( pRequest->pTaskCtrl->pUKT );

  return;
}

/*---------------------------------------------------------------------------*/

static _INLINE void os35k_ReqToDCOM0 ( tos00k_Dlq      *pRequest )
{
  ROUTINE_DBG_MSP00 ("os35k_ReqToDCOM0");
  tos35k_DCOM0Ctrl             &DCOM0 = *kgs.pDCOM0;

  DBG3(( MF__, "[T:0x%03u] Send pRequest: 0x%x, using (DCOM0) pDCOMReqQue: 0x%x",
         pRequest->pTaskCtrl->ulTaskIndex, pRequest, DCOM0.pDCOMReqQue ));

  WAIT_UNTIL_ALONE( &DCOM0.Exclusive_os35k );

  DCOM0.pReqQue_os35k->pRequest = pRequest;
  DCOM0.pReqQue_os35k           = DCOM0.pReqQue_os35k->pNextElem;
  DCOM0.QueLen_os35k++;

  CLEARLOCK( &DCOM0.Exclusive_os35k );

  sql72k_wake_kt ( &DCOM0.ThrdCtrlHdr_os35k );

  return;
}

/*---------------------------------------------------------------------------*/


static _INLINE void os35k_ReqToDCOMi ( tos35k_DCOMiCtrl *pDCOMiCtrl,
                                       tos00k_Dlq       *pRequest )
{
  ROUTINE_DBG_MSP00 ("os35k_ReqToDCOMi");

  WAIT_UNTIL_ALONE( &pDCOMiCtrl->Exclusive_os35k );

  pDCOMiCtrl->pReqQue_os35k->pRequest = pRequest;
  pDCOMiCtrl->pReqQue_os35k           = pDCOMiCtrl->pReqQue_os35k->pNextElem;
  pDCOMiCtrl->QueLen_os35k++;

  CLEARLOCK( &pDCOMiCtrl->Exclusive_os35k );

  if (pDCOMiCtrl->QueLen_os35k > pDCOMiCtrl->MaxQueLen_os35k)
      pDCOMiCtrl->MaxQueLen_os35k = pDCOMiCtrl->QueLen_os35k;

  sql72k_wake_kt ( &pDCOMiCtrl->ThrdCtrlHdr_os35k );

  return;
}

/*---------------------------------------------------------------------------*/

/*
 *   Tries to get a request from the DCOMx 'DCOMWork' queue, if no request
 *   is available then this thread should go to sleep until it is waked
 *   by a task/thread making a DCOM-request to it.
 */
static _INLINE tos00k_Dlq* os35k_DCOMReceive ( tos00k_DCOMQueue        **ppDCOMWorkQue,
                                               ULONG                   *pDCOMQueLen,
                                               LONG                    *pExclusive,
                                               tos00k_ThreadCtrlHeader *pThrdCtrlHeader)
{
  ROUTINE_DBG_MSP00 ("os35k_DCOMReceive");
  
  LONG       rc = NO_ERROR;
  PDLQ_REC   pRequest = (*ppDCOMWorkQue)->pRequest;
  
  while ( pRequest == (PDLQ_REC) 0 )
  {
    DBG3 (( MF__, "DCOM0 DCOMi will sleep on hevSem %d",
            ThrdCtrlHeader.hevSem ));
    
    if (pThrdCtrlHeader->fSemAlreadyPosted == false)
    {
      // --- signal that there are no wake up activities
      pThrdCtrlHeader->ThrdState = KT_SLEEPING;
      
      rc = sql41c_wait_event_sem (pThrdCtrlHeader->hevSem, (ULONG) -1,
                                  "DCOM0 or DCOMi" );

      if ((rc != NO_ERROR) && (rc != ERROR_TIMEOUT))
        ABORT();
      
      pThrdCtrlHeader->ThrdState = KT_RUNNING;
    }
    pThrdCtrlHeader->fSemAlreadyPosted = false;
    
    pRequest = (*ppDCOMWorkQue)->pRequest;
  }
  
  DBG3 (( MF__, "DCOM0 or DCOMi Request: 0x%x from TID:%u, [T:0x%03u]",
          pRequest, pRequest->pTaskCtrl->pUKT->ThrdCtrlHeader.Tid,
          pRequest->pTaskCtrl->ulTaskIndex ));
  
  // --- remove element from queue
  WAIT_UNTIL_ALONE( pExclusive );
  
  (*ppDCOMWorkQue)->pRequest  = 0;
  (*ppDCOMWorkQue)            = (*ppDCOMWorkQue)->pNextElem;
  (*pDCOMQueLen)--;
  
  CLEARLOCK( pExclusive );
  
  return ( pRequest );
}

/*---------------------------------------------------------------------------*/

static tos35k_DCOMiCtrl* os35k_DCOMiInit ( tos00k_TaskCtrl  *pTaskCtrl,
                                           ULONG            ReqType,
                                           tsp00_ErrText    &Errtext ) 
{
  ROUTINE_DBG_MSP00 ("os35k_DCOMiInit");
  tos00k_UKTCtrl                        *pUKT      = pTaskCtrl->pUKT;
  tos00k_Dlq                            &Request   = pTaskCtrl->SpecDLQElem1;
  ULONG                                 Relationship;


  switch ( ReqType )
  {
  case REQ_DCOM_TRANSFER_EO00K:
    if ( pTaskCtrl->pDCOMiCtrl != NULL )
      return pTaskCtrl->pDCOMiCtrl;  // --- DCOMi already running!

    Relationship = TASK_REL_OS35K;
    break;

  default:
    if ( pUKT->pDCOMiCtrl != NULL )
      return pUKT->pDCOMiCtrl;  // --- DCOMi already running!

    if ( !eo38CoInitialize() )
    {
      DBG3 (( MF__, "eo38CoInitialize failed for UKT %d", pUKT->ulUKTIndex ));
      sql46c_build_error_string ( Errtext, ERRMSG_DCOM_COINITIALIZE_FAILED, 0 );
      return NULL;
    }

    Relationship = UKT_REL_OS35K;
    break;
  }

  //
  // --- startup DCOMi
  //
  Request.pTaskCtrl                                  = pTaskCtrl;
  Request.ulReqType                                  = REQ_DCOM_CREATE_DCOMI_EO00K;
  Request.ReqArgs.DCOMCtrlRequest.Relationship_os00k = Relationship;
  
  os35k_ReqToDCOM0( &Request );
  GOTO_DISP(pUKT);
  
  // --- wrong reply?
  if ( pUKT->pCurrReq != &Request )
  {
    MSGCD (( ERR_UNKNOWN_REQUEST ));
    DBG1  (( MF__, "[T:0x%03u] pRequest: 0x%x (SEND) != pUKT->pCurrReq: 0x%x (REPLY)",
             pTaskCtrl->ulTaskIndex, pRequest, pUKT->pCurrReq ));
    ABORT();
  }
  
  if ( Request.ReqArgs.DCOMCtrlRequest.Ok_os00k )
  { 
    if ( Relationship == TASK_REL_OS35K )
      return pTaskCtrl->pDCOMiCtrl;
    else
      return pUKT->pDCOMiCtrl;
  }
  else
  {
    if ( Relationship == UKT_REL_OS35K )
      eo38UnInitialize();

    DBG3 (( MF__, "Can't create DCOMi for UKT %d", pUKT->ulUKTIndex ));
    sql46c_build_error_string ( Errtext,
                                os35k_ErrTab[Request.ReqArgs.DCOMRequest.ErrCode_os00k], 
                                Request.ReqArgs.DCOMRequest.APIRetCode_os00k );
  }
  
  return NULL;
}

/*---------------------------------------------------------------------------*/

static tos35k_DCOMiCtrl* os35k_GetDCOMiCtrl ( tos00k_Dlq  *pRequest )
{
  ROUTINE_DBG_MSP00 ("os35k_GetDCOMiCtrl");

  LONG                             rc               = NO_ERROR;
  tos00k_DlqDCOMCtrlReqArgs        &DCOMCtrlRequest = pRequest->ReqArgs.DCOMCtrlRequest;
  tos35k_DCOMiCtrl                 *pDCOMiCtrl      = NULL;
  tos35k_DCOMiCtrl                 *pTmpDCOMiCtrl;
  ULONG                            DCOMiNo;
  
  switch ( DCOMCtrlRequest.Relationship_os00k )
  {
  case TASK_REL_OS35K :

    rc = os35k_AllocDCOMiCtrl( &pDCOMiCtrl );

    if ( pDCOMiCtrl == NULL )
    {
      DCOMCtrlRequest.ErrCode_os00k    = ERRCODE_DCOM_MAX_DCOMIS_REACHED;
      DCOMCtrlRequest.APIRetCode_os00k = rc;
      DCOMCtrlRequest.Ok_os00k         = false;
    }
    else
      sql70k_init_dcomi_ctrl ( pDCOMiCtrl );

    break;

  case UKT_REL_OS35K:

    for ( DCOMiNo  = 0; 
          DCOMiNo < kgs.ulNumOfDCOMis;
          DCOMiNo++ )
    {
      pTmpDCOMiCtrl = kgs.pFirstDCOMiCtrl + DCOMiNo;
      
      if ( pTmpDCOMiCtrl->ThrdCtrlHdr_os35k.ThrdState == KT_INITIALIZED ) 
      {
        pDCOMiCtrl = pTmpDCOMiCtrl;
        sql70k_init_dcomi_ctrl ( pDCOMiCtrl );
        break;
      }
    }

    if ( pDCOMiCtrl == NULL )
    {
      MSGCD(( ERR_MAX_DCOMIS_REACHED ));
      
      DCOMCtrlRequest.ErrCode_os00k    = ERRCODE_DCOM_MAX_DCOMIS_REACHED;
      DCOMCtrlRequest.APIRetCode_os00k = NO_ERROR;
      DCOMCtrlRequest.Ok_os00k         = false;
    }
    break;
  }
   
  return pDCOMiCtrl;
}

/*---------------------------------------------------------------------------*/

static void os35k_ReleaseDCOMiCtrl ( ULONG                     Relationship,
                                     tos35k_DCOMiCtrl          **ppDCOMiCtrl )
{
  ROUTINE_DBG_MSP00 ("os35k_ReleaseDCOMiCtrl");
  
  
  switch ( Relationship )
  {
  case TASK_REL_OS35K :
    os35k_FreeDCOMiCtrl ( *ppDCOMiCtrl );
    break;

  case UKT_REL_OS35K  :
    // --- init the dcomi control record
    sql70k_init_dcomi_ctrl ( *ppDCOMiCtrl );
    break;
  }

  *ppDCOMiCtrl = NULL; 
  
  return;
}

/*---------------------------------------------------------------------------*/


static bool os35k_DCOMiCreate ( tos00k_Dlq   *pRequest )
{
  ROUTINE_DBG_MSP00 ("os35k_DCOMiCreate");
  LONG                             rc;
  tos35k_DCOMiCtrl                 *pDCOMiCtrl;
  ULONG                            WaitTime;
  tos00k_DlqDCOMCtrlReqArgs        &DCOMCtrlRequest = pRequest->ReqArgs.DCOMCtrlRequest;
  tos00k_TaskCtrl                  *pTaskCtrl       = pRequest->pTaskCtrl;
  tos00k_UKTCtrl                   *pUKT            = pTaskCtrl->pUKT;


  pDCOMiCtrl = os35k_GetDCOMiCtrl ( pRequest );

  if ( pDCOMiCtrl == NULL )
    return ( false );
  
  // --- create DCOMi-Thread
  rc = CREATE_THREAD(&pDCOMiCtrl->ThrdCtrlHdr_os35k.Tid,
                     &pDCOMiCtrl->ThrdCtrlHdr_os35k.hThrd,
                     os35k_DCOMi,
                     pDCOMiCtrl,       // argument
                     CREATE_THRD_SUSPENDED,
                     KT_DCOMi_STACK_SIZE);
  
  if ( rc != NO_ERROR )
  {
    MSGALL((ERR_CREATING_THREAD, "DCOMi", rc));
    DCOMCtrlRequest.ErrCode_os00k    = ERRCODE_DCOM_CREATE_DCOMI;
  }
  else
  {
    pDCOMiCtrl->Relationship_os35k = DCOMCtrlRequest.Relationship_os00k;

    rc = RESUME_THREAD(pDCOMiCtrl->ThrdCtrlHdr_os35k.Tid,
                       pDCOMiCtrl->ThrdCtrlHdr_os35k.hThrd );
    
    if ( rc != NO_ERROR )
    {
      MSGALL((ERR_RESUMING_THREAD, "DCOMi", rc));
      DCOMCtrlRequest.ErrCode_os00k  = ERRCODE_DCOM_RESUME_DCOMI;
    }
  }

  if ( rc != NO_ERROR )
  {
      os35k_KillDCOMi ( &pDCOMiCtrl, DCOMCtrlRequest.Relationship_os00k );

      DCOMCtrlRequest.Ok_os00k         = false;
      DCOMCtrlRequest.APIRetCode_os00k = rc;
      return ( DCOMCtrlRequest.Ok_os00k );
  }

  // --- wait until the DCOMi is up and running
  for ( WaitTime = MAX_DCOMI_WAIT_TIME * 10;
        (pDCOMiCtrl->ThrdCtrlHdr_os35k.ThrdState == KT_INITIALIZED) && WaitTime;
        --WaitTime )
  {
    SLEEP( 1000 / 10 ); // -- sleep a moment
  }
  
  if ( pDCOMiCtrl->ThrdCtrlHdr_os35k.ThrdState != KT_INITIALIZED )
  {
    DCOMCtrlRequest.Ok_os00k = true;

    if ( DCOMCtrlRequest.Relationship_os00k == UKT_REL_OS35K )
      pUKT->pDCOMiCtrl      = pDCOMiCtrl;
    else
      pTaskCtrl->pDCOMiCtrl = pDCOMiCtrl;
  }
  else
  { 
    os35k_KillDCOMi ( &pDCOMiCtrl, DCOMCtrlRequest.Relationship_os00k );

    DCOMCtrlRequest.ErrCode_os00k    = ERRCODE_DCOM_CREATE_DCOMI;
    DCOMCtrlRequest.Ok_os00k         = false;
    DCOMCtrlRequest.APIRetCode_os00k = NO_ERROR;
  }

  return DCOMCtrlRequest.Ok_os00k;
}

/*---------------------------------------------------------------------------*/

static bool os35k_DCOMiRemove ( tos00k_Dlq    *pRequest )
{
  ROUTINE_DBG_MSP00 ("os35k_DCOMiRemove");

  TID                              Tid;
  HANDLE                           ThrdHdl;
  tos35k_DCOMiCtrl                 **ppDCOMiCtrl    = NULL;
  tos00k_DlqDCOMCtrlReqArgs        &DCOMCtrlRequest = pRequest->ReqArgs.DCOMCtrlRequest;
  ULONG                            WaitTime         = MAX_DCOMI_REMOVE_WAIT_TIME;

  if ( DCOMCtrlRequest.Relationship_os00k == UKT_REL_OS35K )
  {
    DCOMCtrlRequest.Ok_os00k         = false;
    DCOMCtrlRequest.ErrCode_os00k    = ERRCODE_DCOM_REM_UKT_DCOMi_NOT_ALLOWED;
    DCOMCtrlRequest.APIRetCode_os00k = NO_ERROR;
    return false;
  }
  else
    ppDCOMiCtrl = &pRequest->pTaskCtrl->pDCOMiCtrl;

  //
  // --- wait until the DCOMi has stop the activities
  //
  while (( (*ppDCOMiCtrl)->ThrdCtrlHdr_os35k.ThrdState != KT_SLEEPING ) ||
         ( (*ppDCOMiCtrl)->pWorkQue_os35k              != (*ppDCOMiCtrl)->pReqQue_os35k ))
  {
    DBG3 (( MF__, "ThrdCtrlHeader.ThrdState = %d",   (*ppDCOMiCtrl)->ThrdCtrlHeader.ThrdState ));
    DBG3 (( MF__, "pWorkQue                 = 0x%x", (*ppDCOMiCtrl)->pWorkQue_os35k ));
    DBG3 (( MF__, "pDCOMReqQue              = 0x%x", (*ppDCOMiCtrl)->pReqQue_os35k ));
    SLEEP( 100 );
  }

  Tid     = (*ppDCOMiCtrl)->ThrdCtrlHdr_os35k.Tid;
  ThrdHdl = (*ppDCOMiCtrl)->ThrdCtrlHdr_os35k.hThrd;

  // --- send request to DCOMi
  os35k_ReqToDCOMi( *ppDCOMiCtrl, pRequest );

  WAIT_THREAD(ThrdHdl, WaitTime);

  *ppDCOMiCtrl = NULL;
  DCOMCtrlRequest.Ok_os00k         = true;
  DCOMCtrlRequest.APIRetCode_os00k = NO_ERROR;

  return true;
}

/*---------------------------------------------------------------------------*/

static void os35k_KillDCOMi ( tos35k_DCOMiCtrl  **ppDCOMiCtrl,
                              ULONG             Relationship )
{
  ROUTINE_DBG_MSP00 ("os35k_KillDCOMi");

  KILL_THREAD ( (*ppDCOMiCtrl)->ThrdCtrlHeader.Tid,
                (*ppDCOMiCtrl)->ThrdCtrlHdr_os35k.hThrd );

  if ( (*ppDCOMiCtrl)->ThrdCtrlHdr_os35k.hevSem != (HEV) INVALID_HANDLE_VALUE )
    sql41c_close_event_sem ( (*ppDCOMiCtrl)->ThrdCtrlHdr_os35k.hevSem, "DEVi" );

  os35k_ReleaseDCOMiCtrl  ( Relationship, ppDCOMiCtrl );

  return;
}

/*---------------------------------------------------------------------------*/

static void os35k_DCOMiTerminate ( tos35k_DCOMiCtrl          *pDCOMiCtrl,
                                   tos00k_DlqDCOMCtrlReqArgs *DCOMCtrlRequest )
{
  ROUTINE_DBG_MSP00 ("os35k_DCOMiTerminate");

  LONG      rc = NO_ERROR;

  pDCOMiCtrl->ThrdCtrlHdr_os35k.ThrdState = KT_TERMINATING;
  
  sql41c_close_event_sem ( pDCOMiCtrl->ThrdCtrlHdr_os35k.hevSem, "DCOMi" );
  os35k_ReleaseDCOMiCtrl ( pDCOMiCtrl->Relationship_os35k, &pDCOMiCtrl );

  return;
}

/*---------------------------------------------------------------------------*/

static void _System os35k_DCOMi ( tos35k_DCOMiCtrl   *pDCOMiCtrl )
{
  ROUTINE_DBG_MSP00 ("os35k_DCOMi");
  
  LONG                          rc         = NO_ERROR;
  bool                          Terminate = false;
  tos00k_Dlq                    *pRequest;
  HANDLE                        ThrdHdl;
  
  #pragma warning( disable : 4509 )
  __try 
  {
    // --- save the current thread handle which is used by 'EXITTHREAD'.
    ThrdHdl = pDCOMiCtrl->ThrdCtrlHdr_os35k.hThrd;
    
    // --- set the thread priority
    rc = SET_THREAD_PRIO ( pDCOMiCtrl->ThrdCtrlHdr_os35k.lThrdPrio,
                           pDCOMiCtrl->ThrdCtrlHdr_os35k.Tid,
                           pDCOMiCtrl->ThrdCtrlHdr_os35k.hThrd );
    
    if ( rc != NO_ERROR )
    {
      MSGCD (( ERR_SETTING_THRD_PRIO, pDCOMiCtrl->ThrdCtrlHdr_os35k.Tid));
      DBG1  ((MF__, "Cannot set the thread priority, TID: %u",
              pDCOMiCtrl->ThrdCtrlHdr_os35k.Tid));
      ABORT();
    }
    
    // --- Create event semaphore (SET) for DCOMi
    rc = sql41c_create_event_sem ( &pDCOMiCtrl->ThrdCtrlHdr_os35k.hevSem,
                                   NULL, NULL, 0, false, NULL );
    
    if ( rc != NO_ERROR )
      ABORT();
    
    if ( eo38CoInitialize() )
    {
      MSGD (( INFO_KERNEL_THREAD_STARTED, "DCOMi" ));
      
      DBG3 (( MF__, "DCOMi TID:%d started", pDCOMiCtrl->ThrdCtrlHeader.Tid ));
      
      // --- set thread control info
      pDCOMiCtrl->ThrdCtrlHdr_os35k.ThrdState   = KT_RUNNING;
      pDCOMiCtrl->ThrdCtrlHdr_os35k.ulStartTime = kgs.ulCurrTime;
      
      // --- Runs until closed
      while ( Terminate == false )
      {
        pRequest = os35k_DCOMReceive (&pDCOMiCtrl->pWorkQue_os35k,
                                      &pDCOMiCtrl->QueLen_os35k,
                                      &pDCOMiCtrl->Exclusive_os35k,
                                      &pDCOMiCtrl->ThrdCtrlHdr_os35k);
        
        switch ( pRequest->ulReqType )
        {
        case REQ_DCOM_TRANSFER_EO00K :
          os39VDcomTransfer ( (teo39_TransferP*)pRequest->ReqArgs.DCOMRequest.pParam_os00k );
          os35k_ReqFromDCOMxToUKT(pRequest); // -- send reply back to UKT
          break;
          
        case REQ_DCOM_PARAMINFO_EO00K :
          os39VDcomParamInfo ( (tsp_dcom_paraminfo_parms*)
                               pRequest->ReqArgs.DCOMRequest.pParam_os00k );
          os35k_ReqFromDCOMxToUKT(pRequest); // -- send reply back to UKT
          break;
          
        case REQ_DCOM_CO_CREATE_INSTANCE_EO00K :
          eo38CoCreateInstance ( (tsp_dcom_co_create_instance_parms*)
                                   pRequest->ReqArgs.DCOMRequest.pParam_os00k );
          os35k_ReqFromDCOMxToUKT(pRequest); // -- send reply back to UKT
          break;
          
        case REQ_DCOM_RELEASE_INSTANCE_EO00K:
          eo38ReleaseInstance ( (tsp_dcom_release_instance_parms*)
                                  pRequest->ReqArgs.DCOMRequest.pParam_os00k );
          os35k_ReqFromDCOMxToUKT(pRequest); // -- send reply back to UKT
          break;
          
        case REQ_DCOM_RELASE_ALL_EO00K:
          eo38VDcomReleaseAll (NULL);
          os35k_ReqFromDCOMxToUKT(pRequest); // -- send reply back to UKT
          break;
          
        case REQ_DCOM_REMOVE_DCOMI_EO00K:
          os35k_DCOMiTerminate ( pDCOMiCtrl, &(pRequest->ReqArgs.DCOMCtrlRequest));
          Terminate = true;                // --- exit for loop
          break;
          
        case REQ_DCOM_CO_GET_CLASSOBJECT_EO00K :
          eo38CoGetClassObject ( (tsp_dcom_co_create_instance_parms*) 
                                   pRequest->ReqArgs.DCOMRequest.pParam_os00k );
          os35k_ReqFromDCOMxToUKT(pRequest); // -- send reply back to UKT
          break;
          
        default          :
          MSGCD (( ERR_ILLEGAL_REQUEST, "DCOMi", pRequest->ulReqType ));
          os35k_ReqFromDCOMxToUKT(pRequest); // -- send reply back to UKT
          break;
        }
        DBG3 (( MF__, "next in loop"));
      }
      
      eo38UnInitialize();
      
      MSGD (( INFO_KERNEL_THREAD_STOPPED, "DCOMi" ));
    }   
  }
  __except( sql50k_excepthandler(GetExceptionInformation()) ) {;}
  
  EXITTHREAD ( 0, ThrdHdl );
  
  return;
}

/*---------------------------------------------------------------------------*/

static LONG os35k_AllocDCOMiCtrl ( tos35k_DCOMiCtrl  **ppDCOMiCtrl )
{
  ROUTINE_DBG_MSP00 ("os35k_AllocDCOMiCtrl");
  LONG              rc             = NO_ERROR;
  PUCHAR            pBuff          = NULL;
  ULONG             NumOfDCOMs     = 2;
  ULONG             TotalSize;


  TotalSize = sizeof(tos35k_DCOMiCtrl) + sizeof(tos00k_DCOMQueue) * NumOfDCOMs;

  rc = ALLOC_MEM ((PPVOID)&pBuff, TotalSize );

  if ( rc != NO_ERROR )
    return rc;

  // --- init memory
  memset (pBuff, 0, TotalSize);
  *ppDCOMiCtrl = (tos35k_DCOMiCtrl*)pBuff;

  (*ppDCOMiCtrl)->pDCOMs_os35k     = (tos00k_DCOMQueue*)((*ppDCOMiCtrl) + 1);
  (*ppDCOMiCtrl)->NumOfDCOMs_os35k = NumOfDCOMs;

  return rc;
}

/*---------------------------------------------------------------------------*/

static _INLINE void os35k_FreeDCOMiCtrl ( tos35k_DCOMiCtrl *pDCOMiCtrl )
{
  ROUTINE_DBG_MSP00 ("os35k_FreeDCOMiCtrl");

  FREE_MEM ( pDCOMiCtrl );

  return;
}


/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/
