/* @lastChanged: "1999-04-21  8:17"
 * @filename:    vos38s.cpp
 * @purpose:     "OLE_Registry_Functions"
 * @release:     7.2.0.0
 * @see:         ""
 *
 * @copyright:   "Copyright by SAP AG, 1998"


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 */

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#define PPVOID  void**
#define tak_all_command_glob void 

#include "ggg00.h"
#include "IliveCacheSink.h"  /* no check */

#include "gco01.h"
#include "hco62.h"
#include "hco60.h"
#include "hco61.h"
#include "hos39.h"
//#include "hos38s.h"
#include "geo38.h"
#include "heo38.h"


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define         STATUS_SETERROR_EXCEPTION 0xC0004711L

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/


/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/


/*===========================================================================*
 *  GLOBAL VARIABLES                                                         *
 *===========================================================================*/

/*===========================================================================*
 *  LOCAL VARIABLES                                                          *
 *===========================================================================*/


/*===========================================================================*
 *  LOCAL FUNCTIONS (PROTOTYPES)                                             *
 *===========================================================================*/

static HRESULT os38_send_dbsink(IliveCacheSink *This, 
                                int *DbgpBufLength, 
                                unsigned char *pDbgBuffer);

/*===========================================================================*
 *  GLOBAL FUNCTIONS (CODE)                                                  *
 *===========================================================================*/

STDMETHODIMP CIliveCacheSink::ABAPRead( 
                      OmsTypeInt4   * TabHandle,
                      tsp00_Int4    * count,
                      tsp00_Int4    * lpbuflen,
                      unsigned char * lpv,
                      short         * lpb_error)

{
  CliveCacheDbg       *pLCD;
  teo39_DebugMsg      *dbgMsgBuf;
  char                *lpc;
  HRESULT             hr;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
    sizeof(dbgMsgBuf->dbgMsgParm.v.ABAPReadP);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_ABAPRead_eo39;
  dbgMsgBuf->dbgMsgParm.v.ABAPReadP.TabHandle = *TabHandle;
  dbgMsgBuf->dbgMsgParm.v.ABAPReadP.count     = *count;
  dbgMsgBuf->dbgMsgParm.v.ABAPReadP.buflen    = *lpbuflen;
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  if SUCCEEDED(hr) {
    *TabHandle = dbgMsgBuf->dbgMsgParm.v.ABAPRead_rplyP.TabHandle;
    *count     = dbgMsgBuf->dbgMsgParm.v.ABAPRead_rplyP.count;
    *lpb_error = dbgMsgBuf->dbgMsgParm.v.ABAPRead_rplyP.e;
    lpc        = ((char*) &dbgMsgBuf->dbgMsgParm.v.ABAPRead_rplyP) +
      sizeof (dbgMsgBuf->dbgMsgParm.v.ABAPRead_rplyP);
    memcpy (lpv, lpc, dbgMsgBuf->dbghdr.hdbgBufLen - 
      sizeof(dbgMsgBuf->dbghdr) - 
      sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) -
      sizeof(dbgMsgBuf->dbgMsgParm.v.ABAPRead_rplyP));
  }
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::ABAPWrite( 
                        OmsTypeInt4 * TabHandle,
                        tsp00_Int4  row_size,
                        tsp00_Int4  row_count,
                        unsigned char  * lpv,
                        short  * lpb_error)
{
  CliveCacheDbg       *pLCD;
  teo39_DebugMsg      *dbgMsgBuf;
  char                *lpc;
  HRESULT             hr;
  tsp00_Int4                len;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
    sizeof(dbgMsgBuf->dbgMsgParm.v.ABAPWriteP);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_ABAPWrite_eo39;
  dbgMsgBuf->dbgMsgParm.v.ABAPWriteP.TabHandle    = *TabHandle;
  dbgMsgBuf->dbgMsgParm.v.ABAPWriteP.row_size     = row_size;
  dbgMsgBuf->dbgMsgParm.v.ABAPWriteP.row_count    = row_count;
  len = row_size * row_count;
  lpc = ((char*) &dbgMsgBuf->dbgMsgParm.v.ABAPWriteP) +
          sizeof (dbgMsgBuf->dbgMsgParm.v.ABAPWriteP);
  memcpy( lpc, lpv, len );
  dbgMsgBuf->dbghdr.hdbgBufLen = dbgMsgBuf->dbghdr.hdbgBufLen + len;
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  if SUCCEEDED(hr) {
    *lpb_error = dbgMsgBuf->dbgMsgParm.v.ABAPWrite_rplyP.e;
  }
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::AllocPacket( 
                      unsigned char  ** lpv,
                      tsp00_Int4 *sz)
{
  CliveCacheDbg       *pLCD;
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_AllocPacket_eo39;
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  *sz = dbgMsgBuf->dbgMsgParm.v.AllocPacket_rplyP.sz;
  *lpv = (unsigned char*)malloc( *sz + sizeof(dbgMsgBuf->dbghdr) + sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind));
  pLCD->m_SQLOutbufPtr = (unsigned char*)malloc( *sz );
  *lpv = *lpv + sizeof(dbgMsgBuf->dbghdr) + sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::CloseVersion (
                      OmsVersionId*  lpVersionId,
                      tsp00_Int4                    lastKey,
                      short*                  lperror)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
    sizeof(dbgMsgBuf->dbgMsgParm.v.CloseVersionP);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_CloseVersion_eo39;
  memcpy(&dbgMsgBuf->dbgMsgParm.v.CloseVersionP.versionId, lpVersionId, 
    sizeof(dbgMsgBuf->dbgMsgParm.v.CloseVersionP.versionId));
  dbgMsgBuf->dbgMsgParm.v.CloseVersionP.lastKey = lastKey;
    
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  
  *lperror = dbgMsgBuf->dbgMsgParm.v.BasisErrorP;
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::CreateOpenVersion(
                      unsigned char  *lpConsistentView,
                      OmsVersionId   *lpVersionId,
                      boolean         createVersion,
                      unsigned char  *lpVersionContext,
                      short          *lperror)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
    sizeof(dbgMsgBuf->dbgMsgParm.v.CreateOpenVersionP);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_CreateOpenVersion_eo39;
  memcpy(&dbgMsgBuf->dbgMsgParm.v.CreateOpenVersionP.consistentView, lpConsistentView,
    sizeof(dbgMsgBuf->dbgMsgParm.v.CreateOpenVersionP.consistentView));
  memcpy(&dbgMsgBuf->dbgMsgParm.v.CreateOpenVersionP.versionId, lpVersionId, 
    sizeof(dbgMsgBuf->dbgMsgParm.v.CreateOpenVersionP.versionId));
  dbgMsgBuf->dbgMsgParm.v.CreateOpenVersionP.createVersion = createVersion;

  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);

  memcpy(lpVersionContext, &dbgMsgBuf->dbgMsgParm.v.CreateOpenVersion_rplyP.VersionContext, 
    sizeof(dbgMsgBuf->dbgMsgParm.v.CreateOpenVersion_rplyP.VersionContext));

  *lperror = dbgMsgBuf->dbgMsgParm.v.CreateOpenVersion_rplyP.e;

  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::SQLRequest( 
                      tsp00_Int4 buflen,
                      unsigned char  * lpv,
                      short *lperror)
{
  HRESULT             hr;
  teo39_DebugMsg      *dbgMsgBuf;
  dbgMsgBuf   = (teo39_DebugMsg*) (lpv - (sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind)));
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
    buflen;
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_SQLRequest_eo39;
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, (unsigned char*)dbgMsgBuf);
  *lperror = 0;
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::SQLReceive( 
                      unsigned char  ** lpv)
{
  CliveCacheDbg       *pLCD;
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  tsp00_Int4                sz;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_SQLReceive_eo39;
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  *lpv = (unsigned char*)&dbgMsgBuf->dbgMsgParm.v;
  sz = dbgMsgBuf->dbghdr.hdbgBufLen - ( sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind));
  memcpy(pLCD->m_SQLOutbufPtr,*lpv,sz);
  *lpv = pLCD->m_SQLOutbufPtr;
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::NewConsistentView(
                      unsigned char  *lpConsistentView,
                      short          *lpb_error) 
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
    sizeof(dbgMsgBuf->dbgMsgParm.v.EndConsistentViewP.consistentView);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_NewConsistentView_eo39;
  memcpy(&dbgMsgBuf->dbgMsgParm.v.EndConsistentViewP.consistentView, lpConsistentView,
    sizeof(dbgMsgBuf->dbgMsgParm.v.EndConsistentViewP.consistentView));
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  memcpy (lpConsistentView, &dbgMsgBuf->dbgMsgParm.v.EndConsistentViewP.consistentView,
    sizeof(dbgMsgBuf->dbgMsgParm.v.EndConsistentViewP.consistentView));
  *lpb_error = dbgMsgBuf->dbgMsgParm.v.EndConsistentView_rplyP.e;
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::EndConsistentView(
                      unsigned char  *lpConsistentView,
                      short          *lpb_error) 
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
    sizeof(dbgMsgBuf->dbgMsgParm.v.EndConsistentViewP.consistentView); /* PTS 1104790 */
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_EndConsistentView_eo39;
  memcpy(&dbgMsgBuf->dbgMsgParm.v.EndConsistentViewP.consistentView, lpConsistentView,
    sizeof(dbgMsgBuf->dbgMsgParm.v.EndConsistentViewP.consistentView));
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  memcpy (lpConsistentView, &dbgMsgBuf->dbgMsgParm.v.EndConsistentViewP.consistentView,
    sizeof(dbgMsgBuf->dbgMsgParm.v.EndConsistentViewP.consistentView));
  *lpb_error = dbgMsgBuf->dbgMsgParm.v.EndConsistentView_rplyP.e;
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::GetErrorMsg( 
                        short errorno,
                        tsp00_Int4 *lpbuflen,
                        unsigned char  * lpv)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
    sizeof(teo39_dbsink_ErrorMsg);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_GetErrorMsg_eo39;
  dbgMsgBuf->dbgMsgParm.v.ErrorMsgP.errorno = errorno;
  dbgMsgBuf->dbgMsgParm.v.ErrorMsgP.msglen  = *lpbuflen;
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  //dbgMsgBuf->dbgMsgParm.v.ErrorMsgP.msgType   *************** TODO
  *lpbuflen = dbgMsgBuf->dbgMsgParm.v.ErrorMsgP.msglen;
  memcpy(lpv, 
    (char*)(&dbgMsgBuf->dbgMsgParm.v.ErrorMsgP.msglen) + sizeof(dbgMsgBuf->dbgMsgParm.v.ErrorMsgP.msglen), 
    *lpbuflen);
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::MultiPurpose (
                        unsigned char MessType,
                        unsigned char MessType2,
                        tsp00_Int4           *lpbuflen,
                        unsigned char  *lpv,
                        short          *lpb_error)
{
  CliveCacheDbg       *pLCD;
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
    sizeof(dbgMsgBuf->dbgMsgParm.v.MultiPurposeP);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_MultiPurpose_eo39;
  dbgMsgBuf->dbgMsgParm.v.MultiPurposeP.MessType  = MessType;
  dbgMsgBuf->dbgMsgParm.v.MultiPurposeP.MessType2 = MessType2;
  dbgMsgBuf->dbgMsgParm.v.MultiPurposeP.buflen    = *lpbuflen;
  memcpy( (char*)dbgMsgBuf + dbgMsgBuf->dbghdr.hdbgBufLen,
          lpv,
          *lpbuflen );
  dbgMsgBuf->dbghdr.hdbgBufLen = dbgMsgBuf->dbghdr.hdbgBufLen + *lpbuflen;
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  if SUCCEEDED(hr) {
    *lpbuflen  = dbgMsgBuf->dbgMsgParm.v.MultiPurpose_rplyP.buflen;
    *lpb_error = dbgMsgBuf->dbgMsgParm.v.MultiPurpose_rplyP.b_error;
    memcpy(lpv, 
      (char*)dbgMsgBuf + 
      sizeof(dbgMsgBuf->dbghdr) + 
      sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
      sizeof(dbgMsgBuf->dbgMsgParm.v.MultiPurpose_rplyP),
      *lpbuflen );
  }
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::SetError( 
                      short error_no,
                      tsp00_Int4 buflen,
                      unsigned char  * lpv,
                      tsp00_CodeType codeType)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
    sizeof(teo39_dbsink_ErrorMsg) + 
    buflen;
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_SetError_eo39;
  dbgMsgBuf->dbgMsgParm.v.ErrorMsgP.errorno = error_no;
  dbgMsgBuf->dbgMsgParm.v.ErrorMsgP.msgType = codeType;
  dbgMsgBuf->dbgMsgParm.v.ErrorMsgP.msglen  = buflen;
  memcpy((char *)(&dbgMsgBuf->dbgMsgParm.v.ErrorMsgP.msglen) + sizeof(dbgMsgBuf->dbgMsgParm.v.ErrorMsgP.msglen), 
    lpv, buflen);
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  
  RaiseException (STATUS_SETERROR_EXCEPTION,
    0,            /* continuable exception */
    0,            /* number of parms       */
    NULL);        /* &parms                */
  return -1;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::BasisError( 
                      short basis_error)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
    sizeof(dbgMsgBuf->dbgMsgParm.v.BasisErrorP);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_BasisError_eo39;
  dbgMsgBuf->dbgMsgParm.v.BasisErrorP = basis_error;
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::HResultError( 
                      HRESULT hresult_error)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
    sizeof(dbgMsgBuf->dbgMsgParm.v.HResultErrorP);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_HResultError_eo39;
  dbgMsgBuf->dbgMsgParm.v.HResultErrorP = hresult_error;
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::Vtrace( 
                      tsp00_Int4 buflen,
                      unsigned char  * lpv)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
    sizeof(dbgMsgBuf->dbgMsgParm.v.BuflenP) + 
    buflen;
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_Vtrace_eo39;
  dbgMsgBuf->dbgMsgParm.v.BuflenP  = buflen;
  memcpy((char *)(&dbgMsgBuf->dbgMsgParm.v.BuflenP) + sizeof(dbgMsgBuf->dbgMsgParm.v.BuflenP), 
    lpv, buflen);
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::ConsoleMsg( 
                      tsp00_Int4 buflen,
                      unsigned char  * lpv)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
    sizeof(dbgMsgBuf->dbgMsgParm.v.BuflenP) + 
    buflen;
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_ConsoleMsg_eo39;
  dbgMsgBuf->dbgMsgParm.v.BuflenP  = buflen;
  memcpy((char *)(&dbgMsgBuf->dbgMsgParm.v.BuflenP) + sizeof(dbgMsgBuf->dbgMsgParm.v.BuflenP), 
    lpv, buflen);
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::EnterCriticalSection(short regionId)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
    sizeof(dbgMsgBuf->dbgMsgParm.v.CriticalSectionP);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_EnterCriticalSection_eo39;
  dbgMsgBuf->dbgMsgParm.v.CriticalSectionP.RegionId = regionId;
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::LeaveCriticalSection(short regionId)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
    sizeof(dbgMsgBuf->dbgMsgParm.v.CriticalSectionP);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  dbgMsgBuf->dbgMsgParm.v.CriticalSectionP.RegionId = regionId;  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_LeaveCriticalSection_eo39;
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::IsLocked (
                      unsigned char  * lpContainerId,
                      OmsTypeOid     * lpoid,
                      short          * lpb_error)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
    sizeof(dbgMsgBuf->dbgMsgParm.v.LockObjP);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_IsLocked_eo39;
  memcpy(&dbgMsgBuf->dbgMsgParm.v.LockObjP.containerId, lpContainerId, sizeof(dbgMsgBuf->dbgMsgParm.v.LockObjP.containerId));
  memcpy(&dbgMsgBuf->dbgMsgParm.v.LockObjP.oid, lpoid, sizeof(dbgMsgBuf->dbgMsgParm.v.LockObjP.oid));
  
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  
  *lpb_error = dbgMsgBuf->dbgMsgParm.v.BasisErrorP;
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::LockObj( 
                      unsigned char  * lpConsistentView,
                      unsigned char  * lpContainerId,
                      OmsTypeOid     * lpoid,
                      unsigned char  * lpobjseq,
                      short          * lpb_error)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
    sizeof(dbgMsgBuf->dbgMsgParm.v.LockObjP);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_LockObj_eo39;
  memcpy(&dbgMsgBuf->dbgMsgParm.v.LockObjP.consistentView, lpConsistentView,
    sizeof(dbgMsgBuf->dbgMsgParm.v.LockObjP.consistentView));
  memcpy(&dbgMsgBuf->dbgMsgParm.v.LockObjP.containerId, lpContainerId, sizeof(dbgMsgBuf->dbgMsgParm.v.LockObjP.containerId));
  memcpy(&dbgMsgBuf->dbgMsgParm.v.LockObjP.oid, lpoid, sizeof(dbgMsgBuf->dbgMsgParm.v.LockObjP.oid));
  memcpy(&dbgMsgBuf->dbgMsgParm.v.LockObjP.WantedSeq, lpobjseq, sizeof(dbgMsgBuf->dbgMsgParm.v.LockObjP.WantedSeq));
  
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  
  *lpb_error = dbgMsgBuf->dbgMsgParm.v.BasisErrorP;
  return hr;

}
/*---------------------------------------------------------------------------*/
        

STDMETHODIMP CIliveCacheSink::UnlockObj( 
                      unsigned char  * lpContainerId,
                      OmsTypeOid     * lpoid,
                      short          * lpb_error)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
    sizeof(dbgMsgBuf->dbgMsgParm.v.UnlockObjP);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_UnlockObj_eo39;
  memcpy(&dbgMsgBuf->dbgMsgParm.v.LockObjP.containerId, lpContainerId, sizeof(dbgMsgBuf->dbgMsgParm.v.UnlockObjP.containerId));
  memcpy(&dbgMsgBuf->dbgMsgParm.v.LockObjP.oid, lpoid, sizeof(dbgMsgBuf->dbgMsgParm.v.UnlockObjP.oid));
    
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  
  *lpb_error = dbgMsgBuf->dbgMsgParm.v.BasisErrorP;
  return hr;

}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::Rollback( 
                      tsp00_Int4 buflen,
                      unsigned char  * lpv,
                      short  * lpb_error)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
    sizeof(dbgMsgBuf->dbgMsgParm.v.BuflenP) + 
    buflen;
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_Rollback_eo39;
  dbgMsgBuf->dbgMsgParm.v.BuflenP  = buflen;
  memcpy((char *)(&dbgMsgBuf->dbgMsgParm.v.BuflenP) + sizeof(dbgMsgBuf->dbgMsgParm.v.BuflenP), 
    lpv, buflen);
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  *lpb_error = dbgMsgBuf->dbgMsgParm.v.BasisErrorP;
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::GetObj (tgg91_TransNo&       consistentView,
                            tgg00_ContainerId&             containerId, 
                            const tgg00_OmsVersionContext* pVersionContext,
                            const OmsTypeOid&              oid,
                            bool                           doLock,
                            tgg91_PageRef&                 objSeq,
                            tsp00_Int4                     bodySize,
                            unsigned char*                 pBody,
                            tsp00_Int4&                    bodyLen,
                            tsp00_Int4&                    histLogCount,
                            tgg00_BasisError&              e)

{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
    sizeof(dbgMsgBuf->dbgMsgParm.v.GetObjP);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_GetObj_eo39;
  memcpy(&dbgMsgBuf->dbgMsgParm.v.GetObjP.consistentView, &consistentView,
    sizeof(dbgMsgBuf->dbgMsgParm.v.GetObjP.consistentView));
  memcpy(&dbgMsgBuf->dbgMsgParm.v.GetObjP.containerId, &containerId, sizeof(dbgMsgBuf->dbgMsgParm.v.GetObjP.containerId));
  memcpy(&dbgMsgBuf->dbgMsgParm.v.GetObjP.oid, &oid, sizeof(dbgMsgBuf->dbgMsgParm.v.GetObjP.oid));
  dbgMsgBuf->dbgMsgParm.v.GetObjP.do_lock  = doLock;
  dbgMsgBuf->dbgMsgParm.v.GetObjP.rec_size = bodySize;
  dbgMsgBuf->dbgMsgParm.v.GetObjP.inVersion = (pVersionContext != NULL);
  if (dbgMsgBuf->dbgMsgParm.v.GetObjP.inVersion)
  {
    memcpy(&dbgMsgBuf->dbgMsgParm.v.GetObjP.versionContext, pVersionContext, 
      sizeof(dbgMsgBuf->dbgMsgParm.v.GetObjP.versionContext));
  }
  
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
    
  memcpy(&consistentView, &dbgMsgBuf->dbgMsgParm.v.GetObj_rplyP.consistentView, 
    sizeof(dbgMsgBuf->dbgMsgParm.v.GetObj_rplyP.consistentView));
  memcpy(&containerId, &dbgMsgBuf->dbgMsgParm.v.GetObj_rplyP.containerId, 
    sizeof(dbgMsgBuf->dbgMsgParm.v.GetObj_rplyP.containerId));
  memcpy(&objSeq, &dbgMsgBuf->dbgMsgParm.v.GetObj_rplyP.ResultSeq, 
    sizeof(dbgMsgBuf->dbgMsgParm.v.GetObj_rplyP.ResultSeq));
  memcpy(pBody, (char*)(&dbgMsgBuf->dbgMsgParm.v.GetObj_rplyP) +
    sizeof(dbgMsgBuf->dbgMsgParm.v.GetObj_rplyP), bodySize );
  bodyLen      = dbgMsgBuf->dbgMsgParm.v.GetObj_rplyP.ObjLen; /* +++ Change Request 1000203 */
  histLogCount = 0;
  e            = dbgMsgBuf->dbgMsgParm.v.GetObj_rplyP.e;
  return hr;
};

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::GetObjWithKey ( 
                      unsigned char  * lpConsistentView,
                      unsigned char  * lptree,
                      unsigned char  * lpVersionContext,
                      tsp00_Int4                      keylen,
                      unsigned char  * lpkey,
                      boolean                   getConsistent,
                      tsp00_Int4                      bodysize,
                      unsigned char  * lpbody,
                      OmsTypeOid     * lpoid,
                      unsigned char  * lpobjseq,
                      int            * lphist_log_count,
                      short          * lpb_error)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
    sizeof(dbgMsgBuf->dbgMsgParm.v.GetObjWithKeyP) + keylen;
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_GetObjWithKey_eo39;
  memcpy(&dbgMsgBuf->dbgMsgParm.v.GetObjWithKeyP.consistentView, 
    lpConsistentView, sizeof(dbgMsgBuf->dbgMsgParm.v.GetObjWithKeyP.consistentView));
  memcpy(&dbgMsgBuf->dbgMsgParm.v.GetObjWithKeyP.containerId, 
    lptree, sizeof(dbgMsgBuf->dbgMsgParm.v.GetObjWithKeyP.containerId));
  dbgMsgBuf->dbgMsgParm.v.GetObjWithKeyP.keylen        = keylen;
  dbgMsgBuf->dbgMsgParm.v.GetObjWithKeyP.getConsistent = getConsistent;
  dbgMsgBuf->dbgMsgParm.v.GetObjWithKeyP.bodysize      = bodysize;

  memcpy ((char*)(&dbgMsgBuf->dbgMsgParm.v.GetObjWithKeyP) +
    sizeof(dbgMsgBuf->dbgMsgParm.v.GetObjWithKeyP), lpkey, keylen);

  dbgMsgBuf->dbgMsgParm.v.GetObjWithKeyP.inVersion = (lpVersionContext != NULL);
  if (dbgMsgBuf->dbgMsgParm.v.GetObjWithKeyP.inVersion)
  {
    memcpy(&dbgMsgBuf->dbgMsgParm.v.GetObjWithKeyP.versionContext, lpVersionContext,
      sizeof(dbgMsgBuf->dbgMsgParm.v.GetObjWithKeyP.versionContext));
  }

  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  
  memcpy(lptree, &dbgMsgBuf->dbgMsgParm.v.GetObjWithKey_rplyP.containerId,
    sizeof(dbgMsgBuf->dbgMsgParm.v.GetObjWithKey_rplyP.containerId));
  memcpy(lpoid, &dbgMsgBuf->dbgMsgParm.v.GetObjWithKey_rplyP.oid,
    sizeof(dbgMsgBuf->dbgMsgParm.v.GetObjWithKey_rplyP.oid));
  memcpy(lpobjseq, dbgMsgBuf->dbgMsgParm.v.GetObjWithKey_rplyP.ResultSeq, 
    sizeof(dbgMsgBuf->dbgMsgParm.v.GetObjWithKey_rplyP.ResultSeq));
  memcpy(lpConsistentView, &dbgMsgBuf->dbgMsgParm.v.GetObjWithKey_rplyP.consistentView, 
    sizeof(dbgMsgBuf->dbgMsgParm.v.GetObjWithKey_rplyP.consistentView));
  memcpy(lpbody, (char*)(&dbgMsgBuf->dbgMsgParm.v.GetObjWithKey_rplyP) +
    sizeof(dbgMsgBuf->dbgMsgParm.v.GetObjWithKey_rplyP), bodysize );
  *lpb_error = dbgMsgBuf->dbgMsgParm.v.GetObjWithKey_rplyP.e;
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::UpdateObj(
                      unsigned char  * lpConsistentView,
                      unsigned char  * lpContainerId,
                      unsigned char  * lpVersionContext,
                      unsigned char  * lpoid,
                      unsigned char  * lpobjseq,
                      tsp00_Int4 bodysize,
                      unsigned char  * lpbody,
                      unsigned char  * lpcoldesc,
                      short  * b_error)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  char                *buf_ptr;
  tsp00_Int4                buf_entry_size;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
    sizeof(dbgMsgBuf->dbgMsgParm.v.UpdateObjP) +
    bodysize +
    ((tgg00_ObjColDesc*)lpcoldesc)->colHashCount_gg00 * sizeof(tgg00_StackEntry) +
    ((tgg00_ObjColDesc*)lpcoldesc)->colInvCount_gg00 * sizeof(tgg00_StackEntry);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_UpdateObj_eo39;
  memcpy(&dbgMsgBuf->dbgMsgParm.v.UpdateObjP.consistentView, lpConsistentView, 
    sizeof(dbgMsgBuf->dbgMsgParm.v.UpdateObjP.consistentView));
  memcpy(&dbgMsgBuf->dbgMsgParm.v.UpdateObjP.containerId, lpContainerId, sizeof(dbgMsgBuf->dbgMsgParm.v.UpdateObjP.containerId));
  memcpy(&dbgMsgBuf->dbgMsgParm.v.UpdateObjP.oid, lpoid, sizeof(dbgMsgBuf->dbgMsgParm.v.UpdateObjP.oid));
  memcpy(&dbgMsgBuf->dbgMsgParm.v.UpdateObjP.WantedSeq, lpobjseq, sizeof(dbgMsgBuf->dbgMsgParm.v.UpdateObjP.WantedSeq));
  dbgMsgBuf->dbgMsgParm.v.UpdateObjP.bodysize = bodysize;
  dbgMsgBuf->dbgMsgParm.v.UpdateObjP.colHashCount_gg00 = ((tgg00_ObjColDesc*)lpcoldesc)->colHashCount_gg00;
  dbgMsgBuf->dbgMsgParm.v.UpdateObjP.colInvCount_gg00 = ((tgg00_ObjColDesc*)lpcoldesc)->colInvCount_gg00;
  buf_ptr = (char*)(&dbgMsgBuf->dbgMsgParm.v.UpdateObjP) + sizeof(dbgMsgBuf->dbgMsgParm.v.UpdateObjP);
  memcpy(buf_ptr, lpbody, bodysize );
  buf_ptr = buf_ptr + bodysize;
  buf_entry_size = dbgMsgBuf->dbgMsgParm.v.UpdateObjP.colHashCount_gg00 * sizeof(tgg00_StackEntry);
  memcpy(buf_ptr, ((tgg00_ObjColDesc*)lpcoldesc)->colHashList_gg00, buf_entry_size );
  buf_ptr = buf_ptr + buf_entry_size;
  buf_entry_size = dbgMsgBuf->dbgMsgParm.v.UpdateObjP.colInvCount_gg00 * sizeof(tgg00_StackEntry);
  memcpy(buf_ptr, ((tgg00_ObjColDesc*)lpcoldesc)->colInvList_gg00, buf_entry_size );
  dbgMsgBuf->dbgMsgParm.v.UpdateObjP.inVersion = (lpVersionContext != NULL);
  if (dbgMsgBuf->dbgMsgParm.v.UpdateObjP.inVersion)
  {
    memcpy(&dbgMsgBuf->dbgMsgParm.v.UpdateObjP.versionContext, lpVersionContext,
      sizeof(dbgMsgBuf->dbgMsgParm.v.UpdateObjP.versionContext));
  }
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  
  memcpy(lpobjseq, &dbgMsgBuf->dbgMsgParm.v.GetObj_rplyP.ResultSeq, 
    sizeof(dbgMsgBuf->dbgMsgParm.v.GetObj_rplyP.ResultSeq));
  *b_error = dbgMsgBuf->dbgMsgParm.v.GetObj_rplyP.e;
  return hr;
};

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::NewObj(
                      unsigned char   * lpConsistentView,
                      unsigned char   * lpContainerId,
                      unsigned char   * lpVersionContext,
                      tsp00_Int4                       KeyPos,
                      tsp00_Int4                       KeyColSize,
                      unsigned char   * KeyColumn, 
                      OmsTypeOid      * lpoid,
                      unsigned char   * lpobjseq,
                      short           * b_error)
{
  teo39_DebugMsg      *dbgMsgBuf;
  char                *lpchar;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
    sizeof(dbgMsgBuf->dbgMsgParm.v.NewObjP);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_NewObj_eo39;
  memcpy(&dbgMsgBuf->dbgMsgParm.v.NewObjP.containerId, lpContainerId, sizeof(dbgMsgBuf->dbgMsgParm.v.NewObjP.containerId));
  memcpy(&dbgMsgBuf->dbgMsgParm.v.NewObjP.consistentView, lpConsistentView,
    sizeof(dbgMsgBuf->dbgMsgParm.v.NewObjP.consistentView));
  dbgMsgBuf->dbgMsgParm.v.NewObjP.KeyPos  = KeyPos;
  dbgMsgBuf->dbgMsgParm.v.NewObjP.KeySize = KeyColSize;
  if (KeyColSize > 0) {
    lpchar = ((char*) &dbgMsgBuf->dbgMsgParm.v.NewObjP + sizeof(dbgMsgBuf->dbgMsgParm.v.NewObjP));
    memcpy (lpchar, KeyColumn, KeyColSize);
    dbgMsgBuf->dbghdr.hdbgBufLen += KeyColSize; 
  }
  dbgMsgBuf->dbgMsgParm.v.NewObjP.inVersion = (lpVersionContext != NULL);
  if (dbgMsgBuf->dbgMsgParm.v.NewObjP.inVersion)
  {
    memcpy(&dbgMsgBuf->dbgMsgParm.v.NewObjP.versionContext, lpVersionContext,
      sizeof(dbgMsgBuf->dbgMsgParm.v.NewObjP.versionContext));
  }
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  memcpy(lpoid, &dbgMsgBuf->dbgMsgParm.v.NewObj_rplyP.oid, sizeof(dbgMsgBuf->dbgMsgParm.v.NewObj_rplyP.oid));
  memcpy(lpobjseq, &dbgMsgBuf->dbgMsgParm.v.NewObj_rplyP.ResultSeq, sizeof(dbgMsgBuf->dbgMsgParm.v.NewObj_rplyP.ResultSeq));
  *b_error = dbgMsgBuf->dbgMsgParm.v.NewObj_rplyP.e;
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::RegisterClass(
                      tsp00_Int4          SchemaId,  
                      GUID                *lpGUID,
                      tsp00_KnlIdentifier *lpClassName,
                      tsp00_Int4           size,
                      tsp00_Uint4          ContainerNo,
                      unsigned char       *lpcoldesc,
                      unsigned char       *lpContainerId,
                      short               *b_error)
{
  teo39_DebugMsg      *dbgMsgBuf;
  tsp00_Int4                invStackLen;
  char*               buf_ptr;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  tgg00_ObjColDesc    *pInvColDesc;
  
  pInvColDesc = (tgg00_ObjColDesc*) lpcoldesc;
  invStackLen = pInvColDesc->colHashCount_gg00 * sizeof (tgg00_StackEntry);
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
    sizeof(dbgMsgBuf->dbgMsgParm.v.RegisterClassP) +
    invStackLen;
  
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_RegisterClass_eo39;
  memcpy(&dbgMsgBuf->dbgMsgParm.v.RegisterClassP.guid, lpGUID, 
    sizeof(dbgMsgBuf->dbgMsgParm.v.RegisterClassP.guid));
  memcpy(&dbgMsgBuf->dbgMsgParm.v.RegisterClassP.className, lpClassName,
    sizeof (dbgMsgBuf->dbgMsgParm.v.RegisterClassP.className));
  dbgMsgBuf->dbgMsgParm.v.RegisterClassP.size     = size;
  dbgMsgBuf->dbgMsgParm.v.RegisterClassP.schemaId = SchemaId;
  dbgMsgBuf->dbgMsgParm.v.RegisterClassP.container_no = ContainerNo;
  dbgMsgBuf->dbgMsgParm.v.RegisterClassP.invcoldesc   = *pInvColDesc;
  if (invStackLen > 0) {
    buf_ptr = (char*)(&dbgMsgBuf->dbgMsgParm.v.RegisterClass_rplyP) + sizeof(dbgMsgBuf->dbgMsgParm.v.RegisterClass_rplyP);
    memcpy (buf_ptr, (char*) pInvColDesc->colHashList_gg00, invStackLen);
  }
  
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  
  memcpy(lpContainerId, &dbgMsgBuf->dbgMsgParm.v.RegisterClass_rplyP.containerId, 
    sizeof(dbgMsgBuf->dbgMsgParm.v.RegisterClass_rplyP.containerId));
  if (invStackLen > 0)
    memcpy((char*) pInvColDesc->colHashList_gg00, buf_ptr, invStackLen);
  *b_error   = dbgMsgBuf->dbgMsgParm.v.RegisterClass_rplyP.e;
  return hr;
};

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::NextObj(
                      unsigned char  * lpConsistentView,
                      unsigned char  * lpContainerId,
                      unsigned char  * lpVersionContext,
                      OmsTypeOid     * lpoid,
                      unsigned char  * lpobjseq,
                      tsp00_Int4                      bodysize,
                      unsigned char  * lpbody,
                      short          * b_error)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
    sizeof(dbgMsgBuf->dbgMsgParm.v.NextObjP);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_NextObj_eo39;
  memcpy(&dbgMsgBuf->dbgMsgParm.v.NextObjP.consistentView, lpConsistentView, 
    sizeof(dbgMsgBuf->dbgMsgParm.v.NextObjP.consistentView));
  memcpy(&dbgMsgBuf->dbgMsgParm.v.NextObjP.containerId, lpContainerId, sizeof(dbgMsgBuf->dbgMsgParm.v.NextObjP.containerId));
  memcpy(&dbgMsgBuf->dbgMsgParm.v.NextObjP.oid, lpoid, sizeof(dbgMsgBuf->dbgMsgParm.v.NextObjP.oid));
  dbgMsgBuf->dbgMsgParm.v.NextObjP.do_lock     = 0;
  dbgMsgBuf->dbgMsgParm.v.NextObjP.with_object = (lpbody != NULL);
  dbgMsgBuf->dbgMsgParm.v.NextObjP.rec_size    = bodysize;
  dbgMsgBuf->dbgMsgParm.v.NextObjP.inVersion   = (lpVersionContext != NULL);
  if (dbgMsgBuf->dbgMsgParm.v.NextObjP.inVersion)
  {
    memcpy(&dbgMsgBuf->dbgMsgParm.v.NextObjP.versionContext, lpVersionContext,
      sizeof(dbgMsgBuf->dbgMsgParm.v.NextObjP.versionContext));
  }
  
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  
  memcpy(lpoid, &dbgMsgBuf->dbgMsgParm.v.NextObj_rplyP.oid, sizeof(dbgMsgBuf->dbgMsgParm.v.NextObj_rplyP.oid));  
  memcpy(lpobjseq, dbgMsgBuf->dbgMsgParm.v.NextObj_rplyP.ResultSeq, 
    sizeof(dbgMsgBuf->dbgMsgParm.v.NextObj_rplyP.ResultSeq));
  memcpy(lpConsistentView, &dbgMsgBuf->dbgMsgParm.v.NextObj_rplyP.consistentView, 
    sizeof(dbgMsgBuf->dbgMsgParm.v.NextObj_rplyP.consistentView));
  if (lpbody) {
    memcpy(lpbody, (char*)(&dbgMsgBuf->dbgMsgParm.v.NextObj_rplyP) +
      sizeof(dbgMsgBuf->dbgMsgParm.v.NextObj_rplyP), bodysize );
  }
  *b_error = dbgMsgBuf->dbgMsgParm.v.NextObj_rplyP.e;
  return hr;
};

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::DeleteAllObj( 
                      unsigned char  * lpContainerId,
                      unsigned char  * lpVersionContext,
                      int            * lpCntDeleted,                    
                      short          * b_error)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
   
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
    sizeof(dbgMsgBuf->dbgMsgParm.v.DeleteAllObjP);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = 1;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_DeleteAllObj_eo39;
  memcpy(&dbgMsgBuf->dbgMsgParm.v.DeleteAllObjP.containerId, lpContainerId, sizeof(dbgMsgBuf->dbgMsgParm.v.DeleteAllObjP.containerId));
  dbgMsgBuf->dbgMsgParm.v.DeleteAllObjP.inVersion = (lpVersionContext != NULL);
  if (dbgMsgBuf->dbgMsgParm.v.DeleteAllObjP.inVersion)
  {
    memcpy(&dbgMsgBuf->dbgMsgParm.v.DeleteAllObjP.VersionContext, lpVersionContext,
      sizeof(dbgMsgBuf->dbgMsgParm.v.DeleteAllObjP.VersionContext));
  }

  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  
  *lpCntDeleted = dbgMsgBuf->dbgMsgParm.v.DeleteAllObj_rplyP.cntDeleted;
  *b_error      = dbgMsgBuf->dbgMsgParm.v.DeleteAllObj_rplyP.e;
  return hr;
};

/*---------------------------------------------------------------------------*/



STDMETHODIMP CIliveCacheSink::DeleteObj(
                      unsigned char  * lpConsistentView,
                      unsigned char  * lpContainerId,
                      unsigned char  * lpVersionContext,
                      unsigned char  * lpoid,
                      unsigned char  * lpobjseq,
                      unsigned char  * lpcoldesc,
                      short  * b_error)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  char                *buf_ptr;
  tsp00_Int4          buf_entry_size;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
    sizeof(dbgMsgBuf->dbgMsgParm.v.UpdateObjP) +
    ((tgg00_ObjColDesc*)lpcoldesc)->colHashCount_gg00 * sizeof(tgg00_StackEntry) +
    ((tgg00_ObjColDesc*)lpcoldesc)->colInvCount_gg00 * sizeof(tgg00_StackEntry);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_DeleteObj_eo39;
  memcpy(&dbgMsgBuf->dbgMsgParm.v.UpdateObjP.consistentView, lpConsistentView,
    sizeof(dbgMsgBuf->dbgMsgParm.v.UpdateObjP.consistentView));
  memcpy(&dbgMsgBuf->dbgMsgParm.v.UpdateObjP.containerId, lpContainerId, sizeof(dbgMsgBuf->dbgMsgParm.v.UpdateObjP.containerId));
  memcpy(&dbgMsgBuf->dbgMsgParm.v.UpdateObjP.oid, lpoid, sizeof(dbgMsgBuf->dbgMsgParm.v.UpdateObjP.oid));
  memcpy(&dbgMsgBuf->dbgMsgParm.v.UpdateObjP.WantedSeq, lpobjseq, sizeof(dbgMsgBuf->dbgMsgParm.v.UpdateObjP.WantedSeq));
  dbgMsgBuf->dbgMsgParm.v.UpdateObjP.bodysize = 0; // not used
  dbgMsgBuf->dbgMsgParm.v.UpdateObjP.colHashCount_gg00 = ((tgg00_ObjColDesc*)lpcoldesc)->colHashCount_gg00;
  dbgMsgBuf->dbgMsgParm.v.UpdateObjP.colInvCount_gg00 = ((tgg00_ObjColDesc*)lpcoldesc)->colInvCount_gg00;
  buf_ptr = (char*)(&dbgMsgBuf->dbgMsgParm.v.UpdateObjP) + sizeof(dbgMsgBuf->dbgMsgParm.v.UpdateObjP);
  buf_entry_size = dbgMsgBuf->dbgMsgParm.v.UpdateObjP.colHashCount_gg00 * sizeof(tgg00_StackEntry);
  memcpy(buf_ptr, ((tgg00_ObjColDesc*)lpcoldesc)->colHashList_gg00, buf_entry_size );
  buf_ptr = buf_ptr + buf_entry_size;
  buf_entry_size = dbgMsgBuf->dbgMsgParm.v.UpdateObjP.colInvCount_gg00 * sizeof(tgg00_StackEntry);
  memcpy(buf_ptr, ((tgg00_ObjColDesc*)lpcoldesc)->colInvList_gg00, buf_entry_size );
  dbgMsgBuf->dbgMsgParm.v.UpdateObjP.inVersion   = (lpVersionContext != NULL);
  if (dbgMsgBuf->dbgMsgParm.v.UpdateObjP.inVersion)
  {
    memcpy(&dbgMsgBuf->dbgMsgParm.v.UpdateObjP.versionContext, lpVersionContext,
      sizeof(dbgMsgBuf->dbgMsgParm.v.UpdateObjP.versionContext));
  }

  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  
  *b_error = dbgMsgBuf->dbgMsgParm.v.BasisErrorP;
  return hr;
};

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::SubtransCommit( 
                      short  * lpb_error)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_SubtransCommit_eo39;
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  *lpb_error = dbgMsgBuf->dbgMsgParm.v.BasisErrorP;
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::GetDefaultContext(void** p)
{
  *p = lpDefaultContext;
  return S_OK;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::SetDefaultContext(void* p)
{
  lpDefaultContext = p;
  return S_OK;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::SubtransRollback( 
                      short  * lpb_error)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount  = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_SubtransRollback_eo39;
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  *lpb_error = dbgMsgBuf->dbgMsgParm.v.BasisErrorP;
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::SubtransStart( 
                      short  * lpb_error)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount  = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_SubtransStart_eo39;
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  *lpb_error = dbgMsgBuf->dbgMsgParm.v.BasisErrorP;
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::DropFile(char      *lpFileName,
                                       short     &error)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  int                 FileNameLen;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount  = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_DropFile_eo39;
  FileNameLen = strlen(lpFileName);
  memcpy(&dbgMsgBuf->dbgMsgParm.v.MessageFileP.msg[0], lpFileName, FileNameLen);
  dbgMsgBuf->dbghdr.hdbgBufLen += sizeof(dbgMsgBuf->dbgMsgParm.v.MessageFileP) - 
    sizeof(dbgMsgBuf->dbgMsgParm.v.MessageFileP.msg) + FileNameLen;
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  error = dbgMsgBuf->dbgMsgParm.v.MessageFileP.e;
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::ExistsFile(char      *lpFileName,
                                       bool&       doesExist,
                                       bool&       isDirectory,
                                       short&      error)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  int                 FileNameLen;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount  = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_ExistsFile_eo39;
  FileNameLen = strlen(lpFileName);
  memcpy(&dbgMsgBuf->dbgMsgParm.v.ExistsFileP.msg[0], lpFileName, FileNameLen);
  dbgMsgBuf->dbghdr.hdbgBufLen += sizeof(dbgMsgBuf->dbgMsgParm.v.ExistsFileP) - 
    sizeof(dbgMsgBuf->dbgMsgParm.v.ExistsFileP.msg) + FileNameLen;
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  doesExist   = dbgMsgBuf->dbgMsgParm.v.ExistsFileP.doesExist;
  isDirectory = dbgMsgBuf->dbgMsgParm.v.ExistsFileP.isDirectory;
  error       = dbgMsgBuf->dbgMsgParm.v.MessageFileP.e;
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::GetWorkingDirectory(tsp00_Pathc& path,
          short  &error) 
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount  = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_GetWorkingDir_eo39;
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  path        = dbgMsgBuf->dbgMsgParm.v.Path_rplyP.path;
  error       = dbgMsgBuf->dbgMsgParm.v.Path_rplyP.e;
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::GetDbrootSapPath(tsp00_Pathc& path,
          short  &error) 
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount  = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_GetDbrootSapDir_eo39;
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  path        = dbgMsgBuf->dbgMsgParm.v.Path_rplyP.path;
  error       = dbgMsgBuf->dbgMsgParm.v.Path_rplyP.e;
  return hr;
}

/*---------------------------------------------------------------------------*/


STDMETHODIMP CIliveCacheSink::OpenFile(char        *lpFileName,
                                       tsp00_Uint4 FileSize,
                                       boolean     doAppend,
                                       tsp00_Int4  *FileHandle,
                                       short       *lpb_error)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  int                 FileNameLen;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount  = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_OpenFile_eo39;
  dbgMsgBuf->dbgMsgParm.v.MessageFileP.i4       = FileSize;
  dbgMsgBuf->dbgMsgParm.v.MessageFileP.doAppend = doAppend;
  FileNameLen = strlen(lpFileName);
  memcpy(&dbgMsgBuf->dbgMsgParm.v.MessageFileP.msg[0], lpFileName, FileNameLen);
  dbgMsgBuf->dbghdr.hdbgBufLen += sizeof(dbgMsgBuf->dbgMsgParm.v.MessageFileP.i4) + FileNameLen;
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  *FileHandle = dbgMsgBuf->dbgMsgParm.v.MessageFileP.i4;
  *lpb_error  = dbgMsgBuf->dbgMsgParm.v.MessageFileP.e;
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::WriteFile(tsp00_Int4 FileHandle,
          char   *msg,
          short  *lpb_error)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  int                 MsgLen;
  
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
  dbgMsgBuf->dbghdr.hdbgMsgKind     = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount  = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_WriteFile_eo39;
  dbgMsgBuf->dbgMsgParm.v.MessageFileP.i4 = FileHandle;
  MsgLen = strlen(msg);
  memcpy(&dbgMsgBuf->dbgMsgParm.v.MessageFileP.msg[0], msg, MsgLen);
  dbgMsgBuf->dbghdr.hdbgBufLen += sizeof(dbgMsgBuf->dbgMsgParm.v.MessageFileP.i4) + MsgLen;
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  *lpb_error = dbgMsgBuf->dbgMsgParm.v.BasisErrorP;
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::CloseFile(tsp00_Int4 FileHandle,
          short  *lpb_error)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
    
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
  dbgMsgBuf->dbghdr.hdbgMsgKind     = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount  = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_CloseFile_eo39;
  dbgMsgBuf->dbgMsgParm.v.MessageFileP.i4 = FileHandle;
  dbgMsgBuf->dbghdr.hdbgBufLen += sizeof(dbgMsgBuf->dbgMsgParm.v.MessageFileP.i4);
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  *lpb_error = dbgMsgBuf->dbgMsgParm.v.BasisErrorP;
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::SyncFile(char* lpFileName,
          tsp00_Int4 FileHandle,
          short  *lpb_error)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
  int                 FileNameLen;        
    
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
  dbgMsgBuf->dbghdr.hdbgMsgKind     = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount  = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_SyncFile_eo39;
  dbgMsgBuf->dbgMsgParm.v.MessageFileP.i4 = FileHandle;
  if (lpFileName != NULL) {
    FileNameLen = strlen(lpFileName);
    memcpy(&dbgMsgBuf->dbgMsgParm.v.MessageFileP.msg[0], lpFileName, FileNameLen);
  }
  else {
    dbgMsgBuf->dbgMsgParm.v.MessageFileP.msg[0] = 0;
    FileNameLen = 1;
  }
  dbgMsgBuf->dbghdr.hdbgBufLen += sizeof(dbgMsgBuf->dbgMsgParm.v.MessageFileP) - 
    sizeof(dbgMsgBuf->dbgMsgParm.v.MessageFileP.msg) + FileNameLen;
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  *lpb_error = dbgMsgBuf->dbgMsgParm.v.BasisErrorP;
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::GetDateTime(tsp00_Date *Date,
          tsp00_Time *Time)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
    
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
  dbgMsgBuf->dbghdr.hdbgMsgKind     = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount  = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_GetDateTime_eo39;
  
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  *Date = dbgMsgBuf->dbgMsgParm.v.GetDateTime_rplyP.date;
  *Time = dbgMsgBuf->dbgMsgParm.v.GetDateTime_rplyP.time;
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::GetClock(tsp00_Int4 *Sec,
          tsp00_Int4 *MSec)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
    
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
  dbgMsgBuf->dbghdr.hdbgMsgKind     = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount  = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_GetClock_eo39;
  
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  *Sec  = dbgMsgBuf->dbgMsgParm.v.GetClock_rplyP.sec;
  *MSec = dbgMsgBuf->dbgMsgParm.v.GetClock_rplyP.msec;
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::DropVersion(unsigned char  *lpVersionContext,
            short  *lpb_error)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
    
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
  dbgMsgBuf->dbghdr.hdbgMsgKind     = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount  = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_DropVersion_eo39;
  memcpy(&dbgMsgBuf->dbgMsgParm.v.DropVersionP, lpVersionContext, 
    sizeof(dbgMsgBuf->dbgMsgParm.v.DropVersionP));
  dbgMsgBuf->dbghdr.hdbgBufLen += sizeof(dbgMsgBuf->dbgMsgParm.v.DropVersionP);
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  *lpb_error = dbgMsgBuf->dbgMsgParm.v.BasisErrorP;
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::CreateSchema(tsp00_KnlIdentifier *SchemaName,
                                          tsp00_Int4           *SchemaId,
                                          short                *lpb_error)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
    
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
  dbgMsgBuf->dbghdr.hdbgMsgKind     = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount  = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_CreateSchema_eo39;
  memcpy(&dbgMsgBuf->dbgMsgParm.v.CreateSchemaP.schemaName, SchemaName,  
    sizeof(dbgMsgBuf->dbgMsgParm.v.CreateSchemaP.schemaName));
  dbgMsgBuf->dbghdr.hdbgBufLen += sizeof(dbgMsgBuf->dbgMsgParm.v.CreateSchemaP);
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  *lpb_error = dbgMsgBuf->dbgMsgParm.v.CreateSchema_rplyP.e;
  *SchemaId  = dbgMsgBuf->dbgMsgParm.v.CreateSchema_rplyP.schemaId;
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::ExistsSchema(tsp00_KnlIdentifier *SchemaName,
                                          tsp00_Int4           *SchemaId,
                                          short                *lpb_error)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
    
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
  dbgMsgBuf->dbghdr.hdbgMsgKind     = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount  = NULL;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_ExistsSchema_eo39;
  memcpy(&dbgMsgBuf->dbgMsgParm.v.CreateSchemaP.schemaName, SchemaName,  
    sizeof(dbgMsgBuf->dbgMsgParm.v.CreateSchemaP.schemaName));
  dbgMsgBuf->dbghdr.hdbgBufLen += sizeof(dbgMsgBuf->dbgMsgParm.v.CreateSchemaP);
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  *lpb_error = dbgMsgBuf->dbgMsgParm.v.CreateSchema_rplyP.e;
  *SchemaId  = dbgMsgBuf->dbgMsgParm.v.CreateSchema_rplyP.schemaId;
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::DropContainer(OmsSchemaHandle schema,
                                         const GUID&        guid,
                                         OmsContainerNo     cno,
                                         tgg00_BasisError&  error)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
    
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
  dbgMsgBuf->dbghdr.hdbgMsgKind     = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount  = NULL;
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_DropContainer_eo39;
  dbgMsgBuf->dbgMsgParm.v.DropContainerP.schemaId = schema;
  dbgMsgBuf->dbgMsgParm.v.DropContainerP.guid     = guid;
  dbgMsgBuf->dbgMsgParm.v.DropContainerP.cno      = cno;
  dbgMsgBuf->dbghdr.hdbgBufLen += sizeof(dbgMsgBuf->dbgMsgParm.v.DropContainerP);
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  error = dbgMsgBuf->dbgMsgParm.v.BasisErrorP;
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::DropSchema(tsp00_Int4 SchemaId,
                                         short      *lpb_error)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
    
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
  dbgMsgBuf->dbghdr.hdbgMsgKind     = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount  = NULL;
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_DropSchema_eo39;
  dbgMsgBuf->dbgMsgParm.v.DropSchemaP.schemaId = SchemaId;  
  dbgMsgBuf->dbghdr.hdbgBufLen += sizeof(dbgMsgBuf->dbgMsgParm.v.DropSchemaP);
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  *lpb_error = dbgMsgBuf->dbgMsgParm.v.BasisErrorP;
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::GetSchemaName(tsp00_Int4 SchemaId,
                                         tsp00_KnlIdentifier *lpSchemaName,
                                         short               *lpb_error)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
    
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
  dbgMsgBuf->dbghdr.hdbgMsgKind     = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount  = NULL;
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_GetSchemaName_eo39;
  dbgMsgBuf->dbgMsgParm.v.GetSchemaNameP.schemaId = SchemaId;  
  dbgMsgBuf->dbghdr.hdbgBufLen += sizeof(dbgMsgBuf->dbgMsgParm.v.DropSchemaP);
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  memcpy (lpSchemaName, &dbgMsgBuf->dbgMsgParm.v.GetSchemaName_rplyP.schemaName[0], sizeof(*lpSchemaName));
  *lpb_error = dbgMsgBuf->dbgMsgParm.v.GetSchemaName_rplyP.e;
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::IsObjHistoryUsed( 
            tgg00_ContainerId& containerId, 
            const OmsTypeOid&  oid,
            bool&              isUsed,
            tgg00_BasisError&  e)
{

  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
    
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
  dbgMsgBuf->dbghdr.hdbgMsgKind     = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount  = NULL;
  dbgMsgBuf->dbgMsgParm.pdbgKind   = dbsink_HistoryUsed_eo39;
  dbgMsgBuf->dbgMsgParm.v.HistoryUsedP.containerId = containerId;
  dbgMsgBuf->dbgMsgParm.v.HistoryUsedP.oid         = oid;
  dbgMsgBuf->dbghdr.hdbgBufLen += sizeof(dbgMsgBuf->dbgMsgParm.v.HistoryUsedP);
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  isUsed = dbgMsgBuf->dbgMsgParm.v.HistoryUsedP.isUsed;
  e      = dbgMsgBuf->dbgMsgParm.v.HistoryUsedP.e;
  return hr;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::LoadLibraryEx(const char  *path, 
                                            bool        loadGlobal,
                                            HANDLE      *hd,
                                            tsp_errtext &errtext,
                                            int         errtext_size)
{
  this->CallbackIsActive = true;
  eo38LoadLibraryEx(path, loadGlobal, hd, errtext, errtext_size);
  this->CallbackIsActive = false;
  return S_OK;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::FreeLibrary(HANDLE      hd,
                                          bool        &ok,
                                          tsp_errtext &errtext,
                                          int         errtext_size)
{
  this->CallbackIsActive = true;
  eo38FreeLibrary(hd, &ok, errtext, errtext_size);
  this->CallbackIsActive = false;
  return S_OK;
}

/*---------------------------------------------------------------------------*/

STDMETHODIMP CIliveCacheSink::GetContainerInfo(tsp00_Uint4 containerHandle,
            GUID&            guid,
            OmsSchemaHandle& schema,
            OmsContainerNo&  cno,
            short&           error)
{
  teo39_DebugMsg      *dbgMsgBuf;
  HRESULT             hr;
  CliveCacheDbg       *pLCD;
    
  pLCD = (CliveCacheDbg*)this->lpSessionContext;
  
  dbgMsgBuf = (teo39_DebugMsg*) pLCD->m_CallBackBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
  dbgMsgBuf->dbghdr.hdbgMsgKind     = iSinkRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount  = NULL;
  dbgMsgBuf->dbgMsgParm.pdbgKind    = dbsink_GetContainerInfo_eo39;
  dbgMsgBuf->dbgMsgParm.v.GetContainerInfoP.containerHandle = containerHandle;  
  dbgMsgBuf->dbghdr.hdbgBufLen += sizeof(dbgMsgBuf->dbgMsgParm.v.GetContainerInfoP);
  hr = os38_send_dbsink( this, &dbgMsgBuf->dbghdr.hdbgBufLen, pLCD->m_CallBackBuf);
  guid   = dbgMsgBuf->dbgMsgParm.v.GetContainerInfo_rplyP.guid;
  schema = dbgMsgBuf->dbgMsgParm.v.GetContainerInfo_rplyP.schema;
  cno    = dbgMsgBuf->dbgMsgParm.v.GetContainerInfo_rplyP.cno;
  error  = dbgMsgBuf->dbgMsgParm.v.GetContainerInfo_rplyP.e;
  return S_OK;
}

/*===========================================================================*
 *  LOCAL FUNCTIONS (CODE)                                                   *
 *===========================================================================*/

static HRESULT os38_send_dbsink(IliveCacheSink *This, 
                                int *DbgpBufLength, 
                                unsigned char *pDbgBuffer)
{
  CliveCacheDbg  *pLCD;
  teo39_DebugMsg *p = (teo39_DebugMsg *) pDbgBuffer;
  HANDLE          hEventArray[1];
  DWORD               WaitRc;
  
  pLCD = *((CliveCacheDbg**)(This + 1));
  switch (p->dbghdr.hdbgMsgKind)
  {
  case iSinkRequest_eo39:
    pLCD->m_buffer_ptr = pDbgBuffer;
    SetEvent(pLCD->m_hCoReplyEvent) ;
    hEventArray[0] = pLCD->m_hCoRequestEvent;
    
    WaitRc = ::WaitForMultipleObjects(1,
      hEventArray,
      FALSE,
      INFINITE);
    switch(WaitRc)
    {
    case WAIT_OBJECT_0:
      p = (teo39_DebugMsg *) pLCD->m_buffer_ptr;
      *DbgpBufLength = p->dbghdr.hdbgBufLen;
      memcpy(pDbgBuffer,pLCD->m_buffer_ptr,*DbgpBufLength);
      break;
    case WAIT_OBJECT_0 + 1:         
    case WAIT_TIMEOUT:
    default:
      break;
    }
    break;
  case vdcomReply_eo39:
  case vdcomRequest_eo39:
  case iSinkReply_eo39:
    // error
    break;
  }
  
  return pLCD->m_hr;
}

/*
 * =============================== END ========================================
 */

