/* @lastChanged: "1999-04-21  17:24"
 * @filename:    vos39.cpp
 * @purpose:     "OLE_Debug_Functions" 
 * @release:     7.2.0.0
 * @see:         ""
 *
 * @copyright:   "Copyright by SAP AG, 1998"


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 */

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "ggg00.h"
#include "gsp00.h"
#include "vsp001.h"

#include "gos00.h"
#include "veo00c.h"
#include "geo007_1.h"
#include "vos00kc.h"
#include "IliveCacheDbg.h"  /* no check */
#include "IliveCacheSink.h" /* no check */

#include "geo00d.h"

#include "hak34.h"
#include "hak341.h"
#include "geo39.h"
#include "hos39.h"
#include "heo38.h"
#include "heo35.h"
//#include "gos35k.h"
#include "heo01.h"         /* PTS 1107355 */


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#ifdef NO_INLINES
# define _INLINE
#else
# define _INLINE	inline
#endif

#define         DEBUG_INTERFACE_LOCAL  -2
#define         DEBUG_INTERFACE_REMOTE -3

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/



/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/

externC void s26first_segment_init (tsp1_packet_ptr, 
                                    tsp1_segment_kind, 
                                    tsp1_segment_ptr*);

externC void s26new_part_init (tsp1_packet_ptr,
                               tsp1_segment*,
                               tsp1_part_ptr*);

externC void s26finish_part (tsp1_packet_ptr,
                             tsp1_part*);

externC const IID IID_IliveCacheDbg;

// {2FCBCE35-1A43-11d2-A977-00A0C94311A5}
externC const CLSID CLSID_liveCacheDbg;

// {2FCBCE36-1A43-11d2-A977-00A0C94311A5}
externC const CLSID CLSID_LiveCacheRMT;

/*===========================================================================*
 *  GLOBAL VARIABLES                                                         *
 *===========================================================================*/




/*===========================================================================*
 *  LOCAL VARIABLES                                                          *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL FUNCTIONS (PROTOTYPES)                                             *
 *===========================================================================*/

static HRESULT os39InstallEnvironment(IliveCacheDbg *pIliveCacheDbg);

static void os39_HandleCallbacks ( IliveCacheSink  *pIliveCacheSink,
                                    teo39_TransferP *transferBuf, 
                                    bool            *rc_ok );


/*===========================================================================*
 *  GLOBAL FUNCTIONS (CODE)                                                  *
 *===========================================================================*/

externC void os39CoCreateInstance (tsp_dcom_co_create_instance_parms * parms)
{
  ROUTINE_DBG_MSP00 ("os39CoCreateInstance");

  /* still a DUMMY, not used by DB-Procedures */ 
  parms->ret = -1;

  return;
}
/*----------------------------------------------------------------------------*/

externC void os39CoGetClassobject (tsp_dcom_co_create_instance_parms * parms)
{
  ROUTINE_DBG_MSP00 ("os39CoGetClassobject");

  teo39_DebugMsg      *dbgMsgBuf;
  teo39_TransferP     transferBuf;
  bool                rc_ok = true;
  tsp_errtext         errtext;
  unsigned char       buf2;
  
  ALLOC_MEM((void **)&dbgMsgBuf,MARSHAL_BUF_MXEO39);
  memcpy(&transferBuf.session, parms->session, sizeof (parms->session));
  
  transferBuf.hr                  = S_OK;
  transferBuf.dbgMsg              = dbgMsgBuf;
  transferBuf.SQLBuffer2_size     = sizeof(buf2);
  transferBuf.SQLBuffer2_length   = 0;
  transferBuf.SQLBuffer_ptr2      = &buf2;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) + 
    sizeof(*parms);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = vdcomRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = 1;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = coGetClassObject_eo39;
  memcpy(&dbgMsgBuf->dbgMsgParm.v.CoCreateP, parms, sizeof(*parms));
  
  sql35_VDCOMTransfer ( (void*)&transferBuf, &rc_ok, &errtext );
  if (( ! rc_ok ) || ( transferBuf.hr != S_OK ))  
  {
    if ( transferBuf.hr != S_OK ) 
      parms->ret = transferBuf.hr;
    else 
      parms->ret = -1;
  }
  else
  {
    os39_HandleCallbacks  ( (IliveCacheSink*)parms->ikernel_sink, &transferBuf, &rc_ok ); 
    if (( ! rc_ok ) || ( transferBuf.hr != S_OK ))  
    {
      if ( transferBuf.hr != S_OK ) 
        parms->ret = transferBuf.hr;
      else 
        parms->ret = -1;
    }
    else
    {
      memcpy(parms, &dbgMsgBuf->dbgMsgParm.v.CoCreateP, sizeof(*parms));
    }
  }
  FREE_MEM(dbgMsgBuf);
};

/*----------------------------------------------------------------------------*/

externC void os39VDcomCreateInstance (tsp_dcom_co_create_instance_parms * parms)
{
  ROUTINE_DBG_MSP00 ("os39VDcomCreateInstance");
  
  teo39_DebugMsg      *dbgMsgBuf;
  teo39_TransferP     transferBuf;
  bool                rc_ok = true;
  tsp_errtext         errtext;
  unsigned char       buf2;
  
  transferBuf.hr                   = S_OK;
  ALLOC_MEM((void**)&dbgMsgBuf,MARSHAL_BUF_MXEO39);
  memcpy(&transferBuf.session, parms->session, sizeof (parms->session));
  transferBuf.dbgMsg = dbgMsgBuf;
  transferBuf.SQLBuffer2_size      = sizeof(buf2);
  transferBuf.SQLBuffer2_length    = 0;
  transferBuf.SQLBuffer_ptr2       = &buf2;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) + 
    sizeof(*parms);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = vdcomRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = 1;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = vdcomCreateInstance_eo39;
  memcpy(&dbgMsgBuf->dbgMsgParm.v.CoCreateP, parms, sizeof(*parms));
  
  sql35_VDCOMTransfer ( (void*)&transferBuf, &rc_ok, &errtext );
  if (( ! rc_ok ) || ( transferBuf.hr != S_OK ))  
  {
    if ( transferBuf.hr != S_OK ) 
      parms->ret = transferBuf.hr;
    else 
      parms->ret = -1;
  }
  else
  {
    os39_HandleCallbacks  ( (IliveCacheSink*)parms->ikernel_sink, &transferBuf, &rc_ok ); 
    if (( ! rc_ok ) || ( transferBuf.hr != S_OK ))  
    {
      if ( transferBuf.hr != S_OK ) 
        parms->ret = transferBuf.hr;
      else 
        parms->ret = -1;
    }
    else
    {
      memcpy(parms, &dbgMsgBuf->dbgMsgParm.v.CoCreateP, sizeof(*parms));
    }
  }   
  FREE_MEM(dbgMsgBuf);
};

/*----------------------------------------------------------------------------*/

externC void os39VDcomDBProc (tsp_dcom_dispatch_parms * parms)
{
  ROUTINE_DBG_MSP00 ("os39VDcomDBProc");
  
  teo39_DebugMsg*     dbgMsgBuf;
  bool                rc_ok = true;
  tsp_errtext         errtext;
  IliveCacheSink      *pIliveCacheSink;
  teo39_TransferP     transferBuf;
  short               param_cnt;   
  int                 i;
  unsigned char*      ptr; 
  unsigned char*      ptr2; 
  teo39_TStack*       ptr_tstack; 
  unsigned char       buf2;
  
  ak341GetDebugContext ((tak_all_command_glob*)parms->session_context,
        *REINTERPRET_CAST(tgg91_SessionNo*, &transferBuf.session,),
        *REINTERPRET_CAST(tsp00_Addr*, &pIliveCacheSink),
        *REINTERPRET_CAST(tsp1_packet_ptr*, &transferBuf.SQLBuffer_ptr));
  transferBuf.hr     = S_OK;
  ALLOC_MEM((void **)&dbgMsgBuf,MARSHAL_BUF_MXEO39);
  transferBuf.dbgMsg = dbgMsgBuf;
  transferBuf.SQLBuffer2_size      = sizeof(buf2);
  transferBuf.SQLBuffer2_length    = 0;
  transferBuf.SQLBuffer_ptr2       = &buf2;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) + 
    sizeof(*parms);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = vdcomRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = 1;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = vdcomDbproc_eo39;
  memcpy(&dbgMsgBuf->dbgMsgParm.v.DispP, parms, sizeof(*parms));
  
  param_cnt = (short)parms->param_stack_len / sizeof(teo39_TStack);
  
  ptr = (unsigned char*)dbgMsgBuf;
  ptr = ptr + MARSHAL_BUF_MXEO39 - parms->param_stack_len;
  memcpy( ptr, parms->param_stack, parms->param_stack_len );

  /* move INOUT parameters to stack */
  ptr_tstack = (teo39_TStack*) ptr;
  for ( i=1; i <= param_cnt; ptr_tstack++, i++ )
  {
    if ( ptr_tstack->inout_param == inout_eeo39 )
    {
      ptr = (unsigned char*)dbgMsgBuf;
      ptr = ptr + dbgMsgBuf->dbghdr.hdbgBufLen;
      memcpy( ptr, ptr_tstack->value.ptr, ptr_tstack->length );
      ptr_tstack->value.ptr = (void*)(tsp00_Longuint) dbgMsgBuf->dbghdr.hdbgBufLen;
      dbgMsgBuf->dbghdr.hdbgBufLen = dbgMsgBuf->dbghdr.hdbgBufLen + 
        ptr_tstack->length;
    }
  }
  
  /* move IN parameters to stack */
  ptr = (unsigned char*)dbgMsgBuf;
  ptr = ptr + MARSHAL_BUF_MXEO39 - parms->param_stack_len;
  ptr_tstack = (teo39_TStack*) ptr;
  for ( i=1; i <= param_cnt; ptr_tstack++, i++ )
  {
    if (( ptr_tstack->inout_param == in_eeo39 ) && ptr_tstack->is_ptr )
    {
      ptr = (unsigned char*)dbgMsgBuf;
      ptr = ptr + dbgMsgBuf->dbghdr.hdbgBufLen;
      memcpy( ptr, ptr_tstack->value.ptr, ptr_tstack->length );
      ptr_tstack->value.ptr = (void*)(tsp00_Longuint) dbgMsgBuf->dbghdr.hdbgBufLen;
      dbgMsgBuf->dbghdr.hdbgBufLen = dbgMsgBuf->dbghdr.hdbgBufLen + 
        ptr_tstack->length;
    }
  }
  
  ptr = (unsigned char*)dbgMsgBuf;
  ptr = ptr + dbgMsgBuf->dbghdr.hdbgBufLen;
  ptr2 = (unsigned char*)dbgMsgBuf;
  ptr2 = ptr2 + MARSHAL_BUF_MXEO39 - parms->param_stack_len;
  memmove( ptr, ptr2, parms->param_stack_len );
  dbgMsgBuf->dbgMsgParm.v.DispP.param_stack = (tsp_moveobj_ptr)(tsp00_Longuint)dbgMsgBuf->dbghdr.hdbgBufLen;
  dbgMsgBuf->dbghdr.hdbgBufLen = dbgMsgBuf->dbghdr.hdbgBufLen + 
    parms->param_stack_len;
  
  sql35_VDCOMTransfer ( (void*)&transferBuf, &rc_ok, &errtext );
  if (( ! rc_ok ) || ( transferBuf.hr != S_OK ))  
  {
    if ( transferBuf.hr != S_OK ) 
      parms->hresult = transferBuf.hr;
    else 
      parms->hresult = -1;
  }
  else
  {
    os39_HandleCallbacks  ( pIliveCacheSink, &transferBuf, &rc_ok ); 
    if ( rc_ok && (transferBuf.hr == S_OK)) 
    {
      ptr = (unsigned char*)dbgMsgBuf;
      ptr = ptr + 
        sizeof(dbgMsgBuf->dbghdr) + 
        sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) + 
        sizeof(*parms);   
      param_cnt = (short)parms->param_stack_len / sizeof(teo39_TStack);
      
      /* move INOUT + OUT parameters from buffer */
      ptr_tstack = (teo39_TStack*) parms->param_stack;
      for ( i=1; i <= param_cnt; ptr_tstack++, i++ )
      {
        if (( ptr_tstack->inout_param == inout_eeo39 ) ||
          ( ptr_tstack->inout_param == out_eeo39 ))
        {
          memcpy( ptr_tstack->value.ptr, ptr, ptr_tstack->length );
          ptr = ptr + ptr_tstack->length;
        }
      }
      parms->hresult = dbgMsgBuf->dbgMsgParm.v.DispP.hresult;
      memcpy(&parms->errtext[0], &dbgMsgBuf->dbgMsgParm.v.DispP.errtext[0], sizeof(parms->errtext));
    }
  }
  FREE_MEM(dbgMsgBuf);
};

/*----------------------------------------------------------------------------*/

externC void os39VDcomGetDispAddr (tsp_dcom_co_create_instance_parms * parms)
{
  ROUTINE_DBG_MSP00 ("os39VDcomGetDispAddr");
  
  teo39_DebugMsg      *dbgMsgBuf;
  teo39_TransferP     transferBuf;
  bool                rc_ok = true;
  tsp_errtext         errtext;
  unsigned char       buf2;
  
  ALLOC_MEM((void **)&dbgMsgBuf,MARSHAL_BUF_MXEO39);
  transferBuf.dbgMsg = dbgMsgBuf;
  transferBuf.SQLBuffer2_size      = sizeof(buf2);
  transferBuf.SQLBuffer2_length    = 0;
  transferBuf.SQLBuffer_ptr2       = &buf2;

  transferBuf.hr     = S_OK;
  memcpy(&transferBuf.session, parms->session, sizeof (parms->session));
  transferBuf.dbgMsg = dbgMsgBuf;
  dbgMsgBuf->dbghdr.hdbgBufSize     = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen      = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) + 
    sizeof(*parms);
  dbgMsgBuf->dbghdr.hdbgMsgKind     = vdcomRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount  = 1;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind    = vdcomGetDispaddr_eo39;
  memcpy(&dbgMsgBuf->dbgMsgParm.v.CoCreateP, parms, sizeof(*parms));
  sql35_VDCOMTransfer ( (void*)&transferBuf, &rc_ok, &errtext );
  if (( ! rc_ok ) || ( transferBuf.hr != S_OK ))  
  {
    if ( transferBuf.hr != S_OK ) 
      parms->ret = transferBuf.hr;
    else 
      parms->ret = -1;
  }
  else
  {
    os39_HandleCallbacks  ( (IliveCacheSink*)parms->ikernel_sink, &transferBuf, &rc_ok ); 
    if (( ! rc_ok ) || ( transferBuf.hr != S_OK ))  
    {
      if ( transferBuf.hr != S_OK ) 
        parms->ret = transferBuf.hr;
      else 
        parms->ret = -1;
    }
    else
    {
      memcpy(parms, &dbgMsgBuf->dbgMsgParm.v.CoCreateP, sizeof(*parms));
    }
  }
  FREE_MEM(dbgMsgBuf);
};


/*----------------------------------------------------------------------------*/

externC void os39VDcomParamInfo (tsp_dcom_paraminfo_parms *parms)
{
  ROUTINE_DBG_MSP00 ("os39VDcomParamInfo");
  
  
  teo39_DebugMsg      *dbgMsgBuf;
  teo39_TransferP     transferBuf;
  tsp1_segment_kind   segm_kind;
  tsp1_segment_ptr    lpSegment;
  tsp1_packet_ptr     packet_ptr;
  bool                rc_ok = true;
  tsp_errtext         errtext;
  
  tsp00_C4          session;
  IliveCacheSink*   pIliveCacheSink;
  tsp1_packet*      SQLBuffer_ptr;
  unsigned char       buf2;
  unsigned char*      ptr;
  
  ak341GetDebugContext ((tak_all_command_glob*)parms->session_context,
        *REINTERPRET_CAST(tgg91_SessionNo*, &session,),
        *REINTERPRET_CAST(tsp00_Addr*, &pIliveCacheSink),
        SQLBuffer_ptr );

  memset (parms->errtext, ' ', sizeof (parms->errtext));
  packet_ptr = (tsp1_packet_ptr)parms->packet;
  
  ALLOC_MEM((void **)&dbgMsgBuf,MARSHAL_BUF_MXEO39);
  transferBuf.dbgMsg = dbgMsgBuf;
  transferBuf.SQLBuffer2_size      = sizeof(buf2);
  transferBuf.SQLBuffer2_length    = 0;
  transferBuf.SQLBuffer_ptr2       = &buf2;
 
  memcpy(&transferBuf.session, parms->session, sizeof (parms->session));
  
  transferBuf.hr     = S_OK;
  
  dbgMsgBuf->dbghdr.hdbgBufSize     = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen      =   sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) + 
    sizeof(*parms);
  dbgMsgBuf->dbghdr.hdbgMsgKind     = vdcomRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount  = 1;
  dbgMsgBuf->dbgMsgParm.pdbgKind     = vdcom_paraminfo_eo39;
  memcpy(&dbgMsgBuf->dbgMsgParm.v.ParamInfoP, parms, sizeof(*parms));
  ptr = (unsigned char*)dbgMsgBuf;
  ptr = ptr + dbgMsgBuf->dbghdr.hdbgBufLen;
  dbgMsgBuf->dbgMsgParm.v.ParamInfoP.packet   =  (tsp_moveobj*)ptr;
  memcpy( ptr, parms->packet, sizeof(packet_ptr->sp1_header) );
  dbgMsgBuf->dbghdr.hdbgBufLen      = dbgMsgBuf->dbghdr.hdbgBufLen + sizeof(packet_ptr->sp1_header);
  
  sql35_VDCOMTransfer ( (void*)&transferBuf, &rc_ok, &errtext );
  if (( ! rc_ok ) || ( transferBuf.hr != S_OK ))  
  {
    segm_kind.becomes(sp1sk_return);
    s26first_segment_init (packet_ptr, segm_kind, &lpSegment);
    lpSegment->sp1s_segm_header().sp1r_returncode() = -1;
    lpSegment->sp1s_segm_header().sp1r_errorpos()   = -1;
    if ( transferBuf.hr != S_OK ) 
      lpSegment->sp1s_segm_header().sp1r_errorpos() = transferBuf.hr;
  }
  else
  {
    os39_HandleCallbacks  ( pIliveCacheSink, &transferBuf, &rc_ok ); 
    if (( ! rc_ok ) || ( transferBuf.hr != S_OK ))  
    {
      segm_kind.becomes(sp1sk_return);
      s26first_segment_init (packet_ptr, segm_kind, &lpSegment);
      lpSegment->sp1s_segm_header().sp1r_returncode() = -1;
      lpSegment->sp1s_segm_header().sp1r_errorpos()   = -1;
      if ( transferBuf.hr != S_OK ) 
        lpSegment->sp1s_segm_header().sp1r_errorpos() = transferBuf.hr;
    }
    else
    {
      memcpy(parms, &dbgMsgBuf->dbgMsgParm.v.ParamInfoP, sizeof(*parms));
      memcpy(packet_ptr, ptr, ((tsp1_packet_ptr)ptr)->sp1_header.sp1h_varpart_len + sizeof(tsp1_packet_header));
      parms->packet =  (tsp_moveobj*)packet_ptr;
    }
  }
  FREE_MEM(dbgMsgBuf);
};

/*----------------------------------------------------------------------------*/

externC void os39VDcomTransEnd (tsp_dcom_dispatch_parms *parms)
{
  ROUTINE_DBG_MSP00 ("os39VDcomTransEnd");

  teo39_DebugMsg*     dbgMsgBuf;
  bool                rc_ok = true;
  tsp_errtext         errtext;
  IliveCacheSink      *pIliveCacheSink;
  teo39_TransferP     transferBuf;
  unsigned char*      ptr;
  teo39_Transend*     lpstack;
  unsigned char       buf2;
  
  
  ak341GetDebugContext ((tak_all_command_glob*)parms->session_context,
        *REINTERPRET_CAST(tgg91_SessionNo*, &transferBuf.session,),
        *REINTERPRET_CAST(tsp00_Addr*, &pIliveCacheSink),
        *REINTERPRET_CAST(tsp1_packet_ptr*, &transferBuf.SQLBuffer_ptr));

  transferBuf.hr     = S_OK;
  ALLOC_MEM((void **)&dbgMsgBuf,MARSHAL_BUF_MXEO39);

  transferBuf.dbgMsg = dbgMsgBuf;
  transferBuf.SQLBuffer2_size      = sizeof(buf2);
  transferBuf.SQLBuffer2_length    = 0;
  transferBuf.SQLBuffer_ptr2       = &buf2;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) + 
    sizeof(*parms);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = vdcomRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = 1;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = vdcomTransEnd_eo39;
  memcpy(&dbgMsgBuf->dbgMsgParm.v.DispP, parms, sizeof(*parms));
  
  ptr = (unsigned char*)dbgMsgBuf;
  ptr = ptr + dbgMsgBuf->dbghdr.hdbgBufLen;
  lpstack = (teo39_Transend*) ptr;
  memcpy( lpstack, parms->param_stack, parms->param_stack_len );
  dbgMsgBuf->dbgMsgParm.v.DispP.param_stack = (tsp_moveobj_ptr)(tsp00_Longuint)dbgMsgBuf->dbghdr.hdbgBufLen;
  dbgMsgBuf->dbghdr.hdbgBufLen = dbgMsgBuf->dbghdr.hdbgBufLen + parms->param_stack_len;

  ptr  = (unsigned char*)dbgMsgBuf;
  ptr  = ptr + dbgMsgBuf->dbghdr.hdbgBufLen;
  lpstack->buflen = *lpstack->lpbuflen;
  memcpy( ptr, lpstack->lpbuf, lpstack->buflen );

  lpstack->lpbuf = (teo39_TransendBuf*)(tsp00_Longuint)dbgMsgBuf->dbghdr.hdbgBufLen;
  dbgMsgBuf->dbghdr.hdbgBufLen = dbgMsgBuf->dbghdr.hdbgBufLen + lpstack->buflen;
  
  sql35_VDCOMTransfer ( (void*)&transferBuf, &rc_ok, &errtext );
  if (( ! rc_ok ) || ( transferBuf.hr != S_OK ))  
  {
    if ( transferBuf.hr != S_OK ) 
      parms->hresult = transferBuf.hr;
    else 
      parms->hresult = -1;
  }
  else
  {
    os39_HandleCallbacks  ( pIliveCacheSink, &transferBuf, &rc_ok ); 
    if (( ! rc_ok ) || ( transferBuf.hr != S_OK ))  
    {
      if ( transferBuf.hr != S_OK ) 
        parms->hresult = transferBuf.hr;
      else 
        parms->hresult = -1;
    }
    else
    {
      ptr = (unsigned char*)dbgMsgBuf;
      ptr = ptr + sizeof(dbgMsgBuf->dbghdr) + 
        sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) + sizeof(*parms);
      lpstack = (teo39_Transend*) parms->param_stack;
      memcpy (lpstack->lpbuflen, ptr, sizeof (*lpstack->lpbuflen));
      ptr += sizeof (*lpstack->lpbuflen);
      memcpy (lpstack->lpbuf, ptr, *lpstack->lpbuflen);
      parms->hresult = dbgMsgBuf->dbgMsgParm.v.DispP.hresult;
      memcpy(&parms->errtext[0], &dbgMsgBuf->dbgMsgParm.v.DispP.errtext[0], sizeof(parms->errtext));
    }
  }
  FREE_MEM(dbgMsgBuf);
}

/*----------------------------------------------------------------------------*/

externC void os39VDcomReleaseAll ( tsp00_Addr SessionContext, bool RemoteFlag, HRESULT *pRc )
{
  ROUTINE_DBG_MSP00 ("eo38VDcomReleaseAll");

  teo39_DebugMsg*     dbgMsgBuf;
  bool                rc_ok = true;
  tsp_errtext         errtext;
  IliveCacheSink      *pIliveCacheSink;
  teo39_TransferP     transferBuf;
  unsigned char       buf2;
  
  ak341GetDebugContext ((tak_all_command_glob*)SessionContext,
        *REINTERPRET_CAST(tgg91_SessionNo*, &transferBuf.session,),
        *REINTERPRET_CAST(tsp00_Addr*, &pIliveCacheSink),
        *REINTERPRET_CAST(tsp1_packet_ptr*, &transferBuf.SQLBuffer_ptr));

  transferBuf.hr     = S_OK;
  ALLOC_MEM((void **)&dbgMsgBuf,MARSHAL_BUF_MXEO39);

  transferBuf.dbgMsg = dbgMsgBuf;
  transferBuf.SQLBuffer2_size      = sizeof(buf2);
  transferBuf.SQLBuffer2_length    = 0;
  transferBuf.SQLBuffer_ptr2       = &buf2;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) +
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind)  + 
    sizeof(teo39_vdcomReleaseAll);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = vdcomRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = 1;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = vdcomReleaseAll_os39;
  dbgMsgBuf->dbgMsgParm.v.ReleaseAllP.remoteFlag = RemoteFlag;
  
  sql35_VDCOMTransfer ( (void*)&transferBuf, &rc_ok, &errtext );
  if (( ! rc_ok ) || ( transferBuf.hr != S_OK ))  
  {
    if ( transferBuf.hr != S_OK ) 
      *pRc = transferBuf.hr;
    else 
      *pRc = -1;
  }
  else
  {
    os39_HandleCallbacks  ( pIliveCacheSink, &transferBuf, &rc_ok ); 
    if (( ! rc_ok ) || ( transferBuf.hr != S_OK ))  
    {
      if ( transferBuf.hr != S_OK ) 
        *pRc = transferBuf.hr;
      else 
        *pRc = -1;
    }
    else
    {
      *pRc = dbgMsgBuf->dbgMsgParm.v.ReleaseAll_rplyP.hr;
    }
  }
  FREE_MEM(dbgMsgBuf);
}

/*----------------------------------------------------------------------------*/

externC void os39VDcomTransfer (teo39_TransferP *p)
{
  ROUTINE_DBG_MSP00 ("os39VDcomTransfer");
  
  tsp_dcom_co_create_instance_parms loc_parms;
  teo39_DebugMsg *dbgMsgBuf = p->dbgMsg;
  
  /* check if LiveCache-DEBUG-SERVER has been activated for       */
  /* actual session                                               */
  
  memcpy (&loc_parms.session, &p->session, sizeof (loc_parms.session));
  memcpy (&(loc_parms.coclsid), &CLSID_liveCacheDbg, sizeof (GUID));
  memcpy (&(loc_parms.iid), &IID_IliveCacheDbg, sizeof (IID));
  loc_parms.debug_mode = 0;
  switch ( dbgMsgBuf->dbgMsgParm.pdbgKind )
  {
  case vdcomTransEnd_eo39 :
  case vdcomDbproc_eo39 :
    loc_parms.debug_location = dbgMsgBuf->dbgMsgParm.v.DispP.debug_location;
    break;
  case vdcomCreateInstance_eo39 : 
  case vdcomGetDispaddr_eo39 :
  case coGetClassObject_eo39 :
    loc_parms.debug_location = dbgMsgBuf->dbgMsgParm.v.CoCreateP.debug_location;
    break;
  case vdcom_paraminfo_eo39 :
    loc_parms.debug_location = dbgMsgBuf->dbgMsgParm.v.ParamInfoP.debug_location;
    break;
  case vdcomReleaseAll_os39:      /* +++ FF PTS 11046689  +++ */
    if ( dbgMsgBuf->dbgMsgParm.v.ReleaseAllP.remoteFlag ) 
      loc_parms.debug_location = (char*)-1;
    else
      loc_parms.debug_location = 0;
    break;
  default:
    if (dbgMsgBuf->dbghdr.hdbgMsgKind == iSinkRmtReply_eo39) 
      loc_parms.debug_location = (char*)-1;
    else
      loc_parms.debug_location = 0;
    break;
  }
  if ( loc_parms.debug_location != 0 ) 
    memcpy (&(loc_parms.coclsid), &CLSID_LiveCacheRMT, sizeof (GUID));
  eo38VDcomGetDispAddr (&loc_parms);
  if ((loc_parms.ret != 0) && (loc_parms.debug_location != (char*)-1))
  {
    if ( vdcomReleaseAll_os39 == dbgMsgBuf->dbgMsgParm.pdbgKind ) {  /* +++ FF PTS 11046689  +++ */
      loc_parms.ret = 0;
      return;
    }
    eo38InitializeDebugServer(&loc_parms);
    /* PTS 1107355  FF 21-Aug-2000 */
    if (loc_parms.ret == 0) { /* install correct environment within proxy */
      loc_parms.ret = os39InstallEnvironment(reinterpret_cast<IliveCacheDbg *>(loc_parms.idispatch));
    }
  }

  /* Send Request to DEBUG-Server                               */
  if (loc_parms.ret == 0) 
  {
    loc_parms.ret = reinterpret_cast<IliveCacheDbg *>(loc_parms.idispatch)->Debug( 
      dbgMsgBuf->dbghdr.hdbgBufSize,
      reinterpret_cast<long*>(&dbgMsgBuf->dbghdr.hdbgBufLen),
      (unsigned char*)dbgMsgBuf,
      p->SQLBuffer2_size,
      &p->SQLBuffer2_length,
      p->SQLBuffer_ptr2);
    if (loc_parms.ret != 0) 
    {
      MSGCD ((ERR_DCOM_HRESULT, "os39VDcomTransfer:Debug", loc_parms.ret ));
      p->hr = loc_parms.ret;
      if (p->hr) eo38RemoveProxy ( p->session );
    }
  }
  else
      p->hr = loc_parms.ret;
};



/*===========================================================================*
 *  LOCAL FUNCTIONS (CODE)                                                   *
 *===========================================================================*/

/* ++++ PTS 1107355 FF 21-Aug-2000 +++++++++++++++++++++++++++++++++++++++++ */
static HRESULT os39InstallEnvironment(IliveCacheDbg *pIliveCacheDbg) {
  teo39_DebugMsg      *dbgMsgBuf;
  teo39_TransferP     transferBuf;
  bool                rc_ok = true;
  unsigned char       buf2;
  HRESULT             hr;
  
  ALLOC_MEM((void **)&dbgMsgBuf,MARSHAL_BUF_MXEO39);

  transferBuf.hr                  = S_OK;
  transferBuf.dbgMsg              = dbgMsgBuf;
  transferBuf.SQLBuffer2_size     = sizeof(buf2);
  transferBuf.SQLBuffer2_length   = 0;
  transferBuf.SQLBuffer_ptr2      = &buf2;
  dbgMsgBuf->dbghdr.hdbgBufSize    = MARSHAL_BUF_MXEO39;
  dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
    sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) + 
    sizeof(teo39_vdcomSetEnvironment);
  dbgMsgBuf->dbghdr.hdbgMsgKind    = vdcomRequest_eo39;
  dbgMsgBuf->dbghdr.hdbgParamCount = 1;
  
  dbgMsgBuf->dbgMsgParm.pdbgKind   = vdcomSetEnvironment_os39;
  
  tsp01_RteError          RteError;
  if ( sqlGetDbrootPath( *(tsp00_Pathc*)&dbgMsgBuf->dbgMsgParm.v.SetEnvironment.envStr[0], 
    TERM_WITH_DELIMITER_EO01, &RteError )) {
    
    hr = pIliveCacheDbg->Debug( 
      dbgMsgBuf->dbghdr.hdbgBufSize,
      reinterpret_cast<long*>(&dbgMsgBuf->dbghdr.hdbgBufLen),
      (unsigned char*)dbgMsgBuf,
      transferBuf.SQLBuffer2_size,
      &transferBuf.SQLBuffer2_length,
      transferBuf.SQLBuffer_ptr2);
  }
  else {
    hr = GEO00D_DCOM_INSTALLATION_PATH_NOT_FOUND;
  }
  
  FREE_MEM(dbgMsgBuf);
  return( hr );
};
/* ---- PTS 1107355 FF 21-Aug-2000 ------------------------------------------ */

/*----------------------------------------------------------------------------*/

static void os39_HandleCallbacks ( IliveCacheSink  *pIliveCacheSink,
                                    teo39_TransferP *transferBuf, 
                                    bool            *rc_ok ) 
{
  ROUTINE_DBG_MSP00 ("os39_HandleCallbacks");
  
  tsp_errtext         errtext;
  HRESULT             hr;
  short				        berror;
  tsp00_Int4          buflen;
  int                 dummy;
  unsigned char       *lpv;
  unsigned char       *lpVersionContext;
  void                *session_context;
  teo39_DebugMsg      *dbgMsgBuf = transferBuf->dbgMsg;
  
  while ((( dbgMsgBuf->dbghdr.hdbgMsgKind == iSinkRequest_eo39 ) ||
    ( dbgMsgBuf->dbghdr.hdbgMsgKind == iSinkRmtRequest_eo39 )) &&
    *rc_ok &&
    ( transferBuf->hr == S_OK ))
  {
    switch ( dbgMsgBuf->dbgMsgParm.pdbgKind )
    {
    case dbsink_ABAPRead_eo39 :
      buflen = dbgMsgBuf->dbgMsgParm.v.ABAPReadP.buflen;
      lpv    = ((unsigned char*) &dbgMsgBuf->dbgMsgParm.v.ABAPRead_rplyP) +
        sizeof (dbgMsgBuf->dbgMsgParm.v.ABAPRead_rplyP);
      hr = pIliveCacheSink->ABAPRead ( 
        &dbgMsgBuf->dbgMsgParm.v.ABAPReadP.TabHandle,
        &dbgMsgBuf->dbgMsgParm.v.ABAPReadP.count,
        &buflen,
        lpv,
        &dbgMsgBuf->dbgMsgParm.v.ABAPRead_rplyP.e );     
      dbgMsgBuf->dbghdr.hdbgBufLen  = sizeof(dbgMsgBuf->dbghdr) + 
        sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
        sizeof(teo39_dbsink_ABAPRead_rply) + buflen;      
      break;
    case dbsink_ABAPWrite_eo39 :
      lpv    = ((unsigned char*) &dbgMsgBuf->dbgMsgParm.v.ABAPWriteP) +
        sizeof (dbgMsgBuf->dbgMsgParm.v.ABAPWriteP);
      hr = pIliveCacheSink->ABAPWrite( 
        &dbgMsgBuf->dbgMsgParm.v.ABAPWriteP.TabHandle,
        dbgMsgBuf->dbgMsgParm.v.ABAPWriteP.row_size,
        dbgMsgBuf->dbgMsgParm.v.ABAPWriteP.row_count,
        lpv,
        &berror );
      dbgMsgBuf->dbgMsgParm.v.ABAPWrite_rplyP.e = berror;
      dbgMsgBuf->dbghdr.hdbgBufLen  = sizeof(dbgMsgBuf->dbghdr) + 
        sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
        sizeof(teo39_dbsink_ABAPWrite_rply);      
      break;
    case dbsink_AllocPacket_eo39:
      dbgMsgBuf->dbghdr.hdbgBufLen  = sizeof(dbgMsgBuf->dbghdr) + 
        sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
        sizeof(teo39_dbsink_AllocPacket_rply);
      hr = pIliveCacheSink->AllocPacket( 
        &transferBuf->SQLBuffer_ptr,
        &dbgMsgBuf->dbgMsgParm.v.AllocPacket_rplyP.sz);
      break;
    case dbsink_SQLRequest_eo39:
      memcpy(transferBuf->SQLBuffer_ptr,
        &dbgMsgBuf->dbgMsgParm.v,
        dbgMsgBuf->dbghdr.hdbgBufLen - (sizeof(dbgMsgBuf->dbghdr) + 
        sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind)));
      hr = pIliveCacheSink->SQLRequest( 
        0,
        transferBuf->SQLBuffer_ptr,
        &berror );
      dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
        sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
      break;
    case dbsink_SQLReceive_eo39:
      hr = pIliveCacheSink->SQLReceive( &lpv );
      dbgMsgBuf->dbghdr.hdbgBufLen = ((tsp1_packet_ptr)lpv)->sp1_header.sp1h_varpart_len + 
        sizeof (tsp1_packet_header);
      memcpy(&dbgMsgBuf->dbgMsgParm.v, lpv, dbgMsgBuf->dbghdr.hdbgBufLen );
      dbgMsgBuf->dbghdr.hdbgBufLen  = sizeof(dbgMsgBuf->dbghdr) + 
        sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
        dbgMsgBuf->dbghdr.hdbgBufLen;
      break;
    case dbsink_GetErrorMsg_eo39:
      dbgMsgBuf->dbgMsgParm.v.ErrorMsgP.msgType = csp_ascii;  //******** TODO
      hr = pIliveCacheSink->GetErrorMsg( 
        dbgMsgBuf->dbgMsgParm.v.ErrorMsgP.errorno,
        &dbgMsgBuf->dbgMsgParm.v.ErrorMsgP.msglen,
        (unsigned char*)(&dbgMsgBuf->dbgMsgParm.v.ErrorMsgP.msglen) +
        sizeof(dbgMsgBuf->dbgMsgParm.v.ErrorMsgP.msglen));
      dbgMsgBuf->dbghdr.hdbgBufLen = dbgMsgBuf->dbghdr.hdbgBufLen +
        dbgMsgBuf->dbgMsgParm.v.ErrorMsgP.msglen; 
      break;    
    case dbsink_MultiPurpose_eo39:
      hr = pIliveCacheSink->MultiPurpose( dbgMsgBuf->dbgMsgParm.v.MultiPurposeP.MessType,
        dbgMsgBuf->dbgMsgParm.v.MultiPurposeP.MessType2,
        &dbgMsgBuf->dbgMsgParm.v.MultiPurposeP.buflen,
        (unsigned char*)&dbgMsgBuf->dbgMsgParm.v.MultiPurposeP.buflen + sizeof(dbgMsgBuf->dbgMsgParm.v.MultiPurposeP.buflen),
        &berror );
      dbgMsgBuf->dbgMsgParm.v.MultiPurpose_rplyP.b_error = berror;
      dbgMsgBuf->dbghdr.hdbgBufLen     = sizeof(dbgMsgBuf->dbghdr) + 
        sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
        sizeof(dbgMsgBuf->dbgMsgParm.v.MultiPurpose_rplyP) +
        dbgMsgBuf->dbgMsgParm.v.MultiPurposeP.buflen;
      break;
    case dbsink_SetError_eo39:
      session_context = *((void**)((char*)pIliveCacheSink) + 1);
      a34SetError (*REINTERPRET_CAST(tak_all_command_glob*, session_context),
        dbgMsgBuf->dbgMsgParm.v.ErrorMsgP.errorno,
        *REINTERPRET_CAST(tsp00_MoveObj *, ((char*)(&dbgMsgBuf->dbgMsgParm.v.ErrorMsgP.msglen) +
        sizeof(dbgMsgBuf->dbgMsgParm.v.ErrorMsgP.msglen))),
        dbgMsgBuf->dbgMsgParm.v.ErrorMsgP.msglen,
        dbgMsgBuf->dbgMsgParm.v.ErrorMsgP.msgType);
      dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
        sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
      break;
    case dbsink_BasisError_eo39:
      hr = pIliveCacheSink->BasisError( dbgMsgBuf->dbgMsgParm.v.BasisErrorP );
      dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
        sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
      break;
    case dbsink_HResultError_eo39:
      hr = pIliveCacheSink->HResultError( dbgMsgBuf->dbgMsgParm.v.HResultErrorP );
      dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
        sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
      break;
    case dbsink_Vtrace_eo39:
      hr = pIliveCacheSink->Vtrace( dbgMsgBuf->dbgMsgParm.v.BuflenP,
        (unsigned char*)(&dbgMsgBuf->dbgMsgParm.v.BuflenP) +
        sizeof(dbgMsgBuf->dbgMsgParm.v.BuflenP));
      dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
        sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
      break;
    case dbsink_ConsoleMsg_eo39:
      hr = pIliveCacheSink->ConsoleMsg( dbgMsgBuf->dbgMsgParm.v.BuflenP,
        (unsigned char*)(&dbgMsgBuf->dbgMsgParm.v.BuflenP) +
        sizeof(dbgMsgBuf->dbgMsgParm.v.BuflenP));
      dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
        sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
      break;
    case dbsink_Rollback_eo39:
      hr = pIliveCacheSink->Rollback( dbgMsgBuf->dbgMsgParm.v.BuflenP,
        (unsigned char*)(&dbgMsgBuf->dbgMsgParm.v.BuflenP) +
        sizeof(dbgMsgBuf->dbgMsgParm.v.BuflenP),
        &berror);
      dbgMsgBuf->dbgMsgParm.v.BasisErrorP = berror;
      dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
        sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
        sizeof(dbgMsgBuf->dbgMsgParm.v.BasisErrorP);
      break;
    case dbsink_GetObjWithKey_eo39:
      {
        teo39_dbsink_GetObjWithKey  GetObjWithKey_l;
        unsigned char               key[4096];
        unsigned char*              lpkey;
        unsigned char*              lpbody;
        memcpy(&GetObjWithKey_l, &dbgMsgBuf->dbgMsgParm.v.GetObjWithKeyP, sizeof(GetObjWithKey_l) );
        lpkey = (unsigned char*) &dbgMsgBuf->dbgMsgParm.v.GetObjWithKeyP + 
          sizeof (dbgMsgBuf->dbgMsgParm.v.GetObjWithKeyP);
        memcpy(key, lpkey, GetObjWithKey_l.keylen);
        lpbody = (unsigned char*) &dbgMsgBuf->dbgMsgParm.v.GetObjWithKey_rplyP + 
          sizeof (dbgMsgBuf->dbgMsgParm.v.GetObjWithKey_rplyP); 
        lpVersionContext = (GetObjWithKey_l.inVersion) ? (unsigned char*) &GetObjWithKey_l.versionContext : NULL;
        hr = pIliveCacheSink->GetObjWithKey (
          (unsigned char*) &dbgMsgBuf->dbgMsgParm.v.GetObjWithKey_rplyP.consistentView,
          (unsigned char*) &GetObjWithKey_l.containerId,
          lpVersionContext,
          GetObjWithKey_l.keylen,
          key,
          GetObjWithKey_l.getConsistent,
          GetObjWithKey_l.bodysize,
          lpbody,
          (OmsTypeOid*) &dbgMsgBuf->dbgMsgParm.v.GetObjWithKey_rplyP.oid,
          (unsigned char*) &dbgMsgBuf->dbgMsgParm.v.GetObjWithKey_rplyP.ResultSeq,
          &dummy,
          &dbgMsgBuf->dbgMsgParm.v.GetObjWithKey_rplyP.e );
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.GetObjWithKey_rplyP) + GetObjWithKey_l.bodysize;
        break;
      }
    case dbsink_GetObj_eo39:
      {
        teo39_dbsink_GetObj  GetObj_l;
        
        memcpy(&GetObj_l, &dbgMsgBuf->dbgMsgParm.v.GetObjP, sizeof(GetObj_l) );
        lpVersionContext = (GetObj_l.inVersion) ? (unsigned char*) &GetObj_l.versionContext : NULL;
        hr = pIliveCacheSink->GetObj(
          dbgMsgBuf->dbgMsgParm.v.GetObj_rplyP.consistentView,
          dbgMsgBuf->dbgMsgParm.v.GetObj_rplyP.containerId,
          REINTERPRET_CAST(tgg00_OmsVersionContext*, lpVersionContext),
          GetObj_l.oid,
          GetObj_l.do_lock,
          dbgMsgBuf->dbgMsgParm.v.GetObj_rplyP.ResultSeq,
          GetObj_l.rec_size,
          (unsigned char*)(&dbgMsgBuf->dbgMsgParm.v.GetObj_rplyP) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.GetObj_rplyP),
          dbgMsgBuf->dbgMsgParm.v.GetObj_rplyP.ObjLen, /* +++ Change Request 1000203, T.A. 01.12.1998 */
          dummy, /* +++ Change Request 1000139 +++ */
          dbgMsgBuf->dbgMsgParm.v.GetObj_rplyP.e );
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.GetObj_rplyP) +
          GetObj_l.rec_size;
        break;
      }
    case dbsink_UpdateObj_eo39:
      {
        teo39_dbsink_UpdateObj  UpdateObj_l;
        char                    *entry1_pos, *entry2_pos, *entry3_pos;
        tgg00_ObjColDesc        coldesc;
        
        memcpy(&UpdateObj_l, &dbgMsgBuf->dbgMsgParm.v.UpdateObjP, sizeof(UpdateObj_l) );
        memcpy(&dbgMsgBuf->dbgMsgParm.v.GetObj_rplyP.ResultSeq,
          &UpdateObj_l.WantedSeq, sizeof(UpdateObj_l.WantedSeq) );
        entry1_pos = (char*)(&dbgMsgBuf->dbgMsgParm.v.UpdateObjP) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.UpdateObjP);
        entry2_pos = entry1_pos + UpdateObj_l.bodysize;
        entry3_pos = entry2_pos + ( UpdateObj_l.colHashCount_gg00 * sizeof(tgg00_StackEntry));
        
        coldesc.colHashCount_gg00   = UpdateObj_l.colHashCount_gg00;
        coldesc.colInvCount_gg00    = UpdateObj_l.colInvCount_gg00;
        coldesc.colHashList_gg00    = (tgg00_StackListPtr) entry2_pos;
        coldesc.colInvList_gg00     = (tgg00_StackListPtr) entry3_pos;

        lpVersionContext = (UpdateObj_l.inVersion) ? (unsigned char*) &UpdateObj_l.versionContext : NULL;
        
        hr = pIliveCacheSink->UpdateObj(
          (unsigned char*) &UpdateObj_l.consistentView,
          (unsigned char*)&UpdateObj_l.containerId,
          lpVersionContext,
          (unsigned char*)&UpdateObj_l.oid,
          (unsigned char*)&dbgMsgBuf->dbgMsgParm.v.GetObj_rplyP.ResultSeq,
          UpdateObj_l.bodysize,
          (unsigned char*)entry1_pos,
          (unsigned char*)&coldesc,
          &dbgMsgBuf->dbgMsgParm.v.GetObj_rplyP.e );
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.GetObj_rplyP);
        break;
      }
    case dbsink_NewObj_eo39:
      {
        teo39_dbsink_NewObj NewObjP_l; 
        unsigned char*      lpkey;
        
        memcpy(&NewObjP_l, &dbgMsgBuf->dbgMsgParm.v.NewObjP, sizeof(NewObjP_l) );
        lpkey = (unsigned char*) &dbgMsgBuf->dbgMsgParm.v.NewObjP + 
          sizeof (dbgMsgBuf->dbgMsgParm.v.NewObjP);
        lpVersionContext = (NewObjP_l.inVersion) ? (unsigned char*) &NewObjP_l.versionContext : NULL;
        hr = pIliveCacheSink->NewObj (
          (unsigned char*)&NewObjP_l.consistentView,
          (unsigned char*)&NewObjP_l.containerId,
          lpVersionContext,
          NewObjP_l.KeyPos,
          NewObjP_l.KeySize,
          lpkey,
          (OmsTypeOid*)&dbgMsgBuf->dbgMsgParm.v.NewObj_rplyP.oid,
          (unsigned char*) &dbgMsgBuf->dbgMsgParm.v.NewObj_rplyP.ResultSeq,
          &dbgMsgBuf->dbgMsgParm.v.NewObj_rplyP.e );
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.NewObj_rplyP);
        break;
      }
    case dbsink_RegisterClass_eo39:
      {
        teo39_dbsink_RegisterClass    RegisterClassP_l;
        
        memcpy(&RegisterClassP_l, &dbgMsgBuf->dbgMsgParm.v.RegisterClassP, sizeof(RegisterClassP_l));
        if (RegisterClassP_l.invcoldesc.colHashCount_gg00 > 0)
          RegisterClassP_l.invcoldesc.colHashList_gg00 = (tgg00_StackListPtr) 
          ((unsigned char*) &dbgMsgBuf->dbgMsgParm.v.RegisterClass_rplyP + 
          sizeof(dbgMsgBuf->dbgMsgParm.v.RegisterClass_rplyP));
        
        hr = pIliveCacheSink->RegisterClass (
          RegisterClassP_l.schemaId,
          (GUID*)&RegisterClassP_l.guid,
          &RegisterClassP_l.className,
          RegisterClassP_l.size,
          RegisterClassP_l.container_no,
          (unsigned char*)&RegisterClassP_l.invcoldesc,
          (unsigned char*)&dbgMsgBuf->dbgMsgParm.v.RegisterClass_rplyP.containerId,
          &dbgMsgBuf->dbgMsgParm.v.RegisterClass_rplyP.e );
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.RegisterClass_rplyP) +
          (RegisterClassP_l.invcoldesc.colHashCount_gg00 *
          sizeof (tgg00_StackEntry));
        break;
      }
    case dbsink_NextObj_eo39:
      {
        teo39_dbsink_NextObj  NextObj_l;
        unsigned char         *pObj;
        
        memcpy(&NextObj_l, &dbgMsgBuf->dbgMsgParm.v.NextObjP, sizeof(NextObj_l) );
        if (NextObj_l.with_object) {
          pObj = (unsigned char*)(&dbgMsgBuf->dbgMsgParm.v.NextObj_rplyP) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.NextObj_rplyP);
        }
        else {
          pObj = NULL;
        }
        lpVersionContext = (NextObj_l.inVersion) ? (unsigned char*) &NextObj_l.versionContext : NULL;
        hr = pIliveCacheSink->NextObj(
          (unsigned char*)&dbgMsgBuf->dbgMsgParm.v.NextObj_rplyP.consistentView,
          (unsigned char*)&NextObj_l.containerId,
          lpVersionContext,
          (OmsTypeOid*)&dbgMsgBuf->dbgMsgParm.v.NextObj_rplyP.oid,
          (unsigned char*)&dbgMsgBuf->dbgMsgParm.v.NextObj_rplyP.ResultSeq,
          NextObj_l.rec_size,
          pObj,
          &dbgMsgBuf->dbgMsgParm.v.NextObj_rplyP.e );
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.NextObj_rplyP);
        if (pObj) {
          dbgMsgBuf->dbghdr.hdbgBufLen += NextObj_l.rec_size;
        }
        break;
      }
    case dbsink_IsLocked_eo39:
      {
        teo39_dbsink_LockObj     LockObj_l;
        
        memcpy(&LockObj_l, &dbgMsgBuf->dbgMsgParm.v.LockObjP, sizeof(LockObj_l) );
        hr = pIliveCacheSink->IsLocked ( (unsigned char*)&LockObj_l.containerId,
          (OmsTypeOid *)&LockObj_l.oid,
          &dbgMsgBuf->dbgMsgParm.v.BasisErrorP );
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.BasisErrorP);
        break;
      }
    case dbsink_LockObj_eo39:
      {
        teo39_dbsink_LockObj     LockObj_l;
        
        memcpy(&LockObj_l, &dbgMsgBuf->dbgMsgParm.v.LockObjP, sizeof(LockObj_l) );
        hr = pIliveCacheSink->LockObj(
          (unsigned char*) &LockObj_l.consistentView,
          (unsigned char*)&LockObj_l.containerId,
          (OmsTypeOid *)&LockObj_l.oid,
          (unsigned char*)&LockObj_l.WantedSeq,
          &dbgMsgBuf->dbgMsgParm.v.BasisErrorP );
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.BasisErrorP);
        break;
      }
    case dbsink_UnlockObj_eo39:
      {
        teo39_dbsink_UnlockObj     UnlockObj_l;
        
        memcpy(&UnlockObj_l, &dbgMsgBuf->dbgMsgParm.v.LockObjP, sizeof(UnlockObj_l) );
        hr = pIliveCacheSink->UnlockObj( (unsigned char*)&UnlockObj_l.containerId,
          (OmsTypeOid *)&UnlockObj_l.oid,
          &dbgMsgBuf->dbgMsgParm.v.BasisErrorP );
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.BasisErrorP);
        break;
      }
    case dbsink_DeleteObj_eo39:
      {
        teo39_dbsink_UpdateObj  UpdateObj_l;
        char                    *entry1_pos, *entry2_pos;
        tgg00_ObjColDesc        coldesc;
        
        memcpy(&UpdateObj_l, &dbgMsgBuf->dbgMsgParm.v.UpdateObjP, sizeof(UpdateObj_l) );
        entry1_pos = (char*)(&dbgMsgBuf->dbgMsgParm.v.UpdateObjP) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.UpdateObjP),
          entry2_pos = entry1_pos + ( UpdateObj_l.colHashCount_gg00 * sizeof(tgg00_StackEntry));
        
        coldesc.colHashCount_gg00   = UpdateObj_l.colHashCount_gg00;
        coldesc.colInvCount_gg00    = UpdateObj_l.colInvCount_gg00;
        coldesc.colHashList_gg00    = (tgg00_StackListPtr) entry1_pos;
        coldesc.colInvList_gg00     = (tgg00_StackListPtr) entry2_pos;

        lpVersionContext = (UpdateObj_l.inVersion) ? (unsigned char*) &UpdateObj_l.versionContext : NULL;
        hr = pIliveCacheSink->DeleteObj(
          (unsigned char*)&UpdateObj_l.consistentView,
          (unsigned char*)&UpdateObj_l.containerId,
          lpVersionContext,
          (unsigned char*)&UpdateObj_l.oid,
          (unsigned char*)&UpdateObj_l.WantedSeq,
          (unsigned char*)&coldesc,
          &dbgMsgBuf->dbgMsgParm.v.BasisErrorP );
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.BasisErrorP);
        break;
      }
    case dbsink_SubtransCommit_eo39 :
      {
        hr = pIliveCacheSink->SubtransCommit (&dbgMsgBuf->dbgMsgParm.v.BasisErrorP);
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.BasisErrorP);
        break;
      }
    case dbsink_SubtransRollback_eo39 :
      {
        hr = pIliveCacheSink->SubtransRollback (&dbgMsgBuf->dbgMsgParm.v.BasisErrorP);
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.BasisErrorP);
        break;
      }
    case dbsink_SubtransStart_eo39 :
      {
        hr = pIliveCacheSink->SubtransStart (&dbgMsgBuf->dbgMsgParm.v.BasisErrorP);
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.BasisErrorP);
        break;
      }
     case dbsink_EnterCriticalSection_eo39 :
      {
        hr = pIliveCacheSink->EnterCriticalSection(dbgMsgBuf->dbgMsgParm.v.CriticalSectionP.RegionId);
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
        break;
      }
    case dbsink_LeaveCriticalSection_eo39 :
      {
        hr = pIliveCacheSink->LeaveCriticalSection(dbgMsgBuf->dbgMsgParm.v.CriticalSectionP.RegionId);
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind);
        break;
      }
    case dbsink_NewConsistentView_eo39 :
      {
        hr = pIliveCacheSink->NewConsistentView(
          (unsigned char*) &dbgMsgBuf->dbgMsgParm.v.EndConsistentViewP.consistentView,
          &dbgMsgBuf->dbgMsgParm.v.EndConsistentView_rplyP.e);
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.EndConsistentView_rplyP);
        break;
      }
    case dbsink_EndConsistentView_eo39 :
      {
        hr = pIliveCacheSink->EndConsistentView(
          (unsigned char*) &dbgMsgBuf->dbgMsgParm.v.EndConsistentViewP.consistentView,
          &dbgMsgBuf->dbgMsgParm.v.EndConsistentView_rplyP.e);
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.EndConsistentView_rplyP);
        break;
      }
    case dbsink_CloseVersion_eo39 :
      {
        hr = pIliveCacheSink->CloseVersion(&dbgMsgBuf->dbgMsgParm.v.CloseVersionP.versionId,
          dbgMsgBuf->dbgMsgParm.v.CloseVersionP.lastKey,
          &berror);
        dbgMsgBuf->dbgMsgParm.v.BasisErrorP = berror;
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.BasisErrorP);
        break;
      }
    case dbsink_CreateOpenVersion_eo39 :
      {
        teo39_dbsink_CreateOpenVersion  CreateOpenVersion_l;
        memcpy (&CreateOpenVersion_l, &dbgMsgBuf->dbgMsgParm.v.CreateOpenVersionP, sizeof(CreateOpenVersion_l));
        hr = pIliveCacheSink->CreateOpenVersion((unsigned char*) &CreateOpenVersion_l.consistentView,
          &CreateOpenVersion_l.versionId,
          CreateOpenVersion_l.createVersion,
          (unsigned char*) &dbgMsgBuf->dbgMsgParm.v.CreateOpenVersion_rplyP.VersionContext, 
          &dbgMsgBuf->dbgMsgParm.v.CreateOpenVersion_rplyP.e);
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.CreateOpenVersion_rplyP);
        break;
      }
    case dbsink_CloseFile_eo39 :
      {
        teo39_dbsink_MessageFile& FileInfo = dbgMsgBuf->dbgMsgParm.v.MessageFileP;
        hr = pIliveCacheSink->CloseFile(FileInfo.i4, &dbgMsgBuf->dbgMsgParm.v.BasisErrorP); 
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.BasisErrorP);
        break;
      }
    case dbsink_DropFile_eo39 :
      {
        teo39_dbsink_MessageFile& FileInfo = dbgMsgBuf->dbgMsgParm.v.MessageFileP;
        hr = pIliveCacheSink->DropFile(&FileInfo.msg[0], FileInfo.e);
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(FileInfo.i4)                    +
          sizeof(FileInfo.e);
        break;
      }
    case dbsink_ExistsFile_eo39 :
      {
        teo39_dbsink_ExistsFile& FileInfo = dbgMsgBuf->dbgMsgParm.v.ExistsFileP;
        hr = pIliveCacheSink->ExistsFile(&FileInfo.msg[0], FileInfo.doesExist, FileInfo.isDirectory, FileInfo.e);
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(FileInfo.doesExist)             +
          sizeof(FileInfo.isDirectory)           +
          sizeof(FileInfo.e);
        break;
      }
    case dbsink_OpenFile_eo39 :
      {
        tsp00_Int4 FileHandle;
        teo39_dbsink_MessageFile& FileInfo = dbgMsgBuf->dbgMsgParm.v.MessageFileP;
        hr = pIliveCacheSink->OpenFile(&FileInfo.msg[0], FileInfo.i4, FileInfo.doAppend, &FileHandle, &FileInfo.e);
        FileInfo.i4 = FileHandle; 
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(FileInfo.i4)                    +
          sizeof(FileInfo.e);
        break;
      }
    case dbsink_SyncFile_eo39 :
      {
        char* pFileName;
        teo39_dbsink_MessageFile& FileInfo = dbgMsgBuf->dbgMsgParm.v.MessageFileP;
        if (0 == strlen(dbgMsgBuf->dbgMsgParm.v.MessageFileP.msg)) {
          pFileName = &dbgMsgBuf->dbgMsgParm.v.MessageFileP.msg[0];
        }
        else {
          pFileName = NULL;
        }
        hr = pIliveCacheSink->SyncFile(pFileName, FileInfo.i4, &berror);
        dbgMsgBuf->dbgMsgParm.v.BasisErrorP = berror; 
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.BasisErrorP);
        break;
      }
    case dbsink_WriteFile_eo39 :
      {
        teo39_dbsink_MessageFile& FileInfo = dbgMsgBuf->dbgMsgParm.v.MessageFileP;
        hr = pIliveCacheSink->WriteFile(FileInfo.i4, &FileInfo.msg[0], &berror);
        dbgMsgBuf->dbgMsgParm.v.BasisErrorP = berror; 
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.BasisErrorP);
        break;
      }
    case dbsink_GetDateTime_eo39 :
      {
        hr = pIliveCacheSink->GetDateTime (&dbgMsgBuf->dbgMsgParm.v.GetDateTime_rplyP.date,
          &dbgMsgBuf->dbgMsgParm.v.GetDateTime_rplyP.time);
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.GetDateTime_rplyP);
        break;
      }
    case dbsink_GetClock_eo39 :
      {
        hr = pIliveCacheSink->GetClock (&dbgMsgBuf->dbgMsgParm.v.GetClock_rplyP.sec,
          &dbgMsgBuf->dbgMsgParm.v.GetClock_rplyP.msec);
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.GetClock_rplyP);
        break;
      }
    case dbsink_DropVersion_eo39 :
      {
        hr = pIliveCacheSink->DropVersion (REINTERPRET_CAST(unsigned char*, &dbgMsgBuf->dbgMsgParm.v.DropVersionP.versionContext), &berror);
        dbgMsgBuf->dbgMsgParm.v.BasisErrorP = berror; 
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.BasisErrorP);
        break;
      }
    case dbsink_CreateSchema_eo39 :
      {
        tsp00_Int4 SchemaId;
        hr = pIliveCacheSink->CreateSchema (&dbgMsgBuf->dbgMsgParm.v.CreateSchemaP.schemaName, &SchemaId, &berror);
        dbgMsgBuf->dbgMsgParm.v.CreateSchema_rplyP.e        = berror;
        dbgMsgBuf->dbgMsgParm.v.CreateSchema_rplyP.schemaId = SchemaId;
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.CreateSchema_rplyP);
        break;
      }
    case dbsink_ExistsSchema_eo39 :
      {
        tsp00_Int4 SchemaId;
        hr = pIliveCacheSink->ExistsSchema (&dbgMsgBuf->dbgMsgParm.v.CreateSchemaP.schemaName, &SchemaId, &berror);
        dbgMsgBuf->dbgMsgParm.v.CreateSchema_rplyP.e        = berror;
        dbgMsgBuf->dbgMsgParm.v.CreateSchema_rplyP.schemaId = SchemaId;
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.CreateSchema_rplyP);
        break;
      }
    case dbsink_DropSchema_eo39 :
      {
        hr = pIliveCacheSink->DropSchema (dbgMsgBuf->dbgMsgParm.v.DropSchemaP.schemaId, &berror);
        dbgMsgBuf->dbgMsgParm.v.BasisErrorP = berror; 
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.BasisErrorP);
        break;
      }
    case dbsink_GetSchemaName_eo39 :
      {
        hr = pIliveCacheSink->GetSchemaName (dbgMsgBuf->dbgMsgParm.v.GetSchemaNameP.schemaId, 
          &dbgMsgBuf->dbgMsgParm.v.GetSchemaName_rplyP.schemaName,
          &dbgMsgBuf->dbgMsgParm.v.GetSchemaName_rplyP.e);
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.GetSchemaName_rplyP);
        break;
      }
    case dbsink_GetWorkingDir_eo39 :
      {
        hr = pIliveCacheSink->GetWorkingDirectory (dbgMsgBuf->dbgMsgParm.v.Path_rplyP.path,
          dbgMsgBuf->dbgMsgParm.v.Path_rplyP.e);
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.Path_rplyP);
        break;
      }
    case dbsink_GetDbrootSapDir_eo39 :
      {
        hr = pIliveCacheSink->GetDbrootSapPath (dbgMsgBuf->dbgMsgParm.v.Path_rplyP.path,
          dbgMsgBuf->dbgMsgParm.v.Path_rplyP.e);
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.Path_rplyP);
        break;
      }
    case dbsink_DropContainer_eo39 :
      {
        hr = pIliveCacheSink->DropContainer (dbgMsgBuf->dbgMsgParm.v.DropContainerP.schemaId,
          dbgMsgBuf->dbgMsgParm.v.DropContainerP.guid,
          dbgMsgBuf->dbgMsgParm.v.DropContainerP.cno,
          berror);
        dbgMsgBuf->dbgMsgParm.v.BasisErrorP = berror; 
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.BasisErrorP);
        break;
      }
    case dbsink_GetContainerInfo_eo39 :
      {
        hr = pIliveCacheSink->GetContainerInfo (dbgMsgBuf->dbgMsgParm.v.GetContainerInfoP.containerHandle,
          dbgMsgBuf->dbgMsgParm.v.GetContainerInfo_rplyP.guid,
          dbgMsgBuf->dbgMsgParm.v.GetContainerInfo_rplyP.schema,
          dbgMsgBuf->dbgMsgParm.v.GetContainerInfo_rplyP.cno,
          dbgMsgBuf->dbgMsgParm.v.GetContainerInfo_rplyP.e);
        dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
          sizeof(dbgMsgBuf->dbgMsgParm.pdbgKind) +
          sizeof(dbgMsgBuf->dbgMsgParm.v.GetContainerInfo_rplyP);
        break;
      }
    case dbsink_HistoryUsed_eo39 :
      hr = pIliveCacheSink->IsObjHistoryUsed(
        dbgMsgBuf->dbgMsgParm.v.HistoryUsedP.containerId,
        dbgMsgBuf->dbgMsgParm.v.HistoryUsedP.oid,
        dbgMsgBuf->dbgMsgParm.v.HistoryUsedP.isUsed,
        dbgMsgBuf->dbgMsgParm.v.HistoryUsedP.e);
      dbgMsgBuf->dbghdr.hdbgBufLen = sizeof(dbgMsgBuf->dbghdr) + 
        sizeof(dbgMsgBuf->dbgMsgParm.v.HistoryUsedP.e) +
        sizeof(dbgMsgBuf->dbgMsgParm.v.HistoryUsedP.isUsed);
      break;
    }
    if ( dbgMsgBuf->dbghdr.hdbgMsgKind == iSinkRmtRequest_eo39 )
      dbgMsgBuf->dbghdr.hdbgMsgKind = iSinkRmtReply_eo39;
    else
      dbgMsgBuf->dbghdr.hdbgMsgKind = iSinkReply_eo39;
    dbgMsgBuf->dbghdr.hdbgBufSize = MARSHAL_BUF_MXEO39;
    sql35_VDCOMTransfer ( (void*)transferBuf, rc_ok, &errtext );
  }
}

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/