/****************************************************************************

  module      : vos54.c

  -------------------------------------------------------------------------

  responsible : FrankH

  special area: kernel / LZU
  description : Disk I/O


  last changed: 2001-01-03  9:27
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1998



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/


#include "veo00c.h"
#include "gos00.h"
#include "geo50_0.h"
#include "geo002.h"
#include "heo01.h"
#include "heo46.h"
#include "heo52.h"
#include "heo54k.h"
#include "heo58.h"

#include "geo007_1.h"
#include "geo007_2.h"
#include "gos44.h"
#include "gos541.h"

#include <winioctl.h>


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/
#define MOD__  "VOS54.C : "
#define MF__   MOD__"UNDEFINED"

#define DEV_OVERLAPPED_IO                  0x1
#define DEV_NO_BUFFERING                   0x2


#define MAX_VDETACH_WAIT_TIME              30        // -- sleep time
#define MAX_VATTACH_WAIT_TIME              10        // -- sleep time


#define MAX_DEVIO_ERRCODE                  27
#define ERRCODE_DEVIO_DEVCREA              0
#define ERRCODE_DEVIO_MAXDEVNO             1
#define ERRCODE_DEVIO_ALRATT               2
#define ERRCODE_DEVIO_DEVICREA             3
#define ERRCODE_DEVIO_WRITE                4
#define ERRCODE_DEVIO_READ                 5
#define ERRCODE_DEVIO_SEEK                 6
#define ERRCODE_DEVIO_INVPNO               7
#define ERRCODE_DEVIO_LOCK                 8
#define ERRCODE_DEVIO_DEV_POS              9
#define ERRCODE_DEVIO_SEEK_POS             10
#define ERRCODE_DEVIO_ERROR_ON_DEVSPACE    11
#define ERRCODE_DEVIO_DEVSPACE_ON_RD_ONLY  12
#define ERRCODE_DEVIO_DEVSPACE_ON_SYS      13
#define ERRCODE_DEVIO_DEVSPACE_ON_DIR      14
#define ERRCODE_DEVIO_GET_FILE_INFO        15
#define ERRCODE_DEVIO_LOCK_DEVSPACE        16
#define ERRCODE_DEVIO_DEVSPACE_LOCKED      17
#define ERRCODE_DEVIO_OPEN_DEVSPACE        18
#define ERRCODE_DEVIO_ACCESS_DENIED        19
#define ERRCODE_DEVIO_WRONG_PATH           20
#define ERRCODE_DEVIO_WRONG_FILE_ATTRIB    21
#define ERRCODE_DEVIO_DEVSPACE_ON_COMP     22
#define ERRCODE_DEVIO_DEVICREA_TIMEOUT     23
#define ERRCODE_DEVIO_EVAL_DEVSPACE        24
#define ERRCODE_DEVIO_CREATE_COMPL_PORT    25
#define ERRCODE_DEVIO_USER_REQ_ERROR       26

#define IO_NOERROR                         0
#define IO_ERROR                           1
#define IO_CONTINUE                        3


#define MAX_DWORD_VAL                      0xFFFFFFFF
#define NUM_OF_SHARING_VIOLATION_RETRIES   3

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/
#define PAGES_TO_QWORD(_Low,_High,_Page) {                                   \
                                        _Low =(_Page)<<DEVSPACE_PAGE_SHFT;   \
                                        _High=(_Page)>>(32-DEVSPACE_PAGE_SHFT);\
                                        }

#define QWORD_TO_PAGES(_Low,_High)      ((_Low   >> DEVSPACE_PAGE_SHFT) |    \
                                        (_High  << (32-DEVSPACE_PAGE_SHFT)))

#if defined(_WIN32)
 #define MAX_DEVSPACE_PAGES              (MAX_DWORD_VAL - 1)

 #define HIGH_VALUE_VALID(_High)         (!(_High >> DEVSPACE_PAGE_SHFT)  ?   \
                                          true : false )

 #define LOW_VALUE_VALID(_Low)           (!(_Low << (32-DEVSPACE_PAGE_SHFT)) ?  \
                                          true : false )

#else
 #define MAX_DEVSPACE_PAGES              ((MAX_DWORD_VAL >>DEVSPACE_PAGE_SHFT)\
                                           - 1)
 #define HIGH_VALUE_VALID(_High)         ((_High == 0)  ? true : false )

 #define LOW_VALUE_VALID(_Low)           ((_Low <=  MAX_DEVSPACE_PAGES) ?     \
                                          true : false )
#endif


#define REQ_TO_DEVI                      __sql54k_req_to_devi
#define REQ_FROM_DEVx_TO_UKT             __sql54k_req_from_devx_to_ukt
#define REQ_FROM_DEV_WOKER_TO_UKT        __sql54k_req_from_dev_worker_to_ukt
#define REQ_TO_DEV0                      __sql54k_req_to_dev0



/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/


/*===========================================================================*
 *  GLOBAL VARIABLES                                                         *
 *===========================================================================*/


/*===========================================================================*
 *  LOCAL VARIABLES                                                          *
 *===========================================================================*/
static ERRORTEXT ErrTab[ MAX_DEVIO_ERRCODE ] =
                        {
                        ERRMSG_DEVIO_DEVCREA ,            // - ERRCODE_DEVIO_DEVCREA
                        ERRMSG_DEVIO_MAXDEVNO ,           // - ERRCODE_DEVIO_MAXDEVNO
                        ERRMSG_DEVIO_ALRATT   ,           // - ERRCODE_DEVIO_ALRATT
                        ERRMSG_DEVIO_DEVICREA ,           // - ERRCODE_DEVIO_DEVICREA
                        ERRMSG_DEVIO_WRITE    ,           // - ERRCODE_DEVIO_WRITE
                        ERRMSG_DEVIO_READ     ,           // - ERRCODE_DEVIO_READ
                        ERRMSG_DEVIO_SEEK     ,           // - ERRCODE_DEVIO_SEEK
                        ERRMSG_DEVIO_INVPNO   ,           // - ERRCODE_DEVIO_INVPNO
                        ERRMSG_DEVIO_LOCK     ,           // - ERRCODE_DEVIO_LOCK
                        ERRMSG_DEVIO_DEV_POS  ,           // - ERRCODE_DEVIO_DEV_POS
                        ERRMSG_DEVIO_SEEK_POS ,           // - ERRCODE_DEVIO_SEEK_POS
                        ERRMSG_DEVIO_ERROR_ON_DEVSPACE,   // - ERRCODE_DEVIO_ERROR_ON_DEVSPACE
                        ERRMSG_DEVIO_DEVSPACE_ON_RD_ONLY, // - ERRCODE_DEVIO_DEVSPACE_ON_RD_ONLY
                        ERRMSG_DEVIO_DEVSPACE_ON_SYS,     // - ERRCODE_DEVIO_DEVSPACE_ON_SYS
                        ERRMSG_DEVIO_DEVSPACE_ON_DIR,     // - ERRCODE_DEVIO_DEVSPACE_ON_DIR
                        ERRMSG_DEVIO_GET_FILE_INFO,       // - ERRCODE_DEVIO_GET_FILE_INFO
                        ERRMSG_DEVIO_LOCK_DEVSPACE,       // - ERRCODE_DEVIO_LOCK_DEVSPACE
                        ERRMSG_DEVIO_DEVSPACE_LOCKED,     // - ERRCODE_DEVIO_DEVSPACE_LOCKED
                        ERRMSG_DEVIO_OPEN_DEVSPACE,       // - ERRCODE_DEVIO_OPEN_DEVSPACE
                        ERRMSG_DEVIO_ACCESS_DENIED,       // - ERRCODE_DEVIO_ACCESS_DENIED
                        ERRMSG_DEVIO_WRONG_PATH,          // - ERRCODE_DEVIO_WRONG_PATH
                        ERRMSG_DEVIO_WRONG_FILE_ATTRIB,   // - ERRCODE_DEVIO_WRONG_FILE_ATTRIB
                        ERRMSG_DEVIO_DEVSPACE_COMPRESSED, // - ERRCODE_DEVIO_DEVSPACE_ON_COMP
                        ERRMSG_DEVIO_DEVICREA_TIMEOUT,    // - ERRCODE_DEVIO_DEVICREA_TIMEOUT
                        ERRMSG_DEVIO_EVAL_DEVSPACE,       // - ERRCODE_DEVIO_EVAL_DEVSPACE
                        ERRMSG_DEVIO_CREATE_COMPL_PORT,   // - ERRCODE_DEVIO_CREATE_COMPL_PORT
                        ERRMSG_DEVIO_USER_REQ_ERROR       // - ERRCODE_DEVIO_USER_REQ_ERROR    
                        };


/*===========================================================================*
 *  LOCAL FUNCTIONS (PROTOTYPES)                                             *
 *===========================================================================*/
APIRET   sql54k_fill_block_0     ( HFILE                     hfDevspace,
                                   PSZ                       pszLogDevName,
                                   ULONG                     ulDevPages,
                                   ULONG                     ulMaxDevPages,
                                   ERRORTEXT                 errtext);

APIRET  sql54k_format_devspace   ( HFILE                     hfDevspace,
                                   PSZ                       pszLogDevName,
                                   ULONG                     ulCurrFileSizeLow,
                                   ULONG                     ulCurrFileSizeHigh,
                                   ULONG                     ulDevPages,
                                   PTASK_CTRL_REC            pTaskCtrl,
                                   ERRORTEXT                 errtext);

LONG     sql54k_evaluate_block_0 ( HFILE                     hfDevspace,
                                   PULONG                    pulDevPages,
                                   PULONG                    pulMaxDevPages,
                                   ULONG                     ulDBPageSize,
                                   BOOL                      fDevspaceIsDrive,
                                   PSZ                       pszLogDevName );

INT      sql54k_vdevsize         ( PSZ                       pszPhysDevName,
                                   PSZ                       pszLogDevName,
                                   INT4                      *devcapacity,
                                   ULONG                     ulCurrFileSizeLow,
                                   ULONG                     ulCurrFileSizeHigh,
                                   ERRORTEXT                 errtext,
                                   BOOLEAN                   *ok );

INT      sql54k_vdevsize_drive   ( PSZ                       pszDriveName,
                                   INT4                      *devcapacity,
                                   ERRORTEXT                 errtext,
                                   BOOLEAN                   *ok );

LONG     sql54k_dev_open         ( PSZ                       pszLogDevName,
                                   LONG                      lDevspaceNo,
                                   ULONG                     ulOptions,
                                   PHFILE                    phfDevspace,
                                   ERRORTEXT                 pcErrText,
                                   PULONG                    pulErrCode,
                                   LONG                     *lDevspaceType );

LONG     sql54k_dev_close        ( HFILE                     hfDevspace,
                                   char                     *pszDriveName,
                                   LONG                      lDevspaceType,
                                   ERRORTEXT                 pcErrText);



VOID     sql54k_dev0_vattach     ( DLQ_VATTACH_REQ_ARGS_REC  *pVattachArgs );

VOID     sql54k_devi_kill        ( LONG                      lDevspaceNo );

PDLQ_REC sql54k_dev_receive      ( PPIO_QUEUE_REC            ppIOWorkQue,
                                   PLONG                     plIOQueLen,
                                   PLONG                     plIOWorkExclusive,
                                   PTHREAD_CTRL_HEADER_REC   pThrdCtrlHeader );

VOID     sql54k_dev0_vdetach     ( PDLQ_REC                  pRequest );

LONG     sql54k_vmark_bad_dev    ( LONG                      lDevspaceNo,
                                   tsp_page_no               tspBadDevspacePage );

VOID _System sql54k_devi         ( PDEVI_CTRL_REC            pDeviCtrl );

LONG     sql54k_set_devspace_lock( HFILE                     hfFileHandle,
                                   PSZ                       pszLogDevName,
                                   BOOL                      fUnlock );

VOID     sql54k_devi_vdetach     ( PDEVI_CTRL_REC            pDeviCtrl,
                                   PDEV_IO_CTRL_REC          pIOCtrl );

BOOL     sql54k_find_devspace_no ( PLONG                     plDevspaceNo,
                                   PSZ                       pszLogDevName,
                                   PULONG                    pulErrCode );

PDEVI_CTRL_REC sql54k_find_devi  ( LONG                      lDevspaceNo,
                                   PULONG                    pulQueueLength );

LONG     sql54k_single_blockio   ( PUKT_CTRL_REC             pUKT,
                                   LONG                      lDevspaceNo,
                                   tsp_page_no               lDevspacePage,
                                   tsp2_io_op_code           OpCode,
                                   BUFFERADDRESS             pBuffAddr,
                                   SHORT                     sNumOfPages,
                                   ERRORTEXT                 pcErrText );

VOID     sql54k_devi_vblockio    ( PDLQ_REC                  pRequest,
                                   PDEV_IO_CTRL_REC          pIOCtrl,
                                   PDEVI_CTRL_REC            pDeviCtrl );

LONG     sql54k_devi_write       ( PDEVI_CTRL_REC             pDeviCtrl,
                                   DLQ_VBLOCKIO_REQ_ARGS_REC  *pVblockIOArgs );

LONG     sql54k_devi_read        ( PDEVI_CTRL_REC             pDeviCtrl,
                                   DLQ_VBLOCKIO_REQ_ARGS_REC  *pVblockIOArgs );

VOID     sql54k_wake_up_tasks    ( PPIO_QUEUE_REC             ppIOWorkQue,
                                   PLONG                      plIOWorkExclusive );

LONG     sql54k_lock_drive       ( HFILE                      hfFileHandle );

LONG     sql54k_unlock_drive     ( HFILE                      hfFileHandle );

LONG     sql54k_overlapped_io    ( PUKT_CTRL_REC              pUKT,
                                   LONG                       lDevspaceNo,
                                   tsp_page_no                lDevspacePage,
                                   tsp2_io_op_code            OpCode,
                                   BUFFERADDRESS              pBuffAddr,
                                   SHORT                      sNumOfPages,
                                   ERRORTEXT                  pcErrText );

LONG sql54k_uncompress		     ( HANDLE					  hfFileHandle );


// --- inline functions
_INLINE BOOL __sql54k_is_user_req_io_error       ( LONG            lDevspaceNo,
                                                   tsp2_io_op_code OpCode );

_INLINE VOID __sql54k_req_to_devi                ( PDEVI_CTRL_REC  pDevCtrl,
                                                   PDLQ_REC        pRequest );

_INLINE VOID __sql54k_req_from_devx_to_ukt       ( PDLQ_REC        pRequest );

_INLINE VOID __sql54k_req_to_dev0                ( PDLQ_REC        pRequest,
                                                   PTASK_CTRL_REC  pTaskCtrl );

_INLINE VOID __sql54k_req_from_dev_worker_to_ukt ( BOOL         fOk,
                                                   ULONG        ulNumOfBytes,
                                                   LONG         lDevspaceNo,
                                                   LPOVERLAPPED pCompletedOverlapped,
                                                   LONG         rc );

_INLINE HANDLE     __sql54k_create_file         ( LPCTSTR                 lpFileName,
                                                  DWORD                   dwDesiredAccess,
                                                  DWORD                   dwShareMode,
                                                  LPSECURITY_ATTRIBUTES   lpSecurityAttributes,
                                                  DWORD                   dwCreationDisposition,
                                                  DWORD                   dwFlagsAndAttributes,
                                                  HANDLE                  hTemplateFile );
#ifdef DEVELOP_CHECKS
 _INLINE LONG __sql54k_check_filepos       ( HFILE            hfFileHandle,
                                             ULONG            lDevspaceNo,
                                             tsp_page_no      lDevspacePage,
                                             SHORT            sNumOfPages,
                                             ULONG            ulDevPages,
                                             PULONG           pulErrCode,
                                             ERRORTEXT        pcErrText );
#endif


/*===========================================================================*
 *  GLOBAL FUNCTIONS (CODE)                                                  *
 *===========================================================================*/

VOID vcurrdevsize ( tsp_int4      devno,
                    tsp_int4      *devcapacity,
                    tsp_errtext   errtext,
                    BOOLEAN       *ok )
  {
  #undef  MF__
  #define MF__  MOD__"vcurrdevsize"

  ULONG                ulCurrFileSizeLow;
  ULONG                ulCurrFileSizeHigh;
  PSZ                  pszLogDevName;
  PATHNAME             szPhysDevName;
  LONG                 lAtt;
  APIRET               rc                 = NO_ERROR;

  DBGIN;

  *ok = false ;
  pszLogDevName = kgs.Dev0.pIOCtrl[devno].szDevspace;
  sql44c_subst_log_parts ( szPhysDevName, pszLogDevName ) ;

  rc = sql44c_get_file_info ( szPhysDevName, &lAtt,
                              &ulCurrFileSizeLow, &ulCurrFileSizeHigh );

  if ( rc == ERROR_ACCESS_DENIED )
    {
    MSGALL(( ERR_DEVSPACE_ACCESS_DENIED, pszLogDevName ));
    DBG1  (( MF__, "Access on devspace '%s' denied", pszLogDevName ));
    sql46c_build_error_string ( errtext, ERRMSG_DEVIO_ACCESS_DENIED, 0 );
    return;
    }
  else if (rc == ERROR_PATH_NOT_FOUND || rc == ERROR_FILE_NOT_FOUND )
    {
    sql46c_build_error_string (errtext, ERRMSG_DEVIO_WRONG_PATH, 0 );
    MSGALL(( ERR_WRONG_PATH, pszLogDevName ));
    DBG1  (( MF__, "Wrong path '%s'", pszLogDevName ));
    return;
    }
  else if ( rc != NO_ERROR )
    {
    sql46c_build_error_string (errtext, ERRMSG_DEVIO_GET_FILE_INFO, rc );
    MSGALL(( ERR_GET_FILE_INFO, pszLogDevName, rc ));
    DBG1  (( MF__, "Query file info error on '%s', rc = %ld",
                    pszLogDevName, rc ));
    return;
    }

  *ok = true;
  *devcapacity = QWORD_TO_PAGES(ulCurrFileSizeLow, ulCurrFileSizeHigh) - 1;

  DBGOUT;
  }

/*------------------------------*/

VOID vdevsize ( tsp_int4       type_spec_devno,
                tsp2_dev_type_Param devtype,
                tsp_int4       *devcapacity,
                tsp_errtext    errtext,
                BOOLEAN        *ok )
  {
  #undef  MF__
  #define MF__  MOD__"vdevsize"

  PATHNAME             szPhysDevName;
  tsp2_devnamec        szLogDevName;
  LONG                 lAtt;
  ULONG                ulCurrFileSizeLow  = 0;
  ULONG                ulCurrFileSizeHigh = 0;
  APIRET               rc                 = NO_ERROR;
  DEV_DESCR            *devsp;
  LONGLONG             PSE36RamdiskSizeAvail ;
  LONGLONG             PSE36RamdiskSizeWanted ;

  DBGIN;

  devsp = sql58k_get_devsp_descr ( (tsp2_dev_type)devtype, --type_spec_devno );

  *ok = (devsp != NULL);

  if ( *ok )
    {
    *ok = true;
    strcpy ( szLogDevName , devsp->szDevspaceName ) ;
    // --- substitute logical path name parts
    sql44c_subst_log_parts ( szPhysDevName, szLogDevName );
    }
  else
    MSGALL(( ERR_INVALID_DEVSP_DESCR,  (tsp2_dev_type)devtype, type_spec_devno+1 )) ;

  if ( *ok )
    {
    *ok = false;

    switch ( sql44c_get_dev_type_by_filename ( szPhysDevName ) )
      {
      case DT_DRIVE:
        sql54k_vdevsize_drive (szPhysDevName, devcapacity, errtext, ok);
        break;

      case DT_PSE36:
        PSE36RamdiskSizeWanted = (LONGLONG )(*devcapacity * XPARAM(ulDBPageSize)) ;
        *ok = os541PSE36GetSize( szPhysDevName, &PSE36RamdiskSizeAvail ) == NO_ERROR 
              && PSE36RamdiskSizeWanted <= PSE36RamdiskSizeAvail ;
        break;

      default :
        DBG3 ((MF__, "devspace: '%s'- devcapacity: %d ",
                     szPhysDevName, *devcapacity));

        rc = sql44c_get_file_info ( szPhysDevName, &lAtt,
                                    &ulCurrFileSizeLow, &ulCurrFileSizeHigh );

        if ( rc == ERROR_ACCESS_DENIED )
          {
          MSGALL(( ERR_DEVSPACE_ACCESS_DENIED, szLogDevName ));
          DBG1  (( MF__, "Access on devspace '%s' denied", szLogDevName ));
          sql46c_build_error_string ( errtext, ERRMSG_DEVIO_ACCESS_DENIED, 0 );
          return;
          }
        else if (rc == ERROR_PATH_NOT_FOUND)
          {
          sql46c_build_error_string (errtext, ERRMSG_DEVIO_WRONG_PATH, 0 );
          MSGALL(( ERR_WRONG_PATH, szLogDevName ));
          DBG1  (( MF__, "Wrong path '%s'", szLogDevName ));
          return;
          }
        else if ( rc == ERROR_FILE_NOT_FOUND )
          {
          rc = NO_ERROR;
          }
        else if ( rc != NO_ERROR )
          {
          sql46c_build_error_string (errtext, ERRMSG_DEVIO_GET_FILE_INFO, rc );
          MSGALL(( ERR_GET_FILE_INFO, szLogDevName, rc ));
          DBG1  (( MF__, "Query file info error on '%s', rc = %ld",
                          szLogDevName, rc ));
          return;
          }

        if ( lAtt == -1 )
          {
          MSGALL(( ERR_UNKNOWN_FILE_ATTRIBUTE, szLogDevName ));
          DBG1  (( MF__, "'%s' has a unknown file attribute", szLogDevName  ));
          sql46c_build_error_string ( errtext,
                                      ERRMSG_DEVIO_WRONG_FILE_ATTRIB, 0 );
          }
        else if ((lAtt & FILE_DIRECTORY) == FILE_DIRECTORY)
          {
          MSGALL(( ERR_IO_ON_DIR, szLogDevName ));
          DBG1  (( MF__, "No Devspaces on directory '%s' !", szLogDevName ));
          sql46c_build_error_string ( errtext,
                                      ERRMSG_DEVIO_DEVSPACE_ON_DIR, 0 );
          }
        else if ((lAtt & FILE_SYSTEM) == FILE_SYSTEM)
          {
          MSGALL(( ERR_IO_ON_SYS, szLogDevName ));
          DBG1  (( MF__, "No Devspaces on system file '%s' !", szLogDevName ));
          sql46c_build_error_string ( errtext,
                                      ERRMSG_DEVIO_DEVSPACE_ON_SYS, 0 );
          }
        else if ((lAtt & FILE_READONLY) == FILE_READONLY)
          {
          MSGALL(( ERR_DEVSPACE_ON_RD_ONLY, szLogDevName ));
          DBG1  (( MF__, "No Devspaces on read-only file '%s' !",
                          szLogDevName  ));
          sql46c_build_error_string ( errtext,
                                      ERRMSG_DEVIO_DEVSPACE_ON_RD_ONLY, 0 );
          }
        else if (((lAtt & FILE_COMPRESSED)       == FILE_COMPRESSED) &&
                 ( XPARAM(fAllowCompressedDevsp) == false ) &&
		  	   ( ulCurrFileSizeLow || ulCurrFileSizeHigh ))
          {
          MSGALL(( ERR_DEVSPACE_COMPRESSED, szLogDevName ));
          DBG1  (( MF__, "No Devspaces on compressed file '%s' !",
                          szLogDevName  ));
          sql46c_build_error_string ( errtext,
                                      ERRMSG_DEVIO_DEVSPACE_COMPRESSED, 0 );
          }
         else
          {
          sql54k_vdevsize ( szPhysDevName, szLogDevName, devcapacity,
                            ulCurrFileSizeLow, ulCurrFileSizeHigh, errtext, ok );
          }
      }

    DBG3 (( MF__, "devcapacity: %d", *devcapacity ));
    }
  DBGOUT;
  }

/*------------------------------*/

VOID vattach ( tsp_int4       type_spec_devno,
               tsp2_dev_type_Param devtype,
               tsp_int4       *devno,
               PROCESS_ID     pid,
               tsp_errtext    errtext,
               BOOLEAN        *ok )
  {
  #undef  MF__
  #define MF__ MOD__"vattach"
  DEV_DESCR                      *devsp ;
  tsp2_devnamec                  szLogDevName;
  PDLQ_REC                       pRequest;
  #ifdef DEVELOP_CHECKS
   PUKT_CTRL_REC                 pUKT      = THIS_UKT_CTRL;
   PTASK_CTRL_REC                pTaskCtrl = pUKT->pCTask;
  #else
   PTASK_CTRL_REC                pTaskCtrl = &kgs.pFirstTaskCtrl[pid - 1];
   PUKT_CTRL_REC                 pUKT      = pTaskCtrl->pUKT;
  #endif

  DBGIN_T (pTaskCtrl->ulTaskIndex);

  #ifdef DEVELOP_CHECKS
   if ( (ULONG)pid != pTaskCtrl->ulTaskIndex )
     {
     MSGD (( ERR_VXXXX_WRONG_TASK, "vattach", pid ));
     ABORT();
     }
  #endif

  if ( kgs.Dev0.ThrdCtrlHeader.ThrdState == KT_INITIALIZED )
    {
    MSGALL(( ERR_THREAD_XXX_NOT_ACTIVE, "DEV0" ));
    sql46c_build_error_string ( errtext, ERRMSG_DEVIO_DEV0_NOT_ACTIVE, 0 );
    *ok = false;
    return;
    }

  pTaskCtrl->TaskState = TSK_VATTACH;

  devsp = sql58k_get_devsp_descr ( (tsp2_dev_type)devtype, --type_spec_devno ) ;

  if ( devsp != NULL )
     strcpy ( szLogDevName , devsp->szDevspaceName ) ;
  else
    {
    MSGALL(( ERR_INVALID_DEVSP_DESCR,  (tsp2_dev_type)devtype, type_spec_devno+1 )) ;
    *ok = false;
    return;
    }

  MSGD (( INFO_ATTACHING_DEVSPACE, szLogDevName ));

  //
  //
  // ---  Send an attach-request.
  //
  //
  pRequest                                             = &pTaskCtrl->SpecDLQElem1;
  pRequest->pTaskCtrl                                  = pTaskCtrl;
  pRequest->ulReqType                                  = REQ_VATTACH;
  pRequest->ReqArgs.VattachRequest.ulDBPageSize        = XPARAM(ulDBPageSize);
  pRequest->ReqArgs.VattachRequest.fOk                 = false;
  pRequest->ReqArgs.VattachRequest.pszDevspaceToAttach = szLogDevName;
  pRequest->ReqArgs.VattachRequest.DevspaceType        = (tsp2_dev_type)devtype ; 

  // --- send request to DEV0 thread
  REQ_TO_DEV0( pRequest, pTaskCtrl );

  // --- goto dispatcher
  GOTO_DISP(pUKT);

  DBG3 (( MF__, "[T:0x%03u] Returned from request pUKT: %x",
          pTaskCtrl->ulTaskIndex, pUKT ));

  // --- wrong reply?
  if ( pUKT->pCurrReq != pRequest )
    {
    MSGALL(( ERR_UNKNOWN_REQUEST ));
    DBG1  (( MF__, "[T:0x%03u] pRequest: 0x%x (SEND) != pUKT->pCurrReq: 0x%x (REPLY)",
            pTaskCtrl->ulTaskIndex, pRequest, pUKT->pCurrReq ));
    ABORT();
    }

  *ok = pRequest->ReqArgs.VattachRequest.fOk;

  if ( *ok == true )
    {
    *devno = pRequest->ReqArgs.VattachRequest.lDevspaceNo;

    DBG3 (( MF__, "[T:0x%03u] Returned OK, pRequest: 0x%x",
            pTaskCtrl->ulTaskIndex, pRequest ));
    }
  else
    {
    //
    // --- replied via UKT queue
    //
    if ( pRequest->ReqArgs.VattachRequest.ulErrCode < MAX_DEVIO_ERRCODE )
      {
      sql46c_build_error_string ( errtext,
                                  ErrTab[ pRequest->ReqArgs.VattachRequest.ulErrCode ], 0 );

      DBG1 (( MF__, "[T:0x%03u] pRequest: 0x%x, %s",
              pTaskCtrl->ulTaskIndex, pRequest,
              ErrTab[ pRequest->ReqArgs.VattachRequest.ulErrCode ] ));
      }
    }

  pTaskCtrl->TaskState = TSK_RUNNING;

  DBGOUT_T (pTaskCtrl->ulTaskIndex);
  return;
  }

/*------------------------------*/

INT4  vget_data_io_cnt ( VOID )
  {
  #undef  MF__
  #define MF__ MOD__"vget_data_io_cnt"

  DBGPAS;

  return ( (INT4 ) kgs.TotalDataIOCnt ) ;
  }

/*------------------------------*/

VOID  vmark_bad_dev ( INT4              devno,
                      tsp_page_no       devpno )
  {
  #undef  MF__
  #define MF__ MOD__"vmark_bad_dev"
  PDLQ_REC                       pRequest;

  PUKT_CTRL_REC   pUKT      = THIS_UKT_CTRL;
  PTASK_CTRL_REC  pTaskCtrl = pUKT->pCTask;

  DBG3 (( MF__, "[T:0x%03u] devno: %u,  devpno: %u ",
          pTaskCtrl->ulTaskIndex, devno, devpno ));

  if (devno < 0 || devno >= XPARAM(lMaxDevspaces) )
    {
    MSGALL(( ERR_INVALID_DEVNO ));
    return;
    }

  pTaskCtrl->TaskState = TSK_IO_WRITE;

  //
  // --- send a mark bad devspace request
  //
  pRequest  = &pTaskCtrl->SpecDLQElem1;
  pRequest->pTaskCtrl                                  = pTaskCtrl;
  pRequest->ulReqType                                  = REQ_MARK_BAD_DEVSPACE;
  pRequest->ReqArgs.MarkBadDevRequest.lDevspaceNo      = devno;
  pRequest->ReqArgs.MarkBadDevRequest.lBadDevspacePage = devpno;

  // --- send request to DEV0 thread
  REQ_TO_DEV0( pRequest, pTaskCtrl );

  // --- goto dispatcher
  GOTO_DISP(pUKT);


  if ( pUKT->pCurrReq != pRequest )
    {
    MSGALL(( ERR_UNKNOWN_REQUEST ));
    ABORT();
    }


  DBG3 (( MF__, "[T:0x%03u] Returned from dev0_vmark_bad_dev, devpno: %u",
                pTaskCtrl->ulTaskIndex, devpno));

  pTaskCtrl->TaskState = TSK_RUNNING;

  DBGOUT_T (pTaskCtrl->ulTaskIndex);
  return;
  }

/*------------------------------*/

_INLINE VOID __os54VblockioFinish ( PTASK_CTRL_REC pTaskCtrl )
  {
  #undef  MF__
  #define MF__ MOD__"__os54VblockioFinish"

  pTaskCtrl->lWaitForDevno    = UNDEF ;
  pTaskCtrl->DevNameWaitingFor= NULL;
  pTaskCtrl->TaskState        = TSK_RUNNING; 

  return ;
  }

/*------------------------------*/

VOID  vblockio ( INT4              devno,
                 tsp_page_no       devpno,
                 tsp2_io_op_code_Param op_code,
                 BUFFERADDRESS     bufadr,
                 INT2              buf_count,
                 PROCESS_ID        pid,
                 ERRORTEXT         errtext,
                 BOOLEAN           * ok )
  {
  #undef  MF__
  #define MF__ MOD__"vblockio"
  PDLQ_REC                       pRequest;
  PDEVI_CTRL_REC                 pDeviCtrl;
  ULONG                          ulDummy;
  LONG                           lIORet,lRetries=0 ;
  #ifdef DEVELOP_CHECKS
   PUKT_CTRL_REC                 pUKT      = THIS_UKT_CTRL;
   PTASK_CTRL_REC                pTaskCtrl = pUKT->pCTask;
  #else
   PTASK_CTRL_REC                pTaskCtrl = &kgs.pFirstTaskCtrl[pid - 1];
   PUKT_CTRL_REC                 pUKT      = pTaskCtrl->pUKT;
  #endif
  BOOLEAN                        fIsReading= op_code == DEV_READ ;
  teo52_TaskDevIOStatistic      *pIOStat;
  BOOL                           fSingleBlockIO = false;
  PDEV_IO_CTRL_REC               pIOCtrl;

  DBGIN_T (pTaskCtrl->ulTaskIndex);

  #ifdef DEVELOP_CHECKS
   if ( (ULONG)pid != pTaskCtrl->ulTaskIndex )
     {
     MSGD (( ERR_VXXXX_WRONG_TASK, "vblockio", pid ));
     ABORT();
     }
  #endif

  #ifdef TEST_BAD_DEV
   {
   static int IOCnt = 1 ;
   if ( ++IOCnt == 20 )
     {
     vmark_bad_dev ( devno, devpno ) ;
     *ok = false ;
     *errtext = '\0';
     return ;
     }
   }
  #endif

  DBG3 (( MF__, "[T:0x%03u] devno: %u,  devpno: %u  op_code: %u",
          pTaskCtrl->ulTaskIndex, devno, devpno, op_code ));

  if (devno < 0 || devno >= XPARAM(lMaxDevspaces) )
    {
    MSGALL(( ERR_INVALID_DEVNO ));
    sql46c_build_error_string ( errtext, ERRMSG_DEVIO_INVALID_DEVNO, 0 );
    *ok = false;
    return;
    }

  // --- devspace not attached correctly?
  pIOCtrl = &kgs.Dev0.pIOCtrl[devno] ;

  if ( pIOCtrl->ulAttached      == 0 ||
       pIOCtrl->ulDetachPending != 0 )
    {
    MSGD (( ERR_DEVSPACE_NOT_ATT ));
    sql46c_build_error_string ( errtext, ERRMSG_DEVIO_NOT_ATT, 0 );
    *ok = false;

    pTaskCtrl->TaskState = TSK_RUNNING;
    DBGOUT_T (pTaskCtrl->ulTaskIndex);
    return;
    }

  if ( pTaskCtrl->TaskType != TT_DW &&
       ( pIOCtrl->DevspaceType == sp2dt_data || 
         pIOCtrl->DevspaceType == sp2dt_mirr_data )
     )
       if ( ++(kgs.TotalDataIOCnt) >= 0x7fffffff )
          kgs.TotalDataIOCnt = 0 ;

  pTaskCtrl->TaskState = fIsReading ? TSK_IO_READ : TSK_IO_WRITE ;

  #ifdef DEVELOPMENT
   pTaskCtrl->ulWaitTime = kgs.ulCurrTime;
  #endif

 
  // ---  simulate an I/O error for test porposes ?
  if ( __sql54k_is_user_req_io_error((long )devno, (tsp2_io_op_code)op_code) )
    {
    sql46c_build_error_string ( errtext, ERRMSG_DEVIO_USER_REQ_ERROR, 0 );
    *ok = false;
    MSGD ((ERR_VBLOCKIO_TEST_IO_ERROR,
           fIsReading ? "reading from" : "writing to", devno, devpno ));

    pTaskCtrl->TaskState = TSK_RUNNING;
    DBGOUT_T (pTaskCtrl->ulTaskIndex);
    return;
    }

 
  pTaskCtrl->lWaitForDevno   = devno ;
  pTaskCtrl->DevNameWaitingFor= pIOCtrl->szDevspace ;

  if ( pIOCtrl->ulDevType == DT_PSE36 ||
       ( pUKT->ulActiveTasks == 1 ) && ( XPARAM(fAllowSingleIO) ))
    {
    teo52_IOStatisticRec  *pSIOStat;
    if ( fIsReading )
      pSIOStat = &pTaskCtrl->TaskStateStat.SelfIOReadStat;
    else
      pSIOStat = &pTaskCtrl->TaskStateStat.SelfIOWriteStat;

    // --- use single block io
    pSIOStat->ulCount++;
    pSIOStat->ulPageCount += buf_count;
    EO52_TAKE_ABS_TIME( (*pSIOStat),
               lIORet = sql54k_single_blockio ( pUKT, devno, devpno, (tsp2_io_op_code)op_code,
                                                bufadr, buf_count, errtext ));

    if ( lIORet != IO_CONTINUE )
      {
      if ( lIORet == IO_NOERROR )
        *ok = true;
      else
        *ok = false;

      // --- increase io counter
      __os54VblockioFinish(pTaskCtrl);
      return;
      }
    //
    // --- could not use single io. continue with DEVi io operation
    //
    DBG3 (( MF__, "[T:0x%03u] Continuing with DEVi io operation",
            pUKT->pCTask->ulTaskIndex ));
    }
  else if ( kgs.ulNumOfDevWorker != 0 )
    {
    if ( fIsReading )
      pIOStat = &pTaskCtrl->TaskStateStat.DevIOReadStat;
    else
      pIOStat = &pTaskCtrl->TaskStateStat.DevIOWriteStat;

    pIOStat->ulCount++;
    pIOStat->ulPageCount += buf_count ;

    pUKT->lWaitForDeviIO++;

    do
    {
    // --- goto dispatcher
      EO52_TAKE_REL_ABS_TIME( pTaskCtrl, (*pIOStat),
                 lIORet = sql54k_overlapped_io( pUKT, devno, devpno, (tsp2_io_op_code)op_code,
                                                bufadr, buf_count, errtext ));
    }
    while (lIORet != IO_CONTINUE && lIORet != IO_NOERROR && ++lRetries < NUMBER_OF_BLOCKIO_RETRIES ); /* PTS 1106176 */

    pUKT->lWaitForDeviIO--;

    if ( lIORet != IO_CONTINUE )
      {
      if ( lIORet == IO_NOERROR )
        *ok = true;
      else
        *ok = false;

      // --- increase io counter
      __os54VblockioFinish(pTaskCtrl);
      return;
      }
    }


  //
  // --- send a vblockio request
  //
  pDeviCtrl = sql54k_find_devi ( devno, &ulDummy );
  pRequest  = &pTaskCtrl->SpecDLQElem1;

  pRequest->pTaskCtrl                              = pTaskCtrl;
  pRequest->ulReqType                              = REQ_VBLOCKIO;
  pRequest->ReqArgs.VblockIORequest.lDevspaceNo    = devno;
  pRequest->ReqArgs.VblockIORequest.lDevspacePage  = devpno;
  pRequest->ReqArgs.VblockIORequest.OpCode         = (tsp2_io_op_code)op_code;
  pRequest->ReqArgs.VblockIORequest.pBuffAddr      = bufadr;
  pRequest->ReqArgs.VblockIORequest.sNumOfPages    = buf_count;
  pRequest->ReqArgs.VblockIORequest.fOk            = false;

  DBG3(( MF__,
          "[T:0x%03u] Send pRequest: 0x%x, using (Devi TID:%u) pIOReqQue: 0x%x",
          pTaskCtrl->ulTaskIndex, pRequest, pDeviCtrl->ThrdCtrlHeader.Tid,
          pDeviCtrl->pIOReqQue ));

  pUKT->lWaitForDeviIO++;
  REQ_TO_DEVI( pDeviCtrl, pRequest );

  if ( fIsReading )
    pIOStat = &pTaskCtrl->TaskStateStat.DevIOReadStat;
  else
    pIOStat = &pTaskCtrl->TaskStateStat.DevIOWriteStat;

  pIOStat->ulCount++;
  pIOStat->ulPageCount += buf_count ;

  // --- goto dispatcher
  EO52_TAKE_REL_ABS_TIME( pTaskCtrl, (*pIOStat), GOTO_DISP(pUKT));

  pUKT->lWaitForDeviIO--;


  if ( pUKT->pCurrReq != pRequest )
    {
    MSGALL(( ERR_UNKNOWN_REQUEST ));
    DBG1  (( MF__, "[T:0x%03u] pRequest: 0x%x (SEND) != pUKT->pCurrReq: 0x%x (REPLY)",
            pTaskCtrl->ulTaskIndex, pRequest, pUKT->pCurrReq ));
    ABORT();
    }

  *ok = pRequest->ReqArgs.VblockIORequest.fOk;
  DBG3 (( MF__, "[T:0x%03u] Returned from devi_vblockio, devpno: %u",
          pTaskCtrl->ulTaskIndex, devpno));


  if ( !*ok )
    {
    if ( pRequest->ReqArgs.VblockIORequest.ulErrCode < MAX_DEVIO_ERRCODE )
      {
      sql46c_build_error_string ( errtext,
                                  ErrTab[ pRequest->ReqArgs.VblockIORequest.ulErrCode ], 0 );
      MSGD (( ERR_ERROR_ON_DEVSPACE, kgs.Dev0.pIOCtrl[devno].szDevspace, devno));
      DBG1 (( MF__, "[T:0x%03u] pRequest: 0x%x, %s",
              pTaskCtrl->ulTaskIndex, pRequest,
              ErrTab[ pRequest->ReqArgs.VblockIORequest.ulErrCode ] ));
      }
    }
  else
    {
    DBG3 (( MF__, "[T:0x%03u] Returned OK, pRequest: 0x%x",
            pTaskCtrl->ulTaskIndex, pRequest ));
    }

  // --- increase io counter

  #ifdef DEVELOPMENT
   pTaskCtrl->ulWaitTime = 0;
  #endif

  __os54VblockioFinish(pTaskCtrl);
  DBGOUT_T (pTaskCtrl->ulTaskIndex);

  return;
  }

/*------------------------------*/

VOID  v2info  ( tsp_int4      devno1,
                tsp_int4      devno2,
                ERRORTEXT     errtext,
                V2_RETURN     *rtc)
  {
  #undef  MF__
  #define MF__ MOD__"v2info"
  PDEVI_CTRL_REC                    pDeviCtrl1;
  PDEVI_CTRL_REC                    pDeviCtrl2;
  PTASK_CTRL_REC                    pTaskCtrl = THIS_UKT_CTRL->pCTask;

  DBGIN_T (pTaskCtrl->ulTaskIndex);

  DBG3 (( MF__, "[T:0x%03u] devno1: %d, devno2: %d",
          pTaskCtrl->ulTaskIndex, devno1, devno1 ));

  pTaskCtrl->TaskState = TSK_V2INFO;

  if ( kgs.Dev0.pIOCtrl[devno1].ulAttached == 0 ||
       kgs.Dev0.pIOCtrl[devno1].ulDetachPending ||
       kgs.Dev0.pIOCtrl[devno2].ulAttached == 0 ||
       kgs.Dev0.pIOCtrl[devno2].ulDetachPending )
    {
    *rtc = V2_NOTOK;
    sql46c_build_error_string ( errtext, ERRMSG_DEVIO_NOT_ATT, 0 );
    DBG3 (( MF__, "[T:0x%03u] Devspace(s) not attached", pTaskCtrl->ulTaskIndex ));
    }
  else
    {
    pDeviCtrl1 = kgs.pFirstDeviCtrl + devno1 * XPARAM(lDeviPerDevspace);
    pDeviCtrl2 = kgs.pFirstDeviCtrl + devno2 * XPARAM(lDeviPerDevspace);

    if ( kgs.Dev0.pIOCtrl[devno1].ulDevPages !=
         kgs.Dev0.pIOCtrl[devno2].ulDevPages )
      {
      MSGALL(( ERR_DIF_SIZE_MIR_DEVSPACES,
               kgs.Dev0.pIOCtrl[devno1].szDevspace,
               kgs.Dev0.pIOCtrl[devno2].szDevspace ));
      DBG1  (( MF__, "[T:0x%03u] Different sizes of mirrored devspaces '%s' and '%s'",
               pTaskCtrl->ulTaskIndex,
               kgs.Dev0.pIOCtrl[devno1].szDevspace,
               kgs.Dev0.pIOCtrl[devno2].szDevspace ));

      *rtc = V2_NOTOK;
      sql46c_build_error_string ( errtext, ERRMSG_DEVIO_DIF_SIZE_OF_MIR, 0 );
      }
    else
      *rtc = V2_OK;
    }

  pTaskCtrl->TaskState = TSK_RUNNING;

  DBG3 (( MF__, "[T:0x%03u] returns rtc: %d", pTaskCtrl->ulTaskIndex, *rtc ));

  DBGOUT_T (pTaskCtrl->ulTaskIndex);
  return;
  }

/*------------------------------*/

VOID  v2blckio ( INT4               devno1,
                 INT4               devno2,
                 tsp_page_no        devpno,
                 tsp2_io_op_code_Param op_code,
                 BUFFERADDRESS      bufadr,
                 INT2               buf_count,
                 PROCESS_ID         pid,
                 V2_DEVNO           * read_dev,
                 V2_DEVNO           * error_dev,
                 ERRORTEXT          errtext1,
                 ERRORTEXT          errtext2,
                 BOOLEAN            * ok )
  {
  #undef  MF__
  #define MF__ MOD__"v2blckio"
  PDLQ_REC                       pRequestA;
  PDLQ_REC                       pRequestB;
  PDEVI_CTRL_REC                 pDeviCtrl1;
  PDEVI_CTRL_REC                 pDeviCtrl2;
  ULONG                          ulQueueLength1;
  ULONG                          ulQueueLength2;
  ULONG                          ulIOCounts;
  #ifdef DEVELOP_CHECKS
   PUKT_CTRL_REC                 pUKT       = THIS_UKT_CTRL;
   PTASK_CTRL_REC                pTaskCtrl  = pUKT->pCTask;
  #else
   PTASK_CTRL_REC                pTaskCtrl  = &kgs.pFirstTaskCtrl[pid - 1];
   PUKT_CTRL_REC                 pUKT       = pTaskCtrl->pUKT;
  #endif
  BOOLEAN                        fIsReading = op_code == DEV_READ ;
  teo52_TaskDevIOStatistic       *pIOStat;

  DBGIN_T (pTaskCtrl->ulTaskIndex);

  #ifdef DEVELOP_CHECKS
   if ( (ULONG)pid != pTaskCtrl->ulTaskIndex )
     {
     MSGD (( ERR_VXXXX_WRONG_TASK, "vb2lckio", pid ));
     ABORT();
     }
  #endif

  DBG3 (( MF__, "[T:0x%03u] devno1: %u,  devno2: %u,  devpno: %u  op_code: %u",
          pTaskCtrl->ulTaskIndex, devno1, devno2, devpno, (tsp2_io_op_code)op_code ));

  if ( devno1 < 0 || devno1 >= XPARAM(lMaxDevspaces) )
    {
    MSGALL(( ERR_INVALID_DEVNO ));
    sql46c_build_error_string ( errtext1, ERRMSG_DEVIO_INVALID_DEVNO, 0 );
    *ok = false;
    return;
    }
  if ( devno2 < 0 || devno2 >= XPARAM(lMaxDevspaces)  )
    {
    MSGALL(( ERR_INVALID_DEVNO ));
    sql46c_build_error_string ( errtext2, ERRMSG_DEVIO_INVALID_DEVNO, 0 );
    *ok = false;
    return;
    }


  // --- devspace not attached correctly?
  if ( kgs.Dev0.pIOCtrl[devno1].ulAttached == 0 ||
       kgs.Dev0.pIOCtrl[devno1].ulDetachPending )
    {
    sql46c_build_error_string ( errtext1, ERRMSG_DEVIO_NOT_ATT, 0 );
    MSGD (( ERR_DEVSPACE_NOT_ATT ));
    *ok = false;
    return;
    }

  if ( kgs.Dev0.pIOCtrl[devno2].ulAttached == 0 ||
       kgs.Dev0.pIOCtrl[devno2].ulDetachPending )
    {
    sql46c_build_error_string ( errtext2, ERRMSG_DEVIO_NOT_ATT, 0 );
    MSGD (( ERR_DEVSPACE_NOT_ATT ));
    *ok = false;
    return;
    }

  #ifdef DEVELOPMENT
   pTaskCtrl->ulWaitTime = kgs.ulCurrTime;
  #endif

  pDeviCtrl1 = sql54k_find_devi ( devno1, &ulQueueLength1 );
  pDeviCtrl2 = sql54k_find_devi ( devno2, &ulQueueLength2 );

  pTaskCtrl->lWaitForDevno    = devno1 ;
  pTaskCtrl->DevNameWaitingFor= kgs.Dev0.pIOCtrl[devno1].szDevspace ;

  //
  // --- send a vblockio request
  //
  pRequestA = &(pTaskCtrl->SpecDLQElem1);

  pRequestA->pTaskCtrl                              = pTaskCtrl;
  pRequestA->ulReqType                              = REQ_VBLOCKIO;
  pRequestA->ReqArgs.VblockIORequest.lDevspacePage  = devpno;
  pRequestA->ReqArgs.VblockIORequest.OpCode         = (tsp2_io_op_code)op_code;
  pRequestA->ReqArgs.VblockIORequest.pBuffAddr      = bufadr;
  pRequestA->ReqArgs.VblockIORequest.sNumOfPages    = buf_count;
  pRequestA->ReqArgs.VblockIORequest.fOk            = false;

  if ( fIsReading )
    {
    pTaskCtrl->TaskState = TSK_IO2_READ ;
    ulIOCounts = 1;

    //
    // --- determine the read devspace
    //
    if (ulQueueLength1 == ulQueueLength2)
      *read_dev = (pDeviCtrl1->DevIOStat.DevRead.ulCount < pDeviCtrl2->DevIOStat.DevRead.ulCount) ? V2_DEV1 : V2_DEV2;
    else
      *read_dev = (ulQueueLength1 < ulQueueLength2) ? V2_DEV1 : V2_DEV2;

    if (*read_dev == V2_DEV1)
      {
      DBG3(( MF__,
             "[T:0x%03u] Send pRequestA: 0x%x, using (Devi TID:%u) pIOReqQue: 0x%x",
             pTaskCtrl->ulTaskIndex, pRequestA, pDeviCtrl1->ThrdCtrlHeader.Tid,
             pDeviCtrl1->pIOReqQue ));

      pRequestA->ReqArgs.VblockIORequest.lDevspaceNo = devno1;
      REQ_TO_DEVI( pDeviCtrl1, pRequestA  );
      }
    else
      {
      DBG3(( MF__,
             "[T:0x%03u] Send pRequestA: 0x%x, using (Devi TID:%u) pIOReqQue: 0x%x",
             pTaskCtrl->ulTaskIndex, pRequestA, pDeviCtrl2->ThrdCtrlHeader.Tid,
             pDeviCtrl2->pIOReqQue ));

      pRequestA->ReqArgs.VblockIORequest.lDevspaceNo = devno2;
      REQ_TO_DEVI( pDeviCtrl2, pRequestA  );
      }
    }
  else
    {
    pTaskCtrl->TaskState = TSK_IO2_WRITE ;
    ulIOCounts = 2;

    pRequestB  = &(pTaskCtrl->SpecDLQElem2);

    pRequestB->pTaskCtrl                              = pTaskCtrl;
    pRequestB->ulReqType                              = REQ_VBLOCKIO;
    pRequestB->ReqArgs.VblockIORequest.lDevspacePage  = devpno;
    pRequestB->ReqArgs.VblockIORequest.OpCode         = (tsp2_io_op_code)op_code;
    pRequestB->ReqArgs.VblockIORequest.pBuffAddr      = bufadr;
    pRequestB->ReqArgs.VblockIORequest.sNumOfPages    = buf_count;
    pRequestB->ReqArgs.VblockIORequest.fOk            = false;

    DBG3(( MF__,
           "[T:0x%03u] Send pRequestA: 0x%x, using (Devi TID:%u) pIOReqQue: 0x%x",
           pTaskCtrl->ulTaskIndex, pRequestA, pDeviCtrl1->ThrdCtrlHeader.Tid,
           pDeviCtrl1->pIOReqQue ));

    pRequestA->ReqArgs.VblockIORequest.lDevspaceNo = devno1;
    REQ_TO_DEVI( pDeviCtrl1, pRequestA );

    DBG3(( MF__,
           "[T:0x%03u] Send pRequestB: 0x%x, using (Devi TID:%u) pIOReqQue: 0x%x",
           pTaskCtrl->ulTaskIndex, pRequestB, pDeviCtrl2->ThrdCtrlHeader.Tid,
           pDeviCtrl2->pIOReqQue ));

    pRequestB->ReqArgs.VblockIORequest.lDevspaceNo = devno2;
    REQ_TO_DEVI( pDeviCtrl2, pRequestB );
    }

  pUKT->lWaitForDeviIO++;

  *ok = true;

  /*
   *  ulIOCounts is 2 if write ( twice ) and 1 if read ( once )
   */
   if ( fIsReading )
      pIOStat = &pTaskCtrl->TaskStateStat.DevIOReadStat;
   else
      pIOStat = &pTaskCtrl->TaskStateStat.DevIOWriteStat;

  pIOStat->ulCount     += ulIOCounts;
  pIOStat->ulPageCount += ulIOCounts * buf_count ;

  for (; ulIOCounts; ulIOCounts--)   // receive and process the replies
    {
    // --- goto dispatcher
    EO52_TAKE_REL_ABS_TIME( pTaskCtrl, (*pIOStat), GOTO_DISP(pUKT));

    #ifdef DEBUG
     if ( pUKT->pCurrReq == pRequestA )
       {
       DBG3 (( MF__, "[T:0x%03u] pRequestA, devspace no. %u",
               pTaskCtrl->ulTaskIndex, devno1));
       }
     if ( pUKT->pCurrReq == pRequestB )
       {
       DBG3 (( MF__, "[T:0x%03u] pRequestB, devspace no. %u",
               pTaskCtrl->ulTaskIndex, devno2));
       }
    #endif


    if (!pUKT->pCurrReq->ReqArgs.VblockIORequest.fOk)
      {
      *ok = false;

      // --- system not aborting?
      if ( pUKT->pCurrReq == pRequestA )
        {
		*error_dev  = fIsReading ? *read_dev : V2_DEV1;

        if ( pRequestA->ReqArgs.VblockIORequest.ulErrCode < MAX_DEVIO_ERRCODE )
          {
          sql46c_build_error_string ( errtext1,
                                      ErrTab[ pRequestA->ReqArgs.VblockIORequest.ulErrCode ], 0 );
          DBG1 (( MF__, "[T:0x%03u] pRequestA: 0x%x, %s",
                  pTaskCtrl->ulTaskIndex, pRequestA,
                  ErrTab[ pUKT->pCurrReq->ReqArgs.VblockIORequest.ulErrCode ] ));
          }
        }
      else if ( pUKT->pCurrReq == pRequestB )
        {
        *error_dev  = V2_DEV2;

        if ( pRequestB->ReqArgs.VblockIORequest.ulErrCode < MAX_DEVIO_ERRCODE )
          {
          sql46c_build_error_string ( errtext2,
                                      ErrTab[ pRequestB->ReqArgs.VblockIORequest.ulErrCode ], 0 );
          DBG1 (( MF__, "[T:0x%03u] pRequestB: 0x%x, %s",
                  pTaskCtrl->ulTaskIndex, pRequestB,
                  ErrTab[ pUKT->pCurrReq->ReqArgs.VblockIORequest.ulErrCode ] ));
          }
        }
      else
        {
        MSGALL(( ERR_UNKNOWN_REQUEST ));
        DBG1  (( MF__, "[T:0x%03u] pRequestA: 0x%x and pRequestB: 0x%x (SEND) "
                "!= pUKT->pCurrReq: 0x%x (REPLY)",
                 pTaskCtrl->ulTaskIndex, pRequestA, pRequestB, pUKT->pCurrReq ));
        }
      }
    }

  #ifdef DEBUG
   if (*ok)
     {
     DBG3 (( MF__, "[T:0x%03u] Returned OK, pRequestA: 0x%x, pRequestB: 0x%x",
             pTaskCtrl->ulTaskIndex, pRequestA, pRequestB ));
     }
  #endif

  pUKT->lWaitForDeviIO--;
  __os54VblockioFinish(pTaskCtrl);

  #ifdef DEVELOPMENT
   pTaskCtrl->ulWaitTime = 0;
  #endif

  DBGOUT_T (pTaskCtrl->ulTaskIndex);
  return;
  }

/*------------------------------*/

VOID vdetach ( INT4        devno,
               PROCESS_ID  pid )
  {
  #undef  MF__
  #define MF__ MOD__"vdetach"
  PDLQ_REC                       pRequest;
  tsp2_devnamec                  szLogDevName;
  #ifdef DEVELOP_CHECKS
   PUKT_CTRL_REC                 pUKT      = THIS_UKT_CTRL;
   PTASK_CTRL_REC                pTaskCtrl = pUKT->pCTask;
  #else
   PTASK_CTRL_REC                pTaskCtrl = &kgs.pFirstTaskCtrl[pid - 1];
   PUKT_CTRL_REC                 pUKT      = pTaskCtrl->pUKT;
  #endif

  DBGIN_T (pTaskCtrl->ulTaskIndex);

  #ifdef DEVELOP_CHECKS
   if ( (ULONG)pid != pTaskCtrl->ulTaskIndex )
     {
     MSGD (( ERR_VXXXX_WRONG_TASK, "vdetach", pid ));
     ABORT();
     }
  #endif

  if ( kgs.Dev0.ThrdCtrlHeader.ThrdState == KT_INITIALIZED )
    return;

  strcpy ( szLogDevName, kgs.Dev0.pIOCtrl[devno].szDevspace );

  MSGD (( INFO_DETACHING_DEVSPACE, szLogDevName ));

  if ( devno < 0 || devno > XPARAM(lMaxDevspaces) )
    {
    MSGALL(( ERR_WRONG_DEVSPACE_NO, devno ));
    DBG1  (( MF__, "[T:0x%03u] Wrong devspace number: %d",
             pTaskCtrl->ulTaskIndex, devno ));
    return;
    }

  pTaskCtrl->TaskState = TSK_VDETACH;

  pRequest                                     = &pTaskCtrl->SpecDLQElem1;
  pRequest->pTaskCtrl                          = pTaskCtrl;
  pRequest->ulReqType                          = REQ_VDETACH;
  pRequest->ReqArgs.VdetachRequest.lDevspaceNo = devno;

  // --- send request to DEV0 thread
  REQ_TO_DEV0( pRequest, pTaskCtrl );

  // --- goto dispatcher
  GOTO_DISP (pUKT);

  // --- wrong reply?
  if ( pUKT->pCurrReq != pRequest )
    {
    MSGALL(( ERR_UNKNOWN_REQUEST ));
    DBG1  (( MF__, "[T:0x%03u] pRequest: 0x%x (SEND) != pUKT->pCurrReq: 0x%x (REPLY)",
             pTaskCtrl->ulTaskIndex, pRequest, pUKT->pCurrReq ));
    ABORT();
    }

  pTaskCtrl->TaskState = TSK_RUNNING;

  DBGOUT_T (pTaskCtrl->ulTaskIndex);
  return;
  }

/*------------------------------*/

VOID vdump_rte ( VOID )
{
  char *RTEDumpName = XPARAM(szRTEDumpFilename) ;

  if ( *RTEDumpName != '\0' ) /* PTS 1109210 */
  {
    int             rc ;
    tsp01_RteError  RteError ;
    tsp00_Pathc     szPgmPath;
    char            szCmdBuf[sizeof(tsp00_Pathc) + sizeof(SQL_DBNAMEC) + 
                             sizeof(VF_FILENAMEC) + 50 ];


    if ( sqlGetDbrootPgmPath (szPgmPath, TERM_WITH_DELIMITER_EO01, &RteError) )
    {
      kgs.ulDebugLevel=3 ;
      sprintf(szCmdBuf, "%scons.exe %s sh all > %s", 
                         szPgmPath, kgs.szServerDB, RTEDumpName);
      rc = system(szCmdBuf) ; 
    }
  }
}

/*------------------------------*/

VOID _System sql54k_dev0 ( VOID )
  {
  #undef  MF__
  #define MF__ MOD__"sql54k_dev0"
  APIRET                         rc = NO_ERROR;
  PDLQ_REC                       pRequest;
  LONG                           lDummy;
  DLQ_MARK_BAD_DEV_REQ_ARGS_REC  *pMarkBadDevRequest;

  DBGIN;

  __try 
  {
    // --- Create event semaphore (SET) for DEV0
    rc = sql41c_create_event_sem (&kgs.Dev0.ThrdCtrlHeader.hevSem,
                                  NULL,
                                  NULL,
                                  0,
                                  false,
                                  NULL);
    if (rc != NO_ERROR)
      ABORT();

    // --- set the thread priority
    rc = SET_THREAD_PRIO ( kgs.Dev0.ThrdCtrlHeader.lThrdPrio,
                           kgs.Dev0.ThrdCtrlHeader.Tid,
                           kgs.Dev0.ThrdCtrlHeader.hThrd );

    if ( rc != NO_ERROR )
      {
      MSGD (( ERR_SETTING_THRD_PRIO, kgs.Dev0.ThrdCtrlHeader.Tid));
      DBG1 ((MF__, "Cannot set the thread priority, TID: %u",
                    kgs.Dev0.ThrdCtrlHeader.Tid));
      ABORT();
      }

    // --- set thread control info
    kgs.Dev0.ThrdCtrlHeader.ThrdState   = KT_RUNNING;
    kgs.Dev0.ThrdCtrlHeader.ulStartTime = kgs.ulCurrTime;

    // --- Runs forever
    for (;; )
      {
      DBG3 (( MF__, "kgs.Dev0.pIOWorkQue           is 0x%x", kgs.Dev0.pIOWorkQue ));
      DBG3 (( MF__, "kgs.Dev0.pIOReqQue            is 0x%x", kgs.Dev0.pIOReqQue ));
      DBG3 (( MF__, "kgs.Dev0.pIOWorkQue->pRequest is 0x%x", kgs.Dev0.pIOWorkQue->pRequest ));

      /*
       *  Wait for a request to DEV0
       */
      pRequest = sql54k_dev_receive (&kgs.Dev0.pIOWorkQue,
                                     &lDummy,
                                     &kgs.Dev0.lExclusive,
                                     &kgs.Dev0.ThrdCtrlHeader);

      switch ( pRequest->ulReqType )
        {
        case REQ_VATTACH :
          sql54k_dev0_vattach ( &pRequest->ReqArgs.VattachRequest );
          break;

        case REQ_VDETACH :
          sql54k_dev0_vdetach ( pRequest );
          break;

        case REQ_MARK_BAD_DEVSPACE :
          pMarkBadDevRequest = &pRequest->ReqArgs.MarkBadDevRequest;

          sql54k_vmark_bad_dev ( pMarkBadDevRequest->lDevspaceNo,
                                 pMarkBadDevRequest->lBadDevspacePage );
          break;

        default          :
          MSGD (( ERR_ILLEGAL_REQUEST, "DEV0", pRequest->ulReqType ));
        }

      // -- send reply back to UKT
      REQ_FROM_DEVx_TO_UKT(pRequest);

      DBG3 (( MF__, "next in loop" ));
      }
  } 
  __except( sql50k_excepthandler(GetExceptionInformation()) ) {;}

  DBGOUT;
  EXITTHREAD ( 0, kgs.Dev0.ThrdCtrlHeader.hThrd );

  return;
  }

/*------------------------------*/

VOID _System sql54k_dev_worker ( PDEV_WORKER_CTRL_REC pDevWorkerCtrl )
  {
  #undef  MF__
  #define MF__ MOD__"sql54k_dev_worker"
  APIRET                        rc = NO_ERROR;
  #if defined (_WIN32)
   BOOL                         fSuccess;
   ULONG                        ulNumOfBytes;
   LPOVERLAPPED                 pCompletedOverlapped;
   LONG                         lDevspaceNo;
  #endif

  DBGIN;

  __try 
  {
    // --- Create event semaphore (SET) for DEV0
    rc = sql41c_create_event_sem (&pDevWorkerCtrl->ThrdCtrlHeader.hevSem,
                                  NULL,
                                  NULL,
                                  0,
                                  false,
                                  NULL);
    if (rc != NO_ERROR)
      ABORT();

    // --- set the thread priority
    rc = SET_THREAD_PRIO ( pDevWorkerCtrl->ThrdCtrlHeader.lThrdPrio,
                           pDevWorkerCtrl->ThrdCtrlHeader.Tid,
                           pDevWorkerCtrl->ThrdCtrlHeader.hThrd );

    if ( rc != NO_ERROR )
      {
      MSGD (( ERR_SETTING_THRD_PRIO, pDevWorkerCtrl->ThrdCtrlHeader.Tid));
      DBG1 ((MF__, "Cannot set the thread priority, TID: %u",
                    pDevWorkerCtrl->ThrdCtrlHeader.Tid));
      ABORT();
      }

    // --- set thread control info
    pDevWorkerCtrl->ThrdCtrlHeader.ThrdState   = KT_RUNNING;
    pDevWorkerCtrl->ThrdCtrlHeader.ulStartTime = kgs.ulCurrTime;

    #if defined (_WIN32)
     // --- Runs forever
     for (;;)
       {
       pDevWorkerCtrl->ThrdCtrlHeader.ThrdState = KT_SLEEPING;

       fSuccess = GetQueuedCompletionStatus( kgs.hCompletionPort,
                                             &ulNumOfBytes,
                                             &lDevspaceNo,
                                             &pCompletedOverlapped,
                                             (DWORD)-1);

       pDevWorkerCtrl->ThrdCtrlHeader.ThrdState = KT_RUNNING;

       if (!fSuccess )
         {
         rc = GetLastError();

         if (pCompletedOverlapped == NULL)
           {
           MSGD (( ERR_GET_QUEUED_COMPL_STATUS, GetLastError()));
           DBG1 ((MF__, "GetQueuedCompletionStatus failed, rc = %d",
                  rc ));
           ABORT();
           }
        }
       else
         rc = NO_ERROR;

       pDevWorkerCtrl->ulIOCount++;

       REQ_FROM_DEV_WOKER_TO_UKT ( fSuccess, ulNumOfBytes,
                                   lDevspaceNo, pCompletedOverlapped, rc );

       DBG3 (( MF__, "next comleted io" ));
       }
    #endif
  } 
  __except( sql50k_excepthandler(GetExceptionInformation()) ) {;}

  DBGOUT;
  EXITTHREAD ( 0, pDevWorkerCtrl->ThrdCtrlHeader.hThrd );

  return;
  }

/*------------------------------*/

//
// --- This function is called by the exception handler to
//     set free the tasks waiting for DEV0 or DEVi activities
//
VOID sql54k_dev0_devi_crash_cleanup ( TID CrashedTid )
  {
  #undef  MF__
  #define MF__ MOD__"sql54k_dev0_devi_crash_cleanup"
  PDEVI_CTRL_REC  pCurrDeviCtrl;
  DBGPAS;

  // --- DEV0 thread crashed
  if ( kgs.Dev0.ThrdCtrlHeader.Tid == CrashedTid )
    {
    // --- set thread state to signal this thread is down
    kgs.Dev0.ThrdCtrlHeader.ThrdState = KT_INITIALIZED;
    sql54k_wake_up_tasks ( &kgs.Dev0.pIOWorkQue,
                           &kgs.Dev0.lExclusive );
    return;
    }

  for (pCurrDeviCtrl =  kgs.pFirstDeviCtrl;
       pCurrDeviCtrl <= kgs.pLastDeviCtrl;
       pCurrDeviCtrl++)
    {
    // --- DEVi thread crashed
    if ( pCurrDeviCtrl->ThrdCtrlHeader.Tid == CrashedTid )
      {
      // --- set thread state to signal this thread is down
      pCurrDeviCtrl->ThrdCtrlHeader.ThrdState = KT_INITIALIZED;
      sql54k_wake_up_tasks ( &pCurrDeviCtrl->pIOWorkQue,
                             &pCurrDeviCtrl->lExclusive );
      return;
      }
    }
  return;
  }

/*------------------------------*/

/* PTS 1106176 */
APIRET  sql54k_ReadBlockWithRetry (HFILE                   hfDevspace,
                                   BUFFERADDRESS           pBuffAddr,
                                   ULONG                   ulLengthWanted,
                                   ULONG                  *ulLength,
                                   ULONG                   lDevspacePage,
                                   PSZ                     szDevspace)
{
  int             Retries = 0 ;
  APIRET          rc = NO_ERROR;
  ULONG           ulFilePosLow;
  ULONG           ulFilePosHigh;

  while (    (rc = READ_FILE ( hfDevspace, pBuffAddr, ulLengthWanted, ulLength) ) != NO_ERROR
          && ++Retries < NUMBER_OF_BLOCKIO_RETRIES )
  {
    MSGALL(( ERR_CANT_READ_FROM_DEVSPACE, lDevspacePage + 1, szDevspace, rc, ulLength ));
    SLEEP(1000); 
    PAGES_TO_QWORD( ulFilePosLow, ulFilePosHigh, lDevspacePage + 1 );

    rc = SET_FILE_PTR ( hfDevspace, ulFilePosLow, FILE_BEGIN,
                        &ulFilePosLow, &ulFilePosHigh );

    if (rc != 0)
    {
      MSGALL(( ERR_CANT_SEEK_ON_DEVSPACE, szDevspace, rc ));
      return(rc);
    }
  }
  return ( rc ) ;
}


/*===========================================================================*
 *  LOCAL FUNCTIONS (CODE)                                                   *
 *===========================================================================*/

static LONG sql54k_overlapped_io( PUKT_CTRL_REC    pUKT,
                                  LONG             lDevspaceNo,
                                  tsp_page_no      lDevspacePage,
                                  tsp2_io_op_code  OpCode,
                                  BUFFERADDRESS    pBuffAddr,
                                  SHORT            sNumOfPages,
                                  ERRORTEXT        pcErrText )
  {
  #undef  MF__
  #define MF__ MOD__"sql54k_overlapped_io"
  ULONG               ulFilePosLow;
  ULONG               ulFilePosHigh;
  ULONG               ulLengthWanted;
  APIRET              rc = NO_ERROR;
  PDEV_IO_CTRL_REC    pIOCtrl;
  OVERLAPPED          *pOverlapped;
  PDLQ_REC            pRequest;

  DBGIN_T (pUKT->pCTask->ulTaskIndex);

  pIOCtrl        = &kgs.Dev0.pIOCtrl[lDevspaceNo];
  pOverlapped    = &pUKT->pCTask->Overlapped;
  ulLengthWanted = PAGES_TO_BYTES(sNumOfPages);
  PAGES_TO_QWORD( ulFilePosLow, ulFilePosHigh, lDevspacePage + 1);

  //
  // --- do read/write operation
  //
  pOverlapped->Offset     = ulFilePosLow;
  pOverlapped->OffsetHigh = ulFilePosHigh;

  if ( OpCode == DEV_READ )
    {
    rc = READ_FILE_OL( pIOCtrl->hfDevspace, pBuffAddr, ulLengthWanted,
                       NULL, pOverlapped );

    // -- are there too many outstanding asynchronous I/O requests?
    if (( rc == ERROR_INVALID_USER_BUFFER ) ||
        ( rc == ERROR_NOT_ENOUGH_QUOTA )    ||
        ( rc == ERROR_NOT_ENOUGH_MEMORY ))
      {
      DBGOUT_T ( pUKT->pCTask->ulTaskIndex );
      return ( IO_CONTINUE );
      }

    if (( rc != NO_ERROR ) && ( rc != ERROR_IO_PENDING ))
      {
      MSGALL(( ERR_CANT_READ_FROM_DEVSPACE, lDevspacePage + 1,
               pIOCtrl->szDevspace, rc, ulLengthWanted ));
      sql46c_build_error_string ( pcErrText, ERRMSG_DEVIO_READ_FROM_DEVSPACE, rc );

      DBGOUT_T ( pUKT->pCTask->ulTaskIndex );
      return ( IO_ERROR );
      }
    pUKT->pSingleBlockIO[lDevspaceNo].lReadCntr++;
    pUKT->pSingleBlockIO[lDevspaceNo].lReadPages += sNumOfPages;
    pIOCtrl->lReadCnt++;
    }
  else
    {
    rc = WRITE_FILE_OL( pIOCtrl->hfDevspace, pBuffAddr, ulLengthWanted,
                        NULL, pOverlapped );

    // -- are there too many outstanding asynchronous I/O requests?
    if (( rc == ERROR_INVALID_USER_BUFFER ) ||
        ( rc == ERROR_NOT_ENOUGH_QUOTA )    ||
        ( rc == ERROR_NOT_ENOUGH_MEMORY ))
      {
      DBGOUT_T ( pUKT->pCTask->ulTaskIndex );
      return ( IO_CONTINUE );
      }

    if (( rc != NO_ERROR ) && ( rc != ERROR_IO_PENDING ))
      {
      MSGALL(( ERR_CANT_WRITE_ON_DEVSPACE, lDevspacePage + 1,
               pIOCtrl->szDevspace, rc, ulLengthWanted ));
      sql46c_build_error_string ( pcErrText, ERRMSG_DEVIO_WRITE_ON_DEVSPACE, rc );

      DBGOUT_T ( pUKT->pCTask->ulTaskIndex );
      return ( IO_ERROR );
      }

    pUKT->pSingleBlockIO[lDevspaceNo].lWriteCntr++;
    pUKT->pSingleBlockIO[lDevspaceNo].lWritePages += sNumOfPages ;
    pIOCtrl->lWriteCnt++;
    }

  GOTO_DISP(pUKT)

  pRequest = pUKT->pCurrReq;

  if (( pRequest            != &pUKT->pCTask->SpecDLQElem1 ) ||
      ( pRequest->ulReqType != REQ_OVERLAPPED_IO_FINISED ))
    {
    MSGALL(( ERR_UNKNOWN_REQUEST ));
    DBG1  (( MF__, "[T:0x%03u] pRequest: 0x%x (SEND) != pUKT->pCurrReq: 0x%x (REPLY)",
            pUKT->pCTask->ulTaskIndex,
            pUKT->pCTask->SpecDLQElem1, pRequest ));
    ABORT();
    }

  if (( pRequest->ReqArgs.DevWorkerRequest.fOk          == false ) ||
      ( pRequest->ReqArgs.DevWorkerRequest.ulNumOfBytes != ulLengthWanted ))
    {
    rc = pRequest->ReqArgs.DevWorkerRequest.APIRetCode;

    if ( OpCode == DEV_READ )
      {
      MSGALL(( ERR_CANT_READ_FROM_DEVSPACE, lDevspacePage + 1,
               pIOCtrl->szDevspace, rc, ulLengthWanted ));
      sql46c_build_error_string ( pcErrText, ERRMSG_DEVIO_READ_FROM_DEVSPACE, rc );
      }
    else
      {
      MSGALL(( ERR_CANT_WRITE_ON_DEVSPACE, lDevspacePage + 1,
               pIOCtrl->szDevspace, rc, ulLengthWanted ));
      sql46c_build_error_string ( pcErrText, ERRMSG_DEVIO_WRITE_ON_DEVSPACE, rc );
      }
    DBGOUT_T ( pUKT->pCTask->ulTaskIndex );
    return ( IO_ERROR );
    }

  DBGOUT_T ( pUKT->pCTask->ulTaskIndex );
  return ( IO_NOERROR );
  }

/*------------------------------*/


static LONG sql54k_single_blockio( PUKT_CTRL_REC    pUKT,
                                   LONG             lDevspaceNo,
                                   tsp_page_no      lDevspacePage,
                                   tsp2_io_op_code  OpCode,
                                   BUFFERADDRESS    pBuffAddr,
                                   SHORT            sNumOfPages,
                                   ERRORTEXT        pcErrText )
  {
  #undef  MF__
  #define MF__ MOD__"sql54k_single_blockio"
  ULONG           ulFilePosLow;
  ULONG           ulFilePosHigh;
  HFILE           hfFileHandle;
  ULONG           ulLengthWanted;
  ULONG           ulLength;
  ULONG           ulTmp;
  LONG            lDevspaceType;
  APIRET          rc          = NO_ERROR;
  bool            fDevspaceIsPSE36 = kgs.Dev0.pIOCtrl[lDevspaceNo].ulDevType == DT_PSE36 ;

  DBGIN_T (pUKT->pCTask->ulTaskIndex);

  DBG3 (( MF__, "[T:0x%03u] lDevspaceNo: %u, using hfFileHandle: %u",
          pUKT->pCTask->ulTaskIndex, lDevspaceNo,
          pUKT->pSingleBlockIO[lDevspaceNo].hfSingleIOFiles ));

  // PSE36 device must opened for ONE TIME ONLY !!! 
  // this is done by vattach                    !!!
  if ( fDevspaceIsPSE36 )
    hfFileHandle = kgs.Dev0.pIOCtrl[lDevspaceNo].hfDevspace ;
  else
    {
    // --- file not open?
    if ( pUKT->pSingleBlockIO[lDevspaceNo].hfSingleIOFiles == (HFILE)INVALID_HANDLE_VALUE )
      {
      if ( kgs.fSingleIOOpenLocked == true )
        {
        DBGOUT_T ( pUKT->pCTask->ulTaskIndex );
        return ( IO_CONTINUE );
        }

      if ( kgs.Dev0.pIOCtrl[lDevspaceNo].szDevspace[0] == '\0' )
        {
        MSGALL(( ERR_SINGLEIO_DEVSPACE_UNDEF, lDevspaceNo ));
        sql46c_build_error_string ( pcErrText,
                                    ERRMSG_DEVIO_SINGLEIO_DEVSP_UNDEF, 0 );
        DBGOUT_T ( pUKT->pCTask->ulTaskIndex );
        return ( IO_ERROR );
        }

      MSGD (( INFO_SINGLEIO_ATTACH, kgs.Dev0.pIOCtrl[lDevspaceNo].szDevspace,
              pUKT->ulUKTIndex));

      rc = sql54k_dev_open ( kgs.Dev0.pIOCtrl[lDevspaceNo].szDevspace,
                             lDevspaceNo,
                             DEV_NO_BUFFERING, // PTS 1110746
                             &hfFileHandle,
                             pcErrText,
                             &ulTmp,
                             &lDevspaceType);

      if ( rc == ERROR_TOO_MANY_OPEN_FILES )
        {
        kgs.fSingleIOOpenLocked = true;
        DBGOUT_T ( pUKT->pCTask->ulTaskIndex );
        return ( IO_CONTINUE );
        }
      else if ( rc != NO_ERROR )
        {
        DBGOUT_T ( pUKT->pCTask->ulTaskIndex );
        return ( IO_ERROR );
        }

      pUKT->pSingleBlockIO[lDevspaceNo].ulOldSingleIOPos = 0;
      pUKT->pSingleBlockIO[lDevspaceNo].hfSingleIOFiles  = hfFileHandle;
      }
    else
      hfFileHandle = pUKT->pSingleBlockIO[lDevspaceNo].hfSingleIOFiles ;

    //
    // --- file-seek necessary?
    //
    if ( pUKT->pSingleBlockIO[lDevspaceNo].ulOldSingleIOPos != (ULONG)(lDevspacePage + 1) )
      {
      PAGES_TO_QWORD( ulFilePosLow, ulFilePosHigh, lDevspacePage + 1);

      rc = SET_FILE_PTR ( hfFileHandle, ulFilePosLow, FILE_BEGIN,
                          &ulFilePosLow, &ulFilePosHigh );

      if (rc != NO_ERROR)
        {
        sql46c_build_error_string ( pcErrText, ERRMSG_DEVIO_SEEK_ON_DEVSPACE, rc );
        MSGALL(( ERR_CANT_SEEK_ON_DEVSPACE, kgs.Dev0.pIOCtrl[lDevspaceNo].szDevspace, rc ));
        DBG1  (( MF__, "[T:0x%03u] Cannot seek on devspace: %s, rc = %d",
                 pUKT->pCTask->ulTaskIndex, kgs.Dev0.pIOCtrl[lDevspaceNo].szDevspace, rc ));

        DBGOUT_T ( pUKT->pCTask->ulTaskIndex );
        return ( IO_ERROR );
        }
      }

    #ifdef DEVELOP_CHECKS
      rc = __sql54k_check_filepos ( hfFileHandle, lDevspaceNo,
                                   lDevspacePage, sNumOfPages,
                                   kgs.Dev0.pIOCtrl[lDevspaceNo].ulDevPages,
                                   NULL, pcErrText );

      if ( rc != NO_ERROR )
        {
        DBGOUT_T ( pUKT->pCTask->ulTaskIndex );
        return ( IO_ERROR );
        }
    #endif
    }
  //
  // --- do read/write operation
  //
  ulLengthWanted = PAGES_TO_BYTES(sNumOfPages);

  if ( OpCode == DEV_READ )
    {
    pUKT->pSingleBlockIO[lDevspaceNo].lReadCntr++;
    pUKT->pSingleBlockIO[lDevspaceNo].lReadPages += sNumOfPages;

    if ( fDevspaceIsPSE36 )
      rc = os541PSE36Read( (HANDLE )hfFileHandle, pBuffAddr, ulLengthWanted, 
                           (LONGLONG)( lDevspacePage+1 ) * XPARAM(ulDBPageSize),
                           &ulLength );
    else
      /* PTS 1106176 */
      rc = sql54k_ReadBlockWithRetry( hfFileHandle, pBuffAddr, ulLengthWanted, &ulLength, 
                                      lDevspacePage, kgs.Dev0.pIOCtrl[lDevspaceNo].szDevspace );

    if ( ( rc != NO_ERROR ) || ( ulLengthWanted != ulLength ) )
      {
      DBG1 (( MF__, "[T:0x%03u] Cannot read from devspace '%s', rc = %d",
              pUKT->pCTask->ulTaskIndex, kgs.Dev0.pIOCtrl[lDevspaceNo].szDevspace, rc ));
      MSGALL(( ERR_CANT_READ_FROM_DEVSPACE, lDevspacePage + 1,
               kgs.Dev0.pIOCtrl[lDevspaceNo].szDevspace, rc, ulLength ));
      sql46c_build_error_string ( pcErrText, ERRMSG_DEVIO_READ_FROM_DEVSPACE, rc );

      DBGOUT_T ( pUKT->pCTask->ulTaskIndex );
      return ( IO_ERROR );
      }
    }
  else
    {
    pUKT->pSingleBlockIO[lDevspaceNo].lWriteCntr ++;
    pUKT->pSingleBlockIO[lDevspaceNo].lWritePages += sNumOfPages ;

    if ( fDevspaceIsPSE36 )
      rc = os541PSE36Write( (HANDLE )hfFileHandle, pBuffAddr, ulLengthWanted, 
                            (LONGLONG)( lDevspacePage+1 ) * XPARAM(ulDBPageSize),
                            &ulLength );
    else
      rc = WRITE_FILE( hfFileHandle, pBuffAddr, ulLengthWanted, &ulLength );

    if ( ( rc != NO_ERROR ) || ( ulLengthWanted != ulLength ) )
      {
      DBG1 (( MF__, "[T:0x%03u] Cannot write on devspace '%s', rc = %d",
              pUKT->pCTask->ulTaskIndex, kgs.Dev0.pIOCtrl[lDevspaceNo].szDevspace, rc ));
      MSGALL(( ERR_CANT_WRITE_ON_DEVSPACE, lDevspacePage + 1,
               kgs.Dev0.pIOCtrl[lDevspaceNo].szDevspace, rc, ulLength ));
      sql46c_build_error_string ( pcErrText, ERRMSG_DEVIO_WRITE_ON_DEVSPACE, rc );

      DBGOUT_T ( pUKT->pCTask->ulTaskIndex );
      return ( IO_ERROR );
      }
    }

  #ifdef DEVELOP_CHECKS
   rc = __sql54k_check_filepos ( hfFileHandle, lDevspaceNo,
                                 lDevspacePage + sNumOfPages, 0,
                                 kgs.Dev0.pIOCtrl[lDevspaceNo].ulDevPages,
                                 NULL, pcErrText );

   if ( rc != NO_ERROR )
     {
     DBGOUT_T ( pUKT->pCTask->ulTaskIndex );
     return ( IO_ERROR );
     }
  #endif

  // --- save current IO-position
  pUKT->pSingleBlockIO[lDevspaceNo].ulOldSingleIOPos = lDevspacePage + sNumOfPages + 1;

  DBGOUT_T ( pUKT->pCTask->ulTaskIndex );
  return ( IO_NOERROR );
  }

/*------------------------------*/

/*
 * Vdevsize for a device being a file for direct io.
 */

#if defined(_WIN32)

 /*
  * Vdevsize for WIN32
  */

 static int sql54k_vdevsize( PSZ            pszPhysDevName,
                             PSZ            pszLogDevName,
                             INT4           *devcapacity,
                             ULONG          ulCurrFileSizeLow,
                             ULONG          ulCurrFileSizeHigh,
                             ERRORTEXT      errtext,
                             BOOLEAN        *ok )
   {
   #undef  MF__
   #define MF__  MOD__"sql54k_vdevsize"

   HANDLE         hfFileHandle;
   ULONG          ulFilePosLow;
   ULONG          ulFilePosHigh;
   APIRET         rc    = NO_ERROR;
   PUKT_CTRL_REC  pUKT  = THIS_UKT_CTRL;

   DBGIN;
   DBG3 ((MF__, "devspace: %s", pszLogDevName));

   *ok = false;

   if (*devcapacity == 0)
     {
     DBG3 ((MF__, "Opening devspace: %s", pszLogDevName));

     hfFileHandle = __sql54k_create_file( pszPhysDevName,
                                          GENERIC_READ      |
                                          GENERIC_WRITE,
                                          FILE_SHARE_READ   |
                                          FILE_SHARE_WRITE,
                                          NULL,
                                          OPEN_EXISTING,
                                          FILE_ATTRIBUTE_NORMAL,
                                          NULL );

     if (hfFileHandle == INVALID_HANDLE_VALUE)
       {
       rc = GetLastError();

       if ( rc == ERROR_ACCESS_DENIED )
         {
         MSGALL(( ERR_DEVSPACE_ACCESS_DENIED, pszLogDevName ));
         DBG1  (( MF__, "Access on devspace '%s' denied", pszLogDevName ));
         sql46c_build_error_string ( errtext, ERRMSG_DEVIO_ACCESS_DENIED, 0 );
         }
       else
         {
         MSGALL(( ERR_CANT_OPEN_DEVSPACE, pszLogDevName, rc ));
         DBG1  (( MF__, "Cannot open devspace '%s', rc = %d", pszLogDevName, rc ));
         sql46c_build_error_string ( errtext, ERRMSG_DEVIO_OPEN_DEVSPACE, rc );
         }

       DBGOUT;
       return ( rc );
       }

     if (HIGH_VALUE_VALID(ulCurrFileSizeHigh) == true)
       *devcapacity = QWORD_TO_PAGES(ulCurrFileSizeLow, ulCurrFileSizeHigh) - 1;
     else
       *devcapacity = MAX_DEVSPACE_PAGES + 1;

     if ((*devcapacity < 1) || ((ULONG)*devcapacity > MAX_DEVSPACE_PAGES))
       {
       *devcapacity += HIGH_VALUE_VALID(ulCurrFileSizeHigh) ? 1 : -1 ;
       MSGALL((ERR_INV_DEVSPACE_SIZE, pszLogDevName, *devcapacity ));
       DBG1  ((MF__, "Invalid devspace size: %u pages", *devcapacity ));
       sql46c_build_error_string ( errtext, ERRMSG_DEVIO_INV_DEVSPACE_SIZE, 0 );

       CloseHandle( hfFileHandle );

       DBGOUT;
       return ( -1 );
       }
     }
   else
     {
     if ((*devcapacity < 1) || ((ULONG)*devcapacity > MAX_DEVSPACE_PAGES))
       {
       MSGALL((ERR_INV_DEVSPACE_SIZE, pszLogDevName, *devcapacity ));
       DBG1  ((MF__, "Invalid devspace size: %u pages", *devcapacity ));
       sql46c_build_error_string ( errtext, ERRMSG_DEVIO_INV_DEVSPACE_SIZE, 0 );

       DBGOUT;
       return ( -1 );
       }

     DBG3 (( MF__, "Creating devspace: %s - capacity: %u pages",
             pszLogDevName, *devcapacity ));

     hfFileHandle = __sql54k_create_file( pszPhysDevName,
                                          GENERIC_READ      |
                                          GENERIC_WRITE,
                                          FILE_SHARE_READ   |
                                          FILE_SHARE_WRITE,
                                          NULL,
                                          OPEN_ALWAYS,
                                          FILE_ATTRIBUTE_NORMAL  |
                                          FILE_FLAG_SEQUENTIAL_SCAN,
                                          NULL );

     if (hfFileHandle == INVALID_HANDLE_VALUE)
       {
       rc = GetLastError();

       if ( rc == ERROR_ACCESS_DENIED )
         {
         MSGALL(( ERR_DEVSPACE_ACCESS_DENIED, pszLogDevName ));
         DBG1  (( MF__, "Access on devspace '%s' denied", pszLogDevName ));
         sql46c_build_error_string ( errtext, ERRMSG_DEVIO_ACCESS_DENIED, 0 );
         }
       else
         {
         MSGALL(( ERR_CANT_CREATE_DEVSPACE, pszLogDevName, rc ));
         DBG1  (( MF__, "Cannot create devspace '%', rc = %d", pszLogDevName,
                  rc ));
         sql46c_build_error_string ( errtext, ERRMSG_DEVIO_CREATE_DEVSPACE, rc );
         }

       DBGOUT;
       return ( rc );
       }

	 if ( XPARAM(fAllowCompressedDevsp) == false )
	   sql54k_uncompress( hfFileHandle );

     if ( XPARAM(fFormatDevspace) == true )
       {
       // --- set the thread priority
       rc = SET_THREAD_PRIO ( THREAD_PRIORITY_IDLE,
                              pUKT->ThrdCtrlHeader.Tid,
                              pUKT->ThrdCtrlHeader.hThrd );

       if ( rc != NO_ERROR )
          {
          MSGD (( ERR_SETTING_THRD_PRIO, pUKT->ThrdCtrlHeader.Tid));
          DBG1 ((MF__, "Cannot set the thread priority, TID: %u, rc=%d",
                 pUKT->ThrdCtrlHeader.Tid, rc ));
          DBGOUT;
          return ( rc );
          }

       rc = sql54k_format_devspace ( (HFILE)hfFileHandle, pszLogDevName,
                                     ulCurrFileSizeLow, ulCurrFileSizeHigh,
                                     *devcapacity + 1,
                                     pUKT->pCTask,
                                     errtext );
       if ( rc != NO_ERROR )
         {
         CloseHandle( hfFileHandle );

         SET_THREAD_PRIO ( pUKT->ThrdCtrlHeader.lThrdPrio,
                           pUKT->ThrdCtrlHeader.Tid,
                           pUKT->ThrdCtrlHeader.hThrd );

         DBGOUT;
         return ( rc );
         }

       // --- set the thread priority
       rc = SET_THREAD_PRIO ( pUKT->ThrdCtrlHeader.lThrdPrio,
                              pUKT->ThrdCtrlHeader.Tid,
                              pUKT->ThrdCtrlHeader.hThrd );

       if ( rc != NO_ERROR )
          {
          MSGD (( ERR_SETTING_THRD_PRIO, pUKT->ThrdCtrlHeader.Tid));
          DBG1 ((MF__, "Cannot set the thread priority, TID: %u, rc=%d",
                 pUKT->ThrdCtrlHeader.Tid, rc ));
          DBGOUT;
          return ( rc );
          }
       }

     PAGES_TO_QWORD( ulFilePosLow, ulFilePosHigh, *devcapacity + 1);

     rc = SET_FILE_PTR ( hfFileHandle, ulFilePosLow, FILE_BEGIN,
                         &ulFilePosLow, &ulFilePosHigh );

     if (rc == ERROR_DISK_FULL)
       {
       MSGALL(( ERR_DISK_FULL, pszLogDevName ));
       DBG1  (( MF__, "Could not create: '%s', not enough space on disk",
                pszLogDevName ));
       sql46c_build_error_string ( errtext, ERRMSG_DEVIO_DISK_FULL, 0 );

       CloseHandle( hfFileHandle );

       DBGOUT;
       return ( rc );
       }
     else if (rc != NO_ERROR)
       {
       MSGALL(( ERR_CANT_SEEK_ON_DEVSPACE, pszLogDevName, rc ));
       DBG1  (( MF__, "Cannot seek on devspace: %s, rc = %d", pszLogDevName,
                rc ));
       sql46c_build_error_string ( errtext, ERRMSG_DEVIO_SEEK_ON_DEVSPACE, rc );

       CloseHandle( hfFileHandle );

       DBGOUT;
       return ( rc );
       }

     if (SetEndOfFile(hfFileHandle ) != true)
       {
       rc = GetLastError();

       if (rc == ERROR_DISK_FULL)
         {
         MSGALL(( ERR_DISK_FULL, pszLogDevName ));
         DBG1  (( MF__, "Could not create: '%s', not enough space on disk",
                  pszLogDevName ));
         sql46c_build_error_string ( errtext, ERRMSG_DEVIO_DISK_FULL, 0 );
         }
       else
         {
         MSGALL(( ERR_CANT_SEEK_ON_DEVSPACE, pszLogDevName, rc ));
         DBG1  (( MF__, "Cannot seek on devspace: %s, rc = %d", pszLogDevName,
                  rc ));
         sql46c_build_error_string ( errtext, ERRMSG_DEVIO_SEEK_ON_DEVSPACE, rc );
         }

       CloseHandle( hfFileHandle );

       DBGOUT;
       return ( rc );
       }
     }

    rc = sql54k_fill_block_0 ( (HFILE)hfFileHandle, pszLogDevName,
                               *devcapacity + 1, *devcapacity + 1, errtext );

    if ( rc != NO_ERROR )
      {
      CloseHandle( hfFileHandle );

      DBGOUT;
      return ( rc );
      }

   if (!CloseHandle( hfFileHandle ))
     {
     rc = GetLastError();

     MSGALL(( ERR_CANT_CLOSE_DEVSPACE, rc ));
     DBG1  (( MF__, "Cannot close devspace, rc = %d", rc ));
     sql46c_build_error_string ( errtext, ERRMSG_DEVIO_CLOSE_DEVSPACE, rc );

     DBGOUT;
     return ( rc );
     }

   rc = sql49c_set_file_security ( pszPhysDevName,
                                   kgs.AdminSA.lpSecurityDescriptor );

   if ( rc != NO_ERROR )
     {
     rc = GetLastError();
     sql46c_build_error_string ( errtext, ERRMSG_DEVIO_SET_DEVSPACE_OWNER, rc );

     DBGOUT;
     return ( rc );
     }

   *ok = true;

   DBGOUT;
   return (NO_ERROR);
   }

#else /*------------------------------*/

 /*
  * Vdevsize for OS/2
  */

 static int sql54k_vdevsize( PSZ            pszPhysDevName,
                             PSZ            pszLogDevName,
                             INT4           *devcapacity,
                             ULONG          ulCurrFileSizeLow,
                             ULONG          ulCurrFileSizeHigh,
                             ERRORTEXT      errtext,
                             BOOLEAN        *ok )
   {
   #undef  MF__
   #define MF__  MOD__"sql54k_vdevsize"

   ULONG          Action;
   FILESTATUS3    FileInfoBuf;
   HFILE          hfFileHandle;
   APIRET         rc = NO_ERROR;
   LONG           devcap;
   PUKT_CTRL_REC  pUKT  = THIS_UKT_CTRL;

   DBGIN;
   DBG3 ((MF__, "devspace: %s", pszLogDevName));

   *ok = false;

   if (*devcapacity == 0)
     {
     DBG3 ((MF__, "Opening devspace: %s", pszLogDevName));

     rc = DosOpen( pszPhysDevName,                 /* File path name */
                   &hfFileHandle,                  /* File handle */
                   &Action,                        /* Action taken */
                   0L,                             /* File primary allocation */
                   FILE_NORMAL,                    /* File attribute */
                   OPEN_ACTION_OPEN_IF_EXISTS |    /* Open function type */
                   OPEN_ACTION_FAIL_IF_NEW,
                   OPEN_FLAGS_FAIL_ON_ERROR   |
                   OPEN_SHARE_DENYWRITE       |
                   OPEN_ACCESS_READWRITE,
                   NOEABUF);                       /* No extended attributes */

     if (rc == ERROR_DISK_FULL)
       {
       MSGALL(( ERR_DISK_FULL, pszLogDevName ));
       DBG1  (( MF__, "Could not create: '%s', not enough space on disk",
                pszLogDevName ));
       sql46c_build_error_string ( errtext, ERRMSG_DEVIO_DISK_FULL, 0 );

       DBGOUT;
       return ( rc );
       }
     else if (rc != NO_ERROR)
       {
       MSGALL(( ERR_CANT_OPEN_DEVSPACE, pszLogDevName, rc ));
       DBG1  (( MF__, "Cannot open devspace '%s', rc = %d", pszLogDevName, rc ));
       sql46c_build_error_string ( errtext, ERRMSG_DEVIO_OPEN_DEVSPACE, rc );

       DBGOUT;
       return ( rc );
       }

     rc = DosQueryFileInfo(hfFileHandle, 1, &FileInfoBuf, sizeof(FILESTATUS3));

     if (rc != NO_ERROR)
       {
       MSGALL(( ERR_DURING_Q_FILE_SIZE, rc ));
       DBG1  (( MF__, "Error during DosQueryFileInfo, rc = %d", rc ));
       sql46c_build_error_string ( errtext, ERRMSG_DEVIO_DURING_Q_FILE_SIZE, rc );

       CLOSE_FILE( hfFileHandle );

       DBGOUT;
       return ( rc );
       }

     *devcapacity = BYTES_TO_PAGES(FileInfoBuf.cbFile) - 1;

     if ((*devcapacity < 1) || ((ULONG)*devcapacity > MAX_DEVSPACE_PAGES))
       {
       MSGALL((ERR_INV_DEVSPACE_SIZE, pszLogDevName, *devcapacity ));
       DBG1  ((MF__, "Invalid devspace size: %u pages", *devcapacity ));
       sql46c_build_error_string ( errtext, ERRMSG_DEVIO_INV_DEVSPACE_SIZE, 0 );

       CLOSE_FILE( hfFileHandle );

       DBGOUT;
       return ( -1 );
       }
     }
   else
     {
     if ((*devcapacity < 1) || ((ULONG)*devcapacity > MAX_DEVSPACE_PAGES))
       {
       MSGALL((ERR_INV_DEVSPACE_SIZE, pszLogDevName, *devcapacity ));
       DBG1  ((MF__, "Invalid devspace size: %u pages", *devcapacity ));
       sql46c_build_error_string ( errtext, ERRMSG_DEVIO_INV_DEVSPACE_SIZE, 0 );

       DBGOUT;
       return ( -1 );
       }

     devcap = (LONG)PAGES_TO_BYTES(*devcapacity + 1);

     DBG3 (( MF__, "Creating devspace: %s - capacity: (%u + 1) * 4096 = %u",
             pszLogDevName, *devcapacity, devcap ));

     rc = DosOpen( pszPhysDevName,
                   &hfFileHandle,
                   &Action,
                   devcap,
                   FILE_NORMAL,
                   OPEN_ACTION_CREATE_IF_NEW  |
                   OPEN_ACTION_REPLACE_IF_EXISTS,
                   OPEN_FLAGS_FAIL_ON_ERROR   |
                   OPEN_SHARE_DENYWRITE       |
                   OPEN_ACCESS_READWRITE,
                   NOEABUF);

     if (rc == ERROR_DISK_FULL)
       {
       MSGALL(( ERR_DISK_FULL, pszLogDevName ));
       DBG1  (( MF__, "Could not create: '%s', not enough space on disk",
                pszLogDevName ));
       sql46c_build_error_string ( errtext, ERRMSG_DEVIO_DISK_FULL, 0 );

       DBGOUT;
       return ( rc );
       }
     else if ( rc != NO_ERROR )
       {
       MSGALL(( ERR_CANT_CREATE_DEVSPACE, pszLogDevName, rc ));
       DBG1  (( MF__, "Cannot create devspace '%', rc = %d", pszLogDevName, rc ));
       sql46c_build_error_string ( errtext, ERRMSG_DEVIO_CREATE_DEVSPACE, rc );

       DBGOUT;
       return ( rc );
       }
     }

   if ( XPARAM(fFormatDevspace) == true )
     {
     // --- set the thread priority
     rc = SET_THREAD_PRIO ( (PRTYC_IDLETIME * 100) + 31,
                            pUKT->ThrdCtrlHeader.Tid,
                            pUKT->ThrdCtrlHeader.hThrd );

     if ( rc != NO_ERROR )
        {
        MSGD (( ERR_SETTING_THRD_PRIO, pUKT->ThrdCtrlHeader.Tid));
        DBG1 ((MF__, "Cannot set the thread priority, TID: %u, rc=%d",
               pUKT->ThrdCtrlHeader.Tid, rc ));
        DBGOUT;
        return ( rc );
        }

     rc = sql54k_format_devspace ( hfFileHandle, pszLogDevName,
                                   ulCurrFileSizeLow, ulCurrFileSizeHigh,
                                   *devcapacity + 1,
                                   pUKT->pCTask,
                                   errtext );
     if ( rc != NO_ERROR )
       {
       CLOSE_FILE( hfFileHandle );

       SET_THREAD_PRIO ( pUKT->ThrdCtrlHeader.lThrdPrio,
                         pUKT->ThrdCtrlHeader.Tid,
                         pUKT->ThrdCtrlHeader.hThrd );

       DBGOUT;
       return ( rc );
       }

     // --- set the thread priority
     rc = SET_THREAD_PRIO ( pUKT->ThrdCtrlHeader.lThrdPrio,
                            pUKT->ThrdCtrlHeader.Tid,
                            pUKT->ThrdCtrlHeader.hThrd );

     if ( rc != NO_ERROR )
        {
        MSGD (( ERR_SETTING_THRD_PRIO, pUKT->ThrdCtrlHeader.Tid));
        DBG1 ((MF__, "Cannot set the thread priority, TID: %u, rc=%d",
               pUKT->ThrdCtrlHeader.Tid, rc ));
        DBGOUT;
        return ( rc );
        }
     }

   rc = sql54k_fill_block_0 ( hfFileHandle, pszLogDevName,
                              *devcapacity + 1, *devcapacity + 1, errtext );

   if ( rc != NO_ERROR )
     {
     CLOSE_FILE( hfFileHandle );

     DBGOUT;
     return ( rc );
     }

   rc = CLOSE_FILE( hfFileHandle );

   if (rc != NO_ERROR)
     {
     MSGALL(( ERR_CANT_CLOSE_DEVSPACE, rc ));
     DBG1  (( MF__, "Cannot close devspace, rc = %d", rc ));
     sql46c_build_error_string ( errtext, ERRMSG_DEVIO_CLOSE_DEVSPACE, rc );

     DBGOUT;
     return ( rc );
     }

   *ok = true;

   DBGOUT;
   return (NO_ERROR);
   }

#endif


/*------------------------------*/

/*
 * Vdevsize for a device being a drive for direct io.
 */
static int sql54k_vdevsize_drive ( PSZ            pszDriveName,
                                   INT4           *devcapacity,
                                   ERRORTEXT      errtext,
                                   BOOLEAN        *ok )
  {
  #undef  MF__
  #define MF__  MOD__"sql54k_vdevsize_drive"

  #if !defined(_WIN32)
    ULONG      Action;
  #endif
  HFILE        hfFileHandle;
  ULONG        ulDummy;
  ULONG        ulMaxPages;
  APIRET
   rc = NO_ERROR;                      /* Return code */

  DBGIN;

  *ok = false;

  DBG3 ((MF__, "Opening devspace: %s", pszDriveName));

  #if defined(_WIN32)
   hfFileHandle = (HFILE)CreateFile( pszDriveName,
                                     GENERIC_READ      |
                                     GENERIC_WRITE,
                                     FILE_SHARE_READ   |
                                     FILE_SHARE_WRITE,
                                     NULL,
                                     OPEN_EXISTING,
                                     FILE_ATTRIBUTE_NORMAL,
                                     NULL );

   if (hfFileHandle == (HFILE)INVALID_HANDLE_VALUE)
     rc = GetLastError();
  #else
   rc = DosOpen( pszDriveName,                 /* File path name */
                 &hfFileHandle,                /* File handle */
                 &Action,                      /* Action taken */
                 0L,                           /* File primary allocation */
                 FILE_NORMAL,                  /* File attribute */
                 OPEN_ACTION_OPEN_IF_EXISTS |  /* Open function type */
                 OPEN_ACTION_FAIL_IF_NEW,
                 OPEN_FLAGS_FAIL_ON_ERROR   |  /* Open mode of the file */
                 OPEN_FLAGS_DASD            |
                 OPEN_SHARE_DENYNONE        |
                 OPEN_ACCESS_READWRITE,
                 NOEABUF);                     /* No extended attributes */
  #endif

  if (rc == ERROR_DRIVE_LOCKED)
    {
    MSGALL (( ERR_DEVSPACE_LOCKED, pszDriveName ));
    DBG1   ( (MF__, "Devspace '%s' locked by another process", pszDriveName ));
    sql46c_build_error_string ( errtext, ERRMSG_DEVIO_DEVSPACE_LOCKED, 0 );

    DBGOUT;
    return ( rc );
    }


  if ( rc == ERROR_ACCESS_DENIED )
    {
    MSGALL(( ERR_DEVSPACE_ACCESS_DENIED, pszDriveName ));
    DBG1  (( MF__, "Access on devspace '%s' denied", pszDriveName ));
    sql46c_build_error_string ( errtext, ERRMSG_DEVIO_ACCESS_DENIED, 0 );

    DBGOUT;
    return ( rc );
    }
  else if ( rc != NO_ERROR )
    {
    MSGALL (( ERR_CANT_OPEN_DEVSPACE, pszDriveName, rc ));
    DBG1   ( (MF__, "Cannot open devspace '%s', rc = %d", pszDriveName, rc ));
    sql46c_build_error_string ( errtext, ERRMSG_DEVIO_OPEN_DEVSPACE, rc );

    DBGOUT;
    return ( rc );
    }


  rc = sql54k_lock_drive ( hfFileHandle );

  if (rc == ERROR_DRIVE_LOCKED)
    {
    MSGALL (( ERR_DEVSPACE_LOCKED, pszDriveName ));
    DBG1   ( (MF__, "Devspace '%s' locked by another process", pszDriveName ));
    sql46c_build_error_string ( errtext, ERRMSG_DEVIO_DEVSPACE_LOCKED, 0 );

    DBGOUT;
    return ( rc );
    }

  if (rc != 0)
    {
    MSGALL (( ERR_CANT_LOCK_DEVSPACE, pszDriveName, rc ));
    DBG1   ( (MF__, "Cannot lock devspace '%s', rc = %d", pszDriveName, rc ));
    sql46c_build_error_string ( errtext, ERRMSG_DEVIO_LOCK_DEVSPACE, rc );

    DBGOUT;
    return ( rc );
    }

  rc = sql54k_evaluate_block_0 ( hfFileHandle,
                                 &ulDummy,
                                 &ulMaxPages,
                                 XPARAM(ulDBPageSize),
                                 true,
                                 pszDriveName );

  if ( rc == NO_ERROR )
    {
    DBG3  (( MF__, "ulMaxPages: %d", ulMaxPages ));

    if (*devcapacity == 0)
      {
      *devcapacity = ulMaxPages - 1;

      if (*devcapacity < 1)
        {
        MSGALL(( ERR_INV_DEVSPACE_SIZE, pszDriveName, *devcapacity ));
        DBG1  (( MF__, "Devspace '%s', invalid size: %u pages", pszDriveName,
                 *devcapacity ));
        sql46c_build_error_string ( errtext, ERRMSG_DEVIO_INV_DEVSPACE_SIZE, rc );
        }
      else
        {
        MSGD (( INFO_DEVSPACE_SIZE, pszDriveName, *devcapacity ));
        DBG3 (( MF__, "Devspace '%s', calculated size: %d pages",
                pszDriveName, *devcapacity ));

        rc = sql54k_fill_block_0 ( hfFileHandle, pszDriveName,
                                   ulMaxPages, ulMaxPages, errtext );
        if ( rc == NO_ERROR )
          *ok = true;
        else
          *ok = false;
        }
      }
    else
      {
      if ((ULONG) *devcapacity > (ulMaxPages - 1))
        {
        MSGD ((INFO_DEV_SIZE_EXC_MAX,
               pszDriveName, (ulMaxPages - 1) ));
        DBG1 ((MF__, "Devspace '%s', requested size exceeds maximum ( %d ).",
               pszDriveName, (ulMaxPages - 1) ));

        *devcapacity  = ulMaxPages - 1;
        }
      else if ( (ULONG) *devcapacity < (ulMaxPages - 1) )
        {
        MSGD (( INFO_DEV_SIZE_LOV_MAX,
                pszDriveName, *devcapacity, (ulMaxPages - 1) ));
        DBG1 (( MF__, "Devspace '%s', requested size ( %d ) is lover than maximum ( %d ).",
                pszDriveName, *devcapacity, (ulMaxPages - 1) ));
        }

      rc = sql54k_fill_block_0 ( hfFileHandle, pszDriveName,
                                 *devcapacity + 1, ulMaxPages, errtext );
      if ( rc == NO_ERROR )
        *ok = true;
      else
        *ok = false;
      }
    }
  else
    sql46c_build_error_string ( errtext, ERRMSG_DEVIO_OPEN_DEVSPACE, 0 );


  if ( sql54k_dev_close ( hfFileHandle, pszDriveName, DT_DRIVE, errtext) != NO_ERROR )
    *ok = false;

  DBGOUT;
  return(0);
  }


/*------------------------------*/

/*
 * initialize block 0 of a devspace
 */

static APIRET  sql54k_fill_block_0 ( HFILE      hfDevspace,
                                     PSZ        pszLogDevName,
                                     ULONG      ulDevPages,
                                     ULONG      ulMaxDevPages,
                                     ERRORTEXT  errtext )
  {
  #undef  MF__
  #define MF__   MOD__"sql54k_fill_block_0"
  ULONG   ulFilePosLow;
  PCHAR   pTmpBuffer;
  ULONG   ulWritten;
  ULONG   ulDBPageSize = XPARAM(ulDBPageSize) ;
  PCHAR   pBuffer      = NULL;
  APIRET  rc           = NO_ERROR;

  DBGIN;

  rc = ALLOC_MEM ((PPVOID)&pBuffer, ulDBPageSize + 
                  kgs.ulSystemPageSize /* alignment */ );

  if( rc != NO_ERROR )
    {
    DBGOUT;
    return ( rc );
    }

  // --- alignment needed for none cached file operations
  pTmpBuffer = (PCHAR)ALIGN( (ULONG)pBuffer, kgs.ulSystemPageSize );

  rc = SET_FILE_PTR ( hfDevspace, 0, FILE_BEGIN, &ulFilePosLow, NULL );

  if ( rc == ERROR_DISK_FULL )
    {
    MSGALL(( ERR_DISK_FULL, pszLogDevName ));
    DBG1  (( MF__, "Could not create: '%s', not enough space on disk",
             pszLogDevName ));
    sql46c_build_error_string ( errtext, ERRMSG_DEVIO_DISK_FULL, 0 );
    FREE_MEM ( (PVOID)pBuffer );

    DBGOUT;
    return ( rc );
    }
  else if ( rc != NO_ERROR )
    {
    MSGALL(( ERR_CANT_SEEK_ON_DEVSPACE, pszLogDevName, rc ));
    DBG1  (( MF__, "Cannot seek on devspace: %s, rc = %d", pszLogDevName, rc ));
    sql46c_build_error_string ( errtext, ERRMSG_DEVIO_SEEK_ON_DEVSPACE, rc );
    FREE_MEM ( (PVOID)pBuffer );

    DBGOUT;
    return ( rc );
    }

  memset (pTmpBuffer, '\0', ulDBPageSize);

  // --- fill buffer
  //
  //     0x12345678 ->  pTmpBuffer[2] = 0x56
  //                    pTmpBuffer[3] = 0x78
  //                    pTmpBuffer[4] = 0x12
  //                    pTmpBuffer[5] = 0x34
  //
  pTmpBuffer[2] = (CHAR) (( ulDBPageSize % KBYTE_64 ) / 256);
  pTmpBuffer[3] = (CHAR) (( ulDBPageSize % KBYTE_64 ) % 256);
  pTmpBuffer[4] = (CHAR) (( ulDBPageSize / KBYTE_64 ) / 256);
  pTmpBuffer[5] = (CHAR) (( ulDBPageSize / KBYTE_64 ) % 256);

  memcpy ( pTmpBuffer + PRODUCT_ID_POS,
           PRODUCT_ID_STRING,
           sizeof (PRODUCT_ID_STRING) - 1);

  memcpy ( pTmpBuffer + DEVSPACE_ID_POS,
           DEVSPACE_ID_STRING,
           sizeof (DEVSPACE_ID_STRING) - 1);

  *(ULONG*)(&pTmpBuffer[ DEVSPACE_SIZE_POS ])    = ulDevPages;
  *(ULONG*)(&pTmpBuffer[ DEVSPACE_MAXSIZE_POS ]) = ulMaxDevPages;

  DBG2 ((MF__, "Block 0 contents", pTmpBuffer, ulDBPageSize));

  rc = WRITE_FILE (hfDevspace, pTmpBuffer, ulDBPageSize, &ulWritten);

  if (rc != NO_ERROR  || ulWritten != ulDBPageSize)
    {
    MSGD  (( ERR_CANT_WRITE_ON_DEVSPACE, 0, pszLogDevName, rc, ulWritten ));
    DBG1 ((MF__, "Cannot write on devspace '%s', rc = %d", pszLogDevName, rc ));
    sql46c_build_error_string ( errtext, ERRMSG_DEVIO_WRITE_ON_DEVSPACE, rc );
    }

  rc = FREE_MEM ( (PVOID)pBuffer );

  if( rc != NO_ERROR )
    {
    DBGOUT;
    return ( rc );
    }

  DBGOUT;

  return (rc);
  }

/*------------------------------*/

static APIRET  sql54k_format_devspace ( HFILE            hfDevspace,
                                        PSZ              pszLogDevName,
                                        ULONG            ulCurrFileSizeLow,
                                        ULONG            ulCurrFileSizeHigh,
                                        ULONG            ulDevPages,
                                        PTASK_CTRL_REC   pTaskCtrl,
                                        ERRORTEXT        errtext )
  {
  #undef  MF__
  #define MF__   MOD__"sql54k_format_devspace"
  ULONG   ulPageNumber;
  ULONG   ulFilePosHigh;
  ULONG   ulFilePosLow;
  PCHAR   pTmpBuffer;
  ULONG   ulWritten;
  ULONG   ulDBPageSize = XPARAM(ulDBPageSize);
  ULONG   ulBuffPages;
  PCHAR   pBuffer      = NULL;
  APIRET  rc           = NO_ERROR;

  DBGIN;

  ulPageNumber = QWORD_TO_PAGES( ulCurrFileSizeLow, ulCurrFileSizeHigh );

  if ( ulPageNumber < ulDevPages )
    {
    ulBuffPages = (32 KB) / ulDBPageSize;

    if ( ulBuffPages == 0 ) ulBuffPages = 1;

    rc = ALLOC_MEM ((PPVOID)&pBuffer,
                    (ulDBPageSize * ulBuffPages) + 
                    kgs.ulSystemPageSize /* alignment */ );


    if( rc != NO_ERROR )
      {
      DBGOUT;
      return ( rc );
      }

    // --- alignment needed for none cached file operations
    pTmpBuffer = (PCHAR)ALIGN( (ULONG)pBuffer, kgs.ulSystemPageSize );

    memset ( pTmpBuffer, '\0', (ulDBPageSize * ulBuffPages) );

    PAGES_TO_QWORD( ulFilePosLow, ulFilePosHigh, ulPageNumber );

    rc = SET_FILE_PTR( hfDevspace, ulFilePosLow, FILE_BEGIN,
                       &ulFilePosLow, &ulFilePosHigh );

    if ( rc == ERROR_DISK_FULL )
      {
      MSGALL(( ERR_DISK_FULL, pszLogDevName ));
      DBG1  (( MF__, "Could not create: '%s', not enough space on disk",
               pszLogDevName ));
      sql46c_build_error_string ( errtext, ERRMSG_DEVIO_DISK_FULL, 0 );
      FREE_MEM ( (PVOID)pBuffer );

      DBGOUT;
      return ( rc );
      }
    else if ( rc != NO_ERROR )
      {
      MSGALL(( ERR_CANT_SEEK_ON_DEVSPACE, pszLogDevName, rc ));
      DBG1  (( MF__, "Cannot seek on devspace: %s, rc = %d", pszLogDevName, rc ));
      sql46c_build_error_string ( errtext, ERRMSG_DEVIO_SEEK_ON_DEVSPACE, rc );
      FREE_MEM ( (PVOID)pBuffer );

      DBGOUT;
      return ( rc );
      }

    for ( ; ulPageNumber + ulBuffPages <= ulDevPages;
            ulPageNumber += ulBuffPages )
      {
      pTaskCtrl->TaskStateStat.DevIOWriteStat.ulCount += ulBuffPages;
      EO52_TAKE_REL_ABS_TIME( pTaskCtrl, pTaskCtrl->TaskStateStat.DevIOWriteStat,
                  rc = WRITE_FILE ( hfDevspace, pTmpBuffer,
                                    ulDBPageSize * ulBuffPages,
                                    &ulWritten ) );

      if (rc == ERROR_DISK_FULL)
        {
        MSGALL(( ERR_DISK_FULL, pszLogDevName ));
        DBG1  (( MF__, "Could not create: '%s', not enough space on disk",
                 pszLogDevName ));
        sql46c_build_error_string ( errtext, ERRMSG_DEVIO_DISK_FULL, 0 );
        FREE_MEM ( (PVOID)pBuffer );

        DBGOUT;
        return ( rc );
        }
      else if (rc        != NO_ERROR  ||
               ulWritten != (ulDBPageSize * ulBuffPages))
        {
        MSGD  (( ERR_CANT_WRITE_ON_DEVSPACE, ulPageNumber, pszLogDevName,
                 rc, ulWritten ));
        DBG1 ((MF__, "Cannot write on devspace '%s', rc = %d", pszLogDevName, rc ));
        sql46c_build_error_string ( errtext, ERRMSG_DEVIO_WRITE_ON_DEVSPACE, rc );
        FREE_MEM ( (PVOID)pBuffer );

        DBGOUT;
        return ( rc );
        }
      }

    if ( ulPageNumber < ulDevPages )
      {
      pTaskCtrl->TaskStateStat.DevIOWriteStat.ulCount +=
                                                    (ulDevPages - ulPageNumber);
      EO52_TAKE_REL_ABS_TIME( pTaskCtrl, pTaskCtrl->TaskStateStat.DevIOWriteStat,
                 rc = WRITE_FILE (hfDevspace, pTmpBuffer,
                                  ulDBPageSize *
                                   (ulDevPages - ulPageNumber),
                                  &ulWritten));

      if (rc == ERROR_DISK_FULL)
        {
        MSGALL(( ERR_DISK_FULL, pszLogDevName ));
        DBG1  (( MF__, "Could not create: '%s', not enough space on disk",
                 pszLogDevName ));
        sql46c_build_error_string ( errtext, ERRMSG_DEVIO_DISK_FULL, 0 );
        FREE_MEM ( (PVOID)pBuffer );

        DBGOUT;
        return ( rc );
        }
      else if (rc        != NO_ERROR  ||
               ulWritten != ulDBPageSize * (ulDevPages - ulPageNumber))
        {
        MSGD  (( ERR_CANT_WRITE_ON_DEVSPACE, ulPageNumber, pszLogDevName,
                 rc, ulWritten ));
        DBG1 ((MF__, "Cannot write on devspace '%s', rc = %d", pszLogDevName, rc ));
        sql46c_build_error_string ( errtext, ERRMSG_DEVIO_WRITE_ON_DEVSPACE, rc );
        FREE_MEM ( (PVOID)pBuffer );

        DBGOUT;
        return ( rc );
        }
      }

    rc = FREE_MEM ( (PVOID)pBuffer );

    if( rc != NO_ERROR )
      {
      DBGOUT;
      return ( rc );
      }
    }

  DBGOUT;
  return (rc);
  }

/*------------------------------*/

static LONG  sql54k_evaluate_block_0  ( HFILE         hfDevspace,
                                        PULONG        pulDevPages,
                                        PULONG        pulMaxDevPages,
                                        ULONG         ulDBPageSize,
                                        BOOL          fDevspaceIsDrive,
                                        PSZ           pszLogDevName )
  {
  #undef  MF__
  #define MF__   MOD__"sql54k_evaluate_block_0"
  ULONG                ulFilePosLow;
  BOOL                 fDevspaceIDOk;
  LONG                 rc      = NO_ERROR;
  PCHAR                pBuffer = NULL;
  ULONG                ulRead;
  PCHAR                pTmpBuffer;
  tsp_page_no          BadDevPos;
  #if defined(_WIN32)
   ULONG               ulFileSizeLow;
   ULONG               ulFileSizeHigh;
  #else
   FILESTATUS3         FileInfoBuf;
  #endif

  DBGIN;

  *pulDevPages = 0;

  rc = ALLOC_MEM ((PPVOID)&pBuffer,
                  ulDBPageSize +
                  kgs.ulSystemPageSize /* alignment */ );

  if( rc != NO_ERROR )
    {
    DBGOUT;
    return ( rc );
    }

  // --- alignment needed for none cached file operations
  pTmpBuffer = (PCHAR)ALIGN( (ULONG)pBuffer, kgs.ulSystemPageSize );

  // --- no devspace file handle defined?
  rc = SET_FILE_PTR ( hfDevspace, 0, FILE_BEGIN, &ulFilePosLow, NULL );

  if ( rc != NO_ERROR )
    {
    MSGALL(( ERR_CANT_SEEK_ON_DEVSPACE, pszLogDevName, rc ));
    DBG1  (( MF__, "Cannot seek on devspace: %s, rc = %d", pszLogDevName, rc ));
    FREE_MEM ( (PVOID)pBuffer );
    DBGOUT;
    return ( rc );
    }

  /* PTS 1106176 */
  rc = sql54k_ReadBlockWithRetry( hfDevspace, (BUFFERADDRESS )pTmpBuffer, ulDBPageSize, &ulRead, 
                                  0, pszLogDevName );

  if (( rc == ERROR_SHARING_VIOLATION ) || ( rc == ERROR_LOCK_VIOLATION ))
    {
    MSGALL(( ERR_DEVSPACE_IN_USE, pszLogDevName ));
    DBG1  (( MF__, "Devspace '%s' is already used by another process, rc = %u", pszLogDevName, rc ));
    FREE_MEM ( (PVOID)pBuffer );
    DBGOUT;
    return (rc);
    }
  else if (rc != NO_ERROR)
    {
    MSGALL(( ERR_CANT_READ_FROM_DEVSPACE, 0, pszLogDevName, rc, ulRead ));
    DBG1  (( MF__, "Cannot read from devspace: '%s', rc = %d", pszLogDevName, rc ));
    FREE_MEM ( (PVOID)pBuffer );
    DBGOUT;
    return (rc);
    }

  DBG2 ((MF__, "Block 0 contents", pTmpBuffer, ulDBPageSize));

  fDevspaceIDOk = !strncmp ( pTmpBuffer + DEVSPACE_ID_POS,
                             DEVSPACE_ID_STRING,
                             sizeof (DEVSPACE_ID_STRING) - 1 );

  if ((ulRead != ulDBPageSize) || (fDevspaceIDOk == false))
    {
    MSGALL(( ERR_NOT_A_VALID_DEVSPACE, pszLogDevName ));
    DBG1  (( MF__, "It's not a valid devspace: '%s'", pszLogDevName ));
    DBG1  (( MF__, "ulRead: %d - fDevspaceIDOk: %f ", ulRead, fDevspaceIDOk ));
    FREE_MEM ( (PVOID)pBuffer );
    DBGOUT;
    return ( -1 );
    }

  if ( fDevspaceIsDrive )
    {
    *pulDevPages    = *(ULONG*)(&pTmpBuffer[ DEVSPACE_SIZE_POS ]);
    *pulMaxDevPages = *(ULONG*)(&pTmpBuffer[ DEVSPACE_MAXSIZE_POS ]);
    }
  else
    {
    #if defined(_WIN32)
     ulFileSizeLow = (ULONG)GetFileSize((HANDLE)hfDevspace, &ulFileSizeHigh);

     if ((ulFileSizeLow == MAX_DWORD_VAL) && ((rc = GetLastError()) != NO_ERROR))
       {
       MSGALL(( ERR_DURING_Q_FILE_SIZE, rc ));
       DBG1  (( MF__, "Error during GetFileSize, rc = %d \n", rc ));
       }
     else
       {
       if (HIGH_VALUE_VALID(ulFileSizeHigh) == true)
         {
         *pulDevPages    = QWORD_TO_PAGES( ulFileSizeLow, ulFileSizeHigh );
         *pulMaxDevPages = QWORD_TO_PAGES( ulFileSizeLow, ulFileSizeHigh );
         }
       else
         {
         MSGALL((ERR_INV_DEVSPACE_SIZE, pszLogDevName, MAX_DEVSPACE_PAGES));
         DBG1  ((MF__, "Invalid devspace size: %u pages", MAX_DEVSPACE_PAGES));
         FREE_MEM ( (PVOID)pBuffer );
         DBGOUT;
         return ( -1 );
         }
       }
    #else
     rc = DosQueryFileInfo( hfDevspace, 1, &FileInfoBuf, sizeof(FILESTATUS3) );

     if ( rc != NO_ERROR )
       {
       MSGALL(( ERR_DURING_Q_FILE_SIZE, rc ));
       DBG1  (( MF__, "Error during DosQueryFileInfo, rc = %d \n", rc ));
       }
     else
       {
       *pulDevPages    = PAGES_TO_BYTES(FileInfoBuf.cbFile);
       *pulMaxDevPages = *pulDevPages;
       }
    #endif
    }

  memcpy ( &BadDevPos, pTmpBuffer + DEVSPACE_BAD_MARK_POS, sizeof(BadDevPos) );


  if ( BadDevPos != 0 )
    {
    rc = !NO_ERROR ;
    MSGALL(( ERR_DEVSP_MARKED_AS_BAD, pszLogDevName, BadDevPos ));
    FREE_MEM ( (PVOID)pBuffer );
    DBGOUT;
    return ( rc );
    }

  rc = FREE_MEM ( (PVOID)pBuffer );

  if( rc != NO_ERROR )
    {
    DBGOUT;
    return ( rc );
    }

  DBGOUT;
  return (rc);
  }

/*------------------------------*/

static PDEVI_CTRL_REC sql54k_find_devi ( LONG    lDevspaceNo,
                                         PULONG  pulQueueLength )
  {
  #undef  MF__
  #define MF__ MOD__"sql54k_find_devi"
  PDEVI_CTRL_REC                    pFirstDeviCtrl;
  PDEVI_CTRL_REC                    pDeviCtrl;
  ULONG                             ulSlot,
                                    ulSlotReminder,
                                    ulMinQueueReminder;
  #ifdef DEBUG
   PTASK_CTRL_REC                   pTaskCtrl = THIS_UKT_CTRL->pCTask;
  #endif

  DBGIN_T (pTaskCtrl->ulTaskIndex);

  DBG3 (( MF__, "[T:0x%03u] Find slot for despace %d",
          pTaskCtrl->ulTaskIndex, lDevspaceNo ));

  pFirstDeviCtrl = kgs.pFirstDeviCtrl + lDevspaceNo * XPARAM(lDeviPerDevspace);

  // --- is there only one DEVi?
  if (( XPARAM(lDeviPerDevspace)  == 1 ) ||
      ( kgs.Dev0.lTraceDevspaceNo == lDevspaceNo ))
    {
    DBG3 (( MF__, "[T:0x%03u] Threre is only one DEVi, devspace  %d",
            pTaskCtrl->ulTaskIndex, lDevspaceNo ));

    *pulQueueLength = pFirstDeviCtrl->DevIOStat.ulIOQueLen;

    DBGOUT_T (pTaskCtrl->ulTaskIndex);
    return ( pFirstDeviCtrl );
    }

  /*
   *  Find a sleeping DEVi Process
   */
  ulSlotReminder     = 0;
  ulMinQueueReminder = (ULONG)-1;

  for ( ulSlot = 0, pDeviCtrl = pFirstDeviCtrl;
        ulSlot < (ULONG) XPARAM(lDeviPerDevspace);
        ulSlot++, pDeviCtrl++ )
    {
    if ( pDeviCtrl->DevIOStat.ulIOQueLen <= (ULONG )XPARAM(lDevQueue) )
      {
      DBG3 (( MF__, "[T:0x%03u] Found sleeping DEVi 0x%x, using slot %u",
              pTaskCtrl->ulTaskIndex, pDeviCtrl, ulSlot ));

      *pulQueueLength = pDeviCtrl->DevIOStat.ulIOQueLen;

      DBGOUT_T (pTaskCtrl->ulTaskIndex);
      return ( pDeviCtrl );
      }

    if (ulMinQueueReminder > (ULONG)pDeviCtrl->DevIOStat.ulIOQueLen)
      {
      // --- set the the reminder to the DEVi with smallest queue length
      ulMinQueueReminder = (ULONG)pDeviCtrl->DevIOStat.ulIOQueLen;
      ulSlotReminder     = ulSlot;
      }
    }

  //
  // --- no sleeping DEVi found. Use the DEVi with the smallest queue length
  //
  *pulQueueLength = ulMinQueueReminder;
  DBG3 (( MF__, "[T:0x%03u] No sleeping DEVi for Devspace No.%u found. Using Slot %u",
                 pTaskCtrl->ulTaskIndex, lDevspaceNo, ulSlotReminder));

  DBGOUT_T (pTaskCtrl->ulTaskIndex);
  return (pFirstDeviCtrl + ulSlotReminder);
  }


/*------------------------------*/

/*
 * find a free devspace number
 */
static BOOL sql54k_find_devspace_no ( PLONG          plDevspaceNo,
                                      PSZ            pszLogDevName,
                                      PULONG         pulErrCode )
  {
  #undef  MF__
  #define MF__ MOD__"sql54k_find_devspace_no"
  LONG                             lDevspaceNo      = 0,
                                   lFirstFree       = -1;
  BOOL                             fAlreadyAttached = false,
                                   fOK;
  PDEV_IO_CTRL_REC                 pIOCtrl;

  DBGIN;

  /*
   * Check and search a free DEVi control record
   */
  while ( lDevspaceNo < XPARAM(lMaxDevspaces)  && !fAlreadyAttached)
    {
    pIOCtrl = &kgs.Dev0.pIOCtrl[lDevspaceNo];

    if ( pIOCtrl->ulAttached )
      {
      fAlreadyAttached = !strcmp( pIOCtrl->szDevspace, pszLogDevName );
      }
    else if ( lFirstFree == -1 )
      lFirstFree = lDevspaceNo;

    lDevspaceNo++;
    }

  if ( fAlreadyAttached )
    {
    MSGALL(( ERR_DEVSPACE_ALREADY_ATT, pszLogDevName ));
    DBG1  (( MF__, "Devspace '%s' already attached", pszLogDevName ));
    *pulErrCode = ERRCODE_DEVIO_ALRATT;

    fOK = false;
    *plDevspaceNo = -1;
    }
  else
    {
    if (lFirstFree != -1)
      {
      DBG3 (( MF__, "First free devspace no. %d", lFirstFree ));
      fOK = true;
      *plDevspaceNo = lFirstFree;
      }
    else
      {
      DBG1  (( MF__, "Maximum number of devspaces reached" ));
      MSGALL(( ERR_MAX_DEVSPACES_REACHED ));
      *pulErrCode = ERRCODE_DEVIO_MAXDEVNO;

      fOK = false;
      *plDevspaceNo = -1;
      }
    }

  DBGOUT;
  return ( fOK );
  }


/*------------------------------*/

/*
 *   Tries to get a request from the DEVx 'IOWork' queue, if no request
 *   is available then this thread should go to sleep until it is waked
 *   by a task/thread making a IO-request to it.
 */

static PDLQ_REC  sql54k_dev_receive ( PPIO_QUEUE_REC          ppIOWorkQue,
                                      PLONG                   plIOQueLen,
                                      PLONG                   plIOWorkExclusive,
                                      PTHREAD_CTRL_HEADER_REC pThrdCtrlHeader )

  {
  #undef  MF__
  #define MF__ MOD__"sql54k_dev_receive"
  LONG       rc = NO_ERROR;
  PDLQ_REC   pRequest;

  DBGIN;

  pRequest = (*ppIOWorkQue)->pRequest;

  while ( pRequest == (PDLQ_REC) 0 )
    {
    DBG3 (( MF__, "DEV or DEVi will sleep on hevSem %d",
            pThrdCtrlHeader->hevSem ));

    if (pThrdCtrlHeader->fSemAlreadyPosted == false)
      {
      // --- signal that there are no wake up activities
      pThrdCtrlHeader->ThrdState = KT_SLEEPING;

//    if (pThrdCtrlHeader->fSemAlreadyPosted == false)
        {
        rc = sql41c_wait_event_sem (pThrdCtrlHeader->hevSem, (ULONG) -1,
                                    "DEV or DEVi" );
        if ((rc != NO_ERROR) && (rc != ERROR_TIMEOUT))
          ABORT();
        }

      pThrdCtrlHeader->ThrdState = KT_RUNNING;
      }
    pThrdCtrlHeader->fSemAlreadyPosted = false;


    pRequest = (*ppIOWorkQue)->pRequest;
    }

  DBG3 (( MF__, "DEV or DEVi Request: 0x%x from TID:%u, [T:0x%03u]",
          pRequest, pRequest->pTaskCtrl->pUKT->ThrdCtrlHeader.Tid,
          pRequest->pTaskCtrl->ulTaskIndex ));

  // --- remove element from queue
  WAIT_UNTIL_ALONE( plIOWorkExclusive );

  (*ppIOWorkQue)->pRequest      = 0;
  (*ppIOWorkQue)                = (*ppIOWorkQue)->pNextElem;
  (*plIOQueLen)--;

  CLEARLOCK( plIOWorkExclusive );

  return ( pRequest );
  }

/*------------------------------*/

static LONG sql54k_dev_open ( PSZ            pszLogDevName,
                              LONG           lDevspaceNo,
                              ULONG          ulOptions,
                              PHFILE         phfDevspace,
                              ERRORTEXT      pcErrText,
                              PULONG         pulErrCode,
                              LONG          *lDevspaceType )
  {
  #undef  MF__
  #define MF__ MOD__"sql54k_dev_open"
  #if defined(_WIN32)
   DWORD               dwFlagsAndAttributes;
  #else
   ULONG               ulFlagsAndAttributes;
   ULONG               Action;
  #endif
  LONG                 rc        = NO_ERROR;
  ULONG                ulTmp     = 0;
  PATHNAME             szPhysDevName;
  LONG                 lAtt;

  DBGIN;

  // --- substitute logical path name parts
  sql44c_subst_log_parts ( szPhysDevName, pszLogDevName );

  #if defined(_WIN32)
   dwFlagsAndAttributes = FILE_ATTRIBUTE_NORMAL;

   if ( ulOptions & DEV_NO_BUFFERING )
     dwFlagsAndAttributes |= FILE_FLAG_NO_BUFFERING;

   if ( ulOptions & DEV_OVERLAPPED_IO )
     dwFlagsAndAttributes |= FILE_FLAG_OVERLAPPED;
  #endif

  *lDevspaceType = sql44c_get_dev_type_by_filename ( szPhysDevName ) ;
  switch ( *lDevspaceType )
    {
    case DT_PSE36:

         rc = os541PSE36Open( szPhysDevName, (HANDLE *) phfDevspace);
         if ( rc != NO_ERROR )
           {
           *pulErrCode = ERRCODE_DEVIO_OPEN_DEVSPACE;
           sql46c_build_error_string ( pcErrText, ERRMSG_DEVIO_OPEN_DEVSPACE, rc );
           MSGALL (( ERR_CANT_OPEN_DEVSPACE, szPhysDevName, rc ));
           DBG1   ( (MF__, "Cannot open devspace '%s', rc = %d", szPhysDevName, rc ));
           *phfDevspace = (HFILE)INVALID_HANDLE_VALUE;

           DBGOUT;
           return ( rc );
           }

    break ;
   
    case DT_DRIVE :
 
    *phfDevspace = (HFILE)__sql54k_create_file( szPhysDevName,
                                                GENERIC_READ      |
                                                GENERIC_WRITE,
                                                FILE_SHARE_READ   |
                                                FILE_SHARE_WRITE,
                                                NULL,
                                                OPEN_EXISTING,
                                                dwFlagsAndAttributes,
                                                NULL );

    if (*phfDevspace == (HFILE)INVALID_HANDLE_VALUE)
      rc = GetLastError();

    if (rc == ERROR_DRIVE_LOCKED)
      {
      *pulErrCode = ERRCODE_DEVIO_DEVSPACE_LOCKED;
      sql46c_build_error_string ( pcErrText, ERRMSG_DEVIO_DEVSPACE_LOCKED, 0 );
      MSGALL (( ERR_DEVSPACE_LOCKED, szPhysDevName ));
      DBG1   ( (MF__, "Devspace '%s' locked by another process", szPhysDevName ));
      *phfDevspace = (HFILE)INVALID_HANDLE_VALUE;

      DBGOUT;
      return ( rc );
      }
    else if ( rc == ERROR_ACCESS_DENIED )
      {
      *pulErrCode = ERRCODE_DEVIO_ACCESS_DENIED;
      MSGALL(( ERR_DEVSPACE_ACCESS_DENIED, szPhysDevName ));
      DBG1  (( MF__, "Access on devspace '%s' denied", szPhysDevName ));
      sql46c_build_error_string ( pcErrText, ERRMSG_DEVIO_ACCESS_DENIED, 0 );
      *phfDevspace = (HFILE)INVALID_HANDLE_VALUE;
      }
    else if ( rc != NO_ERROR )
      {
      *pulErrCode = ERRCODE_DEVIO_OPEN_DEVSPACE;
      sql46c_build_error_string ( pcErrText, ERRMSG_DEVIO_OPEN_DEVSPACE, rc );
      MSGALL (( ERR_CANT_OPEN_DEVSPACE, szPhysDevName, rc ));
      DBG1   ( (MF__, "Cannot open devspace '%s', rc = %d", szPhysDevName, rc ));
      *phfDevspace = (HFILE)INVALID_HANDLE_VALUE;

      DBGOUT;
      return ( rc );
      }

    rc = sql54k_lock_drive ( *phfDevspace );

    if (rc == ERROR_DRIVE_LOCKED)
      {
      *pulErrCode = ERRCODE_DEVIO_DEVSPACE_LOCKED;
      sql46c_build_error_string ( pcErrText, ERRMSG_DEVIO_DEVSPACE_LOCKED, 0 );
      MSGALL (( ERR_DEVSPACE_LOCKED, szPhysDevName ));
      DBG1   ( (MF__, "Devspace '%s' locked by another process", szPhysDevName ));

      CLOSE_FILE( *phfDevspace );
      *phfDevspace = (HFILE)INVALID_HANDLE_VALUE;

      DBGOUT;
      return ( rc );
      }
    else if (rc != 0)
      {
      *pulErrCode = ERRCODE_DEVIO_LOCK_DEVSPACE;
      sql46c_build_error_string ( pcErrText, ERRMSG_DEVIO_LOCK_DEVSPACE, rc );
      MSGALL (( ERR_CANT_LOCK_DEVSPACE, szPhysDevName, rc ));
      DBG1   ( (MF__, "Cannot lock devspace '%s', rc = %d", szPhysDevName, rc ));

      CLOSE_FILE( *phfDevspace );
      *phfDevspace = (HFILE)INVALID_HANDLE_VALUE;

      DBGOUT;
      return ( rc );
      }
      break ;

  default :

    rc = sql44c_get_file_info ( szPhysDevName, &lAtt, &ulTmp, &ulTmp );

    if ( rc == ERROR_ACCESS_DENIED )
      {
      *pulErrCode = ERRCODE_DEVIO_ACCESS_DENIED;
      sql46c_build_error_string ( pcErrText, ERRMSG_DEVIO_ACCESS_DENIED, 0 );
      MSGALL(( ERR_DEVSPACE_ACCESS_DENIED, pszLogDevName ));
      DBG1  (( MF__, "Access on devspace '%s' denied", pszLogDevName ));
      DBGOUT;
      return ( rc );
      }
    else if (rc == ERROR_PATH_NOT_FOUND)
      {
      *pulErrCode = ERRCODE_DEVIO_WRONG_PATH;
      sql46c_build_error_string (pcErrText, ERRMSG_DEVIO_WRONG_PATH, 0 );
      MSGALL(( ERR_WRONG_PATH, pszLogDevName ));
      DBG1  (( MF__, "Wrong path '%s'", pszLogDevName ));
      DBGOUT;
      return ( rc );
      }
    else if ( rc == ERROR_FILE_NOT_FOUND )
      {
      rc = NO_ERROR;
      }
    else if ( rc != NO_ERROR )
      {
      *pulErrCode = ERRCODE_DEVIO_GET_FILE_INFO;
      sql46c_build_error_string (pcErrText, ERRMSG_DEVIO_GET_FILE_INFO, rc );
      MSGALL(( ERR_GET_FILE_INFO, pszLogDevName, rc ));
      DBG1  (( MF__, "Query file info error on '%s', rc = %ld",
               pszLogDevName, rc ));
      DBGOUT;
      return( rc );
      }

    if ( lAtt == -1 )
      {
      rc          = -1;
      *pulErrCode = ERRCODE_DEVIO_WRONG_FILE_ATTRIB;
      MSGALL(( ERR_UNKNOWN_FILE_ATTRIBUTE, pszLogDevName ));
      DBG1  (( MF__, "'%s' has a unknown file attribute", pszLogDevName  ));
      sql46c_build_error_string ( pcErrText, ERRMSG_DEVIO_WRONG_FILE_ATTRIB, 0 );
      }
    else if ((lAtt & FILE_DIRECTORY) == FILE_DIRECTORY)
      {
      rc          = -1;
      *pulErrCode = ERRCODE_DEVIO_DEVSPACE_ON_DIR;
      MSGALL(( ERR_IO_ON_DIR, pszLogDevName ));
      DBG1  (( MF__, "No Devspaces on directory '%s' !", pszLogDevName ));
      sql46c_build_error_string ( pcErrText, ERRMSG_DEVIO_DEVSPACE_ON_DIR, 0 );
      }
    else if ((lAtt & FILE_SYSTEM) == FILE_SYSTEM)
      {
      rc          = -1;
      *pulErrCode = ERRCODE_DEVIO_DEVSPACE_ON_SYS;
      MSGALL(( ERR_IO_ON_SYS, pszLogDevName ));
      DBG1  (( MF__, "No Devspaces on system file '%s' !", pszLogDevName ));
      sql46c_build_error_string ( pcErrText, ERRMSG_DEVIO_DEVSPACE_ON_SYS, 0 );
      }
    else if ((lAtt & FILE_READONLY) == FILE_READONLY)
      {
      rc          = -1;
      *pulErrCode = ERRCODE_DEVIO_DEVSPACE_ON_RD_ONLY;
      MSGALL(( ERR_DEVSPACE_ON_RD_ONLY, pszLogDevName ));
      DBG1  (( MF__, "No Devspaces on read-only file '%s' !",
                      pszLogDevName  ));
      sql46c_build_error_string (pcErrText, ERRMSG_DEVIO_DEVSPACE_ON_RD_ONLY, 0 );
      }
    else if (((lAtt & FILE_COMPRESSED) == FILE_COMPRESSED) &&
			 (  XPARAM(fAllowCompressedDevsp) == false ))
      {
      rc          = -1;
      *pulErrCode = ERRCODE_DEVIO_DEVSPACE_ON_COMP;
      MSGALL(( ERR_DEVSPACE_COMPRESSED, pszLogDevName ));
      DBG1  (( MF__, "No Devspaces on compressed file '%s' !",
                      pszLogDevName  ));
      sql46c_build_error_string (pcErrText, ERRMSG_DEVIO_DEVSPACE_COMPRESSED, 0 );
      }
    else
      {
      DBG3 (( MF__, "Opening devspace: %s", pszLogDevName ));

      *phfDevspace = (HFILE)__sql54k_create_file( szPhysDevName,
                                                  GENERIC_READ      |
                                                  GENERIC_WRITE,
                                                  FILE_SHARE_READ   |
                                                  FILE_SHARE_WRITE,
                                                  NULL,
                                                  OPEN_EXISTING,
                                                  dwFlagsAndAttributes,
                                                  NULL );

      if (*phfDevspace == (HFILE)INVALID_HANDLE_VALUE)
        rc = GetLastError();

      if ( rc == ERROR_ACCESS_DENIED )
        {
        *pulErrCode = ERRCODE_DEVIO_ACCESS_DENIED;
        MSGALL(( ERR_DEVSPACE_ACCESS_DENIED, pszLogDevName ));
        DBG1  (( MF__, "Access on devspace '%s' denied", pszLogDevName ));
        sql46c_build_error_string ( pcErrText, ERRMSG_DEVIO_ACCESS_DENIED, 0 );
        *phfDevspace = (HFILE)INVALID_HANDLE_VALUE;
        }
      else if ( rc != NO_ERROR )
        {
        *pulErrCode = ERRCODE_DEVIO_OPEN_DEVSPACE;
        sql46c_build_error_string ( pcErrText, ERRMSG_DEVIO_OPEN_DEVSPACE, rc );
        MSGALL(( ERR_CANT_OPEN_DEVSPACE, pszLogDevName, rc ));
        DBG1  (( MF__, "Cannot open devspace '%s', rc = %d", pszLogDevName, rc ));
        *phfDevspace = (HFILE)INVALID_HANDLE_VALUE;
        }
      }
    }

  DBGOUT;
  return ( rc );
  }

/*------------------------------*/

LONG     sql54k_dev_close        ( HFILE                     hfDevspace,
                                   char                     *pszDriveName,
                                   LONG                      lDevspaceType,
                                   ERRORTEXT                 errtext)
{
  #undef  MF__
  #define MF__ MOD__"sql54k_dev_close"
  LONG                 rc        = NO_ERROR;

  switch ( lDevspaceType )
    {
      case DT_PSE36:

           rc = os541PSE36Close( (HANDLE )hfDevspace);
           break; 

      case DT_DRIVE :

           rc = sql54k_unlock_drive ( hfDevspace );
           if (rc != 0)
             {
             MSGALL (( ERR_CANT_UNLOCK_DEVSPACE, pszDriveName, rc ));
             DBG1   ( (MF__, "Cannot unlock devspace '%s', rc = %d", pszDriveName, rc ));
             sql46c_build_error_string ( errtext, ERRMSG_DEVIO_UNLOCK_DEVSPACE, rc );
             }
           else
             rc = CLOSE_FILE( hfDevspace );
           break ;

      default :

           rc = CLOSE_FILE( hfDevspace );
    }

  if ( rc != NO_ERROR )
    { 
    MSGALL(( ERR_CANT_CLOSE_DEVSPACE, rc ));
    DBG1  (( MF__, "Cannot close devspace, rc = %d", rc ));
    sql46c_build_error_string ( errtext, ERRMSG_DEVIO_CLOSE_DEVSPACE, rc );
    }

  DBGOUT;
  return ( rc );
  }

/*------------------------------*/

/*
 * May only be called by the DEV0 thread
 * Assumes that the devspace number is known.
 * Thus only opens the devspace, creates the DEVI thread and updates
 * the 'DEVI_CTRL_REC'.
 */
static VOID sql54k_dev0_vattach ( DLQ_VATTACH_REQ_ARGS_REC *pVattachArgs )
  {
  #undef  MF__
  #define MF__ MOD__"sql54k_dev0_vattach"
  PDEVI_CTRL_REC                   pDeviCtrl;
  LONG                             rc;
  ULONG                            ulWaitTime;
  LONG                             lNDevi;
  LONG                             lDevspaceNo;
  ULONG                            ulTmp;
  ERRORTEXT                        DummyErrText;
  bool                             fFirstTime       = true;
  bool                             fIsTraceDevspace = false;
  bool                             fIsPSE36 = false;
  PDEV_IO_CTRL_REC                 pIOCtrl;

  DBGIN;

  DBG3 (( MF__, "Find free devspace number for '%s'", pVattachArgs->pszDevspaceToAttach ));

  // --- get free devspace number
  pVattachArgs->fOk = sql54k_find_devspace_no ( &lDevspaceNo,
                                                pVattachArgs->pszDevspaceToAttach,
                                                &pVattachArgs->ulErrCode );
  if ( pVattachArgs->fOk == true )
    {
    pIOCtrl                   = &kgs.Dev0.pIOCtrl[lDevspaceNo];
    pVattachArgs->lDevspaceNo = lDevspaceNo;

    sql70k_init_dev_io ( pIOCtrl );
    strcpy ( pIOCtrl->szDevspace, pVattachArgs->pszDevspaceToAttach );

    //
    //  - Set number of DEVi threads and save its devspace number
    //    If it's the trace devspace use only 1 DEVi thread
    //
    fIsTraceDevspace  = !strcmp( pVattachArgs->pszDevspaceToAttach, XPARAM(szTraceDevspace));
    fIsPSE36          = os541IsPSE36( pVattachArgs->pszDevspaceToAttach ) != 0;

  
    lNDevi = fIsTraceDevspace || fIsPSE36 ? 1 : XPARAM(lDeviPerDevspace) ;

    DBG3 (( MF__, "Creating %d DEVis for '%s', DevspaceNo: %d",
            lNDevi, pIOCtrl->szDevspace, lDevspaceNo ));

    do
      {
      // --- evaluate the DEVi control record
      pDeviCtrl = kgs.pFirstDeviCtrl                     +
                  lDevspaceNo * XPARAM(lDeviPerDevspace) +
                  lNDevi - 1;

      DBG3 (( MF__, "Using DEVi slot     %d",   lNDevi - 1 ));
      DBG3 (( MF__, "pDeviCtrl is        0x%x", pDeviCtrl ));

      // --- init the devi control record
      sql70k_init_devi_ctrl ( pDeviCtrl, pDeviCtrl->pIOs );

      // --- set devspace related DEVi number and devspace number
      pDeviCtrl->pIOCtrl    = pIOCtrl;
      pIOCtrl->lDevspaceNo  = lDevspaceNo;
      pIOCtrl->ulDBPageSize = pVattachArgs->ulDBPageSize ;

      // --- open the devspace
      rc = sql54k_dev_open ( pIOCtrl->szDevspace,
                             lDevspaceNo,
                             DEV_NO_BUFFERING, // PTS 1110756
                             &pDeviCtrl->hfDevspace,
                             DummyErrText,
                             &pVattachArgs->ulErrCode,
                             &pIOCtrl->ulDevType );

      if ( rc != NO_ERROR )
        {
        pVattachArgs->fOk  = false;
        sql70k_init_devi_ctrl ( pDeviCtrl, pDeviCtrl->pIOs );
        sql70k_init_dev_io    ( pIOCtrl );

        // --- kill all already started DEVi threads
        sql54k_devi_kill (lDevspaceNo);

        DBGOUT;
        return;
        }
      //
      // --- Check block 0, get devspace size and protect file
      //     devspaces against attaches from other databases.
      //
      if ( fFirstTime == true && !fIsPSE36 )
        {
        fFirstTime = false;

        rc = sql54k_evaluate_block_0 ( pDeviCtrl->hfDevspace,
                                       &pIOCtrl->ulDevPages,
                                       &ulTmp,
                                       pIOCtrl->ulDBPageSize,
                                       pIOCtrl->ulDevType == DT_DRIVE,
                                       pIOCtrl->szDevspace );
        if ( rc != NO_ERROR )
          {
          ERRORTEXT   errtext ;
          pVattachArgs->fOk                   = false;
          pVattachArgs->ulErrCode             = ERRCODE_DEVIO_EVAL_DEVSPACE;

          sql54k_dev_close ( pDeviCtrl->hfDevspace, pIOCtrl->szDevspace, 
                             pIOCtrl->ulDevType,    errtext) ;

          sql70k_init_devi_ctrl ( pDeviCtrl, pDeviCtrl->pIOs );
          sql70k_init_dev_io    ( pIOCtrl );

          DBGOUT;
          return;
          }
        //
        // --- protect file devspaces against attaches from other databases
        //     (don't lock the trace devspace !)
        //
        if (( pIOCtrl->ulDevType != DT_DRIVE )  &&
            ( !fIsTraceDevspace && !fIsPSE36 ))
          {
          rc = sql54k_set_devspace_lock ( pDeviCtrl->hfDevspace,
                                          pIOCtrl->szDevspace,
                                          false );
          if ( rc != NO_ERROR )
            {
            pVattachArgs->fOk       = false;
            pVattachArgs->ulErrCode = ERRCODE_DEVIO_LOCK;

            CLOSE_FILE( pDeviCtrl->hfDevspace );

            sql70k_init_devi_ctrl ( pDeviCtrl, pDeviCtrl->pIOs );
            sql70k_init_dev_io    ( pIOCtrl );

            DBGOUT;
            return;
            }
          }
        }

      // --- create DEVi-Thread
      rc = CREATE_THREAD((PULONG)&pDeviCtrl->ThrdCtrlHeader.Tid,
                         &pDeviCtrl->ThrdCtrlHeader.hThrd,
                         sql54k_devi,
                         (ULONG)pDeviCtrl,       // argument
                         CREATE_THRD_SUSPENDED,
                         KT_DEVI_STACK_SIZE);

      if (rc != NO_ERROR)
        {
        MSGD ((ERR_CREATING_THREAD, "DEVi", rc));
        DBG1 ((MF__, "Cannot create thread: '%s', rc = %d", "DEVi", rc))
        }
      else
        {
        rc = RESUME_THREAD(pDeviCtrl->ThrdCtrlHeader.Tid,
                           pDeviCtrl->ThrdCtrlHeader.hThrd );

        if (rc != NO_ERROR)
          {
          MSGD ((ERR_RESUMING_THREAD, "DEVi", rc));
          DBG1 ((MF__, "Cannot resume thread: '%s', rc = %d", "DEVi", rc))
          }
        }

      if ( rc == NO_ERROR )
        {
        // --- wait until the DEVi is up and running
        for ( ulWaitTime = MAX_VATTACH_WAIT_TIME * 10;
              (pDeviCtrl->ThrdCtrlHeader.ThrdState == KT_INITIALIZED) && ulWaitTime;
              --ulWaitTime )
          {
          SLEEP( 1000 / 10 ); // -- sleep a moment
          }
        }

      if (( rc != NO_ERROR ) || ( ulWaitTime == 0 ))
        {
        if ( ulWaitTime == 0 )
          pVattachArgs->ulErrCode  = ERRCODE_DEVIO_DEVICREA_TIMEOUT;
        else
          pVattachArgs->ulErrCode  = ERRCODE_DEVIO_DEVICREA;

        pVattachArgs->fOk = false;

        CLOSE_FILE( pDeviCtrl->hfDevspace );

        sql70k_init_devi_ctrl ( pDeviCtrl, pDeviCtrl->pIOs );
        sql70k_init_dev_io    ( pIOCtrl );

        // --- kill all already started DEVi threads
        sql54k_devi_kill (lDevspaceNo);

        DBGOUT;
        return;
        }

      DBG3 (( MF__, "DEVi TID:%d attached", pDeviCtrl->ThrdCtrlHeader.Tid ));
      }
    while ( (--lNDevi) > 0 );

     if ( fIsPSE36 )
       pIOCtrl->hfDevspace = pDeviCtrl->hfDevspace ;
     else
     // --- use overlapped io?
     if ( kgs.ulNumOfDevWorker != 0 )
       {
       // --- open the devspace
       rc = sql54k_dev_open ( pIOCtrl->szDevspace,
                              lDevspaceNo,
                              DEV_OVERLAPPED_IO | DEV_NO_BUFFERING, // PTS 1110756
                              &pIOCtrl->hfDevspace,
                              DummyErrText,
                              &pVattachArgs->ulErrCode,
                              &pIOCtrl->ulDevType );

       if ( rc != NO_ERROR )
         {
         pVattachArgs->fOk  = false;
         sql70k_init_dev_io ( pIOCtrl );

         // --- kill all already started DEVi threads
         sql54k_devi_kill (lDevspaceNo);
         DBGOUT;
         return;
         }

       pIOCtrl->hCompletionPort =
                    CreateIoCompletionPort( (HANDLE)pIOCtrl->hfDevspace,
                                            kgs.hCompletionPort,
                                            (DWORD)lDevspaceNo, 0);

       if (pIOCtrl->hCompletionPort == NULL)
         {
         pVattachArgs->ulErrCode  = ERRCODE_DEVIO_CREATE_COMPL_PORT;
         pVattachArgs->fOk        = false;
         sql70k_init_dev_io ( pIOCtrl );

         // --- kill all already started DEVi threads
         sql54k_devi_kill (lDevspaceNo);
         DBGOUT;
         return;
         }
       }

    pIOCtrl->ulAttached = kgs.ulCurrTime;
    pIOCtrl->DevspaceType = pVattachArgs->DevspaceType;

    if ( fIsTraceDevspace == true )
      kgs.Dev0.lTraceDevspaceNo = lDevspaceNo;
    }

  DBGOUT;
  return;
  }

/*------------------------------*/

static VOID sql54k_devi_kill ( LONG lDevspaceNo )
  {
  #undef  MF__
  #define MF__ MOD__"sql54k_devi_kill"
  LONG                   lNDevi = XPARAM(lDeviPerDevspace);
  PDEVI_CTRL_REC         pDeviCtrl;
  ERRORTEXT              errtext;

  DBGIN;

  do
    {
    // --- evaluate the DEVi control record
    pDeviCtrl = kgs.pFirstDeviCtrl                     +
                lDevspaceNo * XPARAM(lDeviPerDevspace) +
                lNDevi - 1;

    if (pDeviCtrl->ThrdCtrlHeader.ThrdState != KT_INITIALIZED)
      {
      DBG3 (( MF__, "Killing DEVi Tid:%u, pDeviCtrl is 0x%u",
              pDeviCtrl->ThrdCtrlHeader.Tid, pDeviCtrl ));

      KILL_THREAD ( pDeviCtrl->ThrdCtrlHeader.Tid,
                    pDeviCtrl->ThrdCtrlHeader.hThrd );

      if ( pDeviCtrl->hfDevspace != (HFILE)INVALID_HANDLE_VALUE )
        sql54k_dev_close ( pDeviCtrl->hfDevspace, 
                           pDeviCtrl->pIOCtrl->szDevspace, 
                           pDeviCtrl->pIOCtrl->ulDevType , errtext) ;

      if ( pDeviCtrl->ThrdCtrlHeader.hevSem != (HEV) INVALID_HANDLE_VALUE )
        sql41c_close_event_sem ( pDeviCtrl->ThrdCtrlHeader.hevSem, "DEVi" );

      sql70k_init_devi_ctrl ( pDeviCtrl, pDeviCtrl->pIOs );
      }
    }
  while ( (--lNDevi) > 0 );

  DBGOUT;
  return;
  }

/*------------------------------*/

static VOID  sql54k_dev0_vdetach  ( PDLQ_REC    pRequest )
  {
  #undef  MF__
  #define MF__ MOD__"sql54k_dev0_vdetach"
  PDEVI_CTRL_REC            pDeviCtrl;
  PUKT_CTRL_REC             pTmpUKTCtrl;
  LONG                      lDevspaceNo;
  LONG                      lNDevi;
  TID                       Tid;
  HANDLE                    hThrd;
  APIRET                    rc = NO_ERROR;
  DLQ_VDETACH_REQ_ARGS_REC  *pVdetachArgs;
  PDEV_IO_CTRL_REC          pIOCtrl;
  ULONG                     ulWaitTime = MAX_VDETACH_WAIT_TIME;
  ERRORTEXT                 errtext;

  pVdetachArgs = &pRequest->ReqArgs.VdetachRequest;
  lDevspaceNo  = pVdetachArgs->lDevspaceNo;
  lNDevi       = XPARAM(lDeviPerDevspace);
  pIOCtrl      = &kgs.Dev0.pIOCtrl[lDevspaceNo];

  // --- signal that a DETACH is in progress
  pIOCtrl->ulDetachPending = kgs.ulCurrTime;

  if ( kgs.Dev0.lTraceDevspaceNo == lDevspaceNo )
    kgs.Dev0.lTraceDevspaceNo = UNDEF;

  do
    {
    // --- evaluate the DEVi control record
    pDeviCtrl = kgs.pFirstDeviCtrl                     +
                lDevspaceNo * XPARAM(lDeviPerDevspace) +
                lNDevi - 1;

    if (( pDeviCtrl->ThrdCtrlHeader.ThrdState != KT_INITIALIZED ) &&
        ( pDeviCtrl->ThrdCtrlHeader.ThrdState != KT_TERMINATING ))
      {
      DBG3 (( MF__, "Detaching DEVi slot %d",   lNDevi - 1 ));
      DBG3 (( MF__, "pDeviCtrl is        0x%x", pDeviCtrl ));

      //
      // --- wait until the DEVi has stop the activities
      //
      while (( pDeviCtrl->ThrdCtrlHeader.ThrdState != KT_SLEEPING ) ||
             ( pDeviCtrl->pIOWorkQue               != pDeviCtrl->pIOReqQue ))
        {
        DBG3 (( MF__, "ThrdCtrlHeader.ThrdState = %d",   pDeviCtrl->ThrdCtrlHeader.ThrdState ));
        DBG3 (( MF__, "pIOWorkQue               = 0x%x", pDeviCtrl->pIOWorkQue ));
        DBG3 (( MF__, "pIOReqQue                = 0x%x", pDeviCtrl->pIOReqQue ));
        SLEEP( 100 );
        }

      //
      // --- awaken devi and wait for its termintion
      //
      DBG3(( MF__,
             "(Dev0) Send pRequest: 0x%x, using (Devi TID:%u) pIOReqQue: 0x%x",
             pRequest, pDeviCtrl->ThrdCtrlHeader.Tid,
             pDeviCtrl->pIOReqQue ));

      Tid   = pDeviCtrl->ThrdCtrlHeader.Tid;
      hThrd = pDeviCtrl->ThrdCtrlHeader.hThrd;

      REQ_TO_DEVI ( pDeviCtrl, pRequest );

      WAIT_THREAD(hThrd, ulWaitTime);
      }
    }
  while ( (--lNDevi) > 0 );

  /*
   * We trust that the DEVi threads exits peacefully
   * but cannot be sure that all are gone immediately
   */

  // --- close all single io file of the current UKT
  for (pTmpUKTCtrl = kgs.pFirstUKTCtrl; pTmpUKTCtrl <= kgs.pLastUKTCtrl;
       pTmpUKTCtrl++)
    {
    if (pTmpUKTCtrl->pSingleBlockIO[lDevspaceNo].hfSingleIOFiles !=
                                      (HFILE) INVALID_HANDLE_VALUE )
      {
      DBG3 (( MF__, "Release single io handle of devspace no %d", lDevspaceNo ));

      MSGD (( INFO_SINGLEIO_DETACH, pIOCtrl->szDevspace,
              pTmpUKTCtrl->ulUKTIndex));
 
      rc = sql54k_dev_close ( pTmpUKTCtrl->pSingleBlockIO[lDevspaceNo].hfSingleIOFiles,
                              pIOCtrl->szDevspace, 
                              pIOCtrl->ulDevType , errtext) ;

      pTmpUKTCtrl->pSingleBlockIO[lDevspaceNo].ulOldSingleIOPos = 0;
      pTmpUKTCtrl->pSingleBlockIO[lDevspaceNo].hfSingleIOFiles  = (HFILE) INVALID_HANDLE_VALUE;
      }

    pTmpUKTCtrl->pSingleBlockIO[lDevspaceNo].lReadCntr  = 0;
    pTmpUKTCtrl->pSingleBlockIO[lDevspaceNo].lReadPages  = 0;
    pTmpUKTCtrl->pSingleBlockIO[lDevspaceNo].lWriteCntr = 0;
    pTmpUKTCtrl->pSingleBlockIO[lDevspaceNo].lWritePages = 0;
    }

  // --- overlapped io used?
  if ( kgs.ulNumOfDevWorker != 0 )
    rc = sql54k_dev_close ( pIOCtrl->hfDevspace,
                            pIOCtrl->szDevspace, 
                            pIOCtrl->ulDevType , errtext) ;

  if ( kgs.Dev0.lTraceDevspaceNo == lDevspaceNo )
    kgs.Dev0.lTraceDevspaceNo = (LONG)UNDEF;

  sql70k_init_dev_io ( pIOCtrl );
  kgs.fSingleIOOpenLocked  = false;

  DBGOUT;
  return;
  }

/*------------------------------*/

/*
 * The main loop for the DEVI process, a request is
 * always received and always returned
 */
static VOID _System sql54k_devi ( PDEVI_CTRL_REC  pDeviCtrl )
  {
  #undef  MF__
  #define MF__ MOD__"sql54k_devi"
  APIRET                        rc = NO_ERROR;
  BOOL                          fDetached;
  PDLQ_REC                      pRequest;
  HANDLE                        hThrd;
  PDEV_IO_CTRL_REC              pIOCtrl;

  DBGIN;

  __try 
  {
    // --- save the current thread handle which is used by 'EXITTHREAD'.
    //     Set the Pointer to the IO Control structure.
    hThrd   = pDeviCtrl->ThrdCtrlHeader.hThrd;
    pIOCtrl = pDeviCtrl->pIOCtrl;

    // --- set the thread priority
    rc = SET_THREAD_PRIO ( pDeviCtrl->ThrdCtrlHeader.lThrdPrio,
                           pDeviCtrl->ThrdCtrlHeader.Tid,
                           pDeviCtrl->ThrdCtrlHeader.hThrd );

    if ( rc != NO_ERROR )
      {
      MSGD (( ERR_SETTING_THRD_PRIO, pDeviCtrl->ThrdCtrlHeader.Tid));
      DBG1 ((MF__, "Cannot set the thread priority, TID: %u",
                    kgs.Dev0.ThrdCtrlHeader.Tid));
      ABORT();
      }

    // --- Create event semaphore (SET) for DEVi
    rc = sql41c_create_event_sem (&pDeviCtrl->ThrdCtrlHeader.hevSem,
                                  NULL,
                                  NULL,
                                  0,
                                  false,
                                  NULL);
    if (rc != NO_ERROR)
      ABORT();

    DBG3 (( MF__, "DEVi TID:%d started", pDeviCtrl->ThrdCtrlHeader.Tid ));
    MSGD (( INFO_DEVI_STARTED, pIOCtrl->szDevspace));

    // --- set thread control info
    pDeviCtrl->ThrdCtrlHeader.ThrdState   = KT_RUNNING;
    pDeviCtrl->ThrdCtrlHeader.ulStartTime = kgs.ulCurrTime;

    // --- Runs until detached
    for ( fDetached = false; fDetached == false ; )
      {
      pRequest = sql54k_dev_receive (&pDeviCtrl->pIOWorkQue,
                                     &pDeviCtrl->DevIOStat.ulIOQueLen,
                                     &pDeviCtrl->lExclusive,
                                     &pDeviCtrl->ThrdCtrlHeader);

      switch ( pRequest->ulReqType )
        {
        case REQ_VBLOCKIO:
          sql54k_devi_vblockio ( pRequest, pIOCtrl, pDeviCtrl );

          // -- send reply back to UKT
          REQ_FROM_DEVx_TO_UKT(pRequest);
          break;

        case REQ_VDETACH :
          pDeviCtrl->ThrdCtrlHeader.ThrdState = KT_TERMINATING;

          sql54k_devi_vdetach  ( pDeviCtrl, pIOCtrl );

          fDetached = true; // --- exit for loop
          break;

        default          :
          MSGD (( ERR_ILLEGAL_REQUEST, "DEVi", pRequest->ulReqType ));

          // -- send reply back to UKT
          REQ_FROM_DEVx_TO_UKT(pRequest);
          break;
        }

      DBG3 (( MF__, "next in loop"));
      }
  } 
  __except( sql50k_excepthandler(GetExceptionInformation()) ) {;}

  DBGOUT;
  EXITTHREAD ( 0, hThrd );

  return;
  }


/*------------------------------*/

static VOID sql54k_devi_vdetach ( PDEVI_CTRL_REC    pDeviCtrl,
                                  PDEV_IO_CTRL_REC  pIOCtrl )
  {
  #undef  MF__
  #define MF__ MOD__"sql54k_devi_vdetach"
  APIRET     rc = NO_ERROR;
  ERRORTEXT  errtext;

  DBG3 (( MF__, "Detaching DEVi   TID:%d", pDeviCtrl->ThrdCtrlHeader.Tid ));


  rc = sql54k_dev_close ( pDeviCtrl->hfDevspace,
                          pIOCtrl->szDevspace, 
                          pIOCtrl->ulDevType , errtext) ;

  sql41c_close_event_sem ( pDeviCtrl->ThrdCtrlHeader.hevSem, "DEVi" );

  MSGD (( INFO_DEVI_STOPPED, pDeviCtrl->pIOCtrl->szDevspace ));

  sql70k_init_devi_ctrl ( pDeviCtrl, pDeviCtrl->pIOs );

  DBGOUT;
  return;
  }

/*------------------------------*/

static LONG sql54k_set_devspace_lock ( HFILE          hfFileHandle,
                                       PSZ            pszLogDevName,
                                       BOOL           fUnlock )
  {
  #undef  MF__
  #define MF__ MOD__"sql54k_set_devspace_lock"
  #if !defined(_WIN32)
   FILELOCK      flFileLockPos = { DEVSPACE_ID_POS, sizeof (DEVSPACE_ID_STRING) - 1 };
  #endif
  ULONG          LockTimeout = 1;
  APIRET         rc          = NO_ERROR;
  OVERLAPPED     Overlapped;

  DBGPAS;

  memset ( &Overlapped, 0, sizeof(Overlapped) );

  if ( fUnlock == false )
    {
    if ( !LockFileEx( (HANDLE)hfFileHandle, LOCKFILE_FAIL_IMMEDIATELY | LOCKFILE_EXCLUSIVE_LOCK,
                      0, DEVSPACE_ID_POS + sizeof(DEVSPACE_ID_STRING) - 1, 0,
                      &Overlapped ))
      {
      rc = GetLastError();
      }
    }
  else
    {

    if ( !UnlockFileEx( (HANDLE)hfFileHandle, 0,
                        DEVSPACE_ID_POS + sizeof(DEVSPACE_ID_STRING) - 1, 
                        0, &Overlapped ))
      {
      rc = GetLastError();
      }
    }

  if ( rc == ERROR_LOCK_VIOLATION )
    {
    MSGALL(( ERR_DEVSPACE_IN_USE, pszLogDevName ));
    DBG1  (( MF__, "Devspace '%s' in use", pszLogDevName));
    }
  else if ( rc != NO_ERROR )
    {
    MSGALL(( ERR_CANT_LOCK_UNLOCK_DEVSPACE, pszLogDevName, rc ));
    DBG1  (( MF__, "Cannot lock/unlock the devspace '%s', rc = %d",
             pszLogDevName, rc ));
    }

  return ( rc );
  }

/*------------------------------*/

/*
 *  May not be called in the DEV0 thread, only in the DEVi thread
 *  This is the mother thread controlling the devspace 'i', all
 *  requests are addressed to it.
 */
static VOID sql54k_devi_vblockio ( PDLQ_REC          pRequest,
                                   PDEV_IO_CTRL_REC  pIOCtrl,
                                   PDEVI_CTRL_REC    pDeviCtrl )
  {
  #undef  MF__
  #define MF__ MOD__"sql54k_devi_vblockio"
  DLQ_VBLOCKIO_REQ_ARGS_REC  *pVblockIOArgs;

  DBGIN;

  // --- get pointer to the arguments
  pVblockIOArgs = &pRequest->ReqArgs.VblockIORequest;

  DBG3 (( MF__, "'vblockio' for lDevspaceNo: %d, lDevspacePage %d",
          pVblockIOArgs->lDevspaceNo , pVblockIOArgs->lDevspacePage ));

  if ( __sql54k_is_user_req_io_error(pVblockIOArgs->lDevspaceNo, pVblockIOArgs->OpCode) )
    {
        /*
         *  Simulate an I/O error for test porposes.
         */
        pVblockIOArgs->fOk            = false ;
        pVblockIOArgs->ulErrCode      = ERRCODE_DEVIO_USER_REQ_ERROR ;
        MSGD ((ERR_VBLOCKIO_TEST_IO_ERROR,
              ( pVblockIOArgs->OpCode == DEV_READ) ? "reading from" : "writing to" ,
                pVblockIOArgs->lDevspaceNo , pVblockIOArgs->lDevspacePage ));
    }
  else
  if ( pVblockIOArgs->OpCode == DEV_READ )
    sql54k_devi_read  ( pDeviCtrl, pVblockIOArgs );
  else
    sql54k_devi_write ( pDeviCtrl, pVblockIOArgs );

  DBG3 (( MF__, "ok = %f", pVblockIOArgs->fOk ));

  DBGOUT;
  return;
  }


/*------------------------------*/


static LONG sql54k_uncompress( HANDLE      hfFileHandle )
  {
  #undef  MF__
  #define MF__ MOD__"sql54k_uncompress"
  ULONG                 rc = NO_ERROR;
  USHORT				State = 0; 
  ULONG					Length; 

  DBGIN;

  if (!DeviceIoControl( hfFileHandle, FSCTL_SET_COMPRESSION, &State, 
					    sizeof(USHORT), NULL, 0, &Length, false )) 
	{ 
    rc = GetLastError();
	}

  DBGOUT;
  return ( rc );
  }

/*------------------------------*/

static LONG sql54k_vmark_bad_dev ( LONG               lDevspaceNo,
                                   tsp_page_no        tspBadDevspacePage )
  {
  #undef  MF__
  #define MF__ MOD__"sql54k_vmark_bad_dev"
  ULONG                           rc,
                                  ulFilePosHigh    = 0,
                                  ulLength;
  ULONG                           ulTmp;
  ERRORTEXT                       pcTmpErrText;
  HFILE                           hfDevspace;
  PDEV_IO_CTRL_REC                pIOCtrl;
  ULONG                           lDevspaceType;

  DBGIN;

  pIOCtrl  = &kgs.Dev0.pIOCtrl[lDevspaceNo];

  rc = sql54k_dev_open ( pIOCtrl->szDevspace,
                         pIOCtrl->lDevspaceNo,
                         0,
                         &hfDevspace,
                         pcTmpErrText,
                         &ulTmp,
                         &lDevspaceType );

  if ( rc != NO_ERROR )
    return ( rc );


  rc = SET_FILE_PTR ( hfDevspace, DEVSPACE_BAD_MARK_POS,
                      FILE_BEGIN, &ulFilePosHigh, NULL );

  if ( rc != NO_ERROR )
    {
    MSGD(( ERR_CANT_SEEK_ON_DEVSPACE, pIOCtrl->szDevspace, rc ));
    CLOSE_FILE( hfDevspace );
    return ( rc );
    }

  rc = WRITE_FILE ( hfDevspace, &tspBadDevspacePage,
                    sizeof(tsp_page_no), &ulLength );

  if ( ( rc != NO_ERROR ) || ( ulLength != sizeof(tsp_page_no)) )
    {
    MSGD((ERR_CANT_WRITE_ON_DEVSPACE, 0, pIOCtrl->szDevspace, rc, ulLength));
    }
  else
    {
    MSGALL (( ERR_MARKING_DEVSP_AS_BAD, pIOCtrl->szDevspace,
              tspBadDevspacePage ))
    }


  CLOSE_FILE( hfDevspace );

  DBGOUT;
  return ( rc );
  }

/*------------------------------*/

/*
 * This function performs read io on the devspace
 */
static LONG  sql54k_devi_read ( PDEVI_CTRL_REC             pDeviCtrl,
                                DLQ_VBLOCKIO_REQ_ARGS_REC  *pVblockIOArgs )
  {
  #undef  MF__
  #define MF__ MOD__"sql54k_devi_read"

  ULONG           ulFilePosLow;
  ULONG           ulFilePosHigh;
  ULONG           ulLengthWanted;
  ULONG           ulLength;
  APIRET          rc = NO_ERROR;

  DBGIN;

  if ( (ULONG)(pVblockIOArgs->lDevspacePage + 1) != pDeviCtrl->ulOldIOPos )
    {
    PAGES_TO_QWORD( ulFilePosLow, ulFilePosHigh,
                    pVblockIOArgs->lDevspacePage + 1 );

    rc = SET_FILE_PTR ( pDeviCtrl->hfDevspace, ulFilePosLow, FILE_BEGIN,
                        &ulFilePosLow, &ulFilePosHigh );

    if (rc != 0)
      {
      MSGALL(( ERR_CANT_SEEK_ON_DEVSPACE,
               pDeviCtrl->pIOCtrl->szDevspace, rc ));
      DBG1  (( MF__, "Cannot seek on devspace: %s, rc = %d",
               pDeviCtrl->pIOCtrl->szDevspace, rc ));
      pVblockIOArgs->ulErrCode = ERRCODE_DEVIO_SEEK;
      pVblockIOArgs->fOk       = false;

      DBGOUT;
      return ( rc );
      }
    }

  #ifdef DEVELOP_CHECKS
   if ((ULONG)( pVblockIOArgs->lDevspacePage +
                pVblockIOArgs->sNumOfPages   + 1 ) >
                        pDeviCtrl->pIOCtrl->ulDevPages )
     {
     MSGALL(( ERR_WRNG_DEVSPACE_POS,
              pVblockIOArgs->lDevspacePage,
              pVblockIOArgs->sNumOfPages ));
     DBG1  (( MF__, "Wrong devspace read position: %d - Pages: %d",
              pVblockIOArgs->lDevspacePage,
              pVblockIOArgs->sNumOfPages ));

     pVblockIOArgs->ulErrCode = ERRCODE_DEVIO_DEV_POS;
     pVblockIOArgs->fOk       = false;

     DBGOUT;
     return ( -1 );
     }
  #endif

  #ifdef DEVELOP_CHECKS
   rc = __sql54k_check_filepos ( pDeviCtrl->hfDevspace,
                                 pVblockIOArgs->lDevspaceNo,
                                 pVblockIOArgs->lDevspacePage,
                                 pVblockIOArgs->sNumOfPages,
                                 pDeviCtrl->pIOCtrl->ulDevPages,
                                 &pVblockIOArgs->ulErrCode,
                                 NULL );

   if ( rc != NO_ERROR )
     {
     pVblockIOArgs->fOk = false;
     DBGOUT;
     return ( -1 );
     }
  #endif

  ulLengthWanted = PAGES_TO_BYTES(pVblockIOArgs->sNumOfPages);

  /* PTS 1106176 */
  rc = sql54k_ReadBlockWithRetry( pDeviCtrl->hfDevspace, pVblockIOArgs->pBuffAddr, ulLengthWanted, &ulLength, 
                                  pVblockIOArgs->lDevspacePage, pDeviCtrl->pIOCtrl->szDevspace );

  if ( ( rc != NO_ERROR ) || ( ulLengthWanted != ulLength ) )
    {
    DBG1  (( MF__, "Cannot read from devspace '%s', rc = %d",
             pDeviCtrl->pIOCtrl->szDevspace, rc ));

    MSGALL(( ERR_CANT_READ_FROM_DEVSPACE,
             pVblockIOArgs->lDevspacePage + 1,
             pDeviCtrl->pIOCtrl->szDevspace,
             rc, ulLength ));

    pVblockIOArgs->ulErrCode = ERRCODE_DEVIO_READ;
    pVblockIOArgs->fOk       = false;

    DBGOUT;
    return ( -1 );
    }

  #ifdef DEVELOP_CHECKS
   rc = __sql54k_check_filepos ( pDeviCtrl->hfDevspace,
                                 pVblockIOArgs->lDevspaceNo,
                                 pVblockIOArgs->lDevspacePage +
                                 pVblockIOArgs->sNumOfPages,
                                 0,
                                 pDeviCtrl->pIOCtrl->ulDevPages,
                                 &pVblockIOArgs->ulErrCode,
                                 NULL );

   if ( rc != NO_ERROR )
     {
     pVblockIOArgs->fOk = false;
     DBGOUT;
     return ( -1 );
     }
  #endif

  pDeviCtrl->ulOldIOPos = pVblockIOArgs->lDevspacePage +
                          pVblockIOArgs->sNumOfPages   + 1;
  pVblockIOArgs->fOk    = true;
  pDeviCtrl->DevIOStat.DevRead.ulCount++;
  pDeviCtrl->DevIOStat.DevRead.ulPageCount+= pVblockIOArgs->sNumOfPages;

  DBGOUT;
  return ( rc );
  }

/*------------------------------*/

/*
 * This function performs read io on the devspace
 */
static LONG  sql54k_devi_write ( PDEVI_CTRL_REC             pDeviCtrl,
                                 DLQ_VBLOCKIO_REQ_ARGS_REC  *pVblockIOArgs )
  {
  #undef  MF__
  #define MF__ MOD__"sql54k_devi_write"

  ULONG           ulFilePosLow;
  ULONG           ulFilePosHigh;
  ULONG           ulLengthWanted;
  ULONG           ulLength;
  APIRET          rc = NO_ERROR;

  DBGIN;

  if ( (ULONG)(pVblockIOArgs->lDevspacePage + 1) != pDeviCtrl->ulOldIOPos )
    {
    PAGES_TO_QWORD( ulFilePosLow, ulFilePosHigh,
                    pVblockIOArgs->lDevspacePage + 1 );

    rc = SET_FILE_PTR ( pDeviCtrl->hfDevspace, ulFilePosLow, FILE_BEGIN,
                        &ulFilePosLow, &ulFilePosHigh );

    if ( rc !=  NO_ERROR )
      {
      MSGALL(( ERR_CANT_SEEK_ON_DEVSPACE,
               pDeviCtrl->pIOCtrl->szDevspace, rc ));
      DBG1  (( MF__, "Cannot seek on devspace: %s, rc = %d",
               pDeviCtrl->pIOCtrl->szDevspace, rc ));
      pVblockIOArgs->ulErrCode = ERRCODE_DEVIO_SEEK;
      pVblockIOArgs->fOk       = false;

      DBGOUT;
      return ( rc );
      }
    }

  #ifdef DEVELOP_CHECKS
   rc = __sql54k_check_filepos ( pDeviCtrl->hfDevspace,
                                 pVblockIOArgs->lDevspaceNo,
                                 pVblockIOArgs->lDevspacePage,
                                 pVblockIOArgs->sNumOfPages,
                                 pDeviCtrl->pIOCtrl->ulDevPages,
                                 &pVblockIOArgs->ulErrCode,
                                 NULL );

   if ( rc != NO_ERROR )
     {
     pVblockIOArgs->fOk = false;
     DBGOUT;
     return ( -1 );
     }
  #endif

  ulLengthWanted = PAGES_TO_BYTES(pVblockIOArgs->sNumOfPages);

  rc = WRITE_FILE ( pDeviCtrl->hfDevspace, pVblockIOArgs->pBuffAddr,
                    ulLengthWanted, &ulLength );

  if ( ( rc != NO_ERROR ) || ( ulLengthWanted != ulLength ) )
    {
    DBG1  (( MF__, "Cannot write on devspace '%s', rc = %d",
             pDeviCtrl->pIOCtrl->szDevspace, rc ));

    MSGALL(( ERR_CANT_WRITE_ON_DEVSPACE,
             pVblockIOArgs->lDevspacePage + 1,
             pDeviCtrl->pIOCtrl->szDevspace, rc,
             ulLength ));

    pVblockIOArgs->ulErrCode = ERRCODE_DEVIO_WRITE;
    pVblockIOArgs->fOk       = false;

    DBGOUT;
    return ( -1 );
    }

  #ifdef DEVELOP_CHECKS
   rc = __sql54k_check_filepos ( pDeviCtrl->hfDevspace,
                                 pVblockIOArgs->lDevspaceNo,
                                 pVblockIOArgs->lDevspacePage +
                                 pVblockIOArgs->sNumOfPages,
                                 0,
                                 pDeviCtrl->pIOCtrl->ulDevPages,
                                 &pVblockIOArgs->ulErrCode,
                                 NULL );

   if ( rc != NO_ERROR )
     {
     pVblockIOArgs->fOk = false;
     DBGOUT;
     return ( -1 );
     }
  #endif

  pDeviCtrl->ulOldIOPos = pVblockIOArgs->lDevspacePage +
                          pVblockIOArgs->sNumOfPages   + 1;
  pVblockIOArgs->fOk    = true;
  pDeviCtrl->DevIOStat.DevWrite.ulCount++;
  pDeviCtrl->DevIOStat.DevWrite.ulPageCount+= pVblockIOArgs->sNumOfPages;

  DBGOUT;
  return ( rc );
  }

/*------------------------------*/

static VOID sql54k_wake_up_tasks ( PPIO_QUEUE_REC  ppIOWorkQue,
                                   PLONG           plIOWorkExclusive )
  {
  #undef  MF__
  #define MF__ MOD__"sql54k_wake_up_tasks"

  DBGPAS;

  while ( (*ppIOWorkQue)->pRequest != (PDLQ_REC) 0 )
    {
    // -- send reply back to UKT ( wake up task )
    REQ_FROM_DEVx_TO_UKT((*ppIOWorkQue)->pRequest);

    (*ppIOWorkQue)->pRequest = 0;
    (*ppIOWorkQue)           = (*ppIOWorkQue)->pNextElem;
    }

  return;

  }

/*------------------------------*/

static LONG sql54k_lock_drive ( HFILE hfFileHandle )
  {
  #undef  MF__
  #define MF__  MOD__"sql54k_lock_drive"
  PVOID   ParmList        = NULL;
  ULONG   ParmLengthMax   = 0;
  ULONG   ParmLengthInOut = 0;
  PUCHAR  DataArea        = NULL;
  ULONG   DataLengthMax   = 0;
  ULONG   DataLengthInOut = 0;
  LONG    rc              = NO_ERROR;

  DBGPAS;

  #if defined(_WIN32)
    if (!DeviceIoControl((HANDLE)hfFileHandle, FSCTL_LOCK_VOLUME,
                         NULL, 0, NULL, 0, &DataLengthInOut, NULL))
        rc = GetLastError();
  #else
    rc = DosDevIOCtl( hfFileHandle, IOCTL_DISK, DSK_LOCKDRIVE,
                      ParmList, ParmLengthMax, &ParmLengthInOut,
                      DataArea, DataLengthMax, &DataLengthInOut);
  #endif

  return( rc );
  }

/*------------------------------*/

static LONG sql54k_unlock_drive ( HFILE hfFileHandle )
  {
  #undef  MF__
  #define MF__  MOD__"sql54k_unlock_drive"
  PVOID   ParmList        = NULL;
  ULONG   ParmLengthMax   = 0;
  ULONG   ParmLengthInOut = 0;
  PUCHAR  DataArea        = NULL;
  ULONG   DataLengthMax   = 0;
  ULONG   DataLengthInOut = 0;
  LONG    rc              = NO_ERROR;

  DBGPAS;

  if (!DeviceIoControl((HANDLE)hfFileHandle, FSCTL_UNLOCK_VOLUME,
                        NULL, 0, NULL, 0, &DataLengthInOut, NULL))
      rc = GetLastError();

  return( rc );
  }

/*------------------------------*/

_INLINE VOID __sql54k_req_to_devi ( PDEVI_CTRL_REC  pDeviCtrl,
                                    PDLQ_REC        pRequest )
  {
  #undef  MF__
  #define MF__  MOD__"__sql54k_req_to_devi"

  DBGPAS;

  WAIT_UNTIL_ALONE( &pDeviCtrl->lExclusive );

  pDeviCtrl->pIOReqQue->pRequest = pRequest;
  pDeviCtrl->pIOReqQue           = pDeviCtrl->pIOReqQue->pNextElem;
  pDeviCtrl->DevIOStat.ulIOQueLen++;

  CLEARLOCK( &pDeviCtrl->lExclusive );

  if (pDeviCtrl->DevIOStat.ulIOQueLen > pDeviCtrl->DevIOStat.ulMaxIOQueLen)
      pDeviCtrl->DevIOStat.ulMaxIOQueLen = pDeviCtrl->DevIOStat.ulIOQueLen;

  sql72k_wake_kt ( &pDeviCtrl->ThrdCtrlHeader );
  }

/*------------------------------*/

_INLINE VOID __sql54k_req_from_dev_worker_to_ukt ( BOOL         fOk,
                                                   ULONG        ulNumOfBytes,
                                                   LONG         lDevspaceNo,
                                                   LPOVERLAPPED pCompletedOverlapped,
                                                   LONG         rc )
  {
  #undef  MF__
  #define MF__  MOD__"__sql54k_req_from_dev_worker_to_ukt"
  PDLQ_REC                       pRequest;
  PTASK_CTRL_REC                 pTaskCtrl;

  DBGPAS;

  //
  // --- compute the base pointer to the task control record
  pTaskCtrl = (PTASK_CTRL_REC)((PCHAR)pCompletedOverlapped -
                               offsetof(TASK_CTRL_REC, Overlapped));

  pRequest                                        = &pTaskCtrl->SpecDLQElem1;
  pRequest->pTaskCtrl                             = pTaskCtrl;
  pRequest->ulReqType                             = REQ_OVERLAPPED_IO_FINISED;
  pRequest->ReqArgs.DevWorkerRequest.lDevspaceNo  = lDevspaceNo;
  pRequest->ReqArgs.DevWorkerRequest.ulNumOfBytes = ulNumOfBytes;
  pRequest->ReqArgs.DevWorkerRequest.fOk          = fOk;
  pRequest->ReqArgs.DevWorkerRequest.APIRetCode   = rc;

  DBG3 (( MF__, "Send pRequest: 0x%x to TID:%u, [T:0x%03u]", pRequest,
        pRequest->pTaskCtrl->pUKT->ThrdCtrlHeader.Tid,
        pRequest->pTaskCtrl->ulTaskIndex ));

  sql73k_UKT_enqu ( pRequest->pTaskCtrl, pRequest );
  sql72k_wake_ukt ( pRequest->pTaskCtrl->pUKT );
  }

/*------------------------------*/

_INLINE VOID __sql54k_req_from_devx_to_ukt ( PDLQ_REC pRequest )
  {
  #undef  MF__
  #define MF__  MOD__"__sql54k_req_from_devx_to_ukt"

  DBGPAS;

  DBG3 (( MF__, "Send pRequest: 0x%x to TID:%u, [T:0x%03u]", pRequest,
        pRequest->pTaskCtrl->pUKT->ThrdCtrlHeader.Tid,
        pRequest->pTaskCtrl->ulTaskIndex ));

  sql73k_UKT_enqu ( pRequest->pTaskCtrl, pRequest );
  sql72k_wake_ukt ( pRequest->pTaskCtrl->pUKT );
  }


/*------------------------------*/

_INLINE VOID __sql54k_req_to_dev0 ( PDLQ_REC        pRequest,
                                    PTASK_CTRL_REC  pTaskCtrl )
  {
  #undef  MF__
  #define MF__  MOD__"__sql54k_req_to_dev0"

  DBGPAS;

  DBG3(( MF__, "[T:0x%03u] Send pRequest: 0x%x, using (Dev0) pIOReqQue: 0x%x",
        pTaskCtrl->ulTaskIndex, pRequest, kgs.Dev0.pIOReqQue ));

  WAIT_UNTIL_ALONE( &kgs.Dev0.lExclusive );

  kgs.Dev0.pIOReqQue->pRequest = pRequest;
  kgs.Dev0.pIOReqQue           = kgs.Dev0.pIOReqQue->pNextElem;

  CLEARLOCK( & kgs.Dev0.lExclusive );

  sql72k_wake_kt ( &kgs.Dev0.ThrdCtrlHeader );
  }


/*------------------------------*/


_INLINE BOOL __sql54k_is_user_req_io_error ( LONG            lDevspaceNo,
                                             tsp2_io_op_code OpCode )

  {
  #undef  MF__
  #define MF__  MOD__"__sql54k_is_user_req_io_error"

  PDEVI_CTRL_REC             pDeviCtrl ;
  BOOL                       fForceError    ;

  DBGPAS;

  pDeviCtrl = kgs.pFirstDeviCtrl + lDevspaceNo * XPARAM(lDeviPerDevspace) ;

  fForceError =  pDeviCtrl->lForceIOError && 
                (pDeviCtrl->lForceIOError & ~0x100) == OpCode ; 

  if ( fForceError )
    pDeviCtrl->lForceIOError = false ;

  return ( fForceError ) ;
  }

/*------------------------------*/

_INLINE HANDLE __sql54k_create_file( LPCTSTR                 lpFileName,
                                     DWORD                   dwDesiredAccess,
                                     DWORD                   dwShareMode,
                                     LPSECURITY_ATTRIBUTES   lpSecurityAttributes,
                                     DWORD                   dwCreationDisposition,
                                     DWORD                   dwFlagsAndAttributes,
                                     HANDLE                  hTemplateFile )
  {
  #undef  MF__
  #define MF__ MOD__"__sql54k_create_file"

  HANDLE Handle;
  INT    Retry  = NUM_OF_SHARING_VIOLATION_RETRIES;
  LONG   rc     = NO_ERROR;

  do
    {
    Handle = CreateFile( lpFileName, dwDesiredAccess, dwShareMode,
                         lpSecurityAttributes, dwCreationDisposition,
                         dwFlagsAndAttributes, hTemplateFile );

    if ( Handle == INVALID_HANDLE_VALUE )
      {
      rc = GetLastError();

      if ( rc == ERROR_SHARING_VIOLATION )
        SLEEP(10000 * ((INT)NUM_OF_SHARING_VIOLATION_RETRIES + 1 - Retry)); // linger on sharing violation
      }
    else 
      rc = NO_ERROR;
    }
  while (( rc == ERROR_SHARING_VIOLATION ) && --Retry );

  return Handle;
  }

/*------------------------------*/

#ifdef DEVELOP_CHECKS
 _INLINE LONG __sql54k_check_filepos ( HFILE         hfFileHandle,
                                       ULONG         lDevspaceNo,
                                       tsp_page_no   lDevspacePage,
                                       SHORT         sNumOfPages,
                                       ULONG         ulDevPages,
                                       PULONG        pulErrCode,
                                       ERRORTEXT     pcErrText )
   {
   #undef  MF__
   #define MF__  MOD__"__sql54k_check_filepos"

   ULONG           ulFilePosLow;
   ULONG           ulFilePosHigh;
   ULONG           ulCurrDevPage = (ULONG)UNDEF;
   APIRET          rc            = NO_ERROR;

   if ((ULONG)( lDevspacePage + sNumOfPages + 1 ) > ulDevPages )
     {
     if (pulErrCode)
       *pulErrCode = ERRCODE_DEVIO_DEV_POS;
     if (pcErrText)
       sql46c_build_error_string ( pcErrText, ERRMSG_DEVIO_DEV_POS, 0 );

     MSGALL(( ERR_WRNG_DEVSPACE_POS, lDevspacePage, sNumOfPages ));
     DBG1  (( MF__, "Wrong devspace write position: %u - Pages: %d",
              lDevspacePage, sNumOfPages ));

     DBGOUT;
     return ( -1 );
     }

   ulFilePosHigh = 0;
   rc = SET_FILE_PTR ( hfFileHandle, 0, FILE_CURRENT,
                       &ulFilePosLow, &ulFilePosHigh );

   if (rc == NO_ERROR)
     {
     if ((HIGH_VALUE_VALID(ulFilePosHigh) == true) &&
         (LOW_VALUE_VALID(ulFilePosLow)   == true))
       ulCurrDevPage = QWORD_TO_PAGES(ulFilePosLow, ulFilePosHigh) - 1;
     else
       ulCurrDevPage = MAX_DEVSPACE_PAGES + 1;
     }
   else
     {
     if (pulErrCode)
       *pulErrCode = ERRCODE_DEVIO_SEEK;
     if (pcErrText)
       sql46c_build_error_string ( pcErrText, ERRMSG_DEVIO_SEEK_ON_DEVSPACE, rc );

     MSGALL(( ERR_CANT_SEEK_ON_DEVSPACE,
              kgs.Dev0.pIOCtrl[lDevspaceNo].szDevspace, rc ));
     DBG1  (( MF__, "Cannot seek on devspace: %s, rc = %d",
              kgs.Dev0.pIOCtrl[lDevspaceNo].szDevspace, rc ));

     DBGOUT;
     return ( rc );
     }

   if ( ulCurrDevPage != (ULONG)lDevspacePage )
     {
     if (pulErrCode)
       *pulErrCode = ERRCODE_DEVIO_SEEK_POS;
     if (pcErrText)
       sql46c_build_error_string ( pcErrText, ERRMSG_DEVIO_SEEK_POS, rc );

     MSGALL(( ERR_WRNG_DEVSPACE_SEEK_POS,
              ulCurrDevPage, lDevspacePage ));
     DBG1  (( MF__, "Wrong devspace position. current: %u - expected: %u",
              ulCurrDevPage, lDevspacePage ));

     DBGOUT;
     return ( -1 );
     }

   DBGOUT;
   return ( rc );
   }
#endif
