/****************************************************************************

  module      : vos541.c

  -------------------------------------------------------------------------

  responsible : FrankH

  special area: Intel PSE 36 support 
  description : ramdisk


  last changed: 1999-05-28  10:36
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1998



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include <windows.h>
#include <winioctl.h>
#include "intel/pse36api.h"
#include "gos541.h"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#ifdef NO_INLINES
# define _INLINE
#else
# define _INLINE	__inline
#endif

#define PSE36_REGISTRY_ENTRY "SYSTEM\\CurrentControlSet\\Services\\"

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/



/*===========================================================================*
 *  GLOBAL VARIABLES                                                         *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL VARIABLES                                                          *
 *===========================================================================*/

static LONGLONG os541LastOffset = 0 ; /* !!! for virtual file (diagnose) only !!!*/

/*===========================================================================*
 *  STATIC/INLINE FUNCTIONS (PROTOTYPES)                                     *
 *===========================================================================*/



/*===========================================================================*
 *  PUBLIC FUNCTIONS (CODE)                                                  *
 *===========================================================================*/
ULONG os541IsPSE36(char *FileName)
{
  CHAR driveName[MAX_PATH];

  os541PSE36GetDeviceName(driveName, sizeof(driveName) ) ;

  return( !strcmp( FileName, driveName ) );
}

/*------------------------------*/

char *os541PSE36GetDeviceName (char *DeviceName, 
                               int  MaxNameLen )
{
    CHAR regPath[] = PSE36_REGISTRY_ENTRY;
    CHAR deviceRegEntry[MAX_PATH];
    PCHAR deviceKey;
    HKEY hKeyDriver;
    LONG status;
    ULONG type;
    ULONG size;

    //
    // Determine the name of the ram disk device
    //

    // build up registry path

    strcpy(deviceRegEntry, regPath);
    strcat(deviceRegEntry, PSE36_DEVICE_NAME);
    strcat(deviceRegEntry, "\\");
    deviceKey = deviceRegEntry + strlen(deviceRegEntry);
    strcpy(deviceKey, "Parameters");

    status = RegOpenKeyEx (
                           HKEY_LOCAL_MACHINE,
                           deviceRegEntry,
                           0L,
                           KEY_READ,
                           &hKeyDriver);
	
    if (status == ERROR_SUCCESS) {
        strcpy(DeviceName, "\\\\.\\");
        size = MaxNameLen - 4 ;
        status = RegQueryValueEx(
                                 hKeyDriver, 
                                 "DriveName",
                                 0L,
                                 &type,
                                 (LPBYTE)&DeviceName[4],
                                 &size);
    }

    if (status != ERROR_SUCCESS || DeviceName[4] == '\0') {

        strncpy(DeviceName, PSE36_DEFAULT_DOS_NAME, MaxNameLen);
    }

    RegCloseKey (hKeyDriver); // close key to registry

    return(DeviceName);
}

/*------------------------------*/

DWORD
os541PSE36GetSize( char     *PSE36DriveName,
                   LONGLONG *SizeAvailable )
/*
 *  Initialize the ram disk.  Size is the requested size of the ram disk.
 *
 *  Arguments:
 *
 *     Size - request size of the ram disk in bytes, this value may be ignored.
 *
 *  Returns:
 *
 *     Actual size in bytes of the ram disk.  This may be either small or
 *     larger than the request size.  Returns 0 to indicate ram disk was
 *     not initialized.
 *
 */

{
    HANDLE     hRamDisk;
    PSE36_SIZE diskSize;
    DWORD      rc = NO_ERROR ;
    ULONG      size;

    *SizeAvailable = 0 ;
    if ( PSE36DriveName == NULL )
      PSE36DriveName = PSE36_DEFAULT_DOS_NAME ;

    /*  Try to open it as an asynchronous file */

    if ((hRamDisk = CreateFile(PSE36DriveName,
                         GENERIC_READ|GENERIC_WRITE,
                         0,
                         NULL,
                         OPEN_EXISTING, 
                         FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS |
                         FILE_FLAG_NO_BUFFERING | FILE_FLAG_OVERLAPPED,
                         NULL)) == INVALID_HANDLE_VALUE) {

        //
        // Could not ope ram disk return 0 for length
        //

        return (GetLastError());
    }


    if ( DeviceIoControl(hRamDisk,
                         (DWORD)IOCTL_PSE36_GET_SIZE,
                         NULL,
                         0,
                         &diskSize,
                         sizeof(diskSize),
                         &size,
                         NULL))
      *SizeAvailable = diskSize.Size ;
    else
       rc = GetLastError();

    CloseHandle(hRamDisk);

    return(rc);
}


/*------------------------------*/

DWORD os541PSE36Open( char     *PSE36DriveName, 
                      HANDLE   *hRamDisk )
/*
 *  Open the ram disk.  Size is the requested size of the ram disk.
 *
 *  Arguments:
 *
 *     Size - request size of the ram disk in bytes, this value may be ignored.
 *
 *  Returns:
 *
 *     Actual size in bytes of the ram disk.  This may be either small or
 *     larger than the request size.  Returns 0 to indicate ram disk was
 *     not initialized.
 *
 */

{
    DWORD rc ;

    if ( PSE36DriveName == NULL )
      PSE36DriveName = PSE36_DEFAULT_DOS_NAME ;

    /*  Try to open it as an asynchronous file */
    if ((*hRamDisk = CreateFile(PSE36DriveName,
                         GENERIC_READ|GENERIC_WRITE,
                         0,
                         NULL,
                         OPEN_EXISTING, 
                         FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS |
                         FILE_FLAG_NO_BUFFERING | FILE_FLAG_OVERLAPPED,
                         NULL)) == INVALID_HANDLE_VALUE)
        //
        // Could not ope ram disk return 0 for length
        //
        rc = GetLastError();
    else
        rc = NO_ERROR;
 
   return(rc);
}

/*------------------------------*/

DWORD os541PSE36Close( HANDLE   hRamDisk )
/*
 *  Close the ram disk and release all resources.
 *
 *  Arguments:
 *
 *     None
 *
 *  Returns:
 *
 *     None
 */
{
   DWORD rc ;
   rc = CloseHandle(hRamDisk) == 0 ? NO_ERROR : GetLastError() ;
  
   return(rc);
}


/*------------------------------*/

DWORD os541PSE36Read (
    HANDLE   hRamDisk,
    PVOID    Buffer,
    ULONG    Length,
    LONGLONG Offset,
    ULONG    *read )
/*
 *  Read data from the ram disk
 *
 *  Arguments:
 *
 *     Buffer - address of buffer to receive data
 *
 *     Length - number of bytes to be read
 *
 *     Offset - ram disk offset to read from
 *
 *  Returns:
 *
 *     Number of bytes sucessfully read.
 *
 */

{
    DWORD rc ;
    PPSE36_SCATTER_GATHER_IO pIoRequest;
    ULONG  request[(sizeof(PSE36_SCATTER_GATHER_IO) +
                    sizeof(PSE36_SCATTER_GATHER_DESCRIPTOR)) /
                   sizeof (ULONG)];

    *read = 0;

    pIoRequest = (PPSE36_SCATTER_GATHER_IO)request;

    pIoRequest->DescriptorCount = 1;
    pIoRequest->ScatterGatherList[0].Offset = Offset;
    pIoRequest->ScatterGatherList[0].Buffer = Buffer;
    pIoRequest->ScatterGatherList[0].BufferSize = Length;

    //
    // Use the DeviceIoControl with the file discriptor opened for
    // asynchronous I/O (FILE_FLAG_OVERLAPPED set in the flags and
    // atrribute argument to CreateFile).  Asynchronous calls to
    // ReadFile cause context switches.  Synchronous calls to
    // ReadFile are serialized with other synchronous calls on the
    // same file descriptor.  The DeviceIoControl can be done
    // synchronoulsy (no context switch) and is not serialized
    // because the file is opened for asyncrhonous I/O.  While the
    // other methods will work, this has by far the lowest overhead,
    // and therefore is the fastest.
    //

    if (DeviceIoControl(hRamDisk,
                        (DWORD)IOCTL_PSE36_READ_SCATTER,
                        pIoRequest,
                        sizeof(request),
                        NULL,
                        0,
                        read,
                        NULL)) 
        rc = NO_ERROR;
      else
        rc = GetLastError();
      return(rc);
}


/*------------------------------*/


DWORD os541PSE36Write (
    HANDLE   hRamDisk,
    PVOID    Buffer,
    ULONG    Length,
    LONGLONG Offset,
    ULONG    *written )
/*
 *  Write data to the ram disk
 *
 *  Arguments:
 *
 *     Buffer - address of buffer that contains the data.
 *
 *     Length - number of bytes to be written
 *
 *     Offset - ram disk offset to write to
 *
 *  Returns:
 *
 *     Number of bytes sucessfully read.
 *
 */

{
    DWORD rc ;
    PPSE36_SCATTER_GATHER_IO pIoRequest;
    ULONG  request[(sizeof(PSE36_SCATTER_GATHER_IO) +
                    sizeof(PSE36_SCATTER_GATHER_DESCRIPTOR)) /
                   sizeof (ULONG)];

    *written = 0;

    pIoRequest = (PPSE36_SCATTER_GATHER_IO)request;

    pIoRequest->DescriptorCount = 1;
    pIoRequest->ScatterGatherList[0].Offset = Offset;
    pIoRequest->ScatterGatherList[0].Buffer = Buffer;
    pIoRequest->ScatterGatherList[0].BufferSize = Length;

    //
    // Use the DeviceIoControl with the file discriptor opened for
    // asynchronous I/O (FILE_FLAG_OVERLAPPED set in the flags and
    // atrribute argument to CreateFile).  Asynchronous calls to
    // WriteFile cause context switches.  Synchronous calls to
    // WriteFile are serialized with other synchronous calls on the
    // same file descriptor.  The DeviceIoControl can be done
    // synchronoulsy (no context switch) and is not serialized
    // because the file is opened for asyncrhonous I/O.  While the
    // other methods will work, this has by far the lowest overhead,
    // and therefore is the fastest.
    //

    if (DeviceIoControl(hRamDisk,
                        (DWORD)IOCTL_PSE36_WRITE_GATHER,
                        pIoRequest,
                        sizeof(request),
                        NULL,
                        0,
                        written,
                        NULL))
        rc = NO_ERROR;
      else
        rc = GetLastError();

      return(rc);
}

/*------------------------------*/

/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
/* ATTENTION: This is a HACK                                          */
/* The following function should only be used by the virtual file     */
/* "veo06c" to get access to the PSE device via "diagnose"!           */
/* The problem is. that "seek" is not allowed on a PSE36 device.      */
/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */

LONG sql541PSE36Seek( HANDLE   hRamDisk,
                      LONGLONG Offset,
                      DWORD    Method)
{
  switch (Method)
  {
  case FILE_BEGIN:
    os541LastOffset = Offset ;
    break ;

  case FILE_CURRENT:
    os541LastOffset += Offset ;
    break ;

  case FILE_END :
    os541LastOffset = -1 ;
    break ;

  default:
    os541LastOffset = -1 ;
    break ;
  }
  return( (LONG )(os541LastOffset) ) ;
}

/*------------------------------*/

LONG sql541PSE36Read( HANDLE   hRamDisk,
                      PVOID    Buffer,
                      ULONG    Length,
                      ULONG   *BytesRead )
{
  DWORD rc = os541PSE36Read (hRamDisk, Buffer, Length, os541LastOffset, BytesRead ) ;

  if ( rc == NO_ERROR )
    os541LastOffset += *BytesRead ;
  return (rc);
}

/*===========================================================================*
 *  LOCAL/PRIVATE/PROTECTED FUNCTIONS (CODE)                                 *
 *===========================================================================*/



/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/