/****************************************************************************

  module      : vos542.c

  -------------------------------------------------------------------------

  responsible : FrankH

  special area: Intel PSE 36 support 
  description : special "vblockio"


  last changed: 1999-08-20  16:59
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1999



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "vsp002.h"
#include "geo00_2.h"
#include "geo002.h"
#include "heo46.h"
#include "heo54k.h"
#include "gos00.h"
#include "vos00kc.h"
#include "geo007_1.h"
#include "geo007_2.h"
#include "gos541.h"
#include "gos542.h"
#include "gos543.h"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/
#define MOD__  "VOS542.C : "

#define PSEDBG                         11111,INFO_TYPE,"PSE36DBG"

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

#define WRITE_TO_PSE36_ONLY         0
#define WRITE_TO_PSE36_AND_DEVSPACE 1
#define WRITE_TO_DEVSPACE_ONLY      2


#define PSE36_IS_DATAPAGE_IN_RANGE(_PageNo)  ( (_PageNo) < kgs.PSE36Ctrl->PSE36Stat.MaxDataPages)

/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/

/*===========================================================================*
 *  GLOBAL VARIABLES                                                         *
 *===========================================================================*/


/*===========================================================================*
 *  LOCAL VARIABLES                                                          *
 *===========================================================================*/


/*===========================================================================*
 *  STATIC/INLINE FUNCTIONS (PROTOTYPES)                                     *
 *===========================================================================*/

_INLINE HANDLE __os542PSE36GetHandle ( HANDLE         *PSE36Handle,
                                       tsp_errtext     errtext,
                                       BOOLEAN        *ok );

_INLINE void __os542SetPageCached  ( tsp_page_no _PageNo );

_INLINE void __os542ResetPageCached( tsp_page_no _PageNo );

_INLINE bool __os542IsPageCached   ( tsp_page_no _PageNo );

_INLINE void __os542SetPageChanged ( tsp_page_no _PageNo );

_INLINE void __os542ResetPageChanged(tsp_page_no _PageNo );

_INLINE bool __os542IsPageChanged  ( tsp_page_no _PageNo );

_INLINE tsp_page_no __os542GetNextChangedPage() ;

_INLINE void  __os542PSE36GetMonitorStatistics ( tsp00_Longreal        TotalPSEReads ,
                                                 tsp00_Longreal        TotalPSEReadsFail,
                                                 tsp00_Longreal        TotalPSEWrites, 
                                                 tsp00_Longreal        TotalPSEWritesFail ,
                                                 tsp00_Longreal       *TotalAccess, 
                                                 tsp00_Longreal       *TotalHits,
                                                 tsp00_Longreal       *TotalMiss);

static DWORD  
os542PSE36GetSizeInPages ( int *SizeInPages ) ;

static DWORD  
os542PSE36AllocCtrlStruc ( void **PSE36CtrlStruc,
                           int    Size,
                           char  *CtrlStructId );

static void  
os542HandleDataRead   ( tsp_int4              devno,
                        tsp_page_no           offset,
                        tsp_page_no           PageNo,
                        PTASK_CTRL_REC        pTaskCtrl,
                        BUFFERADDRESS         bufadr,
                        tsp_int2              sNumOfPages,
                        PROCESS_ID            pid,
                        tsp_errtext           ErrText,
                        BOOLEAN              *ok );

static void  
os542HandleDataWrite  ( tsp_int4              devno,
                        tsp_page_no           offset,
                        tsp_page_no           PageNo,
                        PTASK_CTRL_REC        pTaskCtrl,
                        BUFFERADDRESS         bufadr,
                        tsp_int2              sNumOfPages,
                        PROCESS_ID            pid,
                        tsp_errtext           ErrText,
                        BOOLEAN              *ok );

_INLINE void  
__os542GetCachedPage        ( tsp_page_no           PageNo,
                              PTASK_CTRL_REC        pTaskCtrl,
                              BUFFERADDRESS         bufadr,
                              tsp_errtext           ErrText,
                              BOOLEAN              *ok );

static void  
os542GetCachedPageFromPSE36 ( tsp_page_no           PageNo,
                              PTASK_CTRL_REC        pTaskCtrl,
                              BUFFERADDRESS         bufadr,
                              tsp_errtext           ErrText,
                              BOOLEAN              *ok );

static void
os542GetCachedPageFromAWE  ( tsp_page_no           PageNo,
                             PTASK_CTRL_REC        pTaskCtrl,
                             BUFFERADDRESS         bufadr,
                             tsp_errtext           ErrText,
                             BOOLEAN              *ok );

_INLINE void  
__os542CacheDataPage      ( tsp_page_no           PageNo,
                            PTASK_CTRL_REC        pTaskCtrl,
                            BUFFERADDRESS         bufadr,
                            tsp_errtext           ErrText,
                            BOOLEAN              *ok );
static void  
os542CacheDataPageToPSE36 ( tsp_page_no           PageNo,
                            PTASK_CTRL_REC        pTaskCtrl,
                            BUFFERADDRESS         bufadr,
                            tsp_errtext           ErrText,
                            BOOLEAN              *ok );


static void  
os542CacheDataPageToAWE   ( tsp_page_no           PageNo,
                            PTASK_CTRL_REC        pTaskCtrl,
                            BUFFERADDRESS         bufadr,
                            tsp_errtext           ErrText,
                            BOOLEAN              *ok );


static DWORD  os542AWE_Init (int *SizeInPages) ;

/*===========================================================================*
 *  PUBLIC FUNCTIONS (CODE)                                                  *
 *===========================================================================*/

pasbool  vPSE36DataCacheActive()
{
  #undef  MF__
  #define MF__  MOD__"vPSE36DataCacheActive"

  return ( kgs.PSE36Ctrl->PSE36Stat.MaxDataPages > 0 ) ;
}

/*------------------------------*/

tsp_page_no  vPSE36GetLastDataPno()
{
  #undef  MF__
  #define MF__  MOD__"vPSE36GetLastDataPno"

  DBGPAS;

  return ( kgs.PSE36Ctrl->PSE36Stat.MaxDataPages - 1 ) ;
}

/*------------------------------*/

void  vPSE36WriteToPSE36AndDevspace()
{
  #undef  MF__
  #define MF__  MOD__"vPSE36WriteToPSE36AndDevspace"

  DBGPAS;

  // From now, all write operations will be done to the PSE36 Device and
  // the disk. 
  // This function is called by the kernel at start of  savepoint / checkpoint,
  // to indicate that the flush of the datacache will be done to the the physically
  // devspace too.

  kgs.PSE36Ctrl->WriteMode = WRITE_TO_PSE36_AND_DEVSPACE ;
}

/*------------------------------*/

void  vPSE36WriteToDevspaceOnly()
{
  #undef  MF__
  #define MF__  MOD__"vPSE36WriteToDevspaceOnly"

  DBGPAS;

  // From now, all write operations will be done to the devspace only. 
  // This function is called by the kernel in time of savepoint / checkpoint 
  // activity, after the flush of the datacache has been completed.
  // Now the kernel has to flush the changed datapages of the PSE36, so all
  // write operations contains pages from the PSE36 only. 
  // It makes no sense to copy the same pages to the same position within the
  // PSE36.

  kgs.PSE36Ctrl->WriteMode = WRITE_TO_DEVSPACE_ONLY;
}

/*------------------------------*/
void  vPSE36WriteToPSE36Only()
{
  #undef  MF__
  #define MF__  MOD__"vPSE36WriteToPSE36Only"

  DBGPAS;

  // This function is called by the kernel at end of  savepoint / checkpoint,

  kgs.PSE36Ctrl->WriteMode = WRITE_TO_PSE36_ONLY ;
}


/*------------------------------*/

void  vPSE36FreeDataPage(tsp_page_no           PageNo)
{
  #undef  MF__
  #define MF__  MOD__"vPSE36FreePage"

  DBGPAS;

  if ( PSE36_IS_DATAPAGE_IN_RANGE(PageNo))
  {
    WAIT_UNTIL_ALONE( &kgs.PSE36Ctrl->Exclusive );

    if ( __os542IsPageChanged(PageNo) )
      __os542ResetPageChanged(PageNo) ;

    __os542ResetPageCached(PageNo);

    CLEARLOCK( &kgs.PSE36Ctrl->Exclusive );
  }
}

/*------------------------------*/

DWORD  os542PSE36Init ()
{
  #undef  MF__
  #define MF__  MOD__"os542PSE36Init"

  DWORD    rc ;

  rc = os542PSE36AllocCtrlStruc ( &kgs.PSE36Ctrl,sizeof(tos542_PSE36Ctrl),"PSE36Ctrl");

  if ( sql02_get_platform_id () == VER_PLATFORM_WIN32_NT &&
  	   XPARAM(fUseMemEnhancement) && rc == NO_ERROR )
  {
    kgs.PSE36Ctrl->DBPageSize  = XPARAM(ulDBPageSize) ;

    rc = os542PSE36GetSizeInPages(&kgs.PSE36Ctrl->PSE36Stat.TotalPages);

    if ( rc != NO_ERROR ) /* PSE not available, try AWE */
      rc = os542AWE_Init(&kgs.PSE36Ctrl->PSE36Stat.TotalPages);

    if ( rc == NO_ERROR )
    {
      kgs.PSE36Ctrl->InfoPagePos = --kgs.PSE36Ctrl->PSE36Stat.TotalPages ;

      kgs.PSE36Ctrl->PSE36Stat.MaxDataPages  = kgs.PSE36Ctrl->PSE36Stat.TotalPages;

      if ( kgs.PSE36Ctrl->PSE36Stat.MaxDataPages > XPARAM(lMaxDataPages) )
        kgs.PSE36Ctrl->PSE36Stat.MaxDataPages = XPARAM(lMaxDataPages) ;

      MSGD (( INFO_PSE36_PAGES_USED_FOR, 
              kgs.PSE36Ctrl->PSE36Stat.MaxDataPages , "data pages", 1, "info page(s)" ));

      rc = os542PSE36AllocCtrlStruc ( &kgs.PSE36Ctrl->DataPageArray,
                                      kgs.PSE36Ctrl->PSE36Stat.MaxDataPages,"DataPages");

      if ( rc == NO_ERROR )
      {
        kgs.PSE36Ctrl->LowestChangedDataPage     = kgs.PSE36Ctrl->PSE36Stat.MaxDataPages - 1;
        vPSE36WriteToPSE36Only();
      }
    }
  }
  return(rc);
}

/*------------------------------*/

void  vPSE36DataBlockIO ( tsp_int4              devno,
                          tsp_page_no           offset,
                          tsp_page_no           PageNo,
                          tsp2_io_op_code_Param op_code,
                          BUFFERADDRESS         bufadr,
                          tsp_int2              buf_count,
                          PROCESS_ID            pid,
                          tsp_errtext           errtext,
                          BOOLEAN              *ok )
{
  #undef  MF__
  #define MF__  MOD__"vPSE36DataBlockIO"

  if ( kgs.PSE36Ctrl->PSE36Stat.MaxDataPages > 0 )
  {
    PTASK_CTRL_REC pTaskCtrl = &kgs.pFirstTaskCtrl[pid - 1];
    switch ( op_code )
    {
      case sp2ioc_read :
           os542HandleDataRead (devno, offset, PageNo,
                                pTaskCtrl, bufadr, buf_count,
                                pid, errtext, ok ) ;
           break ;
      case sp2ioc_write:
           if ( kgs.PSE36Ctrl->WriteMode != WRITE_TO_DEVSPACE_ONLY ) 
           {
             os542HandleDataWrite(devno, offset, PageNo,
                                  pTaskCtrl, bufadr, buf_count,
                                  pid, errtext, ok ) ;
           }
           else
             vblockio (devno, offset, (tsp2_io_op_code )op_code, bufadr, buf_count, 
                       pid, errtext, ok ) ;

           break ;
      default       : 
           sql46c_build_error_string ( errtext, ERRMSG_INVALID_IO_REQ, op_code );
           *ok = FALSE ; 
           break ;
      }
  }
  else
  {
    /* PSE36 not active, use normal vblockio */
    vblockio (devno, offset, (tsp2_io_op_code )op_code, bufadr, buf_count, 
              pid, errtext, ok ) ;
  }
}

/*------------------------------*/

void  vPSE36GetDataBlockToSave ( PROCESS_ID            pid,
                                 BUFFERADDRESS         bufadr,
                                 tsp_int2              NumBlocksWanted,
                                 tsp_int2             *NumBlocksFound,
                                 tsp_errtext           errtext,
                                 BOOLEAN              *ok )
{
  #undef  MF__
  #define MF__  MOD__"vPSE36GetDataBlockToSave"


  *NumBlocksFound = 0 ;
  *ok             = 1 ;

  if ( kgs.PSE36Ctrl->PSE36Stat.MaxDataPages > 0 )
  {

    PTASK_CTRL_REC                pTaskCtrl = &kgs.pFirstTaskCtrl[pid - 1];
    PUKT_CTRL_REC                 pUKT      = pTaskCtrl->pUKT;
    BUFFERADDRESS                 ActPage = bufadr ;

    pTaskCtrl->TaskState = TSK_PSE36_DATA_READ_EO00;
    WAIT_UNTIL_ALONE( &kgs.PSE36Ctrl->Exclusive );

    if ( kgs.PSE36Ctrl->PSE36Stat.NumChangedDataPages  > 0 && NumBlocksWanted > 0)
    {
      tsp_page_no                   PageNo ;
      do 
	  {
        PageNo = __os542GetNextChangedPage() ;
        DBG3 ((MF__, "Block[%d], ActualPageNo = %d", *NumBlocksFound, PageNo ));
        if ( PageNo >= 0 )
		{
          __os542GetCachedPage( PageNo, pTaskCtrl, ActPage, errtext, ok );
          if ( *ok )

		  {
            __os542ResetPageChanged(PageNo) ;
            ActPage++ ;
            (*NumBlocksFound)++ ;
            DBG3((MF__, "(%d/%d) : pno %d, remaining %d", *NumBlocksFound, 
                         NumBlocksWanted, PageNo, kgs.PSE36Ctrl->PSE36Stat.NumChangedDataPages));
		  }
		}
	  }
      while ( PageNo >= 0 && *ok && 
              kgs.PSE36Ctrl->PSE36Stat.NumChangedDataPages > 0 &&
              NumBlocksWanted > *NumBlocksFound) ;
    }

    CLEARLOCK( &kgs.PSE36Ctrl->Exclusive );
    pTaskCtrl->TaskState = TSK_RUNNING_EO00;

    DBG3 ((MF__, "wanted = %d, found = %d", NumBlocksWanted, *NumBlocksFound ));
  }
}

/* PTS 1103699 */
/*------------------------------*/

void  vPSE36MonitorDataCache( tsp00_Longreal       *TotalAccess,
                              tsp00_Longreal       *TotalHits,
                              tsp00_Longreal       *TotalMiss)
{
  #undef  MF__
  #define MF__  MOD__"vPSE36MonitorDataCache"

  __os542PSE36GetMonitorStatistics(kgs.PSE36Ctrl->PSE36Stat.NumDataReads, 
                                   kgs.PSE36Ctrl->PSE36Stat.NumDataReadsFail,
                                   kgs.PSE36Ctrl->PSE36Stat.NumDataWrites,
                                   kgs.PSE36Ctrl->PSE36Stat.NumDataWritesFail,
                                   TotalAccess, TotalHits, TotalMiss ) ;
}

/*------------------------------*/

void  vPSE36MonitorReset()
{
  #undef  MF__
  #define MF__  MOD__"vPSE36MonitorObjLogCache"

  kgs.PSE36Ctrl->PSE36Stat.NumDataReads     = (tsp00_Longreal )0 ;
  kgs.PSE36Ctrl->PSE36Stat.NumDataReadsFail = (tsp00_Longreal )0 ;
  kgs.PSE36Ctrl->PSE36Stat.NumDataWrites    = (tsp00_Longreal )0 ;
  kgs.PSE36Ctrl->PSE36Stat.NumDataWritesFail= (tsp00_Longreal )0 ;
}


/*===========================================================================*
 *  LOCAL/PRIVATE/PROTECTED FUNCTIONS (CODE)                                 *
 *===========================================================================*/

_INLINE void  __os542PSE36GetMonitorStatistics ( tsp00_Longreal        TotalPSEReads ,
                                                 tsp00_Longreal        TotalPSEReadsFail,
                                                 tsp00_Longreal        TotalPSEWrites, 
                                                 tsp00_Longreal        TotalPSEWritesFail ,
                                                 tsp00_Longreal       *TotalAccess, 
                                                 tsp00_Longreal       *TotalHits,
                                                 tsp00_Longreal       *TotalMiss)
{
  #undef  MF__
  #define MF__  MOD__"__os542PSE36GetMonitorStatistics"

  *TotalAccess = TotalPSEReads     + TotalPSEWrites ;

  *TotalMiss   = TotalPSEReadsFail + TotalPSEWritesFail ;

  *TotalHits   = *TotalAccess - *TotalMiss ;
}


/*------------------------------*/

_INLINE HANDLE  __os542PSE36GetHandle ( HANDLE               *PSE36Handle,
                                        tsp_errtext           errtext,
                                        BOOLEAN              *ok )
{
  #undef  MF__
  #define MF__  MOD__"__os542PSE36GetHandle"

  DWORD                 rc ;
  
  *ok = TRUE; 

  if ( *PSE36Handle == INVALID_HANDLE_VALUE )
    if ( (rc = os541PSE36Open( NULL, PSE36Handle )) != NO_ERROR )
    {
      MSGALL((ERR_PSE36_OPEN_ERROR, rc ));
      *ok = FALSE;
      sql46c_build_error_string ( errtext, ERRMSG_PSE36_OPEN, rc );
    }

  return(*PSE36Handle);
}

/*------------------------------*/


static DWORD  os542PSE36GetSizeInPages ( int *SizeInPages )
{
  #undef  MF__
  #define MF__  MOD__"os542PSE36GetSizeInPages"

  LONGLONG    SizeInBytesAvailable ;
  DWORD       rc = NO_ERROR;
  tsp00_Pathc PSE36DeviceName ;

  DBGPAS;

  os541PSE36GetDeviceName(PSE36DeviceName, sizeof(PSE36DeviceName) ) ;
  MSGD (( INFO_PSE36_DEVICE_NAME, PSE36DeviceName )) ;

  rc = os541PSE36GetSize(PSE36DeviceName, &SizeInBytesAvailable );

  if ( rc == NO_ERROR && SizeInBytesAvailable > 0)
  {
    *SizeInPages = (int )(SizeInBytesAvailable / kgs.PSE36Ctrl->DBPageSize) ;
    MSGD (( INFO_PSE36_PAGES_AVAILABLE, *SizeInPages));
  }
  else
  {
    char SysErrorText [40] ;
    eo46_system_error_text( rc, SysErrorText, sizeof(SysErrorText) ) ;
    MSGD (( INFO_PSE36_NOT_SUPPORTED, SysErrorText ));
    *SizeInPages = -1 ;
  }

  return(rc);
}

/*------------------------------*/

static DWORD  os542PSE36AllocCtrlStruc ( void **PSE36CtrlStruct,
                                         int    Size,
                                         char  *CtrlStructId )
{                                        
  #undef  MF__
  #define MF__  MOD__"os542PSE36AllocCtrlStruc"

  DWORD    rc = NO_ERROR;

  DBGPAS;

  DBG3 ((MF__, "try to allocate %d Bytes for %s", Size, CtrlStructId ));

  rc = ALLOC_MEM ((PPVOID)PSE36CtrlStruct,
                   Size + kgs.ulSystemPageSize );

  if ( rc == NO_ERROR )
  {
    memset(*PSE36CtrlStruct, 0, Size );
  }
  else
  {
    MSGD ((ERR_PSE36_CANT_ALLOC_MEMORY, Size, CtrlStructId, rc ));
  }

  return(rc);
}

/*------------------------------*/

static void  os542CacheDataPageToPSE36 ( tsp_page_no           PageNo,
                                         PTASK_CTRL_REC        pTaskCtrl,
                                         BUFFERADDRESS         bufadr,
                                         tsp_errtext           ErrText,
                                         BOOLEAN              *ok )
{
  #undef  MF__
  #define MF__  MOD__"os542CachePageToPSE36"

  __os542PSE36GetHandle (&pTaskCtrl->pUKT->PSE36_Handle, ErrText, ok ) ;
  if ( *ok )
  {
    DWORD rc ;
    long  lLength;
    LONGLONG PSE36Offset  = (LONGLONG)( PageNo ) * kgs.PSE36Ctrl->DBPageSize ;

    pTaskCtrl->TaskState = TSK_PSE36_DATA_WRITE_EO00;

    rc = os541PSE36Write( pTaskCtrl->pUKT->PSE36_Handle, bufadr, kgs.PSE36Ctrl->DBPageSize, 
                          PSE36Offset, &lLength );

    pTaskCtrl->TaskState = TSK_RUNNING_EO00;

    if ( rc == NO_ERROR && kgs.PSE36Ctrl->DBPageSize == lLength)
      __os542SetPageCached(PageNo) ;
    else
	{
      MSGALL((ERR_PSE36_IO_ERROR, ERRMSG_PSE36_WRITE, rc, PageNo, PSE36Offset ));
      sql46c_build_error_string ( ErrText, ERRMSG_PSE36_WRITE, rc );
      *ok = FALSE ;
	}
  }
}


/*------------------------------*/

static void  os542GetCachedPageFromPSE36 ( tsp_page_no           PageNo,
                                           PTASK_CTRL_REC        pTaskCtrl,
                                           BUFFERADDRESS         bufadr,
                                           tsp_errtext           ErrText,
                                           BOOLEAN              *ok )
{
  #undef  MF__
  #define MF__  MOD__"os542GetCachedPageFromPSE36"

  __os542PSE36GetHandle (&pTaskCtrl->pUKT->PSE36_Handle, ErrText, ok ) ;
  if ( *ok ) 
  {
    DWORD    rc ;
    long  lLength;
    LONGLONG PSE36Offset  = (LONGLONG)( PageNo ) * kgs.PSE36Ctrl->DBPageSize ;

    pTaskCtrl->TaskState = TSK_PSE36_DATA_READ_EO00;

    rc = os541PSE36Read( pTaskCtrl->pUKT->PSE36_Handle, bufadr, kgs.PSE36Ctrl->DBPageSize, 
                         PSE36Offset, &lLength );

    pTaskCtrl->TaskState = TSK_RUNNING_EO00;

    DBG3 ((MF__, "page %d, rc = %d", PageNo, rc ));

    if ( rc != NO_ERROR || lLength != kgs.PSE36Ctrl->DBPageSize )
	{
      MSGALL((ERR_PSE36_IO_ERROR, ERRMSG_PSE36_READ, rc, PageNo, PSE36Offset ));
      sql46c_build_error_string ( ErrText, ERRMSG_PSE36_READ, rc );
      *ok = FALSE ;
	}
  }
}

/*------------------------------*/

static void  os542HandleDataRead ( tsp_int4              devno,
                                   tsp_page_no           offset,
                                   tsp_page_no           PageNo,
                                   PTASK_CTRL_REC        pTaskCtrl,
                                   BUFFERADDRESS         bufadr,
                                   tsp_int2              sNumOfPages,
                                   PROCESS_ID            pid,
                                   tsp_errtext           ErrText,
                                   BOOLEAN              *ok )
{
  #undef  MF__
  #define MF__  MOD__"os542HandleDataRead"

  bool     PageInPSE36Range ;

  DBGPAS;

  kgs.PSE36Ctrl->PSE36Stat.NumDataReads++ ;

  PageInPSE36Range = PSE36_IS_DATAPAGE_IN_RANGE(PageNo);

  if ( !PageInPSE36Range || !__os542IsPageCached(PageNo) )
  {
    /* read page from devspace */
    vblockio (devno, offset, sp2ioc_read, bufadr, sNumOfPages, 
              pid, ErrText, ok ) ;

    if ( *ok && PageInPSE36Range )
    {
      /* save page in PSE36 */
      __os542CacheDataPage (PageNo, pTaskCtrl, bufadr, ErrText, ok ) ;
    }
    else
    {
      kgs.PSE36Ctrl->PSE36Stat.NumDataReadsFail++ ;
      DBG3 ((MF__, "PageNo = %d out of range ( 0..%d)", PageNo,  kgs.PSE36Ctrl->PSE36Stat.MaxDataPages));
      MSGD((PSEDBG, "PageNo = %d out of range ( 0..%d)", PageNo,  kgs.PSE36Ctrl->PSE36Stat.MaxDataPages));
    }
  }
  else
  {
    /* page is in range AND in PSE36 */
    __os542GetCachedPage( PageNo, pTaskCtrl, bufadr, ErrText, ok );
  }
}

/*------------------------------*/

static void  os542HandleDataWrite ( tsp_int4              devno,
                                    tsp_page_no           offset,
                                    tsp_page_no           PageNo,
                                    PTASK_CTRL_REC        pTaskCtrl,
                                    BUFFERADDRESS         bufadr,
                                    tsp_int2              sNumOfPages,
                                    PROCESS_ID            pid,
                                    tsp_errtext           ErrText,
                                    BOOLEAN              *ok )
{
  #undef  MF__
  #define MF__  MOD__"os542HandleDataWrite"

  bool     ForcePhysicalWrite = kgs.PSE36Ctrl->WriteMode != WRITE_TO_PSE36_ONLY;

  int  i ;
  BUFFERADDRESS ActPage ;

  pTaskCtrl->TaskState = TSK_PSE36_DATA_WRITE_EO00;
  
  WAIT_UNTIL_ALONE( &kgs.PSE36Ctrl->Exclusive );
  
  for ( *ok = TRUE, i=0, ActPage = bufadr; 
        i<sNumOfPages && *ok ;
        i++, ActPage++ )
  {
    void b15get_data_page_info();
    BOOLEAN      IsCheckpointRelevant ;
    tsp00_PageNo ActualPageNo;

    b15get_data_page_info (ActPage, &ActualPageNo, &IsCheckpointRelevant ) ;

    DBG3 ((MF__, "%d. ActualPageNo = %d, IsCheckpointRelevant = %d", 
                  i, ActualPageNo, IsCheckpointRelevant ));

    kgs.PSE36Ctrl->PSE36Stat.NumDataWrites++ ;
    if ( PSE36_IS_DATAPAGE_IN_RANGE(ActualPageNo) )
    {
      /* save page in PSE36 */
      __os542CacheDataPage (ActualPageNo, pTaskCtrl, ActPage, ErrText, ok ) ;

      if ( IsCheckpointRelevant )
      {
        if ( !ForcePhysicalWrite )
        {
          if ( !__os542IsPageChanged(ActualPageNo) )
            __os542SetPageChanged(ActualPageNo) ;
        }
        else
        {
          if ( __os542IsPageChanged(ActualPageNo) )
            __os542ResetPageChanged(ActualPageNo) ;
        }
      }
    }
    else
    {
      kgs.PSE36Ctrl->PSE36Stat.NumDataWritesFail++ ;
      ForcePhysicalWrite = true ;
    }
  }
      
  CLEARLOCK( &kgs.PSE36Ctrl->Exclusive );
  pTaskCtrl->TaskState = TSK_RUNNING_EO00;

  if ( ForcePhysicalWrite )
  {
    vblockio (devno, offset, sp2ioc_write, bufadr, sNumOfPages, 
              pid, ErrText, ok ) ;
  }
}

/*------------------------------*/

_INLINE tsp_page_no __os542GetNextChangedPage()
{
  #undef  MF__
  #define MF__  MOD__"__os542GetNextChangedPage"

  tsp_page_no ActPage = kgs.PSE36Ctrl->LowestChangedDataPage ;

  DBG3 ((MF__, "start searching at page %d", ActPage ));

  while ( PSE36_IS_DATAPAGE_IN_RANGE(ActPage) &&
          ! __os542IsPageChanged(ActPage) )
    ActPage++ ;

  if ( !PSE36_IS_DATAPAGE_IN_RANGE(ActPage) )
  {
    ActPage = -1 ;
  }
  else
  {
    kgs.PSE36Ctrl->LowestChangedDataPage = ActPage+1 ;
  }

  DBG3 ((MF__, "returning page %d, LowestChangedDataPage = %d", 
                ActPage,  kgs.PSE36Ctrl->LowestChangedDataPage ));

  return (ActPage);
}

/*------------------------------*/

_INLINE void __os542SetPageCached( tsp_page_no _PageNo )
{
  #undef  MF__
  #define MF__  MOD__"__os542SetPageCached"

  DBG3 ((MF__, "set cached flag for page = %d", _PageNo ));

  kgs.PSE36Ctrl->DataPageArray[_PageNo] |= 1;
}

/*------------------------------*/

_INLINE void __os542ResetPageCached( tsp_page_no _PageNo )
{
  #undef  MF__
  #define MF__  MOD__"__os542ResetPageCached"

  DBG3 ((MF__, "reset flag for page = %d", _PageNo ));

  kgs.PSE36Ctrl->DataPageArray[_PageNo] &= 0xFE ;
}

/*------------------------------*/

_INLINE bool __os542IsPageCached( tsp_page_no _PageNo )
{
  #undef  MF__
  #define MF__  MOD__"__os542IsPageCached"

  bool IsCached = kgs.PSE36Ctrl->DataPageArray[_PageNo] & 1 ;

  DBG3 ((MF__, "page %d is %s cached", _PageNo, IsCached ? "" : "not " ));

  return(IsCached) ;
}

/*------------------------------*/

_INLINE void __os542SetPageChanged( tsp_page_no _PageNo )
{
  #undef  MF__
  #define MF__  MOD__"__os542SetPageChanged"

  kgs.PSE36Ctrl->DataPageArray[_PageNo] |= 2;

  if ( kgs.PSE36Ctrl->LowestChangedDataPage > _PageNo )
    kgs.PSE36Ctrl->LowestChangedDataPage = _PageNo ;

  kgs.PSE36Ctrl->PSE36Stat.NumChangedDataPages++ ;

  DBG3 ((MF__, "set changed flag for page = %d, num changed = %d", 
                _PageNo, kgs.PSE36Ctrl->PSE36Stat.NumChangedDataPages ));
}

/*------------------------------*/

_INLINE void __os542ResetPageChanged( tsp_page_no _PageNo )
{
  #undef  MF__
  #define MF__  MOD__"__os542ResetPageChanged"

  kgs.PSE36Ctrl->DataPageArray[_PageNo] &= 0xFD ;

  --kgs.PSE36Ctrl->PSE36Stat.NumChangedDataPages ;

  DBG3 ((MF__, "reset changed flag for page = %d, num now changed = %d", 
                _PageNo, kgs.PSE36Ctrl->PSE36Stat.NumChangedDataPages ));
}

/*------------------------------*/

_INLINE bool __os542IsPageChanged( tsp_page_no _PageNo )
{
  #undef  MF__
  #define MF__  MOD__"__os542IsPageChanged"

  bool IsChanged = kgs.PSE36Ctrl->DataPageArray[_PageNo] & 0x02 ;

  DBG3 ((MF__, "page %d is %s changed", _PageNo, IsChanged ? "" : "not " ));

  return(IsChanged) ;
}

/*------------------------------*/

_INLINE APIRET os542AWE_AdjustTokenPrivilegeForMemoryLocking()
  {
	APIRET rc = NO_ERROR;
	TOKEN_PRIVILEGES tpPrevious;
	TOKEN_PRIVILEGES tp;
	DWORD  cbPrevious = sizeof(TOKEN_PRIVILEGES);
	LUID   luid;
	HANDLE hTokenUsed;

	// if no token specified open process token
    if (!OpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hTokenUsed))
	{
		rc = GetLastError();
	}

	if (!LookupPrivilegeValue(NULL, SE_LOCK_MEMORY_NAME, &luid ))
	{
		rc = GetLastError();
		CloseHandle(hTokenUsed);
	}

	tp.PrivilegeCount = 1;
	tp.Privileges[0].Luid = luid;
	tp.Privileges[0].Attributes = 0;

	if (!AdjustTokenPrivileges(hTokenUsed, FALSE, &tp, sizeof(TOKEN_PRIVILEGES), &tpPrevious, &cbPrevious))
	{
		rc = GetLastError();
		CloseHandle(hTokenUsed);
	}

	tpPrevious.PrivilegeCount = 1;
	tpPrevious.Privileges[0].Luid = luid;
	tpPrevious.Privileges[0].Attributes |= (SE_PRIVILEGE_ENABLED);

	if (!AdjustTokenPrivileges(hTokenUsed, FALSE, &tpPrevious, cbPrevious, NULL, NULL))
	{
		rc = GetLastError();
	}
	CloseHandle(hTokenUsed);
	
	return(rc);
  }

/*------------------------------*/
_INLINE bool __os542_AWE_MapDBPage   ( tsp_page_no           PageNo,
                                       PTASK_CTRL_REC        pTaskCtrl,
                                       PPVOID                MappingAddress,
                                       tsp_errtext           ErrText,
                                       BOOLEAN              *ok )
{
  #undef  MF__
  #define MF__  MOD__"__os542_AWE_MapPage"

  APIRET rc ;

  *MappingAddress   = kgs.PSE36Ctrl->AWEMappingAdresses [pTaskCtrl->pUKT->ulUKTIndex-1] ;

  rc = os543AWE_Map( *MappingAddress, kgs.PSE36Ctrl->PhysPagesPerDBPage, 
	                  kgs.PSE36Ctrl->PhysPagesFrameNumbers, PageNo * kgs.PSE36Ctrl->PhysPagesPerDBPage ) ;

  if ( rc != NO_ERROR ) 
  {
	// try again, perhaps another thread has mapped the specified address (page) inj this moment 
    int        Retries = 0 ;
	do
	{
      eo46_system_error_text( rc, ErrText, sizeof(ErrText) ) ;
      MSGALL((ERR_AWE_MAP_PHYSICAL_MEMORY, PageNo, ErrText ));
      SLEEP(1000); 
      rc = os543AWE_Map( *MappingAddress, kgs.PSE36Ctrl->PhysPagesPerDBPage, 
	                     kgs.PSE36Ctrl->PhysPagesFrameNumbers, PageNo * kgs.PSE36Ctrl->PhysPagesPerDBPage ) ;
    }
	while ( rc != NO_ERROR && ++Retries < 3 ) ;
  }

  *ok = rc == NO_ERROR ;
  return(*ok);
}

#define IS_AWE_ACTIVE (kgs.PSE36Ctrl->PhysPagesFrameNumbers != NULL ) 

/*------------------------------*/

_INLINE void __os542GetCachedPage ( tsp_page_no           PageNo,
                                    PTASK_CTRL_REC        pTaskCtrl,
                                    BUFFERADDRESS         bufadr,
                                    tsp_errtext           ErrText,
                                    BOOLEAN              *ok )
{
  if( IS_AWE_ACTIVE )
    os542GetCachedPageFromAWE( PageNo, pTaskCtrl, bufadr, ErrText, ok );
  else
    os542GetCachedPageFromPSE36( PageNo, pTaskCtrl, bufadr, ErrText, ok );
}

/*------------------------------*/

_INLINE void  __os542CacheDataPage ( tsp_page_no           PageNo,
                                     PTASK_CTRL_REC        pTaskCtrl,
                                     BUFFERADDRESS         bufadr,
                                     tsp_errtext           ErrText,
                                     BOOLEAN              *ok )
{
  if( IS_AWE_ACTIVE )
    os542CacheDataPageToAWE (PageNo, pTaskCtrl, bufadr, ErrText, ok ) ;
  else
    os542CacheDataPageToPSE36 (PageNo, pTaskCtrl, bufadr, ErrText, ok ) ;
}

/*------------------------------*/

static void os542FreeAWEMemoryRegion ( PVOID        MappingAddress,
					   	      	       tsp_errtext  ErrText,
 									   APIRET      *rc)

{
  #undef  MF__
  #define MF__  MOD__"os542FreeAWEMemoryRegion"

  if ( ( *rc = os543AWE_FreeMappingRegion(&MappingAddress) ) != NO_ERROR )
  {
    eo46_system_error_text( *rc, ErrText, sizeof(ErrText) ) ;
    MSGALL(( ERR_AWE_FREE_MEMORY_REGION, ErrText ));
    sql46c_build_error_string ( ErrText, "os543AWE_FreeMappingRegion failed", *rc );
  }
}

/*------------------------------*/

static void  os542GetCachedPageFromAWE ( tsp_page_no           PageNo,
                                         PTASK_CTRL_REC        pTaskCtrl,
                                         BUFFERADDRESS         bufadr,
                                         tsp_errtext           ErrText,
                                         BOOLEAN              *ok )
{
  #undef  MF__
  #define MF__  MOD__"os542GetCachedPageFromAWE"

  PVOID MappingAddress   = kgs.PSE36Ctrl->AWEMappingAdresses [pTaskCtrl->pUKT->ulUKTIndex-1] ;

  if ( __os542_AWE_MapDBPage ( PageNo, pTaskCtrl, &MappingAddress, ErrText, ok ) )
  {
    memcpy( bufadr, MappingAddress, kgs.PSE36Ctrl->DBPageSize ) ;
	os543AWE_UnMap( MappingAddress, kgs.PSE36Ctrl->PhysPagesPerDBPage ) ;
  }
}

/*------------------------------*/

static void  os542CacheDataPageToAWE   ( tsp_page_no           PageNo,
                                         PTASK_CTRL_REC        pTaskCtrl,
                                         BUFFERADDRESS         bufadr,
                                         tsp_errtext           ErrText,
                                         BOOLEAN              *ok )
{
  #undef  MF__
  #define MF__  MOD__"os542CacheDataPageToAWE"


  PVOID MappingAddress ;

  if ( __os542_AWE_MapDBPage ( PageNo, pTaskCtrl, &MappingAddress, ErrText, ok ) )
  {
    memcpy( MappingAddress, bufadr, kgs.PSE36Ctrl->DBPageSize ) ;
	os543AWE_UnMap( MappingAddress, kgs.PSE36Ctrl->PhysPagesPerDBPage ) ;
    __os542SetPageCached(PageNo) ;
  }
}

/*------------------------------*/

static PVOID  os542AllocAWEMemoryRegion ( int          MappingSizeInBytes,
                                          tsp_errtext  ErrText,
                                          APIRET      *rc)
{
  #undef  MF__
  #define MF__  MOD__"os542AllocAWEMemoryRegion"

  void *MappingAddress ;

  *rc = os543AWE_AllocatMappingRegion ( &MappingAddress, MappingSizeInBytes ) ;

  if ( *rc != NO_ERROR ) 
  {
    eo46_system_error_text( *rc, ErrText, sizeof(ErrText) ) ;
    MSGALL(( ERR_AWE_ALLOC_MEMORY_REGION, ErrText ));
    sql46c_build_error_string ( ErrText, "AWE READ virtual alloc failed", *rc );
  }
  return ( MappingAddress ) ;
}

/*------------------------------*/

static DWORD  os542AWE_Init (int *SizeInPages)
{
  #undef  MF__
  #define MF__  MOD__"os542AWE_Init"

  tsp_errtext  ErrText ;
  APIRET rc ;
  ULONG  ul ;
  ULONG  ulMajorVersion;
  ULONG  ulMinorVersion;
  ULONG  NumberOfPhysPages = 0;
  ULONG  PageLimit;

  sql02_get_os_version ( &ulMajorVersion, &ulMinorVersion );

  if ( ulMajorVersion >= 5 ) // >= Windows 2000
  {
    rc = os543AWE_LoadFunctions();

    if ( rc == NO_ERROR ) 
      rc = os542AWE_AdjustTokenPrivilegeForMemoryLocking();
  }
  else
    rc = !NO_ERROR;

  kgs.PSE36Ctrl->PhysPagesPerDBPage = kgs.PSE36Ctrl->DBPageSize / kgs.ulSystemPageSize;


  if ( rc == NO_ERROR )
  {
    if ( XPARAM(lMemEnhancementLimit) == 0 ) // no limit requested
      PageLimit = (ULONG)-1; // max ULONG
    else
      PageLimit = (XPARAM(lMemEnhancementLimit) * 1024*1024) / kgs.ulSystemPageSize;

    NumberOfPhysPages = min( PageLimit, (ULONG)kgs.PSE36Ctrl->PhysPagesPerDBPage * XPARAM(lMaxDataPages) );

    DBG3 ((MF__, "NumberOfPhysPagesWanted : %d to handle %d DB-Pages", 
                  NumberOfPhysPagesWanted, NumberOfPhysPages ));

    rc = os542PSE36AllocCtrlStruc ( (PPVOID)&kgs.PSE36Ctrl->AWEMappingAdresses, 
                                    kgs.ulNumOfUKTs * sizeof(PVOID),
     	                              "mapping adresses array");

    for ( ul = 0 ; ul < kgs.ulNumOfUKTs && rc == NO_ERROR; ul++ ) 
      kgs.PSE36Ctrl->AWEMappingAdresses[ul] = os542AllocAWEMemoryRegion( kgs.PSE36Ctrl->DBPageSize, ErrText, &rc ) ;
		
    if ( rc == NO_ERROR )
    {
      rc = os543AWE_AllocatPhysicalPages( &NumberOfPhysPages, &kgs.PSE36Ctrl->PhysPagesFrameNumbers) ;

      if ( rc != NO_ERROR )
      {
        *SizeInPages = 0 ;
        eo46_system_error_text( rc, ErrText, sizeof(ErrText) ) ;
        MSGD (( ERR_AWE_ALLOC_PHYSICAL_MEMORY, ErrText ));
      }
      else
      {
        *SizeInPages = NumberOfPhysPages / kgs.PSE36Ctrl->PhysPagesPerDBPage;
        MSGD (( INFO_AWE_PAGES_AVAILABLE, 
                (NumberOfPhysPages * kgs.ulSystemPageSize) / (1024*1024),
                *SizeInPages));
      }
      DBG3 ((MF__, "NumberOfPhysPages : %d to handle %d DB-Pages", 
             NumberOfPhysPages, XPARAM(lMaxDataPages) ));
    }
  }

  if ( rc != NO_ERROR )
  {
    eo46_system_error_text( rc, ErrText, sizeof(ErrText) ) ;
    MSGD (( INFO_AWE_NOT_SUPPORTED, ErrText ));
    
    // cleanup
    if ( kgs.PSE36Ctrl->PhysPagesFrameNumbers && NumberOfPhysPages > 0 )
      os543AWE_FreePhysicalPages ( &NumberOfPhysPages , &kgs.PSE36Ctrl->PhysPagesFrameNumbers );
    
    if ( kgs.PSE36Ctrl->AWEMappingAdresses )
    {
      int DummyRC ; 

      for ( ul = 0 ; ul < kgs.ulNumOfUKTs && kgs.PSE36Ctrl->AWEMappingAdresses[ul] != NULL; ul++ ) 
        os542FreeAWEMemoryRegion( kgs.PSE36Ctrl->AWEMappingAdresses[ul], ErrText, &DummyRC ) ;
      
      FREE_MEM ( (PVOID)kgs.PSE36Ctrl->AWEMappingAdresses );
    }
  }
  return(rc);
}

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/


