/****************************************************************************

  module      : vos543.c

  -------------------------------------------------------------------------

  responsible : FrankH

  special area: Intel extended memory support 
  description : "dynamic loader for Windows 2000 AWE Support"


  last changed: 1999-08-20  16:59
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1999



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gos00.h"
#include "veo00c.h"
#include "heo46.h"
#include "hsp100.h"
#include "geo007_1.h"
#include "gos543.h"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define MOD__  "VOS543L.C : "
#define MF__   MOD__"UNDEFINED"

#define DEFAULT_AWE_DLL                _T("SAPDB_AWE.dll")

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

#define MK_PA(f)       (*lp##f)

#define LOAD_FUNC(f)   ((lp##f = os543AWE_LoadFunc(#f)) == NULL) ? \
                         -1 : NO_ERROR

#define CALL_FUNC(f,p) ( lp##f == NULL ) ? os543AWE_LoadErr(#f),-1 : lp##f##p

/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/

/*===========================================================================*
 *  GLOBAL VARIABLES                                                         *
 *===========================================================================*/


/*===========================================================================*
 *  LOCAL VARIABLES                                                          *
 *===========================================================================*/

static HINSTANCE    hinstAWE_DLL = NULL;
static PSZ          pszDLL;

static int  MK_PA (os543AWE_GetDllVersion)        ( tsp100_VersionID1 *pVersion ) ;

static int  MK_PA (os543AWE_AllocatPhysicalPages) ( ULONG  *NumberOfPhysPagesWanted_Available,
                                                    void  **AWE_SpaceForPhysPagesFrameNumbers ) ;

static int  MK_PA (os543AWE_FreePhysicalPages)    ( ULONG  *NumberOfPhysPagesWanted_Available,
                                                    void  **AWE_SpaceForPhysPagesFrameNumbers ) ;

static int  MK_PA (os543AWE_AllocatMappingRegion) ( void **MappingAddress,
                                                    int    MappingSizeInBytes ) ;

static int  MK_PA (os543AWE_FreeMappingRegion)    ( void **MappingAddress ) ;

static int  MK_PA (os543AWE_Map)                  ( void  *MappingAddress,
                                                    int    NumberOfPhysPages,
                                 		            void  *AWEPhysPagesFrameStart,
                            			            int    AWEOffset) ;

static int  MK_PA (os543AWE_UnMap)                ( void  *MappingAddress,
                                                    int    NumberOfPhysPages ) ;
/*===========================================================================*
 *  STATIC/INLINE FUNCTIONS (PROTOTYPES)                                     *
 *===========================================================================*/

static VOID* os543AWE_LoadFunc(  PSZ  pszFuncName ) ;

static VOID os543AWE_LoadErr( PSZ  pszFuncName ) ;

static int os543AWE_LoadLibrary() ;

/*===========================================================================*
 *  PUBLIC FUNCTIONS (CODE)                                                  *
 *===========================================================================*/

int os543AWE_LoadFunctions ()
  {
  #undef  MF__
  #define MF__ MOD__"os543AWE_LoadFunctions"
  int              rc                  = NO_ERROR;
  static BOOL      fAllFunctionsLoaded = FALSE;

  DBGIN;

  if ( !fAllFunctionsLoaded )
  {
    tsp100_VersionID1 Version ;

	rc = os543AWE_LoadLibrary();
    if ( rc != NO_ERROR ) return ( rc );

    rc = LOAD_FUNC( os543AWE_GetDllVersion );
    if ( rc != NO_ERROR ) return ( rc );

    rc = LOAD_FUNC( os543AWE_AllocatPhysicalPages );
    if ( rc != NO_ERROR ) return ( rc );

    rc = LOAD_FUNC( os543AWE_FreePhysicalPages );
    if ( rc != NO_ERROR ) return ( rc );

    rc = LOAD_FUNC( os543AWE_AllocatMappingRegion );
    if ( rc != NO_ERROR ) return ( rc );

    rc = LOAD_FUNC( os543AWE_FreeMappingRegion );
    if ( rc != NO_ERROR ) return ( rc );

    rc = LOAD_FUNC( os543AWE_Map );
    if ( rc != NO_ERROR ) return ( rc );

    rc = LOAD_FUNC( os543AWE_UnMap );
    if ( rc != NO_ERROR ) return ( rc );

	if ( os543AWE_GetDllVersion ( &Version ) == NO_ERROR )
	{
      MSGD (( INFO_AWE_DLL_VERSION, pszDLL, Version.MajorVersion_sp100 ,
		                                    Version.MinorVersion_sp100 ,
											Version.CorrLevel_sp100, 
											Version.BuildNumberPrefix_sp100)) ;
	}

    fAllFunctionsLoaded = TRUE;
  }

  DBGOUT;
  return ( rc );
  }

/*------------------------------*/

int os543AWE_GetDllVersion ( tsp100_VersionID1 *pVersion )
  {
  #undef  MF__
  #define MF__ MOD__"os543AWE_GetDllVersion"

  DBGPAS;

  return ( CALL_FUNC( os543AWE_GetDllVersion, (pVersion) ) );
  }

/*------------------------------*/

int os543AWE_AllocatPhysicalPages ( ULONG  *NumberOfPhysPagesWanted_Available,
                                    void  **AWE_SpaceForPhysPagesFrameNumbers )
  {
  #undef  MF__
  #define MF__ MOD__"os543AWE_AllocatPhysicalPages"
 
  DBGPAS;
  return ( CALL_FUNC( os543AWE_AllocatPhysicalPages, (NumberOfPhysPagesWanted_Available, AWE_SpaceForPhysPagesFrameNumbers) ) );
  }

/*------------------------------*/

int os543AWE_FreePhysicalPages ( ULONG  *NumberOfPhysPagesWanted_Available,
                                 void  **AWE_SpaceForPhysPagesFrameNumbers )
  {
  #undef  MF__
  #define MF__ MOD__"os543AWE_FreePhysicalPages"
 
  DBGPAS;
  return ( CALL_FUNC( os543AWE_FreePhysicalPages, (NumberOfPhysPagesWanted_Available, AWE_SpaceForPhysPagesFrameNumbers) ) );
  }

/*------------------------------*/

int os543AWE_AllocatMappingRegion ( void **MappingAddress,
                                    int    MappingSizeInBytes ) 
  {
  #undef  MF__
  #define MF__ MOD__"os543AWE_AllocatMappingRegion"

  DBGPAS;

  return ( (CALL_FUNC( os543AWE_AllocatMappingRegion, (MappingAddress, MappingSizeInBytes) )) );
  }

/*------------------------------*/

int os543AWE_FreeMappingRegion ( void **MappingAddress ) 
  {
  #undef  MF__
  #define MF__ MOD__"os543AWE_FreeMappingRegion"

  DBGPAS;

  return ( (CALL_FUNC( os543AWE_FreeMappingRegion, (MappingAddress) )) );
  }

/*------------------------------*/

int os543AWE_Map                  ( void  *MappingAddress,
                                    int    NumberOfPhysPages,
                           		    void  *AWEPhysPagesFrameStart,
                       			    int    AWEOffset) 
  {
  #undef  MF__
  #define MF__ MOD__"os543AWE_Map"

  DBGPAS;

  return ( CALL_FUNC( os543AWE_Map, (MappingAddress, NumberOfPhysPages, AWEPhysPagesFrameStart,AWEOffset ) ) );
  }

/*------------------------------*/

int os543AWE_UnMap                ( void  *MappingAddress,
                                    int    NumberOfPhysPages )
  {
  #undef  MF__
  #define MF__ MOD__"os543AWE_UnMap"

  DBGPAS;

  return ( CALL_FUNC( os543AWE_UnMap, (MappingAddress, NumberOfPhysPages) ) );
  }

/*===========================================================================*
 *  LOCAL/PRIVATE/PROTECTED FUNCTIONS (CODE)                                 *
 *===========================================================================*/

/*------------------------------*/

static int os543AWE_LoadLibrary()
{

  LONG             rc    = NO_ERROR;

  if ( hinstAWE_DLL == NULL )
  {
    HINSTANCE        hinst = NULL;
    tsp00_Pathc      szDLLPath;
    CHAR             *pszTmp;
    DWORD            dwLen;

    pszDLL = DEFAULT_AWE_DLL;
    if (strchr(pszDLL, '\\' ) == NULL)
      {
      // if there is no absolut path, check the directory of the calling program
      dwLen = GetModuleFileName(NULL, szDLLPath, sizeof(szDLLPath) - 1);

      if (dwLen != 0) 
        {
        pszTmp = strrchr(szDLLPath, '\\' );
        
        if ( pszTmp != NULL )
          {
          pszTmp[1] = 0;
          strcat( szDLLPath,pszDLL );
          hinst = LoadLibrary( szDLLPath );
          }
	    }
      }

    if ( hinst == NULL )
      hinst = LoadLibrary( pszDLL );

    if ( hinst == NULL )
    {
      rc = GetLastError();

      MSGALL (( WRN_LOADING_AWE_DLL, pszDLL, rc ));

      DBGOUT;
    }
    else
      hinstAWE_DLL = hinst;
    }
  return ( rc );
}

/*------------------------------*/

static VOID* os543AWE_LoadFunc(  PSZ  pszFuncName )
  {
  #undef  MF__
  #define MF__ MOD__"os543AWE_LoadFunc"
  FARPROC          pFunc ;

  DBGIN;

  if ( hinstAWE_DLL == NULL )
    pFunc = NULL ;
  else  
    pFunc = GetProcAddress( hinstAWE_DLL, pszFuncName);

  if ( pFunc == NULL )
    {
    sql60_enable_message_box ();

    MSGALL (( ERR_LOADING_FUNCTION, pszFuncName, pszDLL ));

    sql60_disable_message_box ();
    }

  DBGOUT;
  return ( pFunc );
  }

/*------------------------------*/

static VOID os543AWE_LoadErr( PSZ  pszFuncName )
  {
  #undef  MF__
  #define MF__ MOD__"os543AWE_LoadErr"

  sql60_enable_message_box();

  MSGALL (( ERR_FUNCTION_NOT_LOADED, pszFuncName ));

  sql60_disable_message_box();

  return;
  }


/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/
