/*!**************************************************************************

  module      : vos57.c

  -------------------------------------------------------------------------

  responsible : RaymondR
  
  special area: Kernel
  description : kernel storage management

  last changed: 2002-12-18  11:39
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2002

  ========== licence begin LGPL
  Copyright (C) 2002 SAP AG

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  ========== licence end

*****************************************************************************/

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gos00.h"
#include "veo00c.h"
#include "geo002.h"
#include "geo007_1.h"
#include "geo50_0.h"
#include "heo57.h"


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#ifdef NO_INLINES
# define _INLINE
#else
# define _INLINE	__inline
#endif

#define MOD__  "VOS57KC : "
#define MF__   MOD__"UNDEFINED"


#define MEM_FRAG_AND_OVERHEAD      300
#define RTE_TMP_ALLOC_POOL_SIZE    ( 1 MB )

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

#define MK_PA(f)                (*lp##f)

#define LOAD_FUNC(f)            ((lp##f = sql57c_load_func(#f)) == NULL) ? \
                                  -1 : NO_ERROR
#define CALL_FUNC(fdll, forg,p) (( lp##fdll == NULL ) ? ##forg##p : lp##fdll##p)



/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/

typedef HANDLE   POOLTYP;

typedef struct mem_info_record
  {
  ULONG    ulSize;
  POOLTYP  Pool;
  } MEM_INFO_REC;

typedef MEM_INFO_REC  *PMEM_INFO_REC;


/*===========================================================================*
 *  GLOBAL VARIABLES                                                         *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL VARIABLES                                                          *
 *===========================================================================*/

static LONG          lExclusive         = 0;
static POOLTYP       KernelMemPool      = NULL;

static HINSTANCE    hinstDLL      = NULL;

static HANDLE MK_PA (HeapCreateP)  ( DWORD  flOptions,
                                     DWORD  dwInitialSize,
                                     DWORD  dwMaximumSize );
static LPVOID MK_PA (HeapAllocP)   ( HANDLE hHeap,
                                     DWORD  dwFlags,
                                     DWORD  dwBytes );
static BOOL   MK_PA (HeapFreeP)    ( HANDLE hHeap,
                                     DWORD  dwFlags,
                                     LPVOID lpMem );
static BOOL   MK_PA (HeapDestroyP) ( HANDLE hHeap );



/*===========================================================================*
 *  LOCAL FUNCTIONS (PROTOTYPES)                                             *
 *===========================================================================*/

static ULONG sql57c_load_functions (  VOID );
static VOID* sql57c_load_func      (  PSZ  pszFuncName );

static LONG  sql57k_pool_malloc  ( POOLTYP *pPool,
                                   ULONG   ulPoolSize,
                                   PSZ     pszPoolName,
                                   PULONG  pulPoolUsed,
                                   PULONG  pulNumOfAllocs,
                                   PPVOID  ppMemory,
                                   ULONG   ulSize );

static LONG  sql57k_pool_free    ( PULONG  pulPoolUsed,
                                   PULONG  pulNumOfAllocs,
                                   PVOID   pMemory );


/*===========================================================================*
 *  GLOBAL FUNCTIONS (CODE)                                                  *
 *===========================================================================*/

VOID  valloc_convcache ( tsp00_Int4      count,
                         tsp00_Int4      *available,
                         tsp00_PageAddr  *bufptr,
                         bool            *ok )

  {
  #undef  MF__
  #define MF__ MOD__"valloc_convcache"
  #ifdef DEBUG
   PUKT_CTRL_REC   pUKT = THIS_UKT_CTRL;
  #endif

  DBGPAS;

  WAIT_UNTIL_ALONE(&kgs.lSharedDynDataExcl);

  *available = (tsp00_Int4)BYTES_TO_PAGES(kgs.pSharedConvCacheTop - kgs.pSharedConvCachePtr);

  DBG3 (( MF__, "T%d - count: %d - available: %d",
                pUKT->pCTask->ulTaskIndex, count, *available ));

  // --- counts in units of BUFFER
  if ( count > *available )
    {
    *ok = FALSE ;
    MSGALL(( WRN_SHRD_DYN_DATA_TOO_SMALL ));
    DBG3  (( MF__, "T%d - converter cache too small in XPARAM", pUKT->pCTask->ulTaskIndex));
    }
  else
    {
    *ok = TRUE ;
    *bufptr = (tsp00_PageAddr ) kgs.pSharedConvCachePtr;
    kgs.pSharedConvCachePtr += PAGES_TO_BYTES(count);   // --- counts in units of BUFFER

    DBG3 (( MF__, "T%d - Selected 0x%08lx", pUKT->pCTask->ulTaskIndex, *bufptr ));
    }

  CLEARLOCK(&kgs.lSharedDynDataExcl);

  }

/*------------------------------*/
/* CR 1107152 */
VOID  valloc_datacache ( tsp00_Int4      count,
                         tsp00_Int4      *available,
                         tsp00_PageAddr  *bufptr,
                         bool            *ok )

{
  #undef  MF__
  #define MF__ MOD__"valloc_datacache"

  tsp00_Longuint  DataSize   = (XPARAM(ulDBPageSize) * (tsp00_Longuint)count);
  tsp00_Longuint  AlignValue = ALIGNMENT_VALUE;
  APIRET		      rc         = NO_ERROR;

  DBGPAS;

  *bufptr = (tsp00_PageAddr ) VirtualAlloc(NULL, DataSize + AlignValue,
                                           MEM_RESERVE | MEM_COMMIT,
                                           PAGE_READWRITE);

  MSGD(( INFO_MEM_ALLOCATED, count, "DATA CACHE", *bufptr ));

  if (!(*bufptr))
  { 
    rc = GetLastError();
    if (rc != NO_ERROR)
    {
      MSGALL(( ERR_ALLOC_MEMORY_XXX_BYTES_FILE_LINE_I64, (__int64)DataSize + AlignValue, rc, __FILE__, __LINE__ ));
      *ok = FALSE ;
      *available = 0 ;
    }
  } 
  else
  {
    *ok = TRUE ;
    *available = count ;
  }

}

/*------------------------------*/

VOID  vnewbuf ( tsp00_Int4      count,
                tsp00_Int4      *available,
                tsp00_PageAddr  *bufptr,
                bool            *ok )

  {
  #undef  MF__
  #define MF__ MOD__"vnewbuf"
  #ifdef DEBUG
   PUKT_CTRL_REC   pUKT = THIS_UKT_CTRL;
  #endif

  DBGPAS;

  WAIT_UNTIL_ALONE(&kgs.lSharedDynDataExcl);

  *available = (tsp00_Int4)BYTES_TO_PAGES(kgs.pSharedDynDataTop - kgs.pSharedDynDataPtr);

  DBG3 (( MF__, "T%d - count: %d - available: %d",
                pUKT->pCTask->ulTaskIndex, count, *available ));

  // --- counts in units of BUFFER
  if ( count > *available )
    {
    *ok = FALSE ;
    MSGALL(( WRN_SHRD_DYN_DATA_TOO_SMALL ));
    DBG3  (( MF__, "T%d - shareddyndata too small in XPARAM", pUKT->pCTask->ulTaskIndex));
    }
  else
    {
    *ok = TRUE ;
    *bufptr = (tsp00_PageAddr) kgs.pSharedDynDataPtr;
    kgs.pSharedDynDataPtr += PAGES_TO_BYTES(count);   // --- counts in units of BUFFER

    DBG3 (( MF__, "T%d - Selected 0x%08lx", pUKT->pCTask->ulTaskIndex, *bufptr ));
    }

  CLEARLOCK(&kgs.lSharedDynDataExcl);

  }

/*------------------------------*/


VOID vallocat ( tsp00_Int4      lengthInBytes,
                void            **objPointer,
                bool            *ok )
  {
  #undef  MF__
  #define MF__ MOD__"vallocat"
  tsp_int4           RealLengthInBytes;
  #ifdef DEBUG
   PUKT_CTRL_REC   pUKT = THIS_UKT_CTRL;
  #endif

  DBGPAS;

  DBG3 (( MF__, "T%d - lengthInBytes %d", pUKT->pCTask->ulTaskIndex, lengthInBytes ));

  RealLengthInBytes = ALIGN_EO00(lengthInBytes, ALIGNMENT_VALUE);
  WAIT_UNTIL_ALONE(&kgs.lSharedDynPoolExcl);

  // --- lengths ok?
  if ( kgs.pSharedDynPoolPtr + lengthInBytes > kgs.pSharedDynPoolTop )
    {
    CLEARLOCK(&kgs.lSharedDynPoolExcl);
    *ok = FALSE ;
    MSGALL(( WRN_SHRD_DYN_POOL_TOO_SMALL ));
    DBG3  (( MF__, "T%d - shareddynpool too small in XPARAM", pUKT->pCTask->ulTaskIndex));
    }
  else
    {
    *objPointer = (void * )kgs.pSharedDynPoolPtr;
    kgs.pSharedDynPoolPtr += RealLengthInBytes;
    CLEARLOCK(&kgs.lSharedDynPoolExcl);
    *ok = TRUE ;
#ifdef SEARCH_UNUSED_SPACE
        if ( RealLengthInBytes > lengthInBytes )
        {
    		MSGD (( 1, INFO_TYPE, _T("vallocat"), _T("%d unused (wanted %d at 0x%lx)"),
                    (RealLengthInBytes - lengthInBytes), lengthInBytes, (long)*objPointer ));
        }
#endif

    DBG3 (( MF__, "T%d - Selected 0x%08lx", pUKT->pCTask->ulTaskIndex, *objPointer ));
    }

  }

/*------------------------------*/

VOID vmalloc ( tsp00_Int4      lengthInBytes,
               void            **objPointer,
               bool            *ok )
  {
  #undef  MF__
  #define MF__ MOD__"vmalloc"
  APIRET           rc = NO_ERROR;
  #ifdef DEBUG
   PUKT_CTRL_REC   pUKT = THIS_UKT_CTRL;
  #endif

  DBGPAS;

  DBG3 (( MF__, "T%d - lengthInBytes %d", pUKT->pCTask->ulTaskIndex, lengthInBytes ));

  rc = sql57k_pool_malloc ( &KernelMemPool,
                            XPARAM(ulKernelAllocPoolSize),
                            "Kernel",
                            &kgs.ulKernelAllocPoolUsed,
                            &kgs.ulKernelAllocPoolCnt,
                            objPointer, lengthInBytes );

  if ( rc == NO_ERROR)
    {
    *ok = TRUE;
    DBG3 (( MF__, "T%d - Selected 0x%08lx", pUKT->pCTask->ulTaskIndex, *objPointer ));
    }
  else
    {
    DBG1 (( MF__, "Cannot allocate memory" ));
    *ok = FALSE;
    }
  }

/*------------------------------*/


VOID vmfree ( void  *objPointer )
  {
  #undef  MF__
  #define MF__ MOD__"vmfree"
  APIRET  rc = NO_ERROR;
  #ifdef DEBUG
   PUKT_CTRL_REC   pUKT = THIS_UKT_CTRL;
  #endif

  DBGPAS;

  DBG3 (( MF__, "T%d - objPointer: 0x%08lx", pUKT->pCTask->ulTaskIndex, objPointer ));

  rc = sql57k_pool_free( &kgs.ulKernelAllocPoolUsed,
                         &kgs.ulKernelAllocPoolCnt, (PUCHAR )objPointer );

  if ( rc != NO_ERROR)
    return;

  return;
  }


/*------------------------------*/

VOID vmnewbuf ( tsp00_Int4      count,
                tsp00_BufAddr  *bufptr,
                bool            *ok )
  {
  #undef  MF__
  #define MF__ MOD__"vmnewbuf"
  LONG             lLength;
  APIRET           rc = NO_ERROR;
  #ifdef DEBUG
   PUKT_CTRL_REC   pUKT = THIS_UKT_CTRL;
  #endif

  DBGPAS;

  DBG3 (( MF__, "T%d - count %d", pUKT->pCTask->ulTaskIndex, count ));

  // --- check overflow
  lLength = count * sizeof(tsp00_Buf);

  if ( (tsp00_Int4 )(lLength / sizeof(tsp00_Buf)) != count )
    {
    *ok = FALSE ;
    MSGD (( ERR_WRONG_VMNEWBUF_COUNT, count ));
    DBG1 (( MF__, "'vmnewbuf' requested count '%d' too large", count ));
    }
  else
    {
    rc = sql57k_pool_malloc ( &KernelMemPool,
                              XPARAM(ulKernelAllocPoolSize),
                              "Kernel",
                              &kgs.ulKernelAllocPoolUsed,
                              &kgs.ulKernelAllocPoolCnt,
                              (PPVOID)bufptr, lLength );

    if ( rc == NO_ERROR)
      {
      *ok = TRUE;
      DBG3 (( MF__, "T%d - Selected 0x%08lx", pUKT->pCTask->ulTaskIndex, *bufptr ));
      }
    else
      {
      *ok = FALSE;
      DBG1 (( MF__, "Cannot allocate memory" ));
      }
    }

  }

/*------------------------------*/

VOID vmfreebuf ( tsp00_PageAddr bufptr )
  {
  #undef  MF__
  #define MF__ MOD__"vmfreebuf"
  APIRET  rc = NO_ERROR;
  #ifdef DEBUG
   PUKT_CTRL_REC   pUKT = THIS_UKT_CTRL;
  #endif

  DBGPAS;

  DBG3 (( MF__, "T%d - bufptr: 0x%08lx", pUKT->pCTask->ulTaskIndex, bufptr ));

  rc = sql57k_pool_free( &kgs.ulKernelAllocPoolUsed,
                         &kgs.ulKernelAllocPoolCnt, bufptr );

  if ( rc != NO_ERROR)
    return;

  return;
  }

/*------------------------------*/

LONG  sql57k_pmalloc  ( ULONG     ulLine,
                        PSZ       pszModuleName,
                        PPVOID    ppMemory,
                        ULONG     ulSize )
  {
  #undef  MF__
  #define MF__ MOD__"sql57k_pmalloc"
  APIRET         rc      = NO_ERROR;
  static POOLTYP MemPool = NULL;
  static POOLTYP TmpPool = NULL;

  DBGIN;

  if ( hinstDLL == NULL )
    sql57c_load_functions();

  if ( kgs.XParam == 0 || XPARAM(ulRTEAllocPoolSize) == 0 )
    {
    rc = sql57k_pool_malloc ( &TmpPool, RTE_TMP_ALLOC_POOL_SIZE,
                              "TMP RTE", &kgs.ulRTEAllocPoolUsed,
                              &kgs.ulRTEAllocPoolCnt, ppMemory, ulSize );
    }
  else
    {
    rc = sql57k_pool_malloc ( &MemPool, XPARAM(ulRTEAllocPoolSize),
                              "RTE", &kgs.ulRTEAllocPoolUsed,
                              &kgs.ulRTEAllocPoolCnt, ppMemory, ulSize );
    }

  #if defined (DEBUG_MEM_ALLOC)
   MSGD(( 19999,INFO_TYPE, "ALLOC   ","'%4d':'%s' - 0x%08x, size: 0x%d",
          ulLine, pszModuleName, *ppMemory, ulSize ));
  #endif

  if ( rc != NO_ERROR)
    {
    DBGOUT;
    return ( rc );
    }



  DBGOUT;
  return ( rc );
  }

/*------------------------------*/

LONG  sql57k_pfree ( ULONG   ulLine,
                     PSZ     pszModuleName,
                     PVOID   pMemory )
  {
  #undef  MF__
  #define MF__ MOD__"sql57k_pfree"
  APIRET  rc = NO_ERROR;

  DBGIN;

  #if defined (DEBUG_MEM_ALLOC)
   MSGD(( 19999,INFO_TYPE, "FREE    ","'%4d':'%s' - 0x%08x ",
          ulLine, pszModuleName, pMemory ));
  #endif

  rc = sql57k_pool_free ( &kgs.ulRTEAllocPoolUsed,
                          &kgs.ulRTEAllocPoolCnt, pMemory );

  if ( rc != NO_ERROR)
    {
    DBGOUT;
    return ( rc );
    }

  DBGOUT;
  return ( rc );
  }


/*------------------------------*/

LONG sql57k_alloc_vnewbuf_pool ( VOID )
  {
  #undef  MF__
  #define MF__ MOD__"sql57k_alloc_vnewbuf_pool"
  APIRET            rc     = NO_ERROR;
  tsp00_Longuint    DataSize, ConvSize;
  tsp00_Longuint    AlignValue = 4096;  // --- must be aligned on disk sector
                                          //     boundaries in memory to allow the
                                          //     use of the NO_BUFFERING/NO_CACHE
                                          //     file open option !

  DBGIN;

  /* PTS 1107152 */
  DataSize = (XPARAM(ulDBPageSize) * 
             (tsp00_Longuint)( XPARAM(lSharedDynData) - XPARAM(lDataCachePages) - XPARAM(lConvCachePages)));

  kgs.pSharedDynDataBot = VirtualAlloc(NULL, DataSize + AlignValue,
                                       MEM_RESERVE | MEM_COMMIT,
                                       PAGE_READWRITE);


  if (!(kgs.pSharedDynDataBot))
  {
    rc = GetLastError();
    MSGALL(( ERR_ALLOC_MEMORY_XXX_BYTES_FILE_LINE_I64, (__int64)DataSize + AlignValue, rc, __FILE__, __LINE__ ));
  }
  else
  {
    MSGD(( INFO_MEM_ALLOCATED, DataSize / XPARAM(ulDBPageSize) , 
                              "SHAREDDYNDATA", kgs.pSharedDynDataBot));
  }


  if (rc == NO_ERROR)
  {
    ConvSize                = (XPARAM(ulDBPageSize) * (tsp00_Longuint)XPARAM(lConvCachePages));
    kgs.pSharedConvCacheBot = VirtualAlloc(NULL, ConvSize + AlignValue,
                                           MEM_RESERVE | MEM_COMMIT,
                                           PAGE_READWRITE);

    if (!(kgs.pSharedConvCacheBot))
		{
        rc = GetLastError();
        MSGALL(( ERR_ALLOC_MEMORY_XXX_BYTES_FILE_LINE_I64, (__int64)ConvSize + AlignValue, rc, __FILE__, __LINE__ ));
		}
		else
		{
        MSGD(( INFO_MEM_ALLOCATED, XPARAM(lConvCachePages), "CONVERTER CACHE", kgs.pSharedConvCacheBot));
		}
  }


  if ( rc != NO_ERROR )
    {
    DBG1  (( MF__, "Cannot allocate memory, rc = %u", rc));
    return ( rc );
    }

  kgs.pSharedDynDataPtr   = (PVOID)ALIGN((tsp00_Longuint)kgs.pSharedDynDataBot, AlignValue);
  kgs.pSharedDynDataTop   = (PVOID)((PUCHAR)kgs.pSharedDynDataPtr + DataSize);

  kgs.pSharedConvCachePtr = (PVOID)ALIGN((tsp00_Longuint)kgs.pSharedConvCacheBot, AlignValue);
  kgs.pSharedConvCacheTop = (PVOID)((PUCHAR)kgs.pSharedConvCachePtr + ConvSize);

  DBGOUT;

  return ( rc );
  }

/*------------------------------*/

LONG sql57k_alloc_vallocat_pool ( VOID )
  {
  #undef  MF__
  #define MF__ MOD__"sql57k_alloc_vallocat_pool"
  APIRET   rc     = NO_ERROR;
  tsp00_Longuint    Size;
  tsp00_Longuint    AlignValue = ALIGNMENT_VALUE;

  DBGIN;

  Size = (1024 * (tsp00_Longuint)XPARAM(lSharedDynPool));
  kgs.pSharedDynPoolBot = VirtualAlloc(NULL, Size + AlignValue,
                                       MEM_RESERVE | MEM_COMMIT,
                                       PAGE_READWRITE);
  if (!(kgs.pSharedDynPoolBot)) /* PTS 1103758 */
   rc = GetLastError();

  if (rc != NO_ERROR)
    {
    MSGALL(( ERR_ALLOC_MEMORY_XXX_BYTES_FILE_LINE_I64, (__int64)Size + AlignValue, rc, __FILE__, __LINE__ ));
    DBG1  (( MF__, "Cannot allocate memory, rc = %u", rc ));
    return ( rc );
    }
  else
  {
    MSGD(( INFO_MEM_ALLOCATED, ALIGN_EO00(Size, XPARAM(ulDBPageSize)) / XPARAM(ulDBPageSize), 
           "SHAREDDYNPOOL", kgs.pSharedDynPoolBot));
  }

  kgs.pSharedDynPoolPtr = (PVOID)ALIGN((tsp00_Longuint)kgs.pSharedDynPoolBot, AlignValue);
  kgs.pSharedDynPoolTop = (PVOID)((PUCHAR)kgs.pSharedDynPoolPtr + Size);

  DBG3 (( MF__, "SharedDynPool      %d     ", XPARAM(lSharedDynPool) ));
  DBG3 (( MF__, "pSharedDynPoolBot: 0x%08lx", kgs.pSharedDynPoolBot ));
  DBG3 (( MF__, "pSharedDynPoolPtr: 0x%08lx", kgs.pSharedDynPoolPtr ));
  DBG3 (( MF__, "pSharedDynPoolTop: 0x%08lx", kgs.pSharedDynPoolTop ));

  DBGOUT;

  return ( rc );
  }


/*===========================================================================*
 *  LOCAL FUNCTIONS (CODE)                                                   *
 *===========================================================================*/


static LONG  sql57k_pool_malloc  ( POOLTYP *pPool,
                                   ULONG   ulPoolSize,
                                   PSZ     pszPoolName,
                                   PULONG  pulPoolUsed,
                                   PULONG  pulNumOfAllocs,
                                   PPVOID  ppMemory,
                                   ULONG   ulSize )
  {
  #undef  MF__
  #define MF__ MOD__"sql57k_pool_malloc"
  APIRET         rc                 = NO_ERROR;
  static BOOL    fMessOutputPending = FALSE;
  PMEM_INFO_REC  pMemInfo;

  DBGIN;

  if ( !*pPool )
    {
     *pPool = CALL_FUNC( HeapCreateP, HeapCreate,(0, 1, 0));

     if ( *pPool == NULL)
       {
       rc = ERROR_NOT_ENOUGH_MEMORY;

       // --- message output might produce additional allocation errors
       if ( !fMessOutputPending )
         {
         fMessOutputPending = TRUE;
         MSGALL(( ERR_ALLOC_MEMORY_XXX_BYTES_FILE_LINE_, ulPoolSize, rc, __FILE__, __LINE__ ));
         DBG1  (( MF__, "Cannot allocate memory, rc = %u", rc ));
         fMessOutputPending = FALSE;
         }

       *pPool = NULL;
       DBGOUT;
       return ( rc );
       }
    }

  ulSize += sizeof (MEM_INFO_REC);

  if ( ulSize < sizeof (MEM_INFO_REC) )  // - overrun error
     rc = ERROR_NOT_ENOUGH_MEMORY;
  else
    {
    *ppMemory = CALL_FUNC( HeapAllocP, HeapAlloc,
                               ( *pPool, 0, ulSize ));

    if ( *ppMemory == NULL )
      rc = ERROR_NOT_ENOUGH_MEMORY;
#ifdef WANT_SPECIAL_MEM_CHUNK_INFO_CODE
    else
      sql57_AddToMemChunkInfo(*ppMemory, ulSize);
#endif /* WANT_SPECIAL_MEM_CHUNK_INFO_CODE */
    }

  if ( rc != NO_ERROR)
    {
    // --- message output might produce additional allocation errors
    if ( !fMessOutputPending )
      {
      fMessOutputPending = TRUE;
      MSGALL(( ERR_CANT_ALLOC_HEAP_MEMORY, rc ));
      DBG1  (( MF__, "Cannot allocate heap memory, rc = %u", rc ));
      fMessOutputPending = FALSE;
      }
    *ppMemory = NULL;
    DBGOUT;
    return ( rc );
    }

  // --- save memory information
  pMemInfo = (PMEM_INFO_REC)*ppMemory;
  pMemInfo->ulSize = ulSize;
  pMemInfo->Pool   = *pPool;
  *(PUCHAR *) ppMemory += sizeof (MEM_INFO_REC);

  WAIT_UNTIL_ALONE( &lExclusive );
  (*pulPoolUsed)    += ulSize;
  (*pulNumOfAllocs)++;
  CLEARLOCK( &lExclusive );

  DBG3 (( MF__, "Memory Addr.: 0x%lx - Size: %d ", *ppMemory,
                ulSize - sizeof (ULONG)));


  DBGOUT;
  return ( rc );
  }

/*------------------------------*/

static LONG  sql57k_pool_free ( PULONG  pulPoolUsed,
                                PULONG  pulNumOfAllocs,
                                PVOID   pMemory )
  {
  #undef  MF__
  #define MF__ MOD__"sql57k_pool_free"
  APIRET         rc = NO_ERROR;
  PMEM_INFO_REC  pMemInfo;

  DBGIN;

  DBG3 (( MF__, "Memory Addr.: 0x%lx ", pMemory ));

  if ( pMemory )
    {
    pMemory  = (PUCHAR) pMemory - sizeof (MEM_INFO_REC);
    pMemInfo = (PMEM_INFO_REC)pMemory;

#ifdef WANT_SPECIAL_MEM_CHUNK_INFO_CODE
    sql57_RemoveFromMemChunkInfo(pMemory);
#endif /* WANT_SPECIAL_MEM_CHUNK_INFO_CODE */

    WAIT_UNTIL_ALONE( &lExclusive );
    (*pulPoolUsed) -= pMemInfo->ulSize;
    (*pulNumOfAllocs)--;
    CLEARLOCK( &lExclusive );

    if ( !CALL_FUNC( HeapFreeP, HeapFree,( pMemInfo->Pool, 0, pMemory )) )
      rc = ERROR_INVALID_BLOCK;
    }
  else
    rc = ERROR_INVALID_BLOCK;

  if ( rc != NO_ERROR)
    {
    MSGD (( ERR_CANT_FREE_HEAP_MEMORY, rc ));
    DBG1 (( MF__, "Cannot free heap memory, rc = %u", rc ));
    DBGOUT;
    return ( rc );
    }

  DBGOUT;

  return ( rc );
  }

/*------------------------------*/

static ULONG sql57c_load_functions(  VOID )
  {
  #undef  MF__
  #define MF__ MOD__"sql57c_load_functions"
  LONG             rc         = NO_ERROR;
  static BOOL      fFirstTime = FALSE;

  DBGIN;

  if ( !fFirstTime )
    {
    rc = LOAD_FUNC( HeapCreateP );

    if (rc == NO_ERROR ) rc = LOAD_FUNC( HeapAllocP );
    if (rc == NO_ERROR ) rc = LOAD_FUNC( HeapFreeP );
    if (rc == NO_ERROR ) rc = LOAD_FUNC( HeapDestroyP );

    if ( rc != NO_ERROR )
      {
      lpHeapCreateP  =  NULL;
      lpHeapAllocP   =  NULL;
      lpHeapFreeP    =  NULL;
      lpHeapDestroyP =  NULL;
      FreeLibrary( hinstDLL);
      hinstDLL = NULL;
      }
    fFirstTime = TRUE;
    }

  DBGOUT;
  return ( rc );
  }

/*------------------------------*/

static VOID* sql57c_load_func(  PSZ  pszFuncName )
  {
  #undef  MF__
  #define MF__ MOD__"sql57c_load_func"
  static PSZ       pszDLL;
  HINSTANCE        hinst = NULL;
  LONG             rc    = NO_ERROR;
  FARPROC          pFunc;

  DBGIN;

  if ( hinstDLL == NULL )
    {
    hinst = LoadLibrary( "adamema.dll" );

    if ( hinst == NULL )
      {
      DBGOUT;
      return ( NULL );
      }

    hinstDLL = hinst;
    }

  pFunc = GetProcAddress( hinstDLL, pszFuncName);

  DBGOUT;
  return ( pFunc );
  }


/*------------------------------*/
#ifdef WANT_SPECIAL_MEM_CHUNK_INFO_CODE

#define SQL57_MEMCHUNK_DUMP_BIN_FILE "MemChunkDump.bin"
#define SQL57_MEMCHUNK_DUMP_TXT_FILE "MemChunkDump.txt"
#define SQL57_CALL_ARRAY_FILL_LEVEL   12
#define SQL57_CALL_ARRAY_MAX_ENTRIES (64*1024)
#define SQL57_TICK_INTERVALL (5000)
#define INFO_MEMCHUNK_INITIALIZATION    N(1899),INFO_TYPE,_T("MEMORY  "), _T("Memory chunk info collection initialized for %d entries")
#define WRN_MEMCHUNK_INITIALIZATION     N(1899), WRN_TYPE,_T("MEMORY  "), _T("Memory chunk info collection initialization failed (out of memory)")
#define WRN_MEMCHUNK_LIMIT_EXCEEDED     N(1899), WRN_TYPE,_T("MEMORY  "), _T("Memory chunk info collection limit exceeded")
#define INFO_MEMCHUNK_ACTIVATED         N(1899),INFO_TYPE,_T("MEMORY  "), _T("Memory chunk info collection activated")
#define INFO_MEMCHUNK_DEACTIVATED       N(1899),INFO_TYPE,_T("MEMORY  "), _T("Memory chunk info collection deactivated")
#define INFO_MEMCHUNK_DUMP_START        N(1899),INFO_TYPE,_T("MEMORY  "), _T("Memory chunk info dump to '%s' %d entries")
#define INFO_MEMCHUNK_DUMP_COMPLETE     N(1899),INFO_TYPE,_T("MEMORY  "), _T("Memory chunk info dump to '%s' completed")
#define INFO_MEMCHUNK_DUMP_ENTRIES_LEFT N(1899),INFO_TYPE,_T("MEMORY  "), _T("Memory chunk info dump %d entries left")
#define WRN_MEMCHUNK_DUMP_OPEN          N(1899), WRN_TYPE,_T("MEMORY  "), _T("Memory chunk info open of '%s' failed [%d]")
#define WRN_MEMCHUNK_DUMP_WRITE         N(1899), WRN_TYPE,_T("MEMORY  "), _T("Memory chunk info write failed [%d]")

#include "heo670.h" /* nocheck */

typedef struct sql57_MemChunkInfoEntry_ {
    void       *addr;
    void       *callArray[SQL57_CALL_ARRAY_FILL_LEVEL];
    tsp00_Int4  callLevel;
    tsp00_Int4 size;
    tsp00_Int4 allocId;
} sql57_MemChunkInfoEntry;
    
static struct sql57_MemChunkInfo_ {
    tsp00_Int4         headerSize;
    bool               initializing;
    bool               active;
    LONG               lExclusive;
    tsp00_Int4         entriesStored;
    long               allocCalls;
    long               freeCalls;
    long               countedAllocCalls;
    long               countedFreeCalls;
    sql57_MemChunkInfoEntry *first;
    sql57_MemChunkInfoEntry *last;
}
 MemChunkInfo =
{
    0,
    false,
    false,
    0,
    0,
    0,
    0,
    0,
    0,
    (sql57_MemChunkInfoEntry *)0,
    (sql57_MemChunkInfoEntry *)0
};

/*----------------------------------------------
   Function: sql57_InitializeMemChunkInfo
   Description: Initialize MemChunkInfo and allocate its memory

   The initializing member is set to true first to allow 'sql57k_pmalloc()'. Otherwise it would
   be called recursive due to sql57_AddToMemChunkInfo() called inside sql57k_pmalloc()...

   The memory chunk info contains its own header size to allow to interpret the binary dump if needed.
   It also counts alloc calls and free calls even if not activated.

   It is activated by the 'kgs.ulDebugLevel' which is set with the x_cons command x_cons SID DEBUGLEVEL 4
   The info dump is triggered with x_cons SID DEBUGLEVEL 5 and will take place with the next allocation
   (i.e. the next connect request...) This is only a 'hack' for the specialists...

   The binary file 'MemChunkDump.bin' is created in the RunDirectory. It consists of the dump of 'MemChunkInfo'
   follwed by the dump of 'MemChunkInfo.entriesStored' * sql57_MemChunkInfoEntry. (see above).

   The text file 'MemChunkDump.txt' is created afterwards in the RunDirectory. It consists of the ascii 
   output of the MemChunkInfo follwed by the interpretation of the sql57_MemChunkInfoEntry array, where
   identical call stacks are dumped only once.
 */
static void sql57_InitializeMemChunkInfo()
{
    void *space;

    MemChunkInfo.initializing = true;

    if ( NO_ERROR == sql57k_pmalloc  ( __LINE__,
                                       __FILE__,
                                       &space,
                                       SQL57_CALL_ARRAY_MAX_ENTRIES * sizeof(sql57_MemChunkInfoEntry) ) )
    {
        WAIT_UNTIL_ALONE( &MemChunkInfo.lExclusive );
            MemChunkInfo.active = false;
            MemChunkInfo.entriesStored = 0;
            MemChunkInfo.first = space;
            MemChunkInfo.last = 0;
            MemChunkInfo.headerSize = sizeof(MemChunkInfo);
        CLEARLOCK( &MemChunkInfo.lExclusive );
        MSGD(( INFO_MEMCHUNK_INITIALIZATION, SQL57_CALL_ARRAY_MAX_ENTRIES ));
    }
    else
    {
        MSGD(( WRN_MEMCHUNK_INITIALIZATION ));
    }
}

/* Only a single output buffer needed, since output is spinlock protected */
static char sql57_bufferLine[1024];

/*----------------------------------------------
   Function: sql57_WriteBufferToFile
   Description: Write a buffer completely into file

   This routine will not return until all bytes given are written, or an error occured
 */   
static APIRET sql57_WriteBufferToFile(HANDLE fileHandle, char *buffer, tsp00_Int4 length)
{
    DWORD bytesToWrite = (DWORD)length;
    DWORD bytesWritten = 0;
    APIRET rc = 0;

    while ( bytesToWrite > 0 )
    {
        if ( !WriteFile( fileHandle,
                        buffer + bytesWritten, 
                        bytesToWrite,
                        &bytesWritten,
                        NULL) )
        {
            return GetLastError();
        }
        bytesToWrite -= bytesWritten;
    }
    return NO_ERROR;
}

/*----------------------------------------------
   Function: sql57_ShowAddressInfo
   Description: Show the content of the address info

   This routine will show symbol and all optional informations like line numbers and module name.
   It writes to given file handle a single entry only
 */   
static void sql57_ShowAddressInfo( HANDLE fileHandle, 
                                   SAPDB_Int4 loopIndex, 
                                   void *pcAddr, 
                                   eo670AddressInfo *addrInfo)
{
    APIRET rc = 0;

    if ( addrInfo->offsetToSymbol < 0 )
    {
      _snprintf( sql57_bufferLine, sizeof(sql57_bufferLine), 
            "   %d: %s - 0x%lx %s\n",
            loopIndex,
            addrInfo->moduleName,
            (SAPDB_Long)((char *)pcAddr-(char *)0),
            addrInfo->symbol);
    }
    else
    {
      _snprintf( sql57_bufferLine, sizeof(sql57_bufferLine), 
            "   %d: %s - 0x%lx %s + 0x%x\n",
            loopIndex,
            addrInfo->moduleName,
            (SAPDB_Long)((char *)pcAddr-(char *)0), 
            (0 != *addrInfo->demangledSymbol) ? addrInfo->demangledSymbol : addrInfo->symbol, 
            addrInfo->offsetToSymbol );
        rc = sql57_WriteBufferToFile( fileHandle,
                                sql57_bufferLine, 
                                strlen(sql57_bufferLine) );
        if ( NO_ERROR != rc ) 
        {
            MSGD (( WRN_MEMCHUNK_DUMP_OPEN, SQL57_MEMCHUNK_DUMP_TXT_FILE, 0 ));
            MSGD (( WRN_MEMCHUNK_DUMP_WRITE, rc ));
        }

      if ( addrInfo->lineNumber >= 0 )
      {
        _snprintf( sql57_bufferLine, sizeof(sql57_bufferLine), 
            "    %s - %s:%d\n",
            addrInfo->moduleName,
            addrInfo->sourceFileName,
            addrInfo->lineNumber );
      }
      else if ( 0 != *addrInfo->sourceFileName )
      {
        _snprintf( sql57_bufferLine, sizeof(sql57_bufferLine), 
            "    %s - %s\n",
            addrInfo->moduleName,
            addrInfo->sourceFileName );
      }
      else
      {
        sql57_bufferLine[0] = 0;
      }

    }

    if ( sql57_bufferLine[0] )
    {
        rc = sql57_WriteBufferToFile( fileHandle,
                                sql57_bufferLine, 
                                strlen(sql57_bufferLine) );
        if ( NO_ERROR != rc ) 
        {
            MSGD (( WRN_MEMCHUNK_DUMP_OPEN, SQL57_MEMCHUNK_DUMP_TXT_FILE, 0 ));
            MSGD (( WRN_MEMCHUNK_DUMP_WRITE, rc ));
        }
    }
}

/*----------------------------------------------
   Function: sql57_DumpCallStack
   Description: Dump a call stack

   This routine will dump all stack entries in symbolic form up to depth given
 */   
static void sql57_DumpCallStack( HANDLE fileHandle, 
                                 tsp00_Int4 depth, 
                                 void **callArray)
{
    void *context;
    tsp00_Int4 iLevel = 0;

    if ( depth > 0 
      && eo670_GetAddressInfoProlog(&context) )
    {
        for ( iLevel = 0; iLevel < depth; iLevel++ )
        {
            eo670AddressInfo localAddrInfo;

            if ( eo670_GetAddressInfo(callArray[iLevel],
                                      &localAddrInfo,
                                      context) )
            {
                sql57_ShowAddressInfo(fileHandle, iLevel, callArray[iLevel], &localAddrInfo);

                eo670_FreeAddressInfo(&localAddrInfo);
            }
        }
        eo670_GetAddressInfoEpilog(context);
    }
}

/*----------------------------------------------
   Function: sql57_CheckForMessage
   Description: Check for time for another message to user

   This routine will check if dump takes that long, that another message should be written into knldiag
   to reduce stress for adminstrators...
 */   
static DWORD sql57_CheckForMessage(DWORD startTick, DWORD interval, tsp00_Int4 entriesLeft)
{
    bool showMessage;
    DWORD currentTick = GetTickCount();

    if ( (startTick + interval) > startTick )
    {
        showMessage = ( currentTick >= (startTick + interval) );
    }
    else
    {
        showMessage = ( (currentTick < startTick) && (currentTick >= (startTick + interval)) );
    }

    if ( showMessage )
    {
        MSGD(( INFO_MEMCHUNK_DUMP_ENTRIES_LEFT, entriesLeft ));
    }

    return ( showMessage ? currentTick : startTick );
}

/*----------------------------------------------
   Function: sql57_DumpMemChunkInfo
   Description: Dump the complete memory chunk info

   This routine will first dump the binary information into a file and after that is done show
   the symbolic information. Doubled call stacks are only shown once to reduce the time needed
   for dumping since symbol resolution takes lot of time. Identical call stacks are marked in the
   output.
 */   
static void sql57_DumpMemChunkInfo()
{
    HANDLE fileHandle;
    APIRET rc = 0;

    if ( !MemChunkInfo.headerSize )
        return;

    /*------ Binary output file ---------*/

    fileHandle = CreateFile( "MemChunkDump.bin",
                            GENERIC_READ |
                            GENERIC_WRITE,
                            FILE_SHARE_READ,
                            NULL,
                            CREATE_ALWAYS,
                            FILE_ATTRIBUTE_NORMAL,
                            NULL );

    if ( fileHandle == INVALID_HANDLE_VALUE )
    {
        rc = GetLastError();
        MSGD (( WRN_MEMCHUNK_DUMP_OPEN, SQL57_MEMCHUNK_DUMP_BIN_FILE, rc ));
        return;
    }

    MSGD(( INFO_MEMCHUNK_DUMP_START, SQL57_MEMCHUNK_DUMP_BIN_FILE, MemChunkInfo.entriesStored ));

    rc = sql57_WriteBufferToFile( fileHandle,
                            (char *)&MemChunkInfo,
                            sizeof(MemChunkInfo) );
    if ( NO_ERROR != rc ) 
    {
        MSGD (( WRN_MEMCHUNK_DUMP_OPEN, SQL57_MEMCHUNK_DUMP_BIN_FILE " Header", 0 ));
        MSGD (( WRN_MEMCHUNK_DUMP_WRITE, rc ));
    }

    rc = sql57_WriteBufferToFile( fileHandle,
                            (char *)MemChunkInfo.first, 
                            MemChunkInfo.entriesStored * sizeof(sql57_MemChunkInfoEntry) );
    if ( NO_ERROR != rc ) 
    {
        MSGD (( WRN_MEMCHUNK_DUMP_OPEN, SQL57_MEMCHUNK_DUMP_BIN_FILE, 0 ));
        MSGD (( WRN_MEMCHUNK_DUMP_WRITE, rc ));
    }

    CloseHandle(fileHandle);

    MSGD(( INFO_MEMCHUNK_DUMP_COMPLETE, SQL57_MEMCHUNK_DUMP_BIN_FILE ));

    /*------ Text output file ---------*/

    fileHandle = CreateFile( "MemChunkDump.txt",
                            GENERIC_READ            |
                            GENERIC_WRITE,
                            FILE_SHARE_READ,
                            NULL,
                            CREATE_ALWAYS,
                            FILE_ATTRIBUTE_NORMAL,
                            NULL);
    if ( fileHandle == INVALID_HANDLE_VALUE )
    {
        rc = GetLastError();
        MSGD (( WRN_MEMCHUNK_DUMP_OPEN, SQL57_MEMCHUNK_DUMP_TXT_FILE, rc ));
        return;
    }

    MSGD(( INFO_MEMCHUNK_DUMP_START, SQL57_MEMCHUNK_DUMP_TXT_FILE, MemChunkInfo.entriesStored ));

    if ( MemChunkInfo.last )
    {
        DWORD startTick = GetTickCount();
        tsp00_Int4 iEntry = 0;
        tsp00_Int4 uniqueCount = 0;

        _snprintf( sql57_bufferLine, sizeof(sql57_bufferLine),
                   "Total alloc calls %ld\n" 
                   "Total free calls %ld\n"
                   "Dump of %d entries\n" 
                   "Counted alloc calls %ld\n" 
                   "Counted free calls %ld\n",
                   MemChunkInfo.allocCalls,
                   MemChunkInfo.freeCalls,
                   MemChunkInfo.entriesStored,
                   MemChunkInfo.countedAllocCalls,
                   MemChunkInfo.countedFreeCalls );
        rc = sql57_WriteBufferToFile( fileHandle,
                                sql57_bufferLine, 
                                strlen(sql57_bufferLine) );
        if ( NO_ERROR != rc ) 
        {
            MSGD (( WRN_MEMCHUNK_DUMP_OPEN, SQL57_MEMCHUNK_DUMP_TXT_FILE, 0 ));
            MSGD (( WRN_MEMCHUNK_DUMP_WRITE, rc ));
        }

        for ( iEntry = 0; iEntry < MemChunkInfo.entriesStored; iEntry++ )
        {
            tsp00_Int4 jEntry = 0;

            startTick = sql57_CheckForMessage(startTick, SQL57_TICK_INTERVALL, MemChunkInfo.entriesStored - iEntry);

            if ( MemChunkInfo.first[iEntry].size >= 0 )
            {
                ++uniqueCount;

                _snprintf( sql57_bufferLine, sizeof(sql57_bufferLine), "\n---- Entry[%d] ---- %6d[%d]: Chunk at 0x%p size 0x%x bytes\n",
                            uniqueCount, iEntry, MemChunkInfo.first[iEntry].allocId,
                            MemChunkInfo.first[iEntry].addr,
                            MemChunkInfo.first[iEntry].size );
                rc = sql57_WriteBufferToFile( fileHandle,
                                        sql57_bufferLine, 
                                        strlen(sql57_bufferLine) );
                if ( NO_ERROR != rc ) 
                {
                    MSGD (( WRN_MEMCHUNK_DUMP_OPEN, SQL57_MEMCHUNK_DUMP_TXT_FILE, 0 ));
                    MSGD (( WRN_MEMCHUNK_DUMP_WRITE, rc ));
                }

                sql57_DumpCallStack(fileHandle, MemChunkInfo.first[iEntry].callLevel, MemChunkInfo.first[iEntry].callArray);

                for ( jEntry = (iEntry+1); jEntry < MemChunkInfo.entriesStored; jEntry++ )
                {
                    if ( (MemChunkInfo.first[jEntry].size >= 0)
                      && (MemChunkInfo.first[iEntry].callLevel == MemChunkInfo.first[jEntry].callLevel) 
                      && !memcmp( MemChunkInfo.first[iEntry].callArray,
                                  MemChunkInfo.first[jEntry].callArray,
                                  sizeof(void *) * MemChunkInfo.first[iEntry].callLevel ) )
                    {
                        _snprintf( sql57_bufferLine, sizeof(sql57_bufferLine), "---> Identical callstack Entry %6d: Chunk at 0x%p size 0x%x bytes\n",
                                    jEntry, 
                                    MemChunkInfo.first[jEntry].addr,
                                    MemChunkInfo.first[jEntry].size );
                        rc = sql57_WriteBufferToFile( fileHandle,
                                                sql57_bufferLine, 
                                                strlen(sql57_bufferLine) );
                        if ( NO_ERROR != rc ) 
                        {
                            MSGD (( WRN_MEMCHUNK_DUMP_OPEN, SQL57_MEMCHUNK_DUMP_TXT_FILE, 0 ));
                            MSGD (( WRN_MEMCHUNK_DUMP_WRITE, rc ));
                        }

                        MemChunkInfo.first[jEntry].size = -1;
                    }
                }
            }
        }
    }

    CloseHandle(fileHandle);

    MSGD(( INFO_MEMCHUNK_DUMP_COMPLETE, SQL57_MEMCHUNK_DUMP_TXT_FILE ));

}

/*----------------------------------------------
   Function: sql57_AddToMemChunkInfo
   Description: Insert function for memory chunk info

   Called after memory was successfully requested. This adds an entry at the end of the memory
   chunk info array and sets 'last' pointer acoordingly. It also does the statistics.
 */   
static void sql57_AddToMemChunkInfo(void *addr, tsp00_Int4 size)
{
    WAIT_UNTIL_ALONE( &MemChunkInfo.lExclusive );

    if ( !MemChunkInfo.headerSize )
    {
        if ( kgs.ulDebugLevel == 4 || kgs.ulDebugLevel == 5 )
        {
            CLEARLOCK( &MemChunkInfo.lExclusive );
            if ( !MemChunkInfo.initializing )
            {
                sql57_InitializeMemChunkInfo();
            }
            WAIT_UNTIL_ALONE( &MemChunkInfo.lExclusive );
        }
    }

    {
        ++MemChunkInfo.allocCalls;

        if ( MemChunkInfo.headerSize )
        {
            if ( kgs.ulDebugLevel == 4 || kgs.ulDebugLevel == 5 )
            {
                if ( !MemChunkInfo.active )
                {
                    MSGD(( INFO_MEMCHUNK_ACTIVATED ));
                    MemChunkInfo.active = true;
                }

                if ( MemChunkInfo.entriesStored < SQL57_CALL_ARRAY_MAX_ENTRIES )
                {
                    sql57_MemChunkInfoEntry *entryPointer = MemChunkInfo.last;

                    if ( !entryPointer )
                    {
                        entryPointer = MemChunkInfo.first;
                    }
                    else
                    {
                        ++entryPointer;
                    }
                    entryPointer->callLevel = eo670_FillCallArray(SQL57_CALL_ARRAY_FILL_LEVEL, entryPointer->callArray);
                    entryPointer->addr = addr;
                    entryPointer->size = size;
                    entryPointer->allocId = MemChunkInfo.allocCalls;

                    MemChunkInfo.last = entryPointer;
                    ++MemChunkInfo.entriesStored;
                    ++MemChunkInfo.countedAllocCalls;
                }

                if ( kgs.ulDebugLevel == 5 )
                {
                    sql57_DumpMemChunkInfo();
                    kgs.ulDebugLevel = 4;
                }
            }
            else
            {
                if ( MemChunkInfo.active )
                {
                    MSGD(( INFO_MEMCHUNK_DEACTIVATED ));
                    MemChunkInfo.active = false;
                }
            }
        }
    }

    CLEARLOCK( &MemChunkInfo.lExclusive );
}

/*----------------------------------------------*/
#include <string.h>

/*----------------------------------------------
   Function: sql57_RemoveFromMemChunkInfo
   Description: Remove function for memory chunk info

   Called before memory is released. This searchs for the corresponding entry and removes it from
   the memory chunk array. The rest of the array is moved, so the array is always in the order of
   oldest memory request first...
 */   
static void sql57_RemoveFromMemChunkInfo(void *addr)
{
    WAIT_UNTIL_ALONE( &MemChunkInfo.lExclusive );
    {
        ++MemChunkInfo.freeCalls;

        if ( MemChunkInfo.headerSize )
        {
            sql57_MemChunkInfoEntry *entryPointer = MemChunkInfo.last;
            if ( entryPointer )
            {
                do 
                {
                    if ( addr == entryPointer->addr )
                    {
                        if ( entryPointer != MemChunkInfo.last )
                        {
                            memmove( entryPointer, (entryPointer+1), 
                                     (MemChunkInfo.last-entryPointer) * sizeof(sql57_MemChunkInfoEntry) );
                            --MemChunkInfo.last;
                        }
                        else
                        {
                            if ( entryPointer == MemChunkInfo.first )
                            {
                                MemChunkInfo.last = 0;
                            }
                            else
                            {
                                --MemChunkInfo.last;
                            }
                        }
                        --MemChunkInfo.entriesStored;
                        ++MemChunkInfo.countedFreeCalls;
                        break;
                    }
                } while ( (entryPointer--) != MemChunkInfo.first );
            }
        }
    }
    CLEARLOCK( &MemChunkInfo.lExclusive );
}

#endif /* WANT_SPECIAL_MEM_CHUNK_INFO_CODE */

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/