/****************************************************************************

  module      : vos92.c

  -------------------------------------------------------------------------

  responsible : FrankH

  special area: cons (x_cons) ...
  description : NT main module ...


  last changed: 2000-03-29  13:24
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1999



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "hsp100.h"
#include "veo00xc.h"
#include "geo00_2.h"
#include "geo002.h"
#include "heo46.h"
#include "heo92.h"
#include "gos00.h"
#include "vos00kc.h"
#include "vos00cnc.h"
#include "geo007_1.h"
#include "geo007_2.h"

#ifdef  RUN_AS_SYSMON
#include "geo921.h" /* nocheck */
#endif

#ifdef  CTRLCOMP
#include "vsp004.h" /* nocheck */
#include "vsp009c.h" /* nocheck */
#include "veo005c.h" /* nocheck */
#endif

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define MOD__  "VOS92.C : "
#define MF__   MOD__"UNDEFINED"

#define MAX_WAIT          30       // --- 10 sec.
#define MAX_ERRSTRING     200
#define ARGS(max, cnt)      ((max-cnt) >= 1)

#define COM_UNDEF           0
#define COM_CANCEL          1
#define COM_KILL            2
#define COM_SHOW_UPDATE     3
#define COM_HELP            4
#define COM_DROP            5
#define COM_TIME_MEASURE    6
#define COM_RESET           7
#define COM_GETPARAM        8
#define COM_PUTPARAM        9
#define COM_CRASH_DUMP     10
#define COM_SET_DEBUGLEVEL 11
#define COM_UKT_AFFINITY   12
#define COM_ERRIOR         13
#define COM_ERRIOW         14
#define COM_DEBUG_BREAK    15
#define COM_PROCESSOR_MASK 16


#define PARA_NONE         0x00
#define PARA_ENABLE       0x01
#define PARA_DISABLE      0x02

#define PARA_IO           0x01
#define PARA_STORAGE      0x02
#define PARA_TASKS        0x04
#define PARA_VERSIONS     0x08
#define PARA_REGIONS      0x10
#define PARA_STATE        0x20
#define PARA_RTE          0x40
#define PARA_TASK_STAT    0x100
#define PARA_TASK_COUNTS  0x200
#define PARA_QUEUE_STAT   0x400
#define PARA_TOTAL_COUNTS 0x800
#define PARA_ACTIVE       0x1000
#define PARA_UKT_SL_STAT  0x2000
#define PARA_SUSPENDS     0x4000
#define PARA_UKT_QUEUES   0x8000
#define PARA_RUNNABLE     0x10000
#define PARA_AIO          0x20000
#define PARA_THREAD_TIMES 0x40000
#define PARA_PSE36_STAT   0x80000
#define PARA_PSE36_FULLSTAT 0x100000

#define PARA_SET          0x01
#define PARA_CLEAR        0x02


#define TASKS_INFO_P_LINE 2


#ifdef  RUN_AS_SYSMON
 #define PARA_DEFAULT     PARA_IO  | PARA_TASKS | PARA_REGIONS |     \
                          PARA_RTE | PARA_STORAGE
 #define PARA_ALL          ( 0xFFFFFFFF - PARA_ACTIVE - PARA_RUNNABLE - PARA_PSE36_STAT)
#else
 #define PARA_DEFAULT     PARA_IO | PARA_TASKS | PARA_VERSIONS
 #define PARA_ALL          ( 0xFFFFFFFF - PARA_ACTIVE - PARA_RUNNABLE - PARA_PSE36_STAT)
#endif


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

#define STATE_TIME_AVAIL(state) ( state.ulSec || state.ulMicroSec )

#define TOT_TIME( State ) \
            ((float )State.ulSec + (float )State.ulMicroSec / 1000000.0 )

#define AVG_TIME( State ) State.ulCount == 0 ? 0 : \
            (((float )State.ulSec + (float )State.ulMicroSec / 1000000.0 ) / (float )State.ulCount)

#define AVG_ABS_TIME( State ) State.ulCount == 0 ? 0 : \
            (((float )State.ulSec + (float )State.ulMicroSec / 1000000.0 ) / (float )State.ulCount)

#define AVG_REL_TIME( State ) State.ulCount == 0 ? 0 : \
            (((float )State.ulRelSec + (float )State.ulRelMicroSec / 1000000.0 ) / (float )State.ulCount)

#define PRI_STAT_TM( _StrFormat, _Stat, _StrCount, _StrRelTm, _StrAbsTm ) \
      if ( _Stat.ulCount && _Stat.ulSec || _Stat.ulMicroSec)              \
        printf(_StrFormat, _StrCount, _Stat.ulCount,                      \
               _StrRelTm, AVG_REL_TIME( _Stat )  ,                        \
               _StrAbsTm, AVG_ABS_TIME( _Stat )  ) ;

#define PRI_PG_PER_IO(_StrFormat, _Stat, _StrPg, _StrPgPerIO )            \
      if ( _Stat.ulCount )                                                \
        printf ( _StrFormat, _StrPg, _Stat.ulPageCount,                   \
                 _StrPgPerIO,                                             \
                 (float ) _Stat.ulPageCount / (float ) _Stat.ulCount ) ;  \

#define OFFSET_OFF(_strct,_elem)   ((size_t)&(((_strct *)0)->_elem))


/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/

typedef struct xcons_record
  {
  BOOL                            fIsGateway;
  PSZ                             pszServerDB;
  PCONSOLE_REC                    pConsoleRec;
  HANDLE                          hConsoleRec;
  C40C                            RTEVersion;
  C40C                            KernelVersion;
  ULONG                           ulShmSize;
  LONG                            lCommand;
  LONG                            lParam;
  LONG                            lValue;
  PSZ                             pszParam;
  LONG                            lExtParam;
  LONG                            lInterval;
  LONG                            lRepeatCount;
  HANDLE                          hHandle;

  PID                             pidQueueOwner; // --- OS/2 only
  SECURITY_ATTRIBUTES             WorldSA;       // --- WIN32 only

  #if defined(_WIN32)
   IPC_BASE_OBJ_NAME              szShmName;
  #endif

  }  XCONS_REC;

typedef XCONS_REC                 *PXCONS_REC;

typedef struct thread_times
  {
    ULONG                         ulDays,
                                  ulHr,
                                  ulMin, 
                                  ulSec,
                                  ulMsec ;
  } XC_THREAD_TIMES ;



/*===========================================================================*
 *  GLOBAL VARIABLES                                                         *
 *===========================================================================*/


/*===========================================================================*
 *  LOCAL VARIABLES                                                          *
 *===========================================================================*/



/*===========================================================================*
 *  STATIC/INLINE FUNCTIONS (PROTOTYPES)                                     *
 *===========================================================================*/

LONG sql92c_get_num_param         ( PSZ                     pszParam,
                                    PLONG                   plNum  );
LONG sql92c_GetProcessorAffinityMask ( PSZ      pszParam,
                                       PLONG    ProcessAffinityMask  );
LONG sql92c_get_idx_param         ( PSZ                     pszParam,
                                    CHAR                    cPrefix,
                                    PLONG                   plNum  );
LONG sql92c_get_set_clear_param   ( PSZ                     pszParam );

LONG sql92c_get_time_param        ( PSZ                     pszParam );
LONG sql92c_get_show_res_param    ( PSZ                     pszParam,
                                    LONG                    lCommand );
LONG sql92c_init                  ( PXCONS_REC              pxcr );
VOID sql92c_release				  ( PXCONS_REC				pxcr );
LONG sql92c_check_accesss         ( PXCONS_REC              pxcr );
LONG sql92c_send_req              ( PXCONS_REC              pxcr );
LONG sql92c_send                  ( PXCONS_REC              pxcr,
                                    ULONG                   ulRequest,
                                    PBOOL                   pfFlag );
BOOL sql92c_wait                  ( PBOOL                   pfFlag);
PSZ  sql92c_kt_state_name         ( ULONG                   ulState );
PSZ  sql92c_ukt_state_name        ( ULONG                   ulState );
PSZ  sql92c_req_type_name         ( ULONG                   ulReqType );
LONG sql92c_show_version_info     ( PXCONS_REC              pxcr );
LONG sql92c_show_task_info        ( PCONSOLE_REC            pConsoleRec );
LONG sql92c_show_io_info          ( PCONSOLE_REC            pConsoleRec );
LONG sql92c_show_aio_info         ( PCONSOLE_REC            pConsoleRec );
LONG sql92c_show_region_info      ( PCONSOLE_REC            pConsoleRec );
LONG sql92c_show_storage_info     ( PCONSOLE_REC            pConsoleRec );
LONG sql92c_show_state_info       ( PCONSOLE_REC            pConsoleRec );
LONG sql92c_show_rte_info         ( PCONSOLE_REC            pConsoleRec );
LONG sql92c_show_queue_info       ( PCONSOLE_REC            pConsoleRec );
LONG sql92c_show_task_count_info  ( PCONSOLE_REC            pConsoleRec );
LONG sql92c_show_total_count_info ( PCONSOLE_REC            pConsoleRec );
LONG sql92c_show_ukt_sl_info      ( PCONSOLE_REC            pConsoleRec );
LONG sql92c_show_ukt_queues       ( PCONSOLE_REC            pConsoleRec );
LONG sql92c_show_UKT_queue        ( ULONG ulUKTQueueLen,
                                    PCONS_SINGLE_QUEUE_INFO_REC pUKTUKTQueue);
LONG sql92c_show_COM_queue        ( ULONG ulCOMQueueLen,
                                    PCONS_SINGLE_QUEUE_INFO_REC pUKTCOMQueue);
LONG sql92c_show_RUN_queue        ( ULONG ulCOMQueueLen,
                                    PCONS_SINGLE_QUEUE_INFO_REC pUKTCOMQueue);
LONG sql92c_show_suspends         ( PCONSOLE_REC            pConsoleRec );
LONG sql92c_show_thread_times     ( PCONSOLE_REC            pConsoleRec );
LONG sql92c_show_PSE36_stat       ( PCONSOLE_REC            pConsoleRec );
LONG sql92c_show_PSE36PageArray   ( PCONSOLE_REC            pConsoleRec );
LONG sql92c_get_com_param         ( PSZ                     pszParam );

LONG sql92c_show_param            ( PSZ                     pszParam,
                                    PCONSOLE_REC            pConsoleRec );


/*===========================================================================*
 *  PUBLIC FUNCTIONS (CODE)                                                  *
 *===========================================================================*/

#if !defined (CTRLCOMP)
INT _CDECL  main (INT argc, CHAR *argv[])
  {
  #undef  MF__
  #define MF__ MOD__"main"
  XCONS_REC               xcr;
  INT                     ArgN          = 1;
  APIRET                  rc            = NO_ERROR;
  LONG                    lCommand;
  LONG                    lParam;


  DBGIN;

  memset ( &xcr, 0, sizeof(xcr) );
  xcr.lInterval     = 0;
  xcr.lRepeatCount  = 1;
  xcr.lCommand      = COM_UNDEF;
  xcr.lParam        = 0;
  xcr.lExtParam     = 0;

  if ( ARGS(argc, ArgN) )
    {
    if ((( argv [ ArgN ] [ 0 ] == '-' ) || ( argv [ ArgN ] [ 0 ] == '/' )) &&
        (( argv [ ArgN ] [ 1 ] == '?' ) || ( argv [ ArgN ] [ 1 ] == 'h' )) &&
        (  argv [ ArgN ] [ 2 ] == '\0' ))
      {
      xcr.lCommand = COM_HELP;
      }
    else
      {
      //
      // --- get command and server database name
      //
      xcr.lCommand = sql92c_get_com_param ( argv[ArgN] );

      if ( xcr.lCommand == COM_UNDEF )
        {
         xcr.pszServerDB = argv [ArgN++];

         if ( ARGS(argc, ArgN) )
           {
           xcr.lCommand = sql92c_get_com_param ( argv[ArgN++] );

           if ( xcr.lCommand == COM_UNDEF )
             rc = ! NO_ERROR;
           }
         else
           rc = ! NO_ERROR;
        }
      else if ( ARGS(argc, ++ArgN) )
        {
        lCommand = sql92c_get_com_param ( argv[ArgN] );

        if ( lCommand != COM_UNDEF ) // - is there another command?
          {
          #ifdef  RUN_AS_SYSMON
           rc = ! NO_ERROR;
          #else
           xcr.pszServerDB = argv [ArgN - 1];
           ArgN++;
          #endif
          }
        }

      if (( rc == NO_ERROR ) && ( xcr.lCommand != COM_HELP ))
        {
        switch ( xcr.lCommand )
          {
          case COM_UKT_AFFINITY:
            if ( ARGS(argc, ArgN) )
              rc = sql92c_get_idx_param ( argv[ArgN++], 'u', &xcr.lParam );
            else
              rc = ! NO_ERROR;
            break;

          // PTS 1107070
          case COM_PROCESSOR_MASK:
            if ( ARGS(argc, ArgN) )
			{
              rc = sql92c_GetProcessorAffinityMask ( argv[ArgN++], &xcr.lParam );

              if ( rc == NO_ERROR && ARGS(argc, ArgN) )
    			  rc = e92_get_tt_to_show (argv[ArgN], &xcr.lExtParam );
			}
            else
              rc = ! NO_ERROR;
            break;

          case COM_CANCEL:
            if ( ARGS(argc, ArgN) )
              rc = sql92c_get_idx_param ( argv[ArgN++], 't', &xcr.lParam );
            else
              rc = ! NO_ERROR;
            break;
            
          case COM_KILL  :
            if ( ARGS(argc, ArgN) )
              rc = sql92c_get_idx_param ( argv[ArgN++], 't', &xcr.lParam );
            else
              rc = ! NO_ERROR;
            break;

          case COM_DROP  :

            if ( !ARGS(argc, ArgN) ||
                 !e92_strcont( XCONS_CMD_TABLES, strupr(argv[ArgN++]),
                                  XCONS_MIN_CMD_LEN_TABLES ))
              {
              rc = ! NO_ERROR;
              }
            break;

          case COM_CRASH_DUMP:
            if ( !ARGS(argc, ArgN) ||
                 !e92_strcont( XCONS_CMD_DUMP, strupr(argv[ArgN++]),
                                  XCONS_MIN_CMD_LEN_DUMP ))
              {
              rc = ! NO_ERROR;
              }
            break;

          case COM_TIME_MEASURE :
            if ( ARGS(argc, ArgN) )
              {
              lParam = sql92c_get_time_param ( argv[ArgN++] );

              if ( lParam != PARA_NONE )
                xcr.lParam |= lParam;
              else
                rc = ! NO_ERROR;
              }
            else
              rc = ! NO_ERROR;
            break;

          case COM_RESET :
            if ( ARGS(argc, ArgN) )
              {
              lParam = sql92c_get_show_res_param( argv[ArgN++], xcr.lCommand );

              if ( lParam != PARA_NONE )
                xcr.lParam |= lParam;
              else
                rc = ! NO_ERROR;
              }
            else
              rc = ! NO_ERROR;
            break;

          case COM_ERRIOR :
          case COM_ERRIOW :
            if ( ARGS(argc, ArgN) )
              rc = sql92c_get_num_param ( argv[ArgN++], &xcr.lParam );
            else
              rc = ! NO_ERROR;
            break;

          case COM_SET_DEBUGLEVEL :
            if ( ARGS(argc, ArgN) )
              rc = sql92c_get_num_param ( argv[ArgN++], &xcr.lParam );
            else
              rc = ! NO_ERROR;
            break;

          case COM_GETPARAM :

            if ( ARGS(argc, ArgN) )
              xcr.pszParam = argv[ArgN++];
            else
              xcr.pszParam = "";

            break;

          case COM_PUTPARAM :

          case COM_SHOW_UPDATE :

            // --- command parameter found?
            if (( ARGS(argc, ArgN)) && ( !isdigit(argv[ArgN][0]) ))
              {
              xcr.lParam = 0;

              do
                {
                lParam = sql92c_get_show_res_param(argv[ArgN++], xcr.lCommand);

                if ( lParam != PARA_NONE )
                  {
                  BOOL fExtParamOk = FALSE ;
                  if (( lParam == PARA_TASK_COUNTS ) &&
                      ( ARGS(argc, ArgN)))
                    {
                    rc = sql92c_get_idx_param(argv[ArgN], 't', &xcr.lExtParam);

                    if ( ! (fExtParamOk = rc == NO_ERROR ) )
                      { rc = e92_get_tt_to_show (argv[ArgN], &xcr.lExtParam );
                        fExtParamOk = rc == NO_ERROR;
                      }
                    else
                      xcr.lExtParam = -xcr.lExtParam ;
                    }
                  else if (lParam == PARA_TASKS || lParam == PARA_ACTIVE || lParam == PARA_RUNNABLE )
                    {
                    xcr.lExtParam = SHOW_TT_ALL_EO92;
                    if ( ARGS(argc, ArgN) )
                      {
                      rc = e92_get_tt_to_show (argv[ArgN], &xcr.lExtParam );
                      fExtParamOk = rc == NO_ERROR;
                      }
                    }

                  xcr.lParam |= lParam;
                  if ( !fExtParamOk )
                    rc = NO_ERROR;
                  else
                    ArgN++;
                  }
                else
                  rc = ! NO_ERROR;
                }
              while (( rc == NO_ERROR )      &&
                     ( ARGS(argc, ArgN))     &&
                     ( !isdigit(argv[ArgN][0]) ));
              }
            else
              xcr.lParam = PARA_DEFAULT;

            // --- try to get interval and repeat count
            if (( rc == NO_ERROR ) && ARGS(argc, ArgN))
              {
              xcr.lRepeatCount = -1;

              // --- try to get interval
              rc = sql92c_get_num_param (argv[ArgN++], &xcr.lInterval);

              if ( rc == NO_ERROR )
                {
                if ( xcr.lInterval < 0 )
                  rc = ! NO_ERROR;
                else if ( ARGS(argc, ArgN))
                  {
                  // --- try to get repeat count
                  rc = sql92c_get_num_param (argv[ArgN++], &xcr.lRepeatCount);

                  if ( xcr.lRepeatCount < 1 )
                    rc = ! NO_ERROR;
                  }
                }
              }
            break;

          case COM_DEBUG_BREAK:

            if ( ARGS(argc, ArgN) )
              {
              xcr.lParam = sql92c_get_set_clear_param ( argv[ArgN++] );

              if ( xcr.lParam == PARA_SET )
                {
                if ( ARGS(argc, ArgN) )
                  rc = sql92c_get_num_param (argv[ArgN++], &xcr.lValue);
                else
                  rc = ! NO_ERROR;
                }
              else if ( xcr.lParam != PARA_CLEAR )
                rc = ! NO_ERROR;
              }
            else
              rc = ! NO_ERROR;

            break;

          default :
            rc = ! NO_ERROR;
          }
        ArgN++;
        }
      }
    }
  else
    rc = ! NO_ERROR;

  //
  // --- display help info ?
  //
  if (( ARGS( argc, ArgN ) )     ||
      ( rc != NO_ERROR  )        ||
      ( xcr.lCommand == COM_HELP   ) ||
      ( xcr.lCommand == COM_UNDEF  ))
    {
    #ifdef  RUN_AS_SYSMON
     printf  (XCONS_USAGE_SYSMON);
    #else
     printf  (XCONS_USAGE);
    #endif

    DBGOUT;
    exit(rc);
    }

  #ifdef  RUN_AS_SYSMON
   rc = eo921SysmonInit ( &xcr.pszServerDB );

   if ( rc != NO_ERROR )
     {
     DBGOUT;
     exit(rc);
     }
  #else
   if ( xcr.pszServerDB == NULL)
     {
     rc = GETENV(SERVERDB_ENV_VAR, &xcr.pszServerDB);
     if (rc != NO_ERROR  ||  *xcr.pszServerDB == '\0')
       {
       rc = GETENV(DBNAME_ENV_VAR, &xcr.pszServerDB);
       if (rc != NO_ERROR  ||  *xcr.pszServerDB == '\0')
         {
          printf  (XCONS_USAGE);

          DBGOUT;
          exit(1);
         }
       }
     }
  #endif


  if ( (rc = sql92c_init ( &xcr )) == NO_ERROR )
    {
    rc = sql92c_send_req ( &xcr );

	sql92c_release( &xcr );
    }

  #ifdef  RUN_AS_SYSMON
  eo921SysmonStop();
  #endif

  DBGOUT;
  exit(rc);

  return(rc);
  }
#endif
/*------------------------------*/

LONG sql92c_init ( PXCONS_REC  pxcr )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_init"
  LONG                    rc       = NO_ERROR;
  HANDLE                  hKSS     = 0;
  KERNEL_SHARED_SEG       *pKSS    = NULL;
  PSZ                     pszPath  = NULL;
  PSZ                     pszName  = NULL;
  tsp00_Versionc          RTEVersionString;
  tsp100_VersionResult    Result;

  ULONG                  ulRetry  = 10;
  ACCESS_RIGHTS_REC      Access;

  DBGIN;


  //
  // --- allocate and initialize security attributes
  //
  pxcr->WorldSA.nLength             = sizeof(pxcr->WorldSA);
  pxcr->WorldSA.bInheritHandle      = FALSE;
  memset ( &Access, 0, sizeof (Access) );   // - this must be done first!

  Access.World.ulAccess = SPECIFIC_RIGHTS_ALL |
                          STANDARD_RIGHTS_ALL;
  rc = sql49c_alloc_and_init_SD( &Access, &pxcr->WorldSA.lpSecurityDescriptor );

  if ( rc != NO_ERROR )
    {
    DBGOUT;
    return ( rc );
    }
  
  //
  // --- open the request queue/mailslot
  //

  strupr ( pxcr->pszServerDB );
  pszPath = MAIL_CONS;
  pszName = pxcr->pszServerDB;

  DBG3 ((MF__, "sql41c_open_queue_mailslot: %s", pszName ));

  rc = sql41c_open_queue_mailslot ( &pxcr->pidQueueOwner,       // - OS/2 only
                                    &pxcr->hHandle,
                                    pszPath, pszName );

 if ( rc == ERROR_FILE_NOT_FOUND )
   {
   FREE_MEM ( pxcr->WorldSA.lpSecurityDescriptor );
   MSGCD ((ERR_DB_OR_GW_NOT_STARTED, pxcr->pszServerDB));
   DBGOUT;
   return ( rc );
   }
 else if ( rc != NO_ERROR )
   {
   FREE_MEM ( pxcr->WorldSA.lpSecurityDescriptor );
   DBGOUT;
   return ( rc );
   }

  //
  // --- check database version
  //
  rc = sql41c_get_named_shrd_mem ( (PPVOID)&pKSS, SHM_KSS, pxcr->pszServerDB,
                                   &hKSS );

  if ( rc == ERROR_INVALID_NAME )
    {
    FREE_MEM ( pxcr->WorldSA.lpSecurityDescriptor );
    sql41c_close_queue_mailslot ( pxcr->hHandle );

    MSGCD ((ERR_DB_OR_GW_NOT_STARTED, pxcr->pszServerDB));
    DBGOUT;
    return ( rc );
    }
  else if (rc != NO_ERROR)
    {
    FREE_MEM ( pxcr->WorldSA.lpSecurityDescriptor );
    sql41c_close_queue_mailslot ( pxcr->hHandle );
    DBGOUT;
    return ( rc );
    }

  // --- set the RTE version string
  Result = sp100_CompareVersionIDs (&pKSS->RTEVersionID);


  if ( Result != IsEqual_esp100)
    {
    FREE_MEM ( pxcr->WorldSA.lpSecurityDescriptor );
    sql41c_close_queue_mailslot ( pxcr->hHandle );
    sql41c_free_shrd_mem ( pKSS, hKSS );

    // --- set the RTE version string
    if ( sql02_get_platform_id() == VER_PLATFORM_WIN32_NT )
      sp100_GetVersionString(COMP_NAME_RTE_NT_SP100, s100buildnumber, RTEVersionString);
    else
      sp100_GetVersionString(COMP_NAME_RTE_WIN9X_SP100, s100buildnumber, RTEVersionString);

    if ( pKSS->fIsGateway )
      MSGCD ((ERR_WRONG_GW_VERSION))
    else
      MSGCD ((ERR_WRONG_DB_VERSION));

    printf (XCONS_WRONG_VERS_KERNEL, pKSS->RTEVersion);
    printf (XCONS_WRONG_VERS_XCONS , RTEVersionString);
    DBGOUT;
    return ( 1 );
    }

  strcpy ( pxcr->RTEVersion,    pKSS->RTEVersion );
  strcpy ( pxcr->KernelVersion, pKSS->KernelVersion );

  pxcr->ulShmSize  = pKSS->ulConsoleShmSize;
  pxcr->fIsGateway = pKSS->fIsGateway;

  rc = sql41c_free_shrd_mem ( pKSS, hKSS );

  if ( rc != NO_ERROR )
    {
    FREE_MEM ( pxcr->WorldSA.lpSecurityDescriptor );
    sql41c_close_queue_mailslot ( pxcr->hHandle );

    DBGOUT;
    return ( rc );
    }

  //
  // --- allocate a shared memory segment
  //
  do
    {
    sql41c_build_unique_obj_name ( pxcr->szShmName );

    rc = sql41c_create_shrd_mem ( (PVOID*) &pxcr->pConsoleRec,
                                   pxcr->ulShmSize, SHM_CONSREC,
                                   pxcr->szShmName,
                                   &pxcr->WorldSA, &pxcr->hConsoleRec );

    // --- unique name already in use?
    if ( rc == ERROR_DUPLICATE_NAME )
      SLEEP ( 1 );
    }
  while (( rc == ERROR_DUPLICATE_NAME ) && (--ulRetry) );

  if ( rc != NO_ERROR )
    {
    FREE_MEM ( pxcr->WorldSA.lpSecurityDescriptor );
    sql41c_free_shrd_mem (pxcr->pConsoleRec, pxcr->hConsoleRec);
    sql41c_close_queue_mailslot ( pxcr->hHandle );
    DBGOUT;
    return ( rc );
    }

  DBGOUT;
  return ( rc );
  }

/*------------------------------*/

VOID sql92c_release ( PXCONS_REC  pxcr )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_release"

  DBGPAS;

  FREE_MEM ( pxcr->WorldSA.lpSecurityDescriptor );
  pxcr->WorldSA.lpSecurityDescriptor = NULL;
  sql41c_free_shrd_mem (pxcr->pConsoleRec, pxcr->hConsoleRec);
  sql41c_close_queue_mailslot ( pxcr->hHandle );
  
  return;
  }

/*------------------------------*/

LONG sql92c_send ( PXCONS_REC     pxcr,
                          ULONG          ulRequest,
                          PBOOL          pfFlag )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_send"
  APIRET            rc = NO_ERROR;
  CONS_REQUEST_REC  Request;

  DBGIN;

  #if !defined(_WIN32)
   rc = sql41c_give_shrd_mem (pxcr->pConsoleRec, pxcr->pidQueueOwner, 0, NULL);

   if ( rc != NO_ERROR )
     {
     DBGOUT;
     return ( rc );
     }
  #endif

  #if defined(_WIN32)
   strcpy ( Request.szClientObjName, pxcr->szShmName );
   Request.ulReqType       = ulRequest;
   Request.ulConsoleRecLen = pxcr->ulShmSize;
  #else
   Request                 = ulRequest;
  #endif

  rc = sql41c_write_queue_mailslot ( pxcr->hHandle,
                                     &Request,
                                     sizeof (CONS_REQUEST_REC),
                                     pxcr->ulShmSize,           // - OS/2 only
                                     pxcr->pConsoleRec );       // - OS/2 only

  if ( rc != NO_ERROR )
    {
    if ( rc == ERROR_HANDLE_EOF )
      MSGCD (( ERR_DB_OR_GW_NOT_STARTED, pxcr->pszServerDB ));
    DBGOUT;
    return ( rc );
    }

  if (!sql92c_wait (pfFlag))
    {
    DBGOUT;
    return ( 1 );
    }

  DBGOUT;
  return ( rc );
  }


/*===========================================================================*
 *  LOCAL/PRIVATE/PROTECTED FUNCTIONS (CODE)                                 *
 *===========================================================================*/


static LONG sql92c_send_req ( PXCONS_REC              pxcr )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_send_req"
  APIRET                  rc = NO_ERROR;

  DBGIN;


  switch ( pxcr->lCommand )
    {
    case COM_UKT_AFFINITY :

      if ( sql92c_check_accesss ( pxcr ) != NO_ERROR )
        return ( 1 );

      memset ( pxcr->pConsoleRec, 0, pxcr->ulShmSize );

      pxcr->pConsoleRec->UKTAffinity.ulUKTIndex = pxcr->lParam;

      rc = sql92c_send (pxcr, CON_REQ_SET_UKT_AFFINITY,
                       &pxcr->pConsoleRec->UKTAffinity.fReturnCodeAvailable);

      if ( rc != NO_ERROR ) return ( rc );

      switch ( pxcr->pConsoleRec->UKTAffinity.usReturnCode )
        {
        case CON_NO_ERROR        :
          break;
        case CON_WRONG_UKT :
          MSGCD (( ERR_XCONS_WRONG_UKT ));
          return ( 1 );
        case CON_ERROR          :
          MSGCD (( ERR_XCONS_SET_UKT_AFFINITY ));
          return ( 1 );
        case CON_NO_PROCESSOR   :
          MSGCD (( ERR_XCONS_AFFIN_NO_PROCESSOR ));
          return ( 1 );
        default:
          MSGCD (( ERR_UNKNOWN_RET_CODE ));
          return ( 1 );
        }

      break;

    case COM_PROCESSOR_MASK :

      if ( sql92c_check_accesss ( pxcr ) != NO_ERROR )
        return ( 1 );

      memset ( pxcr->pConsoleRec, 0, pxcr->ulShmSize );

	  pxcr->pConsoleRec->SetProcessorMask.ProcessAffinityMask = pxcr->lParam;
	  pxcr->pConsoleRec->SetProcessorMask.TaskType            = (UINT1 )pxcr->lExtParam;

      rc = sql92c_send (pxcr, CON_REQ_SET_PROCESSOR_MASK,
                       &pxcr->pConsoleRec->SetProcessorMask.fReturnCodeAvailable);

      if ( rc != NO_ERROR ) return ( rc );

      switch ( pxcr->pConsoleRec->SetProcessorMask.usReturnCode )
        {
        case CON_NO_ERROR        :
          break;
        case CON_ERROR          :
          MSGCD (( ERR_XCONS_SET_UKT_AFFINITY ));
          return ( 1 );
        default:
          MSGCD (( ERR_UNKNOWN_RET_CODE ));
          return ( 1 );
        }
      break;

    case COM_DEBUG_BREAK:

      if ( sql92c_check_accesss ( pxcr ) != NO_ERROR )
        return ( 1 );

      memset ( pxcr->pConsoleRec, 0, pxcr->ulShmSize );

      pxcr->pConsoleRec->DebugBreak.ulBreakPosition = pxcr->lValue;

      if ( pxcr->lParam == PARA_CLEAR )
        pxcr->pConsoleRec->DebugBreak.fClearPosition = TRUE;
      else
        pxcr->pConsoleRec->DebugBreak.fClearPosition = FALSE;

      rc = sql92c_send (pxcr, CON_REQ_DEBUG_BREAK,
                        &pxcr->pConsoleRec->DebugBreak.fReturnCodeAvailable);

      if ( rc != NO_ERROR ) return ( rc );

      switch ( pxcr->pConsoleRec->DebugBreak.usReturnCode )
        {
        case CON_NO_ERROR               :
          break;
        case CON_NO_INTERACT_WITH_DECTOP:
          MSGCD ((ERR_XCONS_INTERACT_WTH_DECTOP ));
          return ( 1 );
        default:
          MSGCD (( ERR_UNKNOWN_RET_CODE ));
          return ( 1 );
        }

      break;


    case COM_CANCEL:

      if ( sql92c_check_accesss ( pxcr ) != NO_ERROR )
        return ( 1 );

      memset ( pxcr->pConsoleRec, 0, pxcr->ulShmSize );

      pxcr->pConsoleRec->CancelTaskCtrl.ulTask = pxcr->lParam;

      rc = sql92c_send (pxcr, CON_REQ_CANCEL,
                       &pxcr->pConsoleRec->CancelTaskCtrl.fReturnCodeAvailable);

      if ( rc != NO_ERROR ) return ( rc );

      switch ( pxcr->pConsoleRec->CancelTaskCtrl.usReturnCode )
        {
        case CON_NO_ERROR       :
          break;
        case CON_WRONG_TASK     :
          MSGCD ((ERR_XCONS_WRONG_TASK ));
          return ( 1 );
        case CON_WRONG_TASK_TYPE:
          MSGCD ((ERR_XCONS_WRONG_TASK_TYPE ));
          return ( 1 );
        default:
          MSGCD (( ERR_UNKNOWN_RET_CODE ));
          return ( 1 );
        }

      break;

    case COM_KILL:

      if ( sql92c_check_accesss ( pxcr ) != NO_ERROR )
        return ( 1 );

      memset ( pxcr->pConsoleRec, 0, pxcr->ulShmSize );

      pxcr->pConsoleRec->KillTaskCtrl.ulTask = pxcr->lParam;

      rc = sql92c_send (pxcr, CON_REQ_KILL,
                        &pxcr->pConsoleRec->KillTaskCtrl.fReturnCodeAvailable);

      if ( rc != NO_ERROR ) return ( rc );

      switch ( pxcr->pConsoleRec->KillTaskCtrl.usReturnCode )
        {
        case CON_NO_ERROR       :
          break;
        case CON_WRONG_TASK     :
          MSGCD ((ERR_XCONS_WRONG_TASK ));
          return ( 1 );
        case CON_WRONG_TASK_TYPE:
          MSGCD ((ERR_XCONS_WRONG_TASK_TYPE ));
          return ( 1 );
        default:
          MSGCD (( ERR_UNKNOWN_RET_CODE ));
          return ( 1 );
        }

      break;

    case COM_CRASH_DUMP:

      if ( sql92c_check_accesss ( pxcr ) != NO_ERROR )
        return (1);

      memset ( pxcr->pConsoleRec, 0, pxcr->ulShmSize );

      rc = sql92c_send (pxcr, CON_REQ_CRASH_DUMP,
                       &pxcr->pConsoleRec->CrashDump.fReturnCodeAvailable);

      if ( rc != NO_ERROR ) return ( rc );

      switch ( pxcr->pConsoleRec->CrashDump.usReturnCode )
        {
        case CON_NO_ERROR :
          break;

        default:
          MSGCD (( ERR_UNKNOWN_RET_CODE ));
          return ( 1 );
        }

      break;

    case COM_DROP:

      #ifdef RUN_AS_SYSMON
        rc = eo921SysmonDropTab ();
      #endif

      break;

    #ifndef RUN_AS_SYSMON

     case COM_GETPARAM:

       if ( sql92c_check_accesss ( pxcr ) != NO_ERROR )
         return ( 1 );

       memset ( pxcr->pConsoleRec, 0, pxcr->ulShmSize );

       rc = sql92c_send (pxcr, CON_REQ_GETPARAM ,
                         &pxcr->pConsoleRec->GetParamInfo.fGetParamAvailable );

       rc = sql92c_show_param ( pxcr->pszParam, pxcr->pConsoleRec );

       if ( rc != NO_ERROR )
         return ( rc );

       break;


     case COM_PUTPARAM      :

       if ( sql92c_check_accesss ( pxcr ) != NO_ERROR )
         return ( 1 );

       rc = sql92c_send (pxcr, CON_REQ_PUTPARAM ,
                         &pxcr->pConsoleRec->PutParamInfo.fPutParamAvailable );

       break;

    #endif
     case COM_ERRIOR:
     case COM_ERRIOW:

       if ( sql92c_check_accesss ( pxcr ) != NO_ERROR )
         return ( 1 );

       pxcr->pConsoleRec->ForceIOError.ulDevno    = pxcr->lParam ;
       pxcr->pConsoleRec->ForceIOError.fErrOnRead = pxcr->lCommand == COM_ERRIOR ;

       rc = sql92c_send (pxcr, CON_REQ_FORCE_IO_ERROR ,
                         &pxcr->pConsoleRec->ForceIOError.fForceIOErrorAvailable);
          break ;

     case COM_SET_DEBUGLEVEL:

       if ( sql92c_check_accesss ( pxcr ) != NO_ERROR )
         return ( 1 );

       pxcr->pConsoleRec->SetDebugLevelCtrl.ulDebugLevel = pxcr->lParam ;

       rc = sql92c_send (pxcr, CON_REQ_SET_DEBUGLEVEL ,
                         &pxcr->pConsoleRec->SetDebugLevelCtrl.fSetDebugLevelAvailable);

       break;

    case COM_TIME_MEASURE  :

      if ( pxcr->lParam & PARA_ENABLE )
          {
          if ( sql92c_check_accesss ( pxcr ) != NO_ERROR )
            return ( 1 );
          rc = sql92c_send (pxcr, CON_REQ_ENABLE_TIME_MEASURE ,
                  &pxcr->pConsoleRec->TimeMeasureCtrl.fTimeMeasureAvailable );
          }

      if ( pxcr->lParam & PARA_DISABLE )
          {
          if ( sql92c_check_accesss ( pxcr ) != NO_ERROR )
            return ( 1 );
          rc = sql92c_send (pxcr, CON_REQ_DISABLE_TIME_MEASURE ,
                   &pxcr->pConsoleRec->TimeMeasureCtrl.fTimeMeasureAvailable );
          }
      break;

    case COM_RESET:
        if ( sql92c_check_accesss ( pxcr ) != NO_ERROR )
          return ( 1 );

        if ( pxcr->lParam & PARA_IO )
            rc = sql92c_send (pxcr, CON_REQ_RES_ALL ,
                    &pxcr->pConsoleRec->ResetCtrl.fResetAvailable );
        else
        if ( pxcr->lParam & PARA_IO )
            rc = sql92c_send (pxcr, CON_REQ_RES_IO ,
                    &pxcr->pConsoleRec->ResetCtrl.fResetAvailable );
        else
        if ( pxcr->lParam & PARA_REGIONS )
             rc = sql92c_send (pxcr, CON_REQ_RES_REGION ,
                    &pxcr->pConsoleRec->ResetCtrl.fResetAvailable );
        else
        if ( pxcr->lParam & PARA_QUEUE_STAT )
            rc = sql92c_send (pxcr, CON_REQ_RES_QUEUE_STAT ,
                    &pxcr->pConsoleRec->ResetCtrl.fResetAvailable );
        else
        if ( pxcr->lParam & PARA_TASK_COUNTS )
            rc = sql92c_send (pxcr, CON_REQ_RES_TASK_COUNTS ,
                    &pxcr->pConsoleRec->ResetCtrl.fResetAvailable );
        else
        if ( pxcr->lParam & PARA_UKT_SL_STAT )
            rc = sql92c_send (pxcr, CON_REQ_RES_UKT_SLEEPTIME_STAT,
                    &pxcr->pConsoleRec->ResetCtrl.fResetAvailable );
        else
        if ( pxcr->lParam & PARA_SUSPENDS    )
            rc = sql92c_send (pxcr, CON_REQ_RES_SUSPENDS,
                    &pxcr->pConsoleRec->ResetCtrl.fResetAvailable );
      break;

    case COM_SHOW_UPDATE:

      #ifndef RUN_AS_SYSMON
       if ( pxcr->fIsGateway )
         printf ( XCONS_GATEWAY, pxcr->pszServerDB );
       else
         printf ( XCONS_SERVERDB, pxcr->pszServerDB );
      #endif

      while ( pxcr->lRepeatCount )
        {
        if ( pxcr->lParam & PARA_VERSIONS )
          {
          rc = sql92c_show_version_info ( pxcr );

          if ( rc != NO_ERROR ) return ( rc );
          }

        if ( pxcr->lParam & PARA_IO )
          {
          memset ( pxcr->pConsoleRec, 0, pxcr->ulShmSize );

          rc = sql92c_send (pxcr, CON_REQ_IO,
                            &pxcr->pConsoleRec->IOInfo.fInfoAvailable);

          if ( rc != NO_ERROR ) return ( rc );

          rc = sql92c_show_io_info ( pxcr->pConsoleRec );

          if ( rc != NO_ERROR ) return ( rc );
          }

        if ( pxcr->lParam & PARA_AIO )
          {
          memset ( pxcr->pConsoleRec, 0, pxcr->ulShmSize );

          rc = sql92c_send (pxcr, CON_REQ_AIO,
                            &pxcr->pConsoleRec->AIOInfo.fInfoAvailable);

          if ( rc != NO_ERROR ) return ( rc );

          rc = sql92c_show_aio_info ( pxcr->pConsoleRec );

          if ( rc != NO_ERROR ) return ( rc );
          }

        if ( pxcr->lParam & PARA_REGIONS )
          {
          memset ( pxcr->pConsoleRec, 0, pxcr->ulShmSize );

          rc = sql92c_send (pxcr, CON_REQ_REGION,
                            &pxcr->pConsoleRec->SemInfo.fInfoAvailable);

          if ( rc != NO_ERROR ) return ( rc );

          rc = sql92c_show_region_info ( pxcr->pConsoleRec );

          if ( rc != NO_ERROR ) return ( rc );
          }

        if ( pxcr->lParam & PARA_RTE )
          {
          memset ( pxcr->pConsoleRec, 0, pxcr->ulShmSize );

          rc = sql92c_send (pxcr, CON_REQ_RTE,
                            &pxcr->pConsoleRec->RTEInfo.fInfoAvailable);

          if ( rc != NO_ERROR ) return ( rc );

          rc = sql92c_show_rte_info ( pxcr->pConsoleRec );

          if ( rc != NO_ERROR ) return ( rc );
          }

        if ( pxcr->lParam & PARA_STORAGE )
          {
          memset ( pxcr->pConsoleRec, 0, pxcr->ulShmSize );

          rc = sql92c_send (pxcr, CON_REQ_STORAGE,
                            &pxcr->pConsoleRec->StorageInfo.fInfoAvailable);

          if ( rc != NO_ERROR ) return ( rc );

          rc = sql92c_show_storage_info ( pxcr->pConsoleRec );

          if ( rc != NO_ERROR ) return ( rc );
          }

        if ( pxcr->lParam & PARA_TASKS  ||
             pxcr->lParam & PARA_ACTIVE ||
             pxcr->lParam & PARA_RUNNABLE)
          {
          memset ( pxcr->pConsoleRec, 0, pxcr->ulShmSize );

          pxcr->pConsoleRec->TaskInfo.StateToShow =
                pxcr->lParam & PARA_TASKS ? SHOW_ALL_EO92 :
                pxcr->lParam & PARA_ACTIVE? SHOW_ACTIVE_EO92 : SHOW_RUNNABLE_EO92 ;

          /* PTS 1106166 */
          pxcr->pConsoleRec->TaskInfo.TypeToShow  = 
                pxcr->lParam == PARA_ALL ? SHOW_TT_ALL_EO92 : (UINT1 )pxcr->lExtParam;

          rc = sql92c_send (pxcr, CON_REQ_TASK,
                            &pxcr->pConsoleRec->TaskInfo.fInfoAvailable);

          if ( rc != NO_ERROR ) return ( rc );

          rc = sql92c_show_task_info ( pxcr->pConsoleRec );

          if ( rc != NO_ERROR ) return ( rc );
          }

        if ( pxcr->lParam & PARA_UKT_SL_STAT )
          {
          memset ( pxcr->pConsoleRec, 0, pxcr->ulShmSize );

          rc = sql92c_send (pxcr, CON_REQ_UKT_SLEEPTIME_STAT ,
                          &pxcr->pConsoleRec->UKTSleepTimeInfo.fInfoAvailable);

          if ( rc != NO_ERROR ) return ( rc );

          rc = sql92c_show_ukt_sl_info ( pxcr->pConsoleRec );

          if ( rc != NO_ERROR ) return ( rc );
          }

        #ifndef RUN_AS_SYSMON

        if ( pxcr->lParam & PARA_THREAD_TIMES )
          {
          memset ( pxcr->pConsoleRec, 0, pxcr->ulShmSize );

          rc = sql92c_send (pxcr, CON_REQ_THREAD_TIMES,
                            &pxcr->pConsoleRec->ThreadTimeInfo.fThreadTimesAvailable);

          if ( rc != NO_ERROR ) return ( rc );

          rc = sql92c_show_thread_times ( pxcr->pConsoleRec );

          if ( rc != NO_ERROR ) return ( rc );
          }

        if ( pxcr->lParam & PARA_PSE36_STAT || pxcr->lParam & PARA_PSE36_FULLSTAT )
          {
          memset ( pxcr->pConsoleRec, 0, pxcr->ulShmSize );

          rc = sql92c_send (pxcr, 
                            pxcr->lParam & PARA_PSE36_STAT ? 
                            CON_REQ_PSE36_STAT : CON_REQ_PSE36_DATA_PAGE_ARRAY,
                            &pxcr->pConsoleRec->PSE36Info.fPSE36StatAvailable);

          if ( rc != NO_ERROR ) return ( rc );

          rc = sql92c_show_PSE36_stat ( pxcr->pConsoleRec );

          if ( rc != NO_ERROR ) return ( rc );
          }

        if ( pxcr->lParam & PARA_SUSPENDS )
          {
          memset ( pxcr->pConsoleRec, 0, pxcr->ulShmSize );

          rc = sql92c_send (pxcr, CON_REQ_SHOW_SUSPENDS,
                         &pxcr->pConsoleRec->SuspendReasons.fSuspendsAvailable);

          if ( rc != NO_ERROR ) return ( rc );

          rc = sql92c_show_suspends ( pxcr->pConsoleRec );

          if ( rc != NO_ERROR ) return ( rc );
          }

        if ( pxcr->lParam & PARA_UKT_QUEUES )
          {
          memset ( pxcr->pConsoleRec, 0, pxcr->ulShmSize );

          rc = sql92c_send (pxcr, CON_REQ_UKT_QUEUES ,
                          &pxcr->pConsoleRec->UKTQueues.fInfoAvailable);

          if ( rc != NO_ERROR ) return ( rc );

          rc = sql92c_show_ukt_queues ( pxcr->pConsoleRec );

          if ( rc != NO_ERROR ) return ( rc );
          }

         if ( pxcr->lParam & PARA_STATE )
           {
           memset ( pxcr->pConsoleRec, 0, pxcr->ulShmSize );

           rc = sql92c_send ( pxcr, CON_REQ_STATE,
                              &pxcr->pConsoleRec->StateInfo.fInfoAvailable);

           if ( rc != NO_ERROR ) return ( rc );

           rc = sql92c_show_state_info ( pxcr->pConsoleRec );

           if ( rc != NO_ERROR ) return ( rc );
           }

        if ( pxcr->lParam & PARA_QUEUE_STAT )
          {
          memset ( pxcr->pConsoleRec, 0, pxcr->ulShmSize );

          rc = sql92c_send (pxcr, CON_REQ_QUEUE_STAT,
                            &pxcr->pConsoleRec->QueInfo.fInfoAvailable);

          if ( rc != NO_ERROR ) return ( rc );

          rc = sql92c_show_queue_info ( pxcr->pConsoleRec );

          if ( rc != NO_ERROR ) return ( rc );
          }

        if ( pxcr->lParam & PARA_TASK_COUNTS )
          {
          memset ( pxcr->pConsoleRec, 0, pxcr->ulShmSize );
          pxcr->pConsoleRec->TaskCountsInfo.lNumOfActiveTasks = pxcr->lExtParam;
          rc = sql92c_send (pxcr, CON_REQ_TASK_COUNTS,
                            &pxcr->pConsoleRec->TaskCountsInfo.fInfoAvailable);

          if ( rc != NO_ERROR ) return ( rc );

          rc = sql92c_show_task_count_info ( pxcr->pConsoleRec );

          if ( rc != NO_ERROR ) return ( rc );
          }

        if ( pxcr->lParam & PARA_TOTAL_COUNTS )
          {
          memset ( pxcr->pConsoleRec, 0, pxcr->ulShmSize );

          rc = sql92c_send (pxcr, CON_REQ_TOTAL_COUNTS,
                            &pxcr->pConsoleRec->TotalCountsInfo.fInfoAvailable);

          if ( rc != NO_ERROR ) return ( rc );

          rc = sql92c_show_total_count_info ( pxcr->pConsoleRec );

          if ( rc != NO_ERROR ) return ( rc );
          }

        #endif

        if (pxcr->lRepeatCount != -1 ) // --- don't loop forever?
          pxcr->lRepeatCount--;

        if (( pxcr->lInterval ) && ( pxcr->lRepeatCount ))
          SLEEP( pxcr->lInterval * 1000 );
        }

      break;

    default:
      return ( 1 );
    }


  DBGOUT;
  return ( rc );
  }

/*------------------------------*/

static LONG sql92c_check_accesss ( PXCONS_REC  pxcr )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_check_accesss"
  APIRET            rc           = NO_ERROR;
  #if defined (_WIN32)
   HEV              hevSemCoord  = 0;
  #endif

  DBGIN;

  #if defined (_WIN32)
   //
   // --- try to open coordinators semaphore to check access rights
   //
   rc = sql41c_open_event_sem ( &hevSemCoord, SEM_COORD, pxcr->pszServerDB,
                                ERROR_ACCESS_DENIED);

   if (rc == NO_ERROR )
     sql41c_close_event_sem ( hevSemCoord, "COORD" );
   else if (rc == ERROR_ACCESS_DENIED)
     MSGCD(( ERR_ACCESS_DENIED ))
   else if (rc == ERROR_SEM_NOT_FOUND)
     MSGCD ((ERR_DB_OR_GW_NOT_STARTED, pxcr->pszServerDB))
  #endif

  DBGOUT;
  return ( rc );
  }

/*------------------------------*/

static LONG sql92c_get_num_param  ( PSZ      pszParam,
                                    PLONG    plNum  )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_get_num_param"
  PSZ      pszTmp;
  DBGIN;

  // --- is the parameter valid?
  pszTmp = pszParam;
  while ( *pszTmp != '\0')
    {
    if (!isdigit ( *pszTmp ))
      return ( 1 );

    pszTmp++;
    }

  *plNum = atol (pszParam);

  return ( NO_ERROR );
  }

/*------------------------------*/

static LONG sql92c_get_idx_param  ( PSZ      pszParam,
                                    CHAR     cPrefix,
                                    PLONG    plNum  )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_get_idx_param"
  PSZ      pszTmp;
  PSZ      pszParamTmp;
  DBGIN;

  // --- is the parameter valid?
  if ( *pszParam == _toupper(cPrefix) || *pszParam == cPrefix )
    {
    pszTmp      = pszParam + 1;
    pszParamTmp = pszParam + 1;
    }
  else
    {
    pszTmp      = pszParam;
    pszParamTmp = pszParam;
    }

  while ( *pszTmp != '\0')
    {
    if (!isdigit ( *pszTmp ))
      return ( 1 );

    pszTmp++;
    }

  *plNum = atol (pszParamTmp);

  return ( NO_ERROR );
  }

/*------------------------------*/

static LONG sql92c_GetProcessorAffinityMask ( PSZ      pszParam,
                                              PLONG    ProcessAffinityMask  )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_GetProcessorAffinityMask"
  APIRET   rc = NO_ERROR;
  int      i, ParamLen ;

  DBGIN;

  for ( i = ParamLen = strlen(pszParam), *ProcessAffinityMask = 0 ; 
        i > 0 && rc == NO_ERROR; i-- )
	if  ( pszParam[i-1] == '1' ) 
      *ProcessAffinityMask |= ( 1 << ( ParamLen-i) ) ;
	else
	  if ( pszParam[i-1] != '0' )
        rc = !NO_ERROR ;

  return ( rc );
  }

/*------------------------------*/

static LONG sql92c_get_time_param  ( PSZ pszParam )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_get_time_param"
  DBGIN;

  strupr (pszParam);

  if ( e92_strcont( XCONS_CMD_ENABLE, pszParam,
                       XCONS_MIN_CMD_LEN_ENABLE ))
   return( PARA_ENABLE );

  if ( e92_strcont( XCONS_CMD_DISABLE,  pszParam,
                       XCONS_MIN_CMD_LEN_DISABLE ))
   return( PARA_DISABLE );

  return ( PARA_NONE );
  }

/*------------------------------*/

static LONG sql92c_get_set_clear_param  ( PSZ    pszParam )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_get_set_clear_param"
  DBGIN;

  strupr (pszParam);

  if ( e92_strcont( XCONS_CMD_SET, pszParam,
                    XCONS_MIN_CMD_LEN_SET ))

   return( PARA_SET );

  if ( e92_strcont( XCONS_CMD_CLEAR, pszParam,
                    XCONS_MIN_CMD_LEN_CLEAR ))
   return( PARA_CLEAR );


  return ( PARA_NONE );
  }

/*------------------------------*/

static LONG sql92c_get_show_res_param  ( PSZ    pszParam,
                                         LONG   lCommand )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_get_show_res_param"
  DBGIN;

  strupr (pszParam);

  if ( e92_strcont( XCONS_CMD_IO, pszParam,
                       XCONS_MIN_CMD_LEN_IO ))
   return( PARA_IO );

  if ( e92_strcont( XCONS_CMD_AIO, pszParam,
                       XCONS_MIN_CMD_LEN_AIO ))
   return( PARA_AIO );

  if ( e92_strcont( XCONS_CMD_REGIONS, pszParam,
                       XCONS_MIN_CMD_LEN_REGIONS  ) )
   return( PARA_REGIONS );

  if ( e92_strcont( XCONS_CMD_UKT_SLEEP_STAT, pszParam,
                       XCONS_MIN_CMD_LEN_UKT_SLEEP_STAT ) )
   return( PARA_UKT_SL_STAT );

  if ( e92_strcont( XCONS_CMD_ALL, pszParam,
                       XCONS_MIN_CMD_LEN_ALL ))
   return( PARA_ALL );

  #ifndef RUN_AS_SYSMON
   if ( e92_strcont( XCONS_CMD_THREAD_TIMES, pszParam,
                       XCONS_MIN_CMD_LEN_THREAD_TIMES ))
   return( PARA_THREAD_TIMES );

   if ( e92_strcont( XCONS_CMD_PSE36_STAT, pszParam,
                       XCONS_MIN_CMD_LEN_PSE36_STAT ))
   return( PARA_PSE36_STAT );

   if ( e92_strcont( XCONS_CMD_PSE36_DATA_PAGE_ARRAY, pszParam,
                       XCONS_MIN_CMD_LEN_PSE36_DATA_PAGE_ARRAY ))
   return( PARA_PSE36_FULLSTAT );

   if ( e92_strcont( XCONS_CMD_UKT_QUEUES, pszParam,
                       XCONS_MIN_CMD_LEN_UKT_QUEUES ) )
    return( PARA_UKT_QUEUES );

   if ( e92_strcont( XCONS_CMD_SUSPENDS, pszParam,
                        XCONS_MIN_CMD_LEN_SUSPENDS ) )
    return( PARA_SUSPENDS );

   if ( e92_strcont( XCONS_CMD_TASK_COUNTS, pszParam,
                        XCONS_MIN_CMD_LEN_TASK_COUNTS ) )
    return( PARA_TASK_COUNTS );

   if ( e92_strcont( XCONS_CMD_QUEUE_STAT, pszParam,
                        XCONS_MIN_CMD_LEN_QUEUE_STAT ))
    return( PARA_QUEUE_STAT );
  #endif

 if ( lCommand == COM_SHOW_UPDATE )
   {
   if ( e92_strcont( XCONS_CMD_VERSIONS, pszParam,
                        XCONS_MIN_CMD_LEN_VERSIONS ))
   return( PARA_VERSIONS );

   if ( e92_strcont( XCONS_CMD_STORAGE,  pszParam,
                        XCONS_MIN_CMD_LEN_STORAGE ))
   return( PARA_STORAGE );

   if ( e92_strcont( XCONS_CMD_RTE, pszParam,
                        XCONS_MIN_CMD_LEN_RTE ))
    return( PARA_RTE );

   if ( e92_strcont( XCONS_CMD_TASKS, pszParam,
                        XCONS_MIN_CMD_LEN_TASKS ))
    return( PARA_TASKS );

   if ( e92_strcont( XCONS_CMD_ACTIVE_TASKS, pszParam,
                        XCONS_MIN_CMD_LEN_ACTIVE_TASKS ))
    return( PARA_ACTIVE );

   if ( e92_strcont( XCONS_CMD_RUNNABLE_TASKS, pszParam,
                        XCONS_MIN_CMD_LEN_RUNNABLE_TASKS ))
    return( PARA_RUNNABLE );

    #ifndef RUN_AS_SYSMON
     if ( e92_strcont( XCONS_CMD_STATE, pszParam,
                          XCONS_MIN_CMD_LEN_STATE ))
      return( PARA_STATE );

     if ( e92_strcont( XCONS_CMD_TASK_STAT, pszParam,
                          XCONS_MIN_CMD_LEN_TASK_STAT ))
      return( PARA_TASK_STAT );

     if ( e92_strcont( XCONS_CMD_TOTAL_COUNTS, pszParam,
                          XCONS_MIN_CMD_LEN_TOTAL_COUNTS ))
      return( PARA_TOTAL_COUNTS );
    #endif
    }

  return ( PARA_NONE );
  }

/*------------------------------*/

static LONG sql92c_get_com_param  ( PSZ    pszParam )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_get_com_param"
  DBGIN;

  strupr (pszParam);


  if ( e92_strcont( XCONS_CMD_HELP1, pszParam,
                       XCONS_MIN_CMD_LEN_HELP1 ))
   return( COM_HELP );

  if ( e92_strcont( XCONS_CMD_HELP2, pszParam,
                       XCONS_MIN_CMD_LEN_HELP2 ))
   return( COM_HELP );

  #ifdef RUN_AS_SYSMON

   if ( e92_strcont( XCONS_CMD_UPDATE, pszParam,
                        XCONS_MIN_CMD_LEN_UPDATE ))
     return( COM_SHOW_UPDATE );

   if ( e92_strcont( XCONS_CMD_DROP, pszParam,
                        XCONS_MIN_CMD_LEN_DROP ))
     return( COM_DROP );

  #else

   if ( e92_strcont( XCONS_CMD_ERRIOR, pszParam,
                        XCONS_MIN_CMD_LEN_ERRIO ))
     return( COM_ERRIOR );

   if ( e92_strcont( XCONS_CMD_ERRIOW, pszParam,
                        XCONS_MIN_CMD_LEN_ERRIO ))
     return( COM_ERRIOW );

   if ( e92_strcont( XCONS_CMD_CANCEL, pszParam,
                        XCONS_MIN_CMD_LEN_CANCEL ))
    return( COM_CANCEL );

   if ( e92_strcont( XCONS_CMD_DEBUG_BREAK, pszParam,
                        XCONS_MIN_CMD_LEN_DEBUG_BREAK ))
    return( COM_DEBUG_BREAK );

   if ( e92_strcont( XCONS_CMD_UKT_AFFINITY, pszParam,
                        XCONS_MIN_CMD_LEN_UKT_AFFINITY ))
    return( COM_UKT_AFFINITY );

   // PTS 1107070
   if ( e92_strcont( XCONS_CMD_PROCESSOR_MASK, pszParam,
                        XCONS_MIN_CMD_LEN_PROCESSOR_MASK ))
    return( COM_PROCESSOR_MASK );

   if ( e92_strcont( XCONS_CMD_KILL, pszParam,
                        XCONS_MIN_CMD_LEN_KILL ))
    return( COM_KILL );

   if ( e92_strcont( XCONS_CMD_SHOW, pszParam,
                        XCONS_MIN_CMD_LEN_SHOW ))
     return( COM_SHOW_UPDATE );

   if ( e92_strcont( XCONS_CMD_TIME_MEASURE,  pszParam,
                        XCONS_MIN_CMD_LEN_TIME_MEASURE ))
     return( COM_TIME_MEASURE );

   if ( e92_strcont( XCONS_CMD_RESET_CNT, pszParam,
                        XCONS_MIN_CMD_LEN_RESET_CNT ))
     return( COM_RESET );

   if ( e92_strcont( XCONS_CMD_GETPARAM, pszParam,
                        XCONS_MIN_CMD_LEN_GETPARAM ))
     return( COM_GETPARAM );

   if ( e92_strcont( XCONS_CMD_PUTPARAM , pszParam,
                        XCONS_MIN_CMD_LEN_PUTPARAM ))
     return( COM_PUTPARAM );

   if ( e92_strcont( XCONS_CMD_DEBUGLEVEL , pszParam,
                        XCONS_MIN_CMD_LEN_DEBUGLEVEL ))
     return( COM_SET_DEBUGLEVEL );

   if ( sql02_get_platform_id () == VER_PLATFORM_WIN32_NT )
     {
     if ( e92_strcont( XCONS_CMD_CRASH, pszParam,
                          XCONS_MIN_CMD_LEN_CRASH ))
       return( COM_CRASH_DUMP );
     }
  #endif

  return ( COM_UNDEF );
  }


/*------------------------------*/


static BOOL  sql92c_wait ( PBOOL pfFlag )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_wait"
  INT i = MAX_WAIT * 10;

  DBGPAS;

  while ( *pfFlag == FALSE && --i )
    SLEEP( 100 );

  if (i == 0)
    MSGCD (( ERR_COMMAND_TIMEOUT ));

  return (i);
  }

/*------------------------------*/

static LONG sql92c_show_version_info ( PXCONS_REC pxcr )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_show_version_info"

  DBGPAS;

  #ifdef RUN_AS_SYSMON

   return ( eo921SysmonUpdateVersionTab (pxcr->KernelVersion, pxcr->RTEVersion) );

  #else

   printf ( XCONS_SHOW_VERS_KERNEL, pxcr->KernelVersion );
   printf ( XCONS_SHOW_VERS_RTE   , pxcr->RTEVersion    );

  #endif

  return ( NO_ERROR );
  }

/*------------------------------*/

static LONG sql92c_show_task_info ( PCONSOLE_REC  pConsoleRec )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_show_task_info"
  INT                         i;
  PCONS_SINGLE_TASK_INFO_REC  pTaskInfos;
  #ifdef RUN_AS_SYSMON
   APIRET                     rc = NO_ERROR;
   teo00_Uint1                ShowId ;
  #else
   CHAR                       szRegionIndex[20];
   CHAR                       szBegExclCnt[20];
   CHAR                       szApplPid[14];
   CHAR                       szTimeOut[20];
   CHAR                       szTid[12];
   ULONG                      ulUnconnectedTasks   = 0;
   ULONG                      ulUnusedServerTasks  = 0;
   ULONG                      ulTimeOut            = 0;
   static  BOOL               FirstCall = TRUE ;
  #endif
  

  DBGPAS;

  pTaskInfos = (PCONS_SINGLE_TASK_INFO_REC)
                   ((PCHAR)pConsoleRec +
                           pConsoleRec->TaskInfo.ulTaskInfoOffset);

  #ifdef RUN_AS_SYSMON

 
   ShowId = pConsoleRec->TaskInfo.StateToShow == SHOW_ALL_EO92 
          ? SHOW_ALL_EO92 : SHOW_ACTIVE_EO92 ;
 
   rc = eo921SysmonCreateShowTaskTab (ShowId);

   if ( rc != NO_ERROR )
     return ( rc );

   for ( i = 0; i < pConsoleRec->TaskInfo.lNumOfActiveTasks; i++ )
     {
     if (( pTaskInfos[i].TaskState == TSK_INITIALIZED_EO00         ||
           pTaskInfos[i].TaskState == TSK_INACTIVE_EO00            ||
           pTaskInfos[i].TaskState == TSK_MARKED_FOR_STARTUP_EO00  ||
           pTaskInfos[i].TaskState == TSK_CONNECTWAIT_EO00       ) &&
           pTaskInfos[i].TaskType  != TT_UT_EO00 )
       continue;

       rc = eo921SysmonUpdateShowTaskTab ( ShowId,
                                           pTaskInfos[i].TaskType,
                                           pTaskInfos[i].CancelFlag,
                                           pTaskInfos[i].ulTaskIndex,
                                           pTaskInfos[i].TaskState,
                                           pTaskInfos[i].ulStateReason,
                                           pTaskInfos[i].Tid,
                                           pTaskInfos[i].ApplPid,
                                           pTaskInfos[i].Sid,
                                           pTaskInfos[i].ulTaskTimeOut,
                                           pTaskInfos[i].lWaitForRoot,
                                           pTaskInfos[i].lWaitForTask,
                                           pTaskInfos[i].DevNameWaitingFor ) ;
     }
   if ( rc == NO_ERROR )
     eo921SysmonCommit();
   else
     return(rc);
  #else

   if ( FirstCall )
     {
       printf ( XCONS_SHOW_TASK_HEADER );
       FirstCall = FALSE ;
     }
   else
      printf ("\n" ) ;

   if ( pConsoleRec->TaskInfo.lNumOfActiveTasks == 0 )
     return ( NO_ERROR );

   for ( i = 0; i < pConsoleRec->TaskInfo.lNumOfActiveTasks; i++)
     {
     if ( pTaskInfos[i].ApplPid != (PID)UNDEF)
       {
       szApplPid[0] = pTaskInfos[i].szClientNode[0] != '\0' ? '*' : ' ';
       szApplPid[1] = ' ';
       _ultoa ( pTaskInfos[i].ApplPid,
                szApplPid + 2, 10 );
       }
     else
      szApplPid[0] = '\0';

     if ( pTaskInfos[i].Sid )
       {
       _ultoa ( pTaskInfos[i].Sid, szRegionIndex, 10 );

       if ( pTaskInfos[i].ulBegExclCnt )
         _ultoa (pTaskInfos[i].ulBegExclCnt, szBegExclCnt, 10);
       else
         szBegExclCnt[0] = '\0';
       }
     else
       {
       szRegionIndex[0] = '\0';
       szBegExclCnt[0]  = '\0';
       }

     if ( pTaskInfos[i].TaskState == TSK_VRECEIVE_EO00 )
       {
       if ( pTaskInfos[i].ulTaskTimeOut == (ULONG) UNDEF )
         strncpy (szTimeOut, XCONS_SHOW_TASK_NO_TIMEOUT, 10);
       else if (pTaskInfos[i].ulTaskTimeOut < (ULONG)time(NULL))
         ulTimeOut = 0;
       else
         ulTimeOut = pTaskInfos[i].ulTaskTimeOut - time (NULL);

       if ( pTaskInfos[i].ulTaskTimeOut != (ULONG) UNDEF )
         _ultoa ( ulTimeOut, szTimeOut, 10 );
       else
         // strncpy (szTimeOut, XCONS_SHOW_TASK_NO_TIMEOUT, 10);

       szTimeOut[10] = '\0';
       }
     else
       szTimeOut[0] = '\0';

     if (( pTaskInfos[i].TaskState == TSK_INITIALIZED_EO00         ||
           pTaskInfos[i].TaskState == TSK_INACTIVE_EO00            ||
           pTaskInfos[i].TaskState == TSK_MARKED_FOR_STARTUP_EO00  ||
           pTaskInfos[i].TaskState == TSK_CONNECTWAIT_EO00       ) &&
           pTaskInfos[i].TaskType  != TT_UT_EO00 )
       {
       if ( pTaskInfos[i].TaskType == TT_SV_EO00 )
         ulUnusedServerTasks++;
       else
         ulUnconnectedTasks++;
       }
     else
       {
       szTid[0] = '0';
       szTid[1] = 'x';
       _ultoa ( pTaskInfos[i].Tid, szTid + 2, 16 );
       strupr ( szTid + 2 );

       printf ( XCONS_SHOW_TASK_INFO,
                pTaskInfos[i].ulTaskIndex,
                szTid,
                eo92GetTaskTypeName ( pTaskInfos[i].TaskType ),
                szApplPid,
                pTaskInfos[i].QueLocation == TASK_IS_IN_RUN_QUEUE_EO00 ? '*' : ' ' ,
                eo92GetTaskStateName (pTaskInfos[i].TaskState,pTaskInfos[i].ulStateReason) ,
                ! pTaskInfos[i].CancelFlag ? ' ' : '!', /* CR 1103476 */
                ! pTaskInfos[i].PrioFlag   ? ' ' :
                         '0' + pTaskInfos[i].PrioFlag
                             + pTaskInfos[i].SavePrioFlag ,
                szTimeOut,
                szBegExclCnt,
                szRegionIndex );
       }
     }

   if ( ulUnusedServerTasks )
     {
     printf ( XCONS_SHOW_SERVER_UNUSED, ulUnusedServerTasks );
     }

   if ( ulUnconnectedTasks )
     {
     printf ( XCONS_SHOW_TASK_CONN_INFO,
              ulUnconnectedTasks, XCONS_TSK_STATE_CONNECTWAIT );
     }
  #endif


  return ( NO_ERROR );
  }

/*------------------------------*/

static LONG sql92c_show_io_info ( PCONSOLE_REC   pConsoleRec )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_show_io_info"
  INT                i;
  ULONG              ulDevspaceNo;
  PCONS_DEV_INFO_REC pDevspaceInfos;

  #ifdef RUN_AS_SYSMON
   APIRET            rc = NO_ERROR;
  #else
   PATHNAME          szPathName;
   ULONG             ulReadCnt  = 0;
   ULONG             ulWriteCnt = 0;
   ULONG             ulTotal    = 0;
  #endif

  DBGPAS;

  pDevspaceInfos = (PCONS_DEV_INFO_REC)
                     ((PCHAR)pConsoleRec +
                             pConsoleRec->IOInfo.ulDevspaceInfoOffset);

  #ifdef RUN_AS_SYSMON
   rc = eo921SysmonCreateIOTab ();

   if ( rc != NO_ERROR )
     return ( rc );
  #else
   printf ( XCONS_SHOW_IO_HEADER );
  #endif

  for ( i = 0; i < pConsoleRec->IOInfo.lNumOfDevspaces; i++)
    {

    ulDevspaceNo = pDevspaceInfos[ i ].ulDevspaceNo;

    #ifdef RUN_AS_SYSMON
 
     rc = eo921SysmonUpdateIOTab ( pDevspaceInfos[ i ].DevspaceName,
                                   ulDevspaceNo,
                                   pDevspaceInfos[ i ].ulReadCnt,
                                   pDevspaceInfos[ i ].ulWriteCnt);


     if ( rc != NO_ERROR )
       return ( rc );
     else
       eo921SysmonCommit();

    #else

     // --- The compress value (44) must be identical to the value
     //     defined in 'XCONS_SHOW_IO_INFO'!!!!!!!!!!!
     eo92_compress_path( szPathName, pDevspaceInfos[ i ].DevspaceName, 44 );

     printf ( XCONS_SHOW_IO_INFO,
              szPathName,
              ulDevspaceNo,
              pDevspaceInfos[ i ].ulReadCnt,
              pDevspaceInfos[ i ].ulWriteCnt,
              pDevspaceInfos[ i ].ulReadCnt +
              pDevspaceInfos[ i ].ulWriteCnt );

     ulReadCnt  += pDevspaceInfos[ i ].ulReadCnt;
     ulWriteCnt += pDevspaceInfos[ i ].ulWriteCnt;
     ulTotal    += pDevspaceInfos[ i ].ulReadCnt +
                   pDevspaceInfos[ i ].ulWriteCnt;
    #endif
    }

  #ifndef RUN_AS_SYSMON
   printf ( XCONS_SHOW_IO_INFO_TOTAL,
            XCONS_SHOW_IO_INFO_TOTAL_STR,
            ulReadCnt,
            ulWriteCnt,
            ulTotal );
  #endif

  return ( NO_ERROR );
  }


/*------------------------------*/


static LONG sql92c_show_aio_info ( PCONSOLE_REC   pConsoleRec )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_show_aio_info"

  INT                        i;
  PCONS_ASYNCI_INFO_REC      pASYNCiInfos;
  #ifdef RUN_AS_SYSMON
   APIRET            rc = NO_ERROR;
  #else
  CHAR                       szTid[12];
  PATHNAME                   szPathName;
  CHAR                       szMode[2];
  #endif

  pASYNCiInfos    = (PCONS_ASYNCI_INFO_REC)
                      ((PCHAR)pConsoleRec +
                          pConsoleRec->AIOInfo.ulASYNCiInfoOffset);


  #ifdef RUN_AS_SYSMON
   rc = eo921SysmonCreateAsynIOTab ();

   if ( rc != NO_ERROR )
     return ( rc );
  #else
   printf ( XCONS_SHOW_ASYNCIO_HEADER );
  #endif

  if ( pConsoleRec->AIOInfo.lNumOfASYNCis )
   {
   for ( i = 0; i < pConsoleRec->AIOInfo.lNumOfASYNCis; i++ )
     {
    #ifdef RUN_AS_SYSMON
     rc =  eo921SysmonUpdateAsynIOTab ( pASYNCiInfos[i].szLogDeviceName,
                                        i, &pASYNCiInfos[i].DevIOStat ) ;

     if ( rc != NO_ERROR )
       return ( rc );
    #else
     // --- The compress value (20) must be identical to the value
     //     defined in 'XCONS_SHOW_ASYNCIO_DEV_INFO'!!!!!!!!!!!
     eo92_compress_path ( szPathName,
                          pASYNCiInfos[i].szLogDeviceName,
                          20 );

     szTid[0] = '0';
     szTid[1] = 'x';
     _ultoa ( pASYNCiInfos[i].Tid, szTid + 2, 16 );
     strupr ( szTid + 2 );

     if ( pASYNCiInfos[i].fForWriting )
       {
       szMode[0] = 'W';
       szMode[1] = '\0';
       }
     else
       {
       szMode[0] = 'R';
       szMode[1] = '\0';
       }

     printf ( XCONS_SHOW_ASYNCIO_INFO,
              szPathName,
              szTid,
              szMode,
              pASYNCiInfos[i].ulMaxBlockCnt,
              pASYNCiInfos[i].ulBlockSize,
              pASYNCiInfos[i].DevIOStat.DevRead.ulCount,
              pASYNCiInfos[i].DevIOStat.DevWrite.ulCount,
              pASYNCiInfos[i].DevIOStat.ulIOQueLen,
              pASYNCiInfos[i].DevIOStat.ulMaxIOQueLen );
  #endif
     }
  }

  #ifdef RUN_AS_SYSMON
  if ( rc == NO_ERROR )
    eo921SysmonCommit();
  #endif

  return ( NO_ERROR );
  }

/*------------------------------*/

static LONG sql92c_show_region_info ( PCONSOLE_REC   pConsoleRec )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_region_info"
  INT                        i;
  INT                        j;
  PCONS_SINGLE_SEM_INFO_REC  pSingleSemInfos;
  PULONG                     pulWaitTasks;
  float                      flPercent;
  #ifdef RUN_AS_SYSMON
   APIRET                    rc = NO_ERROR;
  #else
   CHAR                      szOwnerTaskIndex[20];
  #endif

  DBGPAS;

  pSingleSemInfos = (PCONS_SINGLE_SEM_INFO_REC)
                        ((PCHAR)pConsoleRec +
                                pConsoleRec->SemInfo.ulSingleSemInfoOffset);

  #ifdef RUN_AS_SYSMON
   rc = eo921SysmonCreateRegionTab ();

   if ( rc != NO_ERROR )
     return ( rc );
  #else
   printf ( XCONS_SHOW_REGION_HEADER );
  #endif

  for ( i = 0; i < pConsoleRec->SemInfo.lNumOfSems; i ++ )
    {
    pulWaitTasks = (PULONG)((PCHAR)pConsoleRec +
                                pSingleSemInfos[i].ulWaitTaskOffset);

    if (( ! pSingleSemInfos[i].lGetCount ) ||
        ( ! pSingleSemInfos[i].lCollisionCount ))
      flPercent = 0.0f;
    else
      flPercent = (float) pSingleSemInfos[i].lCollisionCount * 100.0f /
                  (float) pSingleSemInfos[i].lGetCount;


    #ifdef RUN_AS_SYSMON

    rc = eo921SysmonUpdateRegionTab( pSingleSemInfos[i].szSemaName,
                                     pSingleSemInfos[i].ulSid,
                                     pSingleSemInfos[i].ulOwnerTaskIndex , 
                                     pSingleSemInfos[i].lNumWaitTasks <= 0 
                                     ? 0 : pulWaitTasks[0],
                                     pSingleSemInfos[i].lGetCount,
                                     pSingleSemInfos[i].lBegExclTasCount 
                                   + pSingleSemInfos[i].lEndExclTasCount,
                                     pSingleSemInfos[i].lCollisionCount,
                                     (teo00_Longint )flPercent,
                                     pSingleSemInfos[i].lExclusive);

     for ( j = 1; ( rc == NO_ERROR ) &&
           (j < (INT)pSingleSemInfos[i].lNumWaitTasks); j++ )
       {
       // --- set values
       rc = eo921SysmonUpdateRegionTab( pSingleSemInfos[i].szSemaName,
                                        pSingleSemInfos[i].ulSid,
                                        pSingleSemInfos[i].ulOwnerTaskIndex , 
                                        pulWaitTasks[j],
                                        0, 0, 0, 0, 0 ) ;
       }

     if ( rc == NO_ERROR )
       eo921SysmonCommit();
     else
       return ( rc );
 
    #else

     if (pSingleSemInfos[i].ulOwnerTaskIndex)
       {
       szOwnerTaskIndex[0] = 'T';
       _ltoa ( pSingleSemInfos[i].ulOwnerTaskIndex, szOwnerTaskIndex + 1, 10 );
       }
     else
       szOwnerTaskIndex[0] = '\0';

     printf ( XCONS_SHOW_REGION_INFO,
              pSingleSemInfos[i].ulSid,
              pSingleSemInfos[i].szSemaName,
              szOwnerTaskIndex,
              pSingleSemInfos[i].lGetCount,
              pSingleSemInfos[i].lBegExclTasCount +
              pSingleSemInfos[i].lEndExclTasCount,
              pSingleSemInfos[i].lCollisionCount,
              pSingleSemInfos[i].lWaitCount,
              pSingleSemInfos[i].lExclusive,
              flPercent );

     if ( pSingleSemInfos[i].lNumWaitTasks)
       {
       printf ( XCONS_SHOW_REGION_WAIT_HD );

       for ( j = 0; j < (INT)pSingleSemInfos[i].lNumWaitTasks; j++ )
         {
         if ( !(j % TASKS_INFO_P_LINE) && j)
           printf ( XCONS_SHOW_REGION_WAIT_ILF,
                    pulWaitTasks[j] );
         else
           printf ( XCONS_SHOW_REGION_WAIT_I,
                    pulWaitTasks[j] );
         }
       printf ("\n");
       }

    #endif

    }

  return ( NO_ERROR );
  }

/*------------------------------*/

static LONG sql92c_show_storage_info ( PCONSOLE_REC   pConsoleRec )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_show_storage_info"
  PCONS_TASK_STORAGE_INFO_REC pTaskStorInfos;
  #ifdef RUN_AS_SYSMON
   APIRET                     rc = NO_ERROR;
  #else
   INT                        i;
  #endif
  DBGPAS;

  pTaskStorInfos = (PCONS_TASK_STORAGE_INFO_REC)
                       ((PCHAR)pConsoleRec +
                               pConsoleRec->StorageInfo.ulTaskStorInfoOffset);

  #ifdef RUN_AS_SYSMON
   rc = eo921SysmonCreateStorageTab ();

   if ( rc != NO_ERROR )
     return ( rc );

   #if 0
   //
   // --- CODESIZE
   //
   rc = eo921SysmonUpdateStorageTab (XCONS_STOR_DES_CODESIZE, 0, XCONS_STOR_VDES_CODESIZE);

   #endif

   //
   // --- STACKSIZE
   //
   if ( rc == NO_ERROR)
     rc = eo921SysmonUpdateStorageTab (XCONS_STOR_DES_TSK_STACK_SIZE, 
                                       pConsoleRec->StorageInfo.lTaskStacks / 1024,
                                       XCONS_STOR_VDES_TSK_STACK_SIZE);
   //
   // --- KGS Size
   //
   if ( rc == NO_ERROR)
     rc = eo921SysmonUpdateStorageTab (XCONS_STOR_DES_KGS, 
                                       pConsoleRec->StorageInfo.ulKGSSize / 1024,
                                       XCONS_STOR_VDES_KGS);
   //
   // --- KSS Size
   //
   if ( rc == NO_ERROR)
     rc = eo921SysmonUpdateStorageTab (XCONS_STOR_DES_KSS, 
                                       pConsoleRec->StorageInfo.ulKSSSize / 1024,
                                       XCONS_STOR_VDES_KSS);
   //
   // --- SHARED DYN. POOL
   //
   if ( rc == NO_ERROR)
     rc = eo921SysmonUpdateStorageTab (XCONS_STOR_DES_SHRD_DYN_POOLFREE, 
                                       pConsoleRec->StorageInfo.lDynPoolAvail / 1024,
                                       XCONS_STOR_VDES_SHRD_DYN_POOLFREE);

   if ( rc == NO_ERROR)
     rc = eo921SysmonUpdateStorageTab (XCONS_STOR_DES_SHRD_DYN_POOL_USED, 
                                       pConsoleRec->StorageInfo.lDynPoolUsed / 1024,
                                       XCONS_STOR_VDES_SHRD_DYN_POOL_USED);
   //
   // --- SHARED DYN. DATA
   //
   if ( rc == NO_ERROR)
     rc = eo921SysmonUpdateStorageTab (XCONS_STOR_DES_SHRD_DYN_DATFREE, 
                                       pConsoleRec->StorageInfo.lDynDatAvail / sizeof(tsp00_Page),
                                       XCONS_STOR_VDES_SHRD_DYN_DATFREE);

   if ( rc == NO_ERROR)
     rc = eo921SysmonUpdateStorageTab (XCONS_STOR_DES_SHRD_DYN_DAT_USED, 
                                       pConsoleRec->StorageInfo.lDynDatUsed / sizeof(tsp00_Page),
                                       XCONS_STOR_VDES_SHRD_DYN_DAT_USED);
   //
   // --- RTE ALLOCATION POOL
   //
   if ( rc == NO_ERROR)
     rc = eo921SysmonUpdateStorageTab (XCONS_STOR_DES_RTE_ALL_POOL_USED, 
                                       pConsoleRec->StorageInfo.ulRTEAllocPoolUsed / 1024,
                                       XCONS_STOR_VDES_RTE_ALL_POOL_USED);

  #if defined(_WIN32_WORKS_FINE)
   if ( rc == NO_ERROR)
     rc = eo921SysmonUpdateStorageTab (XCONS_STOR_DES_RTE_ALL_POOL_MAX, 
                                       pConsoleRec->StorageInfo.ulRTEAllocPoolMax / (1 MB),
                                       XCONS_STOR_VDES_RTE_ALL_POOL_MAX);
   #endif

   //
   // --- KERNEL ALLOCATION POOL
   //
   if ( rc == NO_ERROR)
     rc = eo921SysmonUpdateStorageTab (XCONS_STOR_DES_KRN_ALL_POOL_USED, 
                                       pConsoleRec->StorageInfo.ulKernelAllocPoolUsed / 1024,
                                       XCONS_STOR_VDES_KRN_ALL_POOL_USED);
 
  #if defined(_WIN32_WORKS_FINE)
   if ( rc == NO_ERROR)
     rc = eo921SysmonUpdateStorageTab (XCONS_STOR_DES_KRN_ALL_POOL_MAX, 
                                       pConsoleRec->StorageInfo.ulKernelAllocPoolMax / 1024,
                                       XCONS_STOR_VDES_KRN_ALL_POOL_MAX);
   #endif

   //
   // --- EXISTING TASKS
   //
   if ( rc == NO_ERROR)
     rc = eo921SysmonUpdateStorageTab (XCONS_STOR_DES_EXISTING_TASKS, 
                                       pConsoleRec->StorageInfo.ulTasksTotal,
                                       XCONS_STOR_VDES_EXISTING_TASKS);

   if ( rc == NO_ERROR)
     eo921SysmonCommit();
   else
     return ( rc );

  #else
   printf ( XCONS_SHOW_STOR_DYN_POOL,
            (pConsoleRec->StorageInfo.lDynPoolUsed +
             pConsoleRec->StorageInfo.lDynPoolAvail  )      / 1024,
            pConsoleRec->StorageInfo.lDynPoolUsed           / 1024,
            pConsoleRec->StorageInfo.lDynPoolAvail          / 1024 );
   printf ( XCONS_SHOW_STOR_DYN_DATA,
            (pConsoleRec->StorageInfo.lDynDatUsed  +
             pConsoleRec->StorageInfo.lDynDatAvail   )      / 1024,
            pConsoleRec->StorageInfo.lDynDatUsed            / 1024,
            pConsoleRec->StorageInfo.lDynDatAvail           / 1024 );

   #if defined(_WIN32)
    printf ( XCONS_SHOW_STOR_RTE_ALLOC_POOL,
            pConsoleRec->StorageInfo.ulRTEAllocPoolUsed     / 1024,
            pConsoleRec->StorageInfo.ulRTEAllocPoolCnt );
    printf ( XCONS_SHOW_STOR_KERN_ALLOC_POOL,
            pConsoleRec->StorageInfo.ulKernelAllocPoolUsed  / 1024,
            pConsoleRec->StorageInfo.ulKernelAllocPoolCnt );
   #else
    printf ( XCONS_SHOW_STOR_RTE_ALLOC_POOL_MX,
            pConsoleRec->StorageInfo.ulRTEAllocPoolUsed     / 1024,
            pConsoleRec->StorageInfo.ulRTEAllocPoolMax      / ( 1 MB ),
            pConsoleRec->StorageInfo.ulRTEAllocPoolCnt );
    printf ( XCONS_SHOW_STOR_KERN_ALLOC_POOL_MX,
            pConsoleRec->StorageInfo.ulKernelAllocPoolUsed  / 1024,
            pConsoleRec->StorageInfo.ulKernelAllocPoolMax   / 1024,
            pConsoleRec->StorageInfo.ulKernelAllocPoolCnt );
   #endif

   printf ( XCONS_SHOW_STOR_ALLOC_STCK,
            pConsoleRec->StorageInfo.lTaskStacks      / 1024 );
   printf ( XCONS_SHOW_STOR_KGS,
            pConsoleRec->StorageInfo.ulKGSSize        / 1024 );
   printf ( XCONS_SHOW_STOR_KSS,
            pConsoleRec->StorageInfo.ulKSSSize        / 1024 );

   printf ( XCONS_SHOW_STOR_TSK_HEADER );

   for ( i = 0; i < pConsoleRec->StorageInfo.lNumOfTasks; i++)
     {
     printf ( XCONS_SHOW_STOR_TSK_INFO,
              pTaskStorInfos[i].ulTaskIndex,
              eo92GetTaskTypeName ( pTaskStorInfos[i].TaskType ),
              (float)pTaskStorInfos[i].ulStackUsed / 1024.0,
              (float)pTaskStorInfos[i].ulStackFree / 1024.0 );
     }
  #endif

  return ( NO_ERROR );
  }

/*------------------------------*/

#define MAX_SECS_PER_LOW_WORD ((double )429.5)

static VOID sql92c_conv_thread_time ( FILETIME        *pThreadTime ,
                                      XC_THREAD_TIMES *pTime ,
                                      XC_THREAD_TIMES *pSumTime )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_conv_thread_time"

  double flHigh = (double ) ( pThreadTime->dwHighDateTime) ;
  double flLow  = (double ) ( pThreadTime->dwLowDateTime ) ;
  double flTotalSec  ;

  flTotalSec = flHigh * MAX_SECS_PER_LOW_WORD + flLow / 10000000.0 ;

  pTime->ulDays = ( ULONG ) floor ( flTotalSec / 3600.0 / 24.0 );
  flTotalSec   -= (double )(pTime->ulDays) * 3600.0 * 24.0 ;

  pTime->ulHr   = ( ULONG ) floor ( flTotalSec / 3600.0 );
  flTotalSec   -= (double )(pTime->ulHr) * 3600.0 ;

  pTime->ulMin = ( ULONG ) floor ( flTotalSec / 60.0 );
  flTotalSec   -= (double )(pTime->ulMin) * 60.0 ;

  pTime->ulSec  = ( ULONG ) floor ( flTotalSec  );
  pTime->ulMsec = ( pThreadTime->dwLowDateTime % 10000000 ) / 10000 ;

  pSumTime->ulMsec  += pTime->ulMsec ;
  pSumTime->ulSec   += pTime->ulSec + (pSumTime->ulMsec / 1000) ;
  pSumTime->ulMsec  %= 1000 ;
  pSumTime->ulMin   += pTime->ulMin + (pSumTime->ulSec  /   60) ;
  pSumTime->ulSec   %= 60 ;
  pSumTime->ulHr    += pTime->ulHr  + (pSumTime->ulMin  /   60) ;
  pSumTime->ulMin   %= 60 ;
  pSumTime->ulDays  += pTime->ulDays+ (pSumTime->ulHr   /   24) ;
  pSumTime->ulHr    %= 24 ;
  return ;
  }


/*------------------------------*/

static LONG sql92c_show_thread_times ( PCONSOLE_REC   pConsoleRec )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_show_thread_times"

  PCONS_SINGLE_THREAD_TIMES_INFO_REC pThreadTimesInfo ;
  LONG                               lActThrdIndex ;
  XC_THREAD_TIMES                    SysTime, UsrTime ;
  XC_THREAD_TIMES                    TotSysTime, TotUsrTime ;
  char                               szSysDays[5], szUsrDays[5] ;

  pThreadTimesInfo       = (PCONS_SINGLE_THREAD_TIMES_INFO_REC)
                           ((PCHAR)pConsoleRec +
                            pConsoleRec->ThreadTimeInfo.ulThreadTimesOffset);


  memset ( &SysTime   , 0, sizeof (SysTime) ) ;
  memset ( &UsrTime   , 0, sizeof (UsrTime) ) ;
  memset ( &TotSysTime, 0, sizeof (TotSysTime) );
  memset ( &TotUsrTime, 0, sizeof (TotUsrTime) );

  printf ( XCONS_SHOW_THREAD_TIMES_HEADER );

  for ( lActThrdIndex = 0; lActThrdIndex < pConsoleRec->ThreadTimeInfo.lNumOfUKTs; 
        lActThrdIndex++, pThreadTimesInfo++ )
    {
    sql92c_conv_thread_time ( &pThreadTimesInfo->SystemTime, &SysTime , &TotSysTime ) ;
    
    sql92c_conv_thread_time ( &pThreadTimesInfo->UserTime,   &UsrTime , &TotUsrTime ) ;

    printf ( XCONS_FORMAT_SHOW_THREAD_TIMES ,
             pThreadTimesInfo->szThreadName,
             pThreadTimesInfo->Tid,
             pThreadTimesInfo->szCluster, 
             SysTime.ulDays ? _ultoa ( SysTime.ulDays, szSysDays, 10 ) : "",
             SysTime.ulHr, SysTime.ulMin, SysTime.ulSec, SysTime.ulMsec ,
             UsrTime.ulDays ? _ultoa ( UsrTime.ulDays, szUsrDays, 10 ) : "",
             UsrTime.ulHr, UsrTime.ulMin, UsrTime.ulSec, UsrTime.ulMsec ) ; 
    }

  printf ( XCONS_FORMAT_SHOW_THREAD_TOT_TIMES ,
           TotSysTime.ulDays ? _ultoa ( TotSysTime.ulDays, szSysDays, 10 ) : "",
           TotSysTime.ulHr, TotSysTime.ulMin, TotSysTime.ulSec, TotSysTime.ulMsec ,
           TotUsrTime.ulDays ? _ultoa ( TotUsrTime.ulDays, szSysDays, 10 ) : "",
           TotUsrTime.ulHr, TotUsrTime.ulMin, TotUsrTime.ulSec, TotUsrTime.ulMsec ) ; 

  return ( NO_ERROR );
  }

/*------------------------------*/

static LONG sql92c_show_PSE36_stat ( PCONSOLE_REC   pConsoleRec )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_show_PSE36_stat"

  tos542_PSE36_Statistics  *PSE36Stat = & pConsoleRec->PSE36Info.PSE36Stat;

  if ( PSE36Stat->MaxDataPages > 0 ) 
  {
    printf ( XCONS_SHOW_PSE36_STAT_HEADER ) ;

    printf ( XCONS_SHOW_PSE36_USED_PAGES ,
                 XCONS_SHOW_PSE36_USED_DATA_PAGES ,
                 PSE36Stat->MaxDataPages,
                 XCONS_SHOW_PSE36_PAGES_TO_FLUSH ,
                 PSE36Stat->NumChangedDataPages ) ;

    printf ( XCONS_SHOW_PSE36_READ_PAGES ,
                 XCONS_SHOW_PSE36_READ_DATA_PAGES ,
                 PSE36Stat->NumDataReads,
                 XCONS_SHOW_PSE36_READ_FAIL_DATA ,
                 PSE36Stat->NumDataReadsFail);

    printf ( XCONS_SHOW_PSE36_WRITE_PAGES ,
                 XCONS_SHOW_PSE36_WRITE_DATA_PAGES ,
                 PSE36Stat->NumDataWrites,
                 XCONS_SHOW_PSE36_WRITE_FAIL_DATA ,
                 PSE36Stat->NumDataWritesFail) ;

    if ( pConsoleRec->PSE36Info.ulPSE36DataPageArrayOffset != 0 )
      sql92c_show_PSE36PageArray ( pConsoleRec ) ;
  }
  else
  {
    printf ( XCONS_SHOW_PSE36_NOT_ACTIVE ) ;
  }

  return ( NO_ERROR );
}
/*------------------------------*/

static LONG sql92c_show_PSE36PageArray ( PCONSOLE_REC   pConsoleRec )
{
  #undef  MF__
  #define MF__ MOD__"sql92c_show_PSE36PageArray"

  long  ActPage; 
  long  PageArraySize = pConsoleRec->PSE36Info.PSE36Stat.MaxDataPages ;
  char *PageArray     = (char *) ((PCHAR)pConsoleRec +
                                         pConsoleRec->PSE36Info.ulPSE36DataPageArrayOffset);
  char Line[80] ;
  int  ActLinePos = 0;
  bool LineHasCachedPages = FALSE;

  if ( PageArraySize > 0 )
    printf ( "\n  PageNo         F = Free    U = Used     C = Changed\n%s\n", XCONS_INFO_LINE);

  for ( ActPage=0; ActPage < PageArraySize ; ActPage++ )
  {
    if ( (ActPage % 50) == 0 )
    {
      Line[ActLinePos] = '\0' ; 

      if ( ActPage > 0 && LineHasCachedPages) 
        printf("\n %s", Line ) ;

      sprintf(Line, "%8u : ", ActPage ) ;
      ActLinePos = strlen(Line) ;

      LineHasCachedPages = FALSE ;
    }
    else
      if ( (ActPage % 10) == 0 )
      {
        Line[ActLinePos] = ' ' ;
        ActLinePos++ ;
      }

    if ( PageArray[ActPage] != 0 )
      LineHasCachedPages = TRUE ;

    switch ( PageArray[ActPage] )
    {
      case 0:  Line[ActLinePos] = 'F'; break ;
      case 1:  Line[ActLinePos] = 'U'; break ;
      case 2:  
      case 3:  Line[ActLinePos] = 'C'; break ;
      default: Line[ActLinePos] = '?'; break ;
    }
    ActLinePos++ ;
  }

  /* print rest */
  if ( (ActPage % 50) != 0 && LineHasCachedPages )
  {
    Line[ActLinePos] = '\0' ; 
    printf("\n %s", Line ) ;
  }

  return ( NO_ERROR );
}

/*------------------------------*/

static LONG sql92c_show_rte_info ( PCONSOLE_REC   pConsoleRec )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_show_rte_info"
  ULONG                       ulDevspaceNo;
  INT                         i;
  INT                         j;
  PCONS_UKT_INFO_REC          pUKTInfos;
  PCONS_DEV_INFO_REC          pDevspaceInfos;
  PCONS_DEVI_INFO_REC         pDEViInfos;
  PCONS_DEV_WORKER_INFO_REC   pDevWorkerInfos;
  BOOL                        fUKTAffinity = FALSE;
  #ifdef RUN_AS_SYSMON
   APIRET                     rc = NO_ERROR;
   char                       szThreadName[100];
  #else
   CHAR                       szTid[12];
   PATHNAME                   szPathName;
  #endif

  DBGPAS;

  pDevWorkerInfos = (PCONS_DEV_WORKER_INFO_REC)
                   ((PCHAR)pConsoleRec +
                       pConsoleRec->RTEInfo.ulDevWorkerInfoOffset);

  pUKTInfos       = (PCONS_UKT_INFO_REC)
                      ((PCHAR)pConsoleRec +
                          pConsoleRec->RTEInfo.ulUKTInfoOffset);


  pDEViInfos      = (PCONS_DEVI_INFO_REC)
                      ((PCHAR)pConsoleRec +
                          pConsoleRec->RTEInfo.ulDEViInfoOffset);


  #ifdef RUN_AS_SYSMON
   rc = eo921SysmonCreateShowRteTabs ();

   if ( rc != NO_ERROR )
     return ( rc );

   //
   // --- COORDINATOR
   //
   if ( pConsoleRec->RTEInfo.CoordTid != (TID)UNDEF)
      rc =  eo921SysmonUpdateDbThreadsTab ( XCONS_SHOW_RTE_KT_COORD,
                                            pConsoleRec->RTEInfo.CoordTid );
   //
   // --- TIMER
   //
   if (( rc == NO_ERROR ) || ( pConsoleRec->RTEInfo.TimerTid != (TID)UNDEF))
      rc =  eo921SysmonUpdateDbThreadsTab ( XCONS_SHOW_RTE_KT_TIMER,
                                            pConsoleRec->RTEInfo.TimerTid );
   //
   // --- CLOCK
   //
   if (( rc == NO_ERROR  ) || ( pConsoleRec->RTEInfo.ClockTid != (TID)UNDEF))
      rc =  eo921SysmonUpdateDbThreadsTab ( XCONS_SHOW_RTE_KT_CLOCK,
                                            pConsoleRec->RTEInfo.ClockTid );
   //
   // --- DEV0
   //
   if (( rc == NO_ERROR  ) || ( pConsoleRec->RTEInfo.Dev0Tid != (TID)UNDEF))
      rc =  eo921SysmonUpdateDbThreadsTab ( XCONS_SHOW_RTE_KT_DEV0,
                                            pConsoleRec->RTEInfo.Dev0Tid );
   //
   // --- ASYNC0
   //
   if (( rc == NO_ERROR  ) || ( pConsoleRec->RTEInfo.Async0Tid != (TID)UNDEF))
      rc =  eo921SysmonUpdateDbThreadsTab ( XCONS_SHOW_RTE_KT_ASYNC0,
                                            pConsoleRec->RTEInfo.Async0Tid );
   //
   // --- CONSOLE
   //
   if ((rc == NO_ERROR ) || (pConsoleRec->RTEInfo.ConsoleTid != (TID)UNDEF))
      rc =  eo921SysmonUpdateDbThreadsTab ( XCONS_SHOW_RTE_KT_CONSOLE,
                                            pConsoleRec->RTEInfo.ConsoleTid );
   //
   // --- REQUESTOR
   //
   if ((rc == NO_ERROR )||(pConsoleRec->RTEInfo.RequestorTid != (TID)UNDEF))
      rc =  eo921SysmonUpdateDbThreadsTab ( XCONS_SHOW_RTE_KT_REQUESTOR,
                                            pConsoleRec->RTEInfo.RequestorTid );
   if ( rc != NO_ERROR  )
     return ( rc );

   //
   // --- UKT's
   //
   for ( i = 0; i < pConsoleRec->RTEInfo.lNumOfUKTs; i++ )
     {
       sprintf ( szThreadName, XCONS_RTE_DESC_UKT, pUKTInfos[i].ulUKTIndex );

       rc = eo921SysmonUpdateDbUktTab ( szThreadName    , pUKTInfos[i].szCluster,
                                        pUKTInfos[i].Tid, pUKTInfos[i].ulActiveTasks ,
                                        pUKTInfos[i].ulTotalTasks,
                                        (ULONG)pUKTInfos[i].dDispatcherCalls ) ;
     }


   //
   // --- UKT I/O
   //

   for ( i = 0; i < pConsoleRec->RTEInfo.lNumOfUKTs && rc == NO_ERROR; i++ )
     {
     pDevspaceInfos = (PCONS_DEV_INFO_REC)
                        ((PCHAR)pConsoleRec +
                                pUKTInfos[i].ulDevspaceInfoOffset);

     for ( j = 0; j < pUKTInfos[i].lNumOfDevspaces && rc == NO_ERROR; j++ )
       {
       ulDevspaceNo = pDevspaceInfos[j].ulDevspaceNo;

       sprintf ( szThreadName, XCONS_RTE_DESC_UKT, pUKTInfos[i].ulUKTIndex );

       rc = eo921SysmonUpdateRteIOTab ( szThreadName, pUKTInfos[i].Tid,
                                        ulDevspaceNo,
                                        pDevspaceInfos[j].ulReadCnt,
                                        pDevspaceInfos[j].ulWriteCnt,0, 0);
       }
     }

   //
   // --- DEVi I/O
   //
   for ( i = 0; i < pConsoleRec->RTEInfo.lNumOfDEVis && rc == NO_ERROR; i++ )
     {
     ulDevspaceNo = pDEViInfos[i].ulDevspaceNo;

     if ( ulDevspaceNo == (ULONG)pConsoleRec->RTEInfo.lTraceDevspaceNo )
       strcpy (szThreadName, XCONS_RTE_DESC_DEV_VTRACE );
     else
       strcpy (szThreadName, XCONS_RTE_DESC_DEV );

     rc = eo921SysmonUpdateRteIOTab ( szThreadName, pDEViInfos[i].Tid,
                                      ulDevspaceNo,
                                      pDEViInfos[i].DevIOStat.DevRead.ulCount,
                                      pDEViInfos[i].DevIOStat.DevWrite.ulCount,
                                      pDEViInfos[i].DevIOStat.ulIOQueLen,
                                      pDEViInfos[i].DevIOStat.ulMaxIOQueLen );
     }

   if ( rc == NO_ERROR )
     eo921SysmonCommit();
   else
     return(rc);
  #else

   // --- show kernel thread info
   printf ( XCONS_SHOW_RTE_KT_HEADER );

   //
   // --- COORDINATOR
   //
   if ( pConsoleRec->RTEInfo.CoordTid != (TID)UNDEF)
     {
     szTid[0] = '0';
     szTid[1] = 'x';
     _ultoa ( pConsoleRec->RTEInfo.CoordTid, szTid + 2, 16 );
     strupr ( szTid + 2 );
     }
   else
     szTid[0] = '\0';

   printf ( XCONS_SHOW_RTE_KT_INFO, XCONS_SHOW_RTE_KT_COORD, szTid,
            sql92c_kt_state_name( pConsoleRec->RTEInfo.CoordState ) );

   //
   // --- TIMER
   //
   if ( pConsoleRec->RTEInfo.TimerTid != (TID)UNDEF)
     {
     szTid[0] = '0';
     szTid[1] = 'x';
     _ultoa ( pConsoleRec->RTEInfo.TimerTid, szTid + 2, 16 );
     strupr ( szTid + 2 );
     }
   else
     szTid[0] = '\0';

   printf ( XCONS_SHOW_RTE_KT_INFO, XCONS_SHOW_RTE_KT_TIMER, szTid,
            sql92c_kt_state_name( pConsoleRec->RTEInfo.TimerState ) );

   //
   // --- CLOCK
   //
   if ( pConsoleRec->RTEInfo.ClockTid != (TID)UNDEF)
     {
     szTid[0] = '0';
     szTid[1] = 'x';
     _ultoa ( pConsoleRec->RTEInfo.ClockTid, szTid + 2, 16 );
     strupr ( szTid + 2 );
     }
   else
     szTid[0] = '\0';

   printf ( XCONS_SHOW_RTE_KT_INFO, XCONS_SHOW_RTE_KT_CLOCK, szTid,
            sql92c_kt_state_name( pConsoleRec->RTEInfo.ClockState ) );

   //
   // --- DEV0
   //
   if ( pConsoleRec->RTEInfo.Dev0Tid != (TID)UNDEF)
     {
     szTid[0] = '0';
     szTid[1] = 'x';
     _ultoa ( pConsoleRec->RTEInfo.Dev0Tid, szTid + 2, 16 );
     strupr ( szTid + 2 );
     }
   else
     szTid[0] = '\0';

   printf ( XCONS_SHOW_RTE_KT_INFO, XCONS_SHOW_RTE_KT_DEV0, szTid,
            sql92c_kt_state_name( pConsoleRec->RTEInfo.Dev0State ) );

   //
   // --- ASYNC0
   //
   if ( pConsoleRec->RTEInfo.Async0Tid != (TID)UNDEF)
     {
     szTid[0] = '0';
     szTid[1] = 'x';
     _ultoa ( pConsoleRec->RTEInfo.Async0Tid, szTid + 2, 16 );
     strupr ( szTid + 2 );
     }
   else
     szTid[0] = '\0';

   printf ( XCONS_SHOW_RTE_KT_INFO, XCONS_SHOW_RTE_KT_ASYNC0, szTid,
            sql92c_kt_state_name( pConsoleRec->RTEInfo.Async0State ) );

   //
   // --- CONSOLE
   //
   if ( pConsoleRec->RTEInfo.ConsoleTid != (TID)UNDEF)
     {
     szTid[0] = '0';
     szTid[1] = 'x';
     _ultoa ( pConsoleRec->RTEInfo.ConsoleTid, szTid + 2, 16 );
     strupr ( szTid + 2 );
     }
   else
     szTid[0] = '\0';

   printf ( XCONS_SHOW_RTE_KT_INFO, XCONS_SHOW_RTE_KT_CONSOLE, szTid,
            sql92c_kt_state_name( pConsoleRec->RTEInfo.ConsoleState ) );

   //
   // --- REQUESTOR
   //
   if ( pConsoleRec->RTEInfo.RequestorTid != (TID)UNDEF)
     {
     szTid[0] = '0';
     szTid[1] = 'x';
     _ultoa ( pConsoleRec->RTEInfo.RequestorTid, szTid + 2, 16 );
     strupr ( szTid + 2 );
     }
   else
     szTid[0] = '\0';

   printf ( XCONS_SHOW_RTE_KT_INFO, XCONS_SHOW_RTE_KT_REQUESTOR, szTid,
            sql92c_kt_state_name( pConsoleRec->RTEInfo.RequestorState ) );

   //
   // --- DCOM0
   //
   if ( pConsoleRec->RTEInfo.DCOM0Tid != (TID)UNDEF)
     {
     szTid[0] = '0';
     szTid[1] = 'x';
     _ultoa ( pConsoleRec->RTEInfo.DCOM0Tid, szTid + 2, 16 );
     strupr ( szTid + 2 );
     }
   else
     szTid[0] = '\0';

   printf ( XCONS_SHOW_RTE_KT_INFO, XCONS_SHOW_RTE_KT_DCOM0, szTid,
            sql92c_kt_state_name( pConsoleRec->RTEInfo.DCOM0State ) );

   if ( pConsoleRec->RTEInfo.lNumOfDevWorkers )
     {
     printf ( XCONS_SHOW_RTE_DEV_WORKER_HEADER );

     for ( i = 0; i < pConsoleRec->RTEInfo.lNumOfDevWorkers; i++ )
       {
       szTid[0] = '0';
       szTid[1] = 'x';
       _ultoa ( pDevWorkerInfos[i].Tid, szTid + 2, 16 );
       strupr ( szTid + 2 );
       printf ( XCONS_SHOW_RTE_DEV_WORKER_INFO,
                i,
                szTid,
                sql92c_ukt_state_name( pDevWorkerInfos[i].ThrdState ),
                pDevWorkerInfos[i].ulIOCount );
       }
     printf ( XCONS_SHOW_RTE_CONCURR_DEV_WORKER,
              pConsoleRec->RTEInfo.ulNumOfConcurrentDevWorkers );
     }

   printf ( XCONS_SHOW_RTE_UKT_HEADER );

   for ( i = 0; i < pConsoleRec->RTEInfo.lNumOfUKTs; i++ )
     {
     szTid[0] = '0';
     szTid[1] = 'x';
     _ultoa ( pUKTInfos[i].Tid, szTid + 2, 16 );
     strupr ( szTid + 2 );

     printf ( XCONS_SHOW_RTE_UKT_INFO,
              pUKTInfos[i].ulUKTIndex,
              szTid,
              sql92c_ukt_state_name( pUKTInfos[i].ThrdState ),
              pUKTInfos[i].dDispatcherCalls,
              pUKTInfos[i].ulActiveTasks,
              pUKTInfos[i].ulTotalTasks,
              pUKTInfos[i].szCluster );


     if ( pUKTInfos[i].ulProcessor )
       fUKTAffinity = TRUE;
     }

   if ( fUKTAffinity )
     {
     printf ( XCONS_SHOW_RTE_UKT_AFFIN_HEADER );

     for ( i = 0; i < pConsoleRec->RTEInfo.lNumOfUKTs; i++ )
       {
       if ( pUKTInfos[i].ulProcessor )
         {
         printf ( XCONS_SHOW_RTE_UKT_AFFIN_INFO,
                  pUKTInfos[i].ulUKTIndex,
                  pUKTInfos[i].ulProcessor );
         }
       }
     }

   printf ( XCONS_SHOW_RTE_IO_HEADER );

   for ( i = 0; i < pConsoleRec->RTEInfo.lNumOfUKTs; i++ )
     {
     pDevspaceInfos = (PCONS_DEV_INFO_REC)
                        ((PCHAR)pConsoleRec +
                                pUKTInfos[i].ulDevspaceInfoOffset);

     for ( j = 0; j < pUKTInfos[i].lNumOfDevspaces; j++ )
       {
       ulDevspaceNo = pDevspaceInfos[j].ulDevspaceNo;

       // --- The compress value (20) must be identical to the value
       //     defined in 'XCONS_SHOW_RTE_IO_UKT_INFO'!!!!!!!!!!!
       eo92_compress_path( szPathName, pDevspaceInfos[j].DevspaceName, 20 );

       szTid[0] = '0';
       szTid[1] = 'x';
       _ultoa ( pUKTInfos[i].Tid, szTid + 2, 16 );
       strupr ( szTid + 2 );

       printf ( XCONS_SHOW_RTE_IO_UKT_INFO,
                pUKTInfos[i].ulUKTIndex,
                szTid,
                szPathName,
                ulDevspaceNo,
                pDevspaceInfos[j].ulReadCnt,
                pDevspaceInfos[j].ulWriteCnt );
       }
     }

   for ( i = 0; i < pConsoleRec->RTEInfo.lNumOfDEVis; i++ )
     {
     ulDevspaceNo = pDEViInfos[i].ulDevspaceNo;

     // --- The compress value (20) must be identical to the value
     //     defined in 'XCONS_SHOW_RTE_IO_DEV_INFO'!!!!!!!!!!!
     eo92_compress_path ( szPathName, pDEViInfos[i].DevspaceName, 20 );

     szTid[0] = '0';
     szTid[1] = 'x';
     _ultoa ( pDEViInfos[i].Tid, szTid + 2, 16 );
     strupr ( szTid + 2 );

     printf ( XCONS_SHOW_RTE_IO_DEV_INFO,
              pDEViInfos[i].ulDEViIndex,
              szTid,
              szPathName,
              ulDevspaceNo,
              pDEViInfos[i].DevIOStat.DevRead.ulCount,
              pDEViInfos[i].DevIOStat.DevWrite.ulCount,
              pDEViInfos[i].DevIOStat.ulIOQueLen,
              pDEViInfos[i].DevIOStat.ulMaxIOQueLen );
     }

  #endif

  return ( NO_ERROR );
  }

/*------------------------------*/

static LONG sql92c_show_total_count_info ( PCONSOLE_REC   pConsoleRec )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_show_total_count_info"

  CONS_TOTAL_COUNTS_INFO_REC      *TotCntInfo ;
  TotCntInfo = & pConsoleRec->TotalCountsInfo ;

  printf ( XCONS_SHOW_TOTAL_STAT_HEADER ) ;

  printf ( XCONS_SHOW_TOTAL_DISP_COMMAND_CNT ,
               XCONS_SHOW_TOTAL_KT_DISP_CALL_CNT ,
               TotCntInfo->ulTotalDispCalls,
               XCONS_SHOW_TOTAL_KT_COMMAND_CNT ,
               TotCntInfo->ulTotalCommands);

  printf ( XCONS_SHOW_TOTAL_KT_STATE_VWAIT_VSUSP_CNT ,
               XCONS_SHOW_TOTAL_KT_STATE_VWAIT_CNT,
               TotCntInfo->ulTotalVwait,
               XCONS_SHOW_TOTAL_KT_STATE_VSUSP_CNT,
               TotCntInfo->ulTotalVsusp );

  printf ( XCONS_SHOW_TOTAL_KT_STATE_VSLEEP,
               XCONS_SHOW_TOTAL_KT_STATE_VSLEEP_CNT,
               TotCntInfo->ulTotalVsleep);

  printf ( XCONS_SHOW_TOTAL_SELFIO_DEVIO_CNT ,
               XCONS_SHOW_TOTAL_KT_SELFIO_CNT ,
               TotCntInfo->ulTotalSelfIO,
               XCONS_SHOW_TOTAL_KT_DEVIO_CNT,
               TotCntInfo->ulTotalDevIO);

  printf ( XCONS_SHOW_TOTAL_REG_ACC_BEGEXCL_COLL,
               XCONS_SHOW_TOTAL_KT_REG_ACCESS,
               TotCntInfo->ulTotalRegAccess,
               XCONS_SHOW_TOTAL_KT_BEGEXCL_COLL,
               TotCntInfo->ulTotalBegExclTasCount);

  printf ( XCONS_SHOW_TOTAL_REG_COLL_ENDEXCL_COLL,
               XCONS_SHOW_TOTAL_KT_REG_COLL,
               TotCntInfo->ulTotalRegColl,
               XCONS_SHOW_TOTAL_KT_ENDEXCL_COLL,
               TotCntInfo->ulTotalEndExclTasCount);

  printf ( XCONS_SHOW_TOTAL_KT_REG_WAIT ,
               XCONS_SHOW_TOTAL_KT_REG_WAIT_CNT ,
               TotCntInfo->ulTotalRegWait);

  if ( TotCntInfo->TotalPrioStat.ulTotalCount )
   { printf ( XCONS_SHOW_TOTAL_PRIO_CNT_AND_OTHER,
                  XCONS_SHOW_TOTAL_KT_PRIO_CNT,
                  TotCntInfo->TotalPrioStat.ulTotalCount,
                  XCONS_SHOW_TOTAL_KT_PRIO_OTHER ,
                  TotCntInfo->TotalPrioStat.ulOtherUkt);
   }

  return ( NO_ERROR );
  }

/*------------------------------*/

static LONG sql92c_show_task_count_info ( PCONSOLE_REC   pConsoleRec )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_show_task_count_info"
  INT                                i;
  CHAR                               szApplPid[12];
  PCONS_SINGLE_TASK_COUNTS_INFO_REC  pTaskCountsInfo;
  teo52_TaskStateStatisticRec        *pTaskStateStat;
  teo52_TaskPrioStatistic            *pTaskPrioStat;
  BOOL                               fTimeMeasureActive;
  #ifndef RUN_AS_SYSMON
  int                                j1;
  #endif

  /* PTS 1104727 */
  if ( pConsoleRec->TaskCountsInfo.usReturnCode == CON_WRONG_TASK )
  {
    MSGCD ((ERR_XCONS_WRONG_TASK ));
    return ( 1 );
  }

  pTaskCountsInfo = (PCONS_SINGLE_TASK_COUNTS_INFO_REC)
                        ((PCHAR)pConsoleRec +
                          pConsoleRec->TaskCountsInfo.ulTaskCountsInfoOffset);

  fTimeMeasureActive = pConsoleRec->TaskCountsInfo.fTimeMeasureActive;

  for ( i = 0;
        i < pConsoleRec->TaskCountsInfo.lNumOfActiveTasks;
        i++, pTaskCountsInfo++ )
    {
    pTaskStateStat  = &pTaskCountsInfo->StateStat;
    pTaskPrioStat   = &pTaskCountsInfo->PrioStat;

    if ( pTaskCountsInfo->ApplPid != (PID)UNDEF )
      {
      _ultoa ( pTaskCountsInfo->ApplPid, szApplPid, 10 );
      }
    else
      {
      szApplPid[0] = '?';
      szApplPid[1] = '\0';
      }


    printf ( XCONS_SHOW_TASK_COUNT_HEADER,
                XCONS_INFO_LINE ,
                pTaskCountsInfo->ulTaskIndex,
                eo92GetTaskTypeName ( pTaskCountsInfo->TaskType ),
                szApplPid,
                XCONS_INFO_LINE ) ;

    if ( pTaskCountsInfo->szClientNode[0] )
      {
        printf ( XCONS_SHOW_TASK_COUNT_REM_NODE_AND_PID ,
                 XCONS_SHOW_TASK_COUNT_REM_NODE ,
                 pTaskCountsInfo->szClientNode ,
                 XCONS_SHOW_TASK_COUNT_REM_PID  ,
                 szApplPid );
      }

    printf ( XCONS_SHOW_TASK_COUNT_DISP_COMMAND_CNT    ,
                XCONS_SHOW_TASK_COUNT_KT_DISP_CALL_CNT,
                pTaskCountsInfo->ulDispatcherCalls,
                XCONS_SHOW_TASK_COUNT_KT_COMMAND_CNT,
                pTaskCountsInfo->ulCommands );

    printf ( XCONS_SHOW_TASK_COUNT_EXCL_SELF_SUSP_CNT  ,
                XCONS_SHOW_TASK_COUNT_KT_EXCLUSIVE_CNT ,
                pTaskCountsInfo->ulBegExclCalls ,
                XCONS_SHOW_TASK_COUNT_KT_SELF_SUSP_CNT ,
                pTaskCountsInfo->ulSelfSuspCount );

  #ifndef RUN_AS_SYSMON
#define XCONS_SHOW_TASK_COUNT_RESUME_CNT " SuspendLevel %d resume count %-2d total %-10d History ["
#define XCONS_SHOW_TASK_COUNT_RESUME_HISTORY " T%d"
#define XCONS_SHOW_TASK_COUNT_RESUME_END_HISTORY " ]\n"
    for ( j1 = 0; j1 < MAX_SUSPEND_LEVEL; j1++ )
    {
      if ( pTaskCountsInfo->totalResumerCount[j1] > 0 )
      {
          int j2;
          int j;
          printf ( XCONS_SHOW_TASK_COUNT_RESUME_CNT,
              j1, (int)pTaskCountsInfo->lResumeCount[j1], (int)pTaskCountsInfo->totalResumerCount[j1]);
          j2 = pTaskCountsInfo->totalResumerCount[j1];
          j = MAX_RESUMER_PIDS;
          while( (j2-- > 0) && (j-- > 0) )
          {
              printf ( " T%d", (int)pTaskCountsInfo->resumerPid[j2%MAX_RESUMER_PIDS][j1] );
          }
          printf ( " ]\n" );
      }
    }
  #endif RUN_AS_SYSMON

    if ( ! fTimeMeasureActive )
      {
      if ( pTaskStateStat->AsynIOReadStat.ulCount )
        printf ( XCONS_SHOW_TASK_COUNT_ASYN_CNT_PG ,
                       XCONS_SHOW_TASK_COUNT_KT_ASYN_R_CNT ,
                       pTaskStateStat->AsynIOReadStat.ulCount   ,
                       XCONS_SHOW_TASK_COUNT_KT_ASYN_R_PG ,
                       pTaskStateStat->AsynIOReadStat.ulPageCount);

      if ( pTaskStateStat->AsynIOWriteStat.ulCount )
        printf ( XCONS_SHOW_TASK_COUNT_ASYN_CNT_PG ,
                       XCONS_SHOW_TASK_COUNT_KT_ASYN_W_CNT ,
                       pTaskStateStat->AsynIOWriteStat.ulCount   ,
                       XCONS_SHOW_TASK_COUNT_KT_ASYN_W_CNT  ,
                       pTaskStateStat->AsynIOWriteStat.ulPageCount);

      if ( pTaskStateStat->SelfIOReadStat.ulCount ||
           pTaskStateStat->DevIOReadStat.ulCount   )
        printf ( XCONS_SHOW_TASK_COUNT_IO_CNT ,
                       XCONS_SHOW_TASK_COUNT_KT_SELF_R_CNT ,
                       pTaskStateStat->SelfIOReadStat.ulCount   ,
                       XCONS_SHOW_TASK_COUNT_KT_DEV_R_CNT ,
                       pTaskStateStat->DevIOReadStat.ulCount  ) ;

      if ( pTaskStateStat->SelfIOWriteStat.ulCount )
        printf ( XCONS_SHOW_TASK_COUNT_IO_CNT ,
                       XCONS_SHOW_TASK_COUNT_KT_SELF_W_CNT ,
                       pTaskStateStat->SelfIOWriteStat.ulCount,
                       XCONS_SHOW_TASK_COUNT_KT_SELF_W_PG ,
                       pTaskStateStat->SelfIOWriteStat.ulPageCount);

      if ( pTaskStateStat->DevIOWriteStat.ulCount )
        printf ( XCONS_SHOW_TASK_COUNT_IO_CNT ,
                       XCONS_SHOW_TASK_COUNT_KT_DEV_W_CNT  ,
                       pTaskStateStat->DevIOWriteStat.ulCount,
                       XCONS_SHOW_TASK_COUNT_KT_DEV_W_PG ,
                       pTaskStateStat->DevIOWriteStat.ulPageCount);

      printf ( XCONS_SHOW_TASK_COUNT_WAIT_SLEEP_SUSP_CNT  ,
                   XCONS_SHOW_TASK_COUNT_KT_STATE_VWAIT_CNT  ,
                   pTaskStateStat->VwaitStat.ulCount,
                   XCONS_SHOW_TASK_COUNT_KT_STATE_VSLEEP_CNT ,
                   pTaskStateStat->VsleepStat.ulCount,
                   XCONS_SHOW_TASK_COUNT_KT_STATE_VSUSP_CNT  ,
                   pTaskStateStat->VsuspStat.ulCount ) ;
      }
    else
      {
      PRI_STAT_TM (  XCONS_SHOW_TASK_COUNT_STATE_CNT_TIME,
                     pTaskStateStat->AsynIOReadStat,
                     XCONS_SHOW_TASK_COUNT_KT_ASYN_R_CNT  ,
                     XCONS_SHOW_TASK_COUNT_KT_ASYN_R_REL_TIME,
                     XCONS_SHOW_TASK_COUNT_KT_ASYN_R_ABS_TIME ) ;

      PRI_PG_PER_IO( XCONS_SHOW_TASK_COUNT_PG_PER_IO,
                     pTaskStateStat->AsynIOReadStat,
                     XCONS_SHOW_TASK_COUNT_KT_ASYN_R_PG,
                     XCONS_SHOW_TASK_COUNT_KT_PG_PER_IO ) ;

      PRI_STAT_TM  ( XCONS_SHOW_TASK_COUNT_STATE_CNT_TIME,
                     pTaskStateStat->AsynIOWriteStat,
                     XCONS_SHOW_TASK_COUNT_KT_ASYN_W_CNT  ,
                     XCONS_SHOW_TASK_COUNT_KT_ASYN_W_REL_TIME,
                     XCONS_SHOW_TASK_COUNT_KT_ASYN_W_ABS_TIME ) ;

      PRI_PG_PER_IO( XCONS_SHOW_TASK_COUNT_PG_PER_IO,
                     pTaskStateStat->AsynIOWriteStat,
                     XCONS_SHOW_TASK_COUNT_KT_ASYN_W_PG,
                     XCONS_SHOW_TASK_COUNT_KT_PG_PER_IO ) ;

      if ( pTaskStateStat->SelfIOReadStat.ulCount )
        printf ( XCONS_SHOW_TASK_COUNT_SELF_R_CNT_AVG_TIME ,
                     XCONS_SHOW_TASK_COUNT_KT_SELF_R_CNT ,
                     pTaskStateStat->SelfIOReadStat.ulCount   ,
                     XCONS_SHOW_TASK_COUNT_KT_SELF_R_AVG_TIME ,
                     AVG_TIME( pTaskStateStat->SelfIOReadStat ) ) ;

      if ( pTaskStateStat->SelfIOWriteStat.ulCount )
        printf ( XCONS_SHOW_TASK_COUNT_IO_CNT_PG_TIME    ,
                     XCONS_SHOW_TASK_COUNT_KT_SELF_W_CNT ,
                     pTaskStateStat->SelfIOWriteStat.ulCount   ,
                     XCONS_SHOW_TASK_COUNT_KT_SELF_W_PG  ,
                     pTaskStateStat->SelfIOWriteStat.ulPageCount   ,
                     XCONS_SHOW_TASK_COUNT_KT_SELF_W_AVG_TIME ,
                     AVG_TIME( pTaskStateStat->SelfIOWriteStat ) ) ;

      PRI_STAT_TM  ( XCONS_SHOW_TASK_COUNT_STATE_CNT_TIME,
                     pTaskStateStat->DevIOReadStat,
                     XCONS_SHOW_TASK_COUNT_KT_DEV_R_CNT  ,
                     XCONS_SHOW_TASK_COUNT_KT_DEV_R_REL_TIME,
                     XCONS_SHOW_TASK_COUNT_KT_DEV_R_ABS_TIME ) ;

      PRI_PG_PER_IO( XCONS_SHOW_TASK_COUNT_PG_PER_IO,
                     pTaskStateStat->DevIOReadStat,
                     XCONS_SHOW_TASK_COUNT_KT_DEV_R_PG,
                     XCONS_SHOW_TASK_COUNT_KT_PG_PER_IO ) ;

      PRI_STAT_TM  ( XCONS_SHOW_TASK_COUNT_STATE_CNT_TIME,
                     pTaskStateStat->DevIOWriteStat,
                     XCONS_SHOW_TASK_COUNT_KT_DEV_W_CNT  ,
                     XCONS_SHOW_TASK_COUNT_KT_DEV_W_REL_TIME,
                     XCONS_SHOW_TASK_COUNT_KT_DEV_W_ABS_TIME ) ;

      PRI_PG_PER_IO( XCONS_SHOW_TASK_COUNT_PG_PER_IO,
                     pTaskStateStat->DevIOWriteStat,
                     XCONS_SHOW_TASK_COUNT_KT_DEV_W_PG,
                     XCONS_SHOW_TASK_COUNT_KT_PG_PER_IO ) ;

      PRI_STAT_TM  ( XCONS_SHOW_TASK_COUNT_STATE_CNT_TIME,
                     pTaskStateStat->VwaitStat,
                     XCONS_SHOW_TASK_COUNT_KT_STATE_VWAIT_CNT  ,
                     XCONS_SHOW_TASK_COUNT_KT_VWAIT_REL_TIME,
                     XCONS_SHOW_TASK_COUNT_KT_VWAIT_ABS_TIME ) ;

      PRI_STAT_TM  ( XCONS_SHOW_TASK_COUNT_STATE_CNT_TIME,
                     pTaskStateStat->VsuspStat,
                     XCONS_SHOW_TASK_COUNT_KT_STATE_VSUSP_CNT  ,
                     XCONS_SHOW_TASK_COUNT_KT_VSUSP_REL_TIME,
                     XCONS_SHOW_TASK_COUNT_KT_VSUSP_ABS_TIME ) ;

      PRI_STAT_TM  ( XCONS_SHOW_TASK_COUNT_STATE_CNT_TIME,
                     pTaskStateStat->VsleepStat,
                     XCONS_SHOW_TASK_COUNT_KT_STATE_VSLEEP_CNT  ,
                     XCONS_SHOW_TASK_COUNT_KT_VSLEEP_REL_TIME,
                     XCONS_SHOW_TASK_COUNT_KT_VSLEEP_ABS_TIME ) ;

      if ( pTaskStateStat->RcvRplStat.ulCount )
        printf ( XCONS_SHOW_TASK_COUNT_RCV_RPL_INFO ,
                     XCONS_SHOW_TASK_COUNT_KT_RCV_RPL_CNT,
                     pTaskStateStat->RcvRplStat.ulCount,
                     XCONS_SHOW_TASK_COUNT_KT_RCV_RPL_TO_LONG,
                     pTaskStateStat->ulRcvRplToLong ,
                     XCONS_SHOW_TASK_COUNT_KT_RCV_RPL_AVG_TIME,
                     AVG_TIME( pTaskStateStat->RcvRplStat )  ) ;

      PRI_STAT_TM  ( XCONS_SHOW_TASK_COUNT_STATE_CNT_TIME,
                     pTaskStateStat->RplRcvStat,
                     XCONS_SHOW_TASK_COUNT_KT_RPL_RCV_CNT  ,
                     XCONS_SHOW_TASK_COUNT_KT_RPL_RCV_REL_TIME,
                     XCONS_SHOW_TASK_COUNT_KT_RPL_RCV_ABS_TIME ) ;
      }

    if ( pTaskPrioStat->ulTotalCount )
       printf ( XCONS_SHOW_TASK_COUNT_PRIO_TOT_OTH,
                      XCONS_SHOW_TASK_COUNT_KT_PRIO_CNT ,
                      pTaskPrioStat->ulTotalCount   ,
                      XCONS_SHOW_TASK_COUNT_KT_PRIO_FROM_OTHER,
                      pTaskPrioStat->ulOtherUkt ) ;

    printf ( XCONS_SHOW_TASK_COUNT_TRAILER ) ;
    }
  return ( NO_ERROR );
  }

/*------------------------------*/

static LONG sql92c_show_queue_info ( PCONSOLE_REC   pConsoleRec )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_show_queue_info"
  INT                              i;
  PCONS_SINGLE_QUEUE_STAT_INFO_REC pQueStatInfo;

  pQueStatInfo  = (PCONS_SINGLE_QUEUE_STAT_INFO_REC)
                      ((PCHAR)pConsoleRec +
                              pConsoleRec->QueInfo.ulQueStatInfoOffset);

  printf ( XCONS_SHOW_QUEUE_STAT_HEADER );

  for ( i = 0; i < pConsoleRec->QueInfo.lNumOfActiveTasks; i++)
    {
    printf ( XCONS_SHOW_QUEUE_STAT_INFO ,
             pQueStatInfo[i].ulTaskIndex,
             pQueStatInfo[i].QueStat.ulUToUQueCount,
             pQueStatInfo[i].QueStat.ulUKTQueCount,
             pQueStatInfo[i].QueStat.ulCOMQueCount ) ;
    }

  return ( NO_ERROR );
  }

/*------------------------------*/

static LONG sql92c_show_ukt_sl_info ( PCONSOLE_REC   pConsoleRec )
{
  #undef  MF__
  #define MF__ MOD__"sql92c_show_ukt_sl_info"
  INT                i;
  char               IdleTimeBuf[20];             /* ddd:hh:mm:ss: */
  char               IOWaitTimeBuf[20];           /* ddd:hh:mm:ss: */
 
  PCONS_SINGLE_UKT_SL_STAT_INFO_REC pSleepStatInfo;

  if ( pConsoleRec->UKTSleepTimeInfo.lNumOfActiveUKTs > 0 )
  {
    pSleepStatInfo = (PCONS_SINGLE_UKT_SL_STAT_INFO_REC)
                         ((PCHAR)pConsoleRec +
                         pConsoleRec->UKTSleepTimeInfo.ulSleepStatInfoOffset);

    printf ( XCONS_SHOW_UKT_SLEEP_STAT_HEADER ) ;

    for ( i = 0; i < pConsoleRec->UKTSleepTimeInfo.lNumOfActiveUKTs; i++)
    {
/* CR 1000097 */
      eo92_SecToDDHHMMSS( pSleepStatInfo[i].UKTIdleStat.ulSec, IdleTimeBuf ) ;

      eo92_SecToDDHHMMSS( pSleepStatInfo[i].UKTIOWaitStat.ulSec, IOWaitTimeBuf ) ; 

      printf( XCONS_SHOW_UKT_SLEEP_STAT_INFO ,
              pSleepStatInfo[i].ulUKTIndex ,
              pSleepStatInfo[i].UKTIdleStat.ulCount,
              IdleTimeBuf,
              (float) pSleepStatInfo[i].UKTIdleStat.ulSec * 100 /
              (float) pConsoleRec->UKTSleepTimeInfo.ulTotalTimeMeasureSec,
              pSleepStatInfo[i].UKTIOWaitStat.ulCount,
              IOWaitTimeBuf,
              (float) pSleepStatInfo[i].UKTIOWaitStat.ulSec * 100 /
              (float) pConsoleRec->UKTSleepTimeInfo.ulTotalTimeMeasureSec,
              pSleepStatInfo[i].ulRUNQueueLen );
    }
  }

  return ( NO_ERROR );
}

#ifndef RUN_AS_SYSMON

/*------------------------------*/

static LONG sql92c_show_ukt_queues ( PCONSOLE_REC   pConsoleRec )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_show_ukt_queues"
  PCONS_SINGLE_UKT_QUEUES_INFO_REC  pUKTQueues;
  ULONG                             ulUKT;


  if ( pConsoleRec->UKTQueues.lNumOfActiveUKTs > 0 )
    {
    pUKTQueues = (PCONS_SINGLE_UKT_QUEUES_INFO_REC)
                         ((PCHAR)pConsoleRec +
                         pConsoleRec->UKTQueues.ulUKTQueuesOffset);

    for ( ulUKT = 0; ulUKT < (ULONG)pConsoleRec->UKTQueues.lNumOfActiveUKTs;
          ulUKT++, pUKTQueues++ )
      if ( pUKTQueues->ulUKTQueueLen ||
           pUKTQueues->ulCOMQueueLen ||
           pUKTQueues->ulRUNQueueLen )
      {
        printf ( XCONS_SHOW_QUEUES_HEAD,
                         pUKTQueues->ulUKTIndex ,
                         pUKTQueues->Tid ) ;

        sql92c_show_UKT_queue ( pUKTQueues->ulUKTQueueLen,
                               (PCONS_SINGLE_QUEUE_INFO_REC )
                                ((PCHAR)pConsoleRec +
                                      pUKTQueues->ulUKTQueueOffset)) ;

        sql92c_show_COM_queue ( pUKTQueues->ulCOMQueueLen,
                               (PCONS_SINGLE_QUEUE_INFO_REC )
                                ((PCHAR)pConsoleRec +
                                      pUKTQueues->ulCOMQueueOffset)) ;

        sql92c_show_RUN_queue ( pUKTQueues->ulRUNQueueLen,
                               (PCONS_SINGLE_QUEUE_INFO_REC )
                                ((PCHAR)pConsoleRec +
                                      pUKTQueues->ulRUNQueueOffset)) ;
      }
    }

  return ( NO_ERROR );
  }

/*------------------------------*/


static LONG sql92c_show_UKT_queue ( ULONG ulUKTQueueLen,
                                    PCONS_SINGLE_QUEUE_INFO_REC pUKTUKTQueue)
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_show_UKT_queue"

  ULONG                             ulQueuElem;

  if ( ulUKTQueueLen > 0 )
    { printf ( XCONS_SHOW_UKT_QUEUE_HEAD ) ;
      for ( ulQueuElem = 0;
            ulQueuElem < ulUKTQueueLen ;
            ulQueuElem++, pUKTUKTQueue++ )
               if ( pUKTUKTQueue->PrioFlag )
                 printf( XCONS_SHOW_UKT_QUEUE_PRIO_ELEMENT,
                          pUKTUKTQueue->ulTaskIndex ,
                          pUKTUKTQueue->PrioFlag ) ;
               else
                 printf( XCONS_SHOW_UKT_QUEUE_ELEMENT,
                          pUKTUKTQueue->ulTaskIndex ) ;
    }

  return ( NO_ERROR );
  }


/*------------------------------*/


static LONG sql92c_show_COM_queue ( ULONG ulCOMQueueLen,
                                    PCONS_SINGLE_QUEUE_INFO_REC pUKTCOMQueue)
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_show_COM_queue"

  ULONG                             ulQueuElem;

  if ( ulCOMQueueLen > 0 )
    { printf ( XCONS_SHOW_COM_QUEUE_HEAD ) ;
      for ( ulQueuElem = 0;
            ulQueuElem < ulCOMQueueLen ;
            ulQueuElem++, pUKTCOMQueue++ )
               if ( pUKTCOMQueue->PrioFlag )
                 printf( XCONS_SHOW_COM_QUEUE_PRIO_ELEMENT,
                            pUKTCOMQueue->ulTaskIndex ,
                            pUKTCOMQueue->PrioFlag ) ;
               else
                 printf( XCONS_SHOW_COM_QUEUE_ELEMENT,
                            pUKTCOMQueue->ulTaskIndex ) ;
    }

  return ( NO_ERROR );
  }


/*------------------------------*/


static LONG sql92c_show_RUN_queue ( ULONG ulRUNQueueLen,
                                    PCONS_SINGLE_QUEUE_INFO_REC pUKTRUNQueue)
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_show_RUN_queue"

  ULONG                             ulQueuElem;
  CHAR                              szFromTask[10] ;

  if ( ulRUNQueueLen > 0 )
    { printf ( XCONS_SHOW_RUN_QUEUE_HEAD ) ;
      for ( ulQueuElem = 0;
            ulQueuElem < ulRUNQueueLen ;
            ulQueuElem++, pUKTRUNQueue++ )
        {   if ( pUKTRUNQueue->ulFromTask != UNDEF )
              sprintf(szFromTask, "(T%d)", pUKTRUNQueue->ulFromTask ) ;
            else
              szFromTask[0] = '\0' ;
            if ( pUKTRUNQueue->ulReqType != UNDEF )
               printf( XCONS_SHOW_RUN_QUEUE_ELEMENT_REQ,
                          pUKTRUNQueue->ulTaskIndex ,
                          eo92GetTaskStateName ( pUKTRUNQueue->TaskState,0 ),
                          pUKTRUNQueue->ulDynPrio ,
                          sql92c_req_type_name ( pUKTRUNQueue->ulReqType ),
                          szFromTask) ;
           else
               printf( XCONS_SHOW_RUN_QUEUE_ELEMENT,
                          pUKTRUNQueue->ulTaskIndex ,
                          eo92GetTaskStateName ( pUKTRUNQueue->TaskState,0 ),
                          pUKTRUNQueue->ulDynPrio , szFromTask) ;
        }
    }

  return ( NO_ERROR );
  }


/*------------------------------*/

static LONG sql92c_show_suspends ( PCONSOLE_REC  pConsoleRec )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_show_suspends"
  INT             i;
  ULONG           ulTotalSuspends = 0 ;

  printf ( XCONS_SHOW_SUSPEND_HEADER );

  for ( i=0; i<MAX_SUSPEND_REASONS ; i++ )
    ulTotalSuspends += pConsoleRec->SuspendReasons.ulSuspends [ i ] ;

  printf ( XCONS_SHOW_SUSPEND_TOTAL , ulTotalSuspends ) ;

  for ( i = 0; i < MAX_SUSPEND_REASONS ; i++ )
    {
    if ( pConsoleRec->SuspendReasons.ulSuspends [ i ] )
      {
      printf( XCONS_SHOW_SUSPEND_CNT_PERCENT,
                SuspendReasonText[i] . szShortReason ,
                pConsoleRec->SuspendReasons.ulSuspends [ i ] ,
                (float ) pConsoleRec->SuspendReasons.ulSuspends [ i ] * 100.0 /
                (float ) ulTotalSuspends ,
                SuspendReasonText[i] . szLongReason ) ;
      }
    }

  return ( NO_ERROR );
  }

/*------------------------------*/


static LONG sql92c_show_param ( PSZ pszParam , PCONSOLE_REC pConsoleRec )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_show_param"

  e92_print_param ( &(pConsoleRec->GetParamInfo.XParam) , pszParam );

  return ( NO_ERROR );
  }

#endif

/*------------------------------*/

static LONG sql92c_show_state_info ( PCONSOLE_REC   pConsoleRec )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_show_storage_info"
  #ifndef RUN_AS_SYSMON
   PSZ      pszStateStr;
  #endif
  DBGPAS;

  #ifndef RUN_AS_SYSMON

   switch (pConsoleRec->StateInfo.DatabaseState)
     {
     case SERVER_STARTING      : pszStateStr = XCONS_SHOW_STATE_STARTING ; break;
     case SERVER_COLD          : pszStateStr = XCONS_SHOW_STATE_COLD     ; break;
     case SERVER_WARM          : pszStateStr = XCONS_SHOW_STATE_WARM     ; break;
     case SERVER_SHUTDOWN      : pszStateStr = XCONS_SHOW_STATE_SHTDWN   ; break;
     case SERVER_SHUTDOWNREINIT: pszStateStr = XCONS_SHOW_STATE_SHTDWNREI; break;
     case SERVER_SHUTDOWNKILL  : pszStateStr = XCONS_SHOW_STATE_SHTDWNKIL; break;
     case SERVER_STOP          : pszStateStr = XCONS_SHOW_STATE_STOP     ; break;
     case SERVER_KILL          : pszStateStr = XCONS_SHOW_STATE_KILL     ; break;
     case SERVER_ABORT         : pszStateStr = XCONS_SHOW_STATE_ABORT    ; break;
     default                   : pszStateStr = XCONS_SHOW_STATE_UNKNOWN  ; break;
     }

   if ( pConsoleRec->StateInfo.ulTerminationTimeOut != 0 )
     {
     printf ( XCONS_SHOW_STATE_INFO_TERM_TIMEOUT, pszStateStr,
              pConsoleRec->StateInfo.ulTerminationTimeOut );
     }
   else
     printf ( XCONS_SHOW_STATE_INFO, pszStateStr );
   

  #endif

  return ( NO_ERROR );
  }


/*------------------------------*/

static PSZ sql92c_kt_state_name ( ULONG ulState )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_kt_state_name"

  DBGPAS;

  switch ( ulState )
    {
    case KT_INITIALIZED:    return ( XCONS_KT_STATE_INITIALIZED );
    case KT_RUNNING    :    return ( XCONS_KT_STATE_RUNNING     );
    case KT_TERMINATING:    return ( XCONS_KT_STATE_TERMINATING );
    case KT_TERMINATED :    return ( XCONS_KT_STATE_TERMINATED  );
    case KT_SLEEPING   :    return ( XCONS_KT_STATE_SLEEPING    );
    case KT_DETACHED   :    return ( XCONS_KT_STATE_DETACHED    );
    case KT_ATTACHED   :    return ( XCONS_KT_STATE_ATTACHED    );
    }

  return ( XCONS_KT_STATE_UNKNOWN );
  }

/*------------------------------*/

static PSZ sql92c_ukt_state_name ( ULONG ulState )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_ukt_state_name"

  DBGPAS;

  switch ( ulState )
    {
    case UKT_INITIALIZED:    return ( XCONS_UKT_STATE_INITIALIZE );
    case UKT_RUNNING    :    return ( XCONS_UKT_STATE_RUNNING    );
    case UKT_TERMINATED :    return ( XCONS_UKT_STATE_TERMINATED );
    case UKT_SLEEPING   :    return ( XCONS_UKT_STATE_SLEEPING   );
    }

  return ( XCONS_UKT_STATE_UNKNOWN );
  }

/*------------------------------*/

static PSZ  sql92c_req_type_name ( ULONG ulReqType )
  {
  #undef  MF__
  #define MF__ MOD__"sql92c_req_type_name"

  DBGPAS;

  switch ( ulReqType )
    {
      case REQ_VATTACH                 : return ( XCONS_REQ_VATTACH ) ;
      case REQ_VBLOCKIO                : return ( XCONS_REQ_VBLOCKIO ) ;
      case REQ_VDETACH                 : return ( XCONS_REQ_VDETACH ) ;
      case REQ_INITOK                  : return ( XCONS_REQ_INITOK ) ;
      case REQ_VRESUME                 : return ( XCONS_REQ_VRESUME ) ;
      case REQ_VSIGNAL                 : return ( XCONS_REQ_VSIGNAL ) ;
      case REQ_CONNECT                 : return ( XCONS_REQ_CONNECT );
      case REQ_RELEASE                 : return ( XCONS_REQ_RELEASE ) ;
      case REQ_VRECEIVE                : return ( XCONS_REQ_VRECEIVE) ;
      case REQ_SHUTDOK                 : return ( XCONS_REQ_SHUTDOK ) ;
      case REQ_SENDER_JOB_FREE         : return ( XCONS_REQ_SENDER_JOB_FREE) ;
      case REQ_VENDEXCL                : return ( XCONS_REQ_VENDEXCL ) ;
      case REQ_RESCHEDULE              : return ( XCONS_REQ_RESCHEDULE ) ;
      case REQ_SHUTDOWN_RESUME_TW      : return ( XCONS_REQ_SHUTDOWN_RESUME_TW);
      case REQ_ASYNOPEN                : return ( XCONS_REQ_ASYNOPEN ) ;
      case REQ_ASYNCLOSE               : return ( XCONS_REQ_ASYNCLOSE ) ;
      case REQ_ASYNIO                  : return ( XCONS_REQ_ASYNIO ) ;
      case REQ_ASYNWAIT                : return ( XCONS_REQ_ASYNWAIT ) ;
      case REQ_VNSHUTDOWN_REPLY        : return ( XCONS_REQ_VNSHUTDOWN_REPLY) ;
      case REQ_OUTGOING_CONNECT        : return ( XCONS_REQ_OUTGOING_CONNECT) ;
      case REQ_OUTGOING_CONNECT_REPLY  : return ( XCONS_REQ_OUTGOING_CONNECT_REPLY) ;
      case REQ_OPEN_WAKEUP_SD          : return ( XCONS_REQ_OPEN_WAKEUP_SD ) ;
      case REQ_CONNECT_RC              : return ( XCONS_REQ_CONNECT_RC ) ;
      case REQ_RELEASE_RC              : return ( XCONS_REQ_RELEASE_RC ) ;
      case REQ_CLOSE_OUTGOING          : return ( XCONS_REQ_CLOSE_OUTGOING ) ;
      case REQ_CLOSE_OUTGOING_REPLY    : return ( XCONS_REQ_CLOSE_OUTGOING_REPLY) ;
      case REQ_FINISH_SV               : return ( XCONS_REQ_FINISH_SV ) ;
      case REQ_ASYNCNTL                : return ( XCONS_REQ_ASYNCNTL ) ;
   }


  return ( XCONS_REQ_TYPE_UNKNOWN );
  }


/*------------------------------*/

#if defined (CTRLCOMP)
/**\
--------------------------------------- MF__ MOD__ (controlling)
\**/

/**\
--------------------------------------- MF__ MOD__ (monitoring)
\**/

typedef struct KernelMonitorT {
    teo5_monitor_object super;
    XCONS_REC           consoleRec;
    int                 requestTag;
    PBOOL               pfFlag;
} KernelMonitorT;

tsp9_cstr os903dbname (void);

    #undef MF__
    #define MF__ MOD__"allocMonitor_MF"
static KernelMonitorT *
allocMonitor_MF (
    tsp9_rte_xerror* xerror)
{
    KernelMonitorT * result;
	LONG             rc;

    rc = ALLOC_MEM ((PPVOID)&result, sizeof (KernelMonitorT) );

    if( rc != NO_ERROR )
	{
        eo44anyError (xerror, "Out of memory");
        return NULL;
    }
    result->consoleRec.pszServerDB = (char*) os903dbname ();
    if ( sql92c_init (&result->consoleRec) != NO_ERROR ) {
        FREE_MEM ( result );
        eo44anyError (xerror, "Allocation of shared memory failed");
        return NULL;
    }
    return result;
}

    #undef MF__
    #define MF__ MOD__"monitorTasks_MF"
void
monitorTasks_MF (
    KernelMonitorT* processes,
    tsp9_rte_readdirection flag,
    char* outbuf,
    int bufSize,
    tsp9_rte_xerror* xerror)
{
}

    #undef MF__
    #define MF__ MOD__"os92allTasks"
teo5_monitor_object*
os92allTasks (
    tsp9_cstr args,
    int flags,
    tsp9_rte_xerror* xerror)
{
    KernelMonitorT* kernelMonitor;

    kernelMonitor = allocMonitor_MF (xerror);
    if (kernelMonitor == NULL) {
        return NULL;
    }
    USEVAR_LINT (args);
    USEVAR_LINT (flags);
    kernelMonitor->super.nextRecord = (teo5_monitor_function*)monitorTasks_MF;
    kernelMonitor->super.description = "";
    return &kernelMonitor->super;
}

    #undef MF__
    #define MF__ MOD__"showState_MF"
static void
showState_MF (
    KernelMonitorT* state,
    tsp9_rte_readdirection direction,
    char* outbuf,
    int bufSize,
    tsp9_rte_xerror* xerror)
{
    tsp9_cstr result;

    USEVAR_LINT (direction);
    if (!state->super.started) {
        state->super.started = TRUE;
        switch (state->consoleRec.pConsoleRec->StateInfo.DatabaseState) {
            case SERVER_STARTING:
                result = XCONS_SHOW_STATE_STARTING ;
                break;
            case SERVER_COLD:
                result = XCONS_SHOW_STATE_COLD;
                break;
            case SERVER_WARM:
                result = XCONS_SHOW_STATE_WARM;
                break;
            case SERVER_SHUTDOWN:
                result = XCONS_SHOW_STATE_SHTDWN;
                break;
            case SERVER_SHUTDOWNREINIT:
                result = XCONS_SHOW_STATE_SHTDWNREI;
                break;
            case SERVER_SHUTDOWNKILL:
                result = XCONS_SHOW_STATE_SHTDWNKIL;
                break;
            case SERVER_STOP:
                result = XCONS_SHOW_STATE_STOP;
                break;
            case SERVER_KILL:
                result = XCONS_SHOW_STATE_KILL;
                break;
            case SERVER_ABORT:
                result = XCONS_SHOW_STATE_ABORT;
                break;
            default:
                fprintf (stdout, "state id %d; info available %s\n",
                    state->consoleRec.pConsoleRec->StateInfo.DatabaseState,
                    state->consoleRec.pConsoleRec->StateInfo.fInfoAvailable
                        ? "True" : "False");
                result = XCONS_SHOW_STATE_UNKNOWN;
                break;
         }
         if ((int) strlen (result) + 1 <= bufSize) {
             strcpy (outbuf, result);
        }
        else {
            eo44anyError (xerror, "buffer too small");
        }
    }
    else {
        eo44eoshowError (xerror);
    }
}

    #undef MF__
    #define MF__ MOD__"os92state"
teo5_monitor_object*
os92state (
    tsp9_cstr args,
    int flags,
    tsp9_rte_xerror* xerror)
{
    KernelMonitorT* result;

    result = allocMonitor_MF (xerror);
    if (result == NULL) {
        return NULL;
    }
    USEVAR_LINT (args);
    USEVAR_LINT (flags);
    result->super.nextRecord = (teo5_monitor_function*)showState_MF;
    result->super.description = "State";
    result->requestTag = CON_REQ_STATE;
    result->pfFlag = &result->consoleRec.pConsoleRec->StateInfo.fInfoAvailable;
    return &result->super;
}

    #undef MF__
    #define MF__ MOD__"os92getRemoteInfo"
void
os92getRemoteInfo (
    teo5_monitor_object* monitorHandle,
    tsp9_rte_xerror* xerror)
{
    int                 rc;
    KernelMonitorT     *monitor = (KernelMonitorT*) ((char*)monitorHandle - OFFSET_OFF(KernelMonitorT, super));
    XCONS_REC          *pxcr = &monitor->consoleRec;

    memset ( pxcr->pConsoleRec, 0, pxcr->ulShmSize );
    rc = sql92c_send (pxcr, monitor->requestTag,
        monitor->pfFlag);
    if (rc != NO_ERROR) {
        eo44anyError (xerror, "Exchange of shared memory failed");
    }
}

    #undef MF__
    #define MF__ MOD__"sqlxknlclosemonitor"
void
sqlxknlclosemonitor (
    void* monitorHandle)
{
    KernelMonitorT     *monitor = (KernelMonitorT*) ((char*)monitorHandle - OFFSET_OFF(KernelMonitorT, super));
    XCONS_REC          *pxcr = &monitor->consoleRec;

	sql92c_release( pxcr );
    FREE_MEM ( monitor );
}
#endif


/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/