/*!================================================================
 module:    vpr01Cursor.c

 responsible: BurkhardD
 
 special area:   CPCDrv | Client | Cursorverwaltung
 
 description: Cursor Verwaltung
 
 see:  
 
 change history:

        2001-06-22 D025086 PTS 1110677 Close the Cursor to drop old ParseIds
        before open a new one

    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 ===================================================================*/

/*==================================================================*
 *  INCLUDES                                                        *
 *==================================================================*/

#include "vpr07Macro.h"
#include "vpr01Cursor.h"
#include "vpr01Con.h"
#include "vpr01Error.h"
#include "vpr01StmtName.h"
#include "gpr03.h"
#include "gpr03m.h"
#include "gmt90.h"

/*==================================================================*
 *  DECLARATIONS                                                    *
 *==================================================================*/

/*==================================================================*
 *  EXTERNAL PROTOTYPES                                             *
 *==================================================================*/

/*==================================================================*
 *  FUNCTION PROTOTYPES                                             *
 *==================================================================*/

static void pr01CursorInitCont(tpr01_CursorContainer *Cont);

static tsp00_Int4 pr01CursorOpenCont(tpr01_CursorContainer *Cursor);

/*static tsp00_Int4 pr01CursorGetCount(tpr01_CursorContainer *Cursor);*/

static void pr01CursorCloseCont(tpr01_CursorContainer *Cursor);

static tpr01_CursorDesc *pr01CursorEnumDesc(tpr01_CursorContainer *Cursor);

static tpr01_CursorDesc *pr01CursorAddDesc(tpr01_CursorContainer *Cursor, tpr05_String *CursorName, tpr01_UniqueID UniqueID, tpr01_ConDesc *ConDesc);

static void pr01CursorDeleteDesc(tpr01_CursorDesc *CursorDesc);

static tpr01_CursorDesc *pr01CursorFindDesc(tpr01_CursorContainer *Cursor, tpr05_String* CursorName, tpr01_UniqueID UniqueID);

static void pr01CursorResetIterator(tpr01_CursorContainer *Cursor);

static void pr01CursorExecute(tpr01_CursorDesc *CursorDesc);

static tpr01_CursorDesc *pr01CursorDeclare (tpr01_CursorContainer *Cursor, tpr01_StmtNameDesc *StmtNameDesc, tpr01_ConDesc *ConDesc, tpr05_String *CursorName, tpr01_UniqueID UniqueID);

static void pr01CursorPrepareOpen(tpr01_CursorDesc *CursorDesc, sqlkaentry *ka);
static void pr01CursorPrepareFetch(tpr01_CursorDesc *CursorDesc, struct tpr01_StmtNameDesc *StmtNameDesc, boolean tryexec);
static void pr01CursorPrepareClose(tpr01_CursorDesc *CursorDesc, sqlkaentry *ka);
static void pr01CursorClose(tpr01_CursorDesc *CursorDesc);

static void pr01CursorAlterCursorName (tpr01_CursorDesc *CursorDesc,  tpr05_String* CursorName);

static tpr01_ConDesc *pr01CursorGetConnection(tpr01_CursorDesc *CursorDesc);

static BOOLEAN pr01CursorSetAttr (tpr01_CursorDesc *CursorDesc, epr01Cursor_Attr Attr, 
				  void *pAttr, tsp00_Int4 cbLen);

static void pr01CursorGetAttr (tpr01_CursorDesc *CursorDesc, epr01Cursor_Attr Attr, 
			       void *pAttr, tsp00_Int4 *cbLen);

static void pr01Cursor_InitCursorName (tpr00_CursorNameDesc *CursorNameDesc);

static tpr09HashValue pr01Cursor_GetKey (const tpr09HashValue data_p, tsp00_Uint4* HashKeyLen);

static tsp00_Int4 pr01Cursor_CompareKey (const tpr09HashKey Item1,tsp00_Uint4 Item1Len, const tpr09HashKey Item2, tsp00_Uint4 Item2Len);

static void pr01Cursor_printKey (const tpr09HashValue data_p, FILE* datei);

/*==================================================================*
 *  CODE                                                            *
 *==================================================================*/

static void pr01CursorInitCont(tpr01_CursorContainer *Cursor)
{
  Cursor->Desc = pr09NewHashDescriptor(20, 
                                       sizeof(tpr01_CursorDesc), 
                                       pr01Cursor_CompareKey, 
                                       0, 
                                       pr01Cursor_GetKey,
                                       pr01Cursor_printKey);

  Cursor->ErrorDesc = pr01NewErrorDesc();
  Cursor->InitCont  = pr01CursorInitCont;
  Cursor->OpenCont  = pr01CursorOpenCont;
  Cursor->CloseCont = pr01CursorCloseCont;
/*  Cursor->GetCount  = NULL; */
  Cursor->EnumDesc  = pr01CursorEnumDesc;
  Cursor->AddDesc   = pr01CursorAddDesc;
  Cursor->DeleteDesc = pr01CursorDeleteDesc;
  Cursor->ResetIterator = pr01CursorResetIterator;
  Cursor->FindDesc  = pr01CursorFindDesc;
  Cursor->Execute   = pr01CursorExecute;
  Cursor->Declare   = pr01CursorDeclare;
  Cursor->Close     = pr01CursorClose;
  Cursor->PrepareOpen     = pr01CursorPrepareOpen;
  Cursor->PrepareFetch    = pr01CursorPrepareFetch;
  Cursor->PrepareClose    = pr01CursorPrepareClose;
  Cursor->AlterCursorName = pr01CursorAlterCursorName;
  Cursor->GetConnection   = pr01CursorGetConnection;
  Cursor->InitCursorName  = pr01Cursor_InitCursorName; 
}

void pr01CursorDeleteCont(tpr01_CursorContainer *Cursor)
{
  if (Cursor) {
    if (pr01CursorOpenCont(Cursor)) {
      tpr01_CursorDesc *CursorDesc;
      while(CursorDesc = Cursor->EnumDesc(Cursor))
	Cursor->DeleteDesc(CursorDesc);
    }
    pr01DeleteErrorDesc(Cursor->ErrorDesc);    
    pr09DeleteHashDescriptor(Cursor->Desc);
    Cursor->Desc = NULL;
    pr03mFree((tsp00_Uint1*)Cursor, "Cursor");
  }
}

tpr01_CursorContainer *pr01CursorNewCont()
{
  tpr01_CursorContainer *Cursor;
  Cursor = pr03mAllocat(sizeof(tpr01_CursorContainer), "Cursor");
  if (Cursor)
    pr01CursorInitCont(Cursor);
  return Cursor;
}


static tsp00_Int4 pr01CursorOpenCont(tpr01_CursorContainer *Cursor)
{
/*  if (!Cursor->Desc) {
    Cursor->Desc = pr09NewDescriptor(10, sizeof(tpr01_CursorDesc));
  }
*/
    if (Cursor->Desc) {
    pr09ResetHashItemIterator(Cursor->Desc);
    return 1;
  } else {
    return 0;
  }
}

/*
static tsp00_Int4 pr01CursorGetCount(tpr01_CursorContainer *Cursor)
{
  return(pr09GetItemCount(Cursor->Desc));
}
*/
static tpr01_CursorDesc *pr01CursorEnumDesc(tpr01_CursorContainer *Cursor)
{
  tpr01_CursorDesc *CursorDesc=pr09HashItemIteratorNextItem(Cursor->Desc);
  return(CursorDesc);
}

static void pr01CursorCloseCont(tpr01_CursorContainer *Cursor)
{
  M90TRACE(M90_TR_HANDLE, "pr01CursorCloseCont", Cursor);
  pr09DeleteAllHashItems(Cursor->Desc);
}

void pr01CursorInitAttr (tpr01_CursorDesc *CursorDesc)
{
  BOOLEAN vfalse = FALSE;
  BOOLEAN vtrue = TRUE;
  CursorDesc->SetAttr (CursorDesc, openkeep_epr01Cursor, &vfalse, 1);
  CursorDesc->SetAttr (CursorDesc, useNewParseInfo_epr01Cursor, &vtrue, 1);
  CursorDesc->SetAttr (CursorDesc, lastDataReceived_epr01Cursor, &vfalse, 1);
}

static tpr01_CursorDesc *pr01CursorAddDesc(tpr01_CursorContainer *Cursor, tpr05_String *CursorName, tpr01_UniqueID UniqueID, tpr01_ConDesc *ConDesc)
{
  M90TRACE(M90_TR_ENTRY, "pr01CursorAddDesc", 0);
  {
    tpr09_HashTableError hTabErr; 
    tpr01_CursorDesc *CursorDesc;      
    CursorDesc = pr09GetNewHashItem(Cursor->Desc);
    CursorDesc->DescType = CursorDesc_epr01;
    CursorDesc->Cursor=Cursor;
    CursorDesc->ConDesc      = ConDesc;
    CursorDesc->StmtNameDesc = NULL;
    CursorDesc->FetchStmtNameDesc = NULL;
    CursorDesc->OpenKa       = NULL;
    CursorDesc->FetchKa      = NULL;
    CursorDesc->CloseKa      = NULL;
    CursorDesc->ErrorDesc    = pr01NewErrorDesc();
    CursorDesc->cue = pr03mAllocat(sizeof(sqlcuentry), "CursorDesc->cu");
    pr03cCuInit(CursorDesc->cue);
    CursorDesc->UniqueID = UniqueID;
    Cursor->InitCursorName(&(CursorDesc->CursorNameDesc));
    if (UniqueID != EmptyUniqueID_cpr01) {
      /*generate an implizit cursor name*/
      CursorDesc->CursorNameDesc.CursorName.cbLen = 
      sp77sprintfUnicode (CursorDesc->CursorNameDesc.CursorName.encodingType,
                          CursorDesc->CursorNameDesc.CursorName.rawString,
                          CursorDesc->CursorNameDesc.CursorName.cbMaxLen,
                          ":%#08lx", 
                          UniqueID);
    }
    else {
      /*take the given cursor name*/  
      tpr05_StringError erg = pr05IfCom_String_Convert (&(CursorDesc->CursorNameDesc.CursorName),
                                                        CursorName );
    }
       /*add the descriptor to the cursor container hash index*/
    hTabErr = pr09AddHashItemToHash(Cursor->Desc, 
                                    CursorDesc->CursorNameDesc.CursorName.rawString, 
                                    CursorDesc->CursorNameDesc.CursorName.cbLen,
                                    (tpr09HashValue*)CursorDesc);

    CursorDesc->SetAttr = pr01CursorSetAttr;
    CursorDesc->GetAttr = pr01CursorGetAttr;
    pr01CursorInitAttr (CursorDesc);
    /* grep shows that CursorDesc->cue->cuname is no longer required
      pr07C2P(CursorDesc->cue->cuname, szCursorName, sizeof(CursorDesc->cue->cuname));*/
    
/*pr09HashDump(Cursor->Desc, "CursorHashDump.txt");    */
      
    M90TRACE(M90_TR_EXIT, "pr01CursorAddDesc", 0);
    return(CursorDesc);
  }
}

static void pr01CursorDeleteDesc(tpr01_CursorDesc *CursorDesc)
{
  if (NULL != pr09FindHashItem(CursorDesc->Cursor->Desc, 
                               CursorDesc->CursorNameDesc.CursorName.rawString, 
                               CursorDesc->CursorNameDesc.CursorName.cbLen)) {  
    tpr05_String* cName = &(CursorDesc->CursorNameDesc.CursorName);                            
    pr01DeleteErrorDesc(CursorDesc->ErrorDesc);    
    if (CursorDesc->cue) {
      pr03mFree(CursorDesc->cue, "CursorDesc->cue");
    }
    /*delete cursor name and release memory*/
    pr09DeleteHashItem(CursorDesc->Cursor->Desc, 
                       CursorDesc->CursorNameDesc.CursorName.rawString, 
                       CursorDesc->CursorNameDesc.CursorName.cbLen);
    if (cName && cName->rawString) pr05IfCom_String_DeleterawString(cName);
  }
}

static void pr01CursorResetIterator(tpr01_CursorContainer *Cont){
  if (Cont && Cont->Desc) pr09ResetHashItemIterator(Cont->Desc);
}

static tpr01_CursorDesc *pr01CursorFindDesc(tpr01_CursorContainer *Cursor, tpr05_String *CursorName, tpr01_UniqueID UniqueID)
{
  M90TRACE(M90_TR_ENTRY, "pr01CursorFindDesc", 0);
  {
     tpr01_CursorDesc *CursorDesc = NULL;

    if ( UniqueID != EmptyUniqueID_cpr01) {
      CursorName->cbLen = sp77sprintfUnicode (CursorName->encodingType,
                                              CursorName->rawString,
                                              CursorName->cbMaxLen,
                                              ":%#08lx", 
                                              UniqueID);      
    }
  
  
    CursorDesc = (tpr01_CursorDesc *)pr09FindHashItem(  Cursor->Desc, 
                                                          CursorName->rawString, 
                                                          CursorName->cbLen);    
  
      
    M90TRACE(M90_TR_EXIT, "pr01CursorFindDesc", 0);
    M90TRACE(M90_TR_HANDLE, "CursorDesc", &CursorDesc);
    return(CursorDesc);    
  }
}

static void pr01CursorExecute(tpr01_CursorDesc *CursorDesc)
{
  CursorDesc->ErrorDesc->Error->SetError(CursorDesc->ErrorDesc, 1, "UNKNOWN", CPR_NTS);
  return;
}

static void pr01CursorClose(tpr01_CursorDesc *CursorDesc)
{
  M90TRACE(M90_TR_ENTRY, "pr01CursorClose", 0);
  M90TRACE(M90_TR_HANDLE, "CursorDesc", &CursorDesc);
  {
    BOOLEAN myVal = FALSE;
    M90TRACE(M90_TR_HANDLE, "StmtNameDesc", &CursorDesc->StmtNameDesc);
    if (CursorDesc->StmtNameDesc) 
    {
      tpr01_CursorContainer *Cursor = CursorDesc->Cursor;
      if (CursorDesc->OpenKa) 
      {
        pr01cFreeKa(CursorDesc->OpenKa);
        CursorDesc->OpenKa = NULL; 
      }
      if (CursorDesc->FetchKa) 
      {
        /* drop parseid  PTS 1106501 */
        sqlcatype *sqlca = CursorDesc->ConDesc->sqlca;
        sqlxatype *sqlxa = CursorDesc->ConDesc->sqlxa;
        sqlgaentry *ga   = CursorDesc->ConDesc->ga;
        sqlkaentry *ka   = CursorDesc->FetchKa;  
        p01bdrop_parsid (sqlca, sqlxa, ka->kaParseInfo.ParseId);
        pr06ParseIdDrop(sqlca, sqlxa, ga, (tpr06_ParseIDCache *)CursorDesc->ConDesc->ParseIdCache, &ka->kaParseInfo.ParseId);
        pr01cFreeKa(CursorDesc->FetchKa);
	CursorDesc->FetchKa = NULL;
      }
      if (CursorDesc->CloseKa) 
      {
	    CursorDesc->CloseKa = NULL;
      }
    }
    /*reset cursor attribut lastDataReceived to false*/
    CursorDesc->SetAttr (CursorDesc, lastDataReceived_epr01Cursor, &myVal, 1);
  }
  M90TRACE(M90_TR_EXIT, "pr01CursorClose", 0);
}

static void pr01CursorPrepareOpen(tpr01_CursorDesc *CursorDesc, sqlkaentry *ka)
{
  BOOLEAN keep;

  M90TRACE(M90_TR_ENTRY, "pr01CursorPrepareOpen", 0);
  PR07_CHECKDESC(CursorDesc, CursorDesc_epr01);
  {
    tpr01_CursorContainer *Cursor = CursorDesc->Cursor;
    /* PTS 1110677 */
    Cursor->Close(CursorDesc);
    if (!CursorDesc->OpenKa) {
      CursorDesc->OpenKa = pr01cDupKa(ka);
      if (CursorDesc->StmtNameDesc) {
	tpr01_StmtNameDesc *StmtNameDesc = CursorDesc->StmtNameDesc;
	if (StmtNameDesc->ka->kapacount > 0) {
	  /* Command with :hostvars */
	  M90TRACE(M90_TR_WORD, "kapacount", &StmtNameDesc->ka->kapacount);
	  M90TRACE(M90_TR_WORD, "kapaindex", &StmtNameDesc->ka->kapaindex);
	  CursorDesc->OpenKa->kapacount = StmtNameDesc->ka->kapacount;
	  CursorDesc->OpenKa->kapaindex = StmtNameDesc->ka->kapaindex;
	}
	CursorDesc->OpenKa->kafaindex = StmtNameDesc->ka->kafaindex;
        if (StmtNameDesc->ka->kafae) {
          memcpy(CursorDesc->OpenKa->kafae, StmtNameDesc->ka->kafae, sizeof(sqlfaentry));
        }
      }
    }
    if (ka->katyp == cpr_com_sql_open_keep) {
      keep = TRUE;
      CursorDesc->SetAttr (CursorDesc, openkeep_epr01Cursor, &keep, sizeof (keep));
    } else {
      keep = FALSE;
      CursorDesc->SetAttr (CursorDesc, openkeep_epr01Cursor, &keep, sizeof (keep));
    }
  }
  M90TRACE(M90_TR_EXIT, "pr01CursorPrepareOpen", 0);
}

static void pr01CursorPrepareFetch(tpr01_CursorDesc *CursorDesc, struct tpr01_StmtNameDesc *StmtNameDesc, boolean tryexec)
{
  M90TRACE(M90_TR_ENTRY, "pr01CursorPrepareFetch", 0);
  PR07_CHECKDESC(CursorDesc, CursorDesc_epr01);
  PR07_CHECKDESC(StmtNameDesc, StmtNameDesc_epr01);
  {
    if (!CursorDesc->FetchKa) {
      CursorDesc->FetchKa = pr01cDupKa(StmtNameDesc->ka);
    }
    if (CursorDesc->FetchStmtNameDesc != StmtNameDesc) {
        pr01cFreeKa(CursorDesc->FetchKa);
        CursorDesc->FetchKa = pr01cDupKa(StmtNameDesc->ka);
    }
    CursorDesc->FetchStmtNameDesc = StmtNameDesc;
    CursorDesc->FetchKa->kacuindex = CursorDesc->OpenKa->kacuindex;
  }
  M90TRACE(M90_TR_EXIT, "pr01CursorPrepareFetch", 0);
}
  
static void pr01CursorPrepareClose(tpr01_CursorDesc *CursorDesc, sqlkaentry *ka)
{
  M90TRACE(M90_TR_ENTRY, "pr01CursorPrepareClose", 0);
  PR07_CHECKDESC(CursorDesc, CursorDesc_epr01);
  {
    CursorDesc->CloseKa = ka;
  }
  M90TRACE(M90_TR_EXIT, "pr01CursorPrepareClose", 0);
}

static tpr01_CursorDesc *pr01CursorDeclare (tpr01_CursorContainer *Cursor, tpr01_StmtNameDesc *StmtNameDesc, tpr01_ConDesc *ConDesc, tpr05_String *CursorName, tpr01_UniqueID UniqueID)
{
  M90TRACE(M90_TR_ENTRY, "pr01CursorDeclare", 0);
  M90TRACE(M90_TR_HANDLE, "StmtNameDesc", StmtNameDesc);
  {
    tpr01_CursorDesc *CursorDesc = Cursor->FindDesc(Cursor, CursorName, UniqueID);
    if (!CursorDesc) {
      CursorDesc = Cursor->AddDesc(Cursor, CursorName, UniqueID, ConDesc);
    }
    if (!StmtNameDesc) {
      tsp00_Int4 ErrorCode;
      tpr01_ErrorText szErrorText;
      pr01ErrorGetErrorText(cpr_unknown_statement_name, &ErrorCode, szErrorText);
      Cursor->ErrorDesc->Error->SetError(Cursor->ErrorDesc, ErrorCode, (char*)szErrorText, CPR_NTS);
      CursorDesc = NULL;
    }
    if (CursorDesc) {
      BOOLEAN vtrue = TRUE;
      CursorDesc->StmtNameDesc = StmtNameDesc;
      CursorDesc->ConDesc = ConDesc;

      /*set useNewParseInfo flag because the following open command must 
        use the parseinfo from stmtnamedesc*/
      CursorDesc->SetAttr (CursorDesc, useNewParseInfo_epr01Cursor, &vtrue, 1);
    }
    M90TRACE(M90_TR_EXIT, "pr01CursorDeclare", 0);
    M90TRACE(M90_TR_HANDLE, "CursorDesc", &CursorDesc);
    return CursorDesc;
  }
}
/*!
  Function:     pr01CursorAlterCursorName

  see also: 

  Description:  alters the cursor name
   
  If no cursor name is given by application a implicit cursor name will 
  be set. After parsing the command by database kernel the implicit cursor 
  name will be replaced with the result set name received from database kernel.
  
  Arguments:
    CursorDesc [in]  cursor descriptor
    CursorName [out] new name of cursor

  return value: void
 */
static void pr01CursorAlterCursorName (tpr01_CursorDesc *CursorDesc, tpr05_String *CursorName)
{
  PR07_CHECKDESC(CursorDesc, CursorDesc_epr01);
  {
    tpr05_StringError erg;
    erg = pr05IfCom_String_Convert (&(CursorDesc->CursorNameDesc.CursorName),CursorName);
    CursorDesc->UniqueID = EmptyUniqueID_cpr01;
  }
}

/*!
  Function:     pr01CursorAlterParseid

  see also: 

  Description:  alters the parseid of current Openka
   
  If the useNewParseinfo flag is set or the parseid of Openka is zero the
  parseid of ka-area in StmtNameDesc will be copied.
  
  Arguments:
    SQLDesc [in]  sql descriptor

  return value: void
 */
void pr01CursorAlterParseid (tpr01_SQLDesc *SQLDesc)
{
  PR07_CHECKDESC(SQLDesc, SQLDesc_epr01);
  {
    tpr01_CursorDesc   *CursorDesc     = SQLDesc->CursorDesc;
    sqlkaentry         *ka             = SQLDesc->ka;
    tpr01_StmtNameDesc *StmtNameDesc   = SQLDesc->StmtNameDesc;
    BOOLEAN            useNewParseInfo = FALSE;
    tsp00_Int4         sz              = 0;
    
    if (   StmtNameDesc 
        &&(StmtNameDesc->State & Prepared_bpr01)
        ) {
         if ( CursorDesc ) {
           CursorDesc->GetAttr (CursorDesc, useNewParseInfo_epr01Cursor, &useNewParseInfo, &sz);
           if (   pr06ParseIdIsEqual(&ka->kaParseInfo.ParseId, ParseIdNull_cpr00)
               || useNewParseInfo ) {
             pr06ParseInfoCopy(&ka->kaParseInfo, &StmtNameDesc->ka->kaParseInfo);
             useNewParseInfo = FALSE;
             CursorDesc->SetAttr (CursorDesc, useNewParseInfo_epr01Cursor, &useNewParseInfo, 1);
           }  
         }   
         else {
           pr06ParseInfoCopy(&ka->kaParseInfo, &StmtNameDesc->ka->kaParseInfo);
         }
         ka->kastate = cpr_state_execute;
         SQLDesc->ore = StmtNameDesc->ore;
      }
  }
}


static tpr01_ConDesc *pr01CursorGetConnection(tpr01_CursorDesc *CursorDesc)
{
  M90TRACE(M90_TR_ENTRY, "pr01CursorGetConnection", 0);
  M90TRACE(M90_TR_EXIT, "pr01CursorGetConnection", 0);
  M90TRACE(M90_TR_HANDLE, "ConDesc", &CursorDesc->ConDesc);
  return CursorDesc->ConDesc;
}
 
static void pr01CursorB2B(void *cout, void *cin, int clen, int len)
{
  if (len == CPR_NTS) 
    len = strlen((char*)cin);
  len = PR07MIN(len, clen);
  memcpy((char*)cout, (char*)cin, len);  
}

static BOOLEAN pr01CursorSetAttr (tpr01_CursorDesc *CursorDesc, epr01Cursor_Attr Attr, 
				  void *pAttr, tsp00_Int4 cbLen)
{
  M90TRACE(M90_TR_ENTRY, "pr01CursorSetAttr", 0);
  {
    BOOLEAN ret = TRUE;
    switch (Attr) {
    case openkeep_epr01Cursor: {
      pr01CursorB2B (&CursorDesc->Attribute.openkeep, pAttr,
		     sizeof (CursorDesc->Attribute.openkeep), cbLen);
      break;
    } 
    case useNewParseInfo_epr01Cursor: {
      pr01CursorB2B (&CursorDesc->Attribute.useNewParseInfo, pAttr,
		     sizeof (CursorDesc->Attribute.useNewParseInfo), cbLen);
      break;
    } 
    case lastDataReceived_epr01Cursor: {
      pr01CursorB2B (&CursorDesc->Attribute.lastDataReceived, pAttr,
		     sizeof (CursorDesc->Attribute.lastDataReceived), cbLen);
      break;
    } 
    default: {
      ret = FALSE;
    }
    }
    M90TRACE(M90_TR_EXIT, "pr01CursorSetAttr", 0);  
    return ret;
  }
}

static void pr01CursorGetAttr (tpr01_CursorDesc *CursorDesc, epr01Cursor_Attr Attr, 
			       void *pAttr, tsp00_Int4 *cbLen)
{
  M90TRACE(M90_TR_ENTRY, "pr01CursorGetAttr", 0);
  {
    switch (Attr) {
    case openkeep_epr01Cursor: {
      *cbLen = sizeof (CursorDesc->Attribute.openkeep);
      pr01CursorB2B (pAttr, &CursorDesc->Attribute.openkeep,
		     sizeof (CursorDesc->Attribute.openkeep), *cbLen);
      break;
    }
    case useNewParseInfo_epr01Cursor: {
      *cbLen = sizeof (CursorDesc->Attribute.useNewParseInfo);
      pr01CursorB2B (pAttr, &CursorDesc->Attribute.useNewParseInfo,
		     sizeof (CursorDesc->Attribute.useNewParseInfo), *cbLen);
      break;
    } 
    case lastDataReceived_epr01Cursor: {
      *cbLen = sizeof (CursorDesc->Attribute.lastDataReceived);
      pr01CursorB2B (pAttr, &CursorDesc->Attribute.lastDataReceived,
		     sizeof (CursorDesc->Attribute.lastDataReceived), *cbLen);
      break;
    } 
    default: {}
    }
  }
  M90TRACE(M90_TR_EXIT, "pr01CursorGetAttr", 0);  
}

/*!
  Function:     pr01Cursor_InitCursorName

  see also: 
  
  Description:  initializes tpr00_CursorNameDesc

  Arguments:
    CursorNameDesc [in] a tpr00_CursorNameDesc
 
  return value: void
 */
static void pr01Cursor_InitCursorName (tpr00_CursorNameDesc *CursorNameDesc)
{
  CursorNameDesc->CursorName.allocationType = constant_epr05;
  pr05IfCom_String_InitString(&(CursorNameDesc->CursorName), 
                              &(CursorNameDesc->private_StrBuf), 
                              sizeof(CursorNameDesc->private_StrBuf), 
                              sp77encodingAscii, 
                              constant_epr05);
  CursorNameDesc->CursorName.cbLen=0;
}

static tpr09HashValue pr01Cursor_GetKey (const tpr09HashValue data_p, tsp00_Uint4* HashKeyLen){
  tpr05_String* string1= &((struct tpr01_CursorDesc *)data_p)->CursorNameDesc.CursorName;
  *HashKeyLen = string1->cbLen;
  return string1->rawString;
}

static tsp00_Int4 pr01Cursor_CompareKey (const tpr09HashKey Item1,tsp00_Uint4 Item1Len, const tpr09HashKey Item2, tsp00_Uint4 Item2Len){
   if (Item1Len != Item2Len)
     return 0;
   return (! memcmp(Item1, Item2, Item1Len));
}

static void pr01Cursor_printKey (const tpr09HashValue data_p, FILE* datei){
  tpr05_String* string1= &((struct tpr01_CursorDesc *)data_p)->CursorNameDesc.CursorName;
  fprintf(datei, "<CursorDesc addr=0x%p name=\"%.*s\" />", data_p, string1->cbLen, string1->rawString);
}
