/*!================================================================

  module:       vpr02Con.c

 -------------------------------------------------------------------

  responsible:  BurkhardD

  special area: Connect Syntax Analyse

  description:  

  last changed: 2001-05-30  18:23 Thomas Simenec  Unicode Analyze Connect
  see also:

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1999



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 ===================================================================*/


/*==================================================================*
 *  INCLUDES                                                        *
 *==================================================================*/

#include "gmt90.h"
#include "vpr01Con.h"
#include "vpr01SQL.h"
#include "vpr01StmtName.h"
#include "vpr07Macro.h"
#include "gpr03.h"
#include "gpr04.h"
#include "gpr05.h"
#include "gpr08.h"
#include "hsp78.h"
#include "vpr02Con.h"
#include "vpr01Trace.h"

/*==================================================================*
 *  DECLARATIONS                                                    *
 *==================================================================*/

static pr02ConKeywordTabInitialized = cpr_is_empty;

static pr05cKeywordTableStruct pr02ConKeywordTab [ ] = {
  { KeywordADABAS_epr01,     {{"ADABAS"    , 0}, {"", 0}, {"", 0}} },
  { KeywordANSI_epr01,       {{"ANSI"      , 0}, {"", 0}, {"", 0}} },
  { KeywordANSIORA_epr01,    {{"ANSIORA"   , 0}, {"", 0}, {"", 0}} },
  { KeywordAT_epr01,         {{"AT"        , 0}, {"", 0}, {"", 0}} },
  { KeywordCACHELIMIT_epr01, {{"CACHELIMIT", 0}, {"", 0}, {"", 0}} },
  { KeywordCHARACTER_epr01,  {{"CHARACTER" , 0}, {"", 0}, {"", 0}} },
  { KeywordCONNECT_epr01,    {{"CONNECT"   , 0}, {"", 0}, {"", 0}} },
  { KeywordCONNECTC_epr01,   {{"CONNECT:"  , 0}, {"", 0}, {"", 0}} },
  { KeywordDB2_epr01,        {{"DB2"       , 0}, {"", 0}, {"", 0}} },
  { KeywordDEFAULT_epr01,    {{"DEFAULT"   , 0}, {"", 0}, {"", 0}} },
  { KeywordEXPLICIT_epr01,   {{"EXPLICIT"  , 0}, {"", 0}, {"", 0}} },
  { KeywordIDENTIFIED_epr01, {{"IDENTIFIED", 0}, {"", 0}, {"", 0}} },
  { KeywordIMPLICIT_epr01,   {{"IMPLICIT"  , 0}, {"", 0}, {"", 0}} },
  { KeywordINTERNAL_epr01,   {{"INTERNAL"  , 0}, {"", 0}, {"", 0}} },
  { KeywordISOLATION_epr01,  {{"ISOLATION" , 0}, {"", 0}, {"", 0}} },
  { KeywordKEY_epr01,        {{"KEY"       , 0}, {"", 0}, {"", 0}} },
  { KeywordLEVEL_epr01,      {{"LEVEL"     , 0}, {"", 0}, {"", 0}} },
  { KeywordLOCK_epr01,       {{"LOCK"      , 0}, {"", 0}, {"", 0}} },
  { KeywordNOLOG_epr01,      {{"NOLOG"     , 0}, {"", 0}, {"", 0}} },
  { KeywordNORMAL_epr01,     {{"NORMAL"    , 0}, {"", 0}, {"", 0}} },
  { KeywordORACLE_epr01,     {{"ORACLE"    , 0}, {"", 0}, {"", 0}} },
  { KeywordSAPR3_epr01,      {{"SAPR3"     , 0}, {"", 0}, {"", 0}} },
  { KeywordSQLDB_epr01,      {{"SQL-DB"    , 0}, {"", 0}, {"", 0}} },
  { KeywordSQLMODE_epr01,    {{"SQLMODE"   , 0}, {"", 0}, {"", 0}} },
  { KeywordTIMEOUT_epr01,    {{"TIMEOUT"   , 0}, {"", 0}, {"", 0}} },
  { KeywordUSING_epr01,      {{"USING"     , 0}, {"", 0}, {"", 0}} }
};

/*==================================================================*
 *  EXTERNAL PROTOTYPES                                             *
 *==================================================================*/

int s30len(tsp_moveobj VAR_ARRAY_REF str, unsigned char val, tsp_int4 cnt);

/*==================================================================*
 *  FUNCTION PROTOTYPES                                             *
 *==================================================================*/

/*==================================================================*
 *  CODE                                                            *
 *==================================================================*/

/*!
  Function:     pr02ConGetConParameter

  see also: 

  Description:  Gets username/password from hostvariables
    
  Arguments:
    sqlca           [in]
    sqlxa           [in]
    ga              [in]
    ka              [in]
    param           [in|out] Index in va area
    userpassw       [out]    username/password from hostvariable
    encodingType    [in]     Unicode/Ascii
    
  return value:
 */
void pr02ConGetConParameter (sqlcatype *sqlca, sqlxatype *sqlxa, 
                             sqlgaentry *ga, sqlkaentry *ka, int *param, 
                             tsp_knl_identifier *userpassw, 
                             tpr05_StringEncoding* encodingType)
{
  M90TRACE(M90_TR_ENTRY, "pr02ConGetConParameter", 0);
  {
    int paind, cnt, vaind, indind, len, pos;
    int va1indva2;
    tsp00_Int2 sint;
    void *vaddr;
    void *vaaddr;
    void *iaddr;
    sqlva2en sqlv2;

    if (sqlca->sqlrap->ralang != CPR_LA_CALL) {
      paind = ka->kapaindex + *param;
      p03getparameteraddr (sqlca, sqlxa, &paind,
                           &cnt, &vaddr, &vaind,
                           &iaddr, &indind);
      paind++;
      *param = paind - ka->kapaindex;
      vaaddr = vaddr;
      va1indva2 = sqlxa->sqlv1p[vaind-1].sc.va1indva2;
      sqlv2 = sqlxa->sqlv2p[va1indva2-1];
      len = sqlv2.va2size;
      if (sqlv2.va2typ == cpr_vcharc) {
        if (sqlv2.va2size == cpr_pointerlen)
          len = mxsp_c40;
        len = s30len ((unsigned char *) vaaddr, '\0', len);
      } else if (sqlv2.va2typ == cpr_vucs2 || sqlv2.va2typ == cpr_vunicodec) {
        void *userpasswbuf;
        tsp00_Uint4 userlen;

        len = 2 * sp81UCS2strlen (vaaddr);
        userpasswbuf = userpassw;
        userlen = sizeof (tsp_knl_identifier);       
        sp77encodingUCS2Native->fillString (&userpasswbuf, 
                                            &userlen,
                                            userlen / 2,
                                            ' ');

        if(sqlLocalMachineSwapKind == sw_normal) {
          *encodingType = sp77encodingUCS2;
        } else if (sqlLocalMachineSwapKind == sw_full_swapped) {
          *encodingType = sp77encodingUCS2Swapped;
        } else {
          /* Error unsupported swap type */
          pr01TraceRuntimeError (sqlca, sqlxa, cpr_paramlist_wrong);
        }
        
      }
      if (len > mxsp_c40)
        len = mxsp_c40;
      if ((sqlv2.va2typ == cpr_vchar) ||
          (sqlv2.va2typ == cpr_vbchar) ||
          (sqlv2.va2typ == cpr_vcharc) ||
          (sqlv2.va2typ == cpr_vbuf) ||
          (sqlv2.va2typ == cpr_vucs2) ||
          (sqlv2.va2typ == cpr_vunicodec)) {
        memcpy (userpassw, vaaddr, len);
      } else {
        if ((sqlv2.va2typ == cpr_vstring) ||
            (sqlv2.va2typ == cpr_vstring1)) {
          if (sqlv2.va2typ != cpr_vstring1) {
            memcpy (&sint, vaaddr, mxsp_c2);
            pos = 3;
          } else {
            sint = ((char *) vaaddr)[0];
            pos = 2;
          }
          len = sint;
          if (len > mxsp_c40) {
            len = mxsp_c40;
          }
          memcpy (userpassw, (char *) vaaddr+pos, len);
        } else {
          pr01TraceRuntimeError (sqlca, sqlxa, cpr_paramlist_wrong);
        }
      }
    }
  }
  M90TRACE(M90_TR_EXIT, "pr02ConGetConParameter", 0);
}

void p08batdbanalyse (sqlcatype*, sqlxatype*, tsp_int4*, tsp_int4*,
                      tpr05_Symbol*, tsp_int4, char*);

void p08busingdbanalyse (sqlcatype*, sqlxatype*, tsp_int4*, tsp_int4*,
                         tpr05_Symbol*, tsp_int4, char*);

/*!
  Function:     pr02ConGetConParameter

  see also: 

  Description:  Analyzes SQL connect statement with hostvariables
                unicode or ascii
    
  Arguments:
    sqlca           [in]
    sqlxa           [in]
    sqlga           [in]
    ga              [in]
    ka              [in]
    buflen          [in]     Length of connect statement
    SQLStatement    [in|out] Statement to be analyzed
    
  return value:
 */
void pr02ConAnalyzeConnectStmt (sqlcatype *sqlca, sqlxatype *sqlxa, sqlgatype* sqlga,
                                sqlgaentry *ga, sqlkaentry *ka, tsp00_Int4 buflen,
                                tpr05_String *SQLStatement)
{
  M90TRACE(M90_TR_ENTRY, "pr02ConAnalyzeConnectStmt", 0);
  {
    tpr05_CharBuffer       *pStatement  = SQLStatement->rawString;    /*pointer to statement*/
    tsp00_Int4             StmtLength   = SQLStatement->cbLen;        /*length of statement*/
    tpr05_StringEncoding   encodingType = SQLStatement->encodingType; /*encoding type of statement*/ 
    tsp00_Int4             pAnalysePos  = 1; /*position from which analyse of statement will be continued*/
    tsp00_Int4             actPos       = 0; /*actual pos of analyse in statement */
    tsp00_Int4             SymbolPos    = 0; /*position of recognized symbol in statement*/
    tsp00_Int4             SymbolLength = 0; /*length of recognized symbol */
    tpr00_Symbol           Symbol       = CPR_S_UNKNOWN; /*recognized symbol*/
    tpr01_KeywordTypeEnum  StmtKeyword;
    tpr05_Symbol           symb;
    tpr05_SymbolIterator   Iter;
    tpr05_SymbolIterator   *pIter;
    tpr05_String           userpassw;
    tsp_knl_identifier     rawString;
    tsp_name               passw;
    tsp00_Int4             sypos;
    tsp00_Int4             sylen;
    tsp00_Int2             mode;
    tsp00_Int2             ansi;
    tsp00_Int2             varplen;
    tsp00_Int2             level;
    tsp78ConversionResult  ConvRes;
    int                    param;
    int                    passwfound;
    int                    i, j, len;
    tsp00_Uint4            dlen, slen;

    pIter = pr05cSymbolOpen (SQLStatement, &Iter);

    if (pr02ConKeywordTabInitialized != cpr_is_init) {
      /*init keywordtab for parsing of connect string*/
      pr05cInitKeywordTable (pr02ConKeywordTab, PR07NUMELEM (pr02ConKeywordTab));
      pr02ConKeywordTabInitialized = cpr_is_init;
    }
    
    if (sqlca->sqlcode == 0) {
      if (ga->gareference != 0) {
        pr01TraceRuntimeError (sqlca, sqlxa, cpr_session_already_connected);      
      } else {
        p05up1casebuf ((void *) SQLStatement->rawString, 1, SQLStatement->cbLen);
        passwfound = 0;
        pr05cSymbolEnum (pIter);
        if (((StmtKeyword = 
              pr05cSymbolGetKeyword (pIter, pr02ConKeywordTab, 
                                     PR07NUMELEM (pr02ConKeywordTab))) 
             == KeywordCONNECT_epr01) ||
            (StmtKeyword == KeywordCONNECTC_epr01)) {
          /*KEYWORD CONNECT found*/
          tpr05_Symbol parsym;
          param = 0;
          if (((parsym = pr05cSymbolEnum (pIter)) == CPR_S_PARAMETER_NAME) ||
              (parsym == CPR_S_QUESTION_MARK)) {
            if (sqlca->sqlrap->ralang != CPR_LA_CALL) {
              /*searching for username or username/passwd from hostvar*/
              memset (&rawString, ' ', sizeof (rawString));
              pr02ConGetConParameter (sqlca, sqlxa, ga, ka, &param, 
                                      &rawString, &encodingType);

              userpassw.allocationType = constant_epr05;
              pr05IfCom_String_InitString (&userpassw, &rawString, sizeof (rawString), 
                                           encodingType, constant_epr05);

              if (!pr05IfCom_String_AsciiCharCompare (&userpassw, 0, "/") &&
                  !pr05IfCom_String_AsciiCharCompare (&userpassw, 1, " ")) {
                passwfound = 1;
                if ((sqlga->gauseropset == cpr_is_false) ||
                    (sqlca->sqlrap->raactsession != cpr_se_primary))
                  p03sqlgaentryinit (ga, sqlca->sqlrap->raactsession, sqlca->sqlemp);
              } else {
                char xuuser[mxsp_c64];

                /*init xuserrec with spaces*/
                memcpy (ga->gaxuserrec.xu_user, bsp_knl_identifier,
                        strlen (bsp_knl_identifier));
                i = 0;

                ConvRes = sp78convertString (sp77encodingAscii, xuuser, sizeof (xuuser),
                                             &dlen, TRUE,			     
                                             encodingType, rawString, sizeof (tsp_knl_identifier),
                                             &slen);
                if (slen < sizeof (tsp_knl_identifier)) {
                  ConvRes = sp78convertString (sp77encodingUCS2Native, xuuser, sizeof (xuuser),
                                               &dlen, TRUE,			     
                                               encodingType, rawString, sizeof (tsp_knl_identifier),
                                               &slen);
                  memcpy (ga->gaxuserrec.xu_userUCS2, xuuser, mxsp_c64);    
                } else {
                  while (pr05IfCom_String_AsciiCharCompare (&userpassw, i, "/") &&
                         (i < mxsp_c64 / 2)) {
                    ga->gaxuserrec.xu_user[i] = xuuser[i];
                    i++;
                  }
                  if (!pr05IfCom_String_AsciiCharCompare (&userpassw, i, "/")) {
                    /*username/password-syntax found, copy passwd into xuserrec*/    
                    passwfound = 1;
                    i++;
                    j = 0;
                    memcpy (passw, bsp_name, sizeof (bsp_name));
                    while ((i < mxsp_c64) && (j < mxsp_name)) {
                      passw[j] = userpassw.rawString[i];
                      i++; j++;
                    }
                    p03encrypt (passw, ga->gaxuserrec.xu_password);
                  }
                }
              }
            }
          } else {
            /*username/password form literal*/
            sylen = pIter->SymbolLength;
            sypos = pIter->SymbolPos;
            if (sylen > mxsp_c64)
              len = mxsp_c64;
            else
              len = sylen;
            memcpy (ga->gaxuserrec.xu_user, bsp_knl_identifier, 
                    strlen (bsp_knl_identifier));
            memset (ga->gaxuserrec.xu_userUCS2, '\0', sizeof (ga->gaxuserrec.xu_userUCS2));
            for (i = 0; i < len; i++) 
              ga->gaxuserrec.xu_user[i] = SQLStatement->rawString[sypos+i-1];

            if (!pr05IfCom_String_AsciiCharCompare (SQLStatement, sypos-1, "\""))
              sylen++;
          }
          symb = pr05cSymbolEnum (pIter);
          if (passwfound) {
            sqlga->gasqlconntyp = cpr_ci_ora_connect;
          } else {
            /*password comes after KEYWORD IDENTIFIED*/
            memcpy (passw, bsp_name, strlen (bsp_name));
            if (symb != CPR_S_EOF) {
              if (pr05cSymbolGetKeyword (pIter, pr02ConKeywordTab, 
                                         PR07NUMELEM (pr02ConKeywordTab))
                  == KeywordIDENTIFIED_epr01) {
                pr05cSymbolEnum (pIter);
                symb = pr05cSymbolEnum (pIter);
                sylen = pIter->SymbolLength;
                sypos = pIter->SymbolPos;

                if ((symb == CPR_S_PARAMETER_NAME) ||
                    (symb == CPR_S_QUESTION_MARK)) {
                  if (sqlca->sqlrap->ralang != CPR_LA_CALL) {
                    memset (&rawString, ' ', sizeof (rawString));
                    pr02ConGetConParameter (sqlca, sqlxa, ga, ka, &param, 
                                            &rawString, &encodingType);
                    memcpy (passw, rawString, mxsp_name);
                    if (!strncmp ((const char *) ga->gaxuserrec.xu_user, bsp_c64, 
                                  sizeof(ga->gaxuserrec.xu_user)))
                      p03encrypt (passw, ga->gaxuserrec.xu_passwordUCS2);
                    else {
                      tsp00_Pw xupassw;

                      ConvRes = sp78convertString (sp77encodingAscii, 
                                                   xupassw, 
                                                   sizeof (xupassw),
                                                   &dlen, 
                                                   TRUE,			     
                                                   encodingType, 
                                                   rawString, 
                                                   sizeof (xupassw),
                                                   &slen);

                      p03encrypt (xupassw, ga->gaxuserrec.xu_password);
                    }
                  }
                } else {
                  if (sylen > mxsp_name)
                    len = mxsp_name;
                  else
                    len = sylen;
                  memcpy (passw, bsp_name, strlen (bsp_name));
                  for (i = 0; i < len; i++)
                    passw[i] = SQLStatement->rawString[sypos+i-1];
                  if (!pr05IfCom_String_AsciiCharCompare (SQLStatement, sypos-1, "\""))
                    sylen++;
                  p03encrypt (passw, ga->gaxuserrec.xu_password);
                }
                symb = pr05cSymbolEnum (pIter);
              }
            }
          }
          if (symb != CPR_S_EOF) {
            if (pr05cSymbolGetKeyword (pIter, pr02ConKeywordTab, 
                                       PR07NUMELEM (pr02ConKeywordTab))
                == KeywordSQLMODE_epr01) {
              pr05cSymbolEnum (pIter);
              if (sqlca->sqlrap->ralang != CPR_LA_CALL) {
                pr01TraceRuntimeError (sqlca, sqlxa, cpr_connect_syntax_wrong);
              } else {
                mode = cpr_kind_empty;
                ansi = cpr_kind_empty;
                StmtKeyword = 
                  pr05cSymbolGetKeyword (pIter, pr02ConKeywordTab, 
                                         PR07NUMELEM (pr02ConKeywordTab));
                switch (StmtKeyword) {
                case KeywordSQLDB_epr01:
                  mode = cpr_kind_internal;
                  sylen = 6;
                  break;
                case KeywordSAPR3_epr01:
                  mode = cpr_kind_sapr3;
                  break;
                case KeywordDB2_epr01:
                  mode = cpr_kind_db2;
                  break;
                case KeywordADABAS_epr01:
                case KeywordINTERNAL_epr01:
                  mode = cpr_kind_internal;
                  sylen = 6;
                  break;
                case KeywordANSI_epr01:
                  mode = cpr_kind_internal;
                  ansi = cpr_kind_ansi;
                  break;
                case KeywordANSIORA_epr01:
                  mode = cpr_kind_oracle;
                  ansi = cpr_kind_ansi;
                  break;
                case KeywordORACLE_epr01:
                  mode = cpr_kind_oracle;
                  break;
                default:
                  mode = cpr_kind_empty;
                  break;
                }
                if (sqlca->sqldbmode == cpr_kind_empty) {
                  sqlca->sqldbmode = mode;
                  sqlca->sqlrap->rasqlansi = ansi;
                } else {
                  if ((sqlca->sqldbmode != mode) ||
                      (sqlca->sqlrap->rasqlansi != ansi) ||
                      (mode == cpr_kind_empty))
                    pr01TraceRuntimeError (sqlca, sqlxa, cpr_differ_sqlmode_not_allowed);
                }
              }
              pr05cSymbolEnum (pIter);
            }
          }
          if (symb != CPR_S_EOF) {
            if (pr05cSymbolGetKeyword (pIter, pr02ConKeywordTab, 
                                       PR07NUMELEM (pr02ConKeywordTab)) 
                == KeywordNOLOG_epr01) {
              pr05cSymbolEnum (pIter);
              ga->ganolog = cpr_is_true;
            }            
          }
          sypos = pIter->SymbolPos;
          if (sqlca->sqlrap->rasqlansi == CPR_KIND_ANSI)
            ga->gaxuserrec.xu_isolation = CPR_LO_ISOLEV3;
          else
            ga->gaxuserrec.xu_isolation = CPR_LO_ISOLEV10;
          varplen = sypos - 1;
          if (symb != CPR_S_EOF) {
            if (pr05cSymbolGetKeyword (pIter, pr02ConKeywordTab, 
                                       PR07NUMELEM (pr02ConKeywordTab)) 
                == KeywordLOCK_epr01) {
              pr05cSymbolEnum (pIter);
              StmtKeyword = pr05cSymbolGetKeyword (pIter, pr02ConKeywordTab, 
                                                   PR07NUMELEM (pr02ConKeywordTab));
              switch (StmtKeyword) {
              case KeywordNORMAL_epr01:
                ga->gaxuserrec.xu_isolation = CPR_LO_ISOLEV10;
                break;
              case KeywordEXPLICIT_epr01:
                ga->gaxuserrec.xu_isolation = CPR_LO_ISOLEV0;
                break;
              case KeywordIMPLICIT_epr01:
                ga->gaxuserrec.xu_isolation = CPR_LO_ISOLEV20;
                break;
              default:
                pr01TraceRuntimeError (sqlca, sqlxa, cpr_connect_syntax_wrong);
                break;
              }
              pr05cSymbolEnum (pIter);
            } else if (pr05cSymbolGetKeyword (pIter, pr02ConKeywordTab, 
                                              PR07NUMELEM (pr02ConKeywordTab)) 
                       == KeywordISOLATION_epr01) {
              pr05cSymbolEnum (pIter);
              if (pr05cSymbolGetKeyword (pIter, pr02ConKeywordTab, 
                                         PR07NUMELEM (pr02ConKeywordTab)) 
                  == KeywordLEVEL_epr01) {
                if (sqlca->sqlrap->rasqlansi == cpr_kind_ansi)
                  ga->gaxuserrec.xu_isolation = CPR_LO_ISOLEV3;
                else
                  ga->gaxuserrec.xu_isolation = CPR_LO_ISOLEV10;
                if (pr05cSymbolEnum (pIter) == CPR_S_UNSIGNED_INTEGER) {
                  level = p05chrtoint2 ((void *) SQLStatement->rawString, 
                                        pIter->SymbolPos,
                                        pIter->SymbolLength);
                  if ((level != CPR_LO_ISOLEV0) &&
                      (level != CPR_LO_ISOLEV1) &&
                      (level != CPR_LO_ISOLEV2) &&
                      (level != CPR_LO_ISOLEV3) &&
                      (level != CPR_LO_ISOLEV4) &&
                      (level != CPR_LO_ISOLEV9) &&
                      (level != CPR_LO_ISOLEV10) &&
                      (level != CPR_LO_ISOLEV15) &&
                      (level != CPR_LO_ISOLEV20) &&
                      (level != CPR_LO_ISOLEV30))
                    pr01TraceRuntimeError (sqlca, sqlxa, cpr_connect_syntax_wrong);
                  else
                    ga->gaxuserrec.xu_isolation = level;
                  pr05cSymbolEnum (pIter);
                } else
                  pr01TraceRuntimeError (sqlca, sqlxa, cpr_connect_syntax_wrong);
              }
            }
          }
          if (symb != CPR_S_EOF) {
            if (pr05cSymbolGetKeyword (pIter, pr02ConKeywordTab, 
                                       PR07NUMELEM (pr02ConKeywordTab)) 
                == KeywordTIMEOUT_epr01) {
              if (pr05cSymbolEnum (pIter) == CPR_S_UNSIGNED_INTEGER) {
                ga->gaxuserrec.xu_timeout = p05chrtoint2 ((void *) SQLStatement->rawString, 
                                                          pIter->SymbolPos,
                                                          pIter->SymbolLength);
                pr05cSymbolEnum (pIter);
              } else
                pr01TraceRuntimeError (sqlca, sqlxa, cpr_connect_syntax_wrong);
            }
          }
          if (symb != CPR_S_EOF) {
            if (pr05cSymbolGetKeyword (pIter, pr02ConKeywordTab, 
                                       PR07NUMELEM (pr02ConKeywordTab)) 
                == KeywordCACHELIMIT_epr01) {
              if (pr05cSymbolEnum (pIter) == CPR_S_UNSIGNED_INTEGER) {
                ga->gaxuserrec.xu_cachelimit = p05chrtoint2 ((void *) SQLStatement->rawString, 
                                                             pIter->SymbolPos,
                                                             pIter->SymbolLength);
                pr05cSymbolEnum (pIter);
              } else
                pr01TraceRuntimeError (sqlca, sqlxa, cpr_connect_syntax_wrong);
            }
          }
          if (symb != CPR_S_EOF) {
            if (pr05cSymbolGetKeyword (pIter, pr02ConKeywordTab, 
                                       PR07NUMELEM (pr02ConKeywordTab)) 
                == KeywordAT_epr01) {
              if (((sqlca->sqldbmode == cpr_kind_oracle) ||
                   (sqlca->sqldbmode == cpr_kind_sapr3)) &&
                  (sqlca->sqlrap->racomponent[0] != 'R')) {
                sqlga->gasqlconntyp = cpr_ci_ora_connect;
                p08batdbanalyse (sqlca, sqlxa,
                                 &pIter->SymbolPos, &pIter->SymbolLength, 
                                 &pIter->Symbol, buflen, SQLStatement->rawString);
              } else
                pr01TraceRuntimeError (sqlca, sqlxa, cpr_connect_syntax_wrong);
            }
          }
          if (symb != CPR_S_EOF) {
            if (pr05cSymbolGetKeyword (pIter, pr02ConKeywordTab, 
                                       PR07NUMELEM (pr02ConKeywordTab)) 
                == KeywordUSING_epr01) {
              if ((sqlca->sqldbmode == cpr_kind_oracle) ||
                  (sqlca->sqldbmode == cpr_kind_sapr3)) {
                sqlga->gasqlconntyp = cpr_ci_ora_connect;
                p08busingdbanalyse (sqlca, sqlxa,
                                    &pIter->SymbolPos, &pIter->SymbolLength, 
                                    &pIter->Symbol, buflen, SQLStatement->rawString);
              } else
                pr01TraceRuntimeError (sqlca, sqlxa, cpr_connect_syntax_wrong);
            }
          }
          if (symb != CPR_S_EOF) {
            if (pr05cSymbolGetKeyword (pIter, pr02ConKeywordTab, 
                                       PR07NUMELEM (pr02ConKeywordTab)) 
                == KeywordKEY_epr01) {
              pr05cSymbolEnum (pIter);
              if (pr05cSymbolGetKeyword (pIter, pr02ConKeywordTab, 
                                       PR07NUMELEM (pr02ConKeywordTab)) 
                == KeywordDEFAULT_epr01) {
                /*KEY DEFAULT found*/
                tsp4_xuser_record    *gaxu = &ga->gaxuserrec;
                struct SQLERROR *sqlemp = sqlca->sqlemp;
                unsigned char   ok;
                sqlindexuser ( 1, gaxu, sqlemp->etext, &ok);
                ga->gastatus = cpr_co_key_default;
                if (ok == FALSE)
                   p04trint4(sqlca->sqlrap, "No XUSER entry found. Continue", sqlca->sqlemp->elzu);
              } else
                pr01TraceRuntimeError (sqlca, sqlxa, cpr_connect_syntax_wrong);
            }
          } /*end if Keyword KEY DEFAULF*/
        }
      }
      p03conninfoget (sqlca->sqlgap, ga, sqlca->sqlrap->raactsession);
    }
  }
  M90TRACE(M90_TR_EXIT, "pr02ConAnalyzeConnectStmt", 0);
}

void pr02ConAnalyseConnect(struct tpr01_ConDesc *ConDesc, struct tpr01_StmtNameDesc *StmtNameDesc)
{
  M90TRACE(M90_TR_ENTRY, "pr02ConAnalyseConnect", 0);
  {
    sqlcatype *sqlca = ConDesc->Connection->sqlca;
    sqlxatype *sqlxa = ConDesc->sqlxa;
    sqlgaentry *gaen = ConDesc->ga;
    sqlgatype *sqlga = sqlca->sqlgap;

    int cbLen;
    char *szSQLStatement=NULL;
    tpr05_String *pConString=NULL;
    tpr01_StmtNameContainer *StmtName = StmtNameDesc->StmtName; 
    
    int ret = StmtName->GetSQL(StmtNameDesc, &pConString, Reference_epr01);
    szSQLStatement= pConString->rawString;
    cbLen = pConString->cbLen;
    
    if (pConString->encodingType == sp77encodingAscii) {
      pr02ConAnalyzeConnectStmt (sqlca, sqlxa, sqlga, gaen, StmtNameDesc->ka, cbLen, pConString); 
    } else {
      /*error*/
    }
  }
  M90TRACE(M90_TR_EXIT, "pr02ConAnalyseConnect", 0);
}

/*!
  Function: pr02ConParseDBString 

  description: Parse einen nach ORACLE Konventionen zusammengesetzten
               Instanznamen.

	       Syntax:   
	              dbstring := [[<networkname>][<servernode>]]<servername>

		      networkname := [<networkstring>]<netsep>
		      servernode  := <nodestring><nodesep>
		      servername  := <dbnamestring>
		      netsep := ':'
		      nodesep := '-'
	       
	       Der Inputstring wird von hinten nach vorne durchsucht um die 
	       Separatoren zu bestimmen. Ist ein <networkname> oder 
	       <servernode> nicht vorhanden oder leer, so zeigen die 
	       Outputwerte auf das Ende des szDBString ('\0').
	       Die Separatoren werden aus szDBString durch '\0' ersetzt.
		      	       
  arguments:
      szDBString   [in/out] Nullterminierter ORACLE Instanzname.
      szNetworName [out] Pointer auf <networkname>.
      szServerNode [out] Pointer auf <servernode>.
      szServerName [out] Pointer auf <servername>.
*/

void pr02ConParseDBString(char *szDBString, char **szNetworkName, char **szServerNode, char **szServerName)
{
  /* points to the end */
  char *szEnd = strrchr(szDBString, '\0');
  char cSep = ':';          /* node separator */
  char *p;
  *szNetworkName = szDBString;
  p = strrchr(szDBString, cSep);
  if (p == NULL) {
    *szServerNode = szDBString;
    *szNetworkName = szEnd;
  }
  else {
    /* drop the separator */
    *p = '\0';
    szDBString = p+1;
  }    
  cSep = '-';
  *szServerNode = szDBString;
  p = strrchr(szDBString, cSep);
  if (p == NULL) {
    cSep = '@';
    p = strrchr(szDBString, cSep);
    if (p == NULL) {
      *szServerName = szDBString;      
      *szServerNode = szEnd;
    }
    else {
      /* drop the separator */
      *p = '\0';
      *szServerName = p+1;
    }
  }
  else {
    *p = '\0';
    *szServerName = p+1;
  }    
}

#ifdef TEST
main()
{
  char *szDBString = "p26235@DB72";
  char *szNetworkName;
  char *szServerNode;
  char *szServerName;
  pr02ConParseDBString(szDBString, &szNetworkName, &szServerNode, &szServerName );
  printf("%s,%s,%s", szNetworkName, szServerNode, szServerName);
}

void p08analyzeconnectstatm ()
{
}
#endif
