/*!**********************************************************************

  module: hwd201sock.h

  -----------------------------------------------------------------------

  responsible:  Holger Bischoff
				Markus Oezgen

  special area: WebAgent

  description:  Socket interface

  see also:

  -----------------------------------------------------------------------

                         Copyright by SAP AG, 1999


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

************************************************************************/

#ifndef HWD201SOCK_H
#define HWD201SOCK_H

/*=======================================================================
 *  INCLUDES
 *=====================================================================*/

#ifdef WIN32
#include <winsock.h>
#else
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <netinet/in.h>
#include <errno.h>
#include <arpa/inet.h>
#endif

#include "gsp00.h"
#include "gwd00.h"

/* mapping of error-codes */
#ifdef WIN32
#define EADDRINUSE_WD201              WSAEADDRINUSE
#else
#define EADDRINUSE_WD201              EINVAL
#endif


#if defined(__cplusplus)
extern "C" {
#endif

    typedef sapdbwa_Int4 twd201_SockRet;

    typedef sapdbwa_Int4 twd201_SockErr;

#ifdef WIN32    
    typedef SOCKET twd201_Socket;
#else
    typedef int twd201_Socket;
#endif

#ifdef WIN32    
    typedef SOCKADDR_IN twd201_SockAddrIn;
#else
    typedef struct sockaddr_in twd201_SockAddrIn;
#endif
    
#ifdef WIN32    
    typedef SOCKADDR twd201_SockAddr;
#else
    typedef struct sockaddr twd201_SockAddr;
#endif    

#define SOCKADDR_LEN sizeof(twd201_SockAddr)
    
#ifndef WIN32
#define INVALID_SOCKET  (twd201_Socket)(~0)
#define SOCKET_ERROR                   (-1)
#endif

    twd201_SockRet wd201SocketStartup();

    twd201_SockRet wd201SocketCleanup();

    twd201_SockErr wd201GetLastError();

    const unsigned char* wd201StrError( twd201_SockErr sockErr );

    twd201_SockRet wd201GetHostname( char *name, sapdbwa_Int4 nameLen );

    twd201_SockRet wd201GetPeername( twd201_Socket    socket,
                                     twd201_SockAddr *sockName,
                                     sapdbwa_Int4      *nameLen );

    twd201_SockRet wd201GetSockName( twd201_Socket    socket,
                                     twd201_SockAddr *sockName,
                                     sapdbwa_Int4      *nameLen );
    
    twd201_Socket wd201Socket( sapdbwa_Int4  af,
                               sapdbwa_Int4  type,
                               sapdbwa_Int4  protocol );

    twd201_SockRet wd201Bind( twd201_Socket    socket,
                              twd201_SockAddr *pAddr,
                              sapdbwa_Int4       nameLen );

    twd201_SockRet wd201CloseSocket( twd201_Socket    socket );

    twd201_SockRet wd201Listen( twd201_Socket  socket,
                                sapdbwa_Int4     backlog );

    twd201_Socket wd201Accept( twd201_Socket    lstnSocket,
                               twd201_SockAddr *pAddr,
                               sapdbwa_Int4      *pNameLen );

    twd201_SockRet wd201Send( twd201_Socket  socket,
                              const char    *buf,
                              sapdbwa_Int4     len,
                              sapdbwa_Int4     flags );

    twd201_SockRet wd201Recv( twd201_Socket socket,
                              char          *buf,
                              sapdbwa_Int4     len,
                              sapdbwa_Int4     flags );

    twd201_SockRet wd201SetSockOpt( twd201_Socket socket,
                                    sapdbwa_Int4    level,
                                    sapdbwa_Int4    opname,
                                    const char   *optval,
                                    sapdbwa_Int4    optlen );

    twd201_SockRet wd201GetSockOpt( twd201_Socket socket,
                                    sapdbwa_Int4    level,
                                    sapdbwa_Int4    opname,
                                    char         *optval,
                                    sapdbwa_Int4   *optlen );

    twd201_SockRet wd201Connect( twd201_Socket          socket,
                                 const twd201_SockAddr *name,
                                 int                    nameLen );

    twd201_SockRet wd201Select( sapdbwa_Int4            nfds,
                                fd_set               *readfds,
                                fd_set               *writefds,
                                fd_set               *exceptfds,
                                const struct timeval *timeout );
    
#if defined(__cplusplus)
} /* extern "C" */
#endif

#endif
