/*!**********************************************************************

  module: hwd204zstring.h

  -----------------------------------------------------------------------

  responsible:  Holger Bischoff
				Markus zgen

  special area: WebAgent

  description:  A string class

  see also:

  -----------------------------------------------------------------------

                         Copyright by SAP AG, 1999



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

************************************************************************/

#if !defined(ZSTRING_H)
#define ZSTRING_H

#include <stdarg.h>

#include "gwd00.h"

// ******************************************************************************
#ifdef  ZStringImport
#undef  ZStringImport
#ifdef WIN32
#define ZStringImport __declspec(dllimport)
#else
#define ZStringImport
#endif
#else
#define ZStringImport
#endif
#ifdef  ZStringExport
#undef  ZStringExport
#ifdef WIN32
#define ZStringExport __declspec(dllexport)
#else
#define ZStringExport
#endif
#else
#define ZStringExport
#endif
// ******************************************************************************
class ZStringImport ZStringExport ZString  
{
protected:
	char*	m_Buf;
	int		m_Length;
	int		m_Alloc;
	int		m_AllocInitalSize;
	int		m_AllocIncrement;
public:
			ZString();
			ZString(int initAlloc,int incrAlloc);
			ZString(const char* s);
			ZString(const char* s,int l);
			ZString(const ZString& s);
	virtual ~ZString();

	int		GetSize()	const 				{ return m_Length;};
	int		GetAllocSize()	const 			{ return m_Alloc;};
	sapdbwa_Bool	SetAllocIncrement(int incrAlloc);
    sapdbwa_Bool	SetSize(int l, sapdbwa_Bool useAllocIncr = sapdbwa_False);
	sapdbwa_Bool	SetBuf( const ZString& s);
	sapdbwa_Bool	SetBuf( const char* s);
	sapdbwa_Bool	SetBuf( const char* s,int l);

	sapdbwa_Bool	Add( const ZString& s);
	sapdbwa_Bool	Add( const char* s);
	sapdbwa_Bool	Add( char  ch);
	sapdbwa_Bool	Add( const char* s,int l);
	sapdbwa_Bool	AddPrefix(ZString& s);
	sapdbwa_Bool	AddPrefix(char* s);
	sapdbwa_Bool	AddPrefix(char* s,int l);

	char&	operator [](int idx);

	operator char*() const;
	ZString& operator =(const char* s);
	ZString& operator =(const ZString& s);

	ZString& operator +=(const ZString& s);
	ZString& operator +=(const char* s);
	ZString& operator +=(char s);

#ifdef WIN32    
//	sapdbwa_Bool	Format(char* fmt,char* s);
//	sapdbwa_Bool	Format(char* fmt,char* s1,char* s2);
	sapdbwa_Bool	Format(char* format, ...);
#endif
    
	sapdbwa_Bool	Break(ZString& s1,ZString& s2,char breakChar);
	sapdbwa_Bool	RBreak(ZString& s1,ZString& s2,char breakChar);
	sapdbwa_Bool	Break(unsigned int len,ZString& s1,ZString& s2);
	sapdbwa_Bool	RBreak(unsigned int len,ZString& s1,ZString& s2);

	char&	Last();

	sapdbwa_Bool	Find(char c,int& pos,int startPos=0);
	sapdbwa_Bool	Find(char* s,int l,int& pos,int startPos=0);
	sapdbwa_Bool	Find(ZString& s,int& pos,int startPos=0);
	sapdbwa_Bool	RFind(char c,int& pos,int startPos=-1);
	sapdbwa_Bool	RFind(char* s,int l,int& pos,int startPos=-1);
	sapdbwa_Bool	RFind(ZString& s,int& pos,int startPos=-1);

	sapdbwa_Bool	operator ==(const ZString& s) const;
	sapdbwa_Bool	operator ==(const char* s) const;
	sapdbwa_Bool	operator !=(const ZString& s) const;
	sapdbwa_Bool	operator !=(const char* s) const;

	sapdbwa_Bool	SubString(ZString& subString,int startPos,int endPos=-1);

	sapdbwa_Bool	CharToHex(char* s,int l,sapdbwa_Bool upperHex=sapdbwa_True);
	sapdbwa_Bool	HexToChar(char* s,int l);
	sapdbwa_Bool	Base64Decode(char* in,int lIn);

	sapdbwa_Bool	RemoveChars(char c);
	sapdbwa_Bool	RemoveChars(char c,sapdbwa_Bool trailing);

	void ToLower();
	void ToUpper();
protected:
	void	Init(int initAlloc,int incrAlloc);

	sapdbwa_Bool	FormatV(char* format, va_list argList);

};
// ******************************************************************************
#endif // !defined(ZSTRING_H)
