/*!**********************************************************************

  module: hwd205fs.h

  -----------------------------------------------------------------------

  responsible:  Holger Bischoff
				Markus zgen

  special area: WebAgent

  description:  FTP Interface to filesystem

  see also:

  -----------------------------------------------------------------------

                         Copyright by SAP AG, 1999


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

************************************************************************/

#ifndef HWD205FS_H
#define HWD205FS_H

/*=======================================================================
 *  INCLUDES
 *=====================================================================*/

#include "geo102.h"

#include "hwd204zstring.h"
#include "hwd207stream.h"

/*!**********************************************************************

  Chapter: twd205_BasicFS

  description:
  
************************************************************************/

class twd205_BasicFS {
public:
    // constructor
    twd205_BasicFS() { };

    // destructor
    virtual ~twd205_BasicFS() { };

    // path can always be an absolute path (beginning with '/') or a path
    // relative to the current working dir (e.g. '../newpath' or
    // 'newpath/sub')

    // change the current working dir relative to the current dir or
    // absolute if dir[0]=='/'
    virtual SAPDB_Bool ChDir( const ZString &dir ) = 0;

    // check if given file or directory exists
    virtual SAPDB_Bool CheckPath( const ZString  &path,
                                  SAPDB_Bool     &pathExists ) = 0;

    // chage to parent dir (cd ..)
    virtual SAPDB_Bool CdUp() = 0;

    // Delete file or empty directory
    virtual SAPDB_Bool Delete( const ZString &path ) = 0;
    
    // get actual working directory
    virtual SAPDB_Bool GetDir( ZString &dir ) = 0;

    // get stream for directory listing
    virtual twd207_BasicInputStream* GetDirStream( const ZString &path,
                                                   SAPDB_Bool     details ) = 0;

    // get stream to retrieve a file
    virtual twd207_BasicInputStream* GetFileInStream( const ZString &path )
        = 0;

    // get stream to store a file
    virtual twd207_BasicOutputStream* GetFileOutStream( const ZString &path )
        = 0;

    // get file size
    virtual SAPDB_Bool GetSize( const ZString &path, SAPDB_Int4 &size ) = 0;

    // build stat string for specified path
    virtual SAPDB_Bool GetStatStr( const ZString &path,
                                   ZString       &statStr ) = 0;

    // create new directory
    virtual SAPDB_Bool MkDir( const ZString &newDir ) = 0;

    // rename file or directory
    virtual SAPDB_Bool RenameFromTo( const ZString &fromName,
                                     const ZString &toName ) = 0;

    // remove directory
    virtual SAPDB_Bool RmDir( const ZString &dir ) = 0;

    virtual const ZString& GetLastErrorMsg() = 0;
    virtual tsp00_Int4 GetLastError() = 0;

}; // class twd205_BasicFS

/*!**********************************************************************

  EndChapter: twd205_BasicFS

************************************************************************/

#endif
