/*!**********************************************************************

  module: hwd206dbfs.h

  -----------------------------------------------------------------------

  responsible:  Holger Bischoff
				Markus zgen

  special area: WebAgent

  description:  FTP Interface to database-filesystem

  see also:

  -----------------------------------------------------------------------

                         Copyright by SAP AG, 1999


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

************************************************************************/

#ifndef HWD206DBFS_H
#define HWD206DBFS_H

/*=======================================================================
 *  INCLUDES
 *=====================================================================*/

#include "gwd00.h"
#include "hwd05wases.h"
#include "hwd101dbfs.h"
#include "hwd204zstring.h"
#include "hwd205fs.h"

/*!**********************************************************************

  Chapter: twd206_DBFS

  description:
  
************************************************************************/

class twd206_Intern;

class twd206_DBFS : public twd205_BasicFS {
public:
    // constructor
    twd206_DBFS( sapdbwa_SessionPoolP pool );

    // destructor
    ~twd206_DBFS();

    // change the current working dir relative to the current dir
    sapdbwa_Bool ChDir( const ZString &dir );

    // check if given file or directory exists
    sapdbwa_Bool CheckPath( const ZString &path, sapdbwa_Bool &pathExists );

    // chage to parent dir (cd ..)
    sapdbwa_Bool CdUp();

    // Delete file or empty directory
    sapdbwa_Bool Delete( const ZString &path );
    
    // get actual working directory
    sapdbwa_Bool GetDir( ZString &dir );

    // get stream for directory listing
    twd207_BasicInputStream* GetDirStream( const ZString &path,
                                           sapdbwa_Bool     details );

    // get stream to retrieve a file
    twd207_BasicInputStream* GetFileInStream( const ZString &path );

    // get stream to store a file
    twd207_BasicOutputStream* GetFileOutStream( const ZString &path );

    // get file size
    sapdbwa_Bool GetSize( const ZString &path, sapdbwa_Int4 &size );

    // build stat string for specified path
    sapdbwa_Bool GetStatStr( const ZString &path, ZString &statStr );

    // create new directory relative to the current working dir
    sapdbwa_Bool MkDir( const ZString &newDir );

    // rename file or directory
    sapdbwa_Bool RenameFromTo( const ZString &fromName,
                             const ZString &toName );

    // remove directory relative to the current working dir
    sapdbwa_Bool RmDir( const ZString &dir );

    const ZString& GetLastErrorMsg();
    sapdbwa_Int4 GetLastError();
private:
    twd206_Intern *in; // we hide all the internal stuff
}; // class twd206_DBFS

/*!**********************************************************************

  EndChapter: twd206_DBFS

************************************************************************/
#endif
