/*!**********************************************************************

  module: hwd208err.h

  -----------------------------------------------------------------------

  responsible:  Holger Bischoff
				Markus zgen

  special area: WebAgent

  description:  C++ Error Object derived from twd26ErrP

  see also:

  -----------------------------------------------------------------------

						Copyright by SAP AG, 1999

    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

************************************************************************/

#ifndef HWD208ERR_H
#define HWD208ERR_H

/*=======================================================================
 *  INCLUDES
 *=====================================================================*/

#include "hwd26err.h"
#include "hwd204zstring.h"

/*!**********************************************************************

  Chapter: twd208_Err

  description: The twd26Err class describes an error with error-id and
               two string arguments. 
  
************************************************************************/

class twd208_Err
{
public:
    // default constructor
    twd208_Err();
    // create from twd26ErrP
    twd208_Err( twd26ErrP err );
    // copy constructor
    twd208_Err( const twd208_Err &errObj );

    // destructor
    ~twd208_Err();

    twd208_Err& operator =( const twd208_Err &err );

    const ZString& GetMsg();

    sapdbwa_UInt4 GetId();

    void SetErr( sapdbwa_Int2     errId,
                 const ZString &arg1 = "",
                 const ZString &arg2 = "" );

    void wd26SetOdbcError( RETCODE sqlRetcode,
                           HENV    henv,
                           HDBC    hdbc,
                           HSTMT   hstmt );
    void ResetErr();

    // void LogErr( log );

    void PrintErr();
private:
    twd26ErrP err;
    ZString   errMsg;
}; // class twd208_Err

/*!**********************************************************************

  EndChapter: twd208_Err

************************************************************************/
/*!**********************************************************************

  Chapter: inline methods

************************************************************************/

inline const ZString& twd208_Err::GetMsg()
{
    return (this->errMsg = wd26GetMsg( this->err ));
}; // GetMsg


inline sapdbwa_UInt4 twd208_Err::GetId()
{
    return wd26GetId( this->err );
}; // GetId


inline void twd208_Err::SetErr( sapdbwa_Int2     errId,
                                const ZString &arg1,
                                const ZString &arg2 )
{
    wd26SetErr( this->err, errId, (const char*) arg1, (const char*) arg2 );
}; // SetErr


inline void twd208_Err::wd26SetOdbcError( RETCODE sqlRetcode,
                                          HENV    henv,
                                          HDBC    hdbc,
                                          HSTMT   hstmt )
{
    //    wd26SetOdbcError( this->err, sqlRetcode, henv, hdbc, hstmt );
}; // SetOdbcError


inline void twd208_Err::ResetErr()
{
    wd26ResetErr( this->err );
}; // ResetError


// void LogErr( log );

inline void twd208_Err::PrintErr()
{
    wd26PrintErr( this->err );
}; // PrintErr

#define ERR_INETD_OFFSET 10000

#define ERR_DIR_ALREADY_OPEN_WD208 (ERR_INETD_OFFSET + 1)
#define ERR_FILE_ALREADY_OPEN_WD208 (ERR_INETD_OFFSET + 2)
#define ERR_INVALID_FILE_OR_DIR_WD208 (ERR_INETD_OFFSET + 3)

#define ERR_MSG_DIR_ALREADY_OPEN_WD208 "Can only open one directory at once"
#define ERR_MSG_FILE_ALREADY_OPEN_WD208 "Can only open one file at once"
#define ERR_MSG_INVALID_FILE_OR_DIR_WD208 "Invalid file or directory"
#endif
