/*!**********************************************************************

  module: hwd21req.h

  -----------------------------------------------------------------------

  responsible:	Holger Bischoff
				Markus Oezgen

  special area: WebAgent

  description:  HttpRequest functions

  see also:

  -----------------------------------------------------------------------

                         Copyright by SAP AG, 1998


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

************************************************************************/

#ifndef HWD21REQ_H
#define HWD21REQ_H

/*=======================================================================
 *  INCLUDES
 *=====================================================================*/

#include "gwd00.h"
#include "hwd03waapi.h"
#include "hwd26err.h"

#if defined(__cplusplus)
extern "C" {
#endif

/*!**********************************************************************

  Chapter: sapdbwa_HttpRequestP

  description:
  
************************************************************************/

typedef sapdbwa_HttpRequestP twd21HttpRequestP;


SAPDB_Bool wd21CreateRequestHandlePool();

void wd21DestroyRequestHandlePool();
    
twd21HttpRequestP wd21CreateRequest( twd00ServerHandle  serverHandle,
                                     twd00ReadBodyFunc  readBody,
                                     twd00GetHeaderFunc getHeader,
                                     twd00IsSecureFunc  isSecure );


void wd21DestroyRequest( twd21HttpRequestP req );

twd00ServerHandle wd21GetServerHandle( twd21HttpRequestP  req );


void wd21SetErr( twd21HttpRequestP req, twd26ErrP err );


void wd21SetUri( twd21HttpRequestP  req,
                 const char        *uri );


void wd21SetIfModifiedSince( twd21HttpRequestP  req,
                             const char        *ifModifiedSince );

void wd21SetQueryString( twd21HttpRequestP  req,
                         const char        *queryString );

void wd21SetPathInfo( twd21HttpRequestP  req,
                      const char        *pathInfo );

void wd21SetRequestMethod( twd21HttpRequestP  req,
                           const char        *requestMethod );

void wd21SetContentType( twd21HttpRequestP  req,
                         const char        *contentType );

void wd21SetContentLength( twd21HttpRequestP  req,
                           const char        *contentLength );

void wd21SetPhysicalPath( twd21HttpRequestP  req,
                          const char        *physicalPath );

void wd21SetServerName( twd21HttpRequestP  req,
                        const char        *serverName );



/*!**********************************************************************

  EndChapter: sapdbwa_HttpRequestP

************************************************************************/
#if defined(__cplusplus)
}
#endif

#endif

