/*!**********************************************************************

  module: hwd220ftp.h

  -----------------------------------------------------------------------

  responsible:	Holger Bischoff
				Markus zgen

  special area: WebAgent

  description:  FTP service

  see also:

  -----------------------------------------------------------------------

                         Copyright by SAP AG, 1999

    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

************************************************************************/

#ifndef HWD220FTP_H
#define HWD220FTP_H

/*=======================================================================
 *  INCLUDES
 *=====================================================================*/

#include "gwd00.h"
#include "hwd205fs.h"
#include "hwd207stream.h"
#include "hwd211service.h"

#define MAX_REPLY_LEN 255

class twd220_FtpService : public twd211_BasicService {
public:
    // constructor
    twd220_FtpService( twd201_Socket      ctrlSocket,
                       twd201_SockAddrIn  ctrlSockName,
                       twd205_BasicFS    &fs,
                       sapdbwa_Bool         debugMode = sapdbwa_False );
    // destructor
    ~twd220_FtpService();

    sapdbwa_Int4 Service();
    
private:
    /*
    class SockOutStream : public twd207_BasicOutputStream {
        SockOutStream() {};
        ~SockOutStream() {};
        sapdbwa_Int4 Write( char *buf, sapdbwa_Int4 bufSize );
        
        const ZString& GetLastErrorMsg();
        sapdbwa_Int4 GetLastError();
    }; // class SockOutStream
    */  
    //    twd201_SockAddrIn         hisAddr;    // remote socket address
    twd201_Socket             dataSocket;     // socket for data connection
    twd201_SockAddrIn         dataSourceAddr; // data addr in passive mode
    twd201_SockAddrIn         dataDestAddr;   // data addr if not in
                                              // passive mode
    sapdbwa_Bool                useDefault; // use default remote port for
                                          // data connection?
    twd201_Socket             pasvSocket;
    twd201_SockAddrIn         pasvSockAddr;
    twd207_BasicOutputStream *dataOutStream;
    sapdbwa_Int4                dataConnType; // TYPE_A (ASCII) or BINARY
    sapdbwa_Int4                formCode;
    twd205_BasicFS           &fs;
    sapdbwa_Bool                debug;
    sapdbwa_Bool                loggedIn;     // is the remote user logged in?
    sapdbwa_Bool                askPassword;  // prompt for password?
    ZString                   userName;     // user logged in
    char                      replyBuf[MAX_REPLY_LEN];
    ZString                   fromName;     // set by the RNFR command

    static const char *ftpVersion;

    // send replies over the control connection:
    void reply( int n, const char *fmt, ...);
    void lreply( int n, const char *fmt, ...);
    void perror_reply( int n, const char *str, const char *msg );
    void sendReply( const char *replyString );
    // send positive reply for cmd
    void acknowledge( const char *cmd );
    // entry function for a ftp session:
    void StartSession();

    // check if user is logged in
    sapdbwa_Bool checkLogin();

    // connection handling:
    void SetRemoteDataPort( const twd201_SockAddrIn &dataSockAddr );
    sapdbwa_Bool GetDataSock();
    twd201_Socket* OpenDataConn( const ZString &name );
    void CloseDataConn( twd201_Socket *dataSockPtr );
    // send data from fromStream to the data connection
    void SendData( twd201_Socket             dataSock,
                   twd207_BasicInputStream  &fromStream );
    // read from the data connection and write to toStream
    void ReceiveData( twd201_Socket              dataSock,
                      twd207_BasicOutputStream  &toStream );

    // commands:
    // change to parent dir
    void CdUp();
    // change working dir
    void Cwd( const ZString &path );
    // delete file or directory
    void Delete( const ZString &path );
    // end session
    void DoLogout( sapdbwa_Int4 status );
    // send dir listing over data connection
    void ListDir( const ZString &path, sapdbwa_Bool details );
    // create new dir
    void MakeDir( const ZString &newDir );
    // login password
    void Pass( const ZString &passwd );
    // switch to passive mode
    void PassiveMode();
    // show current working directory
    void Pwd();
    // remove dir
    void RemoveDir( const ZString &dir );
    // send file over data connection
    void Retrieve( const ZString &path );
    // rename file
    void RenameFrom( const ZString &oldName );
    void RenameTo( const ZString &newName );
    // set type of data connection (ascii or binary)
    void SetTypeCode( sapdbwa_Int4 typeCode, sapdbwa_Int4 formCode );
    // set mode code
    void SetModeCode( sapdbwa_Int4 modeCode );
    // set struct code
    void SetStructCode( sapdbwa_Int4 structCode );
    // STAT command
    void StatCmd();
    // STAT command with file argument
    void StatFileCmd( const ZString &path );
    // store file
    void Store( const ZString &path );
    // login user
    void User( const ZString &user );
    // get size of file
    void Size( const ZString &path );
}; // class twd220_FtpService

#endif
