/*!**********************************************************************

  module: hwd221ftpcmd.h

  -----------------------------------------------------------------------

  responsible:	Holger Bischoff
				Markus zgen

  special area: WebAgent

  description:  Parser for ftp commands

  see also:

  -----------------------------------------------------------------------

                         Copyright by SAP AG, 1999

    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

************************************************************************/

#ifndef HWD221FTPCMD_H
#define HWD221FTPCMD_H

/*=======================================================================
 *  INCLUDES
 *=====================================================================*/

#include "geo102.h"

#include "hwd201sock.h"
#include "hwd204zstring.h"

/* Due to an lack of following declaration.			*/
/* Otherwise I have to include <strings.h>  (MaO)	*/
#if __cplusplus && AIX
extern "C" {
extern void bzero(void *, size_t);
}
#endif

/*!**********************************************************************

  Chapter: twd221_FtpCmd

  description:
  
************************************************************************/

typedef sapdbwa_Int4 twd221_FtpCmdType;

class twd221_Token {
public:
    enum TokenTypes {
        ERR = -1,
        // ftp commands
        USER = 0, PASS, ACCT, SMNT, REIN, QUIT,
        PORT, PASV, TYPE, STRU, MODE, RETR,
        STOR, APPE, MLFL, MAIL, MSND, MSOM,
        MSAM, MRSQ, MRCP, ALLO, REST, RNFR,
        RNTO, ABOR, DELE, CWD,  LIST, NLST,
        SITE, SYST, STAT, HELP, NOOP, MKD,
        RMD,  PWD,  CDUP, STOU, SIZE, MDTM,
        // others
        TOK_A, TOK_B, TOK_C, TOK_E, TOK_F,
        TOK_I, TOK_L, TOK_N, TOK_P, TOK_R,
        TOK_S, TOK_T, TOK_SP,
        FORM_N,
        TOK_CRLF, COMMA, STRING, NUMBER,
        // signals scan error
        LEXERR
    };
}; // class wd221Token

    
class twd221_FtpCmd {
public:
    twd221_FtpCmd( twd221_FtpCmdType        cmdType = -1,
                   const ZString           &arg1 = "",
                   const ZString           &arg2 = "",
                   const twd201_SockAddrIn *sockAddrArg = NULL,
                   const sapdbwa_Int4         codeArg1 = -1,
                   const sapdbwa_Int4         codeArg2 = -1 );
    ~twd221_FtpCmd();

    void Init( twd221_FtpCmdType        cmdType,
               const ZString           &arg1 = "",
               const ZString           &arg2 = "",
               const twd201_SockAddrIn *sockAddrArg = NULL,
               const sapdbwa_Int4         codeArg1 = -1,
               const sapdbwa_Int4         codeArg2 = -1 );
            
    twd221_FtpCmdType CmdType() { return this->cmdType; };
    const ZString& Arg1() { return this->arg1; };
    const ZString& Arg2() { return this->arg2; };
    const twd201_SockAddrIn& SockAddrArg() { return this->sockAddrArg; };
    const sapdbwa_Int4 CodeArg1() { return this->codeArg1; };
    const sapdbwa_Int4 CodeArg2() { return this->codeArg2; };
private:
    twd221_FtpCmdType  cmdType;
    ZString            arg1;
    ZString            arg2;
    twd201_SockAddrIn  sockAddrArg;
    sapdbwa_Int4         codeArg1;
    sapdbwa_Int4         codeArg2;
}; // class twd221_FtpCmd


/*!**********************************************************************

  EndChapter: twd221_FtpCmd
  
************************************************************************/
/*!**********************************************************************

  Chapter: twd221_FtpParser

  description:
  
************************************************************************/

struct ParserCmdEntry;

class twd221_FtpParser {
public:
    twd221_FtpParser( twd201_Socket socket, sapdbwa_Bool debugMode )
        : socket(socket), parsePos(0), state(0), debug(debugMode),
        timeout(sapdbwa_False) {};

    sapdbwa_Bool GetNextCmd( twd221_FtpCmd &ftpCmd, sapdbwa_Bool &timeout );

private:
    twd201_Socket socket;
    sapdbwa_Int4    state;
    char          recvBuf[1000];
    sapdbwa_Int4    parsePos;
    sapdbwa_Int4    recvPos;
    sapdbwa_Bool    debug; // in debug mode we use syslog to log commands
    sapdbwa_Bool    timeout; // sapdbwa_True if timeout occured
    
    // parsing
    ParserCmdEntry* GetCmdEntry();
    ParserCmdEntry* FindCmdEntry( const ZString &cmd );
    sapdbwa_Bool GetPort( twd201_SockAddrIn &sockAddr );
    sapdbwa_Bool GetTypeCode( sapdbwa_Int4 &typeCode, sapdbwa_Int4 &formCode );
    sapdbwa_Bool GetStructCode( sapdbwa_Int4 &structCode );
    sapdbwa_Bool GetModeCode( sapdbwa_Int4 &modeCode );
    
    // scanning
    sapdbwa_Int4 GetNextToken( ZString &value );
    sapdbwa_Bool SkipSpace();
    sapdbwa_Bool GetString( ZString &value );
    sapdbwa_Bool GetNumber( ZString &value );
    void SkipCmd();

    // socket comm
    void NewCmd();
    sapdbwa_Int4 Read();
    sapdbwa_Bool Select();
    sapdbwa_Int4 Receive( char *buf, sapdbwa_Int4 bufLen );
}; // class FtpParser

/*!**********************************************************************

  EndChapter: twd221_FtpParser

************************************************************************/

#endif
