/*!**********************************************************************

  module: vwd120format.c

  -----------------------------------------------------------------------

  responsible:  Holger Bischoff
				Markus Oezgen

  special area: WebAgent

  description:  DB-Filesystem dbformat executable

  see also:

  -----------------------------------------------------------------------

                         Copyright by SAP AG, 1998



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

************************************************************************/

/*=======================================================================
 *  INCLUDES
 *=====================================================================*/

#include <stdio.h>

#include "hwd101dbfs.h"
#include "hwd128help.h"

tsp00_Bool wd120_ReallyFormat();

int main( int argc, char *argv[] )
{
    if (wd128HelpOption( argc, argv )) {
        wd128ShowHelp( DBFORMAT_HELP_WD128,
                       DBFORMAT_USAGE_WD128,
                       HELP_ONLY_OPTIONS_WD128 );
    } else if (argc != 1) {
        wd128ShowUsage( DBFORMAT_USAGE_WD128, HELP_ONLY_OPTIONS_WD128 );
    } else {
        if (wd120_ReallyFormat()) {
            if (!wd101InitDBFS()) {
                printf("Init failed: ");
                printf( "%s \n", wd101CmdGetErrorMsg() );
                
                return 1;
            }; /* if */

            if (!wd101CmdFormatDBFS()) {
                printf( "%s \n", wd101CmdGetErrorMsg() );
            }; /* if */
        
            wd101ExitDBFS();
        }; /* if */
    }; /* else */
    
    return 0;
} /* main */


tsp00_Bool wd120_ReallyFormat()
{
    char key = 'n';
    
    printf( "This will erase all existing files in the dbfs. Really format [y/n]?" );
    key = getc(stdin);

    return (key == 'y' || key == 'Y');
} /* wd128_ReallyFormat */
