/*!**********************************************************************

  module: vwd201sock.c

  -----------------------------------------------------------------------

  responsible:  Holger Bischoff
				Markus Oezgen

  special area: WebAgent

  description:  Socket interface

  see also:

  -----------------------------------------------------------------------

                         Copyright by SAP AG, 1999



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

************************************************************************/

#ifndef WIN32
#endif

#include "hwd201sock.h"
#include "hwd202msg.h"

/* WinSock version information (use WSA_VERSION for WSAStartup()) */
#ifndef MAKEWORD
#define MAKEWORD(l,h) ((WORD)(((BYTE)(l))|(((WORD)(BYTE)(h))<<8)))
#endif
#define WSA_MAJOR_VERSION 1
#define WSA_MINOR_VERSION 1
#define WSA_VERSION MAKEWORD(WSA_MAJOR_VERSION, WSA_MINOR_VERSION)

sapdbwa_Int4 wd201SocketStartup()
{
    /* only required under win32 */
#ifdef WIN32    
    sapdbwa_Int4 ret;
    WSADATA    stWSAData;
    
    ret = WSAStartup( WSA_VERSION, &stWSAData );

    return ret;
#else    
    return 0;
#endif
} /* wd201SocketStartup */


sapdbwa_Int4 wd201SocketCleanup()
{
    /* only required under win32 */
#ifdef WIN32    
    sapdbwa_Int4 ret;
    
    ret = WSACleanup();

    return ret;
#else
    return 0;
#endif    
} /* wd201SocketCleanup */


twd201_SockErr wd201GetLastError()
{
    twd201_SockErr sockErr;

#ifdef WIN32    
    sockErr = WSAGetLastError();
#else
    sockErr = errno;
#endif
    
    return sockErr;
} /* wd201GetLastError */


const unsigned char* wd201StrError( twd201_SockErr sockErr )
{
    switch (sockErr) {
    case 11001: return (unsigned char*) ERR_11001_WD202; break;
    case 11002: return (unsigned char*) ERR_11002_WD202; break;
    case 11003: return (unsigned char*) ERR_11003_WD202; break;
    case 11004: return (unsigned char*) ERR_11004_WD202; break;
                               
    case 10091: return (unsigned char*) ERR_10091_WD202; break;
    case 10092: return (unsigned char*) ERR_10092_WD202; break;
    case 10093: return (unsigned char*) ERR_10093_WD202; break;
                               
    case 10064: return (unsigned char*) ERR_10064_WD202; break;
    case 10065: return (unsigned char*) ERR_10065_WD202; break;
    case 10066: return (unsigned char*) ERR_10066_WD202; break;
    case 10067: return (unsigned char*) ERR_10067_WD202; break;
    case 10068: return (unsigned char*) ERR_10068_WD202; break;
    case 10069: return (unsigned char*) ERR_10069_WD202; break;
    case 10070: return (unsigned char*) ERR_10070_WD202; break;
    case 10071: return (unsigned char*) ERR_10071_WD202; break;
                               
    case 10048: return (unsigned char*) ERR_10048_WD202; break;
    case 10049: return (unsigned char*) ERR_10049_WD202; break;
    case 10050: return (unsigned char*) ERR_10050_WD202; break;
    case 10051: return (unsigned char*) ERR_10051_WD202; break;
    case 10052: return (unsigned char*) ERR_10052_WD202; break;
    case 10053: return (unsigned char*) ERR_10053_WD202; break;
    case 10054: return (unsigned char*) ERR_10054_WD202; break;
    case 10055: return (unsigned char*) ERR_10055_WD202; break;
    case 10056: return (unsigned char*) ERR_10056_WD202; break;
    case 10057: return (unsigned char*) ERR_10057_WD202; break;
    case 10058: return (unsigned char*) ERR_10058_WD202; break;
    case 10059: return (unsigned char*) ERR_10059_WD202; break;
    case 10060: return (unsigned char*) ERR_10060_WD202; break;
    case 10061: return (unsigned char*) ERR_10061_WD202; break;
    case 10062: return (unsigned char*) ERR_10062_WD202; break;
    case 10063: return (unsigned char*) ERR_10063_WD202; break;
                               
    case 10035: return (unsigned char*) ERR_10035_WD202; break;
    case 10036: return (unsigned char*) ERR_10036_WD202; break;
    case 10037: return (unsigned char*) ERR_10037_WD202; break;
    case 10038: return (unsigned char*) ERR_10038_WD202; break;
    case 10039: return (unsigned char*) ERR_10039_WD202; break;
    case 10040: return (unsigned char*) ERR_10040_WD202; break;
    case 10041: return (unsigned char*) ERR_10041_WD202; break;
    case 10042: return (unsigned char*) ERR_10042_WD202; break;
    case 10043: return (unsigned char*) ERR_10043_WD202; break;
    case 10044: return (unsigned char*) ERR_10044_WD202; break;
    case 10045: return (unsigned char*) ERR_10045_WD202; break;
    case 10046: return (unsigned char*) ERR_10046_WD202; break;
    case 10047: return (unsigned char*) ERR_10047_WD202; break;
                               
    case 10022: return (unsigned char*) ERR_10022_WD202; break;
    case 10024: return (unsigned char*) ERR_10024_WD202; break;
                               
    case 10000: return (unsigned char*) ERR_10000_WD202; break;
    case 10004: return (unsigned char*) ERR_10004_WD202; break;
    case 10009: return (unsigned char*) ERR_10009_WD202; break;
    case 10013: return (unsigned char*) ERR_10013_WD202; break;
    case 10014: return (unsigned char*) ERR_10014_WD202; break;
    default: return (unsigned char*) ERR_UNKNOWN_WD202; break;
    }; /* switch */
} /* wd201StrError */


twd201_SockRet wd201GetHostname( char *name, sapdbwa_Int4 nameLen )
{
    twd201_SockRet sockRet;

    sockRet = gethostname( name, nameLen );

    return sockRet;
} /* wd201GetHostname */


twd201_SockRet wd201GetPeername( twd201_Socket    socket,
                                 twd201_SockAddr *sockName,
                                 sapdbwa_Int4      *nameLen )
{
    twd201_SockRet sockRet;

    sockRet = getpeername( socket, sockName, nameLen );
    
    return sockRet;
} /* wd201GetPeername */


twd201_SockRet wd201GetSockName( twd201_Socket    socket,
                                 twd201_SockAddr *sockName,
                                 sapdbwa_Int4      *nameLen )
{
    twd201_SockRet sockRet;

    sockRet = getsockname( socket, sockName, nameLen );
    
    return sockRet;
} /* wd201GetSockName */


twd201_Socket wd201Socket( sapdbwa_Int4  af,
                           sapdbwa_Int4  type,
                           sapdbwa_Int4  protocol )
{
    twd201_Socket newSocket;

    newSocket = socket( af, type, protocol );

    return newSocket;
} /* wd201Socket */


twd201_SockRet wd201Bind( twd201_Socket    socket,
                          twd201_SockAddr *pAddr,
                          sapdbwa_Int4       nameLen )
{
    twd201_SockRet sockRet;

    sockRet = bind( socket, pAddr, nameLen );

    return sockRet;
} /* wd201Bind */


twd201_SockRet wd201CloseSocket( twd201_Socket    socket )
{
    twd201_SockRet sockRet;

#ifdef WIN32    
    sockRet = closesocket( socket );
#else
    /* under UNIX a socket is a file-dscriptor */
    sockRet = close( socket );
#endif
    
    return sockRet;
} /* wd201CloseSocket */


twd201_SockRet wd201Listen( twd201_Socket  socket,
                            sapdbwa_Int4     backlog )
{
    twd201_SockRet sockRet;

    sockRet = listen( socket, backlog );

    return sockRet;
} /* wd201Listen */


twd201_Socket wd201Accept( twd201_Socket    lstnSocket,
                           twd201_SockAddr *pAddr,
                           sapdbwa_Int4      *pNameLen )
{
    twd201_Socket newSocket;

    newSocket = accept( lstnSocket, pAddr, pNameLen );

    return newSocket;    
} /* wd201Accept */


twd201_SockRet wd201Send( twd201_Socket  socket,
                          const char    *buf,
                          sapdbwa_Int4     len,
                          sapdbwa_Int4     flags )
{
    twd201_SockRet sockRet;

    sockRet = send( socket, buf, len, flags );

    return sockRet;
} /* wd201Send */


twd201_SockRet wd201Recv( twd201_Socket  socket,
                          char          *buf,
                          sapdbwa_Int4     len,
                          sapdbwa_Int4     flags )
{
    twd201_SockRet sockRet;

    sockRet = recv( socket, buf, len, flags );

    return sockRet;
} /* wd201Recv */


twd201_SockRet wd201SetSockOpt( twd201_Socket socket,
                                sapdbwa_Int4    level,
                                sapdbwa_Int4    opname,
                                const char   *optval,
                                sapdbwa_Int4    optlen )
{
    twd201_SockRet sockRet;

    sockRet = setsockopt( socket, level, opname, optval, optlen );

    return sockRet;
} /* wd201SetSockOpt */


twd201_SockRet wd201GetSockOpt( twd201_Socket socket,
                                sapdbwa_Int4    level,
                                sapdbwa_Int4    opname,
                                char         *optval,
                                sapdbwa_Int4   *optlen )
{
    twd201_SockRet sockRet;

    sockRet = getsockopt( socket, level, opname, optval, optlen );

    return sockRet;    
} /* wd201GetSockOpt */


twd201_SockRet wd201Connect( twd201_Socket          socket,
                             const twd201_SockAddr *name,
                             int                    nameLen )
{
    twd201_SockRet sockRet;

    sockRet = connect( socket, name, nameLen );

    return sockRet;
} /* wd201Connect */


twd201_SockRet wd201Select( sapdbwa_Int4            nfds,
                            fd_set               *readfds,
                            fd_set               *writefds,
                            fd_set               *exceptfds,
                            const struct timeval *timeout )
{
    twd201_SockRet sockRet;

    sockRet = select( nfds, readfds, writefds, exceptfds,
                      (struct timeval *) timeout );

    return sockRet;
} /* wd201Select */
    
    

