/*!**********************************************************************

  module: vwd203util.c

  -----------------------------------------------------------------------

  responsible:  Holger Bischoff
				Markus zgen

  special area: WebAgent

  description:  Utility functions (UNIX compatibility etc.)

  see also:

  -----------------------------------------------------------------------

                         Copyright by SAP AG, 1999



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

************************************************************************/

/*=======================================================================
 *  INCLUDES
 *=====================================================================*/

#ifndef WIN32
#ifdef LINUX
#include <err.h>    /* warn */
#endif
#include <unistd.h> /* sleep */
#endif

#include "hwd203util.h"


void syslog (int n, const char *fmt, ...)
{
    va_list  ap;    

    va_start(ap, fmt);
    vprintf( fmt, ap );
} /* syslog */


void vsyslog( int n, const char *fmt, va_list ap )
{
    vprintf( fmt, ap );
    printf( "\n" );
} /* vsyslog */


void wd203makeIPAddr( twd201_SockAddrIn sockName, char *buf, sapdbwa_Int4 bufLen )
{
    sprintf( buf, "%d.%d.%d.%d:%d",
#ifdef WIN32             
             sockName.sin_addr.S_un.S_un_b.s_b1,
             sockName.sin_addr.S_un.S_un_b.s_b2,
             sockName.sin_addr.S_un.S_un_b.s_b3,
             sockName.sin_addr.S_un.S_un_b.s_b4,
#else
             ((sapdbwa_UInt1*) &sockName.sin_addr)[0],
             ((sapdbwa_UInt1*) &sockName.sin_addr)[1],
             ((sapdbwa_UInt1*) &sockName.sin_addr)[2],
             ((sapdbwa_UInt1*) &sockName.sin_addr)[3],
#endif             
             sockName.sin_port );
} /* makeIPAddr */


void wd203SockErr( const char *where, const char *sockFunc )
{
    twd201_SockErr  sockErr = wd201GetLastError();
    
    printf( "Error in %s %s: %s (%d)\n", where, sockFunc,
            wd201StrError( sockErr ), sockErr );
#ifdef LINUX
    warn( "%s", sockFunc );
#endif    
} /* wd203SockErr */


void wd203Error( const char *where, const char *sockFunc, const char *errMsg )
{
    printf( "Error in %s %s: %s\n", where, sockFunc, errMsg );
} /* wd210Error */


sapdbwa_Bool wd203Accept( twd201_Socket      lstnSocket,
                  twd201_Socket     *pIncomingSock,
                  twd201_SockAddrIn *pIncomingSockName)
{
    twd201_SockRet sockRet = SOCKET_ERROR;
    sapdbwa_Bool           ok = sapdbwa_True;
    sapdbwa_Int4     nameLen = SOCKADDR_LEN;

    *pIncomingSock = wd201Accept( lstnSocket,
                                  (twd201_SockAddr*) pIncomingSockName,
                                  &nameLen);
    if (*pIncomingSock == INVALID_SOCKET) {
        wd203SockErr( "wd210Accept", "Accept" );
        ok = sapdbwa_False;
    }; /* if */

    return ok;
} /* wd203Accept */


void wd203Sleep( sapdbwa_UInt4 msec )
{
#ifdef WIN32    
    Sleep( msec );
#else
    sleep( msec / 1000 );
#endif    
} /* wd203Sleep */

