/*!**********************************************************************

  module: vwd51ns.c

  -----------------------------------------------------------------------

  responsible:  Holger Bischoff
				Markus Oezgen

  special area: WebAgent

  description:  Functions for Netscape's Web Server

  see also:

  -----------------------------------------------------------------------

                         Copyright by SAP AG, 1998



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

************************************************************************/

/*=======================================================================
 *  INCLUDES
 *=====================================================================*/

#ifdef WIN32
#define XP_WIN32
#define NET_WINSOCK
#else
#define XP_UNIX
#endif
#include <nsapi.h>

#ifdef WIN32
#include <WINDOWS.H>
#endif

#include "gwd00.h"
#include "hwd03waapi.h"
#include "hwd20control.h"
#include "hwd21req.h"
#include "hwd22rep.h"
#include "hwd28dict.h"
#include "hwd29date.h"

/*!**********************************************************************

  Chapter: Filter-Functions

  description: These functions are called by the Enterprise-Server.
  
************************************************************************/

/* NSAPIWebAgentInit is called at startup of the webserver.
 * Here we initialize the WebAgent.
 */
NSAPI_PUBLIC int NSAPIWebAgentInit( pblock *pb, Session *sn, Request *rq );

void NSAPIWebAgentExit( void *data );

/* This functions is called by the WebServer for each HTTP-request
 * that comes in. Here we retrieve the request information and call
 * the WebAgent service function.
 */
NSAPI_PUBLIC int NSAPIWebAgentEntry( pblock         *pb,
                                     Session        *sn,
                                     Request        *rq );

/*!**********************************************************************

  EndChapter: Filter-Functions

************************************************************************/

/*!**********************************************************************

  Chapter: DLL-Main function

  description: We have to clean up resources when exiting
  
************************************************************************/
#if 0
#ifdef _WIN32
BOOL APIENTRY DllMain( HANDLE	hModule,
                       DWORD    dwReason,
                       LPVOID   lpReserved)
{
    switch (dwReason) {
    case DLL_PROCESS_ATTACH:
        break;
    case DLL_PROCESS_DETACH:
        wd20ApiExit();
        break;
    default:
        break;
    }; /* switch */
}; /* DllMain */
#endif /* _WIN32 */
#endif

/*!**********************************************************************

  EndChapter: DLL-Main function

************************************************************************/

/*!**********************************************************************

  Chapter: Call-Back Functions

  description: These functions are stored in the WAHttpRequest and
               WAHttpReply structures. They are used by the WebAgent
               to retrieve body-data from, and send the header and
               content to the web-server.
  
************************************************************************/

sapdbwa_UInt4 wd51MarcsReadBody( twd00ServerHandle   handle,
                               char               *b,
                               sapdbwa_UInt4         len );


const char* wd51MarcsGetHeader( twd00ServerHandle   handle,
                                const char         *fieldName );

sapdbwa_Bool wd51MarcsSendHeader( twd00ServerHandle   handle,
                                sapdbwa_HttpReplyP       rep );


sapdbwa_UInt4 wd51MarcsSendBody( twd00ServerHandle   handle,
                               const void         *Buf,
                               sapdbwa_UInt4         len );

/*!**********************************************************************

  EndChapter: Call-Back Functions

************************************************************************/

/*!**********************************************************************

  Chapter: Internal Functions

************************************************************************/

#define WRITEBUF_SZ		4096

struct st_marcs_conn
{
    char        buf[WRITEBUF_SZ];
    size_t		buflg;
    size_t		bufsz;
};

typedef struct st_marcs_req
{
    pblock      *pb;
    Session     *sn;
    Request     *rq;
    struct st_marcs_conn marcsConn;
    sapdbwa_UInt4  contentLength;
} *twd51MarcsServerHandle;

#ifdef WNT
#define NSAPI_PUBLIC __declspec(dllexport)
#elif WDUNIX
#define NSAPI_PUBLIC
#endif


# define MAX_URI_SZ 1024
# define MAX_PATH_INFO_SZ 1024
# define MAX_REQUEST_METHOD_SZ 1024
# define MAX_CONTENT_TYPE_SZ 1024
# define MAX_CONTENT_LENGTH_SZ 1024
# define MAX_PHYSICALPATH_SZ 1024
# define MAX_SERVERNAME_SZ 1024

twd51MarcsServerHandle wd51_CreateServerHandle( pblock  *pb,
                                                Session *sn,
                                                Request *rq );

void wd51_DestroyServerHandle( twd51MarcsServerHandle handle );

sapdbwa_Bool wd51_MarcsReadPostInit( twd51MarcsServerHandle  handle,
                                   twd21HttpRequestP        req );

sapdbwa_UInt4 wd51_MarcsReadPost( twd51MarcsServerHandle  handle,
                                char                   *Buffer,
                                sapdbwa_UInt4             BufferSize );

sapdbwa_UInt4 wd51_MarcsWrite( twd51MarcsServerHandle   handle,
                             void                    *b,
                             sapdbwa_UInt4              n );

sapdbwa_Bool wd51_MarcsWriteHeader( twd51MarcsServerHandle handle,
                                  sapdbwa_HttpReplyP          rep );

sapdbwa_Int4 wd51_MarcsRequestMap( twd51MarcsServerHandle handle );

sapdbwa_Bool wd51_MarcsFlush( twd51MarcsServerHandle handle );

/* convert CONTENT_TYPE to content-type */
void wd51ConvertFieldName( const char  *waFieldName,
                           char        *nsFieldName,
                           sapdbwa_UInt2  nsSize );

/*!**********************************************************************

  EndChapter: Internal Functions

************************************************************************/

/*!**********************************************************************

  Chapter: Filter-Functions

************************************************************************/

NSAPI_PUBLIC int NSAPIWebAgentInit( pblock *pb, Session *sn, Request *rq )
{

    daemon_atrestart( NSAPIWebAgentExit, NULL );
    
    if (wd20ApiInit(NULL, API_NSAPI_WD20 )) {
        wd20WriteLogMsg("NSAPIWebAgentInit success.\n");
        return(REQ_PROCEED);
    } else {
        return(REQ_ABORTED);
    }; /* else */
} /* NSAPIWebAgentInit */


void NSAPIWebAgentExit( void *data )
{
    wd20ApiExit();
} /* NSAPIWebAgentExit */

NSAPI_PUBLIC int NSAPIWebAgentEntry( pblock         *pb,
                                     Session        *sn,
                                     Request        *rq )
{
    twd51MarcsServerHandle handle;
    sapdbwa_Int4             rtc = REQ_NOACTION;

    /*
    if (!wd20ApiReInit()) {
        return rtc;
    };
    */
    
    handle = wd51_CreateServerHandle( pb, sn, rq );
    if (handle == NULL) {
        /* ... log error */
        return rtc;
    }; /* if */
    
    rtc = wd51_MarcsRequestMap( handle );
    wd51_DestroyServerHandle( handle );

    return rtc;
} /* NSAPIWebAgentEntry */

/*!**********************************************************************

  EndChapter: Filter-Functions

************************************************************************/

/*!**********************************************************************

  Chapter: Call-Back Functions

************************************************************************/

sapdbwa_UInt4 wd51MarcsReadBody( twd00ServerHandle   handle,
                               char               *b,
                               sapdbwa_UInt4         len )
{
    twd51MarcsServerHandle marcsHandle =
        (twd51MarcsServerHandle) handle;
    
    return wd51_MarcsReadPost( marcsHandle, b, len );
} /* wd51MarcsReadBody */


const char* wd51MarcsGetHeader( twd00ServerHandle   handle,
                                const char         *fieldName )
{
    char convertedFieldName[1024];
    twd51MarcsServerHandle marcsHandle = (twd51MarcsServerHandle) handle;
    const char *value = NULL;
    
    if (fieldName != NULL) {
        wd51ConvertFieldName( fieldName,
                              convertedFieldName,
                              sizeof( convertedFieldName ));
        /* try to find entry in headers, vars, srvhedrs, reqpb */
        value = pblock_findval( convertedFieldName, marcsHandle->rq->headers );
        if (value == NULL) {
            value = pblock_findval( convertedFieldName,
                                    marcsHandle->rq->vars );
        };            
        if (value == NULL) {
            value = pblock_findval( convertedFieldName,
                                    marcsHandle->rq->srvhdrs );
        };            
        if (value == NULL) {
            value = pblock_findval( convertedFieldName,
                                    marcsHandle->rq->reqpb );
        };

        return value;
    } else {
        return NULL;
    }; /* else */
} /* wd51MarcsGetHeader */


sapdbwa_Bool wd51MarcsSendHeader( twd00ServerHandle   handle,
                                sapdbwa_HttpReplyP       rep )
{
    twd51MarcsServerHandle marcsHandle =
        (twd51MarcsServerHandle) handle;
    
    return wd51_MarcsWriteHeader( marcsHandle, rep );
} /* wd51MarcsSendHeader */


sapdbwa_UInt4 wd51MarcsSendBody( twd00ServerHandle   handle,
                               const void         *Buf,
                               sapdbwa_UInt4         len )
{
    twd51MarcsServerHandle marcsHandle =
        (twd51MarcsServerHandle) handle;
    
    return wd51_MarcsWrite( marcsHandle, (void*) Buf, len );
} /* wd51MarcsSendBody */


sapdbwa_Bool wd51MarcsIsSecure( twd00ServerHandle   handle )
{
    sapdbwa_Bool isSecure = (sapdbwa_Bool) security_active;
    
    return ( isSecure );
} /* wd51MarcsIsSecure */


/*!**********************************************************************

  EndChapter: Call-Back Functions

************************************************************************/

/*!**********************************************************************

  Chapter: Internal Functions

************************************************************************/

twd51MarcsServerHandle wd51_CreateServerHandle( pblock  *pb,
                                                Session *sn,
                                                Request *rq )
{
    sapdbwa_Bool              allocatOK;
    twd51MarcsServerHandle  newHandle;

    sapdbwa_SQLALLOCAT( sizeof( struct st_marcs_req ),
                (sapdbwa_UInt1**) &newHandle,
                &allocatOK );
    if (allocatOK) {
        newHandle->pb = pb;
        newHandle->sn = sn;
        newHandle->rq = rq;
        newHandle->marcsConn.buflg = 0;
        newHandle->marcsConn.bufsz = WRITEBUF_SZ;
        newHandle->contentLength   = 0;
    } else {
        newHandle = NULL;
    }; /* else */

    return newHandle;
} /* wd51_CreateServerHandle */


void wd51_DestroyServerHandle( twd51MarcsServerHandle handle )
{
    sapdbwa_SQLFREE( (sapdbwa_UInt1*) handle );
} /* wd51_DestroyServerHandle */


sapdbwa_Bool wd51_MarcsReadPostInit( twd51MarcsServerHandle  handle,
                                   twd21HttpRequestP       req )
{
    char        *method;
    int          clen;
    char        *clstr;

    method = pblock_findval("method", handle->rq->reqpb );
    handle->contentLength = 0;

    if (strcmp(method, "POST") && strcmp(method, "PUT"))
        return(true);

    /* if there's no content-length, it's a bogus request */
    if ( request_header( "content-length",
                         &clstr,
                         handle->sn,
                         handle->rq ) == REQ_ABORTED || clstr == NULL)
        {
            return(false);
        }
        
    clen = atoi(clstr);
    /* if the content length is 0 or negative, it's a bogus request */

    if(clen<1)
        {
            return(false);
        }

    handle->contentLength = clen;

    return(true);
} /* wd51_MarcsReadPostInit */


sapdbwa_UInt4 wd51_MarcsReadPost( twd51MarcsServerHandle  handle,
                                char                   *Buffer,
                                sapdbwa_UInt4             BufferSize )
{
    sapdbwa_UInt4 rtc = 0;

    if (handle->contentLength == 0)
        return(0);

    while (handle->contentLength > 0 && BufferSize > 0)
        {
            *Buffer = netbuf_getc( handle->sn->inbuf );
        
            /* check for error in reading */
            if(*Buffer == IO_ERROR || *Buffer == IO_EOF)
                break;
 
            Buffer++;
            BufferSize--;
            handle->contentLength--;
            rtc++;
        }

    return(rtc);
} /* wd51_MarcsReadPost */


sapdbwa_UInt4 wd51_MarcsWrite( twd51MarcsServerHandle   handle,
                             void                    *b,
                             sapdbwa_UInt4              n )
{
    Session *Sn = handle->sn;
    int      rtc;
    size_t   rlg = 0, totn = 0;

    if (handle->marcsConn.bufsz == 0)
        {
            rtc = net_write( Sn->csd, b, n );
            if (rtc == IO_ERROR)
                return(0);
            return(n);
        }

    while (n > 0)
        {
            if (handle->marcsConn.bufsz == handle->marcsConn.buflg)
                {
                    rtc = net_write(Sn->csd,
                                    handle->marcsConn.buf,
                                    handle->marcsConn.buflg);
                    if (rtc == IO_ERROR)
                        return(0);
                    else
                        handle->marcsConn.buflg = 0;
                }

            rlg = min( n,
                       handle->marcsConn.bufsz
                       - handle->marcsConn.buflg );
            memcpy( (char*) handle->marcsConn.buf
                    + handle->marcsConn.buflg,
                    b,
                    rlg );
            handle->marcsConn.buflg += rlg;
            b = (char*) b + rlg;
            n -= rlg;
            totn += rlg;
        }
    return ((sapdbwa_UInt4) totn);
} /* wd51_MarcsWrite */

/*********************************************************/

sapdbwa_Bool wd51_MarcsWriteHeader( twd51MarcsServerHandle  handle,
                                  sapdbwa_HttpReplyP           rep )
{
    char *location      = wd22GetLocation( rep );
    char *lastModified  = wd22GetLastModified( rep );
    char *contentLength = wd22GetContentLength( rep );
    char *contentType   = wd22GetContentType( rep );
    char *expires       = wd22GetExpires( rep );
    char  date[MAX_DATE_LEN_WD29] = "";

    sapdbwa_UInt4       iHeader;
    twd28DictionaryP  headers = wd22GetHeaders( rep );
    const char       *name;
    char             *value;

    if (contentType != NULL) {
        /* We need to get rid of the internal content type. */
        param_free( pblock_remove("content-type", handle->rq->srvhdrs));

        /* Set content-type and begin response */
        pblock_nvinsert( "content-type",
                         contentType,
                         handle->rq->srvhdrs);
    }; /* if */
    if (contentLength != NULL) {
        pblock_nvinsert( "content-length",
                         contentLength,
                         handle->rq->srvhdrs);
    }; /* if */
    if (location != NULL) {
        pblock_nvinsert( "location",
                         location,
                         handle->rq->srvhdrs);
    }; /* if */
    if (lastModified != NULL) {
        if (wd29WADateToHttpDate( lastModified, date ) != 0) {        
            pblock_nvinsert( "last-modified",
                             date,
                             handle->rq->srvhdrs);
        }; /* if */
    }; /* if */
    if (expires != NULL) {
        if (wd29WADateToHttpDate( expires, date ) != 0) {
            pblock_nvinsert( "expires",
                             date,
                             handle->rq->srvhdrs);
        }; /* if */
    }; /* if */

    /* insert all additional header fields */
    for (iHeader = 0; iHeader < wd28NumEntries( headers ); iHeader++) {
        if (wd28GetEntryByIndex( headers, iHeader, &name, &value )) {
            pblock_nvinsert( name,
                             value,
                             handle->rq->srvhdrs);
        }; /* if */
    }; /* for */

    protocol_status( handle->sn, handle->rq, PROTOCOL_OK, NULL);
    
    /* A noaction response from this function means the request was HEAD */
    if (protocol_start_response( handle->sn, handle->rq) == REQ_NOACTION) {
            /*
            WebDbLog("protocol_start_response was HEAD, no body is sent.");
            */
            return(false);
    } else {
        return(true);
    }; /* else */
} /* wd51_MarcsWriteHeader */

/*********************************************************/

sapdbwa_Bool wd51_MarcsFlush( twd51MarcsServerHandle handle )
{
    Session         *Sn = handle->sn;
    sapdbwa_Int4       rtc;

    if (handle->marcsConn.buflg > 0)
        rtc = net_write( Sn->csd,
                         handle->marcsConn.buf,
                         handle->marcsConn.buflg);
    if (rtc == IO_ERROR)
        return(1);

    handle->marcsConn.buflg = 0;
    return(0);
} /* wd51_MarcsFlush */


sapdbwa_Int4 wd51_MarcsRequestMap( twd51MarcsServerHandle handle )
{
    twd21HttpRequestP  req = NULL;
    sapdbwa_HttpReplyP      rep = NULL;

    sapdbwa_Int4  rtc = REQ_NOACTION;
    char        marcs_buf[WRITEBUF_SZ];
    char       *contentLength;
    char       *contentType;
    
    req = wd21CreateRequest( handle,
                             wd51MarcsReadBody,
                             wd51MarcsGetHeader,
                             wd51MarcsIsSecure );
    if (req == NULL) {
        /* memory allocation error */
        /* ... log error */
        return rtc;
    }; /* if */

    rep = wd22CreateReply( handle, wd51MarcsSendHeader, wd51MarcsSendBody );
    if (rep == NULL) {
        /* memory allocation error */
        /* ... log error */
        wd21DestroyRequest( req );        
        return rtc;
    }; /* if */

    wd21SetUri( req, pblock_findval( "uri", handle->rq->reqpb ));    
    wd21SetIfModifiedSince( req, pblock_findval( "if-modified-since",
                                                 handle->rq->headers) );
    wd21SetQueryString( req, pblock_findval("query", handle->rq->reqpb) );
    wd21SetPathInfo( req, NULL );
    wd21SetRequestMethod( req, pblock_findval("method", handle->rq->reqpb) );
    
    if (request_header( "content-type",
                        &contentType,
                        handle->sn,
                        handle->rq) == REQ_ABORTED) {
        contentType = NULL;
    }; /* if */
    wd21SetContentType( req, contentType );

    if ( request_header( "content-length",
                         &contentLength,
                         handle->sn,
                         handle->rq) == REQ_ABORTED) {
        contentLength = NULL;
    }; /* if */
    wd21SetContentLength( req, contentLength );
    
    wd21SetPhysicalPath( req, pblock_findval( "path", handle->rq->vars ) );
    wd21SetServerName( req, pblock_findval( "host", handle->rq->headers ) );
    
    wd51_MarcsReadPostInit( handle, req );

    /*
    my_h.Context->file_handle = &h;
    my_h.WriteHeader = WriteHeader;
    my_h.WriteBody = WriteBody;
    */

    switch ( wd20CallServiceFunc( req, rep )) {
    case sapdbwa_RTC_Ignored:
        rtc = REQ_NOACTION;
        break;
    case sapdbwa_RTC_Accepted:
        rtc = REQ_PROCEED;
        break;
    case sapdbwa_RTC_Error:
        rtc = REQ_ABORTED;
        protocol_status( handle->sn, handle->rq,
                         PROTOCOL_SERVER_ERROR, NULL);
        break;
    default:
        rtc = REQ_ABORTED;
        protocol_status( handle->sn, handle->rq,
                         PROTOCOL_SERVER_ERROR, NULL);
        break;
    }; /* switch */
    wd51_MarcsFlush( handle );

    wd21DestroyRequest( req );
    wd22DestroyReply( rep );
    
    return(rtc);
} /* wd51_MarcsRequestMap */


void wd51ConvertFieldName( const char  *waFieldName,
                           char        *nsFieldName,
                           sapdbwa_UInt2  nsSize )
{
    sapdbwa_UInt2 pos = 0;
    sapdbwa_UInt2 outPos = 0;
    
    if (strcmp( waFieldName, "QUERY_STRING" ) == 0) {
        strcpy( nsFieldName, "query" );
    } else {
        if (strstr( waFieldName, "HTTP_" ) == waFieldName) {
            /* skip leading "HTTP_" */
            pos = strlen( "HTTP_" );
        };
        while (outPos+1 < nsSize && waFieldName[pos] != '\0') {
            if (waFieldName[pos] == '_') {
                nsFieldName[outPos] = '-';
            } else {
                nsFieldName[outPos] = tolower( waFieldName[pos] );
            }; /* else */
            pos++;
            outPos++;
        }; /* while */

        if (nsSize > 0)
            nsFieldName[pos] = '\0';
    };
} /* wd51ConvertFieldName */

/*!**********************************************************************

  EndChapter: Internal Functions

************************************************************************/
