/****************************************************************************

  module      : veo670.c

  -------------------------------------------------------------------------

  responsible : JoergM, RaymondR

  special area: Stack Back Trace
  description : Interface that allows to produce a stack backtrace


  last changed: 2001-11-21  10:33

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

/* PTS 1108958 */
#undef _FILE_OFFSET_BITS 
#include "heo670.h"

#if defined(_WIN32) || defined(SUN) || defined(HPUX) || defined(AIX) || defined(OSF1) || (defined(LINUX) && defined (I386))
#define HAS_STACK_BACKTRACE_SUPPORT
#endif

#if !defined(HAS_STACK_BACKTRACE_SUPPORT)

externC void eo670_CTraceStack(void) {}
externC void vtracestack(void) {}
externC SAPDB_Int4 eo670_FillCallArray(SAPDB_Int4 level, void **callArray) { return 0; }
externC SAPDB_Bool eo670_GetAddressInfo(void *pcAddr, eo670AddressInfo *addrInfo, void *context) { return false; }
externC void eo670_GetAddressInfoEpilog (void* context) {}
externC SAPDB_Bool eo670_GetAddressInfoProlog (void** context) {}
externC void eo670_FreeAddressInfo(eo670AddressInfo *addrInfo) {}
externC void eo670_CTraceContextStack( void* exceptionContext ) {}
externC void eo670WriteString(const char *str) {}

#else /* HAS_STACK_BACKTRACE_SUPPORT */

#include "gsp00.h"
#include "hsp77.h" /* PTS 1110802 */
#include "geo60.h"
#include "RunTime/System/RTESys_Spinlock.h"

#if defined(KERNEL_LZU)
#  undef WANT_DEMANGLING_SUPPORT
#  define WANT_DEMANGLING_SUPPORT
#  include "heo01.h" /* nocheck */
#endif

#if defined(WANT_DEMANGLING_SUPPORT)
#  if defined(SUN) || defined(AIX)
/* only needed if C++ demangling is needed */
#    include <demangle.h> /* PTS 1110802 */
#  endif /* SUN */
#endif /* WANT_DEMANGLING_SUPPORT */

#include <stdio.h>

#if defined(_WIN32) 

#    define FILL_SKIP_COUNT 1
#  if defined (_WIN64) 
#    if !defined (BIT64)
#      define BIT64
#    endif
#  else 
#    define HAS_FAST_FILL_CALL_ARRAY
#    undef BIT64
#  endif

/*WIN*/
#  include <windows.h>
#  if WINVER > 0x0500
#    include <dbghelp.h>
#  else 
#    include <imagehlp.h>
#  endif
/*WIN*/

#else /* only on UNIX */

/*UNIX*/
#  include <unistd.h>
#  include <stdlib.h>
#  include <errno.h>
#  include <string.h>
#  include <signal.h>
#  include <setjmp.h>
/*UNIX*/

/*OSF1*/
#  if defined(OSF1)
#    define HAS_FAST_FILL_CALL_ARRAY
#    define FILL_SKIP_COUNT 2
#    include <loader.h>
#    include <excpt.h>
#    include <ldfcn.h>
#  endif /* OSF1 */
/*OSF1*/

/*LINUX*/
#  ifdef LINUX
#    define HAS_FAST_FILL_CALL_ARRAY
#    define FILL_SKIP_COUNT 1
#    include <sys/types.h>
#    include <fcntl.h>
#    include <elf.h>
#    include <sys/file.h>
#    include <sys/procfs.h>
/* PTS 1106738 */
/* With RedHat6.2 the mmap() prototype was fixed. This prevented the kernel from */
/* starting. The message given was 'invalid argument'. This Message was given due */
/* to the last argument of mmap() which was defined as 'off_t' and later 'off64_t' */
/* So on RedHat6.2 the last argument was expected to be a 64bit value, therefore */
/* the extra 32bits are collected from the stack. Such an offset is not PAGESIZE */
/* aligned, so it leads to 'invalid argument'... */
#    undef __USE_FILE_OFFSET64
#    define __USE_LARGEFILE64
#    include <sys/mman.h>
#    include <sys/param.h>
#    include <sys/stat.h>
#    include <linux/version.h>
/*
  LINUX_VERSION_CODE 132099 maybe to high, but the definite number is not known to me....

  This code is able to produce a stack backtrace even on these old LINUX machines we are currently
  using (Kernel 2.2.11-sap2_serial_sysrq). We have another test machine with a Kernel 2.4.4.
  From this newer machine i took the LINUX_VERSION_CODE. The reason for this effort is that
  even the gdb dumps a core on the old linux machine while trying to load our kernel....
  I took the inline assembler to get the current frame pointer and analyzed the stack frame
  with a memory dump (optimized and debug code). This way i found the offset for the 'sigcontext'
  and put it into the magic number below... I am not happy with this, but it is an old linux
  kernel anyhow... 6.2.2002 jrg

 */
#    if LINUX_VERSION_CODE < 132099
#      define LINUX_WITHOUT_SIGINFO
/* #    include <sys/ucontext.h> can not be used due to conflict with <procfs.h> ... */
/* "asm/sigcontext" and "sys/ucontext" contain the same structure with different naming conventions... */
#      include <asm/sigcontext.h>
#    else
#      undef LINUX_WITHOUT_SIGINFO
#      define __USE_GNU
#      include <sys/ucontext.h>
#    endif
#  endif /* LINUX */
/*LINUX*/

/*SUN*/
#  ifdef SUN
#    define HAS_FAST_FILL_CALL_ARRAY
#    define FILL_SKIP_COUNT 1
#    include <sys/types.h>
#    include <fcntl.h>
#    include <libelf.h>
#    include <sys/frame.h>
#    include <ucontext.h>
#    ifdef BIT64
#      include <sys/stack.h>
#    endif /* BIT64 */
#    include <sys/file.h>
#    include <procfs.h> /* this is the new procfs interface, not based on ioctl!!!! */
#    include <sys/mman.h>
#    include <sys/param.h>
#    include <sys/stat.h>
#  endif /* SUN */
/*SUN*/

/*AIX*/
#  ifdef AIX
#    define HAS_FAST_FILL_CALL_ARRAY
#    define FILL_SKIP_COUNT 2
#    include <sys/debug.h>
#    include <sys/ldr.h>
#  endif /* AIX */
/*AIX*/

/*HPUX*/
#  ifdef HPUX
#    define HAS_FAST_FILL_CALL_ARRAY
#    define FILL_SKIP_COUNT 1
#    include <dl.h>
#    include <fcntl.h>
#    ifdef BIT64
#      include <elf.h>
#    else
#      include <a.out.h>
#    endif
#  endif /* HPUX */
/*HPUX*/

#endif /* only on UNIX */

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/
#define MESSAGE_TEXT_LIMIT  115
#define MAX_SYMBOL_LENGTH 4096

#if defined(_WIN32)
#  define MSG_ID 19999
#  define NONE_VERBOSE_CALL_STACK    0
#  define VERBOSE_CALL_STACK         1

#  define MAXSYMNAMELEN       2048 // max name length for found symbols
#  define TTBUFLEN            65536 // for a temp buffer
#  define MAX_MODULE_NAME32   255
#  define TH32CS_SNAPMODULE   0x00000008
#else
#  define MSG_ID 11599
#endif /* _WIN32 */

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

#if defined(_WIN32)
#  define GIVEUP_TIME_SLICE Sleep(0)
#else
#  define GIVEUP_TIME_SLICE sleep(0)
#endif

/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/

#if defined (_WIN32)
typedef struct eo670_GetAddressInfoContext_
{
  HANDLE m_handle;
} eo670_GetAddressInfoContext;

#else

#  if defined(SUN)
typedef struct 
{
    void *pcAddr;
    void *frame;
    unsigned long arg[6];
} eo670_CallInfo;
#  elif defined(LINUX)
typedef struct 
{
    void *pcAddr;
    void *frame;
    unsigned long arg[4];
} eo670_CallInfo;
#  else
typedef struct 
{
    void *pcAddr;
} eo670_CallInfo;
#  endif /* !SUN */

#endif /* _WIN32 */

typedef struct eo670_ModuleChain_
{
    struct eo670_ModuleChain_ *next;
    char *file_name;
    void *text_mapaddr;
    void *text_endaddr;
    void *osDependend;
} eo670_ModuleChain;

/*===========================================================================*
 *  STATIC VARIABLES                                                         *
 *===========================================================================*/

/* Flag to notice first call ever */
static volatile int eo670_TraceEverCalled = 0;

/* Spinlock used to interlock stack backtrace requests */
static RTE_Lock eo670_inTraceStack;

/* Anchor for list of loaded modules */
static eo670_ModuleChain *eo670_ModuleAnchor = 0;
static eo670_ModuleChain *eo670_LastModuleEntry = 0;

/* A 'lot of' space used to store loaded module infos (not including symbol tables...) */
static char eo670_ModuleInfoBuffer[16*1024];

/* Common Message buffer */
static char eo670MessageBuffer[4096];

/*===========================================================================*
 *  PROTOTYPES                                                               *
 *===========================================================================*/

#if defined(HPUX) && !defined(BIT64)
/*HPUX32*/extern void U_STACK_TRACE(void);
/*HPUX32*/extern void U_TRACEBACK(int,void*);
#endif /* HPUX && !BIT64 */

#if defined(SUN)
/*SUN*/static void eo670_SunShowCallInfoAddon(eo670_CallInfo *callInfo);
#elif defined(LINUX)
/*LINUX*/static void eo670_LinuxShowCallInfoAddon(eo670_CallInfo *callInfo);
#endif /* SUN */

/*ALL*/static void eo670_OsFreeAddressInfo(eo670AddressInfo *addrInfo);

#ifdef WIN32
/*WIN*/static void       eo670_WinShowStack( HANDLE hThread, CONTEXT * c, int VerbosityLevel );
/*WIN*/static SAPDB_Int4 eo670_WinFillCallArray(SAPDB_Int4 level, void **callArray, void *context);
/*WIN*/static void       eo670_WinGetAddressInfoEpilog (void* pContext);
/*WIN*/static SAPDB_Bool eo670_WinGetAddressInfoProlog (void** pContext);
/*WIN*/static SAPDB_Bool eo670_WinGetAddressInfo(void *pcAddr, eo670AddressInfo *addrInfo, void *context);
/*WIN*/static bool       eo670_FillModuleList( DWORD pid, HANDLE hProcess, int VerbosityLevel );
/*WIN*/static char const * const eo670_GetSymSearchPath();
#else
/* 
   Function: eo670_UnixTraceStack
   Description: This routine controls the several steps of stack back trace creation

   It set up signal handler for SIGSEGV and SIGBUS usings siglongjmp() to protect as much as possible.

   The stack back trace creation consists of

    Creating Emergency Dump (non symbolic without any heap usage except for static data)
    Creating Regioster Dump (optional if context != 0 that is in signal context
    Creating Loaded Module List
    Creating Symbolic Stack Back Trace
 */
/*UNIX*/void eo670_UnixTraceStack(void *context);

/* 
   Function: eo670_UpdateModuleChain
   Description: Fill the module chain with loaded module information

   Internally eo670_AddToLoadedModuleChain() is used to actually fill. This routine is responsible
   for gathering text segment start and end address and the module name.

    NOTE: To be implemented individually for each platform...
 */
/*UNIX*/static void eo670_UpdateModuleChain (void); 

/* 
   Function: eo670_FillAddressInfo
   Description: Fill the address info using given program counter

   The given address info is cleared and filled. The entries will point to static storage afterwards,
   so caller is responsible to copy data if needed.

   NOTE To be implemented individually for each platform...

   Arguments: pcAddr   [in] the program counter used to search symbolic information
              addrInfo [out] the address info structure to be filled
   Return value: true if addrInfo contains valid data, false if symbol resoluition failed completely

 */
/*UNIX*/static void eo670_FillAddressInfo(char *pcAddr, eo670AddressInfo *addrInfo);

/* 
   Function: eo670_UnixGetAddressInfo
   Description: Get the symbolic information using given program counter

   The given address info is cleared and filled. The entries will be filled using eo670_FillAddressInfo 
   and copied afterwards. So it is save to call this function for different addresses without side effects.

   NOTE To be implemented individually for each platform...
 
   Arguments: pcAddr   [in] the program counter used to search symbolic information
              addrInfo [out] the address info structure to be filled
   Return value: true if addrInfo contains valid data, false if symbol resoluition failed completely
 */
/*UNIX*/static SAPDB_Bool eo670_UnixGetAddressInfo(void *pcAddr, eo670AddressInfo *addrInfo);

typedef enum { eo670_NoArrayJustOutput = 0, eo670_PointerArray = 1, eo670_FullInfoArray = 2 } eo670_CallArrayType;
/* 
   Function: eo670_UnixFillCallArray
   Description: Fill the given call array up to limit

   The given call array is filled. The type of call array is either void *[] if addCallInfo is false,
   or eo670_CallInfo[] if addCallInfo is true. In the first case, only the program counter value is
   collected and otherwise additional information gathered.

   NOTE To be implemented individually for each platform...

   addCallInfo eo670_NoArrayJustOutput is just for emergency stack back trace (no memory allocation allowed!)
   addCallInfo eo670_PointerArray is just to collect information for deferred stack backtrace (i.e. memory
    leak detection code uses it to find the procedures, that did the allocation calls....)
   addCallInfo eo670_FullInfoArray is used for symbolic stack backtrace
  
   Arguments: skipCount [in] number of stack entries to skip from current context (only used if pContext != 0
              level     [in] size of given array
              addCallInfo [in] flag to indicate type of array or output only mode
              callArray [in] the call array to fill
              pContext [in] the signal context or 0 if current context to be used
   Return value: actual depth filled in array
 */
/*UNIX*/SAPDB_Int4 eo670_UnixFillCallArray(SAPDB_Int4 skipCount, SAPDB_Int4 level, eo670_CallArrayType addCallInfo, void **callArray, void *context);

/* 
   Function: eo670_RegisterDump
   Description: Create a register dump

   Print out register dump using given signal context.

   NOTE To be implemented individually for each platform...

   Arguments: pContext [in] the signal context or 0 if current context to be used
 */
/*UNIX*/static void eo670_RegisterDump(void *pContext);
#endif /* _WIN32 */

/*
  Function: eo670_AddToLoadedModuleChain
  Description: Add the collected information to the module list

  If text_mapaddr is already found in list, the new entry is ignored.
  Otherwise the new element is allocated and added to list anchored on eo670_ModuleAnchor.
  Module list element is allocated via calloc().

  Arguments: file_name [in] module or module file name if available
             text_mapaddr [in] text segment start address
             text_endaddr [in] text segment end address
             osDependend  [in] context pointer used to save local context
  Return value: 0 dupicate entry
                1 entry added
                -1 entry adding failed due to memory problem
 */
static int eo670_AddToLoadedModuleChain(
    const char *file_name,
    void *text_mapaddr,
    void *text_endaddr,
    void *osDependend );

/*
  Function: eo670_ShowLoadedModules

  Description: Show collected list of loaded modules
  
  Using eo670_ModuleAnchor all modules found are listed.
 */
static void eo670_ShowLoadedModules(void);

#ifdef HAS_FAST_FILL_CALL_ARRAY
SAPDB_Int4 eo670_FastFillCallArray(SAPDB_Int4 skipCount, SAPDB_Int4 level, void **callArray);
#endif

/*===========================================================================*
 *  CRITICAL SECTION                                                         *
 *===========================================================================*/

#if defined (_WIN32)

# define CS_TRACE_BACK 1
# define CS_WRITE_DUMP 2
# define CS_LAST       2

typedef struct _tCritSec { CRITICAL_SECTION CritSec;
                            BOOL             CritSecInitialized;
                            INT              CritSecCount; } tCritSec;

static tCritSec CS[CS_LAST];

#  define  ENTER_CRIT_SEC(_No)                                                    \
           if (!CS[_No].CritSecInitialized) { CS[_No].CritSecInitialized = TRUE; \
             InitializeCriticalSection(&CS[_No].CritSec); }                      \
           EnterCriticalSection(&CS[_No].CritSec); CS[_No].CritSecCount++;

#  define  EXIT_CRIT_SEC(_No)                                                     \
           if (CS[_No].CritSecCount) { CS[_No].CritSecCount--;                   \
             LeaveCriticalSection(&CS[_No].CritSec); }

#endif /* WIN32 */

/*===========================================================================*
 *  GLOBAL FUNCTIONS                                                         *
 *===========================================================================*/

/*!
  Function: eo670_FillCallArray(int level, void **callArray)
  Description: Fill the given array with at most level trace back entries.
  Arguments: level [in] the number of level to fill
             callArray [out] the array to fill
  Return value: The actual number of entries filled
 */
SAPDB_Int4 eo670_FillCallArray(SAPDB_Int4 level, void **callArray)
{
    SAPDB_Int4 filledLevel = 0;

    if ( level <= 0 ) return 0;

    memset(callArray, 0, sizeof(void *) * level );

#ifdef HAS_FAST_FILL_CALL_ARRAY

    filledLevel = eo670_FastFillCallArray(FILL_SKIP_COUNT, level, callArray);

#else

    if ( !eo670_TraceEverCalled )
    {
        eo670_TraceEverCalled = 1;
        RTESys_InitLock(&eo670_inTraceStack);
    }

    if ( RTESys_TestAndLock(&eo670_inTraceStack) )
    {
        eo670WriteString("Waiting while stack back trace busy\n");
        while ( RTESys_TestAndLock(&eo670_inTraceStack) )
        {
            GIVEUP_TIME_SLICE;
        }
        eo670WriteString("Finally got the lock\n");
    }

#ifdef WIN32
    filledLevel = eo670_WinFillCallArray(level, callArray, 0);

#else
    filledLevel = eo670_UnixFillCallArray(FILL_SKIP_COUNT, level, eo670_PointerArray, callArray, 0);
#endif /* WIN32 */

    RTESys_Unlock(&eo670_inTraceStack);
#endif
    return filledLevel;
}

/*---------------------------------------------------------------------------
  Function: eo670_GetAddressInfo
  Description: Get mapping information about a given PC address
  Arguments: pcAddr [in] a hopefully valid program counter
             addrInfo [out] structure to be filled with address information
  Return value: true if mapping succeeded, false if not (addrInfo still valid)
 */
SAPDB_Bool eo670_GetAddressInfo(void *pcAddr, eo670AddressInfo *addrInfo, void *context)
{
    SAPDB_Bool callResult;

    memset(addrInfo, 0, sizeof(eo670AddressInfo));

    if ( !eo670_TraceEverCalled )
    {
        eo670_TraceEverCalled = 1;
        RTESys_InitLock(&eo670_inTraceStack);
    }

    if ( RTESys_TestAndLock(&eo670_inTraceStack) )
    {
        eo670WriteString("Waiting while stack back trace busy\n");
        while ( RTESys_TestAndLock(&eo670_inTraceStack) )
        {
            GIVEUP_TIME_SLICE;
        }
        eo670WriteString("Finally got the lock\n");
    }

#if defined(WIN32)
    callResult = eo670_WinGetAddressInfo(pcAddr, addrInfo, context);
#else
    callResult = eo670_UnixGetAddressInfo(pcAddr, addrInfo);
#endif /* WIN32 */

    RTESys_Unlock(&eo670_inTraceStack);

    return callResult;
}

/*---------------------------------------------------------------------------
  Function: eo670_FreeAddressInfo
  Description: Free the mapping information about a given PC address
  Arguments: addrInfo [in] structure to be freed
 */
void eo670_FreeAddressInfo(eo670AddressInfo *addrInfo)
{
    if ( !eo670_TraceEverCalled )
    {
        eo670_TraceEverCalled = 1;
        RTESys_InitLock(&eo670_inTraceStack);
    }

    if ( RTESys_TestAndLock(&eo670_inTraceStack) )
    {
        eo670WriteString("Waiting while stack back trace busy\n");
        while ( RTESys_TestAndLock(&eo670_inTraceStack) )
        {
            GIVEUP_TIME_SLICE;
        }
        eo670WriteString("Finally got the lock\n");
    }

    eo670_OsFreeAddressInfo(addrInfo);

    RTESys_Unlock(&eo670_inTraceStack);
}

/*!---------------------------------------------------------------------------
    Function: eo670_CTraceContextStack
    Description: Produce a Stackbacktrace on stderr using given context
 */
void eo670_CTraceContextStack(void *context)
{
#if defined(_WIN32)

    static HANDLE  thisThread;
    static int verbosity;

    thisThread = GetCurrentThread();

    if ( 0 == context )
    {
        static CONTEXT c;

        memset( &c, '\0', sizeof(c) );
        c.ContextFlags = CONTEXT_FULL;

        // init CONTEXT record so we know where to start the stackwalk

        if ( ! GetThreadContext( thisThread, &c ) )
        {
            eo670WriteString("Failed to get context\n");
            return;
        }

        // --- write a none verbose call stack
        verbosity = NONE_VERBOSE_CALL_STACK;
        context = &c;
    }
    else
    {
        // --- write a verbose call stack
        verbosity = VERBOSE_CALL_STACK;
    }

    eo670_WinShowStack( thisThread, (CONTEXT*)context, verbosity );

#else
/*UNIX*/  eo670_UnixTraceStack(context);
#endif /* WIN32 */
}

/*!---------------------------------------------------------------------------
    Function: eo670_CTraceStack
    Description: Produce a Stackbacktrace on stderr
 */
void eo670_CTraceStack(void)
{
    if ( !eo670_TraceEverCalled )
    {
        eo670_TraceEverCalled = 1;
        RTESys_InitLock(&eo670_inTraceStack);
    }

    if ( RTESys_TestAndLock(&eo670_inTraceStack) )
    {
        eo670WriteString("Ignoring call since stack back trace busy\n");
        return;
    }

    eo670_CTraceContextStack((void *)0);

    RTESys_Unlock(&eo670_inTraceStack);
}

/*!---------------------------------------------------------------------------
    Function: vtracestack
    Description: Pascal wrapper for eo670_CTraceStack
 */
void vtracestack(void)
{
    if ( !eo670_TraceEverCalled )
    {
        eo670_TraceEverCalled = 1;
        RTESys_InitLock(&eo670_inTraceStack);
    }

    if ( RTESys_TestAndLock(&eo670_inTraceStack) )
    {
        eo670WriteString("Ignoring call since stack back trace busy\n");
        return;
    }

    eo670_CTraceContextStack((void *)0);

    RTESys_Unlock(&eo670_inTraceStack);
}

/*!---------------------------------------------------------------------------
  Function: eo670_GetAddressInfoEpilog
  Description: finishes a sequences of eo670_GetAddressInfo calls and
               release all resources hold by eo670_GetAddressInfoProlog.
  Arguments: 
    context [in] pointer returned by eo670_GetAddressInfoProlog
  Return value: none
 */

void eo670_GetAddressInfoEpilog (void* context)
{
#ifdef WIN32
  eo670_WinGetAddressInfoEpilog(context);
#endif /* WIN32 */
}

/*!---------------------------------------------------------------------------
  Function: eo670_GetAddressInfoProlog
  Description: prepares for following calls of eo670_GetAddressInfo
  Arguments: 
    context [out] pointer to context required for eo670_GetAddressInfo.
                  may be null on several plattforms
  Return value: true if mapping returns valid data, false if not
 */
SAPDB_Bool eo670_GetAddressInfoProlog (void** context)
{
#ifdef WIN32
  return eo670_WinGetAddressInfoProlog(context);
#else
/*UNIX*/  eo670_UpdateModuleChain();
#endif /* WIN32 */

  return true;
}

/*===========================================================================*
 *  LOCAL/PRIVATE/PROTECTED FUNCTIONS (CODE)                                 *
 *===========================================================================*/

static void eo670_write_string_with_displacement( const char *str, const char *displacement )
{
#if defined(KERNEL_LZU)
	/* PTS 1112092 */
/* Message Text limit is used for several reasons.
   dbmgetf reads out knldiag on a line by line base and copies atmost
   255 character per line. Each knldiag line begins with 44 characters of 
   header information. Two lines on an 80 character display 160 minus 44 character
   header result in 116 character. One character less to prevent automatic wrap
   around leads to 115.
   This value is only needed due to the line oriented output of knldiag.
 */

  static const char *pc;
	static int msgLength;
	static int Limit;

  pc        = strchr(str,'\n');
	msgLength = ( pc != 0 ? (int)(pc-str) : (int)strlen(str) );

  pc = str;
  MSGD (( MSG_ID, 1, "BTRACE  ", "%.*s", msgLength > MESSAGE_TEXT_LIMIT ? MESSAGE_TEXT_LIMIT : msgLength, pc ));

	msgLength -= MESSAGE_TEXT_LIMIT;
	pc        += MESSAGE_TEXT_LIMIT;

  Limit = MESSAGE_TEXT_LIMIT - (int)strlen(displacement);

	while ( msgLength > 0 )
	{
    MSGD (( MSG_ID, 1, "BTRACE  ", "%s%.*s", displacement, msgLength > Limit ? Limit : msgLength, pc ));

		msgLength -= Limit;
		pc        += Limit;
    }

#else /* not in kernel variant */

#    if defined(_WIN32)
         fprintf(stderr, "%s", str);
#    else
         write(2, str, strlen(str));
#    endif

#endif /* KERNEL_LZU */
}

/*---------------------------------------------------------------------------*/

void eo670WriteString(const char *str)
{
  eo670_write_string_with_displacement(str, "" );
}

/*---------------------------------------------------------------------------
  Function: eo670_AddToLoadedModuleChain
  Description: Add a module entry into sorted list of modules
 */
static int eo670_AddToLoadedModuleChain(
    const char *file_name,
    void *text_mapaddr,
    void *text_endaddr,
    void *osDependend )
{
    static eo670_ModuleChain * modulePointer;
    static eo670_ModuleChain ** pSearchPointer;

    pSearchPointer = &eo670_ModuleAnchor;
    do
    {
        if ( 0 != *pSearchPointer )
        {
            if ( (*pSearchPointer)->text_mapaddr > text_mapaddr )
            {
                break;
            }

            if ( (*pSearchPointer)->text_mapaddr == text_mapaddr )
            {
                return 0;
            }
            pSearchPointer = &((*pSearchPointer)->next);
        }
    } while ( 0 != (*pSearchPointer) );

    if ( eo670_ModuleAnchor )
    {
        /* We need space for the ModuleChain structure and the copy of the file_name + alignment */
        /* We know position of 'last' pointer which was used */

#define ALIGNED_STRLEN_EO670(string_) ( (strlen(string_)+sizeof(void*)) & ~(sizeof(void*)-1) )

        /* First we use the last pointer to calculate the next last pointer, which is used */
        /* later to store the new entry */
        eo670_LastModuleEntry = (eo670_ModuleChain *) ( ((char *)(eo670_LastModuleEntry+1))
                                          + ALIGNED_STRLEN_EO670(eo670_LastModuleEntry->file_name) );
    }
    else
    {
        /* if we never stored anything we begin at start of buffer */
        eo670_LastModuleEntry = (eo670_ModuleChain *)&eo670_ModuleInfoBuffer[0];
    }

    if ( (int)( sizeof(eo670_ModuleChain) + ALIGNED_STRLEN_EO670(file_name) )
       > (int)(&eo670_ModuleInfoBuffer[sizeof(eo670_ModuleInfoBuffer)]  - (((char *)eo670_LastModuleEntry) ) ) )
    {
        eo670WriteString("Out of memory for ModuleChain entry\n");
        return -1;
    }

    eo670_LastModuleEntry->next         = *pSearchPointer;
    eo670_LastModuleEntry->file_name    = (char *)(eo670_LastModuleEntry+1);
    strcpy(eo670_LastModuleEntry->file_name, file_name);
    eo670_LastModuleEntry->text_mapaddr = text_mapaddr;
    eo670_LastModuleEntry->text_endaddr = text_endaddr;
    eo670_LastModuleEntry->osDependend  = osDependend;

    *pSearchPointer = eo670_LastModuleEntry;

    return 1;
}

/*---------------------------------------------------------------------------
  Function: eo670_ShowLoadedModules
  Description: Show sorted list of modules
 */
static void eo670_ShowLoadedModules(void)
{
    static eo670_ModuleChain * modulePointer = 0;

	sp77sprintf( eo670MessageBuffer,
                 sizeof(eo670MessageBuffer),
#if defined(BIT64)
                 "%-29s |%-18s|%s\n", 
#else
                 "%-29s |%-12s|%s\n", 
#endif /* BIT64 */
                 "module file name",
                 ".text start",
                 ".text end" );
    eo670WriteString(eo670MessageBuffer);
    for ( modulePointer = eo670_ModuleAnchor;
          NULL != modulePointer;
          modulePointer = modulePointer->next )
    {
	    sp77sprintf( eo670MessageBuffer,
                     sizeof(eo670MessageBuffer),
#if defined(BIT64)
                     "%-29s |0x%016lx|0x%016lx\n", 
#else
                     "%-29s | 0x%08lx | 0x%08lx\n", 
#endif /* BIT64 */
                     modulePointer->file_name,
                     (SAPDB_Long)modulePointer->text_mapaddr,
                     (SAPDB_Long)modulePointer->text_endaddr );
        eo670WriteString(eo670MessageBuffer);
    }
    eo670WriteString("\n");
}

/*---------------------------------------------------------------------------
  Function: eo670_ShowAddressInfo
  Description: Show address information

  If no symbol is found, the offsetToSymbol is negativ.
  If a offsetToSymbol is given, it is show using the 
  demangledSymbol if available.
  If additonal line number information is found, it is
  show too, otherwise only the module name is given.
 */
static void eo670_ShowAddressInfo(
  int entryIndex,
  void *pcAddr,
  eo670AddressInfo *addrInfo)
{
    if ( addrInfo->offsetToSymbol >= 0 )
    {
        sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
#if defined(BIT64)
                     "%4d: 0x%016lx %s +0x%04x\n",
#else
                     "%4d: 0x%08lx %s +0x%04x\n",
#endif /* BIT64 */
                     entryIndex,
                     (SAPDB_Long)pcAddr,
    ( *addrInfo->demangledSymbol ? addrInfo->demangledSymbol : addrInfo->symbol ),
                     addrInfo->offsetToSymbol );
        eo670WriteString(eo670MessageBuffer);
        if ( addrInfo->lineNumber >= 0 )
        {
            sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
                         "       %s - %s:%d\n",
                         addrInfo->moduleName,
                         addrInfo->sourceFileName,
                         addrInfo->lineNumber );
        }
        else if ( 0 != *addrInfo->sourceFileName )
        {
            sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
                         "       %s - %s\n",
                         addrInfo->moduleName,
                         addrInfo->sourceFileName );
        }
        else
        {
            sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
                         "       %s\n",
                         addrInfo->moduleName );
        }
    }
    else
    {
        sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
                     "%4d: 0x%016lx %s\n",
                     entryIndex,
                     (SAPDB_Long)pcAddr,
                     addrInfo->symbol );
    }
    eo670WriteString(eo670MessageBuffer);
}

/* dump a section of memory as ints */

static void eo670_MemoryDump(void *startAddr, SAPDB_Int4 cnt)
{
    SAPDB_Int4  firstBytes;
    static SAPDB_Int4 iOffset;

    sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
#ifdef BIT64
                 "Memory dump from startAddr 0x%016lx endAddr 0x%016lx\n",
#else
                 "Memory dump from startAddr 0x%08lx endAddr 0x%08lx\n",
#endif /* BIT64 */
                 ((char *)startAddr)-((char *)0),
                 ((char *)startAddr)+cnt-((char *)0) );
    eo670WriteString(eo670MessageBuffer);

    firstBytes = (((char *)startAddr)-((char *)0)) & (sizeof(SAPDB_UInt4)-1);
    if ( firstBytes > 0 )
    {
        firstBytes = sizeof(SAPDB_Int4)-firstBytes;
        cnt -= firstBytes;
        sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
#ifdef BIT64
                     "0x%016lx",
#else
                     "0x%08lx",
#endif /* BIT64 */
                     startAddr );
        while ( firstBytes > 0 )
        {
            sp77sprintf( eo670MessageBuffer + strlen(eo670MessageBuffer), 
                         sizeof(eo670MessageBuffer) - strlen(eo670MessageBuffer),
                         " 0x%02x",
                          *(unsigned char *)startAddr );
            startAddr = ((char *)startAddr)+1;
            --firstBytes;
        }
        sp77sprintf( eo670MessageBuffer + strlen(eo670MessageBuffer), 
                     sizeof(eo670MessageBuffer) - strlen(eo670MessageBuffer),
                     "\n" );
        eo670WriteString(eo670MessageBuffer);
    }
    cnt = (cnt + sizeof(SAPDB_UInt4)-1)/sizeof(SAPDB_UInt4);

    for ( iOffset = 0; iOffset < cnt; iOffset += 4 )
    {
        sp77sprintf( eo670MessageBuffer,
                     sizeof(eo670MessageBuffer),
                     "0x%08lx 0x%08lx 0x%08lx 0x%08lx 0x%08lx\n",
                     ((char *)(((SAPDB_Int4*)startAddr)+iOffset))-((char *)0),
                    *(((SAPDB_UInt4*)startAddr)+iOffset), *(((SAPDB_UInt4*)startAddr)+iOffset+1),
                    *(((SAPDB_UInt4*)startAddr)+iOffset+2), *(((SAPDB_UInt4*)startAddr)+iOffset+3) );
        eo670WriteString(eo670MessageBuffer);
    }
}


#if !defined(WIN32)
/*===========================================================================*
 *  UNIX                                                                     *
 *===========================================================================*/

/*--------------------------
  Signal handling code
 */
/*UNIX*/
/*UNIX*/static struct sigaction oldSigbus;
/*UNIX*/static struct sigaction oldSigsegv;
/*UNIX*/
#  if defined(LINUX) || defined(SUN)
/*SUN+LINUX*/static sigjmp_buf eo670_emergency;
#  else
/*AIX+OSF1*/static jmp_buf eo670_emergency;
#  endif /* LINUX || SUN */
/*UNIX*/
/*UNIX*//*	This is the stack trace back SIGBUS/SIGSEGV signal handler. */
/*UNIX*/static void eo670_StackTraceSegvFunc(int Signal)
/*UNIX*/{
/*UNIX*/    siglongjmp(eo670_emergency, Signal != 0 ? Signal : 1);
/*UNIX*/}
/*UNIX*/
/*--------------------------
  End of signal handling code
 */
/*
  Call Stack array
 */
/* Number of entries to skip... */
#  if defined(OSF1)
#    define SKIP_COUNT 4
#  elif defined(HPUX) || defined(AIX)
#    define SKIP_COUNT 3
#  else
#    define SKIP_COUNT 2
#  endif

/*UNIX*/void eo670_UnixTraceStack(void *pContext)
/*UNIX*/{
/*UNIX*/static void *           old_sigsegv;
/*UNIX*/static void *           old_sigbus;
/*UNIX*/static sigset_t         old_set;
/*UNIX*/static sigset_t         mask_set;
/*UNIX*/static int              rc;
/*UNIX*/static eo670AddressInfo localAddrInfo;
/*UNIX*/static char *           pcAddr;
/*UNIX*/static int              i;
/*UNIX*/static const char *     routine = "<unknown>";
/*UNIX*/static long             stackDepth;
#  define MAX_STACK     (256)
/*UNIX*/static eo670_CallInfo   callInfoArray[MAX_STACK];
/*UNIX*/static int              nextAction;
/*UNIX*/
/*UNIX*//* 	First insure that we do not cause a segment violation when we look for  */
/*UNIX*//* 	the TraceBack table in the code segment.  So, we add a signal handle    */
/*UNIX*//* 	for SIGSEGV and SIGBUS.                                                 */
/*UNIX*/    rc = sigsetjmp( eo670_emergency, 1);
/*UNIX*/    if ( rc == 0 )
/*UNIX*/    {
/*UNIX*/        static struct sigaction sa;
/*UNIX*/
/*UNIX*/        sa.sa_handler = eo670_StackTraceSegvFunc;
/*UNIX*/        sa.sa_flags   = 0;
/*UNIX*/        sigemptyset( &sa.sa_mask );
/*UNIX*/
/*UNIX*/        sigaction(SIGSEGV, &sa, &oldSigsegv);
/*UNIX*/        sigaction(SIGBUS, &sa, &oldSigbus);
/*UNIX*/
/*UNIX*/        sigfillset(&mask_set);
/*UNIX*/        sigdelset(&mask_set, SIGSEGV);
/*UNIX*/        sigdelset(&mask_set, SIGBUS);
/*UNIX*/        sigprocmask(SIG_SETMASK, &mask_set, &old_set);
/*UNIX*/
/*UNIX*/        nextAction = 1;
/*UNIX*/    }
/*UNIX*/    else
/*UNIX*/    {
/*UNIX*/        sp77sprintf(eo670MessageBuffer, sizeof(eo670MessageBuffer), 
/*UNIX*/           "*** Error in stack backtrace: Signal %d in %s\n", rc, routine);
/*UNIX*/        eo670WriteString ( eo670MessageBuffer );
/*UNIX*/        ++nextAction;
/*UNIX*/    }
/*UNIX*/    
/*UNIX*/    while( 0 != nextAction )
/*UNIX*/    {
/*UNIX*/        switch(nextAction)
/*UNIX*/        {
/*UNIX*/        case 1:
/*UNIX*/            eo670WriteString("----> Emergency Stack Back Trace <----\n");
/*UNIX*/            routine = "eo670_UnixFillCallArray(OutputOnly)";
/*UNIX*/            (void)eo670_UnixFillCallArray (SKIP_COUNT, MAX_STACK, eo670_NoArrayJustOutput, (void**)&callInfoArray[0], pContext);
/*UNIX*/            ++nextAction;
/*UNIX*/            break;
/*UNIX*/    
/*UNIX*/        case 2:
/*UNIX*/            if ( pContext )
/*UNIX*/            {
/*UNIX*/                eo670WriteString("----> Register Dump <----\n");
/*UNIX*/                routine = "eo670_RegisterDump";
/*UNIX*/                eo670_RegisterDump(pContext);
/*UNIX*/            }
/*UNIX*/            ++nextAction;
/*UNIX*/            break;
/*UNIX*/    
/*UNIX*/        case 3:
/*UNIX*/            eo670WriteString("----> Module List <----\n");
/*UNIX*/            routine = "eo670_UpdateModuleChain";
/*UNIX*/            eo670_UpdateModuleChain();
/*UNIX*/
/*UNIX*/            routine = "eo670_ShowLoadedModules";
/*UNIX*/            eo670_ShowLoadedModules();
/*UNIX*/            ++nextAction;
/*UNIX*/            break;
/*UNIX*/
/*UNIX*/        case 4:
/*UNIX*/           eo670WriteString("----> Symbolic Stack Back Trace <----\n");
/*UNIX*/
#  if defined(HPUX) && !defined(BIT64)
/*HPUX32*/         fflush(stderr);
/*HPUX32*/         if ( pContext )
/*HPUX32*/         {
/*HPUX32*/             U_TRACEBACK(((struct sigcontext *)pContext)->sc_ctxt.sx.sl_si.si_signo,pContext);
/*HPUX32*/         }
/*HPUX32*/         else
/*HPUX32*/         {
/*HPUX32*/             U_STACK_TRACE();
/*HPUX32*/         }
#  else
/*UNIX*/           routine = "eo670_FillCallArray";
/*UNIX*/           stackDepth = eo670_UnixFillCallArray (SKIP_COUNT, MAX_STACK, eo670_FullInfoArray, (void**)&callInfoArray[0], pContext);
/*UNIX*/           for ( i = 0; i < stackDepth; i++ )
/*UNIX*/           {
/*UNIX*/    	       pcAddr = (char *)(callInfoArray[i].pcAddr);
/*UNIX*/
/*UNIX*/               routine = "eo670_FillAddressInfo";
/*UNIX*/               eo670_FillAddressInfo(pcAddr, &localAddrInfo);
/*UNIX*/
/*UNIX*/               routine = "eo670_ShowAddressInfo";
/*UNIX*/               eo670_ShowAddressInfo(i, pcAddr, &localAddrInfo);
/*UNIX*/
#    if defined(SUN)
/*SUN*/                eo670_SunShowCallInfoAddon(&callInfoArray[i]);
#    elif defined(LINUX)
/*LINUX*/              eo670_LinuxShowCallInfoAddon(&callInfoArray[i]);
#    endif
/*UNIX*/               /* Main stack ends here... */
/*UNIX*/               if (strcmp (localAddrInfo.symbol, "main") == 0)
/*UNIX*/                   break;
/*UNIX*/
/*UNIX*/               /* UKT Stack ends here... */
/*UNIX*/               if (strcmp (localAddrInfo.symbol, "en88_CallA91Main") == 0)
/*UNIX*/                   break;
#    ifdef AIX
/*AIX*/                /* Thread stack ends here... */
/*AIX*/                if (strcmp (localAddrInfo.symbol, "_uc_trampoline") == 0)
/*AIX*/                    break;
/*AIX*/
#    endif /* AIX */
/*UNIX*/            } /* end for loop over stack-array */
#  endif
/*UNIX*/            ++nextAction;
/*UNIX*/            break;
/*UNIX*/
/*UNIX*/        default:
/*UNIX*/            nextAction = 0;
/*UNIX*/            break;
/*UNIX*/        }
/*UNIX*/    }
/*UNIX*/
/*UNIX*/    sigaction(SIGSEGV, &oldSigsegv, 0);
/*UNIX*/    sigaction(SIGBUS, &oldSigbus, 0);
/*UNIX*/    sigprocmask(SIG_SETMASK, &old_set, NULL);
/*UNIX*/    return;
/*UNIX*/}

/*------------------------------------------------------------------------------------------*/

#  if defined(SUN) || defined(LINUX)
/*LINUX+SUN*/
#    define NOHEAP_MALLOC noheap_malloc
#    define NOHEAP_CALLOC noheap_calloc
#    define NOHEAP_FREE noheap_free
#    define USRTEXT  0x2000
#    ifdef BIT64
/*LINUX+SUN*/typedef Elf64_Sym  Elf_Symbol;
/*LINUX+SUN*/typedef Elf64_Ehdr Elf_ElfHeader;
/*LINUX+SUN*/typedef Elf64_Shdr Elf_SectionHeader;
/*LINUX+SUN*/typedef Elf64_Word Elf_Word;
/*LINUX+SUN*/typedef Elf64_Addr Elf_Addr;
#      define ELF_ST_TYPE(_i) ELF64_ST_TYPE(_i)
#    else
/*LINUX+SUN*/typedef Elf32_Sym  Elf_Symbol;
/*LINUX+SUN*/typedef Elf32_Ehdr Elf_ElfHeader;
/*LINUX+SUN*/typedef Elf32_Shdr Elf_SectionHeader;
/*LINUX+SUN*/typedef Elf32_Word Elf_Word;
/*LINUX+SUN*/typedef Elf32_Addr Elf_Addr;
#      define ELF_ST_TYPE(_i) ELF32_ST_TYPE(_i)
#    endif
/*LINUX+SUN*/
/*LINUX+SUN*/typedef struct t_Symbol {
/*LINUX+SUN*/    char *        st_name;
/*LINUX+SUN*/    Elf_Addr      st_value;
/*LINUX+SUN*/    unsigned char st_info;
/*LINUX+SUN*/    unsigned char padding[sizeof(void*)-1];
/*LINUX+SUN*/} Symbol_t;
/*LINUX+SUN*/
/*LINUX+SUN*/typedef struct t_map {
/*LINUX+SUN*/    size_t          nsyms;    /* Number of symbols                */
/*LINUX+SUN*/    char    	   *strtab;   /* String Table                     */
/*LINUX+SUN*/    Symbol_t       *symtab;   /* Symbol Table                     */
/*LINUX+SUN*/    Elf_SectionHeader     *secthdr;
/*LINUX+SUN*/    Elf_SectionHeader     *sym_sh;
/*LINUX+SUN*/    Elf_SectionHeader     *str_sh;
/*LINUX+SUN*/    Elf_SectionHeader     *shent;
/*LINUX+SUN*/    Elf_Word      symsect;
/*LINUX+SUN*/    Elf_Word      strsect;
/*LINUX+SUN*/    char         *filename; /* ELF file name                        */
/*LINUX+SUN*/    int		    fd_proc;
/*LINUX+SUN*/    caddr_t         vaddr;
/*LINUX+SUN*/    struct t_map   *next;    /* Ptr to other shared libraries       */
/*LINUX+SUN*/} map_t;
/*LINUX+SUN*/
/*LINUX+SUN*/
/*LINUX+SUN*//* name to return if symbol name is not found                       */
/*LINUX+SUN*/static char no_symbol_name[9] = "????????";
/*LINUX+SUN*/
#    define USRTEXT  0x2000
#    define PROC_MMAP_FILE_OFFSET 49
#    define PROC_MMAP_ACCESS_OFFSET 18
#    define PROC_MMAP_VADDR_OFFSET 0
#    define PROC_MMAP_VENDADDR_OFFSET 9
/*LINUX+SUN*/
/*LINUX+SUN*//*** Local functions ***/
/*LINUX+SUN*/
/*LINUX+SUN*/static Symbol_t  *tb_get_func(unsigned long addr, map_t *);
/*LINUX+SUN*/static char      *tb_get_filename(map_t *);
/*LINUX+SUN*/
/*LINUX+SUN*/static Symbol_t  *process_syms(Elf_Symbol *, size_t, 
/*LINUX+SUN*/                               size_t *, caddr_t, char *, char **);
/*LINUX+SUN*/static map_t     *get_symtab(int, caddr_t);
/*LINUX+SUN*/static int        compare(Symbol_t *, Symbol_t *);
/*LINUX+SUN*/static void       freemap(map_t *);
/*LINUX+SUN*/
/*LINUX+SUN*//*---------------------------------------------------------------------------*/
/*LINUX+SUN*/
/*LINUX+SUN*/static void *
/*LINUX+SUN*/noheap_malloc(size_t size)
/*LINUX+SUN*/{
/*LINUX+SUN*/  static int fd_dev_zero;
/*LINUX+SUN*/  static caddr_t ra;
/*LINUX+SUN*/
/*LINUX+SUN*/    if ((fd_dev_zero = open("/dev/zero", O_RDWR, 0)) == -1) {
/*LINUX+SUN*/        perror("opening /dev/zero");
/*LINUX+SUN*/        return (NULL);
/*LINUX+SUN*/    }
/*LINUX+SUN*/
/*LINUX+SUN*/    if (size == 0) size = sizeof (int);
/*LINUX+SUN*/
/*LINUX+SUN*/    if ((ra = (caddr_t)mmap((caddr_t) 0, size + sizeof(caddr_t), 
/*LINUX+SUN*/                   (PROT_READ | PROT_WRITE), MAP_PRIVATE,
/*LINUX+SUN*/                   fd_dev_zero, 0)) == (caddr_t)-1) {
/*LINUX+SUN*/        perror("mmap'ing /dev/zero");
/*LINUX+SUN*/        ra = NULL;
/*LINUX+SUN*/    } else {
/*LINUX+SUN*/        *((size_t *)ra) = size;
/*LINUX+SUN*/        ra += sizeof(size_t);
/*LINUX+SUN*/    }
/*LINUX+SUN*/
/*LINUX+SUN*/    close(fd_dev_zero);
/*LINUX+SUN*/    return((void *)ra);
/*LINUX+SUN*/}
/*LINUX+SUN*/
/*LINUX+SUN*/static void *
/*LINUX+SUN*/noheap_calloc(size_t nelem, size_t elsize)
/*LINUX+SUN*/{
/*LINUX+SUN*/    /* While mmap of /dev/zero is being used, the block               */
/*LINUX+SUN*/    /* should come back already zero filled.                          */
/*LINUX+SUN*/    return(noheap_malloc((nelem * elsize)));
/*LINUX+SUN*/}
/*LINUX+SUN*/
/*LINUX+SUN*//*---------------------------------------------------------------------------*/
/*LINUX+SUN*/
/*LINUX+SUN*/static void
/*LINUX+SUN*/noheap_free(void *ptr)
/*LINUX+SUN*/{
/*LINUX+SUN*/  static size_t len;
/*LINUX+SUN*/  static caddr_t real_ptr;
/*LINUX+SUN*/
/*LINUX+SUN*/    if (ptr == NULL) return;
/*LINUX+SUN*/
/*LINUX+SUN*/    real_ptr = (caddr_t)(((size_t *)ptr) - 1);
/*LINUX+SUN*/    len = *((size_t *)real_ptr) + sizeof(caddr_t);
/*LINUX+SUN*/    if (munmap(real_ptr, len))
/*LINUX+SUN*/        perror("munmap");
/*LINUX+SUN*/}
/*LINUX+SUN*/
/*LINUX+SUN*//*---------------------------------------------------------------------------*/
/*LINUX+SUN*//*---------------------------------------------------------------------------*/
/*LINUX+SUN*//*                                                                           */
/*LINUX+SUN*//* tb_get_func() takes an address and dereferences it to give the            */
/*LINUX+SUN*//* function name which starts at the closest lower address. If no            */
/*LINUX+SUN*//* such function exists, it passes an empty string.                          */
/*LINUX+SUN*//* Relies on the qsorted symtab list described above.                        */
/*LINUX+SUN*//* Basically a binary search algorithm.                                      */
/*LINUX+SUN*/
/*LINUX+SUN*/static Symbol_t    *
/*LINUX+SUN*/tb_get_func(unsigned long addr, map_t *map)
/*LINUX+SUN*/{
/*LINUX+SUN*/  static  int         left;
/*LINUX+SUN*/  static  int         right;
/*LINUX+SUN*/  static  int         middle, sclass;
/*LINUX+SUN*/  static  Symbol_t   *symtab;
/*LINUX+SUN*/  static  int i;
/*LINUX+SUN*/
/*LINUX+SUN*/    left = 0;
/*LINUX+SUN*/    right = map->nsyms - 1;
/*LINUX+SUN*/    symtab = map->symtab;
/*LINUX+SUN*/    i = 0;
/*LINUX+SUN*/    if ((symtab->st_value > addr) ||
/*LINUX+SUN*/        ((symtab + right)->st_value < addr)) {
/*LINUX+SUN*/        return (NULL);  /* should be invalid name                     */
/*LINUX+SUN*/    }
/*LINUX+SUN*/    while (left <= right) {  /* Binary Search (sort of)               */
/*LINUX+SUN*/        if ((right-left) <= 1)
/*LINUX+SUN*/            return(symtab + left);
/*LINUX+SUN*/        middle = (left + right) / 2;
/*LINUX+SUN*/        if ((symtab + middle)->st_value <= addr) {
/*LINUX+SUN*/            left = middle;
/*LINUX+SUN*/        } else {
/*LINUX+SUN*/            right = middle;
/*LINUX+SUN*/        }
/*LINUX+SUN*/    }
/*LINUX+SUN*/
/*LINUX+SUN*/    /* should never get here                                          */
/*LINUX+SUN*/    return (NULL);
/*LINUX+SUN*/}
/*LINUX+SUN*/
/*LINUX+SUN*//*---------------------------------------------------------------------------*/
/*LINUX+SUN*/typedef int (*CompareFunc) (const void *, const void *);
/*LINUX+SUN*/
/*LINUX+SUN*/static map_t *
/*LINUX+SUN*/get_symtab(int fd, caddr_t voffset)
/*LINUX+SUN*/{
/*LINUX+SUN*/  static map_t         *map;
/*LINUX+SUN*/  static  Elf_ElfHeader    elfhdr;    /* Elf header                            */
/*LINUX+SUN*/  static  Elf_Symbol     *sp;
/*LINUX+SUN*/  static  int        i;
/*LINUX+SUN*/  static  size_t        syms;
/*LINUX+SUN*/  static  Elf_Symbol     *symtab;    /* Unsorted file image symbol table     */
/*LINUX+SUN*/
/*LINUX+SUN*/    /* Allocate a map structure                                       */
/*LINUX+SUN*/
/*LINUX+SUN*/    map = (map_t *)NOHEAP_CALLOC(sizeof(map_t), 1);
/*LINUX+SUN*/    if(map == 0) {
/*LINUX+SUN*/        eo670WriteString("Out of memory\n");
/*LINUX+SUN*/        return(0);
/*LINUX+SUN*/    }
/*LINUX+SUN*/
/*LINUX+SUN*/    /* Get the ELF header                                             */
/*LINUX+SUN*/
/*LINUX+SUN*/    if (read (fd, &elfhdr, sizeof(elfhdr)) != sizeof(elfhdr)){
/*LINUX+SUN*/        eo670WriteString("Unable to read ELF header \n");
/*LINUX+SUN*/        freemap(map);
/*LINUX+SUN*/        return(0);
/*LINUX+SUN*/    }
/*LINUX+SUN*/
/*LINUX+SUN*/    /* Get the section header table                                   */
/*LINUX+SUN*/
/*LINUX+SUN*/    if (elfhdr.e_shnum == 0) {
/*LINUX+SUN*/        eo670WriteString("No section header entries.\n");
/*LINUX+SUN*/        freemap(map);
/*LINUX+SUN*/        return(0);
/*LINUX+SUN*/    }
/*LINUX+SUN*/
/*LINUX+SUN*/    if (lseek(fd, (long) elfhdr.e_shoff, L_SET) != elfhdr.e_shoff) {
/*LINUX+SUN*/        eo670WriteString("Unable to seek to section header.\n");
/*LINUX+SUN*/        freemap(map);
/*LINUX+SUN*/        return(0);
/*LINUX+SUN*/    }
/*LINUX+SUN*/    if ((map->secthdr = (Elf_SectionHeader *)
/*LINUX+SUN*/         NOHEAP_MALLOC(elfhdr.e_shentsize*elfhdr.e_shnum)) == NULL) {
/*LINUX+SUN*/        eo670WriteString("Unable to allocate section header.\n");
/*LINUX+SUN*/        freemap(map);
/*LINUX+SUN*/        return(0);
/*LINUX+SUN*/    }
/*LINUX+SUN*/
/*LINUX+SUN*/    if (read(fd, map->secthdr, elfhdr.e_shentsize * elfhdr.e_shnum)
/*LINUX+SUN*/        !=  (elfhdr.e_shentsize * elfhdr.e_shnum)) {
/*LINUX+SUN*/        eo670WriteString("Unable to read ELF header section.\n");
/*LINUX+SUN*/        freemap(map);
/*LINUX+SUN*/        return(0);
/*LINUX+SUN*/    }
/*LINUX+SUN*/
/*LINUX+SUN*/    map->shent  = map->secthdr;    /* set to first entry              */
/*LINUX+SUN*/    for (i = 0; i < (int) elfhdr.e_shnum; i++, map->shent++ ){
/*LINUX+SUN*/        /* go thru each section header                                */
/*LINUX+SUN*/        if (map->shent->sh_type ==  SHT_SYMTAB) {
/*LINUX+SUN*/            map->symsect = i;
/*LINUX+SUN*/            map->strsect = map->shent->sh_link;
/*LINUX+SUN*/        }
/*LINUX+SUN*/    }
/*LINUX+SUN*/
/*LINUX+SUN*/    if (map->symsect == 0) {
/*LINUX+SUN*/        freemap(map);
/*LINUX+SUN*/        return(0);
/*LINUX+SUN*/    }
/*LINUX+SUN*/
/*LINUX+SUN*/    map->sym_sh = &map->secthdr[map->symsect];
/*LINUX+SUN*/    syms = map->sym_sh->sh_size/map->sym_sh->sh_entsize;
/*LINUX+SUN*/    if(map->sym_sh->sh_link == elfhdr.e_shnum ||
/*LINUX+SUN*/       map->secthdr[map->sym_sh->sh_link].sh_size == 0) {
/*LINUX+SUN*/        eo670WriteString("Warning - empty string table; no symbols.\n");
/*LINUX+SUN*/        freemap(map);
/*LINUX+SUN*/        return(0);
/*LINUX+SUN*/    }
/*LINUX+SUN*/    map->str_sh = &map->secthdr[map->sym_sh->sh_link];
/*LINUX+SUN*/    if ((map->strtab = (char *)
/*LINUX+SUN*/         NOHEAP_MALLOC(map->str_sh->sh_size +
/*LINUX+SUN*/                       sizeof(map->str_sh->sh_size)+8)) == NULL) {
/*LINUX+SUN*/        eo670WriteString("Warning: Couldnt allocate space for strtab\n");
/*LINUX+SUN*/        freemap(map);
/*LINUX+SUN*/        return(0);
/*LINUX+SUN*/    }
/*LINUX+SUN*/    *(Elf_Word *)map->strtab = map->str_sh->sh_size + 
/*LINUX+SUN*/        sizeof(map->str_sh->sh_size);
/*LINUX+SUN*/    lseek(fd, (long)map->str_sh->sh_offset, L_SET);
/*LINUX+SUN*/    if(read(fd, map->strtab + sizeof(map->str_sh->sh_size), 
/*LINUX+SUN*/	    map->str_sh->sh_size) != map->str_sh->sh_size) {
/*LINUX+SUN*/        eo670WriteString("Warning - error reading string table\n");
/*LINUX+SUN*/        freemap(map);
/*LINUX+SUN*/        return(0);
/*LINUX+SUN*/    }
/*LINUX+SUN*/    lseek(fd, (long)map->sym_sh->sh_offset, L_SET);
/*LINUX+SUN*/    if ((symtab = (Elf_Symbol *) 
/*LINUX+SUN*/         NOHEAP_MALLOC(syms*sizeof(Elf_Symbol))) == NULL) {
/*LINUX+SUN*/        eo670WriteString("Warning: Couldnt allocate space for symtab\n");
/*LINUX+SUN*/        freemap(map);
/*LINUX+SUN*/        return(0);
/*LINUX+SUN*/    }
/*LINUX+SUN*/    if (read(fd, symtab, syms*sizeof(Elf_Symbol)) != 
/*LINUX+SUN*/        syms*sizeof(Elf_Symbol)){
/*LINUX+SUN*/        eo670WriteString("Warning - error reading symbl table\n");
/*LINUX+SUN*/        freemap(map);
/*LINUX+SUN*/        return(0);
/*LINUX+SUN*/    }
/*LINUX+SUN*/
/*LINUX+SUN*/    /* process_syms() returns an array of 'interesting symbols'       */
/*LINUX+SUN*/    /* i.e. those with a non-NULL st_name field.                      */
/*LINUX+SUN*/
/*LINUX+SUN*/    if (elfhdr.e_type == ET_EXEC)
/*LINUX+SUN*/        voffset = 0;
/*LINUX+SUN*/    map->symtab = process_syms(symtab, syms, &map->nsyms, voffset, 
/*LINUX+SUN*/                                map->strtab + sizeof(map->str_sh->sh_size), &map->filename);
/*LINUX+SUN*/    NOHEAP_FREE(symtab);
/*LINUX+SUN*/
/*LINUX+SUN*/    if(map->symtab == NULL) {
/*LINUX+SUN*/        freemap(map);
/*LINUX+SUN*/        return(0);
/*LINUX+SUN*/    }
/*LINUX+SUN*/
/*LINUX+SUN*/    qsort(map->symtab, map->nsyms, sizeof(Symbol_t), (CompareFunc)compare);
/*LINUX+SUN*/
/*LINUX+SUN*/    return(map);
/*LINUX+SUN*/}
/*LINUX+SUN*/
/*LINUX+SUN*//*---------------------------------------------------------------------------*/
/*LINUX+SUN*/
/*LINUX+SUN*/static void
/*LINUX+SUN*/freemap(map_t *map)
/*LINUX+SUN*/{
/*LINUX+SUN*/    if(!map)
/*LINUX+SUN*/        return;
/*LINUX+SUN*/    if(map->secthdr) NOHEAP_FREE(map->secthdr);
/*LINUX+SUN*/    if(map->strtab)  NOHEAP_FREE(map->strtab);
/*LINUX+SUN*/    if(map->symtab)  NOHEAP_FREE(map->symtab);
/*LINUX+SUN*/    NOHEAP_FREE(map);
/*LINUX+SUN*/}
/*LINUX+SUN*/
/*LINUX+SUN*//*---------------------------------------------------------------------------*/
/*LINUX+SUN*//*                                                                           */
/*LINUX+SUN*//* Translate symbol table data particularly for sorting.                     */
/*LINUX+SUN*//* Input is the symbol table data structure, number of symbols,              */
/*LINUX+SUN*//* opened ELF file, and the string table link offset.                        */
/*LINUX+SUN*/
/*LINUX+SUN*/static Symbol_t *
/*LINUX+SUN*/process_syms(Elf_Symbol *data, size_t num, size_t *nsyms,
/*LINUX+SUN*/             caddr_t offset, char* strtab, char **pFilename)
/*LINUX+SUN*/{
/*LINUX+SUN*/  static Symbol_t  *s;
/*LINUX+SUN*/  static Symbol_t  *buf;
/*LINUX+SUN*/  static int i;
/*LINUX+SUN*/  static int useful_sym;   /* Not all symbols read are useful          */
/*LINUX+SUN*/
/*LINUX+SUN*/    useful_sym = 0;
/*LINUX+SUN*/    /* */
/*LINUX+SUN*/    /* This loop exists only to detect number of function names */
/*LINUX+SUN*/    /* so that space is not overallocated for it                */
/*LINUX+SUN*/    /* */
/*LINUX+SUN*/    for (i = 0; i < num; i ++)
/*LINUX+SUN*/    {
/*LINUX+SUN*/        if ( data[i].st_name && 
/*LINUX+SUN*/             (ELF_ST_TYPE(data[i].st_info) == STT_FUNC))
/*LINUX+SUN*/        {
/*LINUX+SUN*/            useful_sym++;
/*LINUX+SUN*/        }
/*LINUX+SUN*/    }
/*LINUX+SUN*/
/*LINUX+SUN*/    s = (Symbol_t *)NOHEAP_MALLOC(useful_sym*sizeof(Symbol_t));
/*LINUX+SUN*/    if(0 == s)
/*LINUX+SUN*/    {
/*LINUX+SUN*/        eo670WriteString("Out of memory\n");
/*LINUX+SUN*/        return NULL;
/*LINUX+SUN*/    }
/*LINUX+SUN*/
/*LINUX+SUN*/    /* */
/*LINUX+SUN*/    /* Skip the first entry. It contains a special STN_UNDEF */
/*LINUX+SUN*/    /* symbol                                                */
/*LINUX+SUN*/    /* */
/*LINUX+SUN*/
/*LINUX+SUN*/    *pFilename = (char *)"";
/*LINUX+SUN*/    for(i=1, buf = s; i<= num; i++, data++)
/*LINUX+SUN*/    {
/*LINUX+SUN*/        if ( data->st_name != (Elf_Word)0 ) 
/*LINUX+SUN*/        {
/*LINUX+SUN*/            if ( 0 == **pFilename ) /* First symbol is module name */
/*LINUX+SUN*/            {
/*LINUX+SUN*/                 *pFilename = strtab + data->st_name;
/*LINUX+SUN*/            }
/*LINUX+SUN*/
/*LINUX+SUN*/            if ( ELF_ST_TYPE(data->st_info) == STT_FUNC )
/*LINUX+SUN*/            {
/*LINUX+SUN*/                buf->st_name  = strtab + data->st_name;
/*LINUX+SUN*/                buf->st_value = data->st_value + (unsigned long) offset;
/*LINUX+SUN*/                buf->st_info  = data->st_info;
/*LINUX+SUN*/                buf++ ;
/*LINUX+SUN*/            }
/*LINUX+SUN*/        }
/*LINUX+SUN*/    }
/*LINUX+SUN*/
/*LINUX+SUN*/    *nsyms = useful_sym;
/*LINUX+SUN*/    return(s);
/*LINUX+SUN*/}
/*LINUX+SUN*/
/*LINUX+SUN*/
/*LINUX+SUN*//*---------------------------------------------------------------------------*/
/*LINUX+SUN*//*                                                                           */
/*LINUX+SUN*//* Compare either by name or by value for sorting.                           */
/*LINUX+SUN*//* This is the comparison function called by qsort to                        */
/*LINUX+SUN*//* sort the symbols either by name or value when requested.                  */
/*LINUX+SUN*/
/*LINUX+SUN*/static int 
/*LINUX+SUN*/compare(Symbol_t *a, Symbol_t *b)
/*LINUX+SUN*/{
/*LINUX+SUN*/    if(a->st_value > b->st_value)
/*LINUX+SUN*/        return 1;
/*LINUX+SUN*/    else if (a->st_value < b->st_value)
/*LINUX+SUN*/        return -1;
/*LINUX+SUN*/    else 
/*LINUX+SUN*/        return (((unsigned char)ELF_ST_TYPE(a->st_info) <
/*LINUX+SUN*/                 (unsigned char)ELF_ST_TYPE(b->st_info)) 
/*LINUX+SUN*/                ? 1 : 0);
/*LINUX+SUN*/
/*LINUX+SUN*/}
/*LINUX+SUN*/
#    ifdef LINUX
/*LINUX*//*---------------------------------------------------------------------------*/
/*LINUX*/
/*LINUX*/static void PrepareProcMap(char *procMap, char *endProcMap)
/*LINUX*/{
/*LINUX*/  static char *p;
/*LINUX*/
/*LINUX*/    p = procMap;
/*LINUX*/    while ( p < endProcMap ) {
/*LINUX*/        if ( *p == '\n' ) *p = 0;
/*LINUX*/        ++p;
/*LINUX*/    }
/*LINUX*/}
/*LINUX*/
/*LINUX*//*---------------------------------------------------------------------------*/
/*LINUX*/
/*LINUX*/static char *NextProcMap(char *procMap, size_t *pBufferedChars, int proc_fd)
/*LINUX*/{
/*LINUX*/  static char *p;
/*LINUX*/  static size_t bytesStored;
/*LINUX*/  static int    bytesRead;
/*LINUX*/
/*LINUX*/    bytesStored = *pBufferedChars;
/*LINUX*/
/*LINUX*/    p = procMap + strlen(procMap);
/*LINUX*/    while ( *p == 0 && p < (procMap + bytesStored) ) {
/*LINUX*/        ++p;
/*LINUX*/    }
/*LINUX*/    bytesStored -= (p-procMap);
/*LINUX*/
/*LINUX*/    memcpy(procMap, p, bytesStored);
/*LINUX*/    bytesRead = read(proc_fd, procMap + bytesStored, 1024 - bytesStored );
/*LINUX*/    if ( bytesRead < 0 )
/*LINUX*/    {
/*LINUX*/        eo670WriteString("cant read from /proc file\n");
/*LINUX*/        return (NULL);
/*LINUX*/    }
/*LINUX*/    PrepareProcMap(procMap + bytesStored, procMap + bytesStored + bytesRead);
/*LINUX*/
/*LINUX*/    *pBufferedChars = bytesStored + bytesRead;
/*LINUX*/
/*LINUX*/    return ( *pBufferedChars > 0 ? procMap : NULL );
/*LINUX*/}
/*LINUX*/
/*LINUX*//*---------------------------------------------------------------------------*/
/*LINUX*/
/*LINUX*/static void eo670_UpdateModuleChain()
/*LINUX*/{
/*LINUX*/  static int       fd_proc;	/* /proc fd                           */
/*LINUX*/  static map_t    *map;      /* Main executable map                */
/*LINUX*/  static char      procMapBuffer[1024];
/*LINUX*/  static char     *procMap;
/*LINUX*/  static size_t    bufferedChars; /* Number of currently buffered character */
/*LINUX*/  static caddr_t   vaddr;
/*LINUX*/  static caddr_t   vendaddr;
/*LINUX*/
/*LINUX*/    sp77sprintf(eo670MessageBuffer, sizeof(eo670MessageBuffer), "/proc/%d/maps", getpid());
/*LINUX*/
/*LINUX*/    if ((fd_proc = open(eo670MessageBuffer, O_RDONLY)) == -1)
/*LINUX*/    {
/*LINUX*/        sp77sprintf(eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*LINUX*/                  "Cant open /proc/%d/maps : %d\n", getpid(), errno);
/*LINUX*/        eo670WriteString(eo670MessageBuffer);
/*LINUX*/        return;
/*LINUX*/    }
/*LINUX*/
/*LINUX*/    if ( (bufferedChars = read(fd_proc, procMapBuffer, sizeof(procMapBuffer))) <= 0 )
/*LINUX*/    {
/*LINUX*/        sp77sprintf(eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*LINUX*/                  "Cant read from /proc/%d/maps : %d\n", getpid(), errno);
/*LINUX*/        eo670WriteString(eo670MessageBuffer);
/*LINUX*/        return;
/*LINUX*/    }
/*LINUX*/
/*LINUX*/    procMap = &procMapBuffer[0];
/*LINUX*/    /* Split lines into zero terminated strings */
/*LINUX*/    PrepareProcMap(procMap, procMap + bufferedChars);
/*LINUX*/
/*LINUX*/    map = 0;
/*LINUX*/
/*LINUX*/    do
/*LINUX*/    {
/*LINUX*/        static int  fd_obj;
/*LINUX*/
/*LINUX*/        if ( strlen(procMap) < PROC_MMAP_FILE_OFFSET )
/*LINUX*/            continue;
/*LINUX*/
/*LINUX*/        if ( strncmp(procMap+PROC_MMAP_ACCESS_OFFSET, "r-xp", 4) != 0 )
/*LINUX*/            continue;
/*LINUX*/
/*LINUX*/        vaddr = (caddr_t)
/*LINUX*/                strtoul(procMap+PROC_MMAP_VADDR_OFFSET, NULL, 16);
/*LINUX*/
/*LINUX*/        vendaddr = (caddr_t)
/*LINUX*/                strtoul(procMap+PROC_MMAP_VENDADDR_OFFSET, NULL, 16);
/*LINUX*/
/*LINUX*/        if ( (fd_obj = open(procMap+PROC_MMAP_FILE_OFFSET, O_RDONLY)) == -1)
/*LINUX*/            continue;
/*LINUX*/        
/*LINUX*/        map = get_symtab(fd_obj, vaddr);
/*LINUX*/        close(fd_obj);
/*LINUX*/
/*LINUX*/        switch( eo670_AddToLoadedModuleChain( procMap + PROC_MMAP_FILE_OFFSET,
/*LINUX*/                                              vaddr,
/*LINUX*/                                              vendaddr,
/*LINUX*/                                              map ) )
/*LINUX*/        {
/*LINUX*/        case 1: break;
/*LINUX*/        case 0: freemap(map); break;
/*LINUX*/        default:goto cleanup;
/*LINUX*/        }
/*LINUX*/    } while ( 0 != NextProcMap(procMap, &bufferedChars, fd_proc) );
/*LINUX*/
/*LINUX*/    (void) close(fd_proc);
/*LINUX*/    return;
/*LINUX*/ cleanup:
/*LINUX*/    (void) close(fd_proc);
/*LINUX*/    freemap(map);
/*LINUX*/    return;
/*LINUX*/}
/*LINUX*/
/*LINUX*//*---------------------------------------------------------------------------*/
#      ifdef LINUX_WITHOUT_SIGINFO
/*OLDLINUX*/
/*OLDLINUX*/        /* libc5 doesn't have <sigcontext.h>: go directly with the kernel   */
/*OLDLINUX*/        /* one.  Check LINUX_VERSION_CODE to see which we should reference. */
/*OLDLINUX*/
/*OLDLINUX*/struct LinuxGccI386ExceptionFrame {
/*OLDLINUX*/    struct LinuxGccI386Frame *fr_savfp;
/*OLDLINUX*/    long   dummy_pc;
/*OLDLINUX*/    long   dummy[40]; /* magic number... but it works.... */
/*OLDLINUX*/    struct sigcontext fr_context;
/*OLDLINUX*/};
/*OLDLINUX*/
/*OLDLINUX*/struct LinuxGccI386SignalHandlerFrame
/*OLDLINUX*/{
/*OLDLINUX*/    struct LinuxGccI386ExceptionFrame *fr_savfp;
/*OLDLINUX*/    long  dummy_pc;
/*OLDLINUX*/    int   signal;
/*OLDLINUX*/    long  dummy_arg2;
/*OLDLINUX*/    long  dummy_arg3;
/*OLDLINUX*/    long  dummy_arg4;
/*OLDLINUX*/};
/*OLDLINUX*/
/*OLDLINUX*/static void eo670_RegisterDump(void *pContext)
/*OLDLINUX*/{
/*OLDLINUX*/    static struct sigcontext *context;
/*OLDLINUX*/    static int iReg;
/*OLDLINUX*/
/*OLDLINUX*/    context = (struct sigcontext *)pContext;
/*OLDLINUX*/
/*OLDLINUX*/    eo670WriteString( "-------------------------------\n" );
/*OLDLINUX*/
/*OLDLINUX*/    sp77sprintf( eo670MessageBuffer,
/*OLDLINUX*/                 sizeof(eo670MessageBuffer),
/*OLDLINUX*/                 "ds   es   0x%08x 0x%08x\n",
/*OLDLINUX*/                 *(long*)&context->ds,
/*OLDLINUX*/                 *(long*)&context->es );
/*OLDLINUX*/    eo670WriteString(eo670MessageBuffer);
/*OLDLINUX*/
/*OLDLINUX*/    sp77sprintf( eo670MessageBuffer,
/*OLDLINUX*/                 sizeof(eo670MessageBuffer),
/*OLDLINUX*/                 "fs   gs   0x%08x 0x%08x\n",
/*OLDLINUX*/                 *(long*)&context->fs,
/*OLDLINUX*/                 *(long*)&context->gs );
/*OLDLINUX*/    eo670WriteString(eo670MessageBuffer);
/*OLDLINUX*/
/*OLDLINUX*/    sp77sprintf( eo670MessageBuffer,
/*OLDLINUX*/                 sizeof(eo670MessageBuffer),
/*OLDLINUX*/                 "trap err  0x%08x 0x%08x\n",
/*OLDLINUX*/                 context->trapno,
/*OLDLINUX*/                 context->err );
/*OLDLINUX*/    eo670WriteString(eo670MessageBuffer);
/*OLDLINUX*/
/*OLDLINUX*/    sp77sprintf( eo670MessageBuffer,
/*OLDLINUX*/                 sizeof(eo670MessageBuffer),
/*OLDLINUX*/                 "efl  ss   0x%08x 0x%08x\n",
/*OLDLINUX*/                 context->eflags,
/*OLDLINUX*/                 *(long*)&context->ss );
/*OLDLINUX*/    eo670WriteString(eo670MessageBuffer);
/*OLDLINUX*/
/*OLDLINUX*/    sp77sprintf( eo670MessageBuffer,
/*OLDLINUX*/                 sizeof(eo670MessageBuffer),
/*OLDLINUX*/                 "edi  esi  0x%08x 0x%08x\n",
/*OLDLINUX*/                 context->edi,
/*OLDLINUX*/                 context->esi );
/*OLDLINUX*/    eo670WriteString(eo670MessageBuffer);
/*OLDLINUX*/
/*OLDLINUX*/    sp77sprintf( eo670MessageBuffer,
/*OLDLINUX*/                 sizeof(eo670MessageBuffer),
/*OLDLINUX*/                 "eax  ebx  0x%08x 0x%08x\n",
/*OLDLINUX*/                 context->eax,
/*OLDLINUX*/                 context->ebx );
/*OLDLINUX*/    eo670WriteString(eo670MessageBuffer);
/*OLDLINUX*/
/*OLDLINUX*/    sp77sprintf( eo670MessageBuffer,
/*OLDLINUX*/                 sizeof(eo670MessageBuffer),
/*OLDLINUX*/                 "ecx  edx  0x%08x 0x%08x\n",
/*OLDLINUX*/                 context->ecx,
/*OLDLINUX*/                 context->edx );
/*OLDLINUX*/    eo670WriteString(eo670MessageBuffer);
/*OLDLINUX*/
/*OLDLINUX*/    sp77sprintf( eo670MessageBuffer,
/*OLDLINUX*/                 sizeof(eo670MessageBuffer),
/*OLDLINUX*/                 "esp  uesp 0x%08x 0x%08x\n",
/*OLDLINUX*/                 context->esp,
/*OLDLINUX*/                 context->esp_at_signal );
/*OLDLINUX*/    eo670WriteString(eo670MessageBuffer);
/*OLDLINUX*/
/*OLDLINUX*/    sp77sprintf( eo670MessageBuffer,
/*OLDLINUX*/                 sizeof(eo670MessageBuffer),
/*OLDLINUX*/                 "ebp  eip  0x%08x 0x%08x\n",
/*OLDLINUX*/                 context->ebp,
/*OLDLINUX*/                 context->eip );
/*OLDLINUX*/    eo670WriteString(eo670MessageBuffer);
/*OLDLINUX*/
/*OLDLINUX*/    sp77sprintf( eo670MessageBuffer,
/*OLDLINUX*/                 sizeof(eo670MessageBuffer),
/*OLDLINUX*/                 "cs   cr2  0x%08x 0x%08x\n",
/*OLDLINUX*/                 *(long*)&context->cs,
/*OLDLINUX*/                 context->cr2);
/*OLDLINUX*/    eo670WriteString(eo670MessageBuffer);
/*OLDLINUX*/ 
/*OLDLINUX*/    eo670WriteString( "-------------------------------\n" );
/*OLDLINUX*/
/*OLDLINUX*/    if ( context->fpstate && 0 == (((long)context->fpstate)&3) )
/*OLDLINUX*/    {
/*OLDLINUX*/        sp77sprintf( eo670MessageBuffer,
/*OLDLINUX*/                 sizeof(eo670MessageBuffer),
/*OLDLINUX*/                 "cw - sw - tag   0x%08x 0x%08x 0x%08x\n",
/*OLDLINUX*/                 context->fpstate->cw,
/*OLDLINUX*/                 context->fpstate->sw,
/*OLDLINUX*/                 context->fpstate->tag);
/*OLDLINUX*/        eo670WriteString(eo670MessageBuffer);
/*OLDLINUX*/
/*OLDLINUX*/        sp77sprintf( eo670MessageBuffer,
/*OLDLINUX*/                 sizeof(eo670MessageBuffer),
/*OLDLINUX*/                 "ipoff - cssel   0x%08x 0x%08x\n",
/*OLDLINUX*/                 context->fpstate->ipoff,
/*OLDLINUX*/                 context->fpstate->cssel);
/*OLDLINUX*/        eo670WriteString(eo670MessageBuffer);
/*OLDLINUX*/
/*OLDLINUX*/        sp77sprintf( eo670MessageBuffer,
/*OLDLINUX*/                 sizeof(eo670MessageBuffer),
/*OLDLINUX*/                 "dataoff-datasel 0x%08x 0x%08x\n",
/*OLDLINUX*/                 context->fpstate->dataoff,
/*OLDLINUX*/                 context->fpstate->datasel);
/*OLDLINUX*/        eo670WriteString(eo670MessageBuffer);
/*OLDLINUX*/
/*OLDLINUX*/        for ( iReg = 0; iReg < sizeof(context->fpstate->_st)/sizeof(struct _fpreg); iReg++ )
/*OLDLINUX*/        {
/*OLDLINUX*/            sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*OLDLINUX*/                     "f%02d      [0x%04x 0x%04x 0x%04x 0x%04x - 0x%04x]\n",
/*OLDLINUX*/                      iReg,
/*OLDLINUX*/                      context->fpstate->_st[iReg].significand[0],
/*OLDLINUX*/                      context->fpstate->_st[iReg].significand[1],
/*OLDLINUX*/                      context->fpstate->_st[iReg].significand[2],
/*OLDLINUX*/                      context->fpstate->_st[iReg].significand[3],
/*OLDLINUX*/                      context->fpstate->_st[iReg].exponent );
/*OLDLINUX*/            eo670WriteString( eo670MessageBuffer );
/*OLDLINUX*/        }
/*OLDLINUX*/    }
/*OLDLINUX*/    else
/*OLDLINUX*/    {
/*OLDLINUX*/        eo670WriteString( "Floating point register unused\n" );
/*OLDLINUX*/    }
/*OLDLINUX*/    eo670WriteString( "-------------------------------\n" );
/*OLDLINUX*/}
#      else /* LINUX_WITHOUT_SIGINFO */
/*LINUX*//*---------------------------------------------------------------------------*/
/*LINUX*/
/*LINUX*/static void eo670_RegisterDump(void *pContext)
/*LINUX*/{
/*LINUX*/    static ucontext_t *context;
/*LINUX*/    static int iReg;
/*LINUX*/
/*LINUX*/    context = (ucontext_t *)pContext;
/*LINUX*/
/*LINUX*/    eo670WriteString( "-------------------------------\n" );
/*LINUX*/
/*LINUX*/    sp77sprintf( eo670MessageBuffer,
/*LINUX*/                 sizeof(eo670MessageBuffer),
/*LINUX*/                 "ds   es   0x%08x 0x%08x\n",
/*LINUX*/                 context->uc_mcontext.gregs[REG_DS],
/*LINUX*/                 context->uc_mcontext.gregs[REG_ES] );
/*LINUX*/    eo670WriteString(eo670MessageBuffer);
/*LINUX*/
/*LINUX*/    sp77sprintf( eo670MessageBuffer,
/*LINUX*/                 sizeof(eo670MessageBuffer),
/*LINUX*/                 "fs   gs   0x%08x 0x%08x\n",
/*LINUX*/                 context->uc_mcontext.gregs[REG_FS],
/*LINUX*/                 context->uc_mcontext.gregs[REG_GS] );
/*LINUX*/    eo670WriteString(eo670MessageBuffer);
/*LINUX*/
/*LINUX*/    sp77sprintf( eo670MessageBuffer,
/*LINUX*/                 sizeof(eo670MessageBuffer),
/*LINUX*/                 "trap err  0x%08x 0x%08x\n",
/*LINUX*/                 context->uc_mcontext.gregs[REG_TRAPNO],
/*LINUX*/                 context->uc_mcontext.gregs[REG_ERR] );
/*LINUX*/    eo670WriteString(eo670MessageBuffer);
/*LINUX*/
/*LINUX*/    sp77sprintf( eo670MessageBuffer,
/*LINUX*/                 sizeof(eo670MessageBuffer),
/*LINUX*/                 "efl  ss   0x%08x 0x%08x\n",
/*LINUX*/                 context->uc_mcontext.gregs[REG_EFL],
/*LINUX*/                 context->uc_mcontext.gregs[REG_SS] );
/*LINUX*/    eo670WriteString(eo670MessageBuffer);
/*LINUX*/
/*LINUX*/    sp77sprintf( eo670MessageBuffer,
/*LINUX*/                 sizeof(eo670MessageBuffer),
/*LINUX*/                 "edi  esi  0x%08x 0x%08x\n",
/*LINUX*/                 context->uc_mcontext.gregs[REG_EDI],
/*LINUX*/                 context->uc_mcontext.gregs[REG_ESI] );
/*LINUX*/    eo670WriteString(eo670MessageBuffer);
/*LINUX*/
/*LINUX*/    sp77sprintf( eo670MessageBuffer,
/*LINUX*/                 sizeof(eo670MessageBuffer),
/*LINUX*/                 "eax  ebx  0x%08x 0x%08x\n",
/*LINUX*/                 context->uc_mcontext.gregs[REG_EAX],
/*LINUX*/                 context->uc_mcontext.gregs[REG_EBX] );
/*LINUX*/    eo670WriteString(eo670MessageBuffer);
/*LINUX*/
/*LINUX*/    sp77sprintf( eo670MessageBuffer,
/*LINUX*/                 sizeof(eo670MessageBuffer),
/*LINUX*/                 "ecx  edx  0x%08x 0x%08x\n",
/*LINUX*/                 context->uc_mcontext.gregs[REG_ECX],
/*LINUX*/                 context->uc_mcontext.gregs[REG_EDX] );
/*LINUX*/    eo670WriteString(eo670MessageBuffer);
/*LINUX*/
/*LINUX*/    sp77sprintf( eo670MessageBuffer,
/*LINUX*/                 sizeof(eo670MessageBuffer),
/*LINUX*/                 "esp  uesp 0x%08x 0x%08x\n",
/*LINUX*/                 context->uc_mcontext.gregs[REG_ESP],
/*LINUX*/                 context->uc_mcontext.gregs[REG_UESP] );
/*LINUX*/    eo670WriteString(eo670MessageBuffer);
/*LINUX*/
/*LINUX*/    sp77sprintf( eo670MessageBuffer,
/*LINUX*/                 sizeof(eo670MessageBuffer),
/*LINUX*/                 "ebp  eip  0x%08x 0x%08x\n",
/*LINUX*/                 context->uc_mcontext.gregs[REG_EBP],
/*LINUX*/                 context->uc_mcontext.gregs[REG_EIP] );
/*LINUX*/    eo670WriteString(eo670MessageBuffer);
/*LINUX*/
/*LINUX*/    sp77sprintf( eo670MessageBuffer,
/*LINUX*/                 sizeof(eo670MessageBuffer),
/*LINUX*/                 "cs   cr2  0x%08x 0x%08x\n",
/*LINUX*/                 context->uc_mcontext.gregs[REG_CS],
/*LINUX*/                 context->uc_mcontext.cr2);
/*LINUX*/    eo670WriteString(eo670MessageBuffer);
/*LINUX*/
/*LINUX*/    eo670WriteString( "-------------------------------\n" );
/*LINUX*/
/*LINUX*/    if ( context->uc_mcontext.fpregs && 0 == (((long)context->uc_mcontext.fpregs)&3) )
/*LINUX*/    {
/*LINUX*/        sp77sprintf( eo670MessageBuffer,
/*LINUX*/                 sizeof(eo670MessageBuffer),
/*LINUX*/                 "cw - sw - tag   0x%08x 0x%08x 0x%08x\n",
/*LINUX*/                 context->uc_mcontext.fpregs->cw,
/*LINUX*/                 context->uc_mcontext.fpregs->sw,
/*LINUX*/                 context->uc_mcontext.fpregs->tag);
/*LINUX*/        eo670WriteString(eo670MessageBuffer);
/*LINUX*/
/*LINUX*/        sp77sprintf( eo670MessageBuffer,
/*LINUX*/                 sizeof(eo670MessageBuffer),
/*LINUX*/                 "ipoff - cssel   0x%08x 0x%08x\n",
/*LINUX*/                 context->uc_mcontext.fpregs->ipoff,
/*LINUX*/                 context->uc_mcontext.fpregs->cssel);
/*LINUX*/        eo670WriteString(eo670MessageBuffer);
/*LINUX*/
/*LINUX*/        sp77sprintf( eo670MessageBuffer,
/*LINUX*/                 sizeof(eo670MessageBuffer),
/*LINUX*/                 "dataoff-datasel 0x%08x 0x%08x\n",
/*LINUX*/                 context->uc_mcontext.fpregs->dataoff,
/*LINUX*/                 context->uc_mcontext.fpregs->datasel);
/*LINUX*/        eo670WriteString(eo670MessageBuffer);
/*LINUX*/
/*LINUX*/        for ( iReg = 0; iReg < sizeof(context->uc_mcontext.fpregs->_st)/sizeof(struct _libc_fpreg); iReg++ )
/*LINUX*/        {
/*LINUX*/            sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*LINUX*/                     "f%02d      [0x%04x 0x%04x 0x%04x 0x%04x - 0x%04x]\n",
/*LINUX*/                      iReg,
/*LINUX*/                      context->uc_mcontext.fpregs->_st[iReg].significand[0],
/*LINUX*/                      context->uc_mcontext.fpregs->_st[iReg].significand[1],
/*LINUX*/                      context->uc_mcontext.fpregs->_st[iReg].significand[2],
/*LINUX*/                      context->uc_mcontext.fpregs->_st[iReg].significand[3],
/*LINUX*/                      context->uc_mcontext.fpregs->_st[iReg].exponent );
/*LINUX*/            eo670WriteString( eo670MessageBuffer );
/*LINUX*/        }
/*LINUX*/    }
/*LINUX*/    else
/*LINUX*/    {
/*LINUX*/        eo670WriteString( "Floating point register unused\n" );
/*LINUX*/    }
/*LINUX*/    eo670WriteString( "-------------------------------\n" );
/*LINUX*/}
#      endif /* LINUX_WITHOUT_SIGINFO */
/*LINUX*/
/*LINUX*//*---------------------------------------------------------------------------*/
/*LINUX*/
/*LINUX*/struct LinuxGccI386Frame {
/*LINUX*/    struct LinuxGccI386Frame *fr_savfp;
/*LINUX*/    unsigned long fr_savpc;
/*LINUX*/    unsigned long fr_arg[4];
/*LINUX*/};
/*LINUX*/
/*LINUX*/SAPDB_Int4 eo670_FastFillCallArray(SAPDB_Int4 skipCount, SAPDB_Int4 level, void **callArray)
/*LINUX*/{
/*LINUX*/    struct LinuxGccI386Frame  *fp;
/*LINUX*/    unsigned long pc;
/*LINUX*/    SAPDB_Int4 depth;
/*LINUX*/
/*LINUX*/    depth = 0;
/*LINUX*/
/*LINUX*/    pc = (unsigned long)eo670_FastFillCallArray;
/*LINUX*/    asm( "movl %%ebp,%0" : "=r" (fp) : /* no input operand */ );
/*LINUX*/
/*LINUX*/    while ( depth < level
/*LINUX*/         && fp 
/*LINUX*/         && ((((long)fp)&3) == 0) 
/*LINUX*/         && (fp != fp->fr_savfp) )
/*LINUX*/    {
/*LINUX*/         if ( skipCount > 0 )
/*LINUX*/         {
/*LINUX*/             --skipCount;
/*LINUX*/         }
/*LINUX*/         else
/*LINUX*/         {
/*LINUX*/            callArray[depth] = ((char *)0) + fp->fr_savpc;
/*LINUX*/            ++depth;
/*LINUX*/         }
/*LINUX*/         fp = fp->fr_savfp;
/*LINUX*/    }
/*LINUX*/    return depth;
/*LINUX*/} /* eo670_FastFillCallArray */
/*LINUX*/
/*LINUX*/SAPDB_Int4 eo670_UnixFillCallArray(SAPDB_Int4 skipCount, SAPDB_Int4 level, eo670_CallArrayType addCallInfo, void **callArray, void *pContext)
/*LINUX*/{
/*LINUX*/    static struct LinuxGccI386Frame  *fp;
/*LINUX*/    static unsigned long pc;
/*LINUX*/    static SAPDB_Int4 depth;
/*LINUX*/
/*LINUX*/    depth = 0;
/*LINUX*/
/*LINUX*/    if ( 0 == pContext )
/*LINUX*/    {
/*LINUX*/        pc = (unsigned long)eo670_UnixFillCallArray;
/*LINUX*/        asm( "movl %%ebp,%0" : "=r" (fp) : /* no input operand */ );
/*LINUX*/    }
/*LINUX*/    else
/*LINUX*/    {
#      ifdef LINUX_WITHOUT_SIGINFO
/*OLDLINUX*/     static struct sigcontext *context;
/*OLDLINUX*/
/*OLDLINUX*/     context = &(((struct LinuxGccI386SignalHandlerFrame *)pContext)->fr_savfp->fr_context);
/*OLDLINUX*/     fp = (struct LinuxGccI386Frame *)(((char *)0)+context->ebp);
/*OLDLINUX*/     pc = context->eip;
#      else /* LINUX_WITHOUT_SIGINFO */
/*LINUX*/        static ucontext_t *context;
/*LINUX*/        context = (ucontext_t *)pContext;
/*LINUX*/        fp = (struct LinuxGccI386Frame *)(((char *)0)+context->uc_mcontext.gregs[REG_EBP]);
/*LINUX*/        pc = context->uc_mcontext.gregs[REG_EIP];
#      endif /* LINUX_WITHOUT_SIGINFO */
/*LINUX*/        skipCount = 0;
/*LINUX*/        switch ( addCallInfo )
/*LINUX*/        {
/*LINUX*/        case eo670_FullInfoArray:
/*LINUX*/            ((eo670_CallInfo *)callArray)[depth].pcAddr = (char *)0 + pc;
/*LINUX*/            ((eo670_CallInfo *)callArray)[depth].frame  = fp;
/*LINUX*/            if ( fp && ((((long)fp)&3) == 0) )
/*LINUX*/            {
/*LINUX*/                ((eo670_CallInfo *)callArray)[depth].arg[0] = fp->fr_arg[0];
/*LINUX*/                ((eo670_CallInfo *)callArray)[depth].arg[1] = fp->fr_arg[1];
/*LINUX*/                ((eo670_CallInfo *)callArray)[depth].arg[2] = fp->fr_arg[2];
/*LINUX*/                ((eo670_CallInfo *)callArray)[depth].arg[3] = fp->fr_arg[3];
/*LINUX*/            }
/*LINUX*/            else
/*LINUX*/            {
/*LINUX*/                ((eo670_CallInfo *)callArray)[depth].arg[0] = 0xdead;
/*LINUX*/                ((eo670_CallInfo *)callArray)[depth].arg[1] = 0xc0de;
/*LINUX*/                ((eo670_CallInfo *)callArray)[depth].arg[2] = 0xdeed;
/*LINUX*/                ((eo670_CallInfo *)callArray)[depth].arg[3] = 0xc0de;
/*LINUX*/            }
/*LINUX*/            break;
/*LINUX*/        case eo670_PointerArray:
/*LINUX*/            callArray[depth] = ((char *)0) + pc;
/*LINUX*/            break;
/*LINUX*/        default: /* eo670_NoArrayJustOutput */
/*LINUX*/            if ( fp && ((((long)fp)&3) == 0) )
/*LINUX*/            {
/*LINUX*/                sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*LINUX*/                         "(%d):0x%lx [0x%lx](0x%lx,0x%lx,0x%lx,0x%lx)\n",
/*LINUX*/                         depth, pc, (long)fp, fp->fr_arg[0], fp->fr_arg[1], fp->fr_arg[2], fp->fr_arg[3] );
/*LINUX*/            }
/*LINUX*/            else
/*LINUX*/            {
/*LINUX*/                sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*LINUX*/                         "(%d):0x%lx [0x%lx](?,?,?,?)",
/*LINUX*/                         depth, pc, (long)fp );
/*LINUX*/            }
/*LINUX*/            eo670WriteString(eo670MessageBuffer);
/*LINUX*/            break;
/*LINUX*/        }
/*LINUX*/        ++depth;
/*LINUX*/    }
/*LINUX*/
/*LINUX*/    while ( depth < level
/*LINUX*/         && fp 
/*LINUX*/         && ((((long)fp)&3) == 0) 
/*LINUX*/         && (fp != fp->fr_savfp)
/*LINUX*/         && ((unsigned long)fp->fr_savfp >= USRTEXT) )
/*LINUX*/    {
/*LINUX*/         if ( skipCount > 0 )
/*LINUX*/         {
/*LINUX*/             --skipCount;
/*LINUX*/         }
/*LINUX*/         else
/*LINUX*/         {
/*LINUX*/            switch ( addCallInfo )
/*LINUX*/            {
/*LINUX*/            case eo670_FullInfoArray:
/*LINUX*/                ((eo670_CallInfo *)callArray)[depth].pcAddr = (char *)0 + fp->fr_savpc;
/*LINUX*/                ((eo670_CallInfo *)callArray)[depth].frame = fp;
/*LINUX*/                ((eo670_CallInfo *)callArray)[depth].arg[0] = fp->fr_arg[0];
/*LINUX*/                ((eo670_CallInfo *)callArray)[depth].arg[1] = fp->fr_arg[1];
/*LINUX*/                ((eo670_CallInfo *)callArray)[depth].arg[2] = fp->fr_arg[2];
/*LINUX*/                ((eo670_CallInfo *)callArray)[depth].arg[3] = fp->fr_arg[3];
/*LINUX*/                break;
/*LINUX*/            case eo670_PointerArray:
/*LINUX*/                callArray[depth] = ((char *)0) + fp->fr_savpc;
/*LINUX*/                break;
/*LINUX*/            default: /* eo670_NoArrayJustOutput */
/*LINUX*/                sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*LINUX*/                         "(%d):0x%lx [0x%lx](0x%lx,0x%lx,0x%lx,0x%lx)\n",
/*LINUX*/                         depth, fp->fr_savpc, (long)fp, fp->fr_arg[0], fp->fr_arg[1], fp->fr_arg[2], fp->fr_arg[3] );
/*LINUX*/                eo670WriteString(eo670MessageBuffer);
/*LINUX*/                break;
/*LINUX*/            }
/*LINUX*/            ++depth;
/*LINUX*/         }
/*LINUX*/         fp = fp->fr_savfp;
/*LINUX*/    }
/*LINUX*/    return depth;
/*LINUX*/}
/*LINUX*/
/*LINUX*//*---------------------------------------------------------------------------*/
/*LINUX*/
#      if defined(WANT_DEMANGLING_SUPPORT)
/*LINUX*/static char   demanglingBuffer[MAX_SYMBOL_LENGTH];
#      endif /* WANT_DEMANGLING_SUPPORT */
/*LINUX*/
/*LINUX*/static void eo670_FillAddressInfo(char *pcAddr, eo670AddressInfo *addrInfo)
/*LINUX*/{
/*LINUX*/  static map_t *map;
/*LINUX*/  static eo670_ModuleChain *modulePointer;
/*LINUX*/  static Symbol_t *funcsym;
/*LINUX*/
/*LINUX*/    memset(addrInfo, 0, sizeof(eo670AddressInfo));
/*LINUX*/    addrInfo->sourceFileName = ""; /* No Source information on LINUX... */
/*LINUX*/    addrInfo->lineNumber = -1;
/*LINUX*/
/*LINUX*/    for ( modulePointer = eo670_ModuleAnchor
/*LINUX*/        ; modulePointer
/*LINUX*/        ; modulePointer = modulePointer->next )
/*LINUX*/    {
/*LINUX*/        if (  ( pcAddr >= ((char *)modulePointer->text_mapaddr) )
/*LINUX*/           && ( pcAddr <  ((char *)modulePointer->text_endaddr) ) )
/*LINUX*/        {
/*LINUX*/            map = (map_t *)modulePointer->osDependend;
/*LINUX*/	           break;
/*LINUX*/   	   }
/*LINUX*/    }
/*LINUX*/
/*LINUX*/    if ( !modulePointer )
/*LINUX*/    {
/*LINUX*/        addrInfo->demangledSymbol = "";
/*LINUX*/        addrInfo->offsetToSymbol = -1;
/*LINUX*/        addrInfo->moduleName = "<not in any loaded module>";
/*LINUX*/        addrInfo->symbol = "<not in any loaded region>";
/*LINUX*/        return;
/*LINUX*/    }
/*LINUX*/
/*LINUX*/    addrInfo->moduleName = modulePointer->file_name;
/*LINUX*/
/*LINUX*/    funcsym = ( map ? tb_get_func(pcAddr-(char *)0, map) : (Symbol_t *)0 );
/*LINUX*/    if ( 0 == funcsym) 
/*LINUX*/    {
/*LINUX*/        addrInfo->symbol = "<symbol not available>";
/*LINUX*/        addrInfo->demangledSymbol = "";
/*LINUX*/        addrInfo->offsetToSymbol = -1;
/*LINUX*/        return;
/*LINUX*/    }
/*LINUX*/
/*LINUX*/    addrInfo->symbol = funcsym->st_name;
/*LINUX*/    addrInfo->offsetToSymbol = pcAddr - (char *)funcsym->st_value;
/*LINUX*/    addrInfo->demangledSymbol = "";
/*LINUX*/    return;
/*LINUX*/}
/*LINUX*/
/*LINUX*/static SAPDB_Bool eo670_UnixGetAddressInfo(void *pcAddr, eo670AddressInfo *addrInfo)
/*LINUX*/{
/*LINUX*/    int    fillResult;
/*LINUX*/    static eo670AddressInfo localAddrInfo;
/*LINUX*/
/*LINUX*/    eo670_FillAddressInfo(pcAddr, &localAddrInfo);
/*LINUX*/    addrInfo->symbol = (char *)calloc( 1, strlen(localAddrInfo.symbol)+1
/*LINUX*/                                 + strlen(localAddrInfo.demangledSymbol)+1
/*LINUX*/                                 + strlen(localAddrInfo.moduleName)+1
/*LINUX*/                                 + strlen(localAddrInfo.sourceFileName)+1 );
/*LINUX*/    if ( !addrInfo->symbol )
/*LINUX*/    {
/*LINUX*/        eo670WriteString("Out of memory for symbol copy\n");
/*LINUX*/        fillResult = 0;
/*LINUX*/    }
/*LINUX*/    else
/*LINUX*/    {
/*LINUX*/        strcpy(addrInfo->symbol, localAddrInfo.symbol);
/*LINUX*/        addrInfo->demangledSymbol = addrInfo->symbol + strlen(localAddrInfo.symbol)+1;
/*LINUX*/        strcpy(addrInfo->demangledSymbol, localAddrInfo.demangledSymbol);
/*LINUX*/        addrInfo->moduleName = addrInfo->demangledSymbol + strlen(localAddrInfo.demangledSymbol)+1;
/*LINUX*/        strcpy(addrInfo->moduleName, localAddrInfo.moduleName);
/*LINUX*/        addrInfo->sourceFileName = addrInfo->moduleName + strlen(localAddrInfo.moduleName)+1;
/*LINUX*/        strcpy(addrInfo->sourceFileName, localAddrInfo.sourceFileName);
/*LINUX*/        addrInfo->offsetToSymbol = localAddrInfo.offsetToSymbol;
/*LINUX*/        addrInfo->lineNumber = localAddrInfo.lineNumber;
/*LINUX*/        fillResult = 1;
/*LINUX*/    }
/*LINUX*/    return fillResult;
/*LINUX*/}
/*LINUX*/
/*LINUX*/static void eo670_OsFreeAddressInfo(eo670AddressInfo *addrInfo)
/*LINUX*/{
/*LINUX*/    free(addrInfo->symbol); /* allocated in one chunk.... */
/*LINUX*/}
/*LINUX*/
/*LINUX*/static void eo670_LinuxShowCallInfoAddon(eo670_CallInfo *callInfo)
/*LINUX*/{
/*LINUX*/    sp77sprintf(eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*LINUX*/         "      [0x%lx] (0x%lx,0x%lx,0x%lx,0x%lx)\n",
/*LINUX*/         (long)callInfo->frame, 
/*LINUX*/         callInfo->arg[0],
/*LINUX*/         callInfo->arg[1],
/*LINUX*/         callInfo->arg[2],
/*LINUX*/         callInfo->arg[3] );
/*LINUX*/    eo670WriteString(eo670MessageBuffer);
/*LINUX*/}
#    endif /* LINUX */

#    ifdef SUN
/*SUN*/static void
/*SUN*/eo670_UpdateModuleChain()
/*SUN*/{
/*SUN*/  static int       fd_proc;	/* /proc fd                           */
/*SUN*/  static map_t    *map;      /* Main executable map                */
/*SUN*/  static prmap_t  *prmaps;    /* temporary allocated process maps   */
/*SUN*/  static int       i;
/*SUN*/  static int bytesRead;
/*SUN*/  static struct stat mapStat;
/*SUN*/    map = 0;
/*SUN*/    prmaps = 0;
/*SUN*/    sp77sprintf(eo670MessageBuffer, sizeof(eo670MessageBuffer), "/proc/%d/map", getpid());
/*SUN*/
/*SUN*/    if ((fd_proc = open(eo670MessageBuffer, O_RDONLY)) == -1)
/*SUN*/    {
/*SUN*/        sp77sprintf(eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*SUN*/                  "Cant open /proc/%d/map : %d\n", getpid(), errno);
/*SUN*/        eo670WriteString(eo670MessageBuffer);
/*SUN*/        return;
/*SUN*/    }
/*SUN*/
/*SUN*/    /* Reading individual entries does not work. (SUN feature....) */
/*SUN*/    if ( -1 == fstat(fd_proc, &mapStat) )
/*SUN*/    {
/*SUN*/        sp77sprintf(eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*SUN*/                  "Cant stat /proc/%d/map : %d\n", getpid(), errno);
/*SUN*/        eo670WriteString(eo670MessageBuffer);
/*SUN*/        (void) close(fd_proc);
/*SUN*/        return;
/*SUN*/    }
/*SUN*/
/*SUN*/    if ( mapStat.st_size <= sizeof(prmap_t) 
/*SUN*/      || (mapStat.st_size % sizeof(prmap_t)) != 0 )
/*SUN*/    {
/*SUN*/        sp77sprintf(eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*SUN*/                  "Cant stat /proc/%d/map : %d\n", getpid(), errno);
/*SUN*/        eo670WriteString(eo670MessageBuffer);
/*SUN*/        (void) close(fd_proc);
/*SUN*/        return;
/*SUN*/    }
/*SUN*/
/*SUN*/    prmaps = NOHEAP_CALLOC(mapStat.st_size, 1);
/*SUN*/    if ( !prmaps )
/*SUN*/    {
/*SUN*/        sp77sprintf(eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*SUN*/                  "Out of memory: Wanted %d bytes for copy of map file\n", 
/*SUN*/                  (int)mapStat.st_size );
/*SUN*/        eo670WriteString(eo670MessageBuffer);
/*SUN*/        (void) close(fd_proc);
/*SUN*/        return;
/*SUN*/    }
/*SUN*/
/*SUN*/    bytesRead = 0; /* Read the complete file in */
/*SUN*/    do
/*SUN*/    {   /* Reading individual entries does not work. (SUN feature....) */
/*SUN*/        i = read(fd_proc, ((char *)prmaps) + bytesRead, mapStat.st_size - bytesRead);
/*SUN*/        if ( i < 0 )
/*SUN*/        {
/*SUN*/            sp77sprintf(eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*SUN*/                      "Failed to read complete map file: %d\n", errno);
/*SUN*/            eo670WriteString(eo670MessageBuffer);
/*SUN*/            (void) close(fd_proc);
/*SUN*/            return;
/*SUN*/        }
/*SUN*/        bytesRead += i;
/*SUN*/    } while ( bytesRead < mapStat.st_size );
/*SUN*/
/*SUN*/    close(fd_proc);
/*SUN*/
/*SUN*/    /* Get address space (shared library) mappings                    */
/*SUN*/    for ( i = 0; (i * sizeof(prmap_t)) < mapStat.st_size; ++i )
/*SUN*/    {
/*SUN*/        static prmap_t *pm;
/*SUN*/        static int      fd_obj;
/*SUN*/
/*SUN*/        pm = prmaps + i;
/*SUN*/        if (pm->pr_mflags != (MA_READ | MA_EXEC))
/*SUN*/            continue;
/*SUN*/
/*SUN*/        sp77sprintf(eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*SUN*/                    "/proc/%d/object/%s", getpid(), pm->pr_mapname);
/*SUN*/        fd_obj = open(eo670MessageBuffer, O_RDONLY);
/*SUN*/        if(fd_obj == -1) 
/*SUN*/        {
/*SUN*/            continue;
/*SUN*/        }
/*SUN*/
/*SUN*/        map = get_symtab(fd_obj, (caddr_t)pm->pr_vaddr);
/*SUN*/        close(fd_obj);
/*SUN*/
/*SUN*/        switch( eo670_AddToLoadedModuleChain( (map ? map->filename : "????"),
/*SUN*/                                              (char *)0 + pm->pr_vaddr,
/*SUN*/                                              ((char *)pm->pr_vaddr) + pm->pr_size,
/*SUN*/                                              map ) )
/*SUN*/        {
/*SUN*/            case 1: break;
/*SUN*/            case 0: freemap(map); break;
/*SUN*/            default:goto cleanup;
/*SUN*/        }
/*SUN*/    }
/*SUN*/    return;
/*SUN*/ cleanup:
/*SUN*/    freemap(map);
/*SUN*/    if ( prmaps ) NOHEAP_FREE(prmaps);
/*SUN*/    return;
/*SUN*/}
/*SUN*/
#      ifdef BIT64
#        define SP_BIAS V9BIAS64
#      else
#        define SP_BIAS 0
#      endif
/*SUN*/
/*SUN*//*---------------------------------------------------------------------------*/
/*SUN*/static struct frame *BiasedFramePtr(void *fp)
/*SUN*/{
/*SUN*/    return (struct frame *)(((unsigned long)fp)+SP_BIAS);
/*SUN*/}
/*SUN*/
/*SUN*//*---------------------------------------------------------------------------*/
/*SUN*/
/*SUN*/static void eo670_RegisterDump(void *pContext)
/*SUN*/{
/*SUN*/    static ucontext_t *context;
/*SUN*/    static int iReg;
/*SUN*/
/*SUN*/    context = (ucontext_t *)pContext;
/*SUN*/
/*SUN*/    eo670WriteString("---------------------------------------\n");
/*SUN*/    eo670WriteString( "Aliases: o0 == r0, o1 == r1, o6 == sp\n" );
/*SUN*/    eo670WriteString( "         o0 == y, fp == i6, rtn == i7\n" );
/*SUN*/    eo670WriteString( "\n" );
/*SUN*/
/*SUN*/    for ( iReg = 0; iReg < 4; iReg++ )
/*SUN*/    {
/*SUN*/        sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*SUN*/                     "g%d  g%d   0x%016lx 0x%016lx\n",
/*SUN*/                      2*iReg,
/*SUN*/                      1+2*iReg,
/*SUN*/    iReg==0 ? (SAPDB_Long)0 : (SAPDB_Long)context->uc_mcontext.gregs[(REG_G1-1)+2*iReg],
/*SUN*/                      (SAPDB_Long)context->uc_mcontext.gregs[REG_G1+2*iReg] );
/*SUN*/        eo670WriteString( eo670MessageBuffer );
/*SUN*/    }
/*SUN*/    eo670WriteString("\n");
/*SUN*/
/*SUN*/    for ( iReg = 0; iReg < 4; iReg++ )
/*SUN*/    {
/*SUN*/        sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*SUN*/                     "o%d  o%d   0x%016lx 0x%016lx\n",
/*SUN*/                      2*iReg,
/*SUN*/                      1+2*iReg,
/*SUN*/                      (SAPDB_Long)context->uc_mcontext.gregs[REG_O0+2*iReg],
/*SUN*/                      (SAPDB_Long)context->uc_mcontext.gregs[REG_O1+2*iReg] );
/*SUN*/        eo670WriteString( eo670MessageBuffer );
/*SUN*/    }
/*SUN*/    eo670WriteString("\n");
/*SUN*/
/*SUN*/    if ( context->uc_mcontext.gwins )
/*SUN*/    {
/*SUN*/      eo670WriteString( "Unexpected register window setting, cant dump %l0-%l7,%i0-%i7\n");
/*SUN*/    }
/*SUN*/    else
/*SUN*/    {
/*SUN*/      static struct frame   *fp;
/*SUN*/      fp = BiasedFramePtr((void *)context->uc_mcontext.gregs[REG_SP]); /* stack pointer*/
/*SUN*/      for ( iReg = 0; iReg < 4; iReg++ )
/*SUN*/      {
/*SUN*/        sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*SUN*/                     "l%d  l%d   0x%016lx 0x%016lx\n",
/*SUN*/                      2*iReg,
/*SUN*/                      1+2*iReg,
/*SUN*/                      fp->fr_local[2*iReg],
/*SUN*/                      fp->fr_local[1+2*iReg] );
/*SUN*/        eo670WriteString( eo670MessageBuffer );
/*SUN*/      }
/*SUN*/      eo670WriteString("\n");
/*SUN*/
/*SUN*/      for ( iReg = 0; iReg < 3; iReg++ )
/*SUN*/      {
/*SUN*/        sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*SUN*/                     "i%d  i%d   0x%016lx 0x%016lx\n",
/*SUN*/                      2*iReg,
/*SUN*/                      1+2*iReg,
/*SUN*/                      fp->fr_arg[2*iReg],
/*SUN*/                      fp->fr_arg[1+2*iReg] );
/*SUN*/        eo670WriteString( eo670MessageBuffer );
/*SUN*/      }
/*SUN*/      sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*SUN*/                   "fp  rtn  0x%016lx 0x%016lx\n",
/*SUN*/                      (SAPDB_Long)fp->fr_savfp,
/*SUN*/                      (SAPDB_Long)fp->fr_savpc );
/*SUN*/      eo670WriteString( eo670MessageBuffer );
/*SUN*/      eo670WriteString("\n");
/*SUN*/    }
/*SUN*/
/*SUN*/    sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*SUN*/                 "y        0x%016lx\n",
/*SUN*/                 (SAPDB_Long)context->uc_mcontext.gregs[REG_Y] );
/*SUN*/    eo670WriteString( eo670MessageBuffer );
#      if defined(__sparcv9)
/*SUN*/    sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*SUN*/                 "ccr      0x%016lx\n",
/*SUN*/                 (SAPDB_Long)context->uc_mcontext.gregs[REG_CCR] );
/*SUN*/    eo670WriteString( eo670MessageBuffer );
/*SUN*/    sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*SUN*/                 "asi      0x%016lx\n",
/*SUN*/                 (SAPDB_Long)context->uc_mcontext.gregs[REG_ASI] );
/*SUN*/    eo670WriteString( eo670MessageBuffer );
/*SUN*/    sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*SUN*/                 "fprs     0x%016lx\n",
/*SUN*/                 (SAPDB_Long)context->uc_mcontext.gregs[REG_FPRS] );
/*SUN*/    eo670WriteString( eo670MessageBuffer );
#      else /* __sparcv9 */
/*SUN*/    sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*SUN*/                 "psr      0x%016lx\n",
/*SUN*/                 (SAPDB_Long)context->uc_mcontext.gregs[REG_PSR] );
/*SUN*/    eo670WriteString( eo670MessageBuffer );
#      endif /* __sparcv9 */
/*SUN*/    sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*SUN*/                 "pc       0x%016lx\n",
/*SUN*/                 (SAPDB_Long)context->uc_mcontext.gregs[REG_PC] );
/*SUN*/    eo670WriteString( eo670MessageBuffer );
/*SUN*/    sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*SUN*/                 "npc      0x%016lx\n",
/*SUN*/                 (SAPDB_Long)context->uc_mcontext.gregs[REG_nPC] );
/*SUN*/    eo670WriteString( eo670MessageBuffer );
/*SUN*/
/*SUN*/    eo670WriteString("---------------------------------------\n");
/*SUN*/
/*SUN*/    if ( context->uc_mcontext.fpregs.fpu_en )
/*SUN*/    {
#      if defined(__sparcv9)
/*SUN*/        for ( iReg = 0; iReg < 32; iReg++ )
#      else /* __sparcv9 */
/*SUN*/        for ( iReg = 0; iReg < 16; iReg++ )
#      endif /* __sparcv9 */
/*SUN*/        {
/*SUN*/            sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*SUN*/                     "f%02df%02d  [0x%016qx]%.20e\n",
/*SUN*/                      2*iReg,
/*SUN*/                      1+2*iReg,
/*SUN*/                      *(SAPDB_UInt8 *)&context->uc_mcontext.fpregs.fpu_fr.fpu_dregs[iReg],
/*SUN*/                      *(double *)&(context->uc_mcontext.fpregs.fpu_fr.fpu_dregs[iReg]) );
/*SUN*/            eo670WriteString( eo670MessageBuffer );
/*SUN*/        }
/*SUN*/        sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*SUN*/                 "fsr      0x%016lx\n",
/*SUN*/                 (SAPDB_Long)context->uc_mcontext.fpregs.fpu_fsr );
/*SUN*/        eo670WriteString( eo670MessageBuffer );
/*SUN*/    }
/*SUN*/    else
/*SUN*/    {
/*SUN*/        eo670WriteString("Floating point unit unused\n");
/*SUN*/    }
/*SUN*/    eo670WriteString("---------------------------------------\n");
/*SUN*/
/*SUN*/}
/*SUN*//*---------------------------------------------------------------------------*/
/*SUN*/
/*SUN*/SAPDB_Int4 eo670_FastFillCallArray(SAPDB_Int4 skipCount, SAPDB_Int4 level, void **callArray)
/*SUN*/{
/*SUN*/    ucontext_t     *context;
/*SUN*/    SAPDB_Int4      depth;
/*SUN*/    unsigned long   pc;
/*SUN*/    struct frame   *fp;
/*SUN*/    ucontext_t      localContext;
/*SUN*/
/*SUN*/    if ( getcontext (&localContext) )
/*SUN*/    {
/*SUN*/        eo670WriteString("Failed to get ucontext\n");
/*SUN*/        return 0;
/*SUN*/    }
/*SUN*/    context = &localContext;
/*SUN*/
/*SUN*/    pc = context->uc_mcontext.gregs[REG_PC]; /* program counter           */
/*SUN*/    fp = BiasedFramePtr((void *)context->uc_mcontext.gregs[REG_SP]); /* stack pointer*/
/*SUN*/
/*SUN*/    depth = 0;
/*SUN*/    if ( depth < level && skipCount <= 0 )
/*SUN*/    {
/*SUN*/        callArray[depth] = ((char *)0) + pc;
/*SUN*/        ++depth;
/*SUN*/    }
/*SUN*/
/*SUN*/    while ( depth < level
/*SUN*/        && (((long)pc)&3) == 0
/*SUN*/        &&  fp 
/*SUN*/        && (((long)fp)&3) == 0
/*SUN*/        && (fp != BiasedFramePtr(fp->fr_savfp)) )
/*SUN*/    {
/*SUN*/        pc = fp->fr_savpc;
/*SUN*/
/*SUN*/        if ( skipCount > 0 )
/*SUN*/        {
/*SUN*/            --skipCount;
/*SUN*/	       }
/*SUN*/	       else
/*SUN*/        {
/*SUN*/            callArray[depth] = ((char *)0) + pc;
/*SUN*/            ++depth;
/*SUN*/        }
/*SUN*/
/*SUN*/        fp = BiasedFramePtr(fp->fr_savfp);
/*SUN*/    }
/*SUN*/    return depth;
/*SUN*/}
/*SUN*/
/*SUN*//*---------------------------------------------------------------------------*/
/*SUN*/
/*SUN*/SAPDB_Int4 eo670_UnixFillCallArray(SAPDB_Int4 skipCount, SAPDB_Int4 level, eo670_CallArrayType addCallInfo, void **callArray, void *pContext)
/*SUN*/{
/*SUN*/    static ucontext_t      *context;
/*SUN*/    static SAPDB_Int4      depth;
/*SUN*/    static unsigned long   pc;
/*SUN*/    static struct frame   *fp;
/*SUN*/
/*SUN*/    if ( 0 == pContext )
/*SUN*/    {
/*SUN*/        static ucontext_t localContext;
/*SUN*/        if ( getcontext (&localContext) )
/*SUN*/        {
/*SUN*/            eo670WriteString("Failed to get ucontext\n");
/*SUN*/            return 0;
/*SUN*/        }
/*SUN*/        context = &localContext;
/*SUN*/    }
/*SUN*/    else
/*SUN*/    {
/*SUN*/        skipCount = 0;
/*SUN*/        context = (ucontext_t *)pContext;
/*SUN*/    }
/*SUN*/
/*SUN*/    pc = context->uc_mcontext.gregs[REG_PC]; /* program counter           */
/*SUN*/    fp = BiasedFramePtr((void *)context->uc_mcontext.gregs[REG_SP]); /* stack pointer*/
/*SUN*/
/*SUN*/    depth = 0;
/*SUN*/    if ( depth < level && skipCount <= 0 )
/*SUN*/    {
/*SUN*/        switch ( addCallInfo )
/*SUN*/        {
/*SUN*/        case eo670_FullInfoArray:
/*SUN*/            ((eo670_CallInfo *)callArray)[depth].pcAddr = (char *)0 + pc;
/*SUN*/            ((eo670_CallInfo *)callArray)[depth].frame = fp;
/*SUN*/            if ( fp && (0 == (((long)fp)&(sizeof(long)-1))) )
/*SUN*/            {
/*SUN*/                ((eo670_CallInfo *)callArray)[depth].arg[0] = fp->fr_arg[0];
/*SUN*/                ((eo670_CallInfo *)callArray)[depth].arg[1] = fp->fr_arg[1];
/*SUN*/                ((eo670_CallInfo *)callArray)[depth].arg[2] = fp->fr_arg[2];
/*SUN*/                ((eo670_CallInfo *)callArray)[depth].arg[3] = fp->fr_arg[3];
/*SUN*/                ((eo670_CallInfo *)callArray)[depth].arg[4] = fp->fr_arg[4];
/*SUN*/                ((eo670_CallInfo *)callArray)[depth].arg[5] = fp->fr_arg[5];
/*SUN*/            }
/*SUN*/            else
/*SUN*/            {
/*SUN*/                ((eo670_CallInfo *)callArray)[depth].arg[0] = 0xdead;
/*SUN*/                ((eo670_CallInfo *)callArray)[depth].arg[1] = 0xc0de;
/*SUN*/                ((eo670_CallInfo *)callArray)[depth].arg[2] = 0xdead;
/*SUN*/                ((eo670_CallInfo *)callArray)[depth].arg[3] = 0xc0de;
/*SUN*/                ((eo670_CallInfo *)callArray)[depth].arg[4] = 0xdead;
/*SUN*/                ((eo670_CallInfo *)callArray)[depth].arg[5] = 0xc0de;
/*SUN*/            }
/*SUN*/            break;
/*SUN*/        case eo670_PointerArray:
/*SUN*/            callArray[depth] = ((char *)0) + pc;
/*SUN*/            break;
/*SUN*/        default: /* eo670_NoArrayJustOutput */
/*SUN*/            if ( fp && (0 == (((long)fp)&(sizeof(long)-1))) )
/*SUN*/            {
/*SUN*/                sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*SUN*/                             "(%d):0x%lx [0x%lx] (0x%lx,0x%lx,0x%lx,0x%lx,0x%lx,0x%lx)\n",
/*SUN*/                             depth, pc, fp, fp->fr_arg[0], fp->fr_arg[1], fp->fr_arg[2], fp->fr_arg[3], fp->fr_arg[4], fp->fr_arg[5] );
/*SUN*/            }
/*SUN*/            else
/*SUN*/            {
/*SUN*/                sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*SUN*/                             "(%d):0x%lx [0x%lx] (?,?,?,?,?,?)\n",
/*SUN*/                             depth, pc, fp );
/*SUN*/            }
/*SUN*/            eo670WriteString(eo670MessageBuffer);
/*SUN*/            break;
/*SUN*/        }
/*SUN*/        ++depth;
/*SUN*/    }
/*SUN*/
/*SUN*/    while ( depth < level
/*SUN*/        && (((long)pc)&3) == 0
/*SUN*/        &&  fp 
/*SUN*/        && (((long)fp)&3) == 0
/*SUN*/        && (fp != BiasedFramePtr(fp->fr_savfp))
/*SUN*/	    && ((((unsigned long) BiasedFramePtr(fp->fr_savfp)) + SP_BIAS) >= USRTEXT) )
/*SUN*/    {
/*SUN*/        pc = fp->fr_savpc;
/*SUN*/
/*SUN*/        if ( skipCount > 0 )
/*SUN*/        {
/*SUN*/            --skipCount;
/*SUN*/	       }
/*SUN*/	       else
/*SUN*/        {
/*SUN*/            switch ( addCallInfo )
/*SUN*/            {
/*SUN*/            case eo670_FullInfoArray:
/*SUN*/                ((eo670_CallInfo *)callArray)[depth].pcAddr = (char *)0 + pc;
/*SUN*/                ((eo670_CallInfo *)callArray)[depth].frame = fp;
/*SUN*/                ((eo670_CallInfo *)callArray)[depth].arg[0] = fp->fr_arg[0];
/*SUN*/                ((eo670_CallInfo *)callArray)[depth].arg[1] = fp->fr_arg[1];
/*SUN*/                ((eo670_CallInfo *)callArray)[depth].arg[2] = fp->fr_arg[2];
/*SUN*/                ((eo670_CallInfo *)callArray)[depth].arg[3] = fp->fr_arg[3];
/*SUN*/                ((eo670_CallInfo *)callArray)[depth].arg[4] = fp->fr_arg[4];
/*SUN*/                ((eo670_CallInfo *)callArray)[depth].arg[5] = fp->fr_arg[5];
/*SUN*/                break;
/*SUN*/            case eo670_PointerArray:
/*SUN*/                callArray[depth] = ((char *)0) + pc;
/*SUN*/                break;
/*SUN*/            default: /* eo670_NoArrayJustOutput */
/*SUN*/                sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*SUN*/                             "(%d):0x%lx [0x%lx] (0x%lx,0x%lx,0x%lx,0x%lx,0x%lx,0x%lx)\n",
/*SUN*/                             depth, pc, fp, fp->fr_arg[0], fp->fr_arg[1], fp->fr_arg[2], fp->fr_arg[3], fp->fr_arg[4], fp->fr_arg[5] );
/*SUN*/                eo670WriteString(eo670MessageBuffer);
/*SUN*/                break;
/*SUN*/            }
/*SUN*/            ++depth;
/*SUN*/        }
/*SUN*/
/*SUN*/        fp = BiasedFramePtr(fp->fr_savfp);
/*SUN*/    }
/*SUN*/    return depth;
/*SUN*/}
/*SUN*/
/*SUN*//*---------------------------------------------------------------------------*/
/*SUN*/
#      if defined(WANT_DEMANGLING_SUPPORT)
/*SUN*/static char   demanglingBuffer[MAX_SYMBOL_LENGTH];
#      endif /* WANT_DEMANGLING_SUPPORT */
/*SUN*/
/*SUN*/static void eo670_FillAddressInfo(char *pcAddr, eo670AddressInfo *addrInfo)
/*SUN*/{
/*SUN*/  static map_t *map;
/*SUN*/  static eo670_ModuleChain *modulePointer;
/*SUN*/  static Symbol_t *funcsym;
/*SUN*/
/*SUN*/    memset(addrInfo, 0, sizeof(eo670AddressInfo));
/*SUN*/    addrInfo->sourceFileName = ""; /* No Source information on SUN... */
/*SUN*/    addrInfo->lineNumber = -1;
/*SUN*/
/*SUN*/    for ( modulePointer = eo670_ModuleAnchor
/*SUN*/        ; modulePointer
/*SUN*/        ; modulePointer = modulePointer->next )
/*SUN*/    {
/*SUN*/        if (  ( pcAddr >= ((char *)modulePointer->text_mapaddr) )
/*SUN*/           && ( pcAddr <  ((char *)modulePointer->text_endaddr) ) )
/*SUN*/        {
/*SUN*/            map = (map_t *)modulePointer->osDependend;
/*SUN*/	           break;
/*SUN*/   	   }
/*SUN*/    }
/*SUN*/
/*SUN*/    if ( !modulePointer )
/*SUN*/    {
/*SUN*/        addrInfo->demangledSymbol = "";
/*SUN*/        addrInfo->offsetToSymbol = -1;
/*SUN*/        addrInfo->moduleName = "<not in any loaded module>";
/*SUN*/        addrInfo->symbol = "<not in any loaded region>";
/*SUN*/        return;
/*SUN*/    }
/*SUN*/
/*SUN*/    addrInfo->moduleName = modulePointer->file_name;
/*SUN*/
/*SUN*/    funcsym = ( map ? tb_get_func(pcAddr-(char *)0, map) : (Symbol_t *)0 );
/*SUN*/    if ( 0 == funcsym) 
/*SUN*/    {
/*SUN*/        addrInfo->symbol = "<symbol not available>";
/*SUN*/        addrInfo->demangledSymbol = "";
/*SUN*/        addrInfo->offsetToSymbol = -1;
/*SUN*/        return;
/*SUN*/    }
/*SUN*/
/*SUN*/    addrInfo->symbol = funcsym->st_name;
/*SUN*/    addrInfo->offsetToSymbol = pcAddr - (char *)funcsym->st_value;
#      if defined(WANT_DEMANGLING_SUPPORT)
/*SUN*/    cplus_demangle(addrInfo->symbol, demanglingBuffer, sizeof(demanglingBuffer));
/*SUN*/    addrInfo->demangledSymbol = demanglingBuffer;
#      else
/*SUN*/    addrInfo->demangledSymbol = "";
#      endif /* WANT_DEMANGLING_SUPPORT */
/*SUN*/    return;
/*SUN*/}
/*SUN*/
/*SUN*/static SAPDB_Bool eo670_UnixGetAddressInfo(void *pcAddr, eo670AddressInfo *addrInfo)
/*SUN*/{
/*SUN*/    int    fillResult;
/*SUN*/    static eo670AddressInfo localAddrInfo;
/*SUN*/
/*SUN*/    eo670_FillAddressInfo(pcAddr, &localAddrInfo);
/*SUN*/    addrInfo->symbol = (char *)calloc( 1, strlen(localAddrInfo.symbol)+1
/*SUN*/                                 + strlen(localAddrInfo.demangledSymbol)+1
/*SUN*/                                 + strlen(localAddrInfo.moduleName)+1
/*SUN*/                                 + strlen(localAddrInfo.sourceFileName)+1 );
/*SUN*/    if ( !addrInfo->symbol )
/*SUN*/    {
/*SUN*/        eo670WriteString("Out of memory for symbol copy\n");
/*SUN*/        fillResult = 0;
/*SUN*/    }
/*SUN*/    else
/*SUN*/    {
/*SUN*/        strcpy(addrInfo->symbol, localAddrInfo.symbol);
/*SUN*/        addrInfo->demangledSymbol = addrInfo->symbol + strlen(localAddrInfo.symbol)+1;
/*SUN*/        strcpy(addrInfo->demangledSymbol, localAddrInfo.demangledSymbol);
/*SUN*/        addrInfo->moduleName = addrInfo->demangledSymbol + strlen(localAddrInfo.demangledSymbol)+1;
/*SUN*/        strcpy(addrInfo->moduleName, localAddrInfo.moduleName);
/*SUN*/        addrInfo->sourceFileName = addrInfo->moduleName + strlen(localAddrInfo.moduleName)+1;
/*SUN*/        strcpy(addrInfo->sourceFileName, localAddrInfo.sourceFileName);
/*SUN*/        addrInfo->offsetToSymbol = localAddrInfo.offsetToSymbol;
/*SUN*/        addrInfo->lineNumber = localAddrInfo.lineNumber;
/*SUN*/        fillResult = 1;
/*SUN*/    }
/*SUN*/    return fillResult;
/*SUN*/}
/*SUN*/
/*SUN*/static void eo670_OsFreeAddressInfo(eo670AddressInfo *addrInfo)
/*SUN*/{
/*SUN*/    free(addrInfo->symbol); /* allocated in one chunk.... */
/*SUN*/}
/*SUN*/
/*SUN*/static void eo670_SunShowCallInfoAddon(eo670_CallInfo *callInfo)
/*SUN*/{
/*SUN*/    sp77sprintf(eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*SUN*/         "      [0x%lx] (0x%lx,0x%lx,0x%lx,0x%lx,0x%lx,0x%lx)\n",
/*SUN*/         (long)callInfo->frame,
/*SUN*/         callInfo->arg[0],
/*SUN*/         callInfo->arg[1],
/*SUN*/         callInfo->arg[2],
/*SUN*/         callInfo->arg[3],
/*SUN*/         callInfo->arg[4],
/*SUN*/         callInfo->arg[5] );
/*SUN*/    eo670WriteString(eo670MessageBuffer);
/*SUN*/}
#    endif /* SUN */
#  endif /* LINUX+SUN */

#  ifdef HPUX
/*HPUX*/static void eo670_RegisterDump(void *pContext)
/*HPUX*/{
#    ifdef BIT64
/*HPUX*/    static ucontext_t *context;
/*HPUX*/    static unsigned long *pRegs;
/*HPUX*/    static int        iReg;
/*HPUX*/
/*HPUX*/    context = (ucontext_t *)pContext;
/*HPUX*/
/*HPUX*/    pRegs = (unsigned long *)&(context->uc_mcontext.ss_wide.ss_64);
/*HPUX*/
/*HPUX*/    eo670WriteString( "-------------------------------------\n" );
/*HPUX*/    eo670WriteString( "Aliases: r2  == rp,   r23 == arg3, r24 == arg2\n" );
/*HPUX*/    eo670WriteString( "         r25 == arg1, r26 == arg0, r27 == dp\n" );
/*HPUX*/    eo670WriteString( "         r28 == ret0, r29 == ret1, r30 == sp\n" );
/*HPUX*/    eo670WriteString( "\n" );
/*HPUX*/    for ( iReg = 0; iReg < 16; iReg++ )
/*HPUX*/    {
/*HPUX*/        sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*HPUX*/                     "r%02d  r%02d 0x%016lx 0x%016lx\n",
/*HPUX*/                      2*iReg,
/*HPUX*/                      1+2*iReg,
/*HPUX*/                      pRegs[2*iReg],
/*HPUX*/                      pRegs[1+2*iReg] );
/*HPUX*/        eo670WriteString( eo670MessageBuffer );
/*HPUX*/    }
/*HPUX*/
/*HPUX*/    eo670WriteString("\n");
/*HPUX*/
/*HPUX*/    sp77sprintf( eo670MessageBuffer,
/*HPUX*/                 sizeof(eo670MessageBuffer),
/*HPUX*/                 "cpustate 0x%016lx\n",
/*HPUX*/                 context->uc_mcontext.ss_wide.ss_64.ss_cpustate);
/*HPUX*/    eo670WriteString(eo670MessageBuffer);
/*HPUX*/
/*HPUX*/    sp77sprintf( eo670MessageBuffer,
/*HPUX*/                 sizeof(eo670MessageBuffer),
/*HPUX*/                 "pcoqHead pcsqHead 0x%016lx 0x%016lx\n",
/*HPUX*/                 context->uc_mcontext.ss_wide.ss_64.ss_pcoq_head,
/*HPUX*/                 context->uc_mcontext.ss_wide.ss_64.ss_pcsq_head);
/*HPUX*/    eo670WriteString(eo670MessageBuffer);
/*HPUX*/
/*HPUX*/    sp77sprintf( eo670MessageBuffer,
/*HPUX*/                 sizeof(eo670MessageBuffer),
/*HPUX*/                 "pcoqTail pcsqTail 0x%016lx 0x%016lx\n",
/*HPUX*/                 context->uc_mcontext.ss_wide.ss_64.ss_pcoq_tail,
/*HPUX*/                 context->uc_mcontext.ss_wide.ss_64.ss_pcsq_tail);
/*HPUX*/    eo670WriteString(eo670MessageBuffer);
/*HPUX*/
/*HPUX*/    eo670WriteString("\n");
/*HPUX*/
/*HPUX*/    sp77sprintf( eo670MessageBuffer,
/*HPUX*/                 sizeof(eo670MessageBuffer),
/*HPUX*/                 "cr0  cr8  0x%016lx 0x%016lx\n",
/*HPUX*/                 context->uc_mcontext.ss_wide.ss_64.ss_cr0,
/*HPUX*/                 context->uc_mcontext.ss_wide.ss_64.ss_cr8);
/*HPUX*/    eo670WriteString(eo670MessageBuffer);
/*HPUX*/
/*HPUX*/    sp77sprintf( eo670MessageBuffer,
/*HPUX*/                 sizeof(eo670MessageBuffer),
/*HPUX*/                 "cr9  cr10 0x%016lx 0x%016lx\n",
/*HPUX*/                 context->uc_mcontext.ss_wide.ss_64.ss_cr9,
/*HPUX*/                 context->uc_mcontext.ss_wide.ss_64.ss_cr10);
/*HPUX*/    eo670WriteString(eo670MessageBuffer);
/*HPUX*/
/*HPUX*/    sp77sprintf( eo670MessageBuffer,
/*HPUX*/                 sizeof(eo670MessageBuffer),
/*HPUX*/                 "sar  cr12 0x%016lx 0x%016lx\n",
/*HPUX*/                 context->uc_mcontext.ss_wide.ss_64.ss_cr11,
/*HPUX*/                 context->uc_mcontext.ss_wide.ss_64.ss_cr12);
/*HPUX*/    eo670WriteString(eo670MessageBuffer);
/*HPUX*/
/*HPUX*/    sp77sprintf( eo670MessageBuffer,
/*HPUX*/                 sizeof(eo670MessageBuffer),
/*HPUX*/                 "cr13 cr15 0x%016lx 0x%016lx\n",
/*HPUX*/                 context->uc_mcontext.ss_wide.ss_64.ss_cr13,
/*HPUX*/                 context->uc_mcontext.ss_wide.ss_64.ss_cr15);
/*HPUX*/    eo670WriteString(eo670MessageBuffer);
/*HPUX*/
/*HPUX*/    sp77sprintf( eo670MessageBuffer,
/*HPUX*/                 sizeof(eo670MessageBuffer),
/*HPUX*/                 "cr19 cr20 0x%016lx 0x%016lx\n",
/*HPUX*/                 context->uc_mcontext.ss_wide.ss_64.ss_cr19,
/*HPUX*/                 context->uc_mcontext.ss_wide.ss_64.ss_cr20);
/*HPUX*/    eo670WriteString(eo670MessageBuffer);
/*HPUX*/
/*HPUX*/    sp77sprintf( eo670MessageBuffer,
/*HPUX*/                 sizeof(eo670MessageBuffer),
/*HPUX*/                 "cr21 cr22 0x%016lx 0x%016lx\n",
/*HPUX*/                 context->uc_mcontext.ss_wide.ss_64.ss_cr21,
/*HPUX*/                 context->uc_mcontext.ss_wide.ss_64.ss_cr22);
/*HPUX*/    eo670WriteString(eo670MessageBuffer);
/*HPUX*/
/*HPUX*/    sp77sprintf( eo670MessageBuffer,
/*HPUX*/                 sizeof(eo670MessageBuffer),
/*HPUX*/                 "cr24 cr25 0x%016lx 0x%016lx\n",
/*HPUX*/                 context->uc_mcontext.ss_wide.ss_64.ss_cr24,
/*HPUX*/                 context->uc_mcontext.ss_wide.ss_64.ss_cr25);
/*HPUX*/    eo670WriteString(eo670MessageBuffer);
/*HPUX*/
/*HPUX*/    sp77sprintf( eo670MessageBuffer,
/*HPUX*/                 sizeof(eo670MessageBuffer),
/*HPUX*/                 "cr26 cr27 0x%016lx 0x%016lx\n",
/*HPUX*/                 context->uc_mcontext.ss_wide.ss_64.ss_cr26,
/*HPUX*/                 context->uc_mcontext.ss_wide.ss_64.ss_cr27);
/*HPUX*/    eo670WriteString(eo670MessageBuffer);
/*HPUX*/
/*HPUX*/    eo670WriteString("\n");
/*HPUX*/
/*HPUX*/    sp77sprintf( eo670MessageBuffer,
/*HPUX*/                 sizeof(eo670MessageBuffer),
/*HPUX*/                 "sr0 sr1 0x%016lx 0x%016lx\n",
/*HPUX*/                 context->uc_mcontext.ss_wide.ss_64.ss_sr0,
/*HPUX*/                 context->uc_mcontext.ss_wide.ss_64.ss_sr1);
/*HPUX*/    eo670WriteString(eo670MessageBuffer);
/*HPUX*/
/*HPUX*/    sp77sprintf( eo670MessageBuffer,
/*HPUX*/                 sizeof(eo670MessageBuffer),
/*HPUX*/                 "sr2 sr3 0x%016lx 0x%016lx\n",
/*HPUX*/                 context->uc_mcontext.ss_wide.ss_64.ss_sr2,
/*HPUX*/                 context->uc_mcontext.ss_wide.ss_64.ss_sr3);
/*HPUX*/    eo670WriteString(eo670MessageBuffer);
/*HPUX*/
/*HPUX*/    sp77sprintf( eo670MessageBuffer,
/*HPUX*/                 sizeof(eo670MessageBuffer),
/*HPUX*/                 "sr4 sr5 0x%016lx 0x%016lx\n",
/*HPUX*/                 context->uc_mcontext.ss_wide.ss_64.ss_sr4,
/*HPUX*/                 context->uc_mcontext.ss_wide.ss_64.ss_sr5);
/*HPUX*/    eo670WriteString(eo670MessageBuffer);
/*HPUX*/
/*HPUX*/    sp77sprintf( eo670MessageBuffer,
/*HPUX*/                 sizeof(eo670MessageBuffer),
/*HPUX*/                 "sr6 sr7 0x%016lx 0x%016lx\n",
/*HPUX*/                 context->uc_mcontext.ss_wide.ss_64.ss_sr6,
/*HPUX*/                 context->uc_mcontext.ss_wide.ss_64.ss_sr7);
/*HPUX*/    eo670WriteString(eo670MessageBuffer);
/*HPUX*/
/*HPUX*/    eo670WriteString( "-------------------------------------\n" );
/*HPUX*/
/*HPUX*/        for ( iReg = 0; iReg < (sizeof(fp_dbl_block_t)/sizeof(double)); iReg++ )
/*HPUX*/        {
/*HPUX*/            sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*HPUX*/                     "f%02d      [0x%016lx]%.20e\n",
/*HPUX*/                      iReg,
/*HPUX*/                      *(SAPDB_UInt8 *)&((double *)&context->uc_mcontext.ss_fpblock.fpdbl)[iReg],
/*HPUX*/                      ((double *)&context->uc_mcontext.ss_fpblock.fpdbl)[iReg] );
/*HPUX*/            eo670WriteString( eo670MessageBuffer );
/*HPUX*/        }
/*HPUX*/
/*HPUX*/    eo670WriteString( "-------------------------------------\n" );
/*HPUX*/
#    endif /* BIT64 */
/*HPUX*/}
#    ifdef BIT64
/*HPUX*/typedef unsigned long general_reg;
/*HUPX*/
/*HPUX*/typedef struct {
/*HPUX*/    unsigned long unwind_descriptor_bits;
/*HPUX*/    unsigned int region_start_address;
/*HPUX*/    unsigned int region_end_address;
/*HPUX*/} uw_rec_def;
/*HUPX*/
/*HPUX*/typedef struct {
/*HPUX*/    unsigned long size;
/*HPUX*/    general_reg   sp;
/*HPUX*/    general_reg   return_link_offset;
/*HPUX*/    general_reg   gp;
/*HPUX*/    general_reg   rp;
/*HPUX*/    general_reg   mrp;
/*HPUX*/    general_reg   r3;
/*HPUX*/    general_reg   r4;
/*HPUX*/    unsigned long reserved[4];
/*HPUX*/} curr_frame_info;
/*HUPX*/
/*HPUX*/typedef struct {
/*HPUX*/    unsigned long size;
/*HPUX*/    general_reg   sp;
/*HPUX*/    general_reg   return_link_offset;
/*HPUX*/    general_reg   gp;
/*HPUX*/    uw_rec_def    uw_rec;
/*HPUX*/    long          uw_index;
/*HPUX*/    general_reg   r3;
/*HPUX*/    general_reg   r4;
/*HPUX*/    unsigned long reserved[4];
/*HPUX*/} prev_frame_info;
/*HUPX*/
/*HPUX*/extern curr_frame_info U_get_current_frame();
/*HPUX*/extern int             U_get_previous_frame(curr_frame_info *pCurr, prev_frame_info *pPrev);
/*HPUX*/extern void            U_copy_frame_info(curr_frame_info *pCurr, prev_frame_info *pPrev);
/*HUPX*/
/*HPUX64*/ /*----------------------------------------------------------------------------------*/
/*HPUX64*/
/*HPUX64*/SAPDB_Int4 eo670_FastFillCallArray(SAPDB_Int4 skipCount, SAPDB_Int4 level, void **callArray)
/*HPUX64*/{
/*HPUX64*/    curr_frame_info   context;
/*HPUX64*/    prev_frame_info   prevContext;
/*HPUX64*/    SAPDB_Int4        depth;
/*HPUX64*/
/*HPUX64*/    depth = 0;
/*HPUX64*/    context = U_get_current_frame();
/*HPUX64*/
/*HPUX64*/    while ( depth < level )
/*HPUX64*/    {
/*HPUX64*/        U_get_previous_frame(&context, &prevContext);
/*HPUX64*/        if ( !prevContext.return_link_offset )
/*HPUX64*/        {
/*HPUX64*/            break;
/*HPUX64*/        }
/*HPUX64*/
/*HPUX64*/        U_copy_frame_info(&context, &prevContext);
/*HPUX64*/        if ( skipCount > 0 )
/*HPUX64*/        {
/*HPUX64*/            --skipCount;
/*HPUX64*/        }
/*HPUX64*/        else
/*HPUX64*/        {
/*HPUX64*/	          callArray[depth] = (void *)(((long)context.return_link_offset)&(~3));
/*HPUX64*/            ++depth;
/*HPUX64*/        }
/*HPUX64*/    }
/*HPUX64*/    return depth;
/*HPUX64*/}
/*HPUX64*/
/*HPUX*/SAPDB_Int4 eo670_UnixFillCallArray(SAPDB_Int4 skipCount, SAPDB_Int4 level, eo670_CallArrayType addCallInfo, void **callArray, void *pContext)
/*HPUX*/{
/*HPUX*/    static curr_frame_info   context;
/*HPUX*/    static prev_frame_info   prevContext;
/*HPUX*/    static SAPDB_Int4        depth;
/*HPUX*/
/*HPUX*/    depth = 0;
/*HPUX*/    if ( pContext )
/*HPUX*/    {
/*HPUX*/        ucontext_t *sigContext = (ucontext_t *)pContext;
/*HPUX*/        context = U_get_current_frame();
/*HPUX*/        switch ( addCallInfo )
/*HPUX*/        {
/*HPUX*/        case eo670_FullInfoArray:
/*HPUX*/	        callArray[depth] = (void *)((char *)0+(((long)sigContext->uc_mcontext.ss_wide.ss_64.ss_pcoq_head)&(~3)));
/*HPUX*/            break;
/*HPUX*/        case eo670_PointerArray:
/*HPUX*/	        callArray[depth] = (void *)((char *)0+(((long)sigContext->uc_mcontext.ss_wide.ss_64.ss_pcoq_head)&(~3)));
/*HPUX*/            break;
/*HPUX*/        default: /* eo670_NoArrayJustOutput */
/*HPUX*/            sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*HPUX*/                         "(%d):0x%lx\n",
/*HPUX*/                         depth, 
/*HPUX*/	                     (((long)sigContext->uc_mcontext.ss_wide.ss_64.ss_pcoq_head)&(~3)) );
/*HPUX*/            eo670WriteString(eo670MessageBuffer);
/*HPUX*/            break;
/*HPUX*/        }
/*HPUX*/        ++depth;
/*HPUX*/        for ( skipCount = 256; skipCount > 0; --skipCount )
/*HPUX*/        {
/*HPUX*/            U_get_previous_frame(&context, &prevContext);
/*HPUX*/            if ( !prevContext.return_link_offset )
/*HPUX*/            {
/*HPUX*/               return depth;
/*HPUX*/            }
/*HPUX*/
/*HPUX*/            U_copy_frame_info(&context, &prevContext);
/*HPUX*/	        if ( ((void *)(((long)context.return_link_offset)&(~3))) == callArray[0] )
/*HPUX*/            {
/*HPUX*/                 break;
/*HPUX*/            }
/*HPUX*/        }
/*HPUX*/        skipCount = 0;
/*HPUX*/    }
/*HPUX*/    else
/*HPUX*/    {
/*HPUX*/        context = U_get_current_frame();
/*HPUX*/    }
/*HPUX*/
/*HPUX*/    while ( depth < level )
/*HPUX*/    {
/*HPUX*/        U_get_previous_frame(&context, &prevContext);
/*HPUX*/        if ( !prevContext.return_link_offset )
/*HPUX*/        {
/*HPUX*/            break;
/*HPUX*/        }
/*HPUX*/
/*HPUX*/        U_copy_frame_info(&context, &prevContext);
/*HPUX*/        if ( skipCount > 0 )
/*HPUX*/        {
/*HPUX*/            --skipCount;
/*HPUX*/        }
/*HPUX*/        else
/*HPUX*/        {
/*HPUX*/            switch ( addCallInfo )
/*HPUX*/            {
/*HPUX*/            case eo670_FullInfoArray:
/*HPUX*/	            callArray[depth] = (void *)(((long)context.return_link_offset)&(~3));
/*HPUX*/                break;
/*HPUX*/            case eo670_PointerArray:
/*HPUX*/	            callArray[depth] = (void *)(((long)context.return_link_offset)&(~3));
/*HPUX*/                break;
/*HPUX*/            default: /* eo670_NoArrayJustOutput */
/*HPUX*/                sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*HPUX*/                             "(%d):0x%lx\n",
/*HPUX*/                             depth, 
/*HPUX*/	                         (((long)context.return_link_offset)&(~3)) );
/*HPUX*/                eo670WriteString(eo670MessageBuffer);
/*HPUX*/                break;
/*HPUX*/            }
/*HPUX*/            ++depth;
/*HPUX*/        }
/*HPUX*/    }
/*HPUX*/    return depth;
/*HPUX*/}
#    else
/*HPUX32*/ /*----------------------------------------------------------------------------------*/
/*HPUX32*/
/*HPUX32*/typedef struct Hp32CurrentFrame_ {
/*HPUX32*/    unsigned curr_frame_size;
/*HPUX32*/    unsigned curr_sp;
/*HPUX32*/    unsigned curr_pcspace;
/*HPUX32*/    unsigned curr_pcoffset;
/*HPUX32*/    unsigned curr_dp;
/*HPUX32*/    unsigned curr_rp;
/*HPUX32*/    unsigned curr_mrp;
/*HPUX32*/    unsigned curr_sr0;
/*HPUX32*/    unsigned curr_sr4;
/*HPUX32*/    unsigned r3;
/*HPUX32*/    unsigned curr_r19;
/*HPUX32*/    int r4;
/*HPUX32*/    int reserved;
/*HPUX32*/} Hp32CurrentFrame;
/*HPUX32*/
/*HPUX32*/typedef struct Hp32PreviousFrame_ {
/*HPUX32*/    unsigned prev_frame_size;
/*HPUX32*/    unsigned prev_sp;
/*HPUX32*/    unsigned prev_pcspace;
/*HPUX32*/    unsigned prev_pcoffset;
/*HPUX32*/    unsigned prev_dp;
/*HPUX32*/    unsigned udescr0;
/*HPUX32*/    unsigned udescr1;
/*HPUX32*/    unsigned ustart;
/*HPUX32*/    unsigned uend;
/*HPUX32*/    unsigned uw_index;
/*HPUX32*/    unsigned prev_r19;
/*HPUX32*/    int r3;
/*HPUX32*/    int r4;
/*HPUX32*/} Hp32PreviousFrame;
/*HPUX32*/      /* defined in libcl */
/*HPUX32*/extern void   U_get_frame_info(Hp32CurrentFrame *);
/*HPUX32*/extern int    U_get_previous_frame_x(Hp32CurrentFrame *, Hp32PreviousFrame *, int);
/*HPUX32*/
/*HPUX32*/static void eo670_Hp32copy_prev_to_curr(Hp32CurrentFrame *currFrame, Hp32PreviousFrame *prevFrame)
/*HPUX32*/{
/*HPUX32*/    currFrame->curr_frame_size = prevFrame->prev_frame_size;
/*HPUX32*/    currFrame->curr_sp         = prevFrame->prev_sp;
/*HPUX32*/    currFrame->curr_pcspace    = prevFrame->prev_pcspace;
/*HPUX32*/    currFrame->curr_pcoffset   = prevFrame->prev_pcoffset;
/*HPUX32*/    currFrame->curr_dp         = prevFrame->prev_dp;
/*HPUX32*/}
/*HPUX32*/
/*HPUX32*/ /*----------------------------------------------------------------------------------*/
/*HPUX32*/
/*HPUX32*/SAPDB_Int4 eo670_FastFillCallArray(SAPDB_Int4 skipCount, SAPDB_Int4 level, void **callArray)
/*HPUX32*/{
/*HPUX32*/    Hp32CurrentFrame  currFrame;
/*HPUX32*/    Hp32PreviousFrame prevFrame;
/*HPUX32*/    SAPDB_Int4        depth;
/*HPUX32*/    int status;
/*HPUX32*/
/*HPUX32*/    depth = 0;
/*HPUX32*/        U_get_frame_info(&currFrame);
/*HPUX32*/        currFrame.curr_sr0 = 0;
/*HPUX32*/        currFrame.curr_sr4 = 0;
/*HPUX32*/
/*HPUX32*/    while ( depth < level )
/*HPUX32*/    {
/*HPUX32*/        status = U_get_previous_frame_x(&currFrame, &prevFrame, sizeof(Hp32CurrentFrame));
/*HPUX32*/        eo670_Hp32copy_prev_to_curr(&currFrame, &prevFrame);
/*HPUX32*/        if ( !currFrame.curr_pcoffset )
/*HPUX32*/        {
/*HPUX32*/            break;
/*HPUX32*/        }
/*HPUX32*/
/*HPUX32*/        if ( status != 0 )
/*HPUX32*/        {
/*HPUX32*/            sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*HPUX32*/                         "Failed to get previous frame:%d\n", status);
/*HPUX32*/            eo670WriteString(eo670MessageBuffer);
/*HPUX32*/            break;
/*HPUX32*/        }
/*HPUX32*/
/*HPUX32*/        if ( skipCount > 0 )
/*HPUX32*/        {
/*HPUX32*/            --skipCount;
/*HPUX32*/        }
/*HPUX32*/        else
/*HPUX32*/        {
/*HPUX32*/	          callArray[depth] = (void *)(((long)currFrame.curr_pcoffset)&(~3));
/*HPUX32*/            ++depth;
/*HPUX32*/        }
/*HPUX32*/    }
/*HPUX32*/    return depth;
/*HPUX32*/}
/*HPUX32*/
/*HPUX*/ /* not yet supported */
/*HPUX*/SAPDB_Int4 eo670_UnixFillCallArray(SAPDB_Int4 skipCount, SAPDB_Int4 level, eo670_CallArrayType addCallInfo, void **callArray, void *pContext)
/*HPUX*/{
/*HPUX*/    return 0;
/*HPUX*/}
#    endif /* BIT64 */
/*HPUX*/
/*HPUX*/static void eo670_UpdateModuleChain(void)
/*HPUX*/{
/*HPUX*/    static int dlIndex;
/*HPUX*/    static int retCode;
/*HPUX*/    static struct shl_descriptor *dlDesc;
/*HPUX*/
/*HPUX*/    for ( dlIndex = 0; 0 == (retCode = shl_get(dlIndex, &dlDesc)); ++dlIndex )
/*HPUX*/    {
/*HPUX*/        /* eo670_AddToLoadedModuleChain would report errors,     */
/*HPUX*/        /* so void is ok since no local stoarge needed.          */
/*HPUX*/        /* The buffer returned is 'static' and content           */
/*HPUX*/        /* is copied in the called eo670_AddToLoadedModuleChain. */
/*HPUX*/        (void)eo670_AddToLoadedModuleChain( dlDesc->filename,
/*HPUX*/                                            (char *)0 + dlDesc->tstart,
/*HPUX*/                                            (char *)0 + dlDesc->tend,
/*HPUX*/                                            (char *)0 + dlIndex );
/*HPUX*/    }
/*HPUX*/}
/*HPUX*/
/*HPUX*/ /*----------------------------------------------------------------------------------*/
#    ifdef BIT64
/*HPUX*/
/*HPUX*/ /* on 64-bit HP/UX the symbol pointer value is a pointer to a Procedure Descriptor */
/*HPUX*/typedef struct {
/*HPUX*/       unsigned long Reserved0;
/*HPUX*/       unsigned long Reserved1;
/*HPUX*/       void *Address;
/*HPUX*/       void *GlobalPointer;
/*HPUX*/} eo670_ProcedureDescriptor_t;
/*HPUX*/
/*HPUX*/static void *eo670_getFuncAddr (void *pFunc)
/*HPUX*/{
/*HPUX*/	eo670_ProcedureDescriptor_t *pPD;
/*HPUX*/	pPD = (eo670_ProcedureDescriptor_t *)pFunc;
/*HPUX*/	return ( pPD ? pPD->Address : (void *)0 );
/*HPUX*/}
#    else
/*HPUX*/
/*HPUX*/static void *eo670_getFuncAddr (void *FuncPtr)
/*HPUX*/{
/*HPUX*/ /* warning: it is assumed that sizeof (unsigned long) == sizeof (void *) */
/*HPUX*/	unsigned long ulFuncPtr;
/*HPUX*/ /* a function pointer on 32-bit HP/UX may be the address of */
/*HPUX*/ /* the pointer to an entry in the Procedure Label Table */
/*HPUX*/    if ( !FuncPtr ) return (void *)0;
/*HPUX*/	ulFuncPtr = (char *)FuncPtr-(char *)0;
/*HPUX*/
/*HPUX*/	/* nothing to do if this is a real function address */	
/*HPUX*/	if ( ulFuncPtr & 2 == 0 )
/*HPUX*/    {
/*HPUX*/        return FuncPtr;
/*HPUX*/    }
/*HPUX*/
/*HPUX*/	/* look up real address in Procedure Label Table (PLT) */
/*HPUX*/	ulFuncPtr &= ~3;
/*HPUX*/	return *(void **)((char *)0+ulFuncPtr);
/*HPUX*/}
#    endif /* BIT64 */
/*HPUX*/
/*HPUX*/struct symbolInfo
/*HPUX*/{
/*HPUX*/    struct symbolInfo *next;
/*HPUX*/    char *name;
/*HPUX*/    char *value;
/*HPUX*/};
/*HPUX*/static struct symbolInfo *eo670_SymbolList = 0;
/*HPUX*/
/*HPUX*/static struct symbolInfo* eo670_addSymbol(char *name, void *value)
/*HPUX*/{
/*HPUX*/    static struct symbolInfo *nextFreeElement = 0;
/*HPUX*/    static int freeSizeInChunk = 0;
/*HPUX*/    static void *lastSymbolChunk = 0;
/*HPUX*/    static struct symbolInfo **symbolSearch = 0;
/*HPUX*/    static int entrySize;
/*HPUX*/    static void *funcAddr;
/*HPUX*/
/*HPUX*/    funcAddr = eo670_getFuncAddr(value);
/*HPUX*/    /* aligned entrysize */
/*HPUX*/    entrySize = (sizeof(struct symbolInfo) + strlen(name) + sizeof(void *))&~(sizeof(void *)-1);
#    define SYMBOL_CHUNK_SIZE (64*1024)
/*HPUX*/    if ( entrySize > SYMBOL_CHUNK_SIZE )
/*HPUX*/    {
/*HPUX*/        eo670WriteString("Rejecting giant symbol: ");
/*HPUX*/        eo670WriteString(name);
/*HPUX*/        eo670WriteString("\n");
/*HPUX*/        return (struct symbolInfo *)0;
/*HPUX*/    }
/*HPUX*/
/*HPUX*/    symbolSearch = &eo670_SymbolList;
/*HPUX*/    while ( *symbolSearch )
/*HPUX*/    {
/*HPUX*/        if ( (*symbolSearch)->value < (char *)funcAddr )
/*HPUX*/        {
/*HPUX*/            symbolSearch = &((*symbolSearch)->next);
/*HPUX*/        }
/*HPUX*/        else if ( (*symbolSearch)->value == (char *)funcAddr )
/*HPUX*/        {
/*HPUX*/            return (*symbolSearch);
/*HPUX*/        }
/*HPUX*/        else
/*HPUX*/        {
/*HPUX*/            break;
/*HPUX*/        }
/*HPUX*/    }
/*HPUX*/
/*HPUX*/    if ( freeSizeInChunk < entrySize )
/*HPUX*/    {
/*HPUX*/        freeSizeInChunk = 0; /* discard rest of chunk */
/*HPUX*/        nextFreeElement = (struct symbolInfo *)malloc(SYMBOL_CHUNK_SIZE);
/*HPUX*/        if ( !nextFreeElement )
/*HPUX*/        {
/*HPUX*/            return nextFreeElement; /* not allocated... */
/*HPUX*/        }
/*HPUX*/        freeSizeInChunk = SYMBOL_CHUNK_SIZE;
/*HPUX*/    }
/*HPUX*/
/*HPUX*/    nextFreeElement->next = *symbolSearch;
/*HPUX*/    nextFreeElement->name = (char *)(nextFreeElement+1);
/*HPUX*/    strcpy(nextFreeElement->name, name);
/*HPUX*/    nextFreeElement->value = (char *)funcAddr;
/*HPUX*/    *symbolSearch = nextFreeElement;
/*HPUX*/    freeSizeInChunk -= entrySize;
/*HPUX*/    nextFreeElement = (struct symbolInfo *)(((char *)nextFreeElement)+entrySize);
/*HPUX*/    return (*symbolSearch);
/*HPUX*/}
/*HPUX*/
/*HPUX*/static void eo670_FillAddressInfo(char *pcAddr, eo670AddressInfo *addrInfo)
/*HPUX*/{
/*HPUX*/   static eo670_ModuleChain *modulePointer;
/*HPUX*/   static struct shl_descriptor *dlDesc;
/*HPUX*/   static int dlIndex;
/*HPUX*/   static int retCode;
/*HPUX*/   static struct shl_symbol *symbolArray;
/*HPUX*/   static int symbolCount;
/*HPUX*/   static char *bestSymbol;
/*HPUX*/   static unsigned long bestOffset;
/*HPUX*/   static int sIndex;
/*HPUX*/
/*HPUX*/   memset(addrInfo, 0, sizeof(eo670AddressInfo));
/*HPUX*/   addrInfo->sourceFileName = ""; /* No Source information on HPUX... */
/*HPUX*/   addrInfo->lineNumber = -1;
/*HPUX*/
/*HPUX*/   for ( modulePointer = eo670_ModuleAnchor
/*HPUX*/       ; modulePointer
/*HPUX*/       ; modulePointer = modulePointer->next )
/*HPUX*/   {
/*HPUX*/       if (  ( pcAddr >= ((char *)modulePointer->text_mapaddr) )
/*HPUX*/          && ( pcAddr <  ((char *)modulePointer->text_endaddr) ) )
/*HPUX*/       {
/*HPUX*/           dlIndex = ((char *)modulePointer->osDependend) - ((char *)0);
/*HPUX*/           break;
/*HPUX*/   	   }
/*HPUX*/   }
/*HPUX*/
/*HPUX*/   if ( !modulePointer )
/*HPUX*/   {
/*HPUX*/       addrInfo->moduleName = "<not in any loaded module>";
/*HPUX*/       addrInfo->symbol = "<not in any loaded region>";
/*HPUX*/       addrInfo->demangledSymbol = "";
/*HPUX*/       addrInfo->offsetToSymbol = -1;
/*HPUX*/       return;
/*HPUX*/   }
/*HPUX*/
/*HPUX*/   addrInfo->moduleName = modulePointer->file_name;
/*HPUX*/
/*HPUX*/   if ( symbolArray )
/*HPUX*/   {
/*HPUX*/       free(symbolArray);
/*HPUX*/   }
/*HPUX*/   symbolArray = 0;
/*HPUX*/
/*HPUX*/   if ( dlIndex != -1 ) /* symbol table already loaded ? */
/*HPUX*/   {
/*HPUX*/       retCode = shl_get(dlIndex, &dlDesc);
/*HPUX*/       if ( retCode != 0 )
/*HPUX*/       {
/*HPUX*/           eo670WriteString("Module index no longer valid\n");
/*HPUX*/           addrInfo->symbol = "<module description no longer available>";
/*HPUX*/           addrInfo->demangledSymbol = "";
/*HPUX*/           addrInfo->offsetToSymbol = -1;
/*HPUX*/           return;
/*HPUX*/       }
/*HPUX*/
/*HPUX*/       symbolCount = shl_getsymbols( dlDesc->handle,
/*HPUX*/                                     TYPE_PROCEDURE,
/*HPUX*/                                     EXPORT_SYMBOLS,
/*HPUX*/                                     malloc,
/*HPUX*/                                    &symbolArray );
/*HPUX*/       if ( symbolCount < 0 )
/*HPUX*/       {
/*HPUX*/           eo670WriteString("Failed to get symbols\n");
/*HPUX*/           addrInfo->symbol = "<no symbols for this region>";
/*HPUX*/           addrInfo->demangledSymbol = "";
/*HPUX*/           addrInfo->offsetToSymbol = -1;
/*HPUX*/           return;
/*HPUX*/       }
/*HPUX*/
/*HPUX*/       for ( sIndex = 0; sIndex < symbolCount; sIndex++ )
/*HPUX*/       {
/*HPUX*/           if ( !eo670_addSymbol(symbolArray[sIndex].name, symbolArray[sIndex].value) )
/*HPUX*/           {
/*HPUX*/               sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*HPUX*/                            "Out of memory while adding symbols from %s (%d:%s)\n", 
/*HPUX*/                             modulePointer->file_name, sIndex, symbolArray[sIndex].name );
/*HPUX*/               eo670WriteString(eo670MessageBuffer);
/*HPUX*/               break;
/*HPUX*/           }
/*HPUX*/       }
/*HPUX*/       if ( sIndex < symbolCount )
/*HPUX*/       {
/*HPUX*/           /* failed to add symbol array, so search in symbolArray... */
/*HPUX*/       }
/*HPUX*/       else
/*HPUX*/       {
/*HPUX*/           free(symbolArray);
/*HPUX*/           symbolArray = 0;
/*HPUX*/           modulePointer->osDependend = (char *)0 - 1; /* mark symbol table read */
/*HPUX*/       }
/*HPUX*/   }
/*HPUX*/
/*HPUX*/   if ( symbolArray )
/*HPUX*/   {
/*HPUX*/       bestSymbol = 0;
/*HPUX*/       bestOffset = pcAddr - ((char *)modulePointer->text_mapaddr);
/*HPUX*/       for ( sIndex = 0; sIndex < symbolCount; sIndex++ )
/*HPUX*/       {
/*HPUX*/           if ( ((char *)symbolArray[sIndex].value) > pcAddr )
/*HPUX*/           {
/*HPUX*/               continue;
/*HPUX*/           }
/*HPUX*/           else if ( ((char *)symbolArray[sIndex].value) == pcAddr )
/*HPUX*/           {
/*HPUX*/               bestSymbol = symbolArray[sIndex].name;
/*HPUX*/               bestOffset = 0;
/*HPUX*/               break;
/*HPUX*/           }
/*HPUX*/           else
/*HPUX*/           {
/*HPUX*/               if ( (pcAddr - ((char *)symbolArray[sIndex].value)) < bestOffset )
/*HPUX*/               {
/*HPUX*/                   bestSymbol = symbolArray[sIndex].name;
/*HPUX*/                   bestOffset = pcAddr - ((char *)symbolArray[sIndex].value);
/*HPUX*/               }
/*HPUX*/           }
/*HPUX*/       }
/*HPUX*/   }
/*HPUX*/   else
/*HPUX*/   {
/*HPUX*/       static struct symbolInfo *pSearch;
/*HPUX*/
/*HPUX*/       bestSymbol = 0;
/*HPUX*/       bestOffset = pcAddr - ((char *)modulePointer->text_mapaddr);
/*HPUX*/       for ( pSearch = eo670_SymbolList; 
/*HPUX*/             pSearch && pSearch->value <= pcAddr;
/*HPUX*/             pSearch = pSearch->next )
/*HPUX*/       {
/*HPUX*/           bestSymbol = pSearch->name;
/*HPUX*/           bestOffset = pcAddr - pSearch->value;
/*HPUX*/       }
/*HPUX*/   }
/*HPUX*/
/*HPUX*/   if ( bestSymbol )
/*HPUX*/   {
/*HPUX*/       addrInfo->symbol = bestSymbol;
/*HPUX*/       addrInfo->offsetToSymbol = bestOffset;
#    if defined(WANT_DEMANGLING_SUPPORT)
/*HPUX*/       addrInfo->demangledSymbol = "";
#    endif /* WANT_DEMANGLING_SUPPORT */
/*HPUX*/   }
/*HPUX*/   else
/*HPUX*/   {
/*HPUX*/       addrInfo->symbol = "<symbol lookup failed>";
/*HPUX*/       addrInfo->demangledSymbol = "";
/*HPUX*/       addrInfo->offsetToSymbol = -1;
/*HPUX*/       addrInfo->lineNumber = -1;
/*HPUX*/   }
/*HPUX*/    return;
/*HPUX*/}
/*HPUX*/
/*HPUX*/static SAPDB_Bool eo670_UnixGetAddressInfo(void *pcAddr, eo670AddressInfo *addrInfo)
/*HPUX*/{
/*HPUX*/    static eo670AddressInfo localAddrInfo;
/*HPUX*/    static SAPDB_Bool fillResult;
/*HPUX*/
/*HPUX*/    eo670_FillAddressInfo((char *)pcAddr, &localAddrInfo);
/*HPUX*/    addrInfo->symbol = (char *)calloc( 1, strlen(localAddrInfo.symbol)+1
/*HPUX*/                                 + strlen(localAddrInfo.demangledSymbol)+1
/*HPUX*/                                 + strlen(localAddrInfo.moduleName)+1
/*HPUX*/                                 + strlen(localAddrInfo.sourceFileName)+1 );
/*HPUX*/    if ( !addrInfo->symbol )
/*HPUX*/    {
/*HPUX*/        eo670WriteString("Out of memory for symbol copy\n");
/*HPUX*/        fillResult = 0;
/*HPUX*/    }
/*HPUX*/    else
/*HPUX*/    {
/*HPUX*/        strcpy(addrInfo->symbol, localAddrInfo.symbol);
/*HPUX*/        addrInfo->moduleName = addrInfo->symbol + strlen(localAddrInfo.symbol)+1;
/*HPUX*/        addrInfo->demangledSymbol = addrInfo->symbol + strlen(localAddrInfo.symbol)+1;
/*HPUX*/        strcpy(addrInfo->demangledSymbol, localAddrInfo.demangledSymbol);
/*HPUX*/        addrInfo->moduleName = addrInfo->demangledSymbol + strlen(localAddrInfo.demangledSymbol)+1;
/*HPUX*/        strcpy(addrInfo->moduleName, localAddrInfo.moduleName);
/*HPUX*/        addrInfo->sourceFileName = addrInfo->moduleName + strlen(localAddrInfo.moduleName)+1;
/*HPUX*/        strcpy(addrInfo->sourceFileName, localAddrInfo.sourceFileName);
/*HPUX*/        addrInfo->offsetToSymbol = localAddrInfo.offsetToSymbol;
/*HPUX*/        addrInfo->lineNumber = localAddrInfo.lineNumber;
/*HPUX*/        fillResult = 1;
/*HPUX*/    }
/*HPUX*/    return fillResult;
/*HPUX*/}
/*HPUX*/
/*HPUX*/static void eo670_OsFreeAddressInfo(eo670AddressInfo *addrInfo)
/*HPUX*/{
/*HPUX*/    free(addrInfo->symbol); /* allocated in one chunk.... */
/*HPUX*/}
/*HPUX*/
#  endif /*HPUX*/

#  if defined(OSF1)
/*OSF1*/struct modul_chain_elem
/*OSF1*/{
/*OSF1*/    ldr_module_info_t	      moduleInfo;
/*OSF1*/    ldr_region_info_t	      textInfo;
/*OSF1*/    elf32_LDFILE	         *ldptr;
/*OSF1*/    int			opened;		/* opened or at at least tried to open */
/*OSF1*/    int			symtab_ok;
/*OSF1*/};
/*OSF1*/typedef struct modul_chain_elem mce;
/*OSF1*/typedef	struct modul_chain_elem	*mce_p;
/*OSF1*/
/*OSF1*/static long ipd_adr	  (LDFILE *ldptr, unsigned long adr, PDR *pApd);
/*OSF1*/extern int ld_ifd_symnum(LDFILE *ldptr, int);
/*OSF1*/extern char * st_str_ifd_iss(int, int);
/*OSF1*/
/*OSF1*/static void eo670_FillAddressInfo(char *pcAddr, eo670AddressInfo *addrInfo)
/*OSF1*/{
/*OSF1*/  static mce_p	       me_p;
/*OSF1*/  static PDR	       apd;
/*OSF1*/  static SYMR	       asym;
/*OSF1*/  static char	      *pname;
/*OSF1*/  static long	       isym;
/*OSF1*/  static int	       ifd;
/*OSF1*/  static eo670_ModuleChain *modulePointer;
/*OSF1*/
/*OSF1*/    memset(addrInfo, 0, sizeof(eo670AddressInfo));
/*OSF1*/
/*OSF1*/    for ( modulePointer = eo670_ModuleAnchor, me_p = 0
/*OSF1*/        ; modulePointer
/*OSF1*/        ; modulePointer = modulePointer->next )
/*OSF1*/    {
/*OSF1*/        if (  ( pcAddr >= ((char *)modulePointer->text_mapaddr) )
/*OSF1*/           && ( pcAddr <  ((char *)modulePointer->text_endaddr) ) )
/*OSF1*/        {
/*OSF1*/            me_p = (mce_p)modulePointer->osDependend;
/*OSF1*/	        break;
/*OSF1*/   	    }
/*OSF1*/    }
/*OSF1*/
/*OSF1*/    addrInfo->demangledSymbol = ""; /* not supported by DEC */
/*OSF1*/    addrInfo->offsetToSymbol = -1;
/*OSF1*/    addrInfo->sourceFileName = "";
/*OSF1*/    addrInfo->lineNumber = -1;
/*OSF1*/
/*OSF1*/   	if ( modulePointer )
/*OSF1*/    {
/*OSF1*/        addrInfo->moduleName = modulePointer->file_name;
/*OSF1*/	    if ( me_p && me_p->opened == 0 )
/*OSF1*/	    {
/*OSF1*/    	    me_p->opened = 1;
/*OSF1*/	       	me_p->ldptr = ldopen(modulePointer->file_name, 0);
/*OSF1*/       		if ( me_p->ldptr == 0 )
/*OSF1*/	        {
/*OSF1*/        	    sp77sprintf ( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*OSF1*/        			        "*** FillAddressInfo: Cannot ldopen %s (errno: %d)\n",
/*OSF1*/			                modulePointer->file_name, errno );
/*OSF1*/                eo670WriteString(eo670MessageBuffer);
/*OSF1*/            }
/*OSF1*/        	else
/*OSF1*/	        {
/*OSF1*/       		    if (PSYMTAB( me_p->ldptr ) == 0)
/*OSF1*/        	    {
/*OSF1*/       		        sp77sprintf ( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*OSF1*/                                      "*** FillAddressInfo: No symbol table for module %s\n", 
/*OSF1*/                                      modulePointer->file_name );
/*OSF1*/                    eo670WriteString(eo670MessageBuffer);
/*OSF1*/                } /* PSYMTAB( me_p->ldptr ) == 0 */
/*OSF1*/      		    else
/*OSF1*/	            {
/*OSF1*/               		me_p->symtab_ok = 1;
/*OSF1*/	           	}
/*OSF1*/	   	    }
/*OSF1*/        } /* if me_p->opened == 0 */
/*OSF1*/
/*OSF1*/	    if ( me_p && me_p->symtab_ok )
/*OSF1*/	    {
/*OSF1*/    	    if ( ipd_adr(me_p->ldptr, (unsigned long)(pcAddr-(char *)0), &apd) == FAILURE )
/*OSF1*/            {
/*OSF1*/	   	        isym = isymNil;
/*OSF1*/            }
/*OSF1*/            else
/*OSF1*/            {
/*OSF1*/	   	        isym = apd.isym;
/*OSF1*/            }
/*OSF1*/
/*OSF1*/	        if (isym != isymNil)
/*OSF1*/    	    {
/*OSF1*/	   	        if (ldtbread(me_p->ldptr, isym, &asym) == FAILURE)
/*OSF1*/                {
/*OSF1*/	                addrInfo->symbol = "<symbol not available>";
/*OSF1*/                }
/*OSF1*/                else
/*OSF1*/                {
/*OSF1*/                    pname = ldgetname(me_p->ldptr, &asym);
/*OSF1*/                    if ( 0 == pname )
/*OSF1*/                    {
/*OSF1*/                        addrInfo->symbol = "<symbol not found>";
/*OSF1*/                    }
/*OSF1*/                    else
/*OSF1*/                    {
/*OSF1*/                         addrInfo->symbol = pname;
/*OSF1*/                         addrInfo->offsetToSymbol = (pcAddr - (char *)apd.adr);
/*OSF1*/   		                 ifd = ld_ifd_symnum ( me_p->ldptr, apd.isym );
/*OSF1*/    	                 if ( ifd != -1 )
/*OSF1*/	                     {
/*OSF1*/                             addrInfo->sourceFileName = st_str_ifd_iss(ifd, 1);
/*OSF1*/                             addrInfo->lineNumber =
/*OSF1*/                                  SYMTAB(me_p->ldptr)->pline[apd.iline
/*OSF1*/						             + (addrInfo->offsetToSymbol/sizeof(int))];
/*OSF1*/		                 }
/*OSF1*/                    }
/*OSF1*/		        }
/*OSF1*/		    } /* isym != isymNil */
/*OSF1*/            else
/*OSF1*/	   	    {
/*OSF1*/	            addrInfo->symbol = "<symbol lookup failed>";
/*OSF1*/    	    }
/*OSF1*/
/*OSF1*/	    } /* me_p->symtab_ok == TRUE */		
/*OSF1*/	    else
/*OSF1*/        {
/*OSF1*/            addrInfo->symbol = "<no symbols available for this region>";
/*OSF1*/        }
/*OSF1*/    }
/*OSF1*/    else
/*OSF1*/    {
/*OSF1*/        addrInfo->moduleName = "<not in any loaded module>";
/*OSF1*/        addrInfo->symbol = "<not in any loaded region>";
/*OSF1*/    }
/*OSF1*/}
/*OSF1*/
/*OSF1*/static void eo670_RegisterDump(void *pContext)
/*OSF1*/{
/*OSF1*/    static CONTEXT *context;
/*OSF1*/    static int iReg;
/*OSF1*/
/*OSF1*/    context = (CONTEXT *)pContext;
/*OSF1*/
/*OSF1*/    eo670WriteString( "-------------------------------------\n" );
/*OSF1*/
/*OSF1*/    for ( iReg = 0; iReg < (sizeof(context->sc_regs)/sizeof(long)/2); iReg++ )
/*OSF1*/    {
/*OSF1*/        sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*OSF1*/                     "r%02d  r%02d 0x%016lx 0x%016lx\n",
/*OSF1*/                      2*iReg,
/*OSF1*/                      1+2*iReg,
/*OSF1*/                      context->sc_regs[2*iReg],
/*OSF1*/                      context->sc_regs[1+2*iReg] );
/*OSF1*/        eo670WriteString( eo670MessageBuffer );
/*OSF1*/    }
/*OSF1*/    eo670WriteString("\n");
/*OSF1*/
/*OSF1*/    sp77sprintf( eo670MessageBuffer,
/*OSF1*/                 sizeof(eo670MessageBuffer),
/*OSF1*/                 "pc       0x%016lx\n",
/*OSF1*/                 context->sc_pc);
/*OSF1*/    eo670WriteString(eo670MessageBuffer);
/*OSF1*/
/*OSF1*/    sp77sprintf( eo670MessageBuffer,
/*OSF1*/                 sizeof(eo670MessageBuffer),
/*OSF1*/                 "ps       0x%016lx\n",
/*OSF1*/                 context->sc_ps);
/*OSF1*/    eo670WriteString(eo670MessageBuffer);
/*OSF1*/
/*OSF1*/    eo670WriteString( "-------------------------------------\n" );
/*OSF1*/
/*OSF1*/    if ( context->sc_ownedfp )
/*OSF1*/    {
/*OSF1*/        for ( iReg = 0; iReg < (sizeof(context->sc_fpregs)/sizeof(long)); iReg++ )
/*OSF1*/        {
/*OSF1*/            sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*OSF1*/                     "f%02d      [0x%016lx]%.20e\n",
/*OSF1*/                      iReg,
/*OSF1*/                      *(SAPDB_UInt8 *)&context->sc_fpregs[iReg],
/*OSF1*/                      *(double *)&context->sc_fpregs[iReg] );
/*OSF1*/            eo670WriteString( eo670MessageBuffer );
/*OSF1*/        }
/*OSF1*/        eo670WriteString("\n");
/*OSF1*/        sp77sprintf( eo670MessageBuffer,
/*OSF1*/                 sizeof(eo670MessageBuffer),
/*OSF1*/                 "fpcr     0x%016lx\n",
/*OSF1*/                 context->sc_fpcr);
/*OSF1*/        eo670WriteString(eo670MessageBuffer);
/*OSF1*/    }
/*OSF1*/    else
/*OSF1*/    {
/*OSF1*/        eo670WriteString( "Floating point unit unused\n");
/*OSF1*/    }
/*OSF1*/    eo670WriteString( "-------------------------------------\n" );
/*OSF1*/
/*OSF1*/}
/*OSF1*/
/*OSF1*/SAPDB_Int4 eo670_FastFillCallArray(SAPDB_Int4 skipCount, SAPDB_Int4 level, void **callArray)
/*OSF1*/{
/*OSF1*/    CONTEXT	         context;
/*OSF1*/    SAPDB_Int4       depth;     /* depth counter */
/*OSF1*/
/*OSF1*/    depth = 0;
/*OSF1*/
/*OSF1*/    exc_capture_context(&context);
/*OSF1*/
/*OSF1*/    do
/*OSF1*/    {
/*OSF1*/        PRUNTIME_FUNCTION prf;
/*OSF1*/
/*OSF1*/        if ( skipCount > 0 )
/*OSF1*/        {
/*OSF1*/            --skipCount;
/*OSF1*/	    }
/*OSF1*/	    else
/*OSF1*/        {
/*OSF1*/	        callArray[depth] = (void *)context.sc_pc;
/*OSF1*/            ++depth;
/*OSF1*/	    }
/*OSF1*/
/*OSF1*/	    prf = exc_lookup_function_entry (context.sc_pc);
/*OSF1*/        if ( 0 == prf ) /* pc did not point into code segment ... */
/*OSF1*/        {
/*OSF1*/            sp77sprintf( eo670MessageBuffer,
/*OSF1*/                         sizeof(eo670MessageBuffer),
/*OSF1*/                         "Damaged stack: exc_lookup_function_entry(0x%016p) failed\n",
/*OSF1*/                         context.sc_pc );
/*OSF1*/            eo670WriteString(eo670MessageBuffer);
/*OSF1*/            return depth;
/*OSF1*/	    }
/*OSF1*/	    exc_virtual_unwind (prf, &context); /* next frame */
/*OSF1*/
/*OSF1*/    } while ( context.sc_pc && (depth < level) );
#    define SKIP_BOT	(1)
/*OSF1*/    if ( depth >= SKIP_BOT ) depth -= SKIP_BOT; /* eliminate last entry */
/*OSF1*/    return depth;
/*OSF1*/}
/*OSF1*/
/*OSF1*/SAPDB_Int4 eo670_UnixFillCallArray(SAPDB_Int4 skipCount, SAPDB_Int4 level, eo670_CallArrayType addCallInfo, void **callArray, void *pContext)
/*OSF1*/{
/*OSF1*/    static CONTEXT	         context;
/*OSF1*/    static SAPDB_Int4        depth;     /* depth counter */
/*OSF1*/
/*OSF1*/    depth = 0;
/*OSF1*/
/*OSF1*/    if ( 0 == pContext )
/*OSF1*/    {
/*OSF1*/        exc_capture_context(&context);
/*OSF1*/    }
/*OSF1*/    else
/*OSF1*/    {
/*OSF1*/        memcpy(&context, pContext, sizeof(CONTEXT));
/*OSF1*/        skipCount = 0;
/*OSF1*/    }
/*OSF1*/
/*OSF1*/    do
/*OSF1*/    {
/*OSF1*/        static PRUNTIME_FUNCTION prf;
/*OSF1*/
/*OSF1*/        if ( skipCount > 0 )
/*OSF1*/        {
/*OSF1*/            --skipCount;
/*OSF1*/	    }
/*OSF1*/	    else
/*OSF1*/        {
/*OSF1*/            switch ( addCallInfo )
/*OSF1*/            {
/*OSF1*/            case eo670_FullInfoArray:
/*OSF1*/	            callArray[depth] = (void *)context.sc_pc;
/*OSF1*/                break;
/*OSF1*/            case eo670_PointerArray:
/*OSF1*/	            callArray[depth] = (void *)context.sc_pc;
/*OSF1*/                break;
/*OSF1*/            default: /* eo670_NoArrayJustOutput */
/*OSF1*/                sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*OSF1*/                             "(%d):0x%lx\n",
/*OSF1*/                             depth, context.sc_pc );
/*OSF1*/                eo670WriteString(eo670MessageBuffer);
/*OSF1*/                break;
/*OSF1*/            }
/*OSF1*/            ++depth;
/*OSF1*/	    }
/*OSF1*/
/*OSF1*/	    prf = exc_lookup_function_entry (context.sc_pc);
/*OSF1*/        if ( 0 == prf ) /* pc did not point into code segment ... */
/*OSF1*/        {
/*OSF1*/            sp77sprintf( eo670MessageBuffer,
/*OSF1*/                         sizeof(eo670MessageBuffer),
/*OSF1*/                         "Damaged stack: exc_lookup_function_entry(0x%016p) failed\n",
/*OSF1*/                         context.sc_pc );
/*OSF1*/            eo670WriteString(eo670MessageBuffer);
/*OSF1*/            return depth;
/*OSF1*/	    }
/*OSF1*/	    exc_virtual_unwind (prf, &context); /* next frame */
/*OSF1*/
/*OSF1*/    } while ( context.sc_pc && (depth < level) );
#    define SKIP_BOT	(1)
/*OSF1*/    if ( depth >= SKIP_BOT ) depth -= SKIP_BOT; /* eliminate last entry */
/*OSF1*/    return depth;
/*OSF1*/}
/*OSF1*/
/*OSF1*/static void eo670_UpdateModuleChain(void)
/*OSF1*/{
/*OSF1*/    static ldr_process_t my_proc;
/*OSF1*/    static ldr_module_t	 module_id;
/*OSF1*/    static mce           moduleChainElement;
/*OSF1*/    static size_t		 ret_size;
/*OSF1*/    static int	         regionIndex;
/*OSF1*/
/*OSF1*/    my_proc = ldr_my_process();
/*OSF1*/    module_id = LDR_NULL_MODULE;
/*OSF1*/
/*OSF1*/	for ( ;; )
/*OSF1*/    {
/*OSF1*/        static int	 rc;
/*OSF1*/        static mce_p me_p; /* ModuleChainElement pointer */
/*OSF1*/        static ldr_module_info_t * mInfo;
/*OSF1*/
/*OSF1*/        me_p = &moduleChainElement;
/*OSF1*/        mInfo = &me_p->moduleInfo;
/*OSF1*/        rc = ldr_next_module ( my_proc, &module_id );
/*OSF1*/        if ( rc )
/*OSF1*/    	{
/*OSF1*/	        sp77sprintf(eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*OSF1*/                        "ldr_next_module failed:%d\n", errno);
/*OSF1*/            eo670WriteString(eo670MessageBuffer);
/*OSF1*/    	    break;
/*OSF1*/	    }
/*OSF1*/
/*OSF1*/        /* end of module list found */
/*OSF1*/    	if ( module_id == LDR_NULL_MODULE)
/*OSF1*/	    {
/*OSF1*/    	    break;
/*OSF1*/	    }
/*OSF1*/
/*OSF1*/        /* inquire module information */
/*OSF1*/    	rc = ldr_inq_module ( my_proc,
/*OSF1*/                              module_id,
/*OSF1*/			                  mInfo,
/*OSF1*/                              sizeof(ldr_module_info_t),
/*OSF1*/			                 &ret_size );
/*OSF1*/	    if ( rc )
/*OSF1*/	    {
/*OSF1*/	        sp77sprintf(eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*OSF1*/                        "ldr_inq_module failed:%d\n", errno);
/*OSF1*/            eo670WriteString(eo670MessageBuffer);
/*OSF1*/    	    continue;
/*OSF1*/	    }
/*OSF1*/
/*OSF1*/        if ( mInfo->lmi_nregion > 0 )
/*OSF1*/        {
/*OSF1*/            static ldr_region_info_t *rInfo;
/*OSF1*/            rInfo = &me_p->textInfo;
/*OSF1*/
/*OSF1*/    	    for ( regionIndex = 0
/*OSF1*/	            ; regionIndex < mInfo->lmi_nregion
/*OSF1*/	            ; regionIndex++ )
/*OSF1*/	        {
/*OSF1*/	            rc = ldr_inq_region ( my_proc, module_id, regionIndex,
/*OSF1*/				                      rInfo, sizeof(ldr_region_info_t),
/*OSF1*/				                      &ret_size );
/*OSF1*/	            if ( rc )
/*OSF1*/	            {
/*OSF1*/                   sp77sprintf(eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*OSF1*/                               "ldr_inq_region failed for %s failed:%d\n",
/*OSF1*/                               mInfo->lmi_name, 
/*OSF1*/                               errno);
/*OSF1*/                   eo670WriteString(eo670MessageBuffer);
/*OSF1*/	    	       break;
/*OSF1*/	            }
/*OSF1*/                if ( 0 == memcmp(rInfo->lri_name, ".text", strlen(".text")+1) )
/*OSF1*/	            {
/*OSF1*/                   regionIndex = mInfo->lmi_nregion;
/*OSF1*/	            }
/*OSF1*/            }
/*OSF1*/
/*OSF1*/            if ( regionIndex < mInfo->lmi_nregion )
/*OSF1*/            {
/*OSF1*/	    	    break; /* no text region found */
/*OSF1*/	        }
/*OSF1*/
/*OSF1*/            /* Copy of entry needed */
/*OSF1*/    	    me_p = (mce_p)calloc ( 1, sizeof(mce) );
/*OSF1*/	        if ( 0 == me_p )
/*OSF1*/            {
/*OSF1*/                eo670WriteString("Out of memory for module list entry copy\n");
/*OSF1*/            }
/*OSF1*/            else
/*OSF1*/            {
/*OSF1*/                memcpy(me_p, &moduleChainElement, sizeof(moduleChainElement));
/*OSF1*/            }
/*OSF1*/
/*OSF1*/            switch ( eo670_AddToLoadedModuleChain( mInfo->lmi_name,
/*OSF1*/                                                   rInfo->lri_mapaddr,
/*OSF1*/                                                   ((char *)rInfo->lri_mapaddr) + rInfo->lri_size,
/*OSF1*/                                                   me_p) )
/*OSF1*/            {
/*OSF1*/            case 1: break;
/*OSF1*/            case 0: if ( me_p ) free(me_p);
/*OSF1*/                    me_p = 0;
/*OSF1*/                    break;
/*OSF1*/            case -1:if ( me_p ) free(me_p);
/*OSF1*/                    me_p = 0;
/*OSF1*/                    break;
/*OSF1*/	        }
/*OSF1*/        }
/*OSF1*/	} /* for(;;) break at end of module list */
/*OSF1*/}
/*OSF1*/
/*OSF1*/static SAPDB_Bool eo670_UnixGetAddressInfo(void *pcAddr, eo670AddressInfo *addrInfo)
/*OSF1*/{
/*OSF1*/  static eo670AddressInfo localAddrInfo;
/*OSF1*/  static SAPDB_Bool fillResult;
/*OSF1*/    eo670_FillAddressInfo((char *)pcAddr, &localAddrInfo);
/*OSF1*/    addrInfo->symbol = (char *)calloc( 1, strlen(localAddrInfo.symbol)+1
/*OSF1*/                                 + strlen(localAddrInfo.demangledSymbol)+1
/*OSF1*/                                 + strlen(localAddrInfo.moduleName)+1
/*OSF1*/                                 + strlen(localAddrInfo.sourceFileName)+1 );
/*OSF1*/    if ( !addrInfo->symbol )
/*OSF1*/    {
/*OSF1*/        eo670WriteString("Out of memory for symbol copy\n");
/*OSF1*/        fillResult = 0;
/*OSF1*/    }
/*OSF1*/    else
/*OSF1*/    {
/*OSF1*/        strcpy(addrInfo->symbol, localAddrInfo.symbol);
/*OSF1*/        addrInfo->moduleName = addrInfo->symbol + strlen(localAddrInfo.symbol)+1;
/*OSF1*/        addrInfo->demangledSymbol = addrInfo->symbol + strlen(localAddrInfo.symbol)+1;
/*OSF1*/        strcpy(addrInfo->demangledSymbol, localAddrInfo.demangledSymbol);
/*OSF1*/        addrInfo->moduleName = addrInfo->demangledSymbol + strlen(localAddrInfo.demangledSymbol)+1;
/*OSF1*/        strcpy(addrInfo->moduleName, localAddrInfo.moduleName);
/*OSF1*/        addrInfo->sourceFileName = addrInfo->moduleName + strlen(localAddrInfo.moduleName)+1;
/*OSF1*/        strcpy(addrInfo->sourceFileName, localAddrInfo.sourceFileName);
/*OSF1*/        addrInfo->offsetToSymbol = localAddrInfo.offsetToSymbol;
/*OSF1*/        addrInfo->lineNumber = localAddrInfo.lineNumber;
/*OSF1*/        fillResult = 1;
/*OSF1*/    }
/*OSF1*/    return fillResult;
/*OSF1*/}
/*OSF1*/
/*OSF1*/static void eo670_OsFreeAddressInfo(eo670AddressInfo *addrInfo)
/*OSF1*/{
/*OSF1*/    free(addrInfo->symbol); /* allocated in one chunk.... */
/*OSF1*/}
/*OSF1*/
/*OSF1*/static long ipd_adr (LDFILE *ldptr, unsigned long adr, PDR *pApd)
/*OSF1*/{
/*OSF1*/    static int ilow;
/*OSF1*/    static int ihigh;
/*OSF1*/    static int ibest;
/*OSF1*/    static unsigned long bestAdr;
/*OSF1*/
/*OSF1*/    ibest = -1;
/*OSF1*/    ihigh = SYMHEADER(ldptr).ipdMax-1;
/*OSF1*/    /* linear search proc table */
/*OSF1*/
/*OSF1*/    for ( ilow = 0; ilow <= ihigh; ++ilow )
/*OSF1*/    {
/*OSF1*/        if ( FAILURE == ldgetpd(ldptr, (long)ilow, pApd) )
/*OSF1*/        {
/*OSF1*/            return FAILURE;
/*OSF1*/        }
/*OSF1*/    	if ( (adr > pApd->adr)
/*OSF1*/          && ( (ibest == -1) || (bestAdr < pApd->adr) ) )
/*OSF1*/	    {
/*OSF1*/	        bestAdr = pApd->adr;
/*OSF1*/            ibest = ilow;
/*OSF1*/	    }
/*OSF1*/        if ( adr == pApd->adr )
/*OSF1*/        {
/*OSF1*/            return SUCCESS;
/*OSF1*/        }
/*OSF1*/    } /* while */
/*OSF1*/
/*OSF1*/    return ldgetpd(ldptr, (long)ibest, pApd);
/*OSF1*/}
#  endif /* OSF1 */

#  ifdef AIX
/*AIX*/
/**************************************************************************/
/* All the following code is required for using the StackTrace() function */
/* Each function has a word of 0000 after it's code followed by a pointer */
/* to a traceback table.  Contained within the traceback table is the name*/
/* of that function.  Due to the possiblity of very large functions or    */
/* functions without this information, there is a literal defined below   */
/* called MAX_FUNC_SEARCH_LEN.  This is the maximum number of words that  */
/* will be searched (following any function return address located on the */
/* stack) before the getname function decides to just return NONAME.      */
/* IMPORTANT NOTE:  This function also sets up a signal handler for       */
/* SIGSEGV.  This is to handle having this function run off the end of    */
/* the code segment when attempting to locate the tracetable and function */
/* name.  Please NOTE that YOU MUST call sigset() after this function to  */
/* re-install/set your signal handler for the signal SIGSEGV!.            */
/**************************************************************************/
/* ADDITIONAL NOTE:  When AIX prepares to call into a new function it will*/
/* have a small section of the stack set aside for parameters/etc.  It    */
/* will then execute a BL instruction which will save the iar (Instruction*/
/* pointer after execution of the BL instruction) into the link regiser.  */
/*  The function that receives control will then normally move the link   */
/* register into the r0 register and from there it will move the r0 info  */
/* into the stack area set aside from the calling process.  It will do    */
/* this at an offset of 2 32 bit words higher than the current stack ptr. */
/*   Unfortunately, some functions (OS specific) will not save off the    */
/* link register information when entered.  This is because they are not  */
/* going to be passing control off to any lower levels using this stack.  */
/* So...You may sometimes notice 1 line of invalid function information   */
/* from this routine if there was a signal involved and the currently     */
/* executing code was somewhere inbetween the creation of the stack area  */
/* and the moving of the link register into that area after function      */
/* transfer via the bl instruction.  Due to this inaccuracy it is very    */
/* important to take all pre signal stack information carefully.  This    */
/* is especially true if the last instruction was a bctr - Russ Blaisdell */
/**************************************************************************/
/*AIX*/void eo670GetFuncName(unsigned int *lr, char *funcname, size_t maxLength);
#    define MAX_FUNC_SEARCH_LEN 5000
/*AIX*/static void eo670GetFuncNameLength(unsigned int *lr, char **pFuncname, size_t *pNameLength, unsigned int *pOffset)
/*AIX*/{
/*AIX*/static struct tbtable *tb;
/*AIX*/static unsigned int   *ip;
/*AIX*/static short          name_len;
/*AIX*/static unsigned int  searchcount;
/*AIX*/ searchcount = 0;
/*AIX*//*	First, search from the input LR address looking for a 32 bit work */
/*AIX*//*	of zeros. */
/*AIX*/	ip = lr;
/*AIX*/	while((*ip != 0) && (searchcount++ < MAX_FUNC_SEARCH_LEN)) ip++;
/*AIX*/
/*AIX*/	if(*ip != 0) return;
/*AIX*/
/*AIX*//* set up addressability to the traceback table. */
/*AIX*/	tb = (struct tbtable *) (ip+1);
/*AIX*/
/*AIX*//* 	Existence of fields in the tbtable extension are       */
/*AIX*//* 	contingent upon specific fields in the base table.     */
/*AIX*//* 	Check for their existence so that we can address the   */
/*AIX*//* 	function name if it exists.                            */
/*AIX*/
/*AIX*/        ip = (unsigned int *) tb +
/*AIX*/         		sizeof(struct tbtable_short)/sizeof(int);
/*AIX*/
/*AIX*/		if ( tb->tb.fixedparms != 0 || tb->tb.floatparms != 0 )
/*AIX*/      			ip++;
/*AIX*/
/*AIX*/		if ( tb->tb.has_tboff == TRUE )
/*AIX*/     {
/*AIX*/         if ( pOffset )
/*AIX*/         {
/*AIX*/             *pOffset = *ip - ((char *)tb - (char *)lr);
/*AIX*/         }
/*AIX*/			ip++;
/*AIX*/     }
/*AIX*/
/*AIX*/		if ( tb->tb.int_hndl == TRUE )
/*AIX*/			ip++;
/*AIX*/
/*AIX*/		if ( tb->tb.has_ctl == TRUE ) 
/*AIX*/			ip += (*ip) + 1; /* don't care */
/*AIX*/
/*AIX*//* Get function name if it exists. */
/*AIX*/		if ( tb->tb.name_present == TRUE )
/*AIX*/		{
/*AIX*/			*pNameLength = * (short *) ip;
/*AIX*/         *pFuncname = (char *)ip+2;
/*AIX*/		}
/*AIX*/     else
/*AIX*/		{
/*AIX*/         *pFuncname = "<symbol not available>";
/*AIX*/         *pNameLength = strlen(*pFuncname);
/*AIX*/		}
/*AIX*/}
/*AIX*/
/*AIX*/
/*AIX*/void eo670GetFuncName(unsigned int *lr, char *funcname, size_t maxLength)
/*AIX*/{
/*AIX*/    static size_t foundNameLength;
/*AIX*/    static char * foundFuncName;
/*AIX*/
/*AIX*/    eo670GetFuncNameLength(lr, &foundFuncName, &foundNameLength, (unsigned int *)0);
/*AIX*/    if ( foundNameLength >= maxLength )
/*AIX*/    {
/*AIX*/        foundNameLength = maxLength-1;
/*AIX*/    }
/*AIX*/    funcname[foundNameLength] = 0;
/*AIX*/    if ( foundNameLength > 0 )
/*AIX*/    {
/*AIX*/        memcpy (funcname, foundFuncName, foundNameLength);
/*AIX*/    }
/*AIX*/}
/*AIX*/
/*AIX*/static void eo670_RegisterDump(void *pContext)
/*AIX*/{
/*AIX*/    static struct sigcontext *context;
/*AIX*/    static int iReg;
/*AIX*/
/*AIX*/    context = (struct sigcontext *)pContext;
/*AIX*/
/*AIX*/    eo670WriteString( "-----------------------------------------------\n" );
/*AIX*/    for ( iReg = 0; iReg < (_NGPRS/2); iReg++ )
/*AIX*/    {
/*AIX*/        sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*AIX*/                     "r%02d r%02d  0x%016qx 0x%016qx\n",
/*AIX*/                      2*iReg,
/*AIX*/                      1+2*iReg,
/*AIX*/                      context->sc_jmpbuf.jmp_context.gpr[2*iReg],
/*AIX*/                      context->sc_jmpbuf.jmp_context.gpr[1+2*iReg] );
/*AIX*/        eo670WriteString( eo670MessageBuffer );
/*AIX*/    }
/*AIX*/    eo670WriteString("\n");
/*AIX*/
/*AIX*/    sp77sprintf( eo670MessageBuffer,
/*AIX*/                 sizeof(eo670MessageBuffer),
/*AIX*/                 "msr      0x%016qx\n",
/*AIX*/                 context->sc_jmpbuf.jmp_context.msr);
/*AIX*/    eo670WriteString(eo670MessageBuffer);
/*AIX*/
/*AIX*/    sp77sprintf( eo670MessageBuffer,
/*AIX*/                 sizeof(eo670MessageBuffer),
/*AIX*/                 "iar      0x%016qx\n",
/*AIX*/                 context->sc_jmpbuf.jmp_context.iar);
/*AIX*/    eo670WriteString(eo670MessageBuffer);
/*AIX*/
/*AIX*/    sp77sprintf( eo670MessageBuffer,
/*AIX*/                 sizeof(eo670MessageBuffer),
/*AIX*/                 "pc       0x%016qx\n",
/*AIX*/                 context->sc_jmpbuf.jmp_context.lr);
/*AIX*/    eo670WriteString(eo670MessageBuffer);
/*AIX*/
/*AIX*/    sp77sprintf( eo670MessageBuffer,
/*AIX*/                 sizeof(eo670MessageBuffer),
/*AIX*/                 "ctr      0x%016qx\n",
/*AIX*/                 context->sc_jmpbuf.jmp_context.ctr);
/*AIX*/    eo670WriteString(eo670MessageBuffer);
/*AIX*/
/*AIX*/    sp77sprintf( eo670MessageBuffer,
/*AIX*/                 sizeof(eo670MessageBuffer),
/*AIX*/                 "cr       0x%08x\n",
/*AIX*/                 context->sc_jmpbuf.jmp_context.cr);
/*AIX*/    eo670WriteString(eo670MessageBuffer);
/*AIX*/
/*AIX*/    sp77sprintf( eo670MessageBuffer,
/*AIX*/                 sizeof(eo670MessageBuffer),
/*AIX*/                 "xer      0x%08x\n",
/*AIX*/                 context->sc_jmpbuf.jmp_context.xer);
/*AIX*/    eo670WriteString(eo670MessageBuffer);
/*AIX*/
/*AIX*/    sp77sprintf( eo670MessageBuffer,
/*AIX*/                 sizeof(eo670MessageBuffer),
/*AIX*/                 "fpscr    0x%08x\n",
/*AIX*/                 context->sc_jmpbuf.jmp_context.fpscr);
/*AIX*/    eo670WriteString(eo670MessageBuffer);
/*AIX*/
/*AIX*/    sp77sprintf( eo670MessageBuffer,
/*AIX*/                 sizeof(eo670MessageBuffer),
/*AIX*/                 "fpscrx   0x%08x\n",
/*AIX*/                 context->sc_jmpbuf.jmp_context.fpscrx);
/*AIX*/    eo670WriteString(eo670MessageBuffer);
/*AIX*/
/*AIX*/    sp77sprintf( eo670MessageBuffer,
/*AIX*/                 sizeof(eo670MessageBuffer),
/*AIX*/                 "except   0x%016qx\n",
/*AIX*/                 context->sc_jmpbuf.jmp_context.except[0]);
/*AIX*/    eo670WriteString(eo670MessageBuffer);
/*AIX*/
/*AIX*/    eo670WriteString( "-----------------------------------------------\n" );
/*AIX*/
/*AIX*/    if ( context->sc_jmpbuf.jmp_context.fpeu )
/*AIX*/    {
/*AIX*/        for ( iReg = 0; iReg < _NFPRS; iReg++ )
/*AIX*/        {
/*AIX*/            sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*AIX*/                     "f%02d      [0x%016qx]%.20e\n",
/*AIX*/                      iReg,
/*AIX*/                      *(SAPDB_UInt8 *)&context->sc_jmpbuf.jmp_context.fpr[iReg],
/*AIX*/                      *(double *)&context->sc_jmpbuf.jmp_context.fpr[iReg] );
/*AIX*/            eo670WriteString( eo670MessageBuffer );
/*AIX*/        }
/*AIX*/        eo670WriteString("\n");
/*AIX*/        sp77sprintf( eo670MessageBuffer,
/*AIX*/                 sizeof(eo670MessageBuffer),
/*AIX*/                 "fpinfo   0x%02x\n",
/*AIX*/                 context->sc_jmpbuf.jmp_context.fpinfo);
/*AIX*/        eo670WriteString(eo670MessageBuffer);
/*AIX*/    }
/*AIX*/    else
/*AIX*/    {
/*AIX*/        eo670WriteString( "Floating point unit unused\n");
/*AIX*/    }
/*AIX*/    eo670WriteString( "-----------------------------------------------\n" );
/*AIX*/
/*AIX*/}
/*AIX*/
/*AIX*/SAPDB_Int4 eo670_FastFillCallArray(SAPDB_Int4 skipCount, SAPDB_Int4 level, void **callArray)
/*AIX*/{
/*AIX*/    unsigned long *sp; /* stack pointer */
/*AIX*/    unsigned long *lr; /* link register */
/*AIX*/    unsigned long *oldsp;
/*AIX*/    ucontext_t localContext;
/*AIX*/    SAPDB_Int4     depth;
/*AIX*/
/*AIX*/    depth = 0;
/*AIX*/
/*AIX*/    if ( getcontext (&localContext) )
/*AIX*/    {
/*AIX*/        eo670WriteString("Failed to get ucontext\n");
/*AIX*/        return 0;
/*AIX*/    }
/*AIX*/    /* On PowerPC-AIX-Maschine the stack pointer is in GPR 1 */
/*AIX*/    sp = (unsigned long *)(localContext.uc_mcontext.jmp_context.gpr[1]);
/*AIX*/
/*AIX*/    /* Start off with our caller */
/*AIX*/    oldsp = sp-10;
/*AIX*/    /* Loop while we still have a valid stack pointer */
/*AIX*/    while( depth < level
/*AIX*/       && ( (( (char*)sp - (char *)0 ) & (sizeof(unsigned long)-1)) == 0 )
/*AIX*/       && (sp > oldsp) )
/*AIX*/    {
/*AIX*/        oldsp = sp;
/*AIX*/
/*AIX*/        /* See if this is a signal.  This is done by looking at the LR field  */
/*AIX*/        /* on the stack.  The LR is at offset long 2 (byte 8 in 32 bit mode   */
/*AIX*/        /* or byte 16 in 64 bit mode).  If the address is from segment 0,     */
/*AIX*/        /* then this must be a signal.                                        */
/*AIX*/        if (*(sp+2) < 0x0fffffff) 	/* check LR field on stack */
/*AIX*/    	   {
/*AIX*/            break;
/*AIX*/    	   }
/*AIX*/        if ( skipCount > 0 )
/*AIX*/        {
/*AIX*/            --skipCount;
/*AIX*/        }
/*AIX*/        else
/*AIX*/        {
/*AIX*/            /* Handle potential local stack allocations within a function  */
/*AIX*/            /* these produces extra frames which are skipped here          */
/*AIX*/            if ( 0 != *(sp+2) )
/*AIX*/            {
/*AIX*/                callArray[depth] = (void *)*(sp+2);
/*AIX*/                ++depth;
/*AIX*/            }
/*AIX*/        }
/*AIX*/
/*AIX*/        /* If here, the stack frame is not the result of a signal.  The LR is */
/*AIX*/        /* accessed at long + 2, but the pointer to the next stack frame is   */
/*AIX*/        /* pointed to by the contents of where the current pointer points.    */
/*AIX*/        sp = (unsigned long *) *sp;
/*AIX*/    }
/*AIX*/    return depth;
/*AIX*/}
/*AIX*/
/*AIX*/SAPDB_Int4 eo670_UnixFillCallArray(SAPDB_Int4 skipCount, SAPDB_Int4 level, eo670_CallArrayType addCallInfo, void **callArray, void *pContext)
/*AIX*/{
/*AIX*/    static unsigned long *sp; /* stack pointer */
/*AIX*/    static unsigned long *lr; /* link register */
/*AIX*/    static unsigned long *oldsp;
/*AIX*/    static SAPDB_Int4     depth;
/*AIX*/
/*AIX*/    depth = 0;
/*AIX*/
/*AIX*/    if ( 0 == pContext )
/*AIX*/    {
/*AIX*/        static ucontext_t localContext;
/*AIX*/        if ( getcontext (&localContext) )
/*AIX*/        {
/*AIX*/            eo670WriteString("Failed to get ucontext\n");
/*AIX*/            return 0;
/*AIX*/        }
/*AIX*/        /* On PowerPC-AIX-Maschine the stack pointer is in GPR 1 */
/*AIX*/        sp = (unsigned long *)(localContext.uc_mcontext.jmp_context.gpr[1]);
/*AIX*/    }
/*AIX*/    else
/*AIX*/    {
/*AIX*/        skipCount = 0;
/*AIX*/        sp = (unsigned long *)(((struct sigcontext *)pContext)->sc_jmpbuf.jmp_context.gpr[1]);
/*AIX*/        switch ( addCallInfo )
/*AIX*/        {
/*AIX*/        case eo670_FullInfoArray:
/*AIX*/            callArray[depth] = (void *)(((struct sigcontext *)pContext)->sc_jmpbuf.jmp_context.iar);
/*AIX*/            break;
/*AIX*/        case eo670_PointerArray:
/*AIX*/            callArray[depth] = (void *)(((struct sigcontext *)pContext)->sc_jmpbuf.jmp_context.iar);
/*AIX*/            break;
/*AIX*/        default: /* eo670_NoArrayJustOutput */
/*AIX*/            sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*AIX*/                         "(%d):0x%lx\n",
/*AIX*/                         depth, (((struct sigcontext *)pContext)->sc_jmpbuf.jmp_context.iar) );
/*AIX*/            eo670WriteString(eo670MessageBuffer);
/*AIX*/            break;
/*AIX*/        }
/*AIX*/        ++depth;
/*AIX*/    }
/*AIX*/
/*AIX*/    /* Start off with our caller */
/*AIX*/    oldsp = sp-10;
/*AIX*/    /* Loop while we still have a valid stack pointer */
/*AIX*/    while( depth < level
/*AIX*/       && ( (( (char*)sp - (char *)0 ) & (sizeof(unsigned long)-1)) == 0 )
/*AIX*/       && (sp > oldsp) )
/*AIX*/    {
/*AIX*/        oldsp = sp;
/*AIX*/
/*AIX*/        /* See if this is a signal.  This is done by looking at the LR field  */
/*AIX*/        /* on the stack.  The LR is at offset long 2 (byte 8 in 32 bit mode   */
/*AIX*/        /* or byte 16 in 64 bit mode).  If the address is from segment 0,     */
/*AIX*/        /* then this must be a signal.                                        */
/*AIX*/        if (*(sp+2) < 0x0fffffff) 	/* check LR field on stack */
/*AIX*/    	   {
/*AIX*/            break;
/*AIX*/    	   }
/*AIX*/        if ( skipCount > 0 )
/*AIX*/        {
/*AIX*/            --skipCount;
/*AIX*/        }
/*AIX*/        else
/*AIX*/        {
/*AIX*/            /* Handle potential local stack allocations within a function  */
/*AIX*/            /* these produces extra frames which are skipped here          */
/*AIX*/            if ( 0 != *(sp+2) )
/*AIX*/            {
/*AIX*/                switch ( addCallInfo )
/*AIX*/                {
/*AIX*/                case eo670_FullInfoArray:
/*AIX*/                    callArray[depth] = (void *)*(sp+2);
/*AIX*/                    break;
/*AIX*/                case eo670_PointerArray:
/*AIX*/                    callArray[depth] = (void *)*(sp+2);
/*AIX*/                    break;
/*AIX*/                default: /* eo670_NoArrayJustOutput */
/*AIX*/                    sp77sprintf( eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*AIX*/                                 "(%d):0x%lx\n",
/*AIX*/                                 depth, *(sp+2) );
/*AIX*/                    eo670WriteString(eo670MessageBuffer);
/*AIX*/                    break;
/*AIX*/                }
/*AIX*/                ++depth;
/*AIX*/            }
/*AIX*/        }
/*AIX*/
/*AIX*/        /* If here, the stack frame is not the result of a signal.  The LR is */
/*AIX*/        /* accessed at long + 2, but the pointer to the next stack frame is   */
/*AIX*/        /* pointed to by the contents of where the current pointer points.    */
/*AIX*/        sp = (unsigned long *) *sp;
/*AIX*/    }
/*AIX*/    return depth;
/*AIX*/}
/*AIX*/
/*AIX*/static void eo670_UpdateModuleChain(void)
/*AIX*/{
/*AIX*/    static char moduleInfo[4096];
/*AIX*/    static struct ld_info *pInfo;
/*AIX*/    static int rc;
/*AIX*/
/*AIX*/    rc = loadquery(L_GETINFO, moduleInfo, sizeof(moduleInfo));
/*AIX*/    if ( rc < 0 )
/*AIX*/    {
/*AIX*/        sp77sprintf(eo670MessageBuffer, sizeof(eo670MessageBuffer),
/*AIX*/                    "No module list since loadquery failed: %d\n", errno);
/*AIX*/        eo670WriteString(eo670MessageBuffer);
/*AIX*/        return;
/*AIX*/    }
/*AIX*/
/*AIX*/    for ( pInfo = (struct ld_info *)moduleInfo;
/*AIX*/          pInfo;
/*AIX*/          pInfo = (struct ld_info *)( ((char *)pInfo) + pInfo->ldinfo_next) )
/*AIX*/    {
/*AIX*/        switch( eo670_AddToLoadedModuleChain( pInfo->ldinfo_filename,
/*AIX*/                                              pInfo->ldinfo_textorg,
/*AIX*/            ((char *)pInfo->ldinfo_textorg) + pInfo->ldinfo_textsize,
/*AIX*/                                              pInfo->ldinfo_filename ) )
/*AIX*/        {
/*AIX*/            case 1:  break; /* new entry */
/*AIX*/            case 0:  break; /* already added */
/*AIX*/            case -1: pInfo->ldinfo_next = 0; /* cannot add module anymore */
/*AIX*/                     break;
/*AIX*/        }
/*AIX*/
/*AIX*/        if ( !pInfo->ldinfo_next )
/*AIX*/        {
/*AIX*/            break;
/*AIX*/        }
/*AIX*/    }
/*AIX*/}
/*AIX*/
/*AIX*/static void eo670_FillAddressInfo(char *pcAddr, eo670AddressInfo *addrInfo)
/*AIX*/{
/*AIX*/    static size_t foundNameLength;
/*AIX*/    static char * foundFuncName;
/*AIX*/    static eo670_ModuleChain *modulePointer;
/*AIX*/    static unsigned int offsetToSymbol;
#    if defined(WANT_DEMANGLING_SUPPORT)
/*AIX*/    static Name *demangledName;
/*AIX*/    static char *restSymbol;
#    endif /* WANT_DEMANGLING_SUPPORT */
/*AIX*/    memset(addrInfo, 0, sizeof(eo670AddressInfo));
/*AIX*/
/*AIX*/    addrInfo->sourceFileName = ""; /* No Source information on AIX... */
/*AIX*/    addrInfo->lineNumber = -1;
/*AIX*/
/*AIX*/    for ( modulePointer = eo670_ModuleAnchor
/*AIX*/        ; modulePointer
/*AIX*/        ; modulePointer = modulePointer->next )
/*AIX*/    {
/*AIX*/        if ( ( (pcAddr) >= ((char *)modulePointer->text_mapaddr) )
/*AIX*/          && ( (pcAddr) <  ((char *)modulePointer->text_endaddr) ) ) 
/*AIX*/        {
/*AIX*/            /* cast is ok, since modulePointer file_name points to */
/*AIX*/            /* dynamically allocated space (see eo670_UpdateModuleChain() */
/*AIX*/            addrInfo->moduleName = modulePointer->file_name;
/*AIX*/	           break;
/*AIX*/   	   }
/*AIX*/    }
/*AIX*/
/*AIX*/    if ( !modulePointer )
/*AIX*/    {
/*AIX*/        addrInfo->offsetToSymbol = -1;
/*AIX*/        addrInfo->symbol = "<not in any loaded region>";
/*AIX*/        addrInfo->demangledSymbol = "";
/*AIX*/        addrInfo->moduleName = "<not in any loaded module>";
/*AIX*/        return;
/*AIX*/    }
/*AIX*/
/*AIX*/    eo670GetFuncNameLength((unsigned int *)pcAddr, &foundFuncName, &foundNameLength, &offsetToSymbol);
/*AIX*/    if ( foundNameLength > 0 )
/*AIX*/    {
/*AIX*/        static char lastSymbolBuffer[MAX_SYMBOL_LENGTH]; /* next call overwrites it! */
/*AIX*/        if ( foundNameLength >= MAX_SYMBOL_LENGTH )
/*AIX*/        {
/*AIX*/            eo670WriteString("Symbol too long, truncated...");
/*AIX*/            memcpy (lastSymbolBuffer, foundFuncName, MAX_SYMBOL_LENGTH-1);
/*AIX*/            lastSymbolBuffer[MAX_SYMBOL_LENGTH-1] = 0;
/*AIX*/        }
/*AIX*/        else
/*AIX*/        {
/*AIX*/            memcpy (lastSymbolBuffer, foundFuncName, foundNameLength);
/*AIX*/            lastSymbolBuffer[foundNameLength] = 0;
/*AIX*/        }
/*AIX*/        addrInfo->symbol = lastSymbolBuffer;
#    if defined(WANT_DEMANGLING_SUPPORT)
/*AIX*/        demangledName = demangle(addrInfo->symbol, &restSymbol,
/*AIX*/           RegularNames | ClassNames | SpecialNames | ParameterText | QualifierText);
/*AIX*/        addrInfo->demangledSymbol = text(demangledName);
#    else  /* !WANT_DEMANGLING_SUPPORT */
/*AIX*/        addrInfo->demangledSymbol = "";
#    endif /* !WANT_DEMANGLING_SUPPORT */
/*AIX*/    }
/*AIX*/    else
/*AIX*/    {
/*AIX*/        addrInfo->symbol = "<symbol not available>";
/*AIX*/        addrInfo->demangledSymbol = "";
/*AIX*/    }
/*AIX*/    addrInfo->offsetToSymbol = offsetToSymbol;
/*AIX*/
/*AIX*/    return;
/*AIX*/}
/*AIX*/
/*AIX*/static SAPDB_Bool eo670_UnixGetAddressInfo(void *pcAddr, eo670AddressInfo *addrInfo)
/*AIX*/{
/*AIX*/    int    result;
/*AIX*/    static eo670AddressInfo localAddrInfo;
/*AIX*/
/*AIX*/    eo670_FillAddressInfo((char *)pcAddr, &localAddrInfo);
/*AIX*/    addrInfo->symbol = (char *)calloc( 1, strlen(localAddrInfo.symbol)+1
/*AIX*/                                 + strlen(localAddrInfo.demangledSymbol)+1
/*AIX*/                                 + strlen(localAddrInfo.moduleName)+1
/*AIX*/                                 + strlen(localAddrInfo.sourceFileName)+1 );
/*AIX*/    if ( !addrInfo->symbol )
/*AIX*/    {
/*AIX*/        eo670WriteString("Out of memory for symbol copy\n");
/*AIX*/        result = false;
/*AIX*/    }
/*AIX*/    else
/*AIX*/    {
/*AIX*/        strcpy(addrInfo->symbol, localAddrInfo.symbol);
/*AIX*/        addrInfo->demangledSymbol = addrInfo->symbol + strlen(localAddrInfo.symbol)+1;
/*AIX*/        strcpy(addrInfo->demangledSymbol, localAddrInfo.demangledSymbol);
/*AIX*/        addrInfo->moduleName = addrInfo->demangledSymbol + strlen(localAddrInfo.demangledSymbol)+1;
/*AIX*/        strcpy(addrInfo->moduleName, localAddrInfo.moduleName);
/*AIX*/        addrInfo->sourceFileName = addrInfo->moduleName + strlen(localAddrInfo.moduleName)+1;
/*AIX*/        strcpy(addrInfo->sourceFileName, localAddrInfo.sourceFileName);
/*AIX*/        addrInfo->offsetToSymbol = localAddrInfo.offsetToSymbol;
/*AIX*/        addrInfo->lineNumber = localAddrInfo.lineNumber;
/*AIX*/        result = true;
/*AIX*/    }
/*AIX*/    return result;
/*AIX*/}
/*AIX*/
/*AIX*/static void eo670_OsFreeAddressInfo(eo670AddressInfo *addrInfo)
/*AIX*/{
/*AIX*/    free(addrInfo->symbol); /* allocated in one chunk.... */
/*AIX*/}
#  endif /* AIX */

#endif /* no _WIN32 */

/*===========================================================================*
 *  WINDOWS                                                                  *
 *===========================================================================*/

#if defined (_WIN32)

// miscellaneous toolhelp32 declarations; we cannot #include the header
// because not all systems may have it
#  pragma pack( push, 8 )
 typedef struct tagMODULEENTRY32
   {
   DWORD   dwSize;
   DWORD   th32ModuleID;       // This module
   DWORD   th32ProcessID;      // owning process
   DWORD   GlblcntUsage;       // Global usage count on the module
   DWORD   ProccntUsage;       // Module usage count in th32ProcessID's context
   BYTE  * modBaseAddr;        // Base address of module in th32ProcessID's context
   DWORD   modBaseSize;        // Size in bytes of module starting at modBaseAddr
   HMODULE hModule;            // The hModule of this module in th32ProcessID's context
   char    szModule[MAX_MODULE_NAME32 + 1];
   char    szExePath[MAX_PATH];
   } MODULEENTRY32;

 typedef MODULEENTRY32 *  PMODULEENTRY32;
 typedef MODULEENTRY32 *  LPMODULEENTRY32;
#  pragma pack( pop )

/*---------------------------------------------------------------------------*/

// SymCleanup()
typedef BOOL (__stdcall *tSC)( IN HANDLE hProcess );

// SymGetOptions()
typedef DWORD (__stdcall *tSGO)( VOID );

 // SymInitialize()
 typedef BOOL (__stdcall *tSI)( IN HANDLE hProcess, IN PCSTR UserSearchPath, IN BOOL fInvadeProcess );

 // SymSetOptions()
 typedef DWORD (__stdcall *tSSO)( IN DWORD SymOptions );

 // UnDecorateSymbolName()
 typedef DWORD (__stdcall WINAPI *tUDSN)( PCSTR DecoratedName, PSTR UnDecoratedName,
	 DWORD UndecoratedLength, DWORD Flags );

#  if defined (_WIN64)
 // SymFunctionTableAccess()
 typedef PVOID (__stdcall *tSFTA)( HANDLE hProcess, DWORD64 AddrBase );

 // SymGetLineFromAddr()
 typedef BOOL (__stdcall *tSGLFA)( IN HANDLE hProcess, IN DWORD64 dwAddr,
	 OUT PDWORD pdwDisplacement, OUT PIMAGEHLP_LINE64 Line );

 // SymGetModuleBase()
 typedef DWORD64 (__stdcall *tSGMB)( IN HANDLE hProcess, IN DWORD64 dwAddr );

 // SymGetModuleInfo()
 typedef BOOL (__stdcall *tSGMI)( IN HANDLE hProcess, IN DWORD64 dwAddr, OUT PIMAGEHLP_MODULE64 ModuleInfo );

 // SymGetSymFromAddr()
 typedef BOOL (__stdcall *tSGSFA)( IN HANDLE hProcess, IN DWORD64 dwAddr,
 	 OUT PDWORD64 pdwDisplacement, OUT PIMAGEHLP_SYMBOL64 Symbol );

 // SymLoadModule()
 typedef DWORD64 (__stdcall *tSLM)( IN HANDLE hProcess, IN HANDLE hFile,
 	 IN PSTR ImageName, IN PSTR ModuleName, IN DWORD64 BaseOfDll, IN DWORD SizeOfDll );

 // StackWalk()
 typedef BOOL (__stdcall *tSW)( DWORD MachineType, HANDLE hProcess,
 	 HANDLE hThread, LPSTACKFRAME64 StackFrame, PVOID ContextRecord,
	 PREAD_PROCESS_MEMORY_ROUTINE64 ReadMemoryRoutine,
	 PFUNCTION_TABLE_ACCESS_ROUTINE64 FunctionTableAccessRoutine,
	 PGET_MODULE_BASE_ROUTINE64 GetModuleBaseRoutine,
	 PTRANSLATE_ADDRESS_ROUTINE64 TranslateAddress );

#  else /* WIN32 only */

 // SymFunctionTableAccess()
 typedef PVOID (__stdcall *tSFTA)( HANDLE hProcess, DWORD AddrBase );

 // SymGetLineFromAddr()
 typedef BOOL (__stdcall *tSGLFA)( IN HANDLE hProcess, IN DWORD dwAddr,
	 OUT PDWORD pdwDisplacement, OUT PIMAGEHLP_LINE Line );

 // SymGetModuleBase()
 typedef DWORD (__stdcall *tSGMB)( IN HANDLE hProcess, IN DWORD dwAddr );

 // SymGetModuleInfo()
 typedef BOOL (__stdcall *tSGMI)( IN HANDLE hProcess, IN DWORD dwAddr, OUT PIMAGEHLP_MODULE ModuleInfo );

 // SymGetSymFromAddr()
 typedef BOOL (__stdcall *tSGSFA)( IN HANDLE hProcess, IN DWORD dwAddr,
 	 OUT PDWORD pdwDisplacement, OUT PIMAGEHLP_SYMBOL Symbol );

 // SymLoadModule()
 typedef BOOL (__stdcall *tSLM)( IN HANDLE hProcess, IN HANDLE hFile,
 	 IN PSTR ImageName, IN PSTR ModuleName, IN DWORD BaseOfDll, IN DWORD SizeOfDll );

 // StackWalk()
 typedef BOOL (__stdcall *tSW)( DWORD MachineType, HANDLE hProcess,
 	 HANDLE hThread, LPSTACKFRAME StackFrame, PVOID ContextRecord,
	 PREAD_PROCESS_MEMORY_ROUTINE ReadMemoryRoutine,
	 PFUNCTION_TABLE_ACCESS_ROUTINE FunctionTableAccessRoutine,
	 PGET_MODULE_BASE_ROUTINE GetModuleBaseRoutine,
	 PTRANSLATE_ADDRESS_ROUTINE TranslateAddress );
#  endif /* _WIN64 */

/*---------------------------------------------------------------------------*/

static tSLM      pSymLoadModule = NULL;
static tSFTA     pSymFunctionTableAccess = NULL;
static tSGMB     pSymGetModuleBase = NULL;
static tSW       pStackWalk = NULL;

static tSC       pSymCleanup = NULL;
static tSGLFA    pSymGetLineFromAddr = NULL;
static tSGMI     pSymGetModuleInfo = NULL;
static tSGO      pSymGetOptions = NULL;
static tSGSFA    pSymGetSymFromAddr = NULL;
static tSI       pSymInitialize = NULL;
static tSSO      pSymSetOptions = NULL;
static tUDSN     pUnDecorateSymbolName = NULL;

/*---------------------------------------------------------------------------*/


/* PTS 1110627 */
static FARPROC myGetProcAddress(HMODULE module, LPCSTR lpProcName, bool Silent )
{
    FARPROC funcPtr = GetProcAddress(module, lpProcName);

    if ( funcPtr == NULL )
    {
        static char MessageLine[1024];
        if ( !Silent )
        {
          sp77sprintf(MessageLine, sizeof(MessageLine), "GetProcAddress(%s) failed: rc = %lu\n", lpProcName, GetLastError() );
          eo670WriteString( MessageLine );
        }
    }
    return funcPtr;
}

/*---------------------------------------------------------------------------*/

static HINSTANCE LoadImageHlpDll()
{
  static HINSTANCE hImagehlpDll = NULL;

  if ( hImagehlpDll != NULL )
    return hImagehlpDll;
  
  hImagehlpDll = LoadLibrary( "imagehlp.dll" );
  
  if ( hImagehlpDll != NULL )
  {
    pSymCleanup             = (tSC) myGetProcAddress( hImagehlpDll,    "SymCleanup", false );
    pSymGetOptions          = (tSGO) myGetProcAddress( hImagehlpDll,   "SymGetOptions", false );
    pSymInitialize          = (tSI) myGetProcAddress( hImagehlpDll,    "SymInitialize", false );
    pSymSetOptions          = (tSSO) myGetProcAddress( hImagehlpDll,   "SymSetOptions", false );
#  if defined (_WIN64)
    pStackWalk              = (tSW) myGetProcAddress( hImagehlpDll,    "StackWalk64", false );
    pSymGetSymFromAddr      = (tSGSFA) myGetProcAddress( hImagehlpDll, "SymGetSymFromAddr64", false );
    pSymGetModuleInfo       = (tSGMI) myGetProcAddress( hImagehlpDll,  "SymGetModuleInfo64", false );
    pSymGetLineFromAddr     = (tSGLFA) GetProcAddress( hImagehlpDll,   "SymGetLineFromAddr64", false );
    pSymFunctionTableAccess = (tSFTA) myGetProcAddress( hImagehlpDll,  "SymFunctionTableAccess64", false );
    pSymGetModuleBase       = (tSGMB) myGetProcAddress( hImagehlpDll,  "SymGetModuleBase64", false );
    pSymLoadModule          = (tSLM) myGetProcAddress( hImagehlpDll,   "SymLoadModule64", false );
#  else /* WIN32 only */
    pStackWalk              = (tSW) myGetProcAddress( hImagehlpDll,    "StackWalk", false );
    pSymGetSymFromAddr      = (tSGSFA) myGetProcAddress( hImagehlpDll, "SymGetSymFromAddr", false );
    pSymGetModuleInfo       = (tSGMI) myGetProcAddress( hImagehlpDll,  "SymGetModuleInfo", false );
    pSymGetLineFromAddr     = (tSGLFA) myGetProcAddress( hImagehlpDll,   "SymGetLineFromAddr", false );
    pSymFunctionTableAccess = (tSFTA) myGetProcAddress( hImagehlpDll,  "SymFunctionTableAccess", false );
    pSymGetModuleBase       = (tSGMB) myGetProcAddress( hImagehlpDll,  "SymGetModuleBase", false );
    pSymLoadModule          = (tSLM) myGetProcAddress( hImagehlpDll,   "SymLoadModule", false );
#  endif /* _WIN64 */
    pUnDecorateSymbolName = (tUDSN) myGetProcAddress( hImagehlpDll, "UnDecorateSymbolName", false );
  }

  if ( pSymCleanup == NULL
    || pSymFunctionTableAccess == NULL
    || pSymGetModuleBase == NULL
    || pSymGetModuleInfo == NULL
    || pSymGetOptions == NULL
    || pSymGetSymFromAddr == NULL
    || pSymInitialize == NULL
    || pSymLoadModule == NULL
    || pSymSetOptions == NULL
    || pStackWalk == NULL
    || pUnDecorateSymbolName == NULL )
  {
    FreeLibrary( hImagehlpDll );
    hImagehlpDll = NULL;
  }

  return hImagehlpDll;
}

/*---------------------------------------------------------------------------*/

#if !defined(_WIN64)
/*WIN32*/struct WinI386Frame
/*WIN32*/{
/*WIN32*/    struct WinI386Frame *fr_savfp;
/*WIN32*/    void *fr_savpc;
/*WIN32*/};
/*WIN32*/
/*WIN32*/SAPDB_Int4 eo670_FastFillCallArray(SAPDB_Int4 skipCount, SAPDB_Int4 level, void **callArray)
/*WIN32*/{
/*WIN32*/    struct WinI386Frame *fp;
/*WIN32*/    SAPDB_Int4 depth;
/*WIN32*/
/*WIN32*/    depth = 0;
/*WIN32*/
/*WIN32*/	 __asm {
/*WIN32*/		mov fp, ebp
/*WIN32*/	 }
/*WIN32*/
/*WIN32*/    while ( depth < level
/*WIN32*/         && fp 
/*WIN32*/         && ((((long)fp)&3) == 0) 
/*WIN32*/         && (fp != fp->fr_savfp) )
/*WIN32*/    {
/*WIN32*/         if ( skipCount > 0 )
/*WIN32*/         {
/*WIN32*/             --skipCount;
/*WIN32*/         }
/*WIN32*/         else
/*WIN32*/         {
/*WIN32*/            callArray[depth] = fp->fr_savpc;
/*WIN32*/            ++depth;
/*WIN32*/         }
/*WIN32*/         fp = fp->fr_savfp;
/*WIN32*/    }
/*WIN32*/    return depth;
/*WIN32*/}
#endif

static SAPDB_Int4 eo670_WinFillCallArray(SAPDB_Int4 level, void **callArray, void *pContext)
{
  CONTEXT	        *context;
#  if defined (_WIN64)
   STACKFRAME64      stackFrame; // in/out stackframe
   DWORD             imageType = IMAGE_FILE_MACHINE_IA64;
#  else /* WIN32 only */
   STACKFRAME        stackFrame; // in/out stackframe
   DWORD             imageType = IMAGE_FILE_MACHINE_I386;
#  endif /* _WIN64 */
  SAPDB_Int4        depth;
  HANDLE            hThread;
  HANDLE            hProcess;

  if( !DuplicateHandle( GetCurrentProcess(), GetCurrentProcess(), GetCurrentProcess(),
                        &hProcess, DUPLICATE_SAME_ACCESS, TRUE, DUPLICATE_SAME_ACCESS )) 
  {
    hProcess = GetCurrentProcess(); 
  }

  if ( !LoadImageHlpDll() )
  {
    CloseHandle(hProcess);
    return 0;
  }

  hThread = GetCurrentThread();

  if ( 0 == pContext )
  {
      CONTEXT localContext;
      memset( &localContext, 0, sizeof(localContext) );
      localContext.ContextFlags = CONTEXT_FULL;
      if ( GetThreadContext( hThread, &localContext ) )
      {
          context = &localContext;
      }
      else
      {
          return 0;
      }
  }
  else
  {
      context = (CONTEXT *)pContext;
  }

  memset( &stackFrame, 0, sizeof(stackFrame) );
  
  depth   = 0;

  __try
  {
      // init STACKFRAME for first call
      // Notes: AddrModeFlat is just an assumption.
#  if defined (_WIN64)
      stackFrame.AddrPC.Offset    = 0;
      stackFrame.AddrPC.Mode      = AddrModeFlat;
      stackFrame.AddrStack.Offset = 0;
      stackFrame.AddrStack.Mode   = AddrModeFlat;
      stackFrame.AddrFrame.Offset = 0;
      stackFrame.AddrFrame.Mode   = AddrModeFlat;
#  else /* WIN32 only */
      stackFrame.AddrPC.Offset    = context->Eip;
      stackFrame.AddrPC.Mode      = AddrModeFlat;
      stackFrame.AddrStack.Offset = context->Esp;
      stackFrame.AddrStack.Mode   = AddrModeFlat;
      stackFrame.AddrFrame.Offset = context->Ebp;
      stackFrame.AddrFrame.Mode   = AddrModeFlat;
#  endif /* _WIN64 */
      while ( depth < level )
      {
        if ( ! pStackWalk( imageType, hProcess, hThread, 
          &stackFrame, context, NULL,
          pSymFunctionTableAccess, pSymGetModuleBase, NULL ) )
          break;
        callArray[depth++] = (void *)stackFrame.AddrPC.Offset;
      }
  }
  __except( EXCEPTION_EXECUTE_HANDLER ) {;}

  if ( GetCurrentProcess() != hProcess ) 
    CloseHandle(hProcess);

  return depth;
}

/*---------------------------------------------------------------------------*/

static SAPDB_Bool eo670_WinGetAddressInfo (void *pcAddr, eo670AddressInfo *addrInfo, void *context)
{
#  if defined (_WIN64)
   static IMAGEHLP_LINE64    Line;
   static DWORD64            offsetFromSymbol; // tells us how far from the symbol we were
   static IMAGEHLP_MODULE64  Module;
   static DWORD              imageType = IMAGE_FILE_MACHINE_IA64;
#  else /* WIN32 only */
   static IMAGEHLP_LINE      Line;
   static DWORD              offsetFromSymbol; // tells us how far from the symbol we were
   static IMAGEHLP_MODULE    Module;
   static DWORD              imageType = IMAGE_FILE_MACHINE_I386;
#  endif /* _WIN64 */
  static double              SymSpace[(sizeof(IMAGEHLP_SYMBOL) + MAXSYMNAMELEN)/sizeof(double)+1];
  static IMAGEHLP_SYMBOL     *pSym = (IMAGEHLP_SYMBOL *) SymSpace;
  static char                undName[MAXSYMNAMELEN];     // undecorated name
  static char                undFullName[MAXSYMNAMELEN]; // undecorated name with all shenanigans
  static char                MessageLine[2048];
  eo670_GetAddressInfoContext* pContext;

  pContext = (eo670_GetAddressInfoContext*) context;

  memset( &Module, '\0', sizeof(Module) );
  Module.SizeOfStruct = sizeof(Module);
  
  if ( ! pSymGetModuleInfo( pContext->m_handle, (unsigned long) pcAddr, &Module ) )
  {
    sp77sprintf( MessageLine, sizeof(MessageLine), "  SymGetModuleInfo() failed, rc = %d\n",
      GetLastError());
    eo670WriteString(MessageLine);
    Module.ModuleName[0] = '?';
    Module.ModuleName[1] = '\0';
  }

  memset( pSym, '\0', sizeof(SymSpace));
  pSym->SizeOfStruct  = sizeof(IMAGEHLP_SYMBOL);
  pSym->MaxNameLength = MAXSYMNAMELEN;
  offsetFromSymbol    = 0;

  if ( ! pSymGetSymFromAddr( pContext->m_handle, (unsigned long) pcAddr, &offsetFromSymbol, pSym ) )
  {
    return false;
  }

  // UnDecorateSymbolName()
    pUnDecorateSymbolName( pSym->Name, undName, MAXSYMNAMELEN, UNDNAME_NAME_ONLY );
    pUnDecorateSymbolName( pSym->Name, undFullName, MAXSYMNAMELEN, UNDNAME_COMPLETE );
    addrInfo->symbol          = (char *)malloc(strlen(undName)+1+strlen(undFullName)+1+strlen(Module.ModuleName)+1);
    strcpy(addrInfo->symbol, undName);
    addrInfo->demangledSymbol = addrInfo->symbol + strlen(undName)+1;
    strcpy(addrInfo->demangledSymbol, undFullName);
    addrInfo->moduleName      = addrInfo->demangledSymbol + strlen(undFullName)+1;
    strcpy(addrInfo->moduleName, Module.ModuleName);
    addrInfo->offsetToSymbol  = offsetFromSymbol;
    addrInfo->sourceFileName = "";
    addrInfo->lineNumber = -1;
    return true;
}

/*---------------------------------------------------------------------------*/

static void eo670_OsFreeAddressInfo(eo670AddressInfo *addrInfo)
{
  free (addrInfo->symbol);   
}

/*---------------------------------------------------------------------------*/

void eo670_WinGetAddressInfoEpilog (void* context)
{
  eo670_GetAddressInfoContext* pContext;
  pContext = (eo670_GetAddressInfoContext*) context;
  pSymCleanup( pContext->m_handle );
  if ( GetCurrentProcess() != pContext->m_handle ) 
    CloseHandle(pContext->m_handle);
  free(context);
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool eo670_WinGetAddressInfoProlog (void** context)
{
  DWORD                        symOptions; // symbol handler settings
  eo670_GetAddressInfoContext* pContext;
  char const *                 symSearchPath;

  pContext = (eo670_GetAddressInfoContext*) malloc (sizeof(eo670_GetAddressInfoContext));
  if (pContext)
  {
    if( !DuplicateHandle( GetCurrentProcess(), GetCurrentProcess(), GetCurrentProcess(),
      &pContext->m_handle, DUPLICATE_SAME_ACCESS, TRUE, DUPLICATE_SAME_ACCESS )) 
    {
      pContext->m_handle = GetCurrentProcess(); 
    }
    if ( LoadImageHlpDll() )
    {
      symSearchPath = eo670_GetSymSearchPath();
      symOptions  = pSymGetOptions();
      symOptions |= SYMOPT_LOAD_LINES;
      symOptions &= ~SYMOPT_UNDNAME;
      pSymSetOptions( symOptions );
    
      if ( ! pSymInitialize( pContext->m_handle, symSearchPath, false ) )
      {
        eo670WriteString("SymInitialize() failed\n");
        pSymCleanup( pContext->m_handle );
      }
      else
      {
         eo670_FillModuleList (GetCurrentProcessId(), pContext->m_handle, 0 );
         *context = pContext;
         return true;
      }
    }
    /* some error occured */
    CloseHandle(pContext->m_handle);
    free(pContext);
  }
  *context = (void*) 0;
  return false;
}

/*---------------------------------------------------------------------------*/

static void eo670_WinShowStack( HANDLE hThread, CONTEXT *context, int VerbosityLevel )
{
  // normally, call ImageNtHeader() and use machine info from PE header
#  if defined (_WIN64)
   static IMAGEHLP_LINE64    Line;
   static STACKFRAME64       sframe; // in/out stackframe
   static DWORD64            offsetFromSymbol; // tells us how far from the symbol we were
   static IMAGEHLP_MODULE64  Module;
   static DWORD              imageType = IMAGE_FILE_MACHINE_IA64;
#  else /* WIN32 only */
   static IMAGEHLP_LINE      Line;
   static STACKFRAME         sframe; // in/out stackframe
   static DWORD              offsetFromSymbol; // tells us how far from the symbol we were
   static IMAGEHLP_MODULE    Module;
   static DWORD              imageType = IMAGE_FILE_MACHINE_I386;
#  endif /* _WIN64 */

  static DWORD               offsetFromLine; 
  static int                 frameNum; // counts walked frames
  static DWORD               symOptions; // symbol handler settings
  static double              SymSpace[(sizeof(IMAGEHLP_SYMBOL) + MAXSYMNAMELEN)/sizeof(double)+1];
  static IMAGEHLP_SYMBOL     *pSym = (IMAGEHLP_SYMBOL *) SymSpace;
  static char                undName[MAXSYMNAMELEN]; // undecorated name
  static char                undFullName[MAXSYMNAMELEN]; // undecorated name with all shenanigans
  static char                MessageLine[2048];
  static char                *p;
  static char const *        symSearchPath;
  static HANDLE              hProcess;

  if( !DuplicateHandle( GetCurrentProcess(), GetCurrentProcess(), GetCurrentProcess(),
                        &hProcess, DUPLICATE_SAME_ACCESS, TRUE, DUPLICATE_SAME_ACCESS )) 
  {
    hProcess = GetCurrentProcess(); 
  }
  
  if ( !LoadImageHlpDll() )
  {
    CloseHandle(hProcess);
    sp77sprintf(MessageLine, sizeof(MessageLine), "Error loading 'imagehlp.dll' or some required functions:%lu\n", GetLastError() );
    eo670WriteString(MessageLine);
    return;
  }

  ENTER_CRIT_SEC( CS_TRACE_BACK );

  __try
    {
    // build symbol search path from:
    symSearchPath = eo670_GetSymSearchPath();

    if ( VerbosityLevel == VERBOSE_CALL_STACK )
    {
      sp77sprintf(MessageLine, sizeof(MessageLine), "SymbolSearchPath:  %s\n", symSearchPath);
      eo670_write_string_with_displacement(MessageLine, "                   " );
    }
  
    symOptions  = pSymGetOptions();
    symOptions |= SYMOPT_LOAD_LINES;
    symOptions &= ~SYMOPT_UNDNAME;

    pSymSetOptions( symOptions );
  
    if ( ! pSymInitialize( hProcess, symSearchPath, false ) )
    {
      eo670WriteString("SymInitialize() failed\n");
      pSymCleanup( hProcess );
    }

    memset( &sframe, '\0', sizeof(sframe) );
  
    // init STACKFRAME for first call
    // Notes: AddrModeFlat is just an assumption. 
#  if defined (_WIN64)
     sframe.AddrPC.Offset    = 0;
     sframe.AddrPC.Mode      = AddrModeFlat;
     sframe.AddrStack.Offset = 0;
     sframe.AddrStack.Mode   = AddrModeFlat;
     sframe.AddrFrame.Offset = 0;
     sframe.AddrFrame.Mode   = AddrModeFlat;
#  else /* WIN32 only */
     sframe.AddrPC.Offset    = context->Eip;
     sframe.AddrPC.Mode      = AddrModeFlat;
     sframe.AddrStack.Offset = context->Esp;
     sframe.AddrStack.Mode   = AddrModeFlat;
     sframe.AddrFrame.Offset = context->Ebp;
     sframe.AddrFrame.Mode   = AddrModeFlat;
#  endif /* _WIN64 */
  
    memset( pSym, '\0', sizeof(SymSpace));
    pSym->SizeOfStruct  = sizeof(IMAGEHLP_SYMBOL);
    pSym->MaxNameLength = MAXSYMNAMELEN;
  
    memset( &Line, '\0', sizeof(Line) );
    Line.SizeOfStruct   = sizeof(Line);
  
    memset( &Module, '\0', sizeof(Module) );
    Module.SizeOfStruct = sizeof(Module);
  
    offsetFromSymbol = 0;
    offsetFromLine   = 0;

    eo670_FillModuleList (GetCurrentProcessId(), hProcess, VerbosityLevel );

    eo670_ShowLoadedModules();

    eo670WriteString( "----> Stack Back Trace <----\n");

    for ( frameNum = 0; frameNum < 100; ++frameNum )
    {
      // get next stack frame (StackWalk(), SymFunctionTableAccess(), SymGetModuleBase())
      // if this returns ERROR_INVALID_ADDRESS (487) or ERROR_NOACCESS (998), you can
      // assume that either you are done, or that the stack is so hosed that the next
      // deeper frame could not be found.
      if ( ! pStackWalk( imageType, hProcess, hThread, &sframe, context, NULL,
                         pSymFunctionTableAccess, pSymGetModuleBase, NULL ) )
      {
        break;
      }

      // display its contents
      if ( sframe.AddrPC.Offset == 0 )
      {
        strcat( MessageLine, "  <nosymbols> (PC == 0)\n" );
        eo670WriteString(MessageLine);
      }
      else
      { // we seem to have a valid PC

        // show module info (SymGetModuleInfo())
        if ( ! pSymGetModuleInfo( hProcess, sframe.AddrPC.Offset, &Module ) )
        {
          sp77sprintf( MessageLine, sizeof(MessageLine), "  SymGetModuleInfo() failed, rc = %d\n",
                       GetLastError());
          eo670WriteString(MessageLine);
          Module.ModuleName[0] = '?';
          Module.ModuleName[1] = '\0';
        }

        // show procedure info (SymGetSymFromAddr())
        if ( ! pSymGetSymFromAddr( hProcess, sframe.AddrPC.Offset, &offsetFromSymbol, pSym ) )
        {
#  if defined (_WIN64)
           _snprintf  ( MessageLine, (int)sizeof(MessageLine), "  %s!%016I64x <nosymbols>\n", 
                        Module.ModuleName, sframe.AddrPC.Offset );
#  else /* WIN32 only */
           sp77sprintf( MessageLine, (int)sizeof(MessageLine), "  %s!%08lx <nosymbols>\n", 
                        Module.ModuleName, sframe.AddrPC.Offset );
#  endif /* WIN64 */

          eo670WriteString(MessageLine);
        }
        else
        {
          // UnDecorateSymbolName()
          pUnDecorateSymbolName( pSym->Name, undName, MAXSYMNAMELEN, UNDNAME_NAME_ONLY );
          pUnDecorateSymbolName( pSym->Name, undFullName, MAXSYMNAMELEN, UNDNAME_COMPLETE );

          if ( offsetFromSymbol != 0 ) 
          {
#  if defined (_WIN64)
             _snprintf  ( MessageLine, (int)sizeof(MessageLine), "  %s!%s + %I64d bytes\n", 
                          Module.ModuleName, undName, offsetFromSymbol );
#  else /* WIN32 only */
             sp77sprintf( MessageLine, (int)sizeof(MessageLine), "  %s!%s + %ld bytes\n", 
                          Module.ModuleName, undName, offsetFromSymbol );
#  endif /* WIN64 */
          }
          else
            sp77sprintf( MessageLine, (int)sizeof(MessageLine), "  %s!%s\n", Module.ModuleName, undName );

          eo670_write_string_with_displacement(MessageLine, "  " );

          if ( strcmp(pSym->Name, undName) != 0 )
          {
            sp77sprintf(MessageLine, sizeof(MessageLine), "      Deco. :  %s\n", pSym->Name );
            eo670_write_string_with_displacement(MessageLine, "               " );
          }
          if ( strcmp(undFullName, undName) != 0 )
          {
            sp77sprintf(MessageLine, sizeof(MessageLine), "      UDeco.:  %s\n", undFullName );
            eo670_write_string_with_displacement(MessageLine, "               " );
          }
        }

        if ( VerbosityLevel == VERBOSE_CALL_STACK )
        {
#  if defined (_WIN64)
           _snprintf( MessageLine, sizeof(MessageLine), "      SFrame:  IP:%016I64x RA:%016I64x FP:%016I64x SP:%016I64x ",
                      sframe.AddrPC.Offset, sframe.AddrReturn.Offset,
                      sframe.AddrFrame.Offset, sframe.AddrStack.Offset );
           eo670WriteString(MessageLine);

           _snprintf( MessageLine, sizeof(MessageLine), "      Params:  01:%016I64x 02:%016I64x 03:%016I64x 04:%016I64x ",
                      sframe.Params[0], sframe.Params[1], sframe.Params[2], sframe.Params[3] );
#  else /* WIN32 only */
           sp77sprintf( MessageLine, sizeof(MessageLine), "      SFrame:  IP:%08lx RA:%08lx FP:%08lx SP:%08lx ",
                        sframe.AddrPC.Offset, sframe.AddrReturn.Offset,
                        sframe.AddrFrame.Offset, sframe.AddrStack.Offset );
           eo670WriteString(MessageLine);

           sp77sprintf( MessageLine, sizeof(MessageLine), "      Params:  01:%08lx 02:%08lx 03:%08lx 04:%08lx ",
                        sframe.Params[0], sframe.Params[1], sframe.Params[2], sframe.Params[3] );
#  endif /* WIN64 */
          eo670WriteString(MessageLine);
      
          // show line number info
          if ( pSymGetLineFromAddr != NULL )
          { 
            if ( ! pSymGetLineFromAddr( hProcess, sframe.AddrPC.Offset, &offsetFromLine, &Line ) )
            {
              if ( GetLastError() != ERROR_INVALID_ADDRESS )
              {
                sp77sprintf( MessageLine, sizeof(MessageLine), "  SymGetLineFromAddr() failed, rc = %d\n", 
                             GetLastError());
                eo670WriteString(MessageLine);
              }
            }
            else
            {
              p = max(strrchr(Line.FileName, '\\'), strrchr(Line.FileName, '/') );
    
              if ( p != NULL ) { *p = '\0'; p++; }
              else             p = Line.FileName;

              sp77sprintf(MessageLine, sizeof(MessageLine), "      Source:  %s ( Line: %lu )\n", p, Line.LineNumber );
              eo670WriteString(MessageLine);
            }
          }
        }
      } // we seem to have a valid PC
      
      // no return address means no deeper stackframe
      if (( sframe.AddrReturn.Offset == 0 ) || 
          ( sframe.AddrReturn.Offset == sframe.AddrPC.Offset ))
      {
        // avoid misunderstandings in the printf() following the loop
        SetLastError( 0 );
        break;
      }
    } // for ( frameNum )

    if ( GetLastError() != 0 ) 
    {
      sp77sprintf(MessageLine, sizeof(MessageLine), "StackWalk(): rc = %lu\n", GetLastError() );
      eo670WriteString(MessageLine);
    }
  }
  __except( EXCEPTION_EXECUTE_HANDLER ) 
  {
    eo670WriteString( "---> StackWalk aborted! <---\n");
  }
  EXIT_CRIT_SEC( CS_TRACE_BACK );

  if ( GetCurrentProcess() != hProcess ) 
    CloseHandle(hProcess);

  pSymCleanup( hProcess );

  return;
}

/*---------------------------------------------------------------------------*/

static bool eo670_FillModuleListK32( DWORD pid, HANDLE hProcess, int VerbosityLevel )
{
  typedef HANDLE (__stdcall *tCT32S)( DWORD dwFlags, DWORD th32ProcessID );  // CreateToolhelp32Snapshot()
  typedef BOOL (__stdcall *tM32F)( HANDLE hSnapshot, LPMODULEENTRY32 lpme ); // Module32First()
  typedef BOOL (__stdcall *tM32N)( HANDLE hSnapshot, LPMODULEENTRY32 lpme ); // Module32Next()
  
  static HINSTANCE hToolhelp;
  static tCT32S    pCT32S;
  static tM32F     pM32F;
  static tM32N     pM32N;
  
  static HANDLE hSnap;
  static MODULEENTRY32 me = { sizeof(me) };
  static bool keepGoing;
  
  hToolhelp = LoadLibrary( "kernel32.dll" );

  if ( hToolhelp != 0 )
  {
    pCT32S = (tCT32S) myGetProcAddress( hToolhelp, "CreateToolhelp32Snapshot", false );
    pM32F  = (tM32F)  myGetProcAddress( hToolhelp, "Module32First", false );
    pM32N  = (tM32N)  myGetProcAddress( hToolhelp, "Module32Next", false );

    if ( pCT32S == 0 || pM32F == 0 || pM32N == 0 )
    {
      FreeLibrary( hToolhelp );
      return false;
    } 
  
    hSnap = pCT32S( TH32CS_SNAPMODULE, pid );

    if ( hSnap == (HANDLE) -1 )
    {
      FreeLibrary( hToolhelp );
      return false;
    } 
  }
  else
  {
    CloseHandle( hSnap );
    FreeLibrary( hToolhelp );
    return false;
  }

  keepGoing = !!pM32F( hSnap, &me );

  if ( !keepGoing ) 
  {
    CloseHandle( hSnap );
    FreeLibrary( hToolhelp );
    return false;
  }
  
  if ( VerbosityLevel == VERBOSE_CALL_STACK )
    eo670WriteString( "------> Module List <-------\n");

  while ( keepGoing )
  {
    static char MessageBuffer[4096];
    // here, we have a filled-in MODULEENTRY32

    if ( VerbosityLevel == VERBOSE_CALL_STACK )
    {
#  if defined (_WIN64)
       _snprintf( MessageBuffer, sizeof(MessageBuffer), "  %016I64X  %8lu %-15.15s %s\n", me.modBaseAddr, me.modBaseSize, me.szModule, me.szExePath );
       eo670_write_string_with_displacement(MessageBuffer, "                                             " );
#  else /* WIN32 only */
       sp77sprintf( MessageBuffer, sizeof(MessageBuffer), "  %08lX  %8lu %-15.15s %s\n", me.modBaseAddr, me.modBaseSize, me.szModule, me.szExePath );
       eo670_write_string_with_displacement(MessageBuffer, "                                   " );
#  endif /* WIN64 */
    }

    if ( !pSymLoadModule( hProcess, 0, me.szExePath, me.szModule, (tsp00_Longuint)me.modBaseAddr, me.modBaseSize ) ) 
    {
      sp77sprintf( MessageBuffer, sizeof(MessageBuffer), "Error loading symbols for \"%s\", rc = %d\n",
                   me.szModule, GetLastError() );
      eo670WriteString(MessageBuffer);
    } 
    else
    {
        eo670_AddToLoadedModuleChain( me.szModule,
                                      me.modBaseAddr,
             ((char *)me.modBaseAddr)+me.modBaseSize,
                                      me.szModule );
    }
#  if !defined(KERNEL_LZU)
    if ( VerbosityLevel == VERBOSE_CALL_STACK )
    {
      sp77sprintf( MessageBuffer, sizeof(MessageBuffer), "Symbols loaded: \"%s\"\n", me.szModule );
      eo670WriteString(MessageBuffer);
    }
#  endif /* !KERNEL_LZU */
    keepGoing = !!pM32N( hSnap, &me );
  }
  
  CloseHandle( hSnap );
  FreeLibrary( hToolhelp );
  
  return true;
}

/*---------------------------------------------------------------------------*/

#  if !defined (_WIN64)

typedef struct _MODULEINFO 
  {
  LPVOID lpBaseOfDll;
  DWORD  SizeOfImage;
  LPVOID EntryPoint;
  } MODULEINFO, *LPMODULEINFO;


 static bool eo670_FillModuleListPSAPI( DWORD pid, HANDLE hProcess, int VerbosityLevel )
  {
   static char MessageBuffer[4096];
   typedef BOOL (__stdcall *tEPM)( HANDLE hProcess, HMODULE *lphModule, DWORD cb, LPDWORD lpcbNeeded );
   typedef DWORD (__stdcall *tGMFNE)( HANDLE hProcess, HMODULE hModule, LPSTR lpFilename, DWORD nSize );
   typedef DWORD (__stdcall *tGMBN)( HANDLE hProcess, HMODULE hModule, LPSTR lpFilename, DWORD nSize );
   typedef BOOL (__stdcall *tGMI)( HANDLE hProcess, HMODULE hModule, LPMODULEINFO pmi, DWORD nSize );
   
   static HINSTANCE    hPsapi;
   static tEPM         pEPM;
   static tGMFNE       pGMFNE;
   static tGMBN        pGMBN;
   static tGMI         pGMI;
   static int          i;
   static DWORD        cbNeeded;
   static MODULEINFO   mi;
   static HMODULE      hMods[sizeof(HMODULE) * (TTBUFLEN / sizeof(HMODULE))];
   static char         tt[TTBUFLEN];
   static char         ttImage[TTBUFLEN];

   hPsapi = LoadLibrary( "psapi.dll" );

   if ( hPsapi == 0 )
     return false;
   
   pEPM   = (tEPM)   myGetProcAddress( hPsapi, "EnumProcessModules", false );
   pGMFNE = (tGMFNE) myGetProcAddress( hPsapi, "GetModuleFileNameExA", false );
   pGMBN  = (tGMFNE) myGetProcAddress( hPsapi, "GetModuleBaseNameA", false );
   pGMI   = (tGMI)   myGetProcAddress( hPsapi, "GetModuleInformation", false );

   if ( pEPM == 0 || pGMFNE == 0 || pGMBN == 0 || pGMI == 0 )
   {
     FreeLibrary( hPsapi );
     return false;
   }
   
   if ( ! pEPM( hProcess, hMods, TTBUFLEN, &cbNeeded ) )
   {
     sp77sprintf( MessageBuffer, sizeof(MessageBuffer), "EPM failed, rc = %lu\n", GetLastError() );
     eo670WriteString(MessageBuffer);
     FreeLibrary( hPsapi );
     return false;
   }
   
   if ( cbNeeded > TTBUFLEN )
   {
     sp77sprintf( MessageBuffer, sizeof(MessageBuffer), "Error more than %lu module handles. \n", (TTBUFLEN/sizeof(HMODULE)) );
     eo670WriteString(MessageBuffer);
     FreeLibrary( hPsapi );
     return false;
   }
   
   if ( VerbosityLevel == VERBOSE_CALL_STACK )
     eo670WriteString( "------> Module List <-------\n");

   for ( i = 0; i < (int)(cbNeeded / sizeof(HMODULE)); ++ i )
   {
     ttImage[0] = '\0';
     tt[0]      = '\0';

     pGMI  ( hProcess, hMods[i], &mi, sizeof(mi) );
     pGMFNE( hProcess, hMods[i], ttImage, TTBUFLEN );
     pGMBN ( hProcess, hMods[i], tt, TTBUFLEN );
     
     if ( !pSymLoadModule( hProcess, 0, ttImage, tt, (tsp00_Longuint)mi.lpBaseOfDll, mi.SizeOfImage ) ) 
     {
       sp77sprintf( MessageBuffer, sizeof(MessageBuffer), "Error loading symbols for \"%s\", rc = %d\n", tt, GetLastError() );
       eo670WriteString(MessageBuffer);
     } 
     else 
     {
         eo670_AddToLoadedModuleChain( tt,
                                       mi.lpBaseOfDll,
              ((char *)mi.lpBaseOfDll)+mi.SizeOfImage,
                                       tt );
         if ( VerbosityLevel == VERBOSE_CALL_STACK )
         {
#    if defined(KERNEL_LZU)
          sp77sprintf( MessageBuffer, sizeof(MessageBuffer), "  %08lX  %8lu %-15.15s\n", mi.lpBaseOfDll, mi.SizeOfImage, tt );
          eo670_write_string_with_displacement(MessageBuffer, "                                   " );
#    else
          sp77sprintf( MessageBuffer, sizeof(MessageBuffer), "Symbols loaded: \"%s\"\n", tt );
          eo670WriteString(MessageBuffer);
#    endif /* KERNEL_LZU */
         }
     }
   }
   FreeLibrary( hPsapi );
   return true;
 }
#  endif /* !WIN64 */

/*---------------------------------------------------------------------------*/

static bool eo670_FillModuleList( DWORD pid, HANDLE hProcess, int VerbosityLevel )
{
   bool Ok;

#  if defined (_WIN64)
     Ok = eo670_FillModuleListK32 (GetCurrentProcessId(), hProcess, VerbosityLevel ); // Windows XP...
#  else
     OSVERSIONINFO      VerInfo;

     memset ( &VerInfo, 0, sizeof(VerInfo) );
     VerInfo.dwOSVersionInfoSize = sizeof(VerInfo);

     if ( GetVersionEx(&VerInfo) )
     {
       if ( VerInfo.dwMajorVersion < 5 )
         Ok = eo670_FillModuleListPSAPI (GetCurrentProcessId(), hProcess, VerbosityLevel );  // Windows NT
       else
         Ok = eo670_FillModuleListK32 (GetCurrentProcessId(), hProcess, VerbosityLevel );    // Windows 2000, Windows XP...
     }
#  endif /* WIN64 */

  return Ok;
}

/*---------------------------------------------------------------------------*/

static char const * const eo670_GetSymSearchPath()
{
  static char              *p;
  static char              tt[TTBUFLEN];
  static char              symSearchPath[4096];


  if ( symSearchPath[0] != '\0' )
    return symSearchPath;

  // current directory
  if ( GetCurrentDirectory( TTBUFLEN, tt ) ) 
  {
    strcat(symSearchPath, tt);
    strcat(symSearchPath, ";");
  }
  // dir with executable
  if ( GetModuleFileName( 0, tt, TTBUFLEN ) )
  {
    for ( p = tt + strlen( tt ) - 1; p >= tt; -- p )
    {
      // locate the rightmost path separator
      if ( *p == '\\' || *p == '/' || *p == ':' )
        break;
    }
    // if we found one, p is pointing at it; if not, tt only contains
    // an exe name (no path), and p points before its first byte
    if ( p != tt ) // path sep found?
    {
      if ( *p == ':' ) // we leave colons in place
        ++ p;
      *p = '\0'; // eliminate the exe name and last path sep
      strcat(symSearchPath, tt);
      strcat(symSearchPath, ";");
    }
  }
  // environment variable _NT_SYMBOL_PATH
  if ( GetEnvironmentVariable( "_NT_SYMBOL_PATH", tt, TTBUFLEN ) ) {
    strcat(symSearchPath, tt);
    strcat(symSearchPath, ";");
  }
  // environment variable _NT_ALTERNATE_SYMBOL_PATH
  if ( GetEnvironmentVariable( "_NT_ALTERNATE_SYMBOL_PATH", tt, TTBUFLEN ) ) {
    strcat(symSearchPath, tt);
    strcat(symSearchPath, ";");
  }
  // environment variable _NT_ALT_SYMBOL_PATH
  if ( GetEnvironmentVariable( "_NT_ALT_SYMBOL_PATH", tt, TTBUFLEN ) ) {
    strcat(symSearchPath, tt);
    strcat(symSearchPath, ";");
  }
  // environment variable SYSTEMROOT
  if ( GetEnvironmentVariable( "SYSTEMROOT", tt, TTBUFLEN ) ) {
    strcat(symSearchPath, tt);
    strcat(symSearchPath, ";");
  }
  
  if ( strlen(symSearchPath) > 0 ) // if we added anything, we have a trailing semicolon
    symSearchPath[strlen(symSearchPath)-1] = 0;
  
#  if defined(KERNEL_LZU)
  {
    static tsp00_Pathc       szInstPath;
    static tsp01_RteError    rteError;
    if ( sqlGetDbrootSapPath ( szInstPath, TERM_WITH_DELIMITER_EO01, &rteError )) 
    {
      strcat(symSearchPath, ";");
      strcat(symSearchPath, &szInstPath[0]); 
    }
  }
#  endif /* KERNEL_LZU */

  return symSearchPath;
}

/*---------------------------------------------------------------------------*/

#ifdef HAS_SUPPORT_FOR_MINI_DUMP
/*
// A user memory dump is written if a Microsoft 'dbghelp.dll' is found 
// This dump can be  evaluated via 'windbg'.
//
// CAUTION: The dbghelp.dll version must be greater or equal 5.1!
*/
bool eo670_WriteDump ( bool WriteFullMemory, bool IfDumpFileExists )
{
  static char MessageBuffer[4096];
  typedef BOOL (__stdcall *tMiniDumpWriteDump)( HANDLE                             hProcess,
                                                DWORD                              ProcessId,
                                                HANDLE                             hFile,
                                                MINIDUMP_TYPE                      DumpType,
                                                PMINIDUMP_EXCEPTION_INFORMATION    ExceptionParam,
                                                PMINIDUMP_USER_STREAM_INFORMATION  UserStreamParam,
                                                PMINIDUMP_CALLBACK_INFORMATION     CallbackParam );

  HANDLE                     hDbgHelp;
  HANDLE                     hFile;
  tMiniDumpWriteDump         pMiniDumpWriteDump;
  char*                      pszFile = "knlcrash.dmp";
  DWORD                      pid     = GetCurrentProcessId(); 
  bool                       Ok      = true;
  HANDLE                     hProcess;

  if( !DuplicateHandle( GetCurrentProcess(), GetCurrentProcess(), GetCurrentProcess(),
                        &hProcess, DUPLICATE_SAME_ACCESS, TRUE, DUPLICATE_SAME_ACCESS )) 
  {
    hProcess = GetCurrentProcess(); 
  }

  hDbgHelp = LoadLibrary( "dbghelp.dll" );  // use dll version >=  5.1

  if ( hDbgHelp == 0 )
  {
    CloseHandle(hProcess);
    return false;
  }

  pMiniDumpWriteDump = (tMiniDumpWriteDump)myGetProcAddress( hDbgHelp, "MiniDumpWriteDump", true );

  if ( pMiniDumpWriteDump == 0 )
  {
    MSGD (( MSG_ID + 1, 0, "INFO    ", "DBGHELP is not suporting 'MiniDumpWriteDump' " ));
    CloseHandle(hProcess);
    FreeLibrary( hDbgHelp );
    return false;
  }

  ENTER_CRIT_SEC( CS_WRITE_DUMP );

  __try
  {

    if ( IfDumpFileExists )
      hFile = CreateFile( pszFile, GENERIC_WRITE, FILE_SHARE_READ, NULL, 
                          OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    else
      hFile = CreateFile( pszFile, GENERIC_WRITE, FILE_SHARE_READ, NULL, 
                          CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);

    if ( hFile == INVALID_HANDLE_VALUE )
    {
      if (( IfDumpFileExists ) && ( GetLastError() == ERROR_FILE_NOT_FOUND ))
      {
        MSGD (( MSG_ID + 1, 0, "INFO    ", "No crash dump file '%s' written ", pszFile  ))
      }
      else
      {
        sp77sprintf( MessageBuffer, sizeof(MessageBuffer), "Error creating mini dump file \"%s\", rc = %d\n", pszFile, GetLastError() );
        eo670WriteString(MessageBuffer);
      }
      Ok = false;
    }
    else
    {
      if ( !pMiniDumpWriteDump( hProcess, pid, hFile, 
                                MiniDumpWithHandleData | 
                                MiniDumpWithDataSegs   |
                                WriteFullMemory ? MiniDumpWithFullMemory : 0,
                                NULL, NULL, NULL ))
      {
        sp77sprintf( MessageBuffer, sizeof(MessageBuffer), "Error writing mini dump, rc = %d\n", GetLastError() );
        eo670WriteString(MessageBuffer);
        Ok = false;
      }
      else
        MSGD (( MSG_ID + 1, 0, "INFO    ", "Crash dump file written!" ));
    }
  }
  __except( EXCEPTION_EXECUTE_HANDLER ) 
  {
    eo670WriteString( "---> Write 'minidump' aborted! <---\n");
  }
  EXIT_CRIT_SEC( CS_WRITE_DUMP );

  CloseHandle(hProcess);
  CloseHandle( hFile );   
  FreeLibrary( hDbgHelp );

  return Ok;
}
#endif /* HAS_SUPPORT_FOR_MINI_DUMP */
#endif /* WIN32 */

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/
#endif /* HAS_STACK_BACKTRACE_SUPPORT */

