
/*!**********************************************************************

  module: vls31_odbc_Result.cpp

  -----------------------------------------------------------------------

  responsible:  Hans-GeorgB, SteffenS
  special area: odbc
  description:
  version    : 7.3
  see also:

  -----------------------------------------------------------------------

   ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end
*/

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gls00.h"
#include "hls30.h"
#include "hls99.h"
#include "hsp100.h"
#include "hls31_odbc_Result.h"


#define SEPARATOR   ""CMD_SEPARATOR_TOKEN_LS00""NEW_LINE""

_TCHAR const *pszSQLModeString_ls31[6] = {_T("NOT USED"),
                                          _T("SQLMODE INTERNAL"NEW_LINE""),
                                          _T("SQLMODE DB2"NEW_LINE""),
                                          _T("SQLMODE ANSI"NEW_LINE""),
                                          _T("SQLMODE ORACLE"NEW_LINE""),
                                          _T("SQLMODE SAPR3"NEW_LINE"")};


/*===========================================================================*
 *    sapdb_Error::sapdb_Error()
 *===========================================================================*/
sapdb_Error::sapdb_Error()
{
    m_isError    = SAPDB_FALSE;
    m_NativeError= 0;
    m_ErrorPos   = 0;
    m_ReturnCode = 0;
    // m_ErrorText;
    // m_NativeErrorAsString;

}
/*===========================================================================*
 *    ~sapdb_Error::sapdb_Error()
 *===========================================================================*/
sapdb_Error::~sapdb_Error()
{
}
/*===========================================================================*
 *     CatalogDefODBC::CatalogDefODBC()
 *===========================================================================*/
CatalogDefODBC::CatalogDefODBC()
{
    // Trace - Writer
    ROUTINE_DBG_MEO00 ("CatalogDefODBC::CatalogDefODBC()");

    // Log   - Writer
    RepServMessages *pMsgInstance  = RepServMessages::Instance();

    // Meta Data
    m_pCatalogExtractDefODBC   = NULL;
    m_pCatalogLoadDefODBC      = NULL;

    m_pUserModeDefODBC         = NULL;
    m_pCURUserDefODBC          = NULL;

    m_pUserDefODBC             = NULL;
    m_pRoleDefODBC             = NULL;
    m_pTableDefODBC            = NULL;
    m_pViewDefODBC             = NULL;
    m_pSynonymDefODBC          = NULL;
    m_pTriggerDefODBC          = NULL;
    m_pDBProcedureDefODBC      = NULL;
    m_pGrantDBProcedureDefODBC = NULL;
    m_pColumnTypeDefODBC       = NULL;
    m_pIndexDefODBC            = NULL;
    m_pSequenceDefODBC         = NULL;
    m_pError                   = NULL;

    m_nSchemaCount             = 0;
    m_nTableCount              = 0;
    m_nSynonymCount            = 0;
    m_nViewCount               = 0;

    // Data
    m_pDataExtractDefODBC      = NULL;

}
/*===========================================================================*
 *     CatalogDefODBC::~CatalogDefODBC
 *===========================================================================*/
CatalogDefODBC::~CatalogDefODBC()
{
    ROUTINE_DBG_MEO00 ("CatalogDefODBC::~CatalogDefODBC()");

    RepServMessages *pMsgInstance = RepServMessages::Instance();

    // Meta Data

    if( m_pCatalogExtractDefODBC  != NULL )
    {
        delete m_pCatalogExtractDefODBC;
        m_pCatalogExtractDefODBC = NULL;
    }

    if ( m_pCatalogLoadDefODBC    != NULL )
    {
        delete m_pCatalogLoadDefODBC;
        m_pCatalogLoadDefODBC = NULL;
    }

    if ( m_pUserModeDefODBC      != NULL )
    {
         delete m_pUserModeDefODBC;
         m_pUserModeDefODBC = NULL;
    }
    if ( m_pCURUserDefODBC  != NULL )
    {
         delete m_pCURUserDefODBC;
         m_pCURUserDefODBC = NULL;
    }
    if ( m_pUserDefODBC       != NULL )
    {
        delete m_pUserDefODBC;
        m_pUserDefODBC = NULL;
    }
    if ( m_pRoleDefODBC       != NULL )
    {
        delete m_pRoleDefODBC;
        m_pRoleDefODBC = NULL;

    }
    if( m_pTableDefODBC      != NULL )
    {
        delete m_pTableDefODBC;
        m_pTableDefODBC = NULL;
    }
    if( m_pViewDefODBC       != NULL )
    {
        delete m_pViewDefODBC;
        m_pViewDefODBC  = NULL;
    }
    if( m_pSynonymDefODBC    != NULL )
    {
        delete m_pSynonymDefODBC;
        m_pSynonymDefODBC = NULL;
    }
    if( m_pTriggerDefODBC    != NULL )
    {
        delete m_pTriggerDefODBC;
        m_pTriggerDefODBC = NULL;
    }
    if( m_pColumnTypeDefODBC != NULL )
    {
        delete m_pColumnTypeDefODBC;
        m_pColumnTypeDefODBC = NULL;
    }
    if( m_pIndexDefODBC      != NULL )
    {
        delete m_pIndexDefODBC ;
        m_pIndexDefODBC = NULL;
    }
    if( m_pSequenceDefODBC   != NULL )
    {
        delete m_pSequenceDefODBC;
        m_pSequenceDefODBC = NULL;
    }
    if( m_pDBProcedureDefODBC != NULL)
    {
        delete m_pDBProcedureDefODBC;
        m_pDBProcedureDefODBC = NULL;
    }
    if( m_pTriggerDefODBC != NULL )
    {
        delete m_pTriggerDefODBC;
        m_pTriggerDefODBC = NULL;
    }
    if ( m_pError != NULL )
    {
        delete m_pError;
        m_pError = NULL;
    }

    // Data
    if ( m_pDataExtractDefODBC != NULL )
    {
        delete m_pDataExtractDefODBC;
        m_pDataExtractDefODBC = NULL;
    }
}
/*===========================================================================*
 *     CatalogDefODBC::initCatalogObjects                                    *
 *===========================================================================*/
SAPDB_Bool CatalogDefODBC::initCatalogObjects( sapdbwa_DBC & odbcConnection)
{
    ROUTINE_DBG_MEO00 ("CatalogDefODBC::initCatalogObjects");

    RepServMessages *pMsgInstance = RepServMessages::Instance();

    m_pError                 = new sapdb_Error();
#ifdef MEMTEST
    if ( m_pError != NULL ) 
    {
        delete m_pError;
        m_pError = NULL;
    }
#endif

    if ( getExtrMode()    ==  emLoadAsXML_els31   || 
         getExtrMode()    ==  emLoadAsDDL_els31     )
    {
         m_pCatalogLoadDefODBC   = new CatalogLoadDefODBC   ( odbcConnection, this );
#ifdef MEMTEST
         if ( m_pCatalogLoadDefODBC    != NULL ) 
         {
              delete m_pCatalogLoadDefODBC;
              m_pCatalogLoadDefODBC = NULL;
         }
#endif
         // Authorization
         m_pUserModeDefODBC       = new UserModeDefODBC      ( odbcConnection, this );
#ifdef MEMTEST_1
        if ( m_pUserModeDefODBC != NULL ) 
        {
             delete m_pUserModeDefODBC;
             m_pUserModeDefODBC = NULL;
        }
#endif
    }

    if ( getExtrMode()    ==  emExtractAsXML_els31 ||
         getExtrMode()    ==  emExtractAsDDL_els31     )
    {

        // Authorization
        m_pUserModeDefODBC       = new UserModeDefODBC      ( odbcConnection, this );
#ifdef MEMTEST_1
        if ( m_pUserModeDefODBC != NULL ) 
        {
             delete m_pUserModeDefODBC;
             m_pUserModeDefODBC = NULL;
        }
#endif
        m_pCURUserDefODBC        = new CURUserDefODBC       ( odbcConnection, this );
#ifdef MEMTEST
        if ( m_pCURUserDefODBC  != NULL ) 
        {
             delete m_pCURUserDefODBC;
             m_pCURUserDefODBC = NULL;
        }
#endif

        // ResultSet of MetaData
        m_pCatalogExtractDefODBC = new CatalogExtractDefODBC( odbcConnection, this );
#ifdef MEMTEST
        if( m_pCatalogExtractDefODBC  != NULL ) 
        {
            delete m_pCatalogExtractDefODBC;
            m_pCatalogExtractDefODBC = NULL;
        }
#endif

        // MetaData
        m_pUserDefODBC           = new UserDefODBC          ( odbcConnection, this );
#ifdef MEMTEST
        if ( m_pUserDefODBC != NULL )       
        {
             delete m_pUserDefODBC;
             m_pUserDefODBC = NULL;
        }
#endif
        m_pRoleDefODBC           = new RoleDefODBC          ( odbcConnection, this );
#ifdef MEMTEST
        if ( m_pRoleDefODBC       != NULL ) 
        {
             delete m_pRoleDefODBC;
             m_pRoleDefODBC = NULL;
        }
#endif
        m_pTableDefODBC          = new TableDefODBC         ( odbcConnection, this );
#ifdef MEMTEST
        if( m_pTableDefODBC      != NULL ) 
        {
            delete m_pTableDefODBC;
            m_pTableDefODBC = NULL;
        }
#endif
        m_pViewDefODBC           = new ViewDefODBC          ( odbcConnection, this );
#ifdef MEMTEST
        if( m_pViewDefODBC       != NULL ) 
        {
            delete m_pViewDefODBC;
            m_pViewDefODBC  = NULL;
        }
#endif
        m_pSynonymDefODBC        = new SynonymDefODBC       ( odbcConnection, this );
#ifdef MEMTEST
        if( m_pSynonymDefODBC    != NULL ) 
        {
            delete m_pSynonymDefODBC;
            m_pSynonymDefODBC = NULL;
        }
#endif
        m_pTriggerDefODBC        = new TriggerDefODBC       ( odbcConnection, this );
#ifdef MEMTEST
        if( m_pTriggerDefODBC    != NULL ) 
        {
            delete m_pTriggerDefODBC;
            m_pTriggerDefODBC = NULL;
        }
#endif
        m_pDBProcedureDefODBC    = new DBProcedureDefODBC   ( odbcConnection, this );
#ifdef MEMTEST
        if( m_pDBProcedureDefODBC != NULL) 
        {
            delete m_pDBProcedureDefODBC;
            m_pDBProcedureDefODBC = NULL;
        }
#endif
        m_pColumnTypeDefODBC     = new ColumnTypeDefODBC    ( odbcConnection, this );
#ifdef MEMTEST
        if( m_pColumnTypeDefODBC != NULL ) 
        {
            delete m_pColumnTypeDefODBC;
            m_pColumnTypeDefODBC = NULL;
        }
#endif
        m_pIndexDefODBC          = new IndexDefODBC         ( odbcConnection, this );
#ifdef MEMTEST
        if( m_pIndexDefODBC      != NULL ) 
        {
            delete m_pIndexDefODBC ;
            m_pIndexDefODBC = NULL;
        }
#endif
        m_pSequenceDefODBC       = new SequenceDefODBC      ( odbcConnection, this );
#ifdef MEMTEST
        if( m_pSequenceDefODBC   != NULL ) 
        {
            delete m_pSequenceDefODBC;
            m_pSequenceDefODBC = NULL;
        }
#endif
        
        // Data
        m_pDataExtractDefODBC    = new DataExtractDefODBC   ( odbcConnection, this );
#ifdef MEMTEST
        if ( m_pDataExtractDefODBC != NULL ) 
        {
             delete m_pDataExtractDefODBC;
             m_pDataExtractDefODBC = NULL;
        }
#endif

        // 1. case : catalogextract  table     <-->  table does not exist
        // 2. case : catalogextract  all       <-->  user  has  not dba privileges
        // 3. case : dataextract for dataload  <-->  tables, indices extracted direct
#ifdef MEMTEST_1        
        if ( m_pCatalogExtractDefODBC->getObjectCount() == 0 )
             return SAPDB_FALSE;
#endif
    }

    return SAPDB_TRUE;
}
/*===========================================================================*
 *     CatalogDefODBC::writeCatalogHeader
 *===========================================================================*/
SAPDB_Bool CatalogDefODBC::writeCatalogHeader ()
{
    SAPDB_Bool isOk = SAPDB_FALSE;

    if ( getExtrMode() == emExtractAsXML_els31 )
    {
    }
    if ( getExtrMode() == emExtractAsDDL_els31 )
    {
    }
    return isOk;
}

/*===========================================================================*
 *     CatalogDefODBC::writeCatalogFooter
 *===========================================================================*/
SAPDB_Bool CatalogDefODBC::writeCatalogFooter ()
{
    SAPDB_Bool isOk = SAPDB_FALSE;

    StudioOAL_String catalogObjects;

    if ( getExtrMode() == emExtractAsXML_els31 )
    {
        catalogObjects  = "</uml:NamespaceOwnedElement>\n"
                          "</sapdbm:DeployedSchema>\n"
                          "</uml:NamespaceOwnedElement>\n"
                          "</dbm:CatalogSchemas>\n"
                          "</dbm:DeployedCatalog>\n"
                          "</dbm:DataSourceDeployedCatalogs>\n"
                          "</dbm:DataSource>\n"
                          "</uml:ModelPackages>\n"
                          "</uml:Model>\n"
                          "</uml:SystemModels>\n"
                          "</uml:System>\n"
                          "</oim:Transfer>\n";
    }
    if ( getExtrMode() == emExtractAsDDL_els31 )
         catalogObjects = SEPARATOR;

    return writeLine ( catalogObjects );
}
/*===========================================================================*
 *     CatalogDefODBC::setODBCDriver
 *===========================================================================*/
SAPDB_Bool CatalogDefODBC::setODBCDriver( tls00_DBInfo *pDBSession )
{
      ROUTINE_DBG_MEO00 ("CatalogDefODBC::setODBCDriver");

      RepServMessages *pMsgInstance = RepServMessages::Instance();

      SAPDB_Bool isOk = SAPDB_FALSE;

      StudioOAL_String odbcDriver;

      if (  pDBSession->dbiDBVersion_ls00.MajorVersion_sp100 == '7' &&
            pDBSession->dbiDBVersion_ls00.MinorVersion_sp100 == '2'    )
      {
            m_odbcDriver = "LiveCache 7.2";
            isOk         = SAPDB_TRUE;
      }

      if  ( pDBSession->dbiDBVersion_ls00.MajorVersion_sp100 == '7' &&
            pDBSession->dbiDBVersion_ls00.MinorVersion_sp100 == '3'    )
      {
            m_odbcDriver = "SAP DB 7.3";
            isOk         = SAPDB_TRUE;
      }
      if  ( pDBSession->dbiDBVersion_ls00.MajorVersion_sp100 == '7' &&
            pDBSession->dbiDBVersion_ls00.MinorVersion_sp100 == '4'    )
      {
            m_odbcDriver = "SAP DB 7.4";
            isOk         = SAPDB_TRUE;
      }

      return isOk;
}
/*===========================================================================*
 *     CatalogDefODBC::Connect
 *===========================================================================*/
SAPDB_Bool CatalogDefODBC::Connect( tls00_DBInfo *pDBSession )
{
    ROUTINE_DBG_MEO00 ("CatalogDefODBC::Connect");

    RepServMessages *pMsgInstance = RepServMessages::Instance();

    setODBCDriver( pDBSession );

    /*-------------------------------------------------------------------------*
     * Connect to the database and set the connection attributes of the table. *
     *-------------------------------------------------------------------------*/

    // SQLSetConnectAttr

    // Mapping of oi tsp1_sqlmode_Enum to odbc tpr00_SQLModeEnum
    switch(pDBSession->dbiSession_ls00.siSQLMode_ls00)
    {
        case sp1sm_db2:
        {
//                 m_odbcConnection.SetAttr( SQL_SQLMODE  , cpr_kind_db2      );
            setSqlMode(cpr_kind_db2);
            break;
        }
        case sp1sm_ansi:
        {
//                 m_odbcConnection.SetAttr( SQL_SQLMODE  , cpr_kind_ansi     );
            setSqlMode(cpr_kind_ansi);
            break;
        }
        case sp1sm_oracle:
        {
//                 m_odbcConnection.SetAttr( SQL_SQLMODE  , cpr_kind_oracle   );
            setSqlMode(cpr_kind_oracle);
            break;
        }
        case sp1sm_adabas:       
        default:
        {
//                 m_odbcConnection.SetAttr( SQL_SQLMODE  , cpr_kind_internal );
            setSqlMode(cpr_kind_internal);
        }
    }     

    setMaxErrorCount(pDBSession->dbiSession_ls00.siMaxAdmissibleErrors);

    // privileges to execute used commands
    if ( m_odbcConnection.AddAttr( SQL_PRODUCER , sp1pr_internal_cmd ) &&
         m_odbcConnection.AddAttr( SQL_COMPONENT, sql_csp_comp_load  )  )
    {
        tsp00_NodeIdc   siDBNodec;
        siDBNodec.p2c(pDBSession->dbiSession_ls00.siDBNode_ls00);

        if ( m_odbcConnectionPool.Connect(m_odbcConnection, NULL,( const SAPDB_Char*) getOdbcDriver().StrPtr() ,
                                          siDBNodec,
                                          pDBSession->dbiSession_ls00.siDBName_ls00,
                                          pDBSession->dbiSession_ls00.siUserName_ls00,
                                          pDBSession->dbiSession_ls00.siUserPwd_ls00 ,
                                          NULL,0,0, sapdbwa_DBC_EXCL ) )
        {
            setODBCConnectionPool( m_odbcConnectionPool );
            setODBCConnection    ( m_odbcConnection );

#ifdef MEMTEST
            delete this;
#endif

            if ( ! initCatalogObjects( m_odbcConnection ) )
                return SAPDB_FALSE;
        }
        else
        {
            sapdbwa_Err Err            = getODBCConnection().GetErr();

            // NULL terminated
            const SAPDB_Char * ErrText = Err.GetErrMsg();

            TRACE_PRNF_MLS99(("failed : ", "%s\n",  ErrText ));

            if ( getExtrMode() == emExtractAsDDL_els31 )
                pMsgInstance->ls98PrintMessage(layCatalogExtract_els98, errODBCErrors_els98, ErrText );
            if ( getExtrMode() == emLoadAsDDL_els31 )
                pMsgInstance->ls98PrintMessage(layCatalogLoad_els98   , errODBCErrors_els98, ErrText );

            m_pError->setErrorText( StudioOAL_String(ErrText) );
            m_pError->isError     ( SAPDB_TRUE );

            return SAPDB_FALSE;
        }
    }
    return SAPDB_TRUE;
}
/*===========================================================================*
 *     CatalogDefODBC::Disconnect
 *===========================================================================*/
SAPDB_Bool CatalogDefODBC::Disconnect( )
{
    ROUTINE_DBG_MEO00 ("CatalogDefODBC::Disconnect");

    RepServMessages *pMsgInstance = RepServMessages::Instance();

    SAPDB_Bool isOk = SAPDB_FALSE;

    if ( m_odbcConnectionPool.Disconnect ( m_odbcConnection ) )
         isOk = SAPDB_TRUE;
    else
    {
         sapdbwa_Err Err            = m_odbcConnection.GetErr();

         // NULL terminated

         const SAPDB_Char * ErrText = Err.GetErrMsg();

         TRACE_PRNF_MLS99(("failed : ", "%s\n", ErrText ));

         if ( getExtrMode() == emExtractAsDDL_els31 )
              pMsgInstance->ls98PrintMessage(layCatalogExtract_els98, errODBCErrors_els98, ErrText );
         if ( getExtrMode() == emLoadAsDDL_els31 )
              pMsgInstance->ls98PrintMessage(layCatalogLoad_els98   , errODBCErrors_els98, ErrText );

         m_pError->setErrorText( StudioOAL_String(ErrText) );
         m_pError->isError     ( SAPDB_TRUE );

         isOk = SAPDB_FALSE;
    }
    return isOk;
}
/*===========================================================================*
 *     CatalogDefODBC::setODBCConnection                                         *
 *===========================================================================*/
SAPDB_Bool CatalogDefODBC::setODBCConnection ( sapdbwa_DBC & odbcConnection )
{
    ROUTINE_DBG_MEO00 ("CatalogDefODBC::setODBCConnection");

    if ( odbcConnection.HDBC() && odbcConnection.HENV() )
         m_odbcConnection = odbcConnection;

    return SAPDB_TRUE;
}
/*===========================================================================*
 *  CatalogDefODBC::setCurrentUser
 *===========================================================================*/
void CatalogDefODBC::setCurrentUser(StudioOAL_String username )
{
    ROUTINE_DBG_MEO00 ("CatalogDefODBC::setCurrentUser");

    if ( (SAPDB_TRUE == getPartSpec().psSelAll_ls00) &&
         (0          != username.Compare( m_currentUser ) ) )
    {
        // catalog extract/load all

        StudioOAL_String  currentUserCmd;
        StudioOAL_String  currentUser;
        StudioOAL_String  currentPassword = "initial ";
        StudioOAL_String  dbaUser;

        /*-------------------------------------------------------------------------*
         * Prepare the security context for the SQL statements.
         *-------------------------------------------------------------------------*/
        currentUserCmd  = "CURRENT USER ";
        currentUserCmd  += username;

        /*-------------------------------------------------------------------------------*
         * Read/Write the catalog definition of the specified object from/to resources   *
         *-------------------------------------------------------------------------------*/

        if ( m_pCURUserDefODBC->executeSQL( currentUserCmd ) )
        {
            m_currentUser = username;

            if (emExtractAsDDL_els31 == getExtrMode())
            {
                currentUser = SEPARATOR;

                if ( getSqlMode() != cpr_kind_internal )
                {
                    currentUser += pszSQLModeString_ls31[cpr_kind_internal];
                    currentUser += SEPARATOR;
                }

                dbaUser = getUserModeDefODBC()->getUserNameDBA();

                if ( dbaUser.Compare( username ) == 0 )
                {
                    SAPDB_Char       destBeg[1000] = "";
                    SAPDB_UInt4      destAt        = 1000;
                    currentPassword.Erase();
                    if (destBeg != NULL)
                    {
                        memset(destBeg,0,1000);
                    }
                    m_odbcConnection.GetPassword( REINTERPRET_CAST(unsigned char*,destBeg ),destAt);
                    currentPassword.ConvertFromASCII_Latin1(destBeg,destBeg+strlen(destBeg));
                }

                currentUser += "USE USER ";
                currentUser += username;
                currentUser += " ";
                currentUser += currentPassword;

                m_pCURUserDefODBC->writeLine( currentUser );
            }
        }
    }
    else if (SAPDB_TRUE == getPartSpec().psSelUser_ls00)
    {
        // catalog extract/load user
        m_currentUser = username;
    }
    else
    {
        // catalog extract/load table
        if ( (0 != getPartSpec().psTabName_ls00.length()) &&
             (0 != getPartSpec().psTabOwnr_ls00.length()) )
        {
            m_currentUser = username;
        }
    }
}
/*===========================================================================*
 *  CatalogDefODBC::getCurrentUser
 *===========================================================================*/
StudioOAL_String CatalogDefODBC::getCurrentUser()
{
    return m_currentUser;
}
/*===========================================================================*
 *  CatalogDefODBC::ExtrCatalogCmd                                           *
 *===========================================================================*/
SAPDB_Int4 CatalogDefODBC::ExtrCatalogCmd( StudioOAL_String username )
{
     ROUTINE_DBG_MEO00 ("CatalogDefODBC::ExtrAllCmd");

     RepServMessages *pMsgInstance = RepServMessages::Instance();

     SAPDB_Int4 rc = errOK_els00;

     // Extract Connection Object
     connectionExtrCmd( );

     // Extract Authorization Objects, schema and privilegies
     userExtrCmd  ( getPartSpec() );
     roleExtrCmd  ( getPartSpec() );

     // Extract Catalog Objects
     tabletypeExtrCmd  ( getPartSpec() );
     sequenceExtrCmd   ( getPartSpec() );
     dbprocedureExtrCmd( getPartSpec() );

     writeCatalogFooter();

     // Statistic of Extracted Catalog Objects
     pMsgInstance->ls98PrintMessage(layCatalogExtract_els98, msgNumTabExtracted_els98     , m_nTableCount   );
     pMsgInstance->ls98PrintMessage(layCatalogExtract_els98, msgNumViewsExtracted_els98   , m_nViewCount    );
     pMsgInstance->ls98PrintMessage(layCatalogExtract_els98, msgNumSynonymsExtracted_els98, m_nSynonymCount );

     setCurrentUser( username );

     return rc;
}
/*===========================================================================*
 *  CatalogDefODBC::ExtrSchemaCmd                                            *
 *===========================================================================*/
SAPDB_Int4 CatalogDefODBC::ExtrSchemaCmd(  StudioOAL_String username )
{
     ROUTINE_DBG_MEO00 ("CatalogDefODBC::ExtrSchemaCmd");

     RepServMessages *pMsgInstance = RepServMessages::Instance();

     SAPDB_Int4 rc = errOK_els00;

     setCurrentUser    ( username );

     // Extract Connection Object
     connectionExtrCmd  ( );

     // Extract Authorization Objects, schema and privilegies
     userExtrCmd  ( getPartSpec() );
     roleExtrCmd  ( getPartSpec() );

     // Extract Catalog Objects
     tabletypeExtrCmd  ( getPartSpec() );
     sequenceExtrCmd   ( getPartSpec() );
     dbprocedureExtrCmd( getPartSpec() );

     writeCatalogFooter();

     // Statistic of Extracted Catalog Objects
     pMsgInstance->ls98PrintMessage(layCatalogExtract_els98, msgNumTabExtracted_els98     , m_nTableCount   );
     pMsgInstance->ls98PrintMessage(layCatalogExtract_els98, msgNumViewsExtracted_els98   , m_nViewCount    );
     pMsgInstance->ls98PrintMessage(layCatalogExtract_els98, msgNumSynonymsExtracted_els98, m_nSynonymCount );

     return rc;
}
/*===========================================================================*
 *  CatalogDefODBC::ExtrTableCmd
 *===========================================================================*/
SAPDB_Int4 CatalogDefODBC::ExtrTableCmd(tls00_PartSpec catpartSpec)
{
     ROUTINE_DBG_MEO00 ("CatalogDefODBC::ExtrTableCmd");

     RepServMessages *pMsgInstance = RepServMessages::Instance();

     SAPDB_Int4 rc = errOK_els00;

     // Extract Connection Object
     connectionExtrCmd();

     // Extract TableTypes
     tabletypeExtrCmd ( catpartSpec );

     writeCatalogFooter();
     
     // Statistic of Extracted Catalog Objects
     pMsgInstance->ls98PrintMessage(layCatalogExtract_els98, msgNumTabExtracted_els98     , m_nTableCount   );
     pMsgInstance->ls98PrintMessage(layCatalogExtract_els98, msgNumViewsExtracted_els98   , m_nViewCount    );
     pMsgInstance->ls98PrintMessage(layCatalogExtract_els98, msgNumSynonymsExtracted_els98, m_nSynonymCount );

     return rc;
}
/*===========================================================================*
 *  CatalogDefODBC::LoadAllCmd                                               *
 *===========================================================================*/
SAPDB_Int4 CatalogDefODBC::LoadCatalogCmd( StudioOAL_String username )
{
    ROUTINE_DBG_MEO00 ("CatalogDefODBC::LoadCatalogCmd");

    SAPDB_Int4 rc = LoadSchemaCmd(username);       // no hierarchy here; its simply the same procedure

    return rc;
}
/*===========================================================================*
 *  CatalogDefODBC::LoadSchemaCmd                                            *
 *===========================================================================*/
SAPDB_Int4 CatalogDefODBC::LoadSchemaCmd( StudioOAL_String username )
{
    ROUTINE_DBG_MEO00 ("CatalogDefODBC::LoadSchemaCmd");

    SAPDB_Bool rcBool = SAPDB_TRUE;
    SAPDB_Int4 rc     = errOK_els00;

    StudioOAL_String  line;
    StudioOAL_String  definition;
    StudioOAL_String  sqlmode    = "SQLMODE";
    StudioOAL_String  use        = "USE";
    StudioOAL_String  user       = "USER";
    StudioOAL_String  currentuser("CURRENT USER ");
    StudioOAL_String  whitespace((SAPDB_UTF8*)" ");
    StudioOAL_String  token      = CMD_SEPARATOR_TOKEN_LS00;

    SAPDB_Int4 lMaxFailedStmt = getMaxErrorCount();

    // read token
    while ( (errOK_els00 == rc) && (SAPDB_TRUE == readLine ( line ) ) )
    {
        line.Trim();

        // Build definition;
        // doing so we'll find command separators which are to be ignored; the thingies between
        // the separators are the commands we are intereseted in
        if ( 0 != token.Compare(line.SubStr(0,CMD_SEPARATOR_TOKEN_LEN_LS00) ) )
        {
            definition += whitespace;   // Assure that the command is properly continued
            definition += line;
        }
        else
        {
            if ( SAPDB_FALSE == definition.Empty() )
            {
                definition.Trim();

                // sqlmode statement
                if ( sqlmode.Compare(definition.SubStr(0,7) ) == 0 )
                {
                    definition = definition.SubStr(7,definition.Length());
                    definition.Trim();
                    
                    m_pCatalogLoadDefODBC->setSQLMode( definition );
                }
                else
                {
                    // convert use user statement to current user statement
                    if ( use.Compare(definition.SubStr(0,3) ) == 0 )
                    {
                        definition = definition.SubStr(3,definition.Length());
                        definition.Trim();

                        if ( user.Compare(definition.SubStr(0,4) ) == 0 )
                        {
                            definition = definition.SubStr(4,definition.Length());
                            definition.Trim();

                            definition = definition.SubStr(0,definition.FindFirstOf(whitespace));

                            currentuser += definition;
                            definition  = currentuser;

                            // Reinitialize currentuser string
                            currentuser = "CURRENT USER ";
                        }
                    }

                    rcBool = m_pCatalogLoadDefODBC->executeSQL( definition );
                    if (SAPDB_FALSE == rcBool)
                    {
                        // Check count of failed statements and report last error to client
                        if (0 == --lMaxFailedStmt)
                        {
                            m_pError->isError(SAPDB_TRUE);
                            m_pError->setErrorText(m_pCatalogLoadDefODBC->getErrorText());

                            rc = errODBCErrors_els98;
                        }
                    }
                }
                definition.Erase();
            
            }   // end if ( SAPDB_FALSE == definition.Empty() )
        
        }   // end else of if ( 0 != token.Compare(line.SubStr(0,CMD_SEPARATOR_TOKEN_LEN_LS00) ) )
        line.Erase();

    }   // end while ( readLine ( line ) == SAPDB_TRUE )

    return rc;
}
/*===========================================================================*
 *  CatalogDefODBC::LoadTableCmd
 *===========================================================================*/
SAPDB_Int4 CatalogDefODBC::LoadTableCmd ( tls00_PartSpec  catpartSpec )
{
    ROUTINE_DBG_MEO00 ("CatalogDefODBC::LoadTableCmd");

    SAPDB_Bool rcBool = SAPDB_TRUE;
    SAPDB_Int4 rc     = errOK_els00;

    StudioOAL_String  line;
    
    StudioOAL_String  definition;
    StudioOAL_String  sqlmode   = "SQLMODE";
    StudioOAL_String  whitespace((SAPDB_UTF8*)" ");
    StudioOAL_String  token     = CMD_SEPARATOR_TOKEN_LS00;

    SAPDB_Int4 lMaxFailedStmt = getMaxErrorCount();

    // read token
    while ( (errOK_els00 == rc) && (SAPDB_TRUE == readLine( line )) )
    {
        line.Trim();

        // Build definition;
        // doing so we'll find command separators which are to be ignored; the thingies between
        // the separators are the commands we are intereseted in

        if (0 != token.Compare(line.SubStr(0, CMD_SEPARATOR_TOKEN_LEN_LS00) ) )
        {
            definition += whitespace;       // Assure that the command is properly continued
            definition += line;
        }
        else
        {
            if ( SAPDB_FALSE == definition.Empty() )
            {
                definition.Trim();

                if (0 == sqlmode.Compare(definition.SubStr(0,7) ) )
                {
                    // SqlMode Statement

                    definition = definition.SubStr(7, definition.Length());
                    definition.Trim();

                    m_pCatalogLoadDefODBC->setSQLMode( definition );
                }
                else
                {
                    // Sql Statement
                    rcBool = m_pCatalogLoadDefODBC->executeSQL( definition );
                    if (SAPDB_FALSE == rcBool)
                    {
                        // Check count of failed statements and report last error to client
                        if (0 == --lMaxFailedStmt)
                        {
                            m_pError->isError(SAPDB_TRUE);
                            m_pError->setErrorText(m_pCatalogLoadDefODBC->getErrorText());

                            rc = errODBCErrors_els98;
                        }
                    }
                }
                definition.Erase();
            }

        }   // end if ( token.Compare(line.SubStr(0,CMD_SEPARATOR_TOKEN_LEN_LS00) ) == 0 && !definition.Empty())
    
    }   // end while ( readLine ( line ) == SAPDB_TRUE )

    return rc;
}
/*===========================================================================*
 *  CatalogDefODBC::readLine
 *===========================================================================*/
SAPDB_Bool CatalogDefODBC::readLine( StudioOAL_String &record  )
{
    SAPDB_Int4 rc = errOK_els00;
    SAPDB_Char ErrText[MAX_REPLY_LENGTH_LS00];

    SAPDB_Bool isOk = SAPDB_FALSE;

    SAPDB_Char szLine[LINELENGTH_MXSP00];

    rc = ls30VFReadLnEx(m_FileSession.vFile, szLine, ErrText, 0);
    if ( rc  >= 0 )
    {
        record = StudioOAL_String( (SAPDB_UTF8* )szLine );
        isOk = SAPDB_TRUE;
    }

    return isOk;
}
/*===========================================================================*
 *     CatalogDefODBC::writeLine
 *===========================================================================*/
SAPDB_Bool CatalogDefODBC::writeLine (StudioOAL_String record )
{
    SAPDB_Int4 rc = errOK_els00;
    SAPDB_Char ErrText[MAX_REPLY_LENGTH_LS00];

    SAPDB_Bool isOk = SAPDB_FALSE;

    rc = ls30VFWriteLn(m_FileSession.vFile, REINTERPRET_CAST(SAPDB_Char*, CONST_CAST(unsigned char*, record.StrPtr())), record.BasisSize(), ErrText);

    record.Erase();

    if (errOK_els00 == rc)
    {
        isOk = SAPDB_TRUE;
    }

    return isOk;
}

// Connection established

/*===========================================================================*
 *     CatalogDefODBC::connectionExtrCmd
 *===========================================================================*/
SAPDB_Int4 CatalogDefODBC::connectionExtrCmd( )
{
    ROUTINE_DBG_MEO00 ("CatalogDefODBC::connectionExtrCmd");

    SAPDB_Int4 rc = errOK_els00;



    if ( getExtrMode() == emExtractAsDDL_els31 )
    {
        StudioOAL_String catalog("SQLMODE ");
        catalog += pszPRSQLModeName_ls00[getSqlMode()];

        writeLine ( catalog );
    } 

    if ( getExtrMode() == emExtractAsXML_els31 )
    {
        // SQLGetConnectAttr
        StudioOAL_String  Version;
        StudioOAL_String  Date;
        StudioOAL_String  Time;
        StudioOAL_String  Driver;
        StudioOAL_String  DataSource;
        StudioOAL_String  ServerNode;
        StudioOAL_String  DBName;
        StudioOAL_String  UserName;
        StudioOAL_String  Password;
        StudioOAL_String  ConnectString; 
        StudioOAL_String  OpenQuoteString  = " [ ";
        StudioOAL_String  CloseQuoteString = " ] ";
        StudioOAL_String  SqlMode(" sapdbm:Sql_Mode=");

        tsp00_Versionc szVersion;
        tsp00_Datec    szDate;
        tsp00_Timec    szTime;

        tsp00_Date     pDate;
        tsp00_Time     pTime;
        sqldattime(pDate, pTime);
        szDate.p2c(pDate);
        szTime.p2c(pTime);

        //Version
        sp100_GetVersionString( COMP_NAME_REPL_MAN_SP100, s100buildnumber, szVersion );
        Version.ConvertFromASCII_Latin1     (szVersion.asCharp(),szVersion.asCharp()+szVersion.length () );
        Version = Version.SubStr(7,Version.Length());
        Version.Trim  ();
        //Date
        Date.ConvertFromASCII_Latin1(szDate.asCharp(),szDate.asCharp()+szDate.length () );
        //Time
        Time.ConvertFromASCII_Latin1(szTime.asCharp(),szTime.asCharp()+szTime.length () );
        
        SAPDB_Char   destBeg[1000] = "";
        SAPDB_UInt4  destAt  = 1000;

        //Driver
        if (destBeg != NULL) 
            memset(destBeg,0,1000);
        m_odbcConnection.GetDriver( REINTERPRET_CAST(unsigned char*,destBeg ),destAt);
        Driver.ConvertFromASCII_Latin1(destBeg,destBeg+strlen(destBeg)); 
                
        //DataSource
        //if (destBeg != NULL) 
        //    memset(destBeg,0,1000);
        //m_odbcConnection.GetDatasource( REINTERPRET_CAST(unsigned char*, destBeg ),destAt);
        //DataSource.ConvertFromASCII_Latin1(destBeg,destBeg+strlen(destBeg)); 
            
        //ServerNode
        m_odbcConnection.GetServerNode( REINTERPRET_CAST(unsigned char*, destBeg ), destAt);
        ServerNode.ConvertFromASCII_Latin1( destBeg, destBeg+strlen(destBeg)); 
    
        //DatabaseName
        if (destBeg != NULL) 
            memset(destBeg,0,1000);
        m_odbcConnection.GetServerDb ( REINTERPRET_CAST(unsigned char*,destBeg ),destAt);
        DBName.ConvertFromASCII_Latin1(destBeg,destBeg+strlen(destBeg)); 
                
        //UserName
        if (destBeg != NULL) 
            memset(destBeg,0,1000);
        m_odbcConnection.GetUser( REINTERPRET_CAST(unsigned char*,destBeg ),destAt);
        UserName.ConvertFromASCII_Latin1(destBeg,destBeg+strlen(destBeg)); 
        
        //Password
        if (destBeg != NULL) 
            memset(destBeg,0,1000);
        m_odbcConnection.GetPassword( REINTERPRET_CAST(unsigned char*,destBeg ),destAt);
        Password.ConvertFromASCII_Latin1(destBeg,destBeg+strlen(destBeg));
            
        // "DRIVER=SAPDB 7.3;SERVERNODE=P52837;SERVERDB=SUTDB73;UID=SUT;PWD=SUT;"
    
        ConnectString       = "DRIVER=";
        ConnectString      += Driver;
        ConnectString      += ";SERVERNODE=";
        ConnectString      += ServerNode;
        ConnectString      += ";SERVERDB=";
        ConnectString      += DBName;
        ConnectString      += ";UID=";
        ConnectString      += UserName;
        ConnectString      += ";PWD=";
        ConnectString      += Password;
        ConnectString      += ";";

            
        StudioOAL_String catalog;

        catalog = "<?xml version='1.0' encoding='ISO-8859-1'?>\n"
                  "<oim:Transfer "
                  " xmlns:oim='http://www.mdcinfo.com/oim/oim.dtd' ";
        //catalog += " xmlns:umx='REPOS:umx.dtd' ";
        //catalog += " xmlns:uml='REPOS:uml.dtd' ";
        //catalog += " xmlns:dbm='REPOS:dbm.dtd' ";
        //catalog += " xmlns:dtm='REPOS:dtm.dtd' ";
        catalog += " xmlns:umx='http://www.mdcinfo.com/oim/umx.dtd' "
                   " xmlns:uml='http://www.mdcinfo.com/oim/uml.dtd' "
                   " xmlns:dbm='http://www.mdcinfo.com/oim/dbm.dtd' "
                   " xmlns:dtm='http://www.mdcinfo.com/oim/dtm.dtd' ";
        //!! ToDo
        catalog += " xmlns:sapdbm='D:\\v73\\sys\\src\\ls\\sapdbm.dtd'>\n";

        catalog += "<oim:TransferHeader Exporter=\"ReplMan\"";
        catalog += " ExporterVersion=\"";
        catalog += Version;
        catalog += "\"";
        catalog += " TransferDateTime=\"";
        catalog += Date;
        catalog += "   ";
        catalog += Time;
        catalog += "\"/>\n";

        catalog += "<uml:System oim:id=\"_0\"  name=\"SAP DB Data Warehousing Framework\">\n"
                   "<uml:SystemModels>\n"
                   "<uml:Model  oim:id=\"_1\" oim:label=\"SAP DB Database Schema\" name=\"SAP DB Database Schema\">\n"
                   "<uml:ModelPackages>\n";

        // Dbm:ConnectionSet
        catalog += "<dbm:ConnectionSet oim:id=\"_2\" oim:label=\"SAP DB WEB Agent ODBC Connection Pooling\""
                   " name=\"SAP DB WEB Agent ODBC Connection Pooling\">\n";

        // Dbm:ConnectionSetConnections
        catalog += "<dbm:ConnectionSetConnections>\n";

        // sapdbm:Connection
        catalog += "<sapdbm:Connection oim:id=\"_4\" oim:supertype=\"Dbm:Connection\" oim:label=\"SAP DB ODBC Connection\" oim:name =\"";
        catalog += Driver;
        catalog += "\" dbm:UserName = \"";
        catalog += UserName;
        catalog += "\" dbm:Password = \"";
        catalog += Password;
        catalog += "\" dbm:ConnectString = \"";
        catalog += ConnectString;
        catalog += "\" dbm:ConnectionTimeOut = \" \""
                   "  dbm:Mode = \"sapdbwa_DBC_EXCL\""
                   "  dbm:IsReusable = \"0\""
                   "  dbm:OpenQuoteString = \"";
        catalog += OpenQuoteString;
        catalog += "\" dbm:CloseQuoteString = \"";
        catalog += CloseQuoteString;
        catalog += "\" dbm:DBName = \"";
        catalog += DBName;
        catalog += "\" dbm:DSN = \"\""
                   "  dbm:IsReadOnly = \"1\""
                   "  sapdbm:Sql_Producer=\"sp1pr_internal_cmd\" sapdbm:Sql_Component=\"sql_csp_comp_load\" ";

        
        switch(getSqlMode())
        {
            case cpr_kind_db2:
            {
                SqlMode += "\"db2\"/>";
                break;
            }
            case cpr_kind_ansi:
            {
                SqlMode += "\"ansi\"/>";
                break;
            }
            case cpr_kind_oracle:
            {
                SqlMode += "\"oracle\"/>";
                break;
            }
            case cpr_kind_internal:      
            default:
            {
                SqlMode += "\"internal\"/>";
            }
        }
        catalog += SqlMode;


        catalog += "</dbm:ConnectionSetConnections>";
        catalog += "</dbm:ConnectionSet>";

        // Dbm:DataSource
        catalog += "<dbm:DataSource oim:id=\"_3\"  oim:label=\"";
        catalog += DataSource;
        catalog += "\" name=\"";
        catalog += DataSource;
        catalog += "\">\n";

        // Dbm:DataSourceDBMS
        catalog += "<dbm:DataSourceDBMS oim:id=\"_5\" ";
        catalog += "  name=\"";
        catalog += DataSource;
        catalog += "\" version=\"";
        catalog += "";
        catalog += "\">\n";

        // Dbm:ColumnTypeSet
        catalog += "<dbm:ColumnTypeSet oim:id=\"8\"";
        catalog += "  name=\"";
        catalog += DataSource;
        catalog += "\" version=\"";
        catalog += "";
        catalog += "\">\n";

        // Dtm:TypeSetTypes
        catalog += "<dtm:TypeSetTypes>\n";

        writeLine ( catalog );
        catalog.Erase();

        columntypesetExtrCmd( );

        catalog += "</dtm:TypeSetTypes>\n"
                   "</dbm:ColumnTypeSet>\n"
                   "</dbm:DataSourceDBMS>\n";

        // Dbm:DataSourceConnections
        catalog += "<dbm:DataSourceConnections>\n";

        // sapdbm:Connection
        catalog += "<sapdbm:Connection oim:id=\"_6\" oim:supertype=\"Dbm:Connection\" oim:label=\"SAP DB ODBC Connection\" oim:name =\"";
        catalog += getOdbcDriver();
        catalog += "\" dbm:UserName = \"";
        catalog += UserName;
        catalog += "\" dbm:Password = \"";
        catalog += Password;
        catalog += "\" dbm:ConnectString = \"";
        catalog += ConnectString;
        catalog += "\" dbm:ConnectionTimeOut = \" \""
                   "  dbm:Mode = \"sapdbwa_DBC_EXCL\""
                   "  dbm:IsReusable = \"0\""
                   "  dbm:OpenQuoteString = \"";
        catalog += OpenQuoteString;
        catalog += "\" dbm:CloseQuoteString = \"";
        catalog += CloseQuoteString;
        catalog += "\" dbm:DBName = \"";
        catalog += DBName;
        catalog += "\" dbm:DSN = \"\""
                   "  dbm:IsReadOnly = \"1\""
                   "  sapdbm:Sql_Producer=\"sp1pr_internal_cmd\" sapdbm:Sql_Component=\"sql_csp_comp_load\" ";

        catalog += SqlMode;

        catalog += "</dbm:DataSourceConnections>\n";

        // Dbm:DataSourceDeployedCatalogs
        catalog += "<dbm:DataSourceDeployedCatalogs>\n";

        // Dbm:DeployedCatalog
        catalog += "<dbm:DeployedCatalog oim:id=\"_7\" ";
        catalog += " RepositoryUpdatedBy=\"";
        catalog += UserName;
        catalog += "\" RepositoryUpdated=\"";
        catalog += Date;
        catalog += "   ";
        catalog += Time;
        catalog += "\" URL=\"";
        catalog += DBName;
        catalog += "\" name=\"";
        catalog += DBName;
        catalog += "\">\n";

        // Dbm:CatalogSchemas
        catalog += "<dbm:CatalogSchemas>\n";

        writeLine ( catalog );
    }
    return rc;
}
/*===========================================================================*
 *  CatalogDefODBC::columntypesExtrCmd                                       *
 *===========================================================================*/
SAPDB_Int4 CatalogDefODBC::columntypesetExtrCmd( )
{
    RepServMessages *pMsgInstance = RepServMessages::Instance();

    SAPDB_Int4 rc = errOK_els00;

    // SYSODBCTYPES  :  0.TYPE_NAME, 1.DATA_TYPE, 2.PRECISION, 3.LITERAL_PREFIX, 4.LITERAL_SUFFIX, 5.CREATE_PARAMS, 6.NULLABLE, 7.CASE_SENSITIVE, 8.SEARCHABLE, 9.UNSIGNED_ATTRIBUTE, 10.MONEY, 11.AUTO_INCREMENT
    //                 12.LOCAL_TYPE_NAME, 13.MINIMUM_SCALE, 14.MAXIMUM_SCALE, 15.SQL_DATA_TYPE, 16.SQL_DATETIME_SUB, 17.NUM_PREC_RADIX, 18.INTERVAL_PRECISION, 19.REMARKS

    // SQLGetTypeInfo:  0.TYPE_NAME, 1.DATA_TYPE, 2.PRECISION, 3.LITERAL_PREFIX, 4.LITERAL_SUFFIX, 5.CREATE_PARAMS, 6.NULLABLE, 7.CASE_SENSITIVE, 8.SEARCHABLE, 9.UNSIGNED_ATTRIBUTE, 10.MONEY, 11.AUTO_INCREMENT
    //                  12.LOCAL_TYPE_NAME, 13.MINIMUM_SCALE, 14.MAXIMUM_SCALE

    if ( getExtrMode() == emExtractAsXML_els31 )
    {
        StudioOAL_String  record[15];

        if( !m_pColumnTypeDefODBC->getDataSourceTypeInfo() )
        {
            StudioOAL_String statement = "SQLGetTypeInfo";
            StudioOAL_String nativeError;
            StudioOAL_String errorText;

            nativeError = m_pColumnTypeDefODBC->getNativeErrorAsString();
            errorText   = m_pColumnTypeDefODBC->getErrorText();

            SAPDB_Char        ErrText[MAX_REPLY_LENGTH_LS00];
            const SAPDB_Char* ErrTextEnd = ErrText + MAX_REPLY_LENGTH_LS00;
            SAPDB_Char*       destAt;

            errorText.ConvertToASCII_Latin1(ErrText, ErrTextEnd, destAt);

            //TRACE_PRNF_MLS99(("failed ", "%s\n"        , statement.StrPtr() ));  !!ToDo
            //TRACE_PRNF_MLS99(("getErrorText() ", "%s\n", errorText.StrPtr() ));

            //if ( getExtrMode() == emExtractAsDDL_els31 )
            pMsgInstance->ls98PrintMessage(layCatalogExtract_els98, msgSimpleMsg_els98 , statement.StrPtr() );
            pMsgInstance->ls98PrintMessage(layCatalogExtract_els98, errODBCErrors_els98, errorText.StrPtr() );
            //if ( getExtrMode() == emLoadAsDDL_els31 )
            //pMsgInstance->ls98PrintMessage(layCatalogLoad_els98, msgSimpleMsg_els98 , statement.StrPtr() );
            //pMsgInstance->ls98PrintMessage(layCatalogLoad_els98, errODBCErrors_els98, errorText.StrPtr() );
        }

        /*-------------------------------------------------------------------------------*
         * Read/Write the catalog definition of the specified object from/to resources   *
         *-------------------------------------------------------------------------------*/
        if (errOK_els00 == rc)
        {
            while ( m_pColumnTypeDefODBC->readResult (record) )
                m_pColumnTypeDefODBC->writeResult(record);
        }
    }
    return rc;
}
// Autorizations Definition

/*===========================================================================*
 *  CatalogDefODBC::schemaExtrCmd                                            *
 *===========================================================================*/
SAPDB_Int4 CatalogDefODBC::schemaExtrCmd( StudioOAL_String user  )
{
    ROUTINE_DBG_MEO00 ("CatalogDefODBC::schemaExtrCmd");

    SAPDB_Int4 rc = errOK_els00;

    if ( getExtrMode() == emExtractAsXML_els31 )
    {
        StudioOAL_String schema;

        /*-------------------------------------------------------------------------*
        * Close the previous Schema                                               *
        *-------------------------------------------------------------------------*/
        if ( m_nSchemaCount > 0 )
        {
            schema += "</uml:NamespaceOwnedElement>\n";
            schema += "</sapdbm:DeployedSchema>\n";
            schema += "</uml:NamespaceOwnedElement>\n";
            writeLine(schema);
        }

        /*-------------------------------------------------------------------------*
        * Open the next Schema                                                    *
        *-------------------------------------------------------------------------*/
        // DOMAIN.USERS: OWNER,GROUPNAME,USERNAME,USERMODE,CONNECTMODE,MAXTIMEOUT,COSTWARNING,COSTLIMIT,CREATEDATE,CREATETIME,ALTERDATE,ALTERTIME,PWCREADATE,PWCREATIME,SERVERDB,SERVERNODE,USER_ID,COMMENT

        StudioOAL_String record[9];

        StudioOAL_String select("SELECT OWNER, GROUPNAME, USERNAME, USERMODE,"
                                 " MAXTIMEOUT, COSTWARNING, COSTLIMIT, CONNECTMODE, COMMENT "
                                 "FROM DOMAIN.USERS ");
        /*-------------------------------------------------------------------------*
        * Prepare the WHERE clause for the SQL statements.                        *
        *-------------------------------------------------------------------------*/
        StudioOAL_String whereClause("WHERE USERNAME = '");
        whereClause  += user;
        whereClause  +="'";
        select       += whereClause;

        /*-------------------------------------------------------------------------------*
        * Read/Write the catalog definition of the specified object from/to resources   *
        *-------------------------------------------------------------------------------*/
        if ( m_pUserDefODBC->executeSQL( select ) )
        {
            if ( m_pUserDefODBC->readResult (record) )
            {
                m_pUserDefODBC->writeResult(record,emExtractAsXML_els31);
                m_nSchemaCount++;
            }
        }
    }
    return rc;
}

/*===========================================================================*
 *  CatalogDefODBC::userExtrCmd                                              *
 *===========================================================================*/
SAPDB_Int4 CatalogDefODBC::userExtrCmd( tls00_PartSpec catpartSpec )
{
    ROUTINE_DBG_MEO00 ("CatalogDefODBC::userExtrCmd");

    SAPDB_Int4 rc = errOK_els00;

    // DOMAIN.USERS  : OWNER,GROUPNAME,USERNAME,USERMODE,CONNECTMODE,MAXTIMEOUT,COSTWARNING,COSTLIMIT,CREATEDATE,CREATETIME,ALTERDATE,ALTERTIME,PWCREADATE,PWCREATIME,SERVERDB,SERVERNODE,USER_ID,COMMENT

    StudioOAL_String  whereClause;
    StudioOAL_String  orderClause;
    StudioOAL_String  record[9];

    StudioOAL_String  select("SELECT OWNER, GROUPNAME, USERNAME, USERMODE,"
                             " MAXTIMEOUT, COSTWARNING, COSTLIMIT, CONNECTMODE,"
                             " COMMENT FROM DOMAIN.USERS ");

    if ( getExtrMode() == emExtractAsDDL_els31 )
    {
        /*-------------------------------------------------------------------------*
         * Prepare the WHERE clause for the SQL statements.                        *
         *-------------------------------------------------------------------------*/

        // catalog extract all
        if  ( getPartSpec().psSelAll_ls00 )
        {
            whereClause  = " WHERE USERMODE <> 'SYSDBA'  "
                           " AND   USERMODE <> 'COLDUSER'"
                           " AND   USERNAME <> 'DOMAIN'  "
                           " AND   USERNAME <> 'SYS'     ";

            orderClause  = " ORDER BY OWNER , GROUPNAME , USERNAME ASC ";
        }
        // catalog extract user
        if ( getPartSpec().psSelUser_ls00 )
        {
            whereClause  = "WHERE OWNER = '";
            whereClause  += getCurrentUser();
            whereClause  +="'";

            orderClause  = " ORDER BY GROUPNAME , USERNAME ASC ";
        }

        select += whereClause;
        select += orderClause;

        /*-------------------------------------------------------------------------------*
        * Read/Write the catalog definition of the specified object from/to resources   *
        *-------------------------------------------------------------------------------*/

        if (  m_pUserDefODBC->executeSQL( select ) )
        {
            while ( m_pUserDefODBC->readResult (record) )
            {
                setCurrentUser ( record[0] );

                m_pUserDefODBC->writeResult(record,emExtractAsDDL_els31);
            }
        }
    }
    return rc;
}
/*===========================================================================*
 *  CatalogDefODBC::roleExtrCmd                                              *
 *===========================================================================*/
SAPDB_Int4 CatalogDefODBC::roleExtrCmd( tls00_PartSpec catpartSpec )
{
    ROUTINE_DBG_MEO00 ("CatalogDefODBC::roleExtrCmd");

    SAPDB_Int4 rc = errOK_els00;

    if ( getExtrMode() == emExtractAsDDL_els31 )
    {
        // <grantee> ::= PUBLIC | <user_name> | <usergroup_name> | <role_name>

        // ROLES          :OWNER,ROLE,PASSWORD_REQUIRED,GRANTED,CREATEDATE,CREATEDATE

        StudioOAL_String  select;
        StudioOAL_String  whereClause;

        StudioOAL_String  record[2];

        select  = "SELECT ROLE,PASSWORD_REQUIRED ";
        select += "FROM DOMAIN.ROLES ";

        if  ( catpartSpec.psSelUser_ls00 )
        {
            whereClause  = "WHERE OWNER = '";
            whereClause += getCurrentUser();
            whereClause += "'";
        }

        select += whereClause;

        /*-------------------------------------------------------------------------------*
         * Read/Write the role definition of the specified table from/to resources
         *-------------------------------------------------------------------------------*/
        if ( m_pRoleDefODBC->executeSQL( select ) )
        {
            while ( m_pRoleDefODBC->readResult (record) )
                m_pRoleDefODBC->writeResult(record);
        }
    }
    return rc;
}

// Data Definition

/*===========================================================================*
 *  CatalogDefODBC::tabletypeExtrCmd                                         *
 *===========================================================================*/
SAPDB_Int4 CatalogDefODBC::tabletypeExtrCmd ( tls00_PartSpec catpartSpec )
{
    ROUTINE_DBG_MEO00 ("CatalogDefODBC::tabletypeExtrCmd");

    SAPDB_Int4 rc = errOK_els00;

    // CREATE CATALOG
    // CATALOG 7.2   : LEVEL,CMD_NO,SEQ_NO,DT_FORMAT,OWNER,TABLENAME,ERROR_CODE,CMD
    // CATALOG 7.3   : LEVEL,CMD_NO,SEQ_NO,DT_FORMAT,OWNER,TABLENAME,ERROR_CODE,SQL_CMD, TYPE, TABLETYPE, CMD

   
    StudioOAL_String  definition;
    StudioOAL_String  view;
    StudioOAL_String  synonym;

    StudioOAL_String  record[4];

    StudioOAL_String  sql;
    StudioOAL_String  table  = "TABLE";
    StudioOAL_String  create = "CREATE";
    StudioOAL_String  grant  = "GRANT";

    StudioOAL_String  cmd_no;

    SAPDB_Bool bGrant = SAPDB_FALSE;

    /*-------------------------------------------------------------------------------*
     * Read/Write the catalog definition of the specified object from/to resources   *
     *-------------------------------------------------------------------------------*/
    while ( m_pCatalogExtractDefODBC->readResult(record) )
    {
        cmd_no = record[0];
        SAPDB_Int4 columnlength = 0;

        //  <sapdbm:DeployedSchema>
        if ( record[1].Compare( getCurrentUser() ) != 0  || m_nSchemaCount == 0 )
        {
            schemaExtrCmd ( record[1] );
        }

        setCurrentUser( record[1] );

        if ( table.Compare( record[3].SubStr(0,5) ) == 0)
        {
            // <dbm:SchemaTables>
            // <dbm:SchemaIndexes>

            // <dbm:DeployedTable>
            // CREATE TABLE
            tableExtrCmdExec  ( record[1], record[2] );
            // <dbm:TableIndices>
            // CREATE INDEX
            indexExtrCmdExecDirect( record[1], record[2] );
            // <dbm:TableTriggers>
            // CREATE TRIGGER
            triggerExtrCmdExec( record[1], record[2] );

            if ( getExtrMode() == emExtractAsXML_els31 )
            {
                StudioOAL_String   deployedTable;
                deployedTable  = "</sapdbm:DeployedTable>";
                m_pCatalogExtractDefODBC->writeResult(deployedTable);
            }
            bGrant = SAPDB_FALSE;
            continue;
        }

        if (create.Compare( record[3].SubStr(0,6) ) == 0)
        {
            definition  = record[3].SubStr(7,record[3].Length());
            definition.Trim()  ;

            // <dbm:SchemaViews>
            // CREATE VIEW
            if(  definition.SubStr(0,4).Compare("VIEW") == 0   )
            {
                viewExtrCmdExecDirect   ( record[1], record[2] );
            }

            // <dbm:SchemaTableSynonyms>
            // CREATE SYNONYM
            if ( definition.SubStr(0,7).Compare("SYNONYM")  == 0 ||
                 definition.SubStr(0,6).Compare("PUBLIC" )  == 0  )
            {
                synonymExtrCmdExecDirect( record[1], record[2] );
            }
            bGrant = SAPDB_FALSE;
            continue;
        }
        // GRANT
        // DOMAIN.SHOW_PRIV_GRANTED
        if (grant.Compare( record[3].SubStr(0,5) ) == 0 )
        {
            if ( getExtrMode() == emExtractAsDDL_els31 )
            {
                sql  = SEPARATOR;
                sql += record[3];
                m_pCatalogExtractDefODBC->writeResult(sql);
            }
            bGrant = SAPDB_TRUE;
            continue;
        }
        
        if ( bGrant == SAPDB_TRUE )
        {
            if ( getExtrMode() == emExtractAsDDL_els31 )
            {
                sql = record[3];
                m_pCatalogExtractDefODBC->writeResult(sql);
            }
        }

    } // while ( getCatalogExtractDefODBC()->readResult (record) )

    return rc;
}
/*===========================================================================*
 *  CatalogDefODBC::tableExtrCmdExecDirect
 *===========================================================================*/
SAPDB_Int4
CatalogDefODBC::tableExtrCmdExecDirect( StudioOAL_String owner, StudioOAL_String tablename )
{
    ROUTINE_DBG_MEO00 ("CatalogDefODBC::tableExtrCmdExecDirect");

    // CREATE TABLE
    // TABLES   :OWNER,TABLENAME,PRIVILEGES,TYPE,CREATEDATE,CREATETIME,UPDSTATDATE,
    //           UPDSTATTIME,ALTERDATE,ALTERTIME,UNLOADED,SAMPLE_PERCENT,SAMPLE_ROWS,COMMENT,TABLEID

    /*-------------------------------------------------------------------------*
     * Read/Write the table definition of the specified table from/to resources*
     *-------------------------------------------------------------------------*/
    SAPDB_Int4 rc = m_pTableDefODBC->readTableDefinitionDirect(owner, tablename);
    if (errOK_els00 == rc)
    {
        m_nTableCount ++;
        m_pTableDefODBC->writeTableDefinition ();
    }

    return rc;
}
/*===========================================================================*
 *  CatalogDefODBC::tableExtrCmdExec
 *===========================================================================*/
SAPDB_Int4 CatalogDefODBC::tableExtrCmdExec ( StudioOAL_String owner, StudioOAL_String tablename )
{
    ROUTINE_DBG_MEO00 ("CatalogDefODBC::tableExtrCmdExec");

    SAPDB_Int4 rc = errOK_els00;

    // CREATE TABLE
    // TABLES   :OWNER,TABLENAME,PRIVILEGES,TYPE,CREATEDATE,CREATETIME,UPDSTATDATE,
    //           UPDSTATTIME,ALTERDATE,ALTERTIME,UNLOADED,SAMPLE_PERCENT,SAMPLE_ROWS,COMMENT,TABLEID

    /*-------------------------------------------------------------------------*
     * Read/Write the table definition of the specified table from/to resources*
     *-------------------------------------------------------------------------*/
    if ( m_pTableDefODBC->readTableDefinition( owner, tablename ) )
    {
        ++m_nTableCount;
        m_pTableDefODBC->writeTableDefinition();
    }

    return rc;
}
/*===========================================================================*
 *  CatalogDefODBC::viewExtrCmdExecDirect
 *===========================================================================*/
SAPDB_Int4 CatalogDefODBC::viewExtrCmdExecDirect( StudioOAL_String owner, StudioOAL_String viewname )
{
    ROUTINE_DBG_MEO00 ("CatalogDefODBC::viewExtrCmdExecDirect");

    SAPDB_Int4 rc     = errOK_els00;
    SAPDB_Bool rcExec = SAPDB_TRUE;

    StudioOAL_String  select;
    StudioOAL_String  whereClause;

    /*-------------------------------------------------------------------------*
     * Prepare the WHERE clause for the SQL statements.                        *
     *-------------------------------------------------------------------------*/
    whereClause += "WHERE owner    = '";
    whereClause += owner ;
    whereClause += "' AND viewname = '";
    whereClause += viewname ;
    whereClause += "' ";

    StudioOAL_String  record[3];

    select  = " SELECT viewname, owner, definition "
              " FROM DOMAIN.viewdefs ";
    select += whereClause;

    SAPDB_Long getNativeError = 0;

    rcExec = m_pViewDefODBC->executeSQL( select );
    if (SAPDB_TRUE == rcExec)
    {
        rcExec = m_pViewDefODBC->readResult( record );
        if (SAPDB_TRUE == rcExec)
        {
            ++m_nViewCount;
            m_pViewDefODBC->writeResult( record );
        }
    }
    if (SAPDB_FALSE == rcExec)
    {
        getNativeError = m_pViewDefODBC->getNativeError();
        rc = STATIC_CAST(SAPDB_Int4, getNativeError);
    }

    return rc;
}
/*===========================================================================*
 *  CatalogDefODBC::viewExtrCmdExec
 *===========================================================================*/
SAPDB_Int4 CatalogDefODBC::viewExtrCmdExec( StudioOAL_String owner, StudioOAL_String viewname )
{
     ROUTINE_DBG_MEO00 ("CatalogDefODBC::viewExtrCmdExec");

     SAPDB_Int4 rc = errOK_els00;

    /*-------------------------------------------------------------------------*
     * Read/Write the view definition of the specified table from/to resources
     *-------------------------------------------------------------------------*/

     if ( m_pViewDefODBC->readViewDefinition  ( owner, viewname ) )
     { 
          m_nViewCount ++ ;
          m_pViewDefODBC->writeViewDefinition ();
     }

     return rc;
}
/*===========================================================================*
 *  CatalogDefODBC::synonymExtrCmdExecDirect
 *===========================================================================*/
SAPDB_Int4 CatalogDefODBC::synonymExtrCmdExecDirect( StudioOAL_String owner, StudioOAL_String synonymname )
{
     ROUTINE_DBG_MEO00 ("CatalogDefODBC::synonymExtrCmdExecDirect");

     SAPDB_Int4 rc = errOK_els00;

     // CREATE SYNONYM
     // SYNONYMS  : OWNER, SYNONYMNAME, PUBLIC, TABLEOWNER, TABLENAME, COMMENT

     StudioOAL_String  select;
     StudioOAL_String  whereClause;

    /*-------------------------------------------------------------------------*
     * Prepare the WHERE clause for the SQL statements.                        *
     *-------------------------------------------------------------------------*/
     whereClause += "WHERE synonymname = '";
     whereClause += synonymname;
     whereClause += "' AND owner = '";
     whereClause += owner;
     whereClause += "' ";

     StudioOAL_String  record[6];

     select  = " SELECT OWNER, SYNONYMNAME, PUBLIC, TABLEOWNER, TABLENAME, COMMENT ";
     select += " FROM DOMAIN.SYNONYMS ";
     select += whereClause;

     SAPDB_Bool      isError       = SAPDB_TRUE;
     SAPDB_Long      getNativeError= 0;
     SAPDB_Int4      getErrorPos   = 0;
     SQLRETURN       getReturnCode = 0;
     StudioOAL_String getErrorText;
     StudioOAL_String getNativeErrorAsString;

     if ( !m_pSynonymDefODBC->executeSQL( select )  )
     {
           isError                = m_pSynonymDefODBC->isError();
           getNativeError         = m_pSynonymDefODBC->getNativeError();
           getErrorPos            = m_pSynonymDefODBC->getErrorPos();
           getReturnCode          = m_pSynonymDefODBC->getReturnCode();
           getErrorText           = m_pSynonymDefODBC->getErrorText();
           getNativeErrorAsString = m_pSynonymDefODBC->getNativeErrorAsString();
           return getNativeError;
     }

     while ( m_pSynonymDefODBC->readResult ( record ) )
     {
             m_nSynonymCount++;
             m_pSynonymDefODBC->writeResult( record );
     }

     return rc;

}
/*===========================================================================*
 *  CatalogDefODBC::synonymExtrCmdExec
 *===========================================================================*/
SAPDB_Int4 CatalogDefODBC::synonymExtrCmdExec( StudioOAL_String owner, StudioOAL_String synonymname )
{
     ROUTINE_DBG_MEO00 ("CatalogDefODBC::synonymExtrCmdExec");

     SAPDB_Int4 rc = errOK_els00;

    /*-------------------------------------------------------------------------*
     * Read/Write the synonym definition of the specified table from/to resources
     *-------------------------------------------------------------------------*/

     if ( m_pSynonymDefODBC->readSynonymDefinition  ( owner, synonymname ) )
     { 
          m_nSynonymCount++;
          m_pSynonymDefODBC->writeSynonymDefinition ();
     }

     return rc;
}
/*===========================================================================*
 *  CatalogDefODBC::indexExtrCmdExecDirect
 *===========================================================================*/
SAPDB_Int4
CatalogDefODBC::indexExtrCmdExecDirect( StudioOAL_String owner, StudioOAL_String tablename )
{
    ROUTINE_DBG_MEO00 ("CatalogDefODBC::indexExtrCmdExecDirect");

    SAPDB_Int4 rc = errOK_els00;

    /*-------------------------------------------------------------------------*
     * Before doing anything else check if more than one table with the same   *
     * name and different owners is found in the DB in case the owner name     *
     * is not specified.                                                       *
     *-------------------------------------------------------------------------*/
    if (SAPDB_TRUE == owner.Empty())
    {
        StudioOAL_String  SelectOwnerCountCmd("SELECT COUNT(DISTINCT OWNER) FROM "
                                              "DOMAIN.COLUMNS WHERE TABLENAME = '");
        SelectOwnerCountCmd += tablename;
        SelectOwnerCountCmd += "'";
        
        if (SAPDB_TRUE == m_pIndexDefODBC->executeSQL( SelectOwnerCountCmd ))
        {
            StudioOAL_String record[1];
        
            m_pIndexDefODBC->readResult(record);
            SAPDB_Int4 lOwnerCount = AtoI( record[0] );
            if (lOwnerCount > 1)
            {
                RepServMessages::Instance()->ls98PrintMessage(layCatalogExtract_els98, errDEAmbiguousTableName_els98);
                rc = errDEAmbiguousTableName_els98;
            }
        }
        else
        {
            rc = errODBCErrors_els98;
        }
    }

    if (errOK_els00 == rc)
    {
        StudioOAL_String  select;
        StudioOAL_String  whereClause;

        /*-------------------------------------------------------------------------*
         * Prepare the WHERE clause for the SQL statements.                        *
         *-------------------------------------------------------------------------*/
        whereClause  = "WHERE tablename = '";
        whereClause += tablename;
        whereClause += "' AND DISABLED = 'NO'";

        if (SAPDB_FALSE == owner.Empty())
        {
            whereClause += " AND owner = '";
            whereClause += owner;
            whereClause += "'";
        }

        // CREATE INDEX
        // INDEXES: OWNER, TABLENAME, INDEXNAME, TYPE, CREATEDATE, CREATETIME, INDEX_USED, DISABLED, COMMENT

        //<create_index_statement> ::=
        // CREATE [UNIQUE] INDEX <index_name> ON <table_name> (<column_name> [ASC | DESC],...)

        select  = "SELECT TYPE, INDEXNAME, COMMENT FROM DOMAIN.INDEXES ";
        select += whereClause;


        if ( !m_pIndexDefODBC->executeSQL( select ) )
        {
            rc = STATIC_CAST(SAPDB_Int4, m_pIndexDefODBC->getNativeError());
        }

        /*-------------------------------------------------------------------------*
         * Read/Write the index definition of the specified table from/to resources
         *-------------------------------------------------------------------------*/
        if (errOK_els00 == rc)
        {
            if (emExtractAsDDL_els31 == getExtrMode())
            {
                while ( m_pIndexDefODBC->readIndexDefinitionDirect( owner, tablename ) )
                {
                    m_pIndexDefODBC->writeIndexDefinition();
                }
            }
            else if (emExtractAsXML_els31 == getExtrMode())
            {
                //SAPDB_Int4 nIndices = 0;
                StudioOAL_String tableIndices;
                SAPDB_Bool       bHasIndex = SAPDB_FALSE;

                while ( m_pIndexDefODBC->readIndexDefinitionDirect ( owner, tablename ) )
                {
                    if (SAPDB_FALSE == bHasIndex)
                    {
                        tableIndices = "<dbm:TableIndices>\n";
                        m_pIndexDefODBC->writeLine( tableIndices );
                        bHasIndex = SAPDB_TRUE;
                    }
                    m_pIndexDefODBC->writeIndexDefinition();
                }

                if (bHasIndex == SAPDB_TRUE)
                {
                    tableIndices = "</dbm:TableIndices>\n";
                    m_pIndexDefODBC->writeLine( tableIndices );
                }
            }
        }
    }   // end if (errOK_els00 == rc)

    return rc;
}
/*===========================================================================*
 *  CatalogDefODBC::indexExtrCmdExec
 *===========================================================================*/
SAPDB_Int4 CatalogDefODBC::indexExtrCmdExec ( StudioOAL_String owner, StudioOAL_String tablename )
{
    ROUTINE_DBG_MEO00 ("CatalogDefODBC::indexExtrCmdExec");

    SAPDB_Int4 rc = errOK_els00;

    /*-------------------------------------------------------------------------*
     * Read/Write the index definition of the specified table from/to resources
     *-------------------------------------------------------------------------*/

    m_pIndexDefODBC->setIndexParameters( owner, tablename ) ;

    if ( getExtrMode() == emExtractAsDDL_els31 )
    {
         while ( m_pIndexDefODBC->readIndexDefinition  ( owner, tablename ) )
                 m_pIndexDefODBC->writeIndexDefinition ();

    }

    if ( getExtrMode() == emExtractAsXML_els31 )
    {

         SAPDB_Int4      nIndices      = 0;
         StudioOAL_String tableIndices;

         while ( m_pIndexDefODBC->readIndexDefinition  ( owner, tablename ) )
         {
                nIndices++;

                if ( getExtrMode() == emExtractAsXML_els31 && nIndices == 1 )
                {
                     tableIndices   = "<dbm:TableIndices>\n";
                     m_pIndexDefODBC->writeLine ( tableIndices );
                }
                m_pIndexDefODBC->writeIndexDefinition ();
         }

         if ( getExtrMode() == emExtractAsXML_els31 && nIndices >= 1 )
         {
              tableIndices   = "</dbm:TableIndices>\n";
              m_pIndexDefODBC->writeLine ( tableIndices );
         }
    }

    return rc;
}

/*===========================================================================*
 *  CatalogDefODBC::sequenceExtrCmd                                          *
 *===========================================================================*/
SAPDB_Int4 CatalogDefODBC::sequenceExtrCmd ( tls00_PartSpec catpartSpec )
{
    ROUTINE_DBG_MEO00 ("CatalogDefODBC::sequenceExtrCmd");

    RepServMessages *pMsgInstance = RepServMessages::Instance();

    SAPDB_Int4 rc = errOK_els00;

    if ( getExtrMode() == emExtractAsDDL_els31 )
    {
        
        SQLRETURN         ReturnCode = 0;
        StudioOAL_String  NativeErrorAsString;

        StudioOAL_String  select;
        StudioOAL_String  whereClause;
        StudioOAL_String  record[9];

        /*-------------------------------------------------------------------------*
         * Prepare the WHERE clause for the SQL statements.                        *
         *-------------------------------------------------------------------------*/
        select = " SELECT DISTINCT OWNER FROM TEMP.LOA_SAVE_CATALOG WHERE OWNER NOT IN ('SYS','PUBLIC') ";

        if ( !m_pCatalogExtractDefODBC->executeSQL( select ) )
        {
            ReturnCode          = m_pCatalogExtractDefODBC->getReturnCode();
            NativeErrorAsString = m_pCatalogExtractDefODBC->getNativeErrorAsString();
            
            pMsgInstance->ls98PrintMessage(layCatalogExtract_els98, errODBCErrors_els98, NativeErrorAsString.StrPtr());

            return ReturnCode;
        }

        /*-------------------------------------------------------------------------------*
         * Read/Write the catalog definition of the specified object from/to resources   *
         *-------------------------------------------------------------------------------*/
        while ( m_pCatalogExtractDefODBC->readResult (record) )
        {
            setCurrentUser( record[0] );

            // CREATE SEQUENCE
            // SEQUENCES  : OWNER,SEQUENCE_NAME,INCREMENT_BY,LAST_NUMBER,MAX_VALUE,MIN_VALUE,
            //              CYCLE_FLAG,CACHE_SIZE,ORDER_FLAG,CREATEDATE,CREATETIME,COMMENT

            select  = "SELECT OWNER,SEQUENCE_NAME,INCREMENT_BY,LAST_NUMBER,MAX_VALUE,MIN_VALUE,CYCLE_FLAG,CACHE_SIZE,ORDER_FLAG"
                      " FROM DOMAIN.SEQUENCES ";

            if ( ! m_pSequenceDefODBC->executeSQL( select )  )
            {
                ReturnCode          = m_pSequenceDefODBC->getReturnCode();
                NativeErrorAsString = m_pSequenceDefODBC->getNativeErrorAsString();

                pMsgInstance->ls98PrintMessage(layCatalogExtract_els98, errODBCErrors_els98, NativeErrorAsString.StrPtr());

                return ReturnCode;
            }

            /*----------------------------------------------------------------------------*
             * Read/Write the sequence definition of the specified table from/to resources*
             *----------------------------------------------------------------------------*/
            while ( m_pSequenceDefODBC->readResult (record) )
                m_pSequenceDefODBC->writeResult(record);
        }
    }// if ( getExtrMode() == emExtractAsDDL_els31 )

    return rc;
}
/*===========================================================================*
 *  CatalogDefODBC::dbprocedureExtrCmd
 *===========================================================================*/
SAPDB_Int4 CatalogDefODBC::dbprocedureExtrCmd( tls00_PartSpec catpartSpec )
{
     ROUTINE_DBG_MEO00 ("CatalogDefODBC::dbprocedureExtrCmd");

     RepServMessages *pMsgInstance = RepServMessages::Instance();

     SAPDB_Int4 rc = errOK_els00;

     SAPDB_Bool        Error       = SAPDB_TRUE;
     SAPDB_Long        NativeError = 0;
     SAPDB_Int4        ErrorPos    = 0;
     SQLRETURN         ReturnCode  = 0;
     StudioOAL_String  ErrorText;
     StudioOAL_String  NativeErrorAsString;

     StudioOAL_String  select;
     StudioOAL_String  whereClause;
     StudioOAL_String  record[4];

     if ( getExtrMode() == emExtractAsDDL_els31 ) 
     {
        /*-------------------------------------------------------------------------*
         * Prepare the WHERE clause for the SQL statements.                        *
         *-------------------------------------------------------------------------*/
         select = " SELECT DISTINCT OWNER FROM TEMP.LOA_SAVE_CATALOG WHERE OWNER NOT IN ('SYS','PUBLIC') ";

         if ( !m_pCatalogExtractDefODBC->executeSQL( select ) )
         {
               Error               = m_pCatalogExtractDefODBC->isError();
               NativeError         = m_pCatalogExtractDefODBC->getNativeError();
               ErrorPos            = m_pCatalogExtractDefODBC->getErrorPos();
               ReturnCode          = m_pCatalogExtractDefODBC->getReturnCode();
               ErrorText           = m_pCatalogExtractDefODBC->getErrorText();
               NativeErrorAsString = m_pCatalogExtractDefODBC->getNativeErrorAsString();

               pMsgInstance->ls98PrintMessage(layCatalogExtract_els98, errODBCErrors_els98, NativeErrorAsString.StrPtr());

               return ReturnCode;
         }

        /*-------------------------------------------------------------------------------*
         * Read/Write the catalog definition of the specified object from/to resources   *
         *-------------------------------------------------------------------------------*/
         while ( m_pCatalogExtractDefODBC->readResult (record) )
         {
             setCurrentUser( record[0] );

             // CREATE DBPROC
             // DBPROCEDURES : OWNER,DBPROCNAME,PACKAGE,PARAMETER,CREATEDATE,CREATETIME,
             //                EXCECUTION_KIND,SQL_SUPPORT,REMOTE_LOCATION,DEFINITION,COMMENT

             StudioOAL_String  dbProcedure;
             StudioOAL_String  comment;

             select  = "SELECT DEFINITION, DBPROCNAME, OWNER, COMMENT ";
             select += "FROM DOMAIN.DBPROCEDURES ";

            /*-------------------------------------------------------------------------*
             * Prepare the WHERE clause for the SQL statements.                        *
             *-------------------------------------------------------------------------*/
             whereClause  = " WHERE definition IS NOT NULL ";

             select      += whereClause;

             if ( ! m_pDBProcedureDefODBC->executeSQL( select )  )
             {
                    Error               = m_pDBProcedureDefODBC->isError();
                    NativeError         = m_pDBProcedureDefODBC->getNativeError();
                    ErrorPos            = m_pDBProcedureDefODBC->getErrorPos();
                    ReturnCode          = m_pDBProcedureDefODBC->getReturnCode();
                    ErrorText           = m_pDBProcedureDefODBC->getErrorText();
                    NativeErrorAsString = m_pDBProcedureDefODBC->getNativeErrorAsString();

                    pMsgInstance->ls98PrintMessage(layCatalogExtract_els98, errODBCErrors_els98, NativeErrorAsString.StrPtr());

                    return ReturnCode;
             }

            /*-------------------------------------------------------------------------------*
             * Read/Write the dbprocedure definition of the specified table from/to resources*
             *-------------------------------------------------------------------------------*/
             while ( m_pDBProcedureDefODBC->readResult  (record) )
                     m_pDBProcedureDefODBC->writeResult (record) ;
         }
     }//if ( getExtrMode() == emExtractAsDDL_els31 )

     return rc;

}
/*===========================================================================*
 *  CatalogDefODBC::triggerExtrCmdExecDirect
 *===========================================================================*/
SAPDB_Int4 CatalogDefODBC::triggerExtrCmdExecDirect( StudioOAL_String owner, StudioOAL_String tablename )
{
     ROUTINE_DBG_MEO00 ("CatalogDefODBC::triggerExtrCmdExecDirect");

     RepServMessages *pMsgInstance = RepServMessages::Instance();

     SAPDB_Int4 rc = errOK_els00;

     // CREATE TRIGGER
     // TRIGGERS   : OWNER,TABLENAME,TRIGGERNAME,INSERT,UPDATE,DELETE,
     //              CREATEDATE,CREATETIME,DEFINITION,COMMENT

     StudioOAL_String  select;
     StudioOAL_String  whereClause;
     StudioOAL_String  record[5];

    /*-------------------------------------------------------------------------*
     * Prepare the WHERE clause for the SQL statement.
     *-------------------------------------------------------------------------*/

     whereClause  = " WHERE definition IS NOT NULL ";
     whereClause += " AND   owner = '";
     whereClause +=   owner;
     whereClause += "'";
     whereClause += " AND tablename = '";
     whereClause += tablename;
     whereClause += "'";

     select  = "SELECT DEFINITION, TRIGGERNAME, TABLENAME, COMMENT, OWNER ";
     select += "FROM DOMAIN.TRIGGERS ";

     select      += whereClause;

     SAPDB_Bool      isError       = SAPDB_TRUE;
     SAPDB_Long      getNativeError= 0;
     SAPDB_Int4      getErrorPos   = 0;
     SQLRETURN       getReturnCode = 0;
     StudioOAL_String getErrorText;
     StudioOAL_String getNativeErrorAsString;

     if ( !m_pTriggerDefODBC->executeSQL( select )  )
     {
           isError                = m_pTriggerDefODBC->isError();
           getNativeError         = m_pTriggerDefODBC->getNativeError();
           getErrorPos            = m_pTriggerDefODBC->getErrorPos();
           getReturnCode          = m_pTriggerDefODBC->getReturnCode();
           getErrorText           = m_pTriggerDefODBC->getErrorText();
           getNativeErrorAsString = m_pTriggerDefODBC->getNativeErrorAsString();
           return getNativeError;
     }

    /*-------------------------------------------------------------------------------*
     * Read/Write the trigger definition of the specified table from/to resources    *
     *-------------------------------------------------------------------------------*/
     while ( m_pTriggerDefODBC->readResult  (record) )
             m_pTriggerDefODBC->writeResult (record);

     return rc;
}
/*===========================================================================*
 *  CatalogDefODBC::triggerExtrCmdExecDirect
 *===========================================================================*/
SAPDB_Int4 CatalogDefODBC::triggerExtrCmdExec( StudioOAL_String owner, StudioOAL_String tablename )
{
     ROUTINE_DBG_MEO00 ("CatalogDefODBC::triggerExtrCmdExec");

     RepServMessages *pMsgInstance = RepServMessages::Instance();

     SAPDB_Int4 rc = errOK_els00;

     /*-------------------------------------------------------------------------*
     * Read/Write the synonym definition of the specified table from/to resources
     *-------------------------------------------------------------------------*/

     if ( m_pTriggerDefODBC->readTriggerDefinition  ( owner, tablename ) )
          m_pTriggerDefODBC->writeTriggerDefinition ();

     return rc;

}
#ifdef UNICODE
/*===========================================================================*
 *   ODBC_WResult::ODBC_WResult
 *===========================================================================*/
ODBC_WResult::ODBC_WResult      ( sapdbwa_DBC & odbcConnection )
             :StudioOAL_WResult ( odbcConnection.HDBC(),odbcConnection.HENV() )
{
}
/*===========================================================================*
 *    ODBC_WResult::~ODBC_WResult
 *===========================================================================*/
ODBC_WResult::~ODBC_WResult ()
{
}
#endif
#ifndef UNICODE
/*===========================================================================*
 *   ODBC_AResult::ODBC_AResult
 *===========================================================================*/
ODBC_AResult::ODBC_AResult ( sapdbwa_DBC & odbcConnection )
             :StudioOAL_AResult ( odbcConnection.HDBC(),odbcConnection.HENV() )
{
}
/*===========================================================================*
 *    ODBC_AResult::~ODBC_AResult
 *===========================================================================*/
ODBC_AResult::~ODBC_AResult ()
{
}
#endif
/*===========================================================================*
 *    CatalogObjectDefODBC::CatalogObjectDefODBC()
 *===========================================================================*/
CatalogObjectDefODBC::CatalogObjectDefODBC( sapdbwa_DBC & odbcConnection, CatalogDefODBC   * pCatalog )
                     : 
#ifdef UNICODE
                      ODBC_WResult( odbcConnection )
#endif
#ifndef UNICODE
                      ODBC_AResult( odbcConnection )
#endif 
                     , m_pCatalogDefODBC( pCatalog )
{
    // Trace - Writer
    ROUTINE_DBG_MEO00 ("CatalogObjectDefODBC::CatalogObjectDefODBC()");

    // Log   - Writer
    RepServMessages *pMsgInstance = RepServMessages::Instance();

}
/*===========================================================================*
 *    CatalogObjectDefODBC::CatalogObjectDefODBC()
 *===========================================================================*/
CatalogObjectDefODBC::CatalogObjectDefODBC( sapdbwa_DBC & odbcConnection )
                     :
#ifdef UNICODE
                      ODBC_WResult( odbcConnection )
#endif
#ifndef UNICODE
                      ODBC_AResult( odbcConnection )
#endif 
{
    // Trace - Writer
    ROUTINE_DBG_MEO00 ("CatalogObjectDefODBC::CatalogObjectDefODBC()");

    // Log   - Writer
    RepServMessages *pMsgInstance = RepServMessages::Instance();

}
/*===========================================================================*
 *    CatalogObjectDefODBC::~CatalogObjectDefODBC()
 *===========================================================================*/
CatalogObjectDefODBC::~CatalogObjectDefODBC()
{
    ROUTINE_DBG_MEO00 ("CatalogObjectDefODBC::~CatalogObjectDefODBC()");
}
/*===========================================================================*
 *     CatalogObjectDefODBC::executeSQL
 *===========================================================================*/
SAPDB_Bool CatalogObjectDefODBC::executeSQL(StudioOAL_String& statement)
{
    ROUTINE_DBG_MEO00 ("CatalogObjectDefODBC::executeSQL()");

    RepServMessages *pMsgInstance = RepServMessages::Instance();

    SAPDB_Bool isOk = SAPDB_TRUE;

    if ( !openResult (statement) )
    {
        StudioOAL_String nativeError;
        StudioOAL_String errorText;

        nativeError = getNativeErrorAsString();
        errorText   = getErrorText();

        if ( filterSQLExceptions(nativeError, errorText))
        {
            isOk = SAPDB_TRUE;
        }
        else
        {
            SAPDB_Char        ErrText[MAX_REPLY_LENGTH_LS00];
            const SAPDB_Char* ErrTextEnd = ErrText + MAX_REPLY_LENGTH_LS00;
            SAPDB_Char*       destAt;

            errorText.ConvertToASCII_Latin1(ErrText, ErrTextEnd, destAt);

            TRACE_PRNF_MLS99(("failed ", "%s\n"        , statement.StrPtr() ));
            TRACE_PRNF_MLS99(("getErrorText() ", "%s\n", errorText.StrPtr() ));

            //if ( getExtrMode() == emExtractAsDDL_els31 )
            pMsgInstance->ls98PrintMessage(layCatalogExtract_els98, msgSimpleMsg_els98 , statement.StrPtr() );
            pMsgInstance->ls98PrintMessage(layCatalogExtract_els98, errODBCErrors_els98, errorText.StrPtr() );
            //if ( getExtrMode() == emLoadAsDDL_els31 )
            //pMsgInstance->ls98PrintMessage(layCatalogLoad_els98, msgSimpleMsg_els98 , statement.StrPtr() );
            //pMsgInstance->ls98PrintMessage(layCatalogLoad_els98, errODBCErrors_els98, errorText.StrPtr() );

            isOk = SAPDB_FALSE;
        }
    }
    return  isOk;
}
/*===========================================================================*
 *     CatalogObjectDefODBC::filterSQLExceptions
 *===========================================================================*/
SAPDB_Bool CatalogObjectDefODBC::filterSQLExceptions(StudioOAL_String nativeError, StudioOAL_String errorText)
{
    SAPDB_Bool isOk = SAPDB_FALSE;

    if ( nativeError.Compare("-4004") == 0 ) //errorText.Length() >= 103  && errorText.SubStr(86,16).Compare("LOA_SAVE_CATALOG") == 0
        isOk = SAPDB_TRUE;

    return isOk;
}
/*===========================================================================*
 *     CatalogObjectDefODBC::readResult                                        *
 *===========================================================================*/
SAPDB_Bool CatalogObjectDefODBC::readResult(StudioOAL_String *record)
{
    ROUTINE_DBG_MEO00 ("CatalogObjectDefODBC::readResult()");

    /*-------------------------------------------------------------------------*
     * Fetch the next record.                                                  *
     *-------------------------------------------------------------------------*/
    if ( !fetch() )
    {
        closeResult();
        return SAPDB_FALSE;
    }

    /*-------------------------------------------------------------------------*
     * If there are no more records, reading the result is finished.           *
     *-------------------------------------------------------------------------*/
    if ( noRows() )
    {
        closeResult();
        return SAPDB_FALSE;
    }
    /*-------------------------------------------------------------------------*
     * Read all columns of the fetched record.                                 *
     *-------------------------------------------------------------------------*/
    SAPDB_Int2       i = 0;
    StudioOAL_String s;

    for (i; i < getColCount(); ++i)
    {
        getColumnDataAsString (i + 1, s);
        record[i] = s;
    }
    return SAPDB_TRUE;
}
/*===========================================================================*
 *     CatalogObjectDefODBC::writeLine
 *===========================================================================*/
SAPDB_Bool CatalogObjectDefODBC::writeLine (StudioOAL_String record )
{
    return getCatalog()->writeLine(record);
}
/*===========================================================================*
 *     CatalogObjectDefODBC::string2XML
 *===========================================================================*/
StudioOAL_String CatalogObjectDefODBC::string2XML( StudioOAL_String string )
{
   StudioOAL_String  patterns((SAPDB_UTF8*)"&\"'<>");
   StudioOAL_String  constraintexpression;

   constraintexpression += string;

   SAPDB_UInt  start = 0, end ;

   while ( (end = constraintexpression.FindFirstOf (patterns, start) ) != StudioOAL_String::NPos )
   {
       switch ( constraintexpression[end] )
       {
           case ( '&' ):
               {
               constraintexpression.Replace(end,1,"&amp;");
               end += 5;
               break;
               }
           case ( '\"' ):
               {
               constraintexpression.Replace(end,1,"&quot;");
               end += 6;
               break;
               }
           case ( '\'' ):
               {
               constraintexpression.Replace(end,1,"&apos;");
               end += 6;
               break;
               }
           case ( '<' ):
               {
               constraintexpression.Replace(end,1,"&lt;");
               end += 4;
               break;
               }
           case ( '>' ):
               {
               constraintexpression.Replace(end,1,"&gt;");
               end += 4;
               break;
               }

           default:
               break;
       }

       start = end + 1;

   }

   return constraintexpression;
}
/*===========================================================================*
 *     CatalogObjectDefODBC::EscapeSingleQuotes
 *===========================================================================*/
StudioOAL_String& CatalogObjectDefODBC::EscapeSingleQuotes(StudioOAL_String& string )
{
    StudioOAL_String  patterns((SAPDB_UTF8*)"'");

    SAPDB_UInt  start = 0;
    SAPDB_UInt  end;

    while (StudioOAL_String::NPos != (end = string.FindFirstOf(patterns, start) ) )
    {
        if ('\'' == string[end])
        {
            string.Replace(end, 1, "\'\'");
            end += 1;
        }
        start = end + 1;
    }
    return string;
}
// EscapeSingleQuotes()


/*===========================================================================*
 *    CatalogExtractDefODBC::CatalogExtractDefODBC
 *===========================================================================*/
CatalogExtractDefODBC::CatalogExtractDefODBC( sapdbwa_DBC & odbcConnection , CatalogDefODBC * pCatalog  )
                      :CatalogObjectDefODBC ( odbcConnection , pCatalog )
{
    ROUTINE_DBG_MEO00 ("CatalogExtractDefODBC::CatalogExtractDefODBC()");

    m_nObjectCount    = 0;
    m_nTableCount     = 0;
    m_nSynonymCount   = 0;
    m_nViewCount      = 0;

    initResultSet( odbcConnection );
}
/*===========================================================================*
 *    CatalogExtractDefODBC::CatalogExtractDefODBC
 *===========================================================================*/
CatalogExtractDefODBC::CatalogExtractDefODBC( sapdbwa_DBC & odbcConnection   )
                      :CatalogObjectDefODBC ( odbcConnection )
{
    ROUTINE_DBG_MEO00 ("CatalogExtractDefODBC::CatalogExtractDefODBC()");

    m_nObjectCount    = 0;
    m_nTableCount     = 0;
    m_nSynonymCount   = 0;
    m_nViewCount      = 0;

}
/*===========================================================================*
 *    CatalogExtractDefODBC::CatalogExtractDefODBC
 *===========================================================================*/
CatalogExtractDefODBC::~CatalogExtractDefODBC()
{
    ROUTINE_DBG_MEO00 ("CatalogExtractDefODBC::~CatalogExtractDefODBC()");

}
/*===========================================================================*
 *     CatalogExtractDefODBC::initResultSet
 *===========================================================================*/
SAPDB_Bool CatalogExtractDefODBC::initResultSet ( sapdbwa_DBC & odbcConnection )
{
    ROUTINE_DBG_MEO00 ("CatalogExtractDefODBC::initResult ()");

    RepServMessages *pMsgInstance  = RepServMessages::Instance();

    SAPDB_Bool       rc            = SAPDB_TRUE;
    SAPDB_Bool       rc_executeCmd = SAPDB_TRUE;

    // In case we have a connection and do not try to load catalog data - go
    if ( (odbcConnection.HDBC() && odbcConnection.HENV())      &&
         ( (emLoadAsDDL_els31 != getCatalog()->getExtrMode() ) &&
           (emLoadAsXML_els31 != getCatalog()->getExtrMode() )  )  )
    {
        StudioOAL_String  SelectCmd;
        
        //*
        //* Drop temp catalog table
        //*
        StudioOAL_String catalogDeleteCmd("DROP TABLE TEMP.LOA_SAVE_CATALOG");

        rc = executeSQL(catalogDeleteCmd);
        
        
        // create temp catalog table
        // CATALOG 7.3   : LEVEL,CMD_NO,SEQ_NO,DT_FORMAT,OWNER,TABLENAME,ERROR_CODE,SQL_CMD,TYPE,TABLETYPE,CMD

        //*
        //*     build SAVE CATALOG commands
        //* We initialize command to the case of CATALOGEXTRACT ALL and overwrite it in other cases
        //*
        StudioOAL_String catalogSaveCmd("SAVE CATALOG ALL INTO TEMP.LOA_SAVE_CATALOG");
        if (SAPDB_TRUE == rc)
        {
            if (SAPDB_TRUE == getCatalog()->getPartSpec().psSelAll_ls00)
            {
                // SAVE CATALOG ALL command

                // First check if user is sysdba; otherwise reject processing
                StudioOAL_String dbaUser = getCatalog()->getUserModeDefODBC()->getUserNameDBA();
                StudioOAL_String User; 

                SAPDB_Char   destBeg[256] = "";
                SAPDB_UInt4  destAt       = 256;

                // get UserName
                odbcConnection.GetUser( REINTERPRET_CAST(unsigned char*, destBeg ), destAt);
                User.ConvertFromASCII_Latin1(destBeg,destBeg+strlen(destBeg));

                if ( dbaUser.Compare(User) != 0 )
                {
                    pMsgInstance->ls98PrintMessage(layCatalogExtract_els98, errRestrictedStmt_els98);
                    rc_executeCmd  = SAPDB_FALSE;
                }
                // else: command already initialized to ALL
            }
            else if ( SAPDB_TRUE == getCatalog()->getPartSpec().psSelUser_ls00 )
            {
                // SAVE CATALOG USER command

                catalogSaveCmd = "SAVE CATALOG USER INTO TEMP.LOA_SAVE_CATALOG";
            }

            else    // single table
            {
                // SAVE CATALOG TABLE command

                // Build 'SAVE CATALOG' command
                catalogSaveCmd = "SAVE CATALOG OF \"";

                // In case the owner name or the table name is empty we do not issue a 
                // 'SAVE CATALOG' cmd. This is a special case for DATAEXTRACT FOR DATALOAD

                SAPDB_Int4 nTabOwnerLength = getCatalog()->getPartSpec().psTabOwnr_ls00.length();
                SAPDB_Int4 nTabNameLength  = getCatalog()->getPartSpec().psTabName_ls00.length();

                if (0 != nTabOwnerLength)
                {
                    StudioOAL_String tabowner_UTF8;
                    SAPDB_Char *pszTabOwner = getCatalog()->getPartSpec().psTabOwnr_ls00.asCharp();
                    tabowner_UTF8.ConvertFromASCII_Latin1(pszTabOwner, pszTabOwner + nTabOwnerLength);

                    catalogSaveCmd += tabowner_UTF8;
                    catalogSaveCmd += "\".\"";
                }

                StudioOAL_String tabname_UTF8;
                SAPDB_Char *pszTabName  = getCatalog()->getPartSpec().psTabName_ls00.asCharp();
                tabname_UTF8.ConvertFromASCII_Latin1(pszTabName, pszTabName + nTabNameLength);

                catalogSaveCmd += tabname_UTF8;
                catalogSaveCmd += "\" INTO TEMP.LOA_SAVE_CATALOG";
            }
        }   // end if (SAPDB_TRUE == rc)



        if ( (SAPDB_TRUE == rc) && (SAPDB_TRUE == rc_executeCmd) )
        {
            tsp00_Date  Date;
            tsp00_Time  Time;
            tsp00_Datec szDate;
            tsp00_Timec szTime;

            sqldattime(Date, Time);

            szDate.p2c(Date);
            szTime.p2c(Time);

            pMsgInstance->ls98PrintMessage(layCatalogExtract_els98, msgStartCommand_els98, szDate.asCharp(), szTime.asCharp() );
            pMsgInstance->ls98PrintMessage(layCatalogExtract_els98, msgCommandExecuting_els98, catalogSaveCmd.StrPtr()            );

            rc = executeSQL(catalogSaveCmd);
            if (SAPDB_TRUE == rc)
            {
                sqldattime(Date, Time);

                szDate.p2c(Date);
                szTime.p2c(Time);

                pMsgInstance->ls98PrintMessage(layCatalogExtract_els98, msgStopCommand_els98 , szDate.asCharp(), szTime.asCharp() );

                StudioOAL_String DropCmd("DELETE FROM TEMP.LOA_SAVE_CATALOG "
                                         "WHERE owner in ( 'SYS' ) OR TYPE = 'SYSTEM'");
                rc = executeSQL(DropCmd);
            }
                    // Get some statistics
           if (SAPDB_TRUE == rc)
           { 
                StudioOAL_String  record[1];

                SelectCmd = " SELECT COUNT (*) FROM  TEMP.LOA_SAVE_CATALOG ";
                if ( executeSQL( SelectCmd ) )
                {
                    readResult( record ) ;
                    m_nObjectCount = AtoI( record[0] );
                }

                SelectCmd = " SELECT COUNT (*) FROM  TEMP.LOA_SAVE_CATALOG  WHERE TABLETYPE = 'TABLE'"
                            "and CMD NOT LIKE 'GRANT *' ";
                if ( executeSQL( SelectCmd ) )
                {
                    readResult( record ) ;
                    m_nTableCount  = AtoI( record[0] );
                    pMsgInstance->ls98PrintMessage(layCatalogExtract_els98, msgNumTabToExtract_els98, m_nTableCount );
                }

                SelectCmd  = " SELECT COUNT (*) FROM  TEMP.LOA_SAVE_CATALOG WHERE CMD LIKE  'CREATE VIEW *' ";
                if ( executeSQL( SelectCmd ) )
                {
                    readResult( record ) ;
                    m_nViewCount   = AtoI( record[0] );
                    pMsgInstance->ls98PrintMessage(layCatalogExtract_els98, msgNumViewToExtract_els98, m_nViewCount );
                }

                SelectCmd  = " SELECT COUNT (*) FROM  TEMP.LOA_SAVE_CATALOG WHERE CMD LIKE  'CREATE SYNONYM * ' ";
                if ( executeSQL( SelectCmd ) )
                {
                    readResult( record ) ;
                    m_nSynonymCount = AtoI( record[0] );
                    pMsgInstance->ls98PrintMessage(layCatalogExtract_els98, msgNumSynonymToExtract_els98, m_nSynonymCount );
                }
            }

            // Now get the data - finally
            if (SAPDB_TRUE == rc)
            {
                StudioOAL_String  orderby(" ORDER BY OWNER");

                // LEVEL, CMD_NO, SEQ_NO, DT_FORMAT, OWNER, TABLENAME , ERROR_CODE, SQLCMD, TYPE, TABLETYPE, CMD

                SelectCmd  = "SELECT CMD_NO,OWNER,TABLENAME,CMD FROM TEMP.LOA_SAVE_CATALOG"
                             " WHERE NOT (TABLETYPE IN ('VIEW', 'SYNONYM') AND SEQ_NO > 0)";

                if ( getCatalog()->getExtrMode() == emExtractAsXML_els31 )
                {
                    SelectCmd += orderby;
                }

                rc = executeSQL( SelectCmd );
            }
        }   // end if ( (SAPDB_TRUE == rc) && (SAPDB_TRUE == rc_executeCmd) )

        if (SAPDB_FALSE == rc)
        {
            SAPDB_Bool       Error               = isError();
            SAPDB_Long       NativeError         = getNativeError();
            SAPDB_Int4       ErrorPos            = getErrorPos();
            SQLRETURN        ReturnCode          = getReturnCode();
            StudioOAL_String ErrorText           = getErrorText();;
            StudioOAL_String NativeErrorAsString = getNativeErrorAsString();
        }
        else
        {
            if (SAPDB_FALSE == rc_executeCmd)
            {
                rc = SAPDB_FALSE;
            }
        }

    }   // end if (   odbcConnection.HDBC() && odbcConnection.HENV() )



    return rc;
}
/*===========================================================================*
 *    CatalogExtractDefODBC::readResult
 *===========================================================================*/
SAPDB_Bool CatalogExtractDefODBC::readResult ( StudioOAL_String *record  )
{
    ROUTINE_DBG_MEO00 ("CatalogExtractDefODBC::readResult ()");

    return CatalogObjectDefODBC::readResult( record);
}
/*===========================================================================*
 *    CatalogExtractDefODBC::writeResult
 *===========================================================================*/
SAPDB_Bool CatalogExtractDefODBC::writeResult( StudioOAL_String record  )
{
    ROUTINE_DBG_MEO00 ("CatalogExtractDefODBC::writeResult ()");

    return CatalogObjectDefODBC::writeLine ( record);
}
/*===========================================================================*
 *    CatalogLoadDefODBC::CatalogLoadDefODBC
 *===========================================================================*/
CatalogLoadDefODBC::CatalogLoadDefODBC   ( sapdbwa_DBC & odbcConnection , CatalogDefODBC * pCatalog  )
                   :CatalogObjectDefODBC ( odbcConnection , pCatalog )
{
    ROUTINE_DBG_MEO00 ("CatalogLoadDefODBC::CatalogLoadDefODBC()");

    initResultSet( odbcConnection );

}
/*===========================================================================*
 *    CatalogLoadDefODBC::CatalogLoadDefODBC
 *===========================================================================*/
CatalogLoadDefODBC::CatalogLoadDefODBC   ( sapdbwa_DBC & odbcConnection )
                   :CatalogObjectDefODBC ( odbcConnection )
{
    ROUTINE_DBG_MEO00 ("CatalogLoadDefODBC::CatalogLoadDefODBC()");
}
/*===========================================================================*
 *    CatalogLoadDefODBC::CatalogLoadDefODBC
 *===========================================================================*/
CatalogLoadDefODBC::~CatalogLoadDefODBC()
{
    ROUTINE_DBG_MEO00 ("CatalogLoadDefODBC::~CatalogLoadDefODBC()");
}
/*===========================================================================*
 *     CatalogLoadDefODBC::initResultSet
 *===========================================================================*/
SAPDB_Bool CatalogLoadDefODBC::initResultSet ( sapdbwa_DBC & odbcConnection )
{
  ROUTINE_DBG_MEO00 ("CatalogLoadDefODBC::initResult ()");

  SAPDB_Bool isOk = SAPDB_FALSE;

  if ( getCatalog()->getExtrMode() == emExtractAsDDL_els31 ||
       getCatalog()->getExtrMode() == emExtractAsXML_els31    )

       return SAPDB_TRUE;

  if (  odbcConnection.HDBC() && odbcConnection.HENV() )
  {
        isOk = SAPDB_TRUE;
  }

  return isOk;
}
/*===========================================================================*
 *    CatalogLoadDefODBC::setSQLMode
 *===========================================================================*/
void CatalogLoadDefODBC::setSQLMode( StudioOAL_String  record )
{
    //typedef enum tpr00_SQLModeEnum {
    //            cpr_kind_empty         =   0,
    //            cpr_kind_internal      =   1,
    //            cpr_kind_db2           =   2,
    //            cpr_kind_ansi          =   3,
    //            cpr_kind_oracle        =   4,
    //            cpr_kind_sapr3         =   5
    //          } tpr00_SQLModeEnum; 

    StudioOAL_String internal = "INTERNAL";
    StudioOAL_String db2      = "DB2";
    StudioOAL_String ansi     = "ANSI";
    StudioOAL_String oracle   = "ORACLE";
    StudioOAL_String sapr3    = "SAPR3";

    if ( internal.Compare(record.SubStr(0,8) ) == 0 )
    {
         setSqlMode( cpr_kind_internal );
         m_sqlmode = internal;
    }
    if ( db2     .Compare(record.SubStr(0,3) ) == 0 )
    {
         setSqlMode( cpr_kind_db2 );
         m_sqlmode = db2;
    }
    if ( ansi    .Compare(record.SubStr(0,4) ) == 0 )
    {
         setSqlMode( cpr_kind_ansi );
         m_sqlmode = ansi ;
    }
    if ( oracle  .Compare(record.SubStr(0,6) ) == 0 )
    {
         setSqlMode( cpr_kind_oracle );
         m_sqlmode = oracle;
    }
    if ( sapr3   .Compare(record.SubStr(0,5) ) == 0 )
    {
         setSqlMode( cpr_kind_sapr3 );
         m_sqlmode = sapr3;
    }
         
}
/*===========================================================================*
 *    CatalogLoadDefODBC::getSQLMode
 *===========================================================================*/
StudioOAL_String CatalogLoadDefODBC::getSQLMode ( )
{
    return m_sqlmode;
}
/*===========================================================================*
 *    CatalogLoadDefODBC::readResult
 *===========================================================================*/
SAPDB_Bool CatalogLoadDefODBC::readResult ( StudioOAL_String *record  )
{
    ROUTINE_DBG_MEO00 ("CatalogLoadDefODBC::readResult ()");

    return CatalogObjectDefODBC::readResult( record);
}
/*===========================================================================*
 *    CatalogLoadDefODBC::writeResult
 *===========================================================================*/
SAPDB_Bool CatalogLoadDefODBC::writeResult( StudioOAL_String record  )
{
    ROUTINE_DBG_MEO00 ("CatalogLoadDefODBC::writeResult ()");

    return CatalogObjectDefODBC::writeLine ( record);
}
/*===========================================================================*
 *    UserDefODBC::UserDefODBC
 *===========================================================================*/
UserDefODBC::UserDefODBC( sapdbwa_DBC & odbcConnection , CatalogDefODBC * pCatalog  )
            :CatalogObjectDefODBC ( odbcConnection , pCatalog )
{
    ROUTINE_DBG_MEO00 ("UserDefODBC::UserDefODBC()");

    initResultSet( odbcConnection );

}
/*===========================================================================*
 *    UserDefODBC::UserDefODBC
 *===========================================================================*/
UserDefODBC::UserDefODBC( sapdbwa_DBC & odbcConnection )
            :CatalogObjectDefODBC ( odbcConnection )
{
}
/*===========================================================================*
 *    UserDefODBC::~UserDefODBC
 *===========================================================================*/
UserDefODBC::~UserDefODBC()
{
   ROUTINE_DBG_MEO00 ("UserDefODBC::~UserDefODBC()");
}
/*===========================================================================*
 *     UserDefODBC::initResultSet
 *===========================================================================*/
SAPDB_Bool UserDefODBC::initResultSet ( sapdbwa_DBC & odbcConnection )
{
    SAPDB_Bool isOk = SAPDB_FALSE;

    if ( odbcConnection.HDBC() && odbcConnection.HENV() )
    {
        // DOMAIN.USERS: OWNER,GROUPNAME,USERNAME,USERMODE,CONNECTMODE,MAXTIMEOUT,COSTWARNING,COSTLIMIT,CREATEDATE,CREATETIME,ALTERDATE,ALTERTIME,PWCREADATE,PWCREATIME,SERVERDB,SERVERNODE,USER_ID,COMMENT

        StudioOAL_String  userDef;
        StudioOAL_String  whereClause;
        StudioOAL_String  orderClause;

        userDef = "SELECT OWNER, GROUPNAME, USERNAME, USERMODE, MAXTIMEOUT, COSTWARNING, COSTLIMIT, CONNECTMODE, COMMENT "
                  "FROM DOMAIN.USERS ";

        /*-------------------------------------------------------------------------*
         * Prepare the WHERE clause for the SQL statements.                        *
         *-------------------------------------------------------------------------*/

        // catalog extract all
        if  ( getCatalog()->getPartSpec().psSelAll_ls00 )
        {
            whereClause  = " WHERE USERMODE <> 'SYSDBA'  "
                           " AND   USERMODE <> 'COLDUSER'"
                           " AND   USERNAME <> 'DOMAIN'  "
                           " AND   USERNAME <> 'SYS'     ";
            orderClause  = " ORDER BY OWNER, GROUPNAME, USERNAME ASC ";
        }
        // catalog extract user
        else if ( getCatalog()->getPartSpec().psSelUser_ls00 )
        {
            whereClause  = "WHERE OWNER = ? ";
            orderClause  = " ORDER BY GROUPNAME , USERNAME ASC ";
        }

        userDef += whereClause;
        userDef += orderClause;

        //m_owner;
        //m_groupName;
        //m_userName;
        //m_userMode;
        //m_maxTimeOut;
        //m_costWarning;
        //m_costLimit;
        //m_connectMode;
        //m_comment;

        isOk = SAPDB_TRUE;
    }
    return isOk;
}
/*===========================================================================*
 *    UserDefODBC::readResult
 *===========================================================================*/
SAPDB_Bool UserDefODBC::readResult (StudioOAL_String *record)
{
    ROUTINE_DBG_MEO00 ("UserDefODBC::readResult()");

    SAPDB_Bool isOk = SAPDB_FALSE;

    isOk = CatalogObjectDefODBC::readResult( record);

    // do this if using prepared statements

    m_owner       = record[0];
    m_groupName   = record[1];
    m_userName    = record[2];
    m_userMode    = record[3];
    m_maxTimeOut  = record[4];
    m_costWarning = record[5];
    m_costLimit   = record[6];
    m_connectMode = record[7];
    m_comment     = record[8];

    return isOk;;
}
/*===========================================================================*
 *    UserDefODBC::writeResult
 *===========================================================================*/
SAPDB_Bool UserDefODBC::writeResult( StudioOAL_String *record ,tls31_ExtrMode catextrMode )
{
    ROUTINE_DBG_MEO00 ("UserDefODBC::writeResult ()");

    StudioOAL_String user;
    StudioOAL_String comment = SEPARATOR;

    //select       =  "SELECT OWNER, GROUPNAME, USERNAME, USERMODE, MAXTIMEOUT, COSTWARNING, COSTLIMIT, CONNECTMODE, COMMENT ";
    //select      +=  "FROM DOMAIN.USERS ";

    // USERGROUP <=> if ( !record[1].Empty() &&  record[2].Empty() )
    // GROUPUSER <=> if ( !record[1].Empty() && !record[2].Empty() )
    // USER      <=> if (  record[1].Empty() && !record[2].Empty() )

    if ( catextrMode == emExtractAsDDL_els31 )
    {
        // <create_usergroup_statement> ::=
        // CREATE USERGROUP <usergroup_name>
        // [<usergroup_mode>]
        // [TIMEOUT   <unsigned_integer>] [COSTWARNING <unsigned_integer>]
        // [COSTLIMIT <unsigned_integer>] [[NOT] EXCLUSIVE]
        // [<usergroup_mode>] ::= RESOURCE | STANDARD

        //<create_user_statement> ::=
        // CREATE USER <user_name> PASSWORD <password> USERGROUP <usergroup_name>

        //<create_user_statement> ::=
        // CREATE USER <user_name> PASSWORD <password>
        //[<user_mode>]
        //[TIMEOUT <unsigned_integer>] [COSTWARNING <unsigned_integer>]
        //[COSTLIMIT <unsigned_integer>] [[NOT] EXCLUSIVE]
        // <user_mode> :: = DBA | RESOURCE | STANDARD

        // GROUPUSER
        if ( !record[1].Empty() && !record[2].Empty() )
        {
             user  = SEPARATOR;

             if ( getCatalog()->getSqlMode() != cpr_kind_internal )
             {
                  user += pszSQLModeString_ls31[cpr_kind_internal];
                  user += SEPARATOR;
             }

             user += "CREATE USER ";
             // <user_name>
             user  += record[2];

             // <password>
             user  += " PASSWORD ";
             user  += "initial"   ;

             user  += " USERGROUP ";
             user  += record[1];

             writeLine( user );

             // COMMENT ON USER <user_name> IS <comment>
             if ( !record[10].Empty() &&  !getColumn(11)->getColIsNull() ) // record[10].Compare( NULL_VALUE ) != 0
             {
                  comment += "COMMENT ON USER ";
                  comment += record[2];
                  comment += " IS ";
                  comment += EscapeSingleQuotes(record[10]);

                  writeLine( comment );
             }
             // <grant_user_statement> ::= GRANT USER  <granted_users> [FROM <user_name>] TO <user_name>
        }
        else
        {
            user  = SEPARATOR;

            if ( getCatalog()->getSqlMode()  != cpr_kind_internal )
            {
                  user += pszSQLModeString_ls31[cpr_kind_internal];
                  user += SEPARATOR;
            }
            // USERGROUP
            if ( !record[1].Empty() &&  record[2].Empty() )
            {
                 user += "CREATE USERGROUP ";
                 // <usergroup_name>
                 user += record[1];
            }
            // USER
            if (  record[1].Empty() && !record[2].Empty() )
            {
                 user += "CREATE USER ";
                 // <user_name>
                 user  += record[2];

                 // <password>
                 user  += " PASSWORD ";
                 user  += "initial" ;
            }

            // [<usergroup_mode>] ::= RESOURCE | STANDARD
            // [<user_mode>     ]:: = DBA | RESOURCE | STANDARD

            if ( record[3].Empty() || getColumn(4)->getColIsNull() ) // record[3].Compare( NULL_VALUE ) == 0 )
                 user += "" ;
            else
            {
                 user += "  ";
                 user += record[3];
            }
            // [TIMEOUT]
            if (record[4].Empty() || getColumn(5)->getColIsNull() ) // record[4].Compare( NULL_VALUE ) == 0)
                user += "" ;
            else
            {
                user += " TIMEOUT   ";
                user += record[4];
            }
            // [COSTWARNING]
            if ( record[5].Empty() || getColumn(6)->getColIsNull() ) // record[5].Compare( NULL_VALUE ) == 0 )
                 user += "" ;
            else
            {
                 user += " COSTWARNING ";
                 user += record[5];
            }
            // [COSTLIMIT]
            if ( record[6].Empty() || getColumn(7)->getColIsNull() ) // record[6].Compare( NULL_VALUE ) == 0 )
                 user += "" ;
            else
            {
                 user += " COSTLIMIT ";
                 user += record[6];
            }
            // [[NOT] EXCLUSIVE]
            if ( record[7].Empty() || getColumn(8)->getColIsNull() ) // record[7].Compare( NULL_VALUE ) == 0)
                 user += "" ;
            else
            {
                if ( record[7].Compare("SINGLE") == 0 )
                     user += " EXCLUSIVE " ;
                else
                     user += " NOT EXCLUSIVE ";

            }

            writeLine(user);

            // USERGROUP
            if ( !record[1].Empty() &&  record[2].Empty() )
            {
                 // COMMENT ON USERGROUP <usergroup_name> IS <comment>
                 if ( !record[8].Empty() &&  !getColumn(9)->getColIsNull() ) // record[8].Compare( NULL_VALUE ) != 0)
                 {
                      comment += "COMMENT ON USERGROUP ";
                      comment += record[1];
                      comment += " IS ";
                      comment += EscapeSingleQuotes(record[8]);
                      writeLine(comment  );
                 }

            }
            // USER
            if (  record[1].Empty() && !record[2].Empty() && !getColumn(9)->getColIsNull() ) //record[8].Compare( NULL_VALUE ) != 0)
            {
                  // COMMENT ON USER <user_name> IS <comment>
                  comment += "COMMENT ON USER ";
                  comment += record[2];
                  comment += " IS ";
                  comment += EscapeSingleQuotes(record[8]);
                  writeLine(comment  );
            }
        }

        // <grant_usergroup_statement> ::= GRANT USERGROUP <granted_usergroups> [FROM <user_name>] TO <user_name>
        // <grant_user_statement>      ::= GRANT USER      <granted_users>      [FROM <user_name>] TO <user_name>
    }
    else if ( catextrMode == emExtractAsXML_els31 )
    {
        StudioOAL_String schema;
        StudioOAL_String label;
        StudioOAL_String name;


        // USERGROUP
        if ( !record[1].Empty() &&  record[2].Empty() )
        {
            label = "USERGROUP";
            name  = record[1];
        }
        // GROUPUSER
        if ( !record[1].Empty() && !record[2].Empty() )
        {
            label = "GROUPUSER";
            name  = record[1];
        }
        // USER
        if (  record[1].Empty() && !record[2].Empty() )
        {
            label = "USER";
            name  = record[2];
        }

        // Uml:NamespaceOwnedElement
        schema  = "<uml:NamespaceOwnedElement>\n";

        // sapdbm:Schema
        schema += "<sapdbm:DeployedSchema oim:id=\"_";
        !getColumn(1)->getColIsNull() ? schema += record[0] : schema += ""; // record[0].Compare( NULL_VALUE )  != 0
        schema += ".";
        if ( !record[1].Empty() )
        {
              schema += record[1];
              schema += ".";
        }
        schema += record[2];
        schema += "\" oim:supertype=\"dbm:DeployedSchema\"";
        schema += " oim:label=\"";
        schema += label;
        schema += "\" oim:name=\"";
        schema += name;
        schema += "\" sapdbm:Owner=\"";
        !getColumn(1)->getColIsNull() ? schema += record[0] : schema += ""; // record[0].Compare( NULL_VALUE )  != 0
        schema += "\" sapdbm:Groupname=\"";
        !getColumn(2)->getColIsNull() ? schema += record[1] : schema += ""; // record[1].Compare( NULL_VALUE )  != 0
        schema += "\" sapdbm:Username=\"";
        !getColumn(3)->getColIsNull() ? schema += record[2] : schema += ""; // record[2].Compare( NULL_VALUE )  != 0
        schema += "\" sapdbm:Usermode=\"";
        !getColumn(4)->getColIsNull() ? schema += record[3] : schema += ""; // record[3].Compare( NULL_VALUE )  != 0
        schema += "\" sapdbm:MaxTimeOut=\"";
        !getColumn(5)->getColIsNull() ? schema += record[4] : schema += ""; // record[4].Compare( NULL_VALUE )  != 0
        schema += "\" sapdbm:Costwarning=\"";
        !getColumn(6)->getColIsNull() ? schema += record[5] : schema += ""; // record[5].Compare( NULL_VALUE )  != 0
        schema += "\" sapdbm:Costlimit=\"";
        !getColumn(7)->getColIsNull() ? schema += record[6] : schema += ""; // record[6].Compare( NULL_VALUE )  != 0
        schema += "\" sapdbm:Connectmode=\"";
        !getColumn(8)->getColIsNull()? schema += record[7] : schema += ""; // record[7].Compare( NULL_VALUE )  != 0
        schema += "\" sapdbm:Comment=\"";
        !getColumn(9)->getColIsNull()? schema += record[8]: schema += ""; // record[8].Compare( NULL_VALUE )  != 0
        schema += "\">\n";

        schema += "<uml:NamespaceOwnedElement>\n";

        writeLine( schema );
    }

    return SAPDB_TRUE;
}
/*===========================================================================*
 *    UserModeDefODBC::UserModeDefODBC
 *===========================================================================*/
UserModeDefODBC::UserModeDefODBC( sapdbwa_DBC & odbcConnection , CatalogDefODBC * pCatalog  )
                :CatalogObjectDefODBC ( odbcConnection , pCatalog )
{
    ROUTINE_DBG_MEO00 ("UserModeDefODBC::UserModeDefODBC()");

    initResultSet( odbcConnection );
}
/*===========================================================================*
 *    UserModeDefODBC
 *===========================================================================*/
UserModeDefODBC::UserModeDefODBC( sapdbwa_DBC & odbcConnection )
               :CatalogObjectDefODBC ( odbcConnection )
{
    ROUTINE_DBG_MEO00 ("UserModeDefODBC::UserModeDefODBC()");

}
/*===========================================================================*
 *    UserModeDefODBC::UserModeDefODBC
 *===========================================================================*/
UserModeDefODBC::~UserModeDefODBC()
{
    ROUTINE_DBG_MEO00 ("UserModeDefODBC::~UserModeDefODBC()");
}
/*===========================================================================*
 *     UserModeDefODBC::initResultSet
 *===========================================================================*/
SAPDB_Bool UserModeDefODBC::initResultSet ( sapdbwa_DBC & odbcConnection )
{
    ROUTINE_DBG_MEO00 ("DBADefODBC::initResult ()");

    SAPDB_Bool rcBool = SAPDB_TRUE;

    if (  odbcConnection.HDBC() && odbcConnection.HENV() )
    {
        StudioOAL_String  usermodeDef;
        StudioOAL_String  record[1];

        // Get SYSDBA name
        usermodeDef = "SELECT USERNAME FROM DOMAIN.USERS WHERE USERMODE IN ( 'SYSDBA' )";
        
        rcBool = executeSQL( usermodeDef );
        if (SAPDB_TRUE == rcBool)
        {
            rcBool = readResult( record );
            if (SAPDB_TRUE == rcBool)
            {
                m_usernameDBA = record[0];
            }
        }

        // Get DBM (COLDUSER) name
        if (SAPDB_TRUE == rcBool)
        {
            usermodeDef = "SELECT USERNAME FROM DOMAIN.USERS WHERE USERMODE IN ( 'COLDUSER' )";

            rcBool = executeSQL( usermodeDef );
            if (SAPDB_TRUE == rcBool)
            {
                rcBool = readResult( record );
                if (SAPDB_TRUE == rcBool)
                {
                    m_usernameDBM = record[0];
                }
            }
        }

        // Handle error situation
        if (SAPDB_FALSE == rcBool)
        {
            getCatalog()->getError()->setErrorText(getErrorText());
            getCatalog()->getError()->isError(isError());
        }
    }

    return rcBool;
}
/*===========================================================================*
 *    UserModeDefODBC::readResult
 *===========================================================================*/
SAPDB_Bool UserModeDefODBC::readResult ( StudioOAL_String *record  )
{
    ROUTINE_DBG_MEO00 ("UserModeDefODBC::readResult ()");

    return CatalogObjectDefODBC::readResult( record);
}
/*===========================================================================*
 *    CURUserDefODBC::CURUserDefODBC
 *===========================================================================*/
CURUserDefODBC::CURUserDefODBC( sapdbwa_DBC & odbcConnection , CatalogDefODBC * pCatalog  )
               :CatalogObjectDefODBC ( odbcConnection , pCatalog )
{
    ROUTINE_DBG_MEO00 ("CURUserDefODBC::CURUserDefODBC()");

    initResultSet( odbcConnection );

}
/*===========================================================================*
 *    CURUserDefODBC::CURUserDefODBC
 *===========================================================================*/
CURUserDefODBC::CURUserDefODBC( sapdbwa_DBC & odbcConnection )
               :CatalogObjectDefODBC ( odbcConnection )
{
    ROUTINE_DBG_MEO00 ("CURUserDefODBC::CURUserDefODBC()");

}
/*===========================================================================*
 *    CURUserDefODBC::CURUserDefODBC
 *===========================================================================*/
CURUserDefODBC::~CURUserDefODBC()
{
    ROUTINE_DBG_MEO00 ("CURUserDefODBC::~CURUserDefODBC()");
}
/*===========================================================================*
 *     CURUserDefODBC::initResultSet
 *===========================================================================*/
SAPDB_Bool CURUserDefODBC::initResultSet ( sapdbwa_DBC & odbcConnection )
{
  ROUTINE_DBG_MEO00 ("CURUserDefODBC::initResult ()");

  SAPDB_Bool isOk = SAPDB_FALSE;

  if (  odbcConnection.HDBC() && odbcConnection.HENV() )
  {
        isOk = SAPDB_TRUE;
  }

  return isOk;
}
/*===========================================================================*
 *    CURUserDefODBC::writeResult
 *===========================================================================*/
SAPDB_Bool CURUserDefODBC::writeResult( StudioOAL_String record  )
{
    ROUTINE_DBG_MEO00 ("CURUserDefODBC::writeResult ()");

    return CatalogObjectDefODBC::writeLine ( record);
}
/*===========================================================================*
 *    RoleDefODBC::RoleDefODBC
 *===========================================================================*/
RoleDefODBC::RoleDefODBC( sapdbwa_DBC & odbcConnection , CatalogDefODBC * pCatalog  )
            :CatalogObjectDefODBC ( odbcConnection , pCatalog )
{
    ROUTINE_DBG_MEO00 ("RoleDefODBC::RoleDefODBC()");

    initResultSet( odbcConnection );

}
/*===========================================================================*
 *    RoleDefODBC ::RoleDefODBC
 *===========================================================================*/
RoleDefODBC::RoleDefODBC( sapdbwa_DBC & odbcConnection )
            :CatalogObjectDefODBC ( odbcConnection )
{
    ROUTINE_DBG_MEO00 ("RoleDefODBC::RoleDefODBC()");

}
/*===========================================================================*
 *    RoleDefODBC::RoleDefODBC
 *===========================================================================*/
RoleDefODBC::~RoleDefODBC()
{
    ROUTINE_DBG_MEO00 ("RoleDefODBC::~RoleDefODBC()");
}
/*===========================================================================*
 *     RoleDefODBC::initResultSet
 *===========================================================================*/
SAPDB_Bool RoleDefODBC::initResultSet ( sapdbwa_DBC & odbcConnection )
{
    SAPDB_Bool isOk = SAPDB_FALSE;

    if (  odbcConnection.HDBC() && odbcConnection.HENV() )
    {
        // <grantee> ::= PUBLIC | <user_name> | <usergroup_name> | <role_name>

        // ROLES          :OWNER,ROLE,PASSWORD_REQUIRED,GRANTED,CREATEDATE,CREATEDATE

        StudioOAL_String  roleDef("SELECT ROLE,PASSWORD_REQUIRED FROM DOMAIN.ROLES ");
        StudioOAL_String  whereClause;

        if  ( getCatalog()->getPartSpec().psSelUser_ls00 )
        {
            whereClause  = "WHERE OWNER = ? ";
        }

        roleDef  += whereClause;
        isOk = SAPDB_TRUE;
    }
    return isOk;
}
/*===========================================================================*
 *    RoleDefODBC::readResult
 *===========================================================================*/
SAPDB_Bool RoleDefODBC::readResult (StudioOAL_String *record)
{
    return CatalogObjectDefODBC::readResult( record);
}
/*===========================================================================*
 *    RoleDefODBC::writeResult
 *===========================================================================*/
SAPDB_Bool RoleDefODBC::writeResult( StudioOAL_String *record )
{
    ROUTINE_DBG_MEO00 ("RoleDefODBC::writeResult ()");

    if ( getCatalog()->getExtrMode() == emExtractAsDDL_els31 )
    {
        // <create_role_statement> ::= CREATE ROLE <role_name> [IDENTIFIED BY <password>]

        StudioOAL_String role;

        role  = SEPARATOR;
        role +="CREATE ROLE ";

        // <role_name>
        role += record[0];

        // [IDENTIFIED BY <password>]
        if ( record[1].Empty() || getColumn(2)->getColIsNull() ) // record[1].Compare( NULL_VALUE ) == 0 )
             role += "" ;
        else
        {
             role += " IDENTIFIED BY ";

             // <password>
             role += "initial" ;
        }

        return writeLine(role);
    }

    return SAPDB_TRUE;
}
/*===========================================================================*
 *    SequenceDefODBC::SequenceDefODBC
 *===========================================================================*/
SequenceDefODBC::SequenceDefODBC( sapdbwa_DBC & odbcConnection , CatalogDefODBC * pCatalog  )
                :CatalogObjectDefODBC ( odbcConnection , pCatalog )
{
    ROUTINE_DBG_MEO00 ("SequenceDefODBC::SequenceDefODBC()");

    initResultSet( odbcConnection );
}
/*===========================================================================*
 *    SequenceDefODBC::SequenceDefODBC
 *===========================================================================*/
SequenceDefODBC::SequenceDefODBC(  sapdbwa_DBC & odbcConnection )
                :CatalogObjectDefODBC ( odbcConnection )
{
    ROUTINE_DBG_MEO00 ("SequenceDefODBC::SequenceDefODBC()");

}
/*===========================================================================*
 *    SequenceDefODBC::SequenceDefODBC
 *===========================================================================*/
SequenceDefODBC::~SequenceDefODBC()
{
    ROUTINE_DBG_MEO00 ("SequenceDefODBC::~SequenceDefODBC()");
}
/*===========================================================================*
 *     SequenceDefODBC::initResultSet
 *===========================================================================*/
SAPDB_Bool SequenceDefODBC::initResultSet ( sapdbwa_DBC & odbcConnection )
{
    ROUTINE_DBG_MEO00 ("SequenceDefODBC::initResult ()");

    SAPDB_Bool isOk = SAPDB_FALSE;

    if (  odbcConnection.HDBC() && odbcConnection.HENV() )
    {
        // CREATE SEQUENCE
        // SEQUENCES: OWNER, SEQUENCE_NAME, INCREMENT_BY, LAST_NUMBER, MAX_VALUE, MIN_VALUE,
        //            CYCLE_FLAG, CACHE_SIZE, ORDER_FLAG, CREATEDATE, CREATETIME, COMMENT
        
        // START_WITH = LAST_NUMBER+1
        StudioOAL_String  sequenceDef("SELECT OWNER, SEQUENCE_NAME, INCREMENT_BY, LAST_NUMBER+1,"
                                      " MAX_VALUE, MIN_VALUE, CYCLE_FLAG, CACHE_SIZE, ORDER_FLAG"
                                      " FROM DOMAIN.SEQUENCES WHERE  OWNER = ?");

        isOk = SAPDB_TRUE;
    }
    return isOk;
}
/*===========================================================================*
 *    SequenceDefODBC::readResult
 *===========================================================================*/
SAPDB_Bool SequenceDefODBC::readResult ( StudioOAL_String *record  )
{
    ROUTINE_DBG_MEO00 ("SequenceDefODBC::readResult ()");

    return CatalogObjectDefODBC::readResult( record);
}
/*===========================================================================*
 *    SequenceDefODBC::writeResult
 *===========================================================================*/
SAPDB_Bool SequenceDefODBC::writeResult( StudioOAL_String *record  )
{
    ROUTINE_DBG_MEO00 ("SequenceDefODBC::writeResult ()");

    RepServMessages *pMsgInstance = RepServMessages::Instance();

    if ( getCatalog()->getExtrMode() == emExtractAsDDL_els31 )
    {

        StudioOAL_String sequence;

        // SEQUENCES  : OWNER,SEQUENCE_NAME,INCREMENT_BY,LAST_NUMBER,MAX_VALUE,MIN_VALUE,
        //              CYCLE_FLAG,CACHE_SIZE,ORDER_FLAG,CREATEDATE,CREATETIME,COMMENT

        //<create_sequence_statement> ::= CREATE SEQUENCE [<owner>.]<sequence_name>
        //[INCREMENT BY <integer>] [START WITH <integer>]
        //[MAXVALUE <integer> | NOMAXVALUE] [MINVALUE <integer> | NOMINVALUE]
        //[CYCLE | NOCYCLE]
        //[CACHE <unsigned_integer> | NOCACHE]
        //[ORDER | NOORDER]

        sequence  =  SEPARATOR;

        // Special handling of 'foreign' sqlmodes
                
        if ( getCatalog()->getSqlMode() != cpr_kind_internal )
        {
             SAPDB_Int4 iSqlMode = getCatalog()->getSqlMode();
             sequence += pszSQLModeString_ls31[iSqlMode];
             sequence += SEPARATOR;
        }
        
        sequence += "CREATE SEQUENCE  ";

        // [<owner>.]
        if ( record[0].Empty() || getColumn(1)->getColIsNull() ) //  record[0].Compare( NULL_VALUE ) == 0
             sequence += "" ;
        else
        {
             sequence += "\"";
             sequence += record[0] ;
             sequence += "\".";
        }
        // <sequence_name>
        if ( record[1].Empty() || getColumn(2)->getColIsNull() ) // record[1].Compare( NULL_VALUE ) == 0
             sequence += "" ;
        else
        {
             sequence += "\"";
             sequence += record[1] ;
             sequence += "\"";
        }
        // [INCREMENT BY <integer>]
        if ( record[2].Empty() || getColumn(3)->getColIsNull() ) // record[2].Compare( NULL_VALUE ) == 0
             sequence += " " ;
        else
        {
             sequence += " INCREMENT BY ";
             sequence += record[2];
        }
        // [START WITH <integer>]
        if ( record[3].Empty()|| getColumn(4)->getColIsNull() ) //  record[3].Compare( NULL_VALUE ) == 0
             sequence += " " ;
        else
        {
             //!! ToDo 'STARTWITH' COLUMN SEQUENCES.DOMAIN
             sequence   += " START WITH ";
             sequence   += record[3];
        }
        // [MAXVALUE <integer> | NOMAXVALUE]
        if ( record[4].Empty() || getColumn(5)->getColIsNull() ) // record[4].Compare( NULL_VALUE ) == 0
            sequence += " NOMAXVALUE ";
        else
        {
            sequence += " MAXVALUE " ;
            sequence += record[4];
        }
        // [MINVALUE <integer> | NOMINVALUE]
        if (  record[5].Empty() || getColumn(6)->getColIsNull() ) // record[5].Compare( NULL_VALUE ) == 0
             sequence += " NOMINVALUE ";
        else
        {
             sequence += " MINVALUE ";
             sequence += record[5];
        }
        // [CYCLE | NOCYCLE]
        if ( record[6].Empty() ||  getColumn(7)->getColIsNull() ) // record[6].Compare( NULL_VALUE ) == 0
             sequence += " NOCYCLE ";
        else
        {
             if ( record[6].Compare("Y") == 0 )
                  sequence += " CYCLE ";
             else
                  sequence += " NOCYCLE ";

        }
        // [CACHE <unsigned_integer> | NOCACHE]
        if ( record[7].Empty() || getColumn(8)->getColIsNull() )  // record[7].Compare( NULL_VALUE ) == 0
             sequence += " NOCACHE ";
        else
        {
             if ( record[7].Compare("0") == 0 )
                 sequence += " NOCACHE ";
             else
             {
                 sequence += " CACHE " ;
                 sequence += record[7];
             }

        }
        // [ORDER | NOORDER]
        if ( record[8].Empty() || getColumn(9)->getColIsNull() ) // record[8].Compare( NULL_VALUE ) == 0
             sequence += " NOORDER ";
        else
        {
            if ( record[8].Compare("Y") == 0 )
                 sequence += " ORDER " ;
            else
                 sequence += " NOORDER ";
        }

        return writeLine(sequence);
    }

    // COMMENT ON not described in reference manual

    return SAPDB_TRUE;

}
/*===========================================================================*
 *    TableDefODBC::TableDefODBC
 *===========================================================================*/
TableDefODBC::TableDefODBC( sapdbwa_DBC & odbcConnection , CatalogDefODBC * pCatalog  )
             :CatalogObjectDefODBC ( odbcConnection , pCatalog )
{
    // Trace - Writer
    ROUTINE_DBG_MEO00 ("TableDefODBC::TableDefODBC()");

    // Log   - Writer
    RepServMessages *pMsgInstance = RepServMessages::Instance();

    setCaseSensitive (SAPDB_TRUE);

#ifdef  UNICODE
    m_pColumnDefResult     = new ODBC_WResult ( odbcConnection );
    m_pConstraintDefResult = new ODBC_WResult ( odbcConnection );
    m_pForeignKeyDefResult = new ODBC_WResult ( odbcConnection );
    m_pTableAttrDefResult  = new ODBC_WResult ( odbcConnection );
#endif
#ifndef UNICODE
    m_pColumnDefResult     = new ODBC_AResult ( odbcConnection );
    m_pConstraintDefResult = new ODBC_AResult ( odbcConnection );
    m_pForeignKeyDefResult = new ODBC_AResult ( odbcConnection );
    m_pTableAttrDefResult  = new ODBC_AResult ( odbcConnection );                
#endif

    initResultSet( odbcConnection );

    setType (ClassTableDefODBC);

}
/*===========================================================================*
 *     TableDefODBC::TableDefODBC()
 *===========================================================================*/
 TableDefODBC::TableDefODBC( sapdbwa_DBC & odbcConnection )
              :CatalogObjectDefODBC ( odbcConnection )
{
    // Trace - Writer
    ROUTINE_DBG_MEO00 ("TableDefODBC::TableDefODBC()");

    // Log   - Writer
    RepServMessages *pMsgInstance = RepServMessages::Instance();

    setCaseSensitive (SAPDB_TRUE);

#ifdef  UNICODE
    m_pColumnDefResult     = new ODBC_WResult ( odbcConnection );
    m_pConstraintDefResult = new ODBC_WResult ( odbcConnection );
    m_pForeignKeyDefResult = new ODBC_WResult ( odbcConnection );
    m_pTableAttrDefResult  = new ODBC_WResult ( odbcConnection );
#endif
#ifndef UNICODE
    m_pColumnDefResult     = new ODBC_AResult ( odbcConnection );
    m_pConstraintDefResult = new ODBC_AResult ( odbcConnection );
    m_pForeignKeyDefResult = new ODBC_AResult ( odbcConnection );
    m_pTableAttrDefResult  = new ODBC_AResult ( odbcConnection );                
#endif

    setType (ClassTableDefODBC);
}
/*===========================================================================*
 *     TableDefODBC::~TableDefODBC
 *===========================================================================*/
TableDefODBC::~TableDefODBC()
{
    ROUTINE_DBG_MEO00 ("TableDefODBC::~TableDefODBC()");

    if (NULL != m_pColumnDefResult)
    {
        delete m_pColumnDefResult;
        m_pColumnDefResult = NULL;
    }
    if (NULL != m_pConstraintDefResult)
    {
        delete m_pConstraintDefResult;
        m_pConstraintDefResult = NULL;
    }
    if (NULL != m_pForeignKeyDefResult)
    {
        delete m_pForeignKeyDefResult;
        m_pForeignKeyDefResult = NULL;
    }
    if (NULL != m_pTableAttrDefResult)
    {
        delete m_pTableAttrDefResult;
        m_pTableAttrDefResult = NULL;
    }
}
/*===========================================================================*
 *     TableDefODBC::executeSQL
 *===========================================================================*/
SAPDB_Bool TableDefODBC::executeSQL (StudioOAL_String statement)
{
    return CatalogObjectDefODBC::executeSQL( statement);
}
/*===========================================================================*
 *     TableDefODBC::readResult
 *===========================================================================*/
SAPDB_Bool TableDefODBC::readResult ( StudioOAL_String *record , SAPDB_Bool* bIsNull )
{
    return CatalogObjectDefODBC::readResult( record );
}
/*===========================================================================*
 *     TableDefODBC::readTableDefinitionDirect
 *===========================================================================*/
SAPDB_Int4
TableDefODBC::readTableDefinitionDirect(StudioOAL_String owner,
                                        StudioOAL_String tablename )
{
    ROUTINE_DBG_MEO00 ("TableDefODBC::readTableDefinitionDirect()");

    SAPDB_Int4 rc = errOK_els00;

    m_tableName = tablename;
    m_owner     = owner;

    StudioOAL_String        columnDef;
    StudioOAL_String        constraintDef;
    StudioOAL_String        forkeyDef;
    StudioOAL_String        attributeDef;
    StudioOAL_String        name;
    StudioOAL_String        record[8];
    StudioOAL_String        s;
    StudioOAL_String        whereClause;
    SAPDB_Bool              bRead = SAPDB_FALSE;
    SAPDB_Bool             *pIsNull = NULL;
    StudioTD_ColumnDef     *col;
    StudioTD_ConstraintDef *constr;
    StudioTD_ForeignKeyDef *forkey;

   /*-------------------------------------------------------------------------*
    * Reset the table definition to its initial values.                       *
    *-------------------------------------------------------------------------*/
    StudioTD_TableDef::reset();

    setCaseSensitive (SAPDB_TRUE);

    mOwner= owner;          // mOwner and mName are members of StudioTD_TableDef
    mName = tablename;

    /*-------------------------------------------------------------------------*
     * Before doing anything else check if more than one table with the same   *
     * name and different owners is found in the DB in case the owner name     *
     * is not specified.                                                       *
     *-------------------------------------------------------------------------*/
    if (SAPDB_TRUE == owner.Empty())
    {
        StudioOAL_String  SelectOwnerCountCmd("SELECT COUNT(DISTINCT OWNER) FROM "
                                              "DOMAIN.COLUMNS WHERE TABLENAME = '");
        SelectOwnerCountCmd += tablename;
        SelectOwnerCountCmd += "'";
        
        if (SAPDB_TRUE == executeSQL( SelectOwnerCountCmd ))
        {
            CatalogObjectDefODBC::readResult(record);
            SAPDB_Int4 lOwnerCount = AtoI( record[0] );
            if (lOwnerCount > 1)
            {
                RepServMessages::Instance()->ls98PrintMessage(layCatalogExtract_els98, errDEAmbiguousTableName_els98);
                rc = errDEAmbiguousTableName_els98;
            }
        }
        else
        {
            rc = errODBCErrors_els98;
        }
    }

        
    /*-------------------------------------------------------------------------*
     * Prepare the WHERE clause for the SQL statements.                        *
     *-------------------------------------------------------------------------*/
    if (errOK_els00 == rc)
    {
        if (SAPDB_FALSE == owner.Empty())
        {
            whereClause += "WHERE A.owner = '";
            whereClause += owner;
            whereClause += "' AND A.tablename = '";
            whereClause += tablename;
            whereClause += "' ";
        }
        else
        {
            whereClause += "WHERE A.tablename = '";
            whereClause += tablename;
            whereClause += "' ";
        }

        columnDef  = " SELECT A.columnname, A.datatype, A.len, A.dec, A.codetype, A.mode, A.default, A.comment "
                     " FROM DOMAIN.columns A ";
        columnDef += whereClause;

        if (!executeSQL( columnDef ))
        {
            rc = errODBCErrors_els98;
        }
        else
        {
            while ( readResult(record, pIsNull) )
            {
                col = addColumn();
                col->setTableOwner(owner);
                col->setTableName (tablename);
                col->setName      (record[0]);
                col->setOldName   (record[0]);
                col->setDataType  (ProperCase (record[1]));

                if ((col->dataType() == "Char") || (col->dataType() == "Varchar") || (col->dataType() == "Float"))
                {
                    col->setDimension (record[2]);
                }

                if (col->dataType() == "Fixed")
                {
                    s  = record[2];
                    s += ",";
                    s += record[3];
                    col->setDimension (s);
                }
                col->setCodeType (record[4]);

                if (record[5] == "KEY")
                {
                    col->setKey(SAPDB_TRUE);
                }

                if (record[5] == "MAN")
                {
                    col->setNotNull(SAPDB_TRUE);
                }

                if ( !getColumn(7)->getColIsNull() )  // record[6] != NULL_VALUE
                {
                    if (record[6].SubStr(0, 7) == "DEFAULT")
                    {
                        col->setDefaultValue (record[6].SubStr(8));
                    }
                    else
                    {
                        if ( record[6].Empty())
                        {
                            record[6] = " ";
                        }
                        col->setDefaultValue (record[6]);
                    }
                }

                if ( !getColumn(8)->getColIsNull() )  // record[7] != NULL_VALUE
                {
                    col->setComment (record[7]);
                }

                //if ( !getColumn(9)->getColIsNull() )  // record[8] != NULL_VALUE
                col->setUnique(SAPDB_FALSE);

                bRead = SAPDB_TRUE;
            }   // end while (readResult(record, pIsNull ) )

            if (!bRead)
            {
                rc = errODBCErrors_els98;
            }
        }
    }   // end if (errOK_els00 == rc)







    /*-------------------------------------------------------------------------*
     * Load the constraint definitions from the database.                      *
     *-------------------------------------------------------------------------*/
    if (errOK_els00 == rc)
    {
        constraintDef  = "SELECT A.constraintname, A.definition ";
        constraintDef += "FROM   domain.constraints A ";
        constraintDef += whereClause;

        if (!executeSQL (constraintDef))
        {
            rc = errODBCErrors_els98;
        }
        else
        {
            while (readResult(record, pIsNull ))
            {
                constr = addConstraint();
                constr->setName         (record[0]);
                constr->setDefinition   (record[1]);
                //constr->parseConditions (record[1]);

                setConstraintName       (record[0]);
                setConstraintExpression (record[1]);
            }
        }
    }

   /*-------------------------------------------------------------------------*
    * Load the foreign key definitions from the database.                     *
    *-------------------------------------------------------------------------*/
    if (errOK_els00 == rc)
    {
        forkeyDef  = "SELECT A.fkeyname, A.columnname, A.reftablename, A.refowner, A.refcolumnname, A.rule, A.comment ";
        forkeyDef += "FROM   foreignkeycolumns A ";
        forkeyDef += whereClause;

        if (!executeSQL (forkeyDef))
        {
            rc = errODBCErrors_els98;
        }
        else
        {
            while (readResult(record, pIsNull ))
            {
                if (name != record[0])
                {
                    forkey = addForeignKey();
                    forkey->setName       (record[0]);
                    s  = record[3];
                    s += ".";
                    s += record[2];
                    s += " (";
                    forkey->setReference (s);

                    if (record[5].SubStr (0, 6) == "DELETE")
                        forkey->setOnDelete (record[5].SubStr (6));
                    else
                        forkey->setOnDelete (record[5]);

                    if ( !getColumn(7)->getColIsNull() )  // record[6] != NULL_VALUE
                        forkey->setComment (record[6]);

                    name = record[0];
                }

                if (!forkey->referencingColumns().Empty())
                {
                    s  = forkey->referencingColumns();
                    s += ",";
                    forkey->setReferencingColumns (s);
                    s  = forkey->reference().SubStr (0, forkey->reference().Length() - 1);
                    s += ",";
                    forkey->setReference (s);
                }

                s  = forkey->referencingColumns();
                s += record[1];
                forkey->setReferencingColumns (s);
                s  = forkey->reference();
                s += record[4];
                s += ")";
                forkey->setReference (s);
            }
        }
    }   // end if (errOK_els00 == rc) before foreign key handling

    /*-------------------------------------------------------------------------*
     * Load the misc attributes of the table definition.                       *
     *-------------------------------------------------------------------------*/
    if (errOK_els00 == rc)
    {
        attributeDef  = "SELECT A.comment, A.sample_percent, A.sample_rows ";
        attributeDef += "FROM   domain.tables A ";
        attributeDef += whereClause;

        if (!executeSQL (attributeDef))
        {
            rc = errODBCErrors_els98;
        }
        else
        {
            if (readResult(record, pIsNull ))
            {
                if ( !getColumn(1)->getColIsNull() ) // record[0] != NULL_VALUE
                    mComment = record[0];

                if ( !getColumn(2)->getColIsNull() )  // record[1] != NULL_VALUE
                {
                    mSampleValue = AtoI (record[1]);
                    mSampleUnit  = "Percent";
                }
                else
                {
                    mSampleValue = AtoI (record[2]);
                    mSampleUnit  = "Rows";
                }

                if (owner.ToUpper() == "TEMP")
                    mTemporary = SAPDB_TRUE;
            }

            setAlterType (AlterNoChange);
        }

    }

    return rc;
}
/*===========================================================================*
 *     TableDefODBC::genCreateStatement
 *===========================================================================*/
StudioOAL_String TableDefODBC::genCreateStatement()
{
  StudioOAL_String          sql;
  SAPDB_Bool               casesensitive = mCaseSensitive;
  SAPDB_Bool               bKey = false;
  StudioOAL_String          s;
  StudioOAL_String          k;
  char                     buf[256];

  StudioTD_ColumnDef      *col    = NULL;
  StudioTD_ConstraintDef  *constr = NULL;
  StudioTD_ForeignKeyDef  *forkey = NULL;

  StudioOAL_String::ConstIterator  iter;

  /*-------------------------------------------------------------------------*
   * Check if a table name was given. If not --> return an empty string.     *
   *-------------------------------------------------------------------------*/
  if (mName.Empty())
    return "";

  /*-------------------------------------------------------------------------*
   * Generate the intro of the CREATE TABLE statement.                       *
   *-------------------------------------------------------------------------*/
  sql += "CREATE TABLE \"";

  if (mTemporary)
    sql += "TEMP\".\"";

  if(mOwner != ""){
      sql += CASE(mOwner);
      sql += "\".\"";
  }
  sql += CASE(mName);
  sql += "\"\n(\n";

  /*-------------------------------------------------------------------------*
   * For each item in the columns list generate its part of the statement.   *
   * In a second run the PRIMARY KEY clause is generated.                    *
   *-------------------------------------------------------------------------*/
  for (col = (StudioTD_ColumnDef *)mColumns.first(); col; col = (StudioTD_ColumnDef *)mColumns.next())
       sql += col->genCreateStatement (mMode, mCaseSensitive);

  /*-------------------------------------------------------------------------*
   * A primary key, constraints and foreign keys are only inserted if the    *
   * mode is not CreateFromQuery (query as a template).                      *
   *-------------------------------------------------------------------------*/
  if (mMode != CreateFromQuery)
  {
    for (col = (StudioTD_ColumnDef *)mColumns.first(); col; col = (StudioTD_ColumnDef *)mColumns.next())
    {
      if (col->key())
      {
        sql += (!bKey ? "\tPRIMARY KEY (" : "");
        sql += "\"";
        sql += CASE(col->name());
        sql += "\", ";
        bKey = true;
      }
    }

    if (bKey)
    {
      iter = sql.End();
      iter-=2;
      sql.Erase (iter, sql.End());
      sql += "),\n";
    }

    /*-----------------------------------------------------------------------*
     * For each item in the constraint list generate its part of the         *
     * statement.                                                            *
     *-----------------------------------------------------------------------*/

    // Column Constraint
    for (constr = (StudioTD_ConstraintDef *)mConstraints.first(); constr; constr = (StudioTD_ConstraintDef *)mConstraints.next())
    {
         // CONSTRAINT <constraint_name> CHECK <search_condition>
        if ( !constr->name().Empty() )
        {
             sql += " CONSTRAINT ";
             sql += "\"";
             sql += constr->name();
             sql += "\"";
             sql += " CHECK ";
             sql += constr->definition();
             sql += ",\n";

        }
        // CONSTRAINT <search_condition>
        else
        {
             sql += " CONSTRAINT ";
             sql += constr->definition();
             sql += ",\n";
        }
    }

    /*
    // Table Constraint
    if ( !constraintexpression().Empty() )
    {
         // CONSTRAINT <constraint_name> CHECK <search_condition>
        if ( !constraintname().Empty() )
        {
             sql += " CONSTRAINT ";
             sql += "\"";
             sql += constraintname();
             sql += "\"";
             sql += " CHECK ";
             sql += constraintexpression();
        }
        // CONSTRAINT <search_condition>
        else
        {
             sql += " CONSTRAINT ";
             sql += constraintexpression();
        }

    }*/

    /*-----------------------------------------------------------------------*
     * For each item in the foreignkey list generate its part of the         *
     * statement.                                                            *
     *-----------------------------------------------------------------------*/
     for (forkey = (StudioTD_ForeignKeyDef *)mForeignKeys.first();forkey; forkey = (StudioTD_ForeignKeyDef *)mForeignKeys.next())
          sql += forkey->genCreateStatement (mMode, mCaseSensitive);

  }// if (mMode != CreateFromQuery)

  iter = sql.End();
  iter-=2;
  sql.Erase (iter, sql.End());
  sql += "\n)";

  /*-------------------------------------------------------------------------*
   * Insert misc table attributes.                                           *
   *-------------------------------------------------------------------------*/
  if (mTemporary && mIgnoreRollback)
    sql += "\nIGNORE ROLLBACK";

  if ((mSampleValue != 20000) || (mSampleUnit != "Rows"))
  {
    sprintf (buf, "\nSAMPLE %d %s", mSampleValue, UpperCase(mSampleUnit).StrPtr());
    sql += buf;
  }

  /*-------------------------------------------------------------------------*
   * If the table is based on a query template the SELECT statement is       *
   * appended in the AS clause.                                              *
   *-------------------------------------------------------------------------*/
  if (mMode == CreateFromQuery)
  {
    sql += "\nAS ";
    sql += mCreateAs;

    if (!mDuplicates.Empty())
    {
      sql += " ";
      sql += UpperCase(mDuplicates);
      sql += " DUPLICATES";
    }
  }

  /*-------------------------------------------------------------------------*
   * Finally, append the comments.                                           *
   *-------------------------------------------------------------------------*/
  if (!mComment.Empty())
  {
    sql += NEW_LINE;
    sql += SEPARATOR;
    sql += "COMMENT ON TABLE \"";
    sql += CASE(mName);
    sql += "\" IS '";
    sql += EscapeSingleQuotes(mComment);
    sql += "'";
  }

  StudioOAL_String TmpCommentString;

  for (col = (StudioTD_ColumnDef *)mColumns.first();
       col;
       col = (StudioTD_ColumnDef *)mColumns.next())
  {
    if (!col->comment().Empty())
    {
      TmpCommentString = col->comment();
      sql += NEW_LINE;
      sql += SEPARATOR;
      sql += "COMMENT ON COLUMN \"";
      sql += CASE(mName);
      sql += "\".\"";
      sql += CASE(col->name());
      sql += "\" IS '";
      sql += EscapeSingleQuotes(TmpCommentString);
      sql += "'";
    }
  }

  if (mMode != CreateFromQuery)
  {
    for (forkey = (StudioTD_ForeignKeyDef *)mForeignKeys.first();
         forkey;
         forkey = (StudioTD_ForeignKeyDef *)mForeignKeys.next())
    {
      if (!forkey->comment().Empty() && (forkey->name() != "<no name>"))
      {
        TmpCommentString = forkey->comment();
        sql += NEW_LINE;
        sql += SEPARATOR;
        sql += "COMMENT ON FOREIGN KEY \"";
        sql += CASE(mName);
        sql += "\".\"";
        sql += CASE(forkey->name());
        sql += "\" IS '";
        sql += EscapeSingleQuotes(TmpCommentString);
        sql += "'";
      }
    }
  }

  return sql;

}
/*===========================================================================*
 *     TableDefODBC::writeTableDefinition
 *===========================================================================*/
SAPDB_Bool TableDefODBC::writeTableDefinition ( )
{
   ROUTINE_DBG_MEO00 ("TableDefODBC::writeTableDefinition ()");

   StudioOAL_String table;

   if ( getCatalog()->getExtrMode() == emExtractAsDDL_els31 )
   {
        table  = SEPARATOR;

        if ( getCatalog()->getSqlMode() != cpr_kind_internal )
        {
             table += pszSQLModeString_ls31[cpr_kind_internal];
             table += SEPARATOR;
        }

        // does not support constraints PTS 1110139
        // table += StudioTD_TableDef::genCreateStatement();
        table += genCreateStatement();
   }
   if ( getCatalog()->getExtrMode() == emExtractAsXML_els31 )
   {
        table += StudioTD_TableDef::genCreateXML();
   }

   return writeLine(table);
}
/*===========================================================================*
 *     TableDefODBC::initResultSet
 *===========================================================================*/
SAPDB_Bool TableDefODBC::initResultSet( sapdbwa_DBC & odbcConnection )
{
    SAPDB_Bool rc = SAPDB_FALSE;

    if ( odbcConnection.HDBC() && odbcConnection.HENV() )
    {
        StudioOAL_String columnDef,constraintDef,foreignKeyDef,tableAttrDef,whereClause;

        /*-------------------------------------------------------------------------*
         * Prepare WHERE clause for SQL statements.                                *
         *-------------------------------------------------------------------------*/
        whereClause    = "WHERE  owner     = ? "
                         " AND   tablename = ? ";

        columnDef      = "SELECT A.columnname, A.datatype, A.len, A.dec, A.codetype, A.mode, A.default, A.comment "
                         "FROM DOMAIN.columns A ";
        columnDef     += whereClause;

        constraintDef = "SELECT constraintname, definition "
                        "FROM  DOMAIN.constraints ";
        constraintDef += whereClause;

        foreignKeyDef  = "SELECT fkeyname, columnname, reftablename, refowner, refcolumnname, rule, comment "
                         "FROM   DOMAIN.foreignkeycolumns ";
        foreignKeyDef += whereClause;

        tableAttrDef   = "SELECT comment, sample_percent, sample_rows "
                         "FROM   DOMAIN.tables ";
        tableAttrDef  += whereClause;

        // SQLAllocStmt     SQLPrepare
        // SQLNumParams     SQLDescribeParam SQLBindParameter
        // SQLNumResultCols SQLDescribeCol   SQLBindCol
        rc = m_pColumnDefResult->init(columnDef);
        if (SAPDB_TRUE == rc)
        {
            rc = m_pConstraintDefResult->init(constraintDef);
        }
        if (SAPDB_TRUE == rc)
        {
            rc = m_pForeignKeyDefResult->init(foreignKeyDef);
        }
        if (SAPDB_TRUE == rc)
        {
            rc = m_pTableAttrDefResult->init(tableAttrDef);
        }
        if (SAPDB_FALSE == rc)
        {
            getCatalog()->getError()->setErrorText(getErrorText() );
            getCatalog()->getError()->isError(isError() );
        }
    }// if ( odbcConnection.HDBC() && odbcConnection.HENV() )

    return rc;
}
/*===========================================================================*
 *     TableDefODBC::readTableDefinition                                     *
 *===========================================================================*/
SAPDB_Bool
TableDefODBC::readTableDefinition(StudioOAL_String owner, StudioOAL_String tablename)
{
    SAPDB_Bool       isOk         = SAPDB_FALSE;

    StudioOAL_String        sql;
    StudioOAL_String        name;
    StudioOAL_String        record[8];
    StudioOAL_String        s;
    StudioOAL_String        whereClause;
    SAPDB_Bool              bRead = false;

    StudioTD_ColumnDef     *col;
    StudioTD_ConstraintDef *constr;
    StudioTD_ForeignKeyDef *forkey;



    /*-------------------------------------------------------------------------*
     * Reset the table attributes to their initial values.                     *
     *-------------------------------------------------------------------------*/
    StudioTD_TableDef::reset();

    setOwner(owner);
    setName(tablename);
    setCaseSensitive(SAPDB_TRUE);

    m_tableName = tablename ;
    m_owner     = owner     ;

    StudioOAL_String parameters[2];

    parameters[0]   = owner     ;
    parameters[1]   = tablename ;

    /*-------------------------------------------------------------------------*
     * Load the column definitions from the database.                          *
     *-------------------------------------------------------------------------*/
    m_pColumnDefResult->reset();

    if ( !m_pColumnDefResult->setParamValues( parameters ) )
    {
        getCatalog()->getError()->setErrorText(getErrorText() );
        getCatalog()->getError()->isError     (isError     () );
        return SAPDB_FALSE;
    }

    /*-------------------------------------------------------------------------*
     * Execute the prepared statement                                          *
     *-------------------------------------------------------------------------*/
    // SQLExecute
    if ( !m_pColumnDefResult->execute() )
    {
        CatalogDefODBC *pCatalog  = getCatalog();
        sapdb_Error    *pError    = pCatalog->getError();

        StudioOAL_String errorText = m_pColumnDefResult->getErrorText();
        SAPDB_Bool      isError   = m_pColumnDefResult->isError     ();

        pError->setErrorText( errorText );
        pError->isError     ( isError   );

        return SAPDB_FALSE;
    }
    // SQlFetch SQLFreeStatement
    while (readColumnDefResult(record))
    {
        col = addColumn();
        col->setTableOwner(owner);
        col->setTableName(tablename);
        col->setName(record[0]);
        col->setOldName(record[0]);
        col->setDataType(ProperCase(record[1]));

        if ((col->dataType() == "Char") || (col->dataType() == "Varchar") || (col->dataType() == "Float"))
          col->setDimension (record[2]);

        if (col->dataType() == "Fixed")
        {
            s  = record[2];
            s += ",";
            s += record[3];
            col->setDimension (s);
        }

        col->setCodeType (record[4]);

        if (record[5] == "KEY")
            col->setKey (SAPDB_TRUE);

        if (record[5] == "MAN")
            col->setNotNull (SAPDB_TRUE);

        if ( !m_pColumnDefResult->getColumn(7)->getColIsNull() ) // record[6] != NULL_VALUE
        {
            if(record[6].SubStr(0, 7) == "DEFAULT")
            {
                col->setDefaultValue (record[6].SubStr(8));
            }
            else
            {
                col->setDefaultValue (record[6]);
            }
        }

        if ( !m_pColumnDefResult->getColumn(8)->getColIsNull() ) // record[7] != NULL_VALUE
            col->setComment (record[7]);

        //if ( !m_pColumnDefResult->getColumn(9)->getColIsNull() ) // record[8] != NULL_VALUE)
        col->setUnique (SAPDB_FALSE);

        bRead = SAPDB_TRUE;
    }

    if (!bRead)
        return isOk;

    /*-------------------------------------------------------------------------*
     * Load the constraint definitions from the database.                      *
     *-------------------------------------------------------------------------*/
    m_pConstraintDefResult->reset();

    if ( !m_pConstraintDefResult->setParamValues( parameters ) )
    {
        getCatalog()->getError()->setErrorText(getErrorText() );
        getCatalog()->getError()->isError(isError() );
        return SAPDB_FALSE;
    }

    /*-------------------------------------------------------------------------*
     * Execute the prepared statement                                          *
     *-------------------------------------------------------------------------*/
    // SQLExecute
    if (!m_pConstraintDefResult->execute() )
    {
        CatalogDefODBC *pCatalog = getCatalog();
        sapdb_Error    *pError   = pCatalog->getError();

        StudioOAL_String errorText = m_pConstraintDefResult->getErrorText();
        SAPDB_Bool      isError   = m_pConstraintDefResult->isError();

        pError->setErrorText( errorText );
        pError->isError( isError );

        return SAPDB_FALSE;
    }
    // SQlFetch SQLFreeStatement
    while (readConstraintDefResult(record))
    {
        constr = addConstraint( );
        constr->setName(record[0]);
        constr->setDefinition(record[1]);
        //constr->parseConditions (record[1]);

        setConstraintName       (record[0]);
        setConstraintExpression (record[1]);

        bRead = true;
    }

    /*-------------------------------------------------------------------------*
     * Load the foreign key definitions from the database.                     *
     *-------------------------------------------------------------------------*/
    m_pForeignKeyDefResult->reset();
    if ( !m_pForeignKeyDefResult->setParamValues( parameters ) )
    {
        getCatalog()->getError()->setErrorText(getErrorText() );
        getCatalog()->getError()->isError     (isError     () );
        return SAPDB_FALSE;
    }

    /*-------------------------------------------------------------------------*
     * Execute the prepared statement                                          *
     *-------------------------------------------------------------------------*/
    // SQLExecute
    if (!m_pForeignKeyDefResult->execute() )
    {
        CatalogDefODBC *pCatalog  = getCatalog();
        sapdb_Error    *pError    = pCatalog->getError();

        StudioOAL_String errorText = m_pForeignKeyDefResult->getErrorText();
        SAPDB_Bool       isError   = m_pForeignKeyDefResult->isError     ();

        pError->setErrorText( errorText );
        pError->isError( isError );

        return SAPDB_FALSE;
    }
    // SQlFetch SQLFreeStatement
    while (readForeignKeyDefResult(record))
    {
        bRead = true;

        if (name != record[0])
        {
            forkey = addForeignKey( );
            forkey->setName(record[0]);
            s  = record[3];
            s += ".";
            s += record[2];
            s += " (";
            forkey->setReference (s);

            if (record[5].SubStr (0, 6) == "DELETE")
                forkey->setOnDelete (record[5].SubStr (6));
            else
                forkey->setOnDelete (record[5]);

            if ( !m_pForeignKeyDefResult->getColumn(7)->getColIsNull() ) // record[6] != NULL_VALUE
                forkey->setComment (record[6]);

            name = record[0];
        }

        if (!forkey->referencingColumns().Empty())
        {
            s  = forkey->referencingColumns();
            s += ",";
            forkey->setReferencingColumns (s);
            s  = forkey->reference().SubStr (0, forkey->reference().Length() - 1);
            s += ",";
            forkey->setReference (s);
        }

        s  = forkey->referencingColumns();
        s += record[1];
        forkey->setReferencingColumns (s);
        s  = forkey->reference();
        s += record[4];
        s += ")";
        forkey->setReference (s);
    }

    /*-------------------------------------------------------------------------*
     * Load the misc attributes of the table definition.                       *
     *-------------------------------------------------------------------------*/
    m_pTableAttrDefResult->reset();

    if ( !m_pTableAttrDefResult->setParamValues( parameters ) )
    {
        getCatalog()->getError()->setErrorText(getErrorText() );
        getCatalog()->getError()->isError     (isError     () );
        return SAPDB_FALSE;
    }

    /*-------------------------------------------------------------------------*
     * Execute the prepared statement                                          *
     *-------------------------------------------------------------------------*/
    // SQLExecute
    if (!m_pTableAttrDefResult->execute() )
    {
        CatalogDefODBC *pCatalog  = getCatalog();
        sapdb_Error    *pError    = pCatalog->getError();

        StudioOAL_String errorText = m_pTableAttrDefResult->getErrorText();
        SAPDB_Bool      isError   = m_pTableAttrDefResult->isError     ();

        pError->setErrorText( errorText );
        pError->isError     ( isError   );

        return SAPDB_FALSE;
    }
    // SQlFetch SQLFreeStatement
    while (readTableAttrDefResult(record))
    {
        bRead = true;

        if ( !m_pTableAttrDefResult->getColumn(1)->getColIsNull() ) // record[0] != NULL_VALUE
            mComment = record[0];

        if ( !m_pTableAttrDefResult->getColumn(2)->getColIsNull() ) // record[1] != NULL_VALUE
        {
            mSampleValue = AtoI (record[1]);
            mSampleUnit  = "Percent";
        }
        else
        {
            mSampleValue = AtoI (record[2]);
            mSampleUnit  = "Rows";
        }

        if (owner.ToUpper() == "TEMP")
              mTemporary = true;
    }

    setAlterType (AlterNoChange);

    isOk = SAPDB_TRUE;

    return isOk;
}
/*===========================================================================*
 *     TableDefODBC::readColumnDefResult
 *===========================================================================*/
SAPDB_Bool TableDefODBC::readColumnDefResult(StudioOAL_String *record)
{

   SAPDB_Int2       i;
   StudioOAL_String  s;

  /*-------------------------------------------------------------------------*
   * Fetch the next record.                                                  *
   *-------------------------------------------------------------------------*/
  // SQLFetch
  if (!m_pColumnDefResult->fetch())
  {
      m_pColumnDefResult->closeCursor();
      return SAPDB_FALSE;
  }

  /*-------------------------------------------------------------------------*
   * If there are no more records, reading the result is finished.           *
   *-------------------------------------------------------------------------*/
  if (m_pColumnDefResult->noRows())
  {
     m_pColumnDefResult->closeCursor();
     return SAPDB_FALSE;
  }

  /*-------------------------------------------------------------------------*
   * Read all columns of the fetched record.                                 *
   *-------------------------------------------------------------------------*/
  for (i = 0; i < m_pColumnDefResult->getColCount(); i++)
  {
    m_pColumnDefResult->bindColumnDataAsString (i + 1, s);
    record[i] = s;
  }

  return SAPDB_TRUE;
}
/*===========================================================================*
 *     TableDefODBC::readConstraintDefResult
 *===========================================================================*/
SAPDB_Bool TableDefODBC::readConstraintDefResult(StudioOAL_String *record)
{
   SAPDB_Int2       i;
   StudioOAL_String  s;

  /*-------------------------------------------------------------------------*
   * Fetch the next record.                                                  *
   *-------------------------------------------------------------------------*/
  // SQLFetch
  if (!m_pConstraintDefResult->fetch())
  {
       m_pConstraintDefResult->closeCursor();
       return SAPDB_FALSE;
  }

  /*-------------------------------------------------------------------------*
   * If there are no more records, reading the result is finished.           *
   *-------------------------------------------------------------------------*/
  if (m_pConstraintDefResult->noRows())
  {
      m_pConstraintDefResult->closeCursor();
      return SAPDB_FALSE;
  }

  /*-------------------------------------------------------------------------*
   * Read all columns of the fetched record.                                 *
   *-------------------------------------------------------------------------*/
  for (i = 0; i < m_pConstraintDefResult->getColCount(); i++)
  {
    m_pConstraintDefResult->bindColumnDataAsString (i + 1, s);
    record[i] = s;
  }

  return SAPDB_TRUE;
}
/*===========================================================================*
 *     TableDefODBC::readForeignKeyDefResult
 *===========================================================================*/
SAPDB_Bool TableDefODBC::readForeignKeyDefResult(StudioOAL_String *record)
{
    SAPDB_Int2       i;
    StudioOAL_String  s;

    /*-------------------------------------------------------------------------*
     * Fetch the next record.                                                  *
     *-------------------------------------------------------------------------*/
    // SQLFetch
    if (!m_pForeignKeyDefResult->fetch())
    {
        m_pForeignKeyDefResult->closeCursor();
        return SAPDB_FALSE;
    }

    /*-------------------------------------------------------------------------*
     * If there are no more records, reading the result is finished.           *
     *-------------------------------------------------------------------------*/
    if (m_pForeignKeyDefResult->noRows())
    {
        m_pForeignKeyDefResult->closeCursor();
        return SAPDB_FALSE;
    }

    /*-------------------------------------------------------------------------*
     * Read all columns of the fetched record.                                 *
     *-------------------------------------------------------------------------*/
    for (i = 0; i < m_pForeignKeyDefResult->getColCount(); ++i)
    {
        m_pForeignKeyDefResult->bindColumnDataAsString (i + 1, s);
        record[i] = s;
    }

    return SAPDB_TRUE;
}
/*===========================================================================*
 *     TableDefODBC::readTableAttrDefResult
 *===========================================================================*/
SAPDB_Bool TableDefODBC::readTableAttrDefResult(StudioOAL_String *record)
{
   SAPDB_Int2       i;
   StudioOAL_String  s;

  /*-------------------------------------------------------------------------*
   * Fetch the next record.                                                  *
   *-------------------------------------------------------------------------*/
  // SQLFetch
  if (!m_pTableAttrDefResult->fetch())
  {
       m_pTableAttrDefResult->closeCursor();
       return SAPDB_FALSE;
  }

  /*-------------------------------------------------------------------------*
   * If there are no more records, reading the result is finished.           *
   *-------------------------------------------------------------------------*/
  if (m_pTableAttrDefResult->noRows())
  {
      m_pTableAttrDefResult->closeCursor();
      return SAPDB_FALSE;
  }

  /*-------------------------------------------------------------------------*
   * Read all columns of the fetched record.                                 *
   *-------------------------------------------------------------------------*/
  for (i = 0; i < m_pTableAttrDefResult->getColCount(); i++)
  {
    m_pTableAttrDefResult->bindColumnDataAsString (i + 1, s);
    record[i] = s;
  }

  return SAPDB_TRUE;
}
/*===========================================================================*
 *    ViewDefODBC::ViewDefODBC
 *===========================================================================*/
ViewDefODBC::ViewDefODBC( sapdbwa_DBC & odbcConnection , CatalogDefODBC * pCatalog  )
            :CatalogObjectDefODBC ( odbcConnection , pCatalog )
{
#ifdef UNICODE
    m_pViewDefResult = new ODBC_WResult ( odbcConnection );
#endif
#ifndef UNICODE
    m_pViewDefResult = new ODBC_AResult ( odbcConnection );
#endif

    initResultSet( odbcConnection );
}
/*===========================================================================*
 *     ViewDefODBC::ViewDefODBC()
 *===========================================================================*/
ViewDefODBC::ViewDefODBC( sapdbwa_DBC & odbcConnection )
            :CatalogObjectDefODBC ( odbcConnection )
{
#ifdef UNICODE
    m_pViewDefResult = new ODBC_WResult ( odbcConnection );
#endif
#ifndef UNICODE
    m_pViewDefResult = new ODBC_AResult ( odbcConnection );
#endif
}
/*===========================================================================*
 *     ViewDefODBC::~ViewDefODBC()
 *===========================================================================*/
ViewDefODBC::~ViewDefODBC()
{
    if ( m_pViewDefResult != NULL )
    {
         delete m_pViewDefResult;
         m_pViewDefResult = NULL;
    }
}
/*===========================================================================*
 *     ViewDefODBC::initResultSet
 *===========================================================================*/
SAPDB_Bool ViewDefODBC::initResultSet ( sapdbwa_DBC & odbcConnection )
{
    SAPDB_Bool isOk = SAPDB_FALSE;

    if (  odbcConnection.HDBC() && odbcConnection.HENV() )
    {
        StudioOAL_String  viewDef;
        StudioOAL_String  whereClause;

       /*-------------------------------------------------------------------------*
        * Prepare the WHERE clause for the SQL statements.                        *
        *-------------------------------------------------------------------------*/
        whereClause += " WHERE owner    = ? "
                       " AND   viewname = ? ";

        viewDef      = "SELECT viewname, owner, definition "
                       "FROM DOMAIN.viewdefs ";
        viewDef     += whereClause;

        // SQLAllocStmt     SQLPrepare
        // SQLNumParams     SQLDescribeParam SQLBindParameter
        // SQLNumResultCols SQLDescribeCol   SQLBindCol
        if ( !m_pViewDefResult->init( viewDef ) )
        {
            getCatalog()->getError()->setErrorText(getErrorText() );
            getCatalog()->getError()->isError     (isError     () );

            return SAPDB_FALSE;
        }

        isOk = SAPDB_TRUE;
    }

    return isOk;
}
/*===========================================================================*
 *     ViewDefODBC::readViewDefinition
 *===========================================================================*/
SAPDB_Bool ViewDefODBC::readViewDefinition( StudioOAL_String owner, StudioOAL_String viewname )
{
   ROUTINE_DBG_MEO00 ("ViewDefODBC::readViewDefinition()");

   RepServMessages *pMsgInstance = RepServMessages::Instance();

   // VIEWS    : OWNER,VIEWNAME,PRIVILEGES,TYPE,CREATEDATE,CREATETIME,UPDSTATDATE,UPDSTATTIME,ALTERDATE,ALTERTIME,UNLOADED,COMMENT
   // VIEWDEFS : OWNER,VIEWNAME,LEN,DEFINITION

   SAPDB_Bool isOk = SAPDB_FALSE;

   StudioOAL_String  viewDef[2];
   StudioOAL_String  record[3];

   viewDef[0] = owner;
   viewDef[1] = viewname;

  /*-------------------------------------------------------------------------*
   * Load the index definitions from the database.                           *
   *-------------------------------------------------------------------------*/
   m_pViewDefResult->reset();

   if ( !m_pViewDefResult->setParamValues( viewDef ) )
   {
        getCatalog()->getError()->setErrorText(getErrorText() );
        getCatalog()->getError()->isError     (isError     () );
        return SAPDB_FALSE;
   }
   /*-------------------------------------------------------------------------*
    * Execute the prepared statement                                          *
    *-------------------------------------------------------------------------*/
   // SQLExecute
   if ( !m_pViewDefResult->execute() )
   {
        StudioOAL_String errorText = m_pViewDefResult->getErrorText();
        SAPDB_Bool       isError   = m_pViewDefResult->isError     ();

        getCatalog()->getError()->setErrorText( errorText );
        getCatalog()->getError()->isError     ( isError   );

        return SAPDB_FALSE;
   }
   // SQlFetch SQLFreeStatement
   while ( readViewDefResult (record) )
   {
         m_viewName       = record[0];
         m_owner          = record[1];
         m_viewDefinition = record[2];

         isOk = SAPDB_TRUE;
   }

   return isOk;

}
/*===========================================================================*
 *     ViewDefODBC::writeViewDefinition
 *===========================================================================*/
SAPDB_Bool ViewDefODBC::writeViewDefinition()
{
   ROUTINE_DBG_MEO00 ("ViewDefODBC::writeViewDefinition");

   RepServMessages *pMsgInstance = RepServMessages::Instance();

   StudioOAL_String schemaview;

   if ( getCatalog()->getExtrMode() == emExtractAsDDL_els31 )
   {
        schemaview  = SEPARATOR;

        if ( getCatalog()->getSqlMode() != cpr_kind_internal )
        {
             SAPDB_Int4 iSqlMode = getCatalog()->getSqlMode();
             schemaview += pszSQLModeString_ls31[iSqlMode];
             schemaview += SEPARATOR;
        }

        schemaview += m_viewDefinition;
           
        //!! ToDo COMMENT ON -> PTS 1109219
        //definition  = record[3].SubStr(11,record[3].Length());
        //definition.Trim();
   }

   if ( getCatalog()->getExtrMode() == emExtractAsXML_els31 )
        schemaview += genCreateXML();

   return writeLine( schemaview );

}
/*===========================================================================*
 *     ViewDefODBC::readViewDefResult
 *===========================================================================*/
SAPDB_Bool ViewDefODBC::readViewDefResult( StudioOAL_String *record )
{
   SAPDB_Int2       i;
   StudioOAL_String  s;

  /*-------------------------------------------------------------------------*
   * Fetch the next record.                                                  *
   *-------------------------------------------------------------------------*/
  // SQLFetch
  if (!m_pViewDefResult->fetch())
  {
      m_pViewDefResult->closeCursor();
      return SAPDB_FALSE;
  }

  /*-------------------------------------------------------------------------*
   * If there are no more records, reading the result is finished.           *
   *-------------------------------------------------------------------------*/
  if (m_pViewDefResult->noRows())
  {
     m_pViewDefResult->closeCursor();
     return SAPDB_FALSE;
  }

  /*-------------------------------------------------------------------------*
   * Read all columns of the fetched record.                                 *
   *-------------------------------------------------------------------------*/
  for (i = 0; i < m_pViewDefResult->getColCount(); i++)
  {
    m_pViewDefResult->bindColumnDataAsString (i + 1, s);
    record[i] = s;
  }

  return SAPDB_TRUE;
}
/*===========================================================================*
 *     ViewDefODBC::readResult()
 *===========================================================================*/
SAPDB_Bool ViewDefODBC::readResult( StudioOAL_String *record )
{
    SAPDB_Bool isOk = SAPDB_FALSE;

    isOk = CatalogObjectDefODBC::readResult( record);

    if ( isOk )
    {
         m_viewName       = record[0];
         m_owner          = record[1];
         // <create_view_statement> ::= CREATE [OR REPLACE] VIEW <table_name> [(<alias_name>,...)] AS <query_expression> [WITH CHECK OPTION]
         m_viewDefinition = record[2];
    }

    return isOk;
}
/*===========================================================================*
 *     ViewDefODBC::writeResult()
 *===========================================================================*/
SAPDB_Bool ViewDefODBC::writeResult( StudioOAL_String *record )
{
      StudioOAL_String schemaview;

      if ( getCatalog()->getExtrMode() == emExtractAsDDL_els31 )
      {
           schemaview  = SEPARATOR;

           if ( getCatalog()->getSqlMode() != cpr_kind_internal )
           {
                 SAPDB_Int4 iSqlMode = getCatalog()->getSqlMode();
                 schemaview += pszSQLModeString_ls31[iSqlMode];
                 schemaview += SEPARATOR;
            }

            schemaview += m_viewDefinition;
               
           //!! ToDo COMMENT ON -> PTS 1109219
           //definition  = record[3].SubStr(11,record[3].Length());
           //definition.Trim();
      }

      if ( getCatalog()->getExtrMode() == emExtractAsXML_els31 )
           schemaview += genCreateXML();

      return writeLine( schemaview );
}
/*===========================================================================*
 *     ViewDefODBC::genCreateXML()
 *===========================================================================*/
StudioOAL_String ViewDefODBC::genCreateXML()
{
    // <dbm:SchemaViews>

   StudioOAL_String  as   ((SAPDB_UTF8*)" AS ");
   StudioOAL_String  check((SAPDB_UTF8*)" WITH CHECK OPTION ");
   StudioOAL_String  schemaView;
   StudioOAL_String  viewQuery;
   StudioOAL_String  checkOption = "";

   viewQuery += m_viewDefinition;

   SAPDB_UInt  start = 0, end ;

   if ( (end = viewQuery.Find (as, start) ) != StudioOAL_String::NPos )
   {
       viewQuery = viewQuery.SubStr( end + as.Length(),viewQuery.Length() );

       start = end +1 ;

       if ( (end = viewQuery.Find (check, start) ) != StudioOAL_String::NPos )
       {
             viewQuery   = viewQuery.SubStr( 0, end );
             checkOption = check;
       }

   }

   schemaView  = "<dbm:DeployedView oim:id=\"_";
   schemaView += m_owner;
   schemaView += ".";
   schemaView += m_viewName;
   schemaView += "\" oim:supertype = \"Query\"";
   schemaView += "   oim:seqno     = \"";
   schemaView += "\" dbm:Body = \"";
   schemaView += string2XML ( viewQuery );
   schemaView += "\" dbm:CheckOption = \"";
   schemaView += checkOption;
   schemaView += "\" dbm:description = \"";
   schemaView += string2XML ( m_viewDefinition );
   schemaView += "\" dbm:isSpecification = \"1";
   schemaView += "\" dbm:comments = \"";
   schemaView += "\" dbm:EstimatedRows = \"";
   schemaView += "\" dbm:EstimatedSize = \"";
   schemaView += "\"/>\n";

   return schemaView;
}
/*===========================================================================*
 *    SynonymDefODBC::SynonymDefODBC
 *===========================================================================*/
SynonymDefODBC::SynonymDefODBC( sapdbwa_DBC & odbcConnection , CatalogDefODBC * pCatalog  )
               :CatalogObjectDefODBC ( odbcConnection , pCatalog )
{
    
#ifdef UNICODE
    m_pSynonymDefResult = new ODBC_WResult ( odbcConnection );
#endif
#ifndef UNICODE
    m_pSynonymDefResult = new ODBC_AResult ( odbcConnection );
#endif

    initResultSet( odbcConnection );
}
/*===========================================================================*
 *     SynonymDefODBC::SynonymDefODBC()
 *===========================================================================*/
SynonymDefODBC::SynonymDefODBC( sapdbwa_DBC & odbcConnection )
               :CatalogObjectDefODBC        ( odbcConnection )
{
#ifdef UNICODE
    m_pSynonymDefResult = new ODBC_WResult  ( odbcConnection );
#endif
#ifndef UNICODE
    m_pSynonymDefResult = new ODBC_AResult  ( odbcConnection );
#endif
}
/*===========================================================================*
 *     SynonymDefODBC::~SynonymDefODBC()
 *===========================================================================*/
SynonymDefODBC::~SynonymDefODBC()
{
    if ( m_pSynonymDefResult != NULL )
    {
        delete m_pSynonymDefResult;
        m_pSynonymDefResult = NULL;
    }
}
/*===========================================================================*
 *     SynonymDefODBC::initResultSet
 *===========================================================================*/
SAPDB_Bool SynonymDefODBC::initResultSet ( sapdbwa_DBC & odbcConnection )
{

  SAPDB_Bool isOk = SAPDB_FALSE;

  if (  odbcConnection.HDBC() && odbcConnection.HENV() )
  {
        // CREATE SYNONYM
        // SYNONYMS  : OWNER, SYNONYMNAME, PUBLIC, TABLEOWNER, TABLENAME, COMMENT

        StudioOAL_String  synonymDef;
        StudioOAL_String  whereClause;

        /*-------------------------------------------------------------------------*
         * Prepare the WHERE clause for the SQL statements.                        *
         *-------------------------------------------------------------------------*/
        whereClause += "WHERE owner       = ? ";
        whereClause += " AND  synonymname = ? ";

        synonymDef  = " SELECT OWNER, SYNONYMNAME, PUBLIC, TABLEOWNER, TABLENAME, COMMENT ";
        synonymDef += " FROM DOMAIN.SYNONYMS ";
        synonymDef += whereClause;

        // SQLAllocStmt     SQLPrepare
        // SQLNumParams     SQLDescribeParam SQLBindParameter
        // SQLNumResultCols SQLDescribeCol   SQLBindCol

        if ( !m_pSynonymDefResult->init( synonymDef ) )
        {
             getCatalog()->getError()->setErrorText(getErrorText() );
             getCatalog()->getError()->isError     (isError     () );

             return SAPDB_FALSE;
        }

        isOk = SAPDB_TRUE;
  }

  return isOk;
}
/*===========================================================================*
 *     SynonymDefODBC::readSynonymDefinition
 *===========================================================================*/
SAPDB_Bool SynonymDefODBC::readSynonymDefinition( StudioOAL_String owner, StudioOAL_String synonymname )
{
   ROUTINE_DBG_MEO00 ("SynonymDefODBC::readSynonymDefinition");

   RepServMessages *pMsgInstance = RepServMessages::Instance();

   SAPDB_Bool isOk = SAPDB_FALSE;

   StudioOAL_String  synonymDef[2];
   StudioOAL_String  record[6];

   synonymDef[0] = owner;
   synonymDef[1] = synonymname;

  /*-------------------------------------------------------------------------*
   * Load the synonym definitions from the database.                           *
   *-------------------------------------------------------------------------*/

   if ( !m_pSynonymDefResult->setParamValues( synonymDef ) )
   {
        getCatalog()->getError()->setErrorText(getErrorText() );
        getCatalog()->getError()->isError     (isError     () );
        return SAPDB_FALSE;
   }
   /*-------------------------------------------------------------------------*
    * Execute the prepared statement                                          *
    *-------------------------------------------------------------------------*/
   // SQLExecute
   if ( !m_pSynonymDefResult->execute() )
   {
        StudioOAL_String errorText = m_pSynonymDefResult->getErrorText();
        SAPDB_Bool      isError   = m_pSynonymDefResult->isError     ();

        getCatalog()->getError()->setErrorText( errorText );
        getCatalog()->getError()->isError     ( isError   );

        return SAPDB_FALSE;
   }
   // SQlFetch SQLFreeStatement
   while ( readSynonymDefResult (record) )
   {
           m_owner             = record[0];
           m_synonymName       = record[1];
           m_public            = record[2];
           m_tableowner        = record[3];
           m_tablename         = record[4];
           m_comment           = record[5];
           
           isOk = SAPDB_TRUE;
   }

   return isOk;

}
/*===========================================================================*
 *     SynonymDefODBC::writeSynonymDefinition
 *===========================================================================*/
SAPDB_Bool SynonymDefODBC::writeSynonymDefinition()
{
   ROUTINE_DBG_MEO00 ("SynonymDefODBC::writeSynonymDefinition");

   RepServMessages *pMsgInstance = RepServMessages::Instance();

   StudioOAL_String tableSynonym;

   if ( getCatalog()->getExtrMode() == emExtractAsXML_els31 )
        tableSynonym += genCreateXML();

   if ( getCatalog()->getExtrMode() == emExtractAsDDL_els31 )
   {
        tableSynonym  = SEPARATOR;

        // Special handling of 'foreign' sqlmodes
        if ( getCatalog()->getSqlMode() != cpr_kind_internal )
        {
             SAPDB_Int4 iSqlMode = getCatalog()->getSqlMode();
             tableSynonym += pszSQLModeString_ls31[iSqlMode];
             tableSynonym += SEPARATOR;
        }

        tableSynonym += genCreateDDL();
        //!! ToDo COMMENT ON
   }
   return writeLine( tableSynonym );

}
/*===========================================================================*
 *     SynonymDefODBC::readSynonymDefResult
 *===========================================================================*/
SAPDB_Bool SynonymDefODBC::readSynonymDefResult( StudioOAL_String *record )
{
   SAPDB_Int2       i;
   StudioOAL_String  s;

  /*-------------------------------------------------------------------------*
   * Fetch the next record.                                                  *
   *-------------------------------------------------------------------------*/
  // SQLFetch
  if (!m_pSynonymDefResult->fetch())
  {
      m_pSynonymDefResult->closeCursor();
      return SAPDB_FALSE;
  }

  /*-------------------------------------------------------------------------*
   * If there are no more records, reading the result is finished.           *
   *-------------------------------------------------------------------------*/
  if (m_pSynonymDefResult->noRows())
  {
     m_pSynonymDefResult->closeCursor();
     return SAPDB_FALSE;
  }

  /*-------------------------------------------------------------------------*
   * Read all columns of the fetched record.                                 *
   *-------------------------------------------------------------------------*/
  for (i = 0; i < m_pSynonymDefResult->getColCount(); i++)
  {
    m_pSynonymDefResult->bindColumnDataAsString (i + 1, s);
    record[i] = s;
  }

  return SAPDB_TRUE;
}
/*===========================================================================*
 *     SynonymDefODBC::readResult()
 *===========================================================================*/
SAPDB_Bool SynonymDefODBC::readResult( StudioOAL_String *record )
{
    SAPDB_Bool isOk = SAPDB_FALSE;

    // SYNONYMS  : OWNER, SYNONYMNAME, PUBLIC, TABLEOWNER, TABLENAME, COMMENT

    isOk = CatalogObjectDefODBC::readResult( record);

    if ( isOk )
    {
         m_owner              = record[0];
         m_synonymName        = record[1];
         m_public             = record[2];
         m_tableowner         = record[3];
         m_tablename          = record[4];
         m_comment            = record[5];
    }

    return isOk;
}
/*===========================================================================*
 *     SynonymDefODBC::writeResult()
 *===========================================================================*/
SAPDB_Bool SynonymDefODBC::writeResult( StudioOAL_String *record )
{
      StudioOAL_String tableSynonym;

      if ( getCatalog()->getExtrMode() == emExtractAsXML_els31 )
      {
           tableSynonym += genCreateXML();
      }

      if ( getCatalog()->getExtrMode() == emExtractAsDDL_els31 )
      {
           tableSynonym  = SEPARATOR;
           tableSynonym += genCreateDDL();
           //!! ToDo COMMENT ON

      }
      return writeLine( tableSynonym );
}
/*===========================================================================*
 *     SynonymDefODBC::genCreateXML()
 *===========================================================================*/
StudioOAL_String SynonymDefODBC::genCreateXML()
{
    // <dbm:SchemaTableSynonyms>

   StudioOAL_String  schemaTableSynonym;

   schemaTableSynonym = "<dbm:TableSynonym oim:id=\"_";
   schemaTableSynonym += string2XML(m_owner);
   schemaTableSynonym += ".";
   schemaTableSynonym += string2XML(m_synonymName);
   schemaTableSynonym += "\" oim:href= \"#_";
   schemaTableSynonym += string2XML(m_tableowner);
   schemaTableSynonym += ".";
   schemaTableSynonym += string2XML(m_tablename);
   schemaTableSynonym += "\" oim:seqno  = \"";
   schemaTableSynonym += "\" dbm:name   = \"";
   schemaTableSynonym += string2XML(m_synonymName);
   schemaTableSynonym += "\" dbm:description = \"";
   schemaTableSynonym += string2XML(m_synonymDefinition);
   schemaTableSynonym += "\" dbm:isSpecification = \"1";
   schemaTableSynonym += "\" dbm:comments = \"";
   schemaTableSynonym += string2XML(m_comment);
   schemaTableSynonym += "\"/>\n";

   return schemaTableSynonym;
}
/*===========================================================================*
 *     SynonymDefODBC::genCreateDDL()
 *===========================================================================*/
StudioOAL_String SynonymDefODBC::genCreateDDL()
{
    // CREATE SYNONYM

   m_synonymDefinition  = "CREATE ";
   m_public.Compare( "NO" )           != 0 ? m_synonymDefinition += m_public : m_synonymDefinition += "";
   m_synonymDefinition += " SYNONYM ";
   m_synonymDefinition += " \"";
   m_owner.Compare( NULL_VALUE )      != 0 ? m_synonymDefinition += m_owner  : m_synonymDefinition += ""; // !getColumn(1)->getColIsNull()
   m_owner.Compare( NULL_VALUE )      != 0 ? m_synonymDefinition += " \".\"" : m_synonymDefinition += ""; // !getColumn(1)->getColIsNull()
   m_synonymDefinition += m_synonymName;
   m_synonymDefinition += " \" FOR \"";
   m_tableowner.Compare( NULL_VALUE ) != 0 ? m_synonymDefinition += m_tableowner : m_synonymDefinition += ""; // !getColumn(4)->getColIsNull()
   m_synonymDefinition += "\".\"";
   m_synonymDefinition += m_tablename;
   m_synonymDefinition += "\"";

   return m_synonymDefinition;
}
/*===========================================================================*
 *    ColumnTypeDefODBC::ColumnTypeDefODBC
 *===========================================================================*/
ColumnTypeDefODBC::ColumnTypeDefODBC( sapdbwa_DBC & odbcConnection , CatalogDefODBC * pCatalog  )
                  :CatalogObjectDefODBC ( odbcConnection , pCatalog )
{
    // Trace - Writer
    ROUTINE_DBG_MEO00 ("ColumnTypeDefODBC::ColumnTypeDefODBC()");

    // Log   - Writer
    RepServMessages *pMsgInstance = RepServMessages::Instance();

    initResultSet( odbcConnection );

}
/*===========================================================================*
 *     ColumnTypeDefODBC::ColumnTypeDefODBC
 *===========================================================================*/
ColumnTypeDefODBC::ColumnTypeDefODBC( sapdbwa_DBC & odbcConnection )
                  :CatalogObjectDefODBC ( odbcConnection )
{
    // Trace - Writer
    ROUTINE_DBG_MEO00 ("ColumnTypeDefODBC::ColumnTypeDefODBC()");

    // Log   - Writer
    RepServMessages *pMsgInstance = RepServMessages::Instance();

}
/*===========================================================================*
 *     ColumnTypeDefODBC::~ColumnTypeDefODBC
 *===========================================================================*/
ColumnTypeDefODBC::~ColumnTypeDefODBC()
{
}
/*===========================================================================*
 *     ColumnTypeDefODBC::initResultSet
 *===========================================================================*/
SAPDB_Bool ColumnTypeDefODBC::initResultSet ( sapdbwa_DBC & odbcConnection )
{

  SAPDB_Bool isOk = SAPDB_FALSE;

  if (  odbcConnection.HDBC() && odbcConnection.HENV() )
  {
        isOk = SAPDB_TRUE;
  }

  return isOk;
}
/*===========================================================================*
 *    ColumnTypeDefODBC::readResult
 *===========================================================================*/
SAPDB_Bool ColumnTypeDefODBC::readResult ( StudioOAL_String *record  )
{
    return CatalogObjectDefODBC::readResult( record);
}
/*===========================================================================*
 *    ColumnTypeDefODBC::writeResult
 *===========================================================================*/
SAPDB_Bool ColumnTypeDefODBC::writeResult( StudioOAL_String *record  )
{
    ROUTINE_DBG_MEO00 ("ColumnTypeDefODBC::writeResult ()");

    RepServMessages *pMsgInstance = RepServMessages::Instance();

    if ( getCatalog()->getExtrMode() == emExtractAsXML_els31 )
    {

        // SYSODBCTYPES  :  0.TYPE_NAME, 1.DATA_TYPE, 2.PRECISION, 3.LITERAL_PREFIX, 4.LITERAL_SUFFIX, 5.CREATE_PARAMS, 6.NULLABLE, 7.CASE_SENSITIVE, 8.SEARCHABLE, 9.UNSIGNED_ATTRIBUTE, 10.MONEY, 11.AUTO_INCREMENT
        //                 12.LOCAL_TYPE_NAME, 13.MINIMUM_SCALE, 14.MAXIMUM_SCALE, 15.SQL_DATA_TYPE, 16.SQL_DATETIME_SUB, 17.NUM_PREC_RADIX, 18.INTERVAL_PRECISION, 19.REMARKS

        // SQLGetTypeInfo( hStmt, SQL_ALL_TYPES )

        // <data_typ> ::=
        //  CHAR[ACTER] [(<unsigned_integer>)] [ASCII | BYTE | EBCDIC]
        //| VARCHAR     [(<unsigned_integer>)] [ASCII | BYTE | EBCDIC]
        //| LONG [VARCHAR]                     [ASCII | BYTE | EBCDIC]
        //| BOOLEAN
        //| FIXED (<unsigned_integer> [,<unsigned_integer>])
        //| FLOAT (<unsigned_integer>)
        //| INT[EGER] | SMALLINT
        //| DATE | TIME | TIMESTAMP

        StudioOAL_String columntype;
        StudioOAL_String columntypeid;
        StudioOAL_String strLong     = "LONG";
        StudioOAL_String strDouble   = "DOUB";

        columntype  = "<sapdbm:";

        if ( strLong  .Compare(record[12].SubStr(0,4)) == 0  && record[12].Length() > 4)
             columntypeid = "LONG_BYTE" ;
        if ( strDouble.Compare(record[12].SubStr(0,4)) == 0  )
             columntypeid = "DOUBLE_PRECISION";
        if ( strLong  .Compare(record[12].SubStr(0,4)) == 0  && record[12].Length() == 4)
             columntypeid = record[12];
        if ( strLong  .Compare(record[12].SubStr(0,4)) != 0 && strDouble.Compare(record[12].SubStr(0,4)) != 0  )
             columntypeid = record[12];

        columntype += columntypeid;

        columntype += " oim:id=\"_";
        columntype += "sapdbm:";
        columntype += columntypeid;
        columntype += "\" oim:label=\"";                                                    // SQLColAttribute
        !getColumn(13)->getColIsNull() ? columntype += record[12] : columntype+= "";        // record[12].Compare( NULL_VALUE )  != 0
        columntype += "\" oim:supertype=\"";
        columntype += "dbm:ColumnType";
        columntype += "\" name=\"";                                                         // SQL_DESC_TYPE(ODBC 3.0), SQL_DESC_TYPE_NAME(ODBC 1.0), SQL_DESC_NAME(ODBC 3.0)
        !getColumn(1)->getColIsNull() ? columntype += record[0]  : columntype+= "";         // SQL_DESC_LABEL(ODBC 2.0) record[0].Compare( NULL_VALUE )   != 0
        columntype += "\" ColumnSize=\"";                                                   // SQL_DESC_LENGTH (ODBC 3.0) ,SQLDescribeCol
        !getColumn(3)->getColIsNull() ? columntype += record[2]  : columntype+= "";         // record[2].Compare( NULL_VALUE )   != 0
        //columntype += " \"IsFixedLength=\"";
        //columntype += record[];
        columntype += "\" LiteralPrefix=\"";                                                // SQL_DESC_LITERAL_PREFIX(ODBC 3.0)
        !getColumn(4)->getColIsNull() ? columntype += record[3]  : columntype+= "";         // record[3].Compare( NULL_VALUE )  != 0
        columntype += "\" LiteralSuffix=\"";                                                // SQL_DESC_LITERAL_SUFFIX(ODBC 3.0)
        !getColumn(5)->getColIsNull() ? columntype += record[4]  : columntype+= "";         // record[4].Compare( NULL_VALUE )  != 0
        columntype += "\" CreateParams=\"";
        !getColumn(6)->getColIsNull() ? columntype += record[5] : columntype+= "";          // record[5].Compare( NULL_VALUE )  != 0
        columntype += "\" IsNullable=\"";                                                   // SQL_DESC_NULLABLE(ODBC 3.0)
        !getColumn(7)->getColIsNull() ? columntype += record[6]  : columntype+= "";         // record[6].Compare( NULL_VALUE )  != 0
        columntype += "\" IsCaseSensitive=\"";                                              // SQL_DESC_CASE_SENSITIVE(ODBC 1.0)
        !getColumn(8)->getColIsNull() ? columntype += record[7]  : columntype+= "";         // record[7].Compare( NULL_VALUE )  != 0
        columntype += "\" Searchable=\"";                                                   // SQL_DESC_SEARCHABLE(ODBC 1.0)
        !getColumn(9)->getColIsNull() ? columntype += record[8] : columntype+= "";          // record[8].Compare( NULL_VALUE )  != 0
        columntype += "\" IsUnsignedAttribute=\"";
        !getColumn(10)->getColIsNull()? columntype += record[9]  : columntype += "";        // record[9].Compare( NULL_VALUE )  != 0
        columntype += "\" IsAutoUniqueValue=\"";                                            // SQL_DESC_AUTO_UNIQUE_VALUE(ODBC 1.0)
        !getColumn(12)->getColIsNull()? columntype += record[11] : columntype+= "";         // record[11].Compare( NULL_VALUE ) != 0
        columntype += "\" MinimumScale=\"";                                                 // SQL_DESC_SCALE(ODBC 3.0)
        !getColumn(14)->getColIsNull()? columntype += record[13]: columntype+= "";          // SQL_DESC_UNSIGNED(ODBC 1.0) record[13].Compare( NULL_VALUE ) != 0
        columntype += "\" MaximumScale=\"";
        !getColumn(15)->getColIsNull()? columntype += record[14]: columntype+= "";          // SQL_DESC_FIXED_PREC_SCALE(ODBC 1.0) record[14].Compare( NULL_VALUE ) != 0
        columntype += "\" IsFixedPrecisionScale=\"";                                        // SQL_DESC_PRECISION(ODBC 3.0)
        !getColumn(3)->getColIsNull() ? columntype += record[2] : columntype+= "";          // record[2].Compare( NULL_VALUE )  != 0

        columntype += "\" IsLong=\"";
        !getColumn(13)->getColIsNull() && strLong.Compare(record[12].SubStr(0,4)) == 0 ? columntype += "1" : columntype += "0"; // record[12].Compare( NULL_VALUE ) != 0

        columntype += "\"/>";

        return writeLine( columntype );
    }

    return SAPDB_TRUE;

}
/*===========================================================================*
 *    TriggerDefODBC::TriggerDefODBC
 *===========================================================================*/
TriggerDefODBC::TriggerDefODBC( sapdbwa_DBC & odbcConnection , CatalogDefODBC * pCatalog  )
               :CatalogObjectDefODBC ( odbcConnection , pCatalog )
{
    // Trace - Writer
    ROUTINE_DBG_MEO00 ("TriggerDefODBC::TriggerDefODBC()");

    // Log   - Writer
    RepServMessages *pMsgInstance = RepServMessages::Instance();

#ifdef UNICODE
    m_pTriggerDefResult = new ODBC_WResult ( odbcConnection );
#endif
#ifndef UNICODE
    m_pTriggerDefResult = new ODBC_AResult ( odbcConnection );
#endif

    initResultSet( odbcConnection );
}
/*===========================================================================*
 *     TriggerDefODBC::TriggerDefODBC()
 *===========================================================================*/
TriggerDefODBC::TriggerDefODBC( sapdbwa_DBC & odbcConnection )
               :CatalogObjectDefODBC ( odbcConnection )
{
    // Trace - Writer
    ROUTINE_DBG_MEO00 ("TriggerDefODBC::TriggerDefODBC()");

    // Log   - Writer
    RepServMessages *pMsgInstance = RepServMessages::Instance();

#ifdef UNICODE
    m_pTriggerDefResult = new ODBC_WResult ( odbcConnection );
#endif
#ifndef UNICODE
    m_pTriggerDefResult = new ODBC_AResult ( odbcConnection );
#endif
}

/*===========================================================================*
 *     TriggerDefODBC::~TriggerDefODBC()
 *===========================================================================*/
TriggerDefODBC::~TriggerDefODBC( )
{
    // Trace - Writer
    ROUTINE_DBG_MEO00 ("TriggerDefODBC::~TriggerDefODBC()");

    // Log   - Writer
    RepServMessages *pMsgInstance = RepServMessages::Instance();

    if ( m_pTriggerDefResult != NULL )
    {
         delete m_pTriggerDefResult;
         m_pTriggerDefResult = NULL;
    }
}
/*===========================================================================*
 *     TriggerDefODBC::initResultSet
 *===========================================================================*/
SAPDB_Bool TriggerDefODBC::initResultSet ( sapdbwa_DBC & odbcConnection )
{

  SAPDB_Bool isOk = SAPDB_FALSE;

  if (  odbcConnection.HDBC() && odbcConnection.HENV() )
  {
        StudioOAL_String  triggerDef;
        StudioOAL_String  whereClause;

        /*-------------------------------------------------------------------------*
         * Prepare the WHERE clause for the SQL statement.
         *-------------------------------------------------------------------------*/

        whereClause  = " WHERE definition IS NOT NULL ";
        whereClause += " AND   owner     = ? ";
        whereClause += " AND   tablename = ? ";

        triggerDef   = "SELECT DEFINITION, TRIGGERNAME, TABLENAME, COMMENT, OWNER ";
        triggerDef  += "FROM DOMAIN.TRIGGERS ";
        triggerDef  += whereClause;

        // SQLAllocStmt     SQLPrepare
        // SQLNumParams     SQLDescribeParam SQLBindParameter
        // SQLNumResultCols SQLDescribeCol   SQLBindCol

        if ( !m_pTriggerDefResult->init( triggerDef ) )
        {
             getCatalog()->getError()->setErrorText(getErrorText() );
             getCatalog()->getError()->isError     (isError     () );

             return SAPDB_FALSE;
        }

        isOk = SAPDB_TRUE;
  }

  return isOk;
}
/*===========================================================================*
 *     TriggerDefODBC::readTriggerDefinition
 *===========================================================================*/
SAPDB_Bool TriggerDefODBC::readTriggerDefinition( StudioOAL_String owner, StudioOAL_String tablename )
{
   ROUTINE_DBG_MEO00 ("ViewDefODBC::readTriggerDefinition()");

   RepServMessages *pMsgInstance = RepServMessages::Instance();

   SAPDB_Bool isOk = SAPDB_FALSE;

   StudioOAL_String  triggerDef[2];
   StudioOAL_String  record[5];

   triggerDef[0] = owner;
   triggerDef[1] = tablename;

  /*-------------------------------------------------------------------------*
   * Load the trigger definitions from the database.                           *
   *-------------------------------------------------------------------------*/

   if ( !m_pTriggerDefResult->setParamValues( triggerDef ) )
   {
        getCatalog()->getError()->setErrorText(getErrorText() );
        getCatalog()->getError()->isError     (isError     () );
        return SAPDB_FALSE;
   }
   /*-------------------------------------------------------------------------*
    * Execute the prepared statement                                          *
    *-------------------------------------------------------------------------*/
   // SQLExecute
   if ( !m_pTriggerDefResult->execute() )
   {
        StudioOAL_String errorText = m_pTriggerDefResult->getErrorText();
        SAPDB_Bool      isError   = m_pTriggerDefResult->isError     ();

        getCatalog()->getError()->setErrorText( errorText );
        getCatalog()->getError()->isError     ( isError   );

        return SAPDB_FALSE;
   }
   // SQlFetch SQLFreeStatement
   if ( readTriggerDefResult (record) )
   {
        m_triggerDefinition = record[0];
        m_triggerName       = record[1];
        m_tableName         = record[2];
        m_comment           = record[3];
        m_owner             = record[4];

        isOk = SAPDB_TRUE;
   }

   return isOk;

}
/*===========================================================================*
 *     TriggerDefODBC::writeTriggerDefinition
 *===========================================================================*/
SAPDB_Bool TriggerDefODBC::writeTriggerDefinition()
{
   ROUTINE_DBG_MEO00 ("TriggerDefODBC::writeTriggerDefinition");

   RepServMessages *pMsgInstance = RepServMessages::Instance();

   SAPDB_Bool isOk = SAPDB_FALSE;

   StudioOAL_String tableTrigger;

   if ( getCatalog()->getExtrMode() == emExtractAsDDL_els31 )
   {
      tableTrigger  = SEPARATOR;

      // Special handling of 'foreign' sqlmodes
      if ( getCatalog()->getSqlMode() != cpr_kind_internal )
      {
           SAPDB_Int4 iSqlMode = getCatalog()->getSqlMode();
           tableTrigger += pszSQLModeString_ls31[iSqlMode];
           tableTrigger += SEPARATOR;
      }

      tableTrigger += m_triggerDefinition;

      // COMMENT ON TRIGGER <trigger_name> ON < table_name> IS <comment>
      if ( !m_comment.Empty() && m_comment.Compare( NULL_VALUE) != 0 )
//    if ( !getColumn(4)->getColIsNull() && !m_comment.Empty() ) // record[3]!= NULL_VALUE
      {
           tableTrigger += NEW_LINE;
           tableTrigger += SEPARATOR;
           tableTrigger += "COMMENT ON TRIGGER \"";
           tableTrigger += m_triggerName;
           tableTrigger += "\" ON \"";
           tableTrigger += m_tableName;
           tableTrigger += "\" IS ' ";
           tableTrigger += EscapeSingleQuotes(m_comment);
           tableTrigger += "'";

      }

   }
   if ( getCatalog()->getExtrMode() == emExtractAsXML_els31 )
       tableTrigger += genCreateXML();

   return writeLine( tableTrigger );

}
/*===========================================================================*
 *     TriggerDefODBC::readTriggerDefResult
 *===========================================================================*/
SAPDB_Bool TriggerDefODBC::readTriggerDefResult( StudioOAL_String *record )
{
   SAPDB_Int2       i;
   StudioOAL_String  s;

  /*-------------------------------------------------------------------------*
   * Fetch the next record.                                                  *
   *-------------------------------------------------------------------------*/
  // SQLFetch
  if (!m_pTriggerDefResult->fetch())
  {
      m_pTriggerDefResult->closeCursor();
      return SAPDB_FALSE;
  }

  /*-------------------------------------------------------------------------*
   * If there are no more records, reading the result is finished.           *
   *-------------------------------------------------------------------------*/
  if (m_pTriggerDefResult->noRows())
  {
     m_pTriggerDefResult->closeCursor();
     return SAPDB_FALSE;
  }

  /*-------------------------------------------------------------------------*
   * Read all columns of the fetched record.                                 *
   *-------------------------------------------------------------------------*/
  for (i = 0; i < m_pTriggerDefResult->getColCount(); i++)
  {
    m_pTriggerDefResult->bindColumnDataAsString (i + 1, s);
    record[i] = s;
  }

  return SAPDB_TRUE;
}
/*===========================================================================*
 *     TriggerDefODBC::readResult()
 *===========================================================================*/
SAPDB_Bool TriggerDefODBC::readResult( StudioOAL_String *record )
{
    SAPDB_Bool isOk = SAPDB_FALSE;

    isOk = CatalogObjectDefODBC::readResult( record);

    if ( isOk )
    {
         m_triggerName       = record[1];
         m_owner             = record[4];
         m_triggerDefinition = record[0];

         m_tableName         = record[2] ;
         m_comment           = record[3];

    }

    return isOk;
}
/*===========================================================================*
 *     TriggerDefODBC::writeResult()
 *===========================================================================*/
SAPDB_Bool TriggerDefODBC::writeResult( StudioOAL_String *record )
{
      StudioOAL_String tableTrigger;

      if ( getCatalog()->getExtrMode() == emExtractAsDDL_els31 )
      {
          tableTrigger  = SEPARATOR;
          tableTrigger += record[0];

          // COMMENT ON TRIGGER <trigger_name> ON < table_name> IS <comment>
          if ( !getColumn(4)->getColIsNull() && !record[3].Empty() ) // record[3]!= NULL_VALUE
          {
               tableTrigger += NEW_LINE;
               tableTrigger += SEPARATOR;
               tableTrigger += "COMMENT ON TRIGGER \"";
               tableTrigger += record[1];
               tableTrigger += "\" ON \"";
               tableTrigger += record[2];
               tableTrigger += "\" IS ' ";
               tableTrigger += EscapeSingleQuotes(record[3]);
               tableTrigger += "'";

          }

      }
      if ( getCatalog()->getExtrMode() == emExtractAsXML_els31 )
           tableTrigger += genCreateXML();

      return writeLine( tableTrigger );
}
/*===========================================================================*
 *     TriggerDefODBC::genCreateXML()
 *===========================================================================*/
StudioOAL_String TriggerDefODBC::genCreateXML()
{
    // <dbm:TableTrigger>

    StudioOAL_String tableTrigger;

   tableTrigger  = "<dbm:DeployedTrigger oim:id=\"_";
   tableTrigger += string2XML( m_owner );
   tableTrigger += ".";
   tableTrigger += string2XML( m_triggerName );
   tableTrigger += ".";
   tableTrigger += string2XML( m_tableName );
   tableTrigger += "\" oim:supertype = \"Trigger";
   tableTrigger += "\" dbm:StatementsLanguage = \"";
   tableTrigger += "\" dbm:Statements         = \"";
   tableTrigger += string2XML ( m_triggerDefinition );
   tableTrigger += "\" dbm:isSpecification = \"1";
   tableTrigger += "\" dbm:comments = \"";
   tableTrigger += string2XML ( m_comment );
   tableTrigger += "\"/>\n";

   //IsDelete, IsInsert, IsUpdate, isQuery, name, ownerScope, Statements, StatementsLanguage

   return tableTrigger;
}
/*===========================================================================*
 *    DBProcedureDefODBC::DBProcedureDefODBC
 *===========================================================================*/
DBProcedureDefODBC::DBProcedureDefODBC( sapdbwa_DBC & odbcConnection , CatalogDefODBC * pCatalog  )
                   :CatalogObjectDefODBC ( odbcConnection , pCatalog )
{
    m_pGrantDBProcedureDefODBC = new GrantDBProcedureDefODBC( odbcConnection, pCatalog );

    initResultSet( odbcConnection );
}
/*===========================================================================*
 *     DBProcedureDefODBC::DBProcedureDefODBC()
 *===========================================================================*/
DBProcedureDefODBC::DBProcedureDefODBC(sapdbwa_DBC & odbcConnection )
                   :CatalogObjectDefODBC ( odbcConnection )
{
    m_pGrantDBProcedureDefODBC = new GrantDBProcedureDefODBC( odbcConnection );
}
/*===========================================================================*
 *     DBProcedureDefODBC::~DBProcedureDefODBC()
 *===========================================================================*/
DBProcedureDefODBC::~DBProcedureDefODBC()
{
    if( m_pGrantDBProcedureDefODBC != NULL )
    {
        delete m_pGrantDBProcedureDefODBC;
        m_pGrantDBProcedureDefODBC = NULL;
    }
}
/*===========================================================================*
 *     DBProcedureDefODBC::initResultSet
 *===========================================================================*/
SAPDB_Bool DBProcedureDefODBC::initResultSet ( sapdbwa_DBC & odbcConnection )
{
  ROUTINE_DBG_MEO00 ("DBProcedureDefODBC::initResult ()");

  SAPDB_Bool isOk = SAPDB_FALSE;

  if (  odbcConnection.HDBC() && odbcConnection.HENV() )
  {
        // CREATE DBPROC
        // DBPROCEDURES : OWNER,DBPROCNAME,PACKAGE,PARAMETER,CREATEDATE,CREATETIME,
        //                EXCECUTION_KIND,SQL_SUPPORT,REMOTE_LOCATION,DEFINITION,COMMENT

        StudioOAL_String  dbProceduredef;
        StudioOAL_String  whereClause;

        //!! ToDo PTS 1109226
        dbProceduredef  = "SELECT DEFINITION, DBPROCNAME, OWNER, COMMENT ";
        dbProceduredef += "FROM DOMAIN.DBPROCEDURES ";

        /*-------------------------------------------------------------------------*
         * Prepare the WHERE clause for the SQL statements.                        *
         *-------------------------------------------------------------------------*/

        whereClause     = " WHERE definition IS NOT NULL ";
        whereClause    += " AND   owner = ? ";

        dbProceduredef += whereClause;

        isOk = SAPDB_TRUE;
  }

  return isOk;
}
/*===========================================================================*
 *     DBProcedureDefODBC::readResult
 *===========================================================================*/
SAPDB_Bool DBProcedureDefODBC::readResult(  StudioOAL_String *record )
{
    SAPDB_Bool isOk = SAPDB_FALSE;

    isOk = CatalogObjectDefODBC::readResult( record);

    if( isOk )
    {
        m_dbprocedureName       = record[1];
        m_owner                 = record[2];
        m_dbprocedureDefinition = record[0];
        m_comment               = record[3];
    }

    return isOk;

}
/*===========================================================================*
 *     DBProcedureDefODBC::writeResult
 *===========================================================================*/
SAPDB_Bool DBProcedureDefODBC::writeResult(  StudioOAL_String *record )
{
    StudioOAL_String dbProcedure;

    if ( getCatalog()->getExtrMode() == emExtractAsDDL_els31 )
    {
        // DEFINITION, DBPROCNAME, OWNER, COMMENT
        dbProcedure  = SEPARATOR;

        // Special handling of 'foreign' sqlmodes
        if ( getCatalog()->getSqlMode() != cpr_kind_internal )
        {
            SAPDB_Int4 iSqlMode = getCatalog()->getSqlMode();
            dbProcedure += pszSQLModeString_ls31[iSqlMode];
            dbProcedure += SEPARATOR;
        }

        dbProcedure += record[0];

        // COMMENT ON DBPROC <dbproc_name> IS <comment>
        if ( !getColumn(4)->getColIsNull() && !record[3].Empty() ) 
        {
            dbProcedure += NEW_LINE;
            dbProcedure += SEPARATOR;
            dbProcedure += "COMMENT ON DBPROC \"";
            dbProcedure += record[1];
            dbProcedure += "\" IS '";
            dbProcedure += EscapeSingleQuotes(record[3]);
            dbProcedure += "'";
        }
        // CREATE DBPROC
        writeLine(dbProcedure);
        // GRANT EXECUTE ON <dbproc_name> TO <grantee>,...
        grantDBProcedureExtrCmd( record[1] );
    }
    if ( getCatalog()->getExtrMode() == emExtractAsXML_els31 )
        dbProcedure  = genCreateXML();

    return SAPDB_TRUE;
}
/*===========================================================================*
 *     DBProcedureDefODBC::genCreateXML()
 *===========================================================================*/
StudioOAL_String  DBProcedureDefODBC::genCreateXML()
{
    // <dbm:StoredProcedure>

    StudioOAL_String dbProcedure;

    dbProcedure  = "<dbm:StoredProcedure oim:id=\"_";
    dbProcedure += string2XML( m_owner );
    dbProcedure += ".";
    dbProcedure += string2XML( m_dbprocedureName );
    dbProcedure += "\" oim:label=\"";
    dbProcedure += string2XML( m_dbprocedureName );
    dbProcedure += "\" dbm:bodyLanguage = \"&lt;create_dbproc_statement&gt;::= see SAPDB reference manual";
    dbProcedure += "\" dbm:body         = \"";
    dbProcedure += string2XML( m_dbprocedureDefinition );
    dbProcedure += "\" dbm:name= \"";
    dbProcedure += string2XML( m_dbprocedureName );
    dbProcedure += "\" dbm:isQuery =\"";
    dbProcedure += "\" dbm:isSpecification = \"1";
    dbProcedure += "\" dbm:comments = \"";
    dbProcedure += string2XML ( m_comment );
    dbProcedure += "\"/>\n";

    return dbProcedure;
}
/*===========================================================================*
 *  DBProcedureDefODBC::grantDBProcedureExtrCmd
 *===========================================================================*/
SAPDB_Int4 DBProcedureDefODBC::grantDBProcedureExtrCmd( StudioOAL_String dbprocname )
{
    ROUTINE_DBG_MEO00 ("DBProcedureDefODBC::grantDBProcedureExtrCmd");

    SAPDB_Int4 rc = errOK_els00;

    // GRANT EXECUTE ON <dbproc_name> TO <grantee>,...
    // <grantee> ::= <usergroup_name> | <user_name> | <role_name> | PUBLIC

    // SYSDD.PRIVILEGES : OWNER, NAME1, NAME2, TYPE, GRANTEE, GRANTEEOWNER,
    //                    PRIVILEGES, GRANTOR, DATE, TIME
    StudioOAL_String  select;
    StudioOAL_String  whereClause;
    StudioOAL_String  record[2];

    StudioOAL_String  grantdbProcedure;

    select  = "SELECT NAME1, GRANTEE FROM SYSDD.PRIVILEGES ";

    /*-------------------------------------------------------------------------*
     * Prepare the WHERE clause for the SQL statements.
     *-------------------------------------------------------------------------*/

    whereClause  = " WHERE type   = 'DBPROC' " //'SEQUENCE'
                   "  AND  name1  = '";
    whereClause +=  m_dbprocedureName ;
    whereClause += "' AND  grantee <> '";
    whereClause +=  m_owner;
    whereClause += "'";

    select      += whereClause;

    SAPDB_Bool      isError       = SAPDB_TRUE;
    SAPDB_Long      getNativeError= 0;
    SAPDB_Int4      getErrorPos   = 0;
//    SQLRETURN       getReturnCode = 0;
//    StudioOAL_String getErrorText;
    StudioOAL_String getNativeErrorAsString;


    if ( !m_pGrantDBProcedureDefODBC->executeSQL( select )  )
    {
        isError                = m_pGrantDBProcedureDefODBC->isError();
        getNativeError         = m_pGrantDBProcedureDefODBC->getNativeError();
        getErrorPos            = m_pGrantDBProcedureDefODBC->getErrorPos();
//      getReturnCode          = m_pGrantDBProcedureDefODBC->getReturnCode();
//      getErrorText           = m_pGrantDBProcedureDefODBC->getErrorText();
        getNativeErrorAsString = m_pGrantDBProcedureDefODBC->getNativeErrorAsString();
        rc = STATIC_CAST(SAPDB_Int4, getNativeError);
    }
    else
    {
        while ( m_pGrantDBProcedureDefODBC->readResult (record ) )
        {
            grantdbProcedure  = SEPARATOR;
            grantdbProcedure += "GRANT EXECUTE ON \"";
            grantdbProcedure += record[0];
            grantdbProcedure += "\" TO ";
            grantdbProcedure += record[1];

            m_pGrantDBProcedureDefODBC->writeResult(grantdbProcedure);
        }
    }
    return rc;
}
/*===========================================================================*
 *    GrantDBProcedureDefODBC::GrantDBProcedureDefODBC
 *===========================================================================*/
GrantDBProcedureDefODBC::GrantDBProcedureDefODBC( sapdbwa_DBC & odbcConnection , CatalogDefODBC * pCatalog  )
                        :CatalogObjectDefODBC   ( odbcConnection , pCatalog )
{
    ROUTINE_DBG_MEO00 ("GrantDBProcedureDefODBC::GrantDBProcedureDefODBC()");
}
/*===========================================================================*
 *    GrantDBProcedureDefODBC::GrantDBProcedureDefODBC
 *===========================================================================*/
GrantDBProcedureDefODBC::GrantDBProcedureDefODBC( sapdbwa_DBC & odbcConnection )
                        :CatalogObjectDefODBC   ( odbcConnection )
{
    ROUTINE_DBG_MEO00 ("GrantDBProcedureDefODBC::GrantDBProcedureDefODBC()");

}
/*===========================================================================*
 *    GrantDBProcedureDefODBC::GrantDBProcedureDefODBC
 *===========================================================================*/
GrantDBProcedureDefODBC::~GrantDBProcedureDefODBC()
{
    ROUTINE_DBG_MEO00 ("GrantDBProcedureDefODBC::~GrantDBProcedureDefODBC()");
}
/*===========================================================================*
 *     GrantDBProcedureDefODBC::initResultSet
 *===========================================================================*/
SAPDB_Bool GrantDBProcedureDefODBC::initResultSet ( sapdbwa_DBC & odbcConnection )
{
    ROUTINE_DBG_MEO00 ("GrantDBProcedureDefODBC::initResult ()");

    SAPDB_Bool isOk = SAPDB_FALSE;

    if (  odbcConnection.HDBC() && odbcConnection.HENV() )
    {
        isOk = SAPDB_TRUE;
    }
    return isOk;
}
/*===========================================================================*
 *    GrantDBProcedureDefODBC::readResult
 *===========================================================================*/
SAPDB_Bool GrantDBProcedureDefODBC::readResult ( StudioOAL_String *record  )
{
    ROUTINE_DBG_MEO00 ("GrantDBProcedureDefODBC::readResult ()");

    return CatalogObjectDefODBC::readResult( record);
}
/*===========================================================================*
 *    GrantDBProcedureDefODBC::writeResult
 *===========================================================================*/
SAPDB_Bool GrantDBProcedureDefODBC::writeResult( StudioOAL_String record  )
{
    ROUTINE_DBG_MEO00 ("GrantDBProcedureDefODBC::writeResult ()");

    return CatalogObjectDefODBC::writeLine ( record);
}
/*===========================================================================*
 *    IndexDefODBC::IndexDefODBC
 *===========================================================================*/
IndexDefODBC::IndexDefODBC( sapdbwa_DBC & odbcConnection , CatalogDefODBC * pCatalog  )
             :CatalogObjectDefODBC ( odbcConnection , pCatalog )
{
    // Trace - Writer
    ROUTINE_DBG_MEO00 ("IndexDefODBC::IndexDefODBC()");

    // Log   - Writer
    RepServMessages *pMsgInstance = RepServMessages::Instance();

#ifdef UNICODE
    m_pIndexDefResult    = new ODBC_WResult ( odbcConnection );
    m_pIndexColDefResult = new ODBC_WResult ( odbcConnection );
#endif
#ifndef UNICODE
    m_pIndexDefResult    = new ODBC_AResult ( odbcConnection );
    m_pIndexColDefResult = new ODBC_AResult ( odbcConnection );
#endif

    initResultSet( odbcConnection );

    m_pIndexColDefODBC   = new IndexColumnDefODBC( odbcConnection, pCatalog );
    m_pIndexColDefODBC->setIndexDefODBC( this );
}
/*===========================================================================*
 *     IndexDefODBC::IndexDefODBC()
 *===========================================================================*/
IndexDefODBC::IndexDefODBC( sapdbwa_DBC & odbcConnection )
             :CatalogObjectDefODBC ( odbcConnection )
{
    // Trace - Writer
    ROUTINE_DBG_MEO00 ("IndexDefODBC::IndexDefODBC()");

    // Log   - Writer
    RepServMessages *pMsgInstance = RepServMessages::Instance();

#ifdef UNICODE
    m_pIndexDefResult    = new ODBC_WResult ( odbcConnection );
    m_pIndexColDefResult = new ODBC_WResult ( odbcConnection );
#endif
#ifndef UNICODE
    m_pIndexDefResult    = new ODBC_AResult ( odbcConnection );
    m_pIndexColDefResult = new ODBC_AResult ( odbcConnection );
#endif

    m_pIndexColDefODBC   = new IndexColumnDefODBC( odbcConnection );

    m_pIndexColDefODBC->setCatalog(getCatalog());
    m_pIndexColDefODBC->setIndexDefODBC( this );

}
/*===========================================================================*
 *     IndexDefODBC::~IndexDefODBC
 *===========================================================================*/
IndexDefODBC::~IndexDefODBC()
{
    ROUTINE_DBG_MEO00 ("IndexDefODBC::~IndexDefODBC()");

    if( m_pIndexColDefODBC != NULL  )
    {
        delete m_pIndexColDefODBC;
        m_pIndexColDefODBC = NULL;
    }
    /*-------------------------------------------------------------------------*
     * Remove ResultSet
     *-------------------------------------------------------------------------*/
    if ( m_pIndexDefResult != NULL )
    {
        delete m_pIndexDefResult;
        m_pIndexDefResult = NULL;
    }
    if ( m_pIndexColDefResult != NULL )
    {
        delete m_pIndexColDefResult;
        m_pIndexColDefResult = NULL;
    }

    /*-------------------------------------------------------------------------*
     * Remove all columns
     *-------------------------------------------------------------------------*/
    IndexColumnDef *pCol = (IndexColumnDef*)m_IndexColumns.first();

    while (pCol != NULL)
    {
        delete pCol;
        pCol = (IndexColumnDef*)m_IndexColumns.next();
    }
    m_IndexColumns.clear();
}
/*===========================================================================*
 *     IndexDefODBC::initResultSet
 *===========================================================================*/
SAPDB_Bool IndexDefODBC::initResultSet ( sapdbwa_DBC & odbcConnection )
{
    SAPDB_Bool rcBool = SAPDB_FALSE;

    if ( odbcConnection.HDBC() && odbcConnection.HENV() )
    {
        StudioOAL_String  indexDef;
        StudioOAL_String  indexColDef;
        StudioOAL_String  whereClauseIndexDef;
        StudioOAL_String  whereClauseIndexColDef;
        StudioOAL_String  orderClauseIndexColDef;

        //<create_index_statement> ::=
        // CREATE [UNIQUE] INDEX <index_name> ON <table_name> (<column_name> [ASC | DESC],...)


        // INDEXES: OWNER, TABLENAME, INDEXNAME, TYPE, CREATEDATE, CREATETIME, INDEX_USED, DISABLED, COMMENT
        /*-------------------------------------------------------------------------*
         * Prepare the WHERE clause for the SQL statements.
         *-------------------------------------------------------------------------*/
        whereClauseIndexDef  = "WHERE  owner     = ? "
                               " AND   tablename = ? "
                               " AND DISABLED    = 'NO' ";

        indexDef  = "SELECT TYPE,INDEXNAME, COMMENT FROM DOMAIN.INDEXES ";
        indexDef += whereClauseIndexDef;


        // INDEXCOLUMNS: OWNER, TABLENAME, INDEXNAME, TYPE, COLUMNNAME, SORT, COLUMNNO, DATATYPE, LEN, DISTINCTVALUES,
        //               PAGECOUNT, AVGLISTLENGTH, CREATEDATE, CREATETIME, INDEX_USED, DISABLED, COMMENT
        /*-------------------------------------------------------------------------*
         * Prepare the WHERE clause for the SQL statements.
         *-------------------------------------------------------------------------*/
        whereClauseIndexColDef = " WHERE owner     = ?"
                                 " AND   tablename = ?"
                                 " AND   indexname = ?"
                                 " AND DISABLED    = 'NO'";
        orderClauseIndexColDef = " ORDER BY COLUMNNO ";

        indexColDef  = "SELECT OWNER,TABLENAME,INDEXNAME,TYPE,COLUMNNAME,SORT,COMMENT";
                       " FROM DOMAIN.INDEXCOLUMNS ";
        indexColDef += whereClauseIndexColDef;
        indexColDef += orderClauseIndexColDef;


        // SQLAllocStmt     SQLPrepare
        // SQLNumParams     SQLDescribeParam SQLBindParameter
        // SQLNumResultCols SQLDescribeCol   SQLBindCol

        rcBool = m_pIndexDefResult->init( indexDef );
        if (SAPDB_TRUE == rcBool)
        {
            rcBool = m_pIndexColDefResult->init( indexColDef );
        }
        if (SAPDB_FALSE == rcBool)
        {
            getCatalog()->getError()->setErrorText(getErrorText() );
            getCatalog()->getError()->isError     (isError     () );
        }
    } // if (  odbcConnection.HDBC() && odbcConnection.HENV() )

    return rcBool;
}
/*===========================================================================*
 *     IndexDefODBC::reset()
 *===========================================================================*/
void IndexDefODBC::reset()
{
    ROUTINE_DBG_MEO00 ("IndexDefODBC::reset()");

    /*-------------------------------------------------------------------------*
     * Remove all columns
     *-------------------------------------------------------------------------*/
    IndexColumnDef *pCol = (IndexColumnDef *)m_IndexColumns.first();

    while (pCol != NULL)
    {
        delete pCol;
        pCol = (IndexColumnDef* )m_IndexColumns.next();
    }
    m_IndexColumns.clear();
}
/*===========================================================================*
 *     IndexDefODBC::readIndexDefinitionDirect
 *===========================================================================*/
SAPDB_Bool IndexDefODBC::readIndexDefinitionDirect(StudioOAL_String owner,
                                                   StudioOAL_String tablename)
{
    ROUTINE_DBG_MEO00 ("IndexDefODBC::readIndexDefinitionDirect ()");

    StudioOAL_String  record[3];

    // INDICES
    if ( !readResult( record ) )
    {
        return SAPDB_FALSE;
    }

    m_owner     = owner;
    m_tableName = tablename;
    m_indexType = record[0];
    m_indexName = record[1];
    m_comment   = record[2];

    // INDEXCOLUMNS
    readIndexColDefinitionDirect( owner, tablename, m_indexName );

    return SAPDB_TRUE;
}
/*===========================================================================*
 *    IndexDefODBC::readIndexColDefinitionDirect
 *===========================================================================*/
SAPDB_Bool
IndexDefODBC::readIndexColDefinitionDirect(StudioOAL_String owner,
                                           StudioOAL_String tablename,
                                           StudioOAL_String indexname )
{
    ROUTINE_DBG_MEO00 ("IndexDefODBC::readIndexColumnDefinitionDirect ()");

    SAPDB_Bool isOk = SAPDB_FALSE;

    StudioOAL_String  indexColDef;
    StudioOAL_String  whereClause;
    StudioOAL_String  orderClause;
    StudioOAL_String  comment;
    StudioOAL_String  record[7];

    reset();

    m_owner     = owner;
    m_tableName = tablename;

    /*-------------------------------------------------------------------------*
     * Prepare the WHERE clause for the SQL statements.
     *-------------------------------------------------------------------------*/
    whereClause   = "WHERE tablename = '";
    whereClause  += tablename;
    whereClause  += "' AND indexname = '";
    whereClause  += indexname;
    whereClause  += "' AND DISABLED = 'NO'";

    if (SAPDB_FALSE == owner.Empty())
    {
        whereClause  += " AND owner = '";
        whereClause  += owner;
        whereClause  += "'";
    }

    orderClause = " ORDER BY COLUMNNO ";

    // INDEXCOLUMNS : OWNER, TABLENAME, INDEXNAME, TYPE, COLUMNNAME, SORT, COLUMNNO, DATATYPE, LEN, DISTINCTVALUES,
    //                PAGECOUNT, AVGLISTLENGTH, CREATEDATE, CREATETIME, INDEX_USED, DISABLED, COMMENT

    //<create_index_statement> ::=
    // CREATE [UNIQUE] INDEX <index_name> ON <table_name> (<column_name> [ASC | DESC],...)
    indexColDef  = "SELECT OWNER, TABLENAME, INDEXNAME, TYPE, COLUMNNAME, SORT, COMMENT FROM DOMAIN.INDEXCOLUMNS ";
    indexColDef += whereClause;
    indexColDef += orderClause;

    IndexColumnDef *pIndexColDef;

    if ( m_pIndexColDefODBC->executeSQL( indexColDef ) )
    {
        while ( m_pIndexColDefODBC->readResult(record) )
        {
            // INDEXCOLUMNS
            pIndexColDef = addColumn();

            pIndexColDef->setOwner    (record[0]);
            pIndexColDef->setTableName(record[1]);
            pIndexColDef->setIndexName(record[2]);
            pIndexColDef->setIndexType(record[3]);

            pIndexColDef->m_column.setName( record[4]);
            pIndexColDef->setSorting( record[5]);

            // INDEXCOLUMNS : COMMENT
            if ( record[6].Compare( NULL_VALUE ) != 0 ) // !m_pIndexColDefResult->getColumn(7)->getColIsNull()
            {
                pIndexColDef->setComment ( record[6] );
            }

            isOk = SAPDB_TRUE;
        }
    }
    return isOk;
}
/*===========================================================================*
 *     IndexDefODBC::setIndexParameters
 *===========================================================================*/
SAPDB_Bool IndexDefODBC::setIndexParameters( StudioOAL_String owner ,
                                             StudioOAL_String tablename )
{
    ROUTINE_DBG_MEO00 ("IndexDefODBC::setIndexParameters()");

    SAPDB_Bool rcBool = SAPDB_TRUE;

    StudioOAL_String  indexDef[2];

    reset();

    m_owner     = indexDef[0] = owner;
    m_tableName = indexDef[1] = tablename;

    /*-------------------------------------------------------------------------*
     * Load the index definitions from the database.                           *
     *-------------------------------------------------------------------------*/
    rcBool = m_pIndexDefResult->setParamValues( indexDef );
    if (SAPDB_TRUE == rcBool)
    {
        /*-------------------------------------------------------------------------*
         * Execute the prepared statement                                          *
         *-------------------------------------------------------------------------*/
        // SQLExecute
        rcBool = m_pIndexDefResult->execute();
    }
    if (SAPDB_FALSE == rcBool)
    {
        getCatalog()->getError()->setErrorText(m_pIndexDefResult->getErrorText());
        getCatalog()->getError()->isError     (m_pIndexDefResult->isError());
    }
    return rcBool;
}
/*===========================================================================*
 *     IndexDefODBC::readIndexDefinition
 *===========================================================================*/
SAPDB_Bool IndexDefODBC::readIndexDefinition( StudioOAL_String owner ,
                                              StudioOAL_String tablename )
{
    ROUTINE_DBG_MEO00 ("IndexDefODBC::readIndexDefinition()");

    SAPDB_Bool isOk = SAPDB_FALSE;

    StudioOAL_String  record[6];

    // SQlFetch SQLFreeStatement
    if ( readIndexDefResult (record) )
    {
        m_indexType = record[0];
        m_indexName = record[1];
        m_comment   = record[2];

        isOk = readIndexColDefinition( owner, tablename, m_indexName );
    }
    return isOk;
}
/*===========================================================================*
 *     IndexDefODBC::readIndexDefResult
 *===========================================================================*/
SAPDB_Bool IndexDefODBC::readIndexDefResult(StudioOAL_String *record)
{
    SAPDB_Int2       i;
    StudioOAL_String s;

    /*-------------------------------------------------------------------------*
     * Fetch the next record.                                                  *
     *-------------------------------------------------------------------------*/
    // SQLFetch
    if (!m_pIndexDefResult->fetch())
    {
        m_pIndexDefResult->closeCursor();
        return SAPDB_FALSE;
    }

    /*-------------------------------------------------------------------------*
     * If there are no more records, reading the result is finished.           *
     *-------------------------------------------------------------------------*/
    if (m_pIndexDefResult->noRows())
    {
        m_pIndexDefResult->closeCursor();
        return SAPDB_FALSE;
    }

    /*-------------------------------------------------------------------------*
     * Read all columns of the fetched record.                                 *
     *-------------------------------------------------------------------------*/
    for (i = 0; i < m_pIndexDefResult->getColCount(); i++)
    {
        m_pIndexDefResult->bindColumnDataAsString (i + 1, s);
        record[i] = s;
    }

    return SAPDB_TRUE;
}
/*===========================================================================*
 *    IndexDefODBC::readIndexColDefinition
 *===========================================================================*/
SAPDB_Bool IndexDefODBC::readIndexColDefinition( StudioOAL_String owner    ,
                                                 StudioOAL_String tablename,
                                                 StudioOAL_String indexname )
{
    ROUTINE_DBG_MEO00 ("IndexDefODBC::readIndexColDefinition()");

    SAPDB_Bool rcBool = SAPDB_TRUE;
    
    reset();

    StudioOAL_String record[7];
    StudioOAL_String indexColDef[3];

    indexColDef[0] = owner;
    indexColDef[1] = tablename;
    indexColDef[2] = indexname;

    /*-------------------------------------------------------------------------*
     * Load the indexcolumn definitions from the database.                     *
     *-------------------------------------------------------------------------*/
    rcBool = m_pIndexColDefResult->setParamValues( indexColDef );
    if (SAPDB_TRUE == rcBool)
    {
        /*-------------------------------------------------------------------------*
         * Execute the prepared statement                                          *
         *-------------------------------------------------------------------------*/
        // SQLExecute
        rcBool = m_pIndexColDefResult->execute();
    }

    if (SAPDB_FALSE == rcBool)
    {
        getCatalog()->getError()->setErrorText(getErrorText());
        getCatalog()->getError()->isError(isError());
        rcBool = SAPDB_FALSE;
    }
    else
    {
        IndexColumnDef *pIndexColDef;

        // SQlFetch SQLFreeStatement
        while( readIndexColDefResult(record) )
        {
            // INDEXCOLUMNS
            pIndexColDef = addColumn();

            pIndexColDef->setOwner    ( record[0]);
            pIndexColDef->setTableName( record[1]);
            pIndexColDef->setIndexName( record[2]);
            pIndexColDef->setIndexType( record[3]);
            pIndexColDef->m_column.setName( record[4]);
            pIndexColDef->setSorting      ( record[5]);

            // INDEXCOLUMNS : COMMENT
            if ( !m_pIndexColDefResult->getColumn(7)->getColIsNull() )
            {
                pIndexColDef->setComment( record[6] );
            }
        }
    }

    return rcBool;
}
/*===========================================================================*
 *     IndexDefODBC::readIndexColDefResult
 *===========================================================================*/
SAPDB_Bool IndexDefODBC::readIndexColDefResult(StudioOAL_String *record)
{
    SAPDB_Int2       i;
    StudioOAL_String s;

    /*-------------------------------------------------------------------------*
     * Fetch the next record.                                                  *
     *-------------------------------------------------------------------------*/
    // SQLFetch
    if (!m_pIndexColDefResult->fetch())
    {
        m_pIndexColDefResult->closeCursor();
        return SAPDB_FALSE;
    }

    /*-------------------------------------------------------------------------*
     * If there are no more records, reading the result is finished.           *
     *-------------------------------------------------------------------------*/
    if (m_pIndexColDefResult->noRows())
    {
        m_pIndexColDefResult->closeCursor();
        return SAPDB_FALSE;
    }

    /*-------------------------------------------------------------------------*
     * Read all columns of the fetched record.                                 *
     *-------------------------------------------------------------------------*/
    for (i = 0; i < m_pIndexColDefResult->getColCount(); i++)
    {
        m_pIndexColDefResult->bindColumnDataAsString (i + 1, s);
        record[i] = s;
    }

    return SAPDB_TRUE;
}
/*===========================================================================*
 *     IndexDefODBC::writeIndexDefinition
 *===========================================================================*/
SAPDB_Bool IndexDefODBC::writeIndexDefinition ()
{
    ROUTINE_DBG_MEO00 ("IndexDefODBC::writeIndexDefinition ()");

    StudioOAL_String tableIndices;

    if (emExtractAsDDL_els31 == getCatalog()->getExtrMode())
    {
        if ( m_IndexColumns.size() > 0 )
        {
            tableIndices = SEPARATOR;

            if (cpr_kind_internal != getCatalog()->getSqlMode())
            {
                SAPDB_Int4 iSqlMode = getCatalog()->getSqlMode();
                tableIndices += pszSQLModeString_ls31[iSqlMode];
                tableIndices += SEPARATOR;
            }

            tableIndices += genCreateStatement();

            // COMMENT ON INDEX <index_name> ON <table_name> IS '  '
            if ( (!m_comment.Empty()) && (0 != m_comment.Compare(NULL_VALUE)) )
            {
                tableIndices += NEW_LINE;
                tableIndices += SEPARATOR;
                tableIndices += "COMMENT ON INDEX \"";
                tableIndices += m_indexName;
                tableIndices += "\" ON \"";

                if (SAPDB_FALSE == m_owner.Empty())
                {
                    tableIndices += m_owner;
                    tableIndices += "\".\"";
                }

                tableIndices += m_tableName;
                tableIndices += "\" IS '";
                tableIndices += EscapeSingleQuotes(m_comment);
                tableIndices += "'";
            }
        }
    }
    else if (emExtractAsXML_els31 == getCatalog()->getExtrMode())
    {
        tableIndices += genCreateXML();
    }

    return writeLine(tableIndices);
}
/*===========================================================================*
 *     IndexDefODBC::addColumn
 *===========================================================================*/
IndexColumnDef * IndexDefODBC::addColumn ()
{
    /*-------------------------------------------------------------------------*
     * Create a new column object.                                             *
     *-------------------------------------------------------------------------*/
    IndexColumnDef *pIndexColDef = new IndexColumnDef( );

    /*-------------------------------------------------------------------------*
     * Append the new column object to the columns list.                       *
     *-------------------------------------------------------------------------*/
    m_IndexColumns.append(pIndexColDef);

    return pIndexColDef;
}
/*===========================================================================*
 *     IndexDefODBC::genCreateStatement
 *===========================================================================*/
StudioOAL_String IndexDefODBC::genCreateStatement ( )
{
    ROUTINE_DBG_MEO00 ("IndexDefODBC::genCreateStatement ()");

    StudioOAL_String  index;
    IndexColumnDef   *pCol;

    // INDEXCOLUMNS: OWNER, TABLENAME, INDEXNAME, TYPE, COLUMNNAME, SORT, COLUMNNO, DATATYPE, LEN, DISTINCTVALUES,
    //               PAGECOUNT, AVGLISTLENGTH, CREATEDATE, CREATETIME, INDEX_USED, DISABLED, COMMENT

    //<create_index_statement> ::=
    // CREATE [UNIQUE] INDEX <index_name> ON <table_name> (<column_name> [ASC | DESC],...)

    index  = "CREATE ";
    index += m_indexType;
    index += " INDEX \"";
    index += m_indexName;
    index += "\" ON \"";
    if (SAPDB_FALSE == m_owner.Empty())
    {
        index += m_owner;
        index += "\".\"";
    }
    index += m_tableName;
    index += "\" (";

    SAPDB_Int4 iIndexColSize  = m_IndexColumns.size();
    SAPDB_Int4 iIndexColCount = 0;

    // CREATE [UNIQUE] INDEX <index_name> ON <table_name> (<column_name> [ASC | DESC],...)  //!! ToDo
    for ( pCol = (IndexColumnDef *)m_IndexColumns.first(); pCol; pCol = (IndexColumnDef *)m_IndexColumns.next())
    {
        iIndexColCount++;

        index += pCol->genCreateStatement();

        if ( iIndexColCount != iIndexColSize )
        {
            index += ",";
        }
        else
        {
            index += " )";
        }
    }
    return index;
}
/*===========================================================================*
 *     IndexDefODBC::genCreateXML()
 *===========================================================================*/
StudioOAL_String IndexDefODBC::genCreateXML()
{
   ROUTINE_DBG_MEO00 ("IndexDefODBC::genCreateXML()");

   StudioOAL_String  index;

   IndexColumnDef  *pCol;

   // INDEXCOLUMNS : OWNER,TABLENAME,INDEXNAME,TYPE,COLUMNNAME,SORT,COLUMNNO,DATATYPE,LEN,DISTINCTVALUES,
   //                PAGECOUNT,AVGLISTLENGTH,CREATEDATE,CREATETIME,INDEX_USED, DISABLED, COMMENT

   //<create_index_statement> ::=
   // CREATE [UNIQUE] INDEX <index_name> ON <table_name> (<column_name> [ASC | DESC],...)

   // <dbm:DeployedTableIndices>
   index  = "<dbm:Index oim:id = \"_";
   index += string2XML( m_owner );
   index += ".";
   index += string2XML( m_tableName );
   index += ".";
   index += string2XML( m_indexName );
   index += "\" oim:href = \"#_";
   index += string2XML( m_owner );
   index += ".";
   index += string2XML( m_tableName );
   index += "\" dbm:name = \"";
   index += string2XML( m_indexName );
   index += "\" dbm:comment = \"";
   m_comment.Compare( NULL_VALUE ) != 0 ? index += string2XML( m_comment ) : index += string2XML("");
   index += "\" dbm:IsUnique = \"";
   m_indexType.Compare("UNIQUE")   == 0 ? index += "1\""     : index += "0\"";
   index += ">\n";

   index +=  "<dbm:IndexIndexColumns>\n";

   for ( pCol   = (IndexColumnDef *)m_IndexColumns.first();pCol; pCol = (IndexColumnDef *)m_IndexColumns.next())
         index += pCol->genCreateXML();

   index +=  "</dbm:IndexIndexColumns>\n";
   index +=  "</dbm:Index>\n";
   // </dbm:DeployedTableIndices>

   return index;
}
/*===========================================================================*
 *    IndexColumnDefODBC::IndexColumnDefODBC
 *===========================================================================*/
IndexColumnDefODBC::IndexColumnDefODBC( sapdbwa_DBC & odbcConnection , CatalogDefODBC * pCatalog  )
                   :CatalogObjectDefODBC ( odbcConnection , pCatalog )
{
}
/*===========================================================================*
 *     IndexColumnDefODBC::IndexColumnDefODBC
 *===========================================================================*/
IndexColumnDefODBC::IndexColumnDefODBC(  sapdbwa_DBC & odbcConnection )
                   :CatalogObjectDefODBC ( odbcConnection )
{
}
/*===========================================================================*
 *    IndexColumnDefODBC::~IndexColumnDefODBC
 *===========================================================================*/
IndexColumnDefODBC::~IndexColumnDefODBC()
{
}
/*===========================================================================*
 *     IndexColumnDefODBC::initResultSet
 *===========================================================================*/
SAPDB_Bool IndexColumnDefODBC::initResultSet ( sapdbwa_DBC & odbcConnection )
{
  ROUTINE_DBG_MEO00 ("IndexColumnDefODBC::initResult ()");

  SAPDB_Bool isOk = SAPDB_FALSE;

  if (  odbcConnection.HDBC() && odbcConnection.HENV() )
  {
        isOk = SAPDB_TRUE;
  }

  return isOk;
}
/*===========================================================================*
 *    IndexColumnDef::setComment
 *===========================================================================*/
void  IndexColumnDef::setComment ( StudioOAL_String comment )
{
      m_column.setComment(comment);
}
/*===========================================================================*
 *    IndexColumnDef::genCreateStatement
 *===========================================================================*/
StudioOAL_String IndexColumnDef::genCreateStatement()
{
     StudioOAL_String  indexColumn;

     indexColumn  = "\"";
     indexColumn += m_column.name ();
     indexColumn += "\"";
     indexColumn += "  ";
     indexColumn += m_sorting;

     return indexColumn;
}
/*===========================================================================*
 *    IndexColumnDef::genCreateXML
 *===========================================================================*/
StudioOAL_String IndexColumnDef::genCreateXML()
{
     ROUTINE_DBG_MEO00 ("IndexColumnDef::genCreateXML()");

     StudioOAL_String  indexColumn;

     // <dbm:IndexIndexColumns>

     indexColumn  =  "<dbm:IndexColumn oim:id = \"_";
     indexColumn += string2XML( m_owner );
     indexColumn += ".";
     indexColumn += string2XML( m_tableName );
     indexColumn += ".";
     indexColumn += string2XML( m_indexName );
     indexColumn += ".";
     indexColumn += string2XML( m_column.name () );
     indexColumn += "\" oim:href = \"#_";
     indexColumn += string2XML( m_owner );
     indexColumn += ".";
     indexColumn += string2XML( m_tableName );
     indexColumn += ".";
     indexColumn += string2XML( m_column.name () );
     indexColumn += "\" dbm:name = \"";
     indexColumn += string2XML( m_column.name () );
     indexColumn += "\" dbm:IsAscending = \"";
     m_sorting.Compare("ASC")   == 0 ? indexColumn += "1\"" : indexColumn += "0\"";
     indexColumn += "/>\n";

     // </dbm:IndexIndexColumns>

     return indexColumn;
}

/*===========================================================================*
 *  CatalogDefODBC::dataExtrCmdExec
 *===========================================================================*/
SAPDB_Int4 CatalogDefODBC::dataExtrCmdExec( StudioOAL_String owner, StudioOAL_String tablename )
{
    ROUTINE_DBG_MEO00 ("CatalogDefODBC::dataExtrCmdExec");

    RepServMessages *pMsgInstance = RepServMessages::Instance();
    SAPDB_Int4       rc           = errOK_els00;

    StudioOAL_String table("\"");
    table  += owner;
    table  += "\".\"";
    table  += tablename;
    table  += "\"";

    StudioOAL_String dataExtractCmd("SELECT * FROM ");
    dataExtractCmd += table;

    StudioOAL_String record[1023];

    /*-------------------------------------------------------------------------*
    * Execute the prepared statement                                          *
    *-------------------------------------------------------------------------*/
    // SQLExecute
    if ( !m_pDataExtractDefODBC->executeSQL( dataExtractCmd ) )
    {
        sapdb_Error    *pError    = getError();

        StudioOAL_String errorText = m_pDataExtractDefODBC->m_pDataExtractDefResult->getErrorText();
        SAPDB_Bool      isError   = m_pDataExtractDefODBC->m_pDataExtractDefResult->isError     ();

        pError->setErrorText( errorText );
        pError->isError     ( isError   );

        return SAPDB_FALSE;
    }

    while ( m_pDataExtractDefODBC->readResult ( record ) )
        m_pDataExtractDefODBC->writeResult( record );

    return rc;
}
/*===========================================================================*
 *    DataExtractDefODBC::DataExtractDefODBC
 *===========================================================================*/
DataExtractDefODBC::DataExtractDefODBC   ( sapdbwa_DBC & odbcConnection , CatalogDefODBC * pCatalog  )
                   :CatalogObjectDefODBC ( odbcConnection , pCatalog )
{
    ROUTINE_DBG_MEO00 ("DataExtractDefODBC::DataExtractDefODBC()");

#ifdef UNICODE
    m_pDataExtractDefResult  = new ODBC_WResult ( odbcConnection );
#endif
#ifndef UNICODE
    m_pDataExtractDefResult  = new ODBC_AResult ( odbcConnection );
#endif

    initResultSet( odbcConnection );

}
/*===========================================================================*
 *    DataExtractDefODBC::DataExtractDefODBC
 *===========================================================================*/
DataExtractDefODBC::DataExtractDefODBC( sapdbwa_DBC & odbcConnection  )
                   :CatalogObjectDefODBC ( odbcConnection )
{
    ROUTINE_DBG_MEO00 ("DataExtractDefODBC::DataExtractDefODBC()");

#ifdef UNICODE
    m_pDataExtractDefResult  = new ODBC_WResult ( odbcConnection );
#endif
#ifndef UNICODE
    m_pDataExtractDefResult  = new ODBC_AResult ( odbcConnection );
#endif

}
/*===========================================================================*
 *    DataExtractDefODBC::DataExtractDefODBC
 *===========================================================================*/
DataExtractDefODBC::~DataExtractDefODBC()
{
    ROUTINE_DBG_MEO00 ("DataExtractDefODBC::~DataExtractDefODBC()");

    if ( m_pDataExtractDefResult != NULL )
    {
        delete m_pDataExtractDefResult;
        m_pDataExtractDefResult = NULL;
    }
}
/*===========================================================================*
 *     DataExtractDefODBC::initResultSet
 *===========================================================================*/
SAPDB_Bool DataExtractDefODBC::initResultSet ( sapdbwa_DBC & odbcConnection )
{
    ROUTINE_DBG_MEO00 ("DataExtractDefODBC::initResultSet()");

    RepServMessages *pMsgInstance = RepServMessages::Instance();

    SAPDB_Bool isOk = SAPDB_TRUE;

    return isOk;
}
/*===========================================================================*
 *    DataExtractDefODBC::readResult
 *===========================================================================*/
SAPDB_Bool DataExtractDefODBC::readResult ( StudioOAL_String *record  )
{
    ROUTINE_DBG_MEO00 ("DataExtractDefODBC::readResult ()");

    return CatalogObjectDefODBC::readResult( record);
}
/*===========================================================================*
 *    DataExtractDefODBC::writeResult
 *===========================================================================*/
SAPDB_Bool DataExtractDefODBC::writeResult( StudioOAL_String *record  )
{
    ROUTINE_DBG_MEO00 ("DataExtractDefODBC::writeResult ()");

    SAPDB_Int4 nCol;

    for ( nCol = 0; nCol < getColCount(); nCol++ )
          writeLine ( record[nCol] );

    return SAPDB_TRUE;
}
/*===========================================================================*
 *     END
 *===========================================================================*/




/*===========================================================================*
 *    class TableExtractDefODBC 
 *===========================================================================*/
/*===========================================================================*
 *    TableExtractDefODBC::TableExtractDefODBC
 *===========================================================================*/
TableExtractDefODBC::TableExtractDefODBC( sapdbwa_DBC & odbcConnection, tls00_PartSpec *pPartSpec )
                      :CatalogObjectDefODBC ( odbcConnection )
{
    ROUTINE_DBG_MEO00 ("TableExtractDefODBC::TableExtractDefODBC()");

    m_nTableCount = 0;

    unsigned char *pUserName = new unsigned char[KNL_IDENTIFIER_SIZE + 1];
    if (SAPDB_TRUE == odbcConnection.GetUser(pUserName, KNL_IDENTIFIER_SIZE))
    {
//      m_username.ConvertFromASCII_Latin1(destBeg,destBeg+strlen(destBeg)); 
            
        m_username = pUserName;
    }

    delete [] pUserName;


    m_pPartSpec = new tls00_PartSpec;
    memcpy(m_pPartSpec, pPartSpec, sizeof(tls00_PartSpec));
    
}
// TableExtractDefODBC::TableExtractDefODBC


/*===========================================================================*
 *    TableExtractDefODBC::~TableExtractDefODBC
 *===========================================================================*/
TableExtractDefODBC::~TableExtractDefODBC()
{
    ROUTINE_DBG_MEO00 ("TableExtractDefODBC::~TableExtractDefODBC()");
    
    if (NULL != m_pPartSpec)
    {
        delete m_pPartSpec;
        m_pPartSpec = NULL;
    }
}
// TableExtractDefODBC::~TableExtractDefODBC()


/*===========================================================================*
 *    TableExtractDefODBC::GetOwnerTablePair
 *===========================================================================*/
SAPDB_Bool
TableExtractDefODBC::GetOwnerTablePair( sapdbwa_DBC & odbcConnection,
                                       StudioOAL_String *pUserName,
                                       StudioOAL_String *pTableName)
{
    ROUTINE_DBG_MEO00 ("TableExtractDefODBC::GetOwnerTablePair()");

    SAPDB_Bool rc = SAPDB_FALSE;

    StudioOAL_String  record[2];        // fitting for owner name and table name

    if ( (SAPDB_FALSE == m_pPartSpec->psSelAll_ls00) && (SAPDB_FALSE == m_pPartSpec->psSelUser_ls00) )
    {
        if (m_nTableCount > 0)
        {
            *pUserName = m_username;
            pTableName->ConvertFromASCII_Latin1(m_pPartSpec->psTabName_ls00.asCharp(),
                                               m_pPartSpec->psTabName_ls00.asCharp() + _tcslen(m_pPartSpec->psTabName_ls00.asCharp()));
            --m_nTableCount;
            rc = SAPDB_TRUE;
        }
    }
    else
    {
        rc = readResult(record);
        if (SAPDB_TRUE == rc)
        {
            *pUserName = record[0];
            *pTableName = record[1];
        }
    }

    return rc;
}
// TableExtractDefODBC::GetOwnerTablePair()


/*===========================================================================*
 *    TableExtractDefODBC::initResultSet
 *===========================================================================*/
SAPDB_Bool
TableExtractDefODBC::initResultSet( sapdbwa_DBC & odbcConnection )
{
    ROUTINE_DBG_MEO00 ("TableExtractDefODBC::initResultSet()");

    SAPDB_Bool rc = SAPDB_TRUE;


    // make sure we have a connection
    if ( odbcConnection.HDBC() && odbcConnection.HENV() ) // sind die sonst NULL???
    {
        StudioOAL_String  catalogDeleteCmd;
        StudioOAL_String  catalogSaveCmd;




        //*
        //*     Drop temp 'save catalog' table
        //*
        catalogDeleteCmd = "DROP TABLE TEMP.LOA_SAVE_CATALOG";
        rc = executeSQL(catalogDeleteCmd);
        if (SAPDB_TRUE == rc)
        {
            // create temp catalog table
            // CATALOG 7.3   : LEVEL,CMD_NO,SEQ_NO,DT_FORMAT,OWNER,TABLENAME,ERROR_CODE,SQL_CMD,TYPE,TABLETYPE,CMD

            //*
            //*     Save catalog TABLE/USER/ALL
            //*
            if ( (SAPDB_FALSE == m_pPartSpec->psSelAll_ls00) && (SAPDB_FALSE == m_pPartSpec->psSelUser_ls00) )
            {
                // Extracting single table means simply to set table count
                m_nTableCount = 1;
            }
            else
            {
                if (SAPDB_TRUE == m_pPartSpec->psSelAll_ls00)   // save catalog all
                {
                    catalogSaveCmd = "SAVE CATALOG ALL INTO TEMP.LOA_SAVE_CATALOG";
                }
                else                                            // save catalog user
                {
                    catalogSaveCmd = "SAVE CATALOG USER INTO TEMP.LOA_SAVE_CATALOG";
                }

                rc = executeSQL(catalogSaveCmd);
                if (SAPDB_TRUE == rc)
                {
                    StudioOAL_String deleteCmd("DELETE FROM TEMP.LOA_SAVE_CATALOG WHERE (TABLENAME) IN "
                                               "(SELECT TABLENAME FROM DOMAIN.TABLES WHERE TYPE = 'SYSTEM' "
                                               "AND TABLETYPE = 'TABLE')");
                    rc = executeSQL(deleteCmd);
                    if (SAPDB_TRUE == rc)
                    {
                        //*
                        //*     Get some statistics
                        //*
                        StudioOAL_String  selectCmd;
                        StudioOAL_String  record[1];

                        selectCmd = "SELECT COUNT (*) FROM TEMP.LOA_SAVE_CATALOG WHERE SQLCMD='N'";
                        rc = executeSQL(selectCmd);
                        if (SAPDB_TRUE == rc)
                        {
                            readResult(record);
                            m_nTableCount  = AtoI( record[0] );

                            //*
                            //*     Finally get the data
                            //*

                            // LEVEL, CMD_NO, SEQ_NO, DT_FORMAT, OWNER, TABLENAME , ERROR_CODE, SQLCMD, TYPE, TABLETYPE, CMD

                            selectCmd  = "SELECT OWNER,TABLENAME FROM TEMP.LOA_SAVE_CATALOG WHERE "
                                         "SQLCMD='N' ORDER BY CMD_NO";
                            rc = executeSQL(selectCmd);
                        }
                    }

                }   // end if (SAPDB_TRUE == rc = executeSQL(catalogSaveCmd))
                
            }   // end else of if ( (SAPDB_FALSE == m_pPartSpec->psSelAll_ls00) && (SAPDB_FALSE == m_pPartSpec->psSelUser_ls00) )   
        
        }   // end else if SAPDB_FALSE == rc = executeSQL(catalogDeleteCmd)
    
    }   // end if (odbcConnection.HDBC() && odbcConnection.HENV())  // sind die sonst NULL???   

    // Check for error while executing
    if (SAPDB_FALSE == rc)
    {
        SAPDB_Bool       Error               = isError();
        SAPDB_Long       NativeError         = getNativeError();
        SAPDB_Int4       ErrorPos            = getErrorPos();
        SQLRETURN        ReturnCode          = getReturnCode();
        StudioOAL_String ErrorText           = getErrorText();
        StudioOAL_String NativeErrorAsString = getNativeErrorAsString();
    }

    return rc;
}
// TableExtractDefODBC::initResultSet( sapdbwa_DBC & odbcConnection )
